/*!
 * 
 *
 * (c) Balsamiq - All rights reserved.
 *
 *
 */
var MacOSExporter;
/******/ (() => { // webpackBootstrap
/******/ 	var __webpack_modules__ = ({

/***/ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchive.js":
/*!************************************************************!*\
  !*** ./node_modules/@balsamiq/bmpr/lib/BalsamiqArchive.js ***!
  \************************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


var Consts = __webpack_require__(/*! ./BalsamiqArchiveConstants */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js");

var CommentPermissionError = "User has no permission to modify or delete this comment";

// BalsamiqArchive class
// ---------------------

//Documentation: https://balsamiq.atlassian.net/wiki/display/intranet/Balsamiq+Archive+Master+API+list

var BalsamiqArchive = function (p_dbDriverInstance, userInfo, log) {
    this.dbDriver = p_dbDriverInstance;
    if (log) {
        this.log = log;
    } else {
        // dummy
        this.log = {
            info: function() {}
        }
    }

    // userinfo.name is the actual user ID
    if (userInfo && userInfo.name) {
        this.userInfo = userInfo;
    }
    else {
        this.userInfo = Consts.defaultUserInfo
    }

    this.userRole = Consts.Role.ROLE_EDITOR;

    this.busy = false;      // it is true when a query is running. trying to run a second query will throw an Error.
};

BalsamiqArchive.prototype.setUserRole = function (role) {
    this.userRole = role;
};

BalsamiqArchive.prototype.setUserInfoName = function (name) {
    this.userInfo.name = name
};

BalsamiqArchive.prototype.getTablesName = function () {
    return ["INFO", "BRANCHES", "RESOURCES", "THUMBNAILS", "USERS", "COMMENTS"] ;
};

BalsamiqArchive.prototype.dump = function (branchID, callback) {
    this._enterCall();

    var dump = {};
    dump.Info = {};
    this._internalGetArchiveData(Consts.Info.SchemaVersion, Consts.Info.SchemaVersion, false, function (obj) {
        if (obj.error) {
            this._exitCall();
            callback(obj);
        } else {
            dump.Info[Consts.Info.SchemaVersion] = obj[Consts.Info.SchemaVersion];
            this._internalGetArchiveData(Consts.Info.ArchiveFormat, "format", false, function (obj) {
                if (obj.error) {
                    this._exitCall();
                    callback(obj);
                } else {
                    dump.Info[Consts.Info.ArchiveFormat] = obj.format;
                    this._getArchiveRevision( function(revision) {
                        dump.Info[Consts.Info.ArchiveRevision] = revision;
                        dump.Info[Consts.Info.ArchiveRevisionUUID] = "";
                        this._internalGetArchiveData(Consts.Info.ArchiveAttributes, "attributes", true, function (obj) {

                            if (obj.error) {
                                this._exitCall();
                                callback(obj);
                            } else {
                                dump.Info[Consts.Info.ArchiveAttributes] = obj.attributes;

                                if (branchID == Consts.Branch.NoBranch) {
                                    this._exitCall();
                                    callback(obj);
                                } else {
                                    this._internalGetBranches(function (obj) {
                                        if (obj.error) {
                                            this._exitCall();
                                            callback(obj);
                                        } else {
                                            dump.Branch = obj.branches;

                                            this._internalGetAllResourcesAttributesForBranch(branchID, function (obj) {
                                                var resourcesIDS = [], i;

                                                if (obj.error) {
                                                    this._exitCall();
                                                    callback(obj);
                                                } else {
                                                    dump.Resources = obj.resources;

                                                    for (i = 0; i < dump.Resources.length; i++) {
                                                        resourcesIDS.push(dump.Resources[i].ID);
                                                    }

                                                    this._internalGetResourcesData(resourcesIDS, branchID, function (obj) {
                                                        if (obj.error) {
                                                            this._exitCall();
                                                            callback(obj);
                                                        } else {

                                                            for (i = 0; i < dump.Resources.length; i++) {
                                                                dump.Resources[i].DATA = obj.data[dump.Resources[i].ID + "|" + dump.Resources[i].BRANCHID];
                                                            }

                                                            this._internalGetThumbnails(function (obj) {
                                                                if (obj.error) {
                                                                    this._exitCall();
                                                                    callback(obj);
                                                                } else {
                                                                    dump.Thumbnails = obj.thumbnails;

                                                                    // more: get users and comments
                                                                    this._internalGetArchiveUsersList(function (obj) {
                                                                        if (obj.error) {
                                                                            this._exitCall();
                                                                            callback(obj);
                                                                        } else {
                                                                            dump.Users = obj.users;

                                                                            this._internalGetCommentsToc(function (obj) {

                                                                                var commentIDs = [], i, j;

                                                                                if (obj.error) {
                                                                                    this._exitCall();
                                                                                    callback(obj);
                                                                                } else {
                                                                                    dump.Comments = obj.comments;

                                                                                    for (i = 0; i < dump.Comments.length; i++) {
                                                                                        commentIDs.push(dump.Comments[i].ID);
                                                                                    }

                                                                                    this._internalGetCommentsData(commentIDs, false, function (obj) {
                                                                                        for (i = 0; i < dump.Comments.length; i++) {
                                                                                            // important: the _internalGetCommentsData DO NOT return data in the same order commentIDs are sent as input!!
                                                                                            for (j=0; j<obj.comments.length; j++) {
                                                                                                if (obj.comments[j].ID === dump.Comments[i].ID) {
                                                                                                    dump.Comments[i].DATA = obj.comments[j].DATA;
                                                                                                    break;
                                                                                                }
                                                                                            }
                                                                                        }

                                                                                        this._exitCall();
                                                                                        callback({dump: dump});
                                                                                    }.bind(this))
                                                                                }
                                                                            }.bind(this));
                                                                        }
                                                                    }.bind(this));
                                                                }
                                                            }.bind(this));
                                                        }
                                                    }.bind(this));
                                                }
                                            }.bind(this));
                                        }
                                    }.bind(this));
                                }
                            }
                        }.bind(this));
                    }.bind(this));
                }
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype.getToc = function (branchID, callback) {
    this._enterCall();
    this._internalGetToc(branchID, function (obj) {
        this._exitCall();
        callback && callback(obj);
    }.bind(this));
};

BalsamiqArchive.prototype._internalGetToc = function (branchID, callback) {

    if (!this.checkInternalCall(callback)) {
        return;
    }

    var dump = {};
    dump.Info = {};
    dump.userInfo = this.userInfo;
    // internally, we are using name but clients just want an ID
    dump.userInfo.ID = dump.userInfo.name;

    this._internalGetArchiveData(Consts.Info.SchemaVersion, Consts.Info.SchemaVersion, false, function (obj) {
        if (obj.error) {
            callback(obj);
        } else {
            dump.Info[Consts.Info.SchemaVersion] = obj[Consts.Info.SchemaVersion];

            this._internalGetArchiveData(Consts.Info.ArchiveFormat, "format", false, function (obj) {

                if (obj.error) {
                    callback(obj);
                } else {
                    dump.Info[Consts.Info.ArchiveFormat] = obj.format;

                    this._getArchiveRevision( function(archiveRevision) {

                        dump.Info[Consts.Info.ArchiveRevision] = archiveRevision;

                        this._internalGetArchiveData(Consts.Info.ArchiveRevisionUUID, "ArchiveRevisionUUID", false, function(obj) {
                            dump.Info[Consts.Info.ArchiveRevisionUUID] = obj.ArchiveRevisionUUID;
                            this._internalGetArchiveData(Consts.Info.ArchiveAttributes, "attributes", true, function (obj) {

                                if (obj.error) {
                                    callback(obj);
                                } else {
                                    dump.Info[Consts.Info.ArchiveAttributes] = obj.attributes;

                                    if (branchID == Consts.Branch.NoBranch) {
                                        callback({dump: dump});
                                    } else {
                                        this._internalGetBranches(function (obj) {
                                            if (obj.error) {
                                                callback(obj);
                                            } else {
                                                dump.Branch = obj.branches;

                                                this._internalGetAllResourcesAttributesForBranch(branchID, function(obj) {

                                                    if (obj.error) {
                                                        callback(obj);
                                                    } else {
                                                        dump.Resources = obj.resources;

                                                        this._internalGetThumbnails( function(obj) {
                                                            if (obj.error) {
                                                                callback(obj);
                                                            } else {

                                                                dump.Thumbnails = obj.thumbnails;
                                                                this._internalGetArchiveUsersList( function(obj) {
                                                                    if (obj.error) {
                                                                        callback(obj);
                                                                    } else {

                                                                        dump.Users = obj.users;
                                                                        this._internalGetCommentsToc( function(obj) {

                                                                            if (obj.error) {
                                                                                callback(obj);
                                                                            } else {

                                                                                dump.Comments = obj.comments;
                                                                                callback({dump: dump});
                                                                            }
                                                                        }.bind(this));
                                                                    }
                                                                }.bind(this));
                                                            }
                                                        }.bind(this));
                                                    }
                                                }.bind(this));
                                            }
                                        }.bind(this));
                                    }
                                }
                            }.bind(this));
                        }.bind(this));
                    }.bind(this));
                }
            }.bind(this));
        }
    }.bind(this));
};


BalsamiqArchive.prototype.createFromDump = function (id, dump, callback) {
    this.archiveID = id;
    // var logObj = {action: "create"};
    // var resourcesNumber = dump.Resources ? dump.Resources.length : 0;
    // this.log.info(logObj, "HUNT: BEGIN create for archive " + id + " with resources " + resourcesNumber);

    // TODO: check if dump is well formed try catch is not obvious with all this callbacks
    // TODO: delete the zombie temporary sqlite archive in case of error (they are located in the "src/archive" folder

    this.create(this.archiveID, dump.Info[Consts.Info.ArchiveFormat], function (obj) {
        if (obj.error) {
            callback(obj);
            return;
        }
        // this.log.info(logObj, "HUNT: after create db " + id);
        this.setArchiveAttributes(dump.Info[Consts.Info.ArchiveAttributes], function (obj) {
            if (obj.error) {
                callback(obj);
                return;
            }
            // delete the default Master branch eventually created in the previous create call
            // this.log.info(logObj, "HUNT: after setting attributes " + id);
            this.deleteBranches([Consts.Branch.MasterBranchID], function (obj) {
                if (obj.error) {
                    callback(obj);
                    return;
                }

                // this.log.info(logObj, "HUNT: after deleting branches " + id);
                var branchesToCreate = dump.Branch;
                var createNextBranch = function(index, localThis) {
                    var toCreate;
                    if (index < branchesToCreate.length) {
                        toCreate = branchesToCreate[index];
                        localThis.createBranch(toCreate.branchID, toCreate.ATTRIBUTES, function (obj) {
                            if (obj.error) {
                                callback(obj);
                                return;
                            }
                            index++;
                            createNextBranch(index, localThis);
                        });
                    } else {
                        // localThis.log.info(logObj, "HUNT: after creating branches " + id);
                        var resourcesToCreate = dump.Resources;

                        var createNextResource = function(index, localThis) {
                            var toCreate;
                            if (index < resourcesToCreate.length) {
                                toCreate = resourcesToCreate[index];
                                localThis.createResource(toCreate.ID, toCreate.BRANCHID, toCreate.ATTRIBUTES, toCreate.DATA, function (obj) {
                                    if (obj.error) {
                                        callback(obj);
                                        return;
                                    }
                                    index++;
                                    createNextResource(index, localThis);
                                });
                            } else {
                                // localThis.log.info(logObj, "HUNT: after creating resources " + id);
                                var thumbnailsToCreate = dump.Thumbnails;
                                var createNextThumbnail = function(index, localThis) {
                                    var toCreate;
                                    if (index < thumbnailsToCreate.length) {
                                        toCreate = thumbnailsToCreate[index];
                                        localThis.createThumbnail(toCreate.ID, toCreate.ATTRIBUTES, function (obj) {
                                            if (obj.error) {
                                                callback(obj);
                                                return;
                                            }
                                            index++;
                                            createNextThumbnail(index, localThis);
                                        });
                                    } else {
                                        // localThis.log.info(logObj, "HUNT: after creating thumbnails " + id);
                                        var usersToCreate = dump.Users;
                                        var createNextUser = function(index, localThis) {
                                            var toCreate;
                                            if (index < usersToCreate.length) {
                                                toCreate = usersToCreate[index];
                                                localThis.reCreateUser(toCreate.ID, toCreate.ATTRIBUTES, function (obj) {
                                                    if (obj.error) {
                                                        callback(obj);
                                                        return;
                                                    }
                                                    index++;
                                                    createNextUser(index, localThis);
                                                });
                                            } else {
                                                // localThis.log.info(logObj, "HUNT: after creating users " + id);
                                                var commentsToCreate = dump.Comments;
                                                var createNextComment = function (index, localThis) {
                                                    var toCreate;
                                                    if (index < commentsToCreate.length) {
                                                        toCreate = commentsToCreate[index];
                                                        // commentID, resourceID, branchID, commentParentID
                                                        localThis.reCreateComment(toCreate.ID, toCreate.RESOURCEID, toCreate.BRANCHID, toCreate.USERID, toCreate.ATTRIBUTES, toCreate.DATA, function (obj) {
                                                            if (obj.error) {
                                                                callback(obj);
                                                                return;
                                                            }
                                                            index++;
                                                            createNextComment(index, localThis);
                                                        });
                                                    } else {
                                                        // localThis.log.info(logObj, "HUNT: after creating comments " + id);
                                                        localThis.setArchiveRevision(dump.Info[Consts.Info.ArchiveRevision], function (obj) {
                                                            // localThis.log.info(logObj, "HUNT: END calling callback " + id);
                                                            if (obj.error) {
                                                                callback(obj);
                                                            }
                                                            else {
                                                                callback(obj);
                                                            }
                                                        });
                                                    }
                                                };
                                                createNextComment(0, localThis);
                                            }
                                        };
                                        createNextUser(0, localThis);
                                    }

                                };
                                createNextThumbnail(0, localThis);
                            }
                        };
                        createNextResource(0, localThis);
                    }
                };

                createNextBranch(0, this);

            }.bind(this));
        }.bind(this));
    }.bind(this));
};

BalsamiqArchive.prototype.createFromBuffer = function (id, buffer, callback) {
    this.archiveID = id;
    this._enterCall();
    this.dbDriver.createFromBuffer(this.archiveID, buffer, function(obj) {

        if (obj.error) {
            this._exitCall();
            this._finalCallbackWithError(callback, obj);
        } else {
            this._internalGetArchiveData(Consts.Info.SchemaVersion, Consts.Info.SchemaVersion, false, function (obj) {

                if (obj.error) {
                    this._exitCall();
                    this._finalCallbackWithError(callback, obj);
                } else {
                    // TODO: how to compare?
                    if (obj[Consts.Info.SchemaVersion] !== Consts.CurrentSchemaVersion) {
                        this._migrate1To2(function(obj) {
                            if (obj.error)
                            {
                                this._exitCall();
                                this._finalCallbackWithError(callback, obj);
                            } else {
                                this._getArchiveRevision(function (p_pv) {
                                    this._exitCall();
                                    this._finalCallback(callback, p_pv, obj);
                                }.bind(this));
                            }
                        }.bind(this));
                    } else {
                        this._getArchiveRevision(function (p_pv) {
                            this._exitCall();
                            this._finalCallback(callback, p_pv, obj);
                        }.bind(this));
                    }
                }
             }.bind(this));
        }

    }.bind(this));
};

BalsamiqArchive.prototype.getBuffer = function (id, callback) {
    this._enterCall();
    this.archiveID = id;
    this.dbDriver.getBuffer(this.archiveID, function (obj) {
        this._exitCall();
        callback && callback(obj);
    }.bind(this));
};

// create a new DBDriver file
BalsamiqArchive.prototype.create = function (id, p_format, callback) {
    this.archiveID = id;
    this._enterCall();
    this.dbDriver.open(this.archiveID, "readWriteOrCreate", function (obj) {
        var query;
        if (obj.error) {
            this._exitCall();
            this._finalCallbackWithError(callback, obj);
        } else {
            this.dbDriver.serialize(this.archiveID, function (obj) {
                if (obj.error) {
                    this._exitCall();
                    this._finalCallbackWithError(callback, obj);
                } else {
                    // create tables
                    if (this.dbDriver.type == "mysql") {
                        query = "CREATE TABLE INFO (NAME VARCHAR(255) PRIMARY KEY, VALUE TEXT CHARACTER SET 'utf8mb4' COLLATE 'utf8mb4_unicode_ci');";
                    } else {
                        query = "CREATE TABLE INFO (NAME VARCHAR(255) PRIMARY KEY, VALUE TEXT);";
                    }
                    this.dbDriver.run(this.archiveID, query, {}, function (obj) {
                        if (obj.error) {
                            this._exitCall();
                            this._finalCallbackWithError(callback, obj);
                        } else {
                            if (this.dbDriver.type == "mysql") {
                                query = "CREATE TABLE BRANCHES (ID VARCHAR(255) PRIMARY KEY, ATTRIBUTES TEXT CHARACTER SET 'utf8mb4' COLLATE 'utf8mb4_unicode_ci');";
                            } else {
                                query = "CREATE TABLE BRANCHES (ID VARCHAR(255) PRIMARY KEY, ATTRIBUTES TEXT);";
                            }
                            this.dbDriver.run(this.archiveID, query, {}, function (obj) {
                                if (obj.error) {
                                    this._exitCall();
                                    this._finalCallbackWithError(callback, obj);
                                } else {
                                    if (this.dbDriver.type == "mysql") {
                                        query = "CREATE TABLE RESOURCES (ID VARCHAR(255), BRANCHID VARCHAR(255), ATTRIBUTES TEXT CHARACTER SET 'utf8mb4' COLLATE 'utf8mb4_unicode_ci', DATA LONGTEXT CHARACTER SET 'utf8mb4' COLLATE 'utf8mb4_unicode_ci', PRIMARY KEY (ID, BRANCHID), FOREIGN KEY (BRANCHID) REFERENCES BRANCHES(ID));";
                                    } else {
                                        query = "CREATE TABLE RESOURCES (ID VARCHAR(255), BRANCHID VARCHAR(255), ATTRIBUTES TEXT, DATA LONGTEXT, PRIMARY KEY (ID, BRANCHID), FOREIGN KEY (BRANCHID) REFERENCES BRANCHES(ID));";
                                    }
                                    this.dbDriver.run(this.archiveID, query, {}, function (obj) {
                                        if (obj.error) {
                                            this._exitCall();
                                            this._finalCallbackWithError(callback, obj);
                                        } else {
                                            this.dbDriver.run(this.archiveID, "CREATE TABLE THUMBNAILS (ID VARCHAR(255) PRIMARY KEY, ATTRIBUTES MEDIUMTEXT);", {}, function (obj) {
                                                if (obj.error) {
                                                    this._exitCall();
                                                    this._finalCallbackWithError(callback, obj);
                                                } else {
                                                    if (this.dbDriver.type == "mysql") {
                                                        query = "CREATE TABLE USERS (ID VARCHAR(255) PRIMARY KEY, ATTRIBUTES TEXT CHARACTER SET 'utf8mb4' COLLATE 'utf8mb4_unicode_ci');";
                                                    } else {
                                                        query = "CREATE TABLE USERS (ID VARCHAR(255) PRIMARY KEY, ATTRIBUTES TEXT);";
                                                    }
                                                    this.dbDriver.run(this.archiveID, query, {}, function (obj) {
                                                        if (obj.error) {
                                                            this._exitCall();
                                                            this._finalCallbackWithError(callback, obj);
                                                        } else {
                                                            if (this.dbDriver.type == "mysql") {
                                                                query = "CREATE TABLE COMMENTS (ID VARCHAR(255) PRIMARY KEY, BRANCHID VARCHAR(255), RESOURCEID VARCHAR(255), DATA LONGTEXT CHARACTER SET 'utf8mb4' COLLATE 'utf8mb4_unicode_ci', USERID VARCHAR(255), ATTRIBUTES TEXT CHARACTER SET 'utf8mb4' COLLATE 'utf8mb4_unicode_ci', FOREIGN KEY (USERID) REFERENCES USERS(ID), FOREIGN KEY (RESOURCEID, BRANCHID) REFERENCES RESOURCES(ID, BRANCHID));";
                                                            } else {
                                                                query = "CREATE TABLE COMMENTS (ID VARCHAR(255) PRIMARY KEY, BRANCHID VARCHAR(255), RESOURCEID VARCHAR(255), DATA LONGTEXT, USERID VARCHAR(255), ATTRIBUTES TEXT, FOREIGN KEY (USERID) REFERENCES USERS(ID), FOREIGN KEY (RESOURCEID, BRANCHID) REFERENCES RESOURCES(ID, BRANCHID));";
                                                            }
                                                            this.dbDriver.run(this.archiveID, query, {}, function (obj) {
                                                                if (obj.error) {
                                                                    this._exitCall();
                                                                    this._finalCallbackWithError(callback, obj);
                                                                } else {
                                                                    // create master branch
                                                                    this.dbDriver.run(this.archiveID, "INSERT INTO BRANCHES (ID, ATTRIBUTES) VALUES (?, ?);", [Consts.Branch.MasterBranchID, JSON.stringify({})], function (obj) {
                                                                        if (obj.error) {
                                                                            this._exitCall();
                                                                            this._finalCallbackWithError(callback, obj);
                                                                        } else {
                                                                            // set archive revision
                                                                            var values = [
                                                                                Consts.Info.SchemaVersion,
                                                                                Consts.CurrentSchemaVersion,
                                                                                Consts.Info.ArchiveRevision,
                                                                                0,
                                                                                Consts.Info.ArchiveFormat,
                                                                                p_format,
                                                                                Consts.Info.ArchiveRevisionUUID,
                                                                                ""
                                                                            ];
                                                                            this.dbDriver.run(this.archiveID, "INSERT INTO INFO (NAME, VALUE) VALUES (?, ?), (?, ?), (?, ?), (?, ?);", values, function (obj) {
                                                                                this._exitCall();
                                                                                if (obj.error) {
                                                                                    this._finalCallbackWithError(callback, obj)
                                                                                } else {
                                                                                    this._getArchiveRevision(function (p_pv) {
                                                                                        this._finalCallback(callback, p_pv);
                                                                                    }.bind(this));
                                                                                }
                                                                            }.bind(this));
                                                                        }
                                                                    }.bind(this));
                                                                }
                                                            }.bind(this));
                                                        }
                                                    }.bind(this));
                                                }
                                            }.bind(this));
                                        }
                                    }.bind(this));
                                }
                            }.bind(this));
                        }
                    }.bind(this));
                }
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype.validateBAR = function (id, callback) {
    // callback({ schemaVersion: null || SchemaVersions, archiveFormat:String?, archiveRevision:Number? || error })

    this.archiveID = id;
    this._enterCall();
    this.dbDriver.open(id, "read", function (obj) {
        if (obj.error) {
            this._exitCall();
            this._finalCallback(callback, { schemaVersion: null });
        } else {
            this.dbDriver.checkIntegrity(this.archiveID, function(obj) {
                if (obj.error) {
                    this.dbDriver.close(id, function () {
                        this._exitCall();
                        this._finalCallbackWithError(callback, obj);
                    }.bind(this));
                } else {
                    // check if upgrade is necessary
                    this._internalGetArchiveData(Consts.Info.SchemaVersion, Consts.Info.SchemaVersion, false, function (obj) {
                        if (obj.error) {
                            this.dbDriver.close(id, function () {
                                this._exitCall();
                                this._finalCallback(callback, { schemaVersion: null });
                            }.bind(this));
                        } else {
                            this._getArchiveRevision(function (archiveRevision) {
                                this.dbDriver.close(id, function () {
                                    this._exitCall();
                                    this._finalCallback(callback, archiveRevision, {
                                        schemaVersion: obj[Consts.Info.SchemaVersion],
                                        archiveFormat: "bmpr"        // TODO one more call to _internalGetArchiveData
                                    });
                                }.bind(this));
                            }.bind(this));
                        }
                    }.bind(this));
                }
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype.open = function (id, callback) {
    this.archiveID = id;
    this._enterCall();
    this.dbDriver.open(this.archiveID, "readWrite", function (obj) {
        if (obj.error) {
            this._exitCall();
            this._finalCallbackWithError(callback, obj);
        } else {
            // check if upgrade is necessary
            this._internalGetArchiveData(Consts.Info.SchemaVersion, Consts.Info.SchemaVersion, false, function (obj) {
                if (obj.error) {
                    this._exitCall();
                    this._finalCallbackWithError(callback, obj);
                } else {
                    // TODO: how to compare?
                    if (obj[Consts.Info.SchemaVersion] !== Consts.CurrentSchemaVersion) { // TODO fix this check!!!
                        this._migrate1To2(function(obj) {
                            if (obj.error) {
                                this._exitCall();
                                this._finalCallbackWithError(callback, obj);
                            } else {
                                this._internalGetToc(Consts.Branch.AllBranches, function(obj) {
                                    if (obj.error) {
                                        this._exitCall();
                                        this._finalCallbackWithError(callback, obj);
                                    } else {
                                        this._getArchiveRevision(function (p_pv) {
                                            this._exitCall();
                                            this._finalCallback(callback, p_pv, obj);
                                        }.bind(this));
                                    }
                                }.bind(this));
                            }
                        }.bind(this));
                    } else {
                        this._internalGetToc(Consts.Branch.AllBranches, function(obj) {
                            if (obj.error) {
                                this._exitCall();
                                this._finalCallbackWithError(callback, obj);
                            } else {
                                this._getArchiveRevision(function (p_pv) {
                                    this._exitCall();

                                    this._finalCallback(callback, p_pv, obj);
                                }.bind(this));
                            }
                        }.bind(this));
                    }
                }
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype._internalUpdateUser = function(attributes, callback) {

    if (!this.checkInternalCall(callback)) {
        return;
    }

    this._internalGetArchiveUsersList( function(obj) {

        if (obj.error) {
            callback(obj);
            return;
        }

        var user = null;
        if (obj.users) {
            user = obj.users.find(function (element/*, index, array*/) {
                return element.ID === this.userInfo.name;
            }.bind(this));
        }

        if (user === null) {
            callback({error: "User to be updated does not exists: " + this.userInfo.name});
            return;
        }

        this.dbDriver.run(this.archiveID, "UPDATE USERS SET ATTRIBUTES=? WHERE ID=?;", [JSON.stringify(attributes), this.userInfo.name], function (obj) {
            callback(obj);
        }.bind(this));
    }.bind(this));
};

BalsamiqArchive.prototype._internalInsertUserIfNeeded = function(attributes, callback) {

    if (!this.checkInternalCall(callback)) {
        return;
    }

    this._internalGetArchiveUsersList( function(obj) {

        if (obj.error) {
            callback(obj);
            return;
        }

        var user = null;
        if (obj.users) {
            user = obj.users.find(function (element/*, index, array*/) {
                return element.ID === this.userInfo.name;
            }.bind(this));
        }

        if (user && !attributes) {
            callback({});
        }
        else if (user && attributes) {
            this._internalUpdateUser(attributes, callback);
        }
        else {

            this.dbDriver.run(this.archiveID, "INSERT INTO USERS (ID, ATTRIBUTES) VALUES (?, ?);", [ this.userInfo.name, JSON.stringify(attributes)], function (obj) {
                if (obj.error) {
                    callback(obj);
                }
                else {
                    callback({});
                }
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype._purgeComments = function (userID, callback) {
    this._internalGetComments(function(obj) {
        if (obj.error) {
            callback(obj);
        } else {
            var commentIDsToPurge = [];
            for (var i = 0; i < obj.comments.length; i++) {
                var comment = obj.comments[i];
                if (comment.ATTRIBUTES[Consts.CommentAttributes.Trashed] === true && comment.ATTRIBUTES[Consts.CommentAttributes.TrashedBy] === userID && comment.DATA !== "{}") {
                    commentIDsToPurge.push(comment.ID);
                }
            }

            makeQueryInBatches(this.dbDriver, 'run', this.archiveID, "UPDATE COMMENTS SET DATA = '{}' WHERE ID in ([!!__VALUES__!!])", [], commentIDsToPurge, function (obj) {
                if (obj.error) {
                    callback(obj);
                } else {
                    callback({ purgedCommentIDs: commentIDsToPurge })
                }
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype.purgeComments = function (userID, callback) {
    this._enterCall();
    this._purgeComments(userID, function (obj) {
        if (obj.error) {
            this._exitCall();
            this._finalCallbackWithError(callback, obj);
        } else {
            this._getArchiveRevision(function (p_ar) {
                this._exitCall();
                this._finalCallback(callback, p_ar, obj);
            }.bind(this));
        }
    }.bind(this));
};

// close the DBDriver db
BalsamiqArchive.prototype.close = function (id, callback) {
    this.archiveID = id;
    this._enterCall();
    this._purgeComments(this.userInfo.name, function () {
        // ignore purge errors. we want to continue with the close anyway.

        this._getArchiveRevision(function (p_ar) {
            this.dbDriver.close(this.archiveID, function (obj) {
                this._exitCall();
                if (obj.error) {
                    this._finalCallbackWithError(callback, obj);
                } else {
                    this._finalCallback(callback, p_ar);
                }
                this.dbDriver = null;
            }.bind(this));
        }.bind(this));
    }.bind(this));
};

BalsamiqArchive.prototype.destroy = function (id, callback) {
    this.archiveID = id;
    this.dbDriver.destroy(this.archiveID, callback);
};

BalsamiqArchive.prototype.destroyExt = function (id, force, callback) {
    this.archiveID = id;
    if (this.dbDriver.destroyExt) {
        this.dbDriver.destroyExt(this.archiveID, force, callback);
    } else {
        // TODO
        // N/A
        this.dbDriver.destroy(this.archiveID, callback);
    }
};

// read and return the archive Revision
BalsamiqArchive.prototype.getArchiveRevision = function (callback) {
    this._getArchiveRevision(function (p_pv) {
        this._finalCallback(callback, p_pv);
    }.bind(this));
};

//Only used for restoring from a previous dump
BalsamiqArchive.prototype.setArchiveRevision = function (p_revision, callback) {
    this._enterCall();
    this.dbDriver.run(this.archiveID, "REPLACE INTO INFO (NAME, VALUE) VALUES (?, ?);", [Consts.Info.ArchiveRevision, p_revision], function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._finalCallback(callback, p_revision);
        }
    }.bind(this));
};

// read attributes for the whole archive
BalsamiqArchive.prototype.getArchiveAttributes = function (callback) {
    this._getArchiveData(Consts.Info.ArchiveAttributes, callback, "attributes", true);
};

// set attributes for the whole archive
BalsamiqArchive.prototype.setArchiveAttributes = function (p_attrs, callback) {
    this._setArchiveData(Consts.Info.ArchiveAttributes, JSON.stringify(p_attrs), callback);
};

BalsamiqArchive.prototype.compact = function (callback) {
    this._enterCall();
    this.dbDriver.compact(this.archiveID, function(obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._getArchiveRevision(function (p_pv) {
                this._finalCallback(callback, p_pv, obj);
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype.download = function (callback/*, progressCallback*/) {
    this.dbDriver.download(this.archiveID, function(uint8Array)
    {
        callback(uint8Array);
    }.bind(this));
};

BalsamiqArchive.prototype.isLocal = function() {
    return true;
};

// read archive format
BalsamiqArchive.prototype.getArchiveFormat = function (callback) {
    this._getArchiveData(Consts.Info.ArchiveFormat, callback, "format");
};

BalsamiqArchive.prototype.getArchiveSchemaVersion = function (callback) {
    this._getArchiveData(Consts.Info.SchemaVersion, callback, Consts.Info.SchemaVersion);
};

// write archive format
BalsamiqArchive.prototype.setArchiveFormat = function (p_format, callback) {
    this._setArchiveData(Consts.Info.ArchiveFormat, p_format, callback);
};

// read the branch list
BalsamiqArchive.prototype.getBranches = function (callback) {
    this._enterCall();

    this._internalGetBranches(function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._getArchiveRevision(function (p_pv) {
                this._finalCallback(callback, p_pv, obj);
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype._internalGetBranches = function (callback) {
    var archiveID;

    if (!this.checkInternalCall(callback)) {
        return;
    }

    archiveID = this.archiveID;
    this.dbDriver.all(archiveID, "SELECT ID, ATTRIBUTES FROM BRANCHES;", {}, function (obj) {
        var res, i, p_record;

        if (obj.error) {
            callback(obj);
        } else {
            res = { branches: [] };
            try {
                for (i=0; i<obj.rows.length; i++) {
                    p_record = obj.rows[i];
                    res.branches.push({
                        branchID: p_record.ID,
                        ATTRIBUTES: JSON.parse(p_record.ATTRIBUTES)
                    });
                }
                callback(res);
            } catch (e) {
                callback({error: "Balsamiq Archive error: Unexpected database response while getting attributes, id from branch: " + archiveID + " " + JSON.stringify(obj) + '. Error: ' + e.message, stack: e.stack});
            }
        }
    }.bind(this));
};

// read the attributes of the given branch
BalsamiqArchive.prototype.getBranchAttributes = function (branchID, callback) {
    this._enterCall();

    this.dbDriver.get(this.archiveID, "SELECT ATTRIBUTES FROM BRANCHES WHERE ID = ?;", [branchID], function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            try {
                if (obj.row) {
                    var res = { attributes: JSON.parse(obj.row.ATTRIBUTES) };
                    this._getArchiveRevision(function (p_pv) {
                        this._finalCallback(callback, p_pv, res);
                    }.bind(this));
                } else {
                    this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: Unexpected db response while getting branch attribute: no attribute for selected branch: " + branchID});
                }
            } catch (e) {
                this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: Unexpected database response while getting branch attribute: " + JSON.stringify(obj)});
            }
        }
    }.bind(this));
};

// write the attributes of a branch
BalsamiqArchive.prototype.setBranchAttributes = function (branchID, attributes, callback) {
    this._enterCall();

    if (!attributes) {
        attributes = {};
    }

    this.dbDriver.run(this.archiveID, "UPDATE BRANCHES SET ATTRIBUTES = ? WHERE ID = ?;", [JSON.stringify(attributes), branchID], function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._finalCallbackWithAffectedRowsCheck(callback, obj);
        }
    }.bind(this));
};

// read the resources belonging to the given branch
BalsamiqArchive.prototype.getAllResourcesAttributesForBranch = function (branchID, callback) {
    this._enterCall();
    this._internalGetAllResourcesAttributesForBranch(branchID, function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj.error);
        } else {
            this._getArchiveRevision(function (p_pv) {
                this._finalCallback(callback, p_pv, obj);
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype._internalGetAllResourcesAttributesForBranch = function (branchID, callback) {

    if (!this.checkInternalCall(callback)) {
        return;
    }

    var query, values;
    var archiveID = this.archiveID;

    if (branchID == Consts.Branch.AllBranches || branchID == null) {
        query =  "SELECT ID, BRANCHID, ATTRIBUTES FROM RESOURCES";
        values = [];
    } else {
        query =  "SELECT ID, BRANCHID, ATTRIBUTES FROM RESOURCES WHERE BRANCHID = ?";
        values = [branchID];
    }

    this.dbDriver.all(this.archiveID, query, values, function (obj) {
        var res, i, p_record, attrs;
        if (obj.error) {
            callback(obj);
        }
        else {
            try {
                res = { resources: [] };
                for (i=0; i<obj.rows.length; i++) {
                    p_record = obj.rows[i];
                    attrs = JSON.parse(p_record.ATTRIBUTES);
                    res.resources.push({
                        ID: p_record.ID,
                        BRANCHID: p_record.BRANCHID,
                        ATTRIBUTES: attrs
                    });
                }
                callback(res);
            } catch (e) {
                callback({
                    error: "Balsamiq Archive error: Unexpected database response while getting resources (" + p_record.ID + ", " + branchID + ")",
                    archiveID: archiveID,
                    attributes: p_record.ATTRIBUTES,
                    errorMessage: e.message,
                    stack: e.stack,
                    bmprObj: obj
                });
            }
        }
    }.bind(this));
};

// create a new resource
BalsamiqArchive.prototype.createResource = function (p_resourceID, branchID, attributes, p_data, callback) {
    this._enterCall();

    if (p_data === undefined) {
        p_data = "";
    }
    if (attributes === null || attributes === undefined) {
        attributes = {};
    }
    var values = [
        p_resourceID,
        branchID,
        p_data,
        JSON.stringify(attributes)
    ];
    this.dbDriver.run(this.archiveID, "INSERT INTO RESOURCES (ID, BRANCHID, DATA, ATTRIBUTES) VALUES (?, ?, ?, ?);", values, function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._finalCallbackWithAffectedRowsCheck(callback, obj, {ID: p_resourceID});
        }
    }.bind(this));
};

BalsamiqArchive.prototype.deleteResources = function (p_resourceIDs, branchID, callback) {
    var query, fixedValues;

    this._enterCall();

    this._internalGetCommentsToc(function (obj) {
        var i, j;
        if (obj.error) {
            this._exitCall();
            this._finalCallbackWithError(callback, obj);
        } else {

            // check if there are comments pointing to these resources and issue an error
            for (i = 0; i < obj.comments.length; i++) {
                var comment = obj.comments[i];

                if (branchID === comment.BRANCHID) {
                    for (j=0; j<p_resourceIDs.length; j++) {
                        if (p_resourceIDs[j] == comment.RESOURCEID) {
                            this._exitCall();
                            this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: Could not delete a resource with comments"});
                            return;
                        }
                    }
                }
            }

            // no comments? go!

            if (branchID == Consts.Branch.AllBranches || branchID == null) {
                query = "DELETE FROM RESOURCES WHERE ID in ([!!__VALUES__!!])";
                fixedValues = [];
            } else {
                query = "DELETE FROM RESOURCES WHERE BRANCHID = ? AND ID in ([!!__VALUES__!!])";
                fixedValues = [branchID];
            }

            makeQueryInBatches(this.dbDriver, 'run', this.archiveID, query, fixedValues, p_resourceIDs, function (obj) {
                this._exitCall();
                if (obj.error) {
                    this._finalCallbackWithError(callback, obj);
                } else {
                    this._finalCallbackWithAffectedRowsCheck(callback, obj);
                }
            }.bind(this));
        }
    }.bind(this));
};

// read the attributes of the selected resources
BalsamiqArchive.prototype.getResourceAttributes = function (p_resourceID, p_branchID, callback) {
    var query, values, allBranch;

    this._enterCall();

    if (p_branchID == Consts.Branch.AllBranches || p_branchID == null) {
        query = "SELECT ID, BRANCHID, ATTRIBUTES FROM RESOURCES WHERE ID = ?;";
        values = [p_resourceID];
        allBranch = true;
    } else {
        query = "SELECT ATTRIBUTES FROM RESOURCES WHERE ID = ? AND BRANCHID = ?;";
        values = [p_resourceID, p_branchID];
        allBranch = false;
    }


    this.dbDriver.all(this.archiveID, query, values, function (obj) {
        var res;
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            try {
                res = {attributes: {}};
                if (allBranch) {
                    for (var i=0; i<obj.rows.length; i++)
                    {
                        var row = obj.rows[i];
                        res.attributes[row.BRANCHID] = JSON.parse(row.ATTRIBUTES);
                    }
                } else {
                    if (obj.rows && obj.rows[0]) {
                        row = obj.rows[0];
                        res.attributes = JSON.parse(row.ATTRIBUTES);
                    }
                }
                this._getArchiveRevision(function (p_pv) {
                    this._finalCallback(callback, p_pv, res);
                }.bind(this));
            } catch (e) {
                this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: Unexpected database response while getting resource attributes: " + JSON.stringify(obj)});
            }
        }
    }.bind(this));
};

// set the attributes of the given resource
BalsamiqArchive.prototype.setResourceAttributes = function (p_resourceID, p_branchID, p_attrs, callback) {
    this._enterCall();

    this.dbDriver.run(this.archiveID, "UPDATE RESOURCES SET ATTRIBUTES = ? WHERE ID = ? AND BRANCHID = ?;", [JSON.stringify(p_attrs), p_resourceID, p_branchID], function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._finalCallbackWithAffectedRowsCheck(callback, obj)
        }
    }.bind(this));
};

// set the branch id of the given resource
BalsamiqArchive.prototype.setResourceBranchID = function (resourceID, oldBranchID, newBranchID, callback) {
    this._enterCall();

    this.dbDriver.get(this.archiveID, "SELECT ID, ATTRIBUTES, DATA FROM RESOURCES WHERE ID = ? AND BRANCHID = ?;", [resourceID, oldBranchID], function (obj) {
        var res;
        if (obj.error) {
            this._exitCall();
            this._finalCallbackWithError(callback, obj);
            return;
        }

        if (!obj.row) {
            this._exitCall();
            this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: no resource found trying to set the branch ID"});
            return;
        }

        res = obj.row;

        this.dbDriver.run(this.archiveID, "INSERT INTO RESOURCES (ID, ATTRIBUTES, DATA, BRANCHID) VALUES (?, ?, ?, ?);", [res.ID, res.ATTRIBUTES, res.DATA, newBranchID], function (obj) {
            if (obj.error) {
                this._exitCall();
                this._finalCallbackWithError(callback, obj);
                return;
            }

            this.dbDriver.run(this.archiveID, "UPDATE COMMENTS SET BRANCHID = ? WHERE RESOURCEID = ? AND BRANCHID = ?;", [newBranchID, resourceID, oldBranchID], function (obj) {
                if (obj.error) {
                    this._exitCall();
                    this._finalCallbackWithError(callback, obj);
                    return;
                }

                this.dbDriver.run(this.archiveID, "DELETE FROM RESOURCES WHERE ID = ? AND BRANCHID = ?;", [resourceID, oldBranchID], function (obj) {
                    this._exitCall();
                    if (obj.error) {
                        this._finalCallbackWithError(callback, obj);
                    } else {
                        this._finalCallbackWithAffectedRowsCheck(callback, obj);
                    }
                }.bind(this));
            }.bind(this));
        }.bind(this));
    }.bind(this));
};

// read the data of the given resource
BalsamiqArchive.prototype.getResourceData = function (p_resourceID, branchID, callback) {
    this.getResourceDataWithOptions(p_resourceID, branchID, {}, callback);
};

BalsamiqArchive.prototype.getResourceDataWithOptions = function (p_resourceID, branchID, options, callback) {
    this._enterCall();

    var computeSHA1 = !!options.computeSHA1;
    var computeSHA1Query = computeSHA1 ? "sha1(DATA) as SHA1, " : "";

    this.dbDriver.get(this.archiveID, "SELECT " + computeSHA1Query + "DATA FROM RESOURCES WHERE ID = ? AND BRANCHID = ?;", [p_resourceID, branchID], function (obj) {
        var res;
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            try {
                if (obj.row && typeof obj.row.DATA === "string") {
                    res = {
                        data: obj.row.DATA
                    };
                    if (computeSHA1) {
                        res.sha1 = obj.row.SHA1;
                    }
                    this._getArchiveRevision(function (p_pv) {
                        this._finalCallback(callback, p_pv, res);
                    }.bind(this));
                } else {
                    this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: Unexpected database response while getting resource data: " + JSON.stringify(obj)});
                }
            }
            catch (e) {
                this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: Unexpected database response while getting resource data: " + JSON.stringify(obj)});
            }
        }
    }.bind(this));
};

BalsamiqArchive.prototype.getResourcesData = function (p_resourceIDs, branchID, callback)
{
    // var query, values, tmp = [], id;

    this._enterCall();

    this._internalGetResourcesData(p_resourceIDs, branchID, function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._getArchiveRevision(function (p_pv) {
                this._finalCallback(callback, p_pv, obj);
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype._internalGetResourcesData = function (p_resourceIDs, branchID, callback)
{
    var query, fixedValues;

    if (p_resourceIDs && p_resourceIDs.length) {
        if (branchID == Consts.Branch.AllBranches || branchID == null) {
            query = "SELECT ID, BRANCHID, DATA FROM RESOURCES WHERE ID in ([!!__VALUES__!!])";
            fixedValues = [];
        } else {
            query = "SELECT ID, BRANCHID, DATA FROM RESOURCES WHERE BRANCHID = ? AND ID in ([!!__VALUES__!!])";
            fixedValues = [branchID];
        }

        makeQueryInBatches(this.dbDriver, 'all', this.archiveID, query, fixedValues, p_resourceIDs, function (obj) {
            if (obj.error) {
                callback(obj);
            } else {
                var res = {data: {}};
                try {
                    for (var i = 0; i < obj.rows.length; i++) {
                        var s = obj.rows[i];
                        res.data[s.ID + "|" + s.BRANCHID] = s.DATA;
                    }
                    callback(res);
                }
                catch (e) {
                    callback({error: "Balsamiq Archive error: Unexpected database response while getting resources data: " + e.stack});
                }
            }
        }.bind(this));
    } else {
        callback({error: "Balsamiq Archive error: Unexpected parameter while getting resources data: p_resourceIDs is null or is not an valid"});
    }
};

// set the data of the given resource
BalsamiqArchive.prototype.setResourceData = function (p_resourceID, branchID, p_data, callback) {
    if (typeof p_data !== 'string') {
        this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: p_data must be a String!"});
    } else {
        this._enterCall();

        this.dbDriver.run(this.archiveID, "UPDATE RESOURCES SET DATA = ? WHERE ID = ? AND BRANCHID = ?;", [p_data, p_resourceID, branchID], function (obj) {
            this._exitCall();
            if (obj.error) {
                this._finalCallbackWithError(callback, obj);
            } else {
                this._finalCallbackWithAffectedRowsCheck(callback, obj);
            }
        }.bind(this));
    }
};

BalsamiqArchive.prototype.getThumbnails = function (callback) {
    this._enterCall();
    this._internalGetThumbnails(function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj.error);
        }
        else {
            this._getArchiveRevision(function (p_pv) {
                this._finalCallback(callback, p_pv, obj);
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype._internalGetThumbnails = function (callback) {

    if (!this.checkInternalCall(callback)) {
        return;
    }
    this.dbDriver.all(this.archiveID, "SELECT ID, ATTRIBUTES FROM THUMBNAILS;", [], function (obj) {
        var attrs, i, p_record, res;
        if (obj.error) {
            callback(obj);
        } else {
            try {
                res = { thumbnails: [] };
                for (i=0; i<obj.rows.length; i++) {
                    p_record = obj.rows[i];
                    attrs = JSON.parse(p_record.ATTRIBUTES);
                    res.thumbnails.push({
                        ID: p_record.ID,
                        ATTRIBUTES: attrs
                    });
                }
                callback(res);
            } catch (e) {
                callback({error: "Balsamiq Archive error: Unexpected database response while getting thumbnails: " + JSON.stringify(obj)});
            }
        }
    }.bind(this));
};

BalsamiqArchive.prototype.createThumbnail = function (thumbnailID, attributes, callback) {
    this._enterCall();

    if (attributes === null || attributes === undefined) {
        attributes = {};
    }
    var values = [
        thumbnailID,
        JSON.stringify(attributes)
    ];
    this.dbDriver.run(this.archiveID, "INSERT INTO THUMBNAILS (ID, ATTRIBUTES) VALUES (?, ?);", values, function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._finalCallbackWithAffectedRowsCheck(callback, obj, {ID: thumbnailID});
        }
    }.bind(this));
};

BalsamiqArchive.prototype.getThumbnail = function (thumbnailID, callback) {
    this._enterCall();
    this.dbDriver.get(this.archiveID, "SELECT ID, ATTRIBUTES FROM THUMBNAILS WHERE ID = ?;", [thumbnailID], function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            try {
                if (obj.row) {
                    var res = {
                        ID: obj.row.ID,
                        ATTRIBUTES: JSON.parse(obj.row.ATTRIBUTES),
                    };
                    this._getArchiveRevision(function (p_pv) {
                        this._finalCallback(callback, p_pv, res);
                    }.bind(this));
                } else {
                    this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: Unexpected database response while getting thumbnail: " + JSON.stringify(obj)});
                }
            } catch (e) {
                this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: Unexpected database response while getting thumbnail: " + JSON.stringify(obj)});
            }
        }
    }.bind(this));
};

// TODO: to evaluate performance
BalsamiqArchive.prototype.getThumbnailFromResourceID = function (resourceID, branchID, callback) {
    var queryResourceID = "%" + resourceID + "%";
    var queryBranchID = "%" + branchID + "%";
    var resourceQuery, thumbnailAttribute;
    var res;

    this._enterCall();

    // getting image
    this.dbDriver.all(this.archiveID, "SELECT * FROM THUMBNAILS WHERE ATTRIBUTES LIKE ? AND ATTRIBUTES LIKE ?", [queryResourceID, queryBranchID], function (thumnbnailObj) {
        if (thumnbnailObj.error) {
            this._exitCall();
            this._finalCallbackWithError(callback, thumnbnailObj);
        } else {
            res = {
                id: "", image: "", resourceName: "", branchName: "", projectName: ""
            };
            if (thumnbnailObj.rows && thumnbnailObj.rows.length) {
                res.id = thumnbnailObj.rows[0].ID;
                try {
                    thumbnailAttribute = JSON.parse(thumnbnailObj.rows[0].ATTRIBUTES);
                } catch (e) {
                    this._exitCall();
                    this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: Unexpected database response while getting thumbnail: " + thumnbnailObj.rows[0].ATTRIBUTES});
                    return;
                }
                res.image = thumbnailAttribute.image;

                // getting project name
                this._internalGetArchiveData(Consts.Info.ArchiveAttributes, "attributes", true, function (archiveObj) {
                    if (archiveObj.error) {
                        this._exitCall();
                        this._finalCallbackWithError(callback, archiveObj);
                    } else {
                        res.projectName = archiveObj.attributes.name;

                        // getting resource name (it is only in master branch)
                        resourceQuery = "SELECT ATTRIBUTES FROM RESOURCES WHERE ID = ? AND BRANCHID = ?;";
                        this.dbDriver.all(this.archiveID, resourceQuery, [resourceID, Consts.Branch.MasterBranchID], function (resourceObj) {
                            var resourceAttributes, row;

                            if (resourceObj.error) {
                                this._exitCall();
                                this._finalCallbackWithError(callback, resourceObj);
                            } else {
                                if (resourceObj.rows && resourceObj.rows.length) {
                                    row = resourceObj.rows[0];
                                    try {
                                        resourceAttributes = JSON.parse(row.ATTRIBUTES);
                                    } catch (e) {
                                        this._exitCall();
                                        this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: Unexpected database response while getting thumbnail: " + row.ATTRIBUTES});
                                        return;
                                    }

                                    res.resourceName = resourceAttributes.name;

                                    // getting branch name (it is only in master branch)
                                    if (branchID !== Consts.Branch.MasterBranchID) {
                                        resourceQuery = "SELECT ATTRIBUTES FROM BRANCHES WHERE ID = ?;";
                                        this.dbDriver.all(this.archiveID, resourceQuery, [branchID], function (resourceObj) {
                                            var resourceAttributes, row;

                                            if (resourceObj.error) {
                                                this._exitCall();
                                                this._finalCallbackWithError(callback, resourceObj);
                                            } else {
                                                if (resourceObj.rows && resourceObj.rows[0]) {
                                                    row = resourceObj.rows[0];
                                                    try {
                                                        resourceAttributes = JSON.parse(row.ATTRIBUTES);
                                                    } catch (e) {
                                                        this._exitCall();
                                                        this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: Unexpected database response while getting thumbnail: " + row.ATTRIBUTES});
                                                        return;
                                                    }
                                                    res.branchName = resourceAttributes.branchName;
                                                }
                                                this._exitCall();
                                                this._getArchiveRevision(function (p_pv) {
                                                    this._finalCallback(callback, p_pv, res);
                                                }.bind(this));
                                            }
                                        }.bind(this));
                                    } else {
                                        this._exitCall();
                                        this._getArchiveRevision(function (p_pv) {
                                            this._finalCallback(callback, p_pv, res);
                                        }.bind(this));
                                    }
                                } else {
                                    // resource does not exit
                                    this._exitCall();
                                    this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: Unexpected database response while getting thumbnail: resource does not existing"});
                                }
                            }
                        }.bind(this));
                    }
                }.bind(this));
            } else {
                // resource does not exit
                this._exitCall();
                this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: Unexpected database response while getting thumbnail: resource does not existing"});
            }
        }
    }.bind(this));
};

BalsamiqArchive.prototype.setThumbnail = function (thumbnailID, attributes, callback) {
    this._enterCall();

    this.dbDriver.run(this.archiveID, "UPDATE THUMBNAILS SET ATTRIBUTES = ? WHERE ID = ?;", [JSON.stringify(attributes), thumbnailID], function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._finalCallbackWithAffectedRowsCheck(callback, obj);
        }
    }.bind(this));
};

BalsamiqArchive.prototype.deleteThumbnails = function (thumbnailIDs, callback) {
    this._enterCall();

    makeQueryInBatches(this.dbDriver, 'run', this.archiveID, "DELETE FROM THUMBNAILS WHERE ID in ([!!__VALUES__!!])", [], thumbnailIDs, function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._finalCallbackWithAffectedRowsCheck(callback, obj);
        }
    }.bind(this));
};

BalsamiqArchive.prototype.deleteBranches = function (branchIDs, callback) {
    this._enterCall();

    makeQueryInBatches(this.dbDriver, 'run', this.archiveID, "DELETE FROM BRANCHES WHERE ID in ([!!__VALUES__!!])", [], branchIDs, function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._finalCallbackWithAffectedRowsCheck(callback, obj);
        }
    }.bind(this));
};

BalsamiqArchive.prototype.createBranch = function(p_branchID, p_attributes, callback) {
    var attr = JSON.stringify(p_attributes);
    this._enterCall();
    this.dbDriver.run(this.archiveID, "INSERT INTO BRANCHES (ID, ATTRIBUTES) VALUES (?, ?);", [p_branchID, attr], function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._finalCallbackWithAffectedRowsCheck(callback, obj);
        }
    }.bind(this));
};

// obj.comments array of {id, data}
BalsamiqArchive.prototype.getCommentsData = function(commentIDs, callback) {
    this._enterCall();

    this._internalGetCommentsData(commentIDs, false, function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._getArchiveRevision(function (p_pv) {
                this._finalCallback(callback, p_pv, obj);
            }.bind(this));
        }

    }.bind(this));
};

BalsamiqArchive.prototype._internalGetCommentsData = function(commentIDs, includeTrashedData, callback) {

    if (!this.checkInternalCall(callback)) {
        return;
    }

    var res = { comments: [] };

    if (commentIDs && commentIDs.length) {
        makeQueryInBatches(this.dbDriver, 'all', this.archiveID, "SELECT ID, BRANCHID, RESOURCEID, DATA, USERID, ATTRIBUTES FROM COMMENTS WHERE ID in ([!!__VALUES__!!])", [], commentIDs, function (obj) {
            var i, row;

            if (obj.error) {
                callback(obj);
            } else {
                try {
                    for (i=0; i<obj.rows.length; i++) {
                        row = obj.rows[i];
                        var attributes = JSON.parse(row.ATTRIBUTES);
                        res.comments.push({
                            ID: row.ID,
                            BRANCHID: row.BRANCHID,
                            RESOURCEID: row.RESOURCEID,
                            USERID: row.USERID,
                            ATTRIBUTES: attributes,
                            DATA: !includeTrashedData && attributes[Consts.CommentAttributes.Trashed] ? "{}" : row.DATA          // this is a string! as for input value
                        });
                    }
                    callback(res);
                } catch (e) {
                    this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: Unexpected database response while getting comments data: " + JSON.stringify(obj)});
                }
            }
        }.bind(this));
    } else {
        callback(res);
    }
};

BalsamiqArchive.prototype.reCreateComment = function(commentID, resourceID, branchID, userID, attributes, data, callback) {

    this._enterCall();

    this.dbDriver.run(this.archiveID, "INSERT INTO COMMENTS (ID, RESOURCEID, BRANCHID, DATA, USERID, ATTRIBUTES) VALUES (?, ?, ?, ?, ?, ?);", [ commentID, resourceID, branchID, data, userID, JSON.stringify(attributes)], function(obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {

            var res = {
                commentID: commentID,
                resourceID: resourceID,
                branchID: branchID,
                userID: this.userInfo.name,
                attributes: attributes
            };

            this._finalCallbackWithAffectedRowsCheck(callback, obj, res);
        }
    }.bind(this));
};

// return id and attributes of newly created comment
// user id is not passed as a param because it is already known
BalsamiqArchive.prototype.createComment = function(commentID, resourceID, branchID, commentParentID, data, callback) {

    this._enterCall();

    var dateTime = Date.now();

    var attributes = {};

    attributes[Consts.CommentAttributes.ParentID] =  commentParentID;
    attributes[Consts.CommentAttributes.ReadBy] =  [this.userInfo.name];
    attributes[Consts.CommentAttributes.Timestamp] =  dateTime;
    attributes[Consts.CommentAttributes.Trashed] =  false;
    attributes[Consts.CommentAttributes.TrashedBy] =  "";
    attributes[Consts.CommentAttributes.LikedBy] =  [];
    attributes[Consts.CommentAttributes.Timestamps] =  {};

    this.dbDriver.run(this.archiveID, "INSERT INTO COMMENTS (ID, BRANCHID, RESOURCEID, DATA, USERID, ATTRIBUTES) VALUES (?, ?, ?, ?, ?, ?);", [ commentID, branchID, resourceID, data, this.userInfo.name, JSON.stringify(attributes)], function(obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {

            var res = {
                commentID: commentID,
                resourceID: resourceID,
                branchID: branchID,
                userID: this.userInfo.name,
                attributes: attributes
            };

            this._finalCallbackWithAffectedRowsCheck(callback, obj, res);
        }
    }.bind(this));
};

/** @typedef {{ imported: boolean; likedBy: string[]; originalAuthor?: string; parentID: string; readBy: string[]; timestamp: number; timestamps: unknown; trashed: boolean; trashedBy: string; }} Attributes */

/** @typedef {{ color: number; id: string; label: string; labelColor: number; tip: "left" | "top" | "right" | "bottom"; x: number; y: number; }} Callout */

/** @typedef {{ id: string; data: { text: string, callouts: Callout[] }; attributes: Attributes; originalAuthorDisplayName: string; }} ApiComment */

/**
 *
 * @param {string} resourceID
 * @param {string} branchID
 * @param {ApiComment} comment
 * @param {() => void} callback
 */
BalsamiqArchive.prototype.importComment = function (resourceID, branchID, comment, callback) {
    this._enterCall();

    var attributes = {};

    attributes[Consts.CommentAttributes.ParentID] = comment.attributes.parentID;
    attributes[Consts.CommentAttributes.ReadBy] = [this.userInfo.name];
    attributes[Consts.CommentAttributes.Timestamp] = comment.attributes.timestamp;
    attributes[Consts.CommentAttributes.Trashed] = comment.attributes.trashed;
    attributes[Consts.CommentAttributes.TrashedBy] = comment.attributes.trashedBy === "" ? comment.attributes.trashedBy : Consts.IMPORTED_USER_ID;
    attributes[Consts.CommentAttributes.LikedBy] = comment.attributes.likedBy.map(function () {
        return Consts.IMPORTED_USER_ID;
    });
    attributes[Consts.CommentAttributes.Timestamps] = comment.attributes.timestamps;
    attributes[Consts.CommentAttributes.Imported] = true;
    attributes[Consts.CommentAttributes.OriginalAuthor] = comment.originalAuthorDisplayName;

    this.dbDriver.run(
        this.archiveID,
        "INSERT INTO COMMENTS (ID, BRANCHID, RESOURCEID, DATA, USERID, ATTRIBUTES) VALUES (?, ?, ?, ?, ?, ?);",
        [comment.id, branchID, resourceID, JSON.stringify(comment.data), this.userInfo.name, JSON.stringify(attributes)],
        function (obj) {
            this._exitCall();
            if (obj.error) {
                this._finalCallbackWithError(callback, obj);
            } else {
                var res = {
                    commentID: comment.id,
                    attributes: attributes, 
                };
                this._finalCallbackWithAffectedRowsCheck(callback, obj, res);
            }
        }.bind(this)
    );
};

// user id is not passed as a param because it is already known
// returns an array of { commentID: xxx, attributes: {} }
BalsamiqArchive.prototype.updateCommentsAttributes = function(commentIDs, attributesArray, callback) {
    this._enterCall();
    this._internalUpdateCommentsAttributes(commentIDs, attributesArray, function(obj, res) {
        this._exitCall();

        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._finalCallbackWithAffectedRowsCheck(callback, obj, { attributes: res });
        }
    }.bind(this));
};

// user id is not passed as a param because it is already known
BalsamiqArchive.prototype._internalUpdateCommentsAttributes = function(commentIDs, attributes, callback) {

    if (!this.checkInternalCall(callback)) {
        return;
    }

    if (commentIDs === null || commentIDs === undefined) {
        callback({ error: "Unexpected error: commentIDs is null" });
        return;
    }

    if (commentIDs.length === 0) {
        callback({ affectedRows: 0 });
        return;
    }

    this._internalGetCommentsData(commentIDs, true, function(obj) {
        var tagList, index, tag, j;

        if (obj.error) {
            callback(obj);
        } else {

            if (obj.comments.length !== commentIDs.length) {
                callback({ affectedRows: 0 });
                return;
            }

            var newAttributesArr = [];
            for (var commentIndex = 0; commentIndex < commentIDs.length; ++commentIndex) {

                // important: the _internalGetCommentsData DO NOT return data in the same order commentIDs are sent as input!!
                var queryResultIndex = obj.comments.findIndex(function (comment) {
                    return comment.ID === commentIDs[commentIndex]
                }.bind(this));

                var oldAttributes = obj.comments[queryResultIndex].ATTRIBUTES;
                var userID = obj.comments[queryResultIndex].USERID;
                var commentID = obj.comments[queryResultIndex].ID;
                var commentData = obj.comments[queryResultIndex].DATA;
                var changed = false;
                var realAttribute;
                var userIndex;

                var newAttributes = {};
                var data;
                for (var attribute in oldAttributes) {
                    newAttributes[attribute] = oldAttributes[attribute]
                }

                if (!canUserUpdateCommentAttributes(this.userInfo.name, this.userRole, attributes[commentIndex], obj.comments[queryResultIndex].USERID)) {
                    callback({ error: CommentPermissionError });
                    return;
                }

                // smart attributes management
                for (var attributeSetter in attributes[commentIndex]) {

                    var attributeValue =  attributes[commentIndex][attributeSetter];

                    if (attributeSetter === Consts.CommentAttributesSetters.setReadStatus) {
                        realAttribute = newAttributes[Consts.CommentAttributes.ReadBy];
                        userIndex = -1;
                        for (j=0; j<realAttribute.length; j++) {
                            if (realAttribute[j] == this.userInfo.name) {
                                userIndex = j;
                                break;
                            }
                        }

                        if (attributeValue === true && userIndex === -1) {
                            realAttribute.push(this.userInfo.name);
                            changed = true;
                        }
                        else if (attributeValue === false && userIndex !== -1) {
                            realAttribute.splice(userIndex, 1);
                            changed = true;
                        }
                        break;
                    }
                    else if (attributeSetter === Consts.CommentAttributesSetters.setLikedStatus) {
                        realAttribute = newAttributes[Consts.CommentAttributes.LikedBy];

                        userIndex = -1;
                        for (j=0; j<realAttribute.length; j++) {
                            if (realAttribute[j] == this.userInfo.name) {
                                userIndex = j;
                                break;
                            }
                        }

                        if (attributeValue === true && userIndex === -1) {
                            realAttribute.push(this.userInfo.name);
                            changed = true;
                        } else if (attributeValue === false && userIndex !== -1) {
                            realAttribute.splice(userIndex, 1);
                            changed = true;
                        }

                        // remove user from readBy list to notify it that someone liked its comment
                        if (changed && userID !== this.userInfo.name) {
                            realAttribute = newAttributes[Consts.CommentAttributes.ReadBy];

                            for (j=0; j<realAttribute.length; j++) {
                                if (realAttribute[j] == userID) {
                                    realAttribute.splice(j, 1);
                                    break;
                                }
                            }
                        }

                        break;
                    }
                    else if (attributeSetter === Consts.CommentAttributesSetters.setTrashedStatus) {
                        if (oldAttributes[Consts.CommentAttributes.Trashed] === false && attributeValue === true) {

                            newAttributes[Consts.CommentAttributes.Trashed] = attributeValue;
                            newAttributes[Consts.CommentAttributes.TrashedBy] = this.userInfo.name;
                            changed = true;
                        }
                        else if (oldAttributes[Consts.CommentAttributes.Trashed] === true && attributeValue === false) {
                            newAttributes[Consts.CommentAttributes.Trashed] = attributeValue;
                            newAttributes[Consts.CommentAttributes.TrashedBy] = "";
                            data = commentData;
                            changed = true;
                        }
                    }
                    else if (attributeSetter === Consts.CommentAttributesSetters.setTimestamps) {
                        if (!newAttributes[Consts.CommentAttributes.Timestamps]) {
                            newAttributes[Consts.CommentAttributes.Timestamps] = {};
                        }

                        tagList = attributeValue;
                        for (index = 0; index < tagList.length; ++index) {
                            tag = tagList[index];
                            newAttributes[Consts.CommentAttributes.Timestamps][tag] = Date.now();
                        }
                        changed = true;
                    }
                }

                if (changed) {
                    newAttributesArr.push( {
                        commentID: commentID,
                        attributes: newAttributes,
                        attributesChangedData: data
                    })
                }
            }

            if (newAttributesArr.length > 0) {
                    this._recurseSetCommentsAttributes(newAttributesArr, 0, function (obj) {
                        if (obj.error) {
                            callback(obj);
                        }
                        else if (obj.affectedRows !== newAttributesArr.length) {
                            callback({ error: "not all changes affected rows", affectedRows: 0});
                        }
                        else {
                            callback({affectedRows: obj.affectedRows}, newAttributesArr);
                        }
                    }.bind(this));
            }
            else {
                callback({affectedRows: commentIDs.length}, newAttributesArr);
            }

        }
    }.bind(this));
};

BalsamiqArchive.prototype._recurseSetCommentsAttributes = function(attributes, affectedRows, callback) {

    this._internalSetCommentAttributes(attributes[0].commentID, attributes[0].attributes, function (obj) {
        if (obj.error || obj.affectedRows == 0) {
            callback(obj);
            return;
        }

        var totalAffectedRows = obj.affectedRows + affectedRows;

        var remainingAttributesToSet = attributes.slice(1);
        if (remainingAttributesToSet.length == 0) {
            callback( { affectedRows: totalAffectedRows });
        } else {
            this._recurseSetCommentsAttributes(remainingAttributesToSet, obj.affectedRows + affectedRows, callback);
        }
    }.bind(this));
};

// returns JSON of attributes
BalsamiqArchive.prototype._internalGetCommentAttributes = function(commentID, callback) {

    if (!this.checkInternalCall(callback)) {
        return;
    }

    this.dbDriver.get(this.archiveID, "SELECT ATTRIBUTES FROM COMMENTS WHERE ID = ?", [commentID], function (obj) {
        if (obj.error) {
            callback(obj);
        }
        else {
            try {
                callback({ attributes: JSON.parse(obj.row.ATTRIBUTES)});

            } catch (e) {
                callback({error: "Balsamiq Archive error: Unexpected database response while getting comment attributes: " + JSON.stringify(obj)});
            }
        }
    });
};

// newAttributes: object to be stringyfied
BalsamiqArchive.prototype._internalSetCommentAttributes = function(commentID, newAttributes, callback) {

    if (!this.checkInternalCall(callback)) {
        return;
    }

    this.dbDriver.run(this.archiveID, "UPDATE COMMENTS SET ATTRIBUTES = ? WHERE ID = ?", [JSON.stringify(newAttributes), commentID], function (obj) {
        callback(obj);
    });
};

BalsamiqArchive.prototype.setCommentData = function(commentID, data, callback) {

    this._enterCall();

    this._internalGetCommentsData([commentID], false, function(obj) {

        if (obj.error) {
            this._exitCall();
            this._finalCallbackWithError(callback, obj);
        } else {
            if (!canUserSetCommentData(this.userInfo.name, this.userRole, obj.comments[0].USERID)) {
                this._exitCall();
                this._getArchiveRevision(function (archiveRevision) {
                    this._finalCallback(callback, archiveRevision, {});
                }.bind(this));
                return;
            }
            var oldAttributes = obj.comments[0].ATTRIBUTES;
            // var userID = obj.comments[0].USERID;

            var newAttributes = {};
            for (var attribute in oldAttributes) {
                newAttributes[attribute] = oldAttributes[attribute]
            }

            // clear readby attributes with the exception of current user which edited it
            newAttributes[Consts.CommentAttributes.ReadBy] = [this.userInfo.name];

            this.dbDriver.run(this.archiveID, "UPDATE COMMENTS SET DATA = ?, ATTRIBUTES = ? WHERE ID = ?;", [data, JSON.stringify(newAttributes), commentID], function (obj) {
                this._exitCall();
                if (obj.error) {
                    this._finalCallbackWithError(callback, obj);
                }
                else {
                    this._finalCallbackWithAffectedRowsCheck(callback, obj, {
                        attributes: newAttributes,
                        data: data
                    });
                }
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype.deleteComments = function(commentIDs, callback) {

    this._enterCall();

    this._internalGetCommentsData(commentIDs, false, function (obj) {
        if (obj.error) {
            this._exitCall();
            this._finalCallbackWithError(callback, obj);
        } else {
            for (var i = 0; i < obj.comments.length; i++) {
                if (!canUserDeleteComment(this.userInfo.name, this.userRole, obj.comments[i].USERID)) {
                    this._exitCall();
                    this._finalCallbackWithError(callback, { error: CommentPermissionError })
                    return;
                }
            }
            makeQueryInBatches(this.dbDriver, 'run', this.archiveID, "DELETE FROM COMMENTS WHERE ID in ([!!__VALUES__!!])", [], commentIDs, function (obj) {
                this._exitCall();
                if (obj.error) {
                    this._finalCallbackWithError(callback, obj);
                } else {
                    this._finalCallbackWithAffectedRowsCheck(callback, obj);
                }
            }.bind(this));
        }
    }.bind(this));
};


// get array of users {id, attributes}
BalsamiqArchive.prototype.getArchiveUsersList = function(callback) {
    this._enterCall();
    // var res = { users: [] };
    this._internalGetArchiveUsersList(function (obj) {

        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        } else {
            this._getArchiveRevision(function (p_pv) {
                this._finalCallback(callback, p_pv, obj);
            }.bind(this));
        }
    }.bind(this));
};

// internal usage only
BalsamiqArchive.prototype.reCreateUser = function(userID, attributes, callback) {

    this._enterCall();
    this.dbDriver.run(this.archiveID, "INSERT INTO USERS (ID, ATTRIBUTES) VALUES (?, ?);", [ userID, JSON.stringify(attributes)], function (obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        }
        else {
            this._getArchiveRevision(function (p_pv) {
                this._finalCallback(callback, p_pv, obj);
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype.createMyUser = function(updateAttributes, callback) {

    this._enterCall();

    this._internalInsertUserIfNeeded(updateAttributes, function(obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        }
        else {
            this._getArchiveRevision(function (p_pv) {
                this._finalCallback(callback, p_pv, obj);
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype.updateMyUser = function(updateAttributes, callback) {

    this._enterCall();

    this._internalUpdateUser(updateAttributes, function(obj) {
        this._exitCall();
        if (obj.error) {
            this._finalCallbackWithError(callback, obj);
        }
        else {
            this._getArchiveRevision(function (p_pv) {
                this._finalCallback(callback, p_pv, obj);
            }.bind(this));
        }
    }.bind(this));
};

BalsamiqArchive.prototype.getHeuristicArchiveSize = function(callback) {
    this._enterCall();

    this.dbDriver.all(this.archiveID, "SELECT SUM(LENGTH(DATA)) AS sizeInBytes FROM RESOURCES", [], function (obj) {
        var size;
        if (obj.error) {
            this._exitCall();
            this._finalCallbackWithError(callback, obj);
        } else {
            if (obj.rows) {
                size = obj.rows[0]['sizeInBytes'];
                this.dbDriver.all(this.archiveID, "SELECT SUM(LENGTH(ATTRIBUTES)) AS sizeInBytes FROM THUMBNAILS", [], function (obj) {
                    this._exitCall();
                    if (obj.rows) {
                        size += obj.rows[0]['sizeInBytes'];
                        this._getArchiveRevision(function (p_pv) {
                            this._finalCallback(callback, p_pv, {heuristicArchiveSize: size});
                        }.bind(this));
                    } else {
                        this._finalCallbackWithError(callback, {error: "unexpected result from query"});
                    }
                }.bind(this));
            } else {
                this._exitCall();
                this._finalCallbackWithError(callback, {error: "unexpected result from query"});
            }
        }
    }.bind(this));
};

BalsamiqArchive.prototype._internalGetArchiveUsersList = function(callback) {

    if (!this.checkInternalCall(callback)) {
        return;
    }

    var res = { users: [] };
    this.dbDriver.all(this.archiveID, "SELECT ID, ATTRIBUTES FROM USERS", [], function(obj) {
        var i, row, attrs;
        if (obj.error) {
            callback(obj);
        } else {
            try {
                for (i=0; i<obj.rows.length; i++) {
                    row = obj.rows[i];
                    attrs = JSON.parse(row.ATTRIBUTES);
                    res.users.push({
                        ID: row.ID,
                        ATTRIBUTES: attrs
                    });
                }
                callback(res);
            } catch (e) {
                callback({error: "Balsamiq Archive error: Unexpected database response while getting users list: " + JSON.stringify(obj)});
            }
        }
    }.bind(this));
};

BalsamiqArchive.prototype._internalGetCommentsToc = function(callback) {

    // TODO: no check here, debug only function
    // if (!this.checkInternalCall(callback)) {
    //     return;
    // }

    var res = { comments: [] };
    this.dbDriver.all(this.archiveID, "SELECT ID, BRANCHID, RESOURCEID, USERID, ATTRIBUTES FROM COMMENTS", [], function (obj) {
        var i, row;

        if (obj.error) {
            callback(obj);
        } else {
            try {
                for (i=0; i<obj.rows.length; i++) {
                    row = obj.rows[i];
                    // var attrs = JSON.parse(row.ATTRIBUTES);
                    res.comments.push({
                        ID: row.ID,
                        BRANCHID: row.BRANCHID,
                        RESOURCEID: row.RESOURCEID,
                        USERID: row.USERID,
                        ATTRIBUTES: JSON.parse(row.ATTRIBUTES)
                    });
                }
                callback(res);
            } catch (e) {
                callback({error: "Balsamiq Archive error: Unexpected database response while getting comments TOC: " + JSON.stringify(obj)});
            }
        }
    }.bind(this));
};

BalsamiqArchive.prototype._internalGetComments = function(callback) {

    // TODO: no check here, debug only function
    // if (!this.checkInternalCall(callback)) {
    //     return;
    // }

    var res = { comments: [] };
    this.dbDriver.all(this.archiveID, "SELECT ID, BRANCHID, RESOURCEID, USERID, ATTRIBUTES, DATA FROM COMMENTS", [], function (obj) {
        var i, row;
        if (obj.error) {
            callback(obj);
        } else {
            try {
                for (i=0; i<obj.rows.length; i++) {
                    row = obj.rows[i];
                    // var attrs = JSON.parse(row.ATTRIBUTES);
                    res.comments.push({
                        ID: row.ID,
                        BRANCHID: row.BRANCHID,
                        RESOURCEID: row.RESOURCEID,
                        USERID: row.USERID,
                        ATTRIBUTES: JSON.parse(row.ATTRIBUTES),
                        DATA: row.DATA          // this is a string! as for input value
                    });
                }
                callback(res);
            } catch (e) {
                callback({error: "Balsamiq Archive error: Unexpected database response while getting comments TOC: " + JSON.stringify(obj)});
            }
        }
    }.bind(this));
};

// call this function at the beginning of each API
BalsamiqArchive.prototype._enterCall = function () {
    // in case the busy guard is inconsistent we restart the server
    if (!this.dbDriver) {
        throw new Error("Balsamiq Archive error: no DBDriver db is open!");
    }
    if (this.busy) {
        throw new Error("Balsamiq Archive error: I am already busy, you can't call me now!");
    }
    this.busy = true;
    if (this.dbDriver.beginSequence) {
        this.dbDriver.beginSequence(this.archiveID);
    }
};

// call this function at the end of each API
BalsamiqArchive.prototype._exitCall = function () {
    // in case the busy guard is inconsistent we restart the server
    if (!this.busy) {
        throw new Error("Balsamiq Archive error: _exitCall called when this.busy is false!");
    }
    this.busy = false;
};

// invoke the user callback returning the archive Revision and an optional result
BalsamiqArchive.prototype._finalCallback = function (callback, p_archiveRevision, p_result) {
    if (this.dbDriver.endSequence) {
        this.dbDriver.endSequence(this.archiveID);
    }
    var obj;
    if (p_archiveRevision.error) {
        callback && callback(p_archiveRevision);
    } else {
        obj = p_result ? p_result : {};
        obj.archiveRevision = p_archiveRevision;
        callback && callback(obj);
    }
};

// invoke the user callback returning an error
BalsamiqArchive.prototype._finalCallbackWithError = function (callback, p_error) {
    if (this.dbDriver.endSequence) {
        this.dbDriver.endSequence(this.archiveID);
    }
    if (callback) {
        callback(p_error);
    }
};

// invoke the user callback returning an error
BalsamiqArchive.prototype._finalCallbackWithAffectedRowsCheck = function (callback, sqlCommandResult, callbackResult) {
    if (sqlCommandResult.affectedRows >= 1) {
        this._incrementArchiveRevision(function (p_pv) {
            this._finalCallback(callback, p_pv, callbackResult);
        }.bind(this));
    } else {
        this._finalCallbackWithError(callback, {error: "Command affected no rows"});
    }
};

// read and return the archive Revision
BalsamiqArchive.prototype._getArchiveRevision = function (callback) {
    // no need to check this.busy here

    // TODO no check for an internal call here

    this.dbDriver.get(this.archiveID, "SELECT VALUE FROM INFO WHERE NAME = ?;", [Consts.Info.ArchiveRevision], function (obj) {
        var resp;
        if (obj.error) {
            callback && callback(obj);
        }
        else {
            try {
                resp = parseInt(obj.row.VALUE, 10);
            }
            catch (e) {
                resp = {error: "Balsamiq Archive error: Unexpected database response while getting archive revision: " + JSON.stringify(obj)};
            }
            callback && callback(resp);
        }
    });
};

// increment and return the archive Revision
BalsamiqArchive.prototype._incrementArchiveRevision = function (callback) {

    // TODO no check for an internal call here
    // if (!this.checkInternalCall(callback)) {
    //     return;
    // }

    // no need to check this.busy here
    this.dbDriver.run(this.archiveID, "UPDATE INFO SET VALUE = VALUE + 1 WHERE NAME = ?;", [Consts.Info.ArchiveRevision], function (obj) {
        if (obj.error) {
            callback && callback(obj);
        }
        else {
            this._getArchiveRevision(callback);
        }
    }.bind(this));
};

// read the value of the given metadata key
BalsamiqArchive.prototype._getArchiveData = function (p_key, callback, p_resultKey, p_isJSON) {

    this._enterCall();
    this._internalGetArchiveData(p_key, p_resultKey, p_isJSON, function (res) {
        this._exitCall();
        if (res.error) {
            this._finalCallbackWithError(callback, res);
        } else {
            this._getArchiveRevision(function (p_pv) {
                this._finalCallback(callback, p_pv, res);
            }.bind(this));
        }
    }.bind(this));
};

// no need to check for busy here
BalsamiqArchive.prototype._internalGetArchiveData = function (p_key, p_resultKey, p_isJSON, callback) {

    if (!this.checkInternalCall(callback)) {
        return;
    }

    if (!Consts.Info[p_key]) {
        callback({error: "Balsamiq Archive error: using unknown key " + p_key + " while getting archive data"});
    } else {

        this.dbDriver.get(this.archiveID, "SELECT NAME, VALUE FROM INFO WHERE NAME = ?;", [p_key], function (obj) {

            if (obj.error) {
                callback(obj);
            } else {
                var res = {};
                if (obj.row) {
                    res[p_resultKey] = p_isJSON ? JSON.parse(obj.row.VALUE) : obj.row.VALUE;
                }
                callback(res);

            }
        }.bind(this));
    }
};

// set the value of the given metadata key
BalsamiqArchive.prototype._setArchiveData = function (p_key, p_value, callback) {
    if (!Consts.Info[p_key]) {
        this._finalCallbackWithError(callback, {error: "Balsamiq Archive error: using unknown key " + p_key + " while setting archive data"});
    } else {
        this._enterCall();

        this.dbDriver.run(this.archiveID, "REPLACE INTO INFO (NAME, VALUE) VALUES (?, ?);", [p_key, p_value], function (obj) {
            this._exitCall();
            if (obj.error) {
                this._finalCallbackWithError(callback, obj);
            } else {
                this._finalCallbackWithAffectedRowsCheck(callback, obj);
            }
        }.bind(this));
    }
};

// TODO no enterCall/existCall: internal usage only - no need to check this.busy
BalsamiqArchive.prototype._migrate1To2 = function(callback) {

    if (!this.checkInternalCall(callback)) {
        return;
    }

    if (this.dbDriver.run) {
        this._internalGetArchiveData(Consts.Info.SchemaVersion, Consts.Info.SchemaVersion, false, function(obj) {
            var query;
            if (obj.error) {
                callback(obj);
            } else {
                if (obj[Consts.Info.SchemaVersion] === Consts.CurrentSchemaVersion) {
                    callback({error: "Balsamiq Archive error: attempt to upgrade new database schema"});
                }
                else {
                    if (this.dbDriver.type == "mysql") {
                        query = "CREATE TABLE USERS (ID VARCHAR(255) PRIMARY KEY, ATTRIBUTES TEXT CHARACTER SET 'utf8mb4' COLLATE 'utf8mb4_unicode_ci');";
                    } else {
                        query = "CREATE TABLE USERS (ID VARCHAR(255) PRIMARY KEY, ATTRIBUTES TEXT);";
                    }
                    this.dbDriver.run(this.archiveID, query, {}, function(obj) {
                        if (obj.error) {
                            callback(obj);
                        } else {
                            if (this.dbDriver.type == "mysql") {
                                query = "CREATE TABLE COMMENTS (ID VARCHAR(255) PRIMARY KEY, BRANCHID VARCHAR(255), RESOURCEID VARCHAR(255), DATA LONGTEXT CHARACTER SET 'utf8mb4' COLLATE 'utf8mb4_unicode_ci', USERID VARCHAR(255), ATTRIBUTES TEXT CHARACTER SET 'utf8mb4' COLLATE 'utf8mb4_unicode_ci', FOREIGN KEY (USERID) REFERENCES USERS(ID), FOREIGN KEY (RESOURCEID, BRANCHID) REFERENCES RESOURCES(ID, BRANCHID));";
                            } else {
                                query = "CREATE TABLE COMMENTS (ID VARCHAR(255) PRIMARY KEY, BRANCHID VARCHAR(255), RESOURCEID VARCHAR(255), DATA LONGTEXT, USERID VARCHAR(255), ATTRIBUTES TEXT, FOREIGN KEY (USERID) REFERENCES USERS(ID), FOREIGN KEY (RESOURCEID, BRANCHID) REFERENCES RESOURCES(ID, BRANCHID));";
                            }
                            this.dbDriver.run(this.archiveID, query, {}, function(obj) {
                                if (obj.error) {
                                    callback(obj);
                                } else {
                                    this.dbDriver.run(this.archiveID, "UPDATE INFO SET VALUE = ? WHERE NAME = ?;", [Consts.CurrentSchemaVersion, Consts.Info.SchemaVersion], function(obj) {
                                        if (obj.error) {
                                            callback(obj);
                                        } else {
                                            // IMPORTANT: avoid to increase the revision number after the migration
                                            // After incrementing the revision number the BMPR cannot be unloaded in case it is opened by user with read only permissions (i.e. not able to save to platform)
                                            // this.dbDriver.run(this.archiveID, "UPDATE INFO SET VALUE = VALUE + 1 WHERE NAME = ?;", [Consts.Info.ArchiveRevision], function (obj) {
                                                obj.migrated = true;
                                                callback(obj);
                                            // }.bind(this))
                                        }
                                    }.bind(this));
                                }
                            }.bind(this));
                        }
                    }.bind(this));
                }
            }
        }.bind(this));
    } else {
        // we are in read only mode (e.g. db opened via createFromURL)
        callback({error: Consts.ErrorCodes.NeedToMigrate})
    }
};

BalsamiqArchive.prototype.checkInternalCall = function(callback) {
    if (!this.busy) {
        callback({ error: "internal call require previous enterCall" });
        return false;
    }

    return true;
};

var getThumbnailFromDump = function (resourceID, branchID, dump, callback) {
    var i, attributes, ret = {};
    ret.projectName = dump.Info.ArchiveAttributes.name;
    ret.archiveRevision = dump.Info.ArchiveRevision;
    for (i=0; i<dump.Resources.length; i++) {
        attributes = dump.Resources[i].ATTRIBUTES;
        if (dump.Resources[i].ID === resourceID && dump.Resources[i].BRANCHID === branchID) {
            ret.resourceName = attributes.name;
            break;
        }
    }

    for (i=0; i<dump.Thumbnails.length; i++) {
        attributes = dump.Thumbnails[i].ATTRIBUTES;
        if (attributes.resourceID === resourceID && attributes.branchID === branchID) {
            ret.image = attributes.image;
            break;
        }
    }
    callback && callback(ret);
};

function makeSqlPlaceholdersForArray(arr) {
    return arr.map(function () { return '?'; }).join(', ');
}


var QUERY_MAX_NUM_VALUES = 500;

function makeQueryInBatches(dbDriver, methodName, dbName, queryTemplate, fixedValues, values, callback) {
    var result = {
        rows: [],
        affectedRows: 0
    };
    if (values.length === 0) {
        callback(result);
        return;
    }
    var driverMethod = dbDriver[methodName].bind(dbDriver);
    var batches = splitArrayInChunks(values, QUERY_MAX_NUM_VALUES);

    var processNextBatch = function () {
        var values = batches[0];
        batches = batches.slice(1);
        var query = queryTemplate.replace('[!!__VALUES__!!]', makeSqlPlaceholdersForArray(values));
        driverMethod(dbName, query, fixedValues.concat(values), function (obj) {
            if (obj.error) {
                callback(obj);
            } else {
                result.affectedRows += obj.affectedRows;
                result.rows = result.rows.concat(obj.rows);
                if (batches.length > 0) {
                    processNextBatch();
                } else {
                    callback(result);
                }
            }
        });
    };
    processNextBatch();
}

function splitArrayInChunks(arr, chunkSize) {
    var chunks = [];
    for (var i=0; i<arr.length; i += chunkSize) {
        chunks.push(arr.slice(i, i+chunkSize));
    }
    return chunks;
}

function canUserUpdateCommentAttributes(userID, userRole, attributesToUpdate, commentUserID) {
    if (userRole < Consts.Role.ROLE_COMMENTER) {
        return false;
    }

    var isCommentAuthor = userID === commentUserID;
    var isCommenter = userRole === Consts.Role.ROLE_COMMENTER;

    for (var attributeSetter in attributesToUpdate) {
        switch (attributeSetter) {
            case Consts.CommentAttributesSetters.setReadStatus:
            case Consts.CommentAttributesSetters.setLikedStatus:
                break;
            case Consts.CommentAttributesSetters.setTrashedStatus:
            case Consts.CommentAttributesSetters.setTimestamps:
                if (isCommenter && !isCommentAuthor) {
                    return false;
                }
                break;
            default:
                return false;
        }
    }

    return true;
}

function canUserSetCommentData(userID, userRole, commentUserID) {
    if (userRole < Consts.Role.ROLE_COMMENTER) {
        return false;
    }

    return userID === commentUserID;
}

function canUserDeleteComment(userID, userRole, commentUserID) {
    if (userRole < Consts.Role.ROLE_COMMENTER) {
        return false;
    }

    if (userRole == Consts.Role.ROLE_COMMENTER) {
        return userID === commentUserID;
    }

    // EDITOR/ADMIN can delete all comments
    return true;
}

exports.MasterBranchID = Consts.Branch.MasterBranchID;
exports.getThumbnailFromDump = getThumbnailFromDump;
exports.BalsamiqArchive = BalsamiqArchive;
exports.canUserUpdateCommentAttributes = canUserUpdateCommentAttributes;
exports.canUserSetCommentData = canUserSetCommentData;
exports.canUserDeleteComment = canUserDeleteComment;


/***/ }),

/***/ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveClient.js":
/*!******************************************************************!*\
  !*** ./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveClient.js ***!
  \******************************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

/*global Uint8Array, */
//Documentation: https://balsamiq.atlassian.net/wiki/display/intranet/Balsamiq+Archive+Master+API+list

var Consts = __webpack_require__(/*! ./BalsamiqArchiveConstants */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js");

var urlEncodeObject = function(obj) {
    var str = [];
    for(var p in obj)
        if (obj.hasOwnProperty(p)) {
            str.push(encodeURIComponent(p) + "=" + encodeURIComponent(obj[p]));
    }
    return str.join("&");
};

var parseHeaders = function(headers) {
    var arr = headers.trim().split(/[\r\n]+/);

    var headerMap = {};
    arr.forEach(function (line) {
      var parts = line.split(': ');
      var header = parts.shift();
      var value = parts.join(': ');
      headerMap[header] = value;
    });

    return headerMap;
}

var defaultNetworkApi = function(method, url, data, callback, timeouts, settings, requestCreatedCallback) {
    var xhr, parseError, responseTimeoutTimer;

    if (method === 'GET_SYNC') {
        method = 'GET';  // Async request are deprecated (and blocked by some browsers
    }

    xhr = new XMLHttpRequest();

    if (settings && settings.binary) {
        xhr.responseType = 'arraybuffer';
    }

    xhr.onreadystatechange = function() {
        var responseObj;

        if (xhr.readyState >= 2 && responseTimeoutTimer) {
            clearTimeout(responseTimeoutTimer);
        }

        if (xhr.readyState !== XMLHttpRequest.DONE) {
            return; // not interested in other events
        }

        if (settings && settings.binary) {
            // get binary data as a response
            if (settings.base64) {
                responseObj = { data : _arrayBufferToBase64(xhr.response) };
            } else {
                responseObj = { data : xhr.response };
            }
        } else {
            try {
                responseObj = JSON.parse(xhr.responseText);
            } catch (e) {
                parseError = { error: xhr.status, error_message: 'INVALID JSON RESPONSE CONTENT' };
            }
        }

        if (settings && settings.progressCallback) {
            xhr.onprogress = function (e) {
                if (e.lengthComputable) {
                    settings.progressCallback(true, e.loaded, e.total);
                } else {
                    settings.progressCallback(false);
                }
            };
        }

        var headers = parseHeaders(xhr.getAllResponseHeaders());
        if (parseError) {
            callback(parseError, headers);
        } else {
            if (xhr.status >= 200 && xhr.status < 400) {
                callback(responseObj, headers);
            } else {
                callback({
                    error: xhr.status,
                    error_message: responseObj.message || responseObj.error || "networkApi error"
                }, headers);
            }
        }
    };

    xhr.ontimeout = function timeout(/*e*/) {
        if (responseTimeoutTimer) {
            clearTimeout(responseTimeoutTimer);
        }
        callback({error_message: 'DEADLINE_TIMEOUT', timeout: true});
    };

    xhr.open(method, url);
    if (settings && settings.headers) {
        for (var header in settings.headers) {
            xhr.setRequestHeader(header, settings.headers[header])
        }
    }
    if (timeouts.deadline) xhr.timeout = timeouts.deadline;
    if (settings && settings.formData) {
        var formData = new FormData();
        for (var i in data) {
            var field = data[i];
            var name = field.name,
                value = field.value,
                type = field.type,
                mimeType = field.mimeType,
                filename = field.filename;
            if (type === 'base64') {
                var binary = atob(value);
                var array = new Uint8Array(binary.length);
                for(var j = 0; j < binary.length; j++) {
                    array[j] = binary.charCodeAt(j);
                }
                formData.append(name, new Blob([array], {type: mimeType}), filename);
            } else {
                formData.append(name, value);
            }
        }
        xhr.send(formData);
    } else {
        xhr.setRequestHeader('Content-Type', 'application/x-www-form-urlencoded; charset=UTF-8');
        xhr.setRequestHeader('Accept', 'application/json');
        xhr.send(urlEncodeObject(data));
    }
    if (timeouts.response) {
        responseTimeoutTimer = setTimeout(function responseTimeout() {
            xhr.abort();
            callback({ error_message: 'RESPONSE_TIMEOUT', timeout: true });
        }, timeouts.response);
    }
    requestCreatedCallback && requestCreatedCallback(xhr);
};

var defaultTimerApi = {};

defaultTimerApi.setTimeout = function () {
    setTimeout.apply(null, arguments);
};

defaultTimerApi.clearTimeout = function () {
    clearTimeout.apply(null, arguments);
};

var BalsamiqArchiveClient = function(serverURL, kind, platformToken, userInfo, networkApi, timerApi, timeouts)
{
    this.serverURL = serverURL;
    this.kind = kind;
    this.platformToken = platformToken;
    this.userInfo = userInfo;
    this.networkApi = networkApi ? networkApi : defaultNetworkApi;
    this.timerApi = timerApi ? timerApi : defaultTimerApi;
    this.timeouts = timeouts || { response: null, deadline: null };

    this.token = "";
    this.userID = "";
    this.rtcChannel = "";
    this.archiveRevision = null;

    this.downloadRequest = null;
    this.downloadCallback = null;

    this.intervalKeepAlive = null;
};

function _arrayBufferToBase64(buffer) {
    var binary = '';
    var bytes = new Uint8Array(buffer);
    var len = bytes.byteLength;
    for (var i = 0; i < len; i++) {
        binary += String.fromCharCode(bytes[i]);
    }
    return window.btoa( binary );
}

BalsamiqArchiveClient.prototype.isLocal = function() {
    return false;
};

BalsamiqArchiveClient.prototype.getKind = function () {
    return this.kind;
};

BalsamiqArchiveClient.prototype.fetch = function(url, callback)
{
    this._getBinary("fetch", "url=" + encodeURI(url), Consts.BinaryDataType.BASE64, function(data) {
        if(!data.error){
            var file = {
                data: data,
                filename: url, //deprecated
                url: url,
                type: ""
            };

            callback(file);
        } else {
            callback(data);
        }
    });
};

BalsamiqArchiveClient.prototype.create = function(platformSiteID, platformArchiveName, platformInfo, callback)
{
    this.platformInfo = platformInfo;

    var postObj = {
        platformToken:this.platformToken,
        platformSiteID:platformSiteID,
        platformArchiveName:platformArchiveName,
        platformInfo:JSON.stringify(this.platformInfo),
        kind:this.kind
    };

    this.networkApi('POST', this.serverURL+"/create", postObj, function(obj) {
        callback && callback(obj);
    }, this.timeouts);
};

BalsamiqArchiveClient.prototype.createWithPlatformArchiveID = function(platformSiteID, platformArchiveID, platformArchiveName, platformInfo, archiveAttributes, callback)
{
    this.platformInfo = platformInfo;

    var postObj = {
        platformToken: this.platformToken,
        platformSiteID: platformSiteID,
        platformArchiveID: platformArchiveID,
        platformArchiveName: platformArchiveName,
        platformInfo: JSON.stringify(this.platformInfo),
        kind: this.kind
    };
    if (archiveAttributes) {
        postObj.archiveAttributes = JSON.stringify(archiveAttributes);
    }

    this.networkApi('POST', this.serverURL+"/create", postObj, function(obj) {
        callback && callback(obj);
    }, this.timeouts);
};

BalsamiqArchiveClient.prototype.startKeepAlive = function() {
    if (this.intervalKeepAlive) {
        this.timerApi.clearTimeout(this.intervalKeepAlive);
    }
    this.intervalKeepAlive = this.timerApi.setTimeout(this.keepAlive.bind(this), 10 * 60 * 1000);
};

BalsamiqArchiveClient.prototype.keepAlive = function() {
    this.intervalKeepAlive = null;
    this.refreshSession(this.platformToken, function (resp) {
        if (resp.error === 403) {
            console.log("BalsamiqArchiveClient keepAlive", resp.error_message);
        } else {
            // console.log("BalsamiqArchiveClient keepAlive", resp);
            this.startKeepAlive();
        }
    }.bind(this));
};

BalsamiqArchiveClient.prototype.setPlatformToken = function(token, refreshSession) {
    this.platformToken = token;
    // console.log("setPlatformToken... " + this.platformToken);

    if (refreshSession !== false) {
        this.refreshSession(this.platformToken, function (/*resp*/) {
            // console.log("setPlatformToken answer: " + resp.success);
            this.startKeepAlive();
        }.bind(this));
    }
};

BalsamiqArchiveClient.prototype.openByPublicShareID = function(publicShareID, options, callback)
{
    var branchID = (options && options.branchID) ? options.branchID : Consts.Branch.MasterBranchID;
    var skipThumbnailImage = options && options.skipThumbnailImage;

    var postObj = {
        publicShareID: publicShareID,
        branchID: branchID,
        skipThumbnailImage: skipThumbnailImage,
        userInfo:JSON.stringify(this.userInfo),
        platformInfo:JSON.stringify(this.platformInfo),
        platformToken:this.platformToken,
        kind:this.kind
    };

    this.networkApi('POST', this.serverURL+"/open", postObj, function(obj)
    {
        if (obj.error) {
            callback && callback(obj);
        } else {
            this.archiveRevision = obj.dump && obj.dump.Info && obj.dump.Info[Consts.Info.ArchiveRevision];
            this.token = obj.token;
            this.userID = obj.userID;
            this.rtcChannel = obj.rtcChannel;

            delete obj.token;
            this.startKeepAlive();
            callback && callback(obj);
        }
    }.bind(this), this.timeouts);
};

BalsamiqArchiveClient.prototype.open = function(platformSiteID, platformArchiveID, platformArchiveName, platformInfo, permissions, options, callback)
{
    var branchID = (options && options.branchID) ? options.branchID : Consts.Branch.MasterBranchID;
    var skipThumbnailImage = options && options.skipThumbnailImage;

    this.platformInfo = platformInfo;

    var postObj = {
        platformToken:this.platformToken,
        platformSiteID:platformSiteID,
        platformArchiveID:platformArchiveID,
        platformArchiveName:platformArchiveName,
        platformInfo:JSON.stringify(this.platformInfo),
        userInfo:JSON.stringify(this.userInfo),
        permissions: permissions,
        kind:this.kind,
        branchID: branchID,
        skipThumbnailImage: skipThumbnailImage
    };

    this.networkApi('POST', this.serverURL+"/open", postObj, function(obj)
    {
        if (obj.error) {
            callback && callback(obj);
        } else {
            this.archiveRevision = obj.dump && obj.dump.Info && obj.dump.Info[Consts.Info.ArchiveRevision];
            this.token = obj.token;
            this.userID = obj.userID;
            this.rtcChannel = obj.rtcChannel;
            this.basArchiveID = obj.basArchiveID;

            //Delete internal things
            // delete obj.userID; // we need the userID in order to correctly instantiate the rtc channel (e.g. presence)
            delete obj.token;
            this.startKeepAlive();
            callback && callback(obj);
        }
    }.bind(this), this.timeouts);
};

BalsamiqArchiveClient.prototype.unloadArchive = function(platformSiteID, platformArchiveID, permissions, callback)
{
    var postObj = {
        platformToken: this.platformToken,
        platformSiteID: platformSiteID,
        platformArchiveID: platformArchiveID,
        userInfo: JSON.stringify(this.userInfo),
        permissions: permissions,
        kind: this.kind
    };

    this.networkApi('POST', this.serverURL+"/unloadArchive", postObj, function(obj) {
        callback && callback(obj);
    }, this.timeouts);
};

BalsamiqArchiveClient.prototype.unloadArchiveIfNotSync = function(platformSiteID, platformArchiveID, permissions, modifiedTimestamp, callback)
{
    var postObj = {
        platformToken: this.platformToken,
        platformSiteID: platformSiteID,
        platformArchiveID: platformArchiveID,
        userInfo: JSON.stringify(this.userInfo),
        permissions: permissions,
        modifiedTimestamp: modifiedTimestamp,
        kind: this.kind
    };

    this.networkApi('POST', this.serverURL+"/unloadArchiveIfNotSync", postObj, function(obj) {
        callback && callback(obj);
    }, this.timeouts);
};

BalsamiqArchiveClient.prototype.close = function(callback)
{
    if (this.intervalKeepAlive) {
        this.timerApi.clearTimeout(this.intervalKeepAlive);
        this.intervalKeepAlive = null;
    }

    if (this.token) { // empty string, which is the default value, means the archive didn't open
        this._get("close", null, callback);
    } else {
        callback({});
    }
};

BalsamiqArchiveClient.prototype.closeAsyncNoFlush = function(callback)
{
    this.timerApi.clearTimeout(this.intervalKeepAlive);
    this._get("close", "noflush=true", callback);
};

BalsamiqArchiveClient.prototype.closeSync = function(callback)
{
    if (this.intervalKeepAlive) {
        this.timerApi.clearTimeout(this.intervalKeepAlive);
        this.intervalKeepAlive = null;
    }
    this._getSync("close", null, callback);
};

BalsamiqArchiveClient.prototype.closeWithBeacon = function()
{
    // This API should be used only when the editor is running in the browser
    // and a "beforeunload" event occurs.
    //
    // It uses the "sendBeacon" api to send the close event to BAS, without
    // blocking the browser with a synchronous xhr (that is deprecated).
    //
    // more at https://developer.mozilla.org/en-US/docs/Web/API/Navigator/sendBeacon
    //
    // If current browser doesn't support the sendBeacon API (like IE), "false" is returned.
    // The caller should always check for the return value and eventually call
    // the deprecated closeSync method when closeWithBeacon returns false.

    var url = this.serverURL+"/close?token="+this.token;
    var requestQueued = !!(
        typeof window !== 'undefined' &&
        window.navigator &&
        window.navigator.sendBeacon &&
        window.navigator.sendBeacon(url, null)
    );

    return requestQueued;
};

BalsamiqArchiveClient.prototype.closeArchiveWithoutFlushing = function(callback)
{
    if (this.intervalKeepAlive) {
        this.timerApi.clearTimeout(this.intervalKeepAlive);
        this.intervalKeepAlive = null;
    }
    this._getSync("close", "noflush=true", callback);
};

BalsamiqArchiveClient.prototype.flush = function(platformArchiveName, callback)
{
    var params = platformArchiveName ? "platformArchiveName=" + platformArchiveName : null;
    this._get("flush", params, callback);
};

BalsamiqArchiveClient.prototype.flushExt = function(opts, callback)
{
    var params = "", i;
    var keys = Object.keys(opts), key;

    for (i=0; i<keys.length; i++) {
        key = keys[i];
        params += (i ? "&" : "") + (key + "=" + opts[key]);
    }
    this._get("flush", params, callback);
};

BalsamiqArchiveClient.prototype.deleteArchive = function(platformSiteID, platformArchiveID, callback)
{
    var postObj = {
        platformToken:this.platformToken,
        platformSiteID:platformSiteID,
        platformArchiveID:platformArchiveID,
        kind:this.kind
    };

    this.networkApi('POST', this.serverURL+"/delete", postObj, function(obj) {
        callback && callback(obj);
    }, this.timeouts);
};

BalsamiqArchiveClient.prototype.deleteArchiveSync = function(platformSiteID, platformArchiveID, callback)
{
    var postObj = {
        platformToken:this.platformToken,
        platformSiteID:platformSiteID,
        platformArchiveID:platformArchiveID,
        kind:this.kind
    };

    this.networkApi('POST', this.serverURL+"/delete?token="+this.token, postObj, function (result) {
        if (result.error) {
            var message = result.error_message ? ("POST sync call failed " + result.error_message) : "Unknown error";
            callback && callback({error: message});
        } else {
            callback && callback(result);
        }
    }, this.timeouts);
};

BalsamiqArchiveClient.prototype.getRTCAuthToken = function(callback) {
    this.networkApi(
        'GET',
        this.serverURL+"/getRTCAuthToken",
        null,
        function (result) {
            callback(result);
        },
        this.timeouts,
        {
            headers: {
                Authorization: 'Bearer ' + this.token,
            },
        }
    );
};

BalsamiqArchiveClient.prototype.download = function (callback, progressCallback) {
    if (!this.downloadRequest) {
        this.downloadCallback = callback;
        this._getBinary("download", null, Consts.BinaryDataType.ARRAYBUFFER, function (blob) {
            this.downloadRequest = null;
            this.downloadCallback = null;
            callback(blob);
        }.bind(this), progressCallback, function(request) {
            this.downloadRequest = request;
        }.bind(this));
    }
};

BalsamiqArchiveClient.prototype.cancelDownload = function () {
    if (this.downloadRequest) {
        this.downloadRequest.abort(); // this will invoke normal callback with error
    }
};

BalsamiqArchiveClient.prototype.getArchiveFormat = function (callback)
{
    this._get("getArchiveFormat", null, callback);
};

BalsamiqArchiveClient.prototype.setArchiveFormat = function (format, callback)
{
    var postObj = {
        format:format,
        archiveRevision:this.archiveRevision
    };
    this._post("setArchiveFormat", postObj, callback);
};

BalsamiqArchiveClient.prototype.getArchiveRevision = function (callback)
{
    this._get("getArchiveRevision", null, callback);
};

BalsamiqArchiveClient.prototype.getArchiveAttributes = function (callback)
{
    this._get("getArchiveAttributes", null, callback);
};

BalsamiqArchiveClient.prototype.setArchiveAttributes = function (attributes, callback)
{
    var postObj = {
        attributes:JSON.stringify(attributes),
        archiveRevision:this.archiveRevision
    };
    this._post("setArchiveAttributes", postObj, callback);
};

BalsamiqArchiveClient.prototype.getBranches = function(callback)
{
    this._get("getBranches", null, callback);
};

BalsamiqArchiveClient.prototype.getBranchAttributes = function (branchID, callback)
{
    this._get("getBranchAttributes", "branchID="+branchID, callback);
};

BalsamiqArchiveClient.prototype.setBranchAttributes = function (branchID, attributes, callback)
{
    var postObj = {
        branchID:branchID,
        attributes: attributes ? JSON.stringify(attributes) : null,
        archiveRevision:this.archiveRevision
    };
    this._post("setBranchAttributes", postObj, callback);
};

BalsamiqArchiveClient.prototype.createResource = function(resourceID, branchID, attributes, data, callback)
{
    var postObj = {
        resourceID: resourceID,
        branchID:branchID,
        attributes:attributes ? JSON.stringify(attributes) : null,
        data: (typeof data === 'string' ? data : JSON.stringify(data)),
        archiveRevision:this.archiveRevision
    };
    this._post("createResource", postObj, callback);
};

BalsamiqArchiveClient.prototype.createBranch = function(branchID, attributes, callback)
{
    var postObj = {
        branchID: branchID,
        attributes: attributes ? JSON.stringify(attributes) : null,
        archiveRevision: this.archiveRevision
    };

    this._post("createBranch", postObj, callback);
};

BalsamiqArchiveClient.prototype.deleteResources = function(resourceIDs, branchID, callback)
{
    var postObj = {
        resourceIDs: JSON.stringify(resourceIDs),
        branchID:branchID,
        archiveRevision:this.archiveRevision
    };
    this._post("deleteResources", postObj, callback);
};

BalsamiqArchiveClient.prototype.deleteBranches = function(branchIDs, callback)
{
    var postObj = {
        branchIDs: JSON.stringify(branchIDs),
        archiveRevision: this.archiveRevision
    };
    this._post("deleteBranches", postObj, callback);
};

BalsamiqArchiveClient.prototype.getResourceAttributes = function(resourceID, branchID, callback)
{
    this._get("getResourceAttributes", "resourceID="+resourceID+"&branchID="+branchID, callback);
};

BalsamiqArchiveClient.prototype.setResourceAttributes = function (resourceID, branchID, attributes, callback)
{
    var postObj = {
        resourceID:resourceID,
        branchID:branchID,
        attributes:JSON.stringify(attributes),
        archiveRevision:this.archiveRevision
    };

    this._post("setResourceAttributes", postObj, callback);
};

BalsamiqArchiveClient.prototype.getTOC = function(branchID, callback)
{
    this._get("getTOC", "branchID="+branchID, callback);
};

BalsamiqArchiveClient.prototype.getResourceData = function(resourceID, branchID, callback)
{
    this.getResourceDataWithOptions(resourceID, branchID, {nocache: false}, callback);
};

BalsamiqArchiveClient.prototype.getResourceDataWithOptions = function(resourceID, branchID, options, callback)
{
    if (options.nocache || !this.basArchiveID) {
        this._get("getResourceData", "resourceID="+resourceID+"&branchID="+branchID, callback);
    } else {
        this.networkApi(
            'GET',
            this.serverURL+"/getResourceData?"+"basArchiveID="+this.basArchiveID+"&resourceID="+resourceID+"&branchID="+branchID,
            null,
            function (result, headers) {
                var revisionHeader = headers && (headers['x-balsamiq-archive-revision'] || headers['X-Balsamiq-Archive-Revision']);
                if (revisionHeader) {
                    result.archiveRevision = revisionHeader;
                }
                this._saveArchiveRevisionAndReturn(result, callback);
            }.bind(this),
            this.timeouts,
            {
                headers: {
                    Authorization: 'Bearer ' + this.token
                }
            }
        );
    }
};

BalsamiqArchiveClient.prototype.getResourcesData = function (resourceIDs, branchID, callback)
{
    var postObj = {
        resourceIDs:JSON.stringify(resourceIDs),
        branchID:branchID,
        archiveRevision:this.archiveRevision
    };

    this._post("getResourcesData", postObj, callback);
};

BalsamiqArchiveClient.prototype.setResourceData = function (resourceID, branchID, data, callback)
{
    var postObj = {
        resourceID:resourceID,
        branchID:branchID,
        data:data,
        archiveRevision:this.archiveRevision
    };
    this._post("setResourceData", postObj, callback);
};

BalsamiqArchiveClient.prototype.setResourceBranchID = function(resourceID, oldBranchID, newBranchID, callback)
{
    var postObj = {
        resourceID: resourceID,
        oldBranchID: oldBranchID,
        newBranchID:  newBranchID,
        archiveRevision: this.archiveRevision
    };

    this._post("setResourceBranchID", postObj, callback);
};

BalsamiqArchiveClient.prototype.createThumbnail = function(thumbnailID, attributes, callback) {
    var postObj = {
        thumbnailID:     thumbnailID,
        attributes:      JSON.stringify(attributes),
        archiveRevision: this.archiveRevision
    };
    this._post("createThumbnail", postObj, callback);
};

BalsamiqArchiveClient.prototype.getThumbnail = function (thumbnailID, callback) {
    this._get("getThumbnail", "thumbnailID="+thumbnailID, callback);
};

BalsamiqArchiveClient.prototype.setThumbnail = function (thumbnailID, attributes, callback) {
    var postObj = {
        thumbnailID: thumbnailID,
        attributes: JSON.stringify(attributes),
        archiveRevision:this.archiveRevision
    };

    this._post("setThumbnail", postObj, callback);
};

BalsamiqArchiveClient.prototype.deleteThumbnails = function(thumbnailIDs, callback) {

    var postObj = {
        thumbnailIDs: JSON.stringify(thumbnailIDs),
        archiveRevision: this.archiveRevision
    };

    this._post("deleteThumbnails", postObj, callback);
};

BalsamiqArchiveClient.prototype.createComment = function(commentID, resourceID, branchID, commentParentID, data, callback)
{
    var postObj = {
        commentID: commentID,
        resourceID: resourceID,
        branchID:branchID,
        commentParentID: commentParentID,
        data: (typeof data === 'string' ? data : JSON.stringify(data)), // data should be a string
        archiveRevision:this.archiveRevision
    };
    this._post("createComment", postObj, callback);
};

BalsamiqArchiveClient.prototype.importComment = function (resourceID, branchID, comment, callback) {
    var postObj = {
        resourceID: resourceID,
        branchID: branchID,
        comment: JSON.stringify(comment),
        archiveRevision: this.archiveRevision,
    };
    this._post("importComment", postObj, callback);
};

BalsamiqArchiveClient.prototype.getCommentsData = function(commentIDs, callback)
{
    var postObj = {
        commentIDs:JSON.stringify(commentIDs),
        archiveRevision:this.archiveRevision
    };

    this._post("getCommentsData", postObj, callback);
};

BalsamiqArchiveClient.prototype.setCommentData = function (commentID, data, callback)
{
    var postObj = {
        commentID:commentID,
        data: data,
        archiveRevision:this.archiveRevision
    };
    this._post("setCommentData", postObj, callback);
};

BalsamiqArchiveClient.prototype.updateCommentsAttributes = function(commentIDs, attributes, callback) {

    var postObj = {
        commentIDs:JSON.stringify(commentIDs),
        attributes:JSON.stringify(attributes),
        archiveRevision:this.archiveRevision
    };

    this._post("updateCommentsAttributes", postObj, callback);
};

BalsamiqArchiveClient.prototype.deleteComments = function(commentIDs, callback)
{
    var postObj = {
        commentIDs: JSON.stringify(commentIDs),
        archiveRevision:this.archiveRevision
    };
    this._post("deleteComments", postObj, callback);
};

BalsamiqArchiveClient.prototype.getArchiveUsersList = function(callback)
{
    this._get("getArchiveUsersList", null, callback);
};

BalsamiqArchiveClient.prototype.refreshSession = function (platformToken, callback)
{
    var postObj = {
        platformToken: platformToken
    };
    this._post("refreshSession", postObj, callback);
};

BalsamiqArchiveClient.prototype.getUsersList = function(option, callback)
{
    var query = option ? "option=" + option : "";
    this._get("getUsersList", query, callback);
};

BalsamiqArchiveClient.prototype.broadcastMessage = function (obj, callback)
{
    var postObj = {
        message: JSON.stringify(obj)
    };
    this._post("broadcastMessage", postObj, callback);
};

BalsamiqArchiveClient.prototype.getUserInfo = function(internalID, callback)
{
    this._get("getUserInfo", "internalID="+internalID, callback);
};

BalsamiqArchiveClient.prototype.updateUserInfo = function(userInfo, callback) {
    this._post("updateUserInfo", userInfo, callback);
};

BalsamiqArchiveClient.prototype.createMyUser = function(archiveUserInfo, callback) {

    var postObj = {
        archiveRevision:this.archiveRevision
    };

    if (archiveUserInfo)
        postObj.updatedUserInfo = JSON.stringify(archiveUserInfo);

    this._post("createMyUser", postObj, callback);
};

BalsamiqArchiveClient.prototype.updateMyUser = function(archiveUserInfo, callback) {
    var postObj = {
        archiveRevision:this.archiveRevision
    };

    if (archiveUserInfo) {
        postObj.updatedUserInfo = JSON.stringify(archiveUserInfo);
    }

    this._post("updateMyUser", postObj, callback);
};

BalsamiqArchiveClient.prototype.uploadTempFile = function(data, callback)
{
    this._post("uploadTempFile", data, callback);
};

BalsamiqArchiveClient.prototype.downloadTempFile = function(key, callback)
{
    var url = this.serverURL+"/downloadTempFile?token="+this.token + "&key=" + key;
    callback(url);
};

BalsamiqArchiveClient.prototype.downloadBmprlLink = function(filename, callback)
{
    var url = this.serverURL+"/download?token="+this.token + "&filename=" + filename;
    callback(url);
};

BalsamiqArchiveClient.prototype.logUserEvent = function (userEvent, callback)
{
    var postObj = {
        userEvent: JSON.stringify(userEvent)
    };
    this._post("logUserEvent", postObj, callback);
};

BalsamiqArchiveClient.prototype.getProjectMembers = function (data, callback)
{
    var postObj = {
        data: JSON.stringify(data)
    };
    this._post("getProjectMembers", postObj, callback);
};

// permalinkInfo.permalinkID
// permalinkInfo.permalinkKind
BalsamiqArchiveClient.prototype.setPermalinkData = function (resourceID, branchID, permalinkInfo, callback)
{
    var permalinkKind = permalinkInfo.permalinkKind || "image";
    delete permalinkInfo.permalinkKind;
    var postObj = {
        resourceID: resourceID,
        branchID: branchID,
        permalinkKind: permalinkKind,
        permalinkInfo: JSON.stringify(permalinkInfo)
    };
    // this._post("permalink", postObj, callback);
    this._post("setPermalinkData", postObj, callback);
};

BalsamiqArchiveClient.prototype.getPermalinkData = function (resourceID, branchID, callback) {
    this._get("getPermalinkData", "resourceID=" + resourceID + "&branchID=" + branchID, callback);
};

BalsamiqArchiveClient.prototype.getPermalinkDataByPermalinkID = function (permalinkID, callback) {
    this._get("getPermalinkData", "permalinkID=" + permalinkID, callback);
};

BalsamiqArchiveClient.prototype.getPermalinksData = function (callback) {
    this._get("getPermalinksData", null, callback);
};

BalsamiqArchiveClient.prototype.deletePermalinkData = function (resourceID, branchID, callback) {
    var postObj = {
        resourceID: resourceID,
        branchID: branchID
    };
    this._post("deletePermalinkData", postObj, callback);
};

BalsamiqArchiveClient.prototype.deletePermalinkDataByPermalinkID = function (permalinkID, callback)
{
    var postObj = {
        permalinkID: permalinkID
    };
    this._post("deletePermalinkData", postObj, callback);
};

BalsamiqArchiveClient.prototype.createOrUpdateImageLink = function (resourceID, branchID, permalinkInfo, base64data, mimeType, callback)
{
    var params = urlEncodeObject({resourceID: resourceID, branchID: branchID});
    // CAVEAT: Order of fields matters in multipart/formdata encoding. Do not shuffle!
    var fields = [{
        name: 'branchID',
        value: branchID,
    },{
        name: 'resourceID',
        value: resourceID,
    }];
    if (permalinkInfo) {
        fields.push({
            name: 'permalinkInfo',
            value: JSON.stringify(permalinkInfo || {}),
        })
    }
    if (base64data) {
        fields.push({
            name: 'image',
            value: base64data,
            type: 'base64',
            mimeType: mimeType
        })
    }
    this._postFormData("createOrUpdateImageLink", fields, params, callback);
};

// if permalink exists return the connected links
// {
//     "image": "https://bas.ngrok.io/p/gntnmnLZUQA4ubtQfnzCTh.png",
//     "edit": "https://balsamiq-ngrok.atlassian.net/browse/TEST-23"
// }
BalsamiqArchiveClient.prototype.checkPermalink = function (resourceID, branchID, callback) {
    var postObj = {
        resourceID: resourceID,
        branchID: branchID,
        // check if the permalink already exists
        check: true
    };
    this._post("permalink", postObj, callback);
};

// TODO DEPRECATED: use setPermalinkData instead
BalsamiqArchiveClient.prototype.setPermalink = function (resourceID, branchID, permalinkInfo, callback) {
    this.setPermalinkData(resourceID, branchID, permalinkInfo, callback);
};

// TODO: DEPRECATED: use checkPermalink instead
BalsamiqArchiveClient.prototype.getPermalink = function (resourceID, branchID, callback) {
    this.checkPermalink(resourceID, branchID, callback);
};

BalsamiqArchiveClient.prototype.getUserAuthToken = function(callback)
{
    // return {
    //     userAuthToken: userAuthToken
    // }
    this._post("getUserAuthToken", null, callback);
};

BalsamiqArchiveClient.prototype._get = function(apiURL, params, callback)
{
    this.networkApi('GET', this.serverURL+"/"+apiURL+"?token="+this.token+(params ? "&"+params : ""), null, function (result) {
        this._saveArchiveRevisionAndReturn(result, callback);
    }.bind(this), this.timeouts);
};

BalsamiqArchiveClient.prototype._getBinary = function(apiURL, params, responseType, callback, progressCallback, requestCreatedCallback)
{
    this.networkApi(
        'GET',
        this.serverURL + "/" + apiURL + "?token=" + this.token + (params ? "&" + params : ""),
        null,
        function (result) {
            if (!result.error) {
                callback(result.data);
            } else {
                callback({ error: "error retrieving the binary: " + result.error, aborted: result.aborted });
            }
        }.bind(this),
        this.timeouts,
        {
            binary : true,
            base64 : responseType == Consts.BinaryDataType.BASE64,
            progressCallback : progressCallback
        },
        function (request) {
            if (typeof(request.abort) !== 'function') {
                throw new Error("Error in getBinary: networkApi returned a request with no abort() function");
            }
            if (requestCreatedCallback) {
                requestCreatedCallback(request);
            }
        });
};

BalsamiqArchiveClient.prototype._getSync = function(apiURL, params, callback)
{
    var url = this.serverURL+"/"+apiURL+"?token="+this.token+(params ? "&"+params : "");
    this.networkApi('GET_SYNC', url, null, function (result) {
        if (result.error) {
            this._saveArchiveRevisionAndReturn(result, callback);
        } else {
            this._saveArchiveRevisionAndReturn({}, callback);
        }
    }.bind(this), this.timeouts);
};

BalsamiqArchiveClient.prototype._post = function(apiName, postObj, callback)
{
    this.networkApi('POST', this.serverURL+"/"+apiName+"?token="+this.token, postObj, function(obj) {
        this._saveArchiveRevisionAndReturn(obj, callback);
    }.bind(this), this.timeouts);
};

BalsamiqArchiveClient.prototype._postFormData = function(apiName, postObj, params, callback)
{
    this.networkApi('POST', this.serverURL+"/"+apiName+"?token="+this.token+(params ? "&"+params : ""), postObj, function(obj) {
        this._saveArchiveRevisionAndReturn(obj, callback);
    }.bind(this), this.timeouts, {formData: true});
};

BalsamiqArchiveClient.prototype._saveArchiveRevisionAndReturn = function(obj, callback)
{
    if (obj.error) {
        callback(obj);
    } else {
        // update the revision if necessary
        if (obj.archiveRevision)
        {
            this.archiveRevision = obj.archiveRevision;
        }
        if (callback) {
            callback(obj);
        }
    }
};


exports.BalsamiqArchiveClient = BalsamiqArchiveClient;


/***/ }),

/***/ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js":
/*!*********************************************************************!*\
  !*** ./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js ***!
  \*********************************************************************/
/***/ ((__unused_webpack_module, exports) => {

//
exports.Branch = {
    MasterBranchID: 'Master',
    AllBranches: 'All',
    NoBranch: 'None'
};

// Keys in INFO
exports.Info = {
    SchemaVersion: 'SchemaVersion', // integer
    ArchiveFormat: 'ArchiveFormat', // string "bmpr"
    ArchiveRevision: 'ArchiveRevision', // integer
    ArchiveAttributes: 'ArchiveAttributes', // json string
    ArchiveRevisionUUID: 'ArchiveRevisionUUID'
};

exports.ErrorCodes = {
    ArchiveNotOpen: 'ArchiveNotOpen',
    UserUnknown: 'UserUnknown',
    OutOfDate: 'OutOfDate',
    CreateFailed: 'CreateFail',
    SchemaVersionNotSupported: 'SchemaVersionNotSupported',
    NeedToMigrate: 'NeedToMigrate',
    ArchiveNotLoaded: 'ArchiveNotLoaded',
    MaintenanceMode: 'MaintenanceMode'
};

exports.Role = {
    ROLE_NO_ACCESS: 0,
    ROLE_VIEWER: 1,
    ROLE_COMMENTER: 2,
    ROLE_EDITOR: 3,
    ROLE_ADMIN: 4
};

exports.LikeStatus = {
    like: "like",
    unlike: "unlike"
};

exports.defaultUserInfo = {
    name: "default",
    avatarURL: ""
};

exports.CommentAttributes = {
    ParentID: 'parentID',               // String UUID
    ReadBy: 'readBy',                   // Array of userIDs
    Timestamp: 'timestamp',             // Integer UNIX Time in ms
    Trashed: 'trashed',                 // Boolean
    TrashedBy: "trashedBy",             // userID
    LikedBy: "likedBy",                 // Array of userIDs
    Timestamps: 'timestamps',           // Object in the form { label: timestamp }
    Imported: "imported",               // Boolean
    OriginalAuthor: "originalAuthor"    // string
};

exports.CommentAttributesSetters = {
    setReadStatus: "setReadStatus",                       // Boolean
    setLikedStatus: "setLikedStatus",                     // Boolean
    setTrashedStatus: "setTrashedStatus",                 // Boolean
    setTimestamps: "setTimestamps",                       // array of strings
    deleteTimestamps: "deleteTimestamps",                 // DEPRECATED array of strings
    setParentID: "setParentID",                           // DEPRECATED String
    setTimestamp: "setTimestamp",                         // DEPRECATED Integer UNIX Time in ms
    setImported: "imported",                              // DEPRECATED Boolean
    setOriginalAuthor: "originalAuthor"                   // DEPRECATED string
};

exports.CommentData = {
    Text: "text",
    Callouts: "callouts"
};

// used when importing comments
exports.IMPORTED_USER_ID = "IMPORTED_USER_ID";

exports.BinaryDataType = {
    ARRAYBUFFER: "arraybuffer",
    BASE64: "base64"
};

exports.SchemaVersions = {
    OneDotTwo: "1.2",
    Two: "2.0"
};

exports.CurrentSchemaVersion = exports.SchemaVersions.Two;


/***/ }),

/***/ "./node_modules/@balsamiq/bmpr/lib/BmprConstants.js":
/*!**********************************************************!*\
  !*** ./node_modules/@balsamiq/bmpr/lib/BmprConstants.js ***!
  \**********************************************************/
/***/ ((__unused_webpack_module, exports) => {

// Archive format
exports.ArchiveFormatBMPR = "bmpr";

// Archive attributes
exports.ArchiveAttributes = {
    Name: 'name', // name of the project
    Theme: 'theme',
    CreationDate: 'creationDate', // integer (epoch)
    Users: 'users', // array of User { id:String, name:String, email:String, ... }
    ModifiedBy: 'modifiedBy', // array of UserIDs
    AnonymousAccessLevel: 'anonymousAccessLevel', // String ("private", "site", "blog", "wiki")
    Slug: 'slug' // String
};

// Branch attributes
exports.BranchAttributes = {
    ProjectDescription: 'projectDescription', // description of the project (here because it can be modified in a branch)
    SkinName: 'skinName', // String
    FontFace: 'fontFace', // String (include here "Use System Font")
    FontSize: "fontSize", // String
    SelectionColor: "selectionColor",
    LinkColor: "linkColor",
    Name: "branchName"
};

// Resource attributes
exports.ResourceAttributes = {
    Kind: 'kind', // String of ResourceKinds
    Name: 'name', // String
    MimeType: 'mimeType', // String, DEPRECATED in Bmpr 2.0
    CreationDate: 'creationDate', // integer (epoch)
    ModifiedBy: 'modifiedBy', // array of UserIDs
    Trashed: 'trashed', // Boolean, necessary in order to implement untrash / undo delete functionality
    //Thumbnail: 'thumbnail' // String, base64 PNG, DEPRECATED
    ThumbnailID: 'thumbnailID', // String, ID of thumbnail, DEPRECATED in Bmpr 2.0
    Notes: 'notes', // String
    ImportedFrom: 'importedFrom' // String, URL of imported file (if any)
};

// Possible values for Kind ResourceAttribute
exports.ResourceKinds = {
    Mockup: "mockup",
    SymbolLibrary: "symbolLibrary",
    Asset: "asset",
    OtherAsset: "otherAsset"        //DEPRECATED in Bmpr 2.0
};

// Other attributes for Mockup resources
exports.MockupAttributes = {
    Order: 'order', // Number
    ParentID: 'parentID' // String, UUID of parent Resource
};

// Other attributes for Asset resources
exports.AssetAttributes = {
    Extension: 'extension'
};

// MimeTypes
// TODO: remove? unused
exports.MimeTypes = {
    Bmml: "text/vnd.balsamiq.bmml",
    Bmpr: "application/vnd.balsamiq.bmpr"
};

// MockupData
exports.MockupData = {
    Bounds: 'bounds', // Rectangle {x, y, width, height}
    Controls: 'controls', // array of ControlData
    Groups: 'groups', // array of GroupData (then unsorted)
    Notes: 'notes' // String
};

exports.GroupData = {
    ID: 'id', // ROOT for the root group, or Control ID (unique ID) of an existing Group Control.
    Controls: 'controls' // Array of Control IDs or sorted by zOrder
};

// Control attributes
exports.ControlData = {
    ID: 'id',   // String, UUID that never changes in the lifetime of this control
    Kind: 'kind',   // the control type (Button, DataGrid...)
    Visible: 'visible', // Boolean
    Trashed: 'trashed', // Boolean
    Locked: 'locked', // Boolean
    Markup: 'markup', // Boolean
    Position: 'position', // { x:int, y:int }
    ExplicitSize: 'explicitSize', // not there unless a user sets it explicitly { width:int, height:int } - only one dimension can be set
    Bounds: 'bounds', // Rectangle {x, y, width, height} - calculated based on Position and natural size or ExplicitSize

    // if the control supports cropping
    CropBounds: 'cropBounds', // {left:0..1000, top:0..1000, right:0..1000, bottom:0..1000}

    // if the control supports linking
    Links: 'links', //Array of LinkData

    //...more properties depending on control type

    // if Kind === Symbol:
    Overrides: 'overrides' //Object(controlID, Object(propName:String, propValue:*))
};

exports.LinkData = {
    Bounds: 'bounds', //Rectangle
    URL: 'url' //URLEncoded String - mockup://mockupID[#controlID] or http://foo or mailto://bar@ciao or BACK
};

exports.UserAttributes = {
    DisplayName: "displayName",
    UserName: "userName",
    Email: "email",
    Anonymous: "anonymous",
    InProject: "inProject"
};

exports.UserEventTypes = {
    AddComment: 'addComment',
    EditComment: 'editComment',
    LikeComment: 'likeComment',
    EditProjectDescription: 'editProjectDescription',
    EditResourceDescription: 'editResourceDescription',
    ChangeResource: 'changeResource',
    CreateResource: 'createResource',
    TrashResource: 'trashResource',
};


/***/ }),

/***/ "./js/core/BalsamiqArchiveInMemory.js":
/*!********************************************!*\
  !*** ./js/core/BalsamiqArchiveInMemory.js ***!
  \********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.BalsamiqArchiveInMemory = BalsamiqArchiveInMemory;
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var Constants = __webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchiveConstants */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js");

function BalsamiqArchiveInMemory(context, dump) {
  _balsamiq.default.exists(context.host);

  this.callAsyncFun = context.host.asyncCallback;
  this.userInfo = dump.userInfo ? dump.userInfo : Constants.defaultUserInfo;
  this.data = dump;
}

;

BalsamiqArchiveInMemory.prototype.replaceDump = function (dump, callback) {
  this.data = dump;
  this.result({}, callback);
};

BalsamiqArchiveInMemory.prototype.open = function (callback) {
  this.result(this.data, callback);
};

BalsamiqArchiveInMemory.prototype.close = function (callback) {
  this.result({}, callback);
};

BalsamiqArchiveInMemory.prototype.getArchiveRevision = function (callback) {
  this.result({}, callback);
};

BalsamiqArchiveInMemory.prototype.getArchiveFormat = function (callback) {
  this.result({
    format: this.data.Info.ArchiveFormat
  }, callback);
};

BalsamiqArchiveInMemory.prototype.getArchiveAttributes = function (callback) {
  this.result({
    attributes: this.data.Info.ArchiveAttributes
  }, callback);
};

BalsamiqArchiveInMemory.prototype.getBranches = function (callback) {
  var branches = [];
  this.data.Branch.forEach(function (branch) {
    branches.push({
      ID: branch.branchID,
      attributes: branch.ATTRIBUTES
    });
  }, this);
  this.result({
    branches: branches
  }, callback);
};

BalsamiqArchiveInMemory.prototype.getBranchAttributes = function (branchID, callback) {
  var branch = this._getBranch(branchID);

  this.result({
    attributes: branch.ATTRIBUTES
  }, callback);
};

BalsamiqArchiveInMemory.prototype.getResourcesOfBranch = function (branchID, callback) {
  var resources = [];
  this.data.Resources.forEach(function (resource) {
    if (!resource.BRANCHID || resource.BRANCHID === branchID) {
      resources.push({
        ID: resource.ID,
        attributes: resource.ATTRIBUTES
      });
    }
  }, this);
  this.result({
    resources: resources
  }, callback);
};

BalsamiqArchiveInMemory.prototype.getResourceAttributes = function (branchID, resourceID, callback) {
  var resource = this._getResource(branchID, resourceID);

  this.result({
    attributes: resource.ATTRIBUTES
  }, callback);
};

BalsamiqArchiveInMemory.prototype.getResourceData = function (branchID, resourceID, callback) {
  var resource = this._getResource(branchID, resourceID);

  this.result({
    data: resource.DATA
  }, callback);
};

BalsamiqArchiveInMemory.prototype.setResourceData = function (resourceID, branchID, data, callback) {
  this._setResourceData(branchID, resourceID, data);

  this.data.Info.ArchiveRevision += 1;
  this.result({
    archiveRevision: this.data.Info.ArchiveRevision
  }, callback);
};

BalsamiqArchiveInMemory.prototype.createResource = function (resourceID, branchID, attributes, data, callback) {
  this.data.Resources.push({
    ATTRIBUTES: attributes,
    BRANCHID: branchID,
    DATA: data,
    ID: resourceID
  });
  this.data.Info.ArchiveRevision += 1;
  this.result({
    archiveRevision: this.data.Info.ArchiveRevision
  }, callback);
};

BalsamiqArchiveInMemory.prototype.setResourceAttributes = function (resourceID, branchID, attributes, callback) {
  for (var i = 0; i < this.data.Resources.length; i++) {
    var resource = this.data.Resources[i];

    if (resource.BRANCHID === branchID && resource.ID === resourceID) {
      resource.ATTRIBUTES = attributes;
      break;
    }
  }

  this.data.Info.ArchiveRevision += 1;
  this.result({
    archiveRevision: this.data.Info.ArchiveRevision
  }, callback);
};

BalsamiqArchiveInMemory.prototype.setThumbnail = function (thumbnailID, attributes, callback) {
  for (let i = 0; i < this.data.Thumbnails.length; i++) {
    const thumbnail = this.data.Thumbnails[i];

    if (thumbnail.ID === thumbnailID) {
      thumbnail.ATTRIBUTES = attributes;
      break;
    }
  }

  this.data.Info.ArchiveRevision += 1;
  this.result({
    archiveRevision: this.data.Info.ArchiveRevision
  }, callback);
};

BalsamiqArchiveInMemory.prototype.dump = function (_branches, callback) {
  this.result({
    dump: this.data
  }, callback);
};

BalsamiqArchiveInMemory.prototype.getResourcesData = function (resourceIDs, branchID, callback) {
  var data = {};
  resourceIDs.forEach(resourceID => {
    const resource = this._getResource(branchID, resourceID);

    data[`${resource.ID}|${resource.BRANCHID}`] = resource.DATA;
  });
  this.result({
    data
  }, callback);
};

BalsamiqArchiveInMemory.prototype.getThumbnail = function (thumbnailID, callback) {
  for (var i = 0; i < this.data.Thumbnails.length; i++) {
    if (this.data.Thumbnails[i].ID === thumbnailID) {
      this.result(this.data.Thumbnails[i], callback);
      break;
    }
  }
};

BalsamiqArchiveInMemory.prototype._getBranch = function (branchID) {
  for (var i = 0; i < this.data.Branch.length; i++) {
    var branch = this.data.Branches[i];

    if (branch.ID === branchID) {
      return branch;
    }
  }

  _balsamiq.default.assert(false, "BalsamiqArchiveInMemory: branchID not found " + branchID);

  return null;
};

BalsamiqArchiveInMemory.prototype._getResource = function (branchID, resourceID) {
  for (var i = 0; i < this.data.Resources.length; i++) {
    var resource = this.data.Resources[i];

    if ((branchID === "All" || resource.BRANCHID === branchID) && resource.ID === resourceID) {
      return resource;
    }
  }

  _balsamiq.default.assert(false, "BalsamiqArchiveInMemory: resourceID not found " + resourceID + " in branchID " + branchID);

  return null;
};

BalsamiqArchiveInMemory.prototype._setResourceData = function (branchID, resourceID, data) {
  for (var i = 0; i < this.data.Resources.length; i++) {
    var resource = this.data.Resources[i];

    if (resource.BRANCHID === branchID && resource.ID === resourceID) {
      resource.DATA = data;
      return;
    }
  }

  _balsamiq.default.assert(false, "BalsamiqArchiveInMemory._setResourceData: resourceID not found " + resourceID + " in branchID " + branchID);

  return null;
};

BalsamiqArchiveInMemory.prototype.result = function (obj, callback) {
  var res = _balsamiq.default.extend({
    archiveRevision: this.data.Info.ArchiveRevision
  }, obj);

  this.callAsyncFun(callback, [res]);
};

BalsamiqArchiveInMemory.prototype.getUserID = function () {
  this.userInfo.ID;
};

BalsamiqArchiveInMemory.prototype.isLocal = function () {
  return false;
};

var _default = BalsamiqArchiveInMemory;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/archive.js":
/*!****************************!*\
  !*** ./js/core/archive.js ***!
  \****************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.EventTypes = exports.Archive = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var EventTypes = {
  ARCHIVE_ATTRIBUTES_CHANGED: "archiveAttributesChanged",
  ARCHIVE_ONLINE: "ArchiveOnline",
  ARCHIVE_OFFLINE: "ArchiveOffline",
  BRANCH_ADDED: "branchAdded",
  BRANCH_ATTRIBUTES_CHANGED: "branchAttributesChanged",
  BRANCHES_DELETED: "branchesDeleted",
  RESOURCE_ADDED: "resourceAdded",
  RESOURCE_DATA_CHANGED: "resourceDataChanged",
  RESOURCE_ATTRIBUTES_CHANGED: "resourceAttributesChanged",
  RESOURCES_DELETED: "resourcesDeleted",
  RESOURCE_BRANCH_CHANGED: "resourceBranchChanged",
  THUMBNAIL_ADDED: "thumbnailAdded",
  THUMBNAILS_DELETED: "thumbnailsDeleted",
  THUMBNAIL_UPDATED: "thumbnailUpdated",
  ONLINE_USERS_LIST_CHANGED: "onlineUsersListChanged",
  COMMENT_CREATED: "commentCreated",
  COMMENTS_ATTRIBUTES_UPDATED: "updateCommentsAttributes",
  COMMENT_DATA_UPDATED: "commentDataUpdated",
  COMMENTS_DELETED: "deleteComments",
  ARCHIVE_USERS_CHANGED: "archiveUsersChanged",
  SERVER_TIMEDIFF: "serverTimeDiff",
  ARCHIVE_RESTORING: "aboutToRestoreArchive",
  ARCHIVE_RESTORED: "archiveRestored",
  ARCHIVE_FLUSHED: "archiveHasBeenFlushedOnPlatform",
  BROADCAST_MESSAGE: "broadcastMessage",
  PERMALINK_CREATED_OR_UPDATED: "permalinkCreatedOrUpdated",
  PERMALINK_DELETED: "permalinkDeleted",
  USER_ROLE_CHANGED: "userRoleChanged"
};
exports.EventTypes = EventTypes;

var Archive = function () {};

exports.Archive = Archive;
var methods = {
  open: function () {},
  getArchiveRevision: function () {},
  //    getArchiveAttributes: function () {},
  //    setArchiveAttributes: function () {},
  getArchiveFormat: function () {},
  //    setArchiveFormat: function () {},
  getBranches: function () {},
  getBranchAttributes: function () {},
  //    setBranchAttributes: function () {},
  //    createResource: function () {},
  //    deleteResources: function () {},
  getResourceAttributes: function () {},
  //    setResourceAttributes: function () {},
  getResourcesData: function () {},
  //    setResourceData: function () {},
  //    getThumbnails: function () {}
  //    createThumbnail: function () {},
  getThumbnail: function () {} //    setThumbnail: function () {},
  //    deleteThumbnails: function () {},
  //    deleteBranches: function () {},
  //    createBranch: function () {}

};

_balsamiq.default.extend(Archive.prototype, methods);

/***/ }),

/***/ "./js/core/balsamiq.js":
/*!*****************************!*\
  !*** ./js/core/balsamiq.js ***!
  \*****************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const ACTIVE = true;

const VOID_FN = () => {};

function callIfActive(func) {
  return ACTIVE ? func : VOID_FN;
}

function assert(p_condition, p_message) {
  if (!p_condition) {
    let m = "Assertion failed: " + p_message; // it's on its own line to make it possible to set a breakpoint!

    throw new Error(m);
  }
}

function assertClass(p_object, p_constructor) {
  if (p_object) {
    assert(p_object instanceof p_constructor, "this object must be instance of class " + p_constructor.name + " instead was " + p_object.constructor.name);
  }
}

function conformsTo(p_objectToCheck, p_referenceObject) {
  var referenceKeys = Object.keys(p_referenceObject);
  return _lodash.default.every(referenceKeys, function (key) {
    var t1 = typeof p_referenceObject[key];
    var t2 = typeof p_objectToCheck[key];
    return t1 === t2;
  });
}

function assertProtocol(p_objectToCheck, p_referenceObject) {
  assert(conformsTo(p_objectToCheck, p_referenceObject), "this object must implement a protocol");
}

function exists(p_value, p_mess) {
  p_mess = isUndefined(p_mess) ? "" : ": " + p_mess;
  assert(!isNull(p_value), "this parameter can not be null" + p_mess);
  assert(!isUndefined(p_value), "this parameter can not be undefined" + p_mess);
  assert(!Number.isNaN(p_value), "this parameter can not be NaN" + p_mess);
}

function existsClass(p_object, p_constructor, p_message) {
  exists(p_object, p_message);
  assertClass(p_object, p_constructor);
}

function existMulti() {
  let [p_mess, ...values] = arguments;
  p_mess = isUndefined(p_mess) ? "" : ": " + p_mess;

  for (let p_value of values) {
    assert(!isNull(p_value), "this parameter can not be null" + p_mess);
    assert(!isUndefined(p_value), "this parameter can not be undefined" + p_mess);
    assert(!Number.isNaN(p_value), "this parameter can not be NaN" + p_mess);
  }
}

function isNull(p_value) {
  return p_value === null;
}

function isUndefined(p_value) {
  return p_value === undefined;
}

function isFunction(p_value) {
  return typeof p_value === 'function';
}

function isString(p_value) {
  return typeof p_value === 'string' || p_value instanceof String;
}

function isNumber(p_value) {
  return typeof p_value === 'number';
}

function isBoolean(p_value) {
  return typeof p_value === 'boolean';
}

function isDate(p_value) {
  return _lodash.default.isDate(p_value); // it should be: p_value instanceof Date; but I can't test it in diff.js
} // function isNan(p_value) {
//     return isNumber(p_value) && Number.isNaN(p_value);
// }


function isObject(p_value) {
  return typeof p_value === 'object' && p_value !== null;
}

function isArray(p_value) {
  return Array.isArray(p_value);
}

function assertFunction(p_value) {
  assert(isFunction(p_value), "this parameter must be a function");
}

function assertString(p_value, p_message) {
  assert(isString(p_value), isUndefined(p_message) ? "this parameter must be a string" : p_message);
}

function assertBool(p_value) {
  assert(isBoolean(p_value), "this parameter must be a boolean");
}

function assertNumber(p_value) {
  assert(isNumber(p_value), `received "${p_value}", expected a number`);
  assert(!Number.isNaN(p_value), "this parameter must be a valid number");
}

function
  /*p_value*/
assertPositiveNumber() {
  _lodash.default.forEach(arguments, value => {
    assertNumber(value);
    assert(value > 0, "this parameter must be a number > 0, received: " + value.toString());
  });
}

function assertSizeNumber(p_value) {
  assertNumber(p_value);
  assert(p_value === -1 || p_value >= 0, "this parameter must be a positive number or -1");
}

function assertNumber01(p_value) {
  assertNumber(p_value);
  assert(p_value >= 0 && p_value <= 1, "this parameter must be a number between 0 and 1");
}

function assertArray(p_value, p_mess) {
  assert(isArray(p_value), `this parameter must be an Array, got value ${p_value} of type ${typeof p_value}` || p_mess);
}

function assertArrayOfStrings(p_value) {
  assert(isArray(p_value), "this parameter must be an Array");

  for (let element of p_value) {
    assertString(element, "this object must be instance a String");
  }
}

function assertArrayOf(p_value, type) {
  assert(isArray(p_value), "this parameter must be an Array");

  for (let element of p_value) {
    assert(typeof element === type, "this object must be of type: " + type);
  }
}

function assertArrayOfClasses(p_value, p_constructor) {
  assert(isArray(p_value), "this parameter must be an Array");

  for (let element of p_value) {
    assert(element instanceof p_constructor, "this object must be instance of class " + p_constructor.name);
  }
}

function assertObject(p_value) {
  assert(isObject(p_value), `Received ${p_value}, expected object`);
}

function assertObjectOrNull(p_value) {
  assert(p_value === null || isObject(p_value), `Received ${p_value}, expected Object or null`);
}

function checkEnum(p_value, p_possibleValues) {
  if (isObject(p_possibleValues)) {
    p_possibleValues = Object.values(p_possibleValues);
  }

  return p_possibleValues.indexOf(p_value) !== -1;
}

function has(p_object, p_path) {
  return _lodash.default.has(p_object, p_path);
}

function isEmpty(p_value) {
  return _lodash.default.isEmpty(p_value);
}

function isEqual(p_value1, p_value2) {
  return _lodash.default.isEqual(p_value1, p_value2);
}

function clone(p_value) {
  return _lodash.default.clone(p_value);
}

function cloneDeep(p_value) {
  return _lodash.default.cloneDeep(p_value);
}

function extend(p_object, p_source) {
  return _lodash.default.extend(p_object, p_source);
}

function get(p_object, p_path, p_defaultValue) {
  return _lodash.default.get(p_object, p_path, p_defaultValue);
}

function set(p_object, p_path, p_value) {
  return _lodash.default.set(p_object, p_path, p_value);
}

function objectForEach(p_values, p_function) {
  return _lodash.default.forEach(p_values, p_function);
}

function objectReduce(p_values, p_function, p_initialObject) {
  return _lodash.default.reduce(p_values, p_function, p_initialObject);
}

function filter(p_collection, p_predicate) {
  return _lodash.default.filter(p_collection, p_predicate);
}

function uniq(p_array) {
  return _lodash.default.uniq(p_array);
}

function uniqBy(p_array, p_predicate) {
  return _lodash.default.uniqBy(p_array, p_predicate);
}

function sortBy(p_collection, p_predicate) {
  return _lodash.default.sortBy(p_collection, p_predicate);
}

function maxBy(p_collection, p_predicate) {
  return _lodash.default.maxBy(p_collection, p_predicate);
}

function countBy(p_collection, p_predicate) {
  return _lodash.default.countBy(p_collection, p_predicate);
}

function chunk(p_array, p_size) {
  return _lodash.default.chunk(p_array, p_size);
}

function assertEnum(p_value, p_possibleValues) {
  if (checkEnum(p_value, p_possibleValues)) {
    return true;
  } else {
    return assert(false, `value "${p_value}" is not a valid ENUM: ${JSON.stringify(p_possibleValues, null, 2)}`);
  }
}

function assertContainsFunctionDeep(p_value, p_message) {
  assert(typeof p_value !== "function", p_message);

  if (Array.isArray(p_value)) {
    p_value.forEach(arrObj => {
      assertContainsFunctionDeep(arrObj, p_message);
    });
  } else if (typeof p_value === "object" && p_value !== null) {
    Object.keys(p_value).forEach(key => {
      assertContainsFunctionDeep(p_value[key], p_message);
    });
  }
}

function assertRect(p_rect, p_message) {
  existMulti(p_message ? p_message : "Rect must have width, height, x, and y", p_rect.x, p_rect.y, p_rect.width, p_rect.height);
}

function assertPoint(p_point, p_message) {
  exists(p_point, "Point must exist");
  existMulti(p_message ? p_message : "Point must and have x and y", p_point.x, p_point.y);
}

function assertSize(p_size, p_message) {
  exists(p_size, "Size must exist");
  existMulti(p_message ? p_message : "Size must have width and height", p_size.width, p_size.height);
}

function assertMaskRect(p_maskRect, p_message) {
  assert(p_maskRect.type === 'rect', p_message);
  existMulti(p_message, p_maskRect.rect.x, p_maskRect.rect.y, p_maskRect.rect.width, p_maskRect.rect.height);
}

const UUIDRegEx = /^[A-Z0-9]{8}-[A-Z0-9]{4}-[A-Z0-9]{4}-[A-Z0-9]{4}-[A-Z0-9]{12}/;

function assertUUID(p_value, p_message) {
  assert(UUIDRegEx.test(p_value), p_message);
}

function assertBranch(p_value, p_message) {
  assert(p_value === 'Master' || UUIDRegEx.test(p_value), p_message);
}

function switchEnum(val, en, cases) {
  assert(isObject(en), `SwitchEnum enum param should be an Object!`);
  assertEnum(val, en);
  assert(isObject(cases), `SwitchEnum cases param should be an Object! (containing funcs)`);

  for (let key in en) {
    if (!isFunction(cases[en[key]])) {
      assert(false, `SwitchEnum case missing or not a Function: ${key} Switching on Enum ${JSON.stringify(en, null, 2)}`);
    }
  }

  assert(Object.keys(en).length == Object.keys(cases).length, `SwitchEnum cases must be same length as enum! We have too many cases.`);
  cases[val]();
} // Custom polyfills
// MDN https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/startsWith


if (!String.prototype.startsWith) {
  String.prototype.startsWith = function (searchString, position) {
    position = position || 0;
    return this.indexOf(searchString, position) === position;
  };
} // MDN https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/find


if (!Array.prototype.find) {
  Object.defineProperty(Array.prototype, 'find', {
    enumerable: false,
    configurable: true,
    writable: true,
    value: function (predicate) {
      if (this == null) {
        throw new TypeError('Array.prototype.find called on null or undefined');
      }

      if (typeof predicate !== 'function') {
        throw new TypeError('predicate must be a function');
      }

      var list = Object(this);
      var length = list.length >>> 0;
      var thisArg = arguments[1];
      var value;

      for (var i = 0; i < length; i++) {
        if (i in list) {
          value = list[i];

          if (predicate.call(thisArg, value, i, list)) {
            return value;
          }
        }
      }

      return undefined;
    }
  });
}

if (!Array.prototype.pushOrRemove) {
  Object.defineProperty(Array.prototype, 'pushOrRemove', {
    enumerable: false,
    configurable: true,
    writable: true,
    value: function (areEqualFun, elem) {
      if (this == null) {
        throw new TypeError('Array.prototype.addOrRemove called on null or undefined');
      }

      if (typeof areEqualFun !== 'function') {
        throw new TypeError('areEqualFun must be a function');
      }

      var list = Object(this);
      var length = list.length >>> 0;
      var value;

      for (var i = 0; i < length; i++) {
        if (i in list) {
          value = list[i];

          if (areEqualFun.call(null, value, elem)) {
            this.splice(i, 1);
            return undefined;
          }
        }
      }

      this.push(elem);
      return undefined;
    }
  });
}

const optional = fn => function (value) {
  for (var _len = arguments.length, rest = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    rest[_key - 1] = arguments[_key];
  }

  if (!isUndefined(value)) fn(value, ...rest);
};

var _default = {
  assert: callIfActive(assert),
  exists: callIfActive(exists),
  existMulti: callIfActive(existMulti),
  conformsTo,
  checkEnum,
  has,
  isEmpty,
  isEqual,
  clone,
  cloneDeep,
  extend,
  objectForEach,
  objectReduce,
  filter,
  uniq,
  uniqBy,
  get,
  set,
  sortBy,
  maxBy,
  countBy,
  chunk,
  assertBool: callIfActive(assertBool),
  assertOptionalBool: callIfActive(optional(assertBool)),
  assertNumber: callIfActive(assertNumber),
  assertOptionalNumber: callIfActive(optional(assertNumber)),
  assertPositiveNumber: callIfActive(assertPositiveNumber),
  assertOptionalPositiveNumber: callIfActive(optional(assertPositiveNumber)),
  assertSizeNumber: callIfActive(assertSizeNumber),
  assertNumber01: callIfActive(assertNumber01),
  assertEnum: callIfActive(assertEnum),
  assertOptionalEnum: callIfActive(optional(assertEnum)),
  assertString: callIfActive(assertString),
  assertOptionalString: callIfActive(optional(assertString)),
  assertArray: callIfActive(assertArray),
  assertOptionalArray: callIfActive(optional(assertArray)),
  assertFunction: callIfActive(assertFunction),
  assertOptionalFunction: callIfActive(optional(assertFunction)),
  assertClass: callIfActive(assertClass),
  assertOptionalClass: callIfActive(optional(assertClass)),
  existsClass: callIfActive(existsClass),
  assertUUID: callIfActive(assertUUID),
  assertBranch: callIfActive(assertBranch),
  assertProtocol: callIfActive(assertProtocol),
  assertContainsFunctionDeep: callIfActive(assertContainsFunctionDeep),
  assertObject: callIfActive(assertObject),
  assertObjectOrNull: callIfActive(assertObjectOrNull),
  assertOptionalObject: callIfActive(optional(assertObject)),
  assertArrayOfClasses: callIfActive(assertArrayOfClasses),
  assertArrayOfStrings: callIfActive(assertArrayOfStrings),
  assertArrayOf: callIfActive(assertArrayOf),
  assertRect: callIfActive(assertRect),
  assertMaskRect: callIfActive(assertMaskRect),
  assertPoint: callIfActive(assertPoint),
  assertSize: callIfActive(assertSize),
  switchEnum,
  isNull,
  isUndefined,
  isString,
  isNumber,
  isObject,
  isArray,
  isFunction,
  isBoolean,
  isDate,
  UUIDRegEx
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/bricks/breakline.js":
/*!*************************************!*\
  !*** ./js/core/bricks/breakline.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.addBreaklineMask = addBreaklineMask;
exports.makeBreaklineDisplayitem = makeBreaklineDisplayitem;
exports.makeFilledBreaklineDisplayitem = makeFilledBreaklineDisplayitem;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function addBreaklineMask(p_control, p_x, p_y, p_position) {
  _balsamiq.default.assert(p_control);

  _balsamiq.default.assertNumber(p_x);

  _balsamiq.default.assertNumber(p_y);

  if (_balsamiq.default.isUndefined(p_position)) {
    p_position = _types.default.Positions.BOTTOM;
  } else {
    _balsamiq.default.assertEnum(p_position, _types.default.Positions);
  }

  var controlSize = p_control.getActualProperty("size");
  var maskPosition = p_position === _types.default.Positions.BOTTOM ? Path.Breakline.CLOSED_BOTTOM : Path.Breakline.CLOSED_TOP;
  var maskId = p_control.addMask(1); // TODO

  var maskPath = Path.newBreaklinePath(controlSize.width, controlSize.height, maskPosition, 0);
  p_control.addPathToMask(maskId, maskPath, p_x, p_y - _types.default.BREAKLINE_VERTICAL_OFFSET); // TODO x, y = 0

  return maskId;
}

;

function makeBreaklineDisplayitem(p_x, p_y, p_width, p_height, p_thickness, p_color) {
  _balsamiq.default.assertNumber(p_x);

  _balsamiq.default.assertNumber(p_y);

  _balsamiq.default.assertPositiveNumber(p_width, p_thickness);

  if (_balsamiq.default.isUndefined(p_color)) {
    p_color = 0x000000;
  } else {
    _balsamiq.default.assertNumber(p_color);
  }

  var breaklineBbox = _types.default.newRectangle(p_x, p_y - _types.default.BREAKLINE_VERTICAL_OFFSET, p_width, _types.default.BREAKLINE_VERTICAL_OFFSET * 2);

  var path = Path.newBreaklinePath(p_width, p_height, Path.Breakline.OPEN, p_thickness);

  var displayItem = _displayitem.default.newSVGPath(breaklineBbox, path, p_thickness, p_color);

  return displayItem;
}

;

function makeFilledBreaklineDisplayitem(p_x, p_y, p_width, p_height, p_position, p_color) {
  _balsamiq.default.assertNumber(p_x);

  _balsamiq.default.assertNumber(p_y);

  if (_balsamiq.default.isUndefined(p_position)) {
    p_position = _types.default.Positions.BOTTOM;
  } else {
    _balsamiq.default.assertEnum(p_position, _types.default.Positions);
  }

  if (_balsamiq.default.isUndefined(p_color)) {
    p_color = 0xFFFFFF;
  } else {
    _balsamiq.default.assertNumber(p_color);
  }

  var position = p_position === _types.default.Positions.BOTTOM ? Path.Breakline.CLOSED_BOTTOM : Path.Breakline.CLOSED_TOP;

  var breaklineBbox = _types.default.newRectangle(p_x, p_y - _types.default.BREAKLINE_VERTICAL_OFFSET, p_width, _types.default.BREAKLINE_VERTICAL_OFFSET * 2);

  var path = Path.newBreaklinePath(p_width, p_height, position, 0);

  var displayItem = _displayitem.default.newSVGPath(breaklineBbox, path, 0, null, p_color);

  return displayItem;
}

;

/***/ }),

/***/ "./js/core/bricks/choice.js":
/*!**********************************!*\
  !*** ./js/core/bricks/choice.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ChoiceType = exports.Choice = void 0;
exports.parseChoiceState = parseChoiceState;
exports.parseChoiceText = parseChoiceText;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var _lodash = __webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

const ChoiceType = {
  CHECKBOX: "CHECKBOX",
  RADIO: "RADIO"
};
exports.ChoiceType = ChoiceType;
const DATA = {
  CHECKBOX: {
    PADDING_VERTICAL: 10,
    PADDING_LEFT: 20,
    ASSET_NAME: "CheckBox",
    SELECTED_ASSET_NAME: "CheckBoxSelected",
    INDETERMINATE_ASSET_NAME: "CheckBoxIndeterminate",
    DELIMITERS: ["[", "]"]
  },
  RADIO: {
    PADDING_VERTICAL: 10,
    PADDING_LEFT: 16,
    ASSET_NAME: "RadioButton",
    SELECTED_ASSET_NAME: "RadioButtonSelected",
    INDETERMINATE_ASSET_NAME: "RadioButtonIndeterminate",
    DELIMITERS: ["(", ")"]
  },
  ASSET_SIZE: 16
};

class Choice {
  constructor(_ref) {
    let {
      type,
      state,
      text,
      textStyle,
      color,
      icon,
      iconRotation,
      customLabelPadding,
      previousChoice,
      context
    } = _ref;

    _defineProperty(this, "getLinks", () => this.text.getLinks());

    _defineProperty(this, "measure", () => {
      let width = this.text.measuredSize().width;

      if (typeof this.customLabelPadding === "number") {
        width += this.customLabelPadding;
      } else {
        width += DATA[this.type].PADDING_LEFT + (this.iconData ? this.iconData.size + this.context.skin.choice_icon_padding : 0) + (this.type === ChoiceType.RADIO ? 4 : 0);
      }

      const height = Math.max(this.text.measuredSize().height + DATA[this.type].PADDING_VERTICAL, this.iconData ? this.iconData.size : 0);
      return _types.default.newSize(width, height);
    });

    _defineProperty(this, "makeDisplayItemsAndLinks", (p_control, p_localBbox) => {
      _balsamiq.default.exists(p_control);

      _balsamiq.default.exists(p_localBbox);

      const data = DATA[this.type];
      const radioPadding = this.type === ChoiceType.RADIO ? 4 : 0;
      let assetName = data.ASSET_NAME;

      if (_types.default.ControlState.isSelectedState(this.state)) {
        assetName = data.SELECTED_ASSET_NAME;
      } else if (_types.default.ControlState.isIndeterminateState(this.state)) {
        assetName = data.INDETERMINATE_ASSET_NAME;
      }

      const assetBounds = _types.default.newRectangle(p_localBbox.x, p_localBbox.y + Math.round((p_localBbox.height - DATA.ASSET_SIZE) / 2), DATA.ASSET_SIZE, DATA.ASSET_SIZE);

      let l = _displayitem.default.createDisplayItemsFromSkin(assetName, this.color, assetBounds, this.context);

      let textX = 0;

      if (this.customLabelPadding) {
        textX = this.customLabelPadding;
      } else {
        textX = data.PADDING_LEFT + (this.iconData ? this.iconData.size + this.context.skin.choice_icon_padding : 0) + radioPadding;
      }

      textX = Math.min(textX, p_localBbox.width);
      let textY = p_localBbox.y + (p_localBbox.height - this.text.measuredSize().height) / 2 + 1;
      const obj = this.text.makeDisplayItemsAndLinks(_types.default.newRectangle(textX, textY, p_localBbox.width - textX, p_localBbox.height));
      l.push(obj.displayItems);

      if (this.iconData) {
        l.push(_icon.default.makeDisplayItemsFromIconData(p_control, this.iconData, data.PADDING_LEFT + radioPadding, (p_localBbox.height - this.iconData.size) / 2, this.color));
      }

      return {
        displayItems: l,
        links: obj.links
      };
    });

    _defineProperty(this, "makeDisplayItems", (p_control, p_localBbox) => this.makeDisplayItemsAndLinks(p_control, p_localBbox).displayItems);

    // customLabelPadding: if defined, wins over icon and related padding
    this.type = type;
    this.state = state;

    const {
      iconData,
      dependencies
    } = _icon.default.processIconProperty(icon, iconRotation);

    this.iconData = iconData;
    this.dependencies = dependencies;
    this.color = _types.default.ControlState.isDisabledState(state) ? context.skin.default_disabled_color : color;
    this.textStyle = (0, _lodash.clone)(textStyle);
    this.textStyle.textColor = _types.default.ControlState.isDisabledState(state) ? context.skin.default_disabled_color : textStyle.textColor;
    this.textStyle.linkColor = _types.default.ControlState.isDisabledState(state) ? context.skin.default_disabled_color : textStyle.linkColor;
    this.text = previousChoice && previousChoice.text;

    if (!this.text || !(0, _text.areTextsEqual)(this.text, text, this.textStyle)) {
      this.text = (0, _text.newSingleLineText)(text, this.textStyle, context);
    }

    this.customLabelPadding = customLabelPadding;
    this.context = context;
  }

}

exports.Choice = Choice;
;

const makeRegExp = function (p_text, p_delimiters) {
  var d = ["\\" + p_delimiters[0], "\\" + p_delimiters[1]];
  return new RegExp(p_text.replace(/€1/g, d[0]).replace(/€2/g, d[1]));
};
/**
 * Return the state of the choice in the text (if any), given the choice delimiters to use
 * @param {!string} p_text
 * @param {!array} p_delimiters
 * @returns {{state:string, match:array}}
 */


function parseChoiceState(p_text, p_delimiters) {
  let choiceState = null;
  let disabled = false;
  let color = null;
  let colorPrefix = "";
  let colorTail = "";
  let origText = p_text;
  let tail = "";
  const colorRegExp = /^{colou?r:(.+?)}(.*){colou?r}/;
  const colorMatch = p_text.match(colorRegExp);

  if (colorMatch !== null) {
    const index = origText.indexOf(p_delimiters[0]);
    colorPrefix = origText.slice(0, index); //store it to move the color markup value into the text when needed

    colorTail = colorMatch[0].slice(colorMatch[0].lastIndexOf("{")); //remember the color markup tag for the text parameter

    tail = origText === colorMatch[0] ? "" : origText.slice(colorMatch[0].length); //in case there is text after a color markup store the value.

    p_text = colorMatch[2];
    color = colorMatch[1];
  } // €1 will be p_delimiters[0]
  // €2 will be p_delimiters[1]


  const normalState = "^€1(?: )?€2(.*)";
  const selectedState = "^€1[oOvVxX\\*]€2(.*)";
  const mixedState = "^€1[-=]€2(.*)";
  const normalStateRegExp = makeRegExp(normalState, p_delimiters);
  const selectedStateRegExp = makeRegExp(selectedState, p_delimiters);
  const mixedStateRegExp = makeRegExp(mixedState, p_delimiters);

  if (p_text.length >= 2 && p_text[0] === '-' && p_text[p_text.length - 1] === '-') {
    disabled = true;
    p_text = p_text.slice(1, p_text.length - 1);
  }

  let m = p_text.match(normalStateRegExp);

  if (m) {
    choiceState = disabled ? _types.default.ControlState.DISABLED : _types.default.ControlState.NORMAL;
  } else {
    m = p_text.match(selectedStateRegExp);

    if (m) {
      choiceState = disabled ? _types.default.ControlState.DISABLED_SELECTED : _types.default.ControlState.SELECTED;
    } else {
      m = p_text.match(mixedStateRegExp);

      if (m) {
        choiceState = disabled ? _types.default.ControlState.DISABLED_INDETERMINATE : _types.default.ControlState.INDETERMINATE;
      }
    }
  }

  let text = "";
  let textcolor = null;
  const colorClosingTag = "^{colou?r}(.*)";

  if (m) {
    if (typeof m[1] === "string" && m[1].length > 0) {
      const subtext = m[1].match(colorClosingTag);

      if (subtext) {
        m[1] = subtext[1];
      } else {
        text = colorPrefix;
      }

      text += m[1].trim() + colorTail + tail;
      textcolor = tail === "" ? color : null;
    } else {
      text = tail.trim();
    }
  }

  return choiceState ? {
    state: choiceState,
    text: text,
    color: color,
    textcolor: textcolor
  } : null;
}

;
/**
 * If the given text is a choice text, then return the type and the state of the choice
 * @param {!string} p_text
 * @returns {{type: string, state: string, match: array}}
 */

function parseChoiceText(p_text) {
  let choiceType = null;
  let choiceState = parseChoiceState(p_text, DATA.CHECKBOX.DELIMITERS);

  if (choiceState) {
    choiceType = ChoiceType.CHECKBOX;
  } else {
    choiceState = parseChoiceState(p_text, DATA.RADIO.DELIMITERS);

    if (choiceState) {
      choiceType = ChoiceType.RADIO;
    }
  }

  return choiceType ? Object.assign(choiceState, {
    type: choiceType
  }) : null;
}

;

/***/ }),

/***/ "./js/core/bricks/combobox.js":
/*!************************************!*\
  !*** ./js/core/bricks/combobox.js ***!
  \************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var object = {
  makeComboboxDisplayItems: function (p_control, p_Bbox, p_comboHeight, p_titleText, p_color, p_state, p_iconData, p_iconColor) {
    var paddingLeft = p_control.style("combobox_padding_left");

    var disabledState = _types.default.ControlState.isDisabledState(p_state);

    var focusedState = _types.default.ControlState.isFocusedState(p_state);

    var background = (0, _rectangle.makeFilledRectangle)(_types.default.paddedRectangle(p_Bbox, 1, 1, 1, 1), disabledState ? p_control.style("combobox_background_disabled_color") : 0xffffff, 1);

    var backgroundBbox = _balsamiq.default.clone(p_Bbox);

    backgroundBbox.height = p_comboHeight;
    var background2 = (0, _rectangle.makeFilledRectangle)(_types.default.paddedRectangle(backgroundBbox, 1, 1, p_Bbox.width - 20, 1), p_control.style("combobox_arrow_background_color"), 1, 1);
    var bordersColor = disabledState ? p_control.style("combobox_border_disabled_color") : focusedState ? p_control.style("default_selection_color") : p_color;

    var comboBounds = _types.default.newRectangle(p_Bbox.x, p_Bbox.y, p_Bbox.width, p_comboHeight);

    var combo = _displayitem.default.createDisplayItemsFromSkin("combobox", bordersColor, comboBounds, p_control._context);

    var x = paddingLeft;
    var iconsize = 10;
    var menuIconX = p_Bbox.x + p_Bbox.width - 5 - iconsize;
    var menuIconY = p_Bbox.y + (p_comboHeight - iconsize * 0.7) / 2;

    var menuIcon = _displayitem.default.newPathsAsset(p_control._context, "DownFillTriangleIcon", _types.default.newRectangle(menuIconX, menuIconY, iconsize, iconsize * 0.7), disabledState ? p_control.style("default_disabled_color") : p_color);

    var icon;

    if (p_iconData) {
      var iconColor = disabledState ? p_control.style("default_disabled_color") : focusedState ? p_control.style("default_selection_color") : p_iconColor ? p_iconColor : p_color;
      icon = _icon.default.makeDisplayItemsFromIconData(p_control, p_iconData, x, (p_comboHeight - p_iconData.size) / 2, iconColor);
      x += p_iconData.size + p_control.style("combobox_icon_padding");
    }

    var title = p_titleText.makeDisplayItems(_types.default.newRectangle(x + p_Bbox.x, p_Bbox.y, p_Bbox.width - p_control.style("combobox_arrow_width") > 0 ? p_Bbox.width - p_control.style("combobox_arrow_width") : 0, p_comboHeight), _types.default.TextAlign.LEFT, _types.default.Positions.CENTER);
    return [background, title, background2, combo, icon, menuIcon];
  }
};
var _default = object;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/bricks/frame.js":
/*!*********************************!*\
  !*** ./js/core/bricks/frame.js ***!
  \*********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var _overlaylabel = __webpack_require__(/*! core/bricks/overlaylabel */ "./js/core/bricks/overlaylabel.js");

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var object = {};
/**
 *
 * @param {Control} p_control
 * @param {Object} p_bounds
 * @param {String?} p_labelText
 * @returns {Array}
 */

object.makeImageFrame = function (p_control, p_bounds, p_labelText) {
  _balsamiq.default.exists(p_control);

  _balsamiq.default.exists(p_bounds);

  if (!_balsamiq.default.isString(p_labelText)) {
    p_labelText = "";
  }

  _balsamiq.default.assertString(p_labelText);

  var gap = p_control._context.skin.image_line_gap;
  var lineWidth = p_control._context.skin.image_line_width;
  var frameColor = p_control._context.skin.image_border_color;
  var crossColor = p_control._context.skin.image_cross_line_color;
  var cross = Path.newPath(p_bounds.width, p_bounds.height);
  cross.moveTo(gap, gap);
  cross.lineTo(p_bounds.width - gap, p_bounds.height - gap);
  cross.moveTo(gap, p_bounds.height - gap);
  cross.lineTo(p_bounds.width - gap, gap);
  var background = (0, _rectangle.makeFilledRectangle)(p_bounds, 0xFFFFFF);

  var frame = _displayitem.default.createDisplayItemsFromSkin("image_frame", frameColor, p_bounds, p_control._context);

  var crossItem = _displayitem.default.newSVGPath(p_bounds, cross, lineWidth, crossColor);

  var res = [background, frame, crossItem];

  if (p_labelText) {
    var obj = (0, _overlaylabel.makeOverlayLabel)(p_control, p_bounds.x + p_bounds.width / 2, p_bounds.y + p_bounds.height / 2, p_labelText);
    res = res.concat(obj.displayItems);
  }

  return res;
};

var _default = object;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/bricks/icon.js":
/*!********************************!*\
  !*** ./js/core/bricks/icon.js ***!
  \********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _frame = _interopRequireDefault(__webpack_require__(/*! core/bricks/frame */ "./js/core/bricks/frame.js"));

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var _fontawesomemap = __webpack_require__(/*! core/skins/fontawesomemap */ "./js/core/skins/fontawesomemap.js");

var IconsMap = _interopRequireWildcard(__webpack_require__(/*! core/bricks/linkiconmap */ "./js/core/bricks/linkiconmap.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var Image = __webpack_require__(/*! core/model/image */ "./js/core/model/image.js");

var object = {};
/**
 *
 * @param {SrcProperty} p_src
 * @param {Number} p_iconSize
 * @constructor
 */

object.IconProperty = function (p_src, p_iconSize) {
  this.src = p_src;
  this.size = p_iconSize;
};

object.IconData = function (p_src, p_iconSize, p_rotation) {
  this.src = p_src;
  this.size = p_iconSize;
  this.rotation = p_rotation;
};

object.newAppIconData = function (p_iconName, p_iconSize, p_rotation) {
  return new object.IconData(new _types.default.SrcProperty(_types.default.SrcLocation.APP_ASSETS, p_iconName), p_iconSize, p_rotation);
};
/**
 *
 * @param {IconProperty} p_iconProp
 * @param {Number} p_rotationProp
 * @returns {{iconData: IconData, dependencies: Array}}
 */


object.processIconProperty = function (p_iconProp, p_rotationProp) {
  var res = {
    iconData: null,
    dependencies: []
  };

  if (p_iconProp) {
    res.iconData = new object.IconData(p_iconProp.src, p_iconProp.size, p_rotationProp);

    if (p_iconProp.src.location !== _types.default.SrcLocation.APP_ASSETS) {
      res.dependencies = [p_iconProp.src];
    } else {
      res.dependencies = [];
    }
  }

  return res;
};
/**
 *
 * @param {Control} p_control
 */


object.processIconData = function (p_control) {
  var res = object.processIconProperty(p_control.getActualProperty("icon"), p_control.getActualProperty("rotation"));
  p_control.iconData = res.iconData;
  p_control.setDependencies(res.dependencies);
};
/**
 * Return an IconData object with empty icon, no rotation.
 * @param {number} p_iconSize
 * @returns {IconData}
 */


object.emptyIconData = function (p_iconSize) {
  return new object.IconData(new _types.default.SrcProperty(_types.default.SrcLocation.APP_ASSETS, "EmptyIcon"), p_iconSize, 0);
};
/**
 * @param {Control} p_control
 * @param {Object} p_iconData
 * @param {Number} p_x
 * @param {Number} p_y
 * @param {Number} p_color
 * @returns {Object}
 */


object.makeDisplayItemsFromIconData = function (p_control, p_iconData, p_x, p_y, p_color) {
  _balsamiq.default.exists(p_control);

  _balsamiq.default.exists(p_iconData);

  _balsamiq.default.assertNumber(p_x);

  _balsamiq.default.assertNumber(p_y);

  _balsamiq.default.assertNumber(p_color);

  var originalBounds = _types.default.newRectangle(p_x, p_y, p_iconData.size, p_iconData.size); //var r = p_iconData.rotation;
  //var x = p_x + p_iconData.size * ((r === 0 || r === 90) ? 0 : 1);
  //var y = p_y + p_iconData.size * ((r === 0 || r === 270) ? 0 : 1);
  //var b = Types.newRectangle(x, y, p_iconData.size, p_iconData.size);


  var res, b;

  if (p_iconData.src.location === _types.default.SrcLocation.APP_ASSETS) {
    if (p_control._context.useFontAwesomeIcons) {
      if (p_iconData.src.resourceID === "EmptyIcon") {
        res = object.makeAwesomeIconBorderDisplayItems(originalBounds, p_color, 0, p_control._context);
      } else {
        res = object.makeAwesomeIconDisplayItems(p_iconData, p_color, p_control, originalBounds);
      }
    } else {
      b = _displayitem.default.rotateBounds90(originalBounds, p_iconData.rotation);
      res = object.makeIconDisplayItems(p_iconData, b, p_color, p_control._context);
    }
  } else {
    b = _displayitem.default.rotateBounds90(originalBounds, p_iconData.rotation);
    res = object.makeCustomIconDisplayItems(p_iconData, p_control, b, p_color, p_x, p_y);
  }

  return res;
};

object.makeIconDisplayItems = function (p_iconData, p_bounds, p_color, p_context) {
  return _displayitem.default.newPathsAsset(p_context, p_iconData.src.resourceID, p_bounds, p_color, 1, p_iconData.rotation);
};

object.makeAwesomeIconBorderDisplayItems = function (p_bounds, p_color, p_rotation, p_context) {
  return _displayitem.default.newPathsAsset(p_context, "EmptyIconFontAwesome", p_bounds, p_color, 1, p_rotation);
};

object.makeAwesomeIconDisplayItems = function (iconData, textColor, control, bounds) {
  const textStyle = {
    fontFamily: "FontAwesome",
    fontSize: iconData.size,
    bold: false,
    italic: false,
    underline: false,
    textColor
  };
  const glyphText = (0, _fontawesomemap.getGlyphByIconName)(iconData.src.resourceID);

  if (!this.glyphText || !(0, _text.areTextsEqual)(this.glyphText, glyphText, textStyle)) {
    this.glyphText = (0, _text.newTextNoWrap)(glyphText, textStyle, control._context);
  }

  var measuredSize = this.glyphText.measuredSize();
  var ratio = measuredSize.width / bounds.width;

  if (ratio > 1) {
    textStyle.fontSize = textStyle.fontSize / ratio;
    this.glyphText = (0, _text.newTextNoWrap)(glyphText, textStyle, control._context);
    measuredSize = this.glyphText.measuredSize();
  }

  var glyphBounds = _types.default.newRectangle((bounds.width - measuredSize.width) / 2 + bounds.x, (bounds.height - measuredSize.height) / 2 + bounds.y, measuredSize.width, measuredSize.height);

  var b = _displayitem.default.rotateBounds90Glyph(glyphBounds, iconData.rotation, textStyle.fontSize * 0.075 * 2);

  var res = this.glyphText.makeDisplayItems(b, _types.default.TextAlign.CENTER, _types.default.Positions.CENTER);

  _displayitem.default.setRotation(res, iconData.rotation);

  return res;
};

object.makeCustomIconDisplayItems = function (p_iconData, p_control, p_bounds, p_color, p_x, p_y) {
  let resourceState = _types.default.ResourceState.NOT_AVAILABLE;
  let res;
  let resource;
  let resourceID = p_iconData.src.resourceID;

  if (resourceID) {
    resource = p_control._context.getDependency(resourceID);
  }

  if (resource) {
    _balsamiq.default.existsClass(resource, Image.Image);

    if (resource.nativeImageID) {
      resourceState = _types.default.ResourceState.AVAILABLE;
      var size = resource.getImageSize();
      res = _displayitem.default.newImage(resource.getImageID(), p_bounds, size.width, size.height);
      res.rotation = p_iconData.rotation;
      res.colorize = p_color;
    } else {
      resourceState = _types.default.ResourceState.PENDING;
      res = _frame.default.makeImageFrame(p_control, _types.default.newRectangle(p_x, p_y, p_iconData.size, p_iconData.size), "Loading...");
    }
  } else {
    // not available
    res = _frame.default.makeImageFrame(p_control, _types.default.newRectangle(p_x, p_y, p_iconData.size, p_iconData.size), "Image not found");
  }

  p_control._context.setReady(p_control.coreObjectID, resourceState === _types.default.ResourceState.AVAILABLE || resourceState === _types.default.ResourceState.NOT_AVAILABLE);

  return res;
};

object.updateIconsOnTextChange = function (p_control, p_newText, p_oldText, p_oldIcons) {
  if (p_oldIcons) {
    var newIcons = [];
    var newTexts = p_control.getLinkableLabels(p_newText);
    var iconMap = new IconsMap.Map();
    var sizeChanged = newTexts.length != p_oldText.length;
    p_oldIcons.forEach((icon, i) => {
      iconMap.mapValue(p_oldText[i], icon);
    });
    newTexts.forEach((label, i) => {
      var icon = iconMap.getValue(label);

      if (iconMap.hasKey(label)) {
        newIcons.push(icon);
      } else {
        if (!sizeChanged) {
          newIcons.push(p_oldIcons[i]);
        } else {
          newIcons.push(undefined);
        }
      }
    });
    return newIcons;
  }

  return null;
};

var _default = object;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/bricks/link.js":
/*!********************************!*\
  !*** ./js/core/bricks/link.js ***!
  \********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var LinksMap = _interopRequireWildcard(__webpack_require__(/*! core/bricks/linkiconmap */ "./js/core/bricks/linkiconmap.js"));

var _textconstants = __webpack_require__(/*! core/texts/textconstants */ "./js/core/texts/textconstants.js");

var _textutilities = __webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const uuidRegExp = new RegExp("[A-Z0-9]{8}-([A-Z0-9]{4}-){3}[A-Z0-9]{12}");
const object = {};
/**
 * Read the 'link' property of a control and, if not null, add the link area with the size of the entire control.
 * @param {Control} p_control
 */

object.processControlLink = function (p_control) {
  const linkProp = p_control.getActualProperty('link');

  if (linkProp) {
    _balsamiq.default.assertString(linkProp);

    p_control.addLinkArea(p_control.localBBox(), linkProp);
  }
};
/**
 * Add to the control the link areas for the given bounds array. Targets are taken from 'link' property.
 * Null targets and null bounds are skipped.
 * @param {Control} p_control
 * @param {Array} p_linkAreaBounds
 */


object.processLinks = function (p_control, p_linkAreaBounds) {
  const linkProp = p_control.getActualProperty('links');

  if (linkProp) {
    p_linkAreaBounds.forEach((bounds, index) => {
      if (index < linkProp.length) {
        const target = linkProp[index];

        if (target && bounds) {
          p_control.addLinkArea(bounds, target);
        }
      }
    });
  }
};
/**
 * Add to the control the link areas coming from bracketed links. If there is a link assigned
 * to the whole control, then the text links are ignored.
 * @param {Control} p_control
 * @param {Array} p_textLinks
 */


object.processTextLinks = function (p_control, p_textLinks) {
  const linkProp = p_control.hasPropertyDefinition('link') ? p_control.getActualProperty('link') : null;

  if (linkProp === "" || linkProp === null) {
    if (Array.isArray(p_textLinks)) {
      p_textLinks.forEach(textLink => {
        if (textLink.target) {
          p_control.addLinkArea(textLink.bounds, object.fixLinkTarget(textLink.target), textLink.continueOnNextLine, textLink.angle);
        }
      });
    }
  }
};

object.fixLinkTarget = function (linkTarget) {
  let newTarget = linkTarget;

  if (uuidRegExp.test(linkTarget)) {
    newTarget = 'resource:' + linkTarget;
  } else if (linkTarget.startsWith('action://')) {// do nothing, it's ok
  } else if (!linkTarget.startsWith('http://') && !linkTarget.startsWith('https://')) {
    newTarget = 'http://' + linkTarget;
  }

  return newTarget;
};

object.addMultiLinksMethods = function (control) {
  control.CONTROL_PROTOTYPE.getIcons = function () {
    let icons = this.getActualProperty("icons");
    const iconsFound = !!icons && icons.some(icon => icon && icon.size);

    if (!iconsFound) {
      icons = null;
    }

    return icons;
  };

  control.CONTROL_PROTOTYPE.getMultiLinks = function () {
    return this.getActualProperty("links");
  };
};

object.updateLinksOnIconChange = function (control, newIcons) {
  const linkmap = new LinksMap.Map();
  const oldLinks = control.getActualProperty("links");
  const newLinks = [];
  const oldIconResourceIDs = []; //remove default icon links from link list

  newLinks.push(oldLinks[0], oldLinks[1]);
  let lastLink = oldLinks[oldLinks.length - 1];
  const icons = control.getActualProperty("icons");

  if (icons !== null) {
    icons.forEach(icon => {
      oldIconResourceIDs.push(icon.src.resourceID);
    });

    for (let i = 2; i < oldLinks.length - 1; i++) {
      linkmap.mapValue(oldIconResourceIDs[i - 2], oldLinks[i]);
    }
  }

  let sizeChanged = oldIconResourceIDs.length !== newIcons.length;
  newIcons.forEach(icon => {
    let link = linkmap.getValue(icon.src.resourceID);

    if (linkmap.hasKey(icon.src.resourceID)) {
      newLinks.push(link);
    } else {
      if (!sizeChanged) {
        newLinks.push(link);
      } else {
        newLinks.push(null);
      }
    }
  });
  newLinks.push(lastLink);
  return newLinks;
};

object.updateLinksOnTextChange = function (control, newText, oldText, oldLinks) {
  if (oldLinks) {
    const newLinks = [];
    newText = control.getLinkableLabels(newText);
    const linkmap = new LinksMap.Map();
    const sizeChanged = newText.length !== oldText.length;
    oldLinks.forEach((link, i) => {
      linkmap.mapValue(oldText[i], link);
    });
    newText.forEach((label, i) => {
      const link = linkmap.getValue(label);

      if (linkmap.hasKey(label)) {
        newLinks.push(link);
      } else {
        if (!sizeChanged) {
          newLinks.push(oldLinks[i]);
        } else {
          newLinks.push(undefined);
        }
      }
    });
    return newLinks;
  }

  return null;
};

object.processNodesLinkTargetsToWireframeNames = function (nodes, getDependency) {
  nodes.forEach(node => {
    if (node.type === _textconstants.NodeTypes.LINK) {
      if (node.target) {
        const target = (0, _textutilities.nodesToString)(node.target);

        if (uuidRegExp.test(target)) {
          const resource = getDependency(target);

          if (resource) {
            node.target = [resource.getName()];
          }
        }
      }
    } else if (node.children) {
      object.processNodesLinkTargetsToWireframeNames(node.children, getDependency);
    }
  });
};

object.getResourceNameIfResource = function (target, getDependency) {
  _balsamiq.default.assertString(target);

  let res = target.startsWith("resource:") ? target.substring(9) : target;

  if (uuidRegExp.test(res)) {
    const resource = getDependency(res);

    if (resource) {
      return resource.getName();
    } else {
      //broken link!
      return null;
    }
  }

  return target;
};

object.processNodesWireframeNamesToLinkTargets = (nodes, getMockupIDByName) => {
  nodes.forEach(node => {
    if (node.type === _textconstants.NodeTypes.LINK && node.target) {
      const mockupName = (0, _textutilities.nodesToString)(node.target);
      const resourceID = getMockupIDByName(mockupName);

      if (resourceID) {
        node.target = [resourceID];
      }
    } else if (node.children) {
      object.processNodesWireframeNamesToLinkTargets(node.children, getMockupIDByName);
    }
  });
};

object.addTextLinksMethods = function (control, options) {
  options = options || {}; // getLinksFunction
  // replaceLinkTargetFunction
  // clearLinksTargetFunction
  // getTextWithMockupNames

  control.CONTROL_PROTOTYPE.getTextLinks = function () {
    let links = null;

    if (options.getLinksFunction) {
      links = options.getLinksFunction.call(this);
    } else {
      if (this.text) {
        links = this.text.getLinks();
      }
    }

    if (links) {
      return links.map(link => {
        if (link.target) {
          return { ...link,
            target: object.fixLinkTarget(link.target)
          };
        }

        return link;
      });
    }

    return [];
  };

  control.CONTROL_PROTOTYPE.replaceTextLinkTarget = function (linkIndex, newTarget) {
    if (newTarget && newTarget.startsWith("resource:")) {
      newTarget = newTarget.substring("resource:".length);
    }

    if (options.replaceLinkTargetFunction) {
      return options.replaceLinkTargetFunction.call(this, linkIndex, newTarget);
    } else {
      return (0, _textutilities.replaceLinkTarget)(this.getActualProperty("text"), linkIndex, newTarget);
    }
  };

  control.CONTROL_PROTOTYPE.clearTextLinks = function () {
    if (options.clearLinksTargetFunction) {
      return options.clearLinksTargetFunction.call(this);
    }

    const linkCount = this.getTextLinks().length;
    let newText = this.getActualProperty("text");

    for (let i = 0; i < linkCount; i++) {
      newText = (0, _textutilities.replaceLinkTarget)(newText, i, null);
    }

    return newText;
  };

  control.CONTROL_PROTOTYPE.getTextWithMockupNames = function () {
    const text = this.getActualProperty("text");
    const nodes = (0, _textutilities.stringToNodes)(text);
    object.processNodesLinkTargetsToWireframeNames(nodes, this._context.getDependency.bind(this));
    return (0, _textutilities.nodesToString)(nodes);
  };

  control.CONTROL_PROTOTYPE.getTextWithResourceIDs = function (textWithMockupNames) {
    const nodes = (0, _textutilities.stringToNodes)(textWithMockupNames);
    object.processNodesWireframeNamesToLinkTargets(nodes, this._context.getMockupIDByName.bind(this));
    return (0, _textutilities.nodesToString)(nodes);
  };
};

var _default = object;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/bricks/linkiconmap.js":
/*!***************************************!*\
  !*** ./js/core/bricks/linkiconmap.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Map = void 0;

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

//we currently use this class to map icons to text, links to text, and links to icons in controls.
class Map {
  constructor() {
    _defineProperty(this, "hasKey", function (key) {
      return key in this.map;
    });

    _defineProperty(this, "mapValue", function (key, value) {
      if (this.map[key] == null) {
        this.map[key] = [];
      }

      this.map[key].push(value);
    });

    _defineProperty(this, "getValue", function (key) {
      if (this.map[key] == null || this.map[key].length == 0) {
        return undefined;
      } else {
        return this.map[key].shift();
      }
    });

    this.map = {};
  }

}

exports.Map = Map;

/***/ }),

/***/ "./js/core/bricks/macro.js":
/*!*********************************!*\
  !*** ./js/core/bricks/macro.js ***!
  \*********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.findMacros = findMacros;
exports.processMacros = processMacros;
exports.processMacrosInText = processMacrosInText;
exports.resolveMacros = resolveMacros;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function findMacros(p_text) {
  var macros = [];

  _balsamiq.default.objectForEach(_types.default.TextMacro, function (macro) {
    if (p_text.indexOf(macro) >= 0) {
      macros.push(macro);
    }
  });

  return macros;
}

function resolveMacros(p_text, p_macros, p_control) {
  p_macros.forEach(function (macro) {
    var resolvedMacro = p_control.resolveMacro(macro); //it shouldn't use special characters to change text appearance when resolving a macro

    resolvedMacro = resolvedMacro.replace(new RegExp('_', "g"), '\\_');
    resolvedMacro = resolvedMacro.replace(new RegExp('\\*', "g"), '\\*');
    resolvedMacro = resolvedMacro.replace(new RegExp('\\[', "g"), '\\[');
    var regexp = new RegExp(macro, "g");
    p_text = p_text.replace(regexp, resolvedMacro);
  });
  return p_text;
}
/**
 * Find text macros in the "text" property of the given control and resolve them.
 * Set the control's macros to remember dependencies.
 * Return the resolved text.
 * @param {Object} p_control
 * @returns {string}
 */


function processMacros(p_control) {
  return processMacrosInText(p_control, p_control.getActualProperty("text"));
}

function processMacrosInText(p_control, p_text) {
  var macros = findMacros(p_text);
  var resolvedText = resolveMacros(p_text, macros, p_control);
  p_control.setMacros(macros);
  return resolvedText;
}

/***/ }),

/***/ "./js/core/bricks/menu.js":
/*!********************************!*\
  !*** ./js/core/bricks/menu.js ***!
  \********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.MenuLine = void 0;
exports.makeDisplayItemFromIconDataInBbox = makeDisplayItemFromIconDataInBbox;
exports.makeDisplayItems = makeDisplayItems;
exports.makeMenuDisplayItems = makeMenuDisplayItems;
exports.makeMenuLine = makeMenuLine;
exports.measureHeight = measureHeight;
exports.measureMenu = measureMenu;
exports.menuInit = menuInit;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var ICON_SIZE = 12;
var Padding = {
  "top": 3,
  "bottom": 3,
  "middle": 3
};

class MenuLine {
  static isSeparator(p_text) {
    var noSpacestext = p_text.replace(/ /g, "");
    return noSpacestext.replace(/-/g, "").length === 0 || noSpacestext.replace(/=/g, "").length === 0;
  }

  static hasDotIcon(p_text) {
    return p_text.substr(0, 2) === "o ";
  }

  static hasCheckmarkIcon(p_text) {
    return p_text.substr(0, 2) === "x ";
  }

  static textEndsWithArrows(p_text) {
    var trailing = p_text.substr(p_text.lastIndexOf(" ") + 1);
    return trailing.length > 0 && trailing.replace(/>/g, "").length === 0;
  }

  static trimTrailingSpacesAndCommas(p_text) {
    while (p_text && (p_text.substring(p_text.length - 1) === ' ' || p_text.substring(p_text.length - 1) === ',')) {
      p_text = p_text.substring(0, p_text.length - 1);
    }

    return p_text;
  }

}

exports.MenuLine = MenuLine;

const Menuline = function (control) {
  this.control = control;
  this.leftIcon = null;
  this.centerText = null;
  this.separator = null;
  this.rightText = null;
  this.rightIcon = null;
};

function menuInit(p_paddingTop, p_paddingBottom, p_paddingMiddle) {
  Padding = {
    top: p_paddingTop,
    bottom: p_paddingBottom,
    middle: p_paddingMiddle
  };
}

function measureMenu(p_control, p_sizeProperty) {
  const size = _balsamiq.default.clone(p_sizeProperty);

  let height = Padding["top"] + Padding["bottom"];
  let maxWidth = 0;
  let width; //calcola la lungezza massima delle righe

  height = p_control._menuLines.reduce((totalHeight, line) => {
    return totalHeight + measureHeight(line, size.width);
  }, height);

  p_control._menuLines.forEach(line => {
    width = 0;

    if (line.separator == null) {
      if (line.leftIcon != null) {
        width += 12;
      }

      if (line.rightIcon != null) {
        width += 12;
      }

      if (line.centerText != null) {
        width += line.centerText.measuredSize().width + 25;
      }

      if (line.rightText != null) {
        width += line.rightText.measuredSize().width;
      }

      maxWidth = Math.max(maxWidth, width);
    }
  });

  return _types.default.newSize(maxWidth, height);
}

function makeMenuDisplayItems(p_control, p_Bbox, p_menulines, p_y, p_selectedIndex, p_menulineX, p_color) {
  var border = _displayitem.default.createDisplayItemsFromSkin("rectangle_border_square", p_control.style("default_border_color"), p_Bbox, p_control._context);

  var linkBounds = [];
  var menulineItems = [];
  var paddingMiddle = Padding["middle"];
  var ruleHeight = p_control.style("menu_rule_padding");
  p_menulines.forEach((menuLine, index) => {
    var lineHeight, bbox;

    if (menuLine.centerText) {
      lineHeight = menuLine.centerText.measuredSize().height + paddingMiddle;
      bbox = _types.default.newRectangle(0, p_y, p_Bbox.width, lineHeight);
      p_y += lineHeight;
    } else if (menuLine.separator) {
      lineHeight = menuLine.control.styleSVGSize(menuLine.control.style("menu_separator_skin")).height;
      bbox = _types.default.newRectangle(0, p_y, p_Bbox.width, lineHeight);
      p_y += ruleHeight;
    }

    menulineItems.push(makeDisplayItems(menuLine, bbox, index === p_selectedIndex, p_menulineX, p_color));
    linkBounds.push(bbox);
  });

  if (p_control.hasPropertyDefinition('links')) {
    _link.default.processLinks(p_control, linkBounds);
  }

  return [border, menulineItems];
}

function makeMenuLine(control, text, selected) {
  let textColor = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : null;
  let useSoftEOL = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : true;
  let previousLine = arguments.length > 5 ? arguments[5] : undefined;
  const menuLine = new Menuline(control);

  if (MenuLine.isSeparator(text)) {
    menuLine.separator = true;
    return menuLine;
  }

  const previousRightText = previousLine && previousLine.rightText;
  const previousCenterText = previousLine && previousLine.centerText;
  textColor = textColor ? textColor : selected ? _colors.default.labelColorGivenBackgroundColor(control.style("default_selection_color")) : control.style("default_text_color");
  const linkColor = selected ? _colors.default.labelColorGivenBackgroundColor(control.style("default_selection_color")) : control.style("default_link_color");
  const style = control.newStandardTextStyle({
    textColor,
    linkColor
  });
  let centerText;
  let rightText;
  const placeholder = String.fromCharCode(0) + String.fromCharCode(0);
  centerText = text.replace("\u2028", "").replace(/\\,/g, placeholder); // escaped commas will be ignored and rendered as commas
  // Check for left icon

  if (MenuLine.hasDotIcon(centerText)) {
    menuLine.leftIcon = "DotIcon";
    centerText = centerText.substring(2);
  } else if (MenuLine.hasCheckmarkIcon(centerText)) {
    menuLine.leftIcon = "CheckMarkIcon";
    centerText = centerText.substring(2);
  } // Check for right arrow OR comma separated string


  if (MenuLine.textEndsWithArrows(centerText)) {
    menuLine.rightIcon = "breadcrumbs_separator";
    centerText = centerText.substr(0, centerText.lastIndexOf(" "));
  } else if (centerText.indexOf(",") !== -1) {
    rightText = MenuLine.trimTrailingSpacesAndCommas(centerText.substr(centerText.lastIndexOf(",") + 1));
    centerText = MenuLine.trimTrailingSpacesAndCommas(centerText.substring(0, centerText.lastIndexOf(",")));
  } // Create Texts if present


  if (rightText != null) {
    const text = rightText.replace(new RegExp(placeholder, 'g'), ",");
    menuLine.rightText = !previousRightText || !(0, _text.areTextsEqual)(previousRightText, text, style) ? useSoftEOL ? (0, _text.newSingleLineText)(text, style, control._context) : (0, _text.newSingleLineNoSoftEOL)(text, style, control._context) : previousRightText;
  }

  if (centerText != null) {
    const text = centerText.replace(new RegExp(placeholder, 'g'), ",");
    menuLine.centerText = !previousCenterText || !(0, _text.areTextsEqual)(previousCenterText, text, style) ? useSoftEOL ? (0, _text.newSingleLineText)(text, style, control._context) : (0, _text.newSingleLineNoSoftEOL)(text, style, control._context) : previousCenterText;
  }

  return menuLine;
}

function makeDisplayItemFromIconDataInBbox(p_menuline, p_name, p_iconWidth, p_iconHeight, p_bbox, p_color, p_halign, p_vcenter) {
  p_halign = _balsamiq.default.isUndefined(p_halign) ? _types.default.TextAlign.LEFT : p_halign;
  p_vcenter = _balsamiq.default.isUndefined(p_vcenter) ? false : p_vcenter;
  var x;

  if (p_halign === _types.default.TextAlign.LEFT) {
    x = p_bbox.x;
  } else if (p_halign === _types.default.TextAlign.CENTER) {
    x = p_bbox.x + (p_bbox.width - p_iconWidth) / 2;
  } else {
    x = p_bbox.x + (p_bbox.width - p_iconWidth);
  }

  var y = p_vcenter ? p_bbox.y + (p_bbox.height - p_iconHeight) / 2 : p_bbox.y;

  var b = _types.default.newRectangle(x, y, p_iconWidth, p_iconHeight);

  var d = _displayitem.default.createDisplayItemsFromSkin(p_name, p_color, b, p_menuline.control._context);

  return d;
} // BBox width is the whole control, here we have to add left and right paddings


function makeDisplayItems(p_menuline, p_bbox, p_selected) {
  let p_leftTextX = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : null;
  let p_color = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : 0;
  var textPaddingLeft = p_menuline.control.style("menu_padding_left");
  var textPaddingRight = p_menuline.control.style("menu_asset_padding_right");
  var iconPaddingLeft = p_menuline.control.style("menu_asset_padding_left");
  var paddingRight = p_menuline.control.style("menu_padding_right"); // Math.max(0, p_w-assetX-getStyle("paddingRight"));
  //iconpaddingx+(paddingx-iconpaddingx-icon.width)/2;
  // Icons will be centered inside this box

  var leftIconBbox = _types.default.newRectangle(iconPaddingLeft, p_bbox.y - 2, textPaddingLeft - iconPaddingLeft, p_bbox.height); //var selectionBbox = Types.newRectangle(p_bbox.x + iconPaddingLeft, p_bbox.y_, Math.max(0, p_bbox.width-iconPaddingLeft-paddingRight), p_bbox.height);


  var selectionBbox = _types.default.paddedRectangle(p_bbox, 0, 0, iconPaddingLeft, paddingRight);

  var textsBbox = _types.default.paddedRectangle(p_bbox, 0, 0, textPaddingLeft, paddingRight + textPaddingRight);

  var displayItems = [];
  var color = p_color;

  if (p_selected) {
    displayItems.push((0, _rectangle.makeFilledRectangle)(selectionBbox, p_menuline.control.style("default_selection_color"), 1));
    color = _colors.default.labelColorGivenBackgroundColor(p_menuline.control.style("default_selection_color"));
  }

  var centerBbox, iconSize;
  var rightItemWidth = 0;

  if (p_menuline.rightText) {
    rightItemWidth = p_menuline.rightText.measuredSizeForWidth(textsBbox.width).width;
    displayItems.push(p_menuline.rightText.makeDisplayItems(textsBbox, _types.default.TextAlign.RIGHT, _types.default.Positions.CENTER));
  }

  if (p_menuline.rightIcon) {
    iconSize = p_menuline.control.styleSVGSize("breadcrumbs_separator");
    rightItemWidth = iconSize.width;
    displayItems.push(makeDisplayItemFromIconDataInBbox(p_menuline, "breadcrumbs_separator", iconSize.width, iconSize.height, textsBbox, color, _types.default.TextAlign.RIGHT, true));
  }

  if (p_menuline.centerText) {
    centerBbox = _types.default.paddedRectangle(textsBbox, 0, 0, 0, rightItemWidth);

    if (_balsamiq.default.isNumber(p_leftTextX)) {
      centerBbox.x = p_leftTextX;
    }

    displayItems.push(p_menuline.centerText.makeDisplayItems(centerBbox, _types.default.TextAlign.LEFT, _types.default.Positions.CENTER));
  }

  if (p_menuline.leftIcon) {
    displayItems.push(makeDisplayItemFromIconDataInBbox(p_menuline, p_menuline.leftIcon, ICON_SIZE, ICON_SIZE, leftIconBbox, color, _types.default.TextAlign.CENTER, true)); // padding to custom position (arbitrary constants to match flex positioning)
  }

  if (p_menuline.separator) {
    var separatorSkin = p_menuline.control.style("menu_separator_skin");
    displayItems.push(makeDisplayItemFromIconDataInBbox(p_menuline, separatorSkin, selectionBbox.width, selectionBbox.height, selectionBbox, color, _types.default.TextAlign.LEFT, false));
  }

  return displayItems;
}

function measureHeight(p_menuline, p_width) {
  //if we have an empty line the line is converted to a separator
  if (!p_menuline.height) {
    if (p_menuline.separator) {
      return p_menuline.control.style("menu_rule_padding");
    }

    return p_menuline.centerText.measuredSizeForWidth(p_width).height + Padding["middle"];
  }
}

/***/ }),

/***/ "./js/core/bricks/overlaylabel.js":
/*!****************************************!*\
  !*** ./js/core/bricks/overlaylabel.js ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.makeOverlayLabel = makeOverlayLabel;
exports.makeTextBackground = makeTextBackground;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");
/**
 * Make a white semi-transparent rounded rectangle for backgrounds of labels.
 * @param {Object} p_textBounds
 * @param {Number?} p_alpha
 * @returns {Array}
 */


function makeTextBackground(p_textBounds, p_alpha) {
  if (!_balsamiq.default.isNumber(p_alpha)) {
    p_alpha = 1;
  }

  var bounds = _types.default.paddedRectangle(p_textBounds, -4, -4, -5, -5);

  var path = Path.newRoundedRectanglePath(bounds.width, bounds.height, 1, 4, 5);

  var fill = _displayitem.default.newSVGPath(bounds, path, 0, null, 0xFFFFFF);

  var border = _displayitem.default.newSVGPath(bounds, path, 1, 0xCCCCCC);

  border.borderAlpha = fill.alpha = 0.85 * p_alpha;
  return [fill, border];
}
/**
 *
 * @param {Control} p_control
 * @param {Number} p_x
 * @param {Number} p_y
 * @param {String} p_labelText
 * @param {Number?} p_alpha
 * @param {Object?} p_textStyle
 * @returns {Object}
 */


function makeOverlayLabel(p_control, p_x, p_y, p_labelText, p_alpha, p_textStyle) {
  _balsamiq.default.assertNumber(p_x);

  _balsamiq.default.assertNumber(p_y);

  _balsamiq.default.assertString(p_labelText);

  if (!_balsamiq.default.isNumber(p_alpha)) {
    p_alpha = 1;
  }

  var textStyle;

  if (p_textStyle) {
    textStyle = p_textStyle;
  } else {
    textStyle = p_control.newTextStyle();
    textStyle.textColor = 0x000000;
  }

  var text = textMod.newText(p_labelText, textStyle, p_control._context);
  var textSize = text.measuredSize();

  var textBounds = _types.default.newRectangle(p_x - textSize.width / 2, p_y - textSize.height / 2, textSize.width, textSize.height);

  var background = makeTextBackground(textBounds, p_alpha);
  var textObj = text.makeDisplayItemsAndLinks(textBounds, _types.default.TextAlign.CENTER);
  textObj.displayItems.forEach(item => {
    item.alpha = p_alpha;
  });
  return {
    displayItems: [background, textObj.displayItems],
    links: textObj.links,
    text: text
  };
}

/***/ }),

/***/ "./js/core/bricks/rectangle.js":
/*!*************************************!*\
  !*** ./js/core/bricks/rectangle.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.makeFilledRectangle = makeFilledRectangle;
exports.makeRectangleBorderDisplayItems = makeRectangleBorderDisplayItems;
exports.makeRectangleDisplayItems = makeRectangleDisplayItems;
exports.makeStrokedRectangle = makeStrokedRectangle;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 *
 * @param {Object} p_bounds
 * @param {Number} p_borderColor
 * @param {Number} p_borderWidth
 * @param {Number?} p_cornerRadius
 * @returns {Array}
 */
function makeStrokedRectangle(p_bounds, p_borderColor, p_borderWidth, p_cornerRadius) {
  _balsamiq.default.exists(p_bounds);

  _balsamiq.default.assertNumber(p_borderColor);

  _balsamiq.default.assertNumber(p_borderWidth);

  if (!_balsamiq.default.isNumber(p_cornerRadius)) {
    p_cornerRadius = 0;
  }

  _balsamiq.default.assertNumber(p_cornerRadius);

  var p;

  if (p_cornerRadius > 0) {
    p = Path.newRoundedRectanglePath(p_bounds.width, p_bounds.height, p_borderWidth, p_cornerRadius);
  } else {
    p = Path.newRectanglePath(p_bounds.width, p_bounds.height, p_borderWidth);
  }

  var items = [_displayitem.default.newSVGPath(p_bounds, p, p_borderWidth, p_borderColor)];
  return items;
}
/**
 * @param {Object} p_bounds
 * @param {Number?} p_fillColor
 * @param {Number?} p_alpha
 * @param {Number?} p_cornerRadius
 * @returns {Array}
 */


function makeFilledRectangle(p_bounds, p_fillColor, p_alpha, p_cornerRadius) {
  if (!_balsamiq.default.isNumber(p_fillColor)) {
    p_fillColor = 0xFFFFFF;
  }

  if (!_balsamiq.default.isNumber(p_cornerRadius)) {
    p_cornerRadius = 0;
  }

  if (!_balsamiq.default.isNumber(p_alpha)) {
    p_alpha = 1;
  }

  _balsamiq.default.assertNumber(p_fillColor);

  _balsamiq.default.assertNumber(p_alpha);

  _balsamiq.default.assertNumber(p_cornerRadius);

  var p;

  if (p_cornerRadius > 0) {
    p = Path.newRoundedRectanglePath(p_bounds.width, p_bounds.height, 0, p_cornerRadius);
  } else {
    p = Path.newRectanglePath(p_bounds.width, p_bounds.height, 0);
  }

  var items = [_displayitem.default.newSVGPath(p_bounds, p, null, null, p_fillColor)];

  _displayitem.default.setAlpha(items, p_alpha);

  return items;
}
/**
 * @param {Object} p_context
 * @param {Object} p_bounds
 * @param {String?} p_borderStyle
 * @param {Number?} p_fillColor
 * @param {Number?} p_alpha
 * @param {Number?} p_borderColor
 * @returns {Array}
 */


function makeRectangleDisplayItems(p_context, p_bounds, p_borderStyle, p_fillColor, p_alpha, p_borderColor) {
  if (!_balsamiq.default.isString(p_borderStyle)) {
    p_borderStyle = Props.BorderStyles.SQUARE;
  }

  _balsamiq.default.assertEnum(p_borderStyle, Props.BorderStyles);

  if (!_balsamiq.default.isNumber(p_fillColor)) {
    p_fillColor = 0xFFFFFF;
  }

  if (!_balsamiq.default.isNumber(p_alpha)) {
    p_alpha = 1;
  }

  if (!_balsamiq.default.isNumber(p_borderColor)) {
    p_borderColor = p_context.skin.default_border_color;
  }

  _balsamiq.default.exists(p_context);

  _balsamiq.default.exists(p_bounds);

  var items = [];
  var backgroundPadding = p_context.skin.rectangle_background_padding.default;

  if (_balsamiq.default.has(p_context.skin.rectangle_background_padding, p_borderStyle)) {
    backgroundPadding = p_context.skin.rectangle_background_padding[p_borderStyle];
  }

  var paddedBbox = _types.default.paddedRectangle(p_bounds, backgroundPadding, backgroundPadding, backgroundPadding, backgroundPadding);

  if (paddedBbox.width > 0 && paddedBbox.height > 0) {
    items = items.concat(makeFilledRectangle(paddedBbox, p_fillColor, p_alpha, p_context.skin.rectangle_corner_radius[p_borderStyle]));
  }

  var assetName = null;

  if (p_borderStyle === Props.BorderStyles.SQUARE || p_borderStyle === Props.BorderStyles.SQUARE_BREAKLINE || p_borderStyle === Props.BorderStyles.SQUARE_BREAKLINE_TOP) {
    assetName = "rectangle_border_square";
  } else if (p_borderStyle === Props.BorderStyles.ROUNDED_SOLID) {
    assetName = "rectangle_border_rounded";
  } else if (p_borderStyle === Props.BorderStyles.ROUNDED_DOTTED) {
    var d = p_context.skin.rectangle_border_rounded_dotted;
    var g = Path.newRoundedRectanglePath(p_bounds.width, p_bounds.height, d.width, d.corner_radius);

    var i = _displayitem.default.newSVGPath(p_bounds, g, d.width, p_borderColor);

    i.dashArray = d.dash_array;
    items.push(i);
  }

  if (assetName) {
    items = items.concat(_displayitem.default.createDisplayItemsFromSkin(assetName, p_borderColor, p_bounds, p_context));
  }

  return items;
}
/**
 * @param {Object} p_context
 * @param {Object} p_bounds
 * @param {Number?} p_borderColor
 * @param {String?} p_borderStyle
 * @returns {Array}
 */


function makeRectangleBorderDisplayItems(p_context, p_bounds, p_borderColor, p_borderStyle) {
  return makeRectangleDisplayItems(p_context, p_bounds, p_borderStyle, 0, 0, p_borderColor);
}

/***/ }),

/***/ "./js/core/bricks/scrollbars.js":
/*!**************************************!*\
  !*** ./js/core/bricks/scrollbars.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Sizes = void 0;
exports.makeDisplayItemsForHorizontalScrollbar = makeDisplayItemsForHorizontalScrollbar;
exports.makeDisplayItemsForVerticalScrollbar = makeDisplayItemsForVerticalScrollbar;
exports.makeVerticalScrollbar = makeVerticalScrollbar;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var rectangleMod = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

const Sizes = {
  ScrollbarFixedSize: 16,
  ThumbLongSize: 21,
  ThumbShortSize: 8,
  ArrowButtonLongSize: 12,
  ScrollbarDefaultVariableSize: 100
};
exports.Sizes = Sizes;

function makeDisplayItemsForVerticalScrollbar(p_x, p_y, p_height, p_thumbPos, p_state, p_context) {
  let p_width = arguments.length > 6 && arguments[6] !== undefined ? arguments[6] : Sizes.ScrollbarFixedSize;

  _balsamiq.default.assertPositiveNumber(p_height);

  _balsamiq.default.assertNumber01(p_thumbPos);

  var bounds = _types.default.newRectangle(p_x, p_y, p_width, p_height);

  var r = []; // Background

  r.push(rectangleMod.makeFilledRectangle(_types.default.shrinkRectangle(bounds, 2), p_context.skin.scrollbar_background_color, 1, p_context.skin.scrollbar_background_radius));
  var color = p_context.skin.default_border_color; //Default skin color for scrollbar

  if (p_state === _types.default.ControlState.DISABLED) {
    color = p_context.skin.default_disabled_color;
  } // Scrollbar


  r = r.concat(_displayitem.default.createDisplayItemsFromSkin("verticalScrollbar", color, bounds, p_context)); // Thumb

  if (p_height - Sizes.ArrowButtonLongSize * 2 > Sizes.ThumbLongSize) {
    var thumbY = p_y + p_context.skin.verticalScrollbar_padding_top + (p_height - Sizes.ThumbLongSize - p_context.skin.verticalScrollbar_padding_top - p_context.skin.verticalScrollbar_padding_bottom) * p_thumbPos;
    var thumbX = p_x + (p_width - Sizes.ThumbShortSize) / 2.0;

    var thumbBounds = _types.default.newRectangle(thumbX, thumbY, Sizes.ThumbShortSize, Sizes.ThumbLongSize);

    r = r.concat(_displayitem.default.newPathsAsset(p_context, "verticalScrollbar_thumb", thumbBounds, color));
  }

  return r;
}

function makeVerticalScrollbar(p_localBbox, p_paddingTop, p_paddingBottom, p_paddingRight, p_value, p_state, p_context, p_width) {
  var width = _balsamiq.default.isUndefined(p_width) ? Sizes.ScrollbarFixedSize : p_width;
  var x = p_localBbox.width - p_paddingRight - width;
  var y = p_paddingTop;
  var height = p_localBbox.height - p_paddingTop - p_paddingBottom;
  var scrollbar;

  if (height > 0) {
    scrollbar = makeDisplayItemsForVerticalScrollbar(x, y, height, p_value, p_state, p_context, width);
  }

  return scrollbar;
}

function makeDisplayItemsForHorizontalScrollbar(p_x, p_y, p_width, p_thumbPos, p_context) {
  _balsamiq.default.assertNumber(p_width);

  var bounds = _types.default.newRectangle(p_x, p_y, p_width, Sizes.ScrollbarFixedSize);

  var r = []; // White Background

  r.push(rectangleMod.makeFilledRectangle(_types.default.shrinkRectangle(bounds, 2), p_context.skin.scrollbar_background_color, 1, p_context.skin.scrollbar_background_radius)); // Scrollbar

  r = r.concat(_displayitem.default.createDisplayItemsFromSkin("horizontalScrollbar", p_context.skin.default_border_color, bounds, p_context)); // Thumb

  if (p_width - Sizes.ArrowButtonLongSize * 2 > Sizes.ThumbLongSize) {
    var thumbY = p_y + (Sizes.ScrollbarFixedSize - Sizes.ThumbShortSize) / 2.0;
    var thumbX = p_x + p_context.skin.horizontalScrollbar_padding_left + (p_width - Sizes.ThumbLongSize - p_context.skin.horizontalScrollbar_padding_left - p_context.skin.horizontalScrollbar_padding_right) * p_thumbPos;

    var thumbBounds = _types.default.newRectangle(thumbX, thumbY, Sizes.ThumbLongSize, Sizes.ThumbShortSize);

    r = r.concat(_displayitem.default.newPathsAsset(p_context, "horizontalScrollbar_thumb", thumbBounds, p_context.skin.default_border_color));
  }

  return r;
}

/***/ }),

/***/ "./js/core/bricks/switch.js":
/*!**********************************!*\
  !*** ./js/core/bricks/switch.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.makeSwitchDisplayItems = makeSwitchDisplayItems;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 *
 * @param {Object} p_context
 * @param {Object} p_bounds
 * @param {Boolean} p_on
 * @param {Number?} p_color
 * @returns {*[]}
 */
function makeSwitchDisplayItems(p_context, p_bounds, p_on, p_color) {
  _balsamiq.default.exists(p_context);

  _balsamiq.default.assert(_balsamiq.default.has(p_context, "skin") && _balsamiq.default.has(p_context.skin, "switch_default_color") && _balsamiq.default.has(p_context.skin, "switch_border_disabled_color") && _balsamiq.default.has(p_context.skin, "switch_background_disabled_color") && _balsamiq.default.has(p_context.skin, "switch_border") && _balsamiq.default.has(p_context.skin, "switch_background"));

  _balsamiq.default.exists(p_bounds);

  _balsamiq.default.assertBool(p_on);

  if (!_balsamiq.default.isNumber(p_color)) {
    p_color = p_context.skin.switch_default_color;
  }

  _balsamiq.default.assertNumber(p_color);

  if (!p_on) {
    p_bounds = _types.default.newRectangle(p_bounds.x + p_bounds.width, p_bounds.y + p_bounds.height, p_bounds.width, p_bounds.height);
  }

  var border = _displayitem.default.newPathsAsset(p_context, "switch_border", p_bounds, p_on ? p_color : p_context.skin.switch_border_disabled_color, 1, p_on ? 0 : 180);

  var background = _displayitem.default.newPathsAsset(p_context, "switch_background", p_bounds, p_on ? p_color : p_context.skin.switch_background_disabled_color, 1, p_on ? 0 : 180);

  return [background, border];
}

/***/ }),

/***/ "./js/core/controllers/dirtywireframescontroller.js":
/*!**********************************************************!*\
  !*** ./js/core/controllers/dirtywireframescontroller.js ***!
  \**********************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.DirtyWireframesController = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var ResourceDisplayVM = _interopRequireWildcard(__webpack_require__(/*! editor/viewmodels/resourcedisplayviewmodel */ "./js/editor/viewmodels/resourcedisplayviewmodel.js"));

var Archive = _interopRequireWildcard(__webpack_require__(/*! core/archive */ "./js/core/archive.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var Events = __webpack_require__(/*! core/events */ "./js/core/events.ts");

var Project = __webpack_require__(/*! core/model/project */ "./js/core/model/project.js");

var CoreObject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var Resource = __webpack_require__(/*! core/model/resource */ "./js/core/model/resource.js");

var Constants = __webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts");

class DirtyWireframesController {
  constructor(context, archiveID, project) {
    _balsamiq.default.exists(context.archives);

    _balsamiq.default.exists(context.eventCenter);

    this.context = context;
    this.archiveID = archiveID;
    this.archive = this.context.archives.getArchive(archiveID);

    _balsamiq.default.assertProtocol(this.archive, Archive.Archive);

    this.project = project;
    this.context.eventCenter.addListener({
      archiveID: this.archive.coreObjectID.archiveID
    }, Events.EventTypes.ANY, this, this.onArchiveEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.RESOURCE_BRANCHID_CHANGED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Resource.EventTypes.RESOURCE_CONTENTS_CHANGED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Resource.EventTypes.RESOURCE_ATTRIBUTES_DIRTY, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.RESOURCE_ADDED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.RESOURCES_DELETED, this, this.onDirtyEvent);
  }

  cleanUp() {
    this.context.eventCenter.removeAllListeners(this);
  }

  reset() {
    this.project.dirtyWireframes.reset();
  }

  onArchiveEvent(event) {
    switch (event.eventType) {
      case Archive.EventTypes.RESOURCES_DELETED:
        event.eventData.resourceIDs.forEach(resourceID => {
          this.removeDirty(CoreObject.makeResourceCoreObjectID(this.project.coreObjectID, event.eventData.branchID, resourceID));
        });
        break;

      case Archive.EventTypes.RESOURCE_ADDED:
        this.addDirty(CoreObject.makeResourceCoreObjectID(this.project.coreObjectID, event.eventData.branchID, event.eventData.resourceID));
        break;

      case Archive.EventTypes.RESOURCE_DATA_CHANGED:
        this.addDirty(CoreObject.makeResourceCoreObjectID(this.project.coreObjectID, event.eventData.branchID, event.eventData.resourceID));
        break;

      default:
        break;
    }
  }

  onDirtyEvent(event) {
    switch (event.eventType) {
      case Project.EventTypes.RESOURCE_ADDED:
        // needed when importing resources
        this.addDirty(event.eventData.coreObjectID);
        break;

      case Project.EventTypes.RESOURCES_DELETED:
        // needed when undoing import resources
        event.eventData.resourceIDs.forEach(resourceID => {
          this.removeDirty(CoreObject.makeResourceCoreObjectID(this.project.coreObjectID, event.eventData.branchID, resourceID));
        });
        break;

      case Resource.EventTypes.RESOURCE_CONTENTS_CHANGED:
        this.addDirty(event.senderCoreObjectID);
        break;

      case Project.EventTypes.RESOURCE_BRANCHID_CHANGED:
        {
          let coreObject = CoreObject.makeResourceCoreObjectID(event.senderCoreObjectID, event.eventData.newBranchID, event.eventData.resourceID);
          this.addDirty(coreObject);
          break;
        }

      case Resource.EventTypes.RESOURCE_ATTRIBUTES_DIRTY:
        {
          let res = this.project.getResource(event.senderCoreObjectID);

          if (res.isTrashed()) {
            this.removeDirty(event.senderCoreObjectID);
          } else {
            this.addDirty(event.senderCoreObjectID);
          }

          break;
        }

      default:
        break;
    }
  }

  addDirty(coreObjectID) {
    let res = this.project.getResource(coreObjectID);

    if (res && res.getType() === Constants.ResourceTypes.MOCKUP) {
      this.project.dirtyWireframes.addDirty(coreObjectID);
    }
  }

  removeDirty(coreObjectID) {
    this.project.dirtyWireframes.removeDirty(coreObjectID);
  }

  isDirty(coreObjectID) {
    return this.project.dirtyWireframes.isDirty(coreObjectID);
  }

  getDirtyWireframes() {
    return this.project.dirtyWireframes.getDirty();
  }

  hasDirtyWireframes() {
    return this.project.dirtyWireframes.hasDirty();
  }

  addAllDirty() {
    const projectResources = this.project.getKeyValue(Project.KeyNames.MOCKUP_LIST_FLAT);
    projectResources.forEach(r => this.addDirty(r.coreObjectID));
  }

  addDirtyResources(resourceList) {
    _balsamiq.default.assertArray(resourceList.resources);

    let projectResources = this.project.getKeyValue(Project.KeyNames.MOCKUP_LIST_FLAT);

    if (projectResources) {
      for (let x = 0; x < projectResources.length; x++) {
        let alreadyExported = false;

        for (let i = 0; i < resourceList.resources.length; i++) {
          let resource = resourceList.resources[i];

          _balsamiq.default.exists(resource.resourceID);

          _balsamiq.default.exists(resource.branchID);

          if (projectResources[x].coreObjectID.branchID == resource.branchID && projectResources[x].coreObjectID.resourceID == resource.resourceID) {
            alreadyExported = true;
            break;
          }
        }

        if (!alreadyExported) {
          this.addDirty(projectResources[x].coreObjectID);
        }
      }
    }
  }

}

exports.DirtyWireframesController = DirtyWireframesController;

/***/ }),

/***/ "./js/core/controllers/emptyqueuecontroller.js":
/*!*****************************************************!*\
  !*** ./js/core/controllers/emptyqueuecontroller.js ***!
  \*****************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.EmptyQueueController = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _archivestate = __webpack_require__(/*! core/model/archivestate */ "./js/core/model/archivestate.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var CoreObject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var Project = __webpack_require__(/*! core/model/project */ "./js/core/model/project.js");

class EmptyQueueController {
  constructor(context, project, callbackIsDirty) {
    _balsamiq.default.exists(context.eventCenter);

    _balsamiq.default.exists(project);

    this.context = context;
    this.quitting = false;
    this.quitCallback = null;
    this.project = project;
    this.callbackIsDirty = callbackIsDirty;
    this.isDirtyVal = false;
    this.projectHasPendingOps = false;
    this.projectControllerHasEvents = false;
    this.context.eventCenter.addListener(this.project.coreObjectID, CoreObject.EventTypes.KEY_VALUE_CHANGED, this, this.projectHasPendingOpsChanged);
  }

  cleanUp() {
    this.context.eventCenter.removeAllListeners(this);
  }

  projectHasPendingOpsChanged(event) {
    if (event.eventData.keyName === Project.KeyNames.HAS_COMING_USER_EVENTS) {
      this.projectHasPendingOps = event.eventData.keyValue;
      this.updateDirty();
    }
  }

  pcEmptyQueue() {
    if (!this.project.hasComingUserEvents) {
      this.projectControllerHasEvents = false;
      this.updateDirty();
    }
  }

  pcNotEmptyQueue() {
    this.projectControllerHasEvents = true;
    this.updateDirty();
  }

  updateDirty() {
    this.isDirtyVal = this.projectControllerHasEvents || this.projectHasPendingOps;
    this.callbackIsDirty && this.callbackIsDirty(this.isDirtyVal);

    if (this.quitting && this.quitCallback && !this.isDirtyVal) {
      this.quitCallback();
      this.resetQuit();
    }
  }

  hasPendingOperations() {
    return this.isDirtyVal;
  }

  hasErrors() {
    const archiveError = this.project.getArchiveState().getKeyValue(_archivestate.KeyNames.ERROR) !== null;
    return archiveError;
  }

  quit(callback) {
    _balsamiq.default.assert(!this.quitCallback, "Second attempt to quit from EmptyQueueController");

    if (!this.hasPendingOperations() || this.hasErrors()) {
      callback();
    } else {
      this.quitting = true;
      this.quitCallback = callback;
    }
  }

  resetQuit() {
    this.quitting = false;
    this.quitCallback = null;
  }

}

exports.EmptyQueueController = EmptyQueueController;

/***/ }),

/***/ "./js/core/controllers/projectcontroller.js":
/*!**************************************************!*\
  !*** ./js/core/controllers/projectcontroller.js ***!
  \**************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = exports.ProjectController = exports.LoadingErrors = exports.EventTypes = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Archive = _interopRequireWildcard(__webpack_require__(/*! core/archive */ "./js/core/archive.js"));

var _operation = __webpack_require__(/*! core/operations/operation */ "./js/core/operations/operation.ts");

var _archiveeventoperation = __webpack_require__(/*! core/operations/archiveeventoperation */ "./js/core/operations/archiveeventoperation.js");

var _BmprConstants = __webpack_require__(/*! @balsamiq/bmpr/lib/BmprConstants */ "./node_modules/@balsamiq/bmpr/lib/BmprConstants.js");

var BARConstants = _interopRequireWildcard(__webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchiveConstants */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js"));

var _BalsamiqArchive = __webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchive */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchive.js");

var _nativeConstants = __webpack_require__(/*! core/native-constants */ "./js/core/native-constants.js");

var _dirtywireframescontroller = __webpack_require__(/*! core/controllers/dirtywireframescontroller */ "./js/core/controllers/dirtywireframescontroller.js");

var _emptyqueuecontroller = __webpack_require__(/*! core/controllers/emptyqueuecontroller */ "./js/core/controllers/emptyqueuecontroller.js");

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var Native = _interopRequireWildcard(__webpack_require__(/*! core/native */ "./js/core/native.js"));

var _userfilesynchronizer = __webpack_require__(/*! core/controllers/userfilesynchronizer */ "./js/core/controllers/userfilesynchronizer.js");

var Settings = _interopRequireWildcard(__webpack_require__(/*! core/core/settings */ "./js/core/core/settings.js"));

var _changetracker = __webpack_require__(/*! core/model/changetracker */ "./js/core/model/changetracker.js");

var _locators = __webpack_require__(/*! core/model/locators */ "./js/core/model/locators.js");

var _licensecontroller = __webpack_require__(/*! core/license/licensecontroller */ "./js/core/license/licensecontroller.js");

var _fileutils = __webpack_require__(/*! core/utils/fileutils */ "./js/core/utils/fileutils.js");

var MimeTypes = _interopRequireWildcard(__webpack_require__(/*! core/mimetypes */ "./js/core/mimetypes.js"));

var _progress = _interopRequireDefault(__webpack_require__(/*! core/utils/progress */ "./js/core/utils/progress.ts"));

var _resourceSkimmer = _interopRequireDefault(__webpack_require__(/*! core/utils/resource-skimmer */ "./js/core/utils/resource-skimmer.ts"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const Events = __webpack_require__(/*! core/events */ "./js/core/events.ts");

const Project = __webpack_require__(/*! core/model/project */ "./js/core/model/project.js");

const Resource = __webpack_require__(/*! core/model/resource */ "./js/core/model/resource.js");

const Branch = __webpack_require__(/*! core/model/branch */ "./js/core/model/branch.js");

const Thumbnail = __webpack_require__(/*! core/model/thumbnail */ "./js/core/model/thumbnail.ts");

const Constants = __webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts");

const CoreObject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

const ArchiveState = __webpack_require__(/*! core/model/archivestate */ "./js/core/model/archivestate.js");

const CommentsTextUtils = __webpack_require__(/*! core/viewmodels/commentstextutils */ "./js/core/viewmodels/commentstextutils.js");

const USE_TIMES = false; // warning: enable this flag only in dev mode, because it generates RSOD in Cloud production

const ASSETS_PER_REQUEST_SQLITE = 16; // read batches of 16 assets from SQLite archives

const DEFAULT_ASSETS_PER_REQUEST_BAS = 0; // read assets in parallel by default from BAS archives

const {
  makeASTFromText,
  getMentionedUserIDs
} = CommentsTextUtils;
const EventTypes = {
  EMPTY_QUEUE: "empty_queue",
  NOT_EMPTY_QUEUE: "not_empty_queue",
  AUTOSAVE_CHANGED: "autosave_changed"
};
exports.EventTypes = EventTypes;
const AUTOSAVE_SETTINGS_KEY = "autosave";
const SESSION_CLOSED_ERROR = "Session closed"; // TODO move elsewhere

const LoadingErrors = {
  invalidSchema: "Not a valid BMPR schema",
  corruptedBMPR: "The BMPR is corrupted",
  openArchive: "Error opening the archive",
  viewerCantMigrate: "VIEWER user can't migrate this file from BMPR 1.2 to 2.0",
  forbiddenLocation: "Sorry, to prevent data loss it is not possible to open a project from the application special folders."
};
exports.LoadingErrors = LoadingErrors;

class ProjectController extends CoreObject.CoreObject {
  constructor(context, archiveID, project, options) {
    super(context, CoreObject.makeVoidCoreObjectID());

    _balsamiq.default.exists(context.archives);

    _balsamiq.default.exists(context.eventCenter);

    _balsamiq.default.exists(context.host);

    _balsamiq.default.exists(context.operationQueue);

    _balsamiq.default.exists(context.settings);

    _balsamiq.default.exists(context.execTime);

    _balsamiq.default.assertClass(project, Project.Project);

    _balsamiq.default.assertString(archiveID);

    this.serverTimeDiff = null;
    this.userID = null;
    this.project = project;
    this.licenseController = null;
    this.options = options;
    this.queue = [];
    this.runningFunc = null;
    this.warningOperationTimeout = null;
    this.lastOperationCompleteDate = null;
    this.blockedByError = false; // true when the Archive returns an error

    this.blockedByOutOfSync = false; // true when the we missed some RTC event

    this.blockedByRestore = false; // true when the project has been restored to a snapshot

    this.blockedBySlowQueue = false; // true when the queue takes too much time to process a function

    this.blockedByCleanUp = false; // true when the cleanUp gets called

    this.blockedByForceQuit = false; // true when users want to force quit

    this.blockedByLoadProjectCanceled = false; // true when users cancel a load project

    this.blockReload = false;
    this.loadProjectCallback = null;
    this.archiveID = null;
    this.archive = null;
    this.lastArchiveRevision = null;
    this.setArchiveID(archiveID);
    this.dirtyWireframesController = new _dirtywireframescontroller.DirtyWireframesController({
      archives: this.context.archives,
      eventCenter: this.context.eventCenter,
      getCoreObject: this.context.getCoreObject
    }, archiveID, project);
    this.emptyQueueController = new _emptyqueuecontroller.EmptyQueueController(this.context, project, isDirty => {
      if (_balsamiq.default.isFunction(options.getIsDirtyCallback) && _balsamiq.default.isFunction(options.getIsDirtyCallback())) {
        options.getIsDirtyCallback()(isDirty);
      }
    });
    this.getBasCfg = options.getBasCfg;
    this.refreshBasCfg = options.refreshBasCfg;
    this.userFile = null; // { userFileURL:String, isEdited:Bool }

    this.lastSynchronizedArchiveRevision = null;
    this.changeTracker = new _changetracker.ChangeTracker(this.logUserEvent.bind(this));
    this.context.eventCenter.addListener({}, Settings.EventTypes.SETTINGS_CHANGED, this, this.onSettingsChanged);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.PROJECT_ATTRIBUTES_DIRTY, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.RESOURCE_ADDED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.COMMENT_ADDED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.COMMENT_IMPORTED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.COMMENTS_ATTRIBUTES_UPDATED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.COMMENT_DATA_UPDATED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.MYUSER_CREATED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.MYUSER_UPDATED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.RESOURCES_DELETED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.BRANCH_ADDED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.BRANCHES_DELETED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.THUMBNAIL_ADDED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.RESOURCE_BRANCHID_CHANGED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Resource.EventTypes.RESOURCE_ATTRIBUTES_DIRTY, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Resource.EventTypes.RESOURCE_DATA_DIRTY, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Branch.EventTypes.BRANCH_ATTRIBUTES_DIRTY, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Thumbnail.EventTypes.THUMBNAIL_DIRTY, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.THUMBNAILS_DELETED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.COMMENTS_DELETED, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.MESSAGE_TO_BROADCAST, this, this.onDirtyEvent);
    this.context.eventCenter.addListener({
      projectID: project.coreObjectID.projectID
    }, Project.EventTypes.PERMALINK_DELETED, this, this.onDirtyEvent);
    this.loadProjectTimeID = null;
  }

  setLicenseController(licenseController) {
    _balsamiq.default.assert(licenseController, _licensecontroller.LicenseController);

    _balsamiq.default.assert(!this.licenseController);

    this.licenseController = licenseController;
    licenseController.init(this.onLicenseChanged.bind(this));
  }

  setBlockedByForceQuit() {
    this.blockedByForceQuit = true;
    this.emptyQueueController.pcEmptyQueue();
  }

  isBlocked() {
    return this.blockedByError || this.blockedByOutOfSync || this.blockedBySlowQueue || this.blockedByRestore || this.blockedByCleanUp || this.blockedByForceQuit || this.blockedByLoadProjectCanceled;
  }

  getArchiveID() {
    return this.archiveID;
  }

  setProjectLocator(projectLocator) {
    _balsamiq.default.assertClass(projectLocator, _locators.ProjectLocator);

    if (this.project) {
      this.project.setProjectLocator(projectLocator);
    }
  }

  setArchiveID(archiveID) {
    _balsamiq.default.assertString(archiveID);

    if (this.archive) {
      this.context.eventCenter.removeListener({
        archiveID: this.archive.coreObjectID.archiveID
      }, Events.EventTypes.ANY, this, this.onArchiveEvent);
    }

    this.archiveID = archiveID;
    this.archive = this.context.archives.getArchive(archiveID);

    _balsamiq.default.assertProtocol(this.archive, Archive.Archive);

    this.lastArchiveRevision = null;
    this.context.eventCenter.addListener({
      archiveID: this.archive.coreObjectID.archiveID
    }, Events.EventTypes.ANY, this, this.onArchiveEvent);
  }

  getProject() {
    return this.project;
  }

  tryCatchCallback(callback, param) {
    if (callback) {
      _balsamiq.default.assertFunction(callback);

      this.context.host.callWithTryCatch(() => callback(param));
    }
  }

  setArchiveError(error) {
    if (this._isCleanedUp) {
      this.context.host.log("Not setting an error on the cleanedUp project controller");
    } else {
      this.project.getArchiveState().setError(error);
    }
  }

  isArchiveSynchronized() {
    return this.getProject().getArchiveState().getKeyValue(ArchiveState.KeyNames.SYNCHRONIZED);
  }

  checkOperationTimeout(shouldCheck) {
    if (!shouldCheck) return;
    this.warningOperationTimeout = this.context.host.timer(_nativeConstants.Timeout.secondsForWarning * 1000, () => {
      this.warningOperationTimeout = null;

      if (this.isBlocked()) {
        return;
      }

      const errorMessage = this.isFileProject() ? "We have some issues saving the changes to your file. We'll keep trying…" : "Your connection is slow or unstable. Attempting to reconnect to save your recent changes…";
      this.setArchiveError({
        type: ArchiveState.ErrorTypes.TRANSIENT,
        messageTextDom: TextUtils.stringToNodes(errorMessage)
      });
    });
  }

  resetOperationTimeout() {
    if (this.warningOperationTimeout) {
      this.warningOperationTimeout = this.context.host.clearTimer(this.warningOperationTimeout);
      const archiveState = this.project.getArchiveState();

      if (archiveState) {
        const error = archiveState.getKeyValue(ArchiveState.KeyNames.ERROR);
        const isPermanentError = error && error.type === ArchiveState.ErrorTypes.PERMANENT;

        if (!isPermanentError) {
          archiveState.setError(null);
        }
      }
    }
  }

  isQueueEmpty() {
    return this.queue.length === 0 && this.runningFunc == null;
  }

  dequeue() {
    _balsamiq.default.assert(this.runningFunc === null);

    if (this.isBlocked()) {
      return;
    }

    if (this.queue.length > 0) {
      this.runningFunc = this.queue[0].func;
      this.checkOperationTimeout(this.queue[0].checkTimeout);
      this.queue = this.queue.slice(1);
      this.runningFunc.apply(this);
    } else {
      this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.EMPTY_QUEUE));
      this.emptyQueueController.pcEmptyQueue();
    }
  }

  enqueue(func) {
    let {
      checkTimeout = true
    } = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    _balsamiq.default.assertFunction(func);

    if (this.queue.length == 0) {
      this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.NOT_EMPTY_QUEUE));
      this.emptyQueueController.pcNotEmptyQueue();
    }

    this.queue.push({
      func,
      checkTimeout
    });

    if (this.queue.length === 1 && this.runningFunc === null) {
      this.dequeue();
    }
  }

  trackEvent(event) {
    if (!this.options.enableChangeTracking) {
      return;
    }

    _balsamiq.default.assertObject(event);

    const getResourceData = (resource, senderCoreObjectID) => {
      const {
        shortResourceID,
        shortBranchID
      } = resource.getShortID();

      if (resource.type === Constants.ResourceTypes.MOCKUP) {
        const branch = this.project.getBranch(senderCoreObjectID);
        const isMaster = branch.isMaster();
        return {
          shortResourceID,
          shortBranchID,
          name: isMaster ? resource.attributes.name : branch.getName(),
          type: isMaster ? resource.type : 'alternate'
        };
      }

      return {
        shortResourceID,
        name: resource.attributes.name,
        type: resource.type
      };
    };

    const {
      senderCoreObjectID
    } = event;

    if (event.eventType === Project.EventTypes.RESOURCE_ADDED) {
      const resource = this.project.getResource(event.eventData.coreObjectID);

      if (!resource) {
        return;
      }

      const edit = {
        id: resource.getShortID().shortResourceID + '-create',
        type: _BmprConstants.UserEventTypes.CreateResource,
        resource: getResourceData(resource, event.eventData.coreObjectID)
      };
      this.changeTracker.register(edit);
    } else if (senderCoreObjectID && senderCoreObjectID.branchID && senderCoreObjectID.resourceID) {
      const resource = this.project.getResource(senderCoreObjectID);

      if (!resource) {
        return;
      }

      const isTrashed = resource.attributes.trashed;
      const type = isTrashed ? _BmprConstants.UserEventTypes.TrashResource : _BmprConstants.UserEventTypes.ChangeResource;
      const edit = {
        id: resource.getShortID().shortResourceID + '-' + type,
        type,
        resource: getResourceData(resource, senderCoreObjectID)
      };
      this.changeTracker.register(edit);
    }
  }

  runningFuncDone(res) {
    let event = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    if (this.isBlocked()) return;

    _balsamiq.default.assert(this.runningFunc !== null);

    this.resetOperationTimeout();
    this.lastOperationCompleteDate = new Date(new Date() - _nativeConstants.Timeout.secondsForError * 1000);

    if (res && res.error) {
      // error!
      this.queue.unshift({
        func: this.runningFunc,
        checkTimeout: false
      });

      if (res.error === SESSION_CLOSED_ERROR) {
        this.runningFunc = null;
        return this._reopenProjectBAS(res => {
          // if we get an error trying to reopen the project, then we don't recover
          if (res && res.error) {
            const errorMessage = this.makeGenericErrorMessage(res);
            this.setArchiveError({
              type: ArchiveState.ErrorTypes.PERMANENT,
              messageTextDom: TextUtils.stringToNodes(errorMessage)
            });
            this.updateArchiveStateShouldReload();
          } else {
            this.dequeue();
          }
        });
      }

      this.blockedByError = true;

      if (res.error === _nativeConstants.Timeout.errorName) {
        this.blockedBySlowQueue = true;
        this.context.host.log(res.error_message, `Timeout: ${_nativeConstants.Timeout.secondsForError * 1000} secs`);
        const lastOperationCompleteTime = `${this.lastOperationCompleteDate.getHours()}:${this.lastOperationCompleteDate.getMinutes()}`;
        const errorMessage = `Unable to reconnect. The last ${_nativeConstants.Timeout.secondsForError} seconds of changes were not saved to the server. Your last successful save was at ${lastOperationCompleteTime}. Please [Reload]`;
        this.setArchiveError({
          type: ArchiveState.ErrorTypes.PERMANENT,
          messageTextDom: TextUtils.stringToNodes(errorMessage)
        });
      } else if (res.error === _nativeConstants.BAS.NotAuthorized) {
        this.blockReload = true;
        this.setArchiveError({
          type: ArchiveState.ErrorTypes.PERMANENT,
          messageTextDom: TextUtils.stringToNodes(`You don't have the necessary permission to work on this project. [Close]`)
        });
      } else {
        const errorMessage = this.makeGenericErrorMessage(res);
        this.setArchiveError({
          type: ArchiveState.ErrorTypes.PERMANENT,
          messageTextDom: TextUtils.stringToNodes(errorMessage)
        });
      }

      this.updateArchiveStateShouldReload();
    } else if (res && res.archiveRevision) {
      if (this.lastArchiveRevision != res.archiveRevision) {
        if (this.userFile) {
          this.userFile.isEdited = true;

          if (this.userFile.synchronizer) {
            this.userFile.synchronizer.onLocalFileChanged(res.archiveRevision);
          } else if (this.userFile.isUnsaved) {
            this.getProject().getArchiveState().setSynchronized(false);
          }
        }
      }

      this.trackEvent(event);
      this.lastArchiveRevision = res.archiveRevision;
      this.context.host.log("runningFuncDone archiveRevision:" + this.lastArchiveRevision);
    }

    this.runningFunc = null;
    this.dequeue();
  }

  downloadURL(url, callback) {
    if (this.archive.fetch) {
      this.archive.fetch(url, data => {
        this.tryCatchCallback(callback, data);
      });
    } else {
      this.context.host.HttpRequest.getBinary(url, {
        base64: true
      }, data => {
        if (!data.error) {
          var file = {
            data: data.data,
            filename: url,
            //deprecated
            url: url,
            type: ""
          };
          this.tryCatchCallback(callback, file);
        } else {
          this.tryCatchCallback(callback, data);
        }
      });
    }
  }

  createProjectBAS(callback) {
    const cfg = this.getBasCfg();

    _balsamiq.default.assertObject(cfg);

    this.enqueue(() => {
      var createDone = res => {
        this.runningFuncDone();
        var dupEntry = res && _balsamiq.default.isString(res.error) && res.error.indexOf("ER_DUP_ENTRY") == 0;

        if (dupEntry) {
          // ignore duplicate project error
          this.tryCatchCallback(callback, {});
        } else if (res.error) {
          this.tryCatchCallback(callback, {
            error: res.error
          });
        } else {
          this.tryCatchCallback(callback, res);
        }
      };

      if (cfg.platformArchiveID) {
        this.archive.createWithPlatformArchiveID(cfg.platformSiteID, cfg.platformArchiveID, cfg.platformArchiveName, cfg.platformInfo, null, res => {
          createDone(res);
        });
      } else {
        this.archive.create(cfg.platformSiteID, cfg.platformArchiveName, cfg.platformInfo, res => {
          createDone(res);
        });
      }
    });
  }

  makeGenericErrorMessage(res) {
    this.context.host.log(`Error: ${res.error}`);

    if (Native.Host.runtime.isBrowser()) {
      return "Oops, something went wrong! Please [reload the project].";
    } else {
      return this.isFileProject() ? `Oops, something went wrong! Project needs to be closed. [Close]` : `Oops, something went wrong! Please reload the project. [Reload]`;
    }
  }

  makeOutOfSyncErrorMessage() {
    return Native.Host.runtime.isBrowser() ? 'Oops. Seems like you missed some changes while you were offline. Please [reload the project] to catch up.' : 'Oops. Seems like you missed some changes while you were offline. Please reload the project to catch up. [Reload]';
  }

  _openProjectBAS(shouldRefreshToken, callback) {
    _balsamiq.default.assertBool(shouldRefreshToken);

    _balsamiq.default.exists(this.getBasCfg());

    _balsamiq.default.assertFunction(this.refreshBasCfg, "refreshBasCfg must be a function");

    _balsamiq.default.assertFunction(callback, "callback must be a function");

    const openArchive = () => {
      const basCfg = this.getBasCfg();

      if (basCfg.platformToken) {
        this.archive.setPlatformToken(basCfg.platformToken, false);
      }

      _balsamiq.default.assertNumber(basCfg.permissions);

      _balsamiq.default.assertClass(basCfg.userInfo, Object);

      if (!basCfg.publicShareID) {
        _balsamiq.default.exists(basCfg.platformSiteID);

        _balsamiq.default.exists(basCfg.platformArchiveID);

        _balsamiq.default.assertString(basCfg.platformArchiveName);

        _balsamiq.default.assertClass(basCfg.platformInfo, Object);

        this.context.host.log("ProjectController: opening BAS archive: platformSiteID=" + basCfg.platformSiteID + " platformArchiveID=" + basCfg.platformArchiveID + " platformArchiveName=" + basCfg.platformArchiveName);
        this.context.execTime.lap(this.loadProjectTimeID, "before archive.open");
        this.archive.open(basCfg.platformSiteID, basCfg.platformArchiveID, basCfg.platformArchiveName, basCfg.platformInfo, basCfg.permissions, {
          branchID: BARConstants.Branch.AllBranches,
          skipThumbnailImage: true
        }, res => {
          this.context.execTime.lap(this.loadProjectTimeID, "after archive.open");
          callback(res);
        });
      } else {
        this.context.host.log("ProjectController: opening BAS archive: publicShareID=" + basCfg.publicShareID);
        this.context.execTime.lap(this.loadProjectTimeID, "before archive.openByPublicShareID");
        this.archive.openByPublicShareID(basCfg.publicShareID, {
          branchID: BARConstants.Branch.AllBranches,
          skipThumbnailImage: true
        }, res => {
          this.context.execTime.lap(this.loadProjectTimeID, "after archive.openByPublicShareID");
          callback(res);
        });
      }
    };

    if (shouldRefreshToken) {
      this.refreshBasCfg(res => {
        if (res.error) {
          callback(res);
        } else {
          openArchive();
        }
      });
    } else {
      openArchive();
    }
  }

  _reopenProjectBAS(callback) {
    _balsamiq.default.assertFunction(callback);

    _balsamiq.default.exists(this.getBasCfg());

    this._openProjectBAS(true, res => {
      if (!res.error) {
        this.checkArchiveRevision(res.dump.Info[BARConstants.Info.ArchiveRevision]);
      }

      callback(res);
    });
  }

  cancelLoadProject() {
    _balsamiq.default.assert(this.blockedByLoadProjectCanceled === false, "cancelLoadProject can only be called once");

    this.blockedByLoadProjectCanceled = true;
    this.emptyQueueController.pcEmptyQueue();
    this.loadProjectCallback({
      canceled: true
    });
  }
  /**
   *
   * @param {Function} callback
   * @param {Function} setProgress
   * @param {boolean} isFirstLoad
   * @param {{ branchID: string; resourceID: string }[]=} selectedResources
   */


  loadProjectBAS(callback, setProgress, isFirstLoad, selectedResources) {
    _balsamiq.default.assertFunction(callback, "callback needs to be a function");

    _balsamiq.default.assertFunction(setProgress, "setProgress needs to be a function");

    _balsamiq.default.assertBool(isFirstLoad, "isFirstLoad needs to be a boolean");

    _balsamiq.default.assertOptionalArray(selectedResources, "selectedResources can be undefined or a list of resources");

    this.loadProjectCallback = callback;
    const PROGRESS_WEIGHTS = new Map([["DOWNLOAD", 0.3], ["LOAD", 0.7]]);
    const progress = new _progress.default(PROGRESS_WEIGHTS, setProgress);
    this.enqueue(() => {
      this.loadProjectTimeID = USE_TIMES ? this.context.execTime.start("Load Project BAS") : null;
      const defaultTimeouts = this.archive.timeouts;
      this.archive.timeouts = {
        response: null,
        deadline: null
      };
      if (this.blockedByLoadProjectCanceled) return;

      this._openProjectBAS(!isFirstLoad, res => {
        if (this.blockedByLoadProjectCanceled) return;

        if (res.error) {
          this.context.execTime.end(this.loadProjectTimeID);
          this.runningFuncDone(res);
          this.tryCatchCallback(callback, {
            error: res.error
          });
        } else {
          this.context.host.callWithTryCatch(() => {
            const cfg = this.getBasCfg();

            if (_balsamiq.default.isNumber(res.timestamp)) {
              this.context.host.log("ProjectController: open timestamp=" + res.timestamp);
            }

            if (cfg.userInfo && cfg.userInfo.name) {
              cfg.userInfo.ID = cfg.userInfo.name;

              if (!cfg.userInfo.displayName || cfg.userInfo.displayName === '') {
                cfg.userInfo.displayName = cfg.userInfo.email;
              }

              this.project.getConnectedUsers().setMyUserInfo(cfg.userInfo);
            }

            this.context.archives.subscribeToArchive(this.archiveID, res.timestamp);
            let assetsPerRequest = DEFAULT_ASSETS_PER_REQUEST_BAS;

            if (_balsamiq.default.isNumber(res.resourcesPerRequest)) {
              this.context.host.log("ProjectController: resourcesPerRequest=" + res.resourcesPerRequest);
              assetsPerRequest = res.resourcesPerRequest;
            }

            const resourceSkimmer = new _resourceSkimmer.default({
              dump: res.dump,
              selectedResources: selectedResources || [],
              loadOnlyImageAssets: true
            });
            this.loadProjectData(res.dump, resourceSkimmer, assetsPerRequest, function setDownloadResourceProgress(title, step, total) {
              progress.step("DOWNLOAD", title, step, total);
            }, dataRes => {
              this.context.execTime.end(this.loadProjectTimeID);
              this.loadProjectTimeID = null;
              if (this.blockedByLoadProjectCanceled) return;
              this.archive.timeouts = defaultTimeouts;

              if (dataRes && dataRes.error) {
                this.context.execTime.end(this.loadProjectTimeID);
                this.runningFuncDone(dataRes);
                this.tryCatchCallback(callback, dataRes);
                return;
              }

              this.doLoad(res.dump, cfg.permissions, resourceSkimmer, loadResult => {
                if (this.blockedByLoadProjectCanceled) return;
                this.runningFuncDone(loadResult);
                this.tryCatchCallback(callback, loadResult);
                this.context.archives.startPingArchive(this.archiveID);
                this.context.archives.startMonitorNetworkStatus(this.archiveID);
                const extract = !!this.options && this.options.filteredProjectMembers;
                const mentionedUserIds = extract ? this.extractMentionedUserIds(res.dump) : null;
                this.updateMentionableUsers({
                  mentionedUserIds
                });
                this.updateAllPermalinks();

                if (this.options.getPngAlreadyExported && this.options.getPngAlreadyExported() != undefined) {
                  this.dirtyWireframesController.addDirtyResources(this.options.getPngAlreadyExported());
                }
              }, function setLoadResourcesProgress(title, step, total) {
                progress.step("LOAD", title, step, total);
              });
            });
          });
        }
      });
    }, {
      checkTimeout: false
    });
  }

  extractMentionedUserIds(dump) {
    const userIds = new Set();

    const addMentionsToSet = text => {
      const ast = makeASTFromText(text);
      const mentions = getMentionedUserIDs(ast);

      for (const mention of mentions) {
        userIds.add(mention);
      }
    };

    for (const branch of dump.Branch) {
      const attributes = branch.ATTRIBUTES;

      if (attributes.projectDescription) {
        addMentionsToSet(attributes.projectDescription);
      }
    }

    for (const resource of dump.Resources) {
      const attributes = resource.ATTRIBUTES;
      let kind = attributes[_BmprConstants.ResourceAttributes.Kind];

      if (kind === _BmprConstants.ResourceKinds.Mockup) {
        addMentionsToSet(attributes.notes);
      }
    }

    for (const comment of dump.Comments) {
      const data = JSON.parse(comment.DATA);

      if (data && data.text) {
        addMentionsToSet(data.text);
      }
    }

    return Array.from(userIds);
  }

  loadProjectSQLite(_ref, callback) {
    let {
      sqliteID,
      userFile,
      userInfo,
      userRole = BARConstants.Role.ROLE_ADMIN,
      setProgress,
      selectedResources = []
    } = _ref;

    _balsamiq.default.assertString(sqliteID);

    _balsamiq.default.assertFunction(callback);

    _balsamiq.default.assertOptionalFunction(setProgress);

    _balsamiq.default.assertEnum(userRole, BARConstants.Role);

    _balsamiq.default.assertArray(selectedResources);

    const PROGRESS_WEIGHTS = new Map([["LOAD_DATA", 0.1], ["LOAD_RESOURCES", 0.9]]);
    const progress = new _progress.default(PROGRESS_WEIGHTS, setProgress);
    this.loadProjectCallback = callback;

    if (userFile) {
      _balsamiq.default.assertString(userFile.userFileURL);

      this.userFile = {
        userFileURL: userFile.userFileURL,
        isEdited: _balsamiq.default.isBoolean(userFile.isEdited) ? userFile.isEdited : null,
        synchronizer: null,
        isUnsaved: _balsamiq.default.isBoolean(userFile.isUnsaved) ? userFile.isUnsaved : null
      };
    }

    this.enqueue(() => {
      if (this.blockedByLoadProjectCanceled) return;
      this.loadProjectTimeID = USE_TIMES ? this.context.execTime.start("Load Project SQLite") : null;
      this.archive.validateBAR(sqliteID, res => {
        this.context.execTime.lap(this.loadProjectTimeID, "archive.validateBAR");

        const exitWithError = error => {
          this.runningFuncDone();
          this.context.execTime.end(this.loadProjectTimeID);
          this.context.host.log(`ProjectController.loadProjectSQLite error: ${res.error}`);
          this.tryCatchCallback(callback, {
            error: error
          });
        };

        if (this.blockedByLoadProjectCanceled) return;

        if (res.error) {
          exitWithError(LoadingErrors.corruptedBMPR);
          return;
        } else if (!_balsamiq.default.checkEnum(res.schemaVersion, BARConstants.SchemaVersions) || res.archiveFormat != "bmpr") {
          exitWithError(LoadingErrors.invalidSchema);
          return;
        } else if (res.schemaVersion === BARConstants.SchemaVersions.OneDotTwo && userRole <= BARConstants.Role.ROLE_VIEWER) {
          exitWithError(LoadingErrors.viewerCantMigrate);
          return;
        }

        if (this.userFile) {
          if (this.userFile.isUnsaved) {
            this.project.archiveState.setSynchronized(!this.userFile.isEdited);
          } else {
            this.userFile.synchronizer = new _userfilesynchronizer.UserFileSynchronizer(this.context, this.context.host.LocalFiles, this.userFile.userFileURL, this._getAutosaveSettings(), isSynchronized => {
              this.project.archiveState.setSynchronized(isSynchronized);
            }, error => {
              this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, Events.EventTypes.SYNC_ERROR, error));
            }, _fileutils.saveUserFile);
            this.userFile.synchronizer.start(this.userFile.isEdited ? res.archiveRevision + 1 : res.archiveRevision, res.archiveRevision);
          }
        }

        this.context.execTime.lap(this.loadProjectTimeID, "before archive.open"); // this.archive.setUserRole(userRole); TODO available from BMPR 3.2.0

        this.archive.open(sqliteID, res => {
          this.context.execTime.lap(this.loadProjectTimeID, "after archive.open");
          if (this.blockedByLoadProjectCanceled) return;

          if (res.error) {
            this.runningFuncDone();
            this.context.host.log("Error opening the archive", res.error);
            this.tryCatchCallback(callback, {
              error: LoadingErrors.openArchive
            });
          } else {
            if (userInfo && userInfo.name) {
              userInfo.ID = userInfo.name;
              userInfo.displayName = userInfo.name !== "" ? userInfo.name : userInfo.email;
              this.project.getConnectedUsers().setMyUserInfo(userInfo);
            }

            const resourceSkimmer = new _resourceSkimmer.default({
              dump: res.dump,
              selectedResources: selectedResources,
              loadOnlyImageAssets: false
            });
            this.loadProjectData(res.dump, resourceSkimmer, ASSETS_PER_REQUEST_SQLITE, function setLoadDataProgress(title, step, total) {
              progress.step("LOAD_DATA", title, step, total);
            }, dataRes => {
              this.context.execTime.end(this.loadProjectTimeID);
              if (this.blockedByLoadProjectCanceled) return;

              if (dataRes && dataRes.error) {
                this.runningFuncDone(dataRes);
                this.tryCatchCallback(callback, dataRes);
                return;
              }

              this.doLoad(res.dump, userRole, resourceSkimmer, loadResult => {
                if (this.blockedByLoadProjectCanceled) return;
                this.runningFuncDone();
                this.tryCatchCallback(callback, loadResult);
              }, function setLoadResourcesProgress(title, step, total) {
                progress.step("LOAD_RESOURCES", title, step, total);
              });
            });
          }
        });
      });
    }, {
      checkTimeout: false
    });
  }

  canReload() {
    return {
      reloadAvailable: this.isFileProject() && !this.isUnsavedFile(),
      reloadEnabled: this.isFileProject() && !this.isUnsavedFile() && !this.isAutosaveEnabled() && !this.isArchiveSynchronized()
    };
  }

  reloadProjectSQLite(sqliteID, callback, setProgress) {
    const PROGRESS_WEIGHTS = new Map([["LOAD_DATA", 0.1], ["LOAD_RESOURCES", 0.9]]);
    const progress = new _progress.default(PROGRESS_WEIGHTS, setProgress);
    this.enqueue(() => {
      this.archive.open(sqliteID, res => {
        if (res.error) {
          this.tryCatchCallback(callback, {
            error: res.error
          });
        } else {
          const resourceSkimmer = new _resourceSkimmer.default({
            dump: res.dump,
            selectedResources: [],
            loadOnlyImageAssets: false
          });
          this.loadProjectData(res.dump, resourceSkimmer, ASSETS_PER_REQUEST_SQLITE, function setLoadDataProgress(title, step, total) {
            progress.step("LOAD_DATA", title, step, total);
          }, dataRes => {
            if (dataRes && dataRes.error) {
              this.runningFuncDone(dataRes);
              this.tryCatchCallback(callback, dataRes);
              return;
            }

            this.doLoad(res.dump, this.project.getUserRole(), resourceSkimmer, loadResult => {
              this.runningFuncDone();
              this.tryCatchCallback(callback, loadResult);
            }, function setLoadResourcesProgress(title, step, total) {
              progress.step("LOAD_RESOURCES", title, step, total);
            });
          });
        }
      });
    }, {
      checkTimeout: false
    });
  }

  loadProjectDump(callback, userInfo) {
    let userRole = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : BARConstants.Role.ROLE_VIEWER;

    _balsamiq.default.assertEnum(userRole, BARConstants.Role);

    this.enqueue(() => {
      this.archive.open(res => {
        if (res.error) {
          this.tryCatchCallback(callback, {
            error: res.error
          });
        } else {
          if (userInfo && userInfo.name) {
            userInfo.ID = userInfo.name;
            userInfo.displayName = userInfo.name !== '' ? userInfo.name : userInfo.email;
            this.project.getConnectedUsers().setMyUserInfo(userInfo);
          }

          const resourceSkimmer = new _resourceSkimmer.default({
            dump: res,
            selectedResources: [],
            loadOnlyImageAssets: false
          });
          this.doLoad(res, userRole, resourceSkimmer, loadResult => {
            this.runningFuncDone();
            this.tryCatchCallback(callback, loadResult);
          });
        }
      });
    }, {
      checkTimeout: false
    });
  }
  /**
   * Fetch resources' data from the archive
   * @param {any} dump TOC
   * @param {ResourceSkimmer} resourceSkimmer
   * @param {number} assetsPerRequest
   * @param {(title: string; step: number; total: number) => void} setProgress
   * @param {(res: unknown) => void} callback
   * @returns 
   */


  loadProjectData(dump, resourceSkimmer, assetsPerRequest, setProgress, callback) {
    _balsamiq.default.assertObject(dump);

    _balsamiq.default.assertClass(resourceSkimmer, _resourceSkimmer.default);

    _balsamiq.default.assertNumber(assetsPerRequest);

    _balsamiq.default.assertFunction(callback);

    this.lastArchiveRevision = dump.Info[BARConstants.Info.ArchiveRevision];
    const resourceIDs = resourceSkimmer.getWireframesAndSLsToFetch();
    let assetResourceIDs = resourceSkimmer.getAllAssets();
    const LOAD_ASSET_DATA_PARALLEL = assetsPerRequest == 0;
    let assetDataToLoadCount = assetResourceIDs.length;
    let loadedAssetDataCount = 0;
    let loadingAssetDataDone = false;

    const checkDoneLoadingAssetData = () => {
      if (loadingAssetDataDone) {
        // In case it is called again
        return true;
      }

      if (loadedAssetDataCount === assetDataToLoadCount) {
        loadingAssetDataDone = true;
        this.context.host.log(`ProjectController: end loading asset data`);
        this.context.execTime.lap(this.loadProjectTimeID, "after loadNextAssetData");
        dump.Resources = dump.Resources.filter(r => r.DATA != null);
        this.tryCatchCallback(callback);
        return true;
      } else {
        return false;
      }
    };

    const loadNextAssetData = () => {
      if (assetResourceIDs.length === assetDataToLoadCount) {
        // first loop, then start the lap
        this.context.host.log(`ProjectController: begin loading asset data: ${assetDataToLoadCount} to load`);
        this.context.execTime.lap(this.loadProjectTimeID, "before loadNextAssetData");
      }

      if (checkDoneLoadingAssetData()) {
        return;
      }

      if (assetResourceIDs.length === 0) {
        return;
      }

      if (LOAD_ASSET_DATA_PARALLEL) {
        const resourceID = assetResourceIDs.shift();
        this.archive.getResourceData(resourceID, BARConstants.Branch.MasterBranchID, res => {
          if (this.isBlocked()) return;

          if (res.error) {
            this.context.host.log(`ProjectController: Error in getResourceData: ${resourceID}, ${res.error}`);
            this.tryCatchCallback(callback, res);
            return;
          }

          resourceSkimmer.setResourceData(resourceID, BARConstants.Branch.MasterBranchID, res.data);
          loadedAssetDataCount += 1;
          setProgress("Downloading Resources...", loadedAssetDataCount, assetDataToLoadCount);
          this.context.host.log(`ProjectController: getResourceData ${loadedAssetDataCount}/${assetDataToLoadCount}: ${resourceID}`);
          checkDoneLoadingAssetData();
        });
        this.context.host.asyncCallback(() => {
          loadNextAssetData();
        });
      } else {
        const ASSETS_PER_REQUESTS = assetsPerRequest;
        const resourceIDs = assetResourceIDs.splice(0, Math.min(ASSETS_PER_REQUESTS, assetResourceIDs.length));
        this.archive.getResourcesData(resourceIDs, BARConstants.Branch.AllBranches, res => {
          if (this.isBlocked()) return;

          if (res.error) {
            this.context.host.log(`ProjectController: Error in getResourcesData, ${res.error}`);
            this.tryCatchCallback(callback, res);
            return;
          }

          loadedAssetDataCount += resourceIDs.length;
          this.context.host.log(`ProjectController: getResourcesData ${loadedAssetDataCount}/${assetDataToLoadCount}`);

          for (const composedID in res.data) {
            const [resourceID, branchID] = composedID.split('|');
            resourceSkimmer.setResourceData(resourceID, branchID, res.data[composedID]);
          }

          if (!checkDoneLoadingAssetData()) {
            loadNextAssetData();
          }
        });
      }
    };

    this.context.execTime.lap(this.loadProjectTimeID, "before getResourcesData"); // Get data of supported resources

    this.archive.getResourcesData(resourceIDs, BARConstants.Branch.AllBranches, res => {
      this.context.execTime.lap(this.loadProjectTimeID, "after getResourcesData");
      if (this.blockedByLoadProjectCanceled) return;

      if (res.error) {
        this.tryCatchCallback(callback, res);
        return;
      }

      this.lastArchiveRevision = res.archiveRevision;

      for (const composedID in res.data) {
        const [resourceID, branchID] = composedID.split('|');

        for (const resource of dump.Resources) {
          if (resource.ID === resourceID && resource.BRANCHID === branchID) {
            resource.DATA = res.data[composedID];
            break;
          }
        }
      }

      const commentIDs = [];

      for (const comment of dump.Comments) {
        commentIDs.push(comment.ID);
      }

      if (commentIDs.length > 0) {
        this.archive.getCommentsData(commentIDs, res => {
          if (this.blockedByLoadProjectCanceled) return;

          if (res.error) {
            this.tryCatchCallback(callback, res);
            return;
          }

          this.lastArchiveRevision = res.archiveRevision;

          for (const comment of res.comments) {
            for (const dumpComment of dump.Comments) {
              if (dumpComment.ID === comment.ID) {
                dumpComment.DATA = comment.DATA;
                break;
              }
            }
          } // remove un-fetched resources since the caller of loadProjectData uses the modified dump to create resources


          assetResourceIDs = resourceSkimmer.removeEmptyWireframesAndExtractUsedAssets();
          assetDataToLoadCount = assetResourceIDs.length;
          loadNextAssetData();
        });
      } else {
        // remove un-fetched resources since the caller of loadProjectData uses the modified dump to create resources
        assetResourceIDs = resourceSkimmer.removeEmptyWireframesAndExtractUsedAssets();
        assetDataToLoadCount = assetResourceIDs.length;
        loadNextAssetData();
      }
    });
  }
  /**
   * Transform data from dump so that it can be loaded by the project
   * @param {import("@balsamiq/bmpr").Dump} dump TOC
   * @param {import("@balsamiq/bmpr/lib/BalsamiqArchiveConstants").Role} userRole
   * @param {import("core/utils/resource-skimmer")} resourceSkimmer
   * @param {(res: unknown) => void} callback
   * @param {((title: string; step: number; total: number) => void)=} setProgress
   * @returns 
   */


  doLoad(dump, userRole, resourceSkimmer, callback, setProgress) {
    const branches = [];
    dump.Branch.forEach(branch => {
      branches.push({
        branchID: branch.branchID,
        attributes: branch.ATTRIBUTES
      });
    });
    const resources = [];
    dump.Resources.forEach(resource => {
      resources.push({
        resourceID: resource.ID,
        branchID: resource.BRANCHID,
        attributes: resource.ATTRIBUTES,
        data: resource.DATA
      });
    });
    const thumbnails = [];
    dump.Thumbnails.forEach(thumbnail => {
      thumbnails.push({
        thumbnailID: thumbnail.ID,
        attributes: thumbnail.ATTRIBUTES
      });
    });
    const users = [];
    dump.Users.forEach(user => {
      users.push({
        userID: user.ID,
        attributes: user.ATTRIBUTES
      });
    });
    const comments = [];
    dump.Comments.forEach(comment => {
      comments.push({
        commentID: comment.ID,
        branchID: comment.BRANCHID,
        resourceID: comment.RESOURCEID,
        userID: comment.USERID,
        attributes: comment.ATTRIBUTES,
        data: comment.DATA
      });
    });
    return this.project.load({
      attributes: dump.Info.ArchiveAttributes,
      branches,
      resources,
      thumbnails,
      userRole,
      users,
      comments,
      startAsReviewer: this.options.startAsReviewer
    }, resourceSkimmer, callback, setProgress, () => this.blockedByLoadProjectCanceled);
  }

  downloadBMPR(callback, progressCallback) {
    _balsamiq.default.assertFunction(callback);

    this.archive.download(res => {
      if (res && res.error) {
        callback(res);
      } else if (res) {
        if (progressCallback) {
          progressCallback(true, 100, 100);
        }

        this.context.host.saveFile(this.project.getName() + ".bmpr", res, MimeTypes.BMPR_MIMETYPE, false, res => {
          callback(res);
        });
      }
    }, progressCallback);
  }

  isPermalinkEnabled() {
    return this.options.permalinkEnabled && this.project.getUserRole() >= Project.Roles.EDITOR;
  }

  isFileProject() {
    return this.archive.isLocal();
  }

  isUnsavedFile() {
    return this.userFile && this.userFile.isUnsaved;
  }

  isUnsavedAndNewFile() {
    return this.isUnsavedFile() && this.isArchiveSynchronized();
  }

  getFileURL() {
    return this.userFile ? this.userFile.userFileURL : null;
  }

  cancelDownloadBMPR() {
    this.archive.cancelDownload();
  }

  getResourceData(branchID, resourceID, callback) {
    this.enqueue(() => {
      this.archive.getResourceData(resourceID, branchID, res => {
        this.runningFuncDone(res);

        if (res) {
          this.tryCatchCallback(callback, res.data);
        }
      });
    });
  }
  /**
   *
   * @param {Object} userEvent
   * @param {((res: unknown) => void)=} callback
   */


  logUserEvent(userEvent, callback) {
    _balsamiq.default.assertObject(userEvent);

    _balsamiq.default.assertEnum(userEvent.type, _BmprConstants.UserEventTypes);

    this.enqueue(() => {
      this.archive.logUserEvent(userEvent, res => {
        if (res.error) {
          this.context.host.log("Error: BalsamiqArchiveClient.logUserEvent ", res.error);
          this.runningFuncDone(res);
          this.tryCatchCallback(callback, res);
        } else {
          this.runningFuncDone();
          this.tryCatchCallback(callback, res.data);
        }
      });
    });
  }

  updateArchiveStateShouldReload() {
    // the user is proposed to reload a project when all the following cases are true:
    // - the archive is online
    // - (for the moment) the archive is on the BAS
    // - the controller is blocked by any condition
    const isArchiveOnline = this.context.archives.getArchiveOnline(this.getArchiveID());
    const canReload = this.getBasCfg() != null;
    const shouldReload = canReload && isArchiveOnline && this.isBlocked();

    if (shouldReload) {
      this._closeArchive(() => {
        if (!this._isCleanedUp) {
          this.project.getArchiveState().setShouldReload(!this.blockReload);
        }
      });
    }
  }

  setReloadProjectState(errorMessage) {
    this.blockedByError = true;
    this.setArchiveError({
      type: ArchiveState.ErrorTypes.PERMANENT,
      messageTextDom: TextUtils.stringToNodes(errorMessage)
    });
    this.updateArchiveStateShouldReload();
  }

  checkArchiveRevision(archiveRevision) {
    if (archiveRevision != null && archiveRevision === this.lastArchiveRevision) {
      if (this.project.getArchiveState() != null) {
        this.project.getArchiveState().setAvailable(true);
      }
    } else {
      this.blockedByOutOfSync = true;
      this.context.host.log(`ProjectController: expected archiveRevision (${this.lastArchiveRevision}), but received (${archiveRevision})`);
      const errorMessage = this.makeOutOfSyncErrorMessage();
      this.setArchiveError({
        type: ArchiveState.ErrorTypes.PERMANENT,
        messageTextDom: TextUtils.stringToNodes(errorMessage)
      });
    }

    this.updateArchiveStateShouldReload();
  }

  onArchiveEvent(event) {
    var _this = this;

    const isArchiveOffline = !this.project.getArchiveState().getKeyValue(ArchiveState.KeyNames.AVAILABLE); // when the archive is offline or blocked by any reason, we must accept only the following events

    if (this.isBlocked() || isArchiveOffline) {
      switch (event.eventType) {
        case Archive.EventTypes.ARCHIVE_ONLINE:
        case Archive.EventTypes.ARCHIVE_OFFLINE:
        case Archive.EventTypes.ONLINE_USERS_LIST_CHANGED:
        case Archive.EventTypes.BROADCAST_MESSAGE:
        case Archive.EventTypes.ARCHIVE_RESTORED:
        case Archive.EventTypes.ARCHIVE_FLUSHED:
          break;

        default:
          return;
      }
    }

    if (event.eventData.archiveRevision) {
      this.lastArchiveRevision = event.eventData.archiveRevision;
      this.context.host.log("event:" + event.eventType + " archiveRevision:" + this.lastArchiveRevision);
    }

    switch (event.eventType) {
      case Archive.EventTypes.ARCHIVE_OFFLINE:
        this.project.archiveState.setAvailable(false);
        this.updateArchiveStateShouldReload();
        break;

      case Archive.EventTypes.ARCHIVE_ONLINE:
        this.archive.getArchiveRevision(function () {
          let res = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

          if (res.error === SESSION_CLOSED_ERROR) {
            _this.enqueue(() => {
              _this._reopenProjectBAS(_this.runningFuncDone.bind(_this));
            }, {
              checkTimeout: false
            });
          } else if (res.error) {
            const errorMessage = _this.makeGenericErrorMessage(res);

            _this.blockedByError = true;

            _this.setArchiveError({
              type: ArchiveState.ErrorTypes.PERMANENT,
              messageTextDom: TextUtils.stringToNodes(errorMessage)
            });
          } else {
            _this.checkArchiveRevision(res.archiveRevision);

            _this.updateAllPermalinks();
          }
        });
        break;

      case Archive.EventTypes.SERVER_TIMEDIFF:
        this.handleServerTimeDiff(event.eventData.serverTimeDiff);
        break;

      case Archive.EventTypes.ARCHIVE_RESTORING:
        this.blockedByRestore = true;

        this._closeArchive(() => {});

        this.setArchiveError({
          info: {
            userName: event.eventData.userName
          },
          type: ArchiveState.ErrorTypes.PERMANENT,
          code: Archive.EventTypes.ARCHIVE_RESTORING
        });
        break;

      case Archive.EventTypes.ARCHIVE_RESTORED:
        this.blockedByOutOfSync = true;
        this.setArchiveError({
          type: ArchiveState.ErrorTypes.PERMANENT,
          code: Archive.EventTypes.ARCHIVE_RESTORED
        });
        break;

      case Archive.EventTypes.ARCHIVE_FLUSHED:
        this.changeTracker && this.changeTracker.reset();
        break;

      case Archive.EventTypes.RESOURCES_DELETED:
      case Archive.EventTypes.BRANCH_ADDED:
      case Archive.EventTypes.RESOURCE_ATTRIBUTES_CHANGED:
      case Archive.EventTypes.BRANCH_ATTRIBUTES_CHANGED:
      case Archive.EventTypes.THUMBNAILS_DELETED:
      case Archive.EventTypes.BRANCHES_DELETED:
      case Archive.EventTypes.ARCHIVE_ATTRIBUTES_CHANGED:
        this.scheduleArchiveOperation(event);
        break;

      case Archive.EventTypes.RESOURCE_DATA_CHANGED:
        this.enqueue(() => {
          this.archive.getResourceData(event.eventData.resourceID, event.eventData.branchID, res => {
            let undefinedDataError = !res.error && !_balsamiq.default.isString(res.data);

            if (undefinedDataError) {
              this.context.host.log("ProjectController: undefined data error on resourceID:" + event.eventData.resourceID + " branchID:" + event.eventData.branchID);

              _balsamiq.default.assert(false, "BalsamiqArchive.getResourceData returned non-String data");
            }

            if (res.error || undefinedDataError) {
              this.runningFuncDone(res);
            } else {
              this.scheduleArchiveOperation(event, {
                data: res.data
              });
              this.runningFuncDone(res);
            }
          });
        });
        break;

      case Archive.EventTypes.RESOURCE_BRANCH_CHANGED:
        this.scheduleArchiveOperation(event);
        break;

      case Archive.EventTypes.RESOURCE_ADDED:
        // in this case we must fake a second event with type RESOURCE_DATA_CHANGED because
        // the resource must be created as soon as possible
        this.scheduleArchiveOperation(event);
        this.enqueue(() => {
          this.archive.getResourceData(event.eventData.resourceID, event.eventData.branchID, res => {
            if (res.data) {
              this.scheduleArchiveOperation({
                eventType: Archive.EventTypes.RESOURCE_DATA_CHANGED,
                eventData: event.eventData
              }, {
                data: res.data
              });
            }

            this.runningFuncDone(res);
          });
        });
        break;

      case Archive.EventTypes.THUMBNAIL_ADDED:
        // TODO create the thumbnail and THEN load the attributes (like RESOURCE_ADDED)
        this.enqueue(() => {
          this.archive.getThumbnail(event.eventData.thumbnailID, res => {
            if (res.error) {
              this._closeArchive(() => {});

              this.blockedByOutOfSync = true;
              const errorMessage = this.makeOutOfSyncErrorMessage();
              this.setArchiveError({
                type: ArchiveState.ErrorTypes.PERMANENT,
                messageTextDom: TextUtils.stringToNodes(errorMessage)
              });
              this.context.host.collectDevData("THUMBNAIL_ADDED", "An error occurred getting a thumbnail as a consequence of a RTC event", null, res.error);
            } else {
              this.scheduleArchiveOperation(event, {
                attributes: res.ATTRIBUTES
              });
            }

            this.runningFuncDone(res);
          });
        });
        break;

      case Archive.EventTypes.ONLINE_USERS_LIST_CHANGED:
        {
          this.project.getConnectedUsers().onUsersListChanged(event.eventData.usersList);
          break;
        }

      case Archive.EventTypes.COMMENT_CREATED:
        this.enqueue(() => {
          this.archive.getCommentsData([event.eventData.commentData.commentID], res => {
            if (res.comments && res.comments.length && res.comments[0].ID === event.eventData.commentData.commentID) {
              var commentData;

              try {
                commentData = JSON.parse(res.comments[0].DATA);
              } catch (syntaxErrorException) {
                commentData = {};
                this.runningFuncDone({
                  error: "Syntax error parsing last comment's JSON"
                });
              }

              this.scheduleArchiveOperation(event, {
                commentID: event.eventData.commentData.commentID,
                parentID: event.eventData.commentData.commentParentID,
                branchID: event.eventData.commentData.branchID,
                resourceID: event.eventData.commentData.resourceID,
                userID: res.comments[0].USERID,
                attributes: res.comments[0].ATTRIBUTES,
                data: commentData
              });
              this.runningFuncDone(res);
            } else {
              this.runningFuncDone(res);
            }
          });
        });
        break;

      case Archive.EventTypes.COMMENTS_ATTRIBUTES_UPDATED:
        this.scheduleArchiveOperation(event, {
          attributes: event.eventData.attributes
        });
        break;

      case Archive.EventTypes.COMMENTS_DELETED:
        this.scheduleArchiveOperation(event, {
          commentIDs: event.eventData.commentIDs
        });
        break;

      case Archive.EventTypes.COMMENT_DATA_UPDATED:
        {
          let commentData;

          try {
            commentData = JSON.parse(event.eventData.data);
          } catch (syntaxErrorException) {
            commentData = {}; // TODO: error !!
          }

          this.scheduleArchiveOperation(event, {
            commentID: event.eventData.commentID,
            data: commentData,
            attributes: event.eventData.attributes
          });
          break;
        }

      case Archive.EventTypes.PERMALINK_CREATED_OR_UPDATED:
        {
          const {
            resourceID,
            branchID
          } = event.eventData;
          this.getPermalinkData(resourceID, branchID, permalink => {
            if (permalink !== null) {
              this.scheduleArchiveOperation(event, permalink);
            }
          });
          break;
        }

      case Archive.EventTypes.PERMALINK_DELETED:
        {
          this.scheduleArchiveOperation(event, event.eventData);
          break;
        }

      case Archive.EventTypes.ARCHIVE_USERS_CHANGED:
        {
          this.enqueue(() => {
            this.archive.getArchiveUsersList(res => {
              this.runningFuncDone(res);

              if (res.error) {
                this.context.host.log("Error: BalsamiqArchiveClient.getArchiveUsersList ", res.error);
              } else {
                _balsamiq.default.assertArray(res.users);

                var users = res.users.map(user => {
                  return {
                    userID: user.ID,
                    attributes: user.ATTRIBUTES
                  };
                });
                this.project.updateUsers(users);
              }
            });
          });
          break;
        }

      case Archive.EventTypes.BROADCAST_MESSAGE:
        this.scheduleArchiveOperation(event, {
          userName: event.eventData.userName,
          message: event.eventData.message,
          timestamp: event.eventData.timestamp
        });
        break;

      case Archive.EventTypes.THUMBNAIL_UPDATED:
        //nothing to do here, in this case we just need to update the lastArchiveRevision
        break;

      case Archive.EventTypes.USER_ROLE_CHANGED:
        _balsamiq.default.assertEnum(event.eventData.role, Constants.UserRoles);

        this.project.setUserRole(event.eventData.role);
        this.updateAllPermalinks();
        break;

      default:
        _balsamiq.default.assert(false, "Unknown Archive Event type: " + event.eventType);

    }
  }

  simulateResourceDataChanged(resourceID, branchID, data) {
    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.assertString(branchID);

    _balsamiq.default.assertString(data);

    this.scheduleArchiveOperation(Events.newCoreEvent(this.archive.coreObjectID, Archive.EventTypes.RESOURCE_DATA_CHANGED, {
      resourceID,
      branchID
    }), {
      data
    });
  }

  onArchiveOperationFinished(res) {
    if (this.isBlocked()) {
      return;
    }

    if (res && res.error) {
      this.blockedByError = true;
      this.setArchiveError({
        type: ArchiveState.ErrorTypes.PERMANENT,
        messageTextDom: TextUtils.stringToNodes(`Error loading the project.`),
        code: res.error
      });
    }
  }

  handleServerTimeDiff(serverTimeDiff) {
    if (this.serverTimeDiff === null) {
      this.serverTimeDiff = serverTimeDiff;
      this.project.setServerTimeDiff(this.serverTimeDiff);
    } else if (this.serverTimeDiff > serverTimeDiff) {
      this.serverTimeDiff = serverTimeDiff;
      this.project.setServerTimeDiff(this.serverTimeDiff);
    }
  }

  forceUpdateArchive() {
    this.project.forceSaveAll();
  }

  scheduleArchiveOperation(event, params) {
    var operation = new _archiveeventoperation.ArchiveEventOperation({
      projectCoreObjectID: this.project.coreObjectID,
      event: event,
      ...params
    });
    this.context.operationQueue.scheduleOperation(operation, res => {
      this.onArchiveOperationFinished(res);
    }, null, _operation.Direction.Do);
  }

  onDirtyEvent(event) {
    switch (event.eventType) {
      case Resource.EventTypes.RESOURCE_ATTRIBUTES_DIRTY:
        {
          let ops = this.context.operationQueue.filterOperations(operation => {
            if (operation instanceof _archiveeventoperation.ArchiveEventOperation) {
              if (operation.params.event.eventType == Archive.EventTypes.RESOURCE_ATTRIBUTES_DIRTY && operation.params.event.eventData.resourceID == event.senderCoreObjectID.resourceID && operation.params.event.eventData.branchID == event.senderCoreObjectID.branchID) {
                return true;
              }
            }

            return false;
          });

          if (ops.length > 0) {
            this.context.host.log("ProjectController: ignored RESOURCE_ATTRIBUTES_DIRTY event of resID:" + event.senderCoreObjectID.resourceID + " branchID:" + event.senderCoreObjectID.branchID);
          } else {
            this.enqueue(() => {
              this.archive.setResourceAttributes(event.senderCoreObjectID.resourceID, event.senderCoreObjectID.branchID, event.eventData.attributes, res => {
                this.runningFuncDone(res, event);
              });
            });
          }
        }
        break;

      case Branch.EventTypes.BRANCH_ATTRIBUTES_DIRTY:
        this.enqueue(() => {
          this.archive.setBranchAttributes(event.senderCoreObjectID.branchID, event.eventData.attributes, res => {
            this.runningFuncDone(res, event);
          });
        });
        break;

      case Thumbnail.EventTypes.THUMBNAIL_DIRTY:
        {
          const {
            resourceID,
            branchID
          } = _balsamiq.default.get(event, "eventData.attributes", {});

          this.context.host.log(`thumbnail dirty: ${event.senderCoreObjectID.thumbnailID}, resourceID=${resourceID} branchID=${branchID}`);
          this.enqueue(() => {
            this.archive.setThumbnail(event.senderCoreObjectID.thumbnailID, event.eventData.attributes, res => {
              this.runningFuncDone(res, event);
            });
          });
          break;
        }

      case Project.EventTypes.PROJECT_ATTRIBUTES_DIRTY:
        this.enqueue(() => {
          this.archive.setArchiveAttributes(event.eventData, res => {
            this.runningFuncDone(res, event);
          });
        });
        break;

      case Project.EventTypes.RESOURCE_ADDED:
        {
          let resource = this.project.getResource(event.eventData.coreObjectID);
          this.enqueue(() => {
            this.archive.createResource(event.eventData.coreObjectID.resourceID, event.eventData.coreObjectID.branchID, resource.getAttributes(), event.eventData.data, res => {
              this.runningFuncDone(res, event);
            });
          });
        }
        break;

      case Project.EventTypes.COMMENT_ADDED:
        {
          this.enqueue(() => {
            this.archive.createComment(event.eventData.coreObjectID.commentID, event.eventData.coreObjectID.resourceID, event.eventData.coreObjectID.branchID, event.eventData.parentID, JSON.stringify(event.eventData.data), res => {
              if (res.attributes) {
                if (!_balsamiq.default.isEmpty(event.eventData.attributes)) {
                  this.archive.updateCommentsAttributes([res.commentID], [event.eventData.attributes], updateRes => {
                    if (updateRes.attributes) {
                      this.project.updateCommentsAttributes(updateRes.attributes);
                    }

                    this.runningFuncDone(updateRes, event);
                  });
                } else {
                  this.project.updateCommentsAttributes([{
                    commentID: res.commentID,
                    attributes: res.attributes
                  }]);
                  this.runningFuncDone(res, event);
                }
              } else {
                this.runningFuncDone(res, event);
              }
            });
          });
        }
        break;

      case Project.EventTypes.COMMENT_IMPORTED:
        {
          this.enqueue(() => {
            this.archive.importComment(event.eventData.coreObjectID.resourceID, event.eventData.coreObjectID.branchID, event.eventData.comment, res => {
              this.project.updateCommentsAttributes([{
                commentID: res.commentID,
                attributes: res.attributes
              }]);
              this.runningFuncDone(res, event);
            });
          });
          break;
        }

      case Project.EventTypes.MYUSER_CREATED:
        {
          this.enqueue(() => {
            this.archive.createMyUser(event.eventData, res => {
              this.runningFuncDone(res);
            });
          });
        }
        break;

      case Project.EventTypes.MYUSER_UPDATED:
        {
          this.enqueue(() => {
            this.archive.updateMyUser(event.eventData, res => {
              this.runningFuncDone(res);
            });
          });
        }
        break;

      case Project.EventTypes.COMMENTS_ATTRIBUTES_UPDATED:
        {
          this.enqueue(() => {
            this.archive.updateCommentsAttributes(event.eventData.commentIDs, event.eventData.attributes, res => {
              if (res.attributes) {
                this.project.updateCommentsAttributes(res.attributes);
              }

              this.runningFuncDone(res, event);
            });
          });
        }
        break;

      case Project.EventTypes.COMMENT_DATA_UPDATED:
        {
          this.enqueue(() => {
            this.archive.setCommentData(event.eventData.commentID, event.eventData.data, res => {
              if (res && res.data) {
                this.project.updateComment(event.eventData.commentID, res.attributes, JSON.parse(res.data));
              }

              this.runningFuncDone(res, event);
            });
          });
        }
        break;

      case Project.EventTypes.RESOURCES_DELETED:
        {
          this.enqueue(() => {
            this.archive.deleteResources(event.eventData.resourceIDs, event.eventData.branchID, res => {
              if (!res.error) {
                if (this.isFileProject()) {
                  this.archive.compact(res => {
                    this.runningFuncDone(res, event);
                  });
                } else {
                  this.runningFuncDone(res, event);
                }
              } else {
                this.runningFuncDone(res, event);
              }
            });
          }, {
            checkTimeout: false
          });
        }
        break;

      case Project.EventTypes.BRANCH_ADDED:
        {
          let branch = this.project.getBranch(event.eventData.coreObjectID);
          this.enqueue(() => {
            this.archive.createBranch(event.eventData.coreObjectID.branchID, branch.getAttributes(), res => {
              this.runningFuncDone(res, event);
            });
          });
        }
        break;

      case Project.EventTypes.THUMBNAIL_ADDED:
        {
          this.enqueue(() => {
            this.archive.createThumbnail(event.eventData.thumbnailID, event.eventData.attributes, res => {
              this.runningFuncDone(res, event);
            });
          });
        }
        break;

      case Project.EventTypes.THUMBNAILS_DELETED:
        {
          this.enqueue(() => {
            this.archive.deleteThumbnails(event.eventData.thumbnailIDs, res => {
              this.runningFuncDone(res, event);
            });
          });
        }
        break;

      case Resource.EventTypes.RESOURCE_DATA_DIRTY:
        {
          let ops = this.context.operationQueue.filterOperations(operation => {
            if (operation instanceof _archiveeventoperation.ArchiveEventOperation) {
              if (operation.params.event.eventType == Archive.EventTypes.RESOURCE_DATA_CHANGED && operation.params.event.eventData.resourceID == event.senderCoreObjectID.resourceID && operation.params.event.eventData.branchID == event.senderCoreObjectID.branchID) {
                return true;
              }
            }

            return false;
          });

          if (ops.length > 0) {
            this.context.host.log("ProjectController: ignored RESOURCE_DATA_DIRTY event of resID:" + event.senderCoreObjectID.resourceID + " branchID:" + event.senderCoreObjectID.branchID);
          } else {
            this.enqueue(() => {
              this.archive.setResourceData(event.senderCoreObjectID.resourceID, event.senderCoreObjectID.branchID, event.eventData.data, res => {
                this.runningFuncDone(res, event);
              });
            });
          }
        }
        break;

      case Project.EventTypes.RESOURCE_BRANCHID_CHANGED:
        {
          this.enqueue(() => {
            this.archive.setResourceBranchID(event.eventData.resourceID, event.eventData.oldBranchID, event.eventData.newBranchID, res => {
              this.runningFuncDone(res, event);
            });
          });
        }
        break;

      case Project.EventTypes.BRANCHES_DELETED:
        {
          this.enqueue(() => {
            this.archive.deleteBranches(event.eventData.branchIDs, res => {
              this.runningFuncDone(res, event);
            });
          });
        }
        break;

      case Project.EventTypes.MESSAGE_TO_BROADCAST:
        {
          if (this.archive.broadcastMessage) {
            this.enqueue(() => {
              this.archive.broadcastMessage(event.eventData, res => {
                if (!res.error) {
                  this.project.updateChatMessage(res);
                } else {
                  this.project.notifyMessageError(event.eventData.id, res.error);
                }

                this.runningFuncDone(res);
              });
            });
          }
        }
        break;

      case Project.EventTypes.COMMENTS_DELETED:
        {
          this.enqueue(() => {
            this.archive.deleteComments(event.eventData.commentIDs, res => {
              this.runningFuncDone(res, event);
            });
          });
        }
        break;

      case Project.EventTypes.PERMALINK_DELETED:
        {
          const {
            resourceID,
            branchID
          } = event.eventData;
          this.enqueue(() => {
            this.archive.deletePermalinkData(resourceID, branchID, res => {
              this.runningFuncDone(res, event);
            });
          });
        }
        break;
    }
  }

  hasComments() {
    return this.project && Object.keys(this.project.comments).length > 0;
  }

  canTakeSnapshot() {
    const basCfg = this.getBasCfg();
    return basCfg && basCfg.canTakeSnapshot;
  }

  canDownload() {
    const basCfg = this.getBasCfg();
    return !!basCfg && basCfg.canDownload !== false;
  }

  flushArchive(callback) {
    this.enqueue(() => {
      this.archive.flushExt({}, res => {
        this.runningFuncDone(res);
        callback(res);
      });
    });
  }

  createOrUpdateImageLink(_ref2) {
    let {
      resourceID,
      branchID,
      permalinkInfo,
      base64data,
      mimeType,
      callback
    } = _ref2;
    this.enqueue(() => {
      this.archive.createOrUpdateImageLink(resourceID, branchID, permalinkInfo, base64data, mimeType, res => {
        this.runningFuncDone(res);

        if (res && !res.error) {
          this.project.onPermalinkCreatedOrUpdated(res);
        }

        callback(res);
      });
    });
  }

  getPermalinkData(resourceID, branchID, callback) {
    this.enqueue(() => {
      this.archive.getPermalinkData(resourceID, branchID, res => {
        this.runningFuncDone(res);

        if (res && !res.error) {
          callback(res);
        } else {
          callback(null);
        }
      });
    });
  }

  updateAllPermalinks(callback) {
    if (!this.isPermalinkEnabled()) {
      callback && callback();
      return;
    }

    this.enqueue(() => {
      this.archive.getPermalinksData(res => {
        if (res && !res.error) {
          const {
            permalinksData
          } = res;
          this.project.onAllPermalinksUpdated(permalinksData);
        }

        this.runningFuncDone(res);
        callback && callback(res);
      });
    });
  }

  updateMentionableUsers() {
    let params = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    const {
      searchString,
      mentionedUserIds
    } = params;

    if (this.isFileProject() || !this.options.enableNotifications) {
      return;
    }

    const data = {
      projectId: this.archive.archiveID,
      searchString,
      mentionedUserIds
    };
    this.archive.getProjectMembers(data, res => {
      if (res.error) {
        this.context.host.log("Error: BalsamiqArchiveClient.getMentionableUsers ", res.error);
      } else {
        const users = res.users || [];
        const replace = !mentionedUserIds && !searchString;
        this.project.updateMentionableUsers(users, replace);
      }
    });
  }

  uploadTempFile(data, callback) {
    this.archive.uploadTempFile(data, res => {
      callback(res);
    });
  }

  downloadTempFile(key, callback) {
    this.archive.downloadTempFile(key, res => {
      callback(res);
    });
  }

  downloadBmprLink(callback) {
    this.archive.downloadBmprlLink(this.project.getName() + ".bmpr", res => {
      callback(res);
    });
  }

  flushAndCloseArchive(waitForClose, callback) {
    _balsamiq.default.assertBool(waitForClose);

    if (this.isBlocked()) {
      callback && callback({
        error: "Archive error, cannot flush"
      });
      return;
    }

    let f = () => {
      // we call runningFuncDone here because archive.close will be the last function called
      this.runningFuncDone();

      let closeCallback = () => {
        callback && callback();
      };

      let skip = () => {};

      let closeAsyncCallback, closeSyncCallback;

      if (waitForClose) {
        closeAsyncCallback = closeCallback;
        closeSyncCallback = skip;
      } else {
        closeAsyncCallback = skip;
        closeSyncCallback = closeCallback;
      }

      this._closeArchive(closeAsyncCallback);

      closeSyncCallback();
    };

    this.enqueue(() => {
      if (this.userFile && this.userFile.synchronizer) {
        this.userFile.synchronizer.stop(f);
      } else {
        f();
      }
    }, {
      checkTimeout: false
    });
  }

  cleanUp() {
    _balsamiq.default.assert(this.isBlocked() || !this.isBlocked() && this.queue.length == 0, "ProjectController.cleanUp called when the queue is not empty");

    _balsamiq.default.assert(this.queueTimer == null, "ProjectController.cleanUp called when the timerQueue is not null");

    if (this.licenseController) {
      this.licenseController.cleanUp();
    }

    this.blockedByCleanUp = true;
    this.resetOperationTimeout();
    this.dirtyWireframesController.cleanUp();
    this.dirtyWireframesController = null;
    this.emptyQueueController.cleanUp();
    this.emptyQueueController = null;

    if (this.userFile) {
      if (this.userFile.synchronizer) {
        this.userFile.synchronizer.cleanUp();
      }

      this.userFile = null;
    }

    super.cleanUp();
  }

  _closeArchive(callback) {
    _balsamiq.default.assertFunction(callback);

    this.context.host.log("ProjectController: close archive");

    if (this.archive instanceof _BalsamiqArchive.BalsamiqArchive) {
      this.archive.close(this.archive.archiveID, callback);
    } else {
      this.archive.close(callback);
    }
  }

  isAutosaveEnabled() {
    if (this.userFile) {
      if (this.userFile.synchronizer) {
        return this.userFile.synchronizer.getAutomaticSynchronization();
      } else {
        return false;
      }
    } else {
      return true;
    }
  }

  canToggleAutosave() {
    if (this.userFile) {
      if (this.userFile.synchronizer) {
        return true;
      } else {
        return false;
      }
    } else {
      return false;
    }
  }

  toggleAutosave() {
    _balsamiq.default.assert(this.canToggleAutosave(), "ProjectController: Autosave can't be toggled");

    this._setAutosave(!this.userFile.synchronizer.getAutomaticSynchronization());

    this.context.settings.setData(AUTOSAVE_SETTINGS_KEY, this.userFile.synchronizer.getAutomaticSynchronization());
  }

  save(callback) {
    _balsamiq.default.exists(this.userFile);

    if (this.userFile.synchronizer) {
      if (this.userFile.synchronizer.getAutomaticSynchronization()) {
        this.userFile.synchronizer.forceSynchronize(callback);
      } else {
        this.userFile.synchronizer.manualSynchronize();
      }
    }
  }

  onSettingsChanged() {
    this._setAutosave(this._getAutosaveSettings());
  }

  onLicenseChanged(data, role) {
    if (role != null) {
      // Set role to project
      _balsamiq.default.assertEnum(role, Constants.UserRoles);

      this.project.setUserRole(role);
    }

    this.project.setLicenseData(data);
  }

  _getAutosaveSettings() {
    let data = this.context.settings.getData(AUTOSAVE_SETTINGS_KEY);
    let autosave = _balsamiq.default.isBoolean(data) ? data : true;
    return autosave;
  }

  _setAutosave(value) {
    _balsamiq.default.assertBool(value);

    if (this.userFile && this.userFile.synchronizer) {
      if (value != this.userFile.synchronizer.getAutomaticSynchronization()) {
        this.userFile.synchronizer.setAutomaticSynchronization(value);
        this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.AUTOSAVE_CHANGED));
      }
    }
  }

  compact() {
    this.enqueue(() => {
      if (this.isFileProject()) {
        this.archive.compact(res => {
          this.runningFuncDone(res);
        });
      } else {
        this.runningFuncDone({});
      }
    }, {
      checkTimeout: false
    });
  }

  getAPIUserAuthToken(callback) {
    this.enqueue(() => {
      this.archive.getUserAuthToken(res => {
        this.runningFuncDone(res);
        callback(res);
      });
    });
  }

}

exports.ProjectController = ProjectController;
var _default = {
  ProjectController,
  EventTypes,
  LoadingErrors
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controllers/userfilesynchronizer.js":
/*!*****************************************************!*\
  !*** ./js/core/controllers/userfilesynchronizer.js ***!
  \*****************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.UserFileSynchronizer = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const AUTOMATIC_SYNCHRONIZATION_DELAY = 5000;

class UserFileSynchronizer {
  constructor(context, localFiles, userFileURL, automaticSynchronization, onIsSynchronizedCallback, errorCallback, saveUserFile) {
    _balsamiq.default.exists(context.host);

    _balsamiq.default.exists(localFiles);

    _balsamiq.default.assertString(userFileURL);

    _balsamiq.default.assertBool(automaticSynchronization);

    _balsamiq.default.assertFunction(onIsSynchronizedCallback);

    _balsamiq.default.assertFunction(errorCallback);

    _balsamiq.default.assertFunction(saveUserFile);

    this.context = context;
    this.userFileURL = userFileURL;
    this.setAutomaticSynchronization(automaticSynchronization);
    this.synchronizedArchiveRevision = null;
    this.onIsSynchronizedCallback = onIsSynchronizedCallback;
    this.errorCallback = errorCallback;
    this.archiveRevision = null;
    this.synchronizingArchiveRevision = null;
    this.timerID = null;
    this.needsAnotherSynchronization = false;
    this.started = false;
    this.stopCallback = null;
    this.saveUserFile = saveUserFile;
    this.localFiles = localFiles;
  }

  cleanUp() {
    _balsamiq.default.assert(this.timerID == null);

    _balsamiq.default.assert(!this.isSynchronizing());
  }

  getAutomaticSynchronization() {
    return this.automaticSynchronization;
  }

  setAutomaticSynchronization(value) {
    _balsamiq.default.assertBool(value);

    if (this.automaticSynchronization == value) {
      return;
    }

    this.automaticSynchronization = value;
    this.context.host.log("UserFilesSynchronizer.automaticSynchronization=" + this.automaticSynchronization);

    if (this.started) {
      if (!this.automaticSynchronization) {
        this._stopTimer();
      } else if (this.automaticSynchronization) {
        this.forceSynchronize();

        this._startTimer();
      }
    }
  }

  isSynchronized() {
    // TODO: check for null
    return this.synchronizedArchiveRevision == this.archiveRevision;
  }

  _setArchiveRevisions(archiveRevision, synchronizedArchiveRevision) {
    _balsamiq.default.assertNumber(synchronizedArchiveRevision);

    _balsamiq.default.assertNumber(archiveRevision);

    let p = this.isSynchronized();
    this.archiveRevision = archiveRevision;
    this.synchronizedArchiveRevision = synchronizedArchiveRevision;

    if (p != this.isSynchronized()) {
      this.onIsSynchronizedCallback(this.isSynchronized());
    }
  }

  isSynchronizing() {
    return this.synchronizingArchiveRevision != null;
  }

  start(synchronizedArchiveRevision, archiveRevision) {
    _balsamiq.default.assertNumber(synchronizedArchiveRevision);

    _balsamiq.default.assertNumber(archiveRevision);

    this.started = true;

    this._setArchiveRevisions(archiveRevision, synchronizedArchiveRevision);

    this.onIsSynchronizedCallback(this.isSynchronized());
  }

  stop(callback) {
    _balsamiq.default.assertFunction(callback);

    _balsamiq.default.assert(this.stopCallback == null);

    this.context.host.log("UserFilesSynchronizer.stop");

    this.stopCallback = () => {
      this.context.host.log("UserFilesSynchronizer.stop: calling stopCallback");
      callback();
    }; // sync may be pending with timer or sync-ing right now


    if (this.automaticSynchronization) {
      // if timer is active, stop it
      this._stopTimer(); // no sync ongoing? sync now


      if (!this.isSynchronized() && !this.isSynchronizing()) {
        this.context.host.log("UserFilesSynchronizer.stop: sync now");

        this._synchronize();
      }
    } // no sync ongoing now: can stop


    if (!this.isSynchronizing()) {
      this.context.host.asyncCallback(this.stopCallback);
    }
  }

  onLocalFileChanged(archiveRevision) {
    _balsamiq.default.assertNumber(archiveRevision);

    _balsamiq.default.assert(this.stopCallback == null);

    if (archiveRevision == this.archiveRevision) {
      return;
    }

    this.context.host.log("UserFilesSynchronizer.onLocalFileChanged: archiveRevision=" + archiveRevision);

    this._setArchiveRevisions(archiveRevision, this.synchronizedArchiveRevision);

    this.localFiles.editLocalFileOf(this.userFileURL, () => {// don't wait for callback because there is a timer to schedule sync
    });

    if (this.automaticSynchronization) {
      this._startTimer();
    }
  }

  forceSynchronize() {
    let callback = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : () => {};

    _balsamiq.default.assertFunction(callback);

    if (this.isSynchronized()) {
      return callback({
        ok: true
      });
    }

    if (this.isSynchronizing()) {
      const stopCallback = this.stopCallback || function () {};

      this.stopCallback = res => {
        this.stopCallback = null;
        stopCallback(res);
        callback(res);
      };

      return;
    }

    this._synchronize(callback);
  }

  manualSynchronize() {
    _balsamiq.default.assert(!this.automaticSynchronization);

    if (this.isSynchronized()) {
      return;
    }

    if (this.isSynchronizing()) {
      this.needsAnotherSynchronization = true;
      return;
    }

    this._synchronize();
  }

  _startTimer() {
    _balsamiq.default.assert(this.automaticSynchronization);

    if (this.timerID) {
      this.context.host.clearTimer(this.timerID);
      this.context.host.log("UserFilesSynchronizer: (re)starting timer");
    } else {
      this.context.host.log("UserFilesSynchronizer: starting timer");
    }

    this.timerID = this.context.host.timer(AUTOMATIC_SYNCHRONIZATION_DELAY, () => {
      this.context.host.log("UserFilesSynchronizer: timer fired");
      this.timerID = null;

      if (!this.automaticSynchronization) {
        return;
      }

      if (this.isSynchronized()) {
        return;
      }

      if (this.isSynchronizing()) {
        this.needsAnotherSynchronization = true;
        return;
      }

      this._synchronize();
    });
  }

  _stopTimer() {
    if (this.timerID) {
      this.context.host.log("UserFilesSynchronizer: stopping timer");
      this.context.host.clearTimer(this.timerID);
      this.timerID = null;
    }
  }

  _synchronize(callback) {
    _balsamiq.default.assert(!this.isSynchronized(), "Already Synchronized");

    _balsamiq.default.assert(!this.isSynchronizing(), "It's Synchronizing");

    this.synchronizingArchiveRevision = this.archiveRevision;
    this.context.host.log("UserFilesSynchronizer._synchronize copy file started synchronizingArchiveRevision=" + this.synchronizingArchiveRevision);

    const finishSynchronize = res => {
      let t = this.synchronizingArchiveRevision;
      this.synchronizingArchiveRevision = null;
      this.context.host.log("UserFilesSynchronizer._synchronize copy file ended res=" + JSON.stringify(res));

      if (callback) {
        if (res.ok) {
          this._setArchiveRevisions(this.archiveRevision, t);

          if (this.needsAnotherSynchronization) {
            this.needsAnotherSynchronization = false;

            if (!this.isSynchronized()) {
              return this._synchronize(callback);
            }
          }
        }

        if (this.stopCallback) {
          this.context.host.asyncCallback(this.stopCallback, [res]);
        }

        return callback(res);
      }

      if (res.ok) {
        this._setArchiveRevisions(this.archiveRevision, t);

        if (this.needsAnotherSynchronization) {
          this.needsAnotherSynchronization = false;

          if (!this.isSynchronized()) {
            return this._synchronize();
          }
        }
      } else {
        this.errorCallback(res);
      }

      if (!this.needsAnotherSynchronization && this.stopCallback) {
        this.context.host.asyncCallback(this.stopCallback, [res]);
      }
    };

    this.saveUserFile(this.context, this.userFileURL, finishSynchronize);
  }

}

exports.UserFileSynchronizer = UserFileSynchronizer;

/***/ }),

/***/ "./js/core/controls/accordion.js":
/*!***************************************!*\
  !*** ./js/core/controls/accordion.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = exports._parseTree = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _controls = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _scrollbars = __webpack_require__(/*! core/bricks/scrollbars */ "./js/core/bricks/scrollbars.js");

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var control = (0, _controls.makeControlClass)({
  controlTypeID: "com.balsamiq.mockups::Accordion",
  controlTypeName: "Accordion",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.SCROLLBAR(),
    ...Props.PropertyDefinitions.TEXT({
      defaultText: "Item One\n" + "Item Two\n" + "- Sub-Item 2.1\n" + "- Sub-Item 2.2\n" + "Item Three\n" + "Item Four",
      multiline: true,
      splitOn: Props.SplitOnTypes.NEWLINE,
      noEmptyText: true
    }),
    ...Props.PropertyDefinitions.links()
  }
});
const ERROR_1ST_ROW = "ERROR: the 1st row must be top level  ";
(0, _controls.addStandardTextPropertyDefinitions)(control);
(0, _controls.addSelectionPropertyDefinitions)(control, 0);

control.CONTROL_PROTOTYPE.getSelectableLabels = function () {
  return this._selectableLabels;
};

control.CONTROL_PROTOTYPE.getLinkNames = function () {
  return this._selectableLabels;
};

control.CONTROL_PROTOTYPE.getLinkableLabels = function (text) {
  return parseTree(text).map(line => TextUtils.unformatText(line.text));
};

control.CONTROL_PROTOTYPE.initData = function () {
  this._lines = [];
  this._texts = [];
};

var parseTree = function (p_text) {
  var newNode = function (p_text, p_isChild, p_fatherIndex) {
    return {
      text: p_text,
      isChild: p_isChild,
      fatherIndex: p_isChild ? p_fatherIndex : 'none',
      numChildren: 0
    };
  };

  var isChildText = function (text) {
    return text.indexOf('- ') === 0 || text.indexOf('\t') === 0;
  };

  var lines = p_text.split('\n');
  var lastFatherIndex = -1;
  var nodes = [];
  var i;

  for (i = 0; i < lines.length; i += 1) {
    var isChild = isChildText(lines[i]);
    var text = lines[i].replace(/^- /, "").replace(/^\t/, "");
    var node = newNode(text, isChild, lastFatherIndex);
    nodes.push(node);

    if (!isChild) {
      lastFatherIndex = i;
    } else {
      if (i == 0) {
        // the first row can't be a child
        node.text = ERROR_1ST_ROW;
        node.isChild = false;
        node.fatherIndex = 'none';
        lastFatherIndex = i;
      } else {
        nodes[lastFatherIndex].numChildren += 1;
      }
    }
  }

  return nodes;
};

var setLinesSelection = function (p_lines, p_selectedIndex) {
  if (p_selectedIndex < 0 || p_lines[p_selectedIndex] == undefined) {
    return;
  }

  p_lines[p_selectedIndex].selected = true;

  if (p_lines[p_selectedIndex].isChild) {
    var fatherIdx = p_lines[p_selectedIndex].fatherIndex;
    p_lines[fatherIdx].selected = true;
  }
};

var setLinesVisibility = function (p_lines, p_selectedIndex) {
  // Visible if top level
  // Visible if selected
  // Visible if child of selected node
  p_lines.forEach(function (node, index) {
    node.visible = node.isChild === false || index === p_selectedIndex || node.isChild && node.fatherIndex === p_selectedIndex;
  }); //avoiding accordion not drawing when last item is a child and is selected
  //pivotal

  if (p_selectedIndex >= 0 && p_lines[p_selectedIndex]) {
    var i = p_lines[p_selectedIndex].fatherIndex + 1;

    while (i < p_lines.length && p_lines[i].isChild == true) {
      p_lines[i].visible = true;
      i++;
    }
  }
};

var getNumberOfVisibleLines = function (p_lines, p_selectedIndex) {
  var results = _balsamiq.default.countBy(p_lines, function (node) {
    return node.visible ? 'visible' : 'hidden';
  });

  return results.visible;
}; // Gets the index of the last child of the selected line


var getLastChildIndex = function (p_lines, p_index) {
  var res;

  if (p_index === -1 || p_lines[p_index] == undefined) {
    res = p_lines.length - 1;
  } else if (p_lines[p_index].isChild) {
    var fatherIndex = p_lines[p_index].fatherIndex;
    res = fatherIndex + p_lines[fatherIndex].numChildren;
  } else {
    res = p_index + p_lines[p_index].numChildren;
  }

  return res;
};

control.CONTROL_PROTOTYPE.makeLineDisplayItems = function (p_line, p_bbox, p_color) {
  let subItemPadding = p_line.isChild ? 10 : 0;
  let line = null; // For top items draw full rectangle and background

  if (!p_line.isChild) {
    line = (0, _rectangle.makeRectangleDisplayItems)(this._context, p_bbox, Props.BorderStyles.SQUARE, p_color);
  } // For child items draw only background color if selected


  if (p_line.selected && p_line.isChild) {
    let topBottomPadding = this.style("accordion_overlap");

    let rect = _types.default.paddedRectangle(p_bbox, topBottomPadding, topBottomPadding, subItemPadding, this.style("accordion_padding_right"));

    if (rect.width > 0 && rect.height > 0) {
      line = (0, _rectangle.makeFilledRectangle)(rect, p_color);
    }
  } // Always draw line text


  let rect = _types.default.paddedRectangle(p_bbox, 0, 0, this.style("accordion_padding_left") + subItemPadding);

  let lineText = null;

  if (rect.width > 0 && rect.height > 0) {
    lineText = p_line.text.makeDisplayItems(rect, _types.default.TextAlign.LEFT, _types.default.Positions.CENTER);
  }

  return [line, lineText];
};

control.CONTROL_PROTOTYPE.updateData = function () {
  this._selectableLabels = this.getLinkableLabels(this.getActualProperty("text")); // Parse the text and make our data structure

  const lines = parseTree(this.getActualProperty("text")); // Mark lines for visibility based on selection

  const selectedIndex = this.getActualProperty("selectedIndex");
  setLinesVisibility(lines, selectedIndex);
  setLinesSelection(lines, selectedIndex); // Create texts

  const selectedTextStyle = this.newStandardTextStyle({
    textColor: _colors.default.labelColorGivenBackgroundColor(this.style('default_selection_color'), 1, this.style("default_text_color")),
    linkColor: _colors.default.labelColorGivenBackgroundColor(this.style('default_selection_color'), 1, this.style("default_link_color"))
  });
  const textStyle = this.newStandardTextStyle();
  const isPreviousTextOfDifferentLength = !this._lines || this._lines.length !== lines.length;
  this._lines = lines.map((line, index) => {
    const style = line.selected ? selectedTextStyle : textStyle;
    line.text = isPreviousTextOfDifferentLength || !(0, _text.areTextsEqual)(this._lines[index].text, line.text, style) ? (0, _text.newSingleLineText)(line.text, style, this._context) : this._lines[index].text;
    return line;
  });
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var size = _balsamiq.default.clone(p_sizeProperty);

  var defaultSize = {
    width: 150,
    height: 186
  };

  if (size.width === _types.default.AUTOSIZE) {
    var textSizes = this._lines.map(line => {
      return line.text.measuredSize();
    });

    var maxTextWidth = _balsamiq.default.maxBy(textSizes, function (size) {
      return size.width;
    }).width;

    size.width = Math.max(maxTextWidth + this.style("accordion_padding_left") + this.style("accordion_padding_right"), defaultSize.width);
  }

  if (size.height === _types.default.AUTOSIZE) {
    var paddingMiddle = this.style("accordion_padding_middle");
    var offset = this.style("accordion_overlap");
    var paddingTop = this.style("accordion_padding_top");
    var paddingBottom = this.style("accordion_padding_bottom");
    var h = paddingTop + paddingBottom + 70;
    h += getNumberOfVisibleLines(this._lines, this.getActualProperty('selectedIndex')) * (this._lines[0].text.measuredSize().height + paddingMiddle - offset);
    size.height = Math.max(h, defaultSize.height);
  }

  return size;
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  // Make border for the whole control
  var background = (0, _rectangle.makeRectangleDisplayItems)(this._context, p_localBbox);
  var linkBounds = [];
  var bounds;
  var paddingMiddle = this.style("accordion_padding_middle");
  var selectedColor = this.style("default_selection_color");
  var selectedIndex = this.getActualProperty("selectedIndex");
  var spaceRemaining = p_localBbox.height;
  let rowOverlap = this.style("accordion_overlap"); // Draw all lines up to all children of selected line

  var i;
  var y = 0;
  var scrollbarY;
  var upToIndex = getLastChildIndex(this._lines, selectedIndex);
  var line, color, rowHeight, lineBBox;
  var displayItems = [];
  rowHeight = this._lines[0].text.measuredSize().height + paddingMiddle;

  for (i = 0; i <= upToIndex; i += 1) {
    line = this._lines[i];

    if (line.visible && rowHeight - rowOverlap <= spaceRemaining) {
      rowHeight = line.text.measuredSize().height + paddingMiddle;
      lineBBox = _types.default.newRectangle(0, y, p_localBbox.width, Math.min(rowHeight, spaceRemaining));
      color = line.selected ? selectedColor : 0xFFFFFF;
      displayItems.push([this.makeLineDisplayItems(line, lineBBox, color)]);

      if (this.getActualProperty('verticalScrollbar') && scrollbarY && y + rowHeight >= scrollbarY) {
        bounds = _types.default.newRectangle(0, y, Math.max(p_localBbox.width - _scrollbars.Sizes.ScrollbarFixedSize, 0), rowHeight - 3);
      } else {
        bounds = _types.default.newRectangle(0, y, p_localBbox.width, rowHeight - 1);
      }

      linkBounds.push(bounds);
      y += rowHeight - rowOverlap;
      spaceRemaining -= rowHeight - rowOverlap; // Starting y of scrollbar is just after the selected top item

      if (!scrollbarY && line.selected && !line.child && spaceRemaining) {
        scrollbarY = y + this.style("accordion_overlap") - 1; // one pixel overlap with top line
      }
    } else {
      //needed otherwise the links are placed in the wrong place after hidden lines
      linkBounds.push(null);
    }
  } // Remains space to fit other Lines??


  var linesToDisplay = 0;
  var nextLineIndex = upToIndex + 1;

  while (nextLineIndex < this._lines.length) {
    if (this._lines[nextLineIndex].visible) {
      linesToDisplay += 1;
    }

    nextLineIndex += 1;
  }

  var overlapSpace,
      spaceNeeded = 0;

  if (linesToDisplay) {
    overlapSpace = (linesToDisplay - 1) * this.style("accordion_overlap"); //rectangles overlap partially!

    spaceNeeded = linesToDisplay * (this._lines[upToIndex + 1].text.measuredSizeForWidth(p_localBbox.width).height + paddingMiddle) - overlapSpace;
  } // Make scrollbar


  var scrollbar, scrollBarHeight, scrollbarX;

  if (scrollbarY && this.getActualProperty('verticalScrollbar')) {
    scrollbarX = p_localBbox.width - _scrollbars.Sizes.ScrollbarFixedSize - this.style("accordion_scrollbar_padding_right");
    scrollBarHeight = p_localBbox.height - scrollbarY - (spaceRemaining >= spaceNeeded ? spaceNeeded - 1 : spaceRemaining > rowHeight ? spaceRemaining - 1 : 0);
    let scrollBarOffset = scrollbarY + scrollBarHeight > p_localBbox.height ? scrollbarY + scrollBarHeight - p_localBbox.height : 0;

    if (scrollBarHeight > rowHeight) {
      scrollbar = (0, _scrollbars.makeDisplayItemsForVerticalScrollbar)(scrollbarX, scrollbarY, scrollBarHeight - scrollBarOffset, this.getActualProperty("value"), _types.default.ControlState.NORMAL, this._context);
    }
  }

  if (linesToDisplay) {
    if (spaceRemaining > spaceNeeded) {
      y = p_localBbox.height - spaceNeeded;
    }

    i = upToIndex + 1;
    rowHeight = this._lines[i].text.measuredSizeForWidth(p_localBbox.width).height + paddingMiddle;

    while (i < this._lines.length && spaceRemaining >= rowHeight - rowOverlap) {
      line = this._lines[i];

      if (line.visible) {
        rowHeight = line.text.measuredSize().height + paddingMiddle;
        lineBBox = _types.default.newRectangle(0, y, p_localBbox.width, Math.min(rowHeight, spaceRemaining));
        displayItems.push([this.makeLineDisplayItems(line, lineBBox, 0xFFFFFF)]);
        bounds = _types.default.newRectangle(0, y, p_localBbox.width, rowHeight);
        linkBounds.push(bounds);
        y += rowHeight - rowOverlap;
        spaceRemaining -= rowHeight - rowOverlap;
      } else {
        linkBounds.push(null);
      }

      i++;
    }
  }

  _link.default.processLinks(this, linkBounds);

  this.pushToDisplayList(background, displayItems.reverse(), scrollbar);
};

control._parseTree = parseTree;
var _default = control;
exports["default"] = _default;
const _parseTree = parseTree;
exports._parseTree = _parseTree;

/***/ }),

/***/ "./js/core/controls/alert.js":
/*!***********************************!*\
  !*** ./js/core/controls/alert.js ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const MIN_WIDTH = 200;
var control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::Alert",
  controlTypeName: "Alert Box",
  controlAlternateNames: ["Alert"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.TEXT({
      defaultText: "Alert\nAlert text goes here\nNo, Yes",
      multiline: true
    }),
    ...Props.PropertyDefinitions.links(),
    ...Props.PropertyDefinitions.BORDER_STYLE(Props.BorderStyles.SQUARE, Props.SimpleBorderStyles),
    device: {
      defaultValue: Props.Devices.IOS,
      enum: Props.Devices
    },
    align: {
      defaultValue: Props.TextAlignments.CENTER,
      enum: Props.TextAlignments
    }
  }
});
Controls.addStandardTextPropertyDefinitions(control, {
  multiline: true
});
Controls.addColorPropertyDefinitions(control, {
  defaultTextColor: "alert_button_color",
  oldTextColorName: 'color'
});

control.CONTROL_PROTOTYPE.getLinkNames = function () {
  return this._linkNames;
};

control.CONTROL_PROTOTYPE.getLinkableLabels = function (text) {
  let texts = text.split("\n");
  let buttons = texts.length >= 3 ? texts[2].split(",") : [];
  let linkableLabels = [];

  if (buttons.length >= 1) {
    linkableLabels.push(buttons[0].trim());
  }

  if (buttons.length >= 2) {
    linkableLabels.push(buttons[1].trim());
  }

  return linkableLabels;
};

control.CONTROL_PROTOTYPE.updateData = function () {
  const texts = this.getActualProperty("text").split("\n");

  if (texts.length > 0) {
    const title = texts[0];
    const textStyle = this.newStandardTextStyle();
    textStyle.fontSize = this.getActualProperty("fontSize") + 2;

    if (!this.titleText || !Text.areTextsEqual(this.titleText, title, textStyle)) {
      this.titleText = Text.newTextNoWrap(title, textStyle, this._context);
    }
  }

  const textStyle = this.newStandardTextStyle();
  this.messageText = texts.length > 1 ? this.messageText && Text.areTextsEqual(this.messageText, texts[1], textStyle) ? this.messageText : Text.newText(texts[1], textStyle, this._context) : null;
  const isiOS = this.getActualProperty("device") != "android";
  const hasColor = this.hasPropertyValue("textColor");
  const color = this.getActualProperty("textColor");
  const buttonTitles = texts.length > 2 ? texts[2].split(",") : [];
  const buttonTextStyle = this.newStandardTextStyle({
    textColor: isiOS || hasColor ? color : this.style("default_link_color"),
    underline: !isiOS
  });
  this._linkNames = [];

  if (buttonTitles.length > 0) {
    const text = buttonTitles[0].trim();

    this._linkNames.push(text);

    if (!this.leftButtonText || !Text.areTextsEqual(this.leftButtonText, text, buttonTextStyle)) {
      this.leftButtonText = Text.newTextNoWrap(text, buttonTextStyle, this._context);
    }
  } else {
    this.leftButtonText = null;
  }

  if (buttonTitles.length > 1) {
    const text = buttonTitles[1].trim();

    this._linkNames.push(text);

    if (!this.rightButtonText || !Text.areTextsEqual(this.rightButtonText, text, buttonTextStyle)) {
      this.rightButtonText = Text.newTextNoWrap(text, buttonTextStyle, this._context);
    }
  } else {
    this.rightButtonText = null;
  }
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var padding = this.style("alert_padding");
  var linePadding = this.style("alert_line_padding");
  var width = this.titleText ? this.titleText.measuredSize().width + 2 * padding : 0;
  var buttonsWidth = 0;

  if (this.leftButtonText) {
    buttonsWidth += this.leftButtonText.measuredSize().width + padding * 2;
  }

  if (this.rightButtonText) {
    buttonsWidth += this.rightButtonText.measuredSize().width + padding * 2;
  }

  width = Math.max(width, buttonsWidth, MIN_WIDTH, this.getActualProperty("fontSize") * 10);
  var height = padding / 2 + (this.titleText ? this.titleText.measuredSize().height + linePadding : 0) + 10;

  if (this.messageText) {
    height += linePadding + this.messageText.measuredSizeForWidth(width).height;
    height += this.titleText ? padding : 0;
  }

  var buttonsHeight = 0;

  if (this.leftButtonText) {
    buttonsHeight = this.leftButtonText.measuredSize().height + 2 * padding;
  }

  if (this.rightButtonText) {
    buttonsHeight = Math.max(buttonsHeight, this.rightButtonText.measuredSize().height + 2 * padding);
  }

  height += padding + buttonsHeight;
  return _types.default.newSize(width, height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (localBbox) {
  if (this.getActualProperty("borderStyle") === Props.BorderStyles.SQUARE) {
    this.pushToDisplayList(Rectangle.makeRectangleDisplayItems(this._context, localBbox, Props.BorderStyles.ROUNDED_SOLID));
  } else {
    this.pushToDisplayList(Rectangle.makeFilledRectangle(localBbox, null, 1, this.style("rectangle_corner_radius").roundedSolid));
  }

  const padding = this.style("alert_padding");
  const linePadding = this.style("alert_line_padding");

  let textBounds = _types.default.newRectangle(padding, padding / 2, Math.max(0, localBbox.width - padding * 2), 1);

  let heightLeft = localBbox.height;

  if (this.titleText) {
    textBounds.y += linePadding;
    textBounds.height = this.titleText.measuredSize().height;
    heightLeft -= textBounds.height + padding;

    if (heightLeft > 0) {
      this.pushToDisplayList(this.titleText.makeDisplayItems(textBounds, this.getActualProperty('align')));
      textBounds.y += linePadding + textBounds.height;
    }
  }

  if (this.messageText && heightLeft > 0) {
    textBounds = _types.default.newRectangle(padding, textBounds.y + (this.titleText ? padding : 0) + linePadding - 3, Math.max(0, localBbox.width - padding * 2), this.messageText.measuredSizeForWidth(localBbox.width - padding * 2).height);
    heightLeft -= textBounds.height + 2 * padding;

    if (heightLeft > 0) {
      this.pushToDisplayList(this.messageText.makeDisplayItems(textBounds, this.getActualProperty('align')));
    }
  }

  const linkBounds = [];
  const isiOS = this.getActualProperty("device") !== "android";
  let buttonsHeight;

  if (this.leftButtonText) {
    buttonsHeight = Math.max(this.leftButtonText.measuredSize().height, this.rightButtonText ? this.rightButtonText.measuredSize().height : 0);
    heightLeft -= buttonsHeight + padding;
  }

  if (this.leftButtonText && heightLeft > 0) {
    const y = Math.max(0, localBbox.height - padding - buttonsHeight + 3);

    if (isiOS) {
      // Horizontal divider
      const ruleHeight = this.styleSVGSize("rule_horizontal").height;

      const dividerHBbox = _types.default.newRectangle(1, localBbox.height - buttonsHeight - 2 * padding, localBbox.width - 2, ruleHeight);

      this.pushToDisplayList(_displayitem.default.createDisplayItemsFromSkin("rule_horizontal", this.style("alert_separator_color"), dividerHBbox, this._context));
    }

    let buttonBounds;

    if (this.rightButtonText) {
      if (isiOS) {
        // Vertical divider
        const ruleWidth = this.styleSVGSize("rule_vertical").width;

        const dividerVBbox = _types.default.newRectangle((localBbox.width - ruleWidth) / 2, localBbox.height - buttonsHeight - 2 * padding + 5, ruleWidth, buttonsHeight + 2 * padding - 1);

        this.pushToDisplayList(_displayitem.default.createDisplayItemsFromSkin("rule_vertical", this.style("alert_separator_color"), dividerVBbox, this._context)); // Two buttons

        textBounds = _types.default.newRectangle(padding - 2, this.leftButtonText.measuredSize().height == buttonsHeight ? y : y + (buttonsHeight - this.leftButtonText.measuredSize().height) / 2, Math.max(0, (localBbox.width - 3 * padding) / 2), buttonsHeight);
        buttonBounds = _types.default.newRectangle(textBounds.x - padding / 2 - 3, Math.max(0, localBbox.height - buttonsHeight - 2 * padding + 5), textBounds.width + padding + 4, buttonsHeight + 2 * padding - 6); // Button 1

        this.pushToDisplayList(this.leftButtonText.makeDisplayItems(textBounds, _types.default.TextAlign.CENTER));
        linkBounds.push(_balsamiq.default.clone(buttonBounds)); // Button 2

        textBounds = _types.default.newRectangle(localBbox.width - padding - textBounds.width + 3, this.rightButtonText.measuredSize().height == buttonsHeight ? y : y + (buttonsHeight - this.rightButtonText.measuredSize().height) / 2, Math.max(0, textBounds.width), this.rightButtonText.measuredSize().height);
        buttonBounds.x = textBounds.x - padding / 2 - 2;
        this.pushToDisplayList(this.rightButtonText.makeDisplayItems(textBounds, _types.default.TextAlign.CENTER));
        linkBounds.push(_balsamiq.default.clone(buttonBounds));
      } else {
        const theX = localBbox.width - 2 * padding - this.rightButtonText.measuredSize().width; // Button2

        textBounds = _types.default.newRectangle(theX, this.rightButtonText.measuredSize().height == buttonsHeight ? y : y + (buttonsHeight - this.rightButtonText.measuredSize().height) / 2, this.rightButtonText.measuredSize().width, this.rightButtonText.measuredSize().height);
        this.pushToDisplayList(this.rightButtonText.makeDisplayItems(textBounds, _types.default.TextAlign.CENTER)); // Button1

        textBounds.x = textBounds.x - 2 * padding - this.leftButtonText.measuredSize().width;
        textBounds.width = this.leftButtonText.measuredSize().width;
        this.pushToDisplayList(this.leftButtonText.makeDisplayItems(textBounds, _types.default.TextAlign.CENTER)); // Button 1 bounds

        const buttonX = Math.max(textBounds.x - padding, 0);
        buttonBounds = _types.default.newRectangle(buttonX, Math.max(0, localBbox.height - buttonsHeight - 2 * padding + 5), textBounds.width + (buttonX < padding ? padding + buttonX : 2 * padding), buttonsHeight + 2 * padding - 6);

        if (textBounds.x >= 0) {
          linkBounds.push(_balsamiq.default.clone(buttonBounds));
        } // Button 2 bounds


        buttonBounds.x = Math.max(theX - padding, textBounds.x + textBounds.width, 0);
        buttonBounds.width = localBbox.width - buttonBounds.x - 2;
        linkBounds.push(_balsamiq.default.clone(buttonBounds));
      }
    } else {
      // One button
      let w;

      if (isiOS) {
        w = Math.min(Math.max(this.leftButtonText.measuredSize().width, localBbox.width / 2), Math.max(0, localBbox.width - padding * 2));
        textBounds = _types.default.newRectangle((localBbox.width - w) / 2, y, w, this.leftButtonText.measuredSize().height);
      } else {
        w = this.leftButtonText.measuredSize().width + 2 * padding;
        textBounds = _types.default.newRectangle(localBbox.width - w - padding, y, w, this.leftButtonText.measuredSize().height);
      }

      buttonBounds = _types.default.newRectangle(0, Math.max(0, localBbox.height - buttonsHeight - 2 * padding + 5), localBbox.width, buttonsHeight + 2 * padding - 6);
      this.pushToDisplayList(this.leftButtonText.makeDisplayItems(textBounds, _types.default.TextAlign.CENTER));
      linkBounds.push(buttonBounds);
    }
  }

  _link.default.processLinks(this, linkBounds);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/all.js":
/*!*********************************!*\
  !*** ./js/core/controls/all.js ***!
  \*********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
Object.defineProperty(exports, "accordion", ({
  enumerable: true,
  get: function () {
    return _accordion.default;
  }
}));
Object.defineProperty(exports, "alert", ({
  enumerable: true,
  get: function () {
    return _alert.default;
  }
}));
Object.defineProperty(exports, "android", ({
  enumerable: true,
  get: function () {
    return _android.default;
  }
}));
Object.defineProperty(exports, "appbar", ({
  enumerable: true,
  get: function () {
    return _appbar.default;
  }
}));
Object.defineProperty(exports, "arrow", ({
  enumerable: true,
  get: function () {
    return _arrow.default;
  }
}));
Object.defineProperty(exports, "barcharts", ({
  enumerable: true,
  get: function () {
    return _barcharts.default;
  }
}));
Object.defineProperty(exports, "blockoftext", ({
  enumerable: true,
  get: function () {
    return _blockoftext.default;
  }
}));
Object.defineProperty(exports, "breadcrumbs", ({
  enumerable: true,
  get: function () {
    return _breadcrumbs.default;
  }
}));
Object.defineProperty(exports, "browserWindow", ({
  enumerable: true,
  get: function () {
    return _browserWindow.default;
  }
}));
Object.defineProperty(exports, "buttonBar", ({
  enumerable: true,
  get: function () {
    return _buttonBar.default;
  }
}));
Object.defineProperty(exports, "buttons", ({
  enumerable: true,
  get: function () {
    return _buttons.default;
  }
}));
Object.defineProperty(exports, "calendar", ({
  enumerable: true,
  get: function () {
    return _calendar.default;
  }
}));
Object.defineProperty(exports, "callout", ({
  enumerable: true,
  get: function () {
    return _callout.default;
  }
}));
Object.defineProperty(exports, "canvas", ({
  enumerable: true,
  get: function () {
    return _canvas.default;
  }
}));
Object.defineProperty(exports, "choicecontrols", ({
  enumerable: true,
  get: function () {
    return _choicecontrols.default;
  }
}));
Object.defineProperty(exports, "choicegroupcontrols", ({
  enumerable: true,
  get: function () {
    return _choicegroupcontrols.default;
  }
}));
Object.defineProperty(exports, "circlebutton", ({
  enumerable: true,
  get: function () {
    return _circlebutton.default;
  }
}));
Object.defineProperty(exports, "colorpicker", ({
  enumerable: true,
  get: function () {
    return _colorpicker.default;
  }
}));
Object.defineProperty(exports, "combobox", ({
  enumerable: true,
  get: function () {
    return _combobox.default;
  }
}));
Object.defineProperty(exports, "coverflow", ({
  enumerable: true,
  get: function () {
    return _coverflow.default;
  }
}));
Object.defineProperty(exports, "curlybraces", ({
  enumerable: true,
  get: function () {
    return _curlybraces.default;
  }
}));
Object.defineProperty(exports, "datagrid", ({
  enumerable: true,
  get: function () {
    return _datagrid.default;
  }
}));
Object.defineProperty(exports, "datechooser", ({
  enumerable: true,
  get: function () {
    return _datechooser.default;
  }
}));
Object.defineProperty(exports, "datepicker", ({
  enumerable: true,
  get: function () {
    return _datepicker.default;
  }
}));
Object.defineProperty(exports, "fieldset", ({
  enumerable: true,
  get: function () {
    return _fieldset.default;
  }
}));
Object.defineProperty(exports, "formattingtoolbar", ({
  enumerable: true,
  get: function () {
    return _formattingtoolbar.default;
  }
}));
Object.defineProperty(exports, "group", ({
  enumerable: true,
  get: function () {
    return _group.default;
  }
}));
Object.defineProperty(exports, "helpbutton", ({
  enumerable: true,
  get: function () {
    return _helpbutton.default;
  }
}));
Object.defineProperty(exports, "icon", ({
  enumerable: true,
  get: function () {
    return _icon.default;
  }
}));
Object.defineProperty(exports, "iconlabel", ({
  enumerable: true,
  get: function () {
    return _iconlabel.default;
  }
}));
Object.defineProperty(exports, "image", ({
  enumerable: true,
  get: function () {
    return _image.default;
  }
}));
Object.defineProperty(exports, "ioskeyboard", ({
  enumerable: true,
  get: function () {
    return _ioskeyboard.default;
  }
}));
Object.defineProperty(exports, "ipad", ({
  enumerable: true,
  get: function () {
    return _ipad.default;
  }
}));
Object.defineProperty(exports, "iphone", ({
  enumerable: true,
  get: function () {
    return _iphone.default;
  }
}));
Object.defineProperty(exports, "iphonemenu", ({
  enumerable: true,
  get: function () {
    return _iphonemenu.default;
  }
}));
Object.defineProperty(exports, "iphonepicker", ({
  enumerable: true,
  get: function () {
    return _iphonepicker.default;
  }
}));
Object.defineProperty(exports, "label", ({
  enumerable: true,
  get: function () {
    return _label.default;
  }
}));
Object.defineProperty(exports, "linechart", ({
  enumerable: true,
  get: function () {
    return _linechart.default;
  }
}));
Object.defineProperty(exports, "lineoftext", ({
  enumerable: true,
  get: function () {
    return _lineoftext.default;
  }
}));
Object.defineProperty(exports, "link", ({
  enumerable: true,
  get: function () {
    return _link.default;
  }
}));
Object.defineProperty(exports, "linkbar", ({
  enumerable: true,
  get: function () {
    return _linkbar.default;
  }
}));
Object.defineProperty(exports, "list", ({
  enumerable: true,
  get: function () {
    return _list.default;
  }
}));
Object.defineProperty(exports, "menu", ({
  enumerable: true,
  get: function () {
    return _menu.default;
  }
}));
Object.defineProperty(exports, "menubar", ({
  enumerable: true,
  get: function () {
    return _menubar.default;
  }
}));
Object.defineProperty(exports, "modalscreen", ({
  enumerable: true,
  get: function () {
    return _modalscreen.default;
  }
}));
Object.defineProperty(exports, "numericstepper", ({
  enumerable: true,
  get: function () {
    return _numericstepper.default;
  }
}));
Object.defineProperty(exports, "paragraph", ({
  enumerable: true,
  get: function () {
    return _paragraph.default;
  }
}));
Object.defineProperty(exports, "piechart", ({
  enumerable: true,
  get: function () {
    return _piechart.default;
  }
}));
Object.defineProperty(exports, "playback", ({
  enumerable: true,
  get: function () {
    return _playback.default;
  }
}));
Object.defineProperty(exports, "popover", ({
  enumerable: true,
  get: function () {
    return _popover.default;
  }
}));
Object.defineProperty(exports, "progressbar", ({
  enumerable: true,
  get: function () {
    return _progressbar.default;
  }
}));
Object.defineProperty(exports, "redx", ({
  enumerable: true,
  get: function () {
    return _redx.default;
  }
}));
Object.defineProperty(exports, "rules", ({
  enumerable: true,
  get: function () {
    return _rules.default;
  }
}));
Object.defineProperty(exports, "scratchout", ({
  enumerable: true,
  get: function () {
    return _scratchout.default;
  }
}));
Object.defineProperty(exports, "scrollbarhorizontal", ({
  enumerable: true,
  get: function () {
    return _scrollbarhorizontal.default;
  }
}));
Object.defineProperty(exports, "scrollbarvertical", ({
  enumerable: true,
  get: function () {
    return _scrollbarvertical.default;
  }
}));
Object.defineProperty(exports, "searchbox", ({
  enumerable: true,
  get: function () {
    return _searchbox.default;
  }
}));
Object.defineProperty(exports, "shape", ({
  enumerable: true,
  get: function () {
    return _shape.default;
  }
}));
Object.defineProperty(exports, "sitemap", ({
  enumerable: true,
  get: function () {
    return _sitemap.default;
  }
}));
Object.defineProperty(exports, "sliders", ({
  enumerable: true,
  get: function () {
    return _sliders.default;
  }
}));
Object.defineProperty(exports, "splitterhorizontal", ({
  enumerable: true,
  get: function () {
    return _splitterhorizontal.default;
  }
}));
Object.defineProperty(exports, "splittervertical", ({
  enumerable: true,
  get: function () {
    return _splittervertical.default;
  }
}));
Object.defineProperty(exports, "stickynote", ({
  enumerable: true,
  get: function () {
    return _stickynote.default;
  }
}));
Object.defineProperty(exports, "streetmap", ({
  enumerable: true,
  get: function () {
    return _streetmap.default;
  }
}));
Object.defineProperty(exports, "switch", ({
  enumerable: true,
  get: function () {
    return _switch.default;
  }
}));
Object.defineProperty(exports, "symbol", ({
  enumerable: true,
  get: function () {
    return _symbol.default;
  }
}));
Object.defineProperty(exports, "tabbars", ({
  enumerable: true,
  get: function () {
    return _tabbars.default;
  }
}));
Object.defineProperty(exports, "tagcloud", ({
  enumerable: true,
  get: function () {
    return _tagcloud.default;
  }
}));
Object.defineProperty(exports, "textarea", ({
  enumerable: true,
  get: function () {
    return _textarea.default;
  }
}));
Object.defineProperty(exports, "textinput", ({
  enumerable: true,
  get: function () {
    return _textinput.default;
  }
}));
Object.defineProperty(exports, "timepicker", ({
  enumerable: true,
  get: function () {
    return _timepicker.default;
  }
}));
Object.defineProperty(exports, "titles", ({
  enumerable: true,
  get: function () {
    return _titles.default;
  }
}));
Object.defineProperty(exports, "tooltip", ({
  enumerable: true,
  get: function () {
    return _tooltip.default;
  }
}));
Object.defineProperty(exports, "tree", ({
  enumerable: true,
  get: function () {
    return _tree.default;
  }
}));
Object.defineProperty(exports, "videoplayer", ({
  enumerable: true,
  get: function () {
    return _videoplayer.default;
  }
}));
Object.defineProperty(exports, "volumeslider", ({
  enumerable: true,
  get: function () {
    return _volumeslider.default;
  }
}));
Object.defineProperty(exports, "webcam", ({
  enumerable: true,
  get: function () {
    return _webcam.default;
  }
}));
Object.defineProperty(exports, "window", ({
  enumerable: true,
  get: function () {
    return _window.default;
  }
}));

var _accordion = _interopRequireDefault(__webpack_require__(/*! core/controls/accordion */ "./js/core/controls/accordion.js"));

var _alert = _interopRequireDefault(__webpack_require__(/*! core/controls/alert */ "./js/core/controls/alert.js"));

var _arrow = _interopRequireDefault(__webpack_require__(/*! core/controls/arrow */ "./js/core/controls/arrow.js"));

var _appbar = _interopRequireDefault(__webpack_require__(/*! core/controls/appbar */ "./js/core/controls/appbar.js"));

var _barcharts = _interopRequireDefault(__webpack_require__(/*! core/controls/barcharts */ "./js/core/controls/barcharts.js"));

var _blockoftext = _interopRequireDefault(__webpack_require__(/*! core/controls/blockoftext */ "./js/core/controls/blockoftext.js"));

var _breadcrumbs = _interopRequireDefault(__webpack_require__(/*! core/controls/breadcrumbs */ "./js/core/controls/breadcrumbs.js"));

var _browserWindow = _interopRequireDefault(__webpack_require__(/*! core/controls/browserWindow */ "./js/core/controls/browserWindow.js"));

var _buttonBar = _interopRequireDefault(__webpack_require__(/*! core/controls/buttonBar */ "./js/core/controls/buttonBar.js"));

var _buttons = _interopRequireDefault(__webpack_require__(/*! core/controls/buttons */ "./js/core/controls/buttons.js"));

var _calendar = _interopRequireDefault(__webpack_require__(/*! core/controls/calendar */ "./js/core/controls/calendar.js"));

var _callout = _interopRequireDefault(__webpack_require__(/*! core/controls/callout */ "./js/core/controls/callout.js"));

var _canvas = _interopRequireDefault(__webpack_require__(/*! core/controls/canvas */ "./js/core/controls/canvas.js"));

var _choicecontrols = _interopRequireDefault(__webpack_require__(/*! core/controls/choicecontrols */ "./js/core/controls/choicecontrols.js"));

var _choicegroupcontrols = _interopRequireDefault(__webpack_require__(/*! core/controls/choicegroupcontrols */ "./js/core/controls/choicegroupcontrols.js"));

var _circlebutton = _interopRequireDefault(__webpack_require__(/*! core/controls/circlebutton */ "./js/core/controls/circlebutton.js"));

var _colorpicker = _interopRequireDefault(__webpack_require__(/*! core/controls/colorpicker */ "./js/core/controls/colorpicker.js"));

var _combobox = _interopRequireDefault(__webpack_require__(/*! core/controls/combobox */ "./js/core/controls/combobox.js"));

var _coverflow = _interopRequireDefault(__webpack_require__(/*! core/controls/coverflow */ "./js/core/controls/coverflow.js"));

var _curlybraces = _interopRequireDefault(__webpack_require__(/*! core/controls/curlybraces */ "./js/core/controls/curlybraces.js"));

var _datagrid = _interopRequireDefault(__webpack_require__(/*! core/controls/datagrid */ "./js/core/controls/datagrid.js"));

var _datechooser = _interopRequireDefault(__webpack_require__(/*! core/controls/datechooser */ "./js/core/controls/datechooser.js"));

var _datepicker = _interopRequireDefault(__webpack_require__(/*! core/controls/datepicker */ "./js/core/controls/datepicker.js"));

var _fieldset = _interopRequireDefault(__webpack_require__(/*! core/controls/fieldset */ "./js/core/controls/fieldset.js"));

var _formattingtoolbar = _interopRequireDefault(__webpack_require__(/*! core/controls/formattingtoolbar */ "./js/core/controls/formattingtoolbar.js"));

var _group = _interopRequireDefault(__webpack_require__(/*! core/controls/group */ "./js/core/controls/group.js"));

var _helpbutton = _interopRequireDefault(__webpack_require__(/*! core/controls/helpbutton */ "./js/core/controls/helpbutton.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/controls/icon */ "./js/core/controls/icon.js"));

var _iconlabel = _interopRequireDefault(__webpack_require__(/*! core/controls/iconlabel */ "./js/core/controls/iconlabel.js"));

var _image = _interopRequireDefault(__webpack_require__(/*! core/controls/image */ "./js/core/controls/image.js"));

var _ipad = _interopRequireDefault(__webpack_require__(/*! core/controls/ipad */ "./js/core/controls/ipad.js"));

var _iphone = _interopRequireDefault(__webpack_require__(/*! core/controls/iphone */ "./js/core/controls/iphone.js"));

var _android = _interopRequireDefault(__webpack_require__(/*! core/controls/android */ "./js/core/controls/android.js"));

var _ioskeyboard = _interopRequireDefault(__webpack_require__(/*! core/controls/ioskeyboard */ "./js/core/controls/ioskeyboard.js"));

var _iphonemenu = _interopRequireDefault(__webpack_require__(/*! core/controls/iphonemenu */ "./js/core/controls/iphonemenu.js"));

var _iphonepicker = _interopRequireDefault(__webpack_require__(/*! core/controls/iphonepicker */ "./js/core/controls/iphonepicker.js"));

var _label = _interopRequireDefault(__webpack_require__(/*! core/controls/label */ "./js/core/controls/label.js"));

var _linechart = _interopRequireDefault(__webpack_require__(/*! core/controls/linechart */ "./js/core/controls/linechart.js"));

var _lineoftext = _interopRequireDefault(__webpack_require__(/*! core/controls/lineoftext */ "./js/core/controls/lineoftext.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/controls/link */ "./js/core/controls/link.js"));

var _linkbar = _interopRequireDefault(__webpack_require__(/*! core/controls/linkbar */ "./js/core/controls/linkbar.js"));

var _list = _interopRequireDefault(__webpack_require__(/*! core/controls/list */ "./js/core/controls/list.js"));

var _menu = _interopRequireDefault(__webpack_require__(/*! core/controls/menu */ "./js/core/controls/menu.js"));

var _menubar = _interopRequireDefault(__webpack_require__(/*! core/controls/menubar */ "./js/core/controls/menubar.js"));

var _modalscreen = _interopRequireDefault(__webpack_require__(/*! core/controls/modalscreen */ "./js/core/controls/modalscreen.js"));

var _numericstepper = _interopRequireDefault(__webpack_require__(/*! core/controls/numericstepper */ "./js/core/controls/numericstepper.js"));

var _paragraph = _interopRequireDefault(__webpack_require__(/*! core/controls/paragraph */ "./js/core/controls/paragraph.js"));

var _piechart = _interopRequireDefault(__webpack_require__(/*! core/controls/piechart */ "./js/core/controls/piechart.js"));

var _playback = _interopRequireDefault(__webpack_require__(/*! core/controls/playback */ "./js/core/controls/playback.js"));

var _popover = _interopRequireDefault(__webpack_require__(/*! core/controls/popover */ "./js/core/controls/popover.js"));

var _progressbar = _interopRequireDefault(__webpack_require__(/*! core/controls/progressbar */ "./js/core/controls/progressbar.js"));

var _redx = _interopRequireDefault(__webpack_require__(/*! core/controls/redx */ "./js/core/controls/redx.js"));

var _rules = _interopRequireDefault(__webpack_require__(/*! core/controls/rules */ "./js/core/controls/rules.js"));

var _scratchout = _interopRequireDefault(__webpack_require__(/*! core/controls/scratchout */ "./js/core/controls/scratchout.js"));

var _scrollbarhorizontal = _interopRequireDefault(__webpack_require__(/*! core/controls/scrollbarhorizontal */ "./js/core/controls/scrollbarhorizontal.js"));

var _scrollbarvertical = _interopRequireDefault(__webpack_require__(/*! core/controls/scrollbarvertical */ "./js/core/controls/scrollbarvertical.js"));

var _searchbox = _interopRequireDefault(__webpack_require__(/*! core/controls/searchbox */ "./js/core/controls/searchbox.js"));

var _shape = _interopRequireDefault(__webpack_require__(/*! core/controls/shape */ "./js/core/controls/shape.js"));

var _sitemap = _interopRequireDefault(__webpack_require__(/*! core/controls/sitemap */ "./js/core/controls/sitemap.js"));

var _sliders = _interopRequireDefault(__webpack_require__(/*! core/controls/sliders */ "./js/core/controls/sliders.js"));

var _splitterhorizontal = _interopRequireDefault(__webpack_require__(/*! core/controls/splitterhorizontal */ "./js/core/controls/splitterhorizontal.js"));

var _splittervertical = _interopRequireDefault(__webpack_require__(/*! core/controls/splittervertical */ "./js/core/controls/splittervertical.js"));

var _stickynote = _interopRequireDefault(__webpack_require__(/*! core/controls/stickynote */ "./js/core/controls/stickynote.js"));

var _streetmap = _interopRequireDefault(__webpack_require__(/*! core/controls/streetmap */ "./js/core/controls/streetmap.js"));

var _switch = _interopRequireDefault(__webpack_require__(/*! core/controls/switch */ "./js/core/controls/switch.js"));

var _symbol = _interopRequireDefault(__webpack_require__(/*! core/controls/symbol */ "./js/core/controls/symbol.js"));

var _tabbars = _interopRequireDefault(__webpack_require__(/*! core/controls/tabbars */ "./js/core/controls/tabbars.js"));

var _tagcloud = _interopRequireDefault(__webpack_require__(/*! core/controls/tagcloud */ "./js/core/controls/tagcloud.js"));

var _textarea = _interopRequireDefault(__webpack_require__(/*! core/controls/textarea */ "./js/core/controls/textarea.js"));

var _textinput = _interopRequireDefault(__webpack_require__(/*! core/controls/textinput */ "./js/core/controls/textinput.js"));

var _titles = _interopRequireDefault(__webpack_require__(/*! core/controls/titles */ "./js/core/controls/titles.js"));

var _tooltip = _interopRequireDefault(__webpack_require__(/*! core/controls/tooltip */ "./js/core/controls/tooltip.js"));

var _tree = _interopRequireDefault(__webpack_require__(/*! core/controls/tree */ "./js/core/controls/tree.js"));

var _videoplayer = _interopRequireDefault(__webpack_require__(/*! core/controls/videoplayer */ "./js/core/controls/videoplayer.js"));

var _volumeslider = _interopRequireDefault(__webpack_require__(/*! core/controls/volumeslider */ "./js/core/controls/volumeslider.js"));

var _webcam = _interopRequireDefault(__webpack_require__(/*! core/controls/webcam */ "./js/core/controls/webcam.js"));

var _window = _interopRequireDefault(__webpack_require__(/*! core/controls/window */ "./js/core/controls/window.js"));

var _timepicker = _interopRequireDefault(__webpack_require__(/*! core/controls/timepicker */ "./js/core/controls/timepicker.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/***/ }),

/***/ "./js/core/controls/android.js":
/*!*************************************!*\
  !*** ./js/core/controls/android.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _snapping = __webpack_require__(/*! core/controls/snapping */ "./js/core/controls/snapping.js");

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

let controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

let textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

let control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::AndroidPhone",
  controlTypeName: "Smartphone",
  controlAlternateNames: ["Android", "Nexus"],
  controlInitialProperties: {
    resizing: Props.iPhoneResizing.STRETCH
  },
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        resize: Props.ResizeOptions.BOTH,
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT
      }
    }),
    orientation: {
      defaultValue: Props.Orientations.PORTRAIT,
      enum: Props.Orientations
    },
    topBar: {
      defaultValue: true,
      dataType: 'bool'
    },
    bgTransparent: {
      defaultValue: false,
      dataType: 'bool'
    },
    resizing: {
      defaultValue: Props.iPhoneResizing.MAINTAIN_PROPORTIONS,
      enum: Props.iPhoneResizing
    }
  }
});

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  let size = _types.default.newSize(356, 725);

  if (this.getActualProperty("orientation") === Props.Orientations.LANDSCAPE) {
    let t = size.width;
    size.width = size.height;
    size.height = t;
  }

  return size;
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  let stretch = this.getActualProperty("resizing") == Props.iPhoneResizing.STRETCH;
  let transparent = this.getActualProperty("bgTransparent");
  let landscape = this.getActualProperty("orientation") === Props.Orientations.LANDSCAPE;
  let skinPadding = this._context.skin["#skinName"] === "sketch" ? 4 : 0;
  let nativeSize = this.measure();
  let nativeRatio = nativeSize.width / nativeSize.height;
  let controlRatio = p_localBbox.width / p_localBbox.height;
  this.androidSnapRect = null; // compute the bounds of the android ignoring the adjustment required by the rotation (landscape)

  let androidBounds = _types.default.newRectangle(0, 0, p_localBbox.width, p_localBbox.height);

  if (!stretch) {
    if (controlRatio > nativeRatio) {
      androidBounds.width = p_localBbox.height * nativeRatio;
      androidBounds.height = p_localBbox.height;
      androidBounds.x = (p_localBbox.width - androidBounds.width) / 2;
      androidBounds.y = 0;
    } else {
      androidBounds.width = p_localBbox.width;
      androidBounds.height = p_localBbox.width / nativeRatio;
      androidBounds.x = 0;
      androidBounds.y = (p_localBbox.height - androidBounds.height) / 2;
    }
  }

  let zoomRatio = stretch ? 1 : androidBounds.width / nativeSize.width;
  let left;
  let right;
  let top;
  let bottom;
  let topBarHeight = zoomRatio * (16 + skinPadding);

  if (!stretch) {
    if (landscape) {
      left = androidBounds.x + zoomRatio * 79;
      right = p_localBbox.width - androidBounds.x - androidBounds.width + zoomRatio * 80;
      top = androidBounds.y + zoomRatio * 19;
      bottom = p_localBbox.height - androidBounds.y - androidBounds.height + zoomRatio * 19;
    } else {
      left = androidBounds.x + zoomRatio * 19;
      right = p_localBbox.width - androidBounds.x - androidBounds.width + zoomRatio * 19;
      top = androidBounds.y + zoomRatio * (79 + skinPadding);
      bottom = p_localBbox.height - androidBounds.y - androidBounds.height + zoomRatio * 80;
    }

    let top2 = top + (this.getActualProperty("topBar") ? topBarHeight : 0); // background pattern

    if (!transparent) {
      this.pushToDisplayList((0, _rectangle.makeFilledRectangle)(_types.default.newRectangle(left, top2, p_localBbox.width - left - right, p_localBbox.height - top2 - bottom), 0xFFFFFF));
    }

    this.androidSnapRect = _types.default.newRectangle(left, top2, Math.max(0, p_localBbox.width - right), Math.max(0, p_localBbox.height - bottom)); // top bar

    if (this.getActualProperty("topBar")) {
      let topBarItemsColor = 0xFFFFFF;

      let topBarBbox = _types.default.newRectangle(left, top, p_localBbox.width - left - right, topBarHeight);

      this.pushToDisplayList((0, _rectangle.makeFilledRectangle)(topBarBbox, 0xCCCCCC));
      let x = p_localBbox.width - right - 5;

      if (topBarBbox) {
        let date = new Date();
        let textStyle = this.newTextStyle();
        textStyle.fontSize = Math.min(12, topBarBbox.height);
        textStyle.textColor = topBarItemsColor;
        let timeText = textMod.newText(date.getHours() + ":" + (date.getMinutes() < 10 ? "0" : "") + date.getMinutes(), textStyle, this._context);
        x -= timeText.measuredSize().width;
        this.pushToDisplayList(timeText.makeDisplayItems(_types.default.newRectangle(x, top + (topBarHeight - timeText.measuredSize().height) / 2 + 1, timeText.measuredSize().width, timeText.measuredSize().height)));
      }

      let iconSize = topBarHeight - 4;

      if (iconSize > 0) {
        let y = top + (topBarHeight - iconSize) / 2;

        let batteryData = _icon.default.newAppIconData("battery-full", iconSize, 90);

        x -= iconSize + 2; //

        this.pushToDisplayList(_icon.default.makeDisplayItemsFromIconData(this, batteryData, x, y, 0xFFFFFF));

        let signalData = _icon.default.newAppIconData("signal", iconSize, 0);

        x -= iconSize + 3; //

        this.pushToDisplayList(_icon.default.makeDisplayItemsFromIconData(this, signalData, x, y, 0xFFFFFF));

        let wifiData = _icon.default.newAppIconData("wifi", iconSize + 2, 0);

        x -= iconSize + 4; //

        this.pushToDisplayList(_icon.default.makeDisplayItemsFromIconData(this, wifiData, x, top + (topBarHeight - iconSize - 2) / 2, 0xFFFFFF));
      }
    } // now adjust the android bounds for the rotation (landscape)


    if (landscape) {
      let t = androidBounds.width;
      androidBounds.width = androidBounds.height;
      androidBounds.height = t;
      androidBounds.y = p_localBbox.height - androidBounds.y;
    }

    this.pushToDisplayList(_displayitem.default.newPathsAsset(this._context, "android_nexus", androidBounds, null, 1, landscape ? 90 : 0));
  } else {
    let buttonSize = this.style("android_nexus_button");
    let earThingSize = this.style("android_nexus_ear_thing");
    let button, earThing;
    let android;
    let smallScreen = false;

    if (landscape) {
      if (p_localBbox.height > 70) {
        top = 19;
        bottom = 19;
      } else {
        top = bottom = 1;
        smallScreen = true;
      }

      if (p_localBbox.width > 167) {
        left = 79;
        right = 80;
      } else {
        left = right = 25;
        smallScreen = true;
      }

      android = _displayitem.default.newRotated90CCWSlice9("android_nexus_scale9", p_localBbox, this._context, null);

      if (p_localBbox.height - top - bottom > buttonSize.width && left > buttonSize.height && !smallScreen) {
        button = _displayitem.default.newPathsAsset(this._context, "android_nexus_button", _types.default.newRectangle(p_localBbox.width - (right + buttonSize.height) / 2 + 1, (p_localBbox.height + buttonSize.width) / 2, buttonSize.width, buttonSize.height), 0, 1, 90);
      }

      if (p_localBbox.height - top - bottom - 22 > earThingSize.width && right > earThingSize.height && !smallScreen) {
        earThing = _displayitem.default.newPathsAsset(this._context, "android_nexus_ear_thing", _types.default.newRectangle((left - earThingSize.height) / 2 - 3, (p_localBbox.height + earThingSize.width) / 2 + 22, earThingSize.width, earThingSize.height), 0, 1, 90);
      }
    } else {
      if (p_localBbox.width > 70) {
        left = 19;
        right = 19;
      } else {
        left = right = 1;
        smallScreen = true;
      }

      if (p_localBbox.height > 167) {
        top = 79 + skinPadding;
        bottom = 80;
      } else {
        top = bottom = 25;
        smallScreen = true;
      }

      android = _displayitem.default.createDisplayItemsFromSkin("android_nexus_scale9", null, p_localBbox, this._context);

      if (p_localBbox.width - right - left > buttonSize.width && bottom > buttonSize.height && !smallScreen) {
        button = _displayitem.default.newPathsAsset(this._context, "android_nexus_button", _types.default.newRectangle((p_localBbox.width - buttonSize.width) / 2 + 1, p_localBbox.height - (bottom + buttonSize.height) / 2 + 1, buttonSize.width, buttonSize.height), 0);
      }

      if (p_localBbox.width - right - left - 21 > earThingSize.width && top > earThingSize.height && !smallScreen) {
        earThing = _displayitem.default.newPathsAsset(this._context, "android_nexus_ear_thing", _types.default.newRectangle((p_localBbox.width - earThingSize.width) / 2 - 21, (top - earThingSize.height) / 2 - 3, earThingSize.width, earThingSize.height), 0);
      }
    }

    let top2 = top + (this.getActualProperty("topBar") ? topBarHeight : 0); // background pattern

    if (!transparent) {
      this.pushToDisplayList((0, _rectangle.makeFilledRectangle)(_types.default.newRectangle(left, top, Math.max(p_localBbox.width - left - right, 1), Math.max(p_localBbox.height - bottom - top, 1)), 0xFFFFFF));
    }

    if (this.getActualProperty("topBar") && !smallScreen) {
      let topBarItemsColor = 0xFFFFFF;

      let topBarBbox = _types.default.newRectangle(left, top, p_localBbox.width - left - right, topBarHeight);

      this.pushToDisplayList((0, _rectangle.makeFilledRectangle)(topBarBbox, 0xCCCCCC));
      let x = p_localBbox.width - right - 5;
      let date = new Date();
      let textStyle = this.newTextStyle();
      textStyle.fontSize = Math.min(12, topBarBbox.height);
      textStyle.textColor = topBarItemsColor;
      let timeText = textMod.newText(date.getHours() + ":" + (date.getMinutes() < 10 ? "0" : "") + date.getMinutes(), textStyle, this._context);
      x -= timeText.measuredSize().width;

      if (x > left && !smallScreen) {
        this.pushToDisplayList(timeText.makeDisplayItems(_types.default.newRectangle(x, top + (topBarHeight - timeText.measuredSize().height) / 2 + 1, timeText.measuredSize().width, timeText.measuredSize().height)));
        let iconSize = topBarHeight - 4; // if (iconSize > 0)
        // {

        let y = top + (topBarHeight - iconSize) / 2;

        let batteryData = _icon.default.newAppIconData("battery-full", iconSize, 90);

        x -= iconSize + 2; //

        if (x > left) {
          this.pushToDisplayList(_icon.default.makeDisplayItemsFromIconData(this, batteryData, x, y, 0xFFFFFF));

          let signalData = _icon.default.newAppIconData("signal", iconSize, 0);

          x -= iconSize + 3; //

          if (x > left) {
            this.pushToDisplayList(_icon.default.makeDisplayItemsFromIconData(this, signalData, x, y, 0xFFFFFF));

            let wifiData = _icon.default.newAppIconData("wifi", iconSize + 2, 0);

            x -= iconSize + 4; //

            if (x > left) {
              this.pushToDisplayList(_icon.default.makeDisplayItemsFromIconData(this, wifiData, x, top + (topBarHeight - iconSize - 2) / 2, 0xFFFFFF));
            }
          }
        }
      }
    }

    this.pushToDisplayList(android, button, earThing);
  }
};

control.CONTROL_PROTOTYPE.getSnapLines = function (withMarkup) {
  let snapLines = (0, _snapping.makeSnapLinesForBbox)(this.getUserBBox(true), false);

  if (this.androidSnapRect) {
    snapLines.push(new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.VERTICAL, this.androidSnapRect.x), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.VERTICAL, this.androidSnapRect.width), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.HORIZONTAL, this.androidSnapRect.y), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.HORIZONTAL, this.androidSnapRect.height));
  }

  return snapLines;
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/appbar.js":
/*!************************************!*\
  !*** ./js/core/controls/appbar.js ***!
  \************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const ICON_SIZE = 16;
const control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::AppBar",
  controlTypeName: "App Bar",
  controlAlternateNames: ["Top Bar"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    filterIcon: {
      defaultValue: true,
      dataType: 'bool'
    },
    menuIcon: {
      defaultValue: true,
      dataType: 'bool'
    },
    navIcon: {
      defaultValue: true,
      dataType: 'bool'
    },
    ...Props.PropertyDefinitions.ICONS({
      doNotUpdateOnTextChange: true
    }),
    //we don't want bold/italic/underline in the inspector
    ...Props.PropertyDefinitions.TEXT({
      defaultText: "Heading"
    }),
    fontSize: {
      defaultValue: 16,
      dataType: 'int'
    },
    ...Props.PropertyDefinitions.links({
      doNotUpdateOnTextChange: true
    })
  }
});
Controls.addColorPropertyDefinitions(control, {
  defaultColor: "appbar_background_color"
});

control.CONTROL_PROTOTYPE.getIconsID = function () {
  return this.iconsForInspector;
};

control.CONTROL_PROTOTYPE.getLinkNames = function () {
  return this.iconsForInspector.map(icon => icon.src.resourceID);
};

control.CONTROL_PROTOTYPE.updateData = function () {
  const text = this.getActualProperty("text");
  const textStyle = this.newStandardTextStyle({
    fontSize: this.hasPropertyValue("fontSize") ? this.getActualProperty("fontSize") : this.style("appbar_font_size"),
    textColor: _colors.default.labelColorGivenBackgroundColor(this.getActualProperty("color")),
    bold: false,
    italic: false,
    underline: false
  });

  if (!this.text || !Text.areTextsEqual(this.text, text, textStyle)) {
    this.text = Text.newTextNoWrap(text, textStyle, this._context);
  }

  this.iconsForInspector = [_icon.default.newAppIconData("bars", ICON_SIZE, 0), _icon.default.newAppIconData("caret-down", ICON_SIZE, 0)];
  const icons = this.getActualProperty("icons") || [];
  const dependencies = [];

  for (const icon of icons) {
    dependencies.push(icon.src);

    const {
      iconData
    } = _icon.default.processIconProperty(icon, 0);

    this.iconsForInspector.push(iconData);
  }

  this.iconsForInspector.push(_icon.default.newAppIconData("ellipsis-v", ICON_SIZE, 0));
  this.setDependencies(dependencies);
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  const iconPadding = this.style("appbar_icon_padding");
  const textSize = this.text.measuredSize();
  const height = Math.max(44, textSize.height + this.style("appbar_padding_bottom") + 1);
  const navIconWidth = this.getActualProperty("navIcon") ? ICON_SIZE + 2 * iconPadding : 0;
  const filterIconWidth = this.getActualProperty("filterIcon") ? ICON_SIZE + iconPadding : 0;
  const menuIconWidth = this.getActualProperty("menuIcon") ? ICON_SIZE + iconPadding : 0;
  const textWidth = this.getActualProperty("text").length > 0 ? textSize.width + iconPadding / 2 : 0;
  const icons = this.getActualProperty("icons");
  const iconsWidth = icons && icons.length > 0 ? icons.length * (ICON_SIZE + iconPadding) : 0;
  const width = this.style("appbar_padding_left") + navIconWidth + filterIconWidth + menuIconWidth + textWidth + iconsWidth;
  return _types.default.newSize(Math.max(320, Math.trunc(width)), height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var iconPadding = this.style("appbar_icon_padding");
  var paddingBottom = this.style("appbar_padding_bottom");
  var paddingTop = this.style("appbar_icon_top_padding");
  this.pushToDisplayList(Rectangle.makeFilledRectangle(p_localBbox, this.getActualProperty("color")));

  var color = _colors.default.labelColorGivenBackgroundColor(this.getActualProperty("color"));

  var theX = this.style("appbar_padding_left");
  let linkBounds = [];
  let bounds;

  if (this.getActualProperty("navIcon") && theX + ICON_SIZE < p_localBbox.width && paddingTop + ICON_SIZE < p_localBbox.height) {
    bounds = _types.default.newRectangle(Math.max(0, theX - 2), paddingTop - 2, ICON_SIZE + 4, ICON_SIZE + 4);
    linkBounds.push(bounds);

    var navIconData = _icon.default.newAppIconData("bars", ICON_SIZE, 0);

    var navIcon = _icon.default.makeDisplayItemsFromIconData(this, navIconData, theX, paddingTop, color);

    theX += ICON_SIZE + 2 * iconPadding;
  } else {
    linkBounds.push(null);
  }

  var textSize = this.text.measuredSize();
  var label = null;

  var labelBbox = _types.default.newRectangle(theX, p_localBbox.height - paddingBottom - textSize.height, textSize.width, textSize.height);

  var filterX = theX + textSize.width + iconPadding / 2;

  if (p_localBbox.height - paddingBottom - textSize.height > 0) {
    label = this.text.makeDisplayItems(labelBbox);
    theX += textSize.width + iconPadding;
  }

  var filterY = this.getActualProperty("text").length > 0 ? labelBbox.y + (textSize.height - ICON_SIZE) / 2 : p_localBbox.height - ICON_SIZE - paddingBottom;

  if (this.getActualProperty("filterIcon") && theX + ICON_SIZE < p_localBbox.width && filterY > 0) {
    bounds = _types.default.newRectangle(Math.max(0, filterX - 2), Math.max(0, filterY - 2), ICON_SIZE + 4, ICON_SIZE + 4);
    linkBounds.push(bounds);

    var filterIconData = _icon.default.newAppIconData("caret-down", ICON_SIZE, 0);

    var filterIcon = _icon.default.makeDisplayItemsFromIconData(this, filterIconData, filterX, filterY, color);

    theX += ICON_SIZE;
  } else {
    linkBounds.push(null);
  }

  let drawMenuIcon = this.getActualProperty("menuIcon") && theX + ICON_SIZE < p_localBbox.width && paddingTop + ICON_SIZE < p_localBbox.height;

  if (drawMenuIcon) {
    var menuIconData = _icon.default.newAppIconData("ellipsis-v", ICON_SIZE, 0);

    var menuIcon = _icon.default.makeDisplayItemsFromIconData(this, menuIconData, p_localBbox.width - iconPadding - ICON_SIZE, paddingTop, color);
  }

  var icons = this.getActualProperty("icons");
  var oldX = theX;

  if (icons) {
    var iconItems = [];
    theX = p_localBbox.width - ((menuIcon ? 1 : 0) + icons.length) * (iconPadding + ICON_SIZE);
    oldX -= !filterIcon ? iconPadding : 0;
    icons.forEach(icon => {
      icon.size = ICON_SIZE;

      var iconProp = _icon.default.processIconProperty(icon, 0);

      if (oldX < theX && paddingTop + ICON_SIZE < p_localBbox.height) {
        iconItems.push(_icon.default.makeDisplayItemsFromIconData(this, iconProp.iconData, theX, paddingTop, color));
        bounds = _types.default.newRectangle(Math.max(0, theX - 2), paddingTop - 2, ICON_SIZE + 4, ICON_SIZE + 4);
        linkBounds.push(bounds);
      } else {
        linkBounds.push(null);
      }

      theX += ICON_SIZE + iconPadding;
    });
  }

  if (drawMenuIcon) {
    bounds = _types.default.newRectangle(Math.max(0, p_localBbox.width - iconPadding - ICON_SIZE - 2), paddingTop - 2, ICON_SIZE + 4, ICON_SIZE + 4);
    linkBounds.push(bounds);
  }

  _link.default.processLinks(this, linkBounds);

  this.pushToDisplayList(filterIcon, label, menuIcon, navIcon, iconItems);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/arrow.js":
/*!***********************************!*\
  !*** ./js/core/controls/arrow.js ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var _overlaylabel = __webpack_require__(/*! core/bricks/overlaylabel */ "./js/core/bricks/overlaylabel.js");

var Geom = _interopRequireWildcard(__webpack_require__(/*! core/utils/geom */ "./js/core/utils/geom.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var MockupCoords = _interopRequireWildcard(__webpack_require__(/*! core/viewcontrollers/mockupcoordinates */ "./js/core/viewcontrollers/mockupcoordinates.js"));

var MockupSize = _interopRequireWildcard(__webpack_require__(/*! core/utils/mockupsize */ "./js/core/utils/mockupsize.js"));

var Constants = _interopRequireWildcard(__webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts"));

var _native = __webpack_require__(/*! core/native */ "./js/core/native.js");

var _selectionutils = __webpack_require__(/*! core/utils/selectionutils */ "./js/core/utils/selectionutils.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var snapTo45 = function (p0, p2, handleID, newPosition) {
  var newPos = _balsamiq.default.clone(newPosition);

  let movePt = newPos;
  let centerPt = handleID === _selectionutils.HandleTypes.P0 ? p2 : p0; // Fix angle because bottom quadrants get negative degrees
  //        90
  //
  // 0           180/-180
  //
  //        -90

  var angleDeg = Math.atan2(centerPt.y - movePt.y, centerPt.x - movePt.x) * 180 / Math.PI;

  if (angleDeg < 0) {
    angleDeg = 360 + angleDeg;
  } //console.debug(angleDeg);


  let newDelta = Math.min(Math.abs(newPos.y - centerPt.y), Math.abs(newPos.x - centerPt.x));
  var STEPS = [22.5, 67.5, 112.5, 157.5, 202.5, 247.5, 292.5, 337.5]; // Each quadrant has its own projection rule
  // top half goes from 0 to 180 deg
  // bottom half goes from 0 to -180

  if (angleDeg < STEPS[0] || angleDeg >= STEPS[7]) {
    //console.debug('LEFT');
    newPos.y = centerPt.y;
  } else if (angleDeg >= STEPS[0] && angleDeg < STEPS[1]) {
    //console.debug('TOP LEFT');
    newPos.y = centerPt.y - newDelta;
    newPos.x = centerPt.x - newDelta;
  } else if (angleDeg >= STEPS[1] && angleDeg < STEPS[2]) {
    //console.debug('TOP');
    newPos.x = centerPt.x;
  } else if (angleDeg >= STEPS[2] && angleDeg < STEPS[3]) {
    //console.debug('TOP RIGHT');
    newPos.y = centerPt.y - newDelta;
    newPos.x = centerPt.x + newDelta;
  } else if (angleDeg >= STEPS[3] && angleDeg < STEPS[4]) {
    //console.debug('RIGHT');
    newPos.y = centerPt.y;
  } else if (angleDeg >= STEPS[4] && angleDeg < STEPS[5]) {
    //console.debug('BOTTOM RIGHT');
    newPos.y = centerPt.y + newDelta;
    newPos.x = centerPt.x + newDelta;
  } else if (angleDeg >= STEPS[5] && angleDeg < STEPS[6]) {
    //console.debug('BOTTOM');
    newPos.x = centerPt.x;
  } else if (angleDeg >= STEPS[6] && angleDeg < STEPS[7]) {
    //console.debug('BOTTOM LEFT');
    newPos.y = centerPt.y + newDelta;
    newPos.x = centerPt.x - newDelta;
  }

  return newPos;
};

var control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::Arrow",
  controlTypeName: "Arrow",
  controlAlternateNames: ["Line"],
  controlInitialProperties: {
    p0: {
      x: 0,
      y: 0
    },
    p1: {
      x: 0.5,
      y: 0
    },
    p2: {
      x: 150,
      y: 100
    },
    shape: 'bezier',
    curvature: 0,
    direction: _types.default.Positions.TOP
  },
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(150, 100),
      options: {
        resize: Props.ResizeOptions.BOTH,
        autosize: Props.AutoSizeOptions.NO
      }
    }),
    // New props
    p0: {
      defaultValue: null,
      dataType: 'object'
    },
    p1: {
      defaultValue: null,
      dataType: 'object'
    },
    p2: {
      defaultValue: null,
      dataType: 'object'
    },
    shape: {
      defaultValue: Props.ArrowShapes.BEZIER,
      enum: Props.ArrowShapes
    },
    labelPosition: {
      defaultValue: function () {
        let hasPointLeft = this.getActualProperty("leftArrow");
        let hasPointRight = this.getActualProperty("rightArrow");
        let labelOffset;

        if (hasPointLeft && hasPointRight || !hasPointLeft && !hasPointRight) {
          labelOffset = 5;
        } else if (hasPointLeft) {
          labelOffset = 8;
        } else {
          labelOffset = 2;
        }

        return labelOffset;
      },
      dataType: 'int'
    },
    // Old props
    leftArrow: {
      defaultValue: true,
      dataType: 'bool'
    },
    rightArrow: {
      defaultValue: true,
      dataType: 'bool'
    },
    stroke: {
      defaultValue: Props.StrokesBasic.SOLID,
      enum: Props.StrokesBasic
    },
    curvature: {
      defaultValue: Props.ArrowCurvatures.CONCAVE,
      enum: Props.ArrowCurvatures
    },
    direction: {
      defaultValue: Props.Positions.TOP,
      enum: Props.Positions
    },
    ...Props.PropertyDefinitions.TEXT({
      defaultText: "",
      multiline: false
    }),
    markup: {
      defaultValue: true,
      dataType: 'bool'
    },
    ...Props.PropertyDefinitions.QUICKDRAW_TIP("You can draw Arrows quickly by holding down the *A key* and dragging. [Learn more…]($DOCURL)", "You can also draw Arrows with your fingers using the Quick Draw Toolbar (in the View Menu). [Learn more…]($DOCURL)")
  }
});
Controls.addStandardTextPropertyDefinitions(control);
Controls.addColorPropertyDefinitions(control, {
  defaultColor: "arrow_default_color",
  hasAlpha: true
});

control.CONTROL_PROTOTYPE.initData = function () {
  this._clipOnBbox = false;
  this.text = null;
};

let loadOldPoints = function (p_w, p_h, p_curvature, p_isTopLeft) {
  p_w = p_w <= 10 ? 1 : p_w;
  p_h = p_h <= 10 ? 1 : p_h;

  if (p_w == 1 || p_h == 1) {
    p_curvature = 0;
  }

  let p0, p2, p1;

  if (p_isTopLeft) {
    p0 = _types.default.newPoint(0, 0);
    p2 = _types.default.newPoint(p_w, p_h);

    if (p_curvature == 0) {
      p1 = _types.default.newPoint((p0.x + p2.x) / 2, (p0.y + p2.y) / 2);
    } else if (p_curvature == 1) {
      p1 = _types.default.newPoint(p_w * 2 / 3, p_h / 3);
    } else if (p_curvature == -1) {
      p1 = _types.default.newPoint(p_w / 3, p_h * 2 / 3);
    }
  } else {
    p0 = _types.default.newPoint(0, p_h);
    p2 = _types.default.newPoint(p_w, 0);

    if (p_curvature == 0) {
      p1 = _types.default.newPoint((p0.x + p2.x) / 2, (p0.y + p2.y) / 2);
    } else if (p_curvature == 1) {
      p1 = _types.default.newPoint(p_w / 3, p_h / 3);
    } else if (p_curvature == -1) {
      p1 = _types.default.newPoint(p_w * 2 / 3, p_h * 2 / 3);
    }
  }

  return {
    p0,
    p1,
    p2
  };
};

_link.default.addTextLinksMethods(control);

control.CONTROL_PROTOTYPE.updateClipBBoxPS = function () {
  this._displayData = this.createDisplayItemsAndLinks(_types.default.newRectangle(0, 0, this._userBBox.width, this._userBBox.height));
  var unionBBox = this._userBBox;

  if (this._displayData.pointLeftBBox) {
    this._displayData.pointLeftBBox.x += this._userBBox.x;
    this._displayData.pointLeftBBox.y += this._userBBox.y;
    unionBBox = _types.default.rectanglesUnion(this._displayData.pointLeftBBox, unionBBox);
  }

  if (this._displayData.pointRightBBox) {
    this._displayData.pointRightBBox.x += this._userBBox.x;
    this._displayData.pointRightBBox.y += this._userBBox.y;
    unionBBox = _types.default.rectanglesUnion(this._displayData.pointRightBBox, unionBBox);
  }

  this._displayData.arrowClipBBox.x += this._userBBox.x;
  this._displayData.arrowClipBBox.y += this._userBBox.y;
  unionBBox = _types.default.rectanglesUnion(this._displayData.arrowClipBBox, unionBBox);

  if (this._displayData.labelClipBBox) {
    this._displayData.labelClipBBox.x += this._userBBox.x;
    this._displayData.labelClipBBox.y += this._userBBox.y;
    unionBBox = _types.default.rectanglesUnion(this._displayData.labelClipBBox, unionBBox);
  }

  this.setClipBBoxPS(unionBBox.x, unionBBox.y, unionBBox.width, unionBBox.height);
};

control.CONTROL_PROTOTYPE.createDisplayItemsAndLinks = function (p_localBbox) {
  var res = {
    displayItems: [],
    links: {},
    pointLeftBBox: null,
    pointRightBBox: null,
    arrowClipBBox: null,
    labelClipBBox: null,
    p0: null,
    pc1: null,
    p2: null,
    arrowPath: null
  };
  var w = p_localBbox.width,
      h = p_localBbox.height;
  var color = this.getActualProperty("color");
  var thickness = this.style("arrow_thickness");
  var thicknessDot = this.style("arrow_thickness_dot");
  var alpha = this.getActualProperty("backgroundAlpha");
  var stroke = this.getActualProperty("stroke");
  var hasPointLeft = this.getActualProperty("leftArrow");
  var hasPointRight = this.getActualProperty("rightArrow"); // get new arrow's props

  let isBezier = this.getActualProperty('shape') != Props.ArrowShapes.LINE;
  let p0 = this.getActualProperty('p0');
  let p2 = this.getActualProperty('p2');
  let p1s = this.getActualProperty('p1');
  let p1 = null; // Load old arrow if needed

  if (p0 === null || p2 === null || p1s === null) {
    let curvature = this.getActualProperty("curvature");

    let isTopLeft = this.getActualProperty("direction") === _types.default.Positions.TOP;

    ({
      p0,
      p1,
      p2
    } = loadOldPoints(w, h, curvature, isTopLeft));
  } else {
    [p0, p2, p1s] = [p0, p2, p1s].map(p => _types.default.newPoint(p.x, p.y));
    p1 = Geom.fromRelativePos(p0, p2, p1s);
  } // compute bbox and choose pc1


  let bBox;
  let pc1;

  if (isBezier) {
    pc1 = Geom.getBezierControlPoint(p0, p2, p1);
    bBox = Geom.getBezierBoundingBox(p0, pc1, p2);
  } else {
    pc1 = p1;
    bBox = Geom.getBoundingBox(p0, p1, p2);
  } // Clip bbox for arrow considers thickness


  res.arrowClipBBox = _types.default.newRectangle(bBox.x - thickness / 2, bBox.y - thickness / 2, bBox.width + thickness, bBox.height + thickness);
  const topLeft = Geom.Rectangle.topLeft(bBox);
  var setAfter = true;

  if (this.parent) {
    if (isBezier) {
      var p1ms = MockupCoords.controlSpaceToMockupSpace(topLeft, this);
      var right = p1ms.x + bBox.width;
      var bottom = p1ms.y + bBox.height;
      if (right > Constants.MaxMockupSize) topLeft.x += right - Constants.MaxMockupSize;
      if (bottom > Constants.MaxMockupSize) topLeft.y += bottom - Constants.MaxMockupSize;
    }

    if (!isBezier || p1ms.x >= 0 && right < Constants.MaxMockupSize && p1ms.y >= 0 && bottom < Constants.MaxMockupSize) {
      res.p0 = p0;
      res.p1 = p1;
      res.pc1 = pc1;
      res.p2 = p2;
      setAfter = false;
    }
  }

  p0 = Geom.Point.sub(p0, topLeft);
  p1 = Geom.Point.sub(p1, topLeft);
  p2 = Geom.Point.sub(p2, topLeft);
  pc1 = Geom.Point.sub(pc1, topLeft);

  if (setAfter) {
    // set it after moving the points, so the control point handles move accordingly
    res.p0 = p0;
    res.p1 = p1;
    res.pc1 = pc1;
    res.p2 = p2;
  } // Arrow line


  let arrowPath = Path.newPath(p_localBbox.width, p_localBbox.height);
  arrowPath.moveTo(p0.x, p0.y);

  if (isBezier) {
    arrowPath.curveTo(pc1.x, pc1.y, p2.x, p2.y);
  } else {
    arrowPath.lineTo(p1.x, p1.y);
    arrowPath.lineTo(p2.x, p2.y);
  }

  let dashes = null;

  if (stroke === _types.default.Stroke.DASHED) {
    dashes = [thickness * 3, thickness * 6];
  } else if (stroke === _types.default.Stroke.DOTTED) {
    dashes = [thicknessDot, thicknessDot * 8];
  }

  let arrow = _displayitem.default.newSVGPath(p_localBbox, arrowPath, thickness, color, null);

  arrow.linecap = _displayitem.default.LineCap.ROUND;
  arrow.borderAlpha = alpha;
  arrow.dashArray = dashes;
  res.displayItems.push(arrow);
  res.arrowPath = arrowPath; // Arrow heads

  let pointSize = this.styleSVGSize("arrow_point");
  let [pointLeft, pointRight] = [null, null]; //eslint-disable-next-line

  let [pointLeftRect, pointRightRect] = [null, null];

  let arrowPointOrigin = _types.default.newPoint(this.style('arrow_point_origin_x'), this.style('arrow_point_origin_y'));

  if (hasPointLeft) {
    const rotL = Math.atan2(pc1.x - p0.x, pc1.y - p0.y) * 180 / Math.PI - 45;

    const pointLeftBounds = _types.default.newRectangle(p0.x, p0.y, pointSize.width, pointSize.height);

    pointLeft = _displayitem.default.newPathsAsset(this._context, "arrow_point", pointLeftBounds, color, alpha, rotL); // Rotate bounds

    let rotatedPointLeftBox = Geom.Rectangle.toPath(pointLeftBounds);
    rotatedPointLeftBox = rotatedPointLeftBox.map(p => Geom.Point.sub(p, Geom.Point.add(p0, arrowPointOrigin))).map(p => Geom.Point.rot(p, -rotL)).map(p => Geom.Point.add(p, p0)); // Build path for rotated bounds

    let rotatePointLeftBoxBounds = Geom.getBoundingBox(...rotatedPointLeftBox);
    res.pointLeftBBox = rotatePointLeftBoxBounds;
    let rotatePath = new Path.newPath(rotatePointLeftBoxBounds.width, rotatePointLeftBoxBounds.height);
    rotatePath.moveTo(rotatedPointLeftBox[0].x - rotatePointLeftBoxBounds.x, rotatedPointLeftBox[0].y - rotatePointLeftBoxBounds.y);

    for (let i = 1; i < 4; i++) {
      rotatePath.lineTo(rotatedPointLeftBox[i].x - rotatePointLeftBoxBounds.x, rotatedPointLeftBox[i].y - rotatePointLeftBoxBounds.y);
    }

    rotatePath.close();
    pointLeftRect = _displayitem.default.newSVGPath(rotatePointLeftBoxBounds, rotatePath, 1, 0, null);
  }

  if (hasPointRight) {
    let rotR = Math.atan2(pc1.x - p2.x, pc1.y - p2.y) * 180 / Math.PI - 45;

    let pointRightBounds = _types.default.newRectangle(p2.x, p2.y, pointSize.width, pointSize.height);

    pointRight = _displayitem.default.newPathsAsset(this._context, "arrow_point", pointRightBounds, color, alpha, rotR); // Rotate bounds

    let rotatedPointRightBox = Geom.Rectangle.toPath(pointRightBounds);
    rotatedPointRightBox = rotatedPointRightBox.map(p => Geom.Point.sub(p, Geom.Point.add(p2, arrowPointOrigin))).map(p => Geom.Point.rot(p, -rotR)).map(p => Geom.Point.add(p, p2)); // Build path for rotated bounds

    let rotatePointRightBoxBounds = Geom.getBoundingBox(...rotatedPointRightBox);
    res.pointRightBBox = rotatePointRightBoxBounds;
    let rotatePath = new Path.newPath(rotatePointRightBoxBounds.width, rotatePointRightBoxBounds.height);
    rotatePath.moveTo(rotatedPointRightBox[0].x - rotatePointRightBoxBounds.x, rotatedPointRightBox[0].y - rotatePointRightBoxBounds.y);

    for (let i = 1; i < 4; i++) {
      rotatePath.lineTo(rotatedPointRightBox[i].x - rotatePointRightBoxBounds.x, rotatedPointRightBox[i].y - rotatePointRightBoxBounds.y);
    }

    rotatePath.close();
    pointRightRect = _displayitem.default.newSVGPath(rotatePointRightBoxBounds, rotatePath, 1, 0, null);
  }

  res.displayItems.push(pointLeft, pointRight
  /*, pointLeftRect, pointRightRect*/
  );
  this.text = null;
  let text = this.getActualProperty("text");

  if (text) {
    let labelOffset = this.getActualProperty('labelPosition');
    let labelPos;

    if (isBezier) {
      labelPos = Geom.evaluateBezier(p0, pc1, p2, labelOffset / 10.0);
    } else {
      labelPos = Geom.evaluateSegments(p0, p1, p2, labelOffset / 10.0);
    }

    let textStyle = this.newStandardTextStyle({
      textColor: this.getActualProperty("color")
    });
    let obj = (0, _overlaylabel.makeOverlayLabel)(this, labelPos.x, labelPos.y, text, alpha, textStyle);
    res.displayItems = res.displayItems.concat(obj.displayItems);
    res.links = obj.links;
    this.text = obj.text;
    res.labelClipBBox = _types.default.newRectangle(0, 0, 0, 0);

    for (const items of obj.displayItems) {
      for (const item of items) {
        const {
          x,
          y,
          width,
          height
        } = item;

        const rectangle = _types.default.newRectangle(x, y, width, height);

        res.labelClipBBox = _types.default.rectanglesUnion(rectangle, res.labelClipBBox);
      }
    }
  }

  return res;
};

control.CONTROL_PROTOTYPE.updateLayout = function (_p_localBbox) {
  this.pushToDisplayList(this._displayData.displayItems);

  _link.default.processTextLinks(this, this._displayData.links);
};

control.getCurveProperties = function (rect, startPointMS, lastPointMS, midPoint) {
  let zoomFactor = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : 1;
  let expandedRect = Geom.Rectangle.clone(rect);

  if (midPoint) {
    expandedRect = Geom.Rectangle.expand(rect, Geom.Point.scale(midPoint, zoomFactor));
  }

  const topLeft = Geom.Rectangle.topLeft(expandedRect);
  const startPoint = Geom.Point.sub(startPointMS, topLeft);
  const lastPoint = Geom.Point.sub(lastPointMS, topLeft);
  let deltaX = lastPointMS.x - startPointMS.x;
  let deltaY = lastPointMS.y - startPointMS.y;
  let p0 = null;
  let p2 = null;
  const rightArrow = deltaX > 0 || deltaX === 0 && deltaY > 0;

  if (rightArrow) {
    p0 = {
      x: startPoint.x,
      y: startPoint.y
    };
    p2 = {
      x: lastPoint.x,
      y: lastPoint.y
    };
  } else {
    p0 = {
      x: lastPoint.x,
      y: lastPoint.y
    };
    p2 = {
      x: startPoint.x,
      y: startPoint.y
    };
  }

  let p1 = _types.default.newPoint(0.5, 0.0);

  if (midPoint) {
    const scaledPointMS = Geom.Point.scale(midPoint, zoomFactor);
    const pointInControlSpace = Geom.Point.sub(scaledPointMS, Geom.Rectangle.topLeft(expandedRect));
    p1 = _types.default.newPoint(pointInControlSpace.x, pointInControlSpace.y);
    p1 = (0, Geom.toRelativePos)(p0, p2, p1);
  }

  const properties = {
    rightArrow,
    leftArrow: !rightArrow,
    p0,
    p1,
    p2,
    expandedRect
  };
  return properties;
};

control.makeDisplayListForPreview = function (context, skin, bounds, startPointMS, lastPointMS, midPoint, zoomFactor) {
  const {
    arrow_thickness: thickness,
    arrow_default_color: color,
    arrow_point: pointSize
  } = skin;
  const alpha = 0.5; // const scaledStartPointMS = { x: startPointMS.x * zoomFactor, y: startPointMS.y * zoomFactor};
  // const scaledLastPointMS = { x: lastPointMS.x * zoomFactor, y: lastPointMS.y * zoomFactor};

  let {
    expandedRect,
    p0,
    p1,
    p2,
    leftArrow
  } = control.getCurveProperties(bounds, startPointMS, lastPointMS, midPoint, zoomFactor);
  let lastPoint = p2;

  if (leftArrow) {
    lastPoint = p0;
  }

  let pointRightBounds = _types.default.newRectangle(lastPoint.x + expandedRect.x, lastPoint.y + expandedRect.y, pointSize.width * zoomFactor, pointSize.height * zoomFactor); // Compute bbox and choose pc1


  let pc1;
  p1 = Geom.fromRelativePos(p0, p2, p1);
  pc1 = Geom.getBezierControlPoint(p0, p2, p1); //const bBox = Geom.getBezierBoundingBox(p0, pc1, p2);
  // Arrow line

  const arrowPath = Path.newPath(expandedRect.width, expandedRect.height);
  arrowPath.moveTo(p0.x, p0.y);
  arrowPath.curveTo(pc1.x, pc1.y, p2.x, p2.y);

  const arrow = _displayitem.default.newSVGPath(expandedRect, arrowPath, thickness * zoomFactor, color, null);

  arrow.linecap = _displayitem.default.LineCap.ROUND;
  arrow.borderAlpha = alpha;
  arrow.dashArray = null; // Right arrow

  const rotR = Math.atan2(pc1.x - lastPoint.x, pc1.y - lastPoint.y) * 180 / Math.PI - 45;

  const pointRight = _displayitem.default.newPathsAsset(context, "arrow_point", pointRightBounds, color, alpha, rotR);

  return [arrow, ...pointRight];
}; // Point is in Parent Space


control.CONTROL_PROTOTYPE.hitTest = function (p_point) {
  let pt = _types.default.newPoint(p_point.x - this._userBBox.x, p_point.y - this._userBBox.y);

  let isBezier = this.getActualProperty('shape') != Props.ArrowShapes.LINE;
  let distance = 10000;

  if (isBezier) {
    let cvLenApprox = Geom.Point.dist(this._displayData.p0, this._displayData.pc1) + Geom.Point.dist(this._displayData.p2, this._displayData.pc1);
    distance = Geom.getBezierMinDistance(this._displayData.p0, this._displayData.pc1, this._displayData.p2, pt, cvLenApprox / 10);
  } else {
    let p0 = this._displayData.p0;
    let p1 = this._displayData.p1;
    let p2 = this._displayData.p2; // distance from first segment

    let d1 = Geom.getPointSegmentDistance(p0, p1, pt); // distance from second segment

    let d2 = Geom.getPointSegmentDistance(p2, p1, pt);
    distance = d1 < d2 ? d1 : d2;
  }

  return distance < 15;
};

control.CONTROL_PROTOTYPE.clipBBoxSelectionPadding = function () {
  return 20;
};

control.CONTROL_PROTOTYPE.getSelectionDisplayItems = function (_ref) {
  let {
    otherUsersSelectionStatus
  } = _ref;

  let arrow = _displayitem.default.newSVGPath(this.localBBox(), this._displayData.arrowPath, 20, this.getSelectionColor(), null);

  arrow.linecap = _displayitem.default.LineCap.ROUND;
  arrow.borderAlpha = this.getSelectionAlpha(otherUsersSelectionStatus);
  let res = [arrow];

  _displayitem.default.setSelectionDisplayItemsFlag(res);

  return res;
};

control.CONTROL_PROTOTYPE.getSelectionRectangleDisplayItemsAndHandlesCS = function () {
  const handleSize1 = _native.Host.runtime.isMobile() ? 12 : 6;
  const handleSize2 = _native.Host.runtime.isMobile() ? 12 : 7;
  const squares = [(0, _selectionutils.getSquareHandleSizeAndPos)(this._displayData.p0.x, this._displayData.p0.y, handleSize1), (0, _selectionutils.getSquareHandleSizeAndPos)(this._displayData.p1.x, this._displayData.p1.y, handleSize2), (0, _selectionutils.getSquareHandleSizeAndPos)(this._displayData.p2.x, this._displayData.p2.y, handleSize1)];
  const handles = [(0, _selectionutils.createHandle)(_selectionutils.HandleTypes.P0, squares[0], "move"), (0, _selectionutils.createHandle)(_selectionutils.HandleTypes.P1, squares[1], "move"), (0, _selectionutils.createHandle)(_selectionutils.HandleTypes.P2, squares[2], "move")];
  const displayItems = squares.map(square => _displayitem.default.newRectangle(0, 1, 0xffffff, null, null, null, (0, _selectionutils.createRectFromSizeAndPos)(square)));

  for (const displayItem of displayItems) {
    _displayitem.default.setOverlayRectFixedSize(displayItem);
  }

  return {
    displayItems,
    handles
  };
};

control.CONTROL_PROTOTYPE.convertToNewFormatIfNeeded = function () {
  let p0 = this.getActualProperty('p0');
  let p2 = this.getActualProperty('p2');
  let p1 = this.getActualProperty('p1'); // Load old arrow if needed

  if (p0 === null || p2 === null || p1 === null) {
    let curvature = this.getActualProperty("curvature");

    let isTopLeft = this.getActualProperty("direction") === _types.default.Positions.TOP;

    let bbox = this.localBBox();
    var w = bbox.width,
        h = bbox.height;
    ({
      p0,
      p1,
      p2
    } = loadOldPoints(w, h, curvature, isTopLeft));
    let p1s = Geom.toRelativePos(p0, p2, p1);
    this.setActualProperty('p0', p0);
    this.setActualProperty('p1', p1s);
    this.setActualProperty('p2', p2);
  }
};

control.CONTROL_PROTOTYPE.hasCustomHandles = function () {
  return true;
};

control.CONTROL_PROTOTYPE.getHandlePosition = function (handleType) {
  switch (handleType) {
    case _selectionutils.HandleTypes.P0:
      {
        return this._displayData.p0;
      }

    case _selectionutils.HandleTypes.P1:
      {
        return this._displayData.p1;
      }

    case _selectionutils.HandleTypes.P2:
      {
        return this._displayData.p2;
      }

    case _selectionutils.HandleTypes.BOTTOM:
      {
        const origin = this.getActualProperty("origin");
        const size = this.getActualProperty("size");
        return _types.default.newPoint(origin.x + size.width / 2, origin.y + size.height);
      }

    case _selectionutils.HandleTypes.RIGHT:
      {
        const origin = this.getActualProperty("origin");
        const size = this.getActualProperty("size");
        return _types.default.newPoint(origin.x + size.width, origin.y + size.height / 2);
      }
  }
};

control.CONTROL_PROTOTYPE.snap = function () {
  let p0 = this.getActualProperty('p0');
  let p2 = this.getActualProperty('p2');

  if (!p0 || !p2) {
    return null;
  }

  let isBezier = this.getActualProperty('shape') != Props.ArrowShapes.LINE;
  let defaultP1 = Geom.Point.scale(Geom.Point.add(p2, p0), 0.5);

  if (isBezier) {
    return this.moveHandle({
      handleID: 'p1'
    }, defaultP1);
  } else {
    let p1s = this.getActualProperty('p1');
    let p1 = Geom.fromRelativePos(p0, p2, p1s);
    let snapPoints = [_types.default.newPoint(p0.x, p2.y), _types.default.newPoint(p2.x, p0.y), defaultP1];
    snapPoints = snapPoints.sort((a, b) => Geom.Point.dist(p1, a) - Geom.Point.dist(p1, b));
    return this.moveHandle({
      handleID: 'p1'
    }, snapPoints[0]);
  }
};

control.CONTROL_PROTOTYPE.getMoveHandleData = function () {
  this.convertToNewFormatIfNeeded();
  return {
    p1s: _balsamiq.default.clone(this.getActualProperty('p1'))
  };
};

control.CONTROL_PROTOTYPE.moveHandle = function (handle, newPosition, shift, moveHandleData) {
  var propNames = [];
  var propValues = [];
  var newPos = newPosition;

  var p0 = _balsamiq.default.clone(this.getActualProperty('p0'));

  var p2 = _balsamiq.default.clone(this.getActualProperty('p2'));

  var p1s, pc1, p1;

  if (shift && (handle.handleID === _selectionutils.HandleTypes.P0 || handle.handleID === _selectionutils.HandleTypes.P2)) {
    newPos = snapTo45(p0, p2, handle.handleID, newPosition);
  }

  var clippedPoint = MockupSize.clipPoint(MockupCoords.controlSpaceToMockupSpace(newPos, this));
  newPos = MockupCoords.mockupSpaceToControlSpace(clippedPoint, this);

  if (moveHandleData) {
    p1s = _balsamiq.default.clone(moveHandleData.p1s); // avoids that the shape changes during the movement after being squeezed against the borders
  } else {
    // coming from undo/redo
    p1s = _balsamiq.default.clone(this.getActualProperty('p1'));
  }

  _balsamiq.default.assertPoint(p0);

  _balsamiq.default.assertPoint(p1s);

  _balsamiq.default.assertPoint(p2);

  switch (handle.handleID) {
    case _selectionutils.HandleTypes.P0:
      {
        p0.x = newPos.x;
        p0.y = newPos.y;
        p1 = Geom.fromRelativePos(p0, p2, p1s);
        pc1 = Geom.getBezierControlPoint(p0, p2, p1);
        break;
      }

    case _selectionutils.HandleTypes.P1:
      {
        p1 = Geom.fromRelativePos(p0, p2, p1s);
        p1.x = newPos.x;
        p1.y = newPos.y;
        pc1 = Geom.getBezierControlPoint(p0, p2, p1);
        p1s = Geom.toRelativePos(p0, p2, p1);
        propNames.push('p1');
        propValues.push(_types.default.newPoint(p1s.x, p1s.y));
        break;
      }

    case _selectionutils.HandleTypes.P2:
      {
        p2.x = newPos.x;
        p2.y = newPos.y;
        p1 = Geom.fromRelativePos(p0, p2, p1s);
        pc1 = Geom.getBezierControlPoint(p0, p2, p1);
        break;
      }
  } // Resulting BBox depends on shape used


  var bBox;
  var shape = this.getActualProperty('shape');

  if (shape === Props.ArrowShapes.LINE) {
    bBox = Geom.getBoundingBox(p0, p1, p2);
  } else {
    bBox = Geom.getBezierBoundingBox(p0, pc1, p2);
  }

  let delta = _types.default.newPoint(bBox.x, bBox.y);

  delta = {
    x: Math.round(delta.x),
    y: Math.round(delta.y)
  };

  let newOrigin = _balsamiq.default.clone(this.getActualProperty('origin'));

  newOrigin.x += delta.x;
  newOrigin.y += delta.y;
  let dx = 0;
  let dy = 0;
  var mockupSpaceOrigin = MockupCoords.controlSpaceToMockupSpace(_types.default.newPoint(bBox.x, bBox.y), this);

  if (mockupSpaceOrigin.x < 0 || mockupSpaceOrigin.y < 0 || mockupSpaceOrigin.x + bBox.width >= Constants.MaxMockupSize || mockupSpaceOrigin.y + bBox.height >= Constants.MaxMockupSize) {
    var mockupSpaceClippedRect = MockupSize.clipRectangle(_types.default.newRectangle(mockupSpaceOrigin.x, mockupSpaceOrigin.y, bBox.width, bBox.height));
    var clippedOrigin = MockupCoords.mockupSpaceToControlSpace(_types.default.newPoint(mockupSpaceClippedRect.x, mockupSpaceClippedRect.y), this);
    dx += Math.round(clippedOrigin.x - bBox.x);
    dy += Math.round(clippedOrigin.y - bBox.y);
    newOrigin.x += dx;
    newOrigin.y += dy;
    delta.x += dx;
    delta.y += dy;

    if (mockupSpaceOrigin.x + bBox.width >= Constants.MaxMockupSize) // clip against the right border
      {
        dx = mockupSpaceClippedRect.width - bBox.width;
      }

    if (mockupSpaceOrigin.y + bBox.height >= Constants.MaxMockupSize) {
      dy = mockupSpaceClippedRect.height - bBox.height;
    }
  }

  const newP2 = _types.default.newPoint(p2.x - delta.x, p2.y - delta.y);

  const newP0 = _types.default.newPoint(p0.x - delta.x, p0.y - delta.y);

  let newP1 = Geom.fromRelativePos(newP0, newP2, p1s);

  if (moveHandleData) {
    if (shape === Props.ArrowShapes.BEZIER) {
      var p1ms = MockupCoords.controlSpaceToMockupSpace(newP1, this);
      pc1 = Geom.getBezierControlPoint(newP0, newP2, newP1);
      bBox = Geom.getBezierBoundingBox(newP0, pc1, newP2);
      mockupSpaceOrigin = MockupCoords.controlSpaceToMockupSpace(_types.default.newPoint(bBox.x, bBox.y), this);
      var right = mockupSpaceOrigin.x + bBox.width;
      var bottom = mockupSpaceOrigin.y + bBox.height;

      var pOffset = _types.default.newPoint(0, 0);

      if (p1ms.x < 0) pOffset.x = p1ms.x;else if (right >= Constants.MaxMockupSize) pOffset.x = right - Constants.MaxMockupSize;
      if (p1ms.y < 0) pOffset.y = p1ms.y;else if (bottom >= Constants.MaxMockupSize) pOffset.y = bottom - Constants.MaxMockupSize;
      const oldP1 = newP1;
      newP1 = Geom.squashBezier(newP0, newP1, newP2, pOffset);

      _balsamiq.default.assertNumber(newP1.x);

      _balsamiq.default.assertNumber(newP1.y);
    } else {
      newP1.x += dx;
      newP1.y += dy;
    }
  } // needed for correct "size" property (that determines the userBBox)


  if (shape === Props.ArrowShapes.LINE) {
    bBox = Geom.getBoundingBox(newP0, newP1, newP2);
  } else {
    pc1 = Geom.getBezierControlPoint(newP0, newP2, newP1);
    bBox = Geom.getBezierBoundingBox(newP0, pc1, newP2);
  }

  propNames.push('origin');
  propValues.push(newOrigin);
  propNames.push('size');
  propValues.push(_types.default.newSize(Math.max(bBox.width, 1), Math.max(bBox.height, 1)));
  propNames.push('p2');
  propValues.push(newP2);
  propNames.push('p0');
  propValues.push(newP0);

  if (handle.handleID != _selectionutils.HandleTypes.P1) {
    const newP1s = Geom.toRelativePos(newP0, newP2, newP1);

    if (!isNaN(newP1s.x) && !isNaN(newP1s.y)) {
      propNames.push('p1');
      propValues.push(newP1s);
    } // else: Arrow collapsed... keep the old p1s

  }

  return {
    propNames: propNames,
    propValues: propValues
  };
};

control.CONTROL_PROTOTYPE.resizeBBox = function (newBBox) {
  this.convertToNewFormatIfNeeded();

  let newSize = _types.default.newSize(newBBox.width, newBBox.height);

  newSize.width = Math.max(newSize.width, 1);
  newSize.height = Math.max(newSize.height, 1);
  const oldSize = this.getActualProperty('size');
  const deltaH = newSize.height / oldSize.height;
  const deltaW = newSize.width / oldSize.width;
  const isBezier = this.getActualProperty('shape') != Props.ArrowShapes.LINE;
  const p0 = this.getActualProperty('p0');
  const p2 = this.getActualProperty('p2');
  const p1s = this.getActualProperty('p1');
  const p1 = Geom.fromRelativePos(p0, p2, p1s);

  const newP0 = _types.default.newPoint(Math.round(p0.x * deltaW), Math.round(p0.y * deltaH));

  const newP2 = _types.default.newPoint(Math.round(p2.x * deltaW), Math.round(p2.y * deltaH));

  let newP1;

  if (isBezier) {
    let pc1 = Geom.getBezierControlPoint(p0, p2, p1);

    let newPC1 = _types.default.newPoint(pc1.x * deltaW, pc1.y * deltaH);

    newP1 = Geom.evaluateBezier(newP0, newPC1, newP2, 0.5);
  } else {
    newP1 = _types.default.newPoint(p1.x * deltaW, p1.y * deltaH);
  }

  const newP1S = Geom.toRelativePos(newP0, newP2, newP1);
  const propNames = ['origin', 'size', 'p0', 'p2', 'p1'];
  const propValues = [_types.default.newPoint(newBBox.x, newBBox.y), newSize, newP0, newP2, newP1S];
  return {
    propNames: propNames,
    propValues: propValues
  };
};

control.CONTROL_PROTOTYPE.changeShape = function (newShape) {
  this.convertToNewFormatIfNeeded();
  var propNames = [];
  var propValues = [];
  var isBezier = this.getActualProperty('shape') != Props.ArrowShapes.LINE;
  var p0 = this.getActualProperty('p0');
  var p2 = this.getActualProperty('p2');
  var p1s = this.getActualProperty('p1');
  var p1 = Geom.fromRelativePos(p0, p2, p1s);
  var pc1 = Geom.getBezierControlPoint(p0, p2, p1);
  var origin = this.getActualProperty('origin');
  var bBox;

  if (newShape === Props.ArrowShapes.LINE && isBezier) {
    bBox = Geom.getBoundingBox(p0, p1, p2);
  } else if (newShape === Props.ArrowShapes.BEZIER && !isBezier) {
    bBox = Geom.getBezierBoundingBox(p0, pc1, p2);
  } else {
    return {
      propNames: [],
      propValues: []
    };
  }

  origin.x += bBox.x;
  origin.y += bBox.y;
  propValues = [origin, newShape];
  propNames = ['origin', 'shape'];
  propNames.push('size');
  propValues.push(_types.default.newSize(Math.max(bBox.width, 1), Math.max(bBox.height, 1)));
  propNames.push('p2');
  propValues.push(_types.default.newPoint(p2.x - bBox.x, p2.y - bBox.y));
  propNames.push('p0');
  propValues.push(_types.default.newPoint(p0.x - bBox.x, p0.y - bBox.y));
  return {
    propNames: propNames,
    propValues: propValues
  };
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/barcharts.js":
/*!***************************************!*\
  !*** ./js/core/controls/barcharts.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = exports.barChartVertical = exports.barChartHorizontal = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var makeBarControlClass = function (p_controlTypeID, p_controlTypeName) {
  var control = controlsMod.makeControlClass({
    controlTypeID: p_controlTypeID,
    controlTypeName: p_controlTypeName,
    controlAlternateNames: [],
    props: { ...Props.PropertyDefinitions.SIZE({
        defaultValue: _types.default.newSize(187, 175),
        options: {
          autosize: Props.AutoSizeOptions.NO,
          resize: Props.ResizeOptions.BOTH
        }
      })
    }
  });

  control.CONTROL_PROTOTYPE.makeAxes = function (p_localBbox, p_thickness, p_color) {
    var offset = p_thickness / 2;
    var path = Path.newPath(p_localBbox.width, p_localBbox.height);
    path.moveTo(offset, 0);
    path.lineTo(offset, p_localBbox.height);
    path.moveTo(0, p_localBbox.height - offset);
    path.lineTo(p_localBbox.width, p_localBbox.height - offset);
    return [_displayitem.default.newSVGPath(p_localBbox, path, p_thickness, p_color)];
  };

  control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
    var thickness = this.style("barchart_axes_thickness");
    var axes = this.makeAxes(p_localBbox, thickness, this.style("barchart_axes_color"));
    var bboxes = this.makeBarBoxes(p_localBbox.width, p_localBbox.height, thickness);
    var color1 = this.style("barchart_series1_color");
    var color2 = this.style("barchart_series2_color");
    var isColor1 = true;
    var bars = bboxes.map(bbox => {
      var result;

      if (bbox.width > 0 && bbox.height > 0) {
        result = (0, _rectangle.makeRectangleDisplayItems)(this._context, bbox, Props.BorderStyles.SQUARE, isColor1 ? color1 : color2);
        isColor1 = !isColor1;
      }

      return result;
    });
    this.pushToDisplayList(axes, bars);
  };

  if (p_controlTypeID === "com.balsamiq.mockups::ColumnChart") {
    // VERTICAL BARS
    control.CONTROL_PROTOTYPE.makeBarBoxes = function (w, h, thickness) {
      var t = thickness - 1;

      var bbox1 = _types.default.newRectangle(Math.round(w * 0.16), h - Math.max(10, h - 89) - t, Math.round(w * 0.15), Math.max(10, h - 89));

      var bbox2 = _types.default.newRectangle(Math.round(w * 0.31), h - Math.max(10, h - 67) - t, Math.round(w * 0.15), Math.max(10, h - 67));

      var bbox3 = _types.default.newRectangle(Math.round(w * 0.58), h - Math.max(10, h - 32) - t, Math.round(w * 0.15), Math.max(10, h - 32));

      var bbox4 = _types.default.newRectangle(Math.round(w * 0.73), h - Math.max(10, h - 48) - t, Math.round(w * 0.15), Math.max(10, h - 48));

      return [bbox1, bbox2, bbox3, bbox4];
    };
  } else {
    // HORIZONTAL BARS
    control.CONTROL_PROTOTYPE.makeBarBoxes = function (w, h, thickness) {
      var t = thickness - 1;

      var bbox1 = _types.default.newRectangle(t, Math.round(h * 0.16), Math.max(10, w - 100), Math.round(h * 0.15));

      var bbox2 = _types.default.newRectangle(t, Math.round(h * 0.31), Math.max(10, w - 79), Math.round(h * 0.15));

      var bbox3 = _types.default.newRectangle(t, Math.round(h * 0.58), Math.max(10, w - 45), Math.round(h * 0.15));

      var bbox4 = _types.default.newRectangle(t, Math.round(h * 0.73), Math.max(10, w - 60), Math.round(h * 0.15));

      return [bbox1, bbox2, bbox3, bbox4];
    };
  }

  return control;
};

const barChartVertical = makeBarControlClass("com.balsamiq.mockups::ColumnChart", "Chart: Column");
exports.barChartVertical = barChartVertical;
const barChartHorizontal = makeBarControlClass("com.balsamiq.mockups::BarChart", "Chart: Bar");
exports.barChartHorizontal = barChartHorizontal;
var _default = {
  barChartVertical,
  barChartHorizontal
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/blockoftext.js":
/*!*****************************************!*\
  !*** ./js/core/controls/blockoftext.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var constantsMod = __webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts");

const PSEUDO_RANDOM = [0, 2, 0, 1, 10, 0, 5, 0, 1, 3, 2, 30, 0, 1, 4, 9, 1, 4, 7, 8, 4, 6];
const RECT_ALPHA = 0.5;
var LOREM_WORDS = constantsMod.Lorem1.split(' ');
var MAX_LINES = 90;
var PARAGRAPH_WORD_COUNT = [105, 63, 48, 115, 49, 18];
var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::BlockOfText",
  controlTypeName: "Block of Text",
  controlAlternateNames: ["Blob", "Greeked", "Fake"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(275, 80),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.QUICKDRAW_TIP("You can draw Blocks of Text quickly by holding down the *T key* and dragging. [Learn more]($DOCURL)", "You can also draw Blocks of Text with your fingers using the Quick Draw Toolbar (in the View Menu). [Learn more]($DOCURL)")
  }
});

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  // Use same code for quick draw and layout
  // but quick draw needs to be scaled for zoom factor while layout doesn't because the proper scaling is at the display list root
  let dl = control.makeDisplayListForPreview(this._context, this._context.skin, p_localBbox, 1);
  this.pushToDisplayList(dl);
  return;
};

control.makeDisplayListForPreview = function (p_context, p_skin, p_localBbox, zoomFactor) {
  var bounds;
  var defaultFontSize = p_context.defaultFontSize;
  var defaultTextColor = p_skin['default_text_color'];
  var rowHeight = defaultFontSize * zoomFactor;
  var padding = rowHeight / 4.64;
  let theY = p_localBbox.y + 3;
  let theX = p_localBbox.x + 0;
  let displayList = [];
  var fakeFont = p_skin["blockoftext_fake_font"];

  if (fakeFont) {
    var skinRatio = p_skin['blockoftext_default_font_size'] / defaultFontSize;
    var spaceWidth = p_skin["blockoftext_space_width"] / skinRatio * zoomFactor;
    let loremWordIndex = 0;
    let addNewLines = 0;
    let paragraphWords = 0;
    let totalLines = 1;
    let firstWordOfLine = true;
    let paragraphIndex = 0;

    while (theY - p_localBbox.y + rowHeight < p_localBbox.height && totalLines < MAX_LINES) {
      // Should simply add a new line in this iteration
      if (addNewLines > 0) {
        theX = p_localBbox.x;
        theY += rowHeight + padding;
        addNewLines -= 1;
        totalLines += 1;
        firstWordOfLine = true;
        continue;
      } // Pick next word to add and the right asset


      let word = LOREM_WORDS[loremWordIndex];
      let assetName = "blockoftext_" + word.length;
      let sk = p_skin[assetName];
      let skinWidth = sk.width / skinRatio * zoomFactor;
      let skinHeight = sk.height / skinRatio * zoomFactor; // If this word overflows the line, add new line in next iteration and keep same word to be added later
      // first word of line never overflows!

      if (!firstWordOfLine && theX - p_localBbox.x + skinWidth > p_localBbox.width) {
        addNewLines += 1;
        continue;
      } // Calculate BBox for new word and add displayitem


      let bbox = _types.default.newRectangle(theX, theY, skinWidth, skinHeight);

      theX += skinWidth + spaceWidth;
      displayList = displayList.concat(_displayitem.default.newPathsAsset(p_context, assetName, bbox, defaultTextColor, null, null));
      paragraphWords += 1;
      firstWordOfLine = false; // Go to next word: cycle thru LOREM words, one at a time

      loremWordIndex = (loremWordIndex + 1) % LOREM_WORDS.length; // Add 2 newLines to begin new paragraph

      if (paragraphWords == PARAGRAPH_WORD_COUNT[paragraphIndex]) {
        paragraphIndex = (paragraphIndex += 1) % PARAGRAPH_WORD_COUNT.length;
        paragraphWords = 0;
        addNewLines += 2;
      }
    }
  } else {
    let theH = defaultFontSize * zoomFactor;
    let row = 0;

    while (theY + rowHeight < p_localBbox.height + p_localBbox.y) {
      var theW = p_localBbox.width - Math.round(p_localBbox.width * PSEUDO_RANDOM[row % PSEUDO_RANDOM.length] / 30);

      if (theW > 0) {
        bounds = _types.default.newRectangle(theX, theY, theW, theH);
        displayList = displayList.concat((0, _rectangle.makeFilledRectangle)(bounds, defaultTextColor, RECT_ALPHA));
      }

      theY += rowHeight + padding;
      row++;
    }
  }

  return displayList;
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/breadcrumbs.js":
/*!*****************************************!*\
  !*** ./js/core/controls/breadcrumbs.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::BreadCrumbs",
  controlTypeName: "Breadcrumbs",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.links()
  }
});
controlsMod.addStandardTextPropertyDefinitions(control, {
  text: "Home, Products, Xyz, Features",
  splitOn: Props.SplitOnTypes.COMMA
});

control.CONTROL_PROTOTYPE.initData = function () {
  this._clipOnBbox = false;
};

control.CONTROL_PROTOTYPE.getLinkNames = function () {
  return this._linkNames;
};

control.CONTROL_PROTOTYPE.getLinkableLabels = function (text) {
  var strings = TextUtils.splitTextOnCommas(text);
  let linkLabels = strings.map(str => TextUtils.unformatText(str));
  return linkLabels;
};

control.CONTROL_PROTOTYPE.updateData = function () {
  this.texts = [];
  let text = this.getActualProperty('text');
  var strings = TextUtils.splitTextOnCommas(text);
  this._linkNames = this.getLinkableLabels(text);

  if (strings.length === 1) {
    strings.push("...");
  }

  var textStyle = this.newStandardTextStyle({
    underline: true,
    textColor: this.style("default_link_color")
  });
  var lastTextStyle = this.newStandardTextStyle({
    underline: false
  });
  strings.forEach((string, index) => {
    this.texts.push(textMod.newText(string, index === strings.length - 1 ? lastTextStyle : textStyle, this._context));
  });
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var width = 0,
      height = 0;
  var separatorSize = this.styleSVGSize('breadcrumbs_separator');
  var separatorPaddingMiddle = this.style('breadcrumbs_padding_middle');
  this.texts.forEach(text => {
    var textSize = text.measuredSize();
    width += textSize.width;
    height = Math.max(height, textSize.height);
  });
  width += (2 * separatorPaddingMiddle + separatorSize.width) * (this.texts.length - 1);
  height += this.style("breadcrumbs_padding_top") + this.style("breadcrumbs_padding_bottom");
  return _types.default.newSize(width, height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var x = 0;
  var separatorPaddingMiddle = this.style('breadcrumbs_padding_middle');
  var separatorSize = this.styleSVGSize('breadcrumbs_separator');
  var separatorBbox;
  var separatorY = (p_localBbox.height - separatorSize.height) / 2;
  var linkBounds = [];
  this.texts.forEach((text, index) => {
    var rect = _types.default.newRectangle(x, 0, text.measuredSize().width + this.style('breadcrumbs_padding_top') + this.style('breadcrumbs_padding_bottom'), p_localBbox.height);

    this.pushToDisplayList(text.makeDisplayItems(rect, _types.default.TextAlign.LEFT, _types.default.Positions.CENTER));
    linkBounds.push(rect);

    if (index !== this.texts.length - 1) {
      x += text.measuredSize().width;
      x += separatorPaddingMiddle;
      separatorBbox = _types.default.newRectangle(x, separatorY, separatorSize.width, separatorSize.height);
      this.pushToDisplayList(_displayitem.default.newPathsAsset(this._context, "breadcrumbs_separator", separatorBbox, this.style("default_icon_color")));
      x += separatorSize.width + separatorPaddingMiddle;
    }
  });

  _link.default.processLinks(this, linkBounds);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/browserWindow.js":
/*!*******************************************!*\
  !*** ./js/core/controls/browserWindow.js ***!
  \*******************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _scrollbars = __webpack_require__(/*! core/bricks/scrollbars */ "./js/core/bricks/scrollbars.js");

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _macro = __webpack_require__(/*! core/bricks/macro */ "./js/core/bricks/macro.js");

var _breakline = __webpack_require__(/*! core/bricks/breakline */ "./js/core/bricks/breakline.js");

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _snapping = __webpack_require__(/*! core/controls/snapping */ "./js/core/controls/snapping.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var browserwindow = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::BrowserWindow",
  controlTypeName: "Browser Window",
  controlAlternateNames: ["Browser"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(450, 400),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.TEXT({
      defaultText: 'A Web Page\nhttps://',
      multiline: true
    }),
    ...Props.PropertyDefinitions.BORDER_STYLE(Props.BorderStyles.SQUARE, Props.BrowserBorderStyles),
    ...Props.PropertyDefinitions.SCROLLBAR()
  }
});
controlsMod.addColorPropertyDefinitions(browserwindow, {
  defaultColor: 0xFFFFFF
});

browserwindow.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  // Split title and url lines
  const text = (0, _macro.processMacros)(this);
  const indexOfLineBreak = text.indexOf('\n');
  const isSingleLine = indexOfLineBreak === -1;
  const title = isSingleLine ? text : text.slice(0, indexOfLineBreak);
  const url = isSingleLine ? "" : text.slice(indexOfLineBreak + 1);
  const textStyle = this.newTextStyle();

  if (!this.titleText || !(0, _text.areTextsEqual)(this.titleText, title, textStyle)) {
    this.titleText = (0, _text.newSingleLineText)(title, textStyle, this._context);
  }

  if (!this.urlText || !(0, _text.areTextsEqual)(this.urlText, url, textStyle)) {
    this.urlText = (0, _text.newSingleLineText)(url, textStyle, this._context);
  }

  var useBreakLine = this.getActualProperty("borderStyle") === Props.BorderStyles.SQUARE_BREAKLINE && p_localBbox.height > _types.default.BREAKLINE_VERTICAL_OFFSET * 2;
  var urlFieldWidth = p_localBbox.width - this.style('browserwindow_url_padding_left') - this.style('browserwindow_url_padding_right'); // White background

  var rect = _types.default.newRectangle(2, 2, p_localBbox.width - 4, useBreakLine ? p_localBbox.height - 2 : this.style("browserwindow_scrollbar_padding_top") - 4);

  var whiteBackground = (0, _rectangle.makeFilledRectangle)(rect, 0xFFFFFF); // Colored background

  var backgroundHeight = p_localBbox.height - this.style("browserwindow_scrollbar_padding_top") - this.style("browserwindow_scrollbar_padding_bottom");
  var background;

  if (backgroundHeight > 0) {
    var backgroundRect = _types.default.newRectangle(1, this.style("browserwindow_scrollbar_padding_top"), p_localBbox.width - 2, backgroundHeight);

    if (useBreakLine) {
      backgroundRect.height += this.style("browserwindow_scrollbar_padding_bottom");
    }

    background = (0, _rectangle.makeFilledRectangle)(backgroundRect, this.getActualProperty("color"));
  } // Browser skin


  var skinBbox = useBreakLine ? _types.default.paddedRectangle(p_localBbox, 0, -this.style("browserwindow_scrollbar_padding_bottom")) : p_localBbox;

  var browser = _displayitem.default.createDisplayItemsFromSkin("browserwindow", null, skinBbox, this._context); // Texts


  var theTitle = this.titleText.makeDisplayItems(_types.default.newRectangle(0, this.style("browserwindow_padding_top"), p_localBbox.width, this.titleText.measuredSize().height), _types.default.TextAlign.CENTER);
  var theURL;

  if (urlFieldWidth > 0) {
    theURL = this.urlText.makeDisplayItems(_types.default.newRectangle(this.style("browserwindow_url_padding_left"), this.style("browserwindow_url_padding_top"), urlFieldWidth, this.urlText.measuredSize().height));
  } // Scrollbar


  var scrollbar = [];

  if (this.getActualProperty('verticalScrollbar')) {
    scrollbar = (0, _scrollbars.makeVerticalScrollbar)(p_localBbox, this.style("browserwindow_scrollbar_padding_top"), useBreakLine ? this.style("browserwindow_scrollbar_padding_bottom_breakline") : this.style("browserwindow_scrollbar_padding_bottom"), this.style("browserwindow_scrollbar_padding_right"), this.getActualProperty('value'), _types.default.ControlState.NORMAL, this._context);
  }

  var breakline, maskId;

  if (useBreakLine && p_localBbox.width > 66) {
    //TODO fix this
    maskId = (0, _breakline.addBreaklineMask)(this, 0, p_localBbox.height - _types.default.BREAKLINE_VERTICAL_OFFSET);

    _displayitem.default.maskDisplayItems([whiteBackground, background, theURL, browser, theTitle, scrollbar], maskId);

    breakline = (0, _breakline.makeBreaklineDisplayitem)(1, p_localBbox.height - _types.default.BREAKLINE_VERTICAL_OFFSET, p_localBbox.width - 2, undefined, this.style("browserwindow_border_thickness"));
  }

  this.pushToDisplayList(whiteBackground, background, browser, theURL, theTitle, scrollbar, breakline);
};

browserwindow.CONTROL_PROTOTYPE.getSnapLines = function (withMarkup) {
  _balsamiq.default.assertBool(withMarkup);

  var paddingTop = this.style("browserwindow_scrollbar_padding_top");
  var paddingBottom = this.style("browserwindow_scrollbar_padding_bottom");
  var bbox = this.getUserBBox(true);
  return [new _snapping.SnapLine(_snapping.SnapTypes.MIDDLE, _snapping.SnapDirections.VERTICAL, bbox.x + bbox.width / 2), new _snapping.SnapLine(_snapping.SnapTypes.MIDDLE, _snapping.SnapDirections.HORIZONTAL, bbox.y + bbox.height / 2), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.VERTICAL, bbox.x), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.VERTICAL, bbox.x + bbox.width), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.HORIZONTAL, bbox.y), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.HORIZONTAL, bbox.y + bbox.height), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.VERTICAL, bbox.x + _snapping.PROXIMITY_DISTANCE), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.VERTICAL, bbox.x + bbox.width - _snapping.PROXIMITY_DISTANCE), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.HORIZONTAL, bbox.y + _snapping.PROXIMITY_DISTANCE + paddingTop), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.HORIZONTAL, bbox.y + bbox.height - _snapping.PROXIMITY_DISTANCE - paddingBottom)];
};

browserwindow.CONTROL_PROTOTYPE.getDescription = function () {
  let res, text;

  if (this.hasPropertyDefinition("text") && (text = this.getActualProperty("text")) !== "") {
    res = TextUtils.unformatText(text);
    res = res.split("↵")[0]; // Shows only the first line
  }

  return res || this.superClass.prototype.getDescription.call(this);
};

var _default = browserwindow;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/buttonBar.js":
/*!***************************************!*\
  !*** ./js/core/controls/buttonBar.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var controlsMod = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var buttonBar = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::ButtonBar",
  controlTypeName: "Button Bar",
  controlAlternateNames: ["Tab Bar"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.WIDTH_AND_OR_HEIGHT,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.links()
  }
}); // PRIVATE

function maxButtonTextWidth(buttonData) {
  const value = _balsamiq.default.maxBy(buttonData.textLines, function (line) {
    return line.measuredSize().width;
  });

  return value.measuredSize().width;
}

function calculateLead(fontSize) {
  return fontSize * 0.25;
}

controlsMod.addStandardTextPropertyDefinitions(buttonBar, {
  text: "One, Two, Three",
  splitOn: Props.SplitOnTypes.COMMA
});
controlsMod.addSelectionPropertyDefinitions(buttonBar, 0);

buttonBar.CONTROL_PROTOTYPE.initData = function () {
  this._columns = [];
  this._buttonTexts = [];
  this._autoWidth = 0; // Store calculated auto width

  this._autoHeight = 0; // Store calculated auto height

  this._selectableLabels = [];
};

buttonBar.CONTROL_PROTOTYPE.getSelectableLabels = function () {
  return this._selectableLabels;
};

buttonBar.CONTROL_PROTOTYPE.getLinkNames = function () {
  return this._selectableLabels;
};

buttonBar.CONTROL_PROTOTYPE.getLinkableLabels = function (text) {
  return text.split(",").map(str => TextUtils.unformatText(str.trim()));
};

buttonBar.CONTROL_PROTOTYPE.updateData = function () {
  this.buildTextLines();
  this.measureAutoWidth();
  this.measureAutoHeight();
}; // Measure automatic width height


buttonBar.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  // respect autosize
  const width = p_sizeProperty.width === _types.default.AUTOSIZE ? this._autoWidth : p_sizeProperty.width;
  const height = p_sizeProperty.height === _types.default.AUTOSIZE ? this._autoHeight : p_sizeProperty.height;
  return _types.default.newSize(width, height);
};

buttonBar.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  // Text lines are ready
  // BBox is ready
  this.positionButtons(p_localBbox); // Calculate rectangle bbox
  // It may be smaller than user set bbox if some buttons are cut away!

  const width = this._columns.reduce((memo, column) => {
    return memo + column.width;
  }, 0); // Outer border, like a rectangle


  let bbox = _balsamiq.default.clone(p_localBbox);

  bbox.x = 0;
  bbox.y = 0;
  bbox.width = width;

  const border = _displayitem.default.createDisplayItemsFromSkin("buttonbar_border", this.style("default_border_color"), bbox, this._context); // Background


  let backgroundBBox = _balsamiq.default.clone(bbox);

  backgroundBBox.x += 1;
  backgroundBBox.y += 1;
  backgroundBBox.width -= 2;
  backgroundBBox.height -= 2;
  let background;

  if (backgroundBBox.width > 0 && backgroundBBox.height > 0) {
    background = (0, _rectangle.makeFilledRectangle)(backgroundBBox, this.style('buttonbar_background_color'));
  } // Link areas are as big as a button


  let linkAreaBounds = []; // The texts

  let texts = []; // Draw vertical separators as lines and texts

  let theX = 0;
  let separatorsPath = Path.newPath(p_localBbox.width, p_localBbox.height);
  const separatorsWidth = this.style('buttonbar_separator_width');
  let rectBbox;

  this._columns.forEach((column, index) => {
    rectBbox = _types.default.newRectangle(theX, 0, column.width, p_localBbox.height);
    linkAreaBounds.push(rectBbox);

    if (index !== this._columns.length - 1) {
      theX += column.width;
      separatorsPath.moveTo(theX - separatorsWidth / 2, 1);
      separatorsPath.lineTo(theX - separatorsWidth / 2, p_localBbox.height - 1);
    }

    const text = this._buttonTexts[index];
    texts.push(text.makeDisplayItems(rectBbox, _types.default.TextAlign.CENTER, _types.default.Positions.CENTER));
  });

  const separators = _displayitem.default.newSVGPath(p_localBbox, separatorsPath, separatorsWidth, this.style('buttonbar_separator_color'));

  _link.default.processLinks(this, linkAreaBounds); // Draw selected button background shrinked a bit


  const selectIndex = this.getActualProperty("selectedIndex");
  let selectionRectangle = [];

  if (selectIndex !== -1 && selectIndex < this._columns.length) {
    const bounds = _types.default.paddedRectangle(this._columns[selectIndex], this.style("buttonbar_selection_padding_top"), this.style("buttonbar_selection_padding_bottom"), selectIndex === 0 ? this.style("buttonbar_selection_padding_left") : 0, selectIndex === this._columns.length - 1 ? this.style("buttonbar_selection_padding_right") : 0);

    selectionRectangle = (0, _rectangle.makeFilledRectangle)(bounds, this.style("default_selection_color"));
  }

  this.pushToDisplayList(background, selectionRectangle, separators, border, texts);
}; // Parse text string and split into text lines


buttonBar.CONTROL_PROTOTYPE.buildTextLines = function () {
  const text = this.getActualProperty("text");
  const items = text.split(",").map(function removeFirstSpace(item) {
    return item[0] === " " ? item.slice(1) : item;
  });
  this._selectableLabels = this.getLinkableLabels(text);

  if (items.length === 1) {
    items.push("...");
  }

  let buttonTexts = [];
  items.forEach((item, i) => {
    let textStyle = this.newStandardTextStyle();
    textStyle.textColor = i == this.getActualProperty("selectedIndex") ? _colors.default.labelColorGivenBackgroundColor(this.style('default_selection_color'), 1, this.style("default_text_color")) : textStyle.textColor;
    textStyle.linkColor = i == this.getActualProperty("selectedIndex") ? _colors.default.labelColorGivenBackgroundColor(this.style('default_selection_color'), 1, this.style("default_link_color")) : textStyle.linkColor;
    const txt = Text.newText(item, textStyle, this._context);
    buttonTexts.push(txt);
  });
  this._buttonTexts = buttonTexts;
}; // Calculate automatic width


buttonBar.CONTROL_PROTOTYPE.measureAutoWidth = function () {
  const paddingLeft = this.style('buttonbar_padding_left');
  const paddingMiddle = this.style('buttonbar_padding_middle');
  const paddingRight = this.style('buttonbar_padding_right');

  const linesWidth = this._buttonTexts.reduce((memo, text) => {
    return memo + text.measuredSize().width;
  }, 0);

  this._autoWidth = paddingLeft + paddingRight + paddingMiddle * 2 * (this._buttonTexts.length - 1) + linesWidth;
}; // Calculate automatic width


buttonBar.CONTROL_PROTOTYPE.measureAutoHeight = function () {
  const paddingTop = this.style('buttonbar_padding_top');
  const paddingBottom = this.style('buttonbar_padding_bottom');

  const maxTextHeight = _balsamiq.default.maxBy(this._buttonTexts, function (text) {
    return text.measuredSize().height;
  }).measuredSize().height;

  this._autoHeight = paddingTop + paddingBottom + maxTextHeight;
}; // Calculate texts and buttons positions


buttonBar.CONTROL_PROTOTYPE.positionButtons = function (p_localBBox) {
  let columns = [];
  const paddingLeft = this.style('buttonbar_padding_left');
  const paddingMiddle = this.style('buttonbar_padding_middle');
  const paddingRight = this.style('buttonbar_padding_right');
  const controlWidth = p_localBBox.width;
  const controlHeight = p_localBBox.height;
  let currX = 0;

  this._buttonTexts.forEach((text, i) => {
    // lines in an Object containing the array of lines for each button
    let minColWidth = 0;

    if (i === 0) {
      minColWidth += paddingLeft;
    } else {
      minColWidth += paddingMiddle;
    } // Get button widest line of text width


    minColWidth += text.measuredSize().width;

    if (i === this._buttonTexts.length - 1) {
      minColWidth += paddingRight;
    } else {
      minColWidth += paddingMiddle;
    } // We have minimum col width (label+paddings) so we can calulate extended col width


    let maxColW = Math.round(Math.max(minColWidth, controlWidth * minColWidth / this._autoWidth));
    let colW;

    if (i === this._buttonTexts.length - 1) {
      colW = currX + maxColW <= controlWidth ? maxColW : controlWidth - currX > minColWidth ? controlWidth - currX : minColWidth;
    } else {
      colW = maxColW;
    } // If button goes beyond user-set width, ignore it


    if (currX + colW > controlWidth) {
      currX += colW;
      return;
    }

    columns.push(_types.default.newRectangle(currX, 0, colW, controlHeight)); // CurrX moved right to beginning of next column

    currX += colW;
  });

  this._columns = columns;
};

var _default = buttonBar;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/buttons.js":
/*!*************************************!*\
  !*** ./js/core/controls/buttons.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.pointybutton = exports.multilinebutton = exports["default"] = exports.button = void 0;

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const MultilineButtonControlTypeID = "com.balsamiq.mockups::MultilineButton";

const makeButtonControlClass = function (_ref) {
  let {
    controlTypeID,
    controlTypeName,
    controlAlternateNames,
    controlInitialProperties,
    stylePrefix,
    sizeOptions,
    defaultText,
    defaultFontSize,
    customProperties,
    alignFunction,
    horizontalPaddingFunction,
    minSizeFunction,
    displayListFunction,
    multiline,
    hasTextLinks
  } = _ref;
  const control = Controls.makeControlClass({
    controlTypeID,
    controlTypeName,
    controlAlternateNames,
    controlInitialProperties,
    props: { ...Props.PropertyDefinitions.SIZE({
        defaultValue: _types.default.newAutoSize(),
        options: sizeOptions
      }),
      ...Props.PropertyDefinitions.MENU_ICON(),
      // used only by Button and PointyButton
      ...Props.PropertyDefinitions.link,
      ...Props.PropertyDefinitions.ROTATION({
        isIcon: true
      })
    }
  });
  Controls.addStandardTextPropertyDefinitions(control, {
    text: defaultText,
    multiline: multiline
  });
  Controls.addColorPropertyDefinitions(control, {
    defaultColor: stylePrefix + "_background_color"
  });
  Controls.addPropertyDefinitions(control, customProperties);

  control.CONTROL_PROTOTYPE.initData = function () {
    this._textLinks = [];
  };

  if (hasTextLinks) {
    _link.default.addTextLinksMethods(control, {
      getLinksFunction: function () {
        return this._textLinks;
      }
    });
  }

  control.CONTROL_PROTOTYPE.updateData = function () {
    _icon.default.processIconData(this);

    this._textLinks = [];
    const textProp = this.getActualProperty("text");
    const newlineIndex = controlTypeID === MultilineButtonControlTypeID ? textProp.indexOf("\n") : -1;
    const firstLine = newlineIndex === -1 ? textProp : textProp.substring(0, newlineIndex);
    const firstTextStyle = this.newStandardTextStyle({
      textColor: this.textAndIconAndLinkColor().textColor,
      linkColor: this.textAndIconAndLinkColor().linkColor
    });

    if (!this.firstText || !Text.areTextsEqual(this.firstText, firstLine, firstTextStyle)) {
      this.firstText = Text.newTextNoWrap(firstLine, firstTextStyle, this._context);
    }

    this._textLinks = this._textLinks.concat(this.firstText.getLinks());

    if (newlineIndex !== -1) {
      const secondLine = textProp.substring(newlineIndex + 1);
      const secondTextStyle = this.newStandardTextStyle({
        textColor: firstTextStyle.textColor,
        fontSize: Math.max(Math.ceil(this.getActualProperty("fontSize") * 0.7), 8),
        linkColor: firstTextStyle.linkColor
      });
      this.secondText = Text.newText(secondLine, secondTextStyle, this._context);
      this._textLinks = this._textLinks.concat(this.secondText.getLinks());
    } else {
      this.secondText = null;
    }
  };

  control.CONTROL_PROTOTYPE.menuIconSize = function () {
    if (!this.getActualProperty("menuIcon")) {
      return 0;
    }

    const fontSize = this.getActualProperty("fontSize");
    return fontSize < 28 ? 10 : fontSize < 60 ? 12 : 24;
  };

  control.CONTROL_PROTOTYPE.textAndIconAndLinkColor = function () {
    const hasStateProp = this.hasPropertyDefinition('state');
    const alpha = this.hasPropertyDefinition('backgroundAlpha') ? this.getActualProperty('backgroundAlpha') : 1;
    let textColor;
    let linkColor;

    if (hasStateProp && _types.default.ControlState.isDisabledState(this.getActualProperty('state'))) {
      textColor = this.style('default_disabled_color');
    } else if (hasStateProp && this.getActualProperty('state') === _types.default.ControlState.SELECTED) {
      textColor = _colors.default.labelColorGivenBackgroundColor(this.style('default_selection_color'), alpha, this.style("default_text_color"));
      linkColor = _colors.default.labelColorGivenBackgroundColor(this.style('default_selection_color'), alpha, this.style("default_link_color"));
    } else {
      textColor = _colors.default.labelColorGivenBackgroundColor(this.getActualProperty('color'), alpha, this.style("default_text_color"));
      linkColor = _colors.default.labelColorGivenBackgroundColor(this.getActualProperty('color'), alpha, this.style("default_link_color"));
    }

    return {
      textColor,
      iconColor: textColor,
      linkColor
    };
  };

  control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
    const minSize = minSizeFunction.call(this);
    const paddingLeft = horizontalPaddingFunction.call(this, _types.default.Positions.LEFT);
    const paddingRight = horizontalPaddingFunction.call(this, _types.default.Positions.RIGHT);
    const iconPadding = this.getActualProperty("text") ? this.style(stylePrefix + "_icon_padding") : 0;
    const textWidth = this.firstText.measuredSize().width;
    let width = textWidth + paddingLeft + paddingRight;

    if (this.iconData) {
      width += this.iconData.size + iconPadding;
    }

    width = Math.max(width, minSize.width);
    let textHeight = this.firstText.measuredSize().height;

    if (this.secondText) {
      const availableWidthForSecondLine = textWidth + paddingLeft + paddingRight;
      textHeight += this.style(stylePrefix + "_padding_middle") + this.secondText.measuredSizeForWidth(availableWidthForSecondLine).height;
    }

    let height = textHeight + this.style(stylePrefix + '_padding_top') + this.style(stylePrefix + '_padding_bottom');

    if (this.iconData) {
      height = Math.max(height, this.iconData.size);
    }

    height = Math.max(height, minSize.height);
    const menuIconSize = this.menuIconSize();

    if (menuIconSize > 0) {
      width += menuIconSize + iconPadding * 3;
      height = Math.max(height, menuIconSize + this.style(stylePrefix + '_padding_top'));
    } // respect autosize


    width = (p_sizeProperty.width === _types.default.AUTOSIZE ? width : p_sizeProperty.width) + 16;
    height = p_sizeProperty.height === _types.default.AUTOSIZE ? height : p_sizeProperty.height;
    return _types.default.newSize(width, height);
  };

  control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
    _link.default.processControlLink(this);

    this.pushToDisplayList(displayListFunction.call(this, p_localBbox));
    let paddingLeft = horizontalPaddingFunction.call(this, _types.default.Positions.LEFT);
    let paddingRight = horizontalPaddingFunction.call(this, _types.default.Positions.RIGHT);
    let menuIconSize = this.menuIconSize();
    let iconPadding = this.style(stylePrefix + "_icon_padding");
    let align = alignFunction.call(this);
    let availableWidthForIconAndText = Math.max(p_localBbox.width - paddingLeft - paddingRight - (menuIconSize ? menuIconSize + iconPadding : 0), 0);
    let measuredTextWidth = Math.max(this.firstText.measuredSize().width, this.secondText ? this.secondText.measuredSize().width : 0);
    let iconAndTextWidth = Math.min(availableWidthForIconAndText, measuredTextWidth + (this.iconData ? this.iconData.size + iconPadding : 0));
    let iconAndTextX = paddingLeft; // LEFT align

    if (align === _types.default.TextAlign.CENTER) {
      iconAndTextX += (availableWidthForIconAndText - iconAndTextWidth) / 2;
    } else if (align === _types.default.TextAlign.RIGHT) {
      iconAndTextX += availableWidthForIconAndText - iconAndTextWidth;
    }

    let colors = this.textAndIconAndLinkColor(); // Icon

    let maxTextWidth = this.secondText ? Math.max(this.firstText.measuredSize().width, this.secondText.measuredSize().width) : this.firstText.measuredSize().width;
    let iconAndTextVisibility = this.iconData ? (this.getActualProperty("text") ? p_localBbox.width >= this.iconData.size + maxTextWidth + menuIconSize + paddingLeft + paddingRight + iconPadding : p_localBbox.width >= this.iconData.size) && p_localBbox.height >= this.iconData.size : false;
    let iconAndMenuiconVisibility = this.iconData ? menuIconSize > 0 ? p_localBbox.width - menuIconSize - 1.5 * iconPadding >= this.iconData.size : p_localBbox.width >= this.iconData.size : false;
    let iconDataVisibility = iconAndMenuiconVisibility && iconAndTextVisibility;

    if (this.iconData && iconDataVisibility) {
      let theX = !this.getActualProperty("text") && !menuIconSize > 0 && p_localBbox.width - paddingLeft - paddingRight < this.iconData.size ? (p_localBbox.width - this.iconData.size) / 2 : iconAndTextX;
      let theY = (p_localBbox.height - this.iconData.size) / 2;

      let icon = _icon.default.makeDisplayItemsFromIconData(this, this.iconData, theX, theY, colors.iconColor);

      this.pushToDisplayList(icon);
    } // Menu icon


    let menuIconVisibility = menuIconSize > 0 ? p_localBbox.width > menuIconSize + 2 * iconPadding + paddingLeft : false;

    if (menuIconSize > 0 && menuIconVisibility) {
      let menuIconX = p_localBbox.width - iconPadding * 1.5 - menuIconSize;
      let menuIconY = (p_localBbox.height - menuIconSize) / 2;
      this.pushToDisplayList(_displayitem.default.newPathsAsset(this._context, "DownFillTriangleIcon", _types.default.newRectangle(menuIconX, menuIconY, menuIconSize, menuIconSize), colors.iconColor));
    } // Text


    let iconWidth = this.iconData && iconDataVisibility ? this.iconData.size + iconPadding : 0;
    let textX = iconAndTextX + iconWidth;
    let textWidth = Math.min(availableWidthForIconAndText - iconWidth, measuredTextWidth);
    let totalTextHeight = this.firstText.measuredSize().height;

    if (this.secondText) {
      totalTextHeight += this.style(stylePrefix + "_padding_middle") + this.secondText.measuredSizeForWidth(textWidth).height;
    }

    let firstTextY = (p_localBbox.height - totalTextHeight) / 2;
    let textAlign = this.iconData ? _types.default.TextAlign.LEFT : align;

    if (textWidth > 0 && this.firstText.measuredSize().height <= p_localBbox.height) {
      let firstTextRect = _types.default.newRectangle(textX, firstTextY, textWidth, this.firstText.measuredSize().height);

      if (!hasTextLinks) {
        this.pushToDisplayList(this.firstText.makeDisplayItems(firstTextRect, textAlign));
      } else {
        let obj = this.firstText.makeDisplayItemsAndLinks(firstTextRect, textAlign);

        _link.default.processTextLinks(this, obj.links);

        this.pushToDisplayList(obj.displayItems);
      }

      if (this.secondText) {
        let secondTextY = firstTextY + this.firstText.measuredSize().height + this.style(stylePrefix + "_padding_middle");

        let secondTextRect = _types.default.newRectangle(textX, secondTextY, textWidth, totalTextHeight - this.firstText.measuredSize().height);

        if (!hasTextLinks) {
          this.pushToDisplayList(this.secondText.makeDisplayItems(secondTextRect, textAlign));
        } else {
          let obj = this.secondText.makeDisplayItemsAndLinks(secondTextRect, textAlign);

          _link.default.processTextLinks(this, obj.links);

          this.pushToDisplayList(obj.displayItems);
        }
      }
    }
  };

  return control;
};

const button = makeButtonControlClass({
  controlTypeID: "com.balsamiq.mockups::Button",
  controlTypeName: "Button",
  stylePrefix: "button",
  controlAlternateNames: [],
  sizeOptions: {
    autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT,
    resize: Props.ResizeOptions.BOTH
  },
  defaultText: "Button",
  defaultFontSize: 13,
  customProperties: {
    align: {
      defaultValue: _types.default.TextAlign.CENTER
    },
    state: {
      defaultValue: Props.States.NORMAL,
      enum: Props.ButtonStates
    }
  },
  alignFunction: function () {
    return this.getActualProperty("align");
  },
  horizontalPaddingFunction: function (p_side) {
    return this.style("button_padding_" + p_side);
  },
  minSizeFunction: function () {
    return _types.default.newSize(0, this.styleSVGSize("button").height);
  },
  displayListFunction: function (p_localBbox) {
    // Background
    let backgroundBounds = _types.default.paddedRectangle(p_localBbox, this.style('button_background_padding_top'), this.style('button_background_padding_bottom'), this.style('button_background_padding_left'), this.style('button_background_padding_right'));

    let backgroundColor;

    if (this.getActualProperty('state') === _types.default.ControlState.SELECTED) {
      backgroundColor = this.style('default_selection_color');
    } else {
      backgroundColor = this.getActualProperty('color');
    }

    let rectangle;

    if (backgroundBounds.width > 0 && backgroundBounds.height > 0) {
      rectangle = Rectangle.makeFilledRectangle(backgroundBounds, backgroundColor);
    } // Border


    let borderColor;

    if (_types.default.ControlState.isDisabledState(this.getActualProperty('state'))) {
      borderColor = this.style('button_disabled_border_color');
    } else if (this.getActualProperty('state') === _types.default.ControlState.FOCUSED) {
      borderColor = this.style('default_selection_color');
    } else {
      borderColor = this.style('button_border_color');
    }

    let borderItems = _displayitem.default.createDisplayItemsFromSkin("button", borderColor, p_localBbox, this._context);

    return [rectangle, borderItems];
  },
  multiline: false,
  hasTextLinks: false
});
exports.button = button;
const pointybutton = makeButtonControlClass({
  controlTypeID: "com.balsamiq.mockups::PointyButton",
  controlTypeName: "Pointy Button",
  stylePrefix: "pointybutton",
  controlInitialProperties: {
    borderStyle: Props.BorderStyles.NONE,
    direction: _types.default.TextAlign.LEFT
  },
  sizeOptions: {
    autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT,
    resize: Props.ResizeOptions.HORIZONTAL
  },
  defaultText: "Button",
  defaultFontSize: 13,
  customProperties: {
    direction: {
      defaultValue: _types.default.TextAlign.RIGHT
    },
    ...Props.PropertyDefinitions.BORDER_STYLE(Props.BorderStyles.SQUARE, Props.SimpleBorderStyles),
    newLayout: {
      // TODO remove this unused property
      defaultValue: ""
    },
    state: {
      defaultValue: Props.States.NORMAL,
      enum: Props.BoolStates
    }
  },
  alignFunction: function () {},
  // Not used
  horizontalPaddingFunction: function () {},
  minSizeFunction: function () {},
  displayListFunction: function () {},
  multiline: false,
  hasTextLinks: false
});
exports.pointybutton = pointybutton;

pointybutton.CONTROL_PROTOTYPE.ios7 = function () {
  return this.getActualProperty("borderStyle") !== Props.BorderStyles.SQUARE;
};

pointybutton.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  let width;
  let height;
  const paddingLeft = this.style("pointybutton_padding_left");
  const paddingRight = this.style("pointybutton_padding_right");
  const paddingTop = this.style("pointybutton_padding_top");
  const paddingBottom = this.style("pointybutton_padding_bottom");
  const iconWidth = this.iconData ? this.iconData.size : 0;
  const iconPadding = this.firstText.measuredSize().width !== 0 && this.iconData ? this.style("pointybutton_icon_padding") : 0;
  const menuIconProperty = this.getActualProperty("menuIcon");
  const menuIconPadding = menuIconProperty ? this.style("pointybutton_icon_padding") : 0;
  const menuIconWidth = menuIconProperty ? this.menuIconSize() : 0;
  const pointPadding = this.style("pointybutton_point_padding");
  const iconLabelandMenuWidth = iconWidth + iconPadding + this.firstText.measuredSize().width + menuIconPadding + menuIconWidth;
  const direction = this.getActualProperty("direction");
  this.buttonArrow = false;

  if (!this.ios7()) {
    const size = this.styleSVGSize("pointybutton_" + direction + "_background");
    this.skinNativeHeight = size.height;
  } else {
    // IOS7
    this.skinNativeHeight = this.style("pointybutton_noborder_height");

    if (direction !== _types.default.TextAlign.CENTER) {
      this.buttonArrow = true;
      this.buttonArrowWidth = 12;
      this.buttonArrowHeight = 12;
      this.buttonArrowRotation = direction === _types.default.TextAlign.LEFT ? 180 : 0;
    }
  }

  if (direction === _types.default.TextAlign.RIGHT) {
    width = paddingLeft + iconLabelandMenuWidth + pointPadding;
  } else if (direction === _types.default.TextAlign.CENTER) {
    width = paddingLeft + iconLabelandMenuWidth + paddingRight;
  } else {
    width = pointPadding + iconLabelandMenuWidth + paddingRight;
  }

  height = Math.max(this.skinNativeHeight, this.firstText.measuredSize().height) + paddingTop + paddingBottom;

  if (this.iconData) {
    height = Math.max(height, this.iconData.size);
  }

  if (menuIconProperty) {
    height = Math.max(height, this.menuIconSize());
  } // respect autosize


  width = p_sizeProperty.width === _types.default.AUTOSIZE ? width : p_sizeProperty.width;
  height = p_sizeProperty.height === _types.default.AUTOSIZE ? height : p_sizeProperty.height;
  return _types.default.newSize(width, height);
};

pointybutton.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this);

  let labelWidth = this.firstText.measuredSize().width;
  let buttonArrowY;

  if (this.buttonArrow) {
    buttonArrowY = (p_localBbox.height - this.buttonArrowHeight) / 2;
  }

  let pointPadding = this.style("pointybutton_point_padding");
  let paddingLeft = this.style("pointybutton_padding_left");
  let paddingRight = this.style("pointybutton_padding_right");
  let iconPadding = this.firstText.measuredSize().width !== 0 && this.iconData ? this.style("pointybutton_icon_padding") : 0;
  const menuIconProperty = this.getActualProperty("menuIcon");
  let menuIconPadding = menuIconProperty ? this.style("pointybutton_icon_padding") : 0;
  let menuIconWidth = menuIconProperty ? this.menuIconSize() : 0;
  let iconWidth = this.iconData ? this.iconData.size : 0;
  let labelPlusIconWidth = iconWidth + iconPadding + Math.min(labelWidth, p_localBbox.width);
  let iconLabelandMenuWidth = labelPlusIconWidth + menuIconPadding + menuIconWidth;
  let direction = this.getActualProperty("direction");

  if (labelWidth > p_localBbox.width) {
    paddingLeft = paddingRight = pointPadding = 0;
    this.firstText.makeLinesForWidth(p_localBbox.width);
  }

  let labelX;
  let buttonArrowX;
  let menuIconX;

  if (direction === _types.default.TextAlign.RIGHT) {
    labelX = Math.max(0, paddingLeft + (p_localBbox.width - paddingLeft - iconLabelandMenuWidth - pointPadding) / 2 + iconWidth + iconPadding);

    if (this.buttonArrow) {
      buttonArrowX = p_localBbox.width - this.buttonArrowWidth - this.style("pointybutton_icon_padding");
    }

    if (menuIconProperty) {
      menuIconX = p_localBbox.width - pointPadding - this.menuIconSize();
    }
  } else if (direction === _types.default.TextAlign.LEFT) {
    labelX = Math.max(0, pointPadding + (p_localBbox.width - pointPadding - iconLabelandMenuWidth - paddingRight) / 2 + iconWidth + iconPadding);

    if (this.buttonArrow) {
      buttonArrowX = this.style("pointybutton_icon_padding");
    }

    if (menuIconProperty) {
      menuIconX = p_localBbox.width - paddingRight - this.menuIconSize();
    }
  } else {
    // MIDDLE
    labelX = Math.max(0, paddingLeft + (p_localBbox.width - paddingLeft - iconLabelandMenuWidth - paddingRight) / 2 + iconWidth + iconPadding);

    if (menuIconProperty) {
      menuIconX = Math.max(0, p_localBbox.width - paddingRight - this.menuIconSize());
    }
  }

  let iconX;

  if (this.iconData) {
    iconX = Math.max(0, labelX - iconPadding - this.iconData.size);
  } // Background and border


  if (this.ios7()) {
    // Draw background
    this.pushToDisplayList(Rectangle.makeFilledRectangle(p_localBbox, this.getActualProperty("color"), 1, this.style("pointybutton_background_border_radius")));
  } else {
    var borderColor = this.getActualProperty("state") === _types.default.ControlState.DISABLED ? this.style("pointybutton_disabled_color") : this.style("default_border_color");
    this.pushToDisplayList(_displayitem.default.createDisplayItemsFromSkin("pointybutton_" + direction + "_background", this.getActualProperty("color"), p_localBbox, this._context));
    this.pushToDisplayList(_displayitem.default.createDisplayItemsFromSkin("pointybutton_" + direction, borderColor, p_localBbox, this._context));
  } // Text


  var colors = this.textAndIconAndLinkColor();
  var labelY = (p_localBbox.height - this.firstText.measuredSize().height) / 2;
  this.pushToDisplayList(this.firstText.makeDisplayItems(_types.default.newRectangle(labelX, labelY, this.firstText.measuredSize().width, this.firstText.measuredSize().height), _types.default.TextAlign.CENTER));

  if (this.iconData) {
    this.pushToDisplayList(_icon.default.makeDisplayItemsFromIconData(this, this.iconData, iconX, (p_localBbox.height - this.iconData.size) / 2, colors.iconColor));
  } // Menu Icon


  if (menuIconProperty) {
    var menuIconY = (p_localBbox.height - this.menuIconSize()) / 2;
    this.pushToDisplayList(_displayitem.default.newPathsAsset(this._context, "DownFillTriangleIcon", _types.default.newRectangle(menuIconX, menuIconY, this.menuIconSize(), this.menuIconSize()), colors.iconColor));
  } // ios7 Arrow


  if (this.buttonArrow) {
    if (direction === _types.default.TextAlign.LEFT) {
      this.pushToDisplayList(_displayitem.default.newRotated180PathsAsset(this._context, "GtIcon", _types.default.newRectangle(buttonArrowX, buttonArrowY, this.buttonArrowWidth, this.buttonArrowHeight), colors.iconColor));
    } else if (direction === _types.default.TextAlign.RIGHT) {
      this.pushToDisplayList(_displayitem.default.newPathsAsset(this._context, "GtIcon", _types.default.newRectangle(buttonArrowX, buttonArrowY, this.buttonArrowWidth, this.buttonArrowHeight), colors.iconColor));
    }
  }
};

const multilinebutton = makeButtonControlClass({
  controlTypeID: MultilineButtonControlTypeID,
  controlTypeName: "Multiline Button",
  stylePrefix: "multilinebutton",
  controlAlternateNames: [],
  sizeOptions: {
    autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT,
    resize: Props.ResizeOptions.BOTH
  },
  defaultText: "Multiline Button\nSecond line of text",
  defaultFontSize: 16,
  splitOn: Props.SplitOnTypes.NEWLINE,
  customProperties: { ...Props.PropertyDefinitions.ALPHA()
  },
  alignFunction: () => _types.default.TextAlign.CENTER,
  horizontalPaddingFunction: function (p_side) {
    return this.style("multilinebutton_padding_" + p_side);
  },
  minSizeFunction: () => _types.default.newSize(28, 66),
  displayListFunction: function (p_localBbox) {
    const background = Rectangle.makeFilledRectangle(p_localBbox, this.getActualProperty("color"), this.getActualProperty("backgroundAlpha"), this.style("multilinebutton_background_radius"));

    const border = _displayitem.default.createDisplayItemsFromSkin("rectangle_border_rounded", this.style("default_border_color"), p_localBbox, this._context);

    return [background, border];
  },
  multiline: true,
  hasTextLinks: true
});
exports.multilinebutton = multilinebutton;
var _default = {
  button,
  pointybutton,
  multilinebutton
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/calendar.js":
/*!**************************************!*\
  !*** ./js/core/controls/calendar.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const Months = ["JANUARY", "FEBRUARY", "MARCH", "APRIL", "MAY", "JUNE", "JULY", "AUGUST", "SEPTEMBER", "OCTOBER", "NOVEMBER", "DECEMBER"];
const Days = ["S", "M", "T", "W", "T", "F", "S"];
const control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::Calendar",
  controlTypeName: "Calendar",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    })
  }
});

control.CONTROL_PROTOTYPE.measure = function () {
  return _types.default.newSize(96, 96);
};

control.CONTROL_PROTOTYPE.initData = function () {
  this.cachedTexts = {};
};

control.CONTROL_PROTOTYPE.updateData = function () {
  const date = new Date();
  const month = date.getMonth();
  const year = date.getFullYear();
  this.date = Months[month] + ' ' + year;
};

control.CONTROL_PROTOTYPE.updateLayout = function (localBbox) {
  if (!this.date) {
    return;
  }

  const today = new Date();
  const iconSize = this.style("calendar_icon_size");
  const paddingTop = this.style("calendar_padding_top");
  const paddingBottom = this.style("calendar_padding_bottom");
  const paddingLeft = this.style("calendar_padding_left");
  const borderLine = this.style("calendar_border_weight");
  const daysLine = this.style("calendar_lines_weight");
  const grey = 0xDDDDDD;
  const textStyle = this.newTextStyle();
  const projectFontSize = textStyle.fontSize;
  let i = 0;

  while (i < 11 && i + 8 < projectFontSize && 97 + 10 * (i + 1) < localBbox.width - borderLine / 2 && 97 + 10 * (i + 1) < localBbox.height - borderLine) {
    i++;
  }

  const fontSize = i + 8;
  textStyle.fontSize = fontSize;

  if (!this.text || !Text.areTextsEqual(this.text, this.date, textStyle)) {
    this.text = Text.newTextNoWrap(this.date, textStyle, this._context);
  }

  if (this.fontSize !== fontSize) {
    this.day = Text.newTextNoWrap(95 .toString(), textStyle, this._context);
    this.fontSize = fontSize;
  }

  const {
    height: dayHeight,
    width: dayWidth
  } = this.day.measuredSize();
  const color = this.style("calendar_color");
  this.pushToDisplayList(Rectangle.makeFilledRectangle(localBbox));
  this.pushToDisplayList(Rectangle.makeStrokedRectangle(localBbox, color, borderLine));
  const dateMeasuredSize = this.text.measuredSize();
  const dateHeight = Math.max(iconSize + 2, dateMeasuredSize.height + 3 * paddingTop + paddingBottom);

  const line = _types.default.newRectangle(0, 10 + dateMeasuredSize.height, localBbox.width, borderLine);

  this.pushToDisplayList(Rectangle.makeStrokedRectangle(line, color, borderLine));
  const firstDay = new Date(today.getFullYear(), today.getMonth(), 1);
  let theX = borderLine - 1.5;
  let theY = dateHeight + dayHeight * 1.5 + 4 * paddingTop;
  const space = (localBbox.height - theY) / 6 - dayHeight;
  let daysVisibility = true;

  if (dayWidth * 7 + paddingLeft + 2 + borderLine * 2 <= localBbox.width && theY + dayHeight * 6 + borderLine * 2 < localBbox.height) {
    // Add day names
    for (let dayIndex = 0; dayIndex < 7; dayIndex++) {
      const x = (localBbox.width - 2 * borderLine) / 7 * dayIndex + paddingLeft / 2 + borderLine;
      const y = dateHeight + dayHeight * 0.5;
      this.addText(Days[dayIndex], x, y, dayWidth, dayHeight, fontSize);
    } // Add numbers in the calendar


    if (7 - firstDay.getDay() > 0) {
      textStyle.color = grey;
      const prevMonthDays = this.daysInMonth(firstDay.getMonth() - 1, firstDay.getYear());

      for (let dayNumber = prevMonthDays - firstDay.getDay() + 1; dayNumber <= prevMonthDays; dayNumber++) {
        const dayWidth = (localBbox.width - 2 * borderLine) / 7;
        this.addText(dayNumber.toString(), theX + paddingLeft, theY, dayWidth, dayHeight, fontSize, grey);
        theX += localBbox.width / 7 - 1.5;
      }
    }

    const currMonthDays = this.daysInMonth(firstDay.getMonth(), firstDay.getYear());
    let dayIndex = firstDay.getDay() - 1;

    for (let dayNumber = 1; dayNumber <= 42 - dayIndex - 1; dayNumber++) {
      if ((dayNumber + dayIndex) % 7 === 0) {
        theX = borderLine - 1.5;
      }

      if (dayNumber === today.getDate() && localBbox.height > 93 && localBbox.width > 93) {
        const x = fontSize * 1.5;
        const path = Path.newEllipsePath(x, x, 1);
        const rectangleX = dayNumber > 9 ? theX + paddingLeft / 2 + 2 : theX + 1;
        const rectangleY = theY + (space + dayHeight) * Math.floor((dayNumber + dayIndex) / 7) - paddingTop - (fontSize < 9 ? 0 : 1);

        const bounds = _types.default.newRectangle(rectangleX, rectangleY, x, x);

        const circle = _displayitem.default.newSVGPath(bounds, path, 1, this.style("circleColor"), this.style('default_selection_color'));

        this.pushToDisplayList(circle);
      }

      const x = theX + paddingLeft;
      const y = theY + (space + dayHeight) * Math.floor((dayNumber + dayIndex) / 7) + 1;
      const width = (localBbox.width - 2 * borderLine) / 7;

      if (dayIndex + dayNumber < 7) {
        this.addText(dayNumber.toString(), x, theY, width, dayHeight, fontSize);
      } else if (dayNumber <= currMonthDays) {
        const color = dayNumber === today.getDate() ? _colors.default.labelColorGivenBackgroundColor(this.style('default_selection_color'), 1, this.style("default_text_color")) : undefined;
        this.addText(dayNumber.toString(), x, y, width, dayHeight, fontSize, color);
      } else {
        const x = theX + paddingLeft;
        const y = theY + (space + dayHeight) * Math.floor((dayNumber + dayIndex) / 7) + 1;
        const width = (localBbox.width - 2 * borderLine) / 7;
        this.addText((dayNumber - currMonthDays).toString(), x, y, width, dayHeight, fontSize, grey);
      }

      theX += localBbox.width / 7 - 1.5;
    }
  } else {
    daysVisibility = false;
  }

  const fullDate = today.getDate() + " " + this.date;

  if (!this.fullDate || !Text.areTextsEqual(this.fullDate, fullDate, textStyle)) {
    this.fullDate = Text.newTextNoWrap(fullDate, textStyle, this._context);
  }

  let drawArrows = true;

  if (!this.topDate) {
    this.topDate = !daysVisibility && this.fullDate.measuredSize().width < localBbox.width ? this.fullDate : this.text;
  }

  if (localBbox.width < this.topDate.measuredSize().width + 2 * paddingLeft) {
    drawArrows = false;
    this.topDate = this.text;

    if (localBbox.width < this.topDate.measuredSize().width + 2 * paddingLeft) {
      const text = this.date.slice(0, 3) + " '" + this.date.slice(-2);

      if (!this.shortTopDate || !Text.areTextsEqual(this.shortTopDate, text, textStyle)) {
        this.shortTopDate = Text.newTextNoWrap(text, textStyle, this._context);
      }

      this.topDate = this.shortTopDate;

      if (localBbox.width < this.topDate.measuredSize().width + 2 * paddingLeft) {
        const text = this.date.slice(0, 3);

        if (!this.moreShortTopDate || !Text.areTextsEqual(this.moreShortTopDate, text, textStyle)) {
          this.moreShortTopDate = Text.newTextNoWrap(text, textStyle, this._context);
        }

        this.topDate = this.text;
      }
    }
  }

  let date;

  if (localBbox.width >= this.topDate.measuredSize().width + 2 * borderLine && localBbox.height >= line.y + line.height) {
    const dateRect = _types.default.newRectangle(borderLine, borderLine, localBbox.width - 2 * borderLine, dateHeight - 2 * borderLine);

    date = this.topDate.makeDisplayItems(dateRect, _types.default.TextAlign.CENTER, _types.default.Positions.CENTER);
  }

  let iconLeft;
  let iconRight;

  if (localBbox.height >= iconSize + 2 * borderLine && localBbox.width >= this.topDate.measuredSize().width + 4 * iconSize && drawArrows) {
    const iconBounds = _types.default.newRectangle(borderLine, (dateHeight - iconSize) / 2, iconSize, iconSize);

    iconLeft = _displayitem.default.newPathsAsset(this._context, "LeftFillTriangleIcon", iconBounds, color);
    iconBounds.x = localBbox.width - borderLine - iconSize;
    iconRight = _displayitem.default.newPathsAsset(this._context, "RightFillTriangleIcon", iconBounds, color);
  } // Draw horizontal and vertical lines


  if (localBbox.height >= 200 && localBbox.width >= 200) {
    theX = borderLine + localBbox.width / 7 - 1.5;

    for (i = 0; i < 6; i++) {
      const rectangleX = borderLine / 2;
      const rectangleY = theY + i * (dayHeight + space) - 5;

      const bounds = _types.default.newRectangle(rectangleX, rectangleY, localBbox.width - borderLine, daysLine);

      this.pushToDisplayList(Rectangle.makeStrokedRectangle(bounds, color, daysLine));
    }

    for (i = 1; i < 7; i++) {
      const rectangleHeight = localBbox.height - theY - dayHeight - borderLine + 1.5 * dayHeight + 5;

      const bounds = _types.default.newRectangle(theX, theY - 5, daysLine, rectangleHeight);

      this.pushToDisplayList(Rectangle.makeStrokedRectangle(bounds, color, daysLine));
      theX += localBbox.width / 7 - 1.5;
    }
  }

  this.pushToDisplayList(date, iconLeft, iconRight);
};

control.CONTROL_PROTOTYPE.addText = function (label, x, y, width, height, fontSize, color) {
  const textStyle = this.newTextStyle();

  if (color) {
    textStyle.textColor = color;
  }

  textStyle.fontSize = fontSize;

  if (!this.cachedTexts[label] || !Text.areTextsEqual(this.cachedTexts[label], label, textStyle)) {
    this.cachedTexts[label] = Text.newText(label, textStyle, this._context);
  }

  const bounds = _types.default.newRectangle(x, y, width, height);

  const displayItems = this.cachedTexts[label].makeDisplayItems(bounds, _types.default.TextAlign.LEFT);
  this.pushToDisplayList(displayItems);
};

control.CONTROL_PROTOTYPE.daysInMonth = function (month, year) {
  return new Date(year, month + 1, 0).getDate();
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/callout.js":
/*!*************************************!*\
  !*** ./js/core/controls/callout.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _macro = __webpack_require__(/*! core/bricks/macro */ "./js/core/bricks/macro.js");

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::CallOut",
  controlTypeName: "Callout",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        resize: Props.ResizeOptions.BOTH,
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT
      }
    }),
    ...Props.PropertyDefinitions.TEXT({
      defaultText: "1",
      multiline: true
    }),
    markup: {
      defaultValue: true,
      dataType: 'bool'
    }
  }
});
controlsMod.addStandardTextPropertyDefinitions(control, {
  text: "1"
});
controlsMod.addColorPropertyDefinitions(control, {
  defaultColor: "callout_default_color",
  hasAlpha: true
});

control.CONTROL_PROTOTYPE.updateData = function () {
  var text = (0, _macro.processMacros)(this);
  var textStyle = this.newStandardTextStyle();
  textStyle.textColor = _colors.default.labelColorGivenBackgroundColor(this.getActualProperty("color"), this.getActualProperty("backgroundAlpha"), this.style('default_text_color'));
  textStyle.linkColor = _colors.default.labelColorGivenBackgroundColor(this.getActualProperty("color"), this.getActualProperty("backgroundAlpha"), this.style('default_link_color'));
  this.text = textMod.newText(text, textStyle, this._context, true);
};

_link.default.addTextLinksMethods(control);

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var width = Math.max(39, this.text.measuredSize().width + this.style("callout_padding_left") + this.style("callout_padding_right"));
  var height = Math.max(39, this.text.measuredSize().height + this.style("callout_padding_top") + this.style("callout_padding_bottom"));
  return _types.default.newSize(width, height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var border;

  if (this.style("callout_skin_count") > 0) {
    var assetNumber = this.controlID % this.style("callout_skin_count") + 1;
    var assetName = "callout_" + assetNumber;
    border = _displayitem.default.createDisplayItemsFromSkin(assetName, this.style("default_border_color"), p_localBbox, this._context);
  }

  var backgroundPath = Path.newRoundedRectanglePath(p_localBbox.width, p_localBbox.height, 0, 30);

  var background = _displayitem.default.newSVGPath(p_localBbox, backgroundPath, 0, null, this.getActualProperty("color"));

  background.alpha = this.getActualProperty("backgroundAlpha");

  var textBounds = _types.default.paddedRectangle(p_localBbox, this.style("callout_padding_top"), this.style("callout_padding_bottom"), this.style("callout_padding_left"), this.style("callout_padding_right"));

  var label = this.text.makeDisplayItemsAndLinks(textBounds, _types.default.TextAlign.LEFT, _types.default.Positions.CENTER);
  this.pushToDisplayList(background, border, label.displayItems);

  _link.default.processTextLinks(this, label.links);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/canvas.js":
/*!************************************!*\
  !*** ./js/core/controls/canvas.js ***!
  \************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _scrollbars = __webpack_require__(/*! core/bricks/scrollbars */ "./js/core/bricks/scrollbars.js");

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var _breakline = __webpack_require__(/*! core/bricks/breakline */ "./js/core/bricks/breakline.js");

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var Snapping = _interopRequireWildcard(__webpack_require__(/*! core/controls/snapping */ "./js/core/controls/snapping.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var canvas = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::Canvas",
  controlTypeName: "Rectangle",
  controlAlternateNames: ["Canvas", "Panel"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(100, 70),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.BORDER_STYLE(Props.BorderStyles.SQUARE, Props.CanvasBorderStyles),
    ...Props.PropertyDefinitions.SCROLLBAR(),
    ...Props.PropertyDefinitions.link,
    ...Props.PropertyDefinitions.QUICKDRAW_TIP("You can draw Rectangles quickly by holding down the *R key* and dragging. [Learn more]($DOCURL)", "You can also draw Rectangles with your fingers using the Quick Draw Toolbar (in the View Menu). [Learn more]($DOCURL)")
  }
});
controlsMod.addColorPropertyDefinitions(canvas, {
  defaultColor: 0xFFFFFF,
  defaultBorderColor: "default_border_color",
  hasAlpha: true
});

canvas.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this);

  var displayItems = [];
  var borderStyle = this.getActualProperty("borderStyle");
  displayItems = displayItems.concat((0, _rectangle.makeRectangleDisplayItems)(this._context, p_localBbox, borderStyle, this.getActualProperty("color"), this.getActualProperty("backgroundAlpha"), this.getActualProperty("borderColor"))); // Scrollbar

  if (this.getActualProperty('verticalScrollbar')) {
    displayItems = displayItems.concat((0, _scrollbars.makeVerticalScrollbar)(p_localBbox, this.style("canvas_scrollbar_padding_top"), this.style("canvas_scrollbar_padding_bottom"), this.style("canvas_scrollbar_padding_right"), this.getActualProperty('value'), _types.default.ControlState.NORMAL, this._context));
  }

  var maskId, y;

  if (borderStyle === Props.BorderStyles.SQUARE_BREAKLINE && p_localBbox.width >= 65 && p_localBbox.height >= 40) {
    y = p_localBbox.y + p_localBbox.height - _types.default.BREAKLINE_VERTICAL_OFFSET;
    maskId = (0, _breakline.addBreaklineMask)(this, p_localBbox.x, y, _types.default.Positions.BOTTOM);

    _displayitem.default.maskDisplayItems(displayItems, maskId);

    displayItems = displayItems.concat((0, _breakline.makeBreaklineDisplayitem)(p_localBbox.x, y, p_localBbox.width, undefined, this.style("canvas_breakline_thickness"), this.getActualProperty("borderColor")));
  } else if (borderStyle === Props.BorderStyles.SQUARE_BREAKLINE_TOP) {
    y = _types.default.BREAKLINE_VERTICAL_OFFSET;
    maskId = (0, _breakline.addBreaklineMask)(this, p_localBbox.x, y, _types.default.Positions.TOP);

    _displayitem.default.maskDisplayItems(displayItems, maskId);

    displayItems = displayItems.concat((0, _breakline.makeBreaklineDisplayitem)(p_localBbox.x, y, p_localBbox.width, undefined, this.style("canvas_breakline_thickness"), this.getActualProperty("borderColor")));
  }

  this.pushToDisplayList(displayItems);
};

canvas.CONTROL_PROTOTYPE.getSnapLines = function (withMarkup) {
  _balsamiq.default.assertBool(withMarkup);

  return Snapping.makeSnapLinesForBbox(this.getUserBBox(withMarkup), true);
};

var _default = canvas;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/choicecontrols.js":
/*!********************************************!*\
  !*** ./js/core/controls/choicecontrols.js ***!
  \********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.radiobutton = exports["default"] = exports.checkbox = void 0;

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _choice = __webpack_require__(/*! core/bricks/choice */ "./js/core/bricks/choice.js");

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

var makeChoiceControlClass = function (controlTypeID, controlTypeName, choiceType, defaultText) {
  var control = Controls.makeControlClass({
    controlTypeID,
    controlTypeName,
    controlAlternateNames: [],
    props: { ...Props.PropertyDefinitions.SIZE({
        defaultValue: _types.default.newAutoSize(),
        options: {
          autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT,
          resize: Props.ResizeOptions.BOTH
        }
      }),
      state: {
        defaultValue: Props.RadioButtonStates.NORMAL,
        enum: Props.RadioButtonStates
      },
      ...Props.PropertyDefinitions.link,
      ...Props.PropertyDefinitions.ROTATION({
        isIcon: true
      })
    }
  });
  Controls.addStandardTextPropertyDefinitions(control, {
    text: defaultText
  });
  Controls.addColorPropertyDefinitions(control, {
    defaultTextColor: "default_text_color",
    oldTextColorName: 'color'
  });

  control.CONTROL_PROTOTYPE.initData = function () {
    this._clipOnBbox = true;
    this.overflow = "visible";
  };

  control.CONTROL_PROTOTYPE.updateData = function () {
    const textColor = this.getActualProperty("textColor");
    this.choice = new _choice.Choice({
      type: choiceType,
      state: this.getActualProperty("state"),
      text: this.getActualProperty("text"),
      textStyle: this.newStandardTextStyle({
        textColor
      }),
      color: textColor,
      icon: this.getActualProperty("icon"),
      iconRotation: this.getActualProperty("rotation"),
      previousChoice: this.choice,
      context: this._context
    });
    this.setDependencies(this.choice.dependencies);
  };

  _link.default.addTextLinksMethods(control, {
    getLinksFunction: function () {
      return this.choice.getLinks();
    }
  });

  control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
    return this.choice.measure();
  };

  control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
    _link.default.processControlLink(this);

    var obj = this.choice.makeDisplayItemsAndLinks(this, p_localBbox);
    this.pushToDisplayList(obj.displayItems);

    _link.default.processTextLinks(this, obj.links);
  };

  return control;
};

const radiobutton = makeChoiceControlClass("com.balsamiq.mockups::RadioButton", "Radio Button", _choice.ChoiceType.RADIO, "Radio Button");
exports.radiobutton = radiobutton;
const checkbox = makeChoiceControlClass("com.balsamiq.mockups::CheckBox", "Checkbox", _choice.ChoiceType.CHECKBOX, "Checkbox");
exports.checkbox = checkbox;
var _default = {
  radiobutton,
  checkbox
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/choicegroupcontrols.js":
/*!*************************************************!*\
  !*** ./js/core/controls/choicegroupcontrols.js ***!
  \*************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.radiobuttongroup = exports["default"] = exports.checkboxgroup = void 0;

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _choice = __webpack_require__(/*! core/bricks/choice */ "./js/core/bricks/choice.js");

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const makeChoiceGroupControlClass = function (p_controlTypeID, p_controlTypeName, p_choiceType, p_defaultText) {
  const ROW_MIN_HEIGHT = 12;
  const control = Controls.makeControlClass({
    controlTypeID: p_controlTypeID,
    controlTypeName: p_controlTypeName,
    controlAlternateNames: [],
    props: { ...Props.PropertyDefinitions.SIZE({
        defaultValue: _types.default.newAutoSize(),
        options: {
          autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT,
          resize: Props.ResizeOptions.BOTH
        }
      }),
      ...Props.PropertyDefinitions.links()
    }
  });
  Controls.addStandardTextPropertyDefinitions(control, {
    text: p_defaultText,
    multiline: true,
    splitOn: Props.SplitOnTypes.NEWLINE,
    noEmptyText: true
  });
  Controls.addColorPropertyDefinitions(control, {
    defaultTextColor: "default_text_color",
    oldTextColorName: 'color'
  });

  control.CONTROL_PROTOTYPE.initData = function () {
    this._clipOnBbox = true;
    this.overflow = "hidden";
    this.lines = [];
  };

  control.CONTROL_PROTOTYPE.getLinkNames = function () {
    return this._linkNames;
  };

  control.CONTROL_PROTOTYPE.getLinkableLabels = function (text) {
    const lines = text.split("\n");
    return lines.map(line => TextUtils.unformatText(line));
  };

  control.CONTROL_PROTOTYPE.parseLine = function (line, index) {
    const res = {
      state: null
    };
    let choice = (0, _choice.parseChoiceText)(line);

    if (choice && choice.type !== p_choiceType) {
      choice = null;
    }

    const textStyle = this.newStandardTextStyle();

    if (choice) {
      const isChoice = this.isPreviousTextOfSameLength && !!this.lines[index].choice;
      const choiceColor = choice.color ? _colors.default.colorFromString(choice.color) : this.getActualProperty("textColor");
      textStyle.textColor = choice.textcolor ? _colors.default.colorFromString(choice.textcolor) : this.getActualProperty("textColor");
      res.choice = new _choice.Choice({
        type: p_choiceType,
        state: choice.state,
        text: choice.text,
        textStyle,
        color: choiceColor,
        customLabelPadding: this.style("choicegroup_label_padding"),
        previousChoice: isChoice ? this.lines[index].choice : null,
        context: this._context
      });
      res.size = res.choice.measure();
      res.size.height = Math.max(ROW_MIN_HEIGHT, res.choice.text.measuredSize().height); // use the text height as line height
    } else {
      const isText = this.isPreviousTextOfSameLength && !this.lines[index].choice;
      textStyle.textColor = this.getActualProperty("textColor");
      res.text = isText && Text.areTextsEqual(this.lines[index].text, line, textStyle) ? this.lines[index].text : Text.newTextNoWrap(line, textStyle, this._context);
      res.size = res.text.measuredSize();
    }

    return res;
  };

  control.CONTROL_PROTOTYPE.updateData = function () {
    const text = this.getActualProperty("text");
    const lines = text.split("\n");
    this.isPreviousTextOfSameLength = this.lines.length === lines.length;
    this.lines = lines.map(this.parseLine, this);
    this._linkNames = this.getLinkableLabels(text);
  };

  control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
    let width = 0,
        height = 0;
    this.lines.forEach(line => {
      width = Math.max(width, line.size.width);
      height += line.size.height;
    });
    height += (this.lines.length - 1) * this.style("choicegroup_line_padding");
    height += this.style("choicegroup_padding_top") + this.style("choicegroup_padding_bottom");
    return _types.default.newSize(width, height);
  };

  control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
    const linePadding = this.style("choicegroup_line_padding");
    let y = this.style("choicegroup_padding_top");
    const linkBounds = [];
    let line = this.lines[0];
    let i = 0;

    while (line != null && y + line.size.height <= p_localBbox.height) {
      var bounds = _types.default.newRectangle(0, y, p_localBbox.width, line.size.height);

      linkBounds.push(bounds);

      if (line.text) {
        this.pushToDisplayList(line.text.makeDisplayItems(bounds));
      } else if (line.choice) {
        this.pushToDisplayList(line.choice.makeDisplayItems(this, bounds));
      }

      y += line.size.height + linePadding;
      i++;
      line = this.lines[i];
    }

    _link.default.processLinks(this, linkBounds);
  };

  return control;
};

const radiobuttongroup = makeChoiceGroupControlClass("com.balsamiq.mockups::RadioButtonGroup", "Radio Button Group", _choice.ChoiceType.RADIO, "(o) option 1 (selected)\n( ) option 2\n(-) option 3 (indeterminate)\n-( ) option 4 (disabled)-\n-(o) option 5\\r(disabled and selected)-\n-(-) option 6\\r(disabled indeterminate)-\nA row without a radio button");
exports.radiobuttongroup = radiobuttongroup;
const checkboxgroup = makeChoiceGroupControlClass("com.balsamiq.mockups::CheckBoxGroup", "Checkbox Group", _choice.ChoiceType.CHECKBOX, "[ ] not selected\n[x] selected\n[-] indeterminate\n-[ ] disabled-\n-[x] disabled selected-\n-[-] disabled indeterminate-\nA row without a checkbox");
exports.checkboxgroup = checkboxgroup;
var _default = {
  radiobuttongroup,
  checkboxgroup
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/circlebutton.js":
/*!******************************************!*\
  !*** ./js/core/controls/circlebutton.js ***!
  \******************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::CircleButton",
  controlTypeName: "Circle Button",
  controlAlternateNames: ["Floating Action Button"],
  controlInitialProperties: {
    icon: _icon.default.newAppIconData("plus", 24, 0)
  },
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.BOTH,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.BORDER_STYLE(Props.SimpleBorderStyles.SQUARE, Props.SimpleBorderStyles),
    state: {
      defaultValue: Props.ButtonStates.NORMAL,
      enum: Props.ButtonStates
    },
    labelPosition: {
      defaultValue: Props.IconLabelPositions.BOTTOM,
      enum: Props.IconLabelPositions
    },
    ...Props.PropertyDefinitions.link,
    ...Props.PropertyDefinitions.ROTATION({
      isIcon: true
    })
  }
});
controlsMod.addStandardTextPropertyDefinitions(control);
controlsMod.addColorPropertyDefinitions(control, {
  defaultColor: "circlebutton_bg_color"
});

control.CONTROL_PROTOTYPE.initData = function () {
  this._clipOnBbox = false;
};

control.CONTROL_PROTOTYPE.getColor = function () {
  var normalStateColor = _colors.default.labelColorGivenBackgroundColor(Props.States.isSelectedState(this.getActualProperty('state')) ? this.style('default_selection_color') : this.getActualProperty('color'), null, this.style("default_text_color"));

  return this.getActualProperty('state') === _types.default.ControlState.DISABLED ? this.style('default_disabled_color') : normalStateColor;
};

control.CONTROL_PROTOTYPE.updateData = function () {
  var text = this.getActualProperty("text");
  var textStyle = this.newStandardTextStyle({
    textColor: this.getColor()
  });
  this.text = text.length > 0 ? textMod.newText(text, textStyle, this._context) : null;

  _icon.default.processIconData(this);
};

control.CONTROL_PROTOTYPE.getTextLinks = function () {
  return this.text ? this.text.getLinks() : [];
};

control.CONTROL_PROTOTYPE.measure = function () {
  return _types.default.newSize(48, 48);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this);

  var state = this.getActualProperty('state');
  var diameter = Math.min(p_localBbox.width, p_localBbox.height);

  var shapeBounds = _types.default.newRectangle((p_localBbox.width - diameter) / 2, (p_localBbox.height - diameter) / 2, diameter, diameter);

  var circlePath = Path.newEllipsePath(diameter, diameter, this.style('circlebutton_border_thickness'));
  var circleBorderColor = Props.States.isFocusedState(state) ? this.style('default_selection_color') : Props.States.isDisabledState(state) ? this.style('default_disabled_color') : this.style('default_border_color');
  var circleBackgroundColor = Props.States.isSelectedState(state) ? this.style('default_selection_color') : this.getActualProperty('color');
  var circleBorderThickness = this.getActualProperty("borderStyle") == Props.SimpleBorderStyles.SQUARE ? this.style('circlebutton_border_thickness') : 0;

  var circle = _displayitem.default.newSVGPath(shapeBounds, circlePath, circleBorderThickness, circleBorderColor, circleBackgroundColor);

  this.pushToDisplayList(circle);
  var labelPosition = this.getActualProperty('labelPosition');
  var iconX, iconY, labelX, labelY;
  var textSize = this.text ? this.text.measuredSize() : null;
  var iconAndLabelHeight = (this.iconData ? this.iconData.size : -2) + (textSize ? textSize.height + 2 : 0);
  var iconAndLabelWidth = (this.iconData ? this.iconData.size : -2) + (textSize ? textSize.width + 2 : 0);
  var labelVisible = false;

  if (textSize) {
    if (labelPosition === _types.default.Positions.BOTTOM) {
      labelX = (p_localBbox.width - textSize.width) / 2;
      labelY = (p_localBbox.height - iconAndLabelHeight) / 2 + (this.iconData ? this.iconData.size : 0);
      labelVisible = textSize.height + (this.iconData ? this.iconData.size + 2 : 0) < diameter && textSize.width < diameter;
    } else if (labelPosition === _types.default.Positions.LEFT) {
      labelX = (p_localBbox.width - iconAndLabelWidth) / 2;
      labelY = (p_localBbox.height - textSize.height) / 2;
      labelVisible = textSize.height < diameter && textSize.width + (this.iconData ? this.iconData.size + 2 : 0) < diameter;
    } else {
      // RIGHT
      labelX = (p_localBbox.width - textSize.width + (this.iconData ? this.iconData.size + 2 : 0)) / 2;
      labelY = (p_localBbox.height - textSize.height) / 2;
      labelVisible = textSize.height < diameter && textSize.width + (this.iconData ? this.iconData.size + 2 : 0) < diameter;
    }

    var label = this.text.makeDisplayItemsAndLinks(_types.default.newRectangle(labelX, labelY, textSize.width, textSize.height));

    _link.default.processTextLinks(this, label.links);

    if (labelVisible) {
      this.pushToDisplayList(label.displayItems);
    }
  }

  if (this.iconData) {
    var iconData = _balsamiq.default.clone(this.iconData);

    if (labelPosition === _types.default.Positions.BOTTOM) {
      iconX = (p_localBbox.width - iconData.size) / 2;
      iconY = (p_localBbox.height - (labelVisible ? iconAndLabelHeight : iconData.size)) / 2;
    } else if (labelPosition === _types.default.Positions.LEFT) {
      iconX = (p_localBbox.width - iconData.size + (labelVisible ? textSize.width + 2 : 0)) / 2; // right aligned

      iconY = (p_localBbox.height - iconData.size) / 2;
    } else {
      // RIGHT
      iconX = (p_localBbox.width - (labelVisible ? iconAndLabelWidth : iconData.size)) / 2;
      iconY = (p_localBbox.height - iconData.size) / 2;
    }

    var icon = _icon.default.makeDisplayItemsFromIconData(this, iconData, iconX, iconY, this.getColor());

    if (iconData.size < diameter) {
      this.pushToDisplayList(icon);
    }
  }
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/colorpicker.js":
/*!*****************************************!*\
  !*** ./js/core/controls/colorpicker.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const colorpicker = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::ColorPicker",
  controlTypeName: "Color Picker",
  controlAlternateNames: []
});
Controls.addColorPropertyDefinitions(colorpicker, {
  defaultColor: 0x2B78E4
});

colorpicker.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  const backRectBBox = _types.default.paddedRectangle(p_localBbox, this.style('colorpicker_padding_top'), this.style('colorpicker_padding_bottom'), this.style('colorpicker_padding_left'), this.style('colorpicker_padding_right'));

  const background = Rectangle.makeFilledRectangle(backRectBBox, this.getActualProperty("color"));

  const skin = _displayitem.default.createDisplayItemsFromSkin("colorpicker", 0, p_localBbox, this._context);

  this.pushToDisplayList(background, skin);
};

colorpicker.CONTROL_PROTOTYPE.measure = function () {
  return _types.default.newSize(26, 28);
};

var _default = colorpicker;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/combobox.js":
/*!**************************************!*\
  !*** ./js/core/controls/combobox.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Menu = _interopRequireWildcard(__webpack_require__(/*! core/bricks/menu */ "./js/core/bricks/menu.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var _combobox = _interopRequireDefault(__webpack_require__(/*! core/bricks/combobox */ "./js/core/bricks/combobox.js"));

var ScrollBar = _interopRequireWildcard(__webpack_require__(/*! core/bricks/scrollbars */ "./js/core/bricks/scrollbars.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::ComboBox",
  controlTypeName: "ComboBox",
  controlAlternateNames: ["PullDown Menu", "DropDown List", "Select List"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT,
        resize: Props.ResizeOptions.HORIZONTAL
      }
    }),
    state: {
      defaultValue: Props.ComboboxStates.NORMAL,
      enum: Props.ComboboxStates
    },
    ...Props.PropertyDefinitions.link,
    ...Props.PropertyDefinitions.SCROLLBAR(),
    ...Props.PropertyDefinitions.ROTATION({
      isIcon: true
    })
  }
});
Controls.addStandardTextPropertyDefinitions(control, {
  text: "ComboBox",
  multiline: true,
  splitOn: Props.SplitOnTypes.NEWLINE
});
Controls.addColorPropertyDefinitions(control, {
  defaultBorderColor: "default_text_color",
  oldBorderColorName: 'color',
  defaultTextColor: "default_text_color"
});
Controls.addSelectionPropertyDefinitions(control, 0, {
  noMinusOne: true
});

control.CONTROL_PROTOTYPE.initData = function () {
  //this._lines is an array, each element contains one row of text except for the first one which is removed.
  this._lines = [];
  this._menuLines = [];
  Menu.menuInit(3, 3, 3);
};

var parseLine = function (p_line) {
  p_line = p_line.replace(new RegExp("^o ", "g"), "");
  p_line = p_line.replace(new RegExp("^x ", "g"), "");
  p_line = p_line.replace(new RegExp(" >$", "g"), "");
  return p_line;
};

control.CONTROL_PROTOTYPE.getSelectableLabels = function () {
  return this._selectableLabels;
};

control.CONTROL_PROTOTYPE.updateData = function () {
  const text = this.getActualProperty("text");
  const lines = text.split("\n");
  this._selectableLabels = lines.map(line => TextUtils.unformatText(line));

  const isDisabled = _types.default.ControlState.isDisabledState(this.getActualProperty("state"));

  this.color = isDisabled ? this.style("default_disabled_color") : this.getActualProperty("borderColor");
  const textColor = isDisabled ? this.style("default_disabled_color") : this.getActualProperty("textColor");
  const linkColor = isDisabled ? this.style("default_disabled_color") : this.style("default_link_color");
  const textStyle = this.newStandardTextStyle({
    textColor,
    linkColor
  });
  const firstRowText = lines[this.getActualProperty("selectedIndex")] !== undefined ? parseLine(lines[this.getActualProperty("selectedIndex")]) : "";

  if (!this.titleText || !(0, _text.areTextsEqual)(this.titleText, firstRowText, textStyle)) {
    this.titleText = (0, _text.newSingleLineText)(firstRowText, textStyle, this._context);
  }

  if (lines !== undefined) {
    this._lines = lines;

    this._lines.shift();
  }

  _icon.default.processIconData(this);
};

control.CONTROL_PROTOTYPE.menuLines = function () {
  const selectedMenuIndex = this.getActualProperty("selectedIndex") - 1;
  const controlState = this.getActualProperty("state");
  const isOpen = !_types.default.ControlState.isClosedState(controlState);
  const isSameLengthOfPreviousState = this._menuLines.length === this._lines.length;

  if (this._lines.length > 0 && isOpen) {
    const textColor = _types.default.ControlState.isDisabledState(controlState) ? this.style("default_disabled_color") : this.getActualProperty("textColor");
    return this._lines.map((line, index) => {
      const text = line.replace(new RegExp(",", "g"), "\\,");
      const selected = selectedMenuIndex === index;
      const previousLine = isSameLengthOfPreviousState ? this._menuLines[index] : null;
      return Menu.makeMenuLine(this, text, selected, textColor, false, previousLine);
    }, this);
  }

  return [];
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  const labelSize = this.titleText.measuredSize();
  const iconSize = this.iconData ? this.iconData.size : 0;
  let width = p_sizeProperty.width === _types.default.AUTOSIZE ? this.style("combobox_padding_left") + labelSize.width + this.style("combobox_padding_right") + this.style("combobox_arrow_width") + iconSize : p_sizeProperty.width;
  let height = this.style("combobox_padding_top") + (labelSize.height === 0 ? this.getActualProperty("fontSize") : labelSize.height) + this.style("combobox_padding_bottom");
  height = Math.max(height, iconSize);
  this.comboHeight = height;
  this._menuLines = this.menuLines();

  if (!_types.default.ControlState.isClosedState(this.getActualProperty("state")) && this._menuLines.length > 0) {
    const size = _balsamiq.default.clone(p_sizeProperty);

    Menu.menuInit(3, 3, 3);
    height = this._menuLines.reduce((totalHeight, line) => totalHeight + Menu.measureHeight(line, size.width), height);
    height += this.style("combobox_padding_bottom");
  }

  return _types.default.newSize(width, height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this);

  var color = _types.default.ControlState.isDisabledState(this.getActualProperty("state")) ? this.style("default_disabled_color") : this.getActualProperty("textColor");

  var items = _combobox.default.makeComboboxDisplayItems(this, p_localBbox, this.comboHeight, this.titleText, this.getActualProperty("borderColor"), this.getActualProperty("state"), this.iconData, color);

  var paddingTop = this.style("combobox_padding_top");
  this._menuLines = this.menuLines();

  if (this._menuLines.length > 0) {
    var borderBox = _balsamiq.default.clone(p_localBbox);

    var background = Rectangle.makeFilledRectangle(borderBox, 0xFFFFFF);
    this.pushToDisplayList(background);
    borderBox.y = Math.max(this.comboHeight - 1, 1);
    borderBox.height = Math.max(borderBox.height - this.comboHeight - 1, 1);
    var border = Rectangle.makeStrokedRectangle(borderBox, _types.default.ControlState.isDisabledState(this.getActualProperty("state")) ? this.style("combobox_border_disabled_color") : this.getActualProperty("borderColor"), this.style("combobox_border_thickness")); //eslint-disable-next-line no-control-regex

    var regExp = new RegExp("(^|\n)(o|x) ");

    if (!regExp.exec(this.getActualProperty("text"))) {
      var menulineX = this.style("combobox_padding_left");
    }

    var display = Menu.makeMenuDisplayItems(this, borderBox, this._menuLines, this.comboHeight + paddingTop, this.getActualProperty("selectedIndex") - 1, menulineX, color);

    if (this.getActualProperty('verticalScrollbar')) {
      var scrollbar = ScrollBar.makeVerticalScrollbar(p_localBbox, this.comboHeight, 2, 0, this.getActualProperty('value'), _types.default.ControlState.isDisabledState(this.getActualProperty("state")) ? _types.default.ControlState.DISABLED : _types.default.ControlState.NORMAL, this._context, 20);
    }
  }

  this.pushToDisplayList(items, display, scrollbar, border);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/controlid.js":
/*!***************************************!*\
  !*** ./js/core/controls/controlid.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.areEqual = areEqual;
exports.areInSameGroup = areInSameGroup;
exports.deserialize = deserialize;
exports.getParent = getParent;
exports.getRelative = getRelative;
exports.isDescendant = isDescendant;
exports.isParentOfChild = isParentOfChild;
exports.serialize = serialize;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function areEqual(a, b) {
  if (!a || !b) {
    return false;
  }

  if (a.length !== b.length) {
    return false;
  }

  for (let i = 0; i < a.length; i += 1) {
    if (a[i] !== b[i]) {
      return false;
    }
  }

  return true;
}

function isDescendant(nephew, parent) {
  _balsamiq.default.assertArray(nephew);

  _balsamiq.default.assertArray(parent);

  if (parent.length >= nephew.length) {
    return false;
  }

  for (let i = 0; i < parent.length; i += 1) {
    if (parent[i] !== nephew[i]) {
      return false;
    }
  }

  return true;
}

function isParentOfChild(parent, child) {
  _balsamiq.default.assertArray(parent);

  _balsamiq.default.assertArray(child);

  return isDescendant(child, parent) && child.length === parent.length + 1;
}

function getParent(controlID) {
  _balsamiq.default.assertArray(controlID);

  if (controlID.length) {
    return controlID.slice(0, -1);
  }

  return [];
}

function getRelative(parentControlID, childControlID) {
  _balsamiq.default.assertArray(parentControlID);

  _balsamiq.default.assertArray(childControlID);

  if (parentControlID.length == childControlID.length) {
    return [-1];
  } else {
    _balsamiq.default.assert(parentControlID.length < childControlID.length);

    return childControlID.slice(parentControlID.length);
  }
}

function areInSameGroup(controlID1, controlID2) {
  return areEqual(getParent(controlID1), getParent(controlID2));
}

function serialize(controlID) {
  _balsamiq.default.assertArray(controlID);

  var res;

  try {
    res = JSON.stringify(controlID);
  } catch (e) {
    res = null;
  }

  return res;
}

function deserialize(controlIDStr) {
  _balsamiq.default.assertString(controlIDStr);

  var res;

  try {
    res = JSON.parse(controlIDStr);
  } catch (e) {
    res = null;
  }

  return res;
}

/***/ }),

/***/ "./js/core/controls/controls.js":
/*!**************************************!*\
  !*** ./js/core/controls/controls.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.addTextClipBBoxCalculation = exports.addStandardTextPropertyDefinitions = exports.addStandardCalculateOriginForNewTextOrSize = exports.addSelectionPropertyDefinitions = exports.addPropertyDefinitions = exports.addColorPropertyDefinitions = exports.Control = void 0;
exports.assertControlsInSameGroup = assertControlsInSameGroup;
exports.classRegistry = void 0;
exports.getControlInitialProperties = getControlInitialProperties;
exports.getControlNames = getControlNames;
exports.isControlCropped = isControlCropped;
exports.isInLockedAncestor = isInLockedAncestor;
exports.lastControlID = lastControlID;
exports.newControl = exports.makeControlClass = exports.makeAssetBasedControlClass = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _displaynode = __webpack_require__(/*! core/displaynode */ "./js/core/displaynode.js");

var CoreObject = _interopRequireWildcard(__webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js"));

var Consts = _interopRequireWildcard(__webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts"));

var _macro = __webpack_require__(/*! core/bricks/macro */ "./js/core/bricks/macro.js");

var textMod = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var Snapping = _interopRequireWildcard(__webpack_require__(/*! core/controls/snapping */ "./js/core/controls/snapping.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var MockupsJSON = _interopRequireWildcard(__webpack_require__(/*! core/serialization/json */ "./js/core/serialization/json.js"));

var UI = _interopRequireWildcard(__webpack_require__(/*! editor/ui_constants */ "./js/editor/ui_constants.js"));

var ControlIDs = _interopRequireWildcard(__webpack_require__(/*! core/controls/controlid */ "./js/core/controls/controlid.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _linkareas = __webpack_require__(/*! core/graphics/linkareas */ "./js/core/graphics/linkareas.js");

var MockupCoords = _interopRequireWildcard(__webpack_require__(/*! core/viewcontrollers/mockupcoordinates */ "./js/core/viewcontrollers/mockupcoordinates.js"));

var _semver = __webpack_require__(/*! core/utils/semver */ "./js/core/utils/semver.ts");

var _wireframeVersion = __webpack_require__(/*! core/wireframe-version */ "./js/core/wireframe-version.ts");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const SurrogateCtor = function () {};

const extend = function (base, sub) {
  SurrogateCtor.prototype = base.prototype;
  sub.prototype = new SurrogateCtor();
  sub.prototype.constructor = sub;
  sub.prototype.superClass = base;
  base.constructor = base; // To allow for "this.superClass.constructor"
};

var classRegistry = {};
exports.classRegistry = classRegistry;

var extendControlClass = function (base, sub, controlTypeID, controlTypeName, controlAlternateNames, controlInitialProperties, subNewPropertyDefinitions) {
  extend(base, sub);
  classRegistry[controlTypeID] = sub;
  sub.prototype.controlTypeName = controlTypeName;
  sub.prototype.controlAlternateNames = controlAlternateNames;
  sub.prototype.controlInitialProperties = controlInitialProperties; // duplicate the property definitions of base and merge the new ones of sub

  sub.prototype.propertyDefinitions = _balsamiq.default.clone(base.prototype.propertyDefinitions);
  let newPropertyNames = subNewPropertyDefinitions ? Object.keys(subNewPropertyDefinitions) : [];
  newPropertyNames.forEach(function (p_newPropertyName) {
    sub.prototype.propertyDefinitions[p_newPropertyName] = subNewPropertyDefinitions[p_newPropertyName];
  });
};

var Control; // forward declaration

exports.Control = Control;

const makeControlClass = function (obj) {
  _balsamiq.default.exists(obj);

  _balsamiq.default.assertString(obj.controlTypeID);

  _balsamiq.default.assertString(obj.controlTypeName);

  const control = {
    CONTROL_TYPE_ID: obj.controlTypeID,
    CONTROL_TYPE_NAME: obj.controlTypeName,
    CONTROL_ALTERNATE_NAMES: obj.controlAlternateNames,
    CONTROL_INITIAL_PROPERTIES: obj.controlInitialProperties
  };

  const controlClass = function (p_context, p_coreObjectID, p_json, allowNegativeCoords) {
    this.superClass.constructor.call(this, p_context, p_coreObjectID, control.CONTROL_TYPE_ID, p_json, allowNegativeCoords);
  };

  extendControlClass(Control, controlClass, control.CONTROL_TYPE_ID, control.CONTROL_TYPE_NAME, control.CONTROL_ALTERNATE_NAMES, control.CONTROL_INITIAL_PROPERTIES, obj.props);
  control.CONTROL_PROTOTYPE = controlClass.prototype;
  return control;
}; // {
//     controlTypeID,
//     controlTypeName,
//     controlAlternateNames,
//     controlInitialProperties,
//     defaultSize
//     resizeOption,
//     skinAssetName,
//     colorStyleName,
//     otherPropertyDefinitions
// }


exports.makeControlClass = makeControlClass;

const makeAssetBasedControlClass = function (params) {
  const propertyDefinitions = _balsamiq.default.extend({ ...Props.PropertyDefinitions.SIZE({
      defaultValue: _balsamiq.default.clone(params.defaultSize),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: params.resizeOption
      }
    })
  }, params.otherPropertyDefinitions);

  const control = makeControlClass({
    controlTypeID: params.controlTypeID,
    controlTypeName: params.controlTypeName,
    controlAlternateNames: params.controlAlternateNames,
    controlInitialProperties: params.controlInitialProperties,
    props: propertyDefinitions
  });

  if (params.skinAssetName) {
    control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
      this.pushToDisplayList(_displayitem.default.createDisplayItemsFromSkin(params.skinAssetName, params.colorStyleName ? this.style(params.colorStyleName) : null, p_localBbox, this._context));
    };
  }

  return control;
};

exports.makeAssetBasedControlClass = makeAssetBasedControlClass;

var addTextClipBBoxCalculation = function (p_control) {
  p_control.CONTROL_PROTOTYPE.updateClipBBoxPS = function () {
    var expandPx = 0;
    var sizeProp = this.getActualProperty("size");

    if (sizeProp.width === Props.AUTOSIZE) {
      if (this.getActualProperty("italic")) {
        let fontSize = this.getActualProperty("fontSize");
        expandPx = fontSize / 6;
      }
    }

    this.setClipBBoxPS(this._userBBox.x, this._userBBox.y, this._userBBox.width + expandPx, this._userBBox.height);
  };
};

exports.addTextClipBBoxCalculation = addTextClipBBoxCalculation;

var addPropertyDefinitions = function (p_control, p_props) {
  p_control.CONTROL_PROTOTYPE.propertyDefinitions = _balsamiq.default.extend(p_props, p_control.CONTROL_PROTOTYPE.propertyDefinitions);
};

exports.addPropertyDefinitions = addPropertyDefinitions;

var addStandardTextPropertyDefinitions = function (p_control, p_defaultValues) {
  if (!p_defaultValues) {
    p_defaultValues = {};
  }

  var propertyDefinitions = { ...Props.PropertyDefinitions.TEXT({
      defaultText: p_defaultValues.text,
      multiline: p_defaultValues.multiline,
      loremable: p_defaultValues.loremable,
      charCountEnabled: p_defaultValues.charCountEnabled,
      splitOn: p_defaultValues.splitOn,
      noEmptyText: p_defaultValues.noEmptyText
    }),
    fontSize: {
      defaultValue: _balsamiq.default.has(p_defaultValues, "fontSize") ? p_defaultValues.fontSize : function () {
        return this._context.defaultFontSize;
      },
      dataType: "int",
      label: "Font Size"
    },
    bold: {
      defaultValue: _balsamiq.default.has(p_defaultValues, "bold") ? p_defaultValues.bold : false,
      dataType: "bool",
      label: "Bold"
    },
    italic: {
      defaultValue: _balsamiq.default.has(p_defaultValues, "italic") ? p_defaultValues.italic : false,
      dataType: "bool",
      label: "Italic"
    },
    underline: {
      defaultValue: _balsamiq.default.has(p_defaultValues, "underline") ? p_defaultValues.underline : false,
      dataType: "bool",
      label: "Underline"
    }
  };
  addPropertyDefinitions(p_control, propertyDefinitions);
};

exports.addStandardTextPropertyDefinitions = addStandardTextPropertyDefinitions;

var addStandardCalculateOriginForNewTextOrSize = function (p_control) {
  p_control.CONTROL_PROTOTYPE.calculateOriginForNewTextOrSize = function (p_text, p_size, fontSize) {
    // Respect right or center alignment when replacing text
    // Origin will move accordingly to respect the alignment
    _balsamiq.default.assertString(p_text);

    _balsamiq.default.assertSize(p_size);

    _balsamiq.default.assertOptionalNumber(fontSize);

    var newOrigin;
    var bbox = this.getUserBBox(true);
    var orientation;

    if (this.hasPropertyDefinition("textOrientation")) {
      orientation = this.getActualProperty("textOrientation");
    }

    if (!orientation || orientation === _types.default.TextOrientation.RIGHT) {
      let textStr = (0, _macro.processMacrosInText)(this, p_text);
      let textStyle = this.newStandardTextStyle({
        fontSize: fontSize || this.getActualProperty("fontSize")
      });
      let text = textMod.newTextNoWrap(textStr, textStyle, this._context);
      var textSize = text.measuredSize();

      if (p_size.width === _types.default.AUTOSIZE) {
        if (this.getActualProperty("align") == _types.default.TextAlign.RIGHT) {
          let right = _types.default.rectangleRight(bbox);

          newOrigin = _types.default.newPoint(right - textSize.width, bbox.y);
        } else if (this.getActualProperty("align") == _types.default.TextAlign.CENTER) {
          let center = bbox.x + Math.floor(bbox.width / 2);
          newOrigin = _types.default.newPoint(center - Math.floor(textSize.width / 2), bbox.y);
        } else {
          newOrigin = _types.default.newPoint(bbox.x, bbox.y);
        }
      } else {
        newOrigin = _types.default.newPoint(bbox.x, bbox.y);
      }
    } else {
      newOrigin = _types.default.newPoint(bbox.x, bbox.y);
    }

    const groupToMockupSpace = MockupCoords.groupToMockupSpace(newOrigin, this.getParent());

    if (groupToMockupSpace.x < 0) {
      newOrigin.x += -groupToMockupSpace.x;
    }

    return newOrigin;
  };
};

exports.addStandardCalculateOriginForNewTextOrSize = addStandardCalculateOriginForNewTextOrSize;

var addSelectionPropertyDefinitions = function (p_control, p_defaultValue) {
  let p_options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;
  let values = {
    selectedIndex: {
      defaultValue: p_defaultValue,
      dataType: "int",
      label: "Selection"
    }
  };

  if (p_options) {
    values.selectedIndex.options = p_options;
  }

  addPropertyDefinitions(p_control, values);

  p_control.CONTROL_PROTOTYPE.calculateSelectionLabelsForEditText = function (text) {
    if (this.getLinkableLabels) {
      return this.getLinkableLabels(text);
    } else {
      let linkLabels = text.split("\n").map(label => {
        return TextUtils.unformatText(label);
      });
      return linkLabels;
    }
  };

  p_control.CONTROL_PROTOTYPE.getDefaultSelectionValue = function () {
    return p_options && p_options.noMinusOne == true ? 0 : -1;
  };

  p_control.CONTROL_PROTOTYPE.updateSelectionOnTextChange = function (control, newText, oldText, oldSelection) {
    let newTextLines = control.calculateSelectionLabelsForEditText(newText);

    if (oldSelection === -1) {
      return -1;
    }

    let oldSelectedText = oldText[oldSelection];

    if (newTextLines.indexOf(oldSelectedText) > -1) {
      //the exact same text has been moved
      return newTextLines.indexOf(oldSelectedText);
    } else {
      if (newTextLines.length == oldText.length) {
        // the text has been changed but the control still has the same number of lines so we keep the same selection
        return oldSelection;
      } else {
        return control.getDefaultSelectionValue();
      }
    }
  };
};

exports.addSelectionPropertyDefinitions = addSelectionPropertyDefinitions;

var addColorPropertyDefinitions = function (p_control, obj) {
  /*p_defaultColor, p_defaultBorderColor, p_hasAlpha, p_defaultTextColor) {*/
  _balsamiq.default.exists(obj); ////you must at least pass a defaultColor or a defaultTextColor
  //Balsamiq.assert(Balsamiq.isString(p_defaultColor) || Balsamiq.isNumber(p_defaultColor) || Balsamiq.isString(p_defaultTextColor) || Balsamiq.isNumber(p_defaultTextColor));


  var mainColorName = obj.mainColorName;

  if (!mainColorName && _balsamiq.default.has(obj, "defaultColor")) {
    mainColorName = "color";
  } else if (!mainColorName && _balsamiq.default.has(obj, "defaultBorderColor")) {
    mainColorName = "borderColor";
  } else if (!mainColorName && _balsamiq.default.has(obj, "defaultTextColor")) {
    mainColorName = "textColor";
  }

  _balsamiq.default.assertString(mainColorName, "Can't assign automatically the main color property of control type id: " + p_control.CONTROL_TYPE_ID);

  var defs = {};

  if (_balsamiq.default.isString(obj.defaultColor) || _balsamiq.default.isNumber(obj.defaultColor)) {
    defs.color = {
      defaultValue: _balsamiq.default.isNumber(obj.defaultColor) ? obj.defaultColor : function () {
        return this.style(obj.defaultColor);
      },
      dataType: "int",
      label: "Color",
      options: {}
    };

    if (_balsamiq.default.isString(obj.oldColorName)) {
      defs.color.options.oldName = obj.oldColorName;
    }

    if (mainColorName == "color") {
      defs.color.options.isMain = true;
    }
  }

  if (_balsamiq.default.isString(obj.defaultBorderColor) || _balsamiq.default.isNumber(obj.defaultBorderColor)) {
    defs.borderColor = {
      defaultValue: _balsamiq.default.isNumber(obj.defaultBorderColor) ? obj.defaultBorderColor : function () {
        return this.style(obj.defaultBorderColor);
      },
      dataType: "int",
      label: "Border Color",
      options: {}
    };

    if (_balsamiq.default.isString(obj.oldBorderColorName)) {
      defs.borderColor.options.oldName = obj.oldBorderColorName;
    }

    if (mainColorName == "borderColor") {
      defs.borderColor.options.isMain = true;
    }
  }

  if (_balsamiq.default.isString(obj.defaultTextColor) || _balsamiq.default.isNumber(obj.defaultTextColor)) {
    defs.textColor = {
      defaultValue: _balsamiq.default.isNumber(obj.defaultTextColor) ? obj.defaultTextColor : function () {
        return this.style(obj.defaultTextColor);
      },
      dataType: "int",
      label: "Text Color",
      options: {}
    };

    if (_balsamiq.default.isString(obj.oldTextColorName)) {
      defs.textColor.options.oldName = obj.oldTextColorName;
    }

    if (mainColorName == "textColor") {
      defs.textColor.options.isMain = true;
    }
  }

  if (_balsamiq.default.isBoolean(obj.hasAlpha)) {
    defs = _balsamiq.default.extend(defs, Props.PropertyDefinitions.ALPHA());
  }

  addPropertyDefinitions(p_control, defs);
  p_control.CONTROL_PROTOTYPE.hasColorProperties = true;

  p_control.CONTROL_PROTOTYPE.getMainColorPropertyName = function () {
    if (this.hasPropertyDefinition("color") && this.getPropertyOptions("color").isMain) {
      return "color";
    } else if (this.hasPropertyDefinition("borderColor") && this.getPropertyOptions("borderColor").isMain) {
      return "borderColor";
    } else if (this.hasPropertyDefinition("textColor") && this.getPropertyOptions("textColor").isMain) {
      return "textColor";
    }

    _balsamiq.default.assert(false, "Can't get the main color property name of control type id: " + this.controlTypeID);
  };
}; // ---------------------------------------------------------

/**
 *
 * @param {Object} p_context
 * @param {Object} p_coreObjectID
 * @param {string} p_controlTypeID
 * @param {Object=} p_json
 * @param {boolean=} allowNegativeCoords
 * @returns {ControlClass}
 */


exports.addColorPropertyDefinitions = addColorPropertyDefinitions;

var newControl = function (p_context, p_coreObjectID, p_controlTypeID, p_json, allowNegativeCoords) {
  _balsamiq.default.exists(p_context);

  _balsamiq.default.exists(p_coreObjectID);

  _balsamiq.default.assertString(p_controlTypeID);

  _balsamiq.default.assertOptionalObject(p_json);

  var ControlClass = classRegistry[p_controlTypeID];
  return ControlClass ? new ControlClass(p_context, p_coreObjectID, p_json, allowNegativeCoords) : null;
}; // ---------------------------------------------------------

/**
 * @constructor
 * @param {any} p_context
 * @param {CoreObject.CoreObjectID} p_coreObjectID
 * @param {string} p_controlTypeID
 * @param {object?} p_json
 * @param {boolean} allowNegativeCoords
 */


exports.newControl = newControl;

exports.Control = Control = function (p_context, p_coreObjectID, p_controlTypeID, p_json, allowNegativeCoords) {
  _balsamiq.default.exists(p_context.textMeter);

  _balsamiq.default.exists(p_context.eventCenter);

  _balsamiq.default.assertFunction(p_context.setDependencies);

  _balsamiq.default.assertFunction(p_context.getDependency);

  _balsamiq.default.assertFunction(p_context.setReady);

  _balsamiq.default.assertFunction(p_context.notifyChange);

  _balsamiq.default.exists(p_context.skin);

  _balsamiq.default.assertNumber(p_context.defaultFontSize);

  _balsamiq.default.assertString(p_context.fontFamily);

  _balsamiq.default.assertNumber(p_context.selectionColor);

  _balsamiq.default.assertNumber(p_context.linkColor);

  _balsamiq.default.assertOptionalObject(p_json);

  this._context = p_context;
  this.children = []; // sorted by z-order

  this.parent = null; // only addChild and removeChild update this

  this._nextChildControlID = 0;
  this.coreObjectID = p_coreObjectID;
  this.controlID = p_coreObjectID.controlID && p_coreObjectID.controlID.length !== 0 ? lastControlID(p_coreObjectID) : -1;
  this.controlTypeID = p_controlTypeID;
  this.properties = {};
  this.overriddenProperties = null; // it exists when the control belongs to a symbol instance

  this.loadedProperties = null; // it exists when the control is loaded from JSON

  this.hasLoadedUnknownProperties = false; // true when loadedProperties contains some unknown properties (or prop. values)

  this._locked = false;
  this._displayList = {
    items: [],
    skinName: this._context.skin["#skinName"]
  };
  this._masks = [];
  this._colorTransforms = [];
  this._clipBBox = null;
  this._userBBox = null;
  this._clipBBoxNoMarkup = null;
  this._userBBoxNoMarkup = null;
  this._clipOnBbox = true;
  this._macros = [];
  this._linkAreas = [];
  this._cropData = null;
  this._dependencies = [];
  this.initData();

  if (p_json) {
    this.loadProperties(p_json, allowNegativeCoords);
  } else {
    this.updateAll();
  }
};
/** @type {CoreObject.CoreObjectID} */


Control.prototype.coreObjectID;
/** @type {number} */

Control.prototype.controlID;
/** @type {string} */

Control.prototype.controlTypeID;
/** @type { { [prop: string]: any }= } */

Control.prototype.properties;
/** @type { Control[]= } */

Control.prototype.children;
/** @type { Control= } */

Control.prototype.parent;

Control.prototype.copyWithChildren = function (p_context, coreObjectID) {
  let controlCopy = this.copy(p_context, coreObjectID);
  let children = [];

  for (let child of this.children) {
    let childCoreObjectID = controlCopy.buildChildCoreObjectID(child.controlID);
    children.push(child.copyWithChildren(p_context, childCoreObjectID));
  }

  controlCopy.addChildren(children);
  controlCopy.updateAll();
  return controlCopy;
};

Control.prototype.copy = function (p_context, p_coreObjectID) {
  var c = newControl(p_context, p_coreObjectID, this.controlTypeID);
  c.properties = _types.default.deepCopy(this.properties);
  c.loadedProperties = _types.default.deepCopy(this.loadedProperties);

  if (this.hasCropData()) {
    c._cropData = _types.default.deepCopy(this._cropData);
  }

  return c;
};

Control.prototype.initData = function () {// called by the constructor
  // subclasses can implement this
}; // TODO rename this to 'reset'


Control.prototype.cleanUp = function () {
  // called when the control is about to be deleted
  this.setDependencies([]);
  this.removeAllChildren();
};

Control.prototype.loadChildren = function (json) {
  let {
    generateNewIDs = false,
    allowNegativeCoords = false
  } = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  _balsamiq.default.exists(json);

  _balsamiq.default.assertBool(generateNewIDs);

  _balsamiq.default.assertBool(allowNegativeCoords);

  if (!_balsamiq.default.has(json, "control")) {
    return; //Empty mockup, it happens
  }

  _balsamiq.default.assertArray(json.control, "Error loading controls: 'control' is not an array");

  var errors = [];

  var jsonControls = _balsamiq.default.sortBy(json.control, function (control) {
    return parseInt(control.zOrder, 10);
  }); // find the last controlID so that, if there are 2 controls with the same controlID, we can use it to fix the broken wireframe


  let lastLoadedControlID = jsonControls.reduce((nextLoadedControlID, jsonControl) => {
    const controlID = parseInt(jsonControl.ID, 10);
    return controlID > nextLoadedControlID ? controlID : nextLoadedControlID;
  }, 0);
  const modelControls = jsonControls.reduce((_ref, jsonControl) => {
    let {
      controlIDs,
      controls
    } = _ref;
    var modelControl;

    try {
      // control ID
      var controlID;

      if (generateNewIDs) {
        controlID = this.getNextControlID();
      } else {
        const tempControlID = parseInt(jsonControl.ID, 10);

        if (controlIDs[tempControlID]) {
          controlID = ++lastLoadedControlID;
        } else {
          controlID = tempControlID;
        }
      }

      controlIDs[controlID] = true; // control instance

      modelControl = newControl(this._context, this.buildChildCoreObjectID(controlID), jsonControl.typeID === "__group__" ? jsonControl.typeID : "com.balsamiq.mockups::" + jsonControl.typeID, // TODO: this is a hack!!
      jsonControl, allowNegativeCoords);

      if (!modelControl) {
        this._context.logger(`Error creating a control: ${jsonControl.typeID}: ${controlID}`);

        errors.push("Error creating a control");
        return null;
      } // locked


      if (_balsamiq.default.isString(jsonControl.locked)) {
        modelControl.setLocked(jsonControl.locked == "true");
      } // children


      if (jsonControl.children) {
        modelControl.loadChildren(jsonControl.children.controls);
      }

      controls.push(modelControl);
    } catch (e) {
      const error = "Error loading control: " + JSON.stringify(jsonControl) + "\n" + e.message + "\n" + e.stack;

      this._context.logger(error);

      errors.push(error);
    }

    return {
      controlIDs,
      controls
    };
  }, {
    controlIDs: {},
    controls: []
  });

  try {
    this.addChildren(modelControls.controls);
  } catch (e) {
    this._context.logger(e.message + "\n" + e.stack);

    errors.push(e.message + e.stack);
  }

  return {
    errors: errors,
    controls: modelControls.controls
  };
};

Control.prototype.loadProperties = function (jsonControl, allowNegativeCoords) {
  // this._context.logger(`Loaded properties: ${JSON.stringify(this.loadedProperties)}`);
  this.hasLoadedUnknownProperties = MockupsJSON.loadProperties(this._context, this, jsonControl, allowNegativeCoords).hasUnknownProperties;
  this.loadedProperties = _balsamiq.default.get(jsonControl, "properties");
};

Control.prototype.getContext = function () {
  return this._context;
};

Control.prototype.reload = function (jsonControl) {
  let {
    allowNegativeCoords = false,
    checkSameID = true
  } = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  if (checkSameID) _balsamiq.default.assert(this.controlID.toString() === jsonControl.ID, "The new and the old control must have the same ID");

  _balsamiq.default.assert(this.controlTypeID === jsonControl.typeID || this.controlTypeID === `com.balsamiq.mockups::${jsonControl.typeID}`, "The new and the old control must have the same typeID");

  this.setLocked(jsonControl.locked === "true");
  this.hasLoadedUnknownProperties = MockupsJSON.loadProperties(this._context, this, jsonControl, allowNegativeCoords).hasUnknownProperties;
  this.loadedProperties = _balsamiq.default.get(jsonControl, "properties");
  if (this.controlTypeID === "Component" || this.controlTypeID === "com.balsamiq.mockups::Component") return;
  if (this.children && this.children.length > 0) this.removeAllChildren();

  const children = _balsamiq.default.get(jsonControl, "children.controls");

  if (children) {
    // When (re)loading children the crop offset must be reset since it must be recalculated from scratch starting from fresh data.
    // Calculations must not rely on previous values.
    this.resetCropData();
    this.loadChildren(jsonControl.children.controls);
  }
};

Control.prototype.saveProperties = function () {
  let jsonProps = {};
  let allPropNames = {};

  for (let propName in this.loadedProperties) {
    allPropNames[propName] = true;
  }

  for (let propName in this.properties) {
    allPropNames[propName] = true;
  }

  for (let propName in allPropNames) {
    if (_balsamiq.default.has(this.loadedProperties, propName)) {
      // save from loaded properties
      jsonProps[propName] = this.loadedProperties[propName];
    } else {
      // save from Control properties
      if (!_balsamiq.default.has(this.properties, propName)) {
        continue;
      }

      if (propName == "origin" || propName == "size") {
        continue;
      }

      let value = this.getProperty(propName, false);

      if (this.getPropertyOptions(propName).oldName) {
        propName = this.getPropertyOptions(propName).oldName;
      }

      MockupsJSON.saveProperty(jsonProps, propName, value);
    }
  } // this._context.logger(`Saved properties: ${JSON.stringify(jsonProps)}`);


  return jsonProps;
};

Control.prototype._loadOverriddenProperties = function (jsonProps, inOverridden) {
  var names = [],
      values = [];

  for (var key in jsonProps) {
    if (key == "x") {
      names.push("origin");
      values.push(_types.default.newPoint(parseInt(jsonProps.x, 10), parseInt(jsonProps.y, 10)));
    } else if (key == "w") {
      names.push("size");
      values.push(_types.default.newSize(parseInt(jsonProps.w, 10), parseInt(jsonProps.h, 10)));
    } else if (["x", "y", "w", "h"].indexOf(key) == -1) {
      var j = MockupsJSON.readProperty(this._context, this, key, jsonProps[key]);

      if (j) {
        // If key is not supported (may happen) we have a warning log from readProperty
        names.push(j.name);
        values.push(j.value);
      }
    }
  }

  for (let index = 0; index < names.length; index++) {
    const name = names[index];
    const value = values[index];

    this._setProperty(name, value, inOverridden, false);
  }

  this.updateAll();
};
/**
 * Get controls' children. It works for groups and symbols
 * 
 * @returns {Control[]}
 */


Control.prototype.getDescendants = function () {
  if (!this.children || this.children.length === 0) {
    return [];
  }

  const descendant = [];
  let guard = 1024;
  let visitQueue = [].concat(this.children);

  while (visitQueue.length > 0 && guard-- > 0) {
    let node = visitQueue.shift();
    descendant.push(node);

    if (node.children && node.children.length > 0) {
      visitQueue = visitQueue.concat(node.children);
    }
  }

  _balsamiq.default.assert(guard > 0, "Descendant limit reached. Maybe a cycle in symbol tree?");

  return descendant;
};
/**
 * Load overriddenProperties as normal properties. This is used when instantiating symbols inside other symbols so they don't result overridden
 * 
 * @param {Object} jsonProps
 */


Control.prototype.loadOverriddenPropertiesAsProperties = function (jsonProps) {
  _balsamiq.default.exists(jsonProps);

  this._loadOverriddenProperties(jsonProps, false);
};

Control.prototype.loadOverriddenProperties = function (jsonProps) {
  _balsamiq.default.exists(jsonProps);

  _balsamiq.default.exists(this.overriddenProperties, "can not load overridden properties in a Control without overridden properties");

  this._loadOverriddenProperties(jsonProps, true);
};

Control.prototype.saveOverriddenProperties = function () {
  //Balsamiq.exists(this.overriddenProperties, "can not save overridden properties of a Control without overridden properties");
  var jsonProps = {};

  if (this.overriddenProperties) {
    for (var propName in this.overriddenProperties) {
      MockupsJSON.saveProperty(jsonProps, propName, this.overriddenProperties[propName]);
    }
  }

  return jsonProps;
};

Control.prototype.pushToDisplayList = function () {
  for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
    args[_key] = arguments[_key];
  }

  const flatArgs = args.flat(Infinity);
  const validArgs = flatArgs.filter(element => !(_balsamiq.default.isUndefined(element) || _balsamiq.default.isNull(element)));

  this._displayList.items.push.apply(this._displayList.items, validArgs);
};

Control.prototype.addMask = function () {
  this._masks.push([]);

  return this._masks.length - 1;
};

Control.prototype.addRectToMask = function (p_maskID, p_rect) {
  _balsamiq.default.assert(p_maskID >= 0 && p_maskID < this._masks.length);

  this._masks[p_maskID].push(_types.default.makeMaskRect(p_rect));
};

Control.prototype.addMaskWithRect = function (p_rect) {
  var maskID = this.addMask();
  this.addRectToMask(maskID, p_rect);
  return maskID;
};

Control.prototype.addMaskWithRectHole = function (p_localBbox, p_holeRect) {
  // TODO add 4 rectangles
  var maskID = this.addMask(); // left

  if (p_holeRect.x > p_localBbox.x) {
    this.addRectToMask(maskID, _types.default.newRectangle(p_localBbox.x, p_localBbox.y, p_holeRect.x - p_localBbox.x, p_localBbox.height));
  } // right


  if (p_holeRect.x + p_holeRect.width < p_localBbox.x + p_localBbox.width) {
    this.addRectToMask(maskID, _types.default.newRectangle(p_holeRect.x + p_holeRect.width, p_localBbox.y, p_localBbox.x + p_localBbox.width - (p_holeRect.x + p_holeRect.width), p_localBbox.height));
  } // top


  if (p_holeRect.y > p_localBbox.y) {
    this.addRectToMask(maskID, _types.default.newRectangle(p_holeRect.x, p_localBbox.y, p_holeRect.x + p_holeRect.width, p_holeRect.y - p_localBbox.y));
  } // bottom


  if (p_holeRect.y + p_holeRect.height < p_localBbox.y + p_localBbox.height) {
    this.addRectToMask(maskID, _types.default.newRectangle(p_holeRect.x, p_holeRect.y + p_holeRect.height, p_holeRect.x + p_holeRect.width, p_localBbox.y + p_localBbox.height - (p_holeRect.y + p_holeRect.height)));
  }

  return maskID;
};

Control.prototype.addPathToMask = function (p_maskID, p_path, p_x, p_y) {
  _balsamiq.default.assert(p_maskID >= 0 && p_maskID < this._masks.length);

  this._masks[p_maskID].push({
    type: "path",
    path: p_path.data,
    x: p_x,
    y: p_y
  });
};

Control.prototype.addColorTransform = function (p_color) {
  _balsamiq.default.assertNumber(p_color);

  this._colorTransforms.push({
    color: p_color
  });

  return this._colorTransforms.length - 1;
};
/**
 *
 * @param {Object} p_bounds
 * @param {Object} p_target
 * @param {Boolean?} p_continueOnNextLine
 */


Control.prototype.addLinkArea = function (p_bounds, p_target) {
  let p_continueOnNextLine = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
  let p_angle = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : 0;

  _balsamiq.default.assertRect(p_bounds);

  _balsamiq.default.assertBool(p_continueOnNextLine);

  _balsamiq.default.assertNumber(p_angle);

  _balsamiq.default.assertString(p_target);

  let isDisabled = false;

  if (this.hasPropertyDefinition("state")) {
    isDisabled = this.getActualProperty("state") === _types.default.ControlState.DISABLED || this.getActualProperty("state") === _types.default.ControlState.DISABLED_SELECTED || this.getActualProperty("state") === _types.default.ControlState.DISABLED_CLOSED || this.getActualProperty("state") === _types.default.ControlState.DISABLED_INDETERMINATE;
  }

  const cropOffset = this.getCropOffset();
  let rect = p_bounds; //Types.rectanglesIntersect(this.localBBox(), p_bounds); // clip the rect to stay inside the bbox

  rect.x = rect.x - cropOffset.x;
  rect.y = rect.y - cropOffset.y;

  this._linkAreas.push({
    bounds: rect,
    target: p_target,
    continueOnNextLine: p_continueOnNextLine,
    angle: p_angle,
    disabled: isDisabled
  });
};

Control.prototype.getLinkAreas = function () {
  return this._linkAreas;
};

Control.prototype.getUserBBox = function (withMarkup) {
  _balsamiq.default.assertBool(withMarkup);

  return withMarkup ? _balsamiq.default.clone(this._userBBox) : _balsamiq.default.clone(this._userBBoxNoMarkup);
};

Control.prototype.getClipBBox = function (withMarkup) {
  _balsamiq.default.assertBool(withMarkup);

  return withMarkup ? _balsamiq.default.clone(this._clipBBox) : _balsamiq.default.clone(this._clipBBoxNoMarkup);
};

Control.prototype.propertyDefinitions = {
  origin: {
    defaultValue: function () {
      return _types.default.newPoint(0, 0);
    },
    dataType: "object",
    label: "Position"
  },
  markup: {
    defaultValue: false,
    dataType: "bool",
    label: "Treat As Markup"
  },
  customID: {
    defaultValue: null,
    dataType: "string"
  },
  customData: {
    defaultValue: null,
    dataType: "string"
  }
};

Control.prototype.toObject = function () {
  return {
    coreObjectID: this.coreObjectID,
    controlTypeID: this.controlTypeID,
    controlIDs: this.getChildrenControlIDs()
  };
};

Control.prototype.getControlTypeName = function () {
  return this.controlTypeName;
};

Control.prototype.getPropertyDefinitions = function () {
  return this.propertyDefinitions;
};
/**
 * Return true if the given property name is supported by the control.
 * @param {string} p_name
 * @returns {Boolean}
 */


Control.prototype.hasPropertyDefinition = function (p_name) {
  return _balsamiq.default.has(this.propertyDefinitions, p_name);
};
/**
 *
 * @param {string} p_name
 * @param {any} p_value
 * @returns {boolean}
 */


Control.prototype.checkPropertyValue = function (p_name, p_value) {
  // check if the property name is known
  _balsamiq.default.assert(_balsamiq.default.has(this.propertyDefinitions, p_name), "this control does not support the property " + p_name);

  let dataType = this.getPropertyDataType(p_name);

  switch (dataType) {
    case "enum":
      {
        return _balsamiq.default.checkEnum(p_value, this.getPropertyEnum(p_name));
      }

    case "int":
      {
        try {
          return Number.isSafeInteger(p_value);
        } catch (e) {
          return false;
        }
      }
  } // TODO other dataTypes (when needed...)


  return true;
};

Control.prototype.hasUnknownProperties = function () {
  return this.hasLoadedUnknownProperties || this.children.find(child => {
    return child.hasUnknownProperties();
  }) !== undefined;
};

Control.prototype.getPropertyLabel = function (p_name) {
  // check if the property name is known
  _balsamiq.default.assert(this.hasPropertyDefinition(p_name), "this control does not support the property " + p_name);

  return this.getPropertyDefinitions()[p_name].label ? this.getPropertyDefinitions()[p_name].label : p_name;
};

Control.prototype.activateOverriddenProperties = function () {
  if (!this.overriddenProperties) {
    this.overriddenProperties = {};
  }

  this.children.forEach(child => {
    child.activateOverriddenProperties();
  });
};

Control.prototype.getControlProperties = function () {
  return this.properties;
};

Control.prototype.getAllActualControlProperties = function () {
  let nonDefaultActualProperties = this.getNonDefaultActualControlProperties();

  for (let propertyName in this.propertyDefinitions) {
    if (!nonDefaultActualProperties[propertyName]) {
      nonDefaultActualProperties[propertyName] = this.getDefaultProperty(propertyName);
    }
  }

  return nonDefaultActualProperties;
};

Control.prototype.getNonDefaultActualControlProperties = function () {
  var actualProperties = {};

  for (let key in this.overriddenProperties) {
    actualProperties[key] = this.overriddenProperties[key];
  }

  for (let key in this.properties) {
    if (_balsamiq.default.has(actualProperties, key)) {
      continue;
    }

    actualProperties[key] = this.properties[key];
  }

  return actualProperties;
};

Control.prototype.hasPropertyValue = function (p_name) {
  // check if the property name is known
  _balsamiq.default.assert(_balsamiq.default.has(this.propertyDefinitions, p_name), "this control does not support the property " + p_name); // 1. overridden


  if (this.overriddenProperties) {
    if (_balsamiq.default.has(this.overriddenProperties, p_name)) {
      return true;
    }
  } // 2. regular properties


  if (_balsamiq.default.has(this.properties, p_name)) {
    return true;
  }

  return false;
};

Control.prototype.getActualProperty = function (p_name) {
  // check if the property name is known
  _balsamiq.default.assert(_balsamiq.default.has(this.propertyDefinitions, p_name), "this control does not support the property " + p_name); // 1. overridden


  if (this.overriddenProperties) {
    if (_balsamiq.default.has(this.overriddenProperties, p_name)) {
      return this.overriddenProperties[p_name];
    }
  } // 2. regular properties


  if (_balsamiq.default.has(this.properties, p_name)) {
    return this.properties[p_name];
  } // 3. default


  return this.getDefaultProperty(p_name);
};

Control.prototype.getProperty = function (p_name, p_overridden) {
  // check if the property name is known
  _balsamiq.default.assert(this.hasPropertyDefinition(p_name), "this control does not support the property " + p_name);

  if (p_overridden) {
    _balsamiq.default.exists(this.overriddenProperties, "can't get an overridden property when overriddenProperties are off");
  }

  var props = p_overridden ? this.overriddenProperties : this.properties;
  return props[p_name];
};

Control.prototype.getDefaultProperty = function (p_name) {
  _balsamiq.default.assert(_balsamiq.default.has(this.propertyDefinitions, p_name), 'property "' + p_name + '" is not supported by this control ' + this.controlTypeID);

  var defaultValue = this.propertyDefinitions[p_name].defaultValue;

  if (_balsamiq.default.isFunction(defaultValue)) {
    defaultValue = defaultValue.call(this);
  }

  return defaultValue;
};

Control.prototype.getPropertyWithDefault = function (name, isOverridden) {
  return this.getProperty(name, isOverridden) || this.getDefaultProperty(name);
};
/**
 * Return the dataType of a property definition. Properties without an explicitly
 * set dataType are strings by default.
 * @param {string} p_name
 * @returns {string}
 */


Control.prototype.getPropertyDataType = function (p_name) {
  _balsamiq.default.assert(_balsamiq.default.has(this.propertyDefinitions, p_name), 'property "' + p_name + '" is not supported by this control ' + this.controlTypeID);

  var dataType = this.propertyDefinitions[p_name].dataType;

  if (!dataType) {
    if (this.propertyDefinitions[p_name].enum) {
      dataType = "enum";
    } else {
      dataType = "string";
    }
  }

  return dataType;
};

Control.prototype.getPropertySetter = function (p_name) {
  _balsamiq.default.assert(_balsamiq.default.has(this.propertyDefinitions, p_name), 'property "' + p_name + '" is not supported by this control ' + this.controlTypeID);

  return this.propertyDefinitions[p_name].setter;
};

Control.prototype.hasAutosize = function () {
  const def = this.propertyDefinitions;
  return def.size && def.size.options && def.size.options.autosize;
};

Control.prototype.getPropertyEnum = function (p_name) {
  _balsamiq.default.assert(_balsamiq.default.has(this.propertyDefinitions, p_name), 'property "' + p_name + '" is not supported by this control ' + this.controlTypeID);

  return this.propertyDefinitions[p_name].enum;
};

Control.prototype.getPropertyOptions = function (p_name) {
  _balsamiq.default.assert(_balsamiq.default.has(this.propertyDefinitions, p_name), 'property "' + p_name + '" is not supported by this control ' + this.controlTypeID);

  var opts = this.propertyDefinitions[p_name].options;
  return opts ? opts : {};
};

Control.prototype.setActualProperty = function (p_name, p_value) {
  this.setActualProperties([p_name], [p_value]);
};

Control.prototype.setActualProperties = function (p_names, p_values) {
  _balsamiq.default.assert(_balsamiq.default.isArray(p_names) && _balsamiq.default.isArray(p_values));

  _balsamiq.default.assert(p_names.length === p_values.length);

  for (let index in p_names) {
    this._setActualProperty(p_names[index], p_values[index]);
  }

  this.updateAll();
};

Control.prototype.hasOverriddenProperties = function () {
  return this.overriddenProperties != null;
};

Control.prototype.isPropertyOverridden = function (p_name) {
  _balsamiq.default.assert(_balsamiq.default.has(this.propertyDefinitions, p_name), 'property "' + p_name + '" is not supported by this control ' + this.controlTypeID);

  return this.overriddenProperties != null && this.overriddenProperties.hasOwnProperty(p_name);
};

Control.prototype.deleteFromLoadProperties = function (p_name) {
  if (this.loadedProperties) {
    const property = this.getPropertyOptions(p_name);
    let propNameToDelete = p_name;

    if (property && property.oldName) {
      propNameToDelete = property.oldName;
    } else if (p_name === "fontSize") {
      propNameToDelete = "size";
    } else if (p_name === "link") {
      propNameToDelete = "href";
    } else if (p_name === "links") {
      propNameToDelete = "hrefs";
    }

    if (p_name === "textColor") {
      delete this.loadedProperties[p_name];
    }

    delete this.loadedProperties[propNameToDelete];
  }
};

Control.prototype._setProperty = function (p_name, p_value, p_overridden, removeOverridden) {
  if (!_balsamiq.default.has(this.propertyDefinitions, p_name)) {
    //this._context.logger('Setting property "' + p_name + '", but is not supported by this control ' + this.controlTypeID);
    return;
  } // Check the enum value in override


  var dataType = this.getPropertyDataType(p_name);

  if (dataType == "enum" && p_value !== undefined && p_overridden) {
    let enumOk = _balsamiq.default.checkEnum(p_value, this.getPropertyEnum(p_name));

    if (!enumOk) {
      // Ignore invalid value in override!
      return;
    }
  } // Strict check if it is not an override


  if (dataType == "enum" && p_value !== undefined && !p_overridden) {
    // we must allow "undefined" to remove the override
    _balsamiq.default.assertEnum(p_value, this.getPropertyEnum(p_name));
  } // choose the right property container


  _balsamiq.default.assert(p_overridden ? this.overriddenProperties !== null : true, "Can't set an overridden property when overridden properties are null");

  var props = p_overridden ? this.overriddenProperties : this.properties;
  let setter = this.getPropertySetter(p_name);

  if (setter) {
    p_value = setter(p_value);
  } // special case for "size": check if autosize (-1) is allowed
  // if not, then use the default value


  if (p_name === "size" && p_value !== undefined) {
    if (p_value) {
      var w = Math.round(p_value.width);
      var h = Math.round(p_value.height);
      var autosizeOption = this.propertyDefinitions.size.options.autosize;

      if (w === Props.AUTOSIZE && autosizeOption !== Props.AutoSizeOptions.ONLY_WIDTH && autosizeOption !== Props.AutoSizeOptions.WIDTH_AND_HEIGHT) {
        w = this.getDefaultProperty("size").width;
      }

      if (h === Props.AUTOSIZE && autosizeOption !== Props.AutoSizeOptions.ONLY_HEIGHT && autosizeOption !== Props.AutoSizeOptions.WIDTH_AND_HEIGHT) {
        h = this.getDefaultProperty("size").height;
      }

      var resizeOption = this.propertyDefinitions.size.options.resize;

      if (resizeOption == Props.ResizeOptions.HORIZONTAL) {
        h = Props.AUTOSIZE;
      }

      if (resizeOption == Props.ResizeOptions.VERTICAL) {
        w = Props.AUTOSIZE;
      }

      p_value = _types.default.newSize(w, h);
    }
  } else if (p_name == "origin" && p_value !== undefined) {
    p_value = _types.default.newPoint(Math.round(p_value.x), Math.round(p_value.y)); // 'origin' property requires a notification

    this._context.notifyChange(this.coreObjectID, Consts.ControlChanges.POSITION);
  } // compare the value inherited form the symbol or from the control's default value


  if (removeOverridden && p_overridden && _balsamiq.default.isEqual(p_value, this.getPropertyWithDefault(p_name, false))) {
    p_value = undefined;
  }

  if (p_name !== "origin" && !p_overridden && _balsamiq.default.isEqual(p_value, this.getDefaultProperty(p_name))) {
    p_value = undefined;
  } // set the value


  if (p_value === undefined) {
    delete props[p_name];
  } else {
    props[p_name] = p_value;
  }

  this.deleteFromLoadProperties(p_name); // call custom onSet, if present

  if (_balsamiq.default.has(this.propertyDefinitions[p_name], "onSet")) {
    this.propertyDefinitions[p_name].onSet.call(this, p_value);
  }

  if (p_name == "controlName") {
    this._context.notifyChange(this.coreObjectID, Consts.ControlChanges.NAME);
  }
};

Control.prototype._setActualProperty = function (p_name, p_value) {
  this._setProperty(p_name, p_value, this.overriddenProperties != null);
};

Control.prototype.setProperty = function (p_name, p_value, p_overridden, removeOverridden) {
  this._setProperty(p_name, p_value, p_overridden, removeOverridden);

  if (p_name == "origin") {
    this.updateBBoxes();
  } else {
    this.updateAll();
  }
};
/**
 * Checks which other properties need to be changed due to a change in a specific one
 *
 * @param {string} name
 * @param {unknown} value
 * @returns {{[propName: string]: unknown}}
 */


Control.prototype.requiredChangesForProperty = function (name, value) {
  return {
    [name]: value
  };
};

Control.prototype.getTextOption = function (option) {
  var res = false;

  if (this.hasPropertyDefinition("text")) {
    var options = this.getPropertyDefinitions()["text"].options;

    if (options && options[option] !== undefined) {
      res = options[option];
    }
  }

  return res;
};

Control.prototype.hasMultilineText = function () {
  return this.getTextOption("multiline");
};

Control.prototype.hasLoremExpansion = function () {
  return this.getTextOption("loremable");
};

Control.prototype.charCountEnabled = function () {
  return this.getTextOption("charCountEnabled");
};

Control.prototype.getDisplayData = function () {
  let {
    selected,
    otherUsersSelectionStatus,
    zoom,
    withLockedBackground
  } = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  var selectionDisplayItems = [];

  if (selected) {
    selectionDisplayItems = this.getSelectionDisplayItems({
      otherUsersSelectionStatus,
      zoom,
      withLockedBackground
    });
  }

  return { //isGroup: this.controlTypeID === "__group__" || this.controlTypeID === "com.balsamiq.mockups::Component",
    ...this.getDisplayDataBBoxes(),
    controlID: this.controlID,
    displayListItems: this._displayList.items.concat(selectionDisplayItems),
    linksDisplayListItems: (0, _linkareas.makeLinkAreasDisplayList)(this._linkAreas, this._context.getDependency),
    displayListSkinName: this._displayList.skinName,
    masks: this._masks,
    colorTransforms: this._colorTransforms,
    clipOnBbox: this._clipOnBbox,
    linkAreas: this._linkAreas,
    isMarkup: this.isMarkup(),
    alpha: 1
  };
};
/**
 * @returns {DisplayNode}
 */


Control.prototype.getDisplayDataForCrop = function () {// subclasses can implement this
};

Control.prototype.getDisplayDataBBoxes = function () {
  let padding = this.clipBBoxSelectionPadding();
  let clipBBox;

  if (padding > 0) {
    clipBBox = _types.default.shrinkRectangle(this._clipBBox, -padding);
  } else {
    clipBBox = _balsamiq.default.clone(this._clipBBox);
  }

  var res = {
    boundsNoMarkup: _balsamiq.default.clone(this._userBBoxNoMarkup),
    clipNoMarkup: _balsamiq.default.clone(this._clipBBoxNoMarkup),
    bounds: _balsamiq.default.clone(this._userBBox),
    clip: clipBBox,
    origin: _balsamiq.default.clone(this.getActualProperty("origin"))
  };

  if (this._cropData !== null) {
    res.cropMask = _balsamiq.default.clone(this._cropData.cropMask);
    res.useCropMask = this._cropData.useCropMask;
    res.cropOffset = _balsamiq.default.clone(this._cropData.cropOffset);
    res.uncroppedSize = this._cropData.uncroppedSize ? _balsamiq.default.clone(this._cropData.uncroppedSize) : {
      width: 10,
      height: 10
    };
  }

  return res;
};

Control.prototype.isGroup = function () {
  return false;
}; // selected param refers only to this control, not to its children!


Control.prototype.getDisplayDataTree = function (selected, zoom) {
  // Balsamiq.assert(arguments.length === 2);
  _balsamiq.default.assert(this.children);

  var children = this.children.map(function (child) {
    return child.getDisplayDataTree(false, zoom);
  }, this);
  return new _displaynode.DisplayNode(this.getDisplayData({
    selected,
    zoom
  }), this.coreObjectID.controlID, children);
};

Control.prototype.updateData = function () {// add/remove children and set their properties here
  // create/destroy other objects and configure them here
  // typically subclasses implement this method
};

Control.prototype.preUpdateLayout = function () {
  this._displayList = {
    items: [],
    skinName: this._context.skin["#skinName"]
  };
  this._masks = [];
  this._colorTransforms = [];
  this._linkAreas = []; // { bounds: Rectangle, target: link, continueOnNextLine: boolean }
};

Control.prototype.updateLayout = function (p_localBbox) {// the bbox is already up-to-date here
  // children and other objects are supposed to be already up-to-date and measured at this time
  // subclasses must implement this method
  // p_localBbox is a rectangle with the bbox in local coordinates
};

Control.prototype.updateAll = function () {
  // rebuild the control
  this.updateData();
  this.updateBBoxes();
  this.preUpdateLayout();
  this.updateLayout(this.localBBox());

  this._context.notifyChange(this.coreObjectID, Consts.ControlChanges.DISPLAY_LIST);
};

Control.prototype.updateBBoxes = function () {
  let oldW = this._userBBox ? this._userBBox.width : -1;
  let oldH = this._userBBox ? this._userBBox.height : -1; // Update BBoxes using control custom method

  var {
    changedBBox,
    changedCropMask
  } = this.updateUserBBoxPS();
  var changedClipBBox = this.updateClipBBoxPS();
  this.updateUserBBoxNoMarkupPS(this._userBBox);
  this.updateClipBBoxNoMarkupPS(this._clipBBox);

  if (changedBBox || changedClipBBox) {
    let changedSize = this._userBBox.width != oldW || this._userBBox.height != oldH;

    if (changedSize) {
      this._context.notifyChange(this.coreObjectID, Consts.ControlChanges.DISPLAY_LIST);
    } else {
      this._context.notifyChange(this.coreObjectID, Consts.ControlChanges.POSITION);
    }
  }

  if (changedCropMask) {
    this._context.notifyChange(this.coreObjectID, Consts.ControlChanges.DISPLAY_LIST);
  }
};

Control.prototype.updateClipBBoxNoMarkupPS = function (clipBBox) {
  var bbox;

  if (this.isMarkup()) {
    let origin = this.getActualProperty("origin");
    bbox = _types.default.newRectangle(origin.x, origin.y, 0, 0);
  } else {
    bbox = _balsamiq.default.clone(clipBBox);
  }

  this.setClipBBoxNoMarkupPS(bbox.x, bbox.y, bbox.width, bbox.height);
};

Control.prototype.getUncroppedUserBBox = function (includeMarkup) {// subclasses can implement this
};

Control.prototype.updateUserBBoxNoMarkupPS = function (userBBox) {
  var bbox;

  if (this.isMarkup()) {
    let origin = this.getActualProperty("origin");
    bbox = _types.default.newRectangle(origin.x, origin.y, 0, 0);
  } else {
    bbox = _balsamiq.default.clone(userBBox);
  }

  this.setUserBBoxNoMarkupPS(bbox.x, bbox.y, bbox.width, bbox.height);
};

Control.prototype.updateUserBBoxPS = function () {
  var changed = {
    changedBBox: false,
    changedCropMask: false
  };

  if (this.isGroup() && isControlCropped(this) && this.hasCropData() && this.children.length > 0) {
    // Needed during interactive move of controls inside cropped group
    // 1 Get children BBox in CS
    // 2 Add origin to move to PS
    // 3 Crop this BBox
    // 4 Update Crop Mask since BBox origin may be moved respect to control origin
    // Get all children bbox in component loacal space (uncropped group) (x,y may be negaive during resize)
    let childrenBBoxCS = this._calcChildrenBBoxCS(false, false);

    this.setUncroppedSize({
      width: childrenBBoxCS.width,
      height: childrenBBoxCS.height
    });
    let origin = this.getActualProperty("origin");

    let crop = _types.default.newRectangleFromCrop(this.getActualProperty("crop"));

    childrenBBoxCS.x += origin.x + this.getCropOffset().x;
    childrenBBoxCS.y += origin.y + this.getCropOffset().y;

    let croppedRect = _types.default.calcCroppedRectangle(crop, childrenBBoxCS); // Crop Mask reference system is in Origin, while control is drawn at croppedrect coordinates, so
    // crop mask relative coordinates are the difference between the two.
    // Crop Offset is not moved since control origin and crop offset are usually kept still while relative coordinates of controls change


    let cropMask = {
      x: croppedRect.x - origin.x,
      y: croppedRect.y - origin.y,
      width: croppedRect.width,
      height: croppedRect.height
    };
    changed.changedCropMask = this.setCropMask(_types.default.makeMaskRect(cropMask), true);
    changed.changedBBox = this.setUserBBoxPS(croppedRect.x, croppedRect.y, croppedRect.width, croppedRect.height);
  } else {
    let origin = this.getActualProperty("origin");

    let childrenOrigin = this._calcChildrenBBoxCS(false, false);

    let size = this.computeSize();
    changed.changedBBox = this.setUserBBoxPS(origin.x + childrenOrigin.x, origin.y + childrenOrigin.y, size.width, size.height);
  }

  return changed;
};

Control.prototype.updateClipBBoxPS = function () {
  var clipBBox = _balsamiq.default.clone(this._userBBox);

  return this.setClipBBoxPS(clipBBox.x, clipBBox.y, clipBBox.width, clipBBox.height);
};

Control.prototype.getParent = function () {
  return this.parent;
};

Control.prototype._calcChildrenBBoxCS = function (clip, skipMarkup) {
  _balsamiq.default.assertBool(skipMarkup);

  _balsamiq.default.assertBool(clip); // return the bounding box of all children in their local reference system


  var children;

  if (skipMarkup) {
    children = _balsamiq.default.filter(this.children, child => {
      return !child.getActualProperty("markup");
    });
  } else {
    children = this.children;
  }

  var bboxprop;

  if (clip) {
    if (skipMarkup) {
      bboxprop = "_clipBBoxNoMarkup";
    } else {
      bboxprop = "_clipBBox";
    }
  } else {
    if (skipMarkup) {
      bboxprop = "_userBBoxNoMarkup";
    } else {
      bboxprop = "_userBBox";
    }
  }

  var bboxes = children.map(child => {
    return child[bboxprop];
  }); // Skip empty bboxes: groups containing only markup?

  bboxes = _balsamiq.default.filter(bboxes, bbox => {
    return bbox.width > 0 && bbox.height > 0;
  });

  var bounds = _types.default.rectangleListUnion(bboxes);

  if (!bounds) {
    bounds = _types.default.newRectangle(0, 0, 0, 0);
  }

  return bounds;
};

Control.prototype.calcChildrenClipBBoxCS = function (skipMarkup) {
  _balsamiq.default.assertBool(skipMarkup);

  return this._calcChildrenBBoxCS(true, skipMarkup);
};

Control.prototype.calcChildrenUserBBoxCS = function (skipMarkup) {
  _balsamiq.default.assertBool(skipMarkup);

  return this._calcChildrenBBoxCS(false, skipMarkup);
};

Control.prototype.calcChildrenOriginCS = function (skipMarkup) {
  // return the top-left 'origin' property of the children in their local reference system
  var origin, childOrigin;
  var i;

  if (this.children.length === 0) {
    origin = _types.default.newPoint(0, 0);
  } else {
    origin = _balsamiq.default.clone(this.children[0].getActualProperty("origin"));

    for (i = 1; i < this.children.length; i += 1) {
      childOrigin = this.children[i].getActualProperty("origin");
      origin.x = Math.min(origin.x, childOrigin.x);
      origin.y = Math.min(origin.y, childOrigin.y);
    }
  }

  return origin;
};

Control.prototype.addChildren = function (p_children) {
  p_children.forEach(p_child => {
    _balsamiq.default.assert(CoreObject.makeParentControlCoreObjectID(p_child.coreObjectID).isEqual(this.coreObjectID), "trying to add a child that doesn't belong here"); //        Balsamiq.assert(p_child.coreObjectID.getParentControlID().isEqual(this.coreObjectID), "trying to add a child that doesn't belong here");


    p_child.parent = this;
    this._nextChildControlID = Math.max(this._nextChildControlID, p_child.controlID + 1);
    this.children.push(p_child);

    this._context.notifyChange(p_child.coreObjectID, Consts.ControlChanges.ADDED);
  });
};

Control.prototype.addChildAtIndex = function (p_child, p_index) {
  _balsamiq.default.assertNumber(p_index);

  _balsamiq.default.assert(p_index >= 0);

  if (p_index > this.children.length) {
    p_index = this.children.length;
  }

  _balsamiq.default.assert(CoreObject.makeParentControlCoreObjectID(p_child.coreObjectID).isEqual(this.coreObjectID), "trying to add a child that doesn't belong here"); //        Balsamiq.assert(p_child.coreObjectID.getParentControlID().isEqual(this.coreObjectID), "trying to add a child that doesn't belong here");


  p_child.parent = this;
  this._nextChildControlID = Math.max(this._nextChildControlID, p_child.controlID + 1);
  this.children.splice(p_index, 0, p_child);

  this._context.notifyChange(p_child.coreObjectID, Consts.ControlChanges.ADDED);
};

Control.prototype.removeChildAtIndex = function (p_index) {
  _balsamiq.default.assert(p_index >= 0 && p_index < this.children.length, "child index out of bounds");

  var c = this.children[p_index];
  c.cleanUp();
  this.children.splice(p_index, 1);

  this._context.notifyChange(c.coreObjectID, Consts.ControlChanges.REMOVED);
};

Control.prototype.removeChild = function (controlID) {
  var childIndex = this.children.findIndex(function (p_control) {
    return _balsamiq.default.isEqual(p_control.controlID, controlID);
  });
  this.removeChildAtIndex(childIndex);
};

Control.prototype.removeAllChildren = function () {
  while (this.children.length > 0) {
    this.removeChildAtIndex(this.children.length - 1);
  }
};

Control.prototype.getChildIndex = function (controlID) {
  _balsamiq.default.assertNumber(controlID);

  return this.children.findIndex(function (p_control) {
    return _balsamiq.default.isEqual(p_control.controlID, controlID);
  });
};

Control.prototype.getChildrenOrder = function () {
  return this.children.map(control => {
    return control.controlID;
  });
};

Control.prototype.setChildIndex = function (childID, index) {
  _balsamiq.default.assertNumber(index);

  _balsamiq.default.assertNumber(childID);

  if (index < 0) {
    return false;
  }

  if (index >= this.children.length) {
    index = this.children.length - 1;
  }

  var oldIndex = this.children.findIndex(control => {
    return control.controlID === childID;
  });

  if (oldIndex === -1) {
    return false;
  }

  var control = this.children[oldIndex];
  this.children.splice(oldIndex, 1);
  this.children.splice(index, 0, control);
  return true;
};

Control.prototype.setChildrenOrder = function (orderedControlIDs) {
  _balsamiq.default.assertArray(orderedControlIDs);

  _balsamiq.default.assert(_balsamiq.default.uniq(orderedControlIDs).length == orderedControlIDs.length, "orderedControlIDs has duplicate controlIDs");

  _balsamiq.default.assert(orderedControlIDs.length == this.children.length, "orderedControlIDs has wrong length");

  var oldChildren = this.children;
  var newChildren = [];
  orderedControlIDs.forEach(controlID => {
    var index = this.getChildIndex(controlID);

    _balsamiq.default.assert(index >= 0, "orderedControlIDs: can't find a control with controlID=" + controlID);

    newChildren.push(oldChildren[index]);
  });
  this.children = newChildren;

  this._context.notifyChange(this.coreObjectID, Consts.ControlChanges.CHILDREN_ORDER);
};

Control.prototype.getChildrenControlIDs = function () {
  // return an array of CoreObjectIDs of the children controls
  return this.children.map(c => {
    return c.coreObjectID;
  });
};

Control.prototype.getControl = function (p_relativeControlID) {
  // p_relativeControlID is the relative path to the control starting from this (array of integers)
  if (!p_relativeControlID || p_relativeControlID.length === 0) {
    return this;
  }

  let child = this.children.find(function (p_control) {
    return _balsamiq.default.isEqual(p_control.controlID, p_relativeControlID[0]);
  });

  if (!child) {
    return null;
  }

  return child.getControl(p_relativeControlID.slice(1));
};

Control.prototype.getNextControlID = function () {
  var r = this._nextChildControlID;
  this._nextChildControlID += 1;
  return r;
};
/**
 * 
 * @param {number=} p_existingID 
 * @returns {import("core/core/coreobject").CoreObjectID}
 */


Control.prototype.buildChildCoreObjectID = function (p_existingID) {
  var childID = p_existingID !== undefined ? p_existingID : this.getNextControlID();
  return CoreObject.makeChildControlCoreObjectID(this.coreObjectID, childID); //    return this.coreObjectID.getChildControlID((p_existingID !== undefined) ? p_existingID : this.getNextControlID());
};

Control.prototype.setClipBBoxNoMarkupPS = function (p_x, p_y, p_width, p_height) {
  this._clipBBoxNoMarkup = _types.default.newRectangle(p_x, p_y, p_width, p_height);
};

Control.prototype.setUserBBoxNoMarkupPS = function (p_x, p_y, p_width, p_height) {
  this._userBBoxNoMarkup = _types.default.newRectangle(p_x, p_y, p_width, p_height);
};

Control.prototype.setUserBBoxPS = function (p_x, p_y, p_width, p_height) {
  // low level function to set this._bbox and send the event
  var changed = this._userBBox === null || this._userBBox.x !== p_x || this._userBBox.y !== p_y || this._userBBox.width !== p_width || this._userBBox.height !== p_height;
  this._userBBox = _types.default.newRectangle(p_x, p_y, p_width, p_height); // if (changed) {
  //     // this notification is good for all bboxes of the control
  //     this._context.notifyChange(this.coreObjectID, ControlChanges.POSITION);
  //
  //     if (this.parent) {
  //         // let the parent control to adjust its bbox
  //         this.parent.onChildBboxChange(this.controlID);
  //     }
  // }

  return changed;
};

Control.prototype.setClipBBoxPS = function (p_x, p_y, p_width, p_height) {
  // low level function to set this._bbox and send the event
  var changed = this._clipBBox === null || this._clipBBox.x !== p_x || this._clipBBox.y !== p_y || this._clipBBox.width !== p_width || this._clipBBox.height !== p_height;
  this._clipBBox = _types.default.newRectangle(p_x, p_y, p_width, p_height); // if (changed) {
  //     if (this.parent) {
  //         this.parent.onChildBboxChange(this.controlID);
  //     }
  // }

  return changed;
};

Control.prototype.localBBox = function () {
  // return the bbox in local coordinates
  var origin = this.getActualProperty("origin");
  return _types.default.newRectangle(this._userBBox.x - origin.x, this._userBBox.y - origin.y, this._userBBox.width, this._userBBox.height);
};

Control.prototype.computeSize = function () {
  // Return the size of the control
  if (this.propertyDefinitions.size) {
    const sizeProp = this.getActualProperty("size");
    return _types.default.isAutoSize(sizeProp) ? this.measure(sizeProp) : sizeProp;
  }

  return this.measure();
};
/**
 * Mandatory to implement for controls which support autosize, or do not have the "size" property.
 * @param {Object=} p_sizeProperty Value of the "size" property (only if the controls supports it)
 * @return {Object} Measured size
 */


Control.prototype.measure = function (p_sizeProperty) {
  throw new Error("Control " + this.controlTypeID + " does not implement measure"); // return Types.newSize(0, 0);
};

Control.prototype.onChildBboxChange = function (p_childControlID) {// called by a child when its bbox changes
  // usually do nothing, but Groups and some compound controls will recompute their bboxes and maybe move their origin
};

Control.prototype.onChildOriginChange = function (p_childControlID) {// called by a child when its 'origin' property changes
  // usually do nothing, but Groups and some compound controls will recompute the origin of the reference system
};

Control.prototype.onReferenceSystemOriginChange = function (p_offset) {
  // called by the parent control when its reference system changes.
  // this control must update its actual 'origin' property and the bbox.
  // the message must not be passed recursively to the children of this control.
  // p_offset is the offset of the reference system origin, so here the opposite must be applied.
  var origin = this.getActualProperty("origin");

  var newOrigin = _types.default.newPoint(origin.x - p_offset.x, origin.y - p_offset.y);

  this._setActualProperty("origin", newOrigin); //    if (this.overriddenProperties) {
  //        this.overriddenProperties.origin = newOrigin;
  //    } else {
  //        this.properties.origin = newOrigin;
  //    }


  this._userBBox.x = this._userBBox.x - p_offset.x;
  this._userBBox.y = this._userBBox.y - p_offset.y;
  this._clipBBox.x = this._clipBBox.x - p_offset.x;
  this._clipBBox.y = this._clipBBox.y - p_offset.y;
  this.updateUserBBoxNoMarkupPS(this._userBBox);
  this.updateClipBBoxNoMarkupPS(this._clipBBox);

  this._context.notifyChange(this.coreObjectID, Consts.ControlChanges.POSITION);
};

Control.prototype.isMarkup = function () {
  return this.getActualProperty("markup");
};

Control.prototype.setLocked = function (value) {
  _balsamiq.default.assertBool(value);

  if (this._locked != value) {
    this._locked = value;

    this._context.notifyChange(this.coreObjectID, Consts.ControlChanges.DISPLAY_LIST); // notify display list to redraw the entire control

  }
};

Control.prototype.isLocked = function () {
  return this._locked;
};

Control.prototype.isLockedOrIsInLockedAncestor = function () {
  return this._locked || isInLockedAncestor(this);
};

Control.prototype.hasLockedChildren = function () {
  return this.children.some(control => control.isLocked());
};

Control.prototype.onContextChange = function () {
  // something in the context did change
  // it could be the font family for the whole mockup, the skin, or anything else
  // typically the control must be rebuilt
  this.children.forEach(function (p_child) {
    p_child.onContextChange();
  }, this);
  this.updateAll();
};

Control.prototype.style = function (p_selector) {
  if (p_selector === "default_selection_color") {
    return this._context.selectionColor;
  } else if (p_selector === "default_font_size") {
    return this._context.defaultFontSize;
  } else if (p_selector === "label_font_family") {
    return this._context.fontFamily;
  } else if (p_selector === "default_link_color") {
    return this._context.linkColor;
  }

  _balsamiq.default.assert(_balsamiq.default.has(this._context.skin, p_selector), "The skin must have " + p_selector);

  return this._context.skin[p_selector];
};
/**
 *
 * @param {string} p_selector
 * @returns {{ width: number, height: number }}
 */


Control.prototype.styleSVGSize = function (p_selector) {
  var value = this.style(p_selector);

  _balsamiq.default.assert(value.type === "SVG_PATHS" || value.type === "SVG_SCALE9", "The skin selector is not an SVG item");

  _balsamiq.default.exists(value.width);

  _balsamiq.default.exists(value.height);

  return _types.default.newSize(value.width, value.height);
};

Control.prototype.newTextStyle = function () {
  var textStyle = {
    fontFamily: this.style("label_font_family"),
    fontSize: this._context.defaultFontSize,
    bold: false,
    italic: false,
    underline: false,
    textColor: this.style("default_text_color"),
    linkColor: this._context.linkColor
  };
  return textStyle;
};

Control.prototype.newStandardTextStyle = function (p_customValues) {
  var textStyle = {
    fontFamily: _balsamiq.default.has(p_customValues, "fontFamily") ? p_customValues.fontFamily : this.style("label_font_family"),
    fontSize: _balsamiq.default.has(p_customValues, "fontSize") ? p_customValues.fontSize : this.getActualProperty("fontSize"),
    bold: _balsamiq.default.has(p_customValues, "bold") ? p_customValues.bold : this.getActualProperty("bold"),
    italic: _balsamiq.default.has(p_customValues, "italic") ? p_customValues.italic : this.getActualProperty("italic"),
    underline: _balsamiq.default.has(p_customValues, "underline") ? p_customValues.underline : this.getActualProperty("underline"),
    textColor: _balsamiq.default.has(p_customValues, "textColor") ? p_customValues.textColor : this.style("default_text_color"),
    linkColor: _balsamiq.default.has(p_customValues, "linkColor") ? p_customValues.linkColor : this._context.linkColor
  };
  return textStyle;
};
/**
 *
 * @param {Array} p_srcProperties
 * @param {String?} p_resourcesType
 */


Control.prototype.setDependencies = function (p_srcProperties) {
  // avoid premature setDependencies if control is not yet fully built
  if (this.coreObjectID.controlID && this.coreObjectID.controlID.length !== 0) {
    if (!_balsamiq.default.isEqual(this._dependencies, p_srcProperties)) {
      this._dependencies = p_srcProperties;

      this._context.setDependencies(this.coreObjectID, p_srcProperties);
    }
  }
};

Control.prototype.onResourceChange = function () {
  this.updateAll();
};
/**
 * 
 * @returns {import("core/model/mockup").Mockup=}
 */


Control.prototype.getMockup = function () {
  return this.parent ? this.parent.getMockup() : null;
};

Control.prototype.setMacros = function (p_macros) {
  this._macros = p_macros;
};

Control.prototype.resolveMacro = function (p_macro) {
  var text = p_macro;

  if (p_macro === _types.default.TextMacro.MOCKUP_NAME && this._context && this._context.getMockupName) {
    text = this._context.getMockupName();
  }

  return text;
};

Control.prototype.updateAllMacros = function () {
  // check if the control depends on a macro
  if (this._macros.length > 0) {
    this.updateAll();
  } // pass the message down


  this.children.forEach(function (child) {
    child.updateAllMacros();
  }, this);
};

Control.prototype.setReady = function (p_ready) {
  _balsamiq.default.assertBool(p_ready);

  this.context.setReady(this.coreObjectID, p_ready);
};
/**
 * Checks wether the point is inside the control's rectangle
 * @param {{ x: number, y: number }} p_point - Point must be in control's group coordinate system
 * @returns {boolean}
 */


Control.prototype.hitTest = function (p_point) {
  return _types.default.testPointInRectangle(p_point, this.getUserBBox(true)); // TODO withMarkup
};
/**
 * Check wether the control is contained by the given rectangle
 * @param {{ x: number, y: number, width: number, height: number }} p_rect - Rect x,y must be in control's group coordinate system
 * @returns {boolean}
 */


Control.prototype.isContainedByRect = function (p_rect) {
  var res = _types.default.rectangleContains(p_rect, this.getUserBBox(true)); // TODO  withMarkup


  return res;
};

Control.prototype.getSelectionBaseColor = function () {
  return 0x000000;
};

Control.prototype.getSelectionColor = function () {
  var color = this.getSelectionBaseColor();

  if (this.isLockedOrIsInLockedAncestor()) {
    color = _colors.default.adjustBrightness(color, 112);
  }

  return color;
};

Control.prototype.getSelectionAlpha = function (otherUsersSelectionStatus) {
  var alpha = 0.2;

  if (otherUsersSelectionStatus !== undefined) {
    _balsamiq.default.assertEnum(otherUsersSelectionStatus, Consts.ConnectedUserStatus);

    switch (otherUsersSelectionStatus) {
      case Consts.ConnectedUserStatus.Active:
        alpha = alpha * UI.OtherUsers.ActiveUserSelectionAlpha;
        break;

      case Consts.ConnectedUserStatus.Waiting:
        alpha = alpha * UI.OtherUsers.WaitingUserSelectionAlpha;
        break;

      default:
        break;
    }
  }

  return alpha;
};

Control.prototype.clipBBoxSelectionPadding = function () {
  return 0;
};
/**
 *
 * @param {*} param0
 * @returns {Object[]}
 */


Control.prototype.getSelectionDisplayItems = function (_ref2) {
  let {
    otherUsersSelectionStatus,
    zoom,
    withLockedBackground = true
  } = _ref2;
  //var color = this.isLocked() ? 0x4A3E18 : 0x000000;
  var bbox = this.localBBox();
  bbox.x -= this._cropData ? this._cropData.cropOffset.x : 0;
  bbox.y -= this._cropData ? this._cropData.cropOffset.y : 0;
  let displayItems = [];

  if (withLockedBackground && this.isLocked()) {
    _balsamiq.default.assertNumber(zoom);

    let adjustedSize = UI.LockedControls.LockRectSize / zoom;
    displayItems.push(_displayitem.default.newFilledRectangle(_types.default.newRectangle(bbox.x - adjustedSize, bbox.y, adjustedSize, adjustedSize), this.getSelectionColor(), 0, this.getSelectionAlpha()));
  }

  displayItems.push(_displayitem.default.newFilledRectangle(bbox, this.getSelectionColor(otherUsersSelectionStatus), 0, this.getSelectionAlpha(otherUsersSelectionStatus)));

  _displayitem.default.setSelectionDisplayItemsFlag(displayItems); // Mark as selection so that they won't be clipped away


  return displayItems;
};

Control.prototype.hasCustomHandles = function () {
  return false;
};

Control.prototype.getResizeOption = function () {
  return this.hasPropertyDefinition("size") ? this.propertyDefinitions.size.options.resize : Props.ResizeOptions.NO;
};

Control.prototype.getSnapLines = function (withMarkup) {
  _balsamiq.default.assertBool(withMarkup);

  return Snapping.makeSnapLinesForBbox(this.getUserBBox(withMarkup), false);
};

Control.prototype.getName = function () {
  if (this.hasPropertyDefinition("controlName")) {
    return this.getActualProperty("controlName");
  } else {
    return "";
  }
};
/**
 * @returns {string}
 */


Control.prototype.getDescription = function () {
  var res, text;

  if (this.hasPropertyDefinition("text") && (text = this.getActualProperty("text")) !== "") {
    res = TextUtils.unformatText(text);
  } else if (this.hasPropertyDefinition("color")) {
    res = _colors.default.newHexColorFromInt(this.getActualProperty("color"));
  } else {
    res = `${this.controlID}`;
  }

  return res;
};

Control.prototype.isSymbol = function () {
  return false;
};

Control.prototype.getVersion = function () {
  let maxVersion = _wireframeVersion.BaseVersion;

  for (const child of this.children) {
    maxVersion = (0, _semver.highestSemVer)(child.getVersion(), maxVersion);
  }

  return maxVersion;
};

Control.prototype.containsSymbolRecursive = function () {
  for (const child of this.children) {
    if (child.containsSymbolRecursive()) {
      return true;
    }
  }

  return this.isSymbol();
};

Control.prototype.isInsideSymbol = function () {
  if (this.parent) {
    return this.parent.isSymbolOrInsideSymbol();
  } else {
    return false;
  }
};

Control.prototype.isSymbolOrInsideSymbol = function () {
  if (this.isSymbol() || this.isInsideSymbol()) {
    return true;
  } else {
    return false;
  }
};

Control.prototype.updateOrigin = function () {
  // update the 'origin' property of the group, based on the top-left origin of the children controls.
  // at the end, the bbox of the group may need an update.
  var uncroppedChildrenBBox = this.calcChildrenUserBBoxCS(false); // uncropped children origin is in Component user Space

  var offset = _types.default.newPoint(uncroppedChildrenBBox.x, uncroppedChildrenBBox.y); // In case of crop we have to recalculate origin even if just the size changes beacuse that will move the cropped origin


  var changedWidth = this.hasCropData() ? this.getUncroppedSize().width !== uncroppedChildrenBBox.width : false;
  var changedHeight = this.hasCropData() ? this.getUncroppedSize().height !== uncroppedChildrenBBox.height : false; // When moving controls inside cropped group we have to recalculate the origin that depends on the crop property

  if (offset.x !== 0 || offset.y !== 0 || changedHeight || changedWidth) {
    var uncroppedGroupOriginPS = this.calcUncroppedOriginPS();

    var newUncroppedGroupOriginPS = _types.default.newPoint(uncroppedGroupOriginPS.x + offset.x, uncroppedGroupOriginPS.y + offset.y); // set the actual 'origin' in a low-level fashion


    var newCroppedGroupOriginPS;

    if (isControlCropped(this)) {
      let newUncroppedGroupRect = _types.default.newRectangle(newUncroppedGroupOriginPS.x, newUncroppedGroupOriginPS.y, uncroppedChildrenBBox.width, uncroppedChildrenBBox.height);

      const crop = _types.default.newRectangleFromCrop(this.getActualProperty("crop"));

      newCroppedGroupOriginPS = _types.default.calcCroppedOrigin(crop, newUncroppedGroupRect);
      this.setCropOffset({
        x: -(newUncroppedGroupRect.width * crop.x) / 1000,
        y: -(newUncroppedGroupRect.height * crop.y) / 1000
      });
    } else {
      newCroppedGroupOriginPS = newUncroppedGroupOriginPS;
    }

    this._setActualProperty("origin", newCroppedGroupOriginPS); // move the children


    this.children.forEach(child => {
      child.onReferenceSystemOriginChange(offset);
    }); // and update the bbox

    this.updateBBoxes();

    if (this.parent) {
      this.parent.onChildOriginChange(this.controlID);
    }

    if (this.hasCropData()) {
      this.updateCropData();
    }
  } else if (this.children.length === 0 && !this.isSymbol()) {
    this.setActualProperty("origin", {
      x: 0,
      y: 0
    });
    this.updateBBoxes();

    if (this.parent) {
      this.parent.onChildOriginChange(this.controlID);
    }

    if (this.hasCropData()) {
      this.updateCropData();
    }
  }
};

Control.prototype.resizeBBox = function (newBBox) {
  var propNames = ["origin"];
  var propValues = [_types.default.newPoint(newBBox.x, newBBox.y)];

  if (this.hasPropertyDefinition("size")) {
    propNames.push("size");
    propValues.push(_types.default.newSize(newBBox.width, newBBox.height));
  }

  return {
    propNames: propNames,
    propValues: propValues
  };
};

Control.prototype.setCropData = function (_ref3) {
  let {
    cropMask,
    cropOffset,
    useCropMask
  } = _ref3;

  _balsamiq.default.assertMaskRect(cropMask);

  _balsamiq.default.assertPoint(cropOffset);

  _balsamiq.default.assertBool(useCropMask);

  this._cropData = {
    cropMask: cropMask,
    cropOffset: cropOffset,
    useCropMask: useCropMask
  };
};

Control.prototype.setCropOffset = function (cropOffset) {
  _balsamiq.default.assertPoint(cropOffset);

  if (!this._cropData) {
    this._cropData = {};
  }

  this._cropData.cropOffset = cropOffset;
};

Control.prototype.setUncroppedSize = function (uncroppedSize) {
  _balsamiq.default.assertSize(uncroppedSize);

  if (!this._cropData) {
    this._cropData = {};
  }

  this._cropData.uncroppedSize = uncroppedSize;
};

Control.prototype.getUncroppedSize = function () {
  if (this._cropData) {
    return this._cropData.uncroppedSize;
  } else {
    return undefined;
  }
};

Control.prototype.getCropOffset = function () {
  if (this._cropData) {
    return _balsamiq.default.clone(this._cropData.cropOffset);
  } else {
    return _types.default.newPoint(0, 0);
  }
};

Control.prototype.setCropMask = function (cropMask, useCropMask) {
  _balsamiq.default.assertMaskRect(cropMask);

  _balsamiq.default.assertBool(useCropMask);

  let changed = true;

  if (this._cropData.cropMask) {
    let oldRect = this._cropData.cropMask.rect;
    let newRect = cropMask.rect;
    changed = oldRect.x != newRect.x || oldRect.y != newRect.y || oldRect.width != newRect.width || oldRect.height != newRect.height;
  }

  this._cropData.cropMask = cropMask;
  this._cropData.useCropMask = useCropMask;
  return changed;
};

Control.prototype.hasCropData = function () {
  return this._cropData !== null;
};

Control.prototype.resetCropData = function () {
  this._cropData = null;
};
/**
 * @returns {{ x: number, y: number }}
 */


Control.prototype.calcUncroppedOriginPS = function () {
  // Control must have active crop (isControlCropped) and must have an updated cropData
  var cropOffsetX = isControlCropped(this) && this.hasCropData() ? this._cropData.cropOffset.x : 0;
  var cropOffsetY = isControlCropped(this) && this.hasCropData() ? this._cropData.cropOffset.y : 0;
  var origin = this.getActualProperty("origin");
  return _types.default.newPoint(origin.x + cropOffsetX, origin.y + cropOffsetY);
};
/**
 * @returns { Control }
 */


Control.prototype.getFirstLevelParent = function () {
  return this.coreObjectID.controlID.length === 1 ? this : this.parent.getFirstLevelParent();
}; // -----------------------


function getControlInitialProperties(controlTypeID) {
  var ControlClass = classRegistry[controlTypeID];

  _balsamiq.default.exists(ControlClass);

  return ControlClass.prototype.controlInitialProperties;
}

function getControlNames(controlTypeID) {
  var ControlClass = classRegistry[controlTypeID];

  _balsamiq.default.exists(ControlClass);

  return {
    typeName: ControlClass.prototype.controlTypeName,
    alternateNames: ControlClass.prototype.controlAlternateNames
  };
}

function lastControlID(p_coreObjectID) {
  _balsamiq.default.exists(p_coreObjectID);

  _balsamiq.default.exists(p_coreObjectID.controlID);

  return p_coreObjectID.controlID[p_coreObjectID.controlID.length - 1];
}

function isControlCropped(p_control) {
  _balsamiq.default.existsClass(p_control, Control);

  if (p_control.hasPropertyDefinition("crop")) {
    return !_types.default.isCropAll(_types.default.newRectangleFromCrop(p_control.getActualProperty("crop")));
  } else {
    return false;
  }
}

function isInLockedAncestor(p_control) {
  _balsamiq.default.existsClass(p_control, Control);

  let res = false;
  let control = p_control;

  while (control && !res) {
    control = control.getParent();

    if (control) {
      res = control.isLocked();
    }
  }

  return res;
}

function assertControlsInSameGroup(controls) {
  _balsamiq.default.assertArray(controls);

  var theGroupID = null;
  controls.forEach(control => {
    var group = control.parent;

    if (theGroupID === null) {
      theGroupID = group.coreObjectID.controlID;
    } else {
      _balsamiq.default.assert(ControlIDs.areEqual(theGroupID, group.coreObjectID.controlID));
    }
  });
}

/***/ }),

/***/ "./js/core/controls/coverflow.js":
/*!***************************************!*\
  !*** ./js/core/controls/coverflow.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _scrollbars = __webpack_require__(/*! core/bricks/scrollbars */ "./js/core/bricks/scrollbars.js");

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var ImageResource = __webpack_require__(/*! core/model/image */ "./js/core/model/image.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::CoverFlow",
  controlTypeName: "Cover Flow",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(493, 198),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.SCROLLBAR(),
    src: {
      defaultValue: null,
      dataType: 'object'
    },
    filter: {
      defaultValue: false,
      dataType: 'bool'
    },
    rotation: {
      defaultValue: 0,
      dataType: 'int'
    }
  }
});

control.CONTROL_PROTOTYPE.updateData = function () {
  var src = this.getActualProperty("src");
  this.setDependencies(src ? [src] : []);
  this.nativeImageSize = null;
  this.nativeImageID = null;
  this.resourceState = null;

  if (src) {
    var resource = this._context.getDependency(src.resourceID);

    if (resource && resource instanceof ImageResource.Image) {
      this.nativeImageSize = resource.getImageSize();
      this.nativeImageID = resource.getImageID();
      this.resourceState = this.nativeImageID ? _types.default.ResourceState.AVAILABLE : _types.default.ResourceState.PENDING;
    } else {
      this.resourceState = _types.default.ResourceState.NOT_AVAILABLE;
    }

    this._context.setReady(this.coreObjectID, this.resourceState === _types.default.ResourceState.AVAILABLE || this.resourceState === _types.default.ResourceState.NOT_AVAILABLE);
  }
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBox) {
  var coverflowBbox = p_localBox;
  var scrollbar;

  if (this.getActualProperty("verticalScrollbar") && p_localBox.height > _scrollbars.Sizes.ScrollbarFixedSize) {
    coverflowBbox = _types.default.newRectangle(0, 0, p_localBox.width, p_localBox.height - _scrollbars.Sizes.ScrollbarFixedSize);
    var width = Math.round(p_localBox.width * 0.75);
    var x = Math.round((p_localBox.width - width) / 2);
    var y = p_localBox.height - _scrollbars.Sizes.ScrollbarFixedSize;
    scrollbar = (0, _scrollbars.makeDisplayItemsForHorizontalScrollbar)(x, y, width, this.getActualProperty('value'), this._context);
  }

  var coverflow = _displayitem.default.newPathsAsset(this._context, "coverflow", coverflowBbox, this.style("coverflow_color"));

  var image;

  if (this.resourceState) {
    var imageWidth = p_localBox.width * 0.31,
        imageHeight = coverflowBbox.height * 0.78;

    var imageBounds = _types.default.newRectangle((p_localBox.width - imageWidth) / 2, (coverflowBbox.height - imageHeight) / 2, imageWidth, imageHeight);

    var state = this.resourceState;

    if (state === _types.default.ResourceState.PENDING) {
      image = (0, _rectangle.makeFilledRectangle)(imageBounds, 0xFF4433);
    } else if (state === _types.default.ResourceState.AVAILABLE) {
      image = _displayitem.default.newImage(this.nativeImageID, imageBounds, this.nativeImageSize.width, this.nativeImageSize.height, this.getActualProperty("filter"), 360 - this.getActualProperty("rotation"));
    }
  }

  this.pushToDisplayList(image, coverflow, scrollbar);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/curlybraces.js":
/*!*****************************************!*\
  !*** ./js/core/controls/curlybraces.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.verticalCurlyBrace = exports.horizontalCurlyBrace = exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _macro = __webpack_require__(/*! core/bricks/macro */ "./js/core/bricks/macro.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var makeCurlyBraceControlClass = function (p_controlTypeID, p_controlTypeName, p_controlAlternateNames, p_defaultDirection, p_allowedDirections, p_defaultMeasure) {
  var control = controlsMod.makeControlClass({
    controlTypeID: p_controlTypeID,
    controlTypeName: p_controlTypeName,
    controlAlternateNames: p_controlAlternateNames,
    props: { ...Props.PropertyDefinitions.SIZE({
        defaultValue: _types.default.newAutoSize(),
        options: {
          resize: Props.ResizeOptions.BOTH,
          autosize: Props.AutoSizeOptions.NO
        }
      }),
      direction: {
        defaultValue: p_defaultDirection,
        enum: p_allowedDirections
      },
      markup: {
        defaultValue: true,
        dataType: 'bool'
      }
    }
  });
  controlsMod.addColorPropertyDefinitions(control, {
    defaultTextColor: "default_text_color",
    oldTextColorName: 'color'
  });
  controlsMod.addStandardTextPropertyDefinitions(control, {
    text: "A paragraph of text.\nA second row of text.",
    multiline: true
  });

  control.CONTROL_PROTOTYPE.updateData = function () {
    var textStyle = this.newStandardTextStyle({
      textColor: this.getActualProperty("textColor")
    });
    let text = (0, _macro.processMacros)(this);
    this.text = textMod.newText(text, textStyle, this._context, true);
  };

  _link.default.addTextLinksMethods(control);

  control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
    return p_defaultMeasure;
  };

  control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
    var direction = this.getActualProperty("direction");
    var isVertical = direction === _types.default.Positions.LEFT || direction === _types.default.Positions.RIGHT;
    var padding = this.style("curlybrace_padding");
    var assetName = isVertical ? "curlybrace_vertical" : "curlybrace_horizontal";
    var assetSize = this.styleSVGSize(assetName);

    var assetBounds = _types.default.newRectangle(0, 0, isVertical ? assetSize.width : p_localBbox.width, isVertical ? p_localBbox.height : assetSize.height);

    var textWidth = isVertical ? p_localBbox.width - padding : p_localBbox.width;
    var textHeight = isVertical ? p_localBbox.height : p_localBbox.height - padding;
    var textBounds;

    if (textWidth > 0 && textHeight > 0) {
      textBounds = _types.default.newRectangle(0, 0, isVertical ? p_localBbox.width - padding : p_localBbox.width, isVertical ? p_localBbox.height : p_localBbox.height - padding);
    }

    var textAlign = _types.default.TextAlign.LEFT;
    var textVerticalPosition = _types.default.Positions.CENTER;
    var rotation = 0;

    if (direction === _types.default.Positions.LEFT) {
      if (textBounds) {
        textBounds.x = padding;
      }
    } else if (direction === _types.default.Positions.RIGHT) {
      assetBounds.x = p_localBbox.width;
      assetBounds.y = p_localBbox.height;
      rotation = 180;
      textAlign = _types.default.TextAlign.RIGHT;
    } else if (direction === _types.default.Positions.TOP) {
      if (textBounds) {
        textBounds.y = padding;
      }

      textAlign = _types.default.TextAlign.CENTER;
      textVerticalPosition = _types.default.Positions.TOP;
    } else if (direction === _types.default.Positions.BOTTOM) {
      assetBounds.x = p_localBbox.width;
      assetBounds.y = p_localBbox.height;
      rotation = 180;
      textAlign = _types.default.TextAlign.CENTER;
      textVerticalPosition = _types.default.Positions.BOTTOM;
    }

    var curlyBrace = _displayitem.default.createDisplayItemsFromSkin(assetName, this.getActualProperty("textColor"), assetBounds, this._context);

    if (rotation) {
      _displayitem.default.setRotation(curlyBrace, rotation);
    }

    var label;

    if (textBounds) {
      label = this.text.makeDisplayItemsAndLinks(textBounds, textAlign, textVerticalPosition);

      _link.default.processTextLinks(this, label.links);
    }

    this.pushToDisplayList(curlyBrace, label ? label.displayItems : null);
  };

  return control;
};

const verticalCurlyBrace = makeCurlyBraceControlClass("com.balsamiq.mockups::VCurly", "V.Curly Brace", ["Vertical Curly Brace"], Props.HorizontalPositions.LEFT, Props.HorizontalPositions, _types.default.newSize(180, 140));
exports.verticalCurlyBrace = verticalCurlyBrace;
const horizontalCurlyBrace = makeCurlyBraceControlClass("com.balsamiq.mockups::HCurly", "H.Curly Brace", ["Horizontal Curly Brace"], Props.VerticalPositions.TOP, Props.VerticalPositions, _types.default.newSize(200, 80));
exports.horizontalCurlyBrace = horizontalCurlyBrace;
var _default = {
  horizontalCurlyBrace,
  verticalCurlyBrace
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/datagrid.js":
/*!**************************************!*\
  !*** ./js/core/controls/datagrid.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var ScrollBars = _interopRequireWildcard(__webpack_require__(/*! core/bricks/scrollbars */ "./js/core/bricks/scrollbars.js"));

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _choice = __webpack_require__(/*! core/bricks/choice */ "./js/core/bricks/choice.js");

var _textconstants = __webpack_require__(/*! core/texts/textconstants */ "./js/core/texts/textconstants.js");

var _textutilities = __webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js");

var _datagrid_utils = __webpack_require__(/*! core/controls/datagrid_utils */ "./js/core/controls/datagrid_utils.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::DataGrid",
  controlTypeName: "Data Grid",
  controlAlternateNames: ["Table"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.BORDER_STYLE(Props.SimpleBorderStyles.SQUARE, Props.SimpleBorderStyles),
    ...Props.PropertyDefinitions.SCROLLBAR(),
    hasHeader: {
      defaultValue: true,
      dataType: 'bool'
    },
    hLines: {
      defaultValue: false,
      dataType: 'bool'
    },
    vLines: {
      defaultValue: true,
      dataType: 'bool'
    },
    rowHeight: {
      defaultValue: 30,
      dataType: 'int'
    },
    rowColor: {
      defaultValue: 0xFFFFFF,
      dataType: 'int',
      label: 'Background',
      options: {
        oldName: 'color'
      }
    },
    alternateRowColor: {
      defaultValue: 0xEEEEEE,
      label: 'Alternate Row',
      dataType: 'int'
    },
    align: {
      defaultValue: Props.TextAlignments.LEFT,
      enum: Props.TextAlignments
    }
  }
});
control.MIN_ROW_HEIGHT = 16;
control.MAX_ROW_HEIGHT = 80;
Controls.addStandardTextPropertyDefinitions(control, {
  text: "Name\\r(job title) ^, Age ^v, Nickname, Employee v\n" + "Giacomo Guilizzoni\\rFounder & CEO, 40, Peldi, (o)\n" + "Marco Botton\\rTuttofare, 38, , [x]\n" + "Mariah Maclachlan\\rBetter Half, 41, Patata, [-]\n" + "Valerie Liberty\\rHead Chef, :), Val, [x]\n" + "[Data Grid Docs     ](https://balsamiq.com/wireframes/desktop/docs/datagrids/), , , [ ]\n" + "{65L, 0R, 35, 0C}",
  multiline: true,
  splitOn: Props.SplitOnTypes.NEWLINE
});
Controls.addSelectionPropertyDefinitions(control, -1);

control.CONTROL_PROTOTYPE.initData = function () {
  this._selectableLabels = [];
  this._rows = [];
};

control.CONTROL_PROTOTYPE.calculateSelectionLabelsForEditText = function () {
  //gets calculated in updateData
  return this._selectableLabels;
};

_link.default.addTextLinksMethods(control, {
  getLinksFunction: function () {
    let links = [];

    this._rows.forEach(row => {
      row.cells.forEach(cellData => {
        if (cellData.text) {
          links = links.concat(cellData.text.getLinks());
        }
      });
    });

    return links;
  },
  replaceLinkTargetFunction: function (linkIndex, newTarget) {
    let linkCount = 0;

    const processNodesAndSetLinkAtIndex = nodes => {
      nodes.forEach(node => {
        if (node.type === _textconstants.NodeTypes.LINK) {
          if (linkCount === linkIndex) {
            if (newTarget === null) {
              node.target = null;
            } else {
              node.target = [newTarget];
            }
          }

          linkCount++;
        } else if (node.children) {
          processNodesAndSetLinkAtIndex(node.children);
        }
      });
    };

    const rowStrings = this._rows.map(row => {
      row.cells.forEach(cell => {
        if (cell.type === _datagrid_utils.CellType.TEXT) {
          processNodesAndSetLinkAtIndex(cell.value);
        }
      });
      return (0, _datagrid_utils.cellsToString)(row.cells);
    });

    if (this._sizesRowString) {
      rowStrings.push(this._sizesRowString);
    }

    return rowStrings.join("\n");
  },
  clearLinksTargetFunction: function () {
    const processNodesAndSetLinkAtIndex = nodes => {
      nodes.forEach(node => {
        if (node.type === _textconstants.NodeTypes.LINK) {
          node.target = null;
        } else if (node.children) {
          processNodesAndSetLinkAtIndex(node.children);
        }
      });
    };

    const rowStrings = this._rows.map(row => {
      row.cells.forEach(cell => {
        if (cell.type === _datagrid_utils.CellType.TEXT) {
          processNodesAndSetLinkAtIndex(cell.value);
        }
      });
      return (0, _datagrid_utils.cellsToString)(row.cells);
    });

    if (this._sizesRowString) {
      rowStrings.push(this._sizesRowString);
    }

    return rowStrings.join("\n");
  }
});

control.CONTROL_PROTOTYPE.getTextWithMockupNames = function () {
  const rowStrings = this._rows.map(row => {
    for (const cell of row.cells) {
      if (cell.type === _datagrid_utils.CellType.TEXT) {
        _link.default.processNodesLinkTargetsToWireframeNames(cell.value, this._context.getDependency.bind(this));
      }
    }

    return (0, _datagrid_utils.cellsToString)(row.cells);
  });

  if (this._sizesRowString) {
    rowStrings.push(this._sizesRowString);
  }

  return rowStrings.join('\n');
};

control.CONTROL_PROTOTYPE.getTextWithResourceIDs = function (textWithMockupNames) {
  const rows = textWithMockupNames.split("\n").map(_datagrid_utils.parseRow);
  const textWithResourceIDs = rows.map(row => {
    for (const cell of row.cells) {
      if (cell.type === _datagrid_utils.CellType.TEXT) {
        cell.value = (0, _textutilities.stringToNodes)(cell.value[0]);

        _link.default.processNodesWireframeNamesToLinkTargets(cell.value, this._context.getMockupIDByName.bind(this));
      }
    }

    return (0, _datagrid_utils.cellsToString)(row.cells);
  });
  return textWithResourceIDs.join('\n');
};

control.CONTROL_PROTOTYPE.getSelectableLabels = function () {
  return this._selectableLabels;
};

control.CONTROL_PROTOTYPE.makeTextStyleWithGivenColor = function (color) {
  const defaultTextColor = this.style("default_text_color");
  const defaultLinkColor = this.style("default_link_color");
  return this.newStandardTextStyle({
    textColor: _colors.default.labelColorGivenBackgroundColor(color, 1, defaultTextColor),
    linkColor: _colors.default.labelColorGivenBackgroundColor(color, 1, defaultLinkColor)
  });
};

control.CONTROL_PROTOTYPE.setColorAndTextStyle = function (row, rowIndex, hasHeader) {
  const selectedIndex = this.getActualProperty("selectedIndex");
  const color = rowIndex === 0 && hasHeader ? 0xCCCCCC : rowIndex === selectedIndex + (hasHeader ? 1 : 0) ? this.style("default_selection_color") : rowIndex % 2 === 0 ? this.getActualProperty("rowColor") : this.getActualProperty("alternateRowColor");
  row.color = color;
  row.textStyle = this.makeTextStyleWithGivenColor(color);
};

control.CONTROL_PROTOTYPE.getColumnWidth = function (columnIndex) {
  let columnWidth = 0;

  for (const row of this._rows) {
    for (const cell of row.cells) {
      if (cell.hasOwnProperty("columnIndex") && cell.columnIndex === columnIndex) {
        columnWidth = Math.max(columnWidth, cell.nativeCellWidth);
        break;
      }
    }
  }

  _balsamiq.default.assert(columnWidth >= 0);

  const paddingMiddle = this.style("datagrid_padding_middle");
  const padding = columnIndex === 0 || columnIndex === this._maxColumnCount - 1 ? paddingMiddle / 2 : paddingMiddle;
  return columnWidth + padding;
};

control.CONTROL_PROTOTYPE.updateData = function () {
  const linePadding = this.style("datagrid_line_padding");
  const rowHeight = this.getActualProperty("rowHeight");
  this._nativeColumnWidths = [];
  this._totalNativeWidth = 0;
  this._totalNativeHeight = 0;
  this._selectableLabels = [];
  this._data = {};
  const lines = this.getActualProperty("text").split("\n");

  const hasSizeRow = _datagrid_utils.isSizeLineRegExp.test(lines[lines.length - 1]);

  const sizeRow = hasSizeRow && lines.pop();
  const hasHeader = this.getActualProperty("hasHeader");
  let maxColumnCount = 0;
  this._rows = lines.map((line, rowIndex) => {
    const row = (0, _datagrid_utils.parseRow)(line, rowIndex, hasHeader);
    this.setColorAndTextStyle(row, rowIndex, hasHeader);
    const previousLine = this._rows.length > rowIndex ? this._rows[rowIndex].cells : null;
    row.cells = row.cells.map((cell, cellIndex) => {
      switch (cell.type) {
        case _datagrid_utils.CellType.CHOICE:
          cell.choice = new _choice.Choice({
            type: cell.choiceData.type,
            state: cell.choiceData.state,
            text: "",
            textStyle: row.textStyle,
            color: cell.choiceData.color ? _colors.default.colorFromString(cell.choiceData.color) : row.textStyle.textColor,
            context: this._context,
            previousChoice: previousLine && previousLine.length > cellIndex && previousLine[cellIndex].type === _datagrid_utils.CellType.CHOICE && previousLine[cellIndex].choice
          });
          cell.nativeCellWidth += 22;
          row.nativeRowHeight = Math.max(row.nativeRowHeight, 22);
          return cell;

        case _datagrid_utils.CellType.TEXT:
          {
            const text = "" + cell.value;
            const isPreviousCellText = previousLine && previousLine.length > cellIndex && previousLine[cellIndex].type === _datagrid_utils.CellType.TEXT;

            if (isPreviousCellText && Text.areTextsEqual(previousLine[cellIndex].text, text, row.textStyle)) {
              cell.text = previousLine[cellIndex].text;
              cell.value = _balsamiq.default.cloneDeep(previousLine[cellIndex].value);
            } else {
              cell.text = Text.newTextNoWrap(text, row.textStyle, this._context);
              cell.value = _balsamiq.default.cloneDeep(cell.text.domTexts[0]);
            }

            cell.nativeCellWidth += cell.text.measuredSize().width;
            row.nativeRowHeight = Math.max(row.nativeRowHeight, cell.text.measuredSize().height + linePadding);
            return cell;
          }

        case _datagrid_utils.CellType.SEPARATOR:
          return cell;
      }
    });
    this._totalNativeHeight += hasHeader && rowIndex === 0 ? row.nativeRowHeight : Math.max(row.nativeRowHeight, rowHeight);

    if (row.selectableLabels) {
      this._selectableLabels.push(row.selectableLabels);
    }

    maxColumnCount = Math.max(maxColumnCount, row.columnCount);
    return row;
  });
  const defaultAlign = this.getActualProperty("align");

  if (hasSizeRow) {
    if (this._sizesRowString !== sizeRow || this._maxColumnCount !== maxColumnCount || this.defaultAlign !== defaultAlign) {
      const sizeRowData = (0, _datagrid_utils.parseSizeLine)(sizeRow, maxColumnCount, defaultAlign);
      this._sizesRowString = sizeRow;
      this._explicitColumnWidths = sizeRowData.explicitColumnWidths;
      this._totalExplicitColumnsWidth = sizeRowData.totalExplicitColumnsWidth;
    }
  } else {
    this._sizesRowString = '';
    this._explicitColumnWidths = null;
    this._totalExplicitColumnsWidth = 0;
  }

  this.defaultAlign = defaultAlign;
  this._maxColumnCount = maxColumnCount; // Compute total native height

  this._totalNativeHeight += this.style("datagrid_padding_top") + this.style("datagrid_padding_bottom"); // Compute native width of each column and total native width

  for (let columnIndex = 0; columnIndex < this._maxColumnCount; columnIndex++) {
    const columnWidth = this.getColumnWidth(columnIndex);
    this._nativeColumnWidths[columnIndex] = columnWidth;
    this._totalNativeWidth += columnWidth;
  }

  if (this.getActualProperty("verticalScrollbar")) {
    this._totalNativeWidth += ScrollBars.Sizes.ScrollbarFixedSize;
  }
};

control.CONTROL_PROTOTYPE.measure = function () {
  const paddingMiddle = this.style("datagrid_padding_middle");
  const paddingLeft = this.style("datagrid_padding_left");
  const paddingRight = this.style("datagrid_padding_right");
  const width = this._totalNativeWidth + paddingLeft + (this._maxColumnCount - 1) * paddingMiddle + paddingRight;
  const height = this._totalNativeHeight;
  return _types.default.newSize(width, height);
};

control.CONTROL_PROTOTYPE.getRowHeights = function (rowHeight, hasHeader) {
  const rowHeights = [];

  for (let rowIndex = 0; rowIndex < this._rows.length; rowIndex++) {
    const nativeRowHeight = this._rows[rowIndex].nativeRowHeight;

    if (hasHeader && rowIndex === 0) {
      rowHeights[rowIndex] = nativeRowHeight;

      if (rowHeights[rowIndex] === 0) {
        rowHeights[rowIndex] = control.MIN_ROW_HEIGHT;
      }
    } else {
      rowHeights[rowIndex] = Math.max(rowHeight, nativeRowHeight);
    }
  }

  return rowHeights;
};

control.CONTROL_PROTOTYPE.drawRow = function (localBbox, y, rowHeight, color) {
  const height = Math.min(rowHeight, localBbox.height - y);
  const displayItem = Rectangle.makeFilledRectangle(_types.default.newRectangle(0, y, localBbox.width, height), color);
  this.pushToDisplayList(displayItem);
};

control.CONTROL_PROTOTYPE.drawHorizontalLine = function (localBbox, y, ruleColor) {
  const {
    height
  } = this.styleSVGSize("rule_horizontal");

  const bounds = _types.default.newRectangle(0, y - height / 2, localBbox.width, height);

  const displayItem = _displayitem.default.createDisplayItemsFromSkin("rule_horizontal", ruleColor, bounds, this._context);

  this.pushToDisplayList(displayItem);
};

control.CONTROL_PROTOTYPE.getColumnWidths = function (totalWidth) {
  const columnWidths = [];

  if (this._totalExplicitColumnsWidth === 0) {
    // The last row is a regular row
    for (let columnIndex = 0; columnIndex < this._maxColumnCount; columnIndex++) {
      columnWidths[columnIndex] = Math.ceil(totalWidth * this._nativeColumnWidths[columnIndex] / this._totalNativeWidth);

      _balsamiq.default.assert(columnWidths[columnIndex] > 0);
    }
  } else {
    // The last row contains column widths
    let leftOverWidth = totalWidth;
    const totalExplicitWidthToUseForCalculating = this._totalExplicitColumnsWidth;
    const columnsAtNativeWidth = {}; // Remove all columns that will need to be sized at their native width because they don't have an explicit width

    for (let columnIndex = 0; columnIndex < this._maxColumnCount; columnIndex++) {
      if (this._explicitColumnWidths[columnIndex].width === 0) {
        columnWidths[columnIndex] = this._nativeColumnWidths[columnIndex];
        leftOverWidth -= columnWidths[columnIndex];
        leftOverWidth = Math.max(0, leftOverWidth);
        columnsAtNativeWidth[columnIndex] = true;

        _balsamiq.default.assert(columnWidths[columnIndex] > 0);
      }
    } // Now calculate the rest based on percentages


    for (let columnIndex = 0; columnIndex < this._maxColumnCount; columnIndex++) {
      if (!columnsAtNativeWidth[columnIndex]) {
        columnWidths[columnIndex] = Math.round(this._explicitColumnWidths[columnIndex].width / totalExplicitWidthToUseForCalculating * leftOverWidth);

        _balsamiq.default.assert(columnWidths[columnIndex] >= 0);
      }
    }
  }

  return columnWidths;
};

control.CONTROL_PROTOTYPE.updateLayout = function (localBbox) {
  // Background
  this.pushToDisplayList(Rectangle.makeFilledRectangle(localBbox, this.getActualProperty("rowColor")));
  const hasBorder = this.getActualProperty("borderStyle") === Props.BorderStyles.SQUARE;
  const paddingMiddle = this.style("datagrid_padding_middle");
  const paddingTop = hasBorder ? this.style("datagrid_padding_top") : 0;
  const paddingLeft = this.style("datagrid_padding_left");
  const paddingRight = this.style("datagrid_padding_right");
  const rowHeight = Math.min(control.MAX_ROW_HEIGHT, Math.max(control.MIN_ROW_HEIGHT, this.getActualProperty("rowHeight")));
  const hasHeader = this.getActualProperty("hasHeader");
  const defaultAlign = this.getActualProperty("align");
  const rowHeights = this.getRowHeights(rowHeight, hasHeader);
  let y = hasBorder ? paddingTop : 0;
  const top = (hasBorder ? paddingTop : 0) + (hasHeader ? rowHeights[0] : 0); // Draw the rows that are visible

  for (let rowIndex = 0; rowIndex < this._rows.length && y < localBbox.height; rowIndex++) {
    this.drawRow(localBbox, y, rowHeights[rowIndex], this._rows[rowIndex].color);
    y += rowHeights[rowIndex];
  } // Fill remaining space with empty rows


  if (this._rows.length % 2 === 0) {
    y += rowHeight;
  }

  const alternateRowColor = this.getActualProperty("alternateRowColor");

  while (y < localBbox.height) {
    this.drawRow(localBbox, y, rowHeight, alternateRowColor);
    y += rowHeight * 2;
  } // Horizontal lines


  const ruleColor = this.style("default_border_color");

  if (this.getActualProperty("hLines")) {
    y = top + (hasHeader ? 0 : rowHeights[0]);

    for (let rowIndex = 1; rowIndex < this._rows.length; rowIndex++) {
      if (y < localBbox.height) {
        this.drawHorizontalLine(localBbox, y, ruleColor);
      }

      y += rowHeights[rowIndex];
    }

    while (y < localBbox.height) {
      this.drawHorizontalLine(localBbox, y, ruleColor);
      y += rowHeight;
    }
  } // Compute real column widths


  const scrollbarWidth = this.getActualProperty("verticalScrollbar") ? ScrollBars.Sizes.ScrollbarFixedSize - 2 : 0;
  const totalWidth = Math.max(localBbox.width - paddingLeft - paddingRight - scrollbarWidth, 1); // We want to draw something in here, there must be at least 1px!

  const columnWidths = this.getColumnWidths(totalWidth);
  let x = 0; // Vertical lines

  if (this.getActualProperty("vLines")) {
    if (this._maxColumnCount > 0) {
      const path = Path.newPath(localBbox.width, localBbox.height);
      x = paddingLeft + columnWidths[0] - paddingMiddle / 2 + 2;
      const lineWidth = this.style("datagrid_vertical_separator_width");

      for (let columnIndex = 1; columnIndex < this._maxColumnCount; columnIndex++) {
        if (x < localBbox.width) {
          path.moveTo(x - lineWidth / 2, 0);
          path.lineTo(x - lineWidth / 2, localBbox.height);
        }

        x += columnWidths[columnIndex];
      }

      this.pushToDisplayList(_displayitem.default.newSVGPath(localBbox, path, lineWidth, ruleColor));
    }
  } // Draw cells


  let columnWidthForCell = 0;
  y = paddingTop;

  for (const [rowIndex, row] of this._rows.entries()) {
    x = paddingLeft; // Render the cells and skip separator cells.

    let columnIndex = 0;

    for (const cell of row.cells) {
      switch (cell.type) {
        case _datagrid_utils.CellType.CHOICE:
        case _datagrid_utils.CellType.TEXT:
          {
            const align = this._explicitColumnWidths ? this._explicitColumnWidths[cell.columnIndex].align : defaultAlign;
            const padding = columnIndex === this._maxColumnCount - 1 || cell.columnIndex === 0 ? paddingMiddle / 2 : paddingMiddle;
            columnWidthForCell = Math.max(0, columnWidths[cell.columnIndex] - padding);

            _balsamiq.default.assert(columnWidthForCell >= 0); // sorting icon


            if (cell.sortingIcon) {
              if (columnIndex === this._maxColumnCount - 1 && hasHeader && rowIndex === 0) {
                columnWidthForCell += scrollbarWidth;
              }

              const iconData = _icon.default.newAppIconData(cell.sortingIcon, _datagrid_utils.SortingIconSize, 0);

              this.pushToDisplayList(_icon.default.makeDisplayItemsFromIconData(this, iconData, x + columnWidthForCell - _datagrid_utils.SortingIconSize - 2, y + (rowHeights[rowIndex] - _datagrid_utils.SortingIconSize) / 2, row.textStyle.textColor));
              columnWidthForCell -= _datagrid_utils.SortingIconSize + 2;
            }

            if (columnWidthForCell > 0) {
              if (cell.type === _datagrid_utils.CellType.TEXT) {
                const textBounds = _types.default.newRectangle(x, y, columnWidthForCell, rowHeights[rowIndex]);

                const {
                  displayItems,
                  links
                } = cell.text.makeDisplayItemsAndLinks(textBounds, align, _types.default.Positions.CENTER);
                this.pushToDisplayList(displayItems);

                _link.default.processTextLinks(this, links);
              } else if (cell.type === _datagrid_utils.CellType.CHOICE) {
                const choiceSize = cell.choice.measure();
                let choiceX = x;

                if (align === _types.default.TextAlign.CENTER) {
                  choiceX += (columnWidthForCell - _datagrid_utils.ChoiceFixedWidth) / 2;
                } else if (align === _types.default.TextAlign.RIGHT) {
                  choiceX += columnWidthForCell - _datagrid_utils.ChoiceFixedWidth;
                }

                choiceX = Math.round(choiceX);
                this.pushToDisplayList(cell.choice.makeDisplayItems(this._context, _types.default.newRectangle(choiceX, y + (rowHeights[rowIndex] - choiceSize.height) / 2, choiceSize.width, choiceSize.height)));
              }
            }

            break;
          }

        case _datagrid_utils.CellType.SEPARATOR:
          x += columnWidths[columnIndex];
          columnIndex++;
          break;
      }
    }

    y += rowHeights[rowIndex];
  } // Scrollbar


  if (this.getActualProperty("verticalScrollbar")) {
    y = (hasHeader ? rowHeights[0] : 0) + (hasBorder ? paddingTop : 0);
    const height = localBbox.height - y - (hasBorder ? this.style("datagrid_scrollbar_padding_bottom") : 0);

    if (height > 0) {
      this.pushToDisplayList(ScrollBars.makeDisplayItemsForVerticalScrollbar(localBbox.width - ScrollBars.Sizes.ScrollbarFixedSize - (hasBorder ? this.style("datagrid_scrollbar_padding_right") : 0), y, height, this.getActualProperty("value"), _types.default.ControlState.NORMAL, this._context));
    }
  } // border


  if (this.getActualProperty("borderStyle") === Props.BorderStyles.SQUARE) {
    this.pushToDisplayList(Rectangle.makeRectangleBorderDisplayItems(this._context, localBbox));
  }
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/datagrid_utils.js":
/*!********************************************!*\
  !*** ./js/core/controls/datagrid_utils.js ***!
  \********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.SortingIconSize = exports.SortingIcon = exports.ChoiceFixedWidth = exports.CellType = void 0;
exports.cellsToString = cellsToString;
exports.isSizeLineRegExp = void 0;
exports.parseRow = parseRow;
exports.parseSizeLine = parseSizeLine;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _choice = __webpack_require__(/*! core/bricks/choice */ "./js/core/bricks/choice.js");

var _textutilities = __webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js");

var DataGridGrammar = _interopRequireWildcard(__webpack_require__(/*! core/texts/datagridGrammar */ "./js/core/texts/datagridGrammar.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const CellType = {
  CHOICE: "choice",
  SEPARATOR: "separator",
  TEXT: "text"
};
exports.CellType = CellType;
const SortingIcon = {
  UP_DOWN: "sort",
  UP: "caret-up",
  DOWN: "caret-down"
};
exports.SortingIcon = SortingIcon;
const SortingIconSize = 15;
exports.SortingIconSize = SortingIconSize;
const ChoiceFixedWidth = 12;
exports.ChoiceFixedWidth = ChoiceFixedWidth;
const isSizeLineRegExp = new RegExp("^{([clrRCL,0-9\\s])+}");
exports.isSizeLineRegExp = isSizeLineRegExp;

function parseSizeLine(p_line, p_maxCount, p_defaultAlign) {
  const result = {};
  result.sizesRowString = p_line;
  result.explicitColumnWidths = [];
  result.totalExplicitColumnsWidth = 0;
  let m;
  const sizes = p_line.replace(/[{} ]/g, "").split(/[,\t]/);
  sizes.forEach(size => {
    m = size.match(/(\d*)([RCLrcl]?)/);
    m[1] = parseInt(m[1], 10);

    if (isNaN(m[1])) {
      m[1] = 0;
    }

    result.totalExplicitColumnsWidth += m[1];
    m[2] = m[2].toLowerCase();
    let align = p_defaultAlign;

    if (m[2] === 'l') {
      align = _types.default.TextAlign.LEFT;
    } else if (m[2] === "r") {
      align = _types.default.TextAlign.RIGHT;
    } else if (m[2] === "c") {
      align = _types.default.TextAlign.CENTER;
    }

    result.explicitColumnWidths.push({
      width: m[1],
      align: align
    });
  }); // Fill missing widths

  for (let i = result.explicitColumnWidths.length; i < p_maxCount; i += 1) {
    result.explicitColumnWidths.push({
      width: 0,
      align: p_defaultAlign
    });
  }

  return result;
}

function parseRow(row, rowIndex, hasHeader) {
  const rowData = {
    nativeRowHeight: 0,
    cells: []
  };
  rowData.rowString = row;
  rowData.rowIndex = rowIndex;
  rowData.totalExplicitColumnsWidth = 0;
  rowData.columnCount = 1;
  let selectableLabel = "";
  const cells = DataGridGrammar.parse(row);
  cells.forEach(cell => {
    let tempText = cell.value;
    const cellData = {
      type: cell.type,
      nativeCellWidth: 0,
      value: tempText
    };

    if (cell.type === CellType.SEPARATOR) {
      rowData.cells.push(cellData);
      selectableLabel += ", ";
      rowData.columnCount += 1;
    } else {
      const tempString = "" + tempText; // the first row can have sorting icons

      if (rowIndex === 0) {
        const last2 = tempString.substr(tempString.length - 2).toLowerCase();
        const last3 = tempString.substr(tempString.length - 3).toLowerCase();

        if (last3 === " v^" || last3 === " ^v") {
          cellData.sortingIcon = SortingIcon.UP_DOWN;
          tempText = tempString.substr(0, tempString.length - 3);
          cellData.sortingIconText = last3;
        } else if (last2 === " v" || last2 === " ^") {
          cellData.sortingIcon = last2 === " v" ? SortingIcon.DOWN : SortingIcon.UP;
          tempText = tempString.substr(0, tempString.length - 2);
          cellData.sortingIconText = last2;
        }

        if (cellData.sortingIcon) {
          cellData.nativeCellWidth += SortingIconSize + 2;
        }
      } // it can be a choice: but only if the text is empty


      const choiceData = (0, _choice.parseChoiceText)(tempString);

      if (choiceData && choiceData.text === "") {
        cellData.type = CellType.CHOICE;
        cellData.choiceData = choiceData;
        cellData.value = cell.value;
        selectableLabel += (0, _textutilities.unformatText)("" + cellData.value);
      } else {
        cellData.type = CellType.TEXT;
        selectableLabel += (0, _textutilities.unformatText)("" + tempText);
        cellData.value = ["" + tempText];
      }

      cellData.columnIndex = rowData.columnCount - 1;

      _balsamiq.default.assert(cellData.nativeCellWidth >= 0);

      rowData.cells.push(cellData);
    }
  });

  if (!(rowIndex === 0 && hasHeader)) {
    rowData.selectableLabels = selectableLabel;
  }

  return rowData;
}

function cellsToString(cells) {
  let resultString = "";
  cells.forEach(cell => {
    let tempText = (0, _textutilities.nodesToString)(cell.value);

    if (cell.sortingIconText) {
      tempText += cell.sortingIconText;
    }

    if (cell.type !== CellType.SEPARATOR) {
      tempText = tempText.replace(/,/g, "\\,");
      tempText = tempText.replace(/\t/g, "\\\t");
    }

    resultString += tempText;
  });
  return resultString;
}

/***/ }),

/***/ "./js/core/controls/datechooser.js":
/*!*****************************************!*\
  !*** ./js/core/controls/datechooser.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::DateChooser",
  controlTypeName: "Date Chooser",
  controlAlternateNames: ["Date Picker"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.ONLY_WIDTH,
        resize: Props.ResizeOptions.HORIZONTAL
      }
    }),
    state: {
      defaultValue: Props.SimpleStates.NORMAL,
      enum: Props.SimpleStates
    },
    ...Props.PropertyDefinitions.TEXT({
      defaultText: "  /  /    ",
      multiline: false
    })
  }
});
controlsMod.addColorPropertyDefinitions(control, {
  defaultBorderColor: "default_border_color",
  oldBorderColorName: 'color'
});
controlsMod.addStandardTextPropertyDefinitions(control);
var BORDER_PADDING_RIGHT = 35;

control.CONTROL_PROTOTYPE.initData = function () {
  this._clipOnBbox = false;
};

control.CONTROL_PROTOTYPE.updateData = function () {
  const textColor = this.getActualProperty('state') !== _types.default.ControlState.DISABLED ? this.style('default_text_color') : this.style('default_disabled_color');
  const textStyle = this.newStandardTextStyle({
    textColor
  });
  const text = this.getActualProperty('text');

  if (!this.text || !(0, _text.areTextsEqual)(this.text, text, textStyle)) {
    this.text = (0, _text.newSingleLineText)(text, textStyle, this._context);
  }
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var textWidth = this.style("datepicker_padding_left") + this.text.measuredSize().width + this.style("datepicker_padding_right");
  var width = p_sizeProperty.width === _types.default.AUTOSIZE ? Math.max(textWidth + BORDER_PADDING_RIGHT, 90) : p_sizeProperty.width;
  var height = Math.max(this.text.measuredSize().height + this.style("datepicker_padding_top") + this.style("datepicker_padding_bottom"), this.style("datepicker_icon_size"));
  return _types.default.newSize(width, height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var iconSize = this.style("datepicker_icon_size");
  var state = this.getActualProperty('state');
  var borderColor = this.getActualProperty("borderColor");

  if (state === _types.default.ControlState.FOCUSED) {
    borderColor = this.style("default_selection_color");
  } else if (state === _types.default.ControlState.DISABLED) {
    borderColor = this.style("datepicker_disabled_color");
  }

  var background;
  var text;
  var borderWidth = p_localBbox.width - this.style("datepicker_border_padding_right");

  if (borderWidth > 0) {
    var borderBounds = _types.default.newRectangle(0, 0, borderWidth, p_localBbox.height);

    background = (0, _rectangle.makeRectangleDisplayItems)(this._context, borderBounds, Props.BorderStyles.SQUARE, 0xFFFFFF, 1, borderColor);
    var verticalPadding = p_localBbox.height / 2 - this.text.measuredSize().height / 2 + (this.getActualProperty('fontSize') - 8) / 12;
    text = this.text.makeDisplayItems(_types.default.paddedRectangle(borderBounds, verticalPadding, 0, this.style("datepicker_padding_left"), 0));
  }

  var iconColor = this.getActualProperty('state') !== _types.default.ControlState.DISABLED ? this.style('default_text_color') : this.style('datepicker_disabled_color');

  var iconData = _icon.default.newAppIconData("calendar", iconSize, 0);

  var icon = _icon.default.makeDisplayItemsFromIconData(this, iconData, p_localBbox.width - iconSize, (p_localBbox.height - iconSize) / 2, iconColor);

  this.pushToDisplayList(background, text, icon);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/datepicker.js":
/*!****************************************!*\
  !*** ./js/core/controls/datepicker.js ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::DatePicker",
  controlTypeName: "Date Picker",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(320, 473),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    datePickerType: {
      defaultValue: Props.DatePickerType.MONTH,
      enum: Props.DatePickerType,
      label: "Date Picker Type"
    },
    ...Props.PropertyDefinitions.links()
  }
});
const DAYS = ["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
const MONTHS = ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];

let getNumberOfDays = function (p_year, p_month) {
  let day = new Date(p_year, p_month + 1, 0);
  return day.getDate();
};

control.CONTROL_PROTOTYPE.getLinkNames = function () {
  return ["CANCEL", "OK"];
};

control.CONTROL_PROTOTYPE.yearsView = function (p_bbox, p_zoom, p_year, p_yearQuantity) {
  let yearTextStyle = this.newTextStyle();
  let yearBbox; //Using a fake label to calculate how much vertical padding we can have between the actual labels

  yearTextStyle.fontSize = 18 * p_zoom;
  let yearLabel = Text.newText("TEST", yearTextStyle, this._context);
  let labelMinSpace = p_yearQuantity * 2 * yearLabel.measuredSize().height;
  yearLabel.textStyle.fontSize = 32 * p_zoom;
  labelMinSpace += yearLabel.measuredSize().height;
  let labelPadding = (p_bbox.height - labelMinSpace) / (2 + 2 * p_yearQuantity); //placing the labels

  if (labelPadding > 0) {
    let theY = labelPadding;

    for (var i = -p_yearQuantity; i <= p_yearQuantity; i++) {
      yearTextStyle.fontSize = (i === 0 ? 32 : 18) * p_zoom;
      yearLabel = Text.newText((p_year + i).toString(), yearTextStyle, this._context);
      yearBbox = _types.default.newRectangle(p_bbox.x + (p_bbox.width - yearLabel.measuredSize().width) / 2, p_bbox.y + theY, yearLabel.measuredSize().width, yearLabel.measuredSize().height);
      this.pushToDisplayList(yearLabel.makeDisplayItems(yearBbox));
      theY += yearLabel.measuredSize().height + labelPadding;
    }
  }
};

control.CONTROL_PROTOTYPE.generateDaysLabels = function (p_date, p_zoom) {
  let labels = [];
  let textColor = this.style("datepicker_color");
  let fontSize = Math.round(this.style("datepicker_fontSize") * p_zoom);

  if (fontSize > 0) {
    let weekDays = ["S", "M", "T", "W", "T", "F", "S"];
    let textStyle = this.newTextStyle();
    textStyle.fontSize = fontSize;
    textStyle.textColor = textColor;
    let text; //creating the days of the week labels

    for (var i = 0; i < 7; i++) {
      text = Text.newText(weekDays[i], textStyle, this._context);
      labels.push(text);
    }

    let firstDay = new Date(p_date.getFullYear(), p_date.getMonth(), 1);
    let daysOfMonth = getNumberOfDays(p_date.getFullYear(), p_date.getMonth());
    let daysToSkip = firstDay.getDay(); //creating empty labels as placeholder for the days of the previous month

    for (i = daysToSkip; i > 0; i--) {
      text = Text.newText("", textStyle, this._context);
      labels.push(text);
    }

    this.today = 6 + daysToSkip;
    let todayTextStyle = this.newTextStyle();
    todayTextStyle.textColor = _colors.default.labelColorGivenBackgroundColor(this.style("default_selection_color"), null, textColor);
    todayTextStyle.bold = true;
    todayTextStyle.fontSize = fontSize; //generating the number of the day labels

    for (i = 1; i <= daysOfMonth; i++) {
      if (i === p_date.getDate()) {
        this.today += i;
      }

      text = Text.newText(i.toString(), i === p_date.getDate() ? todayTextStyle : textStyle, this._context);
      labels.push(text);
    }
  }

  return labels;
};

control.CONTROL_PROTOTYPE.monthView = function (p_bbox, p_zoom, p_date) {
  let padding = 5 * p_zoom;
  let fontSize = this.style("datepicker_fontSize") * p_zoom;
  let daysLabels = this.generateDaysLabels(p_date, p_zoom);

  if (p_bbox.height > 0 && p_bbox.width > 0 && fontSize > 0 && daysLabels.length > 0) {
    let iconSize = this.style("datepicker_iconSize") * p_zoom;
    let iconColor = this.style("datepicker_color");
    let monthTextStyle = this.newTextStyle();
    monthTextStyle.fontSize = fontSize; //placing the month and the < > icons

    let month = Text.newText(MONTHS[p_date.getMonth()].toUpperCase() + " " + p_date.getFullYear().toString(), monthTextStyle, this._context);

    let leftIcon = _icon.default.newAppIconData("angle-left", iconSize, 0);

    let rigthIcon = _icon.default.newAppIconData("angle-right", iconSize, 0);

    let theY = Math.max(iconSize + 2, month.measuredSize().height + 2 * padding);

    let textBbox = _types.default.newRectangle(p_bbox.x + (p_bbox.width - month.measuredSize().width) / 2, p_bbox.y + (theY - month.measuredSize().height) / 2, month.measuredSize().width, month.measuredSize().height);

    this.pushToDisplayList(month.makeDisplayItems(textBbox), _icon.default.makeDisplayItemsFromIconData(this, leftIcon, p_bbox.x, p_bbox.y + (theY - iconSize) / 2, iconColor), _icon.default.makeDisplayItemsFromIconData(this, rigthIcon, p_bbox.x + p_bbox.width - iconSize, p_bbox.y + (theY - iconSize) / 2, iconColor)); //placing the days labels

    let daysHeight = daysLabels[0].measuredSize().height * 1.5;
    let space = (p_bbox.height - theY - daysHeight - 6 * (daysLabels[0].measuredSize().height - 3)) / 6 - fontSize * 0.4;
    let rectWidth = (p_bbox.width - 2) / (p_bbox.height > p_bbox.width ? 6.525 : 6.4);
    let bbox;

    for (let i = 0; i < daysLabels.length; i++) {
      let day = daysLabels[i];
      bbox = _types.default.newRectangle(0, 0, day.measuredSize().width, day.measuredSize().height);
      bbox.x = p_bbox.x + rectWidth * (i % 7) + 1.5 * padding;

      if (i < 7) {
        bbox.y = p_bbox.y + 10 * p_zoom + theY + daysHeight * 0.3;
      } else {
        bbox.y = p_bbox.y + 25 * p_zoom + theY + daysHeight + (space + day.measuredSize().height) * Math.floor((i - 7) / 7) + fontSize * 0.4;
      }

      if (i === this.today) {
        let diameter = fontSize * 2.4;

        let circleBbox = _types.default.newRectangle(bbox.x + (bbox.width - diameter) / 2, bbox.y + (bbox.height - diameter) / 2, diameter, diameter);

        let circle = Path.newEllipsePath(diameter, diameter, 0);
        this.pushToDisplayList(_displayitem.default.newSVGPath(circleBbox, circle, 0, 0, this.style("default_selection_color")));
      }

      this.pushToDisplayList(day.makeDisplayItems(bbox));
    }
  }
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  let linkBounds = [];
  let padding = this.style("datepicker_paddingDate");
  let borderThickness = this.style("datepicker_lineWeightDate");
  let isPortrait = p_localBbox.width < p_localBbox.height; //background

  let background = (0, _rectangle.makeRectangleDisplayItems)(this._context, p_localBbox, borderThickness, 0xffffff);
  this.pushToDisplayList(background); //ok and cancel buttons

  let textStyle = this.newTextStyle();
  textStyle.fontSize = this.style("datepicker_buttonFontSize");
  textStyle.textColor = this.style("default_link_color");
  textStyle.underline = true;
  let ok = Text.newText("OK", textStyle, this._context);
  let cancelText = Text.newText("CANCEL", textStyle, this._context);
  let theX = p_localBbox.width - ok.measuredSize().width - 2 * padding;
  var buttonHeight = Math.max(ok.measuredSize().height, cancelText.measuredSize().height);
  let buttonBbox;

  if (theX - 2 * padding - cancelText.measuredSize().width > 0) {
    buttonBbox = _types.default.newRectangle(theX - 2 * padding - cancelText.measuredSize().width, p_localBbox.height - (buttonHeight + padding), cancelText.measuredSize().width, buttonHeight);
    this.pushToDisplayList(cancelText.makeDisplayItems(buttonBbox));
    linkBounds.push(_balsamiq.default.clone(buttonBbox));
  }

  if (theX > 0) {
    buttonBbox = _types.default.newRectangle(Math.max(1, theX), p_localBbox.height - (buttonHeight + padding), ok.measuredSize().width, buttonHeight);
    this.pushToDisplayList(ok.makeDisplayItems(buttonBbox));
    linkBounds.push(_balsamiq.default.clone(buttonBbox));
  }

  let zoomRatioHeading, zoomRatioDate;
  let date = new Date();
  let yearTextStyle = this.newTextStyle();
  let dateTextStyle = this.newTextStyle();
  let yearLabel = Text.newText(date.getFullYear().toString(), yearTextStyle, this._context);
  let dateLabel = Text.newText(DAYS[date.getDay()] + (isPortrait ? ", " : "\n") + MONTHS[date.getMonth()].slice(0, 3) + " " + (date.getDate() < 10 ? "0" : "") + date.getDate().toString(), dateTextStyle, this._context);
  let yearBbox, dateBbox;

  if (isPortrait) {
    let defaultHeightPortrait = this.style("datepicker_defaultHeightPortrait");
    let defaultWidthPortrait = this.style("datepicker_defaultWidthPortrait");
    let defaultHeadingPortrait = this.style("datepicker_defaultHeadingPortrait");
    let defaultCalendarWidthPortrait = this.style("datepicker_defaultCalendarWidthPortrait"); //calculating zoom ratio

    zoomRatioHeading = Math.min((p_localBbox.height - defaultHeadingPortrait - (ok.measuredSize().height + padding)) / (defaultHeightPortrait - defaultHeadingPortrait - (ok.measuredSize().height + padding)), p_localBbox.width / defaultWidthPortrait);
    defaultHeadingPortrait = defaultHeadingPortrait * zoomRatioHeading;
    zoomRatioDate = Math.min((p_localBbox.height - defaultHeadingPortrait - (ok.measuredSize().height + padding)) / (defaultHeightPortrait - defaultHeadingPortrait - (ok.measuredSize().height + padding)), p_localBbox.width / defaultWidthPortrait);

    if (p_localBbox.width - 2 * borderThickness > 0 && defaultHeadingPortrait - borderThickness > 0) {
      this.pushToDisplayList((0, _rectangle.makeFilledRectangle)(_types.default.newRectangle(borderThickness, borderThickness, p_localBbox.width - 2 * borderThickness, defaultHeadingPortrait - borderThickness), this.style("datepicker_headingPanelBgColor")));
      let distanceRight = p_localBbox.width - borderThickness - padding - (borderThickness + padding + defaultCalendarWidthPortrait * zoomRatioHeading);
      dateTextStyle.fontSize = this.style("datepicker_dateLabelFontSizePortrait") * zoomRatioHeading;
      yearTextStyle.fontSize = this.style("datepicker_yearLabelFontSizePortrait") * zoomRatioHeading;
      yearBbox = _types.default.newRectangle(borderThickness + padding + distanceRight / 2, 2 * padding * zoomRatioHeading, yearLabel.measuredSize().width, yearLabel.measuredSize().height);
      dateBbox = _types.default.newRectangle(yearBbox.x, yearBbox.y + yearBbox.height, dateLabel.measuredSize().width, dateLabel.measuredSize().height);

      if (this.getActualProperty("datePickerType") === Props.DatePickerType.YEAR) {
        let yearPickerBbox = _types.default.newRectangle(borderThickness, defaultHeadingPortrait + borderThickness, Math.max(p_localBbox.width - 2 * borderThickness, 1), Math.max(p_localBbox.height - defaultHeadingPortrait - 2 * borderThickness - buttonHeight, 1));

        this.yearsView(yearPickerBbox, zoomRatioHeading, date.getFullYear(), 3);
      } else {
        padding *= zoomRatioDate;

        let monthBbox = _types.default.newRectangle(p_localBbox.x + padding, defaultHeadingPortrait + borderThickness + padding, p_localBbox.width - 2 * (padding + borderThickness), p_localBbox.height - 2 * (padding + borderThickness) - padding - buttonHeight - defaultHeadingPortrait);

        this.monthView(monthBbox, zoomRatioDate, date);
      }
    }
  } else {
    let defaultHeightLandscape = this.style("datepicker_defaultHeightLandscape");
    let defaultWidthLandscape = this.style("datepicker_defaultWidthLandscape");
    let defaultHeadingLandscape = this.style("datepicker_defaultHeadingLandscape");
    let defaultCalendarHeightLandscape = this.style("datepicker_defaultCalendarHeightLandscape"); //zoom ratio

    zoomRatioHeading = Math.min((p_localBbox.height - (buttonHeight + padding)) / (defaultHeightLandscape - (buttonHeight + padding)), (p_localBbox.width - defaultHeadingLandscape) / (defaultWidthLandscape - defaultHeadingLandscape));
    defaultHeadingLandscape *= zoomRatioHeading;
    zoomRatioDate = Math.min((p_localBbox.height - (buttonHeight + padding)) / (defaultHeightLandscape - (buttonHeight + padding)), (p_localBbox.width - defaultHeadingLandscape) / (defaultWidthLandscape - defaultHeadingLandscape));

    if (defaultHeadingLandscape - borderThickness > 0 && p_localBbox.height - 2 * borderThickness > 0) {
      //heading
      this.pushToDisplayList((0, _rectangle.makeFilledRectangle)(_types.default.newRectangle(borderThickness, borderThickness, defaultHeadingLandscape - borderThickness, p_localBbox.height - 2 * borderThickness), this.style("datepicker_headingPanelBgColor"))); //heading text bbox

      dateTextStyle.fontSize = this.style("datepicker_dateLabelFontSizeLandscape") * zoomRatioHeading;
      yearTextStyle.fontSize = this.style("datepicker_yearLabelFontSizeLandscape") * zoomRatioHeading;
      let distanceBottom = Math.max(0, p_localBbox.height - (buttonHeight + padding) - (2 * padding + borderThickness + defaultCalendarHeightLandscape * zoomRatioDate));
      yearBbox = _types.default.newRectangle(padding * 2 * zoomRatioHeading + borderThickness, borderThickness + padding * zoomRatioHeading + distanceBottom / 2, yearLabel.measuredSize().width, yearLabel.measuredSize().height);
      dateBbox = _types.default.newRectangle(yearBbox.x, yearBbox.y + yearBbox.height, dateLabel.measuredSize().width, dateLabel.measuredSize().height);

      if (this.getActualProperty("datePickerType") === Props.DatePickerType.YEAR) {
        //drawing the year picker view
        let yearPickerBbox = _types.default.newRectangle(defaultHeadingLandscape + borderThickness, borderThickness, Math.max(p_localBbox.width - (defaultHeadingLandscape + borderThickness), 1) - borderThickness, Math.max(p_localBbox.height - 2 * borderThickness - buttonHeight, 1));

        this.yearsView(yearPickerBbox, zoomRatioHeading, date.getFullYear(), 2);
      } else {
        //drawing the month picker view
        padding *= zoomRatioDate;
        let calendarHeight = p_localBbox.height - 2 * (borderThickness + padding) - buttonHeight - padding;

        let monthBbox = _types.default.newRectangle(p_localBbox.x + defaultHeadingLandscape + borderThickness + padding, p_localBbox.y + (p_localBbox.height - calendarHeight - buttonHeight) / 2, Math.max(0, p_localBbox.width - defaultHeadingLandscape - 2 * (padding + borderThickness)), calendarHeight);

        this.monthView(monthBbox, zoomRatioDate, date);
      }
    }
  }

  if (yearBbox && dateBbox) {
    this.pushToDisplayList(dateLabel.makeDisplayItems(dateBbox), yearLabel.makeDisplayItems(yearBbox));
  }

  _link.default.processLinks(this, linkBounds);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/fieldset.js":
/*!**************************************!*\
  !*** ./js/core/controls/fieldset.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _snapping = __webpack_require__(/*! core/controls/snapping */ "./js/core/controls/snapping.js");

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::FieldSet",
  controlTypeName: "Field Set",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.TEXT({
      defaultText: "Group Name",
      multiline: false
    })
  }
});
controlsMod.addStandardTextPropertyDefinitions(control);
controlsMod.addColorPropertyDefinitions(control, {
  defaultColor: 0xFFFFFF,
  hasAlpha: true
});

control.CONTROL_PROTOTYPE.initData = function () {
  this._topHeight = null;
};

control.CONTROL_PROTOTYPE.updateData = function () {
  const textColor = _colors.default.labelColorGivenBackgroundColor(this.getActualProperty("color"), this.getActualProperty("backgroundAlpha"), this.style("default_text_color"));

  const linkColor = _colors.default.labelColorGivenBackgroundColor(this.getActualProperty("color"), this.getActualProperty("backgroundAlpha"), this.style("default_link_color"));

  const textStyle = this.newStandardTextStyle({
    textColor,
    linkColor
  });
  const text = this.getActualProperty('text');

  if (!this.text || !(0, _text.areTextsEqual)(this.text, text, textStyle)) {
    this.text = (0, _text.newSingleLineText)(text, textStyle, this._context);
  }
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var size;
  size = _types.default.newSize(Math.max(200, this.text.measuredSize().width + this.style("fieldset_label_padding_left") + 15), 170);
  return size;
};

var MASK_PADDING = 3;

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var borderBounds = _types.default.paddedRectangle(p_localBbox, this.text.measuredSize().height / 2);

  var border = (0, _rectangle.makeRectangleDisplayItems)(this._context, borderBounds, Props.BorderStyles.SQUARE, 0, 0);
  var background = (0, _rectangle.makeFilledRectangle)(p_localBbox, this.getActualProperty("color"), this.getActualProperty("backgroundAlpha"));
  var textX = this.style("fieldset_label_padding_left");
  var text = this.text.makeDisplayItems(_types.default.newRectangle(textX, this.style("fieldset_label_padding_top"), this.text.measuredSize().width, this.text.measuredSize().height));

  var maskBounds = _types.default.newRectangle(textX - MASK_PADDING, p_localBbox.y, this.text.measuredSize().width + MASK_PADDING * 2, this.text.measuredSize().height);

  var maskID = this.addMaskWithRectHole(p_localBbox, maskBounds);

  _displayitem.default.maskDisplayItems(border, maskID);

  this._topHeight = borderBounds.y;
  this.pushToDisplayList(background, border, text);
};

control.CONTROL_PROTOTYPE.getSnapLines = function (withMarkup) {
  _balsamiq.default.assertBool(withMarkup);

  var bbox = this.getUserBBox(true);
  return [new _snapping.SnapLine(_snapping.SnapTypes.MIDDLE, _snapping.SnapDirections.VERTICAL, bbox.x + bbox.width / 2), new _snapping.SnapLine(_snapping.SnapTypes.MIDDLE, _snapping.SnapDirections.HORIZONTAL, bbox.y + bbox.height / 2), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.VERTICAL, bbox.x), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.VERTICAL, bbox.x + bbox.width), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.HORIZONTAL, bbox.y), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.HORIZONTAL, bbox.y + bbox.height), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.VERTICAL, bbox.x + _snapping.PROXIMITY_DISTANCE), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.VERTICAL, bbox.x + bbox.width - _snapping.PROXIMITY_DISTANCE), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.HORIZONTAL, bbox.y + _snapping.PROXIMITY_DISTANCE + this._topHeight), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.HORIZONTAL, bbox.y + bbox.height - _snapping.PROXIMITY_DISTANCE)];
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/formattingtoolbar.js":
/*!***********************************************!*\
  !*** ./js/core/controls/formattingtoolbar.js ***!
  \***********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var _combobox = _interopRequireDefault(__webpack_require__(/*! core/bricks/combobox */ "./js/core/bricks/combobox.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const Icons = ["bold", "italic", "underline", "strikethrough", "list-ul", "list-ol", "undo", "repeat", "picture-o", "smile-o"];
const control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::FormattingToolbar",
  controlTypeName: "Toolbar",
  controlAlternateNames: ["Formatting Toolbar", "Rich Text Editor Toolbar"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(261, 29),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.HORIZONTAL
      }
    })
  }
});

control.CONTROL_PROTOTYPE.measure = function (sizeProperty) {
  const width = sizeProperty.width === _types.default.AUTOSIZE ? 261 : sizeProperty.width;
  return _types.default.newSize(width, 29);
};

control.CONTROL_PROTOTYPE.initData = function () {
  this.texts = new Array(Icons.length);
};

control.CONTROL_PROTOTYPE.updateData = function () {
  const text = "style";
  const textStyle = {
    textColor: 0x000000,
    fontSize: 10,
    bold: true,
    fontFamily: this.style('label_font_family'),
    italic: true,
    linkColor: this.style('default_link_color'),
    underline: true
  };

  if (!this.text || !Text.areTextsEqual(this.text, text, textStyle)) {
    this.text = Text.newTextNoWrap(text, textStyle, this._context);
  }
};

control.CONTROL_PROTOTYPE.updateLayout = function (localBbox) {
  const iconColor = this.style("formattingtoolbar_icon_color");
  const iconSize = this.style("formattingtoolbar_icon_size");
  this.pushToDisplayList(Rectangle.makeFilledRectangle(localBbox));
  this.pushToDisplayList(Rectangle.makeStrokedRectangle(localBbox, iconColor, this.style("formattingtoolbar_border")));
  this.icon = "IconBold";
  this.rotation = 0;
  let offset = 3;
  let j = 0;
  let comboBox;

  for (let i = 0; i < Icons.length; i++) {
    if (j === 4) {
      if (offset + (j + 3) * (iconSize + 3) + 2 < localBbox.width) {
        const comboBounds = _types.default.newRectangle(offset + 4 * (iconSize + 3), (localBbox.height - 21) / 2, 3 * iconSize + 6, 21);

        comboBox = _combobox.default.makeComboboxDisplayItems(this, comboBounds, 21, this.text, iconColor, _types.default.ControlState.NORMAL);
      }

      j += 3;
    }

    if ((i + 1 === 7 || i + 1 === 9) && offset + j * (iconSize + 3) + 5 < localBbox.width) {
      const line = _types.default.newRectangle(offset + j * (iconSize + 3), (localBbox.height - iconSize) / 2, 2, iconSize);

      this.pushToDisplayList(Rectangle.makeFilledRectangle(line, iconColor));
      offset += 5;
    }

    if (offset + (j + 1) * (iconSize + 3) + 2 < localBbox.width) {
      const iconData = _icon.default.newAppIconData(Icons[i], iconSize, 0);

      const icon = _icon.default.makeDisplayItemsFromIconData(this, iconData, offset + j * (iconSize + 3), (localBbox.height - iconSize) / 2, iconColor);

      this.pushToDisplayList(icon);
    }

    j++;
  }

  this.pushToDisplayList(comboBox);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/group.js":
/*!***********************************!*\
  !*** ./js/core/controls/group.js ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _frame = _interopRequireDefault(__webpack_require__(/*! core/bricks/frame */ "./js/core/bricks/frame.js"));

var Constants = _interopRequireWildcard(__webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _controls = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const group = (0, _controls.makeControlClass)({
  controlTypeID: "__group__",
  controlTypeName: "Group",
  controlAlternateNames: [],
  props: {
    controlName: {
      defaultValue: ""
    },
    crop: {
      defaultValue: _types.default.Crop.ALL
    },
    ...Props.PropertyDefinitions.link,
    notes: {
      defaultValue: ""
    }
  }
});

group.CONTROL_PROTOTYPE.measure = function () {
  // Union of all user bbox in component space
  let res;

  if (this.children.length) {
    let userBBox = this.calcChildrenUserBBoxCS(false); // TODO Markup?

    res = _types.default.calcCroppedSize(_types.default.newRectangleFromCrop(this.getActualProperty('crop')), {
      width: userBBox.width,
      height: userBBox.height
    });
  } else {
    res = _types.default.newSize(150, 150);
  }

  return res;
};

group.CONTROL_PROTOTYPE.updateClipBBoxPS = function () {
  var crop = _types.default.newRectangleFromCrop(this.getActualProperty('crop'));

  if (_types.default.isCropAll(crop)) {
    // Clip is calculated as an offset to the current user bbox
    // The delta can be obtained calculating children's clip vs user bbox delta
    var childrenClipBBox = this.calcChildrenClipBBoxCS(false); //Component Space: can have negative origin if some controls extend bounds on top or left

    var childrenUserBBox = this.calcChildrenUserBBoxCS(false);

    var delta = _types.default.newPoint(childrenClipBBox.x - childrenUserBBox.x, childrenClipBBox.y - childrenUserBBox.y);

    this.setClipBBoxPS(this._userBBox.x + delta.x, this._userBBox.y + delta.y, childrenClipBBox.width, childrenClipBBox.height);
  } else {
    // IF CROP enabled Set Clipping bbox to user bbox cropped!!
    this.setClipBBoxPS(this._userBBox.x, this._userBBox.y, this._userBBox.width, this._userBBox.height);
  }
};

group.CONTROL_PROTOTYPE.updateUserBBoxNoMarkupPS = function (userBBox) {
  // Get children user bbox in component space (0,0) is in Origin
  var childrenBBoxNoMarkup = this.calcChildrenUserBBoxCS(true);

  if (this.getActualProperty('crop') === _types.default.Crop.ALL) {
    let origin = this.getActualProperty('origin');
    this.setUserBBoxNoMarkupPS(origin.x + childrenBBoxNoMarkup.x, origin.y + childrenBBoxNoMarkup.y, childrenBBoxNoMarkup.width, childrenBBoxNoMarkup.height);
  } else {
    let origin = _types.default.calcUncroppedOrigin(_types.default.newRectangleFromCrop(this.getActualProperty('crop')), userBBox); //REVERSE ORIGIN in UNCROPPED POSITION


    let newRect = _types.default.newRectangle(origin.x + childrenBBoxNoMarkup.x, origin.y + childrenBBoxNoMarkup.y, childrenBBoxNoMarkup.width, childrenBBoxNoMarkup.height);

    let intersect = _types.default.rectanglesIntersect(newRect, userBBox);

    this.setUserBBoxNoMarkupPS(intersect.x, intersect.y, intersect.width, intersect.height);
  }
};

group.CONTROL_PROTOTYPE.updateClipBBoxNoMarkupPS = function (clipBBox) {
  var childrenBBoxNoMarkup = this.calcChildrenClipBBoxCS(true);

  if (this.getActualProperty('crop') === _types.default.Crop.ALL) {
    let origin = this.getActualProperty('origin');
    this.setClipBBoxNoMarkupPS(origin.x + childrenBBoxNoMarkup.x, origin.y + childrenBBoxNoMarkup.y, childrenBBoxNoMarkup.width, childrenBBoxNoMarkup.height);
  } else {
    let origin = _types.default.calcUncroppedOrigin(_types.default.newRectangleFromCrop(this.getActualProperty('crop')), this._userBBox); //REVERSE ORIGIN in UNCROPPED POSITION


    let newRect = _types.default.newRectangle(origin.x + childrenBBoxNoMarkup.x, origin.y + childrenBBoxNoMarkup.y, childrenBBoxNoMarkup.width, childrenBBoxNoMarkup.height);

    let intersect = _types.default.rectanglesIntersect(newRect, clipBBox);

    this.setUserBBoxNoMarkupPS(intersect.x, intersect.y, intersect.width, intersect.height);
  }
};

group.CONTROL_PROTOTYPE.getUncroppedUserBBox = function (includeMarkup) {
  _balsamiq.default.assertBool(includeMarkup);

  var bbox;

  if (includeMarkup) {
    bbox = { ...this._userBBox
    };
  } else {
    bbox = { ...this._userBBoxNoMarkup
    };
  }

  var cropProp = this.getActualProperty("crop");

  if (cropProp !== _types.default.Crop.ALL) {
    var cropRect = _types.default.newRectangleFromCrop(cropProp);

    bbox = _types.default.calcUncroppedRectangle(cropRect, bbox);
  }

  return bbox;
};

group.CONTROL_PROTOTYPE.initData = function () {
  this._clipOnBbox = false;
  this.setUserBBoxPS(0, 0, 0, 0);
  this.setClipBBoxPS(0, 0, 0, 0);
  this.setUserBBoxNoMarkupPS(0, 0, 0, 0);
  this.setClipBBoxNoMarkupPS(0, 0, 0, 0);
};

group.CONTROL_PROTOTYPE.updateData = function () {//this._clipOnBbox = false;
};

group.CONTROL_PROTOTYPE.onChildBboxChange = function (p_childControlID) {
  this.updateBBoxes();
};

group.CONTROL_PROTOTYPE.onChildOriginChange = function (p_childControlID) {
  // called by a child when its 'origin' property changes
  // usually do nothing, but Groups and some compound controls will recompute the origin of the reference system
  this.updateOrigin();
};

group.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this);

  if (this.children.length === 0 && this.controlID !== -1) {
    this.pushToDisplayList(_frame.default.makeImageFrame(this, p_localBbox, "Empty Group"));
  }
};

group.CONTROL_PROTOTYPE.updateCropOffsetMask = function () {
  _balsamiq.default.assert(arguments.length === 0);

  if (this.hasCropData()) {
    this.setCropMask(_types.default.makeMaskRect(this.localBBox()), true);
  }
};

group.CONTROL_PROTOTYPE.updateCropData = function () {
  _balsamiq.default.assert(arguments.length === 0);

  const cropRect = _types.default.newRectangleFromCrop(this.getActualProperty('crop'));

  if (_types.default.isCropAll(cropRect)) {
    this.resetCropData();
  } else {
    const bbox = this.calcChildrenUserBBoxCS(false);

    const cropOffset = _types.default.calcCropOffset(cropRect, bbox);

    this.setCropOffset(cropOffset);
    this.setUncroppedSize(_types.default.newSize(bbox.width, bbox.height));
    this.updateBBoxes();
    this.setCropMask(_types.default.makeMaskRect(this.localBBox()), true);

    this._context.notifyChange(this.coreObjectID, Constants.ControlChanges.DISPLAY_LIST);

    this.updateCropOffsetMask();
  }
};

group.CONTROL_PROTOTYPE.updateCropOffsetDescendants = function () {
  this.updateCropData();

  function updateChildren(control) {
    for (const child of control.children) {
      if (child.hasPropertyDefinition('crop')) {
        child.updateCropData();
      }

      if (child.children && child.children.length) {
        updateChildren(child);
      }
    }
  }

  ;
  updateChildren(this);
};

group.CONTROL_PROTOTYPE.addChildren = function (p_controls) {
  // add an array of children and update the reference system origin only at the end.
  // this method is needed when loading a file, so that adding the first child doesn't move
  // the reference system invalidating the coords of the second child.
  // the children must have the 'origin' property already set, otherwise setting it later
  // will cause again the problem.
  this.superClass.prototype.addChildren.call(this, p_controls);
  this.updateAll();
};

group.CONTROL_PROTOTYPE.setMockup = function (p_mockup) {
  _balsamiq.default.assert(this.coreObjectID.controlID.length === 0, "this Group control is not the root");

  this.mockup = p_mockup;
};

group.CONTROL_PROTOTYPE.getMockup = function () {
  var p = null;

  if (this.coreObjectID.controlID.length === 0) {
    p = this.mockup;
  } else if (this.parent) {
    p = this.parent.getMockup();
  }

  return p;
};

group.CONTROL_PROTOTYPE.getSelectionBaseColor = function () {
  return 0x3223b3;
};

group.CONTROL_PROTOTYPE.isGroup = function () {
  return true;
};

group.CONTROL_PROTOTYPE.getResizeOption = function () {
  return Props.ResizeOptions.BOTH;
};

group.CONTROL_PROTOTYPE.getName = function () {
  let asSymbol = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;
  let name = this.superClass.prototype.getName.call(this);

  if (!name || name === '') {
    name = asSymbol ? Constants.UnnamedSymbol : Constants.UnnamedGroup;
  }

  return name;
};

group.CONTROL_PROTOTYPE.getResizeOption = function () {
  return this.children.length > 0 ? Props.ResizeOptions.BOTH : Props.ResizeOptions.NO;
};

group.CONTROL_PROTOTYPE.isEmpty = function () {
  return this.children.length === 0;
};

group.CONTROL_PROTOTYPE.getDescription = function () {
  const controlName = this.hasPropertyValue("controlName") ? this.getActualProperty("controlName") : null;
  return controlName || this.superClass.prototype.getDescription.call(this);
};

var _default = group;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/helpbutton.js":
/*!****************************************!*\
  !*** ./js/core/controls/helpbutton.js ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var controlsMod = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::HelpButton",
  controlTypeName: "Help Button",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.link
  }
});

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this);

  this.pushToDisplayList(_displayitem.default.createDisplayItemsFromSkin("helpbutton", this.style("default_border_color"), p_localBbox, this._context));
};

control.CONTROL_PROTOTYPE.measure = function () {
  return _types.default.newSize(18, 18);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/icon.js":
/*!**********************************!*\
  !*** ./js/core/controls/icon.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var icon = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::Icon",
  controlTypeName: "Icon",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.BORDER_STYLE(Props.BorderStyles.NONE, Props.SimpleBorderStyles),
    ...Props.PropertyDefinitions.link,
    ...Props.PropertyDefinitions.ROTATION({
      isIcon: true
    })
  }
});
controlsMod.addColorPropertyDefinitions(icon, {
  defaultColor: "icon_default_color"
});

icon.CONTROL_PROTOTYPE.initData = function () {
  this._clipOnBbox = false;
};

icon.CONTROL_PROTOTYPE.updateData = function () {
  _icon.default.processIconData(this);

  if (!this.iconData) {
    this.iconData = _icon.default.emptyIconData(_types.default.IconSize.large);
  }
};

icon.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  return _types.default.newSize(this.iconData.size, this.iconData.size);
};

icon.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this);

  var borderStyle = this.getActualProperty('borderStyle');
  var padding = 0;
  var border;

  if (borderStyle !== Props.BorderStyles.NONE) {
    padding = this.style("icon_border_padding");
    border = _icon.default.makeDisplayItemsFromIconData(this, _icon.default.emptyIconData(this.iconData.size), 0, 0, this.getActualProperty('color'));
  }

  var iconData = _balsamiq.default.clone(this.iconData);

  iconData.size -= padding * 2;

  var icon = _icon.default.makeDisplayItemsFromIconData(this, iconData, padding, padding, this.getActualProperty("color"));

  this.pushToDisplayList(icon, border);
};

icon.CONTROL_PROTOTYPE.getDescription = function () {
  const icon = this.hasPropertyValue("icon") ? this.getActualProperty("icon") : null;
  let res;

  if (!icon) {
    res = 'Empty Icon';
  } else if (icon.src && icon.src.resourceID) {
    if (icon.src.location === _types.default.SrcLocation.APP_ASSETS) {
      res = icon.src.resourceID;
    } else if (icon.src.location === _types.default.SrcLocation.PROJECT_ASSETS) {
      const asset = this._context.getDependency(icon.src.resourceID);

      res = asset ? asset.attributes.name : null;
    }
  }

  return res || this.superClass.prototype.getDescription.call(this);
};

var _default = icon;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/iconlabel.js":
/*!***************************************!*\
  !*** ./js/core/controls/iconlabel.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _macro = __webpack_require__(/*! core/bricks/macro */ "./js/core/bricks/macro.js");

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::IconLabel",
  controlTypeName: "Icon and Label",
  controlAlternateNames: ["Icon and Text Label"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.ONLY_WIDTH,
        resize: Props.ResizeOptions.HORIZONTAL
      }
    }),
    ...Props.PropertyDefinitions.BORDER_STYLE(Props.SimpleBorderStyles.NONE, Props.SimpleBorderStyles),
    labelPosition: {
      defaultValue: Props.IconLabelPositions.BOTTOM,
      enum: Props.IconLabelPositions
    },
    state: {
      defaultValue: Props.BoolStates.NORMAL,
      enum: Props.BoolStates
    },
    ...Props.PropertyDefinitions.link,
    ...Props.PropertyDefinitions.ROTATION({
      isIcon: true
    })
  }
});
controlsMod.addStandardTextPropertyDefinitions(control, {
  text: "Icon Name",
  charCountEnabled: true
});
controlsMod.addColorPropertyDefinitions(control, {
  defaultColor: "default_text_color"
});

control.CONTROL_PROTOTYPE.initData = function () {
  this._clipOnBbox = false;
};

control.CONTROL_PROTOTYPE.getColor = function () {
  return this.getActualProperty('state') === _types.default.ControlState.NORMAL ? this.getActualProperty('color') : this.style('default_disabled_color');
};

control.CONTROL_PROTOTYPE.updateData = function () {
  const text = (0, _macro.processMacros)(this);
  const textColor = this.getColor();
  const textStyle = this.newStandardTextStyle({
    textColor
  });

  if (!this.text || !(0, _text.areTextsEqual)(this.text, text, textStyle)) {
    this.text = (0, _text.newSingleLineText)(text, textStyle, this._context);
  }

  _icon.default.processIconData(this);

  if (!this.iconData) {
    this.iconData = _icon.default.emptyIconData(_types.default.IconSize.large);
  }
};

_link.default.addTextLinksMethods(control);

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  const iconSize = this.iconData.size;
  const textSize = this.text.measuredSize();

  let size = _types.default.newSize(0, 0);

  if (this.getActualProperty('labelPosition') === _types.default.Positions.BOTTOM) {
    size.width = p_sizeProperty.width === _types.default.AUTOSIZE ? Math.max(iconSize, textSize.width) : p_sizeProperty.width;
    size.height = iconSize + textSize.height + this.style('icon_label_padding_bottom');
  } else {
    // LEFT or RIGHT
    size.width = p_sizeProperty.width === _types.default.AUTOSIZE ? iconSize + textSize.width + this.style('icon_label_padding') : p_sizeProperty.width;
    let naturalHeight = textSize.height + this.style('label_padding_top') + this.style('label_padding_bottom');
    size.height = Math.max(iconSize, naturalHeight);
  }

  return size;
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this);

  const labelPosition = this.getActualProperty('labelPosition');
  let iconX, iconY, labelX, labelY;
  const textSize = this.text.measuredSize();

  let textBbox = _types.default.newRectangle(0, 0, 1, 1);

  let align = "left";

  if (labelPosition === _types.default.Positions.BOTTOM) {
    iconX = (p_localBbox.width - this.iconData.size) / 2;
    iconY = 0;
    labelY = this.iconData.size + this.style("icon_label_padding_bottom");
    textBbox.width = p_localBbox.width;
    textBbox.height = textSize.height;
    align = "center";
  } else if (labelPosition === _types.default.Positions.LEFT) {
    iconX = p_localBbox.width - this.iconData.size; // right aligned

    iconY = (p_localBbox.height - this.iconData.size) / 2;
    labelY = (p_localBbox.height - textSize.height) / 2;
    textBbox.x = 0;
    textBbox.width = Math.min(textSize.width, p_localBbox.width - this.iconData.size - this.style("icon_label_padding"));
  } else {
    // RIGHT
    iconX = 0;
    iconY = (p_localBbox.height - this.iconData.size) / 2;
    labelY = (p_localBbox.height - textSize.height) / 2;
    textBbox.x = this.iconData.size + this.style("icon_label_padding");
    textBbox.width = p_localBbox.width - textBbox.x;
  }

  textBbox.y = labelY;
  textBbox.height = textSize.height;
  let label = null;

  if (textBbox.width > 0) {
    label = this.text.makeDisplayItemsAndLinks(textBbox, align);

    _link.default.processTextLinks(this, label.links);
  }

  const borderStyle = this.getActualProperty('borderStyle');
  let borderPadding = 0;
  let border;

  if (borderStyle !== Props.BorderStyles.NONE && this.iconData.size + iconX <= p_localBbox.width && iconX >= 0) {
    borderPadding = this.style("icon_border_padding");
    border = _icon.default.makeDisplayItemsFromIconData(this, _icon.default.emptyIconData(this.iconData.size), iconX, iconY, this.getColor());
  }

  let iconData = _balsamiq.default.clone(this.iconData);

  iconData.size -= borderPadding * 2;
  let icon = null;

  if (iconX >= 0 && iconX + borderPadding + iconData.size <= p_localBbox.width) {
    icon = _icon.default.makeDisplayItemsFromIconData(this, iconData, iconX + borderPadding, iconY + borderPadding, this.getColor());
  }

  let labelDisplayItems = label ? label.displayItems : null;
  this.pushToDisplayList(labelDisplayItems, icon, border);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/image.js":
/*!***********************************!*\
  !*** ./js/core/controls/image.js ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var Properties = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var _frame = _interopRequireDefault(__webpack_require__(/*! core/bricks/frame */ "./js/core/bricks/frame.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _overlaylabel = __webpack_require__(/*! core/bricks/overlaylabel */ "./js/core/bricks/overlaylabel.js");

var Image = _interopRequireWildcard(__webpack_require__(/*! core/model/image */ "./js/core/model/image.js"));

var _displaynode = __webpack_require__(/*! core/displaynode */ "./js/core/displaynode.js");

var MockupCoords = _interopRequireWildcard(__webpack_require__(/*! core/viewcontrollers/mockupcoordinates */ "./js/core/viewcontrollers/mockupcoordinates.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Image Control Class
 * @extends {import("core/controls/controls").Control}
 */
const image = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::Image",
  controlTypeName: "Image",
  controlAlternateNames: ["Img"],
  props: { ...Properties.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Properties.AutoSizeOptions.WIDTH_AND_OR_HEIGHT,
        resize: Properties.ResizeOptions.BOTH
      }
    }),
    ...Properties.PropertyDefinitions.BORDER_STYLE(Properties.BorderStyles.NONE, Properties.SimpleBorderStyles),
    filter: {
      dataType: 'bool',
      defaultValue: false
    },
    src: {
      defaultValue: null,
      dataType: 'object',
      setter: value => {
        // we need to set the src to undefined when the value is null to allow
        // the control.setProperty method to remove it. This allows the image to be pasted in B3
        return value === null ? undefined : value;
      }
    },
    crop: {
      defaultValue: _types.default.Crop.ALL
    },
    ...Properties.PropertyDefinitions.link,
    ...Properties.PropertyDefinitions.ROTATION({
      isIcon: false
    }),
    ...Properties.PropertyDefinitions.QUICKDRAW_TIP("You can draw Images quickly by holding down the *I key* and dragging. [Learn more]($DOCURL)", "You can also draw Images with your fingers using the Quick Draw Toolbar (in the View Menu). [Learn more]($DOCURL)")
  }
});
Controls.addStandardTextPropertyDefinitions(image);

image.CONTROL_PROTOTYPE.initData = function () {
  this.nativeImageSize = null;
  this.nativeImageID = null;
  this.resourceState = null;
  this._clipOnBbox = false;
  this._offset = _types.default.newPoint(0, 0);
};

image.CONTROL_PROTOTYPE.updateData = function () {
  const src = this.getActualProperty("src");
  this.setDependencies(src ? [src] : []);
  this.nativeImageSize = null;
  this.nativeImageID = null;
  this.resourceState = null;

  if (src) {
    const resource = this._context.getDependency(src.resourceID);

    if (resource && !(resource instanceof Image.Image)) {
      this._context.host.log('Image control: the Resource ' + src.resourceID + ' is not an Image');
    }

    if (resource && resource instanceof Image.Image) {
      this.nativeImageSize = resource.getImageSize();
      this.nativeImageID = resource.getImageID();
      this.resourceState = this.nativeImageID ? _types.default.ResourceState.AVAILABLE : _types.default.ResourceState.PENDING;
    } else {
      this.resourceState = _types.default.ResourceState.NOT_AVAILABLE;
    }

    this._context.setReady(this.coreObjectID, this.resourceState === _types.default.ResourceState.AVAILABLE || this.resourceState === _types.default.ResourceState.NOT_AVAILABLE);
  }
};

image.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  const crop = _types.default.newRectangleFromCrop(this.getActualProperty("crop"));

  return this.nativeImageSize ? _types.default.newSize(this.nativeImageSize.width * crop.width / 1000, this.nativeImageSize.height * crop.height / 1000) : _types.default.newSize(77, 79);
};

image.CONTROL_PROTOTYPE.getUncroppedUserBBox = function (includeMarkup) {
  _balsamiq.default.assertBool(includeMarkup);

  const imageBounds = _balsamiq.default.clone(this.localBBox());

  const cropProp = this.getActualProperty("crop");
  let realImageBounds;

  if (cropProp === _types.default.Crop.ALL) {
    realImageBounds = imageBounds;
  } else {
    const cropRect = _types.default.newRectangleFromCrop(cropProp);

    const strechedImageSize = _types.default.newSize(imageBounds.width * 1000 / cropRect.width, imageBounds.height * 1000 / cropRect.height);

    const offset = _types.default.newPoint(-strechedImageSize.width * cropRect.x / 1000, -strechedImageSize.height * cropRect.y / 1000);

    realImageBounds = _types.default.newRectangle(imageBounds.x + offset.x, imageBounds.y + offset.y, strechedImageSize.width, strechedImageSize.height);
  }

  const origin = this.getActualProperty('origin');
  realImageBounds.x += origin.x;
  realImageBounds.y += origin.y;
  return realImageBounds;
};

image.CONTROL_PROTOTYPE.updateCropData = function () {
  const cropProp = this.getActualProperty("crop");

  const cropRect = _types.default.newRectangleFromCrop(cropProp);

  if (!_types.default.isCropAll(cropRect)) {
    const imageBounds = this.getImageBounds();

    _balsamiq.default.assertRect(imageBounds);

    const cropOffset = _types.default.newPoint(Math.round(-imageBounds.width / cropRect.width * cropRect.x), Math.round(-imageBounds.height / cropRect.height * cropRect.y));

    this.setCropData({
      cropMask: _types.default.makeMaskRect(this.localBBox()),
      cropOffset: cropOffset,
      useCropMask: true
    });
    this.setUncroppedSize(_types.default.newSize(imageBounds.width, imageBounds.height));
  } else {
    this.resetCropData();
  }
};

function newRotatedOrigin(currentOrigin, currentSize) {
  _balsamiq.default.assertPoint(currentOrigin);

  _balsamiq.default.assertSize(currentSize);

  const center = {
    x: currentOrigin.x + Math.round(currentSize.width / 2),
    y: currentOrigin.y + Math.round(currentSize.height / 2)
  };
  const topRight = {
    x: currentOrigin.x + currentSize.width,
    y: currentOrigin.y
  };
  const x1 = topRight.x - center.x;
  const y1 = topRight.y - center.y;
  return _types.default.newPoint(y1 + center.x, -x1 + center.y);
}

function newCropArea(actualCrop, actualRotation, newRotation) {
  const [x, y, w, h] = actualCrop.split(",");
  const rotation = newRotation - actualRotation;
  const isRotationClockwise = rotation === 90 || rotation === -270;
  return isRotationClockwise ? [1000 - y - h, x, h, w].join(",") : [y, 1000 - x - w, h, w].join(",");
}
/**
 * Checks which other properties need to be changed due to a change in a specific one
 *
 * @param {string} name
 * @param {unknown} value
 * @returns {{[propName: string]: unknown}}
 */


image.CONTROL_PROTOTYPE.requiredChangesForProperty = function (name, value) {
  const changes = {
    [name]: value
  };

  switch (name) {
    case "rotation":
      {
        const actualSize = this.getActualProperty("size");

        const isDefaultSize = _balsamiq.default.isEqual(actualSize, this.getDefaultProperty("size"));

        const size = isDefaultSize ? this.nativeImageSize : actualSize;

        const newSize = _types.default.newSize(size.height, size.width);

        changes["size"] = newSize;
        const origin = this.getActualProperty("origin");
        changes["origin"] = newRotatedOrigin(origin, size);

        if (this.hasCropData()) {
          changes["crop"] = newCropArea(this.getActualProperty("crop"), this.getActualProperty("rotation"), value);
        }

        break;
      }

    case "size":
      {
        const isDefaultSize = _balsamiq.default.isEqual(value, this.getDefaultProperty("size"));

        if (isDefaultSize) {
          const rotation = this.getActualProperty("rotation");

          if (this.nativeImageSize && (rotation === 90 || rotation === 270)) {
            const newSize = _types.default.newSize(this.nativeImageSize.height, this.nativeImageSize.width);

            changes["size"] = newSize;
          }
        }

        break;
      }

    default:
      break;
  }

  return changes;
};

image.CONTROL_PROTOTYPE._getDisplayItems = function (p_localBbox, forCropPreview, withBorders) {
  const imageBounds = forCropPreview ? _balsamiq.default.clone(this.localBBox()) // otherwise the full image displayed in the crop preview gets distorted
  : this.getImageBounds();
  let cropProp = this.getActualProperty("crop");
  let realImageBounds;
  let maskRect;
  let cropData;
  let uncroppedSize;

  var cropOffset = _types.default.newPoint(0, 0);

  var cropBBox = { ...p_localBbox
  };

  if (cropProp === _types.default.Crop.ALL) {
    realImageBounds = imageBounds;
    this.resetCropData();
  } else {
    var cropRect = _types.default.newRectangleFromCrop(cropProp);

    var strechedImageSize = _types.default.newSize(imageBounds.width * 1000 / cropRect.width, imageBounds.height * 1000 / cropRect.height);

    cropOffset = _types.default.newPoint(-strechedImageSize.width * cropRect.x / 1000, -strechedImageSize.height * cropRect.y / 1000);
    realImageBounds = _types.default.newRectangle(imageBounds.x, imageBounds.y, strechedImageSize.width, strechedImageSize.height);
    cropBBox.x -= cropOffset.x;
    cropBBox.y -= cropOffset.y;

    if (this.nativeImageSize) {
      cropData = {
        cropMask: _types.default.makeMaskRect(p_localBbox),
        cropOffset: cropOffset,
        useCropMask: true
      };
      uncroppedSize = _types.default.newSize(realImageBounds.width, realImageBounds.height);
    }
  }

  var res;

  if (this.nativeImageSize) {
    var img = _displayitem.default.newImage(this.nativeImageID, realImageBounds, this.nativeImageSize.width, this.nativeImageSize.height, this.getActualProperty("filter"), 360 - this.getActualProperty("rotation"));

    res = [img];
  } else {
    var errorLabel = this.resourceState === _types.default.ResourceState.NOT_AVAILABLE ? "Image not found" : null;
    res = _frame.default.makeImageFrame(this, maskRect ? maskRect : cropBBox, errorLabel);
  }

  if (withBorders) {
    var borderStyle = this.getActualProperty("borderStyle");

    if (borderStyle === Properties.BorderStyles.SQUARE) {
      var items = Rectangle.makeRectangleDisplayItems(this._context, cropBBox, borderStyle, 0xFFFFFF, 0);
      items.forEach(item => {
        res.push(item);
      });
    }
  }

  return {
    displayItems: res,
    cropData: cropData,
    uncroppedSize: uncroppedSize
  };
};

image.CONTROL_PROTOTYPE.updateLayout = function (localBbox) {
  const res = this._getDisplayItems(localBbox, false, true);

  if (res.cropData) {
    this.setCropData(res.cropData);
  }

  if (res.uncroppedSize) {
    this.setUncroppedSize(res.uncroppedSize);
  }

  for (const item of res.displayItems) {
    this.pushToDisplayList(item);
  }

  _link.default.processControlLink(this); // Label


  const text = this.getActualProperty("text");

  if (text) {
    const textStyle = this.newStandardTextStyle();
    const {
      displayItems
    } = (0, _overlaylabel.makeOverlayLabel)(this, localBbox.width / 2, localBbox.height / 2, text, 1, textStyle);
    this.pushToDisplayList(displayItems);
  }
};

image.makeDisplayListForPreview = function (context, skin, bounds) {
  const {
    image_line_gap: gap,
    image_line_width: lineWidth,
    image_border_color: frameColor,
    image_cross_line_color: crossColor
  } = skin;
  const cross = Path.newPath(bounds.width, bounds.height);
  cross.moveTo(gap, gap);
  cross.lineTo(bounds.width - gap, bounds.height - gap);
  cross.moveTo(gap, bounds.height - gap);
  cross.lineTo(bounds.width - gap, gap);
  const background = Rectangle.makeFilledRectangle(bounds, 0xFFFFFF);

  const frame = _displayitem.default.createDisplayItemsFromSkin("image_frame", frameColor, bounds, context);

  const crossItem = _displayitem.default.newSVGPath(bounds, cross, lineWidth, crossColor);

  const rectangle = Rectangle.makeRectangleDisplayItems(context, bounds, Properties.BorderStyles.SQUARE, 0xFFFFFF, 0);
  return [...background, ...frame, crossItem, ...rectangle];
};

image.CONTROL_PROTOTYPE.getImageBounds = function () {
  var borderStyle = this.getActualProperty("borderStyle");
  var borderPadding = this.style('image_border_padding');
  var localBbox = this.localBBox();
  var imageBounds;

  if (borderStyle === Properties.BorderStyles.SQUARE) {
    imageBounds = _types.default.paddedRectangle(localBbox, borderPadding, borderPadding, borderPadding, borderPadding);
  } else {
    imageBounds = _balsamiq.default.clone(localBbox);
  }

  return imageBounds;
};

image.CONTROL_PROTOTYPE.getDisplayDataForCrop = function () {
  var items;

  if (this.nativeImageSize) {
    var p_localBbox = this.localBBox();

    let res = this._getDisplayItems(p_localBbox, true, false);

    if (res.cropData) {
      this.setCropData(res.cropData);
    }

    if (res.uncroppedSize) {
      this.setUncroppedSize(res.uncroppedSize);
    }

    items = res.displayItems;
  } else {
    items = null;
  }

  var res = new _displaynode.DisplayNode({ //isGroup: this.controlTypeID === "__group__" || this.controlTypeID === "com.balsamiq.mockups::Component",
    ...this.getDisplayDataBBoxes(),
    controlID: this.controlID,
    displayListItems: items,
    displayListSkinName: this._displayList.skinName,
    masks: [],
    colorTransforms: this._colorTransforms,
    clipOnBbox: this._clipOnBbox,
    linkAreas: this._linkAreas,
    isMarkup: this.isMarkup(),
    alpha: 1
  }, this.coreObjectID.controlID, []);
  return res;
};

image.CONTROL_PROTOTYPE.getDescription = function () {
  let res;
  const src = this.hasPropertyDefinition("src") ? this.getActualProperty("src") : null;

  if (!src) {
    res = "Empty Image";
  } else if (src.location === _types.default.SrcLocation.PROJECT_ASSETS) {
    const asset = this._context.getDependency(src.resourceID);

    res = asset ? asset.attributes.name : null;
  }

  return res || this.superClass.prototype.getDescription.call(this);
};

image.CONTROL_PROTOTYPE.getImageResource = function () {
  let res;
  const src = this.hasPropertyDefinition("src") ? this.getActualProperty("src") : null;

  if (!src) {
    res = null;
  } else if (src.location === _types.default.SrcLocation.PROJECT_ASSETS) {
    res = this._context.getDependency(src.resourceID);
  }

  return res;
};

image.CONTROL_PROTOTYPE.getImageBBoxPS = function () {
  if (!this.parent) {
    return undefined;
  }

  let boundsCS = this.getImageBounds();
  let originPS = MockupCoords.controlSpaceToParentSpace(boundsCS, this);
  return { ...originPS,
    width: boundsCS.width,
    height: boundsCS.height
  };
};
/**
 * 
 * @param {boolean} forCropPreview
 * @param {boolean} withBorder
 * @returns {{[key:string]: unknown, bounds: unknown}}
 */


image.CONTROL_PROTOTYPE.getImageOnlyDisplayData = function (forCropPreview, withBorder) {
  const res = this._getDisplayItems(this.localBBox(), forCropPreview, withBorder, false); // override bboxes and crop data with data from custom render


  let bboxes = { ...this.getDisplayDataBBoxes(),
    ...res.cropData,
    uncroppedSize: res.uncroppedSize
  };
  const displayData = { ...bboxes,
    controlID: this.controlID,
    displayListItems: res.displayItems,
    displayListSkinName: this._displayList.skinName,
    masks: [],
    colorTransforms: this._colorTransforms,
    clipOnBbox: this._clipOnBbox,
    linkAreas: [],
    isMarkup: false,
    alpha: 1
  };
  return displayData;
};

var _default = image;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/ioskeyboard.js":
/*!*****************************************!*\
  !*** ./js/core/controls/ioskeyboard.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

let KeySkinOptions = {
  letterKey: "ioskeyboard_letter_key",
  grayLetterKey: "ioskeyboard_letter_key_gray",
  key123: "ioskeyboard_123_key",
  return: "ioskeyboard_return_key",
  spaceBar: "ioskeyboard_space_key"
};
const WhiteButtonLabels = ["Q", "W", "E", "R", "T", "Y", "U", "I", "O", "P", "A", "S", "D", "F", "G", "H", "J", "K", "L", "Z", "X", "C", "V", "B", "N", "M"];
const control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::iPhoneKeyboard",
  controlTypeName: "iOS Keyboard",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        resize: Props.ResizeOptions.BOTH,
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT
      }
    }),
    orientation: {
      defaultValue: Props.Orientations.PORTRAIT,
      enum: Props.Orientations
    },
    device: {
      defaultValue: Props.iOSDevices.iPhone,
      enum: Props.iOSDevices
    }
  }
});

control.CONTROL_PROTOTYPE.initData = function () {
  this.cachedTexts = {};
};

control.CONTROL_PROTOTYPE.measure = function (sizeProperty) {
  const portrait = this.getActualProperty("orientation") === _types.default.Orientations.PORTRAIT;

  const iPhone = this.getActualProperty("device") === _types.default.iOSDevice.iPhone;

  let s;

  if (iPhone) {
    s = portrait ? _types.default.newSize(226, 160) : _types.default.newSize(403, 114);
  } else {
    s = portrait ? _types.default.newSize(546, 188) : _types.default.newSize(728, 251);
  }

  return s;
};

control.CONTROL_PROTOTYPE.updateLayout = function (localBbox) {
  let left, top;
  let width, height;
  const nativeSize = this.measure();
  const nativeRatio = nativeSize.width / nativeSize.height;
  const controlRatio = localBbox.width / localBbox.height;

  if (controlRatio > nativeRatio) {
    width = localBbox.height * nativeRatio;
    height = localBbox.height;
    left = (localBbox.width - width) / 2;
    top = 0;
  } else {
    width = localBbox.width;
    height = localBbox.width / nativeRatio;
    left = 0;
    top = (localBbox.height - height) / 2;
  }

  const scale = width / nativeSize.width;

  const isPortrait = this.getActualProperty("orientation") === _types.default.Orientations.PORTRAIT;

  if (this.getActualProperty("device") === _types.default.iOSDevice.iPhone) {
    this.iPhoneKeyboard(localBbox, isPortrait, left, top, width, height, scale);
  } else {
    this.iPadKeyboard(localBbox, isPortrait, left, top, width, height, scale);
  }
};

control.CONTROL_PROTOTYPE.iPhoneKeyboard = function (p_localBbox, p_portrait, p_left, p_top, p_width, p_height, p_scale) {
  const bW = p_scale * (p_portrait ? 21 : 33);
  const bigBW = p_scale * (p_portrait ? 31 : 40);
  const bH = p_scale * 27;
  let xGap = p_scale * (p_portrait ? 1 : 4);
  const yGap = p_scale * (p_portrait ? 38 : 28);
  let y = p_top + p_scale * (p_portrait ? 11 : 1);
  let x = p_left + p_scale * (p_portrait ? 3 : 17);
  const fontSize = 13;
  this.pushToDisplayList(Rectangle.makeFilledRectangle(_types.default.newRectangle(p_left, p_top, p_width, p_height), this.style("iosKeyboard_background_color"))); // line 1

  let i;

  for (i = 0; i <= 9; i += 1) {
    this.addButtonWithText(WhiteButtonLabels[i], x + i * (bW + xGap), y, bW, bH, p_scale, fontSize, KeySkinOptions.letterKey);
  } // line 2


  y += yGap;
  x = p_left + p_scale * (p_portrait ? 15 : 34);

  for (i = 10; i <= 18; i += 1) {
    this.addButtonWithText(WhiteButtonLabels[i], x + (i - 10) * (bW + xGap), y, bW, bH, p_scale, fontSize, KeySkinOptions.letterKey);
  } // line 3


  y += yGap;
  x = p_left + p_scale * (p_portrait ? 3 : 17);
  this.addButtonWithIcon(false, "arrow-up", x, y, bigBW, bH, p_scale, KeySkinOptions.key123);
  x = p_left + p_scale * (p_portrait ? 35 : 70);

  for (i = 19; i <= 25; i += 1) {
    this.addButtonWithText(WhiteButtonLabels[i], x + (i - 19) * (bW + xGap), y, bW, bH, p_scale, fontSize, KeySkinOptions.letterKey);
  }

  x = p_left + p_scale * (p_portrait ? 190 : 342);
  this.addButtonWithIcon(false, "caret-square-o-left", x, y, bigBW, bH, p_scale, KeySkinOptions.key123); // line 4

  xGap = p_scale * (p_portrait ? 1 : 3);
  y += yGap;
  x = p_left + p_scale * (p_portrait ? 3 : 17);
  this.addButtonWithText("123", x, y, bigBW, bH, p_scale, fontSize, KeySkinOptions.key123);
  x += bigBW + xGap;
  var w = p_portrait ? bW : bigBW;
  this.addButtonWithIcon(false, "globe", x, y, w, bH, p_scale, KeySkinOptions.grayLetterKey);
  x += w + xGap;
  w = bW;
  this.addButtonWithIcon(false, "microphone", x, y, w, bH, p_scale, KeySkinOptions.letterKey);
  x += w + xGap;
  w = p_scale * (p_portrait ? 95 : 170);
  this.addButtonWithText("space", x, y, w, bH, p_scale, fontSize, KeySkinOptions.spaceBar);
  x += w + xGap;
  w = p_scale * (p_portrait ? 47 : 71);
  this.addButtonWithText("return", x, y, w, bH, p_scale, fontSize, KeySkinOptions.return);
};

control.CONTROL_PROTOTYPE.iPadKeyboard = function (p_localBbox, p_portrait, p_left, p_top, p_width, p_height, p_scale) {
  const bW = p_scale * (p_portrait ? 45 : 60);
  const bH = p_scale * (p_portrait ? 41 : 47);
  const yGap = p_scale * (p_portrait ? 46 : 63);
  const xGap = p_scale * (p_portrait ? 4 : 5);
  let x = p_left + p_scale * (p_portrait ? 4 : 8);
  let y = p_top + p_scale * (p_portrait ? 4 : 8);
  const fontSize = p_portrait ? 18 : 20;
  this.pushToDisplayList(Rectangle.makeFilledRectangle(_types.default.newRectangle(p_left, p_top, p_width, p_height), this.style("iosKeyboard_background_color"))); // line 1

  let i;

  for (i = 0; i <= 9; i += 1) {
    this.addButtonWithText(WhiteButtonLabels[i], x + i * (bW + xGap), y, bW, bH, p_scale, fontSize, KeySkinOptions.letterKey);
  }

  this.addButtonWithIcon(true, "caret-square-o-left", x + 10 * (bW + xGap), y, bW, bH, p_scale, KeySkinOptions.grayLetterKey); // line 2

  y += yGap;
  x = p_left + p_scale * (p_portrait ? 21 : 36);

  for (i = 10; i <= 18; i += 1) {
    this.addButtonWithText(WhiteButtonLabels[i], x + (i - 10) * (bW + xGap), y, bW, bH, p_scale, fontSize, KeySkinOptions.letterKey);
  }

  this.addButtonWithText("return", x + 9 * (bW + xGap), y, p_scale * (p_portrait ? 77 : 98), bH, p_scale, fontSize, KeySkinOptions.key123); // line 3

  y += yGap;
  x = p_left + p_scale * (p_portrait ? 4 : 8);
  this.addButtonWithIcon(true, "arrow-up", x, y, bW, bH, p_scale, KeySkinOptions.key123);
  x += bW + xGap;

  for (i = 19; i <= 25; i += 1) {
    this.addButtonWithText(WhiteButtonLabels[i], x + (i - 19) * (bW + xGap), y, bW, bH, p_scale, fontSize, KeySkinOptions.letterKey);
  }

  this.addButtonWithText("!\\r,", x + 7 * (bW + xGap), y, bW, bH, p_scale, fontSize - 4, KeySkinOptions.letterKey);
  this.addButtonWithText("?\\r.", x + 8 * (bW + xGap), y, bW, bH, p_scale, fontSize - 4, KeySkinOptions.letterKey);
  this.addButtonWithIcon(true, "arrow-up", x + 9 * (bW + xGap), y, bW, bH, p_scale, KeySkinOptions.grayLetterKey); // line 4

  y += yGap;
  x = p_left + p_scale * (p_portrait ? 4 : 8);
  var w = bW * (p_portrait ? 1.4 : 1.3);
  this.addButtonWithText(".?123", x, y, w, bH, p_scale, fontSize - 4, KeySkinOptions.key123);
  x += w + xGap;
  this.addButtonWithIcon(true, "globe", x, y, bW, bH, p_scale, KeySkinOptions.grayLetterKey);
  x += bW + xGap;
  this.addButtonWithIcon(true, "microphone", x, y, bW, bH, p_scale, KeySkinOptions.grayLetterKey);
  x += bW + xGap;
  w = p_scale * (p_portrait ? 255 : 349);
  this.addButtonWithText(" ", x, y, w, bH, p_scale, fontSize, KeySkinOptions.spaceBar);
  x += w + xGap;
  w = w = bW * (p_portrait ? 1.4 : 1.3);
  this.addButtonWithText(".?123", x, y, w, bH, p_scale, fontSize - 4, KeySkinOptions.key123);
  x += w + xGap;
  this.addButtonWithIcon(true, "IpadKeyboard", x, y, bW, bH, p_scale, KeySkinOptions.grayLetterKey);
}; // apply a scale to the size and to the origin of the display items


var customScaleDisplayItems = function (p_displayItems, p_x, p_y, p_scale) {
  p_displayItems.forEach(item => {
    item.x = p_x + (item.x - p_x) * p_scale;
    item.y = p_y + (item.y - p_y) * p_scale;
    item.scaleX = item.scaleX * p_scale;
    item.scaleY = item.scaleY * p_scale;
  });
};

control.CONTROL_PROTOTYPE.addButtonWithText = function (label, x, y, width, height, scale, fontSize, keyOption) {
  const bounds = _types.default.newRectangle(x, y, width, height);

  const unscaledBounds = _types.default.newRectangle(x, y, width / scale, height / scale);

  let button = _displayitem.default.createDisplayItemsFromSkin(keyOption, this.style("default_border_color"), unscaledBounds, this._context);

  customScaleDisplayItems(button, x, y, scale);
  let textStyle = this.newTextStyle();
  textStyle.fontSize = fontSize * scale;

  if (!(this.cachedTexts[label] && Text.areTextsEqual(this.cachedTexts[label], label, textStyle))) {
    this.cachedTexts[label] = Text.newText(label, textStyle, this._context);
  }

  const displayLabel = this.cachedTexts[label].makeDisplayItems(bounds, _types.default.TextAlign.CENTER, _types.default.Positions.CENTER);
  this.pushToDisplayList(button, displayLabel);
};

control.CONTROL_PROTOTYPE.addButtonWithIcon = function (isBig, iconName, x, y, width, height, scale, keyOption) {
  const unscaledBounds = _types.default.newRectangle(x, y, width / scale, height / scale);

  const buttonBorder = _displayitem.default.createDisplayItemsFromSkin(keyOption, null, unscaledBounds, this._context);

  customScaleDisplayItems(buttonBorder, x, y, scale);
  const iconSize = (isBig ? 24 : 16) * scale;

  const iconBounds = _types.default.newRectangle(x + (width - iconSize) / 2, y + (height - iconSize) / 2, iconSize, iconSize);

  let icon;

  if (iconName != "IpadKeyboard") {
    var iconData = _icon.default.newAppIconData(iconName, iconSize, 0);

    icon = _icon.default.makeDisplayItemsFromIconData(this, iconData, iconBounds.x, iconBounds.y, this.style("iosKeyboard_icon_color"));
  } else {
    icon = _displayitem.default.newPathsAsset(this._context, iconName, iconBounds, this.style("default_icon_color"));
  }

  this.pushToDisplayList(buttonBorder, icon);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/ipad.js":
/*!**********************************!*\
  !*** ./js/core/controls/ipad.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const INTERNAL_WIDTH_RATIO = 550 / 616;
const INTERNAL_HEIGHT_RATIO = 730 / 916;
const TOPBAR_HEIGHT = 16;
const ICON_SIZE = 14;
const BATTERYICON_RIGHT_PADDING = 10;
const WIFIICON_LEFT_PADDING = 42;

let dateTo12HClock = function (date) {
  let hs = date.getHours();
  let hour = hs > 12 ? hs % 12 : hs;
  let ampm = hs > 11 ? "PM" : "AM";
  return hour + ":" + ("0" + date.getMinutes()).slice(-2) + " " + ampm;
};

const control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::iPad",
  controlTypeName: "iPad",
  controlAlternateNames: [],
  controlInitialProperties: {
    resizing: Props.iPhoneResizing.STRETCH
  },
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        resize: Props.ResizeOptions.BOTH,
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT
      }
    }),
    orientation: {
      defaultValue: Props.Orientations.PORTRAIT,
      enum: Props.Orientations
    },
    topBar: {
      defaultValue: true,
      dataType: 'bool'
    },
    bgTransparent: {
      defaultValue: false,
      dataType: 'bool'
    },
    bgPattern: {
      defaultValue: 'none',
      dataType: 'string'
    },
    resizing: {
      defaultValue: Props.iPhoneResizing.MAINTAIN_PROPORTIONS,
      enum: Props.iPhoneResizing
    }
  }
});

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  let size = _types.default.newSize(616, 916);

  if (this.getActualProperty("orientation") === _types.default.Orientations.LANDSCAPE) {
    let t = size.width;
    size.width = size.height;
    size.height = t;
  }

  return size;
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  let stretch = this.getActualProperty("resizing") == Props.iPhoneResizing.STRETCH;
  let orientation = this.getActualProperty('orientation');

  let portrait = this.getActualProperty('orientation') === _types.default.Orientations.PORTRAIT;

  let skinSize = this.measure();
  let skinRatio = skinSize.width / skinSize.height;
  let bboxRatio = p_localBbox.width / p_localBbox.height;
  let ratioH, ratioV;
  let button, speaker;

  if (!stretch) {
    if (portrait) {
      ratioH = INTERNAL_WIDTH_RATIO;
      ratioV = INTERNAL_HEIGHT_RATIO;
    } else {
      ratioV = INTERNAL_WIDTH_RATIO;
      ratioH = INTERNAL_HEIGHT_RATIO;
    }
  } else {
    ratioH = ratioV = 1;
  }

  let frameBbox = _types.default.newRectangle(0, 0, p_localBbox.width, p_localBbox.height);

  if (!stretch) {
    if (bboxRatio < skinRatio) {
      frameBbox.width = p_localBbox.width;
      frameBbox.height = frameBbox.width / skinRatio;
      frameBbox.x = 0;
      frameBbox.y = (p_localBbox.height - frameBbox.height) / 2;
    } else {
      frameBbox.height = p_localBbox.height;
      frameBbox.width = frameBbox.height * skinRatio;
      frameBbox.y = 0;
      frameBbox.x = (p_localBbox.width - frameBbox.width) / 2;
    }
  }

  let zoomRatio = stretch ? 1 : frameBbox.height / skinSize.height;
  let frame;
  let background = [];
  let topBar, ipadLabel, timeLabel, battery, wifi;
  let topBarBbox;
  let iconSize = ICON_SIZE * zoomRatio;

  let batteryIconData = _icon.default.newAppIconData("battery-three-quarters", iconSize, 0);

  let wifiIconData = _icon.default.newAppIconData("wifi", iconSize, 0);

  let iconColor = this.style("default_icon_color");

  if (!stretch) {
    if (portrait) {
      frame = _displayitem.default.newPathsAsset(this._context, "ipad", frameBbox);
    } else {
      frame = _displayitem.default.newRotated90CWPathsAsset(this._context, "ipad", frameBbox);
    }

    if (!this.getActualProperty('bgTransparent')) {
      let paddingVal = Math.min(frameBbox.height, frameBbox.width) * 0.03;
      background = Rectangle.makeFilledRectangle(_types.default.paddedRectangle(frameBbox, paddingVal, paddingVal, paddingVal, paddingVal), 0xFFFFFF);
    }

    let internalWidth = frameBbox.width * ratioH;
    let internalHeight = frameBbox.height * ratioV;

    if (this.getActualProperty('topBar')) {
      let x = frameBbox.x + (frameBbox.width - internalWidth) / 2;
      let y = frameBbox.y + (frameBbox.height - internalHeight) / 2;
      let w = internalWidth;
      let h = zoomRatio * TOPBAR_HEIGHT;
      topBarBbox = _types.default.newRectangle(x, y, w, h);
      topBar = Rectangle.makeFilledRectangle(topBarBbox, 0xD6DEE1); // Padded bbox for text and Icons

      let iconsBBox = _types.default.paddedRectangle(topBarBbox, 0, 0, h / 4, h / 4); // Style and top labels


      let textStyle = this.newTextStyle();
      textStyle.fontSize = zoomRatio * 10;
      textStyle.textColor = this.style("icon_default_color");

      if (!this.ipadLabelText || !Text.areTextsEqual(this.ipadLabelText, "iPad", textStyle)) {
        this.ipadLabelText = Text.newTextNoWrap("iPad", textStyle, this._context);
      }

      ipadLabel = this.ipadLabelText.makeDisplayItems(iconsBBox, _types.default.TextAlign.LEFT, _types.default.Positions.CENTER);
      const timeText = dateTo12HClock(new Date());

      if (!this.timeText || !Text.areTextsEqual(this.timeText, timeText, textStyle)) {
        this.timeText = Text.newTextNoWrap(timeText, textStyle, this._context);
      }

      timeLabel = this.timeText.makeDisplayItems(iconsBBox, _types.default.TextAlign.CENTER, _types.default.Positions.CENTER); // Top icons

      battery = _icon.default.makeDisplayItemsFromIconData(this, batteryIconData, topBarBbox.x + topBarBbox.width - BATTERYICON_RIGHT_PADDING * zoomRatio - iconSize, topBarBbox.y + (topBarBbox.height - iconSize) / 2, iconColor); // let wifiBbox = Types.newRectangle(topBarBbox.x + (WIFIICON_LEFT_PADDING * zoomRatio) - iconSize, topBarBbox.y + (topBarBbox.height - iconSize) / 2, iconSize, iconSize);

      wifi = _icon.default.makeDisplayItemsFromIconData(this, wifiIconData, topBarBbox.x + WIFIICON_LEFT_PADDING * zoomRatio - iconSize, topBarBbox.y + (topBarBbox.height - iconSize) / 2, iconColor);
    }
  } else {
    let screenBbox;
    let screenX = 32;
    let screenY = 94;
    let screenBottom = 92;
    let noTopBarItems = false;
    let buttonSize = this.style("ipad_button");
    let speakerSize = this.style("ipad_speaker");
    let padding = 10;

    if (portrait) {
      frame = _displayitem.default.createDisplayItemsFromSkin("ipad_scale9", 0, p_localBbox, this._context);
      screenBbox = _types.default.newRectangle(screenX, screenY, Math.max(p_localBbox.width - 2 * screenX, 1), Math.max(p_localBbox.height - (screenBottom + screenY), 1));

      if (p_localBbox.width > buttonSize.width + padding && p_localBbox.height > screenY + screenBottom) {
        button = _displayitem.default.newPathsAsset(this._context, "ipad_button", _types.default.newRectangle((p_localBbox.width - buttonSize.width) / 2, (p_localBbox.height + screenBbox.height + buttonSize.height) / 2 - 6, buttonSize.width, buttonSize.height), 0);
      }

      if (p_localBbox.width > speakerSize.width + padding && p_localBbox.height > screenY + screenBottom) {
        speaker = _displayitem.default.newPathsAsset(this._context, "ipad_speaker", _types.default.newRectangle((p_localBbox.width - speakerSize.width) / 2, (screenBbox.y - speakerSize.height) / 2 - 1, speakerSize.width, speakerSize.height), 0);
      } //Preventing partially empty background


      if (p_localBbox.width < 72) {
        screenBbox.width = p_localBbox.width - 4;
        screenBbox.x = p_localBbox.x + 2;
        noTopBarItems = true;
      }

      if (screenBbox.height < TOPBAR_HEIGHT) {
        screenBbox.height = p_localBbox.height - 4;
        screenBbox.y = p_localBbox.y + 2;
        noTopBarItems = true;
      }
    } else {
      frame = _displayitem.default.newRotated90CCWSlice9("ipad_scale9", p_localBbox, this._context, 0);
      screenBbox = _types.default.newRectangle(screenY, screenX, Math.max(p_localBbox.width - (screenBottom + screenY), 1), Math.max(p_localBbox.height - 2 * screenX, 1));

      if (p_localBbox.height > buttonSize.height + padding && p_localBbox.width > screenY + screenBottom) {
        button = _displayitem.default.newPathsAsset(this._context, "ipad_button", _types.default.newRectangle((screenBbox.x - buttonSize.height) / 2 + 2, (p_localBbox.height + buttonSize.width) / 2, buttonSize.width, buttonSize.height), 0, 1, 90);
      }

      if (p_localBbox.height > speakerSize.height + padding && p_localBbox.width > screenY + screenBottom) {
        speaker = _displayitem.default.newPathsAsset(this._context, "ipad_speaker", _types.default.newRectangle((p_localBbox.width + screenBbox.x + screenBbox.width - speakerSize.height) / 2, (p_localBbox.height + speakerSize.width) / 2, speakerSize.width, speakerSize.height), 0, 1, 90);
      }

      if (p_localBbox.width < 194) {
        screenBbox.width = p_localBbox.width - 4;
        screenBbox.x = p_localBbox.x + 2;
        noTopBarItems = true;
      }

      if (screenBbox.height < TOPBAR_HEIGHT) {
        screenBbox.height = p_localBbox.height - 4;
        screenBbox.y = p_localBbox.y + 2;
        noTopBarItems = true;
      }
    }

    if (!this.getActualProperty('bgTransparent')) {
      background = Rectangle.makeFilledRectangle(screenBbox, 0xFFFFFF);
    }

    if (this.getActualProperty('topBar') && screenBbox.height > TOPBAR_HEIGHT) {
      topBarBbox = _types.default.newRectangle(screenBbox.x, screenBbox.y + (portrait ? 0 : 2), screenBbox.width, TOPBAR_HEIGHT);
      topBar = Rectangle.makeFilledRectangle(topBarBbox, 0xD6DEE1);

      if (!noTopBarItems) {
        // Padded bbox for text and Icons
        let iconsBBox = _types.default.paddedRectangle(topBarBbox, 0, 0, TOPBAR_HEIGHT / 4, TOPBAR_HEIGHT / 4);

        let textStyle = this.newTextStyle();
        textStyle.fontSize = 10;
        textStyle.textColor = this.style("icon_default_color");

        if (!this.ipadLabelText || !Text.areTextsEqual(this.ipadLabelText, "iPad", textStyle)) {
          this.ipadLabelText = Text.newTextNoWrap("iPad", textStyle, this._context);
        }

        if (iconsBBox.x + this.ipadLabelText.measuredSize().width + TOPBAR_HEIGHT / 4 < screenBbox.x + screenBbox.width) {
          ipadLabel = this.ipadLabelText.makeDisplayItems(iconsBBox, _types.default.TextAlign.LEFT, _types.default.Positions.CENTER);
        }

        if (p_localBbox.width > (portrait ? 185 : 306)) {
          const timeText = dateTo12HClock(new Date());

          if (!this.timeText || !Text.areTextsEqual(this.timeText, timeText, textStyle)) {
            this.timeText = Text.newTextNoWrap(timeText, textStyle, this._context);
          }

          timeLabel = this.timeText.makeDisplayItems(iconsBBox, _types.default.TextAlign.CENTER, _types.default.Positions.CENTER);
        } // Top icons


        let iconSize = ICON_SIZE;
        let wifiX = topBarBbox.x + WIFIICON_LEFT_PADDING * zoomRatio - iconSize;
        let batteryX = topBarBbox.x + topBarBbox.width - BATTERYICON_RIGHT_PADDING - iconSize;

        if (batteryX > wifiX + iconSize) {
          battery = _icon.default.makeDisplayItemsFromIconData(this, batteryIconData, batteryX, topBarBbox.y + (topBarBbox.height - iconSize) / 2, iconColor);
        }

        if (wifiX + iconSize < screenBbox.x + screenBbox.width) {
          wifi = _icon.default.makeDisplayItemsFromIconData(this, wifiIconData, wifiX, topBarBbox.y + (topBarBbox.height - iconSize) / 2, iconColor);
        }
      }
    }
  }

  this.pushToDisplayList(background, topBar, frame, ipadLabel, timeLabel, battery, wifi, button, speaker);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/iphone.js":
/*!************************************!*\
  !*** ./js/core/controls/iphone.js ***!
  \************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var controlsMod = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var textMod = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _snapping = __webpack_require__(/*! core/controls/snapping */ "./js/core/controls/snapping.js");

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

let control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::iPhone",
  controlTypeName: "iPhone",
  controlAlternateNames: [],
  controlInitialProperties: {
    model: Props.iPhoneModels.iPhoneX,
    resizing: Props.iPhoneResizing.STRETCH
  },
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        resize: Props.ResizeOptions.BOTH,
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT
      }
    }),
    model: {
      defaultValue: Props.iPhoneModels.iPhone4,
      enum: Props.iPhoneModels
    },
    orientation: {
      defaultValue: Props.Orientations.PORTRAIT,
      enum: Props.Orientations
    },
    topBar: {
      defaultValue: true,
      dataType: 'bool'
    },
    bgPattern: {
      defaultValue: Props.iPhoneBackgrounds.ALL_WHITE,
      enum: Props.iPhoneBackgrounds
    },
    bgTransparent: {
      defaultValue: false,
      dataType: 'bool'
    },
    resizing: {
      defaultValue: Props.iPhoneResizing.MAINTAIN_PROPORTIONS,
      enum: Props.iPhoneResizing
    }
  }
});
const iPhone4 = {
  proportions: {
    left: 24,
    right: 24,
    top: 83,
    bottom: 80
  },
  stretch: {
    left: 23,
    top: 81,
    bottomOffset: 3
  },
  earthing: {
    landscape: {
      xOffset: 5,
      yOffset: 29
    },
    portrait: {
      xOffset: -13,
      yOffset: 5
    }
  },
  button: {
    landscape: {
      xOffset: -4
    },
    portrait: {
      yOffset: -3
    }
  }
};
const iPhone5 = {
  earthing: {
    landscape: {
      xOffset: -10,
      yOffset: 1
    },
    portrait: {
      xOffset: 0,
      yOffset: -7
    }
  },
  button: {
    landscape: {
      xOffset: -3
    },
    portrait: {
      yOffset: 0
    }
  }
};
const iPhone6 = {
  proportions: {
    left: 29,
    right: 29,
    top: 90,
    bottom: 95
  },
  stretch: {
    left: 24,
    top: 87,
    bottomOffset: 7
  },
  earthing: {
    landscape: {
      xOffset: -5,
      yOffset: 30
    },
    portrait: {
      xOffset: -16,
      yOffset: -6
    }
  },
  button: {
    landscape: {
      xOffset: 0
    },
    portrait: {
      yOffset: 2
    }
  }
};
const iPhoneX = {
  proportions: {
    left: 25,
    right: 24,
    top: 22,
    bottom: 22
  },
  stretch: {
    left: 24,
    top: 22,
    bottomOffset: 0
  },
  earthing: {
    landscape: {
      xOffset: 22,
      yOffset: -13
    },
    portrait: {
      xOffset: 14,
      yOffset: 21
    }
  }
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  let model = this.getActualProperty("model");
  let size;

  switch (model) {
    case Props.iPhoneModels.iPhone4:
      {
        size = _types.default.newSize(282, 510);
        break;
      }

    case Props.iPhoneModels.iPhone5:
      {
        size = _types.default.newSize(282, 573);
        break;
      }

    case Props.iPhoneModels.iPhone6:
      {
        size = _types.default.newSize(380, 754);
        break;
      }

    case Props.iPhoneModels.iPhoneX:
      {
        size = _types.default.newSize(436, 867);
        break;
      }
  }

  if (this.getActualProperty("orientation") === Props.Orientations.LANDSCAPE) {
    let t = size.width;
    size.width = size.height;
    size.height = t;
  }

  return size;
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  let model = this.getActualProperty("model");
  let pattern = this.getActualProperty("bgPattern");
  let transparent = this.getActualProperty("bgTransparent");
  let landscape = this.getActualProperty("orientation") === Props.Orientations.LANDSCAPE;
  let nativeSize = this.measure();
  let nativeRatio = nativeSize.width / nativeSize.height;
  let controlRatio = p_localBbox.width / p_localBbox.height;
  let stretch = this.getActualProperty("resizing") == Props.iPhoneResizing.STRETCH;

  let iphoneSnapRect = _types.default.newRectangle(0, 0, 0, 0); // compute the bounds of the iPhone ignoring the adjustment required by the rotation (landscape)


  let iphoneBounds = _types.default.newRectangle(0, 0, p_localBbox.width, p_localBbox.height);

  if (!stretch) {
    if (controlRatio > nativeRatio) {
      iphoneBounds.width = p_localBbox.height * nativeRatio;
      iphoneBounds.height = p_localBbox.height;
      iphoneBounds.x = (p_localBbox.width - iphoneBounds.width) / 2;
      iphoneBounds.y = 0;
    } else {
      iphoneBounds.width = p_localBbox.width;
      iphoneBounds.height = p_localBbox.width / nativeRatio;
      iphoneBounds.x = 0;
      iphoneBounds.y = (p_localBbox.height - iphoneBounds.height) / 2;
    }
  }

  let zoomRatio = !stretch ? iphoneBounds.width / nativeSize.width : 1;
  let left;
  let right;
  let top;
  let bottom;
  let isIphone6 = model === Props.iPhoneModels.iPhone6;
  let isIphone5 = model === Props.iPhoneModels.iPhone5;
  let isIphoneX = model === Props.iPhoneModels.iPhoneX;
  let phoneSkin = this.style("iphone" + (isIphone6 ? "6" : isIphone5 ? "5" : isIphoneX ? "X" : "4") + "_scale9");

  if (!stretch) {
    let defaultValues = isIphone6 ? iPhone6.proportions : isIphoneX ? iPhoneX.proportions : iPhone4.proportions;

    if (landscape) {
      left = iphoneBounds.x + zoomRatio * defaultValues.top;
      right = p_localBbox.width - iphoneBounds.x - iphoneBounds.width + zoomRatio * defaultValues.bottom;
      top = iphoneBounds.y + zoomRatio * defaultValues.right;
      bottom = p_localBbox.height - iphoneBounds.y - iphoneBounds.height + zoomRatio * defaultValues.left;
    } else {
      left = iphoneBounds.x + zoomRatio * defaultValues.left;
      right = p_localBbox.width - iphoneBounds.x - iphoneBounds.width + zoomRatio * defaultValues.right;
      top = iphoneBounds.y + zoomRatio * defaultValues.top;
      bottom = p_localBbox.height - iphoneBounds.y - iphoneBounds.height + zoomRatio * defaultValues.bottom;
    }
  } else {
    let defaultValues = isIphone6 ? iPhone6.stretch : isIphoneX ? iPhoneX.stretch : iPhone4.stretch;

    if (landscape) {
      left = defaultValues.top;
      top = defaultValues.left;
      let minRatioWidth = isIphone6 ? 360 : 254;
      let minRatioHeight = isIphone6 ? 126 : 132;

      if (p_localBbox.height < minRatioHeight) {
        if (isIphone6) {
          top -= (minRatioHeight - p_localBbox.height) / 4;
        } else {
          top -= (minRatioHeight - p_localBbox.height) / 2;
        }
      }

      top = Math.max(top, 0);

      if (p_localBbox.width < minRatioWidth) {
        let div = isIphone6 ? 4 : 3;
        left -= (minRatioWidth - p_localBbox.width) / div;
      }

      bottom = top;
      right = left;

      if (isIphoneX && p_localBbox.height < 230) {
        top = 1;
        bottom = 1;
      }
    } else {
      left = defaultValues.left;
      top = defaultValues.top;
      let minRatioHeight = phoneSkin.scale9[0] + phoneSkin.scale9[3];
      let minRatioWidth = isIphone6 ? 107 : isIphoneX ? 237 : 135;
      let bottomOffset = defaultValues.bottomOffset;

      if (p_localBbox.height < minRatioHeight) {
        if (isIphone6) {
          top -= (minRatioHeight - p_localBbox.height) / 4;
          bottomOffset = bottomOffset / 4;
        } else {
          top -= (minRatioHeight - p_localBbox.height) / 2;
          bottomOffset = bottomOffset / 4;
        }
      }

      top = Math.max(top, 0);

      if (p_localBbox.width < minRatioWidth) {
        left -= (minRatioWidth - p_localBbox.width) / 2;
      }

      bottom = top + bottomOffset;
      right = left;
    }
  }

  let barHeight = zoomRatio * 40;
  let bottomBarHeight = zoomRatio * (landscape ? 30 : 40);
  let blackBarHeight = zoomRatio * (landscape ? 39 : 40);
  let topBarHeight = zoomRatio * (isIphoneX ? 30 : 16);
  let topBarVisible = !stretch || (landscape ? p_localBbox.height > phoneSkin.scale9[1] + phoneSkin.scale9[2] : p_localBbox.height > phoneSkin.scale9[0] + phoneSkin.scale9[3]);
  let iphoneXShowIcons = landscape ? p_localBbox.width > 215 : p_localBbox.width > 260;
  topBarVisible = topBarVisible && (isIphoneX ? iphoneXShowIcons : true);
  let top2 = top + (this.getActualProperty("topBar") ? topBarHeight + (isIphoneX ? -1 : 0) : 0);
  let cornerRadius = 1;
  let bounds; // background pattern

  if (!transparent) {
    // let cornerRadius;
    if (!stretch) {
      let offset = model === Props.iPhoneModels.iPhone6 ? 1 : 0;
      bounds = _types.default.newRectangle(left, Math.max(0, top - offset), p_localBbox.width - left - right, Math.min(p_localBbox.height, p_localBbox.height - top - bottom + offset));
      cornerRadius = 1;
    } else {
      bounds = _types.default.newRectangle(Math.max(left, 1), top, Math.min(p_localBbox.width - left - right, p_localBbox.width - 2), p_localBbox.height - top - bottom);
      cornerRadius = isIphoneX ? 35 : 1;
    }

    let backgroundPath = Path.newRoundedRectanglePath(bounds.width, bounds.height, 0, cornerRadius);

    let background = _displayitem.default.newSVGPath(bounds, backgroundPath, 0, null, pattern === Props.iPhoneBackgrounds.ALL_BLACK ? 0x3F3F3F : 0xFFFFFF);

    this.pushToDisplayList(background);
  }

  cornerRadius = isIphoneX && (!iphoneXShowIcons || !topBarVisible) ? 35 : 1;
  iphoneSnapRect.x = left;
  iphoneSnapRect.y = top2;
  iphoneSnapRect.width = p_localBbox.width - right;
  iphoneSnapRect.height = p_localBbox.height - bottom;
  let patternColor = pattern === Props.iPhoneBackgrounds.ALL_BLACK ? 0x3F3F3F : 0xFFFFFF;
  let bottomBounds = null;
  let patternBottomColor = null;
  bounds = null;

  if (isIphoneX && !topBarVisible && this.getActualProperty("topBar")) {
    top2 = top;
    barHeight += topBarHeight;
  }

  if (pattern === Props.iPhoneBackgrounds.TOP_ONLY && topBarVisible && p_localBbox.height - top2 - bottom > barHeight) {
    iphoneSnapRect.y += barHeight;
    bounds = _types.default.newRectangle(left, top2, p_localBbox.width - left - right, barHeight);
    patternColor = 0x828896;
  } else if (pattern === Props.iPhoneBackgrounds.TOP_BLUE_BOTTOM && p_localBbox.height - top2 - bottom > barHeight + bottomBarHeight) {
    iphoneSnapRect.y += barHeight;
    iphoneSnapRect.height -= bottomBarHeight;
    bounds = _types.default.newRectangle(left, top2, p_localBbox.width - left - right, barHeight);
    bottomBounds = _types.default.newRectangle(left, p_localBbox.height - bottom - bottomBarHeight, p_localBbox.width - left - right, bottomBarHeight);
    patternColor = 0x828896;
    patternBottomColor = 0x828896;
  } else if (pattern === Props.iPhoneBackgrounds.TOP_BLACK_BOTTOM && p_localBbox.height - top2 - bottom > barHeight + bottomBarHeight) {
    iphoneSnapRect.y += barHeight;
    iphoneSnapRect.height -= blackBarHeight;
    bounds = _types.default.newRectangle(left, top2, p_localBbox.width - left - right, barHeight);
    bottomBounds = _types.default.newRectangle(left, p_localBbox.height - bottom - blackBarHeight, p_localBbox.width - left - right, blackBarHeight);
    patternColor = 0x828896;
    patternBottomColor = 0x3F3F3F;
  }

  if (bounds) {
    let roundCorners = isIphoneX && stretch;
    let path = roundedRect(bounds.width, bounds.height, cornerRadius, roundCorners, false);
    this.pushToDisplayList(_displayitem.default.newSVGPath(bounds, path, 0, null, patternColor));

    if (bottomBounds) {
      path = roundedRect(bottomBounds.width, bottomBounds.height, cornerRadius, false, roundCorners);
      this.pushToDisplayList(_displayitem.default.newSVGPath(bottomBounds, path, 0, null, patternBottomColor));
    }
  }

  this.iphoneSnapRect = iphoneSnapRect; // top bar

  let screenWidth = p_localBbox.width - left - right;
  let screenHeight = p_localBbox.height - top - bottom;
  let topBarColor = isIphoneX ? patternColor : pattern !== Props.iPhoneBackgrounds.ALL_BLACK ? 0xD6DEE1 : 0x2E2E2E;

  if (this.getActualProperty("topBar") && topBarVisible) {
    let topBarItemsColor;

    if (!isIphoneX) {
      topBarItemsColor = pattern !== Props.iPhoneBackgrounds.ALL_BLACK ? this.style("iphone_top_bar_items_color") : 0xFFFFFF;
    } else {
      topBarItemsColor = pattern == Props.iPhoneBackgrounds.ALL_WHITE ? this.style("iphone_top_bar_items_color") : 0xFFFFFF;
    }

    let offset = model === Props.iPhoneModels.iPhone6 ? 1 : 0;
    let showBackground = !(isIphoneX && transparent) || pattern === Props.iPhoneBackgrounds.TOP_ONLY || pattern === Props.iPhoneBackgrounds.TOP_BLUE_BOTTOM || pattern === Props.iPhoneBackgrounds.TOP_BLACK_BOTTOM;

    if (showBackground) {
      this.pushToDisplayList((0, _rectangle.makeFilledRectangle)(_types.default.newRectangle(left, top, p_localBbox.width - left - right, topBarHeight), topBarColor));
    }

    let x = left + zoomRatio * 10;

    if ((!isIphoneX && (!stretch || x + topBarHeight < screenWidth) || isIphoneX && iphoneXShowIcons || !stretch) && topBarHeight > 2) {
      let iconSize = isIphoneX ? topBarHeight - 15 : topBarHeight - 2;
      iphoneXShowIcons = iphoneXShowIcons && iconSize > 0;
      let signalIconData = iconSize > 0 ? _icon.default.newAppIconData("signal", iconSize, 0) : null;
      let iconPadding = 5;
      let signalX = !isIphoneX ? x + 1 : left + screenWidth - 25 * zoomRatio - 3 * (iconSize + iconPadding);
      let signalY = !isIphoneX ? top + 1 : top + topBarHeight - iconSize - 1;

      if (signalIconData) {
        this.pushToDisplayList(_icon.default.makeDisplayItemsFromIconData(this, signalIconData, signalX, signalY, topBarItemsColor));
        x = !isIphoneX ? x + iconSize + 2 + zoomRatio * 2 : signalX + iconSize + iconPadding;
      }

      let textStyle = this.newTextStyle();
      textStyle.fontSize = !isIphoneX ? zoomRatio * 10 : zoomRatio * 12;
      textStyle.textColor = topBarItemsColor;
      let carrierText = textMod.newText("ABC", textStyle, this._context);

      if (isIphoneX || !stretch || x + carrierText.measuredSize().width < p_localBbox.width - left) {
        if (!isIphoneX) {
          this.pushToDisplayList(carrierText.makeDisplayItems(_types.default.newRectangle(x, top + offset, carrierText.measuredSize().width, topBarHeight), Props.TextAlignments.LEFT, Props.Positions.CENTER));
        }

        let timeText = textMod.newText("09:52 AM", textStyle, this._context);
        let timeBBox;

        if (!isIphoneX) {
          timeBBox = _types.default.newRectangle(left, top + offset, p_localBbox.width - left - right, topBarHeight);
        } else {
          timeBBox = _types.default.newRectangle(left + 28 * zoomRatio, top + topBarHeight - timeText.measuredSize().height, timeText.measuredSize().width, timeText.measuredSize().height);
        }

        if (textStyle.fontSize > 0 && !stretch || screenWidth - (left + zoomRatio + 2 * topBarHeight + carrierText.measuredSize().width) > timeText.measuredSize().width || isIphoneX && iphoneXShowIcons) {
          this.pushToDisplayList(timeText.makeDisplayItems(timeBBox, Props.TextAlignments.CENTER, Props.Positions.CENTER));
        }

        if (isIphoneX && iphoneXShowIcons) {
          let wifiIconData = _icon.default.newAppIconData("wifi", iconSize, 0);

          this.pushToDisplayList(_icon.default.makeDisplayItemsFromIconData(this, wifiIconData, x, signalY, topBarItemsColor));
          x += iconSize + iconPadding;
        }

        if (!isIphoneX && (!stretch || p_localBbox.width - right - topBarHeight - zoomRatio * 12 > x + carrierText.measuredSize().width) || isIphoneX && iphoneXShowIcons) {
          let batteryIconData = _icon.default.newAppIconData("battery-three-quarters", iconSize, 0);

          let batteryX = !isIphoneX ? p_localBbox.width - right - iconSize - zoomRatio * 12 : x;
          let batteryY = !isIphoneX ? top : signalY;
          this.pushToDisplayList(_icon.default.makeDisplayItemsFromIconData(this, batteryIconData, batteryX, batteryY, topBarItemsColor));
        }
      }
    }
  } // now adjust the iPhone bounds for the rotation (landscape)


  let skinName;
  let iphone;
  let earthing = null;
  let button = null;

  if (!stretch) {
    if (landscape) {
      let t = iphoneBounds.width;
      iphoneBounds.width = iphoneBounds.height;
      iphoneBounds.height = t;
      iphoneBounds.y = p_localBbox.height - iphoneBounds.y;
    }

    skinName = model === Props.iPhoneModels.iPhone5 ? "iphone_5" : model === Props.iPhoneModels.iPhone6 ? "iphone_6" : model === Props.iPhoneModels.iPhoneX ? "iphone_X" : "iphone_4";
    iphone = _displayitem.default.newPathsAsset(this._context, skinName, iphoneBounds, null, 1, landscape ? 90 : 0);
  } else {
    let prefix = model === Props.iPhoneModels.iPhone5 ? "iphone5_" : model === Props.iPhoneModels.iPhone6 ? "iphone6_" : model === Props.iPhoneModels.iPhoneX ? "iphoneX_" : "iphone4_";
    skinName = prefix + "scale9";
    let earthingSkinSize = this.style(prefix + "earthing");
    let buttonSkinSize = isIphoneX ? null : this.style(prefix + "button");
    let offsets = isIphoneX ? iPhoneX : isIphone6 ? iPhone6 : isIphone5 ? iPhone5 : iPhone4;

    if (landscape) {
      iphone = _displayitem.default.newRotated90CCWSlice9(skinName, p_localBbox, this._context, 0);

      if (buttonSkinSize && screenHeight > buttonSkinSize.width && left > buttonSkinSize.height) {
        button = _displayitem.default.newPathsAsset(this._context, prefix + "button", _types.default.newRectangle(p_localBbox.width - (right + buttonSkinSize.height) / 2 + offsets.button.landscape.xOffset, (p_localBbox.height + buttonSkinSize.width) / 2, buttonSkinSize.width, buttonSkinSize.height), 0, 1, 90);
      }

      let showEarthing = !isIphoneX ? screenHeight > earthingSkinSize.width && right > earthingSkinSize.height : p_localBbox.width > 389 && p_localBbox.height > 350;

      if (showEarthing) {
        earthing = _displayitem.default.newPathsAsset(this._context, prefix + "earthing", _types.default.newRectangle((left - earthingSkinSize.height) / 2 + offsets.earthing.landscape.xOffset, (p_localBbox.height + earthingSkinSize.width) / 2 + offsets.earthing.landscape.yOffset, earthingSkinSize.width, earthingSkinSize.height), 0, 1, 90);
      }
    } else {
      iphone = _displayitem.default.createDisplayItemsFromSkin(skinName, 0, p_localBbox, this._context);

      if (buttonSkinSize && screenWidth > buttonSkinSize.width && bottom > buttonSkinSize.height) {
        button = _displayitem.default.newPathsAsset(this._context, prefix + "button", _types.default.newRectangle((p_localBbox.width - buttonSkinSize.width) / 2, p_localBbox.height - (bottom + buttonSkinSize.height) / 2 + offsets.button.portrait.yOffset, buttonSkinSize.width, buttonSkinSize.height));
      }

      let showEarthing = (!isIphoneX ? screenWidth > earthingSkinSize.width : p_localBbox.width > 356) && top > earthingSkinSize.height;

      if (showEarthing) {
        earthing = _displayitem.default.newPathsAsset(this._context, prefix + "earthing", _types.default.newRectangle((p_localBbox.width - earthingSkinSize.width) / 2 + offsets.earthing.portrait.xOffset, (top - earthingSkinSize.height) / 2 + offsets.earthing.portrait.yOffset, earthingSkinSize.width, earthingSkinSize.height));
      }
    }
  }

  this.pushToDisplayList(iphone, earthing, button);
}; //draws a rounded rectangle where you can choose to round the top corners or the bottom ones


let roundedRect = function (width, height, cornerRadius, top, bottom) {
  _balsamiq.default.assertPositiveNumber(width);

  _balsamiq.default.assertPositiveNumber(height);

  _balsamiq.default.assertBool(top);

  _balsamiq.default.assertBool(bottom);

  let w = width;
  let h = height;
  let rx = Math.min(width / 2, cornerRadius);
  let ry = Math.min(width / 2, cornerRadius);
  let path = new Path.Path(w, h);
  path.moveTo(rx, 0);
  path.lineTo(w - rx, 0);

  if (top) {
    path.arc90(rx, ry);
  } else {
    path.lineTo(width, 0);
    path.lineTo(width, ry);
  }

  path.lineTo(width, height - ry);

  if (bottom) {
    path.arc90(-rx, ry);
  } else {
    path.lineTo(width, height);
    path.lineTo(width - rx, height);
  }

  path.lineTo(rx, height);

  if (bottom) {
    path.arc90(-rx, -ry);
  } else {
    path.lineTo(0, height);
    path.lineTo(0, height - ry);
  }

  path.lineTo(0, ry);

  if (top) {
    path.arc90(rx, -ry);
  } else {
    path.lineTo(0, 0);
    path.lineTo(rx, 0);
  }

  return path;
};

control.CONTROL_PROTOTYPE.getSnapLines = function (withMarkup) {
  let snapLines = (0, _snapping.makeSnapLinesForBbox)(this.getUserBBox(true), false);
  snapLines.push(new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.VERTICAL, this.iphoneSnapRect.x), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.VERTICAL, this.iphoneSnapRect.width), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.HORIZONTAL, this.iphoneSnapRect.y), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.HORIZONTAL, this.iphoneSnapRect.height));
  return snapLines;
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/iphonemenu.js":
/*!****************************************!*\
  !*** ./js/core/controls/iphonemenu.js ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _switch = __webpack_require__(/*! core/bricks/switch */ "./js/core/bricks/switch.js");

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::iPhoneMenu",
  controlTypeName: "iOS Menu",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.WIDTH_AND_OR_HEIGHT,
        resize: Props.ResizeOptions.HORIZONTAL
      }
    }),
    ...Props.PropertyDefinitions.TEXT({
      defaultText: "A Simple Label\n" + "+ Add and sub-menu, >\n" + "- Delete\\r(Cancel)\n" + "Two Labels\\, and a comma, yup\n" + "v A Checkmark, (>)\n" + "* A Bullet, =\n" + "_ Space for an icon\n" + "__ Space for a big icon\n" + "On button, ON\n" + "Off button, OFF\n" + "\n" + "v An empty row, (above)",
      multiline: true,
      splitOn: Props.SplitOnTypes.NEWLINE
    }),
    ...Props.PropertyDefinitions.BORDER_STYLE(Props.SimpleBorderStyles.SQUARE, Props.SimpleBorderStyles),
    // why not rounded?
    ...Props.PropertyDefinitions.links()
  }
});
controlsMod.addStandardTextPropertyDefinitions(control);

control.CONTROL_PROTOTYPE.initData = function () {
  this._assetLines = [];
};

var ICON_COLOR = {
  "CircledPlusIcon": 0x009A0F,
  "CircledMinusIcon": 0xCE2A27,
  "CheckMarkIcon": 0x000000,
  "DotIcon": 0x0000FF,
  "ICON": null,
  "BIGICON": null,
  "breadcrumbs_separator": 0x000000,
  "GtCircledIcon": 0x2575DA,
  "ListIcon": 0x999999,
  "ON": null,
  "OFF": null
};
var ICON_SIZE = {
  "CircledPlusIcon": _types.default.newSize(16, 16),
  "CircledMinusIcon": _types.default.newSize(16, 16),
  "CheckMarkIcon": _types.default.newSize(13, 13),
  "DotIcon": _types.default.newSize(12, 12),
  "ICON": _types.default.newSize(16, 16),
  "BIGICON": _types.default.newSize(48, 48),
  "breadcrumbs_separator": null,
  "GtCircledIcon": _types.default.newSize(16, 16),
  "ListIcon": _types.default.newSize(16, 16),
  "ON": _types.default.newSize(49, 28),
  "OFF": _types.default.newSize(49, 28)
};
var ICON_SKIN = {
  "add": "CircledPlusIcon",
  "delete": "CircledMinusIcon",
  "checkmark": "CheckMarkIcon",
  "bullet": "DotIcon",
  "icon": "ICON",
  "bigicon": "BIGICON",
  "submenu": "breadcrumbs_separator",
  "rightarrow": "GtCircledIcon",
  "drag": "ListIcon",
  "on": "ON",
  "off": "OFF"
};
var LEFT_ICON_MARKUP = {
  "add": new RegExp("^\\+ "),
  "delete": new RegExp("^- "),
  "checkmark": new RegExp("^v "),
  "bullet": new RegExp("^\\* "),
  "icon": new RegExp("^_ "),
  "bigicon": new RegExp("^__ ")
};
var RIGHT_ICON_MARKUP = {
  "submenu": new RegExp(" >$"),
  "rightarrow": new RegExp(" \\(>\\)$"),
  "drag": new RegExp(" =$"),
  "on": new RegExp(" ON$"),
  "off": new RegExp(" OFF$")
};

var AssetLine = function (p_text, p_owner) {
  this.control = p_owner;
  this.separator = false;
  this.leftIcon = null;
  this.centerText = null;
  this.rightText = null;
  this.rightIcon = null;
  var text = p_text;
  this.leftIcon = this.parseForIcon(p_text, LEFT_ICON_MARKUP, ICON_SKIN);

  if (this.leftIcon) {
    text = this.removeLeftIconMarkup(p_text);
  }

  var texts = this.parseTexts(text);
  this.rightIcon = this.parseForIcon(texts.right, RIGHT_ICON_MARKUP, ICON_SKIN);

  if (this.rightIcon) {
    texts.right = this.trimTrailingSpaces(this.removeRightIconMarkup(texts.right));
  } //Change strings to texts


  var textStyle = this.control.newStandardTextStyle();

  if (!texts.center || texts.center.replace(/_/g, "").length === 0) {
    this.separator = true;
  } else if (texts.center) {
    this.centerText = textMod.newTextNoWrap(texts.center, textStyle, this.control._context);
  }

  if (texts.right) {
    var textStyle2 = this.control.newTextStyle();
    textStyle2.textColor = this.control.style("default_text_color");
    textStyle2.fontSize = textStyle.fontSize - 2;
    this.rightText = textMod.newTextNoWrap(texts.right, textStyle2, this.control._context);
  }
};

AssetLine.prototype.trimTrailingSpaces = function (p_text) {
  while (p_text && p_text.substring(p_text.length - 1) === ' ') {
    p_text = p_text.substring(0, p_text.length - 1);
  }

  return p_text;
};

AssetLine.prototype.parseForIcon = function (p_text, p_icons, p_skins) {
  if (!p_text) {
    return null;
  }

  var index = 0;
  var result = null;

  for (index in p_icons) {
    if (p_icons.hasOwnProperty(index)) {
      var regEx = p_icons[index];
      var n = p_text.match(regEx);

      if (n) {
        result = index;
      }
    }
  }

  return result ? p_skins[result] : null;
};

AssetLine.prototype.removeLeftIconMarkup = function (p_text) {
  return p_text.substring(p_text.indexOf(" ") + 1);
};

AssetLine.prototype.removeRightIconMarkup = function (p_text) {
  return p_text.substring(0, p_text.lastIndexOf(" "));
};

AssetLine.prototype.parseTexts = function (p_text) {
  let rightStr = null;
  let centertStr = p_text;
  const regEx = new RegExp("\\\\,", "g");
  const placeHolder = String.fromCharCode(1, 1);
  const commaIndex = p_text.replace(regEx, placeHolder).indexOf(',');

  if (commaIndex !== -1) {
    rightStr = p_text.substr(commaIndex + 1).replace(regEx, ",");
    centertStr = p_text.substr(0, commaIndex);
  }

  centertStr = centertStr.replace(regEx, ",");
  return {
    right: rightStr,
    center: centertStr
  };
};

AssetLine.prototype.makeDisplayItemFromIconDataInBbox = function (p_name, p_bbox, p_halign, p_vcenter) {
  var iconSize = this.getIconSize(p_name);

  if (iconSize.width > p_bbox.width) {
    return null;
  }

  p_halign = _balsamiq.default.isUndefined(p_halign) ? _types.default.TextAlign.LEFT : p_halign;
  p_vcenter = _balsamiq.default.isUndefined(p_vcenter) ? false : p_vcenter;
  var color = ICON_COLOR[p_name];
  var x;

  if (p_halign === _types.default.TextAlign.LEFT) {
    x = p_bbox.x;
  } else if (p_halign === _types.default.TextAlign.CENTER) {
    x = p_bbox.x + (p_bbox.width - iconSize.width) / 2;
  } else {
    x = p_bbox.x + (p_bbox.width - iconSize.width);
  }

  var y = p_vcenter ? p_bbox.y + (p_bbox.height - iconSize.height) / 2 : p_bbox.y;

  var b = _types.default.newRectangle(x, y, iconSize.width, iconSize.height);

  var d;

  if (p_name === "ON") {
    d = (0, _switch.makeSwitchDisplayItems)(this.control._context, b, true);
  } else if (p_name === "OFF") {
    d = (0, _switch.makeSwitchDisplayItems)(this.control._context, b, false);
  } else if (p_name === "ICON" || p_name === "BIGICON") {
    d = null;
  } else {
    d = _displayitem.default.createDisplayItemsFromSkin(p_name, color, b, this.control._context);
  }

  return d;
};

AssetLine.prototype.getIconSize = function (p_name) {
  if (!p_name) {
    return _types.default.newSize(0, 0);
  }

  var size;

  if (ICON_SIZE[p_name]) {
    size = ICON_SIZE[p_name];
  } else {
    size = this.control.styleSVGSize(p_name);
  }

  return size;
};

control.CONTROL_PROTOTYPE.getLinkNames = function () {
  return this._linkNames;
};

control.CONTROL_PROTOTYPE.getLinkableLabels = function (text) {
  var lines = text.split("\n");
  let linkLabels = lines.map(line => TextUtils.unformatText(line));
  return linkLabels;
};

control.CONTROL_PROTOTYPE.updateData = function () {
  var text = this.getActualProperty("text");
  var lines = text.split("\n"); // Parse texts and create measurable lines

  this._assetLines = lines.map(line => new AssetLine(line, this));
  this._linkNames = this.getLinkableLabels(text);
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var width = p_sizeProperty.width === _types.default.AUTOSIZE ? this.getActualProperty("fontSize") < 15 ? 215 : this.getActualProperty("fontSize") * 14.5 : p_sizeProperty.width;
  var minRowHeight = this.style("iphonemenu_min_row_height");
  var height = this.style("iphonemenu_padding_top");

  this._assetLines.forEach(line => {
    var rowHeight = minRowHeight;
    var leftIconSize = line.getIconSize(line.leftIcon);
    var rightIconSize = line.getIconSize(line.rightIcon);
    var centerTextSize = line.centerText ? line.centerText.measuredSize() : _types.default.newSize(0, 0);
    var rightTextSize = line.rightText ? line.rightText.measuredSize() : _types.default.newSize(0, 0);

    if (!line.separator) {
      rowHeight = Math.max(rightTextSize.height, centerTextSize.height, rightIconSize.height, leftIconSize.height, minRowHeight);
    }

    rowHeight += this.style("iphonemenu_row_padding");
    line.height = rowHeight;
    height += rowHeight;
  });

  height += this.style("iphonemenu_padding_bottom");
  return _types.default.newSize(width, height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var background = (0, _rectangle.makeFilledRectangle)(p_localBbox, this.style("iphonemenu_background_color"), 1, this.style("iphonemenu_corner_radius"));
  var border;

  if (this.getActualProperty("borderStyle") !== Props.BorderStyles.NONE) {
    border = _displayitem.default.createDisplayItemsFromSkin("rectangle_border_rounded", this.style("default_border_color"), p_localBbox, this._context);
  }

  this.pushToDisplayList(background, border);
  var iconPadding = this.style("iphonemenu_icon_padding");
  var paddingLeft = this.style("iphonemenu_padding_left");
  var paddingRight = this.style("iphonemenu_padding_right");
  var y = this.style("iphonemenu_padding_top") + p_localBbox.y;
  var firstLoop = true;
  var linkAreaBounds = [];

  this._assetLines.forEach(line => {
    var contentBbox = _types.default.newRectangle(p_localBbox.x, y, p_localBbox.width, line.height); // the link area is as big as the whole line


    linkAreaBounds.push(contentBbox); // Calculate right padding: if (right item+padding) does not fit, reduce padding

    var iconW,
        textW,
        actualPaddingRight = paddingRight;

    if (line.rightIcon) {
      iconW = line.getIconSize(line.rightIcon).width;
      actualPaddingRight = Math.max(Math.min(paddingRight, contentBbox.width - iconW), 0);
      contentBbox = _types.default.paddedRectangle(contentBbox, 0, 0, 0, actualPaddingRight);
    } else if (line.rightText) {
      textW = line.rightText.measuredSize().width;
      actualPaddingRight = Math.max(Math.min(paddingRight, contentBbox.width - textW), 0);
      contentBbox = _types.default.paddedRectangle(contentBbox, 0, 0, 0, actualPaddingRight);
    } else {
      contentBbox = _types.default.paddedRectangle(contentBbox, 0, 0, 0, paddingRight);
    } // Right Icon starts on the right at paddingRight distance from right border and can eat up all space on its left up to the left border


    if (line.rightIcon) {
      this.pushToDisplayList(line.makeDisplayItemFromIconDataInBbox(line.rightIcon, contentBbox, _types.default.TextAlign.RIGHT, true));
      contentBbox = _types.default.paddedRectangle(contentBbox, 0, 0, 0, iconW + iconPadding);
    } // Right text starts on the right at paddingright from the border or at left of right icon


    if (line.rightText && contentBbox.width > 0) {
      this.pushToDisplayList(line.rightText.makeDisplayItems(contentBbox, _types.default.TextAlign.RIGHT, _types.default.Positions.CENTER));
      contentBbox = _types.default.paddedRectangle(contentBbox, 0, 0, 0, textW);
    } // Add padding left to draw left side


    contentBbox = _types.default.paddedRectangle(contentBbox, 0, 0, paddingLeft); // Left icon starts on the left at paddingleft from the left border (padding is not reduce if icon does not fit!)

    if (line.leftIcon) {
      // Left icon must have a larger bbox since it can overlap right icon/text
      var leftIconBbox = _types.default.newRectangle(contentBbox.x, contentBbox.y, p_localBbox.width - paddingLeft, contentBbox.height);

      this.pushToDisplayList(line.makeDisplayItemFromIconDataInBbox(line.leftIcon, leftIconBbox, _types.default.TextAlign.LEFT, true));
      contentBbox = _types.default.paddedRectangle(contentBbox, 0, 0, line.getIconSize(line.leftIcon).width + iconPadding);
    } // Center text starts on the left at paddingleft or after the left icon and runs up to right bbox


    if (line.centerText && contentBbox.width > 1) {
      this.pushToDisplayList(line.centerText.makeDisplayItems(contentBbox, _types.default.TextAlign.LEFT, _types.default.Positions.CENTER));
    } // Draw horizontal lines


    if (!firstLoop) {
      var path = Path.newPath(p_localBbox.width, p_localBbox.height);
      path.moveTo(p_localBbox.x + 1, y + 0.5);
      path.lineTo(p_localBbox.x + p_localBbox.width - 1, y + 0.5);
      this.pushToDisplayList(_displayitem.default.newSVGPath(p_localBbox, path, 1, 0, null, null, null));
    } else {
      firstLoop = false;
    }

    y += line.height;
  });

  _link.default.processLinks(this, linkAreaBounds);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/iphonepicker.js":
/*!******************************************!*\
  !*** ./js/core/controls/iphonepicker.js ***!
  \******************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::iPhonePicker",
  controlTypeName: "iOS Picker",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.TEXT({
      defaultText: "7,59\n8,00\n9,01, AM\n10,02, PM\n11,03\n12,04\n1,05",
      multiline: true,
      splitOn: Props.SplitOnTypes.NEWLINE
    })
  }
});

control.CONTROL_PROTOTYPE.measure = function () {
  return _types.default.newSize(214, 160);
};

control.CONTROL_PROTOTYPE.updateData = function () {
  this._data = {
    lines: [],
    columnCount: 0,
    maxLineWidth: 0,
    columnWidths: []
  };
  var text = this.getActualProperty("text");
  var lines = text.split("\n").slice(0, 7);
  var textStyle1 = this.newTextStyle();

  var textStyle2 = _balsamiq.default.extend(_balsamiq.default.clone(textStyle1), {
    textColor: 0x777777,
    fontSize: 12
  });

  var textStyle3 = _balsamiq.default.extend(_balsamiq.default.clone(textStyle1), {
    textColor: 0x999999,
    fontSize: 11
  });

  var textStyle4 = _balsamiq.default.extend(_balsamiq.default.clone(textStyle1), {
    textColor: 0xcccccc,
    fontSize: 10
  });

  var textStyles = [textStyle4, textStyle3, textStyle2, textStyle1, textStyle2, textStyle3, textStyle4];
  lines.forEach((line, lineIndex) => {
    var cellTexts = TextUtils.splitTextOnCommas(line);
    var cells = [];
    var lineWidth = 0;
    cellTexts.forEach((cellText, columnIndex) => {
      cellText = cellText.trim();
      var cell;

      if (line) {
        var t = textMod.newTextNoWrap(cellText, textStyles[lineIndex], this._context);
        var s = t.measuredSize();
        cell = {
          text: t,
          width: s.width
        };
      } else {
        cell = {
          text: null,
          width: 2
        };
      }

      cells.push(cell);

      if (columnIndex < this._data.columnWidths.length) {
        this._data.columnWidths[columnIndex] = Math.max(this._data.columnWidths[columnIndex], cell.width);
      } else {
        this._data.columnWidths.push(cell.width);
      }

      lineWidth += cell.width;
    });

    this._data.lines.push({
      cells: cells,
      cellsWidth: lineWidth
    });

    this._data.columnCount = Math.max(this._data.columnCount, cells.length);
  });

  this._data.columnWidths.forEach(columnWidth => {
    this._data.maxLineWidth += columnWidth;
  });
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var COLUMN_PADDING = 6;
  var HORIZONTAL_PADDING = 7;
  var ROW_HEIGHT = 22;
  var gridMeasuredWidth = HORIZONTAL_PADDING + this._data.maxLineWidth + Math.max(this._data.columnCount - 1, 0) * COLUMN_PADDING + HORIZONTAL_PADDING;
  var gridWidth = Math.max(Math.min(p_localBbox.width, gridMeasuredWidth), 144);
  var factor = Math.max(1, gridWidth / gridMeasuredWidth);
  this.pushToDisplayList((0, _rectangle.makeRectangleDisplayItems)(this._context, p_localBbox, Props.BorderStyles.NONE, 0xFFFFFF, 1, 0xFFFFFF));
  var y = 8;

  this._data.lines.forEach(line => {
    var x = (p_localBbox.width - gridWidth) / 2 + HORIZONTAL_PADDING * factor;
    line.cells.forEach((cell, columnIndex) => {
      var cellWidth = this._data.columnWidths[columnIndex] * factor;

      if (cell.text) {
        this.pushToDisplayList(cell.text.makeDisplayItems(_types.default.newRectangle(x, y, cellWidth, ROW_HEIGHT), _types.default.TextAlign.CENTER));
      }

      x += cellWidth + COLUMN_PADDING * factor;
    });
    y += ROW_HEIGHT;
  });

  var assetHeight = this.styleSVGSize("rule_horizontal").height;
  this.pushToDisplayList(_displayitem.default.createDisplayItemsFromSkin("rule_horizontal", this.style("iphonepicker_line_color"), _types.default.newRectangle(0, 68 - assetHeight / 2, p_localBbox.width, assetHeight), this._context));
  this.pushToDisplayList(_displayitem.default.createDisplayItemsFromSkin("rule_horizontal", this.style("iphonepicker_line_color"), _types.default.newRectangle(0, 90 - assetHeight / 2, p_localBbox.width, assetHeight), this._context));
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/label.js":
/*!***********************************!*\
  !*** ./js/core/controls/label.js ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _macro = __webpack_require__(/*! core/bricks/macro */ "./js/core/bricks/macro.js");

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var label = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::Label",
  controlTypeName: "Label",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.ONLY_WIDTH,
        resize: Props.ResizeOptions.HORIZONTAL
      }
    }),
    state: {
      defaultValue: Props.BoolStates.NORMAL,
      enum: Props.BoolStates
    },
    align: {
      defaultValue: Props.TextAlignments.LEFT,
      enum: Props.TextAlignments
    },
    textOrientation: {
      defaultValue: Props.TextOrientations.RIGHT,
      enum: Props.TextOrientations
    },
    ...Props.PropertyDefinitions.link
  }
});
controlsMod.addStandardTextPropertyDefinitions(label, {
  text: "Some text",
  charCountEnabled: true,
  noEmptyText: true
});
controlsMod.addColorPropertyDefinitions(label, {
  defaultTextColor: "default_text_color",
  oldTextColorName: 'color'
});
controlsMod.addTextClipBBoxCalculation(label);
controlsMod.addStandardCalculateOriginForNewTextOrSize(label);

label.CONTROL_PROTOTYPE.initData = function () {
  this._clipOnBbox = false;
  this.overflow = "visible";
};

label.CONTROL_PROTOTYPE.updateData = function () {
  const text = (0, _macro.processMacros)(this);
  const textColor = this.getActualProperty('state') === _types.default.ControlState.NORMAL ? this.getActualProperty("textColor") : this.style('default_disabled_color');
  const textStyle = this.newStandardTextStyle({
    textColor
  });

  if (!this.text || !(0, _text.areTextsEqual)(this.text, text, textStyle)) {
    this.text = (0, _text.newSingleLineText)(text, textStyle, this._context);
  }
};

_link.default.addTextLinksMethods(label);

label.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var textSize = this.text.measuredSize();
  var autosize = p_sizeProperty.width === _types.default.AUTOSIZE;
  var naturalWidth = textSize.width;
  var naturalHeight = textSize.height + this.style('label_padding_top') + this.style('label_padding_bottom');
  var width = autosize ? naturalWidth : p_sizeProperty.width;
  var height = naturalHeight;
  var orientation = this.getActualProperty("textOrientation");

  if (orientation === _types.default.TextOrientation.UP || orientation === _types.default.TextOrientation.DOWN) {
    if (autosize) {
      width = naturalHeight;
      height = naturalWidth;
    } else {
      width = p_sizeProperty.width;
      height = naturalWidth;
    }
  } else if (orientation === _types.default.TextOrientation.SLANTED) {
    var t1 = naturalWidth,
        t2 = textSize.height;
    var c = Math.cos(Math.PI / 12),
        s = Math.sin(Math.PI / 12);

    if (autosize) {
      width = Math.round(t1 * c + t2 * s);
    } else {
      width = p_sizeProperty.width;
    }

    height = Math.round(t1 * s + t2 * c);
  }

  return _types.default.newSize(width, height);
};

label.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this);

  var orientation = this.getActualProperty("textOrientation");

  if (orientation === _types.default.TextOrientation.RIGHT) {
    var bbox = _balsamiq.default.clone(p_localBbox);

    bbox.y = Math.max((p_localBbox.height - this.text.measuredSize().height) / 2, 0);
    bbox.height = this.text.measuredSize().height;
    var obj = this.text.makeDisplayItemsAndLinks(bbox, this.getActualProperty('align'));
    this.pushToDisplayList(obj.displayItems);

    _link.default.processTextLinks(this, obj.links);
  } else {
    var x = 0,
        y = 0;
    var angle = 0;

    if (orientation === _types.default.TextOrientation.UP) {
      x = 0;
      y = p_localBbox.height;
      angle = 90;
    } else if (orientation === _types.default.TextOrientation.DOWN) {
      x = p_localBbox.width;
      y = 0;
      angle = 270;
    } else if (orientation === _types.default.TextOrientation.SLANTED) {
      x = 0; //this.text.measuredSize().height * Math.sin(Math.PI / 12);

      y = p_localBbox.height - this.text.measuredSize().height;
      angle = 15;
    }

    let bounds = _types.default.newRectangle(x, y, this.text.measuredSize().width, this.text.measuredSize().height);

    let obj = this.text.makeDisplayItemsAndLinks(bounds, this.getActualProperty('align'));

    const adjustXYForOrientation = (x, y) => {
      switch (orientation) {
        case _types.default.TextOrientation.DOWN:
          {
            return {
              y: p_localBbox.y + (x - p_localBbox.width),
              x: p_localBbox.width - (y - p_localBbox.y)
            };
          }

        case _types.default.TextOrientation.UP:
          {
            return {
              y: p_localBbox.height - (x - p_localBbox.x),
              x: p_localBbox.x + (y - p_localBbox.height - p_localBbox.y)
            };
          }

        default:
          return {
            x,
            y
          };
      }
    };

    obj.displayItems.forEach(item => {
      const adjustedXY = adjustXYForOrientation(item.x, item.y);
      item.x = adjustedXY.x;
      item.y = adjustedXY.y;
    });

    _displayitem.default.setRotation(obj.displayItems, angle);

    for (var i = 0; i < obj.links.length; i += 1) {
      var newPt = _types.default.rotatePtDeg(bounds, obj.links[i].bounds, angle);

      obj.links[i].bounds.x = newPt.x;
      obj.links[i].bounds.y = newPt.y;
      obj.links[i].angle = angle;
    }

    _link.default.processTextLinks(this, obj.links);

    this.pushToDisplayList(obj.displayItems);
  }
};

var _default = label;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/linechart.js":
/*!***************************************!*\
  !*** ./js/core/controls/linechart.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::LineChart",
  controlTypeName: "Chart: Line",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(187, 175),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    })
  }
});

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var w = p_localBbox.width,
      h = p_localBbox.height;
  var linesThickness = Math.min(this.style("linechart_max_thickness"), Math.max(this.style("linechart_min_thickness"), w / 40, h / 40));
  var line1 = Path.newPath(w, h);
  line1.moveTo(5, h - 5);
  line1.lineTo(w / 3, h / 3);
  line1.lineTo(w * 2 / 3, h / 2);
  line1.lineTo(w - 15, 25);

  var line1Item = _displayitem.default.newSVGPath(p_localBbox, line1, linesThickness, this.style("linechart_series1_color"));

  line1Item.linecap = _displayitem.default.LineCap.ROUND;
  var line2 = Path.newPath(w, h);
  line2.moveTo(5, h * 123 / 173);
  line2.lineTo(w / 3, h * 3 / 5);
  line2.lineTo(w * 2 / 3, h / 5);
  line2.lineTo(w - 15, h / 3);

  var line2Item = _displayitem.default.newSVGPath(p_localBbox, line2, linesThickness, this.style("linechart_series2_color"));

  line2Item.linecap = _displayitem.default.LineCap.ROUND;
  var axisThickness = this.style("linechart_axis_thickness");
  var a2 = axisThickness / 2;
  var axis = Path.newPath(w, h);
  axis.moveTo(a2, a2);
  axis.lineTo(a2, h - a2);
  axis.lineTo(w - a2, h - a2);

  var axisItem = _displayitem.default.newSVGPath(p_localBbox, axis, axisThickness, this.style("linechart_axis_color"));

  axisItem.linecap = _displayitem.default.LineCap.ROUND;
  this.pushToDisplayList(line1Item, line2Item, axisItem);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/lineoftext.js":
/*!****************************************!*\
  !*** ./js/core/controls/lineoftext.js ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

const ALPHA = 0.5;
var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::LineOfText",
  controlTypeName: "Line of Text",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(200, 27),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.QUICKDRAW_TIP("You can draw Lines of Text quickly by holding down the *Y key* and dragging. [Learn more]($DOCURL)", "You can also draw Lines of Text with your fingers using the Quick Draw Toolbar (in the View Menu). [Learn more]($DOCURL)")
  }
});

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  if (this._context.skin["#skinName"] === "sketch") {
    var width = Math.min(this.styleSVGSize("lineOfText").width, Math.max(this.styleSVGSize("lineOfText").width / this.styleSVGSize("lineOfText").height * p_localBbox.height, p_localBbox.width));

    var bounds = _types.default.newRectangle(0, 0, width, p_localBbox.height);

    let maskID = this.addMaskWithRect(p_localBbox);

    let path = _displayitem.default.newPathsAsset(this._context, "lineOfText", bounds);

    _displayitem.default.maskDisplayItems(path, maskID);

    this.pushToDisplayList(path);
  } else {
    var line = (0, _rectangle.makeFilledRectangle)(p_localBbox, this.style("default_text_color"), ALPHA);
    this.pushToDisplayList(line);
  }
};

control.makeDisplayListForPreview = function (p_context, p_skin, p_localBbox) {
  if (p_skin['#skinName'] === "sketch") {
    let skinVal = p_skin["lineOfText"];
    let width = Math.min(skinVal.width, Math.max(skinVal.width / skinVal.height * p_localBbox.height, p_localBbox.width));

    let bounds = _types.default.newRectangle(p_localBbox.x, p_localBbox.y, width, p_localBbox.height);

    return _displayitem.default.newPathsAsset(p_context, "lineOfText", bounds);
  } else {
    let bounds = _types.default.newRectangle(p_localBbox.x, p_localBbox.y, p_localBbox.width, p_localBbox.height);

    return (0, _rectangle.makeFilledRectangle)(bounds, p_skin["default_text_color"], ALPHA);
  }
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/link.js":
/*!**********************************!*\
  !*** ./js/core/controls/link.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::Link",
  controlTypeName: "Link",
  controlAlternateNames: [],
  props: {
    state: {
      defaultValue: Props.BoolStates.NORMAL,
      enum: Props.BoolStates
    },
    ...Props.PropertyDefinitions.link
  }
});
controlsMod.addStandardTextPropertyDefinitions(control, {
  text: "a link",
  underline: true,
  noEmptyText: true
});

control.CONTROL_PROTOTYPE.initData = function () {
  this._clipOnBbox = false;
  this.overflow = "visible";
};

control.CONTROL_PROTOTYPE.updateData = function () {
  var color = this.getActualProperty('state') === _types.default.ControlState.NORMAL ? this.style('default_link_color') : this.style('default_disabled_color');
  var textStyle = this.newStandardTextStyle({
    textColor: color
  });
  this.text = textMod.newText(this.getActualProperty("text"), textStyle, this._context);
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var textSize = this.text.measuredSize();
  return _types.default.newSize(textSize.width, textSize.height + this.style("link_padding_top") + this.style("link_padding_bottom"));
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this);

  var bbox = _types.default.newRectangle(0, this.style("link_padding_top"), p_localBbox.width, p_localBbox.height);

  this.pushToDisplayList(this.text.makeDisplayItems(bbox, _types.default.TextAlign.LEFT));
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/linkbar.js":
/*!*************************************!*\
  !*** ./js/core/controls/linkbar.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::LinkBar",
  controlTypeName: "Link Bar",
  controlAlternateNames: ["Navigation"],
  props: {
    borderColor: {
      defaultValue: 0x000000,
      dataType: 'int',
      label: 'Separator and Selected Text Color',
      options: {
        oldName: 'borderColor'
      }
    },
    ...Props.PropertyDefinitions.links()
  }
});
controlsMod.addStandardTextPropertyDefinitions(control, {
  text: "Home, Products, Company, Blog",
  underline: true,
  splitOn: Props.SplitOnTypes.COMMA,
  noEmptyText: true
});
controlsMod.addColorPropertyDefinitions(control, {
  defaultTextColor: "default_link_color",
  oldTextColorName: 'color'
});
controlsMod.addSelectionPropertyDefinitions(control, -1);

control.CONTROL_PROTOTYPE.getSelectableLabels = function () {
  return this._selectableLabels;
};

control.CONTROL_PROTOTYPE.getLinkNames = function () {
  return this._selectableLabels;
};

control.CONTROL_PROTOTYPE.getLinkableLabels = function (text) {
  var lines = TextUtils.splitTextOnCommas(text);
  let linkLabels = lines.map(line => TextUtils.unformatText(line));
  return linkLabels;
};

control.CONTROL_PROTOTYPE.initData = function () {
  this._clipOnBbox = false;
  this.overflow = "visible";
};

control.CONTROL_PROTOTYPE.updateData = function () {
  var linkTextStyle = this.newStandardTextStyle({
    textColor: this.getActualProperty("textColor")
  });
  var selectedTextStyle = this.newStandardTextStyle({
    textColor: this.getActualProperty("borderColor"),
    underline: false
  });
  var selectedIndex = this.getActualProperty("selectedIndex");
  var labels = TextUtils.splitTextOnCommas(this.getActualProperty("text"));
  this.texts = labels.map((label, index) => {
    return textMod.newText(label.trim(), index === selectedIndex ? selectedTextStyle : linkTextStyle, this._context);
  });
  this._selectableLabels = this.getLinkableLabels(this.getActualProperty("text"));
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var width = 0,
      height = 0;
  this.texts.forEach(text => {
    width += text.measuredSize().width;
    height = Math.max(height, text.measuredSize().height);
  });
  width += (this.texts.length - 1) * (this.style("linkbar_padding_middle") * 2 + this.styleSVGSize("line_vertical").width);
  height += this.style("linkbar_padding_top") + this.style("linkbar_padding_bottom");
  return _types.default.newSize(width, height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var paddingTop = this.style("linkbar_padding_top");
  var paddingMiddle = this.style("linkbar_padding_middle");
  var separatorSize = this.styleSVGSize("line_vertical");
  var separatorColor = this.getActualProperty("borderColor");
  var labels = [];
  var separators = [];
  var x = 0;
  var linkBounds = [];
  this.texts.forEach((text, index) => {
    var bounds = _types.default.newRectangle(x, paddingTop, text.measuredSize().width, text.measuredSize().height);

    labels.push(text.makeDisplayItems(bounds));
    linkBounds.push(bounds);

    if (index !== this.texts.length - 1) {
      x += text.measuredSize().width + paddingMiddle;
      separators.push(_displayitem.default.createDisplayItemsFromSkin("line_vertical", separatorColor, _types.default.newRectangle(x, 0, separatorSize.width, p_localBbox.height), this._context));
      x += separatorSize.width + paddingMiddle;
    }
  });
  this.pushToDisplayList(labels, separators);

  _link.default.processLinks(this, linkBounds);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/list.js":
/*!**********************************!*\
  !*** ./js/core/controls/list.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _scrollbars = __webpack_require__(/*! core/bricks/scrollbars */ "./js/core/bricks/scrollbars.js");

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::List",
  controlTypeName: "List",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.BORDER_STYLE(Props.BorderStyles.SQUARE, Props.SimpleBorderStyles),
    rowHeight: {
      defaultValue: 21,
      dataType: 'int'
    },
    alternateRowColor: {
      defaultValue: 0xFFFFFF,
      label: 'Alternate Row',
      dataType: 'int'
    },
    hasHeader: {
      defaultValue: false,
      dataType: 'bool'
    },
    ...Props.PropertyDefinitions.SCROLLBAR(),
    ...Props.PropertyDefinitions.ICONS(),
    ...Props.PropertyDefinitions.links()
  }
});
controlsMod.addStandardTextPropertyDefinitions(control, {
  text: "Item One\nItem Two\nItem Three",
  multiline: true,
  splitOn: Props.SplitOnTypes.NEWLINE
});
controlsMod.addColorPropertyDefinitions(control, {
  defaultColor: 0xFFFFFF
});
controlsMod.addSelectionPropertyDefinitions(control, -1);

_link.default.addMultiLinksMethods(control);

control.CONTROL_PROTOTYPE.getSelectableLabels = function () {
  return this._selectableLabels;
};

control.CONTROL_PROTOTYPE.getLinkNames = function () {
  return this.getLinkableLabels(this.getActualProperty("text"));
};

control.CONTROL_PROTOTYPE.getLinkableLabels = function (text) {
  let linkLabels = text.split("\n").map(label => {
    return TextUtils.unformatText(label);
  });
  return linkLabels;
};

control.CONTROL_PROTOTYPE.updateSelectionOnTextChange = function (control, newText, oldText, oldSelection) {
  let newTextLines = control.calculateSelectionLabelsForEditText(newText);
  let offset = this.getActualProperty("hasHeader") ? 1 : 0;

  if (oldSelection === -1) {
    return -1;
  }

  let oldSelectedText = oldText[oldSelection + offset];

  if (newTextLines.indexOf(oldSelectedText) > -1) {
    //the exact same text has been moved
    return newTextLines.indexOf(oldSelectedText) - offset;
  } else {
    if (newTextLines.length == oldText.length) {
      // the text has been changed but the control still has the same number of lines so we keep the same selection
      return oldSelection;
    } else {
      return control.getDefaultSelectionValue();
    }
  }
};

control.CONTROL_PROTOTYPE.updateData = function () {
  this._selectableLabels = [];
  this._linkNames = [];
  this.maxWidth = 0;
  const hasHeader = this.getActualProperty("hasHeader");
  const shift = hasHeader ? 1 : 0;
  const lines = this.getActualProperty("text").split("\n");
  const isPreviousTextOfDifferentLength = !this.texts || this.texts.length !== lines.length;
  this.texts = lines.map((line, index) => {
    let textStyle;

    if (hasHeader && index === 0) {
      textStyle = this.newStandardTextStyle();
    } else {
      const backgroundColor = this.getActualProperty("selectedIndex") + shift == index ? this.style("default_selection_color") : (index + shift) % 2 !== 0 ? this.getActualProperty("alternateRowColor") : this.getActualProperty("color");

      const textColor = _colors.default.labelColorGivenBackgroundColor(backgroundColor, 1, this.style("default_text_color"));

      const linkColor = _colors.default.labelColorGivenBackgroundColor(backgroundColor, 1, this.style("default_link_color"));

      textStyle = this.newStandardTextStyle({
        textColor,
        linkColor
      });
    }

    const text = isPreviousTextOfDifferentLength || !(0, _text.areTextsEqual)(this.texts[index], line, textStyle) ? (0, _text.newSingleLineText)(line, textStyle, this._context) : this.texts[index];
    this.maxWidth = Math.max(this.maxWidth, text.measuredSize().width);
    return text;
  }, this);
  this._selectableLabels = this.getLinkableLabels(this.getActualProperty("text"));

  if (hasHeader) {
    this._selectableLabels.shift();
  }

  let icons = this.getActualProperty("icons") || [];
  let dependencies = icons.map(icon => function () {
    if (icon.src != undefined) {
      return icon.src;
    }
  });
  this.setDependencies(dependencies);

  _balsamiq.default.assert(this.texts.length > 0);
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProp) {
  var measuredHeight = this.style("list_padding_top") + this.style("list_padding_bottom");
  var rowH = Math.max(this.getActualProperty("fontSize") + this.style("list_line_padding"), this.getActualProperty("rowHeight"));
  let icons = this.getActualProperty("icons");
  let iconsFound = !!icons && icons.some(icon => icon && icon.size);
  let iconOffset = iconsFound ? this.getActualProperty("fontSize") + this.style("list_icon_padding") : 0;

  if (this.getActualProperty("hasHeader") && this.texts.length > 0) {
    measuredHeight += this.texts[0].measuredSize().height + 6 + (this.texts.length - 2) * rowH; // -2 is Flex bug: it counts one row less!!!
  } else {
    measuredHeight += this.texts.length * rowH;
  }

  return _types.default.newSize(Math.max(100, this.maxWidth + this.style("list_padding_left") + this.style("list_padding_right") + iconOffset), Math.max(126, measuredHeight));
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  this.iconsFound = false;
  var icons = this.getActualProperty("icons");
  this.iconsFound = !!icons && icons.some(icon => icon && icon.size);
  var fontSize = this.getActualProperty("fontSize");
  var rowHeight = Math.max(this.getActualProperty("fontSize") + this.style("list_line_padding"), this.getActualProperty("rowHeight"));
  var paddingTop = this.style("list_padding_top");
  var paddingLeft = this.style("list_padding_left");
  var paddingRight = this.style("list_scrollbar_padding_right");
  var hasHeader = this.getActualProperty("hasHeader");
  var selectedIndex = this.getActualProperty("selectedIndex");
  var selectionColor = this.style("default_selection_color");
  var theX = this.iconsFound ? fontSize + this.style("list_icon_padding") : paddingLeft;
  var bounds;
  var color;
  var color1 = this.getActualProperty("color");
  var color2 = this.getActualProperty("alternateRowColor");
  var useColor1 = true;
  var linkBounds = [];
  var backgrounds = [];
  var header;
  var headerHeight = 0;

  if (hasHeader) {
    headerHeight = paddingTop + 6 + this.texts[0].measuredSize().height;
    bounds = _types.default.newRectangle(0, 0, p_localBbox.width, headerHeight);
    header = (0, _rectangle.makeRectangleBorderDisplayItems)(this._context, bounds);
    backgrounds.push((0, _rectangle.makeFilledRectangle)(bounds, this.style("list_header_color")));
    linkBounds.push(bounds);
  } else {
    color = selectedIndex === 0 ? selectionColor : color1;
    bounds = _types.default.newRectangle(0, 0, p_localBbox.width, paddingTop);
    backgrounds.push((0, _rectangle.makeFilledRectangle)(bounds, color));
  }

  var y = bounds.height;
  var index = 0;

  while (y < p_localBbox.height) {
    bounds = _types.default.newRectangle(0, y, p_localBbox.width, Math.min(rowHeight, p_localBbox.height - y));
    color = selectedIndex === index ? selectionColor : useColor1 ? color1 : color2;
    backgrounds.push((0, _rectangle.makeFilledRectangle)(bounds, color));
    linkBounds.push(bounds);
    useColor1 = !useColor1;
    y += rowHeight;
    index += 1;
  }

  _link.default.processLinks(this, linkBounds);

  var border;

  if (this.getActualProperty("borderStyle") === Props.BorderStyles.SQUARE) {
    border = (0, _rectangle.makeRectangleBorderDisplayItems)(this._context, p_localBbox);
  }

  y = paddingTop;
  var textItems = [];
  var iconItems = []; //creating the textItems and the Icons if we have them
  // this.maxWidth = 0;

  this.texts.forEach((text, index) => {
    var textY = hasHeader && index === 0 ? y + 3 : y + (rowHeight - text.measuredSize().height) / 2;
    textItems = textItems.concat(text.makeDisplayItems(_types.default.newRectangle(theX, textY, p_localBbox.width - paddingLeft - paddingRight, text.measuredSize().height))); // this.maxWidth = Math.max(this.maxWidth, theX + text.measuredSize().width);

    if (this.iconsFound) {
      var icon = icons[index];

      if (icon && icon.size) {
        //icon has icon data. size is null when there is no icon
        var iconProp = _icon.default.processIconProperty(icon, 0);

        iconProp.iconData.size = fontSize; //changing the size to vary as the textSize

        iconItems.push(_icon.default.makeDisplayItemsFromIconData(this, iconProp.iconData, paddingLeft, textY, text.textStyle.textColor));
      }
    }

    y += hasHeader && index === 0 ? headerHeight - paddingTop : rowHeight;
  });
  var scrollbar;

  if (this.getActualProperty("verticalScrollbar") && p_localBbox.width >= _scrollbars.Sizes.ScrollbarFixedSize) {
    scrollbar = (0, _scrollbars.makeVerticalScrollbar)(p_localBbox, hasHeader ? headerHeight + this.style("list_scrollbar_header_padding_top") : this.getActualProperty("borderStyle") == Props.BorderStyles.SQUARE ? this.style("list_scrollbar_padding_top") : this.style("list_scrollbar_padding_no_border"), this.getActualProperty("borderStyle") == Props.BorderStyles.SQUARE ? this.style("list_scrollbar_padding_bottom") : this.style("list_scrollbar_padding_no_border"), this.getActualProperty("borderStyle") == Props.BorderStyles.SQUARE ? this.style("list_scrollbar_padding_right") : this.style("list_scrollbar_padding_no_border"), this.getActualProperty("value"), _types.default.ControlState.NORMAL, this._context);
  }

  this.pushToDisplayList(backgrounds, header, border, textItems, scrollbar, iconItems);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/menu.js":
/*!**********************************!*\
  !*** ./js/core/controls/menu.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var Menu = _interopRequireWildcard(__webpack_require__(/*! core/bricks/menu */ "./js/core/bricks/menu.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::Menu",
  controlTypeName: "Menu",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.HORIZONTAL
      }
    }),
    ...Props.PropertyDefinitions.TEXT({
      defaultText: "Open,CTRL+O\n" + "Open Recent >\n" + "---\n" + "o Option One\n" + "Option Two\n" + "=\n" + "x Toggle Item\n" + "-Disabled Item-\n" + "Exit,CTRL+Q",
      multiline: true,
      splitOn: Props.SplitOnTypes.NEWLINE
    }),
    ...Props.PropertyDefinitions.links()
  }
});
Controls.addSelectionPropertyDefinitions(control, -1);
Controls.addStandardTextPropertyDefinitions(control);

control.CONTROL_PROTOTYPE.getSelectableLabels = function () {
  return this._selectableLabels;
};

control.CONTROL_PROTOTYPE.getLinkNames = function () {
  return this._selectableLabels;
};

control.CONTROL_PROTOTYPE.getLinkableLabels = function (text) {
  var lines = text.split("\n");
  let linkLabels = lines.map(line => TextUtils.unformatText(line));
  return linkLabels;
};

control.CONTROL_PROTOTYPE.initData = function () {
  this._menuLines = [];
  Menu.menuInit(this.style("menu_padding_top"), this.style("menu_padding_bottom"), this.style("menu_padding_middle"));
};

control.CONTROL_PROTOTYPE.updateData = function () {
  const text = this.getActualProperty("text");
  const lines = text.split("\n");
  this._selectableLabels = this.getLinkableLabels(text); // Parse texts and create measurable lines

  const isSameLengthOfPreviousState = this._menuLines && this._menuLines.length === lines.length;
  this._menuLines = lines.map((line, index) => {
    const isSelected = this.getActualProperty("selectedIndex") === index;
    const previousLine = isSameLengthOfPreviousState ? this._menuLines[index] : null;
    return Menu.makeMenuLine(this, line, isSelected, null, true, previousLine);
  }, this);
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  Menu.menuInit(this.style("menu_padding_top"), this.style("menu_padding_bottom"), this.style("menu_padding_middle"));
  return _types.default.newSize(p_sizeProperty.width === _types.default.AUTOSIZE ? 150 : p_sizeProperty.width, Menu.measureMenu(this, p_sizeProperty.width).height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var background = Rectangle.makeFilledRectangle(p_localBbox, 0xFFFFFF);
  this.pushToDisplayList(background, Menu.makeMenuDisplayItems(this, p_localBbox, this._menuLines, this.style("menu_padding_top"), this.getActualProperty("selectedIndex")));
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/menubar.js":
/*!*************************************!*\
  !*** ./js/core/controls/menubar.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::MenuBar",
  controlTypeName: "Menu Bar",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.TEXT({
      defaultText: "File, Edit, View, Help",
      multiline: false
    }),
    ...Props.PropertyDefinitions.BORDER_STYLE(Props.BorderStyles.SQUARE, Props.SimpleBorderStyles),
    ...Props.PropertyDefinitions.links(),
    ...Props.PropertyDefinitions.ROTATION({
      isIcon: true
    })
  }
});
Controls.addStandardTextPropertyDefinitions(control, {
  splitOn: Props.SplitOnTypes.COMMA
});
Controls.addSelectionPropertyDefinitions(control, -1);

control.CONTROL_PROTOTYPE.getSelectableLabels = function () {
  return this._selectableLabels;
};

control.CONTROL_PROTOTYPE.getLinkNames = function () {
  return this._selectableLabels;
};

control.CONTROL_PROTOTYPE.getLinkableLabels = function (text) {
  const lines = TextUtils.splitTextOnCommas(text);
  return lines.map(TextUtils.unformatText);
};

control.CONTROL_PROTOTYPE.updateData = function () {
  const text = this.getActualProperty("text");
  this._selectableLabels = this.getLinkableLabels(text);
  const labels = TextUtils.splitTextOnCommas(text);

  if (labels.length === 1) {
    labels.push("...");
  }

  const isPreviousTextOfDifferentLength = !this.texts || this.texts.length !== labels.length;
  this.texts = labels.map((label, index) => {
    const color = index === this.getActualProperty("selectedIndex") ? this.style("default_selection_color") : 0xFFFFFF;
    const textStyle = this.newStandardTextStyle({
      textColor: _colors.default.labelColorGivenBackgroundColor(color, 1, this.style("default_text_color")),
      linkColor: _colors.default.labelColorGivenBackgroundColor(color, 1, this.style("default_link_color"))
    });
    return isPreviousTextOfDifferentLength || !Text.areTextsEqual(this.texts[index], label, textStyle) ? Text.newTextNoWrap(label, textStyle, this._context) : this.texts[index];
  }, this);

  _icon.default.processIconData(this);
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var entryPadding = this.style("menubar_entry_padding");
  var width = this.style("menubar_padding_left") + this.style("menubar_padding_right") + (this.iconData ? this.iconData.size + 2 * entryPadding : 0);
  var height = 0;
  this.texts.forEach(text => {
    width += text.measuredSize().width + 2 * entryPadding;
    height = Math.max(height, text.measuredSize().height);
  });
  var svgSize = this.styleSVGSize("rectangle_border_square");
  height = Math.max(height, this.iconData ? this.iconData.size : 0, svgSize.height);
  /* icon */

  height += this.style("menubar_padding_top") + this.style("menubar_padding_bottom");
  return _types.default.newSize(width, height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var background = Rectangle.makeFilledRectangle(p_localBbox, 0xFFFFFF);
  var border;

  if (this.getActualProperty("borderStyle") === Props.BorderStyles.SQUARE) {
    border = _displayitem.default.createDisplayItemsFromSkin("rectangle_border_square", this.style("default_border_color"), p_localBbox, this._context);
  }

  var x = this.style("menubar_padding_left");
  var icon;

  if (this.iconData) {
    x += this.style("menubar_entry_padding");
    icon = _icon.default.makeDisplayItemsFromIconData(this, this.iconData, x, (p_localBbox.height - this.iconData.size) / 2, this.style("default_border_color"));
    x += this.iconData.size + this.style("menubar_entry_padding");
  }

  var selection;
  var labels = [];
  var linkBounds = [];
  this.texts.forEach((text, index) => {
    if (index === this.getActualProperty("selectedIndex")) {
      selection = Rectangle.makeFilledRectangle(_types.default.newRectangle(x, 0, text.measuredSize().width + 2 * this.style("menubar_entry_padding"), p_localBbox.height), this.style("default_selection_color"));
    }

    x += this.style("menubar_entry_padding");
    var availableWidth = p_localBbox.width - x - this.style("menubar_padding_right");

    if (availableWidth >= text.measuredSize().width) {
      var bounds = _types.default.newRectangle(x, this.style("menubar_padding_top"), text.measuredSize().width, p_localBbox.height - (this.style("menubar_padding_top") + this.style("menubar_padding_bottom")));

      labels.push(text.makeDisplayItems(bounds, _types.default.TextAlign.LEFT, _types.default.Positions.CENTER));

      var linkBound = _balsamiq.default.clone(bounds);

      linkBound.x = Math.max(0, x - this.style("menubar_padding_right"));
      linkBound.width += 2 * this.style("menubar_padding_right");
      linkBounds.push(linkBound);
      x += text.measuredSize().width + this.style("menubar_entry_padding");
    }
  });

  _link.default.processLinks(this, linkBounds);

  this.pushToDisplayList(background, border, selection, icon, labels);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/modalscreen.js":
/*!*****************************************!*\
  !*** ./js/core/controls/modalscreen.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeAssetBasedControlClass({
  controlTypeID: "com.balsamiq.mockups::ModalScreen",
  controlTypeName: "Modal Screen",
  controlAlternateNames: ["Overlay"],
  defaultSize: _types.default.newSize(318, 240),
  resizeOption: Props.ResizeOptions.BOTH,
  skinAssetName: null,
  colorStyleName: null,
  otherPropertyDefinitions: null
});
controlsMod.addPropertyDefinitions(control, { ...Props.PropertyDefinitions.link
});

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this);

  this.pushToDisplayList((0, _rectangle.makeFilledRectangle)(p_localBbox, 0x000000, 0.4));
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/numericstepper.js":
/*!********************************************!*\
  !*** ./js/core/controls/numericstepper.js ***!
  \********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::NumericStepper",
  controlTypeName: "Num.Stepper",
  controlAlternateNames: ["Numeric Stepper", "Spinner", "Spin"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.ONLY_WIDTH,
        resize: Props.ResizeOptions.HORIZONTAL
      }
    }),
    state: {
      defaultValue: Props.BoolStates.NORMAL,
      enum: Props.BoolStates
    }
  }
});
controlsMod.addStandardTextPropertyDefinitions(control, {
  text: "3"
});
controlsMod.addColorPropertyDefinitions(control, {
  defaultBorderColor: "default_text_color",
  oldBorderColorName: 'color'
});

control.CONTROL_PROTOTYPE.updateData = function () {
  const isStateNormal = this.getActualProperty('state') === _types.default.ControlState.NORMAL;

  const textColor = isStateNormal ? this.getActualProperty("borderColor") : this.style('default_disabled_color');
  const linkColor = isStateNormal ? this.style("default_link_color") : this.style('default_disabled_color');
  const textStyle = this.newStandardTextStyle({
    textColor,
    linkColor
  });
  const text = this.getActualProperty("text");

  if (!this.text || !(0, _text.areTextsEqual)(this.text, text, textStyle)) {
    this.text = (0, _text.newSingleLineText)(text, textStyle, this._context);
  }
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var textSize = this.text.measuredSize();
  var width = p_sizeProperty.width;

  if (width === _types.default.AUTOSIZE) {
    width = this.text.measuredSize().width + this.style("numericstepper_padding_left") + this.style("numericstepper_padding_right");
  }

  return _types.default.newSize(width, textSize.height + this.style('numericstepper_padding_top') + this.style('numericstepper_padding_bottom'));
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var background = (0, _rectangle.makeFilledRectangle)(p_localBbox, 0xFFFFFF);
  var buttonWidth = this.style("numericstepper_button_width");
  var buttonBackground = (0, _rectangle.makeFilledRectangle)(_types.default.newRectangle(p_localBbox.x + p_localBbox.width - buttonWidth, p_localBbox.y, buttonWidth, p_localBbox.height), this.style("numericstepper_button_background_color"));
  var borderColor = _types.default.ControlState.isDisabledState(this.getActualProperty("state")) ? this.style("numericstepper_disabled_color") : this.getActualProperty("borderColor");

  var border = _displayitem.default.createDisplayItemsFromSkin("numericstepper", borderColor, p_localBbox, this._context);

  var label;

  if (this.getActualProperty("text")) {
    label = this.text.makeDisplayItems(_types.default.paddedRectangle(p_localBbox, this.style('numericstepper_padding_top'), this.style('numericstepper_padding_bottom'), this.style("numericstepper_padding_left"), this.style("numericstepper_padding_right")), _types.default.TextAlign.RIGHT);
  }

  this.pushToDisplayList(background, buttonBackground, border, label);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/paragraph.js":
/*!***************************************!*\
  !*** ./js/core/controls/paragraph.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _macro = __webpack_require__(/*! core/bricks/macro */ "./js/core/bricks/macro.js");

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var paragraph = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::Paragraph",
  controlTypeName: "Text",
  controlAlternateNames: ["Paragraph of Text"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(275, 80),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    align: {
      defaultValue: Props.TextAlignments.LEFT,
      enum: Props.TextAlignments
    },
    state: {
      defaultValue: Props.BoolStates.NORMAL,
      enum: Props.BoolStates
    },
    ...Props.PropertyDefinitions.link
  }
});
controlsMod.addStandardTextPropertyDefinitions(paragraph, {
  text: "A *paragraph* of {color:red}text{color} with an [unassigned link].\nA _second_ &row& of ~text~ with a [web link](https://balsamiq.com/wireframes/desktop/docs/text/#basic-formatting)",
  multiline: true,
  loremable: true,
  charCountEnabled: true,
  splitOn: Props.SplitOnTypes.NEWLINE,
  noEmptyText: true
});
controlsMod.addColorPropertyDefinitions(paragraph, {
  defaultTextColor: "default_text_color",
  oldTextColorName: 'color'
});

paragraph.CONTROL_PROTOTYPE.updateData = function () {
  let resolvedText = (0, _macro.processMacros)(this);
  let textStyle = this.newStandardTextStyle({
    textColor: this.getActualProperty("state") === _types.default.ControlState.DISABLED ? this.style("default_disabled_color") : this.getActualProperty("textColor"),
    linkColor: this.getActualProperty("state") === _types.default.ControlState.DISABLED ? this.style("default_disabled_color") : this.style("default_link_color")
  });

  if (!this.text || !Text.areTextsEqual(this.text, resolvedText, textStyle)) {
    this.text = textMod.newText(resolvedText, textStyle, this._context, true);
  }
};

_link.default.addTextLinksMethods(paragraph);

paragraph.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this);

  const {
    displayItems,
    links
  } = this.text.makeDisplayItemsAndLinks(_types.default.paddedRectangle(p_localBbox, this.style("label_padding_top"), this.style("label_padding_bottom")), this.getActualProperty('align'));
  this.pushToDisplayList(displayItems);

  _link.default.processTextLinks(this, links);
};

var _default = paragraph;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/piechart.js":
/*!**************************************!*\
  !*** ./js/core/controls/piechart.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::PieChart",
  controlTypeName: "Chart: Pie",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(180, 180),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    })
  }
});

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  // bbox for background svg
  var bbox = _balsamiq.default.clone(p_localBbox);

  var diameter = Math.min(p_localBbox.width, p_localBbox.height);
  var offset = Math.abs(p_localBbox.width - p_localBbox.height) / 2.0;
  bbox.width = diameter;
  bbox.height = diameter;

  if (p_localBbox.width > p_localBbox.height) {
    bbox.x = offset;
  } else {
    bbox.y = offset;
  } // For procedural drawing


  var centerX = p_localBbox.width / 2.0;
  var centerY = p_localBbox.height / 2.0;
  var radius = diameter / 2.0;
  var thickness = Math.min(this.style("piechart_max_thickness"), Math.max(this.style("piechart_min_thickness"), diameter / 40));
  var t2 = thickness / 2;
  var color = this.style("piechart_border_color");
  var rads = 120 / 180 * Math.PI;
  var path = Path.newPath(p_localBbox.width, p_localBbox.height);
  path.moveTo(centerX, bbox.y + thickness);
  path.lineTo(centerX, centerY);
  path.lineTo(Math.cos(rads) * (radius - t2) + centerX, Math.sin(rads) * (radius - t2) + centerY);
  path.addCircle(centerX, centerY, radius, thickness);

  var assetBbox = _types.default.paddedRectangle(bbox, thickness, thickness, thickness, thickness);

  this.pushToDisplayList(_displayitem.default.createDisplayItemsFromSkin("piechart_background", null, assetBbox, this._context), // Skin background
  _displayitem.default.newSVGPath(p_localBbox, path, thickness, color));
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/playback.js":
/*!**************************************!*\
  !*** ./js/core/controls/playback.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::MediaControls",
  controlTypeName: "Playback",
  controlAlternateNames: ["Playback Controls"]
});

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  this.pushToDisplayList(_displayitem.default.createDisplayItemsFromSkin('playback', null, p_localBbox, this._context));
};

control.CONTROL_PROTOTYPE.measure = function () {
  return _types.default.newSize(110, 36);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/popover.js":
/*!*************************************!*\
  !*** ./js/core/controls/popover.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::iPadPopover",
  controlTypeName: "Popover",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        resize: Props.ResizeOptions.BOTH,
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT
      }
    }),
    popoverArrowPosition: {
      defaultValue: 2,
      dataType: 'int'
    },
    popoverArrowSide: {
      defaultValue: Props.PopoverPositions.TOP,
      enum: Props.PopoverPositions
    },
    ...Props.PropertyDefinitions.TEXT({
      defaultText: "",
      multiline: true
    })
  }
});
controlsMod.addColorPropertyDefinitions(control, {
  defaultColor: 0xFFFFFF
});
controlsMod.addStandardTextPropertyDefinitions(control);
var ARROW_SIZE = 40;
var ARROW_SIZE_2 = ARROW_SIZE / 2;

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  return _types.default.newSize(222, 167 + ARROW_SIZE / 2);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var arrowSide = this.getActualProperty("popoverArrowSide");
  var arrowPosition = this.getActualProperty("popoverArrowPosition");
  var radius = this.style("popover_radius");
  var radiusAndMargin = radius; // I don't understand why Flex code has a "margin" to add to the radius... I ignore it

  var t = this.style("popover_thickness");
  var t2 = t / 2;
  var textPadding = this.style("popover_text_padding");
  var color = 0x000000;
  var fillColor = this.getActualProperty("color");
  var drawBox = p_localBbox.width > ARROW_SIZE + radiusAndMargin * 2 && p_localBbox.height > ARROW_SIZE + radiusAndMargin * 2;

  if (!drawBox) {
    radiusAndMargin = 0;
  }

  var drawArrow = p_localBbox.width > ARROW_SIZE && p_localBbox.height > ARROW_SIZE;

  if (!drawArrow) {
    arrowSide = null;
  }

  var calcArrowTipCoord = function (p_totalSize) {
    var p = (p_totalSize - radiusAndMargin * 2 - ARROW_SIZE) * arrowPosition * 0.25 + radiusAndMargin + ARROW_SIZE_2;
    p = Math.max(p, ARROW_SIZE_2 + t);
    p = Math.min(p, p_totalSize - ARROW_SIZE_2 - t);
    return p;
  };

  var boxTop = drawArrow ? arrowSide === _types.default.Positions.TOP ? ARROW_SIZE_2 : 0 : 0;
  var boxLeft = drawArrow ? arrowSide === _types.default.Positions.LEFT ? ARROW_SIZE_2 : 0 : 0;
  var boxRigth = drawArrow ? arrowSide === _types.default.Positions.RIGHT ? ARROW_SIZE_2 : 0 : 0;
  var boxBottom = drawArrow ? arrowSide === _types.default.Positions.BOTTOM ? ARROW_SIZE_2 : 0 : 0;
  var boxPath = Path.newPath(p_localBbox.width, p_localBbox.height);
  var arrowTipCoord;
  var arcRadius = radius;
  boxPath.moveTo(boxLeft + t2, boxTop + radiusAndMargin + t2); // top left arc

  if (drawBox) {
    boxPath.arc90(arcRadius, -arcRadius);
  } // top border


  if (arrowSide === _types.default.Positions.TOP) {
    arrowTipCoord = calcArrowTipCoord(p_localBbox.width);
    boxPath.lineTo(arrowTipCoord - ARROW_SIZE_2 + t2, boxTop + t2);
    boxPath.lineTo(arrowTipCoord, t2);
    boxPath.lineTo(arrowTipCoord + ARROW_SIZE_2 - t2, boxTop + t2);
  }

  boxPath.lineTo(p_localBbox.width - boxRigth - t2 - radiusAndMargin, boxTop + t2); // top right arc

  if (drawBox) {
    boxPath.arc90(arcRadius, arcRadius);
  } // right border


  if (arrowSide === _types.default.Positions.RIGHT) {
    arrowTipCoord = calcArrowTipCoord(p_localBbox.height);
    boxPath.lineTo(p_localBbox.width - boxRigth - t2, arrowTipCoord - ARROW_SIZE_2 + t2);
    boxPath.lineTo(p_localBbox.width - boxRigth - t2 + ARROW_SIZE_2, arrowTipCoord - t2);
    boxPath.lineTo(p_localBbox.width - boxRigth - t2, arrowTipCoord + ARROW_SIZE_2 - t2);
  }

  boxPath.lineTo(p_localBbox.width - boxRigth - t2, p_localBbox.height - boxBottom - radiusAndMargin - t2); // bottom right arc

  if (drawBox) {
    boxPath.arc90(-arcRadius, arcRadius);
  } // bottom border


  if (arrowSide === _types.default.Positions.BOTTOM) {
    arrowTipCoord = calcArrowTipCoord(p_localBbox.width);
    boxPath.lineTo(arrowTipCoord + ARROW_SIZE_2 - t2, p_localBbox.height - boxBottom - t2);
    boxPath.lineTo(arrowTipCoord, p_localBbox.height - t2);
    boxPath.lineTo(arrowTipCoord - ARROW_SIZE_2 + t2, p_localBbox.height - boxBottom - t2);
  }

  boxPath.lineTo(boxLeft + radiusAndMargin + t2, p_localBbox.height - boxBottom - t2); // bottom left arc

  if (drawBox) {
    boxPath.arc90(-arcRadius, -arcRadius);
  } // left border


  if (arrowSide === _types.default.Positions.LEFT) {
    arrowTipCoord = calcArrowTipCoord(p_localBbox.height);
    boxPath.lineTo(boxLeft + t2, arrowTipCoord + ARROW_SIZE_2 - t2);
    boxPath.lineTo(t, arrowTipCoord);
    boxPath.lineTo(boxLeft + t2, arrowTipCoord - ARROW_SIZE_2 + t2);
  }

  boxPath.close();
  this.pushToDisplayList(_displayitem.default.newSVGPath(p_localBbox, boxPath, drawBox ? t : null, drawBox ? color : null, fillColor));

  if (!drawBox && drawArrow) {
    var arrowPath;
    var arrowX, arrowY;
    var pxs, pys;

    if (arrowSide === _types.default.Positions.TOP || arrowSide === _types.default.Positions.BOTTOM) {
      arrowPath = Path.newPath(ARROW_SIZE + t, ARROW_SIZE_2 + t);
      arrowX = calcArrowTipCoord(p_localBbox.width) - ARROW_SIZE_2;
      pxs = [t2, ARROW_SIZE_2, ARROW_SIZE + t2];

      if (arrowSide === _types.default.Positions.TOP) {
        pys = [ARROW_SIZE_2 + t2, t2, ARROW_SIZE_2 + t2];
        arrowY = t2;
      } else {
        pys = [t2, ARROW_SIZE_2 + t2, t2];
        arrowY = p_localBbox.height - ARROW_SIZE_2 - t;
      }
    } else {
      arrowPath = Path.newPath(ARROW_SIZE_2 + t, ARROW_SIZE + t);
      arrowY = calcArrowTipCoord(p_localBbox.height) - ARROW_SIZE_2;
      pys = [t2, ARROW_SIZE_2, ARROW_SIZE + t2];

      if (arrowSide === _types.default.Positions.LEFT) {
        pxs = [ARROW_SIZE_2 + t2, t2, ARROW_SIZE_2 + t2];
        arrowX = t2;
      } else {
        pxs = [t2, ARROW_SIZE_2 + t2, t2];
        arrowX = p_localBbox.width - ARROW_SIZE_2 - t;
      }
    }

    arrowPath.moveTo(pxs[0], pys[0]);
    arrowPath.lineTo(pxs[1], pys[1]);
    arrowPath.lineTo(pxs[2], pys[2]);
    this.pushToDisplayList(_displayitem.default.newSVGPath(_types.default.newRectangle(arrowX, arrowY, arrowPath.width, arrowPath.height), arrowPath, t, color, fillColor));
  }

  var textProp = this.getActualProperty("text");

  if (textProp) {
    var radiusPadding = radius * 4 / 6;
    var textWidth = p_localBbox.width - boxLeft - boxRigth - textPadding * 2;
    var textHeight = p_localBbox.height - boxTop - boxBottom - radiusPadding * 2;

    if (textWidth > 0 && textHeight > 0) {
      var textBounds = _types.default.newRectangle(boxLeft + textPadding, boxTop + radiusPadding, textWidth, textHeight);

      var textStyle = this.newStandardTextStyle();
      var hasNewLines = textProp.split("\n").length > 1; //textStyle.fontSize = hasNewLines ? 13 : 14;
      //textStyle.bold = !hasNewLines;

      textStyle.textColor = _colors.default.labelColorGivenBackgroundColor(fillColor, 1, this.style("default_text_color"));
      textStyle.linkColor = _colors.default.labelColorGivenBackgroundColor(fillColor, 1, this.style("default_link_color"));
      var text = textMod.newText(textProp, textStyle, this._context);
      this.pushToDisplayList(text.makeDisplayItems(textBounds, hasNewLines ? _types.default.TextAlign.LEFT : _types.default.TextAlign.CENTER));
    }
  }
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/progressbar.js":
/*!*****************************************!*\
  !*** ./js/core/controls/progressbar.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::ProgressBar",
  controlTypeName: "Progress Bar",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    indeterminate: {
      defaultValue: false,
      dataType: 'bool'
    },
    scrollBarValue: {
      defaultValue: 0.5,
      dataType: 'float01'
    },
    style: {
      defaultValue: Props.ProgressBarStyles.THICK,
      enum: Props.ProgressBarStyles
    }
  }
});
controlsMod.addColorPropertyDefinitions(control, {
  defaultColor: "progressbar_color"
});

control.CONTROL_PROTOTYPE.measure = function () {
  //var height;
  var height = this.getActualProperty("style") != "thin" ? 20 : 10;
  return _types.default.newSize(100, height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var backgroundIndeterminateTiles = [];
  var backgroundProgress = null;
  var border;
  var padding = this.style("progressbar_padding");
  var background, backgroundProgressRect;

  if (this.getActualProperty("style") != "thin") {
    if (this.getActualProperty("indeterminate")) {
      var x = padding;
      var tileWidth = this.styleSVGSize("progressbar_indeterminate").width;

      while (x < p_localBbox.width - padding) {
        var backgroundIndeterminateRect = _types.default.newRectangle(x, padding, tileWidth, p_localBbox.height - padding * 2);

        var backgroundIndeterminate = _displayitem.default.createDisplayItemsFromSkin("progressbar_indeterminate", this.getActualProperty("color"), backgroundIndeterminateRect, this._context);

        backgroundIndeterminateTiles.push(backgroundIndeterminate);
        x += tileWidth;
      }

      background = (0, _rectangle.makeFilledRectangle)(p_localBbox, 0xFFFFFF);
    } else {
      var barWidth = p_localBbox.width * this.getActualProperty('scrollBarValue');

      if (barWidth - padding * 2 > 0) {
        backgroundProgressRect = _types.default.newRectangle(padding, padding, barWidth - padding * 2, p_localBbox.height - padding * 2);
        backgroundProgress = (0, _rectangle.makeFilledRectangle)(backgroundProgressRect, this.getActualProperty("color"));
      }

      background = (0, _rectangle.makeFilledRectangle)(p_localBbox, 0xFFFFFF);
    }

    border = _displayitem.default.createDisplayItemsFromSkin("rectangle_border_square", 0x000000, p_localBbox, this._context);
    this.pushToDisplayList(background, backgroundProgress, backgroundIndeterminateTiles, border);
  } else {
    var bgBounds = _balsamiq.default.clone(p_localBbox);

    bgBounds.y += padding;
    bgBounds.height -= 2 * padding;
    background = (0, _rectangle.makeFilledRectangle)(bgBounds, 0xFFFFFF);
    var background1 = (0, _rectangle.makeFilledRectangle)(bgBounds, this.getActualProperty("color"), 0.45);

    if (this.getActualProperty("indeterminate")) {
      backgroundProgressRect = _types.default.newRectangle(p_localBbox.width * 0.25, padding, p_localBbox.width / 2, p_localBbox.height - 2 * padding);
      backgroundProgress = (0, _rectangle.makeFilledRectangle)(backgroundProgressRect, this.getActualProperty("color"));
    } else if (this.getActualProperty('scrollBarValue') > 0) {
      backgroundProgressRect = _types.default.newRectangle(0, padding, p_localBbox.width * this.getActualProperty('scrollBarValue'), p_localBbox.height - 2 * padding);
      backgroundProgress = (0, _rectangle.makeFilledRectangle)(backgroundProgressRect, this.getActualProperty("color"));
    }

    this.pushToDisplayList(background, background1, backgroundProgress);
  }
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/properties.js":
/*!****************************************!*\
  !*** ./js/core/controls/properties.js ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.newAutoSize = exports.isAutoSize = exports.iPhoneResizing = exports.iPhoneModels = exports.iPhoneBackgrounds = exports.iOSDevices = exports.areValuesEqual = exports.VerticalPositions = exports.TooltipDirections = exports.Thumbs = exports.TextOrientations = exports.TextMacros = exports.TextInputBorderStyles = exports.TextAlignments = exports.StrokesBasic = exports.Strokes = exports.States = exports.SplitOnTypes = exports.SimpleStates = exports.SimpleBorderStyles = exports.Shapes = exports.SearchShapes = exports.ResizeOptions = exports.RadioButtonStates = exports.PropertyDefinitions = exports.PropertyDataTypes = exports.ProgressBarStyles = exports.Positions = exports.PopoverPositions = exports.Orientations = exports.OnOffStates = exports.IconLabelPositions = exports.HorizontalPositions = exports.Directions = exports.Devices = exports.DatePickerType = exports.Crop = exports.ComboboxStates = exports.CanvasBorderStyles = exports.ButtonStates = exports.BrowserBorderStyles = exports.BorderStyles = exports.BoolStates = exports.BREAKLINE_VERTICAL_OFFSET = exports.AutoSizeValues = exports.AutoSizeOptions = exports.ArrowShapes = exports.ArrowCurvatures = exports.AUTOSIZE = void 0;
exports.replaceMultiLink = replaceMultiLink;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const PropertyDataTypes = {
  string: {
    deserialize: function (p_value) {
      return p_value;
    }
  },
  int: {
    deserialize: function (p_value) {
      return parseInt(p_value, 10);
    }
  },
  bool: {
    deserialize: function (p_value) {
      return p_value === 'true';
    }
  },
  float01: {
    deserialize: function (p_value) {
      var value = parseFloat(p_value);

      _balsamiq.default.assert(value >= 0.0 && value <= 1.0);

      return value;
    }
  },
  object: {
    deserialize: function (p_value) {
      return p_value;
    },
    areEqual: function (value1, value2) {
      return JSON.stringify(value1) == JSON.stringify(value2);
    }
  },
  enum: {
    deserialize: function (p_value) {
      return p_value;
    }
  }
};
exports.PropertyDataTypes = PropertyDataTypes;

const areValuesEqual = function (dataType, value1, value2) {
  if (PropertyDataTypes[dataType].areEqual) {
    return PropertyDataTypes[dataType].areEqual(value1, value2);
  } else {
    return value1 === value2;
  }
};

exports.areValuesEqual = areValuesEqual;
const PropertyDefinitions = {
  SIZE: function (obj) {
    return {
      size: {
        dataType: 'object',
        label: 'Size',
        ...obj
      }
    };
  },
  link: {
    link: {
      defaultValue: null,
      dataType: 'string',
      label: 'Link',
      setter: function setIfIsValidUrl(url) {
        return url && url.match(/^(resource:|action:|https*:)/gm) ? encodeURI(url) : undefined;
      }
    }
  },
  links: function (obj) {
    let res = {
      links: {
        defaultValue: null,
        dataType: 'object',
        label: 'Links',
        options: {
          doNotUpdateOnTextChange: false
        }
      }
    };

    if (obj && obj.doNotUpdateOnTextChange) {
      res.links.options.doNotUpdateOnTextChange = obj.doNotUpdateOnTextChange;
    }

    return res;
  },
  ICONS: function (obj) {
    let res = {
      icons: {
        defaultValue: null,
        dataType: 'object',
        options: {
          doNotUpdateOnTextChange: false
        }
      }
    };

    if (obj && obj.doNotUpdateOnTextChange) {
      res.icons.options.doNotUpdateOnTextChange = obj.doNotUpdateOnTextChange;
    }

    return res;
  },
  TEXT: function (_ref) {
    let {
      defaultText,
      multiline,
      onSet,
      loremable,
      charCountEnabled,
      splitOn,
      noEmptyText
    } = _ref;
    var res = {
      defaultValue: defaultText === undefined ? '' : defaultText,
      label: 'Text'
    };

    if (multiline || loremable || charCountEnabled) {
      res.options = {};
    }

    if (multiline) {
      res.options.multiline = true;
    }

    if (loremable) {
      res.options.loremable = true;
    }

    if (charCountEnabled) {
      res.options.charCountEnabled = true;
    }

    if (typeof onSet === 'function') {
      res.onSet = onSet;
    }

    if (splitOn) {
      _balsamiq.default.assertEnum(splitOn, SplitOnTypes);

      res.splitOn = splitOn;
    }

    if (noEmptyText) {
      _balsamiq.default.assertBool(noEmptyText);

      res.setter = function (text) {
        return text == '' ? undefined : text;
      };
    }

    return {
      text: res
    };
  },
  ALPHA: function (obj) {
    var res = {
      defaultValue: 1.0,
      dataType: 'float01',
      label: 'Opacity'
    };

    if (obj && _balsamiq.default.isString(obj.defaultAlpha)) {
      res.defaultValue = function () {
        return this.style(obj.defaultAlpha);
      };
    }

    return {
      backgroundAlpha: res
    };
  },
  BORDER_STYLE: function (defaultValue, enumValues) {
    _balsamiq.default.assertEnum(defaultValue, enumValues);

    return {
      borderStyle: {
        defaultValue: defaultValue,
        enum: enumValues,
        label: "Border Style"
      }
    };
  },
  SCROLLBAR: () => ({
    value: {
      defaultValue: 0.0,
      dataType: 'float01',
      label: 'Scrollbar Value'
    },
    verticalScrollbar: {
      defaultValue: false,
      dataType: 'bool',
      label: 'Scrollbar'
    }
  }),
  MENU_ICON: () => ({
    menuIcon: {
      defaultValue: false,
      dataType: 'bool',
      label: 'Menu Icon'
    }
  }),
  ROTATION: _ref2 => {
    let {
      isIcon
    } = _ref2;
    const defaultRotationProps = {
      defaultValue: 0,
      dataType: "int",
      options: {
        isClockwise: !isIcon
      }
    };
    return isIcon ? {
      icon: {
        defaultValue: null,
        dataType: "object",
        label: "Icon"
      },
      rotation: { ...defaultRotationProps,
        label: "Icon Rotation"
      }
    } : {
      rotation: defaultRotationProps
    };
  },
  QUICKDRAW_TIP: (quickdrawTipDesktopText, quickdrawTipMobileText) => {
    return {
      quickDrawTips: {
        dataType: "object",
        defaultValue: {
          desktop: quickdrawTipDesktopText,
          mobile: quickdrawTipMobileText
        },
        value: {
          desktop: quickdrawTipDesktopText,
          mobile: quickdrawTipMobileText
        }
      }
    };
  }
};
exports.PropertyDefinitions = PropertyDefinitions;
const SplitOnTypes = {
  COMMA: 'comma',
  NEWLINE: 'newline'
};
exports.SplitOnTypes = SplitOnTypes;
const ResizeOptions = {
  NO: 'no',
  HORIZONTAL: 'horizontal',
  VERTICAL: 'vertical',
  BOTH: 'both'
};
exports.ResizeOptions = ResizeOptions;
const AUTOSIZE = -1;
exports.AUTOSIZE = AUTOSIZE;

const newAutoSize = function () {
  return _types.default.newSize(this.AUTOSIZE, this.AUTOSIZE);
};

exports.newAutoSize = newAutoSize;

const isAutoSize = function (size) {
  return size.width === AUTOSIZE || size.height === AUTOSIZE;
};

exports.isAutoSize = isAutoSize;
const AutoSizeOptions = {
  NO: 0,
  ONLY_WIDTH: 1,
  ONLY_HEIGHT: 2,
  WIDTH_AND_OR_HEIGHT: 3,
  WIDTH_AND_HEIGHT: 4
};
exports.AutoSizeOptions = AutoSizeOptions;
const AutoSizeValues = {
  NO: 0,
  WIDTH: 1,
  HEIGHT: 2,
  BOTH: 3
};
exports.AutoSizeValues = AutoSizeValues;
const States = {
  NORMAL: "up",
  SELECTED: "selected",
  FOCUSED: "focused",
  DISABLED: "disabled",
  DISABLED_SELECTED: "disabledSelected",
  CLOSED: "closed",
  DISABLED_CLOSED: "disabledClosed",
  INDETERMINATE: "indeterminate",
  DISABLED_INDETERMINATE: "disabledIndeterminate",
  FOCUSED_CLOSED: "focusedClosed",
  isDisabledState: function (p_state) {
    return p_state === this.DISABLED || p_state === this.DISABLED_SELECTED || p_state === this.DISABLED_CLOSED || p_state === this.DISABLED_INDETERMINATE;
  },
  isSelectedState: function (p_state) {
    return p_state === this.SELECTED || p_state === this.DISABLED_SELECTED;
  },
  isClosedState: function (p_state) {
    return p_state === this.CLOSED || p_state === this.DISABLED_CLOSED || p_state === this.FOCUSED_CLOSED;
  },
  isIndeterminateState: function (p_state) {
    return p_state === this.INDETERMINATE || p_state === this.DISABLED_INDETERMINATE;
  },
  isFocusedState: function (p_state) {
    return p_state === this.FOCUSED || p_state === this.FOCUSED_CLOSED;
  }
};
exports.States = States;
const ButtonStates = {
  NORMAL: "up",
  SELECTED: "selected",
  FOCUSED: "focused",
  DISABLED: "disabled"
};
exports.ButtonStates = ButtonStates;
const BoolStates = {
  NORMAL: "up",
  DISABLED: "disabled"
};
exports.BoolStates = BoolStates;
const ProgressBarStyles = {
  THIN: "thin",
  THICK: "thick"
};
exports.ProgressBarStyles = ProgressBarStyles;
const SimpleStates = {
  NORMAL: "up",
  FOCUSED: "focused",
  DISABLED: "disabled"
};
exports.SimpleStates = SimpleStates;
const RadioButtonStates = {
  NORMAL: "up",
  SELECTED: "selected",
  DISABLED: "disabled",
  DISABLED_SELECTED: "disabledSelected",
  INDETERMINATE: "indeterminate",
  DISABLED_INDETERMINATE: "disabledIndeterminate"
};
exports.RadioButtonStates = RadioButtonStates;
const ComboboxStates = {
  NORMAL: "up",
  FOCUSED: "focused",
  DISABLED: "disabled",
  CLOSED: "closed",
  DISABLED_CLOSED: "disabledClosed",
  FOCUSED_CLOSED: "focusedClosed"
};
exports.ComboboxStates = ComboboxStates;
const BorderStyles = {
  NONE: "none",
  SQUARE: "square",
  ROUNDED_SOLID: "roundedSolid",
  ROUNDED_DOTTED: "roundedDotted",
  SQUARE_BREAKLINE: "squareBreakline",
  SQUARE_BREAKLINE_TOP: "squareBreaklineTop",
  RECTANGLE: "rectangle",
  LINE: "line"
};
exports.BorderStyles = BorderStyles;
const CanvasBorderStyles = {
  NONE: "none",
  SQUARE: "square",
  ROUNDED_SOLID: "roundedSolid",
  ROUNDED_DOTTED: "roundedDotted",
  SQUARE_BREAKLINE: "squareBreakline",
  SQUARE_BREAKLINE_TOP: "squareBreaklineTop"
};
exports.CanvasBorderStyles = CanvasBorderStyles;
const BrowserBorderStyles = {
  SQUARE: "square",
  SQUARE_BREAKLINE: "squareBreakline"
};
exports.BrowserBorderStyles = BrowserBorderStyles;
const SimpleBorderStyles = {
  NONE: "none",
  SQUARE: "square"
};
exports.SimpleBorderStyles = SimpleBorderStyles;
const TextInputBorderStyles = {
  RECTANGLE: "rectangle",
  LINE: "line"
};
exports.TextInputBorderStyles = TextInputBorderStyles;
const TextAlignments = {
  LEFT: "left",
  CENTER: "center",
  RIGHT: "right"
};
exports.TextAlignments = TextAlignments;
const TextOrientations = {
  RIGHT: "right",
  UP: "up",
  DOWN: "down",
  SLANTED: "slanted"
};
exports.TextOrientations = TextOrientations;
const Positions = {
  TOP: "top",
  BOTTOM: "bottom",
  LEFT: "left",
  RIGHT: "right",
  CENTER: "center"
};
exports.Positions = Positions;
const VerticalPositions = {
  TOP: "top",
  BOTTOM: "bottom"
};
exports.VerticalPositions = VerticalPositions;
const HorizontalPositions = {
  LEFT: "left",
  RIGHT: "right"
};
exports.HorizontalPositions = HorizontalPositions;
const IconLabelPositions = {
  BOTTOM: "bottom",
  LEFT: "left",
  RIGHT: "right"
};
exports.IconLabelPositions = IconLabelPositions;
const PopoverPositions = {
  TOP: "top",
  BOTTOM: "bottom",
  LEFT: "left",
  RIGHT: "right"
};
exports.PopoverPositions = PopoverPositions;
const Crop = {
  ALL: "0,0,1000,1000"
};
exports.Crop = Crop;
const OnOffStates = {
  ON: "on",
  OFF: 'off'
};
exports.OnOffStates = OnOffStates;
const Orientations = {
  LANDSCAPE: "landscape",
  PORTRAIT: "portrait"
};
exports.Orientations = Orientations;
const iPhoneModels = {
  iPhone4: "IPhone4",
  iPhone5: "IPhone5",
  iPhone6: "IPhone6",
  iPhoneX: "IPhoneX"
};
exports.iPhoneModels = iPhoneModels;
const iPhoneBackgrounds = {
  ALL_WHITE: "allWhite",
  TOP_ONLY: "topOnly",
  TOP_BLUE_BOTTOM: "topBlueBottom",
  TOP_BLACK_BOTTOM: "topBlackBottom",
  ALL_BLACK: "allBlack"
};
exports.iPhoneBackgrounds = iPhoneBackgrounds;
const iPhoneResizing = {
  STRETCH: "stretch",
  MAINTAIN_PROPORTIONS: "maintainProportions"
};
exports.iPhoneResizing = iPhoneResizing;
const iOSDevices = {
  iPhone: "iPhone",
  iPad: "iPad"
};
exports.iOSDevices = iOSDevices;
const Devices = {
  IOS: 'iOS',
  ANDROID: 'android'
};
exports.Devices = Devices;
const StrokesBasic = {
  SOLID: "solid",
  DASHED: "dashed",
  DOTTED: "dotted"
};
exports.StrokesBasic = StrokesBasic;
const Strokes = {
  SOLID: "solid",
  DASHED: "dashed",
  DOTTED: "dotted",
  BREAKLINE_BOTTOM: "break",
  BREAKLINE_TOP: "breakTop"
};
exports.Strokes = Strokes;
const BREAKLINE_VERTICAL_OFFSET = 19;
exports.BREAKLINE_VERTICAL_OFFSET = BREAKLINE_VERTICAL_OFFSET;
const TextMacros = {
  MOCKUP_NAME: "{mockup-name}"
};
exports.TextMacros = TextMacros;
const ArrowCurvatures = {
  CONVEX: -1,
  STRAIGHT: 0,
  CONCAVE: 1
};
exports.ArrowCurvatures = ArrowCurvatures;
const ArrowShapes = {
  LINE: 'line',
  BEZIER: 'bezier'
};
exports.ArrowShapes = ArrowShapes;
const SearchShapes = {
  CIRCLE: "circle",
  RECTANGLE: "rectangle"
};
exports.SearchShapes = SearchShapes;
const Shapes = {
  CIRCLE: "circle",
  RECTANGLE: "rectangle",
  ROUNDEDRECTANGLE: "roundRect",
  DOTTEDRECTANGLE: "dottedRect",
  DIAMOND: "diamond",
  STAR: "star",
  PARALLELOGRAM: "parallelogram",
  TRIANGLE: "triangle",
  BREAKLINE_BOTTOM: "rectangle breakline bottom ",
  BREAKLINE_TOP: "rectangle breakline top"
};
exports.Shapes = Shapes;
const Directions = {
  VERTICAL: "vertical",
  HORIZONTAL: "horizontal"
};
exports.Directions = Directions;
const TooltipDirections = {
  NW: "NW",
  NE: "NE",
  SW: "SW",
  SE: "SE"
};
exports.TooltipDirections = TooltipDirections;
const Thumbs = {
  POINTY: "pointy",
  CIRCLE: "circle",
  RECTANGLE: "rectangle"
};
exports.Thumbs = Thumbs;
const DatePickerType = {
  MONTH: "month",
  YEAR: "year"
};
exports.DatePickerType = DatePickerType;

function replaceMultiLink(propValue, linkIndex, newLink) {
  if (!propValue) {
    propValue = [];
  }

  var newPropValue = [];
  var maxLength = Math.max(linkIndex + 1, propValue.length);

  for (var i = 0; i < maxLength; i++) {
    if (i == linkIndex) {
      newPropValue.push(newLink);
    } else if (i < propValue.length) {
      newPropValue.push(propValue[i]);
    } else {
      newPropValue.push(null);
    }
  }

  return newPropValue;
}

/***/ }),

/***/ "./js/core/controls/redx.js":
/*!**********************************!*\
  !*** ./js/core/controls/redx.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeAssetBasedControlClass({
  controlTypeID: "com.balsamiq.mockups::RedX",
  controlTypeName: "Red X",
  controlAlternateNames: ["X-Nay"],
  defaultSize: _types.default.newSize(240, 104),
  resizeOption: Props.ResizeOptions.BOTH,
  skinAssetName: "redx",
  colorStyleName: "redx_color",
  otherPropertyDefinitions: {
    markup: {
      defaultValue: true,
      dataType: 'bool'
    }
  }
});
var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/rules.js":
/*!***********************************!*\
  !*** ./js/core/controls/rules.js ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.vrule = exports.hrule = exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _breakline = __webpack_require__(/*! core/bricks/breakline */ "./js/core/bricks/breakline.js");

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var makeRuleControlClass = function (p_controlTypeID, p_controlTypeName, p_controlAlternateNames, p_strokeKind) {
  var isHorizontal = p_controlTypeID === "com.balsamiq.mockups::HRule";
  var assetName = isHorizontal ? "rule_horizontal" : "rule_vertical";
  var autoSizeVal = isHorizontal ? Props.AutoSizeOptions.ONLY_HEIGHT : Props.AutoSizeOptions.ONLY_WIDTH;
  var resizeVal = isHorizontal ? Props.ResizeOptions.HORIZONTAL : Props.ResizeOptions.VERTICAL;
  var defaultSize = isHorizontal ? _types.default.newSize(100, _types.default.AUTOSIZE) : _types.default.newSize(_types.default.AUTOSIZE, 100);
  var control = controlsMod.makeControlClass({
    controlTypeID: p_controlTypeID,
    controlTypeName: p_controlTypeName,
    controlAlternateNames: p_controlAlternateNames,
    props: { ...Props.PropertyDefinitions.SIZE({
        defaultValue: defaultSize,
        options: {
          autosize: autoSizeVal,
          resize: resizeVal
        }
      }),
      stroke: {
        defaultValue: p_strokeKind.SOLID,
        enum: p_strokeKind
      }
    }
  });
  controlsMod.addColorPropertyDefinitions(control, {
    defaultBorderColor: "rule_default_color",
    hasAlpha: true,
    oldBorderColorName: 'color'
  });

  control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
    var w, h, stroke;

    if (isHorizontal) {
      h = 10;
      w = p_sizeProperty.width;
      stroke = this.getActualProperty("stroke");

      if (stroke === _types.default.Stroke.BREAKLINE_BOTTOM || stroke === _types.default.Stroke.BREAKLINE_TOP) {
        h = _types.default.BREAKLINE_VERTICAL_OFFSET * 2;
      }
    } else {
      w = 10;
      h = p_sizeProperty.height;
    }

    return _types.default.newSize(w, h);
  };

  control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
    var color = this.getActualProperty("borderColor");
    var alpha = this.getActualProperty("backgroundAlpha");
    var dashArray, lineCap, background, breakline;
    var stroke = this.getActualProperty("stroke");

    if (stroke === _types.default.Stroke.SOLID) {
      var svgSize = this.styleSVGSize(assetName);
      var bbox = isHorizontal ? _types.default.newRectangle(p_localBbox.x, p_localBbox.y + (p_localBbox.height - svgSize.height) / 2 - 0.5, p_localBbox.width, svgSize.height) : _types.default.newRectangle(p_localBbox.x + (p_localBbox.width - svgSize.width) / 2 - 0.5, p_localBbox.y, svgSize.width, p_localBbox.height);

      var items = _displayitem.default.createDisplayItemsFromSkin(assetName, color, bbox, this._context);

      _displayitem.default.setAlpha(items, alpha);

      this.pushToDisplayList(items);
    } else {
      lineCap = _displayitem.default.LineCap.ROUND;
      var xStart, yStart, xEnd, yEnd;

      if (isHorizontal) {
        xStart = p_localBbox.x;
        yStart = p_localBbox.y + p_localBbox.height / 2;
        xEnd = xStart + p_localBbox.width;
        yEnd = yStart;
      } else {
        xStart = p_localBbox.x + p_localBbox.width / 2;
        yStart = p_localBbox.y;
        xEnd = xStart;
        yEnd = yStart + p_localBbox.height;
      }

      var displayItem, path;

      if (stroke === _types.default.Stroke.DOTTED) {
        path = new Path.Path(p_localBbox.width, p_localBbox.height);
        path.moveTo(xStart, yStart);
        path.lineTo(xEnd, yEnd);
        var sizeDot = this.style("rule_dot_size");
        dashArray = [sizeDot, sizeDot * 8];
        displayItem = _displayitem.default.newSVGPath(p_localBbox, path, this.style("rule_thickness"), color, null, dashArray, lineCap);

        _displayitem.default.setBorderAlpha(displayItem, alpha);

        this.pushToDisplayList(displayItem);
      } else if (stroke === _types.default.Stroke.DASHED) {
        path = new Path.Path(p_localBbox.width, p_localBbox.height);
        path.moveTo(xStart, yStart);
        path.lineTo(xEnd, yEnd);
        var sizeDash = this.style("rule_dash_size");
        dashArray = [sizeDash * 3, sizeDash * 6];
        displayItem = _displayitem.default.newSVGPath(p_localBbox, path, this.style("rule_thickness"), color, null, dashArray, lineCap);

        _displayitem.default.setBorderAlpha(displayItem, alpha);

        this.pushToDisplayList(displayItem);
      } else if (stroke === _types.default.Stroke.BREAKLINE_TOP) {
        _balsamiq.default.assert(p_controlTypeID === "com.balsamiq.mockups::HRule");

        if (p_localBbox.width >= 65) {
          background = (0, _breakline.makeFilledBreaklineDisplayitem)(xStart, yStart, p_localBbox.width, p_localBbox.height, _types.default.Positions.BOTTOM, 0xFFFFFF);
          breakline = (0, _breakline.makeBreaklineDisplayitem)(xStart, yStart, p_localBbox.width, p_localBbox.height, this.style("rule_breakline_thickness"), color);
          breakline.borderAlpha = alpha;
          this.pushToDisplayList(background, breakline);
        } else {
          // not enough space to draw one breakline, so we draw a simple straight line and the white bg
          path = new Path.Path(p_localBbox.width, p_localBbox.height);
          path.moveTo(xStart, yStart);
          path.lineTo(xEnd, yEnd);
          displayItem = _displayitem.default.newSVGPath(p_localBbox, path, this.style("rule_thickness"), color);
          let rectangle = (0, _rectangle.makeFilledRectangle)(_types.default.newRectangle(0, 0, p_localBbox.width, yEnd));

          _displayitem.default.setBorderAlpha(displayItem, alpha);

          this.pushToDisplayList(rectangle, displayItem);
        }
      } else if (stroke === _types.default.Stroke.BREAKLINE_BOTTOM) {
        _balsamiq.default.assert(p_controlTypeID === "com.balsamiq.mockups::HRule");

        if (p_localBbox.width >= 65) {
          background = (0, _breakline.makeFilledBreaklineDisplayitem)(xStart, yStart, p_localBbox.width, p_localBbox.height, _types.default.Positions.TOP, 0xFFFFFF);
          breakline = (0, _breakline.makeBreaklineDisplayitem)(xStart, yStart, p_localBbox.width, p_localBbox.height, this.style("rule_breakline_thickness"), color);
          breakline.borderAlpha = alpha;
          this.pushToDisplayList(background, breakline);
        } else {
          // not enough space to draw one breakline, so we draw a simple straight line and the white bg
          path = new Path.Path(p_localBbox.width, p_localBbox.height);
          path.moveTo(xStart, yStart);
          path.lineTo(xEnd, yEnd);
          displayItem = _displayitem.default.newSVGPath(p_localBbox, path, this.style("rule_thickness"), color);
          let rectangle = (0, _rectangle.makeFilledRectangle)(_types.default.newRectangle(0, yEnd, p_localBbox.width, p_localBbox.height));

          _displayitem.default.setBorderAlpha(displayItem, alpha);

          this.pushToDisplayList(rectangle, displayItem);
        }
      }
    }
  };

  return control;
};

const hrule = makeRuleControlClass("com.balsamiq.mockups::HRule", "H.Rule", ["Horizontal Rule", "Separator", "Line, HR"], _types.default.Stroke);
exports.hrule = hrule;
const vrule = makeRuleControlClass("com.balsamiq.mockups::VRule", "V.Rule", ["Vertical Rule", "Separator", "Line", "VR"], _types.default.StrokeBasic);
exports.vrule = vrule;
var _default = {
  hrule,
  vrule
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/scratchout.js":
/*!****************************************!*\
  !*** ./js/core/controls/scratchout.js ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeAssetBasedControlClass({
  controlTypeID: "com.balsamiq.mockups::ScratchOut",
  controlTypeName: "Scratch-Out",
  controlAlternateNames: [],
  defaultSize: _types.default.newSize(205, 107),
  resizeOption: Props.ResizeOptions.BOTH,
  skinAssetName: null,
  colorStyleName: null,
  otherPropertyDefinitions: {
    markup: {
      defaultValue: true,
      dataType: 'bool'
    }
  }
});
controlsMod.addColorPropertyDefinitions(control, {
  defaultColor: "scratchout_color",
  hasAlpha: true
});

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  let l = _displayitem.default.createDisplayItemsFromSkin("scratchout", this.getActualProperty("color"), p_localBbox, this._context);

  l.forEach(item => {
    item.alpha = this.getActualProperty("backgroundAlpha");
  });
  this.pushToDisplayList(l);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/scrollbarhorizontal.js":
/*!*************************************************!*\
  !*** ./js/core/controls/scrollbarhorizontal.js ***!
  \*************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _scrollbars = __webpack_require__(/*! core/bricks/scrollbars */ "./js/core/bricks/scrollbars.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::HorizontalScrollBar",
  controlTypeName: "H.Scroll Bar",
  controlAlternateNames: ["Horizontal Scrollbar"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.HORIZONTAL
      }
    }),
    value: {
      defaultValue: 0.0,
      dataType: 'float01'
    }
  }
});

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var width = p_sizeProperty.width == Props.AUTOSIZE ? _scrollbars.Sizes.ScrollbarDefaultVariableSize : p_sizeProperty.width;
  return _types.default.newSize(width, _scrollbars.Sizes.ScrollbarFixedSize);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  this.pushToDisplayList((0, _scrollbars.makeDisplayItemsForHorizontalScrollbar)(p_localBbox.x, p_localBbox.y, p_localBbox.width, this.getActualProperty("value"), this._context));
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/scrollbarvertical.js":
/*!***********************************************!*\
  !*** ./js/core/controls/scrollbarvertical.js ***!
  \***********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _scrollbars = __webpack_require__(/*! core/bricks/scrollbars */ "./js/core/bricks/scrollbars.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::VerticalScrollBar",
  controlTypeName: "V.Scroll Bar",
  controlAlternateNames: ["Vertical Scrollbar"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.VERTICAL
      }
    }),
    value: {
      defaultValue: 0.0,
      dataType: 'float01'
    }
  }
});

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var height = p_sizeProperty.height == Props.AUTOSIZE ? _scrollbars.Sizes.ScrollbarDefaultVariableSize : p_sizeProperty.height;
  return _types.default.newSize(_scrollbars.Sizes.ScrollbarFixedSize, height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  this.pushToDisplayList((0, _scrollbars.makeDisplayItemsForVerticalScrollbar)(p_localBbox.x, p_localBbox.y, p_localBbox.height, this.getActualProperty("value"), _types.default.ControlState.NORMAL, this._context));
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/searchbox.js":
/*!***************************************!*\
  !*** ./js/core/controls/searchbox.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::SearchBox",
  controlTypeName: "Search Box",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT,
        resize: Props.ResizeOptions.HORIZONTAL
      }
    }),
    state: {
      defaultValue: _types.default.ControlState.NORMAL,
      enum: Props.SimpleStates
    },
    ...Props.PropertyDefinitions.link,
    shape: {
      defaultValue: Props.SearchShapes.CIRCLE,
      enum: Props.SearchShapes
    },
    searchIcon: {
      defaultValue: true,
      dataType: 'bool'
    },
    microphoneIcon: {
      defaultValue: false,
      dataType: 'bool'
    }
  }
});
controlsMod.addStandardTextPropertyDefinitions(control, {
  text: "search"
});
controlsMod.addColorPropertyDefinitions(control, {
  defaultTextColor: "searchbox_label_color",
  oldTextColorName: 'color'
});

control.CONTROL_PROTOTYPE.updateData = function () {
  const isDisabled = this.getActualProperty('state') === _types.default.ControlState.DISABLED;

  const textColor = isDisabled ? this.style('default_disabled_color') : this.getActualProperty("textColor");
  const linkColor = isDisabled ? this.style('default_disabled_color') : this.style("default_link_color");
  const textStyle = this.newStandardTextStyle({
    textColor,
    linkColor
  });
  const text = this.getActualProperty("text");

  if (!this.text || !(0, _text.areTextsEqual)(this.text, text, textStyle)) {
    this.text = (0, _text.newSingleLineText)(text, textStyle, this._context);
  }
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var textSize = this.text.measuredSize();
  var width = p_sizeProperty.width === _types.default.AUTOSIZE ? Math.max(120, textSize.width + this.style("searchbox_padding_left") + this.style("searchbox_padding_right")) : p_sizeProperty.width;
  var height = textSize.height + this.style("searchbox_padding_top") + this.style("searchbox_padding_bottom");
  return _types.default.newSize(width, height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBox) {
  _link.default.processControlLink(this);

  var state = this.getActualProperty('state');
  var background = (0, _rectangle.makeFilledRectangle)(_types.default.shrinkRectangle(p_localBox, 2), 0xFFFFFF, 1, this.getActualProperty("shape") === Props.SearchShapes.CIRCLE ? 10 : 0);
  var iconColor = state === _types.default.ControlState.DISABLED ? this.style("default_disabled_color") : this.style("default_icon_color");

  var searchIconData = _icon.default.newAppIconData("search", 12, 0);

  var searchIcon = this.getActualProperty("searchIcon") === true ? _icon.default.makeDisplayItemsFromIconData(this, searchIconData, 6, (p_localBox.height - 12) / 2, iconColor) : null;

  var labelBbox = _balsamiq.default.clone(p_localBox);

  labelBbox.x = this.getActualProperty("searchIcon") === true ? this.style("searchbox_padding_left") : 6;
  var label = this.text.makeDisplayItems(_types.default.paddedRectangle(labelBbox, this.style("searchbox_padding_top"), 0, 0));
  var micVisible = p_localBox.width - labelBbox.x - this.text.measuredSize().width > 18;

  var micIconData = _icon.default.newAppIconData("microphone", 12, 0);

  var micIcon = this.getActualProperty("microphoneIcon") === true && micVisible ? _icon.default.makeDisplayItemsFromIconData(this, micIconData, p_localBox.width - 18, (p_localBox.height - 12) / 2, iconColor) : null;
  var borderColor = 0x000000;

  if (state === _types.default.ControlState.FOCUSED) {
    borderColor = this.style("default_selection_color");
  } else if (state === _types.default.ControlState.DISABLED) {
    borderColor = this.style("default_disabled_color");
  }

  var border = _displayitem.default.createDisplayItemsFromSkin(this.getActualProperty("shape") === Props.SearchShapes.CIRCLE ? "searchbox" : "rectangle_border_square", borderColor, p_localBox, this._context);

  this.pushToDisplayList(background, label, searchIcon, micIcon, border);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/shape.js":
/*!***********************************!*\
  !*** ./js/core/controls/shape.js ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _macro = __webpack_require__(/*! core/bricks/macro */ "./js/core/bricks/macro.js");

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var _breakline = __webpack_require__(/*! core/bricks/breakline */ "./js/core/bricks/breakline.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::RoundButton",
  controlTypeName: "Shape",
  controlAlternateNames: ["Geometric Shape", "Circle", "Ellipse", "Rectangle", "Diamond", "Star", "Parallelogram", "Triangle"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(32, 32),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    shape: {
      defaultValue: Props.Shapes.CIRCLE,
      enum: Props.Shapes
    },
    align: {
      defaultValue: Props.TextAlignments.CENTER,
      enum: Props.TextAlignments
    },
    shapeRotation: {
      defaultValue: 0,
      dataType: 'int'
    },
    ...Props.PropertyDefinitions.link,
    ...Props.PropertyDefinitions.ROTATION({
      isIcon: true
    })
  }
});
Controls.addStandardTextPropertyDefinitions(control, {
  splitOn: Props.SplitOnTypes.NEWLINE,
  multiline: true
});
Controls.addColorPropertyDefinitions(control, {
  defaultColor: "shape_default_color",
  defaultBorderColor: "shape_default_border_color",
  hasAlpha: true
});

control.CONTROL_PROTOTYPE.updateData = function () {
  _icon.default.processIconData(this);

  var text = (0, _macro.processMacros)(this);
  var textStyle = this.newStandardTextStyle({
    textColor: _colors.default.labelColorGivenBackgroundColor(this.getActualProperty("color"), this.getActualProperty("backgroundAlpha"), this.style("default_text_color")),
    linkColor: _colors.default.labelColorGivenBackgroundColor(this.getActualProperty("color"), this.getActualProperty("backgroundAlpha"), this.style("default_link_color"))
  });
  this.text = Text.newText(text, textStyle, this._context, true);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this);

  const shape = this.getActualProperty("shape");
  const rotation = this.getActualProperty("shapeRotation");
  let maskID = null; // Border

  let border;
  const w = p_localBbox.width,
        h = p_localBbox.height;
  let t = 0;

  if (w >= 10 && h >= 10) {
    // to avoid RSOD we let t (thickness) = 0 if the control is super small (it can happen with projects exported from myBalsamiq)
    t = Math.floor(Math.min(this.style("shape_max_thickness"), Math.max(this.style("shape_min_thickness"), Math.min(p_localBbox.width, p_localBbox.height) / 40)));
  }

  const t2 = t / 2;
  let path = Path.newPath(w, h);
  let fillColor = this.getActualProperty("color");

  if (shape === Props.Shapes.CIRCLE) {
    path = Path.newEllipsePath(w, h, t);
  } else if (shape === Props.Shapes.RECTANGLE) {
    path = Path.newRectanglePath(w, h, t);
  } else if (shape === Props.Shapes.ROUNDEDRECTANGLE || shape === Props.Shapes.DOTTEDRECTANGLE) {
    const r = shape === Props.Shapes.DOTTEDRECTANGLE ? 10 : Math.min(w / 8, h / 8, 10);
    path = Path.newRoundedRectanglePath(w, h, t, r);

    if (shape === Props.Shapes.DOTTEDRECTANGLE) {
      const backgroundPath = Path.newRoundedRectanglePath(w, h, 0, r);

      const background = _displayitem.default.newSVGPath(p_localBbox, backgroundPath, 0, null, fillColor);

      background.alpha = this.getActualProperty("backgroundAlpha");
      this.pushToDisplayList(background);
      fillColor = null;
    }
  } else if (shape === Props.Shapes.DIAMOND) {
    path.moveTo(w * 0.5, t2);
    path.lineTo(w - t2, h * 0.5);
    path.lineTo(w * 0.5, h - t2);
    path.lineTo(t2, h * 0.5);
    path.close();
  } else if (shape === Props.Shapes.PARALLELOGRAM) {
    path.moveTo(w * 0.17, t2);
    path.lineTo(w - t2, t2);
    path.lineTo(w * 0.83, h - t2);
    path.lineTo(t2, h - t2);
    path.close();
  } else if (shape === Props.Shapes.STAR) {
    path.moveTo(w * 0.64, t);
    path.lineTo(w * 0.67, h * 0.34);
    path.lineTo(w - t, h * 0.5);
    path.lineTo(w * 0.69, h * 0.65);
    path.lineTo(w * 0.62, h - t);
    path.lineTo(w * 0.37, h * 0.75);
    path.lineTo(t, h * 0.8);
    path.lineTo(w * 0.17, h * 0.5);
    path.lineTo(t, h * 0.19);
    path.lineTo(w * 0.37, h * 0.24);
    path.close();
  } else if (shape === Props.Shapes.TRIANGLE) {
    if (rotation === 0) {
      path.moveTo(w / 2, t2);
      path.lineTo(w - t2, h - t2);
      path.lineTo(t2, h - t2);
    } else if (rotation === 90) {
      path.moveTo(t2, h / 2);
      path.lineTo(w - t2, t2);
      path.lineTo(w - t2, h - t2);
    } else if (rotation === 180) {
      path.moveTo(t2, t2);
      path.lineTo(w - t2, t2);
      path.lineTo(w / 2, h - t2);
    } else if (rotation === 270) {
      path.moveTo(t2, t2);
      path.lineTo(w - t2, h / 2);
      path.lineTo(t2, h - t2);
    }

    path.close();
  } else if (shape === Props.Shapes.BREAKLINE_BOTTOM) {
    let drawBreakLine = w >= 65 && p_localBbox.height >= 40;
    path.moveTo(w - t2, h - t2 + (drawBreakLine ? -_types.default.BREAKLINE_VERTICAL_OFFSET : 0));
    path.lineTo(w - t2, t2);
    path.lineTo(t2, t2);

    if (drawBreakLine) {
      path.lineTo(t2, h - t2 - _types.default.BREAKLINE_VERTICAL_OFFSET);
      Path.newBreaklinePath(w, 0, Path.Breakline.OPEN, t, _types.default.newPoint(t2, h - t2 - _types.default.BREAKLINE_VERTICAL_OFFSET), path);
    } else {
      path.lineTo(t2, h - t2);
    }

    path.close();
    maskID = (0, _breakline.addBreaklineMask)(this, 0, h - t2 - _types.default.BREAKLINE_VERTICAL_OFFSET, _types.default.Positions.BOTTOM);
  } else if (shape === Props.Shapes.BREAKLINE_TOP) {
    let drawBreakLine = w >= 65 && p_localBbox.height >= 40;
    path.moveTo(w - t2, drawBreakLine ? t2 + _types.default.BREAKLINE_VERTICAL_OFFSET : t2);
    path.lineTo(w - t2, h - t2);
    path.lineTo(t2, h - t2);

    if (w >= 65 && p_localBbox.height >= 40) {
      path.lineTo(t2, t2 + _types.default.BREAKLINE_VERTICAL_OFFSET);
      Path.newBreaklinePath(w, 0, Path.Breakline.OPEN, t, _types.default.newPoint(t2, t2 + _types.default.BREAKLINE_VERTICAL_OFFSET), path);
    } else {
      path.lineTo(t2, t2);
    }

    path.close();
    maskID = (0, _breakline.addBreaklineMask)(this, 0, t2 + _types.default.BREAKLINE_VERTICAL_OFFSET, _types.default.Positions.TOP);
  }

  border = _displayitem.default.newSVGPath(p_localBbox, path, t, this.getActualProperty("borderColor"), fillColor);
  border.alpha = this.getActualProperty("backgroundAlpha");

  if (shape === Props.Shapes.DOTTEDRECTANGLE) {
    border.dashArray = [8, 8, 8, 8];
  }

  this.pushToDisplayList(border); // Icon and text

  const isEmptyText = this.getActualProperty("text").length === 0;
  const paddingLeft = this.style("shape_padding_left");
  const paddingRight = this.style("shape_padding_right");
  const paddingTop = this.style("shape_padding_top");
  const paddingBottom = this.style("shape_padding_bottom");
  const iconPadding = this.style("shape_icon_padding");
  const align = this.getActualProperty("align");
  let textWidth = isEmptyText ? 0 : Math.min(Math.max(0, p_localBbox.width - paddingLeft - paddingRight - (this.iconData ? this.iconData.size : 0)), this.text.measuredSize().width);
  let iconAndTextWidth = textWidth + (this.iconData ? this.iconData.size + iconPadding : 0);
  let x = 0;

  if (align === _types.default.TextAlign.LEFT) {
    x = paddingLeft;
  } else if (align === _types.default.TextAlign.CENTER) {
    x = (p_localBbox.width - iconAndTextWidth) / 2;
  } else {
    x = p_localBbox.width - paddingRight - iconAndTextWidth;
  }

  x = Math.round(Math.max(0, x));
  let icon, text;

  if (this.iconData) {
    const iconY = Math.round((p_localBbox.height - paddingTop - paddingBottom - this.iconData.size) / 2 + paddingTop);
    icon = _icon.default.makeDisplayItemsFromIconData(this, this.iconData, x, iconY, _colors.default.labelColorGivenBackgroundColor(this.getActualProperty("color"), this.getActualProperty("backgroundAlpha"), this.style("default_icon_color")));
    this.pushToDisplayList(icon);
    x += this.iconData.size + iconPadding;

    if (maskID !== null) {
      _displayitem.default.maskDisplayItems(icon, maskID);
    }
  }

  let spaceForText = textWidth;

  if (!this.iconData && this.getActualProperty("text").length === 1 && spaceForText < p_localBbox.width) {
    //if the control is only one char try to fit it in ignoring the padding
    x = 0;
    spaceForText = p_localBbox.width;
  }

  if (!isEmptyText && p_localBbox.height - this.text.measuredSize().height > 0 && spaceForText > 0) {
    const textBounds = _types.default.newRectangle(x, 0, spaceForText, p_localBbox.height);

    const textAlign = this.iconData ? _types.default.TextAlign.LEFT : align;
    text = this.text.makeDisplayItemsAndLinks(textBounds, textAlign, _types.default.Positions.CENTER);
    this.pushToDisplayList(text.displayItems);

    if (maskID !== null) {
      _displayitem.default.maskDisplayItems(text.displayItems, maskID);
    }

    _link.default.processTextLinks(this, text.links);
  }
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/sitemap.js":
/*!*************************************!*\
  !*** ./js/core/controls/sitemap.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var _sitemapUtils = __webpack_require__(/*! ./sitemapUtils */ "./js/core/controls/sitemapUtils.ts");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::SiteMap",
  controlTypeName: "Site Map",
  controlAlternateNames: [],
  props: {
    direction: {
      defaultValue: Props.Directions.VERTICAL,
      enum: Props.Directions
    },
    ...Props.PropertyDefinitions.TEXT({
      defaultText: "Home\n" + "- Products\n" + "-- Product 1\n" + "-- Product 2\n" + "- About Us\\rCompany\n" + "- Support\n" + "- Blog",
      multiline: true
    }),
    ...Props.PropertyDefinitions.links()
  }
});
Controls.addStandardTextPropertyDefinitions(control, {
  splitOn: Props.SplitOnTypes.NEWLINE
});
const ERROR_LINE = "*{color:#ff0000}Oops!{color}*\\rSitemaps expect a single starting point, like this:\\r \\rHome\\r- Company\\r- Products";
const CONSTANTS = {
  label_padding: 10,
  min_node_width: 100,
  min_node_vertical_height: 70,
  min_node_horizontal_height: 35,
  node_horizontal_padding: 9.25,
  node_vertical_padding: 10,
  node_vertical_spacing: 50,
  node_horizontal_spacing: 70
}; // Node of the tree

const Node = function (control, text, line, textStyle) {
  this.control = control;
  this.level = (0, _sitemapUtils.getLevel)(text);
  const label = (0, _sitemapUtils.getLabel)(text);

  if (!this.label || !Text.areTextsEqual(this.label, label, textStyle)) {
    this.label = Text.newTextNoWrap(label, textStyle, this.control._context);
  }

  this.line = line;
  this.children = [];
  this.father = null;
  this.siblings = false;
};

Node.prototype.addChild = function (p_node) {
  this.children.push(p_node);
};

Node.prototype.measureWidth = function () {
  return Math.max(this.label.measuredSize().width + 2 * CONSTANTS.label_padding, CONSTANTS.min_node_width);
};

Node.prototype.measureHeight = function (direction) {
  return Math.max(this.label.measuredSize().height + 2 * CONSTANTS.label_padding, direction === Props.Directions.VERTICAL ? CONSTANTS.min_node_vertical_height : CONSTANTS.min_node_horizontal_height);
};

Node.prototype.getActualSubtreeWidth = function (direction) {
  let childrenWidth = 0;
  this.children.forEach(node => {
    childrenWidth += node.getSubtreeWidth(direction);
  });

  if (this.children.length > 0) {
    childrenWidth += (this.children.length - 1) * 2 * (direction === Props.Directions.VERTICAL ? CONSTANTS.node_vertical_padding : CONSTANTS.node_horizontal_padding);
  }

  return childrenWidth;
};

Node.prototype.getSubtreeWidth = function (direction) {
  if (!this.subtreeWidth) {
    const ownWidth = this.measureWidth();
    const childrenWidth = this.getActualSubtreeWidth(direction);
    this.subtreeWidth = Math.max(ownWidth, childrenWidth);
  }

  return this.subtreeWidth;
};

Node.prototype.getActualSubtreeHeight = function (direction) {
  let childrenHeight = 0;
  this.children.forEach(node => {
    childrenHeight += node.getSubtreeHeight(direction);
  });

  if (this.children.length > 0) {
    childrenHeight += (this.children.length - 1) * 2 * (direction === Props.Directions.VERTICAL ? CONSTANTS.node_vertical_padding : CONSTANTS.node_horizontal_padding);
  }

  return childrenHeight;
};

Node.prototype.getSubtreeHeight = function (direction) {
  //function used for horizontal sitemap only
  const ownHeight = this.measureHeight(direction);
  const childrenHeight = this.getActualSubtreeHeight(direction); //to avoid complex code later the height is returned as the width

  this.subtreeWidth = Math.max(ownHeight, childrenHeight);
  return this.subtreeWidth;
};

Node.prototype.makeDisplayItems = function (direction) {
  let displayItems = [];

  const localBbox = _types.default.newRectangle(this.x, this.y, this.measureWidth(), this.measureHeight(direction));

  displayItems = displayItems.concat(Rectangle.makeFilledRectangle(localBbox, 0xffffff));
  displayItems = displayItems.concat(_displayitem.default.createDisplayItemsFromSkin("rectangle_border_square", this.control.style("default_border_color"), localBbox, this.control._context));

  if (this.error) {
    displayItems = displayItems.concat(this.label.makeDisplayItems(_types.default.paddedRectangle(localBbox, CONSTANTS.label_padding, 0, CONSTANTS.label_padding), _types.default.TextAlign.LEFT));
  } else {
    displayItems = displayItems.concat(this.label.makeDisplayItems(localBbox, _types.default.TextAlign.CENTER, _types.default.Positions.CENTER));
  }

  return displayItems;
};

Node.prototype.setPosition = function (p_x, p_y) {
  this.x = p_x;
  this.y = p_y;
};

Node.prototype.makeDisplayItemsForEdges = function (p_path, p_direction) {
  if (!this.children || !this.children.length) {
    return null;
  }

  const verticalSpacing = CONSTANTS.node_vertical_spacing;
  let pt1, pt2, centerY;

  if (p_direction === Props.Directions.VERTICAL) {
    // Vertical line from top node
    pt1 = _types.default.newPoint(this.x + this.measureWidth() / 2, this.y + this.measureHeight(p_direction));
    pt2 = _types.default.newPoint(pt1.x, pt1.y + verticalSpacing / 2);
    p_path.moveTo(pt1.x, pt1.y);
    p_path.lineTo(pt2.x, pt2.y); // Horizontal line

    centerY = pt2.y;
    pt1 = _types.default.newPoint(this.children[0].x + this.children[0].measureWidth() / 2, centerY);
    pt2 = _types.default.newPoint(this.children[this.children.length - 1].x + this.children[this.children.length - 1].measureWidth() / 2, centerY);
    p_path.moveTo(pt1.x, pt1.y);
    p_path.lineTo(pt2.x, pt2.y);
    this.children.forEach(node => {
      pt1 = _types.default.newPoint(node.x + node.measureWidth() / 2, node.y);
      pt2 = _types.default.newPoint(pt1.x, centerY);
      p_path.moveTo(pt1.x, pt1.y);
      p_path.lineTo(pt2.x, pt2.y);
    });
  } else {
    // Horizontal line
    pt1 = _types.default.newPoint(this.x + this.measureWidth(), this.y + this.measureHeight(p_direction) / 2);
    pt2 = _types.default.newPoint(pt1.x + verticalSpacing / 2, pt1.y);
    p_path.moveTo(pt1.x, pt1.y);
    p_path.lineTo(pt2.x, pt2.y); // Vertical line

    centerY = pt2.y;
    const centerX = pt2.x;
    pt1 = _types.default.newPoint(centerX, this.children[0].y + this.children[0].measureHeight(p_direction) / 2);
    pt2 = _types.default.newPoint(centerX, this.children[this.children.length - 1].y + this.children[this.children.length - 1].measureHeight(p_direction) / 2);
    p_path.moveTo(pt1.x, pt1.y);
    p_path.lineTo(pt2.x, pt2.y);
    this.children.forEach(node => {
      pt1 = _types.default.newPoint(node.x, node.y + node.measureHeight(p_direction) / 2);
      pt2 = _types.default.newPoint(centerX, pt1.y);
      p_path.moveTo(pt1.x, pt1.y);
      p_path.lineTo(pt2.x, pt2.y);
    });
  }

  return p_path;
}; // Tree as collection of node lines and root


const Tree = function (p_text, p_control, direction) {
  this.root = null;
  this.nRoots = 0;
  this.error = false;
  this.maxLevel = 0;
  this.levels = null;
  this.maxLevelHeights = null;
  this.control = p_control;
  this.direction = direction;
  this.lines = (0, _sitemapUtils.normalizeLevels)(p_text.trim().split("\n"));
  const textStyle = this.control.newStandardTextStyle(); // First, create unlinked nodes

  let levels = [];
  this.lines.forEach((line, index) => {
    const node = new Node(p_control, line, index, textStyle);

    if (node.level > this.maxLevel) {
      this.maxLevel = node.level;
    } // Add to its own level


    if (!levels[node.level]) {
      levels[node.level] = [];
    }

    levels[node.level].push(node); // Add as a child of the last node of the upper level

    if (node.level > 0 && levels[node.level - 1]) {
      const lastIndex = levels[node.level - 1].length - 1;
      levels[node.level - 1][lastIndex].addChild(node);
    } else {
      this.root = node;
      this.nRoots += 1;
    }
  });
  this.levels = levels;

  if (this.nRoots > 1) {
    this.root = new Node(p_control, ERROR_LINE, 0, textStyle);
    this.root.error = true;
  }

  this.hasSibling();
};

Tree.prototype.hasError = function () {
  return this.root.error;
};

Tree.prototype.getLines = function () {
  return this.lines;
};

Tree.prototype.calculateLineHeights = function () {
  // For each line get max height
  this.maxLevelHeights = this.levels.map(level => {
    const maxHeightNode = _balsamiq.default.maxBy(level, node => {
      return node ? node.measureHeight(this.direction) : 0;
    });

    return maxHeightNode.measureHeight ? maxHeightNode.measureHeight(this.direction) : 0;
  });
};

Tree.prototype.calculateLinesWidths = function () {
  // For each line get max width
  let width = this.levels.map(level => {
    const maxHeightNode = _balsamiq.default.maxBy(level, function (node) {
      return node ? node.measureWidth() : 0;
    });

    return maxHeightNode.measureWidth ? maxHeightNode.measureWidth() : 0;
  });
  return width;
};

Tree.prototype.measureHeight = function () {
  // For each line get max height
  this.calculateLineHeights(); // Sum all line heights

  let height = this.maxLevelHeights.reduce((memo, height) => {
    return memo + height;
  }, 0); // Sum vertical spacing

  height += CONSTANTS.node_vertical_spacing * (this.levels.length - 1);
  return height;
};

Tree.prototype.measureWidth = function () {
  return this.root.getSubtreeWidth(this.direction);
};

Tree.prototype.hasSibling = function () {
  let p_node = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;
  const r_node = p_node == null ? this.root : p_node;
  const sibling = r_node.children.length > 1;
  r_node.children.forEach(node => {
    node.siblings = sibling;
    this.hasSibling(node);
  });
  this.root.siblings = false;
};

Tree.prototype.makeDisplayItemsForSubtree = function (p_node, p_localBbox, p_path, p_fatherNode) {
  let parentOffset = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : {};
  const res = {
    displayItems: [],
    linkBounds: []
  };
  const {
    x: parentXOffset = 0,
    y: parentYOffset = 0
  } = parentOffset;
  const childrenOffset = {
    x: parentXOffset,
    y: parentYOffset
  };
  const lineHeight = this.maxLevelHeights[p_node.level];
  const width = p_node.measureWidth();
  let x, y;

  if (this.direction === Props.Directions.VERTICAL) {
    const subtreeWidth = p_node.getSubtreeWidth(this.direction);
    const actualSubtreeWidth = p_node.getActualSubtreeWidth(this.direction);
    x = parentXOffset + p_localBbox.x + (subtreeWidth - width) / 2;
    y = p_localBbox.y;

    if (width > actualSubtreeWidth) {
      childrenOffset.x = (width - actualSubtreeWidth) / 2;
    }
  } else {
    const subtreeHeight = p_node.getSubtreeHeight(this.direction);
    const actualSubtreeHeight = p_node.getActualSubtreeHeight(this.direction);
    const nodeHeight = p_node.measureHeight(this.direction);
    y = parentYOffset + p_localBbox.x + (subtreeHeight - nodeHeight) / 2;

    if (nodeHeight > actualSubtreeHeight) {
      childrenOffset.y = (nodeHeight - actualSubtreeHeight) / 2;
    }

    x = 0;
    const linesWidth = this.calculateLinesWidths();

    for (let i = 0; i < p_node.level; i++) {
      x += linesWidth[i] + CONSTANTS.node_horizontal_spacing;
    }
  }

  res.linkBounds.push(_types.default.newRectangle(x, y, width, lineHeight)); // Set node pos (it will be used to create boxes and edges)

  p_node.setPosition(x, y);
  res.displayItems = res.displayItems.concat(p_node.makeDisplayItems(this.direction));
  let subtreeX = p_localBbox.x;
  const subtreeY = p_localBbox.y + lineHeight + CONSTANTS.node_vertical_spacing;
  let subtreeW = 0;
  const subtreeH = p_localBbox.height - lineHeight - CONSTANTS.node_vertical_spacing;
  p_node.children.forEach(node => {
    subtreeW = node.getSubtreeWidth(this.direction);
    const res2 = this.makeDisplayItemsForSubtree(node, _types.default.newRectangle(subtreeX, subtreeY, subtreeW, subtreeH > 0 ? subtreeH : 0), p_path, p_node, childrenOffset);
    res.displayItems = res.displayItems.concat(res2.displayItems);
    res.linkBounds = res.linkBounds.concat(res2.linkBounds);
    subtreeX += subtreeW + 2 * (this.direction === Props.Directions.VERTICAL ? CONSTANTS.node_vertical_padding : CONSTANTS.node_horizontal_padding);
  });
  p_node.makeDisplayItemsForEdges(p_path, this.direction);
  return res;
};

Tree.prototype.makeDisplayItems = function (p_localBbox) {
  const path = Path.newPath(p_localBbox.width, p_localBbox.height);
  const res = this.makeDisplayItemsForSubtree(this.root, p_localBbox, path);
  const lineWidth = this.control.style("sitemap_edge_width");
  const borderColor = this.control.style("default_border_color");

  const lines = _displayitem.default.newSVGPath(p_localBbox, path, lineWidth, borderColor, null, null, null);

  res.displayItems.push(lines);
  return res;
};

control.CONTROL_PROTOTYPE.updateData = function () {
  const text = this.getActualProperty("text");
  this.tree = new Tree(text, this, this.getActualProperty("direction"));
  this._linkNames = this.tree.hasError() ? [] : this.tree.getLines().map(line => TextUtils.unformatText(line));
};

control.CONTROL_PROTOTYPE.getLinkNames = function () {
  return this._linkNames;
};

control.CONTROL_PROTOTYPE.getLinkableLabels = function () {
  return this.tree.hasError() ? [] : this.tree.getLines().map(TextUtils.unformatText);
};

control.CONTROL_PROTOTYPE.measure = function () {
  const direction = this.getActualProperty("direction");

  if (direction === Props.Directions.VERTICAL) {
    return _types.default.newSize(this.tree.measureWidth(), this.tree.measureHeight());
  } else {
    const widths = this.tree.calculateLinesWidths();
    let sum = 0;
    widths.forEach(w => sum += w);
    const width = sum + (widths.length - 1) * CONSTANTS.node_horizontal_spacing;
    this.tree.measureHeight(); //required call

    return _types.default.newSize(width, this.tree.root.getSubtreeHeight(direction));
  }
};

control.CONTROL_PROTOTYPE.updateLayout = function (localBbox) {
  const {
    linkBounds,
    displayItems
  } = this.tree.makeDisplayItems(localBbox);

  _link.default.processLinks(this, linkBounds);

  this.pushToDisplayList(displayItems);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/sliders.js":
/*!*************************************!*\
  !*** ./js/core/controls/sliders.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.sliderVertical = exports.sliderHorizontal = exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

const DEFAULT_WIDTH = 12;
const DEFAULT_HEIGHT = 12;

var makeDisplayItemsForSlider = function (p_bbox, p_track_selector, p_thumb_selector, p_value, p_color1, p_color2, p_vertical, p_skinOffset) {
  var thumbSize = this.styleSVGSize(p_thumb_selector);
  var trackSize = this.styleSVGSize(p_track_selector);
  var thumbBbox;
  let thumbPadding = this.style(p_thumb_selector + "_padding");

  if (p_vertical) {
    thumbBbox = _types.default.newRectangle(thumbPadding, (p_bbox.height - thumbSize.height) * p_value, DEFAULT_HEIGHT, DEFAULT_HEIGHT);
  } else {
    thumbBbox = _types.default.newRectangle((p_bbox.width - thumbSize.width) * p_value, thumbPadding, DEFAULT_WIDTH, DEFAULT_WIDTH);
  }

  var thumb = _displayitem.default.newPathsAsset(this._context, p_thumb_selector, thumbBbox, p_color2); // thumb position


  var trackbBox1 = _balsamiq.default.clone(p_bbox);

  var trackbBox2 = _balsamiq.default.clone(p_bbox);

  if (p_vertical) {
    trackbBox1.height = thumbBbox.y + thumbSize.height / 2;
    trackbBox1.x = trackbBox2.x = p_bbox.width / 2 - trackSize.width / 2 + p_skinOffset;
    trackbBox2.y = thumbBbox.y + thumbSize.height / 2;
    trackbBox2.height = p_bbox.height - trackbBox2.y;
  } else {
    trackbBox1.width = thumbBbox.x + thumbSize.width / 2;
    trackbBox1.y = trackbBox2.y = (p_bbox.height - trackSize.height) / 2 + p_skinOffset;
    trackbBox2.x = thumbBbox.x + thumbSize.width / 2;
    trackbBox2.width = p_bbox.width - trackbBox2.y;
  }

  var track1 = _displayitem.default.createDisplayItemsFromSkin(p_track_selector, p_color1, trackbBox1, this._context);

  var track2 = _displayitem.default.createDisplayItemsFromSkin(p_track_selector, p_color2, trackbBox2, this._context);

  track1.push(track2);
  track1.push(thumb);
  return track1;
};

var sliderVertical = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::VSlider",
  controlTypeName: "V.Slider",
  controlAlternateNames: ["Vertical Slider"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(DEFAULT_WIDTH, 33),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.VERTICAL
      }
    }),
    value: {
      defaultValue: 0.05,
      dataType: 'float01'
    },
    thumb: {
      defaultValue: Props.Thumbs.POINTY,
      enum: Props.Thumbs
    },
    state: {
      defaultValue: Props.BoolStates.NORMAL,
      enum: Props.BoolStates
    }
  }
});
exports.sliderVertical = sliderVertical;
controlsMod.addColorPropertyDefinitions(sliderVertical, {
  defaultColor: "slider_default_color"
});

sliderVertical.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  return _types.default.newSize(DEFAULT_WIDTH, p_sizeProperty.height);
};

sliderVertical.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var color1 = this.getActualProperty('state') === _types.default.ControlState.NORMAL ? this.getActualProperty("color") : this.style("slider_disabled_color");
  var color2 = this.getActualProperty('state') === _types.default.ControlState.NORMAL ? this.style("slider_default_color") : this.style("slider_disabled_color");
  var thumb = "slider_vertical_thumb_" + this.getActualProperty("thumb");
  var skinOffset = this._context.skin["#skinName"] === "sketch" ? 0 : 1.5;
  var displayList = makeDisplayItemsForSlider.call(this, p_localBbox, "slider_vertical_track", thumb, this.getActualProperty('value'), color1, color2, true, skinOffset);
  this.pushToDisplayList(displayList);
};

var sliderHorizontal = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::HSlider",
  controlTypeName: "H.Slider",
  controlAlternateNames: ["Horizontal Slider"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(33, DEFAULT_HEIGHT),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.HORIZONTAL
      }
    }),
    value: {
      defaultValue: 0.05,
      dataType: 'float01'
    },
    thumb: {
      defaultValue: Props.Thumbs.POINTY,
      enum: Props.Thumbs
    },
    state: {
      defaultValue: Props.BoolStates.NORMAL,
      enum: Props.BoolStates
    }
  }
});
exports.sliderHorizontal = sliderHorizontal;
controlsMod.addColorPropertyDefinitions(sliderHorizontal, {
  defaultColor: "slider_default_color"
});

sliderHorizontal.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  return _types.default.newSize(p_sizeProperty.width, DEFAULT_HEIGHT);
};

sliderHorizontal.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var color1 = this.getActualProperty('state') === _types.default.ControlState.NORMAL ? this.getActualProperty("color") : this.style("slider_disabled_color");
  var color2 = this.getActualProperty('state') === _types.default.ControlState.NORMAL ? this.style("slider_default_color") : this.style("slider_disabled_color");
  var skinOffset = this._context.skin["#skinName"] === "sketch" ? -0.5 : 0.5;
  var thumb = "slider_horizontal_thumb_" + this.getActualProperty("thumb");
  var displayList = makeDisplayItemsForSlider.call(this, p_localBbox, "slider_horizontal_track", thumb, this.getActualProperty('value'), color1, color2, false, skinOffset);
  this.pushToDisplayList(displayList);
};

var _default = {
  sliderVertical,
  sliderHorizontal
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/snapping.js":
/*!**************************************!*\
  !*** ./js/core/controls/snapping.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.SnapTypes = exports.SnapPositions = exports.SnapLine = exports.SnapDirections = exports.SnapConstraints = exports.QUICK_DRAW_GRAVITY = exports.PROXIMITY_LINE_COLOR = exports.PROXIMITY_DISTANCE = exports.NUDGING_GRAVITY = exports.MIDDLE_LINE_COLOR = exports.DRAGGING_GRAVITY = exports.COINCIDENCE_LINE_COLOR = void 0;
exports.makeSnapLinesForBbox = makeSnapLinesForBbox;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _selectionutils = __webpack_require__(/*! core/utils/selectionutils */ "./js/core/utils/selectionutils.js");

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const SnapDirections = {
  HORIZONTAL: "horizontal",
  VERTICAL: "vertical"
};
exports.SnapDirections = SnapDirections;
const SnapTypes = {
  COINCIDENCE: "coincidence",
  PROXIMITY: "proximity",
  MIDDLE: "center"
};
exports.SnapTypes = SnapTypes;
const SnapConstraints = {
  NONE: "none",
  VERTICAL: "vertical",
  HORIZONTAL: "horizontal"
};
exports.SnapConstraints = SnapConstraints;
const PROXIMITY_DISTANCE = 16; // distance between control's bbox and proximity snap lines

exports.PROXIMITY_DISTANCE = PROXIMITY_DISTANCE;
const DRAGGING_GRAVITY = 5; // distance of effect of a snap line when dragging controls with mouse/finger

exports.DRAGGING_GRAVITY = DRAGGING_GRAVITY;
const NUDGING_GRAVITY = 0; // distance of effect of a snap line when nudging with keyboard/inspector

exports.NUDGING_GRAVITY = NUDGING_GRAVITY;
const QUICK_DRAW_GRAVITY = 5; // distance of effect of a snap line when quick drawing

exports.QUICK_DRAW_GRAVITY = QUICK_DRAW_GRAVITY;
const COINCIDENCE_LINE_COLOR = 0xCCCCCC;
exports.COINCIDENCE_LINE_COLOR = COINCIDENCE_LINE_COLOR;
const PROXIMITY_LINE_COLOR = 0xD9D3A4;
exports.PROXIMITY_LINE_COLOR = PROXIMITY_LINE_COLOR;
const MIDDLE_LINE_COLOR = 0x89C9DE;
exports.MIDDLE_LINE_COLOR = MIDDLE_LINE_COLOR;

class SnapLine {
  constructor(type, direction, position) {
    _balsamiq.default.assertEnum(type, SnapTypes);

    _balsamiq.default.assertEnum(direction, SnapDirections);

    this._type = type;
    this._direction = direction;
    this._position = Math.round(position);
  }

  snapType() {
    return this._type;
  }

  snapDirection() {
    return this._direction;
  }

  snapPosition() {
    return this._position;
  }

  cloneWithOffset(offset) {
    var position = this._direction == SnapDirections.HORIZONTAL ? this._position += offset.y : this._position += offset.x;
    return new SnapLine(this._type, this._direction, position);
  }

}

exports.SnapLine = SnapLine;

function makeSnapLinesForBbox(bbox, isContainer) {
  _balsamiq.default.exists(bbox);

  _balsamiq.default.assertBool(isContainer);

  var DISTANCE = isContainer ? PROXIMITY_DISTANCE : -PROXIMITY_DISTANCE;
  return [new SnapLine(SnapTypes.MIDDLE, SnapDirections.VERTICAL, bbox.x + bbox.width / 2), new SnapLine(SnapTypes.MIDDLE, SnapDirections.HORIZONTAL, bbox.y + bbox.height / 2), new SnapLine(SnapTypes.COINCIDENCE, SnapDirections.VERTICAL, bbox.x), new SnapLine(SnapTypes.COINCIDENCE, SnapDirections.VERTICAL, bbox.x + bbox.width), new SnapLine(SnapTypes.COINCIDENCE, SnapDirections.HORIZONTAL, bbox.y), new SnapLine(SnapTypes.COINCIDENCE, SnapDirections.HORIZONTAL, bbox.y + bbox.height), new SnapLine(SnapTypes.PROXIMITY, SnapDirections.VERTICAL, bbox.x + DISTANCE), new SnapLine(SnapTypes.PROXIMITY, SnapDirections.VERTICAL, bbox.x + bbox.width - DISTANCE), new SnapLine(SnapTypes.PROXIMITY, SnapDirections.HORIZONTAL, bbox.y + DISTANCE), new SnapLine(SnapTypes.PROXIMITY, SnapDirections.HORIZONTAL, bbox.y + bbox.height - DISTANCE)];
}

class SnapPositions {
  constructor(gravity, offset) {
    _balsamiq.default.assert(_balsamiq.default.isNumber(gravity) && gravity >= 0, "gravity must a number >= 0");

    _balsamiq.default.exists(offset);

    this._gravity = gravity;
    this._offset = offset;
    /* this._positions =
    {
        [direction: SnapDirections]: {
            [type: SnapTypes]: {
                [position: Integer]: {
                    snapLine: SnapLine,
                    gravity: Integer,
                },
            },
        },
    }
    */

    this._positions = {};
    this.reset();
  }

  reset() {
    this._positions = {};

    for (var direction in SnapDirections) {
      this._positions[SnapDirections[direction]] = {};

      for (var type in SnapTypes) {
        this._positions[SnapDirections[direction]][SnapTypes[type]] = {};
      }
    }
  }

  addSnapLines(snapLines) {
    _balsamiq.default.assertArray(snapLines);

    snapLines.forEach(snapLine => {
      _balsamiq.default.existsClass(snapLine, SnapLine);

      var newSnapLine = snapLine.cloneWithOffset(this._offset);
      var position = newSnapLine.snapPosition();

      for (var i = -this._gravity; i <= this._gravity; i++) {
        this._positions[newSnapLine.snapDirection()][newSnapLine.snapType()][position + i] = {
          snapLine: newSnapLine,
          gravity: i
        };
      }
    });
  }

  _getSnapLine(direction, type, position) {
    return this._positions[direction][type][position] || {};
  }

  snapRectangleForMove(rect, constraint) {
    _balsamiq.default.exists(rect);

    _balsamiq.default.assertEnum(constraint, SnapConstraints);

    var snappedRect = _types.default.newRectangle(Math.round(rect.x), Math.round(rect.y), Math.round(rect.width), Math.round(rect.height));

    var rules = [{
      direction: SnapDirections.VERTICAL,
      position: _types.default.rectangleMiddleX(rect),
      types: [SnapTypes.MIDDLE],
      fun: (rect, snapPosition) => {
        rect.x = rect.x + snapPosition - _types.default.rectangleMiddleX(rect);
      }
    }, {
      direction: SnapDirections.VERTICAL,
      position: rect.x,
      types: [SnapTypes.COINCIDENCE, SnapTypes.PROXIMITY],
      fun: (rect, snapPosition) => {
        rect.x = snapPosition;
      }
    }, {
      direction: SnapDirections.VERTICAL,
      position: _types.default.rectangleRight(rect),
      types: [SnapTypes.COINCIDENCE, SnapTypes.PROXIMITY],
      fun: (rect, snapPosition) => {
        rect.x = snapPosition - rect.width;
      }
    }, {
      direction: SnapDirections.HORIZONTAL,
      position: _types.default.rectangleMiddleY(rect),
      types: [SnapTypes.MIDDLE],
      fun: (rect, snapPosition) => {
        rect.y = rect.y + snapPosition - _types.default.rectangleMiddleY(rect);
      }
    }, {
      direction: SnapDirections.HORIZONTAL,
      position: rect.y,
      types: [SnapTypes.COINCIDENCE, SnapTypes.PROXIMITY],
      fun: (rect, snapPosition) => {
        rect.y = snapPosition;
      }
    }, {
      direction: SnapDirections.HORIZONTAL,
      position: _types.default.rectangleBottom(rect),
      types: [SnapTypes.COINCIDENCE, SnapTypes.PROXIMITY],
      fun: (rect, snapPosition) => {
        rect.y = snapPosition - rect.height;
      }
    }];
    var snappedDirections = {};
    rules.forEach(rule => {
      rule.types.forEach(type => {
        if (!snappedDirections[rule.direction] && constraint !== rule.direction) {
          // if constrained do not snap in same direction
          var {
            snapLine
          } = this._getSnapLine(rule.direction, type, Math.round(rule.position));

          if (snapLine) {
            rule.fun(snappedRect, snapLine.snapPosition());
            snappedDirections[rule.direction] = true;
          }
        }
      });
    });

    const snappedLines = this._findSnappedLines(snappedRect);

    return {
      snappedRect,
      snappedLines
    };
  }

  snapRectangleForResize(rect, handle, ratio) {
    _balsamiq.default.exists(rect);

    _balsamiq.default.assertEnum(handle, _selectionutils.HandleTypes);

    _balsamiq.default.assertOptionalNumber(ratio);

    const snappedRect = _types.default.copyRectangle(rect);

    const rules = [{
      direction: SnapDirections.VERTICAL,
      position: rect.x,
      types: [SnapTypes.COINCIDENCE, SnapTypes.PROXIMITY],
      handles: [_selectionutils.HandleTypes.TOP_LEFT, _selectionutils.HandleTypes.LEFT, _selectionutils.HandleTypes.BOTTOM_LEFT],
      fun: (rect, snapPosition) => {
        rect.width += rect.x - snapPosition;
        rect.x = snapPosition;
      }
    }, {
      direction: SnapDirections.VERTICAL,
      position: _types.default.rectangleRight(rect),
      types: [SnapTypes.COINCIDENCE, SnapTypes.PROXIMITY],
      handles: [_selectionutils.HandleTypes.TOP_RIGHT, _selectionutils.HandleTypes.RIGHT, _selectionutils.HandleTypes.BOTTOM_RIGHT],
      fun: (rect, snapPosition) => {
        const newWidth = snapPosition - rect.x;
        rect.width = newWidth;
      }
    }, {
      direction: SnapDirections.VERTICAL,
      position: _types.default.rectangleMiddleX(rect),
      types: [SnapTypes.MIDDLE],
      handles: [_selectionutils.HandleTypes.TOP_LEFT, _selectionutils.HandleTypes.LEFT, _selectionutils.HandleTypes.BOTTOM_LEFT],
      fun: (rect, snapPosition) => {
        const oldRight = rect.x + rect.width;
        rect.width = (oldRight - snapPosition) * 2;
        rect.x = oldRight - rect.width;
      }
    }, {
      direction: SnapDirections.VERTICAL,
      position: _types.default.rectangleMiddleX(rect),
      types: [SnapTypes.MIDDLE],
      handles: [_selectionutils.HandleTypes.TOP_RIGHT, _selectionutils.HandleTypes.RIGHT, _selectionutils.HandleTypes.BOTTOM_RIGHT],
      fun: (rect, snapPosition) => {
        rect.width = (snapPosition - rect.x) * 2;
      }
    }, {
      direction: SnapDirections.HORIZONTAL,
      position: _types.default.rectangleBottom(rect),
      types: [SnapTypes.COINCIDENCE, SnapTypes.PROXIMITY],
      handles: [_selectionutils.HandleTypes.BOTTOM_LEFT, _selectionutils.HandleTypes.BOTTOM, _selectionutils.HandleTypes.BOTTOM_RIGHT],
      fun: (rect, snapPosition) => {
        rect.height = snapPosition - rect.y;
      }
    }, {
      direction: SnapDirections.HORIZONTAL,
      position: rect.y,
      types: [SnapTypes.COINCIDENCE, SnapTypes.PROXIMITY],
      handles: [_selectionutils.HandleTypes.TOP_LEFT, _selectionutils.HandleTypes.TOP, _selectionutils.HandleTypes.TOP_RIGHT],
      fun: (rect, snapPosition) => {
        rect.height += rect.y - snapPosition;
        rect.y = snapPosition;
      }
    }, {
      direction: SnapDirections.HORIZONTAL,
      position: _types.default.rectangleMiddleY(rect),
      types: [SnapTypes.MIDDLE],
      handles: [_selectionutils.HandleTypes.BOTTOM_LEFT, _selectionutils.HandleTypes.BOTTOM, _selectionutils.HandleTypes.BOTTOM_RIGHT],
      fun: (rect, snapPosition) => {
        rect.height = (snapPosition - rect.y) * 2;
      }
    }, {
      direction: SnapDirections.HORIZONTAL,
      position: _types.default.rectangleMiddleY(rect),
      types: [SnapTypes.MIDDLE],
      handles: [_selectionutils.HandleTypes.TOP_LEFT, _selectionutils.HandleTypes.TOP, _selectionutils.HandleTypes.TOP_RIGHT],
      fun: (rect, snapPosition) => {
        const oldBottom = _types.default.rectangleBottom(rect);

        rect.height = (oldBottom - snapPosition) * 2;
        rect.y = oldBottom - rect.height;
      }
    }];

    const keepRatioFunc = (direction, initialRect, rect) => {
      if (direction === SnapDirections.HORIZONTAL) {
        if (handle === _selectionutils.HandleTypes.TOP_LEFT || handle === _selectionutils.HandleTypes.BOTTOM_LEFT) {
          rect.width = snappedRect.height * ratio;
          rect.x -= rect.width - initialRect.width;
        } else if (handle === _selectionutils.HandleTypes.TOP_RIGHT || _selectionutils.HandleTypes.BOTTOM_RIGHT) {
          rect.width = rect.height * ratio;
        }
      } else {
        if (handle === _selectionutils.HandleTypes.TOP_RIGHT || handle === _selectionutils.HandleTypes.TOP_LEFT) {
          rect.height = rect.width / ratio;
          rect.y -= rect.height - initialRect.height;
        } else if (handle === _selectionutils.HandleTypes.BOTTOM_RIGHT || handle === _selectionutils.HandleTypes.BOTTOM_LEFT) {
          rect.height = rect.width / ratio;
        }
      }
    };

    const snappedDirections = {};
    rules.forEach(rule => {
      if (rule.handles.indexOf(handle) !== -1) {
        rule.types.forEach(type => {
          if (!snappedDirections[rule.direction]) {
            const {
              snapLine,
              gravity
            } = this._getSnapLine(rule.direction, type, Math.round(rule.position)); // consider that since we use integer snap lines coordinates, middle values affect 2 points
            // example: with a width of 10 the middle is 5, but we use the same snap line at 5 with a width of 9
            // so only when resizing we need to half the middle gravity influence


            const gravityThresholdMiddleLines = Math.floor(this._gravity / 2);

            if (snapLine && (type !== SnapTypes.MIDDLE || Math.abs(gravity) <= gravityThresholdMiddleLines)) {
              const initialRect = { ...snappedRect
              };
              rule.fun(snappedRect, snapLine.snapPosition());

              if (ratio) {
                keepRatioFunc(rule.direction, initialRect, snappedRect);
              }

              snappedDirections[rule.direction] = true;
            }
          }
        });
      }
    });

    const snappedLines = this._findSnappedLines(snappedRect);

    return {
      snappedRect,
      snappedLines
    };
  }

  snapLinesForCoords(coords) {
    const candidates = [{
      direction: SnapDirections.VERTICAL,
      position: coords.x,
      types: [SnapTypes.COINCIDENCE, SnapTypes.PROXIMITY]
    }, {
      direction: SnapDirections.VERTICAL,
      position: coords.x,
      types: [SnapTypes.MIDDLE]
    }, {
      direction: SnapDirections.HORIZONTAL,
      position: coords.y,
      types: [SnapTypes.COINCIDENCE, SnapTypes.PROXIMITY]
    }, {
      direction: SnapDirections.HORIZONTAL,
      position: coords.y,
      types: [SnapTypes.MIDDLE]
    }];
    const snapLines = []; // only keep the first vertical and horizontal

    let hasVerticalSnapLine = false;
    let hasHorizontalSnapLine = false;

    for (const candidate of candidates) {
      for (const type of candidate.types) {
        const {
          snapLine
        } = this._getSnapLine(candidate.direction, type, candidate.position);

        if (snapLine) {
          if (candidate.direction === SnapDirections.VERTICAL && !hasVerticalSnapLine) {
            snapLines.push(snapLine);
            hasVerticalSnapLine = true;
          }

          if (candidate.direction === SnapDirections.HORIZONTAL && !hasHorizontalSnapLine) {
            snapLines.push(snapLine);
            hasHorizontalSnapLine = true;
          }
        }
      }
    }

    return snapLines;
  }

  _findSnappedLines(snappedRect) {
    _balsamiq.default.exists(snappedRect);

    const snappedLines = [];
    const candidates = [{
      direction: SnapDirections.VERTICAL,
      position: snappedRect.x,
      types: [SnapTypes.COINCIDENCE, SnapTypes.PROXIMITY]
    }, {
      direction: SnapDirections.VERTICAL,
      position: _types.default.rectangleRight(snappedRect),
      types: [SnapTypes.COINCIDENCE, SnapTypes.PROXIMITY]
    }, {
      direction: SnapDirections.VERTICAL,
      position: _types.default.rectangleMiddleX(snappedRect),
      types: [SnapTypes.MIDDLE]
    }, {
      direction: SnapDirections.HORIZONTAL,
      position: snappedRect.y,
      types: [SnapTypes.COINCIDENCE, SnapTypes.PROXIMITY]
    }, {
      direction: SnapDirections.HORIZONTAL,
      position: _types.default.rectangleBottom(snappedRect),
      types: [SnapTypes.COINCIDENCE, SnapTypes.PROXIMITY]
    }, {
      direction: SnapDirections.HORIZONTAL,
      position: _types.default.rectangleMiddleY(snappedRect),
      types: [SnapTypes.MIDDLE]
    }];
    candidates.forEach(candidate => {
      candidate.types.forEach(type => {
        const {
          snapLine
        } = this._getSnapLine(candidate.direction, type, candidate.position);

        if (snapLine && snapLine.snapPosition() === candidate.position) {
          snappedLines.push(snapLine);
        }
      });
    });
    return snappedLines;
  }

}

exports.SnapPositions = SnapPositions;

/***/ }),

/***/ "./js/core/controls/splitterhorizontal.js":
/*!************************************************!*\
  !*** ./js/core/controls/splitterhorizontal.js ***!
  \************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

const HEIGHT = 12;
var control = controlsMod.makeAssetBasedControlClass({
  controlTypeID: "com.balsamiq.mockups::VSplitter",
  // VSplitter: this is not an error
  controlTypeName: "H.Splitter",
  controlAlternateNames: [],
  defaultSize: _types.default.newSize(100, HEIGHT),
  resizeOption: Props.ResizeOptions.HORIZONTAL,
  skinAssetName: null,
  colorStyleName: null,
  otherPropertyDefinitions: null
});

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  return _types.default.newSize(p_sizeProperty.width, 12);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var track = _displayitem.default.createDisplayItemsFromSkin("splitter_horizontal_track", this.style("default_splitter_color"), p_localBbox, this._context);

  var thumbSize = this.styleSVGSize("splitter_horizontal_thumb");

  var thumbBbox = _types.default.newRectangle((p_localBbox.width - thumbSize.width) / 2, (p_localBbox.height - thumbSize.height) / 2, thumbSize.width, thumbSize.height);

  var thumb = _displayitem.default.newPathsAsset(this._context, "splitter_horizontal_thumb", thumbBbox, this.style("default_splitter_color"));

  this.pushToDisplayList(track, thumb);
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  return _types.default.newSize(p_sizeProperty.width, HEIGHT);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/splittervertical.js":
/*!**********************************************!*\
  !*** ./js/core/controls/splittervertical.js ***!
  \**********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

const WIDTH = 12;
var control = controlsMod.makeAssetBasedControlClass({
  controlTypeID: "com.balsamiq.mockups::HSplitter",
  // HSplitter: this is not an error
  controlTypeName: "V.Splitter",
  controlAlternateNames: ["Vertical Splitter", "Separator", "Drag Bar"],
  defaultSize: _types.default.newSize(WIDTH, 100),
  resizeOption: Props.ResizeOptions.VERTICAL,
  skinAssetName: null,
  colorStyleName: null,
  otherPropertyDefinitions: null
});

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  return _types.default.newSize(12, p_sizeProperty.height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var track = _displayitem.default.createDisplayItemsFromSkin("splitter_vertical_track", this.style("default_splitter_color"), p_localBbox, this._context);

  var thumbSize = this.styleSVGSize("splitter_vertical_thumb");

  var thumbBbox = _types.default.newRectangle((p_localBbox.width - thumbSize.width) / 2, (p_localBbox.height - thumbSize.height) / 2, thumbSize.width, thumbSize.height);

  var thumb = _displayitem.default.newPathsAsset(this._context, "splitter_vertical_thumb", thumbBbox, this.style("default_splitter_color"));

  this.pushToDisplayList(track, thumb);
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  return _types.default.newSize(WIDTH, p_sizeProperty.height);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/stickynote.js":
/*!****************************************!*\
  !*** ./js/core/controls/stickynote.js ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _macro = __webpack_require__(/*! core/bricks/macro */ "./js/core/bricks/macro.js");

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::StickyNote",
  controlTypeName: "Comment",
  controlAlternateNames: ["Sticky Note", "To Do"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        resize: Props.ResizeOptions.BOTH,
        autosize: Props.AutoSizeOptions.NO
      }
    }),
    align: {
      defaultValue: Props.TextAlignments.CENTER,
      enum: Props.TextAlignments
    },
    markup: {
      defaultValue: true,
      dataType: 'bool'
    }
  }
});
Controls.addStandardTextPropertyDefinitions(control, {
  text: "A comment",
  multiline: true
});
Controls.addColorPropertyDefinitions(control, {
  defaultColor: 0xF1C232
});

control.CONTROL_PROTOTYPE.updateData = function () {
  const color = _colors.default.adjustBrightness(this.getActualProperty("color"), this.style("stickynote_brightness_offset"));

  const textStyle = this.newStandardTextStyle({
    textColor: _colors.default.labelColorGivenBackgroundColor(color, 1, this.style("default_text_color")),
    linkColor: _colors.default.labelColorGivenBackgroundColor(color, 1, this.style("default_link_color"))
  });
  const text = (0, _macro.processMacros)(this);

  if (!this.text || !Text.areTextsEqual(this.text, text, textStyle)) {
    this.text = Text.newText(text, textStyle, this._context, true);
  }
};

_link.default.addTextLinksMethods(control);

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  return _types.default.newSize(109, 123);
};

control.CONTROL_PROTOTYPE.updateLayout = function (localBbox) {
  const paddingLeft = this.style("stickynote_padding_left");
  const paddingRight = this.style("stickynote_padding_right");
  const paddingTop = this.style("stickynote_padding_top");
  const paddingBottom = this.style("stickynote_padding_bottom");
  let tape;

  let tapeSkinSize = _types.default.newSize(0, 0);

  if (_balsamiq.default.has(this._context.skin, "stickynote_tape")) {
    tapeSkinSize = this.styleSVGSize("stickynote_tape");
    const tapeWidth = Math.max(tapeSkinSize.width, localBbox.width / 4);

    const tapeBounds = _types.default.newRectangle(Math.round((localBbox.width - tapeWidth) / 2), 0, tapeWidth, tapeSkinSize.height);

    tape = _displayitem.default.createDisplayItemsFromSkin("stickynote_tape", null, tapeBounds, this._context);
  }

  const paperBounds = _types.default.paddedRectangle(localBbox, Math.round(tapeSkinSize.height / 2));

  const paper = _displayitem.default.createDisplayItemsFromSkin("stickynote_paper", null, paperBounds, this._context);

  _displayitem.default.addColorTransformToDisplayItems(this._context.skin, paper, this.getActualProperty("color"));

  const textWidth = Math.min(localBbox.width - paddingLeft - paddingRight, this.text.measuredSize().width);
  const textHeight = localBbox.height - paddingTop - paddingBottom;
  let label;

  if (textHeight > 0 && textWidth > 0) {
    const align = this.getActualProperty('align');
    const textBounds = align === _types.default.TextAlign.LEFT ? _types.default.newRectangle(paddingLeft, paddingTop, textWidth, textHeight) : align === _types.default.TextAlign.LEFT ? _types.default.newRectangle(localBbox.width - paddingLeft - textWidth, paddingTop, textWidth, textHeight) : _types.default.newRectangle((localBbox.width - textWidth) / 2, paddingTop, textWidth, textHeight);
    label = this.text.makeDisplayItemsAndLinks(textBounds, this.getActualProperty("align"), _types.default.Positions.TOP);

    _link.default.processTextLinks(this, label.links);
  }

  this.pushToDisplayList(paper, tape, label ? label.displayItems : null);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/streetmap.js":
/*!***************************************!*\
  !*** ./js/core/controls/streetmap.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var _default = controlsMod.makeAssetBasedControlClass({
  controlTypeID: "com.balsamiq.mockups::Map",
  controlTypeName: "Street Map",
  controlAlternateNames: [],
  defaultSize: _types.default.newSize(252, 222),
  resizeOption: Props.ResizeOptions.BOTH,
  skinAssetName: "streetmap",
  colorStyleName: null,
  otherPropertyDefinitions: null
});

exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/switch.js":
/*!************************************!*\
  !*** ./js/core/controls/switch.js ***!
  \************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var _switch = __webpack_require__(/*! core/bricks/switch */ "./js/core/bricks/switch.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::Switch",
  controlTypeName: "ON/OFF Switch",
  controlAlternateNames: ["Toggle"]
});
Controls.addPropertyDefinitions(control, {
  onOffState: {
    defaultValue: _types.default.OnOffState.ON
  },
  ...Props.PropertyDefinitions.link
});
Controls.addColorPropertyDefinitions(control, {
  defaultColor: "switch_default_color"
});

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this);

  this.pushToDisplayList((0, _switch.makeSwitchDisplayItems)(this._context, p_localBbox, this.getActualProperty("onOffState") === _types.default.OnOffState.ON, this.getActualProperty("color")));
};

control.CONTROL_PROTOTYPE.measure = function () {
  return _types.default.newSize(42, 24);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/symbol.js":
/*!************************************!*\
  !*** ./js/core/controls/symbol.js ***!
  \************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.canConvertToSymbol = canConvertToSymbol;
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _constants = __webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts");

var MockupsJSON = _interopRequireWildcard(__webpack_require__(/*! core/serialization/json */ "./js/core/serialization/json.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _frame = _interopRequireDefault(__webpack_require__(/*! core/bricks/frame */ "./js/core/bricks/frame.js"));

var _symbollibrary = _interopRequireDefault(__webpack_require__(/*! core/model/symbollibrary */ "./js/core/model/symbollibrary.js"));

var _group = _interopRequireDefault(__webpack_require__(/*! core/controls/group */ "./js/core/controls/group.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var ControlID = _interopRequireWildcard(__webpack_require__(/*! core/controls/controlid */ "./js/core/controls/controlid.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _controls = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var _coreobject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var _wireframeVersion = __webpack_require__(/*! core/wireframe-version */ "./js/core/wireframe-version.ts");

var _BmprConstants = __webpack_require__(/*! @balsamiq/bmpr/lib/BmprConstants */ "./node_modules/@balsamiq/bmpr/lib/BmprConstants.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const controlTypeID = "com.balsamiq.mockups::Component";
const defaultSymbolDefinitionProps = {
  crop: _types.default.Crop.ALL,
  markup: false
};
const control = (0, _controls.makeControlClass)({
  controlTypeID,
  controlTypeName: "Symbol",
  controlAlternateNames: [],
  props: {
    src: {
      defaultValue: null,
      dataType: "object",
      onSet: function (p_value) {
        this.state = null;
      }
    },
    overrides: {
      defaultValue: [],
      dataType: "object"
    }
  }
});

control.CONTROL_PROTOTYPE.hasSymbolDefinitionProperty = function (name) {
  return this._symbolDefinitionPropNames.indexOf(name) >= 0;
};

control.CONTROL_PROTOTYPE.getActualSymbolDefinitionProperty = function (name) {
  return _balsamiq.default.has(this._symbolDefinitionOverriddenProps, name) ? this._symbolDefinitionOverriddenProps[name] : this._symbolDefinitionProps[name];
};

control.CONTROL_PROTOTYPE.getNonDefaultSymbolDefinitionProperties = function () {
  return Object.keys(this._symbolDefinitionProps).reduce((nonDefaultSymbolDefinitionsProperties, propName) => {
    const propValue = this._symbolDefinitionProps[propName];

    if (defaultSymbolDefinitionProps[propName] !== propValue) {
      nonDefaultSymbolDefinitionsProperties[propName] = propValue;
    }

    return nonDefaultSymbolDefinitionsProperties;
  }, {});
};

control.CONTROL_PROTOTYPE.getNonDefaultActualControlProperties = function () {
  const actualProperties = this.superClass.prototype.getNonDefaultActualControlProperties.call(this);
  return { ...this.getNonDefaultSymbolDefinitionProperties(),
    ...actualProperties
  };
};

control.CONTROL_PROTOTYPE.getActualProperty = function (p_name) {
  return this.hasSymbolDefinitionProperty(p_name) ? this.getActualSymbolDefinitionProperty(p_name) : this.superClass.prototype.getActualProperty.call(this, ...arguments);
};

control.CONTROL_PROTOTYPE.getProperty = function (p_name, p_overridden) {
  if (this.hasSymbolDefinitionProperty(p_name)) {
    if (p_overridden) {
      return this._symbolDefinitionOverriddenProps[p_name];
    } else {
      return this._symbolDefinitionProps[p_name];
    }
  } else {
    return this.superClass.prototype.getProperty.call(this, p_name, p_overridden);
  }
};

control.CONTROL_PROTOTYPE.hasPropertyDefinition = function (p_name) {
  if (this.hasSymbolDefinitionProperty(p_name)) {
    return true;
  } else {
    return this.superClass.prototype.hasPropertyDefinition.call(this, ...arguments);
  }
};

control.CONTROL_PROTOTYPE.getPropertyDefinitions = function () {
  var defs = this.superClass.prototype.getPropertyDefinitions.call(this, ...arguments);

  this._symbolDefinitionPropNames.forEach(prop => {
    _balsamiq.default.exists(this._symbolDefinitionDataTypes[prop], `The dataType for property ${prop} must be defined`);

    defs[prop] = this._symbolDefinitionDataTypes[prop];
  });

  return defs;
};

control.CONTROL_PROTOTYPE.getPropertyDataType = function (p_name) {
  if (this.hasSymbolDefinitionProperty(p_name)) {
    return this._symbolDefinitionDataTypes[p_name];
  } else {
    return this.superClass.prototype.getPropertyDataType.call(this, ...arguments);
  }
};

control.CONTROL_PROTOTYPE._setProperty = function (p_name, p_value, p_overridden) {
  if (this.hasSymbolDefinitionProperty(p_name)) {
    if (p_value === undefined) {
      delete this._symbolDefinitionOverriddenProps[p_name];
    } else {
      this._symbolDefinitionOverriddenProps[p_name] = p_value;
    }

    this.updateSelfOverrides();
  } else {
    this.superClass.prototype._setProperty.call(this, p_name, p_value, p_overridden);
  }
};

control.CONTROL_PROTOTYPE.getPropertyOptions = function (p_name) {
  return {};
};

control.CONTROL_PROTOTYPE.isPropertyOverridden = function (p_name) {
  if (this.hasSymbolDefinitionProperty(p_name)) {
    return _balsamiq.default.has(this._symbolDefinitionOverriddenProps, p_name);
  } else {
    return this.superClass.prototype.isPropertyOverridden.call(this, ...arguments);
  }
};

control.CONTROL_PROTOTYPE.initSymbolDefinitionPropsData = function () {
  this._symbolDefinitionProps = { ...defaultSymbolDefinitionProps
  };
  this._symbolDefinitionDataTypes = {
    crop: "object",
    markup: "bool"
  };
  this._symbolDefinitionOverriddenProps = {};
};

control.CONTROL_PROTOTYPE.initData = function () {
  this._clipOnBbox = false;
  this.state = _types.default.SymbolState.BROKEN;
  this.isProcessingChildren = false;
  this.initSymbolDefinitionPropsData();
  this._symbolDefinitionPropNames = Object.keys(this._symbolDefinitionProps);
};

control.CONTROL_PROTOTYPE.onChildBboxChange = function (p_childControlID) {
  if (!this.isProcessingChildren) {
    // do not call updateAll() because otherwise the cropOffset is updated and when moving
    // an inner element the other elements move too.
    this.updateBBoxes();
    this.updateCropOffsetMask();

    this._context.notifyChange(this.coreObjectID, _constants.ControlChanges.DISPLAY_LIST);
  }
};

control.CONTROL_PROTOTYPE.onResourceChange = function (resourceID) {
  if (this.isProcessingChildren) return;
  this.processChildren();
  this.updateAll();
};

control.CONTROL_PROTOTYPE.onChildrenUpdate = function () {
  this.updateOverridesFromChildren();
  this.updateSelfOverrides();
  this.processChildren();
  this.updateAll();
};

control.CONTROL_PROTOTYPE.getVersion = function () {
  const mockup = this.getMockup();
  return mockup.getType() === _constants.ResourceTypes.SYMBOLLIBRARY ? _wireframeVersion.NestedSymbolVersion : _wireframeVersion.BaseVersion;
};

control.CONTROL_PROTOTYPE.getSymbolLibrary = function () {
  const src = this.getActualProperty("src");

  if (src) {
    return this._context.getDependency(src.resourceID);
  }

  return null;
};

control.CONTROL_PROTOTYPE.getSymbolDefinition = function () {
  var symbolDefinition = null;
  var src = this.getActualProperty("src");

  if (src) {
    if (this._context.hasSymbolDefinitionALoop(src)) {
      return null;
    }

    let resource = this.getSymbolLibrary();

    if (resource) {
      _balsamiq.default.existsClass(resource, _symbollibrary.default);

      const subResourceID = src.subResourceID === "" ? -1 : parseInt(src.subResourceID);
      symbolDefinition = resource.getSymbolDefinition(subResourceID);
    }
  }

  return symbolDefinition;
};

control.CONTROL_PROTOTYPE.processChildren = function () {
  _balsamiq.default.assert(this.isProcessingChildren === false);

  this.isProcessingChildren = true;
  this.removeAllChildren();
  const symbolDefinition = this.getSymbolDefinition();

  if (symbolDefinition) {
    this.state = _types.default.SymbolState.LOADED;
    copyChildren(this, symbolDefinition);

    if (this.isInsideSymbol()) {
      this.overrideChildrenProperties({
        asProperties: true
      });
    } else {
      this.children.forEach(child => {
        child.activateOverriddenProperties();
      });
      this.overrideChildrenProperties();
    }

    this.children.forEach(child => {
      child.updateAll();
    });
  } else {
    this.state = _types.default.SymbolState.BROKEN;
  }

  this.isProcessingChildren = false;
}; // Read properties from Group that defines the symbol
// Crop


control.CONTROL_PROTOTYPE.getPropsFromSymbolDefinition = function () {
  var symbolDefinition = this.getSymbolDefinition();

  if (symbolDefinition) {
    this._symbolDefinitionPropNames.forEach(prop => this._symbolDefinitionProps[prop] = _balsamiq.default.clone(symbolDefinition.getActualProperty(prop)));
  }
};

control.CONTROL_PROTOTYPE.updateData = function () {
  const src = this.getActualProperty("src");
  this.setDependencies(src ? [src] : []);

  if ((!this.state || this.state === _types.default.SymbolState.BROKEN) && !this.isProcessingChildren) {
    this.processChildren();
  }

  this.initSymbolDefinitionPropsData();
  this.getPropsFromSymbolDefinition();
  this.overrideSelfProperties();

  this._context.setReady(this.coreObjectID, (this.state === _types.default.SymbolState.LOADED || this.state === _types.default.SymbolState.BROKEN) && !this.isProcessingChildren);
};

control.CONTROL_PROTOTYPE.measure = function () {
  // Union of all user bbox in component space
  var userBBox;

  if (this.state === _types.default.SymbolState.LOADED && this.children.length > 0) {
    userBBox = this.calcChildrenUserBBoxCS(false); // TODO markup

    userBBox = _types.default.calcCroppedSize(_types.default.newRectangleFromCrop(this.getActualProperty('crop')), {
      width: userBBox.width,
      height: userBBox.height
    });
  } else {
    userBBox = _types.default.newRectangle(0, 0, 150, 150);
  }

  return userBBox;
};

control.CONTROL_PROTOTYPE.updateClipBBoxPS = function () {
  var crop = _types.default.newRectangleFromCrop(this.getActualProperty('crop'));

  if (_types.default.isCropAll(crop)) {
    let childrenUserBBox = this.calcChildrenUserBBoxCS(false);
    let childrenClipBBox;

    if (!this.state || this.state === _types.default.SymbolState.BROKEN) {
      childrenClipBBox = _types.default.newRectangle(0, 0, 150, 150);
    } else if (this.state === _types.default.SymbolState.LOADED) {
      childrenClipBBox = this.calcChildrenClipBBoxCS(false);
    }

    var delta = _types.default.newPoint(childrenClipBBox.x - childrenUserBBox.x, childrenClipBBox.y - childrenUserBBox.y);

    this.setClipBBoxPS(this._userBBox.x + delta.x, this._userBBox.y + delta.y, childrenClipBBox.width, childrenClipBBox.height);
  } else {
    // If we have crop use userBBox that is already correctly cropped!
    this.setClipBBoxPS(this._userBBox.x, this._userBBox.y, this._userBBox.width, this._userBBox.height);
  }
};

control.CONTROL_PROTOTYPE.updateUserBBoxNoMarkupPS = function (userBBox) {
  // Get children user bbox in component space
  var childrenBBoxNoMarkup = this.calcChildrenUserBBoxCS(true);

  if (this.getActualProperty('crop') === _types.default.Crop.ALL) {
    let origin = this.getActualProperty('origin');
    this.setUserBBoxNoMarkupPS(origin.x + childrenBBoxNoMarkup.x, origin.y + childrenBBoxNoMarkup.y, childrenBBoxNoMarkup.width, childrenBBoxNoMarkup.height);
  } else {
    let origin = _types.default.calcUncroppedOrigin(_types.default.newRectangleFromCrop(this.getActualProperty('crop')), userBBox); //REVERSE ORIGIN in UNCROPPED POSITION //REVERSE ORIGIN in UNCROPPED POSITION


    let newRect = _types.default.newRectangle(origin.x + childrenBBoxNoMarkup.x, origin.y + childrenBBoxNoMarkup.y, childrenBBoxNoMarkup.width, childrenBBoxNoMarkup.height);

    let intersect = _types.default.rectanglesIntersect(newRect, userBBox);

    this.setUserBBoxNoMarkupPS(intersect.x, intersect.y, intersect.width, intersect.height);
  }
};

control.CONTROL_PROTOTYPE.updateClipBBoxNoMarkupPS = function (clipBBox) {
  var childrenBBoxNoMarkup = this.calcChildrenClipBBoxCS(true);

  if (this.getActualProperty('crop') === _types.default.Crop.ALL) {
    let origin = this.getActualProperty('origin');
    this.setClipBBoxNoMarkupPS(origin.x + childrenBBoxNoMarkup.x, origin.y + childrenBBoxNoMarkup.y, childrenBBoxNoMarkup.width, childrenBBoxNoMarkup.height);
  } else {
    let origin = _types.default.calcUncroppedOrigin(_types.default.newRectangleFromCrop(this.getActualProperty('crop')), this._userBBox); //REVERSE ORIGIN in UNCROPPED POSITION


    let newRect = _types.default.newRectangle(origin.x + childrenBBoxNoMarkup.x, origin.y + childrenBBoxNoMarkup.y, childrenBBoxNoMarkup.width, childrenBBoxNoMarkup.height);

    let intersect = _types.default.rectanglesIntersect(newRect, clipBBox);

    this.setClipBBoxNoMarkupPS(intersect.x, intersect.y, intersect.width, intersect.height);
  }
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  if (this.state === _types.default.SymbolState.BROKEN) {
    this.pushToDisplayList(_frame.default.makeImageFrame(this, p_localBbox, "Symbol not found"));
    this.resetCropData();
  } else if (this.state === _types.default.SymbolState.LOADING) {
    this.pushToDisplayList(_frame.default.makeImageFrame(this, p_localBbox));
    this.resetCropData();
  } else if (this.state === _types.default.SymbolState.LOADED && this.isEmpty()) {
    this.pushToDisplayList(_frame.default.makeImageFrame(this, p_localBbox, "Empty Symbol"));
    this.resetCropData();
  }
};

control.CONTROL_PROTOTYPE.getUncroppedUserBBox = function (includeMarkup) {
  return _group.default.CONTROL_PROTOTYPE.getUncroppedUserBBox.call(this, includeMarkup);
};

control.CONTROL_PROTOTYPE.saveProperties = function () {
  this.updateOverridesFromChildren();
  this.updateSelfOverrides();
  return this.superClass.prototype.saveProperties.call(this);
};

control.CONTROL_PROTOTYPE.hasSymbolDefinitionOverriddenProps = function () {
  return !_balsamiq.default.isEmpty(this._symbolDefinitionOverriddenProps);
};

control.CONTROL_PROTOTYPE.updateSelfOverrides = function () {
  const overridesProp = _balsamiq.default.clone(this.getActualProperty('overrides'));

  const index = overridesProp.findIndex(override => override.controlID[0] === -1);
  let changed = false;

  if (this.hasSymbolDefinitionOverriddenProps()) {
    let selfOverrideProps = {};

    for (const [key, value] of Object.entries(this._symbolDefinitionOverriddenProps)) {
      MockupsJSON.saveProperty(selfOverrideProps, key, value);
    }

    if (index === -1) {
      overridesProp.push({
        controlID: [-1],
        properties: selfOverrideProps
      });
    } else {
      overridesProp[index] = {
        controlID: [-1],
        properties: selfOverrideProps
      };
    }

    changed = true;
  } else {
    if (index !== -1) {
      overridesProp.splice(index, 1);
      changed = true;
    }
  }

  if (changed) {
    this.setActualProperty('overrides', overridesProp);
  }
};
/**
 *
 * @param {{ controlID: number[], properties: { [prop: string]: any } }[]} overrides
 */


control.CONTROL_PROTOTYPE.deleteOverridesFromLoadedProps = function (overrides) {
  if (this.loadedProperties == null || this.loadedProperties.overrides == null) {
    return;
  }

  for (let index = 0; index < overrides.length; index++) {
    const override = overrides[index];
    const valueInLoadedProps = this.loadedProperties.overrides.find(loadedProp => loadedProp.ID === override.controlID.join(":"));

    for (const overridePropName in override.properties) {
      const overridePropValue = override.properties[overridePropName];

      if (valueInLoadedProps != null && overridePropValue === valueInLoadedProps[overridePropName]) {
        continue;
      }

      delete this.loadedProperties.overrides;
      return;
    }
  }

  for (let index = 0; index < this.loadedProperties.overrides.length; index++) {
    const loadedPropOverride = this.loadedProperties.overrides[index];
    const valueInOverrides = overrides.find(override => loadedPropOverride.ID === override.controlID.join(":"));

    for (const loadedPropName in loadedPropOverride) {
      if (valueInOverrides != null && valueInOverrides.properties[loadedPropName] != null) {
        continue;
      }

      const controlID = loadedPropOverride.ID.split(":").map(controlID => parseInt(controlID));
      const control = this.getControl(controlID);
      let propertyToCheck = loadedPropName;

      if (["x", "y"].includes(loadedPropName)) {
        propertyToCheck = "origin";
      } else if (["w", "h"].includes(loadedPropName)) {
        propertyToCheck = "size";
      }

      if (control && control.hasPropertyDefinition(propertyToCheck)) {
        delete this.loadedProperties.overrides;
        return;
      }
    }
  }
};

control.CONTROL_PROTOTYPE.updateOverridesFromChildren = function () {
  if (this.getSymbolDefinition() == null) return;
  const overrides = [];

  const getChildOverrides = child => {
    const props = child.saveOverriddenProperties();

    if (!_balsamiq.default.isEmpty(props)) {
      const controlID = ControlID.getRelative(this.coreObjectID.controlID, child.coreObjectID.controlID);

      if (props["overrides"]) {
        props["overrides"] = props["overrides"].map(override => {
          const controlID = override.ID.split(":").map(parseInt);
          return {
            controlID,
            properties: _lodash.default.omit(override, ["ID"])
          };
        });
      }

      overrides.push({
        controlID,
        properties: props
      });
    }

    if (child.children.length > 0) {
      child.children.forEach(getChildOverrides);
    }
  };

  this.children.forEach(getChildOverrides);
  this.deleteOverridesFromLoadedProps(overrides); // update the property without calling the setter

  if (!_balsamiq.default.isEqual(overrides, this.getDefaultProperty('overrides'))) {
    this.properties['overrides'] = overrides;
  } else {
    delete this.properties['overrides'];
  }
};

control.CONTROL_PROTOTYPE.overrideChildrenProperties = function () {
  let {
    asProperties = false
  } = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  let allOps = this.getActualProperty('overrides');
  allOps.forEach(childOps => {
    var childControl = this.getControl(childOps.controlID);

    if (childControl) {
      if (asProperties) {
        childControl.loadOverriddenPropertiesAsProperties(childOps.properties);
      } else {
        childControl.loadOverriddenProperties(childOps.properties);
      }
    } else {//this._context.logger("ignoring overridden properties of controlID:" + JSON.stringify(childOps.controlID));
    }
  });
};

control.CONTROL_PROTOTYPE.overrideSelfProperties = function () {
  var symbolDefinition = this.getSymbolDefinition();

  if (!symbolDefinition) {
    return;
  }

  let allOps = this.getActualProperty('overrides');
  allOps.forEach(childOps => {
    if (childOps.controlID[0] === -1) {
      let names = [];
      let values = [];

      for (let key in childOps.properties) {
        names.push(key);
        values.push(childOps.properties[key]);
      } // Only crop
      //Balsamiq.assert(names.length <= 1);
      //Balsamiq.assert(names.length === 0 || (names.length === 1 && (names[0] === 'crop')));


      names.forEach((name, index) => {
        if (this.hasSymbolDefinitionProperty(name)) {
          if (values[index] === "true" || values[index] === "false") {
            const value = values[index] === "true";
            if (value !== this._symbolDefinitionProps[name]) this._symbolDefinitionOverriddenProps[name] = value;
          } else {
            if (values[index] !== this._symbolDefinitionProps[name]) this._symbolDefinitionOverriddenProps[name] = values[index];
          }
        }
      });
    }
  });
};

control.CONTROL_PROTOTYPE.overrideChildrenOrigins = function () {
  var func = control => {
    if (!control.isPropertyOverridden('origin')) {
      control.setActualProperty('origin', control.getActualProperty('origin'));
    }

    if (control.children.length > 0) {
      control.children.forEach(child => func(child));
    }
  };

  this.children.forEach(child => func(child));
};

control.CONTROL_PROTOTYPE.hasChildrenOverrides = function () {
  return this.getDescendants().some(c => c.overriddenProperties && Object.keys(c.overriddenProperties).length > 0);
};

control.CONTROL_PROTOTYPE.isChildrenOriginOverridden = function () {
  return this.children.some(c => c.isPropertyOverridden("origin"));
};

control.CONTROL_PROTOTYPE.getSelectionBaseColor = function () {
  return 0x239233;
};

control.CONTROL_PROTOTYPE.isGroup = function () {
  return true;
};

control.CONTROL_PROTOTYPE.isSymbol = function () {
  return true;
};

control.CONTROL_PROTOTYPE.getResizeOption = function () {
  var res = Props.ResizeOptions.NO;

  if (this.state === _types.default.SymbolState.LOADED && !this.isEmpty()) {
    res = Props.ResizeOptions.BOTH;
  }

  return res;
};

control.CONTROL_PROTOTYPE.getName = function () {
  let name;
  const symbolDefinition = this.getSymbolDefinition();

  if (symbolDefinition) {
    name = symbolDefinition.getName(true);
  }

  return name && name !== '' ? name : 'Symbol';
};

var copyChildren = function (parentControl, control) {
  _balsamiq.default.existsClass(parentControl, _controls.Control);

  _balsamiq.default.existsClass(control, _controls.Control);

  var newControls = [];
  control.children.forEach(function (child) {
    var c = child.copy(parentControl._context, (0, _coreobject.makeChildControlCoreObjectID)(parentControl.coreObjectID, child.controlID));
    copyChildren(c, child);
    newControls.push(c);
  });
  parentControl.addChildren(newControls);
  parentControl.children.forEach(function (child) {
    child.updateAll();
  });
};

control.findSymbolParent = function (control) {
  var parent = control.parent;

  while (parent) {
    if (parent.isSymbol()) {
      return parent;
    }

    parent = parent.parent;
  }

  return null;
};

const findTopSymbolParent = function (control) {
  let symbolParent = null;
  let parent = control.parent;

  while (parent) {
    if (parent.isSymbol()) {
      symbolParent = parent;
    }

    parent = parent.parent;
  }

  return symbolParent;
};

control.findTopSymbolParent = findTopSymbolParent;

control.insideSymbolInstance = function (currentGroup) {
  return currentGroup.isSymbol() || control.findSymbolParent(currentGroup);
};

control.CONTROL_PROTOTYPE.updateCropData = function () {
  return _group.default.CONTROL_PROTOTYPE.updateCropData.call(this);
};

control.CONTROL_PROTOTYPE.updateCropOffsetMask = function () {
  return _group.default.CONTROL_PROTOTYPE.updateCropOffsetMask.call(this);
};

control.CONTROL_PROTOTYPE.isEmpty = function () {
  return _group.default.CONTROL_PROTOTYPE.isEmpty.call(this);
};

control.CONTROL_PROTOTYPE.getDescription = function () {
  let res;
  let symbolDefinition = this.getSymbolDefinition();

  if (symbolDefinition) {
    res = symbolDefinition.properties ? symbolDefinition.properties.controlName : null;
  }

  return res || this.superClass.prototype.getDescription.call(this);
};

function canConvertToSymbol(controls) {
  // default is true
  if (controls.length == 0 || !controls) {
    return true;
  } // default is true!


  const currentMockup = controls[0].getMockup();

  if (!currentMockup) {
    return true;
  } // always possible in wireframes


  if (currentMockup.getType() !== _BmprConstants.ResourceKinds.SymbolLibrary) {
    return true;
  } // controls must not be a symbol definition


  let isSymbolDefinition = false;

  for (let i = 0; i < controls.length; i += 1) {
    if (controls[i].coreObjectID.controlID.length === 1 && controls[i].isGroup()) {
      isSymbolDefinition = true;
      break;
    }
  }

  return !isSymbolDefinition;
}

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/tabbars.js":
/*!*************************************!*\
  !*** ./js/core/controls/tabbars.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.tabbarVertical = exports.tabbarHorizontal = exports["default"] = void 0;

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Rectangle = _interopRequireWildcard(__webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var ScrollBars = _interopRequireWildcard(__webpack_require__(/*! core/bricks/scrollbars */ "./js/core/bricks/scrollbars.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _snapping = __webpack_require__(/*! core/controls/snapping */ "./js/core/controls/snapping.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

let makeTabBarControlClass = function (p_controlTypeID, p_controlTypeName, p_controlAlternateNames, p_defaultSize, p_defaultText, p_splitOn, p_customProperties, p_textFunction, p_stylePrefix, p_tabsPositionProperty, p_tabsAlignProperty, p_multiline) {
  let control = Controls.makeControlClass({
    controlTypeID: p_controlTypeID,
    controlTypeName: p_controlTypeName,
    controlAlternateNames: p_controlAlternateNames,
    props: { ...Props.PropertyDefinitions.SIZE({
        defaultValue: _types.default.newAutoSize(),
        options: {
          autosize: Props.AutoSizeOptions.NO,
          resize: Props.ResizeOptions.BOTH
        }
      }),
      ...Props.PropertyDefinitions.BORDER_STYLE(Props.BorderStyles.SQUARE, Props.SimpleBorderStyles),
      ...Props.PropertyDefinitions.SCROLLBAR(),
      ...Props.PropertyDefinitions.links()
    }
  });
  Controls.addColorPropertyDefinitions(control, {
    defaultColor: p_stylePrefix + "_default_color",
    hasAlpha: true
  });
  Controls.addStandardTextPropertyDefinitions(control, {
    text: p_defaultText,
    multiline: p_multiline,
    splitOn: p_splitOn
  });
  Controls.addPropertyDefinitions(control, p_customProperties);
  Controls.addSelectionPropertyDefinitions(control, -1);

  control.CONTROL_PROTOTYPE.getSelectableLabels = function () {
    return this._selectableLabels;
  };

  control.CONTROL_PROTOTYPE.getLinkNames = function () {
    return this._selectableLabels;
  };

  control.CONTROL_PROTOTYPE.getLinkableLabels = function (text) {
    const lines = p_textFunction.call(this, text);
    const linkLabels = lines.map(line => TextUtils.unformatText(line));
    return linkLabels;
  };

  control.CONTROL_PROTOTYPE.initData = function () {
    this._backgroundBounds = null;
  };

  control.CONTROL_PROTOTYPE.updateData = function () {
    const text = this.getActualProperty("text");
    const labels = p_textFunction.call(this, text);
    const defaultTextStyle = this.newStandardTextStyle();
    const selectedTextStyle = this.newStandardTextStyle({
      textColor: _colors.default.labelColorGivenBackgroundColor(this.getActualProperty("color"), this.getActualProperty("backgroundAlpha"), this.style("default_text_color")),
      linkColor: _colors.default.labelColorGivenBackgroundColor(this.getActualProperty("color"), this.getActualProperty("backgroundAlpha"), this.style("default_link_color"))
    });
    const tabPaddingLeft = this.style(p_stylePrefix + "_tab_padding_left");
    const tabPaddingRight = this.style(p_stylePrefix + "_tab_padding_right");
    const tabPaddingTop = this.style(p_stylePrefix + "_tab_padding_top");
    const tabPaddingBottom = this.style(p_stylePrefix + "_tab_padding_bottom");
    const tabOverlap = this.style(p_stylePrefix + "_tab_overlap");
    const selectedIndex = this.getActualProperty("selectedIndex");
    const tabsPosition = this.getActualProperty(p_tabsPositionProperty);
    const isVertical = tabsPosition === _types.default.Positions.LEFT || tabsPosition === _types.default.Positions.RIGHT;
    const alpha = this.getActualProperty("backgroundAlpha");
    this._selectableLabels = this.getLinkableLabels(text);
    const isPreviousTextOfDifferentLength = !this.texts || this.texts.length !== labels.length;
    this.texts = labels.map((label, index) => {
      const textStyle = index === selectedIndex ? selectedTextStyle : defaultTextStyle;
      return isPreviousTextOfDifferentLength || !Text.areTextsEqual(this.texts[index], label, textStyle) ? Text.newTextNoWrap(label, textStyle, this._context) : this.texts[index];
    });
    this.maxTextSize = _types.default.newSize(0, 0);
    this.texts.forEach(text => {
      this.maxTextSize.width = Math.max(this.maxTextSize.width, text.measuredSize().width);
      this.maxTextSize.height = Math.max(this.maxTextSize.height, text.measuredSize().height);
    }); // create tabs

    this._tabs = this.texts.map((text, index) => {
      const selected = index === selectedIndex;
      return {
        assetName: p_stylePrefix + "_" + tabsPosition + "_tab",
        backgroundAssetName: p_stylePrefix + "_" + tabsPosition + "_background_tab",
        backgroundColor: selected ? this.getActualProperty("color") : this.style(p_stylePrefix + "_default_tab_color"),
        alpha: selected ? alpha : 1,
        selected: selected,
        text: text,
        textOffset: _types.default.newPoint(tabPaddingLeft, tabPaddingTop + (selected ? this.style(p_stylePrefix + "_" + tabsPosition + "_selected_label_offset") : 0)),
        size: _types.default.newSize(text.measuredSize().width + tabPaddingLeft + tabPaddingRight, (isVertical ? text.measuredSize().height : this.maxTextSize.height) + tabPaddingTop + tabPaddingBottom)
      };
    });
    this._selectedTab = selectedIndex !== -1 ? this._tabs[selectedIndex] : null; // calc control native size

    this._nativeSize = 0;

    if (isVertical) {
      this._nativeSize += this.style(p_stylePrefix + "_padding_top"); // the height of vertical tabs ignores the overlap...

      this._tabs.forEach(tab => {
        this._nativeSize += tab.size.height;
      });

      this._nativeSize += this.style(p_stylePrefix + "_padding_bottom");
    } else {
      this._nativeSize += this.style(p_stylePrefix + "_padding_left"); // the width of horizontal tabs ignores tab padding...

      this._tabs.forEach(tab => {
        this._nativeSize += tab.size.width - tabOverlap; //tab.text.measuredSize().width;
      });

      this._nativeSize += this.style(p_stylePrefix + "_padding_right") + tabOverlap;
    }
  };

  control.CONTROL_PROTOTYPE.measure = function (p_sizeProp) {
    const tabsPosition = this.getActualProperty(p_tabsPositionProperty);
    const isVertical = tabsPosition === _types.default.Positions.LEFT || tabsPosition === _types.default.Positions.RIGHT;
    let width = 0,
        height = 0;

    if (isVertical) {
      width = 200;
      height = Math.max(this._nativeSize, 100);
    } else {
      width = Math.max(this._nativeSize, 150);
      height = 100;
    }

    return _types.default.newSize(width, height);
  };

  control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
    const tabPaddingLeft = this.style(p_stylePrefix + "_tab_padding_left");
    const tabPaddingRight = this.style(p_stylePrefix + "_tab_padding_right");
    const tabPaddingTop = this.style(p_stylePrefix + "_tab_padding_top");
    const tabPaddingBottom = this.style(p_stylePrefix + "_tab_padding_bottom");
    const tabOverlap = this.style(p_stylePrefix + "_tab_overlap");
    const selectedIndex = this.getActualProperty("selectedIndex");
    const tabsPosition = this.getActualProperty(p_tabsPositionProperty);
    const isVertical = tabsPosition === _types.default.Positions.LEFT || tabsPosition === _types.default.Positions.RIGHT;
    const alpha = this.getActualProperty("backgroundAlpha");
    const borderColor = this.style("default_border_color"); // calc tabs size

    let tabsTotalSize = _types.default.newSize(0, 0);

    if (isVertical) {
      tabsTotalSize.width = this.maxTextSize.width + tabPaddingLeft + tabPaddingRight;

      this._tabs.forEach(function (tab) {
        tabsTotalSize.height += tab.size.height;
      });

      tabsTotalSize.height -= tabOverlap * (this._tabs.length - 1);
    } else {
      this._tabs.forEach(function (tab) {
        tabsTotalSize.width += tab.size.width;
      });

      tabsTotalSize.width -= tabOverlap * (this._tabs.length - 1);
      tabsTotalSize.height = this.maxTextSize.height + tabPaddingTop + tabPaddingBottom;
    } // calc tabs origin


    let tabsOrigin = _types.default.newPoint(0, 0);

    if (isVertical) {
      tabsOrigin.x = tabsPosition === _types.default.Positions.LEFT ? 0 : p_localBbox.width - tabsTotalSize.width;
      tabsOrigin.y = this.style(p_stylePrefix + "_padding_top");
    } else {
      const align = this.getActualProperty(p_tabsAlignProperty);
      const paddingLeft = this.style(p_stylePrefix + "_padding_left");
      const paddingRight = this.style(p_stylePrefix + "_padding_right");
      tabsOrigin.x = paddingLeft;

      if (align === _types.default.TextAlign.CENTER) {
        tabsOrigin.x = paddingLeft + (p_localBbox.width - paddingLeft - paddingRight - tabsTotalSize.width) / 2;
      } else if (align === _types.default.TextAlign.RIGHT) {
        tabsOrigin.x = p_localBbox.width - paddingLeft - paddingRight - tabsTotalSize.width;
      }

      tabsOrigin.y = tabsPosition === _types.default.Positions.TOP ? 0 : p_localBbox.height - tabsTotalSize.height;
    } // set origin (and width, if needed) of each tab


    let o = _balsamiq.default.clone(tabsOrigin);

    this._tabs.forEach(tab => {
      tab.origin = o;
      o = _balsamiq.default.clone(o);

      if (isVertical) {
        o.y += tab.size.height - tabOverlap;
        tab.size.width = tabsTotalSize.width;
      } else {
        o.x += tab.size.width - tabOverlap;
      }
    }); // function to create the display list of a tab


    let makeTabDisplayList = function (p_tab) {
      const bounds = _types.default.newRectangle(p_tab.origin.x, p_tab.origin.y, p_tab.size.width, p_tab.size.height); // check if the tab is inside the bbox


      if (isVertical && bounds.y + bounds.height > p_localBbox.y + p_localBbox.height || !isVertical && bounds.x + bounds.width > p_localBbox.x + p_localBbox.width) {
        return null;
      }

      let background = _displayitem.default.createDisplayItemsFromSkin(p_tab.backgroundAssetName, p_tab.backgroundColor, bounds, this._context);

      background.forEach(function (b) {
        b.alpha = p_tab.alpha;
      });

      const border = _displayitem.default.createDisplayItemsFromSkin(p_tab.assetName, borderColor, bounds, this._context);

      const text = p_tab.text.makeDisplayItems(_types.default.newRectangle(p_tab.origin.x + p_tab.textOffset.x, p_tab.origin.y + p_tab.textOffset.y, p_tab.text.measuredSize().width, p_tab.text.measuredSize().height));

      const linkBounds = _types.default.paddedRectangle(bounds, 0, 0, 0, tabOverlap);

      return {
        displayItems: [background, border, text],
        linkBounds: linkBounds
      };
    }; // link bounds


    let linkBounds = []; // create the display list of the selected tab, if any, and create the mask

    let res;
    let selectedTabDisplayItems = [];
    let selectedTabMaskID = null,
        otherTabsMaskID = null;

    if (this._selectedTab) {
      // display list and link
      res = makeTabDisplayList.call(this, this._selectedTab);

      if (res) {
        const linkBound = _types.default.newRectangle(res.linkBounds.x, res.linkBounds.y, Math.min(res.linkBounds.width, p_localBbox.width), Math.min(res.linkBounds.height, p_localBbox.height));

        linkBounds[selectedIndex] = linkBound;
        selectedTabDisplayItems = res.displayItems;
      } // mask to clip the border of the group, only needed if we draw the selected tab


      if (this._selectedTab.origin.x + this._selectedTab.size.width <= p_localBbox.width && this._selectedTab.origin.x >= 0 && this._selectedTab.origin.y >= 0) {
        const mtop = this.style(p_stylePrefix + "_" + tabsPosition + "_selected_tab_mask_padding_top");
        const mbottom = this.style(p_stylePrefix + "_" + tabsPosition + "_selected_tab_mask_padding_bottom");
        const mleft = this.style(p_stylePrefix + "_" + tabsPosition + "_selected_tab_mask_padding_left");
        const mright = this.style(p_stylePrefix + "_" + tabsPosition + "_selected_tab_mask_padding_right");

        let selectedTabMaskBounds = _types.default.newRectangle(this._selectedTab.origin.x, this._selectedTab.origin.y, this._selectedTab.size.width, this._selectedTab.size.height);

        selectedTabMaskBounds = _types.default.paddedRectangle(selectedTabMaskBounds, mtop, mbottom, mleft, mright);
        selectedTabMaskID = this.addMaskWithRectHole(p_localBbox, selectedTabMaskBounds); // only for Sketch skin, horizontal tabs
        // Writing this stuff in the skin would be a nightmare. It's place is here, a special case in the code.
        //cancels the pieces of tabs behind the selected one. this is visible when the opacity is < 100%

        if (this._context.skin["#skinName"] === "sketch" && !isVertical) {
          otherTabsMaskID = this.addMask(); // tabs on the left

          this.addRectToMask(otherTabsMaskID, _types.default.newRectangle(0, selectedTabMaskBounds.y, selectedTabMaskBounds.x, selectedTabMaskBounds.height)); // tabs on the right

          const x = selectedTabMaskBounds.x + selectedTabMaskBounds.width,
                y = selectedTabMaskBounds.y;

          if (tabsPosition === _types.default.Positions.TOP) {
            let gm = Path.newPath(p_localBbox.width, p_localBbox.height);
            let SLANT = 7;
            gm.moveTo(x - SLANT, y);
            gm.lineTo(p_localBbox.width, selectedTabMaskBounds.y);
            gm.lineTo(p_localBbox.width, selectedTabMaskBounds.y + selectedTabMaskBounds.height);
            gm.lineTo(x, selectedTabMaskBounds.y + selectedTabMaskBounds.height);
            gm.lineTo(x - SLANT, selectedTabMaskBounds.y + 0.38 * selectedTabMaskBounds.height);
            gm.close();
            this.addPathToMask(otherTabsMaskID, gm, 0, 0);
          } else {
            if (p_localBbox.width - x + 1 > 0) {
              this.addRectToMask(otherTabsMaskID, _types.default.newRectangle(x - 1, y, p_localBbox.width - x + 1, selectedTabMaskBounds.height));
            }
          }
        }
      }
    } // then create the display list of not selected tabs


    let tabsDisplayItems = [];

    this._tabs.forEach((tab, tabIndex) => {
      if (!tab.selected) {
        res = makeTabDisplayList.call(this, tab);

        if (res) {
          let linkBound = _types.default.newRectangle(res.linkBounds.x, res.linkBounds.y, Math.min(res.linkBounds.width, p_localBbox.width), Math.min(res.linkBounds.height, p_localBbox.height));

          linkBounds[tabIndex] = linkBound;
          tabsDisplayItems = tabsDisplayItems.concat(res.displayItems);
        }
      }
    });

    if (this._selectedTab) {
      _displayitem.default.maskDisplayItems(tabsDisplayItems, otherTabsMaskID);
    } // link areas


    _link.default.processLinks(this, linkBounds); // border and background


    const backgroundOverlap = this.style(p_stylePrefix + "_background_overlap");
    let backgroundBounds;

    if (tabsPosition === _types.default.Positions.TOP) {
      backgroundBounds = _types.default.newRectangle(0, tabsTotalSize.height - backgroundOverlap, p_localBbox.width, Math.max(p_localBbox.height - tabsTotalSize.height + backgroundOverlap, 0));
    } else if (tabsPosition === _types.default.Positions.BOTTOM) {
      backgroundBounds = _types.default.newRectangle(0, 0, p_localBbox.width, Math.max(p_localBbox.height - tabsTotalSize.height + backgroundOverlap, 0));
    } else if (tabsPosition === _types.default.Positions.LEFT) {
      backgroundBounds = _types.default.newRectangle(tabsTotalSize.width - backgroundOverlap, 0, Math.max(p_localBbox.width - tabsTotalSize.width + backgroundOverlap, 0), p_localBbox.height);
    } else {
      backgroundBounds = _types.default.newRectangle(0, 0, Math.max(p_localBbox.width - tabsTotalSize.width + backgroundOverlap, 0), p_localBbox.height);
    }

    let border, backgroundFill;

    if (this.getActualProperty("borderStyle") === Props.BorderStyles.SQUARE) {
      border = _displayitem.default.createDisplayItemsFromSkin("rectangle_border_square", borderColor, backgroundBounds, this._context);
      const bp = this.style(p_stylePrefix + "_background_padding");

      const paddedBounds = _types.default.paddedRectangle(backgroundBounds, bp, bp, bp, bp);

      if (paddedBounds.width > 0 && paddedBounds.height > 0) {
        backgroundFill = Rectangle.makeFilledRectangle(paddedBounds, this.getActualProperty("color"), alpha);
      }
    } else {
      const ruleAsset = isVertical ? this.style("tabbar_vertical_separator_skin") : "rule_horizontal";
      const ruleAssetSize = this.styleSVGSize(ruleAsset);
      let ruleBounds;

      if (tabsPosition === _types.default.Positions.TOP) {
        ruleBounds = _types.default.newRectangle(backgroundBounds.x, Math.round(backgroundBounds.y - ruleAssetSize.height / 2 + backgroundOverlap), backgroundBounds.width, ruleAssetSize.height);
      } else if (tabsPosition === _types.default.Positions.BOTTOM) {
        ruleBounds = _types.default.newRectangle(backgroundBounds.x, Math.round(backgroundBounds.y + backgroundBounds.height - ruleAssetSize.height / 2 - backgroundOverlap), backgroundBounds.width, ruleAssetSize.height);
      } else if (tabsPosition === _types.default.Positions.LEFT) {
        ruleBounds = _types.default.newRectangle(Math.round(backgroundBounds.x - ruleAssetSize.width / 2 + backgroundOverlap), backgroundBounds.y, ruleAssetSize.width, backgroundBounds.height);
      } else {
        ruleBounds = _types.default.newRectangle(Math.round(backgroundBounds.x + backgroundBounds.width - ruleAssetSize.width / 2), backgroundBounds.y, ruleAssetSize.width, backgroundBounds.height);
      }

      border = _displayitem.default.createDisplayItemsFromSkin(ruleAsset, borderColor, ruleBounds, this._context);
    }

    if (this._selectedTab) {
      _displayitem.default.maskDisplayItems([border], selectedTabMaskID);
    } // scrollbar


    let scrollbar;

    if (this.getActualProperty("verticalScrollbar")) {
      const scrollbarPaddingTop = this.style(p_stylePrefix + "_scrollbar_padding_top");
      const scrollbarPaddingBottom = this.style(p_stylePrefix + "_scrollbar_padding_bottom");
      const scrollbarPaddingRight = this.style(p_stylePrefix + "_scrollbar_padding_right");
      const scrollbarX = backgroundBounds.x + backgroundBounds.width - scrollbarPaddingRight - ScrollBars.Sizes.ScrollbarFixedSize;
      const scrollbarY = backgroundBounds.y + scrollbarPaddingTop;
      const scrollbarHeight = backgroundBounds.height - scrollbarPaddingTop - scrollbarPaddingBottom;

      if (scrollbarHeight > 0) {
        scrollbar = ScrollBars.makeDisplayItemsForVerticalScrollbar(scrollbarX, scrollbarY, scrollbarHeight, this.getActualProperty("value"), _types.default.ControlState.NORMAL, this._context);
      }
    }

    this._backgroundBounds = backgroundBounds;
    this.pushToDisplayList(tabsDisplayItems, backgroundFill, scrollbar, border, selectedTabDisplayItems);
  };

  control.CONTROL_PROTOTYPE.getSnapLines = function (withMarkup) {
    _balsamiq.default.assertBool(withMarkup);

    const bbox = this.getUserBBox(true);
    const bbox2 = this._backgroundBounds;
    return [new _snapping.SnapLine(_snapping.SnapTypes.MIDDLE, _snapping.SnapDirections.VERTICAL, bbox.x + bbox.width / 2), new _snapping.SnapLine(_snapping.SnapTypes.MIDDLE, _snapping.SnapDirections.HORIZONTAL, bbox.y + bbox.height / 2), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.VERTICAL, bbox.x), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.VERTICAL, bbox.x + bbox.width), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.HORIZONTAL, bbox.y), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.HORIZONTAL, bbox.y + bbox.height), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.VERTICAL, bbox2.x + _snapping.PROXIMITY_DISTANCE), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.VERTICAL, bbox2.x + bbox2.width - _snapping.PROXIMITY_DISTANCE), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.HORIZONTAL, bbox2.y + _snapping.PROXIMITY_DISTANCE), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.HORIZONTAL, bbox2.y + bbox2.height - _snapping.PROXIMITY_DISTANCE)];
  };

  return control;
};

const tabbarHorizontal = makeTabBarControlClass("com.balsamiq.mockups::TabBar", "Tab Bar", ["Ribbon"], _types.default.newSize(205, 100), "One, Two, Three, Four", Props.SplitOnTypes.COMMA, {
  tabHPosition: {
    defaultValue: _types.default.TextAlign.LEFT
  },
  tabVPosition: {
    defaultValue: _types.default.Positions.TOP
  }
}, function (text) {
  return TextUtils.splitTextOnCommas(text);
}, "tabbar_h", "tabVPosition", "tabHPosition", false);
exports.tabbarHorizontal = tabbarHorizontal;
const tabbarVertical = makeTabBarControlClass("com.balsamiq.mockups::VerticalTabBar", "V.Tabs", ["Vertical Tabs"], _types.default.newSize(200, 182), "First Tab\nSecond Tab\nThird Tab\nFourth Tab", Props.SplitOnTypes.NEWLINE, {
  position: {
    defaultValue: _types.default.Positions.LEFT
  }
}, function (text) {
  return text.split("\n");
}, "tabbar_v", "position", null, true);
exports.tabbarVertical = tabbarVertical;
var _default = {
  tabbarHorizontal,
  tabbarVertical
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/tagcloud.js":
/*!**************************************!*\
  !*** ./js/core/controls/tagcloud.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var Controls = _interopRequireWildcard(__webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const control = Controls.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::TagCloud",
  controlTypeName: "Tag Cloud",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(350, 110),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.TEXT({
      // TODO add a callback to modify the text once inserted, see MTagCloud.as
      defaultText: "software statistics teaching {color:#FF0000}technology{color} tips tool tools toread travel tutorial tutorials tv twitter typography ubuntu usability video videos visualization web web\\ 2.0 web\\ design webdev wiki windows wordpress work writing youtube",
      multiline: true
    })
  }
});

const getNotVeryRandomNumber = function (seed, offset) {
  const sequence = "fdlkgj'[=-erptjfl;kjf[3p9imd=p;o34uf5tmxms.,p3woiugnyiovuwgcw568362';][phmpq";
  return sequence.charCodeAt((seed + offset) % sequence.length) % 16 + 6;
};

control.CONTROL_PROTOTYPE.updateData = function () {
  // This text is used only to extract link targets
  const text = this.getActualProperty("text");
  const textStyle = this.newTextStyle();

  if (!this.text || !Text.areTextsEqual(this.text, text, textStyle)) {
    this.text = Text.newText(text, textStyle, this._context);
  }
};

_link.default.addTextLinksMethods(control);

control.CONTROL_PROTOTYPE.updateLayout = function (localBbox) {
  const text = this.getActualProperty("text");
  const textSavingSpaces = text.replace(/\\ /g, "__m_peldi_m__");
  const labels = textSavingSpaces.split(" ");

  if (labels.length === 1) {
    labels.push("...");
  }

  const LEFT = 9;
  const TOP = 1;
  const STEP_X = 5;
  const STEP_Y = 24;
  let x = LEFT;
  let y = TOP + STEP_Y;
  const textStyle = this.newTextStyle();
  const isPreviousTextOfDifferentLength = !this.labels || this.labels.length !== labels.length;

  if (isPreviousTextOfDifferentLength) {
    this.fontSizes = new Array(labels.length).fill().map((_, index) => getNotVeryRandomNumber(this.controlID, index));
  }

  this.labels = labels.map((label, index) => {
    label = label.replace(/__m_peldi_m__/g, " ");
    textStyle.fontSize = this.fontSizes[index];
    const text = isPreviousTextOfDifferentLength || !Text.areTextsEqual(this.labels[index], label, textStyle) ? Text.newSingleLineText(label, textStyle, this._context) : this.labels[index];
    const textSize = text.measuredSize();

    if (x + textSize.width + STEP_X >= localBbox.width) {
      x = LEFT;
      y += STEP_Y;
    }

    if (y < localBbox.height) {
      // Do not add lines if partially visible
      const rectangle = _types.default.newRectangle(x, y - textSize.height, textSize.width, textSize.height);

      const {
        displayItems,
        links
      } = text.makeDisplayItemsAndLinks(rectangle);
      this.pushToDisplayList(displayItems);

      _link.default.processTextLinks(this, links);
    }

    x += textSize.width + STEP_X;
    return text;
  }, this);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/textarea.js":
/*!**************************************!*\
  !*** ./js/core/controls/textarea.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _scrollbars = __webpack_require__(/*! core/bricks/scrollbars */ "./js/core/bricks/scrollbars.js");

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _macro = __webpack_require__(/*! core/bricks/macro */ "./js/core/bricks/macro.js");

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::TextArea",
  controlTypeName: "Text Area",
  controlAlternateNames: ["Memo Field"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(200, 140),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    align: {
      defaultValue: Props.TextAlignments.LEFT,
      enum: Props.TextAlignments
    },
    ...Props.PropertyDefinitions.SCROLLBAR(),
    state: {
      defaultValue: Props.States.NORMAL,
      enum: Props.SimpleStates
    },
    ...Props.PropertyDefinitions.link
  }
});
controlsMod.addStandardTextPropertyDefinitions(control, {
  multiline: true,
  loremable: true,
  charCountEnabled: true,
  splitOn: Props.SplitOnTypes.NEWLINE
});
controlsMod.addColorPropertyDefinitions(control, {
  defaultColor: 0xFFFFFF,
  defaultBorderColor: "default_border_color",
  defaultTextColor: "default_text_color",
  hasAlpha: true
});

control.CONTROL_PROTOTYPE.initData = function () {
  this.text = null;
};

_link.default.addTextLinksMethods(control);

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  _link.default.processControlLink(this); // BBox for text depends on right scrollbar


  var textBbox = _types.default.paddedRectangle(p_localBbox, this.style("textarea_padding_top"), this.style("textarea_padding_bottom"), this.style("textarea_padding_left"), this.style("textarea_padding_right"));

  var verticalScrollbarDisplayItems = [];

  if (this.getActualProperty('verticalScrollbar')) {
    textBbox.width -= _scrollbars.Sizes.ScrollbarFixedSize;
    verticalScrollbarDisplayItems = (0, _scrollbars.makeVerticalScrollbar)(p_localBbox, this.style("textarea_scrollbar_padding_top"), this.style("textarea_scrollbar_padding_bottom"), this.style("textarea_scrollbar_padding_right"), this.getActualProperty('value'), this.getActualProperty('state'), this._context);
  }

  var state = this.getActualProperty('state'); // Border

  var borderColor;

  if (state === _types.default.ControlState.NORMAL) {
    borderColor = this.getActualProperty('borderColor');
  } else if (state === _types.default.ControlState.DISABLED) {
    borderColor = this.style("default_disabled_color");
  } else if (state === _types.default.ControlState.FOCUSED) {
    borderColor = this.style("default_selection_color");
  }

  var border = (0, _rectangle.makeRectangleBorderDisplayItems)(this._context, p_localBbox, borderColor); // Background

  var background = (0, _rectangle.makeFilledRectangle)(p_localBbox, this.getActualProperty("color"), this.getActualProperty("backgroundAlpha")); // Text color based on background

  var textColor, linkColor;

  if (state === _types.default.ControlState.NORMAL || state === _types.default.ControlState.FOCUSED) {
    if (this.hasPropertyValue("textColor")) {
      textColor = this.getActualProperty("textColor");
    } else {
      textColor = _colors.default.labelColorGivenBackgroundColor(this.getActualProperty('color'), this.getActualProperty("backgroundAlpha"), this.style("default_text_color"));
    }

    linkColor = _colors.default.labelColorGivenBackgroundColor(this.getActualProperty('color'), this.getActualProperty("backgroundAlpha"), this.style("default_link_color"));
  } else if (state === _types.default.ControlState.DISABLED) {
    textColor = this.style("default_disabled_color");
    linkColor = this.style("default_disabled_color");
  }

  var textStyle = this.newStandardTextStyle({
    textColor: textColor,
    linkColor: linkColor
  });
  var resolvedText = (0, _macro.processMacros)(this);
  this.text = textMod.newText(resolvedText, textStyle, this._context, true);
  var obj = this.text.makeDisplayItemsAndLinks(textBbox, this.getActualProperty('align'));

  _link.default.processTextLinks(this, obj.links);

  this.pushToDisplayList(background, obj.displayItems, border, verticalScrollbarDisplayItems);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/textinput.js":
/*!***************************************!*\
  !*** ./js/core/controls/textinput.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::TextInput",
  controlTypeName: "Text Input",
  controlAlternateNames: ["Text Field"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.ONLY_WIDTH,
        resize: Props.ResizeOptions.HORIZONTAL
      }
    }),
    state: {
      defaultValue: Props.States.NORMAL,
      enum: Props.SimpleStates
    },
    align: {
      defaultValue: Props.TextAlignments.LEFT,
      enum: Props.TextAlignments
    },
    ...Props.PropertyDefinitions.BORDER_STYLE(Props.TextInputBorderStyles.RECTANGLE, Props.TextInputBorderStyles)
  }
});
controlsMod.addStandardTextPropertyDefinitions(control, {
  charCountEnabled: true
});
controlsMod.addColorPropertyDefinitions(control, {
  defaultColor: 0xFFFFFF,
  defaultBorderColor: "default_border_color",
  defaultTextColor: "default_text_color",
  hasAlpha: true
});

control.CONTROL_PROTOTYPE.updateData = function () {
  const isDisabled = this.getActualProperty('state') === _types.default.ControlState.DISABLED;

  const textColor = isDisabled ? this.style('default_disabled_color') : this.hasPropertyValue("textColor") ? this.getActualProperty("textColor") : _colors.default.labelColorGivenBackgroundColor(this.getActualProperty("color"), this.getActualProperty("backgroundAlpha"), this.style("default_text_color"));
  const linkColor = isDisabled ? this.style('default_disabled_color') : _colors.default.labelColorGivenBackgroundColor(this.getActualProperty("color"), this.getActualProperty("backgroundAlpha"), this.style("default_link_color"));
  const textStyle = this.newStandardTextStyle({
    textColor,
    linkColor
  });

  if (!this.text || !(0, _text.areTextsEqual)(this.text, this.getActualProperty("text"), textStyle)) {
    this.text = (0, _text.newSingleLineText)(this.getActualProperty("text"), textStyle, this._context);
  }
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var textProp = this.getActualProperty("text");
  var textSize = this.text.measuredSize();
  var width = p_sizeProperty.width;

  if (width === _types.default.AUTOSIZE) {
    if (textProp) {
      width = Math.round(textSize.width + this.style("textinput_label_padding_left") + this.style("textinput_label_padding_right"));
    } else {
      width = 79;
    }
  }

  var textHeight = textProp ? Math.round(textSize.height) : this.getActualProperty("fontSize");
  var height = textHeight + this.style("textinput_label_padding_top") + this.style("textinput_label_padding_bottom") + this.style("textinput_empty_space_bottom");
  return _types.default.newSize(width, height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var realBbox = _types.default.paddedRectangle(p_localBbox, 0, this.style("textinput_empty_space_bottom"));

  var color = this.getActualProperty("color");
  this.pushToDisplayList((0, _rectangle.makeFilledRectangle)(realBbox, color, this.getActualProperty('backgroundAlpha'), 0));
  var borderColor = this.getActualProperty("borderColor");
  var state = this.getActualProperty("state");

  if (state === _types.default.ControlState.DISABLED) {
    borderColor = this.style("default_disabled_color");
  } else if (state === _types.default.ControlState.FOCUSED) {
    borderColor = this.style("default_selection_color");
  }

  realBbox.y += this.getActualProperty("borderStyle") == "line" ? this.style("textinput_border_padding") : 0;
  this.pushToDisplayList(_displayitem.default.createDisplayItemsFromSkin(this.getActualProperty("borderStyle") == "line" ? "slider_horizontal_track" : "rectangle_border_square", borderColor, realBbox, this._context));
  var align = this.getActualProperty("align");
  var label = this.text.makeDisplayItems(_types.default.paddedRectangle(p_localBbox, this.style("textinput_label_padding_top"), 0, this.getActualProperty("borderStyle") == "line" ? 0 : this.style("textinput_label_padding_left"), this.style("textinput_label_padding_right")), align);
  this.pushToDisplayList(label);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/timepicker.js":
/*!****************************************!*\
  !*** ./js/core/controls/timepicker.js ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var Text = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::TimePicker",
  controlTypeName: "Time Picker",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(320, 473),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.links()
  }
});
const Directions = {
  PORTRAIT: "Portrait",
  LANDSCAPE: "Landscape"
};

control.CONTROL_PROTOTYPE.getLinkNames = function () {
  return ["CANCEL", "OK"];
};

control.CONTROL_PROTOTYPE.init = function () {};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  return _types.default.newSize(320, 473);
};

control.CONTROL_PROTOTYPE.digitalClock = function (p_barBbox, p_zoomRatio, p_direction) {
  let activeColor = this.style("timepicker_activeTextColor");
  let disabledColor = this.style("timepicker_inactiveTextColor");
  let displayItems = [];
  let date = new Date();
  let isPortrait = p_direction === Directions.PORTRAIT;
  let isPM = date.getHours() > 11 ? true : false;
  let timeTextStyle = this.newTextStyle();
  let AMTextStyle = this.newTextStyle();
  let PMTextStyle = this.newTextStyle();
  timeTextStyle.fontSize = (isPortrait ? 60 : 54) * p_zoomRatio;
  AMTextStyle.fontSize = PMTextStyle.fontSize = (isPortrait ? 14 : 18) * p_zoomRatio;
  AMTextStyle.textColor = isPM ? disabledColor : activeColor;
  PMTextStyle.textColor = isPM ? activeColor : disabledColor;
  let timeText = (date.getHours() > 12 ? date.getHours() - 12 : date.getHours()) + ":" + ("0" + date.getMinutes()).slice(-2);
  let timeLabel = Text.newText(timeText, timeTextStyle, this._context);
  let AMLabel = Text.newText("AM", AMTextStyle, this._context);
  let PMLabel = Text.newText("PM", PMTextStyle, this._context);
  let timeBbox, usedSpace, AMBbox;
  let theX, theY;
  let timeTextSize = timeLabel.measuredSize();

  if (p_direction === Directions.PORTRAIT) {
    usedSpace = timeTextSize.width + AMLabel.measuredSize().width;
    theX = (p_barBbox.width - usedSpace) / 2;
    theY = (p_barBbox.height - timeTextSize.height) / 2;
    timeBbox = _types.default.newRectangle(theX, theY, timeTextSize.width, timeTextSize.height);
    theX += timeTextSize.width + 20;
    theY += timeTextSize.height / 2 - AMLabel.measuredSize().height;
    AMBbox = _types.default.newRectangle(theX, theY, AMLabel.measuredSize().width, AMLabel.measuredSize().height);
  } else {
    usedSpace = timeTextSize.height + AMLabel.measuredSize().height;
    theX = (p_barBbox.width - timeTextSize.width) / 2;
    theY = (p_barBbox.height - usedSpace) / 2;
    timeBbox = _types.default.newRectangle(theX, theY, timeTextSize.width, timeTextSize.height);
    theX += (timeTextSize.width - AMLabel.measuredSize().width) / 2;
    theY += timeTextSize.height;
    AMBbox = _types.default.newRectangle(theX, theY, AMLabel.measuredSize().width, AMLabel.measuredSize().height);
  }

  displayItems.push(timeLabel.makeDisplayItems(timeBbox));
  displayItems.push(AMLabel.makeDisplayItems(AMBbox));
  displayItems.push(PMLabel.makeDisplayItems(_types.default.newRectangle(theX, theY + AMLabel.measuredSize().height, PMLabel.measuredSize().width, PMLabel.measuredSize().height)));
  return displayItems;
};

control.CONTROL_PROTOTYPE.analogClock = function (p_bbox, p_direction, p_zoom, p_heading) {
  let clockSize = this.style("timepicker_clockSize" + p_direction) * p_zoom;
  let clockBbox;

  if (p_direction === Directions.PORTRAIT) {
    clockBbox = _types.default.newRectangle((p_bbox.width - clockSize) / 2, (p_bbox.height + p_heading - clockSize) / 2, clockSize, clockSize);
  } else {
    clockBbox = _types.default.newRectangle((p_bbox.width + p_heading - clockSize) / 2, (p_bbox.height - clockSize) / 2, clockSize, clockSize);
  }

  if (clockSize > 0) {
    let circle = Path.newEllipsePath(clockSize, clockSize, 0);
    this.pushToDisplayList(_displayitem.default.newSVGPath(clockBbox, circle, 0, 0, this.style("timepicker_clockFaceColor")));
    let centerX = clockBbox.x + clockSize / 2;
    let centerY = clockBbox.y + clockSize / 2;
    let textStyle = this.newTextStyle();
    let selectedTextStyle = this.newTextStyle();
    textStyle.fontSize = selectedTextStyle.fontSize = 14 * p_zoom;
    selectedTextStyle.textColor = _colors.default.labelColorGivenBackgroundColor(this.style("default_selection_color"));
    let date = new Date();
    let hour = date.getHours() % 12;
    let textBbox, angle, label;

    for (let i = 1; i < 13; i++) {
      angle = i * 30 * (Math.PI / 180);
      label = Text.newText(i.toString(), i == hour ? selectedTextStyle : textStyle, this._context);
      textBbox = _types.default.newRectangle(centerX + 0.45 * clockSize * Math.sin(angle) - label.measuredSize().width / 2, centerY - 0.45 * clockSize * Math.cos(angle) - label.measuredSize().height / 2, label.measuredSize().width, label.measuredSize().height);

      if (i == hour || i == 12 && hour == 0) {
        // at 12am/pm i is 12 (used for label and angle computation) and hour is 0
        let handPath = Path.newPath(p_bbox.width, p_bbox.height);
        handPath.moveTo(centerX, centerY);
        handPath.lineTo(textBbox.x + textBbox.width / 2, textBbox.y + textBbox.height / 2);
        this.pushToDisplayList(_displayitem.default.newSVGPath(p_bbox, handPath, 3.0, this.style("default_selection_color")));
        handPath = Path.newEllipsePath(clockSize * 0.0425, clockSize * 0.0425, 0);

        let handBbox = _types.default.newRectangle(textBbox.x + textBbox.width / 2 - clockSize * 0.0425, textBbox.y + textBbox.height / 2 - clockSize * 0.0425, clockSize * 0.085, clockSize * 0.085);

        if (handBbox.width > 0 && handBbox.height > 0) {
          this.pushToDisplayList(_displayitem.default.newSVGPath(handBbox, handPath, 0, 0, this.style("default_selection_color")));
        }
      }

      this.pushToDisplayList(label.makeDisplayItems(textBbox));
    }
  }
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  let linkBounds = [];
  let padding = this.style("timepicker_padding");
  let isPortrait = p_localBbox.width < p_localBbox.height; //background

  let background = (0, _rectangle.makeRectangleDisplayItems)(this._context, p_localBbox, this.style("timepicker_lineWeight"), this.style("timepicker_backgroundColor"));
  this.pushToDisplayList(background); //ok and cancel buttons

  let textStyle = this.newTextStyle();
  textStyle.fontSize = 14;
  textStyle.textColor = this.style("default_link_color");
  textStyle.underline = true;
  let ok = Text.newText("OK", textStyle, this._context);
  let cancelText = Text.newText("CANCEL", textStyle, this._context);
  let theX = p_localBbox.width - ok.measuredSize().width - 2 * padding;
  var buttonHeight = Math.max(ok.measuredSize().height, cancelText.measuredSize().height);
  let bbox;

  if (theX - 2 * padding - cancelText.measuredSize().width > 0) {
    bbox = _types.default.newRectangle(theX - 2 * padding - cancelText.measuredSize().width, p_localBbox.height - (buttonHeight + padding), cancelText.measuredSize().width, buttonHeight);
    this.pushToDisplayList(cancelText.makeDisplayItems(bbox));
    linkBounds.push(_balsamiq.default.clone(bbox));
  }

  if (theX > 0) {
    bbox = _types.default.newRectangle(Math.max(1, theX), p_localBbox.height - (buttonHeight + padding), ok.measuredSize().width, buttonHeight);
    this.pushToDisplayList(ok.makeDisplayItems(bbox));
    linkBounds.push(_balsamiq.default.clone(bbox));
  }

  let zoomRatioHeader;
  let defaultHeaderHeight;
  let lineWeight = this.style("timepicker_lineWeight");

  if (isPortrait) {
    //portrait
    let defaultHeightPortrait = this.style("timepicker_defaultHeightPortrait");
    let defaultWidthPortrait = this.style("timepicker_defaultWidthPortrait");
    defaultHeaderHeight = this.style("timepicker_defaultHeadingPortrait");
    zoomRatioHeader = Math.min((p_localBbox.height - defaultHeaderHeight - (buttonHeight + padding)) / (defaultHeightPortrait - defaultHeaderHeight - (buttonHeight + padding)), p_localBbox.width / defaultWidthPortrait);
    let headerHeight = defaultHeaderHeight * zoomRatioHeader; //topbarbackground

    if (p_localBbox.width - 2 * lineWeight >= 0 && headerHeight - 2 * lineWeight > 0) {
      let topbarBbox = _types.default.newRectangle(lineWeight, lineWeight, p_localBbox.width - 2 * lineWeight, headerHeight - 2 * lineWeight);

      this.pushToDisplayList((0, _rectangle.makeFilledRectangle)(topbarBbox, this.style("timepicker_headingPanelBgColor"))); //Labels

      this.pushToDisplayList(this.digitalClock(topbarBbox, zoomRatioHeader, Directions.PORTRAIT));
    } //Clock


    let zoomRatioClock = Math.min((p_localBbox.height - headerHeight - (buttonHeight + padding)) / (defaultHeightPortrait - defaultHeaderHeight - (buttonHeight + padding)), p_localBbox.width / defaultWidthPortrait);
    this.analogClock(p_localBbox, Directions.PORTRAIT, zoomRatioClock, headerHeight);
  } else {
    //landscape
    let defaultHeightLandscape = this.style("timepicker_defaultHeightLandscape");
    let defaultWidthlandscape = this.style("timepicker_defaultWidthLandscape");
    let defaultHeaderWidth = this.style("timepicker_defaultHeadingLandscape");
    zoomRatioHeader = Math.min((p_localBbox.height - (buttonHeight + padding)) / (defaultHeightLandscape - (buttonHeight + padding)), (p_localBbox.width - defaultHeaderWidth) / (defaultWidthlandscape - defaultHeaderWidth));
    let headerWidth = defaultHeaderWidth * zoomRatioHeader;

    if (p_localBbox.height - 2 * lineWeight >= 0 && headerWidth - 2 * lineWeight > 0) {
      let topbarBbox = _types.default.newRectangle(lineWeight, lineWeight, headerWidth - 2 * lineWeight, p_localBbox.height - 2 * lineWeight);

      this.pushToDisplayList((0, _rectangle.makeFilledRectangle)(topbarBbox, this.style("timepicker_headingPanelBgColor"))); //Labels

      this.pushToDisplayList(this.digitalClock(topbarBbox, zoomRatioHeader, Directions.LANDSCAPE));
    } //Clock


    let zoomRatioClock = p_localBbox.height < 25 ? 0 : Math.abs(Math.min((p_localBbox.height - (buttonHeight + padding)) / (defaultHeightLandscape - (buttonHeight + padding)), (p_localBbox.width - headerWidth) / (defaultWidthlandscape - defaultHeaderWidth)));
    this.analogClock(p_localBbox, Directions.LANDSCAPE, zoomRatioClock, headerWidth);
  }

  _link.default.processLinks(this, linkBounds);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/titles.js":
/*!************************************!*\
  !*** ./js/core/controls/titles.js ***!
  \************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.title = exports.subtitle = exports["default"] = void 0;

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _macro = __webpack_require__(/*! core/bricks/macro */ "./js/core/bricks/macro.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var makeTitleControlClass = function (p_controlTypeID, p_controlTypeName, p_controlAlternateNames, p_defaultText, p_defaultAlign, p_defaultFontSize) {
  var control = controlsMod.makeControlClass({
    controlTypeID: p_controlTypeID,
    controlTypeName: p_controlTypeName,
    controlAlternateNames: p_controlAlternateNames,
    props: { ...Props.PropertyDefinitions.SIZE({
        defaultValue: _types.default.newAutoSize(),
        options: {
          autosize: Props.AutoSizeOptions.ONLY_WIDTH,
          resize: Props.ResizeOptions.HORIZONTAL
        }
      }),
      align: {
        defaultValue: p_defaultAlign,
        enum: Props.TextAlignments
      },
      ...Props.PropertyDefinitions.link
    }
  });
  controlsMod.addStandardTextPropertyDefinitions(control, {
    text: p_defaultText,
    fontSize: p_defaultFontSize,
    charCountEnabled: true,
    noEmptyText: true
  });
  controlsMod.addColorPropertyDefinitions(control, {
    defaultTextColor: "default_text_color",
    oldTextColorName: 'color'
  });
  controlsMod.addTextClipBBoxCalculation(control);
  controlsMod.addStandardCalculateOriginForNewTextOrSize(control);

  control.CONTROL_PROTOTYPE.initData = function () {
    this._clipOnBbox = false;
    this.overflow = "visible";
  };

  control.CONTROL_PROTOTYPE.updateData = function () {
    const text = (0, _macro.processMacros)(this);
    const textStyle = this.newStandardTextStyle({
      textColor: this.getActualProperty("textColor")
    });

    if (!this.text || !(0, _text.areTextsEqual)(this.text, text, textStyle)) {
      this.text = p_controlTypeID === "com.balsamiq.mockups::Title" ? (0, _text.newTextNoWrap)(text, textStyle, this._context) : (0, _text.newSingleLineText)(text, textStyle, this._context);
    }
  };

  control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
    var textSize = this.text.measuredSize();
    var width = p_sizeProperty.width === _types.default.AUTOSIZE ? textSize.width : p_sizeProperty.width;
    return _types.default.newSize(width + this.style("titles_padding_left") + this.style("titles_padding_right"), textSize.height + this.style("titles_padding_bottom"));
  };

  control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
    _link.default.processControlLink(this);

    var textBbox = _types.default.paddedRectangle(p_localBbox, p_controlTypeID == "com.balsamiq.mockups::Title" ? this.style("title_padding_top") : this.style("subtitle_padding_top"), this.style("titles_padding_bottom"), this.style("titles_padding_left"), this.style("titles_padding_right"));

    this.pushToDisplayList(this.text.makeDisplayItems(textBbox, this.getActualProperty("align")));
  };

  return control;
};

const title = makeTitleControlClass("com.balsamiq.mockups::Title", "Title", ["Headline", "Big Title"], "A Big Title", Props.TextAlignments.LEFT, 40);
exports.title = title;
const subtitle = makeTitleControlClass("com.balsamiq.mockups::SubTitle", "Subtitle", [], "A Subtitle", Props.TextAlignments.LEFT, 24);
exports.subtitle = subtitle;
var _default = {
  title,
  subtitle
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/tooltip.js":
/*!*************************************!*\
  !*** ./js/core/controls/tooltip.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::Tooltip",
  controlTypeName: "Tooltip",
  controlAlternateNames: ["Balloon", "Speech Bubble"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newAutoSize(),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    tooltipDirection: {
      defaultValue: Props.TooltipDirections.NW,
      enum: Props.TooltipDirections
    },
    align: {
      defaultValue: Props.TextAlignments.CENTER,
      enum: Props.TextAlignments
    }
  }
});
controlsMod.addStandardTextPropertyDefinitions(control, {
  text: "a tooltip",
  multiline: true
});
var C1 = 8,
    C2 = 7;

control.CONTROL_PROTOTYPE.updateData = function () {
  this.text = textMod.newText(this.getActualProperty("text"), this.newStandardTextStyle(), this._context);
};

_link.default.addTextLinksMethods(control);

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var width = this.style("tooltip_padding_left") + this.style("tooltip_padding_right") + this.text.measuredSize().width + 10;
  var height = Math.max(this.styleSVGSize("tooltip").height + C1, this.text.measuredSize().height);
  return _types.default.newSize(width, height);
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var dir = this.getActualProperty("tooltipDirection");
  var mirrorX = false,
      mirrorY = false;
  var baloonY = 0,
      tipX = 0,
      tipY = 0;
  var tipSize = this.styleSVGSize("tooltip_tip");

  if (dir === Props.TooltipDirections.NW) {
    baloonY = C2;
    tipX = C1;
  } else if (dir === Props.TooltipDirections.NE) {
    mirrorX = true;
    baloonY = C2;
    tipX = p_localBbox.width - C1;
  } else if (dir === Props.TooltipDirections.SW) {
    mirrorY = true;
    tipX = C1;
    tipY = p_localBbox.height;
  } else if (dir === Props.TooltipDirections.SE) {
    mirrorX = true;
    mirrorY = true;
    tipX = p_localBbox.width - C1;
    tipY = p_localBbox.height;
  }

  var baloonBounds = _types.default.newRectangle(0, baloonY, p_localBbox.width, p_localBbox.height - C2);

  var baloon = _displayitem.default.createDisplayItemsFromSkin("tooltip", null, baloonBounds, this._context);

  var tip;

  if (p_localBbox.width >= tipSize.width * 2 && p_localBbox.height >= tipSize.height * 2) {
    var tipBounds = _types.default.newRectangle(tipX, tipY, tipSize.width, tipSize.height);

    tip = _displayitem.default.createDisplayItemsFromSkin("tooltip_tip", null, tipBounds, this._context);

    _displayitem.default.mirrorDisplayItems(tip, tipBounds, mirrorX, mirrorY);
  }

  var label;
  var textWidth = p_localBbox.width - this.style("tooltip_padding_left") - this.style("tooltip_padding_right");

  if (textWidth > 0) {
    var textBounds = _types.default.newRectangle(this.style("tooltip_padding_left"), this.style("tooltip_padding_top") + baloonY, textWidth, baloonBounds.height);

    label = this.text.makeDisplayItemsAndLinks(textBounds, this.getActualProperty("align"), _types.default.Positions.CENTER);

    _link.default.processTextLinks(this, label.links);
  }

  this.pushToDisplayList(baloon, tip, label ? label.displayItems : null);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/tree.js":
/*!**********************************!*\
  !*** ./js/core/controls/tree.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _scrollbars = __webpack_require__(/*! core/bricks/scrollbars */ "./js/core/bricks/scrollbars.js");

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _link = _interopRequireDefault(__webpack_require__(/*! core/bricks/link */ "./js/core/bricks/link.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var wikiMod = __webpack_require__(/*! core/texts/wiki */ "./js/core/texts/wiki.js");

var PADDING_LEFT = 8; // Space between border and text

var PADDING_TOP = 6;
var PADDING_BOTTOM = 6;
var PADDING_MIDDLE = 4;
var TEXT_TOP_SPACE = 2;
var TAB_WIDTH = 25;
var ICON_SIZES = {
  "_": 20,
  "folder": 20,
  "folder-open": 20,
  "plus-square-o": 20,
  "minus-square-o": 20,
  "check-square-o": 20,
  "square-o": 20,
  "caret-right": 26,
  'caret-left': 26,
  // Support for RTL
  "caret-down": 26,
  "file-text-o": 20
};
var ICON_PADDINGS = {
  "_": 0,
  "folder": 1,
  "folder-open": 2,
  "plus-square-o": 0,
  "minus-square-o": 0,
  "check-square-o": 1,
  "square-o": 0,
  "caret-right": 0,
  'caret-left': 0,
  // Support for RTL
  "caret-down": 0,
  "file-text-o": 1
};
var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::Tree",
  controlTypeName: "Tree Pane",
  controlAlternateNames: [],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(300, 285),
      options: {
        autosize: Props.AutoSizeOptions.WIDTH_AND_HEIGHT,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    rtl: {
      defaultValue: false,
      dataType: 'bool'
    },
    ...Props.PropertyDefinitions.BORDER_STYLE(Props.SimpleBorderStyles.SQUARE, Props.SimpleBorderStyles),
    ...Props.PropertyDefinitions.SCROLLBAR(),
    ...Props.PropertyDefinitions.links(),
    state: {
      defaultValue: Props.BoolStates.NORMAL,
      enum: Props.BoolStates
    }
  }
});
controlsMod.addStandardTextPropertyDefinitions(control, {
  text: "f Use f for closed folders\nF Use F for open folders\n[+] You may also use this\n[-] and this\n[x] or this\n[ ] and this\n> or even this\nv and this\n- Use - for a file icon\n_ or _ to leave a space for your own icon\nF use spaces or dots for hierarchy\n v just like\n..- this\n",
  multiline: true,
  splitOn: Props.SplitOnTypes.NEWLINE
});
controlsMod.addColorPropertyDefinitions(control, {
  defaultColor: 0xFFFFFF,
  hasAlpha: true
});
controlsMod.addSelectionPropertyDefinitions(control, -1);

control.CONTROL_PROTOTYPE.getSelectableLabels = function () {
  return this._selectableLabels;
};

control.CONTROL_PROTOTYPE.getLinkNames = function () {
  return this._selectableLabels;
};

control.CONTROL_PROTOTYPE.getLinkableLabels = function (text) {
  var lines = wikiMod.parseTreeText(text);
  return lines.map(line => TextUtils.unformatText(line.text));
};

control.CONTROL_PROTOTYPE.updateData = function () {
  this._textLines = [];
  var text = this.getActualProperty("text");
  var textColor, linkColor;
  this._lines = wikiMod.parseTreeText(text);

  this._lines.forEach((line, rowIndex) => {
    if (this.getActualProperty("state") !== _types.default.ControlState.DISABLED) {
      textColor = _colors.default.labelColorGivenBackgroundColor(rowIndex != this.getActualProperty("selectedIndex") ? this.getActualProperty('color') : this.style("default_selection_color"), this.getActualProperty("backgroundAlpha"), this.style("default_text_color"));
      linkColor = _colors.default.labelColorGivenBackgroundColor(rowIndex != this.getActualProperty("selectedIndex") ? this.getActualProperty('color') : this.style("default_selection_color"), this.getActualProperty("backgroundAlpha"), this.style("default_link_color"));
    } else {
      textColor = this.style("default_disabled_color");
      linkColor = textColor;
    }

    var textStyle = this.newStandardTextStyle({
      textColor: textColor,
      linkColor: linkColor
    });
    var textLine = textMod.newText(line.text, textStyle, this._context);

    this._textLines.push(textLine);
  });

  this._selectableLabels = this._lines.map(line => TextUtils.unformatText(line.text));
};

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  var size = _balsamiq.default.clone(p_sizeProperty);

  if (p_sizeProperty.width === _types.default.AUTOSIZE) {
    size.width = this.propertyDefinitions.size.defaultValue.width;
  }

  var h;

  if (p_sizeProperty.height === _types.default.AUTOSIZE) {
    h = PADDING_BOTTOM + PADDING_TOP;

    this._textLines.forEach(textLine => {
      h += textLine.measuredSize().height + 2 * PADDING_MIDDLE;
    });
  } else {
    h = p_sizeProperty.height;
  }

  size.height = h;
  return size;
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var borderStyle = this.getActualProperty("borderStyle");
  var paddingLeft = borderStyle === Props.BorderStyles.NONE ? 0 : PADDING_LEFT;
  var paddingTop = borderStyle === Props.BorderStyles.NONE ? 0 : PADDING_TOP;

  var disabled = this.getActualProperty("state") === _types.default.ControlState.DISABLED; // White or colored background


  var background = (0, _rectangle.makeFilledRectangle)(p_localBbox, this.getActualProperty("color"), this.getActualProperty('backgroundAlpha'), this.style('rectangle_corner_radius')[borderStyle]); // Solid fill

  this.pushToDisplayList(background); // Actual text

  var theY = paddingTop;
  var selectedIndex = this.getActualProperty("selectedIndex");
  var iconColor;
  var linkBounds = [];

  this._lines.forEach((line, idx) => {
    var textLine = this._textLines[idx];
    var iconX = paddingLeft + TAB_WIDTH * line.level;
    var iconSize = ICON_SIZES[line.symbol];
    var rowHeight = textLine.measuredSize().height + 2 * PADDING_MIDDLE; // Selection

    if (selectedIndex === idx) {
      const previewRowsHeight = this._textLines.slice(0, idx).reduce((previewRowsHeight, textLine) => previewRowsHeight + (textLine.measuredSize().height + 2 * PADDING_MIDDLE), 0);

      const selBBox = _types.default.newRectangle(0, previewRowsHeight + paddingTop, p_localBbox.width, rowHeight);

      this.pushToDisplayList((0, _rectangle.makeFilledRectangle)(selBBox, this.style("default_selection_color")));
    } // Text


    var textBounds;

    if (!this.getActualProperty("rtl")) {
      textBounds = _types.default.newRectangle(iconX + 26, theY + PADDING_MIDDLE + TEXT_TOP_SPACE, textLine.measuredSize().width, textLine.measuredSize().height);
    } else {
      textBounds = _types.default.newRectangle(this._bbox.width - paddingLeft - TAB_WIDTH * line.level, theY + PADDING_MIDDLE + TEXT_TOP_SPACE, textLine.measuredSize().width, textLine.measuredSize().height);
    }

    var textDisplayItem = textLine.makeDisplayItems(textBounds);
    this.pushToDisplayList(textDisplayItem);
    linkBounds.push(textBounds); // Icon

    if (!disabled) {
      iconColor = _colors.default.labelColorGivenBackgroundColor(idx != selectedIndex ? this.getActualProperty('color') : this.style("default_selection_color"), this.getActualProperty('backgroundAlpha'), this.style("default_icon_color"));
    } else {
      iconColor = this.style("default_disabled_color");
    }

    if (line.symbol && line.symbol !== "_") {
      var iconData = _icon.default.newAppIconData(line.symbol, iconSize, 0);

      this.pushToDisplayList(_icon.default.makeDisplayItemsFromIconData(this, iconData, iconX + ICON_PADDINGS[line.symbol], theY + (rowHeight - iconSize) / 2 + 1, iconColor));
    }

    theY = theY + rowHeight;
  });

  _link.default.processLinks(this, linkBounds); // Scrollbar


  if (this.getActualProperty("verticalScrollbar")) {
    this.pushToDisplayList((0, _scrollbars.makeVerticalScrollbar)(p_localBbox, this.style("tree_scrollbar_padding_top"), this.style("tree_scrollbar_padding_bottom"), this.style("tree_scrollbar_padding_right"), this.getActualProperty('value'), this.getActualProperty("state"), this._context));
  } // Black border


  if (borderStyle === Props.BorderStyles.SQUARE) {
    this.pushToDisplayList(_displayitem.default.createDisplayItemsFromSkin("rectangle_border_square", !disabled ? this.style("default_border_color") : this.style("default_disabled_color"), p_localBbox, this._context));
  }
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/videoplayer.js":
/*!*****************************************!*\
  !*** ./js/core/controls/videoplayer.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeAssetBasedControlClass({
  controlTypeID: "com.balsamiq.mockups::VideoPlayer",
  controlTypeName: "Video Player",
  controlAlternateNames: [],
  defaultSize: _types.default.newSize(300, 200),
  resizeOption: Props.ResizeOptions.BOTH,
  skinAssetName: null,
  colorStyleName: null,
  otherPropertyDefinitions: null
});

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  var background = (0, _rectangle.makeFilledRectangle)(_types.default.shrinkRectangle(p_localBbox, 2), 0xFFFFFF);

  var asset = _displayitem.default.createDisplayItemsFromSkin("videoplayer", this.style("videoplayer_color"), p_localBbox, this._context);

  this.pushToDisplayList(background, asset);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/volumeslider.js":
/*!******************************************!*\
  !*** ./js/core/controls/volumeslider.js ***!
  \******************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

const HEIGHT = 16;
var control = controlsMod.makeAssetBasedControlClass({
  controlTypeID: "com.balsamiq.mockups::VolumeSlider",
  controlTypeName: "Volume Slider",
  controlAlternateNames: [],
  defaultSize: _types.default.newSize(72, HEIGHT),
  resizeOption: Props.ResizeOptions.HORIZONTAL,
  skinAssetName: "volumeslider",
  colorStyleName: "default_border_color",
  otherPropertyDefinitions: null
});

control.CONTROL_PROTOTYPE.measure = function (p_sizeProperty) {
  return _types.default.newSize(p_sizeProperty.width, HEIGHT);
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/webcam.js":
/*!************************************!*\
  !*** ./js/core/controls/webcam.js ***!
  \************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var _default = controlsMod.makeAssetBasedControlClass({
  controlTypeID: "com.balsamiq.mockups::Webcam",
  controlTypeName: "Webcam",
  controlAlternateNames: [],
  defaultSize: _types.default.newSize(177, 146),
  resizeOption: Props.ResizeOptions.BOTH,
  skinAssetName: "webcam",
  colorStyleName: "webcam_color",
  otherPropertyDefinitions: null
});

exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/window.js":
/*!************************************!*\
  !*** ./js/core/controls/window.js ***!
  \************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _text = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var _scrollbars = __webpack_require__(/*! core/bricks/scrollbars */ "./js/core/bricks/scrollbars.js");

var _rectangle = __webpack_require__(/*! core/bricks/rectangle */ "./js/core/bricks/rectangle.js");

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var _snapping = __webpack_require__(/*! core/controls/snapping */ "./js/core/controls/snapping.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var controlsMod = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var control = controlsMod.makeControlClass({
  controlTypeID: "com.balsamiq.mockups::TitleWindow",
  controlTypeName: "Window",
  controlAlternateNames: ["Dialog"],
  props: { ...Props.PropertyDefinitions.SIZE({
      defaultValue: _types.default.newSize(450, 400),
      options: {
        autosize: Props.AutoSizeOptions.NO,
        resize: Props.ResizeOptions.BOTH
      }
    }),
    ...Props.PropertyDefinitions.TEXT({
      defaultText: 'Window Name',
      multiline: false
    }),
    ...Props.PropertyDefinitions.SCROLLBAR(),
    bottomheight: {
      defaultValue: 25,
      dataType: 'int'
    },
    topheight: {
      defaultValue: 26,
      dataType: 'int'
    },
    minimize: {
      defaultValue: false,
      dataType: 'bool'
    },
    maximizeRestore: {
      defaultValue: false,
      dataType: 'bool'
    },
    dragger: {
      defaultValue: false,
      dataType: 'bool'
    },
    close: {
      defaultValue: false,
      dataType: 'bool'
    }
  }
});
controlsMod.addStandardTextPropertyDefinitions(control);

control.CONTROL_PROTOTYPE.initData = function () {
  this._topChromeHeight = 0;
  this._bottomChromeHeight = null;
};

control.CONTROL_PROTOTYPE.makeDisplayItemsForIcon = function (p_iconName, p_iconProp, p_localBbox) {
  let iconSize = 13; // var paddingLeft = this.style("window_" + p_iconProp +"_padding_left");

  var paddingRight = this.style("window_" + p_iconProp + "_padding_right");
  var paddingTop = this.style("window_" + p_iconProp + "_padding_top");
  var windowTooSmall = p_localBbox.height < iconSize + paddingTop;
  var x = p_localBbox.x + (p_localBbox.width - paddingRight - iconSize);
  var res;

  if (!windowTooSmall && x > 0) {
    let iconData = _icon.default.newAppIconData(p_iconName, iconSize, 0);

    res = _icon.default.makeDisplayItemsFromIconData(this, iconData, x, p_localBbox.y + paddingTop, 0);
  } else {
    res = null;
  }

  return res;
};

control.CONTROL_PROTOTYPE.updateLayout = function (p_localBbox) {
  const text = this.getActualProperty("text");
  const textStyle = this.newStandardTextStyle();

  if (!this.textLine || !(0, _text.areTextsEqual)(this.textLine, text, textStyle)) {
    this.textLine = (0, _text.newSingleLineText)(text, textStyle, this._context);
  }

  var whiteBackground = (0, _rectangle.makeFilledRectangle)(p_localBbox, 0xFFFFFF);

  var border = _displayitem.default.createDisplayItemsFromSkin("rectangle_border_square", this.style("default_border_color"), p_localBbox, this._context);

  var topHeight = Math.max(this.getActualProperty("topheight"), this.textLine.measuredSize().height);
  this._topChromeHeight = topHeight;

  var topChromeBbox = _types.default.newRectangle(p_localBbox.x, p_localBbox.y, p_localBbox.width, topHeight);

  var topChrome = _displayitem.default.createDisplayItemsFromSkin("rectangle_border_square", this.style("default_border_color"), topChromeBbox, this._context);

  var bottomHeight = this.getActualProperty("bottomheight");
  var bottomChrome;

  if (p_localBbox.height - bottomHeight > topHeight) {
    var bottomChromeBbox = _types.default.newRectangle(p_localBbox.x, p_localBbox.y + p_localBbox.height - bottomHeight, p_localBbox.width, Math.min(bottomHeight, p_localBbox.height));

    bottomChrome = _displayitem.default.createDisplayItemsFromSkin("rectangle_border_square", this.style("default_border_color"), bottomChromeBbox, this._context);
    this._bottomChromeHeight = bottomHeight;
  } else {
    this._bottomChromeHeight = null;
  }

  var scrollbar;

  if (this.getActualProperty("verticalScrollbar") && p_localBbox.width >= this.style("window_scrollbar_padding_right") + _scrollbars.Sizes.ScrollbarFixedSize) {
    scrollbar = (0, _scrollbars.makeVerticalScrollbar)(p_localBbox, topHeight + this.style("window_scrollbar_padding_top"), bottomHeight + this.style("window_scrollbar_padding_bottom"), this.style("window_scrollbar_padding_right"), this.getActualProperty("value"), _types.default.ControlState.NORMAL, this._context);
  }

  var titleBbox = _types.default.newRectangle(p_localBbox.x + this.style("window_title_padding_left"), p_localBbox.y + this.style("window_title_padding_top"), Math.max(p_localBbox.width - this.style("window_title_padding_left") - this.style("window_title_padding_right"), 0), this.textLine.measuredSize().height);

  var title = this.textLine.makeDisplayItems(titleBbox);
  let iconNames = ["window-minimize", "window-maximize", "times"];
  var icons = ["minimize", "maximizeRestore", "close"].map((iconProp, i) => {
    var res = null;

    if (this.getActualProperty(iconProp)) {
      res = this.makeDisplayItemsForIcon(iconNames[i], iconProp, p_localBbox);
    }

    return res;
  });
  var dragger;

  if (this.getActualProperty("dragger")) {
    var size = this.styleSVGSize("window_dragger");
    var paddingBottom = this.style("window_dragger_padding_bottom");
    var windowTooSmall = p_localBbox.height < size.height + paddingBottom;
    var paddingRight = this.style("window_dragger_padding_right");
    var x = p_localBbox.x + p_localBbox.width - paddingRight;

    if (!windowTooSmall && x > 0) {
      var bbox = _types.default.newRectangle(x, p_localBbox.y + p_localBbox.height - paddingBottom, size.width, size.height);

      dragger = _displayitem.default.createDisplayItemsFromSkin("window_dragger", this.style("default_border_color"), bbox, this._context);
    }
  }

  this.pushToDisplayList(whiteBackground, border, topChrome, bottomChrome, scrollbar, title, icons, dragger);
};

control.CONTROL_PROTOTYPE.getSnapLines = function (withMarkup) {
  _balsamiq.default.assertBool(withMarkup);

  var bbox = this.getUserBBox(true);
  var lines = [new _snapping.SnapLine(_snapping.SnapTypes.MIDDLE, _snapping.SnapDirections.VERTICAL, bbox.x + bbox.width / 2), new _snapping.SnapLine(_snapping.SnapTypes.MIDDLE, _snapping.SnapDirections.HORIZONTAL, bbox.y + bbox.height / 2), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.VERTICAL, bbox.x), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.VERTICAL, bbox.x + bbox.width), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.HORIZONTAL, bbox.y), new _snapping.SnapLine(_snapping.SnapTypes.COINCIDENCE, _snapping.SnapDirections.HORIZONTAL, bbox.y + bbox.height), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.VERTICAL, bbox.x + _snapping.PROXIMITY_DISTANCE), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.VERTICAL, bbox.x + bbox.width - _snapping.PROXIMITY_DISTANCE), new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.HORIZONTAL, bbox.y + _snapping.PROXIMITY_DISTANCE + this._topChromeHeight)];

  if (this._bottomChromeHeight !== null) {
    lines.push(new _snapping.SnapLine(_snapping.SnapTypes.PROXIMITY, _snapping.SnapDirections.HORIZONTAL, bbox.y + bbox.height - _snapping.PROXIMITY_DISTANCE - this._bottomChromeHeight));
  }

  return lines;
};

var _default = control;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/core.js":
/*!*************************!*\
  !*** ./js/core/core.js ***!
  \*************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Core = void 0;
exports.makeCore = makeCore;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _sketch = _interopRequireDefault(__webpack_require__(/*! core/skins/sketch */ "./js/core/skins/sketch.js"));

var _wireframe = _interopRequireDefault(__webpack_require__(/*! core/skins/wireframe */ "./js/core/skins/wireframe.js"));

var _textmeter = __webpack_require__(/*! core/textmeter */ "./js/core/textmeter.ts");

var Model = _interopRequireWildcard(__webpack_require__(/*! core/core/model */ "./js/core/core/model.js"));

var _controllers = __webpack_require__(/*! core/core/controllers */ "./js/core/core/controllers.js");

var _viewmodels = __webpack_require__(/*! core/core/viewmodels */ "./js/core/core/viewmodels.js");

var _operationqueue = __webpack_require__(/*! core/operations/operationqueue */ "./js/core/operations/operationqueue.js");

var _operationcenter = __webpack_require__(/*! core/operations/operationcenter */ "./js/core/operations/operationcenter.js");

var _exectime = __webpack_require__(/*! core/utils/exectime */ "./js/core/utils/exectime.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var AllControls = __webpack_require__(/*! core/controls/all */ "./js/core/controls/all.js"); // required to build!


var Events = __webpack_require__(/*! core/events */ "./js/core/events.ts");

var ViewControllers = __webpack_require__(/*! core/core/viewcontrollers */ "./js/core/core/viewcontrollers.js");

var Archives = __webpack_require__(/*! core/core/archives */ "./js/core/core/archives.js");

var Settings = (__webpack_require__(/*! core/core/settings */ "./js/core/core/settings.js").Settings);

var Clipboard = (__webpack_require__(/*! core/core/clipboard */ "./js/core/core/clipboard.js").Clipboard);

var CoreObject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var Tasks = __webpack_require__(/*! core/core/tasks */ "./js/core/core/tasks.ts");

/**
 * The class that provides all the services to the app:
 * - a logger
 * - a tool to monitor execution times: {@link ExecTime}
 * - a text meter to measure bounding box of formatted text strings: {@link TextMeter}
 * - the event center: {@link EventCenter}
 * - a way to load and save user settings: {@link Settings}
 * - the model of BW: {@link Model}
 * - a queue for exclusive access to the Model: {@link OperationQueue}
 * - an undo/redo stack: {@link OperationCenter}
 *
 * Services for dealing with Balsamiq Archives:
 *  - all the open archives are here: {@link Archives}
 *  - the controllers are here: {@link Controllers}
 *
 * These services are for UI stuff:
 * - view models: {@link ViewModels}
 * - a hierarchy of view controllers: {@link ViewControllers}
 * - a stack of ongoing user actions: {@link Tasks}
 *
 * It is meant to be a singleton, at the least for the moment.
 *
 * The Core instance attaches itself to the hooks of native.js.
 *
 * @constructor
 * @param p_context {Object} All the external dependecies
 * @param p_coreID {String} Unique identifiers in case of multiple running Cores
 * */
class Core {
  constructor(p_context, p_coreID, p_readyCallback) {
    _balsamiq.default.exists(p_context);

    _balsamiq.default.exists(p_context.host);

    _balsamiq.default.exists(p_context.guest);

    _balsamiq.default.exists(p_coreID);

    _balsamiq.default.exists(p_context.windowID);

    this.coreID = p_coreID;
    this.windowID = p_context.windowID; // Async initializations ready status check

    this.settingsReady = false; // Bind the methods to this

    this.getCoreObject = this._getCoreObject.bind(this);
    this.findCoreObjectID = this._findCoreObjectID.bind(this);
    this.runTask = this._runTask.bind(this);
    this.runningTask = this._runningTask.bind(this);
    this.taskAction = this._taskAction.bind(this);
    this.runTaskIfPossible = this._runTaskIfPossible.bind(this);
    this.runMainTask = this._runMainTask.bind(this); // Checks if all async initializations have been completed
    // If so, calls the external ready callback

    const readyCallback = () => {
      if (this.settingsReady) {
        // add check for all async constructors here
        p_readyCallback && p_readyCallback(this);
      }
    };

    this.eventCenter = new Events.EventCenter(p_coreID);
    this.skins = {
      sketch: _sketch.default,
      wireframe: _wireframe.default
    };
    this.textMeter = new _textmeter.TextMeter(p_context.host.TextMeterHelper);
    this.logger = p_context.host.log;
    this.execTime = new _exectime.ExecTime({
      host: p_context.host
    });
    this.settings = new Settings({
      getCoreObject: this.getCoreObject,
      host: p_context.host,
      eventCenter: this.eventCenter
    }, 'settings', () => {
      // set settings ready status true
      // and perform global check
      this.settingsReady = true;
      readyCallback();
    });
    this.model = new Model.Model({
      host: p_context.host,
      guest: p_context.guest,
      skins: this.skins,
      textMeter: this.textMeter,
      logger: this.logger,
      eventCenter: this.eventCenter,
      getCoreObject: this.getCoreObject,
      execTime: this.execTime,
      useFontAwesomeIcons: true // !Balsamiq.isUndefined(p_context.useFontAwesomeIcons) ? p_context.useFontAwesomeIcons : false

    });
    this.operationQueue = new _operationqueue.OperationQueue({
      getCoreObject: this.getCoreObject,
      eventCenter: this.eventCenter,
      host: p_context.host
    }, "mainQueue");
    this.operationCenter = new _operationcenter.OperationCenter({
      getCoreObject: this.getCoreObject,
      eventCenter: this.eventCenter,
      operationQueue: this.operationQueue
    }, "mainCenter");
    this.viewModels = new _viewmodels.ViewModels({
      execTime: this.execTime
    });
    this.clipboard = new Clipboard({
      getCoreObject: this.getCoreObject,
      host: p_context.host,
      eventCenter: this.eventCenter
    }, 'clipboard');
    this.archives = new Archives.Archives({
      getCoreObject: this.getCoreObject,
      eventCenter: this.eventCenter,
      host: p_context.host,
      guest: p_context.guest,
      execTime: this.execTime
    });
    this.controllers = new _controllers.Controllers({
      getCoreObject: this.getCoreObject,
      eventCenter: this.eventCenter,
      archives: this.archives,
      host: p_context.host,
      operationQueue: this.operationQueue,
      settings: this.settings,
      execTime: this.execTime
    });
    this.viewControllers = new ViewControllers.ViewControllers({
      getCoreObject: this.getCoreObject,
      coreID: this.coreID,
      eventCenter: this.eventCenter,
      model: this.model,
      settings: this.settings,
      clipboard: this.clipboard,
      controllers: this.controllers,
      host: p_context.host,
      guest: p_context.guest,
      runTask: this.runTask,
      runTaskIfPossible: this.runTaskIfPossible,
      runMainTask: this.runMainTask,
      runningTask: this.runningTask,
      taskAction: this.taskAction,
      textMeter: this.textMeter,
      execTime: this.execTime
    }, ViewControllers.makeViewControllersCoreObjectID());
    this.tasks = new Tasks.Tasks({
      viewControllers: this.viewControllers,
      viewModels: this.viewModels,
      model: this.model,
      getCoreObject: this.getCoreObject,
      findCoreObjectID: this.findCoreObjectID,
      eventCenter: this.eventCenter,
      controllers: this.controllers,
      host: p_context.host,
      archives: this.archives,
      operationCenter: this.operationCenter,
      settings: this.settings,
      textMeter: this.textMeter,
      skins: this.skins,
      execTime: this.execTime,
      clipboard: this.clipboard,
      coreID: this.coreID
    });

    p_context.guest.App._onAwake = () => {
      this.archives.onAwake();
    };

    p_context.guest.App.onFontsLoaded = function () {
      p_context.host.log('Core: fonts loaded');
      this.model.onContextChange();
    }.bind(this);

    p_context.guest.UI._action = function (id, action, params) {
      this.viewControllers.onNativeAction(id, action, params);
    }.bind(this);

    p_context.guest.UI._keyboardEvent = function (id, event) {
      return this.viewControllers.onNativeKeyboardEvent(id, event);
    }.bind(this);

    p_context.guest.UI._mouseEvent = function (id, event) {
      return this.viewControllers.onNativeMouseEvent(id, event);
    }.bind(this);

    p_context.guest.LocalStorage.settingsChangedEvent = function (data) {
      return this.settings.settingsChangedEvent(data);
    }.bind(this);

    p_context.guest.SessionStorage.settingsChangedEvent = function (data) {
      return this.settings.sessionSettingsChangedEvent(data);
    }.bind(this);

    p_context.guest.Clipboard.clipboardChangedEvent = function (data) {
      return this.clipboard.clipboardChangedEvent(data);
    }.bind(this);

    p_context.guest.Cloud._onLogin = (domain, result, userInfo) => {
      let data = {
        domain,
        result,
        userInfo
      };
      this.eventCenter.fireEvent(Events.newCoreEvent(CoreObject.makeVoidCoreObjectID(), Events.EventTypes.CLOUD_LOGIN, data));
    };

    const logNetworkStatus = () => p_context.host.log(`Network is ${p_context.host.Network.isOnline() ? "online" : "offline"}`);

    logNetworkStatus();

    p_context.guest.Network._onNetworkStatusChange = isOnline => {
      logNetworkStatus();
      this.archives.onNetworkStatusChange(isOnline);
    };

    p_context.guest.Desktop.License.onDataChanged = data => {
      p_context.host.log(`Local Files License data changed`);
      this.eventCenter.fireEvent(Events.newCoreEvent(CoreObject.makeVoidCoreObjectID(), Events.EventTypes.DESKTOP_LICENSE_CHANGED, {
        data: data
      }));
    };

    p_context.guest.Desktop._onUserInfoChange = info => {
      if (info && info.email) {
        info.email = info.email.toLocaleLowerCase();
      }

      this.eventCenter.fireEvent(Events.newCoreEvent(CoreObject.makeVoidCoreObjectID(), Events.EventTypes.USER_INFO_CHANGED, info));
    };
  }

  cleanUp() {
    this.model.cleanUp();
    this.model = null;
  } // Methods bound to this
  // ---------------------


  _getCoreObject(coreObjectID) {
    _balsamiq.default.existsClass(coreObjectID, CoreObject.CoreObjectID);

    var object = this.model.getModelObject(coreObjectID);

    if (!object) {
      object = this.viewModels.getViewModel(coreObjectID);
    }

    if (!object) {
      object = this.viewControllers.getViewController(coreObjectID);
    } //Balsamiq.exists(object, "can't find a CoreObject with CoreObjectID: " + coreObjectID.toString());


    return object;
  }

  _findCoreObjectID(genericObjectID) {
    _balsamiq.default.exists(genericObjectID);

    return this.model.findModelObjectID(genericObjectID);
  }

  _runTask(parentTaskID, task, callback) {
    this.tasks.runTask(parentTaskID, task, callback);
  }

  _runTaskIfPossible(parentTaskID, task, callback) {
    return this.tasks.runTaskIfPossible(parentTaskID, task, callback);
  }

  _runningTask(windowID) {
    _balsamiq.default.assertString(windowID);

    return this.tasks.runningTask(windowID);
  }

  _runMainTask(windowID, task, callback) {
    return this.tasks.runMainTask(windowID, task, callback);
  }

  _taskAction(taskID, action, params) {
    return this.tasks.sendAction(taskID, action, params);
  }

}

exports.Core = Core;

function makeCore(host, guest, coreID, windowID, readyCallback) {
  let core = new Core({
    host: host,
    guest: guest,
    windowID: windowID
  }, coreID, readyCallback);
  return core;
}

/***/ }),

/***/ "./js/core/core/archives.js":
/*!**********************************!*\
  !*** ./js/core/core/archives.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Archives = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var ArchiveProtocol = _interopRequireWildcard(__webpack_require__(/*! core/archive */ "./js/core/archive.js"));

var _BalsamiqArchiveInMemory = __webpack_require__(/*! core/BalsamiqArchiveInMemory */ "./js/core/BalsamiqArchiveInMemory.js");

var _sqlitedriver = __webpack_require__(/*! core/sqlitedriver */ "./js/core/sqlitedriver.js");

var Native = _interopRequireWildcard(__webpack_require__(/*! core/native */ "./js/core/native.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var _nativeConstants = __webpack_require__(/*! core/native-constants */ "./js/core/native-constants.js");

var _version = __webpack_require__(/*! core/version */ "./js/core/version.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var ArchiveClient = __webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchiveClient */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveClient.js");

var Archive = __webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchive */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchive.js");

var Events = __webpack_require__(/*! core/events */ "./js/core/events.ts");

var CoreObject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var nativeNetworkApi = function (method, url, params, cb, timeouts, settings, requestCreatedCallback) {
  if (_balsamiq.default.isObject(settings) && settings.binary) {
    _balsamiq.default.assertBool(settings.base64, "nativeNetworkApi requires base64 parameter");

    let hostRequestParams = {
      base64: settings.base64,
      progress: settings.progress,
      timeout: timeouts.deadline
    };
    Native.Host.HttpRequest.getBinary(url, hostRequestParams, cb, requestCreatedCallback);
  } else if (_balsamiq.default.isObject(settings) && settings.formData) {
    Native.Host.HttpRequest.postFormData(url, params, cb);
  } else {
    Native.Host.HttpRequest.ajax({
      method: method,
      url: url,
      params: params,
      timeouts: timeouts,
      headers: settings ? settings.headers : undefined,
      callback: cb
    });
  }
};

var nativeTimerApi = {
  setTimeout: function (callback, ms) {
    return Native.Host.timer(ms, callback);
  },
  clearTimeout: function (timer) {
    Native.Host.clearTimer(timer);
  }
};
const BAS_ARCHIVE_SUCCESSFUL_PING_INTERVAL = 4 * 1000;
const BAS_ARCHIVE_FAILED_PING_INTERVAL = 1 * 1000;
const BAS_ARCHIVE_MAX_ALLOWED_PING_INTERVAL = 10 * 1000;
const BAS_ARCHIVE_GET_USERS_LIST_DELAY = 1 * 1000;

class Archives {
  constructor(context) {
    _balsamiq.default.exists(context);

    _balsamiq.default.exists(context.eventCenter);

    _balsamiq.default.exists(context.host);

    this.context = context;
    /*
    archiveID -> {
        archive: Archive,
        rtcChannel: RTCChannel<Timer>,
        rtcChannelName: String,
        isPinging: bool,
        lastPingFailed: bool,
        lastSuccessfulPingTimestamp: Number?,
        rtcIsDown: bool?,
        isOffline: bool?, // true when lastPingFailed || rtcIsDown
        usersList: String,
        openTimestamp: Number?,
        monitorNetworkStatus: Bool,
    }
    */

    this.archives = {};
    this.sqliteIDs = {};
    this.nextArchiveID = 0;
    this.getUsersListPendingCalls = {};
  }

  makeNewArchiveID() {
    var newID = "archive-" + this.nextArchiveID;
    this.nextArchiveID += 1;
    return newID;
  }

  getArchive(archiveID) {
    _balsamiq.default.assertString(archiveID);

    return this.archives[archiveID] ? this.archives[archiveID].archive : null;
  }

  getArchiveOnline(archiveID) {
    _balsamiq.default.assertString(archiveID);

    return this.archives[archiveID] ? !this.archives[archiveID].isOffline : null;
  }

  addExistingArchive(archive) {
    _balsamiq.default.exists(archive);

    _balsamiq.default.assert(_balsamiq.default.conformsTo(archive, ArchiveProtocol.Archive.prototype));

    var newID = this.makeNewArchiveID();
    this.archives[newID] = {
      archive: archive
    };
    archive.coreObjectID = CoreObject.makeArchiveCoreObjectID(newID); // modify the archive object!

    return newID;
  }

  addArchiveInMemory(archiveDump) {
    _balsamiq.default.exists(archiveDump);

    var archive = new _BalsamiqArchiveInMemory.BalsamiqArchiveInMemory({
      host: this.context.host
    }, archiveDump);
    return this.addExistingArchive(archive);
  }

  addArchiveFromBAS(serverURL, kind, platformToken, userInfo, rtcClient) {
    _balsamiq.default.assertString(serverURL);

    _balsamiq.default.assertString(kind);

    _balsamiq.default.assertString(platformToken);

    _balsamiq.default.assertClass(userInfo, Object);

    var archive = new ArchiveClient.BalsamiqArchiveClient(serverURL, kind, platformToken, userInfo, nativeNetworkApi, nativeTimerApi, {
      deadline: _nativeConstants.Timeout.secondsForError * 1000
    });
    let archiveID = this.addExistingArchive(archive); // Check if we have a client configured

    if (rtcClient) {
      _balsamiq.default.assertObject(rtcClient);

      this.archives[archiveID].rtcClient = rtcClient;
    }

    return archiveID;
  }

  addArchiveFromSQLite(sqliteID, userInfo) {
    let sqliteDriver = new _sqlitedriver.SQLiteDriver({
      host: this.context.host
    }, sqliteID);
    var archive = new Archive.BalsamiqArchive(sqliteDriver, userInfo);
    const archiveID = this.addExistingArchive(archive);
    this.sqliteIDs[archiveID] = sqliteID;
    return archiveID;
  }

  _removeArchive(archiveID, callback) {
    var obj = this.archives[archiveID]; // Stop ping

    if (obj.isPinging) {
      this.stopPingArchive(archiveID);
    } // Cancel getUsersList if scheduled


    if (this.getUsersListPendingCalls[archiveID]) {
      this.context.host.clearTimer(this.getUsersListPendingCalls[archiveID]);
      delete this.getUsersListPendingCalls[archiveID];
    } // Stop monitoring network status


    if (obj.monitorNetworkStatus) {
      this.stopMonitorNetworkStatus(archiveID);
    } // TODO check if close() is implemented everywhere
    // Archive should be already closed by project controller
    //obj.archive.close();


    if (this.sqliteDriver && this.sqliteIDs[archiveID]) {
      this.sqliteDriver.close(this.sqliteIDs[archiveID]);
      delete this.sqliteIDs[archiveID];
    }

    delete this.archives[archiveID];

    if (callback) {
      callback();
    }
  }

  removeArchive(archiveID, callback) {
    var obj = this.archives[archiveID]; // Stop RTC

    if (obj.rtcChannel) {
      this.unsubscribeFromArchive(archiveID, () => {
        this._removeArchive(archiveID, callback);
      });
    } else {
      this._removeArchive(archiveID, callback);
    }
  }

  getRtcClientForArchive(archiveID) {
    _balsamiq.default.exists(this.archives[archiveID]);

    var obj = this.archives[archiveID];
    return obj.rtcClient;
  }

  subscribeToArchive(archiveID, openTimestamp) {
    _balsamiq.default.exists(this.archives[archiveID]);

    var obj = this.archives[archiveID];
    var archive = obj.archive; // Retrieve RTC Client saved when archive was added

    const rtcClient = obj.rtcClient;

    if (!rtcClient) {
      return;
    }

    if (obj.rtcChannel) {
      // Already subscribed
      // Unsubscribe?
      let err = new Error();
      let message = `Archive.subscribeToArchive called when already having a rtcChannel open. Archive ${archiveID}, userID:${archive.userID}, rtcChannelName:${archive.rtcChannelName}\n\n`;
      message += `Stacktrace\n${err.stack}\n\n`;
      message += `Logs\n${this.context.host.Logs.getLogs()}`;
      this.context.host.collectDevData("RTC_DOUBLE_SUBSCRIBE", message, null, null);
      return;
    }

    if (_balsamiq.default.isNumber(openTimestamp)) {
      obj.openTimestamp = openTimestamp;
    }

    _balsamiq.default.assertString(archive.userID);

    _balsamiq.default.assertString(archive.rtcChannel); // Only bas can have RTC right now
    // Pass the BalsamiqArchiveClient function to retreieve the auth token


    const getTokenInfo = () => {
      return new Promise((resolve, reject) => {
        archive.getRTCAuthToken(authResult => {
          if (authResult.error) {
            reject(new Error('Unexpected network error: ' + authResult.error + ' ' + JSON.stringify(authResult)));
          } else {
            resolve(authResult);
          }
        });
      });
    }; // Subscribe


    const userInfo = archive.userInfo;
    const channel = rtcClient.subscribe({
      channelName: archive.rtcChannel,
      getTokenInfo,
      onMessage: message => {
        this.onRTCMessage(archiveID, archive.rtcChannel, message);
      },
      onNetworkDown: () => {//this.onRTCNetworkStatus(archiveID, archive.rtcChannel, false);
      },
      onNetworkUp: () => {//this.onRTCNetworkStatus(archiveID, archive.rtcChannel, true);
      },
      presence: {
        onChanged: () => {
          this.onRTCPresenceMessage(archiveID, archive.rtcChannel);
        },
        userState: {
          avatarURL: userInfo ? userInfo.avatarURL : "",
          displayName: userInfo && userInfo.displayName ? userInfo.displayName : "unknown",
          name: userInfo ? userInfo.name : "unknown"
        },
        userUUID: archive.userID
      }
    });
    obj.rtcChannel = channel;
    obj.rtcIsDown = false;
    this.context.host.log(`Archives.subscribeToArchive: Archive ${archiveID}, userID ${archive.userID}, rtcChannelName ${archive.rtcChannel}`);
  }

  unsubscribeFromArchive(archiveID, callback) {
    _balsamiq.default.exists(this.archives[archiveID]);

    const obj = this.archives[archiveID];

    if (obj.rtcChannel) {
      obj.rtcChannel.unsubscribe().unsubscribePromise.then(() => {
        this.context.host.log(`Archives.unsubscribeFromArchive: Archive ${archiveID}, userID ${obj.archive.userID}, rtcChannelName ${obj.archive.rtcChannel}`);
      }).catch(() => {
        this.context.host.log(`Promise exception in Archives.unsubscribeFromArchive: Archive ${archiveID}, userID ${obj.archive.userID}, rtcChannelName ${obj.archive.rtcChannel}`);
      }).finally(() => {
        obj.rtcChannel = null;
        obj.rtcChannelName = null;
        callback();
      });
    } else {
      callback();
    }
  }

  onNetworkStatusChange(isOnline) {
    _balsamiq.default.assertBool(isOnline);

    for (const archiveID in this.archives) {
      const obj = this.archives[archiveID];
      if (!obj.monitorNetworkStatus) return;
      this.context.host.log(`Archives:NetworkStatusChanged: Archive ${archiveID} is ${isOnline ? "online" : "offline"}`);

      this._updateArchiveOffline(archiveID);
    }
  }

  startMonitorNetworkStatus(archiveID) {
    _balsamiq.default.exists(this.archives[archiveID]);

    const obj = this.archives[archiveID];
    obj.monitorNetworkStatus = true;
  }

  stopMonitorNetworkStatus(archiveID) {
    _balsamiq.default.exists(this.archives[archiveID]);

    const obj = this.archives[archiveID];
    obj.monitorNetworkStatus = false;
  }

  startPingArchive(archiveID) {
    _balsamiq.default.exists(this.archives[archiveID]);

    const obj = this.archives[archiveID];

    _balsamiq.default.assert(!obj.isPinging, "Archives: can not start pinging an already pinging Archive");

    _balsamiq.default.assertString(obj.archive.serverURL, "Archives: can not start pinging an Archive without serverURL");

    const serverURL = obj.archive.serverURL + '/health';
    obj.lastPingFailed = false;

    const timerFunc = () => {
      const obj = this.archives[archiveID];

      if (!obj || !obj.isPinging) {
        return;
      }

      let headers;

      if ((0, _version.getBuildNumber)()) {
        headers = {
          'x-balsamiq-build-num': (0, _version.getBuildNumber)()
        };
      }

      this.context.host.HttpRequest.ajax({
        method: "GET",
        url: serverURL,
        headers: headers,
        callback: (state, headers) => {
          const obj = this.archives[archiveID];

          if (!obj) {
            return; // Archive was closed
          }

          const pingFailed = !!state.error;

          if (obj.lastPingFailed != pingFailed) {
            obj.lastPingFailed = pingFailed;
            this.context.host.log("Archives: Archive " + archiveID + " last ping is " + (obj.lastPingFailed ? "failed" : "ok"));

            if (!pingFailed) {
              const now = Date.now();

              if (obj.lastSuccessfulPingTimestamp && now - obj.lastSuccessfulPingTimestamp > BAS_ARCHIVE_MAX_ALLOWED_PING_INTERVAL) {
                // The last successful ping is too old, we suspect that the we may have lost packets
                // because of network issues or because the computer was in stop. Then, we force the
                // Archive to go offline to simulate network down or stop.
                obj.lastPingFailed = true;
                this.context.host.log("Archives: Archive " + archiveID + " last successful ping is too old");
              }

              obj.lastSuccessfulPingTimestamp = now;
            }

            this._updateArchiveOffline(archiveID);
          }

          const timeDiff = Date.now() - state.now;
          this.context.eventCenter.fireEvent(Events.newCoreEvent(obj.archive.coreObjectID, ArchiveProtocol.EventTypes.SERVER_TIMEDIFF, {
            serverTimeDiff: timeDiff
          }));
        }
      });
      this.context.host.timer(obj.lastPingFailed ? BAS_ARCHIVE_FAILED_PING_INTERVAL : BAS_ARCHIVE_SUCCESSFUL_PING_INTERVAL, timerFunc); // else don't set a new timer
    };

    obj.isPinging = true;
    this.context.host.timer(BAS_ARCHIVE_SUCCESSFUL_PING_INTERVAL, timerFunc);
  }

  stopPingArchive(archiveID) {
    _balsamiq.default.exists(this.archives[archiveID]);

    var obj = this.archives[archiveID];
    obj.isPinging = false;
  }

  onRTCMessage(archiveID, channelName, message) {
    _balsamiq.default.assertString(archiveID);

    _balsamiq.default.assertString(channelName);

    _balsamiq.default.exists(message); // Balsamiq.assert(this.archives[archiveID].rtcChannelName == channelName);


    var archive = this.archives[archiveID].archive;

    if (archive.userID === message.author || archive.userID === message.basSessionId) {
      // ignore messages from myself
      return;
    }

    if (_balsamiq.default.isNumber(message.timestamp) && _balsamiq.default.isNumber(this.archives[archiveID].openTimestamp)) {
      if (message.timestamp < this.archives[archiveID].openTimestamp) {
        this.context.host.log("onRTCMessage: ignoring message older than openTimestamp");
        return;
      }
    }

    var fireEvent = (eventName, eventData) => {
      if (message.archiveRevision) {
        eventData = { ...eventData,
          archiveRevision: message.archiveRevision
        };
      }

      this.context.eventCenter.fireEvent(Events.newCoreEvent(archive.coreObjectID, eventName, eventData));
    };

    switch (message.operation) {
      case "setResourceData":
        fireEvent(ArchiveProtocol.EventTypes.RESOURCE_DATA_CHANGED, {
          resourceID: message.resourceID,
          branchID: message.branchID
        });
        break;

      case "setResourceAttributes":
        fireEvent(ArchiveProtocol.EventTypes.RESOURCE_ATTRIBUTES_CHANGED, {
          resourceID: message.resourceID,
          branchID: message.branchID,
          attributes: message.attributes
        });
        break;

      case "setBranchAttributes":
        fireEvent(ArchiveProtocol.EventTypes.BRANCH_ATTRIBUTES_CHANGED, {
          branchID: message.branchID,
          attributes: message.attributes
        });
        break;

      case "createResource":
        fireEvent(ArchiveProtocol.EventTypes.RESOURCE_ADDED, {
          resourceID: message.resourceID,
          branchID: message.branchID,
          attributes: message.attributes
        });
        break;

      case "createBranch":
        fireEvent(ArchiveProtocol.EventTypes.BRANCH_ADDED, {
          branchID: message.branchID,
          attributes: message.attributes
        });
        break;

      case "createThumbnail":
        fireEvent(ArchiveProtocol.EventTypes.THUMBNAIL_ADDED, {
          thumbnailID: message.thumbnailID
        });
        break;

      case "deleteResources":
        fireEvent(ArchiveProtocol.EventTypes.RESOURCES_DELETED, {
          branchID: message.branchID,
          resourceIDs: message.resourceIDs
        });
        break;

      case "deleteThumbnails":
        fireEvent(ArchiveProtocol.EventTypes.THUMBNAILS_DELETED, {
          thumbnailIDs: message.thumbnailIDs
        });
        break;

      case "deleteBranches":
        fireEvent(ArchiveProtocol.EventTypes.BRANCHES_DELETED, {
          branchIDs: message.branchIDs
        });
        break;

      case "setArchiveAttributes":
        fireEvent(ArchiveProtocol.EventTypes.ARCHIVE_ATTRIBUTES_CHANGED, {
          attributes: message.attributes
        });
        break;

      case "createComment":
        fireEvent(ArchiveProtocol.EventTypes.COMMENT_CREATED, {
          commentData: message
        });
        break;

      case "updateCommentsAttributes":
        fireEvent(ArchiveProtocol.EventTypes.COMMENTS_ATTRIBUTES_UPDATED, {
          attributes: message.attributes
        });
        break;

      case "setCommentData":
        fireEvent(ArchiveProtocol.EventTypes.COMMENT_DATA_UPDATED, {
          commentID: message.commentID,
          attributes: message.attributes,
          data: message.data
        });
        break;

      case "deleteComments":
        fireEvent(ArchiveProtocol.EventTypes.COMMENTS_DELETED, {
          commentIDs: message.commentIDs
        });
        break;

      case "setPermalinkData":
      case "createOrUpdateImageLink":
        {
          fireEvent(ArchiveProtocol.EventTypes.PERMALINK_CREATED_OR_UPDATED, {
            resourceID: message.resourceID,
            branchID: message.branchID
          });
          break;
        }

      case "deletePermalinkData":
        fireEvent(ArchiveProtocol.EventTypes.PERMALINK_DELETED, {
          resourceID: message.resourceID,
          branchID: message.branchID
        });
        break;

      case "createMyUser":
        fireEvent(ArchiveProtocol.EventTypes.ARCHIVE_USERS_CHANGED);
        break;

      case "updateMyUser":
        fireEvent(ArchiveProtocol.EventTypes.ARCHIVE_USERS_CHANGED);
        break;

      case "setThumbnail":
        fireEvent(ArchiveProtocol.EventTypes.THUMBNAIL_UPDATED, {
          thumbnailID: message.thumbnailID
        });
        break;

      case "broadcastMessage":
        fireEvent(ArchiveProtocol.EventTypes.BROADCAST_MESSAGE, {
          userName: message.username,
          message: message.message,
          timestamp: message.timestamp
        });
        break;

      case "aboutToRestoreArchive":
        fireEvent(ArchiveProtocol.EventTypes.ARCHIVE_RESTORING, {
          userName: message.username
        });
        break;

      case "archiveRestored":
        fireEvent(ArchiveProtocol.EventTypes.ARCHIVE_RESTORED, {
          userName: message.username
        });
        break;

      case "archiveHasBeenFlushedOnPlatform":
        fireEvent(ArchiveProtocol.EventTypes.ARCHIVE_FLUSHED, {});
        break;

      case "setResourceBranchID":
        fireEvent(ArchiveProtocol.EventTypes.RESOURCE_BRANCH_CHANGED, {
          newBranchID: message.newBranchID,
          oldBranchID: message.oldBranchID,
          resourceID: message.resourceID
        });
        break;

      case "changeSessionRoles":
        {
          const thisSession = message.sessions.find(session => session.token === archive.token);

          if (thisSession) {
            const role = thisSession.role;
            fireEvent(ArchiveProtocol.EventTypes.USER_ROLE_CHANGED, {
              role
            });
          }

          break;
        }
    }
  }

  onRTCPresenceMessage(archiveID, channelName) {
    _balsamiq.default.assertString(channelName); //        Balsamiq.exists(message); // ignored


    const timerFunc = () => {
      if (!this.archives[archiveID]) {
        return; // just in case...
      }

      delete this.getUsersListPendingCalls[archiveID]; // monitor how often we call the expensive archive.getUsersList

      this.context.host.log('Archives: calling getUsersList');
      var archive = this.archives[archiveID].archive;
      archive.getUsersList(null, res => {
        if (res && res.users) {
          // check if the user list changed since the last time and, if so, remember it
          var jsonUsersList = JSON.stringify(res.users);
          var obj = this.archives[archiveID];

          if (obj && obj.usersList != jsonUsersList) {
            obj.usersList = jsonUsersList;
            this.archives[archiveID].usersListArray = JSON.parse(jsonUsersList);
            this.context.eventCenter.fireEvent(Events.newCoreEvent(archive.coreObjectID, ArchiveProtocol.EventTypes.ONLINE_USERS_LIST_CHANGED, {
              usersList: res.users
            }));
          }
        }
      });
    };

    if (!this.getUsersListPendingCalls[archiveID]) {
      this.context.host.log('Archives: schedule getUsersList');
      this.getUsersListPendingCalls[archiveID] = this.context.host.timer(BAS_ARCHIVE_GET_USERS_LIST_DELAY, timerFunc);
    } else {
      this.context.host.log('Archives: ignore getUsersList');
    }
  }

  onRTCNetworkStatus(archiveID, channelName, isUp) {
    _balsamiq.default.assertString(archiveID);

    _balsamiq.default.assertString(channelName);

    _balsamiq.default.assertBool(isUp);

    var obj = this.archives[archiveID];

    if (obj.rtcIsDown != !isUp) {
      obj.rtcIsDown = !isUp;
      this.context.host.log("Archives: Archive " + archiveID + " has RTC " + (obj.rtcIsDown ? "down" : "up"));

      this._updateArchiveOffline(archiveID);
    }
  }

  onAwake() {
    // send an offline event for all BAS archives with ping
    for (var archiveID in this.archives) {
      let obj = this.archives[archiveID];

      if (obj.isPinging) {
        if (!obj.lastPingFailed) {
          this.context.host.log("Archives: sending Archive " + archiveID + " offline after an awake event");
          obj.lastPingFailed = true;

          this._updateArchiveOffline(archiveID);
        }
      }
    }
  }

  _updateArchiveOffline(archiveID) {
    _balsamiq.default.assertString(archiveID);

    let obj = this.archives[archiveID];
    let oldOffline = obj.isOffline;
    obj.isOffline = obj.lastPingFailed || obj.rtcIsDown || !this.context.host.Network.isOnline();

    if (oldOffline != obj.isOffline) {
      if (obj.isPinging || obj.monitorNetworkStatus) {
        this.context.host.log("Archives: Archive " + archiveID + " is now " + (obj.isOffline ? "offline" : "online"));
        this.context.eventCenter.fireEvent(Events.newCoreEvent(obj.archive.coreObjectID, obj.isOffline ? ArchiveProtocol.EventTypes.ARCHIVE_OFFLINE : ArchiveProtocol.EventTypes.ARCHIVE_ONLINE, {}));
      }
    }
  }

}

exports.Archives = Archives;

/***/ }),

/***/ "./js/core/core/clipboard.js":
/*!***********************************!*\
  !*** ./js/core/core/clipboard.js ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.EventTypes = exports.CoreObjectType = exports.Clipboard = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _events = _interopRequireDefault(__webpack_require__(/*! core/events */ "./js/core/events.ts"));

var CoreObject = _interopRequireWildcard(__webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const CoreObjectType = "clipboard";
exports.CoreObjectType = CoreObjectType;
const EventTypes = {
  CLIPBOARD_CHANGED: "clipboardChanged"
};
exports.EventTypes = EventTypes;

class Clipboard extends CoreObject.CoreObject {
  constructor(context, clipboardId) {
    super(context, CoreObject.makeCoreObjectID(CoreObjectType, {
      clipboardId: clipboardId
    }));

    _balsamiq.default.exists(context.host);

    _balsamiq.default.exists(context.host.LocalStorage);

    this.host = context.host;
  }

  clipboardChangedEvent(data) {
    this.context.eventCenter.fireEvent(_events.default.newCoreEvent(this.coreObjectID, EventTypes.CLIPBOARD_CHANGED, data));
  }
  /**
   *
   * @param {string[]} data
   * @param {((res: any) => void)=} callback
   */


  setData(data, callback) {
    this.host.Clipboard.setData(data, (res, errorMessage) => {
      if (res) {
        this.context.eventCenter.fireEvent(_events.default.newCoreEvent(this.coreObjectID, EventTypes.CLIPBOARD_CHANGED, data));
      }

      if (callback) {
        callback(res, errorMessage);
      }
    });
  }

  getData() {
    return this.host.Clipboard.getData();
  }

  hasData() {
    return true; // To make this work we must get data from clipboard every time the menu
    //  (paste, paste in place) should be shown
    // if (this.canGetDataFromSystem()) {
    //     data = this.host.Clipboard.getData();
    //     res = data && data.length > 0;
    // }
  }

  canGetDataFromSystem() {
    return this.host.Clipboard.canGetDataFromSystem();
  }

  canPasteFromMenu() {
    return this.host.Clipboard.canPasteFromMenu();
  }

  canCopyImageToSystem() {
    return this.host.Clipboard.canCopyImageToSystem();
  }

}

exports.Clipboard = Clipboard;

/***/ }),

/***/ "./js/core/core/controllers.js":
/*!*************************************!*\
  !*** ./js/core/core/controllers.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Controllers = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _projectcontroller = __webpack_require__(/*! core/controllers/projectcontroller */ "./js/core/controllers/projectcontroller.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var CoreObject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var DirtyWireframesController = __webpack_require__(/*! core/controllers/dirtywireframescontroller */ "./js/core/controllers/dirtywireframescontroller.js");

var EmptyQueueController = __webpack_require__(/*! core/controllers/emptyqueuecontroller */ "./js/core/controllers/emptyqueuecontroller.js");

class Controllers {
  constructor(context) {
    _balsamiq.default.exists(context.archives);

    _balsamiq.default.exists(context.eventCenter);

    _balsamiq.default.exists(context.getCoreObject);

    _balsamiq.default.exists(context.operationQueue);

    _balsamiq.default.exists(context.host);

    _balsamiq.default.exists(context.settings);

    _balsamiq.default.exists(context.execTime);

    this.context = context;
    this.controllers = {}; // ID -> Controller instance

    this.nextControllerID = 0;
  }

  makeNewID() {
    var newID = "contrlr-" + this.nextControllerID;
    this.nextControllerID += 1;
    return newID;
  }

  makeNewProjectController(archiveID, project, options) {
    options = options || {};
    const pc = new _projectcontroller.ProjectController({
      archives: this.context.archives,
      eventCenter: this.context.eventCenter,
      operationQueue: this.context.operationQueue,
      getCoreObject: this.context.getCoreObject,
      host: this.context.host,
      settings: this.context.settings,
      execTime: this.context.execTime
    }, archiveID, project, {
      permalinkEnabled: options.permalinkEnabled,
      enableNotifications: options.enableNotifications,
      enableChangeTracking: options.enableChangeTracking,
      filteredProjectMembers: options.filteredProjectMembers,
      getIsDirtyCallback: options.getIsDirtyCallback,
      getPngAlreadyExported: options.getPngAlreadyExported,
      refreshBasCfg: options.refreshBasCfg,
      getBasCfg: options.getBasCfg,
      startAsReviewer: options.startAsReviewer
    });
    this.addController(pc);
    return pc;
  }

  makeNewDirtyWireframesController(archiveID, project) {
    var dwc = new DirtyWireframesController.DirtyWireframesController({
      archives: this.context.archives,
      eventCenter: this.context.eventCenter,
      getCoreObject: this.context.getCoreObject
    }, archiveID, project);
    this.addController(dwc);
    return dwc;
  }

  makeNewEmptyQueueController(projectcontroller, project, callback) {
    var eqc = new EmptyQueueController.EmptyQueueController({
      eventCenter: this.context.eventCenter,
      getCoreObject: this.context.getCoreObject
    }, projectcontroller, project, callback);
    this.addController(eqc);
    return eqc;
  }

  removeController(controller) {
    _balsamiq.default.exists(controller);

    _balsamiq.default.exists(this.controllers[controller.coreObjectID.controllerID]);

    this.controllers[controller.coreObjectID.controllerID].cleanUp();
    delete this.controllers[controller.coreObjectID.controllerID];
  }

  addController(controller) {
    _balsamiq.default.exists(controller);

    var coreObjectID = CoreObject.makeControllerCoreObjectID(this.makeNewID());
    controller.coreObjectID = coreObjectID;
    this.controllers[coreObjectID.controllerID] = controller;
    return controller.coreObjectID;
  }

  getController(coreObjectID) {
    _balsamiq.default.existsClass(coreObjectID, CoreObject.CoreObjectID);

    return this.controllers[coreObjectID.controllerID];
  }

  getProjectControllerForProjectID(projectID) {
    _balsamiq.default.assertString(projectID);

    for (var id in this.controllers) {
      var c = this.controllers[id];

      if (c instanceof _projectcontroller.ProjectController) {
        if (c.project && c.project.coreObjectID.projectID === projectID) {
          return c;
        }
      }
    }

    return null;
  }

}

exports.Controllers = Controllers;

/***/ }),

/***/ "./js/core/core/coreobject.js":
/*!************************************!*\
  !*** ./js/core/core/coreobject.js ***!
  \************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.assertVoidType = exports.assertViewModelType = exports.assertViewControllerType = exports.assertType = exports.assertThumbnailType = exports.assertSiteType = exports.assertResourceType = exports.assertProjectType = exports.assertOperationType = exports.assertControlType = exports.assertBranchType = exports.EventTypes = exports.CoreObjectTypes = exports.CoreObjectID = exports.CoreObject = void 0;
exports.createNewUUID = createNewUUID;
exports.makeArchiveCoreObjectID = makeArchiveCoreObjectID;
exports.makeBranchCoreObjectID = makeBranchCoreObjectID;
exports.makeChildControlCoreObjectID = makeChildControlCoreObjectID;
exports.makeControlCoreObjectID = makeControlCoreObjectID;
exports.makeControlProjectCoreObjectID = makeControlProjectCoreObjectID;
exports.makeControlResourceCoreObjectID = makeControlResourceCoreObjectID;
exports.makeControllerCoreObjectID = makeControllerCoreObjectID;
exports.makeCoreObjectID = makeCoreObjectID;
exports.makeCoreObjectIDFromString = makeCoreObjectIDFromString;
exports.makeOperationCoreObjectID = makeOperationCoreObjectID;
exports.makeParentControlCoreObjectID = makeParentControlCoreObjectID;
exports.makeProjectCoreObjectID = makeProjectCoreObjectID;
exports.makeQueueCoreObjectID = makeQueueCoreObjectID;
exports.makeResourceCoreObjectID = makeResourceCoreObjectID;
exports.makeSiteCoreObjectID = makeSiteCoreObjectID;
exports.makeThumbnailCoreObjectID = makeThumbnailCoreObjectID;
exports.makeViewControllerCoreObjectID = makeViewControllerCoreObjectID;
exports.makeViewModelCoreObjectID = makeViewModelCoreObjectID;
exports.makeVoidCoreObjectID = makeVoidCoreObjectID;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Events = _interopRequireWildcard(__webpack_require__(/*! core/events */ "./js/core/events.ts"));

var _uuidJs = _interopRequireDefault(__webpack_require__(/*! uuid-js */ "./node_modules/uuid-js/lib/uuid.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// CoreObject - base class for all Core object which send events
// -------------------------------------------------------------
const EventTypes = {
  KEY_VALUE_CHANGED: "keyValueChanged"
};
exports.EventTypes = EventTypes;

class CoreObject {
  /**
   *
   * @param {*} context
   * @param {CoreObjectID} coreObjectID
   */
  constructor(context, coreObjectID) {
    _balsamiq.default.exists(context, "context must exist");

    _balsamiq.default.exists(context.eventCenter, "context.eventCenter must exist");

    _balsamiq.default.exists(context.getCoreObject);

    _balsamiq.default.exists(coreObjectID, "coreObjectID must exist");

    this.context = context;
    this.coreObjectID = coreObjectID;
    this.bindings = {}; //  { destKeyName -> { sourceCoreObjectID, sourceKeyName, listener } }

    this._isCleanedUp = false;
  }

  init() {// called when added to the Core
    // subclass
  }

  cleanUp() {
    _balsamiq.default.assert(!this._isCleanedUp, `${this.constructor.name} already cleaned up`);

    this._isCleanedUp = true;
    this.context.eventCenter.removeAllListeners(this);
    this.bindings = {}; //        console.log("CoreObject cleaned up: " + this.coreObjectID.toString());
  }

  onContextChange(params) {// called when something in the context changed
    // subclass
  } // Key-value bindings - public methods
  // -----------------------------------

  /**
   * @returns {string[]}
   */


  getKeyNames() {
    // subclass
    return [];
  }
  /**
   * @param {string} keyName
   * @returns {unknown}
   */


  getKeyValue(keyName) {
    // subclass
    _balsamiq.default.assert(false, "getKeyValue with unknown keyName: " + keyName);

    return null;
  }
  /**
   *
   * @param {string} keyName
   * @param {*} value
   */


  setKeyValue(keyName, value) {
    // subclass
    // set the value, then call notifyKeyValueChanged
    _balsamiq.default.assert(false, "setKeyValue with unknown keyName: " + keyName);
  }
  /**
   *
   * @param {string} keyName
   * @param {CoreObjectID} sourceCoreObjectID
   * @param {string} sourceKeyName
   */


  bindValue(keyName, sourceCoreObjectID, sourceKeyName) {
    _balsamiq.default.assertString(keyName);

    _balsamiq.default.existsClass(sourceCoreObjectID, CoreObjectID);

    _balsamiq.default.assertString(sourceKeyName);

    this.unbindValue(keyName);

    _balsamiq.default.assert(!this.bindings[keyName], "can't bind key named '" + keyName + "' because it was already bound");

    var b = {
      sourceCoreObjectID: sourceCoreObjectID,
      sourceKeyName: sourceKeyName,
      listener: function (event) {
        if (event.eventData.keyName === sourceKeyName) {
          this.setKeyValue(keyName, event.eventData.keyValue);
        }
      }
    };
    this.bindings[keyName] = b;
    this.context.eventCenter.addListener(b.sourceCoreObjectID, EventTypes.KEY_VALUE_CHANGED, this, b.listener);
    this.setKeyValue(keyName, this.context.getCoreObject(sourceCoreObjectID).getKeyValue(sourceKeyName));
  }
  /**
   *
   * @param {string} keyName
   */


  unbindValue(keyName) {
    if (this.bindings[keyName]) {
      var b = this.bindings[keyName];
      this.context.eventCenter.removeListener(b.sourceCoreObjectID, EventTypes.KEY_VALUE_CHANGED, this, b.listener);
      delete this.bindings[keyName];
    }
  } // Key-value bindings - private methods
  // ------------------------------------

  /**
   *
   * @param {string} keyName
   */


  notifyKeyValueChanged(keyName) {
    var e = Events.newCoreEvent(this.coreObjectID, EventTypes.KEY_VALUE_CHANGED, {
      keyName: keyName,
      keyValue: this.getKeyValue(keyName)
    });
    this.context.eventCenter.fireEvent(e);
  }

} // CoreObjectID - to identify an object in the Core
// ------------------------------------------------


exports.CoreObject = CoreObject;
const CoreObjectTypes = {
  VOID: "void",
  // ---
  ARCHIVE: "archive",
  // it's not a CoreObject
  // ---
  SITE: "site",
  PROJECT: "project",
  BRANCH: "branch",
  RESOURCE: "resource",
  CONTROL: "control",
  THUMBNAIL: "thumbnail",
  // ---
  VIEWMODEL: "viewModel",
  VIEWCONTROLLER: "viewController",
  // ---
  CONTROLLER: "controller",
  // ---
  QUEUE: "queue",
  OPERATION: "operation"
};
/**
  * All the possible fields defined in CoreObjectID
  * @typedef {Object} CoreObjectIDData
  * @prop {string=} resourceID
  * @prop {string=} branchID
  * @prop {string=} projectID
  * @prop {string=} thumbnailID
  * @prop {number[]=} controlID
  * @prop {string=} commentID
  * @prop {string=} siteID
  * @prop {string=} controllerID
  * @prop {string=} archiveID
  * @prop {string=} viewControllerID
  * @prop {string=} viewModelID
  * @prop {string=} queueID
  * @prop {string=} operationID
 */

exports.CoreObjectTypes = CoreObjectTypes;

class CoreObjectID {
  /**
   *
   * @param {string} type (one of CoreObjectTypes values)
   * @param {CoreObjectIDData} data
   */
  constructor(type, data) {
    this.type = type;
    Object.assign(this, data); // Annotations used just for defining the props.
    // NOTE: this are no-ops that won't affect performances.

    /** @type {string=} */

    this.resourceID;
    /** @type {string=} */

    this.branchID;
    /** @type {string=} */

    this.projectID;
    /** @type {string=} */

    this.thumbnailID;
    /** @type {number[]=} */

    this.controlID;
    /** @type {string=} */

    this.commentID;
    /** @type {string=} */

    this.siteID;
    /** @type {string=} */

    this.controllerID;
    /** @type {string=} */

    this.archiveID;
    /** @type {string=} */

    this.viewControllerID;
    /** @type {string=} */

    this.viewModelID;
    /** @type {string=} */

    this.queueID;
    /** @type {string=} */

    this.operationID;
  }

  clone() {
    var newID = new CoreObjectID(this.type, {});
    Object.assign(newID, this);
    return newID;
  }

  toString() {
    return JSON.stringify(this);
  }
  /**
   *
   * @param {CoreObjectID} anotherCoreObjectID
   */


  isEqual(anotherCoreObjectID) {
    return _balsamiq.default.isEqual(this, anotherCoreObjectID);
  }
  /**
   *
   * @param {object} filter
   */


  matches(filter) {
    _balsamiq.default.exists(filter);

    for (var key in filter) {
      if (
      /*(key !== "type") &&*/
      filter[key] !== this[key]) {
        return false;
      }
    }

    return true;
  }

} // Factory functions
// -----------------

/**
 *
 * @param {string} str
 */


exports.CoreObjectID = CoreObjectID;

function makeCoreObjectIDFromString(str) {
  _balsamiq.default.assertString(str);

  let res = null;

  try {
    let rawObj = JSON.parse(str);

    _balsamiq.default.assertString(rawObj.type, "Invalid type parsing CoreObjectID, type must be a String: " + str.type);

    res = makeCoreObjectID(rawObj.type, rawObj);
  } catch (err) {
    _balsamiq.default.assert(false, "Error parsing CoreObjectID String: " + str);
  }

  return res;
}
/**
 *
 * @param {string} type
 * @param {CoreObjectIDData} data
 */


function makeCoreObjectID(type, data) {
  return new CoreObjectID(type, data);
}

function makeVoidCoreObjectID() {
  return new CoreObjectID(CoreObjectTypes.VOID, {});
}
/**
 *
 * @param {string} siteID
 */


function makeSiteCoreObjectID(siteID) {
  _balsamiq.default.assertString(siteID);

  return new CoreObjectID(CoreObjectTypes.SITE, {
    siteID: siteID
  });
}
/**
 *
 * @param {CoreObjectID} siteCoreObjectID
 * @param {string} projectID
 */


function makeProjectCoreObjectID(siteCoreObjectID, projectID) {
  assertSiteType(siteCoreObjectID);

  _balsamiq.default.assertString(projectID);

  return new CoreObjectID(CoreObjectTypes.PROJECT, {
    siteID: siteCoreObjectID.siteID,
    projectID: projectID
  });
}
/**
 *
 * @param {CoreObjectID} projectCoreObjectID
 * @param {string} branchID
 */


function makeBranchCoreObjectID(projectCoreObjectID, branchID) {
  assertProjectType(projectCoreObjectID);

  _balsamiq.default.assertString(branchID);

  return new CoreObjectID(CoreObjectTypes.BRANCH, {
    siteID: projectCoreObjectID.siteID,
    projectID: projectCoreObjectID.projectID,
    branchID: branchID
  });
}
/**
 *
 * @param {CoreObjectID} projectCoreObjectID
 * @param {string} branchID
 * @param {string} resourceID
 */


function makeResourceCoreObjectID(projectCoreObjectID, branchID, resourceID) {
  assertProjectType(projectCoreObjectID);

  _balsamiq.default.assertString(branchID);

  _balsamiq.default.assertString(resourceID);

  return new CoreObjectID(CoreObjectTypes.RESOURCE, {
    siteID: projectCoreObjectID.siteID,
    projectID: projectCoreObjectID.projectID,
    branchID: branchID,
    resourceID: resourceID
  });
}
/**
 *
 * @param {CoreObjectID} resourceCoreObjectID
 * @param {number[]} controlID
 */


function makeControlCoreObjectID(resourceCoreObjectID, controlID) {
  assertResourceType(resourceCoreObjectID);

  _balsamiq.default.assertArray(controlID);

  return new CoreObjectID(CoreObjectTypes.CONTROL, {
    siteID: resourceCoreObjectID.siteID,
    projectID: resourceCoreObjectID.projectID,
    branchID: resourceCoreObjectID.branchID,
    resourceID: resourceCoreObjectID.resourceID,
    controlID: controlID
  });
}
/**
 *
 * @param {CoreObjectID} controlCoreObjectID
 * @param {number} childControlID
 */


function makeChildControlCoreObjectID(controlCoreObjectID, childControlID) {
  assertControlType(controlCoreObjectID);

  _balsamiq.default.assertNumber(childControlID);

  return new CoreObjectID(CoreObjectTypes.CONTROL, {
    siteID: controlCoreObjectID.siteID,
    projectID: controlCoreObjectID.projectID,
    branchID: controlCoreObjectID.branchID,
    resourceID: controlCoreObjectID.resourceID,
    controlID: controlCoreObjectID.controlID.concat([childControlID])
  });
}
/**
 *
 * @param {CoreObjectID} controlCoreObjectID
 */


function makeParentControlCoreObjectID(controlCoreObjectID) {
  assertControlType(controlCoreObjectID);

  _balsamiq.default.assert(controlCoreObjectID.controlID.length >= 1);

  return new CoreObjectID(CoreObjectTypes.CONTROL, {
    siteID: controlCoreObjectID.siteID,
    projectID: controlCoreObjectID.projectID,
    branchID: controlCoreObjectID.branchID,
    resourceID: controlCoreObjectID.resourceID,
    controlID: controlCoreObjectID.controlID.slice(0, -1)
  });
}
/**
 *
 * @param {CoreObjectID} controlCoreObjectID
 */


function makeControlResourceCoreObjectID(controlCoreObjectID) {
  assertControlType(controlCoreObjectID);
  return new CoreObjectID(CoreObjectTypes.RESOURCE, {
    siteID: controlCoreObjectID.siteID,
    projectID: controlCoreObjectID.projectID,
    branchID: controlCoreObjectID.branchID,
    resourceID: controlCoreObjectID.resourceID
  });
}
/**
 *
 * @param {CoreObjectID} controlCoreObjectID
 */


function makeControlProjectCoreObjectID(controlCoreObjectID) {
  assertControlType(controlCoreObjectID);
  return new CoreObjectID(CoreObjectTypes.PROJECT, {
    siteID: controlCoreObjectID.siteID,
    projectID: controlCoreObjectID.projectID
  });
}
/**
 *
 * @param {string} archiveID
 */


function makeArchiveCoreObjectID(archiveID) {
  _balsamiq.default.assertString(archiveID);

  return new CoreObjectID(CoreObjectTypes.ARCHIVE, {
    archiveID: archiveID
  });
}
/**
 *
 * @param {string} viewModelID
 */


function makeViewModelCoreObjectID(viewModelID) {
  _balsamiq.default.assertString(viewModelID);

  return new CoreObjectID(CoreObjectTypes.VIEWMODEL, {
    viewModelID: viewModelID
  });
}
/**
 *
 * @param {string} viewControllerID
 */


function makeViewControllerCoreObjectID(viewControllerID) {
  _balsamiq.default.assertString(viewControllerID);

  return new CoreObjectID(CoreObjectTypes.VIEWCONTROLLER, {
    viewControllerID: viewControllerID
  });
}
/**
 *
 * @param {CoreObjectID} projectCoreObjectID
 * @param {string} thumbnailID
 */


function makeThumbnailCoreObjectID(projectCoreObjectID, thumbnailID) {
  assertProjectType(projectCoreObjectID);

  _balsamiq.default.assertString(thumbnailID);

  return new CoreObjectID(CoreObjectTypes.RESOURCE, {
    siteID: projectCoreObjectID.siteID,
    projectID: projectCoreObjectID.projectID,
    thumbnailID: thumbnailID
  });
}
/**
 *
 * @param {string} controllerID
 */


function makeControllerCoreObjectID(controllerID) {
  _balsamiq.default.assertString(controllerID);

  return new CoreObjectID(CoreObjectTypes.CONTROLLER, {
    controllerID: controllerID
  });
}
/**
 *
 * @param {string} queueID
 */


function makeQueueCoreObjectID(queueID) {
  _balsamiq.default.assertString(queueID);

  return new CoreObjectID(CoreObjectTypes.QUEUE, {
    queueID: queueID
  });
}
/**
 *
 * @param {CoreObjectID} queueCoreObjectID
 * @param {string} operationID
 */


function makeOperationCoreObjectID(queueCoreObjectID, operationID) {
  _balsamiq.default.assertString(operationID);

  return new CoreObjectID(CoreObjectTypes.OPERATION, {
    queueID: queueCoreObjectID.queueID,
    operationID: operationID
  });
} // UUID
// ----

/**
 * @returns {string}
 */


function createNewUUID() {
  return _uuidJs.default.create().toString().toUpperCase();
} // Type assertions
// ---------------


var assertType = function (coreObjectID, type) {
  _balsamiq.default.exists(coreObjectID);

  _balsamiq.default.assertEnum(type, CoreObjectTypes);

  _balsamiq.default.assert(coreObjectID.type === type, "CoreObjectID expected type: " + type + " but was: " + coreObjectID.type);
};

exports.assertType = assertType;

var assertSiteType = function (coreObjectID) {
  assertType(coreObjectID, CoreObjectTypes.SITE);
};

exports.assertSiteType = assertSiteType;

var assertProjectType = function (coreObjectID) {
  assertType(coreObjectID, CoreObjectTypes.PROJECT);
};

exports.assertProjectType = assertProjectType;

var assertBranchType = function (coreObjectID) {
  assertType(coreObjectID, CoreObjectTypes.BRANCH);
};

exports.assertBranchType = assertBranchType;

var assertResourceType = function (coreObjectID) {
  assertType(coreObjectID, CoreObjectTypes.RESOURCE);
};

exports.assertResourceType = assertResourceType;

var assertControlType = function (coreObjectID) {
  assertType(coreObjectID, CoreObjectTypes.CONTROL);
};

exports.assertControlType = assertControlType;

var assertViewModelType = function (coreObjectID) {
  assertType(coreObjectID, CoreObjectTypes.VIEWMODEL);
};

exports.assertViewModelType = assertViewModelType;

var assertViewControllerType = function (coreObjectID) {
  assertType(coreObjectID, CoreObjectTypes.VIEWCONTROLLER);
};

exports.assertViewControllerType = assertViewControllerType;

var assertThumbnailType = function (coreObjectID) {
  assertType(coreObjectID, CoreObjectTypes.THUMBNAIL);
};

exports.assertThumbnailType = assertThumbnailType;

var assertVoidType = function (coreObjectID) {
  assertType(coreObjectID, CoreObjectTypes.VOID);
};

exports.assertVoidType = assertVoidType;

var assertOperationType = function (coreObjectID) {
  assertType(coreObjectID, CoreObjectTypes.OPERATION);
};

exports.assertOperationType = assertOperationType;

/***/ }),

/***/ "./js/core/core/model.js":
/*!*******************************!*\
  !*** ./js/core/core/model.js ***!
  \*******************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Model = exports.EventTypes = exports.DEFAULT_SITEID = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _operationqueue = __webpack_require__(/*! core/operations/operationqueue */ "./js/core/operations/operationqueue.js");

var DirtyWireframes = _interopRequireWildcard(__webpack_require__(/*! core/model/dirtywireframes */ "./js/core/model/dirtywireframes.js"));

var Permalinks = _interopRequireWildcard(__webpack_require__(/*! core/model/permalinks */ "./js/core/model/permalinks.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var CoreObject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var Site = __webpack_require__(/*! core/model/site */ "./js/core/model/site.js");

var Project = __webpack_require__(/*! core/model/project */ "./js/core/model/project.js");

var Constants = __webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts");

var Events = __webpack_require__(/*! core/events */ "./js/core/events.ts");

var ConnectedUsers = __webpack_require__(/*! core/model/connectedusers */ "./js/core/model/connectedusers.js");

var ArchiveState = __webpack_require__(/*! core/model/archivestate */ "./js/core/model/archivestate.js");

const DEFAULT_SITEID = "site-0";
exports.DEFAULT_SITEID = DEFAULT_SITEID;
const EventTypes = {
  PROJECT_REMOVED: "projectRemoved",
  PROJECT_ADDED: "projectAdded"
};
exports.EventTypes = EventTypes;

class Model {
  constructor(context) {
    _balsamiq.default.exists(context);

    _balsamiq.default.exists(context.eventCenter);

    _balsamiq.default.exists(context.host);

    this.context = context;
    this.sites = {}; // siteID -> Site

    this.projects = {}; // projectID -> Project

    this.nextSiteID = 0;
    this.addSite();
    this.nextProjectID = 0;
    this.operationQueue = new _operationqueue.OperationQueue(this.context, CoreObject.makeQueueCoreObjectID("modelQueue"));
  }

  cleanUp() {
    for (let siteID in this.sites) {
      this.sites[siteID].cleanUp();
    }

    this.sites = {};

    for (let projectID in this.projects) {
      this.projects[projectID].cleanUp();
    }

    this.projects = {};
    this.operationQueue.cleanUp();
    this.operationQueue = null;
  }

  getOperationQueue() {
    return this.operationQueue;
  } // Generic Model Object
  // --------------------


  getModelObject(p_coreObjectID) {
    _balsamiq.default.existsClass(p_coreObjectID, CoreObject.CoreObjectID);

    var project, resource;

    if (p_coreObjectID.type === CoreObject.CoreObjectTypes.SITE) {
      return this.sites[p_coreObjectID.siteID];
    } else if (p_coreObjectID.type === CoreObject.CoreObjectTypes.PROJECT) {
      return this.projects[p_coreObjectID.projectID];
    } else if (p_coreObjectID.type === CoreObject.CoreObjectTypes.CONFIGURATION) {
      project = this.projects[p_coreObjectID.projectID];
      return project.configuration;
    } else if (p_coreObjectID.type === CoreObject.CoreObjectTypes.RESOURCE) {
      project = this.projects[p_coreObjectID.projectID];
      return project ? project.getResource(p_coreObjectID) : null;
    } else if (p_coreObjectID.type === CoreObject.CoreObjectTypes.CONTROL) {
      project = this.projects[p_coreObjectID.projectID];

      if (project) {
        resource = project.getResource(p_coreObjectID);

        if (resource) {
          _balsamiq.default.assert(resource.type === Constants.ResourceTypes.MOCKUP || resource.type === Constants.ResourceTypes.SYMBOLLIBRARY, "this CoreObjectID must point to a Mockup or SymbolLibrary to get a Control");
        }
      }

      return project && resource ? resource.getControl(p_coreObjectID.controlID) : null;
    } else if (p_coreObjectID.type == ConnectedUsers.CONNECTED_USERS_ID_TYPE) {
      project = this.projects[p_coreObjectID.projectID];
      return project ? project.getConnectedUsers() : null;
    } else if (p_coreObjectID.type == ArchiveState.ARCHIVE_STATE_ID_TYPE) {
      project = this.projects[p_coreObjectID.projectID];
      return project ? project.getArchiveState() : null;
    } else if (p_coreObjectID.type == DirtyWireframes.DIRTY_WIREFRAMES_ID_TYPE) {
      project = this.projects[p_coreObjectID.projectID];
      return project ? project.getDirtyWireframes() : null;
    } else if (p_coreObjectID.type == Permalinks.PERMALINKS_ID_TYPE) {
      project = this.projects[p_coreObjectID.projectID];
      return project ? project.getPermalinks() : null;
    }

    return null;
  } // Find a Model object given a generic objectID
  // To find a resource: { resourceID: string, (optional) branchID: string }
  // Return null if nothing is found


  findModelObjectID(genericObjectID) {
    _balsamiq.default.exists(genericObjectID);

    if (genericObjectID.resourceID) {
      // it should be a Resource of a Project
      for (const projectID in this.projects) {
        let project = this.projects[projectID];

        if (project.hasResourceID(genericObjectID.resourceID)) {
          let branchID = genericObjectID.branchID ? genericObjectID.branchID : Constants.Master;
          let resourceCoreObjectID = CoreObject.makeResourceCoreObjectID(project.coreObjectID, branchID, genericObjectID.resourceID);

          if (project.getResource(resourceCoreObjectID)) {
            return resourceCoreObjectID;
          }
        }
      }
    } else if (genericObjectID.shortResourceID || genericObjectID.shortCommentID) {
      // a short ID. return the first one matching
      for (const projectID in this.projects) {
        let project = this.projects[projectID];
        let matchingCoreObjectIDs = project.findByShortID(genericObjectID);

        if (matchingCoreObjectIDs.length > 0) {
          return matchingCoreObjectIDs[0];
        }
      }
    }

    return null;
  } // Sites
  // -----


  addSite() {
    var siteID = "site-" + this.nextSiteID;
    this.nextSiteID += 1;
    var site = new Site.Site(this.context, CoreObject.makeSiteCoreObjectID(siteID));
    this.sites[siteID] = site;
    return site;
  }

  getSite(coreObjectID) {
    _balsamiq.default.existsClass(coreObjectID, CoreObject.CoreObjectID);

    _balsamiq.default.exists(coreObjectID.siteID);

    return this.sites[coreObjectID.siteID];
  } // Projects
  // --------


  getProject(coreObjectID) {
    _balsamiq.default.existsClass(coreObjectID, CoreObject.CoreObjectID);

    _balsamiq.default.exists(coreObjectID.projectID);

    return this.projects[coreObjectID.projectID];
  }

  getProjectCoreObjectIDs() {
    // TODO sort in some stable way
    var projects = [];

    for (var projectID in this.projects) {
      projects.push(this.projects[projectID]);
    }

    var res = [];
    projects.forEach(project => {
      res.push(project.coreObjectID);
    });
    return res;
  }

  addProject(siteID, externalProjectID) {
    var site = this.sites[siteID];
    var projectID = "project-" + this.nextProjectID;
    this.nextProjectID += 1;
    var project = new Project.Project(this.context, CoreObject.makeProjectCoreObjectID(site.coreObjectID, projectID), externalProjectID);
    this.projects[projectID] = project;
    this.context.eventCenter.fireEvent(Events.newCoreEvent(project.coreObjectID, EventTypes.PROJECT_ADDED));
    return project;
  }

  removeProject(projectID) {
    var project = this.projects[projectID];
    project.cleanUp();
    delete this.projects[projectID];
    this.context.eventCenter.fireEvent(Events.newCoreEvent(project.coreObjectID, EventTypes.PROJECT_REMOVED));
  } // Context
  // -------


  onContextChange(params) {
    this.context.host.log('Model: onContextChange');

    for (var projectID in this.projects) {
      this.projects[projectID].onContextChange(params);
    }
  }

}

exports.Model = Model;

/***/ }),

/***/ "./js/core/core/settings.js":
/*!**********************************!*\
  !*** ./js/core/core/settings.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Settings = exports.EventTypes = exports.CoreObjectType = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _events = _interopRequireDefault(__webpack_require__(/*! core/events */ "./js/core/events.ts"));

var CoreObject = _interopRequireWildcard(__webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const CoreObjectType = "settings";
exports.CoreObjectType = CoreObjectType;
const EventTypes = {
  SETTINGS_CHANGED: "settingsChanged",
  SESSION_SETTINGS_CHANGED: "sessionSettingsChanged"
};
exports.EventTypes = EventTypes;

class Settings extends CoreObject.CoreObject {
  constructor(context, settingsId, doneCallback) {
    super(context, CoreObject.makeCoreObjectID(CoreObjectType, {
      settingsId: settingsId
    }));

    _balsamiq.default.exists(context.host);

    _balsamiq.default.exists(context.host.LocalStorage);

    _balsamiq.default.assertFunction(doneCallback);

    this.host = context.host;
    this.settings = {
      timestamps: {}
    };
    this.sessionData = {};
    let localStorageReady = false;
    let sessionStorageReady = false;
    this.host.LocalStorage.getData(data => {
      try {
        if (!data) {
          this.settings = {
            timestamps: {}
          };
          this.host.LocalStorage.setData(JSON.stringify(this.settings), null);
        } else {
          this.settings = JSON.parse(data);
        }
      } catch (e) {
        this.settings = {
          timestamps: {}
        };
        this.host.LocalStorage.setData(JSON.stringify(this.settings), null);
      }

      localStorageReady = true;

      if (sessionStorageReady) {
        doneCallback();
      }
    });
    this.host.SessionStorage.getData(data => {
      try {
        if (!data) {
          this.sessionData = {};
        } else {
          this.sessionData = JSON.parse(data);
        }
      } catch (e) {
        this.sessionData = {};
      }

      sessionStorageReady = true;

      if (localStorageReady) {
        doneCallback();
      }
    });
  }

  settingsChangedEvent(data) {
    this.context.host.log("Settings.settingsChangedEvent");
    var parsedData,
        emptySettings = {
      timestamps: {}
    };

    try {
      data = data || emptySettings;
      parsedData = JSON.parse(data);
    } catch (e) {
      parsedData = emptySettings;
    }

    this.settings = parsedData;
    this.context.eventCenter.fireEvent(_events.default.newCoreEvent(this.coreObjectID, EventTypes.SETTINGS_CHANGED, parsedData));
  }

  setData(key, data) {
    _balsamiq.default.exists(this.settings);

    this.settings[key] = data;
    this.settings.timestamps[key] = Date.now();
    this.prune();
    this.host.LocalStorage.setData(JSON.stringify(this.settings), null);
  }

  getData(key) {
    _balsamiq.default.exists(this.settings);

    return this.settings[key];
  }

  prune() {
    var size = JSON.stringify(this.settings).length;

    if (size < 4194304) {
      return;
    }

    var sortable = [];

    for (var key in this.settings.timestamps) {
      sortable.push([key, this.settings.timestamps[key]]);
    }

    sortable.sort(function (a, b) {
      return a[1] - b[1];
    });
    var remove = sortable.shift();
    delete this.settings[remove[0]];
    delete this.settings.timestamps[remove[0]];
    this.prune();
  }

  setSessionData(key, data) {
    this.sessionData[key] = data;
    this.host.SessionStorage.setData(JSON.stringify(this.sessionData));
  }

  getSessionData(key) {
    return this.sessionData[key];
  }

  sessionSettingsChangedEvent(data) {
    this.context.host.log("Settings.sessionSettingsChangedEvent");
    var parsedData,
        emptySettings = {};

    try {
      data = data || emptySettings;
      parsedData = JSON.parse(data);
    } catch (e) {
      parsedData = emptySettings;
    }

    this.sessionData = parsedData;
    this.context.eventCenter.fireEvent(_events.default.newCoreEvent(this.coreObjectID, EventTypes.SESSION_SETTINGS_CHANGED, parsedData));
  }

}

exports.Settings = Settings;

/***/ }),

/***/ "./js/core/core/viewcontrollers.js":
/*!*****************************************!*\
  !*** ./js/core/core/viewcontrollers.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.makeViewControllersCoreObjectID = exports.isViewControllersCoreObjectID = exports.computeClickData = exports.ViewControllersCoreObjectType = exports.ViewControllers = exports.TouchEventTypes = exports.TouchEventType = exports.MouseEventType = exports.DOUBLECLICK_TIME = exports.CLICK_TIME = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Keyboard = _interopRequireWildcard(__webpack_require__(/*! core/keyboard */ "./js/core/keyboard.js"));

var _constants = __webpack_require__(/*! core/viewcontrollers/constants */ "./js/core/viewcontrollers/constants.js");

var Native = _interopRequireWildcard(__webpack_require__(/*! core/native */ "./js/core/native.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var CoreObject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var AlertVC = __webpack_require__(/*! core/viewcontrollers/alertvc */ "./js/core/viewcontrollers/alertvc.ts");

const ViewControllersCoreObjectType = "VIEWCONTROLLERS";
exports.ViewControllersCoreObjectType = ViewControllersCoreObjectType;
const DOUBLECLICK_TIME = 350;
exports.DOUBLECLICK_TIME = DOUBLECLICK_TIME;
const CLICK_TIME = 250;
exports.CLICK_TIME = CLICK_TIME;
const MouseEventType = {
  MOUSE_UP: "mouseup",
  MOUSE_DOWN: "mousedown",
  MOUSE_MOVE: "mousemove"
};
exports.MouseEventType = MouseEventType;
const TouchEventType = {
  TOUCH_START: "touchstart",
  TOUCH_MOVE: "touchmove",
  TOUCH_END: "touchend"
};
exports.TouchEventType = TouchEventType;
const TouchEventTypes = [TouchEventType.TOUCH_START, TouchEventType.TOUCH_MOVE, TouchEventType.TOUCH_END];
exports.TouchEventTypes = TouchEventTypes;

const makeViewControllersCoreObjectID = function () {
  return new CoreObject.CoreObjectID(ViewControllersCoreObjectType, {});
};

exports.makeViewControllersCoreObjectID = makeViewControllersCoreObjectID;

const isViewControllersCoreObjectID = function (coreObjectID) {
  _balsamiq.default.existsClass(coreObjectID, CoreObject.CoreObjectID);

  return coreObjectID.type == ViewControllersCoreObjectType;
};

exports.isViewControllersCoreObjectID = isViewControllersCoreObjectID;

const computeClickData = function (id, _ref, lastMouseDownData, penultimateMouseDownData) {
  let {
    isDoubleTap,
    numberOfFingers,
    type,
    time,
    clickCount,
    x,
    y
  } = _ref;

  if (isDoubleTap) {
    return {
      isDoubleClick: true,
      isClick: false
    };
  }

  const isTouchEvent = TouchEventTypes.includes(type);

  if (isTouchEvent && numberOfFingers > 1) {
    return {
      isDoubleClick: false,
      isClick: false
    };
  }

  if (type === MouseEventType.MOUSE_DOWN || type === TouchEventType.TOUCH_START) {
    penultimateMouseDownData.id = lastMouseDownData.id;
    penultimateMouseDownData.time = lastMouseDownData.time;
    penultimateMouseDownData.x = lastMouseDownData.x;
    penultimateMouseDownData.y = lastMouseDownData.y;
    lastMouseDownData.id = id;
    lastMouseDownData.time = time;
    lastMouseDownData.x = x;
    lastMouseDownData.y = y;
  }

  let isClick = false;
  let isDoubleClick = false;
  let inDoubleClickRange = false;

  if (type === MouseEventType.MOUSE_UP || type === TouchEventType.TOUCH_END) {
    inDoubleClickRange = Math.abs(x - penultimateMouseDownData.x) < 4 && Math.abs(y - penultimateMouseDownData.y) < 4;
  }

  if ((type === MouseEventType.MOUSE_UP || type === TouchEventType.TOUCH_END) && clickCount === 2 && inDoubleClickRange) {
    isDoubleClick = true;
  } else if (!clickCount && (type === MouseEventType.MOUSE_UP || type === TouchEventType.TOUCH_END) && inDoubleClickRange) {
    if (penultimateMouseDownData.id == id && time - penultimateMouseDownData.time < DOUBLECLICK_TIME) {
      isDoubleClick = true;
    }
  }

  if ([MouseEventType.MOUSE_UP, MouseEventType.MOUSE_MOVE, TouchEventType.TOUCH_MOVE, TouchEventType.TOUCH_END].includes(type)) {
    isClick = time - lastMouseDownData.time < CLICK_TIME;
  }

  return {
    isDoubleClick,
    isClick
  };
};

exports.computeClickData = computeClickData;

class ViewControllers extends CoreObject.CoreObject {
  constructor(context, coreObjectID) {
    _balsamiq.default.exists(context);

    _balsamiq.default.exists(context.host);

    _balsamiq.default.exists(context.model);

    _balsamiq.default.exists(context.eventCenter);

    _balsamiq.default.exists(context.runTask);

    _balsamiq.default.exists(context.runTaskIfPossible);

    _balsamiq.default.exists(context.runningTask);

    _balsamiq.default.exists(context.taskAction);

    _balsamiq.default.exists(context.controllers);

    _balsamiq.default.exists(context.clipboard);

    _balsamiq.default.exists(context.textMeter);

    _balsamiq.default.exists(context.execTime);

    _balsamiq.default.exists(context.coreID);

    _balsamiq.default.assert(isViewControllersCoreObjectID(coreObjectID));

    super(context, coreObjectID);
    this.taskAction = context.taskAction;
    this.host = context.host;
    this.vcContructors = {}; // types -> ViewController Constructor

    this.vcReservedKeystrokes = [];
    this.vcActionTypes = {}; // types -> ViewController ActionTypes object

    this.vcs = {}; // nativeID -> ViewController

    this.modifierKeys = {
      altKey: false,
      shiftKey: false,
      metaKey: false,
      ctrlKey: false
    };
    this.vcContext = {
      viewControllers: this,
      eventCenter: context.eventCenter,
      model: context.model,
      getCoreObject: context.getCoreObject,
      settings: context.settings,
      host: this.host,
      runTask: context.runTask,
      runTaskIfPossible: context.runTaskIfPossible,
      runMainTask: context.runMainTask,
      clipboard: context.clipboard,
      controllers: context.controllers,
      textMeter: context.textMeter,
      execTime: context.execTime,
      coreID: context.coreID,
      getModifierKeys: () => this.modifierKeys
    };
    this.vcsWaitingForPop = new Set();
    this.vcsWaitingForPush = new Set();
    this.vcsToRemoveOnPop = new Set(); // ViewControllerID -> true

    this.vcsToPushState = [];
    this.vcsToSendCommand = [];
    this.vcsStateChangedCallbacks = new Set();
    this.pushedCount = 0;
    this.lastMouseDownData = {
      id: 0,
      time: 0,
      x: 0,
      y: 0
    };
    this.penultimateMouseDownData = {
      id: 0,
      time: 0,
      x: 0,
      y: 0
    };
    this.applicationActionHandlers = {}; // action:string -> callback(action, params):function
  }

  register(viewControllerMod) {
    if (!this.vcContructors[viewControllerMod.Type]) {
      this.vcContructors[viewControllerMod.Type] = viewControllerMod.ViewController;

      if (viewControllerMod.KeyboardShortcuts) {
        this.vcReservedKeystrokes = this.vcReservedKeystrokes.concat(Object.values(viewControllerMod.KeyboardShortcuts));
      }

      this.vcActionTypes[viewControllerMod.Type] = viewControllerMod.ActionTypes || {};
    }
  }

  isViewControllerTypeRegistered(viewControllerType) {
    return !!this.vcContructors[viewControllerType];
  }

  getViewController(coreObjectID) {
    if (coreObjectID.type === CoreObject.CoreObjectTypes.VIEWCONTROLLER) {
      return this.vcs[coreObjectID.viewControllerID];
    }

    return null;
  }

  make(type, taskID, options, callback) {
    // callback(ViewController instance)
    this.checkViewControllerType(type);

    _balsamiq.default.assertString(taskID, "every ViewController must belong to a Task");

    _balsamiq.default.assertFunction(callback);

    options = options ? options : {};
    var vc = new this.vcContructors[type](this.vcContext, type, options);
    this.host.UI.make(type, options, vc.getState(), obj => {
      vc.coreObjectID = CoreObject.makeViewControllerCoreObjectID(obj.id);
      vc.setNativeID(obj.id);
      vc.setTaskID(taskID);
      this.vcs[obj.id] = vc;
      this.host.log(`made VC of type ${type} with id ${obj.id}`);
      callback(vc);
    });
  }

  makeFromNative(type, id, parentID, options) {
    // return ViewController instance
    this.checkViewControllerType(type);
    var vc = new this.vcContructors[type](this.vcContext, null, options);
    vc.setNativeID(id); // TODO no taskID?

    this.vcs[id] = vc;
    this.pushState(id, vc.getState());
    return vc;
  }

  push(windowID, viewControllerID, options, removeOnPop, callback) {
    this.checkViewControllerID(viewControllerID);
    callback = callback ? callback : () => {};

    _balsamiq.default.assertFunction(callback);

    options = options ? options : {};

    if (removeOnPop) {
      this.vcsToRemoveOnPop.add(viewControllerID);
    }

    this.vcsWaitingForPush.add(viewControllerID);
    this.host.log("begin push nativeID: " + viewControllerID); // set the windowID before the actual push, to let actions flow

    let vc = this.vcs[viewControllerID];
    vc.setWindowID(windowID);
    this.host.UI.push(windowID, viewControllerID, options, obj => {
      this.host.log("end push nativeID: " + viewControllerID);

      if (!this.vcs[viewControllerID]) {
        this.host.log("vc popped before push callback was called: nativeID: " + viewControllerID);
      }

      this.vcsWaitingForPush.delete(viewControllerID);
      this.vcsStateChangedCallbacks.forEach(cb => cb(null, null, viewControllerID));
      callback();
    });
  }

  pop(windowID, viewControllerID, callback) {
    _balsamiq.default.assertString(windowID);

    this.checkViewControllerID(viewControllerID);
    callback = callback ? callback : () => {};

    _balsamiq.default.assertFunction(callback);

    this.host.log("begin pop nativeID: " + viewControllerID);
    this.vcsWaitingForPop.add(viewControllerID);
    this.host.UI.pop(windowID, obj => {
      _balsamiq.default.assert(viewControllerID == obj.id, "ViewControllers.pop: asked to pop ID '" + viewControllerID + "', but on top there was '" + obj.id + "'");

      this.host.log("end pop nativeID: " + obj.id);

      if (this.vcsToRemoveOnPop.has(obj.id)) {
        this.vcsToRemoveOnPop.delete(obj.id);
        this.remove(obj.id);
      }

      this.vcsWaitingForPop.delete(viewControllerID);
      callback();
    });
  }

  getTop(windowID) {
    _balsamiq.default.assertString(windowID);

    return this.host.UI.getTop(windowID);
  }

  setChild(parentID, childName, childID, callback) {
    this.checkViewControllerID(parentID);

    if (childID != null) {
      this.checkViewControllerID(childID);
    }

    callback = callback ? callback : () => {};

    _balsamiq.default.assertFunction(callback);

    var parent = this.vcs[parentID];
    var oldChild = parent.getChild(childName);

    if (oldChild) {
      oldChild.setParentNativeID(null);
      oldChild.setWindowID(null);
    }

    if (childID) {
      var child = this.vcs[childID];
      parent.setChild(childName, child);
      child.setParentNativeID(parent.getNativeID());
      child.setWindowID(parent.getWindowID());
    }

    this.host.UI.setChild(parentID, childName, childID, () => {
      this.vcsStateChangedCallbacks.forEach(cb => cb(null, null, null, null, childID));
      callback();
    });
  }

  remove(id, callback) {
    this.checkViewControllerID(id);
    callback = callback ? callback : () => {};

    _balsamiq.default.assertFunction(callback);

    var removeFun = () => {
      //console.log("remove nativeID: " + id);
      this.vcs[id].cleanUp();
      delete this.vcs[id];
      let index = this.vcsToPushState.indexOf(id);

      if (index >= 0) {
        //console.log("removing vcsToPushState: " + id);
        this.vcsToPushState.splice(index, 1);
      }

      this.host.UI.remove(id, () => {
        callback();
      });
    };

    var vc = this.vcs[id];
    var parentVCNativeID = vc.getParentNativeID();

    if (parentVCNativeID) {
      var parentVC = this.vcs[parentVCNativeID];

      if (parentVC) {
        // ParentVC could have already been removed
        var childName = parentVC.getChildName(vc);

        _balsamiq.default.assertString(childName);

        this.setChild(parentVCNativeID, childName, null, () => {
          removeFun();
        });
      } else {
        removeFun();
      }
    } else {
      removeFun();
    }
  }

  pushAlert(_ref2) {
    let {
      windowID,
      taskID,
      message,
      title,
      buttons,
      callback,
      closable
    } = _ref2;

    _balsamiq.default.assertString(windowID);

    _balsamiq.default.assertString(taskID);

    _balsamiq.default.assertString(message);

    callback = callback ? callback : () => {};

    _balsamiq.default.assertFunction(callback);

    _balsamiq.default.assertBool(closable);

    let alertVC;
    this.make(AlertVC.Type, taskID, {
      initialState: {
        [AlertVC.States.MESSAGE]: TextUtils.stringToNodes(message),
        [AlertVC.States.CLOSABLE]: closable,
        [AlertVC.States.BUTTONS]: buttons || null,
        [AlertVC.States.TITLE]: title || null
      }
    }, _alertVC => {
      alertVC = _alertVC;
      alertVC.setTaskID(taskID);
      alertVC.setCallback(params => {
        this.pop(windowID, alertVC.getNativeID());
        callback(params);
      });
      this.push(windowID, alertVC.getNativeID(), {}, true);
    });

    const closeAlert = () => this.pop(windowID, alertVC.getNativeID());

    return {
      close: closeAlert
    };
  } // Private methods
  // ---------------


  installApplicationActionHandler(action, callback) {
    this.applicationActionHandlers[action] = callback;
  }

  execApplicationAction(action, params) {
    const callback = this.applicationActionHandlers[action];

    if (callback) {
      callback(params);
      return true;
    }

    return false;
  }

  onNativeAction(id, action, params) {
    this.checkViewControllerID(id);
    const vc = this.vcs[id];
    const windowID = vc.getWindowID();
    let vcType = vc.getVCType();
    let isInternalAction = this.vcActionTypes[vcType][action] === _constants.ViewControllerActionTypes.Internal; // force the dispatch for vcs without parent, like the menuBar in the Desktop apps.

    let forceDispatch = isInternalAction || !vc.getParentNativeID();

    if (!forceDispatch && (this.vcsWaitingForPop.has(id) || this.vcsWaitingForPush.size != 0)) {
      this.host.log(`Ignoring action ${action} for VC ${id} during VC pop or push`);
      return; // ASYNC: Don't accept actions during VC pop or push
    }

    if (!forceDispatch && this.isRunningTaskBusy(id)) {
      // the Task is busy, then ignore this action (the UI should not send actions when the Task is busy)
      this.host.log('ViewControllers.onNativeAction: ignoring action ' + action + ' for VC ' + id + ' because the running Task: ' + this.context.runningTask(windowID).name() + ' is busy');
      return;
    } // check if this VC is in the topmost layer of VCs


    var layerRootVCID = id;

    while (this.vcs[layerRootVCID].getParentNativeID()) {
      layerRootVCID = this.vcs[layerRootVCID].getParentNativeID();
    }

    if (!forceDispatch && windowID && layerRootVCID != this.host.UI.getTop(windowID)) {
      // the action comes from a VC NOT in the topmost layer of VCs.
      // we don't allow this. don't process the action
      this.host.log('ViewControllers.onNativeAction: ignoring action ' + action + ' for VC ' + id + ' because this VC is not in the topmost layer of VCs');
      return;
    }

    if (!windowID) {
      this.host.log('ViewControllers.onNativeAction: ignoring action ' + action + ' for VC ' + id + ' because the VC has no windowID');
      return;
    }

    let runningTaskID = this.context.runningTask(windowID).getTaskID();
    let actionHandled = this.taskAction(runningTaskID, action, params);
    let vcTaskID = vc.getTaskID();

    if (!actionHandled && runningTaskID !== vcTaskID) {
      actionHandled = this.taskAction(vcTaskID, action, params);
    }

    var id2 = id;

    while (id2 && !actionHandled) {
      actionHandled = this.vcs[id2].action(action, params);
      id2 = this.vcs[id2].getParentNativeID();
    }

    if (actionHandled) {
      return;
    }

    if (this.execApplicationAction(action, { ...params,
      action,
      vcID: id,
      windowID
    })) {
      return;
    }
  }

  onNativeMouseEvent(id, event) {
    this.checkViewControllerID(id);
    this.updateModifierKeys(event);

    if (this.vcsWaitingForPop.has(id) || this.vcsWaitingForPush.size != 0) {
      this.host.log(`Ignoring mouse event for VC ${id} during VC pop or push`);
      return;
    }

    const windowID = this.vcs[id].getWindowID();

    if (!windowID) {
      // the VC has no windowID, then ignore this mouse event (it is probably an old event, still around)
      this.host.log('ViewControllers.onNativeMouseEvent: ignoring mouse event for VC ' + id + ' because the VC has no windowID');
      return {};
    }

    if (this.isRunningTaskBusy(id)) {
      // the Task is busy, then ignore this mouse events (the UI should not send events when the Task is busy)
      this.host.log('ViewControllers.onNativeMouseEvent: ignoring mouse event for VC ' + id + ' because the running Task is busy');
      return;
    }

    const task = this.context.runningTask(windowID);

    _balsamiq.default.assert(task);

    event.clickData = computeClickData(id, event, this.lastMouseDownData, this.penultimateMouseDownData);

    if (event.button === 0 && event.ctrlKey && Native.Host.runtime.isMac()) {
      event.button = 2;
    }

    let res = task.mouseEvent(event);

    if (!res) {
      // Task did not consume event
      if (task.getTaskID() == this.vcs[id].getTaskID() || task.canInterrupt()) {
        // Forward to VC if it belongs to the running task, or
        // if the task can be interrupted
        res = this.vcs[id].mouseEvent(event);
      }
    }

    return res;
  }

  onNativeKeyboardEvent(id, event) {
    this.checkViewControllerID(id);

    _balsamiq.default.assertString(event.type);

    this.updateModifierKeys(event);

    if (this.vcsWaitingForPop.has(id) || this.vcsWaitingForPush.size != 0) {
      this.host.log(`ViewControllers.onNativeKeyboardEvent: ignoring keyboard event for VC ${id} during VC pop or push`);
      return;
    }

    if (this.isRunningTaskBusy(id)) {
      // the Task is busy, then ignore this keyboard event (the UI should not send events when the Task is busy)
      this.host.log('ViewControllers.onNativeKeyboardEvent: ignoring keyboard event for VC ' + id + ' because the running Task is busy');
      return;
    }

    event = _balsamiq.default.extend({
      stringCode: Keyboard.keystrokeToString(event)
    }, event); // dispatch the event to the designed VC

    const windowID = this.vcs[id].getWindowID();

    if (!windowID) {
      // the VC has no windowID, then ignore this event (it is probably an old event, still around)
      this.host.log('ViewControllers.onNativeKeyboardEvent: ignoring keyboard event for VC ' + id + ' because the VC has no windowID');
      return {};
    }

    let task = this.context.runningTask(windowID);

    _balsamiq.default.exists(task);

    var res = task.keyboardEvent(event, true);

    if (!res) {
      // Task did not consume event
      if (task.getTaskID() == this.vcs[id].getTaskID() || task.canInterrupt()) {
        // Forward to VC if it belongs to the running task, or
        // if the task can be interrupted
        res = this.vcs[id].keyboardEvent(event, true);
      }
    } // if not handled, then dispatch the event to the main VC but with hasFocus=false


    if (!res) {
      var parentID = id;

      while (this.vcs[parentID].getParentNativeID()) {
        parentID = this.vcs[parentID].getParentNativeID();
      }

      res = this.vcs[parentID].keyboardEvent(event, false);
    }

    if (!res) {
      res = this.isReservedKeyboardShortcut(event); // this.host.log("Reserved keyboard shortcut: " + JSON.stringify(event));
    } // if not handled, log it


    if (!res) {
      if (event.type !== 'keyup') {// keyup events are not interesting for log
        // console.log("onNativeKeyboardEvent: unhandled keyboard event: " + JSON.stringify(event) + " for VC: " + id);
      }
    }

    return res;
  }

  updateModifierKeys(event) {
    _balsamiq.default.assertObject(event); // the following line are optional, because not all events may contain modifiers


    if (_balsamiq.default.isBoolean(event.altKey)) {
      this.modifierKeys.altKey = event.altKey;
    }

    if (_balsamiq.default.isBoolean(event.shiftKey)) {
      this.modifierKeys.shiftKey = event.shiftKey;
    }

    if (_balsamiq.default.isBoolean(event.metaKey)) {
      this.modifierKeys.metaKey = event.metaKey;
    }

    if (_balsamiq.default.isBoolean(event.ctrlKey)) {
      this.modifierKeys.ctrlKey = event.ctrlKey;
    }
  }

  isReservedKeyboardShortcut(event) {
    var s = Keyboard.keystrokeToString(event);
    return this.vcReservedKeystrokes.indexOf(s) >= 0;
  }

  schedulePushState(id) {
    this.checkViewControllerID(id); // if scheduling the first push, add the callback

    if (this.vcsToPushState.length === 0) {
      this.host.asyncCallback(function () {
        // this.context.host.log(`ViewControllers: pushing states`);
        var vcs = this.vcsToPushState;
        this.vcsToPushState = [];
        let states = vcs.map(id => this.vcs[id].getState());
        let changedStateKeys = vcs.map(id => this.vcs[id].getChangedStateKeys());
        this.host.UI.setStates(vcs, states, changedStateKeys);
        vcs.forEach(id => this.vcs[id].onPushedState());
        this.vcsStateChangedCallbacks.forEach(cb => cb(vcs, this.vcsWaitingForPush, null));
        this.pushedCount += vcs.length; // this.context.host.log(`ViewControllers: pushedCount = ${this.pushedCount}, +${vcs.length}, ids: ${vcs.join(", ")}`);
      }.bind(this));
    }

    if (this.vcsToPushState.indexOf(id) === -1) {
      this.vcsToPushState.push(id);
    }
  }

  sendCommands(id) {
    this.checkViewControllerID(id);
    this.host.UI.sendCommands([id], [this.vcs[id].popCommands()]);
    this.vcsStateChangedCallbacks.forEach(cb => cb(null, null, null, id));
  }

  pushState(id, state) {
    this.checkViewControllerID(id); // if (false) {
    //     // activate this for debug only
    //     this.host.log("Debug: Checking state variables for function members!");
    //     Balsamiq.assertContainsFunctionDeep(state, "State for Viewcontroller " + this.vcs[id].getName() + " must NOT contain functions!");
    // }

    this.host.UI.setState(id, state);
  }

  checkViewControllerID(id) {
    _balsamiq.default.assert(this.vcs[id], "unknown ViewController with ID: " + id);
  }

  checkViewControllerType(type) {
    _balsamiq.default.assert(this.vcContructors[type], "unknown ViewController type: " + type);
  }

  isRunningTaskBusy(id) {
    let windowID = this.vcs[id].getWindowID();

    if (!windowID) {
      this.context.host.log('ViewControllers: VC ' + id + ' has no windowID');
      return false;
    }

    let task = this.context.runningTask(windowID);

    if (task && task.canInterrupt()) {
      return false; // if the Task can be interrupted, then it's not busy
    }

    let topVCID = this.context.host.UI.getTop(windowID);
    let topVC = this.vcs[topVCID];

    if (topVC && task) {
      if (topVC.getTaskID() == task.getTaskID() || task.getVCID() == id) {
        // if the task wants events from this vc all is fine.
        return false; // if the Task has pushed its own VC, then it's not busy
      }
    }

    return true; // the Task is processing, then it's busy
  }

  removeVCStateChangedCallback(callback) {
    _balsamiq.default.assertFunction(callback);

    this.vcsStateChangedCallbacks.delete(callback);
  }

  addVCStateChangedCallback(callback) {
    _balsamiq.default.assertFunction(callback);

    this.vcsStateChangedCallbacks.add(callback);
  }

}

exports.ViewControllers = ViewControllers;

/***/ }),

/***/ "./js/core/core/viewmodels.js":
/*!************************************!*\
  !*** ./js/core/core/viewmodels.js ***!
  \************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = exports.ViewModels = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var CoreObject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var ViewModel = __webpack_require__(/*! core/viewmodels/viewmodel */ "./js/core/viewmodels/viewmodel.ts");

class ViewModels {
  constructor(context) {
    this.viewModels = {}; // ID -> ViewModel instance

    this.nextViewModelID = 0;
  }

  makeNewID() {
    var newID = "vm-" + this.nextViewModelID;
    this.nextViewModelID += 1;
    return newID;
  }

  addViewModel(viewModel) {
    _balsamiq.default.assertClass(viewModel, ViewModel.ViewModel);

    var coreObjectID = CoreObject.makeViewModelCoreObjectID(this.makeNewID());
    viewModel.coreObjectID = coreObjectID;
    this.viewModels[coreObjectID.viewModelID] = viewModel;
    viewModel.init();
    return viewModel.coreObjectID;
  }

  getViewModel(coreObjectID) {
    _balsamiq.default.existsClass(coreObjectID, CoreObject.CoreObjectID);

    return this.viewModels[coreObjectID.viewModelID];
  }

  removeViewModel(coreObjectID) {
    CoreObject.assertViewModelType(coreObjectID);
    this.viewModels[coreObjectID.viewModelID].cleanUp();
    delete this.viewModels[coreObjectID.viewModelID];
  }

}

exports.ViewModels = ViewModels;
var _default = ViewModels;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/cursors.js":
/*!****************************!*\
  !*** ./js/core/cursors.js ***!
  \****************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.CursorIDs = void 0;
const CursorIDs = {
  DEFAULT: 'default',
  RESIZE_N_S: 'northSouthResize',
  RESIZE_E_W: 'eastWestResize',
  RESIZE_NE_SW: 'neswResize',
  RESIZE_NW_SE: 'nwseResize',
  MOVE: 'move',
  RESIZE_COLUMN: 'columnResize',
  RESIZE_ROW: 'rowResize',
  FINGER: 'pointer',
  HAND_OPEN: 'grab',
  HAND_CLOSED: 'grabbing',
  POINTER_MOVE: 'pointerMove',
  POINTER_OVER: 'pointerOver',
  POINTER_ADD: 'pointerAdd',
  POINTER_REMOVE: 'pointerRemove',
  POINTER_TOGGLE: 'pointerToggle',
  HAND_OPEN_CROP_TARGET: 'handOpenImage',
  HAND_CLOSED_CROP_TARGET: 'handClosedImage',
  HAND_OPEN_CROP_AREA: 'handOpenCrop',
  HAND_CLOSED_CROP_AREA: 'handClosedCrop',
  ADD_ARROW: 'addArrow',
  ADD_IMAGE: 'addImage',
  ADD_BLOCK_OF_TEXT: 'addBlockOfText',
  ADD_LINE_OF_TEXT: 'addLineOfText',
  ADD_RECTANGLE: 'addRectangle',
  UNLOCK: 'unlock',
  CLONE: 'clone'
};
exports.CursorIDs = CursorIDs;

/***/ }),

/***/ "./js/core/displayitem.js":
/*!********************************!*\
  !*** ./js/core/displayitem.js ***!
  \********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/*

 Conventions for Display Items
 -----------------------------

 bounds: types.newRectangle
 points (center, start, end): types.newPoint
 colors (fillColor, borderColor): integer

 */
var displayitemMod = {};
displayitemMod.LINK_ICON_SIZE = 16;
displayitemMod.LINK_ICON_X_OFFSET = 4;
displayitemMod.LINK_ICON_Y_OFFSET = 2;
displayitemMod.LineCap = {
  BUTT: "butt",
  ROUND: "round",
  SQUARE: "square"
};
displayitemMod.Type = {
  SVG_PATH: "SVG_PATH",
  RECTANGLE: "Rectangle",
  LINE: "Line",
  IMAGE: "Image",
  SVG_PATHS_ASSET: "SVG_PATHS_ASSET",
  TEXT: "Text",
  NATIVE_ASSET: "NativeAsset"
};
displayitemMod.GradientType = {
  VERTICAL: "vertical"
};
/**
 * Base factory for all display item objects
 * @param {!string} p_type
 * @param {!number} p_x
 * @param {!number} p_y
 * @param {!number} p_width
 * @param {!number} p_height
 * @param {Object} p_attributes
 * @returns {Object}
 */

displayitemMod.newDisplayItem = function (p_type, p_x, p_y, p_width, p_height, p_attributes) {
  _balsamiq.default.assertEnum(p_type, displayitemMod.Type);

  _balsamiq.default.assertNumber(p_x);

  _balsamiq.default.assertNumber(p_y);

  _balsamiq.default.assertNumber(p_width);

  _balsamiq.default.assertNumber(p_height);

  var r = {
    displayItemType: p_type,
    x: p_x,
    // x position of the item in parent's coordinates
    y: p_y,
    // y position of the item in parent's coordinates
    width: p_width,
    // natural width of the item (before multiplying by scaleX)
    height: p_height,
    // natural height of the item (before multiplying by scaleY)
    scaleX: 1,
    // horizontal scale (can be negative)
    scaleY: 1,
    // vertical scale (can be negative)
    rotation: 0,
    // counter-clockwise, degrees, around (x, y)
    masks: [],
    // mask ids to apply to this item
    overriddenColors: null,
    // color transform id to apply to this item
    alpha: 1,
    // opacity [0..1]
    fillColor: null // number or null

  };
  r = _balsamiq.default.extend(r, p_attributes);
  return r;
};

displayitemMod.newSVGPath = function (p_bounds, p_path, p_borderWidth, p_borderColor, p_fillColor, p_dashArray, p_lineCap) {
  var r = displayitemMod.newDisplayItem(displayitemMod.Type.SVG_PATH, p_bounds.x, p_bounds.y, p_path.width, p_path.height, {
    scaleX: p_bounds.width / p_path.width,
    scaleY: p_bounds.height / p_path.height,
    borderWidth: p_borderWidth,
    pathData: p_path.data,
    dashArray: p_dashArray || null,
    linecap: p_lineCap || displayitemMod.LineCap.BUTT,
    borderAlpha: null
  });

  if (_balsamiq.default.isNumber(p_borderColor)) {
    r.borderColor = p_borderColor;
  }

  if (_balsamiq.default.isNumber(p_fillColor)) {
    r.fillColor = p_fillColor;
  }

  return r;
};

displayitemMod.newRectangle = function (p_borderColor, p_borderWidth, p_fillColor, p_cornerRadius, p_dashArray, p_lineCap, p_bounds) {
  var r = displayitemMod.newDisplayItem(displayitemMod.Type.RECTANGLE, p_bounds.x, p_bounds.y, p_bounds.width, p_bounds.height, {
    borderWidth: p_borderWidth,
    dashArray: p_dashArray,
    lineCap: p_lineCap,
    masks: []
  });

  if (p_borderColor !== null) {
    r.borderColor = p_borderColor;
  }

  if (p_fillColor !== null) {
    r.fillColor = p_fillColor;
  }

  if (p_cornerRadius !== null) {
    r.cornerRadius = p_cornerRadius;
  }

  return r;
};

displayitemMod.newFilledRectangle = function (p_bounds, p_fillColor, p_cornerRadius, p_alpha) {
  _balsamiq.default.exists(p_bounds);

  _balsamiq.default.exists(p_fillColor);

  var r = displayitemMod.newDisplayItem(displayitemMod.Type.RECTANGLE, p_bounds.x, p_bounds.y, p_bounds.width, p_bounds.height, {
    borderWidth: 0,
    fillColor: p_fillColor,
    cornerRadius: !_balsamiq.default.isNumber(p_cornerRadius) ? 0 : p_cornerRadius,
    alpha: !_balsamiq.default.isNumber(p_alpha) ? 1.0 : p_alpha,
    masks: []
  });
  return r;
};

displayitemMod.newFilledRectangleWithRectangularMask = function (p_bounds, p_fillColor, p_alpha, p_maskBounds) {
  _balsamiq.default.assertRect(p_bounds);

  _balsamiq.default.assertRect(p_maskBounds);

  _balsamiq.default.exists(p_fillColor);

  _balsamiq.default.assertNumber(p_alpha); //Back to integer coordinates!


  var boundsRounded = Object.assign({}, p_bounds);

  _types.default.roundCoords(boundsRounded);

  var maskBoundsRounded = Object.assign({}, p_maskBounds);

  _types.default.roundCoords(maskBoundsRounded);

  var displayItems = [];
  var boundsBottom = boundsRounded.y + boundsRounded.height;
  var boundsRight = boundsRounded.x + boundsRounded.width;
  var maskBottom = maskBoundsRounded.y + maskBoundsRounded.height;
  var maskRight = maskBoundsRounded.x + maskBoundsRounded.width;

  if (boundsRounded.y !== maskBoundsRounded.y) {
    // Top rectangle
    let bounds = _types.default.newRectangle(boundsRounded.x, boundsRounded.y, boundsRounded.width, maskBoundsRounded.y - boundsRounded.y);

    displayItems.push(displayitemMod.newFilledRectangle(bounds, p_fillColor, 0, p_alpha));
  }

  if (boundsRounded.x !== maskBoundsRounded.x) {
    // Left rectangle
    let bounds = _types.default.newRectangle(boundsRounded.x, maskBoundsRounded.y, maskBoundsRounded.x - boundsRounded.x, maskBoundsRounded.height);

    displayItems.push(displayitemMod.newFilledRectangle(bounds, p_fillColor, 0, p_alpha));
  }

  if (boundsRight !== maskRight) {
    // Right rectangle
    let bounds = _types.default.newRectangle(maskRight, maskBoundsRounded.y, boundsRight - maskRight, maskBoundsRounded.height);

    displayItems.push(displayitemMod.newFilledRectangle(bounds, p_fillColor, 0, p_alpha));
  }

  if (boundsBottom !== maskBottom) {
    // Bottom rectangle
    let bounds = _types.default.newRectangle(boundsRounded.x, maskBottom, boundsRounded.width, boundsBottom - maskBottom);

    displayItems.push(displayitemMod.newFilledRectangle(bounds, p_fillColor, 0, p_alpha));
  }

  _balsamiq.default.assert(displayItems.length > 0);

  return displayItems;
};

displayitemMod.newLine = function (p_start, p_end, p_width, p_color, p_alpha, p_dashArray, p_lineCap) {
  /*:types.newPoint*/
  _balsamiq.default.assert(false, "displayitemMod.newLine is deprecated, make a SVG path instead.");
};

displayitemMod.rotateImageBounds90 = function (p_bounds, p_imageWidth, p_imageHeight, p_rotation) {
  var res = {};

  switch (p_rotation % 360) {
    case 0:
      res.x = p_bounds.x;
      res.y = p_bounds.y;
      res.scaleX = p_bounds.width / p_imageWidth;
      res.scaleY = p_bounds.height / p_imageHeight;
      break;

    case 90:
      res.x = p_bounds.x;
      res.y = p_bounds.y + p_bounds.height;
      res.scaleX = p_bounds.height / p_imageWidth;
      res.scaleY = p_bounds.width / p_imageHeight;
      break;

    case 180:
      res.x = p_bounds.x + p_bounds.width;
      res.y = p_bounds.y + p_bounds.height;
      res.scaleX = p_bounds.width / p_imageWidth;
      res.scaleY = p_bounds.height / p_imageHeight;
      break;

    case 270:
      res.x = p_bounds.x + p_bounds.width;
      res.y = p_bounds.y;
      res.scaleX = p_bounds.height / p_imageWidth;
      res.scaleY = p_bounds.width / p_imageHeight;
      break;
  }

  return res;
};

displayitemMod.newImage = function (p_nativeImageID, p_bounds, p_imageWidth, p_imageHeight, p_sketchit, p_rotation) {
  var rotation = _balsamiq.default.isNumber(p_rotation) ? p_rotation : 0;
  var bounds = displayitemMod.rotateImageBounds90(p_bounds, p_imageWidth, p_imageHeight, rotation);
  var r = displayitemMod.newDisplayItem(displayitemMod.Type.IMAGE, bounds.x, bounds.y, p_imageWidth, p_imageHeight, {
    scaleX: bounds.scaleX,
    scaleY: bounds.scaleY,
    nativeImageID: p_nativeImageID,
    masks: [],
    sketchit: _balsamiq.default.isBoolean(p_sketchit) ? p_sketchit : false,
    rotation: rotation
  });
  return r;
};

displayitemMod.rotateBounds90 = function (p_bounds, p_rotation) {
  var res = {};

  switch (p_rotation % 360) {
    case 0:
      res.x = p_bounds.x;
      res.y = p_bounds.y;
      res.width = p_bounds.width;
      res.height = p_bounds.height;
      break;

    case 90:
      res.x = p_bounds.x;
      res.y = p_bounds.y + p_bounds.height;
      res.width = p_bounds.height;
      res.height = p_bounds.width;
      break;

    case 180:
      res.x = p_bounds.x + p_bounds.width;
      res.y = p_bounds.y + p_bounds.height;
      res.width = p_bounds.width;
      res.height = p_bounds.height;
      break;

    case 270:
      res.x = p_bounds.x + p_bounds.width;
      res.y = p_bounds.y;
      res.width = p_bounds.height;
      res.height = p_bounds.width;
      break;
  }

  return res;
};

displayitemMod.rotateTextLinksBounds = function (p_bounds, p_rotation) {
  const res = {};
  let rotation = 0;

  switch (p_rotation % 360) {
    case 90:
      res.x = p_bounds.x;
      res.y = p_bounds.y - p_bounds.width;
      res.width = p_bounds.height;
      res.height = p_bounds.width;
      break;

    case 15:
      res.x = p_bounds.x;
      res.y = p_bounds.y;
      res.width = p_bounds.width;
      res.height = p_bounds.height;
      rotation = 15;
      break;

    case 270:
      res.x = p_bounds.x - p_bounds.height;
      res.y = p_bounds.y;
      res.width = p_bounds.height;
      res.height = p_bounds.width;
      break;

    case 0:
      res.x = p_bounds.x;
      res.y = p_bounds.y;
      res.width = p_bounds.width;
      res.height = p_bounds.height;
      break;

    default:
      _balsamiq.default.assert(false, `Rotation value non supported: ${p_rotation}`);

  }

  return {
    bounds: res,
    rotation: rotation
  };
};

displayitemMod.rotateBounds90Glyph = function (p_bounds, p_rotation, p_offsetY) {
  var res = {};
  var p_bounds2;

  if (p_offsetY) {
    p_bounds2 = _balsamiq.default.clone(p_bounds);
    p_bounds2.y += p_offsetY;
    p_bounds2.height -= p_offsetY;
  } else {
    p_bounds2 = p_bounds;
  }

  switch (p_rotation % 360) {
    case 0:
      res.x = p_bounds2.x;
      res.y = p_bounds2.y;
      res.width = p_bounds2.width;
      res.height = p_bounds2.height;
      break;

    case 90:
      res.x = p_bounds2.x - (p_bounds2.height - p_bounds2.width) / 2;
      res.y = p_bounds2.y + (p_bounds2.height - p_bounds2.width) / 2 + p_bounds2.width;
      res.width = p_bounds2.width;
      res.height = p_bounds2.height;
      break;

    case 180:
      res.x = p_bounds2.x + p_bounds2.width;
      res.y = p_bounds2.y + p_bounds2.height;
      res.width = p_bounds2.width;
      res.height = p_bounds2.height;
      break;

    case 270:
      res.x = p_bounds2.x + (p_bounds2.height - p_bounds2.width) / 2 + p_bounds2.width;
      res.y = p_bounds2.y + (p_bounds2.height - p_bounds2.width) / 2;
      res.width = p_bounds2.width;
      res.height = p_bounds2.height;
      break;
  }

  return res;
};

displayitemMod.rotateBounds90RectTall = function (p_bounds, p_rotation) {
  var res = {};

  switch (p_rotation % 360) {
    case 0:
      res.x = p_bounds.x;
      res.y = p_bounds.y;
      res.width = p_bounds.width;
      res.height = p_bounds.height;
      break;

    case 90:
      res.x = p_bounds.x - (p_bounds.height - p_bounds.width) / 2;
      res.y = p_bounds.y + (p_bounds.height - p_bounds.width) / 2 + p_bounds.width;
      res.width = p_bounds.width;
      res.height = p_bounds.height;
      break;

    case 180:
      res.x = p_bounds.x + p_bounds.width;
      res.y = p_bounds.y + p_bounds.height;
      res.width = p_bounds.width;
      res.height = p_bounds.height;
      break;

    case 270:
      res.x = p_bounds.x - (p_bounds.height - p_bounds.width) / 2 - p_bounds.width;
      res.y = p_bounds.y - (p_bounds.height - p_bounds.width) / 2;
      res.width = p_bounds.width;
      res.height = p_bounds.height;
      break;
  }

  return res;
};

displayitemMod.PathsAssetLocation = {
  SKIN: "skin",
  APP: "app",
  NATIVE: "native"
};
displayitemMod.NativeResources = {
  SPLIT_IMAGE_GRIP: "splitImageGrip",
  SPLIT_IMAGE_GRIP_HOVER: "splitImageGripHover"
};
displayitemMod.NativeResourcesData = {
  [displayitemMod.NativeResources.SPLIT_IMAGE_GRIP]: {
    width: 84,
    height: 28,
    center_x: 40,
    center_y: 12
  },
  [displayitemMod.NativeResources.SPLIT_IMAGE_GRIP_HOVER]: {
    width: 84,
    height: 28,
    center_x: 40,
    center_y: 12
  }
};
/**
 * @param {Object} p_context
 * @param {String} p_assetName
 * @param {Object} p_bounds
 * @param {number=} p_fillColor
 * @param {number=} p_rotation
 * @param {number=} p_alpha
 * @returns {Array}
 */

displayitemMod.newPathsAsset = function (p_context, p_assetName, p_bounds, p_fillColor, p_alpha, p_rotation) {
  return this.newCustomPathsAsset(p_context, displayitemMod.PathsAssetLocation.SKIN, p_assetName, p_bounds, p_fillColor, p_alpha, p_rotation);
};
/**
 * @param {Object} p_context
 * @param {String} p_assetName
 * @param {String} p_assetLocation
 * @param {Object} p_bounds
 * @param {number=} p_fillColor
 * @param {number=} p_rotation
 * @param {number=} p_alpha
 * @returns {Array}
 */


displayitemMod.newCustomPathsAsset = function (p_context, p_assetLocation, p_assetName, p_bounds, p_fillColor, p_alpha, p_rotation) {
  _balsamiq.default.exists(p_context);

  _balsamiq.default.assertString(p_assetName);

  _balsamiq.default.exists(p_bounds);

  var asset = p_context.skin[p_assetName];

  _balsamiq.default.assert(asset, "the skin must have an asset called " + p_assetName);

  var r = displayitemMod.newDisplayItem(p_assetLocation === displayitemMod.AppPathsAsset.NATIVE_ASSET ? displayitemMod.Type.NATIVE_ASSET : displayitemMod.Type.SVG_PATHS_ASSET, p_bounds.x, p_bounds.y, asset.width, asset.height, {
    scaleX: p_bounds.width / asset.width,
    scaleY: p_bounds.height / asset.height,
    name: p_assetName,
    fillColor: _balsamiq.default.isNumber(p_fillColor) ? p_fillColor : null,
    rotation: _balsamiq.default.isNumber(p_rotation) ? p_rotation : 0,
    alpha: _balsamiq.default.isNumber(p_alpha) ? p_alpha : 1,
    location: p_assetLocation
  });
  return [r];
};
/**
 * @param {Object} p_context
 * @param {String} p_assetName
 * @param {String} p_assetLocation
 * @param {Object} p_bounds
 * @param {number=} p_fillColor
 * @param {number=} p_rotation
 * @param {number=} p_alpha
 * @returns {Array}
 */


displayitemMod.newNativeAsset = function (p_context, p_assetName, p_bounds, p_fillColor, p_alpha, p_rotation) {
  _balsamiq.default.exists(p_context);

  _balsamiq.default.assertString(p_assetName);

  _balsamiq.default.exists(p_bounds);

  var r = displayitemMod.newDisplayItem(displayitemMod.Type.NATIVE_ASSET, p_bounds.x, p_bounds.y, p_bounds.width, p_bounds.height, {
    scaleX: 1,
    scaleY: 1,
    name: p_assetName,
    fillColor: _balsamiq.default.isNumber(p_fillColor) ? p_fillColor : null,
    rotation: _balsamiq.default.isNumber(p_rotation) ? p_rotation : 0,
    alpha: _balsamiq.default.isNumber(p_alpha) ? p_alpha : 1,
    location: displayitemMod.PathsAssetLocation.NATIVE
  });
  return [r];
};

displayitemMod.newRotated90CCWPathsAsset = function (p_context, p_assetName, p_bounds, p_fillColor, p_alpha) {
  var bounds = _types.default.newRectangle(p_bounds.x, p_bounds.width - p_bounds.y, p_bounds.height, p_bounds.width);

  return displayitemMod.newPathsAsset(p_context, p_assetName, bounds, p_fillColor, p_alpha, 90);
};

displayitemMod.newRotated90CWPathsAsset = function (p_context, p_assetName, p_bounds, p_fillColor, p_alpha) {
  var bounds = _types.default.newRectangle(p_bounds.x + p_bounds.width, p_bounds.y, p_bounds.height, p_bounds.width);

  return displayitemMod.newPathsAsset(p_context, p_assetName, bounds, p_fillColor, p_alpha, -90);
};

displayitemMod.newRotated180PathsAsset = function (p_context, p_assetName, p_bounds, p_fillColor, p_alpha) {
  var bounds = _types.default.newRectangle(p_bounds.x + p_bounds.width, p_bounds.y + p_bounds.height, p_bounds.height, p_bounds.width);

  return displayitemMod.newPathsAsset(p_context, p_assetName, bounds, p_fillColor, p_alpha, 180);
};

displayitemMod.newText = function (p_textDom, p_attributedText, p_textStyle, p_bounds, p_ascender) {
  var r = displayitemMod.newDisplayItem(displayitemMod.Type.TEXT, p_bounds.x, p_bounds.y, p_bounds.width, p_bounds.height, {
    textDom: p_textDom,
    attributedText: p_attributedText,
    fontFamily: p_textStyle.fontFamily,
    fontSize: p_textStyle.fontSize,
    ascender: p_ascender,
    bold: p_textStyle.bold,
    italic: p_textStyle.italic,
    underline: p_textStyle.underline,
    textColor: p_textStyle.textColor,
    linkColor: p_textStyle.linkColor
  });
  return r;
};

displayitemMod.newRotated90CCWSlice9 = function (p_name, p_bounds, p_context, p_fillColor, p_alpha) {
  var skin = p_context.skin; // Get slice sizes

  var scale9Values = skin[p_name].scale9;

  _balsamiq.default.assert(scale9Values !== undefined, "Skin asset " + p_name + ".scale9 not found");

  var left = scale9Values[0];
  var bottom = scale9Values[1];
  var top = scale9Values[2];
  var right = scale9Values[3];

  if (p_bounds.height < top + bottom) {
    var heightRatio = p_bounds.height / (top + bottom);
    top = top * heightRatio;
    bottom = bottom * heightRatio;
  }

  if (p_bounds.width < left + right) {
    var widthRatio = p_bounds.width / (left + right);
    left = left * widthRatio;
    right = right * widthRatio;
  }

  var centerWidth = Math.max(p_bounds.width - (left + right), 0);
  var centerHeight = Math.max(p_bounds.height - (top + bottom), 0); // Overlap is performed streching horizontally N, S and vertically E, W

  var OVERLAP = 0.25; // to hide artifacts on borders due to rendering

  var OVERLAP2 = OVERLAP * 2; // If box is too small N,S,E,W size can be 0, so we must overlap NW, NE, SE, SW

  var LEFT_RIGHT_OVERLAP = centerWidth === 0 ? OVERLAP2 : 0;
  var TOP_BOTTOM_OVERLAP = centerHeight === 0 ? OVERLAP2 : 0;
  var pieces = [//x, y , height, width
  {
    suffix: 'ne',
    bounds: _types.default.newRectangle(p_bounds.x, p_bounds.y + top + TOP_BOTTOM_OVERLAP, top + TOP_BOTTOM_OVERLAP, left + LEFT_RIGHT_OVERLAP)
  }, //boundsNW
  {
    suffix: 'e',
    bounds: _types.default.newRectangle(p_bounds.x + left - OVERLAP, p_bounds.y + top, top, centerWidth + OVERLAP2)
  }, //boundsN
  {
    suffix: 'se',
    bounds: _types.default.newRectangle(p_bounds.x + left + centerWidth - LEFT_RIGHT_OVERLAP, p_bounds.y + top + TOP_BOTTOM_OVERLAP, top + TOP_BOTTOM_OVERLAP, right + LEFT_RIGHT_OVERLAP)
  }, //boundsNE
  {
    suffix: 's',
    bounds: _types.default.newRectangle(p_bounds.x + left + centerWidth, p_bounds.y + top - OVERLAP + centerHeight + OVERLAP2, centerHeight + OVERLAP2, right)
  }, //boundsE
  {
    suffix: 'sw',
    bounds: _types.default.newRectangle(p_bounds.x + left + centerWidth - LEFT_RIGHT_OVERLAP, p_bounds.y + top + centerHeight + bottom, bottom + TOP_BOTTOM_OVERLAP, right + LEFT_RIGHT_OVERLAP)
  }, //boundsSE
  {
    suffix: 'w',
    bounds: _types.default.newRectangle(p_bounds.x + left - OVERLAP, p_bounds.y + top + centerHeight + bottom, bottom, centerWidth + OVERLAP2)
  }, //boundsS
  {
    suffix: 'nw',
    bounds: _types.default.newRectangle(p_bounds.x, p_bounds.y + top + centerHeight + bottom, bottom + TOP_BOTTOM_OVERLAP, left + LEFT_RIGHT_OVERLAP)
  }, //boundsSW
  {
    suffix: 'n',
    bounds: _types.default.newRectangle(p_bounds.x, p_bounds.y + top - OVERLAP + centerHeight + OVERLAP2, centerHeight + OVERLAP2, left)
  }, //boundsW
  {
    suffix: 'c',
    bounds: _types.default.newRectangle(p_bounds.x + left - OVERLAP, p_bounds.y + top - OVERLAP + centerHeight + OVERLAP2, centerHeight + OVERLAP2, centerWidth + OVERLAP2)
  } //boundsC
  ];
  var displayItems = [];
  pieces.forEach(function (skinAsset, i) {
    if (skinAsset.bounds.width > 0 && skinAsset.bounds.height > 0) {
      var name = p_name + '_' + skinAsset.suffix;

      if (skin[name]) {
        _balsamiq.default.exists(skin[name].width);

        displayItems = displayItems.concat(displayitemMod.newPathsAsset(p_context, name, skinAsset.bounds, p_fillColor, p_alpha, 90));
      }
    }
  });
  return displayItems;
};

displayitemMod.newSlice9 = function (p_name, p_bounds, p_context, p_fillColor, p_alpha) {
  var skin = p_context.skin; // Get slice sizes

  var scale9Values = skin[p_name].scale9;

  _balsamiq.default.assert(scale9Values !== undefined, "Skin asset " + p_name + ".scale9 not found");

  var top = scale9Values[0];
  var left = scale9Values[1];
  var right = scale9Values[2];
  var bottom = scale9Values[3];

  if (p_bounds.height < top + bottom) {
    var heightRatio = p_bounds.height / (top + bottom);
    top = top * heightRatio;
    bottom = bottom * heightRatio;
  }

  if (p_bounds.width < left + right) {
    var widthRatio = p_bounds.width / (left + right);
    left = left * widthRatio;
    right = right * widthRatio;
  }

  var centerWidth = Math.max(p_bounds.width - (left + right), 0);
  var centerHeight = Math.max(p_bounds.height - (top + bottom), 0); // Overlap is performed streching horizontally N, S and vertically E, W

  var OVERLAP = 0.25; // to hide artifacts on borders due to rendering

  var OVERLAP2 = OVERLAP * 2; // If box is too small N,S,E,W size can be 0, so we must overlap NW, NE, SE, SW

  var LEFT_RIGHT_OVERLAP = centerWidth === 0 ? OVERLAP2 : 0;
  var TOP_BOTTOM_OVERLAP = centerHeight === 0 ? OVERLAP2 : 0;
  var pieces = [{
    suffix: 'nw',
    bounds: _types.default.newRectangle(p_bounds.x, p_bounds.y, left + LEFT_RIGHT_OVERLAP, top + TOP_BOTTOM_OVERLAP)
  }, //boundsNW
  {
    suffix: 'n',
    bounds: _types.default.newRectangle(p_bounds.x + left - OVERLAP, p_bounds.y, centerWidth + OVERLAP2, top)
  }, //boundsN
  {
    suffix: 'ne',
    bounds: _types.default.newRectangle(p_bounds.x + left + centerWidth - LEFT_RIGHT_OVERLAP, p_bounds.y, right + LEFT_RIGHT_OVERLAP, top + TOP_BOTTOM_OVERLAP)
  }, //boundsNE
  {
    suffix: 'e',
    bounds: _types.default.newRectangle(p_bounds.x + left + centerWidth, p_bounds.y + top - OVERLAP, right, centerHeight + OVERLAP2)
  }, //boundsE
  {
    suffix: 'se',
    bounds: _types.default.newRectangle(p_bounds.x + left + centerWidth - LEFT_RIGHT_OVERLAP, p_bounds.y + top + centerHeight - TOP_BOTTOM_OVERLAP, right + LEFT_RIGHT_OVERLAP, bottom + TOP_BOTTOM_OVERLAP)
  }, //boundsSE
  {
    suffix: 's',
    bounds: _types.default.newRectangle(p_bounds.x + left - OVERLAP, p_bounds.y + top + centerHeight, centerWidth + OVERLAP2, bottom)
  }, //boundsS
  {
    suffix: 'sw',
    bounds: _types.default.newRectangle(p_bounds.x, p_bounds.y - TOP_BOTTOM_OVERLAP + top + centerHeight, left + LEFT_RIGHT_OVERLAP, bottom + TOP_BOTTOM_OVERLAP)
  }, //boundsSW
  {
    suffix: 'w',
    bounds: _types.default.newRectangle(p_bounds.x, p_bounds.y + top - OVERLAP, left, centerHeight + OVERLAP2)
  }, //boundsW
  {
    suffix: 'c',
    bounds: _types.default.newRectangle(p_bounds.x + left - OVERLAP, p_bounds.y + top - OVERLAP, centerWidth + OVERLAP2, centerHeight + OVERLAP2)
  } //boundsC
  ];
  var displayItems = [];
  pieces.forEach(function (skinAsset, i) {
    if (skinAsset.bounds.width > 0 && skinAsset.bounds.height > 0) {
      var name = p_name + '_' + skinAsset.suffix;

      if (skin[name]) {
        _balsamiq.default.exists(skin[name].width);

        displayItems = displayItems.concat(displayitemMod.newPathsAsset(p_context, name, skinAsset.bounds, p_fillColor, p_alpha));
      }
    }
  });
  return displayItems;
};

displayitemMod.createDisplayItemsFromSkin = function (p_name, p_color, p_bounds, p_context) {
  var mainValue = p_context.skin[p_name];

  _balsamiq.default.exists(mainValue, "the skin must have selector " + p_name);

  var displayItems;

  if (mainValue.type === "SVG_PATHS") {
    displayItems = displayitemMod.newPathsAsset(p_context, p_name, p_bounds, p_color);
  } else if (mainValue.type === "SVG_SCALE9") {
    displayItems = displayitemMod.newSlice9(p_name, p_bounds, p_context, p_color);
  }

  return displayItems;
};

displayitemMod.mirrorDisplayItems = function (p_displayItems, p_bounds, p_mirrorX, p_mirrorY) {
  _balsamiq.default.exists(p_displayItems);

  if (!_balsamiq.default.isArray(p_displayItems)) {
    p_displayItems = [p_displayItems];
  }

  if (p_mirrorX) {
    p_displayItems.forEach(function (displayItem) {
      displayItem.x = p_bounds.x + p_bounds.width - (displayItem.x - p_bounds.x + displayItem.width);
      displayItem.scaleX *= -1;
    });
  }

  if (p_mirrorY) {
    p_displayItems.forEach(function (displayItem) {
      displayItem.y = p_bounds.y + p_bounds.height - (displayItem.y - p_bounds.y + displayItem.height);
      displayItem.scaleY *= -1;
    });
  }
};
/**
 *
 * @param {Array} p_displayItems
 * @param {Number?} p_maskID
 */


displayitemMod.maskDisplayItems = function (p_displayItems, p_maskID) {
  if (!_balsamiq.default.isNumber(p_maskID)) {
    return;
  }

  _balsamiq.default.exists(p_displayItems);

  if (!_balsamiq.default.isArray(p_displayItems)) {
    p_displayItems = [p_displayItems];
  }

  p_displayItems = p_displayItems.flat(); // TODO can be done by the caller?

  p_displayItems.forEach(function (displayItem) {
    if (_balsamiq.default.has(displayItem, "masks")) {
      _balsamiq.default.assert(displayItem.masks.indexOf(p_maskID) === -1);

      displayItem.masks.push(p_maskID);
    }
  });
};

displayitemMod.setRotation = function (p_displayItems, p_rotationAngle) {
  _balsamiq.default.exists(p_displayItems);

  if (!_balsamiq.default.isArray(p_displayItems)) {
    p_displayItems = [p_displayItems];
  }

  p_displayItems.forEach(function (displayItem) {
    displayItem.rotation = p_rotationAngle;
  });
};

displayitemMod.setAlpha = function (p_displayItems, p_alpha) {
  _balsamiq.default.assertArray(p_displayItems);

  _balsamiq.default.assertNumber01(p_alpha);

  p_displayItems.forEach(function (displayItem) {
    displayItem.alpha = p_alpha;
  });
};

displayitemMod.setBorderAlpha = function (p_displayItems, p_borderAlpha) {
  if (!_balsamiq.default.isArray(p_displayItems) && _balsamiq.default.isObject(p_displayItems)) {
    p_displayItems = [p_displayItems];
  } else {
    _balsamiq.default.assertArray(p_displayItems);
  }

  _balsamiq.default.assertNumber01(p_borderAlpha);

  p_displayItems.forEach(function (displayItem) {
    displayItem.borderAlpha = p_borderAlpha;
  });
}; // TODO: need arrows for external/broken link


displayitemMod.AppPathsAsset = {
  "#skinName": "links",
  regular: {
    type: "SVG_PATHS",
    xOffset: 4,
    width: 12,
    height: 14,
    colors: [0xdb5048],
    paths: ["M6.81579874,6.81579874 L10.7140176,10.7140176 L10.7140176,1.95855711e-15 L7.10542736e-15,0 L3.35899774,3.35899774 C-0.0821086808,6.53954296 1.43787888,13 2.48224341,13 C3.53328048,13 2.26400891,9.63066706 6.81579874,6.81579874 L6.81579874,6.81579874 Z"]
  },
  broken: {
    type: "SVG_PATHS",
    xOffset: 1,
    width: 14,
    height: 14,
    colors: [0xdb5048, 0xFFFFFF],
    paths: ["M7.5,0 L15,13 L0,13 L7.5,0 L7.5,0 Z", "M7.5,12 C8.05228475,12 8.5,11.5522847 8.5,11 C8.5,10.4477153 8.05228475,10 7.5,10 C6.94771525,10 6.5,10.4477153 6.5,11 C6.5,11.5522847 6.94771525,12 7.5,12 Z M6.5,4.9998423 C6.5,4.44764465 6.94386482,4 7.5,4 C8.05228475,4 8.5,4.44761205 8.5,4.9998423 L8.5,8.2501577 C8.5,8.80235535 8.05613518,9.25 7.5,9.25 C6.94771525,9.25 6.5,8.80238795 6.5,8.2501577 L6.5,4.9998423 Z"]
  },
  external: {
    type: "SVG_PATHS",
    xOffset: 3,
    width: 14,
    height: 14,
    colors: [0xdb5048],
    paths: ["M12,13 L0,13 L0,2 L5,2 L5,5 L3,5 L3,10 L9,10 L9,8 L12,8 L12,13 Z M10.6,4 L6.6,8 L5,6.4 L9,2.4 L6.6,0 L13,4.39115574e-16 L13,6.4 L10.6,4 Z"]
  },
  back: {
    type: "SVG_PATHS",
    xOffset: 4,
    width: 14,
    height: 14,
    colors: [0xdb5048],
    paths: ["M4.547,5.245v-5.512l-7.576,7.576l7.576,7.575v-4.75c4.684,0.185,8.176-5.459,7.438-6.198C11.242,3.194,9.757,6.473,4.547,5.245L4.547,5.245z"]
  } // grip: { type: "SVG_PATHS", xOffset: 0, width: 58, height: 4, colors: [ 0x394045 ], paths: [ "M57,3L58,3L58,4L57,4L57,3ZM57,-1.13686838e-13L58,-1.13686838e-13L58,1L57,1L57,-1.13686838e-13ZM54,3L55,3L55,4L54,4L54,3ZM54,-1.13686838e-13L55,-1.13686838e-13L55,1L54,1L54,-1.13686838e-13ZM51,3L52,3L52,4L51,4L51,3ZM51,-1.13686838e-13L52,-1.13686838e-13L52,1L51,1L51,-1.13686838e-13ZM48,3L49,3L49,4L48,4L48,3ZM48,-1.13686838e-13L49,-1.13686838e-13L49,1L48,1L48,-1.13686838e-13ZM45,3L46,3L46,4L45,4L45,3ZM45,-1.13686838e-13L46,-1.13686838e-13L46,1L45,1L45,-1.13686838e-13ZM42,3L43,3L43,4L42,4L42,3ZM42,-1.13686838e-13L43,-1.13686838e-13L43,1L42,1L42,-1.13686838e-13ZM39,3L40,3L40,4L39,4L39,3ZM39,-1.13686838e-13L40,-1.13686838e-13L40,1L39,1L39,-1.13686838e-13ZM36,3L37,3L37,4L36,4L36,3ZM36,-1.13686838e-13L37,-1.13686838e-13L37,1L36,1L36,-1.13686838e-13ZM33,3L34,3L34,4L33,4L33,3ZM33,-1.13686838e-13L34,-1.13686838e-13L34,1L33,1L33,-1.13686838e-13ZM30,3L31,3L31,4L30,4L30,3ZM28,3L28,4L27,4L27,3L28,3ZM25,3L25,4L24,4L24,3L25,3ZM22,3L22,4L21,4L21,3L22,3ZM19,3L19,4L18,4L18,3L19,3ZM16,3L16,4L15,4L15,3L16,3ZM13,3L13,4L12,4L12,3L13,3ZM10,3L10,4L9,4L9,3L10,3ZM7,3L7,4L6,4L6,3L7,3ZM4,3L4,4L3,4L3,3L4,3ZM1,3L1,4L0,4L0,3L1,3ZM31,-1.13686838e-13L31,1L30,1L30,-1.13686838e-13L31,-1.13686838e-13ZM28,-1.13686838e-13L28,1L27,1L27,-1.13686838e-13L28,-1.13686838e-13ZM25,-1.13686838e-13L25,1L24,1L24,-1.13686838e-13L25,-1.13686838e-13ZM22,-1.13686838e-13L22,1L21,1L21,-1.13686838e-13L22,-1.13686838e-13ZM19,-1.13686838e-13L19,1L18,1L18,-1.13686838e-13L19,-1.13686838e-13ZM16,-1.13686838e-13L16,1L15,1L15,-1.13686838e-13L16,-1.13686838e-13ZM13,-1.13686838e-13L13,1L12,1L12,-1.13686838e-13L13,-1.13686838e-13ZM10,-1.13686838e-13L10,1L9,1L9,-1.13686838e-13L10,-1.13686838e-13ZM7,-1.13686838e-13L7,1L6,1L6,-1.13686838e-13L7,-1.13686838e-13ZM4,-1.13686838e-13L4,1L3,1L3,-1.13686838e-13L4,-1.13686838e-13ZM1,-1.13686838e-13L1,1L0,1L0,-1.13686838e-13L1,-1.13686838e-13Z"]},

};

displayitemMod.getLinkIconBounds = function (p_linkAreaBounds) {
  let p_rotation = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
  let p_xOffset = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : displayitemMod.LINK_ICON_X_OFFSET;

  _balsamiq.default.assertRect(p_linkAreaBounds);

  _balsamiq.default.assertNumber(p_rotation); // Icon bounds in horizontal link area


  var iconBounds = _types.default.newRectangle(p_linkAreaBounds.x + p_linkAreaBounds.width - displayitemMod.LINK_ICON_SIZE + p_xOffset, p_linkAreaBounds.y + p_linkAreaBounds.height - displayitemMod.LINK_ICON_SIZE + displayitemMod.LINK_ICON_Y_OFFSET, displayitemMod.LINK_ICON_SIZE, displayitemMod.LINK_ICON_SIZE);

  if (p_rotation) {
    // If we have rotation, rotate icon origin around linkarea top left
    let rotated = _types.default.rotatePtDeg(p_linkAreaBounds, iconBounds, p_rotation);

    iconBounds.x = rotated.x;
    iconBounds.y = rotated.y;
  }

  return iconBounds;
};

displayitemMod.newLinkDisplayItems = function (p_linkStatus, p_linkAreaBounds, p_linkAreaAngle) {
  let xOffset = displayitemMod.AppPathsAsset[p_linkStatus].xOffset;
  var bounds = displayitemMod.getLinkIconBounds(p_linkAreaBounds, p_linkAreaAngle, xOffset);
  var arrow = displayitemMod.newDisplayItem(displayitemMod.Type.SVG_PATHS_ASSET, bounds.x, bounds.y, bounds.width, bounds.height, {
    name: p_linkStatus,
    location: displayitemMod.PathsAssetLocation.APP,
    alpha: 0.7
  });
  return [arrow];
};

displayitemMod.addColorTransformToDisplayItems = function (p_skin, p_displayItems, p_color) {
  _balsamiq.default.exists(p_skin);

  _balsamiq.default.assertArray(p_displayItems);

  _balsamiq.default.assertNumber(p_color);

  let matrix = _colors.default.colorTransformMatrix(p_color);

  p_displayItems.forEach(item => {
    _balsamiq.default.assert(item.displayItemType === displayitemMod.Type.SVG_PATHS_ASSET);

    let transformedColors = p_skin[item.name].colors.map(color => {
      return _colors.default.applyColorMatrixToColor(matrix, color);
    });
    item.overriddenColors = _balsamiq.default.clone(transformedColors);
  });
};

displayitemMod.setOverlayRectFixedSize = function (p_displayItem) {
  let p_val = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;

  _balsamiq.default.assert(p_displayItem.displayItemType === displayitemMod.Type.RECTANGLE);

  p_displayItem.fixedSize = p_val;
};

displayitemMod.isOverlayRectFixedSize = function (p_displayItem) {
  return p_displayItem.fixedSize ? true : false;
};

displayitemMod.scaleOverlayRectFixedSize = function (p_displayItem, scale) {
  _balsamiq.default.assertNumber(scale);

  if (!p_displayItem || p_displayItem.displayItemType !== displayitemMod.Type.RECTANGLE || !displayitemMod.isOverlayRectFixedSize(p_displayItem)) {
    return undefined;
  }

  var displayItemCopy = Object.assign({}, p_displayItem);
  let xDelta = displayItemCopy.width * (1 - scale) / 2;
  let yDelta = displayItemCopy.height * (1 - scale) / 2;
  displayItemCopy.width = Math.round(displayItemCopy.width);
  displayItemCopy.height = Math.round(displayItemCopy.height);
  displayItemCopy.x = Math.floor(scale * displayItemCopy.x - xDelta) + 0.5;
  displayItemCopy.y = Math.floor(scale * displayItemCopy.y - yDelta) + 0.5;
  return displayItemCopy;
};

displayitemMod.setOverlayRectScalableFixedBorderSize = function (p_displayItem) {
  let p_val = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;

  _balsamiq.default.assert(p_displayItem.displayItemType === displayitemMod.Type.RECTANGLE);

  p_displayItem.fixedBorderSize = p_val;
};

displayitemMod.isOverlayRectScalableFixedBorderSize = function (p_displayItem) {
  return p_displayItem.fixedBorderSize ? true : false;
};

displayitemMod.scaleOverlayRectScalableFixedBorderSize = function (p_displayItem, scale) {
  _balsamiq.default.assertNumber(scale);

  if (!p_displayItem || p_displayItem.displayItemType !== displayitemMod.Type.RECTANGLE || !displayitemMod.isOverlayRectScalableFixedBorderSize(p_displayItem)) {
    return undefined;
  }

  var dl = Object.assign({}, p_displayItem);
  dl.width = Math.round(scale * dl.width) - 1;
  dl.height = Math.round(scale * dl.height) - 1;
  dl.x = Math.floor(scale * dl.x) + 0.5;
  dl.y = Math.floor(scale * dl.y) + 0.5;
  return dl;
};

displayitemMod.scaleOverlayDisplayItems = function (p_displayItems, p_scale) {
  _balsamiq.default.assertArray(p_displayItems);

  _balsamiq.default.assertNumber(p_scale);

  let res = [];

  for (let i = 0; i < p_displayItems.length; i += 1) {
    let di = p_displayItems[i];

    if (displayitemMod.isOverlayRectScalableFixedBorderSize(di)) {
      let newDi = displayitemMod.scaleOverlayRectScalableFixedBorderSize(di, p_scale);
      displayitemMod.setOverlayRectScalableFixedBorderSize(newDi, false);
      res.push(newDi);
    } else if (displayitemMod.isOverlayRectFixedSize(di)) {
      let newDi = displayitemMod.scaleOverlayRectFixedSize(di, p_scale);
      displayitemMod.setOverlayRectFixedSize(newDi, false);
      res.push(newDi);
    } else {
      res.push(di);
    }
  }

  return res;
};

displayitemMod.setSelectionDisplayItemsFlag = function (p_displayItems) {
  let val = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;
  var i;

  for (i = 0; i < p_displayItems.length; i += 1) {
    p_displayItems[i].selection = val;
  }
};

displayitemMod.setDisplayItemsBorder = function (p_displayItems, p_borderWidth, p_borderColor) {
  var i;

  for (i = 0; i < p_displayItems.length; i += 1) {
    p_displayItems[i].borderWidth = p_borderWidth;
    p_displayItems[i].borderColor = p_borderColor;
  }
};

var _default = displayitemMod;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/displaynode.js":
/*!********************************!*\
  !*** ./js/core/displaynode.js ***!
  \********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.DisplayNode = void 0;
exports.applyFuncToDisplayNode = applyFuncToDisplayNode;
exports.copyDisplayDataToExistingDisplayNode = copyDisplayDataToExistingDisplayNode;
exports.newDisplayNode = newDisplayNode;
exports.searchDisplayNodes = searchDisplayNodes;
exports.setCropDataToDisplayNode = setCropDataToDisplayNode;
exports.setCropOffsetToDisplayNode = setCropOffsetToDisplayNode;

var ControlID = _interopRequireWildcard(__webpack_require__(/*! core/controls/controlid */ "./js/core/controls/controlid.js"));

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function newDisplayNode(displayData, controlID, children) {
  _balsamiq.default.exists(displayData);

  _balsamiq.default.assertArray(controlID);

  _balsamiq.default.assertArray(children);

  return {
    displayData: displayData,
    controlID: _balsamiq.default.clone(controlID),
    children: children
  };
}

function setCropOffsetToDisplayNode(displayNode, newOffset) {
  _balsamiq.default.assertPoint(newOffset);

  if (!displayNode.displayData.cropOffset) {
    displayNode.displayData.cropOffset = _types.default.newPoint(0, 0);
  }

  displayNode.displayData.cropOffset.x = newOffset.x;
  displayNode.displayData.cropOffset.y = newOffset.y;
}

function applyFuncToDisplayNode(displayNode, func) {
  displayNode.children.forEach(childNode => {
    applyFuncToDisplayNode(childNode, func);
  });
  func(displayNode);
}

function copyDisplayDataToExistingDisplayNode(displayNodeFrom, displayNodeTo) {
  if (displayNodeTo) {
    displayNodeTo.displayData = _balsamiq.default.clone(displayNodeFrom.displayData); //displayNodeTo.controlID = Balsamiq.clone(displayNodeFrom.controlID);
  }
}

function setCropDataToDisplayNode(displayNode, cropRect, cropOffset, useCropMask) {
  _balsamiq.default.assertPoint(cropOffset);

  _balsamiq.default.assertRect(cropRect);

  _balsamiq.default.assertBool(useCropMask);

  displayNode.displayData.cropOffset = cropOffset;
  displayNode.displayData.cropMask = _types.default.makeMaskRect(cropRect);
  displayNode.displayData.useCropMask = useCropMask;
}

function searchDisplayNodes(displayNode, searchID) {
  if (ControlID.areEqual(displayNode.controlID, searchID)) {
    return displayNode;
  }

  for (let i = 0; i < displayNode.children.length; i += 1) {
    let node = searchDisplayNodes(displayNode.children[i], searchID);

    if (node) {
      return node;
    }
  }

  return null;
}

class DisplayNode {
  constructor(displayData, controlID, children) {
    _balsamiq.default.exists(displayData);

    _balsamiq.default.assertArray(controlID);

    _balsamiq.default.assertArray(children);

    this.displayData = displayData;
    this.controlID = _balsamiq.default.clone(controlID);
    this.children = children;
  }

  setCropOffset(newOffset) {
    _balsamiq.default.assertPoint(newOffset);

    if (!this.displayData.cropOffset) {
      this.displayData.cropOffset = _types.default.newPoint(0, 0);
    }

    this.displayData.cropOffset.x = newOffset.x;
    this.displayData.cropOffset.y = newOffset.y;
  }

  applyFunc(func) {
    this.children.forEach(childNode => {
      childNode.applyFunc(func);
    });
    func(this);
  }

  setCropData(cropRect, cropOffset, useCropMask) {
    _balsamiq.default.assertPoint(cropOffset);

    _balsamiq.default.assertRect(cropRect);

    _balsamiq.default.assertBool(useCropMask);

    this.displayData.cropOffset = cropOffset;
    this.displayData.cropMask = _types.default.makeMaskRect(cropRect);
    this.displayData.useCropMask = useCropMask;
  }

  _searchNode(nodeID) {
    if (ControlID.areEqual(this.controlID, nodeID)) {
      return this;
    }

    for (let i = 0; i < this.children.length; i += 1) {
      let node = this.children[i]._searchNode(nodeID);

      if (node) {
        return node;
      }
    }

    return null;
  }

}

exports.DisplayNode = DisplayNode;

/***/ }),

/***/ "./js/core/graphics/commands.js":
/*!**************************************!*\
  !*** ./js/core/graphics/commands.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.GraphicCommands = void 0;
const GraphicCommands = {
  INIT: "I",
  // [width, height, dpiScale (1,2,3,4)] width, height are not dpi-scaled
  SAVE: "S",
  RESTORE: "R",
  TRANSLATE: "T",
  // [dx, dy]
  ROTATE: "RO",
  // [angle]
  SCALE: "SC",
  // [sx, sy]
  BEGIN: "B",
  CLOSE: "C",
  DRAW: "D",
  // [fillColor, fillAlpha, strokeColor, strokeAlpha]
  CLIP: "CP",
  LINE_DASH: "LD",
  // [numDashes, dash1, dash2, ..]
  LINE_CAP: "LC",
  // ["butt" | "round"]
  LINE_WIDTH: "LW",
  // [value >= 0]
  MOVE: "M",
  // [x, y]
  LINE: "L",
  // [x, y]
  ARC: "A",
  // [x1, y1, x2, y2, rx, ry]
  QUADRATIC: "Q",
  // [x1, y1, x2, y2]
  BEZIER: "BZ",
  // [x1, y1, x2, y2, x3, y3]
  TEXT_STYLE: "TS",
  // [fontFamily, fontSize, bold, italic, underline, textColor]
  TEXT: "TX",
  // [string, x, y]
  TEXT_RTL: "TR",
  // NEVER USED! TO BE REMOVED?
  DRAW_PATH: "DP",
  // [skinName, pathName, fillAlpha, numColors, color1, color2, ..]
  IMAGE: "IM",
  // [nativeImageID, x, y, width, height]
  SKETCHED_IMAGE: "SI",
  // [nativeImageID, x, y, width, height]
  COLORIZED_IMAGE: "CI",
  // [nativeImageID, x, y, width, height, color]
  LINK: "K",
  LINK_PAGE: "KP"
};
exports.GraphicCommands = GraphicCommands;

/***/ }),

/***/ "./js/core/graphics/linkareas.js":
/*!***************************************!*\
  !*** ./js/core/graphics/linkareas.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.LinkAreas = void 0;
exports.makeLinkAreasDisplayList = makeLinkAreasDisplayList;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class LinkAreas {
  constructor(context) {
    _balsamiq.default.exists(context);

    this.context = context;
    this.areas = []; // { bounds: Rectangle, target: String?, controlID: String }
  }

  getAreas() {
    return this.areas;
  }

  build(rootDisplayData, ignoreTopLeft) {
    ignoreTopLeft = _balsamiq.default.isBoolean(ignoreTopLeft) ? ignoreTopLeft : false;
    rootDisplayData.children.forEach(function (child) {
      const x = ignoreTopLeft ? 0 : rootDisplayData.displayData.bounds.x;
      const y = ignoreTopLeft ? 0 : rootDisplayData.displayData.bounds.y;
      let tempAreas = this.processControl(x, y, child, true);

      if (tempAreas != null) {
        tempAreas.forEach(function (area) {
          this.areas.push(area);
        }, this);
      }
    }, this);

    if (this.areas.length > 0) {
      // cut each area with the ones following it (that have higher z-order)
      let tempAreas = [this.areas[0]];

      for (let i = 1; i < this.areas.length; i++) {
        this.removeOverlappingAreas(this.areas[i], tempAreas);
        tempAreas.push(this.areas[i]);
      }

      this.areas = tempAreas;
    }

    if (this.areas.length > 0) {
      this.removeAreasHiddenByControls(rootDisplayData);
    }
  }
  /**
   * Remove link areas that are covered by other controls
   * @private
   * @param {DisplayNode} rootDisplayData
   */


  removeAreasHiddenByControls(rootDisplayData) {
    this.areas = this.areas.reduce((newAreas, area) => {
      const getAreasWithoutCoveredOnes = this.getAreasWithoutCoveredOnes(area, rootDisplayData);
      return newAreas.concat(getAreasWithoutCoveredOnes);
    }, []);
  }
  /**
   * Get new areas considering the areas covered by other controls
   * @private
   * @param {Object} area
   * @param {DisplayNode} rootControl
   * @return {[LinkArea]} new areas
   */


  getAreasWithoutCoveredOnes(area, rootControl) {
    const displayNodes = this.excludePreviousAndCurrentControl(rootControl.children, area.controlID);
    const areasToAdd = this.newAreasForDisplayNodes(displayNodes, area);
    return areasToAdd ? areasToAdd : [];
  }
  /**
   * Exclude controls have a lower z-index or the same in case it's a group
   * If there is a group we need to go in inside the children recursively
   * @private
   * @param {[DisplayNode]} displayNodes
   * @param {[Int]} areaControlID
   * @return {[DisplayNode]}
   */


  excludePreviousAndCurrentControl(displayNodes, areaControlID) {
    const indexOfControlAreaInDisplayNodes = displayNodes.findIndex(d => d.controlID[d.controlID.length - 1] === areaControlID[0]);

    if (areaControlID.length === 1) {
      return displayNodes.slice(indexOfControlAreaInDisplayNodes);
    } else {
      const displayNodesWithoutLowerControls = displayNodes.slice(indexOfControlAreaInDisplayNodes); // remove displayNodes that come before linkArea's control

      return [{ ...displayNodesWithoutLowerControls[0],
        children: this.excludePreviousAndCurrentControl(displayNodesWithoutLowerControls[0].children, areaControlID.slice(1))
      }, ...displayNodesWithoutLowerControls.slice(1)];
    }
  }
  /**
   * Recursively get new areas considering children
   * @private
   * @param {[DisplayNode]} displayNodes
   * @param {Area} area
   * @param {Object} origin
   * @return {[Area]}
   */


  newAreasForDisplayNodes(displayNodes, area) {
    let origin = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {
      x: 0,
      y: 0
    };

    const newAreasForDisplayNodes = (areaToTest, displayNode) => {
      const areaBelongsToDisplayNode = areaToTest && areaToTest.controlID.toString() === displayNode.controlID.toString();
      return displayNode.children.length === 0 || areaBelongsToDisplayNode ? this.newAreasForDisplayNode(displayNode, areaToTest, origin) : this.newAreasForDisplayNodes(displayNode.children, areaToTest, {
        x: displayNode.displayData.origin.x + origin.x,
        y: displayNode.displayData.origin.y + origin.y
      });
    };

    return displayNodes.reduce((areaToTest, displayNode) => {
      const a = areaToTest.map(a => newAreasForDisplayNodes(a, displayNode));
      return a.flat();
    }, [area]);
  }
  /**
   * Get new areas considering controls that are covering it
   * @private
   * @param {[DisplayNode]} displayNode
   * @param {Object} area
   * @param {Object} origin
   * @return {Area}
   */


  newAreasForDisplayNode(displayNode, area, origin) {
    if (displayNode.displayData.isMarkup || _balsamiq.default.get(displayNode, "displayData.displayListItems[0].displayItemType") === "Text") return [area];
    const isDisplayNodeTheParentOfArea = area.controlID.toString().startsWith(displayNode.controlID.toString());
    if (isDisplayNodeTheParentOfArea) return [area];
    const nodeBounds = { ...displayNode.displayData.bounds,
      x: displayNode.displayData.bounds.x + origin.x,
      y: displayNode.displayData.bounds.y + origin.y
    };

    if (area && _types.default.testRectangleIntersect(nodeBounds, area.bounds)) {
      // create new areas if needed
      const intersectionRect = _types.default.rectanglesIntersect(nodeBounds, area.bounds);

      if (!_balsamiq.default.isEqual(intersectionRect, area.bounds)) {
        var rects = this.computeRectangles(area.bounds, intersectionRect);

        if (rects.length === 0) {
          return [];
        } else {
          return rects.map(r => ({
            bounds: r,
            target: area.target,
            controlID: area.controlID
          }));
        }
      } else {
        return [];
      }
    } else {
      return [area];
    }
  }

  removeOverlappingAreas(area, tempAreas) {
    // existing areas that will be removed because (partially) covered by this control
    var areasToRemove = []; // new areas that will be added because have higher z-order

    var areasToAdd = []; // check for intersections with existing areas

    for (let i = 0; i < tempAreas.length; i++) {
      if (_types.default.testRectangleIntersect(tempAreas[i].bounds, area.bounds)) {
        areasToRemove.push(i); // compute and add new areas

        var intersectionRect = _types.default.rectanglesIntersect(tempAreas[i].bounds, area.bounds);

        var rects = this.computeRectangles(tempAreas[i].bounds, intersectionRect);
        rects.forEach(function (rect) {
          areasToAdd.push({
            bounds: rect,
            target: tempAreas[i].target,
            controlID: tempAreas[i].controlID
          });
        }, this);
      }
    } // now remove and add areas


    for (let i = areasToRemove.length - 1; i >= 0; i--) {
      tempAreas.splice(areasToRemove[i], 1);
    }

    for (let i = 0; i < areasToAdd.length; i++) {
      tempAreas.push(areasToAdd[i]);
    }
  }

  processControl(globalX, globalY, controlData) {
    let myAreas = []; // Add Children first, since they are below the container (group, symbol)

    controlData.children.forEach(function (child) {
      let tempAreas = this.processControl(globalX + controlData.displayData.bounds.x, globalY + controlData.displayData.bounds.y, child);

      if (tempAreas != null) {
        tempAreas.forEach(function (area) {
          myAreas.push(area);
        });
      }
    }, this); // add all areas of this control to the list

    controlData.displayData.linkAreas.forEach(function (area) {
      let bounds = _types.default.newRectangle(globalX + controlData.displayData.bounds.x + area.bounds.x, globalY + controlData.displayData.bounds.y + area.bounds.y, area.bounds.width, area.bounds.height);

      let rotation = 0;

      if (area.angle) {
        const adjustedValues = _displayitem.default.rotateTextLinksBounds(bounds, area.angle);

        bounds = adjustedValues.bounds;
        rotation = adjustedValues.rotation;
      }

      if (!area.disabled) {
        myAreas.push({
          bounds: bounds,
          target: area.target,
          controlID: controlData.controlID,
          rotation: rotation
        });
      }
    }, this);

    if (controlData.displayData.cropMask != null) {
      // intersect the linkAreas with the crop rectangle
      let intersectBounds = { ...controlData.displayData.cropMask.rect
      };
      intersectBounds.x += globalX + controlData.displayData.bounds.x - controlData.displayData.cropOffset.x;
      intersectBounds.y += globalY + controlData.displayData.bounds.y - controlData.displayData.cropOffset.y;

      if (myAreas.length > 0) {
        myAreas.forEach(function (area) {
          let bounds = _types.default.rectanglesIntersect(area.bounds, intersectBounds);

          bounds.x += controlData.displayData.cropOffset.x;
          bounds.y += controlData.displayData.cropOffset.y;
          area.bounds = bounds;
        });
      }
    }

    return myAreas;
  }

  computeRectangles(ra, rc) {
    var rects = []; // coordinates of center rectangle

    var l = Math.max(ra.x, rc.x);
    var r = Math.min(ra.x + ra.width, rc.x + rc.width);
    var t = Math.max(ra.y, rc.y);
    var b = Math.min(ra.y + ra.height, rc.y + rc.height); // top left

    if (l > ra.x && t > ra.y) {
      rects.push(_types.default.newRectangle(ra.x, ra.y, l - ra.x, t - ra.y));
    } // top center


    if (r > l && t > ra.y) {
      rects.push(_types.default.newRectangle(l, ra.y, r - l, t - ra.y));
    } // top right


    if (ra.x + ra.width > r && t > ra.y) {
      rects.push(_types.default.newRectangle(r, ra.y, ra.x + ra.width - r, t - ra.y));
    } // left


    if (l > ra.x && b > t) {
      rects.push(_types.default.newRectangle(ra.x, t, l - ra.x, b - t));
    } // right


    if (ra.x + ra.width > r && b > t) {
      rects.push(_types.default.newRectangle(r, t, ra.x + ra.width - r, b - t));
    } // bottom left


    if (l > ra.x && ra.y + ra.height > b) {
      rects.push(_types.default.newRectangle(ra.x, b, l - ra.x, ra.y + ra.height - b));
    } // bottom center


    if (r > l && ra.y + ra.height > b) {
      rects.push(_types.default.newRectangle(l, b, r - l, ra.y + ra.height - b));
    } // bottom right


    if (ra.x + ra.width > r && ra.y + ra.height > b) {
      rects.push(_types.default.newRectangle(r, b, ra.x + ra.width - r, ra.y + ra.height - b));
    }

    return rects;
  }

}

exports.LinkAreas = LinkAreas;

function makeLinkAreasDisplayList(linkAreas, getDependencyFunc) {
  const arrowDisplayItems = linkAreas.reduce((arrowDisplayItems, linkArea) => {
    // arrow icon
    if (!linkArea.continueOnNextLine) {
      var whichArrow = "regular";

      if (linkArea.target === "action://back") {
        whichArrow = "back";
      } else if (linkArea.target.startsWith("http")) {
        whichArrow = "external";
      } else if (linkArea.target.startsWith("resource:")) {
        var plainResourceID = linkArea.target.substr(9);

        if (getDependencyFunc && !getDependencyFunc(plainResourceID)) {
          whichArrow = "broken";
        }
      }

      const arrowDisplayItemsToAdd = _displayitem.default.newLinkDisplayItems(whichArrow, linkArea.bounds, linkArea.angle);

      _displayitem.default.setRotation(arrowDisplayItemsToAdd, linkArea.angle);

      return arrowDisplayItems.concat(arrowDisplayItemsToAdd);
    } else {
      return arrowDisplayItems;
    }
  }, []);
  return arrowDisplayItems;
}

/***/ }),

/***/ "./js/core/graphics/path.js":
/*!**********************************!*\
  !*** ./js/core/graphics/path.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Path = exports.Breakline = void 0;
exports.newBreaklinePath = newBreaklinePath;
exports.newEllipsePath = newEllipsePath;
exports.newPath = newPath;
exports.newRectanglePath = newRectanglePath;
exports.newRoundedRectanglePath = newRoundedRectanglePath;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _commands = __webpack_require__(/*! core/graphics/commands */ "./js/core/graphics/commands.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var Native = __webpack_require__(/*! core/native */ "./js/core/native.js");

class RendererBaseCommands {
  constructor() {
    this.currentPtX = 0;
    this.currentPtY = 0;
  }

  moveTo(dataObj, p_x, p_y) {
    _balsamiq.default.assertNumber(p_x);

    _balsamiq.default.assertNumber(p_y);

    dataObj.data.push(_commands.GraphicCommands.MOVE, p_x, p_y);
    this.currentPtX = p_x;
    this.currentPtY = p_y;
  }

  lineTo(dataObj, p_x, p_y) {
    _balsamiq.default.assertNumber(p_x);

    _balsamiq.default.assertNumber(p_y);

    dataObj.data.push(_commands.GraphicCommands.LINE, p_x, p_y);
    this.currentPtX = p_x;
    this.currentPtY = p_y;
  }

  arc90(dataObj, p_rx, p_ry) {
    var rx = Math.abs(p_rx),
        ry = Math.abs(p_ry),
        rmin = Math.min(rx, ry);

    _balsamiq.default.assertPositiveNumber(rx);

    _balsamiq.default.assertPositiveNumber(ry); // draw a translated, scaled circular arc to simulate an elliptical arc


    var ep = {
      x: p_rx + this.currentPtX,
      y: p_ry + this.currentPtY
    };
    var d = {
      x: ep.x - this.currentPtX,
      y: ep.y - this.currentPtY
    };
    var sameSign = d.x > 0 && d.y > 0 || d.x < 0 && d.y < 0;
    var cp = {
      x: sameSign ? ep.x : this.currentPtX,
      y: sameSign ? this.currentPtY : ep.y
    };
    dataObj.data.push(_commands.GraphicCommands.ARC, cp.x, cp.y, ep.x, ep.y, rx, ry);
    this.currentPtX = ep.x;
    this.currentPtY = ep.y;
  }

  curveTo(dataObj, p_cx, p_cy, p_x, p_y) {
    var cp = {
      x: p_cx,
      y: p_cy
    };
    var ep = {
      x: p_x,
      y: p_y
    };
    dataObj.data.push(_commands.GraphicCommands.QUADRATIC, cp.x, cp.y, ep.x, ep.y);
    this.currentPtX = ep.x;
    this.currentPtY = ep.y;
  }

  close(dataObj) {
    dataObj.data.push(_commands.GraphicCommands.CLOSE);
    this.currentPtX = 0;
    this.currentPtY = 0;
  }

}

class SVGBaseCommands {
  moveTo(dataObj, p_x, p_y) {
    dataObj.data += "M " + p_x + "," + p_y + " ";
  }

  lineTo(dataObj, p_x, p_y) {
    _balsamiq.default.assertNumber(p_x);

    _balsamiq.default.assertNumber(p_y);

    dataObj.data += "L " + p_x + "," + p_y + " ";
  }

  arc90(dataObj, p_rx, p_ry) {
    _balsamiq.default.assertNumber(p_rx);

    _balsamiq.default.assertNumber(p_ry);

    dataObj.data += "a " + Math.abs(p_rx) + "," + Math.abs(p_ry) + " 0 0,1 " + p_rx + "," + p_ry + " ";
  }

  curveTo(dataObj, p_cx, p_cy, p_x, p_y) {
    _balsamiq.default.assertNumber(p_cx);

    _balsamiq.default.assertNumber(p_cy);

    _balsamiq.default.assertNumber(p_x);

    _balsamiq.default.assertNumber(p_y);

    dataObj.data += "Q " + p_cx + "," + p_cy + " " + p_x + "," + p_y + " ";
  }

  close(dataObj) {
    dataObj.data += "Z";
  }

}

class Path {
  /**
   *
   * @param {Number} p_width
   * @param {Number} p_height
   * @param {String?} p_data
   * @returns {Path}
   */
  constructor(p_width, p_height, p_data) {
    _balsamiq.default.assertPositiveNumber(p_width);

    _balsamiq.default.assertPositiveNumber(p_height);

    this.width = p_width;
    this.height = p_height;
    let nativePaths = Native.Host.Paths;

    if (nativePaths.getFormat() == nativePaths.Formats.RENDERER_COMMANDS) {
      this.data = [];
      this.baseCommandsHelper = new RendererBaseCommands();
    } else {
      this.data = _balsamiq.default.isString(p_data) ? p_data : "";
      this.baseCommandsHelper = new SVGBaseCommands();
    }
  }

  moveTo(p_x, p_y) {
    this.baseCommandsHelper.moveTo.call(this.baseCommandsHelper, this, p_x, p_y);
  }

  lineTo(p_x, p_y) {
    this.baseCommandsHelper.lineTo.call(this.baseCommandsHelper, this, p_x, p_y);
  }

  arc90(p_rx, p_ry) {
    this.baseCommandsHelper.arc90.call(this.baseCommandsHelper, this, p_rx, p_ry);
  }

  curveTo(p_cx, p_cy, p_x, p_y) {
    this.baseCommandsHelper.curveTo.call(this.baseCommandsHelper, this, p_cx, p_cy, p_x, p_y);
  }

  close() {
    this.baseCommandsHelper.close.call(this.baseCommandsHelper, this);
  }

  addCircle(p_cx, p_cy, p_r, p_thickness) {
    _balsamiq.default.assertNumber(p_cx);

    _balsamiq.default.assertNumber(p_cy);

    _balsamiq.default.assertNumber(p_r);

    _balsamiq.default.assertNumber(p_thickness);

    var t2 = p_thickness / 2;
    var r = p_r - t2;
    this.moveTo(p_cx - r, p_cy);
    this.arc90(r, -r);
    this.arc90(r, r);
    this.arc90(-r, r);
    this.arc90(-r, -r);
  }

  addRectangle(p_x, p_y, p_width, p_height, p_thickness) {
    var t2 = p_thickness / 2;
    this.moveTo(p_x + t2, p_y + t2);
    this.lineTo(p_x + p_width - t2, p_y + t2);
    this.lineTo(p_x + p_width - t2, p_y + p_height - t2);
    this.lineTo(p_x + t2, p_y + p_height - t2);
    this.close();
  }
  /**
   *
   * @param {Number} p_x
   * @param {Number} p_y
   * @param {Number} p_width
   * @param {Number} p_height
   * @param {Number} p_thickness
   * @param {Number} p_cornerRadius
   * @param {Number?} p_cornerRadiusY
   */


  addRoundedRectangle(p_x, p_y, p_width, p_height, p_thickness, p_cornerRadius, p_cornerRadiusY) {
    _balsamiq.default.assertNumber(p_x);

    _balsamiq.default.assertNumber(p_y);

    _balsamiq.default.assertPositiveNumber(p_width);

    _balsamiq.default.assertPositiveNumber(p_height);

    _balsamiq.default.assertNumber(p_thickness);

    _balsamiq.default.assertNumber(p_cornerRadius);

    if (!_balsamiq.default.isNumber(p_cornerRadiusY)) {
      p_cornerRadiusY = p_cornerRadius;
    }

    var t2 = p_thickness / 2;
    var rX = Math.min(p_width / 2, p_cornerRadius),
        rY = Math.min(p_height / 2, p_cornerRadiusY);
    var w = p_width,
        h = p_height;
    this.moveTo(p_x + rX + t2, p_y + t2);
    this.lineTo(p_x + w - rX - t2, p_y + t2);
    this.arc90(rX, rY);
    this.lineTo(p_x + w - t2, p_y + h - rY - t2);
    this.arc90(-rX, rY);
    this.lineTo(p_x + rX + t2, p_y + h - t2);
    this.arc90(-rX, -rY);
    this.lineTo(p_x + t2, p_y + rY + t2);
    this.arc90(rX, -rY);
  }

}
/**
 *
 * @param {Number} p_width
 * @param {Number} p_height
 * @param {String?} p_data
 * @returns {Path}
 */


exports.Path = Path;

function newPath(p_width, p_height, p_data) {
  return new Path(p_width, p_height, p_data);
}

;
/**
 *
 * @param {Number} p_width
 * @param {Number} p_height
 * @param {Number} p_thickness
 * @returns {Path}
 */

function newEllipsePath(p_width, p_height, p_thickness) {
  var t2 = p_thickness / 2;
  var rx = p_width / 2 - t2;
  var ry = p_height / 2 - t2;
  var cy = p_height / 2;
  var path = newPath(p_width, p_height);
  path.moveTo(t2, cy);
  path.arc90(rx, -ry);
  path.arc90(rx, ry);
  path.arc90(-rx, ry);
  path.arc90(-rx, -ry);
  return path;
}

;
/**
 *
 * @param {Number} p_width
 * @param {Number} p_height
 * @param {Number} p_thickness
 * @returns {Path}
 */

function newRectanglePath(p_width, p_height, p_thickness) {
  var t2 = p_thickness / 2;
  var path = new Path(p_width, p_height);
  path.moveTo(t2, t2);
  path.lineTo(p_width - t2, t2);
  path.lineTo(p_width - t2, p_height - t2);
  path.lineTo(t2, p_height - t2);
  path.close();
  return path;
}

;
/**
 *
 * @param {Number} p_width
 * @param {Number} p_height
 * @param {Number} p_thickness
 * @param {Number} p_cornerRadius
 * @param {Number?} p_cornerRadiusY
 * @returns {Path}
 */

function newRoundedRectanglePath(p_width, p_height, p_thickness, p_cornerRadius, p_cornerRadiusY) {
  _balsamiq.default.assertPositiveNumber(p_width);

  _balsamiq.default.assertPositiveNumber(p_height);

  _balsamiq.default.assertNumber(p_thickness);

  _balsamiq.default.assertNumber(p_cornerRadius);

  if (!_balsamiq.default.isNumber(p_cornerRadiusY)) {
    p_cornerRadiusY = p_cornerRadius;
  }

  var t2 = p_thickness / 2;
  var rX = Math.min(p_width / 2, p_cornerRadius),
      rY = Math.min(p_height / 2, p_cornerRadiusY);
  var w = p_width,
      h = p_height;
  var path = new Path(p_width, p_height);
  path.moveTo(rX + t2, t2);
  path.lineTo(w - rX - t2, t2);
  path.arc90(rX, rY);
  path.lineTo(w - t2, h - rY - t2);
  path.arc90(-rX, rY);
  path.lineTo(rX + t2, h - t2);
  path.arc90(-rX, -rY);
  path.lineTo(t2, rY + t2);
  path.arc90(rX, -rY);
  return path;
}

; // Breaklines

const Breakline = {
  OPEN: "display_item",
  CLOSED_TOP: "mask_top",
  CLOSED_BOTTOM: "mask_bottom"
};
exports.Breakline = Breakline;
var BREAKLINE_UP_HDELTA = 18; // Width of top spike

var BREAKLINE_DOWN_HDELTA = 22; // Width of bottom spike

var BREAKLINE_TOTAL_WIDTH = BREAKLINE_UP_HDELTA * 2 + BREAKLINE_DOWN_HDELTA;
var BREAKLINE_MIN_GAP = 50;
var BREAKLINE_MIDDLE_GAP = 250;

function addBreaklineHeartbeat(p_path, p_x, p_y, p_thickness) {
  _balsamiq.default.assert(p_path instanceof Path);

  _balsamiq.default.assertNumber(p_x);

  _balsamiq.default.assertNumber(p_y);

  _balsamiq.default.assertNumber(p_thickness);

  var vOffset = _types.default.BREAKLINE_VERTICAL_OFFSET - p_thickness / 2; // Draw line up

  var nextX = p_x + BREAKLINE_UP_HDELTA;
  var nextY = p_y - vOffset;
  p_path.lineTo(nextX, nextY); // Draw line down

  nextX += BREAKLINE_DOWN_HDELTA;
  nextY = p_y + vOffset;
  p_path.lineTo(nextX, nextY); // Draw line up

  nextX += BREAKLINE_UP_HDELTA;
  nextY = p_y;
  p_path.lineTo(nextX, nextY);
}

; // Algorithm draws one breakline at 1/10 and the last at 9/10
// then tries to fit some breaklines in the middle
// If too small it fits just one breakline in the middle
// If no p_startPt and p_path provided, creates a path tall as the breakline
// If p_startPt and p_path provided, appends the breakline to the path starting to draw at p_startPt
// TODO as tall as the whole control

function newBreaklinePath(p_width, p_height, p_breakType, p_thickness, p_startPt, p_path) {
  _balsamiq.default.assertPositiveNumber(p_width);

  _balsamiq.default.assertEnum(p_breakType, Breakline);

  _balsamiq.default.assertNumber(p_thickness); // Do not use thickness for masks!


  if (p_breakType !== Breakline.OPEN) {
    p_thickness = 0;
  }

  var firstBreaklineStartX = p_width / 10;
  var firstBreaklineEndX = firstBreaklineStartX + BREAKLINE_TOTAL_WIDTH;
  var lastBreaklineStartX = p_width * 9 / 10 - BREAKLINE_TOTAL_WIDTH;
  var numBreaklines = Math.floor(p_width / firstBreaklineEndX);
  var centerSpace = lastBreaklineStartX - firstBreaklineEndX;
  var drawOneCenterBreakLine = !(numBreaklines === 0 || centerSpace >= BREAKLINE_MIN_GAP);
  var path, x, y;

  if (!_balsamiq.default.isUndefined(p_path) && !_balsamiq.default.isUndefined(p_startPt)) {
    // use old Path and coords
    x = p_startPt.x;
    y = p_startPt.y;
    path = p_path;
  } else {
    // Start a new Path
    x = 0;

    if (p_breakType === Breakline.CLOSED_BOTTOM) {
      y = p_height - _types.default.BREAKLINE_VERTICAL_OFFSET;
    } else if (p_breakType === Breakline.CLOSED_TOP) {
      y = _types.default.BREAKLINE_VERTICAL_OFFSET;
    } else {
      y = _types.default.BREAKLINE_VERTICAL_OFFSET;
    }

    path = new Path(p_width, _types.default.BREAKLINE_VERTICAL_OFFSET * 2);
    path.moveTo(x, y);
  }

  if (drawOneCenterBreakLine) {
    // Line to start of breakline
    x = Math.floor(p_width / 2 - BREAKLINE_TOTAL_WIDTH / 2);
    path.lineTo(x, y); // breakline

    addBreaklineHeartbeat(path, x, y, p_thickness);
  } else {
    // first horizontal segment
    x = firstBreaklineStartX;
    path.lineTo(x, y); // first breakline

    addBreaklineHeartbeat(path, x, y, p_thickness); // breaklines in the middle

    var numMiddleBreakLines = Math.floor(centerSpace / (BREAKLINE_MIDDLE_GAP + BREAKLINE_TOTAL_WIDTH));
    var i, multiplier;

    for (i = 0; i < numMiddleBreakLines; i += 1) {
      multiplier = (i + 1) / (numMiddleBreakLines + 1);
      x = firstBreaklineEndX + centerSpace * multiplier - BREAKLINE_TOTAL_WIDTH / 2;
      path.lineTo(x, y);
      addBreaklineHeartbeat(path, x, y, p_thickness);
    } // Last breakline


    path.lineTo(lastBreaklineStartX, y);
    addBreaklineHeartbeat(path, lastBreaklineStartX, y, p_thickness);
  } // Line to right end


  path.lineTo(p_width, y); // Make a closed path in the right direction if !OPEN

  if (p_breakType === Breakline.CLOSED_BOTTOM) {
    y = 0;
  } else if (p_breakType === Breakline.CLOSED_TOP) {
    y = p_height;
  }

  if (p_breakType !== Breakline.OPEN) {
    path.lineTo(p_width, y);
    path.lineTo(0, y);
    path.close();
  }

  return path;
}

;

/***/ }),

/***/ "./js/core/graphics/simplerenderer.js":
/*!********************************************!*\
  !*** ./js/core/graphics/simplerenderer.js ***!
  \********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
Object.defineProperty(exports, "GraphicCommands", ({
  enumerable: true,
  get: function () {
    return _commands.GraphicCommands;
  }
}));
exports.SimpleRenderer = exports.Modes = void 0;
exports.convertPathsToCommands = convertPathsToCommands;
exports.getDpiScaleFromCommands = getDpiScaleFromCommands;
exports.getImageSizeScaledByDpiFromCommands = getImageSizeScaledByDpiFromCommands;
exports.limitSizeAndScale = limitSizeAndScale;
exports.renderDisplayDataAsImage = renderDisplayDataAsImage;
exports.renderImage = renderImage;
exports.renderTextAsImage = renderTextAsImage;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _sketch = _interopRequireDefault(__webpack_require__(/*! core/skins/sketch */ "./js/core/skins/sketch.js"));

var _wireframe = _interopRequireDefault(__webpack_require__(/*! core/skins/wireframe */ "./js/core/skins/wireframe.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var _controls = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var _commands = __webpack_require__(/*! core/graphics/commands */ "./js/core/graphics/commands.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var Native = __webpack_require__(/*! core/native */ "./js/core/native.js");

const textMod = __webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js");

const Modes = {
  NORMAL: "normal",
  THUMBNAIL: "thumbnail"
};
exports.Modes = Modes;

function makeAbsoluteCoord(p_coord, p_relative, p_previousCoord) {
  return p_relative ? p_coord + p_previousCoord : p_coord;
}

function parseSVGPathData(p_data) {
  let r = /([MmHhLlCcVvAaZzQqSs]) ?((-?[0-9]*\.?[0-9]*([eE][-+]?[0-9]+)?),? ?)*/g;
  let m = p_data.match(r);
  let res = m ? m.map(command => {
    let c = command.charAt(0);
    let coordsToParse = command.substring(1);
    let m2 = coordsToParse.match(/(-?[0-9]*\.?[0-9]*([eE][-+]?[0-9]+)?)/g);
    let coords = [];
    m2.forEach(possibleCoord => {
      if (possibleCoord !== "") {
        coords.push(parseFloat(possibleCoord));
      }
    });
    return {
      command: c,
      coords: coords
    };
  }) : [];
  return res;
}

class SimpleRenderer {
  constructor(context) {
    _balsamiq.default.exists(context.host.TextMeterHelper);

    this.context = context;
    this.debug = false;
    this.showLinkAreas = false;
    this.showMarkup = true;
    this.skins = {};
    this.skins[_sketch.default["#skinName"]] = _sketch.default;
    this.skins[_wireframe.default["#skinName"]] = _wireframe.default;
    this.skin = null; // skin of the control being rendered

    this.commands = [];
  } // Options
  // -------


  setDebug(value) {
    this.debug = value;
  }

  setShowLinkAreas(value) {
    this.showLinkAreas = value;
  }

  setShowMarkup(value) {
    this.showMarkup = value;
  } // Render a displayData tree
  // -------------------------


  renderThumbnail(rootControl, maxSize) {
    let dpiScale = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 1;

    _balsamiq.default.assertNumber(dpiScale);

    return this.renderControls({
      rootControl,
      margins: [0, 0],
      maxSize: maxSize,
      dpiScale,
      whiteBackground: false
    });
  }

  renderFullSize(rootControl, whiteBackground) {
    let dpiScale = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 1;

    _balsamiq.default.exists(rootControl);

    _balsamiq.default.assertBool(whiteBackground);

    _balsamiq.default.assertPositiveNumber(dpiScale);

    return this.renderControls({
      rootControl,
      margins: [0, 0],
      whiteBackground,
      dpiScale
    });
  }

  renderMultipleControlsFullSize(controls, whiteBackground) {
    let dpiScale = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 1;

    _balsamiq.default.exists(controls);

    _balsamiq.default.assertBool(whiteBackground);

    _balsamiq.default.assertPositiveNumber(dpiScale);

    return this.renderMultipleControls({
      controls,
      margins: [0, 0],
      maxSize: -1,
      whiteBackground,
      dpiScale
    });
  }

  renderPage() {
    let {
      rootControl,
      scale,
      margins,
      controlIDsToRenderAsImages
    } = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    return this.renderControls({
      rootControl,
      margins,
      scale,
      whiteBackground: false,
      controlIDsToRenderAsImages
    });
  }

  renderForView(rootControl, scale) {
    return this.renderControls({
      rootControl,
      margins: null,
      scale,
      whiteBackground: true
    });
  } // Private methods
  // ---------------


  calcControlsClipBBox(p_controls, p_showMarkup) {
    var bboxes = p_controls.map(control => {
      return control.getClipBBox(p_showMarkup);
    });
    return _types.default.rectangleListUnion(bboxes);
  }

  calcControlsUserBBox(p_controls, p_showMarkup) {
    var bboxes = p_controls.map(control => {
      return control.getUserBBox(p_showMarkup);
    });
    return _types.default.rectangleListUnion(bboxes);
  }

  drawWhiteBackground(size) {
    this.drawRectangle(0, 0, size.width, size.height);
    this.draw(0xFFFFFF, 1, null, null);
  }

  renderMultipleControls(_ref) {
    let {
      controls,
      margins,
      maxSize,
      scale = 1,
      whiteBackground,
      dpiScale = 1
    } = _ref;

    _balsamiq.default.assertArray(controls);

    (0, _controls.assertControlsInSameGroup)(controls);

    if (margins) {
      _balsamiq.default.assertArray(margins);
    }

    _balsamiq.default.assertSizeNumber(maxSize);

    _balsamiq.default.assertPositiveNumber(scale);

    _balsamiq.default.assertBool(whiteBackground);

    this.commands = [];
    var clipBBox = this.calcControlsClipBBox(controls, this.showMarkup);
    var width = margins ? clipBBox.width + margins[0] : clipBBox.width;
    var height = margins ? clipBBox.height + margins[1] : clipBBox.height;
    var contentsScale = 1;

    if (maxSize !== -1) {
      contentsScale = maxSize / Math.max(width, height, maxSize);
    }

    var imageWidth = Math.max(Math.round(width * scale * contentsScale), 1);
    var imageHeight = Math.max(Math.round(height * scale * contentsScale), 1);
    var data = {
      imageWidth,
      imageHeight,
      scale
    };
    limitSizeAndScale(data, dpiScale, Native.Host.Images.getMaxImageSize());
    imageWidth = data.imageWidth;
    imageHeight = data.imageHeight;
    scale = data.scale;
    this.init(imageWidth, imageHeight, dpiScale);
    this.save();
    this.lineCap("butt");
    this.lineWidth(1);

    if (whiteBackground) {
      this.drawWhiteBackground({
        width: imageWidth,
        height: imageHeight
      });
    }

    var userBBox = this.calcControlsUserBBox(controls, this.showMarkup);
    this.scale(scale * contentsScale, scale * contentsScale);
    this.translate(userBBox.x - clipBBox.x, userBBox.y - clipBBox.y);

    if (!this.showMarkup) {
      // retranslate back if needed
      let userBBoxWithMarkup = this.calcControlsUserBBox(controls, true);
      this.translate(userBBoxWithMarkup.x - userBBox.x, userBBoxWithMarkup.y - userBBox.y);
    } // Translate to leave margin


    if (margins) {
      this.translate(margins[0] / 2, margins[1] / 2);
    } // Translate to normalize controls coordinates


    this.translate(-userBBox.x, -userBBox.y);
    controls.forEach(control => {
      this.renderControl(control);
    });
    this.restore();
    return this.commands;
  }

  renderControls() {
    let {
      rootControl,
      margins,
      maxSize = -1,
      scale = 1,
      dpiScale = 1,
      whiteBackground,
      controlIDsToRenderAsImages
    } = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

    _balsamiq.default.exists(rootControl);

    if (margins) {
      _balsamiq.default.assertArray(margins);
    }

    _balsamiq.default.assertSizeNumber(maxSize);

    _balsamiq.default.assertPositiveNumber(scale);

    _balsamiq.default.assertBool(whiteBackground);

    this.commands = [];
    var rootBounds = rootControl.getClipBBox(this.showMarkup);
    var width = margins ? rootBounds.width + margins[0] : rootBounds.width;
    var height = margins ? rootBounds.height + margins[1] : rootBounds.height;
    let contentsScale = 1;

    if (maxSize !== -1) {
      contentsScale = maxSize / Math.max(width, height, maxSize);
    }

    var imageWidth = Math.max(Math.ceil(width * scale * contentsScale), 1);
    var imageHeight = Math.max(Math.ceil(height * scale * contentsScale), 1);
    var data = {
      imageWidth,
      imageHeight,
      scale
    };
    limitSizeAndScale(data, dpiScale, Native.Host.Images.getMaxImageSize());
    imageWidth = data.imageWidth;
    imageHeight = data.imageHeight;
    scale = data.scale;
    this.init(imageWidth, imageHeight, dpiScale);
    this.save();
    this.lineCap("butt");
    this.lineWidth(1);

    if (whiteBackground) {
      this.drawRectangle(0, 0, imageWidth, imageHeight);
      this.draw(0xFFFFFF, 1, null, null);
    }

    var userBBox = rootControl.getUserBBox(this.showMarkup);
    this.scale(scale * contentsScale, scale * contentsScale);
    this.translate(userBBox.x - rootBounds.x, userBBox.y - rootBounds.y);

    if (!this.showMarkup) {
      // retranslate back if needed
      let userBBoxWithMarkup = rootControl.getUserBBox(true);
      this.translate(userBBoxWithMarkup.x - userBBox.x, userBBoxWithMarkup.y - userBBox.y);
    }

    this.renderControl(rootControl, margins, controlIDsToRenderAsImages);
    this.restore();
    return this.commands;
  }

  renderDisplayData(displayData, margins, cleanCommands) {
    if (cleanCommands) {
      this.commands = [];
    }

    this.save();

    if (!margins) {
      this.translate(displayData.bounds.x, displayData.bounds.y);
    } else {
      this.translate(margins[0], margins[1]);
    }

    if (this.debug) {
      this.drawRectangle(0, 0, displayData.bounds.width, displayData.bounds.height);
      this.draw(null, null, margins ? 0x666666 : 0xFF0000, 1);
    } // set the current skin for the display list items


    this.skin = this.skins[displayData.displayListSkinName]; // clip on control boundaries (if required)

    if (displayData.clipOnBbox) {
      this.begin();
      this.drawRectangle(displayData.clip.x - displayData.bounds.x, displayData.clip.y - displayData.bounds.y, displayData.clip.width, displayData.clip.height);
      this.clip();
    } // clip on crop rectangle


    if (displayData.useCropMask && displayData.cropMask) {
      this.begin();
      this.drawRectangle(displayData.cropMask.rect.x, displayData.cropMask.rect.y, displayData.cropMask.rect.width, displayData.cropMask.rect.height);
      this.clip();
      this.save();
      this.translate(displayData.cropOffset.x, displayData.cropOffset.y);
    } // display items


    displayData.displayListItems.forEach(displayItem => {
      this.save();

      if (this.debug) {
        this.drawRectangle(0, 0, displayData.bounds.width, displayData.bounds.height);
        this.draw(null, null, 0x0088FF, 1);
      } // mask


      if (displayItem.masks && displayItem.masks.length > 0) {
        displayItem.masks.forEach(maskID => {
          var mask = displayData.masks[maskID];

          if (mask.length > 0) {
            this.begin();
            mask.forEach(region => {
              if (region.type === "rect") {
                this.drawRectangle(region.rect.x, region.rect.y, region.rect.width, region.rect.height);
              } else if (region.type === "path") {
                this.drawPathData(region.path);
              }
            }); //var s = this.strokeStyle;
            //this.setStrokeStyle("rgb(255,0,0)";
            //this.strokePath();

            this.clip(); //this.setStrokeStyle(s;
          }
        });
      } // graphics


      if (displayItem.displayItemType === _displayitem.default.Type.TEXT) {
        this.renderTextDisplayItem(displayItem);
      } else if (displayItem.displayItemType === _displayitem.default.Type.SVG_PATH) {
        this.renderSVGPathDisplayItem(displayItem);
      } else if (displayItem.displayItemType === _displayitem.default.Type.SVG_PATHS_ASSET) {
        // this is the only case which requires a skin to be set
        _balsamiq.default.exists(this.skin, "can't set a valid skin object");

        this.renderSVGPathAssetDisplayItem(displayItem);
      } else if (displayItem.displayItemType === _displayitem.default.Type.IMAGE) {
        this.renderImage(displayItem);
      } else if (displayItem.displayItemType === _displayitem.default.Type.RECTANGLE) {
        this.renderRectangle(displayItem);
      }

      this.restore();
    }); // This produces graphic commands to render link areas

    if (this.showLinkAreas) {
      // Graphics commands to render the link areas graphically
      displayData.linkAreas.forEach(linkArea => {
        this.renderLinkArea(linkArea);
      });
    }

    if (displayData.useCropMask && displayData.cropMask) {
      this.restore();
    }

    return this.commands;
  }

  renderControl(p_control, margins, controlIDsToRenderAsImages) {
    var displayData = p_control.getDisplayData();

    if (displayData.isMarkup && !this.showMarkup) {
      return;
    }

    const nativeImageID = controlIDsToRenderAsImages ? controlIDsToRenderAsImages[p_control.coreObjectID.toString()] : undefined;

    if (nativeImageID) {
      const clipBBox = p_control.getClipBBox(this.showMarkup);
      this.renderImage({
        height: clipBBox.height,
        width: clipBBox.width,
        x: clipBBox.x,
        y: clipBBox.y,
        rotation: 0,
        sketchit: false,
        scaleX: 1,
        scaleY: 1,
        nativeImageID
      });
      return;
    } else {
      this.renderDisplayData(displayData, margins, false);
    }

    if (displayData.cropOffset) {
      this.translate(displayData.cropOffset.x, displayData.cropOffset.y);
    } // children controls


    p_control.children.forEach(childControl => {
      this.renderControl(childControl, undefined, controlIDsToRenderAsImages);
    });
    this.restore();
  }

  getLinkAreas(p_rootControl, p_offset, p_linkAreas) {
    var displayData = p_rootControl.getDisplayData();

    if (p_offset === undefined) {
      p_offset = {
        x: displayData.bounds.x,
        y: displayData.bounds.y,
        mainWidth: displayData.bounds.width + displayData.bounds.x,
        mainHeight: displayData.bounds.height + displayData.bounds.y
      };
    }

    if (p_linkAreas === undefined) {
      p_linkAreas = [];
    }

    displayData.linkAreas.forEach(linkArea => {
      var left = linkArea.bounds.x + p_offset.x;
      var top = linkArea.bounds.y + p_offset.y;
      var right = p_offset.mainWidth - left - linkArea.bounds.width;
      var bottom = p_offset.mainHeight - top - linkArea.bounds.height;
      p_linkAreas.push({
        left: left,
        top: top,
        right: right,
        bottom: bottom,
        name: linkArea.target
      });
    });
    p_rootControl.children.forEach(childControl => {
      var childDisplayData = childControl.getDisplayData();
      this.getLinkAreas(childControl, {
        x: p_offset.x + childDisplayData.bounds.x,
        y: p_offset.y + childDisplayData.bounds.y,
        mainWidth: p_offset.mainWidth,
        mainHeight: p_offset.mainHeight
      }, p_linkAreas);
    });
    return p_linkAreas;
  }

  renderLinkArea(p_linkArea) {
    this.save();
    this.begin();
    this.move(p_linkArea.bounds.x, p_linkArea.bounds.y);
    this.line(p_linkArea.bounds.x + p_linkArea.bounds.width, p_linkArea.bounds.y);
    this.line(p_linkArea.bounds.x + p_linkArea.bounds.width, p_linkArea.bounds.y + p_linkArea.bounds.height);
    this.line(p_linkArea.bounds.x, p_linkArea.bounds.y + p_linkArea.bounds.height);
    this.close();
    this.draw(0xFF0000, 0.5, null, null);
    this.restore();
  }

  renderTextDom(p_x, p_y, p_textDom, p_textStyle) {
    this.save(); //var t = textMod.textDom2AttributedText(p_textDom, p_textStyle);

    var x = p_x;
    p_textDom.forEach(p_node => {
      if (_balsamiq.default.isString(p_node)) {
        this.textStyle(p_textStyle.fontFamily, p_textStyle.fontSize, p_textStyle.bold, p_textStyle.italic, p_textStyle.underline, p_textStyle.textColor);
        var w = this.measureTextWidth(p_node);
        this.text(p_node, x, p_y);

        if (p_textStyle.underline || p_textStyle.strikethrough) {
          this.lineCap("round");
          this.lineWidth(Math.max(1, p_textStyle.fontSize * 0.065));

          if (p_textStyle.underline) {
            this.drawHorizontalLine(x, w, p_y + p_textStyle.fontSize * 0.15);
            this.draw(null, null, p_textStyle.textColor, 1);
          }

          if (p_textStyle.strikethrough) {
            this.drawHorizontalLine(x, w, p_y - p_textStyle.fontSize * 0.30);
            this.draw(null, null, p_textStyle.textColor, 1);
          }
        }

        x += w;
      } else {
        var s = {};

        if (p_node.type === "bold") {
          s = {
            bold: true
          };
        } else if (p_node.type === "italic") {
          s = {
            italic: true
          };
        } else if (p_node.type === "color") {
          s = {
            textColor: p_node.color
          };
        } else if (p_node.type === 'size') {
          s = {
            fontSize: p_node.size
          };
        } else if (p_node.type === "underline") {
          s = {
            underline: true
          };
        } else if (p_node.type === "strikethrough") {
          s = {
            strikethrough: true
          };
        } else if (p_node.type === "disabled") {
          s = {
            textColor: this.skin.default_disabled_color
          };
        } else if (p_node.type === "link") {
          s = {
            textColor: p_textStyle.linkColor,
            underline: true
          };
        }

        var newTextStyle = Object.assign({}, p_textStyle, s);
        var newX = this.renderTextDom(x, p_y, p_node.children, newTextStyle);
        x = newX;
      }
    }, this);
    this.restore();
    return x;
  }

  drawHorizontalLine(x, width, y) {
    this.begin();
    this.move(x, y);
    this.line(x + width, y);
  }

  drawRectangle(x, y, width, height, p_clockwise) {
    p_clockwise = _balsamiq.default.isBoolean(p_clockwise) ? p_clockwise : true;

    if (p_clockwise) {
      this.move(x, y);
      this.line(x + width, y);
      this.line(x + width, y + height);
      this.line(x, y + height);
      this.close();
    } else {
      this.move(x, y);
      this.line(x, y + height);
      this.line(x + width, y + height);
      this.line(x + width, y);
      this.close();
    }
  }

  drawSVGPathData(p_data) {
    let c = parseSVGPathData(p_data);
    let currentPoint = {
      x: 0,
      y: 0
    };
    let cpToReflect = null; // { x:, y:} of the second control point of a C/c or S/s command

    c.forEach(o => {
      var cmd = o.command;
      var lcmd = cmd.toLowerCase();
      var relative = cmd === lcmd;
      var cp, cp1, cp2, ep;

      if (lcmd !== 's' && lcmd !== 'c') {
        cpToReflect = null;
      }

      if (lcmd === 'm') {
        _balsamiq.default.assert(o.coords.length === 2);

        currentPoint.x = makeAbsoluteCoord(o.coords[0], relative, currentPoint.x);
        currentPoint.y = makeAbsoluteCoord(o.coords[1], relative, currentPoint.y);
        this.move(currentPoint.x, currentPoint.y);
      } else if (lcmd === 'z') {
        _balsamiq.default.assert(o.coords.length === 0);

        this.close();
      } else if (lcmd === 'l') {
        _balsamiq.default.assert(o.coords.length === 2);

        currentPoint.x = makeAbsoluteCoord(o.coords[0], relative, currentPoint.x);
        currentPoint.y = makeAbsoluteCoord(o.coords[1], relative, currentPoint.y);
        this.line(currentPoint.x, currentPoint.y);
      } else if (lcmd === 'h') {
        _balsamiq.default.assert(o.coords.length === 1);

        currentPoint.x = makeAbsoluteCoord(o.coords[0], relative, currentPoint.x);
        this.line(currentPoint.x, currentPoint.y);
      } else if (lcmd === 'v') {
        _balsamiq.default.assert(o.coords.length === 1);

        currentPoint.y = makeAbsoluteCoord(o.coords[0], relative, currentPoint.y);
        this.line(currentPoint.x, currentPoint.y);
      } else if (lcmd === 'c') {
        _balsamiq.default.assert(o.coords.length === 6);

        cp1 = {
          x: makeAbsoluteCoord(o.coords[0], relative, currentPoint.x),
          y: makeAbsoluteCoord(o.coords[1], relative, currentPoint.y)
        };
        cp2 = {
          x: makeAbsoluteCoord(o.coords[2], relative, currentPoint.x),
          y: makeAbsoluteCoord(o.coords[3], relative, currentPoint.y)
        };
        ep = {
          x: makeAbsoluteCoord(o.coords[4], relative, currentPoint.x),
          y: makeAbsoluteCoord(o.coords[5], relative, currentPoint.y)
        };
        this.bezier(cp1.x, cp1.y, cp2.x, cp2.y, ep.x, ep.y);
        currentPoint = ep;
        cpToReflect = cp2;
      } else if (lcmd === 'a') {
        _balsamiq.default.assert(o.coords.length === 7); // we support only elliptical arcs with x-axis aligned to global x-axis


        _balsamiq.default.assert(o.coords[2] === 0); // large-arc flag is supposed to be false


        _balsamiq.default.assert(o.coords[3] === 0);

        var rx = o.coords[0],
            ry = o.coords[1];

        _balsamiq.default.assertPositiveNumber(rx);

        _balsamiq.default.assertPositiveNumber(ry); // draw a translated, scaled circular arc to simulate an elliptical arc


        ep = {
          x: makeAbsoluteCoord(o.coords[5], relative, currentPoint.x),
          y: makeAbsoluteCoord(o.coords[6], relative, currentPoint.y)
        };
        var d = {
          x: ep.x - currentPoint.x,
          y: ep.y - currentPoint.y
        };
        var sameSign = d.x > 0 && d.y > 0 || d.x < 0 && d.y < 0;
        cp = {
          x: sameSign ? ep.x : currentPoint.x,
          y: sameSign ? currentPoint.y : ep.y
        };
        this.arc(cp.x, cp.y, ep.x, ep.y, rx, ry);
        currentPoint = ep;
      } else if (lcmd === 'q') {
        _balsamiq.default.assert(o.coords.length === 4);

        cp = {
          x: makeAbsoluteCoord(o.coords[0], relative, currentPoint.x),
          y: makeAbsoluteCoord(o.coords[1], relative, currentPoint.y)
        };
        ep = {
          x: makeAbsoluteCoord(o.coords[2], relative, currentPoint.x),
          y: makeAbsoluteCoord(o.coords[3], relative, currentPoint.y)
        };
        this.quadratic(cp.x, cp.y, ep.x, ep.y);
        currentPoint = ep;
      } else if (lcmd === 's') {
        _balsamiq.default.assert(o.coords.length === 4);

        if (cpToReflect) {
          cp1 = {
            x: currentPoint.x + (currentPoint.x - cpToReflect.x),
            y: currentPoint.y + (currentPoint.y - cpToReflect.y)
          };
        } else {
          cp1 = currentPoint;
        }

        cp2 = {
          x: makeAbsoluteCoord(o.coords[0], relative, currentPoint.x),
          y: makeAbsoluteCoord(o.coords[1], relative, currentPoint.y)
        };
        ep = {
          x: makeAbsoluteCoord(o.coords[2], relative, currentPoint.x),
          y: makeAbsoluteCoord(o.coords[3], relative, currentPoint.y)
        };
        this.bezier(cp1.x, cp1.y, cp2.x, cp2.y, ep.x, ep.y);
        currentPoint = ep;
        cpToReflect = cp2;
      } else {
        _balsamiq.default.assert(false, "this SVG path command is not supported: '" + cmd + "'");
      }
    });
  } // Private methods for Display Items
  // ---------------------------------


  renderTextDisplayItem(p_displayItem) {
    this.save();
    var textStyle = {
      bold: p_displayItem.bold,
      italic: p_displayItem.italic,
      underline: p_displayItem.underline,
      fontSize: p_displayItem.fontSize,
      fontFamily: p_displayItem.fontFamily,
      textColor: p_displayItem.textColor,
      linkColor: p_displayItem.linkColor
    };
    this.translate(p_displayItem.x, p_displayItem.y);
    this.rotate(-p_displayItem.rotation * Math.PI / 180);
    this.renderTextDom(0, p_displayItem.ascender, p_displayItem.textDom, textStyle);
    this.restore();
  }

  renderRectangle(p_displayItem) {
    this.save();
    this.lineWidth(_balsamiq.default.isNumber(p_displayItem.borderWidth) ? p_displayItem.borderWidth : 0);
    var fillColor = 0,
        fillAlpha = 0;

    if (_balsamiq.default.isNumber(p_displayItem.fillColor)) {
      fillColor = p_displayItem.fillColor;
      fillAlpha = p_displayItem.alpha;
    }

    if (p_displayItem.dashArray) {
      this.lineDash(p_displayItem.dashArray);
    }

    this.translate(p_displayItem.x, p_displayItem.y);
    this.rotate(-p_displayItem.rotation * Math.PI / 180);
    this.scale(p_displayItem.scaleX, p_displayItem.scaleY);
    this.drawRectangle(0, 0, p_displayItem.width, p_displayItem.height, true);
    var strokeColor = null,
        strokeAlpha = null;

    if (_balsamiq.default.isNumber(p_displayItem.borderWidth) && p_displayItem.borderWidth > 0) {
      this.lineWidth(p_displayItem.borderWidth);
      strokeColor = p_displayItem.borderColor;
      strokeAlpha = p_displayItem.borderAlpha != undefined ? p_displayItem.borderAlpha : 1;
    } else {
      this.lineWidth(0);
    }

    this.draw(fillColor, fillAlpha, strokeColor, strokeAlpha);
    this.restore();
  }

  renderImage(p_displayItem) {
    this.save();
    this.translate(p_displayItem.x, p_displayItem.y);
    this.rotate(-p_displayItem.rotation * Math.PI / 180);

    if (p_displayItem.sketchit) {
      this.sketchedImage(p_displayItem.nativeImageID, 0, 0, p_displayItem.width * p_displayItem.scaleX, p_displayItem.height * p_displayItem.scaleY);
    } else if (_balsamiq.default.isNumber(p_displayItem.colorize)) {
      this.colorizedImage(p_displayItem.nativeImageID, 0, 0, p_displayItem.width * p_displayItem.scaleX, p_displayItem.height * p_displayItem.scaleY, p_displayItem.colorize);
    } else {
      this.image(p_displayItem.nativeImageID, 0, 0, p_displayItem.width * p_displayItem.scaleX, p_displayItem.height * p_displayItem.scaleY);
    }

    this.restore();
  }

  renderSVGPathDisplayItem(p_displayItem) {
    this.save();
    this.lineWidth(_balsamiq.default.isNumber(p_displayItem.borderWidth) ? p_displayItem.borderWidth : 0);
    var fillColor = 0,
        fillAlpha = 0;

    if (_balsamiq.default.isNumber(p_displayItem.fillColor)) {
      fillColor = p_displayItem.fillColor;
      fillAlpha = p_displayItem.alpha;
    }

    this.lineCap(p_displayItem.linecap);

    if (p_displayItem.dashArray) {
      this.lineDash(p_displayItem.dashArray);
    }

    this.translate(p_displayItem.x, p_displayItem.y);
    this.rotate(-p_displayItem.rotation * Math.PI / 180);
    this.scale(p_displayItem.scaleX, p_displayItem.scaleY);
    this.begin();
    this.drawPathData(p_displayItem.pathData);
    var strokeColor = null,
        strokeAlpha = null;

    if (_balsamiq.default.isNumber(p_displayItem.borderWidth) && p_displayItem.borderWidth > 0) {
      this.lineWidth(p_displayItem.borderWidth);
      strokeColor = p_displayItem.borderColor;
      strokeAlpha = p_displayItem.borderAlpha != undefined ? p_displayItem.borderAlpha : 1;
    } else {
      this.lineWidth(0);
    }

    this.draw(fillColor, fillAlpha, strokeColor, strokeAlpha);
    this.restore();
  }

  drawPathData(p_pathData) {
    let nativePaths = Native.Host.Paths;

    if (nativePaths.getFormat() == nativePaths.Formats.RENDERER_COMMANDS) {
      _balsamiq.default.assertArray(p_pathData);

      this.commands = this.commands.concat(p_pathData);
    } else {
      _balsamiq.default.assertString(p_pathData);

      this.drawSVGPathData(p_pathData, this);
    }
  }

  renderSVGPathAssetDisplayItem(p_displayItem) {
    this.save();
    var asset = this.skin[p_displayItem.name];

    _balsamiq.default.assert(asset !== undefined, "the skin must have an asset called " + p_displayItem.name);

    _balsamiq.default.assert(asset.type === "SVG_PATHS", p_displayItem.name + " must be a SVG_PATHS asset");

    this.translate(p_displayItem.x, p_displayItem.y);
    this.rotate(-p_displayItem.rotation * Math.PI / 180);
    this.scale(p_displayItem.scaleX, p_displayItem.scaleY);
    let pathColors = [];
    let pathCount = this.skin[p_displayItem.name].paths.length; // Choose the right color:
    // 1. Overridden
    // 2. Skin Color
    // 3. displayItem fillColor

    for (let i = 0; i < pathCount; i += 1) {
      let pathColor = null;

      if (p_displayItem.overriddenColors && p_displayItem.overriddenColors[i] !== null) {
        pathColor = p_displayItem.overriddenColors[i];
      } else if (this.skin[p_displayItem.name].colors && this.skin[p_displayItem.name].colors[i] !== null) {
        pathColor = this.skin[p_displayItem.name].colors[i];
      } else {
        pathColor = p_displayItem.fillColor;
      }

      if (pathColor == null) {
        // This should never happen, but it does if the skin has some null colors(to be fixed!).
        // Let's work around it for the moment.
        pathColor = 0;
      }

      pathColors.push(pathColor);
    }

    this.drawPath(this.skin["#skinName"], p_displayItem.name, p_displayItem.alpha, pathColors);
    this.restore();
  } // Private methods for graphic commands
  // ------------------------------------


  init(width, height) {
    let dpiScale = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 1;

    _balsamiq.default.assertNumber(width);

    _balsamiq.default.assertNumber(height);

    _balsamiq.default.assertNumber(dpiScale);

    this.commands.push(_commands.GraphicCommands.INIT, width, height, dpiScale);
  }

  save() {
    this.commands.push(_commands.GraphicCommands.SAVE);
  }

  restore() {
    this.commands.push(_commands.GraphicCommands.RESTORE);
  }

  lineWidth(value) {
    _balsamiq.default.assertNumber(value);

    this.commands.push(_commands.GraphicCommands.LINE_WIDTH, value);
  }

  lineCap(value) {
    _balsamiq.default.assertString(value);

    this.commands.push(_commands.GraphicCommands.LINE_CAP, value);
  }

  lineDash(dashes) {
    _balsamiq.default.assertArray(dashes);

    this.commands.push(_commands.GraphicCommands.LINE_DASH, dashes.length);
    this.commands = this.commands.concat(dashes);
  }

  translate(dx, dy) {
    _balsamiq.default.assertNumber(dx);

    _balsamiq.default.assertNumber(dy);

    this.commands.push(_commands.GraphicCommands.TRANSLATE, dx, dy);
  }

  rotate(angle) {
    _balsamiq.default.assertNumber(angle);

    this.commands.push(_commands.GraphicCommands.ROTATE, angle);
  }

  scale(sx, sy) {
    _balsamiq.default.assertNumber(sx);

    _balsamiq.default.assertNumber(sy);

    this.commands.push(_commands.GraphicCommands.SCALE, sx, sy);
  }

  begin() {
    this.commands.push(_commands.GraphicCommands.BEGIN);
  }

  close() {
    this.commands.push(_commands.GraphicCommands.CLOSE);
  }

  move(x, y) {
    _balsamiq.default.assertNumber(x);

    _balsamiq.default.assertNumber(y);

    this.commands.push(_commands.GraphicCommands.MOVE, x, y);
  }

  line(x, y) {
    _balsamiq.default.assertNumber(x);

    _balsamiq.default.assertNumber(y);

    this.commands.push(_commands.GraphicCommands.LINE, x, y);
  }

  draw(fillColor, fillAlpha, strokeColor, strokeAlpha) {
    _balsamiq.default.assert(_balsamiq.default.isNumber(fillColor) || _balsamiq.default.isNull(fillColor));

    _balsamiq.default.assert(_balsamiq.default.isNumber(fillAlpha) || _balsamiq.default.isNull(fillAlpha));

    _balsamiq.default.assert(_balsamiq.default.isNumber(strokeColor) || _balsamiq.default.isNull(strokeColor));

    _balsamiq.default.assert(_balsamiq.default.isNumber(strokeAlpha) || _balsamiq.default.isNull(strokeAlpha));

    this.commands.push(_commands.GraphicCommands.DRAW, fillColor, fillAlpha, strokeColor, strokeAlpha);
  }

  clip() {
    this.commands.push(_commands.GraphicCommands.CLIP);
  }

  arc(x1, y1, x2, y2, rx, ry) {
    _balsamiq.default.assertNumber(x1);

    _balsamiq.default.assertNumber(y1);

    _balsamiq.default.assertNumber(x2);

    _balsamiq.default.assertNumber(y2);

    _balsamiq.default.assertNumber(rx);

    _balsamiq.default.assertNumber(ry);

    this.commands.push(_commands.GraphicCommands.ARC, x1, y1, x2, y2, rx, ry);
  }

  quadratic(x1, y1, x2, y2) {
    _balsamiq.default.assertNumber(x1);

    _balsamiq.default.assertNumber(y1);

    _balsamiq.default.assertNumber(x2);

    _balsamiq.default.assertNumber(y2);

    this.commands.push(_commands.GraphicCommands.QUADRATIC, x1, y1, x2, y2);
  }

  bezier(x1, y1, x2, y2, x3, y3) {
    _balsamiq.default.assertNumber(x1);

    _balsamiq.default.assertNumber(y1);

    _balsamiq.default.assertNumber(x2);

    _balsamiq.default.assertNumber(y2);

    _balsamiq.default.assertNumber(x3);

    _balsamiq.default.assertNumber(y3);

    this.commands.push(_commands.GraphicCommands.BEZIER, x1, y1, x2, y2, x3, y3);
  }

  textStyle(fontFamily, fontSize, bold, italic, underline, textColor) {
    this.commands.push(_commands.GraphicCommands.TEXT_STYLE, fontFamily, fontSize, bold, italic, underline, textColor);
    this.context.host.TextMeterHelper.setTextStyle(fontFamily, fontSize, bold, italic);
  }

  text(string, x, y) {
    this.commands.push(_commands.GraphicCommands.TEXT, string, x, y);
  }

  drawPath(skinName, pathName, fillAlpha, pathColors) {
    _balsamiq.default.assertString(skinName);

    _balsamiq.default.assertString(pathName);

    _balsamiq.default.assert(_balsamiq.default.isNumber(fillAlpha) || _balsamiq.default.isNull(fillAlpha));

    _balsamiq.default.assert(_balsamiq.default.isArray(pathColors)); // This command pushes a variable number of arguments since pathColors array length is variable.
    // The 5th parameter is an Int representing the number of following colors.
    // The total length of this command must thus be calculated as 5 + pathColors.length


    this.commands.push(_commands.GraphicCommands.DRAW_PATH, skinName, pathName, fillAlpha, pathColors.length, ...pathColors);
  }

  image(nativeImageID, x, y, width, height) {
    _balsamiq.default.assertString(nativeImageID);

    _balsamiq.default.assertNumber(x);

    _balsamiq.default.assertNumber(y);

    _balsamiq.default.assertNumber(width);

    _balsamiq.default.assertNumber(height);

    this.commands.push(_commands.GraphicCommands.IMAGE, nativeImageID, x, y, width, height);
  }

  sketchedImage(nativeImageID, x, y, width, height) {
    _balsamiq.default.assertString(nativeImageID);

    _balsamiq.default.assertNumber(x);

    _balsamiq.default.assertNumber(y);

    _balsamiq.default.assertNumber(width);

    _balsamiq.default.assertNumber(height);

    this.commands.push(_commands.GraphicCommands.SKETCHED_IMAGE, nativeImageID, x, y, width, height);
  }

  colorizedImage(nativeImageID, x, y, width, height, color) {
    _balsamiq.default.assertString(nativeImageID);

    _balsamiq.default.assertNumber(x);

    _balsamiq.default.assertNumber(y);

    _balsamiq.default.assertNumber(width);

    _balsamiq.default.assertNumber(height);

    _balsamiq.default.assertNumber(color);

    this.commands.push(_commands.GraphicCommands.COLORIZED_IMAGE, nativeImageID, x, y, width, height, color);
  } // Private methods for text width
  // ------------------------------


  measureTextWidth(string) {
    return this.context.host.TextMeterHelper.measureStringWidth(string);
  }

  static convertPathsToCommands(context, paths) {
    let renderer = new SimpleRenderer(context);
    return paths.map(path => {
      let res = path;
      renderer.drawSVGPathData(path);
      res = renderer.commands;
      renderer.commands = [];
      return res;
    });
  }

}

exports.SimpleRenderer = SimpleRenderer;

function renderImage(context, nativeImageID, size, whiteBackground) {
  _balsamiq.default.exists(size);

  _balsamiq.default.assertPositiveNumber(size.width);

  _balsamiq.default.assertPositiveNumber(size.height);

  _balsamiq.default.assertBool(whiteBackground);

  let imageWidth = Math.max(Math.ceil(size.width), 1);
  let imageHeight = Math.max(Math.ceil(size.height), 1);
  const simpleRenderer = new SimpleRenderer(context);
  simpleRenderer.init(imageWidth, imageHeight, 1);

  if (whiteBackground) {
    simpleRenderer.drawWhiteBackground(size);
  }

  simpleRenderer.image(nativeImageID, 0, 0, imageWidth, imageHeight);
  return simpleRenderer.commands;
}

function renderTextAsImage(_ref2, callback) {
  let {
    textStyle,
    context,
    text,
    margin
  } = _ref2;
  const textNote = textMod.newText(text, textStyle, context);
  const textNoteSize = textNote.measuredSize();
  const displayItems = textNote.makeDisplayItems({
    x: 0,
    y: 0,
    width: textNoteSize.width,
    height: textNoteSize.height
  });
  const textNoteDisplayData = {
    controlID: -1,
    bounds: {
      x: 0,
      y: 0,
      width: textNoteSize.width,
      height: textNoteSize.height
    },
    displayListItems: displayItems,
    displayListSkinName: "wireframe",
    masks: [],
    colorTransforms: null,
    clipOnBbox: false,
    linkAreas: false,
    isMarkup: false
  };
  const simpleRenderer = new SimpleRenderer(context);
  const scale = 2;
  simpleRenderer.init(textNoteSize.width * scale, textNoteSize.height * scale);
  simpleRenderer.save();
  simpleRenderer.scale(scale, scale);
  simpleRenderer.renderDisplayData(textNoteDisplayData, null, false);
  simpleRenderer.restore();
  context.host.Images.make(_ref3 => {
    let {
      id
    } = _ref3;
    context.host.Images.drawCommands(id, simpleRenderer.commands, () => {
      simpleRenderer.commands = [];
      simpleRenderer.image(id, 0, 0, textNoteSize.width, textNoteSize.height);
      callback([_commands.GraphicCommands.SAVE, _commands.GraphicCommands.TRANSLATE, margin.x, margin.y, ...simpleRenderer.commands, _commands.GraphicCommands.RESTORE]);
    });
  });
}

function convertPathsToCommands(paths) {
  let context = {
    host: {
      TextMeterHelper: {}
    }
  };
  let renderer = new SimpleRenderer(context);
  return paths.map(path => {
    let res = path;
    renderer.drawSVGPathData(path);
    res = renderer.commands;
    renderer.commands = [];
    return res;
  });
}

function limitSizeAndScale(data, dpiScale, maxSize) {
  let widthScaledByDPI = data.imageWidth * dpiScale;
  let heightScaledByDPI = data.imageHeight * dpiScale;

  if (widthScaledByDPI > maxSize.width || heightScaledByDPI > maxSize.height) {
    let widthScale = 1;
    let heightScale = 1;

    if (widthScaledByDPI > maxSize.width) {
      widthScale = maxSize.width / widthScaledByDPI;
    }

    if (heightScaledByDPI > maxSize.height) {
      heightScale = maxSize.height / heightScaledByDPI;
    }

    let scale = Math.min(widthScale, heightScale);
    data.imageWidth = Math.round(data.imageWidth * scale);
    data.imageHeight = Math.round(data.imageHeight * scale);
    data.scale = data.scale * scale;
  }
}

function getDpiScaleFromCommands(commands) {
  _balsamiq.default.assert(commands.length >= 3);

  _balsamiq.default.assertPositiveNumber(commands[3]);

  return commands[3];
}

function getImageSizeScaledByDpiFromCommands(imageCommands) {
  _balsamiq.default.assert(imageCommands.length >= 4);

  _balsamiq.default.assertPositiveNumber(imageCommands[1]);

  _balsamiq.default.assertPositiveNumber(imageCommands[2]);

  _balsamiq.default.assertPositiveNumber(imageCommands[3]);

  let dpiScale = imageCommands[3];
  return {
    width: imageCommands[1] * dpiScale,
    height: imageCommands[2] * dpiScale
  };
}

function renderDisplayDataAsImage(displayData, size, scale, context, callback) {
  const simpleRenderer = new SimpleRenderer(context);
  simpleRenderer.init(size.width * scale, size.height * scale);
  simpleRenderer.save();
  simpleRenderer.scale(scale, scale);
  simpleRenderer.translate(-displayData.bounds.x, -displayData.bounds.y);
  simpleRenderer.renderDisplayData(displayData, null, false);
  simpleRenderer.restore();
  context.host.Images.make(_ref4 => {
    let {
      id
    } = _ref4;
    context.host.Images.drawCommands(id, simpleRenderer.commands, () => {
      callback({
        id
      });
    });
  });
}

/***/ }),

/***/ "./js/core/keyboard.js":
/*!*****************************!*\
  !*** ./js/core/keyboard.js ***!
  \*****************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.keystrokeToString = exports.keystrokeFromString = exports.cmdSymbol = exports.altSymbol = exports.Keystroke = exports.Keys = exports.KEYSTROKE_TYPE = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Native = _interopRequireWildcard(__webpack_require__(/*! core/native */ "./js/core/native.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const KEYSTROKE_TYPE = {
  KEYDOWN: "keydown",
  KEYUP: "keyup"
};
exports.KEYSTROKE_TYPE = KEYSTROKE_TYPE;
let Keys = {
  Plus: '+',
  Minus: '-',
  Comma: ',',
  Dot: '.',
  Semicolon: ';',
  Equals: '=',
  SquareBracketLeft: '[',
  SquareBracketRight: ']',
  Canc: 'Delete',
  Zero: '0',
  One: '1',
  Two: '2',
  Three: '3',
  Four: '4',
  Five: '5',
  Six: '6',
  Seven: '7',
  Eight: '8',
  Nine: '9',
  A: 'a',
  B: 'b',
  C: 'c',
  D: 'd',
  E: 'e',
  F: 'f',
  G: 'g',
  H: 'h',
  I: 'i',
  J: 'j',
  K: 'k',
  L: 'l',
  M: 'm',
  N: 'n',
  O: 'o',
  P: 'p',
  R: 'r',
  S: 's',
  T: 't',
  U: 'u',
  V: 'v',
  W: 'w',
  X: 'x',
  Y: 'y',
  Z: 'z',
  Space: ' ',
  Slash: '/',
  Enter: 'Enter',
  Esc: 'Escape',
  Tab: 'Tab',
  Shift: 'Shift',
  Alt: 'Alt',
  Control: 'Control',
  Meta: 'Meta',
  Delete: 'Backspace',
  PgUp: 'PageUp',
  PgDown: 'PageDown',
  UpArrow: 'ArrowUp',
  DownArrow: 'ArrowDown',
  RightArrow: 'ArrowRight',
  LeftArrow: 'ArrowLeft',
  F2: 'F2'
};
exports.Keys = Keys;

class Keystroke {
  constructor(obj) {
    this.code = 0;
    this.cmd = false; // Command on Mac, CTRL on Win/Linux

    this.shift = false;
    this.alt = false;

    if (!obj) {
      return;
    }

    _balsamiq.default.assertString(obj.code);

    this.code = obj.code;
    this.shift = obj.shift == true || obj.shiftKey == true;
    this.alt = obj.alt == true || obj.altKey == true;

    if (Native.Host.runtime.isMac()) {
      this.cmd = obj.cmd == true || obj.meta == true || obj.metaKey == true;
    } else {
      this.cmd = obj.metaKey == true || obj.cmd == true || obj.ctrl == true || obj.ctrlKey == true;
    }
  }

  toString() {
    const a = [this.code];

    if (this.cmd) {
      a.push("c");
    }

    if (this.alt) {
      a.push("a");
    }

    if (this.shift) {
      a.push("s");
    }

    return a.join(".");
  }

  toObject() {
    var o = {
      code: this.code
    };

    if (this.cmd) {
      o.cmd = true;
    }

    if (this.alt) {
      o.alt = true;
    }

    if (this.shift) {
      o.shift = true;
    }

    return o;
  }

}

exports.Keystroke = Keystroke;

var keystrokeToString = function (data) {
  _balsamiq.default.exists(data);

  return new Keystroke(data).toString();
};

exports.keystrokeToString = keystrokeToString;

var keystrokeFromString = function (str) {
  _balsamiq.default.assertString(str);

  var s = str.split(".");

  if (str.startsWith("..")) {
    // fix the case for "dot" shortcuts
    s.shift();
    s.shift();
    s.unshift('.');
  }

  _balsamiq.default.assert(s.length >= 1);

  var k = new Keystroke();
  k.code = s.shift();
  k.cmd = s.indexOf("c") != -1;
  k.alt = s.indexOf("a") != -1;
  k.shift = s.indexOf("s") != -1;
  return k;
}; // Defaults to Ctrl+


exports.keystrokeFromString = keystrokeFromString;
const cmdSymbol = Native.Host.runtime.isMac() ? "⌘" : "Ctrl+"; // Defaults to Alt+

exports.cmdSymbol = cmdSymbol;
const altSymbol = Native.Host.runtime.isMac() ? "⌥" : "Alt+";
exports.altSymbol = altSymbol;

/***/ }),

/***/ "./js/core/license/licensecontroller.js":
/*!**********************************************!*\
  !*** ./js/core/license/licensecontroller.js ***!
  \**********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.LicenseController = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _constants = __webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts");

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class LicenseController {
  constructor(context, platformController) {
    _balsamiq.default.assertObject(platformController);

    _balsamiq.default.assertObject(context.host);

    _balsamiq.default.assertFunction(platformController.getLicenseData);

    _balsamiq.default.assertFunction(platformController.cleanUp);

    this.context = context;
    this.currentLicense = platformController;
    this.updateTimerFlag = false;
    this.timerID = null;
  }

  init(licenseDataChangedCallback) {
    _balsamiq.default.assertFunction(licenseDataChangedCallback);

    this.currentLicense.init((data, role, updateInterval, shouldUpdate) => {
      // This code has to invoke callback and to start update timer
      _balsamiq.default.assertObjectOrNull(data);

      if (!_balsamiq.default.isNull(role)) {
        _balsamiq.default.assertEnum(role, _constants.UserRoles);
      }

      _balsamiq.default.assertNumber(updateInterval);

      _balsamiq.default.assertBool(shouldUpdate);

      if (shouldUpdate === true && data) {
        licenseDataChangedCallback(data, role);
      }

      this.stopUpdateTimer();

      if (updateInterval > 0) {
        this.startUpdateTimer(updateInterval);
      }
    });
    this.getLicenseData();
  }

  startUpdateTimer(msec) {
    _balsamiq.default.assertNumber(msec); // Just in case


    this.stopUpdateTimer();
    this.updateTimerFlag = true;
    this.updateTimer(msec);
  }

  stopUpdateTimer() {
    this.updateTimerFlag = false;

    if (this.timerID) {
      this.context.host.clearTimer(this.timerID);
      this.timerID = null;
    }
  }

  updateTimer(msec) {
    _balsamiq.default.assertNumber(msec);

    this.timerID = this.context.host.timer(msec, () => {
      if (this.updateTimerFlag) {
        this.getLicenseData(); // Will invoke a new startTimer if needed

        this.context.host.log("LicenseController updated license!");
      }
    });
  }

  getLicenseData() {
    if (this.currentLicense) {
      this.currentLicense.getLicenseData(); // invokes callback
    }
  }

  cleanUp() {
    this.currentLicense && this.currentLicense.cleanUp();
    this.stopUpdateTimer();
  }

}

exports.LicenseController = LicenseController;

/***/ }),

/***/ "./js/core/license/licensedata.js":
/*!****************************************!*\
  !*** ./js/core/license/licensedata.js ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.LicenseData = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class LicenseData {
  constructor(obj) {
    _balsamiq.default.assertObjectOrNull(obj.shortLicenseMessage);

    _balsamiq.default.assertObjectOrNull(obj.longLicenseMessage);

    _balsamiq.default.assertObject(obj.readOnlyWarning);

    _balsamiq.default.assertString(obj.purchaseURL);

    _balsamiq.default.assertBool(obj.isValid); // Usually in About box


    this.aboutLicenseMessage = obj.aboutLicenseMessage; // Usually in trial bar

    this.longLicenseMessage = obj.longLicenseMessage; // Usually in open dialog location pick

    this.shortLicenseMessage = obj.shortLicenseMessage;
    this.shortLicenseMessageButtonText = _balsamiq.default.isString(obj.shortLicenseMessageButtonText) ? obj.shortLicenseMessageButtonText : null;
    this.shortLicenseMessageButtonLink = _balsamiq.default.isString(obj.shortLicenseMessageButtonLink) ? obj.shortLicenseMessageButtonLink : null;
    this.readOnlyWarning = obj.readOnlyWarning; // Color scheme for text and backgroud for various levels of warnings (trial expiring days, trial expired, plan issues…)

    this.messageColorSchemeEnum = obj.messageColorSchemeEnum; // Buy url

    this.purchaseURL = obj.purchaseURL; // Valid for editing

    this._isValid = obj.isValid;
  }

  hasLongMessage() {
    return _balsamiq.default.isObject(this.longLicenseMessage);
  }

  hasShortMessage() {
    return _balsamiq.default.isObject(this.shortLicenseMessage);
  }

  hasColorScheme() {
    return _balsamiq.default.isString(this.messageColorSchemeEnum) && this.messageColorSchemeEnum != "";
  }

  isValid() {
    return this._isValid;
  }

}

exports.LicenseData = LicenseData;

/***/ }),

/***/ "./js/core/mimetypes.js":
/*!******************************!*\
  !*** ./js/core/mimetypes.js ***!
  \******************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.WEBP_MIME = exports.TEXT_MIME = exports.SupportedImageTypes = exports.RTF_MIME = exports.PNG_MIME = exports.JPEG_MIME = exports.HTML_MIME = exports.GIF_MIME = exports.CUSTOM_MIME_TYPE = exports.BMPR_TYPE = exports.BMPR_MIMETYPE = exports.BMPRType = void 0;
exports.findEntryForExtension = findEntryForExtension;
exports.findEntryForMimeType = findEntryForMimeType;
exports.findMimetypeFromExtension = findMimetypeFromExtension;
exports.getExtensionAndMimeTypeFromBLOB = getExtensionAndMimeTypeFromBLOB;
exports.getExtensionFromBLOB = void 0;
exports.getExtensionFromMime = getExtensionFromMime;
exports.getImageExtensions = getImageExtensions;
exports.getImageWithMimeTypeFromBase64 = getImageWithMimeTypeFromBase64;
exports.getMimeFromDataAttribute = getMimeFromDataAttribute;
exports.isImageExtension = isImageExtension;
exports.isImageMimeType = isImageMimeType;
exports.isSupportedImageType = isSupportedImageType;
exports.isTextMimeType = isTextMimeType;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const TEXT_MIME = "text/plain";
exports.TEXT_MIME = TEXT_MIME;
const HTML_MIME = "text/html";
exports.HTML_MIME = HTML_MIME;
const RTF_MIME = "text/rtf";
exports.RTF_MIME = RTF_MIME;
const JPEG_MIME = "image/jpeg";
exports.JPEG_MIME = JPEG_MIME;
const GIF_MIME = "image/gif";
exports.GIF_MIME = GIF_MIME;
const PNG_MIME = "image/png";
exports.PNG_MIME = PNG_MIME;
const WEBP_MIME = "image/webp";
exports.WEBP_MIME = WEBP_MIME;
const BMPR_TYPE = "vnd.balsamiq.bmpr";
exports.BMPR_TYPE = BMPR_TYPE;
const BMPR_MIMETYPE = `application/${BMPR_TYPE}`;
exports.BMPR_MIMETYPE = BMPR_MIMETYPE;
const CUSTOM_MIME_TYPE = "__unknown__/";
exports.CUSTOM_MIME_TYPE = CUSTOM_MIME_TYPE;
const Table = {
  PDF: {
    mimetypes: ["application/pdf"],
    extensions: ["pdf"],
    iconID: "file-pdf-o"
  },
  JPEG: {
    base64Headers: ["/9j/"],
    // 0xFF 0xD8 0xFF -> https://en.wikipedia.org/wiki/JPEG#Syntax_and_structure
    mimetypes: [JPEG_MIME, "image/pjpeg"],
    extensions: ["jpg", "jpeg", "jpe"],
    iconID: "file-image-o"
  },
  GIF: {
    base64Headers: ["R0lGOD"],
    mimetypes: [GIF_MIME],
    extensions: ["gif"],
    iconID: "file-image-o"
  },
  WEBP: {
    base64Headers: ["UklGR"],
    mimetypes: [WEBP_MIME],
    extensions: ["webp"],
    iconID: "file-image-o"
  },
  PNG: {
    base64Headers: ["iVBORw"],
    mimetypes: [PNG_MIME],
    extensions: ["png"],
    iconID: "file-image-o"
  },
  TEXT: {
    mimetypes: ["text/plain"],
    extensions: ["txt"],
    iconID: "file-text-o"
  },
  BMPR: {
    mimetypes: [BMPR_MIMETYPE],
    extensions: ["bmpr"],
    iconID: "file-archive-o"
  },
  JSON: {
    mimetypes: ["application/json"],
    extensions: ["txt"],
    iconID: "file-text-o"
  }
};
const SupportedImageTypes = [Table.JPEG, Table.GIF, Table.PNG];
exports.SupportedImageTypes = SupportedImageTypes;
const BMPRType = Table.BMPR;
exports.BMPRType = BMPRType;
const TextMimeTypes = [...Table.JSON.mimetypes, ...Table.TEXT.mimetypes];

function isTextMimeType(mimeType) {
  if (mimeType) {
    _balsamiq.default.assertString(mimeType);

    mimeType = mimeType.toLowerCase();
    return TextMimeTypes.includes(mimeType);
  }

  return false;
}

function isImageExtension(extension) {
  if (typeof extension === 'string') {
    extension = extension.toLowerCase();
    return SupportedImageTypes.some(imageType => imageType.extensions.includes(extension));
  }

  return false;
}

function isImageMimeType(mimeType) {
  if (mimeType) {
    _balsamiq.default.assertString(mimeType);

    mimeType = mimeType.toLowerCase();
    return SupportedImageTypes.some(imageType => imageType.mimetypes.includes(mimeType));
  }

  return false;
}

function getImageExtensions() {
  return SupportedImageTypes.reduce((extensions, imageType) => extensions.concat(imageType.extensions), []);
}

function findMimetypeFromExtension(extension) {
  const imageType = findEntryForExtension(extension);
  return imageType && imageType.mimetypes[0];
}

function findEntryForExtension(extension) {
  if (typeof extension === 'string') {
    extension = extension.toLowerCase();

    for (const imageType of Object.values(Table)) {
      if (imageType.extensions.includes(extension)) {
        return imageType;
      }
    }
  }

  return null;
}

function findEntryForMimeType(mimetype) {
  if (typeof mimetype === 'string') {
    // Search in known mimetypes
    mimetype = mimetype.toLowerCase();

    for (const imageType of Object.values(Table)) {
      if (imageType.mimetypes.includes(mimetype)) {
        return imageType;
      }
    } // Search in known extensions


    const extension = mimetype.split("/")[1];

    for (const imageType of Object.values(Table)) {
      if (imageType.extensions.includes(extension)) {
        return imageType;
      }
    }
  }

  return null;
}

const checkImageType = (magicNumber, base64Headers) => base64Headers.some(header => magicNumber.startsWith(header));

function getExtensionAndMimeTypeFromBLOB(data) {
  _balsamiq.default.assertString(data);

  const magicNumber = data.substring(0, 6);
  const imageType = SupportedImageTypes.find(imageType => checkImageType(magicNumber, imageType.base64Headers));
  const mimeType = imageType && imageType.mimetypes[0];
  const extension = imageType && imageType.extensions[0];
  return {
    extension,
    mimeType
  };
}

const getExtensionFromBLOB = data => getExtensionAndMimeTypeFromBLOB(data).extension;

exports.getExtensionFromBLOB = getExtensionFromBLOB;

function isSupportedImageType(data) {
  const magicNumber = data.substring(0, 6);
  return SupportedImageTypes.some(imageType => checkImageType(magicNumber, imageType.base64Headers));
}

function getMimeFromDataAttribute(data) {
  _balsamiq.default.assertString(data);

  const mimeExp = /^data:(.*?);/;
  const match = mimeExp.exec(data);
  return match ? match[1] : null;
}

function getExtensionFromMime(mimeType) {
  _balsamiq.default.assertString(mimeType);

  const imageType = findEntryForMimeType(mimeType);
  return imageType && imageType.extensions[0];
}

function getImageWithMimeTypeFromBase64(data) {
  _balsamiq.default.assertString(data);

  const extension = getExtensionFromBLOB(data);
  const mime = findMimetypeFromExtension(extension);
  return "data:" + mime + ";base64," + data;
}

/***/ }),

/***/ "./js/core/model/archivestate.js":
/*!***************************************!*\
  !*** ./js/core/model/archivestate.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.KeyNames = exports.ErrorTypes = exports.ArchiveState = exports.ARCHIVE_STATE_ID_TYPE = void 0;
exports.assertArchiveStateType = assertArchiveStateType;
exports.makeArchiveStateCoreObjectID = makeArchiveStateCoreObjectID;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var CoreObject = _interopRequireWildcard(__webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const ARCHIVE_STATE_ID_TYPE = "archiveState";
exports.ARCHIVE_STATE_ID_TYPE = ARCHIVE_STATE_ID_TYPE;

function makeArchiveStateCoreObjectID(projectCoreObjectID) {
  CoreObject.assertProjectType(projectCoreObjectID);
  var newID = projectCoreObjectID.clone();
  newID.type = ARCHIVE_STATE_ID_TYPE;
  return newID;
}

function assertArchiveStateType(coreObjectID) {
  _balsamiq.default.assert(coreObjectID.type == ARCHIVE_STATE_ID_TYPE);
}

const ErrorTypes = {
  TRANSIENT: 'transient',
  PERMANENT: 'permanent'
};
exports.ErrorTypes = ErrorTypes;
const KeyNames = {
  // output
  AVAILABLE: 'available',
  ERROR: 'error',
  SHOULD_RELOAD: 'shouldReload',
  SYNCHRONIZED: 'synchronized'
};
exports.KeyNames = KeyNames;

class ArchiveState extends CoreObject.CoreObject {
  constructor(context, coreObjectID) {
    _balsamiq.default.exists(context);

    _balsamiq.default.exists(context.host);

    assertArchiveStateType(coreObjectID);
    super(context, coreObjectID);
    this._available = true;
    this._error = null; // { type: ErrorTypes, code: string optional, message: string optional }

    this._shouldReload = false;
    this._synchronized = null;
  }

  getKeyValue(keyName) {
    switch (keyName) {
      case KeyNames.AVAILABLE:
        return this._available;

      case KeyNames.ERROR:
        return this._error;

      case KeyNames.SHOULD_RELOAD:
        return this._shouldReload;

      case KeyNames.SYNCHRONIZED:
        return this._synchronized;

      default:
        return super.getKeyValue(keyName);
    }
  }

  setAvailable(value) {
    _balsamiq.default.assertBool(value);

    if (this._available != value) {
      this._available = value;
      this.context.host.log("ArchiveState available: " + value);
      this.notifyKeyValueChanged(KeyNames.AVAILABLE);
    }
  }

  setError(obj) {
    if (!_balsamiq.default.isEqual(this._error, obj)) {
      if (obj) {
        _balsamiq.default.assertEnum(obj.type, ErrorTypes);
      }

      this._error = obj;
      this.context.host.log("ArchiveState error: " + JSON.stringify(obj));
      this.notifyKeyValueChanged(KeyNames.ERROR);
    }
  }

  setShouldReload(value) {
    _balsamiq.default.assertBool(value);

    if (this._shouldReload != value) {
      this._shouldReload = value;
      this.context.host.log("ArchiveState shouldReload: " + value);
      this.notifyKeyValueChanged(KeyNames.SHOULD_RELOAD);
    }
  }

  setSynchronized(value) {
    _balsamiq.default.assert(_balsamiq.default.isNull(value) || _balsamiq.default.isBoolean(value));

    if (this._synchronized !== value) {
      this._synchronized = value;
      this.context.host.log("ArchiveState synchronized: " + value);
      this.notifyKeyValueChanged(KeyNames.SYNCHRONIZED);
    }
  }

}

exports.ArchiveState = ArchiveState;

/***/ }),

/***/ "./js/core/model/asset.js":
/*!********************************!*\
  !*** ./js/core/model/asset.js ***!
  \********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.KeyNames = exports.Asset = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var MimeTypes = _interopRequireWildcard(__webpack_require__(/*! core/mimetypes */ "./js/core/mimetypes.js"));

var _resource = __webpack_require__(/*! core/model/resource */ "./js/core/model/resource.js");

var _constants = __webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts");

var _BmprConstants = __webpack_require__(/*! @balsamiq/bmpr/lib/BmprConstants */ "./node_modules/@balsamiq/bmpr/lib/BmprConstants.js");

var _BalsamiqArchiveConstants = __webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchiveConstants */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js");

var _fontawesomemap = __webpack_require__(/*! core/skins/fontawesomemap */ "./js/core/skins/fontawesomemap.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const KeyNames = {
  FILE_SIZE: "fileSize"
};
exports.KeyNames = KeyNames;

class Asset extends _resource.Resource {
  /**
   *
   * @param {*} context
   * @param {import("core/core/coreobject").CoreObjectID} coreObjectID
   * @param {Object<string, any>} attributes
   * @param {(string|null)} data
   */
  constructor(context, coreObjectID, attributes) {
    let data = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : null;
    super(context, coreObjectID, _constants.ResourceTypes.ASSET);
    this.iconID = null;
    this.extension = null;
    this.data = data;
    /** @type {number|null} */

    this.fileSize = null;

    if (this.coreObjectID.branchID === _BalsamiqArchiveConstants.Branch.MasterBranchID) {
      this.attributes[_BmprConstants.AssetAttributes.Extension] = "";
    }

    this.setAttributes(attributes, null);
    this.updateFileSize();
    this.updateIconIDAndExtension();
  } // Resource methods
  // ----------------


  onAttributesChanged(attributes) {
    super.onAttributesChanged(attributes);
    this.updateIconIDAndExtension();
  }

  _checkAttribute(name, value) {
    if (name === _BmprConstants.ResourceAttributes.MimeType) {
      const mime = value;
      const entry = MimeTypes.findEntryForMimeType(mime);
      const extension = entry ? entry.extensions[0] : mime.split("/")[1];
      return {
        name: _BmprConstants.AssetAttributes.Extension,
        value: extension
      };
    }

    return super._checkAttribute(name, value);
  }

  hasData() {
    return this.data !== null;
  }

  getData() {
    return this.data;
  }

  onDataChanged(data) {
    _balsamiq.default.assertString(data);

    this.data = data;
    this.updateFileSize();
  }

  getIconID() {
    return this.iconID;
  }

  getPreviewData() {
    return this.iconID ? {
      iconID: this.iconID,
      iconText: (0, _fontawesomemap.getGlyphByIconName)(this.iconID)
    } : super.getPreviewData();
  }

  getFullName() {
    return MimeTypes.isImageExtension(this.attributes[_BmprConstants.AssetAttributes.Extension]) ? super.getName() : this.getFileName();
  }

  getFileName() {
    return this.extension ? super.getName() + "." + this.extension : super.getName();
  } // Key-value coding
  // ----------------


  getKeyValue(keyName) {
    switch (keyName) {
      case KeyNames.FILE_SIZE:
        return this.fileSize;

      default:
        return super.getKeyValue(keyName);
    }
  } // Private
  // -------


  updateIconIDAndExtension() {
    const extension = this.attributes[_BmprConstants.AssetAttributes.Extension];
    const entry = MimeTypes.findEntryForExtension(extension);

    if (entry) {
      this.iconID = entry.iconID;
      this.extension = entry.extensions[0];
    } else {
      this.iconID = "file-o";
      this.extension = extension;
    }
  }

  updateFileSize() {
    // Computing original size from base64 size https://developer.mozilla.org/en-US/docs/Glossary/Base64
    if (!this.data || !this.data.length) {
      this.fileSize = null;
      return;
    }

    const base64Size = this.data.length;
    this.fileSize = base64Size * 0.75; // Initial size expressed in Bytes
  }

  getMimeType() {
    const extension = this.attributes[_BmprConstants.AssetAttributes.Extension];
    const entry = MimeTypes.findEntryForExtension(extension);
    return entry ? entry.mimetypes[0] : "application/octet-stream";
  }

}

exports.Asset = Asset;

/***/ }),

/***/ "./js/core/model/branch.js":
/*!*********************************!*\
  !*** ./js/core/model/branch.js ***!
  \*********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.MasterAttributes = exports.EventTypes = exports.Branch = exports.AlternateAttributes = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _coreobject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var _events = _interopRequireDefault(__webpack_require__(/*! core/events */ "./js/core/events.ts"));

var _constants = __webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts");

var _BalsamiqArchiveConstants = _interopRequireDefault(__webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchiveConstants */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js"));

var _BmprConstants = __webpack_require__(/*! @balsamiq/bmpr/lib/BmprConstants */ "./node_modules/@balsamiq/bmpr/lib/BmprConstants.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var MASTER = _BalsamiqArchiveConstants.default.Branch.MasterBranchID;
var EventTypes = {
  BRANCH_CONTENTS_CHANGED: "branchContentsChanged",
  BRANCH_ATTRIBUTES_DIRTY: "branchAttributesDirty" // new attributes are available for save

}; //var KeyNames = {
//    NAME: "name"
//};

exports.EventTypes = EventTypes;
var MasterAttributes = {
  SKIN_NAME: _BmprConstants.BranchAttributes.SkinName,
  FONT_SIZE: _BmprConstants.BranchAttributes.FontSize,
  FONT_FACE: _BmprConstants.BranchAttributes.FontFace,
  SELECTION_COLOR: _BmprConstants.BranchAttributes.SelectionColor,
  PROJECT_DESCRIPTION: _BmprConstants.BranchAttributes.ProjectDescription,
  LINK_COLOR: _BmprConstants.BranchAttributes.LinkColor,
  SYMBOL_LIBRARY_ID: "symbolLibraryID"
};
exports.MasterAttributes = MasterAttributes;
var AlternateAttributes = {
  NAME: _BmprConstants.BranchAttributes.Name
};
exports.AlternateAttributes = AlternateAttributes;
const ObsoleteIgnoredAttributes = {
  MODIFIED_BY: "modifiedBy",
  CREATION_DATE: "creationDate",
  BRANCH_DESCRIPTION: "branchDescription"
};
const AlternateIgnoredAttributes = {
  SKIN_NAME: MasterAttributes.SKIN_NAME,
  FONT_SIZE: MasterAttributes.FONT_SIZE,
  FONT_FACE: MasterAttributes.FONT_FACE,
  SELECTION_COLOR: MasterAttributes.SELECTION_COLOR,
  PROJECT_DESCRIPTION: MasterAttributes.PROJECT_DESCRIPTION,
  LINK_COLOR: MasterAttributes.LINK_COLOR
};

class Branch extends _coreobject.CoreObject {
  constructor(context, coreObjectID, attributes) {
    (0, _coreobject.assertBranchType)(coreObjectID);

    _balsamiq.default.exists(attributes);

    super(context, coreObjectID);
    const initializedAttributes = { ...attributes,
      [MasterAttributes.PROJECT_DESCRIPTION]: attributes[MasterAttributes.PROJECT_DESCRIPTION] || ""
    };
    this.onAttributesChanged(initializedAttributes);
  } //// Key-value
  //// ---------
  //
  //getKeyValue(keyName) {
  //    if (keyName === KeyNames.NAME) {
  //        return this.getName();
  //    }
  //}
  // Getters
  // -------


  isMaster() {
    return this.coreObjectID.branchID === MASTER;
  }

  getAttributes() {
    return this.attributes;
  }

  getAttribute(name) {
    var ignore = this._checkAttribute(name, "");

    _balsamiq.default.assert(!ignore);

    return this.attributes[name];
  }

  getName() {
    if (this.isMaster()) {
      return "Official Version";
    } else {
      return this.getAttribute(AlternateAttributes.NAME);
    }
  } // Setters
  // -------


  setAttribute(name, value) {
    var ignore = this._checkAttribute(name, value);

    if (!ignore && this.attributes[name] !== value) {
      this.attributes[name] = value;
      var eventData = {};
      eventData[name] = value;
      eventData.operationOrigin = _constants.OperationOrigin.USER;
      this.context.eventCenter.fireEvent(_events.default.newCoreEvent(this.coreObjectID, EventTypes.BRANCH_CONTENTS_CHANGED, eventData));
      this.notifyAttributesDirty();
    }
  } // API for lower level
  // -------------------


  onAttributesChanged(attributes) {
    this.attributes = {};
    var eventData = {};

    for (var name in attributes) {
      var ignore = this._checkAttribute(name, attributes[name]);

      if (!ignore) {
        this.attributes[name] = attributes[name];
        eventData[name] = attributes[name];
      }
    }

    eventData.operationOrigin = _constants.OperationOrigin.ARCHIVE;
    this.context.eventCenter.fireEvent(_events.default.newCoreEvent(this.coreObjectID, EventTypes.BRANCH_CONTENTS_CHANGED, attributes));
  } // Private methods
  // ---------------


  notifyAttributesDirty() {
    this.context.eventCenter.fireEvent(_events.default.newCoreEvent(this.coreObjectID, EventTypes.BRANCH_ATTRIBUTES_DIRTY, {
      attributes: this.getAttributes()
    }));
  }

  _checkAttribute(name) {
    var ignore = false;

    if (this.isMaster()) {
      let ignored = Object.values(ObsoleteIgnoredAttributes).find(function (enumValue) {
        return enumValue === name;
      });

      if (ignored !== undefined) {
        ignore = true;
      } else {
        _balsamiq.default.assertEnum(name, MasterAttributes, "unknown attribute for Master branch: " + name);
      }
    } else {
      let ignored = Object.values(AlternateIgnoredAttributes).find(function (enumValue) {
        return enumValue === name;
      });

      if (ignored !== undefined) {
        ignore = true;
      } else {
        ignored = Object.values(ObsoleteIgnoredAttributes).find(function (enumValue) {
          return enumValue === name;
        });

        if (ignored !== undefined) {
          ignore = true;
        } else {
          _balsamiq.default.assertEnum(name, AlternateAttributes, "unknown attribute for Alternate branch: " + name);
        }
      }
    } //Balsamiq.exists(value); relax this assertion until we handle all possible "null" cases


    return ignore;
  }

}

exports.Branch = Branch;

/***/ }),

/***/ "./js/core/model/changetracker.js":
/*!****************************************!*\
  !*** ./js/core/model/changetracker.js ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = exports.ChangeTracker = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _BmprConstants = __webpack_require__(/*! @balsamiq/bmpr/lib/BmprConstants */ "./node_modules/@balsamiq/bmpr/lib/BmprConstants.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class ChangeTracker {
  constructor(logUserEvent) {
    this.changes = new Map();
    this.resourceName = new Map();
    this.logUserEvent = logUserEvent;
    this.wireframeJustAdded;
  }

  register(event) {
    _balsamiq.default.assertObject(event);

    const {
      name,
      shortResourceID
    } = event.resource;

    if (event.type === _BmprConstants.UserEventTypes.CreateResource) {
      this.wireframeJustAdded = event.resource.shortResourceID;
    }

    const previousName = this.resourceName.get(shortResourceID);

    if (previousName !== name) {
      this.resourceName.set(shortResourceID, name);
      this.notify(event);

      if (event.type === _BmprConstants.UserEventTypes.ChangeResource) {
        this.changes.set(event.id, true);
      }
    } else if (event.type !== _BmprConstants.UserEventTypes.ChangeResource) {
      this.notify(event);
    } else if (!this.changes.has(event.id)) {
      if (!this.wireframeJustAdded || this.wireframeJustAdded !== event.resource.shortResourceID) {
        this.changes.set(event.id, true);
        this.notify(event);
      }

      this.wireframeJustAdded = null;
    }
  }

  notify(event) {
    this.logUserEvent(event);
  }

  reset() {
    this.changes.clear();
  }

}

exports.ChangeTracker = ChangeTracker;
var _default = ChangeTracker;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/model/comment.js":
/*!**********************************!*\
  !*** ./js/core/model/comment.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.sortOlderFirst = exports.sortNewerFirst = exports.resourceCoreObjectIDMatchesCommentCoreObjectID = exports.makeCommentResourceCoreObjectID = exports.makeCommentCoreObjectID = exports.assertCommentType = exports.CommentCoreObjectType = exports.Comment = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var CoreObject = _interopRequireWildcard(__webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js"));

var _commentcalloutdata = __webpack_require__(/*! editor/viewcontrollers/commentcalloutdata */ "./js/editor/viewcontrollers/commentcalloutdata.js");

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var textUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var BarConstants = _interopRequireWildcard(__webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchiveConstants */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

/**
 *  @typedef { import("core/core/coreobject").CoreObjectID & { commentID: string } } CommentCoreObjectID
 */

/**
 * @typedef {object} CalloutData
 * @property {number} color
 * @property {string} id
 * @property {string} label
 * @property {number} labelColor
 * @property {string} tip
 * @property {number} x
 * @property {number} y
 */

/**
 * @typedef {object} Attributes
 * @property {string[]} likedBy
 * @property {string} parentID
 * @property {string[]} readBy
 * @property {number} timestamp
 * @property {{ [label: string]: string }} timestamps Object in the form { label: timestamp }
 * @property {boolean} trashed
 * @property {string} trashedBy
 * @property {boolean=} imported
 * @property {string=} originalAuthor
 */

/**
 * @typedef {object} CommentData
 * @property {string} id
 * @property {string} branchID
 * @property {string} resourceID
 * @property {{ text: string, callouts: CalloutData[] }} data
 * @property {string} userID
 * @property {Attributes} attributes
 */
class Comment extends CoreObject.CoreObject {
  /** @type {CommentCoreObjectID} */
  constructor(context, coreObjectID, attributes, userID, data) {
    super(context, coreObjectID);

    _defineProperty(this, "coreObjectID", void 0);

    assertCommentType(coreObjectID);

    _balsamiq.default.assertArray(attributes.readBy);

    _balsamiq.default.assertNumber(attributes.timestamp);

    _balsamiq.default.assertString(attributes.parentID);

    _balsamiq.default.assertBool(attributes.trashed);

    _balsamiq.default.assertString(attributes.trashedBy);

    _balsamiq.default.assertArray(attributes.likedBy);

    if (attributes.timestamps) {
      _balsamiq.default.assertObject(attributes.timestamps);
    }

    _balsamiq.default.assertObject(data);

    this.coreObjectID = coreObjectID;
    this.readBy = _balsamiq.default.clone(attributes.readBy);
    this.timestamp = attributes.timestamp;
    /** @type {string=} */

    this.parentID = _balsamiq.default.clone(attributes.parentID);
    this.trashed = attributes.trashed;
    this.trashedBy = _balsamiq.default.clone(attributes.trashedBy);
    this.likedBy = _balsamiq.default.clone(attributes.likedBy);
    this.timestamps = attributes.timestamps;
    this.imported = attributes.imported;
    /** @type {string=} */

    this.originalAuthor = attributes.originalAuthor;
    /** @type {{ text: string; callouts: unknown; }} */

    this.data = data;
    this.data.callouts = data.callouts ? data.callouts.map(_commentcalloutdata.makeCalloutRawObj) : [];
    /** @type {string} */

    this.userID = userID;
    this.updateAttributedText();
  }

  getID() {
    return this.coreObjectID.commentID;
  }

  getShortID() {
    return {
      shortCommentID: this.coreObjectID.commentID.substring(0, 4)
    };
  }

  matchesShortID(shortID) {
    return _balsamiq.default.isEqual(this.getShortID(), shortID);
  }

  getReadBy() {
    return this.readBy;
  }

  getTimestamp() {
    return this.timestamp;
  }

  getParentID() {
    return this.parentID;
  }

  isTrashed() {
    return this.trashed;
  }

  getTrashedBy() {
    return this.trashedBy;
  }

  getLikedBy() {
    return this.likedBy;
  }

  resourceIs(resourceCoreObjectID) {
    return resourceCoreObjectIDMatchesCommentCoreObjectID(resourceCoreObjectID, this.coreObjectID);
  }

  resourceIDBranchIDIs(resourceID, branchID) {
    return resourceIDBranchIDMatchesCommentCoreObjectID(resourceID, branchID, this.coreObjectID);
  }
  /**
   * @return {string | undefined}
   */


  getText() {
    return this.data && this.data.text ? this.data.text : "";
  }

  getPreviewText() {
    const text = this.getText();
    return text.length < 144 ? text : text.substring(0, 144) + "...";
  }

  getAttributedText() {
    return this.attributedText;
  }

  getTextUpdateTimestamp() {
    return this.timestamps && this.timestamps.text ? this.timestamps.text : -1;
  }

  updateText(text) {
    _balsamiq.default.assertString(text);

    if (!this.data) {
      this.data = {};
    }

    this.data.text = text;
    this.updateAttributedText();
  }

  updateAttributedText() {
    const textNodes = textUtils.stringToNodes(this.data && this.data.text ? this.data.text : "");
    this.attributedText = textUtils.textDom2AttributedText(textNodes, {
      fontFamily: "",
      fontSize: 0,
      textColor: 0x000000
    });
  }

  updateCallouts(callouts) {
    _balsamiq.default.assertArray(callouts);

    if (Array.isArray(callouts)) {
      callouts.forEach(callout => {
        (0, _commentcalloutdata.assertCallout)(callout);
      });
    }

    if (!this.data) {
      this.data = {};
    }

    this.data.callouts = callouts;
  }

  getData() {
    return _balsamiq.default.cloneDeep(this.data);
  }

  getOwnCallouts() {
    return this.data && this.data.callouts ? _balsamiq.default.cloneDeep(this.data.callouts) : [];
  }

  setData(data) {
    _balsamiq.default.assertObject(data);

    let changed = !this.equivalentDataSets(this.data, data);

    if (changed) {
      this.data = data;
      this.updateAttributedText();
    }

    return changed;
  }
  /**
   * @return {string}
   */


  getUserID() {
    return this.userID;
  }

  getAttributes() {
    return _lodash.default.pick(this, ["readBy", "timestamp", "parentID", "trashed", "trashedBy", "likedBy", "timestamps", "imported", "originalAuthor"]);
  }

  updateAttributes(attributes) {
    _balsamiq.default.assertObject(attributes);

    let changed = !this.equivalentAttributeSets(this, attributes);

    if (changed) {
      this.readBy = attributes[BarConstants.CommentAttributes.ReadBy];
      this.timestamp = attributes[BarConstants.CommentAttributes.Timestamp];
      this.parentID = attributes[BarConstants.CommentAttributes.ParentID];
      this.trashed = attributes[BarConstants.CommentAttributes.Trashed];
      this.trashedBy = attributes[BarConstants.CommentAttributes.TrashedBy];
      this.likedBy = attributes[BarConstants.CommentAttributes.LikedBy];
      this.timestamps = attributes[BarConstants.CommentAttributes.Timestamps];
      this.imported = attributes[BarConstants.CommentAttributes.Imported];
      this.originalAuthor = attributes[BarConstants.CommentAttributes.OriginalAuthor];
    }

    return changed;
  }

  setTrashed(trashed) {
    _balsamiq.default.assertBool(trashed);

    this.trashed = trashed;
  }

  setTrashedBy(userID) {
    _balsamiq.default.assertString(userID);

    this.trashedBy = userID;
  }

  isLikedBy(userID) {
    return this.getLikedBy().find(myUserID => {
      return myUserID === userID;
    }) != undefined;
  }

  isReadBy(userID) {
    return this.getReadBy().find(myUserID => {
      return myUserID === userID;
    }) != undefined;
  }

  setLikedBy(userID, liked) {
    _balsamiq.default.assertBool(liked);

    _balsamiq.default.assertString(userID);

    let index = this.likedBy.findIndex(userIDLiked => {
      return userIDLiked === userID;
    });

    if (liked && index == -1) {
      this.likedBy.push(userID);
    } else if (!liked && index != -1) {
      this.likedBy.splice(index, 1);
    }
  }

  setReadBy(userID, read) {
    _balsamiq.default.assertBool(read);

    _balsamiq.default.assertString(userID);

    let index = this.readBy.findIndex(userIDRead => {
      return userIDRead === userID;
    });

    if (read && index == -1) {
      this.readBy.push(userID);
    } else if (!read && index != -1) {
      this.readBy.splice(index, 1);
    }
  }
  /**
   *
   * @param {string} originalAuthor
   */


  setOriginalAuthor(originalAuthor) {
    _balsamiq.default.assertString(originalAuthor);

    this.imported = true;
    this.originalAuthor = originalAuthor;
  }

  equalStringArray(array1, array2) {
    _balsamiq.default.assertArrayOfStrings(array1);

    _balsamiq.default.assertArrayOfStrings(array2);

    if (array1.length !== array2.length) {
      return false;
    }

    let array1Ordered = array1.slice().sort();
    let array2Ordered = array2.slice().sort();

    for (let i = 0; i < array1Ordered.length; ++i) {
      if (array1Ordered[i] !== array2Ordered[i]) {
        return false;
      }
    }

    return true;
  }

  equivalentAttributeSets(attribute1, attribute2) {
    _balsamiq.default.assertObject(attribute1);

    _balsamiq.default.assertObject(attribute2);

    if (attribute1.timestamp !== attribute2.timestamp) {
      return false;
    }

    if (attribute1.parentID !== attribute2.parentID) {
      return false;
    }

    if (attribute1.trashed !== attribute2.trashed) {
      return false;
    }

    if (attribute1.trashedBy !== attribute2.trashedBy) {
      return false;
    }

    if (!this.equalStringArray(attribute1.readBy, attribute2.readBy)) {
      return false;
    }

    if (!this.equalStringArray(attribute1.likedBy, attribute2.likedBy)) {
      return false;
    }

    for (let key in attribute1.timestamps) {
      if (!(key in attribute2.timestamps)) {
        return false;
      }

      if (attribute1[key] !== attribute2[key]) {
        return false;
      }
    }

    for (let key in attribute2.timestamps) {
      if (!(key in attribute1.timestamps)) {
        return false;
      }
    }

    return true;
  }

  equivalentDataSets(data1, data2) {
    _balsamiq.default.assertObject(data1);

    _balsamiq.default.assertObject(data2);

    let text1 = data1.text ? data1.text : "";
    let text2 = data2.text ? data2.text : "";

    if (text1 !== text2) {
      return false;
    }

    let callouts1 = data1.callouts ? data1.callouts : [];
    let callouts2 = data2.callouts ? data2.callouts : [];

    if (callouts1.length !== callouts2.length) {
      return false;
    }

    let array1Ordered = callouts1.slice().sort((calloutData1, calloutData2) => {
      return calloutData1.id.localeCompare(calloutData1.id);
    });
    let array2Ordered = callouts2.slice().sort((calloutData1, calloutData2) => {
      return calloutData1.id.localeCompare(calloutData1.id);
    });
    return array1Ordered.every((_item, i) => (0, _commentcalloutdata.areEqual)(array1Ordered[i], array2Ordered[i]));
  }
  /**
   * @returns {CommentData}
   */


  toObject() {
    return {
      id: this.getID(),
      branchID: this.coreObjectID.branchID,
      resourceID: this.coreObjectID.resourceID,
      data: this.getData(),
      userID: this.getUserID(),
      attributes: this.getAttributes()
    };
  }

}
/**
 *
 * @param {import("core/core/coreobject").CoreObjectID} resourceCoreObjectID
 * @param {string} commentID
 * @return {CommentCoreObjectType}
 */


exports.Comment = Comment;

const makeCommentCoreObjectID = function (resourceCoreObjectID, commentID) {
  CoreObject.assertResourceType(resourceCoreObjectID);

  _balsamiq.default.assertString(commentID);

  return new CoreObject.CoreObjectID(CommentCoreObjectType, {
    siteID: resourceCoreObjectID.siteID,
    projectID: resourceCoreObjectID.projectID,
    branchID: resourceCoreObjectID.branchID,
    resourceID: resourceCoreObjectID.resourceID,
    commentID: commentID
  });
};

exports.makeCommentCoreObjectID = makeCommentCoreObjectID;

const makeCommentResourceCoreObjectID = function (commentCoreObjectID) {
  assertCommentType(commentCoreObjectID);
  return new CoreObject.CoreObjectID(CoreObject.CoreObjectTypes.RESOURCE, {
    siteID: commentCoreObjectID.siteID,
    projectID: commentCoreObjectID.projectID,
    branchID: commentCoreObjectID.branchID,
    resourceID: commentCoreObjectID.resourceID
  });
};

exports.makeCommentResourceCoreObjectID = makeCommentResourceCoreObjectID;
const CommentCoreObjectType = "Comment"; // expanding allowed coreobject types

exports.CommentCoreObjectType = CommentCoreObjectType;
CoreObject.CoreObjectTypes["COMMENT"] = CommentCoreObjectType;

const assertCommentType = function (coreObjectID) {
  CoreObject.assertType(coreObjectID, CommentCoreObjectType);
};

exports.assertCommentType = assertCommentType;

const resourceCoreObjectIDMatchesCommentCoreObjectID = function (resourceCoreObjectID, commentCoreObjectID) {
  return commentCoreObjectID.siteID === resourceCoreObjectID.siteID && commentCoreObjectID.projectID === resourceCoreObjectID.projectID && commentCoreObjectID.branchID === resourceCoreObjectID.branchID && commentCoreObjectID.resourceID === resourceCoreObjectID.resourceID;
};

exports.resourceCoreObjectIDMatchesCommentCoreObjectID = resourceCoreObjectIDMatchesCommentCoreObjectID;

var resourceIDBranchIDMatchesCommentCoreObjectID = function (resourceID, branchID, commentCoreObjectID) {
  return commentCoreObjectID.branchID === branchID && commentCoreObjectID.resourceID === resourceID;
};

const sortNewerFirst = (a, b) => {
  if (a.timestamp === -1) return -1;
  if (b.timestamp === -1) return 1;
  return b.timestamp - a.timestamp;
};
/**
 *
 * @param {Comment} a
 * @param {Comment} b
 * @return number
 */


exports.sortNewerFirst = sortNewerFirst;

const sortOlderFirst = (a, b) => {
  if (a.timestamp === -1) return 1;
  if (b.timestamp === -1) return -1;
  return a.timestamp - b.timestamp;
};

exports.sortOlderFirst = sortOlderFirst;

/***/ }),

/***/ "./js/core/model/connectedusers.js":
/*!*****************************************!*\
  !*** ./js/core/model/connectedusers.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.KeyNames = exports.EventTypes = exports.ConnectedUsers = exports.ConnectedUser = exports.CONNECTED_USERS_ID_TYPE = void 0;
exports.assertConnectedUsersType = assertConnectedUsersType;
exports.makeConnectedUsersCoreObjectID = makeConnectedUsersCoreObjectID;
exports.makeUserInfo = makeUserInfo;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var CoreObject = _interopRequireWildcard(__webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js"));

var Constants = _interopRequireWildcard(__webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts"));

var Events = _interopRequireWildcard(__webpack_require__(/*! core/events */ "./js/core/events.ts"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const CONNECTED_USERS_ID_TYPE = "connectedUsers";
exports.CONNECTED_USERS_ID_TYPE = CONNECTED_USERS_ID_TYPE;
const COMPUTER_USER_PREFIX = "computer-";

function makeUserInfo(userInfo) {
  let {
    userBelongsToProject = true,
    isComputerUser = false
  } = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  if (!userInfo || userInfo.email === "" && isComputerUser) {
    return null;
  }

  _balsamiq.default.assertObject(userInfo);

  const computedUserInfo = { ...userInfo
  };
  computedUserInfo.inProject = userBelongsToProject;
  computedUserInfo.canChat = !userInfo.anonymous;

  if (isComputerUser) {
    computedUserInfo.displayName = computedUserInfo.name && computedUserInfo.name !== '' ? computedUserInfo.name : computedUserInfo.email;
    computedUserInfo.ID = COMPUTER_USER_PREFIX + computedUserInfo.email;
  } else if (computedUserInfo.displayName === '') {
    computedUserInfo.displayName = computedUserInfo.email;
  }

  return computedUserInfo;
}

function makeConnectedUsersCoreObjectID(projectCoreObjectID) {
  CoreObject.assertProjectType(projectCoreObjectID);
  let newID = projectCoreObjectID.clone();
  newID.type = CONNECTED_USERS_ID_TYPE;
  return newID;
}

function assertConnectedUsersType(coreObjectID) {
  _balsamiq.default.assert(coreObjectID.type == CONNECTED_USERS_ID_TYPE);
}

const KeyNames = {
  OTHER_USERS: "otherUsers",
  MY_USER: "myUser",
  ALL_KNOWN_USERS: 'allKnownUsers'
};
exports.KeyNames = KeyNames;
const EventTypes = {
  USER_LEFT: 'userLeft'
};
exports.EventTypes = EventTypes;
const ActivityInterval = {
  Active: 10 * 1000,
  Waiting: 20 * 1000
};

class ConnectedUserTimer {
  constructor(host, callback, name) {
    _balsamiq.default.exists(host);

    _balsamiq.default.assertFunction(callback);

    _balsamiq.default.assertString(name);

    this.host = host;
    this.callback = callback;
    this.name = name;
    this.status = Constants.ConnectedUserStatus.Active;
    this.timerID = null;
  }

  cleanUp() {
    if (this.timerID) {
      this.host.clearTimer(this.timerID);
    }

    this.host = null;
    this.callback = null;
    this.timerID = null;
  }

  restart() {
    if (this.timerID) {
      this.host.clearTimer(this.timerID);
    }

    this._setStatus(Constants.ConnectedUserStatus.Active);

    this.timerID = this.host.timer(ActivityInterval.Active, () => {
      this._setStatus(Constants.ConnectedUserStatus.Waiting);

      this.timerID = this.host.timer(ActivityInterval.Waiting, () => {
        this._setStatus(Constants.ConnectedUserStatus.Away);

        this.timerID = null;
      });
    });
  }

  _setStatus(s) {
    if (this.status != s) {
      this.status = s;
      this.callback(this.name, this.status);
    }
  }

}

const makeKnownUser = function (name, displayName, avatarImageID, avatarImageWidth, avatarImageHeight) {
  _balsamiq.default.assertString(name);

  return {
    name: name,
    displayName: displayName,
    avatarImageID: avatarImageID,
    avatarImageWidth: avatarImageWidth,
    avatarImageHeight: avatarImageHeight
  };
};

class UserAvatar {
  constructor(host, callback, name) {
    _balsamiq.default.exists(host);

    _balsamiq.default.assertFunction(callback);

    _balsamiq.default.assertString(name);

    this.host = host;
    this.callback = callback;
    this.name = name;
    this.url = null;
    this.imageID = null;
    this.imageWidth = -1;
    this.imageHeight = -1;
  }

  cleanUp() {
    if (this.imageID) {
      this.host.Images.remove(this.imageID);
    }
  }

  setURL(url) {
    if (this.url != url) {
      if (this.imageID) {
        this.host.Images.remove(this.imageID);
        this.imageID = null;
      }

      this.url = url;
      this.host.Images.makeFromURL(this.url, res => {
        if (res.id) {
          this.imageID = res.id;
          this.imageWidth = res.width ? res.width : -1;
          this.imageHeight = res.height ? res.height : -1;
          this.callback(this.name, this.url, this.imageID, this.imageWidth, this.imageHeight);
        }
      });
    }
  }

}

class ConnectedUser {
  constructor(_ref) {
    let {
      name,
      displayName,
      role
    } = _ref;

    _balsamiq.default.assertString(name);

    _balsamiq.default.assertString(displayName);

    _balsamiq.default.assertEnum(role, Constants.UserRoles);

    this.name = name;
    this.displayName = displayName;
    this.avatarURL = null;
    this.avatarImageID = null;
    this.avatarImageWidth = -1;
    this.avatarImageHeight = -1;
    this.role = role;
    this.actualRole = role;
    this.anonymous = false;
    this.chat = false;
    this.targetResource = null; // null or { resourceID: string, branchID:string }

    this.selectedControlIDs = []; // null or [ [control ids], [control ids], .. ]

    this.status = Constants.ConnectedUserStatus.Active;
  }

  getName() {
    return this.name;
  }

  getDisplayName() {
    return this.displayName ? this.displayName : this.name;
  }

  setDisplayName(displayName) {
    this.displayName = displayName;
  }

  getAvatarURL() {
    return this.avatarURL;
  }

  getAvatarImageID() {
    return this.avatarImageID;
  }

  getAvatarImageWidth() {
    return this.avatarImageWidth;
  }

  getAvatarImageHeight() {
    return this.avatarImageHeight;
  }

  setAvatar(url, imageID, width, height) {
    _balsamiq.default.assertString(url);

    _balsamiq.default.assertString(imageID);

    this.avatarURL = url;
    this.avatarImageID = imageID;
    this.avatarImageWidth = width;
    this.avatarImageHeight = height;
  }

  getRole() {
    return this.role;
  }

  setRole(role) {
    _balsamiq.default.assertEnum(role, Constants.UserRoles);

    this.role = role;

    if (this.actualRole > this.role) {
      this.actualRole = this.role;
    }
  }

  getActualRole() {
    return this.actualRole;
  }

  setActualRole(role) {
    _balsamiq.default.assertEnum(role, Constants.UserRoles);

    this.actualRole = role;
  }

  getTargetResource() {
    return this.targetResource;
  }

  isAnonymous() {
    return this.anonymous;
  }

  setTargetResource(res) {
    this.targetResource = res;
  }

  getSelectedControlIDs() {
    return this.selectedControlIDs;
  }

  setSelectedControlIDs(controls) {
    this.selectedControlIDs = controls;
  }

  getStatus() {
    return this.status;
  }

  setStatus(s) {
    _balsamiq.default.assertEnum(s, Constants.ConnectedUserStatus);

    this.status = s;
  }

  setAnonymous(value) {
    this.anonymous = value;
  }

  setCanChat(value) {
    this.chat = value ? true : false;
  }

  canChat() {
    return this.chat;
  }

}

exports.ConnectedUser = ConnectedUser;

class ConnectedUsers extends CoreObject.CoreObject {
  constructor(context, coreObjectID) {
    _balsamiq.default.exists(context);

    _balsamiq.default.exists(context.host);

    assertConnectedUsersType(coreObjectID);
    super(context, coreObjectID);
    this.myUserInfo = null;
    /** @type { { role: string; userInfo: import("@balsamiq/bmpr/lib/BmprConstants").UserInfo }[] | null } */

    this.userList = null;
    this.otherUsers = {}; // user name -> ConnectedUser

    this.otherUsersTimers = {}; // user name -> ConnectedUserTimer

    this.sortedOtherUsers = []; // for OTHER_USERS key

    this.allKnownUsers = {}; // user name -> { name, displayName, avatarImageID } of connected and no longer connected users

    this.allKnownAvatars = {}; // user name -> UserAvatar of all known users
  }

  cleanUp() {
    for (let name in this.allKnownAvatars) {
      this.allKnownAvatars[name].cleanUp();
    }

    this.allKnownAvatars = null;

    for (let name in this.otherUsersTimers) {
      this.otherUsersTimers[name].cleanUp();
    }

    this.otherUsersTimers = null;
  }

  setMyUserInfo(userInfo) {
    if (!userInfo) {
      this.myUserInfo = null;
    } else {
      _balsamiq.default.assertString(userInfo.name);

      _balsamiq.default.assertOptionalString(userInfo.email);

      this.myUserInfo = userInfo; // desktop users might not set name since the only mandatory field is email

      if (!this.myUserInfo.ID) {
        this.myUserInfo.ID = userInfo.name;
      }
    }

    this.notifyKeyValueChanged(KeyNames.MY_USER);
    this.rebuildOtherUsers();
  }

  getMyUserInfo() {
    return this.myUserInfo;
  }

  getKeyValue(keyName) {
    switch (keyName) {
      case KeyNames.OTHER_USERS:
        return this.sortedOtherUsers;

      case KeyNames.MY_USER:
        return this.myUserInfo;

      case KeyNames.ALL_KNOWN_USERS:
        return this.allKnownUsers;

      default:
        return super.getKeyValue(keyName);
    }
  }

  onUsersListChanged(userList) {
    _balsamiq.default.assertArray(userList);

    this.userList = userList;
    this.rebuildOtherUsers();
  }

  onUserTargetAndSelectionChanged(userName, actualRole, resourceID, branchID, selectedControlIDs) {
    _balsamiq.default.assertString(userName);

    _balsamiq.default.assertEnum(actualRole, Constants.UserRoles);

    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.assertString(branchID);

    if (selectedControlIDs) {
      _balsamiq.default.assertArray(selectedControlIDs);
    }

    if (this.otherUsers[userName]) {
      let user = this.otherUsers[userName];
      user.setActualRole(actualRole);
      user.setTargetResource({
        resourceID: resourceID,
        branchID: branchID
      });
      user.setSelectedControlIDs(selectedControlIDs ? selectedControlIDs : []);
      this.otherUsersTimers[userName].restart();
      this.notifyKeyValueChanged(KeyNames.OTHER_USERS);
    }
  }

  rebuildOtherUsers() {
    this.sortedOtherUsers = [];
    let userNames = {}; // user names in the userList

    if (this.userList) {
      // add or update the users in the userList
      this.userList.forEach(user => {
        let isMe = this.myUserInfo && user.userInfo.name == this.myUserInfo.name;

        if (!isMe) {
          let name = user.userInfo.name;
          userNames[name] = true;
          let avatar = this.allKnownAvatars[name];
          let avatarImageID = avatar ? avatar.imageID : null;
          let avatarImageWidth = avatar ? avatar.imageWidth : -1;
          let avatarImageHeight = avatar ? avatar.imageHeight : -1;

          if (!avatar) {
            this.allKnownAvatars[name] = new UserAvatar(this.context.host, this._onConnectedUserAvatarImage.bind(this), name);
            this.allKnownAvatars[name].setURL(user.userInfo.avatarURL);
          }

          let otherUser;

          if (this.otherUsers[name]) {
            // known user, then update its object
            otherUser = this.otherUsers[name];
            otherUser.setDisplayName(user.userInfo.displayName ? user.userInfo.displayName : name);
            let parsedRole = parseUserRole(user.role);

            if (parsedRole != null) {
              otherUser.setRole(parsedRole);
            }

            if (user.userInfo.anonymous) {
              otherUser.setAnonymous(true);
            }

            otherUser.setCanChat(user.userInfo.canChat);
          } else {
            // new user, then create its object and its timer
            this.context.host.log("ConnectedUsers: user with name=" + name + " is now connected");
            otherUser = new ConnectedUser({
              name: name,
              displayName: user.userInfo.displayName ? user.userInfo.displayName : name,
              role: parseInt(user.role)
            });

            if (user.userInfo.anonymous) {
              otherUser.setAnonymous(true);
            }

            if (avatarImageID) {
              otherUser.setAvatar(user.userInfo.avatarURL, avatarImageID, avatarImageWidth, avatarImageHeight);
            }

            otherUser.setCanChat(user.userInfo.canChat);
            this.otherUsers[name] = otherUser;
            this.otherUsersTimers[name] = new ConnectedUserTimer(this.context.host, this._onConnectedUserTimer.bind(this), name);
            this.otherUsersTimers[name].restart();
          }

          this.sortedOtherUsers.push(otherUser);
          this.allKnownUsers[name] = makeKnownUser(name, otherUser.getDisplayName(), otherUser.getAvatarImageID(), otherUser.getAvatarImageWidth(), otherUser.getAvatarImageHeight());
        }
      });
      this.sortedOtherUsers.sort((a, b) => a.getName().localeCompare(b.getName())); // remove the users who are no longer in the userList

      let userNamesToDelete = {};

      for (let name in this.otherUsers) {
        if (!userNames[name]) {
          // this user is no longer connected
          userNamesToDelete[name] = true;
          this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.USER_LEFT, this.otherUsers[name]));
          this.context.host.log("ConnectedUsers: user with name=" + name + " is no longer connected");
        }
      }

      for (let name in userNamesToDelete) {
        delete this.otherUsers[name];

        if (this.otherUsersTimers[name]) {
          this.otherUsersTimers[name].cleanUp();
          delete this.otherUsersTimers[name];
        }
      }
    } //console.log("ConnectedUsers: all users: " + JSON.stringify(this.userList) + ", other users: " + JSON.stringify(this.sortedOtherUsers.map((user) => { return { name: user.getName() } })));


    this.notifyKeyValueChanged(KeyNames.ALL_KNOWN_USERS);
    this.notifyKeyValueChanged(KeyNames.OTHER_USERS);
  }

  _onConnectedUserTimer(name, status) {
    if (this.otherUsers[name]) {
      this.context.host.log("ConnectedUsers: user with name=" + name + " has now status=" + status);
      let user = this.otherUsers[name];
      user.setStatus(status);
      this.notifyKeyValueChanged(KeyNames.OTHER_USERS);
    }
  }

  _onConnectedUserAvatarImage(name, url, imageID, imageWidth, imageHeight) {
    this.context.host.log("ConnectedUsers: user with name=" + name + " has avatar with imageID=" + imageID);
    let displayName = this.allKnownUsers[name] ? this.allKnownUsers[name].displayName : undefined;
    this.allKnownUsers[name] = makeKnownUser(name, displayName, imageID, imageWidth, imageHeight);
    this.notifyKeyValueChanged(KeyNames.ALL_KNOWN_USERS);

    if (this.otherUsers[name]) {
      let user = this.otherUsers[name];
      user.setAvatar(url, imageID, imageWidth, imageHeight);
      this.notifyKeyValueChanged(KeyNames.OTHER_USERS);
    }
  }

}

exports.ConnectedUsers = ConnectedUsers;

function parseUserRole(role) {
  let r = null;

  if (_balsamiq.default.isString(role)) {
    let s = parseInt(role);

    if (_balsamiq.default.checkEnum(s, Constants.UserRoles)) {
      r = s;
    }
  }

  return r;
}

/***/ }),

/***/ "./js/core/model/controlschanges.js":
/*!******************************************!*\
  !*** ./js/core/model/controlschanges.js ***!
  \******************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ControlsChanges = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var CoreObject = _interopRequireWildcard(__webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js"));

var _constants = __webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class ControlsChanges {
  constructor() {
    /** @type {Object<string, any>}*/
    this.changes = {}; // stringyfied CoreObjectID -> { position: boolean, displayList: boolean }
  }
  /**
   *
   * @param {import("core/core/coreobject").CoreObjectID>} coreObjectID
   * @param {*} change
   */


  setChange(coreObjectID, change) {
    _balsamiq.default.existsClass(coreObjectID, CoreObject.CoreObjectID);

    _balsamiq.default.assertEnum(change, _constants.ControlChanges);

    this._controlData(coreObjectID)[change] = true;
  }

  getChanges() {
    return this.changes;
  } // Private
  // -------

  /**
   *
   * @param {import("core/core/coreobject").CoreObjectID>} coreObjectID
   */


  _controlData(coreObjectID) {
    var coreObjectIDString = coreObjectID.toString();

    if (!this.changes[coreObjectIDString]) {
      this.changes[coreObjectIDString] = {
        coreObjectID: coreObjectID
      };
    }

    return this.changes[coreObjectIDString];
  }

}

exports.ControlsChanges = ControlsChanges;

/***/ }),

/***/ "./js/core/model/dirtywireframes.js":
/*!******************************************!*\
  !*** ./js/core/model/dirtywireframes.js ***!
  \******************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.KeyNames = exports.DirtyWireframes = exports.DIRTY_WIREFRAMES_ID_TYPE = void 0;
exports.assertDirtyWireframesType = assertDirtyWireframesType;
exports.makeDirtyWireframesCoreObjectID = makeDirtyWireframesCoreObjectID;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var CoreObject = _interopRequireWildcard(__webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const DIRTY_WIREFRAMES_ID_TYPE = "dirtyWireframes";
exports.DIRTY_WIREFRAMES_ID_TYPE = DIRTY_WIREFRAMES_ID_TYPE;

function makeDirtyWireframesCoreObjectID(projectCoreObjectID) {
  CoreObject.assertProjectType(projectCoreObjectID);
  var newID = projectCoreObjectID.clone();
  newID.type = DIRTY_WIREFRAMES_ID_TYPE;
  return newID;
}

function assertDirtyWireframesType(coreObjectID) {
  _balsamiq.default.assert(coreObjectID.type == DIRTY_WIREFRAMES_ID_TYPE);
}

const KeyNames = {
  DIRTY_WIREFRAMES: 'dirtyWireframes'
};
exports.KeyNames = KeyNames;

class DirtyWireframes extends CoreObject.CoreObject {
  constructor(context, coreObjectID) {
    _balsamiq.default.exists(context);

    _balsamiq.default.exists(context.host);

    assertDirtyWireframesType(coreObjectID);
    super(context, coreObjectID);
    this.dirtyWireframes = new Map();
  }

  getKeyValue(keyName) {
    switch (keyName) {
      case KeyNames.DIRTY_WIREFRAMES:
        return this;

      default:
        return super.getKeyValue(keyName);
    }
  }

  reset() {
    if (!this.hasDirty()) {
      return;
    }

    this.dirtyWireframes = new Map();
    this.notifyKeyValueChanged(KeyNames.DIRTY_WIREFRAMES);
  }

  getDirty() {
    return [...this.dirtyWireframes.values()];
  }

  hasDirty() {
    return this.dirtyWireframes.size !== 0;
  }

  isDirty(coreObjectID) {
    const key = coreObjectID.toString();
    return this.dirtyWireframes.has(key);
  }

  addDirty(coreObjectID) {
    const key = coreObjectID.toString();

    if (this.dirtyWireframes.has(key)) {
      return; // Already marked as dirty.
    }

    this.dirtyWireframes.set(key, coreObjectID);
    this.notifyKeyValueChanged(KeyNames.DIRTY_WIREFRAMES);
  }

  removeDirty(coreObjectID) {
    const key = coreObjectID.toString();

    if (!this.dirtyWireframes.has(key)) {
      return; // already deleted
    }

    this.dirtyWireframes.delete(coreObjectID.toString());
    this.notifyKeyValueChanged(KeyNames.DIRTY_WIREFRAMES);
  }

}

exports.DirtyWireframes = DirtyWireframes;

/***/ }),

/***/ "./js/core/model/image.js":
/*!********************************!*\
  !*** ./js/core/model/image.js ***!
  \********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.NativeStatus = exports.KeyNames = exports.Image = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Asset = _interopRequireWildcard(__webpack_require__(/*! core/model/asset */ "./js/core/model/asset.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var MimeTypes = _interopRequireWildcard(__webpack_require__(/*! core/mimetypes */ "./js/core/mimetypes.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const KeyNames = {
  IMAGE_SIZE: "imageSize"
};
exports.KeyNames = KeyNames;
const NativeStatus = {
  PENDING: 'pending',
  OK: 'ok',
  ERROR: 'error'
};
exports.NativeStatus = NativeStatus;

class Image extends Asset.Asset {
  constructor(context, coreObjectID, attributes, data) {
    super(context, coreObjectID, attributes, data);

    _balsamiq.default.assert(context.host);

    _balsamiq.default.assert(context.host.Images);
    /** @type {string|null} */


    this.nativeImageID = null;
    this.imageSize = null;
    this.nativeStatus = NativeStatus.PENDING;
  }

  makeNativeImage(origin) {
    let done = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : () => {};

    this._makeNativeImage(MimeTypes.findMimetypeFromExtension(this.attributes.extension), this.data, origin, done);
  }

  cleanUp() {
    this._removeNativeImage();
  } // Key-value coding
  // ----------------


  getKeyValue(keyName) {
    switch (keyName) {
      case KeyNames.IMAGE_SIZE:
        return this.imageSize;

      default:
        return super.getKeyValue(keyName);
    }
  } // Getters
  // -------


  getImageID() {
    return this.nativeImageID;
  }

  getImageSize() {
    return this.imageSize;
  }

  getPreviewData() {
    if (this.nativeImageID) {
      return {
        imageID: this.nativeImageID
      };
    }

    return super.getPreviewData();
  } // Private
  // -------


  _removeNativeImage() {
    if (this.nativeImageID) {
      this.context.host.log(`Image: deleted ${this.nativeImageID} for resource ${this.coreObjectID.resourceID}`);
      this.context.host.Images.remove(this.nativeImageID);
      this.nativeImageID = null;
      this.nativeStatus = NativeStatus.PENDING;
    }
  }

  _makeNativeImage(mimeType, data, origin, done) {
    // { id: string, width: integer, height: integer, error: * }
    const cb = function (res) {
      // remove the previous image if present
      this._removeNativeImage(); // handle the result


      if (res.error) {
        this.nativeImageID = null;
        this.imageSize = null;
        this.nativeStatus = NativeStatus.ERROR;
      } else {
        // the image is set
        this.context.host.log(`Image: created ${res.id} for resource ${this.coreObjectID.resourceID}`);
        this.nativeImageID = res.id;
        this.imageSize = _types.default.newSize(res.width, res.height);
        this.nativeStatus = NativeStatus.OK;
      }

      this.notifyContentsChanged(null, origin);
      this.notifyKeyValueChanged(KeyNames.IMAGE_SIZE);
      done();
    }.bind(this);

    this.context.host.Images.makeFromBase64Data(mimeType, data, cb);
  }

}

exports.Image = Image;

/***/ }),

/***/ "./js/core/model/locators.js":
/*!***********************************!*\
  !*** ./js/core/model/locators.js ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ProjectPlatformName = exports.ProjectPlatform = exports.ProjectLocatorKey = exports.ProjectLocator = exports.Enums = void 0;
exports.makeCloudBillingURL = makeCloudBillingURL;
exports.makeCloudCommentURL = makeCloudCommentURL;
exports.makeCloudProjectLocator = makeCloudProjectLocator;
exports.makeCloudProjectURL = makeCloudProjectURL;
exports.makeCloudResourceURL = makeCloudResourceURL;
exports.makeCloudShareDialogUrl = makeCloudShareDialogUrl;
exports.makeCloudSpaceSettingsURL = makeCloudSpaceSettingsURL;
exports.makeCloudSuperadminURL = makeCloudSuperadminURL;
exports.makeFileProjectLocator = makeFileProjectLocator;
exports.makeProjectLocatorFromObject = makeProjectLocatorFromObject;
exports.makeWebdemoProjectLocator = makeWebdemoProjectLocator;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Base36 = _interopRequireWildcard(__webpack_require__(/*! core/utils/base36 */ "./js/core/utils/base36.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * @enum
 * The platforms available to {@link ProjectLocator}.
 */
const ProjectPlatform = {
  Cloud: 'cloud',
  File: 'file',
  Webdemo: 'webdemo'
};
/**
 * @enum
 * The user-readable names of the available platforms.
 */

exports.ProjectPlatform = ProjectPlatform;
const ProjectPlatformName = {
  [ProjectPlatform.Cloud]: 'Balsamiq Cloud',
  [ProjectPlatform.File]: 'This Computer',
  [ProjectPlatform.Webdemo]: 'Webdemo'
};
/**
 * @enum
 * Valid keys of {@link ProjectLocator} data.
 * All keys have String value, except where another type is explicitly specified.
 */

exports.ProjectPlatformName = ProjectPlatformName;
const ProjectLocatorKey = {
  Platform: 'platform',
  URL: 'url',
  Name: 'name',
  CloudSiteName: 'cloudSiteName',
  CloudDomain: 'cloudDomain',
  CloudSiteID: 'cloudSiteID',
  // number
  CloudProjectID: 'cloudProjectID',
  // number
  FileIsUnsaved: 'fileIsUnsaved',
  // bool
  WebdemoBASDomain: 'webdemoBASDomain',
  WebdemoPID: 'webdemoPid',
  WebdemoUserName: 'webdemoUserName',
  WebdemoDisplayName: 'webdemoDisplayName',
  WebdemoRole: 'webdemoRole',
  WebdemoAvatarURL: 'webdemoAvatarURL'
};
/**
 * @class
 * A ProjectLocator is a kind of unique resource identifier to locate a Balsamiq Wireframes Project.
 *
 * The minimum set of data for a project locator is made by 3 parts:
 * - {@link Platform} (Cloud, File, Webdemo - and in the future: Google Drive, Confluence...)
 * - URL
 * - Name
 *
 * The URL must be unique for that project.
 *
 * The Name is for displaying purposes only, it does not identify really the project.
 *
 * Depending on the Platform, other data are added to the project locator.
 *
 * For example, Cloud project locators have:
 * - Cloud instance domain (like *balsamiq.cloud* or *staging.balsamiq.cloud*)
 * - siteID (the integer number of the parent Site)
 * - projectID (the integer number of the Project in Cloud)
 *
 * Used mostly for development and debugging, Webdemo projects have data like:
 * - user name, display name, avatarURL
 * - BAS instance domain
 * - PID
 *
 * ProjectLocator objects are designed to be exchanged with native apps. The function
 * {@link makeProjectLocatorFromObject} create a ProjectLocator instance from a simple object,
 * while the method {@link ProjectLocator#toObject} gets a simple object from the ProjectLocator
 * instance.
 */

exports.ProjectLocatorKey = ProjectLocatorKey;

class ProjectLocator {
  /**
   * @constructor
   * @param object {Object} The dictionary with all ProjectLocatorKey and values. The object must contain,
   * at least:
   * -
   */
  constructor(object) {
    _balsamiq.default.assertObject(object);

    _balsamiq.default.assertString(object[ProjectLocatorKey.Platform]);

    _balsamiq.default.assertString(object[ProjectLocatorKey.URL]);

    _balsamiq.default.assertString(object[ProjectLocatorKey.Name]);

    Object.keys(object).forEach(key => _balsamiq.default.assertEnum(key, ProjectLocatorKey));
    this._object = object;
  }
  /**
   * Return a simple object with the data of this ProjectLocator, good for exchange
   * with native environment.
   */


  toObject() {
    return this._object;
  }
  /**
   * Get the value of the given key.
   * @param {ProjectLocatorKey} key
   */


  valueForKey(key) {
    _balsamiq.default.assertEnum(key, ProjectLocatorKey);

    return this._object[key];
  }

  setName(newName) {
    _balsamiq.default.assertString(newName);

    this._object[ProjectLocatorKey.Name] = newName;
  }

  platform() {
    return this.valueForKey(ProjectLocatorKey.Platform);
  }

  url() {
    return this.valueForKey(ProjectLocatorKey.URL);
  }

  name() {
    return this.valueForKey(ProjectLocatorKey.Name);
  }

  isCloudProject() {
    return this.valueForKey(ProjectLocatorKey.Platform) == ProjectPlatform.Cloud;
  }

  cloudDomain() {
    return this.valueForKey(ProjectLocatorKey.CloudDomain);
  }

  cloudSiteID() {
    return this.valueForKey(ProjectLocatorKey.CloudSiteID);
  }

  cloudSiteName() {
    return this.valueForKey(ProjectLocatorKey.CloudSiteName);
  }

  cloudProjectID() {
    return this.valueForKey(ProjectLocatorKey.CloudProjectID);
  }

  isFileProject() {
    return this.valueForKey(ProjectLocatorKey.Platform) == ProjectPlatform.File;
  }

  isUnsavedFileProject() {
    return this.valueForKey(ProjectLocatorKey.FileIsUnsaved);
  }

  isWebdemoProject() {
    return this.valueForKey(ProjectLocatorKey.Platform) == ProjectPlatform.Webdemo;
  }

  webdemoBASDomain() {
    return this.valueForKey(ProjectLocatorKey.WebdemoBASDomain);
  }

  webdemoPID() {
    return this.valueForKey(ProjectLocatorKey.WebdemoPID);
  }

  webdemoUserName() {
    return this.valueForKey(ProjectLocatorKey.WebdemoUserName);
  }

  webdemoDisplayName() {
    return this.valueForKey(ProjectLocatorKey.WebdemoDisplayName);
  }

  webdemoRole() {
    return this.valueForKey(ProjectLocatorKey.WebdemoRole);
  }

  webdemoAvatarURL() {
    return this.valueForKey(ProjectLocatorKey.WebdemoAvatarURL);
  }

}
/**
 * Make a {@link ProjectLocator} of a project in Cloud.
 * @param {String} domain
 * @param {Number} siteID
 * @param {Number} projectID
 * @param {String} projectName
 */


exports.ProjectLocator = ProjectLocator;

function makeCloudProjectLocator(domain, siteID, siteName, projectID, projectName) {
  _balsamiq.default.assertString(domain);

  _balsamiq.default.assertNumber(siteID);

  _balsamiq.default.assertNumber(projectID);

  _balsamiq.default.assertString(projectName);

  _balsamiq.default.assertString(siteName);

  return new ProjectLocator({
    [ProjectLocatorKey.Platform]: ProjectPlatform.Cloud,
    [ProjectLocatorKey.Name]: projectName,
    [ProjectLocatorKey.URL]: makeCloudProjectURL(domain, siteID, projectID),
    [ProjectLocatorKey.CloudDomain]: domain,
    [ProjectLocatorKey.CloudSiteID]: siteID,
    [ProjectLocatorKey.CloudSiteName]: siteName,
    [ProjectLocatorKey.CloudProjectID]: projectID
  });
}
/**
 * Create a query params string that you can concat to a URL
 * @param {Object} queryParams
 * @returns {String}
 */


function encodeQueryObject(query) {
  _balsamiq.default.assertObject(query);

  return Object.keys(query).map(k => `${encodeURIComponent(k)}=${encodeURIComponent(query[k])}`).join('&');
}
/**
 * Make the URL to display superadmin page for user.
 * @param {String} domain
 * @param {Number} userID
 * @returns {String}
 */


function makeCloudSuperadminURL(domain, userID) {
  _balsamiq.default.assertString(domain);

  _balsamiq.default.assertNumber(userID);

  return `https://${domain}/admin/history/u${userID}`;
}
/**
 * Make the URL to display a project share dialog in Cloud.
 * @param {String} domain
 * @param {Number} siteID
 * @param {Number} projectID
 * @returns {String}
 */


function makeCloudShareDialogUrl(domain, siteID, projectID) {
  _balsamiq.default.assertString(domain);

  _balsamiq.default.assertNumber(siteID);

  _balsamiq.default.assertNumber(projectID);

  return `https://${domain}/s${Base36.encode(siteID)}/projects#manage_members=${Base36.encode(projectID)}`;
}
/**
 * Make the URL of a settings page in Cloud with billing anchor.
 * @param {String} domain
 * @param {Number} siteID
 * @returns {String}
 */


function makeCloudBillingURL(domain, siteID) {
  _balsamiq.default.assertString(domain);

  _balsamiq.default.assertNumber(siteID);

  return `https://${domain}/s${Base36.encode(siteID)}/settings#billing-information`;
}
/**
 * Make the URL of a project in Cloud.
 * @param {String} domain
 * @param {Number} siteID
 * @param {Number} projectID
 * @returns {String}
 */


function makeCloudProjectURL(domain, siteID, projectID) {
  _balsamiq.default.assertString(domain);

  _balsamiq.default.assertNumber(siteID);

  _balsamiq.default.assertNumber(projectID);

  return `https://${domain}/s${Base36.encode(siteID)}/p${Base36.encode(projectID)}`;
}
/**
 * Make the URL of the Settings page of a Space in Cloud.
 * @param {String} domain
 * @param {Number} siteID
 * @returns {String}
 */


function makeCloudSpaceSettingsURL(domain, siteID) {
  _balsamiq.default.assertString(domain);

  _balsamiq.default.assertNumber(siteID);

  return `https://${domain}/s${Base36.encode(siteID)}/settings`;
}
/**
 * Make the URL of a Comment in Cloud.
 * @param {String} domain
 * @param {Number} siteID
 * @param {Number} projectID
 * @param {String} shortCommentId
 * @returns {String}
 */


function makeCloudCommentURL(_ref) {
  let {
    domain,
    siteID,
    projectID,
    shortCommentID
  } = _ref;

  _balsamiq.default.assertString(domain);

  _balsamiq.default.assertNumber(siteID);

  _balsamiq.default.assertNumber(projectID);

  _balsamiq.default.assertString(shortCommentID);

  return `https://${domain}/s${Base36.encode(siteID)}/p${Base36.encode(projectID)}/c${shortCommentID}`;
}
/**
 * Make the URL of a Resource in Cloud.
 * @param {String} domain
 * @param {Number} siteID
 * @param {Number} projectID
 * @param {String} shortResourceId
 * @param {String=} shortBranchId
 * @param {Object=} query
 * @returns {String}
 */


function makeCloudResourceURL(_ref2) {
  let {
    domain,
    siteID,
    projectID,
    shortResourceID,
    shortBranchID = null,
    query = null
  } = _ref2;

  _balsamiq.default.assertString(domain);

  _balsamiq.default.assertNumber(siteID);

  _balsamiq.default.assertNumber(projectID);

  _balsamiq.default.assertString(shortResourceID);

  let url = `https://${domain}/s${Base36.encode(siteID)}/p${Base36.encode(projectID)}/r${shortResourceID}`;

  if (shortBranchID) {
    url += `/b${shortBranchID}`;
  }

  if (query) {
    url += `?${encodeQueryObject(query)}`;
  }

  return url;
}
/**
 * Make a {@link ProjectLocator} of a Webdemo project.
 * The URL of the project is a fake one, pointing directly to the BAS instance.
 * @param {String} basDomain
 * @param {String} pid
 * @param {String} userName
 * @param {String} displayName
 * @param {String} role
 * @param {String} avatarURL
 * @returns {ProjectLocator}
 */


function makeWebdemoProjectLocator(basDomain, pid, userName, displayName, role, avatarURL) {
  _balsamiq.default.assertString(basDomain);

  _balsamiq.default.assertString(pid);

  _balsamiq.default.assertString(userName);

  _balsamiq.default.assertString(displayName);

  _balsamiq.default.assertString(role);

  _balsamiq.default.assertString(avatarURL);

  let object = {
    [ProjectLocatorKey.Platform]: ProjectPlatform.Webdemo,
    [ProjectLocatorKey.Name]: "Webdemo pid=" + pid,
    [ProjectLocatorKey.URL]: `https://${basDomain}/?pid=${pid}`,
    [ProjectLocatorKey.WebdemoBASDomain]: basDomain,
    [ProjectLocatorKey.WebdemoUserName]: userName,
    [ProjectLocatorKey.WebdemoDisplayName]: displayName,
    [ProjectLocatorKey.WebdemoRole]: role,
    [ProjectLocatorKey.WebdemoAvatarURL]: avatarURL
  };
  return new ProjectLocator(object);
}

function makeFileProjectLocator(url, name) {
  let isUnsaved = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;

  _balsamiq.default.assertString(url);

  _balsamiq.default.assertString(name);

  _balsamiq.default.assertBool(isUnsaved);

  let object = {
    [ProjectLocatorKey.Platform]: ProjectPlatform.File,
    [ProjectLocatorKey.Name]: name,
    [ProjectLocatorKey.URL]: url,
    [ProjectLocatorKey.FileIsUnsaved]: isUnsaved
  };
  return new ProjectLocator(object);
}
/**
 * Make a {@link ProjectLocator} from a plain object, like the object returned by toObject().
 * @param {Object} object
 */


function makeProjectLocatorFromObject(object) {
  _balsamiq.default.assertObject(object);

  return new ProjectLocator(object);
}

const Enums = {
  ProjectPlatform,
  ProjectLocatorKey,
  ProjectPlatformName
};
exports.Enums = Enums;

/***/ }),

/***/ "./js/core/model/message.js":
/*!**********************************!*\
  !*** ./js/core/model/message.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.MessageTypes = exports.Message = exports.ChatMessage = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const MessageTypes = {
  CHAT: "chat",
  TARGET_AND_SELECTION: 'targetAndSelection'
};
exports.MessageTypes = MessageTypes;

class Message {
  constructor(id, user, type, timestamp) {
    _balsamiq.default.assertString(id);

    _balsamiq.default.assertString(user);

    _balsamiq.default.checkEnum(type, MessageTypes);

    _balsamiq.default.assertNumber(timestamp);

    this.id = id;
    this.user = user;
    this.type = type;
    this.timestamp = timestamp;
    this.error = null;
  }

  getTimestamp() {
    return this.timestamp;
  }

  getData() {
    return {
      id: this.id,
      user: this.user,
      type: this.type,
      timestamp: this.timestamp,
      error: this.error
    };
  }

  update(data) {
    this.id = data.id;
    this.user = data.user;
    this.type = data.type;
    this.timestamp = data.timestamp;
  }

  setError(error) {
    this.error = error;
  }

  getError() {
    return this.error;
  }

}

exports.Message = Message;

class ChatMessage extends Message {
  constructor(id, type, user, displayName, timestamp, text) {
    _balsamiq.default.assertString(text);

    _balsamiq.default.assert(type === MessageTypes.CHAT);

    super(id, user, MessageTypes.CHAT, timestamp);
    this.text = text;
    this.displayName = displayName;
  }

  getText() {
    return this.text;
  }

  getData() {
    var superData = super.getData();
    return { ...superData,
      displayName: this.displayName,
      data: {
        text: this.text
      }
    };
  }

  update(data) {
    _balsamiq.default.assert(data.type === MessageTypes.CHAT);

    super.update(data);
    this.text = data.data.text;
  }

}

exports.ChatMessage = ChatMessage;

/***/ }),

/***/ "./js/core/model/mockup.js":
/*!*********************************!*\
  !*** ./js/core/model/mockup.js ***!
  \*********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.validateMockupData = exports.Mockup = exports.KeyNames = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _controls = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var _group = _interopRequireDefault(__webpack_require__(/*! core/controls/group */ "./js/core/controls/group.js"));

var _json = __webpack_require__(/*! core/serialization/json */ "./js/core/serialization/json.js");

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _events = _interopRequireDefault(__webpack_require__(/*! core/events */ "./js/core/events.ts"));

var _resource = __webpack_require__(/*! core/model/resource */ "./js/core/model/resource.js");

var _constants = __webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts");

var _coreobject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var _BmprConstants = __webpack_require__(/*! @balsamiq/bmpr/lib/BmprConstants */ "./node_modules/@balsamiq/bmpr/lib/BmprConstants.js");

var BarConstants = _interopRequireWildcard(__webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchiveConstants */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js"));

var _controlsdependencies = __webpack_require__(/*! core/model/controlsdependencies */ "./js/core/model/controlsdependencies.ts");

var _controlschanges = __webpack_require__(/*! core/model/controlschanges */ "./js/core/model/controlschanges.js");

var _mockupsize = __webpack_require__(/*! core/utils/mockupsize */ "./js/core/utils/mockupsize.js");

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var _fp = __webpack_require__(/*! lodash/fp */ "./node_modules/lodash/fp.js");

var _diff = __webpack_require__(/*! core/utils/diff */ "./js/core/utils/diff.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const KeyNames = {
  ORDER: _BmprConstants.MockupAttributes.Order,
  PARENTID: _BmprConstants.MockupAttributes.ParentID
};
/**
 * Takes a control and a mockup and returns a new control with relative x and y coordinates
 * @param {*} control
 * @param {*} mockup
 * @returns {*}
 */

exports.KeyNames = KeyNames;

const adjustGroupCoordinates = (control, mockup) => {
  const xDiff = mockup.mockup.measuredW - mockup.mockup.mockupW;
  const yDiff = mockup.mockup.measuredH - mockup.mockup.mockupH;
  return { ...control,
    x: control.x - xDiff,
    y: control.y - yDiff
  };
};

const validateMockupData = function (data) {
  if (data === "") {
    //Data is null in the BMPR. Should the user edit the wireframe the mockup JSON will be correctly set.
    return {
      ok: true
    };
  }

  try {
    const j = JSON.parse(data);
    const v = (0, _json.validateMockupJSON)(j, false);

    if (v.error) {
      return {
        ok: false,
        error: v.error
      };
    }
  } catch (error) {
    return {
      ok: false,
      error: error
    };
  }

  return {
    ok: true
  };
};

exports.validateMockupData = validateMockupData;

class Mockup extends _resource.Resource {
  constructor(context, coreObjectID, properties, attributes, data) {
    let type = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : _constants.ResourceTypes.MOCKUP;
    super(context, coreObjectID, type);

    _balsamiq.default.exists(context.skins, "need skins for new Mockup");

    _balsamiq.default.exists(context.getDependency);

    _balsamiq.default.exists(context.host);

    _balsamiq.default.exists(context.setDependencies);

    _balsamiq.default.assertString(data, "Mockup constructor needs data as String");

    _balsamiq.default.assertObject(attributes, "Mockup constructor needs attributes as Array");

    this.pendingDepsToUpdate = {};
    this.canAddControlToGroupTriggerALoop = this.context.canAddControlToGroupTriggerALoop;
    this.projectSetDependencies = context.setDependencies;

    if (this.coreObjectID.branchID == BarConstants.Branch.MasterBranchID) {
      this.attributes[_BmprConstants.ResourceAttributes.Kind] = _BmprConstants.ResourceKinds.Mockup;
      this.attributes[_BmprConstants.MockupAttributes.Order] = null;
      this.attributes[_BmprConstants.MockupAttributes.ParentID] = null;
    }

    this.properties = {
      skinName: "sketch",
      fontFace: _types.default.fontFaceToFontFamily("Balsamiq Sans"),
      fontSize: 13,
      selectionColor: 0xff0000,
      // must be set by the project, this red color is wrong
      linkColor: 0xff0000 // must be set by the project, this red color is wrong

    };
    this.controlsContext = _balsamiq.default.extend(_balsamiq.default.clone(context), {
      setDependencies: this._setControlDependencies.bind(this),
      notifyChange: this._onControlChange.bind(this),
      setReady: this._setControlReady.bind(this),
      getDependency: this.context.getDependency,
      skin: context.skins[this.properties.skinName],
      defaultFontSize: this.properties.fontSize,
      selectionColor: this.properties.selectionColor,
      fontFamily: this.properties.fontFace,
      linkColor: this.properties.linkColor,
      runtimeInfo: this.context.host.runtimeInfo(),
      getMockupName: this.getName.bind(this),
      hasSymbolDefinitionALoop: context.hasSymbolDefinitionALoop
    });

    if (this.properties) {
      this.setProperty("skinName", properties.skinName);
      this.setProperty("fontFace", properties.fontFace);
      this.setProperty("fontSize", properties.fontSize);
      this.setProperty("selectionColor", properties.selectionColor);
      this.setProperty("linkColor", properties.linkColor);
    }

    this.controlsNotReady = {};
    this.controlsDependencies = new _controlsdependencies.ControlsDependencies();
    this.controlsChanges = null;
    this.isReady = true;
    this.preview = false;
    this.startPreviewSize = null; //this.performEdit(() => {

    this.mainGroupControl = (0, _controls.newControl)(this.controlsContext, this.getMainGroupCoreObjectID(), _group.default.CONTROL_TYPE_ID);
    this.mainGroupControl.setMockup(this); //});

    this.setAttributes(attributes, null); // No origin -> No notifications

    var obj;

    try {
      obj = JSON.parse(data);
    } catch (error) {// TODO
    }

    if (obj) {
      this._load(obj);
    }
  }

  setName(value) {
    super.setName(value);
    this.mainGroupControl.updateAllMacros();
    this.notifyContentsChanged();
  }

  cleanUp() {
    this.mainGroupControl.cleanUp();
    super.cleanUp();
  }

  onContextChange(params) {
    this.context.host.log("Mockup: onContextChange");

    if (this.mainGroupControl) {
      this.mainGroupControl.onContextChange(params);
      this.context.host.log("Mockup: notifyContentsChanged after onContextChange");
      this.notifyContentsChanged();
    }
  } // Key-value coding
  // ----------------


  getKeyNames() {
    return super.getKeyNames().concat(Object.values(KeyNames));
  }

  getKeyValue(keyName) {
    if (keyName === KeyNames.ORDER) {
      return this.getOrder();
    } else if (keyName === KeyNames.PARENTID) {
      return this.getParentID();
    }

    return super.getKeyValue(keyName);
  } // Getters
  // -------


  getProperty(propName) {
    _balsamiq.default.assertString(propName);

    _balsamiq.default.assert(this.properties.hasOwnProperty(propName), "unknown Mockup property: '" + propName + "'");

    return this.properties[propName];
  }

  getSkin() {
    return this.controlsContext.skin;
  }

  getDisplayDataTree(selected, zoom) {
    return this.mainGroupControl.getDisplayDataTree(selected, zoom);
  }

  getMainGroupCoreObjectID() {
    return (0, _coreobject.makeControlCoreObjectID)(this.coreObjectID, []);
  }

  getRootControl() {
    return this.mainGroupControl;
  }
  /**
   *
   * @param {number[]} p_controlIDs
   * @returns {import("core/controls/controls").Control|null}
   */


  getControl(p_controlIDs) {
    if (p_controlIDs) {
      _balsamiq.default.assertArray(p_controlIDs);
    }

    return this.mainGroupControl.getControl(p_controlIDs);
  } // Setters
  // -------


  setProperty(propName, propValue, operationOrigin) {
    _balsamiq.default.assert(Object.keys(this.properties).indexOf(propName) !== -1, "unknown Mockup property: '" + propName + "' value: '" + propValue + "'");

    if (operationOrigin) {
      _balsamiq.default.assertEnum(operationOrigin, _constants.OperationOrigin);
    }

    if (this.properties[propName] !== propValue) {
      this.properties[propName] = propValue;

      if (propName === "skinName") {
        _balsamiq.default.assertString(propValue);

        this.controlsContext.skin = this.context.skins[propValue];
      } else if (propName === "fontSize") {
        _balsamiq.default.assertPositiveNumber(propValue);

        this.controlsContext.defaultFontSize = propValue;
      } else if (propName === "fontFace") {
        _balsamiq.default.assertString(propValue);

        this.controlsContext.fontFamily = _types.default.fontFaceToFontFamily(propValue);
      } else if (propName === "selectionColor") {
        _balsamiq.default.assertNumber(propValue);

        this.controlsContext.selectionColor = propValue;
      } else if (propName === "linkColor") {
        _balsamiq.default.assertNumber(propValue);

        this.controlsContext.linkColor = propValue;
      }

      if (this.mainGroupControl) {
        this.mainGroupControl.onContextChange();
        this.notifyContentsChanged(null, operationOrigin);
      } // TODO integrate readyForSnapshot stuff

    }
  } // Attributes getters and setter
  // -----------------------------

  /**
   * @returns { number }
   */


  getOrder() {
    return this.getAttribute(_BmprConstants.MockupAttributes.Order);
  }

  setOrder(value) {
    _balsamiq.default.assertPositiveNumber(value);

    this.setAttribute(_BmprConstants.MockupAttributes.Order, value, _constants.OperationOrigin.USER);
  }

  getParentID() {
    return this.getAttribute(_BmprConstants.MockupAttributes.ParentID);
  }

  setParentID(value) {
    if (_balsamiq.default.isString(value)) {
      _balsamiq.default.assertUUID(value);
    }

    this.setAttribute(_BmprConstants.MockupAttributes.ParentID, value, _constants.OperationOrigin.USER);
  } // Resource methods
  // ----------------


  onAttributesChanged(attrs) {
    var attributes = _balsamiq.default.clone(attrs);

    if (this.getKind() === _BmprConstants.ResourceKinds.Mockup) {
      // Symbol Libraries do not have order
      if (_BmprConstants.MockupAttributes.Order in attributes) {
        this.setAttribute(_BmprConstants.MockupAttributes.Order, attributes[_BmprConstants.MockupAttributes.Order], _constants.OperationOrigin.ARCHIVE);
        delete attributes[_BmprConstants.MockupAttributes.Order];
      }

      if (_BmprConstants.MockupAttributes.ParentID in attributes) {
        this.setAttribute(_BmprConstants.MockupAttributes.ParentID, attributes[_BmprConstants.MockupAttributes.ParentID], _constants.OperationOrigin.ARCHIVE);
        delete attributes[_BmprConstants.MockupAttributes.ParentID];
      }
    }

    super.onAttributesChanged(attributes); // TODO notify contents change only when the contents actually changed

    this.notifyContentsChanged(null, _constants.OperationOrigin.ARCHIVE);
  }

  setAttributes(attrs, operationOrigin) {
    var attributes = _balsamiq.default.clone(attrs);

    if (this.getKind() === _BmprConstants.ResourceKinds.Mockup) {
      // Symbol Libraries do not have order
      if (_BmprConstants.MockupAttributes.Order in attributes) {
        this.setAttribute(_BmprConstants.MockupAttributes.Order, attributes[_BmprConstants.MockupAttributes.Order], operationOrigin);
        delete attributes[_BmprConstants.MockupAttributes.Order];
      }

      if (_BmprConstants.MockupAttributes.ParentID in attributes) {
        this.setAttribute(_BmprConstants.MockupAttributes.ParentID, attributes[_BmprConstants.MockupAttributes.ParentID], operationOrigin);
        delete attributes[_BmprConstants.MockupAttributes.ParentID];
      }
    }

    super.setAttributes(attributes, operationOrigin);
  }

  onDataChanged(data) {
    var newData;

    try {
      newData = JSON.parse(data);
    } catch (error) {// TODO
    }

    if (newData) {
      this.performEdit(() => {
        const currentData = this.getJSONData();
        const changedGroupControlsIDs = (0, _diff.getChangedGroupControlIDsFromDiff)(currentData, newData);
        let childrenZIndexChanged = false;
        changedGroupControlsIDs.forEach(_ref => {
          let {
            type,
            controlID
          } = _ref;

          switch (type) {
            case _diff.VALUE_CHANGES.UPDATED:
              {
                const cID = controlID.split(",").map(parseInt);
                const control = this.getControl(cID);

                const newControls = _balsamiq.default.get(newData, "mockup.controls.control", []);

                const controlDataToReload = newControls.find(c => c.ID === controlID);
                control.reload(adjustGroupCoordinates(controlDataToReload, currentData), {
                  allowNegativeCoords: true
                });

                if (this.mainGroupControl.getChildIndex(control.controlID) !== parseInt(controlDataToReload.zOrder)) {
                  childrenZIndexChanged = true;
                }

                return;
              }

            case _diff.VALUE_CHANGES.CREATED:
              {
                const newControls = _balsamiq.default.get(newData, "mockup.controls.control", []);

                const controlToAdd = newControls.find(c => c.ID === controlID);
                this.mainGroupControl.loadChildren({
                  control: [adjustGroupCoordinates(controlToAdd, currentData)]
                }, {
                  allowNegativeCoords: true
                });
                return;
              }

            case _diff.VALUE_CHANGES.DELETED:
              {
                const cID = parseInt(controlID);
                this.mainGroupControl.removeChild(cID);
                return;
              }
          }
        });

        if (childrenZIndexChanged) {
          const newOrderedControlIDs = (0, _fp.flow)((0, _fp.sortBy)(function (o) {
            return parseInt(o.zOrder);
          }), (0, _fp.map)(control => parseInt(control.ID)))(_balsamiq.default.get(newData, "mockup.controls.control", []));
          this.mainGroupControl.setChildrenOrder(newOrderedControlIDs);
        }
      }, _constants.OperationOrigin.ARCHIVE);
    }
  }

  getData() {
    return this.export(null, [], false, null, true);
  }
  /**
   * @returns {{ mockup: import("core/serialization/json-types").MockupJSON }}
   */


  getJSONData() {
    return this.export(null, [], false, null, false);
  }

  setTrashed(value) {
    super.setTrashed(value);

    if (!value) {
      for (let id in this.pendingDepsToUpdate) {
        this.onDependencyChanged(id);
      }
    }

    this.pendingDepsToUpdate = {};
  }
  /**
   * Check the controls affected by the resource/subResource change and triggers re-renders of those controls
   * @param {string} resourceID
   * @param {string} origin Origin of the change, ex User or Archive
   * @param {string[]} [subResourceIDs]
   */


  onDependencyChanged(resourceID, origin, subResourceIDs) {
    if (this.isTrashed()) {
      this.pendingDepsToUpdate[resourceID] = true;
      return;
    }

    const controls = subResourceIDs && subResourceIDs.length > 0 ? subResourceIDs.map(c => this.controlsDependencies.getAffectedByResourceChange(resourceID, c.toString())).flat() : this.controlsDependencies.getAffectedByResourceChange(resourceID); // Get changed controls

    const changedControls = [];

    for (let controlCoreID of controls) {
      let control = this.getControl(controlCoreID.controlID); // when symbol that depend on each other it might happen that a control doesn't exist anymore so don't trigger onResourceChange 

      if (control) {
        control.onResourceChange(resourceID);
        changedControls.push(control);
      }
    } // Update bboxes for changed groups


    const groups = this._getAllGroupsFromControls(changedControls);

    groups.forEach(group => {
      group.updateBBoxes();
    });
    this.checkReadyForSnapshot();

    if (controls.length > 0) {
      this.notifyContentsChanged({
        changedCoreObjectIDs: controls
      }, origin);
    }
  }
  /**
   *
   * @param {number[]} p_groupControlID null for main group control, otherwise array of integer controlIDs
   * @param {number[]} controlIDs empty controlIDs array for all controls
   * @param {boolean} doBreakSymbols
   * @param {import("core/model/project").Project} project
   * @param {boolean} asString
   * @param {boolean} doBreakControlsInsideSymbols
   * @returns {{ mockup: import("core/serialization/json-types").MockupJSON } | string} json object or a string
   */


  export(p_groupControlID, controlIDs, doBreakSymbols, project, asString) {
    let doBreakControlsInsideSymbols = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : false;

    if (p_groupControlID != null) {
      _balsamiq.default.assertArray(p_groupControlID);
    } // get the group


    let group = this.getControl(p_groupControlID);

    _balsamiq.default.exists(group);

    let isMockupMainGroup = group === this.getRootControl();
    let exportedObject = (0, _json.saveMockupObject)({
      group,
      controlIDs,
      doBreakSymbols,
      doBreakControlsInsideSymbols,
      project,
      isMockupMainGroup,
      attributes: {
        name: this.getName(),
        order: this.getOrder(),
        parentID: this.getParentID(),
        notes: this.getNotes()
      },
      branchID: this.coreObjectID.branchID,
      resourceID: this.coreObjectID.resourceID
    });

    if (asString) {
      return JSON.stringify(exportedObject, null, 0);
    }

    return exportedObject;
  }
  /**
   *
   * @param {object} p_json
   * @param {number[]} p_groupControlID
   * @param {{ x: number, y: number } | null | undefined} offset
   * @param {boolean} generateNewIDs
   *
   * @returns {{ errors: string[]; controls: any[] }}
   */


  import(p_json, p_groupControlID, offset, generateNewIDs) {
    let result = {
      controls: [],
      errors: []
    };
    let validateResult = (0, _json.validateMockupJSON)(p_json, false);

    if (validateResult.error) {
      result.errors.push(validateResult.error);
      return result;
    }

    let group = this.getControl(p_groupControlID);
    let json = p_json;

    if (offset) {
      json = JSON.parse(JSON.stringify(json));
      let origin = group.getActualProperty("origin");

      for (let control of json.mockup.controls.control) {
        control.x = "" + (parseInt(control.x, 10) + offset.x);
        control.y = "" + (parseInt(control.y, 10) + offset.y);
      }
    } // if (group === this.getRootControl()) {


    if (p_groupControlID.length === 0) {
      // this is a sort of trick because B3 does not consider main group offset
      // clone it because we are going to change it
      json = JSON.parse(JSON.stringify(json));
      let origin = group.getActualProperty("origin");

      for (let control of json.mockup.controls.control) {
        control.x = "" + (parseInt(control.x, 10) - origin.x);
        control.y = "" + (parseInt(control.y, 10) - origin.y);
      }
    }

    let loadResult = group.loadChildren(json.mockup.controls, {
      generateNewIDs,
      allowNegativeCoords: true
    });

    _balsamiq.default.assertArray(loadResult.controls);

    _balsamiq.default.assertArray(loadResult.errors);

    this.updateOriginOnGroupsFromChildToRoot(group);
    this.mainGroupControl.updateAllMacros();
    this.mainGroupControl.updateCropOffsetDescendants(); //this.notifyContentsChanged({ wholeMockupChanged: true }, OperationOrigin.USER);

    loadResult.controls.forEach(control => {
      this._onControlChange(control.coreObjectID, _constants.ControlChanges.ADDED);
    });
    this.notifyDataDirty({
      data: this.getData()
    });
    result.errors = result.errors.concat(loadResult.errors);
    result.controls = result.controls.concat(loadResult.controls);
    return result;
  }

  updateOriginOnGroupsFromChildToRoot(control) {
    if (control.isGroup()) {
      control.children.forEach(c => {
        this.updateOriginOnGroupsFromChildToRoot(c);
      });
      control.updateOrigin();
    }
  }
  /**
   *
   * @param {Record<string, unknown>} json
   */


  load(json) {
    this._load(json);

    this.notifyContentsChanged({
      wholeMockupChanged: true
    }, _constants.OperationOrigin.USER);
    this.notifyDataDirty({
      data: JSON.stringify(json)
    });
  }

  _load(json) {
    // is it better to put the following code inside a performEdit?
    (0, _json.validateMockupJSON)(json);
    this.mainGroupControl.cleanUp(); // TODO rename this method

    this.setDependencies([]);
    this.controlsDependencies = new _controlsdependencies.ControlsDependencies(); // reset the origin of the main group control
    // TODO is it enough to remove the 'origin' property? the default value is (0,0)

    this.mainGroupControl.setActualProperty("origin", _types.default.newPoint(0, 0)); // load all controls

    this.mainGroupControl.loadChildren(json.mockup.controls); // update the origin of the group controls, depending on the children's bbox

    this.updateOriginOnGroupsFromChildToRoot(this.mainGroupControl); // update the macros

    this.mainGroupControl.updateAllMacros();
    this.mainGroupControl.updateCropOffsetDescendants();
    this.loaded = true;
  }

  hasUnknownProperties() {
    return this.mainGroupControl.hasUnknownProperties();
  } // Controls
  // --------


  beginPreview() {
    _balsamiq.default.assert(!this.preview, "Mockup: can't begin another preview");

    this.preview = true;
    this.startPreviewSize = this.getMockupArea();
    this.context.host.log("Mockup: BEGIN preview");
  }

  endPreview(dataDirty, eventData) {
    _balsamiq.default.assert(this.preview, "Mockup: no preview to end"); // call performEdit so it notifies the controlChanges to the svg rendering


    this.performEdit(() => {
      this.updateOriginOnGroupsFromChildToRoot(this.mainGroupControl);
    });
    this.preview = false;
    this.context.host.log("Mockup: END preview, dataDirty: " + dataDirty);
    var endSize = this.getMockupArea();

    if (endSize.width !== this.startPreviewSize.width || endSize.height !== this.startPreviewSize.height) {
      this.onMainGroupControlBBoxChange(false);
    }

    this.notifyContentsChanged({
      endPreview: true,
      ...eventData,
      controlChanges: this.controlsChanges
    }, _constants.OperationOrigin.USER);
    this.controlsChanges = null;

    if (dataDirty) {
      this.notifyDataDirty({
        data: this.getData()
      });
    }
  }
  /**
   *
   * @param {() => void} func
   * @param {import("core/model/constants").OperationOrigin=} operationOrigin
   */


  performEdit(func) {
    let operationOrigin = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : _constants.OperationOrigin.USER;

    _balsamiq.default.assertFunction(func);

    if (operationOrigin == _constants.OperationOrigin.USER) {
      _balsamiq.default.assert(this.context.getUserRole() >= _constants.UserRoles.EDITOR, "Mockup.performEdit can not edit the Mockup because the user role is: " + this.context.getUserRole());
    }

    var startSize = this.getMockupArea();

    if (!this.controlsChanges) {
      this.controlsChanges = new _controlschanges.ControlsChanges();
    }

    func();
    var c = this.controlsChanges; // Get changed controls

    let controlsWithChangedBBox = [];

    for (let key in c.changes) {
      let control = this.getControl(c.changes[key].coreObjectID.controlID);
      controlsWithChangedBBox.push(control);
    } // Update bboxes for changed groups


    let groups = this._getAllGroupsFromControls(controlsWithChangedBBox);

    groups.forEach(group => {
      group.updateAll();
    });

    if (!this.preview) {
      // Update origins and crop data since they may have been changed during move or resize.
      // Set origins to 0,0 and crop offsets to default values calculated from children BBoxes and crop property
      this.updateOriginOnGroupsFromChildToRoot(this.mainGroupControl);
      this.mainGroupControl.updateCropOffsetDescendants();
    }

    if (!this.preview) {
      this.controlsChanges = null;
    }

    this.notifyContentsChanged({
      controlChanges: c
    }, operationOrigin, this.preview);
    let endSize = this.getMockupArea();
    let sizeChanged = endSize.width !== startSize.width || endSize.height !== startSize.height;
    let originChanged = endSize.x !== startSize.x || endSize.y !== startSize.y;

    if (sizeChanged || originChanged) {
      this.onMainGroupControlBBoxChange(this.preview);
    }

    if (!this.preview && operationOrigin == _constants.OperationOrigin.USER) {
      this.notifyDataDirty({
        data: this.getData()
      });
    }
  }

  _getAllGroupsFromControls(controls) {
    let groups = [];

    let addToGroups = function (group) {
      if (groups.indexOf(group) == -1) {
        groups.push(group);
      }
    };

    controls.forEach(control => {
      if (!control) {
        return;
      }

      let parent = control.parent;

      while (parent) {
        addToGroups(parent);
        parent = parent.parent;
      }
    });
    groups.sort(function (a, b) {
      return b.coreObjectID.controlID.length - a.coreObjectID.controlID.length;
    });
    return groups;
  }

  _setControlDependencies(p_coreObjectID, p_srcProperties) {
    this.projectSetDependencies(this, p_coreObjectID, p_srcProperties);
    const result = this.controlsDependencies.setDependency(p_coreObjectID, p_srcProperties); // it may return null in case nothing changed - short circuit

    if (result.changed) {
      // set deps of this resource
      this.setDependencies(result.resourceIDs);
    }
  }

  _onControlChange(coreObjectID, change) {
    //Balsamiq.exists(coreObjectID); disabled for performances
    //Balsamiq.assertEnum(change, ControlChanges); disabled for performances
    //Balsamiq.assert(this.controlsChanges, "the " + change + " of a Control changed because of unknown reason"); enable this only for debugging
    if (this.controlsChanges) {
      this.controlsChanges.setChange(coreObjectID, change);
    }
  }
  /**
   *
   * @param p_coreObjectID
   * @param p_ready
   */


  _setControlReady(p_coreObjectID, p_ready) {
    _balsamiq.default.exists(p_coreObjectID);

    _balsamiq.default.assertBool(p_ready);

    if (p_ready) {
      delete this.controlsNotReady[p_coreObjectID.toString()];
    } else {
      this.controlsNotReady[p_coreObjectID.toString()] = p_coreObjectID;
    }
  }

  checkReadyForSnapshot() {
    var ready = _balsamiq.default.isEmpty(this.controlsNotReady);

    if (!this.isReady && ready) {
      this.isReady = true;

      var e = _events.default.newCoreEvent(this.coreObjectID, _events.default.EventTypes.MOCKUP_READY_FOR_SNAPSHOT);

      this.context.eventCenter.fireEvent(e);
    }
  }

  hasSymbolInstances() {
    //return this.controlsDependencies.hasControlOfType("com.balsamiq.mockups::Component", this.getControl.bind(this));
    let has = false;
    this.controlsDependencies.each(controlCoreObjectID => {
      const control = this.getControl(controlCoreObjectID.controlID);

      _balsamiq.default.exists(control, `control coreObjectID "${controlCoreObjectID}" doesn't exist but it's still in the controlsDependencies`);

      if (control.isSymbol()) {
        has = true;
        return false; // stop iteration
      }

      return true;
    });
    return has;
  }

  getDependenciesOfControlTree(coreObjectID) {
    var _this = this;

    const getDependenciesOfControlTree = function (coreObjectID) {
      let resourceIDs = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];

      const control = _this.mainGroupControl.getControl(coreObjectID.controlID); // add dependencies of itself or its children


      return _balsamiq.default.uniq(resourceIDs.concat(_this.controlsDependencies.getDependenciesOfControl(coreObjectID) || []).concat(...control.children.map(c => getDependenciesOfControlTree(c.coreObjectID, resourceIDs))));
    };

    const resourceIDs = getDependenciesOfControlTree(coreObjectID);
    return resourceIDs;
  } // Areas and Bboxes
  // ----------------


  isEmpty() {
    let rootControl = this.getRootControl();
    return rootControl ? rootControl.children.length === 0 : true;
  }
  /**
   * @returns {{ x: number, y: number, width: number, height: number }}
   */


  getMockupArea() {
    return (0, _mockupsize.clipRectangle)(this.mainGroupControl.getClipBBox(true));
  }
  /**
   * @returns {{ x: number, y: number, width: number, height: number }}
   */


  getMockupAreaNoMarkup() {
    return (0, _mockupsize.clipRectangle)(this.mainGroupControl.getClipBBox(false));
  }
  /**
   * @returns {{ x: number, y: number, width: number, height: number }}
   */


  getMockupUserArea() {
    return (0, _mockupsize.clipRectangle)(this.mainGroupControl.getUserBBox(true));
  }
  /**
   * @returns {{ x: number, y: number, width: number, height: number }}
   */


  getMockupUserAreaNoMarkup() {
    return (0, _mockupsize.clipRectangle)(this.mainGroupControl.getUserBBox(false));
  }

  onMainGroupControlBBoxChange(p_preview) {
    _balsamiq.default.assertBool(p_preview);

    var type = p_preview ? _events.default.EventTypes.MOCKUP_SIZE_PREVIEW : _events.default.EventTypes.MOCKUP_SIZE_CHANGED;

    var e = _events.default.newCoreEvent(this.coreObjectID, type, {});

    this.context.eventCenter.fireEvent(e);
  }

  onResourceChange(p_url, p_ticket, p_state, p_data) {
    //this.context.logger("Mockup.onResourceChange " + p_url + ", " + p_ticket + ", " + p_state + " project" + this.coreObjectID.projectID + " mockup" + this.coreObjectID.mockupID);
    if (this.controlsDependencies.hasControl(p_ticket)) {
      var control = this.getControl(this.controlsDependencies.getControl(p_ticket).controlID);

      _balsamiq.default.assert(control, "cannot find a control to notify a resource change " + this.controlsDependencies[p_ticket].coreID.toString());

      control.onResourceChange();
      this.checkReadyForSnapshot();
    }
  }

}

exports.Mockup = Mockup;

/***/ }),

/***/ "./js/core/model/pdfgenerator.js":
/*!***************************************!*\
  !*** ./js/core/model/pdfgenerator.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.PdfGenerator = exports.PDFPaperSize = exports.PDFOrientation = exports.PDFOptionsSize = exports.MockupSelection = void 0;
exports.defaultPdfOptions = defaultPdfOptions;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Mockup = _interopRequireWildcard(__webpack_require__(/*! core/model/mockup */ "./js/core/model/mockup.js"));

var Project = _interopRequireWildcard(__webpack_require__(/*! core/model/project */ "./js/core/model/project.js"));

var Resource = _interopRequireWildcard(__webpack_require__(/*! core/model/resource */ "./js/core/model/resource.js"));

var SimpleRenderer = _interopRequireWildcard(__webpack_require__(/*! core/graphics/simplerenderer */ "./js/core/graphics/simplerenderer.js"));

var LinkAreas = _interopRequireWildcard(__webpack_require__(/*! core/graphics/linkareas */ "./js/core/graphics/linkareas.js"));

var Constants = _interopRequireWildcard(__webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts"));

var textMod = _interopRequireWildcard(__webpack_require__(/*! core/texts/text */ "./js/core/texts/text.js"));

var BalsamiqArchiveConstants = _interopRequireWildcard(__webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchiveConstants */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var _async = __webpack_require__(/*! core/utils/async */ "./js/core/utils/async.js");

var _commands = __webpack_require__(/*! core/graphics/commands */ "./js/core/graphics/commands.js");

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const AnnotationsFontFamily = 'Arial';
const AnnotationsFontSize = 12;
const AnnotationsTextColor = 0x696969;
const PDFOptionsSize = {
  OPTIMIZEFORSCREEN: "optimizeForScreen",
  FITAPAGE: "fitAPage"
};
exports.PDFOptionsSize = PDFOptionsSize;
const PDFPaperSize = {
  A3: "A3",
  A4: "A4",
  A5: "A5",
  Letter: "Letter",
  Legal: "Legal",
  Tabloid: "Tabloid"
};
exports.PDFPaperSize = PDFPaperSize;
const PDFOrientation = {
  fit: "fit",
  portrait: "portrait",
  landscape: "landscape"
};
exports.PDFOrientation = PDFOrientation;
const MockupSelection = {
  all: "all",
  current: "current",
  set: "set"
};
exports.MockupSelection = MockupSelection;

function defaultPdfOptions() {
  return {
    pdfSize: PDFOptionsSize.FITAPAGE,
    pdfNoMargins: false,
    pdfPaperSize: PDFPaperSize.A4,
    pdfOrientation: PDFOrientation.fit,
    pdfShowLinkingHints: false,
    pdfAddMockupNames: false,
    pdfAddPageNumbers: false,
    pdfIncludeMockupNotes: false
  };
}

; // copied from PDFKit

const PAGE_SIZES = {
  '4A0': [4767.87, 6740.79],
  '2A0': [3370.39, 4767.87],
  A0: [2383.94, 3370.39],
  A1: [1683.78, 2383.94],
  A2: [1190.55, 1683.78],
  A3: [841.89, 1190.55],
  A4: [595.28, 841.89],
  A5: [419.53, 595.28],
  A6: [297.64, 419.53],
  A7: [209.76, 297.64],
  A8: [147.40, 209.76],
  A9: [104.88, 147.40],
  A10: [73.70, 104.88],
  B0: [2834.65, 4008.19],
  B1: [2004.09, 2834.65],
  B2: [1417.32, 2004.09],
  B3: [1000.63, 1417.32],
  B4: [708.66, 1000.63],
  B5: [498.90, 708.66],
  B6: [354.33, 498.90],
  B7: [249.45, 354.33],
  B8: [175.75, 249.45],
  B9: [124.72, 175.75],
  B10: [87.87, 124.72],
  C0: [2599.37, 3676.54],
  C1: [1836.85, 2599.37],
  C2: [1298.27, 1836.85],
  C3: [918.43, 1298.27],
  C4: [649.13, 918.43],
  C5: [459.21, 649.13],
  C6: [323.15, 459.21],
  C7: [229.61, 323.15],
  C8: [161.57, 229.61],
  C9: [113.39, 161.57],
  C10: [79.37, 113.39],
  RA0: [2437.80, 3458.27],
  RA1: [1729.13, 2437.80],
  RA2: [1218.90, 1729.13],
  RA3: [864.57, 1218.90],
  RA4: [609.45, 864.57],
  SRA0: [2551.18, 3628.35],
  SRA1: [1814.17, 2551.18],
  SRA2: [1275.59, 1814.17],
  SRA3: [907.09, 1275.59],
  SRA4: [637.80, 907.09],
  EXECUTIVE: [521.86, 756.00],
  FOLIO: [612.00, 936.00],
  LEGAL: [612.00, 1008.00],
  LETTER: [612.00, 792.00],
  TABLOID: [792.00, 1224.00]
};

const getControlsWithUnsupportedGlyphs = (glyphsInFont, asyncfy, controls, getControlsCallback) => {
  (0, _async.filter)(asyncfy, controls, (control, filterDoneCb) => {
    const textDisplayLists = _balsamiq.default.filter(control.displayListItems, {
      displayItemType: "Text"
    });

    return (0, _async.any)(asyncfy, textDisplayLists, (textDisplayList, anyCallback) => {
      const text = TextUtils.unformatNodes(textDisplayList.textDom);
      glyphsInFont(textDisplayList.fontFamily, textDisplayList.bold, textDisplayList.italic, text, _ref => {
        let {
          res
        } = _ref;
        return anyCallback(!res);
      });
    }, filterDoneCb);
  }, getControlsCallback);
};

const filterByTextType = controlIdsAndDisplayItems => _balsamiq.default.filter(controlIdsAndDisplayItems, c => c.displayListItems.some(d => d.displayItemType === "Text"));

const getControlsDisplayData = displayNodes => {
  return displayNodes.concat(displayNodes.map(dn => getControlsDisplayData(dn.children)).flat());
};

const getControlIdsAndDisplayItems = mockup => {
  const displayNodes = getControlsDisplayData(mockup.getDisplayDataTree().children);
  const displayItems = displayNodes.map(c => ({
    controlID: c.controlID,
    displayListItems: c.displayData.displayListItems
  }));
  const flatDisplayItems = displayItems.flat();
  return flatDisplayItems;
};

const getControlsWithUnsupportedGlyphsFromMockups = (glyphsInFont, asyncfy, mockup, callback) => _lodash.default.flow(getControlIdsAndDisplayItems, filterByTextType, controls => getControlsWithUnsupportedGlyphs(glyphsInFont, asyncfy, controls, controlIdsAndDisplayItems => {
  const controlIDssWithUnsupportedGlyphs = controlIdsAndDisplayItems.map(c => mockup.getControl(c.controlID));
  callback(null, controlIDssWithUnsupportedGlyphs);
}))(mockup);

class PdfGenerator {
  constructor(context) {
    this.context = context;

    _balsamiq.default.exists(this.context.host);

    _balsamiq.default.exists(this.context.eventCenter);

    this.pdfID = null;
    this.options = null;
    this.simpleRenderer = new SimpleRenderer.SimpleRenderer(this.context);
    this.useMarkup = true;
  }

  setShowMarkup(value) {
    this.simpleRenderer.setShowMarkup(value);
    this.useMarkup = value;
  }

  start(options, projectFontFace, callback) {
    if (callback) {
      _balsamiq.default.assertFunction(callback);
    }

    this.options = options;
    let fontFamilies = [AnnotationsFontFamily, 'FontAwesome'];
    fontFamilies.push(projectFontFace);
    this.context.host.PDF.start(options, fontFamilies, data => {
      this.pdfID = data.id;

      if (callback) {
        callback();
      }
    });
  }

  save(saveParams, callback) {
    _balsamiq.default.assertObject(saveParams); //Balsamiq.existMulti("PDF save params is missing something", saveParams.fileName, saveParams.ext);


    if (callback) {
      _balsamiq.default.assertFunction(callback);
    }

    this.context.host.PDF.save(this.pdfID, saveParams, res => {
      // notify the caller that the update is done
      if (callback) {
        callback(res);
      }
    });
  }

  getData(callback) {
    _balsamiq.default.assertFunction(callback);

    this.context.host.PDF.getData(this.pdfID, callback);
  }

  findMockupIndex(mockup, collection) {
    return this.findMockupIndexByID(mockup.coreObjectID.resourceID, mockup.coreObjectID.branchID, collection);
  }

  findMockupIndexByID(resourceID, branchID, collection) {
    for (let i = 0; i < collection.length; ++i) {
      const otherMockup = collection[i];

      if (resourceID === otherMockup.coreObjectID.resourceID && branchID === otherMockup.coreObjectID.branchID) {
        return i;
      }
    }

    return -1;
  }

  findTargetPage(mockup, target, mockups) {
    const resourceID = target.replace('resource:', '');
    const mockupOnMasterIndex = this.findMockupIndexByID(resourceID, Constants.Master, mockups);

    if (mockupOnMasterIndex != -1) {
      return mockupOnMasterIndex;
    } // ok lets find the first one available


    for (let i = 0; i < mockups.length; ++i) {
      const otherMockup = mockups[i];

      if (resourceID === otherMockup.coreObjectID.resourceID) {
        return i;
      }
    }

    return -1;
  }

  createSnapshot(mockup, mockups, callback, project) {
    _balsamiq.default.assertArray(mockups);

    _balsamiq.default.existsClass(mockup, Mockup.Mockup);

    _balsamiq.default.existsClass(project, Project.Project);

    callback && _balsamiq.default.assertFunction(callback);
    const totalPages = mockups.length;
    const currentPage = this.findMockupIndex(mockup, mockups);

    const create = function () {
      const rootBounds = mockup.getRootControl().getClipBBox(this.useMarkup);
      const namePageMargin = this.options.pdfAddMockupNames || this.options.pdfAddPageNumbers ? 26 : 0;
      const text = mockup.getKeyValue(Resource.KeyNames.NOTES);
      const notesMargin = this.options.pdfIncludeMockupNotes && text ? 10 : 0;
      const [marginX, marginY] = this.options.pdfSize === PDFOptionsSize.OPTIMIZEFORSCREEN && this.options.pdfNoMargins ? [0, 0] : [10, 10];
      const annotationsMarginX = marginX;
      const annotationsMarginY = marginY;
      const mockupWidth = rootBounds.width;
      const mockupHeight = rootBounds.height;
      let pageWidth;
      let pageHeight;
      let PDFPageWidth;
      let PDFPageHeight;
      let scaleToFit = 1;
      let layout = 'portrait';

      if (this.options.pdfSize === PDFOptionsSize.FITAPAGE) {
        layout = this.options.pdfOrientation === PDFOrientation.fit ? mockupWidth > mockupHeight ? 'landscape' : 'portrait' : this.options.pdfOrientation;
        const dimensions = PAGE_SIZES[this.options.pdfPaperSize.toUpperCase()];
        pageWidth = dimensions[layout === 'portrait' ? 0 : 1];
        pageHeight = dimensions[layout === 'portrait' ? 1 : 0];
        PDFPageWidth = dimensions[0];
        PDFPageHeight = dimensions[1];
      } else {
        pageWidth = mockupWidth + 2 * marginX;
        pageHeight = mockupHeight + 2 * marginY + namePageMargin + notesMargin;
        PDFPageWidth = pageWidth;
        PDFPageHeight = pageHeight;
      } // take notes into account


      let mockupNotesHeight = 0;
      let textNote = null;
      let textNoteSize;

      if (this.options.pdfIncludeMockupNotes) {
        if (text) {
          const textStyle = {
            fontFamily: AnnotationsFontFamily,
            fontSize: AnnotationsFontSize,
            textColor: AnnotationsTextColor,
            bold: false,
            italic: false,
            underline: false
          };
          const context = {
            textMeter: mockup.context.textMeter,
            host: this.context.host,
            eventCenter: this.context.eventCenter
          };
          textNote = textMod.newText(text, textStyle, context);
          textNoteSize = textNote.measuredSizeForWidth(pageWidth - 2 * marginX);
          textNoteSize.width += 4; // this is because the PDF could crop or do a line break if the width is just half a pixel narrower than expected

          mockupNotesHeight = textNoteSize.height;
        }
      } // now compute the mockup dimensions in print
      // apply corrections for notes


      if (this.options.pdfSize === PDFOptionsSize.FITAPAGE) {
        const ratioW = mockupWidth / (pageWidth - 2 * marginX);
        const ratioH = mockupHeight / (pageHeight - 2 * marginY - mockupNotesHeight - namePageMargin - notesMargin);
        scaleToFit = ratioW > ratioH ? 1. / ratioW : 1. / ratioH;

        if (scaleToFit > 1) {
          scaleToFit = 1;
        }
      } else {
        PDFPageHeight += mockupNotesHeight;
      }

      const drawControlImage = (control, imageDrawnCallback) => {
        this.context.host.Images.make(_ref2 => {
          let {
            id
          } = _ref2;
          const imageCommands = this.simpleRenderer.renderMultipleControls({
            controls: [control],
            margins: [0, 0],
            maxSize: -1,
            dpiScale: 2,
            // increase the image quality by specifying the dpiScale
            whiteBackground: false
          });
          this.context.host.Images.drawCommands(id, imageCommands, () => {
            imageDrawnCallback({
              controlID: control.coreObjectID.toString(),
              imageId: id
            });
          });
        });
      };

      const getControlsToBeRenderedAsBitmaps = doneCallback => getControlsWithUnsupportedGlyphsFromMockups(this.context.host.Fonts.glyphsInFont.bind(this.context.host.Fonts), function fakeAsyncfy(fn) {
        fn();
      }, mockup, doneCallback);

      const createImagesForNotSupportedControls = (controlsToBeRenderedAsBitmaps, createImageCallback) => {
        (0, _async.eachParallel)(this.context.host.asyncCallback, controlsToBeRenderedAsBitmaps, drawControlImage, controlImages => createImageCallback(null, {
          controlImagesMap: _lodash.default.reduce(controlImages, (controlImagesMap, _ref3) => {
            let {
              controlID,
              imageId
            } = _ref3;
            return { ...controlImagesMap,
              [controlID]: imageId
            };
          }, {})
        }));
      };

      const makeCommandsForWireframeNames = (payload, done) => {
        const branchId = mockup.coreObjectID.branchID;
        const isMaster = branchId === BalsamiqArchiveConstants.Branch.MasterBranchID;
        const pageName = mockup.getName() + (isMaster ? '' : ' (' + project.branches[branchId].getName() + ')');

        if (this.options.pdfAddMockupNames) {
          this.context.host.Fonts.glyphsInFont(AnnotationsFontFamily, true, false, pageName, _ref4 => {
            let {
              res
            } = _ref4;
            const mockupNamesCommands = [_commands.GraphicCommands.SAVE, _commands.GraphicCommands.TRANSLATE, annotationsMarginX + 1, annotationsMarginY, _commands.GraphicCommands.TEXT_STYLE, AnnotationsFontFamily, 13, true, false, false, 6710886, _commands.GraphicCommands.TEXT, pageName, 4, 10.236220472440944, _commands.GraphicCommands.RESTORE];

            if (res) {
              done(null, { ...payload,
                mockupNamesCommands,
                pageName
              });
            } else {
              (0, SimpleRenderer.renderTextAsImage)({
                textStyle: {
                  fontFamily: AnnotationsFontFamily,
                  fontSize: 13,
                  textColor: 6710886,
                  bold: true,
                  italic: false,
                  underline: false
                },
                context: this.context,
                text: pageName,
                margin: {
                  x: annotationsMarginX + 1,
                  y: annotationsMarginY
                }
              }, mockupNamesCommands => {
                done(null, { ...payload,
                  mockupNamesCommands,
                  pageName
                });
              });
            }
          });
        } else {
          done(null, { ...payload,
            mockupNamesCommands: null,
            pageName
          });
        }
      };

      const addPage = (_ref5, addPageCallback) => {
        let {
          controlImagesMap,
          mockupNamesCommands,
          pageName
        } = _ref5;
        return this.context.host.PDF.addPage(this.pdfID, PDFPageWidth, PDFPageHeight, layout, pageName, () => {
          let commands = this.simpleRenderer.renderPage({
            rootControl: mockup.getRootControl(),
            scale: scaleToFit,
            margins: [marginX / scaleToFit, (marginY + namePageMargin) / scaleToFit],
            //Fix the top left coords because they get multiplied by scale! #167016743
            controlIDsToRenderAsImages: controlImagesMap
          });
          let displayData = mockup.getRootControl().getDisplayDataTree(false);
          let la = new LinkAreas.LinkAreas({});
          la.build(displayData, true);
          let addLinkCommands = [_commands.GraphicCommands.SAVE, _commands.GraphicCommands.SCALE, scaleToFit, scaleToFit, _commands.GraphicCommands.TRANSLATE, marginX / scaleToFit, (marginY + namePageMargin) / scaleToFit];
          const clipBBoxDelta = {
            x: 0,
            y: 0
          };

          if (!this.useMarkup) {
            const clipBBoxWithMarkup = mockup.getRootControl().getClipBBox(true);
            const clipBBoxWithoutMarkup = mockup.getRootControl().getClipBBox(false);
            clipBBoxDelta.x = clipBBoxWithoutMarkup.x - clipBBoxWithMarkup.x;
            clipBBoxDelta.y = clipBBoxWithoutMarkup.y - clipBBoxWithMarkup.y;
          }

          let links = la.getAreas();

          for (let i = 0; i < links.length; ++i) {
            const bounds = links[i].bounds;
            const target = links[i].target;
            let rotation = _balsamiq.default.isNumber(links[i].rotation) ? -links[i].rotation * Math.PI / 180 : 0;
            let rotationDeg = _balsamiq.default.isNumber(links[i].rotation) ? links[i].rotation : 0;
            let xStart = bounds.x - clipBBoxDelta.x;
            let yStart = bounds.y - clipBBoxDelta.y;
            let theTarget;
            let drawLinks;

            let rotationCenter = _types.default.newPoint(xStart, yStart);

            let rotatedTopLeft = _types.default.rotatePtDeg(rotationCenter, _types.default.newPoint(xStart, yStart), rotationDeg);

            let rotatedTopRight = _types.default.rotatePtDeg(rotationCenter, _types.default.newPoint(xStart + bounds.width, yStart), rotationDeg);

            let rotatedBottomLeft = _types.default.rotatePtDeg(rotationCenter, _types.default.newPoint(xStart, yStart + bounds.height), rotationDeg);

            let rotatedBottomRight = _types.default.rotatePtDeg(rotationCenter, _types.default.newPoint(xStart + bounds.width, yStart + bounds.height), rotationDeg);

            let largerLeft = Math.min(rotatedTopLeft.x, rotatedBottomLeft.x);
            let largerTop = Math.min(rotatedTopLeft.y, rotatedTopRight.y, rotatedBottomLeft.y, rotatedBottomRight.y);
            let largerWidth = Math.max(rotatedBottomRight.x, rotatedTopRight.x) - largerLeft;
            let largerHeight = Math.max(rotatedBottomLeft.y, rotatedBottomRight.y) - largerTop;

            if (target.startsWith("resource:")) {
              // find the mockup
              theTarget = this.findTargetPage(mockup, target, mockups);

              if (theTarget == -1) {
                continue;
              }

              drawLinks = [_commands.GraphicCommands.LINK_PAGE, largerLeft, largerTop, largerWidth, largerHeight, theTarget];
            } else {
              theTarget = target;
              drawLinks = [_commands.GraphicCommands.LINK, largerLeft, largerTop, largerWidth, largerHeight, theTarget];
            }

            if (this.options.pdfShowLinkingHints) {
              let drawLinkHints;

              if (this.options.pdfSize === PDFOptionsSize.OPTIMIZEFORSCREEN) {
                // Draw rectangular red area
                let topLeft = {
                  x: 0,
                  y: 0
                };
                let topRight = {
                  x: bounds.width,
                  y: 0
                };
                let bottomRight = {
                  x: bounds.width,
                  y: bounds.height
                };
                let bottomLeft = {
                  x: 0,
                  y: bounds.height
                };
                drawLinkHints = [_commands.GraphicCommands.SAVE, _commands.GraphicCommands.TRANSLATE, xStart, yStart, _commands.GraphicCommands.ROTATE, rotation, _commands.GraphicCommands.BEGIN, _commands.GraphicCommands.MOVE, topLeft.x, topLeft.y, _commands.GraphicCommands.LINE, topRight.x, topRight.y, _commands.GraphicCommands.LINE, bottomRight.x, bottomRight.y, _commands.GraphicCommands.LINE, bottomLeft.x, bottomLeft.y, _commands.GraphicCommands.LINE, topLeft.x, topLeft.y, _commands.GraphicCommands.DRAW, 0xE06666, 0.5, 0xE06666, 0.5, _commands.GraphicCommands.RESTORE];
              } else {
                // Draw rotated gray arrow
                let s = 1.5;
                let tx = bounds.width - 7 * s;
                let ty = bounds.height - 6 * s;
                drawLinkHints = [_commands.GraphicCommands.SAVE, _commands.GraphicCommands.TRANSLATE, xStart, yStart, _commands.GraphicCommands.ROTATE, rotation, _commands.GraphicCommands.BEGIN, _commands.GraphicCommands.MOVE, tx + 4 * s, ty, _commands.GraphicCommands.LINE, tx + 7 * s, ty + 3 * s, _commands.GraphicCommands.LINE, tx + 4 * s, ty + 6 * s, _commands.GraphicCommands.LINE, tx + 4 * s, ty + 4 * s, _commands.GraphicCommands.LINE, tx, ty + 4 * s, _commands.GraphicCommands.LINE, tx, ty + 2 * s, _commands.GraphicCommands.LINE, tx + 4 * s, ty + 2 * s, _commands.GraphicCommands.LINE, tx + 4 * s, ty, _commands.GraphicCommands.DRAW, 3355443, 0.5, 16777215, 1, _commands.GraphicCommands.RESTORE];
              }

              addLinkCommands = addLinkCommands.concat(drawLinkHints);
            }

            addLinkCommands = addLinkCommands.concat(drawLinks);
          }

          addLinkCommands.push(_commands.GraphicCommands.RESTORE);
          commands = commands.concat(addLinkCommands);

          if (mockupNamesCommands) {
            commands = commands.concat(mockupNamesCommands);
          }

          if (this.options.pdfAddPageNumbers) {
            let addPageNumbersCommand = [_commands.GraphicCommands.SAVE, _commands.GraphicCommands.TRANSLATE, 0, annotationsMarginY, _commands.GraphicCommands.TEXT_STYLE, AnnotationsFontFamily, 13, true, false, false, 6710886, _commands.GraphicCommands.TEXT_RTL, currentPage + 1 + " / " + totalPages, layout === "portrait" ? pageWidth - annotationsMarginX - 1 : pageWidth - annotationsMarginX - 1, 10.236220472440944, _commands.GraphicCommands.RESTORE];
            commands = commands.concat(addPageNumbersCommand);
          } // finally create the notes display list


          if (textNote) {
            const displayItems = textNote.makeDisplayItems({
              x: 0,
              y: 0,
              width: textNoteSize.width,
              height: textNoteSize.height
            });
            const textNoteDisplayData = {
              controlID: -1,
              bounds: {
                x: marginX,
                y: namePageMargin + mockupHeight * scaleToFit + marginY + notesMargin,
                width: pageWidth - 2 * marginX,
                height: mockupNotesHeight
              },
              displayListItems: displayItems,
              displayListSkinName: "wireframe",
              masks: [],
              colorTransforms: null,
              clipOnBbox: false,
              linkAreas: false,
              isMarkup: false
            };
            const textNoteCommands = this.simpleRenderer.renderDisplayData(textNoteDisplayData, null, true);
            commands = commands.concat(textNoteCommands);
          }

          addPageCallback(null, commands);
        });
      };

      const executePDFCommands = pdfID => (commands, callback) => this.context.host.PDF.commands(pdfID, commands, callback);

      (0, _async.queue)([getControlsToBeRenderedAsBitmaps, createImagesForNotSupportedControls, makeCommandsForWireframeNames, addPage, executePDFCommands(this.pdfID)], () => {
        // notify the caller that the update is done
        callback && callback();
      }, {
        asyncfy: this.context.host.asyncCallback
      });
    }.bind(this);

    this.context.host.asyncCallback(create);
  }

}

exports.PdfGenerator = PdfGenerator;

/***/ }),

/***/ "./js/core/model/permalinks.js":
/*!*************************************!*\
  !*** ./js/core/model/permalinks.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Permalinks = exports.PERMALINKS_ID_TYPE = exports.KeyNames = void 0;
exports.assertPermalinkType = assertPermalinkType;
exports.makePermalinksCoreObjectID = makePermalinksCoreObjectID;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var CoreObject = _interopRequireWildcard(__webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const PERMALINKS_ID_TYPE = "permalinks";
exports.PERMALINKS_ID_TYPE = PERMALINKS_ID_TYPE;

function makePermalinksCoreObjectID(projectCoreObjectID) {
  CoreObject.assertProjectType(projectCoreObjectID);
  var newID = projectCoreObjectID.clone();
  newID.type = PERMALINKS_ID_TYPE;
  return newID;
}

function assertPermalinkType(coreObjectID) {
  _balsamiq.default.assert(coreObjectID.type == PERMALINKS_ID_TYPE);
}

const KeyNames = {
  PERMALINKS: 'permalinks'
};
exports.KeyNames = KeyNames;

class Permalinks extends CoreObject.CoreObject {
  constructor(context, coreObjectID) {
    _balsamiq.default.exists(context);

    _balsamiq.default.exists(context.host);

    assertPermalinkType(coreObjectID);
    super(context, coreObjectID);
    this._permalinks = new Map(); // map[resource][branch] = permalink
  }

  getKeyValue(keyName) {
    if (keyName === KeyNames.PERMALINKS) {
      return this;
    }

    return super.getKeyValue(keyName);
  } // permalinks: [{branchID:string, resourceID:string, permalinkID:string, image: string (url)}, ...]


  init(permalinks) {
    this._permalinks = new Map();

    for (let p of permalinks) {
      this._addOrUpdateHelper(p);
    }

    this.notifyKeyValueChanged(KeyNames.PERMALINKS);
  } // permalink: {branchID:string, resourceID:string, permalinkID:string, image: string (url)}


  addOrUpdate(permalink) {
    this._addOrUpdateHelper(permalink);

    this.notifyKeyValueChanged(KeyNames.PERMALINKS);
  }

  delete(_ref) {
    let {
      resourceID,
      branchID
    } = _ref;

    let branches = this._permalinks.get(resourceID);

    if (!branches) {
      return;
    }

    branches.delete(branchID);
    this.notifyKeyValueChanged(KeyNames.PERMALINKS);
  }

  _addOrUpdateHelper(permalink) {
    const {
      branchID,
      resourceID,
      permalinkID,
      image = '',
      dirty
    } = permalink;

    _balsamiq.default.assertString(branchID);

    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.assertString(permalinkID);

    _balsamiq.default.assertString(image);

    _balsamiq.default.assertBool(dirty);

    let permalinkBranchMap = this._permalinks.get(resourceID) || new Map();
    permalinkBranchMap.set(branchID, {
      branchID,
      resourceID,
      permalinkID,
      image,
      dirty
    });

    this._permalinks.set(resourceID, permalinkBranchMap);
  }

  getResourcePermalink(coreObjectID) {
    _balsamiq.default.exists(coreObjectID);

    let branches = this._permalinks.get(coreObjectID.resourceID);

    if (!branches) {
      return null;
    }

    return branches.get(coreObjectID.branchID) || null;
  }

  getIDsOfBranchesWithPermalinks(resourceID) {
    _balsamiq.default.assertString(resourceID);

    const branchPermalinks = this._permalinks.get(resourceID);

    return branchPermalinks ? [...branchPermalinks.keys()] : [];
  }

  hasPermalinkFor(_ref2) {
    let {
      resourceID,
      branchID
    } = _ref2;

    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.assertString(branchID);

    const branchPermalinks = this._permalinks.get(resourceID);

    return branchPermalinks ? branchPermalinks.has(branchID) : false;
  }

  asList() {
    let res = [];

    for (let [_, branches] of this._permalinks) {
      for (let [_, permalink] of branches) {
        res.push(permalink);
      }
    }

    return res;
  }

  getDirty() {
    return this.asList().filter(p => p.dirty === true);
  }

}

exports.Permalinks = Permalinks;

/***/ }),

/***/ "./js/core/model/project.js":
/*!**********************************!*\
  !*** ./js/core/model/project.js ***!
  \**********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Roles = exports.Project = exports.KeyNames = exports.EventTypes = exports.DEFAULT_SKINNAME = exports.DEFAULT_SELECTION_COLOR = exports.DEFAULT_LINK_COLOR = exports.DEFAULT_FONTSIZE = exports.DEFAULT_FONTFACE = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Branch = _interopRequireWildcard(__webpack_require__(/*! core/model/branch */ "./js/core/model/branch.js"));

var Mockup = _interopRequireWildcard(__webpack_require__(/*! core/model/mockup */ "./js/core/model/mockup.js"));

var User = _interopRequireWildcard(__webpack_require__(/*! core/model/user */ "./js/core/model/user.js"));

var Comment = _interopRequireWildcard(__webpack_require__(/*! core/model/comment */ "./js/core/model/comment.js"));

var Message = _interopRequireWildcard(__webpack_require__(/*! core/model/message */ "./js/core/model/message.js"));

var SymbolLibrary = _interopRequireWildcard(__webpack_require__(/*! core/model/symbollibrary */ "./js/core/model/symbollibrary.js"));

var Thumbnail = _interopRequireWildcard(__webpack_require__(/*! core/model/thumbnail */ "./js/core/model/thumbnail.ts"));

var Asset = _interopRequireWildcard(__webpack_require__(/*! core/model/asset */ "./js/core/model/asset.js"));

var Resource = _interopRequireWildcard(__webpack_require__(/*! core/model/resource */ "./js/core/model/resource.js"));

var Constants = _interopRequireWildcard(__webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts"));

var Image = _interopRequireWildcard(__webpack_require__(/*! core/model/image */ "./js/core/model/image.js"));

var ThumbnailQueue = _interopRequireWildcard(__webpack_require__(/*! core/model/thumbnailqueue */ "./js/core/model/thumbnailqueue.js"));

var ConnectedUsers = _interopRequireWildcard(__webpack_require__(/*! core/model/connectedusers */ "./js/core/model/connectedusers.js"));

var ArchiveState = _interopRequireWildcard(__webpack_require__(/*! core/model/archivestate */ "./js/core/model/archivestate.js"));

var DirtyWireframes = _interopRequireWildcard(__webpack_require__(/*! core/model/dirtywireframes */ "./js/core/model/dirtywireframes.js"));

var _sketch = _interopRequireDefault(__webpack_require__(/*! core/skins/sketch */ "./js/core/skins/sketch.js"));

var MimeTypes = _interopRequireWildcard(__webpack_require__(/*! core/mimetypes */ "./js/core/mimetypes.js"));

var CoreObject = _interopRequireWildcard(__webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js"));

var Events = _interopRequireWildcard(__webpack_require__(/*! core/events */ "./js/core/events.ts"));

var BarConstants = _interopRequireWildcard(__webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchiveConstants */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js"));

var _BmprConstants = __webpack_require__(/*! @balsamiq/bmpr/lib/BmprConstants */ "./node_modules/@balsamiq/bmpr/lib/BmprConstants.js");

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var async = _interopRequireWildcard(__webpack_require__(/*! core/utils/async */ "./js/core/utils/async.js"));

var _resourcetree = __webpack_require__(/*! core/utils/resourcetree */ "./js/core/utils/resourcetree.js");

var _locators = __webpack_require__(/*! core/model/locators */ "./js/core/model/locators.js");

var Permalinks = _interopRequireWildcard(__webpack_require__(/*! core/model/permalinks */ "./js/core/model/permalinks.js"));

var _licensedata = __webpack_require__(/*! core/license/licensedata */ "./js/core/license/licensedata.js");

var _textutilities = __webpack_require__(/*! ../texts/textutilities */ "./js/core/texts/textutilities.js");

var _reorderAddedResources = __webpack_require__(/*! core/utils/reorderAddedResources */ "./js/core/utils/reorderAddedResources.ts");

var _symbolMap = __webpack_require__(/*! core/symbol-map */ "./js/core/symbol-map.ts");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var MASTER = BarConstants.Branch.MasterBranchID;
const DEFAULT_FONTSIZE = 13;
exports.DEFAULT_FONTSIZE = DEFAULT_FONTSIZE;
const DEFAULT_FONTFACE = "Balsamiq Sans";
exports.DEFAULT_FONTFACE = DEFAULT_FONTFACE;
const DEFAULT_SKINNAME = _sketch.default["#skinName"];
exports.DEFAULT_SKINNAME = DEFAULT_SKINNAME;
const DEFAULT_SELECTION_COLOR = 0x95BCF2;
exports.DEFAULT_SELECTION_COLOR = DEFAULT_SELECTION_COLOR;
const DEFAULT_LINK_COLOR = 0x085394;
exports.DEFAULT_LINK_COLOR = DEFAULT_LINK_COLOR;
var EventTypes = {
  PROJECT_LOADED: "projectLoaded",
  PROJECT_ATTRIBUTES_DIRTY: "projectAttributesDirty",
  RESOURCE_ADDED: "resourceAdded",
  RESOURCES_DELETED: "resourcesDeleted",
  THUMBNAIL_ADDED: "thumbnailAdded",
  THUMBNAILS_DELETED: "thumbnailsDeleted",
  BRANCH_ADDED: "branchAdded",
  RESOURCE_BRANCHID_CHANGED: "resourceBranchidChanged",
  BRANCHES_DELETED: "branchesDeleted",
  COMMENT_ADDED: "comment_added",
  COMMENT_IMPORTED: "comment_imported",
  COMMENTS_ATTRIBUTES_UPDATED: "commentsAttributesUpdated",
  COMMENT_DATA_UPDATED: "commentDataUpdated",
  COMMENTS_DELETED: "comments_deleted",
  MYUSER_CREATED: "myuser_created",
  MYUSER_UPDATED: "myuser_updated",
  MESSAGE_TO_BROADCAST: "messageToBroadcast",
  PERMALINK_DELETED: "deletePermalinkData",
  UNKNOWN_PROPERTIES_LOADED: "unknownPropertiesLoaded",
  CONTROLS_WITH_LOOPS: "controlsWithLoops"
};
exports.EventTypes = EventTypes;
var KeyNames = {
  NAME: "name",
  CREATION_DATE: "creationDate",
  DESCRIPTION: "description",
  SELECTION_COLOR: "selectionColor",
  FONT_FACE: "fontFace",
  FONT_SIZE: "fontSize",
  SKIN_NAME: "skinName",
  LINK_COLOR: "linkColor",
  USER_ROLE: "userRole",
  DEFAULT_STAFF_ROLE: "defaultStaffRole",
  // lists
  BRANCH_LIST: "branchList",
  MOCKUP_LIST: "mockupList",
  MOCKUP_LIST_FLAT: "mockupListFlat",
  ASSET_LIST: "assetList",
  SYMBOLLIBRARY_LIST: "symbolLibraryList",
  TRASH_LIST: "trashList",
  COMMENTS: "comments",
  USERS: "users",
  MESSAGES: "chatMessages",
  MENTIONABLE_USERS: "mentionableUsers",
  // other stuff
  HAS_COMING_USER_EVENTS: "hasComingUserEvents",
  IS_LOADING: "isLoading",
  LICENSE: "license",
  PROJECTLOCATOR: "projectLocator"
};
exports.KeyNames = KeyNames;
const Roles = Constants.UserRoles;
exports.Roles = Roles;

const recomputeSymbolLibraries = (symbolLibraries, done) => {
  Object.keys(symbolLibraries).forEach(resourceID => {
    Object.keys(symbolLibraries[resourceID]).forEach(branchID => {
      symbolLibraries[resourceID][branchID].updateSymbolDefinitionsWithDependencies();
    });
  });
  done && done();
};
/**
 * @typedef { import("core/model/comment").Comment } Comment
 * @typedef { import("core/model/comment").CommentCoreObjectID } CommentCoreObjectID
 * @typedef { import("core/model/asset").Asset } Asset
 * @typedef { import("core/model/resource").Resource } Resource
 * @typedef { import("core/core/coreobject").CoreObjectID } CoreObjectID
 */


class Project extends CoreObject.CoreObject {
  /** @type { { [branchID: string]: { [attribute: string]: any; } } } */
  constructor(context, _coreObjectID, externalProjectID) {
    super(context, _coreObjectID);

    _defineProperty(this, "branches", {});

    _defineProperty(this, "canAddControlToGroupTriggerALoop", (control, group) => {
      const controlsResource = this.getResource(group.coreObjectID);

      if (!controlsResource || !control.isSymbol() || controlsResource.getType() !== Constants.ResourceTypes.SYMBOLLIBRARY) {
        return false;
      } // Get controls already depending on the group in which I want to add the symbol


      const controlSymbolDefinition = control.getSymbolDefinition();
      const dependantControls = this.symbolMap.controlsDependencies.getAffectedByResourceChange(group.coreObjectID.resourceID, group.coreObjectID.controlID.toString()).map(cID => this.context.getCoreObject(cID));
      const dependantSymbolControlsInsideSLs = dependantControls.filter(c => c.isSymbol() && c.getMockup().getType() === Constants.ResourceTypes.SYMBOLLIBRARY);
      const dependantSymbolDefinitionCoreObjectIDs = dependantSymbolControlsInsideSLs.map(c => {
        const coreObjectID = c.coreObjectID.clone();
        coreObjectID.controlID = coreObjectID.controlID.slice(0, 1);
        return coreObjectID;
      }); // check if the control's symbol definition is on of the dependency of the group

      return dependantSymbolDefinitionCoreObjectIDs.find(symbolDefCID => symbolDefCID.isEqual(controlSymbolDefinition.coreObjectID)) != null;
    });

    _defineProperty(this, "_setSymbolDependencies", (resource, coreObjectID, src) => {
      this.symbolMap.setControlDependencies(resource, coreObjectID, src);
    });

    _defineProperty(this, "getSymbolsDependingOnSymbol", symbolCoreObjectID => {
      if (!this.symbolMap) {
        return [];
      }

      return this.symbolMap.getSymbolsDependingOnSymbol(symbolCoreObjectID);
    });

    _balsamiq.default.exists(context.host);

    CoreObject.assertProjectType(_coreObjectID);
    this.init(externalProjectID);
  }

  init(externalProjectID) {
    this.projectLocator = null;
    this.license = null;
    this.userRole = Roles.EDITOR;
    this.overriddenUserRole = this.userRole;
    this.defaultStaffRole = null;
    this.attributes = {
      name: "Project",
      creationDate: 0
    };
    this.externalProjectID = externalProjectID;
    this.projectIDForCopy = externalProjectID || CoreObject.createNewUUID();
    this.branches[MASTER] = this._makeBranch(MASTER, {});
    this.mockups = {};
    this.assets = {};
    this.symbolLibraries = {};
    this.resourceUsedBy = {}; // resourceID of Asset -> CoreObjectID of Mockup

    this.thumbnails = {};
    this.thumbnailQueue = new ThumbnailQueue.ThumbnailQueue(this.context, ThumbnailQueue.makeThumbnailQueueCoreObjectID(this.coreObjectID));
    this.trashedThumbnailResourceRendered = false;
    this.thumbnailQueuePauseCounter = 0;
    this.connectedUsers = new ConnectedUsers.ConnectedUsers(this.context, ConnectedUsers.makeConnectedUsersCoreObjectID(this.coreObjectID));
    this.mentionableUsers = {};
    this.archiveState = new ArchiveState.ArchiveState(this.context, ArchiveState.makeArchiveStateCoreObjectID(this.coreObjectID));
    this.dirtyWireframes = new DirtyWireframes.DirtyWireframes(this.context, DirtyWireframes.makeDirtyWireframesCoreObjectID(this.coreObjectID));
    this.comments = {};
    this.messages = [];
    /** @type { { [userID: string]: import("core/model/user").User } } */

    this.users = {};
    this.permalinks = new Permalinks.Permalinks(this.context, Permalinks.makePermalinksCoreObjectID(this.coreObjectID));

    this._getMockupDependency = resourceID => {
      let r = this._findResource(MASTER, resourceID);

      return r && !r.isTrashed() ? r : null;
    };

    this._getMockupIDByName = mockupName => {
      return this._findMockupIDByName(mockupName);
    };

    this._getUserRole = () => {
      return this.getUserRole();
    };

    this.addListeners();
    this.serverTimeDiff = 0;
    this.pendingAssetIDs = [];
    this.hasComingUserEvents = false;
    this.maxResourceDataSize = 20 * 1024 * 1024;
    this._isLoading = false; // true when project.load is working

    this.symbolMap = new _symbolMap.SymbolMap(this.context.getCoreObject);
  }

  addListeners() {
    this.context.eventCenter.addListener({
      projectID: this.coreObjectID.projectID
    }, Resource.EventTypes.RESOURCE_CONTENTS_CHANGED, this, this.onResourceContentsChanged);
    this.context.eventCenter.addListener({
      projectID: this.coreObjectID.projectID
    }, Branch.EventTypes.BRANCH_CONTENTS_CHANGED, this, this.onBranchContentsChanged);
    this.context.eventCenter.addListener({
      projectID: this.coreObjectID.projectID
    }, Thumbnail.EventTypes.THUMBNAIL_CONTENTS_CHANGED, this, this.onThumbnailContentsChanged);
    this.context.eventCenter.addListener({
      projectID: this.coreObjectID.projectID
    }, CoreObject.EventTypes.KEY_VALUE_CHANGED, this, this.onKeyValueChanged);
    this.context.eventCenter.addListener(this.archiveState.coreObjectID, CoreObject.EventTypes.KEY_VALUE_CHANGED, this, this.onArchiveStateChanged);
    this.context.eventCenter.addListener(this.connectedUsers.coreObjectID, CoreObject.EventTypes.KEY_VALUE_CHANGED, this, this.onConnectedUsersKeyValueChanged);
    this.context.eventCenter.addListener(this.thumbnailQueue.coreObjectID, ThumbnailQueue.EventTypes.HAS_COMING_USER_EVENTS, this, this.onThumbnailQueueHasArchiveOpsChanged);
  }

  cleanUp() {
    Object.values(this.thumbnails).forEach(thumbnail => {
      thumbnail.cleanUp();
    });
    this.thumbnails = null;
    this.connectedUsers.cleanUp();
    this.connectedUsers = null;
    this.thumbnailQueue.cleanUp();
    this.thumbnailQueue = null;
    this.archiveState.cleanUp();
    this.archiveState = null;
    this.dirtyWireframes.cleanUp();
    this.dirtyWireframes = null;
    this.permalinks.cleanUp();
    this.permalinks = null;
    this.pendingAssetIDs = [];
    this.symbolMap = null;
    super.cleanUp();
  } // Key-value coding
  // ----------------


  getKeyValue(keyName) {
    if (keyName === KeyNames.NAME) {
      return this.getName();
    } else if (keyName === KeyNames.CREATION_DATE) {
      return this.getCreationDate();
    } else if (keyName === KeyNames.BRANCH_LIST) {
      return this.getBranches();
    } else if (keyName === KeyNames.MOCKUP_LIST) {
      return this.getMockupList();
    } else if (keyName === KeyNames.MOCKUP_LIST_FLAT) {
      return this.getFlatMockupList();
    } else if (keyName === KeyNames.ASSET_LIST) {
      return this.getAssetList();
    } else if (keyName === KeyNames.SYMBOLLIBRARY_LIST) {
      return this.getSymbolLibraryList();
    } else if (keyName === KeyNames.TRASH_LIST) {
      return this.getTrashList();
    } else if (keyName === KeyNames.DESCRIPTION) {
      return this.getDescription();
    } else if (keyName === KeyNames.SELECTION_COLOR) {
      return this.getSelectionColor();
    } else if (keyName === KeyNames.FONT_FACE) {
      return this.getFontFace();
    } else if (keyName === KeyNames.FONT_SIZE) {
      return this.getDefaultFontSize();
    } else if (keyName === KeyNames.SKIN_NAME) {
      return this.getSkinName();
    } else if (keyName === KeyNames.LINK_COLOR) {
      return this.getLinkColor();
    } else if (keyName == KeyNames.USER_ROLE) {
      return this.getActualUserRole();
    } else if (keyName == KeyNames.DEFAULT_STAFF_ROLE) {
      return this.getDefaultStaffRole();
    } else if (keyName == KeyNames.COMMENTS) {
      return _balsamiq.default.objectReduce(this.comments, (comments, comment, commentID) => {
        const resource = this.getResourceID(comment.coreObjectID.resourceID, comment.coreObjectID.branchID);

        if (resource && !resource.isTrashed()) {
          comments[commentID] = comment;
        }

        return comments;
      }, {});
    } else if (keyName == KeyNames.USERS) {
      return this.users;
    } else if (keyName == KeyNames.MENTIONABLE_USERS) {
      return this.mentionableUsers;
    } else if (keyName === KeyNames.MESSAGES) {
      return this.messages;
    } else if (keyName === KeyNames.HAS_COMING_USER_EVENTS) {
      return this.hasComingUserEvents;
    } else if (keyName == KeyNames.IS_LOADING) {
      return this.isLoading();
    } else if (keyName == KeyNames.LICENSE) {
      return this.license;
    } else if (keyName == KeyNames.PROJECTLOCATOR) {
      return this.projectLocator;
    }

    return super.getKeyValue(keyName);
  }

  setKeyValue(keyName, value) {
    switch (keyName) {
      case KeyNames.NAME:
        this.setName(value);
        this.projectLocator && this.projectLocator.setName(value);
        break;

      case KeyNames.DESCRIPTION:
        this.setDescription(value);
        break;

      case KeyNames.LINK_COLOR:
        this.setLinkColor(value);
        break;

      case KeyNames.SELECTION_COLOR:
        this.setSelectionColor(value);
        break;

      case KeyNames.SKIN_NAME:
        this.setSkinName(value);
        break;

      case KeyNames.FONT_SIZE:
        this.setFontSize(value);
        break;

      case KeyNames.FONT_FACE:
        this.setFontFace(value);
        break;

      default:
        super.setKeyValue(keyName, value);
    }
  } // Branches
  // --------


  getBranches() {
    // return [ Branch (Master), Branch, .. ]
    const branches = [];

    if (this.branches[MASTER]) {
      branches.push(this.branches[MASTER]);
    }

    for (const branchID in this.branches) {
      if (branchID !== MASTER) {
        branches.push(this.branches[branchID]);
      }
    }

    return branches;
  }

  getBranch(coreObjectID) {
    _balsamiq.default.existsClass(coreObjectID, CoreObject.CoreObjectID);

    _balsamiq.default.assert(coreObjectID.projectID, this.coreObjectID.projectID);

    _balsamiq.default.exists(coreObjectID.branchID);

    return this.branches[coreObjectID.branchID];
  }

  getBranchOfResource(coreResourceID) {
    CoreObject.assertResourceType(coreResourceID);
    return this.branches[coreResourceID.branchID];
  }
  /**
   *
   * @param {CoreObjectID[]} coreResourceIDs
   * @returns {{ [branchID: string]: string }}
   */


  getBranchesOfResources(coreResourceIDs) {
    _balsamiq.default.assertArray(coreResourceIDs);

    return coreResourceIDs.reduce((branchesByBranchID, coreResourceID) => {
      const branch = this.getBranchOfResource(coreResourceID);
      branchesByBranchID[branch.coreObjectID.branchID] = branch.getName();
      return branchesByBranchID;
    }, {});
  }
  /**
   *
   * @param {string} resourceID
   * @return string[] branches
   */


  getBranchesOfMockup(resourceID) {
    if (this.mockups[resourceID]) {
      return Object.keys(this.mockups[resourceID]);
    } else {
      return [];
    }
  }

  getBranchByName(branchName) {
    for (let key in this.branches) {
      let branch = this.branches[key];

      if (branch.getName() === branchName) {
        return branch;
      }
    }

    return null;
  }

  getMentionableUsers() {
    return Object.values(this.mentionableUsers).sort((a, b) => a.fullName.localeCompare(b.fullName));
  }

  getProjectMember(userID) {
    return this.mentionableUsers[userID];
  }

  getProjectMemberByName(name) {
    for (const key in this.mentionableUsers) {
      const projectMember = this.mentionableUsers[key];

      if (projectMember.identifier === name) {
        return projectMember;
      }
    }

    return null;
  }

  getExternalProjectID() {
    return this.externalProjectID;
  }
  /**
   * @return {string}
   */


  getProjectIDForCopy() {
    return this.projectIDForCopy;
  } // Lists
  // -----


  getFlatMockupList() {
    const mockups = [];
    const mockupList = this.getMockupList();

    for (let outerIndex = 0; outerIndex < mockupList.length; ++outerIndex) {
      for (let index = 0; index < mockupList[outerIndex].length; ++index) {
        mockups.push(mockupList[outerIndex][index]);
      }
    }

    return mockups;
  }

  getMockupList() {
    // return [ [ Mockup (Master), Mockup (Branch) ], [ Mockup ], .. ] sorted by mockup order
    const mockups = [];

    for (const resourceID in this.mockups) {
      const mockup = this.mockups[resourceID][MASTER];

      if (mockup && !mockup.isTrashed()) {
        mockups.push(mockup);
      }
    }

    const sortedMockups = _balsamiq.default.sortBy(mockups, mockup => mockup.getAttributes()[_BmprConstants.MockupAttributes.Order]);

    const res = [];
    sortedMockups.forEach(function (mockup) {
      const arr = []; // get all branches of this mockup

      if (this.mockups[mockup.coreObjectID.resourceID][MASTER]) {
        arr.push(this.mockups[mockup.coreObjectID.resourceID][MASTER]);
      }

      for (const branchID in this.mockups[mockup.coreObjectID.resourceID]) {
        if (branchID !== MASTER) {
          arr.push(this.mockups[mockup.coreObjectID.resourceID][branchID]);
        }
      }

      res.push(arr);
    }, this);
    return res;
  }
  /**
   * @return {Asset[]}
   */


  getAssetList() {
    // return [ Asset, Asset .. ] sorted by name
    var assets = [];

    for (var resourceID in this.assets) {
      if (!this.assets[resourceID][MASTER].isTrashed()) {
        assets.push(this.assets[resourceID][MASTER]);
      }
    } // sort by "name"


    var sortedAssets = _balsamiq.default.sortBy(assets, function (asset) {
      return asset.getName().toLowerCase();
    });

    return sortedAssets;
  }

  getSymbolLibraryList() {
    // return [ SymbolLibrary, SymbolLibrary .. ] sorted by name
    var symbolLibraries = [];

    for (var resourceID in this.symbolLibraries) {
      if (!this.symbolLibraries[resourceID][MASTER].isTrashed()) {
        symbolLibraries.push(this.symbolLibraries[resourceID][MASTER]);
      }
    } // sort by "name"


    var sortedSymbolLibraries = _balsamiq.default.sortBy(symbolLibraries, function (symbolLibrary) {
      return symbolLibrary.getAttributes()["name"];
    });

    return sortedSymbolLibraries;
  }

  getTrashList() {
    // return [ Resource, Resource, .. ]
    var resourceID;
    var trashed = [];

    for (resourceID in this.mockups) {
      if (this.mockups[resourceID][MASTER].isTrashed()) {
        trashed.push(this.mockups[resourceID][MASTER]);
      }
    }

    for (resourceID in this.assets) {
      if (this.assets[resourceID][MASTER].isTrashed()) {
        trashed.push(this.assets[resourceID][MASTER]);
      }
    }

    for (resourceID in this.symbolLibraries) {
      if (this.symbolLibraries[resourceID][MASTER].isTrashed()) {
        trashed.push(this.symbolLibraries[resourceID][MASTER]);
      }
    }

    return trashed;
  }
  /**
   * @param {CoreObjectID} coreObjectID
   * @return {Resource | null}
   */


  getResource(coreObjectID) {
    _balsamiq.default.existsClass(coreObjectID, CoreObject.CoreObjectID);

    _balsamiq.default.assert(coreObjectID.projectID === this.coreObjectID.projectID);

    _balsamiq.default.exists(coreObjectID.branchID);

    _balsamiq.default.exists(coreObjectID.resourceID);

    return this._findResource(coreObjectID.branchID, coreObjectID.resourceID);
  }

  hasResourceID(resourceID) {
    _balsamiq.default.assertString(resourceID);

    return this._findResource(MASTER, resourceID) != null;
  }

  findByShortID(shortID) {
    _balsamiq.default.exists(shortID);

    var matchingCoreObjectIDs = [];

    if (shortID.shortResourceID) {
      var functor = list => {
        this._loopOnResources(list, resource => {
          if (resource.matchesShortID(shortID)) {
            matchingCoreObjectIDs.push(resource.coreObjectID);
          }
        });
      };

      functor(this.mockups);
      functor(this.assets);
      functor(this.symbolLibraries);
    }

    if (shortID.shortCommentID) {
      for (let commentID in this.comments) {
        let comment = this.comments[commentID];

        if (comment.matchesShortID(shortID)) {
          matchingCoreObjectIDs.push(comment.coreObjectID);
        }
      }
    }

    return matchingCoreObjectIDs;
  }
  /**
   *
   * @param {string} resourceID
   * @param {string} branchID
   * @return {Resource | null}
   */


  getResourceID(resourceID, branchID) {
    _balsamiq.default.assertString(resourceID, "resourceID must be a string");

    _balsamiq.default.assertString(branchID, "branchID must be a string");

    return this._findResource(branchID, resourceID);
  }

  getAsset(resourceID) {
    _balsamiq.default.assertString(resourceID);

    var res = this.assets[resourceID];

    _balsamiq.default.exists(res);

    return res[MASTER];
  } // Thumbnails
  // ----------


  getThumbnailOfResource(coreResourceID) {
    for (let thumbnailID in this.thumbnails) {
      let thumbnail = this.thumbnails[thumbnailID];
      let resourceCoreObjectID = thumbnail.getResourceCoreObjectID();

      if (resourceCoreObjectID.resourceID === coreResourceID.resourceID && resourceCoreObjectID.branchID === coreResourceID.branchID) {
        return thumbnail;
      }
    }

    return null;
  } // Getters
  // -------


  getName() {
    return this.attributes.name;
  }

  getCreationDate() {
    return this.attributes.creationDate;
  }

  getDefaultSymbolLibrary() {
    var libID = this.branches[MASTER].getAttribute(Branch.MasterAttributes.SYMBOL_LIBRARY_ID);
    var symbolLib = libID && libID in this.symbolLibraries ? this.symbolLibraries[libID][MASTER] : null;

    if (symbolLib && symbolLib.isTrashed()) {
      symbolLib = null;
    }

    return symbolLib;
  }

  getOrCreateDefaultSymbolLibrary() {
    let symbolLib = this.getDefaultSymbolLibrary();

    if (!symbolLib) {
      let symbolLibID = this.newSymbolLibrary(true);
      this.setDefaultSymbolLibraryID(symbolLibID.resourceID);
      symbolLib = this.context.getCoreObject(symbolLibID);
      symbolLib.setName("Project Symbols");
    }

    return symbolLib;
  }

  getDefaultFontSize() {
    var size = this.branches[MASTER].getAttribute(Branch.MasterAttributes.FONT_SIZE);
    return size ? size : DEFAULT_FONTSIZE;
  }

  getFontFace() {
    var face = this.branches[MASTER].getAttribute(Branch.MasterAttributes.FONT_FACE);
    return face ? face : DEFAULT_FONTFACE;
  }

  getSkinName() {
    var skinName = this.branches[MASTER].getAttribute(Branch.MasterAttributes.SKIN_NAME);
    return skinName ? skinName : DEFAULT_SKINNAME;
  }

  getSelectionColor() {
    var color = this.branches[MASTER].getAttribute(Branch.MasterAttributes.SELECTION_COLOR);
    return color != null ? color : DEFAULT_SELECTION_COLOR;
  }

  getLinkColor() {
    var color = this.branches[MASTER].getAttribute(Branch.MasterAttributes.LINK_COLOR);
    return color != null ? color : DEFAULT_LINK_COLOR;
  }

  getDescription() {
    return this.branches[MASTER].getAttribute(Branch.MasterAttributes.PROJECT_DESCRIPTION);
  }

  getActualUserRole() {
    return this.overriddenUserRole;
  }

  getDefaultStaffRole() {
    return this.defaultStaffRole;
  }

  getUserRole() {
    return this.userRole;
  }

  getConnectedUsers() {
    return this.connectedUsers;
  }

  getArchiveState() {
    return this.archiveState;
  }

  getPermalinks() {
    return this.permalinks;
  }

  getDirtyWireframes() {
    return this.dirtyWireframes;
  }

  isAnonymousUser() {
    return this.connectedUsers.getMyUserInfo() ? !!this.connectedUsers.getMyUserInfo().anonymous : true;
  }

  isUserInProject() {
    return this.connectedUsers.getMyUserInfo() ? !!this.connectedUsers.getMyUserInfo().inProject : false;
  }

  getLicenseData() {
    return this.license;
  }

  getProjectLocator() {
    return this.projectLocator;
  } // Setters
  // -------


  setLicenseData(data) {
    _balsamiq.default.assertClass(data, _licensedata.LicenseData);

    this.license = data;
    this.notifyKeyValueChanged(KeyNames.LICENSE);
  }

  setProjectLocator(projectLocator) {
    _balsamiq.default.assertClass(projectLocator, _locators.ProjectLocator);

    this.projectLocator = projectLocator;
    this.notifyKeyValueChanged(KeyNames.PROJECTLOCATOR);
  }

  setUserRole(role) {
    _balsamiq.default.assertEnum(role, Roles);

    if (this.userRole != role) {
      this.userRole = role; // Always set overridden role to current Role.
      // When scaling up I want the current mode to reflect the new capabilities (eg. I register a license -> I go to edit mode)
      // When scaling down we have to restrict current mode

      this.overriddenUserRole = this.userRole;
      this.notifyKeyValueChanged(KeyNames.USER_ROLE);
    }
  }

  setDefaultStaffRole(role) {
    if (role != null) {
      _balsamiq.default.assertEnum(role, Roles);
    }

    if (this.defaultStaffRole === role) {
      return;
    }

    this.defaultStaffRole = role;
    this.notifyKeyValueChanged(KeyNames.DEFAULT_STAFF_ROLE);
  }

  setOverriddenUserRole(role) {
    _balsamiq.default.assertEnum(role, Roles);

    role = Math.min(role, this.userRole);

    if (this.overriddenUserRole != role) {
      this.overriddenUserRole = role;
      this.notifyKeyValueChanged(KeyNames.USER_ROLE);
    }
  }

  resetOverriddenUserRole() {
    this.setOverriddenUserRole(this.userRole);
  }

  setName(value) {
    _balsamiq.default.assertString(value);

    this.attributes.name = value;
    this.notifyKeyValueChanged(KeyNames.NAME);
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.PROJECT_ATTRIBUTES_DIRTY, this.attributes));
  }

  setDescription(value) {
    _balsamiq.default.assertString(value);

    this.branches[MASTER].setAttribute(Branch.MasterAttributes.PROJECT_DESCRIPTION, value);
  }

  setSkinName(value) {
    _balsamiq.default.assertString(value);

    this.branches[MASTER].setAttribute(Branch.MasterAttributes.SKIN_NAME, value);
  }

  setSelectionColor(value) {
    _balsamiq.default.assertNumber(value);

    this.branches[MASTER].setAttribute(Branch.MasterAttributes.SELECTION_COLOR, value);
    this.notifyKeyValueChanged(KeyNames.SELECTION_COLOR);
  }

  setLinkColor(value) {
    _balsamiq.default.assertNumber(value);

    this.branches[MASTER].setAttribute(Branch.MasterAttributes.LINK_COLOR, value);
    this.notifyKeyValueChanged(KeyNames.LINK_COLOR);
  }

  setFontFace(value) {
    _balsamiq.default.assertString(value);

    this.branches[MASTER].setAttribute(Branch.MasterAttributes.FONT_FACE, value);
  }

  setFontSize(value) {
    _balsamiq.default.assertNumber(value);

    this.branches[MASTER].setAttribute(Branch.MasterAttributes.FONT_SIZE, value);
  }

  setDefaultSymbolLibraryID(value) {
    _balsamiq.default.assertUUID(value);

    _balsamiq.default.assert(value in this.symbolLibraries);

    this.branches[MASTER].setAttribute(Branch.MasterAttributes.SYMBOL_LIBRARY_ID, value);
  }

  _makeMockupsTreeData() {
    const mockupList = this.getMockupList();
    return mockupList.map(mockupArray => {
      const mockup = mockupArray[0];
      return {
        resourceID: mockup.coreObjectID.resourceID,
        parentID: mockup.getParentID(),
        order: mockup.getOrder()
      };
    });
  }

  getFlatMockupListWithLevel() {
    const mockups = this.getMockupList();
    const resources = mockups.map(mockupArray => {
      const mockup = mockupArray[0];
      return {
        resourceID: mockup.coreObjectID.resourceID,
        parentID: mockup.getParentID(),
        order: mockup.getOrder()
      };
    });
    const resourceTree = new _resourcetree.ResourceTree(resources);
    const flatMockupList = [];

    for (const mockup of mockups) {
      for (const resource of Array.from(mockup)) {
        const {
          resourceID
        } = resource.coreObjectID;
        const level = resourceTree.resources.find(r => r.resourceID === resourceID).getLevel();
        flatMockupList.push({
          mockup: resource,
          level
        });
      }
    }

    return flatMockupList;
  }

  _makeMockupsTree() {
    const resources = this._makeMockupsTreeData();

    return new _resourcetree.ResourceTree(resources);
  }

  getMockupsTree() {
    return this._makeMockupsTree();
  }
  /**
   * @returns {import("core/utils/resourcetree").ResourceTreeData}
   */


  getMockupsTreeData() {
    return this._makeMockupsTreeData();
  }

  updateResourcesOrderFromTree(tree) {
    this.updateResourcesAttributesFromTree(tree, "order", "getOrder", "setOrder");
  }

  updateResourcesParentIDFromTree(tree) {
    // to avoid loops we first set all the changed parentIDs to null
    //then in the second loop we update them to the correct value
    const mockupsToChange = [];
    this.getMockupList().forEach(mockupArr => {
      // Search resource in tree
      const resID = mockupArr[0].coreObjectID.resourceID;
      const res = tree.getResourceByID(resID);
      const newVal = res ? res.parentID : null;
      const oldVal = mockupArr[0].getParentID();

      if (res && oldVal !== newVal) {
        mockupsToChange.push(mockupArr);
        mockupArr[0].setParentID(null);
      }
    });
    mockupsToChange.forEach(mockupArr => {
      // Search resource in tree
      const resID = mockupArr[0].coreObjectID.resourceID;
      const res = tree.getResourceByID(resID);
      const newVal = res ? res.parentID : null;

      if (newVal !== null) {
        mockupArr[0].setParentID(newVal);
      }
    });
  } // watch out: this changes the master only since it is working on order and parentID


  updateResourcesAttributesFromTree(tree, attrName, attrGetter, attrSetter) {
    this.getMockupList().forEach(mockupArr => {
      // Search resource in tree
      var resID = mockupArr[0].coreObjectID.resourceID;
      var res = tree.getResourceByID(resID);
      let newVal = res ? res[attrName] : null;
      let oldVal = mockupArr[0][attrGetter]();

      if (res && oldVal !== newVal) {
        mockupArr[0][attrSetter](newVal);
      }
    });
  }

  changeLevelToMockupIndexes(resourceIndexesArray, addLevel) {
    _balsamiq.default.assertArray(resourceIndexesArray);

    _balsamiq.default.assertBool(addLevel);

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.changeLevelToMockupIndexes can not edit wireframes because the user role is: " + this.userRole); // indexes are based on getMockupList result


    var tree = this._makeMockupsTree();

    if (addLevel) {
      tree.addLevelToIndexes(resourceIndexesArray);
    } else {
      tree.removeLevelFromIndexes(resourceIndexesArray);
    }

    this.updateResourcesParentIDFromTree(tree);
  }

  reorderMockups(resourceIndexesArray, targetPos) {
    let {
      atSameLevelOfIndex,
      asChildOfIndex
    } = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    _balsamiq.default.assertArray(resourceIndexesArray);

    _balsamiq.default.assertNumber(targetPos);

    if (atSameLevelOfIndex != null) _balsamiq.default.assertNumber(atSameLevelOfIndex, "atSameLevelOfIndex must be the index of the resource at the level of which you want to reorder the targetIndex");
    if (asChildOfIndex != null) _balsamiq.default.assertNumber(asChildOfIndex, "asChildOfIndex must be the index of the parent resource at the level of which you want to reorder the targetIndex");

    _balsamiq.default.assert(!(atSameLevelOfIndex != null && asChildOfIndex != null), "atSameLevelOfIndex and asChildOfIndex are mutually exclusive");

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.reorderMockups can not edit wireframes because the user role is: " + this.userRole); // indexes are based on getMockupList result


    var tree = this._makeMockupsTree();

    tree.reorderResources(resourceIndexesArray, targetPos, {
      atSameLevelOfIndex,
      asChildOfIndex
    });
    this.updateResourcesParentIDFromTree(tree);
    this.updateResourcesOrderFromTree(tree);
  }

  renameResource(resourceID, name) {
    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.assertString(name);

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.renameResource can not edit Resources because the user role is: " + this.userRole);

    var list = this._findResourceList(resourceID);

    _balsamiq.default.exists(list);

    _balsamiq.default.exists(list.list[resourceID]);

    list.list[resourceID][MASTER].setName(name);
  }
  /**
   *
   * @param {string} data
   * @param {{ [attribute: string]: any }} attributes
   * @param {string=} resourceID
   * @param {string} [branchID="Master"]
   * @param {object} [options={}]
   * @param {boolean} [options.notifyListChanges=true]
   */


  addResource(data, attributes, resourceID) {
    let branchID = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : MASTER;
    let {
      notifyListChanges = true
    } = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : {};

    _balsamiq.default.assertString(data);

    if (resourceID) {
      _balsamiq.default.assertString(resourceID);
    }

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.addResource can not edit Resources because the user role is: " + this.userRole);

    let clonedAttributes = _balsamiq.default.clone(attributes);

    this.addMasterAttributes(resourceID, branchID, clonedAttributes);

    const resource = this._makeResource(branchID, resourceID ? resourceID : CoreObject.createNewUUID(), clonedAttributes, data);

    const keyNamesToNotify = this._addResource(resource);

    const newThumbnailAttributes = {
      branchID: resource.coreObjectID.branchID,
      resourceID: resource.coreObjectID.resourceID
    };
    let thumbNail = null;

    switch (resource.type) {
      case Constants.ResourceTypes.MOCKUP:
        thumbNail = this._makeThumbnail(CoreObject.createNewUUID(), newThumbnailAttributes);
        break;

      case Constants.ResourceTypes.ASSET:
        if (resource instanceof Image.Image) {
          resource.makeNativeImage(Constants.OperationOrigin.USER);
          this.addPendingAssetID(resource.coreObjectID.resourceID);
        }

        break;

      case Constants.ResourceTypes.SYMBOLLIBRARY:
        thumbNail = this._makeThumbnail(CoreObject.createNewUUID(), newThumbnailAttributes);
        break;
    }

    if (notifyListChanges) {
      keyNamesToNotify.forEach(keyName => {
        this.notifyKeyValueChanged(keyName);
      });
    }

    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.RESOURCE_ADDED, {
      coreObjectID: resource.coreObjectID,
      data: data
    }));

    if (thumbNail) {
      this._addThumbnail(thumbNail);

      this.updateThumbnail(resource, Constants.OperationOrigin.USER);
      this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.THUMBNAIL_ADDED, {
        thumbnailID: thumbNail.coreObjectID.thumbnailID,
        attributes: newThumbnailAttributes
      }));
    }

    return resource.coreObjectID;
  }
  /**
   *
   * @param {string} resourceID
   * @param {string | undefined} data
   * @param {string} name
   * @param {string} importedFrom
   * @param {() => void} done
   */


  updateAssetData(resourceID, data, name, importedFrom, done) {
    var asset = this.getAsset(resourceID);
    asset.setData(data, Constants.OperationOrigin.USER);
    asset.setAttribute(_BmprConstants.ResourceAttributes.ImportedFrom, importedFrom, Constants.OperationOrigin.USER);
    asset.setAttribute(_BmprConstants.ResourceAttributes.Name, name, Constants.OperationOrigin.USER);

    if (asset instanceof Image.Image) {
      asset.makeNativeImage(Constants.OperationOrigin.USER, done);
      this.addPendingAssetID(resourceID);
    } else {
      done();
    }
  }

  deleteAlternate(resourceID, branchID) {
    const resource = this.mockups[resourceID][branchID];

    _balsamiq.default.exists(resource);

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.deleteAlternate can not edit Resources because the user role is: " + this.userRole);

    this.deletePermalink({
      branchID,
      resourceID
    });
    this.deleteResource(branchID, resourceID);
    this.deleteEmptyBranches();
    const thumbnail = this.getThumbnailOfResource(resource.coreObjectID);

    if (thumbnail) {
      this.deleteThumbnails([thumbnail.coreObjectID.thumbnailID]);
    }

    this.notifyKeyValueChanged(KeyNames.MOCKUP_LIST);
    return resource;
  }

  moveResourcesToTrash(resourceIDs) {
    _balsamiq.default.assertArray(resourceIDs);

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.moveResourcesToTrash can not edit Resources because the user role is: " + this.userRole);

    var mockupResourceIDs = [];
    var allResourceIDs = [];
    resourceIDs.forEach(resourceID => {
      var res = this._findResource(MASTER, resourceID);

      if (res && !res.isTrashed()) {
        if (res.getType() == Constants.ResourceTypes.MOCKUP) {
          mockupResourceIDs.push(resourceID);
        }

        allResourceIDs.push(resourceID);
      }
    });
    mockupResourceIDs.forEach(resourceID => {
      let branchIDs = this.resourceBranches(resourceID);
      branchIDs.forEach(branchID => this.deletePermalink({
        branchID,
        resourceID
      }));
    });

    var tree = this._makeMockupsTree();

    tree.deleteResourceIDs(mockupResourceIDs);
    this.updateResourcesParentIDFromTree(tree);
    allResourceIDs.forEach(resourceID => {
      this._setResourceTrashed(resourceID, true);
    });
    return {
      count: allResourceIDs.length
    };
  }

  restoreResourcesFromTrash(resourceIDs) {
    _balsamiq.default.assertArray(resourceIDs);

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.restoreResourcesFromTrash can not edit Resources because the user role is: " + this.userRole);

    var mockupResourceIDs = [];
    var parentResourceIDs = [];
    var orders = [];
    var allResourceIDs = [];
    resourceIDs.forEach(resourceID => {
      var res = this._findResource(MASTER, resourceID);

      if (res && res.isTrashed()) {
        if (res.getType() == Constants.ResourceTypes.MOCKUP) {
          mockupResourceIDs.push(resourceID);
          parentResourceIDs.push(res.getParentID());
          orders.push(res.getOrder());
        }

        allResourceIDs.push(resourceID);
      }
    });

    var tree = this._makeMockupsTree();

    tree.addResourcesWithParentID(mockupResourceIDs, parentResourceIDs, orders);
    this.updateResourcesParentIDFromTree(tree);
    allResourceIDs.forEach(resourceID => {
      this._setResourceTrashed(resourceID, false);
    }); // fix the potential mismatch between the order in the project and the one in the resource tree

    const newTree = this._makeMockupsTree();

    mockupResourceIDs.forEach(resourceID => {
      const mockupInProject = this._findResource(MASTER, resourceID);

      const mockupInTree = newTree.getResourceByID(resourceID);

      if (mockupInTree.parentID) {
        const parent = newTree.getResourceByID(mockupInTree.parentID);

        if (parent.order >= mockupInTree.order) {
          mockupInProject.setAttribute("parentID", null, Constants.OperationOrigin.USER);
        }
      }
    });
    return {
      count: allResourceIDs.length
    };
  }

  emptyTrash(resourceIDsToDelete) {
    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.emptyTrash can not delete Resources because the user role is: " + this.userRole);

    if (resourceIDsToDelete) {
      _balsamiq.default.assertArray(resourceIDsToDelete);
    } // must split the resources in two:
    // delete alternates first and master resources then
    // this because the alternates refer to their master counterpart when asking master only attributes


    var alternatesResourcesToDelete = {}; // branchID -> [ resourceIDs ]

    var masterResourcesToDelete = []; // [ resourceIDs ]

    let thumbnailsToDelete = {}; // resourceID -> true

    var addResourceToDelete = function (branchID, resourceID) {
      if (branchID === MASTER) {
        masterResourcesToDelete.push(resourceID);
      } else {
        if (!alternatesResourcesToDelete[branchID]) {
          alternatesResourcesToDelete[branchID] = [];
        }

        alternatesResourcesToDelete[branchID].push(resourceID);
      }
    }; // Clean-up thumbnails that should already have been removed from the bmpr
    // (temporary solution for bmprs created before fixing a thumbnails leak)


    for (const thumbnail of Object.values(this.thumbnails)) {
      if (!this.getResource(thumbnail.resourceCoreObjectID)) {
        thumbnailsToDelete[thumbnail.coreObjectID.thumbnailID] = true;
      }
    }

    const leakedThumbnails = Object.keys(thumbnailsToDelete).length;

    if (leakedThumbnails > 0) {
      this.context.host.log("Project: bmpr had leaked thumbnails which have been removed.");
    }

    var commentsToDelete = [];
    var trashedResources = this.getTrashList(); // only MASTER resources are in this list

    trashedResources.forEach(resource => {
      _balsamiq.default.assert(resource.coreObjectID.branchID == MASTER, "trashedResources are expected to be only of MASTER branch");

      let resourceID = resource.coreObjectID.resourceID; // delete this resource if:
      // - resourceIDsToDelete param is not specified, OR
      // - resourceIDsToDelete is specified and contains resourceID

      let deleteThis = resourceIDsToDelete ? resourceIDsToDelete.indexOf(resourceID) >= 0 : true;

      if (deleteThis) {
        let list = this._findResourceList(resourceID);

        for (var branchID in list.list[resourceID]) {
          addResourceToDelete(branchID, resourceID);
          let res = this.getResourceID(resourceID, branchID);
          let thumbnail = this.getThumbnailOfResource(res.coreObjectID);

          if (thumbnail) {
            thumbnailsToDelete[thumbnail.coreObjectID.thumbnailID] = true;
          }

          const comments = this.getCommentsByResource(resourceID, branchID);

          if (comments.length > 0) {
            for (const comment of comments) {
              commentsToDelete.push(comment.getID());
            }
          }
        }
      }
    }); // delete related comments first

    if (commentsToDelete.length > 0) {
      for (var commentID of commentsToDelete) {
        delete this.comments[commentID];
      }

      this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.COMMENTS_DELETED, {
        commentIDs: commentsToDelete
      }));
    }

    var changedKeyNames = {};

    for (let branchID in alternatesResourcesToDelete) {
      alternatesResourcesToDelete[branchID].forEach(resourceID => {
        var keyNames = this._deleteResource(branchID, resourceID);

        keyNames.forEach(keyName => {
          changedKeyNames[keyName] = keyName;
        });
      });
    }

    for (var resourceID of masterResourcesToDelete) {
      var keyNames = this._deleteResource(MASTER, resourceID);

      keyNames.forEach(keyName => {
        changedKeyNames[keyName] = keyName;
      });
    }

    for (let branchID in alternatesResourcesToDelete) {
      this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.RESOURCES_DELETED, {
        branchID: branchID,
        resourceIDs: alternatesResourcesToDelete[branchID]
      }));
    }

    if (masterResourcesToDelete.length > 0) {
      this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.RESOURCES_DELETED, {
        branchID: MASTER,
        resourceIDs: masterResourcesToDelete
      }));
    }

    let thumbnails = Object.keys(thumbnailsToDelete);

    if (thumbnails.length > 0) {
      this.deleteThumbnails(thumbnails);
    }

    for (var keyName in changedKeyNames) {
      this.notifyKeyValueChanged(keyName);
    }

    this.deleteEmptyBranches();
  }

  deleteEmptyBranches() {
    var emptyBranchIDs = [];

    for (var branchID in this.branches) {
      if (branchID != MASTER) {
        var isEmpty = true;

        for (var resourceID in this.mockups) {
          if (this.mockups[resourceID][branchID]) {
            isEmpty = false;
          }
        }

        if (isEmpty) {
          emptyBranchIDs.push(branchID);
        }
      }
    }

    if (emptyBranchIDs.length > 0) {
      emptyBranchIDs.forEach(branchID => {
        this._deleteBranch(branchID);
      });
      this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.BRANCHES_DELETED, {
        branchIDs: emptyBranchIDs
      }));
    }
  }

  appendCopyAndNumber(p_name, p_list) {
    let pieces = p_name.split(" ");

    if (pieces[pieces.length - 1] === "copy" || pieces[pieces.length - 2] === "copy") {
      p_name = p_name.substr(0, p_name.indexOf(" copy"));
    }

    let name = p_name + " copy";
    let num = 1;

    while (this.existsName(name, p_list)) {
      num++;
      name = p_name + " copy " + num;
    }

    return name;
  }

  existsName(name, resources) {
    return resources.some(r => r.getName().toLowerCase() === name.toLowerCase());
  }
  /**
   *
   * @param {string} kind
   * @param {string} originalName
   * @return {string}
   */


  makeNewResourceName(kind, originalName) {
    _balsamiq.default.assertEnum(kind, Constants.ResourceTypes);

    _balsamiq.default.assertString(originalName);

    let resources;

    switch (kind) {
      case Constants.ResourceTypes.MOCKUP:
        resources = this.getFlatMockupList();
        break;

      case Constants.ResourceTypes.SYMBOLLIBRARY:
        resources = this.getSymbolLibraryList();
        break;

      case Constants.ResourceTypes.ASSET:
        resources = this.getAssetList();
        break;

      default:
        _balsamiq.default.asset(false, `${kind} is not a recognized resource type`);

        break;
    }

    return this.existsName(originalName, resources) ? (0, _textutilities.makeUniqueName)(resources.map(r => r.getName()), originalName) : originalName;
  }

  newSymbolLibrary(empty) {
    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.newSymbolLibrary can not edit Resources because the user role is: " + this.userRole);

    var json = empty ? Constants.EmptyMockup : Constants.EmptySymbolLibrary;
    var attributes = {}; // pick a name:

    var nextNewItemID = 0;
    var name = "";
    var libWithSameName;

    do {
      nextNewItemID++;
      name = Constants.UnnamedSymbolLibrary + " " + nextNewItemID;

      for (var resourceID in this.symbolLibraries) {
        libWithSameName = Object.values(this.symbolLibraries[resourceID]).find(symbolLibrary => {
          return name === symbolLibrary.getName();
        });

        if (libWithSameName != undefined) {
          break;
        }
      }
    } while (libWithSameName != undefined);

    attributes[_BmprConstants.ResourceAttributes.Kind] = _BmprConstants.ResourceKinds.SymbolLibrary;
    attributes[_BmprConstants.ResourceAttributes.Name] = name;
    attributes[_BmprConstants.ResourceAttributes.CreationDate] = 0;
    attributes[_BmprConstants.ResourceAttributes.ModifiedBy] = null;
    attributes[_BmprConstants.ResourceAttributes.Trashed] = false;
    attributes[_BmprConstants.ResourceAttributes.Notes] = Constants.EmptySymbolLibraryNotes;
    attributes[_BmprConstants.ResourceAttributes.ImportedFrom] = "";
    var jsonString = JSON.stringify(json);

    var resource = this._makeResource(MASTER, CoreObject.createNewUUID(), attributes, jsonString);

    this._addResource(resource);

    this.notifyKeyValueChanged(KeyNames.SYMBOLLIBRARY_LIST);
    var newThumbnailAttributes = {
      branchID: resource.coreObjectID.branchID,
      resourceID: resource.coreObjectID.resourceID
    };

    var thumbNail = this._makeThumbnail(CoreObject.createNewUUID(), newThumbnailAttributes);

    this._addThumbnail(thumbNail);

    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.RESOURCE_ADDED, {
      coreObjectID: resource.coreObjectID,
      data: jsonString
    }));
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.THUMBNAIL_ADDED, {
      thumbnailID: thumbNail.coreObjectID.thumbnailID,
      attributes: newThumbnailAttributes
    }));
    this.updateThumbnail(resource, Constants.OperationOrigin.USER);
    return resource.coreObjectID;
  }

  cloneSymbolLibrary(resourceID) {
    _balsamiq.default.assert(this.symbolLibraries[resourceID]);

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.cloneSymbolLibrary can not edit Resources because the user role is: " + this.userRole);

    var uuidOfClonedSymbolLibrary = CoreObject.createNewUUID(); // there are no "alternate" symbol libraries, just Master

    let symbolLibrary = this.symbolLibraries[resourceID][MASTER];
    let jsonString = symbolLibrary.getData();

    let clonedAttributes = _balsamiq.default.clone(symbolLibrary.getAttributes());

    var flatListOfSymbolLibraries = Object.values(this.symbolLibraries).reduce((previous, element) => {
      previous.push(element[MASTER]);
      return previous;
    }, []);
    let name = this.appendCopyAndNumber(clonedAttributes[_BmprConstants.ResourceAttributes.Name], flatListOfSymbolLibraries);
    clonedAttributes[_BmprConstants.ResourceAttributes.Name] = name;

    let resource = this._makeResource(symbolLibrary.coreObjectID.branchID, uuidOfClonedSymbolLibrary, clonedAttributes, jsonString);

    this._addResource(resource);

    var newThumbnailAttributes = {
      branchID: resource.coreObjectID.branchID,
      resourceID: resource.coreObjectID.resourceID
    };

    let thumbNail = this._makeThumbnail(CoreObject.createNewUUID(), newThumbnailAttributes);

    this._addThumbnail(thumbNail);

    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.RESOURCE_ADDED, {
      coreObjectID: resource.coreObjectID,
      data: jsonString
    }));
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.THUMBNAIL_ADDED, {
      thumbnailID: thumbNail.coreObjectID.thumbnailID,
      attributes: newThumbnailAttributes
    }));
    this.updateThumbnail(resource, Constants.OperationOrigin.USER);
    this.notifyKeyValueChanged(KeyNames.SYMBOLLIBRARY_LIST);
    return uuidOfClonedSymbolLibrary;
  }

  renameBranch(branchID, newName) {
    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.renameBranch can not edit Resources because the user role is: " + this.userRole); // this will fire the event of branch attrib changed


    this.branches[branchID].setAttribute(Branch.AlternateAttributes.NAME, newName);
    this.notifyKeyValueChanged(KeyNames.MOCKUP_LIST);
  }

  canMoveAlternateToAnotherBranch(resourceID, branchIDFrom, branchIDTo) {
    if (!this.mockups[resourceID][branchIDFrom]) {
      return false;
    }

    if (this.mockups[resourceID][branchIDTo]) {
      return false;
    }

    return true;
  }

  moveAlternateToAnotherBranch(resourceID, branchIDFrom, branchIDTo) {
    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.moveAlternateToAnotherBranch can not edit Resources because the user role is: " + this.userRole);

    _balsamiq.default.assert(this.canMoveAlternateToAnotherBranch(resourceID, branchIDFrom, branchIDTo)); // Get old data


    let mockupToChange = this.mockups[resourceID][branchIDFrom];
    let thumbNail = this.getThumbnailOfResource(mockupToChange.coreObjectID); // update mockup branch

    mockupToChange.coreObjectID.branchID = branchIDTo;
    delete this.mockups[resourceID][branchIDFrom];
    this.mockups[resourceID][branchIDTo] = mockupToChange;
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.RESOURCE_BRANCHID_CHANGED, {
      resourceID: resourceID,
      oldBranchID: branchIDFrom,
      newBranchID: branchIDTo
    })); // update thumbnail resource to reuse it

    thumbNail.setResourceCoreObjectID(mockupToChange.coreObjectID); // Notify the removal of the thumbnail

    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.THUMBNAILS_DELETED, {
      thumbnailIDs: [thumbNail.coreObjectID.thumbnailID]
    })); // Notify the add of the same thumbnail with resource branch changed

    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.THUMBNAIL_ADDED, {
      thumbnailID: thumbNail.coreObjectID.thumbnailID,
      attributes: mockupToChange.coreObjectID
    })); // Before deleting old branch the mockup must have already changed its branch, otherwise we have a DB FOREIGN KEY violation

    this.deleteEmptyBranches();
    this.notifyKeyValueChanged(KeyNames.MOCKUP_LIST);
  }

  moveAlternateToAnotherBranchFromArchive(resourceID, branchIDFrom, branchIDTo) {
    _balsamiq.default.isString(resourceID);

    _balsamiq.default.isString(branchIDFrom);

    _balsamiq.default.isString(branchIDTo); // Get old data


    let mockupToChange = this.mockups[resourceID][branchIDFrom];
    let thumbNail = this.getThumbnailOfResource(mockupToChange.coreObjectID); // update mockup branch

    mockupToChange.coreObjectID.branchID = branchIDTo;
    delete this.mockups[resourceID][branchIDFrom];
    this.mockups[resourceID][branchIDTo] = mockupToChange; // update thumbnail resource to reuse it

    thumbNail.setResourceCoreObjectID(mockupToChange.coreObjectID);
    this.notifyKeyValueChanged(KeyNames.MOCKUP_LIST);
  }
  /**
   *
   * @param {string} branchID
   * @param {{ [key: string]: any }} branchAttributes
   */


  makeBranch(branchID, branchAttributes) {
    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.makeBranch can not edit Resources because the user role is: " + this.userRole);

    this._addBranch(this._makeBranch(branchID, branchAttributes));

    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.BRANCH_ADDED, {
      coreObjectID: this.branches[branchID].coreObjectID
    }));
    this.notifyKeyValueChanged(KeyNames.BRANCH_LIST);
  }
  /**
   *
   * @param {string} referenceMockupResourceID
   * @param {{ [attribute: string]: any }} attributes
   * @return {CoreObjectID}
   */


  newMockupLastChildOf(referenceMockupResourceID, attributes) {
    _balsamiq.default.assertString(referenceMockupResourceID);

    _balsamiq.default.assertObject(attributes);

    return this.newMockup(() => {
      const list = this.getMockupList();
      const index = list.findIndex(function (sublist) {
        return sublist[0].coreObjectID.resourceID == referenceMockupResourceID;
      });
      this.reorderMockups([list.length - 1], index + 1, {
        asChildOfIndex: index
      });
    }, attributes);
  } // creates a new mockup as a first child of the reference if this has children, at same level otherwise


  newMockupChildOfIfParentAlready(referenceMockupResourceID) {
    _balsamiq.default.assertString(referenceMockupResourceID);

    let mtree = this.getMockupsTree();
    let resourceTreeItem = mtree.getResourceByID(referenceMockupResourceID);

    if (!resourceTreeItem.children.length) {
      // ok callee may not know that this mockup has no children, so: create after
      return this.appendNewMockupAfter(referenceMockupResourceID);
    }

    let resourceTreeItemChildIndex = mtree.getIndex(resourceTreeItem.children[0]);
    return this.newMockup(() => {
      var list = this.getMockupList();
      var index = list.findIndex(function (sublist) {
        return sublist[0].coreObjectID.resourceID == referenceMockupResourceID;
      });
      this.reorderMockups([list.length - 1], index + 1, {
        atSameLevelOfIndex: resourceTreeItemChildIndex
      });
    });
  } // creates a new mockup and, if the reference is a mockup, append it at the same level of the reference and as last sibling


  appendNewMockupAtEnd(referenceMockupResourceID) {
    _balsamiq.default.assertString(referenceMockupResourceID);

    let referenceResource = this._findResource(MASTER, referenceMockupResourceID);

    let reorderCallback = null;

    if (referenceResource && referenceResource.getType() == Constants.ResourceTypes.MOCKUP && !referenceResource.isTrashed()) {
      // the reference is a mockup, then the new mockup must be reordered
      let lastMockupResourceID;
      let mtree = this.getMockupsTree();
      let resourceTreeItem = mtree.getResourceByID(referenceMockupResourceID);
      let resourceTreeItemIndex = mtree.getIndex(resourceTreeItem);

      if (!resourceTreeItem.parentID) {
        let allmockups = this.getMockupList();
        lastMockupResourceID = allmockups[allmockups.length - 1][0].coreObjectID.resourceID;
      } else {
        let parentTreeItem = mtree.getResourceByID(resourceTreeItem.parentID);
        let index = mtree.getIndex(parentTreeItem); // new mockup should be dropped AFTER the last descendant index

        let descendantIndexes = mtree.getResourceDescendantsByIndex(index, false);

        if (descendantIndexes.length === 0) {
          lastMockupResourceID = referenceMockupResourceID;
        } else {
          lastMockupResourceID = mtree.getResource(descendantIndexes[descendantIndexes.length - 1]).resourceID;
        }
      }

      reorderCallback = () => {
        var list = this.getMockupList();
        var index = list.findIndex(function (sublist) {
          return sublist[0].coreObjectID.resourceID == lastMockupResourceID;
        }); // explanation: if top level is required, just instruct the resource
        // tree to use the parent of the next resource, we know it is at top
        // level because we are not dropping in the middle of the tree

        this.reorderMockups([list.length - 1], index + 1, {
          atSameLevelOfIndex: resourceTreeItemIndex
        });
      };
    }

    return this.newMockup(reorderCallback);
  } // creates a new mockup and append at the same level of the reference and next to it


  appendNewMockupAfter(referenceMockupResourceID, attributes) {
    let lastMockupResourceID;
    let mtree = this.getMockupsTree();
    let resourceTreeItem = mtree.getResourceByID(referenceMockupResourceID);
    let resourceTreeItemIndex = mtree.getIndex(resourceTreeItem); // new mockup should be dropped AFTER the last descendant index

    let descendantIndexes = mtree.getResourceDescendantsByIndex(resourceTreeItemIndex, false);

    if (descendantIndexes.length === 0) {
      lastMockupResourceID = referenceMockupResourceID;
    } else {
      lastMockupResourceID = mtree.getResource(descendantIndexes[descendantIndexes.length - 1]).resourceID;
    }

    return this.newMockup(() => {
      var list = this.getMockupList();
      var index = list.findIndex(function (sublist) {
        return sublist[0].coreObjectID.resourceID == lastMockupResourceID;
      }); // explanation: if top level is required, just instruct the resource
      // tree to use the parent of the next resource, we know it is at top
      // level because we are not dropping in the middle of the tree

      this.reorderMockups([list.length - 1], index + 1, {
        atSameLevelOfIndex: resourceTreeItemIndex
      });
    }, attributes);
  }
  /**
   * creates a new mockup and invoke a placement/reorder callback if present
   * @param {Function | null} reorderCallback
   * @param {{ name?: string | null; parentID?: string | null }} options
   * @return {CoreObjectID}
   */


  newMockup(reorderCallback) {
    let {
      name: originalName,
      parentID
    } = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    if (reorderCallback != null) _balsamiq.default.assertFunction(reorderCallback);
    if (originalName != null) _balsamiq.default.assertString(originalName);
    if (parentID != null) _balsamiq.default.assertString(parentID);

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.newMockup can not edit Resources because the user role is: " + this.userRole);

    var json = Constants.EmptyMockup;
    var sortedMockups = this.getFlatMockupList(); // pick a name:

    let name = "";

    if (originalName) {
      name = this.makeNewMockupName(originalName);
    } else {
      var nextNewItemID = 0;

      do {
        nextNewItemID++;
        name = Constants.UnnamedMockup + " " + nextNewItemID;
      } while (sortedMockups.find(mockup => {
        return name === mockup.getName();
      }) != undefined);
    }

    var attributes = this.getAttributesForNewMockup(name, parentID);
    var jsonString = JSON.stringify(json);

    var resource = this._makeResource(MASTER, CoreObject.createNewUUID(), attributes, jsonString);

    this._addResource(resource);

    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.RESOURCE_ADDED, {
      coreObjectID: resource.coreObjectID,
      data: jsonString
    })); // fix mockup order

    if (reorderCallback) {
      reorderCallback();
    }

    this.notifyKeyValueChanged(KeyNames.MOCKUP_LIST);
    var newThumbnailAttributes = {
      branchID: resource.coreObjectID.branchID,
      resourceID: resource.coreObjectID.resourceID
    };

    var thumbNail = this._makeThumbnail(CoreObject.createNewUUID(), newThumbnailAttributes);

    this._addThumbnail(thumbNail);

    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.THUMBNAIL_ADDED, {
      thumbnailID: thumbNail.coreObjectID.thumbnailID,
      attributes: newThumbnailAttributes
    }));
    this.updateThumbnail(resource, Constants.OperationOrigin.USER);
    return resource.coreObjectID;
  }

  getAttributesForNewMockup(name, parentID) {
    _balsamiq.default.assertString(name, `name must be a string or null, but it's a ${typeof name}`);

    if (parentID != null) _balsamiq.default.assertString(parentID, `parentID must be a string or null, but it's a ${typeof parentID}`);
    var attributes = {};
    attributes[_BmprConstants.MockupAttributes.Order] = this.getNewMockupOrder();
    attributes[_BmprConstants.ResourceAttributes.Kind] = _BmprConstants.ResourceKinds.Mockup;
    attributes[_BmprConstants.ResourceAttributes.Name] = name;
    attributes[_BmprConstants.ResourceAttributes.CreationDate] = 0;
    attributes[_BmprConstants.ResourceAttributes.ModifiedBy] = null;
    attributes[_BmprConstants.ResourceAttributes.Trashed] = false;
    attributes[_BmprConstants.ResourceAttributes.Notes] = null;
    attributes[_BmprConstants.ResourceAttributes.ImportedFrom] = "";
    attributes[_BmprConstants.MockupAttributes.ParentID] = parentID;
    return attributes;
  }
  /**
   * @return {number}
   */


  getLastMockupOrder() {
    var list = this.getMockupList();
    var len = list.length;
    return len > 0 ? list[len - 1][0].getOrder() : 0;
  }
  /**
   * @param {number?} startFrom
   * @return {number}
   */


  getNewMockupOrder() {
    let startFrom = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 0;
    var list = this.getMockupList();
    var len = list.length;
    return (len == 0 ? Constants.FirstMockupOrder : startFrom + this.getLastMockupOrder() + Constants.FirstMockupOrder) + Math.random() * Constants.FirstMockupOrder;
  }

  getAttributesForNewSymbolLibrary(name) {
    var attributes = {};
    attributes[_BmprConstants.ResourceAttributes.Kind] = _BmprConstants.ResourceKinds.SymbolLibrary;
    attributes[_BmprConstants.ResourceAttributes.Name] = name;
    attributes[_BmprConstants.ResourceAttributes.CreationDate] = 0;
    attributes[_BmprConstants.ResourceAttributes.ModifiedBy] = null;
    attributes[_BmprConstants.ResourceAttributes.Trashed] = false;
    attributes[_BmprConstants.ResourceAttributes.Notes] = Constants.EmptySymbolLibraryNotes;
    attributes[_BmprConstants.ResourceAttributes.ImportedFrom] = "";
    return attributes;
  }

  getAttributesForNewAsset(name, extension) {
    var attributes = {};
    attributes[_BmprConstants.ResourceAttributes.Kind] = _BmprConstants.ResourceKinds.Asset;
    attributes[_BmprConstants.ResourceAttributes.Name] = name;
    attributes[_BmprConstants.ResourceAttributes.CreationDate] = 0;
    attributes[_BmprConstants.ResourceAttributes.ModifiedBy] = null;
    attributes[_BmprConstants.ResourceAttributes.Trashed] = false;
    attributes[_BmprConstants.ResourceAttributes.Notes] = null;
    attributes[_BmprConstants.ResourceAttributes.ImportedFrom] = "";
    attributes[_BmprConstants.AssetAttributes.Extension] = extension;
    return attributes;
  }

  onCommentDeleted(commentIDs) {
    for (let commentID of commentIDs) {
      delete this.comments[commentID];
    }

    this.notifyKeyValueChanged(KeyNames.COMMENTS);
  }

  onCommentAdded(data) {
    _balsamiq.default.assert(!this.comments[data.commentID]);

    let comment = this._makeComment(data.commentID, data.branchID, data.resourceID, data.userID, data.attributes, data.data);

    this.comments[comment.getID()] = comment;
    this.notifyKeyValueChanged(KeyNames.COMMENTS);
  }

  updateComment(commentID, attributes, data) {
    var existingComment = this.comments[commentID];

    if (existingComment) {
      var changed = existingComment.updateAttributes(attributes);
      changed = existingComment.setData(data) || changed;

      if (changed) {
        this.notifyKeyValueChanged(KeyNames.COMMENTS);
      }
    }
  }

  updateCommentsAttributes(updatedAttributes) {
    _balsamiq.default.assertArray(updatedAttributes);

    let notifyNeeded = false;
    updatedAttributes.forEach(updatedAttributeData => {
      var existingComment = this.comments[updatedAttributeData.commentID];

      if (existingComment) {
        notifyNeeded = notifyNeeded || existingComment.updateAttributes(updatedAttributeData.attributes);
      }
    });

    if (notifyNeeded) {
      this.notifyKeyValueChanged(KeyNames.COMMENTS);
    }
  }

  canManageComments() {
    return true;
  }

  hasMyUserInfoIDAndDisplayName() {
    if (this.connectedUsers === null) {
      return false;
    }

    const myUserInfo = this.connectedUsers.getMyUserInfo();
    return !!(myUserInfo && myUserInfo.ID && myUserInfo.displayName && myUserInfo.displayName !== "");
  }

  trashComment(commentID, trash, untrashData) {
    if (!this.createMyUserIfNotPresent()) {
      return;
    }

    this.comments[commentID].setTrashed(trash);
    this.comments[commentID].setData(untrashData ? untrashData : {});

    if (trash) {
      this.comments[commentID].setTrashedBy(this.connectedUsers.getMyUserInfo().ID);
    } else {
      this.comments[commentID].setTrashedBy("");
    }

    let attributes = {};
    attributes[BarConstants.CommentAttributesSetters.setTrashedStatus] = trash;
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.COMMENTS_ATTRIBUTES_UPDATED, {
      commentIDs: [commentID],
      attributes: [attributes]
    }));
    this.notifyKeyValueChanged(KeyNames.COMMENTS);
  }

  updateCommentTextAndCallouts(commentID, text, callouts) {
    _balsamiq.default.assertUUID(commentID);

    this.createMyUserIfNotPresent();
    this.comments[commentID].updateText(text);
    this.comments[commentID].updateCallouts(callouts);
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.COMMENT_DATA_UPDATED, {
      commentID: commentID,
      data: JSON.stringify(this.comments[commentID].getData())
    }));
    let attributes = {};
    attributes[BarConstants.CommentAttributesSetters.setTimestamps] = ["text"];
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.COMMENTS_ATTRIBUTES_UPDATED, {
      commentIDs: [commentID],
      attributes: [attributes]
    }));
    this.notifyKeyValueChanged(KeyNames.COMMENTS);
  }

  likeComment(commentID, like) {
    if (!this.createMyUserIfNotPresent()) {
      return;
    }

    this.comments[commentID].setLikedBy(this.connectedUsers.getMyUserInfo().ID, like);
    let attributes = {};
    attributes[BarConstants.CommentAttributesSetters.setLikedStatus] = like;
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.COMMENTS_ATTRIBUTES_UPDATED, {
      commentIDs: [commentID],
      attributes: [attributes]
    }));
    this.notifyKeyValueChanged(KeyNames.COMMENTS);
  }

  isCommentRead(commentID) {
    return this.comments[commentID].isReadBy(this.connectedUsers.getMyUserInfo().ID);
  }

  readComments(commentIDs, read) {
    _balsamiq.default.assertArray(commentIDs);

    _balsamiq.default.assertBool(read);

    if (!this.createMyUserIfNotPresent()) {
      return;
    }

    let changedCommentIDs = [];

    for (let commentID of commentIDs) {
      if (this.comments[commentID].isReadBy(this.connectedUsers.getMyUserInfo().ID) != read) {
        this.comments[commentID].setReadBy(this.connectedUsers.getMyUserInfo().ID, read);
        changedCommentIDs.push(commentID);
      }
    }

    if (changedCommentIDs.length > 0) {
      let attributes = {};
      attributes[BarConstants.CommentAttributesSetters.setReadStatus] = read;
      this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.COMMENTS_ATTRIBUTES_UPDATED, {
        commentIDs: changedCommentIDs,
        attributes: Array(changedCommentIDs.length).fill(attributes)
      }));
      this.notifyKeyValueChanged(KeyNames.COMMENTS);
    }
  }
  /**
   *
   * @param {string} resourceID
   * @param {string} branchID
   * @return {boolean}
   */


  resourceHasComments(resourceID, branchID) {
    return Object.keys(this.comments).map(k => this.comments[k]).filter(comment => comment.resourceIDBranchIDIs(resourceID, branchID)).length > 0;
  }
  /**
   * @param {string} resourceID
   * @param {string} branchID
   * @return {Comment[]}
   */


  getCommentsByResource(resourceID, branchID) {
    return Object.values(this.comments).filter(comment => comment.resourceIDBranchIDIs(resourceID, branchID));
  }
  /**
   * @param {string} commentID
   * @return {Comment | undefined}
   */


  getCommentByID(commentID) {
    return this.comments[commentID];
  }
  /**
   *
   * @param {string} resourceID
   * @param {string} branchID
   * @param {string | undefined} parentID
   * @param {string} text
   * @param {any} callouts
   * @param {string} originalAuthorDisplayName
   * @param {boolean=} imported
   * @return {CommentCoreObjectID}
   */


  newComment(resourceID, branchID, parentID, text, callouts, originalAuthorDisplayName) {
    let imported = arguments.length > 6 && arguments[6] !== undefined ? arguments[6] : false;

    _balsamiq.default.assert(this.mockups[resourceID][branchID], "Comments allowed for existing mockups only");

    this.createMyUserIfNotPresent();
    var commentID = CoreObject.createNewUUID();
    const importAttributes = imported ? {
      imported: true,
      originalAuthor: originalAuthorDisplayName
    } : {};
    var attributes = {
      parentID: parentID,
      readBy: [],
      likedBy: [],
      trashed: false,
      trashedBy: "",
      timestamp: -1,
      // comment creation is pending
      timestamps: {},
      ...importAttributes
    };
    var data = {
      text: text,
      callouts: callouts
    };

    let comment = this._makeComment(commentID, branchID, resourceID, this.connectedUsers.getMyUserInfo().ID, attributes, data);

    this.comments[comment.getID()] = comment;
    this.comments[commentID].setReadBy(this.connectedUsers.getMyUserInfo().ID, true);
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.COMMENT_ADDED, {
      coreObjectID: comment.coreObjectID,
      data: data,
      parentID: parentID,
      attributes: importAttributes
    }));
    this.notifyKeyValueChanged(KeyNames.COMMENTS);
    return comment.coreObjectID;
  }
  /**
   *
   * @param {object} params
   * @param {string} params.resourceID
   * @param {string} params.branchID
   * @param {string | undefined} params.parentID
   * @param {string | undefined} params.text
   * @param {unknown} params.callouts
   * @param {{ [key: string]: any }} params.attributes
   * @param {string} params.originalAuthorDisplayName
   * @param {string=} params.commentID
   * @return {CommentCoreObjectID}
   */


  importComment(_ref) {
    let {
      resourceID,
      branchID,
      parentID,
      text,
      callouts,
      attributes: originalAttributes,
      originalAuthorDisplayName,
      commentID = CoreObject.createNewUUID()
    } = _ref;

    _balsamiq.default.assert(this.mockups[resourceID][branchID], "Comments allowed for existing mockups only");

    this.createMyUserIfNotPresent();
    const attributes = { ...originalAttributes,
      [BarConstants.CommentAttributes.ParentID]: parentID
    };
    const data = {
      text,
      callouts
    };

    const comment = this._makeComment(commentID, branchID, resourceID, this.connectedUsers.getMyUserInfo().ID, attributes, data);

    comment.setOriginalAuthor(originalAuthorDisplayName);
    this.comments[comment.getID()] = comment;
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.COMMENT_IMPORTED, {
      coreObjectID: comment.coreObjectID,
      comment: {
        id: comment.getID(),
        data,
        attributes,
        originalAuthorDisplayName
      }
    }));
    this.notifyKeyValueChanged(KeyNames.COMMENTS);
    return comment.coreObjectID;
  }
  /**
   * Adds a comment to the list of comments and notifies the change
   * @param {Comment} comment
   */


  addComment(comment) {
    this.comments[comment.getID()] = comment;
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.COMMENT_ADDED, {
      coreObjectID: comment.coreObjectID,
      data: comment.data,
      parentID: comment.parentID,
      attributes: comment.getAttributes()
    }));
    this.notifyKeyValueChanged(KeyNames.COMMENTS);
  }

  canCreateAlternate(resourceID, newBranchID, newBranchName) {
    if (this.branches[newBranchID]) {
      return newBranchName !== this.branches[newBranchID].getName() ? false : this.mockups[resourceID][newBranchID] === undefined;
    } else if (this.getBranchByName(newBranchName) !== null) {
      return false;
    }

    return true;
  }

  doesAlternateExist(resourceID, branchID) {
    return !!this.mockups[resourceID][branchID];
  }

  setServerTimeDiff(serverTimeDiff) {
    this.serverTimeDiff = serverTimeDiff; // TODO: force COMMENTS UPDATE

    this.notifyKeyValueChanged(KeyNames.COMMENTS);
  }

  getServerTimeDiff() {
    return this.serverTimeDiff;
  }

  setMockupDataAndAttributes(resourceID, branchID, attributes, jsonData) {
    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.assertString(branchID);

    _balsamiq.default.assertObject(jsonData);

    var mockupTarget = this.mockups[resourceID][branchID];
    mockupTarget.load(jsonData);
    let filteredAttributes = this.removeNonMasterAttributes(branchID, attributes);

    for (let attribute in filteredAttributes) {
      mockupTarget.setAttribute(attribute, filteredAttributes[attribute], Constants.OperationOrigin.USER);
    }

    this.updateThumbnail(mockupTarget, Constants.OperationOrigin.USER);
  }

  mergeAlternateMockupVersion(resourceID, branchID) {
    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.assertString(branchID);

    _balsamiq.default.assert(branchID != MASTER);

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.promoteAlternateMockupVersion can not edit Resources because the user role is: " + this.userRole);

    var mockupToMerge = this.mockups[resourceID][branchID];
    var mockupTarget = this.mockups[resourceID][MASTER];
    var oldJSONData = mockupTarget.getJSONData();
    var rootControl = mockupTarget.getRootControl();
    var rootBounds = rootControl.getClipBBox(true);

    _balsamiq.default.exists(mockupToMerge);

    let jsonDATA = mockupToMerge.getJSONData();
    var importResult;
    mockupTarget.performEdit(() => {
      importResult = mockupTarget.import(jsonDATA, [], {
        x: rootBounds.width + 20,
        y: 0
      }, true);
    });

    if (importResult.errors.length) {
      return {
        errors: importResult.errors
      };
    }

    return {
      oldData: oldJSONData,
      newData: mockupTarget.getJSONData(),
      import: importResult
    };
  }

  promoteAlternateMockupVersion(resourceID, branchID) {
    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.assertString(branchID);

    _balsamiq.default.assert(branchID != MASTER);

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.promoteAlternateMockupVersion can not edit Resources because the user role is: " + this.userRole);

    var mockupToPromote = this.mockups[resourceID][branchID];
    var mockupTarget = this.mockups[resourceID][MASTER];
    var oldJSONData = mockupTarget.getJSONData();

    var attributesOld = _balsamiq.default.clone(mockupTarget.getAttributes());

    _balsamiq.default.exists(mockupToPromote); // must copy the alternate data OVER the MASTER branch data and extend the attributes to keep those that are missing in the alternate


    let jsonData = mockupToPromote.getJSONData();

    let attributes = _balsamiq.default.extend(_balsamiq.default.clone(attributesOld), mockupToPromote.getAttributes());

    this.setMockupDataAndAttributes(resourceID, MASTER, attributes, jsonData); // this.deleteAlternate(resourceID, branchID);

    this.notifyKeyValueChanged(KeyNames.MOCKUP_LIST);
    return {
      data: oldJSONData,
      attributes: attributesOld
    };
  }
  /**
   *
   * @param {string} resourceID
   * @param {string} branchID
   * @param {string} newBranchID
   * @param {string} newBranchName
   * @param {string} newBranchNamePrefix
   * @returns {Mockup.Mockup}
   */


  createAlternateMockupVersion(resourceID, branchID, newBranchID, newBranchName, newBranchNamePrefix) {
    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.createAlternateMockupVersion can not edit Resources because the user role is: " + this.userRole);

    if (!newBranchID) {
      newBranchID = this.newBranch(newBranchNamePrefix);
    } else {
      _balsamiq.default.exists(newBranchName);

      _balsamiq.default.assertString(newBranchName);

      if (this.branches[newBranchID]) {
        _balsamiq.default.assert(newBranchName === this.branches[newBranchID].getName());
      } else {
        _balsamiq.default.assert(this.getBranchByName(newBranchName) === null); // must be recreated


        var attributes = {
          branchName: newBranchName
        };
        this.makeBranch(newBranchID, attributes);
      } // last condition: branchID, resourceID must not exist


      _balsamiq.default.assert(this.mockups[resourceID][newBranchID] === undefined);
    }

    const mockupToClone = this.mockups[resourceID][branchID];

    _balsamiq.default.exists(mockupToClone);

    let jsonString = mockupToClone.getData();
    let mockupToCloneAttributes = mockupToClone.getAttributes();

    let clonedAttributes = _balsamiq.default.clone(mockupToCloneAttributes); // copying only attributes meaningful for alternates


    clonedAttributes[_BmprConstants.ResourceAttributes.CreationDate] = mockupToCloneAttributes[_BmprConstants.ResourceAttributes.CreationDate];
    clonedAttributes[_BmprConstants.ResourceAttributes.ModifiedBy] = mockupToCloneAttributes[_BmprConstants.ResourceAttributes.ModifiedBy];
    clonedAttributes[_BmprConstants.ResourceAttributes.Notes] = mockupToCloneAttributes[_BmprConstants.ResourceAttributes.Notes];
    this.addMasterAttributes(resourceID, newBranchID, clonedAttributes);

    let resource = this._makeResource(newBranchID, mockupToClone.coreObjectID.resourceID, clonedAttributes, jsonString);

    this._addResource(resource);

    var newThumbnailAttributes = {
      branchID: newBranchID,
      resourceID: resource.coreObjectID.resourceID
    };

    let thumbNail = this._makeThumbnail(CoreObject.createNewUUID(), newThumbnailAttributes);

    this._addThumbnail(thumbNail);

    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.RESOURCE_ADDED, {
      coreObjectID: resource.coreObjectID,
      data: jsonString
    }));
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.THUMBNAIL_ADDED, {
      thumbnailID: thumbNail.coreObjectID.thumbnailID,
      attributes: newThumbnailAttributes
    }));
    this.updateThumbnail(resource, Constants.OperationOrigin.USER);
    this.notifyKeyValueChanged(KeyNames.MOCKUP_LIST);
    return resource;
  }

  pauseThumbnailQueue() {
    this.thumbnailQueuePauseCounter++;

    if (this.thumbnailQueuePauseCounter === 1) {
      this.thumbnailQueue.pause(true);
    }
  }

  playThumbnailQueue() {
    this.thumbnailQueuePauseCounter--;

    _balsamiq.default.assert(this.thumbnailQueuePauseCounter >= 0, "thumbnailQueuePauseCounter must be >= 0");

    if (this.thumbnailQueuePauseCounter === 0) {
      this.thumbnailQueue.pause(false);
    }
  }
  /**
   * @param {object[]} resources
   * @param {Project=} sourceProject
   * @param {{[branchID: string]: string}=} branches
   * @param {((index, total) => void)=} progressCallback
   * @returns {Promise<void>}
   */


  async importResources(resources, sourceProject, branches, progressCallback) {
    const addBranchToProjectIfNeeded = (branchID, branchName) => {
      if (!this.branches[branchID]) {
        if (sourceProject) _balsamiq.default.assertClass(sourceProject, Project, "when there are resources with alternates, the source project must be passed in");
        const branchAttributes = {
          branchName: branchName || sourceProject.branches[branchID].attributes.branchName
        };
        this.makeBranch(branchID, branchAttributes);
      }
    };

    const addResourceToProject = resource => {
      const attributesWithImportedFrom = { ...resource.attributes,
        importedFrom: resource.attributes[_BmprConstants.ResourceAttributes.ImportedFrom] || resource.originalResourceID,
        name: this.makeNewResourceName(resource.attributes[_BmprConstants.ResourceAttributes.Kind], resource.attributes[_BmprConstants.ResourceAttributes.Name])
      };
      addBranchToProjectIfNeeded(resource.branchID, branches && branches[resource.branchID]);
      this.addResource(resource.data, attributesWithImportedFrom, resource.resourceID, resource.branchID, {
        notifyListChanges: false
      });
    };

    this.pauseThumbnailQueue();

    for (let i = 0; i < resources.length; i++) {
      const resource = resources[i];
      await new Promise(resolve => {
        this.context.host.asyncCallback(() => {
          if (progressCallback) progressCallback(i, resources.length);
          addResourceToProject(resource, i);
          resolve();
        });
      });
    }

    recomputeSymbolLibraries(this.symbolLibraries);
    this.playThumbnailQueue();
    this.notifyAllProjectDataChanged();
  }

  addAlternateMockupVersion(mockup, branchName) {
    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.addAlternateMockupVersion can not edit Resources because the user role is: " + this.userRole);

    var branchID = mockup.coreObjectID.branchID;

    if (!this.branches[branchID]) {
      // must be created
      _balsamiq.default.assert(this.getBranchByName(branchName) === null);

      var attributes = {
        branchName: branchName
      };
      this.makeBranch(branchID, attributes);
    }

    var keyNames = this._addResource(mockup);

    keyNames.forEach(keyName => {
      this.notifyKeyValueChanged(keyName);
    });
    let jsonString = mockup.getData();
    this.updateThumbnail(mockup, Constants.OperationOrigin.USER);
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.RESOURCE_ADDED, {
      coreObjectID: mockup.coreObjectID,
      data: jsonString
    }));
  }

  cloneAlternateAsNew(resourceID, branchID) {
    _balsamiq.default.assertUUID(resourceID);

    _balsamiq.default.assertUUID(branchID); // Should not be Master!
    // Get alteranate, master mockup and branch


    var masterMockup = this.getResourceID(resourceID, BarConstants.Branch.MasterBranchID);
    var alternate = this.getResourceID(resourceID, branchID);
    var branch = this.getBranchOfResource(alternate.coreObjectID); // Get data and attributes to clone

    var sortedMockups = this.getFlatMockupList();
    var data = JSON.stringify(alternate.getJSONData());
    var attributes = alternate.getAttributesClone();
    this.addMasterAttributes(resourceID, branchID, attributes);
    let name = branch.getName() + " " + masterMockup.getName();
    if (this.existsName(name, sortedMockups)) name = this.appendCopyAndNumber(name, sortedMockups);
    attributes[_BmprConstants.ResourceAttributes.Name] = name;
    attributes[_BmprConstants.MockupAttributes.Order] = this.getNewMockupOrder();
    var newCoreObjectID = this.addResource(data, attributes); // Put it at the right order
    // cloned should be dropped AFTER the last descendant index
    // fix mockup order: after original but not at top level if we cloned a tree child

    this.reorderClonedMockupAfterOriginal(resourceID, newCoreObjectID.resourceID);
    return newCoreObjectID;
  }

  makeNewMockupName(name) {
    _balsamiq.default.assertString(name);

    const getFlatMockupList = this.getFlatMockupList();
    return this.existsName(name, getFlatMockupList) ? this.appendCopyAndNumber(name, getFlatMockupList) : name;
  }

  cloneMockup(resourceID) {
    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.cloneMockup can not edit Resources because the user role is: " + this.userRole);

    var sortedMockups = this.getMockupList();
    var index = sortedMockups.findIndex(sublist => {
      return sublist[0].coreObjectID.resourceID === resourceID;
    });

    _balsamiq.default.assert(index !== -1);

    var mockupsToClone = sortedMockups[index];
    var len = sortedMockups.length;
    var uuidOfClonedMockups = CoreObject.createNewUUID();
    var mockupOrder = (len == 0 ? 1000000 : sortedMockups[len - 1][0].getOrder() + 1000000) + Math.random() * 1000000;

    for (let mockup of mockupsToClone) {
      let jsonString = mockup.getData();

      let clonedAttributes = _balsamiq.default.clone(mockup.getAttributes());

      this.addMasterAttributes(resourceID, mockup.coreObjectID.branchID, clonedAttributes);
      const name = this.makeNewMockupName(clonedAttributes[_BmprConstants.ResourceAttributes.Name]);
      clonedAttributes[_BmprConstants.ResourceAttributes.Name] = name;
      clonedAttributes[_BmprConstants.MockupAttributes.Order] = mockupOrder;

      let resource = this._makeResource(mockup.coreObjectID.branchID, uuidOfClonedMockups, clonedAttributes, jsonString);

      this._addResource(resource);

      var newThumbnailAttributes = {
        branchID: resource.coreObjectID.branchID,
        resourceID: resource.coreObjectID.resourceID
      };

      let thumbNail = this._makeThumbnail(CoreObject.createNewUUID(), newThumbnailAttributes);

      this._addThumbnail(thumbNail);

      this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.RESOURCE_ADDED, {
        coreObjectID: resource.coreObjectID,
        data: jsonString
      }));
      this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.THUMBNAIL_ADDED, {
        thumbnailID: thumbNail.coreObjectID.thumbnailID,
        attributes: newThumbnailAttributes
      }));
      this.updateThumbnail(resource, Constants.OperationOrigin.USER);
    }

    return uuidOfClonedMockups;
  }
  /**
   *
   * @param {string[]} originalResourceIDs
   * @param {string[]} clonedResourceIDs
   */


  reorderClonedMockups(originalResourceIDs, clonedResourceIDs) {
    let tree = this.getMockupsTree();
    tree = (0, _reorderAddedResources.reorderAddedResources)({
      kind: _reorderAddedResources.OperationType.DUPLICATE,
      currentResourceTree: tree,
      sourceResourceIDs: originalResourceIDs,
      targetResourceIDs: clonedResourceIDs
    });
    this.updateResourcesParentIDFromTree(tree);
    this.updateResourcesOrderFromTree(tree);
  }
  /**
   *
   * @param {ResourceTree} sourceResourceTree
   * @param {string[]} sourceResourceIDs
   * @param {string[]} pastedResourceIDs
   * @param {string[]} selectedResourceIDs
   * @param {import("core/utils/expandable-resource-tree").ExpandableTree } currentResourceTree
   */


  reorderPastedMockups(sourceResourceTree, sourceResourceIDs, pastedResourceIDs, selectedResourceIDs, currentResourceTree) {
    let tree = this.getMockupsTree();
    tree = (0, _reorderAddedResources.reorderAddedResources)({
      kind: _reorderAddedResources.OperationType.PASTE,
      sourceResourceTree,
      currentResourceTree,
      sourceResourceIDs: sourceResourceIDs,
      targetResourceIDs: pastedResourceIDs,
      selectedResourceIDs
    });
    this.updateResourcesParentIDFromTree(tree);
    this.updateResourcesOrderFromTree(tree);
  }

  reorderClonedMockupAfterOriginal(originalResourceID, clonedResourceID) {
    // fix mockup order: after original but not at top level if we cloned a tree child
    let list = this.getMockupList();
    let mtree = this.getMockupsTree();
    let resourceTreeItem = mtree.getResourceByID(originalResourceID);
    let resourceTreeItemIndex = mtree.getIndex(resourceTreeItem);
    let originalResourceIndex = list.findIndex(sublist => {
      return sublist[0].coreObjectID.resourceID === originalResourceID;
    });
    let clonedResourceIndex = list.findIndex(sublist => {
      return sublist[0].coreObjectID.resourceID === clonedResourceID;
    }); // cloned should be dropped AFTER the last descendant index

    let descendantIndexes = mtree.getResourceDescendantsByIndex(resourceTreeItemIndex, false);
    let index = originalResourceIndex;

    if (descendantIndexes.length !== 0) {
      let lastMockupResourceID = mtree.getResource(descendantIndexes[descendantIndexes.length - 1]).resourceID;
      list = this.getMockupList();
      index = list.findIndex(function (sublist) {
        return sublist[0].coreObjectID.resourceID == lastMockupResourceID;
      });
    }

    this.reorderMockups([clonedResourceIndex], index + 1, {
      atSameLevelOfIndex: originalResourceIndex
    });
    this.notifyKeyValueChanged(KeyNames.MOCKUP_LIST);
  }

  newBranch(newBranchNamePrefix) {
    var nameToTry = "";
    var unique = false;

    while (!unique) {
      nameToTry = (newBranchNamePrefix ? newBranchNamePrefix + " " : "Alternate ") + Math.round(Math.random() * 1000) + String.fromCharCode(Math.round(Math.random() * 26) + "a".charCodeAt(0));
      var found = this.getBranchByName(nameToTry) != null;
      unique = !found;
    }

    var attributes = {
      branchName: nameToTry
    };
    var uid = CoreObject.createNewUUID();
    this.makeBranch(uid, attributes);
    return this.branches[uid].coreObjectID.branchID;
  }

  _setResourceTrashed(resourceID, trashed) {
    var list = this._findResourceList(resourceID);

    _balsamiq.default.exists(list);

    _balsamiq.default.exists(list.list[resourceID]);

    list.list[resourceID][MASTER].setTrashed(trashed);
  }
  /**
   *
   * @param {string} branchID
   * @param {string} resourceID
   * @returns {void}
   */


  deleteMockupAndThumbnail(branchID, resourceID) {
    _balsamiq.default.assertString(branchID);

    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.deleteMockupAndThumbnail can not edit Resources because the user role is: " + this.userRole);

    let resource = this._findResource(branchID, resourceID);

    _balsamiq.default.exists(resource);

    let thumbnail = this.getThumbnailOfResource(resource.coreObjectID);

    if (thumbnail) {
      this.deleteThumbnails([thumbnail.coreObjectID.thumbnailID]);
    }

    this.deleteResource(branchID, resourceID);
  }

  deleteThumbnails(thumbnailIDs) {
    for (let thumbnailID of thumbnailIDs) {
      this._deleteThumbnail(thumbnailID);
    }

    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.THUMBNAILS_DELETED, {
      thumbnailIDs: thumbnailIDs
    }));
  }

  deletePermalink(_ref2) {
    let {
      branchID,
      resourceID
    } = _ref2;

    if (!this.permalinks.hasPermalinkFor({
      branchID,
      resourceID
    })) {
      return;
    }

    this.permalinks.delete({
      branchID,
      resourceID
    });
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.PERMALINK_DELETED, {
      branchID,
      resourceID
    }));
  }
  /**
   *
   * @param {string} branchID
   * @param {string} resourceID
   * @param {string[]=} commentIDsToKeep
   * @returns {void}
   */


  deleteResourceComments(branchID, resourceID, commentIDsToKeep) {
    _balsamiq.default.assertString(branchID);

    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.deleteResource can not edit Resources because the user role is: " + this.userRole);

    const comments = this.getCommentsByResource(resourceID, branchID);

    if (comments.length > 0) {
      let commentIDs = [];

      for (let comment of comments) {
        const commentID = comment.getID();

        if (!commentIDsToKeep || !commentIDsToKeep.includes(commentID)) {
          commentIDs.push(commentID);
          delete this.comments[commentID];
        }
      }

      this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.COMMENTS_DELETED, {
        commentIDs: commentIDs
      }));
    }
  }
  /**
   *
   * @param {string} branchID
   * @param {string} resourceID
   * @returns {void}
   */


  deleteResource(branchID, resourceID) {
    _balsamiq.default.assertString(branchID);

    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.assert(this.userRole >= Roles.EDITOR, "Project.deleteResource can not edit Resources because the user role is: " + this.userRole);

    this.deleteResourceComments(branchID, resourceID); // get data and attributes or resource

    var keyNames = this._deleteResource(branchID, resourceID);

    keyNames.forEach(keyName => {
      this.notifyKeyValueChanged(keyName);
    });
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.RESOURCES_DELETED, {
      branchID: branchID,
      resourceIDs: [resourceID]
    })); // TODO return attributes and data of the resource for undo/redo
  } // API for lower level (Archive)
  // -----------------------------
  // 1.2 to 2.0 migration: these attributes are no more allowed in non MASTER resources
  // check on mockups only since there are no alternates for other kinds


  addMasterAttributes(resourceID, branchID, attributes) {
    // must copy some (all for now) attributes from master because they are missing in alternates
    // then, I have to remove them BEFORE really creating the resource
    if (branchID != MASTER) {
      let masterResource = this._findResource(MASTER, resourceID);

      let masterResourceAttributes = masterResource.getAttributes();

      for (let attribute in masterResourceAttributes) {
        if (attribute in attributes) {
          continue;
        }

        attributes[attribute] = masterResourceAttributes[attribute];
      }
    }
  }

  removeNonMasterAttributes(branchID, attributes) {
    let filteredAttributes = _balsamiq.default.clone(attributes);

    if (attributes[_BmprConstants.ResourceAttributes.Kind] === _BmprConstants.ResourceKinds.Mockup && branchID != MASTER) {
      delete filteredAttributes[_BmprConstants.ResourceAttributes.Name];
      delete filteredAttributes[_BmprConstants.ResourceAttributes.Kind];
      delete filteredAttributes[_BmprConstants.ResourceAttributes.Trashed];
      delete filteredAttributes[_BmprConstants.ResourceAttributes.ImportedFrom];
      delete filteredAttributes[_BmprConstants.MockupAttributes.Order];
      delete filteredAttributes[_BmprConstants.MockupAttributes.ParentID];
    }

    return filteredAttributes;
  }

  isLoading() {
    return this._isLoading;
  }

  _setIsLoading(value) {
    _balsamiq.default.assertBool(value);

    let oldValue = this._isLoading;

    if (oldValue != value) {
      this._isLoading = value;
      this.notifyKeyValueChanged(KeyNames.IS_LOADING);
    }
  }

  _removeParentIDfromCommentsWithMissingParent() {
    for (let commentID in this.comments) {
      const comment = this.comments[commentID];

      if (comment.parentID && !(comment.parentID in this.comments)) {
        comment.parentID = "";
      }
    }
  }
  /**
   * Asynchronously create and load projects' resources
   * @param {*} data
   * @param {import("core/utils/resource-skimmer")} resourceSkimmer
   * @param {(res: unknown) => void} callback
   * @param {((title: string; step: number; total: number) => void)=} setProgress
   * @param {(() => boolean)=} isCanceled
   */


  load(data, resourceSkimmer, callback) {
    let setProgress = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : () => {};
    let isCanceled = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : () => false;

    _balsamiq.default.assert(!this._isLoading, "Project.load can not be called when the project is already loading");

    _balsamiq.default.assert(data.attributes);

    _balsamiq.default.assert(data.branches);

    _balsamiq.default.assert(data.resources);

    _balsamiq.default.assert(data.thumbnails);

    _balsamiq.default.assert(data.users);

    _balsamiq.default.assert(data.comments);

    _balsamiq.default.assertNumber(data.userRole);

    _balsamiq.default.assertFunction(callback);

    _balsamiq.default.assertFunction(setProgress);

    _balsamiq.default.assertFunction(isCanceled, "isCanceled is a required function");

    this._setIsLoading(true);

    let timeID = this.context.execTime.start("Load Project");
    var oldProject = {
      userRole: this.userRole,
      overriddenUserRole: this.overriddenUserRole,
      attributes: this.attributes,
      branches: this.branches,
      mockups: this.mockups,
      symbolLibraries: this.symbolLibraries,
      assets: this.assets,
      thumbnails: this.thumbnails,
      permalinks: this.permalinks,
      comments: this.comments,
      // TODO add to notifyError
      users: this.users // TODO add to notifyError

    };

    const notifyError = error => {
      // Cleanup project as it was before
      this.userRole = oldProject.userRole;
      this.overriddenUserRole = oldProject.overriddenUserRole;
      this.attributes = oldProject.attributes;
      this.branches = oldProject.branches;
      this.mockups = oldProject.mockups;
      this.symbolLibraries = oldProject.symbolLibraries;
      this.assets = oldProject.assets;
      this.thumbnails = oldProject.thumbnails;
      this.permalinks = oldProject.permalinks; // Notify all back to normal

      this._setIsLoading(false);

      this.notifyProjectAttributesChanged();
      this.notifyAllProjectDataChanged();

      this._setIsLoading(false); // TODO double?


      return callback({
        error: error
      });
    };

    this.userRole = this.overriddenUserRole = data.userRole;

    if (data.startAsReviewer) {
      this.setOverriddenUserRole(Constants.UserRoles.COMMENTER);
    }

    this.attributes = data.attributes;
    this.branches = {};
    this.mockups = {};
    this.symbolLibraries = {};
    this.assets = {};
    this.thumbnails = {};
    this.comments = {};
    this.users = {};
    this.mentionableUsers = {}; // Notify that old resources have been invalidated

    this.notifyAllProjectDataChanged();
    let hasUnknownProperties = false;

    try {
      _balsamiq.default.assertEnum(data.userRole, Roles); // create branches and resources


      data.branches.forEach(function (branch) {
        this._addBranch(this._makeBranch(branch.branchID, branch.attributes));
      }, this); // sort to create MASTER branch resources first
      // this because only MASTER resources have the complete set of attributes

      data.resources.sort(function (a, b) {
        if (a.branchID === MASTER) {
          return -1;
        } else if (b.branchID === MASTER) {
          return 1;
        }

        return a.branchID.localeCompare(b.branchID);
      }); // find the kind of each resource given the resourceID

      const resourceKinds = {}; // resourceID -> resource kind

      data.resources.forEach(resource => {
        if (resource.branchID === MASTER) {
          _balsamiq.default.exists(resource.attributes, "Resource with no attributes");

          const kind = resource.attributes[_BmprConstants.ResourceAttributes.Kind];

          _balsamiq.default.exists(kind, "Resource with no Kind attribute");

          const expectedKinds = [_BmprConstants.ResourceKinds.Mockup, _BmprConstants.ResourceKinds.SymbolLibrary, _BmprConstants.ResourceKinds.Asset, _BmprConstants.ResourceKinds.OtherAsset];

          _balsamiq.default.assert(expectedKinds.includes(kind), "Unknown resource kind: " + kind);

          resourceKinds[resource.resourceID] = kind;
        }
      }); // create resources in this order to honour the dependencies: Assets, Symbol Libraries, Mockups

      const mockups = [];
      const symbolLibraries = [];
      const assets = [];
      data.resources.forEach(resource => {
        const kind = resourceKinds[resource.resourceID];

        switch (kind) {
          case _BmprConstants.ResourceKinds.Mockup:
            if (!resourceSkimmer.isResourceSelected({
              ID: resource.resourceID,
              BRANCHID: resource.branchID
            })) {
              return;
            }

            mockups.push(resource);
            break;

          case _BmprConstants.ResourceKinds.SymbolLibrary:
            symbolLibraries.push(resource);
            break;

          case _BmprConstants.ResourceKinds.Asset:
          case _BmprConstants.ResourceKinds.OtherAsset:
            assets.push(resource);
            break;
        }
      });
      let resourcesTimeID = this.context.execTime.start("Load Resources");

      let createResource = (resource, done) => {
        if (isCanceled()) return done();

        const doneWithLogging = () => {
          this.context.execTime.lap(resourcesTimeID, `kind:${resource.attributes[_BmprConstants.ResourceAttributes.Kind]}, name:${resource.attributes[_BmprConstants.ResourceAttributes.Name]}, branchID:${resource.branchID.substr(0, 4)}, resourceID:${resource.resourceID.substr(0, 4)}`);
          done();
        };

        let attributes = resource.attributes;
        const isWireframeOrSymbolLibrary = attributes[_BmprConstants.ResourceAttributes.Kind] === _BmprConstants.ResourceKinds.Mockup || attributes[_BmprConstants.ResourceAttributes.Kind] === _BmprConstants.ResourceKinds.SymbolLibrary;

        if (isWireframeOrSymbolLibrary) {
          const v = Mockup.validateMockupData(resource.data);

          if (v.error) {
            done(v.error);
            return;
          }
        }

        this.addMasterAttributes(resource.resourceID, resource.branchID, attributes);

        let res = this._makeResource(resource.branchID, resource.resourceID, attributes, resource.data);

        this._addResource(res);

        if (isWireframeOrSymbolLibrary) {
          hasUnknownProperties = hasUnknownProperties || res.hasUnknownProperties();
        }

        if (res instanceof Image.Image) {
          this.addPendingAssetID(res.coreObjectID.resourceID);
          res.makeNativeImage(Constants.OperationOrigin.ARCHIVE, doneWithLogging);
        } else {
          doneWithLogging();
        }
      };

      const eachParallel = async.eachParallel.bind(null, this.context.host.timer.bind(null, 0));

      const loadNAssetsInParallel = sliceBy => _balsamiq.default.chunk(assets, sliceBy).map(assetsSlice => done => eachParallel(assetsSlice, createResource, done));

      const checkResourcesLoadedCorrectly = done => {
        this.context.execTime.end(resourcesTimeID);
        if (isCanceled()) return done(); // Master branch must exist and must have some attributes

        _balsamiq.default.exists(this.branches[MASTER], "Master branch must exist");

        done();
      };

      const createThumbnail = thumbnail => done => {
        if (isCanceled()) return done();

        this._addThumbnail(this._makeThumbnail(thumbnail.thumbnailID, thumbnail.attributes));

        done();
      };

      const loadUser = user => done => {
        if (isCanceled()) return done();
        this.users[user.userID] = this._makeUser(user.userID, user.attributes);
        done();
      };

      const loadComment = comment => done => {
        if (isCanceled()) return done();
        let dataObj;

        try {
          dataObj = JSON.parse(comment.data);
        } catch (e) {
          this.context.host.log("Project: Cannot parse Comment DATA. " + (comment ? comment.data : "(comment is undefined)"));
          dataObj = {
            text: "<This comment cannot be read>"
          };
        }

        this.comments[comment.commentID] = this._makeComment(comment.commentID, comment.branchID, comment.resourceID, comment.userID, comment.attributes, dataObj);
        done();
      };

      const tasks = [...loadNAssetsInParallel(4), // when mapping here we return a list of async functions (callback) => { doSomeWork(); callback(); }
      // and we save the item (ex: symbolLibrary) in the clojure so that it can be used by the inner function (ex: createResource)
      ...symbolLibraries.map(sl => done => createResource(sl, done)), recomputeSymbolLibraries.bind(this, this.symbolLibraries), ...mockups.map(m => done => createResource(m, done)), checkResourcesLoadedCorrectly, ...data.thumbnails.map(createThumbnail), ...data.users.map(loadUser), ...data.comments.map(loadComment)];

      const progressCallback = index => {
        if (!isCanceled()) {
          setProgress("Loading resources...", index + 1, tasks.length);
        }
      };

      async.queue(tasks, error => {
        this._removeParentIDfromCommentsWithMissingParent();

        if (isCanceled()) {
          return callback();
        }

        if (error) {
          return notifyError(error);
        }

        this.context.execTime.lap(timeID, "all loaded"); // loop to refresh all thumbnails

        this._loopOnResources(this.mockups, this.updateThumbnail.bind(this), {
          includeTrashed: false
        });

        this._loopOnResources(this.symbolLibraries, this.updateThumbnail.bind(this), {
          includeTrashed: false
        });

        this.context.execTime.lap(timeID, "updateThumbnail"); // events and notifications

        this.notifyProjectAttributesChanged();
        this.notifyAllProjectDataChanged();
        this.context.execTime.lap(timeID, "notify");

        this._setIsLoading(false);

        this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.PROJECT_LOADED));
        this.context.execTime.lap(timeID, "PROJECT_LOADED"); // in case user changed his login info in cloud

        if (this.currentUserIsPresentInArchive()) {
          this.updateMyUserIfNeeded();
        }

        this.context.execTime.end(timeID);
        callback({
          hasUnknownProperties
        });

        if (hasUnknownProperties) {
          this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.UNKNOWN_PROPERTIES_LOADED, {
            type: "project"
          }));
        }
      }, {
        progressCb: progressCallback,
        asyncfy: this.context.host.timer.bind(null, 0)
      });
    } catch (error) {
      notifyError(error);
    }
  }

  notifyAllProjectDataChanged() {
    this.notifyKeyValueChanged(KeyNames.BRANCH_LIST);
    this.notifyKeyValueChanged(KeyNames.MOCKUP_LIST);
    this.notifyKeyValueChanged(KeyNames.ASSET_LIST);
    this.notifyKeyValueChanged(KeyNames.SYMBOLLIBRARY_LIST);
    this.notifyKeyValueChanged(KeyNames.TRASH_LIST);
    this.notifyKeyValueChanged(KeyNames.USER_ROLE);
    this.notifyKeyValueChanged(KeyNames.COMMENTS);
    this.notifyKeyValueChanged(KeyNames.USERS);
  }

  notifyProjectAttributesChanged() {
    this.notifyKeyValueChanged(KeyNames.NAME);
    this.notifyKeyValueChanged(KeyNames.DESCRIPTION);
    this.notifyKeyValueChanged(KeyNames.CREATION_DATE);
    this.notifyKeyValueChanged(KeyNames.SKIN_NAME);
    this.notifyKeyValueChanged(KeyNames.FONT_FACE);
    this.notifyKeyValueChanged(KeyNames.FONT_SIZE);
    this.notifyKeyValueChanged(KeyNames.SELECTION_COLOR);
    this.notifyKeyValueChanged(KeyNames.LINK_COLOR);
  }

  onArchiveAttributesChanged(attributes) {
    _balsamiq.default.exists(attributes);

    this.attributes = attributes;
    this.notifyKeyValueChanged(KeyNames.NAME);
  }

  onBranchAdded(branchID, attributes) {
    _balsamiq.default.assertString(branchID);

    _balsamiq.default.exists(attributes);

    this._addBranch(this._makeBranch(branchID, attributes));

    this.notifyKeyValueChanged(KeyNames.BRANCH_LIST);
  }

  onBranchAttributesChanged(branchID, attributes) {
    _balsamiq.default.assertString(branchID);

    _balsamiq.default.exists(attributes);

    _balsamiq.default.exists(this.branches[branchID], "unknown branch with id: " + branchID);

    this.branches[branchID].onAttributesChanged(attributes);
  }

  onBranchesDeleted(branchIDs) {
    _balsamiq.default.assertArray(branchIDs);

    branchIDs.forEach(function (branchID) {
      this._deleteBranch(branchID);
    }, this);
  }

  onResourceAdded(branchID, resourceID, attributes) {
    _balsamiq.default.assertString(branchID);

    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.exists(attributes); // TODO check unique (branchID, resourceID)


    this.addMasterAttributes(resourceID, branchID, attributes);

    const resource = this._makeResource(branchID, resourceID, attributes, "");

    const keyNames = this._addResource(resource);

    keyNames.forEach(keyName => this.notifyKeyValueChanged(keyName));
  }

  forceSaveAll() {
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.PROJECT_ATTRIBUTES_DIRTY, this.attributes));

    var contextChangeFun = function (resource) {
      resource.notifyAttributesDirty();
    };

    var contextChangeDataFun = function (resource) {
      resource.notifyAttributesDirty();
      resource.notifyDataDirty({
        data: resource.getData()
      });
    };

    this._loopOnResources(this.assets, contextChangeFun);

    this._loopOnResources(this.symbolLibraries, contextChangeDataFun);

    this._loopOnResources(this.mockups, contextChangeDataFun);

    for (var branchID in this.branches) {
      this.branches[branchID].notifyAttributesDirty();
    }
  }

  onResourceAttributesChanged(branchID, resourceID, attributes) {
    _balsamiq.default.assertString(branchID);

    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.exists(attributes);

    var r = this._findResource(branchID, resourceID); //Balsamiq.exists(r);


    r && r.onAttributesChanged(this.removeNonMasterAttributes(branchID, attributes));
  }

  onResourceDataChanged(branchID, resourceID, data) {
    return this._updateResourceData({
      branchID,
      resourceID,
      data,
      operationOrigin: Constants.OperationOrigin.ARCHIVE
    });
  }
  /**
   *
   * @param {string} branchID
   * @param {string} resourceID
   * @param {string} data
   * @returns {{ ok: boolean, error?: string }}
   */


  updateResourceData(branchID, resourceID, data) {
    return this._updateResourceData({
      branchID,
      resourceID,
      data,
      operationOrigin: Constants.OperationOrigin.USER
    });
  }

  _updateResourceData(_ref3) {
    let {
      resourceID,
      branchID,
      data,
      operationOrigin
    } = _ref3;

    _balsamiq.default.assertString(branchID, "branchID");

    _balsamiq.default.assertString(resourceID, "resourceID");

    _balsamiq.default.assertString(data, "data");

    _balsamiq.default.assertEnum(operationOrigin, Constants.OperationOrigin);

    const resource = this._findResource(branchID, resourceID);

    if (!resource) {
      return {
        ok: false,
        error: "Resource not found"
      };
    }

    const isMockupOrSymbolLibrary = resource.getType() === _BmprConstants.ResourceKinds.Mockup || resource.getType() === _BmprConstants.ResourceKinds.SymbolLibrary;

    if (isMockupOrSymbolLibrary) {
      const v = Mockup.validateMockupData(data);

      if (v.error) {
        return v;
      }
    }

    resource.setData(data, operationOrigin);

    if (resource.getType() == Constants.ResourceTypes.SYMBOLLIBRARY) {
      this.notifyKeyValueChanged(KeyNames.SYMBOLLIBRARY_LIST);
    }

    if (resource instanceof Image.Image) {
      resource.makeNativeImage(operationOrigin);
      this.addPendingAssetID(resource.coreObjectID.resourceID);
    }

    return {
      ok: true
    };
  }

  onResourcesDeleted(branchID, resourceIDs) {
    _balsamiq.default.assertString(branchID);

    _balsamiq.default.assertArray(resourceIDs);

    var keyNames = [];
    resourceIDs.forEach(resourceID => {
      var branchIDs;

      if (branchID === BarConstants.Branch.AllBranches) {
        var list = this._findResourceList(resourceID).list;

        branchIDs = Object.keys(list[resourceID]);
      } else {
        branchIDs = [branchID];
      }

      branchIDs.forEach(branchID2 => {
        keyNames = keyNames.concat(this._deleteResource(branchID2, resourceID));
      });
    });

    _balsamiq.default.uniq(keyNames);

    keyNames.forEach(keyName => {
      this.notifyKeyValueChanged(keyName);
    });
  }

  onThumbnailAdded(thumbnailID, attributes) {
    _balsamiq.default.assertString(thumbnailID);

    if (!this.thumbnails[thumbnailID]) {
      // create the thumbnail
      this._addThumbnail(this._makeThumbnail(thumbnailID, attributes));
    } // look for an existing resource with this thumbnailID
    // TODO: this should search for mockups only ... ?


    var foundResource = this._findResource(attributes.branchID ? attributes.branchID : MASTER, attributes.resourceID);

    if (foundResource) {
      this.updateThumbnail(foundResource, Constants.OperationOrigin.ARCHIVE);
    }
  }

  onThumbnailsDeleted(thumbnailIDs) {
    _balsamiq.default.assertArray(thumbnailIDs);

    thumbnailIDs.forEach(function (thumbnailID) {
      thumbnailID = thumbnailID.replace(/'/g, ""); // bug coming from PubNub

      this._deleteThumbnail(thumbnailID);
    }, this);
  }

  onPermalinkDataDeleted(_ref4) {
    let {
      resourceID,
      branchID
    } = _ref4;
    this.permalinks.delete({
      resourceID,
      branchID
    });
  }

  onPermalinkCreatedOrUpdated(permalink) {
    this.permalinks.addOrUpdate(permalink);
  }

  onAllPermalinksUpdated(permalinks) {
    this.permalinks.init(permalinks);
  }

  notifyUnknownProperties(resource) {
    if (resource.hasUnknownProperties()) {
      this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.UNKNOWN_PROPERTIES_LOADED, {
        type: "resource"
      }));
    }
  }
  /**
   *
   * @param {{ resourceID: string; subResourceID: string | undefined; }} controlIDs
   */


  notifyControlsWithLoops(controlIDs) {
    _balsamiq.default.assertArray(controlIDs);

    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.CONTROLS_WITH_LOOPS, controlIDs));
  } // Branch, Resource contents and dependencies
  // ------------------------------------------


  getChangedSubResourceIDs(controlChanges) {
    if (!controlChanges) return [];
    const changes = controlChanges.getChanges();

    const relevantChanges = _balsamiq.default.filter(changes, change => {
      const controlID = change.coreObjectID.controlID;
      if (controlID.length === 0) return false; // it's change in the mainGroup => nothing depends on that

      if (controlID.length === 1 && change.position === true && Object.keys(change).length === 2) return false; // it's change of position of mainGroup children => nothing depends on that

      return true;
    });

    if (relevantChanges.length === 0) return [];

    const takeOnlyGroupID = controls => controls.map(_lodash.default.first);

    const removeUndefined = xs => _balsamiq.default.filter(xs, x => x != null);

    const removeDuplicates = _balsamiq.default.uniq;

    const stringify = controls => controls.map(controlID => controlID.toString());

    const subResourceIDs = _lodash.default.flow(takeOnlyGroupID, removeUndefined, removeDuplicates, stringify)(relevantChanges.map(c => c.coreObjectID.controlID));

    return subResourceIDs;
  }

  onResourceContentsChanged(event) {
    let changedResourceID = event.senderCoreObjectID.resourceID;
    let changedResource = this.getResource(event.senderCoreObjectID);

    if (changedResource === null) {
      return;
    }

    if (changedResource instanceof Image.Image && changedResource.nativeStatus !== Image.NativeStatus.PENDING) {
      this.removePendingAssetID(changedResource.coreObjectID.resourceID);
      return;
    }

    const changedResourceType = changedResource.getType();

    switch (changedResourceType) {
      case Constants.ResourceTypes.MOCKUP:
        this.updateThumbnail(changedResource, event.eventData.operationOrigin);
        this.notifyUnknownProperties(changedResource);
        return;
      // nothing depends on a mockup

      case Constants.ResourceTypes.ASSET:
        {
          const relatedControls = this.symbolMap.controlsDependencies.getAffectedByResourceChange(changedResourceID);
          this.updateRelatedControlsThumbnails(relatedControls);
          return;
        }

      case Constants.ResourceTypes.SYMBOLLIBRARY:
        {
          if (event.eventData.wholeMockupChanged) {
            this.updateDependencies(event.senderCoreObjectID.resourceID, event.eventData.operationOrigin, []);
            this.updateThumbnail(changedResource, event.eventData.operationOrigin);
            this.notifyKeyValueChanged(KeyNames.SYMBOLLIBRARY_LIST);
          } else {
            const subResourceIDs = this.getChangedSubResourceIDs(event.eventData.controlChanges);
            if (subResourceIDs.length === 0) return; // we never want to update all controls depending on a SL

            this.updateDependencies(event.senderCoreObjectID.resourceID, event.eventData.operationOrigin, subResourceIDs);
            this.updateThumbnail(changedResource, event.eventData.operationOrigin); // Check if some symbols were added or removed and notify that symbol list changed

            if (this.symbolLibraryListChanged(event)) {
              this.notifyKeyValueChanged(KeyNames.SYMBOLLIBRARY_LIST);
            }
          }

          this.notifyUnknownProperties(changedResource);
          return;
        }
    }
  } // Check if a changes event added or removed some symbols (groups contained directly under the maingroup) from a symbol library


  symbolLibraryListChanged(event) {
    if (!event.eventData.controlChanges) {
      return false;
    }

    for (let key in event.eventData.controlChanges.changes) {
      if (event.eventData.controlChanges.changes[key][Constants.ControlChanges.REMOVED] || event.eventData.controlChanges.changes[key][Constants.ControlChanges.ADDED] || event.eventData.controlChanges.changes[key][Constants.ControlChanges.NAME]) {
        if (event.eventData.controlChanges.changes[key].coreObjectID.controlID.length == 1) {
          return true;
        }
      }
    }

    return false;
  }

  updateThumbnail(resource, operationOrigin) {
    _balsamiq.default.existsClass(resource, Resource.Resource);

    if (resource.type === Constants.ResourceTypes.ASSET) {
      return;
    }

    let thumbnail = this.getThumbnailOfResource(resource.coreObjectID);

    if (thumbnail) {
      this.thumbnailQueue.scheduleThumbnailUpdate(thumbnail, resource.getRootControl(), operationOrigin);
    }
  }

  updateThumbnailForTrashedResources() {
    if (this.trashedThumbnailResourceRendered) return;
    this.trashedThumbnailResourceRendered = true;
    this.getTrashList().map(resource => this.updateThumbnail(resource));
  }
  /**
   * For every resource checks if it has a dependency with the changed resourceID, subResourceID
   * @param {string} changedResourceID
   * @param {string | undefined} origin
   * @param {string[] | undefined} [changedSubResourceIDs]
   */


  updateDependencies(changedResourceID, origin, changedSubResourceIDs) {
    _balsamiq.default.assertString(changedResourceID);

    let changedResource = this.getResourceID(changedResourceID, BarConstants.Branch.MasterBranchID);

    if (!changedResource) {
      return;
    }

    let changedResourceType = changedResource.getType();
    if (changedResourceType === Constants.ResourceTypes.MOCKUP) return; // nothing depends on mockups

    var callDependencyChanged = resource => {
      if (resource.dependsOnResource(changedResourceID)) {
        if (changedResourceType === Constants.ResourceTypes.ASSET) {
          // Update image controls immediately!
          resource.onDependencyChanged(changedResourceID, origin, changedSubResourceIDs);
        } else {
          this.context.host.timer(200, () => {
            resource.onDependencyChanged(changedResourceID, origin, changedSubResourceIDs);
          });
        }
      }
    };

    this._loopOnResources(this.symbolLibraries, callDependencyChanged);

    this._loopOnResources(this.mockups, callDependencyChanged);
  }

  updateResourceUsedBy() {
    this.resourceUsedBy = {};

    var collectDependencies = function (resource) {
      if (!resource.isTrashed()) {
        var deps = resource.getDependencies();
        deps.forEach(function (depResourceObject) {
          if (!this.resourceUsedBy[depResourceObject.resourceID]) {
            this.resourceUsedBy[depResourceObject.resourceID] = {};
          }

          this.resourceUsedBy[depResourceObject.resourceID][resource.coreObjectID.toString()] = resource.coreObjectID;
        }, this);
      }
    }.bind(this);

    this._loopOnResources(this.mockups, collectDependencies);

    this._loopOnResources(this.symbolLibraries, collectDependencies);
  }

  getResourceUsedBy(resourceID) {
    return this.resourceUsedBy[resourceID];
  }

  getSymbolDefinitionsUsedBy(resourceID, subResourceID) {
    const dependencyCoreObjectIDs = this.symbolMap.controlsDependencies.getAffectedByResourceChange(resourceID, subResourceID.toString());
    const controls = dependencyCoreObjectIDs.map(this.context.getCoreObject);
    const mockupsOrSymbols = controls.reduce((all, control) => {
      if (!control) {
        return all;
      }

      const mockup = control.getMockup();

      if (mockup.isTrashed()) {
        return all;
      }

      if (mockup.getType() === Constants.ResourceTypes.SYMBOLLIBRARY) {
        const symbolControlID = control.coreObjectID.controlID.slice(0, 1);

        if (mockup.symbolHasDirectDependency(symbolControlID, resourceID, subResourceID)) {
          const symbol = mockup.getControl(symbolControlID);
          all.add(symbol);
        }
      } else {
        all.add(mockup);
      }

      return all;
    }, new Set());
    return Array.from(mockupsOrSymbols);
  }

  getSymbolInstanceCoreObjectIDs(symbolCoreObjectID) {
    _balsamiq.default.assertClass(symbolCoreObjectID, CoreObject.CoreObjectID);

    _balsamiq.default.exists(symbolCoreObjectID.controlID);

    return this.symbolMap.controlsDependencies.getAffectedByResourceChange(symbolCoreObjectID.resourceID, symbolCoreObjectID.controlID.toString());
  }

  onBranchContentsChanged(event) {
    var branchID = event.senderCoreObjectID.branchID;

    if (!this.branches[branchID]) {
      // probably it's a new Branch not yet added to this.branches
      return;
    }

    _balsamiq.default.exists(this.branches[branchID]); // pass the property to all mockups and symbol libraries


    var setPropsFunction = function (propName, propValue) {
      var f = function (list) {
        this._loopOnResources(list, function (resource) {
          resource.setProperty(propName, propValue, event.eventData.operationOrigin);
        });
      }.bind(this);

      f(this.mockups);
      f(this.symbolLibraries);
    }.bind(this); // check if the event contains the given property,
    // then pass it to the mockups and notify the key value change


    var checkPropFunction = function (propName, keyName) {
      if (event.eventData.hasOwnProperty(propName)) {
        setPropsFunction(propName, event.eventData[propName]);
        this.notifyKeyValueChanged(keyName);
      }
    }.bind(this);

    if (branchID === MASTER) {
      if (event.eventData.hasOwnProperty(Branch.MasterAttributes.PROJECT_DESCRIPTION)) {
        this.notifyKeyValueChanged(KeyNames.DESCRIPTION);
      }

      checkPropFunction(Branch.MasterAttributes.SKIN_NAME, KeyNames.SKIN_NAME);
      checkPropFunction(Branch.MasterAttributes.FONT_SIZE, KeyNames.FONT_SIZE);
      checkPropFunction(Branch.MasterAttributes.FONT_FACE, KeyNames.FONT_FACE);
      checkPropFunction(Branch.MasterAttributes.SELECTION_COLOR, KeyNames.SELECTION_COLOR);
      checkPropFunction(Branch.MasterAttributes.LINK_COLOR, KeyNames.LINK_COLOR);
    }
  }

  onThumbnailContentsChanged() {//        this.context.host.log("thumbnail ready: " + event.senderCoreObjectID.thumbnailID + ", imageID: " + event.eventData.imageID);
  }

  onKeyValueChanged(event) {
    var keysToNotify = {};

    if (event.senderCoreObjectID.type === CoreObject.CoreObjectTypes.RESOURCE) {
      var r = this._findResource(event.senderCoreObjectID.branchID, event.senderCoreObjectID.resourceID);

      if (r) {
        // only resources already added to the project are interesting
        this.updateDependencies(event.senderCoreObjectID.resourceID);

        var listKeyName = this._findResourceList(event.senderCoreObjectID.resourceID).keyName;

        if (event.eventData.keyName === Mockup.KeyNames.PARENTID) {
          keysToNotify[KeyNames.MOCKUP_LIST] = true;
        } else if (event.eventData.keyName === Mockup.KeyNames.ORDER) {
          keysToNotify[KeyNames.MOCKUP_LIST] = true;
        } else if (event.eventData.keyName === _BmprConstants.ResourceAttributes.Trashed) {
          // the resource has been trashed or restored from trash: notify that the two involved lists have been changed
          keysToNotify[listKeyName] = true;
          keysToNotify[KeyNames.TRASH_LIST] = true;
        } else if (event.eventData.keyName === _BmprConstants.ResourceAttributes.Name && (listKeyName === KeyNames.ASSET_LIST || listKeyName === KeyNames.SYMBOLLIBRARY_LIST)) {
          // Asset and Symbol Libraries are sorted by name
          keysToNotify[listKeyName] = true;
        }
      }
    }

    for (let keyName in keysToNotify) {
      this.notifyKeyValueChanged(keyName);
    }
  }

  onConnectedUsersKeyValueChanged(event) {
    if (event.eventData.keyName === ConnectedUsers.KeyNames.MY_USER) {
      if (this.currentUserIsPresentInArchive()) {
        this.updateMyUserIfNeeded();
      }

      this.notifyKeyValueChanged(KeyNames.USERS);
    }
  }

  addUserLeftMessage(user) {
    let message = new Message.ChatMessage(CoreObject.createNewUUID(), Message.MessageTypes.CHAT, "", // No user or displayName. It's an anonymous message from the chat
    "", Date.now(), `User ${user.displayName} left`);
    this.addMessage(message);
  }

  onArchiveStateChanged() {
    this.notifyKeyValueChanged(KeyNames.USER_ROLE);
  }

  updateUsers(users) {
    // redo list from scratch
    this.users = [];
    users.forEach(user => {
      this.users[user.userID] = this._makeUser(user.userID, user.attributes);
    });
    this.notifyKeyValueChanged(KeyNames.USERS);
  }

  getLoggedUserId() {
    if (!this.connectedUsers || !this.connectedUsers.getMyUserInfo()) {
      return null;
    }

    const userInfo = this.connectedUsers.getMyUserInfo();
    return userInfo.platformKind === "cloud" ? userInfo.email : userInfo.name;
  }

  updateMentionableUsers(users) {
    let replace = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;
    const loggedUserId = this.getLoggedUserId();

    if (replace) {
      this.mentionableUsers = {};
    }

    for (const user of users) {
      if (!user.fullName) {
        user.fullName = user.userName;
      }
    }

    users.sort((a, b) => a.fullName.localeCompare(b.fullName)).forEach((user, index) => {
      if (index < users.length - 1) {
        const nextUser = users[index + 1];

        if (user.fullName === nextUser.fullName) {
          user.identifier = user.fullName + " (" + user.userName + ")";
          nextUser.identifier = nextUser.fullName + " (" + nextUser.userName + ")";
        }
      }

      if (!user.identifier) {
        user.identifier = user.fullName;
      }

      user.identifier = user.identifier.replace("]", /%5D/);
      user.identifier = user.identifier.replace("[", /%5B/);
      user.isLoggedUser = user.userName === loggedUserId;
      this.mentionableUsers[user.id] = user;
    });
    this.notifyKeyValueChanged(KeyNames.MENTIONABLE_USERS);
  }

  currentUserIsPresentInArchive() {
    let userInfo = this.getConnectedUsers().getMyUserInfo();

    if (!userInfo) {
      return false;
    }

    return this.users[userInfo.ID] != undefined;
  }

  updateMyUserIfNeeded() {
    if (this.userRole < Roles.COMMENTER) {
      // no changes allowed, even if needed
      return;
    }

    _balsamiq.default.assert(this.currentUserIsPresentInArchive());

    let myUserInfo = this.getConnectedUsers().getMyUserInfo();
    let user_from_bas = this.users[myUserInfo.ID];
    let attributes = {
      userName: myUserInfo.name,
      displayName: myUserInfo.displayName,
      email: myUserInfo.email ? myUserInfo.email : "",
      anonymous: myUserInfo.anonymous ? true : false,
      inProject: _BmprConstants.UserAttributes.InProject in myUserInfo ? myUserInfo.inProject : true
    };

    if (this.hasMyUserInfoIDAndDisplayName() && user_from_bas.attributesDiffer(attributes)) {
      this.users[myUserInfo.ID].setAttributes(attributes);
      this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.MYUSER_UPDATED, attributes));
      this.notifyKeyValueChanged(KeyNames.USERS);
    }
  }

  createMyUserIfNotPresent() {
    if (this.userRole < Roles.COMMENTER) {
      // no changes allowed, even if needed
      return;
    }

    if (this.currentUserIsPresentInArchive()) {
      this.updateMyUserIfNeeded();
      return true;
    }

    var myUserInfo = this.getConnectedUsers().getMyUserInfo();

    if (!myUserInfo) {
      return false;
    }

    var attributes = {
      userName: myUserInfo.name,
      displayName: myUserInfo.displayName,
      email: myUserInfo.email,
      anonymous: myUserInfo.anonymous ? true : false,
      inProject: _BmprConstants.UserAttributes.InProject in myUserInfo ? myUserInfo.inProject : true
    };
    this.users[myUserInfo.ID] = this._makeUser(myUserInfo.ID, attributes);
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.MYUSER_CREATED, attributes));
    this.notifyKeyValueChanged(KeyNames.USERS);
    return true;
  }

  onThumbnailQueueHasArchiveOpsChanged() {
    this.updateHasComingUserEvents();
  }

  updateHasComingUserEvents() {
    let p = this.pendingAssetIDs.length > 0 || this.thumbnailQueue.hasComingUserEvents();

    if (p !== this.hasComingUserEvents) {
      this.hasComingUserEvents = p;
      this.notifyKeyValueChanged(KeyNames.HAS_COMING_USER_EVENTS);
    }
  }

  addPendingAssetID(pendingAssetID) {
    let index = this.pendingAssetIDs.indexOf(pendingAssetID);

    _balsamiq.default.assert(index === -1, "Adding already existing pending asset ID");

    this.pendingAssetIDs.push(pendingAssetID);
    this.updateHasComingUserEvents();
  }

  removePendingAssetID(pendingAssetID) {
    let index = this.pendingAssetIDs.indexOf(pendingAssetID);

    _balsamiq.default.assert(index !== -1, "Removing unknown pending asset ID");

    this.pendingAssetIDs.splice(index, 1);
    this.updateHasComingUserEvents();
  }
  /**
   *
   * @param {string} kind
   * @param {string} src
   * @return {Resource | null}
   */


  getResourceByKindAndImportedFrom(kind, src) {
    _balsamiq.default.assertEnum(kind, Constants.ResourceTypes);

    _balsamiq.default.assertString(src);

    let resources;

    switch (kind) {
      case Constants.ResourceTypes.ASSET:
        resources = this.assets;
        break;

      case Constants.ResourceTypes.SYMBOLLIBRARY:
        resources = this.symbolLibraries;
        break;

      case Constants.ResourceTypes.MOCKUP:
        resources = this.mockups;
        break;
    }

    for (const resourceID in resources) {
      for (const branchID in resources[resourceID]) {
        if (resources[resourceID][branchID].getAttributes()[_BmprConstants.ResourceAttributes.ImportedFrom] === src) {
          return resources[resourceID][branchID];
        }
      }
    }

    return null;
  }
  /**
   *
   * @param {string} kind
   * @param {string} name
   * @return {Resource | null}
   */


  getResourceByKindAndName(kind, name) {
    _balsamiq.default.assertEnum(kind, Constants.ResourceTypes);

    _balsamiq.default.assertString(name);

    let resources;

    switch (kind) {
      case Constants.ResourceTypes.ASSET:
        resources = this.assets;
        break;

      case Constants.ResourceTypes.SYMBOLLIBRARY:
        resources = this.symbolLibraries;
        break;

      case Constants.ResourceTypes.MOCKUP:
        resources = this.mockups;
        break;
    }

    for (const resourceID in resources) {
      for (const branchID in resources[resourceID]) {
        if (resources[resourceID][branchID].getAttributes()[_BmprConstants.ResourceAttributes.Name] && resources[resourceID][branchID].getAttributes()[_BmprConstants.ResourceAttributes.Name].toLowerCase() === name.toLowerCase()) {
          return resources[resourceID][branchID];
        }
      }
    }

    return null;
  } // Private methods
  // ---------------


  _makeBranch(branchID, attributes) {
    return new Branch.Branch(this.context, CoreObject.makeBranchCoreObjectID(this.coreObjectID, branchID), attributes);
  }

  _addBranch(branch) {
    _balsamiq.default.existsClass(branch, Branch.Branch);

    _balsamiq.default.assert(!this.branches[branch.coreObjectID.branchID], "a branch with id='" + branch.coreObjectID.branchID + "' already exists");

    this.branches[branch.coreObjectID.branchID] = branch;
  }

  _deleteBranch(branchID) {
    _balsamiq.default.assertString(branchID);

    _balsamiq.default.exists(this.branches[branchID]);

    var branch = this.branches[branchID];
    delete this.branches[branchID];
    branch.cleanUp();
    this.notifyKeyValueChanged(KeyNames.BRANCH_LIST);
  }

  _makeUser(userID, attributes) {
    _balsamiq.default.assertString(userID, "User creation: userID must be a string.");

    _balsamiq.default.exists(attributes, "User creation: attributes object must exist.");

    var coreObjectID = User.makeUserCoreObjectID(this.coreObjectID, userID);
    var user = new User.User(this.context, coreObjectID, attributes);
    return user;
  }

  _makeComment(commentID, branchID, resourceID, userID, attributes, data) {
    _balsamiq.default.assertString(commentID, "Comment creation: userID must be a string.");

    _balsamiq.default.exists(attributes, "Comment creation: attributes object must exist.");

    _balsamiq.default.exists(this.branches[branchID]);

    let resource = this._findResource(branchID, resourceID);

    _balsamiq.default.exists(resource);

    _balsamiq.default.assertObject(data);

    var coreObjectID = Comment.makeCommentCoreObjectID(resource.coreObjectID, commentID);
    var comment = new Comment.Comment(this.context, coreObjectID, attributes, userID, data);
    return comment;
  }

  _makeResource(branchID, resourceID, attributes, data) {
    _balsamiq.default.assertString(branchID, "Making a resource: branchID must be a string.");

    _balsamiq.default.assertString(resourceID, "Making a resource: resourceID must be a string.");

    _balsamiq.default.exists(attributes, "Making a resource: attributes object must exist."); //Balsamiq.assertString(data, "Making a resource: data must be a string." );


    var res = null;
    const coreObjectID = CoreObject.makeResourceCoreObjectID(this.coreObjectID, branchID, resourceID);

    const resourceContext = _balsamiq.default.extend({
      getUserRole: this._getUserRole,
      getCoreObject: this.context.getCoreObject
    }, this.context);

    const mockupContext = _balsamiq.default.extend({
      getDependency: this._getMockupDependency,
      getMockupIDByName: this._getMockupIDByName,
      getUserRole: this._getUserRole,
      canAddControlToGroupTriggerALoop: this.canAddControlToGroupTriggerALoop,
      setDependencies: this._setSymbolDependencies,
      hasSymbolDefinitionALoop: this.symbolMap.hasSymbolDefinitionALoop
    }, resourceContext);

    const mockupProperties = {
      fontSize: this.getDefaultFontSize(),
      fontFace: this.getFontFace(),
      selectionColor: this.getSelectionColor(),
      skinName: this.getSkinName(),
      linkColor: this.getLinkColor()
    };
    const migratedAttributes = this.removeNonMasterAttributes(branchID, attributes);

    switch (attributes.kind) {
      case _BmprConstants.ResourceKinds.Mockup:
        res = new Mockup.Mockup(mockupContext, coreObjectID, mockupProperties, migratedAttributes, data);
        break;

      case _BmprConstants.ResourceKinds.Asset:
      case _BmprConstants.ResourceKinds.OtherAsset:
        // keep checking the mime type for Bmpr 1.2 format files
        if (attributes.extension && MimeTypes.isImageExtension(attributes.extension) || MimeTypes.isImageMimeType(attributes.mimeType)) {
          res = new Image.Image(resourceContext, coreObjectID, migratedAttributes, data);
        } else {
          res = new Asset.Asset(resourceContext, coreObjectID, migratedAttributes, data);
        }

        break;

      case _BmprConstants.ResourceKinds.SymbolLibrary:
        res = new SymbolLibrary.SymbolLibrary(mockupContext, coreObjectID, mockupProperties, migratedAttributes, data);
        break;
    }

    _balsamiq.default.exists(res, "could not create a Resource of kind: " + attributes.kind);

    return res;
  }

  _addResource(resource) {
    _balsamiq.default.exists(resource);

    _balsamiq.default.assertClass(resource, Resource.Resource);

    const {
      resourceID,
      branchID
    } = resource.coreObjectID;
    const changedKeyNames = [];

    switch (resource.type) {
      case Constants.ResourceTypes.MOCKUP:
        if (!this.mockups[resourceID]) {
          this.mockups[resourceID] = {};
        }

        _balsamiq.default.assert(!this.mockups[resourceID][branchID], "the Wireframe with resourceID=" + resourceID + ", branchID=" + branchID + " is already present in the Project");

        this.mockups[resourceID][branchID] = resource;
        changedKeyNames.push(KeyNames.MOCKUP_LIST);
        break;

      case Constants.ResourceTypes.ASSET:
        _balsamiq.default.assert(!this.assets[resourceID], "the Asset with resourceID=" + resourceID + " is already present in the Project");

        _balsamiq.default.assert(branchID === MASTER);

        this.assets[resourceID] = {};
        this.assets[resourceID][branchID] = resource;
        changedKeyNames.push(KeyNames.ASSET_LIST);
        break;

      case Constants.ResourceTypes.SYMBOLLIBRARY:
        _balsamiq.default.assert(!this.symbolLibraries[resourceID], "the Symbol Library with resourceID=" + resourceID + " is already present in the Project");

        _balsamiq.default.assert(branchID === MASTER);

        this.symbolLibraries[resourceID] = {};
        this.symbolLibraries[resourceID][branchID] = resource;
        changedKeyNames.push(KeyNames.SYMBOLLIBRARY_LIST);
        break;
    }

    _balsamiq.default.assert(changedKeyNames.length >= 1, "could not add a Resource to the project");

    if (resource.isTrashed()) {
      changedKeyNames.push(KeyNames.TRASH_LIST);
    }

    return changedKeyNames;
  }

  _findResource(branchID, resourceID) {
    _balsamiq.default.assertString(branchID);

    _balsamiq.default.assertString(resourceID);

    if (this.mockups[resourceID] && this.mockups[resourceID][branchID]) {
      return this.mockups[resourceID][branchID];
    } else if (this.assets[resourceID]) {
      _balsamiq.default.assert(branchID === MASTER);

      return this.assets[resourceID][MASTER];
    } else if (this.symbolLibraries[resourceID]) {
      _balsamiq.default.assert(branchID === MASTER);

      return this.symbolLibraries[resourceID][MASTER];
    }

    return null;
  }

  _findMockupIDByName(mockupName) {
    _balsamiq.default.assertString(mockupName);

    for (var resourceID in this.mockups) {
      if (this.mockups[resourceID][MASTER]) {
        if (this.mockups[resourceID][MASTER].getName() == mockupName) {
          return resourceID;
        }
      }
    }

    return null;
  }

  _findResourceList(resourceID) {
    _balsamiq.default.assertString(resourceID);

    var list = null;
    var keyName = null;

    if (this.mockups[resourceID]) {
      list = this.mockups;
      keyName = KeyNames.MOCKUP_LIST;
    } else if (this.assets[resourceID]) {
      list = this.assets;
      keyName = KeyNames.ASSET_LIST;
    } else if (this.symbolLibraries[resourceID]) {
      list = this.symbolLibraries;
      keyName = KeyNames.SYMBOLLIBRARY_LIST;
    }

    return list ? {
      list: list,
      keyName: keyName
    } : null;
  }

  resourceBranches(resourceID) {
    _balsamiq.default.assertString(resourceID);

    return Object.keys(this._findResourceList(resourceID).list[resourceID]);
  }

  _deleteResource(branchID, resourceID) {
    _balsamiq.default.assertString(branchID);

    _balsamiq.default.assertString(resourceID);

    var changedKeyNames = [];

    var resource = this._findResource(branchID, resourceID);

    _balsamiq.default.exists(resource);

    var listObj = this._findResourceList(resourceID);

    _balsamiq.default.exists(listObj);

    delete listObj.list[resourceID][branchID];

    if (_balsamiq.default.isEmpty(listObj.list[resourceID])) {
      delete listObj.list[resourceID];
    }

    var isTrashed = resource.isTrashed();
    resource.cleanUp();
    changedKeyNames.push(listObj.keyName);

    if (isTrashed) {
      changedKeyNames.push(KeyNames.TRASH_LIST);
    }

    _balsamiq.default.assert(changedKeyNames.length >= 1, "could not delete a Resource from the project");

    return changedKeyNames;
  }

  _addThumbnail(thumbnail) {
    _balsamiq.default.existsClass(thumbnail, Thumbnail.Thumbnail);

    _balsamiq.default.assert(!this.thumbnails[thumbnail.coreObjectID.thumbnailID], " a thumbnail with this ID already exists");

    this.thumbnails[thumbnail.coreObjectID.thumbnailID] = thumbnail;
  }

  _makeThumbnail(thumbnailID, attributes) {
    _balsamiq.default.assertString(thumbnailID);

    _balsamiq.default.assertString(attributes.resourceID);

    let branchID = attributes.branchID ? attributes.branchID : MASTER;
    var resourceCoreObjectID = CoreObject.makeResourceCoreObjectID(this.coreObjectID, branchID, attributes.resourceID);
    return new Thumbnail.Thumbnail(this.context, CoreObject.makeThumbnailCoreObjectID(this.coreObjectID, thumbnailID), resourceCoreObjectID);
  }

  _deleteThumbnail(thumbnailID) {
    _balsamiq.default.assertString(thumbnailID);

    const thumbnail = this.thumbnails[thumbnailID];

    if (thumbnail) {
      delete this.thumbnails[thumbnailID];
      this.thumbnailQueue.removeThumbnail(thumbnailID);
      thumbnail.cleanUp();
    }
  } // Context
  // -------


  onContextChange(params) {
    this.context.host.log("Project: onContextChange");

    var contextChangeFun = function (resource) {
      resource.onContextChange(params);
    };

    this._loopOnResources(this.assets, contextChangeFun);

    this._loopOnResources(this.symbolLibraries, contextChangeFun);

    this._loopOnResources(this.mockups, contextChangeFun);
  }

  broadcastMessage(type, data) {
    _balsamiq.default.assertString(type);

    _balsamiq.default.exists(data);

    if (!this.connectedUsers.getMyUserInfo()) {
      // can not broadcast messages when anonymous
      return;
    }

    var eventData = {
      id: CoreObject.createNewUUID(),
      user: this.connectedUsers.getMyUserInfo().name,
      type: type,
      data: data
    };
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.MESSAGE_TO_BROADCAST, eventData));
  }

  onBroadcastMessage(id, type, user, timestamp, data) {
    if (!type) {
      this.context.host.collectDevData("DEV_DATA: ON_BROADCAST_MESSAGE", `id: "${id}", type: "${type}", user: "${user}", data: "${_lodash.default.omit(data, ["text"])}"`);
      return;
    }

    this.context.host.log(`Received message of type ${type} with id ${id} from user ${user}`);

    _balsamiq.default.assertString(id, `id must be a string, but instead it has value "${id}" of type "${typeof id}, type: "${type}", user: "${user}", data: "${_lodash.default.omit(data, ["text"])}`);

    _balsamiq.default.assertString(user, `user must be a string, but instead it has value "${user}" of type "${typeof user}"`);

    _balsamiq.default.assertNumber(timestamp);

    _balsamiq.default.exists(data);

    switch (type) {
      case Message.MessageTypes.CHAT:
        {
          // let messageKnownUser = null;
          // if (this.connectedUsers.userList) {
          //     messageKnownUser = this.connectedUsers.userList.find((connectedUser) => {
          //         return connectedUser.userInfo.name === user;
          //     });
          // }
          // this.addMessage(
          //     new Message.ChatMessage(id, type, user, messageKnownUser ? messageKnownUser.userInfo.displayName : "", timestamp, data.text)
          // );
          break;
        }

      case Message.MessageTypes.TARGET_AND_SELECTION:
        this.connectedUsers.onUserTargetAndSelectionChanged(user, data.actualRole, data.resourceID, data.branchID, data.selectedControlIDs);
        break;
    }
  }

  createMessage(text) {
    // TODO: timestamp and user are provided by the BAS so we can avoid sending them
    // Possible changes:
    // 1. create message without timestamp
    // 2. send type/id/text only
    // 3. add timestamp on broadcast function return
    // 4. onArchiveEvent broadcast: fill timestamp and user before sending to model
    var message = new Message.ChatMessage(CoreObject.createNewUUID(), Message.MessageTypes.CHAT, this.connectedUsers.getMyUserInfo().name, this.connectedUsers.getMyUserInfo().displayName, Date.now(), text);
    this.addMessage(message);
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.MESSAGE_TO_BROADCAST, message.getData()));
  }

  sortMessages() {
    this.messages.sort((chatMessage1, chatMessage2) => {
      return chatMessage1.getTimestamp() - chatMessage2.getTimestamp();
    });
  }

  addMessage(chatMessage) {
    _balsamiq.default.assertClass(chatMessage, Message.ChatMessage);

    this.sortMessages();
    this.messages.push(chatMessage);
    this.notifyKeyValueChanged(KeyNames.MESSAGES);
  }

  findMessageByID(id) {
    let messages = this.messages.filter(message => message.id === id);

    if (messages.length > 0) {
      return messages[0];
    }

    return null;
  }

  notifyMessageError(messageID, error) {
    let message = this.findMessageByID(messageID);

    if (message) {
      message.setError(error);
    }
  }

  updateChatMessage(data) {
    let message = this.findMessageByID(data.id);

    if (message) {
      message.update(data);
      this.notifyKeyValueChanged(KeyNames.MESSAGES);
    }
  }

  _loopOnResources(resources, functor) {
    let {
      includeTrashed = true
    } = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    for (var resourceID in resources) {
      for (var branchID in resources[resourceID]) {
        if (includeTrashed || !resources[resourceID][branchID].isTrashed()) {
          functor(resources[resourceID][branchID]);
        }
      }
    }
  }

  _loopOnResourceIDs(resources, functor) {
    for (var resourceID in resources) {
      functor(resourceID);
    }
  }
  /**
   *
   * @param {CoreObjectID} symbolCoreObjectID
   * @returns {CoreObjectID[]}
   */


  /**
   *
   * @param {string} resourceID
   * @param {string=} subResourceID
   * @returns {CoreObjectID[]}
   */
  getAffectedByResourceChange(resourceID, subResourceID) {
    return this.symbolMap.controlsDependencies.getAffectedByResourceChange(resourceID, subResourceID);
  }

}

exports.Project = Project;

/***/ }),

/***/ "./js/core/model/resource.js":
/*!***********************************!*\
  !*** ./js/core/model/resource.js ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Resource = exports.KeyNames = exports.EventTypes = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var CoreObject = _interopRequireWildcard(__webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js"));

var Constants = _interopRequireWildcard(__webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts"));

var Events = _interopRequireWildcard(__webpack_require__(/*! core/events */ "./js/core/events.ts"));

var _BmprConstants = __webpack_require__(/*! @balsamiq/bmpr/lib/BmprConstants */ "./node_modules/@balsamiq/bmpr/lib/BmprConstants.js");

var _BalsamiqArchiveConstants = __webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchiveConstants */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/// <reference types="../../typings" />
const EventTypes = {
  RESOURCE_DEPENDENCIES_CHANGED: "resourceDependenciesChanged",
  // list of dependencies has changed
  RESOURCE_CONTENTS_CHANGED: "resourceContentsChanged",
  // the contents has changed
  RESOURCE_CONTENTS_PREVIEW: "resourceContentsPreview",
  // the contents may change in the next future, this is a preview
  RESOURCE_DATA_DIRTY: "resourceDataDirty",
  // new data are available for save
  RESOURCE_ATTRIBUTES_DIRTY: "resourceAttributesDirty" // new attributes are available for save

};
exports.EventTypes = EventTypes;
const KeyNames = {
  NAME: _BmprConstants.ResourceAttributes.Name,
  NOTES: _BmprConstants.ResourceAttributes.Notes,
  TRASHED: _BmprConstants.ResourceAttributes.Trashed,
  CREATION_DATE: _BmprConstants.ResourceAttributes.CreationDate
};
/**
 * @typedef {Object} ResourceData
 * @prop {string} resourceID
 * @prop {{ [key: string]: any }} attributes
 * @prop {string} data
 * @prop {string=} branchID
 * @prop {Comment[]=} comments
 */

/**
 * @typedef {{ resourceID: string; subResourceID: string | undefined }} Dependency
*/

exports.KeyNames = KeyNames;

class Resource extends CoreObject.CoreObject {
  constructor(context, coreObjectID, type) {
    super(context, coreObjectID);
    CoreObject.assertResourceType(coreObjectID);

    _balsamiq.default.assertEnum(type, Constants.ResourceTypes);

    _balsamiq.default.exists(context.getUserRole);
    /** @type {string} */


    this.type = type;
    /** @type {{ [key: string]: any }} */

    this.attributes = {};

    if (this.coreObjectID.branchID == _BalsamiqArchiveConstants.Branch.MasterBranchID) {
      this.attributes[_BmprConstants.ResourceAttributes.Name] = "";
      this.attributes[_BmprConstants.ResourceAttributes.Kind] = "";
      this.attributes[_BmprConstants.ResourceAttributes.Trashed] = false;
      this.attributes[_BmprConstants.ResourceAttributes.ImportedFrom] = "";
    }

    this.attributes[_BmprConstants.ResourceAttributes.CreationDate] = 0;
    this.attributes[_BmprConstants.ResourceAttributes.ModifiedBy] = "";
    this.attributes[_BmprConstants.ResourceAttributes.Notes] = "";
    /** @type {Dependency[]} */

    this.deps = [];
  }
  /** @returns {string} */


  getType() {
    return this.type;
  }

  dependsOnResource(resourceID) {
    return this.deps.find(resource => resource.resourceID === resourceID);
  }
  /**
   * @returns {Dependency[]} Array<resourceID>
   */


  getDependencies() {
    return this.deps;
  }
  /**
   * @param {Dependency[]} deps
   */


  setDependencies(deps) {
    this.deps = deps;
  }

  onDependencyChanged() {// subclass
  }

  onAttributesChanged(attributes) {
    // subclass
    this.setAttributes(attributes, Constants.OperationOrigin.ARCHIVE);
  }

  onDataChanged() {// subclass
  }
  /** @returns {string=} */


  getData() {
    // subclass
    return null;
  }
  /**
   * @returns {ResourceData}
   */


  toObject() {
    return {
      resourceID: this.coreObjectID.resourceID,
      branchID: this.coreObjectID.branchID,
      attributes: this.getAttributes(),
      data: this.getData()
    };
  } // Key-value
  // ---------


  getKeyNames() {
    return super.getKeyNames().concat(Object.values(KeyNames));
  }

  getKeyValue(keyName) {
    switch (keyName) {
      case KeyNames.NAME:
        return this.getName();

      case KeyNames.CREATION_DATE:
        return this.getCreationDate();

      case KeyNames.NOTES:
        return this.getNotes();

      case KeyNames.TRASHED:
        return this.isTrashed();
    }

    return super.getKeyValue(keyName);
  }

  setKeyValue(keyName, value) {
    switch (keyName) {
      case KeyNames.NOTES:
        this.setNotes(value);
        break;

      default:
        super.setKeyValue(keyName, value);
    }
  } // Attributes private methods
  // --------------------------
  // subclass if needed


  _checkAttribute(name, value) {
    if (name === _BmprConstants.ResourceAttributes.ThumbnailID || name === _BmprConstants.ResourceAttributes.MimeType) {
      return null;
    }

    if (name in this.attributes) {
      if (name === _BmprConstants.ResourceAttributes.Kind && value === _BmprConstants.ResourceKinds.OtherAsset) {
        return {
          name,
          value: _BmprConstants.ResourceKinds.Asset
        };
      }

      return {
        name,
        value
      };
    }

    _balsamiq.default.assert(false, "Unknown resource attribute: " + name);
  }

  setAttribute(name, value, operationOrigin) {
    _balsamiq.default.assertString(name);

    var checkedAttribute = this._checkAttribute(name, value);

    if (!checkedAttribute) {
      return;
    }

    name = checkedAttribute.name;
    value = checkedAttribute.value;

    if (operationOrigin && operationOrigin === Constants.OperationOrigin.USER) {
      _balsamiq.default.assert(this.context.getUserRole() >= Constants.UserRoles.EDITOR, "can not edit this Resource because the user role is: " + this.context.getUserRole());
    }

    if (this.attributes[name] !== value) {
      this.attributes[name] = value;

      if (this.getKeyNames().indexOf(name) != -1) {
        this.notifyKeyValueChanged(name);
      }

      if (operationOrigin && operationOrigin === Constants.OperationOrigin.USER) {
        this.notifyAttributesDirty();
      }
    }
  }

  setAttributes(attributes, operationOrigin) {
    for (const attributeName in attributes) {
      this.setAttribute(attributeName, attributes[attributeName], operationOrigin);
    }
  }

  setData(data, operationOrigin) {
    this.onDataChanged(data);

    if (operationOrigin === Constants.OperationOrigin.USER) {
      this.notifyDataDirty({
        data
      });
    }
  } // Getters - Attributes
  // --------------------

  /**
   *
   * @param {string} name
   * @returns {string=}
   */


  getAttribute(name) {
    if (name in this.attributes) {
      return this.attributes[name];
    }

    const coreSiteId = CoreObject.makeSiteCoreObjectID(this.coreObjectID.siteID);
    const coreProjectId = CoreObject.makeProjectCoreObjectID(coreSiteId, this.coreObjectID.projectID);
    const coreResourceID = CoreObject.makeResourceCoreObjectID(coreProjectId, _BalsamiqArchiveConstants.Branch.MasterBranchID, this.coreObjectID.resourceID);
    const masterObject = this.context.getCoreObject(coreResourceID);
    return masterObject && masterObject.getAttribute(name);
  }

  getAttributes() {
    return this.attributes;
  }

  getAttributesClone() {
    return Object.assign({}, this.attributes);
  }
  /** @returns {boolean} */


  isTrashed() {
    return this.getAttribute(_BmprConstants.ResourceAttributes.Trashed) === true;
  }
  /** @returns {string} */


  getName() {
    return this.getAttribute(_BmprConstants.ResourceAttributes.Name);
  }

  getFullName() {
    return this.getName();
  }

  getFileName() {
    // subclass
    // name for saving the resource to a file
    return this.getName();
  }

  getNotes() {
    return this.getAttribute(_BmprConstants.ResourceAttributes.Notes);
  }

  getCreationDate() {
    return this.getAttribute(_BmprConstants.ResourceAttributes.CreationDate);
  }

  getKind() {
    return this.getAttribute(_BmprConstants.ResourceAttributes.Kind);
  } // Representation
  // --------------


  getPreviewData() {
    // subclass
    // return { imageID } or { iconID, iconText }
    return null;
  } // Setters - Attributes
  // --------------------


  setName(value) {
    _balsamiq.default.assertString(value);

    this.setAttribute(_BmprConstants.ResourceAttributes.Name, value, Constants.OperationOrigin.USER);
  }

  setNotes(value) {
    if (value !== null) {
      _balsamiq.default.assertString(value);

      _balsamiq.default.assert(value.length <= Constants.MaxNoteLength, `The note length ${value.length} is bigger than the max allowed ${Constants.MaxNoteLength}`);
    }

    this.setAttribute(_BmprConstants.ResourceAttributes.Notes, value, Constants.OperationOrigin.USER);
  }

  setTrashed(value) {
    _balsamiq.default.assertBool(value);

    this.setAttribute(_BmprConstants.ResourceAttributes.Trashed, value, Constants.OperationOrigin.USER);
  } // Notifications
  // -------------


  notifyContentsChanged(eventData, origin, preview) {
    if (!eventData) {
      eventData = {};
    }

    if (origin) {
      _balsamiq.default.assertEnum(origin, Constants.OperationOrigin);
    }

    eventData.operationOrigin = origin ? origin : Constants.OperationOrigin.ARCHIVE;
    let eventType = preview === true ? EventTypes.RESOURCE_CONTENTS_PREVIEW : EventTypes.RESOURCE_CONTENTS_CHANGED;
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, eventType, eventData));
  }

  notifyDataDirty(eventData) {
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.RESOURCE_DATA_DIRTY, eventData));
  }

  notifyAttributesDirty() {
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.RESOURCE_ATTRIBUTES_DIRTY, {
      attributes: this.getAttributes()
    }));
  } // Short ID
  // --------


  getShortID() {
    var s = {
      shortResourceID: this.coreObjectID.resourceID.substring(0, 4)
    };

    if (this.coreObjectID.branchID != _BalsamiqArchiveConstants.Branch.MasterBranchID) {
      s.shortBranchID = this.coreObjectID.branchID.substring(0, 4);
    }

    return s;
  }

  matchesShortID(shortID) {
    _balsamiq.default.exists(shortID);

    _balsamiq.default.assertString(shortID.shortResourceID);

    if (shortID.shortBranchID) {
      _balsamiq.default.assertString(shortID.shortBranchID);
    }

    if (this.coreObjectID.resourceID.indexOf(shortID.shortResourceID) == 0) {
      if (!shortID.shortBranchID && this.coreObjectID.branchID == _BalsamiqArchiveConstants.Branch.MasterBranchID || this.coreObjectID.branchID.indexOf(shortID.shortBranchID) == 0) {
        return true;
      }
    }

    return false;
  }

}

exports.Resource = Resource;

/***/ }),

/***/ "./js/core/model/site.js":
/*!*******************************!*\
  !*** ./js/core/model/site.js ***!
  \*******************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = exports.Site = void 0;

var _coreobject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

class Site extends _coreobject.CoreObject {
  constructor(context, coreObjectID) {
    super(context, coreObjectID);
    (0, _coreobject.assertSiteType)(coreObjectID);
  }

}

exports.Site = Site;
var _default = Site;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/model/symbollibrary.js":
/*!****************************************!*\
  !*** ./js/core/model/symbollibrary.js ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = exports.SymbolLibrary = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Mockup = _interopRequireWildcard(__webpack_require__(/*! core/model/mockup */ "./js/core/model/mockup.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Constants = _interopRequireWildcard(__webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts"));

var _group = _interopRequireDefault(__webpack_require__(/*! core/controls/group */ "./js/core/controls/group.js"));

var _symbol = _interopRequireDefault(__webpack_require__(/*! core/controls/symbol */ "./js/core/controls/symbol.js"));

var _json = __webpack_require__(/*! core/serialization/json */ "./js/core/serialization/json.js");

var BmprConstants = _interopRequireWildcard(__webpack_require__(/*! @balsamiq/bmpr/lib/BmprConstants */ "./node_modules/@balsamiq/bmpr/lib/BmprConstants.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/** @typedef { import("core/controls/controls").Control } Control */
class SymbolLibrary extends Mockup.Mockup {
  constructor(context, coreObjectID, properties, attributes, data) {
    super(context, coreObjectID, properties, attributes, data, Constants.ResourceTypes.SYMBOLLIBRARY);
    this.attributes[BmprConstants.ResourceAttributes.Kind] = BmprConstants.ResourceKinds.SymbolLibrary;

    if (!attributes[BmprConstants.ResourceAttributes.Name]) {
      this.attributes[BmprConstants.ResourceAttributes.Name] = "New Symbol Library";
    }

    this.symbolDefinitions = {};
    this.updateSymbolDefinitions();
  }

  _load(json) {
    super._load(json);

    this.updateSymbolDefinitions();
  }

  updateSymbolDefinitions() {
    const res = {};
    this.mainGroupControl.children.forEach(function (control) {
      if (control.controlTypeID === _group.default.CONTROL_TYPE_ID) {
        const nameProp = control.getActualProperty("controlName") || Constants.UnnamedSymbol;
        const controlID = parseInt(control.coreObjectID.controlID[0], 10);
        res[controlID] = nameProp;
      }
    }, this);

    if (_balsamiq.default.isEmpty(res)) {
      res[_types.default.NO_SYMBOL_DEFINITION] = -1;
    }

    this.symbolDefinitions = res;
  }

  getSymbolDefinition(controlID) {
    if (!_balsamiq.default.has(this.symbolDefinitions, controlID)) {
      return null;
    }

    return controlID === _types.default.NO_SYMBOL_DEFINITION ? this.mainGroupControl : this.mainGroupControl.children.find(control => control.controlID === controlID);
  }

  getSymbolDefinitions() {
    return this.symbolDefinitions;
  }
  /**
   * @returns {string[]}
   */


  getSymbolNames() {
    const names = [];

    for (const controlID in this.symbolDefinitions) {
      if (controlID !== _types.default.NO_SYMBOL_DEFINITION) {
        names.push(this.symbolDefinitions[controlID]);
      }
    }

    return names;
  } // Attention please
  // Overriding this method is not the best solution, but for the moment it works.
  // A better solution would be to make a preNotifyContentsChange method for subclasses, or
  // linking the update of symbol definitions to the performEdit method


  notifyContentsChanged(eventData, origin, preview) {
    if (!preview) {
      this.updateSymbolDefinitions();
    }

    super.notifyContentsChanged(eventData, origin, preview);
  }
  /**
   * @override
   * @param {number[]} [controlIDs=[]]
   * @param {{ ensureControlName: boolean }=} options
   * @returns {string?}
   */


  getData() {
    let controlIDs = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];
    let {
      ensureControlName = false
    } = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    const data = this.export(null, controlIDs, false, null, false);

    if (ensureControlName) {
      const dataWithControlName = this.ensureSymbolsHaveControlNameProperty(data);
      return JSON.stringify(dataWithControlName, null, 0);
    } else {
      return JSON.stringify(data, null, 0);
    }
  }
  /**
   * @param {object} data
   * @returns {object}
   */


  ensureSymbolsHaveControlNameProperty(data) {
    data.mockup.controls.control.forEach(symbol => {
      if (!_balsamiq.default.get(symbol, "properties.controlName")) {
        _balsamiq.default.set(symbol, "properties.controlName", Constants.UnnamedSymbol);
      }
    });
    return data;
  }
  /**
   * @override
   * @param {number[]} [controlIDs=[]]
   * @param {unknown=} options
   * @returns {ResourceData}
   */


  toObject() {
    let controlIDs = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];
    let options = arguments.length > 1 ? arguments[1] : undefined;
    return {
      resourceID: this.coreObjectID.resourceID,
      branchID: this.coreObjectID.branchID,
      attributes: this.getAttributes(),
      data: this.getData(controlIDs, options)
    };
  }
  /**
   *
   * @param {string} name
   * @returns {Control | undefined}
   */


  getSymbolByName(name) {
    return Object.values(this.mainGroupControl.children).find(control => {
      if (control.controlTypeID === _group.default.CONTROL_TYPE_ID) {
        const nameProp = control.getActualProperty("controlName") || Constants.UnnamedSymbol;

        if (nameProp === name) {
          const controlID = parseInt(control.coreObjectID.controlID[0], 10);
          return this.getSymbolDefinition(controlID);
        }
      }
    });
  }

  updateSymbolDefinitionsWithDependencies() {
    if (this.hasSymbolInstances()) {
      const symbolsToUpdate = Object.keys(this.controlsDependencies.controlsDependencies).map(key => {
        return this.getControl(this.controlsDependencies.controlsDependencies[key].coreID.controlID);
      }).filter(c => c.controlTypeID === _symbol.default.CONTROL_TYPE_ID);

      if (symbolsToUpdate.length > 0) {
        this.performEdit(() => {
          symbolsToUpdate.forEach(s => s.updateAll());
        }, Constants.OperationOrigin.ARCHIVE);
      }
    }
  }
  /**
   * Checks wether a symbol in the symbol library directly contains a specific resource and subResource ids
   *
   * @param {number[]} controlID
   * @param {string} resourceID The ResourceId of the dependency to check
   * @param {string} subResourceID The SubResourceID of the dependency to check
   * @returns {boolean}
   */


  symbolHasDirectDependency(controlID, resourceID, subResourceID) {
    const mockupJSON = this.export(controlID, [], false, null, false);
    return (0, _json.symbolHasDirectDependency)(mockupJSON.mockup.controls.control, resourceID, subResourceID);
  }

}

exports.SymbolLibrary = SymbolLibrary;
var _default = SymbolLibrary;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/model/thumbnailqueue.js":
/*!*****************************************!*\
  !*** ./js/core/model/thumbnailqueue.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.makeThumbnailQueueCoreObjectID = exports.ThumbnailQueueType = exports.ThumbnailQueue = exports.EventTypes = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _thumbnail = __webpack_require__(/*! core/model/thumbnail */ "./js/core/model/thumbnail.ts");

var _controls = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var _constants = __webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts");

var _coreobject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var _events = __webpack_require__(/*! core/events */ "./js/core/events.ts");

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var EventTypes = {
  QUEUE_EMPTY: "queue_empty",
  HAS_COMING_USER_EVENTS: "HAS_COMING_USER_EVENTS"
};
exports.EventTypes = EventTypes;

class ThumbnailQueue extends _coreobject.CoreObject {
  constructor(context, coreObjectID) {
    super(context, coreObjectID);
    this.queue = []; // thumbnailIDs of thumbnails yet to start

    this.data = {}; // thumbnailID -> { thumbnail: Thumbnail, resource: Resource, dispatchDirty: bool }

    this.updatingThumbnails = {};
    this.comingUserEventsCount = 0;
    this._isPaused = false;
  }

  addComingUserEvent() {
    this.comingUserEventsCount += 1;

    if (this.comingUserEventsCount == 1) {
      this.context.eventCenter.fireEvent((0, _events.newCoreEvent)(this.coreObjectID, EventTypes.HAS_COMING_USER_EVENTS, {
        value: true
      }));
    }
  }

  removeComingUserEvent() {
    _balsamiq.default.assert(this.comingUserEventsCount > 0);

    this.comingUserEventsCount -= 1;

    if (this.comingUserEventsCount === 0) {
      this.context.eventCenter.fireEvent((0, _events.newCoreEvent)(this.coreObjectID, EventTypes.HAS_COMING_USER_EVENTS, {
        value: false
      }));
    }
  }

  removeAllComingUserEvents() {
    if (this.comingUserEventsCount > 0) {
      this.comingUserEventsCount = 0;
      this.context.eventCenter.fireEvent((0, _events.newCoreEvent)(this.coreObjectID, EventTypes.HAS_COMING_USER_EVENTS, {
        value: false
      }));
    }
  }

  scheduleThumbnailUpdate(thumbnail, resourceRootControl, operationOrigin) {
    _balsamiq.default.existsClass(thumbnail, _thumbnail.Thumbnail);

    _balsamiq.default.existsClass(resourceRootControl, _controls.Control);

    if (operationOrigin) {
      _balsamiq.default.assertEnum(operationOrigin, _constants.OperationOrigin);
    }

    var thumbnailID = thumbnail.coreObjectID.thumbnailID; // If no data is present for thumbnailID we should always add it and schedule a new update
    // If data is present it means that an update has been scheduled but has not started yet,
    //  so we may skip scheduling a new update
    // If the update has already started for the same thumbnailID we should schedule a new update anyway
    //  because the running update is rendering an old displaylist

    if (!this.data[thumbnailID]) {
      if (operationOrigin === _constants.OperationOrigin.USER) {
        this.context.host.log("===> scheduleThumbnailUpdate " + thumbnailID);
        this.addComingUserEvent();
      }

      this.data[thumbnailID] = {
        thumbnail: thumbnail,
        resourceRootControl: resourceRootControl,
        dispatchDirty: operationOrigin === _constants.OperationOrigin.USER,
        operationOrigin: operationOrigin
      }; // Enqueue and notify of incoming future UserEvent

      this.queue.push(thumbnailID); // Always advance the queue to fill the pipeline

      this.advanceQueue();
    } else {
      // If data for future update is present, update the dispatchDirty flag
      if (operationOrigin === _constants.OperationOrigin.USER) {
        this.data[thumbnailID].dispatchDirty = true;
      }
    }
  }

  removeThumbnail(thumbnailID) {
    _balsamiq.default.assertString(thumbnailID);

    if (this.data[thumbnailID] && this.data[thumbnailID].operationOrigin === _constants.OperationOrigin.USER) {
      this.removeComingUserEvent();
    }

    delete this.data[thumbnailID];
    this.queue = this.queue.filter(tID => tID !== thumbnailID);
  }
  /**
   *
   * @param {boolean} shouldPause
   */


  pause(shouldPause) {
    _balsamiq.default.assertBool(shouldPause);

    this._isPaused = shouldPause;

    if (!shouldPause) {
      this.advanceQueue();
    }
  } // Private methods
  // ---------------


  advanceQueue() {
    if (this._isPaused || this.queue.length == 0) {
      return;
    }

    let startNextThumbnail = () => {
      _balsamiq.default.assert(this.queue.length > 0); // Find next valid index: I must find a thumbnailID that is not currently being processed inside this.updatingThumbnails
      // We cannot process the same thumbnailID in parallel multiple times!


      let index = 0;
      let thumbnailID = null;

      while (index < this.queue.length && !thumbnailID) {
        let id = this.queue[index];

        if (!this.updatingThumbnails.hasOwnProperty(id)) {
          thumbnailID = id;
        } else {
          index += 1;
        }
      }

      if (thumbnailID) {
        this.queue.splice(index, 1);
        let updatingThumbnail = this.data[thumbnailID];
        this.updatingThumbnails[thumbnailID] = updatingThumbnail;
        delete this.data[thumbnailID];
        this.context.host.asyncCallback(() => {
          if (this._isCleanedUp) {
            return;
          }

          updatingThumbnail.thumbnail.updateThumbnail(updatingThumbnail.resourceRootControl, updatingThumbnail.dispatchDirty, () => {
            if (this._isCleanedUp) {
              return;
            }

            this.onThumbnailUpdated(thumbnailID);
          });
        });
      }
    };

    let oldLen, len, numUpdating;
    let parallelNum = this.context.host.Images.parallelProcessingCapability();

    if (parallelNum == 0) {
      parallelNum = Number.MAX_SAFE_INTEGER;
    } // unlimited parallel processing!


    do {
      oldLen = this.queue.length;
      numUpdating = Object.keys(this.updatingThumbnails).length;

      if (numUpdating < parallelNum) {
        startNextThumbnail();
      }

      len = this.queue.length;
    } while (len > 0 && len != oldLen);
  }

  onThumbnailUpdated(thumbnailID) {
    if (this.updatingThumbnails[thumbnailID] && this.updatingThumbnails[thumbnailID].operationOrigin === _constants.OperationOrigin.USER) {
      this.removeComingUserEvent();
    }

    delete this.updatingThumbnails[thumbnailID];

    if (this.isEmpty()) {
      this.context.eventCenter.fireEvent((0, _events.newCoreEvent)(this.coreObjectID, EventTypes.QUEUE_EMPTY));
    }

    this.advanceQueue();
  }

  hasComingUserEvents() {
    return this.comingUserEventsCount > 0;
  }

  isEmpty() {
    return this.queue.length === 0 && _balsamiq.default.isEmpty(this.updatingThumbnails);
  }

  cleanUp() {
    super.cleanUp();
    this.data.updatingThumbnails = {};
    this.updatingThumbnails = {};
    this.queue = [];
    this.data = {};
    this.removeAllComingUserEvents();
  }

}

exports.ThumbnailQueue = ThumbnailQueue;
var currentThumbnailQueueID = 0;

var makeThumbnailQueueCoreObjectID = function (projectCoreObjectID) {
  (0, _coreobject.assertProjectType)(projectCoreObjectID);
  currentThumbnailQueueID++;
  return new _coreobject.CoreObjectID(ThumbnailQueueType, {
    siteID: projectCoreObjectID.siteID,
    projectID: projectCoreObjectID.projectID,
    thumbnailQueueID: currentThumbnailQueueID.toString()
  });
};

exports.makeThumbnailQueueCoreObjectID = makeThumbnailQueueCoreObjectID;
const ThumbnailQueueType = "ThumbnailQueue";
exports.ThumbnailQueueType = ThumbnailQueueType;

/***/ }),

/***/ "./js/core/model/user.js":
/*!*******************************!*\
  !*** ./js/core/model/user.js ***!
  \*******************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.makeUserCoreObjectID = exports.isValidEmail = exports.assertUserType = exports.User = exports.CoreObjectType = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var CoreObject = _interopRequireWildcard(__webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js"));

var _BmprConstants = __webpack_require__(/*! @balsamiq/bmpr/lib/BmprConstants */ "./node_modules/@balsamiq/bmpr/lib/BmprConstants.js");

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// https://stackoverflow.com/a/1373724
const EmailRegEx = /^[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?$/i;

const isValidEmail = email => _balsamiq.default.isString(email) && EmailRegEx.test(email.trim()); // expanding allowed coreobject types


exports.isValidEmail = isValidEmail;
const CoreObjectType = "User";
exports.CoreObjectType = CoreObjectType;
CoreObject.CoreObjectTypes["USER"] = CoreObjectType;

var assertUserType = function (coreObjectID) {
  CoreObject.assertType(coreObjectID, CoreObjectType);
};

exports.assertUserType = assertUserType;

class User extends CoreObject.CoreObject {
  constructor(context, coreObjectID, attributes) {
    super(context, coreObjectID);
    assertUserType(coreObjectID);

    _balsamiq.default.assertString(attributes.displayName);

    _balsamiq.default.assertString(attributes.userName);
    /** @type { import("@balsamiq/bmpr/lib/BmprConstants").UserInfo } */


    this.attributes = {};
    this.setAttributes(attributes);
  }

  getDisplayName() {
    return this.attributes.displayName;
  }

  getEmail() {
    return this.attributes.email;
  }

  setAttributes(attributes) {
    this.attributes.displayName = attributes.displayName;
    this.attributes.userName = attributes.userName;
    this.attributes.email = isValidEmail(attributes.email) ? attributes.email : null; // defaults to false

    this.attributes.anonymous = attributes.anonymous ? true : false; // defaults to true

    if (_BmprConstants.UserAttributes.InProject in attributes) {
      this.attributes.inProject = attributes.inProject;
    } else if (this.attributes.anonymous) {
      this.attributes.inProject = false;
    } else {
      this.attributes.inProject = true;
    }
  }

  isAnonymous() {
    return this.attributes.anonymous;
  }

  isInProject() {
    return this.attributes.inProject;
  } // returns only non empty/default attributes

  /**
   * @return {import("@balsamiq/bmpr/lib/BmprConstants").UserInfo}
   */


  getAttributes() {
    let attributes = {
      displayName: this.attributes.displayName,
      userName: this.attributes.userName
    };

    if (this.attributes.email) {
      attributes.email = this.attributes.email;
    }

    if (this.attributes.anonymous) {
      attributes.anonymous = this.attributes.anonymous;
    }

    if (!this.attributes.inProject) {
      attributes.inProject = this.attributes.inProject;
    }

    return attributes;
  }

  attributesDiffer(attributes) {
    if (this.attributes.displayName != attributes.displayName) {
      return true;
    }

    if (this.attributes.userName != attributes.userName) {
      return true;
    }

    if (this.attributes.email != attributes.email) {
      return true;
    }

    if (this.attributes.anonymous != attributes.anonymous) {
      return true;
    }

    if (this.attributes.inProject != attributes.inProject) {
      return true;
    }

    return false;
  }

}

exports.User = User;

var makeUserCoreObjectID = function (projectCoreObjectID, userID) {
  CoreObject.assertProjectType(projectCoreObjectID);

  _balsamiq.default.assertString(userID);

  return new CoreObject.CoreObjectID(CoreObjectType, {
    siteID: projectCoreObjectID.siteID,
    projectID: projectCoreObjectID.projectID,
    userID: userID
  });
};

exports.makeUserCoreObjectID = makeUserCoreObjectID;

/***/ }),

/***/ "./js/core/native-constants.js":
/*!*************************************!*\
  !*** ./js/core/native-constants.js ***!
  \*************************************/
/***/ ((module) => {

"use strict";


module.exports = {
  BAS: {
    NotAuthorized: "Not authorized"
  },
  Timeout: {
    errorName: "TimeoutError",
    secondsForError: 120,
    secondsForWarning: 20
  },
  ImageFormats: {
    PNG: "png",
    JPEG: "jpeg"
  },
  ModuleStatus: {
    NotInitialized: 'notInitialized',
    Initializing: 'initializing',
    Initialized: 'initialized',
    Error: 'error'
  },
  ModuleNames: {
    SQLite: 'SQLite',
    PDF: 'PDF'
  },
  CloudLoginResult: {
    Authorized: 'authorized',
    NotAuthorized: 'notAuthorized',
    Error: 'error'
  },
  CloudErrors: {
    NotAuthorized: "NOT_AUTHORIZED",
    NoRightsToCreate: "CREATE_PROJECT_USER_WITH_NO_RIGHT_TO_CREATE",
    TooManyProjects: "CREATE_PROJECT_TOO_MANY_PROJECTS",
    Generic: "GENERIC"
  },
  CloseReason: {
    QuitApp: "QuitApp",
    CloseProject: "CloseProject",
    CloseProjectIfNew: "CloseProjectIfNew",
    CloseWindow: "CloseWindow"
  },
  LocalFilesError: {
    Generic: 'generic',
    DestFileReadOnly: 'destFileReadOnly'
  },
  WindowKind: {
    Editor: 'editor',
    SearchDocumentation: 'searchdocumentation',
    ContactSupport: 'contactsupport'
  },
  ApplicationActionURISchema: "guestaction:",
  ApplicationAction: {
    OpenContactSupport: "OpenContactSupport"
  }
};

/***/ }),

/***/ "./js/core/native.js":
/*!***************************!*\
  !*** ./js/core/native.js ***!
  \***************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Host = exports.Guest = void 0;
Object.defineProperty(exports, "ImageFormats", ({
  enumerable: true,
  get: function () {
    return _nativeConstants.ImageFormats;
  }
}));
Object.defineProperty(exports, "ModuleNames", ({
  enumerable: true,
  get: function () {
    return _nativeConstants.ModuleNames;
  }
}));
Object.defineProperty(exports, "ModuleStatus", ({
  enumerable: true,
  get: function () {
    return _nativeConstants.ModuleStatus;
  }
}));
exports["default"] = void 0;

var Version = _interopRequireWildcard(__webpack_require__(/*! core/version */ "./js/core/version.js"));

var _nativeConstants = __webpack_require__(/*! core/native-constants */ "./js/core/native-constants.js");

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const unimplemented = function () {
  throw new Error("this method must be implemented by the native Host");
};
/**
 * @typedef {Object} RTCSocketIOInitParams
 * @property {string} path
 * @property {boolean} reconnection
 * @property {number} reconnectionDelay
 * @property {number} reconnectionDelayMax
 * @property {number} reconnectionAttempts
 * @property {string[]} transports
 */


var Host = {
  init: function (params, callback) {
    // called when the app is ready to use Host
    unimplemented();
  },
  reset: function () {
    Host._aborted = false;

    Host._reset();
  },
  _reset: function () {
    unimplemented();
  },
  productName: function () {
    // return String
    // a name that identifies the product
    // will be included in RTE reports
    unimplemented();
  },
  runtimeInfo: function () {
    /*
    return {
        type: 'browser' || 'jsvm',
        os: {
            name: 'Mac OS' || 'Windows' || 'iOS' || 'Android' ...
            version: [ 10, 11, 3 ] // for 10.11.3
            fullString: 'macOS High Sierra 10.13.3'
        },
        browser: { // only when type == 'browser'
            name: 'Safari' || 'Chrome' || 'Firefox' || 'IE' || 'Edge' ...
            engineName: 'Gecko' || 'WebKit' ...
            version: [ 32, 1, 64, 826 ] // 32.1.64.826
            fullString: 'Google Chrome 64.12.433 64 bit',
            isMobile: true || false,
            isTouchEnabled: true || false,
        },
        jsvm: {
            name: 'JavascriptCore' || 'Chakra'
        },
        hostName: 'Steve\'s Laptop',
    }
    */
    unimplemented();
  },
  versionInfo: function () {
    /*
    return {
        appVersion: string
        editorVersion: string
    }
    */
  },
  runtime: {
    isBrowser: () => {
      var _Host$runtimeInfo;

      return !!(((_Host$runtimeInfo = Host.runtimeInfo()) === null || _Host$runtimeInfo === void 0 ? void 0 : _Host$runtimeInfo.type) === "browser");
    },
    isJsvm: () => {
      var _Host$runtimeInfo2;

      return !!(((_Host$runtimeInfo2 = Host.runtimeInfo()) === null || _Host$runtimeInfo2 === void 0 ? void 0 : _Host$runtimeInfo2.type) === "jsvm");
    },
    isMac: () => {
      var _Host$runtimeInfo3;

      return !!(((_Host$runtimeInfo3 = Host.runtimeInfo()) === null || _Host$runtimeInfo3 === void 0 ? void 0 : _Host$runtimeInfo3.os.name) === "Mac OS");
    },
    isWindows: () => {
      var _Host$runtimeInfo4;

      return !!(((_Host$runtimeInfo4 = Host.runtimeInfo()) === null || _Host$runtimeInfo4 === void 0 ? void 0 : _Host$runtimeInfo4.os.name.toLowerCase()) === "windows");
    },
    isMobile: () => {
      var _Host$runtimeInfo5, _Host$runtimeInfo5$br;

      return !!((_Host$runtimeInfo5 = Host.runtimeInfo()) !== null && _Host$runtimeInfo5 !== void 0 && (_Host$runtimeInfo5$br = _Host$runtimeInfo5.browser) !== null && _Host$runtimeInfo5$br !== void 0 && _Host$runtimeInfo5$br.isMobile);
    },
    isTouchEnabled: () => {
      var _Host$runtimeInfo6, _Host$runtimeInfo6$br;

      return !!((_Host$runtimeInfo6 = Host.runtimeInfo()) !== null && _Host$runtimeInfo6 !== void 0 && (_Host$runtimeInfo6$br = _Host$runtimeInfo6.browser) !== null && _Host$runtimeInfo6$br !== void 0 && _Host$runtimeInfo6$br.isTouchEnabled);
    }
  },
  log: function (obj) {
    // log a message in the output
    unimplemented();
  },

  /**
   *
   * @param {Function} callback
   * @param {unknown=} params
   */
  asyncCallback: function (callback, params) {
    // asynchronous call of Guest.Callbacks.invoke(callback)
    unimplemented();
  },

  /**
   * @param {number} ms
   * @param {() => void} callback
   * @return {string}
   */
  timer: function (ms, callback) {
    // invoke callback after ms milliseconds
    // returns id that can be used to cancel this timer
    unimplemented();
  },

  /**
   * @param {string} id
   * @return {void}
   */
  clearTimer: function (id) {
    // clears the timer if not yet fired
    unimplemented();
  },
  openURL: function (url) {
    // open an URL, no callback
    unimplemented();
  },
  nativeAction: function (id, params, callback) {
    unimplemented();
  },
  createZipFile: function (snapshotsData, saveParams, callback) {
    // saveParams: { fileName:String, ext:String, folderURL:String? }
    // callback()
    unimplemented();
  },
  canSaveFile: function (callback) {
    // callback(canSave:bool)
    unimplemented();
  },
  saveFile: function (fileName, base64Data, mimeType, isPotentiallyDangerous, callback) {
    // callback({ fileURL: String? })
    unimplemented();
  },
  abort: function (msg, error) {
    // called when the Core traps an exception or when an assertion fails. the app stops here
    if (!Host._aborted) {
      Host._aborted = true;

      Host._abort(msg, error);
    }
  },
  _aborted: false,
  _abort: function (msg) {
    // to be implemented by platforms
    unimplemented();
  },
  collectDevData: function (code, message, details, error) {
    // code: all-capital string, unique in the codebase
    // message: string
    // details: optional plain object
    // error: optional exception
    // called when the app wants to log some data and maybe send it to a server (if possible).
    // if this feature is not available, then an empty function is ok
    unimplemented();
  },
  rte: function (e) {
    var m = "*** INTERNAL ERROR ***\n\n";
    m += "Product: " + Host.productName() + "\n\n";
    m += "Runtime: " + JSON.stringify(Host.runtimeInfo()) + "\n\n";
    m += "Build Number: " + Version.getBuildNumber() + "\n\n";
    var s;

    if (e instanceof Error) {
      s = m + e.message + "\n\n" + e.stack;
    } else {
      s = m + e.toString();
    }

    Host.abort(s, e);
    Host.log(s);
  },
  callWithTryCatch: function (fun) {
    try {
      return fun();
    } catch (e) {
      Host.rte(e);
    }

    return null;
  },
  SQLite: {
    init: function (callback, progressCallback) {
      // callback({ error: string });
      unimplemented();
    },
    status: function () {
      // return ModuleStatus
      unimplemented();
    },
    has: function (id) {
      // return bool
      unimplemented();
    },
    makeFromData: function (uint8Array, callback) {
      // callback({ id: string, error: string });
      unimplemented();
    },
    makeFromFileURL: function (fileURL, callback) {
      // callback({ id: string, error: string });
      unimplemented();
    },
    remove: function (id, callback) {
      unimplemented();
    },
    open: function (id, mode, callback) {
      // callback({ error: string });
      unimplemented();
    },
    close: function (id, callback) {
      // callback({ error: string })
      unimplemented();
    },
    isOpen: function (id) {
      // return bool
      unimplemented();
    },
    query: function (id, query, parameters, callback) {
      // callback({ rows: array, error: string });
      unimplemented();
    },
    tables: function (id, callback) {
      // callback({ tables: array, error: string });
      unimplemented();
    },
    download: function (id, callback, progressCallback) {
      unimplemented();
    },
    compact: function (id, callback) {
      //callback({error:string})
      unimplemented();
    }
  },
  Fonts: {
    getFontFamilies: function () {
      // return Object fontFamily:string -> { regular: bool, bold: bool, italic: bool, boldItalic: bool }
      unimplemented();
    },
    glyphsInFont: function (fontFamily, bold, italic, text, callback) {
      // callback({ res: bool })
      unimplemented();
    }
  },
  TextMeterHelper: {
    setTextStyle: function (fontFamily, fontSize, bold, italic) {
      // the font family should be one the of available font families
      unimplemented();
    },
    measureStringWidth: function (text) {
      // return text width
      unimplemented();
    }
  },
  Images: {
    make: function (callback) {
      // callback({ id: string, width: integer, height: integer, error: * })
      unimplemented();
    },
    makeSync: function () {
      // return string id
      unimplemented();
    },
    drawCommands: function (id, commands, callback) {
      // callback({ })
      unimplemented();
    },
    drawCommandsSync: function (id, commands, callback) {
      // invoke callback immediately! callback({ })
      unimplemented();
    },
    makeFromURL: function (url, callback) {
      // callback({ id: string, width: integer, height: integer, dpi: integer?, error: * })
      unimplemented();
    },
    makeFromBase64Data: function (mimeType, data, callback) {
      // callback({ id: string, width: integer, height: integer, dpi: integer?, error: * })
      unimplemented();
    },
    getImageBase64Data: function (id, callback) {
      // callback({ id: string, data: string })
      // the callback is called synchronously in the web implementation, asynchronously in the others
      unimplemented();
    },
    remove: function (id, callback) {
      // callback({ })
      unimplemented();
    },
    saveImage: function (id, saveParams, isPotentiallyDangerous, callback) {
      // saveParams: { fileName:String, ext:String, folderURL:String? }
      // callback({ })
      unimplemented();
    },
    resize: function (id, width, height, callback) {
      // resize the image and strip dpi info
      // callback({ error: * })
      unimplemented();
    },
    parallelProcessingCapability: function () {
      // returns int value: 0 is maximum concurrency, 1 serial
      unimplemented();
    },
    getJPGBase64Data: function (id, quality, callback) {
      // saveParams: { id: string, quality: number 0 < x < 1.0 }
      unimplemented(); // callback({ data: string })
    },
    usesAlpha: function (id, callback) {
      // saveParams: { id: string }
      unimplemented(); // callback(boolean)
    },
    getMaxImageSize: function () {
      // return {width : number, height: number}
      unimplemented();
    }
  },
  PDF: {
    init: function (callback, progressCallback) {
      // callback({ error: string });
      unimplemented();
    },
    status: function () {
      // return ModuleStatus
      unimplemented();
    },
    start: function (options, fontFamilies, callback) {
      // callback({ id: string })
      unimplemented();
    },

    /**
        * @param {string | undefined} pageName - Optional string to bookmark the page.
     */
    addPage: function (id, width, height, layout, pageName, callback) {
      // callback({ id: string })
      unimplemented();
    },
    commands: function (id, commands, callback) {
      // callback({ id: string })
      unimplemented();
    },
    save: function (id, saveParams, callback) {
      // saveParams: { fileName:String, ext:String, folderURL:String? }
      // callback({ id?: string }) empty object if did not save (because the user canceled)
      unimplemented();
    }
  },
  HttpRequest: {
    ajax: function (params) {
      // params: {
      //   method: "POST" | "GET"
      //   params: { key: val, ... }
      //   url: string,
      //   callback: function
      // }
      unimplemented();
    },
    get: function (url, callback) {
      // callback({ data: string, error: * })
      unimplemented();
    },
    post: function (url, params, callback) {
      // callback({ data: string, error: * })
      unimplemented();
    },
    getBinary: function (url, params, callback, requestCreatedCallback) {
      // params.base64: bool (base64 or binary format, defaults to true)
      // params.timeout: Number (default 0 -> no timeout)
      // callback({ data: *, error: * })
      // requestCreatedCallback(request) where request.abort is a function used to abort the request
      unimplemented();
    },
    postFormData: function (url, paramsArray, callback) {
      // paramsArray: [{name, value, type, mimeType, filename}]
      // name: fieldName : string
      // value: fieldValue : string
      // type: 'base64' | undefined
      // mimeType: string. mime type of value (used only for base64 type)
      // filename: string. filename (used only for base64 type)
      // callback({...response.body}) (a json object like the ajax method)
      unimplemented();
    }
  },
  UI: {
    makeWindow: function (options, callback) {
      // callback({ id: string })
      unimplemented();
    },
    push: function (windowID, viewControllerID, options, callback) {
      // windowID: string, viewControllerID: string, options: object
      // callback({ })
      unimplemented();
    },
    pop: function (windowID, callback) {
      // windowID: string
      // callback({ id: string })
      unimplemented();
    },
    getTop: function (windowID) {
      // windowID: string
      // return topmost viewControllerID or null if empty
      unimplemented();
    },
    make: function (type, options, state, callback) {
      // type: string, options: object, state: object
      // callback({ id: string })
      unimplemented();
    },

    setStates(viewControllerIDs, states, changedStateKeys) {
      // callback({ });
      unimplemented();
    },

    sendCommands(viewControllerIDs, commandArray) {
      // void
      unimplemented();
    },

    setChild: function (parentViewControllerID, childName, childViewControllerID, callback) {
      // callback({ })
      unimplemented();
    },
    remove: function (viewControllerID, callback) {
      // callback({ })
      unimplemented();
    },
    saveFileDialog: function (cfg, callback) {
      // cfg: { title, message, ext }
      // callback({ fileURL: string? })
      unimplemented();
    },
    showOpenFileDialog: function (cfg, callback) {
      // cfg : {accept: [".jpg", ".png"], multiple: false}
      // callback: function(filesInfo: {size, name, type})
      unimplemented();
    },
    getPickedFile: function (fileInfo, callback) {
      // fileInfo: {name, type, size}
      // callback: function(fileObject)
      unimplemented();
    },
    goFullscreen: function (viewControllerID) {
      //
      unimplemented();
    },
    exitFullscreen: function () {
      // void
      unimplemented();
    },

    requestFocus(viewControllerID) {
      // void
      unimplemented();
    }

  },
  Cursors: {
    // for now, cursors are added by the Native side and must implement the IDs definied by cursors.js
    getCursorIDs: function (callback) {
      // callback({ cursorIDs: [String] })
      unimplemented();
    }
  },
  LocalStorage: {
    setData: function (data, callback) {
      // key, data String
      // callback({ })
      unimplemented();
    },
    getData: function (callback) {
      // callback({ data: *, error: * })
      unimplemented();
    }
  },
  SessionStorage: {
    setData: function (data, callback) {
      // data String
      // callback({ })
      unimplemented();
    },
    getData: function (callback) {
      // callback({ data: *, error: * })
      unimplemented();
    }
  },
  Clipboard: {
    setData: function (data, callback) {
      unimplemented();
    },
    getData: function (callback) {
      // callback({ data: *, error: * })
      unimplemented();
    },

    canGetDataFromSystem() {
      // return true if native can access clipboard contents at any time
      unimplemented();
    },

    canPasteFromMenu() {
      // return true if native can get system clipboard triggered from event
      unimplemented();
    },

    canCopyImageToSystem() {
      // return true if native can put image data into clipboard
      unimplemented();
    }

  },
  Paths: {
    Formats: {
      SVG_PATHS: "svgPaths",
      RENDERER_COMMANDS: "rendererCommands"
    },
    getFormat: function () {
      unimplemented();
    },
    addSkinFragments: function (obj, skinName) {
      //obj: {
      //  id: { paths: [String], colors: [int], width: int, height: int }
      //  ...
      //}
      unimplemented();
    },
    supportsPathLibrary: function () {
      unimplemented();
    }
  },
  Cloud: {
    autologin: function (callback) {
      // domain: staging.balsamiq.cloud, balsamiq.cloud, ...
      // This function will trigger calls to Guest.Cloud
      // callback({ })
      unimplemented();
    },
    logout: function (callback) {// This function will trigger calls to Guest.Cloud
      // callback:({ })
    },
    userSettings: function () {
      // callback:({ })
      unimplemented();
    },
    spaceSettings: function (spaceID) {
      // callback:({ })
      unimplemented();
    },
    query: function (query, variables, callback) {
      // callback({ ok:Bool, result:Object })
      unimplemented();
    },
    getOpenToken: function (siteID36, projectID36, callback) {
      // callback({ ok:Bool, token:String, siteID:Number, projectID:Number, role:String, userInfo:Object, serverURL:String })
      unimplemented();
    },
    uploadProject: function (projectFileUrl, projectName, siteID36, progressCallback, finishCallback) {
      // progressCallback({ currentStep:Number, totalSteps:Number })
      // finishCallback({ error:Object?, projectID:String?, projectName:String? })
      unimplemented();
    },
    cancelUploadProject: function (projectFileUrl, siteID36) {
      unimplemented();
    },
    canLogin: function (callback) {
      // callback({ canLogin: Bool })
      unimplemented();
    },
    openLoginURL: function () {
      // callback()
      unimplemented();
    },
    getAPIUserAuthToken: function (callback) {
      // callback({error: string, token: string})
      unimplemented();
    },
    getDomain: function () {
      // returns Cloud base URL
      unimplemented();
    }
  },
  Network: {
    _isOnline: false,
    isOnline: function () {
      return Host.Network._isOnline;
    }
  },
  UnsavedFiles: {
    /**
     *
     * @param {string} fileURL
     * @param {(result: { result: boolean; unsavedFileExists?: boolean }) => void} callback
     */
    isUnsavedFileURL: function (fileURL, callback) {
      // callback({ result:Bool, unsavedFileExists:Bool? })
      unimplemented();
    },
    createUnsavedFile: function (fileName, ext, callback) {
      // callback({ ok:Bool, error:String?, unsavedFileURL:String? })
      unimplemented();
    },
    removeUnsavedFile: function (unsavedFileURL, callback) {
      // callback({ ok:Bool })
      unimplemented();
    },
    copyUnsavedFileTo: function (unsavedFileURL, userFileURL, callback) {
      // callback({ ok:Bool, error:String? })
      unimplemented();
    },
    markUnsavedFileAsCorrupted: function (unsavedFileURL, callback) {
      // callback({ ok:Bool, error:String? })
      unimplemented();
    }
  },
  LocalFiles: {
    isLocalFileURL: function (fileURL, callback) {
      // callback({ result:Bool, localFileExists:Bool? })
      unimplemented();
    },
    getLocalFileFrom: function (userFileURL, callback) {
      // callback({ localFileURL:String, isEdited:Bool, isOlderThanUserFile:Bool } or null)
      unimplemented();
    },
    createLocalFileFrom: function (userFileURL, callback) {
      // callback({ ok:Bool, localFileURL:String, error:String? })
      unimplemented();
    },
    copyLocalFileTo: function (userFileURL, callback) {
      // callback({ ok:Bool, error:LocalFilesError?, errorMessage:String? })
      unimplemented();
    },
    copyLocalFileToNewUserFile: function (userFileURL, newUserFileURL, callback) {
      // callback({ ok:Bool, error:String? })
      unimplemented();
    },
    editLocalFileOf: function (userFileURL, callback) {
      // callback({})
      unimplemented();
    },
    removeLocalFileFromRegistry: function (userFileURL, callback) {
      // callback({})
      unimplemented();
    },
    removeLocalFile: function (userFileURL, callback) {
      // callback({})
      unimplemented();
    },
    updateLocalFileURL: function (userFileURL, localFileURL, callback) {
      // missing entries will be created
      // callback({})
      unimplemented();
    },
    markLocalFileAsCorruptedAndRemoveFromRegistry: function (userFileURL, callback) {
      // callback({ ok:Bool, error:String? })
      unimplemented();
    }
  },
  Desktop: {
    License: {
      getData: function (callback) {
        // callback({})
        unimplemented();
      }
    },
    userInfo: function userInfo(callback) {
      // callback({ email: String, name: String, avatarURL: String })
      unimplemented();
    },
    selectURL: function (url) {
      unimplemented();
    }
  },
  RTCSocketIO: {
    /**
     * @param {string} url
     * @param {RTCSocketIOInitParams} params
     * @return {string}
     */
    init: function (url, params) {
      // returns id
      unimplemented();
    },

    /**
     * @param {string} id
     * @return {void}
     */
    connect: function (id) {
      unimplemented();
    },

    /**
     * @param {string} id
     * @return {void}
     */
    disconnect: function (id) {
      unimplemented();
    },

    /**
     * @param {string} id
     * @param {string} event
     * @param {unknown} payload
     * @return {void}
     */
    emit: function (id, event, payload) {
      unimplemented();
    },

    /**
     * @param {string} id
     * @param {() => void} onConnectCallback
     * @return {void}
     */
    onConnect: function (id, onConnectCallback) {
      // callback()
      unimplemented();
    },

    /**
     * @param {string} id
     * @param {() => void} onDisconnectCallback
     * @return {void}
     */
    onDisconnect: function (id, onDisconnectCallback) {
      // callback()
      unimplemented();
    },

    /**
     * @param {string} id
     * @param {(message: unknown) => void} onMessageCallback
     * @return {void}
     */
    onMessage: function (id, onMessageCallback) {
      // callback(message: object)
      unimplemented();
    },

    /**
     * @param {string} id
     * @param {(data: unknown) => void} onPresenceCallback
     * @return {void}
     */
    onPresence: function (id, onPresenceCallback) {
      // callback(data: object)
      unimplemented();
    }
  }
};
exports.Host = Host;
var Guest = {
  Callbacks: {
    _callbacks: {},
    // token -> function
    _nextToken: 0,
    add: function (fun) {
      var token = "callback-" + Guest.Callbacks._nextToken;
      Guest.Callbacks._nextToken++;
      Guest.Callbacks._callbacks[token] = fun;
      return token;
    },
    invoke: function (token, args) {
      if (token) {
        var fun = Guest.Callbacks._callbacks[token];
        delete Guest.Callbacks._callbacks[token];
        Host.callWithTryCatch(function () {
          fun.apply(null, args);
        });
      }
    },
    invokeAsync: function (token, args) {
      if (token) {
        var token2 = Guest.Callbacks.add(function () {
          Guest.Callbacks.invoke(token, args);
        });
        Host.asyncCallback(token2);
      }
    }
  },
  UI: {
    _focusedID: null,
    _action: null,
    action: function (id, action, params) {
      // id: string, action: string, params: object
      if (Host.onUserActivity) {
        Host.onUserActivity();
      }

      if (Guest.UI._action) {
        Host.callWithTryCatch(function () {
          Guest.UI._action(id, action, params);
        });
      }
    },
    _keyboardEvent: null,
    keyboardEvent: function (id, event) {
      Guest.UI._focusedID = id;

      if (event === null) {
        return true;
      } // id: string, event: object


      if (Guest.UI._keyboardEvent) {
        return Host.callWithTryCatch(function () {
          return Guest.UI._keyboardEvent(id, event);
        });
      } else {
        return false;
      }
    },
    _mouseEvent: null,
    mouseEvent: function (id, event) {
      if (Host.onUserActivity && event.type === "mousedown") {
        Host.onUserActivity();
      }

      if (Guest.UI._mouseEvent) {
        return Host.callWithTryCatch(function () {
          return Guest.UI._mouseEvent(id, event);
        });
      } else {
        return false;
      }
    }
  },
  App: {
    onFontsLoaded: function () {},
    _onAwake: null,
    onAwake: function () {
      if (Guest.App._onAwake) {
        Host.callWithTryCatch(() => {
          Guest.App._onAwake();
        });
      }
    }
  },
  LocalStorage: {
    settingsChangedEvent: function (data) {// new settings data
    }
  },
  SessionStorage: {
    settingsChangedEvent: function (data) {// new settings data
    }
  },
  Clipboard: {
    clipboardChangedEvent: function (data) {// new settings data
    }
  },
  Cloud: {
    _onLogin: null,
    onLogin: function (domain, result, userInfo) {
      _balsamiq.default.assertString(domain);

      _balsamiq.default.assertEnum(result, _nativeConstants.CloudLoginResult);

      if (Guest.Cloud._onLogin) {
        Host.callWithTryCatch(() => {
          Guest.Cloud._onLogin(domain, result, userInfo);
        });
      }
    }
  },
  Network: {
    _onNetworkStatusChange: null,
    onNetworkStatusChange: function (isOnline) {
      _balsamiq.default.assertBool(isOnline);

      Host.Network._isOnline = isOnline;

      if (Guest.Network._onNetworkStatusChange) {
        Guest.Network._onNetworkStatusChange(isOnline);
      }
    }
  },
  LocalFiles: {},
  Desktop: {
    License: {
      onDataChanged: function (data) {// New License data for files
      }
    },
    _onUserInfoChange: null,
    onUserInfoChange: function (info) {
      if (info != null) {
        _balsamiq.default.assertObject(info);

        Host.log(`User Info changed. New info: ${info}`);
      }

      Guest.Desktop._onUserInfoChange(info);
    }
  },
  guestAction: function (id, params, callback) {
    if (Guest.guestActionHandlers[id] && Guest.guestActionHandlers[id].length > 0) {
      Guest.guestActionHandlers[id][Guest.guestActionHandlers[id].length - 1](params, callback);
    } else {
      callback({
        error: "Unknown Guest Action error: " + id
      });
    }
  },
  addGuestActionHandler: function (id, handler) {
    if (!Guest.guestActionHandlers[id]) {
      Guest.guestActionHandlers[id] = [];
    }

    Guest.guestActionHandlers[id].push(handler);
  },
  removeGuestActionHandler: function (id, handler) {
    let handlers = Guest.guestActionHandlers[id];

    if (handlers) {
      let index = handlers.indexOf(handler);

      if (index > -1) {
        handlers.splice(index, 1);
      }
    }
  },
  guestActionHandlers: {}
};
exports.Guest = Guest;
var _default = {
  Guest,
  Host,
  ImageFormats: _nativeConstants.ImageFormats,
  ModuleStatus: _nativeConstants.ModuleStatus,
  ModuleNames: _nativeConstants.ModuleNames
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/operations/archiveeventoperation.js":
/*!*****************************************************!*\
  !*** ./js/core/operations/archiveeventoperation.js ***!
  \*****************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ArchiveEventOperation = void 0;

var Archive = _interopRequireWildcard(__webpack_require__(/*! core/archive */ "./js/core/archive.js"));

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _operation = __webpack_require__(/*! core/operations/operation */ "./js/core/operations/operation.ts");

var ArchiveState = _interopRequireWildcard(__webpack_require__(/*! core/model/archivestate */ "./js/core/model/archivestate.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

class ArchiveEventOperation extends _operation.Operation {
  canUndo() {
    return false;
  }

  name() {
    var n = "Archive Event Operation";

    if (this.params.event) {
      n += ": " + this.params.event.eventType;

      if (this.params.event.eventData.archiveRevision) {
        n += " archiveRevision: " + this.params.event.eventData.archiveRevision;
      }
    }

    return n;
  }

  runDo() {
    _balsamiq.default.exists(this.params.projectCoreObjectID);

    _balsamiq.default.exists(this.params.event);

    const eventType = this.params.event.eventType;
    const eventData = this.params.event.eventData;
    const project = this.context.getCoreObject(this.params.projectCoreObjectID);

    if (!project) {
      this.notifyFinished();
      return;
    }

    const archiveState = project.getArchiveState();

    if (archiveState.getKeyValue(ArchiveState.KeyNames.ERROR)) {
      // Archive has an error, then this operation can not be executed
      this.context.host.log(`${this.name()}: skipping because the ArchiveState has an error`);
      this.notifyFinished();
      return;
    }

    let res;

    switch (eventType) {
      case Archive.EventTypes.RESOURCE_DATA_CHANGED:
        _balsamiq.default.exists(eventData.resourceID, "RESOURCE_DATA_CHANGED: eventData.resourceID");

        _balsamiq.default.exists(eventData.branchID, "RESOURCE_DATA_CHANGED: eventData.branchID");

        _balsamiq.default.exists(this.params.data, "RESOURCE_DATA_CHANGED: this.params.data");

        res = project.onResourceDataChanged(eventData.branchID, eventData.resourceID, this.params.data);
        break;

      case Archive.EventTypes.RESOURCE_ATTRIBUTES_CHANGED:
        _balsamiq.default.exists(eventData.resourceID, "RESOURCE_ATTRIBUTES_CHANGED: eventData.resourceID");

        _balsamiq.default.exists(eventData.branchID, "RESOURCE_ATTRIBUTES_CHANGED: eventData.branchID");

        _balsamiq.default.exists(eventData.attributes, "RESOURCE_ATTRIBUTES_CHANGED: eventData.attributes");

        project.onResourceAttributesChanged(eventData.branchID, eventData.resourceID, eventData.attributes);
        break;

      case Archive.EventTypes.RESOURCE_ADDED:
        _balsamiq.default.exists(eventData.resourceID, "RESOURCE_ADDED: eventData.resourceID");

        _balsamiq.default.exists(eventData.branchID, "RESOURCE_ADDED: eventData.branchID");

        _balsamiq.default.exists(eventData.attributes, "RESOURCE_ADDED: eventData.attributes");

        project.onResourceAdded(eventData.branchID, eventData.resourceID, eventData.attributes);
        break;

      case Archive.EventTypes.RESOURCES_DELETED:
        _balsamiq.default.exists(eventData.resourceIDs, "RESOURCES_DELETED: eventData.resourceID");

        _balsamiq.default.exists(eventData.branchID, "RESOURCES_DELETED: eventData.branchID");

        project.onResourcesDeleted(eventData.branchID, eventData.resourceIDs);
        break;

      case Archive.EventTypes.BRANCH_ADDED:
        _balsamiq.default.exists(eventData.branchID, "BRANCH_ADDED: eventData.branchID");

        project.onBranchAdded(eventData.branchID, eventData.attributes);
        break;

      case Archive.EventTypes.BRANCH_ATTRIBUTES_CHANGED:
        _balsamiq.default.exists(eventData.branchID, "BRANCH_ATTRIBUTES_CHANGED: eventData.branchID");

        _balsamiq.default.exists(eventData.attributes, "RESOURCE_ADDED: eventData.attributes");

        project.onBranchAttributesChanged(eventData.branchID, eventData.attributes);
        break;

      case Archive.EventTypes.THUMBNAIL_ADDED:
        _balsamiq.default.exists(eventData.thumbnailID, "THUMBNAIL_ADDED: eventData.thumbnailID");

        _balsamiq.default.exists(this.params.attributes, "THUMBNAIL_ADDED: eventData.attributes");

        project.onThumbnailAdded(eventData.thumbnailID, this.params.attributes);
        break;

      case Archive.EventTypes.THUMBNAILS_DELETED:
        _balsamiq.default.exists(eventData.thumbnailIDs, "THUMBNAILS_DELETED: eventData.thumbnailIDs");

        project.onThumbnailsDeleted(eventData.thumbnailIDs);
        break;

      case Archive.EventTypes.BRANCHES_DELETED:
        _balsamiq.default.exists(eventData.branchIDs, "BRANCHES_DELETED: eventData.branchIDs");

        project.onBranchesDeleted(eventData.branchIDs);
        break;

      case Archive.EventTypes.ARCHIVE_ATTRIBUTES_CHANGED:
        _balsamiq.default.exists(eventData.attributes, "ARCHIVE_ATTRIBUTES_CHANGED: eventData.attributes");

        project.onArchiveAttributesChanged(eventData.attributes);
        break;

      case Archive.EventTypes.COMMENT_CREATED:
        project.onCommentAdded(this.params);
        break;

      case Archive.EventTypes.COMMENTS_DELETED:
        project.onCommentDeleted(this.params.commentIDs);
        break;

      case Archive.EventTypes.COMMENTS_ATTRIBUTES_UPDATED:
        project.updateCommentsAttributes(this.params.attributes);
        break;

      case Archive.EventTypes.COMMENT_DATA_UPDATED:
        project.updateComment(this.params.commentID, this.params.attributes, this.params.data);
        break;

      case Archive.EventTypes.BROADCAST_MESSAGE:
        project.onBroadcastMessage(this.params.message.id, this.params.message.type, this.params.userName, this.params.timestamp, this.params.message.data);
        break;

      case Archive.EventTypes.RESOURCE_BRANCH_CHANGED:
        {
          const {
            resourceID,
            oldBranchID,
            newBranchID
          } = eventData;

          _balsamiq.default.isString(resourceID);

          _balsamiq.default.isString(oldBranchID);

          _balsamiq.default.isString(newBranchID);

          project.moveAlternateToAnotherBranchFromArchive(resourceID, oldBranchID, newBranchID);
          break;
        }

      case Archive.EventTypes.PERMALINK_DELETED:
        {
          const {
            resourceID,
            branchID
          } = this.params;
          project.onPermalinkDataDeleted({
            resourceID,
            branchID
          });
          break;
        }

      case Archive.EventTypes.PERMALINK_CREATED_OR_UPDATED:
        {
          const permalink = this.params;
          project.onPermalinkCreatedOrUpdated(permalink);
          break;
        }

      default:
        break;
    }

    this.notifyFinished(res);
  }

}

exports.ArchiveEventOperation = ArchiveEventOperation;

/***/ }),

/***/ "./js/core/operations/genericoperation.js":
/*!************************************************!*\
  !*** ./js/core/operations/genericoperation.js ***!
  \************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.GenericOperation = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var Operation = __webpack_require__(/*! core/operations/operation */ "./js/core/operations/operation.ts");

class GenericOperation extends Operation.Operation {
  name() {
    return this.params.operationName ? this.params.operationName : "Generic Operation";
  }

  runDo() {
    _balsamiq.default.exists(this.params.callback, "GenericOperation requires a callback in params"); // optional this.params.operationName:String


    this.params.callback();
  }

}

exports.GenericOperation = GenericOperation;

/***/ }),

/***/ "./js/core/operations/operationcenter.js":
/*!***********************************************!*\
  !*** ./js/core/operations/operationcenter.js ***!
  \***********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.OperationCenter = exports.KeyNames = exports.CoreObjectType = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _coreobject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var _operation = __webpack_require__(/*! core/operations/operation */ "./js/core/operations/operation.ts");

var _operationqueue = __webpack_require__(/*! core/operations/operationqueue */ "./js/core/operations/operationqueue.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const CoreObjectType = "OperationCenter";
exports.CoreObjectType = CoreObjectType;
const KeyNames = {
  HasUndo: "hasUndo",
  HasRedo: "hasRedo",
  UndoOperationName: "undoOperationName",
  RedoOperationName: "redoOperationName",
  StackHead: "stackHead",
  StackSize: "stackSize"
};
exports.KeyNames = KeyNames;

class OperationCenter extends _coreobject.CoreObject {
  constructor(context, centerID) {
    _balsamiq.default.existsClass(context.operationQueue, _operationqueue.OperationQueue);

    super(context, (0, _coreobject.makeCoreObjectID)(CoreObjectType, {
      centerID: centerID
    }));
    this.stack = []; // oldest operation is at position 0, newest at position length-1

    this.stackHead = 0; // can vary from 0 (before the oldest operation) to stack.length (after newest operation)
  }

  getKeyNames() {
    return super.getKeyNames().concat(Object.values(KeyNames));
  }

  getKeyValue(keyName) {
    switch (keyName) {
      case KeyNames.HasUndo:
        return this.stackHead > 0;

      case KeyNames.HasRedo:
        return this.stackHead < this.stack.length;

      case KeyNames.UndoOperationName:
        return this.stackHead > 0 ? this.stack[this.stackHead - 1].name() : "";

      case KeyNames.RedoOperationName:
        return this.stackHead < this.stack.length ? this.stack[this.stackHead].name() : "";

      case KeyNames.StackHead:
        return this.stackHead;

      case KeyNames.StackSize:
        return this.stack.length;
    }

    return super.getKeyValue(keyName);
  }

  doOperation(operation) {
    let finishCallback = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : () => {};
    let progressCallback = arguments.length > 2 ? arguments[2] : undefined;

    _balsamiq.default.existsClass(operation, _operation.Operation);

    if (operation.canUndo()) {
      // the operation supports undo/redo
      // remove newer operations (if any)
      this.stack = this.stack.slice(0, this.stackHead);
      this.context.operationQueue.scheduleOperation(operation, res => {
        if (res.success) {
          // only in case of success, insert the new operation on top and update the head
          this.stack.push(operation);
          this.stackHead += 1;
        }

        this.notifyChange();
        finishCallback(res);
      }, progressCallback, _operation.Direction.Do);
    } else {
      // the operation does not support undo/redo, then it won't be pushed on the stack
      this.context.operationQueue.scheduleOperation(operation, finishCallback, progressCallback, _operation.Direction.Do);
    }
  }

  undoPreviousOperation() {
    let finishCallback = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : () => {};
    let progressCallback = arguments.length > 1 ? arguments[1] : undefined;

    _balsamiq.default.assert(this.stackHead > 0);

    this.context.operationQueue.scheduleOperation(this.stack[this.stackHead - 1], res => {
      if (res.success) {
        this.stackHead -= 1;
      } else {
        // in case of failure remove the operation from the stack
        this.stackHead -= 1;
        this.stack.splice(this.stackHead, 1);
      }

      this.notifyChange();
      finishCallback(res);
    }, progressCallback, _operation.Direction.Undo);
  }

  redoNextOperation() {
    let finishCallback = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : () => {};
    let progressCallback = arguments.length > 1 ? arguments[1] : undefined;

    _balsamiq.default.assert(this.stackHead < this.stack.length);

    this.context.operationQueue.scheduleOperation(this.stack[this.stackHead], res => {
      if (res.success) {
        this.stackHead += 1;
      } else {
        // in case of failure remove the operation from the stack
        this.stack.splice(this.stackHead, 1);
      }

      this.notifyChange();
      finishCallback(res);
    }, progressCallback, _operation.Direction.Do);
  }

  clearHistory() {
    this.stack = []; // oldest operation is at position 0, newest at position length-1

    this.stackHead = 0;
    this.notifyChange();
  }

  notifyChange() {
    this.notifyKeyValueChanged(KeyNames.HasUndo);
    this.notifyKeyValueChanged(KeyNames.HasRedo);
    this.notifyKeyValueChanged(KeyNames.UndoOperationName);
    this.notifyKeyValueChanged(KeyNames.RedoOperationName);
  }

}

exports.OperationCenter = OperationCenter;

/***/ }),

/***/ "./js/core/operations/operationqueue.js":
/*!**********************************************!*\
  !*** ./js/core/operations/operationqueue.js ***!
  \**********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = exports.OperationQueue = exports.EventTypes = exports.CoreObjectType = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var CoreObject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var Operation = __webpack_require__(/*! core/operations/operation */ "./js/core/operations/operation.ts");

var Events = __webpack_require__(/*! core/events */ "./js/core/events.ts");

const CoreObjectType = "OperationQueue";
exports.CoreObjectType = CoreObjectType;
const EventTypes = {
  OPERATION_STARTING: "operationStarting",
  OPERATION_PROGRESS: "operationProgress",
  OPERATION_ENDED: "operationEnded"
};
exports.EventTypes = EventTypes;

class OperationQueue extends CoreObject.CoreObject {
  constructor(context, queueID) {
    _balsamiq.default.exists(context, "context must exist");

    _balsamiq.default.exists(context.eventCenter, "context.eventCenter must exist");

    _balsamiq.default.exists(context.getCoreObject, "context.getCoreObject must exist");

    _balsamiq.default.exists(context.host, "context.host must exist");

    super(context, CoreObject.makeCoreObjectID(CoreObjectType, {
      queueID: queueID
    }));
    this.queue = []; // operationIDs of operations yet to start

    this.data = {}; // operationID -> { operation: Operation, finishCallback: Function, progressCallback: Function }

    this.runningOperation = null; // currently updating { operation: Operation }

    this.nextOperationID = 0;
    this.operationContext = {
      eventCenter: this.context.eventCenter,
      getCoreObject: this.context.getCoreObject,
      host: this.context.host
    };
  }

  cleanUp() {
    this.queue = [];
    this.data = {};
    this.runningOperation = null;
    super.cleanUp();
  }

  scheduleOperation(operation, finishCallback, progressCallback, direction) {
    _balsamiq.default.existsClass(operation, Operation.Operation);

    if (finishCallback) {
      _balsamiq.default.assertFunction(finishCallback);
    }

    if (progressCallback) {
      _balsamiq.default.assertFunction(progressCallback);
    }

    _balsamiq.default.assertEnum(direction, Operation.Direction);

    var operationID = "op-" + this.nextOperationID;
    this.nextOperationID += 1;

    if (!this.data[operationID]) {
      this.data[operationID] = {
        operation: operation,
        finishCallback: finishCallback,
        progressCallback: progressCallback,
        direction: direction
      };
      this.queue.push(operationID);
      this.advanceQueue();
    } else {// already in queue, what to do?
    }

    return operationID;
  }

  filterOperations(filter) {
    _balsamiq.default.assertFunction(filter);

    return this.queue.filter(operationID => {
      return filter(this.data[operationID].operation);
    });
  } //cancelOperation(operationID) {
  //    if (this.runningOperation && this.runningOperation.operation.getOperationID() == operationID) {
  //        this.runningOperation.operation.cancel();
  //    } else {
  //        // TODO test this
  //        for (var i = 0; i < this.queue.length; i++) {
  //            if (this.queue[i] === operationID) {
  //                delete this.data[operationID];
  //                this.queue.splice(i, 1);
  //                break;
  //            }
  //        }
  //    }
  //}
  // Private methods
  // ---------------


  advanceQueue() {
    if (!this.runningOperation) {
      if (this.queue.length > 0) {
        var operationID = this.queue.shift();
        this.runningOperation = this.data[operationID];
        delete this.data[operationID];
        this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.OPERATION_STARTING, {
          operation: this.runningOperation.operation
        }));
        this.context.host.log("OperationQueue: started operation '" + this.runningOperation.operation.name() + "' " + operationID);
        this.runningOperation.operation.run(this.operationContext, operationID, this.onOperationFinished.bind(this), this.onOperationProgress.bind(this), this.runningOperation.direction);
      }
    }
  }

  onOperationFinished(res) {
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.OPERATION_ENDED, {
      operation: this.runningOperation.operation,
      result: res
    }));
    const finishCallback = this.runningOperation.finishCallback;
    this.context.host.log("OperationQueue: ended operation '" + this.runningOperation.operation.name() + "' " + this.runningOperation.operation.operationID);
    this.runningOperation = null;

    if (finishCallback) {
      finishCallback(res);
    }

    this.advanceQueue();
  }

  onOperationProgress(currentStep, maxSteps) {
    this.context.eventCenter.fireEvent(Events.newCoreEvent(this.coreObjectID, EventTypes.OPERATION_PROGRESS, {
      operation: this.runningOperation.operation,
      currentStep: currentStep,
      maxSteps: maxSteps
    }));

    if (this.runningOperation.progressCallback) {
      this.runningOperation.progressCallback(currentStep, maxSteps);
    }
  }

}

exports.OperationQueue = OperationQueue;
var _default = {
  CoreObjectType,
  OperationQueue,
  EventTypes
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/serialization/json.js":
/*!***************************************!*\
  !*** ./js/core/serialization/json.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ValidationErrorCodes = exports.UNKNOWN_PROPERTIES_EXCLUDE_LIST = void 0;
exports.decodeHrefProperty = decodeHrefProperty;
exports.decodeHrefsProperty = decodeHrefsProperty;
exports.decodeIconProperty = decodeIconProperty;
exports.decodeIconsProperty = decodeIconsProperty;
exports.decodeImageSrcProperty = decodeImageSrcProperty;
exports.decodePointProperty = decodePointProperty;
exports.decodeSymbolSrcProperty = decodeSymbolSrcProperty;
exports.encodeLink = encodeLink;
exports.getControlsAndDepsFromContainer = getControlsAndDepsFromContainer;
exports.getControlsBBoxMS = getControlsBBoxMS;
exports.getMockupAttributes = getMockupAttributes;
exports.getMockupResourceID = getMockupResourceID;
exports.getSymbolDefinitions = getSymbolDefinitions;
exports.haveControlsEqualContent = haveControlsEqualContent;
exports.iconSizeNumberToIconSize = iconSizeNumberToIconSize;
exports.isContainerOfControlsAndDeps = isContainerOfControlsAndDeps;
exports.isSymbolToBeBroken = isSymbolToBeBroken;
exports.loadProperties = loadProperties;
exports.makeContainerOfControlsAndDeps = makeContainerOfControlsAndDeps;
exports.readColorPropertyName = readColorPropertyName;
exports.readProperties = readProperties;
exports.readProperty = readProperty;
exports.saveControl = saveControl;
exports.saveMockupObject = saveMockupObject;
exports.saveProperty = saveProperty;
exports.smartCompareOfMockupJSONS = smartCompareOfMockupJSONS;
exports.symbolHasDirectDependency = symbolHasDirectDependency;
exports.validateMockupJSON = validateMockupJSON;

var _semver = __webpack_require__(/*! core/utils/semver */ "./js/core/utils/semver.ts");

var _wireframeVersion = __webpack_require__(/*! core/wireframe-version */ "./js/core/wireframe-version.ts");

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var _fontawesomemap = __webpack_require__(/*! core/skins/fontawesomemap */ "./js/core/skins/fontawesomemap.js");

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const ContainerConstants = {
  CONTROLS_CONTAINER: 'controlsContainer',
  CONTROLS: 'controls',
  CONTROLS_WITH_BROKEN_SYMBOLS: 'controlsWithBrokenSymbols',
  DEPENDENCIES: 'dependencies'
};
const UNKNOWN_PROPERTIES_EXCLUDE_LIST = ["map", "text_links", "uncroppedDims"];
exports.UNKNOWN_PROPERTIES_EXCLUDE_LIST = UNKNOWN_PROPERTIES_EXCLUDE_LIST;
const ValidationErrorCodes = {
  UNSUPPORTED_VERSION: "unsupportedVersion",
  INVALID_FORMAT: "invalidFormat"
};
exports.ValidationErrorCodes = ValidationErrorCodes;

function decodeSymbolSrcProperty(p_src) {
  _balsamiq.default.exists(p_src.ID);

  _balsamiq.default.exists(p_src.Anchor);

  return new _types.default.SrcProperty(_types.default.SrcLocation.PROJECT_ASSETS, p_src.ID, p_src.Anchor.toString());
}

function decodeImageSrcProperty(p_src) {
  if (p_src === null || p_src === undefined) {
    return null;
  }

  _balsamiq.default.exists(p_src.ID); //Balsamiq.exists(p_src.Anchor);


  return new _types.default.SrcProperty(_types.default.SrcLocation.PROJECT_ASSETS, p_src.ID, p_src.Anchor !== undefined ? p_src.Anchor.toString() : undefined);
}

function decodeIconProperty(p_iconProp) {
  //noinspection JSDeprecatedSymbols
  var r = null;

  if (p_iconProp) {
    var res;
    var uuidRegExp = new RegExp("[A-Z0-9]{8}-([A-Z0-9]{4}-){3}[A-Z0-9]{12}");
    var iconID = p_iconProp.ID ? p_iconProp.ID : p_iconProp.name;

    if (uuidRegExp.test(iconID)) {
      // custom icon
      res = new _types.default.SrcProperty(_types.default.SrcLocation.PROJECT_ASSETS, iconID);
    } else {
      // app icon
      iconID = (0, _fontawesomemap.getNewIconName)(iconID); // always use new names
      // Balsamiq.exists(iconID, "could not find a icon with name: " + iconID);

      if (_balsamiq.default.isNull(iconID)) {
        //hack for when an icon is null, it draws an empty icon
        var newProp = _icon.default.emptyIconData(_types.default.IconSize.large);

        newProp.size = p_iconProp.size != null ? p_iconProp.size : newProp.size;
        p_iconProp = newProp;
        p_iconProp.name = "EmptyIcon";
        res = new _types.default.SrcProperty(_types.default.SrcLocation.APP_ASSETS, "EmptyIcon");
      } else {
        res = new _types.default.SrcProperty(_types.default.SrcLocation.APP_ASSETS, iconID);
      }
    }

    var iconSize = _types.default.IconSize[p_iconProp.size];

    if (!iconSize) {
      iconSize = _types.default.IconSize.large;
    }

    r = new _icon.default.IconProperty(res, iconSize);
  }

  return r;
}

function decodeIconsProperty(p_iconsProp) {
  _balsamiq.default.assertArray(p_iconsProp.icons);

  return p_iconsProp.icons.map(icon => {
    if (icon.ID) {
      _balsamiq.default.assertString(icon.ID);

      return decodeIconProperty(icon);
    }

    return {
      "size": null
    };
  });
}

function decodeHrefProperty(p_src) {
  // var location;
  // var resource;
  var str = null;

  if (_balsamiq.default.has(p_src, "URL")) {
    // location = Types.SrcLocation.URI;
    // resource = p_src.URL;
    str = p_src.URL;
  } else if (_balsamiq.default.has(p_src, "ID")) {
    // location = Types.SrcLocation.PROJECT_ASSETS;
    // resource = p_src.ID;
    str = "resource:" + p_src.ID;
  }

  return str; //return [new Types.SrcProperty(location, resource, -1)];
}

function decodeHrefsProperty(p_src) {
  _balsamiq.default.assertArray(p_src.href);

  return p_src.href.map(href => {
    return decodeHrefProperty(href);
  });
}

function decodePointProperty(p_src) {
  return {
    x: p_src.x,
    y: p_src.y
  };
}

function saveMockupObject(_ref) {
  let {
    group,
    controlIDs,
    doBreakSymbols,
    doBreakControlsInsideSymbols = false,
    project,
    isMockupMainGroup,
    attributes,
    resourceID,
    branchID,
    notes
  } = _ref;

  _balsamiq.default.exists(group);

  _balsamiq.default.assertOptionalArray(controlIDs);

  _balsamiq.default.assertBool(doBreakSymbols);

  _balsamiq.default.assertBool(doBreakControlsInsideSymbols);

  _balsamiq.default.assertObjectOrNull(project);

  _balsamiq.default.assertBool(isMockupMainGroup);

  _balsamiq.default.assertObject(attributes);

  _balsamiq.default.assertString(resourceID);

  var json = {
    mockup: {
      controls: {
        control: []
      },
      attributes,
      branchID,
      resourceID,
      mockupH: "",
      mockupW: "",
      measuredW: "",
      measuredH: "",
      version: _wireframeVersion.BaseVersion,
      notes
    }
  };
  const savedControls = [];

  let rootBounds = _types.default.newRectangle(0, 0, 0, 0);

  let mainGroupBBox = group.getUserBBox(true);
  let saveChildrenSelectionOnly = controlIDs && controlIDs.length;

  if (group.children && group.children.length > 0) {
    let childBBox = null;
    let children = group.children;

    if (saveChildrenSelectionOnly) {
      children = children.filter(control => {
        return controlIDs.indexOf(control.controlID) !== -1;
      });
      childBBox = _types.default.rectangleListUnion(children.map(child => {
        return child.getUserBBox(true); // return child['_userBBoxNoMarkup'];
      }));
    }

    for (var i = 0; i < children.length; ++i) {
      savedControls.push(children[i]);
      saveControl(children[i], i, json.mockup.controls.control, doBreakSymbols, project, doBreakControlsInsideSymbols);
    }

    if (!childBBox) {
      rootBounds = mainGroupBBox;
    } else {
      rootBounds = childBBox; // correction for the "infinite" mockup

      if (isMockupMainGroup) {
        rootBounds.x += mainGroupBBox.x;
        rootBounds.y += mainGroupBBox.y;
      }
    }

    json.mockup.controls.control = _balsamiq.default.sortBy(json.mockup.controls.control, function (control) {
      return parseInt(control.ID, 10);
    }); // correction for the "infinite" mockup: moving the offset to all first level children

    if (isMockupMainGroup) {
      for (i = 0; i < json.mockup.controls.control.length; ++i) {
        json.mockup.controls.control[i].x = "" + (parseInt(json.mockup.controls.control[i].x, 10) + mainGroupBBox.x);
        json.mockup.controls.control[i].y = "" + (parseInt(json.mockup.controls.control[i].y, 10) + mainGroupBBox.y);
      }
    }
  }

  json.mockup.mockupH = "" + rootBounds.height;
  json.mockup.mockupW = "" + rootBounds.width; // in other words, this offset correction is not needed in case we are
  // exporting a non root group as a whole

  json.mockup.measuredW = "" + (rootBounds.width + (isMockupMainGroup || saveChildrenSelectionOnly ? rootBounds.x : 0));
  json.mockup.measuredH = "" + (rootBounds.height + (isMockupMainGroup || saveChildrenSelectionOnly ? rootBounds.y : 0));
  json.mockup.version = savedControls.reduce((maxVersion, control) => {
    const version = control.getVersion();
    return (0, _semver.highestSemVer)(version, maxVersion);
  }, _wireframeVersion.BaseVersion);
  return json;
}

function isSymbolToBeBroken(p_control, doBreakSymbols) {
  return p_control.controlTypeID === "com.balsamiq.mockups::Component" && doBreakSymbols;
}

function saveControl(p_control, order, p_json, doBreakSymbols, project, symbolBreakUnderWay) {
  // TODO remove project param and copy the symbol name in the Symbol instance
  const controlProperties = p_control.getNonDefaultActualControlProperties();
  const typeIDPrefix = "com.balsamiq.mockups::";
  let control = {
    ID: _balsamiq.default.isNumber(p_control.controlID) ? "" + p_control.controlID : "1",
    typeID: p_control.controlTypeID === "__group__" ? p_control.controlTypeID : p_control.controlTypeID.substr(typeIDPrefix.length),
    /// strip "com.balsamiq.mockups::"
    zOrder: "" + order
  };
  let jsonProperties = {};
  let skipOverrides = false;
  const breakSymbol = isSymbolToBeBroken(p_control, doBreakSymbols);

  if (breakSymbol) {
    // we are breaking the symbol, overridden control properties will be reported in child controls
    skipOverrides = true; // tweak it transforming it into a group

    control.typeID = "__group__";
    control.ID = "" + p_control.parent.getNextControlID(); // find the name

    const symLibs = project.getSymbolLibraryList();

    for (let symbolLibraryIndex = 0; symbolLibraryIndex < symLibs.length; ++symbolLibraryIndex) {
      const symbLib = symLibs[symbolLibraryIndex];

      if (symbLib.coreObjectID.matches({
        resourceID: p_control.properties.src.resourceID
      })) {
        jsonProperties.controlName = symbLib.symbolDefinitions[p_control.properties.src.subResourceID];
        break;
      }
    }
  }

  const size = p_control.computeSize();

  if (p_control.hasPropertyDefinition("size")) {
    const sizeProp = p_control.getActualProperty("size");
    let defaultSize = p_control.getDefaultProperty("size");

    if (sizeProp.width !== _types.default.AUTOSIZE && size.width != defaultSize.width) {
      control.w = "" + size.width;
    }

    if (sizeProp.height !== _types.default.AUTOSIZE && size.height != defaultSize.height) {
      control.h = "" + size.height;
    } // no default size? find it in measure


    if (defaultSize.width === -1 || defaultSize.height === -1) {
      defaultSize = p_control.measure(defaultSize);
    }

    control.measuredW = "" + (defaultSize.width === -1 ? size.width : defaultSize.width);
    control.measuredH = "" + (defaultSize.height === -1 ? size.height : defaultSize.height);
  } else {
    control.measuredW = "" + size.width;
    control.measuredH = "" + size.height;

    if (p_control.controlTypeID === "com.balsamiq.mockups::Component" || p_control.controlTypeID === "__group__") {
      control.w = "" + size.width;
      control.h = "" + size.height;
    }
  } // TODO: use getProperties otherwise controls without origin explicitly set will generate an internal error


  const origin = controlProperties['origin'];
  control.x = "" + origin.x;
  control.y = "" + origin.y;

  if (p_control.isLocked()) {
    control.locked = "true";
  }

  if (symbolBreakUnderWay || breakSymbol && doBreakSymbols) {
    for (let property in controlProperties) {
      let value = controlProperties[property];

      if (property === "origin" || property === "size") {
        continue;
      }

      if (skipOverrides && property === "overrides") {
        // apply overridden properties of the symbol instance (for example "break apart a cropped nested symbol")
        const selfOverride = value.filter(o => _balsamiq.default.isEqual(o.controlID, [-1]));

        if (selfOverride.length === 1) {
          Object.keys(selfOverride[0].properties).forEach(propName => {
            saveProperty(jsonProperties, propName, selfOverride[0].properties[propName]);
          });
        }

        continue;
      }

      if (property == "src" && breakSymbol) {
        // skip it, we're breaking the symbol
        continue;
      }

      saveProperty(jsonProperties, property, value);
    }
  } else {
    jsonProperties = p_control.saveProperties();
  }

  if (!_balsamiq.default.isEmpty(jsonProperties)) {
    control.properties = jsonProperties;
  }

  p_json.push(control);

  if (p_control.children && p_control.children.length > 0 && control.typeID !== "Component") {
    control.children = {};
    control.children.controls = {};
    control.children.controls.control = [];
    var order2 = 0;
    p_control.children.forEach(function (child) {
      saveControl(child, order2, control.children.controls.control, doBreakSymbols, project, symbolBreakUnderWay || breakSymbol);
      order2++;
    }, this);
    control.children.controls.control = _balsamiq.default.sortBy(control.children.controls.control, function (control) {
      return parseInt(control.ID, 10);
    });
  }
}

function iconSizeNumberToIconSize(value) {
  switch (value) {
    case 16:
      return "xsmall";

    case 24:
      return "small";

    case 32:
      return "medium";

    case 48:
      return "large";

    case 64:
      return "xlarge";

    case 128:
      return "xxlarge";

    default:
      _balsamiq.default.assert(false);

  }
}

function parseOverrides(overrides) {
  return overrides.map((override, i) => {
    const props = {};
    props.ID = overrides[i].controlID.join(":");

    for (let key in overrides[i].properties) {
      if (key === "overrides") {
        props[key] = parseOverrides(overrides[i].properties[key]);
      } else {
        props[key] = overrides[i].properties[key];
      }
    }

    return props;
  });
}

function saveProperty(json, propertyName, property) {
  if (propertyName === "origin") {
    json.x = "" + property.x;
    json.y = "" + property.y;
  } else if (propertyName === "size") {
    if (_balsamiq.default.isObject(property)) {
      json.w = "" + property.width;
      json.h = "" + property.height;
    } else {
      json.size = "" + property;
    }
  } else if (propertyName === "src") {
    if (property === null) {
      json.src = null;
    } else {
      json.src = {};

      if ("subResourceID" in property && property["subResourceID"]) {
        json.src.Anchor = property["subResourceID"];
      }

      if ("resourceID" in property) {
        json.src.ID = property["resourceID"];
      }

      if ("Anchor" in property) {
        json.src.Anchor = property["Anchor"];
      }

      if ("ID" in property) {
        json.src.ID = property["ID"];
      }
    }
  } else if (propertyName === "fontSize") {
    json["size"] = "" + property;
  } else if (propertyName === "icon") {
    if (property != null) {
      json.icon = {};
      json.icon.ID = property.src.resourceID;
      json.icon.size = iconSizeNumberToIconSize(property.size);
    }
  } else if (propertyName === "icons") {
    json.icons = {};
    json.icons.icons = property.map(iconProp => {
      var icon = {};

      if (iconProp != null && iconProp.size != null) {
        icon.ID = iconProp.src.resourceID;
        icon.size = iconSizeNumberToIconSize(iconProp.size);
      } else {
        icon.size = null;
      }

      return icon;
    });
  } else if (propertyName === "value") {
    json.value = "" + Math.round(property * 100);
  } else if (propertyName === "scrollBarValue") {
    json.scrollBarValue = "" + Math.round(property * 100);
  } else if (propertyName === "overrides") {
    json.overrides = parseOverrides(property);
  } else if (['p0', 'p1', 'p2'].indexOf(propertyName) > -1) {
    json[propertyName] = {
      x: property.x,
      y: property.y
    };
  } else if (propertyName == 'links') {
    var jsonLinks = [];

    for (let i = 0; i < property.length; ++i) {
      jsonLinks.push(encodeLink(property[i]));
    }

    json['hrefs'] = {
      href: jsonLinks
    };
  } else if (propertyName == 'link') {
    json['href'] = encodeLink(property);
  } else {
    json[propertyName] = "" + property;
  }
}

function encodeLink(property) {
  let r = {};

  if (property) {
    if (property === "action://back") {
      r = {
        URL: property
      };
    } else if (property.startsWith("http")) {
      r = {
        URL: property
      };
    } else if (property.startsWith("resource:")) {
      let plainResourceID = property.substr(9);
      r = {
        ID: plainResourceID
      };
    }
  }

  return r;
}

function validateMockupJSON(p_json) {
  let shouldAssert = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;
  let error, errorCode;

  if (!p_json) {
    error = "Error loading Mockup from json: json is empty.";
    errorCode = ValidationErrorCodes.INVALID_FORMAT;
  } else if (!p_json.mockup) {
    error = "Error loading Mockup from json: missing main 'mockup' object.";
    errorCode = ValidationErrorCodes.INVALID_FORMAT;
  }

  if (!error) {
    const version = _balsamiq.default.get(p_json, "mockup.version");

    if (!version) {
      error = "Error loading Mockup from json: missing 'version' info.";
      errorCode = ValidationErrorCodes.INVALID_FORMAT;
    } else {
      const r = (0, _semver.readSemVer)(version);

      if (!r) {
        error = "Error loading Mockup from json: 'version' is not valid.";
        errorCode = ValidationErrorCodes.INVALID_FORMAT;
      } else {
        const currentVersion = (0, _semver.readSemVer)(_wireframeVersion.CurrentVersion);

        if (r.major > currentVersion.major) {
          errorCode = ValidationErrorCodes.UNSUPPORTED_VERSION;
          error = errorCode;
        }
      }
    }
  }

  if (shouldAssert) {
    _balsamiq.default.assert(!error, error);
  }

  return error ? {
    error,
    errorCode
  } : {};
}
/**
 * Read the 'overrides' property of a symbol instance
 */


function convertJsonOverridesToControlsOverrides(overriddenJsonValue) {
  // json converts as object the arrays with one item, but we want an array here
  const values = !_balsamiq.default.isArray(overriddenJsonValue) ? [overriddenJsonValue] : overriddenJsonValue; // return the overridden properties for each child control

  return values.reduce((overriddenProps, controlOverrides) => {
    // controlID is a path like this: 1:2:0
    // convert it to an Array, to match the controlID of coreObjectIDs
    const o = {
      controlID: null,
      properties: {}
    };
    const controlID = controlOverrides.ID.split(":").map(stringID => parseInt(stringID, 10));
    o.controlID = controlID;

    for (var key in controlOverrides) {
      if (key === "overrides") {
        o.properties[key] = convertJsonOverridesToControlsOverrides(controlOverrides[key]);
      } else if (key != "ID") {
        o.properties[key] = controlOverrides[key];
      }
    }

    overriddenProps.push(o);
    return overriddenProps;
  }, []);
}

function loadProperties(p_context, p_modelControl, p_jsonControl) {
  let allowNegativeCoords = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : false;

  _balsamiq.default.exists(p_context);

  _balsamiq.default.exists(p_modelControl);

  _balsamiq.default.exists(p_jsonControl);

  let names = [],
      values = [];
  let hasUnknownProperties = false;
  let unknownPropertyNames = []; // unknown property names: [ propName, .. ]

  let unknownPropertyValues = []; // known property names of properties with unknown value: [ [ propName, propValue], .. ]
  // origin

  names.push("origin");
  let x = parseInt(p_jsonControl.x, 10);
  let y = parseInt(p_jsonControl.y, 10);

  if (isNaN(x) || !allowNegativeCoords && x < 0) {
    x = 0;
  }

  if (isNaN(y) || !allowNegativeCoords && y < 0) {
    y = 0;
  }

  values.push(_types.default.newPoint(x, y)); // size (only if the control supports it)

  if (p_modelControl.hasPropertyDefinition("size")) {
    names.push('size');
    var w = _balsamiq.default.has(p_jsonControl, "w") ? parseInt(p_jsonControl.w, 10) : -1;
    var h = _balsamiq.default.has(p_jsonControl, "h") ? parseInt(p_jsonControl.h, 10) : -1;
    values.push(_types.default.newSize(w, h));
  } // other properties


  if (_balsamiq.default.has(p_jsonControl, 'properties')) {
    var propNames = Object.keys(p_jsonControl.properties);
    propNames.forEach(propName => {
      var propValue = p_jsonControl.properties[propName];
      var p;

      if (propName === "overrides") {
        // the 'overrides' property contains the array of overridden properties of children controls
        names.push(propName);
        const overrideValues = convertJsonOverridesToControlsOverrides(propValue);
        values.push(overrideValues);
      } else {
        // regular case
        if (!UNKNOWN_PROPERTIES_EXCLUDE_LIST.includes(propName)) {
          p = readProperty(p_context, p_modelControl, propName, propValue);

          if (p) {
            if (p_modelControl.checkPropertyValue(p.name, p.value)) {
              names.push(p.name);
              values.push(p.value);
            } else {
              hasUnknownProperties = true;
              unknownPropertyValues.push([propName, propValue]);
            }
          } else {
            if (propName !== "hrefs") {
              hasUnknownProperties = true;
              unknownPropertyNames.push(propName);
            }
          }
        }
      }
    });
  }

  if (p_modelControl.hasPropertyDefinition("overrides") && names.indexOf('overrides') === -1) {
    names.push('overrides');
    values.push([]);
  } // set undefined every prop present in the current control but not in the json


  const namesMap = names.reduce((propNames, propName) => {
    propNames[propName] = true;
    return propNames;
  }, {});

  _balsamiq.default.objectForEach(p_modelControl.properties, (_propValue, propName) => {
    if (!namesMap[propName]) {
      names.push(propName);
      values.push(undefined);
    }
  });

  p_modelControl.setActualProperties(names, values); // Log unknown properties to console

  if (unknownPropertyNames.length > 0 || unknownPropertyValues.length > 0) {
    p_context.logger(`Unknown properties in Control with ID: ${JSON.stringify(p_modelControl.coreObjectID.controlID)} type: ${p_modelControl.controlTypeID} branchID: ${p_modelControl.coreObjectID.branchID.substr(0, 4)} resourceID: ${p_modelControl.coreObjectID.resourceID.substr(0, 4)}`);
  }

  if (unknownPropertyNames.length > 0) {
    p_context.logger(`    Unknown Property Name: ${JSON.stringify(unknownPropertyNames)}`);
  }

  if (unknownPropertyValues.length > 0) {
    p_context.logger(`    Unknown Property Values: ${JSON.stringify(unknownPropertyValues)}`);
  }

  return {
    hasUnknownProperties
  };
}

function readProperty(p_context, p_control, p_name, p_value) {
  _balsamiq.default.exists(p_control);

  _balsamiq.default.assertString(p_name);

  _balsamiq.default.assert(p_name !== "map", "this function can't handle 'map' property"); // TODO remove this if statement and add an assertion like for "map" property name


  if (p_name === "overrides") {
    const value = p_value.map(_ref2 => {
      let {
        controlID,
        properties
      } = _ref2;
      const childControl = controlID[0] === -1 ? p_control : p_control.getControl(controlID);
      return {
        controlID,
        properties: childControl ? readProperties(childControl, properties) : {}
      };
    });
    return {
      name: p_name,
      value
    };
  } // fix this property name


  var newName = p_name;

  if (p_name === 'size') {
    newName = 'fontSize';
  } else if (p_name === 'href') {
    newName = 'link';
  } else if (p_name === 'hrefs') {
    newName = 'links';
  }

  if (p_name == 'color' || p_name == 'borderColor' || p_name == 'textColor') {
    newName = readColorPropertyName(p_control, p_name);
  }

  if (!p_control.hasPropertyDefinition(newName)) {
    return undefined;
  }

  var res = {
    name: newName,
    value: null
  };

  if (p_control.controlTypeID === 'com.balsamiq.mockups::Component' && res.name === 'src') {
    res.value = decodeSymbolSrcProperty(p_value);
  } else if (res.name === 'src') {
    res.value = decodeImageSrcProperty(p_value);
  } else if (res.name === 'icon') {
    res.value = decodeIconProperty(p_value);
  } else if (p_name === 'href') {
    res.value = decodeHrefProperty(p_value);
  } else if (p_name === 'hrefs') {
    res.value = decodeHrefsProperty(p_value);
  } else if (p_control.controlTypeID === 'com.balsamiq.mockups::Arrow' && (res.name === 'p0' || res.name === 'p1' || res.name === 'p2')) {
    res.value = decodePointProperty(p_value);
  } else if (p_control.controlTypeID === 'com.balsamiq.mockups::Arrow' && res.name === 'curvature') {
    // some old arrow had problems see: https://www.pivotaltracker.com/story/show/153793729
    res.value = parseInt(p_value);
  } else if (p_name == "icons") {
    res.value = decodeIconsProperty(p_value);
  } else {
    res.value = p_value;
  } // deserialize the property value


  var dataType = p_control.getPropertyDataType(res.name); // fix these properties before deserialization

  if (['value', 'scrollBarValue'].indexOf(res.name) >= 0) {
    res.value = res.value / 100;
  } else if (res.name === 'state') {
    if (res.value === 'indetermined') {
      res.value = 'indeterminate';
    } else if (res.value === 'disabledIndetermined') {
      res.value = 'disabledIndeterminate';
    }
  } else if (res.name === 'fontSize') {
    // Invalid font sizes will be ignored!!
    try {
      let val = parseInt(res.value, 10);

      if (val <= 0) {
        res.value = undefined;

        p_control._context.host.log('WARNING! Ignoring invalid font size in control:' + p_control.coreObjectID.controlID);
      }
    } catch (e) {
      res.value = undefined;

      p_control._context.host.log('WARNING! Ignoring invalid font size in control:' + p_control.coreObjectID.controlID);
    }
  }

  res.value = _types.default.propertyDataTypes[dataType].deserialize(res.value);
  return res;
}

function readProperties(p_control, jsonProperties) {
  const context = p_control.getContext();
  const initialJsonProps = { ...jsonProperties
  };
  const resultProps = {}; // expect x,y,w,z which are mapped to origin and size read every property json via readProperty

  if (initialJsonProps.x != null) {
    resultProps["origin"] = _types.default.newPoint(parseInt(initialJsonProps.x, 10), parseInt(initialJsonProps.y, 10));
    delete initialJsonProps.x;
    delete initialJsonProps.y;
  }

  if (initialJsonProps.w != null) {
    resultProps["size"] = _types.default.newSize(parseInt(initialJsonProps.w, 10), parseInt(initialJsonProps.h, 10));
    delete initialJsonProps.w;
    delete initialJsonProps.h;
  }

  return Object.keys(initialJsonProps).reduce(function readProperties(overriddenProperties, propName) {
    const propValue = initialJsonProps[propName];
    const res = readProperty(context, p_control, propName, propValue);

    if (res) {
      overriddenProperties[res.name] = res.value;
    }

    return overriddenProperties;
  }, { ...resultProps
  });
}

function readColorPropertyName(p_control, p_name) {
  var res = p_name;

  if (p_control.hasPropertyDefinition('color') && p_control.getPropertyOptions('color').oldName == p_name) {
    res = 'color';
  } else if (p_control.hasPropertyDefinition('borderColor') && p_control.getPropertyOptions('borderColor').oldName == p_name) {
    res = 'borderColor';
  } else if (p_control.hasPropertyDefinition('textColor') && p_control.getPropertyOptions('textColor').oldName == p_name) {
    res = 'textColor';
  } else if (p_control.hasPropertyDefinition('rowColor') && p_control.getPropertyOptions('rowColor').oldName == p_name) {
    res = 'rowColor';
  }

  return res;
}
/**
 * Find all top level groups with name. If none is found, then consider the whole mockup as a symbol definition.
 * @param {Object} p_json
 * @returns {Object}
 */


function getSymbolDefinitions(p_json) {
  _balsamiq.default.exists(p_json.mockup);

  var res = {};
  var topLevelControlsJSON = p_json.mockup.controls;
  topLevelControlsJSON = _balsamiq.default.isArray(topLevelControlsJSON.control) ? topLevelControlsJSON.control : [topLevelControlsJSON.control];
  topLevelControlsJSON.forEach(controlJSON => {
    if (controlJSON.typeID === "__group__") {
      if (_balsamiq.default.has(controlJSON, "properties") && _balsamiq.default.has(controlJSON.properties, "controlName")) {
        //noinspection JSDeprecatedSymbols
        res[parseInt(controlJSON.ID, 10)] = parseInt(controlJSON.ID, 10);
      }
    }
  });

  if (_balsamiq.default.isEmpty(res)) {
    res[_types.default.NO_SYMBOL_DEFINITION] = -1;
  }

  return res;
}

function haveChildren(object1, object2, hierarchy) {
  let current1 = object1;
  let current2 = object2;

  for (let child of hierarchy) {
    if (!_balsamiq.default.has(current1, child) || !_balsamiq.default.has(current2, child)) {
      return false;
    }

    current1 = current1[child];
    current2 = current2[child];
  }

  return true;
}

function compareGroupOfControls(controlArray1, controlArray2) {
  let {
    checkPosition
  } = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {
    checkPosition: true
  };

  if (controlArray1.length != controlArray2.length) {
    return false;
  }

  for (let i = 0; i < controlArray1.length; ++i) {
    let control1 = controlArray1[i];
    let control2 = controlArray2[i];

    if (control1.typeID !== control2.typeID) {
      return false;
    }

    if (checkPosition && control1.x && control1.x !== control2.x) {
      return false;
    }

    if (checkPosition && control1.y && control1.y !== control2.y) {
      return false;
    }

    if (control1.typeID !== "__group__" && control1.typeID !== "Component") {
      if (control1.w && control1.w !== control2.w) {
        return false;
      }

      if (control1.h && control1.h !== control2.h) {
        return false;
      }
    }

    if (!_balsamiq.default.isEqual(control1.properties, control2.properties)) {
      return false;
    }

    if (control1.typeID === "__group__") {
      if (!haveChildren(control1, control2, ["children", "controls", "control"])) {
        if (!_balsamiq.default.has(control1, ["children", "controls", "control"]) && !_balsamiq.default.has(control2, ["children", "controls", "control"])) {
          return true;
        } else {
          return false;
        }
      }

      if (!compareGroupOfControls(control1.children.controls.control, control2.children.controls.control)) {
        return false;
      }
    }
  }

  return true;
}

function smartCompareOfMockupJSONS(json1, json2) {
  let {
    checkPosition
  } = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  if (!haveChildren(json1, json2, ["mockup", "controls", "control"])) {
    return false;
  }

  return compareGroupOfControls(json1.mockup.controls.control, json2.mockup.controls.control, {
    checkPosition
  });
}

function getControlsBBoxMS(json) {
  let groupOffset = json.groupOffset || {
    x: 0,
    y: 0
  };
  let coords = {
    x: groupOffset.x + (json.mockup.measuredW - json.mockup.mockupW),
    y: groupOffset.y + (json.mockup.measuredH - json.mockup.mockupH),
    width: parseInt(json.mockup.mockupW),
    height: parseInt(json.mockup.mockupH)
  };
  return coords;
}

function makeContainerOfControlsAndDeps(controls, controlsWithBrokenSymbols, dependencies) {
  _balsamiq.default.assertObject(controls);

  let res = {
    [ContainerConstants.CONTROLS_CONTAINER]: true,
    [ContainerConstants.controls]: controls
  };

  if (controlsWithBrokenSymbols) {
    _balsamiq.default.assertObject(controlsWithBrokenSymbols);

    res[ContainerConstants.CONTROLS_WITH_BROKEN_SYMBOLS] = controlsWithBrokenSymbols;
  }

  if (dependencies) {
    _balsamiq.default.assertArray(dependencies);

    res[ContainerConstants.DEPENDENCIES] = dependencies;
  }

  return res;
}

function isContainerOfControlsAndDeps(json) {
  _balsamiq.default.assertObject(json);

  return json[ContainerConstants.CONTROLS_CONTAINER] === true && _balsamiq.default.isObject(json[ContainerConstants.CONTROLS]); // TODO check if there is the constant
}

function getControlsAndDepsFromContainer(json) {
  _balsamiq.default.assertObject(json);

  _balsamiq.default.assert(json[ContainerConstants.CONTROLS_CONTAINER] === true);

  _balsamiq.default.assertObject(json[ContainerConstants.CONTROLS]);

  let res = {
    [ContainerConstants.CONTROLS]: json[ContainerConstants.CONTROLS]
  };

  if (json.controlsWithBrokenSymbols) {
    res[ContainerConstants.CONTROLS_WITH_BROKEN_SYMBOLS] = json[ContainerConstants.CONTROLS_WITH_BROKEN_SYMBOLS];
  }

  if (json.dependencies) {
    _balsamiq.default.assertArray(json.dependencies);

    res[ContainerConstants.DEPENDENCIES] = json[ContainerConstants.DEPENDENCIES];
  }

  return res;
}

function getMockupAttributes(mockupObj) {
  _balsamiq.default.assertObject(mockupObj);

  return _lodash.default.pick(_balsamiq.default.get(mockupObj, "mockup.attributes"), ["name", "order", "parentID"]);
}

function getMockupResourceID(mockupObj) {
  _balsamiq.default.assertObject(mockupObj);

  return _balsamiq.default.get(mockupObj, "mockup.coreObjectID.resourceID") || _balsamiq.default.get(mockupObj, "mockup.resourceID");
}

function haveControlsEqualContent(c1, c2) {
  return compareGroupOfControls([c1], [c2], {
    checkPosition: false
  });
}
/**
 * Checks wether a symbol in the symbol library directly contains a specific resource and subResource ids
 *
 * @param {import("core/serialization/json-types").ControlJSON[]} controls
 * @param {string} resourceID The ResourceId of the dependency to check
 * @param {string} subResourceID The SubResourceID of the dependency to check
 * @returns {boolean}
 */


function symbolHasDirectDependency(controls, resourceID, subResourceID) {
  for (const control of controls) {
    const hasDirectDependency = control.typeID === "Component" && control.properties.src.ID === resourceID && control.properties.src.Anchor === subResourceID.toString();

    if (hasDirectDependency || control.children && symbolHasDirectDependency(control.children.controls.control, resourceID, subResourceID)) {
      return true;
    }
  }

  return false;
}

/***/ }),

/***/ "./js/core/skins/fontawesomeicons.js":
/*!*******************************************!*\
  !*** ./js/core/skins/fontawesomeicons.js ***!
  \*******************************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;
var _default = [["glass", "Glass", 0xf000, ["Web Application"], ["martini", "drink", "bar", "alcohol", "liquor"]], ["music", "Music", 0xf001, ["Web Application"], ["note", "sound"]], ["search", "Search", 0xf002, ["Web Application"], ["magnify", "zoom", "enlarge", "bigger"]], ["envelope-o", "Envelope Outlined", 0xf003, ["Web Application"], ["email", "e-mail", "letter", "support", "mail", "message", "notification"]], ["heart", "Heart", 0xf004, ["Web Application", "Medical"], ["love", "like", "favorite"]], ["star", "Star", 0xf005, ["Web Application"], ["award", "achievement", "night", "rating", "score", "favorite"]], ["star-o", "Star Outlined", 0xf006, ["Web Application"], ["award", "achievement", "night", "rating", "score", "favorite"]], ["user", "User", 0xf007, ["Web Application"], ["person", "man", "head", "profile"]], ["film", "Film", 0xf008, ["Web Application"], ["movie"]], ["th-large", "th-large", 0xf009, ["Text Editor"], ["blocks", "squares", "boxes", "grid"]], ["th", "th", 0xf00a, ["Text Editor"], ["blocks", "squares", "boxes", "grid"]], ["th-list", "th-list", 0xf00b, ["Text Editor"], ["ul", "ol", "checklist", "finished", "completed", "done", "todo"]], ["check", "Check", 0xf00c, ["Web Application"], ["checkmark", "done", "todo", "agree", "accept", "confirm", "tick", "ok"]], ["times", "Times", 0xf00d, ["Web Application"], ["close", "exit", "x", "cross", "remove", "close"]], ["search-plus", "Search Plus", 0xf00e, ["Web Application"], ["magnify", "zoom", "enlarge", "bigger"]], ["search-minus", "Search Minus", 0xf010, ["Web Application"], ["magnify", "minify", "zoom", "smaller"]], ["power-off", "Power Off", 0xf011, ["Web Application"], ["on"]], ["signal", "signal", 0xf012, ["Web Application"], ["graph", "bars"]], ["cog", "cog", 0xf013, ["Web Application", "Spinner"], ["settings", "gear"]], ["trash-o", "Trash Outlined", 0xf014, ["Web Application"], ["garbage", "delete", "remove", "trash", "hide"]], ["home", "home", 0xf015, ["Web Application"], ["main", "house"]], ["file-o", "File Outlined", 0xf016, ["Text Editor", "File Type"], ["new", "page", "pdf", "document"]], ["clock-o", "Clock Outlined", 0xf017, ["Web Application"], ["watch", "timer", "late", "timestamp"]], ["road", "road", 0xf018, ["Web Application"], ["street"]], ["download", "Download", 0xf019, ["Web Application"], ["import"]], ["arrow-circle-o-down", "Arrow Circle Outlined Down", 0xf01a, ["Directional"], ["download"]], ["arrow-circle-o-up", "Arrow Circle Outlined Up", 0xf01b, ["Directional"]], ["inbox", "inbox", 0xf01c, ["Web Application"]], ["play-circle-o", "Play Circle Outlined", 0xf01d, ["Video Player"]], ["repeat", "Repeat", 0xf01e, ["Text Editor"], ["redo", "forward", "rotate-right"]], ["refresh", "refresh", 0xf021, ["Web Application", "Spinner"], ["reload", "sync"]], ["list-alt", "list-alt", 0xf022, ["Text Editor"], ["ul", "ol", "checklist", "finished", "completed", "done", "todo"]], ["lock", "lock", 0xf023, ["Web Application"], ["protect", "admin", "security"]], ["flag", "flag", 0xf024, ["Web Application"], ["report", "notification", "notify"]], ["headphones", "headphones", 0xf025, ["Web Application"], ["sound", "listen", "music", "audio"]], ["volume-off", "volume-off", 0xf026, ["Web Application"], ["audio", "mute", "sound", "music"]], ["volume-down", "volume-down", 0xf027, ["Web Application"], ["audio", "lower", "quieter", "sound", "music"]], ["volume-up", "volume-up", 0xf028, ["Web Application"], ["audio", "higher", "louder", "sound", "music"]], ["qrcode", "qrcode", 0xf029, ["Web Application"], ["scan"]], ["barcode", "barcode", 0xf02a, ["Web Application"], ["scan"]], ["tag", "tag", 0xf02b, ["Web Application"], ["label"]], ["tags", "tags", 0xf02c, ["Web Application"], ["labels"]], ["book", "book", 0xf02d, ["Web Application"], ["read", "documentation"]], ["bookmark", "bookmark", 0xf02e, ["Web Application"], ["save"]], ["print", "print", 0xf02f, ["Web Application"]], ["camera", "camera", 0xf030, ["Web Application"], ["photo", "picture", "record"]], ["font", "font", 0xf031, ["Text Editor"], ["text"]], ["bold", "bold", 0xf032, ["Text Editor"]], ["italic", "italic", 0xf033, ["Text Editor"], ["italics"]], ["text-height", "text-height", 0xf034, ["Text Editor"]], ["text-width", "text-width", 0xf035, ["Text Editor"]], ["align-left", "align-left", 0xf036, ["Text Editor"], ["text"]], ["align-center", "align-center", 0xf037, ["Text Editor"], ["middle", "text"]], ["align-right", "align-right", 0xf038, ["Text Editor"], ["text"]], ["align-justify", "align-justify", 0xf039, ["Text Editor"], ["text"]], ["list", "list", 0xf03a, ["Text Editor"], ["ul", "ol", "checklist", "finished", "completed", "done", "todo"]], ["outdent", "Outdent", 0xf03b, ["Text Editor"], ["dedent"]], ["indent", "Indent", 0xf03c, ["Text Editor"]], ["video-camera", "Video Camera", 0xf03d, ["Web Application"], ["film", "movie", "record"]], ["picture-o", "Picture Outlined", 0xf03e, ["Web Application"], ["photo", "image"]], ["pencil", "pencil", 0xf040, ["Web Application"], ["write", "edit", "update"]], ["map-marker", "map-marker", 0xf041, ["Web Application"], ["map", "pin", "location", "coordinates", "localize", "address", "travel", "where", "place"]], ["adjust", "adjust", 0xf042, ["Web Application"], ["contrast"]], ["tint", "tint", 0xf043, ["Web Application"], ["raindrop", "waterdrop", "drop", "droplet"]], ["pencil-square-o", "Pencil Square Outlined", 0xf044, ["Web Application"], ["write", "edit", "update", "edit"]], ["share-square-o", "Share Square Outlined", 0xf045, ["Web Application"], ["social", "send", "arrow"]], ["check-square-o", "Check Square Outlined", 0xf046, ["Web Application", "Form Control"], ["todo", "done", "agree", "accept", "confirm", "ok"]], ["arrows", "Arrows", 0xf047, ["Web Application", "Directional"], ["move", "reorder", "resize"]], ["step-backward", "step-backward", 0xf048, ["Video Player"], ["rewind", "previous", "beginning", "start", "first"]], ["fast-backward", "fast-backward", 0xf049, ["Video Player"], ["rewind", "previous", "beginning", "start", "first"]], ["backward", "backward", 0xf04a, ["Video Player"], ["rewind", "previous"]], ["play", "play", 0xf04b, ["Video Player"], ["start", "playing", "music", "sound"]], ["pause", "pause", 0xf04c, ["Video Player"], ["wait"]], ["stop", "stop", 0xf04d, ["Video Player"], ["block", "box", "square"]], ["forward", "forward", 0xf04e, ["Video Player"], ["forward", "next"]], ["fast-forward", "fast-forward", 0xf050, ["Video Player"], ["next", "end", "last"]], ["step-forward", "step-forward", 0xf051, ["Video Player"], ["next", "end", "last"]], ["eject", "eject", 0xf052, ["Video Player"]], ["chevron-left", "chevron-left", 0xf053, ["Directional"], ["bracket", "previous", "back"]], ["chevron-right", "chevron-right", 0xf054, ["Directional"], ["bracket", "next", "forward"]], ["plus-circle", "Plus Circle", 0xf055, ["Web Application"], ["add", "new", "create", "expand"]], ["minus-circle", "Minus Circle", 0xf056, ["Web Application"], ["delete", "remove", "trash", "hide"]], ["times-circle", "Times Circle", 0xf057, ["Web Application"], ["close", "exit", "x"]], ["check-circle", "Check Circle", 0xf058, ["Web Application"], ["todo", "done", "agree", "accept", "confirm", "ok"]], ["question-circle", "Question Circle", 0xf059, ["Web Application"], ["help", "information", "unknown", "support"]], ["info-circle", "Info Circle", 0xf05a, ["Web Application"], ["help", "information", "more", "details"]], ["crosshairs", "Crosshairs", 0xf05b, ["Web Application"], ["picker"]], ["times-circle-o", "Times Circle Outlined", 0xf05c, ["Web Application"], ["close", "exit", "x"]], ["check-circle-o", "Check Circle Outlined", 0xf05d, ["Web Application"], ["todo", "done", "agree", "accept", "confirm", "ok"]], ["ban", "ban", 0xf05e, ["Web Application"], ["delete", "remove", "trash", "hide", "block", "stop", "abort", "cancel"]], ["arrow-left", "arrow-left", 0xf060, ["Directional"], ["previous", "back"]], ["arrow-right", "arrow-right", 0xf061, ["Directional"], ["next", "forward"]], ["arrow-up", "arrow-up", 0xf062, ["Directional"]], ["arrow-down", "arrow-down", 0xf063, ["Directional"], ["download"]], ["share", "Share", 0xf064, ["Web Application"], ["mail-forward"]], ["expand", "Expand", 0xf065, ["Video Player"], ["enlarge", "bigger", "resize"]], ["compress", "Compress", 0xf066, ["Video Player"], ["collapse", "combine", "contract", "merge", "smaller"]], ["plus", "plus", 0xf067, ["Web Application"], ["add", "new", "create", "expand"]], ["minus", "minus", 0xf068, ["Web Application"], ["hide", "minify", "delete", "remove", "trash", "hide", "collapse"]], ["asterisk", "asterisk", 0xf069, ["Web Application"], ["details"]], ["exclamation-circle", "Exclamation Circle", 0xf06a, ["Web Application"], ["warning", "error", "problem", "notification", "alert"]], ["gift", "gift", 0xf06b, ["Web Application"], ["present"]], ["leaf", "leaf", 0xf06c, ["Web Application"], ["eco", "nature", "plant"]], ["fire", "fire", 0xf06d, ["Web Application"], ["flame", "hot", "popular"]], ["eye", "Eye", 0xf06e, ["Web Application"], ["show", "visible", "views"]], ["eye-slash", "Eye Slash", 0xf070, ["Web Application"], ["toggle", "show", "hide", "visible", "visiblity", "views"]], ["exclamation-triangle", "Exclamation Triangle", 0xf071, ["Web Application"], ["warning", "error", "problem", "notification", "alert", "warning"]], ["plane", "plane", 0xf072, ["Web Application", "Transportation"], ["travel", "trip", "location", "destination", "airplane", "fly", "mode"]], ["calendar", "calendar", 0xf073, ["Web Application"], ["date", "time", "when", "event"]], ["random", "random", 0xf074, ["Web Application", "Video Player"], ["sort", "shuffle"]], ["comment", "comment", 0xf075, ["Web Application"], ["speech", "notification", "note", "chat", "bubble", "feedback", "message", "texting", "sms", "conversation"]], ["magnet", "magnet", 0xf076, ["Web Application"]], ["chevron-up", "chevron-up", 0xf077, ["Directional"]], ["chevron-down", "chevron-down", 0xf078, ["Directional"]], ["retweet", "retweet", 0xf079, ["Web Application"], ["refresh", "reload", "share"]], ["shopping-cart", "shopping-cart", 0xf07a, ["Web Application"], ["checkout", "buy", "purchase", "payment"]], ["folder", "Folder", 0xf07b, ["Web Application"]], ["folder-open", "Folder Open", 0xf07c, ["Web Application"]], ["arrows-v", "Arrows Vertical", 0xf07d, ["Web Application", "Directional"], ["resize"]], ["arrows-h", "Arrows Horizontal", 0xf07e, ["Web Application", "Directional"], ["resize"]], ["bar-chart", "Bar Chart", 0xf080, ["Web Application", "Chart"], ["graph", "analytics", "bar-chart-o"]], ["twitter-square", "Twitter Square", 0xf081, ["Brand"], ["tweet", "social network"]], ["facebook-square", "Facebook Square", 0xf082, ["Brand"], ["social network"]], ["camera-retro", "camera-retro", 0xf083, ["Web Application"], ["photo", "picture", "record"]], ["key", "key", 0xf084, ["Web Application"], ["unlock", "password"]], ["cogs", "cogs", 0xf085, ["Web Application"], ["settings", "gears"]], ["comments", "comments", 0xf086, ["Web Application"], ["speech", "notification", "note", "chat", "bubble", "feedback", "message", "texting", "sms", "conversation"]], ["thumbs-o-up", "Thumbs Up Outlined", 0xf087, ["Web Application", "Hand"], ["like", "approve", "favorite", "agree", "hand"]], ["thumbs-o-down", "Thumbs Down Outlined", 0xf088, ["Web Application", "Hand"], ["dislike", "disapprove", "disagree", "hand"]], ["star-half", "star-half", 0xf089, ["Web Application"], ["award", "achievement", "rating", "score"]], ["heart-o", "Heart Outlined", 0xf08a, ["Web Application", "Medical"], ["love", "like", "favorite"]], ["sign-out", "Sign Out", 0xf08b, ["Web Application"], ["log out", "logout", "leave", "exit", "arrow"]], ["linkedin-square", "LinkedIn Square", 0xf08c, ["Brand"]], ["thumb-tack", "Thumb Tack", 0xf08d, ["Web Application"], ["marker", "pin", "location", "coordinates"]], ["external-link", "External Link", 0xf08e, ["Web Application"], ["open", "new"]], ["sign-in", "Sign In", 0xf090, ["Web Application"], ["enter", "join", "log in", "login", "sign up", "sign in", "signin", "signup", "arrow"]], ["trophy", "trophy", 0xf091, ["Web Application"], ["award", "achievement", "cup", "winner", "game"]], ["github-square", "GitHub Square", 0xf092, ["Brand"], ["octocat"]], ["upload", "Upload", 0xf093, ["Web Application"], ["import"]], ["lemon-o", "Lemon Outlined", 0xf094, ["Web Application"], ["food"]], ["phone", "Phone", 0xf095, ["Web Application"], ["call", "voice", "number", "support", "earphone", "telephone"]], ["square-o", "Square Outlined", 0xf096, ["Web Application", "Form Control"], ["block", "square", "box"]], ["bookmark-o", "Bookmark Outlined", 0xf097, ["Web Application"], ["save"]], ["phone-square", "Phone Square", 0xf098, ["Web Application"], ["call", "voice", "number", "support", "telephone"]], ["twitter", "Twitter", 0xf099, ["Brand"], ["tweet", "social network"]], ["facebook", "Facebook", 0xf09a, ["Brand"], ["social network", "facebook-f"]], ["github", "GitHub", 0xf09b, ["Brand"], ["octocat"]], ["unlock", "unlock", 0xf09c, ["Web Application"], ["protect", "admin", "password", "lock"]], ["credit-card", "credit-card", 0xf09d, ["Web Application", "Payment"], ["money", "buy", "debit", "checkout", "purchase", "payment"]], ["rss", "rss", 0xf09e, ["Web Application"], ["blog", "feed"]], ["hdd-o", "HDD", 0xf0a0, ["Web Application"], ["harddrive", "hard drive", "storage", "save"]], ["bullhorn", "bullhorn", 0xf0a1, ["Web Application"], ["announcement", "share", "broadcast", "louder", "megaphone"]], ["bell", "bell", 0xf0f3, ["Web Application"], ["alert", "reminder", "notification"]], ["certificate", "certificate", 0xf0a3, ["Web Application"], ["badge", "star"]], ["hand-o-right", "Hand Outlined Right", 0xf0a4, ["Directional", "Hand"], ["point", "right", "next", "forward", "finger"]], ["hand-o-left", "Hand Outlined Left", 0xf0a5, ["Directional", "Hand"], ["point", "left", "previous", "back", "finger"]], ["hand-o-up", "Hand Outlined Up", 0xf0a6, ["Directional", "Hand"], ["point", "finger"]], ["hand-o-down", "Hand Outlined Down", 0xf0a7, ["Directional", "Hand"], ["point", "finger"]], ["arrow-circle-left", "Arrow Circle Left", 0xf0a8, ["Directional"], ["previous", "back"]], ["arrow-circle-right", "Arrow Circle Right", 0xf0a9, ["Directional"], ["next", "forward"]], ["arrow-circle-up", "Arrow Circle Up", 0xf0aa, ["Directional"]], ["arrow-circle-down", "Arrow Circle Down", 0xf0ab, ["Directional"], ["download"]], ["globe", "Globe", 0xf0ac, ["Web Application"], ["world", "planet", "map", "place", "travel", "earth", "global", "translate", "all", "language", "localize", "location", "coordinates", "country"]], ["wrench", "Wrench", 0xf0ad, ["Web Application"], ["settings", "fix", "update", "spanner"]], ["tasks", "Tasks", 0xf0ae, ["Web Application"], ["progress", "loading", "downloading", "downloads", "settings"]], ["filter", "Filter", 0xf0b0, ["Web Application"], ["funnel", "options"]], ["briefcase", "Briefcase", 0xf0b1, ["Web Application"], ["work", "business", "office", "luggage", "bag"]], ["arrows-alt", "Arrows Alt", 0xf0b2, ["Video Player", "Directional"], ["expand", "enlarge", "fullscreen", "bigger", "move", "reorder", "resize", "arrow"]], ["users", "Users", 0xf0c0, ["Web Application"], ["people", "profiles", "persons", "group"]], ["link", "Link", 0xf0c1, ["Text Editor"], ["chain", "chain"]], ["cloud", "Cloud", 0xf0c2, ["Web Application"], ["save"]], ["flask", "Flask", 0xf0c3, ["Web Application"], ["science", "beaker", "experimental", "labs"]], ["scissors", "Scissors", 0xf0c4, ["Text Editor"], ["cut"]], ["files-o", "Files Outlined", 0xf0c5, ["Text Editor"], ["duplicate", "clone", "copy", "copy"]], ["paperclip", "Paperclip", 0xf0c6, ["Text Editor"], ["attachment"]], ["floppy-o", "Floppy Outlined", 0xf0c7, ["Text Editor"], ["save"]], ["square", "Square", 0xf0c8, ["Web Application", "Form Control"], ["block", "box"]], ["bars", "Bars", 0xf0c9, ["Web Application"], ["menu", "drag", "reorder", "settings", "list", "ul", "ol", "checklist", "todo", "list", "hamburger", "navicon", "reorder"]], ["list-ul", "list-ul", 0xf0ca, ["Text Editor"], ["ul", "ol", "checklist", "todo", "list"]], ["list-ol", "list-ol", 0xf0cb, ["Text Editor"], ["ul", "ol", "checklist", "list", "todo", "list", "numbers"]], ["strikethrough", "Strikethrough", 0xf0cc, ["Text Editor"]], ["underline", "Underline", 0xf0cd, ["Text Editor"]], ["table", "table", 0xf0ce, ["Text Editor"], ["data", "excel", "spreadsheet"]], ["magic", "magic", 0xf0d0, ["Web Application"], ["wizard", "automatic", "autocomplete"]], ["truck", "truck", 0xf0d1, ["Web Application", "Transportation"], ["shipping"]], ["pinterest", "Pinterest", 0xf0d2, ["Brand"]], ["pinterest-square", "Pinterest Square", 0xf0d3, ["Brand"]], ["google-plus-square", "Google Plus Square", 0xf0d4, ["Brand"], ["social network"]], ["google-plus", "Google Plus", 0xf0d5, ["Brand"], ["social network"]], ["money", "Money", 0xf0d6, ["Web Application", "Currency"], ["cash", "money", "buy", "checkout", "purchase", "payment"]], ["caret-down", "Caret Down", 0xf0d7, ["Directional"], ["more", "dropdown", "menu", "triangle down", "arrow"]], ["caret-up", "Caret Up", 0xf0d8, ["Directional"], ["triangle up", "arrow"]], ["caret-left", "Caret Left", 0xf0d9, ["Directional"], ["previous", "back", "triangle left", "arrow"]], ["caret-right", "Caret Right", 0xf0da, ["Directional"], ["next", "forward", "triangle right", "arrow"]], ["columns", "Columns", 0xf0db, ["Text Editor"], ["split", "panes"]], ["sort", "Sort", 0xf0dc, ["Web Application"], ["order", "unsorted"]], ["sort-desc", "Sort Descending", 0xf0dd, ["Web Application"], ["dropdown", "more", "menu", "arrow", "sort-down"]], ["sort-asc", "Sort Ascending", 0xf0de, ["Web Application"], ["arrow", "sort-up"]], ["envelope", "Envelope", 0xf0e0, ["Web Application"], ["email", "e-mail", "letter", "support", "mail", "message", "notification"]], ["linkedin", "LinkedIn", 0xf0e1, ["Brand"]], ["undo", "Undo", 0xf0e2, ["Text Editor"], ["back", "rotate-left"]], ["gavel", "Gavel", 0xf0e3, ["Web Application"], ["judge", "lawyer", "opinion", "legal"]], ["tachometer", "Tachometer", 0xf0e4, ["Web Application"], ["speedometer", "fast", "dashboard"]], ["comment-o", "comment-o", 0xf0e5, ["Web Application"], ["speech", "notification", "note", "chat", "bubble", "feedback", "message", "texting", "sms", "conversation"]], ["comments-o", "comments-o", 0xf0e6, ["Web Application"], ["speech", "notification", "note", "chat", "bubble", "feedback", "message", "texting", "sms", "conversation"]], ["bolt", "Lightning Bolt", 0xf0e7, ["Web Application"], ["lightning", "weather", "flash"]], ["sitemap", "Sitemap", 0xf0e8, ["Web Application"], ["directory", "hierarchy", "organization"]], ["umbrella", "Umbrella", 0xf0e9, ["Web Application"]], ["clipboard", "Clipboard", 0xf0ea, ["Text Editor"], ["copy", "paste"]], ["lightbulb-o", "Lightbulb Outlined", 0xf0eb, ["Web Application"], ["idea", "inspiration"]], ["exchange", "Exchange", 0xf0ec, ["Web Application", "Directional"], ["transfer", "arrows", "arrow"]], ["cloud-download", "Cloud Download", 0xf0ed, ["Web Application"], ["import"]], ["cloud-upload", "Cloud Upload", 0xf0ee, ["Web Application"], ["import"]], ["user-md", "user-md", 0xf0f0, ["Medical"], ["doctor", "profile", "medical", "nurse"]], ["stethoscope", "Stethoscope", 0xf0f1, ["Medical"]], ["suitcase", "Suitcase", 0xf0f2, ["Web Application"], ["trip", "luggage", "travel", "move", "baggage"]], ["bell-o", "Bell Outlined", 0xf0a2, ["Web Application"], ["alert", "reminder", "notification"]], ["coffee", "Coffee", 0xf0f4, ["Web Application"], ["morning", "mug", "breakfast", "tea", "drink", "cafe"]], ["cutlery", "Cutlery", 0xf0f5, ["Web Application"], ["food", "restaurant", "spoon", "knife", "dinner", "eat"]], ["file-text-o", "File Text Outlined", 0xf0f6, ["Text Editor", "File Type"], ["new", "page", "pdf", "document"]], ["building-o", "Building Outlined", 0xf0f7, ["Web Application"], ["work", "business", "apartment", "office", "company"]], ["hospital-o", "hospital Outlined", 0xf0f8, ["Medical"], ["building"]], ["ambulance", "ambulance", 0xf0f9, ["Medical", "Transportation"], ["vehicle", "support", "help"]], ["medkit", "medkit", 0xf0fa, ["Medical"], ["first aid", "firstaid", "help", "support", "health"]], ["fighter-jet", "fighter-jet", 0xf0fb, ["Web Application", "Transportation"], ["fly", "plane", "airplane", "quick", "fast", "travel"]], ["beer", "beer", 0xf0fc, ["Web Application"], ["alcohol", "stein", "drink", "mug", "bar", "liquor"]], ["h-square", "H Square", 0xf0fd, ["Medical"], ["hospital", "hotel"]], ["plus-square", "Plus Square", 0xf0fe, ["Medical", "Web Application", "Form Control"], ["add", "new", "create", "expand"]], ["angle-double-left", "Angle Double Left", 0xf100, ["Directional"], ["laquo", "quote", "previous", "back", "arrows"]], ["angle-double-right", "Angle Double Right", 0xf101, ["Directional"], ["raquo", "quote", "next", "forward", "arrows"]], ["angle-double-up", "Angle Double Up", 0xf102, ["Directional"], ["arrows"]], ["angle-double-down", "Angle Double Down", 0xf103, ["Directional"], ["arrows"]], ["angle-left", "angle-left", 0xf104, ["Directional"], ["previous", "back", "arrow"]], ["angle-right", "angle-right", 0xf105, ["Directional"], ["next", "forward", "arrow"]], ["angle-up", "angle-up", 0xf106, ["Directional"], ["arrow"]], ["angle-down", "angle-down", 0xf107, ["Directional"], ["arrow"]], ["desktop", "Desktop", 0xf108, ["Web Application"], ["monitor", "screen", "desktop", "computer", "demo", "device"]], ["laptop", "Laptop", 0xf109, ["Web Application"], ["demo", "computer", "device"]], ["tablet", "tablet", 0xf10a, ["Web Application"], ["ipad", "device"]], ["mobile", "Mobile Phone", 0xf10b, ["Web Application"], ["cell phone", "cellphone", "text", "call", "iphone", "number", "telephone", "mobile-phone"]], ["circle-o", "Circle Outlined", 0xf10c, ["Web Application", "Form Control"]], ["quote-left", "quote-left", 0xf10d, ["Web Application"]], ["quote-right", "quote-right", 0xf10e, ["Web Application"]], ["spinner", "Spinner", 0xf110, ["Web Application", "Spinner"], ["loading", "progress"]], ["circle", "Circle", 0xf111, ["Web Application", "Form Control"], ["dot", "notification"]], ["reply", "Reply", 0xf112, ["Web Application"], ["mail-reply"]], ["github-alt", "GitHub Alt", 0xf113, ["Brand"], ["octocat"]], ["folder-o", "Folder Outlined", 0xf114, ["Web Application"]], ["folder-open-o", "Folder Open Outlined", 0xf115, ["Web Application"]], ["smile-o", "Smile Outlined", 0xf118, ["Web Application"], ["face", "emoticon", "happy", "approve", "satisfied", "rating"]], ["frown-o", "Frown Outlined", 0xf119, ["Web Application"], ["face", "emoticon", "sad", "disapprove", "rating"]], ["meh-o", "Meh Outlined", 0xf11a, ["Web Application"], ["face", "emoticon", "rating", "neutral"]], ["gamepad", "Gamepad", 0xf11b, ["Web Application"], ["controller"]], ["keyboard-o", "Keyboard Outlined", 0xf11c, ["Web Application"], ["type", "input"]], ["flag-o", "Flag Outlined", 0xf11d, ["Web Application"], ["report", "notification"]], ["flag-checkered", "flag-checkered", 0xf11e, ["Web Application"], ["report", "notification", "notify"]], ["terminal", "Terminal", 0xf120, ["Web Application"], ["command", "prompt", "code"]], ["code", "Code", 0xf121, ["Web Application"], ["html", "brackets"]], ["reply-all", "reply-all", 0xf122, ["Web Application"], ["mail-reply-all"]], ["star-half-o", "Star Half Outlined", 0xf123, ["Web Application"], ["award", "achievement", "rating", "score", "star-half-empty", "star-half-full"]], ["location-arrow", "location-arrow", 0xf124, ["Web Application"], ["map", "coordinates", "location", "address", "place", "where"]], ["crop", "crop", 0xf125, ["Web Application"]], ["code-fork", "code-fork", 0xf126, ["Web Application"], ["git", "fork", "vcs", "svn", "github", "rebase", "version", "merge"]], ["chain-broken", "Chain Broken", 0xf127, ["Text Editor"], ["remove", "unlink"]], ["question", "Question", 0xf128, ["Web Application"], ["help", "information", "unknown", "support"]], ["info", "Info", 0xf129, ["Web Application"], ["help", "information", "more", "details"]], ["exclamation", "exclamation", 0xf12a, ["Web Application"], ["warning", "error", "problem", "notification", "notify", "alert"]], ["superscript", "superscript", 0xf12b, ["Text Editor"], ["exponential"]], ["subscript", "subscript", 0xf12c, ["Text Editor"]], ["eraser", "eraser", 0xf12d, ["Text Editor", "Web Application"], ["remove", "delete"]], ["puzzle-piece", "Puzzle Piece", 0xf12e, ["Web Application"], ["addon", "add-on", "section"]], ["microphone", "microphone", 0xf130, ["Web Application"], ["record", "voice", "sound"]], ["microphone-slash", "Microphone Slash", 0xf131, ["Web Application"], ["record", "voice", "sound", "mute"]], ["shield", "shield", 0xf132, ["Web Application"], ["award", "achievement", "security", "winner"]], ["calendar-o", "calendar-o", 0xf133, ["Web Application"], ["date", "time", "when", "event"]], ["fire-extinguisher", "fire-extinguisher", 0xf134, ["Web Application"]], ["rocket", "rocket", 0xf135, ["Web Application", "Transportation"], ["app"]], ["maxcdn", "MaxCDN", 0xf136, ["Brand"]], ["chevron-circle-left", "Chevron Circle Left", 0xf137, ["Directional"], ["previous", "back", "arrow"]], ["chevron-circle-right", "Chevron Circle Right", 0xf138, ["Directional"], ["next", "forward", "arrow"]], ["chevron-circle-up", "Chevron Circle Up", 0xf139, ["Directional"], ["arrow"]], ["chevron-circle-down", "Chevron Circle Down", 0xf13a, ["Directional"], ["more", "dropdown", "menu", "arrow"]], ["html5", "HTML 5 Logo", 0xf13b, ["Brand"]], ["css3", "CSS 3 Logo", 0xf13c, ["Brand"], ["code"]], ["anchor", "Anchor", 0xf13d, ["Web Application"], ["link"]], ["unlock-alt", "Unlock Alt", 0xf13e, ["Web Application"], ["protect", "admin", "password", "lock"]], ["bullseye", "Bullseye", 0xf140, ["Web Application"], ["target"]], ["ellipsis-h", "Ellipsis Horizontal", 0xf141, ["Web Application"], ["dots"]], ["ellipsis-v", "Ellipsis Vertical", 0xf142, ["Web Application"], ["dots"]], ["rss-square", "RSS Square", 0xf143, ["Web Application"], ["feed", "blog"]], ["play-circle", "Play Circle", 0xf144, ["Video Player"], ["start", "playing"]], ["ticket", "Ticket", 0xf145, ["Web Application"], ["movie", "pass", "support"]], ["minus-square", "Minus Square", 0xf146, ["Web Application", "Form Control"], ["hide", "minify", "delete", "remove", "trash", "hide", "collapse"]], ["minus-square-o", "Minus Square Outlined", 0xf147, ["Web Application", "Form Control"], ["hide", "minify", "delete", "remove", "trash", "hide", "collapse"]], ["level-up", "Level Up", 0xf148, ["Web Application"], ["arrow"]], ["level-down", "Level Down", 0xf149, ["Web Application"], ["arrow"]], ["check-square", "Check Square", 0xf14a, ["Web Application", "Form Control"], ["checkmark", "done", "todo", "agree", "accept", "confirm", "ok"]], ["pencil-square", "Pencil Square", 0xf14b, ["Web Application"], ["write", "edit", "update"]], ["external-link-square", "External Link Square", 0xf14c, ["Web Application"], ["open", "new"]], ["share-square", "Share Square", 0xf14d, ["Web Application"], ["social", "send"]], ["compass", "Compass", 0xf14e, ["Web Application"], ["safari", "directory", "menu", "location"]], ["caret-square-o-down", "Caret Square Outlined Down", 0xf150, ["Web Application", "Directional"], ["more", "dropdown", "menu", "toggle-down"]], ["caret-square-o-up", "Caret Square Outlined Up", 0xf151, ["Web Application", "Directional"], ["toggle-up"]], ["caret-square-o-right", "Caret Square Outlined Right", 0xf152, ["Web Application", "Directional"], ["next", "forward", "toggle-right"]], ["eur", "Euro (EUR)", 0xf153, ["Currency"], ["euro"]], ["gbp", "GBP", 0xf154, ["Currency"]], ["usd", "US Dollar", 0xf155, ["Currency"], ["dollar"]], ["inr", "Indian Rupee (INR)", 0xf156, ["Currency"], ["rupee"]], ["jpy", "Japanese Yen (JPY)", 0xf157, ["Currency"], ["cny", "rmb", "yen"]], ["rub", "Russian Ruble (RUB)", 0xf158, ["Currency"], ["ruble", "rouble"]], ["krw", "Korean Won (KRW)", 0xf159, ["Currency"], ["won"]], ["btc", "Bitcoin (BTC)", 0xf15a, ["Currency", "Brand"], ["bitcoin"]], ["file", "File", 0xf15b, ["Text Editor", "File Type"], ["new", "page", "pdf", "document"]], ["file-text", "File Text", 0xf15c, ["Text Editor", "File Type"], ["new", "page", "pdf", "document"]], ["sort-alpha-asc", "Sort Alpha Ascending", 0xf15d, ["Web Application"]], ["sort-alpha-desc", "Sort Alpha Descending", 0xf15e, ["Web Application"]], ["sort-amount-asc", "Sort Amount Ascending", 0xf160, ["Web Application"]], ["sort-amount-desc", "Sort Amount Descending", 0xf161, ["Web Application"]], ["sort-numeric-asc", "Sort Numeric Ascending", 0xf162, ["Web Application"], ["numbers"]], ["sort-numeric-desc", "Sort Numeric Descending", 0xf163, ["Web Application"], ["numbers"]], ["thumbs-up", "thumbs-up", 0xf164, ["Web Application", "Hand"], ["like", "favorite", "approve", "agree", "hand"]], ["thumbs-down", "thumbs-down", 0xf165, ["Web Application", "Hand"], ["dislike", "disapprove", "disagree", "hand"]], ["youtube-square", "YouTube Square", 0xf166, ["Brand"], ["video", "film"]], ["youtube", "YouTube", 0xf167, ["Brand"], ["video", "film"]], ["xing", "Xing", 0xf168, ["Brand"]], ["xing-square", "Xing Square", 0xf169, ["Brand"]], ["youtube-play", "YouTube Play", 0xf16a, ["Brand", "Video Player"], ["start", "playing"]], ["dropbox", "Dropbox", 0xf16b, ["Brand"]], ["stack-overflow", "Stack Overflow", 0xf16c, ["Brand"]], ["instagram", "Instagram", 0xf16d, ["Brand"]], ["flickr", "Flickr", 0xf16e, ["Brand"]], ["adn", "App.net", 0xf170, ["Brand"]], ["bitbucket", "Bitbucket", 0xf171, ["Brand"], ["git"]], ["bitbucket-square", "Bitbucket Square", 0xf172, ["Brand"], ["git"]], ["tumblr", "Tumblr", 0xf173, ["Brand"]], ["tumblr-square", "Tumblr Square", 0xf174, ["Brand"]], ["long-arrow-down", "Long Arrow Down", 0xf175, ["Directional"]], ["long-arrow-up", "Long Arrow Up", 0xf176, ["Directional"]], ["long-arrow-left", "Long Arrow Left", 0xf177, ["Directional"], ["previous", "back"]], ["long-arrow-right", "Long Arrow Right", 0xf178, ["Directional"]], ["apple", "Apple", 0xf179, ["Brand"], ["osx", "food"]], ["windows", "Windows", 0xf17a, ["Brand"], ["microsoft"]], ["android", "Android", 0xf17b, ["Brand"], ["robot"]], ["linux", "Linux", 0xf17c, ["Brand"], ["tux"]], ["dribbble", "Dribbble", 0xf17d, ["Brand"]], ["skype", "Skype", 0xf17e, ["Brand"]], ["foursquare", "Foursquare", 0xf180, ["Brand"]], ["trello", "Trello", 0xf181, ["Brand"]], ["female", "Female", 0xf182, ["Web Application"], ["woman", "user", "person", "profile"]], ["male", "Male", 0xf183, ["Web Application"], ["man", "user", "person", "profile"]], ["gratipay", "Gratipay (Gittip)", 0xf184, ["Brand"], ["heart", "like", "favorite", "love", "gittip"]], ["sun-o", "Sun Outlined", 0xf185, ["Web Application"], ["weather", "contrast", "lighter", "brighten", "day"]], ["moon-o", "Moon Outlined", 0xf186, ["Web Application"], ["night", "darker", "contrast"]], ["archive", "Archive", 0xf187, ["Web Application"], ["box", "storage"]], ["bug", "Bug", 0xf188, ["Web Application"], ["report", "insect"]], ["vk", "VK", 0xf189, ["Brand"]], ["weibo", "Weibo", 0xf18a, ["Brand"]], ["renren", "Renren", 0xf18b, ["Brand"]], ["pagelines", "Pagelines", 0xf18c, ["Brand"], ["leaf", "leaves", "tree", "plant", "eco", "nature"]], ["stack-exchange", "Stack Exchange", 0xf18d, ["Brand"]], ["arrow-circle-o-right", "Arrow Circle Outlined Right", 0xf18e, ["Directional"], ["next", "forward"]], ["arrow-circle-o-left", "Arrow Circle Outlined Left", 0xf190, ["Directional"], ["previous", "back"]], ["caret-square-o-left", "Caret Square Outlined Left", 0xf191, ["Web Application", "Directional"], ["previous", "back", "toggle-left"]], ["dot-circle-o", "Dot Circle Outlined", 0xf192, ["Web Application", "Form Control"], ["target", "bullseye", "notification"]], ["wheelchair", "Wheelchair", 0xf193, ["Web Application", "Medical", "Transportation", "Accessibility"], ["handicap", "person"]], ["vimeo-square", "Vimeo Square", 0xf194, ["Brand"]], ["try", "Turkish Lira (TRY)", 0xf195, ["Currency"], ["turkish-lira"]], ["plus-square-o", "Plus Square Outlined", 0xf196, ["Web Application", "Form Control"], ["add", "new", "create", "expand"]], ["space-shuttle", "Space Shuttle", 0xf197, ["Web Application", "Transportation"]], ["slack", "Slack Logo", 0xf198, ["Brand"], ["hashtag", "anchor", "hash"]], ["envelope-square", "Envelope Square", 0xf199, ["Web Application"], ["email", "e-mail", "letter", "support", "mail", "message", "notification"]], ["wordpress", "WordPress Logo", 0xf19a, ["Brand"]], ["openid", "OpenID", 0xf19b, ["Brand"]], ["university", "University", 0xf19c, ["Web Application"], ["institution", "bank"]], ["graduation-cap", "Graduation Cap", 0xf19d, ["Web Application"], ["learning", "school", "student", "mortar-board"]], ["yahoo", "Yahoo Logo", 0xf19e, ["Brand"]], ["google", "Google Logo", 0xf1a0, ["Brand"]], ["reddit", "reddit Logo", 0xf1a1, ["Brand"]], ["reddit-square", "reddit Square", 0xf1a2, ["Brand"]], ["stumbleupon-circle", "StumbleUpon Circle", 0xf1a3, ["Brand"]], ["stumbleupon", "StumbleUpon Logo", 0xf1a4, ["Brand"]], ["delicious", "Delicious Logo", 0xf1a5, ["Brand"]], ["digg", "Digg Logo", 0xf1a6, ["Brand"]], ["pied-piper-pp", "Pied Piper PP Logo (Old)", 0xf1a7, ["Brand"]], ["pied-piper-alt", "Pied Piper Alternate Logo", 0xf1a8, ["Brand"]], ["drupal", "Drupal Logo", 0xf1a9, ["Brand"]], ["joomla", "Joomla Logo", 0xf1aa, ["Brand"]], ["language", "Language", 0xf1ab, ["Web Application"]], ["fax", "Fax", 0xf1ac, ["Web Application"]], ["building", "Building", 0xf1ad, ["Web Application"], ["work", "business", "apartment", "office", "company"]], ["child", "Child", 0xf1ae, ["Web Application"]], ["paw", "Paw", 0xf1b0, ["Web Application"], ["pet"]], ["spoon", "spoon", 0xf1b1, ["Web Application"]], ["cube", "Cube", 0xf1b2, ["Web Application"]], ["cubes", "Cubes", 0xf1b3, ["Web Application"]], ["behance", "Behance", 0xf1b4, ["Brand"]], ["behance-square", "Behance Square", 0xf1b5, ["Brand"]], ["steam", "Steam", 0xf1b6, ["Brand"]], ["steam-square", "Steam Square", 0xf1b7, ["Brand"]], ["recycle", "Recycle", 0xf1b8, ["Web Application"]], ["car", "Car", 0xf1b9, ["Web Application", "Transportation"], ["vehicle", "automobile"]], ["taxi", "Taxi", 0xf1ba, ["Web Application", "Transportation"], ["vehicle", "cab"]], ["tree", "Tree", 0xf1bb, ["Web Application"]], ["spotify", "Spotify", 0xf1bc, ["Brand"]], ["deviantart", "deviantART", 0xf1bd, ["Brand"]], ["soundcloud", "SoundCloud", 0xf1be, ["Brand"]], ["database", "Database", 0xf1c0, ["Web Application"]], ["file-pdf-o", "PDF File Outlined", 0xf1c1, ["Web Application", "File Type"]], ["file-word-o", "Word File Outlined", 0xf1c2, ["Web Application", "File Type"]], ["file-excel-o", "Excel File Outlined", 0xf1c3, ["Web Application", "File Type"]], ["file-powerpoint-o", "Powerpoint File Outlined", 0xf1c4, ["Web Application", "File Type"]], ["file-image-o", "Image File Outlined", 0xf1c5, ["Web Application", "File Type"], ["file-photo-o", "file-picture-o"]], ["file-archive-o", "Archive File Outlined", 0xf1c6, ["Web Application", "File Type"], ["file-zip-o"]], ["file-audio-o", "Audio File Outlined", 0xf1c7, ["Web Application", "File Type"], ["file-sound-o"]], ["file-video-o", "Video File Outlined", 0xf1c8, ["Web Application", "File Type"], ["file-movie-o"]], ["file-code-o", "Code File Outlined", 0xf1c9, ["Web Application", "File Type"]], ["vine", "Vine", 0xf1ca, ["Brand"]], ["codepen", "Codepen", 0xf1cb, ["Brand"]], ["jsfiddle", "jsFiddle", 0xf1cc, ["Brand"]], ["life-ring", "Life Ring", 0xf1cd, ["Web Application"], ["life-bouy", "life-buoy", "life-saver", "support"]], ["circle-o-notch", "Circle Outlined Notched", 0xf1ce, ["Web Application", "Spinner"]], ["rebel", "Rebel Alliance", 0xf1d0, ["Brand"], ["ra", "resistance"]], ["empire", "Galactic Empire", 0xf1d1, ["Brand"], ["ge"]], ["git-square", "Git Square", 0xf1d2, ["Brand"]], ["git", "Git", 0xf1d3, ["Brand"]], ["hacker-news", "Hacker News", 0xf1d4, ["Brand"], ["y-combinator-square", "yc-square"]], ["tencent-weibo", "Tencent Weibo", 0xf1d5, ["Brand"]], ["qq", "QQ", 0xf1d6, ["Brand"]], ["weixin", "Weixin (WeChat)", 0xf1d7, ["Brand"], ["wechat"]], ["paper-plane", "Paper Plane", 0xf1d8, ["Web Application"], ["send"]], ["paper-plane-o", "Paper Plane Outlined", 0xf1d9, ["Web Application"], ["send-o"]], ["history", "History", 0xf1da, ["Web Application"]], ["circle-thin", "Circle Outlined Thin", 0xf1db, ["Web Application"]], ["header", "header", 0xf1dc, ["Text Editor"], ["heading"]], ["paragraph", "paragraph", 0xf1dd, ["Text Editor"]], ["sliders", "Sliders", 0xf1de, ["Web Application"], ["settings"]], ["share-alt", "Share Alt", 0xf1e0, ["Web Application", "Brand"]], ["share-alt-square", "Share Alt Square", 0xf1e1, ["Web Application", "Brand"]], ["bomb", "Bomb", 0xf1e2, ["Web Application"]], ["futbol-o", "Futbol Outlined", 0xf1e3, ["Web Application"], ["soccer-ball-o"]], ["tty", "TTY", 0xf1e4, ["Web Application", "Accessibility"]], ["binoculars", "Binoculars", 0xf1e5, ["Web Application"]], ["plug", "Plug", 0xf1e6, ["Web Application"], ["power", "connect"]], ["slideshare", "Slideshare", 0xf1e7, ["Brand"]], ["twitch", "Twitch", 0xf1e8, ["Brand"]], ["yelp", "Yelp", 0xf1e9, ["Brand"]], ["newspaper-o", "Newspaper Outlined", 0xf1ea, ["Web Application"], ["press"]], ["wifi", "WiFi", 0xf1eb, ["Web Application"]], ["calculator", "Calculator", 0xf1ec, ["Web Application"]], ["paypal", "Paypal", 0xf1ed, ["Brand", "Payment"]], ["google-wallet", "Google Wallet", 0xf1ee, ["Brand", "Payment"]], ["cc-visa", "Visa Credit Card", 0xf1f0, ["Brand", "Payment"]], ["cc-mastercard", "MasterCard Credit Card", 0xf1f1, ["Brand", "Payment"]], ["cc-discover", "Discover Credit Card", 0xf1f2, ["Brand", "Payment"]], ["cc-amex", "American Express Credit Card", 0xf1f3, ["Brand", "Payment"], ["amex"]], ["cc-paypal", "Paypal Credit Card", 0xf1f4, ["Brand", "Payment"]], ["cc-stripe", "Stripe Credit Card", 0xf1f5, ["Brand", "Payment"]], ["bell-slash", "Bell Slash", 0xf1f6, ["Web Application"]], ["bell-slash-o", "Bell Slash Outlined", 0xf1f7, ["Web Application"]], ["trash", "Trash", 0xf1f8, ["Web Application"], ["garbage", "delete", "remove", "hide"]], ["copyright", "Copyright", 0xf1f9, ["Web Application"]], ["at", "At", 0xf1fa, ["Web Application"]], ["eyedropper", "Eyedropper", 0xf1fb, ["Web Application"]], ["paint-brush", "Paint Brush", 0xf1fc, ["Web Application"]], ["birthday-cake", "Birthday Cake", 0xf1fd, ["Web Application"]], ["area-chart", "Area Chart", 0xf1fe, ["Web Application", "Chart"], ["graph", "analytics"]], ["pie-chart", "Pie Chart", 0xf200, ["Web Application", "Chart"], ["graph", "analytics"]], ["line-chart", "Line Chart", 0xf201, ["Web Application", "Chart"], ["graph", "analytics"]], ["lastfm", "last.fm", 0xf202, ["Brand"]], ["lastfm-square", "last.fm Square", 0xf203, ["Brand"]], ["toggle-off", "Toggle Off", 0xf204, ["Web Application"]], ["toggle-on", "Toggle On", 0xf205, ["Web Application"]], ["bicycle", "Bicycle", 0xf206, ["Web Application", "Transportation"], ["vehicle", "bike"]], ["bus", "Bus", 0xf207, ["Web Application", "Transportation"], ["vehicle"]], ["ioxhost", "ioxhost", 0xf208, ["Brand"]], ["angellist", "AngelList", 0xf209, ["Brand"]], ["cc", "Closed Captions", 0xf20a, ["Web Application", "Accessibility"]], ["ils", "Shekel (ILS)", 0xf20b, ["Currency"], ["shekel", "sheqel"]], ["meanpath", "meanpath", 0xf20c, ["Brand"]], ["buysellads", "BuySellAds", 0xf20d, ["Brand"]], ["connectdevelop", "Connect Develop", 0xf20e, ["Brand"]], ["dashcube", "DashCube", 0xf210, ["Brand"]], ["forumbee", "Forumbee", 0xf211, ["Brand"]], ["leanpub", "Leanpub", 0xf212, ["Brand"]], ["sellsy", "Sellsy", 0xf213, ["Brand"]], ["shirtsinbulk", "Shirts in Bulk", 0xf214, ["Brand"]], ["simplybuilt", "SimplyBuilt", 0xf215, ["Brand"]], ["skyatlas", "skyatlas", 0xf216, ["Brand"]], ["cart-plus", "Add to Shopping Cart", 0xf217, ["Web Application"], ["add", "shopping"]], ["cart-arrow-down", "Shopping Cart Arrow Down", 0xf218, ["Web Application"], ["shopping"]], ["diamond", "Diamond", 0xf219, ["Web Application"], ["gem", "gemstone"]], ["ship", "Ship", 0xf21a, ["Web Application", "Transportation"], ["boat", "sea"]], ["user-secret", "User Secret", 0xf21b, ["Web Application"], ["whisper", "spy", "incognito", "privacy"]], ["motorcycle", "Motorcycle", 0xf21c, ["Web Application", "Transportation"], ["vehicle", "bike"]], ["street-view", "Street View", 0xf21d, ["Web Application"], ["map"]], ["heartbeat", "Heartbeat", 0xf21e, ["Web Application", "Medical"], ["ekg"]], ["venus", "Venus", 0xf221, ["Gender"], ["female"]], ["mars", "Mars", 0xf222, ["Gender"], ["male"]], ["mercury", "Mercury", 0xf223, ["Gender"], ["transgender"]], ["transgender", "Transgender", 0xf224, ["Gender"], ["intersex"]], ["transgender-alt", "Transgender Alt", 0xf225, ["Gender"]], ["venus-double", "Venus Double", 0xf226, ["Gender"]], ["mars-double", "Mars Double", 0xf227, ["Gender"]], ["venus-mars", "Venus Mars", 0xf228, ["Gender"]], ["mars-stroke", "Mars Stroke", 0xf229, ["Gender"]], ["mars-stroke-v", "Mars Stroke Vertical", 0xf22a, ["Gender"]], ["mars-stroke-h", "Mars Stroke Horizontal", 0xf22b, ["Gender"]], ["neuter", "Neuter", 0xf22c, ["Gender"]], ["genderless", "Genderless", 0xf22d, ["Gender"]], ["facebook-official", "Facebook Official", 0xf230, ["Brand"]], ["pinterest-p", "Pinterest P", 0xf231, ["Brand"]], ["whatsapp", "What's App", 0xf232, ["Brand"]], ["server", "Server", 0xf233, ["Web Application"]], ["user-plus", "Add User", 0xf234, ["Web Application"], ["sign up", "signup"]], ["user-times", "Remove User", 0xf235, ["Web Application"]], ["bed", "Bed", 0xf236, ["Web Application"], ["travel", "hotel"]], ["viacoin", "Viacoin", 0xf237, ["Brand"]], ["train", "Train", 0xf238, ["Transportation"]], ["subway", "Subway", 0xf239, ["Transportation"]], ["medium", "Medium", 0xf23a, ["Brand"]], ["y-combinator", "Y Combinator", 0xf23b, ["Brand"], ["yc"]], ["optin-monster", "Optin Monster", 0xf23c, ["Brand"]], ["opencart", "OpenCart", 0xf23d, ["Brand"]], ["expeditedssl", "ExpeditedSSL", 0xf23e, ["Brand"]], ["battery-full", "Battery Full", 0xf240, ["Web Application"], ["power", "battery-4", "battery"]], ["battery-three-quarters", "Battery 3/4 Full", 0xf241, ["Web Application"], ["power", "battery-3"]], ["battery-half", "Battery 1/2 Full", 0xf242, ["Web Application"], ["power", "battery-2"]], ["battery-quarter", "Battery 1/4 Full", 0xf243, ["Web Application"], ["power", "battery-1"]], ["battery-empty", "Battery Empty", 0xf244, ["Web Application"], ["power", "battery-0"]], ["mouse-pointer", "Mouse Pointer", 0xf245, ["Web Application"]], ["i-cursor", "I Beam Cursor", 0xf246, ["Web Application"]], ["object-group", "Object Group", 0xf247, ["Web Application"]], ["object-ungroup", "Object Ungroup", 0xf248, ["Web Application"]], ["sticky-note", "Sticky Note", 0xf249, ["Web Application"]], ["sticky-note-o", "Sticky Note Outlined", 0xf24a, ["Web Application"]], ["cc-jcb", "JCB Credit Card", 0xf24b, ["Brand", "Payment"]], ["cc-diners-club", "Diner's Club Credit Card", 0xf24c, ["Brand", "Payment"]], ["clone", "Clone", 0xf24d, ["Web Application"], ["copy"]], ["balance-scale", "Balance Scale", 0xf24e, ["Web Application"]], ["hourglass-o", "Hourglass Outlined", 0xf250, ["Web Application"]], ["hourglass-start", "Hourglass Start", 0xf251, ["Web Application"], ["hourglass-1"]], ["hourglass-half", "Hourglass Half", 0xf252, ["Web Application"], ["hourglass-2"]], ["hourglass-end", "Hourglass End", 0xf253, ["Web Application"], ["hourglass-3"]], ["hourglass", "Hourglass", 0xf254, ["Web Application"]], ["hand-rock-o", "Rock (Hand)", 0xf255, ["Web Application", "Hand"], ["hand-grab-o"]], ["hand-paper-o", "Paper (Hand)", 0xf256, ["Web Application", "Hand"], ["stop", "hand-stop-o"]], ["hand-scissors-o", "Scissors (Hand)", 0xf257, ["Web Application", "Hand"]], ["hand-lizard-o", "Lizard (Hand)", 0xf258, ["Web Application", "Hand"]], ["hand-spock-o", "Spock (Hand)", 0xf259, ["Web Application", "Hand"]], ["hand-pointer-o", "Hand Pointer", 0xf25a, ["Web Application", "Hand"]], ["hand-peace-o", "Hand Peace", 0xf25b, ["Web Application", "Hand"]], ["trademark", "Trademark", 0xf25c, ["Web Application"]], ["registered", "Registered Trademark", 0xf25d, ["Web Application"]], ["creative-commons", "Creative Commons", 0xf25e, ["Web Application"]], ["gg", "GG Currency", 0xf260, ["Currency", "Brand"]], ["gg-circle", "GG Currency Circle", 0xf261, ["Currency", "Brand"]], ["tripadvisor", "TripAdvisor", 0xf262, ["Brand"]], ["odnoklassniki", "Odnoklassniki", 0xf263, ["Brand"]], ["odnoklassniki-square", "Odnoklassniki Square", 0xf264, ["Brand"]], ["get-pocket", "Get Pocket", 0xf265, ["Brand"]], ["wikipedia-w", "Wikipedia W", 0xf266, ["Brand"]], ["safari", "Safari", 0xf267, ["Brand"], ["browser"]], ["chrome", "Chrome", 0xf268, ["Brand"], ["browser"]], ["firefox", "Firefox", 0xf269, ["Brand"], ["browser"]], ["opera", "Opera", 0xf26a, ["Brand"]], ["internet-explorer", "Internet-explorer", 0xf26b, ["Brand"], ["browser", "ie"]], ["television", "Television", 0xf26c, ["Web Application"], ["display", "computer", "monitor", "tv"]], ["contao", "Contao", 0xf26d, ["Brand"]], ["500px", "500px", 0xf26e, ["Brand"]], ["amazon", "Amazon", 0xf270, ["Brand"]], ["calendar-plus-o", "Calendar Plus Outlined", 0xf271, ["Web Application"]], ["calendar-minus-o", "Calendar Minus Outlined", 0xf272, ["Web Application"]], ["calendar-times-o", "Calendar Times Outlined", 0xf273, ["Web Application"]], ["calendar-check-o", "Calendar Check Outlined", 0xf274, ["Web Application"], ["ok"]], ["industry", "Industry", 0xf275, ["Web Application"], ["factory"]], ["map-pin", "Map Pin", 0xf276, ["Web Application"]], ["map-signs", "Map Signs", 0xf277, ["Web Application"]], ["map-o", "Map Outlined", 0xf278, ["Web Application"]], ["map", "Map", 0xf279, ["Web Application"]], ["commenting", "Commenting", 0xf27a, ["Web Application"], ["speech", "notification", "note", "chat", "bubble", "feedback", "message", "texting", "sms", "conversation"]], ["commenting-o", "Commenting Outlined", 0xf27b, ["Web Application"], ["speech", "notification", "note", "chat", "bubble", "feedback", "message", "texting", "sms", "conversation"]], ["houzz", "Houzz", 0xf27c, ["Brand"]], ["vimeo", "Vimeo", 0xf27d, ["Brand"]], ["black-tie", "Font Awesome Black Tie", 0xf27e, ["Brand"]], ["fonticons", "Fonticons", 0xf280, ["Brand"]], ["reddit-alien", "reddit Alien", 0xf281, ["Brand"]], ["edge", "Edge Browser", 0xf282, ["Brand"], ["browser", "ie"]], ["credit-card-alt", "Credit Card", 0xf283, ["Payment", "Web Application"], ["money", "buy", "debit", "checkout", "purchase", "payment", "credit card"]], ["codiepie", "Codie Pie", 0xf284, ["Brand"]], ["modx", "MODX", 0xf285, ["Brand"]], ["fort-awesome", "Fort Awesome", 0xf286, ["Brand"]], ["usb", "USB", 0xf287, ["Brand"]], ["product-hunt", "Product Hunt", 0xf288, ["Brand"]], ["mixcloud", "Mixcloud", 0xf289, ["Brand"]], ["scribd", "Scribd", 0xf28a, ["Brand"]], ["pause-circle", "Pause Circle", 0xf28b, ["Video Player"]], ["pause-circle-o", "Pause Circle Outlined", 0xf28c, ["Video Player"]], ["stop-circle", "Stop Circle", 0xf28d, ["Video Player"]], ["stop-circle-o", "Stop Circle Outlined", 0xf28e, ["Video Player"]], ["shopping-bag", "Shopping Bag", 0xf290, ["Web Application"]], ["shopping-basket", "Shopping Basket", 0xf291, ["Web Application"]], ["hashtag", "Hashtag", 0xf292, ["Web Application"]], ["bluetooth", "Bluetooth", 0xf293, ["Web Application", "Brand"]], ["bluetooth-b", "Bluetooth", 0xf294, ["Web Application", "Brand"]], ["percent", "Percent", 0xf295, ["Web Application"]], ["gitlab", "GitLab", 0xf296, ["Brand"]], ["wpbeginner", "WPBeginner", 0xf297, ["Brand"]], ["wpforms", "WPForms", 0xf298, ["Brand"]], ["envira", "Envira Gallery", 0xf299, ["Brand"], ["leaf"]], ["universal-access", "Universal Access", 0xf29a, ["Web Application", "Accessibility"]], ["wheelchair-alt", "Wheelchair Alt", 0xf29b, ["Web Application", "Medical", "Transportation", "Accessibility"], ["handicap", "person"]], ["question-circle-o", "Question Circle Outlined", 0xf29c, ["Web Application", "Accessibility"]], ["blind", "Blind", 0xf29d, ["Web Application", "Accessibility"]], ["audio-description", "Audio Description", 0xf29e, ["Web Application", "Accessibility"]], ["volume-control-phone", "Volume Control Phone", 0xf2a0, ["Web Application", "Accessibility"], ["telephone"]], ["braille", "Braille", 0xf2a1, ["Web Application", "Accessibility"]], ["assistive-listening-systems", "Assistive Listening Systems", 0xf2a2, ["Web Application", "Accessibility"]], ["american-sign-language-interpreting", "American Sign Language Interpreting", 0xf2a3, ["Web Application", "Accessibility"], ["asl-interpreting"]], ["deaf", "Deaf", 0xf2a4, ["Web Application", "Accessibility"], ["deafness", "hard-of-hearing"]], ["glide", "Glide", 0xf2a5, ["Brand"]], ["glide-g", "Glide G", 0xf2a6, ["Brand"]], ["sign-language", "Sign Language", 0xf2a7, ["Web Application", "Accessibility"], ["signing"]], ["low-vision", "Low Vision", 0xf2a8, ["Web Application", "Accessibility"]], ["viadeo", "Viadeo", 0xf2a9, ["Brand"]], ["viadeo-square", "Viadeo Square", 0xf2aa, ["Brand"]], ["snapchat", "Snapchat", 0xf2ab, ["Brand"]], ["snapchat-ghost", "Snapchat Ghost", 0xf2ac, ["Brand"]], ["snapchat-square", "Snapchat Square", 0xf2ad, ["Brand"]], ["pied-piper", "Pied Piper Logo", 0xf2ae, ["Brand"]], ["first-order", "First Order", 0xf2b0, ["Brand"]], ["yoast", "Yoast", 0xf2b1, ["Brand"]], ["themeisle", "ThemeIsle", 0xf2b2, ["Brand"]], ["google-plus-official", "Google Plus Official", 0xf2b3, ["Brand"], ["google-plus-circle"]], ["font-awesome", "Font Awesome", 0xf2b4, ["Brand"], ["fa"]], ["handshake-o", "Handshake Outlined", 0xf2b5, ["Web Application"]], ["envelope-open", "Envelope Open", 0xf2b6, ["Web Application"], ["email", "e-mail", "letter", "support", "mail", "message", "notification"]], ["envelope-open-o", "Envelope Open Outlined", 0xf2b7, ["Web Application"], ["email", "e-mail", "letter", "support", "mail", "message", "notification"]], ["linode", "Linode", 0xf2b8, ["Brand"]], ["address-book", "Address Book", 0xf2b9, ["Web Application"]], ["address-book-o", "Address Book Outlined", 0xf2ba, ["Web Application"]], ["address-card", "Address Card", 0xf2bb, ["Web Application"], ["vcard"]], ["address-card-o", "Address Card Outlined", 0xf2bc, ["Web Application"], ["vcard-o"]], ["user-circle", "User Circle", 0xf2bd, ["Web Application"]], ["user-circle-o", "User Circle Outlined", 0xf2be, ["Web Application"]], ["user-o", "User Outlined", 0xf2c0, ["Web Application"]], ["id-badge", "Identification Badge", 0xf2c1, ["Web Application"]], ["id-card", "Identification Card", 0xf2c2, ["Web Application"], ["drivers-license"]], ["id-card-o", "Identification Card Outlined", 0xf2c3, ["Web Application"], ["drivers-license-o"]], ["quora", "Quora", 0xf2c4, ["Brand"]], ["free-code-camp", "Free Code Camp", 0xf2c5, ["Brand"]], ["telegram", "Telegram", 0xf2c6, ["Brand"]], ["thermometer-full", "Thermometer Full", 0xf2c7, ["Web Application"], ["thermometer-4", "thermometer"]], ["thermometer-three-quarters", "Thermometer 3/4 Full", 0xf2c8, ["Web Application"], ["thermometer-3"]], ["thermometer-half", "Thermometer 1/2 Full", 0xf2c9, ["Web Application"], ["thermometer-2"]], ["thermometer-quarter", "Thermometer 1/4 Full", 0xf2ca, ["Web Application"], ["thermometer-1"]], ["thermometer-empty", "Thermometer Empty", 0xf2cb, ["Web Application"], ["thermometer-0"]], ["shower", "Shower", 0xf2cc, ["Web Application"]], ["bath", "Bath", 0xf2cd, ["Web Application"], ["bathtub", "s15"]], ["podcast", "Podcast", 0xf2ce, ["Web Application"]], ["window-maximize", "Window Maximize", 0xf2d0, ["Web Application"]], ["window-minimize", "Window Minimize", 0xf2d1, ["Web Application"]], ["window-restore", "Window Restore", 0xf2d2, ["Web Application"]], ["window-close", "Window Close", 0xf2d3, ["Web Application"], ["times-rectangle"]], ["window-close-o", "Window Close Outline", 0xf2d4, ["Web Application"], ["times-rectangle-o"]], ["bandcamp", "Bandcamp", 0xf2d5, ["Brand"]], ["grav", "Grav", 0xf2d6, ["Brand"]], ["etsy", "Etsy", 0xf2d7, ["Brand"]], ["imdb", "IMDB", 0xf2d8, ["Brand"]], ["ravelry", "Ravelry", 0xf2d9, ["Brand"]], ["eercast", "Eercast", 0xf2da, ["Brand"]], ["microchip", "Microchip", 0xf2db, ["Web Application"]], ["snowflake-o", "Snowflake Outlined", 0xf2dc, ["Web Application"]], ["superpowers", "Superpowers", 0xf2dd, ["Brand"]], ["wpexplorer", "WPExplorer", 0xf2de, ["Brand"]], ["meetup", "Meetup", 0xf2e0, ["Brand"]]];
exports["default"] = _default;

/***/ }),

/***/ "./js/core/skins/fontawesomemap.js":
/*!*****************************************!*\
  !*** ./js/core/skins/fontawesomemap.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.getGlyphByIconName = getGlyphByIconName;
exports.getNewIconName = getNewIconName;

var _fontawesomeicons = _interopRequireDefault(__webpack_require__(/*! core/skins/fontawesomeicons */ "./js/core/skins/fontawesomeicons.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var old2new = {
  AddCommentIcon: "comment-o",
  AddPageIcon: "plus-square-o",
  //pretty bad
  AIMIcon: "user",
  //bad
  AlarmClockIcon: "clock-o",
  //pretty bad
  Alert2Icon: "exclamation-circle",
  AlertIcon: "exclamation-triangle",
  ArchiveIcon: "archive",
  //bad
  AsteriskIcon: "asterisk",
  BackSpaceIcon: "caret-square-o-left",
  //bad
  BaloonIcon: "comment-o",
  BatteryIcon: "battery-three-quarters",
  BloggerIcon: "stop",
  //REALLY bad
  BoldIcon: "bold",
  BookIcon: "book",
  BulletsIcon: "list-ul",
  CalendarIcon: "calendar",
  CameraIcon: "camera",
  CartIcon: "shopping-card",
  CellBarsIcon: "signal",
  CellPhoneIcon: "mobile",
  ChartColumnIcon: "bar-chart",
  ChartLineIcon: "line-chart",
  ChartPieIcon: "pie-chart",
  CheckMarkIcon: "check",
  CircledMinusIcon: "minus-circle",
  CircledPlusIcon: "plus-circle",
  CircledXIcon: "times-circle",
  CircleIcon: "circle-o",
  ClockIcon: "clock-o",
  CloudIcon: "cloud",
  ClubIcon: "certificate",
  //REALLY bad
  ColumnsIcon: "columns",
  CommentsIcon: "comments-o",
  ContactIcon: "user",
  ContactsIcon: "users",
  CopyIcon: "files-o",
  CopyrightIcon: "copyright",
  CoverFlowIcon: "stop",
  //REALLY bad
  CreativeCommonsIcon: "creative-commons",
  CursorIcon: "mouse-pointer",
  CutIcon: "scissors",
  //pretty bad
  DashIcon: "minus",
  DataBaseIcon: "database",
  DeliciousIcon: "delicious",
  DiggIcon: "digg",
  DocEmptyIcon: "file-o",
  DocExcelIcon: "file-excel-o",
  DocPDFIcon: "file-pdf-o",
  DocWordIcon: "file-word-o",
  DollarIcon: "usd",
  DotHollowIcon: "circle-o",
  //pretty bad
  DotIcon: "circle",
  //pretty bad
  DoubleArrowIcon: "arrows-h",
  DownArrowIcon: "arrow-down",
  DownFillTriangleIcon: "caret-down",
  DownLeftArrowIcon: "arrow-down",
  //REALLY bad
  DownloadIcon: "download",
  DownRightArrowIcon: "arrow-down",
  //REALLY bad
  DownTriangleIcon: "angle-down",
  //pretty bad
  EjectIcon: "eject",
  EllipsisIcon: "ellipsis-h",
  EmailIcon: "envelope-o",
  EmoticonIcon: "smile-o",
  EmptyIcon: "square-o",
  EuroIcon: "eur",
  ExclamationPointIcon: "exclamation",
  ExportIcon: "file-o",
  //bad
  EyeIcon: "eye",
  FacebookIcon: "facebook",
  FileIcon: "file-text-o",
  FilterIcon: "filter",
  FlagIcon: "flag-o",
  FlickrIcon: "flickr",
  FolderClosed: "folder-o",
  FolderIcon: "folder-open-o",
  ForbiddenIcon: "ban",
  ForwardChapterIcon: "fast-forward",
  ForwardIcon: "forward",
  FriendFeedIcon: "stop",
  //REALLY bad
  FullScreenExitIcon: "compress",
  //bad
  FullScreenIcon: "expand",
  //bad
  GearIcon: "cog",
  GeoLocation2Icon: "compass",
  GeoLocationIcon: "location-arrow",
  GMailIcon: "envelope",
  //pretty bad
  GoogleIcon: "google",
  GridIcon: "th",
  GtCircledIcon: "chevron-circle-right",
  GtIcon: "chevron-right",
  HandClosedIcon: "hand-rock-o",
  HandOpenIcon: "hand-paper-o",
  HandPointIcon: "hand-pointer-o",
  HeartIcon: "heart-o",
  Help2Icon: "question",
  HelpIcon: "question-circle",
  HLinesIcon: "bars",
  HomeIcon: "home",
  HomeMobileIcon: "home",
  HourGlassIcon: "spinner",
  //pretty bad
  IBeamIcon: "i-cursor",
  ImageIcon: "picture-o",
  ImportIcon: "file-o",
  //REALLY bad
  IndentIcon: "indent",
  Info2Icon: "info",
  InfoIcon: "info-circle",
  ItalicIcon: "italic",
  JewelIcon: "diamond",
  //bad
  JumpBackIcon: "reply",
  JumpOutIcon: "external-link",
  LeftArrowIcon: "arrow-left",
  LeftFillTriangleIcon: "caret-left",
  LeftTriangleIcon: "angle-left",
  LightBulbIcon: "lightbulb-o",
  LightningIcon: "bolt",
  LineIcon: "minus",
  //bad
  LinkedInIcon: "linkedin",
  LinkGlobeIcon: "globe",
  LinkIcon: "link",
  ListIcon: "bars",
  LockIcon: "lock",
  LockOpenIcon: "unlock",
  MicIcon: "microphone",
  MinusIcon: "minus",
  MusicIcon: "music",
  Number0Icon: "circle-thin",
  //REALLY bad
  Number1Icon: "circle-thin",
  //REALLY bad
  Number2Icon: "circle-thin",
  //REALLY bad
  Number3Icon: "circle-thin",
  //REALLY bad
  Number4Icon: "circle-thin",
  //REALLY bad
  Number5Icon: "circle-thin",
  //REALLY bad
  Number6Icon: "circle-thin",
  //REALLY bad
  Number7Icon: "circle-thin",
  //REALLY bad
  Number8Icon: "circle-thin",
  //REALLY bad
  Number9Icon: "circle-thin",
  //REALLY bad
  NumberedListIcon: "list-ol",
  OpenIDIcon: "openid",
  OutdentIcon: "outdent",
  PagesIcon: "clone",
  PaperClipIcon: "paperclip",
  PasteIcon: "clipboard",
  PauseIcon: "pause",
  PencilIcon: "pencil",
  PhoneDialIcon: "phone",
  PhoneHangUpIcon: "phone",
  //REALLY bad
  PlusBigIcon: "plus",
  PlusIcon: "plus",
  PopOutIcon: "external-link",
  PortletEditIcon: "pencil",
  //bad
  PortletHelpIcon: "question-circle",
  //bad
  PortletMaximizeIcon: "expand",
  //bad
  PortletMinimizeIcon: "compress",
  //bad
  PoundIcon: "gbp",
  PowerIcon: "power-off",
  PrintIcon: "print",
  PushPinIcon: "thumb-tack",
  RecycleIcon: "recycle",
  RedditIcon: "reddit",
  RedoIcon: "repeat",
  RegisteredIcon: "registered",
  ReloadIcon: "refresh",
  RepeatIcon: "retweet",
  RewindChapterIcon: "fast-backward",
  RewindIcon: "backward",
  RightArrowIcon: "arrow-right",
  RightFillTriangleIcon: "caret-right",
  RightTriangleIcon: "angle-right",
  RSSIcon: "rss",
  SadFaceIcon: "frown-o",
  SaveIcon: "floppy-o",
  ScreenIcon: "desktop",
  SearchIcon: "search",
  ShapesIcon: "cubes",
  //bad
  ShareIcon: "share-alt",
  ShiftIcon: "caret-square-o-up",
  //pretty bad
  ShuffleIcon: "random",
  SimpleArrowIcon: "arrow-right",
  SkypeIcon: "skype",
  SlideShareIcon: "slideshare",
  SmallClockIcon: "clock-o",
  SpadesIcon: "tree",
  //REALLY bad
  SpinningIcon: "spinner",
  SquaredMinusIcon: "minus-square-o",
  SquaredPlusIcon: "plus-square-o",
  StarIcon: "star",
  StopIcon: "stop",
  StrikeThroughIcon: "strikethrough",
  StumbleUponIcon: "stumbleupon",
  SuitcaseIcon: "suitcase",
  TableIcon: "table",
  TagIcon: "tag",
  TapeIcon: "stop",
  //REALLY bad
  TelephoneIcon: "phone",
  TextAlignCenterIcon: "align-center",
  TextAlignLeftIcon: "align-left",
  TextAlignRightIcon: "align-right",
  ThreeGIcon: "stop",
  //REALLY bad
  ThumbnailsIcon: "th-large",
  ThumbsDownIcon: "thumbs-o-down",
  ThumbsUpIcon: "thumbs-o-up",
  TrademarkIcon: "trademark",
  //REALLY bad
  Trash2Icon: "trash",
  TrashIcon: "trash-o",
  TruckIcon: "truck",
  TwitterIcon: "twitter",
  UnderlineIcon: "underline",
  UndoIcon: "undo",
  UpArrowIcon: "arrow-up",
  UpdateIcon: "arrow-circle-o-down",
  UpDownFillTriangleIcon: "sort",
  UpFillTriangleIcon: "caret-up",
  UpLeftArrowIcon: "arrow-up",
  //bad
  UpRightArrowIcon: "arrow-up",
  //bad
  UpTriangleIcon: "angle-up",
  UserIcon: "user",
  UsersIcon: "users",
  VideoCameraIcon: "video-camera",
  VLinesIcon: "bars",
  //REALLY bad
  VolumeIcon: "volume-up",
  WirelessIcon: "wifi",
  WorldIcon: "globe",
  WriteIcon: "pencil-square-o",
  XIcon: "times",
  YIMIcon: "smile-o",
  YouTubeIcon: "youtube"
}; // Map Tree custom assets names

old2new.SquaredPlusForTree = "plus-square-o";
old2new.CheckBoxIndeterminate = "minus-square-o";
old2new.CheckBoxSelected = "check-square-o";
old2new.CheckBox = "square-o"; // Icon Data structure
// -------------------
// For faster access to FontAwesome icons

var iconData = {};

for (var i = 0; i < _fontawesomeicons.default.length; i++) {
  var row = _fontawesomeicons.default[i];
  iconData[row[0]] = {
    id: row[0],
    name: row[1],
    charCode: row[2]
  };
}

function getNewIconName(name) {
  var newName = old2new[name] ? old2new[name] : name;
  return iconData[newName] ? newName : null;
}

;

function getGlyphByIconName(name) {
  return iconData[name] ? String.fromCharCode(iconData[name].charCode) : ' ';
}

;

/***/ }),

/***/ "./js/core/skins/sketch.js":
/*!*********************************!*\
  !*** ./js/core/skins/sketch.js ***!
  \*********************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;
var _default = {
  "#skinName": "sketch",
  "default_font_size": 13,
  "default_border_color": 0x000000,
  "default_disabled_color": 0x80808b,
  "default_selection_color": 0x95BCF2,
  "default_disabled_selection_color": 0xcbddf2,
  "default_icon_color": 0x000000,
  "default_text_color": 0x000000,
  "default_link_color": 0x085394,
  "default_splitter_color": 0x000000,
  "alert_separator_color": 0x000000,
  "alert_title_font_size": 18,
  "alert_body_font_size": 13,
  "alert_padding": 10,
  "alert_line_padding": 3,
  "alert_button_color": 0,
  "appbar_background_color": 0xDDDDDD,
  "appbar_padding_left": 13,
  "appbar_icon_padding": 13,
  "appbar_font_size": 16,
  "appbar_icon_top_padding": 14,
  "appbar_padding_bottom": 14,
  "blockoftext_fake_font": true,
  "blockoftext_default_font_size": 13,
  "blockoftext_space_width": 4,
  "blockoftext_2": {
    text: "do",
    type: "SVG_PATHS",
    width: 15,
    height: 8,
    paths: ["M14.609,3.587c-0.234-0.503-0.744-0.814-1.3-0.78c0.158-0.57,0.41-1.303,0.203-1.888c-0.24-0.682-1.014-1.105-1.71-0.839c-0.55,0.21-1.01,0.815-1.455,1.186C9.77,1.748,9.172,2.219,8.512,2.583c0.347-0.95,0.009-2.256-1.198-2.312C6.32,0.224,5.809,1.218,5.224,1.845C5.131,1.021,4.41,0.434,3.531,0.525c-0.943,0.097-1.606,0.95-2.207,1.593C0.887,2.6,0.181,3.153,0.032,3.803C-0.077,4.277,0.09,4.897,0.505,5.186c0.396,0.278,0.127,0.761,0.351,1.209C1.674,8.25,4.031,6.903,5.028,6.07c1.418,1.793,3.812-0.637,5.07-1.391c-0.658,1.316,1.045,2.058,2.015,1.527C13.088,5.683,15.255,5.124,14.609,3.587C14.488,3.327,14.679,3.752,14.609,3.587z"]
  },
  "blockoftext_3": {
    text: "sit",
    type: "SVG_PATHS",
    width: 21.5,
    height: 6.2,
    paths: ["M21.365,3.529c-0.292-0.793-1.097-0.956-1.82-0.728c0.847-1.592,0.019-3.102-1.874-2.75c-1.596,0.297-2.993,1.253-4.381,2.042c0.136-1.382-1.302-1.971-2.416-1.437C9.865,1.165,9.138,2.056,8.313,2.798C9.162,1.207,8.327-0.3,6.438,0.051C4.747,0.366,3.278,1.412,1.804,2.24C1.289,2.529,0.524,2.76,0.196,3.281c-0.377,0.598-0.192,1.41,0.407,1.788c1.229,0.775,3.25-1.067,4.304-1.617C4.321,4.446,4.163,5.939,5.651,6.178C6.81,6.364,8.273,5.583,9.338,5.199c0.055-0.018,0.974,0.736,1.34,0.736c0.599,0,1.249-0.342,1.798-0.547c1.304-0.488,2.434-1.297,3.663-1.936c-0.96,1.6-0.093,3.077,1.795,2.615C19.082,5.785,21.977,5.282,21.365,3.529C21.304,3.365,21.425,3.703,21.365,3.529z"]
  },
  "blockoftext_4": {
    text: "amet",
    type: "SVG_PATHS",
    width: 29.2,
    height: 7.4,
    paths: ["M29.158,3.697c-0.292-0.793-1.097-0.956-1.82-0.728c0.947-1.781-0.21-3.175-2.145-2.691c-1.703,0.334-3.197,1.514-4.7,2.316c0.342-0.946-0.013-2.236-1.205-2.292c-0.993-0.046-1.505,0.948-2.09,1.575c-0.093-0.824-0.813-1.411-1.692-1.32c-0.839,0.086-1.442,0.775-1.982,1.353c0.195-1.1-0.743-2.462-1.955-1.677c-1.083,0.702-1.901,1.747-3.056,2.382C8.86,1.664,8.523,0.358,7.314,0.302C6.321,0.256,5.809,1.25,5.225,1.877c-0.103-0.822-0.806-1.412-1.692-1.32C2.589,0.654,1.925,1.506,1.325,2.15C0.887,2.632,0.184,3.186,0.033,3.835C-0.077,4.309,0.091,4.93,0.506,5.218c0.396,0.279,0.127,0.76,0.351,1.209C1.672,8.283,4.035,6.934,5.03,6.103c1.417,1.793,3.812-0.637,5.069-1.391C9.85,5.21,10.088,5.879,10.502,6.207c0.723,0.613,1.51,0.072,2.196-0.278c0.077,1.104,1.173,1.521,2.133,1.37c0.561-0.088,1.103-0.409,1.561-0.728c-0.108,0.074,0.362-0.267,0.312-0.228c0.38-0.291,0.266-0.165,0.644,0.065c1.202,0.736,2.571-0.106,3.509-0.864c0.935-0.756,1.997-1.364,3.076-1.924c-0.96,1.6-0.093,3.077,1.795,2.615C26.875,5.953,29.77,5.45,29.158,3.697C29.097,3.533,29.218,3.871,29.158,3.697z"]
  },
  "blockoftext_5": {
    text: "lorem",
    type: "SVG_PATHS",
    width: 32.4,
    height: 7.2,
    paths: ["M32.254,3.587c-0.234-0.503-0.744-0.814-1.3-0.78c0.158-0.57,0.41-1.303,0.203-1.888c-0.24-0.682-1.014-1.105-1.71-0.839c-0.55,0.21-1.01,0.815-1.455,1.186c-0.578,0.481-1.176,0.953-1.836,1.316c0.347-0.95,0.009-2.256-1.198-2.312c-0.993-0.046-1.505,0.948-2.09,1.575c-0.119-1.059-1.248-1.611-2.249-1.17c-0.81,0.356-1.454,1.191-2.017,1.85c0.71-1.722-0.687-2.685-2.312-2.278c-1.78,0.349-3.362,1.604-4.935,2.442c0.147-0.538,0.366-1.219,0.171-1.769c-0.24-0.682-1.013-1.105-1.709-0.839c-0.633,0.242-1.196,0.979-1.717,1.4C7.529,1.939,6.946,2.37,6.294,2.709c0.278-1.019,0.478-2.605-1.02-2.703C4.61-0.038,4.208,0.482,3.758,0.88c-0.557,0.492-1.14,0.965-1.756,1.381C1.229,2.784-0.184,3.089,0.02,4.312c0.286,1.712,2.118,0.919,3.037,0.367c-0.565,1.13,0.873,2.14,1.846,1.612c0.393,0,1.109-0.582,1.461-0.75c0.589-0.281,1.188-0.525,1.75-0.862c-0.249,0.498-0.01,1.168,0.403,1.495c0.664,0.561,1.395,0.149,2.035-0.195c1.488-0.797,2.979-1.614,4.479-2.392c-0.479,0.799-0.811,2.022,0.23,2.566c0.498,0.26,1.044,0.176,1.564,0.048c0.138-0.034,1.535-0.366,1.535-0.477c0,2.425,2.927,1.423,4.014,0.586c0.472-0.361,0.527,0.118,1.079,0.241c0.613,0.179,1.258,0.072,1.826-0.2c0.894-0.427,1.619-1.165,2.464-1.672c-0.658,1.316,1.045,2.058,2.015,1.527C30.733,5.683,32.9,5.124,32.254,3.587C32.133,3.327,32.324,3.752,32.254,3.587z"]
  },
  "blockoftext_6": {
    text: "tempor",
    type: "SVG_PATHS",
    width: 42.6,
    height: 7.2,
    paths: ["M42.478,3.665c-0.292-0.793-1.097-0.956-1.82-0.728c0.947-1.781-0.21-3.175-2.145-2.691c-1.776,0.348-3.369,1.6-4.936,2.443c0.174-0.634,0.41-1.4,0.06-2.014c-0.432-0.74-1.434-0.889-2.054-0.299c-1.056,0.978-2.24,2.121-3.64,2.561c0.947-1.781-0.21-3.175-2.145-2.691c-1.779,0.349-3.361,1.604-4.934,2.442c0.14-0.543,0.367-1.213,0.171-1.769c-0.24-0.682-1.014-1.105-1.71-0.839c-0.55,0.21-1.01,0.815-1.455,1.186c-0.578,0.481-1.176,0.953-1.836,1.316c0.347-0.95,0.009-2.256-1.198-2.312c-0.993-0.046-1.505,0.948-2.09,1.575c-0.119-1.059-1.248-1.611-2.249-1.17C9.684,1.033,9.05,1.869,8.479,2.526c0.719-1.723-0.689-2.686-2.311-2.28c-1.678,0.329-3.189,1.482-4.66,2.294C0.98,2.833,0.322,3.026,0.085,3.639C-0.17,4.295,0.167,5.062,0.826,5.316c1.221,0.477,3.055-1.195,4.082-1.729c-0.479,0.799-0.811,2.022,0.23,2.566c1.109,0.58,2.047-0.429,3.101-0.429c0,2.425,2.927,1.423,4.014,0.586c0.472-0.361,0.527,0.118,1.079,0.241c0.613,0.179,1.258,0.072,1.826-0.2c0.894-0.427,1.619-1.165,2.464-1.672c-0.249,0.499-0.011,1.167,0.403,1.495c0.663,0.562,1.395,0.149,2.034-0.195c1.488-0.797,2.979-1.614,4.479-2.392c-0.834,1.391-0.302,3.06,1.547,2.672c1.432-0.304,2.992-0.824,4.25-1.58c-0.249,0.499-0.011,1.167,0.403,1.495c0.663,0.562,1.395,0.149,2.034-0.195c1.489-0.798,2.98-1.613,4.479-2.392c-0.96,1.6-0.093,3.077,1.795,2.615C40.194,5.921,43.09,5.417,42.478,3.665C42.417,3.5,42.538,3.838,42.478,3.665z"]
  },
  "blockoftext_7": {
    text: "eiusmod",
    type: "SVG_PATHS",
    width: 43.1,
    height: 7.3,
    paths: ["M42.994,3.653c-0.179-0.529-0.652-0.867-1.202-0.923c0.308-0.912,0.251-2.081-0.884-2.392c-1.134-0.284-1.724,0.845-2.366,1.534c-0.093-0.824-0.813-1.411-1.692-1.32c-0.835,0.086-1.445,0.772-1.98,1.351c0.213-1.239-1.022-2.598-2.177-1.5c-0.94,0.871-1.923,1.74-3.067,2.333c0.175-0.64,0.418-1.414,0.064-2.033c-0.432-0.74-1.434-0.889-2.054-0.299c-1.057,0.978-2.24,2.12-3.641,2.561c0.423-0.795,0.671-1.86-0.177-2.489c-0.815-0.605-2.07-0.261-2.916,0.098c-1.656,0.703-3.154,1.732-4.758,2.534c0.367-0.859,1.022-2.282-0.156-2.857c-1.254-0.611-2.464,1.147-3.299,1.832c-0.011-0.937-0.856-1.596-1.765-1.471C10.023,0.735,9.273,1.452,8.643,2.05c0.109-1.043-0.8-2.063-1.891-1.666C6.083,0.628,5.675,1.375,5.211,1.873c-0.08-0.716-0.634-1.282-1.407-1.327c-0.892-0.052-1.56,0.599-2.122,1.203C1.219,2.242,0.684,2.736,0.282,3.278C-0.542,4.392,0.7,4.896,0.7,5.786c0,2.182,2.622,1.51,3.691,0.78c0.305-0.195,0.5-0.551,0.782-0.301C5.502,6.556,5.991,6.669,6.42,6.669c0.974,0,1.783-0.589,2.502-1.18c0.828,1.143,2.766,0.296,3.765-0.276c-0.695,2.067,1.637,2.331,2.936,1.495c0.769-0.495,1.426-1.178,2.255-1.573c0.938-0.447,1.793-1.043,2.713-1.521c-0.834,1.391-0.302,3.06,1.547,2.672c1.432-0.304,2.993-0.824,4.251-1.58c-0.564,1.13,0.872,2.14,1.846,1.612c0.393,0,1.109-0.582,1.462-0.75c0.589-0.281,1.188-0.525,1.749-0.862c-0.249,0.499-0.011,1.167,0.403,1.495c0.719,0.609,1.513,0.077,2.195-0.277c0.176,2.499,3.293,1.017,4.304,0.173C39.623,7.713,43.661,5.596,42.994,3.653C42.938,3.489,43.051,3.818,42.994,3.653z"]
  },
  "blockoftext_8": {
    text: "proident",
    type: "SVG_PATHS",
    width: 53.1,
    height: 7.3,
    paths: ["M52.967,3.665c-0.292-0.793-1.097-0.956-1.82-0.728c0.947-1.781-0.21-3.175-2.145-2.691c-1.779,0.349-3.361,1.604-4.934,2.442c0.14-0.543,0.367-1.213,0.171-1.769c-0.24-0.682-1.014-1.105-1.71-0.839c-0.55,0.21-1.01,0.815-1.455,1.186c-0.578,0.481-1.176,0.953-1.836,1.316c0.347-0.95,0.009-2.256-1.198-2.312c-0.993-0.046-1.505,0.948-2.09,1.575c-0.081-0.716-0.634-1.282-1.407-1.327c-0.869-0.05-1.515,0.563-2.069,1.148c-0.056-1.079-1.229-1.729-2.159-1.188c-0.52,0.302-0.88,0.937-1.28,1.367c-0.196-1.741-2.229-1.479-3.192-0.472c-1.055-1.897-3.168-0.031-4.139,0.911c0.17-0.723,0.267-1.475-0.378-1.999c-0.643-0.521-1.357-0.242-1.892,0.253c-0.997,0.925-2.143,1.982-3.464,2.398c0.423-0.795,0.671-1.86-0.177-2.489c-0.729-0.541-1.803-0.308-2.589-0.03c-1.731,0.612-3.16,1.973-4.892,2.519c0.847-1.592,0.019-3.102-1.874-2.75c-1.69,0.314-3.161,1.36-4.634,2.188c-0.516,0.29-1.28,0.521-1.608,1.041c-0.378,0.598-0.192,1.41,0.406,1.788c1.23,0.776,3.248-1.068,4.305-1.617c-0.389,0.648-0.704,1.457-0.231,2.15c0.547,0.801,1.533,0.601,2.321,0.389c2.052-0.55,3.709-1.574,5.567-2.539c-0.834,1.391-0.302,3.06,1.547,2.672c1.432-0.304,2.992-0.824,4.25-1.58c-0.205,0.41-0.068,0.981,0.217,1.316c0.566,0.665,1.313,0.472,1.981,0.112c0.359-0.193,0.721-0.38,1.088-0.556c0.574-0.273,0.417-0.147,0.861,0.213c0.639,0.515,1.338,0.234,2.015,0c0.017,1.743,1.998,1.72,3.133,1.125c0.271-0.143,0.503-0.307,0.731-0.466c0.276-0.194,0.336-0.427,0.608-0.185c0.693,0.615,1.798,0.444,2.55,0.056c0.818,1.936,3.135,0.673,4.209-0.224c1.418,1.793,3.812-0.637,5.07-1.391c-0.249,0.499-0.011,1.167,0.403,1.495c0.663,0.562,1.395,0.149,2.034-0.195c1.488-0.797,2.979-1.614,4.479-2.392c-0.96,1.6-0.093,3.077,1.795,2.615C50.684,5.921,53.579,5.417,52.967,3.665C52.906,3.5,53.027,3.838,52.967,3.665z"]
  },
  "blockoftext_9": {
    text: "consequat",
    type: "SVG_PATHS",
    width: 61.6,
    height: 7.3,
    paths: ["M61.513,3.665c-0.292-0.793-1.097-0.956-1.82-0.728c0.947-1.781-0.21-3.175-2.145-2.691c-1.703,0.334-3.197,1.514-4.7,2.316c0.342-0.946-0.013-2.236-1.205-2.292c-0.993-0.046-1.505,0.948-2.09,1.575c-0.136-1.208-1.504-1.654-2.521-1.027c-0.793,0.525-1.349,1.324-2.003,1.998c0.147-0.426,0.406-0.892,0.441-1.354c0.074-0.956-0.976-1.675-1.842-1.267c-1.349,0.636-2.085,2.272-3.578,2.743c0.947-1.781-0.21-3.175-2.145-2.691c-1.703,0.334-3.197,1.514-4.7,2.316C33.547,1.616,33.192,0.326,32,0.27c-0.993-0.046-1.505,0.948-2.09,1.575c-0.119-1.059-1.248-1.611-2.249-1.17c-0.814,0.358-1.448,1.194-2.019,1.852c0.719-1.723-0.689-2.686-2.311-2.28c-1.779,0.349-3.361,1.604-4.934,2.442c0.14-0.543,0.367-1.213,0.171-1.769c-0.24-0.682-1.014-1.105-1.71-0.839c-0.633,0.242-1.195,0.979-1.717,1.4c-0.571,0.458-1.155,0.89-1.807,1.228c0.148-0.544,0.374-1.233,0.177-1.789c-0.24-0.682-1.014-1.105-1.71-0.839c-0.55,0.21-1.01,0.815-1.455,1.186C9.77,1.748,9.172,2.219,8.512,2.583c0.347-0.95,0.009-2.256-1.198-2.312C6.32,0.224,5.809,1.218,5.224,1.845C5.131,1.021,4.41,0.434,3.531,0.525c-0.943,0.097-1.606,0.95-2.207,1.593C0.887,2.6,0.181,3.153,0.032,3.803C-0.077,4.277,0.09,4.897,0.505,5.186c0.396,0.278,0.127,0.761,0.351,1.209C1.674,8.25,4.031,6.903,5.028,6.07c1.418,1.793,3.812-0.637,5.07-1.391c-0.564,1.13,0.872,2.14,1.846,1.612c0.393,0,1.109-0.582,1.462-0.75c0.589-0.281,1.188-0.525,1.749-0.862c-0.249,0.499-0.011,1.167,0.403,1.495c0.663,0.562,1.395,0.149,2.034-0.195c1.488-0.797,2.979-1.614,4.479-2.392c-0.479,0.799-0.811,2.022,0.23,2.566c1.109,0.58,2.047-0.429,3.101-0.429c0,1.058,0.915,1.653,1.901,1.571c0.66-0.055,1.266-0.385,1.801-0.757c-0.108,0.074,0.362-0.267,0.312-0.228c0.38-0.291,0.266-0.165,0.644,0.065c1.202,0.736,2.571-0.106,3.509-0.864c0.935-0.756,1.997-1.364,3.076-1.924C36.037,4.599,35.947,6.33,37.58,6.33c1.153,0,2.885-0.57,3.888-1.144c-0.263,0.78-0.001,1.764,0.884,1.976c1.056,0.265,1.958-0.393,2.754-0.992c0.292,1.029,1.388,1.303,2.322,1.044c0.471-0.131,0.94-0.427,1.306-0.676c0.164-0.105,0.32-0.221,0.472-0.343c0.256-0.206,0.411,0.167,0.707,0.286c1.344,0.543,2.604-0.372,3.58-1.204c0.822-0.66,1.849-1.199,2.795-1.69c-0.96,1.6-0.093,3.077,1.795,2.615C59.229,5.921,62.125,5.417,61.513,3.665C61.452,3.5,61.573,3.838,61.513,3.665z"]
  },
  "blockoftext_10": {
    text: "incididunt",
    type: "SVG_PATHS",
    width: 56.9,
    height: 7.3,
    paths: ["M56.794,3.684c-0.292-0.793-1.097-0.956-1.82-0.728c0.947-1.781-0.21-3.175-2.145-2.691c-1.779,0.349-3.361,1.604-4.934,2.442c0.187-0.724,0.46-1.586-0.091-2.221c-0.619-0.714-1.471-0.494-2.08,0.071c-1.123,1.042-2.3,1.862-3.659,2.542c0.396-0.926,1.07-2.569-0.416-2.951c-0.706-0.216-1.158,0.244-1.625,0.676c-0.646,0.593-1.289,1.196-2.006,1.705c0.335-0.936-0.071-2.186-1.222-2.24c-0.993-0.046-1.505,0.948-2.09,1.575c-0.196-1.741-2.229-1.479-3.192-0.472c-0.968-1.741-2.908-0.259-3.866,0.649c0.109-1.043-0.8-2.063-1.891-1.666c-0.669,0.245-1.077,0.991-1.541,1.489c-0.201-1.787-2.207-1.417-3.193-0.472c-0.968-1.741-2.907-0.26-3.866,0.649c0.109-1.043-0.8-2.063-1.891-1.666c-0.669,0.245-1.077,0.991-1.541,1.489c-0.093-0.824-0.813-1.411-1.692-1.32c-0.835,0.086-1.445,0.772-1.98,1.351c0.174-1.012-0.512-2.257-1.721-1.795C7.861,0.279,7.49,0.761,7.115,1.083c-0.471,0.406-0.961,0.79-1.474,1.141c0.155-1.848-1.912-1.83-3.01-1.053C1.843,1.731,0.089,2.989,0.004,4.041c-0.047,0.622,0.385,1.221,1.001,1.352c0.458,0.098,0.622-0.178,0.927,0.184c0.274,0.327,0.667,0.512,1.095,0.512c1.125,0,2.653-0.821,3.602-1.391C6.379,5.196,6.617,5.865,7.031,6.193C7.75,6.802,8.544,6.27,9.227,5.916c0.146,2.069,2.649,1.345,3.68,0.641c0.305-0.195,0.5-0.551,0.781-0.301c0.329,0.292,0.818,0.405,1.247,0.405c0.974,0,1.783-0.589,2.502-1.18c0.603,0.832,1.441,0.592,2.27,0.302c0.021,2.174,2.614,1.51,3.691,0.774c0.305-0.195,0.5-0.551,0.782-0.301c0.328,0.292,0.817,0.405,1.246,0.405c0.974,0,1.783-0.589,2.502-1.18c0.606,0.837,1.438,0.583,2.27,0.302c0,2.179,2.626,1.501,3.69,0.774c0.305-0.195,0.5-0.551,0.782-0.301c0.328,0.292,0.817,0.405,1.246,0.405c1.035,0,1.896-0.666,2.651-1.301c-0.29,1.5,0.991,2.21,2.302,1.697c0.579-0.229,1.069-0.634,1.56-1.007c0.712-0.548,1.46-0.894,2.224-1.352c-0.249,0.499-0.011,1.167,0.403,1.495c0.663,0.562,1.395,0.149,2.034-0.195c1.488-0.797,2.979-1.614,4.479-2.392c-0.96,1.6-0.093,3.077,1.795,2.615C54.511,5.939,57.406,5.436,56.794,3.684C56.733,3.519,56.854,3.857,56.794,3.684z"]
  },
  "blockoftext_11": {
    text: "consectetur",
    type: "SVG_PATHS",
    width: 76.2,
    height: 7.3,
    paths: ["M76.087,3.665c-0.292-0.793-1.097-0.956-1.82-0.728c0.423-0.795,0.671-1.86-0.177-2.489c-0.815-0.605-2.07-0.261-2.916,0.098c-1.656,0.703-3.155,1.732-4.759,2.534c0.203-0.476,0.503-1.082,0.544-1.619c0.074-0.956-0.976-1.675-1.842-1.267C63.77,0.83,63.033,2.467,61.54,2.937c0.947-1.781-0.21-3.175-2.145-2.691c-1.703,0.334-3.197,1.514-4.7,2.316c0.342-0.946-0.013-2.236-1.205-2.292c-0.993-0.046-1.505,0.948-2.09,1.575c-0.119-1.059-1.248-1.611-2.249-1.17c-0.814,0.358-1.448,1.194-2.019,1.852c0.719-1.723-0.689-2.686-2.311-2.28c-1.703,0.334-3.197,1.514-4.7,2.316c0.342-0.946-0.013-2.236-1.205-2.292c-0.993-0.046-1.505,0.948-2.09,1.575c-0.08-0.716-0.634-1.282-1.407-1.327c-0.869-0.05-1.515,0.563-2.069,1.148c-0.056-1.079-1.229-1.729-2.159-1.188c-0.52,0.302-0.88,0.937-1.28,1.367c-0.119-1.059-1.248-1.611-2.249-1.17c-0.814,0.358-1.448,1.194-2.019,1.852c0.719-1.723-0.689-2.686-2.311-2.28c-1.779,0.349-3.361,1.604-4.934,2.442c0.14-0.543,0.367-1.213,0.171-1.769c-0.24-0.682-1.014-1.105-1.71-0.839c-0.633,0.242-1.195,0.979-1.717,1.4c-0.571,0.458-1.155,0.89-1.807,1.228c0.148-0.544,0.374-1.233,0.177-1.789c-0.24-0.682-1.014-1.105-1.71-0.839c-0.55,0.21-1.01,0.815-1.455,1.186C9.771,1.748,9.173,2.219,8.513,2.583c0.347-0.95,0.009-2.256-1.198-2.312C6.321,0.224,5.81,1.218,5.225,1.845c-0.093-0.824-0.813-1.411-1.692-1.32C2.587,0.623,1.928,1.476,1.324,2.118C0.887,2.6,0.183,3.154,0.033,3.803c-0.109,0.473,0.059,1.094,0.473,1.383c0.395,0.28,0.127,0.76,0.352,1.209C1.669,8.253,4.036,6.899,5.029,6.07c1.418,1.793,3.812-0.637,5.07-1.391c-0.564,1.13,0.872,2.14,1.846,1.612c0.393,0,1.109-0.582,1.462-0.75c0.589-0.281,1.188-0.525,1.749-0.862c-0.249,0.499-0.011,1.167,0.403,1.495c0.663,0.562,1.395,0.149,2.034-0.195c1.488-0.797,2.979-1.614,4.479-2.392c-0.479,0.799-0.811,2.022,0.23,2.566c1.109,0.58,2.047-0.429,3.101-0.429c0,1.558,1.713,1.845,2.863,1.298c0.289-0.137,0.595-0.327,0.825-0.484c0.305-0.195,0.5-0.551,0.781-0.301c0.693,0.615,1.798,0.445,2.55,0.057c0.388,0.931,1.362,1.174,2.279,0.92c0.471-0.131,0.94-0.427,1.306-0.676c0.164-0.105,0.32-0.221,0.472-0.343c0.256-0.206,0.411,0.167,0.707,0.286c1.344,0.543,2.604-0.372,3.58-1.204c0.822-0.66,1.849-1.199,2.795-1.69c-0.479,0.799-0.811,2.022,0.23,2.566c1.109,0.58,2.047-0.429,3.101-0.429c0,1.058,0.915,1.653,1.901,1.571c0.66-0.055,1.266-0.385,1.801-0.757c-0.108,0.074,0.362-0.267,0.312-0.228c0.38-0.291,0.266-0.165,0.644,0.065c1.202,0.736,2.571-0.106,3.509-0.864c0.935-0.756,1.997-1.364,3.076-1.924C57.527,4.599,57.438,6.33,59.07,6.33c1.153,0,2.884-0.57,3.887-1.144c-0.695,2.067,1.637,2.331,2.936,1.495c0.769-0.495,1.425-1.176,2.253-1.572c0.938-0.447,1.795-1.044,2.716-1.522c-0.96,1.6-0.093,3.077,1.795,2.615C73.804,5.921,76.699,5.417,76.087,3.665C76.026,3.5,76.147,3.838,76.087,3.665z"]
  },
  "blockoftext_12": {
    text: "exercitation",
    type: "SVG_PATHS",
    width: 75.7,
    height: 7.3,
    paths: ["M75.592,3.587c-0.234-0.503-0.744-0.814-1.3-0.78c0.158-0.57,0.41-1.303,0.203-1.888c-0.24-0.682-1.014-1.105-1.71-0.839c-0.633,0.242-1.195,0.979-1.717,1.4c-0.571,0.458-1.155,0.89-1.807,1.228c0.278-1.018,0.478-2.604-1.02-2.703c-0.665-0.043-1.067,0.477-1.518,0.874c-0.536,0.474-1.096,0.922-1.687,1.326c0.115-1.372-1.315-1.943-2.42-1.414c-1.009,0.508-1.735,1.399-2.561,2.142c0.943-1.78-0.212-3.171-2.146-2.688c-1.677,0.329-3.239,1.481-4.699,2.315c0.343-0.946-0.014-2.236-1.205-2.292c-0.993-0.046-1.505,0.948-2.09,1.575c-0.119-1.059-1.248-1.611-2.249-1.17c-0.814,0.358-1.448,1.194-2.019,1.852c0.644-1.541-0.452-2.635-2.04-2.339c-1.596,0.297-2.993,1.253-4.381,2.042c0.099-1.007-0.758-1.781-1.746-1.645c-0.904,0.125-1.646,0.842-2.282,1.438c0.11-1.043-0.801-2.062-1.891-1.666c-0.669,0.245-1.077,0.991-1.541,1.489c-0.119-1.059-1.248-1.611-2.249-1.17c-0.81,0.356-1.454,1.191-2.017,1.85c0.721-1.722-0.694-2.683-2.312-2.278c-1.703,0.334-3.197,1.514-4.7,2.316c0.342-0.946-0.013-2.236-1.205-2.292c-0.993-0.046-1.505,0.948-2.09,1.575c-0.136-1.208-1.504-1.654-2.521-1.027c-0.79,0.523-1.362,1.318-2.005,1.999c0.339-0.909,0.882-2.329-0.518-2.688c-0.706-0.216-1.158,0.244-1.625,0.676C9.882,1.399,9.238,1.999,8.522,2.51c0.334-0.936-0.07-2.187-1.222-2.24c-0.993-0.046-1.505,0.948-2.09,1.575c-0.08-0.716-0.634-1.282-1.407-1.327c-0.892-0.052-1.56,0.599-2.122,1.203C1.219,2.215,0.684,2.709,0.282,3.251C-0.542,4.365,0.7,4.869,0.7,5.758c0,2.182,2.622,1.51,3.691,0.78c0.305-0.195,0.5-0.551,0.781-0.3C5.502,6.529,5.991,6.642,6.42,6.642c1.035,0,1.896-0.666,2.651-1.301c-0.175,0.902,0.19,1.74,1.177,1.87c0.961,0.126,1.781-0.499,2.501-1.042c0.305,1.025,1.379,1.307,2.323,1.045c0.471-0.131,0.94-0.427,1.306-0.676c0.164-0.105,0.32-0.221,0.472-0.343c0.256-0.206,0.411,0.167,0.707,0.286c1.344,0.543,2.604-0.372,3.58-1.204c0.822-0.66,1.849-1.199,2.795-1.69c-0.479,0.799-0.811,2.022,0.23,2.566c0.498,0.26,1.044,0.176,1.564,0.048c0.138-0.034,1.535-0.366,1.535-0.477c0,2.164,2.557,1.611,3.702,0.813c0.173-0.118,0.419-0.308,0.61-0.468c0.268,0.339,0.677,0.489,1.093,0.549c1.085,0.158,2.021-0.506,2.813-1.158c1.333,1.847,5.245-1.171,6.599-1.875c-0.479,0.799-0.811,2.022,0.23,2.566c1.109,0.58,2.047-0.429,3.101-0.429c0,1.058,0.915,1.653,1.901,1.571c0.66-0.055,1.266-0.385,1.801-0.757c-0.108,0.074,0.362-0.267,0.312-0.228c0.38-0.291,0.266-0.165,0.644,0.065c1.202,0.736,2.571-0.106,3.509-0.864c0.935-0.756,1.996-1.364,3.075-1.924c-0.479,0.799-0.811,2.022,0.23,2.566c0.997,0.521,2.4-0.183,3.354-0.517c0.469-0.165,0.766-0.467,1.093-0.078c0.46,0.547,1.04,0.534,1.681,0.408c1.024-0.419,2.059-0.712,3.016-1.287c-0.564,1.13,0.872,2.14,1.846,1.612c0.393,0,1.109-0.582,1.462-0.75c0.589-0.281,1.188-0.525,1.749-0.862c-0.658,1.316,1.045,2.058,2.015,1.527C74.07,5.683,76.237,5.124,75.592,3.587C75.471,3.327,75.661,3.752,75.592,3.587z"]
  },
  "blockoftext_13": {
    text: "reprehenderit",
    type: "SVG_PATHS",
    width: 87.6,
    height: 7.3,
    paths: ["M87.548,3.697c-0.292-0.793-1.097-0.956-1.82-0.728c0.847-1.592,0.019-3.102-1.874-2.75c-1.596,0.297-2.993,1.253-4.381,2.042c0.136-1.382-1.302-1.971-2.416-1.437c-1.006,0.507-1.741,1.394-2.559,2.141c0.945-1.78-0.216-3.169-2.146-2.687c-1.703,0.334-3.197,1.514-4.7,2.316c0.342-0.946-0.013-2.236-1.205-2.292c-0.993-0.046-1.505,0.948-2.09,1.575c-0.08-0.716-0.634-1.282-1.407-1.327C62.08,0.5,61.435,1.114,60.88,1.699c-0.056-1.079-1.229-1.729-2.159-1.188c-0.52,0.302-0.88,0.937-1.28,1.367c-0.093-0.824-0.813-1.411-1.692-1.32c-0.841,0.087-1.438,0.776-1.982,1.352c0.196-1.1-0.744-2.461-1.955-1.676c-1.083,0.702-1.901,1.747-3.056,2.382c0.347-0.95,0.009-2.256-1.198-2.312c-0.993-0.046-1.505,0.948-2.09,1.575c-0.196-1.741-2.229-1.479-3.192-0.472c-0.968-1.741-2.908-0.259-3.866,0.649c0.109-1.043-0.8-2.063-1.891-1.666c-0.669,0.245-1.077,0.991-1.541,1.489c-0.119-1.059-1.248-1.611-2.249-1.17c-0.814,0.358-1.448,1.194-2.019,1.852c0.336-0.804,0.329-1.78-0.574-2.217c-0.817-0.396-1.897-0.097-2.685,0.237c-1.561,0.662-2.958,1.886-4.563,2.391c0.947-1.781-0.21-3.175-2.145-2.691c-1.703,0.334-3.197,1.514-4.7,2.316c0.342-0.946-0.013-2.236-1.205-2.292c-0.993-0.046-1.505,0.948-2.09,1.575c-0.119-1.059-1.248-1.611-2.249-1.17C9.684,1.065,9.05,1.901,8.479,2.559c0.719-1.723-0.689-2.686-2.311-2.28c-1.678,0.329-3.189,1.482-4.66,2.294C0.98,2.865,0.322,3.058,0.085,3.671C-0.17,4.328,0.167,5.095,0.826,5.348c1.221,0.477,3.055-1.195,4.082-1.729c-0.479,0.799-0.811,2.022,0.23,2.566c1.109,0.58,2.047-0.429,3.101-0.429c0,1.058,0.915,1.653,1.901,1.571c0.66-0.055,1.266-0.385,1.801-0.757c-0.108,0.074,0.362-0.267,0.312-0.228c0.38-0.291,0.266-0.165,0.644,0.065c1.202,0.736,2.571-0.106,3.509-0.864c0.935-0.756,1.997-1.364,3.076-1.924c-0.389,0.648-0.704,1.457-0.23,2.15c0.546,0.8,1.532,0.6,2.32,0.389c2.051-0.55,3.709-1.574,5.566-2.539c-0.479,0.799-0.811,2.022,0.23,2.566c1.109,0.58,2.047-0.429,3.101-0.429c0,2.164,2.557,1.611,3.702,0.813c0.173-0.118,0.419-0.308,0.61-0.468c0.268,0.339,0.677,0.489,1.093,0.549c1.085,0.158,2.021-0.506,2.813-1.157c0.604,0.833,1.439,0.589,2.269,0.302c0.009,0.928,0.795,1.541,1.677,1.541c0.949,0,1.793-0.551,2.486-1.11c0.237-0.191,0.588,0.279,0.933,0.356c0.613,0.179,1.258,0.072,1.826-0.2c0.894-0.427,1.619-1.165,2.464-1.672c-0.249,0.499-0.011,1.167,0.403,1.495c0.72,0.61,1.514,0.077,2.196-0.278c0.116,1.658,2.041,1.559,3.121,0.993c0.271-0.143,0.503-0.307,0.731-0.466c0.276-0.194,0.336-0.427,0.609-0.186c0.692,0.615,1.797,0.445,2.549,0.057c0.388,0.931,1.362,1.174,2.279,0.92c0.471-0.131,0.94-0.427,1.306-0.676c0.164-0.105,0.32-0.221,0.472-0.343c0.256-0.206,0.411,0.167,0.707,0.286c1.344,0.543,2.604-0.372,3.58-1.204c0.822-0.66,1.849-1.199,2.795-1.69c-0.596,0.992-0.737,2.489,0.744,2.727c1.157,0.186,2.622-0.595,3.687-0.979c0.055-0.018,0.974,0.736,1.34,0.736c0.599,0,1.249-0.342,1.798-0.547c1.304-0.488,2.434-1.297,3.663-1.936c-0.96,1.6-0.093,3.077,1.795,2.615C85.265,5.953,88.16,5.45,87.548,3.697C87.487,3.533,87.608,3.871,87.548,3.697z"]
  },
  "circlebutton_bg_color": 0xCCCCCC,
  "circlebutton_border_thickness": 3,
  "circlebutton_padding": 2,
  "button": {
    type: "SVG_SCALE9",
    scale9: [4, 5, 6, 6],
    width: 42,
    height: 27
  },
  "button_nw": {
    type: "SVG_PATHS",
    width: 5,
    height: 4,
    paths: ["M2.92,4H0.719l-0.1-3.145L5,0.262v2.22L2.881,2.77L2.92,4z"]
  },
  "button_n": {
    type: "SVG_PATHS",
    width: 31,
    height: 4,
    paths: ["M0,2.481v-2.22l0.378-0.051l0.089,0.002l11.5,0.312L31,0.803v2.2L11.921,2.725L0.497,2.414L0,2.481z"]
  },
  "button_ne": {
    type: "SVG_PATHS",
    width: 6,
    height: 4,
    paths: ["M3.5,4H1.3V3.021L0,3.003v-2.2l3.5,0.051V4z", "M6.063,4h-4.5L1.54,2.101L6.07,2.855L6.063,4z"]
  },
  "button_e": {
    type: "SVG_PATHS",
    width: 6,
    height: 17,
    paths: ["M3.5,0v17H1.3V0H3.5z", "M6.063,0L5.961,17h-4.18L1.564,0H6.063z"]
  },
  "button_se": {
    type: "SVG_PATHS",
    width: 6,
    height: 6,
    paths: ["M3.5,0v2.349L0,2.558V0.354l1.3-0.077V0H3.5z", "M5.961,0L5.928,5.453L0,5.39V0.943L1.793,0.94L1.781,0H5.961z"]
  },
  "button_s": {
    type: "SVG_PATHS",
    width: 31,
    height: 6,
    paths: ["M24.935,0.715L31,0.354v2.204l-5.876,0.35l-3.08,0.513l-7.337-0.632L6.133,2.85L0,3.382V1.175l6.03-0.521l8.763-0.067l0.051,0.005l7.112,0.613L24.935,0.715z", "M0,1l31-0.057V5.39L17.921,5.25L0,6.085V1z"]
  },
  "button_sw": {
    type: "SVG_PATHS",
    width: 5,
    height: 6,
    paths: ["M3.5,1.304L5,1.175v2.207L1.375,3.696L1.258,0h2.201L3.5,1.304z", "M5,1v5.085L4.196,6.122L3.266,1.003L5,1z"]
  },
  "button_w": {
    type: "SVG_PATHS",
    width: 5,
    height: 17,
    paths: ["M2.92,0l0.539,17H1.258L0.719,0H2.92z"]
  },
  "button_c": {
    type: "SVG_PATHS",
    width: 31,
    height: 17,
    paths: []
  },
  "button_padding_left": 3,
  "button_padding_right": 5,
  "button_padding_top": 4,
  "button_padding_bottom": 4,
  "button_background_padding_left": 2,
  "button_background_padding_right": 4,
  "button_background_padding_top": 2,
  "button_background_padding_bottom": 3,
  "button_border_color": 0x000000,
  "button_disabled_border_color": 0x80808b,
  "button_background_color": 0xffffff,
  "button_icon_padding": 4,
  "button_default_icon_color": 0x000000,
  "iosKeyboard_icon_color": 0,
  "iosKeyboard_background_color": 0xEEEEEE,
  "ioskeyboard_letter_key": {
    type: "SVG_PATHS",
    width: 23,
    height: 29,
    colors: [0xFFFFFF, 0x000000],
    paths: ["M4.156,0.112L8.040,0.216L8.040,0.216L18.968,0.131C21.177,0.114,22.982,1.891,22.999,4.100C22.999,4.110,23,4.120,23,4.131L23,24.870C23,27.079,21.209,28.870,19,28.870C18.957,28.870,18.915,28.869,18.872,28.868L16.718,28.8L16.718,28.8L4.442,28.946C2.095,28.974,0.170,27.095,0.142,24.747C0.141,24.685,0.142,24.622,0.144,24.559L0.25,21.297L0.25,21.297L0.048,4.157C0.022,1.948,1.792,0.136,4.001,0.110C4.053,0.110,4.104,0.110,4.156,0.112Z", "M4.001,0.110L4.078,0.110L8.040,0.216L18.968,0.131C21.110,0.114,22.872,1.785,22.993,3.900L23,4.131L23,24.870C23,27.012,21.316,28.761,19.199,28.865L18.936,28.870L16.718,28.8L4.442,28.946C2.095,28.974,0.170,27.095,0.142,24.747L0.144,24.559L0.25,21.297L0.048,4.157C0.022,1.948,1.792,0.136,4.001,0.110ZM4.019,1.610C2.638,1.627,1.532,2.759,1.548,4.140L1.749,21.345L1.643,24.607L1.642,24.730C1.660,26.248,2.906,27.465,4.424,27.446L16.766,27.300L18.949,27.370L19.164,27.365C20.468,27.280,21.5,26.196,21.5,24.870L21.5,4.131L21.493,3.947C21.398,2.644,20.305,1.621,18.980,1.631L7.999,1.716L4.115,1.611L4.019,1.610Z"]
  },
  "ioskeyboard_letter_key_gray": {
    type: "SVG_PATHS",
    width: 23,
    height: 29,
    colors: [0xCCCCCC, 0x000000],
    paths: ["M4.156,0.112L8.040,0.216L8.040,0.216L18.968,0.131C21.177,0.114,22.982,1.891,22.999,4.100C22.999,4.110,23,4.120,23,4.131L23,24.870C23,27.079,21.209,28.870,19,28.870C18.957,28.870,18.915,28.869,18.872,28.868L16.718,28.8L16.718,28.8L4.442,28.946C2.095,28.974,0.170,27.095,0.142,24.747C0.141,24.685,0.142,24.622,0.144,24.559L0.25,21.297L0.25,21.297L0.048,4.157C0.022,1.948,1.792,0.136,4.001,0.110C4.053,0.110,4.104,0.110,4.156,0.112Z", "M4.001,0.110L4.078,0.110L8.040,0.216L18.968,0.131C21.110,0.114,22.872,1.785,22.993,3.900L23,4.131L23,24.870C23,27.012,21.316,28.761,19.199,28.865L18.936,28.870L16.718,28.8L4.442,28.946C2.095,28.974,0.170,27.095,0.142,24.747L0.144,24.559L0.25,21.297L0.048,4.157C0.022,1.948,1.792,0.136,4.001,0.110ZM4.019,1.610C2.638,1.627,1.532,2.759,1.548,4.140L1.749,21.345L1.643,24.607L1.642,24.730C1.660,26.248,2.906,27.465,4.424,27.446L16.766,27.300L18.949,27.370L19.164,27.365C20.468,27.280,21.5,26.196,21.5,24.870L21.5,4.131L21.493,3.947C21.398,2.644,20.305,1.621,18.980,1.631L7.999,1.716L4.115,1.611L4.019,1.610Z"]
  },
  "ioskeyboard_123_key": {
    type: "SVG_PATHS",
    width: 33,
    height: 29,
    colors: [0xCCCCCC, 0x000000],
    paths: ["M4.043,0.044L22.929,0.25L22.929,0.25L27.793,0.380C30.261,0.446,32.215,2.489,32.171,4.957L31.819,25.045C31.780,27.236,29.986,28.989,27.795,28.975L7.628,28.85L7.628,28.85L4.786,28.907C2.301,28.958,0.246,26.984,0.196,24.500C0.195,24.487,0.195,24.475,0.195,24.462L-3.552e-15,8.212L-3.552e-15,8.212L-3.552e-15,4.043C-5.155e-15,1.834,1.790,0.043,4,0.043C4.014,0.043,4.029,0.043,4.043,0.044Z", "M4,0.043L22.929,0.25L27.793,0.380C30.261,0.446,32.215,2.489,32.171,4.957L31.819,25.045C31.780,27.236,29.986,28.989,27.795,28.975L7.628,28.85L4.786,28.907C2.372,28.956,0.364,27.096,0.205,24.711L0.195,24.462L-7.105e-15,8.212L-7.105e-15,4.043C-7.105e-15,1.834,1.790,0.043,4,0.043ZM3.835,1.549C2.531,1.633,1.499,2.718,1.499,4.043L1.499,8.194L1.695,24.444L1.704,24.645C1.827,26.218,3.158,27.440,4.756,27.408L7.637,27.350L27.804,27.475C29.174,27.483,30.295,26.388,30.319,25.019L30.672,4.931C30.700,3.285,29.398,1.924,27.752,1.879L22.912,1.749L4.027,1.543L3.835,1.549Z"]
  },
  "ioskeyboard_return_key": {
    type: "SVG_PATHS",
    width: 47,
    height: 28,
    colors: [0xCCCCCC, 0x000000],
    paths: ["M4,5.506e-14L43,5.506e-14C45.209,5.4668e-14,47,1.790,47,4L47,23C47,25.209139,45.209139,27,43,27L4,27C1.790,27,2.705e-16,25.209,0,23L0,4C-2.705e-16,1.790,1.790,5.547e-14,4,5.506e-14Z", "M4.445,0.084L11.75,0.241L23.5,5.481e-14L42,5.481e-14C44.761,5.481e-14,47,2.158,47,4.821L47,13.5L46.5,20.25L46.693,22.868C46.851,24.992,45.192,26.838,42.989,26.990C42.894,26.996,42.799,27,42.704,27L4.670,27C2.216,27,0.214,25.104,0.171,22.738L0,13.5L0.359,3.803C0.437,1.694,2.257,0.037,4.445,0.084ZM45.5,13.448L45.5,4.821C45.5,2.957,43.932,1.446,42,1.446L23.531,1.446L11.716,1.687L4.412,1.530C3.044,1.501,1.907,2.536,1.858,3.854L1.499,13.474L1.670,22.712C1.700,24.289,3.034,25.553,4.670,25.553L42.704,25.553C42.763,25.553,42.822,25.551,42.882,25.547C44.259,25.452,45.296,24.299,45.197,22.971L45.003,20.353L45.003,20.146L45.5,13.448Z"]
  },
  "ioskeyboard_space_key": {
    type: "SVG_PATHS",
    width: 96,
    height: 29,
    colors: [0xFFFFFF, 0x000000],
    paths: ["M5.037,0.013L78,0.25L78,0.25L90.930,0.070C93.691,0.032,95.961,2.239,95.999,5.000C95.999,5.023,96,5.046,96,5.069L96,14.5L96,14.5L95.650,24.625C95.566,27.060,93.559,28.986,91.123,28.970L21.067,28.5L21.067,28.5L5.630,28.866C2.593,28.938,0.073,26.535,0.001,23.498C0.000,23.454,1.801e-15,23.411,0,23.367L0,14.5L0,14.5L0.528,4.281C0.652,1.884,2.636,0.006,5.037,0.013Z", "M5.037,0.013L78,0.25L90.930,0.070C93.619,0.033,95.841,2.124,95.991,4.783L96,5.069L96,14.5L95.650,24.625C95.566,27.060,93.559,28.986,91.123,28.970L21.067,28.5L5.630,28.866C2.593,28.938,0.073,26.535,0.001,23.498L0,14.5L0.528,4.281C0.652,1.884,2.636,0.006,5.037,0.013ZM5.032,1.513L4.855,1.518C3.335,1.601,2.106,2.819,2.026,4.358L1.5,14.5L1.5,23.367L1.510,23.662C1.664,25.775,3.453,27.417,5.594,27.366L21.077,27.000L91.133,27.470C92.757,27.481,94.095,26.197,94.151,24.574L94.5,14.5L94.5,5.069L94.491,4.829C94.366,2.987,92.819,1.544,90.951,1.570L77.995,1.749L5.032,1.513Z"]
  },
  "pointybutton_default_font_size": 13,
  "pointybutton_background_color": 0xFFFFFF,
  "pointybutton_border_color": 0x000000,
  "pointybutton_padding_left": 10,
  "pointybutton_padding_right": 10,
  "pointybutton_padding_top": 0,
  "pointybutton_padding_bottom": 0,
  "pointybutton_icon_padding": 3,
  "pointybutton_point_padding": 20,
  "pointybutton_disabled_color": 0x999999,
  "pointybutton_noborder_height": 27,
  "pointybutton_background_border_radius": 8,
  "pointybutton_center": {
    type: "SVG_SCALE9",
    scale9: [10, 9, 7, 10],
    width: 36,
    height: 27
  },
  "pointybutton_center_nw": {
    type: "SVG_PATHS",
    width: 9,
    height: 10,
    paths: ["M2.336,10h-2.25C0.195,5.613,2.628,1.038,9,0.406v2.259C3.073,3.359,2.391,8.229,2.336,10z"]
  },
  "pointybutton_center_n": {
    type: "SVG_PATHS",
    width: 20,
    height: 10,
    paths: ["M0,2.665V0.406c0.535-0.053,1.087-0.088,1.678-0.083c2.115,0.023,3.538-0.008,4.917-0.039c1.61-0.036,3.133-0.07,5.521-0.006c0.026-0.003,0.053-0.005,0.08-0.006c0.612-0.023,4.526-0.149,7.872,0.197v2.262c-2.625-0.284-5.986-0.271-7.621-0.217c-0.075,0.014-0.15,0.018-0.23,0.018c-2.424-0.069-3.952-0.035-5.57,0.002C5.249,2.565,3.804,2.596,1.652,2.573C1.055,2.565,0.515,2.605,0,2.665z"]
  },
  "pointybutton_center_ne": {
    type: "SVG_PATHS",
    width: 7,
    height: 10,
    paths: ["M6.957,10h-2.25c0-0.248-0.002-0.491,0-0.761c0.021-2.284,0.048-5.127-1.71-5.895c-0.654-0.286-1.713-0.481-2.93-0.613V0.47c1.512,0.157,2.908,0.41,3.83,0.812c3.122,1.364,3.084,5.345,3.06,7.979C6.955,9.523,6.956,9.76,6.957,10z"]
  },
  "pointybutton_center_e": {
    type: "SVG_PATHS",
    width: 7,
    height: 7,
    paths: ["M6.957,0.345C6.995,2.487,6.992,4.846,6.898,7.02H4.636c0.083-1.845,0.117-4.072,0.071-6.634c-0.002-0.119,0-0.26,0-0.386h2.25C6.957,0.112,6.955,0.238,6.957,0.345z"]
  },
  "pointybutton_center_se": {
    type: "SVG_PATHS",
    width: 7,
    height: 10,
    paths: ["M6.898,0.02C6.757,3.297,6.405,6.147,5.652,7.161C4.22,9.089,1.48,9.813,0.068,9.821V7.569c0.929-0.02,2.875-0.535,3.777-1.75c0.329-0.451,0.641-2.508,0.791-5.8H6.898z"]
  },
  "pointybutton_center_s": {
    type: "SVG_PATHS",
    width: 20,
    height: 10,
    paths: ["M0,9.9V7.645c0.622,0.014,1.269,0.021,1.942,0.013c0.35-0.004,0.837-0.006,1.432-0.006c0.783,0,1.751,0.003,2.84,0.006c4.176,0.015,9.893,0.03,13.792-0.084c0.018,0,0.042-0.004,0.061-0.004v2.252c-0.008,0-0.02,0.002-0.028,0.002c-2.573,0.075-5.944,0.094-9.143,0.094c-1.654,0-3.261-0.005-4.689-0.01c-1.9-0.005-3.433-0.012-4.237,0C1.459,9.913,0.767,9.915,0,9.9z"]
  },
  "pointybutton_center_sw": {
    type: "SVG_PATHS",
    width: 9,
    height: 10,
    paths: ["M4.24,7.203C4.623,7.39,6.551,7.59,9,7.645V9.9C6.739,9.859,3.822,9.677,2.922,9.027C1.248,7.816,0.289,5.097,0.208,3.035C0.182,2.39,0.151,1.296,0.125,0.02h2.25c0.026,1.237,0.055,2.298,0.08,2.926C2.522,4.631,3.323,6.54,4.24,7.203z"]
  },
  "pointybutton_center_w": {
    type: "SVG_PATHS",
    width: 9,
    height: 7,
    paths: ["M2.375,7.02h-2.25C0.08,4.806,0.049,2.042,0.082,0.097c0-0.032,0.004-0.064,0.005-0.097h2.25C2.335,0.044,2.332,0.094,2.332,0.134C2.299,2.069,2.331,4.825,2.375,7.02z"]
  },
  "pointybutton_center_c": {
    type: "SVG_PATHS",
    width: 20,
    height: 7,
    paths: []
  },
  "pointybutton_center_background": {
    type: "SVG_SCALE9",
    scale9: [10, 9, 7, 10],
    width: 36,
    height: 27
  },
  "pointybutton_center_background_nw": {
    type: "SVG_PATHS",
    width: 9,
    height: 10,
    paths: ["M9,10H1.212C1.291,7.195,2.635,2.206,9,1.532V10z"]
  },
  "pointybutton_center_background_n": {
    type: "SVG_PATHS",
    width: 20,
    height: 10,
    paths: ["M20.068,10H0V1.532c0.524-0.056,1.071-0.091,1.665-0.084c4.688,0.055,6.08-0.17,10.583-0.041l-0.01-0.011c0,0,4.343-0.159,7.83,0.209V10z"]
  },
  "pointybutton_center_background_ne": {
    type: "SVG_PATHS",
    width: 7,
    height: 10,
    paths: ["M5.829,10H0.068V1.605c1.358,0.143,2.591,0.363,3.38,0.708C6.139,3.489,5.824,7.43,5.829,10z"]
  },
  "pointybutton_center_background_e": {
    type: "SVG_PATHS",
    width: 7,
    height: 7,
    paths: ["M5.832,0.365C5.854,1.597,5.876,4.338,5.761,7.02H0.068V0h5.761C5.83,0.124,5.83,0.248,5.832,0.365z"]
  },
  "pointybutton_center_background_se": {
    type: "SVG_PATHS",
    width: 7,
    height: 10,
    paths: ["M5.761,0.02C5.638,2.878,5.358,5.669,4.749,6.49C3.575,8.069,1.226,8.688,0.068,8.696V0.02H5.761z"]
  },
  "pointybutton_center_background_s": {
    type: "SVG_PATHS",
    width: 20,
    height: 10,
    paths: ["M20.068,0.02v8.677c-0.008,0-0.02,0.002-0.028,0.002C14.299,8.865,4.5,8.75,1.957,8.782C1.367,8.79,0.694,8.786,0,8.771V0.02H20.068z"]
  },
  "pointybutton_center_background_sw": {
    type: "SVG_PATHS",
    width: 9,
    height: 10,
    paths: ["M9,0.02v8.752c-2.301-0.051-4.842-0.24-5.418-0.656C2.231,7.14,1.402,4.779,1.332,2.99C1.306,2.345,1.276,1.269,1.25,0.02H9z"]
  },
  "pointybutton_center_background_w": {
    type: "SVG_PATHS",
    width: 9,
    height: 7,
    paths: ["M9,0v7.02H1.25C1.205,4.805,1.174,2.038,1.207,0.115C1.207,0.078,1.21,0.037,1.212,0H9z"]
  },
  "pointybutton_center_background_c": {
    type: "SVG_PATHS",
    width: 20,
    height: 7,
    paths: ["M20.068,0v7.02H0V0H20.068z"]
  },
  "pointybutton_left": {
    type: "SVG_SCALE9",
    scale9: [3, 18, 9, 3],
    width: 42,
    height: 27
  },
  "pointybutton_left_nw": {
    type: "SVG_PATHS",
    width: 18,
    height: 3,
    paths: ["M15.296,3H9.667c3.007-1.809,5.874-2.944,7.415-2.968L18,0.018v2.501l-0.881,0.014C16.721,2.538,16.084,2.704,15.296,3z"]
  },
  "pointybutton_left_n": {
    type: "SVG_PATHS",
    width: 15,
    height: 3,
    paths: ["M0,2.519V0.018l0.345-0.005c2.456-0.038,5.483-0.085,8.482-0.085c0.489,0,1.02-0.004,1.573-0.009c0.611-0.004,1.25-0.01,1.894-0.01c1.027,0,1.919,0.015,2.708,0.046v2.504c-0.738-0.03-1.625-0.05-2.708-0.05c-0.637,0-1.27,0.005-1.875,0.01C9.86,2.424,9.323,2.428,8.828,2.428c-2.986,0-6,0.047-8.444,0.085L0,2.519z"]
  },
  "pointybutton_left_ne": {
    type: "SVG_PATHS",
    width: 9,
    height: 3,
    paths: ["M6.688,3H3.441C3.421,2.982,3.4,2.959,3.381,2.942C3.169,2.795,2.422,2.561,0.003,2.459v-2.504c2.844,0.111,4.255,0.438,5.018,1.101C5.561,1.525,6.148,2.203,6.688,3z"]
  },
  "pointybutton_left_e": {
    type: "SVG_PATHS",
    width: 9,
    height: 21,
    paths: ["M8.398,3.989c0.418,2.314,0.512,7.956,0.011,12.04C8.207,17.675,7.977,19.501,7.29,21H4.266c0.929-0.756,1.303-2.348,1.662-5.275c0.506-4.126,0.345-9.443,0.011-11.291C5.625,2.701,4.457,0.912,3.441,0h3.247C7.464,1.146,8.138,2.547,8.398,3.989z"]
  },
  "pointybutton_left_se": {
    type: "SVG_PATHS",
    width: 9,
    height: 3,
    paths: ["M7.29,0C6.493,1.74,5.08,3.037,2.35,3.069C1.742,3.076,0.929,3.084,0.003,3.092v-2.5c0.914-0.008,1.716-0.016,2.318-0.022C3.167,0.56,3.788,0.388,4.266,0H7.29z"]
  },
  "pointybutton_left_s": {
    type: "SVG_PATHS",
    width: 15,
    height: 3,
    paths: ["M8.325,0.62c2.39,0,4.792-0.014,6.678-0.028v2.5c-1.889,0.015-4.29,0.028-6.678,0.028C4.33,3.12,1.553,3.083,0,3.012V0.509C1.446,0.581,4.314,0.62,8.325,0.62z"]
  },
  "pointybutton_left_sw": {
    type: "SVG_PATHS",
    width: 18,
    height: 3,
    paths: ["M17.235,0.457C17.41,0.476,17.682,0.493,18,0.509v2.503c-0.438-0.021-0.795-0.043-1.035-0.068C14.888,2.719,12.366,1.58,9.897,0h5.47C16.031,0.235,16.661,0.395,17.235,0.457z"]
  },
  "pointybutton_left_w": {
    type: "SVG_PATHS",
    width: 18,
    height: 21,
    paths: ["M15.367,21h-5.47c-4.323-2.768-8.483-6.903-9.808-9.959l-0.08-0.224c-0.197-0.585-0.19-0.939,0.035-1.65l0.099-0.229C2.071,5.46,5.975,2.221,9.667,0h5.629C11.75,1.331,5.052,5.363,2.391,10.041c0,0.003-0.001,0.005-0.002,0.007l0.016,0.048C3.955,13.586,10.589,19.31,15.367,21z"]
  },
  "pointybutton_left_c": {
    type: "SVG_PATHS",
    width: 15,
    height: 21,
    paths: []
  },
  "pointybutton_left_background": {
    type: "SVG_SCALE9",
    scale9: [3, 18, 9, 3],
    width: 42,
    height: 27
  },
  "pointybutton_left_background_nw": {
    type: "SVG_PATHS",
    width: 18,
    height: 3,
    paths: ["M18,3h-5.732c2.078-1.066,3.878-1.703,4.833-1.718c0.288-0.004,0.591-0.009,0.9-0.014V3z"]
  },
  "pointybutton_left_background_n": {
    type: "SVG_PATHS",
    width: 15,
    height: 3,
    paths: ["M15.003,3H0V1.269c2.512-0.039,5.683-0.091,8.828-0.091c1.799,0,4.146-0.055,6.175,0.028V3z"]
  },
  "pointybutton_left_background_ne": {
    type: "SVG_PATHS",
    width: 9,
    height: 3,
    paths: ["M5.147,3H0.003V1.206c1.954,0.079,3.614,0.286,4.198,0.793C4.509,2.268,4.832,2.608,5.147,3z"]
  },
  "pointybutton_left_background_e": {
    type: "SVG_PATHS",
    width: 9,
    height: 21,
    paths: ["M7.168,4.211c0.378,2.092,0.497,7.615,0,11.666C6.909,17.989,6.624,19.769,5.881,21H0.003V0h5.144C6.048,1.117,6.889,2.661,7.168,4.211z"]
  },
  "pointybutton_left_background_se": {
    type: "SVG_PATHS",
    width: 9,
    height: 3,
    paths: ["M5.881,0C5.2,1.13,4.134,1.798,2.335,1.819C1.728,1.826,0.928,1.834,0.003,1.842V0H5.881z"]
  },
  "pointybutton_left_background_s": {
    type: "SVG_PATHS",
    width: 15,
    height: 3,
    paths: ["M15.003,0v1.842C10.487,1.878,2.976,1.903,0,1.762V0H15.003z"]
  },
  "pointybutton_left_background_sw": {
    type: "SVG_PATHS",
    width: 18,
    height: 3,
    paths: ["M18,0v1.762c-0.372-0.019-0.685-0.038-0.9-0.062c-1.404-0.152-3.05-0.785-4.745-1.7H18z"]
  },
  "pointybutton_left_background_w": {
    type: "SVG_PATHS",
    width: 18,
    height: 21,
    paths: ["M18,0v21h-5.645c-4.685-2.531-9.744-7.282-11.119-10.456c-0.127-0.397-0.185-0.415,0-1C3.439,5.568,8.354,2.009,12.268,0H18z"]
  },
  "pointybutton_left_background_c": {
    type: "SVG_PATHS",
    width: 15,
    height: 21,
    paths: ["M15.003,0v21H0V0H15.003z"]
  },
  "pointybutton_right": {
    type: "SVG_SCALE9",
    scale9: [3, 8, 18, 3],
    width: 42,
    height: 27
  },
  "pointybutton_right_nw": {
    type: "SVG_PATHS",
    width: 8,
    height: 3,
    paths: ["M5.186,3H1.94c0.54-0.797,1.127-1.475,1.667-1.944c0.708-0.616,1.979-0.941,4.437-1.074v2.507c-1.963,0.112-2.607,0.322-2.81,0.464C5.219,2.967,5.202,2.985,5.186,3z"]
  },
  "pointybutton_right_n": {
    type: "SVG_PATHS",
    width: 16,
    height: 3,
    paths: ["M0,2.488v-2.507c0.92-0.049,2.001-0.072,3.289-0.072c0.644,0,1.283,0.006,1.894,0.01c0.554,0.005,1.084,0.009,1.573,0.009c3,0,6.027,0.047,8.482,0.085l0.762,0.012v2.501l-0.8-0.013C12.8,2.475,9.787,2.428,6.801,2.428c-0.496,0-1.032-0.004-1.593-0.009c-0.604-0.005-1.237-0.01-1.875-0.01C1.95,2.409,0.88,2.44,0.045,2.488z"]
  },
  "pointybutton_right_ne": {
    type: "SVG_PATHS",
    width: 18,
    height: 3,
    paths: ["M7.962,3H2.333C1.544,2.704,0.907,2.538,0.51,2.532L0.045,2.525V0.024l0.502,0.008C2.088,0.056,4.955,1.191,7.962,3z"]
  },
  "pointybutton_right_e": {
    type: "SVG_PATHS",
    width: 18,
    height: 21,
    paths: ["M17.584,9.167c0.225,0.711,0.232,1.065,0.035,1.65l-0.08,0.224c-1.324,3.056-5.484,7.191-9.808,9.959h-5.47c4.778-1.69,11.413-7.415,12.963-10.905l0.016-0.047c0-0.002-0.001-0.004-0.002-0.007C12.576,5.363,5.879,1.331,2.333,0h5.629c3.692,2.221,7.596,5.46,9.523,8.938L17.584,9.167z"]
  },
  "pointybutton_right_se": {
    type: "SVG_PATHS",
    width: 18,
    height: 3,
    paths: ["M7.731,0c-2.469,1.58-4.99,2.719-7.068,2.943C0.505,2.96,0.29,2.976,0.045,2.99V0.484c0.129-0.009,0.258-0.018,0.349-0.027C0.967,0.395,1.597,0.235,2.261,0H7.731z"]
  },
  "pointybutton_right_s": {
    type: "SVG_PATHS",
    width: 16,
    height: 3,
    paths: ["M7.259,0.62c4.437,0,7.485-0.048,8.741-0.136V2.99c-1.44,0.086-4.363,0.13-8.741,0.13c-2.637,0-5.301-0.017-7.259-0.033v-2.5C1.999,0.604,4.666,0.62,7.304,0.62z"]
  },
  "pointybutton_right_sw": {
    type: "SVG_PATHS",
    width: 8,
    height: 3,
    paths: ["M6.308,0.569C6.78,0.575,7.372,0.581,8.045,0.587v2.5C7.36,3.081,6.758,3.075,6.278,3.069C3.549,3.037,2.136,1.74,1.338,0h3.024C4.84,0.388,5.461,0.56,6.308,0.569z"]
  },
  "pointybutton_right_w": {
    type: "SVG_PATHS",
    width: 8,
    height: 21,
    paths: ["M2.69,4.433c-0.334,1.849-0.496,7.166,0.011,11.292C3.06,18.652,3.434,20.244,4.363,21H1.338c-0.686-1.499-0.917-3.325-1.119-4.971C-0.282,11.945-0.188,6.304,0.23,3.988C0.49,2.547,1.164,1.146,1.94,0h3.245C4.179,0.9,3.004,2.695,2.69,4.433z"]
  },
  "pointybutton_right_c": {
    type: "SVG_PATHS",
    width: 16,
    height: 21,
    paths: []
  },
  "pointybutton_right_background": {
    type: "SVG_SCALE9",
    scale9: [3, 8, 18, 3],
    width: 42,
    height: 27
  },
  "pointybutton_right_background_nw": {
    type: "SVG_PATHS",
    width: 8,
    height: 3,
    paths: ["M8.045,3H3.481c0.315-0.392,0.638-0.732,0.946-1.001c0.525-0.456,1.92-0.67,3.617-0.766V3z"]
  },
  "pointybutton_right_background_n": {
    type: "SVG_PATHS",
    width: 16,
    height: 3,
    paths: ["M16,3h-16V1.233c2.149-0.122,4.783-0.056,6.756-0.056c3.323,0,6.671,0.058,9.244,0.097V3z"]
  },
  "pointybutton_right_background_ne": {
    type: "SVG_PATHS",
    width: 18,
    height: 3,
    paths: ["M5.36,3H0.045V1.274C0.206,1.277,0.374,1.28,0.528,1.282C1.482,1.297,3.283,1.934,5.36,3z"]
  },
  "pointybutton_right_background_e": {
    type: "SVG_PATHS",
    width: 18,
    height: 21,
    paths: ["M16.392,9.544c0.185,0.585,0.127,0.603,0,1C15.017,13.718,9.958,18.469,5.273,21H0.045V0H5.36C9.275,2.009,14.189,5.568,16.392,9.544z"]
  },
  "pointybutton_right_background_se": {
    type: "SVG_PATHS",
    width: 18,
    height: 3,
    paths: ["M5.273,0C3.579,0.915,1.932,1.548,0.528,1.7c-0.13,0.014-0.294,0.026-0.483,0.039V0H5.273z"]
  },
  "pointybutton_right_background_s": {
    type: "SVG_PATHS",
    width: 16,
    height: 3,
    paths: ["M16,0v1.739c-2.742,0.174-11.369,0.138-16,0.098V0H16.045z"]
  },
  "pointybutton_right_background_sw": {
    type: "SVG_PATHS",
    width: 8,
    height: 3,
    paths: ["M8.045,0v1.837c-0.68-0.006-1.276-0.012-1.752-0.018C4.494,1.798,3.428,1.13,2.747,0H8.045z"]
  },
  "pointybutton_right_background_w": {
    type: "SVG_PATHS",
    width: 8,
    height: 21,
    paths: ["M8.045,0v21H2.747c-0.743-1.231-1.028-3.011-1.287-5.123c-0.497-4.051-0.378-9.574,0-11.666C1.74,2.661,2.581,1.117,3.481,0H8.045z"]
  },
  "pointybutton_right_background_c": {
    type: "SVG_PATHS",
    width: 16,
    height: 21,
    paths: ["M16,0v21h-16V0H16.045z"]
  },
  "pointybutton_default_icon_color": 0x000000,
  "multilinebutton_default_font_size": 16,
  "multilinebutton_background_color": 0xFFFFFF,
  "multilinebutton_border_color": 0x000000,
  "multilinebutton_padding_left": 14,
  "multilinebutton_padding_right": 14,
  "multilinebutton_padding_top": 16,
  "multilinebutton_padding_bottom": 14,
  "multilinebutton_padding_middle": 8,
  "multilinebutton_icon_padding": 3,
  "multilinebutton_disabled_color": 0x999999,
  "multilinebutton_background_radius": 26,
  "multilinebutton_default_icon_color": 0x000000,
  "label_font_family": "Balsamiq Sans",
  "label_default_font_size": 13,
  "label_padding_top": 4,
  "label_padding_bottom": 4,
  "label_padding_left": 0,
  "label_padding_right": 0,
  "title_padding_top": 4,
  "subtitle_padding_top": 1,
  "titles_padding_bottom": 4,
  "titles_padding_left": 0,
  "titles_padding_right": 0,
  "link_padding_top": 4,
  "link_padding_bottom": 4,
  "linkbar_padding_top": 4,
  "linkbar_padding_bottom": 4,
  "linkbar_padding_middle": 8,
  "rectangle_border_square": {
    type: "SVG_SCALE9",
    scale9: [5, 5, 5, 5],
    width: 41,
    height: 29
  },
  "rectangle_border_square_nw": {
    type: "SVG_PATHS",
    width: 5,
    height: 5,
    paths: ["M2.955,5H0.187l-0.03-0.344L0.125,2.25l1.062-1.969L5,0.164v2.688L2.969,2.875L2.955,5z"]
  },
  "rectangle_border_square_n": {
    type: "SVG_PATHS",
    width: 31,
    height: 5,
    paths: ["M0,2.853V0.164L5.375,0h10L31,0.768v1.979l-3.719,0.097l-1.75-0.375l-2.562,0.312h-6.844L11.594,2.5L6.281,2.781L0,2.853z"]
  },
  "rectangle_border_square_ne": {
    type: "SVG_PATHS",
    width: 5,
    height: 5,
    paths: ["M4.512,5H2.223L2.25,3.5L1.094,2.719L0,2.747V0.768l0.281,0.014L3.812,1L4.512,5z"]
  },
  "rectangle_border_square_e": {
    type: "SVG_PATHS",
    width: 5,
    height: 19,
    paths: ["M5,5.688l-0.5,7.75L4.928,19H2.309l-0.09-3.469l-0.062-4.906l0.125-3.156L2.156,3.688L2.223,0h2.289l0.082,0.469L5,5.688z"]
  },
  "rectangle_border_square_se": {
    type: "SVG_PATHS",
    width: 5,
    height: 5,
    paths: ["M4.928,0L5,0.938L4.125,4.625L2.781,4.8L0,4.471V2.219h2.156l0.156-2.094L2.309,0H4.928z"]
  },
  "rectangle_border_square_s": {
    type: "SVG_PATHS",
    width: 31,
    height: 5,
    paths: ["M0,3.97V1.836l1.812,0.148l4.031,0.078l2.812,0.062l4.875,0.219l2.5-0.031l3.094-0.094h11.25H31v2.252l-0.281-0.033l-7.562-0.031l-4.293,0.457L6.875,4.625L0.438,3.984L0,3.97z"]
  },
  "rectangle_border_square_sw": {
    type: "SVG_PATHS",
    width: 5,
    height: 5,
    paths: ["M2.797,1.656L5,1.836V3.97L0.25,3.812L0.219,2.281L0.009,0h2.554v0.583L2.797,1.656z"]
  },
  "rectangle_border_square_w": {
    type: "SVG_PATHS",
    width: 5,
    height: 19,
    paths: ["M2.955,0l-0.08,12.625l-0.312,5V19H0.009L0,18.906v-3.062l0.594-5.25L0.625,5L0.187,0H2.955z"]
  },
  "rectangle_border_square_c": {
    type: "SVG_PATHS",
    width: 31,
    height: 19,
    paths: []
  },
  "rectangle_border_rounded": {
    type: "SVG_SCALE9",
    scale9: [15, 20, 20, 12],
    width: 69,
    height: 41
  },
  "rectangle_border_rounded_nw": {
    type: "SVG_PATHS",
    width: 20,
    height: 15,
    paths: ["M2.97,15H0.904C3.201,6.808,10.74,0.843,20,0.468v2.021C11.854,2.859,5.202,7.939,2.97,15z"]
  },
  "rectangle_border_rounded_n": {
    type: "SVG_PATHS",
    width: 29,
    height: 15,
    paths: ["M0,2.489V0.468C0.844,0.434,1.701,0.44,2.57,0.502c2.643,0.1,5.512-0.018,8.21-0.136c1.974-0.087,4.015-0.177,6.034-0.177c2.153,0,4.053,0.104,5.807,0.318l1.426,0.179C25.627,0.884,27.3,1.093,29,1.355v2.023c-1.758-0.276-3.515-0.497-5.201-0.707l-1.426-0.18c-1.667-0.203-3.487-0.303-5.559-0.303c-1.975,0-3.994,0.089-5.946,0.175C8.131,2.485,5.29,2.609,2.461,2.499c-0.53-0.036-1.023-0.054-1.514-0.054C0.626,2.445,0.316,2.475,0,2.489z"]
  },
  "rectangle_border_rounded_ne": {
    type: "SVG_PATHS",
    width: 20,
    height: 15,
    paths: ["M19.317,15h-2.009C15.891,7.033,7.945,4.625,0,3.379V1.355C8.579,2.684,17.825,5.41,19.317,15z"]
  },
  "rectangle_border_rounded_e": {
    type: "SVG_PATHS",
    width: 20,
    height: 14,
    paths: ["M19.498,1.431L19.5,9.5c-0.111,1.618-0.462,3.111-0.975,4.5h-2.122c0.598-1.385,0.984-2.909,1.099-4.568L17.5,1.5c-0.037-0.523-0.105-1.019-0.191-1.5h2.009C19.39,0.466,19.463,0.932,19.498,1.431z"]
  },
  "rectangle_border_rounded_se": {
    type: "SVG_PATHS",
    width: 20,
    height: 12,
    paths: ["M18.525,0C15.875,7.179,8.406,11.231,1.539,11.499C1.032,11.555,0.514,11.597,0,11.646V9.638c0.464-0.044,0.932-0.082,1.392-0.132C7.385,9.27,13.847,5.921,16.403,0H18.525z"]
  },
  "rectangle_border_rounded_s": {
    type: "SVG_PATHS",
    width: 29,
    height: 12,
    paths: ["M0,10.721V8.691c2.536,0.457,5.107,0.703,7.543,0.81c7.05,0.767,14.449,0.801,21.457,0.137v2.008c-3.248,0.304-6.617,0.469-10.012,0.469c-3.931,0-7.833-0.209-11.597-0.62C4.741,11.379,2.287,11.111,0,10.721z"]
  },
  "rectangle_border_rounded_sw": {
    type: "SVG_PATHS",
    width: 20,
    height: 12,
    paths: ["M20,8.691v2.029C11.035,9.187,4.915,5.554,1.933,0h2.33C7.521,5.05,13.646,7.547,20,8.691z"]
  },
  "rectangle_border_rounded_w": {
    type: "SVG_PATHS",
    width: 20,
    height: 14,
    paths: ["M2.148,8.642C2.531,10.692,3.269,12.46,4.262,14h-2.33c-0.816-1.521-1.412-3.177-1.749-4.983c-0.218-1.114-0.187-2.989-0.017-3.947C0.184,3.307,0.452,1.614,0.904,0H2.97C2.455,1.63,2.175,3.365,2.167,5.172l-0.02,0.193C2.003,6.072,1.975,7.756,2.148,8.642z"]
  },
  "rectangle_border_rounded_c": {
    type: "SVG_PATHS",
    width: 29,
    height: 14,
    paths: []
  },
  "rectangle_border_rounded_dotted": {
    width: 3,
    corner_radius: 12,
    dash_array: [8, 8]
  },
  "rectangle_corner_radius": {
    roundedSolid: 18,
    roundedDotted: 12,
    square: 0
  },
  "rectangle_background_padding": {
    default: 1,
    none: 0,
    roundedSolid: 2
  },
  "buttonbar_padding_left": 14,
  "buttonbar_padding_middle": 12,
  "buttonbar_padding_right": 14,
  "buttonbar_padding_top": 7,
  "buttonbar_padding_bottom": 7,
  "buttonbar_border": {
    type: "SVG_SCALE9",
    scale9: [4, 5, 7, 5]
  },
  "buttonbar_border_nw": {
    type: "SVG_PATHS",
    width: 5,
    height: 4,
    paths: ["M1.827,2.211c1.058,0.041,2.18,0.075,3.335,0.107V0.333C3.881,0.317,2.345,0.424,0,0.841C0.158,1.905,0.275,2.958,0.36,4h1.589C1.911,3.398,1.873,2.796,1.827,2.211z"]
  },
  "buttonbar_border_n": {
    type: "SVG_PATHS",
    width: 30,
    height: 4,
    paths: ["M30.001,2.852V0.743c-0.88-0.035-1.785-0.101-2.714-0.131c-6.805-0.228-15.953,0.251-22.851,0C2.701,0.549,1.533,0.352,0,0.333v1.985C9.152,2.568,20.979,2.566,30.001,2.852z"]
  },
  "buttonbar_border_ne": {
    type: "SVG_PATHS",
    width: 7,
    height: 4,
    paths: ["M3.686,3.354c-0.388-0.981,0.022-2.206-0.84-2.711C1.936,0.781,0.982,0.782,0,0.743v2.108C0.471,2.866,0.943,2.88,1.398,2.896C1.395,3.268,1.396,3.632,1.395,4h5.262c-0.003-0.211,0.005-0.441,0-0.646C5.794,2.716,4.831,3.557,3.686,3.354z"]
  },
  "buttonbar_border_e": {
    type: "SVG_PATHS",
    width: 7,
    height: 16,
    paths: ["M1.395,0C1.363,5.446,1.602,10.656,1.266,16h4.691c0.072-5.616,0.73-11.584,0.697-16H1.395z"]
  },
  "buttonbar_border_se": {
    type: "SVG_PATHS",
    width: 7,
    height: 7,
    paths: ["M1.171,1.406C0.787,1.467,0.393,1.513,0,1.562v4.727C1.926,6.332,3.844,6.38,5.741,6.434c1.194-1.123,0.266-2.504,0.229-3.885C5.948,1.713,5.945,0.86,5.957,0H1.266C1.234,0.469,1.21,0.936,1.171,1.406z"]
  },
  "buttonbar_border_s": {
    type: "SVG_PATHS",
    width: 30,
    height: 7,
    paths: ["M0,2.243v4.825c10.49-0.987,20.428-0.992,30-0.779V1.562C21.343,2.664,10.674,1.307,0,2.243z"]
  },
  "buttonbar_border_sw": {
    type: "SVG_PATHS",
    width: 5,
    height: 7,
    paths: ["M2.123,2.549C2.136,1.747,2.146,0.888,2.154,0H0.269c0.023,1.547,0.093,3.081,0.254,4.605c1.73,0.166,2.509-0.4,3.428-0.229C3.907,5.488,3.898,6.562,4.408,7.119C4.607,7.1,4.802,7.088,5,7.068V2.243C4.041,2.328,3.081,2.424,2.123,2.549z"]
  },
  "buttonbar_border_w": {
    type: "SVG_PATHS",
    width: 5,
    height: 16,
    paths: ["M1.787,0H0.198c0.456,5.59-0.005,10.868,0.071,16h1.885C2.195,11.207,2.131,5.335,1.787,0z"]
  },
  "buttonbar_separator_width": 2,
  "buttonbar_separator_color": 0x000000,
  'buttonbar_background_color': 0xFFFFFF,
  "buttonbar_selection_padding_top": 1,
  "buttonbar_selection_padding_bottom": 2,
  "buttonbar_selection_padding_left": 1,
  "buttonbar_selection_padding_right": 4,
  "canvas_scrollbar_padding_top": 2,
  "canvas_scrollbar_padding_bottom": 0,
  "canvas_scrollbar_padding_right": 2,
  "canvas_breakline_thickness": 3,
  // Tree
  "tree_scrollbar_padding_top": 4,
  "tree_scrollbar_padding_bottom": 2,
  "tree_scrollbar_padding_right": 2,
  // Image
  "image_border_padding": 5,
  "image_border_color": 0x000000,
  "image_frame": {
    type: "SVG_SCALE9",
    scale9: [7, 5, 6, 5],
    width: 77,
    height: 79
  },
  "image_frame_nw": {
    type: "SVG_PATHS",
    width: 5,
    height: 7,
    paths: ["M2.755,7h-2.5L0.25,1.501c0-0.69,0.559-1.25,1.249-1.251c0,0,0,0,0.001,0c0.139,0,0.272,0.022,0.396,0.064C2.02,0.273,2.151,0.251,2.288,0.25L5,0.239v2.5L2.751,2.748L2.755,7z"]
  },
  "image_frame_n": {
    type: "SVG_PATHS",
    width: 66,
    height: 7,
    paths: ["M0,2.739v-2.5l66-0.26v2.5L0,2.739z"]
  },
  "image_frame_ne": {
    type: "SVG_PATHS",
    width: 6,
    height: 7,
    paths: ["M5.98,7h-2.5l0.001-4.534L0,2.479v-2.5l5.983-0.023L5.98,7z"]
  },
  "image_frame_e": {
    type: "SVG_PATHS",
    width: 6,
    height: 67,
    paths: ["M5.98,0L5.96,67h-2.5L3.48,0H5.98z"]
  },
  "image_frame_se": {
    type: "SVG_PATHS",
    width: 6,
    height: 5,
    paths: ["M5.96,0L5.958,4.704L0,4.709V2.212l3.46-0.004L3.46,0H5.96z"]
  },
  "image_frame_s": {
    type: "SVG_PATHS",
    width: 66,
    height: 5,
    paths: ["M0,4.886V2.385l0.452,0.009L31.5,2.25L66,2.212v2.497L20.5,4.75L0.437,4.894L0,4.886z"]
  },
  "image_frame_sw": {
    type: "SVG_PATHS",
    width: 5,
    height: 5,
    paths: ["M2.82,2.345L5,2.385v2.501L0.321,4.8V3.357L0.318,0h2.5L2.82,2.345z"]
  },
  "image_frame_w": {
    type: "SVG_PATHS",
    width: 5,
    height: 67,
    paths: ["M2.755,0l0.063,67h-2.5L0.255,0H2.755z"]
  },
  "image_frame_c": {
    type: "SVG_PATHS",
    width: 66,
    height: 67,
    paths: []
  },
  "image_line_gap": 5,
  "image_line_width": 3,
  "image_cross_line_color": 0x5C5C5E,
  "icon_default_color": 0x000000,
  "icon_border_padding": 4,
  "icon_label_padding": 3,
  "icon_label_padding_bottom": 2,
  // Choices
  "choice_icon_padding": 3,
  "choicegroup_padding_top": 2,
  "choicegroup_padding_bottom": 2,
  "choicegroup_line_padding": 9,
  "choicegroup_label_padding": 20,
  "checkbox": {
    type: "SVG_PATHS",
    width: 20,
    height: 17,
    paths: ["M7.342,16.214c-1.738,0-3.365-0.132-4.833-0.392L1.87,15.711l-0.158-0.629c-0.914-3.641-0.98-7.836-0.923-11.457l0.017-1.067L1.87,2.645c1.994,0.161,4.007,0.288,6.016,0.414c1.57,0.1,3.139,0.198,4.692,0.312l0.806,0.061l0.11,0.801c0.455,3.283,0.476,6.996,0.06,10.734l-0.086,0.779l-0.777,0.102C10.831,16.091,9.032,16.214,7.342,16.214zM3.501,13.957c2.364,0.342,5.156,0.341,8.146,0.002c0.283-3.004,0.271-5.965-0.032-8.651c-1.28-0.091-2.567-0.171-3.855-0.253C6.098,4.951,4.432,4.846,2.776,4.721C2.755,7.729,2.863,11.035,3.501,13.957z"]
  },
  "checkbox_selected": {
    type: "SVG_PATHS",
    width: 20,
    height: 17,
    paths: ["M7.567,16.073c-1.738,0-3.365-0.131-4.833-0.391L2.095,15.57l-0.158-0.629C1.023,11.301,0.958,7.106,1.015,3.484l0.017-1.067l1.064,0.087C4.09,2.665,6.103,2.792,8.111,2.918c1.571,0.1,3.139,0.198,4.693,0.313l0.806,0.06l0.11,0.801c0.455,3.283,0.476,6.996,0.06,10.734l-0.087,0.779l-0.777,0.102C11.058,15.95,9.258,16.073,7.567,16.073zM3.728,13.816c2.365,0.342,5.156,0.341,8.146,0.003c0.284-3.005,0.272-5.966-0.032-8.652c-1.279-0.091-2.566-0.171-3.855-0.253C6.323,4.81,4.658,4.705,3.002,4.58C2.981,7.588,3.089,10.894,3.728,13.816z", "M6.907,13.461c-0.078,0-0.145-0.007-0.206-0.014c-0.522-0.058-0.962-0.382-1.178-0.866c-0.478-1.074-1.871-2.696-2.367-3.23c-0.47-0.505-0.441-1.296,0.064-1.767c0.506-0.47,1.298-0.439,1.767,0.065c0.168,0.181,1.387,1.509,2.248,2.87c1.577-1.442,4.993-5.634,7.761-9.358c0.412-0.554,1.194-0.669,1.748-0.258c0.555,0.411,0.67,1.194,0.258,1.749C9.544,12.691,7.695,13.461,6.907,13.461z"]
  },
  "checkbox_indeterminate": {
    type: "SVG_PATHS",
    width: 20,
    height: 17,
    paths: ["M7.342,16.214c-1.738,0-3.365-0.132-4.833-0.392L1.87,15.711l-0.158-0.629c-0.914-3.641-0.98-7.836-0.923-11.457l0.017-1.066L1.87,2.645c1.994,0.162,4.007,0.289,6.016,0.414c1.57,0.101,3.139,0.198,4.692,0.312l0.806,0.061l0.11,0.801c0.455,3.283,0.477,6.996,0.061,10.734l-0.087,0.779l-0.776,0.102C10.831,16.091,9.032,16.214,7.342,16.214zM3.501,13.957c2.364,0.342,5.156,0.341,8.146,0.002c0.283-3.004,0.271-5.965-0.031-8.65c-1.28-0.092-2.567-0.172-3.855-0.254C6.098,4.951,4.432,4.846,2.776,4.721C2.755,7.729,2.863,11.035,3.501,13.957z", "M10.531,11H4.094V8.5h6.437V11z"]
  },
  "radio": {
    type: "SVG_PATHS",
    width: 15,
    height: 16,
    paths: ["M6.489,13.958c-3.038,0-6.115-1.74-6.26-5.071C0.102,8.578,0.074,8.286,0.074,7.907L0.087,6.75c0.694-4.396,3.301-5.578,5.366-5.795C6.291,0.87,7.861,1.322,8.645,1.756c0.873,0.48,1.859,1.14,2.573,2.383c1.384,2.41,1.47,6.532-0.489,8.442l-0.092,0.08C9.535,13.498,8.062,13.958,6.489,13.958zM2.075,6.986v0.92c0,0.104-0.002,0.177,0.01,0.236l0.132,0.23l0.008,0.271c0.003,2.174,2.148,3.312,4.265,3.312c1.113,0,2.134-0.301,2.884-0.851c1.104-1.148,1.167-4.133,0.11-5.974C9.018,4.32,8.34,3.871,7.678,3.506C7.183,3.232,6.171,2.943,5.709,2.943C3.664,3.154,2.49,4.477,2.075,6.986z"]
  },
  "radio_selected": {
    type: "SVG_PATHS",
    width: 15,
    height: 16,
    paths: ["M8,10c-1.014,0.43-1.846,0.48-3,0C3.892,9.396,2.706,6.612,3.995,5.529c1.701-1.443,3.284-0.738,4.444,0.102C9.409,7,8.91,9.036,8,10z", "M6.489,13.958c-3.038,0-6.115-1.74-6.26-5.071C0.102,8.578,0.074,8.286,0.074,7.907L0.087,6.75c0.694-4.396,3.301-5.578,5.366-5.795C6.291,0.87,7.861,1.322,8.645,1.756c0.873,0.48,1.859,1.14,2.573,2.383c1.384,2.41,1.47,6.532-0.489,8.442l-0.092,0.08C9.535,13.498,8.062,13.958,6.489,13.958zM2.075,6.986v0.92c0,0.104-0.002,0.177,0.01,0.236l0.132,0.23l0.008,0.271c0.003,2.174,2.148,3.312,4.265,3.312c1.113,0,2.134-0.301,2.884-0.851c1.104-1.148,1.167-4.133,0.11-5.974C9.018,4.32,8.34,3.871,7.678,3.506C7.183,3.232,6.171,2.943,5.709,2.943C3.664,3.154,2.49,4.477,2.075,6.986z"]
  },
  "radio_indeterminate": {
    type: "SVG_PATHS",
    width: 15,
    height: 16,
    paths: ["M6.489,13.958c-3.038,0-6.115-1.74-6.26-5.071C0.102,8.578,0.074,8.286,0.074,7.907L0.087,6.75c0.694-4.396,3.301-5.578,5.366-5.795C6.291,0.87,7.861,1.322,8.645,1.756c0.873,0.48,1.859,1.141,2.573,2.383c1.384,2.41,1.47,6.532-0.489,8.442l-0.092,0.08C9.535,13.498,8.062,13.958,6.489,13.958zM2.075,6.986v0.92c0,0.104-0.002,0.177,0.01,0.236l0.132,0.229l0.008,0.271c0.003,2.174,2.148,3.312,4.265,3.312c1.113,0,2.134-0.301,2.884-0.852c1.104-1.147,1.167-4.133,0.11-5.974C9.018,4.32,8.34,3.871,7.678,3.506C7.183,3.232,6.171,2.943,5.709,2.943C3.664,3.154,2.49,4.477,2.075,6.986z", "M9.371,8.876H2.934V6.375h6.437V8.876z"]
  },
  "CheckBox": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M37.1,25c-0.133,4.899-0.367,11.517-0.699,19.851c-2.867,0.066-8.567-0.284-17.101-1.051c-6.634-0.566-12.25-0.483-16.851,0.25c-0.666-4.333-0.967-10.934-0.899-19.8C1.65,12.351,1.684,6.066,1.65,5.399C5.316,5.533,11.4,5.851,19.9,6.351C27.3,6.616,32.8,6.3,36.4,5.399C37.1,9.467,37.333,16,37.1,25zM5.4,9.399v30.75c0.699,0.033,5.066,0.201,13.1,0.5c4.666,0.167,9.316,0.518,13.949,1.051C33.783,32.733,34.3,22.467,34,10.899c-3.167-1.033-7.65-1.566-13.45-1.6C13.783,9.467,8.732,9.5,5.4,9.399z"]
  },
  "CheckBoxIndeterminate": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M37.1,25c-0.134,4.9-0.366,11.517-0.7,19.85c-2.866,0.066-8.566-0.283-17.1-1.05c-6.633-0.566-12.25-0.483-16.85,0.25c-0.667-4.333-0.967-10.933-0.9-19.8c0.1-11.9,0.133-18.184,0.1-18.85c3.667,0.133,9.75,0.449,18.25,0.949c7.4,0.268,12.9-0.05,16.5-0.949C37.1,9.467,37.333,16,37.1,25zM5.399,9.4v30.75c0.7,0.033,5.066,0.199,13.101,0.5c4.667,0.166,9.316,0.516,13.95,1.049C33.783,32.733,34.3,22.467,34,10.9C30.833,9.867,26.35,9.333,20.55,9.3C13.783,9.467,8.733,9.5,5.399,9.4z", "M19.649,22.35c4.134,0.033,7.184-0.016,9.15-0.149c0.4,0.6,0.533,1.566,0.4,2.899c-0.066,0.7-0.2,1.667-0.4,2.9c-1.633,0-4.8-0.05-9.5-0.15c-3.666-0.1-6.783-0.1-9.35,0c-0.367-0.633-0.533-1.6-0.5-2.899C9.483,23.217,9.5,22.3,9.5,22.2C11.533,22.233,14.917,22.283,19.649,22.35z"]
  },
  "CheckBoxSelected": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M37.1,25c-0.133,4.899-0.367,11.517-0.699,19.851c-2.867,0.066-8.567-0.284-17.101-1.051c-6.634-0.566-12.25-0.483-16.851,0.25c-0.666-4.333-0.967-10.934-0.899-19.8C1.65,12.351,1.684,6.066,1.65,5.399C5.316,5.533,11.4,5.851,19.9,6.351C27.3,6.616,32.8,6.3,36.4,5.399C37.1,9.467,37.333,16,37.1,25zM5.4,9.399v30.75c0.699,0.033,5.066,0.201,13.1,0.5c4.666,0.167,9.316,0.518,13.949,1.051C33.783,32.733,34.3,22.467,34,10.899c-3.167-1.033-7.65-1.566-13.45-1.6C13.783,9.467,8.732,9.5,5.4,9.399z", "M46.5,6.25c-10.801,9.933-20.566,20.267-29.301,31c-1.1-0.2-2.299-1.25-3.6-3.149c-1.333-2.301-2.267-3.884-2.8-4.75c-1.467-2.534-2.384-4.184-2.75-4.951c-0.866-1.799-1.384-3.583-1.55-5.35c1.9-2.033,3.934-1.449,6.1,1.75c1.167,1.7,2.566,4.184,4.2,7.45c1.366-1.2,5.567-5.2,12.601-12c5.433-5.233,9.783-9.067,13.049-11.5c1.234-0.934,1.951-1.434,2.15-1.5c0.768-0.267,1.4,0.233,1.9,1.5c0.066,0.2-0.051,0.566-0.35,1.101C45.916,6.183,46.033,6.316,46.5,6.25z"]
  },
  "CheckMarkIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M46.45,8.3C34.184,19.566,23.233,31.167,13.6,43.101c-1.266-0.201-2.616-1.367-4.049-3.5C8.017,37,6.95,35.217,6.35,34.25c-1.6-2.733-2.633-4.583-3.1-5.55c-0.934-2-1.483-4-1.65-6.001C3.7,20.4,5.967,21.05,8.4,24.649c1.333,1.934,2.9,4.734,4.699,8.4c1.667-1.467,6.4-5.967,14.201-13.5C33.4,13.684,38.267,9.383,41.9,6.65c1.4-1.067,2.217-1.634,2.449-1.701c0.834-0.266,1.533,0.301,2.101,1.701C46.517,6.883,46.367,7.283,46,7.85C45.801,8.217,45.95,8.367,46.45,8.3z"]
  },
  "CircledMinusIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M42.5,20.1c-0.567-3.066-1.851-5.133-3.851-6.199c2.2,3.799,2.967,7.799,2.3,12c-0.6,3.832-2.283,7.199-5.05,10.1c-2.733,2.867-5.933,4.633-9.601,5.299C22.3,42.033,18.483,41.266,14.85,39c-3.5-2.201-5.75-5.734-6.75-10.6c-0.934-4.602-0.567-9.018,1.1-13.25c-2.4,3.566-3.483,7.266-3.25,11.1c0.233,3.566,1.517,6.832,3.85,9.799c2.3,2.9,5.25,5.018,8.85,6.352c3.8,1.432,7.75,1.75,11.85,0.949c2.633-0.5,5.016-2.133,7.149-4.9C39.583,35.982,41,33,41.899,29.5C42.799,26.033,43,22.9,42.5,20.1zM24.75,11.05c-3.533-0.2-6.833,0.967-9.9,3.5c-2.967,3.5-4.1,7.134-3.4,10.899c0.633,3.301,2.483,6.1,5.55,8.4c3.033,2.232,6.25,3.283,9.649,3.15c3.767-0.133,6.717-1.768,8.851-4.9c2-2.867,2.699-5.967,2.1-9.3c-0.533-3.101-2.017-5.767-4.45-8C30.683,12.5,27.883,11.25,24.75,11.05zM15.05,4.449C22.183,0.449,29,0.05,35.5,3.25c6.799,3.333,10.166,9.783,10.1,19.35c-0.033,9.232-2.983,16.566-8.85,22c-4.367,1.768-8.783,2.518-13.25,2.25c-4.333-0.268-8.2-1.482-11.6-3.65C4.267,38.367,1.267,30.9,2.9,20.8C4.033,13.8,8.083,8.35,15.05,4.449z", "M31.799,22.65c0.067,0.399,0.234,1.066,0.5,2c0.168,0.832,0.134,1.383-0.1,1.648c-3.5,0.033-6.283,0.051-8.349,0.051c-3.867,0.033-6.383,0.018-7.55-0.051c-0.633-2.566,1-3.832,4.9-3.799c0.133,0,1.933,0.1,5.399,0.3C28.866,22.934,30.6,22.883,31.799,22.65z"]
  },
  "CircledPlusIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M25.451,16.85c0.133,1.633,0.332,2.768,0.6,3.4c1.133,1.166,2.982,1.783,5.55,1.85v3.051c-0.366,0.634-1.351,0.899-2.95,0.8c-1.633-0.133-2.717,0.217-3.25,1.051c0.566,1.199,0.85,2.299,0.85,3.299c0.033,1.033-0.25,2.201-0.85,3.5h-3.05c-0.533-1.232-0.733-3.5-0.6-6.799C21.283,26.4,20.1,26.134,18.2,26.2c-1.733,0.066-2.817-0.483-3.25-1.651c0.267-1.398,1.083-2.1,2.45-2.1c2.033,0,3.267-0.116,3.7-0.35c-0.133-2.867-0.083-4.883,0.15-6.05c0.367-1.967,1.55-2.833,3.55-2.601C25.067,14.083,25.283,15.217,25.451,16.85z", "M32.951,14.949C30.518,12.717,27.75,11.5,24.65,11.3c-3.433-0.233-6.667,0.899-9.7,3.399c-2.9,3.434-4.017,7-3.35,10.701c0.633,3.267,2.467,6.033,5.5,8.3c2.967,2.2,6.133,3.233,9.5,3.1c3.667-0.166,6.566-1.783,8.7-4.85c1.934-2.833,2.616-5.883,2.05-9.15C36.817,19.767,35.351,17.15,32.951,14.949zM15.2,4.8c7-3.933,13.7-4.333,20.101-1.2c6.633,3.301,9.934,9.65,9.9,19.051c-0.034,9.101-2.934,16.3-8.701,21.601c-4.299,1.732-8.633,2.466-13,2.199c-4.3-0.267-8.1-1.449-11.4-3.55c-7.5-4.8-10.45-12.149-8.85-22.051C4.35,13.982,8.333,8.633,15.2,4.8zM18.7,42.101c3.733,1.4,7.617,1.717,11.65,0.949c2.601-0.5,4.933-2.116,7-4.85c1.899-2.434,3.3-5.367,4.2-8.8c0.9-3.399,1.1-6.5,0.6-9.301c-0.533-3-1.8-5.017-3.8-6.05c2.167,3.767,2.933,7.7,2.3,11.801c-0.6,3.767-2.267,7.066-5,9.9c-2.699,2.866-5.85,4.616-9.449,5.25C22.267,41.7,18.517,40.934,14.95,38.7c-3.434-2.133-5.65-5.6-6.65-10.4c-0.933-4.501-0.566-8.833,1.1-13c-2.333,3.5-3.383,7.134-3.15,10.9c0.2,3.533,1.45,6.75,3.75,9.65C12.267,38.684,15.167,40.767,18.7,42.101z"]
  },
  "DotIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M31.05,11.4c4.567,4.467,6.5,8.982,5.8,13.55c-0.6,3.933-2.883,7.083-6.85,9.45c-3.834,2.333-7.584,2.982-11.25,1.949c-4.1-1.166-6.6-4.199-7.5-9.1c-0.867-4.8,0.334-8.75,3.6-11.85C18.483,11.934,23.883,10.6,31.05,11.4z"]
  },
  "DownFillTriangleIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M2.1,6.65c6.133-0.234,13.483-0.234,22.05,0c4.9,0.133,12.167,0.434,21.8,0.899c-0.566,2.967-2.184,6.067-4.85,9.3c-3.201,3.434-5.451,6-6.75,7.7c-1.101,1.434-2.967,4.301-5.601,8.6c-2.2,3.634-4.217,6.434-6.05,8.4c-4.933-7.4-8.617-12.967-11.05-16.699C7.383,18.184,4.2,12.117,2.1,6.65z"]
  },
  "GtCircledIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M42.05,12.25C40.55,9.717,37.634,7.634,33.3,6c-4.633-1.767-9.083-2.267-13.35-1.5c-4.3,0.8-7.917,2.884-10.85,6.25c-2.466,2.899-4.033,6.167-4.7,9.8c-0.9,5.066-0.15,9.617,2.25,13.65c2.167,3.633,5.35,6.4,9.55,8.3c4.067,1.866,8.233,2.566,12.5,2.101c4.533-0.534,8.25-2.268,11.149-5.201c2.767-2.799,4.417-6.967,4.95-12.5C45.399,21.167,44.483,16.283,42.05,12.25zM46.35,27.101c-0.767,7.433-3.116,12.633-7.05,15.6c-5.033,3.833-11.367,5-19,3.5c-7.567-1.467-12.867-4.85-15.9-10.15c-2.033-3.6-3.017-7.5-2.95-11.699c0.067-4,1.067-7.768,3-11.301c1.934-3.566,4.533-6.4,7.8-8.5C15.75,2.316,19.6,1.2,23.8,1.2c3.867,0,7.684,0.883,11.45,2.65c3.6,1.666,6.25,3.732,7.95,6.199C46.033,14.183,47.083,19.866,46.35,27.101z", "M21.65,9.2c-2,0.333-4.15,1.467-6.45,3.4c-3.133,2.699-4.8,6.466-5,11.299c-0.2,4.967,1.35,8.734,4.65,11.301c3.034,2.434,6.684,3.566,10.95,3.4c4.367-0.134,7.783-1.601,10.25-4.4c1.8-2.033,2.95-4.768,3.45-8.2c0.533-3.7,0.1-7.033-1.3-10c-1.066-2.267-2.95-4.067-5.65-5.399C29.35,9.033,25.716,8.566,21.65,9.2zM34,7.899c3.533,1.734,5.85,3.734,6.95,6c3.5,7.167,3.066,13.783-1.301,19.851c-4.199,5.866-9.85,8.483-16.95,7.851c-4.733-0.4-8.6-2.434-11.6-6.101c-2.767-3.366-4.25-7.333-4.45-11.899c-0.2-4.634,1-8.601,3.6-11.9c2.9-3.667,7.05-5.7,12.45-6.1C26.466,5.3,30.233,6.066,34,7.899z", "M34.3,22.8c-4.2,6.566-8.434,11.35-12.7,14.35c-1.6-2.732-1.167-5.266,1.3-7.6l3.3-2.75c1.3-1.066,2.166-2.033,2.6-2.9c-0.7-1.566-2-3.033-3.9-4.399c-2.267-1.399-3.917-2.517-4.95-3.351c0.4-4.199,2.517-4.75,6.35-1.649C26.533,14.7,29.2,17.467,34.3,22.8z"]
  },
  "GtIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M35.4,15.9c2.6,1.766,4.533,3.283,5.799,4.549c-2.366,3.935-6.75,8.45-13.149,13.551c-7.067,5.6-11.767,9.933-14.101,13c-1.699-0.601-3.166-1.767-4.399-3.5c-0.5-0.733-1.417-2.45-2.75-5.15c3.366-1.899,6-3.417,7.899-4.55c3.467-2.033,6.084-3.85,7.851-5.45c3.833-4.233,5.75-6.383,5.75-6.449c-0.033-0.567-1.217-1.834-3.55-3.801c-1.934-1.666-3.383-2.8-4.35-3.4c-1.234-0.799-3.567-1.85-7-3.149c-2.867-1.067-4.834-2.167-5.9-3.3c0.267-6.434,2.65-8.566,7.15-6.4C17.217,3.083,21.3,5.934,26.9,10.4C31.533,13.367,34.367,15.199,35.4,15.9z"]
  },
  "IpadKeyboard": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M44.703,32.047c-0.008,0-0.016,0-0.023,0L3.568,31.401c-0.819-0.013-1.477-0.681-1.477-1.5V8.54c0-0.435,0.188-0.848,0.517-1.133c0.328-0.284,0.765-0.411,1.193-0.353c0.039,0.006,3.944,0.554,8.33,0.854c5.609,0.381,10.907,0.461,14.364,0.461c2.247,0,3.619-0.034,3.629-0.035l14.557-0.215c0.399-0.008,0.789,0.15,1.075,0.432c0.286,0.282,0.447,0.667,0.447,1.068v20.928c0,0.402-0.161,0.787-0.448,1.069C45.475,31.893,45.097,32.047,44.703,32.047zM5.092,28.425l38.111,0.599V11.142l-13.019,0.192c0,0-1.399,0.035-3.689,0.035c-3.503,0-8.874-0.081-14.567-0.469c-2.741-0.187-5.254-0.462-6.836-0.654V28.425z", "M8.666,13.48c0.615-0.23,2.764,0.682,2.022,1.742c-0.147,0.211-0.833,0.805-1.28,0.973c-1.152,0.436-1.921-0.182-1.921-1.689C7.488,14.506,7.751,13.823,8.666,13.48z", "M13.659,13.48c0.615-0.23,2.765,0.682,2.024,1.742c-0.147,0.211-0.834,0.805-1.279,0.973c-1.153,0.436-1.922-0.182-1.922-1.689C12.481,14.506,12.745,13.823,13.659,13.48z", "M18.654,13.48c0.614-0.23,2.764,0.682,2.023,1.742c-0.148,0.211-0.835,0.805-1.281,0.973c-1.153,0.436-1.922-0.182-1.921-1.689C17.476,14.506,17.74,13.823,18.654,13.48z", "M23.648,13.48c0.615-0.23,2.764,0.682,2.024,1.742c-0.147,0.211-0.836,0.805-1.281,0.973c-1.153,0.436-1.922-0.182-1.921-1.689C22.47,14.506,22.734,13.823,23.648,13.48z", "M28.641,13.48c0.616-0.23,2.766,0.682,2.026,1.742c-0.149,0.211-0.835,0.805-1.281,0.973c-1.153,0.436-1.923-0.182-1.922-1.689C27.464,14.506,27.729,13.823,28.641,13.48z", "M33.636,13.48c0.615-0.23,2.765,0.682,2.021,1.742c-0.146,0.211-0.832,0.805-1.278,0.973c-1.155,0.436-1.923-0.182-1.922-1.689C32.457,14.506,32.721,13.823,33.636,13.48z", "M38.63,13.48c0.614-0.23,2.765,0.682,2.023,1.742c-0.147,0.211-0.833,0.805-1.278,0.973c-1.154,0.436-1.924-0.182-1.923-1.689C37.452,14.506,37.717,13.823,38.63,13.48z", "M8.666,18.483c0.615-0.231,2.764,0.681,2.022,1.741c-0.147,0.211-0.833,0.805-1.28,0.973c-1.152,0.437-1.921-0.182-1.921-1.689C7.488,19.508,7.751,18.826,8.666,18.483z", "M13.659,18.483c0.615-0.231,2.765,0.681,2.024,1.741c-0.147,0.211-0.834,0.805-1.279,0.973c-1.153,0.437-1.922-0.182-1.922-1.689C12.481,19.508,12.745,18.826,13.659,18.483z", "M18.654,18.483c0.614-0.231,2.764,0.681,2.023,1.741c-0.148,0.211-0.835,0.805-1.281,0.973c-1.153,0.437-1.922-0.182-1.921-1.689C17.476,19.508,17.74,18.826,18.654,18.483z", "M23.648,18.483c0.615-0.231,2.764,0.681,2.024,1.741c-0.147,0.211-0.836,0.805-1.281,0.973c-1.153,0.437-1.922-0.182-1.921-1.689C22.47,19.508,22.734,18.826,23.648,18.483z", "M28.641,18.483c0.616-0.231,2.766,0.681,2.026,1.741c-0.149,0.211-0.835,0.805-1.281,0.973c-1.153,0.437-1.923-0.182-1.922-1.689C27.464,19.508,27.729,18.826,28.641,18.483z", "M33.636,18.483c0.615-0.231,2.765,0.681,2.021,1.741c-0.146,0.211-0.832,0.805-1.278,0.973c-1.155,0.437-1.923-0.182-1.922-1.689C32.457,19.508,32.721,18.826,33.636,18.483z", "M38.63,18.483c0.614-0.231,2.765,0.681,2.023,1.741c-0.147,0.211-0.833,0.805-1.278,0.973c-1.154,0.437-1.924-0.182-1.923-1.689C37.452,19.508,37.717,18.826,38.63,18.483z", "M8.666,23.486c0.615-0.23,2.764,0.682,2.022,1.742c-0.147,0.211-0.833,0.804-1.28,0.973c-1.152,0.437-1.921-0.181-1.921-1.689C7.488,24.512,7.751,23.828,8.666,23.486z", "M16.534,24.739c15.117-0.472,20.656-0.076,18.662,0.947c-0.915,0.47-6.216,0.765-9.498,0.865c-3.958,0.122-7.217-0.176-9.656-0.319c-5.14-0.301-3.537-0.912-3.538-0.789C12.503,25.443,12.758,24.857,16.534,24.739z", "M38.63,23.486c0.614-0.23,2.765,0.682,2.023,1.742c-0.147,0.211-0.833,0.804-1.278,0.973c-1.154,0.437-1.924-0.181-1.923-1.689C37.452,24.512,37.717,23.828,38.63,23.486z", "M24.059,45.209c-0.275,0-0.551-0.075-0.794-0.228l-1.916-1.196c-0.031-0.02-0.061-0.04-0.09-0.062l-3.111-2.273l-3.518-2.787c-0.649-0.515-0.758-1.459-0.243-2.107c0.515-0.65,1.458-0.757,2.107-0.244l3.471,2.752l3.02,2.206l1.006,0.628l7.319-5.38c0.667-0.49,1.606-0.346,2.097,0.32c0.491,0.668,0.347,1.606-0.32,2.098l-8.139,5.982C24.684,45.111,24.372,45.209,24.059,45.209z"]
  },
  "LeftFillTriangleIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M42.1,1.2c0.233,6.366,0.217,14.016-0.05,22.949c-0.134,5.066-0.433,12.617-0.899,22.65c-3.101-0.566-6.317-2.25-9.65-5.05c-3.566-3.333-6.233-5.684-8-7.05c-1.467-1.134-4.45-3.066-8.95-5.801c-3.767-2.266-6.683-4.366-8.75-6.3c3.967-2.433,9.767-6.25,17.399-11.45C30.1,6.683,36.4,3.366,42.1,1.2z"]
  },
  "ListIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M47.199,12.4c-5.066,1.033-12.684,1.199-22.849,0.5c-5.467-0.4-13-1.184-22.6-2.351C1.25,9.283,1.067,8.033,1.2,6.8C1.367,5.334,1.25,3.867,0.85,2.4C3.45,3.233,7.667,3.633,13.5,3.6c3.3-0.033,8.434-0.1,15.4-0.199C34.699,3.5,39.016,4.1,41.85,5.2C45.482,6.633,47.266,9.033,47.199,12.4z", "M46.301,28.75c-7.635,1.166-15.434,1.617-23.401,1.35C14.633,29.767,7.283,28.7,0.85,26.9C1.083,25.8,1.167,24.434,1.1,22.8c-0.067-1.366,0.15-2.716,0.65-4.05c4.9-0.2,11.85-0.05,20.85,0.45c8.966,0.533,15.951,0.683,20.951,0.45c0.166,0.766,0.682,2.232,1.549,4.399C45.766,25.816,46.166,27.383,46.301,28.75z", "M44.449,45.1C35.383,45.3,28.5,45.45,23.8,45.55c-8.4,0.134-15.45-0.017-21.15-0.45c-0.566-1.732-0.966-3.5-1.2-5.3c-0.3-2.3-0.2-3.866,0.3-4.7c15.8-0.732,29.732-0.434,41.801,0.9c0.199,1.133,0.566,2.733,1.1,4.8C44.982,42.533,44.916,43.967,44.449,45.1z"]
  },
  "MicIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M32.65,9.5c-0.066-0.934-0.2-1.617-0.4-2.051c-0.3-0.566-0.816-0.85-1.55-0.85C30.533,8.434,31.184,9.4,32.65,9.5zM30.25,9.949c-0.434,1.201-0.416,2.234,0.05,3.101c0.434,0.866,1.217,1.434,2.351,1.7c0.133-1.233-0.033-2.25-0.5-3.051C31.75,11.066,31.117,10.483,30.25,9.949zM26.45,3.15c-0.267-0.167-0.767-0.301-1.5-0.4c0.033,1.233,0.683,1.867,1.95,1.9C27.033,3.949,26.883,3.449,26.45,3.15zM27.85,4.199C27.65,6.033,28.283,7,29.75,7.1c-0.066-0.366,0.05-0.75,0.35-1.15c0.234-0.366,0.283-0.799,0.15-1.299c-0.334,0.066-0.7-0.033-1.1-0.301C28.75,4.117,28.316,4.066,27.85,4.199zM23.05,5.65c0.233-2.434-0.566-3.567-2.399-3.4C20.483,4.117,21.283,5.25,23.05,5.65zM21.85,6.65c-0.467-1-1.184-1.5-2.149-1.5C19.334,7.316,19.967,8.767,21.6,9.5C22.233,8.633,22.316,7.684,21.85,6.65zM23.05,8.5c-0.767,2.166-0.283,3.617,1.45,4.35c0.533-0.732,0.684-1.65,0.45-2.75C24.684,8.9,24.05,8.367,23.05,8.5zM25.9,11.9c-0.801,2.533-0.15,3.967,1.949,4.299c-0.066-1.533,0.084-2.5,0.45-2.899c-0.3-0.033-0.649-0.316-1.05-0.851C26.916,11.983,26.467,11.8,25.9,11.9zM29.3,9c-0.267,0-0.55-0.2-0.85-0.6c-0.267-0.367-0.634-0.484-1.101-0.351C26.816,8.65,26.7,9.383,27,10.25c0.334,0.867,0.934,1.25,1.8,1.15C28.666,10.434,28.834,9.633,29.3,9zM29.1,2.15c2.4,1.133,4.018,2.649,4.851,4.549c1.2,2.768,1.466,5.534,0.8,8.301c-0.633,2.6-1.967,4.699-4,6.3c1.334,3.233,3.533,8.168,6.6,14.801c1.834,5.4,1.234,9.166-1.8,11.3c-2.733-1.399-5.083-4.55-7.05-9.45c-3.133-7.633-4.9-11.8-5.3-12.5c-5.333,0.268-9.167-1.651-11.5-5.751C9.733,16.3,9.184,12.566,10.05,8.5C11.15,3.367,15,0.75,21.6,0.65C24.334,0.583,26.834,1.083,29.1,2.15zM25.9,8.5c0.066-1.1,0.066-1.883,0-2.35c-0.134-0.801-0.601-1.134-1.4-1c-0.566,0.633-0.633,1.299-0.2,2C24.666,7.717,25.2,8.166,25.9,8.5zM28.85,16.75c-0.066,0.867,0.084,1.65,0.45,2.35c0.667,0,1.184-0.283,1.55-0.85c0.234-0.301,0.518-0.816,0.851-1.551c-0.167-0.133-0.534-0.582-1.101-1.35c-0.399-0.533-0.833-0.883-1.3-1.05C29.2,14.867,29.05,15.684,28.85,16.75zM23.05,17.65c0.4-0.934,0.567-1.884,0.5-2.851c-0.1-1.167-0.583-1.816-1.45-1.95C20.967,15.483,21.283,17.083,23.05,17.65zM20.15,20.5c0.232,0.066,0.55,0.233,0.949,0.5c0.4,0.233,0.734,0.4,1,0.5c0.234-1,0.268-1.717,0.101-2.15c-0.2-0.533-0.717-0.949-1.55-1.25c0.066,0.367-0.033,0.75-0.301,1.15C20.084,19.617,20.017,20.033,20.15,20.5zM26.65,19.55c0.3-0.467,0.383-0.95,0.25-1.45c-0.434-0.8-1.084-1.434-1.95-1.9C24.416,17,24.267,18.017,24.5,19.25c0.233,1.333,0.867,1.916,1.9,1.75C26.3,20.533,26.383,20.05,26.65,19.55zM27.25,24.1c-0.733,0.133-1.217,0.35-1.45,0.65c1.434,2.167,2.983,5.468,4.65,9.9c1.667,4.467,3.2,7.767,4.6,9.899c-0.1-2.967-0.884-6.732-2.35-11.299c-1.333-4.268-2.717-7.584-4.15-9.951C28.35,23.732,27.916,24,27.25,24.1zM20.55,11.05C20.25,9.916,19.633,9.233,18.7,9c-0.5,1.033-0.616,2.1-0.351,3.199c0.268,1.168,0.867,1.867,1.801,2.101C20.684,13.267,20.816,12.184,20.55,11.05zM18.25,6.1C18.184,5.767,18.283,5.4,18.55,5c0.233-0.4,0.284-0.834,0.15-1.301C18.1,3.267,17.467,3.117,16.8,3.25C16.6,4.717,17.084,5.666,18.25,6.1zM13.25,6.55c-0.166,0.7,0.217,1.2,1.15,1.5c0.232-0.733,0.399-1.333,0.5-1.8c0.199-0.834,0.033-1.367-0.5-1.6C13.733,5.517,13.35,6.15,13.25,6.55zM13.45,9.25c-0.067-0.767-0.4-1.167-1-1.2c-0.434,0.467-0.601,1.184-0.5,2.149c0.134,1,0.634,1.4,1.5,1.201C13.483,10.4,13.483,9.684,13.45,9.25zM16.8,10.449c0.467-1.166,0.7-2,0.7-2.5c0-0.866-0.55-1.316-1.65-1.35c0,0.7-0.066,1.5-0.199,2.4C15.684,9.6,16.066,10.083,16.8,10.449zM14.85,10.9c-0.333,0.799-0.434,1.566-0.3,2.299c0.167,0.768,0.601,1.284,1.3,1.551c0.5-0.566,0.617-1.301,0.351-2.2S15.483,11.1,14.85,10.9zM19.2,18.1c0.1-1.232,0.05-2.133-0.15-2.699c-0.267-0.768-0.85-1.134-1.75-1.101C16.7,16.5,17.334,17.767,19.2,18.1zM13.2,17.55c-0.067,0.8,0.333,1.317,1.2,1.55c0.434-1.166,0.666-1.949,0.699-2.35c0.066-0.801-0.333-1.301-1.199-1.5C13.467,16.35,13.233,17.117,13.2,17.55zM15.85,20.5c0.566,0.467,1.367,0.633,2.4,0.5c-0.2-0.733-0.4-1.25-0.6-1.551c-0.301-0.5-0.75-0.783-1.351-0.85c0.034,0.301-0.033,0.617-0.2,0.95C15.9,19.916,15.816,20.233,15.85,20.5z"]
  },
  "RadioButton": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M37.354,19.13c-0.598-1.588-1.834-4.021-3.707-7.299c-3.305-2.057-6.714-3.006-10.227-2.85c-3.279,0.156-6.206,1.21-8.782,3.161c-2.576,1.952-4.358,4.45-5.348,7.495c-1.067,3.252-1.067,6.647,0,10.187c0.573,1.899,2.16,3.917,4.763,6.05c2.602,2.134,5.008,3.357,7.221,3.67c4.293,0.65,8.157-0.312,11.592-2.889c3.305-2.473,5.191-5.621,5.66-9.445C38.839,24.659,38.448,21.966,37.354,19.13zM40.009,33.181c-1.092,2.992-3.227,5.257-6.4,6.792c-2.941,1.405-6.284,2.017-10.031,1.835c-3.721-0.183-7.078-1.094-10.07-2.732c-3.175-1.796-5.322-4.138-6.44-7.025c-1.431-7.703-0.573-13.961,2.575-18.775c3.201-4.814,8.08-7.168,14.638-7.064c6.193,0.104,10.968,2.823,14.325,8.158C42.091,19.91,42.56,26.181,40.009,33.181z"]
  },
  "RadioButtonIndeterminate": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M40.01,33.181c-1.094,2.992-3.227,5.257-6.402,6.792c-2.939,1.405-6.283,2.017-10.031,1.835c-3.721-0.183-7.077-1.094-10.07-2.732c-3.174-1.796-5.32-4.138-6.439-7.025c-1.432-7.703-0.572-13.961,2.576-18.775c3.2-4.814,8.08-7.168,14.637-7.064c6.193,0.104,10.969,2.823,14.324,8.158C42.091,19.91,42.56,26.181,40.01,33.181zM37.355,19.13c-0.6-1.588-1.834-4.021-3.709-7.299c-3.304-2.057-6.713-3.006-10.226-2.85c-3.278,0.156-6.206,1.21-8.782,3.161c-2.576,1.952-4.359,4.45-5.348,7.495c-1.066,3.252-1.066,6.647,0,10.187c0.572,1.899,2.16,3.917,4.762,6.05c2.603,2.134,5.01,3.357,7.221,3.67c4.294,0.65,8.158-0.312,11.593-2.889c3.305-2.473,5.191-5.621,5.659-9.445C38.838,24.659,38.448,21.966,37.355,19.13z", "M24.3,21.25c4.134,0.033,7.184-0.017,9.149-0.149c0.4,0.6,0.534,1.566,0.4,2.899c-0.066,0.7-0.199,1.667-0.4,2.899c-1.633,0-4.799-0.049-9.5-0.149c-3.666-0.101-6.783-0.101-9.35,0c-0.366-0.634-0.533-1.601-0.5-2.899c0.033-1.734,0.051-2.65,0.051-2.75C16.184,21.134,19.566,21.183,24.3,21.25z"]
  },
  "RadioButtonSelected": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M23.4,9c-3.268,0.134-6.184,1.183-8.75,3.149c-2.601,1.934-4.384,4.434-5.351,7.5c-1.066,3.234-1.066,6.617,0,10.15c0.567,1.9,2.15,3.917,4.75,6.051c2.601,2.133,5,3.366,7.2,3.699c4.3,0.633,8.166-0.333,11.6-2.9c3.301-2.467,5.2-5.616,5.7-9.449c0.3-2.533-0.1-5.217-1.2-8.051c-0.6-1.6-1.833-4.033-3.699-7.299C30.35,9.783,26.934,8.833,23.4,9zM38.6,14.351c3.5,5.566,3.967,11.85,1.4,18.85c-1.1,2.967-3.233,5.217-6.4,6.75c-2.934,1.434-6.266,2.05-10,1.85c-3.732-0.166-7.1-1.083-10.1-2.75c-3.166-1.767-5.316-4.1-6.45-7c-1.433-7.699-0.566-13.967,2.601-18.8C12.85,8.45,17.733,6.101,24.3,6.2C30.467,6.3,35.233,9.017,38.6,14.351z", "M17.75,30.649c-2.5-2.666-3.467-6.6-2.9-11.799c3.268-3.301,6.551-4.701,9.851-4.201c2.866,0.434,5.149,2.084,6.85,4.951c1.7,2.799,2.184,5.533,1.45,8.199c-0.867,2.967-3.066,4.783-6.6,5.45C22.9,33.884,20.017,33.017,17.75,30.649z"]
  },
  "RightFillTriangleIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M6.199,1.45c5.668,2.134,11.9,5.399,18.701,9.8c3.866,2.5,9.582,6.283,17.149,11.35c-2,1.9-4.866,3.984-8.601,6.25c-4.433,2.668-7.383,4.568-8.85,5.701c-3.434,3.133-6.066,5.467-7.9,7c-3.332,2.766-6.533,4.434-9.6,5C6.633,36.584,6.333,29.1,6.199,24.1C5.967,15.268,5.967,7.717,6.199,1.45z"]
  },
  "EmptyIconFontAwesome": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M43.099,47.049H4.878C2.74,47.049,1,45.309,1,43.17V5.049C1,2.91,2.74,1.17,4.878,1.17h38.221c2.139,0,3.879,1.74,3.879,3.879V43.17C46.978,45.309,45.238,47.049,43.099,47.049L43.099,47.049zM4.878,3.17C3.843,3.17,3,4.013,3,5.049V43.17c0,1.036,0.844,1.879,1.879,1.879h38.221c1.037,0,1.879-0.843,1.879-1.879V5.049c0-1.036-0.842-1.879-1.879-1.879H4.878L4.878,3.17z"]
  },
  // Browser Window
  "browserwindow": {
    type: "SVG_SCALE9",
    scale9: [77, 162, 93, 34],
    width: 455,
    height: 404
  },
  "browserwindow_nw": {
    type: "SVG_PATHS",
    width: 162,
    height: 77,
    colors: [0xc1c1c1, 0xffffff, 0xffffff, 0xffffff, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000],
    paths: ["M156.796,49.019c1.743,0.008,3.468,0.027,5.204,0.039v23.104L2.167,71.977V1.416c0,0,72.73-0.073,159.833-0.037v29.242l-4-0.039C155.43,35.17,155.111,43.227,156.796,49.019z", "M140,38c-4.801,1.729-3.339,3.818-4,10c-3.443,0.776-4.776-0.557-4-4c-1,0-2,0-3,0c-2.168-0.168-0.453,3.547-1,5c-7.909,2.156-0.435-13.834-8-9c1.687-6.312,7.555-8.445,9-15C134.801,27.199,135.19,34.811,140,38z", "M17.542,36.431c2.92-4.413,7.228-7.438,12-10c-1.748,7.081,7.791,2.876,12,4c0,4,0,8,0,12c-4.407,1.259-14.625-3.292-13,4C24.857,43.115,20.894,40.078,17.542,36.431z", "M79,37c-2.92,4.414-7.228,7.438-12,10c1.748-7.082-7.791-2.876-12-4c0-4,0-8,0-12c4.407-1.26,14.625,3.293,13-4C71.685,30.315,75.648,33.352,79,37z", "M2.167,2.834", "M3.419,77H0.436c0.078-19.752,0.159-38.33,0.232-55.118L0.762,2.811L1.08,0.554l1.084-0.001C41.179,0.496,99.815,0.471,162,0.462v2.5C100.388,2.971,42.262,2.995,3.253,3.051L3.25,3.074L3.168,21.893C3.102,37.04,3.03,53.65,2.959,71.26c52.67,0.68,105.947,0.884,159.041,0.864v2.501C109.062,74.644,55.942,74.44,3.417,73.766L3.419,77z", "M29.655,49.02c-0.167,0-0.335-0.033-0.494-0.102c-2.498-1.072-5.015-3.345-7.45-5.542c-1.895-1.71-3.685-3.325-5.343-4.298c-0.221-0.129-0.397-0.323-0.504-0.556c-0.569-1.237-0.209-2.187,10.712-12.436c0.768-0.72,1.459-1.368,2.02-1.901c0.329-0.313,0.801-0.424,1.235-0.287c0.433,0.135,0.759,0.494,0.851,0.939c0.219,1.061,0.381,2.117,0.481,2.852c1.129,0.111,2.614,0.096,4.064,0.055c0.672-0.019,1.344-0.037,2.012-0.037c1.493,0,2.638,0.094,3.714,0.301c0.496,0.096,0.885,0.48,0.987,0.976c0.744,3.612,1.889,10.368,1.191,15.024c-0.095,0.629-0.625,1.092-1.284,1.063l-10.89-0.415c0.033,1.094,0.048,2.549-0.067,3.303c-0.059,0.383-0.291,0.717-0.629,0.904C30.073,48.967,29.864,49.02,29.655,49.02zM18.603,37.541c1.58,1.088,3.204,2.553,4.784,3.979c1.698,1.533,3.44,3.105,5.092,4.184c-0.009-0.801-0.041-1.666-0.073-2.28c-0.019-0.351,0.111-0.693,0.357-0.944c0.247-0.25,0.589-0.379,0.938-0.371l11.082,0.422c0.308-3.86-0.486-8.938-1.122-12.195c-0.699-0.087-1.478-0.127-2.422-0.127c-0.646,0-1.294,0.018-1.943,0.036c-0.674,0.019-1.348,0.037-2.021,0.037c-1.352,0-2.445-0.079-3.44-0.247c-0.551-0.094-0.973-0.541-1.034-1.096c-0.001-0.01-0.062-0.559-0.177-1.344c-0.11,0.104-0.223,0.209-0.336,0.316C25.543,30.485,20.175,35.523,18.603,37.541z", "M67.051,50.451c-0.395,0-0.778-0.188-1.018-0.523c-0.584-0.819-0.687-2.018-0.786-3.176c-0.036-0.418-0.088-1.027-0.173-1.451c-1.134-0.109-2.629-0.092-4.097-0.052c-0.688,0.019-1.377,0.038-2.062,0.038c0,0,0,0,0,0c-1.584,0-2.798-0.103-3.936-0.333c-0.406-0.082-0.745-0.359-0.905-0.741c-0.213-0.508-0.317-1.108-0.427-1.744c-0.091-0.527-0.185-1.074-0.345-1.613c-0.054-0.184-0.065-0.375-0.034-0.562c0.186-1.104-0.024-2.359-0.247-3.688c-0.381-2.273-0.813-4.852,0.876-7.062c0.127-0.167,0.295-0.299,0.488-0.385c0.757-0.334,1.699-0.496,2.877-0.496c1.318,0,2.779,0.199,4.191,0.393c1.67,0.229,3.298,0.434,4.367,0.322c0.068-0.666,0.183-1.574,0.334-2.74c0.077-0.59,0.153-1.176,0.216-1.696c0.051-0.425,0.316-0.794,0.703-0.978c0.386-0.186,0.84-0.158,1.203,0.069c3.346,2.103,6.613,5.747,9.773,9.271c1.063,1.187,2.067,2.306,3.036,3.321c0.244,0.257,0.368,0.604,0.342,0.957c-0.026,0.353-0.201,0.678-0.481,0.895c-1.772,1.373-3.491,3.213-5.154,4.992c-2.572,2.754-5.232,5.602-8.253,6.883C67.38,50.419,67.214,50.451,67.051,50.451zM62.97,42.713c1.235,0,2.241,0.062,3.165,0.198c0.338,0.05,0.641,0.235,0.838,0.514c0.567,0.8,0.664,1.924,0.765,3.114c0.021,0.245,0.047,0.555,0.083,0.855c2.069-1.271,4.059-3.4,6.144-5.632c1.41-1.509,2.858-3.06,4.392-4.392c-0.703-0.763-1.422-1.564-2.169-2.398c-2.388-2.664-5.038-5.621-7.601-7.652c-0.16,1.24-0.336,2.637-0.336,3.069c0,0.521-0.323,0.987-0.812,1.171c-0.63,0.235-1.396,0.35-2.34,0.35c-1.223,0.001-2.627-0.191-3.985-0.378c-2.069-0.284-4.294-0.521-5.42-0.194c-0.801,1.309-0.524,2.961-0.207,4.854c0.231,1.38,0.47,2.805,0.29,4.226c0.154,0.58,0.25,1.132,0.335,1.626c0.035,0.204,0.069,0.405,0.105,0.586c0.775,0.108,1.638,0.157,2.698,0.157c0,0,0,0,0,0c0.663,0,1.328-0.019,1.994-0.037C61.596,42.73,62.285,42.713,62.97,42.713z", "M141.114,38.029c-3.295-6.477-6.644-11.275-10.238-14.67c-0.495-0.467-1.271-0.454-1.749,0.031c-3.986,4.042-7.421,9.674-10.328,14.637c-0.178,0.303-0.219,0.668-0.115,1.004c0.105,0.336,0.346,0.611,0.665,0.76c1.021,0.477,2.045,0.551,2.793,0.604c0.028,0.002,0.057,0.004,0.086,0.006c0.263,1.164,0.325,3.301,0.376,5.051c0.042,1.42,0.081,2.761,0.21,3.701c0.086,0.624,0.62,1.079,1.237,1.079c0.045,0,0.09-0.003,0.136-0.007c1.811-0.196,1.811-0.196,3.92-0.255c0.253-0.007,0.474-0.107,0.667-0.243c0.616-0.045,1.141-0.525,1.172-1.166c0.029-0.6,0.002-1.119-0.02-1.538c-0.014-0.271-0.027-0.521-0.02-0.714c0.25-0.026,0.462-0.045,0.634-0.057c0.193,0.791,0.433,1.867,0.481,2.584c0.005,0.081,0.04,0.15,0.06,0.226c0.018,0.069,0.025,0.141,0.054,0.204c0.039,0.086,0.098,0.154,0.154,0.229c0.039,0.051,0.068,0.109,0.114,0.154c0.073,0.07,0.163,0.118,0.251,0.17c0.048,0.027,0.087,0.066,0.14,0.088c0.147,0.061,0.306,0.096,0.474,0.096c0,0,0.001,0,0.001,0l0,0c0.683,0,1.162,0.057,1.625,0.11c0.846,0.099,1.646,0.192,3.085-0.119c0.551-0.119,0.955-0.593,0.984-1.156c0.054-1.027-0.044-2.298-0.148-3.644c-0.125-1.617-0.277-3.577-0.068-4.744c0.615-0.252,1.569-0.58,2.293-0.604c0.427-0.014,0.818-0.246,1.035-0.613C141.293,38.864,141.308,38.41,141.114,38.029zM136.141,38.429c-0.289,0.143-0.512,0.392-0.622,0.694c-0.562,1.545-0.375,3.943-0.196,6.263c0.064,0.823,0.125,1.618,0.148,2.311c-0.458,0.032-0.823-0.013-1.286-0.067c-0.239-0.027-0.507-0.059-0.816-0.083c-0.154-0.894-0.393-1.881-0.619-2.753l-0.15-0.582l-0.548-0.246c-0.658-0.295-1.593-0.25-2.984-0.074l-0.678,0.082l-0.318,0.376c-0.742,0.878-0.69,1.857-0.641,2.804c0.006,0.109,0.009,0.224,0.014,0.335c-1.1,0.032-1.549,0.052-2.257,0.118c-0.037-0.662-0.06-1.432-0.083-2.227c-0.064-2.172-0.129-4.411-0.51-5.824c0.005-0.309-0.082-0.604-0.256-0.86c-0.464-0.683-1.253-0.739-2.017-0.794c-0.165-0.012-0.34-0.023-0.521-0.042c2.419-4.077,5.18-8.44,8.248-11.806c2.78,2.855,5.433,6.68,8.062,11.619C137.097,37.967,136.276,38.362,136.141,38.429z", "M157.194,47.758c1.599,0.02,3.204,0.041,4.806,0.062v2.501c-1.839-0.023-3.681-0.048-5.515-0.07c-0.402-0.005-0.778-0.203-1.009-0.533c-2.875-4.105-1.937-10.364-1.183-15.393c0.153-1.017,0.298-1.984,0.401-2.873c0.074-0.637,0.653-1.09,1.259-1.105c1.995,0.028,4.009,0.054,6.046,0.077v2.5c-1.669-0.019-3.327-0.039-4.966-0.062c-0.082,0.588-0.173,1.201-0.268,1.834C156.113,39.049,155.308,44.412,157.194,47.758z", "M106.803,45.004c-3.074,0-4.136-2.5-4.912-4.325c-0.242-0.568-0.474-1.116-0.739-1.567c-0.7,0.467-1.094,1.15-1.535,1.918c-0.762,1.324-1.711,2.973-4.449,2.973l-1.107-0.027l-0.059-0.91c-0.254-3.947,1.833-5.559,3.673-6.98c0.28-0.216,0.55-0.424,0.81-0.639c-0.267-0.777-0.835-1.559-1.43-2.377c-1.203-1.654-2.566-3.529-2.193-6.206l0.136-0.981l0.983,0.127c2.842,0.369,4.993,3.82,5.983,6.252c0.399-0.465,0.741-1.07,1.094-1.697c1.022-1.811,2.293-4.113,5.729-3.689l0.821,0.102l0.053,0.826c0.256,3.98-2.601,6.227-4.686,7.868c-0.283,0.222-0.558,0.438-0.816,0.651c0.358,0.889,0.924,1.614,1.517,2.375c1.021,1.309,2.177,2.793,2.32,5.246l0.059,1.016L106.803,45.004z", "M103.74,34.097", "M101.114,36.902"]
  },
  "browserwindow_n": {
    type: "SVG_PATHS",
    width: 200,
    height: 77,
    colors: [0xc1c1c1, 0xc1c1c1, 0x000000, 0x000000, 0x000000, 0x000000],
    paths: ["M0,72.162V49.058c68.193,0.473,131.252,4.375,200,1.228v22.109L0,72.162z", "M0,30.621V1.379C67.017,1.407,142.542,1.5,200,1.742v30.811L0,30.621z", "M0,74.625v-2.501c49.237-0.018,98.316-0.228,146.604-0.437c17.619-0.076,35.463-0.152,53.396-0.221v2.5c-17.928,0.068-35.77,0.145-53.385,0.221C98.323,74.396,49.241,74.606,0,74.625z", "M0,2.962v-2.5c69.762-0.01,143.99,0.003,200,0.017v2.5C143.991,2.965,69.761,2.952,0,2.962z", "M0,32.923v-2.5c10.352,0.118,21.363,0.174,34.133,0.174c12.979,0,26.197-0.06,38.98-0.116c12.814-0.058,26.065-0.117,39.11-0.117c24.191,0,55.807,0.183,87.776,1.542v2.502c-31.96-1.361-63.583-1.544-87.776-1.544c-13.039,0-26.287,0.06-39.1,0.117c-12.786,0.057-26.007,0.116-38.991,0.116C21.365,33.097,10.354,33.041,0,32.923z", "M0,50.32v-2.501c21.099,0.274,42.487,0.646,63.23,1.009c40.805,0.713,82.999,1.449,124.327,1.449c4.15,0,8.297-0.011,12.443-0.025v2.498c-4.146,0.015-8.292,0.027-12.443,0.027c-41.35,0-83.555-0.736-124.37-1.449C42.457,50.965,21.083,50.594,0,50.32z"]
  },
  "browserwindow_ne": {
    type: "SVG_PATHS",
    width: 93,
    height: 77,
    colors: [0xc1c1c1, 0xffffff, 0x000000, 0x000000, 0x000000, 0x000000],
    paths: ["M90.711,72.5L0,72.395V50.285c0.861-0.04,1.713-0.066,2.576-0.108C4.093,45.441,4.366,37.539,3,32.582l-3-0.029V1.742c45.018,0.189,78.951,0.471,89,0.883C93.765,2.82,90.711,72.5,90.711,72.5z", "M33.089,50.707C29.363,50.148,23.708,49.275,23,42c-1.312-13.473,53.15-15.494,55.405-2.557C81.644,58.021,40.607,51.832,33.089,50.707z", "M91.998,77h-2.5c0.002-1.119,0.003-2.243,0.005-3.355c-0.154,0.068-0.323,0.105-0.502,0.105C59.513,73.767,29.638,73.854,0,73.967v-2.5c29.637-0.112,59.511-0.2,88.999-0.217c0.181,0,0.353,0.038,0.508,0.107c0.041-24.824,0.116-48.018,0.235-68.351C83.203,3.004,48.404,2.99,0,2.979v-2.5c53.044,0.014,89.75,0.028,90.834,0.029h1.424L92.25,1.765C92.116,23.92,92.038,49.528,91.998,77z", "M4.533,49.603c-0.158,0.741-0.311,1.455-0.444,2.133c-0.115,0.583-0.626,1.005-1.221,1.008C1.913,52.748,0.956,52.746,0,52.75v-2.498c0.613-0.002,1.228-0.001,1.841-0.004c0.079-0.381,0.162-0.771,0.247-1.168c0.998-4.674,2.228-10.438,0.38-14.568C1.645,34.475,0.822,34.442,0,34.407v-2.502c1.105,0.047,2.21,0.092,3.315,0.142c0.436,0.02,0.83,0.266,1.04,0.648C7.079,37.67,5.668,44.286,4.533,49.603z", "M57.357,53.295c-3.663,0-7.16-0.232-10.49-0.453c-4.982-0.332-9.688-0.645-14.064-0.125c-0.496,0.062-0.976-0.188-1.221-0.6c-0.025,0-0.05,0-0.075,0c-4.123,0-8.45-2.688-10.24-6.606c-0.928-2.03-2.384-7.287,4.111-12.216c8.529-6.473,45.484-6.346,51.354,0.174c3.144,3.492,4.039,9.334,3.088,12.964c-0.537,2.05-1.624,3.405-3.146,3.919C69.8,52.672,63.356,53.295,57.357,53.295zM37.994,49.939c2.918,0,5.926,0.2,9.039,0.407c8.551,0.569,18.24,1.214,28.842-2.364c0.911-0.309,1.347-1.5,1.526-2.184c0.769-2.937-0.063-7.92-2.527-10.657c-5.077-5.64-40.504-5.532-47.984,0.145c-3.613,2.741-4.802,6.004-3.349,9.187c1.574,3.447,5.834,5.738,9.29,5.011c0.543-0.112,1.072,0.14,1.336,0.591C35.427,49.979,36.702,49.939,37.994,49.939z", "M33.562,48.133c-0.365,0-0.728-0.159-0.975-0.467c-0.856-1.063-1.29-1.824-1.633-2.482c-0.284,0.001-0.602,0.025-0.931,0.05c-1.285,0.097-3.043,0.23-4.21-1.181c-0.956-1.01-1.251-2.605-0.775-4.43l0.08-0.307l0.217-0.232c0.743-0.795,1.391-1.203,1.911-1.533c0.247-0.156,0.461-0.291,0.628-0.436l0.21-0.182l0.268-0.075c2.467-0.695,4.646,0.831,5.526,2.698c0.784,1.666,0.539,3.448-0.574,4.722c0.275,0.503,0.623,1.066,1.23,1.821c0.433,0.537,0.348,1.324-0.189,1.757C34.114,48.043,33.838,48.133,33.562,48.133zM27.389,40.561c-0.11,0.572-0.15,1.363,0.299,1.84c0.372,0.445,1.058,0.423,2.148,0.34c0.502-0.038,1.016-0.077,1.534-0.045c0.737-0.757,0.432-1.681,0.246-2.074c-0.402-0.855-1.335-1.576-2.365-1.405c-0.225,0.169-0.449,0.311-0.668,0.448C28.189,39.913,27.814,40.15,27.389,40.561z"]
  },
  "browserwindow_e": {
    type: "SVG_PATHS",
    width: 93,
    height: 293,
    colors: [0x000000],
    paths: ["M92.128,209.112c0.049,30.582,0.095,59.18,0.114,83.888h-2.5c-0.021-24.708-0.065-53.301-0.114-83.884C89.523,143.368,89.399,66.261,89.498,0h2.5C91.9,66.257,92.023,143.365,92.128,209.112z"]
  },
  "browserwindow_se": {
    type: "SVG_PATHS",
    width: 93,
    height: 34,
    colors: [0xc1c1c1, 0x000000, 0x000000, 0x000000, 0x000000],
    paths: ["M89,34L0,33.737V6.559c30.58,0.176,60.633,0.396,89,0.673C97.165,7.311,89,34,89,34z", "M87.335,18.096c-4.375,6.457-7.805,7.892-11.986,9.842c-2.346-2.643,4.027-6.218,8.437-9.808C84.812,16.867,85.86,17.401,87.335,18.096z", "M80.631,15.917c-3.029,6-5.576,7.214-8.673,8.89c-1.916-2.651,2.775-5.785,5.968-9.024C78.648,14.615,79.473,15.177,80.631,15.917z", "M90.929,11.214c-0.032-0.126-0.063-0.254-0.096-0.381", "M92.096,21.156c0.177,3.572,0.359,7.266-0.146,10.67l-0.991-0.076c-0.054,0.635-0.582,1.136-1.232,1.143C59.976,33.203,29.961,33.351,0,33.396v-2.496c29.874-0.044,59.801-0.195,89.471-0.504l0.055-10.626c-0.117-2.561-0.188-5.153,0.04-7.617l0.012-2.122C59.964,8.764,30.025,8.031,0,7.633V5.136c30.08,0.398,60.076,1.13,89.748,2.399C89.746,5.071,89.744,2.549,89.742,0h2.5c0.004,3.831,0.006,7.576,0.008,11.213l0.001,1.251h-0.187l-0.038,7.254C92.048,20.198,92.072,20.679,92.096,21.156z"]
  },
  "browserwindow_s": {
    type: "SVG_PATHS",
    width: 200,
    height: 34,
    colors: [0xc1c1c1, 0x000000, 0x000000],
    paths: ["M200,33.737L0,33.146V5.944c60.675,0.057,131.654,0.223,200,0.614V33.737z", "M0,32.251v-2.51c9.376,0.062,18.738,0.217,28.04,0.51c55.555,0.405,113.654,0.734,171.96,0.648v2.496c-6.282,0.009-12.565,0.02-18.839,0.02c-51.975,0-103.577-0.303-153.17-0.665C18.706,32.457,9.359,32.312,0,32.251z", "M0,8.019v-2.5c10.734-0.077,21.44-0.16,32.07-0.245c33.902-0.271,68.265-0.546,102.733-0.546c21.719,0,43.479,0.12,65.196,0.408v2.497C143.831,6.888,87.351,7.331,32.09,7.773C21.454,7.858,10.741,7.941,0,8.019z"]
  },
  "browserwindow_sw": {
    type: "SVG_PATHS",
    width: 162,
    height: 34,
    colors: [0xc1c1c1, 0x000000],
    paths: ["M162,5.944v27.202L2.5,32.676L2.167,6.018C2.167,6.018,70.321,5.857,162,5.944z", "M162,5.519v2.5C109.414,8.397,55.806,8.584,3.258,7.554c0.098,7.993,0.208,15.596,0.332,22.737c31.249,1.05,63.208,0.522,94.129,0.012c17.853-0.295,36.064-0.595,54.286-0.595c3.332,0,6.665,0.012,9.995,0.033v2.51c-21.532-0.144-43.135,0.202-64.24,0.552c-31.344,0.519-63.757,1.052-95.447-0.054l-1.207-0.042v-1.42C0.937,21.655,0.793,11.165,0.67,0h2.932l0.003,5.061C56.037,6.082,109.525,5.896,162,5.519z"]
  },
  "browserwindow_w": {
    type: "SVG_PATHS",
    width: 162,
    height: 293,
    colors: [0x000000],
    paths: ["M3.419,0l0.183,293H0.67c-0.896-81.764-0.6-200.342-0.234-293H3.419z"]
  },
  "browserwindow_c": {
    type: "SVG_PATHS",
    width: 200,
    height: 293,
    paths: []
  },
  "browserwindow_padding_top": 8,
  "browserwindow_url_padding_top": 35,
  "browserwindow_url_padding_left": 160,
  "browserwindow_url_padding_right": 94,
  "browserwindow_scrollbar_padding_top": 74,
  "browserwindow_scrollbar_padding_bottom": 27,
  "browserwindow_scrollbar_padding_bottom_breakline": 0,
  "browserwindow_scrollbar_padding_right": 4,
  "browserwindow_border_thickness": 3,
  // All scrollbars
  "scrollbar_background_color": 0xFFFFFF,
  "scrollbar_background_radius": 3,
  // Vertical Scrollbar
  "verticalScrollbar": {
    type: "SVG_SCALE9",
    scale9: [14, 1, 2, 15],
    width: 16,
    height: 61
  },
  "verticalScrollbar_e": {
    type: "SVG_PATHS",
    width: 2,
    height: 32,
    paths: ["M1.931,0c0.223,9.823,0.206,22.011-0.004,32H0v-4.05C0.12,20.06,0.127,11.266,0,3.513V0H1.931z"]
  },
  "verticalScrollbar_n": {
    type: "SVG_PATHS",
    width: 13,
    height: 14,
    paths: ["M3.953,12.174c1.584,0,3.204-0.039,4.708-0.074c1.625-0.039,3.086-0.073,4.223-0.055c0.017,0.635,0.032,1.296,0.047,1.955H13V0.833c-0.034-0.018-0.065-0.04-0.104-0.05C9.262-0.157,4.109-0.119,0,0.862V14h1.149c-0.004-0.644-0.008-1.276-0.013-1.885C2.043,12.152,2.985,12.174,3.953,12.174zM1.047,5.679C1.086,4.7,1.12,3.689,1.051,2.684c3.673-0.77,8.217-0.773,11.342,0.035c0.03,0.008,0.06,0,0.09,0.006c0.138,2.067,0.251,4.551,0.341,7.32c-1.169-0.016-2.61,0.018-4.211,0.055c-2.37,0.057-5.011,0.112-7.323,0.018C0.926,8.734,0.984,7.244,1.047,5.679z", "M11.037,8.615c-2.312,0-4.625,0-6.938,0c0.296-2.53,2.254-3.399,3.083-5.396C10.083,3.402,10.188,6.38,11.037,8.615z"]
  },
  "verticalScrollbar_ne": {
    type: "SVG_PATHS",
    width: 2,
    height: 14,
    paths: ["M1.931,14H0V0.833c0.007,0.004,0.014,0.008,0.021,0.011c0.101-0.043,0.206-0.081,0.321-0.09c0.551-0.033,1.031,0.371,1.072,0.922c0.177,2.355,0.316,5.356,0.422,8.756c0.172,0.194,0.27,0.454,0.249,0.734c-0.015,0.2-0.101,0.37-0.213,0.519C1.894,12.438,1.913,13.214,1.931,14z"]
  },
  "verticalScrollbar_nw": {
    type: "SVG_PATHS",
    width: 1,
    height: 14,
    paths: ["M1,14H0.15c-0.01-1.685-0.024-3.255-0.042-4.681C-0.048,8.073,0,6.822,0.049,5.599c0.041-1.021,0.08-1.984,0.004-2.927C0.04,2.514,0.073,2.365,0.128,2.228C0.113,2.189,0.089,2.156,0.079,2.115c-0.133-0.536,0.194-1.078,0.73-1.212C0.871,0.888,0.938,0.878,1,0.862V14z"]
  },
  "verticalScrollbar_s": {
    type: "SVG_PATHS",
    width: 13,
    height: 15,
    paths: ["M12.925,0c-0.019,0.894-0.038,1.782-0.06,2.637c-0.073-0.018-0.143-0.045-0.221-0.045H1.188C1.184,1.749,1.182,0.87,1.18,0H0v14.564c0.096,0.03,0.193,0.06,0.299,0.06c0.007,0,0.014,0,0.021,0c0.196-0.004,0.367-0.078,0.52-0.179c1.816,0.572,5.165,0.588,9.263,0.531c1.203-0.017,2.263-0.03,2.898-0.001V0H12.925zM12.458,12.964c-0.646-0.008-1.443,0-2.383,0.013c-2.625,0.036-7.392,0.099-8.796-0.48c-0.04-2.332-0.065-5.007-0.082-7.904h11.447c0.06,0,0.111-0.023,0.167-0.034C12.716,7.799,12.598,10.668,12.458,12.964z", "M4.099,6.644c2.312,0.001,4.625,0,6.937,0.001c-0.296,2.529-2.254,3.399-3.083,5.396C5.053,11.856,4.948,8.879,4.099,6.644z"]
  },
  "verticalScrollbar_se": {
    type: "SVG_PATHS",
    width: 2,
    height: 15,
    paths: ["M1.393,14.067c0,0.006-0.004,0.01-0.004,0.016C1.388,14.087,1.39,14.09,1.39,14.094c-0.006,0.061-0.033,0.111-0.049,0.168c-0.018,0.062-0.024,0.127-0.053,0.183c-0.03,0.061-0.077,0.107-0.118,0.16c-0.034,0.043-0.06,0.094-0.101,0.131c-0.055,0.051-0.124,0.083-0.188,0.12c-0.041,0.022-0.075,0.056-0.118,0.073c-0.115,0.045-0.238,0.07-0.367,0.07c-0.022,0-0.046-0.001-0.069-0.002c-0.003,0-0.005-0.002-0.007-0.002c-0.007-0.001-0.013,0-0.02-0.001C0.216,14.986,0.112,14.98,0,14.976V0h1.926C1.807,5.682,1.626,10.652,1.393,14.067z"]
  },
  "verticalScrollbar_sw": {
    type: "SVG_PATHS",
    width: 1,
    height: 15,
    paths: ["M1,0v14.564c-0.396-0.126-0.691-0.481-0.701-0.921C0.226,9.923,0.192,5.174,0.179,0H1z"]
  },
  "verticalScrollbar_w": {
    type: "SVG_PATHS",
    width: 1,
    height: 32,
    paths: ["M1,0v32H0.179c-0.012-4.596-0.009-9.525-0.006-14.382C0.176,11.217,0.18,5.084,0.15,0H1z"]
  },
  "verticalScrollbar_c": {
    type: "SVG_PATHS",
    width: 13,
    height: 32,
    paths: ["M0,0h1.149c0.03,5.09,0.027,11.212,0.023,17.619C1.169,22.482,1.167,27.401,1.18,32H0V0z", "M12.931,0H13v3.513C12.98,2.312,12.957,1.146,12.931,0z", "M13,32h-0.075c0.028-1.319,0.054-2.664,0.075-4.05V32z"]
  },
  "verticalScrollbar_padding_top": 13,
  "verticalScrollbar_padding_bottom": 14,
  "verticalScrollbar_thumb": {
    type: "SVG_PATHS",
    width: 8,
    height: 21,
    paths: ["M0,0.769c1.157-1.083,5.073-0.668,7.407,1.084C7.407,8.853,8,13.02,8,20.02c-2.834,1.331-6.426,1.081-8-1.25C0,11.77,0.824,7.686,0,0.769z"]
  },
  // Horizontal Scrollbar
  "horizontalScrollbar": {
    type: "SVG_SCALE9",
    scale9: [1, 14, 15, 2],
    width: 61,
    height: 16
  },
  "horizontalScrollbar_e": {
    type: "SVG_PATHS",
    width: 15,
    height: 13,
    paths: ["M14.928,0.512c-0.009-0.061-0.006-0.123-0.024-0.179c-0.021-0.063-0.062-0.116-0.094-0.175C14.778,0.105,14.757,0.048,14.717,0H0v1.113c0.881,0.02,1.75,0.039,2.594,0.061C2.576,1.247,2.549,1.316,2.549,1.395v11.459C1.72,12.857,0.855,12.859,0,12.861V13h14.452c0.519-1.854,0.534-5.101,0.481-9.062c-0.019-1.416-0.035-2.639,0.018-3.198c0-0.004-0.001-0.007-0.001-0.011c0-0.006,0.004-0.01,0.004-0.016C14.959,0.643,14.937,0.578,14.928,0.512zM12.453,12.761c-2.332,0.04-5.007,0.065-7.904,0.083V1.395c0-0.06-0.023-0.111-0.034-0.168c3.241,0.096,6.11,0.215,8.406,0.354c-0.009,0.647,0,1.445,0.013,2.385C12.969,6.59,13.032,11.357,12.453,12.761z", "M6.602,3.003c2.529,0.296,3.399,2.254,5.396,3.083c-0.184,2.9-3.161,3.006-5.396,3.854C6.602,7.627,6.601,5.314,6.602,3.003z"]
  },
  "horizontalScrollbar_n": {
    type: "SVG_PATHS",
    width: 32,
    height: 1,
    paths: ["M32,1H0V0.106c9.83-0.223,22.018-0.204,32,0.007V1z"]
  },
  "horizontalScrollbar_ne": {
    type: "SVG_PATHS",
    width: 15,
    height: 1,
    paths: ["M14.717,1H0V0.113c5.663,0.119,10.619,0.3,14.024,0.533c0.005,0.001,0.008,0.004,0.013,0.004s0.009-0.002,0.014-0.002c0.064,0.007,0.118,0.037,0.179,0.055c0.063,0.019,0.131,0.027,0.188,0.057c0.058,0.03,0.1,0.077,0.149,0.117C14.618,0.917,14.675,0.949,14.717,1L14.717,1z"]
  },
  "horizontalScrollbar_nw": {
    type: "SVG_PATHS",
    width: 14,
    height: 1,
    paths: ["M14,1H0.932c0.167-0.209,0.413-0.354,0.701-0.376C4.793,0.387,9.105,0.217,14,0.106V1z"]
  },
  "horizontalScrollbar_s": {
    type: "SVG_PATHS",
    width: 32,
    height: 2,
    paths: ["M32,0v1.861c-4.603,0.013-9.542,0.01-14.409,0.005C11.202,1.863,5.08,1.858,0,1.889V0H32z"]
  },
  "horizontalScrollbar_se": {
    type: "SVG_PATHS",
    width: 15,
    height: 2,
    paths: ["M14.401,0.2c0.102,0.153,0.176,0.324,0.18,0.521c0.011,0.552-0.428,1.009-0.98,1.02C9.89,1.813,5.157,1.847,0,1.861V0h14.452C14.434,0.063,14.42,0.14,14.401,0.2z"]
  },
  "horizontalScrollbar_sw": {
    type: "SVG_PATHS",
    width: 14,
    height: 2,
    paths: ["M14,0v1.889c-1.706,0.01-3.294,0.023-4.735,0.042C8.635,2.01,8.005,2.04,7.377,2.04c-0.612,0-1.221-0.024-1.824-0.049c-1.02-0.042-1.983-0.08-2.925-0.004c-0.16,0.011-0.309-0.022-0.448-0.078C2.143,1.924,2.112,1.951,2.071,1.961C1.991,1.98,1.91,1.99,1.83,1.99c-0.449,0-0.856-0.304-0.97-0.759C0.761,0.833,0.68,0.417,0.599,0h5.244c0.851,0.033,1.677,0.049,2.477,0H14z"]
  },
  "horizontalScrollbar_w": {
    type: "SVG_PATHS",
    width: 14,
    height: 13,
    paths: ["M12.568,1.14c-0.03,1.017-0.003,2.235,0.028,3.519c0.067,2.794,0.141,5.915-0.361,8.242c0.571-0.005,1.165-0.008,1.766-0.012V13H8.319c0.651-0.04,1.288-0.118,1.9-0.293c0.514-2.079,0.445-5.206,0.377-8.001c-0.031-1.259-0.06-2.465-0.033-3.509C7.566,1.289,4.888,1.408,2.681,1.557C2.676,1.586,2.684,1.615,2.676,1.645c-0.81,3.127-0.807,7.671-0.035,11.344c1.005-0.068,2.015-0.035,2.993,0.005c0.07,0.003,0.139,0.004,0.208,0.007H0.599C-0.159,9.087-0.124,4.479,0.739,1.145c0.012-0.048,0.043-0.083,0.062-0.127C0.757,0.917,0.719,0.812,0.71,0.695C0.69,0.433,0.779,0.19,0.932,0H14v1.105C13.519,1.116,13.037,1.127,12.568,1.14z", "M8.572,3.002c0,2.312,0,4.625,0,6.938c-2.53-0.297-3.399-2.254-5.396-3.084C3.359,3.956,6.336,3.851,8.572,3.002z"]
  },
  "horizontalScrollbar_c": {
    type: "SVG_PATHS",
    width: 32,
    height: 13,
    paths: ["M32,13H0v-0.111c5.086-0.03,11.195-0.027,17.592-0.022c4.874,0.005,9.803,0.008,14.408-0.005V13z", "M32,0v1.113C22.021,0.902,9.823,0.881,0,1.105V0H32z"]
  },
  "horizontalScrollbar_padding_left": 13,
  "horizontalScrollbar_padding_right": 14,
  "horizontalScrollbar_thumb": {
    type: "SVG_PATHS",
    width: 21,
    height: 8,
    paths: ["M20.332,0C21.415,1.157,21,5.073,19.248,7.407C12.248,7.407,8.082,8,1.082,8C-0.25,5.166,0,1.574,2.332,0C9.332,0,13.415,0.824,20.332,0z"]
  },
  "calendar": {
    type: "SVG_PATHS",
    width: 96,
    height: 96,
    colors: [0xFFFFFF, 0x000000, 0x000000, 0x000000],
    paths: ["M0,0L0,96L96,96L96,0z", "M96,2.75l-1.5-1.625L16.875,0L0,1.125v93.75L1.375,96l20.375-0.625h73.375L96,62.875l-0.625-50.25L96,2.75zM93,93.875H46.625l-43.75-1v-71l89.875,1.25l0.875,42.75L93,93.875zM93,20.875l-44.75-0.5l-45.375-1.5v-16L34,2l58.75,0.875l0.875,1.25L93,20.875z", "M5.625,11.25l6.75-4.875V15.5L5.625,11.25zM82,7.25v9.625l7.625-4.25L82,7.25zM20.375,8l-0.125,7L22,15.625l0.375-3l2.875-0.75l-0.5-1H22.5l-0.375-2.75l4.25-0.25v-1.5h-6.75l0.125,1.25L20.375,8zM28.75,15.625h5l0.25-2L31.25,13.5v-1.25l2.375-0.375V10.75L31.5,10.5L31,9.125l2.75-0.875l-0.125-1.375L28.75,6.5L28,8.375l0.875,1L29.125,14l-0.375,0.938M43.25,11.938L42.938,15l-1.688,1.188l-3.188,0.125l-1-0.375l-0.688-1.25l0.562-1l-0.125-5.25l-0.5-0.688l0.625-0.625H40.5L41.812,8.5l-0.375,2.625L43.25,11.938zM38.188,9.188l0.438,1.188h1.188l0.312-0.938L39.5,8.188l-1,0.125L38.188,9.188zM41.5,13.312l-1-1L38.812,12.5l-0.562,1.188l0.312,0.938l1.125,0.562l1.562-0.375L41.5,13.312zM50,10.438l0.625,0.375l1-0.875l0.938-0.5l0.5,0.75l-0.312,1.688L49.812,14.5l-0.875,1.125l1.25,0.625l4.5-0.062l0.438-1.25l-2.688-0.25l0.25-0.938l1.375-1.25l0.688-1.938l-0.812-2.25L52.25,7.75L50.5,8.938L50,10.438zM62.188,12.812l-0.312,1.438l-1.438,1.625l-2.375,0.188l-1.625-2.125l-0.188-1.875l1.375-2.875l1.75-1.375l1.25,0.875l1.125,1.25L62.188,12.812zM60.812,12.062L60.5,10.75l-1.188-0.562l-1.5,1.875l0.25,1.5l0.812,1l1.5-0.562L60.812,12.062zM68.938,12.25l-0.25,2.375L67,16.188l-2.062,0.188l-1.75-2.125l-0.5-2.062l0.812-2.125l2.25-1.875l1.938,1.125L68.938,12.25zM67.938,12.312l-0.812-1.562l-1.188-0.25l-1.625,1.625v1.625l1.375,1.125h1l0.812-1.188L67.938,12.312zM76.188,15.062l-1.625,1.75l-2.125,0.562l-2.312-1.188l-0.562-2.125l0.625-1.375l-0.875-0.812l0.125-1.75l0.938-2.25l2-1l1.5,1.062l0.688,1.25l0.25,1.5l-0.75,1.25l1.812,1.312L76.188,15.062zM71.062,9.875L71.75,11l1.062-0.062l0.5-1.188l-0.938-0.875L71.062,9.875zM74.812,14.312l-1.438-1.188l-1.75,0.688l-0.125,1L72.25,16l1.5-0.062L74.812,14.312z", "M16.375,27.75L15,27.438l0.062,1.062l1.75,0.562l1.125,1v1.125l-0.375,1.562L15,33.062l-2-0.438v-1.562l0.5-0.188l0.875,0.75l1.562-0.125l0.125-1.062L15,29.812l-1.125-0.375l-0.938-1.375l0.25-1.5l2-1.062l1.625,0.5l0.438,1.062L16.375,27.75zM23.125,26.375l0.25,3.188l-0.125,2.25l0.062,1.688l0.812,0.062l0.938-1.688l-0.25-2.125l0.875,0.875l1.25-0.688L27,31.312l0.188,1.875l1.5,0.5l0.438-1.5l-0.25-5.312l-1.188-1.062L27,25.875l-0.562,1.938l-1.125,0.062l-0.875-1.688L23.125,26.375zM32.875,26.188l0.25,1.688l2.25,0.375l0.688,5.625L37,34l0.25-5.75l2.562-0.438L39.938,27l-1.188-0.75l-2.625-0.062L34.5,26.75L33.812,26l-1-0.125M44.042,26.542v1.583l2.125,5.917l0.583,0.25l0.75-1.625l1.042-0.208l0.625,1.458L50,33.959c0,0,1.042-3.375,1.042-3.625s0.208-4.25,0.208-4.25l-1.375-0.209l-1.042,4.5l-1.917,0.417l-0.958-3.625l-0.833-0.792L44.042,26.542zM53.917,27l0.208,0.917L56,28.584l0.917,0.5l-0.084,5.5l1.542,0.458l0.542-0.625l0.125-5.5l2-0.542l0.458-1.333l-2.417-0.375l-3.125,0.083L53.917,27zM64.042,27.25l0.291,1.542l0.792,0.958l-0.25,5.167l2.167-0.042l0.208-2.75l2.875-0.416v-0.667l-3.083-0.125l-0.375-1.792L70,28.792l-0.042-1.208l-4.208-0.167l-1.708-0.333V27.25zM76.584,27.084l-1.625,0.75l-1.041,1l0.125,1.666l1.416,0.791l1.709,0.584l1,0.75l-0.209,1.209c0,0-1.833,0.332-2.083,0.041s-1.125-1.084-1.125-1.084l-0.958,0.043l0.041,0.916l1.5,1.166l1.459,0.834l2.125-0.666l0.958-1.709c0,0-0.125-1.541-0.25-1.666s-0.875-0.834-1-0.875c-0.125-0.043-1.583-0.543-1.583-0.543l-1.625-0.75l0.125-0.707l1.666-0.084l1.459,0.209l0.166-0.793l-0.916-0.832L76.584,27.084zM67.166,39.333l-1.084,0.833l-0.041,0.917l0.708,0.542l0.208,0.083l0.042,4.625l1,0.208l0.083-5.458l-0.166-1.542L67.166,39.333zM74.957,40.875l0.25,1.083l0.959,0.375l0.041,0.833l-1.25,1.167l-1.166,0.875l0.166,1.125l3,0.125h2.125l0.084-1.292l-1.209-0.416l-2.083-0.084l1.958-1.583l0.209-1.958l-0.75-0.959L76.082,40L74.957,40.875zM11.916,47.25l0.208,0.875l1.458,0.708l1.042,0.375l0.167,1.167L13.916,50.5l-0.042,0.75l0.708,0.166L14.499,52l-1.583,0.25l-0.042,0.625l1.958,0.875h1.625l1.375-1.542c0,0-0.083-1.041-0.167-1.208s-1.542-1-1.542-1l0.167-1l0.708-0.875l-0.667-0.917l-3.292-0.167h-1.083L11.916,47.25zM23.124,47.083l-0.042,2.458l1.417,1.375l2.042,0.459l0.083,2.208l1.167,0.167l1-3.084l-0.708-0.583l-0.333-3.25l-1.75,0.208l0.292,1.084l0.25,1.291l-1.375,0.125l-0.542-0.458l0.167-1.792l-0.583-0.25L23.124,47.083zM35.166,47.166l-0.25,3.459l1.792,0.916l1.375,0.417l-0.25,1.083l-2.75-0.208l-0.083,0.583l1.458,1.375l1.75,0.042l1.417-2.083l0.125-1.667l-1.083-0.708l-2.333-0.417l0.042-1.208h2l0.333-1.209l-1.625-0.416L35.166,47.166zM51.082,53.833l-1.583,1.292l-2.667-0.334l-1.5-1.25l-0.333-3.166l0.75-1.584l1.5-1.625l2.542,0.125v0.75l-1.5,0.584l-1.25,0.958l0.167,1.167L49.915,51l1.042,1L51.082,53.833zM49.332,52.708l-0.708-0.833l-1.458,0.166v1.125l1.542,0.459L49.332,52.708zM55.332,48.083l0.292,0.833l1.542,0.959l1.333,0.25l-1.292,3.583l-0.166,1.083h1.875l1.375-4.583l0.541-1.583l-0.25-0.709l-4.416,0.042L55.332,48.083zM70.291,54.75c-0.084,0.167-1,1.25-1.125,1.25s-2.125-0.167-2.125-0.167L65.25,55l-0.417-1.292l0.708-1.125l0.792-1.125l-0.625-1.083l0.208-1.292l1.5-0.875L68.5,48l1.458,0.875c0,0-0.042,1.416-0.083,1.541s-1.542,0.875-1.542,0.875l1.25,0.459l0.75,0.791C70.333,52.541,70.375,54.583,70.291,54.75zM67,49.833l0.791,0.417l0.667-0.542l-0.083-0.583l-0.834,0.041L67,49.833zM68.875,52.958l-1.292-0.167l-0.792,0.667l0.084,0.917l1.083,0.041l0.917-0.458V52.958zM78.458,53.666l-0.25,2.25l-1.083-0.083l-0.292-3.542c0,0-1.75-0.333-1.958-0.375s-1.334-0.958-1.334-0.958L73.375,50l1-1.5l1.583-0.375l1.083,0.083L78,49.125L78.458,53.666zM76.625,49.666l-1.125-0.25L75,50.208l0.75,0.542h0.791L76.625,49.666zM11.042,57.166L11,58.833l0.125,5.083l0.917,0.042l0.333-3.292l-0.458-3.416L11.042,57.166zM18.083,59.166l-0.042,2.709l-0.958,1.666l-1.333,0.542l-1.458-0.792l-0.875-1.666l0.542-3l1.542-1l1.292,0.583L18.083,59.166zM16.542,59.291l-0.875-0.041l-0.833,1.5l0.208,1.708H16l0.542-1.958V59.291zM24.167,57.916l-0.042,1.334L24.542,60l0.167,2.166l0.042,1.75l0.917,0.042l0.292-5.625l-0.417-1L24.167,57.916zM28,57.5l-1.042,0.875c0,0-0.083,1.334,0.042,1.375s0.5,0.5,0.5,0.5l0.5,2.583L27.792,64l0.875,0.166L28.917,60l-0.333-2.25L28,57.5zM35.458,57.708l-0.333,0.708l-0.292,1l0.25,0.917v2.375l0.167,0.958h0.542l0.25-4.208L36,57.75L35.458,57.708zM36.916,59.041l-0.083,0.709l1.125,0.125l0.625,0.291L38.5,61.791l-0.625,0.667l-1.292,1.125l0.292,0.917h2.083L41,64.541l-0.167-1.166l-1.792-0.25H38.25l1.417-1.334l0.292-1.875l-0.292-1.458l-1.292-0.292l-1.458,0.709V59.041zM46.333,57.166l-0.625,1.209l-0.25,0.458l0.417,0.583l0.125,4.5L47.083,64l0.167-5.084l-0.167-1.083l-0.042-0.792L46.333,57.166zM48.999,58.541l0.5-0.375c0,0,0.709,0.167,0.709,0.292s-0.334,0.792-0.334,0.792l-0.875,0.166l-0.416,0.75l0.666,0.667l0.667,0.292l0.375,0.25l-0.042,0.75l-0.458,0.333l-0.667-0.375c0,0-0.541,0.75-0.625,0.917s0.125,0.541,0.125,0.541l0.834,0.417l1.333-0.083l1.042-0.417l0.041-1.292l-0.083-1.041l-1.083-0.917l0.708-0.667l0.417-0.875l-0.292-1.125l-1.375-0.5h-1.042l-0.958,1.084L48.999,58.541zM56.916,58.75l0.167,5.75l0.875-0.042l0.125-1.333l-0.167-4.292l-0.125-1.25l-0.708-0.208l-0.834,0.875L56.916,58.75zM59.291,60.708l0.667,0.625l0.958,0.25c0,0,0.292,2.458,0.292,2.792s0.5,0.458,0.5,0.458l0.958-0.25l0.333-3.917l-0.125-3.208l-1.375-0.083L61.083,58.5l-0.334,1l-0.5-0.625l-1.041-0.084L59.041,60L59.291,60.708zM66.291,58.166l-0.333,1.042l0.25,0.5l0.25,5l0.75,0.042l0.25-3.542l-0.125-2.25l-0.209-1h-0.833V58.166zM72.166,58l-1.25-0.5l-1.208,0.083l-0.625,0.875v1.583l0.041,0.709c0,0,1.75,0.208,1.875,0.208s0.5,0,0.5,0l0.125,1.208l-0.833,0.875l-1.5,0.209l-0.375,0.583l0.042,0.542l1.791,0.375l1.084-0.084c0,0,1.042-0.542,1.125-0.875s0.166-1.708,0.166-1.708l-0.208-1.458l-0.833-0.834l-1.584-0.166l-0.625-0.209l0.167-0.958l1.75,0.083L72.166,58zM76.291,58.166l-0.458,1.084l0.458,1l-0.125,4.583l1,0.125L77.624,60l-0.25-1.459l-0.375-0.416L76.291,58.166zM83.708,63.458c-0.083,0.167-0.375,1.25-0.417,1.375s-1.458,0.417-1.458,0.417l-1.792-0.667l-0.833-1.333V61l0.75-2.292l1.5-1.083h1.25v1.416h-0.625l-1.084,0.625l-0.333,1.125l0.667,0.042L82.666,61c0,0,0.958,0.75,1.042,0.875S83.791,63.291,83.708,63.458zM82.333,62.875l-0.792-0.709l-0.667,0.584l0.459,1.166h0.875L82.333,62.875zM12.123,68.792l-0.458,1.292l0.5,5.041l-0.042,1.542l1.458,0.167l0.083-1.542l-0.5-5.208l-0.083-1.209L12.123,68.792zM13.957,69l0.042,0.667l1.958,0.417l0.042,2.541l-0.083,2.25L15.248,76.5l1.75,0.417c0,0,0.042-2.126,0.042-2.292s0.625-3.958,0.625-3.958l0.042-1.917l-0.75-0.291l-2.542,0.125L13.957,69zM24.25,69l-0.688,1.062l0.562,0.812L24.25,76H25l0.25-6.375l-0.188-0.688L24.25,69zM30.667,75.417c0,0.167-1.042,1.333-1.208,1.416S27.625,76.5,27.625,76.5L26.5,75.167l-0.25-1.292l0.917-0.333l0.25-1.042l-0.958-1.292l0.125-1.25l1.333-1.166l0.833-0.084l1.292,0.917v1.5l-1.25,1.25l0.583,0.417c0,0,1.125,0.75,1.25,1S30.667,75.25,30.667,75.417zM27.625,70.667l0.417,0.875l0.875-0.375L28.625,70l-0.5-0.083L27.625,70.667zM29.375,74.458l-0.625-0.75l-0.625,0.084l-0.5,0.75c0,0,0.5,1,0.667,1s0.917,0,0.917,0L29.375,74.458zM36.083,68.625L35.5,69.667l0.5,0.625l0.042,5.5L36.958,76L37,69.083L36.5,68.5L36.083,68.625zM41.75,70.083l-0.083,6.667h-0.958L40.5,73.667l-1.25-0.542l-0.958-0.958l-0.583-1.125l0.333-1.542l1.833-0.708l1.167,0.166L41.75,70.083zM40.625,70.667l-0.875-0.75l-0.833,0.708l0.417,1.042l0.917,0.5L40.625,70.667zM47.417,70.042l0.708,0.458l-0.25,1.083l-1.75,1.5c0,0-0.417,0.834-0.417,0.959s0.583,0.916,0.583,1.041s1.458,0.084,1.458,0.084l1.417-0.292l-0.25-0.917l-1.25-0.208l0.458-0.583l0.917-0.959l0.125-2.125l-1.042-1L47.25,69c0,0-0.875,0.625-1,0.75s-0.125,0.792-0.125,0.792l0.542,0.416l0.5-0.208L47.417,70.042zM53.084,70.333c0,0-0.042,1.792-0.042,1.917s-0.333,2.458-0.333,2.458l-1.042,0.5l-1.375-0.375l-0.458-1v-1.5l0.416-2.125l1.209-0.916l0.791,0.291L53.084,70.333zM52,71.375l-0.333-0.667L51.292,71L51,72.083v1.625l0.334,0.125c0,0,0.457-0.375,0.541-0.583S52,71.375,52,71.375zM58.292,70.208l0.583,0.125l0.25,1.167l-0.958,1.167l-1.5,1.416l-0.333,1l0.375,0.667L58.25,76L60,75.875V75l-0.791-0.042l-1.042-0.333l0.792-1.083l0.958-0.917l0.042-1.917L59.417,69.5l-0.5-0.625l-1.042,0.042l-0.708,0.541l-0.333,0.834l0.666,0.583l0.709-0.167L58.292,70.208zM62.125,69.083L61.5,70.458l0.417,0.709l0.042,4.583h0.75L63,69.167l-0.375-0.084H62.125zM67.834,69.917l0.5,0.208l0.25,0.833l-0.459,1.25L66.5,73.833l-0.375,1l0.459,0.917c0,0,1.875,0.208,2.041,0.208s1.542-0.125,1.542-0.125l0.167-1.041l-1.292,0.041l-1.125-0.25l0.875-1.041l0.583-1.209l0.459-1.416l-0.209-1.042l-0.5-1.042L68.25,68.5l-1,0.667l-0.458,0.708v1L67.667,71L67.834,69.917zM72.042,70l0.667,0.125l0.25,0.958l-0.667,1.084L70.667,74l-0.333,0.792l0.25,0.916L72.792,76l1.792-0.167l-0.25-0.916l-2.459-0.209l0.709-0.833L73.417,73L74,71.708L73.75,70l-0.708-1.125h-0.875l-0.833,0.292l-0.5,1.041L71.084,71l0.583,0.042L72.042,70zM78.667,70.291l0.708,0.5l-0.291,1.25L77.5,73.583l-1.416,1.125l-0.209,0.75l0.959,0.375L79.209,76l1.458-0.042l-0.042-0.875l-2.208-0.125l1.042-1.125l0.75-1.292L80.5,70.5l-0.583-1.292h-1.542L77,69.958l-0.333,1.083l0.958,0.542l0.584-0.625L78.667,70.291zM80.834,69.583l0.5,0.625L82,69.833l1.292-0.5l0.417,0.25l-0.625,0.5l-0.834,0.625c0,0-0.041,0.417-0.125,0.542s0.042,1.083,0.042,1.083l0.625,0.208c0,0,0.542,0.292,0.667,0.417s0.625,0.542,0.625,0.542s-0.042,0.791-0.042,1s-0.333,0.208-0.5,0.208s-0.792,0-0.792,0l-1.125,0.042L81.5,75.458l0.417,0.5l1.333,0.417l2.084-0.334l0.291-1.541l-0.166-1.792l-0.667-0.75l-0.708-0.333l0.083-0.875l0.75-0.584l-0.292-1.125l-1.208-0.833l-1.042,0.542l-1.125,0.666L80.834,69.583zM12.834,82.291l0.667,0.5c0,0,0.083,1.125,0.042,1.25s-1.208,1.042-1.208,1.042l-1.583,1.666l-0.417,0.875l1.25,1.209l3.375,0.166l-0.042-1.041l-2.708-0.334l0.417-0.541l1.083-0.834l0.75-1.166c0,0,0.292-1.375,0.292-1.542s-0.042-1.625-0.042-1.625l-0.958-0.75c0,0-0.958-0.208-1.333-0.167s-1.167,0.667-1.333,0.792s0.167,0.917,0.167,0.917l0.542,0.541l0.583-0.125L12.834,82.291zM15.042,81.541v1.708l0.208,1.25l0.958,0.459l0.625,0.416l0.292,3l1.375,0.167c0,0,0.458-1.208,0.458-1.333s0.208-2.375,0.208-2.5s0-1.209,0-1.209l-0.208-2.625l-1.542-0.333l-0.125,2.042l-0.375,1.208l-0.833-0.25l-0.083-2.458L15.042,81.541zM24.334,82.708l0.458-0.625l0.875,0.083l0.333,0.375l-0.25,1.208l-0.583,0.75c0,0-1.167,0.5-1.375,0.625s-0.875,0.959-1,1.167s-0.458,1.083-0.458,1.083l0.458,0.792l0.958,0.208l2.083,0.167l1.042-0.333v-1.167l-2.167-0.125l0.375-0.833l1.792-1.292l0.25-1.417c0,0-0.292-2.041-0.292-2.208s-0.792-0.5-0.792-0.5s-0.917-0.125-1.083-0.125s-1.208,0.708-1.25,0.833s-0.708,0.959-0.708,0.959v0.916l1.167,0.292L24.334,82.708zM32.001,80.666l-1.875,0.083l-1.375,0.125l-0.042,1.334l-0.417,1.75l0.833,0.791l1.5,0.584l0.125,1.166l-0.083,0.584l-1.833-0.125l-0.667,0.375c0,0,0.25,0.541,0.375,0.666s2.167,0.834,2.167,0.834l1.333-0.667c0,0,0.417-0.959,0.542-1.292s0.208-1.167,0.208-1.375s-0.208-1.041-0.292-1.291s-1.292-0.792-1.292-0.792l-0.958,0.042l-0.833-0.209l0.25-1.166l0.542-0.334h1.25l0.917,0.084l0.083-1.125L32.001,80.666zM37.209,82.124h0.625l0.167,0.833l-0.5,1c0,0-1.292,1.5-1.458,1.584s-1.042,1.333-1.042,1.333l0.125,0.917l1.708,0.25l2.125-0.167l-0.125-0.875l-1.958-0.042l0.375-0.583l1.333-1.333l0.708-1.584l-0.25-1.291l-0.5-1.042l-0.958-0.25l-1.5,0.417c0,0-0.292,0.708-0.292,0.833s0.333,0.667,0.333,0.667l0.792-0.042L37.209,82.124zM43.792,85.999c0,0.167,0,1,0,1.125s-0.417,0.875-0.417,0.875l-0.875,0.542l-1.208-0.25l-1.417-0.667c0,0-0.25-0.667-0.25-0.875s-0.25-1.458-0.25-1.625s0.333-1.083,0.333-1.083l0.875-2.042l1-1.042l0.708-0.041l0.208,0.583l-0.25,0.875c0,0-0.625,0.417-0.75,0.5s-0.625,0.5-0.708,0.667s-0.042,0.958-0.042,0.958l0.583,0.25c0,0,1.125,0.083,1.292,0.083s0.833,0.209,0.833,0.209S43.792,85.832,43.792,85.999zM42.417,85.624l-1.042-0.042l-0.542,0.417l0.083,0.875l1.083,0.417l0.417-0.625V85.624zM48.293,81.582l0.75,0.25l0.167,0.959c0,0-0.417,1.25-0.5,1.375s-1.5,0.625-1.5,0.625l-0.833,1.25l0.042,1.291l0.875,0.584l2.792-0.25l0.292-0.417v-0.75l-2.042-0.042l0.125-0.5l1.25-1.041l0.542-1.25c0,0-0.167-1.709-0.167-2.042s-0.208-0.833-0.208-0.833s-1.292-0.167-1.417-0.167s-0.875,0.208-0.875,0.208l-0.292,0.917l0.458,0.458L48.293,81.582zM51.96,81.499h0.625l0.292,0.25c0,0-0.584,1.583-0.709,1.958s-0.333,1.667-0.416,2.042s-0.292,0.75-0.417,1s0.083,0.75,0.083,0.75l0.792-0.042l0.5-0.75c0,0,0.333-1.708,0.333-1.875s0.459-1.291,0.542-1.5s0.625-1.333,0.625-1.458s0.208-0.75,0.208-0.917s-0.333-0.541-0.333-0.541l-0.667-0.042c0,0-1.541-0.042-1.708-0.042s-0.583,0.167-0.667,0.292s0,0.792,0,0.792l0.584,0.5L51.96,81.499zM58.419,81.874l0.084-0.542l0.708-0.125l0.792,0.125l0.291,0.75c0,0-0.916,0.958-1.083,1.083s-0.625,0.709-0.833,0.917s-1,1-1.209,1.167s-0.875,1.291-0.875,1.458s0.375,1.292,0.375,1.292s2.125-0.042,2.25-0.042s0.875,0.042,1.125,0.042s0.417-0.167,0.417-0.167l-0.042-0.833l-0.666-0.167h-1.5l0.375-0.417c0,0,0.708-1.291,0.875-1.458s1.166-1.042,1.166-1.042s0.542-1.291,0.542-1.416s-0.042-1.459-0.042-1.459s-1.041-0.958-1.208-1s-1.125,0-1.417,0.125s-0.833,0.209-1.041,0.334s-0.334,1.166-0.334,1.333s0.459,0.792,0.459,0.792h0.833L58.419,81.874zM65.544,86.624c0,0.209-0.25,1.208-0.416,1.541s-1.25,0.5-1.417,0.5s-1.417-0.333-1.417-0.333l-1.083-0.833c0,0-0.083-1.209-0.042-1.459s0.25-0.75,0.25-0.75l0.375-0.458l0.125-0.458l-0.333-0.167l-0.292-0.5l-0.083-1.208l0.708-1.334l0.917-0.666l0.458,0.041l0.959,0.417l0.458,1.208l-0.708,1.25l-0.417,0.709c0,0,0.875,0.458,1,0.5s0.542,0.333,0.667,0.458S65.544,86.415,65.544,86.624zM62.336,82.915l0.417,0.25c0,0,0.625-0.375,0.625-0.583s0.041-0.958,0.041-0.958l-0.125-0.334l-0.75,0.459L62.336,82.915zM64.128,85.999l-0.459-0.709l-0.75,0.292l-0.208,0.708v0.834l0.833,0.208l0.584-0.208V85.999zM69.378,82.04l0.666,0.042l0.584,0.417l-0.125,1.291l-1.792,1.375c0,0-1.417,1.584-1.542,1.667s-0.375,0.5-0.375,0.5l0.084,1.333l1.5,0.084l3-0.084l0.333-0.458c0,0-0.417-0.583-0.667-0.583s-2.125-0.334-2.125-0.334l0.5-0.458l1.125-1.208c0,0,0.709-0.917,0.834-1.042s0.666-1.042,0.666-1.042l-0.375-1.583l-0.5-0.958l-1.416-0.417l-0.625,0.333l-1.042,1.209l0.125,0.75l1,0.041L69.378,82.04zM76.003,86.374c0,0.125,0,1.083,0,1.083l0.083,1.333l-0.958,0.209l-0.542-0.292c0,0-0.042-1.749-0.042-1.958s0-2.459,0-2.459l-1.041,0.125l-0.792-0.458l-0.208-0.75l-0.292-0.5l0.542-1.375l0.583-0.917l1-0.166l1.125,0.083l0.75,0.5l-0.125,1.917l-0.083,1.708C76.003,84.457,76.003,86.249,76.003,86.374zM75.086,81.915l-0.542-0.583l-0.166,0.042c0,0-0.709,0.291-0.792,0.458s0.292,1.25,0.292,1.25l0.708,0.083L75.086,81.915z"]
  },
  "calendar_color": 0x000000,
  "circleColor": 0xCCCCCC,
  "calendar_padding_top": 2,
  "calendar_padding_bottom": 5,
  "calendar_padding_left": 6,
  "calendar_border_weight": 3,
  "calendar_lines_weight": 2,
  "calendar_icon_size": 12,
  "redx": {
    type: "SVG_PATHS",
    width: 240,
    height: 104,
    paths: ["M205.995,79c-11.521-4.029-21.931-11.279-33-15c-6.87-2.311-23.101-8.55-36.16-13.62c-0.02-0.008-0.038-0.015-0.058-0.022c0.627-0.188,1.218-0.357,1.218-0.357c34.489-10.18,63.979-25.359,100-34c1.069-6.4-1.91-8.76-4-12c-32,8.18-64.851,22.7-95,33c-1.983,0.684-10.679,3.441-19.992,6.401C118.35,43.139,117.995,43,117.995,43c-4.56-0.109-12.94-4.561-20-7c-31.341-10.82-57.25-23.33-84-36h-3c-13.32,11.57,7.739,15.82,18,20c22.659,9.23,48.14,22.85,69,28l2.68,1.21c-10.19,3.21-19.069,5.99-21.68,6.79c-14.261,4.359-40.28,15.73-56,20c-9.64,2.62-20.55,0.59-23,9v2c1.49,0.852,1.45,3.221,2,5c30.069-4.029,56.569-17.33,86-26c4.88-1.439,17.84-5.59,29.34-9.3l0.66,0.3c8.93,2.59,89.72,32.58,116,47h3c1.01-1.33,0.95-3.721,3-4v-7C230.346,86.26,218.064,83.221,205.995,79z"]
  },
  "redx_color": 0x771822,
  "textinput_label_padding_top": 7,
  "textinput_label_padding_bottom": 7,
  "textinput_label_padding_left": 7,
  "textinput_label_padding_right": 7,
  "textinput_empty_space_bottom": 0,
  "textinput_border_padding": 4,
  "colorpicker": {
    type: "SVG_PATHS",
    width: 26,
    height: 28,
    paths: ["M7.311,24.095C7.311,24.095,7.311,24.095,7.311,24.095c-1.955,0-3.671-0.129-5.248-0.393l-0.743-0.125L1.235,22.83c-0.439-3.833-0.249-7.73-0.064-11.5c0.146-2.994,0.297-6.09,0.116-9.04L1.224,1.28l1.011-0.05C6.19,1.032,9.796,0.936,13.258,0.936c3.74,0,7.302,0.118,10.584,0.351l0.909,0.064l0.021,0.911c0.032,1.421,0.092,2.867,0.153,4.326c0.219,5.267,0.446,10.712-0.596,15.801l-0.111,0.547l-0.523,0.191c-0.691,0.254-1.452,0.283-1.965,0.283c-0.317,0-0.636-0.013-0.943-0.025c-0.262-0.011-0.515-0.021-0.751-0.021c-0.436,0-0.603,0.041-0.663,0.064L19.2,23.497L19.016,23.5c-1.675,0.025-3.397,0.15-5.221,0.283C11.68,23.937,9.492,24.095,7.311,24.095z M3.144,21.843c1.262,0.169,2.633,0.252,4.167,0.252h0c2.109,0,4.259-0.156,6.339-0.307c1.796-0.13,3.496-0.253,5.176-0.285c0.486-0.146,1.192-0.153,2.042-0.116c0.491,0.021,0.99,0.065,1.621-0.019c0.854-4.681,0.643-9.77,0.437-14.698c-0.048-1.16-0.096-2.312-0.13-3.45c-2.982-0.189-6.187-0.285-9.538-0.285c-3.135,0-6.392,0.08-9.925,0.243c0.104,2.739-0.032,5.535-0.165,8.249C3,14.863,2.827,18.407,3.144,21.843z", "M13.072,18.061c3.643-0.345,6.508-1.046,9-0.106c-1.319,3.368-1.835,6.428-5.293,9.046C15.987,25.476,12.456,22.847,13.072,18.061z"]
  },
  "colorpicker_padding_left": 2,
  "colorpicker_padding_right": 2,
  "colorpicker_padding_top": 2,
  "colorpicker_padding_bottom": 6,
  "textarea_padding_left": 8,
  "textarea_padding_right": 6,
  "textarea_padding_top": 8,
  "textarea_padding_bottom": 4,
  "textarea_scrollbar_padding_right": 2,
  "textarea_scrollbar_padding_top": 2,
  "textarea_scrollbar_padding_bottom": 2,
  "breadcrumbs_separator": {
    type: "SVG_PATHS",
    width: 8,
    height: 10,
    paths: ["M8,4c0,0.333,0,0.667,0,1C6.219,6.337,5.374,7.365,4,9c-0.812,0.278-0.501,1-2.418,1C4.757,5.8,4.791,4.157,2,0C3.749,0,4.791,2.449,8,4z"]
  },
  "breadcrumbs_padding_top": 4,
  "breadcrumbs_padding_bottom": 4,
  "breadcrumbs_padding_middle": 4,
  "piechart_background": {
    type: "SVG_PATHS",
    width: 180,
    height: 180,
    colors: [0xA7A7A7, 0xE6E6E6],
    paths: ["M90,90l-45.006,77.953C18.097,152.391,0,123.31,0,90C0,40.295,40.294,0,90,0V90z", "M180,90c0,49.706-40.294,90-90,90c-16.396,0-31.767-4.387-45.006-12.047L90,90V0C139.706,0,180,40.295,180,90z"]
  },
  "piechart_border_color": 0x000000,
  "piechart_min_thickness": 4,
  "piechart_max_thickness": 8,
  "linechart_axis_color": 0x000000,
  "linechart_axis_thickness": 4,
  "linechart_min_thickness": 4,
  "linechart_max_thickness": 8,
  "linechart_series1_color": 0xE6E6E6,
  "linechart_series2_color": 0xA7A7A7,
  "webcam": {
    type: "SVG_PATHS",
    width: 177,
    height: 146,
    colors: [0xffffff, 0x000000],
    paths: ["M174.84,143.78H1.4V1.38h173.44V143.78z", "M73.586,43.336c0.884,1.374,1.264,2.769,0,5.25c-1.863,1.533-3.723,0.857-5.25,0C64.252,44.969,69.969,39.252,73.586,43.336zM103,48c1.367-1.279,2.516-4.275,0-6c-3.15-1.688-6.693,1.886-5,5C99.252,48.388,101.09,49.295,103,48zM176.499,143.748c-0.001,0.02-0.011,0.034-0.013,0.053c-0.008,0.086-0.033,0.163-0.057,0.244c-0.022,0.075-0.037,0.15-0.072,0.219c-0.035,0.067-0.084,0.122-0.129,0.184c-0.052,0.066-0.098,0.134-0.16,0.188c-0.053,0.047-0.114,0.078-0.176,0.115c-0.078,0.048-0.152,0.094-0.24,0.124c-0.022,0.008-0.039,0.024-0.062,0.03c-2.308,0.656-23.944,0.885-53.972,0.885c-30.635,0-70.002-0.236-106.494-0.497c-7.737-0.056-12.843-0.09-13.838-0.089l-1.254,0.004v-1.254c0-17.38-0.06-35.633-0.117-53.283C-0.18,60.814-0.28,29.941-0.107,1.114L-0.1-0.132l1.246,0.003C22.556-0.07,38.648,0.046,55.687,0.17C84.679,0.379,114.652,0.588,175.5,0.583c0.689,0,1.25,0.56,1.25,1.25V16.5c0,0.008-0.005,0.015-0.005,0.023c0,0.008,0.004,0.014,0.004,0.022c-0.748,20.888-0.431,42.571-0.124,63.541C176.932,101.083,177.249,122.795,176.499,143.748zM15.144,142.792c1.897,0.015,3.773,0.026,5.637,0.04c1.584-23.075,19.102-41.98,36.499-51.135C45.563,81.691,40.335,67.9,42.926,53.502c2.544-14.135,12.331-26.261,25.444-32.02c1.016-0.489,2.372-1.06,3.894-1.5c0.145-0.047,0.284-0.106,0.429-0.152c16.864-5.304,33.165-0.582,43.605,12.633c11.33,14.342,12.334,34.501,2.71,51.775c21.713,10.59,32.09,38.033,34.424,58.926c11.258-0.104,18.137-0.281,20.6-0.531c0.701-20.574,0.396-41.886,0.094-62.512c-0.307-20.989-0.623-42.692,0.125-63.634V3.084C114.279,3.101,84.463,2.878,55.668,2.669C38.965,2.549,23.172,2.435,2.386,2.375C2.223,30.833,2.321,61.241,2.417,90.662c0.056,17.23,0.114,35.035,0.117,52.044C4.676,142.717,9.113,142.748,15.144,142.792zM115.65,84.064c9.812-16.49,9.146-36.006-1.707-49.74c-7.814-9.892-19.213-14.619-31.576-13.433c0.198,0.139,0.392,0.287,0.583,0.447c0.635,0.532,0.719,1.478,0.187,2.113c-0.53,0.635-1.477,0.72-2.113,0.187c-2.168-1.815-5.257-1.546-7.879-0.786c-1.103,0.362-2.18,0.779-3.239,1.232c-1.103,0.517-1.824,0.95-1.849,0.965c-0.201,0.124-0.421,0.182-0.641,0.203c-11.12,5.695-19.311,16.412-21.538,28.782C43.363,68.007,48.834,81.4,60.887,90.788C76.747,101.42,103.1,101.271,115.65,84.064zM150.415,143.189c-2.327-20.196-12.317-46.547-33.073-56.422c-7.344,9.32-19.505,14.273-32.101,14.273c-8.734,0-17.666-2.382-25.334-7.318c-16.991,8.471-34.493,26.752-36.119,49.129C83.751,143.266,126.066,143.377,150.415,143.189zM93.395,90.688c7.889-1.694,15.104-9.008,17.551-17.785c0.223-0.798-0.243-1.625-1.043-1.848c-0.797-0.223-1.623,0.244-1.848,1.042c-2.125,7.625-8.556,14.209-15.291,15.657c-8.962,1.93-19.705-3.24-24.978-12.025c-0.426-0.709-1.347-0.941-2.058-0.515c-0.71,0.427-0.94,1.349-0.514,2.058c5.062,8.436,14.693,13.877,23.744,13.877C90.46,91.148,91.947,90.998,93.395,90.688z"]
  },
  "webcam_color": 0x000000,
  "searchbox_padding_left": 22,
  "searchbox_padding_right": 4,
  "searchbox_padding_top": 6,
  "searchbox_padding_bottom": 6,
  "searchbox_label_color": 0x80808B,
  "searchbox_disabled_color": 0x80808b,
  "searchbox": {
    type: "SVG_SCALE9",
    scale9: [13, 26, 19, 11]
  },
  "searchbox_nw": {
    type: "SVG_PATHS",
    width: 26,
    height: 13,
    paths: ["M4.788,7.085c4.037-3.216,11.155-5.06,19.53-5.06c0.562,0,1.123,0.013,1.682,0.03V0.054  c-0.559-0.017-1.119-0.028-1.682-0.028c-5.183,0-14.773,0.713-20.776,5.495C1.113,7.456,0.087,10.249-0.017,13h1.988  C2.07,10.801,2.888,8.598,4.788,7.085z"]
  },
  "searchbox_n": {
    type: "SVG_PATHS",
    width: 6,
    height: 13,
    paths: ["M6,2.555V0.529C4.06,0.273,2.045,0.112,0,0.054v2.002C2.042,2.118,4.061,2.289,6,2.555z"]
  },
  "searchbox_ne": {
    type: "SVG_PATHS",
    width: 19,
    height: 13,
    paths: ["M16.127,6.537C13.312,3.686,7.171,1.476,0,0.529v2.025c6.586,0.903,12.242,2.895,14.703,5.389  c1.485,1.505,2.124,3.285,2.183,5.057H18.9C18.847,10.749,18.035,8.472,16.127,6.537z"]
  },
  "searchbox_e": {
    type: "SVG_PATHS",
    width: 19,
    height: 1,
    paths: ["M16.886,0c0.011,0.334,0.009,0.668-0.02,1h2.01c0.024-0.332,0.032-0.666,0.024-1H16.886z"]
  },
  "searchbox_se": {
    type: "SVG_PATHS",
    width: 19,
    height: 11,
    paths: ["M16.866,0c-0.144,1.644-0.758,3.232-1.642,4.536C13.408,7.096,8.288,8.571,0,8.942v2.008  c9.202-0.391,14.607-2.088,16.868-5.273C17.971,4.049,18.725,2.06,18.876,0H16.866z"]
  },
  "searchbox_s": {
    type: "SVG_PATHS",
    width: 6,
    height: 11,
    paths: ["M1.595,11.039c1.558,0,3.022-0.03,4.405-0.089V8.942C4.617,9.004,3.155,9.039,1.595,9.039  C1.049,9.039,0.52,9.034,0,9.028v2C0.522,11.034,1.051,11.039,1.595,11.039z"]
  },
  "searchbox_sw": {
    type: "SVG_PATHS",
    width: 26,
    height: 11,
    paths: ["M17.437,8.777C9.532,8.628,4.76,8.537,2.701,3.378C2.295,2.361,2.035,1.196,1.973,0h-1.994  C0.041,1.464,0.35,2.884,0.843,4.12c2.551,6.392,8.425,6.503,16.556,6.657c2.067,0.039,4.205,0.079,6.577,0.22  c0.646,0.014,1.327,0.024,2.024,0.031v-2c-0.677-0.007-1.333-0.018-1.944-0.029C21.682,8.858,19.427,8.815,17.437,8.777z"]
  },
  "searchbox_w": {
    type: "SVG_PATHS",
    width: 26,
    height: 1,
    paths: ["M1.972,0h-1.988c-0.013,0.334-0.018,0.669-0.004,1h1.994C1.956,0.668,1.957,0.334,1.972,0z"]
  },
  "coverflow": {
    type: "SVG_PATHS",
    width: 493,
    height: 198,
    paths: ["M9,0c0.667,0,1.333,0,2,0c9.138,6.862,18.295,13.705,29,19c3.37-3.473-2.4-9.87,2-12c14.539,5.461,25.502,14.498,41,19c0.751-7.154,0.317-11.981-1-19c24.94,7.099,53.366,19.272,80,28c3.442-5.327,1.951-17.121,1-24c56.834,5.365,112.391,6.169,169,6c0,7.333,0,14.667,0,22c23.641-9.393,51.253-20.035,76-25c2.124,4.209,0.485,12.182,1,18c14.916-4.449,26.698-17.259,42-18c2.356,3.596-1.233,8.353,0,14c10.632-6.149,24.343-14.168,36-16c4.241,32.239-2.168,69.137-1,104c0.71,21.188-0.097,45.115,3,66c0.681,4.589-0.654,6.156,4,9c0,0.333,0,0.667,0,1c-13.526-0.291-26.499-9.391-39-14c-1.149,9.363,1.269,15.15-1,20c-1,0-2,0-3,0c-10.271-8.396-23.147-14.186-37-19c-1.028,3.898-0.494,8.579,0,14c-28.449-7.884-54.6-18.067-81-28c-2.851,5.875,2.049,12.599,2,19c-56.418,1.166-111.169,0.287-169,2c-3.758-7.081-1.148-15.346-1-22c-22.422,4.264-47.688,14.298-74,18c-4.5-1.167-1.08,5.587-6,4c-4.693-1.307-2.551-9.449-3-15c-11.972,1.466-28.755,11.772-42,12c-2.68-4.193,0.237-10.058,0-15c-15.461,2.205-23.191,12.143-37,16c-0.86-0.807-0.594-2.739-2-3c0-0.333,0-0.667,0-1c7.916-22.422,4.344-56.793,5-90c0.258-13.031,2.333-28.325,3-45C8.54,31.5,5.594,4.103,9,0z M7,175c9.825-5.842,21.409-9.925,32-15c-0.577-40.949,1.855-90.352,2-134c-9.29-5.376-17.325-12.008-27-17C11.46,64.126,13.234,123.567,7,175z M86,176c28.257-4.076,53.291-11.375,78-19c-4.041-36.708-3.192-81.019-1-116c-24.685-8.315-49.824-16.176-73-26C89.091,67.137,91.331,120.79,86,176z M44,176c13-4,26.464-7.536,38-13c1.288-35.132,3.505-89.513,1-132c-14.801-1.533-21.69-10.977-35-14C46.543,70.07,44.556,128.049,44,176z M169,180c54.012-0.595,105.038-0.621,158-2c-1.466-53.525-2.416-102.781-1-156c-54.284,3.716-103.75-2.624-156-4C163.711,70.765,167.8,124.794,169,180z M451,36c-1.899,38.019-3.038,91.805,2,136c8.696,5.03,20.604,11.981,31,12c-4.917-44.717-2.265-98.797-1-143c0.205-7.163,2.804-15.995-2-21C469.756,24.089,458.868,28.535,451,36z M331,46c-1.579,43.245,2.43,78.435,0,112c23.997,10.336,49.149,19.517,76,27c-3.906-54.721-4.351-110.793-5-164C377.521,28.521,352.622,35.622,331,46z M445,22c-0.057-0.61-0.149-1.184-1-1c-12.125,6.208-24.533,12.134-36,19c-0.405,40.785,1.792,90.683,4,133c12.575,4.425,23.574,10.426,35,16C443.348,133.183,443.605,76.471,445,22z"]
  },
  "coverflow_color": 0x000000,
  "scratchout": {
    type: "SVG_PATHS",
    width: 205,
    height: 107,
    paths: ["M183.938,103.357c-0.301-1.293-1.361-1.826-0.956-3.824c-3.021-2.322-3.755,4.031-8.605,2.868c-3.061-0.127-4.206-2.168-3.824-5.737c-3.772,1.729-6.709,6.935-11.474,4.781c-2.123-11.135,8.937-17.133,13.387-24.859c-12.04,6.128-14.488,21.845-31.553,22.947c0.637-14.024,11.619-17.702,18.166-25.815c-11.682,4.293-16.145,20.771-27.728,21.99c-2.158-7.69,5.622-14.984,10.517-21.034c5.463-6.753,11.587-13.253,16.255-19.122  c-15.683,8.222-40.248,40.252-45.895,39.201c-6.314-10.054,4.456-20.203,10.518-26.771c7.21-7.812,16.645-14.776,21.991-21.991C120.975,56.969,87.271,99.278,79.72,96.664c-0.215-21.002,19.665-29.999,29.64-43.981c-9.693,4.716-19.203,12.874-28.684,21.035  c-9.371,8.066-21.857,25.229-27.729,22.946c-6.371-8.138,16.167-26.514,23.904-39.201c-17.33,10.717-37.883,40.988-42.35,38.342  c-8.602-2.978-2.875-14.846,1.236-20.176c4.974-6.447,11.133-13.549,15.297-18.166c-12.631,8.088-31.09,42.974-35.377,38.245C3.402,82.239,29.556,62.055,36.693,47.902c-5.64,1.699-13.936,18.787-16.254,17.21c-6.822-4.848-0.16-13.57,2.869-19.122c3.852-7.063,8.428-13.826,12.43-18.167c-8.547,3.251-22.905,24.31-25.816,21.034c-7.842-5.822,9.826-19.441,14.342-28.684  c-4.297,0.803-5.265,4.934-9.561,5.736c-5.009-0.408-2.414-8.422-8.605-7.648C0.261,15.574,0.324,7.191,3.404,5.814  c2.789-1.035,4.38-4.032,8.43-3.807c1.779,1.424,5.25,4.031,2.869,6.693c3.428,0.242,3.502-2.871,6.692-2.869  c2.689,0.18,4.073,1.664,3.825,4.781c5.727-1.142,9.983-8.263,16.254-5.736c3.433,6.877-4.328,10.143-6.693,14.341c6.717-2.844,10.459-8.663,19.123-9.562c8.389,2.279-9.452,14.809-13.387,21.992c9.719-2.752,24.092-26.33,30.598-20.079  c2.923,4.337-12.131,18.266-19.123,26.771C44.593,47.34,38.44,57.074,30.957,65.112C44.203,54.984,60.55,36.35,75.895,24.955c4.805-3.568,12.36-10.76,18.167-5.737c3.796,4.628-31.744,31.36-44.938,49.72c18.27-11.348,38.271-30.186,57.368-43.027c6.133-4.123,19.158-13.312,21.991-8.604c3.453,4.224-7.134,9.329-11.474,12.43c-14.82,10.591-27.06,22.726-38.245,36.333c23.612-13.678,63.758-51.32,65.973-45.895c2.134,8.822-7.204,16.146-13.386,21.991c-7.073,6.687-15.973,12.132-21.035,19.123c20.496-12.651,57.352-44.06,60.236-39.201c2.407,5.067-7.816,15.064-13.386,21.034c-5.897,6.323-13.34,11.612-18.166,18.167c16.445-10.646,43.885-40.751,46.035-34.961c5.129,7.443-13.82,20.325-20.22,31.136c12.378-5.789,26.011-27.828,32.509-21.99c1.417,5.412-14.039,18.47-21.035,27.728c4.49-0.304,8.569-9.266,17.21-8.605c5.242,6.307-11.161,17.467-16.254,25.816c5.629-0.844,12.811-17.216,21.991-11.474c0,0.955,0,1.911,0,2.868c-2.645,5.005-7.286,8.012-10.519,12.43  c3.869-0.352,6.095-6.107,10.519-2.868c1.551,1.705,0,3.187,0,4.78c-3.196,6.365-9.475,9.647-11.475,17.21C186.605,104.908,184.87,105.729,183.938,103.357z"]
  },
  "scratchout_color": 0x000000,
  "volumeslider": {
    type: "SVG_SCALE9",
    scale9: [1, 52, 2, 14]
  },
  "volumeslider_nw": "",
  "volumeslider_n": "",
  "volumeslider_ne": "",
  "volumeslider_e": "",
  "volumeslider_se": {
    type: "SVG_PATHS",
    width: 2,
    height: 14,
    paths: ["M0.247,3.516C0.168,3.53,0.08,3.54,0,3.554v2.027C0.199,5.548,0.407,5.52,0.603,5.484c0.543-0.098,0.904-0.619,0.807-1.162  C1.311,3.779,0.799,3.414,0.247,3.516z"]
  },
  "volumeslider_s": {
    type: "SVG_PATHS",
    width: 18,
    height: 14,
    paths: ["M0,4.527v1.995c0.841,0.012,1.683,0.022,2.524,0.022C8.025,6.544,13.438,6.34,18,5.58V3.554C12.824,4.455,6.191,4.584,0,4.527z"]
  },
  "volumeslider_sw": {
    type: "SVG_PATHS",
    width: 52,
    height: 14,
    paths: ["M14,2c2.254,2.595-0.803,6.697,1,8C16.344,8.538,17.044,2.08,14,2z M11.083,3.764c2.254,1.498-0.803,3.867,1,4.619  C13.427,7.539,14.127,3.81,11.083,3.764z M43,5c-1.001,1-0.668,2.175,0,3c2.182,0.182,3.182-0.818,3-3C44.999,4.445,44.193,4.5,43,5  z M43.564,0.611c-2.504,0.3-4.52,2.311-5.42,5.295c-1.672,0.018-3.502,0.133-5.447,0.251c-4.241,0.26-9.047,0.553-12.591-0.139  c-0.544-0.107-1.067,0.248-1.173,0.79s0.248,1.067,0.79,1.172c1.779,0.349,3.81,0.467,5.915,0.467c2.385,0,4.865-0.152,7.18-0.295  c1.908-0.115,3.711-0.221,5.32-0.243c0.951,2.84,3.752,4.356,6.397,4.356c0.274,0,0.547-0.017,0.815-0.05  c2.314-0.28,4.907-1.958,5.137-5.708C50.992,6.516,51.497,6.517,52,6.522V4.527c-0.643-0.005-1.286-0.012-1.917-0.021  C48.538,1.799,46.088,0.306,43.564,0.611z M45.109,10.23c-2.129,0.252-4.6-0.924-5.172-3.337c0.614-2.474,2.043-4.078,3.864-4.297  c1.773-0.217,3.562,0.997,4.695,3.159C48.564,8.292,47.314,9.963,45.109,10.23z M10.038,0H7.207C6.025,0.822,4.736,1.824,3.58,3.014  L3.33,3.021C2.76,3.064,2.25,3.11,1.785,3.24L1.48,3.324L1.28,3.57C0.072,5.047,0.417,7.024,0.695,8.614  c0.044,0.251,0.086,0.495,0.121,0.727l0.091,0.618l0.596,0.189c0.775,0.246,1.585,0.367,2.461,0.367c0.08,0,0.163-0.006,0.244-0.008  c1.135,1.062,2.334,1.956,3.425,2.689c0.071,0.232,0.22,0.441,0.446,0.572c0.159,0.092,0.332,0.137,0.503,0.137  c0.074,0,0.147-0.012,0.219-0.028C8.867,13.893,8.934,13.906,9,13.906c0.329,0,0.65-0.161,0.841-0.458  c0.219-0.341,0.202-0.761-0.003-1.081C10.682,9.219,10.551,2.106,10.038,0z M3.505,7.492C3.494,7.686,3.489,7.882,3.48,8.076  C3.428,8.109,3.369,8.129,3.322,8.173C3.234,8.254,3.173,8.35,3.121,8.45C2.976,8.43,2.826,8.422,2.686,8.393  c-0.007-0.041-0.014-0.08-0.021-0.121C2.469,7.151,2.25,5.898,2.658,5.098c0.233-0.035,0.484-0.057,0.697-0.073  C3.591,5.746,3.557,6.596,3.505,7.492z M8.097,11.09c-0.883-0.619-1.776-1.329-2.62-2.133c-0.015-0.43-0.002-0.875,0.025-1.352  c0.062-1.069,0.121-2.255-0.3-3.385c0.98-0.979,2.048-1.824,3.087-2.544C8.499,3.992,8.521,8.648,8.097,11.09z"]
  },
  "volumeslider_w": {
    type: "SVG_PATHS",
    width: 52,
    height: 1,
    paths: ["M9.862,0.515C9.853,0.496,9.852,0.476,9.84,0.458C9.815,0.42,9.777,0.397,9.749,0.363C9.707,0.316,9.668,0.269,9.619,0.23  C9.551,0.176,9.478,0.138,9.4,0.104c-0.023-0.01-0.043-0.021-0.067-0.029C9.05-0.027,8.731-0.016,8.458,0.16  C8.065,0.413,7.642,0.696,7.207,1h2.832C9.995,0.821,9.949,0.665,9.9,0.564C9.891,0.545,9.873,0.533,9.862,0.515z"]
  },
  "videoplayer_color": 0x000000,
  "videoplayer": {
    type: "SVG_SCALE9",
    scale9: [7, 146, 74, 43]
  },
  "videoplayer_nw": {
    type: "SVG_PATHS",
    width: 146,
    height: 7,
    paths: ["M2.996,7h-3.252C-0.155,4.053,0.202,1.798,1,1c3.19-0.47,71.288-0.61,145-0.346v2.548C98.793,3.095,51.623,3.074,4,3C2.835,3.168,2.914,4.58,3,6C3,6.334,2.996,6.666,2.996,7z"]
  },
  "videoplayer_n": {
    type: "SVG_PATHS",
    width: 80,
    height: 7,
    paths: ["M0,3.202V0.654C26.878,0.75,54.503,0.9,80,1.108v2.384C53.262,3.35,26.625,3.263,0,3.202z"]
  },
  "videoplayer_ne": {
    type: "SVG_PATHS",
    width: 74,
    height: 7,
    paths: ["M72.843,7h-2.77C70.048,6.002,70.029,4.994,70,4C46.55,3.777,23.239,3.616,0,3.492V1.108c28.248,0.229,53.886,0.53,73,0.906C72.942,3.663,72.895,5.337,72.843,7z"]
  },
  "videoplayer_e": {
    type: "SVG_PATHS",
    width: 74,
    height: 150,
    paths: ["M72.147,150h-2.219c-0.71-50.354,1.388-102.257,0.145-150h2.77C71.346,47.974,72.046,101.611,72.147,150z"]
  },
  "videoplayer_se": {
    type: "SVG_PATHS",
    width: 74,
    height: 43,
    paths: ["M57,27c4.201,0.548-18.556-1.115-1,0C56.328,27.021,56.674,26.957,57,27z", "M45.866,16.565C44.245,15.429,43.899,13.246,42,15c0.547-2.521,0.686-2.647,1-4c0.804,0.529,1.646,0.839,3,1c-1.158,1.479-0.437,2.502,0.722,3.981C46.941,16.473,46.722,16.874,45.866,16.565z", "M60.622,12.149c0.879,0.392,2.646,0.266,3.997,0c0.433,0.924,0,3.75,0,3.75c-0.796-0.491-1.085-0.817-1.372-0.649c-1.025,0.6-2.584,1.165-3.348,1.851C58.031,16.361,62.774,14.978,60.622,12.149z", "M41.487,31.003c0.867,0.867,1.252,0.792,1.77,0.486c1.064-0.631,1.865-1.673,2.65-3.136c1.549,1.157,0.718,1.47-1.054,4.142c-0.033,0.507,0.774,0.973,1.054,1.505c-1.13,0-2.259,0-3.388,0C42.282,33.27,41.439,31.449,41.487,31.003z", "M59.067,28.331c1.547-0.502,2.514,4.261,4.841,2.229c0,1.486,0.537,2.683,0.239,4.258c-1.339,0-2.203-0.96-3.87-0.912C62.362,32.42,57.286,29.885,59.067,28.331z", "M72.147,0C72.179,14.577,72.155,28.68,72,42c-23.774,0.024-47.848,0.044-72,0.03v-2.004C23.327,40.023,46.674,40.001,70,40c-0.314-10.745,0.448-25.421,0-33C46.655,6.778,23.322,6.618,0,6.494V4.689C23.213,4.812,46.521,4.923,70,5c-0.03-1.663-0.047-3.333-0.071-5H72.147z", "M33.274,15.756c5.08,0.134,3.91,10.906,1.669,13.345C31.936,26.928,37.035,20.084,33.274,15.756z", "M28.408,18.697c5.078,0.077,3.911,6.298,1.669,7.706C27.069,25.148,32.168,21.196,28.408,18.697z", "M16.535,29.294c-1.394,0-2.679-0.192-3.907-0.583l-0.596-0.189l-0.091-0.618c-0.058-0.392-0.13-0.805-0.204-1.229c-0.44-2.517-0.987-5.649,0.836-7.877l0.199-0.244l0.304-0.085c0.717-0.201,1.534-0.274,2.451-0.343l1.009-0.032c0.553,0.009,1.026,0.418,1.043,0.97s-0.404,1.013-0.957,1.029l-0.987,0.029c-0.624,0.048-1.213,0.099-1.701,0.197c-1.009,1.535-0.595,3.906-0.227,6.011c0.038,0.218,0.075,0.432,0.111,0.642c1.096,0.276,2.263,0.373,3.55,0.298c0.575-0.027,1.025,0.388,1.058,0.938c0.033,0.552-0.388,1.025-0.938,1.058C17.164,29.284,16.847,29.294,16.535,29.294z", "M16.593,20.092c-0.243,0-0.486-0.088-0.678-0.265c-0.406-0.375-0.432-1.008-0.058-1.413c2.695-2.922,5.977-5.188,8.534-6.836c0.465-0.299,1.082-0.166,1.383,0.299c0.299,0.465,0.165,1.083-0.299,1.383c-2.611,1.682-5.62,3.771-8.147,6.51C17.131,19.983,16.862,20.092,16.593,20.092z", "M24.933,34.95c-0.186,0-0.373-0.052-0.541-0.159c-2.558-1.649-5.84-3.917-8.533-6.835c-0.375-0.406-0.35-1.039,0.056-1.413c0.406-0.375,1.04-0.35,1.413,0.056c2.526,2.737,5.537,4.827,8.148,6.511c0.464,0.3,0.598,0.919,0.299,1.383C25.583,34.789,25.261,34.95,24.933,34.95z", "M24.236,34.95c-0.171,0-0.344-0.044-0.503-0.136c-0.477-0.278-0.638-0.891-0.36-1.368c1.729-2.967,1.451-18.924,0.66-20.592c-0.236-0.498-0.026-1.097,0.47-1.334c0.497-0.239,1.091-0.033,1.331,0.463c1.021,2.113,1.364,18.872-0.732,22.471C24.915,34.772,24.58,34.95,24.236,34.95z", "M17.427,29.267c-0.502,0-0.935-0.376-0.992-0.887c-0.131-1.139-0.064-2.302,0-3.427c0.112-1.965,0.218-3.82-0.699-5.347c-0.284-0.473-0.131-1.087,0.343-1.372c0.475-0.285,1.088-0.131,1.372,0.343c1.234,2.054,1.105,4.309,0.981,6.489c-0.062,1.089-0.121,2.118-0.011,3.087c0.063,0.549-0.331,1.044-0.88,1.106C17.503,29.265,17.465,29.267,17.427,29.267z", "M48.45,28.277c-0.259,0-0.509-0.101-0.695-0.281c-0.195-0.188-0.305-0.448-0.305-0.719v-8.58c0-0.553,0.447-1,1-1h9.078c0.276,0,0.54,0.114,0.729,0.315c0.188,0.202,0.286,0.473,0.269,0.748l-0.528,8.303c-0.033,0.514-0.451,0.919-0.966,0.936l-8.55,0.277C48.472,28.277,48.461,28.277,48.45,28.277zM49.45,19.697v6.547l6.609-0.214l0.403-6.333H49.45z", "M4.602,18.801l-0.879,8.417c-0.053,0.51-0.482,0.896-0.994,0.896c-0.001,0-0.003,0-0.004,0L0,28.104v-2l1.828,0.007l0.671-6.423L0,19.665v-2.001l3.616,0.033c0.282,0.003,0.55,0.124,0.737,0.334S4.631,18.521,4.602,18.801z"]
  },
  "videoplayer_s": {
    type: "SVG_PATHS",
    width: 80,
    height: 43,
    paths: ["M0,41.825v-1.939c26.599,0.115,53.284,0.144,80,0.141v2.004C53.331,42.016,26.569,41.958,0,41.825z", "M0,6.213V4.267C26.607,4.397,53.238,4.55,80,4.689v1.805C53.316,6.353,26.653,6.273,0,6.213z", "M0,18.938v-2l80,0.727v2.001L0,18.938z", "M0,27.799v-2l80,0.305v2L0,27.799z"]
  },
  "videoplayer_sw": {
    type: "SVG_PATHS",
    width: 146,
    height: 43,
    colors: [0xcdcccc, 0x000000, 0x000000, 0x000000, 0x000000],
    paths: ["M144.1,27.114H42.766V17H144.1V27.114z", "M18,9c18.511-3.5,22.394,23.341,6,26C5.177,38.054-0.555,12.508,18,9zM9.542,26.13C12.554,35.705,30.62,35.478,32,25c1.317-10-5.794-15.269-13.599-13.696C9.629,13.07,7.257,18.863,9.542,26.13z", "M18.408,17.163c0.259,4.589,0.259,4.589,0.064,9.265C21.297,25.2,22.383,24.254,25,23C22.855,19.599,21.51,18.395,18.408,17.163zM27,22c-1.666,4.001-6.267,5.066-10,7c0-4.667,0-9.333,0-14C21.507,16.159,24.398,18.936,27,22z", "M146,4.267v1.946C97.893,6.104,49.833,6.075,1.836,6c0,7.884-0.167,23.316,1.164,32c47.118,1.178,94.913,1.678,143,1.886v1.939C95.984,41.575,46.661,41.052,0,40C0,26.667,0,13.333,0,0h1.969C1.983,1.337,1.983,2.66,2,4C50.33,3.872,98.125,4.032,146,4.267z", "M42.984,25.407L146,25.799v2L41.883,27.403c-0.281-0.001-0.548-0.12-0.737-0.328c-0.188-0.208-0.281-0.485-0.255-0.765l0.879-9.403C41.818,16.393,42.25,16,42.766,16c0.003,0,0.006,0,0.009,0L146,16.938v2l-102.324-0.93L42.984,25.407z"]
  },
  "videoplayer_w": {
    type: "SVG_PATHS",
    width: 146,
    height: 150,
    paths: ["M1.969,150H0c0-32.334,0-64.667,0-97c1.843-8.287,1-15.516,1-29C1,20.396-0.524,7.775-0.257,0h3.252C2.922,50.269,1.427,99.115,1.969,150z"]
  },
  "streetmap": {
    type: "SVG_PATHS",
    width: 252,
    height: 222,
    colors: [0xffffff, 0xebe9e2, 0xebe9e2, 0xebe9e2, 0xebe9e2, 0xebe9e2, 0xebe9e2, 0xebe9e2, 0xebe9e2, 0xebe9e2, 0xebe9e2, 0xebe9e2, 0xebe9e2, 0xebe9e2, 0xebe9e2, 0xbcd5ac, 0xfaf078],
    paths: ["M252.001,222.172H0V0h252.001V222.172z", "M0,0h56.571l11.356,60.686L0,72", "M134.743,0l8.071,48.082L73.029,59.657L62.743,0H134.743z", "M140.915,0h111.086v31.886L149.12,47.314L140.915,0z", "M252.001,38.057v21.601l-97.715,17.485l-4.114-22.629L252.001,38.057z", "M252.001,65.828v73.029l-87.12,8.406L159.43,108l75.379-11.994l-2.061-6.52L158.4,102.857l-4.114-20.572L252.001,65.828z", "M144.001,54.514l9.44,59.778l-30.667,3.994l-10.661-58.629L144.001,54.514z", "M74.058,65.828l31.886-5.143l9.257,59.657l-31.886,3.363L74.058,65.828z", "M36.106,73.293l32.925-5.262l8.113,55.675l-31.376,4.71L36.106,73.293z", "M0,78.172l29.829-4.115L38.058,126l-30.41,11.734L0,94.897V78.172z", "M0,132.686l6.171,30.604l60.371-6.342L72,222.172H0V132.686z", "M78.172,222.172L72,156.947l48.343-5.747l4.114,70.972H78.172z", "M129.471,221.891l-5.013-68.728l45.257-2.743l5.01,71.471H129.471z", "M180.001,222.172l-5.144-72.704l46.285-3.409L226.286,222L180.001,222.172z", "M226.286,146.059L252.001,144v78.172H231.43L226.286,146.059z", "M10.286,159.574l-3.904-28.934l36.819-10.298l73.472-4.118l35.478-9.252l7.279,40.293l-149.144,13.56", "M218.059,0l24.686,145.312L0,163.543l0.337-4.113l236.843-20.572L211.695,0H218.059z"]
  },
  "playback": {
    type: "SVG_PATHS",
    width: 110,
    height: 36,
    colors: [0xffffff, 0xffffff, 0xffffff, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000],
    paths: ["M72.127,17.732c0,8.964-7.048,16.231-15.742,16.231c-8.695,0-15.743-7.268-15.743-16.231C40.642,8.768,47.69,1.5,56.385,1.5C65.079,1.5,72.127,8.768,72.127,17.732z", "M108.5,20.5C94.668,51.052,62.216,6.068,92.697,3.08c7.433-0.729,14.239,5,15.803,13.42", "M2.833,14.834c5.887-20.715,34.132-6.582,26.89,9.014C23.721,36.772-2.188,32.259,2.833,14.834", "M55.245,34.964c-10.249,0-15.584-8.04-16.205-15.511c-0.653-7.861,3.542-17.141,14.295-18.939L53.5,0.5l6.192,0.019c10.215,2.003,14.752,11.365,13.955,19.579c-0.712,7.335-5.949,14.735-16.044,14.735h-0.049c-0.78,0.087-1.557,0.131-2.308,0.131C55.247,34.964,55.246,34.964,55.245,34.964zM53.584,2.5c-9.445,1.623-13.129,9.83-12.551,16.787c0.566,6.808,5.181,13.676,14.212,13.677c0.001,0,0,0,0.001,0c0.695,0,1.416-0.042,2.141-0.124l0.117-0.007c9.545,0,13.572-6.965,14.151-12.929C72.364,12.613,68.381,4.312,59.401,2.5H53.584z", "M95.833,32.088c-8.812,0-16.376-9.549-16.763-17.455C78.906,11.275,79.896,3.33,92.6,2.084c0.454-0.044,0.915-0.066,1.37-0.066c7.45,0,13.975,6.014,15.514,14.3L109.5,16.5v4.216l-0.089,0.196C105.229,30.148,99.719,32.088,95.833,32.088zM93.97,4.018c-0.389,0-0.781,0.019-1.175,0.057C84.94,4.845,80.775,8.56,81.067,14.535c0.343,7.001,7.224,15.553,14.766,15.553c4.682,0,8.713-3.39,11.667-9.806v-3.688C106.108,9.301,100.431,4.018,93.97,4.018z", "M16.619,32.281L16.619,32.281c-5.413,0-10.286-2.184-13.038-5.841c-2.428-3.227-3.019-7.336-1.708-11.884C3.877,7.501,8.713,3.453,15.139,3.453c5.646,0,11.507,3.242,14.587,8.066c2.547,3.992,2.869,8.52,0.904,12.749C28.335,29.211,22.966,32.281,16.619,32.281zM15.139,5.453c-3.211,0-8.956,1.254-11.343,9.654c-1.131,3.924-0.652,7.426,1.383,10.131c2.341,3.11,6.724,5.043,11.439,5.043c4.358,0,9.851-1.801,12.198-6.854c1.667-3.589,1.391-7.435-0.776-10.831C25.313,8.323,20.128,5.453,15.139,5.453z", "M15.5,17.5L15.5,17.5", "M25.97,11.141l-1.284,0.092c-2.755,0.197-4.963,1.899-7.1,3.545c-0.534,0.412-1.05,0.799-1.563,1.164c-0.059-0.497-0.039-1.05-0.007-1.627c0.07-1.281,0.151-2.734-0.804-3.873l-0.439-0.524l-0.648,0.22c-2.297,0.777-3.952,2.59-5.552,4.344c-0.753,0.824-1.465,1.604-2.208,2.234l-0.559,0.475l0.284,0.676c1.157,2.754,4.071,4.127,6.641,5.338c0.431,0.203,0.854,0.402,1.257,0.604l0.617,0.307l0.507-0.468c1.074-0.991,1.003-2.393,0.946-3.518c-0.018-0.352-0.03-0.688-0.016-1.002c0.161,0.117,0.322,0.23,0.483,0.349c2.268,1.664,4.612,3.385,7.36,4.071l0.936,0.234l0.268-0.927c0.535-1.85,0.499-3.938,0.463-5.957c-0.028-1.611-0.055-3.133,0.189-4.488L25.97,11.141zM14.039,21.608c-0.15-0.071-0.302-0.143-0.455-0.215c-2.108-0.993-4.281-2.017-5.33-3.688c0.621-0.59,1.216-1.241,1.796-1.877c1.216-1.332,2.467-2.703,3.928-3.459c0.115,0.503,0.08,1.137,0.042,1.835c-0.051,0.922-0.085,1.933,0.22,2.892l-0.566,0.345l0.576,0.403c-0.25,0.794-0.226,1.625-0.187,2.385C14.091,20.81,14.104,21.265,14.039,21.608zM23.552,16.932c0.026,1.49,0.053,3.02-0.165,4.369c-1.993-0.734-3.862-2.105-5.678-3.438c-0.151-0.11-0.302-0.222-0.453-0.332c0.526-0.378,1.044-0.776,1.552-1.168c1.602-1.234,3.132-2.413,4.792-2.906C23.511,14.596,23.532,15.777,23.552,16.932z", "M53.255,26.349l-0.438-0.829c-1.067-2.019-0.953-4.295-0.843-6.497c0.093-1.863,0.181-3.622-0.408-5.167l-0.062-0.271c-0.034-0.402-0.104-0.803-0.167-1.156c-0.059-0.34-0.11-0.645-0.131-0.89l-0.063-0.758l0.713-0.264c0.552-0.204,1.143-0.308,1.757-0.308c4.125,0,8.629,4.727,10.499,7.705l0.288,0.458l-0.227,0.492c-1.548,3.359-5.705,5.185-9.045,6.65L53.255,26.349zM53.483,13.276c0.684,1.897,0.584,3.903,0.487,5.846c-0.083,1.658-0.162,3.241,0.26,4.603l0.093-0.041c2.847-1.25,6.348-2.786,7.781-5.182c-1.909-2.787-5.814-6.531-8.774-6.281C53.388,12.552,53.447,12.911,53.483,13.276z", "M95.5,17.5L95.5,17.5", "M96.345,24.126l-0.501-0.505c-0.972-0.982-0.915-2.28-0.864-3.426c0.016-0.373,0.032-0.736,0.012-1.068c-2.381,1.766-4.864,3.486-7.682,4.413l-0.844,0.277l-0.375-0.805c-0.837-1.796-0.741-3.982-0.647-6.099c0.067-1.544,0.138-3.142-0.184-4.428l-0.326-1.305l1.343,0.062c2.737,0.129,4.919,1.824,7.029,3.463c0.567,0.441,1.116,0.867,1.662,1.252c0.051-0.482,0.019-1.025-0.015-1.584c-0.078-1.315-0.166-2.807,0.869-3.955l0.448-0.498l0.631,0.225c2.283,0.814,3.928,2.596,5.518,4.318c0.756,0.818,1.469,1.591,2.215,2.219l0.563,0.475l-0.285,0.678c-1.196,2.846-4.42,4.347-7.011,5.554c-0.312,0.146-0.619,0.288-0.915,0.43L96.345,24.126zM96.769,17.8c0.281,0.822,0.243,1.691,0.209,2.482c-0.024,0.55-0.035,0.986,0.019,1.322c0.021-0.01,0.04-0.019,0.061-0.028c2.157-1.004,4.573-2.13,5.689-3.898c-0.622-0.584-1.217-1.229-1.796-1.856c-1.22-1.321-2.475-2.681-3.951-3.452c-0.131,0.511-0.092,1.166-0.05,1.887c0.054,0.902,0.112,1.895-0.174,2.836l0.525,0.311L96.769,17.8zM87.439,13.458c0.106,1.156,0.054,2.364,0.002,3.544c-0.069,1.569-0.136,3.069,0.164,4.298c2.162-0.883,4.188-2.294,6.151-3.747c-0.571-0.406-1.13-0.841-1.678-1.267C90.526,15.08,89.042,13.928,87.439,13.458z"]
  },
  "rule_default_color": 0x000000,
  "rule_thickness": 3,
  "rule_breakline_thickness": 3,
  "rule_dash_size": 4,
  "rule_dot_size": 1,
  "rule_horizontal": {
    type: "SVG_SCALE9",
    scale9: [2, 8, 7, 2],
    width: 67,
    height: 10
  },
  "rule_horizontal_nw": {
    type: "SVG_PATHS",
    width: 8,
    height: 2,
    paths: []
  },
  "rule_horizontal_n": {
    type: "SVG_PATHS",
    width: 52,
    height: 2,
    paths: []
  },
  "rule_horizontal_ne": {
    type: "SVG_PATHS",
    width: 7,
    height: 2,
    paths: []
  },
  "rule_horizontal_e": {
    type: "SVG_PATHS",
    width: 7,
    height: 6,
    paths: ["M7,4C4.87,3.882,2.513,3.789,0,3.715V2.299C2.356,2.498,4.546,2.729,6.5,3C7,3.345,7,3.667,7,4z"]
  },
  "rule_horizontal_se": {
    type: "SVG_PATHS",
    width: 7,
    height: 2,
    paths: []
  },
  "rule_horizontal_s": {
    type: "SVG_PATHS",
    width: 52,
    height: 2,
    paths: []
  },
  "rule_horizontal_sw": {
    type: "SVG_PATHS",
    width: 8,
    height: 2,
    paths: []
  },
  "rule_horizontal_w": {
    type: "SVG_PATHS",
    width: 8,
    height: 6,
    paths: ["M8,1.119v1.592C3.079,2.373,0,1.847,0,1C0.811,1.104,3.742,1.12,8,1.119z"]
  },
  "rule_horizontal_c": {
    type: "SVG_PATHS",
    width: 52,
    height: 6,
    paths: ["M52,3.715C35.498,3.228,12.125,3.543,0,2.711V1.119c12.389-0.004,36.079-0.166,52,1.18V3.715z"]
  },
  "rule_vertical": {
    type: "SVG_SCALE9",
    scale9: [8, 2, 2, 9],
    width: 10,
    height: 67
  },
  "rule_vertical_nw": {
    type: "SVG_PATHS",
    width: 2,
    height: 8,
    paths: []
  },
  "rule_vertical_n": {
    type: "SVG_PATHS",
    width: 6,
    height: 8,
    paths: ["M3.751,8H1.485c0.152-2.535,0.371-4.985,0.691-7.3C2.655-0.088,3.683-0.053,4,0.7C3.87,3.036,3.794,5.483,3.751,8z"]
  },
  "rule_vertical_ne": {
    type: "SVG_PATHS",
    width: 2,
    height: 8,
    paths: []
  },
  "rule_vertical_e": {
    type: "SVG_PATHS",
    width: 2,
    height: 50,
    paths: []
  },
  "rule_vertical_se": {
    type: "SVG_PATHS",
    width: 2,
    height: 9,
    paths: []
  },
  "rule_vertical_s": {
    type: "SVG_PATHS",
    width: 6,
    height: 9,
    paths: ["M3.41,0C3.188,2.56,2.893,4.984,2.5,7.227C2.107,9.144,1.183,8.965,1,7.965C1.381,5.55,1.615,2.869,1.75,0H3.41z"]
  },
  "rule_vertical_sw": {
    type: "SVG_PATHS",
    width: 2,
    height: 9,
    paths: []
  },
  "rule_vertical_w": {
    type: "SVG_PATHS",
    width: 2,
    height: 50,
    paths: []
  },
  "rule_vertical_c": {
    type: "SVG_PATHS",
    width: 6,
    height: 50,
    paths: ["M3.41,50H1.75c0.69-14.56-1.22-34.102-0.266-50h2.267C3.478,16.119,4.678,35.333,3.41,50z"]
  },
  "splitter_horizontal_track": {
    type: "SVG_SCALE9",
    scale9: [4, 4, 5, 5],
    width: 50,
    height: 12
  },
  "splitter_horizontal_track_nw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M4,1.104V3C2.667,3,1.333,3,0,3c0-0.666,0-1.334,0-2C1.316,1.052,2.654,1.082,4,1.104z"]
  },
  "splitter_horizontal_track_n": {
    type: "SVG_PATHS",
    width: 41,
    height: 4,
    paths: ["M41,3C27.333,3,13.667,3,0,3V1.104c13.542,0.22,28.412-0.886,41,0.289V3z"]
  },
  "splitter_horizontal_track_ne": {
    type: "SVG_PATHS",
    width: 5,
    height: 4,
    paths: ["M5,3C3.333,3,1.667,3,0,3V1.394C1.709,1.553,3.384,1.747,5,2C5,2.334,5,2.666,5,3z"]
  },
  "splitter_horizontal_track_e": {
    type: "SVG_PATHS",
    width: 5,
    height: 3,
    paths: []
  },
  "splitter_horizontal_track_se": {
    type: "SVG_PATHS",
    width: 5,
    height: 5,
    paths: ["M5,4C3.333,4,1.667,4,0,4V2.084C1.643,2.067,3.302,2.044,5,2C5,2.667,5,3.333,5,4z"]
  },
  "splitter_horizontal_track_s": {
    type: "SVG_PATHS",
    width: 41,
    height: 5,
    paths: ["M41,4C27.333,4,13.667,4,0,4V1.79c12.492-0.502,25.587,0.442,41,0.294V4z"]
  },
  "splitter_horizontal_track_sw": {
    type: "SVG_PATHS",
    width: 4,
    height: 5,
    paths: ["M4,1.79V4C2.667,4,1.333,4,0,4c0-0.667,0-1.333,0-2C1.327,1.909,2.661,1.845,4,1.79z"]
  },
  "splitter_horizontal_track_w": {
    type: "SVG_PATHS",
    width: 4,
    height: 3,
    paths: []
  },
  "splitter_horizontal_track_c": {
    type: "SVG_PATHS",
    width: 41,
    height: 3,
    paths: []
  },
  "splitter_horizontal_thumb": {
    type: "SVG_PATHS",
    width: 17,
    height: 4,
    paths: ["M7,1c2.007-0.34,3.129,0.205,3,2C9,3,8,3,7,3C7,2.333,7,1.666,7,1z", "M13,1c2.007-0.34,3.129,0.205,3,2c-1,0-2,0-3,0C13,2.333,13,1.666,13,1z", "M0,2c-0.368-1.701,0.78-1.887,2-2c0.027,0.973,1.224,0.775,2,1C3.721,3.318,1.329,3.299,0,2z"]
  },
  "splitter_vertical_track": {
    type: "SVG_SCALE9",
    scale9: [4, 4, 5, 5],
    width: 12,
    height: 50
  },
  "splitter_vertical_track_nw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M3,4H1.493C1.635,2.64,1.796,1.299,2,0c0.333,0,0.667,0,1,0C3,1.333,3,2.667,3,4z"]
  },
  "splitter_vertical_track_n": {
    type: "SVG_PATHS",
    width: 3,
    height: 4,
    paths: []
  },
  "splitter_vertical_track_ne": {
    type: "SVG_PATHS",
    width: 5,
    height: 4,
    paths: ["M4,4H2.074C2.057,2.682,2.035,1.354,2,0c0.667,0,1.333,0,2,0C4,1.333,4,2.667,4,4z"]
  },
  "splitter_vertical_track_e": {
    type: "SVG_PATHS",
    width: 5,
    height: 41,
    paths: ["M4,0c0,13.666,0,27.333,0,41H1.752C1.329,28.539,2.271,15.436,2.074,0H4z"]
  },
  "splitter_vertical_track_se": {
    type: "SVG_PATHS",
    width: 5,
    height: 5,
    paths: ["M4,0c0,1.667,0,3.333,0,5C3.333,5,2.667,5,2,5C1.887,3.343,1.81,1.676,1.752,0H4z"]
  },
  "splitter_vertical_track_s": {
    type: "SVG_PATHS",
    width: 3,
    height: 5,
    paths: []
  },
  "splitter_vertical_track_sw": {
    type: "SVG_PATHS",
    width: 4,
    height: 5,
    paths: ["M3,0c0,1.667,0,3.333,0,5C2.333,5,1.667,5,1,5c0.064-1.641,0.098-3.313,0.117-5H3z"]
  },
  "splitter_vertical_track_w": {
    type: "SVG_PATHS",
    width: 4,
    height: 41,
    paths: ["M3,0c0,13.666,0,27.333,0,41H1.117C1.273,27.364,0.19,12.489,1.493,0H3z"]
  },
  "splitter_vertical_track_c": {
    type: "SVG_PATHS",
    width: 3,
    height: 41,
    paths: []
  },
  "splitter_vertical_thumb": {
    type: "SVG_PATHS",
    width: 4,
    height: 16,
    paths: ["M1,1c2.917-1.471,2.917,4.471,0,3C1,3,1,2,1,1z", "M1,7c2.917-1.471,2.917,4.47,0,3C1,9,1,8,1,7z", "M3,16c-1.419,0.086-2.832,0.165-3-1C1.167,13.106,2.888,12.307,3,16z"]
  },
  "datepicker_padding_left": 6,
  "datepicker_padding_right": 6,
  "datepicker_padding_top": 7,
  "datepicker_padding_bottom": 5,
  "datepicker_icon_size": 24,
  "datepicker_disabled_color": 0x80808b,
  "datepicker_border_padding_right": 35,
  "fieldset_border_padding_top": 6,
  "fieldset_label_padding_top": 0,
  "fieldset_label_padding_left": 15,
  "fieldset_label_padding_right": 4,
  "list_padding_top": 3,
  "list_padding_bottom": 3,
  "list_padding_left": 8,
  "list_padding_right": 8,
  "list_line_padding": 8,
  "list_header_color": 0xCCCCCC,
  "list_scrollbar_padding_top": 2,
  "list_scrollbar_padding_bottom": 2,
  "list_scrollbar_padding_right": 2,
  "list_scrollbar_padding_no_border": 0,
  "list_scrollbar_header_padding_top": -1,
  "list_icon_padding": 16,
  "progressbar_color": 0xCCCCCC,
  "progressbar_indeterminate": {
    type: "SVG_PATHS",
    width: 25,
    height: 11,
    paths: ["M12.549,0H25v0.035L14.036,11H1.551L12.549,0z"]
  },
  "progressbar_padding": 3,
  "menubar_entry_padding": 9,
  "menubar_padding_left": 9,
  "menubar_padding_right": 9,
  "menubar_padding_top": 2,
  "menubar_padding_bottom": 1,
  "slider_disabled_color": 0x898989,
  "slider_default_color": 0,
  "slider_vertical_thumb_circle_padding": 0,
  "slider_vertical_thumb_pointy_padding": 0,
  "slider_vertical_thumb_rectangle_padding": 0,
  "slider_horizontal_thumb_circle_padding": 0.5,
  "slider_horizontal_thumb_pointy_padding": 0,
  "slider_horizontal_thumb_rectangle_padding": 0,
  "slider_vertical_track": {
    type: "SVG_SCALE9",
    scale9: [5, 1, 4, 4],
    width: 11,
    height: 33
  },
  "slider_vertical_track_nw": {
    type: "SVG_PATHS",
    width: 1,
    height: 5,
    paths: []
  },
  "slider_vertical_track_n": {
    type: "SVG_PATHS",
    width: 6,
    height: 5,
    paths: ["M5.574,5H3.36c0.026-1.102,0.046-2.171,0.051-3.176c0.176-1.175,0.888-1.503,1.316-1.503c0.571,0,0.777,0.585,0.883,1.729C5.602,3.042,5.588,4.022,5.574,5z"]
  },
  "slider_vertical_track_ne": {
    type: "SVG_PATHS",
    width: 4,
    height: 5,
    paths: []
  },
  "slider_vertical_track_e": {
    type: "SVG_PATHS",
    width: 4,
    height: 24,
    paths: []
  },
  "slider_vertical_track_se": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: []
  },
  "slider_vertical_track_s": {
    type: "SVG_PATHS",
    width: 6,
    height: 4,
    paths: ["M4.988,0c-0.021,1.173-0.039,2.337-0.05,3.489C4.899,4,3.587,4.209,3.411,3.211C3.336,2.124,3.277,1.058,3.227,0H4.988z"]
  },
  "slider_vertical_track_sw": {
    type: "SVG_PATHS",
    width: 1,
    height: 4,
    paths: []
  },
  "slider_vertical_track_w": {
    type: "SVG_PATHS",
    width: 1,
    height: 24,
    paths: []
  },
  "slider_vertical_track_c": {
    type: "SVG_PATHS",
    width: 6,
    height: 24,
    paths: ["M5.574,0C5.449,8.494,5.126,16.466,4.988,24H3.227C2.776,14.663,3.206,6.512,3.36,0H5.574z"]
  },
  "slider_vertical_thumb_pointy": {
    type: "SVG_PATHS",
    width: 14,
    height: 14,
    paths: ["M9.363,1.904c0.502,0.098,3.479,3.416,3.426,5.314c-0.029,1.031-3.291,4.763-3.697,4.872c-1.998,0.167-6.617,0.569-7.456-0.5c-0.62-0.79-0.361-7.122-0.427-8.751C1.209,1.318,7.43,1.666,9.363,1.904z"]
  },
  "slider_vertical_thumb_circle": {
    type: "SVG_PATHS",
    width: 14,
    height: 14,
    paths: ["M10.354,2.4c2.703,2.181,2.805,8.146-0.408,9.688C7.345,13.1,1.868,12.162,1.631,6.232C1.631,2.133,8.925,0.376,10.354,2.4z"]
  },
  "slider_vertical_thumb_rectangle": {
    type: "SVG_PATHS",
    width: 14,
    height: 14,
    paths: ["M11.575,4.239c1.008,0.198,0.954,4.926,0.142,5.145c-1.887,0.732-8.781,1.339-9.62,0.27c-0.62-0.789-0.36-3.261-0.426-4.89C1.67,3.243,9.673,3.806,11.575,4.239z"]
  },
  "slider_horizontal_track": {
    type: "SVG_SCALE9",
    scale9: [2, 5, 5, 7],
    width: 33,
    height: 11
  },
  "slider_horizontal_track_nw": {
    type: "SVG_PATHS",
    width: 5,
    height: 2,
    paths: []
  },
  "slider_horizontal_track_n": {
    type: "SVG_PATHS",
    width: 23,
    height: 2,
    paths: []
  },
  "slider_horizontal_track_ne": {
    type: "SVG_PATHS",
    width: 5,
    height: 2,
    paths: []
  },
  "slider_horizontal_track_e": {
    type: "SVG_PATHS",
    width: 5,
    height: 2,
    paths: []
  },
  "slider_horizontal_track_se": {
    type: "SVG_PATHS",
    width: 5,
    height: 7,
    paths: ["M4.615,2.029c0,0.571-0.585,0.777-1.729,0.883C1.915,2.904,0.958,2.889,0,2.875V0.664C1.083,0.69,2.123,0.708,3.112,0.713C4.287,0.889,4.615,1.601,4.615,2.029z"]
  },
  "slider_horizontal_track_s": {
    type: "SVG_PATHS",
    width: 23,
    height: 7,
    paths: ["M23,0.664v2.211C14.879,2.757,7.241,2.459,0,2.312V0.479C8.917,0.109,16.729,0.517,23,0.664z"]
  },
  "slider_horizontal_track_sw": {
    type: "SVG_PATHS",
    width: 5,
    height: 7,
    paths: ["M0.726,0.713C2.179,0.612,3.6,0.538,5,0.479v1.833C3.467,2.281,1.945,2.254,0.447,2.24C-0.063,2.201-0.272,0.889,0.726,0.713z"]
  },
  "slider_horizontal_track_w": {
    type: "SVG_PATHS",
    width: 5,
    height: 2,
    paths: []
  },
  "slider_horizontal_track_c": {
    type: "SVG_PATHS",
    width: 23,
    height: 2,
    paths: []
  },
  "slider_horizontal_thumb_pointy": {
    type: "SVG_PATHS",
    width: 14,
    height: 14,
    paths: ["M12.019,9.324c-0.099,0.503-3.416,3.479-5.315,3.426c-1.031-0.029-4.762-3.29-4.872-3.697c-0.167-1.998-0.569-6.617,0.5-7.456c0.79-0.62,7.122-0.361,8.75-0.427C12.604,1.17,12.256,7.391,12.019,9.324z"]
  },
  "slider_horizontal_thumb_circle": {
    type: "SVG_PATHS",
    width: 14,
    height: 14,
    paths: ["M11.599,9.771c-2.181,2.703-8.146,2.805-9.688-0.408C0.899,6.763,1.837,1.286,7.767,1.049C11.866,1.049,13.623,8.343,11.599,9.771z"]
  },
  "slider_horizontal_thumb_rectangle": {
    type: "SVG_PATHS",
    width: 14,
    height: 14,
    paths: ["M9.67,11.576c-0.198,1.008-4.925,0.953-5.144,0.141C3.793,9.83,3.187,2.936,4.255,2.097c0.79-0.62,3.261-0.36,4.89-0.426C10.667,1.671,10.104,9.674,9.67,11.576z"]
  },
  "combobox": {
    type: "SVG_SCALE9",
    scale9: [3, 3, 21, 3],
    width: 51,
    height: 24
  },
  "combobox_nw": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    paths: ["M2.27,3H0.268c0.02-0.356,0.027-0.703,0.049-1.062c0.034-0.551,0.505-0.971,1.059-0.937c0.551,0.033,0.971,0.508,0.937,1.06C2.294,2.379,2.288,2.685,2.27,3z", "M3,0.495v2C2.377,2.498,1.761,2.5,1.175,2.5c-0.552,0-1-0.447-1-1s0.448-1,1-1C1.761,0.5,2.377,0.498,3,0.495z"]
  },
  "combobox_n": {
    type: "SVG_PATHS",
    width: 27,
    height: 3,
    paths: ["M27,0.438v2C21.238,2.285,14.017,2.359,7.556,2.43C4.885,2.458,2.327,2.484,0,2.495v-2c2.32-0.011,4.871-0.037,7.534-0.065C13.994,0.361,21.208,0.286,27,0.438z"]
  },
  "combobox_ne": {
    type: "SVG_PATHS",
    width: 21,
    height: 3,
    paths: ["M20.73,3h-2.011c-0.087-0.629-0.189-0.77-0.328-0.86c-1.105,0.008-2.277,0-3.485-0.005C10.693,2.116,5.92,2.097,2.104,2.494C1.548,2.553,1.062,2.153,1.005,1.604C0.948,1.055,1.347,0.562,1.896,0.506c3.924-0.409,8.755-0.39,13.018-0.371c1.302,0.006,2.563,0.01,3.745,0.004l0.292-0.002l0.374,0.231C20.103,0.817,20.536,1.359,20.73,3z", "M3.001,3H1.002C1.003,2.74,0.993,2.481,1,2.221C1.017,1.67,1.469,1.268,2.029,1.25C2.581,1.267,3.016,1.728,3,2.279C2.993,2.52,3.002,2.76,3.001,3z", "M2.999,1.539C2.979,2.077,2.535,2.5,2,2.5c-0.013,0-0.026,0-0.04-0.001C1.334,2.475,0.671,2.456,0,2.438v-2c0.698,0.019,1.386,0.038,2.039,0.063C2.591,0.522,3.021,0.987,2.999,1.539z"]
  },
  "combobox_e": {
    type: "SVG_PATHS",
    width: 21,
    height: 18,
    paths: ["M20.742,0.088c0.19,5.059,0.346,11.524-0.004,17.912h-2.003c0.354-6.342,0.198-12.791,0.013-17.766c-0.01-0.087-0.02-0.162-0.03-0.234h2.011C20.733,0.031,20.739,0.056,20.742,0.088z", "M3.466,6.837C3.854,10.453,4.241,14.186,3.597,18H1.57c0.682-3.647,0.31-7.198-0.093-10.95C1.231,4.76,0.992,2.396,1.002,0h1.999C2.991,2.291,3.226,4.598,3.466,6.837z"]
  },
  "combobox_se": {
    type: "SVG_PATHS",
    width: 21,
    height: 3,
    paths: ["M20.738,0c-0.02,0.354-0.034,0.71-0.057,1.063C20.647,1.594,20.208,2,19.685,2c-0.021,0-0.043-0.001-0.065-0.002c-0.551-0.035-0.969-0.511-0.934-1.062c0.02-0.312,0.032-0.624,0.05-0.937H20.738z", "M20.948,1.443c0.047,0.55-0.361,1.034-0.912,1.081C16.277,2.846,8.413,2.971,0,2.999V1c8.355-0.027,16.151-0.15,19.866-0.468C20.414,0.473,20.9,0.893,20.948,1.443z", "M3.597,0C3.557,0.233,3.527,0.466,3.48,0.699C3.384,1.173,2.967,1.5,2.501,1.5c-0.066,0-0.133-0.007-0.2-0.02C1.76,1.37,1.41,0.842,1.52,0.301C1.541,0.2,1.551,0.101,1.57,0H3.597z", "M20.95,1.474c0.03,0.552-0.392,1.023-0.944,1.053C16.284,2.73,8.451,2.842,0,2.891v-2c8.41-0.048,16.196-0.158,19.896-0.36C20.438,0.474,20.919,0.921,20.95,1.474z"]
  },
  "combobox_s": {
    type: "SVG_PATHS",
    width: 27,
    height: 3,
    paths: ["M27,1v1.999c-1.316,0.005-2.645,0.007-3.975,0.007C13.856,3.006,4.699,2.91,0,2.843v-2C5.358,0.92,16.604,1.035,27,1z", "M27,0.891v2C23.453,2.91,19.798,2.92,16.289,2.92C9.81,2.92,3.844,2.889,0,2.841v-2C5.909,0.914,16.896,0.947,27,0.891z"]
  },
  "combobox_sw": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    paths: ["M2.302,0C2.32,0.263,2.327,0.539,2.346,0.8c0.041,0.551-0.373,1.03-0.923,1.071c-0.025,0.002-0.05,0.003-0.075,0.003c-0.519,0-0.958-0.4-0.996-0.926C0.329,0.639,0.321,0.312,0.3,0H2.302z", "M1.366,0.817C1.818,0.825,2.379,0.834,3,0.843v2C2.359,2.834,1.796,2.825,1.332,2.817C0.78,2.808,0.34,2.353,0.349,1.801c0.009-0.547,0.455-0.983,1-0.983C1.354,0.817,1.36,0.817,1.366,0.817z", "M1.366,0.817C1.85,0.825,2.409,0.833,3,0.841v2C2.39,2.834,1.829,2.826,1.332,2.817C0.78,2.808,0.34,2.353,0.349,1.801c0.009-0.547,0.455-0.983,1-0.983C1.354,0.817,1.36,0.817,1.366,0.817z"]
  },
  "combobox_w": {
    type: "SVG_PATHS",
    width: 3,
    height: 18,
    paths: ["M2.302,18H0.3C-0.071,12.395-0.081,6.352,0.268,0H2.27C1.917,6.36,1.926,12.406,2.302,18z"]
  },
  "combobox_c": {
    type: "SVG_PATHS",
    width: 27,
    height: 18,
    paths: []
  },
  "combobox_padding_left": 6,
  "combobox_padding_right": 2,
  "combobox_padding_top": 4,
  "combobox_padding_bottom": 10,
  "combobox_padding_middle": 3,
  "combobox_icon_padding": 3,
  "combobox_arrow_width": 22,
  "combobox_pulldown_overlap": 4,
  "combobox_pulldown_padding_bottom": 6,
  "combobox_border_thickness": 3,
  "combobox_background_disabled_color": 0xffffff,
  "combobox_arrow_background_color": 0xffffff,
  "combobox_border_disabled_color": 0x898989,
  "barchart_axes_thickness": 4,
  "barchart_axes_color": 0x000000,
  "barchart_series1_color": 0xDFDFDF,
  "barchart_series2_color": 0x959595,
  "formattingtoolbar_border": 2,
  "formattingtoolbar_icon_size": 15,
  "formattingtoolbar_icon_color": 0x000000,
  "accordion_padding_middle": 17,
  "accordion_padding_left": 6,
  "accordion_padding_right": 3,
  "accordion_overlap": 3,
  "accordion_padding_top": 4,
  "accordion_padding_bottom": 4,
  "accordion_scrollbar_padding_right": 2,
  "accordion_selected_subrow_padding": 6,
  "menu_padding_middle": 8,
  "menu_padding_top": 5,
  "menu_padding_bottom": 5,
  "menu_padding_right": 5,
  "menu_padding_left": 20,
  "menu_asset_padding_left": 5,
  "menu_asset_padding_right": 0,
  "menu_separator_skin": "rule_horizontal",
  "menu_rule_padding": 5,
  "line_vertical": {
    type: "SVG_PATHS",
    width: 2,
    height: 48,
    paths: ["M0,43.548V2.484C0,2.484,0.415,0,2,0c0,15.963,0,47.947-0.668,47.889C0,48.534,0,43.548,0,43.548z"]
  },
  "sitemap_edge_width": 2,
  "tabbar_vertical_separator_skin": "rule_vertical",
  "tabbar_h_default_tab_color": 0xCCCCCC,
  "tabbar_h_default_color": 0xFFFFFF,
  "tabbar_h_tab_padding_left": 6,
  "tabbar_h_tab_padding_right": 14,
  "tabbar_h_tab_padding_top": 7,
  "tabbar_h_tab_padding_bottom": 7,
  "tabbar_h_tab_overlap": 8,
  "tabbar_h_padding_left": 20,
  "tabbar_h_padding_right": 21,
  "tabbar_h_padding_top": 0,
  "tabbar_h_padding_bottom": 0,
  "tabbar_h_background_overlap": 2,
  "tabbar_h_background_padding": 2,
  "tabbar_h_top_selected_label_offset": 2,
  "tabbar_h_bottom_selected_label_offset": -2,
  "tabbar_h_scrollbar_padding_top": 4,
  "tabbar_h_scrollbar_padding_bottom": 1,
  "tabbar_h_scrollbar_padding_right": 2,
  "tabbar_h_top_selected_tab_mask_padding_top": 0,
  "tabbar_h_top_selected_tab_mask_padding_bottom": -3,
  "tabbar_h_top_selected_tab_mask_padding_left": 2,
  "tabbar_h_top_selected_tab_mask_padding_right": 2,
  "tabbar_h_bottom_selected_tab_mask_padding_top": -3,
  "tabbar_h_bottom_selected_tab_mask_padding_bottom": 0,
  "tabbar_h_bottom_selected_tab_mask_padding_left": 3,
  "tabbar_h_bottom_selected_tab_mask_padding_right": 6,
  "tabbar_h_top_tab": {
    type: "SVG_SCALE9",
    scale9: [1, 6, 17, 2],
    width: 45,
    height: 36
  },
  "tabbar_h_top_tab_nw": {
    type: "SVG_PATHS",
    width: 6,
    height: 1,
    paths: []
  },
  "tabbar_h_top_tab_n": {
    type: "SVG_PATHS",
    width: 22,
    height: 1,
    paths: []
  },
  "tabbar_h_top_tab_ne": {
    type: "SVG_PATHS",
    width: 17,
    height: 1,
    paths: []
  },
  "tabbar_h_top_tab_e": {
    type: "SVG_PATHS",
    width: 17,
    height: 33,
    paths: ["M11.275,19.835c1.163,4.502,2.353,9.114,3.31,13.165h-2.058c-0.932-3.919-2.072-8.338-3.188-12.665C7.885,14.703,6.511,9.381,5.559,5.081l-0.02-0.045C5.022,3.828,4.438,2.463,2.736,2.107C1.86,2.088,0.945,2.074,0,2.062v-2C0.993,0.074,1.95,0.09,2.867,0.11c0.055,0.001,0.108,0.007,0.162,0.017c2.809,0.526,3.772,2.777,4.349,4.122l0.056,0.13c0.024,0.057,0.044,0.117,0.058,0.179C8.437,8.838,9.815,14.18,11.275,19.835z"]
  },
  "tabbar_h_top_tab_se": {
    type: "SVG_PATHS",
    width: 17,
    height: 2,
    paths: ["M14.585,0c0.161,0.683,0.317,1.354,0.464,2h-2.054c-0.148-0.647-0.306-1.318-0.468-2H14.585z"]
  },
  "tabbar_h_top_tab_s": {
    type: "SVG_PATHS",
    width: 22,
    height: 2,
    paths: []
  },
  "tabbar_h_top_tab_sw": {
    type: "SVG_PATHS",
    width: 6,
    height: 2,
    paths: ["M2.928,0c0.008,0.688,0.018,1.361,0.029,2H0.956C0.944,1.36,0.935,0.688,0.926,0H2.928z"]
  },
  "tabbar_h_top_tab_w": {
    type: "SVG_PATHS",
    width: 6,
    height: 33,
    paths: ["M2.106,0.109C3.399,0.128,4.701,0.124,6,0.127V2.13C5.016,2.127,4.028,2.133,3.048,2.122L2.841,6.838c0.053,3.357,0.049,7.883,0.045,12.675C2.882,24.195,2.879,28.978,2.928,33H0.926c-0.048-4.031-0.044-8.812-0.041-13.489C0.89,14.728,0.894,10.209,0.842,6.81l0.251-5.743C1.116,0.525,1.578,0.074,2.106,0.109z"]
  },
  "tabbar_h_top_tab_c": {
    type: "SVG_PATHS",
    width: 22,
    height: 33,
    paths: ["M17.821,0.039c1.487,0,2.866,0.009,4.179,0.023v2c-1.312-0.015-2.69-0.023-4.179-0.023c-2.791,0-5.588,0.025-8.384,0.053C6.282,2.12,3.135,2.138,0,2.13V0.127c3.129,0.009,6.269-0.006,9.418-0.035C12.221,0.064,15.025,0.039,17.821,0.039z"]
  },
  "tabbar_h_top_background_tab": {
    type: "SVG_SCALE9",
    scale9: [1, 6, 17, 2],
    width: 45,
    height: 36
  },
  "tabbar_h_top_background_tab_nw": {
    type: "SVG_PATHS",
    width: 6,
    height: 1,
    paths: []
  },
  "tabbar_h_top_background_tab_n": {
    type: "SVG_PATHS",
    width: 22,
    height: 1,
    paths: []
  },
  "tabbar_h_top_background_tab_ne": {
    type: "SVG_PATHS",
    width: 17,
    height: 1,
    paths: []
  },
  "tabbar_h_top_background_tab_e": {
    type: "SVG_PATHS",
    width: 17,
    height: 33,
    paths: ["M12.596,33H0V2.101c0.765,0.01,1.535,0.005,2.298,0.022c2.233,0.418,2.937,2.223,3.471,3.465C7.435,13.134,10.605,24.491,12.596,33z"]
  },
  "tabbar_h_top_background_tab_se": {
    type: "SVG_PATHS",
    width: 17,
    height: 2,
    paths: ["M12.596,0c0.16,0.685,0.314,1.356,0.458,2H0V0H12.596z"]
  },
  "tabbar_h_top_background_tab_s": {
    type: "SVG_PATHS",
    width: 22,
    height: 2,
    paths: ["M22,0v2H0V0H22z"]
  },
  "tabbar_h_top_background_tab_sw": {
    type: "SVG_PATHS",
    width: 6,
    height: 2,
    paths: ["M6,0v2H2.99C2.977,1.367,2.966,0.692,2.956,0H6z"]
  },
  "tabbar_h_top_background_tab_w": {
    type: "SVG_PATHS",
    width: 6,
    height: 33,
    paths: ["M6,33H2.956c-0.109-7.758,0.01-18.912-0.091-25.5l0.237-5.377C4.064,2.137,5.034,2.133,6,2.138V33z"]
  },
  "tabbar_h_top_background_tab_c": {
    type: "SVG_PATHS",
    width: 22,
    height: 33,
    paths: ["M22,33H0V2.138c7.3,0.038,14.69-0.134,22-0.037V33z"]
  },
  "tabbar_h_bottom_tab": {
    type: "SVG_SCALE9",
    scale9: [2, 6, 13, 6],
    width: 45,
    height: 36
  },
  "tabbar_h_bottom_tab_nw": {
    type: "SVG_PATHS",
    width: 6,
    height: 2,
    paths: []
  },
  "tabbar_h_bottom_tab_n": {
    type: "SVG_PATHS",
    width: 26,
    height: 2,
    paths: []
  },
  "tabbar_h_bottom_tab_ne": {
    type: "SVG_PATHS",
    width: 13,
    height: 2,
    paths: []
  },
  "tabbar_h_bottom_tab_e": {
    type: "SVG_PATHS",
    width: 13,
    height: 28,
    paths: ["M9.137,0c-0.925,3.797-2.028,8.072-3.11,12.264C4.499,18.183,3.058,23.771,2.108,28H0.05c0.004-0.007,0.007-0.013,0.011-0.021c0.954-4.306,2.448-10.092,4.029-16.216C5.126,7.751,6.18,3.665,7.078,0H9.137z"]
  },
  "tabbar_h_bottom_tab_se": {
    type: "SVG_PATHS",
    width: 13,
    height: 6,
    paths: ["M2.108,0c-0.038,0.167-0.08,0.346-0.115,0.508C1.979,0.569,1.96,0.629,1.936,0.688c-0.529,1.23-1.19,1.705-1.829,2.163C0.072,2.875,0.035,2.902,0,2.928V0.094C0.017,0.061,0.034,0.033,0.05,0H2.108z"]
  },
  "tabbar_h_bottom_tab_s": {
    type: "SVG_PATHS",
    width: 26,
    height: 6,
    paths: ["M26,0.094v2.834c-0.161,0.117-0.33,0.24-0.513,0.393c-0.174,0.146-0.392,0.228-0.618,0.232c-4.419,0.099-8.958,0.31-13.347,0.513C7.735,4.241,3.852,4.421,0,4.532V2.53c3.816-0.11,7.67-0.288,11.43-0.463c4.29-0.198,8.721-0.404,13.057-0.507c0.158-0.122,0.31-0.23,0.454-0.334C25.445,0.863,25.725,0.648,26,0.094z"]
  },
  "tabbar_h_bottom_tab_sw": {
    type: "SVG_PATHS",
    width: 6,
    height: 6,
    paths: ["M6,2.53v2.002C4.697,4.569,3.398,4.6,2.108,4.618c-0.005,0-0.01,0-0.015,0c-0.545,0-0.991-0.438-1-0.985C1.086,3.08,1.527,2.626,2.079,2.618C3.378,2.6,4.687,2.568,6,2.53z", "M3.068,3.394c0.124,0.539-0.212,1.075-0.75,1.199C2.243,4.61,2.167,4.618,2.093,4.618c-0.456,0-0.867-0.313-0.974-0.775C0.821,2.548,0.825,1.256,0.845,0h2C2.826,1.188,2.821,2.316,3.068,3.394z"]
  },
  "tabbar_h_bottom_tab_w": {
    type: "SVG_PATHS",
    width: 6,
    height: 28,
    paths: ["M3.144,5.037c0.202,7.428,0.431,15.847-0.311,20.975c-0.079,0.547-0.591,0.935-1.133,0.846c-0.547-0.079-0.926-0.586-0.847-1.133c0.717-4.957,0.491-13.285,0.291-20.633C1.097,3.332,1.051,1.619,1.019,0h2C3.051,1.6,3.097,3.296,3.144,5.037z", "M2.845,28h-2c0.001-0.083,0.001-0.167,0.003-0.25c0.011-0.615,0.021-1.23-0.003-1.841c-0.023-0.552,0.406-1.018,0.958-1.04c0.566-0.052,1.018,0.406,1.041,0.958c0.027,0.648,0.016,1.303,0.005,1.957C2.847,27.857,2.846,27.928,2.845,28z"]
  },
  "tabbar_h_bottom_tab_c": {
    type: "SVG_PATHS",
    width: 26,
    height: 28,
    paths: []
  },
  "tabbar_h_bottom_background_tab": {
    type: "SVG_SCALE9",
    scale9: [1, 6, 17, 2],
    width: 45,
    height: 36
  },
  "tabbar_h_bottom_background_tab_nw": {
    type: "SVG_PATHS",
    width: 6,
    height: 1,
    paths: ["M6,1H2.003C1.998,0.665,1.992,0.324,1.988,0H6V1z"]
  },
  "tabbar_h_bottom_background_tab_n": {
    type: "SVG_PATHS",
    width: 22,
    height: 1,
    paths: ["M22,1H0V0h22V1z"]
  },
  "tabbar_h_bottom_background_tab_ne": {
    type: "SVG_PATHS",
    width: 17,
    height: 1,
    paths: ["M12.349,1H0V0h12.59C12.513,0.325,12.43,0.666,12.349,1z"]
  },
  "tabbar_h_bottom_background_tab_e": {
    type: "SVG_PATHS",
    width: 17,
    height: 33,
    paths: ["M12.349,0c-2.232,9.254-5.674,21.781-7.333,29.292c-0.565,1.314-1.227,1.474-2.17,2.261C1.901,31.574,0.948,31.617,0,31.646V0H12.349z"]
  },
  "tabbar_h_bottom_background_tab_se": {
    type: "SVG_PATHS",
    width: 17,
    height: 2,
    paths: []
  },
  "tabbar_h_bottom_background_tab_s": {
    type: "SVG_PATHS",
    width: 22,
    height: 2,
    paths: []
  },
  "tabbar_h_bottom_background_tab_sw": {
    type: "SVG_PATHS",
    width: 6,
    height: 2,
    paths: []
  },
  "tabbar_h_bottom_background_tab_w": {
    type: "SVG_PATHS",
    width: 6,
    height: 33,
    paths: ["M6,0v32.518c-1.303,0.037-2.609,0.082-3.906,0.101l-0.25-5.75C2.782,20.379,2.142,8.488,2.003,0H6z", "M2.094,32.618c-0.433-1.881-0.171-3.837-0.25-5.75"]
  },
  "tabbar_h_bottom_background_tab_c": {
    type: "SVG_PATHS",
    width: 22,
    height: 33,
    paths: ["M22,0v31.646c-7.311,0.231-14.695,0.664-22,0.871V0H22z"]
  },
  "tabbar_v_default_tab_color": 0xCCCCCC,
  "tabbar_v_default_color": 0xFFFFFF,
  "tabbar_v_tab_padding_left": 6,
  "tabbar_v_tab_padding_right": 4,
  "tabbar_v_tab_padding_top": 9,
  "tabbar_v_tab_padding_bottom": 8,
  "tabbar_v_tab_overlap": 2,
  "tabbar_v_padding_left": 6,
  "tabbar_v_padding_right": 4,
  "tabbar_v_padding_top": 20,
  "tabbar_v_padding_bottom": 42,
  "tabbar_v_background_overlap": 2,
  "tabbar_v_background_padding": 2,
  "tabbar_v_left_selected_label_offset": 0,
  "tabbar_v_right_selected_label_offset": 0,
  "tabbar_v_scrollbar_padding_top": 3,
  "tabbar_v_scrollbar_padding_bottom": 1,
  "tabbar_v_scrollbar_padding_right": 2,
  "tabbar_v_left_selected_tab_mask_padding_top": 2,
  "tabbar_v_left_selected_tab_mask_padding_bottom": 2,
  "tabbar_v_left_selected_tab_mask_padding_left": 0,
  "tabbar_v_left_selected_tab_mask_padding_right": -1,
  "tabbar_v_right_selected_tab_mask_padding_top": 2,
  "tabbar_v_right_selected_tab_mask_padding_bottom": 2,
  "tabbar_v_right_selected_tab_mask_padding_left": -1,
  "tabbar_v_right_selected_tab_mask_padding_right": 0,
  "tabbar_v_left_tab": {
    type: "SVG_SCALE9",
    scale9: [5, 4, 2, 6],
    width: 42,
    height: 29
  },
  "tabbar_v_left_tab_nw": {
    type: "SVG_PATHS",
    width: 4,
    height: 5,
    paths: ["M2.641,5H0.638l0.127-2.308L1.56,1.391l0.532-0.024c0.629-0.028,1.283-0.046,1.93-0.067V3.3c-0.429,0.015-0.869,0.023-1.29,0.041L2.641,5z"]
  },
  "tabbar_v_left_tab_n": {
    type: "SVG_PATHS",
    width: 36,
    height: 5,
    paths: ["M0.022,3.3V1.299c7.735-0.249,16.643,0.032,25.292,0.31c3.752,0.12,7.319,0.232,10.69,0.3v2.001c-3.394-0.068-6.98-0.181-10.754-0.302C16.606,3.328,7.706,3.047,0.022,3.3z"]
  },
  "tabbar_v_left_tab_ne": {
    type: "SVG_PATHS",
    width: 2,
    height: 5,
    paths: ["M0.003,3.909V1.908c0.674,0.014,1.353,0.028,2.011,0.038l-0.029,2C1.336,3.937,0.667,3.922,0.003,3.909z"]
  },
  "tabbar_v_left_tab_e": {
    type: "SVG_PATHS",
    width: 2,
    height: 18,
    paths: []
  },
  "tabbar_v_left_tab_se": {
    type: "SVG_PATHS",
    width: 2,
    height: 6,
    paths: ["M1.902,5.765C1.283,5.704,0.648,5.65,0.003,5.602V3.597c0.71,0.054,1.412,0.111,2.093,0.178L1.902,5.765z"]
  },
  "tabbar_v_left_tab_s": {
    type: "SVG_PATHS",
    width: 36,
    height: 6,
    paths: ["M0.022,3.691v-2.03c6.557,1.033,14.487,1.225,22.197,1.399c4.912,0.112,9.587,0.22,13.785,0.536v2.005c-4.183-0.321-8.887-0.43-13.83-0.541C14.521,4.887,6.646,4.701,0.022,3.691z"]
  },
  "tabbar_v_left_tab_sw": {
    type: "SVG_PATHS",
    width: 4,
    height: 6,
    paths: ["M2.585,1.432C3.05,1.517,3.542,1.586,4.022,1.661v2.03c-0.78-0.119-1.553-0.243-2.296-0.387L1.243,3.212L0.5,1.738V0.049h2v1.213L2.585,1.432z"]
  },
  "tabbar_v_left_tab_w": {
    type: "SVG_PATHS",
    width: 4,
    height: 18,
    paths: ["M2.499,2.556L2.5,18.049h-2V2.5L0.638,0h2.002L2.499,2.556z"]
  },
  "tabbar_v_left_tab_c": {
    type: "SVG_PATHS",
    width: 36,
    height: 18,
    paths: []
  },
  "tabbar_v_left_background_tab": {
    type: "SVG_SCALE9",
    scale9: [5, 4, 2, 6],
    width: 42,
    height: 29
  },
  "tabbar_v_left_background_tab_nw": {
    type: "SVG_PATHS",
    width: 4,
    height: 5,
    paths: ["M4.022,5H1.639L1.75,3c0.129-0.211,0.259-0.423,0.388-0.635c0.618-0.028,1.249-0.051,1.884-0.071V5z"]
  },
  "tabbar_v_left_background_tab_n": {
    type: "SVG_PATHS",
    width: 36,
    height: 5,
    paths: ["M36.003,5H0.022V2.294C11.096,1.938,24.789,2.685,36.003,2.91V5z"]
  },
  "tabbar_v_left_background_tab_ne": {
    type: "SVG_PATHS",
    width: 2,
    height: 5,
    paths: ["M2,5H0.003V2.91C0.677,2.924,1.345,2.937,2,2.946V5z"]
  },
  "tabbar_v_left_background_tab_e": {
    type: "SVG_PATHS",
    width: 2,
    height: 18,
    paths: ["M2,0v18.049H0.003V0H2z"]
  },
  "tabbar_v_left_background_tab_se": {
    type: "SVG_PATHS",
    width: 2,
    height: 6,
    paths: ["M2,0.049V4.77C1.351,4.706,0.682,4.651,0.003,4.6V0.049H2z"]
  },
  "tabbar_v_left_background_tab_s": {
    type: "SVG_PATHS",
    width: 36,
    height: 6,
    paths: ["M36.003,0.049V4.6C25.181,3.779,10.809,4.356,0.022,2.688V0.049H36.003z"]
  },
  "tabbar_v_left_background_tab_sw": {
    type: "SVG_PATHS",
    width: 4,
    height: 6,
    paths: ["M4.022,0.049v2.639C3.303,2.576,2.6,2.455,1.915,2.323L1.5,1.5c0-0.483,0-0.968,0-1.451H4.022z"]
  },
  "tabbar_v_left_background_tab_w": {
    type: "SVG_PATHS",
    width: 4,
    height: 18,
    paths: ["M4.022,0v18.049H1.5c0-5.184,0-10.366,0-15.549L1.639,0H4.022z"]
  },
  "tabbar_v_left_background_tab_c": {
    type: "SVG_PATHS",
    width: 36,
    height: 18,
    paths: ["M36.003,0v18.049H0.022V0H36.003z"]
  },
  "tabbar_v_right_tab": {
    type: "SVG_SCALE9",
    scale9: [5, 2, 5, 6],
    width: 42,
    height: 29
  },
  "tabbar_v_right_tab_nw": {
    type: "SVG_PATHS",
    width: 2,
    height: 5,
    paths: ["M2,1.962v2C1.333,3.976,0.661,3.99,0.015,4l-0.029-2C0.641,1.99,1.323,1.976,2,1.962z"]
  },
  "tabbar_v_right_tab_n": {
    type: "SVG_PATHS",
    width: 35,
    height: 5,
    paths: ["M35,1.252v1.999c-7.297-0.16-15.711,0.119-23.886,0.39C7.195,3.771,3.476,3.891,0,3.962v-2c3.453-0.071,7.151-0.191,11.047-0.32C19.236,1.369,27.656,1.093,35,1.252z"]
  },
  "tabbar_v_right_tab_ne": {
    type: "SVG_PATHS",
    width: 5,
    height: 5,
    paths: ["M5.019,5H3.017L2.924,3.341C1.973,3.302,0.995,3.272,0,3.251V1.252c1.217,0.026,2.413,0.062,3.564,0.114l0.532,0.024l0.794,1.301L5.019,5z"]
  },
  "tabbar_v_right_tab_e": {
    type: "SVG_PATHS",
    width: 5,
    height: 18,
    paths: ["M5.155,2.444L5.156,18h-2V2.5L3.017,0h2.002L5.155,2.444z"]
  },
  "tabbar_v_right_tab_se": {
    type: "SVG_PATHS",
    width: 5,
    height: 6,
    paths: ["M5.156,0v1.738L4.413,3.212L3.931,3.305C2.685,3.545,1.358,3.735,0,3.904v-2.01C1.056,1.76,2.089,1.61,3.071,1.432l0.085-0.17V0H5.156z"]
  },
  "tabbar_v_right_tab_s": {
    type: "SVG_PATHS",
    width: 35,
    height: 6,
    paths: ["M35,1.895v2.01C28.761,4.68,21.483,4.788,14.386,4.882C9.216,4.949,4.298,5.016,0,5.334V3.33c4.323-0.314,9.217-0.381,14.36-0.448C21.495,2.787,28.822,2.685,35,1.895z"]
  },
  "tabbar_v_right_tab_sw": {
    type: "SVG_PATHS",
    width: 2,
    height: 6,
    paths: ["M2,3.33v2.004C1.352,5.382,0.716,5.435,0.097,5.495l-0.194-1.99C0.583,3.438,1.286,3.382,2,3.33z"]
  },
  "tabbar_v_right_tab_w": {
    type: "SVG_PATHS",
    width: 2,
    height: 18,
    paths: []
  },
  "tabbar_v_right_tab_c": {
    type: "SVG_PATHS",
    width: 35,
    height: 18,
    paths: []
  },
  "tabbar_v_right_background_tab": {
    type: "SVG_SCALE9",
    scale9: [5, 2, 5, 6],
    width: 42,
    height: 29
  },
  "tabbar_v_right_background_tab_nw": {
    type: "SVG_PATHS",
    width: 2,
    height: 5,
    paths: ["M2,5H0V3c0.654-0.01,1.322-0.021,2-0.035V5z"]
  },
  "tabbar_v_right_background_tab_n": {
    type: "SVG_PATHS",
    width: 35,
    height: 5,
    paths: ["M35,5H0V2.965c10.732-0.22,24.133-0.945,35-0.711V5z"]
  },
  "tabbar_v_right_background_tab_ne": {
    type: "SVG_PATHS",
    width: 5,
    height: 5,
    paths: ["M4.018,5H0V2.254c1.203,0.025,2.385,0.06,3.519,0.111C3.647,2.577,3.776,2.789,3.906,3L4.018,5z"]
  },
  "tabbar_v_right_background_tab_e": {
    type: "SVG_PATHS",
    width: 5,
    height: 18,
    paths: ["M4.018,0l0.139,2.5c0,5.166,0,10.333,0,15.5H0V0H4.018z"]
  },
  "tabbar_v_right_background_tab_se": {
    type: "SVG_PATHS",
    width: 5,
    height: 6,
    paths: ["M4.156,0c0,0.5,0,1,0,1.5L3.741,2.323C2.555,2.552,1.295,2.737,0,2.9V0H4.156z"]
  },
  "tabbar_v_right_background_tab_s": {
    type: "SVG_PATHS",
    width: 35,
    height: 6,
    paths: ["M35,0v2.9C24.32,4.245,10.351,3.572,0,4.33V0H35z"]
  },
  "tabbar_v_right_background_tab_sw": {
    type: "SVG_PATHS",
    width: 2,
    height: 6,
    paths: ["M2,0v4.33C1.317,4.381,0.649,4.437,0,4.5V0H2z"]
  },
  "tabbar_v_right_background_tab_w": {
    type: "SVG_PATHS",
    width: 2,
    height: 18,
    paths: ["M2,0v18H0V0H2z"]
  },
  "tabbar_v_right_background_tab_c": {
    type: "SVG_PATHS",
    width: 35,
    height: 18,
    paths: ["M35,0v18H0V0H35z"]
  },
  "window_title_padding_left": 6,
  "window_title_padding_top": 7,
  "window_title_padding_right": 53,
  "window_scrollbar_padding_top": 0,
  "window_scrollbar_padding_right": 2,
  "window_scrollbar_padding_bottom": -2,
  "window_maximizeRestore_padding_right": 22,
  "window_maximizeRestore_padding_top": 6,
  "window_minimize_padding_right": 38,
  "window_minimize_padding_top": 6,
  "window_close_padding_right": 7,
  "window_close_padding_top": 6,
  "window_dragger": {
    type: "SVG_PATHS",
    width: 13,
    height: 11,
    paths: ["M9.938,2.65c0.475-0.73,0.45-1.779-0.775-2.103c-0.533-0.354-1.49-0.096-1.986,0.364C4.837,3.574,1.416,3.762,0.654,8c0.224,2.857,2.017,0.987,3.56-0.092L9.938,2.65z", "M9.219,9.164c-1.543,1.078-3.336,2.949-3.56,0.092c2.088-3.756,2.962-2.893,5.3-5.555c0.496-0.46,0.729-0.209,0.847,0.205c1.012,1.743-0.196,2.475-0.671,3.206L9.219,9.164z"]
  },
  "window_dragger_padding_right": 17,
  "window_dragger_padding_left": -1,
  "window_dragger_padding_bottom": 17,
  "switch_border": {
    type: "SVG_PATHS",
    width: 49,
    height: 28,
    colors: [0xffffff, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M36.025,3.508c5.734,0,10.385,4.648,10.385,10.385s-4.65,10.385-10.385,10.385c-5.736,0-10.385-4.648-10.385-10.385S30.289,3.508,36.025,3.508z", "M9.825,4.072c-0.606,0-1.138-0.441-1.234-1.059C8.485,2.331,8.953,1.692,9.635,1.587c11.01-1.706,20.622-1.771,28.569-0.194c0.677,0.135,1.117,0.793,0.983,1.47c-0.135,0.678-0.796,1.112-1.469,0.983c-7.66-1.521-16.979-1.448-27.7,0.212C9.953,4.067,9.888,4.072,9.825,4.072z", "M35.915,26.445c-1.012,0-2.297-0.165-4.192-0.57c-0.675-0.145-1.105-0.81-0.96-1.484c0.145-0.675,0.812-1.103,1.484-0.961c3.752,0.804,4.5,0.548,5.992,0.041c0.699-0.238,1.557-0.53,2.837-0.792c1.165-0.998,1.817-2.135,2.565-3.439c0.312-0.544,0.634-1.104,1.004-1.678c0.373-0.581,1.147-0.747,1.728-0.374s0.748,1.147,0.374,1.728c-0.345,0.536-0.645,1.06-0.937,1.569c-0.846,1.474-1.721,2.998-3.404,4.337l-0.242,0.192l-0.303,0.059c-1.286,0.244-2.099,0.521-2.816,0.765C37.995,26.194,37.162,26.445,35.915,26.445z", "M45.695,19.488c-0.69,0-1.25-0.56-1.25-1.25c0-0.524,0.159-1.03,0.312-1.521c0.079-0.251,0.198-0.631,0.198-0.772c0-0.69,0.56-1.25,1.25-1.25s1.25,0.56,1.25,1.25c0,0.524-0.159,1.03-0.312,1.521c-0.079,0.251-0.198,0.631-0.198,0.772C46.945,18.929,46.386,19.488,45.695,19.488z", "M22.709,27.267c-4.574,0-8.638-0.648-15.971-2.703l-0.194-0.055l-0.167-0.112c-2.186-1.473-4.065-5.056-4.373-8.335c-0.064-0.687,0.441-1.296,1.128-1.36c0.688-0.067,1.297,0.44,1.361,1.128c0.255,2.722,1.791,5.369,3.129,6.386c10.989,3.065,14.061,2.827,24.104,2.053c1.282-0.1,2.675-0.207,4.208-0.317c0.69-0.055,1.287,0.469,1.336,1.157s-0.468,1.287-1.157,1.337c-1.529,0.109-2.917,0.217-4.195,0.315C28.187,27.048,25.362,27.267,22.709,27.267z", "M3.353,17.971c-0.68,0-1.237-0.546-1.25-1.229C1.95,7.813,4.618,2.433,9.618,1.59c0.683-0.112,1.326,0.344,1.44,1.024c0.115,0.681-0.344,1.326-1.024,1.44C4.95,4.912,4.542,13.183,4.604,16.699c0.012,0.69-0.538,1.26-1.229,1.271C3.368,17.971,3.36,17.971,3.353,17.971z", "M32.653,26.041c-0.188,0-0.378-0.042-0.558-0.132c-5.302-2.651-6.223-4.587-7.394-7.779c-0.238-0.648,0.095-1.366,0.743-1.604c0.646-0.237,1.366,0.096,1.604,0.743c0.962,2.623,1.493,4.068,6.165,6.404c0.617,0.309,0.868,1.06,0.559,1.677C33.553,25.788,33.112,26.041,32.653,26.041z", "M25.876,19.195c-0.458,0-0.9-0.253-1.119-0.691c-1.372-2.743-1.185-6.808,0.476-10.354c1.45-3.097,3.807-5.232,6.465-5.858c0.673-0.156,1.346,0.259,1.503,0.931c0.158,0.672-0.258,1.345-0.93,1.503c-1.91,0.45-3.65,2.084-4.774,4.484c-1.329,2.837-1.531,6.123-0.504,8.178c0.309,0.617,0.058,1.368-0.559,1.677C26.254,19.153,26.064,19.195,25.876,19.195z", "M46.204,17.195c-0.591,0-1.117-0.421-1.228-1.023c-0.178-0.969-0.264-1.829-0.347-2.66c-0.169-1.688-0.314-3.145-1.211-5.021c-0.326-0.65-0.899-1.185-1.506-1.75c-0.397-0.37-0.808-0.753-1.154-1.186c-2.341-2.195-3.244-1.997-7.428-1.063l-1.078,0.238c-0.679,0.147-1.341-0.281-1.488-0.955c-0.147-0.675,0.28-1.341,0.955-1.488l1.067-0.235c4.394-0.979,6.4-1.425,9.756,1.75l0.131,0.146c0.234,0.305,0.579,0.625,0.944,0.966c0.707,0.658,1.508,1.405,2.046,2.479c1.096,2.293,1.278,4.113,1.454,5.872c0.078,0.781,0.159,1.589,0.319,2.456c0.125,0.679-0.324,1.331-1.003,1.456C46.355,17.188,46.279,17.195,46.204,17.195z"]
  },
  "switch_background": {
    type: "SVG_PATHS",
    width: 49,
    height: 28,
    paths: ["M45.695,14.018c0,6.627-5.373,12-11.998,12h-18.35c-6.627,0-12-5.373-12-12V13.5c0-6.627,5.373-12,12-12h18.35c6.625,0,11.998,5.373,11.998,12V14.018z"]
  },
  "switch_default_color": 0x6aa84f,
  "switch_border_disabled_color": 0xE2E2E2,
  "switch_background_disabled_color": 0xE2E2E2,
  "tooltip": {
    type: "SVG_SCALE9",
    scale9: [12, 16, 16, 12],
    width: 77,
    height: 25
  },
  "tooltip_nw": {
    type: "SVG_PATHS",
    width: 16,
    height: 12,
    colors: [0xffffff, 0x000000],
    paths: ["M16,12H1.052c0.022-0.482,0.069-0.978,0.168-1.5H1.5c0,0,1.01-4.454,14.5-6.837V12z", "M2.5,12h-2v-1.5c0-0.242,0.087-0.475,0.247-0.657C4.782,5.216,10.176,3.184,16,2.321v2.011C10.514,5.167,5.992,7.027,2.5,10.881V12z"]
  },
  "tooltip_n": {
    type: "SVG_PATHS",
    width: 45,
    height: 12,
    colors: [0xffffff, 0x000000],
    paths: ["M45,12H0V3.663C6.934,2.438,17.162,1.76,32.25,2.516c0,0,6.279-0.875,12.75-0.177V12z", "M0,4.332V2.321c3.155-0.467,6.434-0.596,9.695-0.596c1.978,0,3.975,0.044,5.968,0.089c1.967,0.044,3.929,0.088,5.862,0.088c4.676,0,8.075-0.271,11.022-0.882C32.614,1.007,32.682,1,32.75,1h4.5c0.146,0,0.291,0.032,0.424,0.094c1.059,0.496,2.24,0.435,3.365,0.409c0.47-0.011,0.938-0.021,1.401-0.006c0.845-0.13,1.703-0.195,2.56-0.224v2.001c-0.785,0.028-1.57,0.091-2.343,0.214c-0.065,0.009-0.133,0.013-0.198,0.011c-0.454-0.02-0.914-0.009-1.375,0.003c-0.299,0.007-0.598,0.014-0.896,0.014c-0.956,0-2.067-0.064-3.149-0.516h-4.188c-3.062,0.624-6.561,0.902-11.326,0.902c-1.948,0-3.925-0.044-5.907-0.088c-1.978-0.045-3.96-0.089-5.923-0.089C6.186,3.726,2.955,3.883,0,4.332z"]
  },
  "tooltip_ne": {
    type: "SVG_PATHS",
    width: 16,
    height: 12,
    colors: [0xffffff, 0x000000],
    paths: ["M13.929,12H0V2.339C6.597,3.05,13.376,5.408,13.929,12z", "M15.5,12h-2v-0.377C11.916,5.45,5.862,3.246,0.702,3.246c-0.234,0-0.468,0.02-0.702,0.028V1.273c0.234-0.008,0.468-0.027,0.702-0.027c5.95,0,12.955,2.631,14.769,10.016C15.49,11.34,15.5,11.42,15.5,11.5V12z"]
  },
  "tooltip_e": {
    type: "SVG_PATHS",
    width: 16,
    height: 1,
    colors: [0xffffff, 0x000000],
    paths: ["M13.978,0.5c0,0,0.028,0.185,0.024,0.5H0V0h13.929C13.943,0.168,13.972,0.326,13.978,0.5z", "M15.5,0v1h-2V0H15.5z"]
  },
  "tooltip_se": {
    type: "SVG_PATHS",
    width: 16,
    height: 12,
    colors: [0xffffff, 0x000000],
    paths: ["M14.001,0C13.979,1.779,12.933,7.8,0,9.335V0H14.001z", "M15.5,0v1.5c0,0.162-0.039,0.322-0.115,0.466C12.066,8.273,6.093,9.822,0,10.294V8.293c5.529-0.441,10.631-1.812,13.5-7.045V0H15.5z"]
  },
  "tooltip_s": {
    type: "SVG_PATHS",
    width: 45,
    height: 12,
    colors: [0xffffff, 0x000000],
    paths: ["M45,0v9.335c-2.542,0.302-5.528,0.436-9.072,0.324C13.546,8.954,12.612,12.248,0,10.646V0H45z", "M0,11.437V9.434c1.793,0.09,3.674,0.14,5.706,0.14c3.83,0,7.733-0.156,11.508-0.308c4.899-0.196,9.972-0.397,14.978-0.268c2.367-0.34,4.968-0.406,7.483-0.472C41.455,8.48,43.249,8.432,45,8.293v2.001c-1.769,0.137-3.547,0.188-5.273,0.232c-2.602,0.066-5.06,0.13-7.331,0.463C32.339,10.998,32.282,11,32.223,11c-4.972-0.136-10.033,0.068-14.929,0.264c-3.794,0.152-7.718,0.31-11.588,0.31C3.678,11.573,1.796,11.524,0,11.437z"]
  },
  "tooltip_sw": {
    type: "SVG_PATHS",
    width: 16,
    height: 12,
    colors: [0xffffff, 0x000000],
    paths: ["M16,0v10.646c-1.346-0.171-2.823-0.396-4.476-0.69c0,0-10.044-2.188-10.479-9.955H16z", "M12.787,9.223c1.035,0.09,2.111,0.156,3.213,0.211v2.003c-1.183-0.058-2.335-0.13-3.444-0.227c-0.037-0.003-0.074-0.009-0.111-0.016C6.37,9.969,2.033,6.544,0.546,1.799C0.516,1.702,0.5,1.602,0.5,1.5V0h2v1.344C4.334,6.911,10.251,8.703,12.787,9.223z"]
  },
  "tooltip_w": {
    type: "SVG_PATHS",
    width: 16,
    height: 1,
    colors: [0xffffff, 0x000000],
    paths: ["M16,0v1H1.045C1.027,0.678,1.036,0.341,1.052,0H16z", "M2.5,0v1h-2V0H2.5z"]
  },
  "tooltip_c": {
    type: "SVG_PATHS",
    width: 45,
    height: 1,
    colors: [0xffffff],
    paths: ["M45,0v1H0V0H45z"]
  },
  "tooltip_tip": {
    type: "SVG_PATHS",
    width: 16,
    height: 14,
    colors: [0xffffff, 0x000000],
    paths: ["M2.753,2.369L16,10.23V14H0v-1.525l3.82-0.322L2.753,2.369z", "M2.012,0.287C2.778-0.213,3.578,0.154,4,0.654C6.884,4.77,11.323,6.677,16,9c0,1,0.871,2,0.871,3C15.654,12.127,5.855,4.854,3.553,3.734c0.1,3.053,1.542,6.684,0.699,9.25C2.52,13.23,1.045,13.114,0,12.948C0,12.281,0,11.667,0,11c0.667,0,1.333,0,2,0C5.524,7.953-1.43,3.416,2.012,0.287z"]
  },
  "tooltip_padding_left": 12,
  "tooltip_padding_right": 18,
  "tooltip_padding_top": 2,
  "tooltip_padding_bottom": 6,
  "iphonemenu_right_text_color": 0x324F85,
  "iphonemenu_min_row_height": 27,
  "iphonemenu_padding_top": 5,
  "iphonemenu_padding_bottom": 5,
  "iphonemenu_row_padding": 2,
  "iphonemenu_background_color": 0xFFFFFF,
  "iphonemenu_padding_left": 8,
  "iphonemenu_padding_right": 6,
  "iphonemenu_icon_padding": 4,
  "iphonemenu_corner_radius": 30,
  "shape_default_color": 0xFFFFFF,
  "shape_default_border_color": 0x000000,
  "shape_min_thickness": 2,
  "shape_max_thickness": 4,
  "shape_padding_left": 8,
  "shape_padding_right": 8,
  "shape_padding_top": 4,
  "shape_padding_bottom": 4,
  "shape_icon_padding": 3,
  "curlybrace_vertical": {
    type: "SVG_PATHS",
    width: 15,
    height: 388,
    paths: ["M15,193c0,2.333-0.069,4.818-1,7C-8.898,254.707,29.316,334.251,4,388c-0.333,0-0.667,0-1,0c6.151-25.139,8.342-50.567,6-79c-2.392-29.036-9.9-57.505-5-86c1.607-9.35,6.955-18.614,7-28c0.052-10.883-5.16-22.471-6-35C1.301,104.825,19.615,46.492,0,6c0-1,0-2,0-3c7.357-8.023,9.897,21.396,11,30C17.577,84.299-0.699,146.711,15,193z"]
  },
  "curlybrace_horizontal": {
    type: "SVG_PATHS",
    width: 388,
    height: 15,
    paths: ["M195,14.568c-2.22,0.432-4.522,0.722-7,0C135.069-10.561,53.491,29.863,0,4c0-0.334,0-0.666,0-1c25.138,6.151,50.567,8.342,79,6c29.036-2.392,57.505-9.9,86-5c9.349,1.607,18.614,6.955,28,7c10.883,0.052,22.471-5.16,35-6c55.175-3.699,113.508,14.615,154-5c1,0,2,0,3,0c0.247,1.086,1.988,0.678,2,2c-6.341,7.849-18.659,6.546-28,8C306.812,18.123,245.17-1.109,195,14.568z"]
  },
  "curlybrace_default_color": 0x000000,
  "curlybrace_padding": 20,
  "numericstepper": {
    type: "SVG_SCALE9",
    scale9: [4, 3, 19, 5],
    width: 44,
    height: 26
  },
  "numericstepper_nw": {
    type: "SVG_PATHS",
    width: 3,
    height: 4,
    paths: ["M2.656,4H0.654c0.017-0.701,0.036-1.415,0.05-2.102V0.42H3v2H2.693C2.682,2.941,2.668,3.475,2.656,4z"]
  },
  "numericstepper_n": {
    type: "SVG_PATHS",
    width: 22,
    height: 4,
    paths: ["M0,2.42v-2h3.346c4.609,0,11.869,0,18.654,0.001V2.42c-6.785,0-14.045,0-18.654,0H0z"]
  },
  "numericstepper_ne": {
    type: "SVG_PATHS",
    width: 19,
    height: 4,
    paths: ["M18.414,4h-2.007c-0.039-0.488-0.067-0.985-0.122-1.459L16.221,2.42c-2.082,0.001-7.261,0.003-13.212,0C3.013,2.84,3.012,3.381,3.007,4h-2c0.005-0.624,0.006-1.167,0.001-1.58c-0.334,0-0.671,0-1.008,0V0.421c7.893,0,15.143,0,16.828-0.001h0.609l0.79,1.524l0.021,0.181C18.322,2.733,18.366,3.368,18.414,4z"]
  },
  "numericstepper_e": {
    type: "SVG_PATHS",
    width: 19,
    height: 17,
    paths: ["M9.506,1.244c3.763,0.236,3.899,4.1,5,7c-3,0-6,0-9,0C5.891,4.961,8.431,3.834,9.506,1.244z", "M10.63,17H9.278c-2.582-0.929-2.806-4.213-3.772-6.756c3,0,6,0,9,0C14.134,13.415,11.761,14.586,10.63,17z", "M18.262,9.24c0.296,2.64,0.364,5.266,0.23,7.76h-1.994c0.138-2.446,0.062-5.038-0.235-7.643L16.25,9.248l0.012-0.11c0.335-3.135,0.372-6.25,0.145-9.138h2.007C18.633,2.932,18.594,6.079,18.262,9.24z", "M1.006,0h2C2.995,1.386,2.964,3.203,2.929,5.211C2.86,9.188,2.78,13.841,2.823,17h-2C0.78,13.775,0.86,9.141,0.93,5.176C0.964,3.178,0.995,1.37,1.006,0z"]
  },
  "numericstepper_se": {
    type: "SVG_PATHS",
    width: 19,
    height: 5,
    paths: ["M10.506,0.244C10.044,0.215,9.639,0.13,9.278,0h1.353C10.591,0.084,10.542,0.157,10.506,0.244z", "M18.492,0c-0.058,1.077-0.142,2.137-0.279,3.15L18.12,3.831l-0.669,0.158c-1.556,0.366-3.147,0.31-4.688,0.258c-0.937-0.033-1.821-0.064-2.69-0.005L9.92,4.252L9.771,4.216C8.864,3.997,7.842,4.153,6.759,4.321c-1.314,0.202-2.761,0.426-4.125-0.08c-0.162,0.17-0.385,0.284-0.637,0.306C1.348,4.603,0.679,4.646,0,4.682V2.677C0.304,2.66,0.613,2.646,0.91,2.625C0.865,1.947,0.838,1.047,0.824,0h2c0.011,0.829,0.031,1.562,0.062,2.145c0.968,0.602,2.183,0.415,3.569,0.2c1.179-0.182,2.396-0.369,3.632-0.107c0.921-0.053,1.85-0.021,2.748,0.011c1.221,0.043,2.386,0.084,3.485-0.078C16.4,1.464,16.457,0.737,16.498,0H18.492z"]
  },
  "numericstepper_s": {
    type: "SVG_PATHS",
    width: 22,
    height: 5,
    paths: ["M0,4.531V2.527c0.203-0.001,0.404-0.002,0.608-0.002c2.565,0,5.286,0.07,7.936,0.14C13.135,2.785,17.856,2.904,22,2.677v2.005c-1.7,0.091-3.483,0.127-5.312,0.127c-2.687,0-5.465-0.072-8.196-0.144C5.512,4.586,2.684,4.518,0,4.531z"]
  },
  "numericstepper_sw": {
    type: "SVG_PATHS",
    width: 3,
    height: 5,
    paths: ["M2.584,2.529C2.722,2.528,2.862,2.528,3,2.527v2.004c-0.45,0.003-0.905,0.002-1.346,0.01L0.695,4.558L0.639,3.599C0.57,2.416,0.535,1.209,0.501,0h2.004C2.529,0.846,2.544,1.697,2.584,2.529z"]
  },
  "numericstepper_w": {
    type: "SVG_PATHS",
    width: 3,
    height: 17,
    paths: ["M2.606,2.085C2.481,7.038,2.37,12.113,2.506,17H0.501C0.367,12.081,0.482,6.994,0.606,2.035C0.624,1.345,0.638,0.679,0.654,0h2.001C2.639,0.693,2.624,1.382,2.606,2.085z"]
  },
  "numericstepper_c": {
    type: "SVG_PATHS",
    width: 22,
    height: 17,
    paths: []
  },
  "numericstepper_padding_left": 9,
  "numericstepper_padding_right": 24,
  "numericstepper_padding_top": 6,
  "numericstepper_padding_bottom": 5,
  "numericstepper_button_width": 18,
  "numericstepper_button_background_color": 0xFFFFFF,
  "numericstepper_disabled_color": 0x80808b,
  "iphone_4": {
    type: "SVG_PATHS",
    width: 282,
    height: 510,
    colors: [0xc0c0c0, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M109.698,508.242c-7.667,0-15.333,0-23,0c-25.613-2.186-56.52,3.198-71.009-13.06C1.983,479.805,5.061,443.491,4,415.174c-3.471-92.69-0.621-227.107,0.098-325.32c0-1,0-2.586,0-3.586c1.043-35.139-5.403-71.507,26.01-80.519c12.961-3.718,36.673-0.02,54.728-1.285c35.879-2.515,73.681-0.616,106.862-0.222c24.371,0.289,54.25-1.282,74.184,7.898c24.459,22.768,13.119,80.865,13.119,131.102c0,84.936,1.185,200.157-0.303,295c-0.202,12.895-0.859,36.507-5.018,45.36c-7.53,16.032-16.072,18.638-37.982,21.64C200.35,510.085,144.743,502.602,109.698,508.242zM25.698,83.242c0.664,113.002-1.329,231.569,1,342.907c70.857,3.94,153.006,0.43,226,0c-0.089-118.478,3.145-231.931,2.919-343.465C185.875,77.756,99.405,81.497,25.698,83.242z", "M134,443c-0.333,0-0.667,0-1,0c-6.362,3.638-11.098,8.902-13,17c0,0.333,0,0.667,0,1c-6.437,14.771,6.681,30.839,20,29c0.333,0,0.667,0,1,0C180.58,497.661,168.586,424.902,134,443zM161.597,470.736c-0.657,7.044-5.753,14.482-11.63,15.858c-38.066,8.912-35.564-46.759-3.173-41.234c8.328,1.422,11.87,7.172,14.803,16.917C162.057,465.845,161.597,470.03,161.597,470.736z", "M130.039,456.09c-0.407,7.074,1.78,11.554,1,19c4.16,4.969,17.2,3.144,23,1c0.229-7.229-0.373-13.628-2-19C143.861,455.361,136.613,455.301,130.039,456.09zM134.039,472.09c0-0.333,0-0.667,0-1c-0.074-4.26-1.673-6.994-1-12c4.774-2.043,10.196,1.052,17,0c0.662,4.671,1.572,9.095,2,14C147.793,475.978,136.572,478.124,134.039,472.09z", "M256,357c2.312-26.023,0.362-63.636,1-88c0-0.333,0-0.667,0-1c1.063-4.169,0.259-4.411,0-9c0-1,0-2,0-3c1.801-57.264-0.962-120.526-1-175c-6.474-5.065-19.8-1.301-28-2c-0.333,0-0.667,0-1,0c-3.356-2.311-10.983-0.35-16-1c-0.333,0-0.667,0-1,0c-1.667,0-3.333,0-5,0c-0.333,0-0.667,0-1,0c-52.793,2.207-119.116-0.449-175,2c-0.333,0-0.667,0-1,0c-1.373-2.578-3.127-0.222-5,0c0.842,44.967-0.149,89.588,1,134c0,1-0.008,2,0,3c0.435,53.834,0.593,111.301,0.208,164.268c-0.06,8.237,0.451,16.879,0.323,12.954C24.887,406.621,24.366,414.614,25,426c11.4,3.362,21.915,2.679,34,3c66.267,1.761,132.138-2.346,198-2C257.608,420.24,257.802,381.719,256,357C256,356.333,256,357.667,256,357zM254,308c-0.333,0-1,12.667-1,19c0,2,1.597,78.329,0.141,96.592C253.088,424.257,250,422.667,250,423c-42,0-136,0.667-136,1c-9.333,0-18.667,0-28,0c-9,0-18,0-27,0c-11.112,0.112-20.81-1.189-30-3c-2.174-10.159-0.449-22.217-1-34c0-1.667,0-3.333,0-5c0-2.333,0-4.667,0-7s0-4.667,0-7c0-33.333,0-66.667,0-100c0-0.667,0-1.333,0-2c0-14.666,0-29.334,0-44c0-1,0-2,0-3c0-19.666-0.667-59-1-59c0-4.666,0-9.334,0-14c0-4,0-8,0-12c-0.645-4.355-0.609-20.609,0-23c0.333,0-0.458-0.014-0.125-0.014c0-2,0-4,0-6c-0.333,0,0.458,0.014,0.125,0.014c0-1.666,0-3.334,0-5s0-3.334,0-5c0-1,0-2,0-3c0-2,0.037-5.639,0.037-7.639c0-1.992,121.436,0.127,162.002-0.93c2.333,0,16.336,0.398,17.003,0.398c3,0,35.912,0.387,44.772-0.266C253.963,83.332,253,95.334,253,98c0,9.334,0,18.666,0,28c-0.163,1.162,0.221,47.779,1,48c0,0.666,0,60.666,0,61c0.661,9.006-0.255,28.489-0.255,28.822", "M121.119,35.809c-5.431,0.879-10.983,5.519-9,10.555c1.55,3.935,13.167,4.613,20,4.871c18.578,0.699,36.074,1.447,38-11.366C167.313,32.842,132.967,33.893,121.119,35.809zM115.35,41.249c2.017-4.235,44.511-3.589,46.935-3.363c2.844,0.501,3.268,2.306,3.268,2.681c-0.248,8.734-18.256,7.534-32.348,7.084C132.83,47.65,115.642,48.446,115.35,41.249z", "M281,398c0-1.333,0-2.667,0-4c-0.333,0,0.333,0,0,0c0-3.333,0-6.667,0-10c0-15.667,0-31.333,0-47c0-0.667,0-1.333,0-2s0-1.333,0-2c-0.333,0-0.667,0-1,0c0-4.667,1-9.333,1-14c0-3.667,0-7.333,0-11c0-1.333,0-2.667,0-4c0-6.333,0.333-22,0-22c0-1.333,0-2.667,0-4c-0.646-4.355,1.305-11.305-1-14c-0.828-3.589,2.255-4.838,0-7c0-4,0-8,0-12c0.731-9.785-0.31-17.531,0-29c0.324-0.273,0.324-10.727,0-11c0.011-0.078-0.409-6.391,0-6c0.301,0.083,0.301-5.083,0-5c1.252-8.346-0.58-14.125,0-26c0.316-0.085,0.316-7.915,0-8c0-1,0-2,0-3c0.324-0.338,0.324-11.662,0-12c-0.498-4.641,1.263-5.223,0-8c0-4,0-8,0-12c0-0.334,0-0.666,0-1c-0.333,0-0.667,0-1,0c0-3,0-6,0-9c0.333,0,0.667,0,1,0c0-3.334,0-6.666,0-10c-0.333,0-0.667,0-1,0c0.547-1.453-1.169-5.168,1-5c0.826-2.826-1.409-2.591-1-5c2.169,0.168,0.453-3.547,1-5c-0.333,0-0.667,0-1,0c0-1.666,0-3.334,0-5c-0.333,0-0.667,0-1,0c0-4.334,1.52-7.902,1.52-12.236c-2.27-1.064,0.101-6.492-0.519-9.206c0-0.667,0,0,0-0.667c-2.169,0.169,0.547-1.786,0-3.239c-1.874-1.904-0.371-2.525-0.262-5.013c-1.48-0.186-0.487-6.232-0.57-7.815c-3.822-10.177-1.733-20.306-6.552-29.486c-2.055-0.278-5.621-5.026-6.625-6.356L245,7c-16.969-4.697-44.737-0.427-63.189-3.642C178.811,3.358,176,4,173,4c-5.681-0.652-13.981,1.314-18-1c-2.193,2.209-4.086-0.989-8,0c-1.333,0-2.667,0-4,0c-9.898,0.639-23.382-0.401-36,0c-1.711,2.289-7.633,0.367-11,1c-2.667,0-5.333,0-8,0c-0.745,2.256-5.609,0.391-8,1c-1,0-2,0-3,0s-2,0-3,0c-1.333,0-2.667,0-4,0c-3.367,0.633-9.289-1.289-11,1c-1,0-2,0-3,0c-1.583,0.083-3.629-1.962-3.814-0.48c-1.333,0-2.783-1.605-4.117-1.605c-2.391,0.608-6.206-1.939-6.951,0.316c-4.835-0.169-4.645-1.219-7.351,0.74c-4.78,1.082-5.15,0.989-5.926,1.213c-6.514,1.819-6.997,0.572-10.785,5.117c0,0.334-5.007,4.379-5.007,4.713c-2.211,0.789-3.476,7.04-4.973,8.543C4.9,29.381,4.925,38.925,5,46c0,0.334-1.917,0.49-1.917,0.823C1.986,49.727,4.097,54.098,3,57c-0.333,0,0.333,0,0,0c0,1,0,2,0,3c-0.328,0.535-0.328,13.465,0,14c0,0.334-1,88-1,90c-0.583,11.582-0.287,21.287,0,32c0.333,0,0.667,0,1,0v7c0,1,0,2,0,3c-0.333,0-0.667,0-1,0c0,4.334,0,8.666,0,13c0,0.666,0,1.334,0,2c0,35.667,0,71.333,0,107c-0.333,0,0.333,0,0,0c0,6.667,0,13.333,0,20c0,1.333,1,2.667,1,4c0.333,0-0.333,0,0,0c-0.665,17.335,1.331,37.331-1,53c0,0.667,0,1.333,0,2c-0.333,0-0.333,4,0,4c0,2.667,0,5.333,0,8c0,1.333,0,2.667,0,4c0,2.333,3.733,52.767,3.733,53.1c-0.423,2.09,1.858,7.393,2.191,7.393c2.972,5.361,5.111,11.294,10.241,14.498c0.903,1.763,7.568,5.722,9.331,6.625C32.534,506.91,36.847,510.162,43,509c2.356-0.022,18.547-0.547,20,0c6.677,0.656,32.975-0.358,37-1c3.333,0,6.667,0,10,0c0.333,0,0.667,0,1,0c4,0,8,0,12,0c1.453-0.547,5.168,2.169,5,0c1,0,2,0,3,0c0.21,0.323,9.79,0.323,10,0c1,0,2,0,3,0c2,0,4,0,6,0c2.333,0,4.667,0,7,0c8.333,0,16.613,0.49,24.946,0.49c3,0,28.461,0.162,29.011,0.266C212.765,508.521,239,507.333,239,507c10.321-2.346,18.229-7.104,26-12c3.087-2.247,5.753-4.913,8-8c0.891-5.442,3.926-8.741,4-15c0.171-1.496-0.357-3.69,1-4c0-0.333,0-0.667,0-1c0.575-1.758-1.209-5.875,1-6c-0.827-4.076-0.153-3.443,0-8c0.333,0,0.667,0,1,0c0-1.333,0-2.667,0-4c0-0.333,0-0.667,0-1c-0.132-4.59-0.853-12.564,0-17c0-1,0-2,0-3c-0.799-8.087,1.423-15.207,0-21C282.27,405.937,280.381,400.714,281,398C281.333,398,280.667,398,281,398zM276.633,431.107c0,5.397-3.025,40.115-3.025,40.452c-0.942,25.681-14.789,27.22-38.452,33.403c-0.337,0-0.676,0-1.013,0c-3.74,0.646-8.083-3.311-10.146-0.987c-0.337,0-10.675,0-11.012,0c-14.531-0.43-19.799-0.984-36.538,0.013c-0.337,0-0.675,0-1.013,0c-7.976-1.481-22.447-1.834-30.367,0c-0.337,0-0.675,0-1.013,0c-2.361,0-4.723,0-7.085,0c-1.012,0-2.024,0-3.036,0c-21.394,0.543-35.896,0.891-54.661,2.012c-42.027,2.242-50.492-3.839-50.829-3.839c-15.407-4.839-20.098-17.686-21.277-32.421C2,436,5.403,215.164,5.403,214.489c2.569-20.155,1-59.722,1-60.733c-0.048-9.737-2.024-28.006-2.024-28.344c2.937-22.585,0.09-50.82,3.024-73.904c0-0.338,0.674-13.645,0.674-13.982C11.114,19.307,14.608,13.607,33.157,8c17.417-3.031,69.826-3.705,103.812-3.055c33.755,0.645,66.3-0.662,95.138,2.023C251.897,8.813,266.173,8,270.868,24.396c3.364,8.881,4.415,15.367,4.559,22.427c0,1.454-0.062,0.562,0.27,10.577l0.365,5.522c0.834,16.409,0.83,16.292,0.952,26.667c0,12.146-1,25.312-1,37.459c0,20.583,0.952,39.418,0.952,60c0,1.35-1.934,93.486-0.333,99.31c0,7.76,0,15.521,0,23.28c0,3.374,0,6.749,0,10.122c0.647,4.179,1.907,8.304,0.012,12.147c0.956,23.234-1.225,49.363,0,72.881C278.927,406.978,276.633,426.383,276.633,431.107z", "M257,10l-1.1-0.367C256.298,9.725,256.679,9.832,257,10z", "M249,8l6.9,1.633C254.126,9.229,249.825,9.353,249,8z", "M90.809,37.74c-1.079,0.869-2.494,4.244-2.043,8.118c0.352,3.026,2.99,3.55,4.542,3.747c4.219,0.539,8.192,1.113,8.629-8.742C101.299,35.457,94.232,34.982,90.809,37.74z", "M191.643-0.281c-3.252,0.107-2.297,2.094-3.004,3.907c-0.612,1.569,9.704,0.772,14.366,0.874c25.295,0.548,25.896,2.33,25.896-3.523C226.361-0.644,196.933-0.456,191.643-0.281z", "M0.362,67.547c0.242,2.225,2.094,1.574,3.907,2.06c1.569,0.42,0.772-6.653,0.874-9.851C5.691,42.41,7.474,41.998,1.62,41.998C-0.559,43.309-0.045,63.805,0.362,67.547z", "M0.362,111.805c0.242,0.956,2.094,0.677,3.907,0.885c1.569,0.181,0.772-2.857,0.874-4.231c0.548-7.451,2.33-7.628-3.523-7.628C-0.559,101.393-0.045,110.197,0.362,111.805z", "M1.62,158.697c5.854,0,4.071-0.177,3.523-7.628c-0.102-1.374,0.695-4.412-0.874-4.231c-1.813,0.208-3.665-0.071-3.907,0.885C-0.045,149.33-0.559,158.135,1.62,158.697z", "M48.678,509.938c-11.147,0-23.667-1.261-32.006-10.616C2.419,483.33,0.892,479.42,2,415.14c1.154-66.876,0.541-158.233,0-238.836c-0.208-31.013-0.405-60.306-0.5-87.055l-0.022-0.382C0.47,71.017-0.649,12.16,27.297,4.26c8.497-2.402,22.223-2.452,35.496-2.5c7.49-0.027,15.235-0.055,21.766-0.513c9.365-0.657,19.458-0.976,30.855-0.976c14.746,0,29.525,0.533,43.818,1.05c11.349,0.41,22.068,0.798,32.49,0.921c5.096,0.061,10.359-0.056,15.965-0.18c4.631-0.102,9.42-0.207,14.178-0.207c15.3,0,32.732,0.996,46.022,8.387c14.692,8.721,14.018,58.253,13.364,106.154c-0.128,9.446-0.25,18.369-0.25,26.846c0,26.602,0.096,55.029,0.197,85.124c0.242,72.03,0.493,146.512-0.5,209.907c-0.198,12.647-2.177,41.774-7.027,51.081c-6.853,13.15-23.183,15.427-38.975,17.628l-1.72,0.24c-7.887,1.107-17.365,1.646-28.978,1.646c-10.079,0-20.825-0.396-31.218-0.777c-10.406-0.383-21.167-0.778-31.271-0.778c-12.91,0-22.854,0.633-31.296,1.991l-0.318,0.025l-23.154-0.007c-2.801-0.239-5.802-0.355-9.174-0.355c-4.932,0-9.964,0.251-14.83,0.494C57.848,509.707,53.23,509.938,48.678,509.938zM115.414,4.271c-11.303,0-21.305,0.316-30.576,0.966c-6.663,0.468-14.475,0.495-22.03,0.522C49.803,5.807,36.356,5.855,28.385,8.108C2.657,15.383,4.784,76.47,5.472,88.642c0.02,0.345,0.028,0.55,0.028,0.601c0.095,26.735,0.292,56.025,0.5,87.035c0.541,80.625,1.155,172.007,0,238.931c-1.119,64.871,0.563,66.76,13.659,81.451c7.287,8.177,18.759,9.278,29.02,9.278c4.453,0,9.022-0.229,13.86-0.47c4.915-0.246,9.998-0.5,15.03-0.5c3.453,0,6.538,0.118,9.429,0.362h22.741c8.623-1.376,18.725-2.017,31.773-2.017c10.178,0,20.976,0.396,31.419,0.78c10.356,0.381,21.064,0.775,31.07,0.775c11.424,0,20.721-0.526,28.421-1.607l1.724-0.241c14.789-2.062,30.081-4.193,35.98-15.514c4.227-8.112,6.368-36.109,6.574-49.296c0.993-63.357,0.742-137.82,0.5-209.831c-0.102-30.099-0.197-58.528-0.197-85.138c0-8.504,0.122-17.439,0.251-26.9c0.577-42.379,1.297-95.12-11.357-102.632c-12.443-6.917-29.244-7.854-44.03-7.854c-4.714,0-9.48,0.105-14.09,0.207c-5.64,0.124-10.952,0.239-16.1,0.18c-10.471-0.125-21.214-0.513-32.588-0.924C144.83,4.805,130.087,4.271,115.414,4.271z"]
  },
  "iphone_5": {
    type: "SVG_PATHS",
    width: 282,
    height: 573,
    colors: [0xc0c0c0, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M109.698,571.742c-7.667,0-15.333-1-23-1c-25.613-2.186-56.52,4.198-71.009-12.06C1.983,543.305,5.061,503.991,4,475.674c-3.471-92.69-0.621-286.107,0.098-384.32c0-1,0-2.586,0-3.586c1.043-35.139-6.403-72.507,25.01-81.519c12.961-3.718,37.673-0.02,55.728-1.285c35.879-2.515,72.681,0.384,105.862,0.778c24.371,0.29,55.25-2.282,75.184,6.898c18.45,19.958,11.119,81.866,11.119,132.102c0,84.936,1.185,263.157-0.303,358c-0.202,12.895-0.016,37.995-4.536,46.67c-6.615,12.694-24.617,18.104-40.464,20.33C196.366,574.704,144.743,566.102,109.698,571.742zM25.698,82.742c0.664,113.002-1.329,294.569,1,405.907c70.857,3.94,152.006,0.43,225,0c-0.089-118.478,3.145-292.931,2.919-404.465C184.875,79.256,99.405,80.997,25.698,82.742z", "M130,507.743c-0.333,0-0.667,0-1,0c-6.362,3.638-11.098,8.902-13,17c0,0.333,0,0.667,0,1c-6.437,14.771,6.681,30.839,20,29c0.333,0,0.667,0,1,0C176.58,562.404,164.586,489.646,130,507.743zM157.597,535.479c-0.657,7.044-5.753,14.482-11.63,15.858c-38.066,8.912-35.564-46.759-3.173-41.234c8.328,1.422,11.87,7.172,14.803,16.917C158.057,530.588,157.597,534.773,157.597,535.479z", "M126.039,520.833c-0.407,7.074,1.78,11.554,1,19c4.16,4.969,17.2,3.144,23,1c0.229-7.229-0.373-13.628-2-19C139.861,520.104,132.613,520.044,126.039,520.833zM130.039,536.833c0-0.333,0-0.667,0-1c-0.074-4.26-1.673-6.994-1-12c4.774-2.043,10.196,1.052,17,0c0.662,4.671,1.572,9.095,2,14C143.793,540.721,132.572,542.867,130.039,536.833z", "M256,416.743c2.312-26.023,0.362-126.636,1-151c0-0.333,0-0.667,0-1c1.063-4.169,0.259-4.411,0-9c0-1,0-2,0-3c2.048-60.364-0.962-116.526-1-171c-6.474-5.065-19.8-1.301-28-2c-0.333,0-0.667,0-1,0c-3.356-2.311-11.983,0.649-17,0c-0.333,0-0.667,0-1,0c-1.667,0-3.333,0-5,0c-0.333,0-0.667,0-1,0c-52.793,2.206-118.116-2.45-174,0c-0.333,0-0.667,0-1,0c-1.373-2.579-3.127-0.222-5,0c0.842,44.968-1.432,84.992,1,131c0,1,0,2,0,3c-1.017,51.576-1.936,173.996-0.792,227.268c0.148,8.236,0.399,16.482,0.792,24.732c-0.187-3.927-0.341-7.854-0.468-11.778c0.355,12.399,0.834,27.393,1.468,38.778c11.4,3.362,21.915,0.679,34,1c66.267,1.761,132.829-3.829,198-1C257.685,470.489,256.652,449.666,256,416.743C256,416.076,256,417.41,256,416.743zM254,367.743c-0.333,0-1,12.667-1,19c0,2-0.101,83.316,0.22,101.487c0.012,0.667-3.22-0.153-3.22,0.18c-42,0-136,0-136,0.333c-9.333,0-18.667,0-28,0c-9,0-18,0-27,0c-11.112,0.112-21.81,0.811-31-1c-2.174-10.159-0.449-29.217-1-41c0-1.667,0-3.333,0-5c0-2.333,0-4.667,0-7s0-4.667,0-7c0-33.333,0-129.667,0-163c0-0.667,0-1.333,0-2c0-14.667,0-29.333,0-44c0-1,0-2,0-3c0-19.667,0.333-59,0-59c0-4.667,0-6.333,0-11c0-4,0-8,0-12c-0.645-4.354-0.609-20.608,0-23c0.333,0-0.458-0.015-0.125-0.015c0-2,0-4,0-6c-0.333,0,0.458,0.015,0.125,0.015c0-1.667,0-4.333,0-6s0-3.333,0-5c0-1,0-2,0-3c0-2-0.321-5.047,0.037-7.015c0,0,127.348-1.35,167.963,0.015c2.332,0.078,4.667,0,7,0c0.667,0,1.333,0,2,0c3,0,6.128-0.374,9.128-0.374c6,0,37.02,0.816,37.687,0.816c3.124-0.457,2.186,9.891,2.186,12.558c0,9.333,0,19.667,0,29c-0.163,1.163-0.779,44.779,0,45c0,0.667,0,60.667,0,61c0.661,9.006-0.255,28.489-0.255,28.822", "M120.119,43.552c-5.431,0.878-10.983,5.519-9,10.554c1.55,3.936,13.167,4.614,20,4.871c18.578,0.701,36.074,1.447,38-11.365C166.313,40.585,131.967,41.635,120.119,43.552zM114.35,48.992c2.017-4.235,44.511-3.589,46.935-3.363c2.844,0.501,3.268,2.307,3.268,2.681c-0.248,8.734-18.256,7.534-32.348,7.083C131.83,55.393,114.642,56.189,114.35,48.992z", "M279,453.743C278.667,453.743,279.333,453.743,279,453.743c0-3.333,0-6.667,0-10c0-15.667,0-31.333,0-47c0-0.667,0-1.333,0-2s0-1.333,0-2c2.571,0,3.237-51,0-51c0-1.333,0-65.667,0-67c-0.646-4.355,1.305-11.305-1-14c-0.828-3.589,2.255-4.838,0-7c0-4,0-8,0-12c0.731-9.785-0.31-17.53,0-29c0.324-0.274,0.324-10.726,0-11c0.011-0.079-0.409-6.39,0-6c0.301,0.083,0.301-5.083,0-5c1.252-8.346-0.58-14.126,0-26c0.316-0.085,0.316-7.915,0-8c0-1,0,1,0,0c0.324-0.337,0.324-11.663,0-12c-0.498-4.64,1.263-5.223,0-8c0-4,0-8,0-12c0-0.333,0-0.667,0-1c-0.333,0-0.667,0-1,0c0-3,0-6,0-9c0.333,0,0.667,0,1,0c0-3.333,0-6.667,0-10c-0.333,0-0.667,0-1,0c0.547-1.453-1.169-5.169,1-5c0.826-2.826-0.409-2.591,0-5c2.169,0.169,0.453-3.547,1-5c-0.333,0-0.667,0-1,0c0-1.667,0-3.333,0-5c-0.333,0-0.667,0-1,0c0-5.333,1.52-8.903,1.52-12.236c-2.27-1.063,0.101-6.492-0.519-9.206c0-0.666,0,0,0-0.666c-2.169,0.168,0.547-1.787,0-3.24c-1.874-1.904-0.371-2.525-0.262-5.014c-1.48-0.186-0.487-6.23-0.57-7.814c-3.822-10.178-1.733-20.305-6.552-29.486c-1.363-2.962-3.772-4.037-6.625-6.356L244,6.743c-16.969-4.698-68-3-71-3c-5.681-0.653-13.981,1.315-18-1c-2.193,2.209-4.086-0.989-8,0c-1.333,0-2.667,0-4,0c-9.898,0.638-23.382-0.401-36,0c-3.792-0.196-7.633,0.367-11,1c-2.667,0-19.979-0.972-20,0c-0.021,1-48.383,0.96-49.159,1.184c-6.514,1.819-6.997,0.571-10.785,5.117c0,0.333-4.007,5.379-4.007,5.712c-2.211,0.789-3.476,7.041-4.973,8.543C2.294,34.904,4.863,32.888,5,45.743c0,0.333-1.917,0.489-1.917,0.823C1.986,49.469,4.097,53.84,3,56.743c-0.333,0,0.333,0,0,0c0,1,0,2,0,3c-0.328,0.534-0.328,13.466,0,14c0,0.333,0,0.667,0,1c-2.255,0.745,0.609,5.608,0,8c0,1,0,2,0,3c-0.333,0-0.667,0-1,0c0,1.667,0,3.333,0,5c0.333,0,0.667,0,1,0c0,0.333,0,0.667,0,1c0.664,13.336-3.33,29.329-1,41c0.435-1.417-0.565,7.417-1,6c-0.031-0.284,2.036,20.944,1,19c0,2,0,1,0,3c-0.583,11.583-0.287,21.286,0,32c0.333,0,0.667,0,1,0v7c0,1,0,2,0,3c-0.333,0-0.667,0-1,0c0,4.333,0,8.667,0,13c0,0.667,0,1.333,0,2c0,35.667,0,134.333,0,170c-0.333,0,0.333,0,0,0c0,6.667,0.667,24,1,24c-0.665,17.335-1,54.333-1,55c-0.333,0-0.333,3,0,3c0,2.667,0,5.333,0,8c0,1.333,0,2.667,0,4c0,2.333,0.667,7,1,7c0,2,0,4,0,6c0,0.333,0,0.667,0,1c-0.772,4.772,1.564,10.437,1,15c0,0.333,0,0.667,0,1c1.697,6.478,1.733,26.767,1.733,27.1c-0.423,2.09,1.858,6.393,2.191,6.393c2.972,5.361,5.111,11.294,10.241,14.498c0.903,1.763,7.568,5.722,9.331,6.625c5.038,2.295,8.764,4.854,14.917,3.692c2.356-0.022,19.096,0.824,20.549,1.371c6.677,0.656,22.736-0.68,23.736-0.68c-3.828,3.72,26.164-0.999,23.302-0.999c0.333,0,0.667,0,1,0c4,0,8,0,12,0c1.453-0.547,5.168,2.169,5,0c1,0,2,0,3,0c0.21,0.323,9.789,0.323,10,0c1,0,2,1,3,1c2,0,4,0,6,0c2.333,0,4.667,0,7,0c8.333,0,16.613,0.49,24.946,0.49c3,0,27.461,0.162,28.011,0.266c1.808-0.235,28.043-1.423,28.043-1.756c10.321-2.346,18.642-5.183,26.413-10.078c3.087-2.247,4.34-6.835,6.587-9.922c0.891-5.442,4.926-8.741,5-15c0.171-1.496,3-21.667,3-23C279,511.41,279,455.076,279,453.743zM274.935,513.792c-1.065,4.375-2.193,21.596-2.193,21.933c-4.927,28.494-15.353,25.354-38.586,31.981c-0.337,0-20.876,0.579-21.212,0.602c-16.308,1.076-37.172,0.682-37.51,0.682c-7.976-1.481-31.042,0-31.38,0c-2.361,0-4.724-1-7.085-1c-1.012,0-2.024,0-3.036,0c-21.394,0.543-35.896-0.109-54.661,1.012c-0.675,0-1.35,0-2.024,0c-13.783-1.064-26.836-0.763-39.262-3.185c-0.338,0-10.886-3.904-11.223-3.904c-14.476-6.344-17.39-18.229-18.569-32.964c-8.817-61.959-2.79-317.041-2.79-317.716c2.569-20.154,1-56.722,1-57.734c-0.048-9.736-2.024-28.005-2.024-28.342c2.937-22.586-1.91-50.822,1.024-73.905c0-1.056,0.674-12.927,0.674-13.981c3.037-18.22,7.531-24.919,26.08-30.526c17.417-3.031,70.826-2.705,104.812-2.055c33.755,0.646,65.3-0.662,94.138,2.024c19.791,1.844,35.203,0.893,39.898,17.29c3.364,8.881,3.278,15.504,3.422,22.564c0,0.338,1.27,10.577,1.27,10.577c0.187,20.851,0.184,20.851,0.317,32.188c0,12.146-1,25.313-1,37.461c0,20.582,0.952,36.417,0.952,56.999c0,1.351,1.133,150.438,2.359,173.986C280.607,359.967,274.935,508.394,274.935,513.792z", "M256,9.743l-1.1-0.367C255.298,9.467,255.679,9.576,256,9.743z", "M248,7.743l6.9,1.633C253.126,8.973,248.825,9.096,248,7.743z", "M135.396,23.834c-0.869,0.701-2.01,3.421-1.646,6.542c0.284,2.439,2.409,2.861,3.66,3.02c3.399,0.435,6.602,0.898,6.954-7.045C143.85,21.996,138.154,21.612,135.396,23.834z", "M191.643,0.219c-3.252,0.107-2.297,2.094-3.004,3.907c-0.612,1.569,9.704,0.773,14.366,0.874c25.295,0.548,25.896,2.33,25.896-3.523C226.361-0.144,196.933,0.044,191.643,0.219z", "M0.362,106.047c0.242,2.225,2.094,1.574,3.907,2.06c1.569,0.42,0.772-6.653,0.874-9.851C5.691,80.91,7.474,80.498,1.62,80.498C-0.559,81.809-0.045,102.305,0.362,106.047z", "M0.362,153.682c0.242,1.526,2.094,1.08,3.907,1.413c1.569,0.288,0.772-4.563,0.874-6.756c0.548-11.896,2.33-12.179-3.523-12.179C-0.559,137.059-0.045,151.115,0.362,153.682z", "M1.62,201.512c5.854,0,4.071-0.283,3.523-12.184c-0.102-2.193,0.695-7.047-0.874-6.76c-1.813,0.334-3.665-0.112-3.907,1.414C-0.045,186.549-0.559,200.611,1.62,201.512z", "M184.507,573.11c-10.344,0-20.898-0.105-31.104-0.208c-9.922-0.1-19.294-0.194-28.228-0.194c-5.759,0-10.745,0.04-15.243,0.122c-3.903,0.015-7.734-0.105-11.444-0.207c-3.684-0.102-7.493-0.206-11.298-0.206c-2.428-0.117-4.917-0.173-7.705-0.173c-4.689,0-9.549,0.161-14.249,0.316c-4.608,0.152-9.373,0.311-13.884,0.311c-12.948,0-27.197-1.198-34.714-10.09C2.807,546.423,0.893,539.844,2,475.64c1.141-66.146,0.294-186.833-0.387-283.808C1.348,154.005,1.098,118.275,1,90.749v-3.007c0.19-6.429,0.091-12.971-0.004-19.297C0.555,39.188,0.139,11.55,27.29,3.762c6.194-1.777,15.04-2.005,22.054-2.005c3.887,0,7.993,0.08,11.963,0.158c3.98,0.078,8.097,0.159,12.002,0.159c4.439,0,8.119-0.107,11.25-0.326c9.365-0.656,19.458-0.976,30.855-0.976c14.746,0,29.525,0.534,43.818,1.05c11.349,0.41,22.068,0.797,32.49,0.921c5.079,0.063,10.381-0.055,16.033-0.178c4.722-0.103,9.604-0.208,14.471-0.208c15.657,0,33.456,0.996,46.746,8.386c13.931,8.269,12.815,52.452,11.634,99.23c-0.311,12.308-0.604,23.934-0.604,34.77c0,28.58,0.252,68.464,0.519,110.69c0.547,86.417,1.166,184.365,0.179,247.341l-0.004,0.265c-0.281,17.952-0.571,36.516-8.023,50.816c-7.095,13.615-23.332,16.641-39.822,17.882C220.392,572.674,205.03,573.11,184.507,573.11zM125.175,568.708c8.954,0,18.335,0.095,28.268,0.194c10.196,0.103,20.74,0.208,31.064,0.208c20.421,0,35.687-0.433,48.041-1.362c15.338-1.154,30.385-3.861,36.576-15.741c7.015-13.462,7.297-31.544,7.57-49.031l0.004-0.265c0.986-62.932,0.368-160.856-0.179-247.254c-0.267-42.232-0.519-82.123-0.519-110.715c0-10.887,0.294-22.537,0.605-34.871c1.049-41.537,2.238-88.618-9.627-95.661c-12.443-6.918-29.609-7.854-44.754-7.854c-4.824,0-9.684,0.105-14.384,0.208c-5.687,0.123-11.038,0.24-16.167,0.179c-10.471-0.125-21.214-0.513-32.588-0.923c-14.257-0.515-29-1.047-43.673-1.047c-11.303,0-21.305,0.316-30.576,0.966c-3.225,0.226-6.996,0.336-11.529,0.336c-3.944,0-8.081-0.081-12.08-0.159c-3.952-0.078-8.038-0.157-11.885-0.157c-6.752,0-15.235,0.21-20.951,1.85C4.185,14.551,4.561,39.499,4.996,68.386c0.096,6.37,0.195,12.957,0.003,19.416L5,90.742c0.098,27.512,0.348,63.238,0.614,101.062C6.294,288.8,7.142,409.514,6,475.708c-1.123,65.081,0.959,69.429,13.693,84.491c6.461,7.642,19.63,8.672,31.659,8.672c4.445,0,9.177-0.156,13.752-0.308c4.732-0.157,9.625-0.319,14.381-0.319c2.855,0,5.407,0.058,7.803,0.176c0.042-0.001,0.085-0.001,0.128-0.001c3.722,0,7.513,0.104,11.182,0.205c3.684,0.102,7.493,0.206,11.298,0.206C114.382,568.748,119.391,568.708,125.175,568.708z"]
  },
  "iphone_6": {
    type: "SVG_PATHS",
    width: 380,
    height: 754,
    colors: [0x010101, 0x010101, 0x010101, 0x010101, 0xc1c0c0, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M2,90.642h8.418c1.105,0,2,0.896,2,2v24.559c0,1.104-0.895,2-2,2H2c-1.104,0-2-0.896-2-2V92.642C0,91.538,0.896,90.642,2,90.642", "M2,145.275h8.418c1.105,0,2,0.896,2,2v48.15c0,1.104-0.895,2-2,2H2c-1.104,0-2-0.896-2-2v-48.15C0,146.171,0.896,145.275,2,145.275", "M369.582,145.275H378c1.104,0,2,0.896,2,2v48.15c0,1.104-0.896,2-2,2h-8.418c-1.105,0-2-0.896-2-2v-48.15C367.582,146.171,368.477,145.275,369.582,145.275", "M2,203.633h8.418c1.105,0,2,0.896,2,2v48.15c0,1.104-0.895,2-2,2H2c-1.104,0-2-0.896-2-2v-48.15C0,204.529,0.896,203.633,2,203.633", "M375.872,115.945l1-65.887c1.963-28.228-21.387-47.001-49-47.001l-28.842-1H134.369H55.872c-32.953,0-50,10.852-49,48.001l-1,367.603L4.168,672.778l1.704,31.278c0,27.613,22.386,47,50,47h103.639h168.361c27.619,0,50-19.387,50-47L375.872,115.945zM353.872,660.057h-324V660H28V88h324v0.057h1.872V660.057z", "M209.683,53.017h-39.357c-1.763,0-3.13-0.472-4.06-1.404c-1.27-1.272-1.267-2.903-1.266-3.686c-0.004-0.043-0.021-0.098-0.035-0.151c-0.117-0.45-0.39-1.506,0.372-2.49C166.174,44.205,167.622,44,170,44h39.947c2.851,0,5,1.661,5,3.862c0,1.009,0.092,1.328,0.172,1.494c0.345,0.714,0.303,1.533-0.111,2.194C214.23,52.788,212.604,53.016,209.683,53.017M209.673,49.017h0.008c0.536,0,0.965-0.01,1.306-0.025c-0.021-0.253-0.031-0.514-0.036-0.785c-0.187-0.095-0.535-0.207-1.003-0.207H170c-0.407,0-0.736,0.008-0.999,0.021c0.002,0.283,0.015,0.684,0.096,0.766c0.064,0.064,0.409,0.23,1.229,0.23H209.673L209.673,49.017z", "M187.995,736.06c-7.36,0-14.091-3.305-18.952-9.304c-4.476-5.524-7.043-13.041-7.043-20.623C162,690.014,172.76,680,190.078,680c9.826,0,17.938,3.076,22.841,8.661c4.604,5.246,6.348,12.44,5.18,21.386c-1.562,11.958-5.891,18.939-14.038,22.637C199.473,734.766,193.316,736.06,187.995,736.06M190.078,684C175.001,684,166,692.274,166,706.133c0,16.229,11.183,25.927,21.995,25.927c4.716,0,10.374-1.185,14.414-3.019c6.837-3.103,10.343-8.938,11.724-19.512c1.013-7.754-0.407-13.887-4.22-18.229C205.78,686.593,198.736,684,190.078,684", "M190.347,32.691c-3.74,0-6.347-3.688-6.347-6.997c0-3.353,2.61-5.694,6.347-5.694c3.5,0,6.347,2.847,6.347,6.346C196.693,27.915,196.693,32.691,190.347,32.691M190.347,24C189.64,24,188,24.165,188,25.694c0,1.456,1.206,2.997,2.347,2.997c2.247,0,2.347-0.524,2.347-2.345C192.693,25.052,191.641,24,190.347,24", "M139.706,55.392c-4.327,0-7.59-3.035-7.59-7.06l-0.001-0.275c-0.007-1.516-0.019-4.055,1.802-5.884c1.291-1.298,3.239-1.956,5.789-1.956c4.184,0,7.588,3.404,7.588,7.587C147.294,55.392,141.583,55.392,139.706,55.392M139.706,44.217c-0.995,0-2.314,0.135-2.954,0.777c-0.649,0.652-0.643,1.923-0.637,3.045l0.001,0.293c0,2.114,1.803,3.06,3.59,3.06c2.549,0,3.588-0.324,3.588-3.588C143.294,45.826,141.685,44.217,139.706,44.217", "M350,92H28.5c-1.104,0-2-0.896-2-2s0.896-2,2-2H350c1.104,0,2,0.896,2,2S351.104,92,350,92", "M353,660.5c-1.104,0-2-0.896-2-2V89c0-1.104,0.896-2,2-2s2,0.896,2,2v569.5C355,659.604,354.104,660.5,353,660.5", "M348.054,660H28.942c-1.105,0-2-0.896-2-2s0.895-2,2-2h319.112c1.104,0,2,0.896,2,2S349.157,660,348.054,660", "M326,753h-0.035l-57.5-1l-214.338-0.999c-2.653,0.346-5.328,0.521-7.953,0.521c-14.051,0-23.937-4.781-31.467-13.462C7.502,729.755,4,718.614,4,703.999V46c0-16.592,4.578-28.055,13.996-35.044C26.23,4.846,36.306,2,54,2l261.263,1l10.829,1C357.176,4.03,373,19.335,373,46.001V121.5l1.879,511.993L376,703.967C376,731.935,355.645,753,326,753M53.873,746.999L54.01,747l214.5,1l57.508,1C353.374,748.993,372,729.649,372,703.999l-1.121-70.467L369,121.507V46.001C369,12.937,342.806,8,326,8l-0.184-0.008L315.071,7L53.993,6C19.458,6,8,17.215,8,46v657.999c0,13.61,3.182,23.894,9.728,31.439c6.759,7.792,15.572,12.083,28.446,12.083c2.495,0,5.04-0.17,7.564-0.504L53.873,746.999zM352,660H28V88h324V660zM32,656h316V92H32V656z", "M351,659H29V89h322V659zM31,657h318V91H31V657z", "M327.872,753.057H55.873c-29.625,0-49.971-21.039-50.001-48.945l-1.705-31.285l0.001-0.062l1.704-255.118l1-367.594C6.4,32.554,9.948,20.137,17.72,12.153C25.754,3.901,36.512,0.057,55.873,0.057H299.03l28.911,1.001c15.573-0.001,28.001,5.849,37.516,16.047c8.335,8.936,12.281,20.359,11.414,33.037l-0.999,65.833l2,588.075C377.872,731.991,357.517,753.057,327.872,753.057M8.168,672.732l1.704,31.324c0,25.654,18.636,45,46.001,45h271.999c27.364,0,46-19.346,46-45l-2-588.105l1.005-66.032c0.804-11.565-2.773-21.968-10.345-30.085c-8.761-9.391-20.123-14.777-34.66-14.777l-28.911-1.001L55.873,4.057c-18.208,0-30.057,3.459-37.287,10.886c-6.979,7.17-8.157,18.639-7.714,35.061l-1,367.662L8.168,672.732zM353.872,660.057h-324v-572h324V660.057zM33.872,656.057h316v-564h-316V656.057z"]
  },
  "iphone_X": {
    type: "SVG_PATHS",
    width: 434,
    height: 867,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M367.04,0.471h-300c-35.2,0-62.94,28.8-62.94,64v738c0,35.2,27.74,64,62.94,64h300c35.2,0,64-28.8,64-64v-738C431.04,29.271,402.24,0.471,367.04,0.471z M407,803.728c0,22.524-18.296,40.849-40.785,40.849H67.785C45.296,844.576,27,826.252,27,803.728V63.424c0-22.523,17.627-40.848,39.293-40.848h38.299v0.05h2.984c3.84,0,6.963,3.128,6.963,6.974v4.135c0,10.438,8.479,18.93,18.901,18.93h167.121c10.422,0,18.9-8.492,18.9-18.93V29.6c0-3.846,3.124-6.974,6.963-6.974h2.984v-0.05h38.299c21.666,0,39.293,18.325,39.293,40.848V803.728z", "M431.397,197.821c0-38.25,0-76.499,0-114.749c0-5.902,0.057-11.805,0-17.706c-0.396-41.358-26.289-65.229-67.992-64.817c-47.242,0.467-94.488,0-141.732,0c-50.049,0-100.245-0.55-149.848,0C51.127,0.779,49.626,1.15,39.926,5.95C19.365,16.123,7.798,27.301,3.239,55.051C1.927,63.04,2.487,71.622,2.487,79.702c0,11.502,0,23.005,0,34.508c-0.763,0.006-1.438,0.186-1.658,0.716c-0.943,2.267-0.399,5.274-0.399,7.675c0,8.129,0,16.258,0,24.387c0,0.031,0.009,0.068,0.012,0.101c-0.051,0.317,0.022,0.784,0.3,1.452c0.357,0.857,0.919,1.979,1.746,2.652c0,8.773,0,17.547,0,26.32c-0.82,0.438-1.514,1.127-1.904,2.064c-0.302,0.725-0.02,2.134-0.02,2.875c0,2.685,0,5.37,0,8.055c0,14.82,0,29.641,0,44.461c0,1.448,0,2.896,0,4.343c-0.086,0.208-0.1,0.433,0,0.672c0,0.375,0.19,0.648,0.491,0.834c0.391,0.517,0.881,0.892,1.433,1.115c0,4.46,0,8.92,0,13.381c-1.059-0.035-2.011,0.421-2.011,1.726c0,4.291,0,8.582,0,12.873c0,14.549,0,29.097,0,43.646c0,1.131,0,2.263,0,3.394c0,0.016,0.005,0.036,0.006,0.053c-0.065,0.307,0.005,0.784,0.3,1.493c0.425,1.021,0.987,1.922,1.705,2.518c0,17.558,0,35.116,0,52.674c0,65.552-1.453,130.929,0,196.653c1.175,53.154,0,106.338,0,159.506c0,23.225-0.055,46.449,0,69.672c0.115,48.59,31.006,66.432,67.083,66.432c14.911,0,29.241,1.05,44.152,1.05c61.978,0,123.03-0.45,185.005-0.45c20.076,0,41.181-0.961,61.255-0.961c8.125,0,22.651-1.09,30.845-1.839c16.4-1.5,35.924-15.86,39-43c0.792-6.984,1.493-13.888,1.571-20.89c0-21.266,0.803-42.545,0.803-63.811c0-51.935,0.1-103.231,0.1-155.5c0-61.321-0.564-126.933-0.564-196.333c0-40.794-0.146-81.792-0.247-122.724c1.397-0.356,2.512-1.308,2.512-2.921c0-5.649,0-11.298,0-16.946c0-13.561,0-27.122,0-40.683C434,199.459,432.84,198.369,431.397,197.821z M426.749,415.992c0,64.408,0,128.814,0,193.222c0,48.338-0.649,96.674-0.649,145.012c0,15.336,0.101,30.672,0.101,46.008c-0.26,23.879-3.4,41.316-24.601,55.616c-8.901,6.004-30.229,5.601-40.791,5.601c-19.322,0-38.646,1.055-57.969,1.055c-62.049,0-124.096,0-186.144,0c-15.671,0-31.342,0-47.013,0c-36.688,0-62.024-22.794-62.024-62.693c0-24.758,0-49.518,0-74.277c0-54,0.837-102.384,0-162.004c-0.922-65.673,0-131.361,0-197.042c0-42.408,0-84.816,0-127.224c0.12-0.368,0.2-0.758,0.2-1.18c0-5.443,0-10.886,0-16.329c0-13.767-0.2-109.977-0.2-143.71c0-11.816-0.076-22.92,4.161-34.364C19.983,21.632,37.005,7.028,64.271,4.35c4.67-0.458,9.54,0.1,14.226,0.1c54.59,0,109.28,0.25,163.87,0.25c28.965,0,57.931,0,86.896,0c13.42,0,26.992-0.594,40.402,0.137c23.887,1.302,44.435,16.801,53.059,38.674c3.056,7.75,2.947,16.009,2.947,24.276c0,9.295,0,18.59,0,27.885c0,43.781,0.157,78.398,0.251,102.895c-0.468,0.286-0.115,62.717,0.579,63.083C426.792,313.29,426.749,364.349,426.749,415.992z", "M410.054,82.574c-0.021-5.994-0.044-11.988-0.066-17.981c-0.037-10.231-0.726-22.981-8.412-30.629c-0.237-0.236-0.497-0.447-0.743-0.672c-6.89-8.817-17.813-13.019-29.409-13.187c-13.931,0.075-27.861,0.15-41.792,0.226c-0.164,0.001-0.335,0.026-0.506,0.065c-0.562,0.011-1.097,0.048-1.619,0.096c-3.056,0.018-6.124-0.229-9.167,0.053c-1.139,0.105-2.039,1.206-1.738,3.306c-0.101,2.7-0.267,6.358-0.267,8.213c0,4.417-1.062,8.761-3.148,12.341c-2.985,4.946-7.323,6.014-12.684,6.014c-6.988,0-13.976,0-20.963,0c-30.726,0-61.451,0-92.177,0c-16.23,0-32.46,0-48.691,0c-6.479,0-12.132-0.354-17.18-5.41c-4.918-4.926-4.11-11.073-4.669-17.297C116.7,26.347,116.4,24.549,115,22.85c-2-2.373-4.54-2.8-5.8-2.805c-16.734-0.059-26.067,0.071-42.8-0.096c-1.032-0.01-4.654,0.153-4.724,0.889c-11.958,1.74-22.345,8.024-29.172,17.733c-4.816,6.399-7.719,14.25-8.5,22.227c-0.354,3.62-0.479,6.259-0.503,9.952c-0.376,0.187-0.627,4.678-0.627,5.354c0.003,21.212,0.623,39.734,0.626,60.946c0.007,51.91-0.604,103.301-0.596,155.211c0.009,65.286,0.018,130.571,0.026,195.857c0.008,61.29,0.017,122.581,0.025,183.871c0.005,39.924,0.011,79.849,0.016,119.772c0.001,6.56,0.002,13.119,0.003,19.679c0,1.188,1.101,4.438,2.207,4.909c2.22,8.766,7.183,14.304,14.136,19.97c8.512,7.546,19.5,10.503,30.723,10.503c17.818,0,35.636,0,53.455,0c62.188,0,124.376,0,186.565,0c4.067,0,8.135,0,12.202,0c0.336,0.201,1.344,0.226,1.738,0.228c16.433,0.066,30.066-0.57,46.5-0.7c0.74-0.006,3.7,0.2,4.3-0.6c3.196-0.704,22.159-5.035,30.301-20.8c3.21-6.016,4.5-10.399,4.318-22.143c-0.392-25.332,1.072-50.67,1.072-76.005c0-7.837,0-15.675,0-23.512c0.069-12.397,0.14-24.794,0.209-37.191c0.358-63.719-0.356-127.437,0.002-191.155c0.163-29.095,0.587-58.202,0.479-87.298c-0.166-44.891-0.332-89.782-0.498-134.673C410.474,196.174,410.264,139.374,410.054,82.574z M404.581,68.749c0,11.315,0,22.629,0,33.944c0,46.415,0,92.83,0,139.245c0,63.733,0,127.466,0,191.199c0,63.776,0,127.553,0,191.329c0,18.031,0,36.063,0,54.095c-0.113,20.1-0.226,40.2-0.339,60.3c-0.104,18.611-0.209,37.223-0.314,55.834c-0.018,3.199-0.035,6.398-0.054,9.598c-0.044,7.769-2.331,15.181-6.252,21.503c-6.901,9.926-18.288,16.535-31.202,16.535c-42.429,0-84.856,0-127.285,0c-54.111,0-108.223,0-162.334,0c-5.005,0-10.103,0.238-15.062-0.522c-18.496-2.838-32.319-19.916-32.319-38.398c0-6.932,0-13.864,0-20.796c0-40.436,0-80.871,0-121.306c0-61.468,0-122.936,0-184.402c0-65.226,0-130.45,0-195.675c0-51.708-1.02-103.415-1.02-155.122c0-21.409,0.687-42.753,1.043-64.177c0.191-11.5,5.797-22.174,14.583-29.404c1.549-1.225,3.191-2.333,4.903-3.324c5.971-3.246,12.739-4.381,19.622-4.381c11.77,0,23.54,0,35.311,0c0.161,0.031,0.33,0.05,0.508,0.05c1.536,0,2.905-0.099,4.329,0.318c1.938,1.323,2.658,3.882,2.793,6.874c0.168,3.724,0.187,6.894,1.603,10.403c1.362,3.373,3.071,6.659,5.601,9.302c2.497,2.609,6.553,2.745,9.898,3.255c9.407,1.435,19.508,0.404,28.987,0.396c21.368-0.02,42.745,0.151,64.112-0.125c9.916-0.128,19.832-0.256,29.748-0.385c14.63,0,29.26,0,43.89,0c5.296,0,10.076-0.096,15.05-2.513c1.1-0.534,2.152-1.199,3.146-1.961c1.429-1.017,2.629-2.235,3.637-3.587c2.169-2.712,3.72-5.938,4.3-9.205c0.467-2.629,0.403-5.219,0.418-7.876c0.021-3.818,1.221-4.618,3.891-4.65c15.062-0.088,30.126-0.175,45.189-0.263c0.354-0.002,0.64-0.074,0.868-0.193c4.338,0.527,8.827,1.268,13.052,2.598c5.507,2.262,10.242,6.032,13.713,11.59c4.516,7.234,5.988,16.281,5.988,24.68C404.581,65.27,404.581,67.009,404.581,68.749z", "M187.865,37.61c2.181,1.006,5.174,0.768,7.5,0.803c5.292,0.082,10.582-0.257,15.868-0.431c6.398-0.211,12.798-0.422,19.196-0.633c4.864-0.16,11.897,1.883,13.157-4.357c0.04-0.198,0.055-0.384,0.052-0.562c0.014-0.198,0.01-0.402-0.034-0.618c-1.062-5.258-5.702-4.92-9.896-4.92c-6.634,0-13.269,0-19.902,0c-6.349,0-12.698,0-19.047,0c-4.378,0-9.331,0.959-8.863,6.645C186.058,35.508,185.786,36.651,187.865,37.61z M197.742,31.407c5.324,0,10.648,0,15.973,0c5.5,0,10.998,0,16.497,0c2.101,0,7.131-1.109,8.36,1.172c-1.502,2.304-8.778,1.28-10.863,1.349c-5.177,0.171-10.354,0.341-15.53,0.512c-3.237,0.106-6.473,0.256-9.71,0.362c-2.944,0.095-5.89,0.139-8.834,0.059c-0.896-0.024-1.782-0.092-2.668-0.213c-0.036-0.431-0.071-0.862-0.106-1.292C190.604,30.257,195.819,31.407,197.742,31.407z", "M264.078,39.22c4.203,0.872,7.949-2.65,8.176-6.851c-0.238-4.431-4.11-7.834-8.023-9.368c-4.633-1.817-9.039,1.961-9.285,6.53c-0.008,0.138,0.004,0.273,0.006,0.41c-0.277,0.726-0.452,1.525-0.499,2.396C254.141,38.122,259.439,40.262,264.078,39.22z M262.761,26.437c0.06-0.031,0.123-0.052,0.184-0.08c0.066,0.008,0.133,0.008,0.2,0.019c2.73,0.444,4.316,3.041,4.649,5.571c-0.632,1.502-1.772,2.696-3.356,3.516c-1.347,0.698-3.028,0.643-4.442,0.254c-1.01-0.278-1.097-2.524-1.142-3.349C258.714,29.765,260.592,27.561,262.761,26.437z"]
  },
  "iphone_top_bar_items_color": 0x000000,
  "iphone4_scale9": {
    type: "SVG_SCALE9",
    scale9: [87, 161, 94, 90],
    width: 282,
    height: 510
  },
  "iphone4_scale9_nw": {
    type: "SVG_PATHS",
    width: 161,
    height: 87,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101],
    paths: ["M25.715,87H4.098c0-0.261,0-0.515,0-0.732c1.043-35.139-5.403-71.507,26.01-80.519c12.961-3.718,36.673-0.02,54.728-1.285C110.037,2.697,136.173,3.107,161,3.645v76.637c-45.562,0.274-92.572,1.949-135.302,2.961C25.706,84.491,25.708,85.75,25.715,87z", "M0.362,67.547c0.242,2.225,2.094,1.574,3.907,2.06c1.569,0.42,0.772-6.653,0.874-9.851C5.691,42.41,7.474,41.998,1.62,41.998C-0.559,43.309-0.045,63.805,0.362,67.547z", "M27.026,87h-3.928c-0.034-2.334-0.055-4.664-0.099-7c1.873-0.222,3.627-2.578,5,0c0.333,0,0.667,0,1,0c41.52-1.82,88.787-0.825,132-1.098v4.832c-51.727,0.132-133.963-1.013-133.963,0.627C27.037,85.079,27.032,86.012,27.026,87z", "M5.889,87H1.382c-0.361-7.635-0.704-23.47,1.441-39.208c0.62-4.55,1.398-8.75,2.328-12.608c0.236-4.156,0.767-8.061,1.924-10.627c1.498-1.503,2.762-7.754,4.973-8.543c0-0.137,0.836-0.893,1.822-1.76c3.698-5.168,8.178-8.511,13.426-9.994c8.497-2.402,22.223-2.452,35.496-2.5c7.49-0.027,15.235-0.055,21.766-0.513c9.365-0.657,19.458-0.976,30.855-0.976c14.746,0,29.525,0.533,43.818,1.05c0.597,0.021,1.175,0.042,1.769,0.063v4.002c-0.641-0.023-1.268-0.045-1.913-0.068c-1.77-0.063-3.566-0.125-5.352-0.188c-5.548-0.026-11.139-0.077-16.767-0.185c-13.531-0.26-29.981-0.307-45.982-0.081c-2.085,0.11-4.138,0.231-6.148,0.373c-6.663,0.468-14.475,0.495-22.03,0.522c-1.195,0.004-2.393,0.01-3.59,0.016C59.143,5.848,59.063,5.916,59,6c-1,0-2,0-3,0c-0.175,0.01-0.357-0.011-0.541-0.047c1.157-0.061,2.342-0.12,3.545-0.177c-1.364,0.007-2.726,0.017-4.08,0.031c0.179,0.062,0.357,0.112,0.535,0.146C45.997,6.459,38.141,7.133,33.157,8c-6.044,1.827-10.482,3.667-13.834,5.943c-6.048,6.622-9.486,17.34-11.437,27.621c-0.199,4.053-0.483,9.725-0.483,9.943C5.999,62.556,5.915,74.779,5.889,87z"]
  },
  "iphone4_scale9_n": {
    type: "SVG_PATHS",
    width: 27,
    height: 87,
    colors: [0xc1c0c0, 0x010101, 0x010101],
    paths: ["M0,80.281V3.645C9.218,3.844,18.261,4.06,27,4.189v76.105C18.09,80.222,9.071,80.227,0,80.281z", "M0,83.734v-4.832c9.213-0.058,18.252-0.17,27-0.383v4.938C19.932,83.627,10.605,83.707,0,83.734z", "M0,5.387V1.385c9.352,0.337,18.299,0.651,27,0.803v4C18.287,6.036,9.344,5.724,0,5.387z"]
  },
  "iphone4_scale9_ne": {
    type: "SVG_PATHS",
    width: 94,
    height: 87,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101],
    paths: ["M93.246,87H67.617c-0.002-1.438,0.002-2.879-0.001-4.315C46.355,81.182,23.535,80.488,0,80.295V4.189c1.234,0.018,2.476,0.038,3.698,0.053c24.371,0.289,54.25-1.282,74.184,7.898C93.381,26.568,94.502,55.188,93.246,87z", "M3.642-0.281C0.39-0.174,1.345,1.812,0.638,3.626C0.026,5.195,10.343,4.398,15.005,4.5C40.3,5.048,40.901,6.83,40.901,0.977C38.36-0.644,8.933-0.456,3.642-0.281z", "M68.021,87H64.84c-0.307-2.046-0.906-3.519-2.027-3.436c-8.859,0.652-41.771,0.266-44.771,0.266c-0.668,0-14.67-0.398-17.004-0.398C0.702,83.44,0.347,83.448,0,83.457V78.52c5.448-0.133,10.792-0.302,15.999-0.52c0.334,0,0.668,0,1,0c1.668,0,3.334,0,5,0c0.334,0,0.668,0,1,0c5.018,0.65,12.645-1.311,16,1c0.334,0,0.668,0,1,0c8.201,0.699,21.527-3.065,28,2C68,82.98,68.013,84.997,68.021,87z", "M93.522,87h-4.541c-0.097-7.898-0.162-9.161-0.92-24.077L87.696,57.4c-0.332-10.016-0.27-9.123-0.27-10.577c-0.145-7.06-1.195-13.546-4.559-22.427c-2.725-9.514-8.68-13.229-17.061-15.036c-1.359-0.132-2.895-0.284-3.891-0.692c-5.223-0.762-11.209-1.084-17.811-1.699c-5.062-0.471-10.244-0.814-15.521-1.072c-2.965,0.04-5.92,0.102-8.811,0.166c-5.641,0.124-10.953,0.239-16.1,0.18C2.454,6.228,1.229,6.209,0,6.188v-4c1.245,0.021,2.486,0.04,3.722,0.055c5.096,0.061,10.359-0.056,15.965-0.18c4.631-0.102,9.42-0.207,14.178-0.207c15.299,0,32.732,0.996,46.021,8.387C91.557,17.169,93.531,49.846,93.522,87z"]
  },
  "iphone4_scale9_e": {
    type: "SVG_PATHS",
    width: 94,
    height: 333,
    colors: [0xc1c0c0, 0x010101, 0x010101],
    paths: ["M91.001,56.242c0,79.56,1.039,185.692-0.052,276.758H64.704c0.022-114.728,3.047-224.788,2.913-333h25.628C92.519,18.389,91.001,37.841,91.001,56.242z", "M68.999,169c0,1,0,2,0,3c0.26,4.59,1.064,4.832,0,8.999c0,0.334,0,0.668,0,1c-0.637,24.365,1.312,61.978-1,88c1.46,20.016,1.61,49.076,1.292,63.001h-3.935c1.067-22.978-0.356-91.111-0.356-93c0-6.332,0.668-19,1-19l-0.254-44.178c0-0.332,0.916-19.816,0.254-28.822c0-0.334,0-60.334,0-61c-0.779-0.221-1.162-46.838-1-48c0-9.334,0-18.666,0-28c0-1.717,0.397-7.301-0.159-11h3.181C68.24,53.09,70.734,113.818,68.999,169z", "M93.251,29.396c-0.129,9.446-0.25,18.369-0.25,26.846c0,26.602,0.096,55.029,0.197,85.124c0.199,59.098,0.402,119.844-0.08,174.766c0.043,0.689,0.053,1.344-0.016,1.922c-0.046,5.036-0.099,10.015-0.157,14.946h-4.002c0.056-4.752,0.108-9.536,0.153-14.386c-0.127-0.367-0.273-0.654-0.453-0.826c-1.225-23.518,0.957-49.646,0-72.88c0.367-0.744,0.613-1.498,0.771-2.262c0-1.748,0-3.498,0-5.25c-0.215-1.529-0.541-3.078-0.783-4.635c0-3.373,0-6.748,0-10.123c0-7.76,0-15.521,0-23.28c-1.6-5.824,0.334-97.96,0.334-99.31c0-20.582-0.953-39.417-0.953-60c0-12.146,1-25.312,1-37.459C89.001,1.646,88.992,0.791,88.982,0h4.541C93.52,9.622,93.385,19.543,93.251,29.396z"]
  },
  "iphone4_scale9_se": {
    type: "SVG_PATHS",
    width: 94,
    height: 90,
    colors: [0xc1c0c0, 0x010101, 0x010101],
    paths: ["M90.949,0c-0.074,6.153-0.157,12.238-0.251,18.241c-0.203,12.895-0.859,36.508-5.018,45.361c-7.531,16.031-16.072,18.637-37.982,21.639C34.039,87.113,17.353,87.143,0,86.801V7.033C21.998,6.65,43.727,6.272,64.698,6.149C64.697,4.094,64.704,2.052,64.704,0H90.949z", "M69.291,0c-0.075,3.271-0.175,5.713-0.292,6.999C46.006,6.878,23.012,7.303,0,7.828V3.23c22.84-0.136,46.131-0.231,61.999-0.231c0-0.332,3.088,1.258,3.141,0.592C65.221,2.57,65.292,1.359,65.355,0H69.291z", "M92.946,0c-0.073,6.18-0.154,12.281-0.248,18.272c-0.199,12.648-2.178,41.775-7.027,51.082c-5.498,10.551-17.098,14.1-29.625,16.229c-1.627,0.527-3.291,1.018-5.047,1.416c0,0.143-4.781,0.441-10.4,0.754c-6.965,0.744-15.051,1.115-24.6,1.115c-5.201,0-10.583-0.108-15.999-0.262v-5.121c8.744-0.2,14.416,0.178,24.983,0.49c0.338,0,10.676,0,11.012,0c0.99-1.115,2.51-0.781,4.232-0.205c1.441-0.156,2.85-0.32,4.191-0.51l1.725-0.24c14.789-2.062,30.08-4.193,35.98-15.514c1.188-2.281,2.211-6.133,3.072-10.783c0.201-1.609,0.344-3.318,0.41-5.164c0-0.338,3.025-35.055,3.025-40.453c0-0.971,0.098-2.57,0.232-4.506C88.893,4.419,88.917,2.205,88.943,0H92.946z"]
  },
  "iphone4_scale9_s": {
    type: "SVG_PATHS",
    width: 27,
    height: 90,
    colors: [0xc1c0c0, 0x010101, 0x010101],
    paths: ["M0,86.181V7.484c9.022-0.141,18.034-0.296,27-0.451v79.768C18.069,86.625,8.964,86.353,0,86.181z", "M0,8.465V3.413C8.579,3.348,17.75,3.286,27,3.23v4.598C18.003,8.034,9.003,8.254,0,8.465z", "M0,88.048v-5.291c5.344,0.077,10.628,0.524,14.433,1.23c0.338,0,0.676,0,1.014,0c4.565-0.271,8.274-0.428,11.553-0.502v5.121c-1.208-0.035-2.416-0.069-3.626-0.107c-1.154-0.006-2.012-0.01-2.428-0.01C13.948,88.489,6.987,88.159,0,88.048z"]
  },
  "iphone4_scale9_sw": {
    type: "SVG_PATHS",
    width: 161,
    height: 90,
    colors: [0xc1c0c0, 0x010101, 0x010101],
    paths: ["M161,7.484v78.696c-18.695-0.359-36.767-0.278-51.302,2.061c-7.667,0-15.333,0-23,0c-25.613-2.186-56.52,3.199-71.009-13.059C2.75,60.664,4.767,27.493,4.141,0h22.443c0.04,2.047,0.071,4.107,0.114,6.149C68.901,8.496,115.106,8.197,161,7.484z", "M114,3.999c0-0.156,20.695-0.386,47-0.586v5.052C127.07,9.26,93.087,9.905,59,8.999c-12.085-0.32-22.6,0.363-34-3C24.883,3.9,24.806,1.917,24.754,0h4.057C28.874,0.333,28.93,0.67,29,0.999c9.19,1.811,18.888,3.113,30,3c9,0,18,0,27,0C95.333,3.999,104.667,3.999,114,3.999z", "M161,82.757v5.291c-1.333-0.021-2.666-0.049-4.001-0.049c-2.332,0-4.666,0-7,0c-2,0-4,0-6,0c-1,0-2,0-3,0c-0.21,0.324-9.79,0.324-10,0c-1,0-2,0-3,0c0.168,2.17-3.547-0.547-5,0c-0.532,0-1.064,0-1.597,0c-4.044,0.32-7.745,0.75-11.189,1.305l-0.318,0.025l-23.154-0.006c-1.733-0.148-3.558-0.244-5.485-0.299c-4.668,0.137-9.349,0.215-12.99,0.178c-1.856,0.08-3.706,0.168-5.529,0.26c-4.889,0.244-9.507,0.475-14.06,0.475c-11.147,0-23.667-1.26-32.006-10.615C2.788,63.743,0.979,59.62,1.92,0h4.001C5.696,14.28,5.633,25.3,5.838,33.983C6.2,40.897,6.638,46.294,7.165,49.739c0.271,3.387,0.731,6.672,1.461,9.785c2.118,7.129,5.588,11.027,11.032,17.135c1.28,1.436,2.695,2.643,4.204,3.676c1.412,0.693,2.931,1.309,4.58,1.826c0.22,0,3.891,2.572,18.003,3.754c0.749,0.014,1.494,0.021,2.232,0.021c4.453,0,9.022-0.229,13.86-0.469c4.915-0.246,9.998-0.5,15.03-0.5c3.453,0,6.538,0.117,9.429,0.361h4.862c8.158-0.381,15.962-0.643,24.543-0.883c7.115-0.77,15.275-1.133,25.109-1.133c2.317,0,4.671,0.021,7.042,0.059C152.228,82.877,156.634,82.694,161,82.757z"]
  },
  "iphone4_scale9_w": {
    type: "SVG_PATHS",
    width: 161,
    height: 333,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M26.584,333H4.141c-0.037-1.63-0.082-3.242-0.141-4.827C0.529,235.484,3.379,101.066,4.098,2.854c0-0.782,0-1.915,0-2.854h21.617C26.303,109.798,24.475,224.683,26.584,333z", "M0.362,24.805c0.242,0.956,2.094,0.677,3.907,0.885c1.569,0.181,0.772-2.857,0.874-4.231c0.548-7.451,2.33-7.628-3.523-7.628C-0.559,14.393-0.045,23.197,0.362,24.805z", "M1.62,71.697c5.854,0,4.071-0.177,3.523-7.628c-0.102-1.374,0.695-4.412-0.874-4.231c-1.813,0.208-3.665-0.071-3.907,0.885C-0.045,62.33-0.559,71.135,1.62,71.697z", "M28.812,333h-4.057c-0.22-8.257,0.05-15.286-0.203-24.978c-0.001-0.154-0.007-0.414-0.02-0.801c0.008,0.273,0.013,0.533,0.02,0.801c0.012,1.42-0.397-6.33-0.343-13.756C24.593,241.301,24.435,183.834,24,130c-0.008-1,0-2,0-3c-1.09-42.105-0.273-84.402-0.901-127h3.928C27.015,1.766,27,3.718,27,5c0,1,0,2,0,3c0,1.666,0,3.334,0,5c0,1.665,0,3.33,0,4.995c-0.04-0.001-0.085-0.003-0.125-0.005c0-0.001,0-0.003,0-0.004c-0.135,0-0.084,0.002,0,0.004c0,1.999,0,3.997,0,5.996C26.542,23.986,27.333,24,27,24c-0.609,2.391-0.645,18.645,0,23c0,4,0,8,0,12c0,4.666,0,9.334,0,14c0.333,0,1,39.334,1,59c0,1,0,2,0,3c0,14.666,0,29.334,0,43.999c0,0.669,0,1.335,0,2c0,33.335,0,66.668,0,100c0,2.334,0,4.668,0,7c0,2.334,0,4.668,0,7c0,1.668,0,3.334,0,5C28.533,311.4,26.946,323.054,28.812,333z", "M27,18c0-0.002,0-0.004,0-0.005C27.084,17.998,27.135,18,27,18z", "M6.196,118.936C6.663,191.646,7.015,269.376,6,328.208c-0.028,1.642-0.054,3.23-0.079,4.792H1.92c0.025-1.584,0.051-3.197,0.08-4.86c0-1.381,0-2.76,0-4.141c-0.333,0-0.333-4,0-4c0-0.666,0-1.332,0-2c0.066-0.447,0.118-0.91,0.177-1.363c0.23-16.652,0.365-34.596,0.429-53.346C2.326,262.526,2,261.763,2,260.999c0-6.666,0-13.332,0-19.999c0-35.667,0-71.333,0-107c0-0.666,0-1.334,0-2c0-4.334,0-8.666,0-13c0.065,0,0.131,0,0.196,0c-0.021-3.346-0.043-6.678-0.065-10c-0.043,0-0.087,0-0.131,0C1.728,98.821,1.456,89.534,1.929,78.701C1.748,51.605,1.584,25.929,1.5,2.249L1.478,1.867C1.447,1.318,1.414,0.689,1.382,0h4.507c-0.017,7.71-0.011,15.419-0.3,22.83C5.64,32.91,5.701,43.299,5.768,53.948c0.342,4.548,0.618,9.247,0.635,12.808C6.403,67.639,7.596,97.855,6.196,118.936z"]
  },
  "iphone4_scale9_c": {
    type: "SVG_PATHS",
    width: 27,
    height: 333,
    paths: []
  },
  "iphone4_button": {
    type: "SVG_PATHS",
    width: 52,
    height: 52,
    colors: [0x010101, 0x010101],
    paths: ["M17.963,3.668c-0.333,0-0.667,0-1,0c-6.362,3.638-11.098,8.902-13,17c0,0.333,0,0.667,0,1c-6.437,14.771,6.681,30.839,20,29c0.333,0,0.667,0,1,0C64.543,58.329,52.549-14.43,17.963,3.668zM45.56,31.404c-0.657,7.044-5.753,14.482-11.63,15.858C-4.136,56.175-1.634,0.504,30.757,6.028C39.085,7.45,42.627,13.2,45.56,22.945C46.02,26.513,45.56,30.698,45.56,31.404z", "M14.002,16.758c-0.407,7.074,1.78,11.554,1,19c4.16,4.969,17.2,3.144,23,1c0.229-7.229-0.373-13.628-2-19C27.825,16.029,20.576,15.969,14.002,16.758zM18.002,32.758c0-0.333,0-0.667,0-1c-0.074-4.26-1.673-6.994-1-12c4.774-2.043,10.196,1.052,17,0c0.662,4.671,1.572,9.095,2,14C31.756,36.646,20.536,38.792,18.002,32.758z"]
  },
  "iphone4_earthing": {
    type: "SVG_PATHS",
    width: 84,
    height: 24,
    colors: [0x010101, 0x010101],
    paths: ["M44.72,20.263c18.578,0.699,36.074,1.447,38-11.366c-2.805-7.026-37.151-5.976-49-4.06c-5.431,0.879-10.983,5.519-9,10.555C26.27,19.326,37.887,20.005,44.72,20.263zM74.886,6.914c2.844,0.501,3.268,2.306,3.268,2.681c-0.248,8.734-18.256,7.534-32.348,7.084c-0.375,0-17.562,0.796-17.854-6.401C29.968,6.042,72.462,6.688,74.886,6.914z", "M1.367,14.887c-0.451-3.874,0.965-7.249,2.043-8.118c3.423-2.758,10.49-2.283,11.127,3.123c-0.438,9.855-4.411,9.281-8.629,8.742C4.357,18.437,1.719,17.913,1.367,14.887z"]
  },
  "iphone5_scale9": {
    type: "SVG_SCALE9",
    scale9: [204, 62, 97, 88],
    width: 282,
    height: 573
  },
  "iphone5_scale9_nw": {
    type: "SVG_PATHS",
    width: 62,
    height: 204,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M25.756,204H3.07c0.38-42.789,0.806-82.272,1.028-112.646c0-1,0-2.586,0-3.586c1.043-35.139-6.403-72.507,25.01-81.519C37.064,3.967,49.447,4.479,62,4.922v77.009c-12.325,0.253-24.469,0.531-36.302,0.812C25.905,117.914,25.854,159.729,25.756,204z", "M27,104.743c0-0.002,0-0.004,0-0.005C27.084,104.741,27.135,104.743,27,104.743z", "M26.875,104.729c0,0.002,0,0.004,0,0.005C26.791,104.73,26.741,104.729,26.875,104.729z", "M27.025,204h-3.342C21.793,160.793,23.8,122.497,23,79.743c1.873-0.222,3.627-2.579,5,0c0.333,0,0.667,0,1,0c10.702-0.47,21.761-0.668,33-0.715v4.397c-20.605,0.15-34.963,0.303-34.963,0.303C26.679,85.696,27,88.743,27,90.743c0,1,0,2,0,3c0,1.667,0,3.333,0,5c0,1.665,0,4.327,0,5.995c-0.04-0.002-0.085-0.003-0.125-0.005c0-0.001,0-0.003,0-0.005c-0.135,0-0.084,0.002,0,0.005c0,1.999,0,3.997,0,5.995c-0.333,0,0.458,0.015,0.125,0.015c-0.609,2.392-0.645,18.646,0,23c0,4,0,8,0,12c0,4.667,0,6.333,0,11C27.275,156.743,27.096,183.578,27.025,204z", "M6.109,204H2c0-0.419,0-0.838,0-1.257c0.333,0,0.667,0,1,0c0-1,0-2,0-3v-7c-0.333,0-0.667,0-1,0c-0.287-10.714-0.583-20.417,0-32c0-2,0-1,0-3c1.036,1.944-1.031-19.284-1-19c0.435,1.417,1.435-7.417,1-6c-2.33-11.671,1.664-27.664,1-41c0-0.333,0-0.667,0-1c-0.333,0-0.667,0-1,0c0-1.667,0-3.333,0-5c0.333,0,0.667,0,1,0c0-1,0-2,0-3c0.609-2.392-2.255-7.255,0-8c0-0.333,0-0.667,0-1c-0.328-0.534-0.328-13.466,0-14c0-1,0-2,0-3c0.042,0,0.083,0,0.094,0c-0.01,0-0.052,0-0.094,0c1.097-2.903-1.014-7.274,0.083-10.177c0-0.334,1.917-0.49,1.917-0.823C4.863,32.888,2.294,34.904,7.075,24.299c1.498-1.502,2.762-7.754,4.973-8.543c0-0.333,4.007-5.379,4.007-5.712c3.788-4.546,4.271-3.298,10.785-5.116c0.5-0.145,20.805-0.18,35.159-0.455v0.174C49.156,5.011,38.365,5.663,32.157,6.743C13.608,12.35,9.114,19.05,6.077,37.27c0,1.055-0.674,12.926-0.674,13.981c-2.935,23.083,1.913,51.319-1.024,73.905c0,0.337,1.977,18.605,2.024,28.342C6.403,154.398,7.641,183.411,6.109,204z", "M0.362,106.047c0.242,2.225,2.094,1.574,3.907,2.06c1.569,0.42,0.772-6.653,0.874-9.851C5.691,80.91,7.474,80.498,1.62,80.498C-0.559,81.809-0.045,102.305,0.362,106.047z", "M0.362,153.682c0.242,1.526,2.094,1.08,3.907,1.413c1.569,0.288,0.772-4.563,0.874-6.756c0.548-11.896,2.33-12.179-3.523-12.179C-0.559,137.059-0.045,151.115,0.362,153.682z", "M1.62,201.512c5.854,0,4.071-0.283,3.523-12.184c-0.102-2.193,0.695-7.047-0.874-6.76c-1.813,0.334-3.665-0.112-3.907,1.414C-0.045,186.549-0.559,200.611,1.62,201.512z", "M5.699,204h-4c-0.029-4.092-0.057-8.152-0.085-12.168C1.348,154.005,1.098,118.275,1,90.749v-3.007c0.19-6.429,0.091-12.97-0.004-19.296C0.555,39.187,0.139,11.55,27.29,3.761c6.194-1.776,15.039-2.004,22.052-2.004c3.887,0,7.993,0.08,11.964,0.158C61.536,1.92,61.77,1.924,62,1.929v4c-0.257-0.005-0.517-0.01-0.773-0.015c-3.952-0.077-8.038-0.157-11.886-0.157c-6.751,0-15.234,0.21-20.949,1.85C4.185,14.551,4.561,39.499,4.996,68.386c0.096,6.37,0.195,12.957,0.003,19.416L5,90.742c0.098,27.512,0.348,63.238,0.614,101.062C5.642,195.83,5.67,199.898,5.699,204z"]
  },
  "iphone5_scale9_n": {
    type: "SVG_PATHS",
    width: 123,
    height: 204,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101],
    paths: ["M0,81.931V4.922c7.898,0.279,15.863,0.531,22.836,0.042C56.675,2.592,91.331,5.033,123,5.65v75.656C82.623,80.576,40.303,81.105,0,81.931z", "M0,83.426v-4.397c40.944-0.169,84.42,1.808,123,1.211v3.251C92.217,82.9,36.371,83.16,0,83.426z", "M0,4.646V4.473c7.896-0.15,13.993-0.375,14-0.729c0.021-0.972,17.333,0,20,0c3.367-0.634,7.208-1.196,11-1c12.618-0.401,26.102,0.638,36,0c1.333,0,2.667,0,4,0c3.914-0.989,5.807,2.21,8,0c4.019,2.315,12.319,0.347,18,1c0.839,0,5.439-0.133,12-0.222v1.41c-15.44-0.079-31.468,0.073-48.032-0.244C53.096,4.27,23.19,3.987,0,4.646z", "M0,5.929v-4c3.765,0.072,7.629,0.145,11.309,0.145c4.439,0,8.119-0.106,11.25-0.326c9.366-0.656,19.46-0.976,30.859-0.976c14.746,0,29.527,0.533,43.821,1.05C106.162,2.144,114.693,2.45,123,2.627v4.001c-8.361-0.178-16.936-0.485-25.906-0.81C82.835,5.304,68.091,4.771,53.417,4.771c-11.305,0-21.307,0.315-30.58,0.966c-3.225,0.227-6.996,0.336-11.529,0.336C7.616,6.073,3.76,6.002,0,5.929z"]
  },
  "iphone5_scale9_ne": {
    type: "SVG_PATHS",
    width: 97,
    height: 204,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M92.108,204h-23.09c0.4-43.68,0.67-84.842,0.599-119.815C47.683,82.635,24.191,81.745,0,81.307V5.65C1.91,5.688,3.812,5.72,5.698,5.742c24.371,0.289,55.25-2.282,75.184,6.898c18.45,19.958,11.119,81.865,11.119,132.102C92.001,161.361,92.046,181.561,92.108,204z", "M72.702,204H68c0-15.945,0-32.904,0-33.257c-0.779-0.221-0.163-43.837,0-45c0-9.333,0-19.667,0-29c0-2.667,0.938-13.015-2.186-12.558c-0.667,0-31.687-0.816-37.687-0.816c-3,0-6.128,0.374-9.128,0.374c-0.667,0-1.333,0-2,0c-2.333,0-4.668,0.078-7,0c-2.91-0.098-6.278-0.181-10-0.253v-3.251c6.132-0.095,12.146-0.251,18-0.496c0.333,0,0.667,0,1,0c1.667,0,3.333,0,5,0c0.333,0,0.667,0,1,0c5.017,0.649,13.644-2.311,17,0c0.333,0,0.667,0,1,0c8.2,0.699,21.526-3.065,28,2C71.028,121.282,72.613,161.721,72.702,204z", "M93.198,204h-2.072c-0.103-12.125-0.16-19.904-0.16-20.208c0-20.582-0.952-36.417-0.952-56.999c0-12.147,1-25.314,1-37.461c-0.134-11.338-0.131-11.338-0.317-32.188c0,0-1.27-10.239-1.27-10.577c-0.144-7.061-0.058-13.684-3.422-22.564C81.31,7.605,65.897,8.557,46.106,6.713C31.419,5.344,16.026,5.014,0,4.932v-1.41c16.889-0.228,46.778-0.162,59,3.222l19.992,3.981c2.853,2.318,5.262,3.393,6.625,6.355c4.818,9.182,2.729,19.309,6.552,29.486c0.083,1.584-0.91,7.629,0.57,7.814c-0.109,2.488-1.612,3.109,0.262,5.014c0.547,1.453-2.169,3.408,0,3.24c0,0.666,0,0,0,0.666c0.619,2.714-1.751,8.143,0.519,9.206c0,3.333-1.52,6.903-1.52,12.236c0.333,0,0.667,0,1,0c0,1.667,0,3.333,0,5c0.333,0,0.667,0,1,0c-0.547,1.453,1.169,5.169-1,5c-0.409,2.409,0.826,2.174,0,5c-2.169-0.169-0.453,3.547-1,5c0.333,0,0.667,0,1,0c0,3.333,0,6.667,0,10c-0.333,0-0.667,0-1,0c0,3,0,6,0,9c0.333,0,0.667,0,1,0c0,0.333,0,0.667,0,1c0,4,0,8,0,12c1.263,2.777-0.498,3.36,0,8c0.324,0.337,0.324,11.663,0,12c0-0.188,0-0.27,0-0.286c0,0.017,0,0.099,0,0.286c0,0.156,0,0.312,0,0.286c0,0.026,0-0.13,0-0.286c0.316,0.085,0.316,7.915,0,8c-0.58,11.874,1.252,17.654,0,26c0.301-0.083,0.301,5.083,0,5c-0.409-0.39,0.011,5.921,0,6C93.093,201.821,93.159,202.733,93.198,204z", "M71,9.743l-1.1-0.366C70.298,9.467,70.679,9.576,71,9.743z", "M63,7.743l6.9,1.634C68.126,8.973,63.825,9.096,63,7.743z", "M6.643,0.219C3.391,0.326,4.346,2.312,3.639,4.126C3.026,5.695,13.343,4.898,18.005,5C43.3,5.548,43.901,7.33,43.901,1.477C41.361-0.144,11.933,0.044,6.643,0.219z", "M95.208,204h-4c-0.123-22.377-0.207-42.712-0.207-59.258c0-10.887,0.294-22.537,0.605-34.871c1.049-41.537,2.238-88.617-9.627-95.661C69.536,7.292,52.37,6.355,37.226,6.355c-4.825,0-9.686,0.105-14.387,0.208C17.154,6.687,11.805,6.799,6.675,6.742C4.464,6.716,2.238,6.676,0,6.628V2.627c2.255,0.048,4.497,0.089,6.722,0.115c5.074,0.063,10.38-0.055,16.03-0.178c4.723-0.103,9.605-0.209,14.474-0.209c15.657,0,33.455,0.996,46.746,8.387c13.931,8.269,12.815,52.452,11.634,99.23c-0.311,12.308-0.604,23.933-0.604,34.77C95.001,161.285,95.085,181.62,95.208,204z"]
  },
  "iphone5_scale9_e": {
    type: "SVG_PATHS",
    width: 97,
    height: 281,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101],
    paths: ["M92.108,0c0.235,85.43,0.709,203.435-0.173,281H66.697c-0.015-80.883,1.458-186.801,2.321-281H92.108z", "M72,48.743c0,1,0,2,0,3c0.259,4.589,1.063,4.831,0,9c0,0.333,0,0.667,0,1c-0.638,24.364,1.312,124.977-1,151c0-0.167,0-0.208,0-0.188c0-0.021,0,0.021,0,0.188c0,0.083,0,0.167,0,0.188c0-0.021,0-0.104,0-0.188c0.576,29.086,1.448,48.729,1.172,68.257h-3.998C67.909,258.03,68,184.642,68,182.743c0-6.333,0.667-19,1-19L67.745,56.565c0-0.333,0.916-19.816,0.255-28.822C68,27.586,68,14.194,68,0h4.702C72.735,15.97,72.562,32.193,72,48.743z", "M94,188.743c0,0.667,0,1.333,0,2s0,1.333,0,2c0,15.667,0,31.333,0,47c0,3.333,0,6.667,0,10c-0.083,0-0.104,0-0.094,0c-0.01,0,0.011,0,0.094,0c0,0.696,0,16.376,0,31.257h-3.132c1.505-44.702,4.138-125.61,2.458-127.222C92.374,135.521,91.48,41.827,91.126,0h2.072c0.1,3.163,0.033,8.547-0.198,8.743c-0.31,11.47,0.731,19.215,0,29c0,4,0,8,0,12c2.255,2.162-0.828,3.411,0,7c2.305,2.695,0.354,9.645,1,14c0,1.333,0,65.667,0,67C97.237,137.743,96.571,188.743,94,188.743z", "M95.52,51.433c0.498,78.664,1.055,166.87,0.406,229.567h-4c0.648-62.663,0.092-150.876-0.406-229.542C91.407,33.665,91.297,16.292,91.208,0h4C95.298,16.284,95.407,33.648,95.52,51.433z"]
  },
  "iphone5_scale9_se": {
    type: "SVG_PATHS",
    width: 97,
    height: 88,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101],
    paths: ["M91.936,0c-0.07,6.181-0.149,12.112-0.237,17.742c-0.202,12.895-0.016,37.995-4.536,46.67c-6.615,12.694-24.617,18.104-40.464,20.33C32.953,86.673,16.741,86.548,0,86.01V4.571c22.629-0.396,45.024-0.794,66.698-0.922C66.697,2.441,66.697,1.219,66.697,0H91.936z", "M72.172,0C72.135,2.576,72.08,5.149,72,7.743C48.159,6.708,24.131,6.804,0,7.251v-3.82C23.784,3.41,48.436,3.41,65,3.41c0-0.333,3.231,0.487,3.22-0.18C68.204,2.312,68.188,1.224,68.174,0H72.172z", "M94,0c0,13.628,0,26.584,0,26.743c0,1.333-2.829,21.504-3,23c-0.074,6.259-4.109,9.558-5,15c-2.247,3.087-3.5,7.675-6.587,9.922C71.642,79.561,63.321,82.397,53,84.743c0,0.333-26.235,1.521-28.043,1.756C24.498,86.413,7.393,86.285,0,86.246V84.06c7.625,0,18.538-0.131,27.943-0.752c0.336-0.022,20.875-0.602,21.212-0.602c23.233-6.627,33.659-3.487,38.586-31.981c0-0.337,1.128-17.558,2.193-21.933c0-1.423,0.395-12.79,0.933-28.792H94z", "M95.926,0c-0.064,6.181-0.139,12.123-0.228,17.773l-0.004,0.265c-0.281,17.952-0.571,36.516-8.023,50.816C80.576,82.47,64.339,85.495,47.849,86.736C35.493,87.666,20.273,88.103,0,88.109v-4c20.17-0.007,35.294-0.439,47.548-1.361c15.338-1.154,30.385-3.861,36.576-15.741c7.015-13.462,7.297-31.544,7.57-49.031l0.004-0.265C91.787,12.08,91.861,6.158,91.925,0H95.926z"]
  },
  "iphone5_scale9_s": {
    type: "SVG_PATHS",
    width: 123,
    height: 88,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101],
    paths: ["M0,85.25V5.019c39.844,0.971,81.808,0.272,123-0.447V86.01c-26.292-0.846-53.89-2.714-75.302,0.732c-7.667,0-15.333-1-23-1C16.759,85.064,8.311,85.211,0,85.25z", "M0,8.811V3.743c8,0,16,0,24,0c9.333,0,18.667,0,28,0c0-0.202,34.477-0.281,71-0.312v3.82C82.163,8.006,41.03,9.765,0,8.811z", "M0,86.237v-3.075c4.983,0.231,10.063,0.438,15.247,0.839c0.675,0,1.35,0,2.024,0c18.765-1.121,33.267-0.469,54.661-1.012c1.012,0,2.024,0,3.036,0c2.362,0,4.724,1,7.085,1c0.338,0,23.404-1.481,31.38,0c0.143,0,3.971,0.07,9.566,0.07v2.187c-1.465-0.008-2.558-0.013-3.054-0.013c-8.333,0-16.613-0.49-24.946-0.49c-2.333,0-4.667,0-7,0c-2,0-4,0-6,0c-1,0-2-1-3-1c-0.211,0.323-9.79,0.323-10,0c-1,0-2,0-3,0c0.168,2.169-3.547-0.547-5,0c-4,0-8,0-12,0c-0.333,0-0.667,0-1,0c2.862,0-27.13,4.719-23.302,0.999c-1,0-17.059,1.336-23.736,0.68C0.807,86.363,0.468,86.302,0,86.237z", "M0,87.665v-4.001c1.039-0.032,2.076-0.066,3.104-0.101c4.732-0.157,9.625-0.319,14.381-0.319c2.855,0,5.407,0.058,7.803,0.175c3.739-0.042,7.599,0.103,11.31,0.205c3.684,0.102,7.493,0.206,11.298,0.206c4.485-0.082,9.494-0.122,15.277-0.122c8.954,0,18.336,0.095,28.269,0.194c10.197,0.103,20.741,0.208,31.066,0.208c0.167,0,0.326-0.001,0.492-0.001v4c-0.166,0-0.326,0.001-0.492,0.001c-10.345,0-20.899-0.105-31.106-0.208c-9.923-0.1-19.295-0.194-28.229-0.194c-5.758,0-10.743,0.04-15.241,0.122c-3.875-0.015-7.734-0.105-11.444-0.207c-3.684-0.102-7.493-0.206-11.298-0.206c-2.428-0.117-4.916-0.173-7.705-0.173c-4.689,0-9.549,0.161-14.249,0.316C2.165,87.596,1.083,87.632,0,87.665z"]
  },
  "iphone5_scale9_sw": {
    type: "SVG_PATHS",
    width: 62,
    height: 88,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101],
    paths: ["M26.698,3.649C38.217,4.29,50.017,4.727,62,5.019V85.25c-18.498,0.089-36.313-0.35-46.311-11.567C3.434,59.933,4.596,27.052,4.221,0h22.404C26.649,1.221,26.673,2.444,26.698,3.649z", "M59,3.743c1,0,2,0,3,0v5.067c-1-0.023-2-0.041-3-0.067c-12.085-0.321-22.6,2.362-34-1C24.864,5.309,24.736,2.703,24.615,0h2.927C27.673,0.964,27.818,1.895,28,2.743C37.19,4.554,47.888,3.855,59,3.743z", "M37.985,80.816c7.752,1.511,15.75,1.962,24.015,2.346v3.075c-3.914-0.533-17.483-1.207-19.587-1.187c-6.153,1.162-9.879-1.397-14.917-3.692c-1.763-0.903-8.428-4.862-9.331-6.625c-5.129-3.204-7.269-9.137-10.241-14.498c-0.333,0-2.615-4.303-2.191-6.393c0-0.333-0.036-20.622-1.733-27.1c0-0.333,0-0.667,0-1c0.564-4.563-1.772-10.228-1-15c0-0.333,0-0.667,0-1c0-2,0-4,0-6C2.792,3.743,2.455,1.931,2.23,0h2.735C5.705,18.511,6.75,33.813,8.192,43.948c1.18,14.735,4.093,26.62,18.569,32.964C27.099,76.912,37.647,80.816,37.985,80.816z", "M51.352,83.871c3.447,0,7.064-0.094,10.648-0.207v4.001c-3.575,0.112-7.186,0.206-10.648,0.206c-12.948,0-27.197-1.198-34.714-10.09C3.511,62.256,1.119,55.531,1.856,0h4C5.112,56.243,7.594,60.889,19.692,75.199C26.153,82.841,39.322,83.871,51.352,83.871z"]
  },
  "iphone5_scale9_w": {
    type: "SVG_PATHS",
    width: 62,
    height: 281,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101],
    paths: ["M26.625,281H4.221c-0.044-3.203-0.109-6.328-0.221-9.326C1.603,207.649,2.221,95.57,3.07,0h22.687C25.544,96.424,25.109,204.502,26.625,281z", "M27.542,281h-2.927c-0.447-9.941-0.804-21.287-1.083-31.035c0.127,3.925,0.281,7.852,0.468,11.778c-0.393-8.25-0.643-16.496-0.792-24.732C22.064,183.739,22.983,61.319,24,9.743c0-1,0-2,0-3C23.88,4.478,23.781,2.237,23.683,0h3.342C27.01,4.303,27,8.321,27,11.743c0,1,0,2,0,3c0,14.667,0,29.333,0,44c0,0.667,0,1.333,0,2c0,33.333,0,129.667,0,163c0,2.333,0,4.667,0,7s0,4.667,0,7c0,1.667,0,3.333,0,5C27.505,253.542,26.107,270.431,27.542,281z", "M5.403,7.232c0,0.564-4.217,179.114-0.437,273.768H2.23C2.095,279.833,2,278.622,2,277.743c0-1.333,0-2.667,0-4c0-2.667,0-5.333,0-8c-0.333,0-0.333-3,0-3c0-0.667,0.335-37.665,1-55c-0.333,0-1-17.333-1-24c0.042,0,0.083,0,0.094,0c-0.01,0-0.052,0-0.094,0c0-35.667,0-134.333,0-170c0-0.667,0-1.333,0-2C2,7.829,2,3.914,2,0h4.109C5.918,2.562,5.687,5.002,5.403,7.232z", "M5.699,0C6.361,94.728,7.094,208.261,6,271.708c-0.056,3.248-0.104,6.338-0.143,9.292h-4c0.04-2.979,0.088-6.091,0.144-9.36C3.094,208.232,2.361,94.713,1.699,0H5.699z"]
  },
  "iphone5_scale9_c": {
    type: "SVG_PATHS",
    width: 123,
    height: 281,
    paths: []
  },
  "iphone5_button": {
    type: "SVG_PATHS",
    width: 52,
    height: 52,
    colors: [0x010101, 0x010101],
    paths: ["M17.964,3.668c-0.333,0-0.667,0-1,0c-6.362,3.638-11.098,8.902-13,17c0,0.333,0,0.667,0,1c-6.437,14.771,6.681,30.839,20,29c0.333,0,0.667,0,1,0C64.544,58.329,52.55-14.43,17.964,3.668zM45.561,31.404c-0.657,7.044-5.753,14.482-11.63,15.858C-4.135,56.175-1.634,0.504,30.758,6.028c8.328,1.422,11.87,7.172,14.803,16.917C46.021,26.513,45.561,30.698,45.561,31.404z", "M14.003,16.758c-0.407,7.074,1.78,11.554,1,19c4.16,4.969,17.2,3.144,23,1c0.229-7.229-0.373-13.628-2-19C27.825,16.029,20.577,15.969,14.003,16.758zM18.003,32.758c0-0.333,0-0.667,0-1c-0.074-4.26-1.673-6.994-1-12c4.774-2.043,10.196,1.052,17,0c0.662,4.671,1.572,9.095,2,14C31.757,36.646,20.536,38.792,18.003,32.758z"]
  },
  "iphone5_earthing": {
    type: "SVG_PATHS",
    width: 60,
    height: 38,
    colors: [0x010101, 0x010101],
    paths: ["M10.206,21.648c-5.431,0.878-10.983,5.519-9,10.554c1.55,3.936,13.167,4.614,20,4.871c18.578,0.701,36.074,1.447,38-11.365C56.4,18.682,22.054,19.731,10.206,21.648zM4.437,27.089c2.017-4.235,44.511-3.589,46.935-3.363c2.844,0.501,3.268,2.307,3.268,2.681c-0.248,8.734-18.256,7.534-32.348,7.083C21.917,33.489,4.729,34.286,4.437,27.089z", "M25.482,1.931c-0.869,0.701-2.01,3.422-1.646,6.543c0.284,2.439,2.409,2.861,3.66,3.02c3.399,0.435,6.602,0.898,6.954-7.045C33.937,0.093,28.241-0.291,25.482,1.931z"]
  },
  "iphone6_scale9": {
    type: "SVG_SCALE9",
    scale9: [263, 61, 73, 99],
    width: 380,
    height: 754
  },
  "iphone6_scale9_nw": {
    type: "SVG_PATHS",
    width: 61,
    height: 263,
    colors: [0x010101, 0x010101, 0x010101, 0xc1c0c0, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M2,90.642h8.418c1.105,0,2,0.896,2,2v24.56c0,1.104-0.895,2-2,2H2c-1.104,0-2-0.896-2-2v-24.56C0,91.537,0.896,90.642,2,90.642", "M2,145.275h8.418c1.105,0,2,0.896,2,2v48.149c0,1.104-0.895,2-2,2H2c-1.104,0-2-0.896-2-2v-48.149C0,146.171,0.896,145.275,2,145.275", "M2,203.633h8.418c1.105,0,2,0.896,2,2v48.15c0,1.104-0.895,2-2,2H2c-1.104,0-2-0.896-2-2v-48.15C0,204.529,0.896,203.633,2,203.633", "M28,263H6.293L6.872,50.058c-1-37.149,16.047-48.001,49-48.001H61V88H28V263z", "M61,88v4H28.5c-1.104,0-2-0.896-2-2s0.896-2,2-2H61z", "M8,263H4V46c0-16.592,4.578-28.055,13.996-35.044C26.23,4.846,36.306,2,54,2l7,0.026v4.001L53.993,6C19.458,6,8,17.215,8,46V263z", "M32,263h-4V88h33v4H32V263z", "M31,263h-2V89h32v2H31V263z", "M33.872,263h-4V88.057H61v4H33.872V263z", "M10.293,263h-4L6.872,50.053C6.4,32.555,9.948,20.137,17.72,12.152C25.754,3.9,36.512,0.057,55.873,0.057H61v4h-5.127c-18.208,0-30.057,3.459-37.287,10.887c-6.979,7.17-8.157,18.639-7.714,35.061L10.293,263z"]
  },
  "iphone6_scale9_n": {
    type: "SVG_PATHS",
    width: 246,
    height: 263,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M0,88V2.057h73.369H238.03L246,2.333V88H0z", "M246,92H0v-4h246V92z", "M0,6.027V2.026l246,0.942v4L0,6.027z", "M0,92v-4h246v4H0z", "M0,91v-2h246v2H0z", "M0,92.057v-4h246v4H0z", "M0,4.057v-4h238.03L246,0.333v4.002l-8.039-0.278H0z"]
  },
  "iphone6_scale9_ne": {
    type: "SVG_PATHS",
    width: 73,
    height: 263,
    colors: [0x010101, 0xc1c0c0, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M62.582,145.275H71c1.104,0,2,0.896,2,2v48.149c0,1.104-0.896,2-2,2h-8.418c-1.105,0-2-0.896-2-2v-48.149C60.582,146.171,61.477,145.275,62.582,145.275", "M69.372,263h-22.5V88.057H45V88H0V2.333l20.872,0.724c27.613,0,50.963,18.773,49,47.001l-1,65.888L69.372,263z", "M45,90c0,1.104-0.896,2-2,2H0v-4h43C44.104,88,45,88.896,45,90z", "M48,263h-4V89c0-1.104,0.896-2,2-2s2,0.896,2,2V263z", "M66.52,263h-4L62,121.507V46.001C62,12.938,35.806,8,19,8l-0.184-0.008L8.071,7L0,6.969v-4L8.263,3l10.829,1C50.176,4.03,66,19.335,66,46.001V121.5L66.52,263z", "M45,263h-4V92H0v-4h45V263z", "M44,263h-2V91H0v-2h44V263z", "M69.372,263h-4l-0.5-147.049l1.005-66.032c0.804-11.565-2.773-21.968-10.345-30.085c-8.761-9.391-20.123-14.777-34.66-14.777L0,4.335V0.333l20.941,0.726c15.573-0.002,28.001,5.849,37.516,16.047c8.335,8.936,12.281,20.358,11.414,33.037l-0.999,65.832L69.372,263z", "M46.872,263h-4V92.057H0v-4h46.872V263z"]
  },
  "iphone6_scale9_e": {
    type: "SVG_PATHS",
    width: 73,
    height: 392,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M69.372,0l1.333,392H46.872V0H69.372z", "M48,0v392h-4V0H48z", "M67.879,370.493L68.221,392h-4l-0.342-21.468L62.52,0h4L67.879,370.493z", "M45,0v392h-4V0H45z", "M44,0v392h-2V0H44z", "M69.372,0l1.333,392h-4L65.372,0H69.372z", "M46.872,0v392h-4V0H46.872z"]
  },
  "iphone6_scale9_se": {
    type: "SVG_PATHS",
    width: 73,
    height: 99,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M70.705,0l0.167,49.057c0,27.613-22.381,47-50,47H0v-91h46.872V0H70.705z", "M48,0v3.5c0,1.104-0.896,2-2,2s-2-0.896-2-2V0H48z", "M43.054,3c0,1.104-0.896,2-2,2H0V1h41.054C42.157,1,43.054,1.896,43.054,3z", "M68.221,0L69,48.967C69,76.935,48.645,98,19,98h-0.035L0,97.67v-4.001L19.018,94C46.374,93.993,65,74.649,65,48.999L64.221,0H68.221z", "M45,0v5H0V1h41V0H45z", "M44,0v4H0V2h42V0H44z", "M70.705,0l0.167,49.05c0,27.941-20.355,49.007-50,49.007H0v-4h20.872c27.364,0,46-19.346,46-45L66.705,0H70.705z", "M46.872,0v5.057H0v-4h42.872V0H46.872z"]
  },
  "iphone6_scale9_s": {
    type: "SVG_PATHS",
    width: 246,
    height: 99,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M0,96.057v-91h246v91H98.511H0z", "M246,5H0V1h246V5z", "M0,96.033v-4.001L207.51,93L246,93.669v4.001L207.465,97L0,96.033z", "M0,5V1h246v4H0z", "M0,4V2h246v2H0z", "M0,5.057v-4h246v4H0z", "M0,98.057v-4h246v4H0z"]
  },
  "iphone6_scale9_sw": {
    type: "SVG_PATHS",
    width: 61,
    height: 99,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M29.872,5.057H61v91h-5.128c-27.614,0-50-19.387-50-47L4.168,17.778L4.287,0H28v5h1.872V5.057z", "M61,1v4H28.942c-1.105,0-2-0.896-2-2s0.895-2,2-2H61z", "M54.01,92L61,92.032v4.001l-6.873-0.032c-2.653,0.346-5.328,0.521-7.953,0.521c-14.051,0-23.937-4.781-31.467-13.462C7.502,74.755,4,63.614,4,48.999V0h4v48.999c0,13.61,3.182,23.894,9.728,31.439C24.487,88.23,33.3,92.521,46.174,92.521c2.495,0,5.04-0.17,7.564-0.504l0.135-0.019L54.01,92z", "M32,1h29v4H28V0h4V1z", "M31,2h30v2H29V0h2V2z", "M33.872,1.057H61v4H29.872V0h4V1.057z", "M55.873,94.057H61v4h-5.127c-29.625,0-49.971-21.039-50.001-48.945L4.167,17.826l0.001-0.062L4.287,0h4L8.168,17.732l1.704,31.324C9.872,74.711,28.508,94.057,55.873,94.057z"]
  },
  "iphone6_scale9_w": {
    type: "SVG_PATHS",
    width: 61,
    height: 392,
    colors: [0xc1c0c0, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M28,0v392H4.287l1.585-237.34L6.293,0H28z", "M8,0v392H4V0H8z", "M32,0v392h-4V0H32z", "M31,0v392h-2V0H31z", "M33.872,0v392h-4V0H33.872z", "M10.293,0L9.872,154.666L8.287,392h-4l1.585-237.354L6.293,0H10.293z"]
  },
  "iphone6_scale9_c": {
    type: "SVG_PATHS",
    width: 246,
    height: 392,
    paths: []
  },
  "iphone6_button": {
    type: "SVG_PATHS",
    width: 58,
    height: 58,
    colors: [0x010101],
    paths: ["M26.77,57.029c-7.36,0-14.091-3.305-18.952-9.303c-4.476-5.525-7.043-13.041-7.043-20.623c0-16.119,10.76-26.133,28.079-26.133c9.826,0,17.938,3.076,22.841,8.66c4.604,5.246,6.348,12.441,5.18,21.387c-1.562,11.957-5.891,18.939-14.038,22.637C38.248,55.736,32.091,57.029,26.77,57.029M28.853,4.971c-15.077,0-24.079,8.273-24.079,22.133c0,16.229,11.183,25.926,21.995,25.926c4.716,0,10.374-1.184,14.414-3.018c6.837-3.104,10.343-8.939,11.724-19.512c1.013-7.754-0.407-13.887-4.22-18.229C44.555,7.562,37.511,4.971,28.853,4.971"]
  },
  "iphone6_earthing": {
    type: "SVG_PATHS",
    width: 84,
    height: 38,
    colors: [0x010101, 0x010101, 0x010101],
    paths: ["M77.95,34.322H38.593c-1.763,0-3.13-0.473-4.06-1.404c-1.27-1.271-1.267-2.904-1.266-3.686c-0.004-0.043-0.021-0.1-0.035-0.152c-0.117-0.449-0.39-1.506,0.372-2.49c0.837-1.08,2.285-1.285,4.663-1.285h39.948c2.851,0,5,1.66,5,3.861c0,1.01,0.092,1.328,0.172,1.494c0.345,0.715,0.303,1.533-0.111,2.195C82.498,34.092,80.872,34.32,77.95,34.322M77.94,30.322h0.008c0.536,0,0.965-0.01,1.306-0.025c-0.021-0.254-0.031-0.514-0.036-0.785c-0.187-0.096-0.535-0.207-1.003-0.207H38.267c-0.407,0-0.736,0.008-0.999,0.021c0.002,0.283,0.015,0.684,0.096,0.766c0.064,0.064,0.409,0.23,1.229,0.23H77.94L77.94,30.322z", "M58.614,13.996c-3.74,0-6.347-3.689-6.347-6.998c0-3.354,2.61-5.693,6.347-5.693c3.5,0,6.347,2.848,6.347,6.346C64.961,9.219,64.961,13.996,58.614,13.996M58.614,5.305c-0.707,0-2.347,0.164-2.347,1.693c0,1.457,1.206,2.998,2.347,2.998c2.248,0,2.347-0.525,2.347-2.346C60.961,6.355,59.908,5.305,58.614,5.305", "M7.973,36.695c-4.327,0-7.59-3.033-7.59-7.059l-0.001-0.275c-0.007-1.516-0.019-4.055,1.802-5.885c1.291-1.297,3.239-1.955,5.789-1.955c4.184,0,7.588,3.404,7.588,7.586C15.561,36.695,9.85,36.695,7.973,36.695M7.973,25.521c-0.995,0-2.314,0.135-2.954,0.777c-0.649,0.652-0.643,1.922-0.637,3.045l0.001,0.293c0,2.113,1.803,3.059,3.59,3.059c2.549,0,3.588-0.322,3.588-3.588C11.561,27.131,9.952,25.521,7.973,25.521"]
  },
  "iphoneX_scale9": {
    type: "SVG_SCALE9",
    scale9: [331, 140, 137, 71],
    width: 434,
    height: 867
  },
  "iphoneX_scale9_nw": {
    type: "SVG_PATHS",
    width: 140,
    height: 331,
    colors: [0xc1c0c0, 0x010101, 0x010101],
    paths: ["M27,331H4.1V64.471c0-35.199,27.74-64,62.94-64H140v52.194h-6.561c-10.422,0-18.9-8.492-18.9-18.931V29.6c0-3.846-3.123-6.974-6.963-6.974h-2.984v-0.05H66.293C44.627,22.576,27,40.9,27,63.424V331z", "M7.659,331H2.488c0-3.328,0-6.656,0-9.984c-0.718-0.596-1.28-1.496-1.705-2.518c-0.295-0.709-0.365-1.186-0.3-1.493c-0.001-0.017-0.006-0.037-0.006-0.053c0-1.132,0-2.263,0-3.394c0-14.55,0-29.098,0-43.646c0-4.291,0-8.582,0-12.873c0-1.305,0.952-1.762,2.011-1.727c0-4.461,0-8.92,0-13.381c-0.552-0.223-1.042-0.598-1.433-1.115c-0.301-0.186-0.491-0.459-0.491-0.834c-0.1-0.238-0.086-0.463,0-0.672c0-1.447,0-2.895,0-4.343c0-14.819,0-29.641,0-44.461c0-2.685,0-5.37,0-8.055c0-0.741-0.282-2.15,0.02-2.875c0.39-0.937,1.084-1.626,1.904-2.063c0-8.773,0-17.548,0-26.32c-0.827-0.674-1.389-1.795-1.746-2.652c-0.278-0.668-0.351-1.135-0.3-1.452c-0.003-0.032-0.012-0.069-0.012-0.101c0-8.129,0-16.258,0-24.387c0-2.401-0.544-5.408,0.399-7.676c0.22-0.529,0.895-0.71,1.658-0.716c0-11.503,0-23.006,0-34.508c0-8.08-0.56-16.662,0.752-24.651C7.798,27.301,19.365,16.123,39.926,5.95c9.7-4.8,11.201-5.171,31.899-5.401C94.449,0.298,117.199,0.28,140,0.33v4.202C119.492,4.488,98.987,4.45,78.497,4.45c-4.687,0-9.556-0.558-14.226-0.101C37.005,7.027,19.982,21.632,11.82,43.684C7.583,55.127,7.659,66.23,7.659,78.047c0,33.732,0.2,129.943,0.2,143.71c0,5.443,0,10.886,0,16.329c0,0.422-0.08,0.812-0.2,1.18C7.659,269.844,7.659,300.422,7.659,331z", "M29.42,331h-6.511c-0.001-12.913-0.003-25.826-0.005-38.739C22.896,240.352,23.507,188.96,23.5,137.05c-0.003-21.212-0.623-39.733-0.626-60.946c0-0.676,0.251-5.166,0.627-5.354c0.024-3.693,0.149-6.332,0.503-9.952c0.781-7.978,3.684-15.827,8.5-22.228c6.827-9.709,17.214-15.992,29.172-17.732c0.07-0.736,3.692-0.898,4.725-0.889c16.732,0.167,26.065,0.037,42.8,0.096c1.26,0.005,3.8,0.432,5.8,2.805c1.4,1.699,1.7,3.498,1.822,4.862c0.56,6.224-0.248,12.371,4.67,17.297c5.048,5.056,10.701,5.41,17.18,5.41c0.443,0,0.885,0,1.328,0v5.365c-3.856-0.012-7.686-0.191-11.406-0.759c-3.345-0.51-7.4-0.646-9.898-3.254c-2.529-2.644-4.238-5.93-5.6-9.303c-1.416-3.509-1.436-6.679-1.604-10.402c-0.135-2.992-0.855-5.551-2.793-6.875c-1.424-0.416-2.793-0.317-4.329-0.317c-0.179,0-0.347-0.019-0.509-0.05c-11.771,0-23.541,0-35.311,0c-6.883,0-13.65,1.135-19.621,4.381c-1.713,0.99-3.354,2.1-4.904,3.324c-8.785,7.229-14.392,17.904-14.582,29.404c-0.356,21.424-1.043,42.768-1.043,64.176c0,51.707,1.02,103.414,1.02,155.123C29.42,297.821,29.42,314.411,29.42,331z"]
  },
  "iphoneX_scale9_n": {
    type: "SVG_PATHS",
    width: 157,
    height: 331,
    colors: [0xc1c0c0, 0x010101, 0x010101],
    paths: ["M0,52.665V0.471h157v52.194H0z", "M0,4.532V0.33c27.189,0.06,54.452,0.219,81.673,0.219c25.109,0,50.218,0.129,75.327,0.183V4.7c-18.211,0-36.422,0-54.633,0C68.267,4.7,34.129,4.605,0,4.532z", "M0,55.784v-5.365c15.788,0,31.576,0,47.363,0c30.726,0,61.451,0,92.177,0c5.82,0,11.64,0,17.46,0v4.485c-0.552,0.005-1.106,0.008-1.669,0.008c-14.63,0-29.261,0-43.89,0c-9.916,0.129-19.832,0.257-29.748,0.385c-21.367,0.276-42.744,0.105-64.112,0.125C11.851,55.427,5.894,55.803,0,55.784z"]
  },
  "iphoneX_scale9_ne": {
    type: "SVG_PATHS",
    width: 137,
    height: 331,
    colors: [0xc1c0c0, 0x010101, 0x010101],
    paths: ["M134.04,331H110V63.424c0-22.523-17.627-40.848-39.293-40.848H32.408v0.05h-2.984c-3.839,0-6.963,3.128-6.963,6.974v4.135c0,10.438-8.478,18.931-18.9,18.931H0V0.471h70.04c35.2,0,64,28.801,64,64V331z", "M134.669,331h-4.949c-0.03-23.071-0.088-46.15-0.219-69.35c-0.694-0.366-1.047-62.797-0.579-63.084c-0.094-24.496-0.251-59.113-0.251-102.895c0-9.295,0-18.59,0-27.885c0-8.268,0.109-16.526-2.947-24.276C117.1,21.639,96.552,6.139,72.665,4.837C59.255,4.105,45.683,4.7,32.263,4.7c-10.754,0-21.509,0-32.263,0V0.731c22.136,0.047,44.271,0.036,66.405-0.183c41.703-0.412,67.596,23.459,67.992,64.817c0.057,5.901,0,11.804,0,17.706c0,38.25,0,76.498,0,114.748c1.442,0.549,2.603,1.639,2.605,3.125c0,13.561,0,27.121,0,40.683c0,5.647,0,11.297,0,16.946c0,1.613-1.115,2.564-2.512,2.921C134.548,284.66,134.617,307.842,134.669,331z", "M113.971,331h-6.39c0-29.688,0-59.375,0-89.062c0-46.414,0-92.83,0-139.244c0-11.315,0-22.629,0-33.944c0-1.74,0-3.479,0.002-5.217c0-8.399-1.472-17.446-5.988-24.681c-3.471-5.558-8.206-9.328-13.713-11.59c-4.225-1.33-8.714-2.07-13.052-2.598c-0.229,0.119-0.514,0.191-0.868,0.193c-15.062,0.088-30.127,0.175-45.188,0.263c-2.671,0.032-3.87,0.832-3.892,4.649c-0.015,2.657,0.049,5.248-0.418,7.877c-0.58,3.267-2.131,6.492-4.3,9.205c-1.008,1.352-2.208,2.569-3.637,3.586c-0.994,0.763-2.046,1.428-3.146,1.961C8.936,54.56,4.643,54.864,0,54.904v-4.485c1.167,0,2.335,0,3.503,0c5.361,0,9.699-1.067,12.684-6.015c2.087-3.579,3.148-7.924,3.148-12.34c0-1.855,0.166-5.514,0.267-8.213c-0.301-2.101,0.6-3.201,1.738-3.307c3.043-0.281,6.111-0.035,9.167-0.053c0.522-0.049,1.058-0.085,1.619-0.096c0.171-0.039,0.342-0.064,0.506-0.065c13.931-0.076,27.861-0.151,41.792-0.226c11.596,0.168,22.52,4.369,29.409,13.187c0.246,0.226,0.506,0.437,0.743,0.672c7.686,7.647,8.375,20.397,8.412,30.629c0.021,5.993,0.045,11.987,0.065,17.981c0.21,56.8,0.42,113.6,0.629,170.398C113.779,278.982,113.875,304.991,113.971,331z"]
  },
  "iphoneX_scale9_e": {
    type: "SVG_PATHS",
    width: 137,
    height: 465,
    colors: [0xc1c0c0, 0x010101, 0x010101],
    paths: ["M134.04,0v465H110V0H134.04z", "M134.738,53.219c0,69.4,0.563,135.013,0.563,196.333c0,52.27-0.1,103.565-0.1,155.5c0,19.979-0.7,39.968-0.786,59.948h-5.225c-0.015-13.925-0.092-27.85-0.092-41.774c0-48.338,0.649-96.674,0.649-145.012c0-64.408,0-128.813,0-193.222c0-28.441,0.008-56.71-0.029-84.992h4.949C134.708,17.762,134.738,35.512,134.738,53.219z", "M113.702,143.944c-0.358,63.718,0.355,127.436-0.002,191.154c-0.069,12.397-0.141,24.794-0.209,37.191c0,7.837,0,15.675,0,23.512c0,23.065-1.198,46.134-1.12,69.198h-5.451c0.003-0.435,0.006-0.87,0.008-1.305c0.105-18.611,0.21-37.224,0.314-55.834c0.113-20.101,0.226-40.2,0.339-60.301c0-18.031,0-36.063,0-54.095c0-63.775,0-127.553,0-191.329c0-34.045,0-68.091,0-102.137h6.39c0.07,18.882,0.141,37.764,0.21,56.646C114.289,85.742,113.865,114.849,113.702,143.944z"]
  },
  "iphoneX_scale9_se": {
    type: "SVG_PATHS",
    width: 137,
    height: 71,
    colors: [0xc1c0c0, 0x010101, 0x010101],
    paths: ["M134.04,0v6.471c0,35.2-28.8,64-64,64H0V48.577h69.215C91.704,48.577,110,30.252,110,7.729V0H134.04z", "M134.416,0c-0.006,1.288-0.017,2.575-0.017,3.863c-0.078,7.002-0.779,13.905-1.571,20.89c-3.076,27.14-22.601,41.5-39,43c-8.194,0.749-22.72,1.839-30.845,1.839c-20.074,0-41.18,0.961-61.256,0.961c-0.577,0-1.151,0.001-1.728,0.002v-4.049c1.947,0,3.893,0,5.84,0c19.323,0,38.647-1.055,57.969-1.055c10.562,0,31.891,0.402,40.791-5.602c21.201-14.3,24.342-31.736,24.602-55.615c0-1.411-0.008-2.823-0.01-4.234H134.416z", "M112.419,6.807c0.182,11.744-1.108,16.127-4.318,22.144C99.959,44.715,80.996,49.046,77.8,49.75c-0.6,0.8-3.56,0.594-4.3,0.6c-16.434,0.131-30.067,0.767-46.5,0.7c-0.394-0.002-1.402-0.026-1.738-0.228c-4.066,0-8.135,0-12.202,0c-4.353,0-8.707,0-13.06,0v-4.491c23.14,0,46.279,0,69.42,0c12.914,0,24.301-6.609,31.202-16.535c3.921-6.322,6.208-13.733,6.252-21.503c0.016-2.766,0.031-5.529,0.046-8.293h5.451C112.379,2.269,112.384,4.538,112.419,6.807z"]
  },
  "iphoneX_scale9_s": {
    type: "SVG_PATHS",
    width: 157,
    height: 71,
    colors: [0xc1c0c0, 0x010101, 0x010101],
    paths: ["M0,70.471V48.577h157v21.894H0z", "M0,70.977v-4.471c52.333,0,104.666,0,157,0v4.049C104.393,70.562,52.434,70.884,0,70.977z", "M0,50.822v-4.491c33.045,0,66.09,0,99.135,0c19.289,0,38.577,0,57.865,0v4.491C104.667,50.822,52.333,50.822,0,50.822z"]
  },
  "iphoneX_scale9_sw": {
    type: "SVG_PATHS",
    width: 140,
    height: 71,
    colors: [0xc1c0c0, 0x010101, 0x010101],
    paths: ["M67.785,48.577H140v21.894H67.04c-35.2,0-62.94-28.8-62.94-64V0H27v7.729C27,30.252,45.296,48.576,67.785,48.577z", "M116.695,66.506c7.769,0,15.537,0,23.305,0v4.471c-8.745,0.016-17.5,0.026-26.277,0.026c-14.91,0-29.24-1.05-44.152-1.05c-36.076,0-66.967-17.842-67.083-66.432C2.485,2.348,2.488,1.174,2.486,0h5.173c0,1.271,0,2.542,0,3.812c0,39.899,25.336,62.693,62.024,62.693C85.354,66.506,101.025,66.506,116.695,66.506z", "M76.801,46.331c21.066,0,42.133,0,63.199,0v4.491c-5.501,0-11.003,0-16.505,0c-17.819,0-35.638,0-53.455,0c-11.224,0-22.211-2.957-30.724-10.503C32.363,34.653,27.4,29.115,25.182,20.349c-1.106-0.471-2.207-3.721-2.207-4.908C22.973,10.293,22.973,5.146,22.972,0h6.448c0,2.471,0,4.94,0,7.411c0,18.481,13.823,35.56,32.319,38.397C66.698,46.569,71.796,46.331,76.801,46.331z"]
  },
  "iphoneX_scale9_w": {
    type: "SVG_PATHS",
    width: 140,
    height: 465,
    colors: [0xc1c0c0, 0x010101, 0x010101],
    paths: ["M27,0v465H4.1V0H27z", "M7.659,232.532c0.837,59.62,0,108.004,0,162.004c0,23.488,0,46.977,0,70.464H2.486c-0.042-22.05,0.002-44.1,0.002-66.151c0-53.168,1.175-106.352,0-159.506c-1.453-65.724,0-131.101,0-196.653c0-14.229,0-28.459,0-42.689h5.171c0,11.83,0,23.66,0,35.49C7.659,101.171,6.737,166.859,7.659,232.532z", "M29.42,0c0,48.636,0,97.271,0,145.907c0,61.466,0,122.934,0,184.401c0,40.436,0,80.87,0,121.307c0,4.461,0,8.923,0,13.385h-6.448c0-1.413,0-2.826-0.001-4.239c-0.005-39.923-0.011-79.848-0.016-119.771c-0.008-61.29-0.018-122.581-0.025-183.871C22.923,104.745,22.917,52.373,22.909,0H29.42z"]
  },
  "iphoneX_scale9_c": {
    type: "SVG_PATHS",
    width: 157,
    height: 465,
    paths: []
  },
  "iphoneX_earthing": {
    type: "SVG_PATHS",
    width: 87,
    height: 17,
    colors: [0x010101, 0x010101],
    paths: ["M2,15.074c2.181,1.006,5.174,0.768,7.5,0.803c5.292,0.082,10.582-0.257,15.868-0.431c6.398-0.211,12.798-0.422,19.197-0.633c4.863-0.16,11.896,1.883,13.156-4.357c0.04-0.198,0.055-0.384,0.053-0.562c0.014-0.198,0.01-0.402-0.035-0.618c-1.062-5.258-5.701-4.92-9.896-4.92c-6.634,0-13.269,0-19.902,0c-6.349,0-12.698,0-19.047,0c-4.378,0-9.331,0.959-8.863,6.645C0.193,12.972-0.079,14.115,2,15.074zM11.877,8.872c5.324,0,10.648,0,15.973,0c5.5,0,10.998,0,16.497,0c2.101,0,7.131-1.109,8.36,1.172c-1.502,2.304-8.778,1.28-10.863,1.349c-5.177,0.171-10.354,0.341-15.53,0.512c-3.237,0.106-6.473,0.256-9.71,0.362c-2.944,0.095-5.89,0.139-8.834,0.059c-0.896-0.024-1.782-0.092-2.668-0.213c-0.036-0.431-0.071-0.862-0.106-1.292C4.739,7.721,9.954,8.872,11.877,8.872z", "M78.213,16.685c4.203,0.872,7.949-2.65,8.176-6.851C86.15,5.402,82.279,2,78.365,0.465c-4.633-1.817-9.039,1.961-9.285,6.53c-0.008,0.138,0.005,0.273,0.007,0.41c-0.277,0.726-0.452,1.525-0.499,2.396C68.275,15.586,73.574,17.727,78.213,16.685zM76.896,3.901c0.06-0.031,0.123-0.052,0.185-0.08c0.065,0.008,0.133,0.008,0.199,0.019c2.73,0.444,4.316,3.041,4.649,5.571c-0.632,1.502-1.772,2.696-3.356,3.516c-1.347,0.698-3.027,0.643-4.441,0.254c-1.01-0.278-1.098-2.524-1.143-3.349C72.849,7.229,74.727,5.025,76.896,3.901z"]
  },
  "android_nexus": {
    type: "SVG_PATHS",
    width: 356,
    height: 725,
    colors: [0x010101, 0x010101, 0x010101, 0x010101, 0xffffff, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, null, 0x010101, null, 0x010101, null, 0x010101],
    paths: ["M340,648H16V76h324V648zM20,644h316V80H20V644z", "M221.848,705.061H135c-3.674,0-22-0.313-22-6.529c0-6.217,18.326-6.53,22-6.53h86.848c3.674,0,22,0.313,22,6.53C243.848,704.747,225.521,705.061,221.848,705.061zM117.473,698.53c1.927,1.046,7.946,2.53,17.527,2.53h86.848c9.58,0,15.6-1.484,17.525-2.53c-1.928-1.046-7.945-2.53-17.525-2.53H135C125.419,696,119.4,697.484,117.473,698.53z", "M221.848,41.061H135c-3.674,0-22-0.314-22-6.529S131.326,28,135,28h86.848c3.674,0,22,0.314,22,6.529C243.848,40.745,225.521,41.061,221.848,41.061zM117.473,34.529c1.927,1.047,7.946,2.531,17.527,2.531h86.848c9.58,0,15.6-1.484,17.525-2.531C237.445,33.484,231.428,32,221.848,32H135C125.419,32,119.4,33.484,117.473,34.529z", "M73.352,42.055c-3.601,0-6.53-2.931-6.53-6.531c0-3.6,2.93-6.529,6.53-6.529s6.53,2.93,6.53,6.529C79.882,39.124,76.952,42.055,73.352,42.055zM73.352,32.994c-1.395,0-2.53,1.135-2.53,2.529c0,1.396,1.135,2.531,2.53,2.531s2.53-1.136,2.53-2.531C75.882,34.129,74.747,32.994,73.352,32.994z", "M353.9,436.214L354,94.283l-0.1-60.285C353.9,16.32,339.676,2,322.005,2l-84.407,0.1L145.38,2L54.866,2.2L33.995,2C22.648,2,14.221,6.477,10.116,13.384C6.385,19.664,3,25.796,3,33.999v592.458v64.545c0,3.699,0.966,7.151,1.781,10.556C8.709,717.971,19.556,720.631,33.995,723l35.192-0.725L248.812,723h73.191c9.212,0,17.192-2.706,23.029-8.933c5.363-5.724,8.966-14.604,8.966-23.066l-0.5-12.884L353.9,436.214zM340,648.865H20v-568h320V648.865z", "M33.853,725.003l-0.181-0.029c-15.445-2.533-26.698-5.664-30.836-22.95l-0.294-1.215C1.784,697.707,1,694.499,1,691.001V33.999c0-8.496,3.35-14.825,7.397-21.636C13.065,4.506,22.396,0,33.995,0l20.89,0.2L145.376,0L237.6,0.1L322.003,0C340.695,0,355.9,15.252,355.9,33.999l0.1,60.28l-0.1,341.936l-0.4,241.905l0.498,12.804c0.002,8.893-3.641,18.255-9.505,24.511c-5.949,6.349-14.188,9.565-24.488,9.565h-73.192l-179.633-0.725L33.853,725.003zM69.146,720.275L248.82,721h73.185c9.149,0,16.407-2.793,21.57-8.301c5.118-5.461,8.425-13.979,8.425-21.698l-0.498-12.807l0.398-241.983L352,94.283l-0.1-60.281C351.9,17.457,338.489,4,322.005,4L237.6,4.1L145.378,4L54.87,4.2L33.976,4C23.842,4,15.765,7.793,11.835,14.406C7.962,20.924,5,26.505,5,33.999v657.002c0,3.017,0.726,5.986,1.427,8.858l0.299,1.23c3.566,14.897,12.738,17.489,27.413,19.906L69.146,720.275zM342,650.865H18v-572h324V650.865zM22,646.865h316v-564H22V646.865z", "M26.204,724.266c-10.909,0-18.292-3.257-21.388-19.469c-0.055-0.285-0.169-0.659-0.307-1.127C3.945,701.748,3,698.533,3,693.867v-64.545l1.566-416.413L3,36.882c0-8.304,2.206-16.979,6.053-23.819C11.396,8.895,17.36,3.48,32.466,1.243V1.008h25.105l89.885,2.982l92.117-1.024l84.429-0.1c17.569,0,29.838,14.392,29.838,34.999L354,97.145l-0.16,342.213l-0.34,241.629l0.495,12.726c1.116,9.604-1.76,14.787-7.102,20.484c-6.745,7.194-14.028,8.67-22.891,8.67l-67.95-0.898L71.2,723.141l-35.164,0.726c-0.889,0.042-1.812,0.092-2.717,0.14C30.812,724.139,28.445,724.266,26.204,724.266zM34.169,5.008c-0.285,0.072-0.551,0.11-0.749,0.139c-13.777,1.971-18.959,6.457-20.881,9.875C9.019,21.283,7,29.244,7,36.864l1.566,176.034L7,629.33v64.537c0,4.09,0.843,6.96,1.347,8.675c0.177,0.601,0.316,1.076,0.398,1.505c2.618,13.71,7.934,16.219,17.459,16.219c2.134,0,4.45-0.124,6.902-0.255c0.913-0.049,1.844-0.099,2.794-0.144l35.246-0.728l184.92-1.174l67.964,0.898c9.472,0,14.885-2.007,19.945-7.405c4.765-5.083,7.007-9.163,6.037-17.358l-0.012-0.156l-0.5-12.886l0.338-241.707L350,97.148l-0.16-59.278c0-21.304-13.393-31.004-25.836-31.004l-84.406,0.1l-92.185,1.023L57.505,5.007L34.169,5.008zM342,650.865H18v-572h324V650.865zM22,646.865h316v-564H22V646.865z", "M178.197,706.38l-0.061-0.002l-43.423-1.318c-0.005,0-0.011,0-0.016,0c-3.649,0-21.923-0.321-21.923-6.528c0-5.005,5.14-6.53,22-6.53l40.751-0.438L221.641,692c0.045,0,0.092,0,0.137,0c12.209,0,21.845,0.023,21.845,6.53c0,4.489-5.539,6.138-21.952,6.528L178.197,706.38zM116.836,698.268c1.313,1.038,7.519,2.793,17.938,2.793l43.423,1.317l43.364-1.315c15.767-0.377,17.812-2.021,18.062-2.518C239.621,696,229.218,696,221.621,696l-46.093-0.438L134.795,696C120.495,696,117.29,697.318,116.836,698.268z", "M125.461,41.178c-4.746,0-12.688,0-12.688-6.648c0-6.346,20.629-6.529,22.982-6.529l63.095,0.475L221.579,28c0.003,0,0.006,0,0.008,0c3.731,0,22.034,0.318,22.034,6.529c0,6.216-18.326,6.531-22,6.531h-86.847c-1.748,0-3.443,0.031-5.058,0.062C128.223,41.15,126.798,41.178,125.461,41.178zM116.793,34.886c0.149,1.246,1.335,2.292,8.668,2.292c1.312,0,2.713-0.027,4.179-0.054c1.639-0.031,3.359-0.063,5.133-0.063h86.847c9.581,0,15.601-1.483,17.526-2.529c-1.928-1.047-7.945-2.53-17.526-2.53l-22.743,0.474L135.741,32C125.229,32,118.211,33.775,116.793,34.886z", "M73.125,42.055c-6.53,0-6.53-4.744-6.53-6.531c0-3.303,2.499-5.918,6.218-6.506c4.244-0.668,6.843,1.871,6.843,6.506C79.656,39.125,76.726,42.055,73.125,42.055zM74.117,32.911c-0.205,0-0.434,0.021-0.679,0.06c-1.312,0.206-2.843,1-2.843,2.555c0,1.994,0,2.529,2.53,2.529c1.395,0,2.53-1.135,2.53-2.529C75.656,32.911,74.73,32.911,74.117,32.911z", "M17,74.925v575.288", "M17,652.213c-1.104,0-2-0.896-2-2V74.925c0-1.104,0.896-2,2-2s2,0.896,2,2v575.288C19,651.317,18.104,652.213,17,652.213z", "M15.209,78l325.641-0.161", "M15.209,80c-1.104,0-2-0.895-2-1.999s0.895-2,1.999-2.001l325.64-0.161h0.001c1.104,0,1.999,0.896,2,1.999s-0.895,2-1.999,2.001L15.209,80C15.21,80,15.21,80,15.209,80z", "M15.521,648.404l323.749,1.189", "M339.27,651.594c-0.002,0-0.005,0-0.008,0l-323.748-1.188c-1.104-0.004-1.997-0.902-1.993-2.009c0.004-1.102,0.899-1.991,2-1.991c0.002,0,0.005,0,0.007,0l323.75,1.188c1.104,0.004,1.996,0.902,1.991,2.008C341.266,650.703,340.371,651.594,339.27,651.594z"]
  },
  "android_nexus_scale9": {
    type: "SVG_SCALE9",
    scale9: [86, 43, 38, 85],
    width: 356,
    height: 725
  },
  "android_nexus_scale9_nw": {
    type: "SVG_PATHS",
    width: 43,
    height: 86,
    colors: [0xffffff, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M20,86H3V33.999c0-8.203,3.385-14.335,7.116-20.615C14.221,6.477,22.648,2,33.995,2L43,2.086v78.779H20V86z", "M22,86h-4v-7.135h25v4H22V86z", "M5,86H1V33.999c0-8.496,3.35-14.825,7.397-21.637C13.066,4.506,22.396,0,33.995,0L43,0.086v4.001L33.976,4c-10.134,0-18.21,3.793-22.141,10.405C7.962,20.924,5,26.506,5,33.999V86z", "M22,86h-4v-7.134h25v4H22V86z", "M7,36.864c0-7.62,2.019-15.581,5.539-21.842c1.922-3.418,7.104-7.905,20.881-9.876c0.198-0.028,0.465-0.066,0.75-0.139H43v-4H32.466v0.235C17.36,3.48,11.396,8.895,9.053,13.062C5.206,19.902,3,28.578,3,36.882L3.437,86h4L7,36.864zM32.466,1.644v2.783c-0.308-0.316-0.529-0.747-0.562-1.259C31.862,2.554,32.104,2.019,32.466,1.644z", "M19,86h-4V74.926c0-1.105,0.896-2,2-2s2,0.895,2,2V86z", "M43,75.986v4L15.21,80c0,0,0,0-0.001,0c-1.104,0-2-0.895-2-1.999S14.104,76,15.208,76L43,75.986z", "M20,86h-4V76h27v4H20V86z"]
  },
  "android_nexus_scale9_n": {
    type: "SVG_PATHS",
    width: 275,
    height: 86,
    colors: [0xffffff, 0x010101, 0x010101, 0x010101, 0x010101, 0xffffff, 0x010101, 0x010101],
    paths: ["M0,80.865V2.086L11.866,2.2L102.38,2l92.217,0.1L275,2.005v78.86H0z", "M0,82.865v-4h275v4H0z", "M0,4.087V0.086L11.885,0.2l90.49-0.2l92.223,0.1L275,0.005v4L194.599,4.1L102.378,4L11.871,4.2L0,4.087z", "M0,82.866v-4h275v4H0z", "M0,5.008v-4h14.572l89.885,2.982l92.117-1.024L275,2.873v4l-78.402,0.093l-92.186,1.023L14.505,5.007L0,5.008z", "M-27.791,78l325.64-0.161z", "M275,79.851L0,79.986v-4l275-0.136V79.851z", "M0,80v-4h275v4H0z"]
  },
  "android_nexus_scale9_ne": {
    type: "SVG_PATHS",
    width: 38,
    height: 86,
    colors: [0xffffff, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M35.986,86H22v-5.135H0V2.005L4.005,2C21.676,2,35.9,16.32,35.9,33.999L35.986,86z", "M37.986,86h-4L33.9,34.002C33.9,17.457,20.489,4,4.005,4L0,4.005v-4L4.003,0C22.695,0,37.9,15.252,37.9,33.999L37.986,86z", "M24,86h-4v-3.135H0v-4h24V86z", "M35.97,86h-4l-0.13-48.13c0-14.935-8.085-31.004-25.836-31.004L0,6.873v-4l6.002-0.007c17.569,0,29.838,14.393,29.838,34.999L35.97,86z", "M24,86h-4v-3.134H0v-4h24V86z", "M24.85,77.838c0.001,1.104-0.895,2-1.999,2.001L0,79.851v-4l22.849-0.012h0.001C23.954,75.839,24.849,76.733,24.85,77.838z", "M22,86h-4v-6H0v-4h22V86z"]
  },
  "android_nexus_scale9_e": {
    type: "SVG_PATHS",
    width: 38,
    height: 554,
    colors: [0xffffff, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M35.986,0L36,8.283l-0.1,341.931L35.563,554H22V0H35.986z", "M37.986,0L38,8.28l-0.1,341.935L37.563,554h-3.999L33.9,350.211L34,8.282L33.986,0H37.986z", "M24,0v554h-4V0H24z", "M35.97,0L36,11.143l-0.16,342.214L35.558,554H31.56l0.28-200.647L32,11.147L31.97,0H35.97z", "M24,0v554h-4V0H24z", "M22,0v554h-4V0H22z"]
  },
  "android_nexus_scale9_se": {
    type: "SVG_PATHS",
    width: 38,
    height: 85,
    colors: [0xffffff, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M35.5,38.117L36,51.001c0,8.463-3.603,17.344-8.966,23.066C21.197,80.294,13.217,83,4.005,83H0V8.865h22V0h13.563L35.5,38.117z", "M37.5,38.12l0.498,12.804c0.002,8.892-3.641,18.255-9.505,24.511C22.544,81.782,14.305,85,4.005,85H0v-4h4.005c9.149,0,16.407-2.793,21.57-8.301C30.693,67.238,34,58.722,34,51.001l-0.498-12.807L33.564,0h3.999L37.5,38.12z", "M24,0v10.865H0v-4h20V0H24z", "M35.995,53.711c1.116,9.603-1.76,14.787-7.101,20.484c-6.746,7.195-14.029,8.67-22.891,8.67L0,82.786v-4.001l6.03,0.08c9.472,0,14.885-2.007,19.946-7.405c4.764-5.083,7.006-9.163,6.037-17.358l-0.012-0.157l-0.5-12.885L31.56,0h3.998L35.5,40.985L35.995,53.711z", "M24,0v10.866H0v-4h20V0H24z", "M23.27,9.602c-0.004,1.102-0.898,1.992-2,1.992c-0.002,0-0.005,0-0.008,0L0,11.516v-4l21.277,0.078C22.382,7.598,23.273,8.496,23.27,9.602z", "M22,0v8H0V4h18V0H22z"]
  },
  "android_nexus_scale9_s": {
    type: "SVG_PATHS",
    width: 275,
    height: 85,
    colors: [0xffffff, 0x010101, 0x010101, 0x010101, 0x010101, 0xffffff, 0x010101, 0x010101],
    paths: ["M0,82.814V8.865h275V83h-69.188L26.187,82.275L0,82.814z", "M0,10.865v-4h275v4H0z", "M0,84.814v-4l26.146-0.539L205.82,81H275v4h-69.188L26.179,84.275L0,84.814z", "M0,10.866v-4h275v4H0z", "M0,83.722v-4.001l28.146-0.58l184.921-1.174L275,78.785v4.001l-61.946-0.819L28.2,83.141L0,83.722z", "M-27.479,8.404L296.27,9.594z", "M275,11.516L0,10.505v-4l275,1.011V11.516z", "M0,8V4h275v4H0z"]
  },
  "android_nexus_scale9_sw": {
    type: "SVG_PATHS",
    width: 43,
    height: 85,
    colors: [0xffffff, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101, 0x010101],
    paths: ["M20,8.865h23v73.949L33.995,83C19.556,80.631,8.709,77.971,4.781,61.557C3.966,58.152,3,54.7,3,51.001V0h17V8.865z", "M22,6.865h21v4H18V0h4V6.865z", "M34.138,80.997L43,80.814v4l-9.147,0.188l-0.181-0.029C18.226,82.439,6.973,79.31,2.835,62.022l-0.294-1.214C1.784,57.707,1,54.499,1,51.001V0h4v51.001c0,3.017,0.726,5.986,1.427,8.858l0.299,1.231C10.291,75.988,19.463,78.58,34.138,80.997z", "M22,6.866h21v4H18V0h4V6.866z", "M35.9,79.867l7.1-0.146v4.001l-6.964,0.144c-0.889,0.042-1.812,0.092-2.716,0.14c-2.507,0.134-4.875,0.261-7.116,0.261c-10.909,0-18.292-3.257-21.388-19.469c-0.055-0.285-0.169-0.659-0.307-1.127C3.945,61.748,3,58.533,3,53.867V0h4v53.867c0,4.09,0.843,6.96,1.347,8.675c0.177,0.601,0.316,1.076,0.398,1.505c2.618,13.71,7.934,16.219,17.459,16.219c2.134,0,4.45-0.124,6.902-0.255C34.019,79.962,34.95,79.912,35.9,79.867z", "M19,0v10.213c0,1.104-0.896,2-2,2s-2-0.896-2-2V0H19z", "M43,6.505v4l-27.486-0.101c-1.104-0.004-1.997-0.902-1.993-2.008c0.004-1.102,0.899-1.992,2-1.992c0.002,0,0.005,0,0.007,0L43,6.505z", "M20,4h23v4H16V0h4V4z"]
  },
  "android_nexus_scale9_w": {
    type: "SVG_PATHS",
    width: 43,
    height: 554,
    colors: [0xffffff, 0x010101, 0x010101, 0x010101, 0x010101, 0xffffff, 0x010101, 0x010101],
    paths: ["M20,0v554H3v-13.544V0H20z", "M22,0v554h-4V0H22z", "M5,0v554H1V0H5z", "M22,0v554h-4V0H22z", "M7.437,0l1.129,126.899L7,543.33V554H3v-10.678l1.566-416.413L3.437,0H7.437z", "M17-11.074v575.287", "M19,0v554h-4V0H19z", "M20,0v554h-4V0H20z"]
  },
  "android_nexus_scale9_c": {
    type: "SVG_PATHS",
    width: 275,
    height: 554,
    colors: [],
    paths: []
  },
  "android_nexus_button": {
    type: "SVG_PATHS",
    width: 131,
    height: 16,
    colors: [0x010101],
    paths: ["M65.5,15.409l-0.061-0.002l-43.424-1.318c-0.005,0-0.011,0-0.016,0c-3.649,0-21.923-0.322-21.923-6.53c0-5.004,5.14-6.529,22-6.529l40.752-0.438l46.115,0.438c0.045,0,0.092,0,0.137,0c12.208,0,21.844,0.023,21.844,6.529c0,4.49-5.54,6.139-21.952,6.53L65.5,15.409zM4.138,7.296c1.313,1.039,7.52,2.794,17.938,2.794L65.5,11.407l43.363-1.316c15.768-0.377,17.812-2.022,18.063-2.518c-0.003-2.544-10.407-2.544-18.003-2.544L62.831,4.592L22.098,5.029C7.797,5.029,4.592,6.347,4.138,7.296z"]
  },
  "android_nexus_ear_thing": {
    type: "SVG_PATHS",
    width: 177,
    height: 14,
    colors: [0x010101, 0x010101],
    paths: ["M58.851,13.151c-4.745,0-12.686,0-12.686-6.648c0-6.347,20.629-6.53,22.982-6.53l63.095,0.475l22.729-0.475c0.002,0,0.005,0,0.007,0c3.733,0,22.034,0.319,22.034,6.53c0,6.216-18.326,6.53-22,6.53H68.166c-1.748,0-3.444,0.032-5.059,0.062C61.613,13.124,60.188,13.151,58.851,13.151zM50.185,6.858c0.149,1.246,1.334,2.293,8.667,2.293c1.313,0,2.713-0.027,4.18-0.055c1.639-0.031,3.36-0.063,5.134-0.063h86.847c9.582,0,15.601-1.484,17.527-2.53c-1.926-1.047-7.945-2.53-17.527-2.53l-22.744,0.475L69.132,3.973C58.62,3.973,51.602,5.748,50.185,6.858z", "M6.517,14.027c-6.53,0-6.53-4.745-6.53-6.531c0-3.303,2.499-5.917,6.217-6.505c2.627-0.414,4.717,0.429,5.859,2.315c0.652,1.08,0.983,2.489,0.983,4.189C13.047,11.098,10.118,14.027,6.517,14.027zM7.51,4.884c-0.205,0-0.434,0.02-0.68,0.059c-1.312,0.207-2.842,1-2.842,2.554c0,1.995,0,2.531,2.53,2.531c1.395,0,2.53-1.136,2.53-2.531C9.047,4.884,8.122,4.884,7.51,4.884z"]
  },
  "datagrid_line_padding": 3,
  "datagrid_padding_top": 2,
  "datagrid_padding_bottom": 3,
  "datagrid_padding_left": 7,
  "datagrid_padding_right": 7,
  "datagrid_padding_middle": 6,
  "datagrid_selected_padding_left": 3,
  "datagrid_selected_padding_right": 3,
  "datagrid_scrollbar_padding_right": 2,
  "datagrid_scrollbar_padding_top": 4,
  "datagrid_scrollbar_padding_bottom": 2,
  "datagrid_vertical_separator_width": 2,
  "iphonepicker_line_color": 0x000000,
  "helpbutton": {
    type: "SVG_PATHS",
    width: 18,
    height: 18,
    colors: [0xffffff, null, 0x010101, 0x231f20],
    paths: ["M9.184,0.747c-4.889,1.668-8.039,5.834-7.35,9.755c0.678,3.848,2.566,5.881,5.459,5.881c3.929,0,8.334-3.645,9.062-7.499c0.63-3.322-1.647-6.206-6.418-8.138H9.184z", "M7.595,17.18c-1.638,0-5.562-0.646-6.618-6.636C0.214,6.21,3.523,1.678,8.675,0.002L8.833-0.05h1.365l0.186,0.072c7.198,2.825,7.173,7.374,6.827,9.199C16.423,13.387,11.84,17.18,7.595,17.18zM9.17,2.05C5.096,3.44,2.47,6.911,3.045,10.18c0.565,3.205,2.138,4.9,4.549,4.9c3.274,0,6.945-3.037,7.553-6.249c0.524-2.769-1.374-5.172-5.349-6.781H9.17z", "M8.595,11.865c-0.023,0-0.047-0.001-0.071-0.003c-0.482-0.038-0.841-0.46-0.802-0.942c0.138-1.719,0.989-2.465,1.61-3.01c0.525-0.461,0.788-0.691,0.787-1.424c-0.001-0.635-0.471-0.979-0.907-1.028C8.656,5.395,8.232,5.784,8.056,6.529c-0.11,0.472-0.582,0.766-1.052,0.652C6.534,7.07,6.242,6.6,6.353,6.129C6.79,4.271,8.198,3.58,9.412,3.719c1.42,0.164,2.454,1.326,2.457,2.765c0.003,1.526-0.766,2.2-1.383,2.741c-0.545,0.479-0.938,0.823-1.02,1.836C9.429,11.519,9.046,11.865,8.595,11.865z", "M9.484,13.359c0,0.437-0.357,0.791-0.797,0.791c-0.439,0-0.797-0.354-0.797-0.791c0-0.438,0.357-0.792,0.797-0.792C9.127,12.567,9.484,12.922,9.484,13.359z"]
  },
  "callout_default_color": 0xFFFF00,
  "callout_default_background_alpha": 0.5,
  "callout_padding_left": 14,
  "callout_padding_right": 12,
  "callout_padding_top": 9,
  "callout_padding_bottom": 6,
  "callout_skin_count": 9,
  "callout_1": {
    type: "SVG_SCALE9",
    scale9: [14, 16, 11, 14],
    width: 33,
    height: 39
  },
  "callout_1_nw": {
    type: "SVG_PATHS",
    width: 16,
    height: 14,
    paths: ["M6.417,14H1.347C3.33,7.522,7.729,1.173,16,0.138v4.585C12.109,6.271,8.913,10.291,6.417,14z"]
  },
  "callout_1_n": {
    type: "SVG_PATHS",
    width: 6,
    height: 14,
    paths: ["M0,4.723V0.138C1.044,0.007,2.142-0.049,3.312,0C4.276,0.041,5.167,0.146,6,0.3v3.781C4.835,3.999,3.709,4.016,2.74,4.082C1.791,4.147,0.878,4.373,0,4.723z"]
  },
  "callout_1_ne": {
    type: "SVG_PATHS",
    width: 11,
    height: 14,
    paths: ["M10.853,14H8.95c0.058-0.792,0.087-1.599,0.05-2.438C8.756,5.961,4.116,4.37,0,4.081V0.3C8.384,1.847,10.379,8.768,10.853,14z"]
  },
  "callout_1_e": {
    type: "SVG_PATHS",
    width: 11,
    height: 11,
    paths: ["M10.853,0C11.021,1.86,10.999,3.51,11,4.625c0.002,1.648-0.259,3.934-0.901,6.375H4.821C7.04,7.882,8.638,4.267,8.95,0H10.853z"]
  },
  "callout_1_se": {
    type: "SVG_PATHS",
    width: 11,
    height: 14,
    paths: ["M10.099,0C8.772,5.044,5.798,10.769,0,13.019V5.222c0.718-0.548,1.354-1.107,1.875-1.659C2.926,2.448,3.922,1.264,4.821,0H10.099z"]
  },
  "callout_1_s": {
    type: "SVG_PATHS",
    width: 6,
    height: 14,
    paths: ["M0,13.985V8.645c2.186-0.928,4.321-2.142,6-3.423v7.797C4.392,13.643,2.567,14,0.5,14C0.329,14,0.169,13.988,0,13.985z"]
  },
  "callout_1_sw": {
    type: "SVG_PATHS",
    width: 16,
    height: 14,
    paths: ["M16,8.645v5.341C6.336,13.845,0.788,9.685,0.082,0h3.646c-0.09,3.706,0.479,7.23,1.835,9C7.226,11.17,11.713,10.465,16,8.645z"]
  },
  "callout_1_w": {
    type: "SVG_PATHS",
    width: 16,
    height: 11,
    paths: ["M6.417,0C6.403,0.021,6.389,0.042,6.375,0.062C4.757,2.473,3.829,6.851,3.728,11H0.082C0.03,10.298,0,9.573,0,8.812C0,6.224,0.398,3.098,1.347,0H6.417z"]
  },
  "callout_1_c": {
    type: "SVG_PATHS",
    width: 6,
    height: 11,
    paths: []
  },
  "callout_2": {
    type: "SVG_SCALE9",
    scale9: [6, 19, 16, 11],
    width: 36,
    height: 39
  },
  "callout_2_nw": {
    type: "SVG_PATHS",
    width: 19,
    height: 6,
    paths: ["M17.065,6H9.443C12.189,3.158,15.439,1.058,19,0.459V5.09C18.351,5.355,17.705,5.662,17.065,6z"]
  },
  "callout_2_n": {
    type: "SVG_PATHS",
    width: 1,
    height: 6,
    paths: ["M0,5.09V0.459C0.331,0.404,0.664,0.36,1,0.331v4.381C0.666,4.827,0.333,4.953,0,5.09z"]
  },
  "callout_2_ne": {
    type: "SVG_PATHS",
    width: 16,
    height: 6,
    paths: ["M13.07,6h-1.704C9.255,4.31,6.71,3.812,4.938,3.812C3.321,3.812,1.656,4.138,0,4.712V0.331c0.396-0.034,0.793-0.054,1.195-0.05C7.266,0.347,10.935,2.604,13.07,6z"]
  },
  "callout_2_e": {
    type: "SVG_PATHS",
    width: 16,
    height: 22,
    paths: ["M13.07,0c1.954,3.107,2.625,7.166,2.625,11.406c0,3.312-0.151,7.121-1.26,10.594H9.422c3.474-3.275,5.766-7.716,5.766-12.5c0-4.868-1.689-7.793-3.821-9.5H13.07z"]
  },
  "callout_2_se": {
    type: "SVG_PATHS",
    width: 16,
    height: 11,
    paths: ["M14.435,0C12.758,5.254,8.886,9.741,0,10.578V5.063C3.503,4.283,6.799,2.473,9.422,0H14.435z"]
  },
  "callout_2_s": {
    type: "SVG_PATHS",
    width: 1,
    height: 11,
    paths: ["M0,10.647v-5.39c0.335-0.057,0.668-0.12,1-0.194v5.515C0.677,10.608,0.336,10.627,0,10.647z"]
  },
  "callout_2_sw": {
    type: "SVG_PATHS",
    width: 19,
    height: 11,
    paths: ["M16.188,5.5c0.943,0,1.883-0.086,2.812-0.242v5.39c-0.681,0.042-1.379,0.071-2.117,0.071C4.011,10.719,1.26,3.803,0.719,0h2.461C3.283,0.353,3.41,0.696,3.582,1.024C4.061,1.027,4.57,0.879,4.929,0h1.204C6.5,1.623,7.031,3.578,7.84,4.338C9.879,5.077,12.619,5.5,16.188,5.5z"]
  },
  "callout_2_w": {
    type: "SVG_PATHS",
    width: 19,
    height: 22,
    colors: [0x231f20, 0x231f20],
    paths: ["M17.065,0C12.786,2.259,8.804,6.05,6.167,10.311c-1.062,3.236,0.231,6.566-2.962,7.293c-0.171,0.912-0.267,1.818-0.267,2.709c0,0.588,0.087,1.149,0.243,1.688H0.719c-0.205-1.441-0.094-2.438-0.094-2.438C0.625,13.985,4.063,5.567,9.443,0H17.065z", "M6.132,22H4.929c0.166-0.408,0.303-0.958,0.384-1.75C5.483,18.579,5.69,20.05,6.132,22z"]
  },
  "callout_2_c": {
    type: "SVG_PATHS",
    width: 1,
    height: 22,
    paths: []
  },
  "callout_3": {
    type: "SVG_SCALE9",
    scale9: [15, 19, 11, 15],
    width: 41,
    height: 39
  },
  "callout_3_nw": {
    type: "SVG_PATHS",
    width: 19,
    height: 15,
    paths: ["M7.637,15H3.458C6.645,8.929,12.142,3.363,19,1.391v5.975C14.91,8.871,10.639,11.496,7.637,15z"]
  },
  "callout_3_n": {
    type: "SVG_PATHS",
    width: 11,
    height: 15,
    paths: ["M0,7.365V1.391c1.7-0.488,3.48-0.764,5.331-0.764c2.024,0,3.92,0.43,5.669,1.168v4.366C9.796,5.979,8.512,5.899,7.17,5.899C5.111,5.899,2.595,6.41,0,7.365z"]
  },
  "callout_3_ne": {
    type: "SVG_PATHS",
    width: 11,
    height: 15,
    paths: ["M10.401,15H8.449C8.028,9.176,4.636,6.866,0,6.161V1.795C5.197,3.991,9,9.089,10.401,15z"]
  },
  "callout_3_e": {
    type: "SVG_PATHS",
    width: 11,
    height: 9,
    paths: ["M10.401,0C10.786,1.621,11,3.3,11,5.001C11,6.392,10.877,7.727,10.636,9H6.558c1.26-2.423,1.937-5.065,1.937-7.738c0-0.438-0.016-0.857-0.045-1.262H10.401z"]
  },
  "callout_3_se": {
    type: "SVG_PATHS",
    width: 11,
    height: 15,
    paths: ["M10.636,0C9.561,5.684,6.069,10.13,0,12.659V7.172C2.892,5.271,5.111,2.78,6.558,0H10.636z"]
  },
  "callout_3_s": {
    type: "SVG_PATHS",
    width: 11,
    height: 15,
    paths: ["M0,14.927v-4.083c4.303-0.373,8.008-1.706,11-3.672v5.487C7.979,13.919,4.32,14.703,0,14.927z"]
  },
  "callout_3_sw": {
    type: "SVG_PATHS",
    width: 19,
    height: 15,
    paths: ["M16.544,10.961c0.841,0,1.655-0.048,2.456-0.117v4.083C18.163,14.97,17.31,15,16.421,15C8.267,15,0.419,10.471,0.419,1.948c0-0.632,0.048-1.285,0.122-1.948h3.126c-0.026,0.341-0.06,0.68-0.06,1.028C3.607,8.509,9.555,10.961,16.544,10.961z"]
  },
  "callout_3_w": {
    type: "SVG_PATHS",
    width: 19,
    height: 9,
    paths: ["M7.637,0c-2.193,2.56-3.713,5.586-3.97,9H0.541c0.322-2.877,1.342-5.999,2.917-9H7.637z"]
  },
  "callout_3_c": {
    type: "SVG_PATHS",
    width: 11,
    height: 9,
    paths: []
  },
  "callout_4": {
    type: "SVG_SCALE9",
    scale9: [13, 11, 17, 10],
    width: 36,
    height: 43
  },
  "callout_4_nw": {
    type: "SVG_PATHS",
    width: 11,
    height: 13,
    paths: ["M7.091,13H4.016C5.788,9.438,8.147,6.086,11,3.704v4.037C9.529,9.293,8.214,11.079,7.091,13z"]
  },
  "callout_4_n": {
    type: "SVG_PATHS",
    width: 8,
    height: 13,
    paths: ["M0,7.741V3.704c2.354-1.965,5.037-3.276,8-3.497v2.282C5.056,3.391,2.337,5.276,0,7.741z"]
  },
  "callout_4_ne": {
    type: "SVG_PATHS",
    width: 17,
    height: 13,
    paths: ["M16.808,13h-6.072c0.026-0.311,0.054-0.621,0.075-0.932c0.035-0.528,0.049-1.023,0.049-1.5C9.928,9.572,8.775,8.484,7.868,8.082c-2.76-1.225-0.736-2.391,0.98-0.98c0.481,0.395,1.207,0.922,1.889,1.4c-0.613-4.646-3.605-6.98-8.081-6.551C1.754,2.037,0.868,2.224,0,2.489V0.207c0.55-0.041,1.106-0.058,1.675-0.021C11.278,0.801,15.863,4.199,16.808,13z"]
  },
  "callout_4_e": {
    type: "SVG_PATHS",
    width: 17,
    height: 20,
    paths: ["M16.808,0C16.933,1.169,17,2.424,17,3.789C17,9.165,15.252,15.123,12.143,20H4.641c3.33-5.151,5.47-12.566,6.095-20H16.808z"]
  },
  "callout_4_se": {
    type: "SVG_PATHS",
    width: 17,
    height: 10,
    paths: ["M12.143,0C9.205,4.607,5.053,8.25,0,9.512V5.026C1.733,3.838,3.283,2.102,4.641,0H12.143z"]
  },
  "callout_4_s": {
    type: "SVG_PATHS",
    width: 8,
    height: 10,
    paths: ["M0,9.479V6.617C0.75,6.75,1.546,6.842,2.44,6.842c2.008,0,3.868-0.655,5.56-1.815v4.485C6.751,9.824,5.45,10,4.096,10C2.601,10,1.247,9.806,0,9.479z"]
  },
  "callout_4_sw": {
    type: "SVG_PATHS",
    width: 11,
    height: 10,
    paths: ["M11,6.617v2.862C5.861,8.135,2.774,4.281,1.324,0H3.71C4.638,3.023,6.658,5.85,11,6.617z"]
  },
  "callout_4_w": {
    type: "SVG_PATHS",
    width: 11,
    height: 20,
    paths: ["M2.957,14.273c0,1.638,0.14,3.725,0.754,5.727H1.324c-0.681-2.009-1.004-4.111-1.004-6.095V13.66C0.32,9.942,1.638,4.781,4.016,0h3.075C4.509,4.419,2.957,9.562,2.957,14.273z"]
  },
  "callout_4_c": {
    type: "SVG_PATHS",
    width: 8,
    height: 20,
    paths: []
  },
  "callout_5": {
    type: "SVG_SCALE9",
    scale9: [10, 19, 18, 16],
    width: 38,
    height: 40
  },
  "callout_5_nw": {
    type: "SVG_PATHS",
    width: 19,
    height: 10,
    paths: ["M9.104,10H4.452C7.893,5.513,13.185,2.196,19,1.104V5.16C15.035,5.405,11.64,7.213,9.104,10z"]
  },
  "callout_5_n": {
    type: "SVG_PATHS",
    width: 1,
    height: 10,
    paths: ["M0,5.16V1.104c0.333-0.062,0.664-0.139,1-0.187v4.191c-0.024,0-0.049-0.002-0.073-0.002C0.612,5.107,0.309,5.141,0,5.16z"]
  },
  "callout_5_ne": {
    type: "SVG_PATHS",
    width: 18,
    height: 10,
    paths: ["M16.708,10h-2.163c-0.587-0.726-1.597-1.333-2.584-1.793C9.188,6.003,4.449,5.119,0,5.109V0.918c1.028-0.146,2.069-0.225,3.117-0.225C9.64,0.693,14.514,4.661,16.708,10z"]
  },
  "callout_5_e": {
    type: "SVG_PATHS",
    width: 18,
    height: 14,
    paths: ["M16.708,0c0.831,2.021,1.289,4.233,1.29,6.51c0,2.909-0.315,5.381-0.902,7.49H8.652c1.791-1.986,3.241-4.183,4.351-6.277c1.009-1.906,1.415-3.553,1.363-4.973c0.953-1.036,0.815-1.963,0.179-2.75H16.708z"]
  },
  "callout_5_se": {
    type: "SVG_PATHS",
    width: 18,
    height: 16,
    paths: ["M17.095,0C14.824,8.167,8.426,10.812,0,11.59V5.979C3.452,4.787,6.336,2.569,8.652,0H17.095z"]
  },
  "callout_5_s": {
    type: "SVG_PATHS",
    width: 1,
    height: 16,
    paths: ["M0,11.668V6.281c0.337-0.095,0.673-0.189,1-0.303v5.611C0.671,11.62,0.334,11.643,0,11.668z"]
  },
  "callout_5_sw": {
    type: "SVG_PATHS",
    width: 19,
    height: 16,
    paths: ["M14.444,6.924c1.615,0,3.127-0.24,4.556-0.643v5.387c-0.234,0.018-0.464,0.04-0.702,0.055C15.046,13.407,14.954,16,14.954,16s-4.656-4.293-8.227-6.271C6.134,9.401,5.806,8.896,5.631,8.35C3.154,6.613,1.246,3.92,0.637,0h3.642C4.687,5.255,10.327,6.924,14.444,6.924z"]
  },
  "callout_5_w": {
    type: "SVG_PATHS",
    width: 19,
    height: 14,
    paths: ["M4.24,13.398c0,0.209,0.024,0.402,0.04,0.602H0.637c-0.132-0.85-0.224-1.737-0.224-2.707c0-4.126,1.53-8.02,4.039-11.293h4.651C6.046,3.359,4.24,8.146,4.24,13.398z"]
  },
  "callout_5_c": {
    type: "SVG_PATHS",
    width: 1,
    height: 14,
    paths: []
  },
  "callout_6": {
    type: "SVG_SCALE9",
    scale9: [13, 13, 14, 10],
    width: 36,
    height: 37
  },
  "callout_6_nw": {
    type: "SVG_PATHS",
    width: 13,
    height: 13,
    paths: ["M4.067,13H0.535C1.858,7.279,5.745,2.465,13,0.858v2.101C8.227,4.771,5.385,8.87,4.067,13z"]
  },
  "callout_6_n": {
    type: "SVG_PATHS",
    width: 9,
    height: 13,
    paths: ["M0,2.959V0.858c1.673-0.37,3.504-0.591,5.546-0.591c1.167,0,2.323,0.063,3.454,0.191v1.993C7.95,2.159,6.806,1.99,5.546,1.99C3.464,1.99,1.621,2.343,0,2.959z"]
  },
  "callout_6_ne": {
    type: "SVG_PATHS",
    width: 14,
    height: 13,
    paths: ["M13.998,13h-6.13C7.234,7.728,4.634,3.746,0,2.452V0.459C7.14,1.271,13.224,4.799,13.998,13z"]
  },
  "callout_6_e": {
    type: "SVG_PATHS",
    width: 14,
    height: 14,
    paths: ["M13.998,0c0.055,0.58,0.086,1.181,0.086,1.809c0,3.664-0.698,8.184-2.54,12.191H4.881c2.001-3.383,3.144-7.474,3.144-11.362c0-0.91-0.056-1.789-0.158-2.638H13.998z"]
  },
  "callout_6_se": {
    type: "SVG_PATHS",
    width: 14,
    height: 10,
    paths: ["M11.544,0C9.44,4.58,5.839,8.489,0,9.648V5.425C1.94,4.08,3.588,2.187,4.881,0H11.544z"]
  },
  "callout_6_s": {
    type: "SVG_PATHS",
    width: 9,
    height: 10,
    paths: ["M0,9.283V7.257c0.77,0.211,1.559,0.322,2.357,0.322C4.811,7.579,7.05,6.776,9,5.425v4.224C7.851,9.877,6.617,10,5.291,10C3.445,10,1.672,9.75,0,9.283z"]
  },
  "callout_6_sw": {
    type: "SVG_PATHS",
    width: 13,
    height: 10,
    paths: ["M13,7.257v2.026C8.159,7.933,4.203,4.694,1.973,0h3.113C6.881,3.694,9.767,6.368,13,7.257z"]
  },
  "callout_6_w": {
    type: "SVG_PATHS",
    width: 13,
    height: 14,
    paths: ["M3.176,5.5c0,3.144,0.716,6.045,1.91,8.5H1.973C0.715,11.353,0,8.246,0,4.742C0,3.124,0.181,1.531,0.535,0h3.532C3.461,1.9,3.176,3.807,3.176,5.5z"]
  },
  "callout_6_c": {
    type: "SVG_PATHS",
    width: 9,
    height: 14,
    paths: []
  },
  "callout_7": {
    type: "SVG_SCALE9",
    scale9: [23, 24, 12, 15],
    width: 39,
    height: 39
  },
  "callout_7_nw": {
    type: "SVG_PATHS",
    width: 24,
    height: 23,
    paths: ["M24,5.367V0.36c-0.486-0.022-0.979-0.051-1.441-0.051c-5.105,0-21.802,4.024-21.802,20.24c0,0.862,0.047,1.668,0.113,2.45h5.023C8.339,16.261,16.445,6.815,24,5.367zM9.825,9.318c4.955-4.955,12.072-5.886,12.072-5.886s-6.036,3.153-9.189,6.306c-3.153,3.154-8.048,10.15-8.048,10.15L4.3,19.5C4.3,19.5,4.871,14.273,9.825,9.318z"]
  },
  "callout_7_n": {
    type: "SVG_PATHS",
    width: 3,
    height: 23,
    paths: ["M0,5.367V0.36C0.982,0.406,1.987,0.482,3,0.64v4.553C2.647,5.18,2.289,5.174,1.922,5.174C1.288,5.174,0.646,5.243,0,5.367z"]
  },
  "callout_7_ne": {
    type: "SVG_PATHS",
    width: 12,
    height: 23,
    paths: ["M11.603,23H6.721c1.281-2.675,2.051-5.753,2.051-9.178c0-6.805-3.582-8.448-8.771-8.63V0.64c2.789,0.432,5.517,1.34,7.21,3.033c4.204,4.204,4.864,6.427,4.864,12.853C12.074,18.94,11.901,21.076,11.603,23z"]
  },
  "callout_7_e": {
    type: "SVG_PATHS",
    width: 12,
    height: 1,
    paths: ["M11.603,0c-0.052,0.336-0.1,0.678-0.16,1H6.212C6.39,0.674,6.559,0.339,6.721,0H11.603z"]
  },
  "callout_7_se": {
    type: "SVG_PATHS",
    width: 12,
    height: 15,
    paths: ["M11.443,0C9.749,9.098,4.879,12.703,0,14.111V7.016C2.44,5.257,4.63,2.89,6.212,0H11.443z"]
  },
  "callout_7_s": {
    type: "SVG_PATHS",
    width: 3,
    height: 15,
    paths: ["M0,14.734V8.804c1.023-0.501,2.032-1.091,3-1.788v7.096C1.991,14.402,0.981,14.601,0,14.734z"]
  },
  "callout_7_sw": {
    type: "SVG_PATHS",
    width: 24,
    height: 15,
    paths: ["M16.853,10.604c2.287,0,4.762-0.632,7.147-1.8v5.931C22.382,14.954,20.84,15,19.5,15C13.011,15,2.519,13.083,0.96,0h4.596C5.287,0.93,5.141,1.794,5.141,2.556C5.141,7.31,11.507,10.604,16.853,10.604z"]
  },
  "callout_7_w": {
    type: "SVG_PATHS",
    width: 24,
    height: 1,
    paths: ["M5.894,0C5.771,0.339,5.649,0.677,5.556,1H0.96C0.921,0.677,0.898,0.335,0.87,0H5.894z"]
  },
  "callout_7_c": {
    type: "SVG_PATHS",
    width: 3,
    height: 1,
    paths: []
  },
  "callout_8": {
    type: "SVG_SCALE9",
    scale9: [7, 23, 21, 9],
    width: 45,
    height: 40
  },
  "callout_8_nw": {
    type: "SVG_PATHS",
    width: 23,
    height: 7,
    paths: ["M12.479,7H6.706c3.469-3.823,8.377-6.451,14.829-6.451c0.472,0,0.964,0.049,1.465,0.121v2.427c-0.168-0.006-0.327-0.024-0.5-0.024C18.963,3.072,15.462,4.599,12.479,7z"]
  },
  "callout_8_n": {
    type: "SVG_PATHS",
    width: 1,
    height: 7,
    paths: ["M0,3.097V0.67c0.33,0.047,0.661,0.097,1,0.172v2.305C0.677,3.117,0.337,3.109,0,3.097z"]
  },
  "callout_8_ne": {
    type: "SVG_PATHS",
    width: 21,
    height: 7,
    paths: ["M11.167,7H8.221C6.25,4.964,3.592,3.472,0,3.146V0.842C3.532,1.619,7.576,3.916,11.167,7z"]
  },
  "callout_8_e": {
    type: "SVG_PATHS",
    width: 21,
    height: 24,
    paths: ["M21,16.913c0,2.421-0.982,4.862-2.634,7.087H9.731c2.52-2.512,3.983-5.479,3.983-8.296c0-2.618-0.716-10.769-5.494-15.704h2.946C16.603,4.667,21,11.141,21,16.913z"]
  },
  "callout_8_se": {
    type: "SVG_PATHS",
    width: 21,
    height: 9,
    paths: ["M18.366,0C14.609,5.062,7.372,9,0.237,9C0.16,9,0.079,8.999,0,8.999V5.401C3.984,4.366,7.335,2.389,9.731,0H18.366z"]
  },
  "callout_8_s": {
    type: "SVG_PATHS",
    width: 1,
    height: 9,
    paths: ["M0,8.99V5.628c0.335-0.072,0.673-0.142,1-0.227v3.598C0.675,8.998,0.34,8.995,0,8.99z"]
  },
  "callout_8_sw": {
    type: "SVG_PATHS",
    width: 23,
    height: 9,
    paths: ["M17.811,6.187c1.815,0,3.547-0.205,5.189-0.559V8.99C15.764,8.885,5.437,7.714,1.697,0h4.775C8.424,3.269,11.871,6.187,17.811,6.187z"]
  },
  "callout_8_w": {
    type: "SVG_PATHS",
    width: 23,
    height: 24,
    paths: ["M4.117,15.952c0,1.548,0.508,4.954,2.355,8.048H1.697c-0.901-1.858-1.424-4.092-1.424-6.787c0-5.746,2.083-12.42,6.433-17.213h5.773C7.607,3.921,4.117,10.178,4.117,15.952z"]
  },
  "callout_8_c": {
    type: "SVG_PATHS",
    width: 1,
    height: 24,
    paths: []
  },
  "callout_9": {
    type: "SVG_SCALE9",
    scale9: [13, 12, 15, 10],
    width: 37,
    height: 39
  },
  "callout_9_nw": {
    type: "SVG_PATHS",
    width: 12,
    height: 13,
    paths: ["M7.022,13H3.027C5.084,8.799,8.208,4.905,12,2.497v6.072C10.104,9.8,8.428,11.335,7.022,13z"]
  },
  "callout_9_n": {
    type: "SVG_PATHS",
    width: 10,
    height: 13,
    paths: ["M0,8.569V2.497C2.448,0.941,5.169,0,8.057,0C8.709,0,9.358,0.066,10,0.181v5.504c-0.207-0.012-0.41-0.03-0.622-0.03C5.91,5.654,2.73,6.798,0,8.569z"]
  },
  "callout_9_ne": {
    type: "SVG_PATHS",
    width: 15,
    height: 13,
    paths: ["M13.556,13H8.329C7.061,8.404,4.046,5.916,0,5.685V0.181C5.851,1.224,11.115,6.804,13.556,13z"]
  },
  "callout_9_e": {
    type: "SVG_PATHS",
    width: 15,
    height: 16,
    paths: ["M13.556,0C14.469,2.319,15,4.722,15,7.009c0,3.03-0.724,6.162-2.27,8.991H5.226c2.335-3.034,3.763-6.739,3.763-10.613c0-2.034-0.23-3.83-0.66-5.387H13.556z"]
  },
  "callout_9_se": {
    type: "SVG_PATHS",
    width: 15,
    height: 10,
    paths: ["M12.73,0C10.447,4.178,6.353,7.69,0,9.19V4.723C2.007,3.483,3.788,1.868,5.226,0H12.73z"]
  },
  "callout_9_s": {
    type: "SVG_PATHS",
    width: 10,
    height: 10,
    paths: ["M0,9.749V7.356c0.299,0.022,0.596,0.049,0.91,0.049c3.25,0,6.372-1.005,9.09-2.683V9.19C7.867,9.694,5.487,9.977,2.832,9.977C1.834,9.977,0.897,9.889,0,9.749z"]
  },
  "callout_9_sw": {
    type: "SVG_PATHS",
    width: 12,
    height: 10,
    paths: ["M12,7.356v2.393C6.259,8.854,2.573,5.203,1.095,0h1.998C4.399,3.964,7.31,7.006,12,7.356z"]
  },
  "callout_9_w": {
    type: "SVG_PATHS",
    width: 12,
    height: 16,
    paths: ["M2.279,10.732c0,1.833,0.273,3.627,0.813,5.268H1.095c-0.489-1.721-0.737-3.611-0.737-5.628C0.357,7.143,1.33,3.467,3.027,0h3.995C4.038,3.534,2.279,7.659,2.279,10.732z"]
  },
  "callout_9_c": {
    type: "SVG_PATHS",
    width: 10,
    height: 16,
    paths: []
  },
  "stickynote_padding_top": 31,
  "stickynote_padding_bottom": 10,
  "stickynote_padding_left": 10,
  "stickynote_padding_right": 14,
  "stickynote_brightness_offset": 90,
  "stickynote_paper": {
    type: "SVG_SCALE9",
    scale9: [8, 5, 10, 6],
    width: 105,
    height: 106
  },
  "stickynote_paper_nw": {
    type: "SVG_PATHS",
    width: 5,
    height: 8,
    colors: [0x5c5c5c],
    paths: ["M5,8H3.003c0.079-3.457,0.203-7.251,0.203-7.251S4.056,0.673,5,0.604V8z"]
  },
  "stickynote_paper_n": {
    type: "SVG_PATHS",
    width: 90,
    height: 8,
    colors: [0x5c5c5c, 0x4f4f4f, 0x404040, 0x232323],
    paths: ["M90.125,8H0V0.604C0.757,0.55,1.575,0.5,2.062,0.5c3.312,0,42.224,0.576,48.062,1.062c3.048,0.254,28.889,1.045,40,1.466V8z", "M45.281,8H21.625c-0.224-1.222-1.3-0.642-2.361,0h-2.892c-0.826-1.609-0.538-5.002-0.538-5.002s28.833,2.834,29.333,4.167C45.199,7.25,45.237,7.538,45.281,8z", "M89.82,8H24.368c-0.484-1.484-1.313-2.877-2.493-2.877c-2,0-9.625-3.748-9.625-3.748S36.876,2.623,38.709,2.29s11.162-0.569,12.328-0.616c4.846-0.193,21.281,1.7,23.336,1.867c2.343,0.191,12.75-0.249,14.75,0.918C89.869,4.894,89.977,6.348,89.82,8z", "M90.125,8h-4.448c1.439-1.416,0.905-4.416,0.905-4.416l3.543,0.265V8z"]
  },
  "stickynote_paper_ne": {
    type: "SVG_PATHS",
    width: 10,
    height: 8,
    colors: [0x5c5c5c, 0x232323],
    paths: ["M9.589,8H0.125V3.028C2.684,3.126,4.462,3.203,5,3.25c1.438,0.125,4.541,0.271,4.541,0.271S9.519,5.756,9.582,7C9.592,7.199,9.594,7.544,9.589,8z", "M8.759,8H0.125V3.849L8.834,4.5L8.759,8z"]
  },
  "stickynote_paper_e": {
    type: "SVG_PATHS",
    width: 10,
    height: 92,
    colors: [0x5c5c5c, 0x232323, 0x383838],
    paths: ["M9.589,0C9.537,4.712,8.727,21.968,8.25,28c-0.342,4.327-0.372,53.568-0.375,64h-7.75V0H9.589z", "M8.759,0l-1.98,92H0.125V0H8.759z", "M6,57.165c0,0-0.333,2.5-3,2.333s-3.166-2-2.333-3.333s4-5.333,5.167-5.167S6,57.165,6,57.165z"]
  },
  "stickynote_paper_se": {
    type: "SVG_PATHS",
    width: 10,
    height: 6,
    colors: [0x5c5c5c, 0x232323],
    paths: ["M7.875,0c0,1.019,0,1.675,0,1.875c0,1.24-0.433,3.844-0.433,3.844s-3.04,0.156-4.567,0.156c-0.347,0-1.333-0.008-2.75-0.021V0H7.875z", "M6.778,0L6.667,5.166c0,0-2.519-0.06-6.542-0.16V0H6.778z"]
  },
  "stickynote_paper_s": {
    type: "SVG_PATHS",
    width: 90,
    height: 6,
    colors: [0x5c5c5c, 0x232323],
    paths: ["M90.125,0v5.854C81.595,5.775,56.912,5.5,54.875,5.5c-11.062,0-34.562-2.312-38.75-2.438C12.888,2.966,3.614,2.617,0,2.523V0H90.125z", "M90.125,0v5.006C75.097,4.632,38.996,3.691,34,3.166C32.584,3.018,8.032,2.158,0,1.699V0H90.125z"]
  },
  "stickynote_paper_sw": {
    type: "SVG_PATHS",
    width: 5,
    height: 6,
    colors: [0x5c5c5c, 0x232323],
    paths: ["M5,0v2.523C4.427,2.509,3.992,2.5,3.75,2.5c-1.003,0-3.46-0.209-3.46-0.209S0.243,1.391,0.187,0H5z", "M5,0v1.699C3.588,1.618,2.675,1.55,2.5,1.5C1.635,1.252,1.229,0.819,1.145,0H5z"]
  },
  "stickynote_paper_w": {
    type: "SVG_PATHS",
    width: 5,
    height: 92,
    colors: [0x5c5c5c, 0x4f4f4f, 0x232323, 0x383838],
    paths: ["M5,0v92H0.187C0.104,89.927,0,86.751,0,83.75c0-7.375,1.434-37.301,1.688-40.562c0.438-5.625,1.438-26.312,1.438-31.875C3.125,7.311,2.972,1.375,3,0.125C3.001,0.084,3.002,0.042,3.003,0H5z", "M5,6.179v53.271c-0.824-0.052-1.504-0.302-1.666-0.951c-0.333-1.334-0.5-25-0.167-28.166s1-23.334,1.667-24C4.876,6.29,4.932,6.238,5,6.179z", "M5,92H1.145c-0.029-0.284-0.021-0.613,0.022-1c0.167-1.5,0.5-16.5,0-18.333s0-7.333,0-8.5S3,61,4.334,59.834C4.465,59.719,4.696,59.605,5,59.495V92z", "M5,58.947v4.276c-1.232,0.012-2.03-0.148-2.167-0.557C2,60.167,2.667,59.5,2.667,59.5C3.109,59.344,3.927,59.156,5,58.947z"]
  },
  "stickynote_paper_c": {
    type: "SVG_PATHS",
    width: 90,
    height: 92,
    colors: [0x5c5c5c, 0x4f4f4f, 0x4f4f4f, 0x404040, 0x232323, 0x383838, 0x383838, 0x383838],
    paths: ["M90.125,0v92H0V0H90.125z", "M16.372,0h2.892c-0.671,0.406-1.333,0.832-1.763,0.832C16.988,0.832,16.628,0.499,16.372,0z", "M45.281,0c0.651,6.817,2.532,51.916,2.219,53.165c-0.333,1.334-12.166,6-13.166,6.167s-30.167-0.5-31.667-0.167C1.897,59.336,0.868,59.503,0,59.449V6.179c1.021-0.897,4.839-3.681,6.667-3.681c3.333,0,4.333,3.834,9,3.834c2.167,0,6.064-4.668,6-6C21.661,0.204,21.643,0.102,21.625,0H45.281z", "M89.82,0c-0.264,2.78-1.278,6.123-1.278,6.123L46.875,55.75c-1.447,1.449-3.375-0.625-5.625-3c-1.17-1.234-13.541-3.126-15.541-2.293c-2.688,1.119-14.167,3-16.167,3.5s-8.167,1-8.5-1.167s2.834-6.666,5.167-6.833s9.833-0.167,10.5-1.834c0.667-1.666-8-8.166-7.833-10c0.167-1.833,5.464-2.24,7.5-3.166c3.667-1.667,0.833-5.834,3-7.334c0.998-0.689,3.834-0.166,5.334-1.5c1.5-1.333,4-6.333,5.5-6.833s7-0.167,8-1.167s0.833-3.666-1.167-3.666s-4.833-2.666-6.333-2.5s-5.834-3.5-5.834-5.167c0-0.684-0.171-1.759-0.507-2.79H89.82z", "M90.125,0v92H0V59.495c2.782-1.015,12.015-1.828,13.667-1.828c1.833,0,13.5-1.833,15.5-2s11.667-1.5,13.167-1.333s3.833-1.167,3-2.5s-6-2.834-7.667-3.5s-5.167-1.667-4.833-2.667S38.5,43,41.667,43s2.5-0.666,2-1.333s-0.167-3,0.667-3.5s2-5.5,2.333-8.167c0.333-2.666,2.833-6.833,5.167-6.833S60.5,22,60.834,20S66,14.167,67.5,13.334C69,12.5,71,9,72.5,7.667s7.334-1.167,7.334-2.5s4-3.667,5.666-5c0.064-0.052,0.119-0.11,0.177-0.167H90.125z", "M55,54.165c1.833-0.5,4,1,6,1.833s3.667,1.834-0.666,5.834S55.5,66.998,55.5,66.998s-1.666,2-3.166,0.167s-0.667-8-0.5-9.167S53.167,54.665,55,54.165z", "M60.167,50.165c0,0,3-0.667,5.333-0.5s4.834,0.833,6.667,2.5s-0.833,2-0.833,2s-1.334,1.333-2.834,1s-5.166-1-7-2.167S60.167,50.165,60.167,50.165z", "M40.5,56.667c-2,0.167-13.833,1.499-15.166,1.666C24.218,58.473,6.311,63.159,0,63.224v-4.276c5.808-1.126,19.067-2.832,20.334-3.114c1.5-0.334,21.07-2.476,22.236-2.309S42.5,56.5,40.5,56.667z"]
  },
  "stickynote_tape": {
    type: "SVG_SCALE9",
    scale9: [1, 4, 5, 2],
    width: 47,
    height: 27
  },
  "stickynote_tape_nw": {
    type: "SVG_PATHS",
    width: 4,
    height: 1,
    paths: []
  },
  "stickynote_tape_n": {
    type: "SVG_PATHS",
    width: 38,
    height: 1,
    colors: [0xe0554b],
    paths: ["M38,1h-2.202L38,0.711V1z"]
  },
  "stickynote_tape_ne": {
    type: "SVG_PATHS",
    width: 5,
    height: 1,
    colors: [0xe0554b],
    paths: ["M2.783,1H0V0.711l2.799-0.367C2.799,0.344,2.791,0.614,2.783,1z"]
  },
  "stickynote_tape_e": {
    type: "SVG_PATHS",
    width: 5,
    height: 24,
    colors: [0xe0554b, 0xdd3127],
    paths: ["M2.781,2.156C2.938,6,3.225,7.986,3.203,10.781c-0.031,4.031-0.391,10.125-0.391,10.125L0,21.174V0h2.783C2.77,0.632,2.758,1.574,2.781,2.156z", "M2.562,20.719L0,20.967v-9.221l3.062,0.004L2.562,20.719z"]
  },
  "stickynote_tape_se": {
    type: "SVG_PATHS",
    width: 5,
    height: 2,
    paths: []
  },
  "stickynote_tape_s": {
    type: "SVG_PATHS",
    width: 38,
    height: 2,
    colors: [0xe0554b, 0xdd3127],
    paths: ["M0,0.793V0h8.326L0,0.793z", "M0,0.65V0h6.709L0,0.65z"]
  },
  "stickynote_tape_sw": {
    type: "SVG_PATHS",
    width: 4,
    height: 2,
    colors: [0xe0554b, 0xdd3127],
    paths: ["M4,0v0.793L2.812,0.906c0,0-0.106-0.348-0.267-0.906H4z", "M4,0v0.65l-1.031,0.1c0,0-0.076-0.285-0.196-0.75H4z"]
  },
  "stickynote_tape_w": {
    type: "SVG_PATHS",
    width: 4,
    height: 24,
    colors: [0xe0554b, 0xdd3127],
    paths: ["M4,24H2.545c-0.57-1.98-1.857-6.77-1.827-9.5C0.781,8.938,1,5.094,1,5.094L4,4.7V24z", "M4,24H2.772c-0.505-1.944-1.816-7.113-1.866-8.5c-0.062-1.719,0.031-3.812,0.031-3.812L4,11.691V24z"]
  },
  "stickynote_tape_c": {
    type: "SVG_PATHS",
    width: 38,
    height: 24,
    colors: [0xe0554b, 0xdd3127, 0xdd3127, 0xdd3127],
    paths: ["M38,0v21.174L8.326,24H0V4.7L35.798,0H38z", "M6.709,24H0V11.691l38,0.055v9.221L6.709,24z", "M27.875,2.167C28,2.5,27.667,3.5,25.834,3.667s-1.459-1.625-1.459-1.625C25.125,1.5,27.75,1.834,27.875,2.167z", "M36.625,0.584C37,1.25,35.755,2.336,34.959,2.25c-1.167-0.125-1.334-1.541-1.334-1.541C34.25,0,36.25-0.083,36.625,0.584z"]
  },
  "arrow_thickness": 4,
  "arrow_thickness_dot": 1,
  "arrow_default_color": 0x000000,
  "arrow_padding_left": 3,
  "arrow_padding_right": 3,
  "arrow_point_origin_x": 1.9,
  "arrow_point_origin_y": 1.9,
  "arrow_point": {
    type: "SVG_PATHS",
    width: 16,
    height: 16,
    paths: ["M-1.826-1.796c-0.158,0.108-0.037,3.418,0.229,5.938C-1.452,5.512-0.815,6.6-0.61,7.189c0.43,1.246,1.875,3.908,2.66,5.168c0.396,0.639,1.604,0.18,1.604,0.18l1.944-1.66c0,0-0.549-4.693-0.121-5.213c0.428-0.52,4.802-0.093,4.802-0.093l2.783-3.854c0,0-0.976-1.21-2.108-1.498C8.814-0.325,5.042-1.299,2.021-1.619C0.995-1.728-1.437-2.061-1.826-1.796z"]
  },
  "popover_radius": 15,
  "popover_radius_margin": 5,
  "popover_thickness": 3,
  "popover_cover1_padding": 6,
  "popover_cover2_offset": 2,
  "popover_cover2_padding": 4,
  "popover_text_padding": 8,
  "ipad": {
    type: "SVG_PATHS",
    width: 616,
    height: 916,
    colors: [0x454545, 0x000000, 0x000000, 0x000000, 0x000000],
    paths: ["M599.676,5.5c6.531,0,11.824,5.314,11.824,11.868v882.263c0,6.555-5.293,11.869-11.824,11.869H16.325c-6.53,0-11.825-5.314-11.825-11.869V17.368C4.5,10.814,9.795,5.5,16.325,5.5H599.676zM31,92v733h552V92H31z", "M614,17.368v882.263c0,7.923-6.426,14.369-14.324,14.369H16.325C8.426,914,2,907.554,2,899.631V17.368C2,9.445,8.426,3,16.325,3h583.351C607.574,3,614,9.445,614,17.368zM7,899.631c0,5.166,4.183,9.369,9.325,9.369h583.351c5.142,0,9.324-4.203,9.324-9.369V17.368C609,12.202,604.817,8,599.676,8H16.325C11.183,8,7,12.202,7,17.368V899.631zM585,89.5v738H28v-738H585zM580,822V94H33v728H580z", "M308.087,38.196c3.001,0,5.435,2.434,5.435,5.435s-2.434,5.435-5.435,5.435c-3.002,0-5.435-2.434-5.435-5.435S305.085,38.196,308.087,38.196z", "M334.5,866.729c0,14.414-11.728,26.141-26.142,26.141s-26.141-11.727-26.141-26.141s11.727-26.142,26.141-26.142S334.5,852.314,334.5,866.729zM287.217,866.729c0,11.657,9.484,21.141,21.141,21.141s21.142-9.483,21.142-21.141c0-11.658-9.484-21.142-21.142-21.142S287.217,855.07,287.217,866.729z", "M318.739,861.305v10.848c0,3.033-2.468,5.5-5.5,5.5h-10.305c-3.033,0-5.5-2.467-5.5-5.5v-10.848c0-3.033,2.467-5.5,5.5-5.5h10.305C316.271,855.805,318.739,858.271,318.739,861.305zM302.435,872.152c0,0.275,0.224,0.5,0.5,0.5h10.305c0.275,0,0.5-0.225,0.5-0.5v-10.848c0-0.276-0.225-0.5-0.5-0.5h-10.305c-0.276,0-0.5,0.224-0.5,0.5V872.152z"]
  },
  "ipad_scale9": {
    type: "SVG_SCALE9",
    scale9: [96, 37, 38, 97],
    width: 616,
    height: 916
  },
  "ipad_scale9_nw": {
    type: "SVG_PATHS",
    width: 37,
    height: 96,
    colors: [0x454545, null, null],
    paths: ["M31,96H4.5V17.368C4.5,10.814,9.795,5.5,16.325,5.5H37V92h-6V96z", "M33,96h-5v-6.5h9V94h-4V96z", "M7,96H2V17.368C2,9.445,8.426,3,16.325,3H37v5H16.325C11.183,8,7,12.202,7,17.368V96z"]
  },
  "ipad_scale9_n": {
    type: "SVG_PATHS",
    width: 541,
    height: 96,
    colors: [0x454545, null, null],
    paths: ["M0,92V5.5h541V92H0z", "M0,94v-4.5h541V94H0z", "M0,8V3h541v5H0z"]
  },
  "ipad_scale9_ne": {
    type: "SVG_PATHS",
    width: 38,
    height: 96,
    colors: [0x454545, null, null],
    paths: ["M33.5,96H5v-4H0V5.5h21.676c6.531,0,11.824,5.314,11.824,11.868V96z", "M36,96h-5V17.368C31,12.202,26.817,8,21.676,8H0V3h21.676C29.574,3,36,9.445,36,17.368V96z", "M7,96H2v-2H0v-4.5h7V96z"]
  },
  "ipad_scale9_e": {
    type: "SVG_PATHS",
    width: 38,
    height: 723,
    colors: [0x454545, null, null],
    paths: ["M33.5,0v723H5V0H33.5z", "M36,0v723h-5V0H36z", "M7,0v723H2V0H7z"]
  },
  "ipad_scale9_se": {
    type: "SVG_PATHS",
    width: 38,
    height: 97,
    colors: [0x454545, null, null],
    paths: ["M33.5,0v80.631c0,6.555-5.293,11.869-11.824,11.869H0V6h5V0H33.5z", "M36,0v80.631C36,88.554,29.574,95,21.676,95H0v-5h21.676C26.817,90,31,85.797,31,80.631V0H36z", "M7,0v8.5H0V3h2V0H7z"]
  },
  "ipad_scale9_s": {
    type: "SVG_PATHS",
    width: 541,
    height: 97,
    colors: [0x454545, null, null],
    paths: ["M0,92.5V6h541v86.5H0z", "M0,8.5V3h541v5.5H0z", "M0,95v-5h541v5H0z"]
  },
  "ipad_scale9_sw": {
    type: "SVG_PATHS",
    width: 37,
    height: 97,
    colors: [0x454545, null, null],
    paths: ["M31,6h6v86.5H16.325C9.795,92.5,4.5,87.186,4.5,80.631V0H31V6z", "M33,3h4v5.5h-9V0h5V3z", "M16.325,90H37v5H16.325C8.426,95,2,88.554,2,80.631V0h5v80.631C7,85.797,11.183,90,16.325,90z"]
  },
  "ipad_scale9_w": {
    type: "SVG_PATHS",
    width: 37,
    height: 723,
    colors: [0x454545, null, null],
    paths: ["M31,0v723H4.5V0H31z", "M33,0v723h-5V0H33z", "M7,0v723H2V0H7z"]
  },
  "ipad_scale9_c": {
    type: "SVG_PATHS",
    width: 541,
    height: 723,
    paths: []
  },
  "ipad_button": {
    type: "SVG_PATHS",
    width: 54,
    height: 54,
    paths: ["M53.142,27c0,14.414-11.728,26.141-26.142,26.141S0.858,41.414,0.858,27S12.586,0.859,27,0.859S53.142,12.586,53.142,27zM5.858,27c0,11.658,9.484,21.141,21.142,21.141S48.142,38.658,48.142,27S38.657,5.859,27,5.859S5.858,15.342,5.858,27z", "M37.381,21.576v10.848c0,3.033-2.468,5.5-5.5,5.5H21.576c-3.033,0-5.5-2.467-5.5-5.5V21.576c0-3.033,2.467-5.5,5.5-5.5h10.305C34.913,16.076,37.381,18.543,37.381,21.576zM21.076,32.424c0,0.275,0.224,0.5,0.5,0.5h10.305c0.275,0,0.5-0.225,0.5-0.5V21.576c0-0.275-0.225-0.5-0.5-0.5H21.576c-0.276,0-0.5,0.225-0.5,0.5V32.424z"]
  },
  "ipad_speaker": {
    type: "SVG_PATHS",
    width: 12,
    height: 12,
    paths: ["M6,0.565c3.001,0,5.435,2.434,5.435,5.435S9.001,11.435,6,11.435c-3.002,0-5.435-2.434-5.435-5.435S2.998,0.565,6,0.565z"]
  },
  "lineOfText": {
    type: "SVG_PATHS",
    width: 4056,
    height: 72,
    colors: [0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000],
    paths: ["M423,1c2.333,0,4.667,0,7,0c5.727,2.94,10.462,6.871,9,17c4.907-7.903,33.229-21.244,36,1c5.166-3.564,7.902-13.415,15-15c12.882-2.877,21.318,6.958,17,22c31.659,14.105-11.922,52.83-34,34c-6.101,4.899-12.92,9.08-21,12c-3,0-6,0-9,0c-5.849-2.484-10.324-6.343-12-13c-10.186,7.593-29.738,4.71-25-12c-11.262,7.732-77.423,31.206-57-10c-17.128,4.085-44.488,34.099-65,16c-12.817,3.651-45.21,18.837-48-2c-0.828-6.185,1.934-8.805,4-14c-21.226,2.688-38.802,40.129-67,23c-6.101,4.899-12.92,9.08-21,12c-3,0-6,0-9,0c-6.291-1.709-8.524-7.476-13-11c-6.918,13.186-41.224,14.638-34-9c-12.642,2.35-46.449,20.391-51,1c-1.775-7.562,1.697-11.463,4-17c-12.618,1.209-21.368,19.83-38,17C5.2,52.013,0.702,46.933,1,40c0.444-10.331,19.33-16.176,31-23C46.852,8.315,58.089,0.457,72,2c14.797,1.642,16.917,17.068,10,27c16.195,0.151,23.575-26.865,40-27c13.482-0.111,14.03,14.59,10,26c6.228-14.365,35.544-34.363,44-9c2.97-17.144,37.609-22.795,32,6c12.548-3.215,34.24-25.093,55-23c10.941,1.103,19.155,14.381,11,25c3.452,1.125,7.529-5.342,13-10c10.36-8.821,33.705-18.275,35,5c10.68-5.874,58.444-35.597,64-9c1.467,7.025-1.396,10.157-3,16C400.294,23.627,409.672,10.339,423,1z", "M637,1c2.333,0,4.667,0,7,0c2.772,1.895,5.483,3.851,8,6c0.493,3.507,1.491,6.509,1,11c4.907-7.903,33.229-21.244,36,1c5.166-3.564,7.902-13.415,15-15c12.869-2.874,21.389,7.012,17,22c11.657-3.6,52.498-37.748,68-19c4.911,5.939,2.605,16.255-1,22c11.93-2.667,21.479,5.849,18,17c-4.66,14.936-76.666,28.402-51-10c-21.43,5.204-44.308,41.98-68,24c-6.101,4.899-12.92,9.08-21,12c-3,0-6,0-9,0c-5.98-2.687-10.899-6.434-12-14c-9.156,9.354-30.757,5.811-25-12c-11.098,4.061-23.416,14.941-36,14c-8.062-0.604-11.221-10.311-21-7c-5.219-3.862-10.634-10.162-7-18c5.081-10.961,52.327-48.799,54-13C620.487,16.487,627.429,7.429,637,1z", "M2307,1c2.334,0,4.666,0,7,0c5.727,2.94,10.462,6.871,9,17c4.907-7.903,33.229-21.244,36,1c5.166-3.564,7.902-13.415,15-15c12.882-2.877,21.318,6.958,17,22c31.659,14.105-11.923,52.83-34,34c-6.101,4.899-12.92,9.08-21,12c-3,0-6,0-9,0c-5.849-2.484-10.324-6.343-12-13c-10.187,7.593-29.738,4.71-25-12c-11.262,7.732-77.423,31.206-57-10c-17.164,4.334-45.447,33.879-66,16c-12.816,3.662-45.212,18.82-48-2c-0.828-6.185,1.934-8.805,4-14c-21.226,2.688-38.803,40.129-67,23c-6.101,4.899-12.92,9.08-21,12c-3,0-6,0-9,0c-4.819-2.515-9.777-4.89-11-11c-8.725,11.331-42.459,16.253-36-9c-17.757,8.817-65.869,19.529-47-16c-10.788,1.359-24.553,20.253-39,17c-6.646-1.496-12.421-12.112-6-21c14.833-3.144,42.667-32.028,67-30c13.513,1.126,19.048,16.201,11,27c15.083,0.432,23.55-24.759,37-27c14.457-2.408,19.621,12.35,12,25c6.191-8.671,35.909-36.215,44-8c4.934-14.987,36.617-25.455,33,6c13.951-11.082,81.902-44.184,66,3c6.747-8.361,43.204-38.821,48-5c13.761-12.607,78.706-39.508,62,6C2284.295,23.627,2293.672,10.339,2307,1z", "M2521,1c2.334,0,4.666,0,7,0c2.772,1.895,5.482,3.851,8,6c0.493,3.507,1.49,6.509,1,11c4.907-7.903,33.229-21.244,36,1c5.166-3.564,7.902-13.415,15-15c12.869-2.874,21.389,7.012,17,22c11.656-3.6,52.498-37.748,68-19c4.911,5.939,2.605,16.255-1,22c11.93-2.667,21.479,5.849,18,17c-4.66,14.936-76.666,28.402-51-10c-21.43,5.204-44.309,41.98-68,24c-6.101,4.899-12.92,9.08-21,12c-3,0-6,0-9,0c-5.98-2.687-10.899-6.434-12-14c-9.155,9.354-30.757,5.811-25-12c-11.098,4.061-23.416,14.941-36,14c-8.062-0.604-11.221-10.311-21-7c-5.219-3.862-10.634-10.162-7-18c5.081-10.961,52.327-48.799,54-13C2504.486,16.487,2511.429,7.429,2521,1z", "M1047,72c-3,0-6,0-9,0c-6.135-2.865-11.425-6.575-12-15c-16.896,10.979-48.731,2.722-33-20c-21.226,2.688-38.802,40.129-67,23c-6.101,4.899-12.92,9.08-21,12c-3,0-6,0-9,0c-6.135-2.865-11.425-6.575-12-15c-13.687,8.481-18.535-4.91-33-4c-5.512-4.206-10.614-10.203-7-18c3.688-7.956,28.583-30.182,45-28c2.422,2.244,5.675,3.659,7,7c7.804-9.029,30.828-12.227,32,5c4.933-14.987,36.618-25.455,33,6c14.61-8.205,72.035-44.796,68-2c5.983-7.878,35.053-30.162,41-4c5.166-3.564,7.902-13.415,15-15c12.882-2.877,21.319,6.958,17,22c31.659,14.105-11.922,52.83-34,34C1061.899,64.899,1055.08,69.08,1047,72z", "M1466,72c-6.135-2.865-11.425-6.575-12-15c-18.438,10.766-48.068,2.858-32-21c-12.079,5.737-39.837,23.758-61,26c-18.584,1.97-21.27-14.364-14-25c-12.981,2.348-46.32,30.758-69,25c-11.201-2.844-10.895-17.814-5-26c-10.576,5.174-36.141,20.837-55,25c-21.847,4.822-29.668-9.747-20-25c-10.788,1.359-24.553,20.253-39,17c-6.646-1.496-12.421-12.112-6-21c14.833-3.144,42.667-32.028,67-30c13.513,1.126,19.049,16.201,11,27c15.327-1.513,40.548-28.954,64-27c13.551,1.129,18.961,16.065,11,27c15.009-1.909,60.292-43.441,76-21c4.081,5.83,1.639,14.971-1,21c16.398-3.741,41.008-28.833,63-27c10.207,0.851,17.349,8.7,14,20c5.686-7.929,34.721-28.435,40-3c3.523-13.196,33.005-24.46,33,0c6.773-3.582,28.636-21.894,38-5c7.804-9.029,30.828-12.227,32,5c5.166-3.564,7.902-13.415,15-15c12.869-2.874,21.389,7.012,17,22c11.656-3.6,52.499-37.748,68-19c4.911,5.939,2.605,16.255-1,22c11.93-2.667,21.479,5.849,18,17c-4.702,15.07-77.296,27.766-51-10c-21.43,5.204-44.308,41.98-68,24c-6.101,4.899-12.92,9.08-21,12c-3,0-6,0-9,0c-6.135-2.865-11.425-6.575-12-15c-7.84,3.872-18.426,3.604-23-3c-6.674,8.177-27.497,16.682-38,6c-6.101,4.899-12.92,9.08-21,12C1472,72,1469,72,1466,72z", "M2932,72c-3,0-6,0-9,0c-6.135-2.865-11.425-6.575-12-15c-16.896,10.979-48.731,2.722-33-20c-21.226,2.688-38.803,40.129-67,23c-6.101,4.899-12.92,9.08-21,12c-3,0-6,0-9,0c-6.135-2.865-11.425-6.575-12-15c-13.687,8.481-18.535-4.91-33-4c-5.513-4.207-10.614-10.202-7-18c3.688-7.956,28.583-30.182,45-28c2.422,2.244,5.675,3.659,7,7c7.804-9.029,30.828-12.227,32,5c4.941-14.99,36.622-25.454,33,6c14.61-8.205,72.035-44.796,68-2c5.982-7.878,35.053-30.162,41-4c5.166-3.564,7.902-13.415,15-15c12.882-2.877,21.318,6.958,17,22c31.659,14.105-11.923,52.83-34,34C2946.899,64.899,2940.08,69.08,2932,72z", "M3351,72c-6.135-2.865-11.425-6.575-12-15c-18.438,10.766-48.068,2.858-32-21c-12.079,5.737-39.837,23.757-61,26c-18.584,1.97-21.27-14.364-14-25c-12.98,2.348-46.32,30.758-69,25c-11.201-2.844-10.895-17.814-5-26c-10.575,5.173-36.141,20.838-55,25c-21.847,4.822-29.668-9.747-20-25c-10.788,1.359-24.553,20.253-39,17c-6.646-1.496-12.422-12.112-6-21c14.833-3.144,42.666-32.028,67-30c13.513,1.126,19.049,16.201,11,27c15.327-1.513,40.549-28.955,64-27c13.551,1.129,18.961,16.065,11,27c15.008-1.909,60.294-43.442,76-21c4.08,5.83,1.64,14.971-1,21c16.398-3.741,41.008-28.833,63-27c10.207,0.851,17.35,8.7,14,20c5.686-7.929,34.721-28.435,40-3c3.523-13.196,33.005-24.46,33,0c6.773-3.582,28.636-21.894,38-5c7.804-9.029,30.828-12.227,32,5c5.166-3.564,7.902-13.415,15-15c12.869-2.875,21.389,7.012,17,22c11.654-3.599,52.501-37.75,68-19c4.91,5.94,2.606,16.254-1,22c11.93-2.667,21.479,5.849,18,17c-4.703,15.07-77.296,27.766-51-10c-21.43,5.204-44.309,41.98-68,24c-6.101,4.899-12.92,9.08-21,12c-3,0-6,0-9,0c-6.135-2.865-11.425-6.575-12-15c-7.84,3.872-18.427,3.604-23-3c-6.675,8.177-27.497,16.682-38,6c-6.101,4.899-12.92,9.08-21,12C3357,72,3354,72,3351,72z", "M4056,7c0,16.333,0,32.667,0,49c-8.462,4.95-20.98,4.98-26-3c-12.818,3.651-45.209,18.837-48-2c-0.828-6.185,1.934-8.805,4-14c-21.226,2.688-38.802,40.129-67,23c-6.101,4.899-12.92,9.08-21,12c-3,0-6,0-9,0c-4.818-2.515-9.777-4.89-11-11c-8.725,11.331-42.46,16.253-36-9c-12.064,2.024-45.673,20.152-51,2c-2.111-7.193,1.672-12.639,4-18c-10.788,1.359-24.553,20.253-39,17c-6.646-1.496-12.421-12.112-6-21c14.834-3.144,42.666-32.028,67-30c13.513,1.126,19.049,16.201,11,27c15.084,0.432,23.551-24.759,37-27c14.456-2.408,19.621,12.35,12,25c6.19-8.671,35.909-36.215,44-8c4.941-14.99,36.622-25.454,33,6c13.951-11.082,81.902-44.184,66,3C4030.24,23.236,4036.62,5.956,4056,7z", "M1781,72c-3,0-6,0-9,0c-8.802-1.638-8.093-9.909-11-17c-1.62-3.951-6.491-7.166-7-12c-1.556-14.762,40.478-56.954,50-25c7.019-2.19,8.45-12.354,16-14c13.168-2.87,22.097,8.591,16,22c32.805,12.88-10.505,53.002-33,34C1796.218,64.552,1789.858,69.524,1781,72z", "M3667,72c-3,0-6,0-9,0c-8.803-1.638-8.093-9.909-11-17c-1.62-3.951-6.49-7.166-7-12c-1.557-14.762,40.479-56.954,50-25c7.019-2.19,8.45-12.354,16-14c13.168-2.87,22.097,8.591,16,22c32.805,12.88-10.505,53.002-33,34C3682.219,64.552,3675.857,69.524,3667,72z"]
  },
  "timepicker_lineWeight": 3,
  "timepicker_backgroundColor": 0xFFFFFF,
  "timepicker_headingPanelBgColor": 0xDDDDDD,
  "timepicker_clockFaceColor": 0xEEEEEE,
  "timepicker_activeTextColor": 0x000000,
  "timepicker_inactiveTextColor": 0x999999,
  "timepicker_padding": 10,
  "timepicker_defaultHeadingPortrait": 95,
  "timepicker_defaultHeadingLandscape": 188,
  "timepicker_defaultWidthLandscape": 570,
  "timepicker_defaultHeightLandscape": 303,
  "timepicker_defaultWidthPortrait": 320,
  "timepicker_defaultHeightPortrait": 473,
  "timepicker_clockSizePortrait": 285,
  "timepicker_clockSizeLandscape": 250,
  "datepicker_iconSize": 24,
  "datepicker_color": 0,
  "datepicker_paddingDate": 10,
  "datepicker_lineWeightDate": 3,
  "datepicker_fontSize": 13,
  "datepicker_buttonFontSize": 14,
  "datepicker_headingPanelBgColor": 0xDDDDDD,
  "datepicker_defaultHeadingLandscape": 188,
  "datepicker_defaultWidthLandscape": 570,
  "datepicker_defaultHeightLandscape": 303,
  "datepicker_defaultCalendarHeightLandscape": 260,
  "datepicker_yearLabelFontSizeLandscape": 18,
  "datepicker_dateLabelFontSizeLandscape": 40,
  "datepicker_defaultHeadingPortrait": 95,
  "datepicker_defaultWidthPortrait": 320,
  "datepicker_defaultHeightPortrait": 473,
  "datepicker_defaultCalendarWidthPortrait": 290,
  "datepicker_yearLabelFontSizePortrait": 14,
  "datepicker_dateLabelFontSizePortrait": 32,
  "__last": null
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/skins/wireframe.js":
/*!************************************!*\
  !*** ./js/core/skins/wireframe.js ***!
  \************************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;
var _default = {
  "#skinName": "wireframe",
  "default_font_size": 13,
  "default_border_color": 0x333333,
  "default_disabled_color": 0x80808b,
  "default_selection_color": 0x97bdf0,
  "default_disabled_selection_color": 0xcbddf2,
  "default_icon_color": 0x333333,
  "default_text_color": 0x333333,
  "default_link_color": 0x085394,
  "default_splitter_color": 0x666666,
  "blockoftext_font_style": "",
  "accordion_padding_middle": 17,
  "accordion_padding_left": 4,
  "accordion_padding_right": 4,
  "accordion_overlap": 1,
  "accordion_padding_top": 6,
  "accordion_padding_bottom": 4,
  "accordion_scrollbar_padding_right": 0,
  "accordion_selected_subrow_padding": 6,
  "alert_separator_color": 0xcccccc,
  "alert_title_font_size": 18,
  "alert_body_font_size": 13,
  "alert_padding": 10,
  "alert_line_padding": 3,
  "alert_button_color": 0x333333,
  "arrow_thickness": 2,
  "arrow_thickness_dot": 1,
  "arrow_default_color": 0x333333,
  "arrow_padding_left": 3,
  "arrow_padding_right": 3,
  "arrow_point_origin_x": 2,
  "arrow_point_origin_y": 2,
  "arrow_point": {
    type: "SVG_PATHS",
    width: 14,
    height: 14,
    paths: ["M11.498,-2L-2,-2L-2,11.498L0.498,11.498L0.498,0.498L11.498,0.498z"]
  },
  "appbar_background_color": 0xDDDDDD,
  "appbar_padding_left": 13,
  "appbar_icon_padding": 13,
  "appbar_font_size": 16,
  "appbar_icon_top_padding": 14,
  "appbar_padding_bottom": 14,
  "barchart_axes_thickness": 2,
  "barchart_axes_color": 0x333333,
  "barchart_series1_color": 0xDFDFDF,
  "barchart_series2_color": 0x959595,
  "breadcrumbs_separator": {
    type: "SVG_PATHS",
    width: 8,
    height: 10,
    paths: ["M7.328,4.997L1.887,9l-0.76-1.037L5.173,5L1.127,2.037L1.887,1L7.328,4.997z"]
  },
  "breadcrumbs_padding_top": 4,
  "breadcrumbs_padding_bottom": 4,
  "breadcrumbs_padding_middle": 4,
  "browserwindow": {
    type: "SVG_SCALE9",
    scale9: [73, 123, 38, 4],
    width: 455,
    height: 404
  },
  "browserwindow_nw": {
    type: "SVG_PATHS",
    width: 123,
    height: 73,
    colors: [0xdededd, 0x333333, 0x231f20, 0x6b6b6b, 0x6b6b6b, 0x6b6b6b, 0xffffff, 0x6b6b6b, 0x6b6b6b, 0x231f20],
    paths: ["M123,1v71.5H1V3c0-1.104,0.896-2,2-2H123z", "M123,31.902v1H0v-1H123z", "M123,72v1H0v-1H123z", "M11.471,51.578l6.005-6.012c0.754-0.756,2.069-0.755,2.824,0c0.377,0.378,0.585,0.879,0.584,1.414c0,0.533-0.208,1.035-0.585,1.412l-2.597,2.6h7.191c1.101,0,1.997,0.896,1.997,1.998s-0.896,1.998-1.997,1.998h-7.23l2.626,2.59c0.784,0.773,0.793,2.041,0.021,2.825C19.932,60.788,19.427,61,18.888,61c-0.528,0-1.025-0.204-1.401-0.574l-6.006-5.922c-0.391-0.385-0.606-0.924-0.59-1.451C10.874,52.493,11.08,51.969,11.471,51.578z", "M51.999,53.053c0.016,0.527-0.2,1.066-0.59,1.451l-6.006,5.922C45.027,60.796,44.53,61,44.002,61c-0.539,0-1.044-0.212-1.423-0.597c-0.772-0.784-0.762-2.052,0.021-2.825l2.626-2.59h-7.23c-1.101,0-1.997-0.896-1.997-1.998s0.896-1.998,1.997-1.998h7.191l-2.597-2.6c-0.377-0.377-0.585-0.879-0.585-1.412c0-0.535,0.207-1.036,0.584-1.414c0.754-0.755,2.07-0.756,2.824,0l6.005,6.012C51.81,51.969,52.016,52.493,51.999,53.053z", "M73.385,45.227h0.002c2.185,0,4.243,0.914,5.721,2.479l2.386-1.38l0.006,8.726l-7.563-4.357l2.449-1.415c-0.848-0.663-1.896-1.043-2.999-1.043h-0.001c-2.689,0-4.877,2.188-4.877,4.878s2.188,4.879,4.877,4.879c1.526,0,2.936-0.695,3.87-1.906c0.506-0.656,1.451-0.779,2.109-0.271c0.658,0.507,0.78,1.451,0.273,2.109C78.13,59.879,75.852,61,73.386,61c-4.348,0-7.886-3.538-7.886-7.887C65.5,48.766,69.038,45.227,73.385,45.227z", "M123,41.5v23H99.5v-23H123z", "M100,64h23v1H99V41h24v1h-23V64z", "M120.054,58.408l-4.262-4.256c-0.009-0.008-0.021-0.01-0.03-0.02c0.585-0.947,0.903-2.037,0.903-3.179c0-1.625-0.633-3.152-1.783-4.302c-1.148-1.148-2.676-1.781-4.3-1.781s-3.152,0.633-4.301,1.781c-1.149,1.149-1.782,2.677-1.782,4.302c0,1.624,0.633,3.152,1.782,4.301c1.148,1.149,2.676,1.784,4.301,1.784c1.077,0,2.109-0.285,3.019-0.809c0.019,0.022,0.026,0.051,0.046,0.07l4.262,4.256c0.296,0.295,0.684,0.443,1.072,0.443c0.389,0,0.778-0.148,1.074-0.445C120.648,59.961,120.648,59,120.054,58.408zM107.713,53.823c-0.767-0.767-1.189-1.785-1.189-2.869c0-1.085,0.422-2.104,1.189-2.871c0.767-0.767,1.786-1.188,2.87-1.188s2.103,0.422,2.869,1.188c0.767,0.768,1.19,1.786,1.19,2.871c0,1.084-0.423,2.103-1.19,2.87C111.92,55.36,109.247,55.36,107.713,53.823z", "M2,73H0V3c0-1.654,1.346-3,3-3h120v2H3C2.449,2,2,2.448,2,3V73z"]
  },
  "browserwindow_n": {
    type: "SVG_PATHS",
    width: 294,
    height: 73,
    colors: [0xdededd, 0x333333, 0x231f20, 0xffffff, 0x6b6b6b, 0x6b6b6b, 0x231f20],
    paths: ["M294,72.5H0V1h294V72.5z", "M294,32.902H0v-1h294V32.902z", "M294,73H0v-1h294V73z", "M294,64.5H0v-23h294V64.5z", "M0,65v-1h294v1H0z", "M0,42v-1h294v1H0z", "M0,2V0h294v2H0z"]
  },
  "browserwindow_ne": {
    type: "SVG_PATHS",
    width: 38,
    height: 73,
    colors: [0xdededd, 0x333333, 0x231f20, 0xffffff, 0x6b6b6b, 0x6b6b6b, 0x6b6b6b, 0x6b6b6b, 0x231f20],
    paths: ["M37,72.5H0V1h35c1.104,0,2,0.896,2,2V72.5z", "M38,32.902H0v-1h38V32.902z", "M38,73H0v-1h38V73z", "M3.5,64.5H0v-23h3.5V64.5z", "M4,65H0v-1h3V42H0v-1h4V65z", "M28,47c0,1.104-0.896,2-2,2H14c-1.104,0-2-0.896-2-2l0,0c0-1.105,0.896-2,2-2h12C27.104,45,28,45.895,28,47L28,47z", "M28,53c0,1.104-0.896,2-2,2H14c-1.104,0-2-0.896-2-2l0,0c0-1.104,0.896-2,2-2h12C27.104,51,28,51.896,28,53L28,53z", "M28,59c0,1.104-0.896,2-2,2H14c-1.104,0-2-0.896-2-2l0,0c0-1.104,0.896-2,2-2h12C27.104,57,28,57.896,28,59L28,59z", "M38,73h-2V3c0-0.552-0.448-1-1-1H0V0h35c1.654,0,3,1.346,3,3V73z"]
  },
  "browserwindow_e": {
    type: "SVG_PATHS",
    width: 38,
    height: 327,
    colors: [0x333333],
    paths: ["M38,0v327h-2V0H38z"]
  },
  "browserwindow_se": {
    type: "SVG_PATHS",
    width: 38,
    height: 4,
    colors: [0x333333],
    paths: ["M38,0v1c0,1.654-1.346,3-3,3H0V2h35c0.552,0,1-0.448,1-1V0H38z"]
  },
  "browserwindow_s": {
    type: "SVG_PATHS",
    width: 294,
    height: 4,
    colors: [0x333333],
    paths: ["M0,4V2h294v2H0z"]
  },
  "browserwindow_sw": {
    type: "SVG_PATHS",
    width: 123,
    height: 4,
    colors: [0x333333],
    paths: ["M3,2h120v2H3C1.346,4,0,2.654,0,1V0h2v1C2,1.552,2.449,2,3,2z"]
  },
  "browserwindow_w": {
    type: "SVG_PATHS",
    width: 123,
    height: 327,
    colors: [0x333333],
    paths: ["M2,0v327H0V0H2z"]
  },
  "browserwindow_c": {
    type: "SVG_PATHS",
    width: 294,
    height: 327,
    paths: []
  },
  "browserwindow_padding_top": 8,
  "browserwindow_url_padding_top": 48,
  "browserwindow_url_padding_left": 126,
  "browserwindow_url_padding_right": 46,
  "browserwindow_scrollbar_padding_top": 72,
  "browserwindow_scrollbar_padding_bottom": 0,
  "browserwindow_scrollbar_padding_bottom_breakline": 0,
  "browserwindow_scrollbar_padding_right": 0,
  "browserwindow_border_thickness": 1,
  "button": {
    type: "SVG_SCALE9",
    scale9: [4, 4, 4, 4],
    width: 37,
    height: 24
  },
  "button_nw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M1,4H0V3.5C0,1.57,1.57,0,3.5,0H4v1H3.5C2.098,1,1,2.098,1,3.5V4z"]
  },
  "button_n": {
    type: "SVG_PATHS",
    width: 29,
    height: 4,
    paths: ["M0,1V0h29v1H0z"]
  },
  "button_ne": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M4,4H3V3.5C3,2.098,1.902,1,0.5,1H0V0h0.5C2.43,0,4,1.57,4,3.5V4z"]
  },
  "button_e": {
    type: "SVG_PATHS",
    width: 4,
    height: 16,
    paths: ["M4,0v16H3V0H4z"]
  },
  "button_se": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M4,0v0.5C4,2.43,2.43,4,0.5,4H0V3h0.5C1.902,3,3,1.902,3,0.5V0H4z"]
  },
  "button_s": {
    type: "SVG_PATHS",
    width: 29,
    height: 4,
    paths: ["M0,4V3h29v1H0z"]
  },
  "button_sw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M3.5,3H4v1H3.5C1.57,4,0,2.43,0,0.5V0h1v0.5C1,1.902,2.098,3,3.5,3z"]
  },
  "button_w": {
    type: "SVG_PATHS",
    width: 4,
    height: 16,
    paths: ["M1,0v16H0V0H1z"]
  },
  "button_c": {
    type: "SVG_PATHS",
    width: 29,
    height: 16,
    paths: []
  },
  "button_padding_left": 3,
  "button_padding_right": 3,
  "button_padding_top": 6,
  "button_padding_bottom": 3,
  "button_background_padding_left": 1,
  "button_background_padding_right": 1,
  "button_background_padding_top": 1,
  "button_background_padding_bottom": 1,
  "button_border_color": 0x333333,
  "button_disabled_border_color": 0xcccccc,
  "button_background_color": 0xdddddd,
  "button_icon_padding": 3,
  "button_default_icon_color": 0x333333,
  "buttonbar_border": {
    type: "SVG_SCALE9",
    scale9: [4, 4, 4, 4],
    width: 24,
    height: 24
  },
  "buttonbar_border_nw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    colors: [0x333333],
    paths: ["M1,4H0V2.5C0,1.121,1.122,0,2.5,0H4v1H2.5C1.673,1,1,1.673,1,2.5V4z"]
  },
  "buttonbar_border_n": {
    type: "SVG_PATHS",
    width: 16,
    height: 4,
    colors: [0x333333],
    paths: ["M0,1V0h16v1H0z"]
  },
  "buttonbar_border_ne": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    colors: [0x333333],
    paths: ["M4,4H3V2.5C3,1.673,2.327,1,1.5,1H0V0h1.5C2.878,0,4,1.121,4,2.5V4z"]
  },
  "buttonbar_border_e": {
    type: "SVG_PATHS",
    width: 4,
    height: 16,
    colors: [0x333333],
    paths: ["M4,0v16H3V0H4z"]
  },
  "buttonbar_border_se": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    colors: [0x333333],
    paths: ["M4,0v1.5C4,2.879,2.878,4,1.5,4H0V3h1.5C2.327,3,3,2.327,3,1.5V0H4z"]
  },
  "buttonbar_border_s": {
    type: "SVG_PATHS",
    width: 16,
    height: 4,
    colors: [0x333333],
    paths: ["M0,4V3h16v1H0z"]
  },
  "buttonbar_border_sw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    colors: [0x333333],
    paths: ["M2.5,3H4v1H2.5C1.122,4,0,2.879,0,1.5V0h1v1.5C1,2.327,1.673,3,2.5,3z"]
  },
  "buttonbar_border_w": {
    type: "SVG_PATHS",
    width: 4,
    height: 16,
    colors: [0x333333],
    paths: ["M1,0v16H0V0H1z"]
  },
  "buttonbar_border_c": {
    type: "SVG_PATHS",
    width: 16,
    height: 16,
    paths: []
  },
  "buttonbar_padding_left": 14,
  "buttonbar_padding_middle": 12,
  "buttonbar_padding_right": 14,
  "buttonbar_padding_top": 4,
  "buttonbar_padding_bottom": 4,
  "buttonbar_separator_width": 1,
  "buttonbar_separator_color": 0x333333,
  "buttonbar_selection_color": 0x95BCF2,
  'buttonbar_background_color': 0xdddddd,
  "buttonbar_selection_padding_top": 1,
  "buttonbar_selection_padding_bottom": 1,
  "buttonbar_selection_padding_left": 1,
  "buttonbar_selection_padding_right": 1,
  "calendar": {
    type: "SVG_PATHS",
    width: 96,
    height: 96,
    colors: [0xFFFFFF, 0x676767, 0x676767, 0x676767, 0x676767],
    paths: ["M0,0L0,96L96,96L96,0z", "M96.5,96.5H-0.5v-97H96.5V96.5zM0.5,95.5H95.5v-95H0.5V95.5z", "M96,21.045H0v-1H96V21.045z", "M12,16l-6-5.5L12,5V16z", "M84,5l6,5.5L84,16V5z", "M21.838,14.5V5.828h5.455v1.189h-3.65v2.637h3v1.195h-3v3.65H21.838z", "M33.967,14.283c-0.824,0.242-1.605,0.363-2.344,0.363c-1.074,0-1.922-0.305-2.543-0.914s-0.932-1.441-0.932-2.496c0-0.996,0.284-1.798,0.853-2.405S30.32,7.92,31.254,7.92c0.941,0,1.629,0.297,2.062,0.891s0.65,1.533,0.65,2.818h-3.99c0.117,1.227,0.791,1.84,2.021,1.84c0.582,0,1.238-0.135,1.969-0.404V14.283zM29.953,10.65h2.309c0-1.098-0.354-1.646-1.061-1.646C30.482,9.004,30.066,9.553,29.953,10.65z", "M35.76,14.5V5.248h1.734v4.031c0.445-0.906,1.146-1.359,2.104-1.359c0.773,0,1.382,0.283,1.825,0.85s0.665,1.346,0.665,2.338c0,1.074-0.25,1.933-0.75,2.575s-1.168,0.964-2.004,0.964c-0.719,0-1.332-0.266-1.84-0.797l-0.322,0.65H35.76zM37.494,12.818c0.422,0.469,0.871,0.703,1.348,0.703c0.426,0,0.766-0.206,1.02-0.618s0.381-0.962,0.381-1.649c0-1.316-0.424-1.975-1.271-1.975c-0.512,0-1.004,0.33-1.477,0.99V12.818z", "M47.443,14.5v-1.371L47.537,13c0.465-0.621,0.941-1.145,1.43-1.57l0.709-0.627c1.167-1.031,1.751-1.936,1.751-2.713c0-0.906-0.515-1.359-1.547-1.359c-0.637,0-1.375,0.215-2.214,0.645V6.104c0.882-0.328,1.736-0.492,2.561-0.492c0.921,0,1.659,0.218,2.215,0.653c0.555,0.436,0.831,1.019,0.831,1.749c0,0.504-0.134,0.963-0.404,1.377c-0.269,0.414-0.743,0.895-1.424,1.441l-0.585,0.469c-0.816,0.656-1.301,1.266-1.454,1.828h3.833V14.5H47.443z", "M58.418,14.717c-0.984,0-1.784-0.422-2.4-1.266c-0.615-0.844-0.922-1.939-0.922-3.287c0-1.363,0.309-2.463,0.926-3.299s1.43-1.254,2.438-1.254c1,0,1.809,0.417,2.429,1.251c0.618,0.834,0.929,1.927,0.929,3.278c0,1.387-0.311,2.496-0.929,3.328C60.268,14.301,59.445,14.717,58.418,14.717zM58.435,13.633c1.028,0,1.542-1.156,1.542-3.469s-0.506-3.469-1.518-3.469s-1.518,1.154-1.518,3.463C56.942,12.475,57.439,13.633,58.435,13.633z", "M66.334,14.717c-0.984,0-1.784-0.422-2.399-1.266c-0.616-0.844-0.924-1.939-0.924-3.287c0-1.363,0.309-2.463,0.927-3.299c0.617-0.836,1.43-1.254,2.438-1.254c1,0,1.81,0.417,2.428,1.251c0.619,0.834,0.929,1.927,0.929,3.278c0,1.387-0.31,2.496-0.929,3.328C68.185,14.301,67.361,14.717,66.334,14.717zM66.352,13.633c1.027,0,1.541-1.156,1.541-3.469s-0.507-3.469-1.518-3.469c-1.012,0-1.519,1.154-1.519,3.463C64.857,12.475,65.356,13.633,66.352,13.633z", "M72.897,9.854c-0.837-0.551-1.254-1.215-1.254-1.992c0-0.672,0.263-1.216,0.791-1.632c0.526-0.416,1.218-0.624,2.074-0.624c0.805,0,1.447,0.178,1.931,0.533c0.482,0.355,0.724,0.83,0.724,1.424c0,0.766-0.489,1.49-1.465,2.174c1.233,0.645,1.851,1.455,1.851,2.432c0,0.762-0.301,1.377-0.901,1.846c-0.603,0.469-1.392,0.703-2.367,0.703c-0.934,0-1.687-0.215-2.259-0.645c-0.573-0.43-0.858-0.996-0.858-1.699C71.163,11.354,71.74,10.514,72.897,9.854zM73.751,10.434c-0.629,0.523-0.942,1.076-0.942,1.658c0,0.453,0.156,0.826,0.472,1.119c0.313,0.293,0.713,0.439,1.198,0.439c0.426,0,0.781-0.115,1.068-0.346s0.432-0.518,0.432-0.861c0-0.297-0.098-0.551-0.29-0.762c-0.194-0.211-0.57-0.475-1.129-0.791L73.751,10.434zM74.794,9.244c0.578-0.449,0.868-0.93,0.868-1.441c0-0.336-0.119-0.604-0.357-0.806s-0.554-0.302-0.944-0.302c-0.363,0-0.656,0.088-0.881,0.264s-0.337,0.406-0.337,0.691c0,0.281,0.111,0.528,0.334,0.741C73.699,8.604,74.139,8.889,74.794,9.244z", "M8.689,35.162c-0.437,0-0.995-0.095-1.674-0.285v-0.914c0.732,0.34,1.334,0.51,1.806,0.51c0.363,0,0.655-0.096,0.876-0.285c0.221-0.191,0.332-0.441,0.332-0.752c0-0.255-0.072-0.472-0.218-0.65C9.667,32.606,9.4,32.407,9.01,32.188l-0.448-0.26c-0.554-0.316-0.944-0.613-1.171-0.894c-0.228-0.28-0.341-0.606-0.341-0.978c0-0.502,0.182-0.914,0.545-1.237s0.826-0.485,1.389-0.485c0.501,0,1.03,0.083,1.586,0.25v0.844c-0.686-0.27-1.197-0.404-1.534-0.404c-0.319,0-0.583,0.085-0.791,0.255s-0.312,0.384-0.312,0.642c0,0.217,0.076,0.408,0.229,0.576c0.152,0.166,0.431,0.365,0.835,0.598l0.466,0.264c0.562,0.318,0.955,0.621,1.178,0.904c0.223,0.285,0.334,0.626,0.334,1.024c0,0.565-0.208,1.02-0.626,1.362C9.931,34.991,9.378,35.162,8.689,35.162z", "M19.046,35v-6.504h1.279l1.802,5.031l1.854-5.031h1.143V35h-0.861v-5.291l-1.793,4.855h-0.892l-1.74-4.869V35H19.046z", "M34.707,35v-5.814H32.4v-0.689h5.537v0.689H35.63V35H34.707z", "M46.083,35l-1.674-6.504h0.888l1.332,5.137l1.226-5.137h0.889l1.159,5.066l1.433-5.066h0.751L50.261,35h-0.918l-1.146-5.014L46.997,35H46.083z", "M60.873,35v-5.814h-2.306v-0.689h5.537v0.689h-2.307V35H60.873z", "M72.843,35v-6.504h3.634v0.689h-2.712v2.18h2.278v0.682h-2.278V35H72.843z", "M87.189,35.162c-0.435,0-0.993-0.095-1.674-0.285v-0.914c0.732,0.34,1.335,0.51,1.808,0.51c0.362,0,0.654-0.096,0.875-0.285c0.222-0.191,0.333-0.441,0.333-0.752c0-0.255-0.073-0.472-0.218-0.65s-0.412-0.378-0.803-0.598l-0.447-0.26c-0.554-0.316-0.944-0.613-1.171-0.894c-0.228-0.28-0.341-0.606-0.341-0.978c0-0.502,0.181-0.914,0.545-1.237c0.363-0.323,0.826-0.485,1.389-0.485c0.5,0,1.029,0.083,1.586,0.25v0.844c-0.686-0.27-1.197-0.404-1.533-0.404c-0.319,0-0.583,0.085-0.791,0.255s-0.312,0.384-0.312,0.642c0,0.217,0.076,0.408,0.229,0.576c0.151,0.166,0.431,0.365,0.835,0.598l0.466,0.264c0.562,0.318,0.955,0.621,1.177,0.904c0.224,0.285,0.335,0.626,0.335,1.024c0,0.565-0.208,1.02-0.627,1.362C88.431,34.991,87.879,35.162,87.189,35.162z", "M73.089,45v-0.578h1.157v-4.562l-1.157,0.289v-0.594l1.93-0.48v5.348h1.157V45H73.089z", "M85.635,45v-0.676c0.225-0.523,0.678-1.094,1.358-1.711l0.441-0.395c0.569-0.514,0.853-1.021,0.853-1.527c0-0.322-0.099-0.576-0.291-0.762c-0.194-0.186-0.462-0.277-0.799-0.277c-0.4,0-0.874,0.154-1.418,0.465v-0.68c0.514-0.242,1.021-0.363,1.526-0.363c0.541,0,0.978,0.146,1.305,0.438c0.328,0.291,0.493,0.678,0.493,1.16c0,0.346-0.082,0.654-0.248,0.922c-0.167,0.268-0.473,0.594-0.925,0.977l-0.296,0.254c-0.617,0.523-0.976,1.023-1.071,1.5h2.513V45H85.635z", "M7.283,55.957v-0.719c0.552,0.219,0.987,0.328,1.305,0.328c0.367,0,0.666-0.107,0.896-0.322S9.83,54.752,9.83,54.41c0-0.812-0.569-1.219-1.707-1.219H7.818v-0.52l0.27-0.004c1.068,0,1.602-0.377,1.602-1.129c0-0.592-0.352-0.887-1.055-0.887c-0.385,0-0.81,0.107-1.273,0.324v-0.672c0.456-0.154,0.898-0.23,1.328-0.23c1.182,0,1.773,0.447,1.773,1.34c0,0.68-0.402,1.164-1.207,1.453c0.932,0.217,1.398,0.727,1.398,1.531c0,0.545-0.18,0.971-0.539,1.281s-0.853,0.465-1.48,0.465C8.277,56.145,7.827,56.082,7.283,55.957z", "M22.565,56v-1.637h-2.602v-0.582l2.602-3.562h0.723v3.516h0.773v0.629h-0.773V56H22.565zM20.718,53.734h1.898v-2.57L20.718,53.734z", "M33.61,56.023v-0.688c0.383,0.154,0.748,0.23,1.094,0.23c0.38,0,0.679-0.111,0.896-0.336s0.326-0.533,0.326-0.926c0-0.867-0.598-1.301-1.793-1.301c-0.135,0-0.277,0.01-0.426,0.027v-2.812h2.926v0.672h-2.25v1.523c0.742,0,1.322,0.17,1.738,0.512s0.625,0.814,0.625,1.418c0,0.562-0.189,1.004-0.568,1.322s-0.904,0.479-1.576,0.479C34.311,56.145,33.979,56.104,33.61,56.023z", "M47.095,53.004c0.367-0.477,0.83-0.715,1.387-0.715c0.506,0,0.919,0.168,1.24,0.502s0.483,0.764,0.483,1.287c0,0.602-0.182,1.096-0.541,1.484c-0.362,0.389-0.818,0.582-1.374,0.582c-0.638,0-1.137-0.256-1.496-0.77s-0.539-1.225-0.539-2.137c0-1.002,0.207-1.781,0.619-2.336s0.994-0.832,1.744-0.832c0.345,0,0.774,0.072,1.29,0.219v0.672c-0.545-0.209-0.99-0.312-1.336-0.312C47.587,50.648,47.095,51.434,47.095,53.004zM49.431,54.25c0-0.438-0.098-0.781-0.295-1.031s-0.469-0.375-0.818-0.375c-0.326,0-0.605,0.109-0.84,0.328s-0.352,0.482-0.352,0.789c0,0.484,0.111,0.873,0.334,1.166s0.518,0.439,0.885,0.439c0.331,0,0.594-0.119,0.791-0.357S49.431,54.65,49.431,54.25z", "M60.006,56c0.075-0.451,0.183-0.84,0.323-1.168c0.139-0.328,0.385-0.779,0.739-1.352l1.578-2.539h-2.976v-0.723h3.71v0.723c-1.48,2.186-2.312,3.871-2.488,5.059H60.006z", "M73.727,52.852c-0.564-0.393-0.848-0.828-0.848-1.305c0-0.438,0.16-0.793,0.48-1.066s0.737-0.41,1.254-0.41c0.476,0,0.863,0.115,1.158,0.348c0.296,0.232,0.442,0.533,0.442,0.902c0,0.561-0.358,1.07-1.078,1.531c0.874,0.428,1.31,0.979,1.31,1.652c0,0.484-0.183,0.879-0.549,1.184s-0.838,0.457-1.416,0.457c-0.567,0-1.026-0.145-1.377-0.432c-0.35-0.287-0.526-0.666-0.526-1.135C72.578,53.885,72.96,53.311,73.727,52.852zM74.144,53.164c-0.31,0.24-0.519,0.449-0.63,0.627s-0.167,0.4-0.167,0.666c0,0.33,0.108,0.598,0.321,0.803s0.495,0.307,0.843,0.307c0.336,0,0.611-0.088,0.823-0.262c0.213-0.174,0.317-0.4,0.317-0.68c0-0.225-0.065-0.414-0.196-0.568c-0.132-0.154-0.387-0.344-0.764-0.568L74.144,53.164zM74.707,52.609c0.54-0.354,0.809-0.748,0.809-1.184c0-0.232-0.089-0.418-0.269-0.561s-0.416-0.213-0.712-0.213c-0.283,0-0.513,0.07-0.691,0.209c-0.176,0.139-0.266,0.32-0.266,0.541c0,0.217,0.081,0.41,0.24,0.58C73.979,52.152,74.276,52.361,74.707,52.609z", "M85.88,55.965v-0.652c0.491,0.17,0.907,0.254,1.25,0.254c0.479,0,0.854-0.203,1.124-0.611c0.269-0.408,0.404-0.975,0.404-1.701c-0.391,0.482-0.835,0.723-1.332,0.723c-0.487,0-0.9-0.18-1.238-0.537s-0.508-0.795-0.508-1.311c0-0.592,0.18-1.082,0.541-1.471c0.36-0.389,0.813-0.584,1.36-0.584c0.628,0,1.124,0.252,1.485,0.756c0.36,0.504,0.543,1.193,0.543,2.068c0,0.994-0.221,1.785-0.663,2.369c-0.441,0.584-1.038,0.877-1.791,0.877C86.754,56.145,86.364,56.084,85.88,55.965zM86.35,51.977c0,0.447,0.099,0.801,0.298,1.059s0.474,0.387,0.819,0.387c0.326,0,0.604-0.111,0.834-0.332c0.229-0.221,0.346-0.488,0.346-0.801c0-0.492-0.112-0.889-0.334-1.189c-0.224-0.301-0.516-0.451-0.878-0.451c-0.335,0-0.601,0.119-0.795,0.357C86.448,51.244,86.35,51.568,86.35,51.977z", "M5.144,67v-0.578H6.3v-4.562l-1.156,0.289v-0.594l1.93-0.48v5.348H8.23V67H5.144z", "M11.507,67.145c-0.596,0-1.079-0.281-1.447-0.842s-0.553-1.293-0.553-2.197c0-0.912,0.186-1.645,0.557-2.199s0.86-0.832,1.467-0.832s1.096,0.277,1.467,0.832s0.557,1.283,0.557,2.188c0,0.928-0.186,1.668-0.557,2.221S12.13,67.145,11.507,67.145zM11.515,66.566c0.815,0,1.223-0.828,1.223-2.484c0-1.619-0.402-2.43-1.207-2.43c-0.802,0-1.203,0.818-1.203,2.453C10.328,65.746,10.724,66.566,11.515,66.566z", "M18.228,67v-0.578h1.156v-4.562l-1.156,0.289v-0.594l1.93-0.48v5.348h1.156V67H18.228z", "M23.286,67v-0.578h1.156v-4.562l-1.156,0.289v-0.594l1.93-0.48v5.348h1.156V67H23.286z", "M31.311,67v-0.578h1.156v-4.562l-1.156,0.289v-0.594l1.93-0.48v5.348h1.156V67H31.311z", "M35.831,67v-0.676c0.224-0.523,0.677-1.094,1.359-1.711l0.441-0.395c0.568-0.514,0.852-1.021,0.852-1.527c0-0.322-0.097-0.576-0.291-0.762s-0.46-0.277-0.799-0.277c-0.401,0-0.874,0.154-1.418,0.465v-0.68c0.513-0.242,1.022-0.363,1.527-0.363c0.542,0,0.977,0.146,1.305,0.438s0.492,0.678,0.492,1.16c0,0.346-0.083,0.654-0.248,0.922s-0.473,0.594-0.924,0.977l-0.297,0.254c-0.617,0.523-0.974,1.023-1.07,1.5h2.512V67H35.831z", "M44.394,67v-0.578h1.156v-4.562l-1.156,0.289v-0.594l1.93-0.48v5.348h1.156V67H44.394z", "M49.063,66.957v-0.719c0.551,0.219,0.985,0.328,1.305,0.328c0.367,0,0.666-0.107,0.896-0.322c0.229-0.215,0.346-0.492,0.346-0.834c0-0.812-0.57-1.219-1.708-1.219h-0.305v-0.52l0.271-0.004c1.067,0,1.601-0.377,1.601-1.129c0-0.592-0.351-0.887-1.054-0.887c-0.387,0-0.811,0.107-1.274,0.324v-0.672c0.456-0.154,0.899-0.23,1.328-0.23c1.183,0,1.774,0.447,1.774,1.34c0,0.68-0.403,1.164-1.208,1.453c0.933,0.217,1.399,0.727,1.399,1.531c0,0.545-0.18,0.971-0.54,1.281c-0.359,0.311-0.854,0.465-1.479,0.465C50.056,67.145,49.606,67.082,49.063,66.957z", "M57.477,67v-0.578h1.157v-4.562l-1.157,0.289v-0.594l1.93-0.48v5.348h1.157V67H57.477z", "M64.344,67v-1.637h-2.601v-0.582l2.601-3.562h0.724v3.516h0.772v0.629h-0.772V67H64.344zM62.498,64.734h1.898v-2.57L62.498,64.734z", "M70.56,67v-0.578h1.157v-4.562l-1.157,0.289v-0.594l1.93-0.48v5.348h1.157V67H70.56z", "M75.389,67.023v-0.688c0.383,0.154,0.747,0.23,1.093,0.23c0.382,0,0.681-0.111,0.897-0.336c0.218-0.225,0.326-0.533,0.326-0.926c0-0.867-0.599-1.301-1.793-1.301c-0.136,0-0.277,0.01-0.427,0.027v-2.812h2.927v0.672h-2.25v1.523c0.742,0,1.322,0.17,1.738,0.512c0.417,0.342,0.625,0.814,0.625,1.418c0,0.562-0.19,1.004-0.569,1.322s-0.904,0.479-1.576,0.479C76.089,67.145,75.759,67.104,75.389,67.023z", "M83.644,67v-0.578h1.157v-4.562l-1.157,0.289v-0.594l1.93-0.48v5.348h1.157V67H83.644z", "M88.875,64.004c0.367-0.477,0.83-0.715,1.386-0.715c0.507,0,0.919,0.168,1.24,0.502c0.322,0.334,0.483,0.764,0.483,1.287c0,0.602-0.18,1.096-0.542,1.484c-0.361,0.389-0.817,0.582-1.372,0.582c-0.64,0-1.137-0.256-1.497-0.77c-0.359-0.514-0.539-1.225-0.539-2.137c0-1.002,0.208-1.781,0.62-2.336s0.993-0.832,1.743-0.832c0.345,0,0.774,0.072,1.29,0.219v0.672c-0.545-0.209-0.99-0.312-1.336-0.312C89.368,61.648,88.875,62.434,88.875,64.004zM91.21,65.25c0-0.438-0.097-0.781-0.295-1.031c-0.197-0.25-0.468-0.375-0.818-0.375c-0.325,0-0.604,0.109-0.839,0.328c-0.235,0.219-0.353,0.482-0.353,0.789c0,0.484,0.112,0.873,0.334,1.166c0.224,0.293,0.519,0.439,0.886,0.439c0.33,0,0.593-0.119,0.79-0.357C91.114,65.971,91.21,65.65,91.21,65.25z", "M5.144,78v-0.578H6.3v-4.562l-1.156,0.289v-0.594l1.93-0.48v5.348H8.23V78H5.144z", "M10.203,78c0.076-0.451,0.183-0.84,0.322-1.168s0.386-0.779,0.74-1.352l1.578-2.539H9.867v-0.723h3.711v0.723c-1.482,2.186-2.311,3.871-2.488,5.059H10.203z", "M18.228,78v-0.578h1.156v-4.562l-1.156,0.289v-0.594l1.93-0.48v5.348h1.156V78H18.228z", "M23.923,74.852c-0.565-0.393-0.848-0.828-0.848-1.305c0-0.438,0.16-0.793,0.48-1.066s0.738-0.41,1.254-0.41c0.477,0,0.863,0.115,1.158,0.348s0.443,0.533,0.443,0.902c0,0.561-0.359,1.07-1.078,1.531c0.873,0.428,1.309,0.979,1.309,1.652c0,0.484-0.183,0.879-0.549,1.184s-0.838,0.457-1.416,0.457c-0.568,0-1.027-0.145-1.377-0.432s-0.525-0.666-0.525-1.135C22.774,75.885,23.157,75.311,23.923,74.852zM24.341,75.164c-0.31,0.24-0.52,0.449-0.631,0.627s-0.166,0.4-0.166,0.666c0,0.33,0.107,0.598,0.32,0.803s0.495,0.307,0.844,0.307c0.336,0,0.61-0.088,0.822-0.262s0.318-0.4,0.318-0.68c0-0.225-0.066-0.414-0.197-0.568s-0.386-0.344-0.764-0.568L24.341,75.164zM24.903,74.609c0.539-0.354,0.809-0.748,0.809-1.184c0-0.232-0.09-0.418-0.27-0.561s-0.417-0.213-0.711-0.213c-0.284,0-0.514,0.07-0.691,0.209s-0.266,0.32-0.266,0.541c0,0.217,0.08,0.41,0.24,0.58S24.471,74.361,24.903,74.609z", "M31.311,78v-0.578h1.156v-4.562l-1.156,0.289v-0.594l1.93-0.48v5.348h1.156V78H31.311z", "M36.077,77.965v-0.652c0.49,0.17,0.906,0.254,1.25,0.254c0.479,0,0.854-0.203,1.123-0.611s0.404-0.975,0.404-1.701c-0.391,0.482-0.834,0.723-1.332,0.723c-0.487,0-0.9-0.18-1.238-0.537s-0.508-0.795-0.508-1.311c0-0.592,0.18-1.082,0.541-1.471s0.814-0.584,1.361-0.584c0.627,0,1.123,0.252,1.484,0.756s0.543,1.193,0.543,2.068c0,0.994-0.221,1.785-0.662,2.369s-1.039,0.877-1.791,0.877C36.95,78.145,36.559,78.084,36.077,77.965zM36.545,73.977c0,0.447,0.1,0.801,0.299,1.059s0.472,0.387,0.818,0.387c0.326,0,0.604-0.111,0.834-0.332s0.346-0.488,0.346-0.801c0-0.492-0.111-0.889-0.334-1.189s-0.515-0.451-0.877-0.451c-0.336,0-0.601,0.119-0.795,0.357S36.545,73.568,36.545,73.977z", "M43.855,78v-0.676c0.224-0.523,0.677-1.094,1.359-1.711l0.441-0.395c0.568-0.514,0.852-1.021,0.852-1.527c0-0.322-0.097-0.576-0.291-0.762s-0.46-0.277-0.799-0.277c-0.401,0-0.874,0.154-1.418,0.465v-0.68c0.513-0.242,1.022-0.363,1.527-0.363c0.542,0,0.977,0.146,1.305,0.438s0.492,0.678,0.492,1.16c0,0.346-0.083,0.654-0.248,0.922s-0.473,0.594-0.924,0.977l-0.297,0.254c-0.617,0.523-0.974,1.023-1.07,1.5h2.512V78H43.855z", "M50.757,78.145c-0.597,0-1.079-0.281-1.447-0.842s-0.553-1.293-0.553-2.197c0-0.912,0.186-1.645,0.557-2.199s0.859-0.832,1.467-0.832s1.096,0.277,1.467,0.832s0.558,1.283,0.558,2.188c0,0.928-0.187,1.668-0.558,2.221S51.38,78.145,50.757,78.145zM50.765,77.566c0.815,0,1.224-0.828,1.224-2.484c0-1.619-0.403-2.43-1.208-2.43c-0.802,0-1.203,0.818-1.203,2.453C49.578,76.746,49.973,77.566,50.765,77.566z", "M56.939,78v-0.676c0.224-0.523,0.676-1.094,1.358-1.711l0.441-0.395c0.567-0.514,0.852-1.021,0.852-1.527c0-0.322-0.097-0.576-0.29-0.762c-0.194-0.186-0.461-0.277-0.799-0.277c-0.401,0-0.875,0.154-1.418,0.465v-0.68c0.513-0.242,1.022-0.363,1.526-0.363c0.542,0,0.977,0.146,1.305,0.438c0.329,0.291,0.492,0.678,0.492,1.16c0,0.346-0.083,0.654-0.247,0.922c-0.166,0.268-0.475,0.594-0.925,0.977l-0.296,0.254c-0.617,0.523-0.975,1.023-1.07,1.5h2.512V78H56.939z", "M62.536,78v-0.578h1.157v-4.562l-1.157,0.289v-0.594l1.93-0.48v5.348h1.157V78H62.536z", "M70.022,78v-0.676c0.225-0.523,0.677-1.094,1.358-1.711l0.442-0.395c0.567-0.514,0.851-1.021,0.851-1.527c0-0.322-0.097-0.576-0.29-0.762c-0.194-0.186-0.461-0.277-0.799-0.277c-0.4,0-0.874,0.154-1.419,0.465v-0.68c0.515-0.242,1.023-0.363,1.527-0.363c0.542,0,0.977,0.146,1.305,0.438c0.329,0.291,0.492,0.678,0.492,1.16c0,0.346-0.081,0.654-0.247,0.922s-0.474,0.594-0.925,0.977l-0.296,0.254c-0.617,0.523-0.975,1.023-1.07,1.5h2.512V78H70.022z", "M75.081,78v-0.676c0.225-0.523,0.678-1.094,1.358-1.711l0.441-0.395c0.568-0.514,0.852-1.021,0.852-1.527c0-0.322-0.097-0.576-0.29-0.762c-0.194-0.186-0.461-0.277-0.799-0.277c-0.4,0-0.874,0.154-1.418,0.465v-0.68c0.514-0.242,1.022-0.363,1.526-0.363c0.542,0,0.978,0.146,1.305,0.438c0.329,0.291,0.492,0.678,0.492,1.16c0,0.346-0.081,0.654-0.247,0.922s-0.474,0.594-0.925,0.977l-0.296,0.254c-0.617,0.523-0.975,1.023-1.07,1.5h2.512V78H75.081z", "M83.106,78v-0.676c0.225-0.523,0.678-1.094,1.358-1.711l0.441-0.395c0.569-0.514,0.853-1.021,0.853-1.527c0-0.322-0.099-0.576-0.291-0.762c-0.194-0.186-0.462-0.277-0.799-0.277c-0.4,0-0.874,0.154-1.418,0.465v-0.68c0.514-0.242,1.021-0.363,1.526-0.363c0.541,0,0.978,0.146,1.305,0.438c0.328,0.291,0.492,0.678,0.492,1.16c0,0.346-0.081,0.654-0.247,0.922c-0.167,0.268-0.473,0.594-0.925,0.977l-0.296,0.254c-0.617,0.523-0.976,1.023-1.071,1.5h2.513V78H83.106z", "M88.313,77.957v-0.719c0.552,0.219,0.985,0.328,1.305,0.328c0.367,0,0.666-0.107,0.896-0.322c0.229-0.215,0.346-0.492,0.346-0.834c0-0.812-0.569-1.219-1.708-1.219h-0.305v-0.52l0.271-0.004c1.067,0,1.601-0.377,1.601-1.129c0-0.592-0.351-0.887-1.054-0.887c-0.385,0-0.811,0.107-1.274,0.324v-0.672c0.456-0.154,0.899-0.23,1.328-0.23c1.183,0,1.774,0.447,1.774,1.34c0,0.68-0.403,1.164-1.208,1.453c0.933,0.217,1.399,0.727,1.399,1.531c0,0.545-0.18,0.971-0.54,1.281c-0.359,0.311-0.854,0.465-1.479,0.465C89.306,78.145,88.857,78.082,88.313,77.957z", "M4.605,89v-0.676c0.224-0.523,0.677-1.094,1.359-1.711l0.441-0.395c0.568-0.514,0.852-1.021,0.852-1.527c0-0.322-0.097-0.576-0.291-0.762s-0.46-0.277-0.799-0.277c-0.401,0-0.874,0.154-1.418,0.465v-0.68c0.513-0.242,1.022-0.363,1.527-0.363c0.542,0,0.977,0.146,1.305,0.438s0.492,0.678,0.492,1.16c0,0.346-0.083,0.654-0.248,0.922s-0.473,0.594-0.924,0.977l-0.297,0.254c-0.617,0.523-0.974,1.023-1.07,1.5h2.512V89H4.605z", "M12.011,89v-1.637H9.41v-0.582l2.602-3.562h0.723v3.516h0.773v0.629h-0.773V89H12.011zM10.164,86.734h1.898v-2.57L10.164,86.734z", "M17.688,89v-0.676c0.224-0.523,0.677-1.094,1.359-1.711l0.441-0.395c0.568-0.514,0.852-1.021,0.852-1.527c0-0.322-0.097-0.576-0.291-0.762s-0.46-0.277-0.799-0.277c-0.401,0-0.874,0.154-1.418,0.465v-0.68c0.513-0.242,1.022-0.363,1.527-0.363c0.542,0,0.977,0.146,1.305,0.438s0.492,0.678,0.492,1.16c0,0.346-0.083,0.654-0.248,0.922s-0.473,0.594-0.924,0.977l-0.297,0.254c-0.617,0.523-0.974,1.023-1.07,1.5h2.512V89H17.688z", "M23.056,89.023v-0.688c0.383,0.154,0.748,0.23,1.094,0.23c0.38,0,0.679-0.111,0.896-0.336s0.326-0.533,0.326-0.926c0-0.867-0.598-1.301-1.793-1.301c-0.135,0-0.277,0.01-0.426,0.027v-2.812h2.926v0.672h-2.25v1.523c0.742,0,1.322,0.17,1.738,0.512s0.625,0.814,0.625,1.418c0,0.562-0.189,1.004-0.568,1.322s-0.904,0.479-1.576,0.479C23.756,89.145,23.425,89.104,23.056,89.023z", "M30.772,89v-0.676c0.224-0.523,0.677-1.094,1.359-1.711l0.441-0.395c0.568-0.514,0.852-1.021,0.852-1.527c0-0.322-0.097-0.576-0.291-0.762s-0.46-0.277-0.799-0.277c-0.401,0-0.874,0.154-1.418,0.465v-0.68c0.513-0.242,1.022-0.363,1.527-0.363c0.542,0,0.977,0.146,1.305,0.438s0.492,0.678,0.492,1.16c0,0.346-0.083,0.654-0.248,0.922s-0.473,0.594-0.924,0.977l-0.297,0.254c-0.617,0.523-0.974,1.023-1.07,1.5h2.512V89H30.772z", "M36.542,86.004c0.367-0.477,0.83-0.715,1.387-0.715c0.505,0,0.918,0.168,1.24,0.502s0.482,0.764,0.482,1.287c0,0.602-0.18,1.096-0.541,1.484s-0.818,0.582-1.373,0.582c-0.638,0-1.137-0.256-1.496-0.77s-0.539-1.225-0.539-2.137c0-1.002,0.207-1.781,0.619-2.336s0.994-0.832,1.744-0.832c0.344,0,0.773,0.072,1.289,0.219v0.672c-0.544-0.209-0.99-0.312-1.336-0.312C37.034,83.648,36.542,84.434,36.542,86.004zM38.877,87.25c0-0.438-0.098-0.781-0.295-1.031s-0.469-0.375-0.818-0.375c-0.326,0-0.605,0.109-0.84,0.328s-0.352,0.482-0.352,0.789c0,0.484,0.111,0.873,0.334,1.166s0.518,0.439,0.885,0.439c0.331,0,0.594-0.119,0.791-0.357S38.877,87.65,38.877,87.25z", "M43.855,89v-0.676c0.224-0.523,0.677-1.094,1.359-1.711l0.441-0.395c0.568-0.514,0.852-1.021,0.852-1.527c0-0.322-0.097-0.576-0.291-0.762s-0.46-0.277-0.799-0.277c-0.401,0-0.874,0.154-1.418,0.465v-0.68c0.513-0.242,1.022-0.363,1.527-0.363c0.542,0,0.977,0.146,1.305,0.438s0.492,0.678,0.492,1.16c0,0.346-0.083,0.654-0.248,0.922s-0.473,0.594-0.924,0.977l-0.297,0.254c-0.617,0.523-0.974,1.023-1.07,1.5h2.512V89H43.855z", "M49.453,89c0.075-0.451,0.183-0.84,0.323-1.168c0.139-0.328,0.385-0.779,0.739-1.352l1.578-2.539h-2.976v-0.723h3.71v0.723c-1.481,2.186-2.311,3.871-2.488,5.059H49.453z", "M56.939,89v-0.676c0.224-0.523,0.676-1.094,1.358-1.711l0.441-0.395c0.567-0.514,0.852-1.021,0.852-1.527c0-0.322-0.097-0.576-0.29-0.762c-0.194-0.186-0.461-0.277-0.799-0.277c-0.401,0-0.875,0.154-1.418,0.465v-0.68c0.513-0.242,1.022-0.363,1.526-0.363c0.542,0,0.977,0.146,1.305,0.438c0.329,0.291,0.492,0.678,0.492,1.16c0,0.346-0.083,0.654-0.247,0.922c-0.166,0.268-0.475,0.594-0.925,0.977l-0.296,0.254c-0.617,0.523-0.975,1.023-1.07,1.5h2.512V89H56.939z", "M63.172,85.852c-0.565-0.393-0.847-0.828-0.847-1.305c0-0.438,0.159-0.793,0.48-1.066c0.32-0.273,0.737-0.41,1.254-0.41c0.476,0,0.862,0.115,1.158,0.348s0.442,0.533,0.442,0.902c0,0.561-0.359,1.07-1.078,1.531c0.873,0.428,1.31,0.979,1.31,1.652c0,0.484-0.183,0.879-0.549,1.184s-0.838,0.457-1.417,0.457c-0.566,0-1.025-0.145-1.376-0.432c-0.352-0.287-0.526-0.666-0.526-1.135C62.024,86.885,62.407,86.311,63.172,85.852zM63.59,86.164c-0.31,0.24-0.521,0.449-0.631,0.627c-0.109,0.178-0.166,0.4-0.166,0.666c0,0.33,0.107,0.598,0.32,0.803c0.214,0.205,0.496,0.307,0.845,0.307c0.335,0,0.609-0.088,0.822-0.262c0.212-0.174,0.317-0.4,0.317-0.68c0-0.225-0.064-0.414-0.196-0.568c-0.133-0.154-0.387-0.344-0.764-0.568L63.59,86.164zM64.153,85.609c0.54-0.354,0.809-0.748,0.809-1.184c0-0.232-0.089-0.418-0.269-0.561s-0.417-0.213-0.712-0.213c-0.283,0-0.514,0.07-0.691,0.209c-0.176,0.139-0.266,0.32-0.266,0.541c0,0.217,0.081,0.41,0.24,0.58C63.425,85.152,63.721,85.361,64.153,85.609z"]
  },
  "calendar_color": 0x333333,
  //used in calendar control
  "circleColor": 0xCCCCCC,
  "circlebutton_bg_color": 0xCCCCCC,
  "circlebutton_border_thickness": 1,
  "circlebutton_padding": 2,
  "calendar_padding_top": 2,
  "calendar_padding_bottom": 5,
  "calendar_padding_left": 6,
  "calendar_border_weight": 1,
  "calendar_lines_weight": 1,
  "calendar_icon_size": 12,
  "callout_default_color": 0x2675e7,
  "callout_default_background_alpha": 1,
  "callout_padding_left": 14,
  "callout_padding_right": 12,
  "callout_padding_top": 9,
  "callout_padding_bottom": 6,
  "callout_skin_count": 0,
  "canvas_scrollbar_padding_top": 0,
  "canvas_scrollbar_padding_bottom": 0,
  "canvas_scrollbar_padding_right": 0,
  "canvas_breakline_thickness": 1,
  "choice_icon_padding": 3,
  "choicegroup_padding_top": 2,
  "choicegroup_padding_bottom": 2,
  "choicegroup_line_padding": 9,
  "choicegroup_label_padding": 20,
  "choice_color": 0x333333,
  "colorpicker": {
    type: "SVG_PATHS",
    width: 26,
    height: 28,
    colors: [0xffffff, 0xffffff, 0x676767, 0x676767],
    paths: ["M23,24H1V2h22V24zM2,23h20V3H2V23z", "M22.999,24H12v-9h10.999V24z", "M22,17l-4.5,5L13,17H22z", "M24,25H0V1h24V25zM1,24h22V2H1V24z"]
  },
  "colorpicker_padding_left": 2,
  "colorpicker_padding_right": 2,
  "colorpicker_padding_top": 2,
  "colorpicker_padding_bottom": 4,
  "combobox": {
    type: "SVG_SCALE9",
    scale9: [4, 2, 21, 4],
    width: 51,
    height: 24
  },
  "combobox_nw": {
    type: "SVG_PATHS",
    width: 2,
    height: 4,
    paths: ["M1,4H0V0h2v1H1V4z"]
  },
  "combobox_n": {
    type: "SVG_PATHS",
    width: 28,
    height: 4,
    paths: ["M0,1V0h28v1H0z"]
  },
  "combobox_ne": {
    type: "SVG_PATHS",
    width: 21,
    height: 4,
    paths: ["M21,4h-1V2.5C20,1.673,19.327,1,18.5,1H0V0h18.5C19.878,0,21,1.121,21,2.5V4z", "M2,4H1V1h1V4z"]
  },
  "combobox_e": {
    type: "SVG_PATHS",
    width: 21,
    height: 16,
    paths: ["M21,0v16h-1V0H21z", "M2,0v16H1V0H2z"]
  },
  "combobox_se": {
    type: "SVG_PATHS",
    width: 21,
    height: 4,
    paths: ["M21,0v1.5C21,2.879,19.878,4,18.5,4H0V3h18.5C19.327,3,20,2.327,20,1.5V0H21z", "M2,0v3H1V0H2z"]
  },
  "combobox_s": {
    type: "SVG_PATHS",
    width: 28,
    height: 4,
    paths: ["M0,4V3h28v1H0z"]
  },
  "combobox_sw": {
    type: "SVG_PATHS",
    width: 2,
    height: 4,
    paths: ["M1,3h1v1H0V0h1V3z"]
  },
  "combobox_w": {
    type: "SVG_PATHS",
    width: 2,
    height: 16,
    paths: ["M1,0v16H0V0H1z"]
  },
  "combobox_c": {
    type: "SVG_PATHS",
    width: 28,
    height: 16,
    paths: []
  },
  "combobox_padding_left": 4,
  "combobox_padding_right": 6,
  "combobox_padding_top": 2,
  "combobox_padding_bottom": 9,
  "combobox_padding_middle": 3,
  "combobox_icon_padding": 3,
  "combobox_arrow_width": 20,
  "combobox_pulldown_overlap": 1,
  "combobox_pulldown_padding_bottom": 6,
  "combobox_border_thickness": 1,
  "combobox_background_disabled_color": 0xdddddd,
  "combobox_arrow_background_color": 0xdddddd,
  "combobox_border_disabled_color": 0xcccccc,
  "coverflow": {
    type: "SVG_PATHS",
    width: 493,
    height: 198,
    paths: ["M3,180.651V7.285l41,14.997v147.42l-0.371,0.1L3,180.651zM4,8.715v170.634l39-10.414V22.981L4,8.715z", "M489,180.653l-41-11.085V22.467l41-15.186V180.653zM449,168.803l39,10.544V8.719l-39,14.443V168.803z", "M324,177H169V21h155V177zM170,176h153V22H170V176z", "M408,180.651l-85-22.697V38.377l85-31.092V180.651zM324,157.187l83,22.162V8.715l-83,30.361V157.187z", "M449,180.653l-42-11.353V22.832l42-15.551V180.653zM408,168.535l40,10.812V8.719l-40,14.809V168.535z", "M85,180.651V7.284l0.672,0.246L170,38.469v119.419l-0.371,0.1L85,180.651zM86,8.716v170.633l83-22.229V39.168L86,8.716z", "M43,180.65V7.286l43,15.635v146.315L43,180.65zM44,8.714V179.35l41-10.883V23.622L44,8.714z"]
  },
  "coverflow_color": 0x676767,
  "curlybrace_horizontal": {
    type: "SVG_PATHS",
    width: 388,
    height: 15,
    paths: ["M195,15C155.176-12.447,55.376,21.792,3,4c54.123,11.296,144.04-17.662,193,7c44.622-25.627,134.694,7.814,190-10c-38.688,23.759-129.25-9.365-178,7c-3.854,1.294-8.238,3.421-10,7C197,15,196,15,195,15z"]
  },
  "curlybrace_vertical": {
    type: "SVG_PATHS",
    width: 15,
    height: 388,
    paths: ["M15,191C-12.448,230.824,21.792,330.625,4,383C15.296,328.877-13.663,238.961,11,190C-14.627,145.379,18.814,55.307,1,0C24.759,38.688-8.367,129.25,8,178c1.293,3.854,3.421,8.238,7,10C15,189,15,190,15,191z"]
  },
  "curlybrace_padding": 20,
  "datagrid_line_padding": 3,
  "datagrid_padding_top": 1,
  "datagrid_padding_bottom": 1,
  "datagrid_padding_left": 7,
  "datagrid_padding_right": 7,
  "datagrid_padding_middle": 6,
  "datagrid_selected_padding_left": 1,
  "datagrid_selected_padding_right": 1,
  "datagrid_scrollbar_padding_right": 0,
  "datagrid_scrollbar_padding_top": 0,
  "datagrid_scrollbar_padding_bottom": 0,
  "datagrid_vertical_separator_width": 1,
  "datepicker_padding_left": 6,
  "datepicker_padding_right": 6,
  "datepicker_padding_top": 6,
  "datepicker_padding_bottom": 5,
  "datepicker_icon_size": 24,
  "datepicker_disabled_color": 0xcfcfcf,
  "datepicker_border_padding_right": 30,
  "fieldset_border_padding_top": 6,
  "fieldset_label_padding_top": 0,
  "fieldset_label_padding_left": 15,
  "fieldset_label_padding_right": 4,
  "formattingtoolbar_border": 1,
  "formattingtoolbar_icon_size": 15,
  "formattingtoolbar_icon_color": 0x333333,
  "helpbutton": {
    type: "SVG_PATHS",
    width: 18,
    height: 18,
    colors: [0x676767, 0x676767],
    paths: ["M9,17c-4.411,0-8-3.589-8-8s3.589-8,8-8s8,3.589,8,8S13.411,17,9,17zM9,3C5.691,3,3,5.691,3,9s2.691,6,6,6s6-2.691,6-6S12.309,3,9,3z", "M6.144,6.338c0.144-0.369,0.344-0.689,0.603-0.961c0.259-0.271,0.571-0.481,0.936-0.634c0.365-0.151,0.773-0.228,1.225-0.228c0.583,0,1.069,0.08,1.459,0.24s0.704,0.358,0.943,0.597c0.236,0.238,0.406,0.495,0.509,0.77c0.103,0.275,0.154,0.531,0.154,0.77c0,0.394-0.052,0.718-0.154,0.974c-0.103,0.254-0.229,0.471-0.381,0.652c-0.151,0.18-0.32,0.334-0.505,0.461c-0.184,0.127-0.359,0.254-0.524,0.381C10.245,9.488,10.1,9.634,9.973,9.796c-0.127,0.166-0.207,0.37-0.239,0.617v0.468H8.071v-0.555C8.095,9.974,8.162,9.678,8.274,9.44c0.11-0.238,0.24-0.441,0.388-0.61c0.147-0.168,0.303-0.312,0.467-0.438c0.165-0.122,0.316-0.245,0.455-0.369c0.141-0.122,0.253-0.258,0.34-0.405c0.085-0.147,0.124-0.333,0.116-0.554c0-0.379-0.092-0.656-0.276-0.838c-0.186-0.18-0.442-0.271-0.77-0.271C8.772,5.956,8.581,6,8.421,6.085C8.262,6.172,8.13,6.286,8.027,6.43C7.925,6.574,7.85,6.742,7.8,6.936c-0.05,0.191-0.074,0.4-0.074,0.621h-1.81C5.923,7.113,6,6.707,6.144,6.338zM9.831,11.62v1.896H7.898V11.62H9.831z"]
  },
  "horizontalScrollbar": {
    type: "SVG_SCALE9",
    scale9: [2, 12, 12, 2],
    width: 61,
    height: 16
  },
  "horizontalScrollbar_nw": {
    type: "SVG_PATHS",
    width: 12,
    height: 2,
    colors: [0x515151],
    paths: ["M1,2H0V0h12v1H1V2z"]
  },
  "horizontalScrollbar_n": {
    type: "SVG_PATHS",
    width: 37,
    height: 2,
    colors: [0x515151],
    paths: ["M0,1V0h37v1H0z"]
  },
  "horizontalScrollbar_ne": {
    type: "SVG_PATHS",
    width: 12,
    height: 2,
    colors: [0x515151],
    paths: ["M12,2h-1V1H0V0h12V2z"]
  },
  "horizontalScrollbar_e": {
    type: "SVG_PATHS",
    width: 12,
    height: 12,
    colors: [0x515151, 0x515151],
    paths: ["M12,0v12h-1V0H12z", "M4,1.5L9,6l-5,4.5V1.5z"]
  },
  "horizontalScrollbar_se": {
    type: "SVG_PATHS",
    width: 12,
    height: 2,
    colors: [0x515151],
    paths: ["M12,0v2H0V1h11V0H12z"]
  },
  "horizontalScrollbar_s": {
    type: "SVG_PATHS",
    width: 37,
    height: 2,
    colors: [0x515151],
    paths: ["M0,2V1h37v1H0z"]
  },
  "horizontalScrollbar_sw": {
    type: "SVG_PATHS",
    width: 12,
    height: 2,
    colors: [0x515151],
    paths: ["M1,1h11v1H0V0h1V1z"]
  },
  "horizontalScrollbar_w": {
    type: "SVG_PATHS",
    width: 12,
    height: 12,
    colors: [0x515151, 0x515151],
    paths: ["M1,0v12H0V0H1z", "M8,10.5L3,6l5-4.5V10.5z"]
  },
  "horizontalScrollbar_c": {
    type: "SVG_PATHS",
    width: 37,
    height: 12,
    colors: [0x515151],
    paths: []
  },
  "horizontalScrollbar_padding_left": 13,
  "horizontalScrollbar_padding_right": 14,
  "horizontalScrollbar_thumb": {
    type: "SVG_PATHS",
    width: 21,
    height: 8,
    colors: [0x515151],
    paths: ["M21,4c0,2.209-1.791,4-4,4H4C1.791,8,0,6.209,0,4l0,0c0-2.209,1.791-4,4-4h13C19.209,0,21,1.791,21,4L21,4z"]
  },
  "icon_default_color": 0x333333,
  "icon_border_padding": 4,
  "icon_label_padding": 3,
  "icon_label_padding_bottom": 2,
  "image_border_padding": 5,
  "image_border_color": 0x333333,
  "image_frame": {
    type: "SVG_SCALE9",
    scale9: [1, 1, 1, 1],
    width: 42,
    height: 21
  },
  "image_frame_nw": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,1H0V0h1V1z"]
  },
  "image_frame_n": {
    type: "SVG_PATHS",
    width: 40,
    height: 1,
    paths: ["M40,0v1H0h0V0H40z"]
  },
  "image_frame_ne": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,1H0h0V0h1V1z"]
  },
  "image_frame_e": {
    type: "SVG_PATHS",
    width: 1,
    height: 19,
    paths: ["M1,0v19H0V0H1z"]
  },
  "image_frame_se": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,0v1H0V0h0H1z"]
  },
  "image_frame_s": {
    type: "SVG_PATHS",
    width: 40,
    height: 1,
    paths: ["M0,0h40v1H0L0,0L0,0z"]
  },
  "image_frame_sw": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,0v1H0V0H1z"]
  },
  "image_frame_w": {
    type: "SVG_PATHS",
    width: 1,
    height: 19,
    paths: ["M1,0v19H0V0H1z"]
  },
  "image_frame_c": {
    type: "SVG_PATHS",
    width: 40,
    height: 19,
    paths: ["M0,0L0,0l0,19h0V0z"]
  },
  "image_line_gap": 5,
  "image_line_width": 1,
  "image_cross_line_color": 0x333333,
  "iosKeyboard_icon_color": 0x666666,
  "iosKeyboard_background_color": 0xEEEEEE,
  "ioskeyboard_letter_key": {
    type: "SVG_PATHS",
    width: 23,
    height: 29,
    colors: [0xFFFFFF, 0x000000],
    paths: ["M4.156,0.112L8.040,0.216L8.040,0.216L18.968,0.131C21.177,0.114,22.982,1.891,22.999,4.100C22.999,4.110,23,4.120,23,4.131L23,24.870C23,27.079,21.209,28.870,19,28.870C18.957,28.870,18.915,28.869,18.872,28.868L16.718,28.8L16.718,28.8L4.442,28.946C2.095,28.974,0.170,27.095,0.142,24.747C0.141,24.685,0.142,24.622,0.144,24.559L0.25,21.297L0.25,21.297L0.048,4.157C0.022,1.948,1.792,0.136,4.001,0.110C4.053,0.110,4.104,0.110,4.156,0.112Z", "M5,0L18,0C20.761,0,23,2.238,23,5L23,24C23,26.761,20.761,29,18,29L5,29C2.238,29,0,26.761,0,24L0,5C0,2.238,2.238,0,5,0ZM5,1C2.790,1,1,2.790,1,5L1,24C1,26.209,2.790,28,5,28L18,28C20.209,28,22,26.209,22,24L22,5C22,2.790,20.209,1,18,1L5,1Z"]
  },
  "ioskeyboard_letter_key_gray": {
    type: "SVG_PATHS",
    width: 23,
    height: 29,
    colors: [0xCCCCCC, 0x333333],
    paths: ["M4.156,0.112L8.040,0.216L8.040,0.216L18.968,0.131C21.177,0.114,22.982,1.891,22.999,4.100C22.999,4.110,23,4.120,23,4.131L23,24.870C23,27.079,21.209,28.870,19,28.870C18.957,28.870,18.915,28.869,18.872,28.868L16.718,28.8L16.718,28.8L4.442,28.946C2.095,28.974,0.170,27.095,0.142,24.747C0.141,24.685,0.142,24.622,0.144,24.559L0.25,21.297L0.25,21.297L0.048,4.157C0.022,1.948,1.792,0.136,4.001,0.110C4.053,0.110,4.104,0.110,4.156,0.112Z", "M5,0L18,0C20.761,0,23,2.238,23,5L23,24C23,26.761,20.761,29,18,29L5,29C2.238,29,0,26.761,0,24L0,5C0,2.238,2.238,0,5,0ZM5,1C2.790,1,1,2.790,1,5L1,24C1,26.209,2.790,28,5,28L18,28C20.209,28,22,26.209,22,24L22,5C22,2.790,20.209,1,18,1L5,1Z"]
  },
  "ioskeyboard_123_key": {
    type: "SVG_PATHS",
    width: 33,
    height: 29,
    colors: [0xCCCCCC, 0x333333],
    paths: ["M5.043,0.044L23.929,0.25L23.929,0.25L28.793,0.380C31.261,0.446,33.215,2.489,33.171,4.957L32.819,25.045C32.780,27.236,30.986,28.989,28.795,28.975L8.628,28.85L8.628,28.85L5.786,28.907C3.301,28.958,1.246,26.984,1.196,24.500C1.195,24.487,1.195,24.475,1.195,24.462L1,8.212L1,8.212L1,4.043C1,1.834,2.790,0.043,5,0.043C5.014,0.043,5.029,0.043,5.043,0.044Z", "M5.5,0L28.5,0C31.261,0,33.5,2.238,33.5,5L33.5,24C33.5,26.761,31.261,29,28.5,29L5.5,29C2.738,29,0.5,26.761,0.5,24L0.5,5C0.5,2.238,2.738,0,5.5,0ZM5.5,1C3.290,1,1.5,2.790,1.5,5L1.5,24C1.5,26.209,3.290,28,5.5,28L28.5,28C30.709,28,32.5,26.209,32.5,24L32.5,5C32.5,2.790,30.709,1,28.5,1L5.5,1Z"]
  },
  "ioskeyboard_return_key": {
    type: "SVG_PATHS",
    width: 47,
    height: 27,
    colors: [0xCCCCCC, 0x333333],
    paths: ["M4,5.506e-14L43,5.506e-14C45.209,5.466e-14,47,1.790,47,4L47,23C47,25.209,45.209,27,43,27L4,27C1.790,27,2.705e-16,25.209,0,23L0,4C-2.705e-16,1.790,1.790,5.547e-14,4,5.506e-14Z", "M5,5.684e-14L42,5.684e-14C44.761,5.684e-14,47,2.238,47,5L47,22C47,24.761,44.761,27,42,27L5,27C2.238,27,0,24.761,0,22L0,5C0,2.238,2.238,5.684e-14,5,5.684e-14ZM5,1C2.790,1,1,2.790,1,5L1,22C1,24.209,2.790,26,5,26L42,26C44.209,26,46,24.209,46,22L46,5C46,2.790,44.209,1,42,1L5,1Z"]
  },
  "ioskeyboard_space_key": {
    type: "SVG_PATHS",
    width: 96,
    height: 29,
    colors: [0xFFFFFF, 0x333333],
    paths: ["M5.037,0.013L78,0.25L90.930,0.070C93.619,0.033,95.841,2.124,95.991,4.783L96,5.069L96,5.069L96,14.5L95.650,24.625C95.566,27.060,93.559,28.986,91.123,28.970L21.067,28.5L5.630,28.866C2.593,28.938,0.073,26.535,0.001,23.498L0,14.5L0.528,4.281C0.652,1.884,2.636,0.006,5.037,0.013Z", "M5,0L91,0C93.761,0,96,2.238,96,5L96,24C96,26.761,93.761,29,91,29L5,29C2.238,29,0,26.761,0,24L0,5C0,2.238,2.238,0,5,0ZM5,1C2.790,1,1,2.790,1,5L1,24C1,26.209,2.790,28,5,28L91,28C93.209,28,95,26.209,95,24L95,5C95,2.790,93.209,1,91,1L5,1Z"]
  },
  "ipad": {
    type: "SVG_PATHS",
    width: 616,
    height: 916,
    colors: [0xdedddd, 0x666766, 0x666766, 0x666766, 0x666766, 0x666766],
    paths: ["M603.517,2.5H12.484C6.97,2.5,2.5,6.973,2.5,12.489v891.021c0,5.517,4.47,9.989,9.984,9.989h591.033c5.513,0,9.983-4.473,9.983-9.989V12.489C613.5,6.973,609.029,2.5,603.517,2.5zM582.5,823.5h-548v-730h548V823.5z", "M308,53.5c-4.136,0-7.5-3.364-7.5-7.5s3.364-7.5,7.5-7.5s7.5,3.364,7.5,7.5S312.136,53.5,308,53.5zM308,39.5c-3.584,0-6.5,2.916-6.5,6.5s2.916,6.5,6.5,6.5s6.5-2.916,6.5-6.5S311.584,39.5,308,39.5z", "M308,891.5c-13.509,0-24.5-10.99-24.5-24.5s10.991-24.5,24.5-24.5c13.51,0,24.5,10.99,24.5,24.5S321.51,891.5,308,891.5zM308,843.5c-12.958,0-23.5,10.542-23.5,23.5s10.542,23.5,23.5,23.5s23.5-10.542,23.5-23.5S320.958,843.5,308,843.5z", "M314,875.5h-12c-1.378,0-2.5-1.121-2.5-2.5v-12c0-1.379,1.122-2.5,2.5-2.5h12c1.379,0,2.5,1.121,2.5,2.5v12C316.5,874.379,315.379,875.5,314,875.5zM302,859.5c-0.827,0-1.5,0.673-1.5,1.5v12c0,0.827,0.673,1.5,1.5,1.5h12c0.827,0,1.5-0.673,1.5-1.5v-12c0-0.827-0.673-1.5-1.5-1.5H302z", "M603.517,915H12.484C6.152,915,1,909.846,1,903.511V12.489C1,6.154,6.152,1,12.484,1h591.033C609.849,1,615,6.154,615,12.489v891.021C615,909.846,609.849,915,603.517,915zM12.484,4C7.806,4,4,7.809,4,12.489v891.021C4,908.191,7.806,912,12.484,912h591.033c4.678,0,8.483-3.809,8.483-8.489V12.489C612,7.809,608.194,4,603.517,4H12.484z", "M583,824H34V93h549V824zM35,823h547V94H35V823z"]
  },
  "ipad_scale9": {
    type: "SVG_SCALE9",
    scale9: [97, 38, 39, 96],
    width: 616,
    height: 916
  },
  "ipad_scale9_nw": {
    type: "SVG_PATHS",
    width: 38,
    height: 97,
    colors: [0xdedddd, 0x666766, 0x666766],
    paths: ["M34.5,97h-32V12.489C2.5,6.973,6.97,2.5,12.484,2.5H38v91h-3.5V97z", "M4,97H1V12.489C1,6.154,6.152,1,12.484,1H38v3H12.484C7.806,4,4,7.809,4,12.489V97z", "M35,97h-1v-4h4v1h-3V97z"]
  },
  "ipad_scale9_n": {
    type: "SVG_PATHS",
    width: 539,
    height: 97,
    colors: [0xdedddd, 0x666766, 0x666766],
    paths: ["M0,93.5v-91h539v91H0z", "M0,4V1h539v3H0z", "M0,94v-1h539v1H0z"]
  },
  "ipad_scale9_ne": {
    type: "SVG_PATHS",
    width: 39,
    height: 97,
    colors: [0xdedddd, 0x666766, 0x666766],
    paths: ["M36.5,97h-31v-3.5H0v-91h26.517c5.513,0,9.983,4.473,9.983,9.989V97z", "M38,97h-3V12.489C35,7.809,31.194,4,26.517,4H0V1h26.517C32.849,1,38,6.154,38,12.489V97z", "M6,97H5v-3H0v-1h6V97z"]
  },
  "ipad_scale9_e": {
    type: "SVG_PATHS",
    width: 39,
    height: 723,
    colors: [0xdedddd, 0x666766, 0x666766],
    paths: ["M36.5,0v723h-31V0H36.5z", "M38,0v723h-3V0H38z", "M6,0v723H5V0H6z"]
  },
  "ipad_scale9_se": {
    type: "SVG_PATHS",
    width: 39,
    height: 96,
    colors: [0xdedddd, 0x666766, 0x666766],
    paths: ["M36.5,0v83.511c0,5.517-4.471,9.989-9.983,9.989H0v-90h5.5V0H36.5z", "M38,0v83.511C38,89.846,32.849,95,26.517,95H0v-3h26.517C31.194,92,35,88.191,35,83.511V0H38z", "M6,0v4H0V3h5V0H6z"]
  },
  "ipad_scale9_s": {
    type: "SVG_PATHS",
    width: 539,
    height: 96,
    colors: [0xdedddd, 0x666766, 0x666766],
    paths: ["M0,93.5v-90h539v90H0z", "M0,95v-3h539v3H0z", "M0,4V3h539v1H0z"]
  },
  "ipad_scale9_sw": {
    type: "SVG_PATHS",
    width: 38,
    height: 96,
    colors: [0xdedddd, 0x666766, 0x666766],
    paths: ["M34.5,3.5H38v90H12.484C6.97,93.5,2.5,89.027,2.5,83.511V0h32V3.5z", "M12.484,92H38v3H12.484C6.152,95,1,89.846,1,83.511V0h3v83.511C4,88.191,7.806,92,12.484,92z", "M35,3h3v1h-4V0h1V3z"]
  },
  "ipad_scale9_w": {
    type: "SVG_PATHS",
    width: 38,
    height: 723,
    colors: [0xdedddd, 0x666766, 0x666766],
    paths: ["M34.5,0v723h-32V0H34.5z", "M4,0v723H1V0H4z", "M35,0v723h-1V0H35z"]
  },
  "ipad_scale9_c": {
    type: "SVG_PATHS",
    width: 539,
    height: 723,
    paths: []
  },
  "ipad_button": {
    type: "SVG_PATHS",
    width: 50,
    height: 50,
    colors: [0x666766, 0x666766],
    paths: ["M25,49.5C11.491,49.5,0.5,38.51,0.5,25S11.491,0.5,25,0.5S49.5,11.49,49.5,25S38.509,49.5,25,49.5zM25,1.5C12.042,1.5,1.5,12.042,1.5,25S12.042,48.5,25,48.5S48.5,37.958,48.5,25S37.958,1.5,25,1.5z", "M31,33.5H19c-1.378,0-2.5-1.121-2.5-2.5V19c0-1.379,1.122-2.5,2.5-2.5h12c1.378,0,2.5,1.121,2.5,2.5v12C33.5,32.379,32.378,33.5,31,33.5zM19,17.5c-0.827,0-1.5,0.673-1.5,1.5v12c0,0.827,0.673,1.5,1.5,1.5h12c0.827,0,1.5-0.673,1.5-1.5V19c0-0.827-0.673-1.5-1.5-1.5H19z"]
  },
  "ipad_speaker": {
    type: "SVG_PATHS",
    width: 16,
    height: 16,
    colors: [0x666766],
    paths: ["M8,15.5c-4.136,0-7.5-3.364-7.5-7.5S3.864,0.5,8,0.5s7.5,3.364,7.5,7.5S12.136,15.5,8,15.5zM8,1.5C4.416,1.5,1.5,4.416,1.5,8s2.916,6.5,6.5,6.5s6.5-2.916,6.5-6.5S11.584,1.5,8,1.5z"]
  },
  "iphone_4": {
    type: "SVG_PATHS",
    width: 282,
    height: 510,
    colors: [0x676767, 0x676767, 0x676767, 0x676767, 0xffffff, 0x676767, 0x676767, 0x676767, 0x676767, 0x676767, 0x676767],
    paths: ["M226,4h-38.75V0H226V4z", "M4,65.336H0V44.439h4V65.336z", "M4,115.336H0V99.785h4V115.336z", "M4,162.336H0v-15.551h4V162.336z", "M248.496,3h-215c-16.568,0-30,13.432-30,30v446c0,16.568,13.432,30,30,30h215c16.568,0,30-13.432,30-30V33C278.496,16.432,265.064,3,248.496,3zM254.188,422.691H27.771V82.465h226.417V422.691z", "M162.516,49.121h-43.039c-2.416,0-4.381-1.965-4.381-4.381s1.965-4.381,4.381-4.381h43.039c2.416,0,4.381,1.965,4.381,4.381S164.931,49.121,162.516,49.121zM119.477,41.359c-1.864,0-3.381,1.517-3.381,3.381s1.517,3.381,3.381,3.381h43.039c1.864,0,3.381-1.517,3.381-3.381s-1.517-3.381-3.381-3.381H119.477z", "M140.996,490.799c-13.311,0-24.141-10.829-24.141-24.141s10.83-24.141,24.141-24.141c13.312,0,24.141,10.829,24.141,24.141S154.307,490.799,140.996,490.799zM140.996,443.518c-12.76,0-23.141,10.381-23.141,23.141s10.381,23.141,23.141,23.141c12.76,0,23.141-10.381,23.141-23.141S153.756,443.518,140.996,443.518z", "M145.625,474.215h-9.258c-1.93,0-3.5-1.57-3.5-3.5v-8.113c0-1.93,1.57-3.5,3.5-3.5h9.258c1.93,0,3.5,1.57,3.5,3.5v8.113C149.125,472.645,147.555,474.215,145.625,474.215zM136.367,460.102c-1.378,0-2.5,1.121-2.5,2.5v8.113c0,1.379,1.122,2.5,2.5,2.5h9.258c1.378,0,2.5-1.121,2.5-2.5v-8.113c0-1.379-1.122-2.5-2.5-2.5H136.367z", "M95.533,52.677c-4.079,0-7.396-3.318-7.396-7.396s3.318-7.396,7.396-7.396s7.396,3.318,7.396,7.396S99.612,52.677,95.533,52.677zM95.533,38.884c-3.527,0-6.396,2.869-6.396,6.396s2.87,6.396,6.396,6.396s6.396-2.869,6.396-6.396S99.06,38.884,95.533,38.884z", "M250,5c14.888,0,27,12.112,27,27v448c0,14.888-12.112,27-27,27H32c-14.888,0-27-12.112-27-27V32C5,17.112,17.112,5,32,5H250M250,2H32C15.432,2,2,15.432,2,32v448c0,16.568,13.432,30,30,30h218c16.568,0,30-13.432,30-30V32C280,15.432,266.568,2,250,2L250,2z", "M254.98,422.964h-228V81.965h228V422.964zM27.98,421.964h226V82.965h-226V421.964z"]
  },
  "iphone_5": {
    type: "SVG_PATHS",
    width: 282,
    height: 573,
    colors: [0x676767, 0x676767, 0x676767, 0x676767, 0xffffff, 0x676767, 0x676767, 0x676767, 0x676767, 0x676767, 0x676767],
    paths: ["M226,4.25h-38.75V0H226V4.25z", "M4,65.586H0V44.689h4V65.586z", "M4,115.586H0v-15.551h4V115.586z", "M4,162.586H0v-15.551h4V162.586z", "M249,3H33C16.432,3,3,16.432,3,33v510c0,16.568,13.432,30,30,30h216c16.568,0,30-13.432,30-30V33C279,16.432,265.568,3,249,3zM254.188,486.691H27.771V82.715h226.417V486.691z", "M162.516,49.371h-43.039c-2.416,0-4.381-1.965-4.381-4.381s1.965-4.381,4.381-4.381h43.039c2.416,0,4.381,1.965,4.381,4.381S164.931,49.371,162.516,49.371zM119.477,41.609c-1.864,0-3.381,1.517-3.381,3.381s1.517,3.381,3.381,3.381h43.039c1.864,0,3.381-1.517,3.381-3.381s-1.517-3.381-3.381-3.381H119.477z", "M140.996,553.799c-13.311,0-24.141-10.829-24.141-24.141s10.83-24.141,24.141-24.141c13.312,0,24.141,10.829,24.141,24.141S154.307,553.799,140.996,553.799zM140.996,506.518c-12.76,0-23.141,10.381-23.141,23.141s10.381,23.141,23.141,23.141c12.76,0,23.141-10.381,23.141-23.141S153.756,506.518,140.996,506.518z", "M145.625,537.215h-9.258c-1.93,0-3.5-1.57-3.5-3.5v-8.113c0-1.93,1.57-3.5,3.5-3.5h9.258c1.93,0,3.5,1.57,3.5,3.5v8.113C149.125,535.645,147.555,537.215,145.625,537.215zM136.367,523.102c-1.378,0-2.5,1.121-2.5,2.5v8.113c0,1.379,1.122,2.5,2.5,2.5h9.258c1.378,0,2.5-1.121,2.5-2.5v-8.113c0-1.379-1.122-2.5-2.5-2.5H136.367z", "M140.996,32.927c-4.079,0-7.396-3.318-7.396-7.396s3.318-7.396,7.396-7.396s7.396,3.318,7.396,7.396S145.075,32.927,140.996,32.927zM140.996,19.134c-3.527,0-6.396,2.869-6.396,6.396s2.87,6.396,6.396,6.396s6.396-2.869,6.396-6.396S144.523,19.134,140.996,19.134z", "M250,5c14.888,0,27,12.112,27,27v511c0,14.888-12.112,27-27,27H32c-14.888,0-27-12.112-27-27V32C5,17.112,17.112,5,32,5H250M250,2H32C15.432,2,2,15.432,2,32v511c0,16.568,13.432,30,30,30h218c16.568,0,30-13.432,30-30V32C280,15.432,266.568,2,250,2L250,2z", "M254.98,486.715h-228v-404.5h228V486.715zM27.98,485.715h226v-402.5h-226V485.715z"]
  },
  "iphone_6": {
    type: "SVG_PATHS",
    width: 380,
    height: 754,
    colors: [0x676767, 0x676767, 0x676767, 0x676767, 0x676767, 0xffffff, 0x676767, 0x676767, 0x676767, 0x676767, 0x676767],
    paths: ["M0,92.65c0-1.109,0.893-2.008,1.995-2.008h8.428c1.102,0,1.995,0.886,1.995,2.008v24.542c0,1.109-0.893,2.008-1.995,2.008H1.995C0.893,119.2,0,118.314,0,117.192V92.65z", "M0,205.63c0-1.103,0.893-1.997,1.995-1.997h8.428c1.102,0,1.995,0.906,1.995,1.997v48.157c0,1.103-0.893,1.996-1.995,1.996H1.995c-1.102,0-1.995-0.905-1.995-1.996V205.63z", "M0,147.271c0-1.103,0.893-1.997,1.995-1.997h8.428c1.102,0,1.995,0.906,1.995,1.997v48.157c0,1.103-0.893,1.996-1.995,1.996H1.995c-1.102,0-1.995-0.904-1.995-1.996V147.271z", "M367.582,147.271c0-1.103,0.893-1.997,1.994-1.997h8.429c1.103,0,1.995,0.906,1.995,1.997v48.157c0,1.103-0.893,1.996-1.995,1.996h-8.429c-1.102,0-1.994-0.904-1.994-1.996V147.271z", "M327,0H53C25.38,0,3,22.386,3,50.001v653.998C3,731.613,25.387,754,53,754h274c27.619,0,50-22.387,50-50.001V50.001C377,22.387,354.613,0,327,0zM350,658H30V90h320V658z", "M327,3H53C27.04,3,6,24.041,6,50.001v653.998C6,729.957,27.043,751,53,751h274c25.961,0,47-21.041,47-47.001V50.001C374,24.043,352.957,3,327,3zM350,658H30V90h320V658z", "M27,87v574h326V87H27zM350,658H30V90h320V658z", "M190.078,735.15c14.955,0,27.078-12.122,27.078-27.075S205.033,681,190.078,681C175.124,681,163,693.122,163,708.075S175.124,735.15,190.078,735.15M190.078,733.15c-13.851,0-25.079-11.227-25.079-25.075S176.228,683,190.078,683s25.078,11.227,25.078,25.075S203.929,733.15,190.078,733.15", "M167.25,46.362L167.25,46.362c0,1.036,0.832,1.863,1.858,1.863h42.232c1.021,0,1.857-0.834,1.857-1.863c0-1.036-0.832-1.862-1.857-1.862h-42.232C168.087,44.5,167.25,45.334,167.25,46.362M165.25,46.362L165.25,46.362c0-2.131,1.73-3.862,3.858-3.862h42.232c2.133,0,3.857,1.724,3.857,3.862c0,2.131-1.73,3.863-3.857,3.863h-42.232C166.975,50.226,165.25,48.5,165.25,46.362", "M190.347,31.692c2.952,0,5.347-2.394,5.347-5.346c0-2.952-2.395-5.346-5.347-5.346c-2.954,0-5.347,2.394-5.347,5.346C185,29.298,187.393,31.692,190.347,31.692M190.347,29.692c-1.849,0-3.347-1.498-3.347-3.346c0-1.848,1.498-3.346,3.347-3.346s3.347,1.498,3.347,3.346C193.693,28.194,192.195,29.692,190.347,29.692", "M139.706,54.392c3.638,0,6.588-2.949,6.588-6.587c0-3.639-2.95-6.588-6.588-6.588c-3.639,0-6.588,2.949-6.588,6.588C133.118,51.442,136.067,54.392,139.706,54.392M139.706,52.392c-2.534,0-4.588-2.054-4.588-4.587c0-2.534,2.054-4.588,4.588-4.588s4.588,2.054,4.588,4.588C144.294,50.338,142.24,52.392,139.706,52.392"]
  },
  "iphone_X": {
    type: "SVG_PATHS",
    width: 434,
    height: 867,
    colors: [0xFFFFFF, 0x676767, 0x676767, 0x676767, 0x676767, 0x676767, 0x676767, 0x676767, 0x676767],
    paths: ["M432,197.55h-0.96V64.471c0-35.2-28.8-64-64-64h-300c-35.2,0-64,28.8-64,64v51.083c-0.014,0-0.026-0.004-0.04-0.004H2c-1.103,0-2,0.897-2,2v31c0,1.103,0.897,2,2,2h1c0.014,0,0.026-0.004,0.04-0.004v27.008c-0.014,0-0.026-0.004-0.04-0.004H2c-1.103,0-2,0.897-2,2v60c0,1.103,0.897,2,2,2h1c0.014,0,0.026-0.004,0.04-0.004v15.008c-0.014,0-0.026-0.004-0.04-0.004H2c-1.103,0-2,0.897-2,2v60c0,1.103,0.897,2,2,2h1c0.014,0,0.026-0.004,0.04-0.004v481.925c0,35.2,28.8,64,64,64h300c35.2,0,64-28.8,64-64V261.55H432c1.103,0,2-0.897,2-2v-60C434,198.447,433.103,197.55,432,197.55zM407,803.728c0,22.524-18.296,40.849-40.785,40.849H67.785C45.296,844.576,27,826.252,27,803.728V63.424c0-22.523,17.627-40.848,39.293-40.848h38.299v0.05h2.984c3.84,0,6.963,3.128,6.963,6.974v4.135c0,10.438,8.479,18.93,18.901,18.93h167.121c10.422,0,18.9-8.492,18.9-18.93V29.6c0-3.846,3.124-6.974,6.963-6.974h2.984v-0.05h38.299c21.666,0,39.293,18.325,39.293,40.848V803.728z", "M3,150.55H2c-1.103,0-2-0.897-2-2v-31c0-1.103,0.897-2,2-2h1c1.103,0,2,0.897,2,2v31C5,149.653,4.103,150.55,3,150.55z M3,148.55v1V148.55L3,148.55L3,148.55z M2,117.55v31h0.997l0.003-31H2z", "M3,241.55H2c-1.103,0-2-0.897-2-2v-60c0-1.103,0.897-2,2-2h1c1.103,0,2,0.897,2,2v60C5,240.653,4.103,241.55,3,241.55z M3,239.55v1V239.55L3,239.55L3,239.55z M2,179.55v60h0.997l0.003-60H2z", "M3,320.55H2c-1.103,0-2-0.897-2-2v-60c0-1.103,0.897-2,2-2h1c1.103,0,2,0.897,2,2v60C5,319.653,4.103,320.55,3,320.55z M3,318.55v1V318.55L3,318.55L3,318.55z M2,258.55v60h0.997l0.003-60H2z", "M432,261.55h-1c-1.103,0-2-0.897-2-2v-60c0-1.103,0.897-2,2-2h1c1.103,0,2,0.897,2,2v60C434,260.653,433.103,261.55,432,261.55z M432,259.55v1V259.55L432,259.55L432,259.55z M431,199.55v60h0.997l0.003-60H431z", "M367.04,2.471c34.187,0,62,27.813,62,62v738c0,34.187-27.813,62-62,62h-300c-34.187,0-62-27.813-62-62v-738c0-34.187,27.813-62,62-62H367.04 M367.04,0.471h-300c-35.2,0-64,28.8-64,64v738c0,35.2,28.8,64,64,64h300c35.2,0,64-28.8,64-64v-738C431.04,29.271,402.24,0.471,367.04,0.471L367.04,0.471z", "M367.707,22.576c21.666,0,39.293,18.325,39.293,40.848v740.303c0,22.524-18.296,40.849-40.785,40.849H67.785C45.296,844.576,27,826.252,27,803.728V63.424c0-22.523,17.627-40.848,39.293-40.848h38.299v0.05h2.984c3.84,0,6.963,3.128,6.963,6.974v4.135c0,10.438,8.479,18.93,18.901,18.93h167.121c10.422,0,18.9-8.492,18.9-18.93V29.6c0-3.846,3.124-6.974,6.963-6.974h2.984v-0.05H367.707 M367.707,20.576h-38.299h-2v0.05h-0.984c-4.942,0-8.963,4.026-8.963,8.974v4.135c0,9.335-7.581,16.93-16.9,16.93H133.44c-9.319,0-16.901-7.595-16.901-16.93V29.6c0-4.948-4.021-8.974-8.963-8.974h-0.984l-2-0.05H66.293C43.524,20.576,25,39.798,25,63.424v740.303c0,23.627,19.193,42.849,42.785,42.849h298.43c23.592,0,42.785-19.222,42.785-42.849V63.424C409,39.798,390.476,20.576,367.707,20.576L367.707,20.576z", "M262.949,40.05c-4.245,0-7.699-3.454-7.699-7.7c0-4.246,3.454-7.7,7.699-7.7c4.246,0,7.701,3.454,7.701,7.7C270.65,36.596,267.195,40.05,262.949,40.05z M262.949,26.65c-3.143,0-5.699,2.557-5.699,5.7s2.557,5.7,5.699,5.7c3.144,0,5.701-2.557,5.701-5.7S266.093,26.65,262.949,26.65z", "M240.701,36.65H194.5c-2.398,0-4.35-1.952-4.35-4.35s1.952-4.35,4.35-4.35h46.201c2.398,0,4.35,1.952,4.35,4.35S243.1,36.65,240.701,36.65z M194.5,29.95c-1.296,0-2.35,1.054-2.35,2.35s1.054,2.35,2.35,2.35h46.201c1.296,0,2.35-1.054,2.35-2.35s-1.054-2.35-2.35-2.35H194.5z"]
  },
  "iphone4_scale9": {
    type: "SVG_SCALE9",
    scale9: [164, 42, 98, 90],
    width: 282,
    height: 510
  },
  "iphone4_scale9_nw": {
    type: "SVG_PATHS",
    width: 42,
    height: 164,
    colors: [0x676767, 0x676767, 0x676767, 0xffffff, 0x676767, 0x676767],
    paths: ["M4,65.336H0V44.439h4V65.336z", "M4,115.336H0V99.785h4V115.336z", "M4,162.336H0v-15.551h4V162.336z", "M27.771,164H3.496V33c0-16.568,13.432-30,30-30H42v79.465H27.771V164z", "M5,164H2V32C2,15.432,15.432,2,32,2h10v3H32C17.112,5,5,17.112,5,32V164z", "M27.98,164h-1V81.965H42v1H27.98V164z"]
  },
  "iphone4_scale9_n": {
    type: "SVG_PATHS",
    width: 142,
    height: 164,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M0,82.465V3h142v79.465H0z", "M0,5V2h142v3H0z", "M0,82.965v-1h142v1H0z"]
  },
  "iphone4_scale9_ne": {
    type: "SVG_PATHS",
    width: 98,
    height: 164,
    colors: [0x676767, 0xffffff, 0x676767, 0x676767],
    paths: ["M42,4H3.25V0H42V4z", "M94.496,164H70.188V82.465H0V3h64.496c16.568,0,30,13.432,30,30V164z", "M96,164h-3V32C93,17.112,80.888,5,66,5H0V2h66c16.568,0,30,13.432,30,30V164z", "M70.98,164h-1V82.965H0v-1h70.98V164z"]
  },
  "iphone4_scale9_e": {
    type: "SVG_PATHS",
    width: 98,
    height: 256,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M94.496,0v256H70.188V0H94.496z", "M96,0v256h-3V0H96z", "M70.98,0v256h-1V0H70.98z"]
  },
  "iphone4_scale9_se": {
    type: "SVG_PATHS",
    width: 98,
    height: 90,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M94.496,0v59c0,16.568-13.432,30-30,30H0V2.691h70.188V0H94.496z", "M96,0v60c0,16.568-13.432,30-30,30H0v-3h66c14.888,0,27-12.112,27-27V0H96z", "M70.98,0v2.964H0v-1h69.98V0H70.98z"]
  },
  "iphone4_scale9_s": {
    type: "SVG_PATHS",
    width: 142,
    height: 90,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M0,89V2.691h142V89H0z", "M0,90v-3h142v3H0z", "M0,2.964v-1h142v1H0z"]
  },
  "iphone4_scale9_sw": {
    type: "SVG_PATHS",
    width: 42,
    height: 90,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M27.771,2.691H42V89h-8.504c-16.568,0-30-13.432-30-30V0h24.275V2.691z", "M32,87h10v3H32C15.432,90,2,76.568,2,60V0h3v60C5,74.888,17.112,87,32,87z", "M27.98,1.964H42v1H26.98V0h1V1.964z"]
  },
  "iphone4_scale9_w": {
    type: "SVG_PATHS",
    width: 42,
    height: 256,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M27.771,0v256H3.496V0H27.771z", "M5,0v256H2V0H5z", "M27.98,0v256h-1V0H27.98z"]
  },
  "iphone4_scale9_c": {
    type: "SVG_PATHS",
    width: 142,
    height: 256,
    paths: []
  },
  "iphone4_button": {
    type: "SVG_PATHS",
    width: 50,
    height: 50,
    colors: [0x676767, 0x676767],
    paths: ["M25,49.141C11.689,49.141,0.859,38.312,0.859,25S11.689,0.859,25,0.859S49.141,11.688,49.141,25S38.311,49.141,25,49.141zM25,1.859C12.24,1.859,1.859,12.24,1.859,25S12.24,48.141,25,48.141S48.141,37.76,48.141,25S37.76,1.859,25,1.859z", "M29.629,32.557h-9.258c-1.93,0-3.5-1.57-3.5-3.5v-8.113c0-1.93,1.57-3.5,3.5-3.5h9.258c1.93,0,3.5,1.57,3.5,3.5v8.113C33.129,30.986,31.559,32.557,29.629,32.557zM20.371,18.443c-1.378,0-2.5,1.121-2.5,2.5v8.113c0,1.379,1.122,2.5,2.5,2.5h9.258c1.378,0,2.5-1.121,2.5-2.5v-8.113c0-1.379-1.122-2.5-2.5-2.5H20.371z"]
  },
  "iphone4_earthing": {
    type: "SVG_PATHS",
    width: 80,
    height: 16,
    colors: [0x676767, 0x676767],
    paths: ["M74.999,11.841H31.96c-2.416,0-4.381-1.965-4.381-4.381s1.965-4.381,4.381-4.381h43.039c2.416,0,4.381,1.965,4.381,4.381S77.415,11.841,74.999,11.841zM31.96,4.079c-1.864,0-3.381,1.517-3.381,3.381s1.517,3.381,3.381,3.381h43.039c1.864,0,3.381-1.517,3.381-3.381s-1.517-3.381-3.381-3.381H31.96z", "M8.017,15.396C3.938,15.396,0.62,12.078,0.62,8s3.318-7.396,7.396-7.396S15.413,3.922,15.413,8S12.095,15.396,8.017,15.396zM8.017,1.604C4.49,1.604,1.62,4.473,1.62,8s2.87,6.396,6.396,6.396S14.413,11.527,14.413,8S11.543,1.604,8.017,1.604z"]
  },
  "iphone5_scale9": {
    type: "SVG_SCALE9",
    scale9: [167, 37, 97, 93],
    width: 282,
    height: 573
  },
  "iphone5_scale9_nw": {
    type: "SVG_PATHS",
    width: 37,
    height: 167,
    colors: [0x676767, 0x676767, 0x676767, 0xffffff, 0x676767, 0x676767],
    paths: ["M4,65.586H0V44.689h4V65.586z", "M4,115.586H0v-15.551h4V115.586z", "M4,162.586H0v-15.551h4V162.586z", "M27.771,167H3V33C3,16.432,16.432,3,33,3h4v79.715h-9.229V167z", "M5,167H2V32C2,15.432,15.432,2,32,2h5v3h-5C17.112,5,5,17.112,5,32V167z", "M27.98,167h-1V82.215H37v1h-9.02V167z"]
  },
  "iphone5_scale9_n": {
    type: "SVG_PATHS",
    width: 148,
    height: 167,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M0,82.715V3h148v79.715H0z", "M0,5V2h148v3H0z", "M0,83.215v-1h148v1H0z"]
  },
  "iphone5_scale9_ne": {
    type: "SVG_PATHS",
    width: 97,
    height: 167,
    colors: [0x676767, 0xffffff, 0x676767, 0x676767],
    paths: ["M41,4.25H2.25V0H41V4.25z", "M94,167H69.188V82.715H0V3h64c16.568,0,30,13.432,30,30V167z", "M95,167h-3V32C92,17.112,79.888,5,65,5H0V2h65c16.568,0,30,13.432,30,30V167z", "M69.98,167h-1V83.215H0v-1h69.98V167z"]
  },
  "iphone5_scale9_e": {
    type: "SVG_PATHS",
    width: 97,
    height: 313,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M94,0v313H69.188V0H94z", "M95,0v313h-3V0H95z", "M69.98,0v313h-1V0H69.98z"]
  },
  "iphone5_scale9_se": {
    type: "SVG_PATHS",
    width: 97,
    height: 93,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M94,0v63c0,16.568-13.432,30-30,30H0V6.691h69.188V0H94z", "M95,0v63c0,16.568-13.432,30-30,30H0v-3h65c14.888,0,27-12.112,27-27V0H95z", "M69.98,0v6.715H0v-1h68.98V0H69.98z"]
  },
  "iphone5_scale9_s": {
    type: "SVG_PATHS",
    width: 148,
    height: 93,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M0,93V6.691h148V93H0z", "M0,93v-3h148v3H0z", "M0,6.715v-1h148v1H0z"]
  },
  "iphone5_scale9_sw": {
    type: "SVG_PATHS",
    width: 37,
    height: 93,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M27.771,6.691H37V93h-4C16.432,93,3,79.568,3,63V0h24.771V6.691z", "M32,90h5v3h-5C15.432,93,2,79.568,2,63V0h3v63C5,77.888,17.112,90,32,90z", "M27.98,5.715H37v1H26.98V0h1V5.715z"]
  },
  "iphone5_scale9_w": {
    type: "SVG_PATHS",
    width: 37,
    height: 313,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M27.771,0v313H3V0H27.771z", "M5,0v313H2V0H5z", "M27.98,0v313h-1V0H27.98z"]
  },
  "iphone5_scale9_c": {
    type: "SVG_PATHS",
    width: 148,
    height: 313,
    paths: []
  },
  "iphone5_button": {
    type: "SVG_PATHS",
    width: 50,
    height: 50,
    colors: [0x676767, 0x676767],
    paths: ["M25,49.141C11.689,49.141,0.859,38.312,0.859,25S11.689,0.859,25,0.859S49.141,11.688,49.141,25S38.311,49.141,25,49.141zM25,1.859C12.24,1.859,1.859,12.24,1.859,25S12.24,48.141,25,48.141S48.141,37.76,48.141,25S37.76,1.859,25,1.859z", "M29.629,32.557h-9.258c-1.93,0-3.5-1.57-3.5-3.5v-8.113c0-1.93,1.57-3.5,3.5-3.5h9.258c1.93,0,3.5,1.57,3.5,3.5v8.113C33.129,30.986,31.559,32.557,29.629,32.557zM20.371,18.443c-1.378,0-2.5,1.121-2.5,2.5v8.113c0,1.379,1.122,2.5,2.5,2.5h9.258c1.378,0,2.5-1.121,2.5-2.5v-8.113c0-1.379-1.122-2.5-2.5-2.5H20.371z"]
  },
  "iphone5_earthing": {
    type: "SVG_PATHS",
    width: 52,
    height: 32,
    colors: [0x676767, 0x676767],
    paths: ["M47.52,31.619H4.48c-2.416,0-4.381-1.965-4.381-4.381s1.965-4.381,4.381-4.381H47.52c2.416,0,4.381,1.965,4.381,4.381S49.935,31.619,47.52,31.619zM4.48,23.857c-1.864,0-3.381,1.517-3.381,3.381s1.517,3.381,3.381,3.381H47.52c1.864,0,3.381-1.517,3.381-3.381s-1.517-3.381-3.381-3.381H4.48z", "M26,15.174c-4.079,0-7.396-3.318-7.396-7.396S21.921,0.381,26,0.381s7.396,3.318,7.396,7.396S30.079,15.174,26,15.174zM26,1.381c-3.527,0-6.396,2.869-6.396,6.396s2.87,6.396,6.396,6.396s6.396-2.869,6.396-6.396S29.527,1.381,26,1.381z"]
  },
  "iphone6_scale9": {
    type: "SVG_SCALE9",
    scale9: [265, 57, 69, 101],
    width: 380,
    height: 754
  },
  "iphone6_scale9_nw": {
    type: "SVG_PATHS",
    width: 57,
    height: 265,
    colors: [0xffffff, 0x676767, 0xffffff, 0x676767, 0xffffff, 0x676767, 0xffffff, 0x676767, 0x676767],
    paths: ["M1,92.65c0-1.109,0.893-2.008,1.995-2.008h8.428c1.102,0,1.995,0.885,1.995,2.008v24.541c0,1.109-0.893,2.009-1.995,2.009H2.995C1.893,119.2,1,118.314,1,117.191V92.65z", "M11.423,120.2H2.995c-1.651,0-2.995-1.35-2.995-3.009V92.65c0-1.658,1.344-3.008,2.995-3.008h8.428c1.651,0,2.995,1.35,2.995,3.008v24.541C14.418,118.851,13.074,120.2,11.423,120.2zM2.995,91.643C2.446,91.643,2,92.095,2,92.65v24.541c0,0.565,0.437,1.009,0.995,1.009h8.428c0.549,0,0.995-0.452,0.995-1.009V92.65c0-0.565-0.437-1.008-0.995-1.008H2.995z", "M1,205.63c0-1.103,0.893-1.997,1.995-1.997h8.428c1.102,0,1.995,0.906,1.995,1.997v48.157c0,1.103-0.893,1.996-1.995,1.996H2.995c-1.102,0-1.995-0.905-1.995-1.996V205.63z", "M11.423,256.783H2.995c-1.651,0-2.995-1.344-2.995-2.996V205.63c0-1.652,1.344-2.997,2.995-2.997h8.428c1.651,0,2.995,1.345,2.995,2.997v48.157C14.418,255.439,13.074,256.783,11.423,256.783zM2.995,204.633C2.446,204.633,2,205.08,2,205.63v48.157c0,0.54,0.456,0.996,0.995,0.996h8.428c0.549,0,0.995-0.447,0.995-0.996V205.63c0-0.54-0.456-0.997-0.995-0.997H2.995z", "M1,147.271c0-1.104,0.893-1.997,1.995-1.997h8.428c1.102,0,1.995,0.906,1.995,1.997v48.157c0,1.103-0.893,1.995-1.995,1.995H2.995c-1.102,0-1.995-0.903-1.995-1.995V147.271z", "M11.423,198.424H2.995c-1.651,0-2.995-1.344-2.995-2.995v-48.157c0-1.652,1.344-2.997,2.995-2.997h8.428c1.651,0,2.995,1.345,2.995,2.997v48.157C14.418,197.08,13.074,198.424,11.423,198.424zM2.995,146.274c-0.549,0-0.995,0.447-0.995,0.997v48.157c0,0.549,0.446,0.995,0.995,0.995h8.428c0.549,0,0.995-0.446,0.995-0.995v-48.157c0-0.54-0.456-0.997-0.995-0.997H2.995z", "M28,265H4V50.001C4,22.386,26.38,0,54,0h3v87H28V265z", "M6,265H4V50.001C4,22.386,26.38,0,54,0h3v2h-3C27.533,2,6,23.533,6,50.001V265z", "M29,265h-2V87h30v2H29V265z"]
  },
  "iphone6_scale9_n": {
    type: "SVG_PATHS",
    width: 254,
    height: 265,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M0,87V0h254v87H0z", "M0,2V0h254v2H0z", "M0,89v-2h254v2H0z"]
  },
  "iphone6_scale9_ne": {
    type: "SVG_PATHS",
    width: 69,
    height: 265,
    colors: [0xffffff, 0x676767, 0xffffff, 0x676767, 0x676767],
    paths: ["M55.582,147.271c0-1.104,0.893-1.997,1.994-1.997h8.429c1.103,0,1.995,0.906,1.995,1.997v48.157c0,1.103-0.893,1.995-1.995,1.995h-8.429c-1.102,0-1.994-0.903-1.994-1.995V147.271z", "M66.005,198.424h-8.429c-1.651,0-2.994-1.344-2.994-2.995v-48.157c0-1.652,1.343-2.997,2.994-2.997h8.429c1.651,0,2.995,1.345,2.995,2.997v48.157C69,197.08,67.656,198.424,66.005,198.424zM57.576,146.274c-0.548,0-0.994,0.447-0.994,0.997v48.157c0,0.549,0.446,0.995,0.994,0.995h8.429c0.549,0,0.995-0.446,0.995-0.995v-48.157c0-0.54-0.456-0.997-0.995-0.997H57.576z", "M65,265H40V87H0V0h15c27.613,0,50,22.387,50,50.001V265z", "M65,265h-2V50.001C63,23.533,41.467,2,15,2H0V0h15c27.613,0,50,22.387,50,50.001V265z", "M42,265h-2V89H0v-2h42V265z"]
  },
  "iphone6_scale9_e": {
    type: "SVG_PATHS",
    width: 69,
    height: 388,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M65,0v388H40V0H65z", "M65,0v388h-2V0H65z", "M42,0v388h-2V0H42z"]
  },
  "iphone6_scale9_se": {
    type: "SVG_PATHS",
    width: 69,
    height: 101,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M65,0v50.999C65,78.613,42.619,101,15,101H0V8h40V0H65z", "M65,0v50.999C65,78.613,42.619,101,15,101H0v-2h15c26.467,0,48-21.533,48-48.001V0H65z", "M42,0v8H0V6h40V0H42z"]
  },
  "iphone6_scale9_s": {
    type: "SVG_PATHS",
    width: 254,
    height: 101,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M0,101V8h254v93H0z", "M0,101v-2h254v2H0z", "M0,8V6h254v2H0z"]
  },
  "iphone6_scale9_sw": {
    type: "SVG_PATHS",
    width: 57,
    height: 101,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M28,8h29v93h-3C26.386,101,4,78.613,4,50.999V0h24V8z", "M54,99h3v2h-3C26.386,101,4,78.613,4,50.999V0h2v50.999C6,77.467,27.533,99,54,99z", "M29,6h28v2H27V0h2V6z"]
  },
  "iphone6_scale9_w": {
    type: "SVG_PATHS",
    width: 57,
    height: 388,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M28,0v388H4V0H28z", "M6,0v388H4V0H6z", "M29,0v388h-2V0H29z"]
  },
  "iphone6_scale9_c": {
    type: "SVG_PATHS",
    width: 254,
    height: 388,
    paths: []
  },
  "iphone6_button": {
    type: "SVG_PATHS",
    width: 56,
    height: 56,
    colors: [0x676767],
    paths: ["M28,55.075c14.955,0,27.078-12.122,27.078-27.075S42.955,0.925,28,0.925C13.046,0.925,0.922,13.047,0.922,28S13.046,55.075,28,55.075M28,53.075C14.149,53.075,2.922,41.849,2.922,28S14.149,2.925,28,2.925S53.078,14.151,53.078,28S41.851,53.075,28,53.075"]
  },
  "iphone6_earthing": {
    type: "SVG_PATHS",
    width: 84,
    height: 34,
    colors: [0x676767, 0x676767, 0x676767],
    paths: ["M34.967,27.166L34.967,27.166c0,1.037,0.832,1.863,1.858,1.863h42.232c1.021,0,1.857-0.834,1.857-1.863c0-1.035-0.832-1.861-1.857-1.861H36.825C35.804,25.305,34.967,26.139,34.967,27.166M32.967,27.166L32.967,27.166c0-2.131,1.73-3.861,3.858-3.861h42.232c2.133,0,3.857,1.725,3.857,3.861c0,2.131-1.73,3.863-3.857,3.863H36.825C34.692,31.029,32.967,29.305,32.967,27.166", "M58.314,10.996c2.952,0,5.347-2.395,5.347-5.346s-2.395-5.346-5.347-5.346c-2.954,0-5.347,2.395-5.347,5.346S55.36,10.996,58.314,10.996M58.314,8.996c-1.849,0-3.347-1.498-3.347-3.346s1.498-3.346,3.347-3.346s3.347,1.498,3.347,3.346S60.163,8.996,58.314,8.996", "M7.673,33.695c3.638,0,6.588-2.949,6.588-6.586c0-3.639-2.95-6.588-6.588-6.588c-3.639,0-6.588,2.949-6.588,6.588C1.085,30.746,4.034,33.695,7.673,33.695M7.673,31.695c-2.534,0-4.588-2.053-4.588-4.586c0-2.535,2.054-4.588,4.588-4.588s4.588,2.053,4.588,4.588C12.261,29.643,10.208,31.695,7.673,31.695"]
  },
  "iphoneX_scale9": {
    type: "SVG_SCALE9",
    scale9: [325, 142, 136, 62],
    width: 434,
    height: 867
  },
  "iphoneX_scale9_nw": {
    type: "SVG_PATHS",
    width: 142,
    height: 325,
    colors: [0xffffff, 0x676767, 0x676767, 0x676767, 0x676767, 0x676767],
    paths: ["M27,325H3.04V64.471c0-35.199,28.8-64,64-64H142v52.194h-8.561c-10.422,0-18.9-8.492-18.9-18.931V29.6c0-3.846-3.123-6.974-6.963-6.974h-2.984v-0.05H66.293C44.627,22.576,27,40.9,27,63.424V325z", "M3,150.55H2c-1.103,0-2-0.897-2-2v-31c0-1.103,0.897-2,2-2h1c1.103,0,2,0.897,2,2v31C5,149.652,4.103,150.55,3,150.55zM3,148.55v1V148.55L3,148.55L3,148.55zM2,117.55v31h0.997l0.003-31H2z", "M3,241.55H2c-1.103,0-2-0.897-2-2v-60c0-1.103,0.897-2,2-2h1c1.103,0,2,0.897,2,2v60C5,240.652,4.103,241.55,3,241.55zM3,239.55v1V239.55L3,239.55L3,239.55zM2,179.55v60h0.997l0.003-60H2z", "M3,320.55H2c-1.103,0-2-0.897-2-2v-60c0-1.103,0.897-2,2-2h1c1.103,0,2,0.897,2,2v60C5,319.652,4.103,320.55,3,320.55zM3,318.55v1V318.55L3,318.55L3,318.55zM2,258.55v60h0.997l0.003-60H2z", "M5.04,325h-2V64.471c0-35.199,28.8-64,64-64H142v2H67.04c-34.187,0-62,27.813-62,62V325z", "M27,325h-2V63.424c0-23.626,18.523-42.848,41.293-42.848h38.299l2,0.05h0.984c4.941,0,8.963,4.026,8.963,8.974v4.135c0,9.336,7.582,16.931,16.9,16.931H142v2h-8.561c-10.422,0-18.9-8.492-18.9-18.931V29.6c0-3.846-3.123-6.974-6.963-6.974h-2.984v-0.05H66.293C44.627,22.576,27,40.9,27,63.424V325z"]
  },
  "iphoneX_scale9_n": {
    type: "SVG_PATHS",
    width: 156,
    height: 325,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M0,52.665V0.471h156v52.194H0z", "M0,2.471v-2h156v2H0z", "M0,52.665v-2h156v2H0z"]
  },
  "iphoneX_scale9_ne": {
    type: "SVG_PATHS",
    width: 136,
    height: 325,
    colors: [0xffffff, 0x676767, 0x676767, 0x676767],
    paths: ["M133.04,325H109V63.424c0-22.523-17.627-40.848-39.293-40.848H31.408v0.05h-2.984c-3.839,0-6.963,3.128-6.963,6.974v4.135c0,10.438-8.478,18.931-18.9,18.931H0V0.471h69.04c35.2,0,64,28.801,64,64V325z", "M134,261.55h-1c-1.103,0-2-0.897-2-2v-60c0-1.103,0.897-2,2-2h1c1.103,0,2,0.897,2,2v60C136,260.652,135.103,261.55,134,261.55zM134,259.55v1V259.55L134,259.55L134,259.55zM133,199.55v60h0.997l0.003-60H133L133,199.55z", "M133.04,325h-2V64.471c0-34.187-27.813-62-62-62H0v-2h69.04c35.2,0,64,28.801,64,64V325z", "M111,325h-2V63.424c0-22.523-17.627-40.848-39.293-40.848H31.408v0.05h-2.984c-3.839,0-6.963,3.128-6.963,6.974v4.135c0,10.438-8.478,18.931-18.9,18.931H0v-2h2.561c9.319,0,16.9-7.595,16.9-16.931V29.6c0-4.947,4.021-8.974,8.963-8.974h0.984v-0.05h2h38.299C92.476,20.576,111,39.798,111,63.424V325z"]
  },
  "iphoneX_scale9_e": {
    type: "SVG_PATHS",
    width: 136,
    height: 480,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M133.04,0v477.471c0,0.848-0.031,1.689-0.064,2.529h-24.008c0.013-0.423,0.032-0.845,0.032-1.271V0H133.04z", "M133.04,0v477.471c0,0.848-0.031,1.689-0.064,2.529h-2c0.034-0.84,0.064-1.681,0.064-2.529V0H133.04z", "M111,0v478.727c0,0.427-0.02,0.85-0.032,1.273h-2c0.013-0.424,0.032-0.847,0.032-1.273V0H111z"]
  },
  "iphoneX_scale9_se": {
    type: "SVG_PATHS",
    width: 136,
    height: 62,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M132.976,0c-1.342,34.038-29.583,61.471-63.936,61.471H0V39.577h68.215c22.063,0,40.078-17.641,40.753-39.577H132.976z", "M132.976,0c-1.342,34.038-29.583,61.471-63.936,61.471H0v-2h69.04c33.338,0,60.603-26.453,61.936-59.471H132.976z", "M110.968,0c-0.677,23.039-19.588,41.576-42.753,41.576H0v-2h68.215c22.062,0,40.077-17.64,40.753-39.576H110.968z"]
  },
  "iphoneX_scale9_s": {
    type: "SVG_PATHS",
    width: 156,
    height: 62,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M0,61.471V39.577h156v21.894H0z", "M0,61.471v-2h156v2H0z", "M0,41.576v-2h156v2H0z"]
  },
  "iphoneX_scale9_sw": {
    type: "SVG_PATHS",
    width: 142,
    height: 62,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M67.785,39.577H142v21.894H67.04C32.688,61.471,4.446,34.038,3.104,0h23.928C27.707,21.937,45.722,39.576,67.785,39.577z", "M67.04,59.471H142v2H67.04C32.688,61.471,4.446,34.038,3.104,0h2C6.437,33.018,33.702,59.471,67.04,59.471z", "M67.785,39.576H142v2H67.785C44.62,41.576,25.709,23.039,25.032,0h2C27.707,21.937,45.722,39.576,67.785,39.576z"]
  },
  "iphoneX_scale9_w": {
    type: "SVG_PATHS",
    width: 142,
    height: 480,
    colors: [0xffffff, 0x676767, 0x676767],
    paths: ["M27,478.729c0,0.427,0.019,0.849,0.032,1.271H3.104c-0.033-0.84-0.064-1.682-0.064-2.529V0H27V478.729z", "M5.04,477.471c0,0.849,0.03,1.689,0.064,2.529h-2c-0.033-0.84-0.064-1.682-0.064-2.529V0h2V477.471z", "M27,478.729c0,0.427,0.019,0.849,0.032,1.271h-2C25.02,479.576,25,479.153,25,478.727V0h2V478.729z"]
  },
  "iphoneX_scale9_c": {
    type: "SVG_PATHS",
    width: 156,
    height: 480,
    paths: []
  },
  "iphoneX_earthing": {
    type: "SVG_PATHS",
    width: 81,
    height: 16,
    colors: [0x676767, 0x676767],
    paths: ["M72.799,15.4c-4.244,0-7.699-3.454-7.699-7.7S68.555,0,72.799,0C77.045,0,80.5,3.454,80.5,7.7S77.045,15.4,72.799,15.4zM72.799,2C69.656,2,67.1,4.557,67.1,7.7s2.557,5.7,5.699,5.7c3.145,0,5.701-2.557,5.701-5.7S75.943,2,72.799,2z", "M50.551,12H4.35C1.952,12,0,10.049,0,7.65s1.952-4.35,4.35-4.35h46.201c2.398,0,4.35,1.951,4.35,4.35S52.949,12,50.551,12zM4.35,5.301C3.054,5.301,2,6.354,2,7.65C2,8.946,3.054,10,4.35,10h46.201c1.297,0,2.35-1.054,2.35-2.35c0-1.297-1.053-2.35-2.35-2.35H4.35z"]
  },
  "iphone_top_bar_items_color": 0x333333,
  "iphonemenu_right_text_color": 0x324F85,
  "iphonemenu_min_row_height": 27,
  "iphonemenu_padding_top": 5,
  "iphonemenu_padding_bottom": 5,
  "iphonemenu_row_padding": 2,
  "iphonemenu_background_color": 0xFFFFFF,
  "iphonemenu_padding_left": 8,
  "iphonemenu_padding_right": 6,
  "iphonemenu_icon_padding": 4,
  "iphonemenu_corner_radius": 7,
  "iphonepicker_line_color": 0xDDDDDD,
  "android_nexus": {
    type: "SVG_PATHS",
    width: 356,
    height: 725,
    colors: [0xffffff, 0x676767, 0x676767, 0x676767, 0x676767, 0x676767],
    paths: ["M322.005,2H33.995C16.317,2,2,16.326,2,33.998v657.003C2,708.679,16.325,723,33.995,723h288.01C339.683,723,354,708.674,354,691.001V33.998C354,16.321,339.676,2,322.005,2zM338,646H18V78h320V646z", "M322.005,724H33.995C15.802,724,1,709.196,1,691.001V33.998C1,15.803,15.802,1,33.995,1h288.01C340.198,1,355,15.803,355,33.998v657.003C355,709.196,340.198,724,322.005,724zM33.995,3C16.904,3,3,16.906,3,33.998v657.003C3,708.094,16.904,722,33.995,722h288.01C339.096,722,353,708.094,353,691.001V33.998C353,16.906,339.096,3,322.005,3H33.995zM339,647H17V77h322V647zM19,645h318V79H19V645z", "M339,647H17V77h322V647zM19,645h318V79H19V645z", "M221.847,704.061H135c-7.264,0-21-1.156-21-5.53s13.736-5.53,21-5.53h86.847c7.264,0,21,1.156,21,5.53S229.11,704.061,221.847,704.061zM135,695c-12.536,0-19,2.475-19,3.53s6.464,3.53,19,3.53h86.847c12.536,0,19-2.475,19-3.53s-6.464-3.53-19-3.53H135z", "M221.847,40.061H135c-7.264,0-21-1.156-21-5.53S127.736,29,135,29h86.847c7.264,0,21,1.156,21,5.53S229.11,40.061,221.847,40.061zM135,31c-12.536,0-19,2.475-19,3.53s6.464,3.53,19,3.53h86.847c12.536,0,19-2.475,19-3.53s-6.464-3.53-19-3.53H135z", "M73.352,41.055c-3.049,0-5.53-2.481-5.53-5.531c0-3.049,2.481-5.529,5.53-5.529s5.53,2.48,5.53,5.529C78.882,38.573,76.401,41.055,73.352,41.055zM73.352,31.994c-1.947,0-3.53,1.583-3.53,3.529c0,1.947,1.583,3.531,3.53,3.531s3.53-1.584,3.53-3.531C76.882,33.577,75.298,31.994,73.352,31.994z"]
  },
  "android_nexus_scale9": {
    type: "SVG_SCALE9",
    scale9: [83, 37, 38, 85],
    width: 356,
    height: 725
  },
  "android_nexus_scale9_nw": {
    type: "SVG_PATHS",
    width: 37,
    height: 83,
    colors: [0xffffff, 0x666666, 0x666666],
    paths: ["M18,83H2V33.999C2,16.327,16.317,2,33.995,2H37v76H18V83z", "M2,83H0V33.999C0,15.22,15.214,0,33.995,0H37v2h-3.005C16.317,2,2,16.327,2,33.999V83z", "M20,83h-2v-5h19v2H20V83z"]
  },
  "android_nexus_scale9_n": {
    type: "SVG_PATHS",
    width: 281,
    height: 83,
    colors: [0xffffff, 0x666666, 0x666666],
    paths: ["M0,78V2h281v76H0z", "M0,2V0h281v2H0z", "M0,80v-2h281v2H0z"]
  },
  "android_nexus_scale9_ne": {
    type: "SVG_PATHS",
    width: 38,
    height: 83,
    colors: [0xffffff, 0x666666, 0x666666],
    paths: ["M36,83H20v-5H0V2h4.005C21.676,2,36,16.32,36,33.999V83z", "M38,83h-2V33.999C36,16.32,21.676,2,4.005,2H0V0h4.005C22.781,0,38,15.219,38,33.999V83z", "M20,83h-2v-3H0v-2h20V83z"]
  },
  "android_nexus_scale9_e": {
    type: "SVG_PATHS",
    width: 38,
    height: 557,
    colors: [0xffffff, 0x666666, 0x666666],
    paths: ["M36,0v557H20V0H36z", "M38,0v557h-2V0H38z", "M20,0v557h-2V0H20z"]
  },
  "android_nexus_scale9_se": {
    type: "SVG_PATHS",
    width: 38,
    height: 85,
    colors: [0xffffff, 0x666666, 0x666666],
    paths: ["M36,0v51.001C36,68.674,21.682,83,4.005,83H0V6h20V0H36z", "M38,0v51.001C38,69.78,22.785,85,4.005,85H0v-2h4.005C21.682,83,36,68.674,36,51.001V0H38z", "M20,0v6H0V4h18V0H20z"]
  },
  "android_nexus_scale9_s": {
    type: "SVG_PATHS",
    width: 281,
    height: 85,
    colors: [0xffffff, 0x666666, 0x666666],
    paths: ["M0,83V6h281v77H0z", "M0,85v-2h281v2H0z", "M0,6V4h281v2H0z"]
  },
  "android_nexus_scale9_sw": {
    type: "SVG_PATHS",
    width: 37,
    height: 85,
    colors: [0xffffff, 0x666666, 0x666666],
    paths: ["M18,6h19v77h-3.005C16.325,83,2,68.678,2,51.001V0h16V6z", "M33.995,83H37v2h-3.005C15.218,85,0,69.781,0,51.001V0h2v51.001C2,68.678,16.325,83,33.995,83z", "M20,4h17v2H18V0h2V4z"]
  },
  "android_nexus_scale9_w": {
    type: "SVG_PATHS",
    width: 37,
    height: 557,
    colors: [0xffffff, 0x666666, 0x666666],
    paths: ["M18,0v557H2V0H18z", "M2,0v557H0V0H2z", "M20,0v557h-2V0H20z"]
  },
  "android_nexus_scale9_c": {
    type: "SVG_PATHS",
    width: 281,
    height: 557,
    paths: []
  },
  "android_nexus_ear_thing": {
    type: "SVG_PATHS",
    width: 177,
    height: 14,
    colors: [0x666666, 0x666666],
    paths: ["M2,7.494C2,5.002,4.029,2.982,6.53,2.982c2.5,0,4.529,2.02,4.529,4.512s-2.028,4.514-4.529,4.514C4.029,12.008,2,9.986,2,7.494M0,7.494C0,11.088,2.923,14,6.53,14c3.605,0,6.528-2.912,6.528-6.506c0-3.593-2.923-6.504-6.528-6.504C2.923,0.99,0,3.901,0,7.494", "M48.172,6.505c0-2.493,2.031-4.513,4.534-4.513h117.763c2.503,0,4.532,2.025,4.532,4.513c0,2.492-2.029,4.513-4.532,4.513H52.706C50.202,11.018,48.172,8.992,48.172,6.505M52.706,0c-3.608,0-6.534,2.912-6.534,6.505c0,3.589,2.927,6.505,6.534,6.505h117.763c3.606,0,6.531-2.912,6.531-6.505C177,2.915,174.074,0,170.469,0H52.706z"]
  },
  "android_nexus_button": {
    type: "SVG_PATHS",
    width: 131,
    height: 13,
    colors: [0x666666],
    paths: ["M2.077,6.5c0-2.49,2.031-4.509,4.534-4.509h117.78c2.504,0,4.532,2.023,4.532,4.509c0,2.49-2.029,4.509-4.532,4.509H6.611C4.106,11.009,2.077,8.986,2.077,6.5M6.611,0C3.002,0,0.077,2.91,0.077,6.5c0,3.587,2.927,6.5,6.534,6.5h117.78c3.606,0,6.532-2.911,6.532-6.5c0-3.588-2.927-6.5-6.532-6.5H6.611z"]
  },
  "label_font_family": "Balsamiq Sans",
  "label_default_font_size": 13,
  "label_padding_top": 4,
  "label_padding_bottom": 4,
  "label_padding_left": 4,
  "label_padding_right": 4,
  "title_padding_top": 4,
  "subtitle_padding_top": 1,
  "titles_padding_bottom": 4,
  "titles_padding_left": 0,
  "titles_padding_right": 0,
  "line_vertical": {
    type: "SVG_PATHS",
    width: 2,
    height: 48,
    paths: ["M0,43.548V2.484C0,2.484,0.415,0,2,0c0,15.963,0,47.947-0.668,47.889C0,48.534,0,43.548,0,43.548z"]
  },
  "linechart_axis_color": 0x333333,
  "linechart_axis_thickness": 2,
  "linechart_min_thickness": 2,
  "linechart_max_thickness": 2,
  "linechart_series1_color": 0xE6E6E6,
  "linechart_series2_color": 0xA7A7A7,
  "link_padding_top": 4,
  "link_padding_bottom": 4,
  "linkbar_padding_top": 4,
  "linkbar_padding_bottom": 4,
  "linkbar_padding_middle": 8,
  "list_padding_top": 2,
  "list_padding_bottom": 3,
  "list_padding_left": 8,
  "list_padding_right": 8,
  "list_line_padding": 8,
  "list_header_color": 0xCCCCCC,
  "list_scrollbar_padding_top": 0,
  "list_scrollbar_padding_bottom": 0,
  "list_scrollbar_padding_right": 0,
  "list_scrollbar_padding_no_border": 0,
  "list_scrollbar_header_padding_top": -1,
  "list_icon_padding": 16,
  "menu_padding_middle": 8,
  "menu_padding_top": 1,
  "menu_padding_bottom": 1,
  "menu_padding_right": 1,
  "menu_padding_left": 20,
  "menu_asset_padding_left": 1,
  "menu_asset_padding_right": 4,
  "menu_separator_skin": "pixel",
  "menu_rule_padding": 1,
  "menubar_entry_padding": 9,
  "menubar_padding_left": 9,
  "menubar_padding_right": 9,
  "menubar_padding_top": 2,
  "menubar_padding_bottom": 1,
  "multilinebutton_default_font_size": 16,
  "multilinebutton_background_color": 0xFFFFFF,
  "multilinebutton_border_color": 0x000000,
  "multilinebutton_padding_left": 5,
  "multilinebutton_padding_right": 5,
  "multilinebutton_padding_top": 16,
  "multilinebutton_padding_bottom": 14,
  "multilinebutton_padding_middle": 8,
  "multilinebutton_icon_padding": 3,
  "multilinebutton_disabled_color": 0x999999,
  "multilinebutton_background_radius": 6,
  "multilinebutton_default_icon_color": 0x333333,
  "numericstepper": {
    type: "SVG_SCALE9",
    scale9: [3, 3, 20, 3],
    width: 44,
    height: 24
  },
  "numericstepper_nw": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    paths: ["M1,3H0V0h3v1H1V3z"]
  },
  "numericstepper_n": {
    type: "SVG_PATHS",
    width: 21,
    height: 3,
    paths: ["M0,1V0h21v1H0z"]
  },
  "numericstepper_ne": {
    type: "SVG_PATHS",
    width: 20,
    height: 3,
    paths: ["M20,3h-1V1H0V0h20V3z", "M2.99,3h-1V0h1V3z"]
  },
  "numericstepper_e": {
    type: "SVG_PATHS",
    width: 20,
    height: 18,
    paths: ["M20,0v18h-1V0H20z", "M2.99,0v18h-1V0H2.99z", "M6.162,7.947l5-6l5,6H6.162z", "M16.162,9.952l-5,6l-5-6H16.162z"]
  },
  "numericstepper_se": {
    type: "SVG_PATHS",
    width: 20,
    height: 3,
    paths: ["M20,0v3H0V2h19V0H20z", "M2.99,0v2.997h-1V0H2.99z"]
  },
  "numericstepper_s": {
    type: "SVG_PATHS",
    width: 21,
    height: 3,
    paths: ["M0,3V2h21v1H0z"]
  },
  "numericstepper_sw": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    paths: ["M1,2h2v1H0V0h1V2z"]
  },
  "numericstepper_w": {
    type: "SVG_PATHS",
    width: 3,
    height: 18,
    paths: ["M1,0v18H0V0H1z"]
  },
  "numericstepper_c": {
    type: "SVG_PATHS",
    width: 21,
    height: 18,
    paths: []
  },
  "numericstepper_padding_left": 9,
  "numericstepper_padding_right": 24,
  "numericstepper_padding_top": 6,
  "numericstepper_padding_bottom": 5,
  "numericstepper_button_width": 18,
  "numericstepper_button_background_color": 0xDDDDDD,
  "numericstepper_disabled_color": 0xCDCCCC,
  "piechart_background": {
    type: "SVG_PATHS",
    width: 180,
    height: 180,
    colors: [0xA7A7A7, 0xE6E6E6],
    paths: ["M90,90l-45.006,77.953C18.097,152.391,0,123.31,0,90C0,40.295,40.294,0,90,0V90z", "M180,90c0,49.706-40.294,90-90,90c-16.396,0-31.767-4.387-45.006-12.047L90,90V0C139.706,0,180,40.295,180,90z"]
  },
  "piechart_border_color": 0x333333,
  "piechart_min_thickness": 2,
  "piechart_max_thickness": 2,
  "pixel": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,1H0V0h1V1z"]
  },
  "playback": {
    type: "SVG_PATHS",
    width: 110,
    height: 36,
    colors: [0XDDDDDD, 0X666666, 0X666666, 0XDDDDDD, 0X666666, 0X666666, 0XDDDDDD, 0X666666],
    paths: ["M108.013,19.211c0,7.168-5.831,13-12.999,13c-7.169,0-13.001-5.832-13.001-13s5.832-13,13.001-13C102.182,6.211,108.013,12.043,108.013,19.211z", "M97.2,14.211l6,5l-6,5V14.211z", "M89.7,14.211l6,5l-6,5V14.211z", "M28.013,19.211c0,7.168-5.832,13-13.001,13c-7.167,0-12.999-5.832-12.999-13s5.832-13,12.999-13C22.181,6.211,28.013,12.043,28.013,19.211z", "M12.825,24.211l-6-5l6-5V24.211z", "M20.325,24.211l-6-5l6-5V24.211z", "M73,18c0,9.926-8.075,18-18,18c-9.925,0-18-8.074-18-18S45.075,0,55,0C64.925,0,73,8.074,73,18z", "M50.377,9.092L63.85,18l-13.473,8.908V9.092z"]
  },
  "pointybutton_default_font_size": 13,
  "pointybutton_background_color": 0xFFFFFF,
  "pointybutton_border_color": 0x999999,
  "pointybutton_padding_left": 10,
  "pointybutton_padding_right": 10,
  "pointybutton_padding_top": 0,
  "pointybutton_padding_bottom": 0,
  "pointybutton_icon_padding": 3,
  "pointybutton_point_padding": 20,
  "pointybutton_disabled_color": 0x80808B,
  "pointybutton_noborder_height": 27,
  "pointybutton_background_border_radius": 5,
  "pointybutton_center": {
    type: "SVG_SCALE9",
    scale9: [4, 4, 4, 4],
    width: 36,
    height: 27
  },
  "pointybutton_center_nw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M0.966,4h-1c0.007-2.206,1.802-4,4.01-4H4v1H3.976C2.319,1,0.973,2.345,0.966,4z"]
  },
  "pointybutton_center_n": {
    type: "SVG_PATHS",
    width: 28,
    height: 4,
    paths: ["M0,1V0h28v1H0z"]
  },
  "pointybutton_center_ne": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M4.033,4h-1C3.026,2.345,1.68,1,0.023,1H0V0h0.024C2.231,0,4.026,1.794,4.033,4z"]
  },
  "pointybutton_center_e": {
    type: "SVG_PATHS",
    width: 4,
    height: 19,
    paths: ["M4.033,0c0,0.004,0.001,0.008,0.001,0.012v18.977c0,0.004-0.001,0.008-0.001,0.012h-1c0-0.004,0.001-0.008,0.001-0.012V0.012c0-0.004-0.001-0.008-0.001-0.012H4.033z"]
  },
  "pointybutton_center_se": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M4.033,0c-0.007,2.206-1.802,4-4.01,4H0V3h0.024c1.656,0,3.003-1.345,3.01-3H4.033z"]
  },
  "pointybutton_center_s": {
    type: "SVG_PATHS",
    width: 28,
    height: 4,
    paths: ["M0,4V3h28v1H0z"]
  },
  "pointybutton_center_sw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M3.976,3H4v1H3.976c-2.208,0-4.003-1.794-4.01-4h1C0.973,1.655,2.319,3,3.976,3z"]
  },
  "pointybutton_center_w": {
    type: "SVG_PATHS",
    width: 4,
    height: 19,
    paths: ["M0.965,18.988c0,0.004,0.001,0.008,0.001,0.012h-1c0-0.004-0.001-0.008-0.001-0.012V0.012c0-0.004,0.001-0.008,0.001-0.012h1c0,0.004-0.001,0.008-0.001,0.012V18.988z"]
  },
  "pointybutton_center_c": {
    type: "SVG_PATHS",
    width: 28,
    height: 19,
    paths: []
  },
  "pointybutton_center_background": {
    type: "SVG_SCALE9",
    scale9: [4, 4, 4, 4],
    width: 36,
    height: 27
  },
  "pointybutton_center_background_nw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M4,4H0.5V2.5c0-1.104,0.896-2,2-2H4V4z"]
  },
  "pointybutton_center_background_n": {
    type: "SVG_PATHS",
    width: 28,
    height: 4,
    paths: ["M28,4H0V0.5h28V4z"]
  },
  "pointybutton_center_background_ne": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M3.499,4H0V0.5h1.5c1.104,0,2,0.896,2,2V4z"]
  },
  "pointybutton_center_background_e": {
    type: "SVG_PATHS",
    width: 4,
    height: 19,
    paths: ["M3.499,0v19H0V0H3.499z"]
  },
  "pointybutton_center_background_se": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M3.499,0v1.5c0,1.104-0.896,2-2,2H0V0H3.499z"]
  },
  "pointybutton_center_background_s": {
    type: "SVG_PATHS",
    width: 28,
    height: 4,
    paths: ["M28,0v3.5H0V0H28z"]
  },
  "pointybutton_center_background_sw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M4,0v3.5H2.5c-1.104,0-2-0.896-2-2V0H4z"]
  },
  "pointybutton_center_background_w": {
    type: "SVG_PATHS",
    width: 4,
    height: 19,
    paths: ["M4,0v19H0.5V0H4z"]
  },
  "pointybutton_center_background_c": {
    type: "SVG_PATHS",
    width: 28,
    height: 19,
    paths: ["M28,0v19H0V0H28z"]
  },
  "pointybutton_left": {
    type: "SVG_SCALE9",
    scale9: [3, 17, 3, 3],
    width: 42,
    height: 27
  },
  "pointybutton_left_nw": {
    type: "SVG_PATHS",
    width: 17,
    height: 3,
    paths: ["M12.293,3h-1.818c2.258-1.455,4.409-2.574,5.891-2.982L16.499,0L17,0.001V1h-0.432C15.366,1.342,13.874,2.063,12.293,3z"]
  },
  "pointybutton_left_n": {
    type: "SVG_PATHS",
    width: 22,
    height: 3,
    paths: ["M0,1V0.001L21.69,0.038c0.096,0.04,0.203,0.101,0.309,0.158V1.38C21.782,1.221,21.573,1.085,21.393,1H0z"]
  },
  "pointybutton_left_ne": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    paths: ["M2.819,3H1.648C1.232,2.432,0.58,1.803,0,1.38V0.196C0.982,0.733,2.296,1.981,2.819,3z"]
  },
  "pointybutton_left_e": {
    type: "SVG_PATHS",
    width: 3,
    height: 21,
    paths: ["M2.972,0.338L2.999,0.5L2.961,20.691C2.921,20.788,2.864,20.894,2.806,21H1.619c0.159-0.218,0.294-0.426,0.38-0.606V0.591C1.926,0.411,1.802,0.21,1.648,0h1.171C2.878,0.115,2.935,0.229,2.972,0.338z"]
  },
  "pointybutton_left_se": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    paths: ["M2.806,0C2.268,0.98,1.019,2.293,0,2.816V1.646C0.569,1.229,1.195,0.578,1.619,0H2.806z"]
  },
  "pointybutton_left_s": {
    type: "SVG_PATHS",
    width: 22,
    height: 3,
    paths: ["M0,2.982V2h21.409C21.587,1.927,21.79,1.799,22,1.646v1.171c-0.115,0.06-0.23,0.119-0.337,0.156L21.583,3L0,2.982z"]
  },
  "pointybutton_left_sw": {
    type: "SVG_PATHS",
    width: 17,
    height: 3,
    paths: ["M16.569,2H17v0.982l-0.634-0.001C14.847,2.559,12.686,1.444,10.431,0h1.879C13.884,0.934,15.371,1.655,16.569,2z"]
  },
  "pointybutton_left_w": {
    type: "SVG_PATHS",
    width: 17,
    height: 21,
    paths: ["M12.31,21h-1.879C5.446,17.809,0,13.008,0,10.5C0,8.046,5.499,3.207,10.476,0h1.818C7.12,3.065,1,8.45,1,10.5C1,12.539,7.133,17.93,12.31,21z"]
  },
  "pointybutton_left_c": {
    type: "SVG_PATHS",
    width: 22,
    height: 21,
    paths: []
  },
  "pointybutton_left_background": {
    type: "SVG_SCALE9",
    scale9: [3, 16, 3, 3],
    width: 42,
    height: 27
  },
  "pointybutton_left_background_nw": {
    type: "SVG_PATHS",
    width: 16,
    height: 3,
    paths: ["M16,3h-4.653C13.053,1.941,14.684,1.1,16,0.654V3z"]
  },
  "pointybutton_left_background_n": {
    type: "SVG_PATHS",
    width: 23,
    height: 3,
    paths: ["M23,3H0V0.654c0.172-0.059,0.339-0.11,0.5-0.154h22c0.152,0.062,0.323,0.162,0.5,0.274V3z"]
  },
  "pointybutton_left_background_ne": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    paths: ["M2.245,3H0V0.774C0.792,1.28,1.757,2.213,2.245,3z"]
  },
  "pointybutton_left_background_e": {
    type: "SVG_PATHS",
    width: 3,
    height: 21,
    paths: ["M2.245,0c0.108,0.175,0.201,0.346,0.254,0.5v20c-0.063,0.151-0.162,0.323-0.275,0.5H0V0H2.245z"]
  },
  "pointybutton_left_background_se": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    paths: ["M2.224,0C1.72,0.79,0.789,1.756,0,2.245V0H2.224z"]
  },
  "pointybutton_left_background_s": {
    type: "SVG_PATHS",
    width: 23,
    height: 3,
    paths: ["M23,0v2.245c-0.176,0.108-0.346,0.202-0.5,0.255h-22C0.339,2.455,0.172,2.403,0,2.345V0H23z"]
  },
  "pointybutton_left_background_sw": {
    type: "SVG_PATHS",
    width: 16,
    height: 3,
    paths: ["M16,2.345C14.685,1.896,13.056,1.056,11.354,0H16V2.345z"]
  },
  "pointybutton_left_background_w": {
    type: "SVG_PATHS",
    width: 16,
    height: 21,
    paths: ["M16,0v21h-4.646C6.258,17.84,0.5,12.753,0.5,10.5C0.5,8.249,6.255,3.159,11.346,0H16z"]
  },
  "pointybutton_left_background_c": {
    type: "SVG_PATHS",
    width: 23,
    height: 21,
    paths: ["M23,0v21H0V0H23z"]
  },
  "pointybutton_right": {
    type: "SVG_SCALE9",
    scale9: [3, 3, 17, 3],
    width: 42,
    height: 27
  },
  "pointybutton_right_nw": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    paths: ["M1.35,3H0.18C0.703,1.98,2.017,0.732,3,0.195v1.184C2.419,1.802,1.766,2.432,1.35,3z"]
  },
  "pointybutton_right_n": {
    type: "SVG_PATHS",
    width: 22,
    height: 3,
    paths: ["M0,1.379V0.195c0.105-0.058,0.213-0.117,0.309-0.157L0.5,0L22,0.018V1H0.606C0.426,1.085,0.217,1.221,0,1.379z"]
  },
  "pointybutton_right_ne": {
    type: "SVG_PATHS",
    width: 17,
    height: 3,
    paths: ["M6.524,3H4.707c-1.58-0.937-3.073-1.658-4.275-2H0V0.018h0.633C2.114,0.426,4.266,1.545,6.524,3z"]
  },
  "pointybutton_right_e": {
    type: "SVG_PATHS",
    width: 17,
    height: 21,
    paths: ["M16.999,10.5c0,2.509-5.447,7.31-10.432,10.5H4.688c5.177-3.07,11.311-8.461,11.311-10.5c0-2.051-6.12-7.436-11.292-10.5h1.817C11.501,3.207,16.999,8.045,16.999,10.5z"]
  },
  "pointybutton_right_se": {
    type: "SVG_PATHS",
    width: 17,
    height: 3,
    paths: ["M6.567,0C4.312,1.443,2.151,2.558,0.633,2.98L0.567,2.999L0,2.998V1.999h0.43C1.627,1.654,3.114,0.934,4.688,0H6.567z"]
  },
  "pointybutton_right_s": {
    type: "SVG_PATHS",
    width: 22,
    height: 3,
    paths: ["M0,2.816V1.646c0.209,0.152,0.412,0.281,0.59,0.354H22v0.999L0.337,2.972C0.23,2.935,0.115,2.875,0,2.816z"]
  },
  "pointybutton_right_sw": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    paths: ["M3,1.646v1.171C1.98,2.293,0.732,0.981,0.194,0h1.187C1.804,0.578,2.431,1.229,3,1.646z"]
  },
  "pointybutton_right_w": {
    type: "SVG_PATHS",
    width: 3,
    height: 21,
    paths: ["M1,20.393c0.086,0.181,0.221,0.39,0.381,0.607H0.194c-0.058-0.106-0.116-0.213-0.156-0.31L0,20.499L0.027,0.338C0.064,0.229,0.121,0.115,0.18,0h1.17C1.196,0.209,1.073,0.411,1,0.59V20.393z"]
  },
  "pointybutton_right_c": {
    type: "SVG_PATHS",
    width: 22,
    height: 21,
    paths: []
  },
  "pointybutton_right_background": {
    type: "SVG_SCALE9",
    scale9: [3, 3, 17, 3],
    width: 42,
    height: 27
  },
  "pointybutton_right_background_nw": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    paths: ["M3,3H0.754C1.242,2.212,2.207,1.278,3,0.774V3z"]
  },
  "pointybutton_right_background_n": {
    type: "SVG_PATHS",
    width: 22,
    height: 3,
    paths: ["M22,3H0V0.774C0.177,0.661,0.349,0.562,0.5,0.5H22V3z"]
  },
  "pointybutton_right_background_ne": {
    type: "SVG_PATHS",
    width: 17,
    height: 3,
    paths: ["M5.652,3H0V0.5h0.5C1.889,0.882,3.723,1.803,5.652,3z"]
  },
  "pointybutton_right_background_e": {
    type: "SVG_PATHS",
    width: 17,
    height: 21,
    paths: ["M16.499,10.5c0,2.252-5.761,7.34-10.858,10.5H0V0h5.652C10.743,3.158,16.499,8.248,16.499,10.5z"]
  },
  "pointybutton_right_background_se": {
    type: "SVG_PATHS",
    width: 17,
    height: 3,
    paths: ["M5.641,0C3.717,1.192,1.888,2.112,0.499,2.499H0V0H5.641z"]
  },
  "pointybutton_right_background_s": {
    type: "SVG_PATHS",
    width: 22,
    height: 3,
    paths: ["M22,0v2.499H0.5C0.346,2.446,0.175,2.354,0,2.245V0H22z"]
  },
  "pointybutton_right_background_sw": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    paths: ["M3,0v2.245C2.211,1.757,1.28,0.791,0.775,0H3z"]
  },
  "pointybutton_right_background_w": {
    type: "SVG_PATHS",
    width: 3,
    height: 21,
    paths: ["M3,0v21H0.775c-0.113-0.178-0.212-0.349-0.275-0.501V0.5C0.553,0.346,0.646,0.175,0.754,0H3z"]
  },
  "pointybutton_right_background_c": {
    type: "SVG_PATHS",
    width: 22,
    height: 21,
    paths: ["M22,0v21H0V0H22z"]
  },
  "pointybutton_default_icon_color": 0x333333,
  "popover_radius": 16,
  "popover_radius_margin": 3,
  "popover_thickness": 1,
  "popover_cover1_padding": 2,
  "popover_cover2_offset": 1,
  "popover_cover2_padding": 1,
  "popover_text_padding": 5,
  "progressbar_color": 0xCCCCCC,
  "progressbar_indeterminate": {
    type: "SVG_PATHS",
    width: 25,
    height: 11,
    paths: ["M12.549,0H25v0.035L14.036,11H1.551L12.549,0z"]
  },
  "progressbar_padding": 3,
  "rectangle_border_square": {
    type: "SVG_SCALE9",
    scale9: [1, 1, 1, 1],
    width: 42,
    height: 21
  },
  "rectangle_border_square_nw": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,1H0V0h1V1z"]
  },
  "rectangle_border_square_n": {
    type: "SVG_PATHS",
    width: 40,
    height: 1,
    paths: ["M40,0v1H0h0V0H40z"]
  },
  "rectangle_border_square_ne": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,1H0h0V0h1V1z"]
  },
  "rectangle_border_square_e": {
    type: "SVG_PATHS",
    width: 1,
    height: 19,
    paths: ["M1,0v19H0V0H1z"]
  },
  "rectangle_border_square_se": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,0v1H0V0h0H1z"]
  },
  "rectangle_border_square_s": {
    type: "SVG_PATHS",
    width: 40,
    height: 1,
    paths: ["M0,0h40v1H0L0,0L0,0z"]
  },
  "rectangle_border_square_sw": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,0v1H0V0H1z"]
  },
  "rectangle_border_square_w": {
    type: "SVG_PATHS",
    width: 1,
    height: 19,
    paths: ["M1,0v19H0V0H1z"]
  },
  "rectangle_border_square_c": {
    type: "SVG_PATHS",
    width: 40,
    height: 19,
    paths: ["M0,0L0,0l0,19h0V0z"]
  },
  "rectangle_border_rounded": {
    type: "SVG_SCALE9",
    scale9: [7, 7, 7, 7],
    width: 68,
    height: 41
  },
  "rectangle_border_rounded_nw": {
    type: "SVG_PATHS",
    width: 7,
    height: 7,
    paths: ["M1,7H0V5.5C0,2.468,2.467,0,5.5,0h1.559v1H5.5C3.019,1,1,3.019,1,5.5V7z"]
  },
  "rectangle_border_rounded_n": {
    type: "SVG_PATHS",
    width: 54,
    height: 7,
    paths: ["M0.059,1V0H54v1H0.059z"]
  },
  "rectangle_border_rounded_ne": {
    type: "SVG_PATHS",
    width: 7,
    height: 7,
    paths: ["M7,7H6V5.5C6,3.019,3.981,1,1.5,1H0V0h1.5C4.533,0,7,2.468,7,5.5V7z"]
  },
  "rectangle_border_rounded_e": {
    type: "SVG_PATHS",
    width: 7,
    height: 27,
    paths: ["M7,0v27H6V0H7z"]
  },
  "rectangle_border_rounded_se": {
    type: "SVG_PATHS",
    width: 7,
    height: 7,
    paths: ["M7,0v1.5C7,4.532,4.533,7,1.5,7H0V6h1.5C3.981,6,6,3.981,6,1.5V0H7z"]
  },
  "rectangle_border_rounded_s": {
    type: "SVG_PATHS",
    width: 54,
    height: 7,
    paths: ["M0.059,7V6H54v1H0.059z"]
  },
  "rectangle_border_rounded_sw": {
    type: "SVG_PATHS",
    width: 7,
    height: 7,
    paths: ["M5.5,6h1.559v1H5.5C2.467,7,0,4.532,0,1.5V0h1v1.5C1,3.981,3.019,6,5.5,6z"]
  },
  "rectangle_border_rounded_w": {
    type: "SVG_PATHS",
    width: 7,
    height: 27,
    paths: ["M1,0v27H0V0H1z"]
  },
  "rectangle_border_rounded_c": {
    type: "SVG_PATHS",
    width: 54,
    height: 27,
    paths: []
  },
  "rectangle_border_rounded_dotted": {
    width: 1,
    corner_radius: 8,
    dash_array: [8, 8]
  },
  "rectangle_corner_radius": {
    roundedSolid: 4,
    roundedDotted: 8,
    square: 0
  },
  "rectangle_background_padding": {
    default: 1,
    none: 0
  },
  "redx": {
    type: "SVG_PATHS",
    width: 240,
    height: 104,
    paths: ["M238.363,97.116c-0.428,3.152-2.683,5.189-5.745,5.189c-1.924,0.001-4.118-0.75-6.522-2.229c-11.821-7.275-62.758-26.416-63.271-26.608l-42.671-18.097c-4.853,1.782-99.574,36.628-106.109,43.516c-1.699,1.791-3.631,2.737-5.586,2.737c-4.028,0-7.115-3.915-7.366-7.43c-0.263-3.683,2.248-6.264,6.716-6.906c6.154-0.885,45.34-17.834,45.736-18.005l51.522-18.467C94.28,46.652,17.693,17.138,9.393,15.322c-4.82-1.055-6.229-4.283-5.94-6.953c0.318-2.941,2.864-5.984,6.574-5.984c1.448,0,2.929,0.455,4.4,1.355c13.34,8.152,100.028,39.641,104.411,41.229c4.516-1.55,95.795-32.926,109.232-41.661c1.268-0.824,2.553-1.242,3.821-1.242c3.658,0,6.087,3.398,6.239,6.548c0.138,2.867-1.56,6.11-6.74,6.61c-9.449,0.915-87.066,31.476-97.545,35.614c9.852,3.982,79.343,31.987,98.646,37.33C236.928,89.396,238.828,93.695,238.363,97.116z"]
  },
  "redx_color": 0xcf2b28,
  "rule_default_color": 0x333333,
  "rule_thickness": 2,
  "rule_breakline_thickness": 1,
  "rule_dash_size": 2,
  "rule_dot_size": 1,
  "rule_horizontal": {
    type: "SVG_PATHS",
    width: 10,
    height: 10,
    paths: ["M10,6H0V5h10V6z"]
  },
  "rule_vertical": {
    type: "SVG_PATHS",
    width: 10,
    height: 10,
    paths: ["M6,10.0H5V0h1V10.0z"]
  },
  "searchbox": {
    type: "SVG_SCALE9",
    scale9: [12, 13, 13, 12],
    width: 51,
    height: 25
  },
  "searchbox_nw": {
    type: "SVG_PATHS",
    width: 13,
    height: 12,
    paths: ["M1.025,12h-1C0.291,5.34,5.776,0,12.5,0h0.496v1H12.5C6.328,1,1.29,5.892,1.025,12z"]
  },
  "searchbox_n": {
    type: "SVG_PATHS",
    width: 25,
    height: 12,
    paths: ["M-0.004,1V0h25v1H-0.004z"]
  },
  "searchbox_ne": {
    type: "SVG_PATHS",
    width: 13,
    height: 12,
    paths: ["M12.974,12h-1C11.709,5.892,6.671,1,0.499,1h-0.503V0h0.503C7.223,0,12.708,5.34,12.974,12z"]
  },
  "searchbox_e": {
    type: "SVG_PATHS",
    width: 13,
    height: 1,
    paths: ["M12.974,0c0.007,0.167,0.025,0.331,0.025,0.5S12.98,0.833,12.974,1h-1c0.007-0.167,0.025-0.331,0.025-0.5S11.981,0.167,11.974,0H12.974z"]
  },
  "searchbox_se": {
    type: "SVG_PATHS",
    width: 13,
    height: 12,
    paths: ["M12.974,0C12.708,6.66,7.223,12,0.499,12h-0.503v-1h0.503c6.172,0,11.21-4.892,11.475-11H12.974z"]
  },
  "searchbox_s": {
    type: "SVG_PATHS",
    width: 25,
    height: 12,
    paths: ["M-0.004,12v-1h25v1H-0.004z"]
  },
  "searchbox_sw": {
    type: "SVG_PATHS",
    width: 13,
    height: 12,
    paths: ["M12.5,11h0.496v1H12.5C5.776,12,0.291,6.66,0.025,0h1C1.29,6.108,6.328,11,12.5,11z"]
  },
  "searchbox_w": {
    type: "SVG_PATHS",
    width: 13,
    height: 1,
    paths: ["M1,0.5C1,0.669,1.018,0.833,1.025,1h-1C0.019,0.833,0,0.669,0,0.5S0.019,0.167,0.025,0h1C1.018,0.167,1,0.331,1,0.5z"]
  },
  "searchbox_c": {
    type: "SVG_PATHS",
    width: 25,
    height: 1,
    paths: []
  },
  "searchbox_label_color": 0x333333,
  "searchbox_padding_left": 22,
  "searchbox_padding_right": 4,
  "searchbox_padding_top": 7,
  "searchbox_padding_bottom": 6,
  "scratchout": {
    type: "SVG_PATHS",
    width: 205,
    height: 107,
    paths: ["M183.938,103.357c-0.301-1.293-1.361-1.826-0.956-3.824c-3.021-2.322-3.755,4.031-8.605,2.868c-3.061-0.127-4.206-2.168-3.824-5.737c-3.772,1.729-6.709,6.935-11.474,4.781c-2.123-11.135,8.937-17.133,13.387-24.859c-12.04,6.128-14.488,21.845-31.553,22.947c0.637-14.024,11.619-17.702,18.166-25.815c-11.682,4.293-16.145,20.771-27.728,21.99c-2.158-7.69,5.622-14.984,10.517-21.034c5.463-6.753,11.587-13.253,16.255-19.122c-15.683,8.222-40.248,40.252-45.895,39.201c-6.314-10.054,4.456-20.203,10.518-26.771c7.21-7.812,16.645-14.776,21.991-21.991C120.975,56.969,87.271,99.278,79.72,96.664c-0.215-21.002,19.665-29.999,29.64-43.981c-9.693,4.716-19.203,12.874-28.684,21.035c-9.371,8.066-21.857,25.229-27.729,22.946c-6.371-8.138,16.167-26.514,23.904-39.201c-17.33,10.717-37.883,40.988-42.35,38.342c-8.602-2.978-2.875-14.846,1.236-20.176c4.974-6.447,11.133-13.549,15.297-18.166c-12.631,8.088-31.09,42.974-35.377,38.245C3.402,82.239,29.556,62.055,36.693,47.902c-5.64,1.699-13.936,18.787-16.254,17.21c-6.822-4.848-0.16-13.57,2.869-19.122c3.852-7.063,8.428-13.826,12.43-18.167c-8.547,3.251-22.905,24.31-25.816,21.034c-7.842-5.822,9.826-19.441,14.342-28.684c-4.297,0.803-5.265,4.934-9.561,5.736c-5.009-0.408-2.414-8.422-8.605-7.648C0.261,15.574,0.324,7.191,3.404,5.814c2.789-1.035,4.38-4.032,8.43-3.807c1.779,1.424,5.25,4.031,2.869,6.693c3.428,0.242,3.502-2.871,6.692-2.869c2.689,0.18,4.073,1.664,3.825,4.781c5.727-1.142,9.983-8.263,16.254-5.736c3.433,6.877-4.328,10.143-6.693,14.341c6.717-2.844,10.459-8.663,19.123-9.562c8.389,2.279-9.452,14.809-13.387,21.992c9.719-2.752,24.092-26.33,30.598-20.079c2.923,4.337-12.131,18.266-19.123,26.771C44.593,47.34,38.44,57.074,30.957,65.112C44.203,54.984,60.55,36.35,75.895,24.955c4.805-3.568,12.36-10.76,18.167-5.737c3.796,4.628-31.744,31.36-44.938,49.72c18.27-11.348,38.271-30.186,57.368-43.027c6.133-4.123,19.158-13.312,21.991-8.604c3.453,4.224-7.134,9.329-11.474,12.43c-14.82,10.591-27.06,22.726-38.245,36.333c23.612-13.678,63.758-51.32,65.973-45.895c2.134,8.822-7.204,16.146-13.386,21.991c-7.073,6.687-15.973,12.132-21.035,19.123c20.496-12.651,57.352-44.06,60.236-39.201c2.407,5.067-7.816,15.064-13.386,21.034c-5.897,6.323-13.34,11.612-18.166,18.167c16.445-10.646,43.885-40.751,46.035-34.961c5.129,7.443-13.82,20.325-20.22,31.136c12.378-5.789,26.011-27.828,32.509-21.99c1.417,5.412-14.039,18.47-21.035,27.728c4.49-0.304,8.569-9.266,17.21-8.605c5.242,6.307-11.161,17.467-16.254,25.816c5.629-0.844,12.811-17.216,21.991-11.474c0,0.955,0,1.911,0,2.868c-2.645,5.005-7.286,8.012-10.519,12.43c3.869-0.352,6.095-6.107,10.519-2.868c1.551,1.705,0,3.187,0,4.78c-3.196,6.365-9.475,9.647-11.475,17.21C186.605,104.908,184.87,105.729,183.938,103.357z"]
  },
  "scratchout_color": 0x333333,
  "scrollbar_background_color": 0xFFFFFF,
  "scrollbar_background_radius": 0,
  "shape_default_color": 0xFFFFFF,
  "shape_default_border_color": 0x333333,
  "shape_min_thickness": 1,
  "shape_max_thickness": 1,
  "shape_padding_left": 8,
  "shape_padding_right": 8,
  "shape_padding_top": 4,
  "shape_padding_bottom": 4,
  "shape_icon_padding": 3,
  "sitemap_edge_width": 1,
  "slider_disabled_color": 0xcccccc,
  "slider_default_color": 0x666666,
  "slider_vertical_thumb_circle_padding": 0.5,
  "slider_vertical_thumb_pointy_padding": 1,
  "slider_vertical_thumb_rectangle_padding": 1,
  "slider_horizontal_thumb_circle_padding": 0.5,
  "slider_horizontal_thumb_pointy_padding": 1,
  "slider_horizontal_thumb_rectangle_padding": 1,
  "slider_horizontal_track": {
    type: "SVG_SCALE9",
    scale9: [2, 3, 3, 8],
    width: 33,
    height: 11
  },
  "slider_horizontal_track_nw": {
    type: "SVG_PATHS",
    width: 3,
    height: 2,
    paths: []
  },
  "slider_horizontal_track_n": {
    type: "SVG_PATHS",
    width: 27,
    height: 2,
    paths: []
  },
  "slider_horizontal_track_ne": {
    type: "SVG_PATHS",
    width: 3,
    height: 2,
    paths: []
  },
  "slider_horizontal_track_e": {
    type: "SVG_PATHS",
    width: 3,
    height: 1,
    paths: []
  },
  "slider_horizontal_track_se": {
    type: "SVG_PATHS",
    width: 3,
    height: 8,
    paths: ["M0,0.961h3v1H0V0.961z"]
  },
  "slider_horizontal_track_s": {
    type: "SVG_PATHS",
    width: 27,
    height: 8,
    paths: ["M0,0.961h27v1H0V0.961z"]
  },
  "slider_horizontal_track_sw": {
    type: "SVG_PATHS",
    width: 3,
    height: 8,
    paths: ["M0,0.961h3v1H0V0.961z"]
  },
  "slider_horizontal_track_w": {
    type: "SVG_PATHS",
    width: 3,
    height: 1,
    paths: []
  },
  "slider_horizontal_track_c": {
    type: "SVG_PATHS",
    width: 27,
    height: 1,
    paths: []
  },
  "slider_horizontal_thumb": {
    type: "SVG_PATHS",
    width: 10,
    height: 10,
    paths: ["M10,5c0,2.764-2.239,5.002-5,5.002c-2.762,0-5-2.238-5-5.002c0-2.761,2.239-5,5-5C7.761,0,10,2.239,10,5z"]
  },
  "slider_vertical_track": {
    type: "SVG_SCALE9",
    scale9: [4, 6, 3, 4],
    width: 11,
    height: 33
  },
  "slider_vertical_track_nw": {
    type: "SVG_PATHS",
    width: 6,
    height: 4,
    paths: ["M5,4H4V0h1V4z"]
  },
  "slider_vertical_track_n": {
    type: "SVG_PATHS",
    width: 2,
    height: 4,
    paths: []
  },
  "slider_vertical_track_ne": {
    type: "SVG_PATHS",
    width: 3,
    height: 4,
    paths: []
  },
  "slider_vertical_track_e": {
    type: "SVG_PATHS",
    width: 3,
    height: 25,
    paths: []
  },
  "slider_vertical_track_se": {
    type: "SVG_PATHS",
    width: 3,
    height: 4,
    paths: []
  },
  "slider_vertical_track_s": {
    type: "SVG_PATHS",
    width: 2,
    height: 4,
    paths: []
  },
  "slider_vertical_track_sw": {
    type: "SVG_PATHS",
    width: 6,
    height: 4,
    paths: ["M5,0v4H4V0H5z"]
  },
  "slider_vertical_track_w": {
    type: "SVG_PATHS",
    width: 6,
    height: 25,
    paths: ["M5,0v25H4V0H5z"]
  },
  "slider_vertical_track_c": {
    type: "SVG_PATHS",
    width: 2,
    height: 25,
    paths: []
  },
  "slider_vertical_thumb_pointy": {
    type: "SVG_PATHS",
    width: 14,
    height: 14,
    paths: ["M10.795,5.29c0.388,0.391,0.387,1.029-0.001,1.418l-3.575,3.584C6.831,10.681,6.063,11,5.513,11H1.5c-0.55,0-1-0.45-1-1V2c0-0.55,0.45-1,1-1h4.038c0.55,0,1.317,0.319,1.705,0.71L10.795,5.29z"]
  },
  "slider_vertical_thumb_circle": {
    type: "SVG_PATHS",
    width: 12,
    height: 12,
    paths: ["M6,0.5c3.037,0,5.5,2.463,5.5,5.5S9.037,11.5,6,11.5c-3.038,0-5.5-2.463-5.5-5.5S2.962,0.5,6,0.5z"]
  },
  "slider_vertical_thumb_rectangle": {
    type: "SVG_PATHS",
    width: 14,
    height: 14,
    paths: ["M10.5,3.5c0.55,0,1,0.45,1,1v3c0,0.55-0.45,1-1,1h-9c-0.55,0-1-0.45-1-1v-3c0-0.55,0.45-1,1-1H10.5z"]
  },
  "splitter_horizontal_track": {
    type: "SVG_SCALE9",
    scale9: [4, 5, 6, 4],
    width: 50,
    height: 12
  },
  "splitter_horizontal_track_nw": {
    type: "SVG_PATHS",
    width: 5,
    height: 4,
    paths: ["M5,1v1H0V1H5z"]
  },
  "splitter_horizontal_track_n": {
    type: "SVG_PATHS",
    width: 39,
    height: 4,
    paths: ["M39,2H0V1h39V2z"]
  },
  "splitter_horizontal_track_ne": {
    type: "SVG_PATHS",
    width: 6,
    height: 4,
    paths: ["M6,2H0V1h6V2z"]
  },
  "splitter_horizontal_track_e": {
    type: "SVG_PATHS",
    width: 6,
    height: 4,
    paths: []
  },
  "splitter_horizontal_track_se": {
    type: "SVG_PATHS",
    width: 6,
    height: 4,
    paths: ["M6,3H0V2h6V3z"]
  },
  "splitter_horizontal_track_s": {
    type: "SVG_PATHS",
    width: 39,
    height: 4,
    paths: ["M39,3H0V2h39V3z"]
  },
  "splitter_horizontal_track_sw": {
    type: "SVG_PATHS",
    width: 5,
    height: 4,
    paths: ["M5,2v1H0V2H5z"]
  },
  "splitter_horizontal_track_w": {
    type: "SVG_PATHS",
    width: 5,
    height: 4,
    paths: []
  },
  "splitter_horizontal_track_c": {
    type: "SVG_PATHS",
    width: 39,
    height: 4,
    paths: []
  },
  "splitter_horizontal_thumb": {
    type: "SVG_PATHS",
    width: 16,
    height: 4,
    paths: ["M3,1v2H1V1H3z", "M9,1v2H7V1H9z", "M15,1v2h-2V1H15z"]
  },
  "slider_horizontal_thumb_pointy": {
    type: "SVG_PATHS",
    width: 14,
    height: 14,
    paths: ["M6.71,10.795c-0.391,0.388-1.029,0.387-1.418-0.001L1.708,7.219C1.319,6.831,1,6.063,1,5.513V1.5c0-0.55,0.45-1,1-1h8c0.55,0,1,0.45,1,1v4.038c0,0.55-0.319,1.317-0.71,1.705L6.71,10.795z"]
  },
  "slider_horizontal_thumb_circle": {
    type: "SVG_PATHS",
    width: 12,
    height: 12,
    paths: ["M6,0.5c3.037,0,5.5,2.463,5.5,5.5S9.037,11.5,6,11.5c-3.038,0-5.5-2.463-5.5-5.5S2.962,0.5,6,0.5z"]
  },
  "slider_horizontal_thumb_rectangle": {
    type: "SVG_PATHS",
    width: 14,
    height: 14,
    paths: ["M8.5,10.5c0,0.55-0.45,1-1,1h-3c-0.55,0-1-0.45-1-1v-9c0-0.55,0.45-1,1-1h3c0.55,0,1,0.45,1,1V10.5z"]
  },
  "splitter_vertical_track": {
    type: "SVG_SCALE9",
    scale9: [3, 4, 4, 3],
    width: 12,
    height: 50
  },
  "splitter_vertical_track_nw": {
    type: "SVG_PATHS",
    width: 4,
    height: 3,
    paths: ["M2,3H1V0h1V3z"]
  },
  "splitter_vertical_track_n": {
    type: "SVG_PATHS",
    width: 4,
    height: 3,
    paths: []
  },
  "splitter_vertical_track_ne": {
    type: "SVG_PATHS",
    width: 4,
    height: 3,
    paths: ["M3,3H2V0h1V3z"]
  },
  "splitter_vertical_track_e": {
    type: "SVG_PATHS",
    width: 4,
    height: 44,
    paths: ["M3,0v44H2V0H3z"]
  },
  "splitter_vertical_track_se": {
    type: "SVG_PATHS",
    width: 4,
    height: 3,
    paths: ["M3,0v3H2V0H3z"]
  },
  "splitter_vertical_track_s": {
    type: "SVG_PATHS",
    width: 4,
    height: 3,
    paths: []
  },
  "splitter_vertical_track_sw": {
    type: "SVG_PATHS",
    width: 4,
    height: 3,
    paths: ["M2,0v3H1V0H2z"]
  },
  "splitter_vertical_track_w": {
    type: "SVG_PATHS",
    width: 4,
    height: 44,
    paths: ["M2,0v44H1V0H2z"]
  },
  "splitter_vertical_track_c": {
    type: "SVG_PATHS",
    width: 4,
    height: 44,
    paths: []
  },
  "splitter_vertical_thumb": {
    type: "SVG_PATHS",
    width: 4,
    height: 16,
    paths: ["M3,3H1V1h2V3z", "M3,9H1V7h2V9z", "M3,15H1v-2h2V15z"]
  },
  "streetmap": {
    type: "SVG_PATHS",
    width: 252,
    height: 222,
    colors: [0xebe9e1, 0xebe9e1, 0xebe9e1, 0xebe9e1, 0xebe9e1, 0xebe9e1, 0xebe9e1, 0xebe9e1, 0xebe9e1, 0xbdd5ac, 0xebe9e1, 0xebe9e1, 0xebe9e1, 0xebe9e1, 0xebe9e1, 0xfaf077, 0xebe9e1, 0xebe9e1],
    paths: ["M0,0h57.206L66.6,61.287L0,73.029V0z", "M135.292,0l7.633,48.957L73.058,60.406L63.077,0H135.292z", "M0,79.193l28.73-5.283l7.632,46.675c0,0-27.3,6.165-29.355,7.633S0,79.193,0,79.193z", "M35.189,72.441l32.292-5.283l9.394,50.785l-35.227,2.642L35.189,72.441z", "M74.819,66.57l29.943-4.99l11.155,54.602l-32.291,1.468L74.819,66.57z", "M143.805,55.122l7.927,51.372l-30.237,7.633l-9.687-53.721L143.805,55.122z", "M149.97,53.947L252,36.041v22.604l-99.095,18.494L149.97,53.947z", "M155.547,83.01L252,65.396v72.215l-9.854,2.643l-7.633-49.904l-76.911,12.328L155.547,83.01z", "M159.363,108.256l69.867-10.274l7.926,42.272l-72.803,5.283L159.363,108.256z", "M9.65,157.867L7.008,127.54l29.355-7.808l79.554-4.504l35.814-9.908l7.339,39.935L9.65,157.867z", "M0,164.031l66.893-4.403L71.59,222H0V164.031z", "M73.058,158.454l46.675-2.936L124.43,222H77.168L73.058,158.454z", "M124.724,154.638l44.914-2.936L174.041,222h-44.033L124.724,154.638z", "M176.39,150.821l46.089-2.936L225.707,222h-45.795L176.39,150.821z", "M227.762,146.418L252,144.363V222l-20.128-1.019L227.762,146.418z", "M0.285,165.465l-0.569-6.977l237.111-19.353L217.568,34.915l6.884-1.271l20.667,111.838L0.285,165.465z", "M0,158.748v-34.934l3.779,34.053L0,158.748z", "M252,0v31.344L148.208,48.371L141.163,0H252z"]
  },
  "stickynote_padding_top": 13,
  "stickynote_padding_bottom": 10,
  "stickynote_padding_left": 10,
  "stickynote_padding_right": 10,
  "stickynote_brightness_offset": 0,
  "stickynote_paper": {
    type: "SVG_SCALE9",
    scale9: [8, 5, 6, 8],
    width: 105,
    height: 106
  },
  "stickynote_paper_nw": {
    type: "SVG_PATHS",
    width: 5,
    height: 8,
    colors: [0x000000],
    paths: ["M5,8H3.679C3.686,4.449,3.632,1.972,3.5,1L5,1.041V8z"]
  },
  "stickynote_paper_n": {
    type: "SVG_PATHS",
    width: 94,
    height: 8,
    colors: [0x000000],
    paths: ["M94,8H0V1.041l94,2.545V8z"]
  },
  "stickynote_paper_ne": {
    type: "SVG_PATHS",
    width: 6,
    height: 8,
    colors: [0x000000],
    paths: ["M2.949,8H0V3.586l3,0.081C3,3.667,2.979,5.237,2.949,8z"]
  },
  "stickynote_paper_e": {
    type: "SVG_PATHS",
    width: 6,
    height: 90,
    colors: [0x000000],
    paths: ["M3.113,90H0V0h2.949C2.789,14.616,2.344,62.75,3.113,90z"]
  },
  "stickynote_paper_se": {
    type: "SVG_PATHS",
    width: 6,
    height: 8,
    colors: [0x000000],
    paths: ["M3.113,0C3.13,0.621,3.148,1.235,3.167,1.834C2.211,2.117,1.151,2.376,0,2.615V0H3.113z"]
  },
  "stickynote_paper_s": {
    type: "SVG_PATHS",
    width: 94,
    height: 8,
    colors: [0x000000],
    paths: ["M94,0v2.615C72.226,7.129,17.322,4.01,0,3.048V0H94z"]
  },
  "stickynote_paper_sw": {
    type: "SVG_PATHS",
    width: 5,
    height: 8,
    colors: [0x000000],
    paths: ["M5,0v3.048C3.27,2.952,1.91,2.877,1,2.834c0,0,0.039-1.02,0.106-2.834H5z"]
  },
  "stickynote_paper_w": {
    type: "SVG_PATHS",
    width: 5,
    height: 90,
    colors: [0x000000],
    paths: ["M5,0v90H1.106C1.597,76.741,3.639,20.439,3.679,0H5z"]
  },
  "stickynote_paper_c": {
    type: "SVG_PATHS",
    width: 94,
    height: 90,
    colors: [0x000000],
    paths: ["M94,0v90H0V0H94z"]
  },
  "switch_border": {
    type: "SVG_PATHS",
    width: 49,
    height: 28,
    colors: [null, 0xffffff, null],
    paths: ["M34,28H14C6.28,28,0,21.72,0,14S6.28,0,14,0h20V28zM14,2C7.383,2,2,7.383,2,14s5.383,12,12,12h18V2H14z", "M48,14c0,7.18-5.82,13-13,13s-13-5.82-13-13S27.82,1,35,1S48,6.82,48,14z", "M35,28c-7.72,0-14-6.28-14-14S27.28,0,35,0s14,6.28,14,14S42.72,28,35,28zM35,2c-6.617,0-12,5.383-12,12s5.383,12,12,12s12-5.383,12-12S41.617,2,35,2z"]
  },
  "switch_background": {
    type: "SVG_PATHS",
    width: 49,
    height: 28,
    paths: ["M49,14c0,7.732-6.268,14-14,14H14C6.268,28,0,21.732,0,14l0,0C0,6.268,6.268,0,14,0h21C42.732,0,49,6.268,49,14L49,14z"]
  },
  "switch_default_color": 0x6aa84f,
  "switch_border_disabled_color": 0x999999,
  "switch_background_disabled_color": 0xFFFFFF,
  "tabbar_vertical_separator_skin": "pixel",
  "tabbar_h_default_tab_color": 0xdddddd,
  "tabbar_h_default_color": 0xFFFFFF,
  "tabbar_h_tab_padding_left": 6,
  "tabbar_h_tab_padding_right": 6,
  "tabbar_h_tab_padding_top": 7,
  "tabbar_h_tab_padding_bottom": 7,
  "tabbar_h_tab_overlap": 1,
  "tabbar_h_padding_left": 20,
  "tabbar_h_padding_right": 24,
  "tabbar_h_padding_top": 7,
  "tabbar_h_padding_bottom": 0,
  "tabbar_h_background_overlap": 1,
  "tabbar_h_background_padding": 1,
  "tabbar_h_top_selected_label_offset": 0,
  "tabbar_h_bottom_selected_label_offset": 0,
  "tabbar_h_scrollbar_padding_top": 0,
  "tabbar_h_scrollbar_padding_bottom": 0,
  "tabbar_h_scrollbar_padding_right": 0,
  "tabbar_h_top_selected_tab_mask_padding_top": 0,
  "tabbar_h_top_selected_tab_mask_padding_bottom": -1,
  "tabbar_h_top_selected_tab_mask_padding_left": 1,
  "tabbar_h_top_selected_tab_mask_padding_right": 1,
  "tabbar_h_bottom_selected_tab_mask_padding_top": -1,
  "tabbar_h_bottom_selected_tab_mask_padding_bottom": 0,
  "tabbar_h_bottom_selected_tab_mask_padding_left": 1,
  "tabbar_h_bottom_selected_tab_mask_padding_right": 1,
  "tabbar_h_top_tab": {
    type: "SVG_SCALE9",
    scale9: [4, 4, 4, 4],
    width: 35,
    height: 31
  },
  "tabbar_h_top_tab_nw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M1,4H0V0h4v1H1V4z"]
  },
  "tabbar_h_top_tab_n": {
    type: "SVG_PATHS",
    width: 27,
    height: 4,
    paths: ["M27,1H0V0h27V1z"]
  },
  "tabbar_h_top_tab_ne": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M3.999,4h-1V1H0V0h4V4z"]
  },
  "tabbar_h_top_tab_e": {
    type: "SVG_PATHS",
    width: 4,
    height: 23,
    paths: ["M3.999,0v23h-1V0H3.999z"]
  },
  "tabbar_h_top_tab_se": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M3.999,0v4h-1V0H3.999z"]
  },
  "tabbar_h_top_tab_s": {
    type: "SVG_PATHS",
    width: 27,
    height: 4,
    paths: []
  },
  "tabbar_h_top_tab_sw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M1,0v4H0V0H1z"]
  },
  "tabbar_h_top_tab_w": {
    type: "SVG_PATHS",
    width: 4,
    height: 23,
    paths: ["M1,0v23H0V0H1z"]
  },
  "tabbar_h_top_tab_c": {
    type: "SVG_PATHS",
    width: 27,
    height: 23,
    paths: []
  },
  "tabbar_h_top_background_tab": {
    type: "SVG_SCALE9",
    scale9: [4, 4, 4, 4],
    width: 35,
    height: 31
  },
  "tabbar_h_top_background_tab_nw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M4,4H0.5V0.5H4V4z"]
  },
  "tabbar_h_top_background_tab_n": {
    type: "SVG_PATHS",
    width: 27,
    height: 4,
    paths: ["M27,4H0V0.5h27V4z"]
  },
  "tabbar_h_top_background_tab_ne": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M3.499,4H0V0.5h3.5V4z"]
  },
  "tabbar_h_top_background_tab_e": {
    type: "SVG_PATHS",
    width: 4,
    height: 23,
    paths: ["M3.499,0v23H0V0H3.499z"]
  },
  "tabbar_h_top_background_tab_se": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M3.499,0v4H0V0H3.499z"]
  },
  "tabbar_h_top_background_tab_s": {
    type: "SVG_PATHS",
    width: 27,
    height: 4,
    paths: ["M27,0v4H0V0H27z"]
  },
  "tabbar_h_top_background_tab_sw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M4,0v4H0.5V0H4z"]
  },
  "tabbar_h_top_background_tab_w": {
    type: "SVG_PATHS",
    width: 4,
    height: 23,
    paths: ["M4,0v23H0.5V0H4z"]
  },
  "tabbar_h_top_background_tab_c": {
    type: "SVG_PATHS",
    width: 27,
    height: 23,
    paths: ["M27,0v23H0V0H27z"]
  },
  "tabbar_h_bottom_tab": {
    type: "SVG_SCALE9",
    scale9: [4, 4, 4, 4],
    width: 35,
    height: 31
  },
  "tabbar_h_bottom_tab_nw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M1,4H0V0h1V4z"]
  },
  "tabbar_h_bottom_tab_n": {
    type: "SVG_PATHS",
    width: 27,
    height: 4,
    paths: []
  },
  "tabbar_h_bottom_tab_ne": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M3.995,4h-1V0h1V4z"]
  },
  "tabbar_h_bottom_tab_e": {
    type: "SVG_PATHS",
    width: 4,
    height: 23,
    paths: ["M3.995,0v23h-1V0H3.995z"]
  },
  "tabbar_h_bottom_tab_se": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M3.995,0v3.996H0v-1h2.996V0H3.995z"]
  },
  "tabbar_h_bottom_tab_s": {
    type: "SVG_PATHS",
    width: 27,
    height: 4,
    paths: ["M27,3.996H0v-1h27V3.996z"]
  },
  "tabbar_h_bottom_tab_sw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M1,2.996h3v1H0V0h1V2.996z"]
  },
  "tabbar_h_bottom_tab_w": {
    type: "SVG_PATHS",
    width: 4,
    height: 23,
    paths: ["M1,0v23H0V0H1z"]
  },
  "tabbar_h_bottom_tab_c": {
    type: "SVG_PATHS",
    width: 27,
    height: 23,
    paths: []
  },
  "tabbar_h_bottom_background_tab": {
    type: "SVG_SCALE9",
    scale9: [4, 4, 4, 4],
    width: 35,
    height: 31
  },
  "tabbar_h_bottom_background_tab_nw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M4,4H0.5V0H4V4z"]
  },
  "tabbar_h_bottom_background_tab_n": {
    type: "SVG_PATHS",
    width: 27,
    height: 4,
    paths: ["M27,4H0V0h27V4z"]
  },
  "tabbar_h_bottom_background_tab_ne": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M3.495,4H0V0h3.496V4z"]
  },
  "tabbar_h_bottom_background_tab_e": {
    type: "SVG_PATHS",
    width: 4,
    height: 23,
    paths: ["M3.495,0v23H0V0H3.495z"]
  },
  "tabbar_h_bottom_background_tab_se": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M3.495,0v3.496H0V0H3.495z"]
  },
  "tabbar_h_bottom_background_tab_s": {
    type: "SVG_PATHS",
    width: 27,
    height: 4,
    paths: ["M27,0v3.496H0V0H27z"]
  },
  "tabbar_h_bottom_background_tab_sw": {
    type: "SVG_PATHS",
    width: 4,
    height: 4,
    paths: ["M4,0v3.496H0.5V0H4z"]
  },
  "tabbar_h_bottom_background_tab_w": {
    type: "SVG_PATHS",
    width: 4,
    height: 23,
    paths: ["M4,0v23H0.5V0H4z"]
  },
  "tabbar_h_bottom_background_tab_c": {
    type: "SVG_PATHS",
    width: 27,
    height: 23,
    paths: ["M27,0v23H0V0H27z"]
  },
  "tabbar_v_default_tab_color": 0xdddddd,
  "tabbar_v_default_color": 0xFFFFFF,
  "tabbar_v_tab_padding_left": 6,
  "tabbar_v_tab_padding_right": 4,
  "tabbar_v_tab_padding_top": 8,
  "tabbar_v_tab_padding_bottom": 8,
  "tabbar_v_tab_overlap": 1,
  "tabbar_v_padding_left": 6,
  "tabbar_v_padding_right": 4,
  "tabbar_v_padding_top": 20,
  "tabbar_v_padding_bottom": 42,
  "tabbar_v_background_overlap": 1,
  "tabbar_v_background_padding": 1,
  "tabbar_v_left_selected_label_offset": 0,
  "tabbar_v_right_selected_label_offset": 0,
  "tabbar_v_scrollbar_padding_top": 0,
  "tabbar_v_scrollbar_padding_bottom": 0,
  "tabbar_v_scrollbar_padding_right": 0,
  "tabbar_v_left_selected_tab_mask_padding_top": 1,
  "tabbar_v_left_selected_tab_mask_padding_bottom": 1,
  "tabbar_v_left_selected_tab_mask_padding_left": 0,
  "tabbar_v_left_selected_tab_mask_padding_right": -1,
  "tabbar_v_right_selected_tab_mask_padding_top": 1,
  "tabbar_v_right_selected_tab_mask_padding_bottom": 1,
  "tabbar_v_right_selected_tab_mask_padding_left": -1,
  "tabbar_v_right_selected_tab_mask_padding_right": 0,
  "tabbar_v_left_tab": {
    type: "SVG_SCALE9",
    scale9: [1, 1, 1, 1],
    width: 41,
    height: 27
  },
  "tabbar_v_left_tab_nw": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,0v1h0H0V0H1z"]
  },
  "tabbar_v_left_tab_n": {
    type: "SVG_PATHS",
    width: 39,
    height: 1,
    paths: ["M0,1V0h39v1H0z"]
  },
  "tabbar_v_left_tab_ne": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M0,1V0h1v1H0z"]
  },
  "tabbar_v_left_tab_e": {
    type: "SVG_PATHS",
    width: 1,
    height: 25,
    paths: []
  },
  "tabbar_v_left_tab_se": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M0,0h1v1H0V0z"]
  },
  "tabbar_v_left_tab_s": {
    type: "SVG_PATHS",
    width: 39,
    height: 1,
    paths: ["M0,0h39v1H0V0z"]
  },
  "tabbar_v_left_tab_sw": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,0L1,0l0,1H0V0H1z"]
  },
  "tabbar_v_left_tab_w": {
    type: "SVG_PATHS",
    width: 1,
    height: 25,
    paths: ["M0,0h1v25H0V0z"]
  },
  "tabbar_v_left_tab_c": {
    type: "SVG_PATHS",
    width: 39,
    height: 25,
    paths: []
  },
  "tabbar_v_left_background_tab": {
    type: "SVG_SCALE9",
    scale9: [1, 1, 1, 1],
    width: 41,
    height: 27
  },
  "tabbar_v_left_background_tab_nw": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,1H0.5V0.5H1V1z"]
  },
  "tabbar_v_left_background_tab_n": {
    type: "SVG_PATHS",
    width: 39,
    height: 1,
    paths: ["M39,1H0V0.5h39V1z"]
  },
  "tabbar_v_left_background_tab_ne": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,1H0V0.5h1V1z"]
  },
  "tabbar_v_left_background_tab_e": {
    type: "SVG_PATHS",
    width: 1,
    height: 25,
    paths: ["M1,0v25H0V0H1z"]
  },
  "tabbar_v_left_background_tab_se": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,0v0.5H0V0H1z"]
  },
  "tabbar_v_left_background_tab_s": {
    type: "SVG_PATHS",
    width: 39,
    height: 1,
    paths: ["M39,0v0.5H0V0H39z"]
  },
  "tabbar_v_left_background_tab_sw": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,0v0.5H0.5V0H1z"]
  },
  "tabbar_v_left_background_tab_w": {
    type: "SVG_PATHS",
    width: 1,
    height: 25,
    paths: ["M1,0v25H0.5V0H1z"]
  },
  "tabbar_v_left_background_tab_c": {
    type: "SVG_PATHS",
    width: 39,
    height: 25,
    paths: ["M39,0v25H0V0H39z"]
  },
  "tabbar_v_right_tab": {
    type: "SVG_SCALE9",
    scale9: [1, 1, 1, 1],
    width: 41,
    height: 27
  },
  "tabbar_v_right_tab_nw": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,0v1H0V0H1z"]
  },
  "tabbar_v_right_tab_n": {
    type: "SVG_PATHS",
    width: 39,
    height: 1,
    paths: ["M39,1H0V0h39V1z"]
  },
  "tabbar_v_right_tab_ne": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,1H0V0h1V1z"]
  },
  "tabbar_v_right_tab_e": {
    type: "SVG_PATHS",
    width: 1,
    height: 25,
    paths: ["M1,0v25H0V0H1z"]
  },
  "tabbar_v_right_tab_se": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,0v1H0V0H1z"]
  },
  "tabbar_v_right_tab_s": {
    type: "SVG_PATHS",
    width: 39,
    height: 1,
    paths: ["M0,0h39v1H0V0z"]
  },
  "tabbar_v_right_tab_sw": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M0,0h1v1H0V0z"]
  },
  "tabbar_v_right_tab_w": {
    type: "SVG_PATHS",
    width: 1,
    height: 25,
    paths: []
  },
  "tabbar_v_right_tab_c": {
    type: "SVG_PATHS",
    width: 39,
    height: 25,
    paths: []
  },
  "tabbar_v_right_background_tab": {
    type: "SVG_SCALE9",
    scale9: [1, 1, 1, 1],
    width: 41,
    height: 27
  },
  "tabbar_v_right_background_tab_nw": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,1H0V0.5h1V1z"]
  },
  "tabbar_v_right_background_tab_n": {
    type: "SVG_PATHS",
    width: 39,
    height: 1,
    paths: ["M39,1H0V0.5h39V1z"]
  },
  "tabbar_v_right_background_tab_ne": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M0.5,1H0V0.5h0.5V1z"]
  },
  "tabbar_v_right_background_tab_e": {
    type: "SVG_PATHS",
    width: 1,
    height: 25,
    paths: ["M0.5,0v25H0V0H0.5z"]
  },
  "tabbar_v_right_background_tab_se": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M0.5,0v0.5H0V0H0.5z"]
  },
  "tabbar_v_right_background_tab_s": {
    type: "SVG_PATHS",
    width: 39,
    height: 1,
    paths: ["M39,0v0.5H0V0H39z"]
  },
  "tabbar_v_right_background_tab_sw": {
    type: "SVG_PATHS",
    width: 1,
    height: 1,
    paths: ["M1,0v0.5H0V0H1z"]
  },
  "tabbar_v_right_background_tab_w": {
    type: "SVG_PATHS",
    width: 1,
    height: 25,
    paths: ["M1,0v25H0V0H1z"]
  },
  "tabbar_v_right_background_tab_c": {
    type: "SVG_PATHS",
    width: 39,
    height: 25,
    paths: ["M39,0v25H0V0H39z"]
  },
  "textarea_padding_left": 6,
  "textarea_padding_right": 6,
  "textarea_padding_top": 6,
  "textarea_padding_bottom": 2,
  "textarea_scrollbar_padding_right": 0,
  "textarea_scrollbar_padding_top": 0,
  "textarea_scrollbar_padding_bottom": 0,
  "textinput_label_padding_top": 6,
  "textinput_label_padding_bottom": 5,
  "textinput_label_padding_left": 5,
  "textinput_label_padding_right": 5,
  "textinput_empty_space_bottom": 0,
  "textinput_border_padding": 6,
  "tooltip": {
    type: "SVG_SCALE9",
    scale9: [3, 3, 3, 3],
    width: 77,
    height: 25
  },
  "tooltip_nw": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    colors: [0xffffff, 0x676767],
    paths: ["M3,3H0.5V2.5c0-1.105,0.896-2,2-2H3V3z", "M1,3H0V2.5C0,1.121,1.122,0,2.5,0H3v1H2.5C1.673,1,1,1.673,1,2.5V3z"]
  },
  "tooltip_n": {
    type: "SVG_PATHS",
    width: 71,
    height: 3,
    colors: [0xffffff, 0x676767],
    paths: ["M70.938,3H0V0.5h70.938V3z", "M0,1V0h70.938v1H0z"]
  },
  "tooltip_ne": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    colors: [0xffffff, 0x676767],
    paths: ["M2.5,3h-2.562V0.5H0.5c1.104,0,2,0.895,2,2V3z", "M3,3H2V2.5C2,1.673,1.327,1,0.5,1h-0.562V0H0.5C1.878,0,3,1.121,3,2.5V3z"]
  },
  "tooltip_e": {
    type: "SVG_PATHS",
    width: 3,
    height: 19,
    colors: [0xffffff, 0x676767],
    paths: ["M2.5,0v19h-2.562V0H2.5z", "M3,0v19H2V0H3z"]
  },
  "tooltip_se": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    colors: [0xffffff, 0x676767],
    paths: ["M2.5,0v0.5c0,1.105-0.896,2-2,2h-0.562V0H2.5z", "M3,0v0.5C3,1.879,1.878,3,0.5,3h-0.562V2H0.5C1.327,2,2,1.327,2,0.5V0H3z"]
  },
  "tooltip_s": {
    type: "SVG_PATHS",
    width: 71,
    height: 3,
    colors: [0xffffff, 0x676767],
    paths: ["M70.938,0v2.5H0V0H70.938z", "M0,3V2h70.938v1H0z"]
  },
  "tooltip_sw": {
    type: "SVG_PATHS",
    width: 3,
    height: 3,
    colors: [0xffffff, 0x676767],
    paths: ["M3,0v2.5H2.5c-1.104,0-2-0.895-2-2V0H3z", "M2.5,2H3v1H2.5C1.122,3,0,1.879,0,0.5V0h1v0.5C1,1.327,1.673,2,2.5,2z"]
  },
  "tooltip_w": {
    type: "SVG_PATHS",
    width: 3,
    height: 19,
    colors: [0xffffff, 0x676767],
    paths: ["M3,0v19H0.5V0H3z", "M1,0v19H0V0H1z"]
  },
  "tooltip_c": {
    type: "SVG_PATHS",
    width: 71,
    height: 19,
    colors: [0xffffff],
    paths: ["M70.938,0v19H0V0H70.938z"]
  },
  "tooltip_tip": {
    type: "SVG_PATHS",
    width: 16,
    height: 14,
    colors: [0xffffff, 0xffffff, 0x676767, 0xffffff],
    paths: ["M8.016,1L15,8v2.5L1,10.516V8L8.016,1z", "M0.989,7.739L8.029,0.7l6.972,6.971", "M0.989,8.446L0.282,7.739l7.747-7.746l7.679,7.678l-0.707,0.707L8.029,1.407L0.989,8.446z", "M16,10.516H0v-2.5h16V10.516z"]
  },
  "tooltip_padding_left": 12,
  "tooltip_padding_right": 12,
  "tooltip_padding_top": 1,
  "tooltip_padding_bottom": 1,
  // Tree
  "tree_scrollbar_padding_top": 0,
  "tree_scrollbar_padding_bottom": 0,
  "tree_scrollbar_padding_right": 0,
  "verticalScrollbar": {
    type: "SVG_SCALE9",
    scale9: [12, 2, 2, 12],
    width: 16,
    height: 61
  },
  "verticalScrollbar_nw": {
    type: "SVG_PATHS",
    width: 2,
    height: 12,
    colors: [0x515151],
    paths: ["M1,12H0V0h2v1H1V12z"]
  },
  "verticalScrollbar_n": {
    type: "SVG_PATHS",
    width: 12,
    height: 12,
    colors: [0x515151, 0x515151],
    paths: ["M0,1V0h12v1H0z", "M2,8l4-4l4,4H2z"]
  },
  "verticalScrollbar_ne": {
    type: "SVG_PATHS",
    width: 2,
    height: 12,
    colors: [0x515151],
    paths: ["M2,12H1V1H0V0h2V12z"]
  },
  "verticalScrollbar_e": {
    type: "SVG_PATHS",
    width: 2,
    height: 37,
    colors: [0x515151],
    paths: ["M2,0v37H1V0H2z"]
  },
  "verticalScrollbar_se": {
    type: "SVG_PATHS",
    width: 2,
    height: 12,
    colors: [0x515151],
    paths: ["M2,0v12H0v-1h1V0H2z"]
  },
  "verticalScrollbar_s": {
    type: "SVG_PATHS",
    width: 12,
    height: 12,
    colors: [0x515151, 0x515151],
    paths: ["M0,12v-1h12v1H0z", "M10,4L6,8L2,4H10z"]
  },
  "verticalScrollbar_sw": {
    type: "SVG_PATHS",
    width: 2,
    height: 12,
    colors: [0x515151],
    paths: ["M1,11h1v1H0V0h1V11z"]
  },
  "verticalScrollbar_w": {
    type: "SVG_PATHS",
    width: 2,
    height: 37,
    colors: [0x515151],
    paths: ["M1,0v37H0V0H1z"]
  },
  "verticalScrollbar_c": {
    type: "SVG_PATHS",
    width: 12,
    height: 37,
    colors: [0x515151],
    paths: []
  },
  "verticalScrollbar_padding_top": 13,
  "verticalScrollbar_padding_bottom": 14,
  "verticalScrollbar_thumb": {
    type: "SVG_PATHS",
    width: 8,
    height: 21,
    colors: [0x515151],
    paths: ["M4,21c-2.209,0-4-1.791-4-4V4c0-2.209,1.791-4,4-4l0,0c2.209,0,4,1.791,4,4v13C8,19.209,6.209,21,4,21L4,21z"]
  },
  "videoplayer_color": 0x000000,
  // unused
  "videoplayer": {
    type: "SVG_SCALE9",
    scale9: [1, 137, 71, 39],
    width: 300,
    height: 200
  },
  "videoplayer_nw": {
    type: "SVG_PATHS",
    width: 137,
    height: 1,
    colors: [0x676767],
    paths: ["M137,0v1H1H0V0H137z"]
  },
  "videoplayer_n": {
    type: "SVG_PATHS",
    width: 92,
    height: 1,
    colors: [0x676767],
    paths: ["M0,1V0h92v1H0z"]
  },
  "videoplayer_ne": {
    type: "SVG_PATHS",
    width: 71,
    height: 1,
    colors: [0x676767],
    paths: ["M71,1h-1H0V0h71V1z"]
  },
  "videoplayer_e": {
    type: "SVG_PATHS",
    width: 71,
    height: 160,
    colors: [0x676767],
    paths: ["M71,0v160h-1V0H71z"]
  },
  "videoplayer_se": {
    type: "SVG_PATHS",
    width: 71,
    height: 39,
    colors: [0x676767, 0x676767, 0x676767, 0x676767, 0x676767, 0x676767, 0x676767],
    paths: ["M1,25H0v-1V14v-1h1V25z", "M25.382,27.271l-1.595-1.596c1.621-1.619,2.56-3.717,2.643-5.909c0.082-2.158-0.678-4.153-2.141-5.616l1.595-1.596c1.912,1.913,2.906,4.505,2.801,7.299C28.578,22.615,27.405,25.25,25.382,27.271L25.382,27.271z", "M20.246,29.787l-5.954-6.219H9v-7.227h5.292l5.954-6.219V29.787z", "M24.05,22.045c-0.597,0.596-1.373,0.895-2.154,0.912v-6.261c0.782,0.016,1.56,0.315,2.155,0.913C25.273,18.832,25.273,20.82,24.05,22.045z", "M58.986,13.975C58.985,11.784,57.302,10,55.234,10l-16.808,0.019c-2.066,0.004-3.745,1.789-3.743,3.982l0.014,12.023C34.7,28.216,36.383,30,38.448,30c0.001,0,0.002,0,0.002,0l16.809-0.019c0.962-0.002,1.873-0.387,2.569-1.086c0.748-0.75,1.175-1.808,1.172-2.896L58.986,13.975zM46.529,28.312l-8.414,0.012c0,0-0.001,0-0.002,0c-0.783,0-1.419-0.636-1.421-1.42l-0.007-5.966c-0.001-0.376,0.147-0.738,0.414-1.005c0.267-0.268,0.627-0.416,1.005-0.418l8.414-0.008h0.001c0.784,0,1.42,0.635,1.421,1.42l0.008,5.963C47.95,27.673,47.313,28.311,46.529,28.312zM56.089,17.489c0,0.393-0.318,0.71-0.711,0.71s-0.71-0.317-0.71-0.71v-2.645l-3.235,3.233c-0.14,0.14-0.32,0.209-0.502,0.209c-0.183,0-0.364-0.069-0.502-0.209c-0.278-0.276-0.278-0.726,0-1.004l3.253-3.253l-2.637,0.036c-0.382-0.048-0.714-0.31-0.72-0.701c-0.006-0.393,0.308-0.714,0.7-0.72l4.344-0.059c0.181,0.009,0.374,0.072,0.509,0.205c0.135,0.134,0.211,0.315,0.211,0.505V17.489z", "M71,0v39H0v-1h70V0H71z", "M0,0h71v1H0V0z"]
  },
  "videoplayer_s": {
    type: "SVG_PATHS",
    width: 92,
    height: 39,
    colors: [0x676767, 0x676767, 0x676767, 0x676767],
    paths: ["M0,25v-1h92v1H0z", "M0,14v-1h92v1H0z", "M0,39v-1h92v1H0z", "M0,0h92v1H0V0z"]
  },
  "videoplayer_sw": {
    type: "SVG_PATHS",
    width: 137,
    height: 39,
    colors: [0x989898, 0x676767, 0xffffff, 0x676767, 0x676767, 0x676767],
    paths: ["M41.5,13.5H137v11H41.5V13.5z", "M30,18.5C30,24.289,25.29,29,19.5,29C13.709,29,9,24.289,9,18.5S13.709,8,19.5,8C25.29,8,30,12.711,30,18.5z", "M16.803,13.304l7.858,5.196l-7.858,5.195V13.304z", "M42,24h95v1H41V13h96v1H42V24z", "M1,38h136v1H0V0h1V38z", "M0,0h137v1H0V0z"]
  },
  "videoplayer_w": {
    type: "SVG_PATHS",
    width: 137,
    height: 160,
    colors: [0x676767],
    paths: ["M0,0h1v160H0V0z"]
  },
  "videoplayer_c": {
    type: "SVG_PATHS",
    width: 92,
    height: 160,
    paths: []
  },
  "volumeslider": {
    type: "SVG_SCALE9",
    scale9: [1, 52, 2, 1],
    width: 72,
    height: 16
  },
  "volumeslider_nw": {
    type: "SVG_PATHS",
    width: 52,
    height: 1,
    paths: ["M9.138,1H8.18l0.958-1V1z"]
  },
  "volumeslider_n": {
    type: "SVG_PATHS",
    width: 18,
    height: 1,
    paths: []
  },
  "volumeslider_ne": {
    type: "SVG_PATHS",
    width: 2,
    height: 1,
    paths: []
  },
  "volumeslider_e": {
    type: "SVG_PATHS",
    width: 2,
    height: 14,
    paths: []
  },
  "volumeslider_se": {
    type: "SVG_PATHS",
    width: 2,
    height: 1,
    paths: []
  },
  "volumeslider_s": {
    type: "SVG_PATHS",
    width: 18,
    height: 1,
    paths: []
  },
  "volumeslider_sw": {
    type: "SVG_PATHS",
    width: 52,
    height: 1,
    paths: ["M9.138,0.977L8.202,0h0.936V0.977z"]
  },
  "volumeslider_w": {
    type: "SVG_PATHS",
    width: 52,
    height: 14,
    colors: [0x666666, 0x666666, 0x666666, 0x666666, 0xFFFFFF, 0x666666],
    paths: ["M13.718,0.976c1.554,1.554,2.361,3.66,2.275,5.931c-0.086,2.244-1.039,4.385-2.684,6.027l-1.295-1.297c1.317-1.314,2.079-3.02,2.147-4.802c0.066-1.754-0.551-3.374-1.739-4.562L13.718,0.976z", "M9.138,0v14H8.202L4.3,9.926H0V4.053h4.3L8.18,0H9.138z", "M12.228,8.686c-0.484,0.486-1.116,0.729-1.75,0.742V4.34c0.636,0.013,1.268,0.258,1.752,0.743C13.223,6.076,13.223,7.691,12.228,8.686z", "M52.001,7v1H20V7H52.001z", "M49.501,7.5c0,2.76-2.24,5-5,5c-2.762,0-5-2.24-5-5c0-2.762,2.238-5,5-5C47.261,2.5,49.501,4.738,49.501,7.5z", "M44.501,13c-3.033,0-5.5-2.468-5.5-5.5s2.467-5.5,5.5-5.5s5.5,2.468,5.5,5.5S47.534,13,44.501,13zM44.501,3c-2.481,0-4.5,2.019-4.5,4.5s2.019,4.5,4.5,4.5s4.5-2.019,4.5-4.5S46.982,3,44.501,3z"]
  },
  "volumeslider_c": {
    type: "SVG_PATHS",
    width: 18,
    height: 14,
    colors: [0x666666],
    paths: ["M18.001,8h-18V7h18V8z"]
  },
  "webcam_color": 0x000000,
  // unused
  "webcam": {
    type: "SVG_PATHS",
    width: 177,
    height: 146,
    colors: [0x676767, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0x676767],
    paths: ["M177,146H0V0h177V146z", "M128.967,57.982c0,22.193-18.117,42.066-40.467,42.066S48.033,80.176,48.033,57.982c0-22.191,18.117-44.855,40.467-44.855S128.967,35.791,128.967,57.982z", "M121.45,80.148c0,22.192-21.895,40.182-32.95,40.182c-9.957,0-32.95-17.989-32.95-40.182c0-22.191,14.752-40.182,32.95-40.182C106.697,39.967,121.45,57.957,121.45,80.148z", "M88.5,152.75", "M104.612,138.867H72.388l3-59.17h26.225L104.612,138.867z", "M11.062,148c0,0,0.578-8.345,3.601-9.311c6.384-2.04,42.202-12.292,57.358-19.269c15.695-7.228,21.291-3.674,21.291-3.674V148H11.062z", "M83.688,148v-32.253c0,0,5.596-3.554,21.291,3.674c15.156,6.977,50.975,17.229,57.358,19.269c3.022,0.966,3.601,9.311,3.601,9.311H83.688z", "M60.76,71.002v18.007l-7.275-3.747l-4.244-14.26C51.222,62.349,60.76,71.002,60.76,71.002z", "M127.76,71.002l-4.244,14.26l-7.275,3.747V71.002C116.24,71.002,125.778,62.349,127.76,71.002z", "M178,147H-1V-1h179V147zM1,145h175V1H1V145z"]
  },
  "window_title_padding_left": 5,
  "window_title_padding_top": 8,
  "window_title_padding_right": 53,
  "window_scrollbar_padding_top": -1,
  "window_scrollbar_padding_right": 0,
  "window_scrollbar_padding_bottom": -1,
  "window_maximizeRestore_padding_right": 22,
  "window_minimize_padding_right": 38,
  "window_close_padding_right": 8,
  "window_maximizeRestore_padding_top": 6,
  "window_minimize_padding_top": 6,
  "window_close_padding_top": 6,
  "window_dragger": {
    type: "SVG_PATHS",
    width: 13,
    height: 11.001,
    paths: ["M11.5,1L2.5,10", "M3.03,10.53L1.969,9.47l9.001-9l1.061,1.061L3.03,10.53z", "M12,6l-4,4", "M8.531,10.53L7.47,9.47l4-4l1.061,1.061L8.531,10.53z"]
  },
  "window_dragger_padding_right": 17,
  "window_dragger_padding_left": -1,
  "window_dragger_padding_bottom": 17,
  // Temporary, copied from Sketch
  "Search": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M23.354,29.177c-4.516,0.738-9.184,1.213-13.549-1.484C0.668,22.042,3.885,5.81,12.701,2.22C27.41-3.771,41.004,9.884,31.4,24.169c0.406,4.076,10.229,12.113,11.984,17.238c-0.984,4.023-2.217,5.008-7.527,4.811C31.99,40.302,24.832,29.915,23.354,29.177z M23.699,4.536c-2.369-0.57-7.197-0.738-9.26,0c-11.758,4.189-9.07,25.035,6.943,21.998C31.348,24.646,36.121,7.521,23.699,4.536z M29.02,26.222c-1.396,0.82-0.74,0.574-1.971,1.559c3.637,5.816,10.623,15.158,13.053,15.27C40.791,40.64,32.9,32.397,29.02,26.222z"]
  },
  "CheckBox": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M41,42H4V4h37V42zM6,40h33V6H6V40z"]
  },
  "CheckBoxIndeterminate": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M33,25H12v-4h21V25z", "M41,42H4V4h37V42zM6,40h33V6H6V40z"]
  },
  "CheckBoxSelected": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M41,42H4V4h37V42zM6,40h33V6H6V40z", "M17.312,38.312c-1.057-0.635-6.382-6.067-7.25-7.562c-1.125-1.938-1.125-2.812-0.375-3.812s2.125-2.062,3.812-1.188c1.688,1.125,4.938,5,4.562,4.562s1.562,1.812,2.812,1.875s7.395-7.35,8.438-8.5C31.75,21,41.188,7.125,42.438,5.625S44.875,4.062,46,4.875s1.447,3.132,0,5.188C38.562,20.625,26.369,36.17,24.188,38.125C21.188,40.812,19.188,39.438,17.312,38.312z"]
  },
  "CheckMarkIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M17.144,46.968c0,0-0.003,0-0.006,0c-1.517,0-2.97-0.605-4.04-1.681L2.642,34.784c-2.225-2.235-2.216-5.848,0.018-8.073c2.234-2.222,5.847-2.216,8.074,0.018l6.418,6.448L37.23,2.724c2.23-2.233,5.846-2.233,8.075-0.003c2.23,2.229,1.673,5.328,0.003,8.072L21.181,45.296C20.111,46.366,18.659,46.968,17.144,46.968L17.144,46.968z"]
  },
  "CircledMinusIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M40.16,7.871c-8.942-8.939-23.437-8.939-32.383,0c-8.939,8.941-8.939,23.438,0,32.381c8.946,8.939,23.44,8.939,32.383,0C49.101,31.31,49.101,16.812,40.16,7.871zM38.303,26.911H9.638v-5.707h28.665V26.911z"]
  },
  "CircledPlusIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M40.051,7.859c-8.92-8.915-23.375-8.915-32.291,0c-8.917,8.917-8.917,23.372,0,32.292c8.916,8.914,23.371,8.914,32.291,0C48.965,31.231,48.965,16.776,40.051,7.859zM38.197,26.855H26.759v11.439h-5.707V26.855H9.614v-5.706h11.438V9.711h5.707v11.438h11.438V26.855z"]
  },
  "DotIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M24.021,6.975c-9.387,0-17.023,7.638-17.023,17.022c0,9.39,7.636,17.028,17.023,17.028s17.023-7.639,17.023-17.028C41.044,14.612,33.408,6.975,24.021,6.975z"]
  },
  "DownFillTriangleIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M1.443,8.064c0-0.43,0.106-0.858,0.318-1.246C2.212,5.994,3.079,5.48,4.024,5.48l40.883,0.006c0.941,0,1.809,0.513,2.263,1.338c0.455,0.824,0.424,1.828-0.08,2.626L26.643,41.696c-0.474,0.741-1.296,1.195-2.183,1.195c-0.886,0-1.708-0.454-2.183-1.195L1.847,9.447C1.579,9.025,1.443,8.547,1.443,8.064L1.443,8.064z"]
  },
  "GtCircledIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M35.298,24.533c0,0.157-0.064,0.309-0.18,0.416l-13.773,12.95c-0.109,0.103-0.248,0.153-0.389,0.153c-0.15,0-0.299-0.06-0.412-0.179l-3.318-3.528c-0.102-0.109-0.158-0.255-0.152-0.405c0.004-0.149,0.068-0.292,0.178-0.395l9.529-8.959l-9.502-8.719c-0.111-0.102-0.178-0.243-0.184-0.393s0.047-0.297,0.148-0.407l3.275-3.57c0.211-0.23,0.57-0.246,0.799-0.034L35.113,24.12C35.23,24.227,35.296,24.376,35.298,24.533z", "M24.04,1.08C11.315,1.08,1,11.396,1,24.123c0,12.726,10.315,23.045,23.04,23.045c12.73,0,23.046-10.319,23.046-23.045C47.085,11.396,36.77,1.08,24.04,1.08zM24.042,42.27c-10.021,0-18.146-8.125-18.146-18.146c0-10.022,8.125-18.147,18.146-18.147c10.021,0,18.148,8.125,18.148,18.147C42.19,34.145,34.063,42.27,24.042,42.27z"]
  },
  "GtIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M39.883,23.831c0.002,0.27-0.108,0.529-0.307,0.714L15.912,46.794c-0.187,0.177-0.426,0.264-0.665,0.264c-0.258,0-0.516-0.103-0.708-0.307l-5.701-6.062c-0.176-0.188-0.271-0.438-0.263-0.695c0.008-0.257,0.118-0.502,0.306-0.678l16.373-15.393L8.926,8.943C8.736,8.77,8.623,8.527,8.611,8.269C8.601,8.012,8.692,7.76,8.866,7.57l5.628-6.135c0.363-0.396,0.979-0.422,1.374-0.059l23.7,21.743C39.768,23.304,39.881,23.561,39.883,23.831z"]
  },
  "IpadKeyboard": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M48,30H0V7h48V30zM2,28h44V9H2V28z", "M23.402,42.068l-0.136-0.099l-8.585-6.255l1.178-1.617l7.543,5.495l7.541-5.495l1.178,1.617l-8.584,6.255L23.402,42.068z", "M7,15H5v-2h2V15z", "M11,15H9v-2h2V15z", "M15,15h-2v-2h2V15z", "M19,15h-2v-2h2V15z", "M23,15h-2v-2h2V15z", "M27,15h-2v-2h2V15z", "M31,15h-2v-2h2V15z", "M35,15h-2v-2h2V15z", "M39,15h-2v-2h2V15z", "M43,15h-2v-2h2V15z", "M7,20H5v-2h2V20z", "M11,20H9v-2h2V20z", "M15,20h-2v-2h2V20z", "M19,20h-2v-2h2V20z", "M23,20h-2v-2h2V20z", "M27,20h-2v-2h2V20z", "M31,20h-2v-2h2V20z", "M35,20h-2v-2h2V20z", "M39,20h-2v-2h2V20z", "M43,20h-2v-2h2V20z", "M7,24.5H5v-2h2V24.5z", "M11,24.5H9v-2h2V24.5z", "M35,24.5H13v-2h22V24.5z", "M39,24.5h-2v-2h2V24.5z", "M43,24.5h-2v-2h2V24.5z"]
  },
  "LeftFillTriangleIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M40.586,1.163c0.429,0,0.857,0.106,1.245,0.317c0.825,0.452,1.338,1.318,1.338,2.264l-0.005,40.884c0,0.942-0.513,1.81-1.339,2.264c-0.824,0.454-1.828,0.424-2.625-0.081L6.953,26.364c-0.741-0.473-1.195-1.297-1.195-2.182c0-0.888,0.454-1.709,1.195-2.184l32.25-20.432C39.625,1.3,40.104,1.163,40.586,1.163L40.586,1.163z"]
  },
  "ListIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M5.726,1.947h36.815c2.45,0,4.438,2.166,4.438,4.84c0,2.671-1.987,4.837-4.438,4.837H5.726c-2.454,0-4.441-2.166-4.441-4.837l0,0C1.284,4.113,3.271,1.947,5.726,1.947z", "M5.726,18.563h36.815c2.45,0,4.438,2.166,4.438,4.839s-1.987,4.836-4.438,4.836H5.726c-2.454,0-4.441-2.163-4.441-4.836l0,0C1.284,20.729,3.271,18.563,5.726,18.563z", "M5.726,35.182h36.815c2.45,0,4.438,2.166,4.438,4.839c0,2.676-1.987,4.836-4.438,4.836H5.726c-2.454,0-4.441-2.16-4.441-4.836l0,0C1.284,37.348,3.271,35.182,5.726,35.182z"]
  },
  "MicIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M15.158,25.064V9.786c0-4.843,3.922-8.782,8.744-8.782c4.821,0,8.739,3.939,8.739,8.782v15.278c0,4.845-3.918,8.783-8.739,8.783C19.08,33.848,15.158,29.909,15.158,25.064z", "M40.21,20.987v3.163c0,8.138-5.633,14.91-13.016,16.288v2.983h5.402v3.559H15.202v-3.559h5.403v-2.983C13.22,39.061,7.591,32.29,7.591,24.15v-3.163c0-1.191,0.967-2.159,2.159-2.159c1.192,0,2.159,0.968,2.159,2.159v3.163c0,6.749,5.141,12.239,11.458,12.239h1.064c6.32,0,11.46-5.49,11.46-12.239v-3.163c0-1.191,0.967-2.159,2.158-2.159C39.244,18.828,40.21,19.796,40.21,20.987z"]
  },
  "RadioButton": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M24,41.5c-9.649,0-17.5-7.851-17.5-17.5S14.351,6.5,24,6.5S41.5,14.351,41.5,24S33.649,41.5,24,41.5zM24,9.5C16.005,9.5,9.5,16.005,9.5,24S16.005,38.5,24,38.5S38.5,31.995,38.5,24S31.995,9.5,24,9.5z"]
  },
  "RadioButtonIndeterminate": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M24,41.5c-9.649,0-17.5-7.851-17.5-17.5S14.351,6.5,24,6.5S41.5,14.351,41.5,24S33.649,41.5,24,41.5zM24,9.5C16.005,9.5,9.5,16.005,9.5,24S16.005,38.5,24,38.5S38.5,31.995,38.5,24S31.995,9.5,24,9.5z", "M33,26H15v-4h18V26z"]
  },
  "RadioButtonSelected": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M24,41.5c-9.649,0-17.5-7.851-17.5-17.5S14.351,6.5,24,6.5S41.5,14.351,41.5,24S33.649,41.5,24,41.5zM24,9.5C16.005,9.5,9.5,16.005,9.5,24S16.005,38.5,24,38.5S38.5,31.995,38.5,24S31.995,9.5,24,9.5z", "M32,24c0,4.418-3.582,8-8,8s-8-3.582-8-8s3.582-8,8-8S32,19.582,32,24z", "M24,32.5c-4.687,0-8.5-3.813-8.5-8.5s3.813-8.5,8.5-8.5s8.5,3.813,8.5,8.5S28.687,32.5,24,32.5zM24,16.5c-4.136,0-7.5,3.364-7.5,7.5s3.364,7.5,7.5,7.5s7.5-3.364,7.5-7.5S28.136,16.5,24,16.5z"]
  },
  "RightFillTriangleIcon": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M11.194,47.212c-0.429,0-0.856-0.106-1.245-0.317c-0.824-0.452-1.337-1.319-1.337-2.264L8.617,3.746c0-0.943,0.513-1.81,1.339-2.263c0.825-0.456,1.829-0.425,2.625,0.081l32.247,20.446c0.741,0.474,1.195,1.297,1.195,2.182c0,0.888-0.454,1.709-1.195,2.184L12.578,46.808C12.157,47.075,11.678,47.212,11.194,47.212L11.194,47.212z"]
  },
  // Temporary, copied to Sketch
  "EmptyIconFontAwesome": {
    type: "SVG_PATHS",
    width: 48,
    height: 48,
    paths: ["M43.099,47.049H4.878C2.74,47.049,1,45.309,1,43.17V5.049C1,2.91,2.74,1.17,4.878,1.17h38.221c2.139,0,3.879,1.74,3.879,3.879V43.17C46.978,45.309,45.238,47.049,43.099,47.049L43.099,47.049zM4.878,3.17C3.843,3.17,3,4.013,3,5.049V43.17c0,1.036,0.844,1.879,1.879,1.879h38.221c1.037,0,1.879-0.843,1.879-1.879V5.049c0-1.036-0.842-1.879-1.879-1.879H4.878L4.878,3.17z"]
  },
  "timepicker_lineWeight": 1,
  "timepicker_backgroundColor": 0xFFFFFF,
  "timepicker_headingPanelBgColor": 0xDDDDDD,
  "timepicker_clockFaceColor": 0xEEEEEE,
  "timepicker_activeTextColor": 0x000000,
  "timepicker_inactiveTextColor": 0x999999,
  "timepicker_padding": 10,
  "timepicker_defaultHeadingPortrait": 95,
  "timepicker_defaultHeadingLandscape": 188,
  "timepicker_defaultWidthLandscape": 570,
  "timepicker_defaultWidthPortrait": 320,
  "timepicker_defaultHeightLandscape": 303,
  "timepicker_defaultHeightPortrait": 473,
  "timepicker_clockSizePortrait": 285,
  "timepicker_clockSizeLandscape": 250,
  "datepicker_iconSize": 24,
  "datepicker_color": 0x333333,
  "datepicker_paddingDate": 10,
  "datepicker_lineWeightDate": 1,
  "datepicker_fontSize": 13,
  "datepicker_buttonFontSize": 14,
  "datepicker_headingPanelBgColor": 0xDDDDDD,
  "datepicker_defaultHeadingLandscape": 188,
  "datepicker_defaultWidthLandscape": 570,
  "datepicker_defaultHeightLandscape": 303,
  "datepicker_defaultCalendarHeightLandscape": 260,
  "datepicker_yearLabelFontSizeLandscape": 18,
  "datepicker_dateLabelFontSizeLandscape": 40,
  "datepicker_defaultHeadingPortrait": 95,
  "datepicker_defaultWidthPortrait": 320,
  "datepicker_defaultHeightPortrait": 473,
  "datepicker_defaultCalendarWidthPortrait": 290,
  "datepicker_yearLabelFontSizePortrait": 14,
  "datepicker_dateLabelFontSizePortrait": 32,
  "__last": null
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/sqlitedriver.js":
/*!*********************************!*\
  !*** ./js/core/sqlitedriver.js ***!
  \*********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = exports.SQLiteDriver = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class SQLiteDriver {
  constructor(context, sqliteID) {
    _balsamiq.default.exists(context.host);

    _balsamiq.default.assertString(sqliteID);

    this.context = context;
    this.sqliteID = sqliteID;
    this.sequenceCounter = 0;
    this.openMode = null; // String
  }

  _checkID(id) {
    _balsamiq.default.assert(id === this.sqliteID, "can not call SQLiteDriver with another id");
  }

  _openIfNeeded(callback) {
    _balsamiq.default.assert(this.sequenceCounter >= 1);

    _balsamiq.default.assertString(this.openMode, "Can not open the file without openMode. Call 'open' before.");

    if (!this.context.host.SQLite.isOpen(this.sqliteID)) {
      this.context.host.SQLite.open(this.sqliteID, this.openMode, res => {
        this.context.host.asyncCallback(callback, [res]); // res.error
      });
    } else {
      this.context.host.asyncCallback(callback, [{}]);
    }
  }

  beginSequence(id) {
    this._checkID(id);

    this.sequenceCounter += 1;
  }

  endSequence(id) {
    this._checkID(id);

    _balsamiq.default.assert(this.sequenceCounter >= 1);

    this.sequenceCounter -= 1;

    if (this.sequenceCounter == 0 && this.context.host.SQLite.isOpen(this.sqliteID)) {
      this.context.host.SQLite.close(this.sqliteID, res => {// TODO what to do with the res?
      });
    }
  }

  open(id, mode, callback) {
    this._checkID(id);

    _balsamiq.default.assert(this.openMode === null, "The SQLite db is already 'open'.");

    _balsamiq.default.assertString(mode);

    this.openMode = mode;

    this._openIfNeeded(callback); // // TODO check permissions: that mode <= than host sql mode
    // if (this.context.host.SQLite.has(id)) {
    //     // TODO check res
    //     this.context.host.asyncCallback(callback, [{}]);
    //         });
    //     });
    // } else {
    //     this.context.host.asyncCallback(callback, [{ error: `No db with id ${id}` }]);
    // }

  }

  checkIntegrity(id, callback) {
    this._checkID(id);

    this.context.host.SQLite.query(id, "pragma integrity_check;", {}, result => {
      if (result.error || result.rows[0]["integrity_check"] !== "ok") {
        callback({
          error: `the db ${id} is malformed`
        });
      } else {
        callback({});
      }
    });
  }

  close(id, callback) {
    this._checkID(id);

    _balsamiq.default.assert(this.sequenceCounter == 1); // Balsamiq.assert(!this.context.host.SQLite.isOpen(this.sqliteID));


    this.openMode = null;
    this.context.host.asyncCallback(callback, [{}]);
  }

  run(id, query, params, callback) {
    this._checkID(id);

    _balsamiq.default.assertString(query);

    this._openIfNeeded(res => {
      if (!res.error) {
        this.context.host.SQLite.query(id, query, params, result => {
          this.context.host.SQLite.query(id, "SELECT CHANGES();", [], obj => {
            result.affectedRows = obj.rows[0]["CHANGES()"];
            callback(result);
          });
        });
      } else {
        callback(res);
      }
    });
  }

  all(id, query, params, callback) {
    this._checkID(id);

    _balsamiq.default.assertString(query);

    this._openIfNeeded(res => {
      if (!res.error) {
        this.context.host.SQLite.query(id, query, params, function (res) {
          callback({
            rows: res.rows
          });
        });
      } else {
        callback(res);
      }
    });
  }

  get(id, query, params, callback) {
    this._checkID(id);

    _balsamiq.default.assertString(query);

    this._openIfNeeded(res => {
      if (!res.error) {
        this.context.host.SQLite.query(id, query, params, function (res) {
          if (res.error) {
            callback(res);
          } else {
            callback({
              row: res.rows[0]
            });
          }
        });
      } else {
        callback(res);
      }
    });
  }

  serialize(id, callback) {
    this.context.host.asyncCallback(callback, [{}]);
  }

  download(id, callback) {
    this.context.host.SQLite.download(id, callback, null);
  }

  compact(id, callback) {
    this._checkID(id);

    this._openIfNeeded(res => {
      if (!res.error) {
        this.context.host.SQLite.compact(id, callback);
      } else {
        callback(res);
      }
    });
  }

}

exports.SQLiteDriver = SQLiteDriver;
var _default = SQLiteDriver;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/texts/datagridGrammar.js":
/*!******************************************!*\
  !*** ./js/core/texts/datagridGrammar.js ***!
  \******************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.SyntaxError = peg$SyntaxError;
exports.parse = peg$parse;

var balsamiqTextDomUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textDomUtils */ "./js/core/texts/textDomUtils.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

/* eslint-disable */
// Generated by Peggy 1.2.0.
//
// https://peggyjs.org/
function peg$subclass(child, parent) {
  function C() {
    this.constructor = child;
  }

  C.prototype = parent.prototype;
  child.prototype = new C();
}

function peg$SyntaxError(message, expected, found, location) {
  var self = Error.call(this, message);

  if (Object.setPrototypeOf) {
    Object.setPrototypeOf(self, peg$SyntaxError.prototype);
  }

  self.expected = expected;
  self.found = found;
  self.location = location;
  self.name = "SyntaxError";
  return self;
}

peg$subclass(peg$SyntaxError, Error);

function peg$padEnd(str, targetLength, padString) {
  padString = padString || " ";

  if (str.length > targetLength) {
    return str;
  }

  targetLength -= str.length;
  padString += padString.repeat(targetLength);
  return str + padString.slice(0, targetLength);
}

peg$SyntaxError.prototype.format = function (sources) {
  var str = "Error: " + this.message;

  if (this.location) {
    var src = null;
    var k;

    for (k = 0; k < sources.length; k++) {
      if (sources[k].source === this.location.source) {
        src = sources[k].text.split(/\r\n|\n|\r/g);
        break;
      }
    }

    var s = this.location.start;
    var loc = this.location.source + ":" + s.line + ":" + s.column;

    if (src) {
      var e = this.location.end;
      var filler = peg$padEnd("", s.line.toString().length);
      var line = src[s.line - 1];
      var last = s.line === e.line ? e.column : line.length + 1;
      str += "\n --> " + loc + "\n" + filler + " |\n" + s.line + " | " + line + "\n" + filler + " | " + peg$padEnd("", s.column - 1) + peg$padEnd("", last - s.column, "^");
    } else {
      str += "\n at " + loc;
    }
  }

  return str;
};

peg$SyntaxError.buildMessage = function (expected, found) {
  var DESCRIBE_EXPECTATION_FNS = {
    literal: function (expectation) {
      return "\"" + literalEscape(expectation.text) + "\"";
    },
    class: function (expectation) {
      var escapedParts = expectation.parts.map(function (part) {
        return Array.isArray(part) ? classEscape(part[0]) + "-" + classEscape(part[1]) : classEscape(part);
      });
      return "[" + (expectation.inverted ? "^" : "") + escapedParts + "]";
    },
    any: function () {
      return "any character";
    },
    end: function () {
      return "end of input";
    },
    other: function (expectation) {
      return expectation.description;
    }
  };

  function hex(ch) {
    return ch.charCodeAt(0).toString(16).toUpperCase();
  }

  function literalEscape(s) {
    return s.replace(/\\/g, "\\\\").replace(/"/g, "\\\"").replace(/\0/g, "\\0").replace(/\t/g, "\\t").replace(/\n/g, "\\n").replace(/\r/g, "\\r").replace(/[\x00-\x0F]/g, function (ch) {
      return "\\x0" + hex(ch);
    }).replace(/[\x10-\x1F\x7F-\x9F]/g, function (ch) {
      return "\\x" + hex(ch);
    });
  }

  function classEscape(s) {
    return s.replace(/\\/g, "\\\\").replace(/\]/g, "\\]").replace(/\^/g, "\\^").replace(/-/g, "\\-").replace(/\0/g, "\\0").replace(/\t/g, "\\t").replace(/\n/g, "\\n").replace(/\r/g, "\\r").replace(/[\x00-\x0F]/g, function (ch) {
      return "\\x0" + hex(ch);
    }).replace(/[\x10-\x1F\x7F-\x9F]/g, function (ch) {
      return "\\x" + hex(ch);
    });
  }

  function describeExpectation(expectation) {
    return DESCRIBE_EXPECTATION_FNS[expectation.type](expectation);
  }

  function describeExpected(expected) {
    var descriptions = expected.map(describeExpectation);
    var i, j;
    descriptions.sort();

    if (descriptions.length > 0) {
      for (i = 1, j = 1; i < descriptions.length; i++) {
        if (descriptions[i - 1] !== descriptions[i]) {
          descriptions[j] = descriptions[i];
          j++;
        }
      }

      descriptions.length = j;
    }

    switch (descriptions.length) {
      case 1:
        return descriptions[0];

      case 2:
        return descriptions[0] + " or " + descriptions[1];

      default:
        return descriptions.slice(0, -1).join(", ") + ", or " + descriptions[descriptions.length - 1];
    }
  }

  function describeFound(found) {
    return found ? "\"" + literalEscape(found) + "\"" : "end of input";
  }

  return "Expected " + describeExpected(expected) + " but " + describeFound(found) + " found.";
};

function peg$parse(input, options) {
  options = options !== undefined ? options : {};
  var peg$FAILED = {};
  var peg$source = options.grammarSource;
  var peg$startRuleFunctions = {
    normalSequenceOrEmptySequence: peg$parsenormalSequenceOrEmptySequence
  };
  var peg$startRuleFunction = peg$parsenormalSequenceOrEmptySequence;
  var peg$c0 = " ";
  var peg$c1 = "\\";
  var peg$r0 = /^[,\t]/;
  var peg$e0 = peg$anyExpectation();
  var peg$e1 = peg$classExpectation([",", "\t"], false, false);
  var peg$e2 = peg$literalExpectation(" ", false);
  var peg$e3 = peg$literalExpectation("\\", false);

  var peg$f0 = function () {
    return [];
  };

  var peg$f1 = function (u) {
    return balsamiqTextDomUtils.joinStringUnits(u);
  };

  var peg$f2 = function (sep, hasSpace) {
    return {
      type: 'separator',
      value: balsamiqTextDomUtils.joinStringUnits([sep, !!hasSpace ? " " : ""])
    };
  };

  var peg$f3 = function (ch) {
    return ch;
  };

  var peg$f4 = function (cellText) {
    return {
      type: 'cell',
      value: balsamiqTextDomUtils.joinStringUnits([cellText])
    };
  };

  var peg$f5 = function (u) {
    return u;
  };

  var peg$currPos = 0;
  var peg$savedPos = 0;
  var peg$posDetailsCache = [{
    line: 1,
    column: 1
  }];
  var peg$maxFailPos = 0;
  var peg$maxFailExpected = [];
  var peg$silentFails = 0;
  var peg$result;

  if ("startRule" in options) {
    if (!(options.startRule in peg$startRuleFunctions)) {
      throw new Error("Can't start parsing from rule \"" + options.startRule + "\".");
    }

    peg$startRuleFunction = peg$startRuleFunctions[options.startRule];
  }

  function text() {
    return input.substring(peg$savedPos, peg$currPos);
  }

  function offset() {
    return peg$savedPos;
  }

  function range() {
    return {
      source: peg$source,
      start: peg$savedPos,
      end: peg$currPos
    };
  }

  function location() {
    return peg$computeLocation(peg$savedPos, peg$currPos);
  }

  function expected(description, location) {
    location = location !== undefined ? location : peg$computeLocation(peg$savedPos, peg$currPos);
    throw peg$buildStructuredError([peg$otherExpectation(description)], input.substring(peg$savedPos, peg$currPos), location);
  }

  function error(message, location) {
    location = location !== undefined ? location : peg$computeLocation(peg$savedPos, peg$currPos);
    throw peg$buildSimpleError(message, location);
  }

  function peg$literalExpectation(text, ignoreCase) {
    return {
      type: "literal",
      text: text,
      ignoreCase: ignoreCase
    };
  }

  function peg$classExpectation(parts, inverted, ignoreCase) {
    return {
      type: "class",
      parts: parts,
      inverted: inverted,
      ignoreCase: ignoreCase
    };
  }

  function peg$anyExpectation() {
    return {
      type: "any"
    };
  }

  function peg$endExpectation() {
    return {
      type: "end"
    };
  }

  function peg$otherExpectation(description) {
    return {
      type: "other",
      description: description
    };
  }

  function peg$computePosDetails(pos) {
    var details = peg$posDetailsCache[pos];
    var p;

    if (details) {
      return details;
    } else {
      p = pos - 1;

      while (!peg$posDetailsCache[p]) {
        p--;
      }

      details = peg$posDetailsCache[p];
      details = {
        line: details.line,
        column: details.column
      };

      while (p < pos) {
        if (input.charCodeAt(p) === 10) {
          details.line++;
          details.column = 1;
        } else {
          details.column++;
        }

        p++;
      }

      peg$posDetailsCache[pos] = details;
      return details;
    }
  }

  function peg$computeLocation(startPos, endPos) {
    var startPosDetails = peg$computePosDetails(startPos);
    var endPosDetails = peg$computePosDetails(endPos);
    return {
      source: peg$source,
      start: {
        offset: startPos,
        line: startPosDetails.line,
        column: startPosDetails.column
      },
      end: {
        offset: endPos,
        line: endPosDetails.line,
        column: endPosDetails.column
      }
    };
  }

  function peg$fail(expected) {
    if (peg$currPos < peg$maxFailPos) {
      return;
    }

    if (peg$currPos > peg$maxFailPos) {
      peg$maxFailPos = peg$currPos;
      peg$maxFailExpected = [];
    }

    peg$maxFailExpected.push(expected);
  }

  function peg$buildSimpleError(message, location) {
    return new peg$SyntaxError(message, null, null, location);
  }

  function peg$buildStructuredError(expected, found, location) {
    return new peg$SyntaxError(peg$SyntaxError.buildMessage(expected, found), expected, found, location);
  }

  function peg$parsenormalSequenceOrEmptySequence() {
    var s0, s1;
    s0 = peg$parsenormalSequence();

    if (s0 === peg$FAILED) {
      s0 = peg$currPos;
      s1 = '';
      peg$savedPos = s0;
      s1 = peg$f0();
      s0 = s1;
    }

    return s0;
  }

  function peg$parsenormalSequence() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = [];
    s2 = peg$parseunit();

    if (s2 !== peg$FAILED) {
      while (s2 !== peg$FAILED) {
        s1.push(s2);
        s2 = peg$parseunit();
      }
    } else {
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      peg$savedPos = s0;
      s1 = peg$f1(s1);
    }

    s0 = s1;
    return s0;
  }

  function peg$parseunit() {
    var s0;
    s0 = peg$parseisSeparatorLiteral();

    if (s0 === peg$FAILED) {
      s0 = peg$parsecellSequence();

      if (s0 === peg$FAILED) {
        s0 = peg$parsechar();
      }
    }

    return s0;
  }

  function peg$parsechar() {
    var s0;
    s0 = peg$parseescapedChar();

    if (s0 === peg$FAILED) {
      if (input.length > peg$currPos) {
        s0 = input.charAt(peg$currPos);
        peg$currPos++;
      } else {
        s0 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e0);
        }
      }
    }

    return s0;
  }

  function peg$parseisSeparatorLiteral() {
    var s0, s1, s2;
    s0 = peg$currPos;

    if (peg$r0.test(input.charAt(peg$currPos))) {
      s1 = input.charAt(peg$currPos);
      peg$currPos++;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e1);
      }
    }

    if (s1 !== peg$FAILED) {
      if (input.charCodeAt(peg$currPos) === 32) {
        s2 = peg$c0;
        peg$currPos++;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e2);
        }
      }

      if (s2 === peg$FAILED) {
        s2 = null;
      }

      peg$savedPos = s0;
      s0 = peg$f2(s1, s2);
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parseescape() {
    var s0;

    if (input.charCodeAt(peg$currPos) === 92) {
      s0 = peg$c1;
      peg$currPos++;
    } else {
      s0 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e3);
      }
    }

    return s0;
  }

  function peg$parseescapedChar() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$parseescape();

    if (s1 !== peg$FAILED) {
      if (peg$r0.test(input.charAt(peg$currPos))) {
        s2 = input.charAt(peg$currPos);
        peg$currPos++;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e1);
        }
      }

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f3(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsecellSequence() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = [];
    s2 = peg$parsenotSeparator();

    if (s2 !== peg$FAILED) {
      while (s2 !== peg$FAILED) {
        s1.push(s2);
        s2 = peg$parsenotSeparator();
      }
    } else {
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      peg$savedPos = s0;
      s1 = peg$f4(s1);
    }

    s0 = s1;
    return s0;
  }

  function peg$parsenotSeparator() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;
    s2 = peg$parseisSeparatorLiteral();
    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parsechar();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f5(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  peg$result = peg$startRuleFunction();

  if (peg$result !== peg$FAILED && peg$currPos === input.length) {
    return peg$result;
  } else {
    if (peg$result !== peg$FAILED && peg$currPos < input.length) {
      peg$fail(peg$endExpectation());
    }

    throw peg$buildStructuredError(peg$maxFailExpected, peg$maxFailPos < input.length ? input.charAt(peg$maxFailPos) : null, peg$maxFailPos < input.length ? peg$computeLocation(peg$maxFailPos, peg$maxFailPos + 1) : peg$computeLocation(peg$maxFailPos, peg$maxFailPos));
  }
}

/***/ }),

/***/ "./js/core/texts/text.js":
/*!*******************************!*\
  !*** ./js/core/texts/text.js ***!
  \*******************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.TextLine = exports.Text = void 0;
exports.areTextsEqual = areTextsEqual;
exports.newSingleLineNoSoftEOL = newSingleLineNoSoftEOL;
exports.newSingleLineText = newSingleLineText;
exports.newText = newText;
exports.newTextNoWrap = newTextNoWrap;
exports.splitText = splitText;
exports.testBuildLines = testBuildLines;
exports.wrapDomText = wrapDomText;

var TextGrammar = _interopRequireWildcard(__webpack_require__(/*! core/texts/textGrammar */ "./js/core/texts/textGrammar.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _textconstants = __webpack_require__(/*! core/texts/textconstants */ "./js/core/texts/textconstants.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

//{ TextUtils.flattenTextDom }
const BULLET_LINE_INDENT = "         ";
const BULLET_LINE_PREFIX = "     •   ";

const isWhitespace = function (text) {
  return text.trim() === '';
};

const WordsRE = /[\s]+|\\?[^\s\\]+/g;

const splitInWords = function (text) {
  return text.match(WordsRE);
}; // backslash splits words. it can be only at the beginning of a word.


const LEADING_FACTOR = 1.2;

class TreeBuilder {
  constructor() {
    let maxLength = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 0;
    let textMeter = arguments.length > 1 ? arguments[1] : undefined;
    let textStyle = arguments.length > 2 ? arguments[2] : undefined;
    let truncate = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : false;
    let respectSoftEol = arguments.length > 4 ? arguments[4] : undefined;
    let singleLine = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : false;

    _balsamiq.default.assertNumber(maxLength);

    _balsamiq.default.exists(textMeter);

    _balsamiq.default.exists(textStyle);

    _balsamiq.default.assertBool(respectSoftEol);

    _balsamiq.default.assertBool(truncate);

    _balsamiq.default.assertBool(singleLine);

    this.respectSoftEol = respectSoftEol;
    this.textMeter = textMeter;
    this.maxLength = maxLength; // maxLength == 0 means no wrap

    this.lines = [];
    this.linkData = []; // { target: taken from the textDom, startX: number, endX: number, continueOnNextLine (optional): boolean }

    this.nodes = [{
      children: []
    }];
    this.textStyle = textStyle;
    this.textMeter.initializeMeasurement(this.textStyle);
    this.emptyLine = true;
    let height = 1;

    if (this.textStyle) {
      height = this.textStyle.fontSize;
    }

    this.currentLineSize = _types.default.newTextSize(0, height, 1);
    this.insideBullet = false;
    this.addingBullet = false;
    this.truncate = truncate;
    this.singleLine = singleLine;
  }

  traverseTextDomElements(elements) {
    for (const element of elements) {
      if (element.type === _textconstants.NodeTypes.EOL) {
        this.breakLine();
        this.insideBullet = false;
      } else if (element.type === _textconstants.NodeTypes.BULLET) {
        this.insideBullet = true;
        this.addBullet(BULLET_LINE_PREFIX);
        this.traverseTextDomElements(element.children);
      } else if (typeof element === 'string') {
        if (this.truncate) {
          this.addText(this.tokenize(element));
        } else {
          const words = splitInWords(element);

          if (words) {
            for (const word of words) {
              this.addText(this.tokenize(word));
            }
          }
        }
      } else if (element.type === _textconstants.NodeTypes.SOFT_EOL) {
        if (this.respectSoftEol) {
          this.breakLine();

          if (this.insideBullet) {
            this.addBullet(BULLET_LINE_INDENT);
          }
        } else {
          this.addText(this.tokenize(element.value));
        }

        this.traverseTextDomElements(element.children);
      } else if (element.type === _textconstants.NodeTypes.ESCAPEDCHAR) {
        this.addText(this.tokenize(element.value));
      } else if (element.type === _textconstants.NodeTypes.MENTION) {
        this.addText(this.tokenize(element.value || element.children[0]));
      } else {
        if (element.type === _textconstants.NodeTypes.COLOR) {
          // NOTE from Marco and Paolo: this code should not modify the element,
          // but we don't want to change it right now because
          // it's here since forever
          if (typeof element.color === 'string') {
            element.color = _colors.default.colorFromString(element.color);
          }
        }

        this.openFormat(_lodash.default.omit(element, 'children'));
        this.traverseTextDomElements(element.children);
        this.closeFormat();
      }
    }
  }

  addBullet(text) {
    _balsamiq.default.assert(this.emptyLine, "bullet text must be added at the beginning of a line");

    this.addText(this.tokenize(text));
    this.emptyLine = true;
  } // Text tokenization aimed to deal with surrogate pairs (e.g. '𝛀') and combining marks (e.g. 'cafe\u0301').
  // See https://dmitripavlutin.com/what-every-javascript-developer-should-know-about-unicode/#3-unicode-in-javascript


  tokenize(text) {
    return [...text.normalize()];
  }

  addText(tokens) {
    let start = 0;
    let end = tokens.length;
    let lastLowerEnd = start;
    let lastUpperEnd = end;

    while (end > start) {
      // Iterative Binary Search
      const subStringToTest = tokens.slice(start, end).join('');
      const currentText = this.textMeter.currentText;
      this.textMeter.addText(subStringToTest);
      const sizeToTest = this.textMeter.getSize(); // If the text cannot be wrapped or we still have available space...

      if (this.maxLength === 0 || sizeToTest.width <= this.maxLength) {
        // Can we try to fit a larger chunk of text?
        if (sizeToTest.width < this.maxLength && end < lastUpperEnd) {
          lastLowerEnd = end;
          const largerEnd = Math.ceil((end + lastUpperEnd) / 2);

          if (largerEnd === lastUpperEnd) {
            // No, we reached the end.
            // Add what we know fits.
            this.appendOrMergeTextToCurrentNode(subStringToTest);
            this.currentLineSize = sizeToTest; // Can we go further?

            if (this.truncate || this.singleLine || end === tokens.length || isWhitespace(tokens.slice(end, tokens.length).join(''))) return; // No.

            start = end;
            end = tokens.length;
            lastLowerEnd = start;
            lastUpperEnd = end;
            this.breakLine();

            if (this.insideBullet) {
              this.addBullet(BULLET_LINE_INDENT);
              continue;
            }
          } else {
            end = largerEnd; // Yes, let's try to fit a larger chunk of text.
          }
        } else {
          // If we reached the end of the string and we cannot truncate, we leave.
          if (!this.emptyLine && !this.truncate && end < tokens.length) return; // We cannot fit a larger chunk of text, there we add what we have.

          this.appendOrMergeTextToCurrentNode(subStringToTest);
          this.currentLineSize = sizeToTest; // Can we go further?

          if (this.truncate || this.singleLine || end === tokens.length || isWhitespace(tokens.slice(end, tokens.length).join(''))) return; // No.

          start = end;
          end = tokens.length;
          lastLowerEnd = start;
          lastUpperEnd = end;
          this.breakLine();

          if (this.insideBullet) {
            this.addBullet(BULLET_LINE_INDENT);
            continue;
          }
        }
      } else {
        // Text can be wrapped. Is current chunk is too big?
        if (sizeToTest.width > this.maxLength) {
          // If we can try a smaller chunk...
          if (end > start) {
            // If current line is not empty and we want non-truncated multi-lines...
            if (!this.emptyLine && !this.truncate && !this.singleLine) {
              // Does it make sense to go further?
              if (isWhitespace(tokens.slice(start, end).join(''))) return; // No.
              // Let's make a new line.

              this.breakLine();

              if (this.insideBullet) {
                this.addBullet(BULLET_LINE_INDENT);
                continue;
              }
            } else {
              // Current line is empty, thus we need to try to fit a smaller chunk.
              lastUpperEnd = end;
              end = Math.floor((lastLowerEnd + end) / 2); // Are we sure this is a legit smaller chunk?

              if (end === start) {
                // Hell no, we already tried the smallest chunk.
                // Can we go further?
                if (this.truncate || this.singleLine || end === tokens.length || isWhitespace(tokens.slice(end, tokens.length).join(''))) return; // No.

                start = end + 1;
                end = tokens.length;
                lastLowerEnd = start;
                lastUpperEnd = end;
                this.breakLine();

                if (this.insideBullet) {
                  this.addBullet(BULLET_LINE_INDENT);
                  continue;
                }
              }
            }
          } else if (this.emptyLine) {
            // Text can be wrapped, current line is empty and we cannot have a smaller chunk, thus let's add what we have.
            this.appendOrMergeTextToCurrentNode(subStringToTest);
            this.currentLineSize = sizeToTest;
          } else {
            return;
          }
        } else {
          return;
        }
      }

      this.textMeter.currentText = this.emptyLine && !this.insideBullet ? '' : currentText;
    }
  }

  addTextToMeasure(text) {
    this.textMeter.addText(text);
    return this.textMeter.getSize();
  }

  appendOrMergeTextToCurrentNode(text) {
    const children = this.nodes[this.nodes.length - 1].children;

    if (children.length > 0 && typeof children[children.length - 1] === 'string') {
      children[children.length - 1] += text;
    } else {
      children.push(text);
    }

    this.emptyLine = false;
  }

  breakLine() {
    // If a link is open, then close it
    this.linkData.forEach(linkData => {
      if (!linkData.endX) {
        linkData.endX = this.currentLineSize.width;
        linkData.continueOnNextLine = true;
      }
    });
    this.lines.push({
      textDom: this.nodes[0].children,
      size: this.currentLineSize,
      links: this.linkData
    });
    this.linkData = [];
    this.textMeter.terminateMeasurement();
    this.textMeter.initializeMeasurement(this.textStyle);
    this.currentLineSize = this.textMeter.getSize();
    const nodes = this.nodes;
    this.nodes = [{
      children: []
    }];

    for (const node of nodes.slice(1)) {
      this.openFormat(_lodash.default.omit(node, 'children'));
    } // Strip away the rightmost leaf in the tree if it's an empty format (i.e. children == [])


    if (nodes.length >= 2 && nodes[nodes.length - 1].children.length === 0) {
      nodes[nodes.length - 2].children.pop();
    }

    this.nodes[this.nodes.length - 1].children.push("");
    this.emptyLine = true;
  } // Meant to be called with a newly created format object without "children" (i.e. {type: 'bold'})


  openFormat(format) {
    if (format.type === _textconstants.NodeTypes.LINK) {
      const target = _balsamiq.default.isArray(format.target) ? format.target[0] : ""; //// TODO: this sucks, we should add "resource:" when loading controls
      //if (target.length > 0 && !target.startsWith("resource:") && !target.startsWith("http")) {
      //    target = "resource:" + target;
      //}

      _balsamiq.default.assertString(target);

      this.linkData.push({
        target,
        startX: this.currentLineSize.width
      });
    }

    this.textMeter.openFormat(format);
    this.currentLineSize = this.textMeter.getSize();
    const newNode = Object.assign({}, {
      children: []
    }, format);
    this.nodes[this.nodes.length - 1].children.push(newNode);
    this.nodes.push(newNode);
  }

  closeFormat() {
    var node = this.nodes.pop(); // close the link if open, then move it to the top of the stack, so that the next link is ready to be pop-ed.

    if (node.type === _textconstants.NodeTypes.LINK) {
      const ld = this.linkData.pop();
      ld.endX = this.currentLineSize.width;
      this.linkData.unshift(ld);
    }

    this.textMeter.closeFormat();
  }

  finish() {
    if (this.nodes[0].children.length > 0) {
      if (this.nodes.length > 0 && this.nodes[this.nodes.length - 1].children.length === 0) {
        this.nodes.pop();
      }

      this.lines.push({
        textDom: this.nodes[0].children,
        size: this.currentLineSize,
        links: this.linkData
      });
    }

    this.textMeter.terminateMeasurement();
  }

}

function wrapDomText(sourceDomText, maxWidth, textMeter, p_textStyle) {
  var builder = new TreeBuilder(maxWidth, textMeter, p_textStyle, false);
  builder.traverseTextDomElements(sourceDomText);
  builder.finish();
  return builder.lines;
}

class TextLine {
  constructor(textDom, textSize, textStyle, links) {
    const TextStyleObject = {
      fontFamily: "",
      fontSize: 1,
      textColor: 0x000000,
      bold: false,
      italic: false,
      underline: false
    };

    _balsamiq.default.assert(_balsamiq.default.conformsTo(textStyle, TextStyleObject), "textStyle parameter is not a valid TextStyle");

    this.textDom = textDom;
    this.textSize = textSize;
    this.textStyle = textStyle;
    this.links = links;
    this.attributedText = TextUtils.textDom2AttributedText(textDom, textStyle);
  }

  getTextDisplayItemAt(x, y) {
    return _displayitem.default.newText(this.textDom, this.attributedText, this.textStyle, _types.default.newRectangle(x, y, this.textSize.width, this.textSize.height), this.textSize.ascender);
  }

  getTextDisplayItemAtRight(x, y) {
    return _displayitem.default.newText(this.textDom, this.attributedText, this.textStyle, _types.default.newRectangle(x - this.textSize.width, y, this.textSize.width, this.textSize.height), this.textSize.ascender);
  }

  getTextDisplayItemInBbox(p_bbox, p_align, p_y) {
    let x = 0;

    if (p_align === _types.default.TextAlign.CENTER) {
      x = (p_bbox.width - this.textSize.width) / 2;
    } else if (p_align === _types.default.TextAlign.RIGHT) {
      x = p_bbox.width - this.textSize.width;
    }

    return this.getTextDisplayItemAt(x, p_y);
  }

  getWidth() {
    return this.textSize.width;
  }

  getHeight() {
    return this.textSize.height;
  }

}

exports.TextLine = TextLine;

function ensureCorrectDomText(domTexts, textStyle) {
  return domTexts.map(domText => {
    if (domText.type === _textconstants.NodeTypes.SIZE && domText.size <= 0) {
      domText.size = textStyle.fontSize;
    }

    if (domText.children && domText.children.length > 0) {
      domText.children = ensureCorrectDomText(domText.children, textStyle);
    }

    return domText;
  });
}

class Text {
  constructor(texts, textStyle, respectSoftEol, truncate, context, parseBullet) {
    let singleLine = arguments.length > 6 && arguments[6] !== undefined ? arguments[6] : false;

    _balsamiq.default.assertArrayOfStrings(texts);

    this.textStyle = textStyle;
    this.truncate = truncate;
    this.context = context;
    this.lines = texts;
    this.domTexts = texts.map(text => {
      return ensureCorrectDomText(TextGrammar.parse(text, {
        parseBullet: parseBullet
      }), textStyle);
    });
    this.respectSoftEol = respectSoftEol;
    this.cachedSizes = {};
    this.cachedTextLines = {};
    this.singleLine = singleLine;
  }

  measureTextLines(p_textLines) {
    // calc width and height
    var width = 0;
    var height = 0;
    p_textLines.forEach((textLine, index) => {
      let lineHeight = 0;

      if (textLine) {
        width = Math.max(width, textLine.getWidth());
        lineHeight = textLine.getHeight();
      } else {
        lineHeight = this.textStyle.fontSize;
      }

      height += lineHeight;

      if (index !== p_textLines.length - 1) {
        height += lineHeight * (LEADING_FACTOR - 1);
      }
    });
    return _types.default.newSize(Math.round(width), Math.round(height));
  }

  measuredSizeForWidth(width) {
    if (!this.cachedSizes[width]) {
      this.makeLinesForWidth(width);
    }

    return this.cachedSizes[width];
  }

  measuredSize() {
    return this.measuredSizeForWidth(0);
  }

  makeLinesForWidth(p_width) {
    let textLines = [];

    for (const domText of this.domTexts) {
      const lines = buildLines(domText, this.textStyle, p_width, this.truncate, this.context.textMeter, this.respectSoftEol, this.singleLine); // In case of empty line, the buildLines function returns an empty array. Then we insert an empty
      // TextLine (the null value) so that the vertical space of the empty line is occupied.

      if (lines.length === 0) {
        textLines.push(null);
      } else {
        textLines = textLines.concat(lines);
      }
    }

    this.cachedTextLines[p_width] = textLines;
    this.cachedSizes[p_width] = this.measureTextLines(textLines);
  }

  makeDisplayItemsAndLinks(p_bbox, p_halign, p_vposition) {
    // p_halign == undefined -> default Left
    // p_vposition == undefined -> p_vposition = default top
    p_halign = _balsamiq.default.isUndefined(p_halign) ? _types.default.TextAlign.LEFT : p_halign;
    p_vposition = _balsamiq.default.isUndefined(p_vposition) ? _types.default.Positions.TOP : p_vposition;

    _balsamiq.default.assertEnum(p_halign, _types.default.TextAlign);

    _balsamiq.default.assertEnum(p_vposition, _types.default.Positions);

    var textLines;
    var textSize;

    if (this.cachedSizes[0]) {
      if (p_bbox.width >= this.cachedSizes[0].width) {
        textLines = this.cachedTextLines[0];
        textSize = this.cachedSizes[0];
      }
    }

    if (!textLines && this.cachedTextLines[p_bbox.width]) {
      textLines = this.cachedTextLines[p_bbox.width];
      textSize = this.cachedSizes[p_bbox.width];
    }

    if (!textLines) {
      this.makeLinesForWidth(p_bbox.width);
      textLines = this.cachedTextLines[p_bbox.width];
      textSize = this.cachedSizes[p_bbox.width];
    }

    var y = 0;

    if (p_vposition === _types.default.Positions.TOP) {
      y = p_bbox.y;
    } else if (p_vposition === _types.default.Positions.CENTER) {
      y = p_bbox.y + (p_bbox.height - textSize.height) / 2;
    } else {
      y = p_bbox.y + p_bbox.height - textSize.height;
    }

    var items = [],
        links = [];
    textLines.forEach(textLine => {
      if (textLine) {
        if (y + textLine.getHeight() >= p_bbox.y && y < p_bbox.y + p_bbox.height) {
          var x;

          if (p_halign === _types.default.TextAlign.LEFT) {
            x = p_bbox.x;
          } else if (p_halign === _types.default.TextAlign.CENTER) {
            x = p_bbox.x + (p_bbox.width - textLine.getWidth()) / 2;
          } else {
            x = p_bbox.x + p_bbox.width - textLine.getWidth();
          }

          items.push(textLine.getTextDisplayItemAt(x, y));
          textLine.links.forEach(link => {
            links.push({
              target: link.target,
              bounds: _types.default.newRectangle(x + link.startX, y, link.endX - link.startX, textLine.getHeight()),
              continueOnNextLine: _balsamiq.default.isBoolean(link.continueOnNextLine) ? link.continueOnNextLine : false
            });
          });
        }
      }

      let lineHeight = textLine ? textLine.getHeight() : this.textStyle.fontSize;
      y += lineHeight * LEADING_FACTOR;
    });
    return {
      displayItems: items,
      links: links
    };
  }

  makeDisplayItems(p_bbox, p_halign, p_vposition) {
    return this.makeDisplayItemsAndLinks(p_bbox, p_halign, p_vposition).displayItems;
  }

  getLinks() {
    var links = [];
    this.lines.forEach(line => {
      var domText = TextUtils.stringToNodes(line);

      var collectLinks = nodes => {
        nodes.forEach(node => {
          if (!_balsamiq.default.isString(node)) {
            if (node.type == _textconstants.NodeTypes.LINK) {
              links.push({
                target: node.target ? TextUtils.unformatNodes(node.target) : "",
                label: TextUtils.unformatNodes(node.children)
              });
            }

            if (node.children) {
              collectLinks(node.children);
            }
          }
        });
      };

      collectLinks(domText);
    }); //console.log("text links: " + JSON.stringify(links));

    return links;
  }

}

exports.Text = Text;

function buildLines(domText, textStyle, maxWidth, truncate, textMeter, respectSoftEol, singleLine) {
  _balsamiq.default.assertObject(domText);

  _balsamiq.default.assertBool(respectSoftEol);

  _balsamiq.default.assertBool(singleLine); // maxWidth == 0 -> never wrap
  // truncate == true -> if maxWidth > 0 then create only first line


  const treeBuilder = new TreeBuilder(maxWidth, textMeter, textStyle, truncate, respectSoftEol, singleLine);
  treeBuilder.traverseTextDomElements(domText);
  treeBuilder.finish();
  return treeBuilder.lines.map(line => new TextLine(line.textDom, line.size, textStyle, line.links));
}

function newText(text, textStyle, context, parseBullet) {
  _balsamiq.default.assertString(text);

  const splitText = this.splitText(text);
  return new Text(splitText, textStyle, true, false, context, parseBullet);
}

function newTextNoWrap(text, textStyle, context, parseBullet) {
  _balsamiq.default.assertString(text);

  return new Text([text], textStyle, true, true, context, parseBullet);
}

function newSingleLineNoSoftEOL(text, textStyle, context) {
  _balsamiq.default.assertString(text);

  return new Text([text], textStyle, false, true, context, null, true);
}

function newSingleLineText(text, textStyle, context) {
  _balsamiq.default.assertString(text);

  return new Text([text], textStyle, true, true, context, null, true);
}

function areTextsEqual(textObject, text, textStyle) {
  const joined = textObject.lines.join("\n");
  return joined === text && _balsamiq.default.isEqual(textObject.textStyle, textStyle);
}

function splitText(p_text) {
  return p_text.split(/[\n\u2028\u2029]/g);
}

function testBuildLines(text, maxLength, textMeter, textStyle, truncate, useSoftEol, isSingleLine) {
  const treeBuilder = new TreeBuilder(maxLength, textMeter, textStyle, truncate, useSoftEol, isSingleLine);
  treeBuilder.traverseTextDomElements(text);
  treeBuilder.finish();
  return treeBuilder.lines.map(line => line.textDom[0]);
}

/***/ }),

/***/ "./js/core/texts/textDomUtils.js":
/*!***************************************!*\
  !*** ./js/core/texts/textDomUtils.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.flatten = flatten;
exports.joinStringUnits = joinStringUnits;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Dependencies of non-AMD compliant modules must also be non-AMD compliant (i.e. populate a global variable)
// This module is a dependency of the PEG auto-generated grammar, which is not AMD-compliant, therefore it also must
// define a global variable
//noinspection JSHint, JSLint
function flatten(arr) {
  var res = [];
  arr.forEach(el => {
    if (Array.isArray(el)) {
      res = res.concat(flatten(el));
    } else {
      res.push(el);
    }
  });
  return res;
}

;

function joinStringUnits(elements) {
  var result = []; // Flatten

  elements = flatten(elements); // Join adjacent strings

  elements.forEach(function (el) {
    if (result.length > 0 && _balsamiq.default.isString(el) && _balsamiq.default.isString(result[result.length - 1])) {
      result[result.length - 1] = result[result.length - 1] + el;
    } else {
      result.push(el);
    }
  });
  return result;
}

;

/***/ }),

/***/ "./js/core/texts/textGrammar.js":
/*!**************************************!*\
  !*** ./js/core/texts/textGrammar.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.SyntaxError = peg$SyntaxError;
exports.parse = peg$parse;

var balsamiqTextDomUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textDomUtils */ "./js/core/texts/textDomUtils.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

/* eslint-disable */
// Generated by Peggy 1.2.0.
//
// https://peggyjs.org/
function peg$subclass(child, parent) {
  function C() {
    this.constructor = child;
  }

  C.prototype = parent.prototype;
  child.prototype = new C();
}

function peg$SyntaxError(message, expected, found, location) {
  var self = Error.call(this, message);

  if (Object.setPrototypeOf) {
    Object.setPrototypeOf(self, peg$SyntaxError.prototype);
  }

  self.expected = expected;
  self.found = found;
  self.location = location;
  self.name = "SyntaxError";
  return self;
}

peg$subclass(peg$SyntaxError, Error);

function peg$padEnd(str, targetLength, padString) {
  padString = padString || " ";

  if (str.length > targetLength) {
    return str;
  }

  targetLength -= str.length;
  padString += padString.repeat(targetLength);
  return str + padString.slice(0, targetLength);
}

peg$SyntaxError.prototype.format = function (sources) {
  var str = "Error: " + this.message;

  if (this.location) {
    var src = null;
    var k;

    for (k = 0; k < sources.length; k++) {
      if (sources[k].source === this.location.source) {
        src = sources[k].text.split(/\r\n|\n|\r/g);
        break;
      }
    }

    var s = this.location.start;
    var loc = this.location.source + ":" + s.line + ":" + s.column;

    if (src) {
      var e = this.location.end;
      var filler = peg$padEnd("", s.line.toString().length);
      var line = src[s.line - 1];
      var last = s.line === e.line ? e.column : line.length + 1;
      str += "\n --> " + loc + "\n" + filler + " |\n" + s.line + " | " + line + "\n" + filler + " | " + peg$padEnd("", s.column - 1) + peg$padEnd("", last - s.column, "^");
    } else {
      str += "\n at " + loc;
    }
  }

  return str;
};

peg$SyntaxError.buildMessage = function (expected, found) {
  var DESCRIBE_EXPECTATION_FNS = {
    literal: function (expectation) {
      return "\"" + literalEscape(expectation.text) + "\"";
    },
    class: function (expectation) {
      var escapedParts = expectation.parts.map(function (part) {
        return Array.isArray(part) ? classEscape(part[0]) + "-" + classEscape(part[1]) : classEscape(part);
      });
      return "[" + (expectation.inverted ? "^" : "") + escapedParts + "]";
    },
    any: function () {
      return "any character";
    },
    end: function () {
      return "end of input";
    },
    other: function (expectation) {
      return expectation.description;
    }
  };

  function hex(ch) {
    return ch.charCodeAt(0).toString(16).toUpperCase();
  }

  function literalEscape(s) {
    return s.replace(/\\/g, "\\\\").replace(/"/g, "\\\"").replace(/\0/g, "\\0").replace(/\t/g, "\\t").replace(/\n/g, "\\n").replace(/\r/g, "\\r").replace(/[\x00-\x0F]/g, function (ch) {
      return "\\x0" + hex(ch);
    }).replace(/[\x10-\x1F\x7F-\x9F]/g, function (ch) {
      return "\\x" + hex(ch);
    });
  }

  function classEscape(s) {
    return s.replace(/\\/g, "\\\\").replace(/\]/g, "\\]").replace(/\^/g, "\\^").replace(/-/g, "\\-").replace(/\0/g, "\\0").replace(/\t/g, "\\t").replace(/\n/g, "\\n").replace(/\r/g, "\\r").replace(/[\x00-\x0F]/g, function (ch) {
      return "\\x0" + hex(ch);
    }).replace(/[\x10-\x1F\x7F-\x9F]/g, function (ch) {
      return "\\x" + hex(ch);
    });
  }

  function describeExpectation(expectation) {
    return DESCRIBE_EXPECTATION_FNS[expectation.type](expectation);
  }

  function describeExpected(expected) {
    var descriptions = expected.map(describeExpectation);
    var i, j;
    descriptions.sort();

    if (descriptions.length > 0) {
      for (i = 1, j = 1; i < descriptions.length; i++) {
        if (descriptions[i - 1] !== descriptions[i]) {
          descriptions[j] = descriptions[i];
          j++;
        }
      }

      descriptions.length = j;
    }

    switch (descriptions.length) {
      case 1:
        return descriptions[0];

      case 2:
        return descriptions[0] + " or " + descriptions[1];

      default:
        return descriptions.slice(0, -1).join(", ") + ", or " + descriptions[descriptions.length - 1];
    }
  }

  function describeFound(found) {
    return found ? "\"" + literalEscape(found) + "\"" : "end of input";
  }

  return "Expected " + describeExpected(expected) + " but " + describeFound(found) + " found.";
};

function peg$parse(input, options) {
  options = options !== undefined ? options : {};
  var peg$FAILED = {};
  var peg$source = options.grammarSource;
  var peg$startRuleFunctions = {
    normalSequenceOrEmptySequence: peg$parsenormalSequenceOrEmptySequence
  };
  var peg$startRuleFunction = peg$parsenormalSequenceOrEmptySequence;
  var peg$c0 = "*";
  var peg$c1 = "-";
  var peg$c2 = "&";
  var peg$c3 = "_";
  var peg$c4 = "~";
  var peg$c5 = "[";
  var peg$c6 = "]";
  var peg$c7 = "{color}";
  var peg$c8 = "{color:";
  var peg$c9 = "{colour:";
  var peg$c10 = "{colour}";
  var peg$c11 = "{size}";
  var peg$c12 = "{size:";
  var peg$c13 = "(";
  var peg$c14 = ")";
  var peg$c15 = " ";
  var peg$c16 = "\t";
  var peg$c17 = "\\";
  var peg$c18 = "\n";
  var peg$c19 = "\r";
  var peg$c20 = "\\n";
  var peg$c21 = "\\r";
  var peg$c22 = "* ";
  var peg$c23 = "- ";
  var peg$c24 = "](";
  var peg$c25 = "#";
  var peg$c26 = "}";
  var peg$c27 = "{";
  var peg$c28 = "color";
  var peg$c29 = "{color";
  var peg$c30 = "colour";
  var peg$c31 = "{colour";
  var peg$c32 = "{size";
  var peg$c33 = "@[";
  var peg$r0 = /^[0-9]/;
  var peg$r1 = /^[\\&~*_\-[\]]/;
  var peg$r2 = /^[a-zA-Z0-9]/;
  var peg$r3 = /^[ a-zA-Z0-9@=&?.:\/#,\-_+%'()]/;
  var peg$r4 = /^[ a-zA-Z0-9@=&?.:\/#,\-_+%']/;
  var peg$e0 = peg$anyExpectation();
  var peg$e1 = peg$classExpectation([["0", "9"]], false, false);
  var peg$e2 = peg$literalExpectation("*", false);
  var peg$e3 = peg$literalExpectation("-", false);
  var peg$e4 = peg$literalExpectation("&", false);
  var peg$e5 = peg$literalExpectation("_", false);
  var peg$e6 = peg$literalExpectation("~", false);
  var peg$e7 = peg$literalExpectation("[", false);
  var peg$e8 = peg$literalExpectation("]", false);
  var peg$e9 = peg$literalExpectation("{color}", true);
  var peg$e10 = peg$literalExpectation("{color:", true);
  var peg$e11 = peg$literalExpectation("{colour:", true);
  var peg$e12 = peg$literalExpectation("{colour}", true);
  var peg$e13 = peg$literalExpectation("{size}", true);
  var peg$e14 = peg$literalExpectation("{size:", true);
  var peg$e15 = peg$literalExpectation("(", false);
  var peg$e16 = peg$literalExpectation(")", false);
  var peg$e17 = peg$literalExpectation(" ", false);
  var peg$e18 = peg$literalExpectation("\t", false);
  var peg$e19 = peg$literalExpectation("\\", false);
  var peg$e20 = peg$literalExpectation("\n", false);
  var peg$e21 = peg$literalExpectation("\r", false);
  var peg$e22 = peg$literalExpectation("\\n", false);
  var peg$e23 = peg$literalExpectation("\\r", false);
  var peg$e24 = peg$classExpectation(["\\", "&", "~", "*", "_", "-", "[", "]"], false, false);
  var peg$e25 = peg$literalExpectation("* ", false);
  var peg$e26 = peg$literalExpectation("- ", false);
  var peg$e27 = peg$literalExpectation("](", false);
  var peg$e28 = peg$literalExpectation("#", false);
  var peg$e29 = peg$classExpectation([["a", "z"], ["A", "Z"], ["0", "9"]], false, false);
  var peg$e30 = peg$literalExpectation("}", false);
  var peg$e31 = peg$literalExpectation("{", false);
  var peg$e32 = peg$literalExpectation("color", true);
  var peg$e33 = peg$literalExpectation("{color", true);
  var peg$e34 = peg$literalExpectation("colour", true);
  var peg$e35 = peg$literalExpectation("{colour", true);
  var peg$e36 = peg$literalExpectation("{size", true);
  var peg$e37 = peg$literalExpectation("@[", false);
  var peg$e38 = peg$classExpectation([" ", ["a", "z"], ["A", "Z"], ["0", "9"], "@", "=", "&", "?", ".", ":", "/", "#", ",", "-", "_", "+", "%", "'", "(", ")"], false, false);
  var peg$e39 = peg$classExpectation([" ", ["a", "z"], ["A", "Z"], ["0", "9"], "@", "=", "&", "?", ".", ":", "/", "#", ",", "-", "_", "+", "%", "'"], false, false);

  var peg$f0 = function () {
    return [];
  };

  var peg$f1 = function (u) {
    return balsamiqTextDomUtils.joinStringUnits(u);
  };

  var peg$f2 = function () {
    return location().start.column == 1;
  };

  var peg$f3 = function (w, inner) {
    return [w ? w : "", inner];
  };

  var peg$f4 = function () {
    return {
      type: 'eol'
    };
  };

  var peg$f5 = function (m, start, end) {
    return {
      type: 'softEol',
      value: m,
      children: balsamiqTextDomUtils.joinStringUnits([start ? start : "", end])
    };
  };

  var peg$f6 = function (u) {
    return u;
  };

  var peg$f7 = function (ch) {
    return {
      type: 'escapedChar',
      value: ch
    };
  };

  var peg$f8 = function () {
    return location().start.column == 1;
  };

  var peg$f9 = function () {
    return options.parseBullet;
  };

  var peg$f10 = function (v, internalStartUnit, internalNotBulletUnit) {
    return {
      type: 'bullet',
      value: v,
      children: balsamiqTextDomUtils.joinStringUnits([internalStartUnit ? internalStartUnit : "", internalNotBulletUnit])
    };
  };

  var peg$f11 = function (internalStartUnit, internalUnits) {
    return {
      type: 'bold',
      children: balsamiqTextDomUtils.joinStringUnits([internalStartUnit, internalUnits])
    };
  };

  var peg$f12 = function (internalStartUnit, internalUnits) {
    return {
      type: 'italic',
      children: balsamiqTextDomUtils.joinStringUnits([internalStartUnit, internalUnits])
    };
  };

  var peg$f13 = function (internalStartUnit, internalUnits) {
    return {
      type: 'disabled',
      children: balsamiqTextDomUtils.joinStringUnits([internalStartUnit, internalUnits])
    };
  };

  var peg$f14 = function (inner) {
    return ["", inner];
  };

  var peg$f15 = function (internalStartUnit, internalUnits) {
    return {
      type: 'underline',
      children: balsamiqTextDomUtils.joinStringUnits([internalStartUnit, internalUnits])
    };
  };

  var peg$f16 = function (internalStartUnit, internalUnits) {
    return {
      type: 'strikethrough',
      children: balsamiqTextDomUtils.joinStringUnits([internalStartUnit, internalUnits])
    };
  };

  var peg$f17 = function (linkStartText, linkText, linkTarget) {
    var type = "link";
    var target = balsamiqTextDomUtils.joinStringUnits(linkTarget);
    type = target[0].substring(0, 12) == "guestaction:" ? "actionLink" : "link";
    return {
      type: type,
      children: balsamiqTextDomUtils.joinStringUnits([linkStartText, linkText]),
      target: target
    };
  };

  var peg$f18 = function (w, inner) {
    return [w ? w : "", inner];
  };

  var peg$f19 = function (internalStartUnit, internalUnits) {
    return {
      type: 'link',
      children: balsamiqTextDomUtils.joinStringUnits([internalStartUnit, internalUnits])
    };
  };

  var peg$f20 = function (hasSpace, hashtag, color, internalStartUnit, internalUnits, v) {
    return {
      type: 'color',
      value: v,
      color: color.join(''),
      hasSpace: !!hasSpace,
      hashtag: !!hashtag,
      children: balsamiqTextDomUtils.joinStringUnits([internalStartUnit, internalUnits])
    };
  };

  var peg$f21 = function (hasSpace, hashtag, color, internalStartUnit, internalUnits, v) {
    return {
      type: 'color',
      value: v,
      color: color.join(''),
      hasSpace: !!hasSpace,
      hashtag: !!hashtag,
      children: balsamiqTextDomUtils.joinStringUnits([internalStartUnit, internalUnits])
    };
  };

  var peg$f22 = function (hasSpace, size) {
    return {
      type: 'size',
      size: parseInt(size.join('')),
      hasSpace: !!hasSpace,
      children: balsamiqTextDomUtils.joinStringUnits([""])
    };
  };

  var peg$f23 = function (hasSpace, size, internalStartUnit, internalUnits) {
    return {
      type: 'size',
      size: parseInt(size.join('')),
      hasSpace: !!hasSpace,
      children: balsamiqTextDomUtils.joinStringUnits([internalStartUnit, internalUnits])
    };
  };

  var peg$f24 = function (mentionUser, target) {
    return {
      type: 'mention',
      children: balsamiqTextDomUtils.joinStringUnits(mentionUser),
      target: target
    };
  };

  var peg$f25 = function (target) {
    return target && balsamiqTextDomUtils.joinStringUnits(target);
  };

  var peg$currPos = 0;
  var peg$savedPos = 0;
  var peg$posDetailsCache = [{
    line: 1,
    column: 1
  }];
  var peg$maxFailPos = 0;
  var peg$maxFailExpected = [];
  var peg$silentFails = 0;
  var peg$result;

  if ("startRule" in options) {
    if (!(options.startRule in peg$startRuleFunctions)) {
      throw new Error("Can't start parsing from rule \"" + options.startRule + "\".");
    }

    peg$startRuleFunction = peg$startRuleFunctions[options.startRule];
  }

  function text() {
    return input.substring(peg$savedPos, peg$currPos);
  }

  function offset() {
    return peg$savedPos;
  }

  function range() {
    return {
      source: peg$source,
      start: peg$savedPos,
      end: peg$currPos
    };
  }

  function location() {
    return peg$computeLocation(peg$savedPos, peg$currPos);
  }

  function expected(description, location) {
    location = location !== undefined ? location : peg$computeLocation(peg$savedPos, peg$currPos);
    throw peg$buildStructuredError([peg$otherExpectation(description)], input.substring(peg$savedPos, peg$currPos), location);
  }

  function error(message, location) {
    location = location !== undefined ? location : peg$computeLocation(peg$savedPos, peg$currPos);
    throw peg$buildSimpleError(message, location);
  }

  function peg$literalExpectation(text, ignoreCase) {
    return {
      type: "literal",
      text: text,
      ignoreCase: ignoreCase
    };
  }

  function peg$classExpectation(parts, inverted, ignoreCase) {
    return {
      type: "class",
      parts: parts,
      inverted: inverted,
      ignoreCase: ignoreCase
    };
  }

  function peg$anyExpectation() {
    return {
      type: "any"
    };
  }

  function peg$endExpectation() {
    return {
      type: "end"
    };
  }

  function peg$otherExpectation(description) {
    return {
      type: "other",
      description: description
    };
  }

  function peg$computePosDetails(pos) {
    var details = peg$posDetailsCache[pos];
    var p;

    if (details) {
      return details;
    } else {
      p = pos - 1;

      while (!peg$posDetailsCache[p]) {
        p--;
      }

      details = peg$posDetailsCache[p];
      details = {
        line: details.line,
        column: details.column
      };

      while (p < pos) {
        if (input.charCodeAt(p) === 10) {
          details.line++;
          details.column = 1;
        } else {
          details.column++;
        }

        p++;
      }

      peg$posDetailsCache[pos] = details;
      return details;
    }
  }

  function peg$computeLocation(startPos, endPos) {
    var startPosDetails = peg$computePosDetails(startPos);
    var endPosDetails = peg$computePosDetails(endPos);
    return {
      source: peg$source,
      start: {
        offset: startPos,
        line: startPosDetails.line,
        column: startPosDetails.column
      },
      end: {
        offset: endPos,
        line: endPosDetails.line,
        column: endPosDetails.column
      }
    };
  }

  function peg$fail(expected) {
    if (peg$currPos < peg$maxFailPos) {
      return;
    }

    if (peg$currPos > peg$maxFailPos) {
      peg$maxFailPos = peg$currPos;
      peg$maxFailExpected = [];
    }

    peg$maxFailExpected.push(expected);
  }

  function peg$buildSimpleError(message, location) {
    return new peg$SyntaxError(message, null, null, location);
  }

  function peg$buildStructuredError(expected, found, location) {
    return new peg$SyntaxError(peg$SyntaxError.buildMessage(expected, found), expected, found, location);
  }

  function peg$parsenormalSequenceOrEmptySequence() {
    var s0, s1;
    s0 = peg$parsenormalSequence();

    if (s0 === peg$FAILED) {
      s0 = peg$currPos;
      s1 = '';
      peg$savedPos = s0;
      s1 = peg$f0();
      s0 = s1;
    }

    return s0;
  }

  function peg$parsenormalSequence() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = [];
    s2 = peg$parseunit();

    if (s2 !== peg$FAILED) {
      while (s2 !== peg$FAILED) {
        s1.push(s2);
        s2 = peg$parseunit();
      }
    } else {
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      peg$savedPos = s0;
      s1 = peg$f1(s1);
    }

    s0 = s1;
    return s0;
  }

  function peg$parseunit() {
    var s0;
    s0 = peg$parsesoftEndOfLine();

    if (s0 === peg$FAILED) {
      s0 = peg$parsebullet();

      if (s0 === peg$FAILED) {
        s0 = peg$parseendOfLine();

        if (s0 === peg$FAILED) {
          s0 = peg$parseescapedChar();

          if (s0 === peg$FAILED) {
            s0 = peg$parsestylesThatMayNotStartInText();

            if (s0 === peg$FAILED) {
              s0 = peg$parseunderlineSequence();

              if (s0 === peg$FAILED) {
                s0 = peg$parseurlLinkSequence();

                if (s0 === peg$FAILED) {
                  s0 = peg$parselinkSequence();

                  if (s0 === peg$FAILED) {
                    s0 = peg$parsecolorSequence();

                    if (s0 === peg$FAILED) {
                      s0 = peg$parsecolourSequence();

                      if (s0 === peg$FAILED) {
                        s0 = peg$parsesizeSequenceNoText();

                        if (s0 === peg$FAILED) {
                          s0 = peg$parsesizeSequence();

                          if (s0 === peg$FAILED) {
                            s0 = peg$parsementionSequence();

                            if (s0 === peg$FAILED) {
                              s0 = peg$parsechar();
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }

    return s0;
  }

  function peg$parsestartUnit() {
    var s0;
    s0 = peg$parsesoftEndOfLine();

    if (s0 === peg$FAILED) {
      s0 = peg$parsebullet();

      if (s0 === peg$FAILED) {
        s0 = peg$parseendOfLine();

        if (s0 === peg$FAILED) {
          s0 = peg$parseescapedChar();

          if (s0 === peg$FAILED) {
            s0 = peg$parseinnerBoldSequence();

            if (s0 === peg$FAILED) {
              s0 = peg$parseinnerItalicSequence();

              if (s0 === peg$FAILED) {
                s0 = peg$parseinnerDisabledSequence();

                if (s0 === peg$FAILED) {
                  s0 = peg$parseinnerUnderlineSequence();

                  if (s0 === peg$FAILED) {
                    s0 = peg$parseinnerStrikethroughSequence();

                    if (s0 === peg$FAILED) {
                      s0 = peg$parseinnerUrlLinkSequence();

                      if (s0 === peg$FAILED) {
                        s0 = peg$parseinnerLinkSequence();

                        if (s0 === peg$FAILED) {
                          s0 = peg$parsecolorSequence();

                          if (s0 === peg$FAILED) {
                            s0 = peg$parsecolourSequence();

                            if (s0 === peg$FAILED) {
                              s0 = peg$parsesizeSequenceNoText();

                              if (s0 === peg$FAILED) {
                                s0 = peg$parsesizeSequence();

                                if (s0 === peg$FAILED) {
                                  s0 = peg$parsechar();
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }

    return s0;
  }

  function peg$parsestartLinkUnit() {
    var s0;
    s0 = peg$parsesoftEndOfLine();

    if (s0 === peg$FAILED) {
      s0 = peg$parsebullet();

      if (s0 === peg$FAILED) {
        s0 = peg$parseendOfLine();

        if (s0 === peg$FAILED) {
          s0 = peg$parseescapedChar();

          if (s0 === peg$FAILED) {
            s0 = peg$parseinnerBoldSequence();

            if (s0 === peg$FAILED) {
              s0 = peg$parseinnerItalicSequence();

              if (s0 === peg$FAILED) {
                s0 = peg$parseinnerDisabledSequence();

                if (s0 === peg$FAILED) {
                  s0 = peg$parseinnerUnderlineSequence();

                  if (s0 === peg$FAILED) {
                    s0 = peg$parseinnerStrikethroughSequence();

                    if (s0 === peg$FAILED) {
                      s0 = peg$parsecolorSequence();

                      if (s0 === peg$FAILED) {
                        s0 = peg$parsecolourSequence();

                        if (s0 === peg$FAILED) {
                          s0 = peg$parsesizeSequenceNoText();

                          if (s0 === peg$FAILED) {
                            s0 = peg$parsesizeSequence();

                            if (s0 === peg$FAILED) {
                              s0 = peg$parsechar();
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }

    return s0;
  }

  function peg$parselinkUnit() {
    var s0;
    s0 = peg$parsesoftEndOfLine();

    if (s0 === peg$FAILED) {
      s0 = peg$parsebullet();

      if (s0 === peg$FAILED) {
        s0 = peg$parseendOfLine();

        if (s0 === peg$FAILED) {
          s0 = peg$parseescapedChar();

          if (s0 === peg$FAILED) {
            s0 = peg$parsestylesThatMayNotStartInText();

            if (s0 === peg$FAILED) {
              s0 = peg$parseunderlineSequence();

              if (s0 === peg$FAILED) {
                s0 = peg$parseurlLinkSequence();

                if (s0 === peg$FAILED) {
                  s0 = peg$parsecolorSequence();

                  if (s0 === peg$FAILED) {
                    s0 = peg$parsecolourSequence();

                    if (s0 === peg$FAILED) {
                      s0 = peg$parsesizeSequenceNoText();

                      if (s0 === peg$FAILED) {
                        s0 = peg$parsesizeSequence();

                        if (s0 === peg$FAILED) {
                          s0 = peg$parsementionSequence();

                          if (s0 === peg$FAILED) {
                            s0 = peg$parsechar();
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }

    return s0;
  }

  function peg$parsechar() {
    var s0;

    if (input.length > peg$currPos) {
      s0 = input.charAt(peg$currPos);
      peg$currPos++;
    } else {
      s0 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e0);
      }
    }

    return s0;
  }

  function peg$parsedigit() {
    var s0;

    if (peg$r0.test(input.charAt(peg$currPos))) {
      s0 = input.charAt(peg$currPos);
      peg$currPos++;
    } else {
      s0 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e1);
      }
    }

    return s0;
  }

  function peg$parseisMarkupStartLiteral() {
    var s0;

    if (input.charCodeAt(peg$currPos) === 42) {
      s0 = peg$c0;
      peg$currPos++;
    } else {
      s0 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e2);
      }
    }

    if (s0 === peg$FAILED) {
      if (input.charCodeAt(peg$currPos) === 45) {
        s0 = peg$c1;
        peg$currPos++;
      } else {
        s0 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e3);
        }
      }

      if (s0 === peg$FAILED) {
        if (input.charCodeAt(peg$currPos) === 38) {
          s0 = peg$c2;
          peg$currPos++;
        } else {
          s0 = peg$FAILED;

          if (peg$silentFails === 0) {
            peg$fail(peg$e4);
          }
        }

        if (s0 === peg$FAILED) {
          if (input.charCodeAt(peg$currPos) === 95) {
            s0 = peg$c3;
            peg$currPos++;
          } else {
            s0 = peg$FAILED;

            if (peg$silentFails === 0) {
              peg$fail(peg$e5);
            }
          }

          if (s0 === peg$FAILED) {
            if (input.charCodeAt(peg$currPos) === 126) {
              s0 = peg$c4;
              peg$currPos++;
            } else {
              s0 = peg$FAILED;

              if (peg$silentFails === 0) {
                peg$fail(peg$e6);
              }
            }

            if (s0 === peg$FAILED) {
              if (input.charCodeAt(peg$currPos) === 91) {
                s0 = peg$c5;
                peg$currPos++;
              } else {
                s0 = peg$FAILED;

                if (peg$silentFails === 0) {
                  peg$fail(peg$e7);
                }
              }

              if (s0 === peg$FAILED) {
                if (input.charCodeAt(peg$currPos) === 93) {
                  s0 = peg$c6;
                  peg$currPos++;
                } else {
                  s0 = peg$FAILED;

                  if (peg$silentFails === 0) {
                    peg$fail(peg$e8);
                  }
                }

                if (s0 === peg$FAILED) {
                  if (input.substr(peg$currPos, 7).toLowerCase() === peg$c7) {
                    s0 = input.substr(peg$currPos, 7);
                    peg$currPos += 7;
                  } else {
                    s0 = peg$FAILED;

                    if (peg$silentFails === 0) {
                      peg$fail(peg$e9);
                    }
                  }

                  if (s0 === peg$FAILED) {
                    if (input.substr(peg$currPos, 7).toLowerCase() === peg$c8) {
                      s0 = input.substr(peg$currPos, 7);
                      peg$currPos += 7;
                    } else {
                      s0 = peg$FAILED;

                      if (peg$silentFails === 0) {
                        peg$fail(peg$e10);
                      }
                    }

                    if (s0 === peg$FAILED) {
                      if (input.substr(peg$currPos, 8).toLowerCase() === peg$c9) {
                        s0 = input.substr(peg$currPos, 8);
                        peg$currPos += 8;
                      } else {
                        s0 = peg$FAILED;

                        if (peg$silentFails === 0) {
                          peg$fail(peg$e11);
                        }
                      }

                      if (s0 === peg$FAILED) {
                        if (input.substr(peg$currPos, 8).toLowerCase() === peg$c10) {
                          s0 = input.substr(peg$currPos, 8);
                          peg$currPos += 8;
                        } else {
                          s0 = peg$FAILED;

                          if (peg$silentFails === 0) {
                            peg$fail(peg$e12);
                          }
                        }

                        if (s0 === peg$FAILED) {
                          if (input.substr(peg$currPos, 6).toLowerCase() === peg$c11) {
                            s0 = input.substr(peg$currPos, 6);
                            peg$currPos += 6;
                          } else {
                            s0 = peg$FAILED;

                            if (peg$silentFails === 0) {
                              peg$fail(peg$e13);
                            }
                          }

                          if (s0 === peg$FAILED) {
                            if (input.substr(peg$currPos, 6).toLowerCase() === peg$c12) {
                              s0 = input.substr(peg$currPos, 6);
                              peg$currPos += 6;
                            } else {
                              s0 = peg$FAILED;

                              if (peg$silentFails === 0) {
                                peg$fail(peg$e14);
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }

    return s0;
  }

  function peg$parsestylesThatMayNotStartInText() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$parseallowedStarterPrefix();

    if (s1 === peg$FAILED) {
      peg$savedPos = peg$currPos;
      s1 = peg$f2();

      if (s1) {
        s1 = undefined;
      } else {
        s1 = peg$FAILED;
      }
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parseinnerItalicSequence();

      if (s2 === peg$FAILED) {
        s2 = peg$parseinnerBoldSequence();

        if (s2 === peg$FAILED) {
          s2 = peg$parseinnerDisabledSequence();

          if (s2 === peg$FAILED) {
            s2 = peg$parseinnerStrikethroughSequence();
          }
        }
      }

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f3(s1, s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parseallowedStarterPrefix() {
    var s0;
    s0 = peg$parsewhiteSpace();

    if (s0 === peg$FAILED) {
      if (input.charCodeAt(peg$currPos) === 40) {
        s0 = peg$c13;
        peg$currPos++;
      } else {
        s0 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e15);
        }
      }

      if (s0 === peg$FAILED) {
        if (input.charCodeAt(peg$currPos) === 41) {
          s0 = peg$c14;
          peg$currPos++;
        } else {
          s0 = peg$FAILED;

          if (peg$silentFails === 0) {
            peg$fail(peg$e16);
          }
        }
      }
    }

    return s0;
  }

  function peg$parsewhiteSpace() {
    var s0;

    if (input.charCodeAt(peg$currPos) === 32) {
      s0 = peg$c15;
      peg$currPos++;
    } else {
      s0 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e17);
      }
    }

    if (s0 === peg$FAILED) {
      if (input.charCodeAt(peg$currPos) === 9) {
        s0 = peg$c16;
        peg$currPos++;
      } else {
        s0 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e18);
        }
      }
    }

    return s0;
  }

  function peg$parsewhiteSpaceOrEmptySequence() {
    var s0, s1;
    s0 = peg$parsewhiteSpace();

    if (s0 === peg$FAILED) {
      s0 = peg$currPos;
      s1 = '';
      peg$savedPos = s0;
      s1 = peg$f0();
      s0 = s1;
    }

    return s0;
  }

  function peg$parseescape() {
    var s0;

    if (input.charCodeAt(peg$currPos) === 92) {
      s0 = peg$c17;
      peg$currPos++;
    } else {
      s0 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e19);
      }
    }

    return s0;
  }

  function peg$parseendOfLine() {
    var s0, s1;
    s0 = peg$currPos;

    if (input.charCodeAt(peg$currPos) === 10) {
      s1 = peg$c18;
      peg$currPos++;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e20);
      }
    }

    if (s1 === peg$FAILED) {
      if (input.charCodeAt(peg$currPos) === 13) {
        s1 = peg$c19;
        peg$currPos++;
      } else {
        s1 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e21);
        }
      }
    }

    if (s1 !== peg$FAILED) {
      peg$savedPos = s0;
      s1 = peg$f4();
    }

    s0 = s1;
    return s0;
  }

  function peg$parsesoftEndOfLine() {
    var s0, s1, s2, s3, s4;
    s0 = peg$currPos;

    if (input.substr(peg$currPos, 2) === peg$c20) {
      s1 = peg$c20;
      peg$currPos += 2;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e22);
      }
    }

    if (s1 === peg$FAILED) {
      if (input.substr(peg$currPos, 2) === peg$c21) {
        s1 = peg$c21;
        peg$currPos += 2;
      } else {
        s1 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e23);
        }
      }
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parsesoftEndOfLineChildren();

      if (s2 === peg$FAILED) {
        s2 = null;
      }

      s3 = [];
      s4 = peg$parsesoftEndOfLineUnit();

      while (s4 !== peg$FAILED) {
        s3.push(s4);
        s4 = peg$parsesoftEndOfLineUnit();
      }

      peg$savedPos = s0;
      s0 = peg$f5(s1, s2, s3);
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsesoftEndOfLineChildren() {
    var s0, s1;
    s0 = peg$currPos;
    s1 = peg$parsestartUnit();

    if (s1 !== peg$FAILED) {
      peg$savedPos = s0;
      s1 = peg$f6(s1);
    }

    s0 = s1;
    return s0;
  }

  function peg$parsesoftEndOfLineUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;
    s2 = peg$parseisMarkupStartLiteral();
    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parseunit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parseescapedChar() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$parseescape();

    if (s1 !== peg$FAILED) {
      if (peg$r1.test(input.charAt(peg$currPos))) {
        s2 = input.charAt(peg$currPos);
        peg$currPos++;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e24);
        }
      }

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f7(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsebullet() {
    var s0, s1, s2, s3, s4, s5, s6;
    s0 = peg$currPos;
    peg$savedPos = peg$currPos;
    s1 = peg$f8();

    if (s1) {
      s1 = undefined;
    } else {
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      peg$savedPos = peg$currPos;
      s2 = peg$f9();

      if (s2) {
        s2 = undefined;
      } else {
        s2 = peg$FAILED;
      }

      if (s2 !== peg$FAILED) {
        if (input.substr(peg$currPos, 2) === peg$c22) {
          s3 = peg$c22;
          peg$currPos += 2;
        } else {
          s3 = peg$FAILED;

          if (peg$silentFails === 0) {
            peg$fail(peg$e25);
          }
        }

        if (s3 === peg$FAILED) {
          if (input.substr(peg$currPos, 2) === peg$c23) {
            s3 = peg$c23;
            peg$currPos += 2;
          } else {
            s3 = peg$FAILED;

            if (peg$silentFails === 0) {
              peg$fail(peg$e26);
            }
          }
        }

        if (s3 !== peg$FAILED) {
          s4 = peg$parsenotBulletStartUnit();

          if (s4 === peg$FAILED) {
            s4 = null;
          }

          s5 = [];
          s6 = peg$parsenotBulletUnit();

          while (s6 !== peg$FAILED) {
            s5.push(s6);
            s6 = peg$parsenotBulletUnit();
          }

          peg$savedPos = s0;
          s0 = peg$f10(s3, s4, s5);
        } else {
          peg$currPos = s0;
          s0 = peg$FAILED;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotBulletStartUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.substr(peg$currPos, 2) === peg$c22) {
      s2 = peg$c22;
      peg$currPos += 2;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e25);
      }
    }

    if (s2 === peg$FAILED) {
      if (input.substr(peg$currPos, 2) === peg$c23) {
        s2 = peg$c23;
        peg$currPos += 2;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e26);
        }
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parsestartUnit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotBulletUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.substr(peg$currPos, 2) === peg$c22) {
      s2 = peg$c22;
      peg$currPos += 2;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e25);
      }
    }

    if (s2 === peg$FAILED) {
      if (input.substr(peg$currPos, 2) === peg$c23) {
        s2 = peg$c23;
        peg$currPos += 2;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e26);
        }
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parseunit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parseinnerBoldSequence() {
    var s0, s1, s2, s3, s4, s5;
    s0 = peg$currPos;

    if (input.charCodeAt(peg$currPos) === 42) {
      s1 = peg$c0;
      peg$currPos++;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e2);
      }
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$currPos;
      peg$silentFails++;
      s3 = peg$parsewhiteSpace();
      peg$silentFails--;

      if (s3 === peg$FAILED) {
        s2 = undefined;
      } else {
        peg$currPos = s2;
        s2 = peg$FAILED;
      }

      if (s2 !== peg$FAILED) {
        s3 = peg$parsenotBoldStartUnit();

        if (s3 !== peg$FAILED) {
          s4 = [];
          s5 = peg$parsenotBoldUnit();

          while (s5 !== peg$FAILED) {
            s4.push(s5);
            s5 = peg$parsenotBoldUnit();
          }

          if (input.charCodeAt(peg$currPos) === 42) {
            s5 = peg$c0;
            peg$currPos++;
          } else {
            s5 = peg$FAILED;

            if (peg$silentFails === 0) {
              peg$fail(peg$e2);
            }
          }

          if (s5 !== peg$FAILED) {
            peg$savedPos = s0;
            s0 = peg$f11(s3, s4);
          } else {
            peg$currPos = s0;
            s0 = peg$FAILED;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$FAILED;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotBoldStartUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 42) {
      s2 = peg$c0;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e2);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parsestartUnit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotBoldUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 42) {
      s2 = peg$c0;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e2);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parseunit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parseinnerItalicSequence() {
    var s0, s1, s2, s3, s4, s5;
    s0 = peg$currPos;

    if (input.charCodeAt(peg$currPos) === 95) {
      s1 = peg$c3;
      peg$currPos++;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e5);
      }
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$currPos;
      peg$silentFails++;
      s3 = peg$parsewhiteSpace();
      peg$silentFails--;

      if (s3 === peg$FAILED) {
        s2 = undefined;
      } else {
        peg$currPos = s2;
        s2 = peg$FAILED;
      }

      if (s2 !== peg$FAILED) {
        s3 = peg$parsenotItalicStartUnit();

        if (s3 !== peg$FAILED) {
          s4 = [];
          s5 = peg$parsenotItalicUnit();

          while (s5 !== peg$FAILED) {
            s4.push(s5);
            s5 = peg$parsenotItalicUnit();
          }

          if (input.charCodeAt(peg$currPos) === 95) {
            s5 = peg$c3;
            peg$currPos++;
          } else {
            s5 = peg$FAILED;

            if (peg$silentFails === 0) {
              peg$fail(peg$e5);
            }
          }

          if (s5 !== peg$FAILED) {
            peg$savedPos = s0;
            s0 = peg$f12(s3, s4);
          } else {
            peg$currPos = s0;
            s0 = peg$FAILED;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$FAILED;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotItalicUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 95) {
      s2 = peg$c3;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e5);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parseunit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotItalicStartUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 95) {
      s2 = peg$c3;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e5);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parsestartUnit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parseinnerDisabledSequence() {
    var s0, s1, s2, s3, s4, s5;
    s0 = peg$currPos;

    if (input.charCodeAt(peg$currPos) === 45) {
      s1 = peg$c1;
      peg$currPos++;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e3);
      }
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$currPos;
      peg$silentFails++;
      s3 = peg$parsewhiteSpace();
      peg$silentFails--;

      if (s3 === peg$FAILED) {
        s2 = undefined;
      } else {
        peg$currPos = s2;
        s2 = peg$FAILED;
      }

      if (s2 !== peg$FAILED) {
        s3 = peg$parsenotDisabledStartUnit();

        if (s3 !== peg$FAILED) {
          s4 = [];
          s5 = peg$parsenotDisabledUnit();

          while (s5 !== peg$FAILED) {
            s4.push(s5);
            s5 = peg$parsenotDisabledUnit();
          }

          if (input.charCodeAt(peg$currPos) === 45) {
            s5 = peg$c1;
            peg$currPos++;
          } else {
            s5 = peg$FAILED;

            if (peg$silentFails === 0) {
              peg$fail(peg$e3);
            }
          }

          if (s5 !== peg$FAILED) {
            peg$savedPos = s0;
            s0 = peg$f13(s3, s4);
          } else {
            peg$currPos = s0;
            s0 = peg$FAILED;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$FAILED;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotDisabledUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 45) {
      s2 = peg$c1;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e3);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parseunit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotDisabledStartUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 45) {
      s2 = peg$c1;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e3);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parsestartUnit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parseunderlineSequence() {
    var s0, s1;
    s0 = peg$currPos;
    s1 = peg$parseinnerUnderlineSequence();

    if (s1 !== peg$FAILED) {
      peg$savedPos = s0;
      s1 = peg$f14(s1);
    }

    s0 = s1;
    return s0;
  }

  function peg$parseinnerUnderlineSequence() {
    var s0, s1, s2, s3, s4, s5;
    s0 = peg$currPos;

    if (input.charCodeAt(peg$currPos) === 38) {
      s1 = peg$c2;
      peg$currPos++;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e4);
      }
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$currPos;
      peg$silentFails++;
      s3 = peg$parsewhiteSpace();
      peg$silentFails--;

      if (s3 === peg$FAILED) {
        s2 = undefined;
      } else {
        peg$currPos = s2;
        s2 = peg$FAILED;
      }

      if (s2 !== peg$FAILED) {
        s3 = peg$parsenotUnderlinedStartUnit();

        if (s3 !== peg$FAILED) {
          s4 = [];
          s5 = peg$parsenotUnderlinedUnit();

          while (s5 !== peg$FAILED) {
            s4.push(s5);
            s5 = peg$parsenotUnderlinedUnit();
          }

          if (input.charCodeAt(peg$currPos) === 38) {
            s5 = peg$c2;
            peg$currPos++;
          } else {
            s5 = peg$FAILED;

            if (peg$silentFails === 0) {
              peg$fail(peg$e4);
            }
          }

          if (s5 !== peg$FAILED) {
            peg$savedPos = s0;
            s0 = peg$f15(s3, s4);
          } else {
            peg$currPos = s0;
            s0 = peg$FAILED;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$FAILED;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotUnderlinedUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 38) {
      s2 = peg$c2;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e4);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parseunit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotUnderlinedStartUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 38) {
      s2 = peg$c2;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e4);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parsestartUnit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parseinnerStrikethroughSequence() {
    var s0, s1, s2, s3, s4, s5;
    s0 = peg$currPos;

    if (input.charCodeAt(peg$currPos) === 126) {
      s1 = peg$c4;
      peg$currPos++;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e6);
      }
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$currPos;
      peg$silentFails++;
      s3 = peg$parsewhiteSpace();
      peg$silentFails--;

      if (s3 === peg$FAILED) {
        s2 = undefined;
      } else {
        peg$currPos = s2;
        s2 = peg$FAILED;
      }

      if (s2 !== peg$FAILED) {
        s3 = peg$parsenotStrikethroughStartUnit();

        if (s3 !== peg$FAILED) {
          s4 = [];
          s5 = peg$parsenotStrikethroughUnit();

          while (s5 !== peg$FAILED) {
            s4.push(s5);
            s5 = peg$parsenotStrikethroughUnit();
          }

          if (input.charCodeAt(peg$currPos) === 126) {
            s5 = peg$c4;
            peg$currPos++;
          } else {
            s5 = peg$FAILED;

            if (peg$silentFails === 0) {
              peg$fail(peg$e6);
            }
          }

          if (s5 !== peg$FAILED) {
            peg$savedPos = s0;
            s0 = peg$f16(s3, s4);
          } else {
            peg$currPos = s0;
            s0 = peg$FAILED;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$FAILED;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotStrikethroughUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 126) {
      s2 = peg$c4;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e6);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parseunit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotStrikethroughStartUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 126) {
      s2 = peg$c4;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e6);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parsestartUnit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parseurlLinkSequence() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$parseallowedStarterPrefix();

    if (s1 === peg$FAILED) {
      peg$savedPos = peg$currPos;
      s1 = peg$f2();

      if (s1) {
        s1 = undefined;
      } else {
        s1 = peg$FAILED;
      }
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parseinnerUrlLinkSequence();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f3(s1, s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parseinnerUrlLinkSequence() {
    var s0, s1, s2, s3, s4, s5, s6, s7, s8;
    s0 = peg$currPos;

    if (input.charCodeAt(peg$currPos) === 91) {
      s1 = peg$c5;
      peg$currPos++;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e7);
      }
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$currPos;
      peg$silentFails++;
      s3 = peg$parsewhiteSpace();
      peg$silentFails--;

      if (s3 === peg$FAILED) {
        s2 = undefined;
      } else {
        peg$currPos = s2;
        s2 = peg$FAILED;
      }

      if (s2 !== peg$FAILED) {
        s3 = peg$parsenotStartLinkUnit();

        if (s3 !== peg$FAILED) {
          s4 = [];
          s5 = peg$parsenotLinkUnit();

          while (s5 !== peg$FAILED) {
            s4.push(s5);
            s5 = peg$parsenotLinkUnit();
          }

          if (input.substr(peg$currPos, 2) === peg$c24) {
            s5 = peg$c24;
            peg$currPos += 2;
          } else {
            s5 = peg$FAILED;

            if (peg$silentFails === 0) {
              peg$fail(peg$e27);
            }
          }

          if (s5 !== peg$FAILED) {
            s6 = peg$currPos;
            peg$silentFails++;
            s7 = peg$parsewhiteSpace();
            peg$silentFails--;

            if (s7 === peg$FAILED) {
              s6 = undefined;
            } else {
              peg$currPos = s6;
              s6 = peg$FAILED;
            }

            if (s6 !== peg$FAILED) {
              s7 = [];
              s8 = peg$parsenotLinkTargetUnit();

              if (s8 !== peg$FAILED) {
                while (s8 !== peg$FAILED) {
                  s7.push(s8);
                  s8 = peg$parsenotLinkTargetUnit();
                }
              } else {
                s7 = peg$FAILED;
              }

              if (s7 !== peg$FAILED) {
                if (input.charCodeAt(peg$currPos) === 41) {
                  s8 = peg$c14;
                  peg$currPos++;
                } else {
                  s8 = peg$FAILED;

                  if (peg$silentFails === 0) {
                    peg$fail(peg$e16);
                  }
                }

                if (s8 !== peg$FAILED) {
                  peg$savedPos = s0;
                  s0 = peg$f17(s3, s4, s7);
                } else {
                  peg$currPos = s0;
                  s0 = peg$FAILED;
                }
              } else {
                peg$currPos = s0;
                s0 = peg$FAILED;
              }
            } else {
              peg$currPos = s0;
              s0 = peg$FAILED;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$FAILED;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$FAILED;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotLinkTargetUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 91) {
      s2 = peg$c5;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e7);
      }
    }

    if (s2 === peg$FAILED) {
      if (input.charCodeAt(peg$currPos) === 93) {
        s2 = peg$c6;
        peg$currPos++;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e8);
        }
      }

      if (s2 === peg$FAILED) {
        if (input.charCodeAt(peg$currPos) === 40) {
          s2 = peg$c13;
          peg$currPos++;
        } else {
          s2 = peg$FAILED;

          if (peg$silentFails === 0) {
            peg$fail(peg$e15);
          }
        }

        if (s2 === peg$FAILED) {
          if (input.charCodeAt(peg$currPos) === 41) {
            s2 = peg$c14;
            peg$currPos++;
          } else {
            s2 = peg$FAILED;

            if (peg$silentFails === 0) {
              peg$fail(peg$e16);
            }
          }
        }
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parsechar();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parselinkSequence() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$parseallowedStarterPrefix();

    if (s1 === peg$FAILED) {
      peg$savedPos = peg$currPos;
      s1 = peg$f2();

      if (s1) {
        s1 = undefined;
      } else {
        s1 = peg$FAILED;
      }
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parseinnerLinkSequence();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f18(s1, s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parseinnerLinkSequence() {
    var s0, s1, s2, s3, s4, s5;
    s0 = peg$currPos;

    if (input.charCodeAt(peg$currPos) === 91) {
      s1 = peg$c5;
      peg$currPos++;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e7);
      }
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$currPos;
      peg$silentFails++;
      s3 = peg$parsewhiteSpace();
      peg$silentFails--;

      if (s3 === peg$FAILED) {
        s2 = undefined;
      } else {
        peg$currPos = s2;
        s2 = peg$FAILED;
      }

      if (s2 !== peg$FAILED) {
        s3 = peg$parsenotStartLinkUnit();

        if (s3 !== peg$FAILED) {
          s4 = [];
          s5 = peg$parsenotLinkUnit();

          while (s5 !== peg$FAILED) {
            s4.push(s5);
            s5 = peg$parsenotLinkUnit();
          }

          if (input.charCodeAt(peg$currPos) === 93) {
            s5 = peg$c6;
            peg$currPos++;
          } else {
            s5 = peg$FAILED;

            if (peg$silentFails === 0) {
              peg$fail(peg$e8);
            }
          }

          if (s5 !== peg$FAILED) {
            peg$savedPos = s0;
            s0 = peg$f19(s3, s4);
          } else {
            peg$currPos = s0;
            s0 = peg$FAILED;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$FAILED;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotStartLinkUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 91) {
      s2 = peg$c5;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e7);
      }
    }

    if (s2 === peg$FAILED) {
      if (input.charCodeAt(peg$currPos) === 93) {
        s2 = peg$c6;
        peg$currPos++;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e8);
        }
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parsestartLinkUnit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotLinkUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 91) {
      s2 = peg$c5;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e7);
      }
    }

    if (s2 === peg$FAILED) {
      if (input.charCodeAt(peg$currPos) === 93) {
        s2 = peg$c6;
        peg$currPos++;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e8);
        }
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parselinkUnit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsecolorSequence() {
    var s0, s1, s2, s3, s4, s5, s6, s7, s8, s9, s10;
    s0 = peg$currPos;

    if (input.substr(peg$currPos, 7).toLowerCase() === peg$c8) {
      s1 = input.substr(peg$currPos, 7);
      peg$currPos += 7;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e10);
      }
    }

    if (s1 !== peg$FAILED) {
      if (input.charCodeAt(peg$currPos) === 32) {
        s2 = peg$c15;
        peg$currPos++;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e17);
        }
      }

      if (s2 === peg$FAILED) {
        s2 = null;
      }

      if (input.charCodeAt(peg$currPos) === 35) {
        s3 = peg$c25;
        peg$currPos++;
      } else {
        s3 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e28);
        }
      }

      if (s3 === peg$FAILED) {
        s3 = null;
      }

      s4 = [];

      if (peg$r2.test(input.charAt(peg$currPos))) {
        s5 = input.charAt(peg$currPos);
        peg$currPos++;
      } else {
        s5 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e29);
        }
      }

      if (s5 !== peg$FAILED) {
        while (s5 !== peg$FAILED) {
          s4.push(s5);

          if (peg$r2.test(input.charAt(peg$currPos))) {
            s5 = input.charAt(peg$currPos);
            peg$currPos++;
          } else {
            s5 = peg$FAILED;

            if (peg$silentFails === 0) {
              peg$fail(peg$e29);
            }
          }
        }
      } else {
        s4 = peg$FAILED;
      }

      if (s4 !== peg$FAILED) {
        if (input.charCodeAt(peg$currPos) === 125) {
          s5 = peg$c26;
          peg$currPos++;
        } else {
          s5 = peg$FAILED;

          if (peg$silentFails === 0) {
            peg$fail(peg$e30);
          }
        }

        if (s5 !== peg$FAILED) {
          s6 = peg$parsenotStartColorUnit();

          if (s6 !== peg$FAILED) {
            s7 = [];
            s8 = peg$parsenotColorUnit();

            while (s8 !== peg$FAILED) {
              s7.push(s8);
              s8 = peg$parsenotColorUnit();
            }

            if (input.charCodeAt(peg$currPos) === 123) {
              s8 = peg$c27;
              peg$currPos++;
            } else {
              s8 = peg$FAILED;

              if (peg$silentFails === 0) {
                peg$fail(peg$e31);
              }
            }

            if (s8 !== peg$FAILED) {
              if (input.substr(peg$currPos, 5).toLowerCase() === peg$c28) {
                s9 = input.substr(peg$currPos, 5);
                peg$currPos += 5;
              } else {
                s9 = peg$FAILED;

                if (peg$silentFails === 0) {
                  peg$fail(peg$e32);
                }
              }

              if (s9 !== peg$FAILED) {
                if (input.charCodeAt(peg$currPos) === 125) {
                  s10 = peg$c26;
                  peg$currPos++;
                } else {
                  s10 = peg$FAILED;

                  if (peg$silentFails === 0) {
                    peg$fail(peg$e30);
                  }
                }

                if (s10 !== peg$FAILED) {
                  peg$savedPos = s0;
                  s0 = peg$f20(s2, s3, s4, s6, s7, s9);
                } else {
                  peg$currPos = s0;
                  s0 = peg$FAILED;
                }
              } else {
                peg$currPos = s0;
                s0 = peg$FAILED;
              }
            } else {
              peg$currPos = s0;
              s0 = peg$FAILED;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$FAILED;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$FAILED;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotStartColorUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.substr(peg$currPos, 6).toLowerCase() === peg$c29) {
      s2 = input.substr(peg$currPos, 6);
      peg$currPos += 6;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e33);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parsestartUnit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotColorUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.substr(peg$currPos, 6).toLowerCase() === peg$c29) {
      s2 = input.substr(peg$currPos, 6);
      peg$currPos += 6;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e33);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parseunit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsecolourSequence() {
    var s0, s1, s2, s3, s4, s5, s6, s7, s8, s9, s10;
    s0 = peg$currPos;

    if (input.substr(peg$currPos, 8).toLowerCase() === peg$c9) {
      s1 = input.substr(peg$currPos, 8);
      peg$currPos += 8;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e11);
      }
    }

    if (s1 !== peg$FAILED) {
      if (input.charCodeAt(peg$currPos) === 32) {
        s2 = peg$c15;
        peg$currPos++;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e17);
        }
      }

      if (s2 === peg$FAILED) {
        s2 = null;
      }

      if (input.charCodeAt(peg$currPos) === 35) {
        s3 = peg$c25;
        peg$currPos++;
      } else {
        s3 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e28);
        }
      }

      if (s3 === peg$FAILED) {
        s3 = null;
      }

      s4 = [];

      if (peg$r2.test(input.charAt(peg$currPos))) {
        s5 = input.charAt(peg$currPos);
        peg$currPos++;
      } else {
        s5 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e29);
        }
      }

      if (s5 !== peg$FAILED) {
        while (s5 !== peg$FAILED) {
          s4.push(s5);

          if (peg$r2.test(input.charAt(peg$currPos))) {
            s5 = input.charAt(peg$currPos);
            peg$currPos++;
          } else {
            s5 = peg$FAILED;

            if (peg$silentFails === 0) {
              peg$fail(peg$e29);
            }
          }
        }
      } else {
        s4 = peg$FAILED;
      }

      if (s4 !== peg$FAILED) {
        if (input.charCodeAt(peg$currPos) === 125) {
          s5 = peg$c26;
          peg$currPos++;
        } else {
          s5 = peg$FAILED;

          if (peg$silentFails === 0) {
            peg$fail(peg$e30);
          }
        }

        if (s5 !== peg$FAILED) {
          s6 = peg$parsenotStartColourUnit();

          if (s6 !== peg$FAILED) {
            s7 = [];
            s8 = peg$parsenotColourUnit();

            while (s8 !== peg$FAILED) {
              s7.push(s8);
              s8 = peg$parsenotColourUnit();
            }

            if (input.charCodeAt(peg$currPos) === 123) {
              s8 = peg$c27;
              peg$currPos++;
            } else {
              s8 = peg$FAILED;

              if (peg$silentFails === 0) {
                peg$fail(peg$e31);
              }
            }

            if (s8 !== peg$FAILED) {
              if (input.substr(peg$currPos, 6).toLowerCase() === peg$c30) {
                s9 = input.substr(peg$currPos, 6);
                peg$currPos += 6;
              } else {
                s9 = peg$FAILED;

                if (peg$silentFails === 0) {
                  peg$fail(peg$e34);
                }
              }

              if (s9 !== peg$FAILED) {
                if (input.charCodeAt(peg$currPos) === 125) {
                  s10 = peg$c26;
                  peg$currPos++;
                } else {
                  s10 = peg$FAILED;

                  if (peg$silentFails === 0) {
                    peg$fail(peg$e30);
                  }
                }

                if (s10 !== peg$FAILED) {
                  peg$savedPos = s0;
                  s0 = peg$f21(s2, s3, s4, s6, s7, s9);
                } else {
                  peg$currPos = s0;
                  s0 = peg$FAILED;
                }
              } else {
                peg$currPos = s0;
                s0 = peg$FAILED;
              }
            } else {
              peg$currPos = s0;
              s0 = peg$FAILED;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$FAILED;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$FAILED;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotStartColourUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.substr(peg$currPos, 7).toLowerCase() === peg$c31) {
      s2 = input.substr(peg$currPos, 7);
      peg$currPos += 7;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e35);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parsestartUnit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotColourUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.substr(peg$currPos, 7).toLowerCase() === peg$c31) {
      s2 = input.substr(peg$currPos, 7);
      peg$currPos += 7;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e35);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parseunit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsesizeSequenceNoText() {
    var s0, s1, s2, s3, s4, s5;
    s0 = peg$currPos;

    if (input.substr(peg$currPos, 6).toLowerCase() === peg$c12) {
      s1 = input.substr(peg$currPos, 6);
      peg$currPos += 6;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e14);
      }
    }

    if (s1 !== peg$FAILED) {
      if (input.charCodeAt(peg$currPos) === 32) {
        s2 = peg$c15;
        peg$currPos++;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e17);
        }
      }

      if (s2 === peg$FAILED) {
        s2 = null;
      }

      s3 = [];
      s4 = peg$parsedigit();

      if (s4 !== peg$FAILED) {
        while (s4 !== peg$FAILED) {
          s3.push(s4);
          s4 = peg$parsedigit();
        }
      } else {
        s3 = peg$FAILED;
      }

      if (s3 !== peg$FAILED) {
        if (input.charCodeAt(peg$currPos) === 125) {
          s4 = peg$c26;
          peg$currPos++;
        } else {
          s4 = peg$FAILED;

          if (peg$silentFails === 0) {
            peg$fail(peg$e30);
          }
        }

        if (s4 !== peg$FAILED) {
          if (input.substr(peg$currPos, 6).toLowerCase() === peg$c11) {
            s5 = input.substr(peg$currPos, 6);
            peg$currPos += 6;
          } else {
            s5 = peg$FAILED;

            if (peg$silentFails === 0) {
              peg$fail(peg$e13);
            }
          }

          if (s5 !== peg$FAILED) {
            peg$savedPos = s0;
            s0 = peg$f22(s2, s3);
          } else {
            peg$currPos = s0;
            s0 = peg$FAILED;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$FAILED;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsesizeSequence() {
    var s0, s1, s2, s3, s4, s5, s6, s7;
    s0 = peg$currPos;

    if (input.substr(peg$currPos, 6).toLowerCase() === peg$c12) {
      s1 = input.substr(peg$currPos, 6);
      peg$currPos += 6;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e14);
      }
    }

    if (s1 !== peg$FAILED) {
      if (input.charCodeAt(peg$currPos) === 32) {
        s2 = peg$c15;
        peg$currPos++;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e17);
        }
      }

      if (s2 === peg$FAILED) {
        s2 = null;
      }

      s3 = [];
      s4 = peg$parsedigit();

      if (s4 !== peg$FAILED) {
        while (s4 !== peg$FAILED) {
          s3.push(s4);
          s4 = peg$parsedigit();
        }
      } else {
        s3 = peg$FAILED;
      }

      if (s3 !== peg$FAILED) {
        if (input.charCodeAt(peg$currPos) === 125) {
          s4 = peg$c26;
          peg$currPos++;
        } else {
          s4 = peg$FAILED;

          if (peg$silentFails === 0) {
            peg$fail(peg$e30);
          }
        }

        if (s4 !== peg$FAILED) {
          s5 = peg$parsenotStartSizeUnit();

          if (s5 !== peg$FAILED) {
            s6 = [];
            s7 = peg$parsenotSizeUnit();

            while (s7 !== peg$FAILED) {
              s6.push(s7);
              s7 = peg$parsenotSizeUnit();
            }

            if (input.substr(peg$currPos, 6).toLowerCase() === peg$c11) {
              s7 = input.substr(peg$currPos, 6);
              peg$currPos += 6;
            } else {
              s7 = peg$FAILED;

              if (peg$silentFails === 0) {
                peg$fail(peg$e13);
              }
            }

            if (s7 !== peg$FAILED) {
              peg$savedPos = s0;
              s0 = peg$f23(s2, s3, s5, s6);
            } else {
              peg$currPos = s0;
              s0 = peg$FAILED;
            }
          } else {
            peg$currPos = s0;
            s0 = peg$FAILED;
          }
        } else {
          peg$currPos = s0;
          s0 = peg$FAILED;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotStartSizeUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.substr(peg$currPos, 5).toLowerCase() === peg$c32) {
      s2 = input.substr(peg$currPos, 5);
      peg$currPos += 5;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e36);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parsestartUnit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsenotSizeUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.substr(peg$currPos, 5).toLowerCase() === peg$c32) {
      s2 = input.substr(peg$currPos, 5);
      peg$currPos += 5;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e36);
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      s2 = peg$parseunit();

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsementionSequence() {
    var s0, s1, s2, s3, s4;
    s0 = peg$currPos;

    if (input.substr(peg$currPos, 2) === peg$c33) {
      s1 = peg$c33;
      peg$currPos += 2;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e37);
      }
    }

    if (s1 !== peg$FAILED) {
      s2 = [];
      s3 = peg$parsementionUserUnit();

      if (s3 !== peg$FAILED) {
        while (s3 !== peg$FAILED) {
          s2.push(s3);
          s3 = peg$parsementionUserUnit();
        }
      } else {
        s2 = peg$FAILED;
      }

      if (s2 !== peg$FAILED) {
        if (input.charCodeAt(peg$currPos) === 93) {
          s3 = peg$c6;
          peg$currPos++;
        } else {
          s3 = peg$FAILED;

          if (peg$silentFails === 0) {
            peg$fail(peg$e8);
          }
        }

        if (s3 !== peg$FAILED) {
          s4 = peg$parsementionTarget();

          if (s4 === peg$FAILED) {
            s4 = null;
          }

          peg$savedPos = s0;
          s0 = peg$f24(s2, s4);
        } else {
          peg$currPos = s0;
          s0 = peg$FAILED;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsementionTarget() {
    var s0, s1, s2, s3;
    s0 = peg$currPos;

    if (input.charCodeAt(peg$currPos) === 40) {
      s1 = peg$c13;
      peg$currPos++;
    } else {
      s1 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e15);
      }
    }

    if (s1 !== peg$FAILED) {
      s2 = [];
      s3 = peg$parsementionTargetUnit();

      if (s3 !== peg$FAILED) {
        while (s3 !== peg$FAILED) {
          s2.push(s3);
          s3 = peg$parsementionTargetUnit();
        }
      } else {
        s2 = peg$FAILED;
      }

      if (s2 !== peg$FAILED) {
        if (input.charCodeAt(peg$currPos) === 41) {
          s3 = peg$c14;
          peg$currPos++;
        } else {
          s3 = peg$FAILED;

          if (peg$silentFails === 0) {
            peg$fail(peg$e16);
          }
        }

        if (s3 !== peg$FAILED) {
          peg$savedPos = s0;
          s0 = peg$f25(s2);
        } else {
          peg$currPos = s0;
          s0 = peg$FAILED;
        }
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsementionUserUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 91) {
      s2 = peg$c5;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e7);
      }
    }

    if (s2 === peg$FAILED) {
      if (input.charCodeAt(peg$currPos) === 93) {
        s2 = peg$c6;
        peg$currPos++;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e8);
        }
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      if (peg$r3.test(input.charAt(peg$currPos))) {
        s2 = input.charAt(peg$currPos);
        peg$currPos++;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e38);
        }
      }

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  function peg$parsementionTargetUnit() {
    var s0, s1, s2;
    s0 = peg$currPos;
    s1 = peg$currPos;
    peg$silentFails++;

    if (input.charCodeAt(peg$currPos) === 91) {
      s2 = peg$c5;
      peg$currPos++;
    } else {
      s2 = peg$FAILED;

      if (peg$silentFails === 0) {
        peg$fail(peg$e7);
      }
    }

    if (s2 === peg$FAILED) {
      if (input.charCodeAt(peg$currPos) === 93) {
        s2 = peg$c6;
        peg$currPos++;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e8);
        }
      }
    }

    peg$silentFails--;

    if (s2 === peg$FAILED) {
      s1 = undefined;
    } else {
      peg$currPos = s1;
      s1 = peg$FAILED;
    }

    if (s1 !== peg$FAILED) {
      if (peg$r4.test(input.charAt(peg$currPos))) {
        s2 = input.charAt(peg$currPos);
        peg$currPos++;
      } else {
        s2 = peg$FAILED;

        if (peg$silentFails === 0) {
          peg$fail(peg$e39);
        }
      }

      if (s2 !== peg$FAILED) {
        peg$savedPos = s0;
        s0 = peg$f6(s2);
      } else {
        peg$currPos = s0;
        s0 = peg$FAILED;
      }
    } else {
      peg$currPos = s0;
      s0 = peg$FAILED;
    }

    return s0;
  }

  peg$result = peg$startRuleFunction();

  if (peg$result !== peg$FAILED && peg$currPos === input.length) {
    return peg$result;
  } else {
    if (peg$result !== peg$FAILED && peg$currPos < input.length) {
      peg$fail(peg$endExpectation());
    }

    throw peg$buildStructuredError(peg$maxFailExpected, peg$maxFailPos < input.length ? input.charAt(peg$maxFailPos) : null, peg$maxFailPos < input.length ? peg$computeLocation(peg$maxFailPos, peg$maxFailPos + 1) : peg$computeLocation(peg$maxFailPos, peg$maxFailPos));
  }
}

/***/ }),

/***/ "./js/core/texts/textconstants.js":
/*!****************************************!*\
  !*** ./js/core/texts/textconstants.js ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.NodeTypes = exports.NodeIdentifiers = exports.FontAwesomeIconPrefix = exports.ApplicationIconURISchema = void 0;
const NodeTypes = {
  BOLD: 'bold',
  ITALIC: 'italic',
  UNDERLINE: 'underline',
  SIZE: 'size',
  COLOR: 'color',
  LINK: 'link',
  DISABLED: 'disabled',
  STRIKETHROUGH: 'strikethrough',
  EOL: 'eol',
  SOFT_EOL: 'softEol',
  BULLET: 'bullet',
  ESCAPEDCHAR: 'escapedChar',
  MENTION: 'mention',
  ACTION_LINK: 'actionLink'
};
exports.NodeTypes = NodeTypes;
const NodeIdentifiers = {
  MentionAll: 'all'
};
exports.NodeIdentifiers = NodeIdentifiers;
const ApplicationIconURISchema = "icon:";
exports.ApplicationIconURISchema = ApplicationIconURISchema;
const FontAwesomeIconPrefix = "fontAwesomeIcon:";
exports.FontAwesomeIconPrefix = FontAwesomeIconPrefix;

/***/ }),

/***/ "./js/core/texts/textutilities.js":
/*!****************************************!*\
  !*** ./js/core/texts/textutilities.js ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.biggestFontSize = biggestFontSize;
exports.flattenTextDom = flattenTextDom;
exports.makeUniqueName = makeUniqueName;
exports.nodesToString = nodesToString;
exports.replaceLinkTarget = replaceLinkTarget;
exports.splitTextOnCommas = splitTextOnCommas;
exports.stringToNodes = stringToNodes;
exports.textDom2AttributedText = textDom2AttributedText;
exports.unformatNodes = unformatNodes;
exports.unformatText = unformatText;
exports.unformatTextReplacingNewlines = unformatTextReplacingNewlines;

var _textconstants = __webpack_require__(/*! core/texts/textconstants */ "./js/core/texts/textconstants.js");

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var TextGrammar = _interopRequireWildcard(__webpack_require__(/*! core/texts/textGrammar */ "./js/core/texts/textGrammar.js"));

var _colors = _interopRequireDefault(__webpack_require__(/*! core/colors */ "./js/core/colors.ts"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * @typedef {Object} ComplexNode
 * @property {string} type
 * @property {Array<Node>} [children]
 * @property {Array<string>} [target]
 * @property {string} [value]
 * @property {string} [hashtag]
 * @property {string} [color]
 * @property {number} [size]
 * @property {boolean} [hasSpace]
 */

/**
 * @typedef {ComplexNode | string} Node
 * @typedef {{ fontFamily: string; fontSize: number; textColor: string }} AttributedTextBaseAttributes
 * @typedef {{ string: string, nodes: Node[] }} AttributedTextAttribute
 * @typedef {{ attributes: AttributedTextAttribute[]; baseAttributes: AttributedTextBaseAttributes }} AttributedText
 */
function flattenTextDom(textDomElements) {
  var result = [];
  textDomElements.forEach(function (textDomNode) {
    if (_balsamiq.default.isString(textDomNode)) {
      result.push(textDomNode);
    } else {
      if (textDomNode.type == _textconstants.NodeTypes.EOL) {
        result.push("\n");
      } else {
        var openFormat = {
          type: textDomNode.type,
          open: true
        };

        if (textDomNode.type === _textconstants.NodeTypes.COLOR) {
          openFormat.color = textDomNode.color;
        } else if (textDomNode.type === _textconstants.NodeTypes.SIZE) {
          openFormat.size = textDomNode.size;
        } else if (textDomNode.type === _textconstants.NodeTypes.LINK) {
          openFormat.target = textDomNode.target;
        } else if (textDomNode.type === _textconstants.NodeTypes.ACTION_LINK) {
          openFormat.target = textDomNode.target;
        }

        result.push(openFormat);

        if (textDomNode.children) {
          result = result.concat(flattenTextDom(textDomNode.children));
        }

        result.push({
          type: textDomNode.type,
          open: false
        });
      }
    }
  });
  return result;
}

function buildAttributedString(flatTextDom, textStyle) {
  var str = "";
  var attributes = [];
  var stack = []; // if we have a string, push it and its attributes from the stack

  let pushElem = function () {
    if (str) {
      var elem = {
        nodes: stack.map(_balsamiq.default.cloneDeep),
        string: str
      };
      str = "";
      attributes.push(elem);
    }
  };

  flatTextDom.forEach(item => {
    if (_balsamiq.default.isString(item)) {
      str += item;
    } else {
      // Push the current string and attr before changing attributes
      pushElem();

      if (item.open) {
        stack.push(makeAttributedStringItem(item));
      } else {
        stack.pop();
      }
    }
  }); // push remaining string

  pushElem();
  var baseAttributes = {
    fontFamily: textStyle.fontFamily,
    fontSize: textStyle.fontSize,
    textColor: textStyle.textColor
  };
  return {
    attributes: attributes,
    baseAttributes: baseAttributes
  };
}

function textDom2AttributedText(textDom, textStyle) {
  return buildAttributedString(flattenTextDom(textDom), textStyle);
}
/**
 *
 * @param {*} textDom
 * @param {number} fontSize
 * @returns {string}
 */


function textDomNodes2html(textDom, fontSize) {
  var html = "";
  textDom.forEach(function (textDomNode) {
    if (_balsamiq.default.isString(textDomNode)) {
      html += textDomNode; // TODO convert special chars
    } else {
      var tag = "";
      var attributes = "";
      var newSize = fontSize;

      if (textDomNode.type === _textconstants.NodeTypes.BOLD) {
        tag = "b";
      } else if (textDomNode.type === _textconstants.NodeTypes.ITALIC) {
        tag = "i";
      } else if (textDomNode.type === _textconstants.NodeTypes.UNDERLINE) {
        tag = "u";
      } else if (textDomNode.type === _textconstants.NodeTypes.SIZE) {
        tag = "span";
        newSize = parseInt(fontSize, 10);

        if (textDomNode.size[0] === "+" || textDomNode.size[0] === "-") {
          newSize += parseInt(textDomNode.size, 10);
        } else {
          newSize = parseInt(textDomNode.size, 10);
        }

        attributes = "style='font-size:" + newSize + "'";
      } else if (textDomNode.type === _textconstants.NodeTypes.COLOR) {
        tag = "span";
        attributes = "style='color:" + textDomNode.color + "'";
      } else if (textDomNode.type === _textconstants.NodeTypes.LINK) {
        tag = "a";
        attributes = "href"; // // #085394
      } else if (textDomNode.type === _textconstants.NodeTypes.DISABLED) {
        tag = "span";
        attributes = "style='color:#80808B'";
      } else if (textDomNode.type === _textconstants.NodeTypes.STRIKETHROUGH) {
        tag = "strike";
      }

      if (attributes.length > 0) {
        attributes = " " + attributes;
      }

      html += "<" + tag + attributes + ">";
      html += textDomNodes2html(textDomNode.children, newSize);
      html += "</" + tag + ">";
    }
  });
  return html;
}

function makeAttributedStringItem(item) {
  let domItem = _balsamiq.default.cloneDeep(item); // Items are modified to be better consumed by natives


  _balsamiq.default.assertEnum(domItem.type, _textconstants.NodeTypes); // Color Items should have INT colors


  if (domItem.type === _textconstants.NodeTypes.COLOR && _balsamiq.default.isString(domItem.color)) {
    domItem.color = _colors.default.colorFromString(domItem.color);
  }

  delete domItem.open;
  return domItem;
}

function biggestFontSize(p_textDom, p_baseFontSize) {
  let biggestSize = 0;
  p_textDom.forEach(textDomNode => {
    let nodeBiggestSize = 0;

    if (_balsamiq.default.isObject(textDomNode)) {
      if (textDomNode.type === _textconstants.NodeTypes.SIZE) {
        const nodeSize = parseInt(textDomNode.size, 10);
        nodeBiggestSize = isNaN(nodeSize) ? biggestSize : nodeSize;
      }

      nodeBiggestSize = Math.max(nodeBiggestSize, biggestFontSize(textDomNode.children, p_baseFontSize));
    } else {
      nodeBiggestSize = p_baseFontSize;
    }

    biggestSize = Math.max(biggestSize, nodeBiggestSize);
  });
  return biggestSize;
}
/**
 *
 * @param {Node[]} nodes
 * @returns {string}
 */


function unformatNodes(nodes) {
  _balsamiq.default.assertArray(nodes);

  let res = "";
  nodes.forEach(item => {
    if (_balsamiq.default.isString(item)) {
      res += item;
    } else if (item.type === _textconstants.NodeTypes.SOFT_EOL || item.type === _textconstants.NodeTypes.EOL) {
      res += "↵";

      if (item.children) {
        res += unformatNodes(item.children);
      }
    } else if (item.type) if (item.children) {
      res += unformatNodes(item.children);
    }
  });
  return res;
}
/**
 * @param {string} p_text
 * @returns {string[]}
 */


function splitTextOnCommas(p_text) {
  p_text = p_text.replace(/\\,/g, "__m_peldi_m__");
  p_text = p_text.replace(/, /g, ","); //remove the first space after each comma

  let res = p_text.split(",");
  const l = res.length;
  let i;

  for (i = 0; i < l; i += 1) {
    res[i] = res[i].replace(/__m_peldi_m__/g, ",");
  }

  return res;
}
/**
 *
 * @param {string} text
 * @returns {string}
 */


function unformatText(text) {
  _balsamiq.default.assertString(text);

  let textDom = stringToNodes(text);
  return unformatNodes(textDom);
}
/**
 *
 * @param {string} text
 * @returns {string}
 */


function unformatTextReplacingNewlines(text) {
  let s = unformatText(text);
  s = s.replace(/↵+/g, " ");
  return s;
}
/**
 *
 * @param {string} text
 * @returns {Node[]}
 */


function stringToNodes(text) {
  _balsamiq.default.assertString(text);

  if (text === "") {
    return [];
  }

  let splitText = text.split("\n");
  let nodes = [];
  splitText.forEach((str, index) => {
    nodes = nodes.concat(TextGrammar.parse(str));

    if (index < splitText.length - 1) {
      nodes.push({
        type: _textconstants.NodeTypes.EOL
      });
    }
  });
  return nodes;
}
/**
 *
 * @param {Node[]} nodes
 * @returns {string}
 */


function nodesToString(nodes) {
  _balsamiq.default.assertArray(nodes);

  let str = "";
  nodes.forEach(node => {
    if (_balsamiq.default.isString(node)) {
      str += node;
    } else {
      let pre = "";
      let post = "";

      switch (node.type) {
        case _textconstants.NodeTypes.BOLD:
          {
            pre = post = "*";
            break;
          }

        case _textconstants.NodeTypes.ITALIC:
          {
            pre = post = "_";
            break;
          }

        case _textconstants.NodeTypes.UNDERLINE:
          {
            pre = post = "&";
            break;
          }

        case _textconstants.NodeTypes.STRIKETHROUGH:
          {
            pre = post = "~";
            break;
          }

        case _textconstants.NodeTypes.DISABLED:
          {
            pre = post = "-";
            break;
          }

        case _textconstants.NodeTypes.BULLET:
          {
            pre = node.value;
            break;
          }

        case _textconstants.NodeTypes.EOL:
          {
            post = "\n";
            break;
          }

        case _textconstants.NodeTypes.SOFT_EOL:
          {
            pre = node.value;
            break;
          }

        case _textconstants.NodeTypes.MENTION:
          {
            pre = "@[";

            if (node.target) {
              post = "](" + nodesToString(node.target) + ")";
            } else {
              post = "]";
            }

            break;
          }

        case _textconstants.NodeTypes.ACTION_LINK:
        case _textconstants.NodeTypes.LINK:
          {
            pre = "[";

            if (node.target) {
              post = "](" + nodesToString(node.target) + ")";
            } else {
              post = "]";
            }

            break;
          }

        case _textconstants.NodeTypes.COLOR:
          {
            pre = "{" + node.value + ":" + (node.hasSpace ? " " : "") + (node.hashtag ? "#" : "") + node.color + "}"; // TODO check color

            post = "{" + node.value + "}";
            break;
          }

        case _textconstants.NodeTypes.SIZE:
          {
            pre = "{size:" + (node.hasSpace ? " " : "") + node.size + "}"; // TODO check size

            post = "{size}";
            break;
          }

        case _textconstants.NodeTypes.ESCAPEDCHAR:
          {
            post = "\\" + node.value;
            break;
          }

        default:
          {
            _balsamiq.default.assert(false, "nodesToString does not handle unknown node type: " + node.type);
          }
      }

      str += pre;

      if (node.children) {
        str += nodesToString(node.children);
      }

      str += post;
    }
  });
  return str;
}
/**
 *
 * @param {string} text
 * @param {number} linkIndex
 * @param {string|null} newTarget
 */


function replaceLinkTarget(text, linkIndex, newTarget) {
  _balsamiq.default.assertString(text);

  _balsamiq.default.assertNumber(linkIndex);

  if (newTarget !== null) {
    _balsamiq.default.assertString(newTarget);
  }

  let nodes = stringToNodes(text);
  let i = 0;
  let found = false;

  let processNodes = function (nodes) {
    nodes.forEach(node => {
      if (!found) {
        if (node.type === _textconstants.NodeTypes.LINK) {
          if (i === linkIndex) {
            if (newTarget) {
              node.target = [newTarget];
            } else {
              delete node.target;
            }

            found = true;
          } else {
            i += 1;
          }
        } else if (!_balsamiq.default.isString(node) && node.children) {
          processNodes(node.children);
        }
      }
    });
  };

  processNodes(nodes);
  return nodesToString(nodes);
}
/**
 *
 * @param {string[]} existingNames
 * @param {string} initialName
 * @returns {string}
 */


function makeUniqueName(existingNames, initialName) {
  let existingNamesLowerCased = existingNames.filter(n => !!n).map(n => n.toLowerCase());
  let uniqueName = initialName;
  const copyRegex = /\(\d+\)$/;
  let copyIndex = 1;

  while (existingNamesLowerCased.includes(uniqueName.toLowerCase())) {
    if (uniqueName.search(copyRegex) !== -1) {
      uniqueName = uniqueName.replace(copyRegex, `(${copyIndex})`);
    } else {
      uniqueName = uniqueName + " (1)";
    }

    copyIndex = copyIndex + 1;
  }

  return uniqueName;
}

/***/ }),

/***/ "./js/core/texts/wiki.js":
/*!*******************************!*\
  !*** ./js/core/texts/wiki.js ***!
  \*******************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.parseTreeText = parseTreeText;

function parseTreeText(text) {
  if (text === "") {
    return [];
  }

  const symbols = {
    '_': "_",
    'f': "folder",
    'F': "folder-open",
    '[+]': "plus-square-o",
    '[-]': "minus-square-o",
    '[=]': "minus-square-o",
    '[x]': "check-square-o",
    '[*]': "check-square-o",
    '[ ]': "square-o",
    '[]': "square-o",
    '>': "caret-right",
    '<': 'caret-left',
    // Support for RTL
    'v': "caret-down",
    '-': "file-text-o"
  };
  const lines = [];
  text = text.replace(/\s+$/, '');
  text.split('\n').forEach(line => {
    let level = 0;

    while (line.length > 0 && (line[0] === ' ' || line[0] === '.' || line[0] === '\t')) {
      level += 1;
      line = line.slice(1);
    }

    let symbolName = null;
    let symbolLength = 0;
    let potentialSymbol;
    Object.keys(symbols).forEach(symbol => {
      potentialSymbol = line.slice(0, symbol.length + 1);

      if (potentialSymbol === symbol.concat(' ') || potentialSymbol === symbol) {
        symbolName = symbols[symbol];
        symbolLength = symbol.length + 1;
      }
    });
    line = line.slice(symbolLength);
    lines.push({
      symbol: symbolName,
      text: line,
      level: level
    });
  });
  return lines;
}

/***/ }),

/***/ "./js/core/types.js":
/*!**************************!*\
  !*** ./js/core/types.js ***!
  \**************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/** @module Types */
var types = {};

types.rotatePtDeg = function (p_center, p_pt, angle) {
  var radians = Math.PI / 180 * angle;
  var cos = Math.cos(radians);
  var sin = Math.sin(radians);
  var nx = cos * (p_pt.x - p_center.x) + sin * (p_pt.y - p_center.y) + p_center.x;
  var ny = cos * (p_pt.y - p_center.y) - sin * (p_pt.x - p_center.x) + p_center.y;
  return types.newPoint(nx, ny);
};
/**
 * @typedef {{ width: number; height: number; }} Size
 * @typedef {{ x: number; y: number; }} Point
 * @typedef {{ x: number; y: number; width: number; height: number; }} Rect
 */

/**
 * Round the coordinates of a generic point/rectangle
 * @function
 * @param p_obj {Object} A generic point or rectangle
 */


types.roundCoords = function (p_obj) {
  if (p_obj.x) {
    p_obj.x = Math.round(p_obj.x);
  }

  if (p_obj.y) {
    p_obj.y = Math.round(p_obj.y);
  }

  if (p_obj.width) {
    p_obj.width = Math.round(p_obj.width);
  }

  if (p_obj.height) {
    p_obj.height = Math.round(p_obj.height);
  }
}; // Rectangles
// ----------


types.rectanglesUnion = function (p_a, p_b) {
  var minX = Math.min(p_a.x, p_b.x);
  var minY = Math.min(p_a.y, p_b.y);
  var maxX = Math.max(p_a.x + p_a.width, p_b.x + p_b.width);
  var maxY = Math.max(p_a.y + p_a.height, p_b.y + p_b.height);
  return types.newRectangle(minX, minY, maxX - minX, maxY - minY);
};

types.rectangleListUnion = function (p_rectangles) {
  if (!p_rectangles || p_rectangles.length === 0) {
    return null;
  }

  if (p_rectangles.length === 1) {
    return _balsamiq.default.clone(p_rectangles[0]);
  }

  return p_rectangles.slice(1).reduce(function (p_a, p_b) {
    return types.rectanglesUnion(p_a, p_b);
  }, p_rectangles[0]);
};

types.newRectangleFromCrop = function (p_cropProp) {
  const a = p_cropProp.split(",");
  const width = Math.max(1, +a[2]);
  const height = Math.max(1, +a[3]);
  return types.newRectangle(+a[0], +a[1], width, height);
};

types.calcCropOffset = function (p_cropRectangle, p_rect) {
  _balsamiq.default.assertRect(p_rect);

  _balsamiq.default.assertRect(p_cropRectangle);

  return types.newPoint(Math.round(-p_rect.width * p_cropRectangle.x / 1000), Math.round(-p_rect.height * p_cropRectangle.y / 1000));
};

types.calcCroppedSize = function (p_cropRectangle, p_size) {
  _balsamiq.default.existMulti("Invalid rect!", p_cropRectangle.width, p_cropRectangle.height, p_size.width, p_size.height);

  var width = Math.round(p_size.width * p_cropRectangle.width / 1000);
  var height = Math.round(p_size.height * p_cropRectangle.height / 1000);
  return types.newSize(width, height);
};

types.calcCroppedOrigin = function (p_cropRectangle, p_rect) {
  _balsamiq.default.assertRect(p_rect);

  _balsamiq.default.assertRect(p_cropRectangle);

  var x = Math.round(p_rect.x + p_rect.width * p_cropRectangle.x / 1000);
  var y = Math.round(p_rect.y + p_rect.height * p_cropRectangle.y / 1000);
  return types.newPoint(x, y);
};

types.calcCroppedRectangle = function (p_cropRectangle, p_rect) {
  _balsamiq.default.assertRect(p_rect);

  _balsamiq.default.assertRect(p_cropRectangle);

  var width = Math.round(p_rect.width * p_cropRectangle.width / 1000);
  var height = Math.round(p_rect.height * p_cropRectangle.height / 1000);
  var x = Math.round(p_rect.x + p_rect.width * p_cropRectangle.x / 1000);
  var y = Math.round(p_rect.y + p_rect.height * p_cropRectangle.y / 1000);
  return types.newRectangle(x, y, width, height);
};

types.calcCropProperty = function (p_croppedRectangle, p_uncroppedRectangle) {
  //Balsamiq.assert(types.rectangleContains(p_uncroppedRectangle, p_croppedRectangle));
  const w = Math.max(1, Math.round(p_croppedRectangle.width / p_uncroppedRectangle.width * 1000));
  const h = Math.max(1, Math.round(p_croppedRectangle.height / p_uncroppedRectangle.height * 1000));
  const x = Math.round((p_croppedRectangle.x - p_uncroppedRectangle.x) / p_uncroppedRectangle.width * 1000) || 0;
  const y = Math.round((p_croppedRectangle.y - p_uncroppedRectangle.y) / p_uncroppedRectangle.height * 1000) || 0;
  return `${x},${y},${w},${h}`;
};

types.isCropAll = function (p_cropRectangle) {
  _balsamiq.default.existMulti("Invalid crop rect", p_cropRectangle.width, p_cropRectangle.height, p_cropRectangle.x, p_cropRectangle.y);

  return p_cropRectangle.x === 0 && p_cropRectangle.y === 0 && p_cropRectangle.width === 1000 && p_cropRectangle.height === 1000;
};

types.calcUncroppedOrigin = function (p_cropRect, p_croppedBBox) {
  _balsamiq.default.assertRect(p_croppedBBox);

  let point = types.newPoint(p_croppedBBox.x - p_cropRect.x * p_croppedBBox.width / p_cropRect.width, p_croppedBBox.y - p_cropRect.y * p_croppedBBox.height / p_cropRect.height);
  return isNaN(point.x) && isNaN(point.y) ? {
    x: 0,
    y: 0
  } : point;
};

types.calcUncroppedRectangle = function (p_cropRect, p_croppedBBox) {
  _balsamiq.default.assertRect(p_croppedBBox);

  var x = Math.round(p_croppedBBox.x - p_cropRect.x * p_croppedBBox.width / p_cropRect.width) || 0;
  var y = Math.round(p_croppedBBox.y - p_cropRect.y * p_croppedBBox.height / p_cropRect.height) || 0;
  var w = Math.round(p_croppedBBox.width * 1000 / p_cropRect.width) || 0;
  var h = Math.round(p_croppedBBox.height * 1000 / p_cropRect.height) || 0;
  return types.newRectangle(x, y, w, h);
};

types.shrinkRectangle = function (rect, amount) {
  // width and height must be greater than 0 to avoid RSODs when drawing rectangles in controls
  let width = Math.max(1, rect.width - amount);
  let height = Math.max(1, rect.height - amount);
  return types.newRectangle(rect.x + amount / 2.0, rect.y + amount / 2.0, width, height);
};

types.paddedRectangle = function (rect, top, bottom, left, right) {
  // add a padding to rect. the values can be left undefined
  // like for the shrink rectangle the min value for the width and height should be 1 to avoid RSODs
  top = top || 0;
  bottom = bottom || 0;
  left = left || 0;
  right = right || 0;
  return types.newRectangle(rect.x + left, rect.y + top, Math.max(rect.width - left - right, 1), Math.max(rect.height - top - bottom, 1));
};

types.scaledRectangle = function (rect, scale) {
  _balsamiq.default.assertNumber(scale);

  return types.newRectangle(rect.x * scale, rect.y * scale, rect.width * scale, rect.height * scale);
};

types.newRectangleFromSize = function (p_size) {
  return types.newRectangle(0, 0, p_size.width, p_size.height);
};

types.testPointInRectangle = function (p_point, p_rectangle) {
  return p_point.x >= p_rectangle.x && p_point.x <= p_rectangle.x + p_rectangle.width && p_point.y >= p_rectangle.y && p_point.y <= p_rectangle.y + p_rectangle.height;
};

types.testRectangleIntersect = function (a, b) {
  return !(b.x > a.x + a.width || b.x + b.width < a.x || b.y > a.y + a.height || b.y + b.height < a.y);
};

types.testMultipleRectanglesIntersect = function (p_rectangle, p_rectangles) {
  var intersect = false;

  for (let rectIndex = 0; rectIndex < p_rectangles.length && !intersect; rectIndex += 1) {
    intersect = types.testRectangleIntersect(p_rectangles[rectIndex], p_rectangle);
  }

  return intersect;
};

types.rectangleContains = function (a, b) {
  // a contains b
  return b.x >= a.x && b.x <= a.x + a.width && b.x + b.width >= a.x && b.x + b.width <= a.x + a.width && b.y >= a.y && b.y <= a.y + a.height && b.y + b.height >= a.y && b.y + b.height <= a.y + a.height;
};

types.rectanglesIntersect = function (a, b) {
  var maxL = Math.max(a.x, b.x);
  var maxT = Math.max(a.y, b.y);
  var minR = Math.min(a.x + a.width, b.x + b.width);
  var minB = Math.min(a.y + a.height, b.y + b.height);
  return types.newRectangle(maxL, maxT, Math.max(minR - maxL, 0), Math.max(minB - maxT, 0));
};

types.rectangleMiddleX = function (rect) {
  _balsamiq.default.existMulti(rect.x, rect.width);

  return Math.round(rect.x + rect.width / 2);
};

types.rectangleMiddleY = function (rect) {
  _balsamiq.default.existMulti(rect.y, rect.height);

  return Math.round(rect.y + rect.height / 2);
};

types.rectangleRight = function (rect) {
  _balsamiq.default.existMulti(rect.x, rect.width);

  return rect.x + rect.width;
};

types.rectangleBottom = function (rect) {
  _balsamiq.default.existMulti(rect.y, rect.height);

  return rect.y + rect.height;
};

types.makeMaskRect = function (p_rect) {
  return {
    type: "rect",
    rect: p_rect
  };
};
/**
 * New Size object
 * @param {number} p_width
 * @param {number} p_height
 * @returns {Size}
 */


types.newSize = function (p_width, p_height) {
  _balsamiq.default.assertSizeNumber(p_width);

  _balsamiq.default.assertSizeNumber(p_height);

  return {
    width: p_width,
    height: p_height
  };
};

types.newAutoSize = function () {
  return types.newSize(this.AUTOSIZE, this.AUTOSIZE);
};
/**
 * Scale a Size object by a scale factor
 * @param {Size} size
 * @param {number} scale
 * @returns {Size}
 */


types.scaledSize = function (size, scale) {
  return types.newSize(size.width * scale, size.height * scale);
};
/**
 *
 * @param {number} x X
 * @param {number} y Y
 * @param {number} width Width
 * @param {number} height Height
 * @returns {Rect}
 */


types.newRectangle = function (x, y, width, height) {
  _balsamiq.default.assertNumber(x);

  _balsamiq.default.assertNumber(y);

  _balsamiq.default.assertSizeNumber(width);

  _balsamiq.default.assertSizeNumber(height);

  return {
    x,
    y,
    width,
    height
  };
};

types.copyRectangle = function (rect) {
  return _balsamiq.default.clone(rect);
};
/**
 * New Point object
 * @param {number} p_x X
 * @param {number} p_y Y
 * @returns {{ x: number, y: number }}
 */


types.newPoint = function (p_x, p_y) {
  return {
    x: p_x,
    y: p_y
  };
};
/**
 * New Text Size object
 * @param {number} p_width Width
 * @param {number} p_height Height, must be > 0
 * @param {number} p_ascender Ascender, must be > 0
 * @returns {{ width: number, height: number, ascender: number }}
 */


types.newTextSize = function (p_width, p_height, p_ascender) {
  _balsamiq.default.assertNumber(p_width);

  _balsamiq.default.assertPositiveNumber(p_height);

  _balsamiq.default.assertPositiveNumber(p_ascender);

  return {
    width: p_width,
    height: p_height,
    ascender: p_ascender
  };
};

types.propertyDataTypes = {
  string: {
    deserialize: function (p_value) {
      return p_value;
    }
  },
  int: {
    deserialize: function (p_value) {
      return parseInt(p_value, 10);
    }
  },
  bool: {
    deserialize: function (p_value) {
      return p_value === 'true';
    }
  },
  float01: {
    deserialize: function (p_value) {
      var value = parseFloat(p_value);

      _balsamiq.default.assert(value >= 0.0 && value <= 1.0);

      return value;
    }
  },
  object: {
    deserialize: function (p_value) {
      return p_value;
    }
  },
  enum: {
    deserialize: function (p_value) {
      return p_value;
    }
  }
}; // Used by size property

types.AUTOSIZE = -1;

types.isAutoSize = function (size) {
  return size.width === types.AUTOSIZE || size.height === types.AUTOSIZE;
};

types.ControlState = {
  NORMAL: "up",
  SELECTED: "selected",
  FOCUSED: "focused",
  DISABLED: "disabled",
  DISABLED_SELECTED: "disabledSelected",
  CLOSED: "closed",
  DISABLED_CLOSED: "disabledClosed",
  INDETERMINATE: "indeterminate",
  DISABLED_INDETERMINATE: "disabledIndeterminate",
  FOCUSED_CLOSED: "focusedClosed",
  isDisabledState: function (p_state) {
    return p_state === this.DISABLED || p_state === this.DISABLED_SELECTED || p_state === this.DISABLED_CLOSED || p_state === this.DISABLED_INDETERMINATE;
  },
  isSelectedState: function (p_state) {
    return p_state === this.SELECTED || p_state === this.DISABLED_SELECTED;
  },
  isClosedState: function (p_state) {
    return p_state === this.CLOSED || p_state === this.DISABLED_CLOSED || p_state === this.FOCUSED_CLOSED;
  },
  isIndeterminateState: function (p_state) {
    return p_state === this.INDETERMINATE || p_state === this.DISABLED_INDETERMINATE;
  },
  isFocusedState: function (p_state) {
    return p_state === this.FOCUSED || p_state === this.FOCUSED_CLOSED;
  }
};
types.TextAlign = {
  LEFT: "left",
  CENTER: "center",
  RIGHT: "right"
};
types.TextOrientation = {
  RIGHT: "right",
  UP: "up",
  DOWN: "down",
  SLANTED: "slanted"
};
types.Positions = {
  TOP: "top",
  BOTTOM: "bottom",
  LEFT: "left",
  RIGHT: "right",
  CENTER: "center"
};
types.Crop = {
  ALL: "0,0,1000,1000"
};
types.OnOffState = {
  ON: "on",
  OFF: 'off'
};
types.Orientations = {
  LANDSCAPE: "landscape",
  PORTRAIT: "portrait"
};
types.iOSDevice = {
  iPhone: "iPhone",
  iPad: "iPad"
};
types.StrokeBasic = {
  SOLID: "solid",
  DASHED: "dashed",
  DOTTED: "dotted"
};
types.Stroke = {
  SOLID: "solid",
  DASHED: "dashed",
  DOTTED: "dotted",
  BREAKLINE_BOTTOM: "break",
  BREAKLINE_TOP: "breakTop"
};
types.BREAKLINE_VERTICAL_OFFSET = 19;
types.TextMacro = {
  MOCKUP_NAME: "{mockup-name}"
};
types.ArrowCurvature = {
  CONVEX: -1,
  STRAIGHT: 0,
  CONCAVE: 1
};
types.IconSize = {
  xsmall: 16,
  small: 24,
  medium: 32,
  large: 48,
  xlarge: 64,
  xxlarge: 128
};
types.NO_SYMBOL_DEFINITION = "_NO_SYMBOL_DEFINITION_";
types.SymbolState = {
  LOADED: "loaded",
  BROKEN: "broken",
  LOADING: "loading"
};
types.ResourceState = {
  UNKNOWN: "unknown",
  PENDING: "pending",
  AVAILABLE: "available",
  NOT_AVAILABLE: "not_available"
};
types.ResourceType = {
  UNKNOWN: "unknown",
  BMML: "bmml",
  TEXT: "text",
  JSON: "json",
  IMAGE: "image"
};
types.Platform = {
  Cloud: "cloud",
  Desktop: "desktop",
  GoogleDrive: "google-drive",
  Confluence: "confluence",
  ConfluenceCloud: "confluenceCloud",
  Jira: "jira",
  JiraCloud: "jiraCloud"
}; //types.ResourceLocation = {
//    APP: "app",
//    SITE_ASSETS: "siteAssets",
//    PROJECT_ASSETS: "projectAssets",
//    MOCKUPS: "mockups",
//    URI: "uri"
//};
//types.SiteType = {
//    MYBALSAMIQ: "mybalsamiq",
//    DESKTOP: "desktop",
//    GDRIVE: "gdrive"
//};

types.LocationRule = {
  CLASSIC: "classic",
  MYBALSAMIQ: "mybalsamiq",
  BMPR: "bmpr",
  ARCHIVE: "archive",
  MEMORY: "memory",
  URI: "uri"
};
types.SrcLocation = {
  APP_ASSETS: "app",
  SITE_ASSETS: "siteAssets",
  PROJECT_ASSETS: "projectAssets",
  URI: "uri"
};

class SrcProperty {
  /**
   *
   * @param {String} p_location
   * @param {String} p_resourceID
   * @param {String=} p_subResourceID
   */
  constructor(p_location, p_resourceID, p_subResourceID) {
    _balsamiq.default.assertEnum(p_location, types.SrcLocation);

    _balsamiq.default.assertString(p_resourceID);

    if (p_subResourceID !== undefined) {
      _balsamiq.default.assertString(p_subResourceID);
    }

    this.location = p_location;
    this.resourceID = p_resourceID;
    this.subResourceID = p_subResourceID;
    this.resolvedUrl = ""; // To be filled
  }

}

types.SrcProperty = SrcProperty;
/**
 * 
 * @param {unknown[]} p_oldArray
 * @param {unknown[]} p_newArray
 * @param {(a: unknown[], b: unknown[]) => boolean | undefined} p_isEqualFunc
 * @returns {{ removed: unknown[], added: unknown[] }}
 */

types.diffArray = function (p_oldArray, p_newArray) {
  let p_isEqualFunc = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : _balsamiq.default.isEqual;

  var toRemove = _balsamiq.default.filter(p_oldArray, oldItem => {
    return !p_newArray.some(newItem => {
      return p_isEqualFunc(newItem, oldItem);
    });
  });

  var toAdd = _balsamiq.default.filter(p_newArray, newItem => {
    return !p_oldArray.some(oldItem => {
      return p_isEqualFunc(newItem, oldItem);
    });
  });

  return {
    removed: toRemove,
    added: toAdd
  };
};

types.equalArray = function (a, b) {
  let isEqualFunc = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : _balsamiq.default.isEqual;

  if (a.length !== b.length) {
    return false;
  }

  for (let i = 0; i < a.length; i += 1) {
    if (!isEqualFunc(a[i], b[i])) {
      return false;
    }
  }

  return true;
};
/**
 * Deep diff between two object, using lodash
 * @param  {Object} object Object compared
 * @param  {Object} base   Object to compare with
 * @return {Object}        Return a new object who represent the diff
 */


types.diffObject = function (object, base) {
  function changes(object, base) {
    let result = {};
    Object.entries(object).forEach(_ref => {
      let [key, value] = _ref;

      if (!_balsamiq.default.isEqual(value, base[key])) {
        result[key] = _balsamiq.default.isObject(value) && _balsamiq.default.isObject(base[key]) ? changes(value, base[key]) : value;
      }
    });
    return result;
  }

  return changes(object, base);
}; //If Object.create isn't already defined, we just do the simple shim, without the second argument,
//since that's all we need here


var object_create = Object.create;

if (typeof object_create !== 'function') {
  object_create = function (o) {
    function F() {}

    F.prototype = o;
    return new F();
  };
}

types.deepCopy = function (obj) {
  if (obj == null || typeof obj !== 'object') {
    return obj;
  }

  var ret; // Handle Date

  if (obj instanceof Date) {
    ret = new Date();
    ret.setTime(obj.getTime());
    return ret;
  } // Handle Array


  if (obj instanceof Array) {
    ret = [];

    for (var i = 0, len = obj.length; i < len; i++) {
      ret[i] = types.deepCopy(obj[i]);
    }

    return ret;
  } //make sure the returned object has the same prototype as the original


  ret = object_create(obj.constructor.prototype);

  for (var key in obj) {
    ret[key] = types.deepCopy(obj[key]);
  }

  return ret;
};

types.makeSubEnum = function (enumerated, keys) {
  var c = {};

  for (var key in keys) {
    c[key] = enumerated[key];
  }

  return c;
}; // Font face util
// --------------


types.fontFaceToFontFamily = function (fontFace) {
  // TODO check this code
  return fontFace; // if (fontFace === "Balsamiq Sans") {
  //     return fontFace;
  // } else {
  //     return fontFace.replace(/ /g, "-");
  // }
}; // Menus
// -----


types.MENU_ITEM_CLICK = "menuItemClick"; // Other stuff
// -----------

types.padNumber = function (number, width) {
  _balsamiq.default.assertNumber(number);

  _balsamiq.default.assertPositiveNumber(width);

  let r = '' + number;

  while (r.length < width) {
    r = '0' + r;
  }

  return r;
};

var _default = types;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/utils/async.js":
/*!********************************!*\
  !*** ./js/core/utils/async.js ***!
  \********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.queue = exports.filter = exports.eachParallel = exports.async = exports.any = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var async = function (generator) {
  var iteratorFunc = function (it, oldResult) {
    var {
      value,
      done
    } = it.next(oldResult);

    if (!done) {
      value.then(iteratorFunc.bind(null, it));
    }
  };

  return function () {
    var it = new generator();
    iteratorFunc(it);
  };
};

exports.async = async;

const asyncEachParallelIterator = (asyncfy, item, predicate, callback) => {
  asyncfy(() => {
    try {
      predicate(item, (err, result) => {
        callback(err, result);
      });
    } catch (err) {
      callback(err);
    }
  });
};
/**
 * An async utility function that executes a function on a series of items in parallel. It makes the predicate function async.
 * @param {function} asyncfy - A Function that takes a function and returns an async version of it
 * @param {*[]} items - A list of items that will be given as first argument to the predicate function
 * @param {function} predicate - A Function that takes an item and a callback as input, does some operations and must call the callback when it is done
 * @param {function} [finishCb=() => {}] - A Function that gets called after all items have been processed by the predicate function
 */


const eachParallel = function (asyncfy, items, predicate) {
  let finishCb = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : () => {};
  if (items.length === 0) return finishCb();
  const itemCount = items.length;
  let callbackCount = 0;
  const errors = [];
  const results = [];

  const callback = (err, result) => {
    callbackCount = callbackCount + 1;
    if (err) errors.push(err);
    if (result) results.push(result);

    if (callbackCount === itemCount) {
      return finishCb(errors.length > 0 ? errors : undefined, results.length > 0 ? results : undefined);
    }
  };

  items.forEach(item => asyncEachParallelIterator(asyncfy, item, predicate, callback));
};
/**
 * An function that makes the passed in functions asynchronous and then execute them in order, passing the results of the previous function to the next one
 * @param {function} asyncfy - A Function that takes a function and returns an async version of it
 * @param {function[]} funcs - List of functions that will be "asyncfied" and executed
 * @param {number} index - index of the function to be executed
 * @param {function} finishCb - Function that gets called when queue is empty or one element return an error
 * @param {function} progressCb - Function that gets called after every iteration
 * @param {any?} previousFuncResult - The result of the previous function. If it's not undefined it will be passed to next function in the list
 */


exports.eachParallel = eachParallel;

const asyncQueueIterator = (asyncfy, funcs, index, finishCb, progressCb, previousFuncResult) => {
  asyncfy(() => {
    const func = previousFuncResult === undefined ? funcs[index] : funcs[index].bind(null, previousFuncResult);
    func((err, funcResult) => {
      if (err) {
        return finishCb(err);
      }

      progressCb(index);

      if (index + 1 < funcs.length) {
        asyncQueueIterator(asyncfy, funcs, index + 1, finishCb, progressCb, funcResult);
      } else {
        finishCb(null, funcResult);
      }
    });
  });
};
/**
 * An async queue that accepts a list of functions and execute them in order
 * @param {function[]} funcs - List of functions that will be "asyncfied" and executed
 * @param {function} finishCb - Function that gets called when queue is empty or one element return an error
 * @param {Object} [options]
 * @param {function} [options.asyncfy=cb => cb()] - A Function that takes a function and executes it. Useful to asyncfy sync operations
 * @param {function} [options.progressCb=() => {}] - Function that gets called after every iteration
 */


const queue = function (funcs, finishCb) {
  let {
    progressCb = () => {},
    asyncfy = cb => cb()
  } = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  if (funcs.length === 0) return finishCb();

  _balsamiq.default.assertArrayOf(funcs, "function");

  asyncQueueIterator(asyncfy, funcs, 0, finishCb, progressCb);
};

exports.queue = queue;

const asyncFilterIterator = (asyncfy, item, predicate, callback) => {
  asyncfy(() => {
    try {
      predicate(item, result => {
        callback(!!result);
      });
    } catch (err) {
      callback(false);
    }
  });
};
/**
 * An async utility function that filters a series of items in parallel. It makes the predicate function async. It keeps the original order.
 * @param {function} asyncfy - A Function that takes a function and returns an async version of it
 * @param {any[]} items - A list of items that will be given as first argument to the predicate function
 * @param {function} predicate - A Function that takes an item and a callback as input, does some operations and must call the callback when it is done with true/false
 * @param {function} finishCb - A Function that gets called after all items have been processed by the predicate function returning the items that passed the predicate
 */


const filter = (asyncfy, items, predicate, finishCb) => {
  _balsamiq.default.assertArray(items);

  _balsamiq.default.assertFunction(predicate);

  if (items.length === 0) return finishCb([]);
  const itemCount = items.length;
  let callbackCount = 0;
  const filteredItems = {};
  const indexes = [];

  const callback = (index, result) => {
    callbackCount = callbackCount + 1;

    if (result === true) {
      filteredItems[index] = items[index];
      indexes.push(index);
    }

    if (callbackCount === itemCount) {
      indexes.sort();
      return finishCb(indexes.map(i => filteredItems[i]));
    }
  };

  items.forEach((item, index) => asyncFilterIterator(asyncfy, item, predicate, result => callback(index, result)));
};

exports.filter = filter;

const asyncAnyIterator = (asyncfy, items, index, predicate, finishCb) => {
  asyncfy(() => {
    try {
      const item = items[index];
      predicate(item, funcResult => {
        if (funcResult === true) {
          finishCb(true);
          return;
        }

        if (index + 1 < items.length) {
          asyncAnyIterator(asyncfy, items, index + 1, predicate, finishCb);
        } else {
          finishCb(false);
        }
      });
    } catch (err) {
      finishCb(err);
    }
  });
};
/**
 * An async utility function that checks if predicate returns truthy for any item of collection. It makes the predicate function async. Returns as soon as one is true.
 * @param {function} asyncfy - A Function that takes a function and returns an async version of it
 * @param {any[]} items - A list of items that will be given as first argument to the predicate function
 * @param {function} predicate - A Function that takes an item and a callback as input, does some operations and must call the callback when it is done with true/false
 * @param {function} finishCb - A Function that gets called with the either true or false
 */


const any = (asyncfy, items, predicate, finishCb) => {
  _balsamiq.default.assertArray(items);

  _balsamiq.default.assertFunction(predicate);

  if (items.length === 0) return finishCb(false);
  asyncAnyIterator(asyncfy, items, 0, predicate, finishCb);
};

exports.any = any;

/***/ }),

/***/ "./js/core/utils/base36.js":
/*!*********************************!*\
  !*** ./js/core/utils/base36.js ***!
  \*********************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.decode = decode;
exports.encode = encode;
const CHARACTER_SET = "0123456789abcdefghijklmnopqrstuvwxyz".split("");

function encode(integer) {
  if (integer === 0) {
    return '0';
  }

  var s = '';

  while (integer > 0) {
    s = CHARACTER_SET[integer % 36] + s;
    integer = Math.floor(integer / 36);
  }

  return s;
}

function decode(base36String) {
  var val = 0,
      base36Chars = base36String.split("").reverse();
  base36Chars.forEach(function (character, index) {
    val += CHARACTER_SET.indexOf(character) * Math.pow(36, index);
  });
  return val;
}

/***/ }),

/***/ "./js/core/utils/diff.js":
/*!*******************************!*\
  !*** ./js/core/utils/diff.js ***!
  \*******************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.getChangedGroupControlIDsFromDiff = exports.getChangedGroupControlIDs = exports.deepDiff = exports.VALUE_CHANGES = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const VALUE_CHANGES = {
  CREATED: "created",
  UPDATED: "updated",
  DELETED: "deleted",
  UNCHANGED: "unchanged"
};
exports.VALUE_CHANGES = VALUE_CHANGES;

const compareValues = (value1, value2) => {
  if (value1 === value2) {
    return VALUE_CHANGES.UNCHANGED;
  }

  if (_balsamiq.default.isDate(value1) && _balsamiq.default.isDate(value2) && value1.getTime() === value2.getTime()) {
    return VALUE_CHANGES.UNCHANGED;
  }

  if ('undefined' == typeof value1) {
    return VALUE_CHANGES.CREATED;
  }

  if ('undefined' == typeof value2) {
    return VALUE_CHANGES.DELETED;
  }

  return VALUE_CHANGES.UPDATED;
};

const isValue = obj => {
  return !_balsamiq.default.isObject(obj) && !_balsamiq.default.isArray(obj);
};

const deepDiff = (oldObj, newObj, arraySearchFieldPropNames) => {
  if (_balsamiq.default.isFunction(oldObj) || _balsamiq.default.isFunction(newObj)) {
    throw "Invalid argument. Function given, but was expecting an object";
  }

  if (isValue(oldObj) || isValue(newObj)) {
    const changeType = compareValues(oldObj, newObj);
    const changeValue = changeType === VALUE_CHANGES.UPDATED ? {
      prevValue: oldObj,
      newValue: newObj
    } : oldObj === undefined ? newObj : oldObj;
    return {
      type: changeType,
      data: changeValue
    };
  }

  const diff = {};

  for (let key in oldObj) {
    if (_balsamiq.default.isFunction(oldObj[key])) {
      continue;
    }

    let value2 = undefined;

    if (_balsamiq.default.isArray(oldObj) && arraySearchFieldPropNames != null) {
      const searchKeys = _lodash.default.pick(oldObj[key], arraySearchFieldPropNames);

      value2 = _lodash.default.find(newObj, searchKeys);
    } else if ('undefined' !== typeof newObj[key]) {
      value2 = newObj[key];
    }

    diff[key] = deepDiff(oldObj[key], value2, arraySearchFieldPropNames);
  }

  for (let key in newObj) {
    if (_balsamiq.default.isArray(newObj) && arraySearchFieldPropNames != null) {
      const searchKeys = _lodash.default.pick(newObj[key], arraySearchFieldPropNames);

      if (!_lodash.default.find(oldObj, searchKeys)) {
        diff["created-" + key] = deepDiff(undefined, newObj[key], arraySearchFieldPropNames);
      }

      continue;
    }

    if (_balsamiq.default.isFunction(newObj[key]) || 'undefined' !== typeof diff[key]) {
      continue;
    }

    diff[key] = deepDiff(undefined, newObj[key], arraySearchFieldPropNames);
  }

  return diff;
};

exports.deepDiff = deepDiff;

const hasBeenUpdated = node => {
  if (!node) return false;
  if (Array.isArray(node) && node.some(hasBeenUpdated)) return true;

  if (typeof node === 'object') {
    if (node.type && node.type !== VALUE_CHANGES.UNCHANGED) return true;
    if (!node.type && Object.values(node).some(hasBeenUpdated)) return true;
  }

  return false;
};

const getChangedGroupControlIDs = controls => Object.values(controls).map(control => [VALUE_CHANGES.DELETED, VALUE_CHANGES.CREATED].includes(control.type) ? {
  type: control.type,
  controlID: control.data.ID
} : hasBeenUpdated(control) ? {
  type: VALUE_CHANGES.UPDATED,
  controlID: control.ID.data
} : null).filter(x => x);

exports.getChangedGroupControlIDs = getChangedGroupControlIDs;

const diffControls = (oldControls, newControls) => deepDiff(oldControls, newControls, ["ID", "typeID"]);

const getChangedGroupControlIDsFromDiff = (oldMockup, newMockup) => {
  const oldControls = oldMockup.mockup.controls.control;
  const newControls = newMockup.mockup.controls.control;
  const diff = diffControls(oldControls, newControls);
  return getChangedGroupControlIDs(diff);
};

exports.getChangedGroupControlIDsFromDiff = getChangedGroupControlIDsFromDiff;

/***/ }),

/***/ "./js/core/utils/exectime.js":
/*!***********************************!*\
  !*** ./js/core/utils/exectime.js ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ExecTime = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class ExecTime {
  constructor(context) {
    _balsamiq.default.exists(context);

    _balsamiq.default.exists(context.host);

    this.context = context;
    this.times = {}; // id -> { name:string, startTime:integer, lapCount:integer, lastLapTime:integer }

    this.nextID = 0;
  }
  /**
   *
   * @param {string} name
   * @returns {string}
   */


  start(name) {
    _balsamiq.default.assertString(name);

    let id = "time-" + this.nextID;
    this.nextID += 1;
    let now = Date.now();
    this.times[id] = {
      name: name,
      startTime: now,
      lapCount: 0,
      lastLapTime: now
    };
    this.context.host.log(`time "${this.times[id].name}": started`);
    return id;
  }
  /**
   *
   * @param {string} id
   * @param {string} msg
   */


  lap(id, msg) {
    if (id === null) return;

    _balsamiq.default.assertString(id);

    _balsamiq.default.exists(this.times[id], 'ExecTime.lap: unknown id=' + id);

    msg = msg ? ', ' + msg : '';
    let now = Date.now();
    let lapTime = now - this.times[id].lastLapTime;
    this.times[id].lapCount += 1;
    this.times[id].lastLapTime = now;
    this.context.host.log(`time "${this.times[id].name}": lap ${this.times[id].lapCount}: ${formatTime(lapTime)}${msg}`);
  }
  /**
   *
   * @param {string} id
   */


  end(id) {
    if (id === null) return;

    _balsamiq.default.assertString(id);

    _balsamiq.default.exists(this.times[id], 'ExecTime.end: unknown id=' + id);

    let elapsedTime = Date.now() - this.times[id].startTime;
    this.context.host.log(`time "${this.times[id].name}": ended: ${formatTime(elapsedTime)}`);
    delete this.times[id];
  }

}
/**
 *
 * @param {number} t
 * @returns {string}
 */


exports.ExecTime = ExecTime;

function formatTime(t) {
  _balsamiq.default.assertNumber(t);

  if (t < 2000) {
    return `${t}ms`;
  } else {
    let s = (t / 1000).toFixed(2);
    return `${s}s`;
  }
}

/***/ }),

/***/ "./js/core/utils/fileutils.js":
/*!************************************!*\
  !*** ./js/core/utils/fileutils.js ***!
  \************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.sanitizeFileName = sanitizeFileName;
exports.saveFileToUserURL = saveFileToUserURL;
exports.saveUserFile = saveUserFile;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _locators = __webpack_require__(/*! core/model/locators */ "./js/core/model/locators.js");

var _urls = __webpack_require__(/*! core/urls */ "./js/core/urls.ts");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Remove illegal charaters for file names.
function sanitizeFileName(fileName) {
  return fileName.replace(/[/\\?%*:|"<>]/g, '_');
}

function copyUnsavedFileTo(context, projectLocator, fileURL, cb) {
  context.host.UnsavedFiles.copyUnsavedFileTo(projectLocator.url(), fileURL, copyRes => {
    if (copyRes.ok) {
      // remove unsaved file
      context.host.UnsavedFiles.removeUnsavedFile(projectLocator.url(), removeRes => {
        if (removeRes.ok) {
          const newProjectLocator = (0, _locators.makeFileProjectLocator)(fileURL, (0, _urls.decomposeURL)(fileURL).nameNoExtension, false);
          cb({
            ok: true,
            projectLocator: newProjectLocator
          });
        } else {
          context.host.log(`Error trying to delete unsavedFile ${projectLocator.url()}. Error: ${removeRes.error}`);
          cb(removeRes);
        }
      });
    } else {
      context.host.log(`Error trying to copy unsavedFile ${projectLocator.url()} to userFileURL: ${fileURL}. Error: ${copyRes.error}`);
      cb(copyRes);
    }
  });
}

function copyLocalFileToNewUserFile(context, projectLocator, fileURL, cb) {
  // autosave -> old userFile is synced then close and open a new window with new project
  // manual save -> open a new window with new project, then remove entry from registry
  context.host.LocalFiles.copyLocalFileToNewUserFile(projectLocator.url(), fileURL, res => {
    if (res.ok) {
      const newProjectLocator = (0, _locators.makeFileProjectLocator)(fileURL, (0, _urls.decomposeURL)(fileURL).nameNoExtension, false);
      cb({
        ok: true,
        projectLocator: newProjectLocator
      });
    } else {
      context.host.log(`Error trying to copy file to ${fileURL}. Error: ${res.error}`);
      cb(res);
    }
  });
}

function saveFileToUserURL(context, projectLocator, fileURL, cb) {
  _balsamiq.default.assertObject(context);

  _balsamiq.default.assertClass(projectLocator, _locators.ProjectLocator);

  _balsamiq.default.assertString(fileURL);

  _balsamiq.default.assertFunction(cb);

  if (projectLocator.isUnsavedFileProject()) {
    copyUnsavedFileTo(context, projectLocator, fileURL, cb);
  } else {
    copyLocalFileToNewUserFile(context, projectLocator, fileURL, cb);
  }
}

function saveUserFile(context, userFileURL, callback) {
  _balsamiq.default.assertObject(context);

  _balsamiq.default.assertString(userFileURL);

  _balsamiq.default.assertFunction(callback);

  context.host.LocalFiles.copyLocalFileTo(userFileURL, res => {
    if (!res.ok) {
      // retry after 1 sec
      context.host.timer(1000, () => {
        context.host.LocalFiles.copyLocalFileTo(userFileURL, callback);
      });
    } else {
      callback(res);
    }
  });
}

/***/ }),

/***/ "./js/core/utils/geom.js":
/*!*******************************!*\
  !*** ./js/core/utils/geom.js ***!
  \*******************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Rectangle = exports.Point = void 0;
exports.evaluateBezier = evaluateBezier;
exports.evaluateSegments = evaluateSegments;
exports.fromRelativePos = fromRelativePos;
exports.getBezierBoundingBox = getBezierBoundingBox;
exports.getBezierControlPoint = getBezierControlPoint;
exports.getBezierMinDistance = getBezierMinDistance;
exports.getBoundingBox = getBoundingBox;
exports.getPointLineDistance = getPointLineDistance;
exports.getPointSegmentDistance = getPointSegmentDistance;
exports.interpolate = interpolate;
exports.squashBezier = squashBezier;
exports.toRelativePos = toRelativePos;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const {
  newPoint,
  newRectangle
} = _types.default;
/** @namespace Geom */

/**
 * @memberOf Geom
 */

class Rectangle {
  static topLeft(p_rect) {
    return newPoint(p_rect.x, p_rect.y);
  }

  static toPath(p_rect) {
    const {
      x,
      y,
      width,
      height
    } = p_rect;
    return [newPoint(x, y), newPoint(x + width, y), newPoint(x + width, y + height), newPoint(x, y + height)];
  }

  static center(p_rect) {
    return newPoint(p_rect.x + p_rect.width / 2.0, p_rect.y + p_rect.height / 2.0);
  }

  static expand(rect, point) {
    const expandedRect = Rectangle.clone(rect);
    const isPointMoreLeft = point.x < rect.x;
    const isPointMoreRight = point.x > rect.x + rect.width;
    const isPointMoreHigh = point.y < rect.y;
    const isPointMoreLow = point.y > rect.y + rect.height;

    if (isPointMoreLeft) {
      expandedRect.x = point.x;
      expandedRect.width = rect.width + rect.x - point.x;
    } else if (isPointMoreRight) {
      expandedRect.width = point.x - rect.x;
    }

    if (isPointMoreHigh) {
      expandedRect.y = point.y;
      expandedRect.height = rect.height + rect.y - point.y;
    } else if (isPointMoreLow) {
      expandedRect.height = point.y - rect.y;
    }

    return expandedRect;
  }

  static clone(p_rect) {
    return newRectangle(p_rect.x, p_rect.y, p_rect.width, p_rect.height);
  }

}
/**
 * @memberOf Geom
 */


exports.Rectangle = Rectangle;

class Point {
  static len(p_pt) {
    return Math.sqrt(p_pt.x * p_pt.x + p_pt.y * p_pt.y);
  }

  static add(p_p0, p_p1) {
    return newPoint(p_p0.x + p_p1.x, p_p0.y + p_p1.y);
  }

  static sub(p_p0, p_p1) {
    return newPoint(p_p0.x - p_p1.x, p_p0.y - p_p1.y);
  }

  static dot(p_p0, p_p1) {
    return p_p0.x * p_p1.x + p_p0.y * p_p1.y;
  }

  static clone(p_pt) {
    return newPoint(p_pt.x, p_pt.y);
  }

  static equals(p_p0, p_p1) {
    return p_p0.x == p_p1.x && p_p0.y == p_p1.y;
  }

  static scale(p_pt, p_x, p_y) {
    if (p_y === undefined) {
      p_y = p_x;
    }

    return newPoint(p_pt.x * p_x, p_pt.y * p_y);
  }

  static rot(p_pt, p_degree) {
    const rad = p_degree * Math.PI / 180.0;
    return newPoint(p_pt.x * Math.cos(rad) - p_pt.y * Math.sin(rad), p_pt.x * Math.sin(rad) + p_pt.y * Math.cos(rad));
  }

  static dist(p_p1, p_p2) {
    return Math.sqrt(Math.pow(p_p1.x - p_p2.x, 2) + Math.pow(p_p1.y - p_p2.y, 2));
  }

}
/**
 * Compute the bounding box containing all the Point passed as args
 * @memberOf Geom
 */


exports.Point = Point;

function getBoundingBox(p0) {
  let minP = Point.clone(p0);
  let maxP = Point.clone(p0);

  for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    args[_key - 1] = arguments[_key];
  }

  for (let p of args) {
    minP.x = Math.min(minP.x, p.x);
    minP.y = Math.min(minP.y, p.y);
    maxP.x = Math.max(maxP.x, p.x);
    maxP.y = Math.max(maxP.y, p.y);
  }

  return newRectangle(minP.x, minP.y, maxP.x - minP.x, maxP.y - minP.y);
}
/**
 * Compute the bounding box of a Bezier
 * @memberOf Geom
 */


function getBezierBoundingBox(p_p0, p_p1, p_p2) {
  let t0x = (p_p0.x - p_p1.x) / (p_p0.x - 2 * p_p1.x + p_p2.x);
  let t0y = (p_p0.y - p_p1.y) / (p_p0.y - 2 * p_p1.y + p_p2.y); // Wrap t0 between 0 and 1

  t0x = Math.max(Math.min(t0x, 1), 0);
  t0y = Math.max(Math.min(t0y, 1), 0); // Handle the division by 0, occurring when p0 and p2 are aligned

  t0x = isNaN(t0x) ? 0.5 : t0x;
  t0y = isNaN(t0y) ? 0.5 : t0y; // Compute the point for t0

  let pt = newPoint(evaluateBezier(p_p0, p_p1, p_p2, t0x).x, evaluateBezier(p_p0, p_p1, p_p2, t0y).y);
  let maxP = newPoint(Math.max(p_p0.x, pt.x, p_p2.x), Math.max(p_p0.y, pt.y, p_p2.y));
  let minP = newPoint(Math.min(p_p0.x, pt.x, p_p2.x), Math.min(p_p0.y, pt.y, p_p2.y));
  return newRectangle(minP.x, minP.y, maxP.x - minP.x, maxP.y - minP.y);
}
/** Compute the position of p_pt with respect to p_p2-p_p0 */


function toRelativePos(p_p0, p_p2, p_pt) {
  let s = newPoint(p_p2.x - p_p0.x, p_p2.y - p_p0.y);
  let dS = 1 / (-s.x * s.x - s.y * s.y);
  let d = newPoint(p_pt.x - p_p0.x, p_pt.y - p_p0.y);
  return newPoint(dS * (-s.x * d.x - s.y * d.y), dS * (-s.y * d.x + s.x * d.y));
}
/** Inverse of 'toRelativePos' */


function fromRelativePos(p_p0, p_p2, p_sp) {
  const s = newPoint(p_p2.x - p_p0.x, p_p2.y - p_p0.y);
  const s1 = newPoint(s.y, -s.x);
  const res = interpolate(p_p0, newPoint(p_p0.x + s.x, p_p0.y + s.y), p_sp.x);
  return interpolate(res, newPoint(res.x + s1.x, res.y + s1.y), p_sp.y);
}
/** Fine the Point for t on a Bezier curve defined by p_p0, p_p1, p_p2*/


function evaluateBezier(p_p0, p_p1, p_p2, t) {
  return newPoint((1 - t) * (1 - t) * p_p0.x + 2 * (1 - t) * t * p_p1.x + t * t * p_p2.x, (1 - t) * (1 - t) * p_p0.y + 2 * (1 - t) * t * p_p1.y + t * t * p_p2.y);
}
/** Find the Point t (t between 0, 1) on a two segment line defined by p_p0, p_p1, p_p2, in t */


function evaluateSegments(p_p0, p_p1, p_p2, t) {
  let l1 = Point.len(Point.sub(p_p1, p_p0));
  let l2 = Point.len(Point.sub(p_p2, p_p1));
  let lt = l1 + l2;
  l1 = l1 / lt;
  l2 = l2 / lt;

  if (t < l1) {
    return interpolate(p_p0, p_p1, t / l1);
  } else {
    return interpolate(p_p1, p_p2, (t - l1) / l2);
  }
}
/** Combines linearly p_p0 and p_p1 respectively with params (1-t) and t */


function interpolate(p_p0, p_p1, t) {
  return newPoint(p_p0.x * (1 - t) + p_p1.x * t, p_p0.y * (1 - t) + p_p1.y * t);
}
/** Find the control point that defines a Bezier curve interpolating p_p0, p_p1, p_p2 */


function getBezierControlPoint(p_p0, p_p2, p_ip) {
  // handle a special case that would lead to a division by 0.
  if (Point.equals(p_ip, p_p0) || Point.equals(p_ip, p_p2)) {
    return interpolate(p_p0, p_p2, 0.5);
  }

  let t = 0.5;
  return newPoint((p_p0.x * (t - 1) / t + p_ip.x / (t - t * t) + t * p_p2.x / (t - 1)) / 2, (p_p0.y * (t - 1) / t + p_ip.y / (t - t * t) + t * p_p2.y / (t - 1)) / 2);
}

function getBezierMinDistance(p0, p1, p2, pt, steps) {
  var curvePts = [];

  for (let t = 0; t < 1; t += 1 / steps) {
    curvePts.push(evaluateBezier(p0, p1, p2, t));
  }

  var distances = curvePts.map(bezierPt => Point.dist(bezierPt, pt));
  return Math.min(...distances);
}

function squashBezier(p_p0, p_p1, p_p2, p_ptDiff) {
  var pc1 = getBezierControlPoint(p_p0, p_p2, p_p1); // Finds the *t* in which the derivative of the curve is 0

  var t0x = (p_p0.x - pc1.x) / (p_p0.x - 2 * pc1.x + p_p2.x);
  var t0y = (p_p0.y - pc1.y) / (p_p0.y - 2 * pc1.y + p_p2.y);
  var replaceX = isNaN(t0x) || t0x >= 1 || t0x <= 0;
  var replaceY = isNaN(t0y) || t0y >= 1 || t0y <= 0;

  var cp = _balsamiq.default.clone(pc1);

  if (!replaceX) {
    var ptx = evaluateBezier(p_p0, pc1, p_p2, t0x);
    ptx.x -= p_ptDiff.x;
    cp.x = (p_p0.x * (t0x - 1) / t0x + ptx.x / (t0x - t0x * t0x) + t0x * p_p2.x / (t0x - 1)) / 2;
  }

  if (!replaceY) {
    var pty = evaluateBezier(p_p0, pc1, p_p2, t0y);
    pty.y -= p_ptDiff.y;
    cp.y = (p_p0.y * (t0y - 1) / t0y + pty.y / (t0y - t0y * t0y) + t0y * p_p2.y / (t0y - 1)) / 2;
  } // Find the interpolated point


  var res = evaluateBezier(p_p0, cp, p_p2, 0.5);
  return res;
}

function getPointLineDistance(p0, p1, pt) {
  // https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line
  return Math.abs((p1.y - p0.y) * pt.x - (p1.x - p0.x) * pt.y + p1.x * p0.y - p1.y * p0.x) / Math.sqrt(Math.pow(p1.y - p0.y, 2) + Math.pow(p1.x - p0.x, 2));
}

function getPointSegmentDistance(p0, p1, pt) {
  // http://geomalgorithms.com/a02-_lines.html : dist_Point_to_Segment
  let v = Point.sub(p1, p0);
  let w = Point.sub(pt, p0);
  let c1 = Point.dot(w, v);

  if (c1 <= 0) {
    return Point.dist(pt, p0);
  }

  let c2 = Point.dot(v, v);

  if (c2 <= c1) {
    return Point.dist(pt, p1);
  }

  let b = c1 / c2;
  let pb = Point.add(p0, Point.scale(v, b));
  return Point.dist(pt, pb);
}

/***/ }),

/***/ "./js/core/utils/mockupsize.js":
/*!*************************************!*\
  !*** ./js/core/utils/mockupsize.js ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.clipPoint = clipPoint;
exports.clipRectangle = clipRectangle;
exports.moveRectangle = moveRectangle;

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _constants = __webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const MaxMockupArea = _types.default.newRectangle(0, 0, _constants.MaxMockupSize, _constants.MaxMockupSize);

function clipRectangle(rect) {
  return _types.default.rectanglesIntersect(rect, MaxMockupArea);
}

function clipPoint(point) {
  return _types.default.newPoint(Math.min(_constants.MaxMockupSize, Math.max(0, point.x)), Math.min(_constants.MaxMockupSize, Math.max(0, point.y)));
}

function moveRectangle(rect) {
  let x = rect.x;
  let y = rect.y;

  if (rect.width <= _constants.MaxMockupSize) {
    x = Math.max(0, Math.min(_constants.MaxMockupSize - rect.width, x));
  } else {
    x = 0;
  }

  if (rect.height <= _constants.MaxMockupSize) {
    y = Math.max(0, Math.min(_constants.MaxMockupSize - rect.height, y));
  } else {
    y = 0;
  }

  return _types.default.newRectangle(x, y, rect.width, rect.height);
}

/***/ }),

/***/ "./js/core/utils/resourcetree.js":
/*!***************************************!*\
  !*** ./js/core/utils/resourcetree.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ResourceTreeItem = exports.ResourceTree = exports.MAX_RESOURCE_LEVELS = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var _native = __webpack_require__(/*! core/native */ "./js/core/native.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const MAX_RESOURCE_LEVELS = 5;
exports.MAX_RESOURCE_LEVELS = MAX_RESOURCE_LEVELS;

const sortAscending = (a, b) => a - b;

const sortDescending = (a, b) => b - a;
/**
 *
 * @param {ResourceTreeItem} resourceTreeItem
 * @param {ResourceTreeItem} parentTreeItem
 * @returns {boolean}
 */


function hasCircularReferences(resourceTreeItem, parentTreeItem) {
  if (parentTreeItem == null) {
    return false;
  }

  const parentResourceIDs = [];

  while (parentTreeItem != null) {
    parentResourceIDs.push(parentTreeItem.resourceID);
    parentTreeItem = parentTreeItem.parent;
  }

  if (parentResourceIDs.includes(resourceTreeItem.resourceID)) {
    return true;
  }

  return false;
}
/**
 *
 * @typedef { { resourceID: string; parentID: string | null; order: number }[] } ResourceTreeData
 */


class ResourceTreeItem {
  constructor(resourceID, parentID, index, order) {
    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.assertNumber(index);

    this.resourceID = resourceID; // B3 projects might have a broken relationship see https://www.pivotaltracker.com/story/show/172331028

    this.parentID = parentID;

    if (parentID === resourceID) {
      _native.Host.log(`ResourceTreeItem: found a resource with same parentID equal to resourceID: ${resourceID}. Setting parentID to null`);

      this.parentID = null;
    }

    this.parent = null;
    this.children = [];
    this.index = index;
    this.order = order;
  }

  clone() {
    return new ResourceTreeItem(this.resourceID, this.parentID, this.index, this.order);
  }

  setParent(newParent) {
    if (newParent) {
      _balsamiq.default.existsClass(newParent, ResourceTreeItem);
    }

    if (hasCircularReferences(this, newParent)) {
      _native.Host.log(`ResourceTreeItem: found a resource with parentID or ancestorID equal to resourceID: ${this.resourceID}. Setting parentID to null`);

      newParent = null;
    }

    var oldParent = this.parent;

    if (oldParent) {
      oldParent.removeChild(this);
    }

    this.parent = newParent;
    this.parentID = newParent ? newParent.resourceID : null;

    if (newParent) {
      newParent._addChild(this);
    }
  }

  _addChild(item) {
    _balsamiq.default.existsClass(item, ResourceTreeItem);

    if (this.children.indexOf(item) === -1) {
      this.children.push(item);
      this.children.sort((a, b) => a.index - b.index);
      return true;
    }

    return false;
  }

  forEachChild(func) {
    this.children.forEach(func, this);
  }

  removeChild(item) {
    _balsamiq.default.existsClass(item, ResourceTreeItem);

    const index = this.children.indexOf(item);

    if (index !== -1) {
      this.children.splice(index, 1);
      return true;
    }

    return false;
  }

  getLevel() {
    if (!this.parent) {
      return 0;
    } else {
      return this.parent.getLevel() + 1;
    }
  }

  delete() {
    this.setParent(null);
  }

  hasChildren() {
    return this.children.length > 0;
  }

}

exports.ResourceTreeItem = ResourceTreeItem;

class ResourceTree {
  /**
   *
   * @param {ResourceTreeData} resources
   */
  constructor(resources) {
    _balsamiq.default.assertArray(resources);

    const resourceIDSet = {};
    let order = Number.NEGATIVE_INFINITY;
    let needsRebuilding = false;
    /** @property {ResourceTreeItem} */

    this.resources = resources.map((res, index) => {
      if (order === res.order) {
        needsRebuilding = true;
      }

      _balsamiq.default.assert(order <= res.order);

      order = res.order;
      resourceIDSet[res.resourceID] = 1;
      return new ResourceTreeItem(res.resourceID, res.parentID, index, res.order);
    });

    _balsamiq.default.assert(Object.keys(resourceIDSet).length === resources.length);

    this._calculateParents();

    if (needsRebuilding) {
      this._rebuildTreeOrder();
    }

    this.enforceCorrectOrder();
  }

  clone() {
    const new_resources = this.resources.map(resourceTreeItem => resourceTreeItem.clone());
    return new ResourceTree(new_resources);
  }
  /*
  Rationale: old projects from B3 may have parents defined after their children,
  causing an assertion to fail in this.selfCheck().
  */


  enforceCorrectOrder() {
    this.resources.forEach((resource, index) => {
      if (resource.parent && resource.parent.index >= index) {
        _native.Host.log("ResourceTree: fixed wrong order between parent and children.");

        resource.parent = null;
      }
    });
  }

  _arrayHasIncreasingIndexes(array) {
    var prevIndex = -1;
    var ok = true;

    for (let i = 0; i < array.length && ok; i += 1) {
      ok = array[i].index > prevIndex;
    }

    return ok;
  }

  _rebuildTreeOrder() {
    _balsamiq.default.assert(this.resources.length >= 2, "It shouldn't be needed to rebuild the tree if there are less then 2 resources");

    const resourcesToReorder = this.resources.slice(1);
    this.resources = this.resources.slice(0, 1);
    resourcesToReorder.forEach((resource, i) => {
      const targetIndex = i + 1;

      var [order] = this._calculateOrderForNewElementsAtIndex(targetIndex, 1);

      resource.order = order;
      this.resources.push(resource);
    });

    this._setInternalResourceIndexes(); // recheck for errors


    this.selfCheck();
  }

  _resetOrder() {
    this.resources[0].order = 1;

    this._rebuildTreeOrder();
  }

  selfCheck() {
    let prevOrder = -1;

    for (let i = 0; i < this.resources.length; ++i) {
      const resource = this.resources[i];

      if (resource.order >= Number.MAX_SAFE_INTEGER) {
        this._resetOrder();

        return;
      } // normally order should be grater than prevOrder
      // if it's equal => there is decimal precision issue => rebuild everything
      // if it's greater => we want to fail because it should not happen


      if (resource.order === prevOrder) {
        this._rebuildTreeOrder();

        return;
      }

      _balsamiq.default.assert(resource.order > prevOrder, "Tree order of elements must be ascending at index:" + i + " order:" + resource.order + " prev:" + prevOrder);

      _balsamiq.default.assert(resource.index === i, "Tree internal index must be coherent with element array position");

      _balsamiq.default.assert(resource.parent ? resource.parent.index < i : true, "Tree: parent must have lesser index than child");

      _balsamiq.default.assert(resource.children.length ? this._arrayHasIncreasingIndexes(resource.children) : true, "Tree: children must be ordered in increasing index order");

      prevOrder = resource.order;
    }
  }

  _setInternalResourceIndexes() {
    this.resources.forEach((res, index) => res.index = index);
    this.resources.forEach(res => res.children.sort((a, b) => a.index - b.index));
  }

  _createResourcesObj(resources) {
    var resourcesObj = {};

    for (let i = 0; i < resources.length; i += 1) {
      resourcesObj[resources[i].resourceID] = resources[i];
    }

    return resourcesObj;
  }

  _calculateParents() {
    var resourcesObj = this._createResourcesObj(this.resources);

    this.resources.forEach(res => {
      if (res.parentID) {
        let parent = resourcesObj[res.parentID]; //Balsamiq.assert(parent);

        res.setParent(parent);
      }
    });
  }

  _changeLevelToIndexes(indexArray, canChangeLevelFun, changeLevelFun) {
    _balsamiq.default.assertArray(indexArray);

    indexArray.sort(sortAscending);
    var toChange = [];
    var moveResourceIDs = {};

    for (let lev = 0; lev < MAX_RESOURCE_LEVELS; lev += 1) {
      indexArray.forEach(index => {
        let resLevel = this.getResourceLevel(index);

        if (resLevel === lev) {
          if (canChangeLevelFun(index, moveResourceIDs)) {
            toChange.push(index);
            moveResourceIDs[this.getResource(index).resourceID] = true;
          }
        }
      });

      if (toChange.length === indexArray.length) {
        break;
      }
    }

    toChange.forEach(index => changeLevelFun(index));
  }

  _getAllSiblings(index) {
    var resource = this.getResource(index);
    var allSiblings = [];

    if (resource && resource.parent) {
      allSiblings = resource.parent.children;
    } else {
      allSiblings = this.getMainLevelResources();
    }

    return allSiblings;
  }

  _getFollowingSiblings(index) {
    var siblings = [];

    this._getAllSiblings(index).forEach(sibling => {
      if (sibling.index > index) {
        siblings.push(sibling);
      }
    });

    return siblings;
  }

  _isIndentedRespectToPrev(index, levDiff) {
    _balsamiq.default.assert(index !== 0);

    _balsamiq.default.assertNumber(levDiff);

    return this.getResourceLevel(index) === this.getResourceLevel(index - 1) + levDiff;
  }

  _getPrevResourceWithSameLevel(index) {
    return this._getPrevResourceWithLevelDelta(index, 0);
  }

  _getPrevResourceWithLevelDelta(index, level, blackList, whiteList) {
    _balsamiq.default.assertNumber(index);

    _balsamiq.default.assertNumber(level);

    if (!blackList) {
      blackList = [];
    }

    var startResource = this.getResource(index);

    while (index > 0) {
      index -= 1;
      let isBlacklisted = blackList.indexOf(index) !== -1;
      let isWhitelisted = whiteList ? whiteList.indexOf(index) !== -1 : true;

      if (this.getResourceLevel(index) === startResource.getLevel() + level && !isBlacklisted && isWhitelisted) {
        return this.getResource(index);
      }
    }

    return null;
  }

  _searchForNewParent(oldParentIndex, blackList, whiteList) {
    var newParent = null;
    var lev = this.getResourceLevel(oldParentIndex);
    var delta = 0;

    while (!newParent && delta <= lev) {
      newParent = this._getPrevResourceWithLevelDelta(oldParentIndex, -delta, blackList, whiteList);
      delta += 1;
    }

    return newParent;
  }

  _removeResourcesAndKeepParents(indexArray) {
    _balsamiq.default.assertArray(indexArray);

    const isToRemove = function (index) {
      return indexArray.indexOf(index) !== -1;
    };

    var removed = [];
    indexArray.forEach(index => {
      // Search a new father for its orphans that won't be deleted
      var newParent = this._searchForNewParent(index, indexArray); // Get element children and change parent only if they are not being removed as well


      var ch = _balsamiq.default.clone(this.getResource(index).children);

      ch.forEach(child => {
        if (!isToRemove(child.index)) {
          child.setParent(newParent);
        }
      }); // Search new father for deleted resource

      var resource = this.getResource(index);

      if (resource.parent && !isToRemove(resource.parent.index)) {
        // Search for a new Parent only inside removed objects
        let newParent = this._searchForNewParent(resource.parent.index, null, indexArray);

        resource.setParent(newParent);
      }

      removed.push(resource);
    }); // Remove physically at the end so taht indexes are coherent

    for (let j = indexArray.length - 1; j >= 0; j -= 1) {
      this.resources.splice(indexArray[j], 1);
    }

    this._setInternalResourceIndexes();

    return removed;
  }

  _randomSign() {
    return Math.random() < 0.5 ? -1 : 1;
  }

  _calculateOrderForNewElementsAtIndex(index, numElementsToAdd) {
    var delta, startOrder;
    var result = [];

    if (index === 0) {
      if (this.resources.length > 0) {
        delta = this.resources[0].order / (numElementsToAdd + 1);
      } else {
        delta = 1;
      }

      startOrder = 0;
    } else if (index === this.resources.length) {
      delta = 10;
      startOrder = this.resources[this.resources.length - 1].order;
    } else {
      delta = (this.resources[index].order - this.resources[index - 1].order) / (numElementsToAdd + 1);
      startOrder = this.resources[index - 1].order;
    }

    for (let i = 1; i <= numElementsToAdd; i += 1) {
      let value = startOrder + delta * i + this._randomSign() * Math.random() * (delta / 4);
      result.push(value);
    }

    return result;
  }

  _getDescendantIndexes(resource) {
    if (resource.children.length > 0) {
      var descendants = [];
      resource.children.forEach(child => {
        descendants.push(child.index);
        descendants = descendants.concat(this._getDescendantIndexes(child));
      });
      return descendants;
    } else {
      return [];
    }
  }

  _findInsertionIndexFromParent(parentID, order) {
    _balsamiq.default.assertPositiveNumber(order);

    var index;
    var allSiblings;
    var parent = this.getResourceByID(parentID);

    if (parent) {
      allSiblings = parent.children;
    } else {
      allSiblings = this.getMainLevelResources();
    }

    var siblingIndex = 0;

    while (siblingIndex < allSiblings.length && allSiblings[siblingIndex].order < order) {
      siblingIndex += 1;
    }

    if (allSiblings.length > 0) {
      if (siblingIndex === 0) {
        index = allSiblings[0].index;
      } else if (siblingIndex === allSiblings.length) {
        let lastSibling = allSiblings[allSiblings.length - 1];
        index = lastSibling.index + this._getDescendantIndexes(lastSibling).length + 1;
      } else {
        index = allSiblings[siblingIndex].index;
      }
    } else {
      index = 0;
      parent = null;
    }

    return {
      index,
      parent
    };
  }

  getIndex(resource) {
    _balsamiq.default.existsClass(resource, ResourceTreeItem);

    return this.resources.indexOf(resource);
  }

  getResourceByID(resourceID) {
    var index = this.resources.findIndex(res => {
      return res.resourceID === resourceID;
    });

    if (index !== -1) {
      return this.resources[index];
    } else {
      return null;
    }
  }

  getResource(index) {
    _balsamiq.default.assertNumber(index);

    _balsamiq.default.assert(index >= 0);

    return this.resources[index]; //may be undefined
  }

  getResourceLevel(index) {
    return this.getResource(index).getLevel();
  }

  addPayloadToResources(keyName, values) {
    _balsamiq.default.assert(values.length === this.resources.length);

    this.resources.forEach((resource, index) => {
      resource[keyName] = values[index];
    });
  }

  removeLevelFromIndexes(indexArray) {
    this._changeLevelToIndexes(indexArray, this.canRemoveLevelFromIndex.bind(this), this.removeLevelFromIndex.bind(this));
  }

  addLevelToIndexes(indexArray) {
    this._changeLevelToIndexes(indexArray, this.canAddLevelToIndex.bind(this), this.addLevelToIndex.bind(this));
  }

  canAddLevelToIndex(index, moveResourceIDs) {
    if (index === 0 || this.getResourceLevel(index) >= MAX_RESOURCE_LEVELS - 1) return false;
    const prevResource = this.getResource(index - 1);
    if (!this._isIndentedRespectToPrev(index, +1) || this._isIndentedRespectToPrev(index, +1) && moveResourceIDs[prevResource.resourceID]) return true;
    return false;
  }

  addLevelToIndex(index) {
    _balsamiq.default.assertNumber(index); // Do not indent if first or if already indented respect to prev item


    if (index === 0 || this._isIndentedRespectToPrev(index, +1)) {
      return false;
    }

    var element = this.getResource(index);

    var newParent = this._getPrevResourceWithSameLevel(index);

    element.setParent(newParent); // MOD CHILDREN: make them children of the first sibling of original element!

    for (let i = element.children.length - 1; i >= 0; i -= 1) {
      element.children[i].setParent(newParent);
    }

    return true;
  }

  canRemoveLevelFromIndex(index) {
    _balsamiq.default.assertNumber(index);

    if (index === 0 || this.getResourceLevel(index) === 0) {
      return false;
    } else {
      return true;
    }
  }

  removeLevelFromIndex(index) {
    _balsamiq.default.assertNumber(index);

    if (index === 0 || this.getResourceLevel(index) === 0) {
      return false;
    } // Get immediately following siblings of node coming after node children


    var followingSiblings = this._getFollowingSiblings(index); // Change parent


    var res = this.getResource(index);
    res.setParent(res.parent.parent); // Immediately following siblings become my children!

    followingSiblings.forEach(sibling => sibling.setParent(res));
    return true;
  }
  /**
   *
   * @param {ResourceTreeItem[]=} array
   */


  getMainLevelResources(array) {
    if (!_balsamiq.default.isArray(array)) {
      array = this.resources;
    }

    return array.filter(res => {
      return res.getLevel() === 0;
    });
  }

  deleteResourceIDs(resourceIDs) {
    _balsamiq.default.assertArray(resourceIDs);

    var indexArray = resourceIDs.map(resourceID => {
      return this.getIndex(this.getResourceByID(resourceID));
    });
    indexArray.sort(sortDescending);
    indexArray.forEach(index => {
      this.deleteResourceAtIndex(index);
    });
  }

  deleteResourceAtIndex(index) {
    // Search for new father for its orphans
    var newParent = null;
    var lev = this.getResourceLevel(index);
    var delta = 0;

    while (!newParent && delta <= lev) {
      newParent = this._getPrevResourceWithLevelDelta(index, -delta);
      delta += 1;
    }

    var ch = _balsamiq.default.clone(this.getResource(index).children);

    ch.forEach(child => {
      child.setParent(newParent);
    });
    var resource = this.getResource(index);
    resource.delete();
    this.resources.splice(index, 1);

    this._setInternalResourceIndexes();

    this.selfCheck();
  }

  reorderResources(indexArray, targetIndex) {
    let {
      atSameLevelOfIndex,
      asChildOfIndex
    } = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    _balsamiq.default.assertArray(indexArray);

    _balsamiq.default.assertNumber(targetIndex);

    if (atSameLevelOfIndex != null) _balsamiq.default.assertNumber(atSameLevelOfIndex, "atSameLevelOfIndex must be the index of the resource at the level of which you want to reorder the targetIndex");
    if (asChildOfIndex != null) _balsamiq.default.assertNumber(asChildOfIndex, "asChildOfIndex must be the index of the parent resource at the level of which you want to reorder the targetIndex");

    _balsamiq.default.assert(!(atSameLevelOfIndex != null && asChildOfIndex != null), "atSameLevelOfIndex and asChildOfIndex are mutually exclusive");

    indexArray.sort(sortAscending);
    var deltaIndex = 0; // calculate insertion index after the removal of moving resources

    for (let i = 0; i <= indexArray.length; i += 1) {
      if (indexArray[i] < targetIndex) {
        deltaIndex += 1;
      }
    }

    let requiredParent = null;

    if (atSameLevelOfIndex != null) {
      const resourceReferenceForParent = this.getResource(atSameLevelOfIndex);
      requiredParent = resourceReferenceForParent ? resourceReferenceForParent.parent : null;
    } else if (asChildOfIndex != null) {
      const resourceReferenceForParent = this.getResource(asChildOfIndex);
      requiredParent = resourceReferenceForParent ? resourceReferenceForParent : null;
    }

    targetIndex -= deltaIndex; // Remove from tree the moving resources

    var removed = this._removeResourcesAndKeepParents(indexArray); // Calculate new order for moving resources


    var newOrders = this._calculateOrderForNewElementsAtIndex(targetIndex, indexArray.length);

    removed.forEach((res, index) => res.order = newOrders[index]); // Add removed resources at the right spot

    var nextResource = this.getResource(targetIndex);
    var currentIndex = targetIndex - 1;

    if (currentIndex < 0) {
      currentIndex = 0;
    }

    this.resources.splice(targetIndex, 0, ...removed); // Calculate parent for resources dropped in the middle of a tree!!!

    const newParentForRemoved = requiredParent ? requiredParent : nextResource ? nextResource.parent : null;
    this.getMainLevelResources(removed).forEach(removedResource => {
      removedResource.setParent(newParentForRemoved);
    }); // Restore internal indexes

    this._setInternalResourceIndexes(); // Check for errors


    this.selfCheck();
  }

  addResourcesWithParentID(resourceIDs, parentIDs, orders) {
    _balsamiq.default.assertArray(resourceIDs);

    _balsamiq.default.assertArray(parentIDs);

    _balsamiq.default.assertArray(orders);

    _balsamiq.default.assert(resourceIDs.length === parentIDs.length && parentIDs.length === orders.length);

    resourceIDs.forEach((resourceID, index) => {
      this.addResourceWithParentID(resourceID, parentIDs[index], orders[index]);
    });
  }
  /**
   *
   * @param {string[]} resourceIDs
   * @param {string|null} parentID
   */


  appendResourcesToParent(resourceIDs, parentID) {
    _balsamiq.default.assertArray(resourceIDs);

    const parent = this.getResourceByID(parentID);
    resourceIDs.forEach(resourceID => {
      let allSiblings;

      if (parent) {
        allSiblings = parent.children;
      } else {
        allSiblings = this.getMainLevelResources();
      }

      let index = -1;

      if (allSiblings.length > 0) {
        let lastSibling = allSiblings[allSiblings.length - 1];
        index = lastSibling.index + this._getDescendantIndexes(lastSibling).length + 1;
      } else {
        index = parent.index + 1;
      }

      let order;
      [order] = this._calculateOrderForNewElementsAtIndex(index, 1);
      const newResource = new ResourceTreeItem(resourceID, parent ? parent.resourceID : null, index, order);

      if (parent) {
        newResource.setParent(parent);
      }

      this.resources.splice(index, 0, newResource); // Restore internal indexes

      this._setInternalResourceIndexes(); // Check for errors


      this.selfCheck();
    });
  }

  addResourceWithParentID(resourceID, parentID, order) {
    _balsamiq.default.assertString(resourceID);

    _balsamiq.default.assertPositiveNumber(order);

    var {
      index,
      parent
    } = this._findInsertionIndexFromParent(parentID, order);

    [order] = this._calculateOrderForNewElementsAtIndex(index, 1);
    var newResource = new ResourceTreeItem(resourceID, parent ? parent.resourceID : null, index, order);

    if (parent) {
      newResource.setParent(parent);
    }

    this.resources.splice(index, 0, newResource); // Restore internal indexes

    this._setInternalResourceIndexes(); // Check for errors


    this.selfCheck();
  }

  addResourceAtIndex(resourceID, targetIndex) {
    _balsamiq.default.assertString(resourceID);

    var nextResource = this.getResource(targetIndex);

    var [order] = this._calculateOrderForNewElementsAtIndex(targetIndex, 1); // Calculate parent for resources dropped in the middle of a tree!!!


    var newParent = nextResource ? nextResource.parent : null;
    var newResource = new ResourceTreeItem(resourceID, newParent ? newParent.resourceID : null, targetIndex, order);

    if (newParent) {
      newResource.setParent(newParent);
    }

    this.resources.splice(targetIndex, 0, newResource); // Restore internal indexes

    this._setInternalResourceIndexes(); // Check for errors


    this.selfCheck();
  }

  cloneResourceSubTree(index, resourceIDGenerator) {
    _balsamiq.default.assertNumber(index); //Balsamiq.assert(this.getResourceLevel(index) === 0);


    var toClone = this.getResourceDescendantsByIndex(index, true);
    var clones = [];
    var resourceIDMap = {};
    toClone.forEach((descIndex, index) => {
      // Clone resource
      let resource = this.getResource(descIndex);
      let cloned = new ResourceTreeItem(resource.resourceID, resource.parentID, resource.index, resource.order); //cloned = Object.assign(cloned, resource);
      // Replace resourceID

      let newResourceID = resourceIDGenerator();
      resourceIDMap[resource.resourceID] = newResourceID;
      cloned.resourceID = newResourceID; // Reset parentship

      let originalParentID = resource.parentID;
      let newParentID;
      let newParent;

      if (resourceIDMap[originalParentID]) {
        newParentID = resourceIDMap[originalParentID];
        newParent = clones.find(clone => clone.resourceID === newParentID);
      } else {
        newParentID = originalParentID;
        newParent = this.getResourceByID(originalParentID);
      }

      cloned.children = [];
      cloned.index = index;
      cloned.setParent(newParent);
      clones.push(cloned);
    }); // Calculate new Orders

    var insertionIndex = Math.max(...toClone) + 1;

    var orderArray = this._calculateOrderForNewElementsAtIndex(insertionIndex, toClone.length);

    orderArray.forEach((newOrder, index) => clones[index].order = newOrder); // add to array

    this.resources.splice(insertionIndex, 0, ...clones); // Restore internal indexes

    this._setInternalResourceIndexes(); // Check for errors


    this.selfCheck();
    return clones.map(clone => clone.index);
  }
  /**
   *
   * @param {number} index
   * @param {boolean=} includeIndex
   * @returns { number[] }
   */


  getResourceDescendantsByIndex(index) {
    let includeIndex = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

    var res = this._getDescendantIndexes(this.getResource(index));

    if (includeIndex) {
      res.unshift(index);
    }

    return res;
  }

  hasChildren(index) {
    _balsamiq.default.assertNumber(index);

    var resource = this.getResource(index);
    return resource ? resource.children.length > 0 : false;
  }
  /**
   *
   * @param {(resource: ResourceTreeItem, index: number) => void} fun
   */


  forEachResource(fun) {
    this.resources.forEach(fun);
  }

  dump(log) {
    this.resources.forEach(res => {
      let indent = '';

      for (let i = 0; i < res.getLevel(); i++) {
        indent += '-';
      }

      log(indent + res.resourceID);
    });
  }

}

exports.ResourceTree = ResourceTree;

/***/ }),

/***/ "./js/core/utils/selectionutils.js":
/*!*****************************************!*\
  !*** ./js/core/utils/selectionutils.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.HandleTypes = exports.Handle = void 0;
exports.createHandle = createHandle;
exports.createRectFromSizeAndPos = createRectFromSizeAndPos;
exports.getBBoxPtForHandle = getBBoxPtForHandle;
exports.getDragHandleTypeFromDragPoints = getDragHandleTypeFromDragPoints;
exports.getRectHandleSizeAndPos = getRectHandleSizeAndPos;
exports.getSelectionRectangleDisplayItems = getSelectionRectangleDisplayItems;
exports.getSelectionRectangleDisplayItemsAndHandlesMS = getSelectionRectangleDisplayItemsAndHandlesMS;
exports.getSelectionRectangleHandlesDisplayList = getSelectionRectangleHandlesDisplayList;
exports.getSelectionRectangleHandlesForHitTest = getSelectionRectangleHandlesForHitTest;
exports.getSplitImageDisplayItemsAndHandles = getSplitImageDisplayItemsAndHandles;
exports.getSquareHandleSizeAndPos = getSquareHandleSizeAndPos;
exports.isArrowHandle = isArrowHandle;
exports.makeLockIconAndSelectionRectDisplayItems = makeLockIconAndSelectionRectDisplayItems;
exports.resizeBBoxFromHandle = resizeBBoxFromHandle;
exports.snapCoordinatesToHalfPixel = snapCoordinatesToHalfPixel;

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var Props = _interopRequireWildcard(__webpack_require__(/*! core/controls/properties */ "./js/core/controls/properties.js"));

var MockupCoords = _interopRequireWildcard(__webpack_require__(/*! core/viewcontrollers/mockupcoordinates */ "./js/core/viewcontrollers/mockupcoordinates.js"));

var _icon = _interopRequireDefault(__webpack_require__(/*! core/bricks/icon */ "./js/core/bricks/icon.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

var Path = _interopRequireWildcard(__webpack_require__(/*! core/graphics/path */ "./js/core/graphics/path.js"));

var UI = _interopRequireWildcard(__webpack_require__(/*! editor/ui_constants */ "./js/editor/ui_constants.js"));

var _cursors = __webpack_require__(/*! core/cursors */ "./js/core/cursors.js");

var _native = __webpack_require__(/*! core/native */ "./js/core/native.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const HandleTypes = {
  TOP_LEFT: 'topLeft',
  TOP: 'top',
  TOP_RIGHT: 'topRight',
  RIGHT: 'right',
  BOTTOM_RIGHT: 'bottomRight',
  BOTTOM: 'bottom',
  BOTTOM_LEFT: 'bottomLeft',
  LEFT: 'left',
  P0: 'p0',
  P1: 'p1',
  P2: 'p2',
  SPLIT_HORIZONTAL: 'SPLIT_HORIZONTAL',
  SPLIT_HORIZONTAL_BUTTON: 'SPLIT_HORIZONTAL_BUTTON'
};
exports.HandleTypes = HandleTypes;

function isArrowHandle(handleType) {
  const customHandles = [HandleTypes.P0, HandleTypes.P1, HandleTypes.P2];
  return customHandles.includes(handleType);
} // Start from top left, clockwise


function getSelectionRectangleHandlesDisplayList(bbox, resizeOption) {
  _balsamiq.default.existMulti(bbox.x, bbox.y, bbox.height, bbox.width);

  _balsamiq.default.assertEnum(resizeOption, Props.ResizeOptions);

  const size = _native.Host.runtime.isMobile() ? 12 : 4;
  const x = bbox.x;
  const y = bbox.y;
  const w = bbox.width;
  const h = bbox.height;
  const w2 = bbox.width / 2;
  const h2 = bbox.height / 2;

  switch (resizeOption) {
    case Props.ResizeOptions.BOTH:
      return [createHandle(HandleTypes.TOP_LEFT, getSquareHandleSizeAndPos(x - 1, y - 1, size)), createHandle(HandleTypes.TOP, getSquareHandleSizeAndPos(x + w2, y, size)), createHandle(HandleTypes.TOP_RIGHT, getSquareHandleSizeAndPos(x + w + 1, y - 1, size)), createHandle(HandleTypes.RIGHT, getSquareHandleSizeAndPos(x + w, y + h2, size)), createHandle(HandleTypes.BOTTOM_RIGHT, getSquareHandleSizeAndPos(x + w + 1, y + h + 1, size)), createHandle(HandleTypes.BOTTOM, getSquareHandleSizeAndPos(x + w2, y + h, size)), createHandle(HandleTypes.BOTTOM_LEFT, getSquareHandleSizeAndPos(x - 1, y + h + 1, size)), createHandle(HandleTypes.LEFT, getSquareHandleSizeAndPos(x, y + h2, size))];

    case Props.ResizeOptions.HORIZONTAL:
      return [createHandle(HandleTypes.RIGHT, getSquareHandleSizeAndPos(x + w, y + h2, size)), createHandle(HandleTypes.LEFT, getSquareHandleSizeAndPos(x, y + h2, size))];

    case Props.ResizeOptions.VERTICAL:
      return [createHandle(HandleTypes.TOP, getSquareHandleSizeAndPos(x + w2, y, size)), createHandle(HandleTypes.BOTTOM, getSquareHandleSizeAndPos(x + w2, y + h, size))];

    case Props.ResizeOptions.NO:
      return [];
  }
} // Start from top left, clockwise


function getSelectionRectangleHandlesForHitTest(bbox, resizeOption) {
  _balsamiq.default.existMulti(bbox.x, bbox.y, bbox.height, bbox.width);

  _balsamiq.default.assertEnum(resizeOption, Props.ResizeOptions);

  const isMobile = _native.Host.runtime.isMobile();

  const size = isMobile ? 24 : 6;
  const x = bbox.x;
  const y = bbox.y;
  const w = bbox.width;
  const h = bbox.height;
  const w2 = bbox.width / 2;
  const h2 = bbox.height / 2;

  switch (resizeOption) {
    case Props.ResizeOptions.BOTH:
      return [createHandle(HandleTypes.BOTTOM_RIGHT, getSquareHandleSizeAndPos(x + w, y + h, size), _cursors.CursorIDs.RESIZE_NW_SE), createHandle(HandleTypes.BOTTOM_LEFT, getSquareHandleSizeAndPos(x, y + h, size), _cursors.CursorIDs.RESIZE_NE_SW), createHandle(HandleTypes.TOP_RIGHT, getSquareHandleSizeAndPos(x + w, y, size), _cursors.CursorIDs.RESIZE_NE_SW), createHandle(HandleTypes.TOP_LEFT, getSquareHandleSizeAndPos(x, y, size), _cursors.CursorIDs.RESIZE_NW_SE), createHandle(HandleTypes.TOP, getRectHandleSizeAndPos(x + w2, y, isMobile ? size : w, size), _cursors.CursorIDs.RESIZE_N_S), createHandle(HandleTypes.RIGHT, getRectHandleSizeAndPos(x + w, y + h2, size, isMobile ? size : h), _cursors.CursorIDs.RESIZE_E_W), createHandle(HandleTypes.BOTTOM, getRectHandleSizeAndPos(x + w2, y + h, isMobile ? size : w, size), _cursors.CursorIDs.RESIZE_N_S), createHandle(HandleTypes.LEFT, getRectHandleSizeAndPos(x, y + h2, size, isMobile ? size : h), _cursors.CursorIDs.RESIZE_E_W)];

    case Props.ResizeOptions.HORIZONTAL:
      return [createHandle(HandleTypes.RIGHT, getRectHandleSizeAndPos(x + w, y + h2, size, isMobile ? size : h), _cursors.CursorIDs.RESIZE_E_W), createHandle(HandleTypes.LEFT, getRectHandleSizeAndPos(x, y + h2, size, isMobile ? size : h), _cursors.CursorIDs.RESIZE_E_W)];

    case Props.ResizeOptions.VERTICAL:
      return [createHandle(HandleTypes.TOP, getRectHandleSizeAndPos(x + w2, y, isMobile ? size : w, size), _cursors.CursorIDs.RESIZE_N_S), createHandle(HandleTypes.BOTTOM, getRectHandleSizeAndPos(x + w2, y + h, isMobile ? size : w, size), _cursors.CursorIDs.RESIZE_N_S)];

    case Props.ResizeOptions.NO:
      return [];
  }
}

function getSquareHandleSizeAndPos(x, y, size) {
  _balsamiq.default.assertNumber(x);

  _balsamiq.default.assertNumber(y);

  _balsamiq.default.assertNumber(size);

  return {
    center: {
      x,
      y
    },
    width: size,
    height: size
  };
}

function getRectHandleSizeAndPos(x, y, width, height) {
  _balsamiq.default.assertNumber(x);

  _balsamiq.default.assertNumber(y);

  _balsamiq.default.assertNumber(width);

  _balsamiq.default.assertNumber(height);

  return {
    center: {
      x,
      y
    },
    width,
    height
  };
}

class Handle {
  constructor(handleID, sizeAndPos, cursor) {
    _balsamiq.default.assertString(handleID);

    _balsamiq.default.existMulti(sizeAndPos.center, sizeAndPos.width, sizeAndPos.height);

    this.handleID = handleID;
    this.sizeAndPos = sizeAndPos;
    this.cursor = cursor;
  }

  clone() {
    return new Handle(this.handleID, _balsamiq.default.clone(this.sizeAndPos), this.cursor);
  }

  hitTest(p_pointMS, p_zoom) {
    _balsamiq.default.assertNumber(p_zoom);

    let distW, distH;

    switch (this.handleID) {
      case HandleTypes.BOTTOM:
      case HandleTypes.TOP:
        {
          distW = this.sizeAndPos.width / 2;
          distH = this.sizeAndPos.height / 2 / p_zoom;
          break;
        }

      case HandleTypes.LEFT:
      case HandleTypes.RIGHT:
        {
          distW = this.sizeAndPos.width / 2 / p_zoom;
          distH = this.sizeAndPos.height / 2;
          break;
        }

      default:
        {
          distW = this.sizeAndPos.width / 2 / p_zoom;
          distH = this.sizeAndPos.height / 2 / p_zoom;
        }
    }

    const testX = Math.abs(p_pointMS.x - this.sizeAndPos.center.x) < distW;
    const testY = Math.abs(p_pointMS.y - this.sizeAndPos.center.y) < distH;
    return testX && testY;
  }

  getRect() {
    return createRectFromSizeAndPos(this.sizeAndPos);
  }

  offset(x, y) {
    this.sizeAndPos.center.x += x;
    this.sizeAndPos.center.y += y;
  }

}

exports.Handle = Handle;

function createHandle(handleID, sizeAndPos, cursor) {
  return new Handle(handleID, sizeAndPos, cursor);
}

function getSelectionRectangleDisplayItems(_ref) {
  let {
    bbox,
    scale,
    borderColor,
    borderAlpha,
    fillColor,
    fillAlpha
  } = _ref;

  _balsamiq.default.exists(bbox);

  _balsamiq.default.assertPositiveNumber(scale);

  _balsamiq.default.assertNumber(borderColor);

  const STROKE_WIDTH = 1;
  const W = STROKE_WIDTH / 2;

  let scaledBBox = _types.default.scaledRectangle(bbox, scale);

  let adjustedBBox = _types.default.newRectangle(Math.floor(scaledBBox.x) + W, Math.floor(scaledBBox.y) + W, Math.round(scaledBBox.width) - STROKE_WIDTH, Math.round(scaledBBox.height) - STROKE_WIDTH);

  let items = [_displayitem.default.newRectangle(borderColor, 1, _balsamiq.default.isNumber(fillColor) ? fillColor : null, null, null, null, adjustedBBox)];

  if (_balsamiq.default.isNumber(borderAlpha)) {
    _displayitem.default.setBorderAlpha(items, borderAlpha);
  }

  if (_balsamiq.default.isNumber(fillAlpha)) {
    _displayitem.default.setAlpha(items, fillAlpha);
  }

  return items;
}

function getSelectionRectangleDisplayItemsAndHandlesMS(p_bboxMS, p_resizeOp, fillColor, isResizing) {
  const bboxMS = Object.assign({}, p_bboxMS);
  let displayItems;

  if (fillColor === undefined || fillColor === null) {
    displayItems = [_displayitem.default.newRectangle(0x0, 1, null, null, null, null, bboxMS)];
  } else {
    displayItems = [_displayitem.default.newRectangle(fillColor, 1, fillColor, null, null, null, bboxMS)];
    displayItems[0].alpha = 0.1;
    displayItems[0].borderAlpha = 0.3;
  }

  _displayitem.default.setOverlayRectScalableFixedBorderSize(displayItems[0]);

  const handles = getSelectionRectangleHandlesForHitTest(p_bboxMS, p_resizeOp);
  const handlesScaled = getSelectionRectangleHandlesDisplayList(bboxMS, p_resizeOp);
  const handleFillColor = isResizing ? 0x009e0f : 0xffffff;

  for (const handle of handlesScaled) {
    const bbox = handle.getRect();

    const displayItem = _displayitem.default.newRectangle(0, 1, handleFillColor, null, null, null, bbox);

    displayItem.cursor = handle.cursor;

    _displayitem.default.setOverlayRectFixedSize(displayItem);

    displayItems.push(displayItem);
  }

  return {
    displayItems,
    handles
  };
}
/**
 *
 * @param {import("core/types").Rect} imageBounds
 * @param {number} ySplitValue
 * @returns {displayItems: Array, handles: Array}
 */


function getSplitImageDisplayItemsAndHandles(imageBounds, imageBoundsMS, ySplitValue, cursorIsOver) {
  const splitValueMS = imageBoundsMS.height * ySplitValue / 1000;
  const splitValue = imageBounds.height * ySplitValue / 1000;
  let displayItems;

  const frameRectangleBounds = _types.default.newRectangle(Math.round(imageBounds.x), Math.round(imageBounds.y), imageBounds.width, imageBounds.height);

  displayItems = [_displayitem.default.newRectangle(0x0, 1, null, null, null, null, frameRectangleBounds)];

  _displayitem.default.setOverlayRectScalableFixedBorderSize(displayItems[0]);

  let boundsForLines = _types.default.newRectangle(imageBoundsMS.x + 1, imageBoundsMS.y, imageBoundsMS.width - 2, imageBoundsMS.height);

  [0, -1, +1].forEach(element => {
    let p = new Path.Path(imageBoundsMS.width, imageBoundsMS.height);
    p.moveTo(0, splitValueMS + element);
    p.lineTo(imageBoundsMS.width, splitValueMS + element);

    let splitImageHandle1 = _displayitem.default.newSVGPath(boundsForLines, p, 1, element === 0 ? UI.Colors.Black : UI.Colors.White, null);

    displayItems.push(splitImageHandle1);
  });
  const context = {
    skin: _displayitem.default.AppPathsAsset
  };
  let nativeResourceName;

  if (!cursorIsOver) {
    nativeResourceName = _displayitem.default.NativeResources.SPLIT_IMAGE_GRIP;
  } else {
    nativeResourceName = _displayitem.default.NativeResources.SPLIT_IMAGE_GRIP_HOVER;
  }

  const grip_width = _displayitem.default.NativeResourcesData[nativeResourceName].width;
  const grip_height = _displayitem.default.NativeResourcesData[nativeResourceName].height;
  const center_offset_x = grip_width / 2 - _displayitem.default.NativeResourcesData[nativeResourceName].center_x;
  const center_offset_y = grip_height / 2 - _displayitem.default.NativeResourcesData[nativeResourceName].center_y;

  const gripBoundsMS = _types.default.newRectangle(Math.round(imageBoundsMS.x + imageBoundsMS.width / 2 - grip_width / 2 + center_offset_x), Math.round(imageBoundsMS.y + splitValueMS - grip_height / 2 + center_offset_y), grip_width, grip_height);

  if (grip_width < imageBoundsMS.width * 0.6) {
    const gripIconBoundsMS = _types.default.newRectangle(gripBoundsMS.x, gripBoundsMS.y, grip_width, grip_height);

    let gripIcon = _displayitem.default.newNativeAsset(context, nativeResourceName, gripIconBoundsMS);

    displayItems = displayItems.concat(gripIcon);
  }

  const wholeGrabHandle = createHandle(HandleTypes.SPLIT_HORIZONTAL, {
    center: {
      x: imageBounds.x + imageBounds.width / 2,
      y: imageBounds.y + splitValue
    },
    width: imageBoundsMS.width,
    height: gripBoundsMS.height
  }, _cursors.CursorIDs.RESIZE_N_S);
  const dragButtonHandle = createHandle(HandleTypes.SPLIT_HORIZONTAL_BUTTON, {
    center: {
      x: imageBounds.x + imageBounds.width / 2,
      y: imageBounds.y + splitValue
    },
    width: grip_width,
    height: grip_height
  }, _cursors.CursorIDs.RESIZE_N_S);
  return {
    displayItems,
    handles: [wholeGrabHandle, dragButtonHandle]
  };
}

function createRectFromSizeAndPos(sizeAndPos) {
  return _types.default.newRectangle(sizeAndPos.center.x - sizeAndPos.width / 2, sizeAndPos.center.y - sizeAndPos.height / 2, sizeAndPos.width, sizeAndPos.height);
}

function snapCoordinatesToHalfPixel(geom) {
  geom.x = Math.floor(geom.x) + 0.5;
  geom.y = Math.floor(geom.y) + 0.5;

  if (Number.isFinite(geom.width)) {
    geom.width = Math.round(geom.width);
  }

  if (Number.isFinite(geom.height)) {
    geom.height = Math.round(geom.height);
  }

  return geom;
}

function resizeBBoxLeft(pt, bbox, minW) {
  var bboxRight = bbox.x + bbox.width;

  if (pt.x < 0) {
    pt.x = 0;
  }

  if (bbox.x + bbox.width - pt.x < minW) {
    pt.x = bbox.x + bbox.width - minW;
  }

  bbox.width = bbox.x + bbox.width - pt.x;
  bbox.x = pt.x;

  _balsamiq.default.assert(bboxRight === bbox.width + bbox.x);
}

function resizeBBoxTop(pt, bbox, minH) {
  var bboxBottom = bbox.y + bbox.height;

  if (pt.y < 0) {
    pt.y = 0;
  }

  if (bbox.y + bbox.height - pt.y < minH) {
    pt.y = bbox.y + bbox.height - minH;
  }

  bbox.height = bbox.y + bbox.height - pt.y;
  bbox.y = pt.y;

  _balsamiq.default.assert(bboxBottom === bbox.y + bbox.height);
}

function resizeBBoxRight(pt, bbox, minW) {
  if (pt.x - bbox.x < minW) {
    pt.x = bbox.x + minW;
  }

  bbox.width = pt.x - bbox.x;
}

function resizeBBoxBottom(pt, bbox, minH) {
  if (pt.y - bbox.y < minH) {
    pt.y = bbox.y + minH;
  }

  bbox.height = pt.y - bbox.y;
}

function resizeTopLeftKeepingRatio(pt1, bbox, ratio, minH, minW) {
  var pt = _balsamiq.default.clone(pt1);

  var bboxBottom = bbox.y + bbox.height;
  var bboxRight = bbox.x + bbox.width; // respect minW

  if (pt.x > bboxRight - minW) {
    pt.x = bboxRight - minW;
  } // no negative X


  if (pt.x < 0) {
    pt.x = 0;
  } // no negative Y


  if (bboxBottom - (bboxRight - pt.x) / ratio < 0) {
    pt.x = bboxRight - ratio * bboxBottom;
  } // respect minH


  if ((bboxRight - pt.x) / ratio < minH) {
    pt.x = bboxRight - minH * ratio;
  }

  var newX = pt.x;
  var newW = bboxRight - newX;
  var newH = newW / ratio;
  var newY = bboxBottom - newH;
  bbox.x = newX;
  bbox.y = newY;
  bbox.width = newW;
  bbox.height = newH;
}

function resizeBottomRightKeepingRatio(pt1, bbox, ratio, minH, minW) {
  var pt = _balsamiq.default.clone(pt1); // respect minW


  if (pt.x - bbox.x < minW) {
    pt.x = bbox.x + minW;
  } // respect minH


  if ((pt.x - bbox.x) / ratio < minH) {
    pt.x = minH * ratio + bbox.x;
  }

  var newW = pt.x - bbox.x;
  var newH = newW / ratio;
  bbox.height = newH;
  bbox.width = newW;
}

function resizeBottomLeftKeepingAspect(pt1, bbox, ratio, minH, minW) {
  var pt = _balsamiq.default.clone(pt1);

  var bboxRight = bbox.x + bbox.width; // respect minW

  if (pt.x > bboxRight - minW) {
    pt.x = bboxRight - minW;
  } // no negative X


  if (pt.x < 0) {
    pt.x = 0;
  } // respect minH


  if ((bboxRight - pt.x) / ratio < minH) {
    pt.x = bboxRight - minH * ratio;
  }

  var newX = pt.x;
  var newW = bboxRight - newX;
  var newH = newW / ratio;
  bbox.height = newH;
  bbox.width = newW;
  bbox.x = newX;
}

function resizeTopRightKeepingAspect(pt1, bbox, ratio, minH, minW) {
  var pt = _balsamiq.default.clone(pt1);

  var bboxBottom = bbox.y + bbox.height; // no negative Y

  if (bboxBottom - (pt.x - bbox.x) / ratio < 0) {
    pt.x = ratio * bboxBottom + bbox.x;
  } // respect minW


  if (pt.x - bbox.x < minW) {
    pt.x = bbox.x + minW;
  } // respect minH


  if ((pt.x - bbox.x) / ratio < minH) {
    pt.x = minH * ratio + bbox.x;
  }

  var newW = pt.x - bbox.x;
  var newH = newW / ratio;
  var newY = bboxBottom - newH;
  bbox.height = newH;
  bbox.width = newW;
  bbox.y = newY;
}

function resizeBBoxFromHandle(handle, newPt, origBBox) {
  let minW = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : 10;
  let minH = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : 10;
  let keepRatio = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : false;
  var ratio = origBBox.width / origBBox.height;
  var bbox = Object.assign({}, origBBox);
  var pt = Object.assign({}, newPt);

  switch (handle.handleID) {
    case HandleTypes.TOP_LEFT:
      {
        if (keepRatio) {
          resizeTopLeftKeepingRatio(pt, bbox, ratio, minH, minW);
        } else {
          resizeBBoxLeft(pt, bbox, minW);
          resizeBBoxTop(pt, bbox, minH);
        }

        break;
      }

    case HandleTypes.TOP:
      {
        resizeBBoxTop(pt, bbox, minH);
        break;
      }

    case HandleTypes.TOP_RIGHT:
      {
        if (keepRatio) {
          resizeTopRightKeepingAspect(pt, bbox, ratio, minH, minW);
        } else {
          resizeBBoxTop(pt, bbox, minH);
          resizeBBoxRight(pt, bbox, minW);
        }

        break;
      }

    case HandleTypes.RIGHT:
      {
        resizeBBoxRight(pt, bbox, minW);
        break;
      }

    case HandleTypes.BOTTOM_RIGHT:
      {
        if (keepRatio) {
          resizeBottomRightKeepingRatio(pt, bbox, ratio, minH, minW);
        } else {
          resizeBBoxRight(pt, bbox, minW);
          resizeBBoxBottom(pt, bbox, minH);
        }

        break;
      }

    case HandleTypes.BOTTOM:
      {
        resizeBBoxBottom(pt, bbox, minH);
        break;
      }

    case HandleTypes.BOTTOM_LEFT:
      {
        if (keepRatio) {
          resizeBottomLeftKeepingAspect(pt, bbox, ratio, minH, minW);
        } else {
          resizeBBoxLeft(pt, bbox, minW);
          resizeBBoxBottom(pt, bbox, minH);
        }

        break;
      }

    case HandleTypes.LEFT:
      {
        resizeBBoxLeft(pt, bbox, minW);
        break;
      }
  }

  bbox.height = Math.round(bbox.height);
  bbox.width = Math.round(bbox.width);
  bbox.x = Math.round(bbox.x);
  bbox.y = Math.round(bbox.y);
  return bbox;
}

function getBBoxPtForHandle(handle, bbox) {
  _balsamiq.default.assertClass(handle, Handle);

  _balsamiq.default.assertRect(bbox);

  _balsamiq.default.assertEnum(handle.handleID, HandleTypes);

  var res;

  switch (handle.handleID) {
    case HandleTypes.TOP:
      res = _types.default.newPoint(bbox.x + bbox.width / 2, bbox.y);
      break;

    case HandleTypes.TOP_RIGHT:
      res = _types.default.newPoint(bbox.x + bbox.width, bbox.y);
      break;

    case HandleTypes.RIGHT:
      res = _types.default.newPoint(bbox.x + bbox.width, bbox.y + bbox.height / 2);
      break;

    case HandleTypes.BOTTOM_RIGHT:
      res = _types.default.newPoint(bbox.x + bbox.width, bbox.y + bbox.height);
      break;

    case HandleTypes.BOTTOM:
      res = _types.default.newPoint(bbox.x + bbox.width / 2, bbox.y + bbox.height);
      break;

    case HandleTypes.BOTTOM_LEFT:
      res = _types.default.newPoint(bbox.x, bbox.y + bbox.height);
      break;

    case HandleTypes.LEFT:
      res = _types.default.newPoint(bbox.x, bbox.y + bbox.height / 2);
      break;

    case HandleTypes.TOP_LEFT:
      res = _types.default.newPoint(bbox.x, bbox.y);
      break;
  }

  return res;
}

function getDragHandleTypeFromDragPoints(startDragPt, endDragPt) {
  // Dragging Right
  if (endDragPt.x > startDragPt.x) {
    // Down
    if (endDragPt.y > startDragPt.y) {
      // Bottom right
      return HandleTypes.BOTTOM_RIGHT;
    } else {
      return HandleTypes.TOP_RIGHT;
    }
  } else {
    // Dragging left
    if (endDragPt.y > startDragPt.y) {
      // Bottom left
      return HandleTypes.BOTTOM_LEFT;
    } else {
      return HandleTypes.TOP_LEFT;
    }
  }
}

function makeLockIconAndSelectionRectDisplayItems(control, lockRectSize, lockIconSize, zoom, useMarkup) {
  _balsamiq.default.assertPositiveNumber(lockRectSize);

  _balsamiq.default.assertPositiveNumber(zoom);

  _balsamiq.default.assertBool(useMarkup);

  if (control.isMarkup() && !useMarkup) {
    return [];
  } // Lock Icon


  let bbox = MockupCoords.groupToMockupSpace(control.getUserBBox(true), control.parent); //we need the bbox that includes the (possibly) hidden markup controls

  bbox = _types.default.scaledRectangle(bbox, zoom);
  let iconData = new _icon.default.IconData(new _types.default.SrcProperty(_types.default.SrcLocation.APP_ASSETS, UI.LockedControls.LockIconFAName), lockIconSize, 0);
  let iconXDelta = lockIconSize + (lockRectSize - lockIconSize) / 2;
  let iconYDelta = -(lockRectSize - lockIconSize) / 2;

  let displayItems = _icon.default.makeDisplayItemsFromIconData(control, iconData, bbox.x - iconXDelta, bbox.y - iconYDelta, UI.LockedControls.LockIconBackgoundColor);

  _displayitem.default.setDisplayItemsBorder(displayItems, UI.LockedControls.LockIconBorderWidth, UI.LockedControls.LockIconBorderColor); // Dotted rectangle around individual control and lock icon
  // values given to Path.Path must be positive numbers


  let smallestPosNum = Number.MIN_VALUE;
  let p = new Path.Path(Math.max(bbox.width, smallestPosNum), Math.max(bbox.height, smallestPosNum));
  p.moveTo(-lockRectSize, 0);
  p.lineTo(bbox.width, 0);
  p.lineTo(bbox.width, bbox.height);
  p.lineTo(0, bbox.height);
  p.lineTo(0, lockRectSize);
  p.lineTo(-lockRectSize, lockRectSize);
  p.close();

  let dotted = _displayitem.default.newSVGPath(bbox, p, UI.LockedControls.SelectionRectBorderWidth, UI.LockedControls.SelectionRectColor, null, UI.LockedControls.SelectionRectDashArray);

  displayItems.push(dotted);
  return displayItems;
}

/***/ }),

/***/ "./js/core/version.js":
/*!****************************!*\
  !*** ./js/core/version.js ***!
  \****************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";
/* global BUILD_APPLICATION_NAME, BUILD_NUMBER, EDITOR_BUILD_NUMBER */


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.getEditorBuildNumber = exports.getBuildNumber = void 0;

var getBuildNumber = function () {
  if (true) {
    return "macos-exporter" + '-' + "0";
  } else {}
};

exports.getBuildNumber = getBuildNumber;

var getEditorBuildNumber = function () {
  if (false) {} else {
    return null;
  }
};

exports.getEditorBuildNumber = getEditorBuildNumber;

/***/ }),

/***/ "./js/core/viewcontrollers/constants.js":
/*!**********************************************!*\
  !*** ./js/core/viewcontrollers/constants.js ***!
  \**********************************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ViewControllerActionTypes = void 0;
const ViewControllerActionTypes = {
  User: 'user',
  // originated by the user, can be ignored on certain conditions
  Internal: 'internal' // directed to the Core part of the viewcontroller, not ignored

};
exports.ViewControllerActionTypes = ViewControllerActionTypes;

/***/ }),

/***/ "./js/core/viewcontrollers/mockupcoordinates.js":
/*!******************************************************!*\
  !*** ./js/core/viewcontrollers/mockupcoordinates.js ***!
  \******************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.controlSpaceToMockupSpace = controlSpaceToMockupSpace;
exports.controlSpaceToParentSpace = controlSpaceToParentSpace;
exports.getControlOffset = getControlOffset;
exports.getGroupOffset = getGroupOffset;
exports.groupToMockupSpace = groupToMockupSpace;
exports.mockupSpaceToControlSpace = mockupSpaceToControlSpace;
exports.mockupToGroupSpace = mockupToGroupSpace;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function mockupSpaceToControlSpace(p_point, p_control) {
  var groupCoords = mockupToGroupSpace(p_point, p_control.parent);
  var origin = p_control.getActualProperty('origin');
  const cropOffset = p_control.getCropOffset();

  var controlCoords = _types.default.newPoint(groupCoords.x - origin.x - cropOffset.x, groupCoords.y - origin.y - cropOffset.y);

  return controlCoords;
}

function controlSpaceToMockupSpace(p_point, p_control) {
  var origin = p_control.getActualProperty('origin');

  var groupCoords = _types.default.newPoint(p_point.x + origin.x, p_point.y + origin.y);

  var mockupCoords = groupToMockupSpace(groupCoords, p_control.parent);
  return mockupCoords;
}

function controlSpaceToParentSpace(p_point, p_control) {
  _balsamiq.default.exists(p_control);

  _balsamiq.default.assertPoint(p_point);

  var origin = p_control.getActualProperty('origin');

  var groupCoords = _types.default.newPoint(p_point.x + origin.x, p_point.y + origin.y);

  return groupCoords;
}

function mockupToGroupSpace(p_point, p_group) {
  _balsamiq.default.exists(p_group);

  _balsamiq.default.assertPoint(p_point);

  var offset = getGroupOffset(p_group);

  var res = _balsamiq.default.clone(p_point);

  res.x -= offset.x;
  res.y -= offset.y;
  return res;
}

function groupToMockupSpace(p_point, p_group) {
  _balsamiq.default.exists(p_group);

  _balsamiq.default.assertPoint(p_point);

  var offset = getGroupOffset(p_group);

  var res = _balsamiq.default.clone(p_point);

  res.x += offset.x;
  res.y += offset.y;
  return res;
}

function getGroupOffset(p_group) {
  var pt = _types.default.newPoint(0, 0);

  while (p_group) {
    let origin = p_group.getActualProperty("origin");
    pt.x += origin.x;
    pt.y += origin.y;

    if (p_group.hasCropData()) {
      let cropOffset = p_group.getCropOffset();
      pt.x += cropOffset.x;
      pt.y += cropOffset.y;
    }

    p_group = p_group.parent;
  }

  return pt;
}

function getControlOffset(p_control) {
  var origin = p_control.getActualProperty('origin');
  var offset = groupToMockupSpace(origin, p_control.parent);
  return offset;
}

/***/ }),

/***/ "./js/core/viewcontrollers/operationprogressviewcontroller.js":
/*!********************************************************************!*\
  !*** ./js/core/viewcontrollers/operationprogressviewcontroller.js ***!
  \********************************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const ViewController = __webpack_require__(/*! core/viewcontrollers/viewcontroller */ "./js/core/viewcontrollers/viewcontroller.ts");

const Actions = {
  CANCEL: "cancel"
};
const KeyNames = {};
const States = {
  CANCELLABLE: "cancellable",
  // bool
  CANCELLED: "cancelled",
  // bool
  CURRENT_STEP: "currentStep",
  // int
  MAX_STEPS: "maxSteps",
  // int
  OPERATION_NAME: "operationName",
  // string
  MESSAGE: "message",
  // string (nullable)
  QUOTE: "quote",
  // object (nullable) -- DEPRECATED
  CANCEL_BUTTON_LABEL: "cancelButtonLabel" // string

};

class OperationProgressViewController extends ViewController.ViewController {
  init(options) {
    this.state[States.CURRENT_STEP] = 0;
    this.state[States.MAX_STEPS] = 0;
    this.state[States.OPERATION_NAME] = '';
    this.state[States.MESSAGE] = null;
    this.state[States.QUOTE] = null;
    this.state[States.CANCELLABLE] = false;
    this.state[States.CANCELLED] = false;
    this.state[States.QUOTE] = null;
    this.cancelCallback = null;
    this.state[States.CANCEL_BUTTON_LABEL] = options.cancelButtonLabel || "Cancel";
  }

  cleanUp() {
    super.cleanUp();
  }

  action(action, params) {
    switch (action) {
      case Actions.CANCEL:
        {
          if (this.cancelCallback && !this.state.cancelled) {
            this.cancelCallback();
            this.setState(States.CANCELLED, true);
          }

          return true;
        }
    }

    return super.action(action, params);
  }

  setCancelCallback(callback) {
    if (callback) {
      _balsamiq.default.assertFunction(callback);
    }

    this.cancelCallback = callback;
    this.setState(States.CANCELLABLE, this.cancelCallback !== null);
  }

  setProgress(operationName, currentStep, maxSteps, message) {
    this.setState(States.CURRENT_STEP, currentStep);
    this.setState(States.MAX_STEPS, maxSteps);
    this.setState(States.OPERATION_NAME, operationName);
    this.setState(States.MESSAGE, message ? message : null);
  }

}

var _default = {
  ViewController: OperationProgressViewController,
  Type: "operationProgress",
  ChildNames: {},
  Actions,
  KeyNames,
  States
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/viewmodels/commentstextutils.js":
/*!*************************************************!*\
  !*** ./js/core/viewmodels/commentstextutils.js ***!
  \*************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;
exports.extractNewItemsOneTime = extractNewItemsOneTime;
exports.getCalloutIDsFromText = getCalloutIDsFromText;
exports.getMentionedUserIDs = getMentionedUserIDs;
exports.makeASTFromText = makeASTFromText;
exports.makeTextFromAST = makeTextFromAST;
exports.processCommentAST = processCommentAST;
exports.processMockupNotesAST = processMockupNotesAST;
exports.processNodesForAlternateNamesToLinkIDs = processNodesForAlternateNamesToLinkIDs;
exports.processNodesForCalloutNamesToLinkIDs = processNodesForCalloutNamesToLinkIDs;
exports.processNodesForUserNamesToUserIDs = processNodesForUserNamesToUserIDs;
exports.processProjectNotesAST = processProjectNotesAST;
exports.transformAlternateNamesToLinkIDs = transformAlternateNamesToLinkIDs;
exports.transformCalloutIDsToCalloutText = transformCalloutIDsToCalloutText;
exports.transformCalloutNamesToLinkIDs = transformCalloutNamesToLinkIDs;
exports.transformLinkIDsToAlternateNames = transformLinkIDsToAlternateNames;
exports.transformUserIDsToUserNames = transformUserIDsToUserNames;
exports.transformUserNamesToUserIDs = transformUserNamesToUserIDs;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _commentcalloutdata = __webpack_require__(/*! editor/viewcontrollers/commentcalloutdata */ "./js/editor/viewcontrollers/commentcalloutdata.js");

var CoreObject = _interopRequireWildcard(__webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js"));

var _textconstants = __webpack_require__(/*! core/texts/textconstants */ "./js/core/texts/textconstants.js");

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const {
  MentionAll
} = _textconstants.NodeIdentifiers;

function transformLinkIDsToAlternateNames(text, project) {
  if (!text || !project) {
    return '';
  }

  _balsamiq.default.assertString(text);

  _balsamiq.default.assertFunction(project.getBranch);

  _balsamiq.default.exists(project.coreObjectID);

  const uuidRegExp = new RegExp("[A-Z0-9]{8}-([A-Z0-9]{4}-){3}[A-Z0-9]{12}|Master");

  const processNodes = nodes => {
    nodes.forEach(node => {
      if (node.type == _textconstants.NodeTypes.LINK && node.target) {
        const target = TextUtils.nodesToString(node.target);

        if (uuidRegExp.test(target)) {
          const branchCoreeObjectID = CoreObject.makeBranchCoreObjectID(project.coreObjectID, target);
          const branch = project.getBranch(branchCoreeObjectID);

          if (branch) {
            // node.target = [ branch.getName() ];
            node.children = [branch.getName()];
            node.target = null;
          }
        }
      } else if (node.children) {
        processNodes(node.children);
      }
    });
  };

  let nodes = TextUtils.stringToNodes(text);
  processNodes(nodes);
  return TextUtils.nodesToString(nodes);
}

;

function processNodesForAlternateNamesToLinkIDs(nodes, project) {
  nodes.forEach(node => {
    if (node.type == _textconstants.NodeTypes.LINK) {
      const alternateName = TextUtils.nodesToString(node.children);
      const branch = project.getBranchByName(alternateName);

      if (branch) {
        node.target = [branch.coreObjectID.branchID];
      }
    } else if (node.children) {
      processNodesForAlternateNamesToLinkIDs(node.children, project);
    }
  });
}

;

function transformAlternateNamesToLinkIDs(text, project) {
  if (!text && !project) {
    return '';
  }

  _balsamiq.default.assertString(text);

  _balsamiq.default.assertFunction(project.getBranchByName);

  let nodes = TextUtils.stringToNodes(text);
  processNodesForAlternateNamesToLinkIDs(nodes, project);
  return TextUtils.nodesToString(nodes);
}

;

function processNodesForUserNamesToUserIDs(nodes, project) {
  nodes.forEach(node => {
    if (node.type === _textconstants.NodeTypes.MENTION) {
      const identifier = TextUtils.nodesToString(node.children);

      if (identifier === MentionAll) {
        node.target = [MentionAll];
      } else {
        const user = project.getProjectMemberByName(identifier);

        if (user) {
          node.target = [user.id.toString()];
        }
      }
    } else if (node.children) {
      processNodesForUserNamesToUserIDs(node.children, project);
    }
  });
}

function transformUserNamesToUserIDs(text, project) {
  if (!text) {
    return '';
  }

  _balsamiq.default.assertString(text);

  const nodes = TextUtils.stringToNodes(text);
  processNodesForUserNamesToUserIDs(nodes, project);
  return TextUtils.nodesToString(nodes);
}

;

function transformUserIDsToUserNames(text, project) {
  if (!text) {
    return '';
  }

  _balsamiq.default.assertString(text);

  const processNodes = nodes => {
    nodes.forEach(node => {
      if (node.type == _textconstants.NodeTypes.MENTION) {
        let userId = null;

        if (node.target) {
          userId = TextUtils.nodesToString(node.target);
        } else if (node.children) {
          // CAVEAT: avoids breaking old mention syntax where userId is stored *children* prop instead of *target*
          const tmp = parseInt(TextUtils.nodesToString(node.children));
          userId = isNaN(tmp) ? null : tmp.toString();
        }

        if (node.target && node.target[0] === MentionAll) {
          node.target = null; // CAVEAT: removes the target so that it's not rendered in comment's body, when the user is editing.
        } else if (project) {
          const user = project.getProjectMember(userId);

          if (user) {
            node.children = [user.identifier];
            node.target = null; // CAVEAT: removes the target so that it's not rendered in comment's body, when the user is editing.
          }
        }
      } else if (node.children) {
        processNodes(node.children);
      }
    });
  };

  const nodes = TextUtils.stringToNodes(text);
  processNodes(nodes);
  return TextUtils.nodesToString(nodes);
}

;

function getMentionedUserIDs(nodes) {
  let mentionedUserIDs = [];

  const visitNodes = nodes => {
    nodes.forEach(node => {
      if (node.type == _textconstants.NodeTypes.MENTION && node.target) {
        mentionedUserIDs.push(node.target[0]);
      } else if (node.children) {
        visitNodes(node.children);
      }
    });
  };

  visitNodes(nodes);
  return mentionedUserIDs;
}

function processNodesForCalloutNamesToLinkIDs(nodes, commentsViewModel) {
  nodes.forEach(node => {
    if (node.type == _textconstants.NodeTypes.LINK) {
      var calloutText = TextUtils.nodesToString(node.children);

      if ((0, _commentcalloutdata.isValidCalloutTextLink)(calloutText)) {
        let id = commentsViewModel.getCalloutIDByTextLink(calloutText);

        if (id) {
          node.target = [id];
        }
      }
    } else if (node.children) {
      processNodesForCalloutNamesToLinkIDs(node.children, commentsViewModel);
    }
  });
}

function transformCalloutNamesToLinkIDs(text, commentsViewModel) {
  if (!text) {
    return '';
  }

  _balsamiq.default.assertString(text);

  _balsamiq.default.assertFunction(commentsViewModel.getCalloutIDByTextLink);

  let nodes = TextUtils.stringToNodes(text);
  processNodesForCalloutNamesToLinkIDs(nodes, commentsViewModel);
  return TextUtils.nodesToString(nodes);
}

;

function transformCalloutIDsToCalloutText(text, commentsViewModel) {
  if (!text) {
    return '';
  }

  _balsamiq.default.assertString(text);

  _balsamiq.default.assertFunction(commentsViewModel.getTextLinkByCalloutID);

  var processNodes = nodes => {
    nodes.forEach(node => {
      if (node.type == _textconstants.NodeTypes.LINK && node.target) {
        var calloutString = TextUtils.nodesToString(node.children);

        if ((0, _commentcalloutdata.isValidCalloutTextLink)(calloutString)) {
          let calloutID = TextUtils.nodesToString(node.target);
          let calloutText = commentsViewModel.getTextLinkByCalloutID(calloutID) || calloutString;
          node.children = [calloutText];
          node.target = null;
        }
      } else if (node.children) {
        processNodes(node.children);
      }
    });
  };

  var nodes = TextUtils.stringToNodes(text);
  processNodes(nodes);
  return TextUtils.nodesToString(nodes);
}

;

function getCalloutIDsFromText(text) {
  if (!text) {
    return '';
  }

  _balsamiq.default.assertString(text);

  var ids = [];

  let processNodes = nodes => {
    nodes.forEach(node => {
      if (node.type == _textconstants.NodeTypes.LINK) {
        var calloutText = TextUtils.nodesToString(node.children);

        if ((0, _commentcalloutdata.isValidCalloutTextLink)(calloutText) && node.target) {
          ids.push(TextUtils.nodesToString(node.target));
        }
      } else if (node.children) {
        processNodes(node.children);
      }
    });
  };

  let nodes = TextUtils.stringToNodes(text);
  processNodes(nodes);
  return ids;
}

;

function makeASTFromText(text) {
  if (!text) {
    return [];
  }

  return TextUtils.stringToNodes(text);
}

;

function processCommentAST(nodes, project, commentsViewModel) {
  _balsamiq.default.assert(project);

  _balsamiq.default.assert(commentsViewModel);

  processNodesForAlternateNamesToLinkIDs(nodes, project);
  processNodesForCalloutNamesToLinkIDs(nodes, commentsViewModel);
  processNodesForUserNamesToUserIDs(nodes, project);
}

;

function processMockupNotesAST(nodes, project) {
  _balsamiq.default.assert(project);

  processNodesForUserNamesToUserIDs(nodes, project);
}

function processProjectNotesAST(nodes, project) {
  _balsamiq.default.assert(project);

  processNodesForUserNamesToUserIDs(nodes, project);
}

function makeTextFromAST(nodes) {
  return TextUtils.nodesToString(nodes);
}

;

function extractNewItemsOneTime(before, after) {
  const occurrences = (array, occuredItem) => array.reduce((occurrences, item) => item === occuredItem ? occurrences + 1 : occurrences, 0);

  const uniqueItems = [...new Set(after)];
  const newItems = [];

  for (const item of uniqueItems) {
    if (occurrences(after, item) > occurrences(before, item)) {
      newItems.push(item);
    }
  }

  return newItems;
}

var _default = {
  transformLinkIDsToAlternateNames,
  transformAlternateNamesToLinkIDs,
  transformCalloutNamesToLinkIDs,
  transformCalloutIDsToCalloutText,
  transformUserNamesToUserIDs,
  transformUserIDsToUserNames,
  getCalloutIDsFromText,
  makeASTFromText,
  makeTextFromAST,
  processMockupNotesAST,
  processProjectNotesAST,
  processCommentAST,
  getMentionedUserIDs,
  extractNewItemsOneTime
};
exports["default"] = _default;

/***/ }),

/***/ "./js/editor/constants/nativeactions.js":
/*!**********************************************!*\
  !*** ./js/editor/constants/nativeactions.js ***!
  \**********************************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.RequestOpenFileResponse = exports.RequestLoadProjectLocatorResponse = exports.RequestCreateNewProjectResponse = exports.NativeActions = exports.NativeActionParams = exports.Enums = void 0;
const NativeActions = {
  APP_ENDED: 'appEnded',
  // params:
  REQUEST_CREATE_NEW_PROJECT: "requestCreateNewProject",
  // params:
  OPEN_PROJECT: "openProject",
  // params: 
  REQUEST_OPEN_FILE: "requestOpenFile",
  // params: { }, expects a callback with { response: RequestOpenFileResponse }
  REQUEST_LOAD_PROJECT_LOCATOR: "requestLoadProjectLocator",
  // params: { NativeActionParams.ProjectLocator }, expects a callback with { response: RequestLoadProjectLocatorResponse }
  NOTIFY_OPEN_PROJECT: "notifyOpenProject",
  // params: { projectLocator }
  NOTIFY_TARGET_RESOURCE_NAME: "notifyResourceName",
  // params: { NativeActionParams.TargetResourceName:String? }
  NOTIFY_PROJECT_NAME: "notifyProjectName",
  // params: { NativeActionParams.ProjectName:String },
  NOTIFY_CLOUD_PROJECTS: "notifyCloudProjects",
  // params: { NativeActionParams.CloudProjects }
  OPEN_USER_INFO: "openUserInfo",
  // params: { message }
  OPEN_REGISTER_WINDOW: "openRegisterWindow",
  OPEN_CONTACT_SUPPORT: "openContactSupport",
  OPEN_SEARCH_DOCUMENTATION: "openSearchDocumentation",
  TAKE_MAIN_PROJECT_WIREFRAME_SCREENSHOT: "takeMainProjectWireframeScreenshot",
  GET_MAIN_PROJECT_USER_INFO: "getMainProjectUserInfo",
  GET_MAIN_PROJECT_URL: "getMainProjectUrl",
  OPEN_CHECK_FOR_UPDATE: "openCheckForUpdate"
};
exports.NativeActions = NativeActions;
const NativeActionParams = {
  ProjectName: "projectName",
  // String
  ProjectLocator: "projectLocator",
  // ProjectLocator object
  TargetResourceName: "targetResourceName",
  // String
  CloudProjects: "cloudProjects",
  // [ ProjectLocator object ]
  IsFinished: "isFinished",
  // Bool
  DesktopUserInfo: "desktopUserInfo",
  // Object { email: String, name: String }
  Message: "message" // String

};
exports.NativeActionParams = NativeActionParams;
const RequestOpenFileResponse = {
  Open: "open",
  // with param projectLocator:Object
  Cancel: "cancel",
  End: "end"
};
exports.RequestOpenFileResponse = RequestOpenFileResponse;
const RequestLoadProjectLocatorResponse = {
  Load: "load",
  End: "end"
};
exports.RequestLoadProjectLocatorResponse = RequestLoadProjectLocatorResponse;
const RequestCreateNewProjectResponse = {
  Create: "Create",
  End: "end"
};
exports.RequestCreateNewProjectResponse = RequestCreateNewProjectResponse;
const Enums = {
  NativeActions,
  NativeActionParams,
  RequestOpenFileResponse,
  RequestLoadProjectLocatorResponse,
  RequestCreateNewProjectResponse
};
exports.Enums = Enums;

/***/ }),

/***/ "./js/editor/operations/updateprojectfromdumpoperation.js":
/*!****************************************************************!*\
  !*** ./js/editor/operations/updateprojectfromdumpoperation.js ***!
  \****************************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var Operation = __webpack_require__(/*! core/operations/operation */ "./js/core/operations/operation.ts");

class UpdateProjectFromDumpOperation extends Operation.Operation {
  name() {
    return "Update Project From Dump";
  }

  isCancelable() {
    return false;
  }

  runDo() {
    _balsamiq.default.exists(this.params.projectController, "UpdateProjectFromDumpOperation requires a project controller in params");

    _balsamiq.default.exists(this.params.dump, "UpdateProjectFromDumpOperation requires a dump in params");

    var pc = this.params.projectController;
    pc.archive.replaceDump(this.params.dump);
    pc.loadProjectDump(() => {
      this.notifyFinished();
    });
  }

}

var _default = {
  Operation: UpdateProjectFromDumpOperation,
  Type: "UpdateProjectFromDump"
};
exports["default"] = _default;

/***/ }),

/***/ "./js/editor/projectsettings.js":
/*!**************************************!*\
  !*** ./js/editor/projectsettings.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.RightPanels = exports.ProjectSettings = exports.NewProjectAttributes = exports.NavigatorViewModes = void 0;
exports.isUnknownPropertiesNotificationDismissedForProject = isUnknownPropertiesNotificationDismissedForProject;
exports.loadNewProjectMasterBranchAttributes = loadNewProjectMasterBranchAttributes;
exports.loadProjectSettings = loadProjectSettings;
exports.saveNewProjectAttribute = saveNewProjectAttribute;
exports.saveProjectDismissedUnknownPropertiesNotification = saveProjectDismissedUnknownPropertiesNotification;
exports.saveProjectSettings = saveProjectSettings;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _project = __webpack_require__(/*! core/model/project */ "./js/core/model/project.js");

var _settings = __webpack_require__(/*! core/core/settings */ "./js/core/core/settings.js");

var _BmprConstants = __webpack_require__(/*! @balsamiq/bmpr/lib/BmprConstants */ "./node_modules/@balsamiq/bmpr/lib/BmprConstants.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const RightPanels = {
  None: 'none',
  Inspector: 'inspector',
  ProjectInspector: 'projectInspector',
  Comments: 'comments'
};
exports.RightPanels = RightPanels;
const NavigatorViewModes = {
  Thumbnails: 'thumbnails',
  List: 'list'
};
exports.NavigatorViewModes = NavigatorViewModes;
const ProjectSettings = {
  ChatPosition: {
    key: 'chatPosition',
    defaultValue: {
      x: 0
    }
  },
  UiLibraryVisible: {
    key: 'uilibraryVisible',
    defaultValue: true
  },
  UiLibraryCategories: {
    key: 'uilibraryCategories',
    defaultValue: []
  },
  MockupsNavigatorVisible: {
    key: 'mockupsNavigatorVisible',
    defaultValue: true
  },
  MockupsNavigatorWidth: {
    key: 'mockupsNavigatorWidth',
    defaultValue: 192
  },
  MockupsCommentsWidth: {
    key: 'mockupsCommentsPanelWidth',
    defaultValue: 340
  },
  MockupsNavigatorViewMode: {
    key: 'mockupsNavigatorViewMode',
    defaultValue: NavigatorViewModes.Thumbnails
  },
  MockupsGridMode: {
    key: 'mockupsGridMode',
    defaultValue: false
  },
  AssetsNavigatorVisible: {
    key: 'assetsNavigatorVisible',
    defaultValue: true
  },
  AssetsNavigatorWidth: {
    key: 'assetsNavigatorWidth',
    defaultValue: 192
  },
  AssetsNavigatorViewMode: {
    key: 'assetsNavigatorViewMode',
    defaultValue: NavigatorViewModes.Thumbnails
  },
  AssetsGridMode: {
    key: 'assetsGridMode',
    defaultValue: false
  },
  SymbolsNavigatorWidth: {
    key: 'symbolsNavigatorWidth',
    defaultValue: 192
  },
  RightPanel: {
    key: 'rightPanelContent',
    defaultValue: RightPanels.Inspector
  },
  LinkInspectorsCollapsed: {
    key: 'linkInspectorCollapsed',
    defaultValue: false
  },
  IconInspectorsCollapsed: {
    key: 'iconInspectorCollapsed',
    defaultValue: false
  },
  OtherSymbolsVisible: {
    key: 'otherSymbolsVisible',
    defaultValue: false
  },
  MarkupVisible: {
    key: 'markupVisible',
    defaultValue: true
  },
  GroupBreadcrumbsPosition: {
    key: 'groupBreadcrumbsPosition',
    defaultValue: {
      x: 100,
      y: 100
    }
  },
  // { x: Number, y : Number }
  AlternatesPanelHeight: {
    key: 'alternatesPanelHeight',
    defaultValue: 200
  },
  SelectedMockup: {
    key: 'selectedMockup',
    defaultValue: null
  },
  // resourceID
  SelectedMockupsAlternates: {
    key: 'selectedMockupsAlternates',
    defaultValue: {}
  },
  // resourceID -> branchID of non-Master branchIDs
  SelectedAsset: {
    key: 'selectedAsset',
    defaultValue: null
  },
  ClosedParentMockups: {
    key: 'closedParentMockups',
    defaultValue: []
  },
  // resourceIDs,
  Fullscreen: {
    key: 'FullScreenSettings',
    defaultValue: null
  },
  Presentation: {
    key: 'PresentationSettings',
    defaultValue: {}
  },
  Export: {
    key: 'ExportSettings',
    defaultValue: null
  },
  LastExternalUrl: {
    key: 'LastExternalUrl',
    defaultValue: []
  },
  LastImageFromWeb: {
    key: 'LastImageFromWeb',
    defaultValue: []
  },
  LastImageFromFlickr: {
    key: 'LastImageFromFlickr',
    defaultValue: []
  },
  MockupsViewportData: {
    key: 'MockupsScrollAndZoom',
    defaultValue: {}
  },
  // ControlID-BranchID -> { controlID: ID, branchID: ID, zoom: Number, scrollTop: Number, scrollLeft: Number }
  LastExportPNGFolderURL: {
    key: 'lastExportPNGFolderURL',
    defaultValue: null
  },
  LastExportPNGZIPName: {
    key: 'lastExportPNGZipName',
    defaultValue: null
  },
  LastExportPDFFileURL: {
    key: 'lastExportPDFFileURL',
    defaultValue: null
  },
  QuickDrawToolbarPosition: {
    key: 'quickDrawToolbarPosition',
    defaultValue: {
      x: null,
      y: null
    }
  },
  // { x: Number, y : Number }
  QuickDrawToolbarVisible: {
    key: 'quickDrawToolbarVisible',
    defaultValue: false
  },
  OLD_PngExportTransparentBackground: {
    key: 'pngExportTransparentBackground',
    defaultValue: null
  },
  OLD_PdfExport: {
    key: 'PDFSettings',
    defaultValue: null
  }
};
exports.ProjectSettings = ProjectSettings;
const CustomDefaultSettings = [ProjectSettings.UiLibraryVisible, ProjectSettings.UiLibraryCategories, ProjectSettings.MockupsNavigatorVisible, ProjectSettings.MockupsNavigatorWidth, ProjectSettings.MockupsCommentsWidth, ProjectSettings.MockupsNavigatorViewMode, ProjectSettings.MockupsGridMode, ProjectSettings.AssetsNavigatorVisible, ProjectSettings.AssetsNavigatorWidth, ProjectSettings.AssetsNavigatorViewMode, ProjectSettings.AssetsGridMode, ProjectSettings.SymbolsNavigatorWidth, ProjectSettings.RightPanel, ProjectSettings.LinkInspectorsCollapsed, ProjectSettings.IconInspectorsCollapsed, ProjectSettings.OtherSymbolsVisible, ProjectSettings.MarkupVisible, ProjectSettings.GroupBreadcrumbsPosition, ProjectSettings.AlternatesPanelHeight];
const NewProjectAttributes = {
  FontFace: {
    key: 'newProjectFontFace',
    branchAttribute: _BmprConstants.BranchAttributes.FontFace,
    defaultValue: _project.DEFAULT_FONTFACE
  },
  FontSize: {
    key: 'newProjectFontSize',
    branchAttribute: _BmprConstants.BranchAttributes.FontSize,
    defaultValue: _project.DEFAULT_FONTSIZE
  },
  LinkColor: {
    key: 'newProjectLinkColor',
    branchAttribute: _BmprConstants.BranchAttributes.LinkColor,
    defaultValue: _project.DEFAULT_LINK_COLOR
  },
  SelectionColor: {
    key: 'newProjectSelectionColor',
    branchAttribute: _BmprConstants.BranchAttributes.SelectionColor,
    defaultValue: _project.DEFAULT_SELECTION_COLOR
  },
  SkinName: {
    key: 'newProjectSkinName',
    branchAttribute: _BmprConstants.BranchAttributes.SkinName,
    defaultValue: _project.DEFAULT_SKINNAME
  }
};
exports.NewProjectAttributes = NewProjectAttributes;

function loadProjectSettings(settings, project, projectSettings) {
  _balsamiq.default.existsClass(settings, _settings.Settings);

  _balsamiq.default.existsClass(project, _project.Project);

  _balsamiq.default.exists(projectSettings);

  let customDefaultValue = settings.getData(makeCustomDefaultKey(projectSettings.key));
  let defaultValue = customDefaultValue !== undefined ? customDefaultValue : projectSettings.defaultValue;
  return getProjectData(settings, projectSettings.key, project, defaultValue);
}

function saveProjectSettings(settings, project, projectSettings, value) {
  _balsamiq.default.existsClass(settings, _settings.Settings, "Settings");

  _balsamiq.default.existsClass(project, _project.Project, "Project");

  _balsamiq.default.exists(projectSettings, "projectSettings");

  if (isInCustomDefaultSettings(projectSettings)) {
    settings.setData(makeCustomDefaultKey(projectSettings.key), value);
  }

  setProjectData(settings, projectSettings.key, project, value);
}

function isInCustomDefaultSettings(projectSettings) {
  return CustomDefaultSettings.find(obj => projectSettings.key == obj.key) != undefined;
}

function embedProjectInKey(key, project) {
  if (project && project.getExternalProjectID()) {
    return key + "_projectId_" + project.getExternalProjectID();
  }

  return key;
}

function makeCustomDefaultKey(key) {
  return key + "_default";
}

function setProjectData(settings, key, project, data) {
  _balsamiq.default.existsClass(settings, _settings.Settings);

  _balsamiq.default.assertString(key);

  if (project !== undefined) {
    _balsamiq.default.existsClass(project, _project.Project);
  }

  if (project && project.getExternalProjectID()) {
    settings.setData(embedProjectInKey(key, project), data);
  }
}

function getProjectData(settings, key, project, defaultValue) {
  _balsamiq.default.existsClass(settings, _settings.Settings);

  _balsamiq.default.assertString(key);

  if (project !== undefined) {
    _balsamiq.default.existsClass(project, _project.Project);
  }

  let data;

  if (project && project.getExternalProjectID()) {
    data = settings.getData(embedProjectInKey(key, project));
  }

  if (data == undefined) {
    data = defaultValue;
  }

  return data;
}

function saveNewProjectAttribute(settings, projectAttribute, value) {
  _balsamiq.default.existsClass(settings, _settings.Settings);

  _balsamiq.default.exists(projectAttribute);

  settings.setData(projectAttribute.key, value); // let test = loadNewProjectMasterBranchAttributes(settings);
  // console.log(test);
}

function loadNewProjectMasterBranchAttributes(settings) {
  _balsamiq.default.existsClass(settings, _settings.Settings);

  let attributes = {};

  for (let name in NewProjectAttributes) {
    let obj = NewProjectAttributes[name];
    let data = settings.getData(obj.key);

    if (data === undefined) {
      data = obj.defaultValue;
    }

    _balsamiq.default.assert(data !== undefined);

    attributes[obj.branchAttribute] = data;
  }

  return attributes;
}

const DismissedUnknownPropertiesNotificationStorageKey = "DismissedUnknownPropertiesNotificationStorageKey";

function saveProjectDismissedUnknownPropertiesNotification(settings, coreID, projectCoreObjectID) {
  let data = settings.getSessionData(DismissedUnknownPropertiesNotificationStorageKey);

  if (!data) {
    data = {};
  }

  const id = coreID + "--" + projectCoreObjectID.toString();
  data[id] = true;
  settings.setSessionData(DismissedUnknownPropertiesNotificationStorageKey, data);
}

function isUnknownPropertiesNotificationDismissedForProject(settings, coreID, projectCoreObjectID) {
  let data = settings.getSessionData(DismissedUnknownPropertiesNotificationStorageKey);

  if (!data) {
    data = {};
  }

  const id = coreID + "--" + projectCoreObjectID.toString();
  return data[id] === true;
}

/***/ }),

/***/ "./js/editor/tasks/alerttask.js":
/*!**************************************!*\
  !*** ./js/editor/tasks/alerttask.js ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.AlertTask = void 0;

var _task = __webpack_require__(/*! core/tasks/task */ "./js/core/tasks/task.ts");

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class AlertTask extends _task.TaskBase {
  constructor(params) {
    super(params);

    _balsamiq.default.assertString(params.message);

    _balsamiq.default.assertString(params.title);

    this.message = params.message;
    this.title = params.title;
  }

  name() {
    return "AlertTask";
  }

  doRun() {
    this.context.viewControllers.pushAlert({
      windowID: this.context.windowID,
      taskID: this.getTaskID(),
      message: this.message,
      title: this.title,
      buttons: null,
      callback: () => this.notifyFinished(),
      closable: true
    });
  }

}

exports.AlertTask = AlertTask;

/***/ }),

/***/ "./js/editor/ui_constants.js":
/*!***********************************!*\
  !*** ./js/editor/ui_constants.js ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ZIndexes = exports.Utility = exports.Users = exports.UserImagePlaceholder = exports.UILibrary = exports.Transform = exports.Tooltip = exports.Toolbar = exports.TextEdit = exports.Text = exports.Tables = exports.StatusBar = exports.SmallInputs = exports.SmallButtons = exports.SizeOverlay = exports.Selects = exports.QuickDrawToolbar = exports.QuickAdd = exports.Panels = exports.OtherUsers = exports.Onboarding = exports.Notifications = exports.Nav = exports.Mysites = exports.MockupCanvas = exports.Menus = exports.MediaQueries = exports.LockedControls = exports.LargeButtons = exports.Label = exports.Inputs = exports.Grid = exports.Forms = exports.Focus = exports.FloatingBar = exports.EditorConstants = exports.Dialog = exports.Defaults = exports.Colors = exports.ColorSchemes = exports.Chat = exports.Callouts = exports.Buttons = exports.Breakpoints = exports.Boxes = exports.AppBarButtonContainerRecessedWithLeftMargin = exports.AppBarButtonContainerRecessedMedium = exports.AppBarButtonContainerRecessed = exports.AppBarButtonContainer = void 0;
// This is the file where cross platform UI constants should be. (eg: color palette, some b4 style, etc...)
//
// Every native platform will import and convert properly the following values,
// so they should be as "agnostic" as possible
// RULES TO FOLLOW
//
// - UseCamelCase to define props
// - Colors properties should contain the "Color" prop (except for the ones in "Colors" object)
// - Size properties should contains "Size", "Width", "Height", "Margin" or "Padding"
//
// These rules can be changed easily in cssConstantsAdapter.js
const Colors = {
  // BLACK WHITE
  White: 0xFFFFFF,
  White1: 0xFAFAFA,
  Black: 0x222222,
  // GRAY
  Gray: 0x999999,
  Gray1: 0xDDDDDD,
  Gray2: 0xC9CBCC,
  Gray2HalfAlpha: 0xE0E1E1,
  Gray3: 0xEEEEEE,
  Gray4: 0x666666,
  Gray5: 0x535353,
  Gray6: 0xBBBBBB,
  Gray7: 0xF0F0F0,
  Gray8: 0xCCCCCC,
  // GREEN
  Green: 0x74C41D,
  GreenLight: 0xABDB77,
  GreenDark: 0x68AB28,
  // RED
  Red: 0xCC0000,
  RedLight: 0xE06666,
  RedDark: 0xAF0F11,
  RedLipstick: 0xd31f38,
  // ORANGE
  Orange: 0xef571c,
  // YELLOW
  Yellow: 0xFFFF00,
  Yellow1: 0xF9DA56,
  Yellow2: 0xf4cc1e,
  Yellow3: 0xfedc49,
  // BLUE
  BlueCerulean: 0x0070BC,
  BlueDarkSky: 0x4a9ed9,
  // BACKGROUNDS
  Dark: 0x394045,
  // Text
  LightBg: 0x595e62,
  DarkBg: 0x3D4A54,
  // PRIMARY
  Primary: 0x3D9BE8,
  PrimaryHover: 0x64AFED,
  PrimaryActive: 0x67BBA,
  PrimaryActiveDarkBg: 0x64AFED,
  PrimarySelectedLight: 0xB1D7F5,
  // SECONDARY
  Secondary: 0x666666,
  SecondaryHover: 0x999999,
  SecondaryActive: 0x535353,
  // PRIMARY LIGHT
  PrimaryLight: 0xE2F0FC,
  PrimaryLightHover: 0xEEF6FD,
  PrimaryLightActive: 0xCAD7E2,
  // SUCCESS
  Success: 0x74C41D,
  // Green
  SuccessHover: 0xABDB77,
  // GreenLight
  SuccessActive: 0x68AB28,
  // GreenDark
  // ALERT
  Alert: 0xCC0000,
  // Red
  AlertHover: 0xE06666,
  // RedLight
  AlertActive: 0xAF0F11,
  // RedDark
  AlertLightBg: 0xF8E0E0,
  // RedVeryLight
  // TEXT
  Text: 0x394045,
  // Dark
  Link: 0x067BBA,
  LinkLight: 0xE2F0FC,
  Active: 0x3D9BE8,
  ActiveLight: 0x3D9BE8,
  Selected: 0x3D9BE8,
  // TODO
  Selected2: 0x76b9ef,
  // TODO multiple selection
  Disabled: 0xB3B3B3,
  // TODO
  SelectedDisabled: 0x477090
}; // DEFAULTS

exports.Colors = Colors;
const Defaults = {
  BodyColor: Colors.Gray7,
  BorderRadius: "3px",
  BorderColor: Colors.Gray2,
  BoxShadow: "0 2px 3px rgba(0,0,0,.4);",
  IconColor: Colors.Black,
  IconMediumColor: Colors.Gray,
  ViewMaxWidth: '1200px'
};
exports.Defaults = Defaults;
const MockupCanvas = {
  VoidBackgroundColor: Colors.Gray7,
  EmptyBackgroundColor: Colors.White,
  GraphPaperBackgroundColor: Colors.White1,
  Padding: 400
};
exports.MockupCanvas = MockupCanvas;
const Text = {
  DefaultColor: Colors.Dark,
  DarkColor: Colors.Black,
  MediumColor: Colors.Gray,
  LightColor: Colors.White,
  LinkColor: Colors.Link,
  LinkColorLight: Colors.LinkLight,
  FontFamily: "Open Sans",
  LargeFontFamily: "Open Sans",
  FontSize: 12
};
exports.Text = Text;
const Focus = {
  Color: Colors.Selected,
  OutlineWidth: 2
};
exports.Focus = Focus;
const Tooltip = {
  BackgroundColor: Colors.LightBg,
  BoxShadow: '0px 1px 2px 0px rgba(50, 50, 50, 0.55)',
  Color: Colors.White,
  LinkColor: Colors.White,
  PaddingTop: '6px',
  PaddingBottom: '6px',
  PaddingLeft: '12px',
  PaddingRight: '12px',
  BorderRadius: '5px',
  MaxWidth: '305px'
};
exports.Tooltip = Tooltip;
const AppBarButtonContainer = {
  MarginRight: '6px',
  Height: '20px'
};
exports.AppBarButtonContainer = AppBarButtonContainer;
const AppBarButtonContainerRecessed = {
  PaddingTop: '2px',
  PaddingBottom: '2px',
  PaddingLeft: '6px',
  PaddingRight: '6px',
  MarginRight: '12px',
  BackgroundColor: Colors.Gray4,
  Color: Colors.White,
  BorderRadius: Defaults.BorderRadius
};
exports.AppBarButtonContainerRecessed = AppBarButtonContainerRecessed;
const AppBarButtonContainerRecessedMedium = {
  MarginRight: '6px'
};
exports.AppBarButtonContainerRecessedMedium = AppBarButtonContainerRecessedMedium;
const AppBarButtonContainerRecessedWithLeftMargin = {
  MarginLeft: '6px'
};
exports.AppBarButtonContainerRecessedWithLeftMargin = AppBarButtonContainerRecessedWithLeftMargin;
const StatusBar = {
  PaddingTop: '12px',
  PaddingBottom: '12px',
  PaddingLeft: '12px',
  PaddingRight: '12px',
  MinHeight: '40px',
  FontFamily: Text.LargeFontFamily,
  FontSize: 14,
  LineHeight: '14px',
  WarningBackgroundColor: Colors.Yellow3,
  WarningTextColor: Colors.Black,
  AlertBackgroundColor: Colors.Alert,
  AlertTextColor: Colors.White
};
exports.StatusBar = StatusBar;
const Toolbar = {
  BackgroundColor: Colors.Dark,
  ReadOnlyBackgroundColor: Colors.Gray5,
  Color: Colors.White,
  ColorActive: Colors.PrimaryActiveDarkBg,
  FontSize: '14px',
  IconColor: Colors.White,
  IconSize: '20px',
  HoverIconColor: Colors.Primary,
  SelectedIconColor: Colors.PrimaryHover,
  DisabledIconColor: Colors.Gray4,
  DisabledIconColorWTG: Colors.Gray8,
  SelectedDisabledColor: Colors.SelectedDisabled,
  SeparatorColor: Colors.Gray4,
  SeparatorWidth: '1px',
  ProjectNameFontSize: 14,
  TitleFontSize: '21px',
  TitleLineHeight: '21px',
  UsernameFontSizeLarge: '16px',
  UsernameLineHeightLarge: '16px',
  UsernameFontSizeSmall: '12px',
  UsernameLineHeightSmall: '12px',
  AvatarBorderWidth: 1,
  AvatarWidth: 18,
  AvatarHeight: 18,
  AvatarBorderRadius: 0
};
exports.Toolbar = Toolbar;
const Panels = {
  BorderColor: Defaults.BorderColor,
  HeadingBackgroundColor: Colors.Gray2,
  InactiveHeadingBackgroundColor: Colors.Gray2HalfAlpha,
  HeadingFontSize: 14,
  BackgroundColor: Colors.Gray3,
  EmptyBackgroundColor: Colors.Gray4,
  TriangleColor: Colors.Gray4,
  TriangleSelectedColor: Colors.White,
  IndentSize: 16,
  NavigatorIndentSize: 10,
  LineHeightSize: 12,
  TitleHeight: 36,
  TitlePaddingTop: 12,
  TitlePaddingBottom: 12
};
exports.Panels = Panels;
const Menus = {
  MenuBarFontSize: 14,
  MenuItemFontSize: 14,
  MenuItemHeight: 22,
  MenuItemPaddingLeft: 6,
  MenuItemPaddingRight: 24,
  MenuItemPaddingTop: 4,
  MenuItemPaddingBottom: 4,
  MenuMarkWidth: 20,
  MenuBarFontFamily: Text.LargeFontFamily,
  BackgroundColor: Colors.White,
  BackgroundColorAlpha: 0.97,
  SeparatorWidth: 1,
  SeparatorHeight: 14,
  SeparatorColor: Colors.Gray3,
  //0xEBEBEB,
  TextColor: Colors.Text,
  DisabledTextColor: Colors.Disabled,
  HighlightedBackgroundColor: Colors.Active,
  HighlightedTextColor: Colors.White,
  ShortcutLeftPadding: 12,
  ShortcutRightPadding: 6
};
exports.Menus = Menus;
const UILibrary = {
  ButtonFontSize: 11,
  ButtonTextColor: Text.DarkColor,
  ButtonBackgroundColor: Colors.White,
  SelectedButtonTextColor: Text.LightColor,
  SelectedButtonBackgroundColor: Colors.Gray
};
exports.UILibrary = UILibrary;
const QuickDrawToolbar = {
  BackgroundColor: Colors.Gray2,
  BackgroundOpacity: 0.9,
  Height: 80,
  ElementWidth: 80,
  SeparatorBorderColor: Defaults.BorderColor
};
exports.QuickDrawToolbar = QuickDrawToolbar;
const FloatingBar = {
  BackgroundColor: Colors.Gray2,
  BackgroundOpacity: 0.9,
  Height: 32,
  DoubleLineHeight: 58,
  HandleWidth: 18,
  SeparatorBorderColor: Defaults.BorderColor,
  PathElementHorizontalPadding: 10
};
exports.FloatingBar = FloatingBar;
const TextEdit = {
  BackgroundColor: Colors.Gray3,
  BackgroundOpacity: 0.9,
  BorderColor: Defaults.BorderColor,
  BorderWidth: 1,
  BorderRadius: 5
};
exports.TextEdit = TextEdit;
const QuickAdd = {
  FontSize: 11,
  BorderRadius: 5,
  // 3 + 2 of border
  BackgroundColor: Colors.Gray4,
  FocusedBackgroundColor: Colors.White,
  TextColor: Text.LightColor,
  TextColorDisabled: Colors.Gray,
  FocusedTextColor: Text.DefaultColor,
  TooltipText: "Type the first few letters of a control's name for suggestions, then use ↑ and ↓ to select and hit ENTER to add the control to your wireframe (⇧ENTER to add it at your mouse's position). Use / or + to send focus to Quick Add."
};
exports.QuickAdd = QuickAdd;
const OtherUsers = {
  SelectionColor: Colors.Gray4,
  ActiveUserSelectionAlpha: 0.5,
  WaitingUserSelectionAlpha: 0.25
};
exports.OtherUsers = OtherUsers;
const LockedControls = {
  LockIconFAName: 'lock',
  LockIconSize: 14,
  LockRectSize: 18,
  LockIconBorderColor: Colors.Black,
  LockIconBorderWidth: 1,
  LockIconBackgoundColor: Colors.White,
  SelectionRectColor: Colors.Gray,
  SelectionRectDashArray: [4, 4],
  SelectionRectBorderWidth: 1
};
exports.LockedControls = LockedControls;
const ZIndexes = {
  StatusBar: 5,
  Modal: 10,
  Utility: 15,
  Menu: 20,
  Tooltip: 30,
  BrowserWarning: 1000,
  Top: 9999
}; // CLOUD
// BREAKPOINTS

exports.ZIndexes = ZIndexes;
const Breakpoints = {
  Xsmall: '320px',
  Small: '768px',
  Medium: '992px',
  Large: '1200px',
  Modal: '650px',
  TableMin: '650px'
}; // GRID

exports.Breakpoints = Breakpoints;
const Grid = {
  Gutter: "15px"
}; // FORMS

exports.Grid = Grid;
const Forms = {
  BorderColor: Defaults.BorderColor
}; // BUTTONS

exports.Forms = Forms;
const Buttons = {
  BorderRadius: '5px',
  WhiteSpace: 'nowrap',
  // All
  FontColor: Colors.White,
  FontColorActive: Colors.White,
  FontColorLight: Colors.Text,
  FontColorLightActive: Colors.Text,
  FontColorDisabled: Colors.Gray,
  FontSize: 16,
  LineHeight: 21,
  MinHeight: 31,
  MinWidth: 78,
  Padding: '.5rem .25rem',
  // TODO: remove and use the underlying values
  PaddingLeft: 8,
  PaddingRight: 8,
  PaddingTop: 4,
  PaddingBottom: 4,
  // Disabled
  BackgroundColorDisabled: Colors.Gray6,
  BorderColorDisabled: Colors.Gray6,
  // Default
  BackgroundColor: Colors.Gray,
  BackgroundColorActive: Colors.Gray4,
  BackgroundColorHover: Colors.Gray2,
  BorderColor: Colors.Gray,
  BorderColorActive: Colors.Gray4,
  BorderColorHover: Colors.Gray2,
  // Primary
  BackgroundColorPrimary: Colors.Primary,
  BackgroundColorPrimaryActive: Colors.PrimaryActive,
  BackgroundColorPrimaryHover: Colors.PrimaryHover,
  BorderColorPrimary: Colors.Primary,
  BorderColorPrimaryActive: Colors.PrimaryActive,
  BorderColorPrimaryHover: Colors.PrimaryHover,
  // Primary Light
  BackgroundColorLight: Colors.PrimaryLight,
  BackgroundColorLightActive: Colors.PrimaryLightActive,
  BackgroundColorLightHover: Colors.PrimaryLightHover,
  BorderColorLight: Colors.PrimaryLight,
  BorderColorLightActive: Colors.PrimaryLightActive,
  BorderColorLightHover: Colors.PrimaryLightHover
};
exports.Buttons = Buttons;
const SmallButtons = {
  FontSize: 14,
  LineHeight: 20,
  PaddingLeft: 8,
  PaddingRight: 8,
  PaddingTop: 4,
  PaddingBottom: 4,
  MarginBottom: 2
};
exports.SmallButtons = SmallButtons;
const LargeButtons = {
  FontSize: 18,
  LineHeight: 24,
  PaddingLeft: 24,
  PaddingRight: 24,
  PaddingTop: 8,
  PaddingBottom: 8,
  LetterSpacing: 0.8,
  MarginBottom: 2
}; // INPUTS

exports.LargeButtons = LargeButtons;
const Inputs = {
  BackgroundColor: Colors.White,
  ReadOnlyBackgroundColor: Colors.Gray3,
  ReadOnlyInputColor: Colors.Gray5,
  BorderColor: Colors.Gray2,
  BorderFocusColor: Colors.Gray,
  BorderWidth: '2px',
  BorderRadius: '5px',
  BorderStyle: 'solid',
  FontColor: Colors.Dark,
  FontSize: '16px',
  LineHeight: '21px',
  MinHeight: '32px',
  PaddingTop: '4px',
  PaddingBottom: '4px',
  PaddingLeft: '4px',
  PaddingRight: '4px'
};
exports.Inputs = Inputs;
const SmallInputs = {
  FontSize: '14px',
  LineHeight: '20px',
  MinHeight: '32px',
  PaddingTop: '4px',
  PaddingBottom: '4px',
  PaddingLeft: '4px',
  PaddingRight: '4px'
};
exports.SmallInputs = SmallInputs;
const Label = {
  FontColor: Colors.Dark,
  FontSize: '16px',
  LineHeight: '21px',
  FontWeight: 'bold'
}; // SELECTS

exports.Label = Label;
const Selects = {
  BackgroundColor: Colors.White,
  BorderColor: Colors.Gray2,
  FontColor: Colors.Text,
  FontSize: '16px',
  LineHeight: '21px',
  MinHeight: '32px',
  Padding: '4px 4px'
}; // GRID OF SITES

exports.Selects = Selects;
const Mysites = {
  BackgroundColor: Colors.White,
  BackgroundColorActive: Colors.PrimaryLight,
  BackgroundColorDown: Colors.PrimaryLightActive,
  BorderColor: Colors.White,
  BorderColorActive: Colors.PrimaryLightActive,
  MetaFontColor: Colors.Gray4
}; // TABLES

exports.Mysites = Mysites;
const Tables = {
  BackgroundAlternateColor: Colors.Gray3,
  BorderColor: Defaults.BorderColor,
  CellPadding: '8px 8px 8px 0',
  CellMargin: '0',
  FontSize: '16px',
  LineHeight: '24px'
}; // RESOURCE NAVIGATOR

exports.Tables = Tables;
const Nav = {
  BackgroundColorActive: Colors.Primary,
  BackgroundColorDown: Colors.PrimaryActive,
  FontSize: '1rem',
  FontColor: Colors.Dark,
  FontColorActive: Colors.White,
  FooterBorderColor: Colors.Gray2,
  LineHeight: '2rem',
  ListLineHeight: 12,
  ListTitleHeight: 28,
  ListTitlePaddingTop: 8,
  ListTitlePaddingBottom: 8
};
exports.Nav = Nav;
const Boxes = {
  BorderRadius: 4
};
exports.Boxes = Boxes;
const Notifications = {
  FontSize: Text.FontSize,
  FontColor: Colors.White,
  BackgroundColor: Colors.Black,
  BorderRadius: 8,
  Padding: 10,
  Margin: 16,
  BorderColor: Colors.Black,
  BorderSize: 2
};
exports.Notifications = Notifications;
const Onboarding = {
  TitleFontSize: 18,
  FontSize: 13,
  FontColor: Text.DefaultColor,
  BackgroundColor: Colors.White1,
  BorderRadius: 8,
  Padding: 16,
  BorderColor: Colors.Gray1,
  BorderSize: 1
};
exports.Onboarding = Onboarding;
const Dialog = {
  BackgroundColor: Colors.White,
  BorderRadius: 5,
  BorderColor: Colors.Gray6,
  OverlayAlpha: 0.15,
  OverlayColor: Colors.Black,
  FontSize: 14,
  Padding: 16,
  TitleBackgroundColor: Colors.White,
  TitleFontSize: 16,
  TitleFontWeight: 'normal',
  TitleFontFamily: Text.FontFamily,
  TitleHeight: 24
};
exports.Dialog = Dialog;
const Utility = {
  BackgroundColor: Colors.White,
  BorderRadius: 5,
  BorderColor: Colors.Gray6,
  TitleBackgroundColor: Colors.White,
  TitleFontSize: 16,
  TitleFontWeight: 'normal',
  TitleFontFamily: Text.FontFamily,
  Padding: 16
};
exports.Utility = Utility;
const Chat = {
  BorderRadius: 5,
  BorderColor: Colors.Gray,
  BackgroundColor: Colors.Gray3,
  TitleBackgroundColor: Colors.Gray2,
  TitleColor: Text.DefaultColor,
  TitleFontSize: 14,
  InputBackgroundColor: Colors.White,
  Padding: 6
};
exports.Chat = Chat;
const Transform = {
  SelectedBackgroundColor: Colors.Selected,
  SelectedColor: Colors.White,
  ListBackgroundColor: Colors.White,
  TitleFontSize: 14,
  FontSize: 14
};
exports.Transform = Transform;
const SizeOverlay = {
  BackgroundColor: Colors.Gray4,
  TextColor: Colors.White,
  BackgroundColorAlpha: 0.7,
  FontSize: 36,
  FontFamily: "Open Sans",
  CornerRadius: 8,
  PaddingBottom: -5,
  Padding: -10
};
exports.SizeOverlay = SizeOverlay;
const Users = {
  AdminColor: 0x98C36A,
  // GREEN
  OwnerColor: 0x98C36A,
  // GREEN
  EditorColor: 0xDDDDDD,
  // GRAY LIGHT
  CommenterColor: 0xE2A846,
  // ORANGE
  ViewerColor: 0xE2A846 // ORANGE

};
exports.Users = Users;
const Callouts = {
  EditBackgroundColor: Colors.Gray4,
  EditBackgroundAlpha: 0.3,
  TooltipBackgroundColor: FloatingBar.BackgroundColor,
  TooltipBackgroundOpacity: FloatingBar.BackgroundOpacity,
  UnselectedTipColor: Colors.Gray,
  Colors: {
    Blue: 0x2b78e4,
    Green: 0x009e0f,
    Yellow: 0xf1c232,
    Red: 0xcf2a27,
    Gray: Colors.Gray1
  },
  TextColors: {
    Light: Colors.White1,
    Dark: Colors.Black
  },
  ReadOnlyColor: Colors.Gray
};
exports.Callouts = Callouts;
const UserImagePlaceholder = "/cloud/images/UserAvatarAnonymous120.png";
exports.UserImagePlaceholder = UserImagePlaceholder;
const EditorConstants = {
  TOPBAR_H: 26,
  TOPBAR_H_MOBILE: 26 * 2 + 5,
  TOPBAR_H_WTG: 65,
  APPBAR_H: 40,
  TOPPANEL_H: 120,
  TOPMESSAGE_H: 39,
  RIGHTPANEL_W: 300,
  SEPARATOR_W: 1,
  SEPARATOR_OFFSET: 5,
  LEFTPANEL_MAX_W: 400,
  LEFT_PANEL_MIN_W: 180,
  LEFTPANEL_START_W: 198,
  RIGHTPANEL_COMMENTS_MAX_W: 600,
  RIGHT_PANEL_COMMENTS_MIN_W: 300,
  RIGHTPANEL_COMMENTS_START_W: 340,
  PRESENTATION_MOCKUPINFO_MAX_H: 600,
  PRESENTATION_MOCKUPINFO_MIN_H: 100,
  PRESENTATION_MOCKUPINFO_START_H: 300,
  BOTTOMBAR_H: 25
};
exports.EditorConstants = EditorConstants;
const ColorSchemes = {
  // DESKTOP TRIAL
  WARNING0: {
    textColor: Colors.White,
    backgroundColor: Colors.PrimaryActive
  },
  WARNING1: {
    textColor: Colors.White,
    backgroundColor: Colors.Primary
  },
  WARNING2: {
    textColor: Colors.Black,
    backgroundColor: Colors.Yellow3
  },
  WARNING3: {
    textColor: Colors.White,
    backgroundColor: Colors.Orange
  },
  WARNING4: {
    textColor: Colors.White,
    backgroundColor: Colors.Red
  }
};
exports.ColorSchemes = ColorSchemes;
const MediaQueries = {
  VeryLarge: "screen and (max-width: 1054px)",
  Large: "screen and (max-width: 979px)",
  MediumLarge: "screen and (max-width: 870px)",
  Medium: "screen and (max-width: 767px)",
  Small: "screen and (max-width: 666px)",
  VerySmall: "screen and (max-width: 410px)",
  VeryVerySmall: "screen and (max-width: 361px)",
  NotSupported: "screen and (max-width: 319px)",
  Values: {
    VerySmall: 410,
    Small: 666
  }
};
exports.MediaQueries = MediaQueries;

/***/ }),

/***/ "./js/editor/utils/assets.js":
/*!***********************************!*\
  !*** ./js/editor/utils/assets.js ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.TYPICAL_PHOTO_DPI = exports.DEFAULT_JPG_QUALITY = void 0;
exports.calcDpiScale = calcDpiScale;
exports.calcSizeForMaxMegapixels = calcSizeForMaxMegapixels;
exports.calcSizeForResize = calcSizeForResize;
exports.convertPNGToJPGIfPossible = convertPNGToJPGIfPossible;
exports.resizeImageAssetIfNeeded = resizeImageAssetIfNeeded;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _mimetypes = __webpack_require__(/*! core/mimetypes */ "./js/core/mimetypes.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const MAX_MEGAPIXELS = 2;
const TYPICAL_PHOTO_DPI = [150, 300, 600, 1200]; // photos are often made with these dpi values

exports.TYPICAL_PHOTO_DPI = TYPICAL_PHOTO_DPI;
const DEFAULT_JPG_QUALITY = 0.92;
exports.DEFAULT_JPG_QUALITY = DEFAULT_JPG_QUALITY;

function calcDpiScale(width, height, dpi, isMac) {
  _balsamiq.default.assertPositiveNumber(width);

  _balsamiq.default.assertPositiveNumber(height);

  _balsamiq.default.assertOptionalPositiveNumber(dpi);

  _balsamiq.default.assertBool(isMac); // If the dpi info is not specified, then ignore it


  if (!_balsamiq.default.isNumber(dpi)) {
    return 1;
  } // Try to recognize a photo vs a screenshot from the dpi.
  // If it is a photo, then do not scale down.


  if (TYPICAL_PHOTO_DPI.indexOf(dpi) >= 0) {
    return 1;
  }

  const PlatformBaseDpi = isMac ? 72 : 96; // Windows and Linux use 96
  // If it is a rare image with very low dpi, then do not scale it

  if (dpi <= PlatformBaseDpi) {
    return 1;
  } // When a Windows/Linux image is opened on Mac, do not scale it


  if (isMac && dpi < 144) {
    return 1;
  } // It seems a screenshot: then scale it down to bring to the PlatformBaseDpi


  return dpi / PlatformBaseDpi;
}

function calcSizeForMaxMegapixels(width, height, maxMegapixels) {
  _balsamiq.default.assertPositiveNumber(width);

  _balsamiq.default.assertPositiveNumber(height);

  _balsamiq.default.assertPositiveNumber(maxMegapixels);

  const MaxPixelsCount = 1024 * 1024 * maxMegapixels;
  let imagePixelsCount = width * height;

  if (imagePixelsCount > MaxPixelsCount) {
    const ratio = width / height;
    height = Math.round(Math.sqrt(MaxPixelsCount / ratio));
    width = Math.round(height * ratio);
  }

  return {
    width,
    height
  };
}

function calcSizeForResize(width, height, dpi, isMac) {
  // Check if the image must be resized because of the high dpi
  let dpiScale = calcDpiScale(width, height, dpi, isMac);

  if (dpiScale > 1) {
    width = Math.round(width / dpiScale);
    height = Math.round(height / dpiScale);
  } // Check if the image must be resized because of the large size


  let res = calcSizeForMaxMegapixels(width, height, MAX_MEGAPIXELS);
  width = res.width;
  height = res.height;
  return {
    width,
    height
  };
}

function resizeImageAssetIfNeeded(context, data, callback) {
  _balsamiq.default.exists(context.host);

  _balsamiq.default.assertString(data);

  _balsamiq.default.assertFunction(callback);

  let {
    extension,
    mimeType
  } = (0, _mimetypes.getExtensionAndMimeTypeFromBLOB)(data);

  function doNothing() {
    context.host.asyncCallback(() => {
      callback({
        extension: extension,
        mimeType: mimeType,
        data: data
      });
    });
  }

  function deleteImageAndReturnData(imageID) {
    context.host.Images.getImageBase64Data(imageID, res => {
      _balsamiq.default.assertString(res.data);

      let imageData = res.data;
      context.host.Images.remove(imageID, res => {
        callback({
          isImage: true,
          extension: extension,
          mimeType: mimeType,
          data: imageData
        });
      });
    });
  }

  if (mimeType && (0, _mimetypes.isImageMimeType)(mimeType)) {
    context.host.Images.makeFromBase64Data(mimeType, data, res => {
      if (res.error) {
        doNothing();
      } else {
        let imageID = res.id;
        let width = res.width;
        let height = res.height;
        let dpi = res.dpi;
        let s = calcSizeForResize(width, height, dpi, context.host.runtime.isMac()); // Do the resize if needed

        if (s.width == width && s.height == height) {
          deleteImageAndReturnData(imageID);
        } else {
          context.host.Images.resize(imageID, s.width, s.height, res => {
            if (res.error) {
              doNothing();
            } else {
              deleteImageAndReturnData(imageID);
            }
          });
        }
      }
    });
  } else {
    doNothing();
  }
}

function convertPNGToJPGIfPossible(context, data, callback) {
  let {
    _,
    mimeType
  } = (0, _mimetypes.getExtensionAndMimeTypeFromBLOB)(data);

  if (!(0, _mimetypes.isImageMimeType)(mimeType) || mimeType !== _mimetypes.PNG_MIME) {
    callback({
      ok: false
    });
    return;
  } // Convert image to JPG


  context.host.Images.makeFromBase64Data(mimeType, data, image => {
    context.host.Images.usesAlpha(image.id, usesAlpha => {
      if (usesAlpha) {
        context.host.Images.remove(image.id, () => {
          callback({
            ok: false
          });
        });
      } else {
        context.host.Images.getJPGBase64Data(image.id, DEFAULT_JPG_QUALITY, _ref => {
          let {
            data
          } = _ref;
          context.host.Images.remove(image.id, () => {
            callback({
              data,
              ok: true
            });
          });
        });
      }
    });
  });
}

/***/ }),

/***/ "./js/editor/viewcontrollers/commentcalloutdata.js":
/*!*********************************************************!*\
  !*** ./js/editor/viewcontrollers/commentcalloutdata.js ***!
  \*********************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.CalloutTipType = exports.CALLOUT_WIDTH = exports.CALLOUT_NICE_PADDING = exports.CALLOUT_HEIGHT = exports.CALLOUT_CIRCLE_RADIUS = void 0;
exports.areEqual = areEqual;
exports.assertCallout = assertCallout;
exports.assertColor = assertColor;
exports.fixDuplicateTextLabelsForCallouts = fixDuplicateTextLabelsForCallouts;
exports.getCalloutRectMS = getCalloutRectMS;
exports.getCalloutTextFromTextLink = getCalloutTextFromTextLink;
exports.getTextColorForCalloutColor = void 0;
exports.isValidCalloutTextLink = isValidCalloutTextLink;
exports.makeCalloutRawObj = makeCalloutRawObj;
exports.makeCalloutTextLinkFromText = makeCalloutTextLinkFromText;
exports.makeTextLinkToCalloutIDMap = makeTextLinkToCalloutIDMap;
exports.makeTextLinkToCalloutInfoMap = makeTextLinkToCalloutInfoMap;
exports.makeTextMarkupForCallout = makeTextMarkupForCallout;
exports.parseCalloutIDsFromCommentText = parseCalloutIDsFromCommentText;
exports.removeSuffixFromCalloutText = removeSuffixFromCalloutText;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _ui_constants = __webpack_require__(/*! editor/ui_constants */ "./js/editor/ui_constants.js");

var _textconstants = __webpack_require__(/*! core/texts/textconstants */ "./js/core/texts/textconstants.js");

var TextUtils = _interopRequireWildcard(__webpack_require__(/*! core/texts/textutilities */ "./js/core/texts/textutilities.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

//referring to callout LEFT tip
const CALLOUT_CIRCLE_RADIUS = 20;
exports.CALLOUT_CIRCLE_RADIUS = CALLOUT_CIRCLE_RADIUS;
const CALLOUT_HEIGHT = 40;
exports.CALLOUT_HEIGHT = CALLOUT_HEIGHT;
const CALLOUT_WIDTH = 65;
exports.CALLOUT_WIDTH = CALLOUT_WIDTH;
const CALLOUT_NICE_PADDING = 65;
exports.CALLOUT_NICE_PADDING = CALLOUT_NICE_PADDING;

const getTextColorForCalloutColor = calloutColor => {
  switch (calloutColor) {
    case _ui_constants.Callouts.Colors.Blue:
    case _ui_constants.Callouts.Colors.Green:
    case _ui_constants.Callouts.Colors.Red:
      return _ui_constants.Callouts.TextColors.Light;

    case _ui_constants.Callouts.Colors.Grey:
    case _ui_constants.Callouts.Colors.Yellow:
      return _ui_constants.Callouts.TextColors.Dark;

    default:
      return _ui_constants.Callouts.TextColors.Dark;
  }
};

exports.getTextColorForCalloutColor = getTextColorForCalloutColor;

function assertCallout(callout) {
  _balsamiq.default.exists(callout);

  _balsamiq.default.assertEnum(callout.tip, CalloutTipType);

  _balsamiq.default.assertString(callout.label);

  _balsamiq.default.assertNumber(callout.x);

  _balsamiq.default.assertNumber(callout.y);

  _balsamiq.default.assertUUID(callout.id);

  _balsamiq.default.assertNumber(callout.color);

  _balsamiq.default.assertNumber(callout.labelColor);
}

function assertColor(color) {
  _balsamiq.default.assertEnum(color, _ui_constants.Callouts.Colors);
}

const CalloutTipType = {
  LEFT: "left",
  TOP: "top",
  RIGHT: "right",
  BOTTOM: "bottom",
  NONE: "none"
};
exports.CalloutTipType = CalloutTipType;

function makeCalloutRawObj(_ref) {
  let {
    x,
    y,
    tip = CalloutTipType.LEFT,
    id,
    label,
    color = _ui_constants.Callouts.Colors.Blue
  } = _ref;
  return {
    x,
    y,
    tip,
    id,
    label,
    color,
    labelColor: getTextColorForCalloutColor(color)
  };
} // Markup to be inserted in comment text
// [(ID)](CALLOUT_ID)


function makeTextMarkupForCallout(callout) {
  return `[(${callout.label})](${callout.id})`;
}

function makeTextLinkToCalloutIDMap(callouts) {
  _balsamiq.default.assertArray(callouts);

  let calloutsMap = {};
  callouts.forEach(callout => {
    calloutsMap[makeCalloutTextLinkFromText(callout.label)] = callout.id;
  });
  return calloutsMap;
}

function makeTextLinkToCalloutInfoMap(callouts) {
  _balsamiq.default.assertArray(callouts);

  return callouts.reduce((calloutsInfoByIDs, callout) => {
    calloutsInfoByIDs[makeCalloutTextLinkFromText(callout.label)] = _lodash.default.pick(callout, ['id', 'color', 'tip', 'label', 'labelColor']);
    return calloutsInfoByIDs;
  }, {});
} // Check if the format of the link is:
// (ID.SUB)


function isValidCalloutTextLink(textLink) {
  _balsamiq.default.assertString(textLink);

  return /\(\d+(.\d+)?\)/.test(textLink);
} // Add ( ) to text


function makeCalloutTextLinkFromText(text) {
  _balsamiq.default.assertString(text);

  return `(${text})`;
} // Parse ID from (ID)


function getCalloutTextFromTextLink(textLink) {
  _balsamiq.default.assertString(textLink);

  var res = null;

  if (isValidCalloutTextLink(textLink)) {
    res = textLink.slice(1, textLink.length - 1);
  }

  return res;
}

function parseCalloutIDsFromCommentText(commentText) {
  var calloutIDs = [];

  var processNodes = nodes => {
    nodes.forEach(node => {
      if (node.type == _textconstants.NodeTypes.LINK && node.target) {
        var calloutString = TextUtils.nodesToString(node.children);

        if (isValidCalloutTextLink(calloutString)) {
          let calloutID = TextUtils.nodesToString(node.target);
          calloutIDs.push(calloutID);
        }
      } else if (node.children) {
        processNodes(node.children);
      }
    });
  };

  var nodes = TextUtils.stringToNodes(commentText);
  processNodes(nodes);
  return calloutIDs;
}

function getCalloutRectMS(callout) {
  assertCallout(callout);
  var w, h;

  switch (callout.tip) {
    case CalloutTipType.LEFT:
    case CalloutTipType.RIGHT:
      w = CALLOUT_WIDTH;
      h = CALLOUT_HEIGHT;
      break;

    case CalloutTipType.TOP:
    case CalloutTipType.BOTTOM:
      w = CALLOUT_HEIGHT;
      h = CALLOUT_WIDTH;
      break;

    default:
      h = CALLOUT_HEIGHT;
      w = CALLOUT_HEIGHT;
  }

  return _types.default.newRectangle(callout.x - w / 2, callout.y - h / 2, w, h);
}

function removeSuffixFromCalloutText(text) {
  _balsamiq.default.assertString(text);

  var index = text.indexOf('.');

  if (index !== -1) {
    return text.replace(/\.(\d)+/, '');
  } else {
    return text;
  }
}

function fixDuplicateTextLabelsForCallouts(callouts) {
  var buckets = {};
  callouts.forEach(callout => {
    if (!buckets[callout.label]) {
      buckets[callout.label] = [];
    }

    buckets[callout.label].push(callout);
  });
  Object.keys(buckets).forEach(text => {
    if (buckets[text].length > 1) {
      buckets[text].sort();
      buckets[text].forEach((callout, index) => {
        callout.label += index ? "." + index : '';
      });
    }
  });
  return callouts;
}

function areEqual(callout1, callout2) {
  assertCallout(callout1);
  assertCallout(callout2);
  return _balsamiq.default.isEqual(callout1, callout2);
}

/***/ }),

/***/ "./js/editor/viewcontrollers/filepickerviewcontroller.js":
/*!***************************************************************!*\
  !*** ./js/editor/viewcontrollers/filepickerviewcontroller.js ***!
  \***************************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ViewController = exports.Type = exports.ChildNames = exports.Actions = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _viewcontroller = __webpack_require__(/*! core/viewcontrollers/viewcontroller */ "./js/core/viewcontrollers/viewcontroller.ts");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const Actions = {
  PICK: "pick",
  CANCEL: "cancel"
};
exports.Actions = Actions;

class FilePickerViewController extends _viewcontroller.ViewController {
  init() {
    this.callback = null;
  }

  setup(type, allowedExtensions, allowedMimeTypes, numFiles, maxFileMbSize, title, callback) {
    this.setState('allowedExtensions', allowedExtensions);
    this.setState('allowedMimeTypes', allowedMimeTypes);
    this.setState('numFiles', numFiles);
    this.setState('title', title);
    this.setState('type', type);
    this.setState('maxFileMbSize', maxFileMbSize);
    this.callback = callback;
  }

  action(action, params) {
    _balsamiq.default.assertString(action);

    switch (action) {
      case Actions.PICK:
        this.callback(true, params.files);
        return true;

      case Actions.CANCEL:
        this.callback(false);
        return true;
    }

    return super.action(action, params);
  }

}

exports.ViewController = FilePickerViewController;
const Type = 'filepicker';
exports.Type = Type;
const ChildNames = {};
exports.ChildNames = ChildNames;

/***/ }),

/***/ "./js/editor/viewmodels/mockupviewport.js":
/*!************************************************!*\
  !*** ./js/editor/viewmodels/mockupviewport.js ***!
  \************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ZOOM_STOPS = exports.Viewport = exports.MockupViewport = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const MinZoom = 0.1;
const MaxZoom = 3.0;
const ZOOM_STOPS = [MinZoom, 0.25, 0.33, 0.5, 0.66, 0.75, 1.0, 1.5, 2.0, 2.5, MaxZoom];
exports.ZOOM_STOPS = ZOOM_STOPS;

class Viewport {
  constructor(data) {
    this.zoom = 1.0;
    this.scrollTop = 0;
    this.scrollLeft = 0;
    this.canZoomIn = true;
    this.canZoomOut = true;
    this.size = {
      height: undefined,
      width: undefined
    };
    this.scrollToRectData = null;
    this.initFromObject(data);
  }

  initFromObject(data) {
    if (data) {
      _balsamiq.default.assertObject(data);

      this.scrollToRectData = null;

      if (data.canZoomIn !== undefined) {
        _balsamiq.default.assertBool(data.canZoomIn);

        this.canZoomIn = data.canZoomIn;
      }

      if (data.canZoomOut !== undefined) {
        _balsamiq.default.assertBool(data.canZoomOut);

        this.canZoomOut = data.canZoomOut;
      }

      if (data.zoom !== undefined) {
        _balsamiq.default.assertNumber(data.zoom);

        if (data.zoom > ZOOM_STOPS[ZOOM_STOPS.length - 1]) {
          this.zoom = ZOOM_STOPS[ZOOM_STOPS.length - 1];
          this.canZoomIn = false;
        } else if (data.zoom < ZOOM_STOPS[0]) {
          this.zoom = ZOOM_STOPS[0];
          this.canZoomOut = false;
        } else {
          this.zoom = data.zoom;
        }
      }

      if (data.scrollTop !== undefined) {
        _balsamiq.default.assertNumber(data.scrollTop);

        this.scrollTop = data.scrollTop;
      }

      if (data.scrollLeft !== undefined) {
        _balsamiq.default.assertNumber(data.scrollLeft);

        this.scrollLeft = data.scrollLeft;
      }

      if (data.height !== undefined) {
        _balsamiq.default.assertNumber(data.height);

        this.size.height = data.height;
      }

      if (data.width !== undefined) {
        _balsamiq.default.assertNumber(data.width);

        this.size.width = data.width;
      }

      if (data.size !== undefined && data.size.height !== undefined && data.size.width !== undefined) {
        _balsamiq.default.assertNumber(data.size.height);

        this.size.height = data.size.height;

        _balsamiq.default.assertNumber(data.size.width);

        this.size.width = data.size.width;
      }
    }
  }

}

exports.Viewport = Viewport;

class MockupViewport extends Viewport {
  constructor(context, resourceID, branchID, data) {
    super(data);

    _balsamiq.default.assertFunction(context.onViewportChanged);

    _balsamiq.default.assertUUID(resourceID);

    _balsamiq.default.assertBranch(branchID);

    this.context = context;
    this.resourceID = resourceID;
    this.branchID = branchID;
  }

  setChangedCallback(callback) {
    if (callback) {
      _balsamiq.default.assertFunction(callback);
    }

    this.onChangedCallback = callback;
  }

  onChanged() {
    this.context.onViewportChanged(this);
    this.onChangedCallback && this.onChangedCallback(this);
  }

  initFromViewport(viewport) {
    _balsamiq.default.assertClass(viewport, MockupViewport);

    this.initFromObject(viewport);
    this.onChanged();
  }

  setZoom(zoom) {
    _balsamiq.default.assertNumber(zoom);

    this.zoom = zoom;
    this.updateCanZoom();
    this.onChanged();
  }

  setViewScroll(top, left) {
    _balsamiq.default.assertNumber(top);

    _balsamiq.default.assertNumber(left);

    _balsamiq.default.assert(top >= 0);

    _balsamiq.default.assert(left >= 0);

    this.scrollTop = top;
    this.scrollLeft = left;
    this.onChanged();
    return this;
  }

  setViewSize(size) {
    this.size = size; // Now I can execute pending scroll to rect

    if (this.hasScrollToRectData()) {
      this.scrollToRectMS(...this.getScrollToRectData());
      this.resetScrollToRectData();
    }

    this.onChanged();
  }

  getViewSize() {
    return _balsamiq.default.clone(this.size);
  }

  viewportToMockupCoords(pointOrRect) {
    pointOrRect = _balsamiq.default.clone(pointOrRect);
    pointOrRect.x += this.scrollLeft;
    pointOrRect.y += this.scrollTop;
    pointOrRect.x /= this.zoom;
    pointOrRect.y /= this.zoom;

    if (Number.isFinite(pointOrRect.width)) {
      pointOrRect.width /= this.zoom;
    }

    if (Number.isFinite(pointOrRect.height)) {
      pointOrRect.height /= this.zoom;
    }

    return pointOrRect;
  }

  mockupToViewportCoords(pointOrRect) {
    pointOrRect = _balsamiq.default.clone(pointOrRect);
    pointOrRect.x *= this.zoom;
    pointOrRect.y *= this.zoom;

    if (Number.isFinite(pointOrRect.width)) {
      pointOrRect.width *= this.zoom;
    }

    if (Number.isFinite(pointOrRect.height)) {
      pointOrRect.height *= this.zoom;
    }

    pointOrRect.x -= this.scrollLeft;
    pointOrRect.y -= this.scrollTop;
    return pointOrRect;
  }

  updateCanZoom() {
    this.canZoomIn = this.zoom < MaxZoom;
    this.canZoomOut = this.zoom > MinZoom;
  }

  zoomInScrollWheel(mouseCoords, delta, payload) {
    _balsamiq.default.assertPoint(mouseCoords);

    _balsamiq.default.assertNumber(delta);

    _balsamiq.default.assertObject(payload);

    const now = Date.now();
    const isOldEnough = payload.lastEventDate ? now - payload.lastEventDate > 500 : true;
    const zoomToUseForComputing = payload.computedZoom && !isOldEnough ? payload.computedZoom : this.zoom;
    const newZoom = delta > 0 ? zoomToUseForComputing * (1 + delta) : zoomToUseForComputing / (1 - delta);
    const computedZoom = Math.min(MaxZoom, Math.max(newZoom, MinZoom));
    const snappedZoom = computedZoom >= 0.85 && computedZoom <= 1.15 ? 1 : computedZoom;
    const currentZoom = this.zoom;
    this.zoom = snappedZoom;
    this.updateCanZoom();
    this.scrollLeft = Math.max(0, Math.round(this.scrollLeft + mouseCoords.x * (this.zoom - currentZoom)));
    this.scrollTop = Math.max(0, Math.round(this.scrollTop + mouseCoords.y * (this.zoom - currentZoom)));
    payload.computedZoom = computedZoom;
    payload.lastEventDate = now;
    this.onChanged();
    return this;
  }

  zoomIn(selectionRectMS, viewWidth, viewHeight) {
    if (!this.canZoomIn) {
      return this;
    }

    _balsamiq.default.assertNumber(viewWidth);

    _balsamiq.default.assertNumber(viewHeight);

    const currentZoom = this.zoom;
    let zoom_stop_index = 0;

    while (currentZoom >= ZOOM_STOPS[zoom_stop_index] && zoom_stop_index < ZOOM_STOPS.length - 1) {
      zoom_stop_index++;
    }

    this.zoom = ZOOM_STOPS[zoom_stop_index];
    this.updateCanZoom();

    if (selectionRectMS) {
      _balsamiq.default.assertRect(selectionRectMS);

      this.scrollToDesiredVisibleCenter(selectionRectMS, viewWidth, viewHeight);
    } else {
      const currentViewRect = _types.default.newRectangle(Math.max(0, this.scrollLeft / currentZoom), Math.max(0, this.scrollTop / currentZoom), viewWidth / currentZoom, viewHeight / currentZoom);

      this.scrollToDesiredVisibleCenter(currentViewRect, viewWidth, viewHeight);
    }

    this.onChanged();
    return this;
  }

  scrollToDesiredVisibleCenter(centerRect, viewWidth, viewHeight) {
    _balsamiq.default.assertRect(centerRect);

    _balsamiq.default.assertNumber(viewWidth);

    _balsamiq.default.assertNumber(viewHeight);

    const dsv = _types.default.newRectangle(centerRect.x + centerRect.width / 2, centerRect.y + centerRect.height / 2, 0, 0);

    this.scrollLeft = Math.max(0, Math.round(dsv.x * this.zoom - viewWidth / 2));
    this.scrollTop = Math.max(0, Math.round(dsv.y * this.zoom - viewHeight / 2));
  }

  zoomOut(selectionRectMS, viewWidth, viewHeight) {
    if (!this.canZoomOut) {
      return this;
    }

    _balsamiq.default.assertNumber(viewWidth);

    _balsamiq.default.assertNumber(viewHeight);

    const currentZoom = this.zoom;
    let zoomIndex = ZOOM_STOPS.length - 1;

    while (currentZoom <= ZOOM_STOPS[zoomIndex] && zoomIndex >= 1) {
      zoomIndex--;
    }

    this.zoom = ZOOM_STOPS[zoomIndex];
    this.updateCanZoom();

    if (selectionRectMS) {
      _balsamiq.default.assertRect(selectionRectMS);

      this.scrollToDesiredVisibleCenter(selectionRectMS, viewWidth, viewHeight);
    } else {
      const currentViewRect = _types.default.newRectangle(Math.max(0, this.scrollLeft / currentZoom), Math.max(0, this.scrollTop / currentZoom), viewWidth / currentZoom, viewHeight / currentZoom);

      this.scrollToDesiredVisibleCenter(currentViewRect, viewWidth, viewHeight);
    }

    this.onChanged();
    return this;
  }

  zoom100(viewWidth, viewHeight) {
    const currentZoom = this.zoom;
    this.zoom = 1;
    this.canZoomIn = true;
    this.canZoomOut = true;

    if (viewWidth && viewHeight) {
      const currentViewRect = _types.default.newRectangle(Math.max(0, this.scrollLeft / currentZoom), Math.max(0, this.scrollTop / currentZoom), viewWidth / currentZoom, viewHeight / currentZoom);

      this.scrollToDesiredVisibleCenter(currentViewRect, viewWidth, viewHeight);
    } else {
      //Zoom All to actual size moves everything back to top left.
      this.scrollLeft = 0;
      this.scrollTop = 0;
    }

    this.onChanged();
    return this;
  }

  _calculateZoomToFit(viewWidth, viewHeight, mockupWidth, mockupHeight) {
    const ZOOM_TO_FIT_PADDING = 10;

    if (viewHeight > 0 && viewWidth > 0 && mockupWidth > 0 && mockupHeight > 0) {
      let computedZoom = Math.min(viewHeight / (mockupHeight + ZOOM_TO_FIT_PADDING * 2), viewWidth / (mockupWidth + ZOOM_TO_FIT_PADDING * 2));
      return Math.min(MaxZoom, Math.max(computedZoom, MinZoom));
    }

    return 1;
  }

  zoomFit(viewWidth, viewHeight, mockupWidth, mockupHeight, mockupTop, mockupLeft) {
    const ZOOM_TO_FIT_PADDING = 10;
    this.zoom = this._calculateZoomToFit(viewWidth, viewHeight, mockupWidth, mockupHeight);
    this.updateCanZoom();
    this.scrollTop = Math.max(0, Math.round((mockupTop - ZOOM_TO_FIT_PADDING) * this.zoom));
    this.scrollLeft = Math.max(0, Math.round((mockupLeft - ZOOM_TO_FIT_PADDING) * this.zoom));
    this.onChanged();
    return this;
  }

  scrollToRectMS(rectMS, padding, callback) {
    // Do it later, view has not been mounted yet!
    if (this.size.height < 0 || this.size.width < 0) {
      this.setScrollToRectData(rectMS, padding, callback);
      return this;
    }

    _balsamiq.default.assertRect(rectMS);

    var scrollLeft = this.scrollLeft;
    var scrollTop = this.scrollTop;
    var rectVS = this.mockupToViewportCoords(Object.assign({}, rectMS));

    if (Number.isInteger(padding) && padding > 0) {
      rectVS = _types.default.paddedRectangle(rectVS, -padding, -padding, -padding, -padding);
    }

    if (rectVS.x < 0) {
      scrollLeft = Math.round(this.scrollLeft + rectVS.x);
    }

    if (rectVS.y < 0) {
      scrollTop = Math.round(this.scrollTop + rectVS.y);
    }

    var rectRight = _types.default.rectangleRight(rectVS);

    if (rectRight > this.size.width) {
      scrollLeft = Math.round(this.scrollLeft + rectRight - this.size.width);
    }

    var rectBottom = _types.default.rectangleBottom(rectVS);

    if (rectBottom > this.size.height) {
      scrollTop = Math.round(this.scrollTop + rectBottom - this.size.height);
    }

    if (scrollTop < 0) {
      scrollTop = 0;
    }

    if (scrollLeft < 0) {
      scrollLeft = 0;
    }

    this.scrollTop = scrollTop;
    this.scrollLeft = scrollLeft;
    callback && callback(this);
    this.onChanged();
    return this;
  }

  hasScrollToRectData() {
    return this.scrollToRectData !== null;
  }

  getScrollToRectData() {
    return [this.scrollToRectData.rectMS, this.scrollToRectData.padding, this.scrollToRectData.callback];
  }

  setScrollToRectData(rectMS, padding, callback) {
    this.scrollToRectData = {
      rectMS: rectMS,
      padding: padding,
      callback: callback
    };
  }

  resetScrollToRectData() {
    this.scrollToRectData = null;
  }

}

exports.MockupViewport = MockupViewport;

/***/ }),

/***/ "./js/editor/viewmodels/resourcedisplayviewmodel.js":
/*!**********************************************************!*\
  !*** ./js/editor/viewmodels/resourcedisplayviewmodel.js ***!
  \**********************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ResourceDisplayViewModel = exports.KeyNames = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _mockupviewport = __webpack_require__(/*! editor/viewmodels/mockupviewport */ "./js/editor/viewmodels/mockupviewport.js");

var _projectsettings = __webpack_require__(/*! editor/projectsettings */ "./js/editor/projectsettings.js");

var _events = _interopRequireDefault(__webpack_require__(/*! core/events */ "./js/core/events.ts"));

var _debounce = __webpack_require__(/*! core/utils/debounce */ "./js/core/utils/debounce.ts");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var ViewModel = __webpack_require__(/*! core/viewmodels/viewmodel */ "./js/core/viewmodels/viewmodel.ts");

var Project = __webpack_require__(/*! core/model/project */ "./js/core/model/project.js");

var Constants = __webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts");

var KeyNames = {
  // input
  // -----
  PROJECT: "project",
  CURRENT_MOCKUP_INDEX: "currentMockupIndex",
  // output
  // ------
  VIEWPORT: "viewport",
  SHOW_MARKUP: "showMarkup",
  SHOW_OTHER_USERS_SELECTION: "showOtherUsersSelection",
  HIDE_OTHER_CONTROLS: "hideOtherControls" // in symbols view

};
exports.KeyNames = KeyNames;
const SETTINGS_SAVE_TIMEOUT = 1000;

class ResourceDisplayViewModel extends ViewModel.ViewModel {
  init() {
    this.viewports = {};
    this.project = null;
    this.showMarkup = true; // this.pngTransparentBackground = false;

    this.hideOtherControls = false;
    this.showOtherUsersSelection = true;
    this.targetWireframeMouseCoords = _types.default.newPoint(0, 0);
    this.viewportsContext = {
      onViewportChanged: this.onViewportChanged.bind(this)
    };
    this.mockupsViewportData = {};
    this.saveDisabled = false;
    this.saveViewportSettingsDebouncedByID = {};
  }

  onViewportChanged(viewport) {
    if (!this.saveViewportSettingsDebouncedByID[viewport.resourceID]) {
      this.saveViewportSettingsDebouncedByID[viewport.resourceID] = (0, _debounce.debounce)(this.saveViewportSettings.bind(this), SETTINGS_SAVE_TIMEOUT);
    }

    this.saveViewportSettingsDebouncedByID[viewport.resourceID](viewport);
  }

  loadSettings() {
    _balsamiq.default.exists(this.project);

    this.showMarkup = (0, _projectsettings.loadProjectSettings)(this.context.settings, this.project, _projectsettings.ProjectSettings.MarkupVisible); // this.pngTransparentBackground = loadProjectSettings(this.context.settings, this.project, ProjectSettings.PngExportTransparentBackground);

    this.hideOtherControls = !(0, _projectsettings.loadProjectSettings)(this.context.settings, this.project, _projectsettings.ProjectSettings.OtherSymbolsVisible);
    this.saveDisabled = true;
    this.mockupsViewportData = (0, _projectsettings.loadProjectSettings)(this.context.settings, this.project, _projectsettings.ProjectSettings.MockupsViewportData);

    for (let key in this.mockupsViewportData) {
      this.createViewportFromDump(this.mockupsViewportData[key]);
    }

    this.saveDisabled = false;
    this.notifyKeyValueChanged(KeyNames.SHOW_MARKUP);
    this.notifyKeyValueChanged(KeyNames.HIDE_OTHER_CONTROLS);
    this.notifyKeyValueChanged(KeyNames.SHOW_OTHER_USERS_SELECTION);
  }

  saveViewportSettings(viewport) {
    if (this.project && !this.saveDisabled) {
      this.mockupsViewportData[this.getKeyForViewport(viewport)] = this.dumpViewport(viewport);
      (0, _projectsettings.saveProjectSettings)(this.context.settings, this.project, _projectsettings.ProjectSettings.MockupsViewportData, this.mockupsViewportData);
    }
  }

  getTargetWireframeMouseCoords() {
    return _types.default.newPoint(this.targetWireframeMouseCoords.x, this.targetWireframeMouseCoords.y);
  }

  setTargetWireframeMouseCoords(x, y) {
    this.targetWireframeMouseCoords.x = x;
    this.targetWireframeMouseCoords.y = y;
  } // Add viewport to my list, replace object if already there


  setViewport(resourceID, branchID, viewport) {
    _balsamiq.default.exists(resourceID);

    _balsamiq.default.exists(branchID);

    _balsamiq.default.assertClass(viewport, _mockupviewport.MockupViewport);

    if (!this.viewports[resourceID]) {
      this.viewports[resourceID] = {};
    }

    if (this.viewports[resourceID][branchID]) {
      this.viewports[resourceID][branchID].initFromViewport(viewport);
    } else {
      this.viewports[resourceID][branchID] = viewport;
    } // Update settings for viewport


    this.saveViewportSettings(viewport);
  }

  getViewport(resourceID, branchID) {
    if (this.viewports[resourceID]) {
      return this.viewports[resourceID][branchID];
    }

    return null;
  }

  getKeyValue(keyName) {
    switch (keyName) {
      case KeyNames.SHOW_MARKUP:
        return this.showMarkup;

      case KeyNames.VIEWPORT:
        return this.viewports;

      case KeyNames.HIDE_OTHER_CONTROLS:
        return this.hideOtherControls;

      case KeyNames.SHOW_OTHER_USERS_SELECTION:
        return this.showOtherUsersSelection;

      case KeyNames.PROJECT:
        return this.project;

      default:
        return super.getKeyValue(keyName);
    }
  }

  setKeyValue(keyName, keyValue) {
    switch (keyName) {
      case KeyNames.PROJECT:
        if (this.project && _balsamiq.default.isEqual(this.project.coreObjectID, _balsamiq.default.get(keyValue, "coreObjectID"))) {
          return;
        }

        if (this.project && this.project !== keyValue) {
          this.context.eventCenter.removeListener({
            projectID: this.project.coreObjectID.projectID
          }, Project.EventTypes.RESOURCE_BRANCHID_CHANGED, this, this.onResourceBranchChanged);
        }

        this.project = keyValue;

        if (this.project != null) {
          this.context.eventCenter.addListener({
            projectID: this.project.coreObjectID.projectID
          }, Project.EventTypes.RESOURCE_BRANCHID_CHANGED, this, this.onResourceBranchChanged);
          this.loadSettings();
        }

        break;

      case KeyNames.CURRENT_MOCKUP_INDEX:
        Object.keys(this.saveViewportSettingsDebouncedByID).forEach(id => {
          this.saveViewportSettingsDebouncedByID[id].flush();
          delete this.saveViewportSettingsDebouncedByID[id];
        });
        break;

      default:
        super.setKeyValue(keyName, keyValue);
    }
  }

  onResourceBranchChanged(event) {
    var resourceID = event.eventData.resourceID;
    var oldBranchID = event.eventData.oldBranchID;
    var newBranchID = event.eventData.newBranchID;
    var viewportsForResource = this.viewports[resourceID];

    if (viewportsForResource && viewportsForResource[oldBranchID]) {
      let viewport = viewportsForResource[oldBranchID]; // Remove settings data relative to old branch

      delete this.mockupsViewportData[this.getKeyForViewport(viewport)]; // Move same viewport to new branch and delete old reference

      viewportsForResource[newBranchID] = viewport;
      delete viewportsForResource[oldBranchID]; // Change branch also inside viewport structure

      viewport.branchID = newBranchID; // Save viewport with new branch in settings

      this.saveViewportSettings(viewport);
    }
  }

  toggleShowMarkup() {
    this.showMarkup = !this.showMarkup;
    this.notifyKeyValueChanged(KeyNames.SHOW_MARKUP);
    (0, _projectsettings.saveProjectSettings)(this.context.settings, this.project, _projectsettings.ProjectSettings.MarkupVisible, this.showMarkup);
  }

  toggleHideOtherControls() {
    this.hideOtherControls = !this.hideOtherControls;
    this.notifyKeyValueChanged(KeyNames.HIDE_OTHER_CONTROLS);
    (0, _projectsettings.saveProjectSettings)(this.context.settings, this.project, _projectsettings.ProjectSettings.OtherSymbolsVisible, !this.hideOtherControls);
  }

  toggleShowOtherUsersSelection() {
    this.showOtherUsersSelection = !this.showOtherUsersSelection;
    this.notifyKeyValueChanged(KeyNames.SHOW_OTHER_USERS_SELECTION);
  }

  zoomAll100(_ref) {
    let {
      type
    } = _ref;

    _balsamiq.default.assertEnum(type, Constants.ResourceTypes);

    var mockupsToZoom;

    if (type === Constants.ResourceTypes.MOCKUP) {
      mockupsToZoom = this.project.getFlatMockupList();
    } else {
      mockupsToZoom = this.project.getSymbolLibraryList();
    }

    for (var i = 0; i < mockupsToZoom.length; ++i) {
      let mockup = mockupsToZoom[i];
      let viewport = this.getOrCreateViewport(mockup.coreObjectID.resourceID, mockup.coreObjectID.branchID);
      viewport.zoom100();
      viewport.setViewScroll(0, 0);
    }
  }

  zoomAllFit(_ref2) {
    let {
      viewWidth,
      viewHeight,
      type
    } = _ref2;

    _balsamiq.default.assertEnum(type, Constants.ResourceTypes);

    _balsamiq.default.assertNumber(viewHeight);

    _balsamiq.default.assertNumber(viewWidth);

    var mockupsToZoom;

    if (type === Constants.ResourceTypes.MOCKUP) {
      mockupsToZoom = this.project.getFlatMockupList();
    } else {
      mockupsToZoom = this.project.getSymbolLibraryList();
    }

    for (var i = 0; i < mockupsToZoom.length; ++i) {
      let mockup = mockupsToZoom[i];
      let bbox = mockup.getMockupArea();
      let viewport = this.getOrCreateViewport(mockup.coreObjectID.resourceID, mockup.coreObjectID.branchID);
      viewport.zoomFit(viewWidth, viewHeight, bbox.width, bbox.height, bbox.y, bbox.x);
    }
  }

  createViewport(resourceID, branchID, data) {
    var viewport = new _mockupviewport.MockupViewport(this.viewportsContext, resourceID, branchID, data);
    this.setViewport(resourceID, branchID, viewport);
    return viewport;
  }

  getOrCreateViewport(resourceID, branchID) {
    var viewport = this.getViewport(resourceID, branchID);

    if (!viewport) {
      viewport = this.createViewport(resourceID, branchID);
      this.setViewport(resourceID, branchID, viewport);
    }

    return viewport;
  }

  dumpViewport(viewport) {
    return {
      resourceID: viewport.resourceID,
      branchID: viewport.branchID,
      zoom: viewport.zoom,
      scrollLeft: viewport.scrollLeft,
      scrollTop: viewport.scrollTop
    };
  }

  createViewportFromDump(dump) {
    _balsamiq.default.assertUUID(dump.resourceID);

    _balsamiq.default.assertBranch(dump.branchID);

    var viewport = new _mockupviewport.MockupViewport(this.viewportsContext, dump.resourceID, dump.branchID, {
      scrollLeft: dump.scrollLeft,
      scrollTop: dump.scrollTop,
      zoom: dump.zoom
    });
    this.setViewport(dump.resourceID, dump.branchID, viewport);
  }

  getKeyForViewport(viewport) {
    return `${viewport.resourceID}-${viewport.branchID}`;
  }

}

exports.ResourceDisplayViewModel = ResourceDisplayViewModel;

/***/ }),

/***/ "./js/macos/common/native-macos.js":
/*!*****************************************!*\
  !*** ./js/macos/common/native-macos.js ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));

var Native = _interopRequireWildcard(__webpack_require__(/*! core/native */ "./js/core/native.js"));

Object.keys(Native).forEach(function (key) {
  if (key === "default" || key === "__esModule") return;
  if (key in exports && exports[key] === Native[key]) return;
  Object.defineProperty(exports, key, {
    enumerable: true,
    get: function () {
      return Native[key];
    }
  });
});

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var Constants = _interopRequireWildcard(__webpack_require__(/*! core/native-constants */ "./js/core/native-constants.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

/* global HostResetNative, RTCSocketIONative, HttpRequestNative, PathLibraryNative, CloudNative, PDFNative, ClipboardNative, UnsavedFilesNative, LocalFilesNative, UserInfoNative, LicenseNative, URLSelectorNative */
Native.Host.runtimeInfo = function () {
  return Host.runtimeInfo();
};

Native.Host.versionInfo = function () {
  return Host.versionInfo();
};

Native.Host.productName = function () {
  return "Balsamiq Wireframes";
};

Native.Host.init = function (params, callback) {
  Native.Host.log("Initializing host");

  if (callback) {
    Native.Host.asyncCallback(callback, []);
  }
};

Native.Host.log = function (obj) {
  Host.logWithMessage(obj);
};

Native.Host.collectDevData = function (code, message, details, error) {
  Host.collectDevDataWithCodeMessageDetailsError(code, message, details, error ? error.toString() : "-");
};

Native.Host.asyncCallback = function (callback, params) {
  // alternate implementation: do not pass params to native, rather
  // use them in the closure
  // e.g.
  // _asyncCallback( () => { callback(params)})
  Host.asyncCallbackWithCallbackParams(() => {
    callback.apply(null, params);
  }, {});
};

Native.Host.timer = function (ms, callback) {
  // returns id that can be used to cancel this timer
  return Host.timerWithMsCallback(ms, callback);
};

Native.Host.clearTimer = function (id) {
  Host.clearTimerWithId(id);
};

Native.Host.openURL = function (url) {
  Host.openURLWithUrl(url);
};

Native.Host.nativeAction = function (id, params, callback) {
  Host.nativeActionWithIdParamsCallback(id, params, callback);
};

Native.Host.createZipFile = function (snapshotsData, saveParams, callback) {
  Host.createZipFileWithSnapshotsDataSaveParamsCallback(snapshotsData, saveParams, callback);
};

Native.Host.canSaveFile = function (callback) {
  Native.Host.asyncCallback(callback, [true]);
};

Native.Host.saveFile = function (fileName, base64Data, mimeType, isPotentiallyDangerous, callback) {
  Host.saveFileWithFileNameBase64DataMimeTypeIsPotentiallyDangerousCallback(fileName, base64Data, mimeType, isPotentiallyDangerous, callback);
};

const convertErrorToMap = errorStack => Object.getOwnPropertyNames(errorStack).reduce((errorObj, key) => {
  errorObj[key] = errorStack[key];
  return errorObj;
}, {});

Native.Host._abort = function (message, errorStack) {
  Host.abortWithMessageErrorStack(message, convertErrorToMap(errorStack));
};

Native.Host._reset = function () {
  Host.reset();
};

Native.Host.UI = {
  layers: {},
  make: function (type, options, state, callback) {
    UINative.makeWithTypeOptionsStateCallback(type, options, state, callback);
  },
  setStates: function (ids, states, changedStateKeys) {
    _balsamiq.default.assert(ids.length === states.length);

    UINative.setStatesWithIdsStatesChangedStateKeys(ids, states, changedStateKeys);
  },
  renderRoot: function () {
    Native.Host.log("Native.Host.UI.renderRoot");
  },
  getRootDivID: function () {
    Native.Host.log("Native.Host.UI.getRootDivID");
  },
  rootSizeChanged: function () {
    Native.Host.log("Native.Host.UI.rootSizeChanged");
  },
  enableKeyboardEvents: function () {
    Native.Host.log("Native.Host.UI.enableKeyboardEvents");
  },
  abort: function (message) {
    Native.Host.log("Native.Host.UI.abort: " + message);
  },
  // public
  makeWindow: function (options, callback) {
    UINative.makeWindowWithOptionsCallback(options, res => {
      this.layers[res.id] = [];
      callback(res);
    });
  },
  push: function (windowID, viewControllerID, options, callback) {
    Native.Host.log("pushing: " + windowID + " -> " + viewControllerID);
    UINative.pushWithWindowIDControllerIDOptionsCallback(windowID, viewControllerID, options, () => {
      this.layers[windowID].push(viewControllerID);
      callback();
    });
  },
  pop: function (windowID, callback) {
    Native.Host.log("popping: " + windowID);
    UINative.popWithWindowIDCallback(windowID, res => {
      this.layers[windowID].pop();
      Native.Host.asyncCallback(callback, [res]);
    });
  },
  getTop: function (windowID) {
    return this.layers[windowID].length > 0 ? this.layers[windowID][this.layers[windowID].length - 1] : null;
  },
  sendCommands: function (ids, commandArrays) {
    UINative.sendCommandsWithIdsCommands(ids, commandArrays);
  },
  setChild: function (parentID, childName, childID, callback) {
    UINative.setChildWithParentIDChildNameChildIDCallback(parentID, childName, childID, callback);
  },
  remove: function (id, callback) {
    UINative.removeWithIdCallback(id, callback);
  },
  saveFileDialog: function (cfg, callback) {
    UINative.saveFileDialogWithCfgCallback(cfg, callback);
  },

  showOpenFileDialog(cfg, callback) {
    UINative.showOpenFileDialogWithCfgCallback(cfg, callback);
  },

  getPickedFile(fileInfo, callback) {
    callback(fileInfo);
  },

  goFullscreen: function () {
    Native.Host.log("Native.Host.UI.goFullscreen");
  },
  exitFullscreen: function () {
    Native.Host.log("Native.Host.UI.exitFullscreen");
  },
  requestFocus: function (id) {
    UINative.requestFocusWithId(id);
  }
};
Native.Host.LocalStorage = {
  setData: function (data, callback) {
    Host.localStorageSetDataWithDataCallback(data, callback);
  },
  getData: function (callback) {
    Host.localStorageGetDataWithCallback(callback);
  }
};
Native.Host.SessionStorage = {
  setData: function (data, callback) {
    Host.sessionStorageSetDataWithDataCallback(data, callback);
  },
  getData: function (callback) {
    Host.sessionStorageGetDataWithCallback(callback);
  }
};
Native.Host.SQLite = {
  init: function (callback, progressCallback) {
    Native.Host.abort("Native.Host.SQLite.init unimplemented");
  },
  status: function () {
    return Constants.ModuleStatus.Initialized;
  },
  has: function (id) {
    return SQLiteNative.hasWithId(id);
  },
  makeFromFileInfo: function (fileInfo, callback) {
    _balsamiq.default.assertFunction(callback);

    SQLiteNative.makeFromFileInfoWithFileInfoCallback(fileInfo, callback);
  },
  makeFromData: function (uint8Array, callback) {
    _balsamiq.default.assertFunction(callback); // callback({ id: string, error: string });


    SQLiteNative.makeFromDataWithUint8ArrayCallback(uint8Array, callback);
  },
  makeFromFileURL: function (fileURL, callback) {
    _balsamiq.default.assertString(fileURL);

    _balsamiq.default.assertFunction(callback); // callback({ id: string, error: string });


    SQLiteNative.makeFromFileURLWithFileURLCallback(fileURL, callback);
  },
  remove: function (id, callback) {
    _balsamiq.default.assertString(id);

    _balsamiq.default.assertFunction(callback);

    SQLiteNative.removeWithIdCallback(id, callback);
  },
  open: function (id, mode, callback) {
    _balsamiq.default.assertString(id);

    _balsamiq.default.assertFunction(callback);

    SQLiteNative.openWithIdModeCallback(id, mode, callback);
  },
  close: function (id, callback) {
    _balsamiq.default.assertString(id);

    _balsamiq.default.assertFunction(callback);

    SQLiteNative.closeWithIdCallback(id, callback); // callback({ error: string })
  },
  isOpen: function (id) {
    _balsamiq.default.assertString(id);

    return SQLiteNative.isOpenWithId(id);
  },
  query: function (id, query, parameters, callback) {
    _balsamiq.default.assertString(id);

    _balsamiq.default.assertString(query);

    _balsamiq.default.assertFunction(callback);

    SQLiteNative.queryWithIdQueryParametersCallback(id, query, parameters, callback);
  },
  tables: function (id, callback) {
    _balsamiq.default.assertString(id);

    _balsamiq.default.assertFunction(callback);

    SQLiteNative.tablesWithIdCallback(id, callback);
  },
  download: function
    /*id, callback, progressCallback*/
  () {
    Native.Host.abort("Native.Host.SQLite.download unimplemented");
  },
  reset: function (callback) {
    _balsamiq.default.assertFunction(callback);

    SQLiteNative.resetWithCallback(callback);
  },
  compact: function (id, callback) {
    _balsamiq.default.assertString(id);

    _balsamiq.default.assertFunction(callback);

    SQLiteNative.compactWithIdCallback(id, callback);
  }
};
Native.Host.Fonts = {
  getFontFamilies: function () {
    return FontsNative.getFontFamilies(); // return Object fontFamily:string -> { regular: bool, bold: bool, italic: bool, boldItalic: bool }
  },
  glyphsInFont: function (fontFamily, bold, italic, text, callback) {
    FontsNative.glyphsInFontWithFontFamilyBoldItalicTextCallback(fontFamily, bold, italic, text, callback);
  }
};
Native.Host.TextMeterHelper = {
  setTextStyle: function (fontFamily, fontSize, bold, italic) {
    TextMeterNative.setTextStyleWithFamilySizeBoldItalic(fontFamily, fontSize, bold, italic);
  },
  measureStringWidth: function (text) {
    return TextMeterNative.measureStringWidthWithText(text);
  }
};
Native.Host.Images = {
  make: function (callback) {
    ImagesNative.makeWithCallback(callback);
  },
  makeSync: function () {
    // return string id
    return ImagesNative.makeSync();
  },
  drawCommands: function (id, commands, callback) {
    ImagesNative.drawCommandsWithIdCommandsCallback(id, commands, callback);
  },
  drawCommandsSync: function (id, commands, callback) {
    // invoke callback immediately! callback({ })
    ImagesNative.drawCommandsSyncWithIdCommandsCallback(id, commands, callback);
  },
  makeFromURL: function (url, callback) {
    ImagesNative.makeFromURLWithUrlCallback(url, callback);
  },
  makeFromBase64Data: function (mimeType, data, callback) {
    ImagesNative.makeFromBase64DataWithMimeTypeDataCallback(mimeType, data, callback);
  },
  getImageBase64Data: function (id, callback) {
    ImagesNative.getImageBase64DataWithIdCallback(id, callback);
  },
  remove: function (id, callback) {
    ImagesNative.removeWithIdCallback(id, callback);
  },
  saveImage: function (id, saveParams, isPotentiallyDangerous, callback) {
    ImagesNative.saveImageWithIdSaveParamsIsPotentiallyDangerousCallback(id, saveParams, isPotentiallyDangerous, callback);
  },
  parallelProcessingCapability: function () {
    return 1;
  },
  resize: function (id, width, height, callback) {
    ImagesNative.resizeWithIdWidthHeightCallback(id, width, height, callback);
  },
  getJPGBase64Data: function (id, quality, callback) {
    ImagesNative.getJPGBase64DataWithIdQualityCallback(id, quality, callback);
  },
  usesAlpha: function (id, callback) {
    ImagesNative.usesAlphaWithIdCallback(id, callback);
  },
  getMaxImageSize: function () {
    return {
      width: 20000,
      height: 20000
    };
  }
};
Native.Host.PDF = {
  init: function (callback, progressCallback) {
    Native.Host.asyncCallback(() => {
      callback({});
    });
  },
  status: function () {
    return Native.ModuleStatus.Initialized;
  },
  start: function (options, fontFamilies, callback) {
    PDFNative.startWithOptionsFontFamiliesCallback(options, fontFamilies, callback); // callback({ id: string })
  },
  addPage: function (id, width, height, layout, pageName, callback) {
    PDFNative.addPageWithIdWidthHeightLayoutPageNameCallback(id, width, height, layout, pageName, callback);
  },
  commands: function (id, commands, callback) {
    PDFNative.commandsWithIdCommandsCallback(id, commands, callback);
  },
  save: function (id, saveParams, callback) {
    PDFNative.saveWithIdSaveParamsCallback(id, saveParams, callback);
  }
};
Native.Host.Cursors = {
  // for now, cursors are added by the Native side and must implement the IDs definied by cursors.js
  getCursorIDs: function (callback) {
    // callback({ cursorIDs: [String] })
    Native.Host.abort("Native.Host.Cursors.getCursorIDs unimplemented");
  }
};
Native.Host.Clipboard = {
  setData: function (data, callback) {
    ClipboardNative.setDataWithDataCallback(data, callback);
  },
  getData: function (callback) {
    return ClipboardNative.getDataWithCallback(callback);
  },

  canGetDataFromSystem() {
    return true;
  },

  canPasteFromMenu() {
    return true;
  },

  canCopyImageToSystem() {
    return true;
  }

};
Native.Host.HttpRequest = {
  ajax: function (options) {
    return HttpRequestNative.ajaxWithOptionsCallback(options, options.callback);
  },
  get: function (url, callback) {
    return HttpRequestNative.getWithUrlValueCallback(url, callback);
  },
  getBinary: function (url, params, callback, requestCreatedCallback) {
    return HttpRequestNative.getBinaryWithUrlValueParamsCallbackRequestCreatedCallback(url, params, callback, function (id) {
      var nativeRequest = {
        id: id,
        abort: function () {
          HttpRequestNative.abortWithId(id);
        }
      };

      if (requestCreatedCallback) {
        requestCreatedCallback(nativeRequest);
      }
    });
  },
  post: function () {
    Native.Host.abort("Native.Host.HttpRequest.post unimplemented");
  },
  postFormData: function (url, paramsArray, callback) {
    return HttpRequestNative.postFormDataWithUrlValueParamsCallback(url, paramsArray, callback);
  }
};
Native.Host.Paths = {
  Formats: Native.Host.Paths.Formats,
  getFormat: function () {
    return Native.Host.Paths.Formats.RENDERER_COMMANDS;
  },
  addSkinFragments: function (obj, skinName) {
    _balsamiq.default.assertString(skinName, "Error in adding skin fragments. skinName should be a string");

    for (let key in obj) {
      _balsamiq.default.assertNumber(obj[key].width, "Error in skin fragment. width should be a number: " + JSON.stringify(obj[key]));

      _balsamiq.default.assertNumber(obj[key].height, "Error in skin fragment. height should be a number: " + JSON.stringify(obj[key]));

      _balsamiq.default.assertArray(obj[key].paths, "Error in skin fragment. paths should be an array: " + JSON.stringify(obj[key]));

      _balsamiq.default.assertArray(obj[key].colors, "Error in skin fragment. colors should be an array: " + JSON.stringify(obj[key]));
    }

    PathLibraryNative.addSkinFragmentsWithObjSkinName(obj, skinName);
  },
  supportsPathLibrary: function () {
    return false;
  }
};

Native.Host.Cloud.autologin = function (callback) {
  CloudNative.autologinWithCallback(callback);
};

Native.Host.Cloud.logout = function (callback) {
  CloudNative.logoutWithCallback(callback);
};

Native.Host.Cloud.query = function (query, variables, callback) {
  _balsamiq.default.assertString(query);

  _balsamiq.default.assertObject(variables);

  _balsamiq.default.assertFunction(callback);

  CloudNative.queryWithQueryVariablesCallback(query, variables, callback);
};

Native.Host.Cloud.getOpenToken = function (siteID36, projectID36, callback) {
  _balsamiq.default.assertString(siteID36);

  _balsamiq.default.assertString(projectID36);

  CloudNative.getOpenTokenWithSiteID36ProjectID36Callback(siteID36, projectID36, callback);
};

Native.Host.Cloud.uploadProject = function (projectFileUrl, projectName, siteID36, progressCallback, finishCallback) {
  _balsamiq.default.assertString(projectFileUrl);

  _balsamiq.default.assertString(projectName);

  _balsamiq.default.assertString(siteID36);

  _balsamiq.default.assertFunction(progressCallback);

  _balsamiq.default.assertFunction(finishCallback);

  CloudNative.uploadProjectWithProjectFileUrlProjectNameSiteID36ProgressCallbackFinishCallback(projectFileUrl, projectName, siteID36, progressCallback, finishCallback);
};

Native.Host.Cloud.cancelUploadProject = function (projectFileUrl, siteID36) {
  _balsamiq.default.assertString(projectFileUrl);

  _balsamiq.default.assertString(siteID36);

  CloudNative.cancelUploadProjectWithProjectFileUrlSiteID36(projectFileUrl, siteID36);
};

Native.Host.Cloud.canLogin = function (callback) {
  _balsamiq.default.assertFunction(callback);

  CloudNative.canLoginWithCallback(callback);
};

Native.Host.Cloud.userSettings = function () {
  CloudNative.userSettings();
};

Native.Host.Cloud.openLoginURL = function () {
  CloudNative.openLoginURL();
};

Native.Host.Cloud.getAPIUserAuthToken = function (callback) {
  _balsamiq.default.assertFunction(callback);

  CloudNative.getAPIUserAuthTokenWithCallback(callback);
};

Native.Host.Cloud.spaceSettings = function (spaceID) {
  CloudNative.spaceSettingsWithSpaceID(spaceID);
};

Native.Host.Cloud.getDomain = function () {
  return CloudNative.getDomain();
};

Native.Host.UnsavedFiles.isUnsavedFileURL = function (fileURL, callback) {
  _balsamiq.default.assertString(fileURL);

  _balsamiq.default.assertFunction(callback);

  UnsavedFilesNative.isUnsavedFileURLWithFileURLCallback(fileURL, callback);
};

Native.Host.UnsavedFiles.createUnsavedFile = function (fileName, ext, callback) {
  _balsamiq.default.assertString(fileName);

  _balsamiq.default.assertString(ext);

  _balsamiq.default.assertFunction(callback);

  UnsavedFilesNative.createUnsavedFileWithFileNameExtCallback(fileName, ext, callback);
};

Native.Host.UnsavedFiles.removeUnsavedFile = function (unsavedFileURL, callback) {
  _balsamiq.default.assertString(unsavedFileURL);

  _balsamiq.default.assertFunction(callback);

  UnsavedFilesNative.removeUnsavedFileWithUnsavedFileURLCallback(unsavedFileURL, callback);
};

Native.Host.UnsavedFiles.copyUnsavedFileTo = function (unsavedFileURL, userFileURL, callback) {
  _balsamiq.default.assertString(unsavedFileURL);

  _balsamiq.default.assertFunction(callback);

  UnsavedFilesNative.copyUnsavedFileToUnsavedFileURLUserFileURLCallback(unsavedFileURL, userFileURL, callback);
};

Native.Host.UnsavedFiles.markUnsavedFileAsCorrupted = function (unsavedFileURL, callback) {
  _balsamiq.default.assertString(unsavedFileURL);

  _balsamiq.default.assertFunction(callback);

  UnsavedFilesNative.markUnsavedFileAsCorruptedWithUnsavedFileURLCallback(unsavedFileURL, callback);
};

Native.Host.LocalFiles.isLocalFileURL = function (fileURL, callback) {
  _balsamiq.default.assertString(fileURL);

  _balsamiq.default.assertFunction(callback);

  LocalFilesNative.isLocalFileURLWithFileURLCallback(fileURL, callback);
};

Native.Host.LocalFiles.getLocalFileFrom = function (userFileURL, callback) {
  _balsamiq.default.assertString(userFileURL);

  _balsamiq.default.assertFunction(callback);

  LocalFilesNative.getLocalFileFromUserFileURLCallback(userFileURL, callback);
};

Native.Host.LocalFiles.createLocalFileFrom = function (userFileURL, callback) {
  _balsamiq.default.assertString(userFileURL);

  _balsamiq.default.assertFunction(callback);

  LocalFilesNative.createLocalFileFromUserFileURLCallback(userFileURL, callback);
};

Native.Host.LocalFiles.copyLocalFileTo = function (userFileURL, callback) {
  _balsamiq.default.assertString(userFileURL);

  _balsamiq.default.assertFunction(callback);

  LocalFilesNative.copyLocalFileToUserFileURLCallback(userFileURL, callback);
};

Native.Host.LocalFiles.copyLocalFileToNewUserFile = function (userFileURL, newUserFileURL, callback) {
  _balsamiq.default.assertString(userFileURL);

  _balsamiq.default.assertString(newUserFileURL);

  _balsamiq.default.assertFunction(callback);

  LocalFilesNative.copyLocalFileToNewUserFileWithUserFileURLNewUserFileURLCallback(userFileURL, newUserFileURL, callback);
};

Native.Host.LocalFiles.editLocalFileOf = function (userFileURL, callback) {
  _balsamiq.default.assertString(userFileURL);

  _balsamiq.default.assertFunction(callback);

  LocalFilesNative.editLocalFileOfUserFileURLCallback(userFileURL, callback);
};

Native.Host.LocalFiles.removeLocalFileFromRegistry = function (userFileURL, callback) {
  _balsamiq.default.assertString(userFileURL);

  _balsamiq.default.assertFunction(callback);

  LocalFilesNative.removeLocalFileFromRegistryWithUserFileURLCallback(userFileURL, callback);
};

Native.Host.LocalFiles.removeLocalFile = function (userFileURL, callback) {
  _balsamiq.default.assertString(userFileURL);

  _balsamiq.default.assertFunction(callback);

  LocalFilesNative.removeLocalFileWithUserFileURLCallback(userFileURL, callback);
};

Native.Host.LocalFiles.updateLocalFileURL = function (userFileURL, localFileURL, callback) {
  _balsamiq.default.assertString(userFileURL);

  _balsamiq.default.assertString(localFileURL);

  _balsamiq.default.assertFunction(callback);

  LocalFilesNative.updateLocalFileURLWithUserFileURLLocalFileURLCallback(userFileURL, localFileURL, callback);
};

Native.Host.LocalFiles.markLocalFileAsCorruptedAndRemoveFromRegistry = function (userFileURL, callback) {
  _balsamiq.default.assertString(userFileURL);

  _balsamiq.default.assertFunction(callback);

  LocalFilesNative.markLocalFileAsCorruptedAndRemoveFromRegistryWithUserFileURLCallback(userFileURL, callback);
};

Native.Host.Desktop.License.getData = function (callback) {
  _balsamiq.default.assertFunction(callback);

  LicenseNative.getDataWithCallback(callback);
};

Native.Host.Desktop.userInfo = function (callback) {
  _balsamiq.default.assertFunction(callback);

  UserInfoNative.userInfoWithCallback(callback);
};

Native.Host.Desktop.selectURL = function (url) {
  _balsamiq.default.assertString(url);

  URLSelectorNative.selectURLWithUrl(url);
};

Native.Host._reset = function () {
  HostResetNative.resetHost();
};
/**
 * @param {string} url
 * @param {import('core/native').RTCSocketIOInitParams} params
 * @return {string}
 */


Native.Host.RTCSocketIO.init = function (url, params) {
  return RTCSocketIONative.initializeWithUrlParams(url, params);
};
/**
 * @param {string} id
 * @return {void}
 */


Native.Host.RTCSocketIO.connect = function (id) {
  RTCSocketIONative.connectWithId(id);
};
/**
 * @param {string} id
 * @return {void}
 */


Native.Host.RTCSocketIO.disconnect = function (id) {
  RTCSocketIONative.disconnectWithId(id);
};
/**
 * @param {string} id
 * @param {string} event
 * @param {unknown} payload
 * @return {void}
 */


Native.Host.RTCSocketIO.emit = function (id, event, payload) {
  RTCSocketIONative.emitWithIdEventPayload(id, event, payload);
};
/**
 * @param {string} id
 * @param {() => void} onConnectCallback
 * @return {void}
 */


Native.Host.RTCSocketIO.onConnect = function (id, onConnectCallback) {
  RTCSocketIONative.onConnectWithIdCallback(id, onConnectCallback);
};
/**
 * @param {string} id
 * @param {() => void} onDisconnectCallback
 * @return {void}
 */


Native.Host.RTCSocketIO.onDisconnect = function (id, onDisconnectCallback) {
  RTCSocketIONative.onDisconnectWithIdCallback(id, onDisconnectCallback);
};
/**
 * @param {string} id
 * @param {(message: unknown) => void} onMessageCallback
 * @return {void}
 */


Native.Host.RTCSocketIO.onMessage = function (id, onMessageCallback) {
  RTCSocketIONative.onMessageWithIdCallback(id, onMessageCallback);
};
/**
 * @param {string} id
 * @param {(data: unknown) => void} onPresenceCallback
 * @return {void}
 */


Native.Host.RTCSocketIO.onPresence = function (id, onPresenceCallback) {
  RTCSocketIONative.onPresenceWithIdCallback(id, onPresenceCallback);
};

/***/ }),

/***/ "./js/macos/exporter/exporter.js":
/*!***************************************!*\
  !*** ./js/macos/exporter/exporter.js ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ExporterAppDesktop = void 0;

var Native = _interopRequireWildcard(__webpack_require__(/*! macos/common/native-macos */ "./js/macos/common/native-macos.js"));

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _exportertask = __webpack_require__(/*! exporter/exportertask */ "./js/exporter/exportertask.ts");

var Core = _interopRequireWildcard(__webpack_require__(/*! core/core */ "./js/core/core.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

class ExporterAppDesktop {
  constructor(params) {
    _balsamiq.default.exists(params);

    _balsamiq.default.assertString(params.bmprFileURL);

    _balsamiq.default.assertString(params.pdfFileURL);

    this.params = params;
  }

  makeMainTask() {
    return new _exportertask.ExporterTask(this.params);
  }

  onMainTaskFinished(res) {
    const params = res && res.error ? {
      error: res.error
    } : {};
    Native.Host.nativeAction("ExporterEnded", params);
    this.mainTask = undefined;
  }

  startMainTask() {
    if (this.shouldQuit) {
      this.onMainTaskFinished();
      return;
    }

    _balsamiq.default.assert(this.mainTask == null, "can not start this App again because the main task is still running");

    this.mainTask = this.makeMainTask();
    this.core.tasks.runMainTask(this.windowID, this.mainTask, res => {
      this.onMainTaskFinished(res);
    });
  }

  start() {
    Native.Host.init(this.params, () => {
      Native.Host.UI.makeWindow({}, res => {
        this.windowID = res.id;
        const coreID = this.params.coreID || 0;
        this.core = Core.makeCore(Native.Host, Native.Guest, `core-${coreID}`, this.windowID, core => {
          this.startMainTask();

          if (this.params.mainTaskStartedCallback) {
            this.params.mainTaskStartedCallback();
          }
        });
      });
    });
  }

  tryQuit(closeReason) {
    this.shouldQuit = true;
    this.mainTask.tryQuit(closeReason);
  }

}

exports.ExporterAppDesktop = ExporterAppDesktop;

/***/ }),

/***/ "./js/core/colors.ts":
/*!***************************!*\
  !*** ./js/core/colors.ts ***!
  \***************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.rgb2int = exports.palette = exports.lab2int = exports.int2rgb = exports.int2lab = exports["default"] = exports.colorNames = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const colors = {};

const isNumber = function (value) {
  return typeof value === 'number' && !isNaN(value);
};

const rgb2int = function (r, g, b) {
  /*jslint bitwise: true */

  /*jshint bitwise: false */
  const res = r << 16 | g << 8 | b;
  /*jshint bitwise: true */

  /*jslint bitwise: false */

  return res;
};

exports.rgb2int = rgb2int;

const int2rgb = function (p_color) {
  // turn off bitwise operators warnings here, we know that we operating on 32-bit integers

  /*jslint bitwise: true */

  /*jshint bitwise: false */
  const red = p_color >> 16 & 0xFF;
  const green = p_color >> 8 & 0xFF;
  const blue = p_color & 0xFF;
  /*jshint bitwise: true */

  /*jslint bitwise: false */

  return [red, green, blue];
};

exports.int2rgb = int2rgb;

const perceivedBrightness = function (p_color) {
  const rgb = int2rgb(p_color);
  return (rgb[0] * 299 + rgb[1] * 587 + rgb[2] * 114) / 2550;
};

colors.labelColorGivenBackgroundColor = function (p_backgroundColor, p_alpha, p_defaultColor) {
  if (!isNumber(p_defaultColor)) {
    p_defaultColor = 0x000000;
  }

  if (!isNumber(p_alpha)) {
    p_alpha = 1;
  }

  let labelColor;

  if (p_alpha < 0.75) {
    labelColor = p_defaultColor;
  } else {
    labelColor = perceivedBrightness(p_backgroundColor) < 60 ? 0xFFFFFF : p_defaultColor;
  }

  return labelColor;
};

colors.newHexColorFromInt = function (p_color) {
  const hexStr = (+p_color).toString(16);
  return '#' + '000000'.substring(0, 6 - hexStr.length) + hexStr;
};

colors.newHexColorFromRGB = function (p_color) {
  const intCol = rgb2int(p_color[0], p_color[1], p_color[2]);
  const hexStr = (+intCol).toString(16);
  return '#' + '000000'.substring(0, 6 - hexStr.length) + hexStr;
};

colors.newHexGrayFromFloat01 = function (p_gray) {
  _balsamiq.default.assertNumber01(p_gray);

  let hexStr = (p_gray * 255).toString(16);

  if (hexStr.length === 1) {
    hexStr = '0' + hexStr;
  }

  return '#' + hexStr + hexStr + hexStr;
};
/**
 *
 * @param {Number} p_color
 * @param {Number?} p_alpha
 * @returns {string}
 */


colors.newRgbColorFromInt = function (p_color, p_alpha) {
  _balsamiq.default.assertNumber(p_color);

  if (!isNumber(p_alpha)) {
    p_alpha = 1;
  }

  _balsamiq.default.assertNumber01(p_alpha);

  const rgb = int2rgb(p_color);
  let str;

  if (p_alpha === 1) {
    str = "rgb(" + rgb[0] + "," + rgb[1] + "," + rgb[2] + ")";
  } else {
    str = "rgba(" + rgb[0] + "," + rgb[1] + "," + rgb[2] + "," + p_alpha + ")";
  }

  return str;
};

colors.applyColorMatrixToColor = function (p_matrix, p_color) {
  _balsamiq.default.assertNumber(p_color);

  const rgb = int2rgb(p_color);
  const r = rgb[0] / 255.0;
  const g = rgb[1] / 255.0;
  const b = rgb[2] / 255.0; // Alpha is ignored

  const resultR = r * p_matrix[0] + g * p_matrix[1] + b * p_matrix[2] + p_matrix[4];
  const resultG = r * p_matrix[5] + g * p_matrix[6] + b * p_matrix[7] + p_matrix[9];
  const resultB = r * p_matrix[10] + g * p_matrix[11] + b * p_matrix[12] + p_matrix[14];
  return rgb2int(resultR * 255, resultG * 255, resultB * 255);
};

colors.colorTransformMatrix = function (p_color) {
  _balsamiq.default.assertNumber(p_color);

  const rgb = int2rgb(p_color);
  const r = rgb[0] / 255.0,
        g = rgb[1] / 255.0,
        b = rgb[2] / 255.0;
  const matrix = [1 - r, 0, 0, 0, r, 0, 1 - g, 0, 0, g, 0, 0, 1 - b, 0, b, 0, 0, 0, 1, 0];
  return matrix;
};

colors.adjustBrightness = function (p_color, p_offset) {
  _balsamiq.default.assertNumber(p_color);

  const rgb = int2rgb(p_color);
  const rgbArray = rgb.map(function (c) {
    return Math.min(255, c + p_offset);
  });
  /*jslint bitwise: true */

  /*jshint bitwise: false */

  const newColor = (rgbArray[0] << 16) + (rgbArray[1] << 8) + rgbArray[2];
  /*jshint bitwise: true */

  /*jslint bitwise: false */

  return newColor;
};

const palette = [0x000000, 0x333333, 0x666666, 0x999999, 0xCCCCCC, 0xDDDDDD, 0xEEEEEE, 0xFFFFFF, 0xCF2A27, 0xFF9900, 0xFFFF00, 0x009E0F, 0x00FFFF, 0x2B78E4, 0x9900FF, 0xFF00FF, 0xEA9999, 0xF9CB9C, 0xFFE599, 0xB6D7A8, 0xA2C4C9, 0x9FC5F8, 0xB4A7D6, 0xD5A6BD, 0xE06666, 0xF6B26B, 0xFFD966, 0x93C47D, 0x76A5AF, 0x6FA8DC, 0x8E7CC3, 0xC27BA0, 0xCC0000, 0xE69138, 0xF1C232, 0x6AA84F, 0x45818E, 0x597EAA, 0x674EA7, 0xA64D79, 0x990000, 0xB45F06, 0xBF9000, 0x38761D, 0x134F5C, 0x085394, 0x351C75, 0x741B47, 0x660000, 0x783F04, 0x7F6000, 0x274E13, 0x0C343D, 0x073763, 0x20124D, 0x4C1130];
exports.palette = palette;
const colorNames = ["black", "gray1", "gray2", "gray3", "gray4", "gray5", "gray6", "white", "red", "orange", "yellow", "green", "cyan", "blue", "purple", "pink", "red2", "orange2", "yellow2", "green2", "cyan2", "blue2", "purple2", "pink2", "red3", "orange3", "yellow3", "green3", "cyan3", "blue3", "purple3", "pink3", "red4", "orange4", "yellow4", "green4", "cyan4", "blue4", "purple4", "pink4", "red5", "orange5", "yellow5", "green5", "cyan5", "blue5", "purple5", "pink5", "red6", "orange6", "yellow6", "green6", "cyan6", "blue6", "purple6", "pink6"];
exports.colorNames = colorNames;
const alternativeGrayNames = {
  grey1: "gray1",
  grey2: "gray2",
  grey3: "gray3",
  grey4: "gray4",
  grey5: "gray5",
  grey6: "gray6"
};

colors.colorFromString = function (p_string) {
  _balsamiq.default.assertString(p_string);

  p_string = p_string.toLowerCase();
  let color; // 1. alternative color name

  if (alternativeGrayNames[p_string]) {
    p_string = alternativeGrayNames[p_string];
  } // 2. color name


  const index = colorNames.indexOf(p_string);

  if (index !== -1) {
    color = palette[index];
  } // 3. hex color


  if (!isNumber(color)) {
    // remove # if present
    if (p_string.length >= 1 && p_string.charAt(0) === '#') {
      p_string = p_string.substr(1);
    } // accept only 3 or 6 digit hex code


    if (p_string.length === 3) {
      p_string = p_string[0] + p_string[0] + p_string[1] + p_string[1] + p_string[2] + p_string[2];
    }

    if (p_string.length === 6) {
      color = parseInt(p_string, 16);
    }
  } //4. the color is none of the above (it could be an incorrect color name or and incorrect hex value)


  if (!isNumber(color)) {
    color = 0;
  }

  return color;
};

colors.lab2rgb = function (lab) {
  let y = (lab[0] + 16) / 116,
      x = lab[1] / 500 + y,
      z = y - lab[2] / 200,
      r,
      g,
      b;
  x = 0.95047 * (x * x * x > 0.008856 ? x * x * x : (x - 16 / 116) / 7.787);
  y = 1.00000 * (y * y * y > 0.008856 ? y * y * y : (y - 16 / 116) / 7.787);
  z = 1.08883 * (z * z * z > 0.008856 ? z * z * z : (z - 16 / 116) / 7.787);
  r = x * 3.2406 + y * -1.5372 + z * -0.4986;
  g = x * -0.9689 + y * 1.8758 + z * 0.0415;
  b = x * 0.0557 + y * -0.2040 + z * 1.0570;
  r = r > 0.0031308 ? 1.055 * Math.pow(r, 1 / 2.4) - 0.055 : 12.92 * r;
  g = g > 0.0031308 ? 1.055 * Math.pow(g, 1 / 2.4) - 0.055 : 12.92 * g;
  b = b > 0.0031308 ? 1.055 * Math.pow(b, 1 / 2.4) - 0.055 : 12.92 * b;
  return [Math.max(0, Math.min(1, r)) * 255, Math.max(0, Math.min(1, g)) * 255, Math.max(0, Math.min(1, b)) * 255];
};

const rgb2lab = function (rgb) {
  let r = rgb[0] / 255,
      g = rgb[1] / 255,
      b = rgb[2] / 255,
      x,
      y,
      z;
  r = r > 0.04045 ? Math.pow((r + 0.055) / 1.055, 2.4) : r / 12.92;
  g = g > 0.04045 ? Math.pow((g + 0.055) / 1.055, 2.4) : g / 12.92;
  b = b > 0.04045 ? Math.pow((b + 0.055) / 1.055, 2.4) : b / 12.92;
  x = (r * 0.4124 + g * 0.3576 + b * 0.1805) / 0.95047;
  y = (r * 0.2126 + g * 0.7152 + b * 0.0722) / 1.00000;
  z = (r * 0.0193 + g * 0.1192 + b * 0.9505) / 1.08883;
  x = x > 0.008856 ? Math.pow(x, 1 / 3) : 7.787 * x + 16 / 116;
  y = y > 0.008856 ? Math.pow(y, 1 / 3) : 7.787 * y + 16 / 116;
  z = z > 0.008856 ? Math.pow(z, 1 / 3) : 7.787 * z + 16 / 116;
  return [116 * y - 16, 500 * (x - y), 200 * (y - z)];
};

const int2lab = function (intColor) {
  const rgb = int2rgb(intColor);
  return rgb2lab(rgb);
};

exports.int2lab = int2lab;

const lab2int = function (labColor) {
  const rgb = colors.lab2rgb(labColor);
  return colors.rgb2int(...rgb);
}; // calculate the perceptual distance between colors in CIELAB
// https://github.com/THEjoezack/ColorMine/blob/master/ColorMine/ColorSpaces/Comparisons/Cie94Comparison.cs


exports.lab2int = lab2int;

colors.deltaE = function (labA, labB) {
  const deltaL = labA[0] - labB[0];
  const deltaA = labA[1] - labB[1];
  const deltaB = labA[2] - labB[2];
  const c1 = Math.sqrt(labA[1] * labA[1] + labA[2] * labA[2]);
  const c2 = Math.sqrt(labB[1] * labB[1] + labB[2] * labB[2]);
  const deltaC = c1 - c2;
  let deltaH = deltaA * deltaA + deltaB * deltaB - deltaC * deltaC;
  deltaH = deltaH < 0 ? 0 : Math.sqrt(deltaH);
  const sc = 1.0 + 0.045 * c1;
  const sh = 1.0 + 0.015 * c1;
  const deltaLKlsl = deltaL / 1.0;
  const deltaCkcsc = deltaC / sc;
  const deltaHkhsh = deltaH / sh;
  const i = deltaLKlsl * deltaLKlsl + deltaCkcsc * deltaCkcsc + deltaHkhsh * deltaHkhsh;
  return i < 0 ? 0 : Math.sqrt(i);
};

colors.rgb2int = rgb2int;
colors.int2rgb = int2rgb;
colors.palette = palette;
colors.colorNames = colorNames;
colors.int2lab = int2lab;
colors.lab2int = lab2int;
colors.rgb2lab = rgb2lab;
var _default = colors;
exports["default"] = _default;

/***/ }),

/***/ "./js/core/controls/sitemapUtils.ts":
/*!******************************************!*\
  !*** ./js/core/controls/sitemapUtils.ts ***!
  \******************************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.getLabel = getLabel;
exports.getLevel = getLevel;
exports.normalizeLevels = normalizeLevels;
const INDENT_CHARS = ["-", "\t"]; // Get indentation level of line
// Root has level 0

function getLevel(line) {
  //This check stays because we call the function from a js file. Obsolete once all controls are converted to ts.
  if (!line) {
    return 0;
  }

  let level = 0;

  while (INDENT_CHARS.indexOf(line[level]) !== -1) {
    level += 1;
  }

  return level;
} // Remove leading indent chars


function getLabel(p_line) {
  //This check stays because we call the function from a js file. Obsolete once all controls are converted to ts.
  if (!p_line) {
    return "";
  }

  let i = 0;

  while (INDENT_CHARS.indexOf(p_line[i]) !== -1) {
    i += 1;
  }

  return p_line.substring(i).trim();
} // If each line is starting with one or more tabs/dashes remove them so that the root will be at level 0


function normalizeLevels(p_lines) {
  const res = [...p_lines]; //This check stays because we call the function from a js file. Obsolete once all controls are converted to ts.

  if (!p_lines) {
    return res;
  }

  let currentMaxLevel = 0;
  let lastLevel = -1;
  let clippedPreviousLine = false;

  for (let i = 0; i < res.length; i++) {
    const currentIndent = getLevel(res[i]);

    if (clippedPreviousLine && currentIndent === lastLevel) {
      //If we clipped the previous line, we need to clip this one as well. currentMaxLevel will not change.
      res[i] = res[i].substring(currentIndent - currentMaxLevel + 1);
    } else if (currentIndent > currentMaxLevel) {
      //Fix indention in case of jumps
      res[i] = res[i].substring(currentIndent - currentMaxLevel);
      currentMaxLevel++;
      clippedPreviousLine = true;
    } else {
      //We are moving up in the Tree or stay on the same level
      currentMaxLevel = currentIndent + 1;
      clippedPreviousLine = false;
    }

    lastLevel = currentIndent;
  }

  return res;
}

/***/ }),

/***/ "./js/core/core/tasks.ts":
/*!*******************************!*\
  !*** ./js/core/core/tasks.ts ***!
  \*******************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Tasks = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _task = __webpack_require__(/*! core/tasks/task */ "./js/core/tasks/task.ts");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

class Tasks {
  constructor(context) {
    _defineProperty(this, "context", void 0);

    _defineProperty(this, "data", new Map());

    _balsamiq.default.exists(context.viewControllers);

    _balsamiq.default.exists(context.viewModels);

    _balsamiq.default.exists(context.controllers);

    _balsamiq.default.exists(context.host);

    _balsamiq.default.exists(context.operationCenter);

    _balsamiq.default.exists(context.eventCenter);

    _balsamiq.default.exists(context.getCoreObject);

    _balsamiq.default.exists(context.archives);

    _balsamiq.default.exists(context.settings);

    _balsamiq.default.exists(context.model);

    _balsamiq.default.exists(context.textMeter);

    _balsamiq.default.exists(context.skins);

    _balsamiq.default.exists(context.execTime);

    _balsamiq.default.exists(context.coreID);

    this.context = context;
  }

  runMainTask(windowID, task, callback) {
    const windowTasks = this.getWindowTasks(windowID, true);

    this._runTask(windowTasks, undefined, undefined, task, false, callback);
  }

  runTask(parentTaskID, task, callback) {
    const parent = this.getTask(parentTaskID);

    if (parent) {
      this._runTask(parent.windowTasks, parent.task, parent.taskIndex, task, true, callback);
    } else {
      _balsamiq.default.assert(false, `Tasks: can not find Task with id: ${parentTaskID}`);
    }
  }

  runTaskIfPossible(parentTaskID, task, callback) {
    const parent = this.getTask(parentTaskID);

    if (parent) {
      return this._runTask(parent.windowTasks, parent.task, parent.taskIndex, task, false, callback);
    } else {
      _balsamiq.default.assert(false, `Tasks: can not find Task with id: ${parentTaskID}`);

      return false;
    }
  }

  _runTask(windowTasks, parentTask, parentTaskIndex, task, mustInterruptSubtasks, callback) {
    // try to run a task
    // if the parent task has subtasks, there are two options depending on mustInterruptSubtasks:
    // 1. true: try to interrupt the subtasks and to run the given task (async), return true
    // 2. false: don't run the given task, return false
    // check that parentTask and parentTaskIndex are both undefined or valued
    _balsamiq.default.assert(parentTask === undefined && parentTaskIndex === undefined || parentTaskIndex !== undefined && parentTaskIndex !== undefined, "Tasks: bad parameters");

    if (!parentTask) {
      // main task
      _balsamiq.default.assert(parentTaskIndex === undefined, `Tasks: bad parameters in _runTask`);

      _balsamiq.default.assert(windowTasks.stack.length === 0, `Tasks: main Task is already running in ${windowTasks.windowID}, can not run Task ${task.name()}`);
    } // the following 2 assertions can be removed once all client code is written in Typescript


    _balsamiq.default.existsClass(task, _task.TaskBase);

    _balsamiq.default.assertBool(mustInterruptSubtasks);

    if (parentTaskIndex !== undefined && parentTaskIndex !== windowTasks.stack.length - 1) {
      // the running task is not the parent task
      const tasksToInterrupt = windowTasks.stack.slice(parentTaskIndex + 1);
      const tasksToInterruptCanBeInterrupted = tasksToInterrupt.findIndex(obj => {
        return !obj.task.canInterrupt();
      }) === -1; // if interrupting the not interruptible subtasks is not required, then return false

      if (!tasksToInterruptCanBeInterrupted && !mustInterruptSubtasks) {
        return false;
      }

      if (windowTasks.afterInterrupt && !mustInterruptSubtasks) {
        return false; // a task is already waiting to run
      }

      _balsamiq.default.assert(!windowTasks.afterInterrupt, "cannot run a task when another task is waiting to run"); // interrupting is required: then the first subtask of the running task must be interrupted before running the new subtask


      const taskToInterrupt = windowTasks.stack[parentTaskIndex + 1].task;

      _balsamiq.default.assert(taskToInterrupt.canInterrupt(), "cannot run " + task.name() + " as subtask of " + windowTasks.stack[parentTaskIndex].task.name() + " because the running subtask " + taskToInterrupt.name() + " cannot be interrupted");

      windowTasks.afterInterrupt = {
        taskToRun: task,
        taskIDToWait: taskToInterrupt.getTaskID(),
        parentTaskID: parentTask.getTaskID(),
        callback: callback
      };
      taskToInterrupt.interrupt();
    } else {
      // it is the main task or the parent task is the running task. The task can run!
      const taskContext = this.makeTaskContext(windowTasks.windowID);
      const runningTask = this.runningTask(windowTasks.windowID);

      if (runningTask) {
        _balsamiq.default.extend(taskContext, runningTask.context);

        _balsamiq.default.extend(taskContext, runningTask.subtaskContext());
      }

      const taskID = windowTasks.windowID + "|task-" + windowTasks.nextTaskID;
      windowTasks.nextTaskID += 1;
      this.context.host.log(`Tasks: starting task '${task.name()}', '${taskID}'`);
      windowTasks.stack.push({
        task: task,
        callback: callback
      });
      task.run(taskContext, taskID);
    }

    return true;
  }

  onTaskFinished(windowID, taskID, res) {
    const windowTasks = this.getWindowTasks(windowID);

    _balsamiq.default.assert(windowTasks.stack[windowTasks.stack.length - 1].task.getTaskID() == taskID, "finished taskID '" + taskID + "' is not the running task");

    const stackItem = windowTasks.stack.pop();
    const error = res && res.error ? ", with error: " + res.error : "";
    this.context.host.log("Tasks: finished task '" + stackItem.task.name() + "', '" + taskID + "'" + error);

    if (stackItem && stackItem.callback) {
      stackItem.callback(res || {});
    }

    if (windowTasks.afterInterrupt) {
      if (windowTasks.afterInterrupt.taskIDToWait == taskID) {
        const obj = windowTasks.afterInterrupt;
        windowTasks.afterInterrupt = undefined;
        this.runTask(obj.parentTaskID, obj.taskToRun, obj.callback);
      }
    }
  }

  runningTask(windowID) {
    const windowTasks = this.getWindowTasks(windowID);
    return windowTasks.stack.length > 0 ? windowTasks.stack[windowTasks.stack.length - 1].task : undefined;
  }

  getTaskIndex(windowID, taskID) {
    const windowTasks = this.getWindowTasks(windowID);
    const taskIndex = windowTasks.stack.findIndex(obj => {
      return obj.task.getTaskID() == taskID;
    });
    return taskIndex;
  }

  taskHasSubtasks(windowID, taskID) {
    const windowTasks = this.getWindowTasks(windowID);

    _balsamiq.default.assertString(taskID);

    const taskIndex = this.getTaskIndex(windowID, taskID);

    _balsamiq.default.assert(taskIndex >= 0, "taskHasSubtasks: can not find taskID: " + taskID);

    return taskIndex < windowTasks.stack.length - 1;
  }

  canInterruptSubtasksOfTask(windowID, taskID) {
    const windowTasks = this.getWindowTasks(windowID);

    _balsamiq.default.assertString(taskID);

    const taskIndex = this.getTaskIndex(windowID, taskID);

    _balsamiq.default.assert(taskIndex >= 0, "canInterruptSubtasksOfTask:can not find taskID: " + taskID);

    return taskIndex == windowTasks.stack.length - 1 ? true : windowTasks.stack[taskIndex + 1].task.canInterrupt();
  }

  interruptSubtasksOfTask(windowID, taskID, callback) {
    const windowTasks = this.getWindowTasks(windowID);

    _balsamiq.default.assertString(taskID);

    _balsamiq.default.assertFunction(callback);

    _balsamiq.default.assert(this.canInterruptSubtasksOfTask(windowID, taskID), "can not interrupt subtasks of taskID: " + taskID);

    const taskIndex = this.getTaskIndex(windowID, taskID);

    _balsamiq.default.assert(taskIndex >= 0, "interruptSubtasksOfTask: can not find taskID: " + taskID);

    if (taskIndex == windowTasks.stack.length - 1) {
      this.context.host.asyncCallback(callback);
    } else {
      windowTasks.stack[taskIndex + 1].task.interrupt(callback);
    }
  }

  sendAction(taskID, action, params) {
    _balsamiq.default.assertString(taskID);

    const obj = this.getTask(taskID);

    if (!obj) {
      return false;
    }

    return obj.windowTasks.stack[obj.taskIndex].task.action(action, params);
  }

  makeTaskContext(windowID) {
    return {
      runningTask: () => {
        return this.runningTask(windowID);
      },
      runSubtask: (parentTaskID, subtask, callback) => {
        this.runTask(parentTaskID, subtask, callback);
      },
      runSubtaskIfPossible: (parentTaskID, subtask, callback) => {
        return this.runTaskIfPossible(parentTaskID, subtask, callback);
      },
      runMainTask: (windowID, task, callback) => {
        this.runMainTask(windowID, task, callback);
      },
      onTaskFinished: (taskID, res) => {
        this.onTaskFinished(windowID, taskID, res);
      },
      hasSubtasks: taskID => {
        return this.taskHasSubtasks(windowID, taskID);
      },
      canInterruptSubtasks: taskID => {
        return this.canInterruptSubtasksOfTask(windowID, taskID);
      },
      interruptSubtasks: (taskID, callback) => {
        this.interruptSubtasksOfTask(windowID, taskID, callback);
      },
      viewControllers: this.context.viewControllers,
      viewModels: this.context.viewModels,
      eventCenter: this.context.eventCenter,
      controllers: this.context.controllers,
      host: this.context.host,
      windowID: windowID,
      operationCenter: this.context.operationCenter,
      archives: this.context.archives,
      settings: this.context.settings,
      model: this.context.model,
      getCoreObject: this.context.getCoreObject,
      findCoreObjectID: this.context.findCoreObjectID,
      textMeter: this.context.textMeter,
      skins: this.context.skins,
      execTime: this.context.execTime,
      clipboard: this.context.clipboard,
      coreID: this.context.coreID
    };
  }

  getTask(taskID) {
    for (const [windowID] of this.data) {
      const windowTasks = this.getWindowTasks(windowID);

      for (let taskIndex = 0; taskIndex < windowTasks.stack.length; taskIndex++) {
        const task = windowTasks.stack[taskIndex].task;

        if (task.getTaskID() === taskID) {
          return {
            windowTasks,
            taskIndex,
            task
          };
        }
      }
    }

    return undefined;
  }

  getWindowTasks(windowID) {
    let createIfNeeded = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

    if (!this.data.has(windowID) && createIfNeeded) {
      this.data.set(windowID, {
        windowID,
        taskContext: this.makeTaskContext(windowID),
        nextTaskID: 0,
        stack: []
      });
    }

    const t = this.data.get(windowID);

    _balsamiq.default.assert(t !== undefined, `Tasks: unknown windowID: ${windowID}`);

    return t;
  }

}

exports.Tasks = Tasks;

/***/ }),

/***/ "./js/core/events.ts":
/*!***************************!*\
  !*** ./js/core/events.ts ***!
  \***************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.newCoreEventWithIndex = exports.newCoreEvent = exports["default"] = exports.EventTypes = exports.EventCenter = exports.CoreEvent = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var EventTypes;
exports.EventTypes = EventTypes;

(function (EventTypes) {
  EventTypes["ANY"] = "any";
  EventTypes["PROJECT_PROPERTY_CHANGED"] = "projectPropertyChanged";
  EventTypes["PROJECT_LOADED"] = "projectLoaded";
  EventTypes["MOCKUP_ADDED"] = "mockupAdded";
  EventTypes["MOCKUP_REMOVED"] = "mockupRemoved";
  EventTypes["MOCKUP_LOADED"] = "mockupLoaded";
  EventTypes["MOCKUP_PROPERTY_CHANGED"] = "mockupPropertyChanged";
  EventTypes["MOCKUP_SIZE_CHANGED"] = "mockupSizeChanged";
  EventTypes["MOCKUP_READY_FOR_SNAPSHOT"] = "mockupReadyForSnapshot";
  EventTypes["MOCKUP_REORDERED"] = "mockupReordered";
  EventTypes["MOCKUP_SIZE_PREVIEW"] = "mockupSizePreview";
  EventTypes["CLOUD_LOGIN"] = "cloudLogin";
  EventTypes["DESKTOP_LICENSE_CHANGED"] = "desktopLicenseChanged";
  EventTypes["SYNC_ERROR"] = "syncError";
  EventTypes["USER_INFO_CHANGED"] = "userInfoChanged";
})(EventTypes || (exports.EventTypes = EventTypes = {}));

class CoreEvent {
  constructor(senderCoreObjectID, eventType, eventData) {
    _defineProperty(this, "eventType", void 0);

    _defineProperty(this, "senderCoreObjectID", void 0);

    _defineProperty(this, "eventData", void 0);

    _defineProperty(this, "coreID", void 0);

    _balsamiq.default.assertString(eventType);

    this.senderCoreObjectID = senderCoreObjectID;
    this.eventType = eventType;
    this.eventData = eventData;
    this.coreID = null; // will be set when fired
  }

}

exports.CoreEvent = CoreEvent;

const newCoreEvent = function (coreObjectID, eventType, data) {
  return new CoreEvent(coreObjectID, eventType, data);
};

exports.newCoreEvent = newCoreEvent;

const newCoreEventWithIndex = function (coreObjectID, eventType, index) {
  return new CoreEvent(coreObjectID, eventType, {
    index: index
  });
};

exports.newCoreEventWithIndex = newCoreEventWithIndex;

class EventCenter {
  constructor(coreID) {
    _defineProperty(this, "_coreID", void 0);

    _defineProperty(this, "_listeners", void 0);

    this._coreID = coreID;
    this._listeners = {}; // key: eventType, value: Array of listeners
  }

  addListener(filter, eventType, listenerObject, listenerFunction) {
    _balsamiq.default.exists(filter);

    _balsamiq.default.assertString(eventType);

    _balsamiq.default.exists(listenerObject);

    _balsamiq.default.assertFunction(listenerFunction); // create an entry for this event type if not already there


    if (!_balsamiq.default.has(this._listeners, eventType)) {
      this._listeners[eventType] = [];
    } // search the object related to this filter, if already there


    let filterData = _lodash.default.find(this._listeners[eventType], obj => {
      return _lodash.default.isEqual(obj.filter, filter);
    }); // create an array of listeners for this sender filter if not already there


    if (!filterData) {
      filterData = {
        filter: filter,
        listeners: []
      };

      this._listeners[eventType].push(filterData);
    } // verify that the pair (obj, fun) is not already in the list


    _balsamiq.default.assert(_lodash.default.find(filterData.listeners, function (listener) {
      return listenerObject === listener.obj && listenerFunction === listener.fun;
    }) === undefined); // add the pair (object, function) to the array of listeners


    filterData.listeners.push({
      obj: listenerObject,
      fun: listenerFunction
    });
  }

  removeListener(filter, eventType, listenerObject, listenerFunction) {
    _balsamiq.default.exists(filter);

    _balsamiq.default.assertString(eventType);

    _balsamiq.default.exists(listenerObject);

    _balsamiq.default.assertFunction(listenerFunction); // verify that the event type has been already added


    _balsamiq.default.assert(_balsamiq.default.has(this._listeners, eventType), "event type not found in listeners: " + eventType); // find the filter for the coreObjectID


    const filterData = _lodash.default.find(this._listeners[eventType], function (obj) {
      return _lodash.default.isEqual(obj.filter, filter);
    });

    _balsamiq.default.assert(filterData, "CoreObjectID not found in listeners"); // remove the listener


    const filteredListeners = filterData.listeners.filter(function (listener) {
      return !(listenerObject === listener.obj && listenerFunction === listener.fun);
    });

    _balsamiq.default.assert(filteredListeners.length === filterData.listeners.length - 1, "listener not found");

    filterData.listeners = filteredListeners;
  }

  removeAllListeners(listenerObject) {
    _balsamiq.default.exists(listenerObject);

    Object.keys(this._listeners).forEach(eventType => {
      this._listeners[eventType].forEach(obj => {
        const filteredListeners = _balsamiq.default.filter(obj.listeners, function (listener) {
          return !(listenerObject === listener.obj);
        });

        obj.listeners = filteredListeners;
      });
    });
  }

  fireEvent(event) {
    // console.log("fireEvent: " + event.eventType + ", sender: " + JSON.stringify(event.senderCoreObjectID));
    event.coreID = this._coreID;
    let possibleListeners = [];

    if (this._listeners[EventTypes.ANY]) {
      possibleListeners = possibleListeners.concat(this._listeners[EventTypes.ANY]);
    }

    if (this._listeners[event.eventType]) {
      possibleListeners = possibleListeners.concat(this._listeners[event.eventType]);
    }

    possibleListeners.forEach(filterData => {
      if (event.senderCoreObjectID.matches(filterData.filter)) {
        // call the listeners
        filterData.listeners.forEach(listener => {
          listener.fun.call(listener.obj, event);
        });
      }
    });
  }

}

exports.EventCenter = EventCenter;
var _default = {
  EventTypes,
  CoreEvent,
  newCoreEvent,
  newCoreEventWithIndex,
  EventCenter
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/graphics/pathlibrary.ts":
/*!*****************************************!*\
  !*** ./js/core/graphics/pathlibrary.ts ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.addPathLibrary = addPathLibrary;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _simplerenderer = __webpack_require__(/*! core/graphics/simplerenderer */ "./js/core/graphics/simplerenderer.js");

var _sketch = _interopRequireDefault(__webpack_require__(/*! core/skins/sketch */ "./js/core/skins/sketch.js"));

var _wireframe = _interopRequireDefault(__webpack_require__(/*! core/skins/wireframe */ "./js/core/skins/wireframe.js"));

var _displayitem = _interopRequireDefault(__webpack_require__(/*! core/displayitem */ "./js/core/displayitem.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function addPathLibrary(context) {
  let nativePaths = context.host.Paths;

  if (!nativePaths.supportsPathLibrary()) {
    return;
  }

  let simpleRenderer = new _simplerenderer.SimpleRenderer(context);
  let convertToCommandsIfNeeded;

  if (nativePaths.getFormat() === nativePaths.Formats.RENDERER_COMMANDS) {
    convertToCommandsIfNeeded = paths => {
      return paths.map(path => {
        let res = path;
        simpleRenderer.drawSVGPathData(path);
        res = simpleRenderer.commands;
        simpleRenderer.commands = [];
        return res;
      });
    };
  } else {
    convertToCommandsIfNeeded = a => a;
  }

  function addSkinObjToPathLibrary(theSkin) {
    let skinObj = {};

    for (let key in theSkin) {
      if (_balsamiq.default.isObject(theSkin[key]) && theSkin[key].type === "SVG_PATHS") {
        let svg_path = theSkin[key];
        skinObj[key] = {
          paths: convertToCommandsIfNeeded(svg_path.paths),
          colors: svg_path.colors ? svg_path.colors : [],
          width: svg_path.width,
          height: svg_path.height
        };
      }
    }

    context.host.Paths.addSkinFragments(skinObj, theSkin["#skinName"]);
  }

  addSkinObjToPathLibrary(_sketch.default);
  addSkinObjToPathLibrary(_wireframe.default);
  addSkinObjToPathLibrary(_displayitem.default.AppPathsAsset);
}

/***/ }),

/***/ "./js/core/model/constants.ts":
/*!************************************!*\
  !*** ./js/core/model/constants.ts ***!
  \************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = exports.UserRoles = exports.UnnamedSymbolLibrary = exports.UnnamedSymbol = exports.UnnamedMockup = exports.UnnamedGroup = exports.ResourceTypes = exports.ProjectFontSizes = exports.OperationOrigin = exports.MockupViewPadding = exports.MaxNoteLength = exports.MaxMockupSize = exports.Master = exports.Lorem1 = exports.LicenseType = exports.FontSizes = exports.FontFamilies = exports.FirstMockupOrder = exports.EmptySymbolLibraryNotes = exports.EmptySymbolLibrary = exports.EmptyMockup = exports.ControlChanges = exports.ConnectedUserStatus = void 0;

var BarConstants = _interopRequireWildcard(__webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchiveConstants */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js"));

var BmprConstants = _interopRequireWildcard(__webpack_require__(/*! @balsamiq/bmpr/lib/BmprConstants */ "./node_modules/@balsamiq/bmpr/lib/BmprConstants.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const ControlChanges = {
  POSITION: "position",
  DISPLAY_LIST: "displayList",
  ADDED: "added",
  REMOVED: "removed",
  CHILDREN_ORDER: "childrenOrder",
  NAME: "name"
};
exports.ControlChanges = ControlChanges;
const ResourceTypes = {
  MOCKUP: BmprConstants.ResourceKinds.Mockup,
  SYMBOLLIBRARY: BmprConstants.ResourceKinds.SymbolLibrary,
  ASSET: BmprConstants.ResourceKinds.Asset
};
exports.ResourceTypes = ResourceTypes;
var OperationOrigin;
exports.OperationOrigin = OperationOrigin;

(function (OperationOrigin) {
  OperationOrigin["USER"] = "user";
  OperationOrigin["ARCHIVE"] = "archive";
})(OperationOrigin || (exports.OperationOrigin = OperationOrigin = {}));

const UserRoles = {
  NO_ACCESS: BarConstants.Role.ROLE_NO_ACCESS,
  VIEWER: BarConstants.Role.ROLE_VIEWER,
  COMMENTER: BarConstants.Role.ROLE_COMMENTER,
  EDITOR: BarConstants.Role.ROLE_EDITOR,
  ADMIN: BarConstants.Role.ROLE_ADMIN
};
exports.UserRoles = UserRoles;
const ConnectedUserStatus = {
  Active: 0,
  Waiting: 1,
  Away: 2
};
exports.ConnectedUserStatus = ConnectedUserStatus;
const LicenseType = {
  Trial: "trial",
  TrialExpired: "trialExpired",
  Ok: "ok"
};
exports.LicenseType = LicenseType;
const Master = BarConstants.Branch.MasterBranchID;
exports.Master = Master;
const EmptyMockup = {
  mockup: {
    controls: {},
    measuredH: "0",
    measuredW: "0",
    mockupH: "0",
    mockupW: "0",
    version: "1.0"
  }
};
exports.EmptyMockup = EmptyMockup;
const EmptySymbolLibrary = {
  mockup: {
    controls: {
      control: [{
        ID: "0",
        h: "248",
        measuredH: "123",
        measuredW: "109",
        properties: {
          markup: "false",
          text: "{size:16}This is a Symbol Library{size}\n \nA *Symbol Library* is a collection of *Symbols*. You can have one Symbol Library for all of your Symbols or you can create separate Symbol Libraries for different kinds of Symbols.\n \nSymbols can be used for master templates or reusable widgets such as a set of controls that are used frequently. \n \n*Double-click on a Symbol in this Symbol Library to edit it.* Or click the New Symbol icon to create a new one.\n \nLearn more at [https://balsamiq.com/wireframes/desktop/docs/symbols/](https://balsamiq.com/wireframes/desktop/docs/symbols/)."
        },
        typeID: "StickyNote",
        w: "412",
        x: "668",
        y: "20",
        zOrder: "0"
      }, {
        ID: "5",
        children: {
          controls: {
            control: [{
              ID: "0",
              h: "458",
              measuredH: "400",
              measuredW: "450",
              properties: {
                text: "A Web Page\nhttp://"
              },
              typeID: "BrowserWindow",
              w: "617",
              x: "0",
              y: "0",
              zOrder: "0"
            }, {
              ID: "1",
              h: "152",
              measuredH: "123",
              measuredW: "109",
              properties: {
                align: "left",
                color: "6989903",
                size: "14",
                text: "This is a *Symbol*. It will show up in the UI Library in the *Symbols* category. \n \nAny edits will get applied everywhere it is used. You can also override individual properties (like the selected link above - _*try it!*_)"
              },
              typeID: "StickyNote",
              w: "357",
              x: "130",
              y: "211",
              zOrder: "1"
            }, {
              ID: "4",
              h: "29",
              measuredH: "29",
              measuredW: "346",
              properties: {
                borderStyle: "none",
                selectedIndex: "0",
                text: "Home, Products, Company, Support, Contact"
              },
              typeID: "MenuBar",
              w: "609",
              x: "3",
              y: "73",
              zOrder: "2"
            }, {
              ID: "5",
              measuredH: "43",
              measuredW: "180",
              properties: {
                align: "left",
                text: "Page Title"
              },
              typeID: "Title",
              x: "16",
              y: "118",
              zOrder: "3"
            }, {
              ID: "6",
              measuredH: "25",
              measuredW: "120",
              typeID: "SearchBox",
              x: "481",
              y: "75",
              zOrder: "4"
            }]
          }
        },
        h: "458",
        measuredH: "458",
        measuredW: "617",
        properties: {
          controlName: "Sample Master Template",
          notes: "This is a sample symbol for a master template.\r\rIt is a set of controls that have been grouped and given a name.\r "
        },
        typeID: "__group__",
        w: "617",
        x: "14",
        y: "27",
        zOrder: "1"
      }, {
        ID: "7",
        children: {
          controls: {
            control: [{
              ID: "0",
              measuredH: "24",
              measuredW: "24",
              properties: {
                borderStyle: "square",
                icon: {
                  ID: "FacebookIcon",
                  size: "small"
                }
              },
              typeID: "Icon",
              x: "16",
              y: "19",
              zOrder: "1"
            }, {
              ID: "1",
              h: "54",
              measuredH: "170",
              measuredW: "200",
              properties: {
                text: "Share"
              },
              typeID: "FieldSet",
              w: "214",
              x: "0",
              y: "0",
              zOrder: "0"
            }, {
              ID: "2",
              measuredH: "24",
              measuredW: "24",
              properties: {
                borderStyle: "square",
                icon: {
                  ID: "TwitterIcon",
                  size: "small"
                }
              },
              typeID: "Icon",
              x: "48",
              y: "19",
              zOrder: "2"
            }, {
              ID: "3",
              measuredH: "24",
              measuredW: "24",
              properties: {
                borderStyle: "square",
                icon: {
                  ID: "GoogleIcon",
                  size: "small"
                }
              },
              typeID: "Icon",
              x: "79",
              y: "19",
              zOrder: "3"
            }, {
              ID: "4",
              measuredH: "24",
              measuredW: "24",
              properties: {
                borderStyle: "square",
                icon: {
                  ID: "EmailIcon",
                  size: "small"
                }
              },
              typeID: "Icon",
              x: "111",
              y: "19",
              zOrder: "4"
            }, {
              ID: "5",
              measuredH: "24",
              measuredW: "24",
              properties: {
                borderStyle: "square",
                icon: {
                  ID: "ShareIcon",
                  size: "small"
                }
              },
              typeID: "Icon",
              x: "142",
              y: "19",
              zOrder: "5"
            }, {
              ID: "6",
              measuredH: "24",
              measuredW: "24",
              properties: {
                borderStyle: "square",
                icon: {
                  ID: "EllipsisIcon",
                  size: "small"
                }
              },
              typeID: "Icon",
              x: "174",
              y: "19",
              zOrder: "6"
            }, {
              ID: "7",
              h: "56",
              measuredH: "123",
              measuredW: "109",
              properties: {
                text: "This is a Symbol."
              },
              typeID: "StickyNote",
              w: "133",
              x: "149",
              y: "35",
              zOrder: "7"
            }]
          }
        },
        h: "91",
        measuredH: "91",
        measuredW: "282",
        properties: {
          controlName: "Sample Widget",
          notes: "This is a sample symbol for a reusable widget.\r\rIt is a set of controls that have been grouped and given a name."
        },
        typeID: "__group__",
        w: "282",
        x: "668",
        y: "372",
        zOrder: "2"
      }, {
        ID: "11",
        h: "80",
        measuredH: "80",
        measuredW: "200",
        properties: {
          text: "These notes don't appear as a Symbol because they are not in a group."
        },
        typeID: "HCurly",
        w: "412",
        x: "668",
        y: "268",
        zOrder: "3"
      }]
    },
    measuredH: "485",
    measuredW: "1080",
    mockupH: "465",
    mockupW: "1066",
    version: "1.0"
  }
};
exports.EmptySymbolLibrary = EmptySymbolLibrary;
const EmptySymbolLibraryNotes = "These are the notes for this Symbol Library. Click to edit them.\n\nUse them to add your name and email if you want to share on Wireframes To Go";
exports.EmptySymbolLibraryNotes = EmptySymbolLibraryNotes;
const FirstMockupOrder = 1000000;
exports.FirstMockupOrder = FirstMockupOrder;
const UnnamedMockup = "New Wireframe";
exports.UnnamedMockup = UnnamedMockup;
const UnnamedSymbolLibrary = "New Symbol Library";
exports.UnnamedSymbolLibrary = UnnamedSymbolLibrary;
const UnnamedSymbol = "New Symbol";
exports.UnnamedSymbol = UnnamedSymbol;
const UnnamedGroup = "Group";
exports.UnnamedGroup = UnnamedGroup;
const MaxMockupSize = 10000;
exports.MaxMockupSize = MaxMockupSize;
const MockupViewPadding = 400;
exports.MockupViewPadding = MockupViewPadding;
const Lorem1 = "lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.";
exports.Lorem1 = Lorem1;
const FontSizes = [8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 20, 24, 28, 32, 40, 60, 72];
exports.FontSizes = FontSizes;
const ProjectFontSizes = [8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18];
exports.ProjectFontSizes = ProjectFontSizes;
const FontFamilies = ["Balsamiq Sans", "Open Sans", "Roboto", "Lato", "PT Sans", "PT Serif", "Vollkorn", "Roboto Slab", "Caveat", "Roboto Mono", "Inconsolata"];
exports.FontFamilies = FontFamilies;
const MaxNoteLength = 10000;
exports.MaxNoteLength = MaxNoteLength;
var _default = {
  ControlChanges,
  ResourceTypes,
  UserRoles,
  ConnectedUserStatus,
  LicenseType,
  Master,
  EmptyMockup,
  EmptySymbolLibrary,
  EmptySymbolLibraryNotes,
  FirstMockupOrder,
  UnnamedMockup,
  UnnamedSymbolLibrary,
  UnnamedSymbol,
  UnnamedGroup,
  MaxMockupSize,
  MockupViewPadding,
  Lorem1,
  FontSizes,
  ProjectFontSizes,
  FontFamilies,
  MaxNoteLength
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/model/controlsdependencies.ts":
/*!***********************************************!*\
  !*** ./js/core/model/controlsdependencies.ts ***!
  \***********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ControlsDependencies = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _coreobject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var _lodash = _interopRequireDefault(__webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

class ControlsDependencies {
  constructor() {
    _defineProperty(this, "affectedByResourceChangeCache", {});

    _defineProperty(this, "controlsDependencies", {});
  }

  // stringyfied Control CoreObjectID -> { resourceIDs: Array of {resourceID, subResourceID}, coreID: CoreObjectID };
  getAffectedByResourceChange(resourceID, subResourceID) {
    const searchKeys = {
      resourceID: resourceID
    };

    if (subResourceID !== undefined) {
      _balsamiq.default.assertString(subResourceID);

      searchKeys.subResourceID = subResourceID;
    }

    const searchKeyString = JSON.stringify(searchKeys);
    const cachedValue = this.affectedByResourceChangeCache[searchKeyString];

    if (cachedValue) {
      return cachedValue;
    }

    const relatedControls = [];

    for (const key in this.controlsDependencies) {
      const controlsDepsForCoreObjectID = this.controlsDependencies[key];

      const resource = _lodash.default.find(controlsDepsForCoreObjectID.resourceIDs, searchKeys);

      if (resource !== undefined) {
        relatedControls.push(controlsDepsForCoreObjectID.coreID);
      }
    }

    const result = relatedControls.sort(function childrenBeforeParents(c1, c2) {
      return c2.controlID.length - c1.controlID.length;
    });
    this.affectedByResourceChangeCache[searchKeyString] = result;
    return result;
  } // returns a set of all dependencies
  // it may return null in case nothing changed - short circuit


  setDependency(coreObjectID, srcProperties) {
    let returnAllResources = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : true;

    _balsamiq.default.existsClass(coreObjectID, _coreobject.CoreObjectID);

    _balsamiq.default.assertArray(srcProperties); // resources that are deps of this control


    const resourceIDs = [];
    srcProperties.forEach(function (srcProp) {
      if (srcProp && srcProp.location === _types.default.SrcLocation.PROJECT_ASSETS) {
        resourceIDs.push({
          resourceID: srcProp.resourceID,
          subResourceID: srcProp.subResourceID
        });
      }
    });
    const coreObjectIDString = coreObjectID.toString(); // short circuit: if the control had no dependencies and still has none, then return

    if (resourceIDs.length === 0 && this.controlsDependencies[coreObjectIDString] == null) {
      return {
        changed: false
      };
    } // compute diffs: new urls, old urls by comparing new and old dependencies


    const oldResourceIDs = this.controlsDependencies[coreObjectIDString] ? this.controlsDependencies[coreObjectIDString].resourceIDs : [];

    const diff = _types.default.diffArray(oldResourceIDs, resourceIDs); // short circuit: if the dependencies are still the same, then return


    if (diff.removed.length === 0 && diff.added.length === 0) {
      return {
        changed: false
      };
    }

    diff.removed.concat(diff.added).forEach(resourceID => {
      delete this.affectedByResourceChangeCache[JSON.stringify(resourceID)];
    }); // remember the dependencies of this control, or remove from the list if they are empty

    if (resourceIDs.length > 0) {
      this.controlsDependencies[coreObjectIDString] = {
        resourceIDs: resourceIDs,
        coreID: coreObjectID
      };
    } else {
      delete this.controlsDependencies[coreObjectIDString];
    } // update all deps of this mockup


    if (!returnAllResources) {
      return {
        changed: true,
        resourceIDs: []
      };
    }

    const allResources = new Map();

    for (const coreObjectIDString in this.controlsDependencies) {
      this.controlsDependencies[coreObjectIDString].resourceIDs.forEach(resourceObject => {
        const resourceID = {
          resourceID: resourceObject.resourceID,
          subResourceID: resourceObject.subResourceID
        };
        const resourceIDStr = JSON.stringify(resourceID);
        allResources.set(resourceIDStr, resourceID);
      });
    } // set deps of this resource


    return {
      changed: true,
      resourceIDs: Array.from(allResources.values())
    };
  }

  hasControl(stringyfiedControlCoreObjectID) {
    return this.controlsDependencies[stringyfiedControlCoreObjectID] != null;
  }

  getControl(stringyfiedControlCoreObjectID) {
    _balsamiq.default.assert(this.hasControl(stringyfiedControlCoreObjectID));

    return this.controlsDependencies[stringyfiedControlCoreObjectID].coreID;
  }

  each(callback) {
    for (const key in this.controlsDependencies) {
      const obj = this.controlsDependencies[key];

      if (!callback(obj.coreID)) {
        break;
      }
    }
  }

  hasControlOfType(controlTypeID, getControlFunc) {
    for (const key in this.controlsDependencies) {
      const obj = this.controlsDependencies[key];
      const control = getControlFunc(obj.coreID.controlID);

      if (control && control.controlTypeID === controlTypeID) {
        return true;
      }
    }

    return false;
  }

  getDependenciesOfControl(coreObjectID) {
    _balsamiq.default.existsClass(coreObjectID, _coreobject.CoreObjectID);

    const coreObjectIDString = coreObjectID.toString();

    if (this.hasControl(coreObjectIDString)) {
      return this.controlsDependencies[coreObjectIDString].resourceIDs;
    } else {
      return null;
    }
  }

}

exports.ControlsDependencies = ControlsDependencies;

/***/ }),

/***/ "./js/core/model/thumbnail.ts":
/*!************************************!*\
  !*** ./js/core/model/thumbnail.ts ***!
  \************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ThumbnailSizeValues = exports.ThumbnailQualityValues = exports.Thumbnail = exports.EventTypes = void 0;
exports.getLowResThumbnailSize = getLowResThumbnailSize;
exports.getThumbnailJpgQualityIndex = getThumbnailJpgQualityIndex;
exports.getThumbnailJpgSize = getThumbnailJpgSize;
exports.getThumbnailJpgSizeIndex = getThumbnailJpgSizeIndex;
exports.setThumbnailJpgQualityIndex = setThumbnailJpgQualityIndex;
exports.setThumbnailJpgSizeIndex = setThumbnailJpgSizeIndex;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _coreobject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var _simplerenderer = __webpack_require__(/*! core/graphics/simplerenderer */ "./js/core/graphics/simplerenderer.js");

var _events = __webpack_require__(/*! core/events */ "./js/core/events.ts");

var _controls = __webpack_require__(/*! core/controls/controls */ "./js/core/controls/controls.js");

var _exectime = __webpack_require__(/*! core/utils/exectime */ "./js/core/utils/exectime.js");

var _assets = __webpack_require__(/*! editor/utils/assets */ "./js/editor/utils/assets.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

const MEASURE_TIME = false;
const HIGH_RES_THUMBNAIL_SIZE = 196;
const HIGH_RES_THUMBNAIL_DPI_SCALE = 2;
const LOW_RES_THUMBNAIL_SIZE = 320;
const LOW_RES_THUMBNAIL_DPI_SCALE = 1;
const EventTypes = {
  THUMBNAIL_CONTENTS_CHANGED: "thumbnailContentsChanged",
  THUMBNAIL_DIRTY: "thumbnailDirty"
};
exports.EventTypes = EventTypes;
let thumbnailJpgQualityIndex = 0;
let thumbnailJpgSizeIndex = 0;
const ThumbnailQualityValues = [0.60, 0.65, 0.70, 0.75, 0.80, 0.85, _assets.DEFAULT_JPG_QUALITY];
exports.ThumbnailQualityValues = ThumbnailQualityValues;
const ThumbnailSizeValues = [320, 360, 380, 400, 420, 440, 460, 480];
exports.ThumbnailSizeValues = ThumbnailSizeValues;

function setThumbnailJpgQualityIndex(qualityIndex) {
  _balsamiq.default.assert(qualityIndex >= 0 && qualityIndex < ThumbnailQualityValues.length);

  thumbnailJpgQualityIndex = qualityIndex;
}

function setThumbnailJpgSizeIndex(sizeIndex) {
  _balsamiq.default.assert(sizeIndex >= 0 && sizeIndex < ThumbnailSizeValues.length);

  thumbnailJpgSizeIndex = sizeIndex;
}

function getThumbnailJpgQualityIndex() {
  return thumbnailJpgQualityIndex;
}

function getThumbnailJpgSizeIndex() {
  return thumbnailJpgSizeIndex;
}

function getThumbnailJpgSize() {
  return ThumbnailSizeValues[getThumbnailJpgSizeIndex()];
}

class Thumbnail extends _coreobject.CoreObject {
  constructor(context, coreObjectID, resourceCoreObjectID) {
    let realSize = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : false;
    super(context, coreObjectID);

    _defineProperty(this, "simpleRenderer", void 0);

    _defineProperty(this, "nativeImageID", void 0);

    _defineProperty(this, "nativeLowResImageID", void 0);

    _defineProperty(this, "realSize", void 0);

    _defineProperty(this, "resourceCoreObjectID", void 0);

    _defineProperty(this, "execTime", void 0);

    _defineProperty(this, "execTimeID", void 0);

    (0, _coreobject.assertResourceType)(resourceCoreObjectID);
    this.simpleRenderer = new _simplerenderer.SimpleRenderer(this.context);
    this.nativeImageID = undefined;
    this.nativeLowResImageID = undefined;
    this.realSize = realSize;
    this.resourceCoreObjectID = resourceCoreObjectID;
    this.execTime = MEASURE_TIME ? new _exectime.ExecTime({
      host: this.context.host
    }) : undefined;
    this.execTimeID = undefined;
  }

  setResourceCoreObjectID(resourceCoreObjectID) {
    (0, _coreobject.assertResourceType)(resourceCoreObjectID);
    this.resourceCoreObjectID = resourceCoreObjectID.clone();
  }

  getResourceCoreObjectID() {
    return this.resourceCoreObjectID;
  }

  getImageID() {
    return this.nativeImageID;
  }

  getLowResImageID() {
    return this.nativeLowResImageID;
  }

  renderCommand(mockupRootControl, dpiScale) {
    if (this.realSize) {
      return this.simpleRenderer.renderFullSize(mockupRootControl, false, dpiScale);
    } else {
      return this.simpleRenderer.renderThumbnail(mockupRootControl, HIGH_RES_THUMBNAIL_SIZE, dpiScale);
    }
  }

  updateThumbnail(mockupRootControl, dispatchDirty, callback) {
    _balsamiq.default.existsClass(mockupRootControl, _controls.Control);

    _balsamiq.default.assertBool(dispatchDirty);

    if (callback) {
      _balsamiq.default.assertFunction(callback);
    }

    const finish = () => {
      if (this.execTimeID && this.execTime) {
        this.execTime.end(this.execTimeID);
        this.execTimeID = undefined;
      }

      callback && callback();
    };

    this.execTimeID = this.execTime ? this.execTime.start("updateThumbnail " + this.resourceCoreObjectID.resourceID) : undefined; // function that draws the commands in the thumbnail image

    const fun = () => {
      // CAVEAT: This function is invoked asynchronously, so we need to check if the thumbnail is still valid
      if (this._isCleanedUp) {
        finish();
        return;
      }

      this._measureTime("before renderCommand"); // get the commands to draw the thumbnail


      const commands = this.renderCommand(mockupRootControl, HIGH_RES_THUMBNAIL_DPI_SCALE);

      this._measureTime("after renderCommand, before drawCommands");

      this.context.host.Images.drawCommands(this.nativeImageID, commands, () => {
        // CAVEAT: check if thumbnail is still here. If it isn't, just skip this job
        if (this._isCleanedUp) {
          finish();
          return;
        }

        const fun2 = () => {
          if (this._isCleanedUp || !this.nativeImageID) {
            finish();
            return;
          }

          const highResImageSize = (0, _simplerenderer.getImageSizeScaledByDpiFromCommands)(commands);
          const lowResSize = getLowResThumbnailSize(highResImageSize);
          const lowResCommands = (0, _simplerenderer.renderImage)(this.context, this.nativeImageID, lowResSize, true);
          this.context.host.Images.drawCommands(this.nativeLowResImageID, lowResCommands, () => {
            this._measureTime("after drawCommands, before THUMBNAIL_CONTENTS_CHANGED");

            if (this._isCleanedUp) {
              finish();
              return;
            } // notify that the image has changed


            this.context.eventCenter.fireEvent((0, _events.newCoreEvent)(this.coreObjectID, EventTypes.THUMBNAIL_CONTENTS_CHANGED, {
              imageID: this.nativeImageID,
              resourceCoreObjectID: this.resourceCoreObjectID
            }));

            this._measureTime("after THUMBNAIL_CONTENTS_CHANGED"); // if requested, then notify that the thumbnail is ready to be saved


            if (dispatchDirty) {
              this._measureTime("before getJPGBase64Data"); // TODO put image data in the eventData field?


              this.context.host.Images.getJPGBase64Data(this.nativeLowResImageID, ThumbnailQualityValues[thumbnailJpgQualityIndex], res => {
                this._measureTime("after getJPGBase64Data");

                const attributes = {
                  branchID: this.resourceCoreObjectID.branchID,
                  resourceID: this.resourceCoreObjectID.resourceID,
                  image: res.data
                };

                if (this._isCleanedUp) {
                  finish();
                  return;
                }

                this._measureTime("before THUMBNAIL_DIRTY");

                this.context.eventCenter.fireEvent((0, _events.newCoreEvent)(this.coreObjectID, EventTypes.THUMBNAIL_DIRTY, {
                  imageID: this.nativeImageID,
                  attributes: attributes
                }));

                this._measureTime("after THUMBNAIL_DIRTY");

                finish();
              }); // notify the caller that the update is done
            } else {
              // notify the caller that the update is done
              finish();
            }
          });
        };

        if (!this.nativeLowResImageID) {
          this.context.host.Images.make(res => {
            this.nativeLowResImageID = res.id;
            fun2();
          });
        } else {
          this.context.host.asyncCallback(fun2);
        }
      });
    }; // function that creates the image if needed, then calls fun()


    if (!this.nativeImageID) {
      this.context.host.Images.make(res => {
        this.nativeImageID = res.id;
        fun();
      });
    } else {
      this.context.host.asyncCallback(fun);
    }
  }

  cleanUp() {
    super.cleanUp();

    if (this.nativeImageID) {
      this.context.host.Images.remove(this.nativeImageID);
    }

    this.nativeImageID = undefined;
  }

  _measureTime(message) {
    if (this.execTime && this.execTimeID) {
      this.execTime.lap(this.execTimeID, message);
    }
  }

}

exports.Thumbnail = Thumbnail;

function getLowResThumbnailSize(highResThumbnailSize) {
  const lowResMaxLength = getThumbnailJpgSize() * LOW_RES_THUMBNAIL_DPI_SCALE;
  const highToLowResRatio = getThumbnailJpgSize() * LOW_RES_THUMBNAIL_DPI_SCALE / (HIGH_RES_THUMBNAIL_SIZE * HIGH_RES_THUMBNAIL_DPI_SCALE);
  const maxLength = Math.max(highResThumbnailSize.width, highResThumbnailSize.height);
  let ratio = highToLowResRatio;

  if (maxLength * ratio > lowResMaxLength) {
    ratio = lowResMaxLength / maxLength;
  }

  return {
    width: highResThumbnailSize.width * ratio,
    height: highResThumbnailSize.height * ratio
  };
}

/***/ }),

/***/ "./js/core/operations/operation.ts":
/*!*****************************************!*\
  !*** ./js/core/operations/operation.ts ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = exports.Operation = exports.Direction = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var Direction;
exports.Direction = Direction;

(function (Direction) {
  Direction["Do"] = "do";
  Direction["Undo"] = "undo";
})(Direction || (exports.Direction = Direction = {}));

class Operation {
  constructor(params) {
    _defineProperty(this, "context", void 0);

    _defineProperty(this, "operationID", void 0);

    _defineProperty(this, "finishCallback", void 0);

    _defineProperty(this, "progressCallback", void 0);

    _defineProperty(this, "params", void 0);

    _defineProperty(this, "undoData", {});

    _defineProperty(this, "running", false);

    _defineProperty(this, "direction", void 0);

    _defineProperty(this, "isFirstRun", true);

    this.operationID = null;
    this.finishCallback = null;
    this.progressCallback = null;
    this.params = params;
    this.direction = null;
  }

  getOperationID() {
    return this.operationID;
  }

  run(context, operationID, finishCallback, progressCallback, direction) {
    _balsamiq.default.exists(context);

    _balsamiq.default.exists(context.getCoreObject);

    _balsamiq.default.exists(context.eventCenter);

    _balsamiq.default.exists(context.host);

    _balsamiq.default.assertString(operationID);

    _balsamiq.default.assertFunction(finishCallback);

    _balsamiq.default.assertFunction(progressCallback);

    _balsamiq.default.assertEnum(direction, Direction);

    if (direction == Direction.Undo) {
      _balsamiq.default.assert(this.canUndo(), "cannot undo an Operation that is not undoable. name: " + this.name());
    }

    _balsamiq.default.assert(!this.running, "cannot run an already running Operation");

    this.context = context;
    this.operationID = operationID;
    this.finishCallback = finishCallback;
    this.progressCallback = progressCallback;
    this.direction = direction;
    this.running = true;

    if (direction == Direction.Do) {
      this.runDo();
    } else {
      this.runUndo();
    }
  }

  executionName() {
    let p;

    if (this.direction == Direction.Do) {
      p = this.isFirstRun ? "" : "redo ";
    } else {
      p = "undo ";
    }

    return p + this.name();
  } // Subclasses methods
  // ------------------


  name() {
    // subclass
    return this.constructor.name;
  }

  canUndo() {
    // subclass
    // return true if the operation implements runUndo() method
    return false;
  }

  runDo() {// subclass
    // must call notifyFinished(..) when finished
  }

  runUndo() {// subclass
    // called only if canUndo() returns true
    // must call notifyFinished(..) when finished
  } // Notification methods
  // --------------------


  notifyProgress(currentStep, maxSteps) {
    if (this.progressCallback) {
      this.progressCallback(currentStep, maxSteps);
    }
  }

  notifyFinished() {
    let res = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

    _balsamiq.default.assertObject(res);

    if (this.running) {
      this.running = false;
      this.isFirstRun = false;

      if (typeof res.success !== "boolean") {
        res.success = true;
      }

      this.finishCallback(res);
    }
  }

}

exports.Operation = Operation;
var _default = {
  Operation,
  Direction
};
exports["default"] = _default;

/***/ }),

/***/ "./js/core/symbol-map.ts":
/*!*******************************!*\
  !*** ./js/core/symbol-map.ts ***!
  \*******************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.SymbolMap = void 0;

var _constants = __webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts");

var _controlsdependencies = __webpack_require__(/*! core/model/controlsdependencies */ "./js/core/model/controlsdependencies.ts");

var _tarjanScc = __webpack_require__(/*! core/utils/tarjan-scc */ "./js/core/utils/tarjan-scc.ts");

var _coreobject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function getSymbolIDString(c) {
  return `${c.resourceID}:${c.subResourceID ? c.subResourceID : "-1"}`;
}

class SymbolMap {
  constructor(getCoreObject) {
    _defineProperty(this, "controlsDependencies", new _controlsdependencies.ControlsDependencies());

    _defineProperty(this, "componentsWithLoops", new Set());

    _defineProperty(this, "getCoreObject", void 0);

    _defineProperty(this, "setControlDependencies", (resource, p_coreObjectID, p_srcProperties) => {
      const symbolSrcProperties = p_srcProperties.filter(src => src.subResourceID != null);
      const {
        changed
      } = this.controlsDependencies.setDependency(p_coreObjectID, symbolSrcProperties, false);

      if (changed && resource.getType() === _constants.ResourceTypes.SYMBOLLIBRARY) {
        this.updateComponentsWithLoops();
      }
    });

    _defineProperty(this, "hasSymbolDefinitionALoop", src => {
      return this.componentsWithLoops.has(getSymbolIDString(src));
    });

    _defineProperty(this, "getSymbolsDependingOnSymbol", symbolCoreObjectID => {
      _balsamiq.default.existsClass(symbolCoreObjectID, _coreobject.CoreObjectID);

      _balsamiq.default.assert(symbolCoreObjectID.controlID != null, "A control is expected here");

      const dependantControls = this.controlsDependencies.getAffectedByResourceChange(symbolCoreObjectID.resourceID, symbolCoreObjectID.controlID.toString()).map(cID => this.getCoreObject(cID));
      const dependantSymbolControlsInsideSLs = dependantControls.filter(c => c && c.isSymbol() && c.getMockup().getType() === _constants.ResourceTypes.SYMBOLLIBRARY);
      const dependantSymbolDefinitions = dependantSymbolControlsInsideSLs.map(control => {
        const coreObjectID = control.coreObjectID.clone();
        coreObjectID.controlID = coreObjectID.controlID.slice(0, 1);
        return coreObjectID;
      });
      return dependantSymbolDefinitions;
    });

    this.getCoreObject = getCoreObject;
  }

  /**
   * Create a list of nodes with their connections from controlsDependencies
   */
  updateComponentsWithLoops() {
    const nodesMap = new Map(); // create a node for every control and it's dependencies

    for (const coreObjectIDStr of Object.keys(this.controlsDependencies.controlsDependencies)) {
      const resource = this.controlsDependencies.controlsDependencies[coreObjectIDStr];
      const controlID = getSymbolIDString({
        resourceID: resource.coreID.resourceID,
        subResourceID: resource.coreID.controlID.slice(0, 1).toString()
      });

      if (!nodesMap.has(controlID)) {
        const node = new _tarjanScc.Node(controlID);
        nodesMap.set(controlID, node);
      }

      resource.resourceIDs.forEach(s => {
        const symbolIDStr = getSymbolIDString(s);

        if (!nodesMap.has(symbolIDStr)) {
          const node = new _tarjanScc.Node(symbolIDStr);
          nodesMap.set(symbolIDStr, node);
        }
      });
    } // create an array of unique nodes and add its dependencies as node connections


    const nodes = [];

    for (const coreObjectIDStr of Object.keys(this.controlsDependencies.controlsDependencies)) {
      const controlDependencies = this.controlsDependencies.controlsDependencies[coreObjectIDStr];
      const symbolIDStr = getSymbolIDString({
        resourceID: controlDependencies.coreID.resourceID,
        subResourceID: controlDependencies.coreID.controlID.slice(0, 1).toString()
      });
      const node = nodesMap.get(symbolIDStr);
      controlDependencies.resourceIDs.forEach(s => {
        const symbolIDStr = getSymbolIDString(s);

        if (!node.connections.find(c => c.name === symbolIDStr)) {
          const connection = nodesMap.get(symbolIDStr);
          node.connections.push(connection);
        }
      });

      if (!nodes.find(c => c.name === symbolIDStr)) {
        nodes.push(node);
      }
    }

    const graph = new _tarjanScc.Graph(nodes);
    const tarjan = new _tarjanScc.Tarjan(graph);
    const scc = tarjan.run();
    let componentsWithLoops = new Set();
    scc.forEach(component => {
      if (component.length > 1) {
        componentsWithLoops = new Set([...componentsWithLoops, ...component.map(c => c.name)]);
      }
    });
    this.componentsWithLoops = componentsWithLoops;
  }

}

exports.SymbolMap = SymbolMap;

/***/ }),

/***/ "./js/core/tasks/initnativemoduletask.ts":
/*!***********************************************!*\
  !*** ./js/core/tasks/initnativemoduletask.ts ***!
  \***********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.InitNativeModuleTask = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _task = __webpack_require__(/*! core/tasks/task */ "./js/core/tasks/task.ts");

var _native = __webpack_require__(/*! core/native */ "./js/core/native.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class InitNativeModuleTask extends _task.TaskBase {
  name() {
    return 'Init Module';
  }

  doRun() {
    _balsamiq.default.assertEnum(this.params.moduleName, _native.ModuleNames);

    if (this.params.disableProgress) {
      this.enableProgress(false);
    }

    let module;

    switch (this.params.moduleName) {
      case _native.ModuleNames.SQLite:
        module = this.context.host.SQLite;
        break;

      case _native.ModuleNames.PDF:
        module = this.context.host.PDF;
        break;

      default:
        _balsamiq.default.assert(false, "InitModuleTask unknown module name to init: " + this.params.moduleName);

    }

    const moduleStatus = module.status();

    switch (moduleStatus) {
      case _native.ModuleStatus.NotInitialized:
        {
          // init it
          const title = this.params.title || "Initializing " + this.params.moduleName + "…";
          this.startProgress(title);
          module.init(res => {
            this.notifyFinished(res);
          }, (currentStep, totalSteps) => {
            this.setProgress(title, currentStep, totalSteps);
          });
        }
        break;

      case _native.ModuleStatus.Initializing:
        // what to do??
        _balsamiq.default.assert(false, "InitModuleTask can not init module: " + this.params.moduleName + " because already initializing");

        break;

      case _native.ModuleStatus.Error:
        this.context.host.asyncCallback(() => {
          this.notifyFinished({
            error: true
          });
        });
        break;

      case _native.ModuleStatus.Initialized:
        this.context.host.asyncCallback(() => {
          this.notifyFinished({});
        });
        break;

      default:
        _balsamiq.default.assert(false, "InitModuleTask unknown module status: " + moduleStatus);

    }
  }

}

exports.InitNativeModuleTask = InitNativeModuleTask;

/***/ }),

/***/ "./js/core/tasks/task.ts":
/*!*******************************!*\
  !*** ./js/core/tasks/task.ts ***!
  \*******************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.TaskBase = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _operationprogressviewcontroller = _interopRequireDefault(__webpack_require__(/*! core/viewcontrollers/operationprogressviewcontroller */ "./js/core/viewcontrollers/operationprogressviewcontroller.js"));

var _genericoperation = __webpack_require__(/*! core/operations/genericoperation */ "./js/core/operations/genericoperation.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

const SHOW_PROGRESS_DELAY = 300;

class TaskBase {
  // ViewController<State, NoOptions>; OperationProgressVC
  // GenericOperation
  constructor(params) {
    _defineProperty(this, "params", void 0);

    _defineProperty(this, "context", void 0);

    _defineProperty(this, "response", void 0);

    _defineProperty(this, "taskID", void 0);

    _defineProperty(this, "vcID", void 0);

    _defineProperty(this, "running", false);

    _defineProperty(this, "canceled", false);

    _defineProperty(this, "finished", false);

    _defineProperty(this, "interrupted", false);

    _defineProperty(this, "interruptedCallback", void 0);

    _defineProperty(this, "progressEnabled", true);

    _defineProperty(this, "progressVC", void 0);

    _defineProperty(this, "progressStartTime", void 0);

    _defineProperty(this, "progressTimer", void 0);

    _defineProperty(this, "progressTitle", void 0);

    _defineProperty(this, "progressCurrentStep", 0);

    _defineProperty(this, "progressTotalSteps", 0);

    _defineProperty(this, "cancelButtonLabel", void 0);

    _defineProperty(this, "timerID", void 0);

    _defineProperty(this, "pendingOperation", void 0);

    _defineProperty(this, "_onTimeout", () => {
      this.timerID = undefined;
      this.onTimeout();
    });

    this.params = params;
  }

  cleanUp() {// subclass
  }

  getVCID() {
    return this.vcID || "";
  }

  getTaskID() {
    return this.taskID || "";
  }

  run(context, taskID) {
    _balsamiq.default.assert(!this.running, "could not run() an already running Task");

    _balsamiq.default.assert(!this.canceled, "could not run() a canceled Task");

    _balsamiq.default.assert(!this.finished, "could not run() a finished Task");

    _balsamiq.default.assert(!this.interrupted, "could not run() an interrupted Task");

    _balsamiq.default.exists(context);

    _balsamiq.default.assertFunction(context.runSubtask);

    _balsamiq.default.assertFunction(context.onTaskFinished);

    _balsamiq.default.assertFunction(context.hasSubtasks);

    _balsamiq.default.assertFunction(context.canInterruptSubtasks);

    _balsamiq.default.assertFunction(context.interruptSubtasks);

    _balsamiq.default.exists(context.windowID);

    _balsamiq.default.exists(context.operationCenter);

    _balsamiq.default.exists(context.eventCenter);

    this.context = context;

    _balsamiq.default.exists(taskID);

    this.taskID = taskID;
    this.running = true;
    this.doRun();
  }

  cancel() {
    _balsamiq.default.assert(this.running, "could not cancel() a not running Task");

    _balsamiq.default.assert(!this.canceled, "could not cancel() an already canceled Task");

    _balsamiq.default.assert(!this.finished, "could not cancel() a finished Task");

    _balsamiq.default.assert(!this.interrupted, "could not cancel() an interrupted Task");

    if (this.isCancelable()) {
      this.canceled = true;
      this.onCanceled();
    }
  }

  interrupt(callback) {
    _balsamiq.default.assert(this.running, "could not interrupt() a not running Task");

    _balsamiq.default.assert(!this.canceled, "could not interrupt() a canceled Task");

    _balsamiq.default.assert(!this.finished, "could not interrupt() a finished Task");

    _balsamiq.default.assert(!this.interrupted, "could not interrupt() an already interrupted Task");

    if (this.canInterrupt()) {
      this.interrupted = true;
      this.interruptedCallback = callback;
      this.onInterrupted();
    }
  }

  startOperation(callback, operationName) {
    this.pendingOperation = new _genericoperation.GenericOperation({
      callback: callback,
      operationName: operationName
    });
    this.scheduleOperation(this.pendingOperation, () => {
      this.pendingOperation = null;
    });
  }

  finishOperation() {
    _balsamiq.default.exists(this.pendingOperation);

    this.pendingOperation.notifyFinished();
  }

  notifyFinished(res) {
    _balsamiq.default.assert(this.running, "could not notifyFinished() on a not running Task");

    _balsamiq.default.assert(!this.finished, "could not notifyFinished() on an already finished Task");

    this.response = res || {};

    if (this.canceled) {
      this.response.canceled = true;
    }

    this.endProgress();

    if (this.pendingOperation) {
      this.pendingOperation.notifyFinished();
    }

    this.finished = true;
    this.running = false;
    this.context.onTaskFinished(this.taskID, this.response);

    if (this.interrupted && this.interruptedCallback) {
      this.interruptedCallback();
    }

    if (this.timerID) {
      this.context.host.clearTimer(this.timerID);
    }

    this.cleanUp();
  } // Progress
  // --------


  enableProgress(value) {
    _balsamiq.default.assertBool(value);

    if (value) {
      this.progressEnabled = this.context.viewControllers.isViewControllerTypeRegistered(_operationprogressviewcontroller.default.Type);
    } else {
      this.progressEnabled = false;
    }
  }

  setCancelButtonLabel(value) {
    _balsamiq.default.assertString(value);

    this.cancelButtonLabel = value;
  }

  startProgress(title) {
    if (this.progressVCAvailable() && this.progressEnabled) {
      _balsamiq.default.assert(this.progressStartTime == null, "Task progress start time must be null when starting a new progress");

      _balsamiq.default.assert(this.progressVC == null, "Task progress VC must be null when starting a new progress");

      _balsamiq.default.assert(this.progressTimer == null, "Task progress timer must be null when starting a new progress");

      this.progressStartTime = Date.now();
      this.progressTimer = this.context.host.timer(SHOW_PROGRESS_DELAY, this.onProgressDelay.bind(this));
      this.progressTitle = title ? title : this.name();
      this.progressCurrentStep = 0;
      this.progressTotalSteps = 0;
    }
  }

  setProgress(title, currentStep, totalSteps) {
    if (this.progressVCAvailable() && this.progressEnabled) {
      _balsamiq.default.assert(this.progressStartTime, "Task progress was not started");

      this.progressTitle = title;
      this.progressCurrentStep = currentStep;
      this.progressTotalSteps = totalSteps;

      if (this.progressVC) {
        this.progressVC.setProgress(this.progressTitle, this.progressCurrentStep, this.progressTotalSteps);
      }
    }
  }

  endProgress() {
    if (this.progressVCAvailable() && this.progressEnabled) {
      if (!this.progressStartTime) {
        return;
      }

      this.progressStartTime = undefined;
      this.progressTimer = undefined;

      if (this.progressVC) {
        this.context.viewControllers.pop(this.context.windowID, this.progressVC.getNativeID());
        this.progressVC = undefined;
      }
    }
  }

  onProgressDelay() {
    if (this.progressVCAvailable() && this.progressEnabled) {
      if (!this.finished && this.progressStartTime && !this.progressVC) {
        this.context.viewControllers.make(_operationprogressviewcontroller.default.Type, this.taskID, {
          cancelButtonLabel: this.cancelButtonLabel
        }, (vc
        /* OperationProgressVC */
        ) => {
          if (!this.progressStartTime) {
            this.context.viewControllers.remove(vc.getNativeID());
          } else {
            this.progressVC = vc;
            this.progressVC.setProgress(this.progressTitle, this.progressCurrentStep, this.progressTotalSteps);

            if (this.isCancelable()) {
              this.progressVC.setCancelCallback(this.cancel.bind(this));
            }

            this.context.viewControllers.push(this.context.windowID, this.progressVC.getNativeID(), null, true);
          }
        });
      }
    }
  }

  progressVCAvailable() {
    return this.context && this.context.viewControllers.isViewControllerTypeRegistered(_operationprogressviewcontroller.default.Type);
  } // Operations
  // ----------


  scheduleOperation(operation, finishCallback) {
    _balsamiq.default.exists(operation);

    _balsamiq.default.assertFunction(finishCallback);

    this.startProgress(operation.name() || "");
    this.context.operationCenter.doOperation(operation, res => {
      if (this.progressStartTime) {
        this.endProgress();
      }

      finishCallback(res);
    }, (currentStep, totalSteps) => {
      this.setProgress(operation.name() || "", currentStep, totalSteps);
    });
    return operation;
  } // Subtasks
  // --------


  runSubtask(subtask, callback) {
    this.context.runSubtask(this.taskID, subtask, callback);
  }

  runSubtaskIfPossible(subtask, callback) {
    return this.context.runSubtaskIfPossible(this.taskID, subtask, callback);
  }

  hasSubtasks() {
    return this.context.hasSubtasks(this.taskID);
  }

  canInterruptSubtasks() {
    return this.context.canInterruptSubtasks(this.taskID);
  }

  interruptSubtasks(callback) {
    this.context.interruptSubtasks(this.taskID, callback);
  } // Subclasses
  // ----------


  name() {
    // subclass
    return "Unnamed Task";
  }

  doRun() {// subclass
  }

  isCancelable() {
    // subclass
    return false;
  }

  onCanceled() {// subclass
  }

  canInterrupt() {
    // subclass
    return false;
  }

  onInterrupted() {// subclass
  }

  subtaskContext() {
    // subclass
    // the returned object will be merged with the default context and passed to subtasks
    return {};
  }

  setTimer(ms) {
    if (this.timerID) {
      this.context.host.clearTimer(this.timerID);
    }

    this.timerID = this.context.host.timer(ms, this._onTimeout);
  }

  onTimeout() {// invoked when timer expires
    // subclass
  }

  action(action, params) {
    // subclass
    return false;
  }

  mouseEvent(event) {
    // subclass
    return false;
  }

  keyboardEvent(event, hasFocus) {
    // subclass
    return false;
  }

}

exports.TaskBase = TaskBase;

/***/ }),

/***/ "./js/core/textmeter.ts":
/*!******************************!*\
  !*** ./js/core/textmeter.ts ***!
  \******************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.TextMeter = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _types = _interopRequireDefault(__webpack_require__(/*! core/types */ "./js/core/types.js"));

var _lodash = __webpack_require__(/*! lodash */ "./node_modules/lodash/lodash.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

class TextMeter {
  constructor(p_helper) {
    _defineProperty(this, "maxFontSize", void 0);

    _defineProperty(this, "previousWidth", void 0);

    _defineProperty(this, "currentText", void 0);

    _defineProperty(this, "textStyles", void 0);

    _defineProperty(this, "helper", void 0);

    _balsamiq.default.exists(p_helper); // Balsamiq.assert(Balsamiq.conformsTo(p_helper, TextMeterHelper.prototype));


    this.maxFontSize = 1;
    this.previousWidth = 0;
    this.currentText = "";
    this.textStyles = [];
    this.helper = p_helper;
  }

  initializeMeasurement(p_textStyle) {
    this.helper.setTextStyle(p_textStyle.fontFamily, p_textStyle.fontSize, p_textStyle.bold, p_textStyle.italic);
    this.textStyles = [(0, _lodash.clone)(p_textStyle)];
    this.maxFontSize = 1;
    this.previousWidth = 0;
    this.currentText = "";
  }

  addText(text) {
    const currentTextStyle = this.textStyles[this.textStyles.length - 1];
    this.maxFontSize = Math.max(this.maxFontSize, currentTextStyle.fontSize);
    this.currentText += text;
  }

  openFormat(format) {
    // TODO save previous width only when the format is one in the list below
    this.previousWidth += this.helper.measureStringWidth(this.currentText);
    this.currentText = "";
    const newTextStyle = (0, _lodash.clone)(this.textStyles[this.textStyles.length - 1]);

    if (format.type === "bold") {
      newTextStyle.bold = true;
    } else if (format.type === "italic") {
      newTextStyle.italic = true;
    } else if (format.type === "size") {
      newTextStyle.fontSize = format.size;
      this.maxFontSize = Math.max(this.maxFontSize, format.size);
    }

    this.helper.setTextStyle(newTextStyle.fontFamily, newTextStyle.fontSize, newTextStyle.bold, newTextStyle.italic);
    this.textStyles.push(newTextStyle);
  }

  closeFormat() {
    this.previousWidth += this.helper.measureStringWidth(this.currentText);
    this.currentText = "";
    this.textStyles.pop();
    const oldTextStyle = this.textStyles[this.textStyles.length - 1];
    this.helper.setTextStyle(oldTextStyle.fontFamily, oldTextStyle.fontSize, oldTextStyle.bold, oldTextStyle.italic);
  }

  getSize() {
    const currentWidth = this.helper.measureStringWidth(this.currentText);
    const totalWidth = Math.ceil(this.previousWidth + currentWidth);
    const height = this.currentText === "" ? this.textStyles[this.textStyles.length - 1].fontSize : this.maxFontSize;
    return _types.default.newTextSize(Math.max(0, totalWidth), height, height / 1.27);
  }

  terminateMeasurement() {
    this.textStyles = [];
    this.previousWidth = 0;
    this.currentText = "";
  }

}

exports.TextMeter = TextMeter;

/***/ }),

/***/ "./js/core/urls.ts":
/*!*************************!*\
  !*** ./js/core/urls.ts ***!
  \*************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.decomposeURL = decomposeURL;
exports.deserializeParamString = deserializeParamString;
exports.getFileExt = getFileExt;
exports.getFileNameAndExt = getFileNameAndExt;
exports.getFileNameNoExt = getFileNameNoExt;
exports.isImageURL = isImageURL;
exports.serializeParamObj = serializeParamObj;
exports.stripAnchor = stripAnchor;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function decomposeURL(p_url) {
  _balsamiq.default.assertString(p_url);

  const res = {
    protocol: "",
    parentURL: "",
    name: "",
    extension: "",
    nameNoExtension: "",
    anchor: "",
    params: ""
  };

  if (p_url.charAt(p_url.length - 1) === '/') {
    p_url = p_url.substr(0, p_url.length - 2);
  }

  res.protocol = p_url.substr(0, p_url.indexOf(':'));
  const lastSlashIndex = p_url.lastIndexOf('/');
  res.parentURL = p_url.substr(0, lastSlashIndex); // special case: fixing file:///c: without the final /
  // with check for length and other parts too because ":" allowed in the fragment part
  // https://stackoverflow.com/questions/2053132/is-a-colon-safe-for-friendly-url-use

  if (res.parentURL.length == 10 && res.parentURL.startsWith("file:/") && res.parentURL.endsWith(":")) {
    res.parentURL += "/";
  }

  let nameExtAnchorParams;

  try {
    nameExtAnchorParams = decodeURIComponent(p_url.substr(lastSlashIndex + 1));
  } catch (e) {
    if (e instanceof URIError) {
      nameExtAnchorParams = p_url.substr(lastSlashIndex + 1);
    } else {
      throw e;
    }
  }

  const paramIndex = nameExtAnchorParams.indexOf('?');

  if (paramIndex >= 0) {
    res.params = nameExtAnchorParams.substr(paramIndex + 1);
    nameExtAnchorParams = nameExtAnchorParams.substr(0, paramIndex);
  }

  const anchorIndex = nameExtAnchorParams.indexOf('#');

  if (anchorIndex >= 0) {
    res.anchor = nameExtAnchorParams.substr(anchorIndex + 1);
    nameExtAnchorParams = nameExtAnchorParams.substr(0, anchorIndex);
  }

  res.name = nameExtAnchorParams;
  const lastDotIndex = nameExtAnchorParams.lastIndexOf('.');

  if (lastDotIndex > 0) {
    res.extension = nameExtAnchorParams.substr(lastDotIndex + 1);
    nameExtAnchorParams = nameExtAnchorParams.substr(0, lastDotIndex);
  }

  res.nameNoExtension = nameExtAnchorParams;
  return res;
}

function stripAnchor(p_url) {
  const res = {
    url: p_url,
    anchor: ""
  };
  const anchorIndex = p_url.indexOf('#');

  if (anchorIndex >= 0) {
    res.url = p_url.substr(0, anchorIndex);
    res.anchor = p_url.substr(anchorIndex + 1);
  }

  return res;
}

function isImageURL(p_url) {
  _balsamiq.default.assertString(p_url);

  const extension = decomposeURL(p_url).extension.toLowerCase();
  return ['gif', 'png', 'jpeg', 'jpg'].indexOf(extension) !== -1;
}

function serializeParamObj(paramObj) {
  const str = [];
  let p;

  for (p in paramObj) {
    if (Object.prototype.hasOwnProperty.call(paramObj, p)) {
      str.push(encodeURIComponent(p) + "=" + encodeURIComponent(paramObj[p]));
    }
  }

  return str.join("&");
}

function deserializeParamString(paramStr) {
  const origPairs = paramStr[0] === '?' ? paramStr.slice(1) : paramStr.slice();
  const pairs = origPairs.split('&');
  const result = {};
  pairs.forEach(function (p) {
    if (p) {
      const pair = p.split('=');
      result[decodeURIComponent(pair[0] || '')] = decodeURIComponent(pair[1] || '');
    }
  });
  return result;
}

function getFileExt(fileName) {
  _balsamiq.default.assertString(fileName);

  return fileName.slice((Math.max(0, fileName.lastIndexOf(".")) || Infinity) + 1);
}

function getFileNameNoExt(fileName) {
  _balsamiq.default.assertString(fileName);

  return fileName.slice(0, Math.max(0, fileName.lastIndexOf(".")) || Infinity);
}

function getFileNameAndExt(file) {
  _balsamiq.default.assertObject(file);

  let ext, nameNoExt, originalName;

  if (file.url) {
    const decomposed = decomposeURL(file.url);
    ext = decomposed.extension;
    nameNoExt = decomposed.nameNoExtension;
    originalName = file.url;
  } else if (file.filename) {
    ext = getFileExt(file.filename);
    nameNoExt = getFileNameNoExt(file.filename);
    originalName = file.filename;
  } else {
    _balsamiq.default.assert(false, "File obj must have a filename or url key.");
  }

  return {
    extension: ext || "",
    nameNoExtension: nameNoExt || "",
    originalName: originalName || ""
  };
}

/***/ }),

/***/ "./js/core/utils/debounce.ts":
/*!***********************************!*\
  !*** ./js/core/utils/debounce.ts ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.debounceAsync = exports.debounce = void 0;

var Native = _interopRequireWildcard(__webpack_require__(/*! core/native */ "./js/core/native.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const debounce = function (func, delay) {
  let timerID;
  let args;

  const f = function () {
    for (var _len = arguments.length, argv = new Array(_len), _key = 0; _key < _len; _key++) {
      argv[_key] = arguments[_key];
    }

    args = argv;
    timerID && Native.Host.clearTimer(timerID);
    timerID = Native.Host.timer(delay, function () {
      timerID = null;
      func(...args);
    });
  };

  f.flush = function () {
    if (timerID) {
      Native.Host.clearTimer(timerID);
      timerID = null;
      func(...args);
    }
  };

  return f;
};

exports.debounce = debounce;

const debounceAsync = function (f, delay) {
  let timerID;
  let args;

  const debounced = function () {
    for (var _len2 = arguments.length, argv = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
      argv[_key2] = arguments[_key2];
    }

    return new Promise((resolve, reject) => {
      args = argv;
      timerID && Native.Host.clearTimer(timerID);
      timerID = Native.Host.timer(delay, function () {
        timerID = null;
        resolve(f(...args));
      });
    });
  };

  return debounced;
};

exports.debounceAsync = debounceAsync;

/***/ }),

/***/ "./js/core/utils/progress.ts":
/*!***********************************!*\
  !*** ./js/core/utils/progress.ts ***!
  \***********************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function mapToString(map) {
  const arr = [];

  for (const [phaseName, phaseWeight] of map.entries()) {
    arr.push(`[${phaseName}, ${phaseWeight}]`);
  }

  return arr.join(", ");
}

class Progress {
  constructor(phases, setProgress) {
    _defineProperty(this, "phases", void 0);

    _defineProperty(this, "setProgress", void 0);

    this.phases = phases;
    this.setProgress = setProgress;
  }

  previousPhasesCompoundProgress(currentPhase) {
    let previousPhasesProgress = 0;

    for (const [phaseName, phaseWeight] of this.phases.entries()) {
      if (phaseName !== currentPhase) {
        previousPhasesProgress += phaseWeight;
      } else {
        break;
      }
    }

    return previousPhasesProgress;
  }

  step(phase, title, step, total) {
    const previousPhaseCompoundProgress = this.previousPhasesCompoundProgress(phase);
    const phaseWeight = this.phases.get(phase);

    if (phaseWeight) {
      this.setProgress(title, step / total * phaseWeight + previousPhaseCompoundProgress, 1);
    } else {
      throw new Error(`phase "${phase}" is not a phase. Phases are "${mapToString(this.phases)}"`);
    }
  }

}

exports["default"] = Progress;

/***/ }),

/***/ "./js/core/utils/reorderAddedResources.ts":
/*!************************************************!*\
  !*** ./js/core/utils/reorderAddedResources.ts ***!
  \************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.reorderAddedResources = exports.rebuildHierarchyLevels = exports.OperationType = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _resourcetree = __webpack_require__(/*! core/utils/resourcetree */ "./js/core/utils/resourcetree.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var OperationType;
exports.OperationType = OperationType;

(function (OperationType) {
  OperationType["DUPLICATE"] = "duplicate";
  OperationType["PASTE"] = "paste";
})(OperationType || (exports.OperationType = OperationType = {}));

const addToParent = (output, parentID, items, afterResourceID) => {
  if (afterResourceID) {
    const afterResource = output.getResourceByID(afterResourceID); // must calculate an order half-way between afterResource and the next sibling if one

    let allSiblings = [];

    if (afterResource.parent) {
      allSiblings = afterResource.parent.children;
    } else {
      allSiblings = output.getMainLevelResources();
    }

    if (allSiblings.length == 1) {
      // easy, add a bit to current order
      output.appendResourcesToParent(items, parentID);
    } else {
      const afterResourceIndex = allSiblings.findIndex(element => element.resourceID === afterResourceID);

      if (afterResourceIndex === allSiblings.length - 1) {
        output.appendResourcesToParent(items, parentID);
      } else {
        const order_base = afterResource.order;
        const order_increment = (allSiblings[afterResourceIndex + 1].order - afterResource.order) / (items.length + 1);
        const orders = [];
        const parents = [];
        items.forEach((item, index) => {
          orders.push(order_base + order_increment * (index + 1));
          parents.push(parentID);
        });
        output.addResourcesWithParentID(items, parents, orders);
      }
    }
  } else {
    // push it as a first child
    const parentItem = output.getResourceByID(parentID);

    if (parentItem.children.length == 0) {
      output.appendResourcesToParent(items, parentID);
    } else {
      const order_base = 0;
      const order_increment = parentItem.children[0].order / (items.length + 1);
      const orders = [];
      const parents = [];
      items.forEach((item, index) => {
        orders.push(order_base + order_increment * (index + 1));
        parents.push(parentID);
      });
      output.addResourcesWithParentID(items, parents, orders);
    }
  }
};

const placeOneElement = _ref => {
  let {
    currentResourceTree,
    targetResourceID,
    parentID,
    afterResourceID
  } = _ref;
  const output = currentResourceTree.clone(); // remove target resources if existing since they will be added at the right place

  if (output.getResourceByID(targetResourceID)) {
    output.deleteResourceIDs([targetResourceID]);
  }

  addToParent(output, parentID, [targetResourceID], afterResourceID);
  return output;
};

const rebuildHierarchyLevels = function (levels) {
  let offset = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
  const lowest_level = Math.min(...levels);
  const new_levels = [];
  let previous_level = -1;
  let previous_adjusted_level = -1;
  levels.forEach(level => {
    if (previous_level === -1) {
      if (level > lowest_level) {
        new_levels.push(lowest_level + offset);
        previous_level = level;
        previous_adjusted_level = lowest_level;
      } else {
        new_levels.push(level + offset);
        previous_level = level;
        previous_adjusted_level = level;
      }
    } else {
      if (level > previous_level) {
        new_levels.push(previous_adjusted_level + 1 + offset);
        previous_level = level;
        previous_adjusted_level = previous_adjusted_level + 1;
      } else if (level === previous_level) {
        new_levels.push(previous_adjusted_level + offset);
        previous_level = level;
      } else {
        new_levels.push(lowest_level + offset);
        previous_level = level;
        previous_adjusted_level = lowest_level;
      }
    }
  });
  return new_levels;
};

exports.rebuildHierarchyLevels = rebuildHierarchyLevels;

const rebuildHierarchy = (input, levels, parentID, afterResourceID, description) => {
  const parent_item = input.getResourceByID(parentID); // I'm adding an offset to levels to reveal the REAL level
  // the resource will have so that we can limit it
  // in fact, nesting level cannot be more than MAX_RESOURCE_LEVELS

  const new_levels = rebuildHierarchyLevels(levels, parent_item ? parent_item.getLevel() + 1 : 0);
  const output = input.clone(); // remove target resources if existing since they will be added at the right place

  description.targetResourceIDs.forEach((resourceID, index) => {
    if (output.getResourceByID(resourceID)) {
      output.deleteResourceIDs([resourceID]);
    }
  });
  let resource_at_current_level = "";
  let prev_level = -1;
  const last_resource_at_level = {};
  description.targetResourceIDs.forEach((resourceID, index) => {
    // remove target resources if existing since they will be added at the right place
    if (output.getResourceByID(resourceID)) {
      output.deleteResourceIDs([resourceID]);
    }

    let resource_level = new_levels[index];

    if (resource_level >= _resourcetree.MAX_RESOURCE_LEVELS - 1) {
      resource_level = _resourcetree.MAX_RESOURCE_LEVELS - 1;
    }

    if (index === 0) {
      addToParent(output, parentID, [resourceID], afterResourceID);
      last_resource_at_level[resource_level] = resourceID;
      prev_level = resource_level;
      resource_at_current_level = resourceID;
    } else {
      if (resource_level > prev_level) {
        last_resource_at_level[resource_level] = resourceID;
        output.appendResourcesToParent([resourceID], resource_at_current_level);
      } else {
        const latest_resource_id = last_resource_at_level[resource_level];
        const latest_resource = output.getResourceByID(latest_resource_id);
        addToParent(output, latest_resource.parentID, [resourceID], latest_resource_id);
      }

      prev_level = resource_level;
      resource_at_current_level = resourceID;
    }
  });
  return output;
};

const getPastePosition = (tree, selectedResourceIDs) => {
  // state open/closed of the farthest ancestor of the last resourceID belonging to the selection
  // closed -> return it
  // open ->
  // 1 find the farthest closed
  const currentResourceID = selectedResourceIDs[selectedResourceIDs.length - 1];
  let currentResourceTreeItem = tree.getResourceByID(currentResourceID);
  let closedResourceID = null;

  do {
    currentResourceTreeItem = currentResourceTreeItem.parent;

    if (!currentResourceTreeItem || !selectedResourceIDs.includes(currentResourceTreeItem.resourceID)) {
      break;
    }

    if (!currentResourceTreeItem.open) {
      closedResourceID = currentResourceTreeItem;
    }
  } while (currentResourceTreeItem.parent);

  if (!closedResourceID) {
    closedResourceID = tree.getResourceByID(currentResourceID);
  }

  return {
    parentID: closedResourceID.parent ? closedResourceID.parent.resourceID : null,
    afterResourceID: closedResourceID.resourceID
  };
};

const reorderAddedResources = params => {
  _balsamiq.default.assert(params.sourceResourceIDs.length === params.targetResourceIDs.length);

  _balsamiq.default.assert(params.sourceResourceIDs.length !== 0);

  let selection_list = null;

  switch (params.kind) {
    case OperationType.DUPLICATE:
      selection_list = params.sourceResourceIDs;
      break;

    case OperationType.PASTE:
      selection_list = params.selectedResourceIDs ? params.selectedResourceIDs : params.sourceResourceIDs;
      break;
  }

  const last_index = selection_list.length - 1;
  const last_item = params.currentResourceTree.getResourceByID(selection_list[last_index]);
  const lastIsParent = last_item.children.length !== 0;
  const lastParentID = last_item.parentID;
  const lastItemID = last_item.resourceID;

  switch (params.kind) {
    case OperationType.DUPLICATE:
      {
        if (params.targetResourceIDs.length === 1) {
          return placeOneElement({
            currentResourceTree: params.currentResourceTree,
            targetResourceID: params.targetResourceIDs[0],
            parentID: lastParentID,
            afterResourceID: lastItemID
          });
        }

        const levels = params.sourceResourceIDs.map(resourceID => {
          const resourceTreeItem = params.currentResourceTree.getResourceByID(resourceID);
          return resourceTreeItem.getLevel();
        }); // duplicate: the new hierarchy is attached
        //            1: at the level of the lowest level in the selection
        //            2: immediately after the last selected item (or parent item)

        let level = 0;
        const lowest_level = Math.min(...levels);
        let cur_item = params.currentResourceTree.getResourceByID(lastItemID);
        let parentResourceID = null;
        let afterResourceID = null;

        do {
          level = cur_item.getLevel();
          afterResourceID = cur_item.resourceID;

          if (level === 0) {
            parentResourceID = null;
            break;
          }

          cur_item = cur_item.parent;
          parentResourceID = cur_item ? cur_item.resourceID : null;
        } while (level > lowest_level || cur_item == null); // find the last selected ancestor with the lowest_level


        return rebuildHierarchy(params.currentResourceTree, levels, parentResourceID, afterResourceID, params);
      }

    case OperationType.PASTE:
      {
        if (params.targetResourceIDs.length === 1) {
          return placeOneElement({
            currentResourceTree: params.currentResourceTree,
            targetResourceID: params.targetResourceIDs[0],
            parentID: lastIsParent ? lastItemID : lastParentID,
            afterResourceID: lastIsParent ? null : lastItemID
          });
        }

        const levels = params.sourceResourceIDs.map(resourceID => {
          const resourceTreeItem = params.sourceResourceTree.getResourceByID(resourceID);
          return resourceTreeItem.getLevel();
        }); // paste: 
        // with no closed nodes the new hierarchy is attached
        //        as child of the last selected if it has children
        //        as a sibling of the last selected otherwise
        // in case the last selected resources belong to a closed hierarchy
        //        as a sibling of the last selected otherwise
        // if last selected resource is a parent, then its children are not included in the selection ...
        // so it cannot be a closed node

        if (lastIsParent) {
          return rebuildHierarchy(params.currentResourceTree, levels, lastItemID, null, params);
        } else {
          const {
            parentID,
            afterResourceID
          } = getPastePosition(params.currentResourceTree, selection_list);
          return rebuildHierarchy(params.currentResourceTree, levels, parentID, afterResourceID, params);
        }
      }
  }
};

exports.reorderAddedResources = reorderAddedResources;

/***/ }),

/***/ "./js/core/utils/resource-skimmer.ts":
/*!*******************************************!*\
  !*** ./js/core/utils/resource-skimmer.ts ***!
  \*******************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports["default"] = void 0;

var _BalsamiqArchiveConstants = __webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchiveConstants */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js");

var _BmprConstants = __webpack_require__(/*! @balsamiq/bmpr/lib/BmprConstants */ "./node_modules/@balsamiq/bmpr/lib/BmprConstants.js");

var MimeTypes = _interopRequireWildcard(__webpack_require__(/*! core/mimetypes */ "./js/core/mimetypes.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function isImageAsset(resource) {
  const {
    kind,
    mimeType,
    extension
  } = resource.ATTRIBUTES;
  return kind === _BmprConstants.ResourceKinds.OtherAsset || kind === _BmprConstants.ResourceKinds.Asset ? extension && MimeTypes.isImageExtension(extension) || mimeType && MimeTypes.isImageMimeType(mimeType) : false;
}

class ResourceSkimmer {
  constructor(_ref) {
    let {
      dump,
      selectedResources,
      loadOnlyImageAssets = false
    } = _ref;

    _defineProperty(this, "dump", void 0);

    _defineProperty(this, "selectedResources", void 0);

    _defineProperty(this, "loadOnlyImageAssets", void 0);

    _defineProperty(this, "wireframesAndSLToFetch", void 0);

    _defineProperty(this, "allAsset", void 0);

    this.dump = dump;
    this.selectedResources = selectedResources;
    this.loadOnlyImageAssets = loadOnlyImageAssets;
    const [wireframesAndSLsToFetch, assetsToFetch] = this.setResourceToFetch();
    this.wireframesAndSLToFetch = wireframesAndSLsToFetch;
    this.allAsset = assetsToFetch;
  }

  setResourceToFetch() {
    const wireframesAndSLsToFetch = [];
    const assetsToFetch = [];
    this.dump.Resources.forEach(resource => {
      if (resource.BRANCHID !== _BalsamiqArchiveConstants.Branch.MasterBranchID) return;

      switch (resource.ATTRIBUTES.kind) {
        case _BmprConstants.ResourceKinds.Mockup:
          {
            if (!this.isResourceSelected(resource)) {
              // do not load the mockup data if the list of resources is passed in and if it doesn't contain the specified resourceID
              return;
            }

            wireframesAndSLsToFetch.push(resource.ID);
            break;
          }

        case _BmprConstants.ResourceKinds.SymbolLibrary:
          wireframesAndSLsToFetch.push(resource.ID);
          break;

        case _BmprConstants.ResourceKinds.Asset:
        case _BmprConstants.ResourceKinds.OtherAsset:
          if (!this.loadOnlyImageAssets || isImageAsset(resource)) {
            assetsToFetch.push(resource.ID);
          }

          break;

        default:
          // callback({ error: "There is an unknown Resource type inside this project" });
          return;
      }
    });
    return [wireframesAndSLsToFetch, assetsToFetch];
  }

  getWireframesAndSLsToFetch() {
    return this.wireframesAndSLToFetch;
  }

  getAllAssets() {
    return this.allAsset;
  }

  hasSelectedResources() {
    return this.selectedResources.length !== 0;
  }

  setResourceData(resourceID, branchID, data) {
    for (const resource of this.dump.Resources) {
      if (resource.ID === resourceID && resource.BRANCHID === branchID) {
        resource.DATA = data;
        break;
      }
    }
  }
  /**
   * Resource is selected if the selection is empty or, when it is specified, if it is in the selection
   */


  isResourceSelected(resource) {
    return !this.hasSelectedResources() || this.selectedResources.find(selectedResource => selectedResource.resourceID === resource.ID && (resource.BRANCHID === "Master" || selectedResource.branchID === resource.BRANCHID)) != null;
  }

  removeEmptyWireframes() {
    this.dump.Resources = this.dump.Resources.filter(r => r.DATA != null || r.ATTRIBUTES.kind === _BmprConstants.ResourceKinds.Asset || r.ATTRIBUTES.kind === _BmprConstants.ResourceKinds.OtherAsset);
  }
  /**
   * Modify the dump by removing empty wireframes/SLs resources and then parse the fetched ones and only return list of assets used by the controls
   */


  removeEmptyWireframesAndExtractUsedAssets() {
    const requiredAssets = new Set();

    function extractDepsFromJson(controls) {
      controls.forEach(c => {
        if (c.properties && c.properties.src) {
          requiredAssets.add(c.properties.src.ID);
        }

        if (c.children && c.children.controls) {
          extractDepsFromJson(c.children.controls.control);
        }
      });
    }

    if (this.hasSelectedResources()) {
      this.removeEmptyWireframes();
      this.dump.Resources.forEach(r => {
        if (r.ATTRIBUTES.kind === _BmprConstants.ResourceKinds.Asset || r.ATTRIBUTES.kind === _BmprConstants.ResourceKinds.OtherAsset) return;
        const data = JSON.parse(r.DATA);
        extractDepsFromJson(data.mockup.controls.control);
      });
      return Array.from(requiredAssets);
    } else {
      return this.getAllAssets();
    }
  }

}

exports["default"] = ResourceSkimmer;

/***/ }),

/***/ "./js/core/utils/semver.ts":
/*!*********************************!*\
  !*** ./js/core/utils/semver.ts ***!
  \*********************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.highestSemVer = highestSemVer;
exports.readSemVer = readSemVer;

var semver = _interopRequireWildcard(__webpack_require__(/*! semver */ "./node_modules/semver/semver.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function readSemVer(version) {
  const r = semver.parse(fixSemVer(version));
  return r ? {
    major: r.major,
    minor: r.minor,
    patch: r.patch
  } : undefined;
}

function fixSemVer(version) {
  let r = version;
  const s = version.split(".").length;

  for (let i = 0; i < 3 - s; i++) {
    r += ".0";
  }

  return r;
}

function highestSemVer(v1, v2) {
  const v1SemVer = readSemVer(v1);
  const v2SemVer = readSemVer(v2);

  if (v1SemVer.major > v2SemVer.major) {
    return v1;
  } else if (v1SemVer.major < v2SemVer.major) {
    return v2;
  } else {
    if (v1SemVer.minor > v2SemVer.minor) {
      return v1;
    } else if (v1SemVer.minor < v2SemVer.minor) {
      return v2;
    } else {
      if (v1SemVer.patch > v2SemVer.patch) {
        return v1;
      } else if (v1SemVer.patch < v2SemVer.patch) {
        return v2;
      } else {
        return v1;
      }
    }
  }
}

/***/ }),

/***/ "./js/core/utils/tarjan-scc.ts":
/*!*************************************!*\
  !*** ./js/core/utils/tarjan-scc.ts ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.Tarjan = exports.Node = exports.Graph = void 0;

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

class Node {
  // this is the node's label in the graph
  // lowest index reachable from the node including itself
  constructor(name) {
    _defineProperty(this, "name", void 0);

    _defineProperty(this, "connections", void 0);

    _defineProperty(this, "index", void 0);

    _defineProperty(this, "lowLink", void 0);

    this.name = name;
    this.connections = [];
    this.index = -1;
    this.lowLink = -1;
  }

  equals(node) {
    return this.name == node.name;
  }

}

exports.Node = Node;

class Graph {
  constructor(nodes) {
    _defineProperty(this, "nodes", void 0);

    this.nodes = nodes;
  }

}

exports.Graph = Graph;

class Nodes {
  constructor(vertices) {
    _defineProperty(this, "nodes", void 0);

    this.nodes = vertices;
  }

  contains(node) {
    for (const i in this.nodes) {
      if (this.nodes[i].equals(node)) {
        return true;
      }
    }

    return false;
  }

}
/*
if we perform a Depth First Search (DPF)
all nodes with the same lowLink are Strong Connected Components (SCC)
but the order is which you start exploring the graph might impact the lowLink value
tarjan adds an invariant to prevent this to happen
it uses a stack in which every node is pushed when visited in the DFS
if you find a node already on the stack, mark the current node lowLink value with the lowest lowLink between the current and the last visited node
if the current node started a SCC then pop all the nodes from the stack until you reach the node again.
A node started a SCC if its id equals its lowLink value
*/


class Tarjan {
  constructor(graph) {
    _defineProperty(this, "index", void 0);

    _defineProperty(this, "stack", void 0);

    _defineProperty(this, "graph", void 0);

    _defineProperty(this, "scc", void 0);

    this.index = 0;
    this.stack = new Nodes([]);
    this.graph = graph;
    this.scc = [];
  }

  run() {
    for (const i in this.graph.nodes) {
      if (this.graph.nodes[i].index < 0) {
        this.strongConnect(this.graph.nodes[i]);
      }
    }

    return this.scc;
  }

  strongConnect(node) {
    // Set the depth index for v to the smallest unused index
    node.index = this.index;
    node.lowLink = this.index;
    this.index = this.index + 1;
    this.stack.nodes.push(node); // Consider successors of v
    // aka... consider each node in node.connections

    for (const i in node.connections) {
      const v = node;
      const w = node.connections[i];

      if (w.index < 0) {
        // Successor w has not yet been visited; recurse on it
        this.strongConnect(w);
        v.lowLink = Math.min(v.lowLink, w.lowLink);
      } else if (this.stack.contains(w)) {
        // Successor w is in stack S and hence in the current SCC
        v.lowLink = Math.min(v.lowLink, w.index);
      }
    } // If v is a root node, pop the stack and generate an SCC


    if (node.lowLink == node.index) {
      // start a new strongly connected component
      const nodes = [];
      let w;

      if (this.stack.nodes.length > 0) {
        do {
          w = this.stack.nodes.pop(); // add w to current strongly connected component

          nodes.push(w);
        } while (!node.equals(w));
      } // output the current strongly connected component
      // ... i'm going to push the results to a member scc array variable


      if (nodes.length > 0) {
        this.scc.push(nodes);
      }
    }
  }

}

exports.Tarjan = Tarjan;

/***/ }),

/***/ "./js/core/viewcontrollers/alertvc.ts":
/*!********************************************!*\
  !*** ./js/core/viewcontrollers/alertvc.ts ***!
  \********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ViewController = exports.Type = exports.States = exports.KeyNames = exports.Enums = exports.ChildNames = exports.Actions = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var ViewController = _interopRequireWildcard(__webpack_require__(/*! core/viewcontrollers/viewcontroller */ "./js/core/viewcontrollers/viewcontroller.ts"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var Actions;
exports.Actions = Actions;

(function (Actions) {
  Actions["BUTTON"] = "button";
})(Actions || (exports.Actions = Actions = {}));

const ButtonThemes = {
  ALERT: "alert",
  DEFAULT: "default",
  PRIMARY: "primary"
};
const States = {
  MESSAGE: "MESSAGE",
  TITLE: "TITLE",
  CLOSABLE: "CLOSABLE",
  BUTTONS: "BUTTONS"
};
exports.States = States;
const ButtonIdentifiers = {
  CANCEL: "cancel",
  MAIN: "main"
};
const KeyNames = {};
exports.KeyNames = KeyNames;

class AlertViewController extends ViewController.ViewController {
  constructor(context, vcType, options) {
    const initialState = options && options.initialState || {
      MESSAGE: [],
      TITLE: "Alert",
      CLOSABLE: true,
      BUTTONS: []
    };
    super(context, vcType, options, initialState);

    _defineProperty(this, "callback", void 0);

    this.callback = null; // function ( { index: button index } )

    this.setButtons(initialState.BUTTONS || [{
      text: "Close",
      type: ButtonIdentifiers.MAIN
    }]);
  }

  action(action, params) {
    switch (action) {
      case Actions.BUTTON:
        if (this.callback) {
          this.callback(params);
        }

        return true;
    }

    return super.action(action, params);
  }

  setCallback(callback) {
    _balsamiq.default.assertFunction(callback);

    this.callback = callback;
  }

  setMessage(msg) {
    this.setStates({
      MESSAGE: msg
    });
  }

  setButtons(buttons) {
    _balsamiq.default.assertArray(buttons);

    buttons.forEach(button => {
      _balsamiq.default.assertString(button.text);

      if (button.name) _balsamiq.default.assertString(button.name);
      if (button.action) _balsamiq.default.assertString(button.action);
      if (button.theme) _balsamiq.default.assertString(button.theme);
      if (button.type) _balsamiq.default.assertEnum(button.type, ButtonIdentifiers);
    });
    this.setStates({
      BUTTONS: buttons
    });
  }

  setTitle(title) {
    _balsamiq.default.assertString(title);

    this.setStates({
      TITLE: title
    });
  }

  setClosable(closable) {
    _balsamiq.default.assertBool(closable);

    this.setStates({
      CLOSABLE: closable
    });
  }

}

exports.ViewController = AlertViewController;
const Type = "alert";
exports.Type = Type;
const ChildNames = {};
exports.ChildNames = ChildNames;
const Enums = {
  ButtonIdentifiers,
  ButtonThemes
};
exports.Enums = Enums;

/***/ }),

/***/ "./js/core/viewcontrollers/viewcontroller.ts":
/*!***************************************************!*\
  !*** ./js/core/viewcontrollers/viewcontroller.ts ***!
  \***************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ViewController = exports.Type = exports.States = exports.KeyboardShortcuts = exports.KeyNames = exports.ChildNames = exports.Actions = exports.ActionTypes = void 0;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _coreobject = __webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

class ViewController extends _coreobject.CoreObject {
  constructor(context, vcType, options, defaultState) {
    _balsamiq.default.exists(context);

    _balsamiq.default.exists(context.host);

    _balsamiq.default.exists(context.viewControllers);

    _balsamiq.default.exists(context.model);

    _balsamiq.default.exists(context.eventCenter);

    _balsamiq.default.exists(context.runTask);

    _balsamiq.default.assertString(vcType);

    super(context, (0, _coreobject.makeVoidCoreObjectID)());

    _defineProperty(this, "state", void 0);

    _defineProperty(this, "nativeID", void 0);

    _defineProperty(this, "parentNativeID", void 0);

    _defineProperty(this, "windowID", void 0);

    _defineProperty(this, "taskID", void 0);

    _defineProperty(this, "changedKeys", void 0);

    _defineProperty(this, "children", void 0);

    _defineProperty(this, "vcType", void 0);

    _defineProperty(this, "name", void 0);

    _defineProperty(this, "commandQueue", void 0);

    this.context = context;
    this.changedKeys = {}; // String -> true

    this.children = {}; // name -> ViewController instance

    this.vcType = vcType;
    this.name = "GenericVC";
    this.commandQueue = [];
    this.state = defaultState || {};
    this.init(options);

    for (const key in this.state) {
      if (this.state[key] instanceof Object) {
        this.state[key] = _balsamiq.default.cloneDeep(this.state[key]);
      }
    }
  }
  /**
   * @deprecated Available only in JS classes.
   */


  init(options) {// subclass only for JS classes
  }

  cleanUp() {
    this.nativeID = undefined;
    this.taskID = undefined;
    this.children = {};
    super.cleanUp();
  }

  getState() {
    return this.state;
  }

  setStates(s) {
    for (const k in s) {
      this.setState(k, s[k]);
    }
  }

  setState(key, value) {
    _balsamiq.default.assertString(key);

    if (_balsamiq.default.isEqual(value, this.state[key])) {
      // this.context.host.log(`VC: skip setState ${key} of VC ${this.nativeID} ${this.vcType}`); // FOR DEBUG PURPOSES ONLY
      return;
    }

    this.state[key] = value instanceof Object ? _balsamiq.default.cloneDeep(value) : value;
    this.changedKeys[key] = true;

    if (this.nativeID) {
      // this.context.host.log(`VC: schedule setState ${key} of VC ${this.nativeID} ${this.vcType}`); // FOR DEBUG PURPOSES ONLY
      this.context.viewControllers.schedulePushState(this.nativeID);
    }
  }

  getChangedStateKeys() {
    return _balsamiq.default.clone(this.changedKeys);
  }

  onPushedState() {
    // let changedKeys = Object.keys(this.changedKeys);
    // let allKeys = Object.keys(this.state);
    // this.context.host.log(`VC: reset changed state with ${changedKeys.length} of ${allKeys.length} states of VC ${this.nativeID} ${this.vcType}`);
    this.changedKeys = {};
  }

  sendCommand(commandKey, data) {
    _balsamiq.default.assertString(commandKey);

    this.commandQueue.push({
      commandKey,
      data
    });

    if (this.nativeID) {
      this.context.viewControllers.sendCommands(this.nativeID);
    }
  }

  popCommands() {
    const c = this.commandQueue;
    this.commandQueue = [];
    return c;
  }

  getNativeID() {
    return this.nativeID;
  }

  setNativeID(nativeID) {
    this.nativeID = nativeID;
  }

  getParentNativeID() {
    return this.parentNativeID;
  }

  setParentNativeID(parentNativeID) {
    this.parentNativeID = parentNativeID;
  }

  getWindowID() {
    return this.windowID;
  }

  setWindowID(value) {
    this.windowID = value;

    for (const childName in this.children) {
      if (this.children[childName]) {
        this.children[childName].setWindowID(value);
      }
    }
  }

  getTaskID() {
    return this.taskID;
  }

  setTaskID(taskID) {
    this.taskID = taskID;
  }

  action(action, params) {
    // subclass
    return false;
  }

  mouseEvent(event) {
    const names = this.getActiveChildNames();
    const child = names.find(name => {
      const child = this.getChild(name);

      if (child && child.mouseEvent(event)) {
        return true;
      }

      return false;
    });
    return child != null;
  }

  keyboardEvent(event, hasFocus) {
    _balsamiq.default.assertBool(hasFocus, "viewcontroller keyboardEvent needs hasFocus parameter"); // subclass but then call super.keyboardEvent(event, hasFocus)


    if (hasFocus) {
      return false;
    } else {
      // pass it to the active children
      const names = this.getActiveChildNames();
      const child = names.find(name => {
        const child = this.getChild(name);

        if (child && this.getChild(name).keyboardEvent(event, hasFocus)) {
          return true;
        }

        return false;
      });
      return child != null;
    }
  }

  getChildNames() {
    // subclass
    return [];
  }

  getActiveChildNames() {
    // subclass
    return this.getChildNames();
  }

  getChild(childName) {
    _balsamiq.default.assertEnum(childName, this.getChildNames());

    return this.children[childName];
  }

  setChild(childName, child) {
    _balsamiq.default.assertEnum(childName, this.getChildNames());

    this.children[childName] = child;
  }

  getChildName(child) {
    _balsamiq.default.existsClass(child, ViewController);

    for (const childName in this.children) {
      if (this.children[childName] == child) {
        return childName;
      }
    }

    return null;
  }

  getVCType() {
    return this.vcType;
  }

  getName() {
    return this.name;
  }

  setName(name) {
    _balsamiq.default.assertString(name);

    this.name = name;
  }

  runTask(task, callback) {
    _balsamiq.default.assertString(this.taskID, "cannot run a Task because vc.taskID is not set");

    this.context.runTask(this.taskID, task, callback);
  }

  runTaskIfPossible(task, callback) {
    _balsamiq.default.assertString(this.taskID, "cannot run a Task because vc.taskID is not set");

    return this.context.runTaskIfPossible(this.taskID, task, callback);
  }

}

exports.ViewController = ViewController;
const Type = "";
exports.Type = Type;
const ChildNames = {};
exports.ChildNames = ChildNames;
const Actions = {};
exports.Actions = Actions;
const KeyNames = {};
exports.KeyNames = KeyNames;
const States = {};
exports.States = States;
const KeyboardShortcuts = {};
exports.KeyboardShortcuts = KeyboardShortcuts;
const ActionTypes = {}; // optional, action -> ViewControllerActionTypes

exports.ActionTypes = ActionTypes;

/***/ }),

/***/ "./js/core/viewmodels/viewmodel.ts":
/*!*****************************************!*\
  !*** ./js/core/viewmodels/viewmodel.ts ***!
  \*****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ViewModel = exports.EventTypes = void 0;

var CoreObject = _interopRequireWildcard(__webpack_require__(/*! core/core/coreobject */ "./js/core/core/coreobject.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

const EventTypes = {};
exports.EventTypes = EventTypes;

class ViewModel extends CoreObject.CoreObject {
  constructor(context) {
    super(context, CoreObject.makeVoidCoreObjectID());
  }

}

exports.ViewModel = ViewModel;

/***/ }),

/***/ "./js/core/wireframe-version.ts":
/*!**************************************!*\
  !*** ./js/core/wireframe-version.ts ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.NestedSymbolVersion = exports.CurrentVersion = exports.BaseVersion = void 0;
const BaseVersion = "1.0";
exports.BaseVersion = BaseVersion;
const NestedSymbolVersion = "2.0";
exports.NestedSymbolVersion = NestedSymbolVersion;
const CurrentVersion = NestedSymbolVersion; // keep the currentVersion updated with the code that can support it

exports.CurrentVersion = CurrentVersion;

/***/ }),

/***/ "./js/editor/tasks/loadprojecttask.ts":
/*!********************************************!*\
  !*** ./js/editor/tasks/loadprojecttask.ts ***!
  \********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.LoadProjectTask = void 0;
Object.defineProperty(exports, "LoadingErrors", ({
  enumerable: true,
  get: function () {
    return _projectcontroller.LoadingErrors;
  }
}));
exports.UpdateProjectTask = exports.SourceTypes = void 0;
exports.buildLoadProjectAlertTask = buildLoadProjectAlertTask;

var _balsamiq = _interopRequireDefault(__webpack_require__(/*! core/balsamiq */ "./js/core/balsamiq.js"));

var _task = __webpack_require__(/*! core/tasks/task */ "./js/core/tasks/task.ts");

var Model = _interopRequireWildcard(__webpack_require__(/*! core/core/model */ "./js/core/core/model.js"));

var ModelConstants = _interopRequireWildcard(__webpack_require__(/*! core/model/constants */ "./js/core/model/constants.ts"));

var _updateprojectfromdumpoperation = _interopRequireDefault(__webpack_require__(/*! editor/operations/updateprojectfromdumpoperation */ "./js/editor/operations/updateprojectfromdumpoperation.js"));

var MimeTypes = _interopRequireWildcard(__webpack_require__(/*! core/mimetypes */ "./js/core/mimetypes.js"));

var FilePickerVC = _interopRequireWildcard(__webpack_require__(/*! editor/viewcontrollers/filepickerviewcontroller */ "./js/editor/viewcontrollers/filepickerviewcontroller.js"));

var _native = _interopRequireWildcard(__webpack_require__(/*! core/native */ "./js/core/native.js"));

var _alerttask = __webpack_require__(/*! editor/tasks/alerttask */ "./js/editor/tasks/alerttask.js");

var _newerbmprversionalerttask = __webpack_require__(/*! editor/tasks/newerbmprversionalerttask */ "./js/editor/tasks/newerbmprversionalerttask.ts");

var _initnativemoduletask = __webpack_require__(/*! core/tasks/initnativemoduletask */ "./js/core/tasks/initnativemoduletask.ts");

var _projectcontroller = __webpack_require__(/*! core/controllers/projectcontroller */ "./js/core/controllers/projectcontroller.js");

var _json = __webpack_require__(/*! core/serialization/json */ "./js/core/serialization/json.js");

var BARConstants = _interopRequireWildcard(__webpack_require__(/*! @balsamiq/bmpr/lib/BalsamiqArchiveConstants */ "./node_modules/@balsamiq/bmpr/lib/BalsamiqArchiveConstants.js"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var SourceTypes;
exports.SourceTypes = SourceTypes;

(function (SourceTypes) {
  SourceTypes["BAS"] = "bas";
  SourceTypes["DUMP"] = "dump";
  SourceTypes["FILE_URL"] = "fileUrl";
  SourceTypes["SQLITE"] = "sqlite";
  SourceTypes["URL"] = "url";
})(SourceTypes || (exports.SourceTypes = SourceTypes = {}));

class LoadProjectTask extends _task.TaskBase {
  constructor(params) {
    super(params);

    _defineProperty(this, "cancelableOp", null);
  }

  name() {
    return "Load Project";
  }

  startProgress(title) {
    if (this.params.useEditorProgress && this.context.startEditorProgress) {
      this.context.startEditorProgress("Loading project…");
    } else {
      super.startProgress(title);
    }
  }

  setProgress(title, currentStep, totalSteps) {
    if (this.params.useEditorProgress && this.context.setEditorProgress) {
      this.context.setEditorProgress("Loading project…", currentStep, totalSteps);
    } else {
      super.setProgress(title, currentStep, totalSteps);
    }
  }

  endProgress() {
    if (this.params.useEditorProgress && this.context.endEditorProgress) {
      this.context.endEditorProgress();
    } else {
      super.endProgress();
    }
  }

  getOptions() {
    return this.context.editorOptions ? this.context.editorOptions : this.context.viewerOptions;
  }

  doRun() {
    _balsamiq.default.assertEnum(this.params.sourceType, SourceTypes);

    if (this.params.projectLoadedCallback) {
      _balsamiq.default.assertFunction(this.params.projectLoadedCallback);
    }

    if (this.params.isNewUnsavedFile !== undefined) {
      _balsamiq.default.assertBool(this.params.isNewUnsavedFile);
    }

    if (this.params.skipSelectProject !== undefined) {
      _balsamiq.default.assertBool(this.params.skipSelectProject);
    }

    if (this.params.cancelButtonLabel !== undefined) {
      _balsamiq.default.assertString(this.params.cancelButtonLabel);

      this.setCancelButtonLabel(this.params.cancelButtonLabel);
    } else {
      this.setCancelButtonLabel("Close Project");
    }

    _balsamiq.default.assertOptionalBool(this.params.useEditorProgress); // params.cfg.createProject (optional) boolean: call create on BAS. by default it is false.
    // if params.cfg.platformArchiveID is not specified, then call create on BAS.


    switch (this.params.sourceType) {
      case SourceTypes.BAS:
        this.params.isFirstLoad = !!this.params.isFirstLoad;
        this.loadProjectFromBAS();
        break;

      case SourceTypes.DUMP:
        _balsamiq.default.exists(this.params.dump);

        _balsamiq.default.assertOptionalEnum(this.params.userRole, BARConstants.Role);

        this.loadProjectFromDump(this.params.dump, this.params.dumpID, this.params.userRole);
        break;

      case SourceTypes.FILE_URL:
        _balsamiq.default.exists(this.params.filePath);

        _balsamiq.default.assertOptionalEnum(this.params.userRole, BARConstants.Role);

        this.loadProjectFromFilePath(this.params.filePath);
        break;

      case SourceTypes.SQLITE:
        if (this.params.sqliteID) {
          _balsamiq.default.assertOptionalEnum(this.params.userRole, BARConstants.Role);

          this.loadProjectFromSQLite(this.params.sqliteID, this.getOptions() ? this.getOptions().localUserInfo : null);
        } else {
          this.runSubtask(new _initnativemoduletask.InitNativeModuleTask({
            moduleName: _native.default.ModuleNames.SQLite
          }), res => {
            if (res.error) {
              this.notifyFinished(res);
            } else {
              this.context.viewControllers.make(FilePickerVC.Type, this.getTaskID(), null, filePickerVC => {
                const filePickerCallback = (ok, res) => {
                  this.context.viewControllers.pop(this.context.windowID, filePickerVC.getNativeID());

                  if (ok && res[0] && res[0].id) {
                    this.loadProjectFromSQLite(res[0].id, this.getOptions() ? this.getOptions().localUserInfo : null);
                  } else {
                    this.notifyFinished();
                  }
                };

                filePickerVC.setup(ModelConstants.ResourceTypes.MOCKUP, [MimeTypes.findEntryForMimeType(MimeTypes.BMPR_MIMETYPE).extensions[0]], [], 1, -1, "Select a project to open", filePickerCallback);
                this.context.viewControllers.push(this.context.windowID, filePickerVC.getNativeID(), {}, true);
              });
            }
          });
        }

        break;

      case SourceTypes.URL:
        if (this.params.url) {
          this.loadProjectFromURL(this.params.url);
        }

        break;
    }
  }

  loadProjectFromURL(url) {
    this.runSubtask(new _initnativemoduletask.InitNativeModuleTask({
      moduleName: _native.ModuleNames.SQLite
    }), () => {
      if (this.canceled) return this.notifyFinished();
      this.startOperation(() => {
        if (this.canceled) return this.notifyFinished();
        this.setProgress("Loading Project…", 0, 0);
        this.context.host.HttpRequest.getBinary(url, {
          base64: false
        }, resp => {
          this.cancelableOp = null;
          if (this.canceled) return this.notifyFinished();

          if (!resp.error) {
            const sqliteBuffer = new Uint8Array(resp.data);

            _native.default.Host.SQLite.makeFromData(sqliteBuffer, res => {
              if (!res.error) {
                this._loadFromSQLite(res.id, null);
              } else {
                this._onProjectLoaded(res, null, null);
              }
            });
          } else {
            this._onProjectLoaded(resp, null, null);
          }
        }, req => {
          this.cancelableOp = () => req.abort();
        });
      });
    });
  }

  loadProjectFromFilePath(filePath) {
    // TODO rename to loadProjectFromFileURL
    this.runSubtask(new _initnativemoduletask.InitNativeModuleTask({
      moduleName: _native.ModuleNames.SQLite
    }), () => {
      this.startOperation(() => {
        this.setProgress("Loading Project from file", 0, 0);
        this.context.host.UnsavedFiles.isUnsavedFileURL(filePath, res => {
          _balsamiq.default.assertBool(res.result);

          const isUnsaved = res.result;

          if (isUnsaved) {
            _balsamiq.default.assertBool(res.unsavedFileExists);

            if (!res.unsavedFileExists) {
              // TODO show an error, the file does not exists
              _balsamiq.default.assert(false, "The UnsavedFile does not exists: " + filePath);
            } else {
              const loadFileParams = {
                filePath: filePath,
                userFileURL: filePath,
                isEdited: _balsamiq.default.isBoolean(this.params.isNewUnsavedFile) ? !this.params.isNewUnsavedFile : true,
                isUnsaved: true
              };
              this.loadFile(loadFileParams, unsavedFileLoadRes => {
                if (unsavedFileLoadRes.error) {
                  this.context.host.UnsavedFiles.markUnsavedFileAsCorrupted(filePath, () => {
                    this._onProjectLoaded(unsavedFileLoadRes, null, null);
                  });
                } else {
                  if (unsavedFileLoadRes.error) {
                    this._onProjectLoaded({ ...unsavedFileLoadRes,
                      didLoadPreviousLocalFileURL: false
                    }, null, null);
                  } else {
                    this._onProjectLoaded({ ...unsavedFileLoadRes,
                      didLoadPreviousLocalFileURL: false
                    }, unsavedFileLoadRes.pc, unsavedFileLoadRes.archiveID);
                  }
                }
              });
            }
          } else {
            this.context.host.LocalFiles.isLocalFileURL(filePath, _ref => {
              let {
                result
              } = _ref;

              if (result) {
                // user is not allowed to open local files directly
                this._onProjectLoaded({
                  error: _projectcontroller.LoadingErrors.forbiddenLocation
                }, null, null);
              } else {
                // { localFileURL:String, isEdited:Bool, isOlderThanUserFile:Bool } or null
                this.context.host.LocalFiles.getLocalFileFrom(filePath, getLocalFileFromRes => {
                  if (getLocalFileFromRes) {
                    if (getLocalFileFromRes.isOlderThanUserFile) {
                      this.loadUneditedFile(filePath, getLocalFileFromRes.localFileURL);
                    } else if (getLocalFileFromRes.isEdited) {
                      this.loadEditedFile(filePath, getLocalFileFromRes.localFileURL);
                    } else {
                      this.loadUneditedFile(filePath, getLocalFileFromRes.localFileURL);
                    }
                  }
                });
              }
            });
          }
        });
      });
    });
  }

  loadUneditedFile(filePath, previousLocalFileURL) {
    // 1. load user file (creating a new local file)
    // 2. if corruted search for a previous local file
    // 3. if corrupted: fail
    this.context.host.LocalFiles.createLocalFileFrom(filePath, localFileCreationRes => {
      if (localFileCreationRes.ok) {
        const loadFileParams = {
          filePath: localFileCreationRes.localFileURL,
          userFileURL: filePath,
          isEdited: false,
          isUnsaved: false
        };
        this.loadFile(loadFileParams, newLocalFileRes => {
          if (newLocalFileRes.error) {
            this.context.host.LocalFiles.markLocalFileAsCorruptedAndRemoveFromRegistry(filePath, () => {
              // local file is corrupted -> user file is corrupted too, search for a previous local file
              if (previousLocalFileURL) {
                // set previous as the local file for our user file
                this.context.host.LocalFiles.updateLocalFileURL(filePath, previousLocalFileURL, () => {
                  // try opening it
                  const previousLoadFileParams = {
                    filePath: previousLocalFileURL,
                    userFileURL: filePath,
                    isEdited: true,
                    isUnsaved: false
                  };
                  this.loadFile(previousLoadFileParams, previousLocalFileRes => {
                    if (previousLocalFileRes.error) {
                      this.context.host.LocalFiles.markLocalFileAsCorruptedAndRemoveFromRegistry(filePath, () => {
                        this._onProjectLoaded({ ...previousLocalFileRes,
                          didLoadPreviousLocalFileURL: false
                        }, null, null);
                      });
                    } else {
                      this._onProjectLoaded({ ...previousLocalFileRes,
                        didLoadPreviousLocalFileURL: true
                      }, previousLocalFileRes.pc, previousLocalFileRes.archiveID);
                    }
                  });
                });
              } else {
                this._onProjectLoaded({ ...newLocalFileRes,
                  didLoadPreviousLocalFileURL: false
                }, null, null);
              }
            });
          } else {
            this._onProjectLoaded({ ...newLocalFileRes,
              didLoadPreviousLocalFileURL: false
            }, newLocalFileRes.pc, newLocalFileRes.archiveID);
          }
        });
      } else {
        this._onProjectLoaded({ ...localFileCreationRes,
          didLoadPreviousLocalFileURL: false
        }, null, null);
      }
    });
  }

  loadEditedFile(filePath, localFileURL) {
    // 1. load local file
    // 2. if corrupted load user file (creating a new local file)
    // 3. if corrupted: fail
    const loadFileParams = {
      filePath: localFileURL,
      userFileURL: filePath,
      isEdited: true,
      isUnsaved: false
    };
    this.loadFile(loadFileParams, loadLocalFileRes => {
      if (loadLocalFileRes.error) {
        this.context.host.LocalFiles.markLocalFileAsCorruptedAndRemoveFromRegistry(filePath, () => {
          // restart with user file creating a new local file
          this.context.host.LocalFiles.createLocalFileFrom(filePath, localFileCreationRes => {
            if (localFileCreationRes.ok) {
              const localFileParams = {
                filePath: localFileCreationRes.localFileURL,
                userFileURL: filePath,
                isEdited: false,
                isUnsaved: false
              };
              this.loadFile(localFileParams, newLocalFileRes => {
                if (newLocalFileRes.error) {
                  this.context.host.LocalFiles.markLocalFileAsCorruptedAndRemoveFromRegistry(filePath, () => {
                    this._onProjectLoaded({ ...newLocalFileRes,
                      didLoadPreviousLocalFileURL: false
                    }, null, null);
                  });
                } else {
                  this._onProjectLoaded({ ...newLocalFileRes,
                    didLoadPreviousLocalFileURL: true
                  }, newLocalFileRes.pc, newLocalFileRes.archiveID);
                }
              });
            } else {
              this._onProjectLoaded({ ...localFileCreationRes,
                didLoadPreviousLocalFileURL: false
              }, null, null);
            }
          });
        });
      } else {
        this._onProjectLoaded({ ...loadLocalFileRes,
          didLoadPreviousLocalFileURL: false
        }, loadLocalFileRes.pc, loadLocalFileRes.archiveID);
      }
    });
  }

  loadFile(_ref2, callback) {
    let {
      filePath,
      userFileURL,
      isEdited,
      isUnsaved
    } = _ref2;
    this.context.host.SQLite.makeFromFileURL(filePath, res => {
      if (res.error) {
        callback(res);
      } else {
        this._loadFromSQLiteExt(res.id, {
          userFileURL: userFileURL,
          isEdited: isEdited,
          isUnsaved: isUnsaved
        }, undefined, (res, pc, archiveID) => {
          callback({ ...res,
            pc,
            archiveID
          });
        });
      }
    });
  }

  loadProjectFromBAS() {
    const cfg = this.getOptions().getBasCfg();

    _balsamiq.default.assertString(cfg.serverURL);

    _balsamiq.default.assertString(cfg.kind);

    if (cfg.userInfo) {
      _balsamiq.default.assertString(cfg.userInfo.name);
    }

    this.startOperation(() => {
      this.setProgress("Loading Project…", 0, 0);
      const archiveID = this.context.archives.addArchiveFromBAS(cfg.serverURL, cfg.kind, cfg.platformToken, cfg.userInfo, this.params.rtcClient);
      const project = this.context.model.addProject(Model.DEFAULT_SITEID, cfg.platformSiteID + ":" + cfg.platformArchiveID);

      if (cfg.maxPostSize) {
        project.maxResourceDataSize = cfg.maxPostSize - 64 * 1024; //64K is a conservative overhead for the POST request
      }

      const pc = this.context.controllers.makeNewProjectController(archiveID, project, this.getOptions());

      this.cancelableOp = () => pc.cancelLoadProject();

      if (cfg.createProject || !cfg.platformArchiveID) {
        pc.createProjectBAS(res => {
          if (res.error) {
            // console.log("Error creating new project on BAS: " + JSON.stringify(res));
            this._onProjectLoaded(res, pc, archiveID);
          } else {
            this._continueLoadProjectFromBAS(pc, archiveID, res);
          }
        });
      } else {
        this._continueLoadProjectFromBAS(pc, archiveID);
      }
    });
  }

  _continueLoadProjectFromBAS(pc, archiveID, createRes) {
    if (this.params.sourceType !== SourceTypes.BAS) {
      return _balsamiq.default.assert(false, "SourceType must be BAS");
    }

    if (createRes && !createRes.error) {
      const basCfg = this.getOptions().getBasCfg();

      if (_balsamiq.default.isString(createRes.platformArchiveID)) {
        basCfg.platformArchiveID = createRes.platformArchiveID;
      }

      if (_balsamiq.default.isString(createRes.platformArchiveName)) {
        basCfg.platformArchiveName = createRes.platformArchiveName;
      }

      this.getOptions().setBasCfg(basCfg);
    }

    pc.loadProjectBAS(res => {
      this.cancelableOp = null;

      this._onProjectLoaded(res, pc, archiveID);
    }, this.setProgress.bind(this), !!this.params.isFirstLoad, this.params.selectedResources);
  }

  loadProjectFromDump(dump, dumpID, userRole) {
    _balsamiq.default.exists(dump);

    if (!_balsamiq.default.isString(dumpID)) {
      dumpID = null;
    }

    _balsamiq.default.assertOptionalEnum(userRole, BARConstants.Role);

    this.startOperation(() => {
      this.setProgress("Loading Project from Memory…", 0, 0);
      const archiveID = this.context.archives.addArchiveInMemory(dump);
      const project = this.context.model.addProject(Model.DEFAULT_SITEID, dumpID);
      const pc = this.context.controllers.makeNewProjectController(archiveID, project, this.getOptions());

      this.cancelableOp = () => pc.cancelLoadProject();

      pc.loadProjectDump(res => {
        this.cancelableOp = null;

        this._onProjectLoaded(res, pc, archiveID);
      }, undefined, userRole);
    });
  }

  loadProjectFromSQLite(sqliteID, userInfo) {
    _balsamiq.default.assertString(sqliteID);

    this.startOperation(() => {
      this.setProgress("Loading Project from BMPR archive…", 0, 0);

      this._loadFromSQLite(sqliteID, userInfo);
    });
  }

  _loadFromSQLite(sqliteID, userInfo) {
    this._loadFromSQLiteExt(sqliteID, undefined, userInfo, this._onProjectLoaded.bind(this));
  }

  _loadFromSQLiteExt(sqliteID, fileInfo, userInfo, callback) {
    _balsamiq.default.assertString(sqliteID);

    _balsamiq.default.assertFunction(callback);

    const archiveID = this.context.archives.addArchiveFromSQLite(sqliteID, userInfo);
    const fileUrl = fileInfo ? fileInfo.userFileURL : null;
    const project = this.context.model.addProject(Model.DEFAULT_SITEID, fileUrl);
    const projectController = this.context.controllers.makeNewProjectController(archiveID, project, this.getOptions());

    this.cancelableOp = () => projectController.cancelLoadProject();

    projectController.loadProjectSQLite({
      sqliteID,
      userFile: fileInfo,
      userInfo,
      userRole: _balsamiq.default.checkEnum(this.params.userRole, BARConstants.Role) ? this.params.userRole : undefined,
      setProgress: this.setProgress.bind(this),
      selectedResources: this.params.selectedResources
    }, res => {
      this.cancelableOp = null;
      callback(res, projectController, archiveID);
    });
  }

  _onProjectLoaded(res, projectController, archiveID) {
    this.finishOperation();
    const resPayload = {
      projectCoreObjectID: _balsamiq.default.get(projectController, "project.coreObjectID"),
      projectController,
      archiveID
    };

    if (res.canceled) {
      if (this.params.projectLoadedCallback) {
        this.params.projectLoadedCallback({ ...resPayload,
          canceled: true
        }, projectController);
      }

      this.notifyFinished();
      return;
    }

    if (this.context.projectListViewModel) {
      if (this.params.skipSelectProject == undefined || this.params.skipSelectProject == false) {
        this.context.projectListViewModel.selectProjectID(_balsamiq.default.get(projectController, "project.coreObjectID.projectID"));
      }
    }

    if (this.params.projectLoadedCallback) {
      if (res.error) {
        this.params.projectLoadedCallback({ ...res,
          ...resPayload
        }, projectController);
      } else {
        this.params.projectLoadedCallback(resPayload, projectController);
      }
    }

    if (this.getOptions() && this.getOptions().getProjectLoadedCallback && this.getOptions().getProjectLoadedCallback()) {
      const basCfg = this.getOptions().getBasCfg();
      this.getOptions().getProjectLoadedCallback()(basCfg ? {
        platformArchiveName: basCfg.platformArchiveName,
        platformArchiveID: basCfg.platformArchiveID
      } : undefined);
    }

    if (this.context.viewerOptions && this.context.viewerOptions.getProjectLoadedCallback()) {
      this.context.viewerOptions.getProjectLoadedCallback()();
    }

    if (res.error) {
      if (this.context.viewControllers && !this.params.disableProgress && !this.params.disableAlertOnError) {
        this.runSubtask(buildLoadProjectAlertTask(res), () => {
          this.notifyFinished({ ...resPayload,
            error: res.error,
            filePath: res.filePath
          });
        });
      } else {
        this.notifyFinished({ ...resPayload,
          error: res.error,
          filePath: res.filePath
        });
      }
    } else {
      this.notifyFinished({ ...res,
        ...resPayload
      });
    }
  }

  isCancelable() {
    return true;
  }

  onCanceled() {
    if (this.cancelableOp) {
      this.cancelableOp();
    }
  }

}

exports.LoadProjectTask = LoadProjectTask;

class UpdateProjectTask extends _task.TaskBase {
  constructor() {
    super(...arguments);

    _defineProperty(this, "operation", void 0);
  }

  doRun() {
    _balsamiq.default.assertEnum(this.params.sourceType, SourceTypes);

    if (this.params.projectLoadedCallback) {
      _balsamiq.default.assertFunction(this.params.projectLoadedCallback);
    }

    switch (this.params.sourceType) {
      case SourceTypes.BAS:
        _balsamiq.default.assert(false);

        break;

      case SourceTypes.DUMP:
        _balsamiq.default.exists(this.params.dump);

        this.updateProjectFromDump(this.params.dump, this.params.projectController);
        break;

      case SourceTypes.SQLITE:
        _balsamiq.default.assert(false);

        break;

      case SourceTypes.URL:
        _balsamiq.default.assertString(this.params.url);

        _balsamiq.default.exists(this.params.projectController);

        this.updateProjectFromURL();
        break;
    }
  }

  updateProjectFromDump(dump, projectController) {
    _balsamiq.default.exists(dump);

    _balsamiq.default.exists(projectController); // Start Operation


    this.operation = this.scheduleOperation(new _updateprojectfromdumpoperation.default.Operation({
      projectController: projectController,
      dump: dump
    }), () => {
      if (this.params.projectLoadedCallback) {
        this.params.projectLoadedCallback();
      }

      this.notifyFinished();
    });
  }

  updateProjectFromURL() {
    this.startOperation(() => {
      this.setProgress("Loading Project…", 0, 0);
      this.context.host.HttpRequest.getBinary(this.params.url, {
        base64: false
      }, resp => {
        if (!resp.error) {
          const sqliteBuffer = new Uint8Array(resp.data);

          _native.default.Host.SQLite.makeFromData(sqliteBuffer, res => {
            if (!res.error) {
              const archiveID = this.context.archives.addArchiveFromSQLite(res.id, null);
              this.params.projectController.setArchiveID(archiveID);
              this.params.projectController.reloadProjectSQLite(res.id, res => {
                if (this.params.projectLoadedCallback) {
                  this.params.projectLoadedCallback(res);
                }

                this.notifyFinished(res);
              }, this.setProgress.bind(this));
            } else {
              this.notifyFinished(res);
            }
          });
        } else {
          this.notifyFinished(resp);
        }
      });
    });
  }

}

exports.UpdateProjectTask = UpdateProjectTask;

function buildLoadProjectAlertTask(res) {
  if (res.error === _json.ValidationErrorCodes.UNSUPPORTED_VERSION) {
    return new _newerbmprversionalerttask.NewerBmprVersionAlertTask({});
  } else {
    return new _alerttask.AlertTask({
      message: "An error occurred loading the project.\n\n" + res.error,
      title: "Load Error"
    });
  }
}

/***/ }),

/***/ "./js/editor/tasks/newerbmprversionalerttask.ts":
/*!******************************************************!*\
  !*** ./js/editor/tasks/newerbmprversionalerttask.ts ***!
  \******************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.NewerBmprVersionAlertTask = exports.Case = void 0;

var _task = __webpack_require__(/*! core/tasks/task */ "./js/core/tasks/task.ts");

var _nativeactions = __webpack_require__(/*! editor/constants/nativeactions */ "./js/editor/constants/nativeactions.js");

var AlertVC = _interopRequireWildcard(__webpack_require__(/*! core/viewcontrollers/alertvc */ "./js/core/viewcontrollers/alertvc.ts"));

var _nativeConstants = __webpack_require__(/*! core/native-constants */ "./js/core/native-constants.js");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var Case;
exports.Case = Case;

(function (Case) {
  Case["Wireframe"] = "wireframe";
  Case["Project"] = "project";
})(Case || (exports.Case = Case = {}));

class NewerBmprVersionAlertTask extends _task.TaskBase {
  constructor() {
    super(...arguments);

    _defineProperty(this, "closeAlert", void 0);

    _defineProperty(this, "onButtonClick", () => {
      const isBrowser = this.context.host.runtime.isBrowser();

      if (isBrowser) {
        window.location.reload();
      } else {
        this.context.host.nativeAction(_nativeactions.NativeActions.OPEN_CHECK_FOR_UPDATE);
      }

      this.notifyFinished();
    });
  }

  cleanUp() {
    this.closeAlert = undefined;
    super.cleanUp();
  }

  name() {
    return "Newer BMPR Version";
  }

  doRun() {
    const word = this.params.case || Case.Project;
    const messageCommon = `The ${word} you're trying to load was created with a version of Balsamiq Wireframes more recent than what this version can understand.`;
    const messageBrowser = `You must reload this browser page to load the new version and continue working.`;
    const messageDesktop = `You must update Balsamiq for Desktop in order to continue to work on this project.`;
    const messageSupport = `If the error persists, please send us an email at [support@balsamiq.com](${_nativeConstants.ApplicationActionURISchema + _nativeConstants.ApplicationAction.OpenContactSupport}) and we'll do our best to help.`;
    const isBrowser = this.context.host.runtime.isBrowser();
    const message = `${messageCommon}\n\n${isBrowser ? messageBrowser : messageDesktop}\n\n${messageSupport}`;
    const buttonTitle = isBrowser ? "Reload" : "Close Project and Update";
    this.closeAlert = this.context.viewControllers.pushAlert({
      windowID: this.context.windowID,
      taskID: this.getTaskID(),
      title: "Ooops! Newer version detected",
      message: message,
      buttons: [{
        text: buttonTitle,
        type: AlertVC.Enums.ButtonIdentifiers.MAIN
      }],
      callback: this.onButtonClick,
      closable: false
    }).close;
  }

  canInterrupt() {
    return true;
  }

  onInterrupted() {
    if (this.closeAlert) {
      this.closeAlert();
    }

    this.notifyFinished();
  }

}

exports.NewerBmprVersionAlertTask = NewerBmprVersionAlertTask;

/***/ }),

/***/ "./js/exporter/exportertask.ts":
/*!*************************************!*\
  !*** ./js/exporter/exportertask.ts ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {

"use strict";


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
exports.ExporterTask = void 0;

var _task = __webpack_require__(/*! core/tasks/task */ "./js/core/tasks/task.ts");

var _loadprojecttask = __webpack_require__(/*! editor/tasks/loadprojecttask */ "./js/editor/tasks/loadprojecttask.ts");

var _initnativemoduletask = __webpack_require__(/*! core/tasks/initnativemoduletask */ "./js/core/tasks/initnativemoduletask.ts");

var _native = __webpack_require__(/*! core/native */ "./js/core/native.js");

var _project = __webpack_require__(/*! core/model/project */ "./js/core/model/project.js");

var _pdfgenerator = __webpack_require__(/*! core/model/pdfgenerator */ "./js/core/model/pdfgenerator.js");

var _pathlibrary = __webpack_require__(/*! core/graphics/pathlibrary */ "./js/core/graphics/pathlibrary.ts");

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

class ExporterTask extends _task.TaskBase {
  constructor() {
    super(...arguments);

    _defineProperty(this, "project", void 0);

    _defineProperty(this, "sqliteID", "");
  }

  name() {
    return "ExporterTask";
  }

  doRun() {
    this.context.host.log(`ExporterTask: convert ${this.params.bmprFileURL} into ${this.params.pdfFileURL}`);
    this.context.host.log(`ExporterTask: adding PathLibrary...`);
    (0, _pathlibrary.addPathLibrary)(this.context);
    this.context.host.log(`ExporterTask: done`);
    this.initSQLiteModule();
  }

  initSQLiteModule() {
    this.runSubtask(new _initnativemoduletask.InitNativeModuleTask({
      moduleName: _native.ModuleNames.SQLite
    }), res => {
      if (res.error) {
        this.context.host.log(`ExporterTask: error initializing SQLite module`);
        this.notifyFinished(res);
      } else {
        this.context.host.log(`ExporterTask: SQLite module initialized`);
        this.loadProject();
      }
    });
  }

  loadProject() {
    this.context.host.SQLite.makeFromFileURL(this.params.bmprFileURL, res => {
      if (res.error) {
        this.notifyFinished(res);
      } else {
        this.sqliteID = res.id;
        this.context.host.log(`ExporterTask: SQLite archive: ${this.sqliteID}`);
        const task = new _loadprojecttask.LoadProjectTask({
          sourceType: _loadprojecttask.SourceTypes.SQLITE,
          disableProgress: true,
          sqliteID: this.sqliteID,
          userRole: _project.Roles.VIEWER
        });
        this.runSubtask(task, res => {
          if (res.error) {
            this.context.host.log(`ExporterTask: error loading the project`);
            this.notifyFinished(res);
          } else {
            this.context.host.log(`ExporterTask: project loaded`);
            this.project = res.projectController.getProject();
            this.initPDFModule();
          }
        });
      }
    });
  }

  initPDFModule() {
    this.runSubtask(new _initnativemoduletask.InitNativeModuleTask({
      moduleName: _native.ModuleNames.PDF
    }), res => {
      if (res.error) {
        this.context.host.log(`ExporterTask: error initializing PDF module`);
        this.notifyFinished(res);
      } else {
        this.context.host.log(`ExporterTask: PDF module initialized`);
        this.export();
      }
    });
  }

  export() {
    const pdfGenerator = new _pdfgenerator.PdfGenerator({
      host: this.context.host,
      eventCenter: this.context.eventCenter
    });
    const mockups = this.project.getFlatMockupList();
    let mockupCount = 0;
    this.context.host.log(`ExporterTask: ${mockups.length} wireframes`);

    const pdfGeneratorCallback = () => {
      if (mockupCount == mockups.length) {
        // done
        pdfGenerator.save({
          url: this.params.pdfFileURL,
          noUI: true
        }, res => {
          if (res.error) {
            this.context.host.log(`ExporterTask: error saving PDF file`);
            this.notifyFinished(res);
          } else {
            this.context.host.log(`ExporterTask: PDF file saved`);
            this.closeProject();
          }
        });
      } else {
        // take another snapshot
        const mockup = mockups[mockupCount];
        this.context.host.log(`ExporterTask: wireframe ${mockupCount}`);
        mockupCount += 1;
        pdfGenerator.createSnapshot(mockup, mockups, pdfGeneratorCallback, this.project);
      }
    };

    pdfGenerator.start({
      size: _pdfgenerator.PDFOptionsSize.FITAPAGE,
      PaperSize: _pdfgenerator.PDFPaperSize.Letter,
      PDFOrientation: _pdfgenerator.PDFOrientation.landscape // noMargins: true,

    }, this.project.getFontFace(), pdfGeneratorCallback);
  }

  closeProject() {
    // close the project for real? ;-)
    this.context.host.log(`ExporterTask: done`);
    this.notifyFinished();
  }

}

exports.ExporterTask = ExporterTask;

/***/ }),

/***/ "./node_modules/lodash/fp.js":
/*!***********************************!*\
  !*** ./node_modules/lodash/fp.js ***!
  \***********************************/
/***/ ((module, __unused_webpack_exports, __webpack_require__) => {

var _ = (__webpack_require__(/*! ./lodash.min */ "./node_modules/lodash/lodash.min.js").runInContext)();
module.exports = __webpack_require__(/*! ./fp/_baseConvert */ "./node_modules/lodash/fp/_baseConvert.js")(_, _);


/***/ }),

/***/ "./node_modules/lodash/fp/_baseConvert.js":
/*!************************************************!*\
  !*** ./node_modules/lodash/fp/_baseConvert.js ***!
  \************************************************/
/***/ ((module, __unused_webpack_exports, __webpack_require__) => {

var mapping = __webpack_require__(/*! ./_mapping */ "./node_modules/lodash/fp/_mapping.js"),
    fallbackHolder = __webpack_require__(/*! ./placeholder */ "./node_modules/lodash/fp/placeholder.js");

/** Built-in value reference. */
var push = Array.prototype.push;

/**
 * Creates a function, with an arity of `n`, that invokes `func` with the
 * arguments it receives.
 *
 * @private
 * @param {Function} func The function to wrap.
 * @param {number} n The arity of the new function.
 * @returns {Function} Returns the new function.
 */
function baseArity(func, n) {
  return n == 2
    ? function(a, b) { return func.apply(undefined, arguments); }
    : function(a) { return func.apply(undefined, arguments); };
}

/**
 * Creates a function that invokes `func`, with up to `n` arguments, ignoring
 * any additional arguments.
 *
 * @private
 * @param {Function} func The function to cap arguments for.
 * @param {number} n The arity cap.
 * @returns {Function} Returns the new function.
 */
function baseAry(func, n) {
  return n == 2
    ? function(a, b) { return func(a, b); }
    : function(a) { return func(a); };
}

/**
 * Creates a clone of `array`.
 *
 * @private
 * @param {Array} array The array to clone.
 * @returns {Array} Returns the cloned array.
 */
function cloneArray(array) {
  var length = array ? array.length : 0,
      result = Array(length);

  while (length--) {
    result[length] = array[length];
  }
  return result;
}

/**
 * Creates a function that clones a given object using the assignment `func`.
 *
 * @private
 * @param {Function} func The assignment function.
 * @returns {Function} Returns the new cloner function.
 */
function createCloner(func) {
  return function(object) {
    return func({}, object);
  };
}

/**
 * A specialized version of `_.spread` which flattens the spread array into
 * the arguments of the invoked `func`.
 *
 * @private
 * @param {Function} func The function to spread arguments over.
 * @param {number} start The start position of the spread.
 * @returns {Function} Returns the new function.
 */
function flatSpread(func, start) {
  return function() {
    var length = arguments.length,
        lastIndex = length - 1,
        args = Array(length);

    while (length--) {
      args[length] = arguments[length];
    }
    var array = args[start],
        otherArgs = args.slice(0, start);

    if (array) {
      push.apply(otherArgs, array);
    }
    if (start != lastIndex) {
      push.apply(otherArgs, args.slice(start + 1));
    }
    return func.apply(this, otherArgs);
  };
}

/**
 * Creates a function that wraps `func` and uses `cloner` to clone the first
 * argument it receives.
 *
 * @private
 * @param {Function} func The function to wrap.
 * @param {Function} cloner The function to clone arguments.
 * @returns {Function} Returns the new immutable function.
 */
function wrapImmutable(func, cloner) {
  return function() {
    var length = arguments.length;
    if (!length) {
      return;
    }
    var args = Array(length);
    while (length--) {
      args[length] = arguments[length];
    }
    var result = args[0] = cloner.apply(undefined, args);
    func.apply(undefined, args);
    return result;
  };
}

/**
 * The base implementation of `convert` which accepts a `util` object of methods
 * required to perform conversions.
 *
 * @param {Object} util The util object.
 * @param {string} name The name of the function to convert.
 * @param {Function} func The function to convert.
 * @param {Object} [options] The options object.
 * @param {boolean} [options.cap=true] Specify capping iteratee arguments.
 * @param {boolean} [options.curry=true] Specify currying.
 * @param {boolean} [options.fixed=true] Specify fixed arity.
 * @param {boolean} [options.immutable=true] Specify immutable operations.
 * @param {boolean} [options.rearg=true] Specify rearranging arguments.
 * @returns {Function|Object} Returns the converted function or object.
 */
function baseConvert(util, name, func, options) {
  var isLib = typeof name == 'function',
      isObj = name === Object(name);

  if (isObj) {
    options = func;
    func = name;
    name = undefined;
  }
  if (func == null) {
    throw new TypeError;
  }
  options || (options = {});

  var config = {
    'cap': 'cap' in options ? options.cap : true,
    'curry': 'curry' in options ? options.curry : true,
    'fixed': 'fixed' in options ? options.fixed : true,
    'immutable': 'immutable' in options ? options.immutable : true,
    'rearg': 'rearg' in options ? options.rearg : true
  };

  var defaultHolder = isLib ? func : fallbackHolder,
      forceCurry = ('curry' in options) && options.curry,
      forceFixed = ('fixed' in options) && options.fixed,
      forceRearg = ('rearg' in options) && options.rearg,
      pristine = isLib ? func.runInContext() : undefined;

  var helpers = isLib ? func : {
    'ary': util.ary,
    'assign': util.assign,
    'clone': util.clone,
    'curry': util.curry,
    'forEach': util.forEach,
    'isArray': util.isArray,
    'isError': util.isError,
    'isFunction': util.isFunction,
    'isWeakMap': util.isWeakMap,
    'iteratee': util.iteratee,
    'keys': util.keys,
    'rearg': util.rearg,
    'toInteger': util.toInteger,
    'toPath': util.toPath
  };

  var ary = helpers.ary,
      assign = helpers.assign,
      clone = helpers.clone,
      curry = helpers.curry,
      each = helpers.forEach,
      isArray = helpers.isArray,
      isError = helpers.isError,
      isFunction = helpers.isFunction,
      isWeakMap = helpers.isWeakMap,
      keys = helpers.keys,
      rearg = helpers.rearg,
      toInteger = helpers.toInteger,
      toPath = helpers.toPath;

  var aryMethodKeys = keys(mapping.aryMethod);

  var wrappers = {
    'castArray': function(castArray) {
      return function() {
        var value = arguments[0];
        return isArray(value)
          ? castArray(cloneArray(value))
          : castArray.apply(undefined, arguments);
      };
    },
    'iteratee': function(iteratee) {
      return function() {
        var func = arguments[0],
            arity = arguments[1],
            result = iteratee(func, arity),
            length = result.length;

        if (config.cap && typeof arity == 'number') {
          arity = arity > 2 ? (arity - 2) : 1;
          return (length && length <= arity) ? result : baseAry(result, arity);
        }
        return result;
      };
    },
    'mixin': function(mixin) {
      return function(source) {
        var func = this;
        if (!isFunction(func)) {
          return mixin(func, Object(source));
        }
        var pairs = [];
        each(keys(source), function(key) {
          if (isFunction(source[key])) {
            pairs.push([key, func.prototype[key]]);
          }
        });

        mixin(func, Object(source));

        each(pairs, function(pair) {
          var value = pair[1];
          if (isFunction(value)) {
            func.prototype[pair[0]] = value;
          } else {
            delete func.prototype[pair[0]];
          }
        });
        return func;
      };
    },
    'nthArg': function(nthArg) {
      return function(n) {
        var arity = n < 0 ? 1 : (toInteger(n) + 1);
        return curry(nthArg(n), arity);
      };
    },
    'rearg': function(rearg) {
      return function(func, indexes) {
        var arity = indexes ? indexes.length : 0;
        return curry(rearg(func, indexes), arity);
      };
    },
    'runInContext': function(runInContext) {
      return function(context) {
        return baseConvert(util, runInContext(context), options);
      };
    }
  };

  /*--------------------------------------------------------------------------*/

  /**
   * Casts `func` to a function with an arity capped iteratee if needed.
   *
   * @private
   * @param {string} name The name of the function to inspect.
   * @param {Function} func The function to inspect.
   * @returns {Function} Returns the cast function.
   */
  function castCap(name, func) {
    if (config.cap) {
      var indexes = mapping.iterateeRearg[name];
      if (indexes) {
        return iterateeRearg(func, indexes);
      }
      var n = !isLib && mapping.iterateeAry[name];
      if (n) {
        return iterateeAry(func, n);
      }
    }
    return func;
  }

  /**
   * Casts `func` to a curried function if needed.
   *
   * @private
   * @param {string} name The name of the function to inspect.
   * @param {Function} func The function to inspect.
   * @param {number} n The arity of `func`.
   * @returns {Function} Returns the cast function.
   */
  function castCurry(name, func, n) {
    return (forceCurry || (config.curry && n > 1))
      ? curry(func, n)
      : func;
  }

  /**
   * Casts `func` to a fixed arity function if needed.
   *
   * @private
   * @param {string} name The name of the function to inspect.
   * @param {Function} func The function to inspect.
   * @param {number} n The arity cap.
   * @returns {Function} Returns the cast function.
   */
  function castFixed(name, func, n) {
    if (config.fixed && (forceFixed || !mapping.skipFixed[name])) {
      var data = mapping.methodSpread[name],
          start = data && data.start;

      return start  === undefined ? ary(func, n) : flatSpread(func, start);
    }
    return func;
  }

  /**
   * Casts `func` to an rearged function if needed.
   *
   * @private
   * @param {string} name The name of the function to inspect.
   * @param {Function} func The function to inspect.
   * @param {number} n The arity of `func`.
   * @returns {Function} Returns the cast function.
   */
  function castRearg(name, func, n) {
    return (config.rearg && n > 1 && (forceRearg || !mapping.skipRearg[name]))
      ? rearg(func, mapping.methodRearg[name] || mapping.aryRearg[n])
      : func;
  }

  /**
   * Creates a clone of `object` by `path`.
   *
   * @private
   * @param {Object} object The object to clone.
   * @param {Array|string} path The path to clone by.
   * @returns {Object} Returns the cloned object.
   */
  function cloneByPath(object, path) {
    path = toPath(path);

    var index = -1,
        length = path.length,
        lastIndex = length - 1,
        result = clone(Object(object)),
        nested = result;

    while (nested != null && ++index < length) {
      var key = path[index],
          value = nested[key];

      if (value != null &&
          !(isFunction(value) || isError(value) || isWeakMap(value))) {
        nested[key] = clone(index == lastIndex ? value : Object(value));
      }
      nested = nested[key];
    }
    return result;
  }

  /**
   * Converts `lodash` to an immutable auto-curried iteratee-first data-last
   * version with conversion `options` applied.
   *
   * @param {Object} [options] The options object. See `baseConvert` for more details.
   * @returns {Function} Returns the converted `lodash`.
   */
  function convertLib(options) {
    return _.runInContext.convert(options)(undefined);
  }

  /**
   * Create a converter function for `func` of `name`.
   *
   * @param {string} name The name of the function to convert.
   * @param {Function} func The function to convert.
   * @returns {Function} Returns the new converter function.
   */
  function createConverter(name, func) {
    var realName = mapping.aliasToReal[name] || name,
        methodName = mapping.remap[realName] || realName,
        oldOptions = options;

    return function(options) {
      var newUtil = isLib ? pristine : helpers,
          newFunc = isLib ? pristine[methodName] : func,
          newOptions = assign(assign({}, oldOptions), options);

      return baseConvert(newUtil, realName, newFunc, newOptions);
    };
  }

  /**
   * Creates a function that wraps `func` to invoke its iteratee, with up to `n`
   * arguments, ignoring any additional arguments.
   *
   * @private
   * @param {Function} func The function to cap iteratee arguments for.
   * @param {number} n The arity cap.
   * @returns {Function} Returns the new function.
   */
  function iterateeAry(func, n) {
    return overArg(func, function(func) {
      return typeof func == 'function' ? baseAry(func, n) : func;
    });
  }

  /**
   * Creates a function that wraps `func` to invoke its iteratee with arguments
   * arranged according to the specified `indexes` where the argument value at
   * the first index is provided as the first argument, the argument value at
   * the second index is provided as the second argument, and so on.
   *
   * @private
   * @param {Function} func The function to rearrange iteratee arguments for.
   * @param {number[]} indexes The arranged argument indexes.
   * @returns {Function} Returns the new function.
   */
  function iterateeRearg(func, indexes) {
    return overArg(func, function(func) {
      var n = indexes.length;
      return baseArity(rearg(baseAry(func, n), indexes), n);
    });
  }

  /**
   * Creates a function that invokes `func` with its first argument transformed.
   *
   * @private
   * @param {Function} func The function to wrap.
   * @param {Function} transform The argument transform.
   * @returns {Function} Returns the new function.
   */
  function overArg(func, transform) {
    return function() {
      var length = arguments.length;
      if (!length) {
        return func();
      }
      var args = Array(length);
      while (length--) {
        args[length] = arguments[length];
      }
      var index = config.rearg ? 0 : (length - 1);
      args[index] = transform(args[index]);
      return func.apply(undefined, args);
    };
  }

  /**
   * Creates a function that wraps `func` and applys the conversions
   * rules by `name`.
   *
   * @private
   * @param {string} name The name of the function to wrap.
   * @param {Function} func The function to wrap.
   * @returns {Function} Returns the converted function.
   */
  function wrap(name, func, placeholder) {
    var result,
        realName = mapping.aliasToReal[name] || name,
        wrapped = func,
        wrapper = wrappers[realName];

    if (wrapper) {
      wrapped = wrapper(func);
    }
    else if (config.immutable) {
      if (mapping.mutate.array[realName]) {
        wrapped = wrapImmutable(func, cloneArray);
      }
      else if (mapping.mutate.object[realName]) {
        wrapped = wrapImmutable(func, createCloner(func));
      }
      else if (mapping.mutate.set[realName]) {
        wrapped = wrapImmutable(func, cloneByPath);
      }
    }
    each(aryMethodKeys, function(aryKey) {
      each(mapping.aryMethod[aryKey], function(otherName) {
        if (realName == otherName) {
          var data = mapping.methodSpread[realName],
              afterRearg = data && data.afterRearg;

          result = afterRearg
            ? castFixed(realName, castRearg(realName, wrapped, aryKey), aryKey)
            : castRearg(realName, castFixed(realName, wrapped, aryKey), aryKey);

          result = castCap(realName, result);
          result = castCurry(realName, result, aryKey);
          return false;
        }
      });
      return !result;
    });

    result || (result = wrapped);
    if (result == func) {
      result = forceCurry ? curry(result, 1) : function() {
        return func.apply(this, arguments);
      };
    }
    result.convert = createConverter(realName, func);
    result.placeholder = func.placeholder = placeholder;

    return result;
  }

  /*--------------------------------------------------------------------------*/

  if (!isObj) {
    return wrap(name, func, defaultHolder);
  }
  var _ = func;

  // Convert methods by ary cap.
  var pairs = [];
  each(aryMethodKeys, function(aryKey) {
    each(mapping.aryMethod[aryKey], function(key) {
      var func = _[mapping.remap[key] || key];
      if (func) {
        pairs.push([key, wrap(key, func, _)]);
      }
    });
  });

  // Convert remaining methods.
  each(keys(_), function(key) {
    var func = _[key];
    if (typeof func == 'function') {
      var length = pairs.length;
      while (length--) {
        if (pairs[length][0] == key) {
          return;
        }
      }
      func.convert = createConverter(key, func);
      pairs.push([key, func]);
    }
  });

  // Assign to `_` leaving `_.prototype` unchanged to allow chaining.
  each(pairs, function(pair) {
    _[pair[0]] = pair[1];
  });

  _.convert = convertLib;
  _.placeholder = _;

  // Assign aliases.
  each(keys(_), function(key) {
    each(mapping.realToAlias[key] || [], function(alias) {
      _[alias] = _[key];
    });
  });

  return _;
}

module.exports = baseConvert;


/***/ }),

/***/ "./node_modules/lodash/fp/_mapping.js":
/*!********************************************!*\
  !*** ./node_modules/lodash/fp/_mapping.js ***!
  \********************************************/
/***/ ((__unused_webpack_module, exports) => {

/** Used to map aliases to their real names. */
exports.aliasToReal = {

  // Lodash aliases.
  'each': 'forEach',
  'eachRight': 'forEachRight',
  'entries': 'toPairs',
  'entriesIn': 'toPairsIn',
  'extend': 'assignIn',
  'extendAll': 'assignInAll',
  'extendAllWith': 'assignInAllWith',
  'extendWith': 'assignInWith',
  'first': 'head',

  // Methods that are curried variants of others.
  'conforms': 'conformsTo',
  'matches': 'isMatch',
  'property': 'get',

  // Ramda aliases.
  '__': 'placeholder',
  'F': 'stubFalse',
  'T': 'stubTrue',
  'all': 'every',
  'allPass': 'overEvery',
  'always': 'constant',
  'any': 'some',
  'anyPass': 'overSome',
  'apply': 'spread',
  'assoc': 'set',
  'assocPath': 'set',
  'complement': 'negate',
  'compose': 'flowRight',
  'contains': 'includes',
  'dissoc': 'unset',
  'dissocPath': 'unset',
  'dropLast': 'dropRight',
  'dropLastWhile': 'dropRightWhile',
  'equals': 'isEqual',
  'identical': 'eq',
  'indexBy': 'keyBy',
  'init': 'initial',
  'invertObj': 'invert',
  'juxt': 'over',
  'omitAll': 'omit',
  'nAry': 'ary',
  'path': 'get',
  'pathEq': 'matchesProperty',
  'pathOr': 'getOr',
  'paths': 'at',
  'pickAll': 'pick',
  'pipe': 'flow',
  'pluck': 'map',
  'prop': 'get',
  'propEq': 'matchesProperty',
  'propOr': 'getOr',
  'props': 'at',
  'symmetricDifference': 'xor',
  'symmetricDifferenceBy': 'xorBy',
  'symmetricDifferenceWith': 'xorWith',
  'takeLast': 'takeRight',
  'takeLastWhile': 'takeRightWhile',
  'unapply': 'rest',
  'unnest': 'flatten',
  'useWith': 'overArgs',
  'where': 'conformsTo',
  'whereEq': 'isMatch',
  'zipObj': 'zipObject'
};

/** Used to map ary to method names. */
exports.aryMethod = {
  '1': [
    'assignAll', 'assignInAll', 'attempt', 'castArray', 'ceil', 'create',
    'curry', 'curryRight', 'defaultsAll', 'defaultsDeepAll', 'floor', 'flow',
    'flowRight', 'fromPairs', 'invert', 'iteratee', 'memoize', 'method', 'mergeAll',
    'methodOf', 'mixin', 'nthArg', 'over', 'overEvery', 'overSome','rest', 'reverse',
    'round', 'runInContext', 'spread', 'template', 'trim', 'trimEnd', 'trimStart',
    'uniqueId', 'words', 'zipAll'
  ],
  '2': [
    'add', 'after', 'ary', 'assign', 'assignAllWith', 'assignIn', 'assignInAllWith',
    'at', 'before', 'bind', 'bindAll', 'bindKey', 'chunk', 'cloneDeepWith',
    'cloneWith', 'concat', 'conformsTo', 'countBy', 'curryN', 'curryRightN',
    'debounce', 'defaults', 'defaultsDeep', 'defaultTo', 'delay', 'difference',
    'divide', 'drop', 'dropRight', 'dropRightWhile', 'dropWhile', 'endsWith', 'eq',
    'every', 'filter', 'find', 'findIndex', 'findKey', 'findLast', 'findLastIndex',
    'findLastKey', 'flatMap', 'flatMapDeep', 'flattenDepth', 'forEach',
    'forEachRight', 'forIn', 'forInRight', 'forOwn', 'forOwnRight', 'get',
    'groupBy', 'gt', 'gte', 'has', 'hasIn', 'includes', 'indexOf', 'intersection',
    'invertBy', 'invoke', 'invokeMap', 'isEqual', 'isMatch', 'join', 'keyBy',
    'lastIndexOf', 'lt', 'lte', 'map', 'mapKeys', 'mapValues', 'matchesProperty',
    'maxBy', 'meanBy', 'merge', 'mergeAllWith', 'minBy', 'multiply', 'nth', 'omit',
    'omitBy', 'overArgs', 'pad', 'padEnd', 'padStart', 'parseInt', 'partial',
    'partialRight', 'partition', 'pick', 'pickBy', 'propertyOf', 'pull', 'pullAll',
    'pullAt', 'random', 'range', 'rangeRight', 'rearg', 'reject', 'remove',
    'repeat', 'restFrom', 'result', 'sampleSize', 'some', 'sortBy', 'sortedIndex',
    'sortedIndexOf', 'sortedLastIndex', 'sortedLastIndexOf', 'sortedUniqBy',
    'split', 'spreadFrom', 'startsWith', 'subtract', 'sumBy', 'take', 'takeRight',
    'takeRightWhile', 'takeWhile', 'tap', 'throttle', 'thru', 'times', 'trimChars',
    'trimCharsEnd', 'trimCharsStart', 'truncate', 'union', 'uniqBy', 'uniqWith',
    'unset', 'unzipWith', 'without', 'wrap', 'xor', 'zip', 'zipObject',
    'zipObjectDeep'
  ],
  '3': [
    'assignInWith', 'assignWith', 'clamp', 'differenceBy', 'differenceWith',
    'findFrom', 'findIndexFrom', 'findLastFrom', 'findLastIndexFrom', 'getOr',
    'includesFrom', 'indexOfFrom', 'inRange', 'intersectionBy', 'intersectionWith',
    'invokeArgs', 'invokeArgsMap', 'isEqualWith', 'isMatchWith', 'flatMapDepth',
    'lastIndexOfFrom', 'mergeWith', 'orderBy', 'padChars', 'padCharsEnd',
    'padCharsStart', 'pullAllBy', 'pullAllWith', 'rangeStep', 'rangeStepRight',
    'reduce', 'reduceRight', 'replace', 'set', 'slice', 'sortedIndexBy',
    'sortedLastIndexBy', 'transform', 'unionBy', 'unionWith', 'update', 'xorBy',
    'xorWith', 'zipWith'
  ],
  '4': [
    'fill', 'setWith', 'updateWith'
  ]
};

/** Used to map ary to rearg configs. */
exports.aryRearg = {
  '2': [1, 0],
  '3': [2, 0, 1],
  '4': [3, 2, 0, 1]
};

/** Used to map method names to their iteratee ary. */
exports.iterateeAry = {
  'dropRightWhile': 1,
  'dropWhile': 1,
  'every': 1,
  'filter': 1,
  'find': 1,
  'findFrom': 1,
  'findIndex': 1,
  'findIndexFrom': 1,
  'findKey': 1,
  'findLast': 1,
  'findLastFrom': 1,
  'findLastIndex': 1,
  'findLastIndexFrom': 1,
  'findLastKey': 1,
  'flatMap': 1,
  'flatMapDeep': 1,
  'flatMapDepth': 1,
  'forEach': 1,
  'forEachRight': 1,
  'forIn': 1,
  'forInRight': 1,
  'forOwn': 1,
  'forOwnRight': 1,
  'map': 1,
  'mapKeys': 1,
  'mapValues': 1,
  'partition': 1,
  'reduce': 2,
  'reduceRight': 2,
  'reject': 1,
  'remove': 1,
  'some': 1,
  'takeRightWhile': 1,
  'takeWhile': 1,
  'times': 1,
  'transform': 2
};

/** Used to map method names to iteratee rearg configs. */
exports.iterateeRearg = {
  'mapKeys': [1],
  'reduceRight': [1, 0]
};

/** Used to map method names to rearg configs. */
exports.methodRearg = {
  'assignInAllWith': [1, 0],
  'assignInWith': [1, 2, 0],
  'assignAllWith': [1, 0],
  'assignWith': [1, 2, 0],
  'differenceBy': [1, 2, 0],
  'differenceWith': [1, 2, 0],
  'getOr': [2, 1, 0],
  'intersectionBy': [1, 2, 0],
  'intersectionWith': [1, 2, 0],
  'isEqualWith': [1, 2, 0],
  'isMatchWith': [2, 1, 0],
  'mergeAllWith': [1, 0],
  'mergeWith': [1, 2, 0],
  'padChars': [2, 1, 0],
  'padCharsEnd': [2, 1, 0],
  'padCharsStart': [2, 1, 0],
  'pullAllBy': [2, 1, 0],
  'pullAllWith': [2, 1, 0],
  'rangeStep': [1, 2, 0],
  'rangeStepRight': [1, 2, 0],
  'setWith': [3, 1, 2, 0],
  'sortedIndexBy': [2, 1, 0],
  'sortedLastIndexBy': [2, 1, 0],
  'unionBy': [1, 2, 0],
  'unionWith': [1, 2, 0],
  'updateWith': [3, 1, 2, 0],
  'xorBy': [1, 2, 0],
  'xorWith': [1, 2, 0],
  'zipWith': [1, 2, 0]
};

/** Used to map method names to spread configs. */
exports.methodSpread = {
  'assignAll': { 'start': 0 },
  'assignAllWith': { 'start': 0 },
  'assignInAll': { 'start': 0 },
  'assignInAllWith': { 'start': 0 },
  'defaultsAll': { 'start': 0 },
  'defaultsDeepAll': { 'start': 0 },
  'invokeArgs': { 'start': 2 },
  'invokeArgsMap': { 'start': 2 },
  'mergeAll': { 'start': 0 },
  'mergeAllWith': { 'start': 0 },
  'partial': { 'start': 1 },
  'partialRight': { 'start': 1 },
  'without': { 'start': 1 },
  'zipAll': { 'start': 0 }
};

/** Used to identify methods which mutate arrays or objects. */
exports.mutate = {
  'array': {
    'fill': true,
    'pull': true,
    'pullAll': true,
    'pullAllBy': true,
    'pullAllWith': true,
    'pullAt': true,
    'remove': true,
    'reverse': true
  },
  'object': {
    'assign': true,
    'assignAll': true,
    'assignAllWith': true,
    'assignIn': true,
    'assignInAll': true,
    'assignInAllWith': true,
    'assignInWith': true,
    'assignWith': true,
    'defaults': true,
    'defaultsAll': true,
    'defaultsDeep': true,
    'defaultsDeepAll': true,
    'merge': true,
    'mergeAll': true,
    'mergeAllWith': true,
    'mergeWith': true,
  },
  'set': {
    'set': true,
    'setWith': true,
    'unset': true,
    'update': true,
    'updateWith': true
  }
};

/** Used to map real names to their aliases. */
exports.realToAlias = (function() {
  var hasOwnProperty = Object.prototype.hasOwnProperty,
      object = exports.aliasToReal,
      result = {};

  for (var key in object) {
    var value = object[key];
    if (hasOwnProperty.call(result, value)) {
      result[value].push(key);
    } else {
      result[value] = [key];
    }
  }
  return result;
}());

/** Used to map method names to other names. */
exports.remap = {
  'assignAll': 'assign',
  'assignAllWith': 'assignWith',
  'assignInAll': 'assignIn',
  'assignInAllWith': 'assignInWith',
  'curryN': 'curry',
  'curryRightN': 'curryRight',
  'defaultsAll': 'defaults',
  'defaultsDeepAll': 'defaultsDeep',
  'findFrom': 'find',
  'findIndexFrom': 'findIndex',
  'findLastFrom': 'findLast',
  'findLastIndexFrom': 'findLastIndex',
  'getOr': 'get',
  'includesFrom': 'includes',
  'indexOfFrom': 'indexOf',
  'invokeArgs': 'invoke',
  'invokeArgsMap': 'invokeMap',
  'lastIndexOfFrom': 'lastIndexOf',
  'mergeAll': 'merge',
  'mergeAllWith': 'mergeWith',
  'padChars': 'pad',
  'padCharsEnd': 'padEnd',
  'padCharsStart': 'padStart',
  'propertyOf': 'get',
  'rangeStep': 'range',
  'rangeStepRight': 'rangeRight',
  'restFrom': 'rest',
  'spreadFrom': 'spread',
  'trimChars': 'trim',
  'trimCharsEnd': 'trimEnd',
  'trimCharsStart': 'trimStart',
  'zipAll': 'zip'
};

/** Used to track methods that skip fixing their arity. */
exports.skipFixed = {
  'castArray': true,
  'flow': true,
  'flowRight': true,
  'iteratee': true,
  'mixin': true,
  'rearg': true,
  'runInContext': true
};

/** Used to track methods that skip rearranging arguments. */
exports.skipRearg = {
  'add': true,
  'assign': true,
  'assignIn': true,
  'bind': true,
  'bindKey': true,
  'concat': true,
  'difference': true,
  'divide': true,
  'eq': true,
  'gt': true,
  'gte': true,
  'isEqual': true,
  'lt': true,
  'lte': true,
  'matchesProperty': true,
  'merge': true,
  'multiply': true,
  'overArgs': true,
  'partial': true,
  'partialRight': true,
  'propertyOf': true,
  'random': true,
  'range': true,
  'rangeRight': true,
  'subtract': true,
  'zip': true,
  'zipObject': true,
  'zipObjectDeep': true
};


/***/ }),

/***/ "./node_modules/lodash/fp/placeholder.js":
/*!***********************************************!*\
  !*** ./node_modules/lodash/fp/placeholder.js ***!
  \***********************************************/
/***/ ((module) => {

/**
 * The default argument placeholder value for methods.
 *
 * @type {Object}
 */
module.exports = {};


/***/ }),

/***/ "./node_modules/lodash/lodash.js":
/*!***************************************!*\
  !*** ./node_modules/lodash/lodash.js ***!
  \***************************************/
/***/ (function(module, exports, __webpack_require__) {

/* module decorator */ module = __webpack_require__.nmd(module);
var __WEBPACK_AMD_DEFINE_RESULT__;/**
 * @license
 * Lodash <https://lodash.com/>
 * Copyright OpenJS Foundation and other contributors <https://openjsf.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */
;(function() {

  /** Used as a safe reference for `undefined` in pre-ES5 environments. */
  var undefined;

  /** Used as the semantic version number. */
  var VERSION = '4.17.21';

  /** Used as the size to enable large array optimizations. */
  var LARGE_ARRAY_SIZE = 200;

  /** Error message constants. */
  var CORE_ERROR_TEXT = 'Unsupported core-js use. Try https://npms.io/search?q=ponyfill.',
      FUNC_ERROR_TEXT = 'Expected a function',
      INVALID_TEMPL_VAR_ERROR_TEXT = 'Invalid `variable` option passed into `_.template`';

  /** Used to stand-in for `undefined` hash values. */
  var HASH_UNDEFINED = '__lodash_hash_undefined__';

  /** Used as the maximum memoize cache size. */
  var MAX_MEMOIZE_SIZE = 500;

  /** Used as the internal argument placeholder. */
  var PLACEHOLDER = '__lodash_placeholder__';

  /** Used to compose bitmasks for cloning. */
  var CLONE_DEEP_FLAG = 1,
      CLONE_FLAT_FLAG = 2,
      CLONE_SYMBOLS_FLAG = 4;

  /** Used to compose bitmasks for value comparisons. */
  var COMPARE_PARTIAL_FLAG = 1,
      COMPARE_UNORDERED_FLAG = 2;

  /** Used to compose bitmasks for function metadata. */
  var WRAP_BIND_FLAG = 1,
      WRAP_BIND_KEY_FLAG = 2,
      WRAP_CURRY_BOUND_FLAG = 4,
      WRAP_CURRY_FLAG = 8,
      WRAP_CURRY_RIGHT_FLAG = 16,
      WRAP_PARTIAL_FLAG = 32,
      WRAP_PARTIAL_RIGHT_FLAG = 64,
      WRAP_ARY_FLAG = 128,
      WRAP_REARG_FLAG = 256,
      WRAP_FLIP_FLAG = 512;

  /** Used as default options for `_.truncate`. */
  var DEFAULT_TRUNC_LENGTH = 30,
      DEFAULT_TRUNC_OMISSION = '...';

  /** Used to detect hot functions by number of calls within a span of milliseconds. */
  var HOT_COUNT = 800,
      HOT_SPAN = 16;

  /** Used to indicate the type of lazy iteratees. */
  var LAZY_FILTER_FLAG = 1,
      LAZY_MAP_FLAG = 2,
      LAZY_WHILE_FLAG = 3;

  /** Used as references for various `Number` constants. */
  var INFINITY = 1 / 0,
      MAX_SAFE_INTEGER = 9007199254740991,
      MAX_INTEGER = 1.7976931348623157e+308,
      NAN = 0 / 0;

  /** Used as references for the maximum length and index of an array. */
  var MAX_ARRAY_LENGTH = 4294967295,
      MAX_ARRAY_INDEX = MAX_ARRAY_LENGTH - 1,
      HALF_MAX_ARRAY_LENGTH = MAX_ARRAY_LENGTH >>> 1;

  /** Used to associate wrap methods with their bit flags. */
  var wrapFlags = [
    ['ary', WRAP_ARY_FLAG],
    ['bind', WRAP_BIND_FLAG],
    ['bindKey', WRAP_BIND_KEY_FLAG],
    ['curry', WRAP_CURRY_FLAG],
    ['curryRight', WRAP_CURRY_RIGHT_FLAG],
    ['flip', WRAP_FLIP_FLAG],
    ['partial', WRAP_PARTIAL_FLAG],
    ['partialRight', WRAP_PARTIAL_RIGHT_FLAG],
    ['rearg', WRAP_REARG_FLAG]
  ];

  /** `Object#toString` result references. */
  var argsTag = '[object Arguments]',
      arrayTag = '[object Array]',
      asyncTag = '[object AsyncFunction]',
      boolTag = '[object Boolean]',
      dateTag = '[object Date]',
      domExcTag = '[object DOMException]',
      errorTag = '[object Error]',
      funcTag = '[object Function]',
      genTag = '[object GeneratorFunction]',
      mapTag = '[object Map]',
      numberTag = '[object Number]',
      nullTag = '[object Null]',
      objectTag = '[object Object]',
      promiseTag = '[object Promise]',
      proxyTag = '[object Proxy]',
      regexpTag = '[object RegExp]',
      setTag = '[object Set]',
      stringTag = '[object String]',
      symbolTag = '[object Symbol]',
      undefinedTag = '[object Undefined]',
      weakMapTag = '[object WeakMap]',
      weakSetTag = '[object WeakSet]';

  var arrayBufferTag = '[object ArrayBuffer]',
      dataViewTag = '[object DataView]',
      float32Tag = '[object Float32Array]',
      float64Tag = '[object Float64Array]',
      int8Tag = '[object Int8Array]',
      int16Tag = '[object Int16Array]',
      int32Tag = '[object Int32Array]',
      uint8Tag = '[object Uint8Array]',
      uint8ClampedTag = '[object Uint8ClampedArray]',
      uint16Tag = '[object Uint16Array]',
      uint32Tag = '[object Uint32Array]';

  /** Used to match empty string literals in compiled template source. */
  var reEmptyStringLeading = /\b__p \+= '';/g,
      reEmptyStringMiddle = /\b(__p \+=) '' \+/g,
      reEmptyStringTrailing = /(__e\(.*?\)|\b__t\)) \+\n'';/g;

  /** Used to match HTML entities and HTML characters. */
  var reEscapedHtml = /&(?:amp|lt|gt|quot|#39);/g,
      reUnescapedHtml = /[&<>"']/g,
      reHasEscapedHtml = RegExp(reEscapedHtml.source),
      reHasUnescapedHtml = RegExp(reUnescapedHtml.source);

  /** Used to match template delimiters. */
  var reEscape = /<%-([\s\S]+?)%>/g,
      reEvaluate = /<%([\s\S]+?)%>/g,
      reInterpolate = /<%=([\s\S]+?)%>/g;

  /** Used to match property names within property paths. */
  var reIsDeepProp = /\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,
      reIsPlainProp = /^\w*$/,
      rePropName = /[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g;

  /**
   * Used to match `RegExp`
   * [syntax characters](http://ecma-international.org/ecma-262/7.0/#sec-patterns).
   */
  var reRegExpChar = /[\\^$.*+?()[\]{}|]/g,
      reHasRegExpChar = RegExp(reRegExpChar.source);

  /** Used to match leading whitespace. */
  var reTrimStart = /^\s+/;

  /** Used to match a single whitespace character. */
  var reWhitespace = /\s/;

  /** Used to match wrap detail comments. */
  var reWrapComment = /\{(?:\n\/\* \[wrapped with .+\] \*\/)?\n?/,
      reWrapDetails = /\{\n\/\* \[wrapped with (.+)\] \*/,
      reSplitDetails = /,? & /;

  /** Used to match words composed of alphanumeric characters. */
  var reAsciiWord = /[^\x00-\x2f\x3a-\x40\x5b-\x60\x7b-\x7f]+/g;

  /**
   * Used to validate the `validate` option in `_.template` variable.
   *
   * Forbids characters which could potentially change the meaning of the function argument definition:
   * - "()," (modification of function parameters)
   * - "=" (default value)
   * - "[]{}" (destructuring of function parameters)
   * - "/" (beginning of a comment)
   * - whitespace
   */
  var reForbiddenIdentifierChars = /[()=,{}\[\]\/\s]/;

  /** Used to match backslashes in property paths. */
  var reEscapeChar = /\\(\\)?/g;

  /**
   * Used to match
   * [ES template delimiters](http://ecma-international.org/ecma-262/7.0/#sec-template-literal-lexical-components).
   */
  var reEsTemplate = /\$\{([^\\}]*(?:\\.[^\\}]*)*)\}/g;

  /** Used to match `RegExp` flags from their coerced string values. */
  var reFlags = /\w*$/;

  /** Used to detect bad signed hexadecimal string values. */
  var reIsBadHex = /^[-+]0x[0-9a-f]+$/i;

  /** Used to detect binary string values. */
  var reIsBinary = /^0b[01]+$/i;

  /** Used to detect host constructors (Safari). */
  var reIsHostCtor = /^\[object .+?Constructor\]$/;

  /** Used to detect octal string values. */
  var reIsOctal = /^0o[0-7]+$/i;

  /** Used to detect unsigned integer values. */
  var reIsUint = /^(?:0|[1-9]\d*)$/;

  /** Used to match Latin Unicode letters (excluding mathematical operators). */
  var reLatin = /[\xc0-\xd6\xd8-\xf6\xf8-\xff\u0100-\u017f]/g;

  /** Used to ensure capturing order of template delimiters. */
  var reNoMatch = /($^)/;

  /** Used to match unescaped characters in compiled string literals. */
  var reUnescapedString = /['\n\r\u2028\u2029\\]/g;

  /** Used to compose unicode character classes. */
  var rsAstralRange = '\\ud800-\\udfff',
      rsComboMarksRange = '\\u0300-\\u036f',
      reComboHalfMarksRange = '\\ufe20-\\ufe2f',
      rsComboSymbolsRange = '\\u20d0-\\u20ff',
      rsComboRange = rsComboMarksRange + reComboHalfMarksRange + rsComboSymbolsRange,
      rsDingbatRange = '\\u2700-\\u27bf',
      rsLowerRange = 'a-z\\xdf-\\xf6\\xf8-\\xff',
      rsMathOpRange = '\\xac\\xb1\\xd7\\xf7',
      rsNonCharRange = '\\x00-\\x2f\\x3a-\\x40\\x5b-\\x60\\x7b-\\xbf',
      rsPunctuationRange = '\\u2000-\\u206f',
      rsSpaceRange = ' \\t\\x0b\\f\\xa0\\ufeff\\n\\r\\u2028\\u2029\\u1680\\u180e\\u2000\\u2001\\u2002\\u2003\\u2004\\u2005\\u2006\\u2007\\u2008\\u2009\\u200a\\u202f\\u205f\\u3000',
      rsUpperRange = 'A-Z\\xc0-\\xd6\\xd8-\\xde',
      rsVarRange = '\\ufe0e\\ufe0f',
      rsBreakRange = rsMathOpRange + rsNonCharRange + rsPunctuationRange + rsSpaceRange;

  /** Used to compose unicode capture groups. */
  var rsApos = "['\u2019]",
      rsAstral = '[' + rsAstralRange + ']',
      rsBreak = '[' + rsBreakRange + ']',
      rsCombo = '[' + rsComboRange + ']',
      rsDigits = '\\d+',
      rsDingbat = '[' + rsDingbatRange + ']',
      rsLower = '[' + rsLowerRange + ']',
      rsMisc = '[^' + rsAstralRange + rsBreakRange + rsDigits + rsDingbatRange + rsLowerRange + rsUpperRange + ']',
      rsFitz = '\\ud83c[\\udffb-\\udfff]',
      rsModifier = '(?:' + rsCombo + '|' + rsFitz + ')',
      rsNonAstral = '[^' + rsAstralRange + ']',
      rsRegional = '(?:\\ud83c[\\udde6-\\uddff]){2}',
      rsSurrPair = '[\\ud800-\\udbff][\\udc00-\\udfff]',
      rsUpper = '[' + rsUpperRange + ']',
      rsZWJ = '\\u200d';

  /** Used to compose unicode regexes. */
  var rsMiscLower = '(?:' + rsLower + '|' + rsMisc + ')',
      rsMiscUpper = '(?:' + rsUpper + '|' + rsMisc + ')',
      rsOptContrLower = '(?:' + rsApos + '(?:d|ll|m|re|s|t|ve))?',
      rsOptContrUpper = '(?:' + rsApos + '(?:D|LL|M|RE|S|T|VE))?',
      reOptMod = rsModifier + '?',
      rsOptVar = '[' + rsVarRange + ']?',
      rsOptJoin = '(?:' + rsZWJ + '(?:' + [rsNonAstral, rsRegional, rsSurrPair].join('|') + ')' + rsOptVar + reOptMod + ')*',
      rsOrdLower = '\\d*(?:1st|2nd|3rd|(?![123])\\dth)(?=\\b|[A-Z_])',
      rsOrdUpper = '\\d*(?:1ST|2ND|3RD|(?![123])\\dTH)(?=\\b|[a-z_])',
      rsSeq = rsOptVar + reOptMod + rsOptJoin,
      rsEmoji = '(?:' + [rsDingbat, rsRegional, rsSurrPair].join('|') + ')' + rsSeq,
      rsSymbol = '(?:' + [rsNonAstral + rsCombo + '?', rsCombo, rsRegional, rsSurrPair, rsAstral].join('|') + ')';

  /** Used to match apostrophes. */
  var reApos = RegExp(rsApos, 'g');

  /**
   * Used to match [combining diacritical marks](https://en.wikipedia.org/wiki/Combining_Diacritical_Marks) and
   * [combining diacritical marks for symbols](https://en.wikipedia.org/wiki/Combining_Diacritical_Marks_for_Symbols).
   */
  var reComboMark = RegExp(rsCombo, 'g');

  /** Used to match [string symbols](https://mathiasbynens.be/notes/javascript-unicode). */
  var reUnicode = RegExp(rsFitz + '(?=' + rsFitz + ')|' + rsSymbol + rsSeq, 'g');

  /** Used to match complex or compound words. */
  var reUnicodeWord = RegExp([
    rsUpper + '?' + rsLower + '+' + rsOptContrLower + '(?=' + [rsBreak, rsUpper, '$'].join('|') + ')',
    rsMiscUpper + '+' + rsOptContrUpper + '(?=' + [rsBreak, rsUpper + rsMiscLower, '$'].join('|') + ')',
    rsUpper + '?' + rsMiscLower + '+' + rsOptContrLower,
    rsUpper + '+' + rsOptContrUpper,
    rsOrdUpper,
    rsOrdLower,
    rsDigits,
    rsEmoji
  ].join('|'), 'g');

  /** Used to detect strings with [zero-width joiners or code points from the astral planes](http://eev.ee/blog/2015/09/12/dark-corners-of-unicode/). */
  var reHasUnicode = RegExp('[' + rsZWJ + rsAstralRange  + rsComboRange + rsVarRange + ']');

  /** Used to detect strings that need a more robust regexp to match words. */
  var reHasUnicodeWord = /[a-z][A-Z]|[A-Z]{2}[a-z]|[0-9][a-zA-Z]|[a-zA-Z][0-9]|[^a-zA-Z0-9 ]/;

  /** Used to assign default `context` object properties. */
  var contextProps = [
    'Array', 'Buffer', 'DataView', 'Date', 'Error', 'Float32Array', 'Float64Array',
    'Function', 'Int8Array', 'Int16Array', 'Int32Array', 'Map', 'Math', 'Object',
    'Promise', 'RegExp', 'Set', 'String', 'Symbol', 'TypeError', 'Uint8Array',
    'Uint8ClampedArray', 'Uint16Array', 'Uint32Array', 'WeakMap',
    '_', 'clearTimeout', 'isFinite', 'parseInt', 'setTimeout'
  ];

  /** Used to make template sourceURLs easier to identify. */
  var templateCounter = -1;

  /** Used to identify `toStringTag` values of typed arrays. */
  var typedArrayTags = {};
  typedArrayTags[float32Tag] = typedArrayTags[float64Tag] =
  typedArrayTags[int8Tag] = typedArrayTags[int16Tag] =
  typedArrayTags[int32Tag] = typedArrayTags[uint8Tag] =
  typedArrayTags[uint8ClampedTag] = typedArrayTags[uint16Tag] =
  typedArrayTags[uint32Tag] = true;
  typedArrayTags[argsTag] = typedArrayTags[arrayTag] =
  typedArrayTags[arrayBufferTag] = typedArrayTags[boolTag] =
  typedArrayTags[dataViewTag] = typedArrayTags[dateTag] =
  typedArrayTags[errorTag] = typedArrayTags[funcTag] =
  typedArrayTags[mapTag] = typedArrayTags[numberTag] =
  typedArrayTags[objectTag] = typedArrayTags[regexpTag] =
  typedArrayTags[setTag] = typedArrayTags[stringTag] =
  typedArrayTags[weakMapTag] = false;

  /** Used to identify `toStringTag` values supported by `_.clone`. */
  var cloneableTags = {};
  cloneableTags[argsTag] = cloneableTags[arrayTag] =
  cloneableTags[arrayBufferTag] = cloneableTags[dataViewTag] =
  cloneableTags[boolTag] = cloneableTags[dateTag] =
  cloneableTags[float32Tag] = cloneableTags[float64Tag] =
  cloneableTags[int8Tag] = cloneableTags[int16Tag] =
  cloneableTags[int32Tag] = cloneableTags[mapTag] =
  cloneableTags[numberTag] = cloneableTags[objectTag] =
  cloneableTags[regexpTag] = cloneableTags[setTag] =
  cloneableTags[stringTag] = cloneableTags[symbolTag] =
  cloneableTags[uint8Tag] = cloneableTags[uint8ClampedTag] =
  cloneableTags[uint16Tag] = cloneableTags[uint32Tag] = true;
  cloneableTags[errorTag] = cloneableTags[funcTag] =
  cloneableTags[weakMapTag] = false;

  /** Used to map Latin Unicode letters to basic Latin letters. */
  var deburredLetters = {
    // Latin-1 Supplement block.
    '\xc0': 'A',  '\xc1': 'A', '\xc2': 'A', '\xc3': 'A', '\xc4': 'A', '\xc5': 'A',
    '\xe0': 'a',  '\xe1': 'a', '\xe2': 'a', '\xe3': 'a', '\xe4': 'a', '\xe5': 'a',
    '\xc7': 'C',  '\xe7': 'c',
    '\xd0': 'D',  '\xf0': 'd',
    '\xc8': 'E',  '\xc9': 'E', '\xca': 'E', '\xcb': 'E',
    '\xe8': 'e',  '\xe9': 'e', '\xea': 'e', '\xeb': 'e',
    '\xcc': 'I',  '\xcd': 'I', '\xce': 'I', '\xcf': 'I',
    '\xec': 'i',  '\xed': 'i', '\xee': 'i', '\xef': 'i',
    '\xd1': 'N',  '\xf1': 'n',
    '\xd2': 'O',  '\xd3': 'O', '\xd4': 'O', '\xd5': 'O', '\xd6': 'O', '\xd8': 'O',
    '\xf2': 'o',  '\xf3': 'o', '\xf4': 'o', '\xf5': 'o', '\xf6': 'o', '\xf8': 'o',
    '\xd9': 'U',  '\xda': 'U', '\xdb': 'U', '\xdc': 'U',
    '\xf9': 'u',  '\xfa': 'u', '\xfb': 'u', '\xfc': 'u',
    '\xdd': 'Y',  '\xfd': 'y', '\xff': 'y',
    '\xc6': 'Ae', '\xe6': 'ae',
    '\xde': 'Th', '\xfe': 'th',
    '\xdf': 'ss',
    // Latin Extended-A block.
    '\u0100': 'A',  '\u0102': 'A', '\u0104': 'A',
    '\u0101': 'a',  '\u0103': 'a', '\u0105': 'a',
    '\u0106': 'C',  '\u0108': 'C', '\u010a': 'C', '\u010c': 'C',
    '\u0107': 'c',  '\u0109': 'c', '\u010b': 'c', '\u010d': 'c',
    '\u010e': 'D',  '\u0110': 'D', '\u010f': 'd', '\u0111': 'd',
    '\u0112': 'E',  '\u0114': 'E', '\u0116': 'E', '\u0118': 'E', '\u011a': 'E',
    '\u0113': 'e',  '\u0115': 'e', '\u0117': 'e', '\u0119': 'e', '\u011b': 'e',
    '\u011c': 'G',  '\u011e': 'G', '\u0120': 'G', '\u0122': 'G',
    '\u011d': 'g',  '\u011f': 'g', '\u0121': 'g', '\u0123': 'g',
    '\u0124': 'H',  '\u0126': 'H', '\u0125': 'h', '\u0127': 'h',
    '\u0128': 'I',  '\u012a': 'I', '\u012c': 'I', '\u012e': 'I', '\u0130': 'I',
    '\u0129': 'i',  '\u012b': 'i', '\u012d': 'i', '\u012f': 'i', '\u0131': 'i',
    '\u0134': 'J',  '\u0135': 'j',
    '\u0136': 'K',  '\u0137': 'k', '\u0138': 'k',
    '\u0139': 'L',  '\u013b': 'L', '\u013d': 'L', '\u013f': 'L', '\u0141': 'L',
    '\u013a': 'l',  '\u013c': 'l', '\u013e': 'l', '\u0140': 'l', '\u0142': 'l',
    '\u0143': 'N',  '\u0145': 'N', '\u0147': 'N', '\u014a': 'N',
    '\u0144': 'n',  '\u0146': 'n', '\u0148': 'n', '\u014b': 'n',
    '\u014c': 'O',  '\u014e': 'O', '\u0150': 'O',
    '\u014d': 'o',  '\u014f': 'o', '\u0151': 'o',
    '\u0154': 'R',  '\u0156': 'R', '\u0158': 'R',
    '\u0155': 'r',  '\u0157': 'r', '\u0159': 'r',
    '\u015a': 'S',  '\u015c': 'S', '\u015e': 'S', '\u0160': 'S',
    '\u015b': 's',  '\u015d': 's', '\u015f': 's', '\u0161': 's',
    '\u0162': 'T',  '\u0164': 'T', '\u0166': 'T',
    '\u0163': 't',  '\u0165': 't', '\u0167': 't',
    '\u0168': 'U',  '\u016a': 'U', '\u016c': 'U', '\u016e': 'U', '\u0170': 'U', '\u0172': 'U',
    '\u0169': 'u',  '\u016b': 'u', '\u016d': 'u', '\u016f': 'u', '\u0171': 'u', '\u0173': 'u',
    '\u0174': 'W',  '\u0175': 'w',
    '\u0176': 'Y',  '\u0177': 'y', '\u0178': 'Y',
    '\u0179': 'Z',  '\u017b': 'Z', '\u017d': 'Z',
    '\u017a': 'z',  '\u017c': 'z', '\u017e': 'z',
    '\u0132': 'IJ', '\u0133': 'ij',
    '\u0152': 'Oe', '\u0153': 'oe',
    '\u0149': "'n", '\u017f': 's'
  };

  /** Used to map characters to HTML entities. */
  var htmlEscapes = {
    '&': '&amp;',
    '<': '&lt;',
    '>': '&gt;',
    '"': '&quot;',
    "'": '&#39;'
  };

  /** Used to map HTML entities to characters. */
  var htmlUnescapes = {
    '&amp;': '&',
    '&lt;': '<',
    '&gt;': '>',
    '&quot;': '"',
    '&#39;': "'"
  };

  /** Used to escape characters for inclusion in compiled string literals. */
  var stringEscapes = {
    '\\': '\\',
    "'": "'",
    '\n': 'n',
    '\r': 'r',
    '\u2028': 'u2028',
    '\u2029': 'u2029'
  };

  /** Built-in method references without a dependency on `root`. */
  var freeParseFloat = parseFloat,
      freeParseInt = parseInt;

  /** Detect free variable `global` from Node.js. */
  var freeGlobal = typeof __webpack_require__.g == 'object' && __webpack_require__.g && __webpack_require__.g.Object === Object && __webpack_require__.g;

  /** Detect free variable `self`. */
  var freeSelf = typeof self == 'object' && self && self.Object === Object && self;

  /** Used as a reference to the global object. */
  var root = freeGlobal || freeSelf || Function('return this')();

  /** Detect free variable `exports`. */
  var freeExports =  true && exports && !exports.nodeType && exports;

  /** Detect free variable `module`. */
  var freeModule = freeExports && "object" == 'object' && module && !module.nodeType && module;

  /** Detect the popular CommonJS extension `module.exports`. */
  var moduleExports = freeModule && freeModule.exports === freeExports;

  /** Detect free variable `process` from Node.js. */
  var freeProcess = moduleExports && freeGlobal.process;

  /** Used to access faster Node.js helpers. */
  var nodeUtil = (function() {
    try {
      // Use `util.types` for Node.js 10+.
      var types = freeModule && freeModule.require && freeModule.require('util').types;

      if (types) {
        return types;
      }

      // Legacy `process.binding('util')` for Node.js < 10.
      return freeProcess && freeProcess.binding && freeProcess.binding('util');
    } catch (e) {}
  }());

  /* Node.js helper references. */
  var nodeIsArrayBuffer = nodeUtil && nodeUtil.isArrayBuffer,
      nodeIsDate = nodeUtil && nodeUtil.isDate,
      nodeIsMap = nodeUtil && nodeUtil.isMap,
      nodeIsRegExp = nodeUtil && nodeUtil.isRegExp,
      nodeIsSet = nodeUtil && nodeUtil.isSet,
      nodeIsTypedArray = nodeUtil && nodeUtil.isTypedArray;

  /*--------------------------------------------------------------------------*/

  /**
   * A faster alternative to `Function#apply`, this function invokes `func`
   * with the `this` binding of `thisArg` and the arguments of `args`.
   *
   * @private
   * @param {Function} func The function to invoke.
   * @param {*} thisArg The `this` binding of `func`.
   * @param {Array} args The arguments to invoke `func` with.
   * @returns {*} Returns the result of `func`.
   */
  function apply(func, thisArg, args) {
    switch (args.length) {
      case 0: return func.call(thisArg);
      case 1: return func.call(thisArg, args[0]);
      case 2: return func.call(thisArg, args[0], args[1]);
      case 3: return func.call(thisArg, args[0], args[1], args[2]);
    }
    return func.apply(thisArg, args);
  }

  /**
   * A specialized version of `baseAggregator` for arrays.
   *
   * @private
   * @param {Array} [array] The array to iterate over.
   * @param {Function} setter The function to set `accumulator` values.
   * @param {Function} iteratee The iteratee to transform keys.
   * @param {Object} accumulator The initial aggregated object.
   * @returns {Function} Returns `accumulator`.
   */
  function arrayAggregator(array, setter, iteratee, accumulator) {
    var index = -1,
        length = array == null ? 0 : array.length;

    while (++index < length) {
      var value = array[index];
      setter(accumulator, value, iteratee(value), array);
    }
    return accumulator;
  }

  /**
   * A specialized version of `_.forEach` for arrays without support for
   * iteratee shorthands.
   *
   * @private
   * @param {Array} [array] The array to iterate over.
   * @param {Function} iteratee The function invoked per iteration.
   * @returns {Array} Returns `array`.
   */
  function arrayEach(array, iteratee) {
    var index = -1,
        length = array == null ? 0 : array.length;

    while (++index < length) {
      if (iteratee(array[index], index, array) === false) {
        break;
      }
    }
    return array;
  }

  /**
   * A specialized version of `_.forEachRight` for arrays without support for
   * iteratee shorthands.
   *
   * @private
   * @param {Array} [array] The array to iterate over.
   * @param {Function} iteratee The function invoked per iteration.
   * @returns {Array} Returns `array`.
   */
  function arrayEachRight(array, iteratee) {
    var length = array == null ? 0 : array.length;

    while (length--) {
      if (iteratee(array[length], length, array) === false) {
        break;
      }
    }
    return array;
  }

  /**
   * A specialized version of `_.every` for arrays without support for
   * iteratee shorthands.
   *
   * @private
   * @param {Array} [array] The array to iterate over.
   * @param {Function} predicate The function invoked per iteration.
   * @returns {boolean} Returns `true` if all elements pass the predicate check,
   *  else `false`.
   */
  function arrayEvery(array, predicate) {
    var index = -1,
        length = array == null ? 0 : array.length;

    while (++index < length) {
      if (!predicate(array[index], index, array)) {
        return false;
      }
    }
    return true;
  }

  /**
   * A specialized version of `_.filter` for arrays without support for
   * iteratee shorthands.
   *
   * @private
   * @param {Array} [array] The array to iterate over.
   * @param {Function} predicate The function invoked per iteration.
   * @returns {Array} Returns the new filtered array.
   */
  function arrayFilter(array, predicate) {
    var index = -1,
        length = array == null ? 0 : array.length,
        resIndex = 0,
        result = [];

    while (++index < length) {
      var value = array[index];
      if (predicate(value, index, array)) {
        result[resIndex++] = value;
      }
    }
    return result;
  }

  /**
   * A specialized version of `_.includes` for arrays without support for
   * specifying an index to search from.
   *
   * @private
   * @param {Array} [array] The array to inspect.
   * @param {*} target The value to search for.
   * @returns {boolean} Returns `true` if `target` is found, else `false`.
   */
  function arrayIncludes(array, value) {
    var length = array == null ? 0 : array.length;
    return !!length && baseIndexOf(array, value, 0) > -1;
  }

  /**
   * This function is like `arrayIncludes` except that it accepts a comparator.
   *
   * @private
   * @param {Array} [array] The array to inspect.
   * @param {*} target The value to search for.
   * @param {Function} comparator The comparator invoked per element.
   * @returns {boolean} Returns `true` if `target` is found, else `false`.
   */
  function arrayIncludesWith(array, value, comparator) {
    var index = -1,
        length = array == null ? 0 : array.length;

    while (++index < length) {
      if (comparator(value, array[index])) {
        return true;
      }
    }
    return false;
  }

  /**
   * A specialized version of `_.map` for arrays without support for iteratee
   * shorthands.
   *
   * @private
   * @param {Array} [array] The array to iterate over.
   * @param {Function} iteratee The function invoked per iteration.
   * @returns {Array} Returns the new mapped array.
   */
  function arrayMap(array, iteratee) {
    var index = -1,
        length = array == null ? 0 : array.length,
        result = Array(length);

    while (++index < length) {
      result[index] = iteratee(array[index], index, array);
    }
    return result;
  }

  /**
   * Appends the elements of `values` to `array`.
   *
   * @private
   * @param {Array} array The array to modify.
   * @param {Array} values The values to append.
   * @returns {Array} Returns `array`.
   */
  function arrayPush(array, values) {
    var index = -1,
        length = values.length,
        offset = array.length;

    while (++index < length) {
      array[offset + index] = values[index];
    }
    return array;
  }

  /**
   * A specialized version of `_.reduce` for arrays without support for
   * iteratee shorthands.
   *
   * @private
   * @param {Array} [array] The array to iterate over.
   * @param {Function} iteratee The function invoked per iteration.
   * @param {*} [accumulator] The initial value.
   * @param {boolean} [initAccum] Specify using the first element of `array` as
   *  the initial value.
   * @returns {*} Returns the accumulated value.
   */
  function arrayReduce(array, iteratee, accumulator, initAccum) {
    var index = -1,
        length = array == null ? 0 : array.length;

    if (initAccum && length) {
      accumulator = array[++index];
    }
    while (++index < length) {
      accumulator = iteratee(accumulator, array[index], index, array);
    }
    return accumulator;
  }

  /**
   * A specialized version of `_.reduceRight` for arrays without support for
   * iteratee shorthands.
   *
   * @private
   * @param {Array} [array] The array to iterate over.
   * @param {Function} iteratee The function invoked per iteration.
   * @param {*} [accumulator] The initial value.
   * @param {boolean} [initAccum] Specify using the last element of `array` as
   *  the initial value.
   * @returns {*} Returns the accumulated value.
   */
  function arrayReduceRight(array, iteratee, accumulator, initAccum) {
    var length = array == null ? 0 : array.length;
    if (initAccum && length) {
      accumulator = array[--length];
    }
    while (length--) {
      accumulator = iteratee(accumulator, array[length], length, array);
    }
    return accumulator;
  }

  /**
   * A specialized version of `_.some` for arrays without support for iteratee
   * shorthands.
   *
   * @private
   * @param {Array} [array] The array to iterate over.
   * @param {Function} predicate The function invoked per iteration.
   * @returns {boolean} Returns `true` if any element passes the predicate check,
   *  else `false`.
   */
  function arraySome(array, predicate) {
    var index = -1,
        length = array == null ? 0 : array.length;

    while (++index < length) {
      if (predicate(array[index], index, array)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Gets the size of an ASCII `string`.
   *
   * @private
   * @param {string} string The string inspect.
   * @returns {number} Returns the string size.
   */
  var asciiSize = baseProperty('length');

  /**
   * Converts an ASCII `string` to an array.
   *
   * @private
   * @param {string} string The string to convert.
   * @returns {Array} Returns the converted array.
   */
  function asciiToArray(string) {
    return string.split('');
  }

  /**
   * Splits an ASCII `string` into an array of its words.
   *
   * @private
   * @param {string} The string to inspect.
   * @returns {Array} Returns the words of `string`.
   */
  function asciiWords(string) {
    return string.match(reAsciiWord) || [];
  }

  /**
   * The base implementation of methods like `_.findKey` and `_.findLastKey`,
   * without support for iteratee shorthands, which iterates over `collection`
   * using `eachFunc`.
   *
   * @private
   * @param {Array|Object} collection The collection to inspect.
   * @param {Function} predicate The function invoked per iteration.
   * @param {Function} eachFunc The function to iterate over `collection`.
   * @returns {*} Returns the found element or its key, else `undefined`.
   */
  function baseFindKey(collection, predicate, eachFunc) {
    var result;
    eachFunc(collection, function(value, key, collection) {
      if (predicate(value, key, collection)) {
        result = key;
        return false;
      }
    });
    return result;
  }

  /**
   * The base implementation of `_.findIndex` and `_.findLastIndex` without
   * support for iteratee shorthands.
   *
   * @private
   * @param {Array} array The array to inspect.
   * @param {Function} predicate The function invoked per iteration.
   * @param {number} fromIndex The index to search from.
   * @param {boolean} [fromRight] Specify iterating from right to left.
   * @returns {number} Returns the index of the matched value, else `-1`.
   */
  function baseFindIndex(array, predicate, fromIndex, fromRight) {
    var length = array.length,
        index = fromIndex + (fromRight ? 1 : -1);

    while ((fromRight ? index-- : ++index < length)) {
      if (predicate(array[index], index, array)) {
        return index;
      }
    }
    return -1;
  }

  /**
   * The base implementation of `_.indexOf` without `fromIndex` bounds checks.
   *
   * @private
   * @param {Array} array The array to inspect.
   * @param {*} value The value to search for.
   * @param {number} fromIndex The index to search from.
   * @returns {number} Returns the index of the matched value, else `-1`.
   */
  function baseIndexOf(array, value, fromIndex) {
    return value === value
      ? strictIndexOf(array, value, fromIndex)
      : baseFindIndex(array, baseIsNaN, fromIndex);
  }

  /**
   * This function is like `baseIndexOf` except that it accepts a comparator.
   *
   * @private
   * @param {Array} array The array to inspect.
   * @param {*} value The value to search for.
   * @param {number} fromIndex The index to search from.
   * @param {Function} comparator The comparator invoked per element.
   * @returns {number} Returns the index of the matched value, else `-1`.
   */
  function baseIndexOfWith(array, value, fromIndex, comparator) {
    var index = fromIndex - 1,
        length = array.length;

    while (++index < length) {
      if (comparator(array[index], value)) {
        return index;
      }
    }
    return -1;
  }

  /**
   * The base implementation of `_.isNaN` without support for number objects.
   *
   * @private
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is `NaN`, else `false`.
   */
  function baseIsNaN(value) {
    return value !== value;
  }

  /**
   * The base implementation of `_.mean` and `_.meanBy` without support for
   * iteratee shorthands.
   *
   * @private
   * @param {Array} array The array to iterate over.
   * @param {Function} iteratee The function invoked per iteration.
   * @returns {number} Returns the mean.
   */
  function baseMean(array, iteratee) {
    var length = array == null ? 0 : array.length;
    return length ? (baseSum(array, iteratee) / length) : NAN;
  }

  /**
   * The base implementation of `_.property` without support for deep paths.
   *
   * @private
   * @param {string} key The key of the property to get.
   * @returns {Function} Returns the new accessor function.
   */
  function baseProperty(key) {
    return function(object) {
      return object == null ? undefined : object[key];
    };
  }

  /**
   * The base implementation of `_.propertyOf` without support for deep paths.
   *
   * @private
   * @param {Object} object The object to query.
   * @returns {Function} Returns the new accessor function.
   */
  function basePropertyOf(object) {
    return function(key) {
      return object == null ? undefined : object[key];
    };
  }

  /**
   * The base implementation of `_.reduce` and `_.reduceRight`, without support
   * for iteratee shorthands, which iterates over `collection` using `eachFunc`.
   *
   * @private
   * @param {Array|Object} collection The collection to iterate over.
   * @param {Function} iteratee The function invoked per iteration.
   * @param {*} accumulator The initial value.
   * @param {boolean} initAccum Specify using the first or last element of
   *  `collection` as the initial value.
   * @param {Function} eachFunc The function to iterate over `collection`.
   * @returns {*} Returns the accumulated value.
   */
  function baseReduce(collection, iteratee, accumulator, initAccum, eachFunc) {
    eachFunc(collection, function(value, index, collection) {
      accumulator = initAccum
        ? (initAccum = false, value)
        : iteratee(accumulator, value, index, collection);
    });
    return accumulator;
  }

  /**
   * The base implementation of `_.sortBy` which uses `comparer` to define the
   * sort order of `array` and replaces criteria objects with their corresponding
   * values.
   *
   * @private
   * @param {Array} array The array to sort.
   * @param {Function} comparer The function to define sort order.
   * @returns {Array} Returns `array`.
   */
  function baseSortBy(array, comparer) {
    var length = array.length;

    array.sort(comparer);
    while (length--) {
      array[length] = array[length].value;
    }
    return array;
  }

  /**
   * The base implementation of `_.sum` and `_.sumBy` without support for
   * iteratee shorthands.
   *
   * @private
   * @param {Array} array The array to iterate over.
   * @param {Function} iteratee The function invoked per iteration.
   * @returns {number} Returns the sum.
   */
  function baseSum(array, iteratee) {
    var result,
        index = -1,
        length = array.length;

    while (++index < length) {
      var current = iteratee(array[index]);
      if (current !== undefined) {
        result = result === undefined ? current : (result + current);
      }
    }
    return result;
  }

  /**
   * The base implementation of `_.times` without support for iteratee shorthands
   * or max array length checks.
   *
   * @private
   * @param {number} n The number of times to invoke `iteratee`.
   * @param {Function} iteratee The function invoked per iteration.
   * @returns {Array} Returns the array of results.
   */
  function baseTimes(n, iteratee) {
    var index = -1,
        result = Array(n);

    while (++index < n) {
      result[index] = iteratee(index);
    }
    return result;
  }

  /**
   * The base implementation of `_.toPairs` and `_.toPairsIn` which creates an array
   * of key-value pairs for `object` corresponding to the property names of `props`.
   *
   * @private
   * @param {Object} object The object to query.
   * @param {Array} props The property names to get values for.
   * @returns {Object} Returns the key-value pairs.
   */
  function baseToPairs(object, props) {
    return arrayMap(props, function(key) {
      return [key, object[key]];
    });
  }

  /**
   * The base implementation of `_.trim`.
   *
   * @private
   * @param {string} string The string to trim.
   * @returns {string} Returns the trimmed string.
   */
  function baseTrim(string) {
    return string
      ? string.slice(0, trimmedEndIndex(string) + 1).replace(reTrimStart, '')
      : string;
  }

  /**
   * The base implementation of `_.unary` without support for storing metadata.
   *
   * @private
   * @param {Function} func The function to cap arguments for.
   * @returns {Function} Returns the new capped function.
   */
  function baseUnary(func) {
    return function(value) {
      return func(value);
    };
  }

  /**
   * The base implementation of `_.values` and `_.valuesIn` which creates an
   * array of `object` property values corresponding to the property names
   * of `props`.
   *
   * @private
   * @param {Object} object The object to query.
   * @param {Array} props The property names to get values for.
   * @returns {Object} Returns the array of property values.
   */
  function baseValues(object, props) {
    return arrayMap(props, function(key) {
      return object[key];
    });
  }

  /**
   * Checks if a `cache` value for `key` exists.
   *
   * @private
   * @param {Object} cache The cache to query.
   * @param {string} key The key of the entry to check.
   * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
   */
  function cacheHas(cache, key) {
    return cache.has(key);
  }

  /**
   * Used by `_.trim` and `_.trimStart` to get the index of the first string symbol
   * that is not found in the character symbols.
   *
   * @private
   * @param {Array} strSymbols The string symbols to inspect.
   * @param {Array} chrSymbols The character symbols to find.
   * @returns {number} Returns the index of the first unmatched string symbol.
   */
  function charsStartIndex(strSymbols, chrSymbols) {
    var index = -1,
        length = strSymbols.length;

    while (++index < length && baseIndexOf(chrSymbols, strSymbols[index], 0) > -1) {}
    return index;
  }

  /**
   * Used by `_.trim` and `_.trimEnd` to get the index of the last string symbol
   * that is not found in the character symbols.
   *
   * @private
   * @param {Array} strSymbols The string symbols to inspect.
   * @param {Array} chrSymbols The character symbols to find.
   * @returns {number} Returns the index of the last unmatched string symbol.
   */
  function charsEndIndex(strSymbols, chrSymbols) {
    var index = strSymbols.length;

    while (index-- && baseIndexOf(chrSymbols, strSymbols[index], 0) > -1) {}
    return index;
  }

  /**
   * Gets the number of `placeholder` occurrences in `array`.
   *
   * @private
   * @param {Array} array The array to inspect.
   * @param {*} placeholder The placeholder to search for.
   * @returns {number} Returns the placeholder count.
   */
  function countHolders(array, placeholder) {
    var length = array.length,
        result = 0;

    while (length--) {
      if (array[length] === placeholder) {
        ++result;
      }
    }
    return result;
  }

  /**
   * Used by `_.deburr` to convert Latin-1 Supplement and Latin Extended-A
   * letters to basic Latin letters.
   *
   * @private
   * @param {string} letter The matched letter to deburr.
   * @returns {string} Returns the deburred letter.
   */
  var deburrLetter = basePropertyOf(deburredLetters);

  /**
   * Used by `_.escape` to convert characters to HTML entities.
   *
   * @private
   * @param {string} chr The matched character to escape.
   * @returns {string} Returns the escaped character.
   */
  var escapeHtmlChar = basePropertyOf(htmlEscapes);

  /**
   * Used by `_.template` to escape characters for inclusion in compiled string literals.
   *
   * @private
   * @param {string} chr The matched character to escape.
   * @returns {string} Returns the escaped character.
   */
  function escapeStringChar(chr) {
    return '\\' + stringEscapes[chr];
  }

  /**
   * Gets the value at `key` of `object`.
   *
   * @private
   * @param {Object} [object] The object to query.
   * @param {string} key The key of the property to get.
   * @returns {*} Returns the property value.
   */
  function getValue(object, key) {
    return object == null ? undefined : object[key];
  }

  /**
   * Checks if `string` contains Unicode symbols.
   *
   * @private
   * @param {string} string The string to inspect.
   * @returns {boolean} Returns `true` if a symbol is found, else `false`.
   */
  function hasUnicode(string) {
    return reHasUnicode.test(string);
  }

  /**
   * Checks if `string` contains a word composed of Unicode symbols.
   *
   * @private
   * @param {string} string The string to inspect.
   * @returns {boolean} Returns `true` if a word is found, else `false`.
   */
  function hasUnicodeWord(string) {
    return reHasUnicodeWord.test(string);
  }

  /**
   * Converts `iterator` to an array.
   *
   * @private
   * @param {Object} iterator The iterator to convert.
   * @returns {Array} Returns the converted array.
   */
  function iteratorToArray(iterator) {
    var data,
        result = [];

    while (!(data = iterator.next()).done) {
      result.push(data.value);
    }
    return result;
  }

  /**
   * Converts `map` to its key-value pairs.
   *
   * @private
   * @param {Object} map The map to convert.
   * @returns {Array} Returns the key-value pairs.
   */
  function mapToArray(map) {
    var index = -1,
        result = Array(map.size);

    map.forEach(function(value, key) {
      result[++index] = [key, value];
    });
    return result;
  }

  /**
   * Creates a unary function that invokes `func` with its argument transformed.
   *
   * @private
   * @param {Function} func The function to wrap.
   * @param {Function} transform The argument transform.
   * @returns {Function} Returns the new function.
   */
  function overArg(func, transform) {
    return function(arg) {
      return func(transform(arg));
    };
  }

  /**
   * Replaces all `placeholder` elements in `array` with an internal placeholder
   * and returns an array of their indexes.
   *
   * @private
   * @param {Array} array The array to modify.
   * @param {*} placeholder The placeholder to replace.
   * @returns {Array} Returns the new array of placeholder indexes.
   */
  function replaceHolders(array, placeholder) {
    var index = -1,
        length = array.length,
        resIndex = 0,
        result = [];

    while (++index < length) {
      var value = array[index];
      if (value === placeholder || value === PLACEHOLDER) {
        array[index] = PLACEHOLDER;
        result[resIndex++] = index;
      }
    }
    return result;
  }

  /**
   * Converts `set` to an array of its values.
   *
   * @private
   * @param {Object} set The set to convert.
   * @returns {Array} Returns the values.
   */
  function setToArray(set) {
    var index = -1,
        result = Array(set.size);

    set.forEach(function(value) {
      result[++index] = value;
    });
    return result;
  }

  /**
   * Converts `set` to its value-value pairs.
   *
   * @private
   * @param {Object} set The set to convert.
   * @returns {Array} Returns the value-value pairs.
   */
  function setToPairs(set) {
    var index = -1,
        result = Array(set.size);

    set.forEach(function(value) {
      result[++index] = [value, value];
    });
    return result;
  }

  /**
   * A specialized version of `_.indexOf` which performs strict equality
   * comparisons of values, i.e. `===`.
   *
   * @private
   * @param {Array} array The array to inspect.
   * @param {*} value The value to search for.
   * @param {number} fromIndex The index to search from.
   * @returns {number} Returns the index of the matched value, else `-1`.
   */
  function strictIndexOf(array, value, fromIndex) {
    var index = fromIndex - 1,
        length = array.length;

    while (++index < length) {
      if (array[index] === value) {
        return index;
      }
    }
    return -1;
  }

  /**
   * A specialized version of `_.lastIndexOf` which performs strict equality
   * comparisons of values, i.e. `===`.
   *
   * @private
   * @param {Array} array The array to inspect.
   * @param {*} value The value to search for.
   * @param {number} fromIndex The index to search from.
   * @returns {number} Returns the index of the matched value, else `-1`.
   */
  function strictLastIndexOf(array, value, fromIndex) {
    var index = fromIndex + 1;
    while (index--) {
      if (array[index] === value) {
        return index;
      }
    }
    return index;
  }

  /**
   * Gets the number of symbols in `string`.
   *
   * @private
   * @param {string} string The string to inspect.
   * @returns {number} Returns the string size.
   */
  function stringSize(string) {
    return hasUnicode(string)
      ? unicodeSize(string)
      : asciiSize(string);
  }

  /**
   * Converts `string` to an array.
   *
   * @private
   * @param {string} string The string to convert.
   * @returns {Array} Returns the converted array.
   */
  function stringToArray(string) {
    return hasUnicode(string)
      ? unicodeToArray(string)
      : asciiToArray(string);
  }

  /**
   * Used by `_.trim` and `_.trimEnd` to get the index of the last non-whitespace
   * character of `string`.
   *
   * @private
   * @param {string} string The string to inspect.
   * @returns {number} Returns the index of the last non-whitespace character.
   */
  function trimmedEndIndex(string) {
    var index = string.length;

    while (index-- && reWhitespace.test(string.charAt(index))) {}
    return index;
  }

  /**
   * Used by `_.unescape` to convert HTML entities to characters.
   *
   * @private
   * @param {string} chr The matched character to unescape.
   * @returns {string} Returns the unescaped character.
   */
  var unescapeHtmlChar = basePropertyOf(htmlUnescapes);

  /**
   * Gets the size of a Unicode `string`.
   *
   * @private
   * @param {string} string The string inspect.
   * @returns {number} Returns the string size.
   */
  function unicodeSize(string) {
    var result = reUnicode.lastIndex = 0;
    while (reUnicode.test(string)) {
      ++result;
    }
    return result;
  }

  /**
   * Converts a Unicode `string` to an array.
   *
   * @private
   * @param {string} string The string to convert.
   * @returns {Array} Returns the converted array.
   */
  function unicodeToArray(string) {
    return string.match(reUnicode) || [];
  }

  /**
   * Splits a Unicode `string` into an array of its words.
   *
   * @private
   * @param {string} The string to inspect.
   * @returns {Array} Returns the words of `string`.
   */
  function unicodeWords(string) {
    return string.match(reUnicodeWord) || [];
  }

  /*--------------------------------------------------------------------------*/

  /**
   * Create a new pristine `lodash` function using the `context` object.
   *
   * @static
   * @memberOf _
   * @since 1.1.0
   * @category Util
   * @param {Object} [context=root] The context object.
   * @returns {Function} Returns a new `lodash` function.
   * @example
   *
   * _.mixin({ 'foo': _.constant('foo') });
   *
   * var lodash = _.runInContext();
   * lodash.mixin({ 'bar': lodash.constant('bar') });
   *
   * _.isFunction(_.foo);
   * // => true
   * _.isFunction(_.bar);
   * // => false
   *
   * lodash.isFunction(lodash.foo);
   * // => false
   * lodash.isFunction(lodash.bar);
   * // => true
   *
   * // Create a suped-up `defer` in Node.js.
   * var defer = _.runInContext({ 'setTimeout': setImmediate }).defer;
   */
  var runInContext = (function runInContext(context) {
    context = context == null ? root : _.defaults(root.Object(), context, _.pick(root, contextProps));

    /** Built-in constructor references. */
    var Array = context.Array,
        Date = context.Date,
        Error = context.Error,
        Function = context.Function,
        Math = context.Math,
        Object = context.Object,
        RegExp = context.RegExp,
        String = context.String,
        TypeError = context.TypeError;

    /** Used for built-in method references. */
    var arrayProto = Array.prototype,
        funcProto = Function.prototype,
        objectProto = Object.prototype;

    /** Used to detect overreaching core-js shims. */
    var coreJsData = context['__core-js_shared__'];

    /** Used to resolve the decompiled source of functions. */
    var funcToString = funcProto.toString;

    /** Used to check objects for own properties. */
    var hasOwnProperty = objectProto.hasOwnProperty;

    /** Used to generate unique IDs. */
    var idCounter = 0;

    /** Used to detect methods masquerading as native. */
    var maskSrcKey = (function() {
      var uid = /[^.]+$/.exec(coreJsData && coreJsData.keys && coreJsData.keys.IE_PROTO || '');
      return uid ? ('Symbol(src)_1.' + uid) : '';
    }());

    /**
     * Used to resolve the
     * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
     * of values.
     */
    var nativeObjectToString = objectProto.toString;

    /** Used to infer the `Object` constructor. */
    var objectCtorString = funcToString.call(Object);

    /** Used to restore the original `_` reference in `_.noConflict`. */
    var oldDash = root._;

    /** Used to detect if a method is native. */
    var reIsNative = RegExp('^' +
      funcToString.call(hasOwnProperty).replace(reRegExpChar, '\\$&')
      .replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g, '$1.*?') + '$'
    );

    /** Built-in value references. */
    var Buffer = moduleExports ? context.Buffer : undefined,
        Symbol = context.Symbol,
        Uint8Array = context.Uint8Array,
        allocUnsafe = Buffer ? Buffer.allocUnsafe : undefined,
        getPrototype = overArg(Object.getPrototypeOf, Object),
        objectCreate = Object.create,
        propertyIsEnumerable = objectProto.propertyIsEnumerable,
        splice = arrayProto.splice,
        spreadableSymbol = Symbol ? Symbol.isConcatSpreadable : undefined,
        symIterator = Symbol ? Symbol.iterator : undefined,
        symToStringTag = Symbol ? Symbol.toStringTag : undefined;

    var defineProperty = (function() {
      try {
        var func = getNative(Object, 'defineProperty');
        func({}, '', {});
        return func;
      } catch (e) {}
    }());

    /** Mocked built-ins. */
    var ctxClearTimeout = context.clearTimeout !== root.clearTimeout && context.clearTimeout,
        ctxNow = Date && Date.now !== root.Date.now && Date.now,
        ctxSetTimeout = context.setTimeout !== root.setTimeout && context.setTimeout;

    /* Built-in method references for those with the same name as other `lodash` methods. */
    var nativeCeil = Math.ceil,
        nativeFloor = Math.floor,
        nativeGetSymbols = Object.getOwnPropertySymbols,
        nativeIsBuffer = Buffer ? Buffer.isBuffer : undefined,
        nativeIsFinite = context.isFinite,
        nativeJoin = arrayProto.join,
        nativeKeys = overArg(Object.keys, Object),
        nativeMax = Math.max,
        nativeMin = Math.min,
        nativeNow = Date.now,
        nativeParseInt = context.parseInt,
        nativeRandom = Math.random,
        nativeReverse = arrayProto.reverse;

    /* Built-in method references that are verified to be native. */
    var DataView = getNative(context, 'DataView'),
        Map = getNative(context, 'Map'),
        Promise = getNative(context, 'Promise'),
        Set = getNative(context, 'Set'),
        WeakMap = getNative(context, 'WeakMap'),
        nativeCreate = getNative(Object, 'create');

    /** Used to store function metadata. */
    var metaMap = WeakMap && new WeakMap;

    /** Used to lookup unminified function names. */
    var realNames = {};

    /** Used to detect maps, sets, and weakmaps. */
    var dataViewCtorString = toSource(DataView),
        mapCtorString = toSource(Map),
        promiseCtorString = toSource(Promise),
        setCtorString = toSource(Set),
        weakMapCtorString = toSource(WeakMap);

    /** Used to convert symbols to primitives and strings. */
    var symbolProto = Symbol ? Symbol.prototype : undefined,
        symbolValueOf = symbolProto ? symbolProto.valueOf : undefined,
        symbolToString = symbolProto ? symbolProto.toString : undefined;

    /*------------------------------------------------------------------------*/

    /**
     * Creates a `lodash` object which wraps `value` to enable implicit method
     * chain sequences. Methods that operate on and return arrays, collections,
     * and functions can be chained together. Methods that retrieve a single value
     * or may return a primitive value will automatically end the chain sequence
     * and return the unwrapped value. Otherwise, the value must be unwrapped
     * with `_#value`.
     *
     * Explicit chain sequences, which must be unwrapped with `_#value`, may be
     * enabled using `_.chain`.
     *
     * The execution of chained methods is lazy, that is, it's deferred until
     * `_#value` is implicitly or explicitly called.
     *
     * Lazy evaluation allows several methods to support shortcut fusion.
     * Shortcut fusion is an optimization to merge iteratee calls; this avoids
     * the creation of intermediate arrays and can greatly reduce the number of
     * iteratee executions. Sections of a chain sequence qualify for shortcut
     * fusion if the section is applied to an array and iteratees accept only
     * one argument. The heuristic for whether a section qualifies for shortcut
     * fusion is subject to change.
     *
     * Chaining is supported in custom builds as long as the `_#value` method is
     * directly or indirectly included in the build.
     *
     * In addition to lodash methods, wrappers have `Array` and `String` methods.
     *
     * The wrapper `Array` methods are:
     * `concat`, `join`, `pop`, `push`, `shift`, `sort`, `splice`, and `unshift`
     *
     * The wrapper `String` methods are:
     * `replace` and `split`
     *
     * The wrapper methods that support shortcut fusion are:
     * `at`, `compact`, `drop`, `dropRight`, `dropWhile`, `filter`, `find`,
     * `findLast`, `head`, `initial`, `last`, `map`, `reject`, `reverse`, `slice`,
     * `tail`, `take`, `takeRight`, `takeRightWhile`, `takeWhile`, and `toArray`
     *
     * The chainable wrapper methods are:
     * `after`, `ary`, `assign`, `assignIn`, `assignInWith`, `assignWith`, `at`,
     * `before`, `bind`, `bindAll`, `bindKey`, `castArray`, `chain`, `chunk`,
     * `commit`, `compact`, `concat`, `conforms`, `constant`, `countBy`, `create`,
     * `curry`, `debounce`, `defaults`, `defaultsDeep`, `defer`, `delay`,
     * `difference`, `differenceBy`, `differenceWith`, `drop`, `dropRight`,
     * `dropRightWhile`, `dropWhile`, `extend`, `extendWith`, `fill`, `filter`,
     * `flatMap`, `flatMapDeep`, `flatMapDepth`, `flatten`, `flattenDeep`,
     * `flattenDepth`, `flip`, `flow`, `flowRight`, `fromPairs`, `functions`,
     * `functionsIn`, `groupBy`, `initial`, `intersection`, `intersectionBy`,
     * `intersectionWith`, `invert`, `invertBy`, `invokeMap`, `iteratee`, `keyBy`,
     * `keys`, `keysIn`, `map`, `mapKeys`, `mapValues`, `matches`, `matchesProperty`,
     * `memoize`, `merge`, `mergeWith`, `method`, `methodOf`, `mixin`, `negate`,
     * `nthArg`, `omit`, `omitBy`, `once`, `orderBy`, `over`, `overArgs`,
     * `overEvery`, `overSome`, `partial`, `partialRight`, `partition`, `pick`,
     * `pickBy`, `plant`, `property`, `propertyOf`, `pull`, `pullAll`, `pullAllBy`,
     * `pullAllWith`, `pullAt`, `push`, `range`, `rangeRight`, `rearg`, `reject`,
     * `remove`, `rest`, `reverse`, `sampleSize`, `set`, `setWith`, `shuffle`,
     * `slice`, `sort`, `sortBy`, `splice`, `spread`, `tail`, `take`, `takeRight`,
     * `takeRightWhile`, `takeWhile`, `tap`, `throttle`, `thru`, `toArray`,
     * `toPairs`, `toPairsIn`, `toPath`, `toPlainObject`, `transform`, `unary`,
     * `union`, `unionBy`, `unionWith`, `uniq`, `uniqBy`, `uniqWith`, `unset`,
     * `unshift`, `unzip`, `unzipWith`, `update`, `updateWith`, `values`,
     * `valuesIn`, `without`, `wrap`, `xor`, `xorBy`, `xorWith`, `zip`,
     * `zipObject`, `zipObjectDeep`, and `zipWith`
     *
     * The wrapper methods that are **not** chainable by default are:
     * `add`, `attempt`, `camelCase`, `capitalize`, `ceil`, `clamp`, `clone`,
     * `cloneDeep`, `cloneDeepWith`, `cloneWith`, `conformsTo`, `deburr`,
     * `defaultTo`, `divide`, `each`, `eachRight`, `endsWith`, `eq`, `escape`,
     * `escapeRegExp`, `every`, `find`, `findIndex`, `findKey`, `findLast`,
     * `findLastIndex`, `findLastKey`, `first`, `floor`, `forEach`, `forEachRight`,
     * `forIn`, `forInRight`, `forOwn`, `forOwnRight`, `get`, `gt`, `gte`, `has`,
     * `hasIn`, `head`, `identity`, `includes`, `indexOf`, `inRange`, `invoke`,
     * `isArguments`, `isArray`, `isArrayBuffer`, `isArrayLike`, `isArrayLikeObject`,
     * `isBoolean`, `isBuffer`, `isDate`, `isElement`, `isEmpty`, `isEqual`,
     * `isEqualWith`, `isError`, `isFinite`, `isFunction`, `isInteger`, `isLength`,
     * `isMap`, `isMatch`, `isMatchWith`, `isNaN`, `isNative`, `isNil`, `isNull`,
     * `isNumber`, `isObject`, `isObjectLike`, `isPlainObject`, `isRegExp`,
     * `isSafeInteger`, `isSet`, `isString`, `isUndefined`, `isTypedArray`,
     * `isWeakMap`, `isWeakSet`, `join`, `kebabCase`, `last`, `lastIndexOf`,
     * `lowerCase`, `lowerFirst`, `lt`, `lte`, `max`, `maxBy`, `mean`, `meanBy`,
     * `min`, `minBy`, `multiply`, `noConflict`, `noop`, `now`, `nth`, `pad`,
     * `padEnd`, `padStart`, `parseInt`, `pop`, `random`, `reduce`, `reduceRight`,
     * `repeat`, `result`, `round`, `runInContext`, `sample`, `shift`, `size`,
     * `snakeCase`, `some`, `sortedIndex`, `sortedIndexBy`, `sortedLastIndex`,
     * `sortedLastIndexBy`, `startCase`, `startsWith`, `stubArray`, `stubFalse`,
     * `stubObject`, `stubString`, `stubTrue`, `subtract`, `sum`, `sumBy`,
     * `template`, `times`, `toFinite`, `toInteger`, `toJSON`, `toLength`,
     * `toLower`, `toNumber`, `toSafeInteger`, `toString`, `toUpper`, `trim`,
     * `trimEnd`, `trimStart`, `truncate`, `unescape`, `uniqueId`, `upperCase`,
     * `upperFirst`, `value`, and `words`
     *
     * @name _
     * @constructor
     * @category Seq
     * @param {*} value The value to wrap in a `lodash` instance.
     * @returns {Object} Returns the new `lodash` wrapper instance.
     * @example
     *
     * function square(n) {
     *   return n * n;
     * }
     *
     * var wrapped = _([1, 2, 3]);
     *
     * // Returns an unwrapped value.
     * wrapped.reduce(_.add);
     * // => 6
     *
     * // Returns a wrapped value.
     * var squares = wrapped.map(square);
     *
     * _.isArray(squares);
     * // => false
     *
     * _.isArray(squares.value());
     * // => true
     */
    function lodash(value) {
      if (isObjectLike(value) && !isArray(value) && !(value instanceof LazyWrapper)) {
        if (value instanceof LodashWrapper) {
          return value;
        }
        if (hasOwnProperty.call(value, '__wrapped__')) {
          return wrapperClone(value);
        }
      }
      return new LodashWrapper(value);
    }

    /**
     * The base implementation of `_.create` without support for assigning
     * properties to the created object.
     *
     * @private
     * @param {Object} proto The object to inherit from.
     * @returns {Object} Returns the new object.
     */
    var baseCreate = (function() {
      function object() {}
      return function(proto) {
        if (!isObject(proto)) {
          return {};
        }
        if (objectCreate) {
          return objectCreate(proto);
        }
        object.prototype = proto;
        var result = new object;
        object.prototype = undefined;
        return result;
      };
    }());

    /**
     * The function whose prototype chain sequence wrappers inherit from.
     *
     * @private
     */
    function baseLodash() {
      // No operation performed.
    }

    /**
     * The base constructor for creating `lodash` wrapper objects.
     *
     * @private
     * @param {*} value The value to wrap.
     * @param {boolean} [chainAll] Enable explicit method chain sequences.
     */
    function LodashWrapper(value, chainAll) {
      this.__wrapped__ = value;
      this.__actions__ = [];
      this.__chain__ = !!chainAll;
      this.__index__ = 0;
      this.__values__ = undefined;
    }

    /**
     * By default, the template delimiters used by lodash are like those in
     * embedded Ruby (ERB) as well as ES2015 template strings. Change the
     * following template settings to use alternative delimiters.
     *
     * @static
     * @memberOf _
     * @type {Object}
     */
    lodash.templateSettings = {

      /**
       * Used to detect `data` property values to be HTML-escaped.
       *
       * @memberOf _.templateSettings
       * @type {RegExp}
       */
      'escape': reEscape,

      /**
       * Used to detect code to be evaluated.
       *
       * @memberOf _.templateSettings
       * @type {RegExp}
       */
      'evaluate': reEvaluate,

      /**
       * Used to detect `data` property values to inject.
       *
       * @memberOf _.templateSettings
       * @type {RegExp}
       */
      'interpolate': reInterpolate,

      /**
       * Used to reference the data object in the template text.
       *
       * @memberOf _.templateSettings
       * @type {string}
       */
      'variable': '',

      /**
       * Used to import variables into the compiled template.
       *
       * @memberOf _.templateSettings
       * @type {Object}
       */
      'imports': {

        /**
         * A reference to the `lodash` function.
         *
         * @memberOf _.templateSettings.imports
         * @type {Function}
         */
        '_': lodash
      }
    };

    // Ensure wrappers are instances of `baseLodash`.
    lodash.prototype = baseLodash.prototype;
    lodash.prototype.constructor = lodash;

    LodashWrapper.prototype = baseCreate(baseLodash.prototype);
    LodashWrapper.prototype.constructor = LodashWrapper;

    /*------------------------------------------------------------------------*/

    /**
     * Creates a lazy wrapper object which wraps `value` to enable lazy evaluation.
     *
     * @private
     * @constructor
     * @param {*} value The value to wrap.
     */
    function LazyWrapper(value) {
      this.__wrapped__ = value;
      this.__actions__ = [];
      this.__dir__ = 1;
      this.__filtered__ = false;
      this.__iteratees__ = [];
      this.__takeCount__ = MAX_ARRAY_LENGTH;
      this.__views__ = [];
    }

    /**
     * Creates a clone of the lazy wrapper object.
     *
     * @private
     * @name clone
     * @memberOf LazyWrapper
     * @returns {Object} Returns the cloned `LazyWrapper` object.
     */
    function lazyClone() {
      var result = new LazyWrapper(this.__wrapped__);
      result.__actions__ = copyArray(this.__actions__);
      result.__dir__ = this.__dir__;
      result.__filtered__ = this.__filtered__;
      result.__iteratees__ = copyArray(this.__iteratees__);
      result.__takeCount__ = this.__takeCount__;
      result.__views__ = copyArray(this.__views__);
      return result;
    }

    /**
     * Reverses the direction of lazy iteration.
     *
     * @private
     * @name reverse
     * @memberOf LazyWrapper
     * @returns {Object} Returns the new reversed `LazyWrapper` object.
     */
    function lazyReverse() {
      if (this.__filtered__) {
        var result = new LazyWrapper(this);
        result.__dir__ = -1;
        result.__filtered__ = true;
      } else {
        result = this.clone();
        result.__dir__ *= -1;
      }
      return result;
    }

    /**
     * Extracts the unwrapped value from its lazy wrapper.
     *
     * @private
     * @name value
     * @memberOf LazyWrapper
     * @returns {*} Returns the unwrapped value.
     */
    function lazyValue() {
      var array = this.__wrapped__.value(),
          dir = this.__dir__,
          isArr = isArray(array),
          isRight = dir < 0,
          arrLength = isArr ? array.length : 0,
          view = getView(0, arrLength, this.__views__),
          start = view.start,
          end = view.end,
          length = end - start,
          index = isRight ? end : (start - 1),
          iteratees = this.__iteratees__,
          iterLength = iteratees.length,
          resIndex = 0,
          takeCount = nativeMin(length, this.__takeCount__);

      if (!isArr || (!isRight && arrLength == length && takeCount == length)) {
        return baseWrapperValue(array, this.__actions__);
      }
      var result = [];

      outer:
      while (length-- && resIndex < takeCount) {
        index += dir;

        var iterIndex = -1,
            value = array[index];

        while (++iterIndex < iterLength) {
          var data = iteratees[iterIndex],
              iteratee = data.iteratee,
              type = data.type,
              computed = iteratee(value);

          if (type == LAZY_MAP_FLAG) {
            value = computed;
          } else if (!computed) {
            if (type == LAZY_FILTER_FLAG) {
              continue outer;
            } else {
              break outer;
            }
          }
        }
        result[resIndex++] = value;
      }
      return result;
    }

    // Ensure `LazyWrapper` is an instance of `baseLodash`.
    LazyWrapper.prototype = baseCreate(baseLodash.prototype);
    LazyWrapper.prototype.constructor = LazyWrapper;

    /*------------------------------------------------------------------------*/

    /**
     * Creates a hash object.
     *
     * @private
     * @constructor
     * @param {Array} [entries] The key-value pairs to cache.
     */
    function Hash(entries) {
      var index = -1,
          length = entries == null ? 0 : entries.length;

      this.clear();
      while (++index < length) {
        var entry = entries[index];
        this.set(entry[0], entry[1]);
      }
    }

    /**
     * Removes all key-value entries from the hash.
     *
     * @private
     * @name clear
     * @memberOf Hash
     */
    function hashClear() {
      this.__data__ = nativeCreate ? nativeCreate(null) : {};
      this.size = 0;
    }

    /**
     * Removes `key` and its value from the hash.
     *
     * @private
     * @name delete
     * @memberOf Hash
     * @param {Object} hash The hash to modify.
     * @param {string} key The key of the value to remove.
     * @returns {boolean} Returns `true` if the entry was removed, else `false`.
     */
    function hashDelete(key) {
      var result = this.has(key) && delete this.__data__[key];
      this.size -= result ? 1 : 0;
      return result;
    }

    /**
     * Gets the hash value for `key`.
     *
     * @private
     * @name get
     * @memberOf Hash
     * @param {string} key The key of the value to get.
     * @returns {*} Returns the entry value.
     */
    function hashGet(key) {
      var data = this.__data__;
      if (nativeCreate) {
        var result = data[key];
        return result === HASH_UNDEFINED ? undefined : result;
      }
      return hasOwnProperty.call(data, key) ? data[key] : undefined;
    }

    /**
     * Checks if a hash value for `key` exists.
     *
     * @private
     * @name has
     * @memberOf Hash
     * @param {string} key The key of the entry to check.
     * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
     */
    function hashHas(key) {
      var data = this.__data__;
      return nativeCreate ? (data[key] !== undefined) : hasOwnProperty.call(data, key);
    }

    /**
     * Sets the hash `key` to `value`.
     *
     * @private
     * @name set
     * @memberOf Hash
     * @param {string} key The key of the value to set.
     * @param {*} value The value to set.
     * @returns {Object} Returns the hash instance.
     */
    function hashSet(key, value) {
      var data = this.__data__;
      this.size += this.has(key) ? 0 : 1;
      data[key] = (nativeCreate && value === undefined) ? HASH_UNDEFINED : value;
      return this;
    }

    // Add methods to `Hash`.
    Hash.prototype.clear = hashClear;
    Hash.prototype['delete'] = hashDelete;
    Hash.prototype.get = hashGet;
    Hash.prototype.has = hashHas;
    Hash.prototype.set = hashSet;

    /*------------------------------------------------------------------------*/

    /**
     * Creates an list cache object.
     *
     * @private
     * @constructor
     * @param {Array} [entries] The key-value pairs to cache.
     */
    function ListCache(entries) {
      var index = -1,
          length = entries == null ? 0 : entries.length;

      this.clear();
      while (++index < length) {
        var entry = entries[index];
        this.set(entry[0], entry[1]);
      }
    }

    /**
     * Removes all key-value entries from the list cache.
     *
     * @private
     * @name clear
     * @memberOf ListCache
     */
    function listCacheClear() {
      this.__data__ = [];
      this.size = 0;
    }

    /**
     * Removes `key` and its value from the list cache.
     *
     * @private
     * @name delete
     * @memberOf ListCache
     * @param {string} key The key of the value to remove.
     * @returns {boolean} Returns `true` if the entry was removed, else `false`.
     */
    function listCacheDelete(key) {
      var data = this.__data__,
          index = assocIndexOf(data, key);

      if (index < 0) {
        return false;
      }
      var lastIndex = data.length - 1;
      if (index == lastIndex) {
        data.pop();
      } else {
        splice.call(data, index, 1);
      }
      --this.size;
      return true;
    }

    /**
     * Gets the list cache value for `key`.
     *
     * @private
     * @name get
     * @memberOf ListCache
     * @param {string} key The key of the value to get.
     * @returns {*} Returns the entry value.
     */
    function listCacheGet(key) {
      var data = this.__data__,
          index = assocIndexOf(data, key);

      return index < 0 ? undefined : data[index][1];
    }

    /**
     * Checks if a list cache value for `key` exists.
     *
     * @private
     * @name has
     * @memberOf ListCache
     * @param {string} key The key of the entry to check.
     * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
     */
    function listCacheHas(key) {
      return assocIndexOf(this.__data__, key) > -1;
    }

    /**
     * Sets the list cache `key` to `value`.
     *
     * @private
     * @name set
     * @memberOf ListCache
     * @param {string} key The key of the value to set.
     * @param {*} value The value to set.
     * @returns {Object} Returns the list cache instance.
     */
    function listCacheSet(key, value) {
      var data = this.__data__,
          index = assocIndexOf(data, key);

      if (index < 0) {
        ++this.size;
        data.push([key, value]);
      } else {
        data[index][1] = value;
      }
      return this;
    }

    // Add methods to `ListCache`.
    ListCache.prototype.clear = listCacheClear;
    ListCache.prototype['delete'] = listCacheDelete;
    ListCache.prototype.get = listCacheGet;
    ListCache.prototype.has = listCacheHas;
    ListCache.prototype.set = listCacheSet;

    /*------------------------------------------------------------------------*/

    /**
     * Creates a map cache object to store key-value pairs.
     *
     * @private
     * @constructor
     * @param {Array} [entries] The key-value pairs to cache.
     */
    function MapCache(entries) {
      var index = -1,
          length = entries == null ? 0 : entries.length;

      this.clear();
      while (++index < length) {
        var entry = entries[index];
        this.set(entry[0], entry[1]);
      }
    }

    /**
     * Removes all key-value entries from the map.
     *
     * @private
     * @name clear
     * @memberOf MapCache
     */
    function mapCacheClear() {
      this.size = 0;
      this.__data__ = {
        'hash': new Hash,
        'map': new (Map || ListCache),
        'string': new Hash
      };
    }

    /**
     * Removes `key` and its value from the map.
     *
     * @private
     * @name delete
     * @memberOf MapCache
     * @param {string} key The key of the value to remove.
     * @returns {boolean} Returns `true` if the entry was removed, else `false`.
     */
    function mapCacheDelete(key) {
      var result = getMapData(this, key)['delete'](key);
      this.size -= result ? 1 : 0;
      return result;
    }

    /**
     * Gets the map value for `key`.
     *
     * @private
     * @name get
     * @memberOf MapCache
     * @param {string} key The key of the value to get.
     * @returns {*} Returns the entry value.
     */
    function mapCacheGet(key) {
      return getMapData(this, key).get(key);
    }

    /**
     * Checks if a map value for `key` exists.
     *
     * @private
     * @name has
     * @memberOf MapCache
     * @param {string} key The key of the entry to check.
     * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
     */
    function mapCacheHas(key) {
      return getMapData(this, key).has(key);
    }

    /**
     * Sets the map `key` to `value`.
     *
     * @private
     * @name set
     * @memberOf MapCache
     * @param {string} key The key of the value to set.
     * @param {*} value The value to set.
     * @returns {Object} Returns the map cache instance.
     */
    function mapCacheSet(key, value) {
      var data = getMapData(this, key),
          size = data.size;

      data.set(key, value);
      this.size += data.size == size ? 0 : 1;
      return this;
    }

    // Add methods to `MapCache`.
    MapCache.prototype.clear = mapCacheClear;
    MapCache.prototype['delete'] = mapCacheDelete;
    MapCache.prototype.get = mapCacheGet;
    MapCache.prototype.has = mapCacheHas;
    MapCache.prototype.set = mapCacheSet;

    /*------------------------------------------------------------------------*/

    /**
     *
     * Creates an array cache object to store unique values.
     *
     * @private
     * @constructor
     * @param {Array} [values] The values to cache.
     */
    function SetCache(values) {
      var index = -1,
          length = values == null ? 0 : values.length;

      this.__data__ = new MapCache;
      while (++index < length) {
        this.add(values[index]);
      }
    }

    /**
     * Adds `value` to the array cache.
     *
     * @private
     * @name add
     * @memberOf SetCache
     * @alias push
     * @param {*} value The value to cache.
     * @returns {Object} Returns the cache instance.
     */
    function setCacheAdd(value) {
      this.__data__.set(value, HASH_UNDEFINED);
      return this;
    }

    /**
     * Checks if `value` is in the array cache.
     *
     * @private
     * @name has
     * @memberOf SetCache
     * @param {*} value The value to search for.
     * @returns {number} Returns `true` if `value` is found, else `false`.
     */
    function setCacheHas(value) {
      return this.__data__.has(value);
    }

    // Add methods to `SetCache`.
    SetCache.prototype.add = SetCache.prototype.push = setCacheAdd;
    SetCache.prototype.has = setCacheHas;

    /*------------------------------------------------------------------------*/

    /**
     * Creates a stack cache object to store key-value pairs.
     *
     * @private
     * @constructor
     * @param {Array} [entries] The key-value pairs to cache.
     */
    function Stack(entries) {
      var data = this.__data__ = new ListCache(entries);
      this.size = data.size;
    }

    /**
     * Removes all key-value entries from the stack.
     *
     * @private
     * @name clear
     * @memberOf Stack
     */
    function stackClear() {
      this.__data__ = new ListCache;
      this.size = 0;
    }

    /**
     * Removes `key` and its value from the stack.
     *
     * @private
     * @name delete
     * @memberOf Stack
     * @param {string} key The key of the value to remove.
     * @returns {boolean} Returns `true` if the entry was removed, else `false`.
     */
    function stackDelete(key) {
      var data = this.__data__,
          result = data['delete'](key);

      this.size = data.size;
      return result;
    }

    /**
     * Gets the stack value for `key`.
     *
     * @private
     * @name get
     * @memberOf Stack
     * @param {string} key The key of the value to get.
     * @returns {*} Returns the entry value.
     */
    function stackGet(key) {
      return this.__data__.get(key);
    }

    /**
     * Checks if a stack value for `key` exists.
     *
     * @private
     * @name has
     * @memberOf Stack
     * @param {string} key The key of the entry to check.
     * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
     */
    function stackHas(key) {
      return this.__data__.has(key);
    }

    /**
     * Sets the stack `key` to `value`.
     *
     * @private
     * @name set
     * @memberOf Stack
     * @param {string} key The key of the value to set.
     * @param {*} value The value to set.
     * @returns {Object} Returns the stack cache instance.
     */
    function stackSet(key, value) {
      var data = this.__data__;
      if (data instanceof ListCache) {
        var pairs = data.__data__;
        if (!Map || (pairs.length < LARGE_ARRAY_SIZE - 1)) {
          pairs.push([key, value]);
          this.size = ++data.size;
          return this;
        }
        data = this.__data__ = new MapCache(pairs);
      }
      data.set(key, value);
      this.size = data.size;
      return this;
    }

    // Add methods to `Stack`.
    Stack.prototype.clear = stackClear;
    Stack.prototype['delete'] = stackDelete;
    Stack.prototype.get = stackGet;
    Stack.prototype.has = stackHas;
    Stack.prototype.set = stackSet;

    /*------------------------------------------------------------------------*/

    /**
     * Creates an array of the enumerable property names of the array-like `value`.
     *
     * @private
     * @param {*} value The value to query.
     * @param {boolean} inherited Specify returning inherited property names.
     * @returns {Array} Returns the array of property names.
     */
    function arrayLikeKeys(value, inherited) {
      var isArr = isArray(value),
          isArg = !isArr && isArguments(value),
          isBuff = !isArr && !isArg && isBuffer(value),
          isType = !isArr && !isArg && !isBuff && isTypedArray(value),
          skipIndexes = isArr || isArg || isBuff || isType,
          result = skipIndexes ? baseTimes(value.length, String) : [],
          length = result.length;

      for (var key in value) {
        if ((inherited || hasOwnProperty.call(value, key)) &&
            !(skipIndexes && (
               // Safari 9 has enumerable `arguments.length` in strict mode.
               key == 'length' ||
               // Node.js 0.10 has enumerable non-index properties on buffers.
               (isBuff && (key == 'offset' || key == 'parent')) ||
               // PhantomJS 2 has enumerable non-index properties on typed arrays.
               (isType && (key == 'buffer' || key == 'byteLength' || key == 'byteOffset')) ||
               // Skip index properties.
               isIndex(key, length)
            ))) {
          result.push(key);
        }
      }
      return result;
    }

    /**
     * A specialized version of `_.sample` for arrays.
     *
     * @private
     * @param {Array} array The array to sample.
     * @returns {*} Returns the random element.
     */
    function arraySample(array) {
      var length = array.length;
      return length ? array[baseRandom(0, length - 1)] : undefined;
    }

    /**
     * A specialized version of `_.sampleSize` for arrays.
     *
     * @private
     * @param {Array} array The array to sample.
     * @param {number} n The number of elements to sample.
     * @returns {Array} Returns the random elements.
     */
    function arraySampleSize(array, n) {
      return shuffleSelf(copyArray(array), baseClamp(n, 0, array.length));
    }

    /**
     * A specialized version of `_.shuffle` for arrays.
     *
     * @private
     * @param {Array} array The array to shuffle.
     * @returns {Array} Returns the new shuffled array.
     */
    function arrayShuffle(array) {
      return shuffleSelf(copyArray(array));
    }

    /**
     * This function is like `assignValue` except that it doesn't assign
     * `undefined` values.
     *
     * @private
     * @param {Object} object The object to modify.
     * @param {string} key The key of the property to assign.
     * @param {*} value The value to assign.
     */
    function assignMergeValue(object, key, value) {
      if ((value !== undefined && !eq(object[key], value)) ||
          (value === undefined && !(key in object))) {
        baseAssignValue(object, key, value);
      }
    }

    /**
     * Assigns `value` to `key` of `object` if the existing value is not equivalent
     * using [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
     * for equality comparisons.
     *
     * @private
     * @param {Object} object The object to modify.
     * @param {string} key The key of the property to assign.
     * @param {*} value The value to assign.
     */
    function assignValue(object, key, value) {
      var objValue = object[key];
      if (!(hasOwnProperty.call(object, key) && eq(objValue, value)) ||
          (value === undefined && !(key in object))) {
        baseAssignValue(object, key, value);
      }
    }

    /**
     * Gets the index at which the `key` is found in `array` of key-value pairs.
     *
     * @private
     * @param {Array} array The array to inspect.
     * @param {*} key The key to search for.
     * @returns {number} Returns the index of the matched value, else `-1`.
     */
    function assocIndexOf(array, key) {
      var length = array.length;
      while (length--) {
        if (eq(array[length][0], key)) {
          return length;
        }
      }
      return -1;
    }

    /**
     * Aggregates elements of `collection` on `accumulator` with keys transformed
     * by `iteratee` and values set by `setter`.
     *
     * @private
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} setter The function to set `accumulator` values.
     * @param {Function} iteratee The iteratee to transform keys.
     * @param {Object} accumulator The initial aggregated object.
     * @returns {Function} Returns `accumulator`.
     */
    function baseAggregator(collection, setter, iteratee, accumulator) {
      baseEach(collection, function(value, key, collection) {
        setter(accumulator, value, iteratee(value), collection);
      });
      return accumulator;
    }

    /**
     * The base implementation of `_.assign` without support for multiple sources
     * or `customizer` functions.
     *
     * @private
     * @param {Object} object The destination object.
     * @param {Object} source The source object.
     * @returns {Object} Returns `object`.
     */
    function baseAssign(object, source) {
      return object && copyObject(source, keys(source), object);
    }

    /**
     * The base implementation of `_.assignIn` without support for multiple sources
     * or `customizer` functions.
     *
     * @private
     * @param {Object} object The destination object.
     * @param {Object} source The source object.
     * @returns {Object} Returns `object`.
     */
    function baseAssignIn(object, source) {
      return object && copyObject(source, keysIn(source), object);
    }

    /**
     * The base implementation of `assignValue` and `assignMergeValue` without
     * value checks.
     *
     * @private
     * @param {Object} object The object to modify.
     * @param {string} key The key of the property to assign.
     * @param {*} value The value to assign.
     */
    function baseAssignValue(object, key, value) {
      if (key == '__proto__' && defineProperty) {
        defineProperty(object, key, {
          'configurable': true,
          'enumerable': true,
          'value': value,
          'writable': true
        });
      } else {
        object[key] = value;
      }
    }

    /**
     * The base implementation of `_.at` without support for individual paths.
     *
     * @private
     * @param {Object} object The object to iterate over.
     * @param {string[]} paths The property paths to pick.
     * @returns {Array} Returns the picked elements.
     */
    function baseAt(object, paths) {
      var index = -1,
          length = paths.length,
          result = Array(length),
          skip = object == null;

      while (++index < length) {
        result[index] = skip ? undefined : get(object, paths[index]);
      }
      return result;
    }

    /**
     * The base implementation of `_.clamp` which doesn't coerce arguments.
     *
     * @private
     * @param {number} number The number to clamp.
     * @param {number} [lower] The lower bound.
     * @param {number} upper The upper bound.
     * @returns {number} Returns the clamped number.
     */
    function baseClamp(number, lower, upper) {
      if (number === number) {
        if (upper !== undefined) {
          number = number <= upper ? number : upper;
        }
        if (lower !== undefined) {
          number = number >= lower ? number : lower;
        }
      }
      return number;
    }

    /**
     * The base implementation of `_.clone` and `_.cloneDeep` which tracks
     * traversed objects.
     *
     * @private
     * @param {*} value The value to clone.
     * @param {boolean} bitmask The bitmask flags.
     *  1 - Deep clone
     *  2 - Flatten inherited properties
     *  4 - Clone symbols
     * @param {Function} [customizer] The function to customize cloning.
     * @param {string} [key] The key of `value`.
     * @param {Object} [object] The parent object of `value`.
     * @param {Object} [stack] Tracks traversed objects and their clone counterparts.
     * @returns {*} Returns the cloned value.
     */
    function baseClone(value, bitmask, customizer, key, object, stack) {
      var result,
          isDeep = bitmask & CLONE_DEEP_FLAG,
          isFlat = bitmask & CLONE_FLAT_FLAG,
          isFull = bitmask & CLONE_SYMBOLS_FLAG;

      if (customizer) {
        result = object ? customizer(value, key, object, stack) : customizer(value);
      }
      if (result !== undefined) {
        return result;
      }
      if (!isObject(value)) {
        return value;
      }
      var isArr = isArray(value);
      if (isArr) {
        result = initCloneArray(value);
        if (!isDeep) {
          return copyArray(value, result);
        }
      } else {
        var tag = getTag(value),
            isFunc = tag == funcTag || tag == genTag;

        if (isBuffer(value)) {
          return cloneBuffer(value, isDeep);
        }
        if (tag == objectTag || tag == argsTag || (isFunc && !object)) {
          result = (isFlat || isFunc) ? {} : initCloneObject(value);
          if (!isDeep) {
            return isFlat
              ? copySymbolsIn(value, baseAssignIn(result, value))
              : copySymbols(value, baseAssign(result, value));
          }
        } else {
          if (!cloneableTags[tag]) {
            return object ? value : {};
          }
          result = initCloneByTag(value, tag, isDeep);
        }
      }
      // Check for circular references and return its corresponding clone.
      stack || (stack = new Stack);
      var stacked = stack.get(value);
      if (stacked) {
        return stacked;
      }
      stack.set(value, result);

      if (isSet(value)) {
        value.forEach(function(subValue) {
          result.add(baseClone(subValue, bitmask, customizer, subValue, value, stack));
        });
      } else if (isMap(value)) {
        value.forEach(function(subValue, key) {
          result.set(key, baseClone(subValue, bitmask, customizer, key, value, stack));
        });
      }

      var keysFunc = isFull
        ? (isFlat ? getAllKeysIn : getAllKeys)
        : (isFlat ? keysIn : keys);

      var props = isArr ? undefined : keysFunc(value);
      arrayEach(props || value, function(subValue, key) {
        if (props) {
          key = subValue;
          subValue = value[key];
        }
        // Recursively populate clone (susceptible to call stack limits).
        assignValue(result, key, baseClone(subValue, bitmask, customizer, key, value, stack));
      });
      return result;
    }

    /**
     * The base implementation of `_.conforms` which doesn't clone `source`.
     *
     * @private
     * @param {Object} source The object of property predicates to conform to.
     * @returns {Function} Returns the new spec function.
     */
    function baseConforms(source) {
      var props = keys(source);
      return function(object) {
        return baseConformsTo(object, source, props);
      };
    }

    /**
     * The base implementation of `_.conformsTo` which accepts `props` to check.
     *
     * @private
     * @param {Object} object The object to inspect.
     * @param {Object} source The object of property predicates to conform to.
     * @returns {boolean} Returns `true` if `object` conforms, else `false`.
     */
    function baseConformsTo(object, source, props) {
      var length = props.length;
      if (object == null) {
        return !length;
      }
      object = Object(object);
      while (length--) {
        var key = props[length],
            predicate = source[key],
            value = object[key];

        if ((value === undefined && !(key in object)) || !predicate(value)) {
          return false;
        }
      }
      return true;
    }

    /**
     * The base implementation of `_.delay` and `_.defer` which accepts `args`
     * to provide to `func`.
     *
     * @private
     * @param {Function} func The function to delay.
     * @param {number} wait The number of milliseconds to delay invocation.
     * @param {Array} args The arguments to provide to `func`.
     * @returns {number|Object} Returns the timer id or timeout object.
     */
    function baseDelay(func, wait, args) {
      if (typeof func != 'function') {
        throw new TypeError(FUNC_ERROR_TEXT);
      }
      return setTimeout(function() { func.apply(undefined, args); }, wait);
    }

    /**
     * The base implementation of methods like `_.difference` without support
     * for excluding multiple arrays or iteratee shorthands.
     *
     * @private
     * @param {Array} array The array to inspect.
     * @param {Array} values The values to exclude.
     * @param {Function} [iteratee] The iteratee invoked per element.
     * @param {Function} [comparator] The comparator invoked per element.
     * @returns {Array} Returns the new array of filtered values.
     */
    function baseDifference(array, values, iteratee, comparator) {
      var index = -1,
          includes = arrayIncludes,
          isCommon = true,
          length = array.length,
          result = [],
          valuesLength = values.length;

      if (!length) {
        return result;
      }
      if (iteratee) {
        values = arrayMap(values, baseUnary(iteratee));
      }
      if (comparator) {
        includes = arrayIncludesWith;
        isCommon = false;
      }
      else if (values.length >= LARGE_ARRAY_SIZE) {
        includes = cacheHas;
        isCommon = false;
        values = new SetCache(values);
      }
      outer:
      while (++index < length) {
        var value = array[index],
            computed = iteratee == null ? value : iteratee(value);

        value = (comparator || value !== 0) ? value : 0;
        if (isCommon && computed === computed) {
          var valuesIndex = valuesLength;
          while (valuesIndex--) {
            if (values[valuesIndex] === computed) {
              continue outer;
            }
          }
          result.push(value);
        }
        else if (!includes(values, computed, comparator)) {
          result.push(value);
        }
      }
      return result;
    }

    /**
     * The base implementation of `_.forEach` without support for iteratee shorthands.
     *
     * @private
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @returns {Array|Object} Returns `collection`.
     */
    var baseEach = createBaseEach(baseForOwn);

    /**
     * The base implementation of `_.forEachRight` without support for iteratee shorthands.
     *
     * @private
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @returns {Array|Object} Returns `collection`.
     */
    var baseEachRight = createBaseEach(baseForOwnRight, true);

    /**
     * The base implementation of `_.every` without support for iteratee shorthands.
     *
     * @private
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} predicate The function invoked per iteration.
     * @returns {boolean} Returns `true` if all elements pass the predicate check,
     *  else `false`
     */
    function baseEvery(collection, predicate) {
      var result = true;
      baseEach(collection, function(value, index, collection) {
        result = !!predicate(value, index, collection);
        return result;
      });
      return result;
    }

    /**
     * The base implementation of methods like `_.max` and `_.min` which accepts a
     * `comparator` to determine the extremum value.
     *
     * @private
     * @param {Array} array The array to iterate over.
     * @param {Function} iteratee The iteratee invoked per iteration.
     * @param {Function} comparator The comparator used to compare values.
     * @returns {*} Returns the extremum value.
     */
    function baseExtremum(array, iteratee, comparator) {
      var index = -1,
          length = array.length;

      while (++index < length) {
        var value = array[index],
            current = iteratee(value);

        if (current != null && (computed === undefined
              ? (current === current && !isSymbol(current))
              : comparator(current, computed)
            )) {
          var computed = current,
              result = value;
        }
      }
      return result;
    }

    /**
     * The base implementation of `_.fill` without an iteratee call guard.
     *
     * @private
     * @param {Array} array The array to fill.
     * @param {*} value The value to fill `array` with.
     * @param {number} [start=0] The start position.
     * @param {number} [end=array.length] The end position.
     * @returns {Array} Returns `array`.
     */
    function baseFill(array, value, start, end) {
      var length = array.length;

      start = toInteger(start);
      if (start < 0) {
        start = -start > length ? 0 : (length + start);
      }
      end = (end === undefined || end > length) ? length : toInteger(end);
      if (end < 0) {
        end += length;
      }
      end = start > end ? 0 : toLength(end);
      while (start < end) {
        array[start++] = value;
      }
      return array;
    }

    /**
     * The base implementation of `_.filter` without support for iteratee shorthands.
     *
     * @private
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} predicate The function invoked per iteration.
     * @returns {Array} Returns the new filtered array.
     */
    function baseFilter(collection, predicate) {
      var result = [];
      baseEach(collection, function(value, index, collection) {
        if (predicate(value, index, collection)) {
          result.push(value);
        }
      });
      return result;
    }

    /**
     * The base implementation of `_.flatten` with support for restricting flattening.
     *
     * @private
     * @param {Array} array The array to flatten.
     * @param {number} depth The maximum recursion depth.
     * @param {boolean} [predicate=isFlattenable] The function invoked per iteration.
     * @param {boolean} [isStrict] Restrict to values that pass `predicate` checks.
     * @param {Array} [result=[]] The initial result value.
     * @returns {Array} Returns the new flattened array.
     */
    function baseFlatten(array, depth, predicate, isStrict, result) {
      var index = -1,
          length = array.length;

      predicate || (predicate = isFlattenable);
      result || (result = []);

      while (++index < length) {
        var value = array[index];
        if (depth > 0 && predicate(value)) {
          if (depth > 1) {
            // Recursively flatten arrays (susceptible to call stack limits).
            baseFlatten(value, depth - 1, predicate, isStrict, result);
          } else {
            arrayPush(result, value);
          }
        } else if (!isStrict) {
          result[result.length] = value;
        }
      }
      return result;
    }

    /**
     * The base implementation of `baseForOwn` which iterates over `object`
     * properties returned by `keysFunc` and invokes `iteratee` for each property.
     * Iteratee functions may exit iteration early by explicitly returning `false`.
     *
     * @private
     * @param {Object} object The object to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @param {Function} keysFunc The function to get the keys of `object`.
     * @returns {Object} Returns `object`.
     */
    var baseFor = createBaseFor();

    /**
     * This function is like `baseFor` except that it iterates over properties
     * in the opposite order.
     *
     * @private
     * @param {Object} object The object to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @param {Function} keysFunc The function to get the keys of `object`.
     * @returns {Object} Returns `object`.
     */
    var baseForRight = createBaseFor(true);

    /**
     * The base implementation of `_.forOwn` without support for iteratee shorthands.
     *
     * @private
     * @param {Object} object The object to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @returns {Object} Returns `object`.
     */
    function baseForOwn(object, iteratee) {
      return object && baseFor(object, iteratee, keys);
    }

    /**
     * The base implementation of `_.forOwnRight` without support for iteratee shorthands.
     *
     * @private
     * @param {Object} object The object to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @returns {Object} Returns `object`.
     */
    function baseForOwnRight(object, iteratee) {
      return object && baseForRight(object, iteratee, keys);
    }

    /**
     * The base implementation of `_.functions` which creates an array of
     * `object` function property names filtered from `props`.
     *
     * @private
     * @param {Object} object The object to inspect.
     * @param {Array} props The property names to filter.
     * @returns {Array} Returns the function names.
     */
    function baseFunctions(object, props) {
      return arrayFilter(props, function(key) {
        return isFunction(object[key]);
      });
    }

    /**
     * The base implementation of `_.get` without support for default values.
     *
     * @private
     * @param {Object} object The object to query.
     * @param {Array|string} path The path of the property to get.
     * @returns {*} Returns the resolved value.
     */
    function baseGet(object, path) {
      path = castPath(path, object);

      var index = 0,
          length = path.length;

      while (object != null && index < length) {
        object = object[toKey(path[index++])];
      }
      return (index && index == length) ? object : undefined;
    }

    /**
     * The base implementation of `getAllKeys` and `getAllKeysIn` which uses
     * `keysFunc` and `symbolsFunc` to get the enumerable property names and
     * symbols of `object`.
     *
     * @private
     * @param {Object} object The object to query.
     * @param {Function} keysFunc The function to get the keys of `object`.
     * @param {Function} symbolsFunc The function to get the symbols of `object`.
     * @returns {Array} Returns the array of property names and symbols.
     */
    function baseGetAllKeys(object, keysFunc, symbolsFunc) {
      var result = keysFunc(object);
      return isArray(object) ? result : arrayPush(result, symbolsFunc(object));
    }

    /**
     * The base implementation of `getTag` without fallbacks for buggy environments.
     *
     * @private
     * @param {*} value The value to query.
     * @returns {string} Returns the `toStringTag`.
     */
    function baseGetTag(value) {
      if (value == null) {
        return value === undefined ? undefinedTag : nullTag;
      }
      return (symToStringTag && symToStringTag in Object(value))
        ? getRawTag(value)
        : objectToString(value);
    }

    /**
     * The base implementation of `_.gt` which doesn't coerce arguments.
     *
     * @private
     * @param {*} value The value to compare.
     * @param {*} other The other value to compare.
     * @returns {boolean} Returns `true` if `value` is greater than `other`,
     *  else `false`.
     */
    function baseGt(value, other) {
      return value > other;
    }

    /**
     * The base implementation of `_.has` without support for deep paths.
     *
     * @private
     * @param {Object} [object] The object to query.
     * @param {Array|string} key The key to check.
     * @returns {boolean} Returns `true` if `key` exists, else `false`.
     */
    function baseHas(object, key) {
      return object != null && hasOwnProperty.call(object, key);
    }

    /**
     * The base implementation of `_.hasIn` without support for deep paths.
     *
     * @private
     * @param {Object} [object] The object to query.
     * @param {Array|string} key The key to check.
     * @returns {boolean} Returns `true` if `key` exists, else `false`.
     */
    function baseHasIn(object, key) {
      return object != null && key in Object(object);
    }

    /**
     * The base implementation of `_.inRange` which doesn't coerce arguments.
     *
     * @private
     * @param {number} number The number to check.
     * @param {number} start The start of the range.
     * @param {number} end The end of the range.
     * @returns {boolean} Returns `true` if `number` is in the range, else `false`.
     */
    function baseInRange(number, start, end) {
      return number >= nativeMin(start, end) && number < nativeMax(start, end);
    }

    /**
     * The base implementation of methods like `_.intersection`, without support
     * for iteratee shorthands, that accepts an array of arrays to inspect.
     *
     * @private
     * @param {Array} arrays The arrays to inspect.
     * @param {Function} [iteratee] The iteratee invoked per element.
     * @param {Function} [comparator] The comparator invoked per element.
     * @returns {Array} Returns the new array of shared values.
     */
    function baseIntersection(arrays, iteratee, comparator) {
      var includes = comparator ? arrayIncludesWith : arrayIncludes,
          length = arrays[0].length,
          othLength = arrays.length,
          othIndex = othLength,
          caches = Array(othLength),
          maxLength = Infinity,
          result = [];

      while (othIndex--) {
        var array = arrays[othIndex];
        if (othIndex && iteratee) {
          array = arrayMap(array, baseUnary(iteratee));
        }
        maxLength = nativeMin(array.length, maxLength);
        caches[othIndex] = !comparator && (iteratee || (length >= 120 && array.length >= 120))
          ? new SetCache(othIndex && array)
          : undefined;
      }
      array = arrays[0];

      var index = -1,
          seen = caches[0];

      outer:
      while (++index < length && result.length < maxLength) {
        var value = array[index],
            computed = iteratee ? iteratee(value) : value;

        value = (comparator || value !== 0) ? value : 0;
        if (!(seen
              ? cacheHas(seen, computed)
              : includes(result, computed, comparator)
            )) {
          othIndex = othLength;
          while (--othIndex) {
            var cache = caches[othIndex];
            if (!(cache
                  ? cacheHas(cache, computed)
                  : includes(arrays[othIndex], computed, comparator))
                ) {
              continue outer;
            }
          }
          if (seen) {
            seen.push(computed);
          }
          result.push(value);
        }
      }
      return result;
    }

    /**
     * The base implementation of `_.invert` and `_.invertBy` which inverts
     * `object` with values transformed by `iteratee` and set by `setter`.
     *
     * @private
     * @param {Object} object The object to iterate over.
     * @param {Function} setter The function to set `accumulator` values.
     * @param {Function} iteratee The iteratee to transform values.
     * @param {Object} accumulator The initial inverted object.
     * @returns {Function} Returns `accumulator`.
     */
    function baseInverter(object, setter, iteratee, accumulator) {
      baseForOwn(object, function(value, key, object) {
        setter(accumulator, iteratee(value), key, object);
      });
      return accumulator;
    }

    /**
     * The base implementation of `_.invoke` without support for individual
     * method arguments.
     *
     * @private
     * @param {Object} object The object to query.
     * @param {Array|string} path The path of the method to invoke.
     * @param {Array} args The arguments to invoke the method with.
     * @returns {*} Returns the result of the invoked method.
     */
    function baseInvoke(object, path, args) {
      path = castPath(path, object);
      object = parent(object, path);
      var func = object == null ? object : object[toKey(last(path))];
      return func == null ? undefined : apply(func, object, args);
    }

    /**
     * The base implementation of `_.isArguments`.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an `arguments` object,
     */
    function baseIsArguments(value) {
      return isObjectLike(value) && baseGetTag(value) == argsTag;
    }

    /**
     * The base implementation of `_.isArrayBuffer` without Node.js optimizations.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an array buffer, else `false`.
     */
    function baseIsArrayBuffer(value) {
      return isObjectLike(value) && baseGetTag(value) == arrayBufferTag;
    }

    /**
     * The base implementation of `_.isDate` without Node.js optimizations.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a date object, else `false`.
     */
    function baseIsDate(value) {
      return isObjectLike(value) && baseGetTag(value) == dateTag;
    }

    /**
     * The base implementation of `_.isEqual` which supports partial comparisons
     * and tracks traversed objects.
     *
     * @private
     * @param {*} value The value to compare.
     * @param {*} other The other value to compare.
     * @param {boolean} bitmask The bitmask flags.
     *  1 - Unordered comparison
     *  2 - Partial comparison
     * @param {Function} [customizer] The function to customize comparisons.
     * @param {Object} [stack] Tracks traversed `value` and `other` objects.
     * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
     */
    function baseIsEqual(value, other, bitmask, customizer, stack) {
      if (value === other) {
        return true;
      }
      if (value == null || other == null || (!isObjectLike(value) && !isObjectLike(other))) {
        return value !== value && other !== other;
      }
      return baseIsEqualDeep(value, other, bitmask, customizer, baseIsEqual, stack);
    }

    /**
     * A specialized version of `baseIsEqual` for arrays and objects which performs
     * deep comparisons and tracks traversed objects enabling objects with circular
     * references to be compared.
     *
     * @private
     * @param {Object} object The object to compare.
     * @param {Object} other The other object to compare.
     * @param {number} bitmask The bitmask flags. See `baseIsEqual` for more details.
     * @param {Function} customizer The function to customize comparisons.
     * @param {Function} equalFunc The function to determine equivalents of values.
     * @param {Object} [stack] Tracks traversed `object` and `other` objects.
     * @returns {boolean} Returns `true` if the objects are equivalent, else `false`.
     */
    function baseIsEqualDeep(object, other, bitmask, customizer, equalFunc, stack) {
      var objIsArr = isArray(object),
          othIsArr = isArray(other),
          objTag = objIsArr ? arrayTag : getTag(object),
          othTag = othIsArr ? arrayTag : getTag(other);

      objTag = objTag == argsTag ? objectTag : objTag;
      othTag = othTag == argsTag ? objectTag : othTag;

      var objIsObj = objTag == objectTag,
          othIsObj = othTag == objectTag,
          isSameTag = objTag == othTag;

      if (isSameTag && isBuffer(object)) {
        if (!isBuffer(other)) {
          return false;
        }
        objIsArr = true;
        objIsObj = false;
      }
      if (isSameTag && !objIsObj) {
        stack || (stack = new Stack);
        return (objIsArr || isTypedArray(object))
          ? equalArrays(object, other, bitmask, customizer, equalFunc, stack)
          : equalByTag(object, other, objTag, bitmask, customizer, equalFunc, stack);
      }
      if (!(bitmask & COMPARE_PARTIAL_FLAG)) {
        var objIsWrapped = objIsObj && hasOwnProperty.call(object, '__wrapped__'),
            othIsWrapped = othIsObj && hasOwnProperty.call(other, '__wrapped__');

        if (objIsWrapped || othIsWrapped) {
          var objUnwrapped = objIsWrapped ? object.value() : object,
              othUnwrapped = othIsWrapped ? other.value() : other;

          stack || (stack = new Stack);
          return equalFunc(objUnwrapped, othUnwrapped, bitmask, customizer, stack);
        }
      }
      if (!isSameTag) {
        return false;
      }
      stack || (stack = new Stack);
      return equalObjects(object, other, bitmask, customizer, equalFunc, stack);
    }

    /**
     * The base implementation of `_.isMap` without Node.js optimizations.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a map, else `false`.
     */
    function baseIsMap(value) {
      return isObjectLike(value) && getTag(value) == mapTag;
    }

    /**
     * The base implementation of `_.isMatch` without support for iteratee shorthands.
     *
     * @private
     * @param {Object} object The object to inspect.
     * @param {Object} source The object of property values to match.
     * @param {Array} matchData The property names, values, and compare flags to match.
     * @param {Function} [customizer] The function to customize comparisons.
     * @returns {boolean} Returns `true` if `object` is a match, else `false`.
     */
    function baseIsMatch(object, source, matchData, customizer) {
      var index = matchData.length,
          length = index,
          noCustomizer = !customizer;

      if (object == null) {
        return !length;
      }
      object = Object(object);
      while (index--) {
        var data = matchData[index];
        if ((noCustomizer && data[2])
              ? data[1] !== object[data[0]]
              : !(data[0] in object)
            ) {
          return false;
        }
      }
      while (++index < length) {
        data = matchData[index];
        var key = data[0],
            objValue = object[key],
            srcValue = data[1];

        if (noCustomizer && data[2]) {
          if (objValue === undefined && !(key in object)) {
            return false;
          }
        } else {
          var stack = new Stack;
          if (customizer) {
            var result = customizer(objValue, srcValue, key, object, source, stack);
          }
          if (!(result === undefined
                ? baseIsEqual(srcValue, objValue, COMPARE_PARTIAL_FLAG | COMPARE_UNORDERED_FLAG, customizer, stack)
                : result
              )) {
            return false;
          }
        }
      }
      return true;
    }

    /**
     * The base implementation of `_.isNative` without bad shim checks.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a native function,
     *  else `false`.
     */
    function baseIsNative(value) {
      if (!isObject(value) || isMasked(value)) {
        return false;
      }
      var pattern = isFunction(value) ? reIsNative : reIsHostCtor;
      return pattern.test(toSource(value));
    }

    /**
     * The base implementation of `_.isRegExp` without Node.js optimizations.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a regexp, else `false`.
     */
    function baseIsRegExp(value) {
      return isObjectLike(value) && baseGetTag(value) == regexpTag;
    }

    /**
     * The base implementation of `_.isSet` without Node.js optimizations.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a set, else `false`.
     */
    function baseIsSet(value) {
      return isObjectLike(value) && getTag(value) == setTag;
    }

    /**
     * The base implementation of `_.isTypedArray` without Node.js optimizations.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a typed array, else `false`.
     */
    function baseIsTypedArray(value) {
      return isObjectLike(value) &&
        isLength(value.length) && !!typedArrayTags[baseGetTag(value)];
    }

    /**
     * The base implementation of `_.iteratee`.
     *
     * @private
     * @param {*} [value=_.identity] The value to convert to an iteratee.
     * @returns {Function} Returns the iteratee.
     */
    function baseIteratee(value) {
      // Don't store the `typeof` result in a variable to avoid a JIT bug in Safari 9.
      // See https://bugs.webkit.org/show_bug.cgi?id=156034 for more details.
      if (typeof value == 'function') {
        return value;
      }
      if (value == null) {
        return identity;
      }
      if (typeof value == 'object') {
        return isArray(value)
          ? baseMatchesProperty(value[0], value[1])
          : baseMatches(value);
      }
      return property(value);
    }

    /**
     * The base implementation of `_.keys` which doesn't treat sparse arrays as dense.
     *
     * @private
     * @param {Object} object The object to query.
     * @returns {Array} Returns the array of property names.
     */
    function baseKeys(object) {
      if (!isPrototype(object)) {
        return nativeKeys(object);
      }
      var result = [];
      for (var key in Object(object)) {
        if (hasOwnProperty.call(object, key) && key != 'constructor') {
          result.push(key);
        }
      }
      return result;
    }

    /**
     * The base implementation of `_.keysIn` which doesn't treat sparse arrays as dense.
     *
     * @private
     * @param {Object} object The object to query.
     * @returns {Array} Returns the array of property names.
     */
    function baseKeysIn(object) {
      if (!isObject(object)) {
        return nativeKeysIn(object);
      }
      var isProto = isPrototype(object),
          result = [];

      for (var key in object) {
        if (!(key == 'constructor' && (isProto || !hasOwnProperty.call(object, key)))) {
          result.push(key);
        }
      }
      return result;
    }

    /**
     * The base implementation of `_.lt` which doesn't coerce arguments.
     *
     * @private
     * @param {*} value The value to compare.
     * @param {*} other The other value to compare.
     * @returns {boolean} Returns `true` if `value` is less than `other`,
     *  else `false`.
     */
    function baseLt(value, other) {
      return value < other;
    }

    /**
     * The base implementation of `_.map` without support for iteratee shorthands.
     *
     * @private
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @returns {Array} Returns the new mapped array.
     */
    function baseMap(collection, iteratee) {
      var index = -1,
          result = isArrayLike(collection) ? Array(collection.length) : [];

      baseEach(collection, function(value, key, collection) {
        result[++index] = iteratee(value, key, collection);
      });
      return result;
    }

    /**
     * The base implementation of `_.matches` which doesn't clone `source`.
     *
     * @private
     * @param {Object} source The object of property values to match.
     * @returns {Function} Returns the new spec function.
     */
    function baseMatches(source) {
      var matchData = getMatchData(source);
      if (matchData.length == 1 && matchData[0][2]) {
        return matchesStrictComparable(matchData[0][0], matchData[0][1]);
      }
      return function(object) {
        return object === source || baseIsMatch(object, source, matchData);
      };
    }

    /**
     * The base implementation of `_.matchesProperty` which doesn't clone `srcValue`.
     *
     * @private
     * @param {string} path The path of the property to get.
     * @param {*} srcValue The value to match.
     * @returns {Function} Returns the new spec function.
     */
    function baseMatchesProperty(path, srcValue) {
      if (isKey(path) && isStrictComparable(srcValue)) {
        return matchesStrictComparable(toKey(path), srcValue);
      }
      return function(object) {
        var objValue = get(object, path);
        return (objValue === undefined && objValue === srcValue)
          ? hasIn(object, path)
          : baseIsEqual(srcValue, objValue, COMPARE_PARTIAL_FLAG | COMPARE_UNORDERED_FLAG);
      };
    }

    /**
     * The base implementation of `_.merge` without support for multiple sources.
     *
     * @private
     * @param {Object} object The destination object.
     * @param {Object} source The source object.
     * @param {number} srcIndex The index of `source`.
     * @param {Function} [customizer] The function to customize merged values.
     * @param {Object} [stack] Tracks traversed source values and their merged
     *  counterparts.
     */
    function baseMerge(object, source, srcIndex, customizer, stack) {
      if (object === source) {
        return;
      }
      baseFor(source, function(srcValue, key) {
        stack || (stack = new Stack);
        if (isObject(srcValue)) {
          baseMergeDeep(object, source, key, srcIndex, baseMerge, customizer, stack);
        }
        else {
          var newValue = customizer
            ? customizer(safeGet(object, key), srcValue, (key + ''), object, source, stack)
            : undefined;

          if (newValue === undefined) {
            newValue = srcValue;
          }
          assignMergeValue(object, key, newValue);
        }
      }, keysIn);
    }

    /**
     * A specialized version of `baseMerge` for arrays and objects which performs
     * deep merges and tracks traversed objects enabling objects with circular
     * references to be merged.
     *
     * @private
     * @param {Object} object The destination object.
     * @param {Object} source The source object.
     * @param {string} key The key of the value to merge.
     * @param {number} srcIndex The index of `source`.
     * @param {Function} mergeFunc The function to merge values.
     * @param {Function} [customizer] The function to customize assigned values.
     * @param {Object} [stack] Tracks traversed source values and their merged
     *  counterparts.
     */
    function baseMergeDeep(object, source, key, srcIndex, mergeFunc, customizer, stack) {
      var objValue = safeGet(object, key),
          srcValue = safeGet(source, key),
          stacked = stack.get(srcValue);

      if (stacked) {
        assignMergeValue(object, key, stacked);
        return;
      }
      var newValue = customizer
        ? customizer(objValue, srcValue, (key + ''), object, source, stack)
        : undefined;

      var isCommon = newValue === undefined;

      if (isCommon) {
        var isArr = isArray(srcValue),
            isBuff = !isArr && isBuffer(srcValue),
            isTyped = !isArr && !isBuff && isTypedArray(srcValue);

        newValue = srcValue;
        if (isArr || isBuff || isTyped) {
          if (isArray(objValue)) {
            newValue = objValue;
          }
          else if (isArrayLikeObject(objValue)) {
            newValue = copyArray(objValue);
          }
          else if (isBuff) {
            isCommon = false;
            newValue = cloneBuffer(srcValue, true);
          }
          else if (isTyped) {
            isCommon = false;
            newValue = cloneTypedArray(srcValue, true);
          }
          else {
            newValue = [];
          }
        }
        else if (isPlainObject(srcValue) || isArguments(srcValue)) {
          newValue = objValue;
          if (isArguments(objValue)) {
            newValue = toPlainObject(objValue);
          }
          else if (!isObject(objValue) || isFunction(objValue)) {
            newValue = initCloneObject(srcValue);
          }
        }
        else {
          isCommon = false;
        }
      }
      if (isCommon) {
        // Recursively merge objects and arrays (susceptible to call stack limits).
        stack.set(srcValue, newValue);
        mergeFunc(newValue, srcValue, srcIndex, customizer, stack);
        stack['delete'](srcValue);
      }
      assignMergeValue(object, key, newValue);
    }

    /**
     * The base implementation of `_.nth` which doesn't coerce arguments.
     *
     * @private
     * @param {Array} array The array to query.
     * @param {number} n The index of the element to return.
     * @returns {*} Returns the nth element of `array`.
     */
    function baseNth(array, n) {
      var length = array.length;
      if (!length) {
        return;
      }
      n += n < 0 ? length : 0;
      return isIndex(n, length) ? array[n] : undefined;
    }

    /**
     * The base implementation of `_.orderBy` without param guards.
     *
     * @private
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function[]|Object[]|string[]} iteratees The iteratees to sort by.
     * @param {string[]} orders The sort orders of `iteratees`.
     * @returns {Array} Returns the new sorted array.
     */
    function baseOrderBy(collection, iteratees, orders) {
      if (iteratees.length) {
        iteratees = arrayMap(iteratees, function(iteratee) {
          if (isArray(iteratee)) {
            return function(value) {
              return baseGet(value, iteratee.length === 1 ? iteratee[0] : iteratee);
            }
          }
          return iteratee;
        });
      } else {
        iteratees = [identity];
      }

      var index = -1;
      iteratees = arrayMap(iteratees, baseUnary(getIteratee()));

      var result = baseMap(collection, function(value, key, collection) {
        var criteria = arrayMap(iteratees, function(iteratee) {
          return iteratee(value);
        });
        return { 'criteria': criteria, 'index': ++index, 'value': value };
      });

      return baseSortBy(result, function(object, other) {
        return compareMultiple(object, other, orders);
      });
    }

    /**
     * The base implementation of `_.pick` without support for individual
     * property identifiers.
     *
     * @private
     * @param {Object} object The source object.
     * @param {string[]} paths The property paths to pick.
     * @returns {Object} Returns the new object.
     */
    function basePick(object, paths) {
      return basePickBy(object, paths, function(value, path) {
        return hasIn(object, path);
      });
    }

    /**
     * The base implementation of  `_.pickBy` without support for iteratee shorthands.
     *
     * @private
     * @param {Object} object The source object.
     * @param {string[]} paths The property paths to pick.
     * @param {Function} predicate The function invoked per property.
     * @returns {Object} Returns the new object.
     */
    function basePickBy(object, paths, predicate) {
      var index = -1,
          length = paths.length,
          result = {};

      while (++index < length) {
        var path = paths[index],
            value = baseGet(object, path);

        if (predicate(value, path)) {
          baseSet(result, castPath(path, object), value);
        }
      }
      return result;
    }

    /**
     * A specialized version of `baseProperty` which supports deep paths.
     *
     * @private
     * @param {Array|string} path The path of the property to get.
     * @returns {Function} Returns the new accessor function.
     */
    function basePropertyDeep(path) {
      return function(object) {
        return baseGet(object, path);
      };
    }

    /**
     * The base implementation of `_.pullAllBy` without support for iteratee
     * shorthands.
     *
     * @private
     * @param {Array} array The array to modify.
     * @param {Array} values The values to remove.
     * @param {Function} [iteratee] The iteratee invoked per element.
     * @param {Function} [comparator] The comparator invoked per element.
     * @returns {Array} Returns `array`.
     */
    function basePullAll(array, values, iteratee, comparator) {
      var indexOf = comparator ? baseIndexOfWith : baseIndexOf,
          index = -1,
          length = values.length,
          seen = array;

      if (array === values) {
        values = copyArray(values);
      }
      if (iteratee) {
        seen = arrayMap(array, baseUnary(iteratee));
      }
      while (++index < length) {
        var fromIndex = 0,
            value = values[index],
            computed = iteratee ? iteratee(value) : value;

        while ((fromIndex = indexOf(seen, computed, fromIndex, comparator)) > -1) {
          if (seen !== array) {
            splice.call(seen, fromIndex, 1);
          }
          splice.call(array, fromIndex, 1);
        }
      }
      return array;
    }

    /**
     * The base implementation of `_.pullAt` without support for individual
     * indexes or capturing the removed elements.
     *
     * @private
     * @param {Array} array The array to modify.
     * @param {number[]} indexes The indexes of elements to remove.
     * @returns {Array} Returns `array`.
     */
    function basePullAt(array, indexes) {
      var length = array ? indexes.length : 0,
          lastIndex = length - 1;

      while (length--) {
        var index = indexes[length];
        if (length == lastIndex || index !== previous) {
          var previous = index;
          if (isIndex(index)) {
            splice.call(array, index, 1);
          } else {
            baseUnset(array, index);
          }
        }
      }
      return array;
    }

    /**
     * The base implementation of `_.random` without support for returning
     * floating-point numbers.
     *
     * @private
     * @param {number} lower The lower bound.
     * @param {number} upper The upper bound.
     * @returns {number} Returns the random number.
     */
    function baseRandom(lower, upper) {
      return lower + nativeFloor(nativeRandom() * (upper - lower + 1));
    }

    /**
     * The base implementation of `_.range` and `_.rangeRight` which doesn't
     * coerce arguments.
     *
     * @private
     * @param {number} start The start of the range.
     * @param {number} end The end of the range.
     * @param {number} step The value to increment or decrement by.
     * @param {boolean} [fromRight] Specify iterating from right to left.
     * @returns {Array} Returns the range of numbers.
     */
    function baseRange(start, end, step, fromRight) {
      var index = -1,
          length = nativeMax(nativeCeil((end - start) / (step || 1)), 0),
          result = Array(length);

      while (length--) {
        result[fromRight ? length : ++index] = start;
        start += step;
      }
      return result;
    }

    /**
     * The base implementation of `_.repeat` which doesn't coerce arguments.
     *
     * @private
     * @param {string} string The string to repeat.
     * @param {number} n The number of times to repeat the string.
     * @returns {string} Returns the repeated string.
     */
    function baseRepeat(string, n) {
      var result = '';
      if (!string || n < 1 || n > MAX_SAFE_INTEGER) {
        return result;
      }
      // Leverage the exponentiation by squaring algorithm for a faster repeat.
      // See https://en.wikipedia.org/wiki/Exponentiation_by_squaring for more details.
      do {
        if (n % 2) {
          result += string;
        }
        n = nativeFloor(n / 2);
        if (n) {
          string += string;
        }
      } while (n);

      return result;
    }

    /**
     * The base implementation of `_.rest` which doesn't validate or coerce arguments.
     *
     * @private
     * @param {Function} func The function to apply a rest parameter to.
     * @param {number} [start=func.length-1] The start position of the rest parameter.
     * @returns {Function} Returns the new function.
     */
    function baseRest(func, start) {
      return setToString(overRest(func, start, identity), func + '');
    }

    /**
     * The base implementation of `_.sample`.
     *
     * @private
     * @param {Array|Object} collection The collection to sample.
     * @returns {*} Returns the random element.
     */
    function baseSample(collection) {
      return arraySample(values(collection));
    }

    /**
     * The base implementation of `_.sampleSize` without param guards.
     *
     * @private
     * @param {Array|Object} collection The collection to sample.
     * @param {number} n The number of elements to sample.
     * @returns {Array} Returns the random elements.
     */
    function baseSampleSize(collection, n) {
      var array = values(collection);
      return shuffleSelf(array, baseClamp(n, 0, array.length));
    }

    /**
     * The base implementation of `_.set`.
     *
     * @private
     * @param {Object} object The object to modify.
     * @param {Array|string} path The path of the property to set.
     * @param {*} value The value to set.
     * @param {Function} [customizer] The function to customize path creation.
     * @returns {Object} Returns `object`.
     */
    function baseSet(object, path, value, customizer) {
      if (!isObject(object)) {
        return object;
      }
      path = castPath(path, object);

      var index = -1,
          length = path.length,
          lastIndex = length - 1,
          nested = object;

      while (nested != null && ++index < length) {
        var key = toKey(path[index]),
            newValue = value;

        if (key === '__proto__' || key === 'constructor' || key === 'prototype') {
          return object;
        }

        if (index != lastIndex) {
          var objValue = nested[key];
          newValue = customizer ? customizer(objValue, key, nested) : undefined;
          if (newValue === undefined) {
            newValue = isObject(objValue)
              ? objValue
              : (isIndex(path[index + 1]) ? [] : {});
          }
        }
        assignValue(nested, key, newValue);
        nested = nested[key];
      }
      return object;
    }

    /**
     * The base implementation of `setData` without support for hot loop shorting.
     *
     * @private
     * @param {Function} func The function to associate metadata with.
     * @param {*} data The metadata.
     * @returns {Function} Returns `func`.
     */
    var baseSetData = !metaMap ? identity : function(func, data) {
      metaMap.set(func, data);
      return func;
    };

    /**
     * The base implementation of `setToString` without support for hot loop shorting.
     *
     * @private
     * @param {Function} func The function to modify.
     * @param {Function} string The `toString` result.
     * @returns {Function} Returns `func`.
     */
    var baseSetToString = !defineProperty ? identity : function(func, string) {
      return defineProperty(func, 'toString', {
        'configurable': true,
        'enumerable': false,
        'value': constant(string),
        'writable': true
      });
    };

    /**
     * The base implementation of `_.shuffle`.
     *
     * @private
     * @param {Array|Object} collection The collection to shuffle.
     * @returns {Array} Returns the new shuffled array.
     */
    function baseShuffle(collection) {
      return shuffleSelf(values(collection));
    }

    /**
     * The base implementation of `_.slice` without an iteratee call guard.
     *
     * @private
     * @param {Array} array The array to slice.
     * @param {number} [start=0] The start position.
     * @param {number} [end=array.length] The end position.
     * @returns {Array} Returns the slice of `array`.
     */
    function baseSlice(array, start, end) {
      var index = -1,
          length = array.length;

      if (start < 0) {
        start = -start > length ? 0 : (length + start);
      }
      end = end > length ? length : end;
      if (end < 0) {
        end += length;
      }
      length = start > end ? 0 : ((end - start) >>> 0);
      start >>>= 0;

      var result = Array(length);
      while (++index < length) {
        result[index] = array[index + start];
      }
      return result;
    }

    /**
     * The base implementation of `_.some` without support for iteratee shorthands.
     *
     * @private
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} predicate The function invoked per iteration.
     * @returns {boolean} Returns `true` if any element passes the predicate check,
     *  else `false`.
     */
    function baseSome(collection, predicate) {
      var result;

      baseEach(collection, function(value, index, collection) {
        result = predicate(value, index, collection);
        return !result;
      });
      return !!result;
    }

    /**
     * The base implementation of `_.sortedIndex` and `_.sortedLastIndex` which
     * performs a binary search of `array` to determine the index at which `value`
     * should be inserted into `array` in order to maintain its sort order.
     *
     * @private
     * @param {Array} array The sorted array to inspect.
     * @param {*} value The value to evaluate.
     * @param {boolean} [retHighest] Specify returning the highest qualified index.
     * @returns {number} Returns the index at which `value` should be inserted
     *  into `array`.
     */
    function baseSortedIndex(array, value, retHighest) {
      var low = 0,
          high = array == null ? low : array.length;

      if (typeof value == 'number' && value === value && high <= HALF_MAX_ARRAY_LENGTH) {
        while (low < high) {
          var mid = (low + high) >>> 1,
              computed = array[mid];

          if (computed !== null && !isSymbol(computed) &&
              (retHighest ? (computed <= value) : (computed < value))) {
            low = mid + 1;
          } else {
            high = mid;
          }
        }
        return high;
      }
      return baseSortedIndexBy(array, value, identity, retHighest);
    }

    /**
     * The base implementation of `_.sortedIndexBy` and `_.sortedLastIndexBy`
     * which invokes `iteratee` for `value` and each element of `array` to compute
     * their sort ranking. The iteratee is invoked with one argument; (value).
     *
     * @private
     * @param {Array} array The sorted array to inspect.
     * @param {*} value The value to evaluate.
     * @param {Function} iteratee The iteratee invoked per element.
     * @param {boolean} [retHighest] Specify returning the highest qualified index.
     * @returns {number} Returns the index at which `value` should be inserted
     *  into `array`.
     */
    function baseSortedIndexBy(array, value, iteratee, retHighest) {
      var low = 0,
          high = array == null ? 0 : array.length;
      if (high === 0) {
        return 0;
      }

      value = iteratee(value);
      var valIsNaN = value !== value,
          valIsNull = value === null,
          valIsSymbol = isSymbol(value),
          valIsUndefined = value === undefined;

      while (low < high) {
        var mid = nativeFloor((low + high) / 2),
            computed = iteratee(array[mid]),
            othIsDefined = computed !== undefined,
            othIsNull = computed === null,
            othIsReflexive = computed === computed,
            othIsSymbol = isSymbol(computed);

        if (valIsNaN) {
          var setLow = retHighest || othIsReflexive;
        } else if (valIsUndefined) {
          setLow = othIsReflexive && (retHighest || othIsDefined);
        } else if (valIsNull) {
          setLow = othIsReflexive && othIsDefined && (retHighest || !othIsNull);
        } else if (valIsSymbol) {
          setLow = othIsReflexive && othIsDefined && !othIsNull && (retHighest || !othIsSymbol);
        } else if (othIsNull || othIsSymbol) {
          setLow = false;
        } else {
          setLow = retHighest ? (computed <= value) : (computed < value);
        }
        if (setLow) {
          low = mid + 1;
        } else {
          high = mid;
        }
      }
      return nativeMin(high, MAX_ARRAY_INDEX);
    }

    /**
     * The base implementation of `_.sortedUniq` and `_.sortedUniqBy` without
     * support for iteratee shorthands.
     *
     * @private
     * @param {Array} array The array to inspect.
     * @param {Function} [iteratee] The iteratee invoked per element.
     * @returns {Array} Returns the new duplicate free array.
     */
    function baseSortedUniq(array, iteratee) {
      var index = -1,
          length = array.length,
          resIndex = 0,
          result = [];

      while (++index < length) {
        var value = array[index],
            computed = iteratee ? iteratee(value) : value;

        if (!index || !eq(computed, seen)) {
          var seen = computed;
          result[resIndex++] = value === 0 ? 0 : value;
        }
      }
      return result;
    }

    /**
     * The base implementation of `_.toNumber` which doesn't ensure correct
     * conversions of binary, hexadecimal, or octal string values.
     *
     * @private
     * @param {*} value The value to process.
     * @returns {number} Returns the number.
     */
    function baseToNumber(value) {
      if (typeof value == 'number') {
        return value;
      }
      if (isSymbol(value)) {
        return NAN;
      }
      return +value;
    }

    /**
     * The base implementation of `_.toString` which doesn't convert nullish
     * values to empty strings.
     *
     * @private
     * @param {*} value The value to process.
     * @returns {string} Returns the string.
     */
    function baseToString(value) {
      // Exit early for strings to avoid a performance hit in some environments.
      if (typeof value == 'string') {
        return value;
      }
      if (isArray(value)) {
        // Recursively convert values (susceptible to call stack limits).
        return arrayMap(value, baseToString) + '';
      }
      if (isSymbol(value)) {
        return symbolToString ? symbolToString.call(value) : '';
      }
      var result = (value + '');
      return (result == '0' && (1 / value) == -INFINITY) ? '-0' : result;
    }

    /**
     * The base implementation of `_.uniqBy` without support for iteratee shorthands.
     *
     * @private
     * @param {Array} array The array to inspect.
     * @param {Function} [iteratee] The iteratee invoked per element.
     * @param {Function} [comparator] The comparator invoked per element.
     * @returns {Array} Returns the new duplicate free array.
     */
    function baseUniq(array, iteratee, comparator) {
      var index = -1,
          includes = arrayIncludes,
          length = array.length,
          isCommon = true,
          result = [],
          seen = result;

      if (comparator) {
        isCommon = false;
        includes = arrayIncludesWith;
      }
      else if (length >= LARGE_ARRAY_SIZE) {
        var set = iteratee ? null : createSet(array);
        if (set) {
          return setToArray(set);
        }
        isCommon = false;
        includes = cacheHas;
        seen = new SetCache;
      }
      else {
        seen = iteratee ? [] : result;
      }
      outer:
      while (++index < length) {
        var value = array[index],
            computed = iteratee ? iteratee(value) : value;

        value = (comparator || value !== 0) ? value : 0;
        if (isCommon && computed === computed) {
          var seenIndex = seen.length;
          while (seenIndex--) {
            if (seen[seenIndex] === computed) {
              continue outer;
            }
          }
          if (iteratee) {
            seen.push(computed);
          }
          result.push(value);
        }
        else if (!includes(seen, computed, comparator)) {
          if (seen !== result) {
            seen.push(computed);
          }
          result.push(value);
        }
      }
      return result;
    }

    /**
     * The base implementation of `_.unset`.
     *
     * @private
     * @param {Object} object The object to modify.
     * @param {Array|string} path The property path to unset.
     * @returns {boolean} Returns `true` if the property is deleted, else `false`.
     */
    function baseUnset(object, path) {
      path = castPath(path, object);
      object = parent(object, path);
      return object == null || delete object[toKey(last(path))];
    }

    /**
     * The base implementation of `_.update`.
     *
     * @private
     * @param {Object} object The object to modify.
     * @param {Array|string} path The path of the property to update.
     * @param {Function} updater The function to produce the updated value.
     * @param {Function} [customizer] The function to customize path creation.
     * @returns {Object} Returns `object`.
     */
    function baseUpdate(object, path, updater, customizer) {
      return baseSet(object, path, updater(baseGet(object, path)), customizer);
    }

    /**
     * The base implementation of methods like `_.dropWhile` and `_.takeWhile`
     * without support for iteratee shorthands.
     *
     * @private
     * @param {Array} array The array to query.
     * @param {Function} predicate The function invoked per iteration.
     * @param {boolean} [isDrop] Specify dropping elements instead of taking them.
     * @param {boolean} [fromRight] Specify iterating from right to left.
     * @returns {Array} Returns the slice of `array`.
     */
    function baseWhile(array, predicate, isDrop, fromRight) {
      var length = array.length,
          index = fromRight ? length : -1;

      while ((fromRight ? index-- : ++index < length) &&
        predicate(array[index], index, array)) {}

      return isDrop
        ? baseSlice(array, (fromRight ? 0 : index), (fromRight ? index + 1 : length))
        : baseSlice(array, (fromRight ? index + 1 : 0), (fromRight ? length : index));
    }

    /**
     * The base implementation of `wrapperValue` which returns the result of
     * performing a sequence of actions on the unwrapped `value`, where each
     * successive action is supplied the return value of the previous.
     *
     * @private
     * @param {*} value The unwrapped value.
     * @param {Array} actions Actions to perform to resolve the unwrapped value.
     * @returns {*} Returns the resolved value.
     */
    function baseWrapperValue(value, actions) {
      var result = value;
      if (result instanceof LazyWrapper) {
        result = result.value();
      }
      return arrayReduce(actions, function(result, action) {
        return action.func.apply(action.thisArg, arrayPush([result], action.args));
      }, result);
    }

    /**
     * The base implementation of methods like `_.xor`, without support for
     * iteratee shorthands, that accepts an array of arrays to inspect.
     *
     * @private
     * @param {Array} arrays The arrays to inspect.
     * @param {Function} [iteratee] The iteratee invoked per element.
     * @param {Function} [comparator] The comparator invoked per element.
     * @returns {Array} Returns the new array of values.
     */
    function baseXor(arrays, iteratee, comparator) {
      var length = arrays.length;
      if (length < 2) {
        return length ? baseUniq(arrays[0]) : [];
      }
      var index = -1,
          result = Array(length);

      while (++index < length) {
        var array = arrays[index],
            othIndex = -1;

        while (++othIndex < length) {
          if (othIndex != index) {
            result[index] = baseDifference(result[index] || array, arrays[othIndex], iteratee, comparator);
          }
        }
      }
      return baseUniq(baseFlatten(result, 1), iteratee, comparator);
    }

    /**
     * This base implementation of `_.zipObject` which assigns values using `assignFunc`.
     *
     * @private
     * @param {Array} props The property identifiers.
     * @param {Array} values The property values.
     * @param {Function} assignFunc The function to assign values.
     * @returns {Object} Returns the new object.
     */
    function baseZipObject(props, values, assignFunc) {
      var index = -1,
          length = props.length,
          valsLength = values.length,
          result = {};

      while (++index < length) {
        var value = index < valsLength ? values[index] : undefined;
        assignFunc(result, props[index], value);
      }
      return result;
    }

    /**
     * Casts `value` to an empty array if it's not an array like object.
     *
     * @private
     * @param {*} value The value to inspect.
     * @returns {Array|Object} Returns the cast array-like object.
     */
    function castArrayLikeObject(value) {
      return isArrayLikeObject(value) ? value : [];
    }

    /**
     * Casts `value` to `identity` if it's not a function.
     *
     * @private
     * @param {*} value The value to inspect.
     * @returns {Function} Returns cast function.
     */
    function castFunction(value) {
      return typeof value == 'function' ? value : identity;
    }

    /**
     * Casts `value` to a path array if it's not one.
     *
     * @private
     * @param {*} value The value to inspect.
     * @param {Object} [object] The object to query keys on.
     * @returns {Array} Returns the cast property path array.
     */
    function castPath(value, object) {
      if (isArray(value)) {
        return value;
      }
      return isKey(value, object) ? [value] : stringToPath(toString(value));
    }

    /**
     * A `baseRest` alias which can be replaced with `identity` by module
     * replacement plugins.
     *
     * @private
     * @type {Function}
     * @param {Function} func The function to apply a rest parameter to.
     * @returns {Function} Returns the new function.
     */
    var castRest = baseRest;

    /**
     * Casts `array` to a slice if it's needed.
     *
     * @private
     * @param {Array} array The array to inspect.
     * @param {number} start The start position.
     * @param {number} [end=array.length] The end position.
     * @returns {Array} Returns the cast slice.
     */
    function castSlice(array, start, end) {
      var length = array.length;
      end = end === undefined ? length : end;
      return (!start && end >= length) ? array : baseSlice(array, start, end);
    }

    /**
     * A simple wrapper around the global [`clearTimeout`](https://mdn.io/clearTimeout).
     *
     * @private
     * @param {number|Object} id The timer id or timeout object of the timer to clear.
     */
    var clearTimeout = ctxClearTimeout || function(id) {
      return root.clearTimeout(id);
    };

    /**
     * Creates a clone of  `buffer`.
     *
     * @private
     * @param {Buffer} buffer The buffer to clone.
     * @param {boolean} [isDeep] Specify a deep clone.
     * @returns {Buffer} Returns the cloned buffer.
     */
    function cloneBuffer(buffer, isDeep) {
      if (isDeep) {
        return buffer.slice();
      }
      var length = buffer.length,
          result = allocUnsafe ? allocUnsafe(length) : new buffer.constructor(length);

      buffer.copy(result);
      return result;
    }

    /**
     * Creates a clone of `arrayBuffer`.
     *
     * @private
     * @param {ArrayBuffer} arrayBuffer The array buffer to clone.
     * @returns {ArrayBuffer} Returns the cloned array buffer.
     */
    function cloneArrayBuffer(arrayBuffer) {
      var result = new arrayBuffer.constructor(arrayBuffer.byteLength);
      new Uint8Array(result).set(new Uint8Array(arrayBuffer));
      return result;
    }

    /**
     * Creates a clone of `dataView`.
     *
     * @private
     * @param {Object} dataView The data view to clone.
     * @param {boolean} [isDeep] Specify a deep clone.
     * @returns {Object} Returns the cloned data view.
     */
    function cloneDataView(dataView, isDeep) {
      var buffer = isDeep ? cloneArrayBuffer(dataView.buffer) : dataView.buffer;
      return new dataView.constructor(buffer, dataView.byteOffset, dataView.byteLength);
    }

    /**
     * Creates a clone of `regexp`.
     *
     * @private
     * @param {Object} regexp The regexp to clone.
     * @returns {Object} Returns the cloned regexp.
     */
    function cloneRegExp(regexp) {
      var result = new regexp.constructor(regexp.source, reFlags.exec(regexp));
      result.lastIndex = regexp.lastIndex;
      return result;
    }

    /**
     * Creates a clone of the `symbol` object.
     *
     * @private
     * @param {Object} symbol The symbol object to clone.
     * @returns {Object} Returns the cloned symbol object.
     */
    function cloneSymbol(symbol) {
      return symbolValueOf ? Object(symbolValueOf.call(symbol)) : {};
    }

    /**
     * Creates a clone of `typedArray`.
     *
     * @private
     * @param {Object} typedArray The typed array to clone.
     * @param {boolean} [isDeep] Specify a deep clone.
     * @returns {Object} Returns the cloned typed array.
     */
    function cloneTypedArray(typedArray, isDeep) {
      var buffer = isDeep ? cloneArrayBuffer(typedArray.buffer) : typedArray.buffer;
      return new typedArray.constructor(buffer, typedArray.byteOffset, typedArray.length);
    }

    /**
     * Compares values to sort them in ascending order.
     *
     * @private
     * @param {*} value The value to compare.
     * @param {*} other The other value to compare.
     * @returns {number} Returns the sort order indicator for `value`.
     */
    function compareAscending(value, other) {
      if (value !== other) {
        var valIsDefined = value !== undefined,
            valIsNull = value === null,
            valIsReflexive = value === value,
            valIsSymbol = isSymbol(value);

        var othIsDefined = other !== undefined,
            othIsNull = other === null,
            othIsReflexive = other === other,
            othIsSymbol = isSymbol(other);

        if ((!othIsNull && !othIsSymbol && !valIsSymbol && value > other) ||
            (valIsSymbol && othIsDefined && othIsReflexive && !othIsNull && !othIsSymbol) ||
            (valIsNull && othIsDefined && othIsReflexive) ||
            (!valIsDefined && othIsReflexive) ||
            !valIsReflexive) {
          return 1;
        }
        if ((!valIsNull && !valIsSymbol && !othIsSymbol && value < other) ||
            (othIsSymbol && valIsDefined && valIsReflexive && !valIsNull && !valIsSymbol) ||
            (othIsNull && valIsDefined && valIsReflexive) ||
            (!othIsDefined && valIsReflexive) ||
            !othIsReflexive) {
          return -1;
        }
      }
      return 0;
    }

    /**
     * Used by `_.orderBy` to compare multiple properties of a value to another
     * and stable sort them.
     *
     * If `orders` is unspecified, all values are sorted in ascending order. Otherwise,
     * specify an order of "desc" for descending or "asc" for ascending sort order
     * of corresponding values.
     *
     * @private
     * @param {Object} object The object to compare.
     * @param {Object} other The other object to compare.
     * @param {boolean[]|string[]} orders The order to sort by for each property.
     * @returns {number} Returns the sort order indicator for `object`.
     */
    function compareMultiple(object, other, orders) {
      var index = -1,
          objCriteria = object.criteria,
          othCriteria = other.criteria,
          length = objCriteria.length,
          ordersLength = orders.length;

      while (++index < length) {
        var result = compareAscending(objCriteria[index], othCriteria[index]);
        if (result) {
          if (index >= ordersLength) {
            return result;
          }
          var order = orders[index];
          return result * (order == 'desc' ? -1 : 1);
        }
      }
      // Fixes an `Array#sort` bug in the JS engine embedded in Adobe applications
      // that causes it, under certain circumstances, to provide the same value for
      // `object` and `other`. See https://github.com/jashkenas/underscore/pull/1247
      // for more details.
      //
      // This also ensures a stable sort in V8 and other engines.
      // See https://bugs.chromium.org/p/v8/issues/detail?id=90 for more details.
      return object.index - other.index;
    }

    /**
     * Creates an array that is the composition of partially applied arguments,
     * placeholders, and provided arguments into a single array of arguments.
     *
     * @private
     * @param {Array} args The provided arguments.
     * @param {Array} partials The arguments to prepend to those provided.
     * @param {Array} holders The `partials` placeholder indexes.
     * @params {boolean} [isCurried] Specify composing for a curried function.
     * @returns {Array} Returns the new array of composed arguments.
     */
    function composeArgs(args, partials, holders, isCurried) {
      var argsIndex = -1,
          argsLength = args.length,
          holdersLength = holders.length,
          leftIndex = -1,
          leftLength = partials.length,
          rangeLength = nativeMax(argsLength - holdersLength, 0),
          result = Array(leftLength + rangeLength),
          isUncurried = !isCurried;

      while (++leftIndex < leftLength) {
        result[leftIndex] = partials[leftIndex];
      }
      while (++argsIndex < holdersLength) {
        if (isUncurried || argsIndex < argsLength) {
          result[holders[argsIndex]] = args[argsIndex];
        }
      }
      while (rangeLength--) {
        result[leftIndex++] = args[argsIndex++];
      }
      return result;
    }

    /**
     * This function is like `composeArgs` except that the arguments composition
     * is tailored for `_.partialRight`.
     *
     * @private
     * @param {Array} args The provided arguments.
     * @param {Array} partials The arguments to append to those provided.
     * @param {Array} holders The `partials` placeholder indexes.
     * @params {boolean} [isCurried] Specify composing for a curried function.
     * @returns {Array} Returns the new array of composed arguments.
     */
    function composeArgsRight(args, partials, holders, isCurried) {
      var argsIndex = -1,
          argsLength = args.length,
          holdersIndex = -1,
          holdersLength = holders.length,
          rightIndex = -1,
          rightLength = partials.length,
          rangeLength = nativeMax(argsLength - holdersLength, 0),
          result = Array(rangeLength + rightLength),
          isUncurried = !isCurried;

      while (++argsIndex < rangeLength) {
        result[argsIndex] = args[argsIndex];
      }
      var offset = argsIndex;
      while (++rightIndex < rightLength) {
        result[offset + rightIndex] = partials[rightIndex];
      }
      while (++holdersIndex < holdersLength) {
        if (isUncurried || argsIndex < argsLength) {
          result[offset + holders[holdersIndex]] = args[argsIndex++];
        }
      }
      return result;
    }

    /**
     * Copies the values of `source` to `array`.
     *
     * @private
     * @param {Array} source The array to copy values from.
     * @param {Array} [array=[]] The array to copy values to.
     * @returns {Array} Returns `array`.
     */
    function copyArray(source, array) {
      var index = -1,
          length = source.length;

      array || (array = Array(length));
      while (++index < length) {
        array[index] = source[index];
      }
      return array;
    }

    /**
     * Copies properties of `source` to `object`.
     *
     * @private
     * @param {Object} source The object to copy properties from.
     * @param {Array} props The property identifiers to copy.
     * @param {Object} [object={}] The object to copy properties to.
     * @param {Function} [customizer] The function to customize copied values.
     * @returns {Object} Returns `object`.
     */
    function copyObject(source, props, object, customizer) {
      var isNew = !object;
      object || (object = {});

      var index = -1,
          length = props.length;

      while (++index < length) {
        var key = props[index];

        var newValue = customizer
          ? customizer(object[key], source[key], key, object, source)
          : undefined;

        if (newValue === undefined) {
          newValue = source[key];
        }
        if (isNew) {
          baseAssignValue(object, key, newValue);
        } else {
          assignValue(object, key, newValue);
        }
      }
      return object;
    }

    /**
     * Copies own symbols of `source` to `object`.
     *
     * @private
     * @param {Object} source The object to copy symbols from.
     * @param {Object} [object={}] The object to copy symbols to.
     * @returns {Object} Returns `object`.
     */
    function copySymbols(source, object) {
      return copyObject(source, getSymbols(source), object);
    }

    /**
     * Copies own and inherited symbols of `source` to `object`.
     *
     * @private
     * @param {Object} source The object to copy symbols from.
     * @param {Object} [object={}] The object to copy symbols to.
     * @returns {Object} Returns `object`.
     */
    function copySymbolsIn(source, object) {
      return copyObject(source, getSymbolsIn(source), object);
    }

    /**
     * Creates a function like `_.groupBy`.
     *
     * @private
     * @param {Function} setter The function to set accumulator values.
     * @param {Function} [initializer] The accumulator object initializer.
     * @returns {Function} Returns the new aggregator function.
     */
    function createAggregator(setter, initializer) {
      return function(collection, iteratee) {
        var func = isArray(collection) ? arrayAggregator : baseAggregator,
            accumulator = initializer ? initializer() : {};

        return func(collection, setter, getIteratee(iteratee, 2), accumulator);
      };
    }

    /**
     * Creates a function like `_.assign`.
     *
     * @private
     * @param {Function} assigner The function to assign values.
     * @returns {Function} Returns the new assigner function.
     */
    function createAssigner(assigner) {
      return baseRest(function(object, sources) {
        var index = -1,
            length = sources.length,
            customizer = length > 1 ? sources[length - 1] : undefined,
            guard = length > 2 ? sources[2] : undefined;

        customizer = (assigner.length > 3 && typeof customizer == 'function')
          ? (length--, customizer)
          : undefined;

        if (guard && isIterateeCall(sources[0], sources[1], guard)) {
          customizer = length < 3 ? undefined : customizer;
          length = 1;
        }
        object = Object(object);
        while (++index < length) {
          var source = sources[index];
          if (source) {
            assigner(object, source, index, customizer);
          }
        }
        return object;
      });
    }

    /**
     * Creates a `baseEach` or `baseEachRight` function.
     *
     * @private
     * @param {Function} eachFunc The function to iterate over a collection.
     * @param {boolean} [fromRight] Specify iterating from right to left.
     * @returns {Function} Returns the new base function.
     */
    function createBaseEach(eachFunc, fromRight) {
      return function(collection, iteratee) {
        if (collection == null) {
          return collection;
        }
        if (!isArrayLike(collection)) {
          return eachFunc(collection, iteratee);
        }
        var length = collection.length,
            index = fromRight ? length : -1,
            iterable = Object(collection);

        while ((fromRight ? index-- : ++index < length)) {
          if (iteratee(iterable[index], index, iterable) === false) {
            break;
          }
        }
        return collection;
      };
    }

    /**
     * Creates a base function for methods like `_.forIn` and `_.forOwn`.
     *
     * @private
     * @param {boolean} [fromRight] Specify iterating from right to left.
     * @returns {Function} Returns the new base function.
     */
    function createBaseFor(fromRight) {
      return function(object, iteratee, keysFunc) {
        var index = -1,
            iterable = Object(object),
            props = keysFunc(object),
            length = props.length;

        while (length--) {
          var key = props[fromRight ? length : ++index];
          if (iteratee(iterable[key], key, iterable) === false) {
            break;
          }
        }
        return object;
      };
    }

    /**
     * Creates a function that wraps `func` to invoke it with the optional `this`
     * binding of `thisArg`.
     *
     * @private
     * @param {Function} func The function to wrap.
     * @param {number} bitmask The bitmask flags. See `createWrap` for more details.
     * @param {*} [thisArg] The `this` binding of `func`.
     * @returns {Function} Returns the new wrapped function.
     */
    function createBind(func, bitmask, thisArg) {
      var isBind = bitmask & WRAP_BIND_FLAG,
          Ctor = createCtor(func);

      function wrapper() {
        var fn = (this && this !== root && this instanceof wrapper) ? Ctor : func;
        return fn.apply(isBind ? thisArg : this, arguments);
      }
      return wrapper;
    }

    /**
     * Creates a function like `_.lowerFirst`.
     *
     * @private
     * @param {string} methodName The name of the `String` case method to use.
     * @returns {Function} Returns the new case function.
     */
    function createCaseFirst(methodName) {
      return function(string) {
        string = toString(string);

        var strSymbols = hasUnicode(string)
          ? stringToArray(string)
          : undefined;

        var chr = strSymbols
          ? strSymbols[0]
          : string.charAt(0);

        var trailing = strSymbols
          ? castSlice(strSymbols, 1).join('')
          : string.slice(1);

        return chr[methodName]() + trailing;
      };
    }

    /**
     * Creates a function like `_.camelCase`.
     *
     * @private
     * @param {Function} callback The function to combine each word.
     * @returns {Function} Returns the new compounder function.
     */
    function createCompounder(callback) {
      return function(string) {
        return arrayReduce(words(deburr(string).replace(reApos, '')), callback, '');
      };
    }

    /**
     * Creates a function that produces an instance of `Ctor` regardless of
     * whether it was invoked as part of a `new` expression or by `call` or `apply`.
     *
     * @private
     * @param {Function} Ctor The constructor to wrap.
     * @returns {Function} Returns the new wrapped function.
     */
    function createCtor(Ctor) {
      return function() {
        // Use a `switch` statement to work with class constructors. See
        // http://ecma-international.org/ecma-262/7.0/#sec-ecmascript-function-objects-call-thisargument-argumentslist
        // for more details.
        var args = arguments;
        switch (args.length) {
          case 0: return new Ctor;
          case 1: return new Ctor(args[0]);
          case 2: return new Ctor(args[0], args[1]);
          case 3: return new Ctor(args[0], args[1], args[2]);
          case 4: return new Ctor(args[0], args[1], args[2], args[3]);
          case 5: return new Ctor(args[0], args[1], args[2], args[3], args[4]);
          case 6: return new Ctor(args[0], args[1], args[2], args[3], args[4], args[5]);
          case 7: return new Ctor(args[0], args[1], args[2], args[3], args[4], args[5], args[6]);
        }
        var thisBinding = baseCreate(Ctor.prototype),
            result = Ctor.apply(thisBinding, args);

        // Mimic the constructor's `return` behavior.
        // See https://es5.github.io/#x13.2.2 for more details.
        return isObject(result) ? result : thisBinding;
      };
    }

    /**
     * Creates a function that wraps `func` to enable currying.
     *
     * @private
     * @param {Function} func The function to wrap.
     * @param {number} bitmask The bitmask flags. See `createWrap` for more details.
     * @param {number} arity The arity of `func`.
     * @returns {Function} Returns the new wrapped function.
     */
    function createCurry(func, bitmask, arity) {
      var Ctor = createCtor(func);

      function wrapper() {
        var length = arguments.length,
            args = Array(length),
            index = length,
            placeholder = getHolder(wrapper);

        while (index--) {
          args[index] = arguments[index];
        }
        var holders = (length < 3 && args[0] !== placeholder && args[length - 1] !== placeholder)
          ? []
          : replaceHolders(args, placeholder);

        length -= holders.length;
        if (length < arity) {
          return createRecurry(
            func, bitmask, createHybrid, wrapper.placeholder, undefined,
            args, holders, undefined, undefined, arity - length);
        }
        var fn = (this && this !== root && this instanceof wrapper) ? Ctor : func;
        return apply(fn, this, args);
      }
      return wrapper;
    }

    /**
     * Creates a `_.find` or `_.findLast` function.
     *
     * @private
     * @param {Function} findIndexFunc The function to find the collection index.
     * @returns {Function} Returns the new find function.
     */
    function createFind(findIndexFunc) {
      return function(collection, predicate, fromIndex) {
        var iterable = Object(collection);
        if (!isArrayLike(collection)) {
          var iteratee = getIteratee(predicate, 3);
          collection = keys(collection);
          predicate = function(key) { return iteratee(iterable[key], key, iterable); };
        }
        var index = findIndexFunc(collection, predicate, fromIndex);
        return index > -1 ? iterable[iteratee ? collection[index] : index] : undefined;
      };
    }

    /**
     * Creates a `_.flow` or `_.flowRight` function.
     *
     * @private
     * @param {boolean} [fromRight] Specify iterating from right to left.
     * @returns {Function} Returns the new flow function.
     */
    function createFlow(fromRight) {
      return flatRest(function(funcs) {
        var length = funcs.length,
            index = length,
            prereq = LodashWrapper.prototype.thru;

        if (fromRight) {
          funcs.reverse();
        }
        while (index--) {
          var func = funcs[index];
          if (typeof func != 'function') {
            throw new TypeError(FUNC_ERROR_TEXT);
          }
          if (prereq && !wrapper && getFuncName(func) == 'wrapper') {
            var wrapper = new LodashWrapper([], true);
          }
        }
        index = wrapper ? index : length;
        while (++index < length) {
          func = funcs[index];

          var funcName = getFuncName(func),
              data = funcName == 'wrapper' ? getData(func) : undefined;

          if (data && isLaziable(data[0]) &&
                data[1] == (WRAP_ARY_FLAG | WRAP_CURRY_FLAG | WRAP_PARTIAL_FLAG | WRAP_REARG_FLAG) &&
                !data[4].length && data[9] == 1
              ) {
            wrapper = wrapper[getFuncName(data[0])].apply(wrapper, data[3]);
          } else {
            wrapper = (func.length == 1 && isLaziable(func))
              ? wrapper[funcName]()
              : wrapper.thru(func);
          }
        }
        return function() {
          var args = arguments,
              value = args[0];

          if (wrapper && args.length == 1 && isArray(value)) {
            return wrapper.plant(value).value();
          }
          var index = 0,
              result = length ? funcs[index].apply(this, args) : value;

          while (++index < length) {
            result = funcs[index].call(this, result);
          }
          return result;
        };
      });
    }

    /**
     * Creates a function that wraps `func` to invoke it with optional `this`
     * binding of `thisArg`, partial application, and currying.
     *
     * @private
     * @param {Function|string} func The function or method name to wrap.
     * @param {number} bitmask The bitmask flags. See `createWrap` for more details.
     * @param {*} [thisArg] The `this` binding of `func`.
     * @param {Array} [partials] The arguments to prepend to those provided to
     *  the new function.
     * @param {Array} [holders] The `partials` placeholder indexes.
     * @param {Array} [partialsRight] The arguments to append to those provided
     *  to the new function.
     * @param {Array} [holdersRight] The `partialsRight` placeholder indexes.
     * @param {Array} [argPos] The argument positions of the new function.
     * @param {number} [ary] The arity cap of `func`.
     * @param {number} [arity] The arity of `func`.
     * @returns {Function} Returns the new wrapped function.
     */
    function createHybrid(func, bitmask, thisArg, partials, holders, partialsRight, holdersRight, argPos, ary, arity) {
      var isAry = bitmask & WRAP_ARY_FLAG,
          isBind = bitmask & WRAP_BIND_FLAG,
          isBindKey = bitmask & WRAP_BIND_KEY_FLAG,
          isCurried = bitmask & (WRAP_CURRY_FLAG | WRAP_CURRY_RIGHT_FLAG),
          isFlip = bitmask & WRAP_FLIP_FLAG,
          Ctor = isBindKey ? undefined : createCtor(func);

      function wrapper() {
        var length = arguments.length,
            args = Array(length),
            index = length;

        while (index--) {
          args[index] = arguments[index];
        }
        if (isCurried) {
          var placeholder = getHolder(wrapper),
              holdersCount = countHolders(args, placeholder);
        }
        if (partials) {
          args = composeArgs(args, partials, holders, isCurried);
        }
        if (partialsRight) {
          args = composeArgsRight(args, partialsRight, holdersRight, isCurried);
        }
        length -= holdersCount;
        if (isCurried && length < arity) {
          var newHolders = replaceHolders(args, placeholder);
          return createRecurry(
            func, bitmask, createHybrid, wrapper.placeholder, thisArg,
            args, newHolders, argPos, ary, arity - length
          );
        }
        var thisBinding = isBind ? thisArg : this,
            fn = isBindKey ? thisBinding[func] : func;

        length = args.length;
        if (argPos) {
          args = reorder(args, argPos);
        } else if (isFlip && length > 1) {
          args.reverse();
        }
        if (isAry && ary < length) {
          args.length = ary;
        }
        if (this && this !== root && this instanceof wrapper) {
          fn = Ctor || createCtor(fn);
        }
        return fn.apply(thisBinding, args);
      }
      return wrapper;
    }

    /**
     * Creates a function like `_.invertBy`.
     *
     * @private
     * @param {Function} setter The function to set accumulator values.
     * @param {Function} toIteratee The function to resolve iteratees.
     * @returns {Function} Returns the new inverter function.
     */
    function createInverter(setter, toIteratee) {
      return function(object, iteratee) {
        return baseInverter(object, setter, toIteratee(iteratee), {});
      };
    }

    /**
     * Creates a function that performs a mathematical operation on two values.
     *
     * @private
     * @param {Function} operator The function to perform the operation.
     * @param {number} [defaultValue] The value used for `undefined` arguments.
     * @returns {Function} Returns the new mathematical operation function.
     */
    function createMathOperation(operator, defaultValue) {
      return function(value, other) {
        var result;
        if (value === undefined && other === undefined) {
          return defaultValue;
        }
        if (value !== undefined) {
          result = value;
        }
        if (other !== undefined) {
          if (result === undefined) {
            return other;
          }
          if (typeof value == 'string' || typeof other == 'string') {
            value = baseToString(value);
            other = baseToString(other);
          } else {
            value = baseToNumber(value);
            other = baseToNumber(other);
          }
          result = operator(value, other);
        }
        return result;
      };
    }

    /**
     * Creates a function like `_.over`.
     *
     * @private
     * @param {Function} arrayFunc The function to iterate over iteratees.
     * @returns {Function} Returns the new over function.
     */
    function createOver(arrayFunc) {
      return flatRest(function(iteratees) {
        iteratees = arrayMap(iteratees, baseUnary(getIteratee()));
        return baseRest(function(args) {
          var thisArg = this;
          return arrayFunc(iteratees, function(iteratee) {
            return apply(iteratee, thisArg, args);
          });
        });
      });
    }

    /**
     * Creates the padding for `string` based on `length`. The `chars` string
     * is truncated if the number of characters exceeds `length`.
     *
     * @private
     * @param {number} length The padding length.
     * @param {string} [chars=' '] The string used as padding.
     * @returns {string} Returns the padding for `string`.
     */
    function createPadding(length, chars) {
      chars = chars === undefined ? ' ' : baseToString(chars);

      var charsLength = chars.length;
      if (charsLength < 2) {
        return charsLength ? baseRepeat(chars, length) : chars;
      }
      var result = baseRepeat(chars, nativeCeil(length / stringSize(chars)));
      return hasUnicode(chars)
        ? castSlice(stringToArray(result), 0, length).join('')
        : result.slice(0, length);
    }

    /**
     * Creates a function that wraps `func` to invoke it with the `this` binding
     * of `thisArg` and `partials` prepended to the arguments it receives.
     *
     * @private
     * @param {Function} func The function to wrap.
     * @param {number} bitmask The bitmask flags. See `createWrap` for more details.
     * @param {*} thisArg The `this` binding of `func`.
     * @param {Array} partials The arguments to prepend to those provided to
     *  the new function.
     * @returns {Function} Returns the new wrapped function.
     */
    function createPartial(func, bitmask, thisArg, partials) {
      var isBind = bitmask & WRAP_BIND_FLAG,
          Ctor = createCtor(func);

      function wrapper() {
        var argsIndex = -1,
            argsLength = arguments.length,
            leftIndex = -1,
            leftLength = partials.length,
            args = Array(leftLength + argsLength),
            fn = (this && this !== root && this instanceof wrapper) ? Ctor : func;

        while (++leftIndex < leftLength) {
          args[leftIndex] = partials[leftIndex];
        }
        while (argsLength--) {
          args[leftIndex++] = arguments[++argsIndex];
        }
        return apply(fn, isBind ? thisArg : this, args);
      }
      return wrapper;
    }

    /**
     * Creates a `_.range` or `_.rangeRight` function.
     *
     * @private
     * @param {boolean} [fromRight] Specify iterating from right to left.
     * @returns {Function} Returns the new range function.
     */
    function createRange(fromRight) {
      return function(start, end, step) {
        if (step && typeof step != 'number' && isIterateeCall(start, end, step)) {
          end = step = undefined;
        }
        // Ensure the sign of `-0` is preserved.
        start = toFinite(start);
        if (end === undefined) {
          end = start;
          start = 0;
        } else {
          end = toFinite(end);
        }
        step = step === undefined ? (start < end ? 1 : -1) : toFinite(step);
        return baseRange(start, end, step, fromRight);
      };
    }

    /**
     * Creates a function that performs a relational operation on two values.
     *
     * @private
     * @param {Function} operator The function to perform the operation.
     * @returns {Function} Returns the new relational operation function.
     */
    function createRelationalOperation(operator) {
      return function(value, other) {
        if (!(typeof value == 'string' && typeof other == 'string')) {
          value = toNumber(value);
          other = toNumber(other);
        }
        return operator(value, other);
      };
    }

    /**
     * Creates a function that wraps `func` to continue currying.
     *
     * @private
     * @param {Function} func The function to wrap.
     * @param {number} bitmask The bitmask flags. See `createWrap` for more details.
     * @param {Function} wrapFunc The function to create the `func` wrapper.
     * @param {*} placeholder The placeholder value.
     * @param {*} [thisArg] The `this` binding of `func`.
     * @param {Array} [partials] The arguments to prepend to those provided to
     *  the new function.
     * @param {Array} [holders] The `partials` placeholder indexes.
     * @param {Array} [argPos] The argument positions of the new function.
     * @param {number} [ary] The arity cap of `func`.
     * @param {number} [arity] The arity of `func`.
     * @returns {Function} Returns the new wrapped function.
     */
    function createRecurry(func, bitmask, wrapFunc, placeholder, thisArg, partials, holders, argPos, ary, arity) {
      var isCurry = bitmask & WRAP_CURRY_FLAG,
          newHolders = isCurry ? holders : undefined,
          newHoldersRight = isCurry ? undefined : holders,
          newPartials = isCurry ? partials : undefined,
          newPartialsRight = isCurry ? undefined : partials;

      bitmask |= (isCurry ? WRAP_PARTIAL_FLAG : WRAP_PARTIAL_RIGHT_FLAG);
      bitmask &= ~(isCurry ? WRAP_PARTIAL_RIGHT_FLAG : WRAP_PARTIAL_FLAG);

      if (!(bitmask & WRAP_CURRY_BOUND_FLAG)) {
        bitmask &= ~(WRAP_BIND_FLAG | WRAP_BIND_KEY_FLAG);
      }
      var newData = [
        func, bitmask, thisArg, newPartials, newHolders, newPartialsRight,
        newHoldersRight, argPos, ary, arity
      ];

      var result = wrapFunc.apply(undefined, newData);
      if (isLaziable(func)) {
        setData(result, newData);
      }
      result.placeholder = placeholder;
      return setWrapToString(result, func, bitmask);
    }

    /**
     * Creates a function like `_.round`.
     *
     * @private
     * @param {string} methodName The name of the `Math` method to use when rounding.
     * @returns {Function} Returns the new round function.
     */
    function createRound(methodName) {
      var func = Math[methodName];
      return function(number, precision) {
        number = toNumber(number);
        precision = precision == null ? 0 : nativeMin(toInteger(precision), 292);
        if (precision && nativeIsFinite(number)) {
          // Shift with exponential notation to avoid floating-point issues.
          // See [MDN](https://mdn.io/round#Examples) for more details.
          var pair = (toString(number) + 'e').split('e'),
              value = func(pair[0] + 'e' + (+pair[1] + precision));

          pair = (toString(value) + 'e').split('e');
          return +(pair[0] + 'e' + (+pair[1] - precision));
        }
        return func(number);
      };
    }

    /**
     * Creates a set object of `values`.
     *
     * @private
     * @param {Array} values The values to add to the set.
     * @returns {Object} Returns the new set.
     */
    var createSet = !(Set && (1 / setToArray(new Set([,-0]))[1]) == INFINITY) ? noop : function(values) {
      return new Set(values);
    };

    /**
     * Creates a `_.toPairs` or `_.toPairsIn` function.
     *
     * @private
     * @param {Function} keysFunc The function to get the keys of a given object.
     * @returns {Function} Returns the new pairs function.
     */
    function createToPairs(keysFunc) {
      return function(object) {
        var tag = getTag(object);
        if (tag == mapTag) {
          return mapToArray(object);
        }
        if (tag == setTag) {
          return setToPairs(object);
        }
        return baseToPairs(object, keysFunc(object));
      };
    }

    /**
     * Creates a function that either curries or invokes `func` with optional
     * `this` binding and partially applied arguments.
     *
     * @private
     * @param {Function|string} func The function or method name to wrap.
     * @param {number} bitmask The bitmask flags.
     *    1 - `_.bind`
     *    2 - `_.bindKey`
     *    4 - `_.curry` or `_.curryRight` of a bound function
     *    8 - `_.curry`
     *   16 - `_.curryRight`
     *   32 - `_.partial`
     *   64 - `_.partialRight`
     *  128 - `_.rearg`
     *  256 - `_.ary`
     *  512 - `_.flip`
     * @param {*} [thisArg] The `this` binding of `func`.
     * @param {Array} [partials] The arguments to be partially applied.
     * @param {Array} [holders] The `partials` placeholder indexes.
     * @param {Array} [argPos] The argument positions of the new function.
     * @param {number} [ary] The arity cap of `func`.
     * @param {number} [arity] The arity of `func`.
     * @returns {Function} Returns the new wrapped function.
     */
    function createWrap(func, bitmask, thisArg, partials, holders, argPos, ary, arity) {
      var isBindKey = bitmask & WRAP_BIND_KEY_FLAG;
      if (!isBindKey && typeof func != 'function') {
        throw new TypeError(FUNC_ERROR_TEXT);
      }
      var length = partials ? partials.length : 0;
      if (!length) {
        bitmask &= ~(WRAP_PARTIAL_FLAG | WRAP_PARTIAL_RIGHT_FLAG);
        partials = holders = undefined;
      }
      ary = ary === undefined ? ary : nativeMax(toInteger(ary), 0);
      arity = arity === undefined ? arity : toInteger(arity);
      length -= holders ? holders.length : 0;

      if (bitmask & WRAP_PARTIAL_RIGHT_FLAG) {
        var partialsRight = partials,
            holdersRight = holders;

        partials = holders = undefined;
      }
      var data = isBindKey ? undefined : getData(func);

      var newData = [
        func, bitmask, thisArg, partials, holders, partialsRight, holdersRight,
        argPos, ary, arity
      ];

      if (data) {
        mergeData(newData, data);
      }
      func = newData[0];
      bitmask = newData[1];
      thisArg = newData[2];
      partials = newData[3];
      holders = newData[4];
      arity = newData[9] = newData[9] === undefined
        ? (isBindKey ? 0 : func.length)
        : nativeMax(newData[9] - length, 0);

      if (!arity && bitmask & (WRAP_CURRY_FLAG | WRAP_CURRY_RIGHT_FLAG)) {
        bitmask &= ~(WRAP_CURRY_FLAG | WRAP_CURRY_RIGHT_FLAG);
      }
      if (!bitmask || bitmask == WRAP_BIND_FLAG) {
        var result = createBind(func, bitmask, thisArg);
      } else if (bitmask == WRAP_CURRY_FLAG || bitmask == WRAP_CURRY_RIGHT_FLAG) {
        result = createCurry(func, bitmask, arity);
      } else if ((bitmask == WRAP_PARTIAL_FLAG || bitmask == (WRAP_BIND_FLAG | WRAP_PARTIAL_FLAG)) && !holders.length) {
        result = createPartial(func, bitmask, thisArg, partials);
      } else {
        result = createHybrid.apply(undefined, newData);
      }
      var setter = data ? baseSetData : setData;
      return setWrapToString(setter(result, newData), func, bitmask);
    }

    /**
     * Used by `_.defaults` to customize its `_.assignIn` use to assign properties
     * of source objects to the destination object for all destination properties
     * that resolve to `undefined`.
     *
     * @private
     * @param {*} objValue The destination value.
     * @param {*} srcValue The source value.
     * @param {string} key The key of the property to assign.
     * @param {Object} object The parent object of `objValue`.
     * @returns {*} Returns the value to assign.
     */
    function customDefaultsAssignIn(objValue, srcValue, key, object) {
      if (objValue === undefined ||
          (eq(objValue, objectProto[key]) && !hasOwnProperty.call(object, key))) {
        return srcValue;
      }
      return objValue;
    }

    /**
     * Used by `_.defaultsDeep` to customize its `_.merge` use to merge source
     * objects into destination objects that are passed thru.
     *
     * @private
     * @param {*} objValue The destination value.
     * @param {*} srcValue The source value.
     * @param {string} key The key of the property to merge.
     * @param {Object} object The parent object of `objValue`.
     * @param {Object} source The parent object of `srcValue`.
     * @param {Object} [stack] Tracks traversed source values and their merged
     *  counterparts.
     * @returns {*} Returns the value to assign.
     */
    function customDefaultsMerge(objValue, srcValue, key, object, source, stack) {
      if (isObject(objValue) && isObject(srcValue)) {
        // Recursively merge objects and arrays (susceptible to call stack limits).
        stack.set(srcValue, objValue);
        baseMerge(objValue, srcValue, undefined, customDefaultsMerge, stack);
        stack['delete'](srcValue);
      }
      return objValue;
    }

    /**
     * Used by `_.omit` to customize its `_.cloneDeep` use to only clone plain
     * objects.
     *
     * @private
     * @param {*} value The value to inspect.
     * @param {string} key The key of the property to inspect.
     * @returns {*} Returns the uncloned value or `undefined` to defer cloning to `_.cloneDeep`.
     */
    function customOmitClone(value) {
      return isPlainObject(value) ? undefined : value;
    }

    /**
     * A specialized version of `baseIsEqualDeep` for arrays with support for
     * partial deep comparisons.
     *
     * @private
     * @param {Array} array The array to compare.
     * @param {Array} other The other array to compare.
     * @param {number} bitmask The bitmask flags. See `baseIsEqual` for more details.
     * @param {Function} customizer The function to customize comparisons.
     * @param {Function} equalFunc The function to determine equivalents of values.
     * @param {Object} stack Tracks traversed `array` and `other` objects.
     * @returns {boolean} Returns `true` if the arrays are equivalent, else `false`.
     */
    function equalArrays(array, other, bitmask, customizer, equalFunc, stack) {
      var isPartial = bitmask & COMPARE_PARTIAL_FLAG,
          arrLength = array.length,
          othLength = other.length;

      if (arrLength != othLength && !(isPartial && othLength > arrLength)) {
        return false;
      }
      // Check that cyclic values are equal.
      var arrStacked = stack.get(array);
      var othStacked = stack.get(other);
      if (arrStacked && othStacked) {
        return arrStacked == other && othStacked == array;
      }
      var index = -1,
          result = true,
          seen = (bitmask & COMPARE_UNORDERED_FLAG) ? new SetCache : undefined;

      stack.set(array, other);
      stack.set(other, array);

      // Ignore non-index properties.
      while (++index < arrLength) {
        var arrValue = array[index],
            othValue = other[index];

        if (customizer) {
          var compared = isPartial
            ? customizer(othValue, arrValue, index, other, array, stack)
            : customizer(arrValue, othValue, index, array, other, stack);
        }
        if (compared !== undefined) {
          if (compared) {
            continue;
          }
          result = false;
          break;
        }
        // Recursively compare arrays (susceptible to call stack limits).
        if (seen) {
          if (!arraySome(other, function(othValue, othIndex) {
                if (!cacheHas(seen, othIndex) &&
                    (arrValue === othValue || equalFunc(arrValue, othValue, bitmask, customizer, stack))) {
                  return seen.push(othIndex);
                }
              })) {
            result = false;
            break;
          }
        } else if (!(
              arrValue === othValue ||
                equalFunc(arrValue, othValue, bitmask, customizer, stack)
            )) {
          result = false;
          break;
        }
      }
      stack['delete'](array);
      stack['delete'](other);
      return result;
    }

    /**
     * A specialized version of `baseIsEqualDeep` for comparing objects of
     * the same `toStringTag`.
     *
     * **Note:** This function only supports comparing values with tags of
     * `Boolean`, `Date`, `Error`, `Number`, `RegExp`, or `String`.
     *
     * @private
     * @param {Object} object The object to compare.
     * @param {Object} other The other object to compare.
     * @param {string} tag The `toStringTag` of the objects to compare.
     * @param {number} bitmask The bitmask flags. See `baseIsEqual` for more details.
     * @param {Function} customizer The function to customize comparisons.
     * @param {Function} equalFunc The function to determine equivalents of values.
     * @param {Object} stack Tracks traversed `object` and `other` objects.
     * @returns {boolean} Returns `true` if the objects are equivalent, else `false`.
     */
    function equalByTag(object, other, tag, bitmask, customizer, equalFunc, stack) {
      switch (tag) {
        case dataViewTag:
          if ((object.byteLength != other.byteLength) ||
              (object.byteOffset != other.byteOffset)) {
            return false;
          }
          object = object.buffer;
          other = other.buffer;

        case arrayBufferTag:
          if ((object.byteLength != other.byteLength) ||
              !equalFunc(new Uint8Array(object), new Uint8Array(other))) {
            return false;
          }
          return true;

        case boolTag:
        case dateTag:
        case numberTag:
          // Coerce booleans to `1` or `0` and dates to milliseconds.
          // Invalid dates are coerced to `NaN`.
          return eq(+object, +other);

        case errorTag:
          return object.name == other.name && object.message == other.message;

        case regexpTag:
        case stringTag:
          // Coerce regexes to strings and treat strings, primitives and objects,
          // as equal. See http://www.ecma-international.org/ecma-262/7.0/#sec-regexp.prototype.tostring
          // for more details.
          return object == (other + '');

        case mapTag:
          var convert = mapToArray;

        case setTag:
          var isPartial = bitmask & COMPARE_PARTIAL_FLAG;
          convert || (convert = setToArray);

          if (object.size != other.size && !isPartial) {
            return false;
          }
          // Assume cyclic values are equal.
          var stacked = stack.get(object);
          if (stacked) {
            return stacked == other;
          }
          bitmask |= COMPARE_UNORDERED_FLAG;

          // Recursively compare objects (susceptible to call stack limits).
          stack.set(object, other);
          var result = equalArrays(convert(object), convert(other), bitmask, customizer, equalFunc, stack);
          stack['delete'](object);
          return result;

        case symbolTag:
          if (symbolValueOf) {
            return symbolValueOf.call(object) == symbolValueOf.call(other);
          }
      }
      return false;
    }

    /**
     * A specialized version of `baseIsEqualDeep` for objects with support for
     * partial deep comparisons.
     *
     * @private
     * @param {Object} object The object to compare.
     * @param {Object} other The other object to compare.
     * @param {number} bitmask The bitmask flags. See `baseIsEqual` for more details.
     * @param {Function} customizer The function to customize comparisons.
     * @param {Function} equalFunc The function to determine equivalents of values.
     * @param {Object} stack Tracks traversed `object` and `other` objects.
     * @returns {boolean} Returns `true` if the objects are equivalent, else `false`.
     */
    function equalObjects(object, other, bitmask, customizer, equalFunc, stack) {
      var isPartial = bitmask & COMPARE_PARTIAL_FLAG,
          objProps = getAllKeys(object),
          objLength = objProps.length,
          othProps = getAllKeys(other),
          othLength = othProps.length;

      if (objLength != othLength && !isPartial) {
        return false;
      }
      var index = objLength;
      while (index--) {
        var key = objProps[index];
        if (!(isPartial ? key in other : hasOwnProperty.call(other, key))) {
          return false;
        }
      }
      // Check that cyclic values are equal.
      var objStacked = stack.get(object);
      var othStacked = stack.get(other);
      if (objStacked && othStacked) {
        return objStacked == other && othStacked == object;
      }
      var result = true;
      stack.set(object, other);
      stack.set(other, object);

      var skipCtor = isPartial;
      while (++index < objLength) {
        key = objProps[index];
        var objValue = object[key],
            othValue = other[key];

        if (customizer) {
          var compared = isPartial
            ? customizer(othValue, objValue, key, other, object, stack)
            : customizer(objValue, othValue, key, object, other, stack);
        }
        // Recursively compare objects (susceptible to call stack limits).
        if (!(compared === undefined
              ? (objValue === othValue || equalFunc(objValue, othValue, bitmask, customizer, stack))
              : compared
            )) {
          result = false;
          break;
        }
        skipCtor || (skipCtor = key == 'constructor');
      }
      if (result && !skipCtor) {
        var objCtor = object.constructor,
            othCtor = other.constructor;

        // Non `Object` object instances with different constructors are not equal.
        if (objCtor != othCtor &&
            ('constructor' in object && 'constructor' in other) &&
            !(typeof objCtor == 'function' && objCtor instanceof objCtor &&
              typeof othCtor == 'function' && othCtor instanceof othCtor)) {
          result = false;
        }
      }
      stack['delete'](object);
      stack['delete'](other);
      return result;
    }

    /**
     * A specialized version of `baseRest` which flattens the rest array.
     *
     * @private
     * @param {Function} func The function to apply a rest parameter to.
     * @returns {Function} Returns the new function.
     */
    function flatRest(func) {
      return setToString(overRest(func, undefined, flatten), func + '');
    }

    /**
     * Creates an array of own enumerable property names and symbols of `object`.
     *
     * @private
     * @param {Object} object The object to query.
     * @returns {Array} Returns the array of property names and symbols.
     */
    function getAllKeys(object) {
      return baseGetAllKeys(object, keys, getSymbols);
    }

    /**
     * Creates an array of own and inherited enumerable property names and
     * symbols of `object`.
     *
     * @private
     * @param {Object} object The object to query.
     * @returns {Array} Returns the array of property names and symbols.
     */
    function getAllKeysIn(object) {
      return baseGetAllKeys(object, keysIn, getSymbolsIn);
    }

    /**
     * Gets metadata for `func`.
     *
     * @private
     * @param {Function} func The function to query.
     * @returns {*} Returns the metadata for `func`.
     */
    var getData = !metaMap ? noop : function(func) {
      return metaMap.get(func);
    };

    /**
     * Gets the name of `func`.
     *
     * @private
     * @param {Function} func The function to query.
     * @returns {string} Returns the function name.
     */
    function getFuncName(func) {
      var result = (func.name + ''),
          array = realNames[result],
          length = hasOwnProperty.call(realNames, result) ? array.length : 0;

      while (length--) {
        var data = array[length],
            otherFunc = data.func;
        if (otherFunc == null || otherFunc == func) {
          return data.name;
        }
      }
      return result;
    }

    /**
     * Gets the argument placeholder value for `func`.
     *
     * @private
     * @param {Function} func The function to inspect.
     * @returns {*} Returns the placeholder value.
     */
    function getHolder(func) {
      var object = hasOwnProperty.call(lodash, 'placeholder') ? lodash : func;
      return object.placeholder;
    }

    /**
     * Gets the appropriate "iteratee" function. If `_.iteratee` is customized,
     * this function returns the custom method, otherwise it returns `baseIteratee`.
     * If arguments are provided, the chosen function is invoked with them and
     * its result is returned.
     *
     * @private
     * @param {*} [value] The value to convert to an iteratee.
     * @param {number} [arity] The arity of the created iteratee.
     * @returns {Function} Returns the chosen function or its result.
     */
    function getIteratee() {
      var result = lodash.iteratee || iteratee;
      result = result === iteratee ? baseIteratee : result;
      return arguments.length ? result(arguments[0], arguments[1]) : result;
    }

    /**
     * Gets the data for `map`.
     *
     * @private
     * @param {Object} map The map to query.
     * @param {string} key The reference key.
     * @returns {*} Returns the map data.
     */
    function getMapData(map, key) {
      var data = map.__data__;
      return isKeyable(key)
        ? data[typeof key == 'string' ? 'string' : 'hash']
        : data.map;
    }

    /**
     * Gets the property names, values, and compare flags of `object`.
     *
     * @private
     * @param {Object} object The object to query.
     * @returns {Array} Returns the match data of `object`.
     */
    function getMatchData(object) {
      var result = keys(object),
          length = result.length;

      while (length--) {
        var key = result[length],
            value = object[key];

        result[length] = [key, value, isStrictComparable(value)];
      }
      return result;
    }

    /**
     * Gets the native function at `key` of `object`.
     *
     * @private
     * @param {Object} object The object to query.
     * @param {string} key The key of the method to get.
     * @returns {*} Returns the function if it's native, else `undefined`.
     */
    function getNative(object, key) {
      var value = getValue(object, key);
      return baseIsNative(value) ? value : undefined;
    }

    /**
     * A specialized version of `baseGetTag` which ignores `Symbol.toStringTag` values.
     *
     * @private
     * @param {*} value The value to query.
     * @returns {string} Returns the raw `toStringTag`.
     */
    function getRawTag(value) {
      var isOwn = hasOwnProperty.call(value, symToStringTag),
          tag = value[symToStringTag];

      try {
        value[symToStringTag] = undefined;
        var unmasked = true;
      } catch (e) {}

      var result = nativeObjectToString.call(value);
      if (unmasked) {
        if (isOwn) {
          value[symToStringTag] = tag;
        } else {
          delete value[symToStringTag];
        }
      }
      return result;
    }

    /**
     * Creates an array of the own enumerable symbols of `object`.
     *
     * @private
     * @param {Object} object The object to query.
     * @returns {Array} Returns the array of symbols.
     */
    var getSymbols = !nativeGetSymbols ? stubArray : function(object) {
      if (object == null) {
        return [];
      }
      object = Object(object);
      return arrayFilter(nativeGetSymbols(object), function(symbol) {
        return propertyIsEnumerable.call(object, symbol);
      });
    };

    /**
     * Creates an array of the own and inherited enumerable symbols of `object`.
     *
     * @private
     * @param {Object} object The object to query.
     * @returns {Array} Returns the array of symbols.
     */
    var getSymbolsIn = !nativeGetSymbols ? stubArray : function(object) {
      var result = [];
      while (object) {
        arrayPush(result, getSymbols(object));
        object = getPrototype(object);
      }
      return result;
    };

    /**
     * Gets the `toStringTag` of `value`.
     *
     * @private
     * @param {*} value The value to query.
     * @returns {string} Returns the `toStringTag`.
     */
    var getTag = baseGetTag;

    // Fallback for data views, maps, sets, and weak maps in IE 11 and promises in Node.js < 6.
    if ((DataView && getTag(new DataView(new ArrayBuffer(1))) != dataViewTag) ||
        (Map && getTag(new Map) != mapTag) ||
        (Promise && getTag(Promise.resolve()) != promiseTag) ||
        (Set && getTag(new Set) != setTag) ||
        (WeakMap && getTag(new WeakMap) != weakMapTag)) {
      getTag = function(value) {
        var result = baseGetTag(value),
            Ctor = result == objectTag ? value.constructor : undefined,
            ctorString = Ctor ? toSource(Ctor) : '';

        if (ctorString) {
          switch (ctorString) {
            case dataViewCtorString: return dataViewTag;
            case mapCtorString: return mapTag;
            case promiseCtorString: return promiseTag;
            case setCtorString: return setTag;
            case weakMapCtorString: return weakMapTag;
          }
        }
        return result;
      };
    }

    /**
     * Gets the view, applying any `transforms` to the `start` and `end` positions.
     *
     * @private
     * @param {number} start The start of the view.
     * @param {number} end The end of the view.
     * @param {Array} transforms The transformations to apply to the view.
     * @returns {Object} Returns an object containing the `start` and `end`
     *  positions of the view.
     */
    function getView(start, end, transforms) {
      var index = -1,
          length = transforms.length;

      while (++index < length) {
        var data = transforms[index],
            size = data.size;

        switch (data.type) {
          case 'drop':      start += size; break;
          case 'dropRight': end -= size; break;
          case 'take':      end = nativeMin(end, start + size); break;
          case 'takeRight': start = nativeMax(start, end - size); break;
        }
      }
      return { 'start': start, 'end': end };
    }

    /**
     * Extracts wrapper details from the `source` body comment.
     *
     * @private
     * @param {string} source The source to inspect.
     * @returns {Array} Returns the wrapper details.
     */
    function getWrapDetails(source) {
      var match = source.match(reWrapDetails);
      return match ? match[1].split(reSplitDetails) : [];
    }

    /**
     * Checks if `path` exists on `object`.
     *
     * @private
     * @param {Object} object The object to query.
     * @param {Array|string} path The path to check.
     * @param {Function} hasFunc The function to check properties.
     * @returns {boolean} Returns `true` if `path` exists, else `false`.
     */
    function hasPath(object, path, hasFunc) {
      path = castPath(path, object);

      var index = -1,
          length = path.length,
          result = false;

      while (++index < length) {
        var key = toKey(path[index]);
        if (!(result = object != null && hasFunc(object, key))) {
          break;
        }
        object = object[key];
      }
      if (result || ++index != length) {
        return result;
      }
      length = object == null ? 0 : object.length;
      return !!length && isLength(length) && isIndex(key, length) &&
        (isArray(object) || isArguments(object));
    }

    /**
     * Initializes an array clone.
     *
     * @private
     * @param {Array} array The array to clone.
     * @returns {Array} Returns the initialized clone.
     */
    function initCloneArray(array) {
      var length = array.length,
          result = new array.constructor(length);

      // Add properties assigned by `RegExp#exec`.
      if (length && typeof array[0] == 'string' && hasOwnProperty.call(array, 'index')) {
        result.index = array.index;
        result.input = array.input;
      }
      return result;
    }

    /**
     * Initializes an object clone.
     *
     * @private
     * @param {Object} object The object to clone.
     * @returns {Object} Returns the initialized clone.
     */
    function initCloneObject(object) {
      return (typeof object.constructor == 'function' && !isPrototype(object))
        ? baseCreate(getPrototype(object))
        : {};
    }

    /**
     * Initializes an object clone based on its `toStringTag`.
     *
     * **Note:** This function only supports cloning values with tags of
     * `Boolean`, `Date`, `Error`, `Map`, `Number`, `RegExp`, `Set`, or `String`.
     *
     * @private
     * @param {Object} object The object to clone.
     * @param {string} tag The `toStringTag` of the object to clone.
     * @param {boolean} [isDeep] Specify a deep clone.
     * @returns {Object} Returns the initialized clone.
     */
    function initCloneByTag(object, tag, isDeep) {
      var Ctor = object.constructor;
      switch (tag) {
        case arrayBufferTag:
          return cloneArrayBuffer(object);

        case boolTag:
        case dateTag:
          return new Ctor(+object);

        case dataViewTag:
          return cloneDataView(object, isDeep);

        case float32Tag: case float64Tag:
        case int8Tag: case int16Tag: case int32Tag:
        case uint8Tag: case uint8ClampedTag: case uint16Tag: case uint32Tag:
          return cloneTypedArray(object, isDeep);

        case mapTag:
          return new Ctor;

        case numberTag:
        case stringTag:
          return new Ctor(object);

        case regexpTag:
          return cloneRegExp(object);

        case setTag:
          return new Ctor;

        case symbolTag:
          return cloneSymbol(object);
      }
    }

    /**
     * Inserts wrapper `details` in a comment at the top of the `source` body.
     *
     * @private
     * @param {string} source The source to modify.
     * @returns {Array} details The details to insert.
     * @returns {string} Returns the modified source.
     */
    function insertWrapDetails(source, details) {
      var length = details.length;
      if (!length) {
        return source;
      }
      var lastIndex = length - 1;
      details[lastIndex] = (length > 1 ? '& ' : '') + details[lastIndex];
      details = details.join(length > 2 ? ', ' : ' ');
      return source.replace(reWrapComment, '{\n/* [wrapped with ' + details + '] */\n');
    }

    /**
     * Checks if `value` is a flattenable `arguments` object or array.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is flattenable, else `false`.
     */
    function isFlattenable(value) {
      return isArray(value) || isArguments(value) ||
        !!(spreadableSymbol && value && value[spreadableSymbol]);
    }

    /**
     * Checks if `value` is a valid array-like index.
     *
     * @private
     * @param {*} value The value to check.
     * @param {number} [length=MAX_SAFE_INTEGER] The upper bounds of a valid index.
     * @returns {boolean} Returns `true` if `value` is a valid index, else `false`.
     */
    function isIndex(value, length) {
      var type = typeof value;
      length = length == null ? MAX_SAFE_INTEGER : length;

      return !!length &&
        (type == 'number' ||
          (type != 'symbol' && reIsUint.test(value))) &&
            (value > -1 && value % 1 == 0 && value < length);
    }

    /**
     * Checks if the given arguments are from an iteratee call.
     *
     * @private
     * @param {*} value The potential iteratee value argument.
     * @param {*} index The potential iteratee index or key argument.
     * @param {*} object The potential iteratee object argument.
     * @returns {boolean} Returns `true` if the arguments are from an iteratee call,
     *  else `false`.
     */
    function isIterateeCall(value, index, object) {
      if (!isObject(object)) {
        return false;
      }
      var type = typeof index;
      if (type == 'number'
            ? (isArrayLike(object) && isIndex(index, object.length))
            : (type == 'string' && index in object)
          ) {
        return eq(object[index], value);
      }
      return false;
    }

    /**
     * Checks if `value` is a property name and not a property path.
     *
     * @private
     * @param {*} value The value to check.
     * @param {Object} [object] The object to query keys on.
     * @returns {boolean} Returns `true` if `value` is a property name, else `false`.
     */
    function isKey(value, object) {
      if (isArray(value)) {
        return false;
      }
      var type = typeof value;
      if (type == 'number' || type == 'symbol' || type == 'boolean' ||
          value == null || isSymbol(value)) {
        return true;
      }
      return reIsPlainProp.test(value) || !reIsDeepProp.test(value) ||
        (object != null && value in Object(object));
    }

    /**
     * Checks if `value` is suitable for use as unique object key.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is suitable, else `false`.
     */
    function isKeyable(value) {
      var type = typeof value;
      return (type == 'string' || type == 'number' || type == 'symbol' || type == 'boolean')
        ? (value !== '__proto__')
        : (value === null);
    }

    /**
     * Checks if `func` has a lazy counterpart.
     *
     * @private
     * @param {Function} func The function to check.
     * @returns {boolean} Returns `true` if `func` has a lazy counterpart,
     *  else `false`.
     */
    function isLaziable(func) {
      var funcName = getFuncName(func),
          other = lodash[funcName];

      if (typeof other != 'function' || !(funcName in LazyWrapper.prototype)) {
        return false;
      }
      if (func === other) {
        return true;
      }
      var data = getData(other);
      return !!data && func === data[0];
    }

    /**
     * Checks if `func` has its source masked.
     *
     * @private
     * @param {Function} func The function to check.
     * @returns {boolean} Returns `true` if `func` is masked, else `false`.
     */
    function isMasked(func) {
      return !!maskSrcKey && (maskSrcKey in func);
    }

    /**
     * Checks if `func` is capable of being masked.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `func` is maskable, else `false`.
     */
    var isMaskable = coreJsData ? isFunction : stubFalse;

    /**
     * Checks if `value` is likely a prototype object.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a prototype, else `false`.
     */
    function isPrototype(value) {
      var Ctor = value && value.constructor,
          proto = (typeof Ctor == 'function' && Ctor.prototype) || objectProto;

      return value === proto;
    }

    /**
     * Checks if `value` is suitable for strict equality comparisons, i.e. `===`.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` if suitable for strict
     *  equality comparisons, else `false`.
     */
    function isStrictComparable(value) {
      return value === value && !isObject(value);
    }

    /**
     * A specialized version of `matchesProperty` for source values suitable
     * for strict equality comparisons, i.e. `===`.
     *
     * @private
     * @param {string} key The key of the property to get.
     * @param {*} srcValue The value to match.
     * @returns {Function} Returns the new spec function.
     */
    function matchesStrictComparable(key, srcValue) {
      return function(object) {
        if (object == null) {
          return false;
        }
        return object[key] === srcValue &&
          (srcValue !== undefined || (key in Object(object)));
      };
    }

    /**
     * A specialized version of `_.memoize` which clears the memoized function's
     * cache when it exceeds `MAX_MEMOIZE_SIZE`.
     *
     * @private
     * @param {Function} func The function to have its output memoized.
     * @returns {Function} Returns the new memoized function.
     */
    function memoizeCapped(func) {
      var result = memoize(func, function(key) {
        if (cache.size === MAX_MEMOIZE_SIZE) {
          cache.clear();
        }
        return key;
      });

      var cache = result.cache;
      return result;
    }

    /**
     * Merges the function metadata of `source` into `data`.
     *
     * Merging metadata reduces the number of wrappers used to invoke a function.
     * This is possible because methods like `_.bind`, `_.curry`, and `_.partial`
     * may be applied regardless of execution order. Methods like `_.ary` and
     * `_.rearg` modify function arguments, making the order in which they are
     * executed important, preventing the merging of metadata. However, we make
     * an exception for a safe combined case where curried functions have `_.ary`
     * and or `_.rearg` applied.
     *
     * @private
     * @param {Array} data The destination metadata.
     * @param {Array} source The source metadata.
     * @returns {Array} Returns `data`.
     */
    function mergeData(data, source) {
      var bitmask = data[1],
          srcBitmask = source[1],
          newBitmask = bitmask | srcBitmask,
          isCommon = newBitmask < (WRAP_BIND_FLAG | WRAP_BIND_KEY_FLAG | WRAP_ARY_FLAG);

      var isCombo =
        ((srcBitmask == WRAP_ARY_FLAG) && (bitmask == WRAP_CURRY_FLAG)) ||
        ((srcBitmask == WRAP_ARY_FLAG) && (bitmask == WRAP_REARG_FLAG) && (data[7].length <= source[8])) ||
        ((srcBitmask == (WRAP_ARY_FLAG | WRAP_REARG_FLAG)) && (source[7].length <= source[8]) && (bitmask == WRAP_CURRY_FLAG));

      // Exit early if metadata can't be merged.
      if (!(isCommon || isCombo)) {
        return data;
      }
      // Use source `thisArg` if available.
      if (srcBitmask & WRAP_BIND_FLAG) {
        data[2] = source[2];
        // Set when currying a bound function.
        newBitmask |= bitmask & WRAP_BIND_FLAG ? 0 : WRAP_CURRY_BOUND_FLAG;
      }
      // Compose partial arguments.
      var value = source[3];
      if (value) {
        var partials = data[3];
        data[3] = partials ? composeArgs(partials, value, source[4]) : value;
        data[4] = partials ? replaceHolders(data[3], PLACEHOLDER) : source[4];
      }
      // Compose partial right arguments.
      value = source[5];
      if (value) {
        partials = data[5];
        data[5] = partials ? composeArgsRight(partials, value, source[6]) : value;
        data[6] = partials ? replaceHolders(data[5], PLACEHOLDER) : source[6];
      }
      // Use source `argPos` if available.
      value = source[7];
      if (value) {
        data[7] = value;
      }
      // Use source `ary` if it's smaller.
      if (srcBitmask & WRAP_ARY_FLAG) {
        data[8] = data[8] == null ? source[8] : nativeMin(data[8], source[8]);
      }
      // Use source `arity` if one is not provided.
      if (data[9] == null) {
        data[9] = source[9];
      }
      // Use source `func` and merge bitmasks.
      data[0] = source[0];
      data[1] = newBitmask;

      return data;
    }

    /**
     * This function is like
     * [`Object.keys`](http://ecma-international.org/ecma-262/7.0/#sec-object.keys)
     * except that it includes inherited enumerable properties.
     *
     * @private
     * @param {Object} object The object to query.
     * @returns {Array} Returns the array of property names.
     */
    function nativeKeysIn(object) {
      var result = [];
      if (object != null) {
        for (var key in Object(object)) {
          result.push(key);
        }
      }
      return result;
    }

    /**
     * Converts `value` to a string using `Object.prototype.toString`.
     *
     * @private
     * @param {*} value The value to convert.
     * @returns {string} Returns the converted string.
     */
    function objectToString(value) {
      return nativeObjectToString.call(value);
    }

    /**
     * A specialized version of `baseRest` which transforms the rest array.
     *
     * @private
     * @param {Function} func The function to apply a rest parameter to.
     * @param {number} [start=func.length-1] The start position of the rest parameter.
     * @param {Function} transform The rest array transform.
     * @returns {Function} Returns the new function.
     */
    function overRest(func, start, transform) {
      start = nativeMax(start === undefined ? (func.length - 1) : start, 0);
      return function() {
        var args = arguments,
            index = -1,
            length = nativeMax(args.length - start, 0),
            array = Array(length);

        while (++index < length) {
          array[index] = args[start + index];
        }
        index = -1;
        var otherArgs = Array(start + 1);
        while (++index < start) {
          otherArgs[index] = args[index];
        }
        otherArgs[start] = transform(array);
        return apply(func, this, otherArgs);
      };
    }

    /**
     * Gets the parent value at `path` of `object`.
     *
     * @private
     * @param {Object} object The object to query.
     * @param {Array} path The path to get the parent value of.
     * @returns {*} Returns the parent value.
     */
    function parent(object, path) {
      return path.length < 2 ? object : baseGet(object, baseSlice(path, 0, -1));
    }

    /**
     * Reorder `array` according to the specified indexes where the element at
     * the first index is assigned as the first element, the element at
     * the second index is assigned as the second element, and so on.
     *
     * @private
     * @param {Array} array The array to reorder.
     * @param {Array} indexes The arranged array indexes.
     * @returns {Array} Returns `array`.
     */
    function reorder(array, indexes) {
      var arrLength = array.length,
          length = nativeMin(indexes.length, arrLength),
          oldArray = copyArray(array);

      while (length--) {
        var index = indexes[length];
        array[length] = isIndex(index, arrLength) ? oldArray[index] : undefined;
      }
      return array;
    }

    /**
     * Gets the value at `key`, unless `key` is "__proto__" or "constructor".
     *
     * @private
     * @param {Object} object The object to query.
     * @param {string} key The key of the property to get.
     * @returns {*} Returns the property value.
     */
    function safeGet(object, key) {
      if (key === 'constructor' && typeof object[key] === 'function') {
        return;
      }

      if (key == '__proto__') {
        return;
      }

      return object[key];
    }

    /**
     * Sets metadata for `func`.
     *
     * **Note:** If this function becomes hot, i.e. is invoked a lot in a short
     * period of time, it will trip its breaker and transition to an identity
     * function to avoid garbage collection pauses in V8. See
     * [V8 issue 2070](https://bugs.chromium.org/p/v8/issues/detail?id=2070)
     * for more details.
     *
     * @private
     * @param {Function} func The function to associate metadata with.
     * @param {*} data The metadata.
     * @returns {Function} Returns `func`.
     */
    var setData = shortOut(baseSetData);

    /**
     * A simple wrapper around the global [`setTimeout`](https://mdn.io/setTimeout).
     *
     * @private
     * @param {Function} func The function to delay.
     * @param {number} wait The number of milliseconds to delay invocation.
     * @returns {number|Object} Returns the timer id or timeout object.
     */
    var setTimeout = ctxSetTimeout || function(func, wait) {
      return root.setTimeout(func, wait);
    };

    /**
     * Sets the `toString` method of `func` to return `string`.
     *
     * @private
     * @param {Function} func The function to modify.
     * @param {Function} string The `toString` result.
     * @returns {Function} Returns `func`.
     */
    var setToString = shortOut(baseSetToString);

    /**
     * Sets the `toString` method of `wrapper` to mimic the source of `reference`
     * with wrapper details in a comment at the top of the source body.
     *
     * @private
     * @param {Function} wrapper The function to modify.
     * @param {Function} reference The reference function.
     * @param {number} bitmask The bitmask flags. See `createWrap` for more details.
     * @returns {Function} Returns `wrapper`.
     */
    function setWrapToString(wrapper, reference, bitmask) {
      var source = (reference + '');
      return setToString(wrapper, insertWrapDetails(source, updateWrapDetails(getWrapDetails(source), bitmask)));
    }

    /**
     * Creates a function that'll short out and invoke `identity` instead
     * of `func` when it's called `HOT_COUNT` or more times in `HOT_SPAN`
     * milliseconds.
     *
     * @private
     * @param {Function} func The function to restrict.
     * @returns {Function} Returns the new shortable function.
     */
    function shortOut(func) {
      var count = 0,
          lastCalled = 0;

      return function() {
        var stamp = nativeNow(),
            remaining = HOT_SPAN - (stamp - lastCalled);

        lastCalled = stamp;
        if (remaining > 0) {
          if (++count >= HOT_COUNT) {
            return arguments[0];
          }
        } else {
          count = 0;
        }
        return func.apply(undefined, arguments);
      };
    }

    /**
     * A specialized version of `_.shuffle` which mutates and sets the size of `array`.
     *
     * @private
     * @param {Array} array The array to shuffle.
     * @param {number} [size=array.length] The size of `array`.
     * @returns {Array} Returns `array`.
     */
    function shuffleSelf(array, size) {
      var index = -1,
          length = array.length,
          lastIndex = length - 1;

      size = size === undefined ? length : size;
      while (++index < size) {
        var rand = baseRandom(index, lastIndex),
            value = array[rand];

        array[rand] = array[index];
        array[index] = value;
      }
      array.length = size;
      return array;
    }

    /**
     * Converts `string` to a property path array.
     *
     * @private
     * @param {string} string The string to convert.
     * @returns {Array} Returns the property path array.
     */
    var stringToPath = memoizeCapped(function(string) {
      var result = [];
      if (string.charCodeAt(0) === 46 /* . */) {
        result.push('');
      }
      string.replace(rePropName, function(match, number, quote, subString) {
        result.push(quote ? subString.replace(reEscapeChar, '$1') : (number || match));
      });
      return result;
    });

    /**
     * Converts `value` to a string key if it's not a string or symbol.
     *
     * @private
     * @param {*} value The value to inspect.
     * @returns {string|symbol} Returns the key.
     */
    function toKey(value) {
      if (typeof value == 'string' || isSymbol(value)) {
        return value;
      }
      var result = (value + '');
      return (result == '0' && (1 / value) == -INFINITY) ? '-0' : result;
    }

    /**
     * Converts `func` to its source code.
     *
     * @private
     * @param {Function} func The function to convert.
     * @returns {string} Returns the source code.
     */
    function toSource(func) {
      if (func != null) {
        try {
          return funcToString.call(func);
        } catch (e) {}
        try {
          return (func + '');
        } catch (e) {}
      }
      return '';
    }

    /**
     * Updates wrapper `details` based on `bitmask` flags.
     *
     * @private
     * @returns {Array} details The details to modify.
     * @param {number} bitmask The bitmask flags. See `createWrap` for more details.
     * @returns {Array} Returns `details`.
     */
    function updateWrapDetails(details, bitmask) {
      arrayEach(wrapFlags, function(pair) {
        var value = '_.' + pair[0];
        if ((bitmask & pair[1]) && !arrayIncludes(details, value)) {
          details.push(value);
        }
      });
      return details.sort();
    }

    /**
     * Creates a clone of `wrapper`.
     *
     * @private
     * @param {Object} wrapper The wrapper to clone.
     * @returns {Object} Returns the cloned wrapper.
     */
    function wrapperClone(wrapper) {
      if (wrapper instanceof LazyWrapper) {
        return wrapper.clone();
      }
      var result = new LodashWrapper(wrapper.__wrapped__, wrapper.__chain__);
      result.__actions__ = copyArray(wrapper.__actions__);
      result.__index__  = wrapper.__index__;
      result.__values__ = wrapper.__values__;
      return result;
    }

    /*------------------------------------------------------------------------*/

    /**
     * Creates an array of elements split into groups the length of `size`.
     * If `array` can't be split evenly, the final chunk will be the remaining
     * elements.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Array
     * @param {Array} array The array to process.
     * @param {number} [size=1] The length of each chunk
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {Array} Returns the new array of chunks.
     * @example
     *
     * _.chunk(['a', 'b', 'c', 'd'], 2);
     * // => [['a', 'b'], ['c', 'd']]
     *
     * _.chunk(['a', 'b', 'c', 'd'], 3);
     * // => [['a', 'b', 'c'], ['d']]
     */
    function chunk(array, size, guard) {
      if ((guard ? isIterateeCall(array, size, guard) : size === undefined)) {
        size = 1;
      } else {
        size = nativeMax(toInteger(size), 0);
      }
      var length = array == null ? 0 : array.length;
      if (!length || size < 1) {
        return [];
      }
      var index = 0,
          resIndex = 0,
          result = Array(nativeCeil(length / size));

      while (index < length) {
        result[resIndex++] = baseSlice(array, index, (index += size));
      }
      return result;
    }

    /**
     * Creates an array with all falsey values removed. The values `false`, `null`,
     * `0`, `""`, `undefined`, and `NaN` are falsey.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Array
     * @param {Array} array The array to compact.
     * @returns {Array} Returns the new array of filtered values.
     * @example
     *
     * _.compact([0, 1, false, 2, '', 3]);
     * // => [1, 2, 3]
     */
    function compact(array) {
      var index = -1,
          length = array == null ? 0 : array.length,
          resIndex = 0,
          result = [];

      while (++index < length) {
        var value = array[index];
        if (value) {
          result[resIndex++] = value;
        }
      }
      return result;
    }

    /**
     * Creates a new array concatenating `array` with any additional arrays
     * and/or values.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The array to concatenate.
     * @param {...*} [values] The values to concatenate.
     * @returns {Array} Returns the new concatenated array.
     * @example
     *
     * var array = [1];
     * var other = _.concat(array, 2, [3], [[4]]);
     *
     * console.log(other);
     * // => [1, 2, 3, [4]]
     *
     * console.log(array);
     * // => [1]
     */
    function concat() {
      var length = arguments.length;
      if (!length) {
        return [];
      }
      var args = Array(length - 1),
          array = arguments[0],
          index = length;

      while (index--) {
        args[index - 1] = arguments[index];
      }
      return arrayPush(isArray(array) ? copyArray(array) : [array], baseFlatten(args, 1));
    }

    /**
     * Creates an array of `array` values not included in the other given arrays
     * using [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
     * for equality comparisons. The order and references of result values are
     * determined by the first array.
     *
     * **Note:** Unlike `_.pullAll`, this method returns a new array.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @param {...Array} [values] The values to exclude.
     * @returns {Array} Returns the new array of filtered values.
     * @see _.without, _.xor
     * @example
     *
     * _.difference([2, 1], [2, 3]);
     * // => [1]
     */
    var difference = baseRest(function(array, values) {
      return isArrayLikeObject(array)
        ? baseDifference(array, baseFlatten(values, 1, isArrayLikeObject, true))
        : [];
    });

    /**
     * This method is like `_.difference` except that it accepts `iteratee` which
     * is invoked for each element of `array` and `values` to generate the criterion
     * by which they're compared. The order and references of result values are
     * determined by the first array. The iteratee is invoked with one argument:
     * (value).
     *
     * **Note:** Unlike `_.pullAllBy`, this method returns a new array.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @param {...Array} [values] The values to exclude.
     * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
     * @returns {Array} Returns the new array of filtered values.
     * @example
     *
     * _.differenceBy([2.1, 1.2], [2.3, 3.4], Math.floor);
     * // => [1.2]
     *
     * // The `_.property` iteratee shorthand.
     * _.differenceBy([{ 'x': 2 }, { 'x': 1 }], [{ 'x': 1 }], 'x');
     * // => [{ 'x': 2 }]
     */
    var differenceBy = baseRest(function(array, values) {
      var iteratee = last(values);
      if (isArrayLikeObject(iteratee)) {
        iteratee = undefined;
      }
      return isArrayLikeObject(array)
        ? baseDifference(array, baseFlatten(values, 1, isArrayLikeObject, true), getIteratee(iteratee, 2))
        : [];
    });

    /**
     * This method is like `_.difference` except that it accepts `comparator`
     * which is invoked to compare elements of `array` to `values`. The order and
     * references of result values are determined by the first array. The comparator
     * is invoked with two arguments: (arrVal, othVal).
     *
     * **Note:** Unlike `_.pullAllWith`, this method returns a new array.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @param {...Array} [values] The values to exclude.
     * @param {Function} [comparator] The comparator invoked per element.
     * @returns {Array} Returns the new array of filtered values.
     * @example
     *
     * var objects = [{ 'x': 1, 'y': 2 }, { 'x': 2, 'y': 1 }];
     *
     * _.differenceWith(objects, [{ 'x': 1, 'y': 2 }], _.isEqual);
     * // => [{ 'x': 2, 'y': 1 }]
     */
    var differenceWith = baseRest(function(array, values) {
      var comparator = last(values);
      if (isArrayLikeObject(comparator)) {
        comparator = undefined;
      }
      return isArrayLikeObject(array)
        ? baseDifference(array, baseFlatten(values, 1, isArrayLikeObject, true), undefined, comparator)
        : [];
    });

    /**
     * Creates a slice of `array` with `n` elements dropped from the beginning.
     *
     * @static
     * @memberOf _
     * @since 0.5.0
     * @category Array
     * @param {Array} array The array to query.
     * @param {number} [n=1] The number of elements to drop.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {Array} Returns the slice of `array`.
     * @example
     *
     * _.drop([1, 2, 3]);
     * // => [2, 3]
     *
     * _.drop([1, 2, 3], 2);
     * // => [3]
     *
     * _.drop([1, 2, 3], 5);
     * // => []
     *
     * _.drop([1, 2, 3], 0);
     * // => [1, 2, 3]
     */
    function drop(array, n, guard) {
      var length = array == null ? 0 : array.length;
      if (!length) {
        return [];
      }
      n = (guard || n === undefined) ? 1 : toInteger(n);
      return baseSlice(array, n < 0 ? 0 : n, length);
    }

    /**
     * Creates a slice of `array` with `n` elements dropped from the end.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Array
     * @param {Array} array The array to query.
     * @param {number} [n=1] The number of elements to drop.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {Array} Returns the slice of `array`.
     * @example
     *
     * _.dropRight([1, 2, 3]);
     * // => [1, 2]
     *
     * _.dropRight([1, 2, 3], 2);
     * // => [1]
     *
     * _.dropRight([1, 2, 3], 5);
     * // => []
     *
     * _.dropRight([1, 2, 3], 0);
     * // => [1, 2, 3]
     */
    function dropRight(array, n, guard) {
      var length = array == null ? 0 : array.length;
      if (!length) {
        return [];
      }
      n = (guard || n === undefined) ? 1 : toInteger(n);
      n = length - n;
      return baseSlice(array, 0, n < 0 ? 0 : n);
    }

    /**
     * Creates a slice of `array` excluding elements dropped from the end.
     * Elements are dropped until `predicate` returns falsey. The predicate is
     * invoked with three arguments: (value, index, array).
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Array
     * @param {Array} array The array to query.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @returns {Array} Returns the slice of `array`.
     * @example
     *
     * var users = [
     *   { 'user': 'barney',  'active': true },
     *   { 'user': 'fred',    'active': false },
     *   { 'user': 'pebbles', 'active': false }
     * ];
     *
     * _.dropRightWhile(users, function(o) { return !o.active; });
     * // => objects for ['barney']
     *
     * // The `_.matches` iteratee shorthand.
     * _.dropRightWhile(users, { 'user': 'pebbles', 'active': false });
     * // => objects for ['barney', 'fred']
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.dropRightWhile(users, ['active', false]);
     * // => objects for ['barney']
     *
     * // The `_.property` iteratee shorthand.
     * _.dropRightWhile(users, 'active');
     * // => objects for ['barney', 'fred', 'pebbles']
     */
    function dropRightWhile(array, predicate) {
      return (array && array.length)
        ? baseWhile(array, getIteratee(predicate, 3), true, true)
        : [];
    }

    /**
     * Creates a slice of `array` excluding elements dropped from the beginning.
     * Elements are dropped until `predicate` returns falsey. The predicate is
     * invoked with three arguments: (value, index, array).
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Array
     * @param {Array} array The array to query.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @returns {Array} Returns the slice of `array`.
     * @example
     *
     * var users = [
     *   { 'user': 'barney',  'active': false },
     *   { 'user': 'fred',    'active': false },
     *   { 'user': 'pebbles', 'active': true }
     * ];
     *
     * _.dropWhile(users, function(o) { return !o.active; });
     * // => objects for ['pebbles']
     *
     * // The `_.matches` iteratee shorthand.
     * _.dropWhile(users, { 'user': 'barney', 'active': false });
     * // => objects for ['fred', 'pebbles']
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.dropWhile(users, ['active', false]);
     * // => objects for ['pebbles']
     *
     * // The `_.property` iteratee shorthand.
     * _.dropWhile(users, 'active');
     * // => objects for ['barney', 'fred', 'pebbles']
     */
    function dropWhile(array, predicate) {
      return (array && array.length)
        ? baseWhile(array, getIteratee(predicate, 3), true)
        : [];
    }

    /**
     * Fills elements of `array` with `value` from `start` up to, but not
     * including, `end`.
     *
     * **Note:** This method mutates `array`.
     *
     * @static
     * @memberOf _
     * @since 3.2.0
     * @category Array
     * @param {Array} array The array to fill.
     * @param {*} value The value to fill `array` with.
     * @param {number} [start=0] The start position.
     * @param {number} [end=array.length] The end position.
     * @returns {Array} Returns `array`.
     * @example
     *
     * var array = [1, 2, 3];
     *
     * _.fill(array, 'a');
     * console.log(array);
     * // => ['a', 'a', 'a']
     *
     * _.fill(Array(3), 2);
     * // => [2, 2, 2]
     *
     * _.fill([4, 6, 8, 10], '*', 1, 3);
     * // => [4, '*', '*', 10]
     */
    function fill(array, value, start, end) {
      var length = array == null ? 0 : array.length;
      if (!length) {
        return [];
      }
      if (start && typeof start != 'number' && isIterateeCall(array, value, start)) {
        start = 0;
        end = length;
      }
      return baseFill(array, value, start, end);
    }

    /**
     * This method is like `_.find` except that it returns the index of the first
     * element `predicate` returns truthy for instead of the element itself.
     *
     * @static
     * @memberOf _
     * @since 1.1.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @param {number} [fromIndex=0] The index to search from.
     * @returns {number} Returns the index of the found element, else `-1`.
     * @example
     *
     * var users = [
     *   { 'user': 'barney',  'active': false },
     *   { 'user': 'fred',    'active': false },
     *   { 'user': 'pebbles', 'active': true }
     * ];
     *
     * _.findIndex(users, function(o) { return o.user == 'barney'; });
     * // => 0
     *
     * // The `_.matches` iteratee shorthand.
     * _.findIndex(users, { 'user': 'fred', 'active': false });
     * // => 1
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.findIndex(users, ['active', false]);
     * // => 0
     *
     * // The `_.property` iteratee shorthand.
     * _.findIndex(users, 'active');
     * // => 2
     */
    function findIndex(array, predicate, fromIndex) {
      var length = array == null ? 0 : array.length;
      if (!length) {
        return -1;
      }
      var index = fromIndex == null ? 0 : toInteger(fromIndex);
      if (index < 0) {
        index = nativeMax(length + index, 0);
      }
      return baseFindIndex(array, getIteratee(predicate, 3), index);
    }

    /**
     * This method is like `_.findIndex` except that it iterates over elements
     * of `collection` from right to left.
     *
     * @static
     * @memberOf _
     * @since 2.0.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @param {number} [fromIndex=array.length-1] The index to search from.
     * @returns {number} Returns the index of the found element, else `-1`.
     * @example
     *
     * var users = [
     *   { 'user': 'barney',  'active': true },
     *   { 'user': 'fred',    'active': false },
     *   { 'user': 'pebbles', 'active': false }
     * ];
     *
     * _.findLastIndex(users, function(o) { return o.user == 'pebbles'; });
     * // => 2
     *
     * // The `_.matches` iteratee shorthand.
     * _.findLastIndex(users, { 'user': 'barney', 'active': true });
     * // => 0
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.findLastIndex(users, ['active', false]);
     * // => 2
     *
     * // The `_.property` iteratee shorthand.
     * _.findLastIndex(users, 'active');
     * // => 0
     */
    function findLastIndex(array, predicate, fromIndex) {
      var length = array == null ? 0 : array.length;
      if (!length) {
        return -1;
      }
      var index = length - 1;
      if (fromIndex !== undefined) {
        index = toInteger(fromIndex);
        index = fromIndex < 0
          ? nativeMax(length + index, 0)
          : nativeMin(index, length - 1);
      }
      return baseFindIndex(array, getIteratee(predicate, 3), index, true);
    }

    /**
     * Flattens `array` a single level deep.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Array
     * @param {Array} array The array to flatten.
     * @returns {Array} Returns the new flattened array.
     * @example
     *
     * _.flatten([1, [2, [3, [4]], 5]]);
     * // => [1, 2, [3, [4]], 5]
     */
    function flatten(array) {
      var length = array == null ? 0 : array.length;
      return length ? baseFlatten(array, 1) : [];
    }

    /**
     * Recursively flattens `array`.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Array
     * @param {Array} array The array to flatten.
     * @returns {Array} Returns the new flattened array.
     * @example
     *
     * _.flattenDeep([1, [2, [3, [4]], 5]]);
     * // => [1, 2, 3, 4, 5]
     */
    function flattenDeep(array) {
      var length = array == null ? 0 : array.length;
      return length ? baseFlatten(array, INFINITY) : [];
    }

    /**
     * Recursively flatten `array` up to `depth` times.
     *
     * @static
     * @memberOf _
     * @since 4.4.0
     * @category Array
     * @param {Array} array The array to flatten.
     * @param {number} [depth=1] The maximum recursion depth.
     * @returns {Array} Returns the new flattened array.
     * @example
     *
     * var array = [1, [2, [3, [4]], 5]];
     *
     * _.flattenDepth(array, 1);
     * // => [1, 2, [3, [4]], 5]
     *
     * _.flattenDepth(array, 2);
     * // => [1, 2, 3, [4], 5]
     */
    function flattenDepth(array, depth) {
      var length = array == null ? 0 : array.length;
      if (!length) {
        return [];
      }
      depth = depth === undefined ? 1 : toInteger(depth);
      return baseFlatten(array, depth);
    }

    /**
     * The inverse of `_.toPairs`; this method returns an object composed
     * from key-value `pairs`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} pairs The key-value pairs.
     * @returns {Object} Returns the new object.
     * @example
     *
     * _.fromPairs([['a', 1], ['b', 2]]);
     * // => { 'a': 1, 'b': 2 }
     */
    function fromPairs(pairs) {
      var index = -1,
          length = pairs == null ? 0 : pairs.length,
          result = {};

      while (++index < length) {
        var pair = pairs[index];
        result[pair[0]] = pair[1];
      }
      return result;
    }

    /**
     * Gets the first element of `array`.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @alias first
     * @category Array
     * @param {Array} array The array to query.
     * @returns {*} Returns the first element of `array`.
     * @example
     *
     * _.head([1, 2, 3]);
     * // => 1
     *
     * _.head([]);
     * // => undefined
     */
    function head(array) {
      return (array && array.length) ? array[0] : undefined;
    }

    /**
     * Gets the index at which the first occurrence of `value` is found in `array`
     * using [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
     * for equality comparisons. If `fromIndex` is negative, it's used as the
     * offset from the end of `array`.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @param {*} value The value to search for.
     * @param {number} [fromIndex=0] The index to search from.
     * @returns {number} Returns the index of the matched value, else `-1`.
     * @example
     *
     * _.indexOf([1, 2, 1, 2], 2);
     * // => 1
     *
     * // Search from the `fromIndex`.
     * _.indexOf([1, 2, 1, 2], 2, 2);
     * // => 3
     */
    function indexOf(array, value, fromIndex) {
      var length = array == null ? 0 : array.length;
      if (!length) {
        return -1;
      }
      var index = fromIndex == null ? 0 : toInteger(fromIndex);
      if (index < 0) {
        index = nativeMax(length + index, 0);
      }
      return baseIndexOf(array, value, index);
    }

    /**
     * Gets all but the last element of `array`.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Array
     * @param {Array} array The array to query.
     * @returns {Array} Returns the slice of `array`.
     * @example
     *
     * _.initial([1, 2, 3]);
     * // => [1, 2]
     */
    function initial(array) {
      var length = array == null ? 0 : array.length;
      return length ? baseSlice(array, 0, -1) : [];
    }

    /**
     * Creates an array of unique values that are included in all given arrays
     * using [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
     * for equality comparisons. The order and references of result values are
     * determined by the first array.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Array
     * @param {...Array} [arrays] The arrays to inspect.
     * @returns {Array} Returns the new array of intersecting values.
     * @example
     *
     * _.intersection([2, 1], [2, 3]);
     * // => [2]
     */
    var intersection = baseRest(function(arrays) {
      var mapped = arrayMap(arrays, castArrayLikeObject);
      return (mapped.length && mapped[0] === arrays[0])
        ? baseIntersection(mapped)
        : [];
    });

    /**
     * This method is like `_.intersection` except that it accepts `iteratee`
     * which is invoked for each element of each `arrays` to generate the criterion
     * by which they're compared. The order and references of result values are
     * determined by the first array. The iteratee is invoked with one argument:
     * (value).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {...Array} [arrays] The arrays to inspect.
     * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
     * @returns {Array} Returns the new array of intersecting values.
     * @example
     *
     * _.intersectionBy([2.1, 1.2], [2.3, 3.4], Math.floor);
     * // => [2.1]
     *
     * // The `_.property` iteratee shorthand.
     * _.intersectionBy([{ 'x': 1 }], [{ 'x': 2 }, { 'x': 1 }], 'x');
     * // => [{ 'x': 1 }]
     */
    var intersectionBy = baseRest(function(arrays) {
      var iteratee = last(arrays),
          mapped = arrayMap(arrays, castArrayLikeObject);

      if (iteratee === last(mapped)) {
        iteratee = undefined;
      } else {
        mapped.pop();
      }
      return (mapped.length && mapped[0] === arrays[0])
        ? baseIntersection(mapped, getIteratee(iteratee, 2))
        : [];
    });

    /**
     * This method is like `_.intersection` except that it accepts `comparator`
     * which is invoked to compare elements of `arrays`. The order and references
     * of result values are determined by the first array. The comparator is
     * invoked with two arguments: (arrVal, othVal).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {...Array} [arrays] The arrays to inspect.
     * @param {Function} [comparator] The comparator invoked per element.
     * @returns {Array} Returns the new array of intersecting values.
     * @example
     *
     * var objects = [{ 'x': 1, 'y': 2 }, { 'x': 2, 'y': 1 }];
     * var others = [{ 'x': 1, 'y': 1 }, { 'x': 1, 'y': 2 }];
     *
     * _.intersectionWith(objects, others, _.isEqual);
     * // => [{ 'x': 1, 'y': 2 }]
     */
    var intersectionWith = baseRest(function(arrays) {
      var comparator = last(arrays),
          mapped = arrayMap(arrays, castArrayLikeObject);

      comparator = typeof comparator == 'function' ? comparator : undefined;
      if (comparator) {
        mapped.pop();
      }
      return (mapped.length && mapped[0] === arrays[0])
        ? baseIntersection(mapped, undefined, comparator)
        : [];
    });

    /**
     * Converts all elements in `array` into a string separated by `separator`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The array to convert.
     * @param {string} [separator=','] The element separator.
     * @returns {string} Returns the joined string.
     * @example
     *
     * _.join(['a', 'b', 'c'], '~');
     * // => 'a~b~c'
     */
    function join(array, separator) {
      return array == null ? '' : nativeJoin.call(array, separator);
    }

    /**
     * Gets the last element of `array`.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Array
     * @param {Array} array The array to query.
     * @returns {*} Returns the last element of `array`.
     * @example
     *
     * _.last([1, 2, 3]);
     * // => 3
     */
    function last(array) {
      var length = array == null ? 0 : array.length;
      return length ? array[length - 1] : undefined;
    }

    /**
     * This method is like `_.indexOf` except that it iterates over elements of
     * `array` from right to left.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @param {*} value The value to search for.
     * @param {number} [fromIndex=array.length-1] The index to search from.
     * @returns {number} Returns the index of the matched value, else `-1`.
     * @example
     *
     * _.lastIndexOf([1, 2, 1, 2], 2);
     * // => 3
     *
     * // Search from the `fromIndex`.
     * _.lastIndexOf([1, 2, 1, 2], 2, 2);
     * // => 1
     */
    function lastIndexOf(array, value, fromIndex) {
      var length = array == null ? 0 : array.length;
      if (!length) {
        return -1;
      }
      var index = length;
      if (fromIndex !== undefined) {
        index = toInteger(fromIndex);
        index = index < 0 ? nativeMax(length + index, 0) : nativeMin(index, length - 1);
      }
      return value === value
        ? strictLastIndexOf(array, value, index)
        : baseFindIndex(array, baseIsNaN, index, true);
    }

    /**
     * Gets the element at index `n` of `array`. If `n` is negative, the nth
     * element from the end is returned.
     *
     * @static
     * @memberOf _
     * @since 4.11.0
     * @category Array
     * @param {Array} array The array to query.
     * @param {number} [n=0] The index of the element to return.
     * @returns {*} Returns the nth element of `array`.
     * @example
     *
     * var array = ['a', 'b', 'c', 'd'];
     *
     * _.nth(array, 1);
     * // => 'b'
     *
     * _.nth(array, -2);
     * // => 'c';
     */
    function nth(array, n) {
      return (array && array.length) ? baseNth(array, toInteger(n)) : undefined;
    }

    /**
     * Removes all given values from `array` using
     * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
     * for equality comparisons.
     *
     * **Note:** Unlike `_.without`, this method mutates `array`. Use `_.remove`
     * to remove elements from an array by predicate.
     *
     * @static
     * @memberOf _
     * @since 2.0.0
     * @category Array
     * @param {Array} array The array to modify.
     * @param {...*} [values] The values to remove.
     * @returns {Array} Returns `array`.
     * @example
     *
     * var array = ['a', 'b', 'c', 'a', 'b', 'c'];
     *
     * _.pull(array, 'a', 'c');
     * console.log(array);
     * // => ['b', 'b']
     */
    var pull = baseRest(pullAll);

    /**
     * This method is like `_.pull` except that it accepts an array of values to remove.
     *
     * **Note:** Unlike `_.difference`, this method mutates `array`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The array to modify.
     * @param {Array} values The values to remove.
     * @returns {Array} Returns `array`.
     * @example
     *
     * var array = ['a', 'b', 'c', 'a', 'b', 'c'];
     *
     * _.pullAll(array, ['a', 'c']);
     * console.log(array);
     * // => ['b', 'b']
     */
    function pullAll(array, values) {
      return (array && array.length && values && values.length)
        ? basePullAll(array, values)
        : array;
    }

    /**
     * This method is like `_.pullAll` except that it accepts `iteratee` which is
     * invoked for each element of `array` and `values` to generate the criterion
     * by which they're compared. The iteratee is invoked with one argument: (value).
     *
     * **Note:** Unlike `_.differenceBy`, this method mutates `array`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The array to modify.
     * @param {Array} values The values to remove.
     * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
     * @returns {Array} Returns `array`.
     * @example
     *
     * var array = [{ 'x': 1 }, { 'x': 2 }, { 'x': 3 }, { 'x': 1 }];
     *
     * _.pullAllBy(array, [{ 'x': 1 }, { 'x': 3 }], 'x');
     * console.log(array);
     * // => [{ 'x': 2 }]
     */
    function pullAllBy(array, values, iteratee) {
      return (array && array.length && values && values.length)
        ? basePullAll(array, values, getIteratee(iteratee, 2))
        : array;
    }

    /**
     * This method is like `_.pullAll` except that it accepts `comparator` which
     * is invoked to compare elements of `array` to `values`. The comparator is
     * invoked with two arguments: (arrVal, othVal).
     *
     * **Note:** Unlike `_.differenceWith`, this method mutates `array`.
     *
     * @static
     * @memberOf _
     * @since 4.6.0
     * @category Array
     * @param {Array} array The array to modify.
     * @param {Array} values The values to remove.
     * @param {Function} [comparator] The comparator invoked per element.
     * @returns {Array} Returns `array`.
     * @example
     *
     * var array = [{ 'x': 1, 'y': 2 }, { 'x': 3, 'y': 4 }, { 'x': 5, 'y': 6 }];
     *
     * _.pullAllWith(array, [{ 'x': 3, 'y': 4 }], _.isEqual);
     * console.log(array);
     * // => [{ 'x': 1, 'y': 2 }, { 'x': 5, 'y': 6 }]
     */
    function pullAllWith(array, values, comparator) {
      return (array && array.length && values && values.length)
        ? basePullAll(array, values, undefined, comparator)
        : array;
    }

    /**
     * Removes elements from `array` corresponding to `indexes` and returns an
     * array of removed elements.
     *
     * **Note:** Unlike `_.at`, this method mutates `array`.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Array
     * @param {Array} array The array to modify.
     * @param {...(number|number[])} [indexes] The indexes of elements to remove.
     * @returns {Array} Returns the new array of removed elements.
     * @example
     *
     * var array = ['a', 'b', 'c', 'd'];
     * var pulled = _.pullAt(array, [1, 3]);
     *
     * console.log(array);
     * // => ['a', 'c']
     *
     * console.log(pulled);
     * // => ['b', 'd']
     */
    var pullAt = flatRest(function(array, indexes) {
      var length = array == null ? 0 : array.length,
          result = baseAt(array, indexes);

      basePullAt(array, arrayMap(indexes, function(index) {
        return isIndex(index, length) ? +index : index;
      }).sort(compareAscending));

      return result;
    });

    /**
     * Removes all elements from `array` that `predicate` returns truthy for
     * and returns an array of the removed elements. The predicate is invoked
     * with three arguments: (value, index, array).
     *
     * **Note:** Unlike `_.filter`, this method mutates `array`. Use `_.pull`
     * to pull elements from an array by value.
     *
     * @static
     * @memberOf _
     * @since 2.0.0
     * @category Array
     * @param {Array} array The array to modify.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @returns {Array} Returns the new array of removed elements.
     * @example
     *
     * var array = [1, 2, 3, 4];
     * var evens = _.remove(array, function(n) {
     *   return n % 2 == 0;
     * });
     *
     * console.log(array);
     * // => [1, 3]
     *
     * console.log(evens);
     * // => [2, 4]
     */
    function remove(array, predicate) {
      var result = [];
      if (!(array && array.length)) {
        return result;
      }
      var index = -1,
          indexes = [],
          length = array.length;

      predicate = getIteratee(predicate, 3);
      while (++index < length) {
        var value = array[index];
        if (predicate(value, index, array)) {
          result.push(value);
          indexes.push(index);
        }
      }
      basePullAt(array, indexes);
      return result;
    }

    /**
     * Reverses `array` so that the first element becomes the last, the second
     * element becomes the second to last, and so on.
     *
     * **Note:** This method mutates `array` and is based on
     * [`Array#reverse`](https://mdn.io/Array/reverse).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The array to modify.
     * @returns {Array} Returns `array`.
     * @example
     *
     * var array = [1, 2, 3];
     *
     * _.reverse(array);
     * // => [3, 2, 1]
     *
     * console.log(array);
     * // => [3, 2, 1]
     */
    function reverse(array) {
      return array == null ? array : nativeReverse.call(array);
    }

    /**
     * Creates a slice of `array` from `start` up to, but not including, `end`.
     *
     * **Note:** This method is used instead of
     * [`Array#slice`](https://mdn.io/Array/slice) to ensure dense arrays are
     * returned.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Array
     * @param {Array} array The array to slice.
     * @param {number} [start=0] The start position.
     * @param {number} [end=array.length] The end position.
     * @returns {Array} Returns the slice of `array`.
     */
    function slice(array, start, end) {
      var length = array == null ? 0 : array.length;
      if (!length) {
        return [];
      }
      if (end && typeof end != 'number' && isIterateeCall(array, start, end)) {
        start = 0;
        end = length;
      }
      else {
        start = start == null ? 0 : toInteger(start);
        end = end === undefined ? length : toInteger(end);
      }
      return baseSlice(array, start, end);
    }

    /**
     * Uses a binary search to determine the lowest index at which `value`
     * should be inserted into `array` in order to maintain its sort order.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Array
     * @param {Array} array The sorted array to inspect.
     * @param {*} value The value to evaluate.
     * @returns {number} Returns the index at which `value` should be inserted
     *  into `array`.
     * @example
     *
     * _.sortedIndex([30, 50], 40);
     * // => 1
     */
    function sortedIndex(array, value) {
      return baseSortedIndex(array, value);
    }

    /**
     * This method is like `_.sortedIndex` except that it accepts `iteratee`
     * which is invoked for `value` and each element of `array` to compute their
     * sort ranking. The iteratee is invoked with one argument: (value).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The sorted array to inspect.
     * @param {*} value The value to evaluate.
     * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
     * @returns {number} Returns the index at which `value` should be inserted
     *  into `array`.
     * @example
     *
     * var objects = [{ 'x': 4 }, { 'x': 5 }];
     *
     * _.sortedIndexBy(objects, { 'x': 4 }, function(o) { return o.x; });
     * // => 0
     *
     * // The `_.property` iteratee shorthand.
     * _.sortedIndexBy(objects, { 'x': 4 }, 'x');
     * // => 0
     */
    function sortedIndexBy(array, value, iteratee) {
      return baseSortedIndexBy(array, value, getIteratee(iteratee, 2));
    }

    /**
     * This method is like `_.indexOf` except that it performs a binary
     * search on a sorted `array`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @param {*} value The value to search for.
     * @returns {number} Returns the index of the matched value, else `-1`.
     * @example
     *
     * _.sortedIndexOf([4, 5, 5, 5, 6], 5);
     * // => 1
     */
    function sortedIndexOf(array, value) {
      var length = array == null ? 0 : array.length;
      if (length) {
        var index = baseSortedIndex(array, value);
        if (index < length && eq(array[index], value)) {
          return index;
        }
      }
      return -1;
    }

    /**
     * This method is like `_.sortedIndex` except that it returns the highest
     * index at which `value` should be inserted into `array` in order to
     * maintain its sort order.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Array
     * @param {Array} array The sorted array to inspect.
     * @param {*} value The value to evaluate.
     * @returns {number} Returns the index at which `value` should be inserted
     *  into `array`.
     * @example
     *
     * _.sortedLastIndex([4, 5, 5, 5, 6], 5);
     * // => 4
     */
    function sortedLastIndex(array, value) {
      return baseSortedIndex(array, value, true);
    }

    /**
     * This method is like `_.sortedLastIndex` except that it accepts `iteratee`
     * which is invoked for `value` and each element of `array` to compute their
     * sort ranking. The iteratee is invoked with one argument: (value).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The sorted array to inspect.
     * @param {*} value The value to evaluate.
     * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
     * @returns {number} Returns the index at which `value` should be inserted
     *  into `array`.
     * @example
     *
     * var objects = [{ 'x': 4 }, { 'x': 5 }];
     *
     * _.sortedLastIndexBy(objects, { 'x': 4 }, function(o) { return o.x; });
     * // => 1
     *
     * // The `_.property` iteratee shorthand.
     * _.sortedLastIndexBy(objects, { 'x': 4 }, 'x');
     * // => 1
     */
    function sortedLastIndexBy(array, value, iteratee) {
      return baseSortedIndexBy(array, value, getIteratee(iteratee, 2), true);
    }

    /**
     * This method is like `_.lastIndexOf` except that it performs a binary
     * search on a sorted `array`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @param {*} value The value to search for.
     * @returns {number} Returns the index of the matched value, else `-1`.
     * @example
     *
     * _.sortedLastIndexOf([4, 5, 5, 5, 6], 5);
     * // => 3
     */
    function sortedLastIndexOf(array, value) {
      var length = array == null ? 0 : array.length;
      if (length) {
        var index = baseSortedIndex(array, value, true) - 1;
        if (eq(array[index], value)) {
          return index;
        }
      }
      return -1;
    }

    /**
     * This method is like `_.uniq` except that it's designed and optimized
     * for sorted arrays.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @returns {Array} Returns the new duplicate free array.
     * @example
     *
     * _.sortedUniq([1, 1, 2]);
     * // => [1, 2]
     */
    function sortedUniq(array) {
      return (array && array.length)
        ? baseSortedUniq(array)
        : [];
    }

    /**
     * This method is like `_.uniqBy` except that it's designed and optimized
     * for sorted arrays.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @param {Function} [iteratee] The iteratee invoked per element.
     * @returns {Array} Returns the new duplicate free array.
     * @example
     *
     * _.sortedUniqBy([1.1, 1.2, 2.3, 2.4], Math.floor);
     * // => [1.1, 2.3]
     */
    function sortedUniqBy(array, iteratee) {
      return (array && array.length)
        ? baseSortedUniq(array, getIteratee(iteratee, 2))
        : [];
    }

    /**
     * Gets all but the first element of `array`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The array to query.
     * @returns {Array} Returns the slice of `array`.
     * @example
     *
     * _.tail([1, 2, 3]);
     * // => [2, 3]
     */
    function tail(array) {
      var length = array == null ? 0 : array.length;
      return length ? baseSlice(array, 1, length) : [];
    }

    /**
     * Creates a slice of `array` with `n` elements taken from the beginning.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Array
     * @param {Array} array The array to query.
     * @param {number} [n=1] The number of elements to take.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {Array} Returns the slice of `array`.
     * @example
     *
     * _.take([1, 2, 3]);
     * // => [1]
     *
     * _.take([1, 2, 3], 2);
     * // => [1, 2]
     *
     * _.take([1, 2, 3], 5);
     * // => [1, 2, 3]
     *
     * _.take([1, 2, 3], 0);
     * // => []
     */
    function take(array, n, guard) {
      if (!(array && array.length)) {
        return [];
      }
      n = (guard || n === undefined) ? 1 : toInteger(n);
      return baseSlice(array, 0, n < 0 ? 0 : n);
    }

    /**
     * Creates a slice of `array` with `n` elements taken from the end.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Array
     * @param {Array} array The array to query.
     * @param {number} [n=1] The number of elements to take.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {Array} Returns the slice of `array`.
     * @example
     *
     * _.takeRight([1, 2, 3]);
     * // => [3]
     *
     * _.takeRight([1, 2, 3], 2);
     * // => [2, 3]
     *
     * _.takeRight([1, 2, 3], 5);
     * // => [1, 2, 3]
     *
     * _.takeRight([1, 2, 3], 0);
     * // => []
     */
    function takeRight(array, n, guard) {
      var length = array == null ? 0 : array.length;
      if (!length) {
        return [];
      }
      n = (guard || n === undefined) ? 1 : toInteger(n);
      n = length - n;
      return baseSlice(array, n < 0 ? 0 : n, length);
    }

    /**
     * Creates a slice of `array` with elements taken from the end. Elements are
     * taken until `predicate` returns falsey. The predicate is invoked with
     * three arguments: (value, index, array).
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Array
     * @param {Array} array The array to query.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @returns {Array} Returns the slice of `array`.
     * @example
     *
     * var users = [
     *   { 'user': 'barney',  'active': true },
     *   { 'user': 'fred',    'active': false },
     *   { 'user': 'pebbles', 'active': false }
     * ];
     *
     * _.takeRightWhile(users, function(o) { return !o.active; });
     * // => objects for ['fred', 'pebbles']
     *
     * // The `_.matches` iteratee shorthand.
     * _.takeRightWhile(users, { 'user': 'pebbles', 'active': false });
     * // => objects for ['pebbles']
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.takeRightWhile(users, ['active', false]);
     * // => objects for ['fred', 'pebbles']
     *
     * // The `_.property` iteratee shorthand.
     * _.takeRightWhile(users, 'active');
     * // => []
     */
    function takeRightWhile(array, predicate) {
      return (array && array.length)
        ? baseWhile(array, getIteratee(predicate, 3), false, true)
        : [];
    }

    /**
     * Creates a slice of `array` with elements taken from the beginning. Elements
     * are taken until `predicate` returns falsey. The predicate is invoked with
     * three arguments: (value, index, array).
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Array
     * @param {Array} array The array to query.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @returns {Array} Returns the slice of `array`.
     * @example
     *
     * var users = [
     *   { 'user': 'barney',  'active': false },
     *   { 'user': 'fred',    'active': false },
     *   { 'user': 'pebbles', 'active': true }
     * ];
     *
     * _.takeWhile(users, function(o) { return !o.active; });
     * // => objects for ['barney', 'fred']
     *
     * // The `_.matches` iteratee shorthand.
     * _.takeWhile(users, { 'user': 'barney', 'active': false });
     * // => objects for ['barney']
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.takeWhile(users, ['active', false]);
     * // => objects for ['barney', 'fred']
     *
     * // The `_.property` iteratee shorthand.
     * _.takeWhile(users, 'active');
     * // => []
     */
    function takeWhile(array, predicate) {
      return (array && array.length)
        ? baseWhile(array, getIteratee(predicate, 3))
        : [];
    }

    /**
     * Creates an array of unique values, in order, from all given arrays using
     * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
     * for equality comparisons.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Array
     * @param {...Array} [arrays] The arrays to inspect.
     * @returns {Array} Returns the new array of combined values.
     * @example
     *
     * _.union([2], [1, 2]);
     * // => [2, 1]
     */
    var union = baseRest(function(arrays) {
      return baseUniq(baseFlatten(arrays, 1, isArrayLikeObject, true));
    });

    /**
     * This method is like `_.union` except that it accepts `iteratee` which is
     * invoked for each element of each `arrays` to generate the criterion by
     * which uniqueness is computed. Result values are chosen from the first
     * array in which the value occurs. The iteratee is invoked with one argument:
     * (value).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {...Array} [arrays] The arrays to inspect.
     * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
     * @returns {Array} Returns the new array of combined values.
     * @example
     *
     * _.unionBy([2.1], [1.2, 2.3], Math.floor);
     * // => [2.1, 1.2]
     *
     * // The `_.property` iteratee shorthand.
     * _.unionBy([{ 'x': 1 }], [{ 'x': 2 }, { 'x': 1 }], 'x');
     * // => [{ 'x': 1 }, { 'x': 2 }]
     */
    var unionBy = baseRest(function(arrays) {
      var iteratee = last(arrays);
      if (isArrayLikeObject(iteratee)) {
        iteratee = undefined;
      }
      return baseUniq(baseFlatten(arrays, 1, isArrayLikeObject, true), getIteratee(iteratee, 2));
    });

    /**
     * This method is like `_.union` except that it accepts `comparator` which
     * is invoked to compare elements of `arrays`. Result values are chosen from
     * the first array in which the value occurs. The comparator is invoked
     * with two arguments: (arrVal, othVal).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {...Array} [arrays] The arrays to inspect.
     * @param {Function} [comparator] The comparator invoked per element.
     * @returns {Array} Returns the new array of combined values.
     * @example
     *
     * var objects = [{ 'x': 1, 'y': 2 }, { 'x': 2, 'y': 1 }];
     * var others = [{ 'x': 1, 'y': 1 }, { 'x': 1, 'y': 2 }];
     *
     * _.unionWith(objects, others, _.isEqual);
     * // => [{ 'x': 1, 'y': 2 }, { 'x': 2, 'y': 1 }, { 'x': 1, 'y': 1 }]
     */
    var unionWith = baseRest(function(arrays) {
      var comparator = last(arrays);
      comparator = typeof comparator == 'function' ? comparator : undefined;
      return baseUniq(baseFlatten(arrays, 1, isArrayLikeObject, true), undefined, comparator);
    });

    /**
     * Creates a duplicate-free version of an array, using
     * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
     * for equality comparisons, in which only the first occurrence of each element
     * is kept. The order of result values is determined by the order they occur
     * in the array.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @returns {Array} Returns the new duplicate free array.
     * @example
     *
     * _.uniq([2, 1, 2]);
     * // => [2, 1]
     */
    function uniq(array) {
      return (array && array.length) ? baseUniq(array) : [];
    }

    /**
     * This method is like `_.uniq` except that it accepts `iteratee` which is
     * invoked for each element in `array` to generate the criterion by which
     * uniqueness is computed. The order of result values is determined by the
     * order they occur in the array. The iteratee is invoked with one argument:
     * (value).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
     * @returns {Array} Returns the new duplicate free array.
     * @example
     *
     * _.uniqBy([2.1, 1.2, 2.3], Math.floor);
     * // => [2.1, 1.2]
     *
     * // The `_.property` iteratee shorthand.
     * _.uniqBy([{ 'x': 1 }, { 'x': 2 }, { 'x': 1 }], 'x');
     * // => [{ 'x': 1 }, { 'x': 2 }]
     */
    function uniqBy(array, iteratee) {
      return (array && array.length) ? baseUniq(array, getIteratee(iteratee, 2)) : [];
    }

    /**
     * This method is like `_.uniq` except that it accepts `comparator` which
     * is invoked to compare elements of `array`. The order of result values is
     * determined by the order they occur in the array.The comparator is invoked
     * with two arguments: (arrVal, othVal).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @param {Function} [comparator] The comparator invoked per element.
     * @returns {Array} Returns the new duplicate free array.
     * @example
     *
     * var objects = [{ 'x': 1, 'y': 2 }, { 'x': 2, 'y': 1 }, { 'x': 1, 'y': 2 }];
     *
     * _.uniqWith(objects, _.isEqual);
     * // => [{ 'x': 1, 'y': 2 }, { 'x': 2, 'y': 1 }]
     */
    function uniqWith(array, comparator) {
      comparator = typeof comparator == 'function' ? comparator : undefined;
      return (array && array.length) ? baseUniq(array, undefined, comparator) : [];
    }

    /**
     * This method is like `_.zip` except that it accepts an array of grouped
     * elements and creates an array regrouping the elements to their pre-zip
     * configuration.
     *
     * @static
     * @memberOf _
     * @since 1.2.0
     * @category Array
     * @param {Array} array The array of grouped elements to process.
     * @returns {Array} Returns the new array of regrouped elements.
     * @example
     *
     * var zipped = _.zip(['a', 'b'], [1, 2], [true, false]);
     * // => [['a', 1, true], ['b', 2, false]]
     *
     * _.unzip(zipped);
     * // => [['a', 'b'], [1, 2], [true, false]]
     */
    function unzip(array) {
      if (!(array && array.length)) {
        return [];
      }
      var length = 0;
      array = arrayFilter(array, function(group) {
        if (isArrayLikeObject(group)) {
          length = nativeMax(group.length, length);
          return true;
        }
      });
      return baseTimes(length, function(index) {
        return arrayMap(array, baseProperty(index));
      });
    }

    /**
     * This method is like `_.unzip` except that it accepts `iteratee` to specify
     * how regrouped values should be combined. The iteratee is invoked with the
     * elements of each group: (...group).
     *
     * @static
     * @memberOf _
     * @since 3.8.0
     * @category Array
     * @param {Array} array The array of grouped elements to process.
     * @param {Function} [iteratee=_.identity] The function to combine
     *  regrouped values.
     * @returns {Array} Returns the new array of regrouped elements.
     * @example
     *
     * var zipped = _.zip([1, 2], [10, 20], [100, 200]);
     * // => [[1, 10, 100], [2, 20, 200]]
     *
     * _.unzipWith(zipped, _.add);
     * // => [3, 30, 300]
     */
    function unzipWith(array, iteratee) {
      if (!(array && array.length)) {
        return [];
      }
      var result = unzip(array);
      if (iteratee == null) {
        return result;
      }
      return arrayMap(result, function(group) {
        return apply(iteratee, undefined, group);
      });
    }

    /**
     * Creates an array excluding all given values using
     * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
     * for equality comparisons.
     *
     * **Note:** Unlike `_.pull`, this method returns a new array.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Array
     * @param {Array} array The array to inspect.
     * @param {...*} [values] The values to exclude.
     * @returns {Array} Returns the new array of filtered values.
     * @see _.difference, _.xor
     * @example
     *
     * _.without([2, 1, 2, 3], 1, 2);
     * // => [3]
     */
    var without = baseRest(function(array, values) {
      return isArrayLikeObject(array)
        ? baseDifference(array, values)
        : [];
    });

    /**
     * Creates an array of unique values that is the
     * [symmetric difference](https://en.wikipedia.org/wiki/Symmetric_difference)
     * of the given arrays. The order of result values is determined by the order
     * they occur in the arrays.
     *
     * @static
     * @memberOf _
     * @since 2.4.0
     * @category Array
     * @param {...Array} [arrays] The arrays to inspect.
     * @returns {Array} Returns the new array of filtered values.
     * @see _.difference, _.without
     * @example
     *
     * _.xor([2, 1], [2, 3]);
     * // => [1, 3]
     */
    var xor = baseRest(function(arrays) {
      return baseXor(arrayFilter(arrays, isArrayLikeObject));
    });

    /**
     * This method is like `_.xor` except that it accepts `iteratee` which is
     * invoked for each element of each `arrays` to generate the criterion by
     * which by which they're compared. The order of result values is determined
     * by the order they occur in the arrays. The iteratee is invoked with one
     * argument: (value).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {...Array} [arrays] The arrays to inspect.
     * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
     * @returns {Array} Returns the new array of filtered values.
     * @example
     *
     * _.xorBy([2.1, 1.2], [2.3, 3.4], Math.floor);
     * // => [1.2, 3.4]
     *
     * // The `_.property` iteratee shorthand.
     * _.xorBy([{ 'x': 1 }], [{ 'x': 2 }, { 'x': 1 }], 'x');
     * // => [{ 'x': 2 }]
     */
    var xorBy = baseRest(function(arrays) {
      var iteratee = last(arrays);
      if (isArrayLikeObject(iteratee)) {
        iteratee = undefined;
      }
      return baseXor(arrayFilter(arrays, isArrayLikeObject), getIteratee(iteratee, 2));
    });

    /**
     * This method is like `_.xor` except that it accepts `comparator` which is
     * invoked to compare elements of `arrays`. The order of result values is
     * determined by the order they occur in the arrays. The comparator is invoked
     * with two arguments: (arrVal, othVal).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Array
     * @param {...Array} [arrays] The arrays to inspect.
     * @param {Function} [comparator] The comparator invoked per element.
     * @returns {Array} Returns the new array of filtered values.
     * @example
     *
     * var objects = [{ 'x': 1, 'y': 2 }, { 'x': 2, 'y': 1 }];
     * var others = [{ 'x': 1, 'y': 1 }, { 'x': 1, 'y': 2 }];
     *
     * _.xorWith(objects, others, _.isEqual);
     * // => [{ 'x': 2, 'y': 1 }, { 'x': 1, 'y': 1 }]
     */
    var xorWith = baseRest(function(arrays) {
      var comparator = last(arrays);
      comparator = typeof comparator == 'function' ? comparator : undefined;
      return baseXor(arrayFilter(arrays, isArrayLikeObject), undefined, comparator);
    });

    /**
     * Creates an array of grouped elements, the first of which contains the
     * first elements of the given arrays, the second of which contains the
     * second elements of the given arrays, and so on.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Array
     * @param {...Array} [arrays] The arrays to process.
     * @returns {Array} Returns the new array of grouped elements.
     * @example
     *
     * _.zip(['a', 'b'], [1, 2], [true, false]);
     * // => [['a', 1, true], ['b', 2, false]]
     */
    var zip = baseRest(unzip);

    /**
     * This method is like `_.fromPairs` except that it accepts two arrays,
     * one of property identifiers and one of corresponding values.
     *
     * @static
     * @memberOf _
     * @since 0.4.0
     * @category Array
     * @param {Array} [props=[]] The property identifiers.
     * @param {Array} [values=[]] The property values.
     * @returns {Object} Returns the new object.
     * @example
     *
     * _.zipObject(['a', 'b'], [1, 2]);
     * // => { 'a': 1, 'b': 2 }
     */
    function zipObject(props, values) {
      return baseZipObject(props || [], values || [], assignValue);
    }

    /**
     * This method is like `_.zipObject` except that it supports property paths.
     *
     * @static
     * @memberOf _
     * @since 4.1.0
     * @category Array
     * @param {Array} [props=[]] The property identifiers.
     * @param {Array} [values=[]] The property values.
     * @returns {Object} Returns the new object.
     * @example
     *
     * _.zipObjectDeep(['a.b[0].c', 'a.b[1].d'], [1, 2]);
     * // => { 'a': { 'b': [{ 'c': 1 }, { 'd': 2 }] } }
     */
    function zipObjectDeep(props, values) {
      return baseZipObject(props || [], values || [], baseSet);
    }

    /**
     * This method is like `_.zip` except that it accepts `iteratee` to specify
     * how grouped values should be combined. The iteratee is invoked with the
     * elements of each group: (...group).
     *
     * @static
     * @memberOf _
     * @since 3.8.0
     * @category Array
     * @param {...Array} [arrays] The arrays to process.
     * @param {Function} [iteratee=_.identity] The function to combine
     *  grouped values.
     * @returns {Array} Returns the new array of grouped elements.
     * @example
     *
     * _.zipWith([1, 2], [10, 20], [100, 200], function(a, b, c) {
     *   return a + b + c;
     * });
     * // => [111, 222]
     */
    var zipWith = baseRest(function(arrays) {
      var length = arrays.length,
          iteratee = length > 1 ? arrays[length - 1] : undefined;

      iteratee = typeof iteratee == 'function' ? (arrays.pop(), iteratee) : undefined;
      return unzipWith(arrays, iteratee);
    });

    /*------------------------------------------------------------------------*/

    /**
     * Creates a `lodash` wrapper instance that wraps `value` with explicit method
     * chain sequences enabled. The result of such sequences must be unwrapped
     * with `_#value`.
     *
     * @static
     * @memberOf _
     * @since 1.3.0
     * @category Seq
     * @param {*} value The value to wrap.
     * @returns {Object} Returns the new `lodash` wrapper instance.
     * @example
     *
     * var users = [
     *   { 'user': 'barney',  'age': 36 },
     *   { 'user': 'fred',    'age': 40 },
     *   { 'user': 'pebbles', 'age': 1 }
     * ];
     *
     * var youngest = _
     *   .chain(users)
     *   .sortBy('age')
     *   .map(function(o) {
     *     return o.user + ' is ' + o.age;
     *   })
     *   .head()
     *   .value();
     * // => 'pebbles is 1'
     */
    function chain(value) {
      var result = lodash(value);
      result.__chain__ = true;
      return result;
    }

    /**
     * This method invokes `interceptor` and returns `value`. The interceptor
     * is invoked with one argument; (value). The purpose of this method is to
     * "tap into" a method chain sequence in order to modify intermediate results.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Seq
     * @param {*} value The value to provide to `interceptor`.
     * @param {Function} interceptor The function to invoke.
     * @returns {*} Returns `value`.
     * @example
     *
     * _([1, 2, 3])
     *  .tap(function(array) {
     *    // Mutate input array.
     *    array.pop();
     *  })
     *  .reverse()
     *  .value();
     * // => [2, 1]
     */
    function tap(value, interceptor) {
      interceptor(value);
      return value;
    }

    /**
     * This method is like `_.tap` except that it returns the result of `interceptor`.
     * The purpose of this method is to "pass thru" values replacing intermediate
     * results in a method chain sequence.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Seq
     * @param {*} value The value to provide to `interceptor`.
     * @param {Function} interceptor The function to invoke.
     * @returns {*} Returns the result of `interceptor`.
     * @example
     *
     * _('  abc  ')
     *  .chain()
     *  .trim()
     *  .thru(function(value) {
     *    return [value];
     *  })
     *  .value();
     * // => ['abc']
     */
    function thru(value, interceptor) {
      return interceptor(value);
    }

    /**
     * This method is the wrapper version of `_.at`.
     *
     * @name at
     * @memberOf _
     * @since 1.0.0
     * @category Seq
     * @param {...(string|string[])} [paths] The property paths to pick.
     * @returns {Object} Returns the new `lodash` wrapper instance.
     * @example
     *
     * var object = { 'a': [{ 'b': { 'c': 3 } }, 4] };
     *
     * _(object).at(['a[0].b.c', 'a[1]']).value();
     * // => [3, 4]
     */
    var wrapperAt = flatRest(function(paths) {
      var length = paths.length,
          start = length ? paths[0] : 0,
          value = this.__wrapped__,
          interceptor = function(object) { return baseAt(object, paths); };

      if (length > 1 || this.__actions__.length ||
          !(value instanceof LazyWrapper) || !isIndex(start)) {
        return this.thru(interceptor);
      }
      value = value.slice(start, +start + (length ? 1 : 0));
      value.__actions__.push({
        'func': thru,
        'args': [interceptor],
        'thisArg': undefined
      });
      return new LodashWrapper(value, this.__chain__).thru(function(array) {
        if (length && !array.length) {
          array.push(undefined);
        }
        return array;
      });
    });

    /**
     * Creates a `lodash` wrapper instance with explicit method chain sequences enabled.
     *
     * @name chain
     * @memberOf _
     * @since 0.1.0
     * @category Seq
     * @returns {Object} Returns the new `lodash` wrapper instance.
     * @example
     *
     * var users = [
     *   { 'user': 'barney', 'age': 36 },
     *   { 'user': 'fred',   'age': 40 }
     * ];
     *
     * // A sequence without explicit chaining.
     * _(users).head();
     * // => { 'user': 'barney', 'age': 36 }
     *
     * // A sequence with explicit chaining.
     * _(users)
     *   .chain()
     *   .head()
     *   .pick('user')
     *   .value();
     * // => { 'user': 'barney' }
     */
    function wrapperChain() {
      return chain(this);
    }

    /**
     * Executes the chain sequence and returns the wrapped result.
     *
     * @name commit
     * @memberOf _
     * @since 3.2.0
     * @category Seq
     * @returns {Object} Returns the new `lodash` wrapper instance.
     * @example
     *
     * var array = [1, 2];
     * var wrapped = _(array).push(3);
     *
     * console.log(array);
     * // => [1, 2]
     *
     * wrapped = wrapped.commit();
     * console.log(array);
     * // => [1, 2, 3]
     *
     * wrapped.last();
     * // => 3
     *
     * console.log(array);
     * // => [1, 2, 3]
     */
    function wrapperCommit() {
      return new LodashWrapper(this.value(), this.__chain__);
    }

    /**
     * Gets the next value on a wrapped object following the
     * [iterator protocol](https://mdn.io/iteration_protocols#iterator).
     *
     * @name next
     * @memberOf _
     * @since 4.0.0
     * @category Seq
     * @returns {Object} Returns the next iterator value.
     * @example
     *
     * var wrapped = _([1, 2]);
     *
     * wrapped.next();
     * // => { 'done': false, 'value': 1 }
     *
     * wrapped.next();
     * // => { 'done': false, 'value': 2 }
     *
     * wrapped.next();
     * // => { 'done': true, 'value': undefined }
     */
    function wrapperNext() {
      if (this.__values__ === undefined) {
        this.__values__ = toArray(this.value());
      }
      var done = this.__index__ >= this.__values__.length,
          value = done ? undefined : this.__values__[this.__index__++];

      return { 'done': done, 'value': value };
    }

    /**
     * Enables the wrapper to be iterable.
     *
     * @name Symbol.iterator
     * @memberOf _
     * @since 4.0.0
     * @category Seq
     * @returns {Object} Returns the wrapper object.
     * @example
     *
     * var wrapped = _([1, 2]);
     *
     * wrapped[Symbol.iterator]() === wrapped;
     * // => true
     *
     * Array.from(wrapped);
     * // => [1, 2]
     */
    function wrapperToIterator() {
      return this;
    }

    /**
     * Creates a clone of the chain sequence planting `value` as the wrapped value.
     *
     * @name plant
     * @memberOf _
     * @since 3.2.0
     * @category Seq
     * @param {*} value The value to plant.
     * @returns {Object} Returns the new `lodash` wrapper instance.
     * @example
     *
     * function square(n) {
     *   return n * n;
     * }
     *
     * var wrapped = _([1, 2]).map(square);
     * var other = wrapped.plant([3, 4]);
     *
     * other.value();
     * // => [9, 16]
     *
     * wrapped.value();
     * // => [1, 4]
     */
    function wrapperPlant(value) {
      var result,
          parent = this;

      while (parent instanceof baseLodash) {
        var clone = wrapperClone(parent);
        clone.__index__ = 0;
        clone.__values__ = undefined;
        if (result) {
          previous.__wrapped__ = clone;
        } else {
          result = clone;
        }
        var previous = clone;
        parent = parent.__wrapped__;
      }
      previous.__wrapped__ = value;
      return result;
    }

    /**
     * This method is the wrapper version of `_.reverse`.
     *
     * **Note:** This method mutates the wrapped array.
     *
     * @name reverse
     * @memberOf _
     * @since 0.1.0
     * @category Seq
     * @returns {Object} Returns the new `lodash` wrapper instance.
     * @example
     *
     * var array = [1, 2, 3];
     *
     * _(array).reverse().value()
     * // => [3, 2, 1]
     *
     * console.log(array);
     * // => [3, 2, 1]
     */
    function wrapperReverse() {
      var value = this.__wrapped__;
      if (value instanceof LazyWrapper) {
        var wrapped = value;
        if (this.__actions__.length) {
          wrapped = new LazyWrapper(this);
        }
        wrapped = wrapped.reverse();
        wrapped.__actions__.push({
          'func': thru,
          'args': [reverse],
          'thisArg': undefined
        });
        return new LodashWrapper(wrapped, this.__chain__);
      }
      return this.thru(reverse);
    }

    /**
     * Executes the chain sequence to resolve the unwrapped value.
     *
     * @name value
     * @memberOf _
     * @since 0.1.0
     * @alias toJSON, valueOf
     * @category Seq
     * @returns {*} Returns the resolved unwrapped value.
     * @example
     *
     * _([1, 2, 3]).value();
     * // => [1, 2, 3]
     */
    function wrapperValue() {
      return baseWrapperValue(this.__wrapped__, this.__actions__);
    }

    /*------------------------------------------------------------------------*/

    /**
     * Creates an object composed of keys generated from the results of running
     * each element of `collection` thru `iteratee`. The corresponding value of
     * each key is the number of times the key was returned by `iteratee`. The
     * iteratee is invoked with one argument: (value).
     *
     * @static
     * @memberOf _
     * @since 0.5.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [iteratee=_.identity] The iteratee to transform keys.
     * @returns {Object} Returns the composed aggregate object.
     * @example
     *
     * _.countBy([6.1, 4.2, 6.3], Math.floor);
     * // => { '4': 1, '6': 2 }
     *
     * // The `_.property` iteratee shorthand.
     * _.countBy(['one', 'two', 'three'], 'length');
     * // => { '3': 2, '5': 1 }
     */
    var countBy = createAggregator(function(result, value, key) {
      if (hasOwnProperty.call(result, key)) {
        ++result[key];
      } else {
        baseAssignValue(result, key, 1);
      }
    });

    /**
     * Checks if `predicate` returns truthy for **all** elements of `collection`.
     * Iteration is stopped once `predicate` returns falsey. The predicate is
     * invoked with three arguments: (value, index|key, collection).
     *
     * **Note:** This method returns `true` for
     * [empty collections](https://en.wikipedia.org/wiki/Empty_set) because
     * [everything is true](https://en.wikipedia.org/wiki/Vacuous_truth) of
     * elements of empty collections.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {boolean} Returns `true` if all elements pass the predicate check,
     *  else `false`.
     * @example
     *
     * _.every([true, 1, null, 'yes'], Boolean);
     * // => false
     *
     * var users = [
     *   { 'user': 'barney', 'age': 36, 'active': false },
     *   { 'user': 'fred',   'age': 40, 'active': false }
     * ];
     *
     * // The `_.matches` iteratee shorthand.
     * _.every(users, { 'user': 'barney', 'active': false });
     * // => false
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.every(users, ['active', false]);
     * // => true
     *
     * // The `_.property` iteratee shorthand.
     * _.every(users, 'active');
     * // => false
     */
    function every(collection, predicate, guard) {
      var func = isArray(collection) ? arrayEvery : baseEvery;
      if (guard && isIterateeCall(collection, predicate, guard)) {
        predicate = undefined;
      }
      return func(collection, getIteratee(predicate, 3));
    }

    /**
     * Iterates over elements of `collection`, returning an array of all elements
     * `predicate` returns truthy for. The predicate is invoked with three
     * arguments: (value, index|key, collection).
     *
     * **Note:** Unlike `_.remove`, this method returns a new array.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @returns {Array} Returns the new filtered array.
     * @see _.reject
     * @example
     *
     * var users = [
     *   { 'user': 'barney', 'age': 36, 'active': true },
     *   { 'user': 'fred',   'age': 40, 'active': false }
     * ];
     *
     * _.filter(users, function(o) { return !o.active; });
     * // => objects for ['fred']
     *
     * // The `_.matches` iteratee shorthand.
     * _.filter(users, { 'age': 36, 'active': true });
     * // => objects for ['barney']
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.filter(users, ['active', false]);
     * // => objects for ['fred']
     *
     * // The `_.property` iteratee shorthand.
     * _.filter(users, 'active');
     * // => objects for ['barney']
     *
     * // Combining several predicates using `_.overEvery` or `_.overSome`.
     * _.filter(users, _.overSome([{ 'age': 36 }, ['age', 40]]));
     * // => objects for ['fred', 'barney']
     */
    function filter(collection, predicate) {
      var func = isArray(collection) ? arrayFilter : baseFilter;
      return func(collection, getIteratee(predicate, 3));
    }

    /**
     * Iterates over elements of `collection`, returning the first element
     * `predicate` returns truthy for. The predicate is invoked with three
     * arguments: (value, index|key, collection).
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Collection
     * @param {Array|Object} collection The collection to inspect.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @param {number} [fromIndex=0] The index to search from.
     * @returns {*} Returns the matched element, else `undefined`.
     * @example
     *
     * var users = [
     *   { 'user': 'barney',  'age': 36, 'active': true },
     *   { 'user': 'fred',    'age': 40, 'active': false },
     *   { 'user': 'pebbles', 'age': 1,  'active': true }
     * ];
     *
     * _.find(users, function(o) { return o.age < 40; });
     * // => object for 'barney'
     *
     * // The `_.matches` iteratee shorthand.
     * _.find(users, { 'age': 1, 'active': true });
     * // => object for 'pebbles'
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.find(users, ['active', false]);
     * // => object for 'fred'
     *
     * // The `_.property` iteratee shorthand.
     * _.find(users, 'active');
     * // => object for 'barney'
     */
    var find = createFind(findIndex);

    /**
     * This method is like `_.find` except that it iterates over elements of
     * `collection` from right to left.
     *
     * @static
     * @memberOf _
     * @since 2.0.0
     * @category Collection
     * @param {Array|Object} collection The collection to inspect.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @param {number} [fromIndex=collection.length-1] The index to search from.
     * @returns {*} Returns the matched element, else `undefined`.
     * @example
     *
     * _.findLast([1, 2, 3, 4], function(n) {
     *   return n % 2 == 1;
     * });
     * // => 3
     */
    var findLast = createFind(findLastIndex);

    /**
     * Creates a flattened array of values by running each element in `collection`
     * thru `iteratee` and flattening the mapped results. The iteratee is invoked
     * with three arguments: (value, index|key, collection).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @returns {Array} Returns the new flattened array.
     * @example
     *
     * function duplicate(n) {
     *   return [n, n];
     * }
     *
     * _.flatMap([1, 2], duplicate);
     * // => [1, 1, 2, 2]
     */
    function flatMap(collection, iteratee) {
      return baseFlatten(map(collection, iteratee), 1);
    }

    /**
     * This method is like `_.flatMap` except that it recursively flattens the
     * mapped results.
     *
     * @static
     * @memberOf _
     * @since 4.7.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @returns {Array} Returns the new flattened array.
     * @example
     *
     * function duplicate(n) {
     *   return [[[n, n]]];
     * }
     *
     * _.flatMapDeep([1, 2], duplicate);
     * // => [1, 1, 2, 2]
     */
    function flatMapDeep(collection, iteratee) {
      return baseFlatten(map(collection, iteratee), INFINITY);
    }

    /**
     * This method is like `_.flatMap` except that it recursively flattens the
     * mapped results up to `depth` times.
     *
     * @static
     * @memberOf _
     * @since 4.7.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @param {number} [depth=1] The maximum recursion depth.
     * @returns {Array} Returns the new flattened array.
     * @example
     *
     * function duplicate(n) {
     *   return [[[n, n]]];
     * }
     *
     * _.flatMapDepth([1, 2], duplicate, 2);
     * // => [[1, 1], [2, 2]]
     */
    function flatMapDepth(collection, iteratee, depth) {
      depth = depth === undefined ? 1 : toInteger(depth);
      return baseFlatten(map(collection, iteratee), depth);
    }

    /**
     * Iterates over elements of `collection` and invokes `iteratee` for each element.
     * The iteratee is invoked with three arguments: (value, index|key, collection).
     * Iteratee functions may exit iteration early by explicitly returning `false`.
     *
     * **Note:** As with other "Collections" methods, objects with a "length"
     * property are iterated like arrays. To avoid this behavior use `_.forIn`
     * or `_.forOwn` for object iteration.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @alias each
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @returns {Array|Object} Returns `collection`.
     * @see _.forEachRight
     * @example
     *
     * _.forEach([1, 2], function(value) {
     *   console.log(value);
     * });
     * // => Logs `1` then `2`.
     *
     * _.forEach({ 'a': 1, 'b': 2 }, function(value, key) {
     *   console.log(key);
     * });
     * // => Logs 'a' then 'b' (iteration order is not guaranteed).
     */
    function forEach(collection, iteratee) {
      var func = isArray(collection) ? arrayEach : baseEach;
      return func(collection, getIteratee(iteratee, 3));
    }

    /**
     * This method is like `_.forEach` except that it iterates over elements of
     * `collection` from right to left.
     *
     * @static
     * @memberOf _
     * @since 2.0.0
     * @alias eachRight
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @returns {Array|Object} Returns `collection`.
     * @see _.forEach
     * @example
     *
     * _.forEachRight([1, 2], function(value) {
     *   console.log(value);
     * });
     * // => Logs `2` then `1`.
     */
    function forEachRight(collection, iteratee) {
      var func = isArray(collection) ? arrayEachRight : baseEachRight;
      return func(collection, getIteratee(iteratee, 3));
    }

    /**
     * Creates an object composed of keys generated from the results of running
     * each element of `collection` thru `iteratee`. The order of grouped values
     * is determined by the order they occur in `collection`. The corresponding
     * value of each key is an array of elements responsible for generating the
     * key. The iteratee is invoked with one argument: (value).
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [iteratee=_.identity] The iteratee to transform keys.
     * @returns {Object} Returns the composed aggregate object.
     * @example
     *
     * _.groupBy([6.1, 4.2, 6.3], Math.floor);
     * // => { '4': [4.2], '6': [6.1, 6.3] }
     *
     * // The `_.property` iteratee shorthand.
     * _.groupBy(['one', 'two', 'three'], 'length');
     * // => { '3': ['one', 'two'], '5': ['three'] }
     */
    var groupBy = createAggregator(function(result, value, key) {
      if (hasOwnProperty.call(result, key)) {
        result[key].push(value);
      } else {
        baseAssignValue(result, key, [value]);
      }
    });

    /**
     * Checks if `value` is in `collection`. If `collection` is a string, it's
     * checked for a substring of `value`, otherwise
     * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
     * is used for equality comparisons. If `fromIndex` is negative, it's used as
     * the offset from the end of `collection`.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Collection
     * @param {Array|Object|string} collection The collection to inspect.
     * @param {*} value The value to search for.
     * @param {number} [fromIndex=0] The index to search from.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.reduce`.
     * @returns {boolean} Returns `true` if `value` is found, else `false`.
     * @example
     *
     * _.includes([1, 2, 3], 1);
     * // => true
     *
     * _.includes([1, 2, 3], 1, 2);
     * // => false
     *
     * _.includes({ 'a': 1, 'b': 2 }, 1);
     * // => true
     *
     * _.includes('abcd', 'bc');
     * // => true
     */
    function includes(collection, value, fromIndex, guard) {
      collection = isArrayLike(collection) ? collection : values(collection);
      fromIndex = (fromIndex && !guard) ? toInteger(fromIndex) : 0;

      var length = collection.length;
      if (fromIndex < 0) {
        fromIndex = nativeMax(length + fromIndex, 0);
      }
      return isString(collection)
        ? (fromIndex <= length && collection.indexOf(value, fromIndex) > -1)
        : (!!length && baseIndexOf(collection, value, fromIndex) > -1);
    }

    /**
     * Invokes the method at `path` of each element in `collection`, returning
     * an array of the results of each invoked method. Any additional arguments
     * are provided to each invoked method. If `path` is a function, it's invoked
     * for, and `this` bound to, each element in `collection`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Array|Function|string} path The path of the method to invoke or
     *  the function invoked per iteration.
     * @param {...*} [args] The arguments to invoke each method with.
     * @returns {Array} Returns the array of results.
     * @example
     *
     * _.invokeMap([[5, 1, 7], [3, 2, 1]], 'sort');
     * // => [[1, 5, 7], [1, 2, 3]]
     *
     * _.invokeMap([123, 456], String.prototype.split, '');
     * // => [['1', '2', '3'], ['4', '5', '6']]
     */
    var invokeMap = baseRest(function(collection, path, args) {
      var index = -1,
          isFunc = typeof path == 'function',
          result = isArrayLike(collection) ? Array(collection.length) : [];

      baseEach(collection, function(value) {
        result[++index] = isFunc ? apply(path, value, args) : baseInvoke(value, path, args);
      });
      return result;
    });

    /**
     * Creates an object composed of keys generated from the results of running
     * each element of `collection` thru `iteratee`. The corresponding value of
     * each key is the last element responsible for generating the key. The
     * iteratee is invoked with one argument: (value).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [iteratee=_.identity] The iteratee to transform keys.
     * @returns {Object} Returns the composed aggregate object.
     * @example
     *
     * var array = [
     *   { 'dir': 'left', 'code': 97 },
     *   { 'dir': 'right', 'code': 100 }
     * ];
     *
     * _.keyBy(array, function(o) {
     *   return String.fromCharCode(o.code);
     * });
     * // => { 'a': { 'dir': 'left', 'code': 97 }, 'd': { 'dir': 'right', 'code': 100 } }
     *
     * _.keyBy(array, 'dir');
     * // => { 'left': { 'dir': 'left', 'code': 97 }, 'right': { 'dir': 'right', 'code': 100 } }
     */
    var keyBy = createAggregator(function(result, value, key) {
      baseAssignValue(result, key, value);
    });

    /**
     * Creates an array of values by running each element in `collection` thru
     * `iteratee`. The iteratee is invoked with three arguments:
     * (value, index|key, collection).
     *
     * Many lodash methods are guarded to work as iteratees for methods like
     * `_.every`, `_.filter`, `_.map`, `_.mapValues`, `_.reject`, and `_.some`.
     *
     * The guarded methods are:
     * `ary`, `chunk`, `curry`, `curryRight`, `drop`, `dropRight`, `every`,
     * `fill`, `invert`, `parseInt`, `random`, `range`, `rangeRight`, `repeat`,
     * `sampleSize`, `slice`, `some`, `sortBy`, `split`, `take`, `takeRight`,
     * `template`, `trim`, `trimEnd`, `trimStart`, and `words`
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @returns {Array} Returns the new mapped array.
     * @example
     *
     * function square(n) {
     *   return n * n;
     * }
     *
     * _.map([4, 8], square);
     * // => [16, 64]
     *
     * _.map({ 'a': 4, 'b': 8 }, square);
     * // => [16, 64] (iteration order is not guaranteed)
     *
     * var users = [
     *   { 'user': 'barney' },
     *   { 'user': 'fred' }
     * ];
     *
     * // The `_.property` iteratee shorthand.
     * _.map(users, 'user');
     * // => ['barney', 'fred']
     */
    function map(collection, iteratee) {
      var func = isArray(collection) ? arrayMap : baseMap;
      return func(collection, getIteratee(iteratee, 3));
    }

    /**
     * This method is like `_.sortBy` except that it allows specifying the sort
     * orders of the iteratees to sort by. If `orders` is unspecified, all values
     * are sorted in ascending order. Otherwise, specify an order of "desc" for
     * descending or "asc" for ascending sort order of corresponding values.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Array[]|Function[]|Object[]|string[]} [iteratees=[_.identity]]
     *  The iteratees to sort by.
     * @param {string[]} [orders] The sort orders of `iteratees`.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.reduce`.
     * @returns {Array} Returns the new sorted array.
     * @example
     *
     * var users = [
     *   { 'user': 'fred',   'age': 48 },
     *   { 'user': 'barney', 'age': 34 },
     *   { 'user': 'fred',   'age': 40 },
     *   { 'user': 'barney', 'age': 36 }
     * ];
     *
     * // Sort by `user` in ascending order and by `age` in descending order.
     * _.orderBy(users, ['user', 'age'], ['asc', 'desc']);
     * // => objects for [['barney', 36], ['barney', 34], ['fred', 48], ['fred', 40]]
     */
    function orderBy(collection, iteratees, orders, guard) {
      if (collection == null) {
        return [];
      }
      if (!isArray(iteratees)) {
        iteratees = iteratees == null ? [] : [iteratees];
      }
      orders = guard ? undefined : orders;
      if (!isArray(orders)) {
        orders = orders == null ? [] : [orders];
      }
      return baseOrderBy(collection, iteratees, orders);
    }

    /**
     * Creates an array of elements split into two groups, the first of which
     * contains elements `predicate` returns truthy for, the second of which
     * contains elements `predicate` returns falsey for. The predicate is
     * invoked with one argument: (value).
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @returns {Array} Returns the array of grouped elements.
     * @example
     *
     * var users = [
     *   { 'user': 'barney',  'age': 36, 'active': false },
     *   { 'user': 'fred',    'age': 40, 'active': true },
     *   { 'user': 'pebbles', 'age': 1,  'active': false }
     * ];
     *
     * _.partition(users, function(o) { return o.active; });
     * // => objects for [['fred'], ['barney', 'pebbles']]
     *
     * // The `_.matches` iteratee shorthand.
     * _.partition(users, { 'age': 1, 'active': false });
     * // => objects for [['pebbles'], ['barney', 'fred']]
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.partition(users, ['active', false]);
     * // => objects for [['barney', 'pebbles'], ['fred']]
     *
     * // The `_.property` iteratee shorthand.
     * _.partition(users, 'active');
     * // => objects for [['fred'], ['barney', 'pebbles']]
     */
    var partition = createAggregator(function(result, value, key) {
      result[key ? 0 : 1].push(value);
    }, function() { return [[], []]; });

    /**
     * Reduces `collection` to a value which is the accumulated result of running
     * each element in `collection` thru `iteratee`, where each successive
     * invocation is supplied the return value of the previous. If `accumulator`
     * is not given, the first element of `collection` is used as the initial
     * value. The iteratee is invoked with four arguments:
     * (accumulator, value, index|key, collection).
     *
     * Many lodash methods are guarded to work as iteratees for methods like
     * `_.reduce`, `_.reduceRight`, and `_.transform`.
     *
     * The guarded methods are:
     * `assign`, `defaults`, `defaultsDeep`, `includes`, `merge`, `orderBy`,
     * and `sortBy`
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @param {*} [accumulator] The initial value.
     * @returns {*} Returns the accumulated value.
     * @see _.reduceRight
     * @example
     *
     * _.reduce([1, 2], function(sum, n) {
     *   return sum + n;
     * }, 0);
     * // => 3
     *
     * _.reduce({ 'a': 1, 'b': 2, 'c': 1 }, function(result, value, key) {
     *   (result[value] || (result[value] = [])).push(key);
     *   return result;
     * }, {});
     * // => { '1': ['a', 'c'], '2': ['b'] } (iteration order is not guaranteed)
     */
    function reduce(collection, iteratee, accumulator) {
      var func = isArray(collection) ? arrayReduce : baseReduce,
          initAccum = arguments.length < 3;

      return func(collection, getIteratee(iteratee, 4), accumulator, initAccum, baseEach);
    }

    /**
     * This method is like `_.reduce` except that it iterates over elements of
     * `collection` from right to left.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @param {*} [accumulator] The initial value.
     * @returns {*} Returns the accumulated value.
     * @see _.reduce
     * @example
     *
     * var array = [[0, 1], [2, 3], [4, 5]];
     *
     * _.reduceRight(array, function(flattened, other) {
     *   return flattened.concat(other);
     * }, []);
     * // => [4, 5, 2, 3, 0, 1]
     */
    function reduceRight(collection, iteratee, accumulator) {
      var func = isArray(collection) ? arrayReduceRight : baseReduce,
          initAccum = arguments.length < 3;

      return func(collection, getIteratee(iteratee, 4), accumulator, initAccum, baseEachRight);
    }

    /**
     * The opposite of `_.filter`; this method returns the elements of `collection`
     * that `predicate` does **not** return truthy for.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @returns {Array} Returns the new filtered array.
     * @see _.filter
     * @example
     *
     * var users = [
     *   { 'user': 'barney', 'age': 36, 'active': false },
     *   { 'user': 'fred',   'age': 40, 'active': true }
     * ];
     *
     * _.reject(users, function(o) { return !o.active; });
     * // => objects for ['fred']
     *
     * // The `_.matches` iteratee shorthand.
     * _.reject(users, { 'age': 40, 'active': true });
     * // => objects for ['barney']
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.reject(users, ['active', false]);
     * // => objects for ['fred']
     *
     * // The `_.property` iteratee shorthand.
     * _.reject(users, 'active');
     * // => objects for ['barney']
     */
    function reject(collection, predicate) {
      var func = isArray(collection) ? arrayFilter : baseFilter;
      return func(collection, negate(getIteratee(predicate, 3)));
    }

    /**
     * Gets a random element from `collection`.
     *
     * @static
     * @memberOf _
     * @since 2.0.0
     * @category Collection
     * @param {Array|Object} collection The collection to sample.
     * @returns {*} Returns the random element.
     * @example
     *
     * _.sample([1, 2, 3, 4]);
     * // => 2
     */
    function sample(collection) {
      var func = isArray(collection) ? arraySample : baseSample;
      return func(collection);
    }

    /**
     * Gets `n` random elements at unique keys from `collection` up to the
     * size of `collection`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Collection
     * @param {Array|Object} collection The collection to sample.
     * @param {number} [n=1] The number of elements to sample.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {Array} Returns the random elements.
     * @example
     *
     * _.sampleSize([1, 2, 3], 2);
     * // => [3, 1]
     *
     * _.sampleSize([1, 2, 3], 4);
     * // => [2, 3, 1]
     */
    function sampleSize(collection, n, guard) {
      if ((guard ? isIterateeCall(collection, n, guard) : n === undefined)) {
        n = 1;
      } else {
        n = toInteger(n);
      }
      var func = isArray(collection) ? arraySampleSize : baseSampleSize;
      return func(collection, n);
    }

    /**
     * Creates an array of shuffled values, using a version of the
     * [Fisher-Yates shuffle](https://en.wikipedia.org/wiki/Fisher-Yates_shuffle).
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Collection
     * @param {Array|Object} collection The collection to shuffle.
     * @returns {Array} Returns the new shuffled array.
     * @example
     *
     * _.shuffle([1, 2, 3, 4]);
     * // => [4, 1, 3, 2]
     */
    function shuffle(collection) {
      var func = isArray(collection) ? arrayShuffle : baseShuffle;
      return func(collection);
    }

    /**
     * Gets the size of `collection` by returning its length for array-like
     * values or the number of own enumerable string keyed properties for objects.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Collection
     * @param {Array|Object|string} collection The collection to inspect.
     * @returns {number} Returns the collection size.
     * @example
     *
     * _.size([1, 2, 3]);
     * // => 3
     *
     * _.size({ 'a': 1, 'b': 2 });
     * // => 2
     *
     * _.size('pebbles');
     * // => 7
     */
    function size(collection) {
      if (collection == null) {
        return 0;
      }
      if (isArrayLike(collection)) {
        return isString(collection) ? stringSize(collection) : collection.length;
      }
      var tag = getTag(collection);
      if (tag == mapTag || tag == setTag) {
        return collection.size;
      }
      return baseKeys(collection).length;
    }

    /**
     * Checks if `predicate` returns truthy for **any** element of `collection`.
     * Iteration is stopped once `predicate` returns truthy. The predicate is
     * invoked with three arguments: (value, index|key, collection).
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {boolean} Returns `true` if any element passes the predicate check,
     *  else `false`.
     * @example
     *
     * _.some([null, 0, 'yes', false], Boolean);
     * // => true
     *
     * var users = [
     *   { 'user': 'barney', 'active': true },
     *   { 'user': 'fred',   'active': false }
     * ];
     *
     * // The `_.matches` iteratee shorthand.
     * _.some(users, { 'user': 'barney', 'active': false });
     * // => false
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.some(users, ['active', false]);
     * // => true
     *
     * // The `_.property` iteratee shorthand.
     * _.some(users, 'active');
     * // => true
     */
    function some(collection, predicate, guard) {
      var func = isArray(collection) ? arraySome : baseSome;
      if (guard && isIterateeCall(collection, predicate, guard)) {
        predicate = undefined;
      }
      return func(collection, getIteratee(predicate, 3));
    }

    /**
     * Creates an array of elements, sorted in ascending order by the results of
     * running each element in a collection thru each iteratee. This method
     * performs a stable sort, that is, it preserves the original sort order of
     * equal elements. The iteratees are invoked with one argument: (value).
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Collection
     * @param {Array|Object} collection The collection to iterate over.
     * @param {...(Function|Function[])} [iteratees=[_.identity]]
     *  The iteratees to sort by.
     * @returns {Array} Returns the new sorted array.
     * @example
     *
     * var users = [
     *   { 'user': 'fred',   'age': 48 },
     *   { 'user': 'barney', 'age': 36 },
     *   { 'user': 'fred',   'age': 30 },
     *   { 'user': 'barney', 'age': 34 }
     * ];
     *
     * _.sortBy(users, [function(o) { return o.user; }]);
     * // => objects for [['barney', 36], ['barney', 34], ['fred', 48], ['fred', 30]]
     *
     * _.sortBy(users, ['user', 'age']);
     * // => objects for [['barney', 34], ['barney', 36], ['fred', 30], ['fred', 48]]
     */
    var sortBy = baseRest(function(collection, iteratees) {
      if (collection == null) {
        return [];
      }
      var length = iteratees.length;
      if (length > 1 && isIterateeCall(collection, iteratees[0], iteratees[1])) {
        iteratees = [];
      } else if (length > 2 && isIterateeCall(iteratees[0], iteratees[1], iteratees[2])) {
        iteratees = [iteratees[0]];
      }
      return baseOrderBy(collection, baseFlatten(iteratees, 1), []);
    });

    /*------------------------------------------------------------------------*/

    /**
     * Gets the timestamp of the number of milliseconds that have elapsed since
     * the Unix epoch (1 January 1970 00:00:00 UTC).
     *
     * @static
     * @memberOf _
     * @since 2.4.0
     * @category Date
     * @returns {number} Returns the timestamp.
     * @example
     *
     * _.defer(function(stamp) {
     *   console.log(_.now() - stamp);
     * }, _.now());
     * // => Logs the number of milliseconds it took for the deferred invocation.
     */
    var now = ctxNow || function() {
      return root.Date.now();
    };

    /*------------------------------------------------------------------------*/

    /**
     * The opposite of `_.before`; this method creates a function that invokes
     * `func` once it's called `n` or more times.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Function
     * @param {number} n The number of calls before `func` is invoked.
     * @param {Function} func The function to restrict.
     * @returns {Function} Returns the new restricted function.
     * @example
     *
     * var saves = ['profile', 'settings'];
     *
     * var done = _.after(saves.length, function() {
     *   console.log('done saving!');
     * });
     *
     * _.forEach(saves, function(type) {
     *   asyncSave({ 'type': type, 'complete': done });
     * });
     * // => Logs 'done saving!' after the two async saves have completed.
     */
    function after(n, func) {
      if (typeof func != 'function') {
        throw new TypeError(FUNC_ERROR_TEXT);
      }
      n = toInteger(n);
      return function() {
        if (--n < 1) {
          return func.apply(this, arguments);
        }
      };
    }

    /**
     * Creates a function that invokes `func`, with up to `n` arguments,
     * ignoring any additional arguments.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Function
     * @param {Function} func The function to cap arguments for.
     * @param {number} [n=func.length] The arity cap.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {Function} Returns the new capped function.
     * @example
     *
     * _.map(['6', '8', '10'], _.ary(parseInt, 1));
     * // => [6, 8, 10]
     */
    function ary(func, n, guard) {
      n = guard ? undefined : n;
      n = (func && n == null) ? func.length : n;
      return createWrap(func, WRAP_ARY_FLAG, undefined, undefined, undefined, undefined, n);
    }

    /**
     * Creates a function that invokes `func`, with the `this` binding and arguments
     * of the created function, while it's called less than `n` times. Subsequent
     * calls to the created function return the result of the last `func` invocation.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Function
     * @param {number} n The number of calls at which `func` is no longer invoked.
     * @param {Function} func The function to restrict.
     * @returns {Function} Returns the new restricted function.
     * @example
     *
     * jQuery(element).on('click', _.before(5, addContactToList));
     * // => Allows adding up to 4 contacts to the list.
     */
    function before(n, func) {
      var result;
      if (typeof func != 'function') {
        throw new TypeError(FUNC_ERROR_TEXT);
      }
      n = toInteger(n);
      return function() {
        if (--n > 0) {
          result = func.apply(this, arguments);
        }
        if (n <= 1) {
          func = undefined;
        }
        return result;
      };
    }

    /**
     * Creates a function that invokes `func` with the `this` binding of `thisArg`
     * and `partials` prepended to the arguments it receives.
     *
     * The `_.bind.placeholder` value, which defaults to `_` in monolithic builds,
     * may be used as a placeholder for partially applied arguments.
     *
     * **Note:** Unlike native `Function#bind`, this method doesn't set the "length"
     * property of bound functions.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Function
     * @param {Function} func The function to bind.
     * @param {*} thisArg The `this` binding of `func`.
     * @param {...*} [partials] The arguments to be partially applied.
     * @returns {Function} Returns the new bound function.
     * @example
     *
     * function greet(greeting, punctuation) {
     *   return greeting + ' ' + this.user + punctuation;
     * }
     *
     * var object = { 'user': 'fred' };
     *
     * var bound = _.bind(greet, object, 'hi');
     * bound('!');
     * // => 'hi fred!'
     *
     * // Bound with placeholders.
     * var bound = _.bind(greet, object, _, '!');
     * bound('hi');
     * // => 'hi fred!'
     */
    var bind = baseRest(function(func, thisArg, partials) {
      var bitmask = WRAP_BIND_FLAG;
      if (partials.length) {
        var holders = replaceHolders(partials, getHolder(bind));
        bitmask |= WRAP_PARTIAL_FLAG;
      }
      return createWrap(func, bitmask, thisArg, partials, holders);
    });

    /**
     * Creates a function that invokes the method at `object[key]` with `partials`
     * prepended to the arguments it receives.
     *
     * This method differs from `_.bind` by allowing bound functions to reference
     * methods that may be redefined or don't yet exist. See
     * [Peter Michaux's article](http://peter.michaux.ca/articles/lazy-function-definition-pattern)
     * for more details.
     *
     * The `_.bindKey.placeholder` value, which defaults to `_` in monolithic
     * builds, may be used as a placeholder for partially applied arguments.
     *
     * @static
     * @memberOf _
     * @since 0.10.0
     * @category Function
     * @param {Object} object The object to invoke the method on.
     * @param {string} key The key of the method.
     * @param {...*} [partials] The arguments to be partially applied.
     * @returns {Function} Returns the new bound function.
     * @example
     *
     * var object = {
     *   'user': 'fred',
     *   'greet': function(greeting, punctuation) {
     *     return greeting + ' ' + this.user + punctuation;
     *   }
     * };
     *
     * var bound = _.bindKey(object, 'greet', 'hi');
     * bound('!');
     * // => 'hi fred!'
     *
     * object.greet = function(greeting, punctuation) {
     *   return greeting + 'ya ' + this.user + punctuation;
     * };
     *
     * bound('!');
     * // => 'hiya fred!'
     *
     * // Bound with placeholders.
     * var bound = _.bindKey(object, 'greet', _, '!');
     * bound('hi');
     * // => 'hiya fred!'
     */
    var bindKey = baseRest(function(object, key, partials) {
      var bitmask = WRAP_BIND_FLAG | WRAP_BIND_KEY_FLAG;
      if (partials.length) {
        var holders = replaceHolders(partials, getHolder(bindKey));
        bitmask |= WRAP_PARTIAL_FLAG;
      }
      return createWrap(key, bitmask, object, partials, holders);
    });

    /**
     * Creates a function that accepts arguments of `func` and either invokes
     * `func` returning its result, if at least `arity` number of arguments have
     * been provided, or returns a function that accepts the remaining `func`
     * arguments, and so on. The arity of `func` may be specified if `func.length`
     * is not sufficient.
     *
     * The `_.curry.placeholder` value, which defaults to `_` in monolithic builds,
     * may be used as a placeholder for provided arguments.
     *
     * **Note:** This method doesn't set the "length" property of curried functions.
     *
     * @static
     * @memberOf _
     * @since 2.0.0
     * @category Function
     * @param {Function} func The function to curry.
     * @param {number} [arity=func.length] The arity of `func`.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {Function} Returns the new curried function.
     * @example
     *
     * var abc = function(a, b, c) {
     *   return [a, b, c];
     * };
     *
     * var curried = _.curry(abc);
     *
     * curried(1)(2)(3);
     * // => [1, 2, 3]
     *
     * curried(1, 2)(3);
     * // => [1, 2, 3]
     *
     * curried(1, 2, 3);
     * // => [1, 2, 3]
     *
     * // Curried with placeholders.
     * curried(1)(_, 3)(2);
     * // => [1, 2, 3]
     */
    function curry(func, arity, guard) {
      arity = guard ? undefined : arity;
      var result = createWrap(func, WRAP_CURRY_FLAG, undefined, undefined, undefined, undefined, undefined, arity);
      result.placeholder = curry.placeholder;
      return result;
    }

    /**
     * This method is like `_.curry` except that arguments are applied to `func`
     * in the manner of `_.partialRight` instead of `_.partial`.
     *
     * The `_.curryRight.placeholder` value, which defaults to `_` in monolithic
     * builds, may be used as a placeholder for provided arguments.
     *
     * **Note:** This method doesn't set the "length" property of curried functions.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Function
     * @param {Function} func The function to curry.
     * @param {number} [arity=func.length] The arity of `func`.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {Function} Returns the new curried function.
     * @example
     *
     * var abc = function(a, b, c) {
     *   return [a, b, c];
     * };
     *
     * var curried = _.curryRight(abc);
     *
     * curried(3)(2)(1);
     * // => [1, 2, 3]
     *
     * curried(2, 3)(1);
     * // => [1, 2, 3]
     *
     * curried(1, 2, 3);
     * // => [1, 2, 3]
     *
     * // Curried with placeholders.
     * curried(3)(1, _)(2);
     * // => [1, 2, 3]
     */
    function curryRight(func, arity, guard) {
      arity = guard ? undefined : arity;
      var result = createWrap(func, WRAP_CURRY_RIGHT_FLAG, undefined, undefined, undefined, undefined, undefined, arity);
      result.placeholder = curryRight.placeholder;
      return result;
    }

    /**
     * Creates a debounced function that delays invoking `func` until after `wait`
     * milliseconds have elapsed since the last time the debounced function was
     * invoked. The debounced function comes with a `cancel` method to cancel
     * delayed `func` invocations and a `flush` method to immediately invoke them.
     * Provide `options` to indicate whether `func` should be invoked on the
     * leading and/or trailing edge of the `wait` timeout. The `func` is invoked
     * with the last arguments provided to the debounced function. Subsequent
     * calls to the debounced function return the result of the last `func`
     * invocation.
     *
     * **Note:** If `leading` and `trailing` options are `true`, `func` is
     * invoked on the trailing edge of the timeout only if the debounced function
     * is invoked more than once during the `wait` timeout.
     *
     * If `wait` is `0` and `leading` is `false`, `func` invocation is deferred
     * until to the next tick, similar to `setTimeout` with a timeout of `0`.
     *
     * See [David Corbacho's article](https://css-tricks.com/debouncing-throttling-explained-examples/)
     * for details over the differences between `_.debounce` and `_.throttle`.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Function
     * @param {Function} func The function to debounce.
     * @param {number} [wait=0] The number of milliseconds to delay.
     * @param {Object} [options={}] The options object.
     * @param {boolean} [options.leading=false]
     *  Specify invoking on the leading edge of the timeout.
     * @param {number} [options.maxWait]
     *  The maximum time `func` is allowed to be delayed before it's invoked.
     * @param {boolean} [options.trailing=true]
     *  Specify invoking on the trailing edge of the timeout.
     * @returns {Function} Returns the new debounced function.
     * @example
     *
     * // Avoid costly calculations while the window size is in flux.
     * jQuery(window).on('resize', _.debounce(calculateLayout, 150));
     *
     * // Invoke `sendMail` when clicked, debouncing subsequent calls.
     * jQuery(element).on('click', _.debounce(sendMail, 300, {
     *   'leading': true,
     *   'trailing': false
     * }));
     *
     * // Ensure `batchLog` is invoked once after 1 second of debounced calls.
     * var debounced = _.debounce(batchLog, 250, { 'maxWait': 1000 });
     * var source = new EventSource('/stream');
     * jQuery(source).on('message', debounced);
     *
     * // Cancel the trailing debounced invocation.
     * jQuery(window).on('popstate', debounced.cancel);
     */
    function debounce(func, wait, options) {
      var lastArgs,
          lastThis,
          maxWait,
          result,
          timerId,
          lastCallTime,
          lastInvokeTime = 0,
          leading = false,
          maxing = false,
          trailing = true;

      if (typeof func != 'function') {
        throw new TypeError(FUNC_ERROR_TEXT);
      }
      wait = toNumber(wait) || 0;
      if (isObject(options)) {
        leading = !!options.leading;
        maxing = 'maxWait' in options;
        maxWait = maxing ? nativeMax(toNumber(options.maxWait) || 0, wait) : maxWait;
        trailing = 'trailing' in options ? !!options.trailing : trailing;
      }

      function invokeFunc(time) {
        var args = lastArgs,
            thisArg = lastThis;

        lastArgs = lastThis = undefined;
        lastInvokeTime = time;
        result = func.apply(thisArg, args);
        return result;
      }

      function leadingEdge(time) {
        // Reset any `maxWait` timer.
        lastInvokeTime = time;
        // Start the timer for the trailing edge.
        timerId = setTimeout(timerExpired, wait);
        // Invoke the leading edge.
        return leading ? invokeFunc(time) : result;
      }

      function remainingWait(time) {
        var timeSinceLastCall = time - lastCallTime,
            timeSinceLastInvoke = time - lastInvokeTime,
            timeWaiting = wait - timeSinceLastCall;

        return maxing
          ? nativeMin(timeWaiting, maxWait - timeSinceLastInvoke)
          : timeWaiting;
      }

      function shouldInvoke(time) {
        var timeSinceLastCall = time - lastCallTime,
            timeSinceLastInvoke = time - lastInvokeTime;

        // Either this is the first call, activity has stopped and we're at the
        // trailing edge, the system time has gone backwards and we're treating
        // it as the trailing edge, or we've hit the `maxWait` limit.
        return (lastCallTime === undefined || (timeSinceLastCall >= wait) ||
          (timeSinceLastCall < 0) || (maxing && timeSinceLastInvoke >= maxWait));
      }

      function timerExpired() {
        var time = now();
        if (shouldInvoke(time)) {
          return trailingEdge(time);
        }
        // Restart the timer.
        timerId = setTimeout(timerExpired, remainingWait(time));
      }

      function trailingEdge(time) {
        timerId = undefined;

        // Only invoke if we have `lastArgs` which means `func` has been
        // debounced at least once.
        if (trailing && lastArgs) {
          return invokeFunc(time);
        }
        lastArgs = lastThis = undefined;
        return result;
      }

      function cancel() {
        if (timerId !== undefined) {
          clearTimeout(timerId);
        }
        lastInvokeTime = 0;
        lastArgs = lastCallTime = lastThis = timerId = undefined;
      }

      function flush() {
        return timerId === undefined ? result : trailingEdge(now());
      }

      function debounced() {
        var time = now(),
            isInvoking = shouldInvoke(time);

        lastArgs = arguments;
        lastThis = this;
        lastCallTime = time;

        if (isInvoking) {
          if (timerId === undefined) {
            return leadingEdge(lastCallTime);
          }
          if (maxing) {
            // Handle invocations in a tight loop.
            clearTimeout(timerId);
            timerId = setTimeout(timerExpired, wait);
            return invokeFunc(lastCallTime);
          }
        }
        if (timerId === undefined) {
          timerId = setTimeout(timerExpired, wait);
        }
        return result;
      }
      debounced.cancel = cancel;
      debounced.flush = flush;
      return debounced;
    }

    /**
     * Defers invoking the `func` until the current call stack has cleared. Any
     * additional arguments are provided to `func` when it's invoked.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Function
     * @param {Function} func The function to defer.
     * @param {...*} [args] The arguments to invoke `func` with.
     * @returns {number} Returns the timer id.
     * @example
     *
     * _.defer(function(text) {
     *   console.log(text);
     * }, 'deferred');
     * // => Logs 'deferred' after one millisecond.
     */
    var defer = baseRest(function(func, args) {
      return baseDelay(func, 1, args);
    });

    /**
     * Invokes `func` after `wait` milliseconds. Any additional arguments are
     * provided to `func` when it's invoked.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Function
     * @param {Function} func The function to delay.
     * @param {number} wait The number of milliseconds to delay invocation.
     * @param {...*} [args] The arguments to invoke `func` with.
     * @returns {number} Returns the timer id.
     * @example
     *
     * _.delay(function(text) {
     *   console.log(text);
     * }, 1000, 'later');
     * // => Logs 'later' after one second.
     */
    var delay = baseRest(function(func, wait, args) {
      return baseDelay(func, toNumber(wait) || 0, args);
    });

    /**
     * Creates a function that invokes `func` with arguments reversed.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Function
     * @param {Function} func The function to flip arguments for.
     * @returns {Function} Returns the new flipped function.
     * @example
     *
     * var flipped = _.flip(function() {
     *   return _.toArray(arguments);
     * });
     *
     * flipped('a', 'b', 'c', 'd');
     * // => ['d', 'c', 'b', 'a']
     */
    function flip(func) {
      return createWrap(func, WRAP_FLIP_FLAG);
    }

    /**
     * Creates a function that memoizes the result of `func`. If `resolver` is
     * provided, it determines the cache key for storing the result based on the
     * arguments provided to the memoized function. By default, the first argument
     * provided to the memoized function is used as the map cache key. The `func`
     * is invoked with the `this` binding of the memoized function.
     *
     * **Note:** The cache is exposed as the `cache` property on the memoized
     * function. Its creation may be customized by replacing the `_.memoize.Cache`
     * constructor with one whose instances implement the
     * [`Map`](http://ecma-international.org/ecma-262/7.0/#sec-properties-of-the-map-prototype-object)
     * method interface of `clear`, `delete`, `get`, `has`, and `set`.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Function
     * @param {Function} func The function to have its output memoized.
     * @param {Function} [resolver] The function to resolve the cache key.
     * @returns {Function} Returns the new memoized function.
     * @example
     *
     * var object = { 'a': 1, 'b': 2 };
     * var other = { 'c': 3, 'd': 4 };
     *
     * var values = _.memoize(_.values);
     * values(object);
     * // => [1, 2]
     *
     * values(other);
     * // => [3, 4]
     *
     * object.a = 2;
     * values(object);
     * // => [1, 2]
     *
     * // Modify the result cache.
     * values.cache.set(object, ['a', 'b']);
     * values(object);
     * // => ['a', 'b']
     *
     * // Replace `_.memoize.Cache`.
     * _.memoize.Cache = WeakMap;
     */
    function memoize(func, resolver) {
      if (typeof func != 'function' || (resolver != null && typeof resolver != 'function')) {
        throw new TypeError(FUNC_ERROR_TEXT);
      }
      var memoized = function() {
        var args = arguments,
            key = resolver ? resolver.apply(this, args) : args[0],
            cache = memoized.cache;

        if (cache.has(key)) {
          return cache.get(key);
        }
        var result = func.apply(this, args);
        memoized.cache = cache.set(key, result) || cache;
        return result;
      };
      memoized.cache = new (memoize.Cache || MapCache);
      return memoized;
    }

    // Expose `MapCache`.
    memoize.Cache = MapCache;

    /**
     * Creates a function that negates the result of the predicate `func`. The
     * `func` predicate is invoked with the `this` binding and arguments of the
     * created function.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Function
     * @param {Function} predicate The predicate to negate.
     * @returns {Function} Returns the new negated function.
     * @example
     *
     * function isEven(n) {
     *   return n % 2 == 0;
     * }
     *
     * _.filter([1, 2, 3, 4, 5, 6], _.negate(isEven));
     * // => [1, 3, 5]
     */
    function negate(predicate) {
      if (typeof predicate != 'function') {
        throw new TypeError(FUNC_ERROR_TEXT);
      }
      return function() {
        var args = arguments;
        switch (args.length) {
          case 0: return !predicate.call(this);
          case 1: return !predicate.call(this, args[0]);
          case 2: return !predicate.call(this, args[0], args[1]);
          case 3: return !predicate.call(this, args[0], args[1], args[2]);
        }
        return !predicate.apply(this, args);
      };
    }

    /**
     * Creates a function that is restricted to invoking `func` once. Repeat calls
     * to the function return the value of the first invocation. The `func` is
     * invoked with the `this` binding and arguments of the created function.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Function
     * @param {Function} func The function to restrict.
     * @returns {Function} Returns the new restricted function.
     * @example
     *
     * var initialize = _.once(createApplication);
     * initialize();
     * initialize();
     * // => `createApplication` is invoked once
     */
    function once(func) {
      return before(2, func);
    }

    /**
     * Creates a function that invokes `func` with its arguments transformed.
     *
     * @static
     * @since 4.0.0
     * @memberOf _
     * @category Function
     * @param {Function} func The function to wrap.
     * @param {...(Function|Function[])} [transforms=[_.identity]]
     *  The argument transforms.
     * @returns {Function} Returns the new function.
     * @example
     *
     * function doubled(n) {
     *   return n * 2;
     * }
     *
     * function square(n) {
     *   return n * n;
     * }
     *
     * var func = _.overArgs(function(x, y) {
     *   return [x, y];
     * }, [square, doubled]);
     *
     * func(9, 3);
     * // => [81, 6]
     *
     * func(10, 5);
     * // => [100, 10]
     */
    var overArgs = castRest(function(func, transforms) {
      transforms = (transforms.length == 1 && isArray(transforms[0]))
        ? arrayMap(transforms[0], baseUnary(getIteratee()))
        : arrayMap(baseFlatten(transforms, 1), baseUnary(getIteratee()));

      var funcsLength = transforms.length;
      return baseRest(function(args) {
        var index = -1,
            length = nativeMin(args.length, funcsLength);

        while (++index < length) {
          args[index] = transforms[index].call(this, args[index]);
        }
        return apply(func, this, args);
      });
    });

    /**
     * Creates a function that invokes `func` with `partials` prepended to the
     * arguments it receives. This method is like `_.bind` except it does **not**
     * alter the `this` binding.
     *
     * The `_.partial.placeholder` value, which defaults to `_` in monolithic
     * builds, may be used as a placeholder for partially applied arguments.
     *
     * **Note:** This method doesn't set the "length" property of partially
     * applied functions.
     *
     * @static
     * @memberOf _
     * @since 0.2.0
     * @category Function
     * @param {Function} func The function to partially apply arguments to.
     * @param {...*} [partials] The arguments to be partially applied.
     * @returns {Function} Returns the new partially applied function.
     * @example
     *
     * function greet(greeting, name) {
     *   return greeting + ' ' + name;
     * }
     *
     * var sayHelloTo = _.partial(greet, 'hello');
     * sayHelloTo('fred');
     * // => 'hello fred'
     *
     * // Partially applied with placeholders.
     * var greetFred = _.partial(greet, _, 'fred');
     * greetFred('hi');
     * // => 'hi fred'
     */
    var partial = baseRest(function(func, partials) {
      var holders = replaceHolders(partials, getHolder(partial));
      return createWrap(func, WRAP_PARTIAL_FLAG, undefined, partials, holders);
    });

    /**
     * This method is like `_.partial` except that partially applied arguments
     * are appended to the arguments it receives.
     *
     * The `_.partialRight.placeholder` value, which defaults to `_` in monolithic
     * builds, may be used as a placeholder for partially applied arguments.
     *
     * **Note:** This method doesn't set the "length" property of partially
     * applied functions.
     *
     * @static
     * @memberOf _
     * @since 1.0.0
     * @category Function
     * @param {Function} func The function to partially apply arguments to.
     * @param {...*} [partials] The arguments to be partially applied.
     * @returns {Function} Returns the new partially applied function.
     * @example
     *
     * function greet(greeting, name) {
     *   return greeting + ' ' + name;
     * }
     *
     * var greetFred = _.partialRight(greet, 'fred');
     * greetFred('hi');
     * // => 'hi fred'
     *
     * // Partially applied with placeholders.
     * var sayHelloTo = _.partialRight(greet, 'hello', _);
     * sayHelloTo('fred');
     * // => 'hello fred'
     */
    var partialRight = baseRest(function(func, partials) {
      var holders = replaceHolders(partials, getHolder(partialRight));
      return createWrap(func, WRAP_PARTIAL_RIGHT_FLAG, undefined, partials, holders);
    });

    /**
     * Creates a function that invokes `func` with arguments arranged according
     * to the specified `indexes` where the argument value at the first index is
     * provided as the first argument, the argument value at the second index is
     * provided as the second argument, and so on.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Function
     * @param {Function} func The function to rearrange arguments for.
     * @param {...(number|number[])} indexes The arranged argument indexes.
     * @returns {Function} Returns the new function.
     * @example
     *
     * var rearged = _.rearg(function(a, b, c) {
     *   return [a, b, c];
     * }, [2, 0, 1]);
     *
     * rearged('b', 'c', 'a')
     * // => ['a', 'b', 'c']
     */
    var rearg = flatRest(function(func, indexes) {
      return createWrap(func, WRAP_REARG_FLAG, undefined, undefined, undefined, indexes);
    });

    /**
     * Creates a function that invokes `func` with the `this` binding of the
     * created function and arguments from `start` and beyond provided as
     * an array.
     *
     * **Note:** This method is based on the
     * [rest parameter](https://mdn.io/rest_parameters).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Function
     * @param {Function} func The function to apply a rest parameter to.
     * @param {number} [start=func.length-1] The start position of the rest parameter.
     * @returns {Function} Returns the new function.
     * @example
     *
     * var say = _.rest(function(what, names) {
     *   return what + ' ' + _.initial(names).join(', ') +
     *     (_.size(names) > 1 ? ', & ' : '') + _.last(names);
     * });
     *
     * say('hello', 'fred', 'barney', 'pebbles');
     * // => 'hello fred, barney, & pebbles'
     */
    function rest(func, start) {
      if (typeof func != 'function') {
        throw new TypeError(FUNC_ERROR_TEXT);
      }
      start = start === undefined ? start : toInteger(start);
      return baseRest(func, start);
    }

    /**
     * Creates a function that invokes `func` with the `this` binding of the
     * create function and an array of arguments much like
     * [`Function#apply`](http://www.ecma-international.org/ecma-262/7.0/#sec-function.prototype.apply).
     *
     * **Note:** This method is based on the
     * [spread operator](https://mdn.io/spread_operator).
     *
     * @static
     * @memberOf _
     * @since 3.2.0
     * @category Function
     * @param {Function} func The function to spread arguments over.
     * @param {number} [start=0] The start position of the spread.
     * @returns {Function} Returns the new function.
     * @example
     *
     * var say = _.spread(function(who, what) {
     *   return who + ' says ' + what;
     * });
     *
     * say(['fred', 'hello']);
     * // => 'fred says hello'
     *
     * var numbers = Promise.all([
     *   Promise.resolve(40),
     *   Promise.resolve(36)
     * ]);
     *
     * numbers.then(_.spread(function(x, y) {
     *   return x + y;
     * }));
     * // => a Promise of 76
     */
    function spread(func, start) {
      if (typeof func != 'function') {
        throw new TypeError(FUNC_ERROR_TEXT);
      }
      start = start == null ? 0 : nativeMax(toInteger(start), 0);
      return baseRest(function(args) {
        var array = args[start],
            otherArgs = castSlice(args, 0, start);

        if (array) {
          arrayPush(otherArgs, array);
        }
        return apply(func, this, otherArgs);
      });
    }

    /**
     * Creates a throttled function that only invokes `func` at most once per
     * every `wait` milliseconds. The throttled function comes with a `cancel`
     * method to cancel delayed `func` invocations and a `flush` method to
     * immediately invoke them. Provide `options` to indicate whether `func`
     * should be invoked on the leading and/or trailing edge of the `wait`
     * timeout. The `func` is invoked with the last arguments provided to the
     * throttled function. Subsequent calls to the throttled function return the
     * result of the last `func` invocation.
     *
     * **Note:** If `leading` and `trailing` options are `true`, `func` is
     * invoked on the trailing edge of the timeout only if the throttled function
     * is invoked more than once during the `wait` timeout.
     *
     * If `wait` is `0` and `leading` is `false`, `func` invocation is deferred
     * until to the next tick, similar to `setTimeout` with a timeout of `0`.
     *
     * See [David Corbacho's article](https://css-tricks.com/debouncing-throttling-explained-examples/)
     * for details over the differences between `_.throttle` and `_.debounce`.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Function
     * @param {Function} func The function to throttle.
     * @param {number} [wait=0] The number of milliseconds to throttle invocations to.
     * @param {Object} [options={}] The options object.
     * @param {boolean} [options.leading=true]
     *  Specify invoking on the leading edge of the timeout.
     * @param {boolean} [options.trailing=true]
     *  Specify invoking on the trailing edge of the timeout.
     * @returns {Function} Returns the new throttled function.
     * @example
     *
     * // Avoid excessively updating the position while scrolling.
     * jQuery(window).on('scroll', _.throttle(updatePosition, 100));
     *
     * // Invoke `renewToken` when the click event is fired, but not more than once every 5 minutes.
     * var throttled = _.throttle(renewToken, 300000, { 'trailing': false });
     * jQuery(element).on('click', throttled);
     *
     * // Cancel the trailing throttled invocation.
     * jQuery(window).on('popstate', throttled.cancel);
     */
    function throttle(func, wait, options) {
      var leading = true,
          trailing = true;

      if (typeof func != 'function') {
        throw new TypeError(FUNC_ERROR_TEXT);
      }
      if (isObject(options)) {
        leading = 'leading' in options ? !!options.leading : leading;
        trailing = 'trailing' in options ? !!options.trailing : trailing;
      }
      return debounce(func, wait, {
        'leading': leading,
        'maxWait': wait,
        'trailing': trailing
      });
    }

    /**
     * Creates a function that accepts up to one argument, ignoring any
     * additional arguments.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Function
     * @param {Function} func The function to cap arguments for.
     * @returns {Function} Returns the new capped function.
     * @example
     *
     * _.map(['6', '8', '10'], _.unary(parseInt));
     * // => [6, 8, 10]
     */
    function unary(func) {
      return ary(func, 1);
    }

    /**
     * Creates a function that provides `value` to `wrapper` as its first
     * argument. Any additional arguments provided to the function are appended
     * to those provided to the `wrapper`. The wrapper is invoked with the `this`
     * binding of the created function.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Function
     * @param {*} value The value to wrap.
     * @param {Function} [wrapper=identity] The wrapper function.
     * @returns {Function} Returns the new function.
     * @example
     *
     * var p = _.wrap(_.escape, function(func, text) {
     *   return '<p>' + func(text) + '</p>';
     * });
     *
     * p('fred, barney, & pebbles');
     * // => '<p>fred, barney, &amp; pebbles</p>'
     */
    function wrap(value, wrapper) {
      return partial(castFunction(wrapper), value);
    }

    /*------------------------------------------------------------------------*/

    /**
     * Casts `value` as an array if it's not one.
     *
     * @static
     * @memberOf _
     * @since 4.4.0
     * @category Lang
     * @param {*} value The value to inspect.
     * @returns {Array} Returns the cast array.
     * @example
     *
     * _.castArray(1);
     * // => [1]
     *
     * _.castArray({ 'a': 1 });
     * // => [{ 'a': 1 }]
     *
     * _.castArray('abc');
     * // => ['abc']
     *
     * _.castArray(null);
     * // => [null]
     *
     * _.castArray(undefined);
     * // => [undefined]
     *
     * _.castArray();
     * // => []
     *
     * var array = [1, 2, 3];
     * console.log(_.castArray(array) === array);
     * // => true
     */
    function castArray() {
      if (!arguments.length) {
        return [];
      }
      var value = arguments[0];
      return isArray(value) ? value : [value];
    }

    /**
     * Creates a shallow clone of `value`.
     *
     * **Note:** This method is loosely based on the
     * [structured clone algorithm](https://mdn.io/Structured_clone_algorithm)
     * and supports cloning arrays, array buffers, booleans, date objects, maps,
     * numbers, `Object` objects, regexes, sets, strings, symbols, and typed
     * arrays. The own enumerable properties of `arguments` objects are cloned
     * as plain objects. An empty object is returned for uncloneable values such
     * as error objects, functions, DOM nodes, and WeakMaps.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to clone.
     * @returns {*} Returns the cloned value.
     * @see _.cloneDeep
     * @example
     *
     * var objects = [{ 'a': 1 }, { 'b': 2 }];
     *
     * var shallow = _.clone(objects);
     * console.log(shallow[0] === objects[0]);
     * // => true
     */
    function clone(value) {
      return baseClone(value, CLONE_SYMBOLS_FLAG);
    }

    /**
     * This method is like `_.clone` except that it accepts `customizer` which
     * is invoked to produce the cloned value. If `customizer` returns `undefined`,
     * cloning is handled by the method instead. The `customizer` is invoked with
     * up to four arguments; (value [, index|key, object, stack]).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to clone.
     * @param {Function} [customizer] The function to customize cloning.
     * @returns {*} Returns the cloned value.
     * @see _.cloneDeepWith
     * @example
     *
     * function customizer(value) {
     *   if (_.isElement(value)) {
     *     return value.cloneNode(false);
     *   }
     * }
     *
     * var el = _.cloneWith(document.body, customizer);
     *
     * console.log(el === document.body);
     * // => false
     * console.log(el.nodeName);
     * // => 'BODY'
     * console.log(el.childNodes.length);
     * // => 0
     */
    function cloneWith(value, customizer) {
      customizer = typeof customizer == 'function' ? customizer : undefined;
      return baseClone(value, CLONE_SYMBOLS_FLAG, customizer);
    }

    /**
     * This method is like `_.clone` except that it recursively clones `value`.
     *
     * @static
     * @memberOf _
     * @since 1.0.0
     * @category Lang
     * @param {*} value The value to recursively clone.
     * @returns {*} Returns the deep cloned value.
     * @see _.clone
     * @example
     *
     * var objects = [{ 'a': 1 }, { 'b': 2 }];
     *
     * var deep = _.cloneDeep(objects);
     * console.log(deep[0] === objects[0]);
     * // => false
     */
    function cloneDeep(value) {
      return baseClone(value, CLONE_DEEP_FLAG | CLONE_SYMBOLS_FLAG);
    }

    /**
     * This method is like `_.cloneWith` except that it recursively clones `value`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to recursively clone.
     * @param {Function} [customizer] The function to customize cloning.
     * @returns {*} Returns the deep cloned value.
     * @see _.cloneWith
     * @example
     *
     * function customizer(value) {
     *   if (_.isElement(value)) {
     *     return value.cloneNode(true);
     *   }
     * }
     *
     * var el = _.cloneDeepWith(document.body, customizer);
     *
     * console.log(el === document.body);
     * // => false
     * console.log(el.nodeName);
     * // => 'BODY'
     * console.log(el.childNodes.length);
     * // => 20
     */
    function cloneDeepWith(value, customizer) {
      customizer = typeof customizer == 'function' ? customizer : undefined;
      return baseClone(value, CLONE_DEEP_FLAG | CLONE_SYMBOLS_FLAG, customizer);
    }

    /**
     * Checks if `object` conforms to `source` by invoking the predicate
     * properties of `source` with the corresponding property values of `object`.
     *
     * **Note:** This method is equivalent to `_.conforms` when `source` is
     * partially applied.
     *
     * @static
     * @memberOf _
     * @since 4.14.0
     * @category Lang
     * @param {Object} object The object to inspect.
     * @param {Object} source The object of property predicates to conform to.
     * @returns {boolean} Returns `true` if `object` conforms, else `false`.
     * @example
     *
     * var object = { 'a': 1, 'b': 2 };
     *
     * _.conformsTo(object, { 'b': function(n) { return n > 1; } });
     * // => true
     *
     * _.conformsTo(object, { 'b': function(n) { return n > 2; } });
     * // => false
     */
    function conformsTo(object, source) {
      return source == null || baseConformsTo(object, source, keys(source));
    }

    /**
     * Performs a
     * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
     * comparison between two values to determine if they are equivalent.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to compare.
     * @param {*} other The other value to compare.
     * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
     * @example
     *
     * var object = { 'a': 1 };
     * var other = { 'a': 1 };
     *
     * _.eq(object, object);
     * // => true
     *
     * _.eq(object, other);
     * // => false
     *
     * _.eq('a', 'a');
     * // => true
     *
     * _.eq('a', Object('a'));
     * // => false
     *
     * _.eq(NaN, NaN);
     * // => true
     */
    function eq(value, other) {
      return value === other || (value !== value && other !== other);
    }

    /**
     * Checks if `value` is greater than `other`.
     *
     * @static
     * @memberOf _
     * @since 3.9.0
     * @category Lang
     * @param {*} value The value to compare.
     * @param {*} other The other value to compare.
     * @returns {boolean} Returns `true` if `value` is greater than `other`,
     *  else `false`.
     * @see _.lt
     * @example
     *
     * _.gt(3, 1);
     * // => true
     *
     * _.gt(3, 3);
     * // => false
     *
     * _.gt(1, 3);
     * // => false
     */
    var gt = createRelationalOperation(baseGt);

    /**
     * Checks if `value` is greater than or equal to `other`.
     *
     * @static
     * @memberOf _
     * @since 3.9.0
     * @category Lang
     * @param {*} value The value to compare.
     * @param {*} other The other value to compare.
     * @returns {boolean} Returns `true` if `value` is greater than or equal to
     *  `other`, else `false`.
     * @see _.lte
     * @example
     *
     * _.gte(3, 1);
     * // => true
     *
     * _.gte(3, 3);
     * // => true
     *
     * _.gte(1, 3);
     * // => false
     */
    var gte = createRelationalOperation(function(value, other) {
      return value >= other;
    });

    /**
     * Checks if `value` is likely an `arguments` object.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an `arguments` object,
     *  else `false`.
     * @example
     *
     * _.isArguments(function() { return arguments; }());
     * // => true
     *
     * _.isArguments([1, 2, 3]);
     * // => false
     */
    var isArguments = baseIsArguments(function() { return arguments; }()) ? baseIsArguments : function(value) {
      return isObjectLike(value) && hasOwnProperty.call(value, 'callee') &&
        !propertyIsEnumerable.call(value, 'callee');
    };

    /**
     * Checks if `value` is classified as an `Array` object.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an array, else `false`.
     * @example
     *
     * _.isArray([1, 2, 3]);
     * // => true
     *
     * _.isArray(document.body.children);
     * // => false
     *
     * _.isArray('abc');
     * // => false
     *
     * _.isArray(_.noop);
     * // => false
     */
    var isArray = Array.isArray;

    /**
     * Checks if `value` is classified as an `ArrayBuffer` object.
     *
     * @static
     * @memberOf _
     * @since 4.3.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an array buffer, else `false`.
     * @example
     *
     * _.isArrayBuffer(new ArrayBuffer(2));
     * // => true
     *
     * _.isArrayBuffer(new Array(2));
     * // => false
     */
    var isArrayBuffer = nodeIsArrayBuffer ? baseUnary(nodeIsArrayBuffer) : baseIsArrayBuffer;

    /**
     * Checks if `value` is array-like. A value is considered array-like if it's
     * not a function and has a `value.length` that's an integer greater than or
     * equal to `0` and less than or equal to `Number.MAX_SAFE_INTEGER`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is array-like, else `false`.
     * @example
     *
     * _.isArrayLike([1, 2, 3]);
     * // => true
     *
     * _.isArrayLike(document.body.children);
     * // => true
     *
     * _.isArrayLike('abc');
     * // => true
     *
     * _.isArrayLike(_.noop);
     * // => false
     */
    function isArrayLike(value) {
      return value != null && isLength(value.length) && !isFunction(value);
    }

    /**
     * This method is like `_.isArrayLike` except that it also checks if `value`
     * is an object.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an array-like object,
     *  else `false`.
     * @example
     *
     * _.isArrayLikeObject([1, 2, 3]);
     * // => true
     *
     * _.isArrayLikeObject(document.body.children);
     * // => true
     *
     * _.isArrayLikeObject('abc');
     * // => false
     *
     * _.isArrayLikeObject(_.noop);
     * // => false
     */
    function isArrayLikeObject(value) {
      return isObjectLike(value) && isArrayLike(value);
    }

    /**
     * Checks if `value` is classified as a boolean primitive or object.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a boolean, else `false`.
     * @example
     *
     * _.isBoolean(false);
     * // => true
     *
     * _.isBoolean(null);
     * // => false
     */
    function isBoolean(value) {
      return value === true || value === false ||
        (isObjectLike(value) && baseGetTag(value) == boolTag);
    }

    /**
     * Checks if `value` is a buffer.
     *
     * @static
     * @memberOf _
     * @since 4.3.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a buffer, else `false`.
     * @example
     *
     * _.isBuffer(new Buffer(2));
     * // => true
     *
     * _.isBuffer(new Uint8Array(2));
     * // => false
     */
    var isBuffer = nativeIsBuffer || stubFalse;

    /**
     * Checks if `value` is classified as a `Date` object.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a date object, else `false`.
     * @example
     *
     * _.isDate(new Date);
     * // => true
     *
     * _.isDate('Mon April 23 2012');
     * // => false
     */
    var isDate = nodeIsDate ? baseUnary(nodeIsDate) : baseIsDate;

    /**
     * Checks if `value` is likely a DOM element.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a DOM element, else `false`.
     * @example
     *
     * _.isElement(document.body);
     * // => true
     *
     * _.isElement('<body>');
     * // => false
     */
    function isElement(value) {
      return isObjectLike(value) && value.nodeType === 1 && !isPlainObject(value);
    }

    /**
     * Checks if `value` is an empty object, collection, map, or set.
     *
     * Objects are considered empty if they have no own enumerable string keyed
     * properties.
     *
     * Array-like values such as `arguments` objects, arrays, buffers, strings, or
     * jQuery-like collections are considered empty if they have a `length` of `0`.
     * Similarly, maps and sets are considered empty if they have a `size` of `0`.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is empty, else `false`.
     * @example
     *
     * _.isEmpty(null);
     * // => true
     *
     * _.isEmpty(true);
     * // => true
     *
     * _.isEmpty(1);
     * // => true
     *
     * _.isEmpty([1, 2, 3]);
     * // => false
     *
     * _.isEmpty({ 'a': 1 });
     * // => false
     */
    function isEmpty(value) {
      if (value == null) {
        return true;
      }
      if (isArrayLike(value) &&
          (isArray(value) || typeof value == 'string' || typeof value.splice == 'function' ||
            isBuffer(value) || isTypedArray(value) || isArguments(value))) {
        return !value.length;
      }
      var tag = getTag(value);
      if (tag == mapTag || tag == setTag) {
        return !value.size;
      }
      if (isPrototype(value)) {
        return !baseKeys(value).length;
      }
      for (var key in value) {
        if (hasOwnProperty.call(value, key)) {
          return false;
        }
      }
      return true;
    }

    /**
     * Performs a deep comparison between two values to determine if they are
     * equivalent.
     *
     * **Note:** This method supports comparing arrays, array buffers, booleans,
     * date objects, error objects, maps, numbers, `Object` objects, regexes,
     * sets, strings, symbols, and typed arrays. `Object` objects are compared
     * by their own, not inherited, enumerable properties. Functions and DOM
     * nodes are compared by strict equality, i.e. `===`.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to compare.
     * @param {*} other The other value to compare.
     * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
     * @example
     *
     * var object = { 'a': 1 };
     * var other = { 'a': 1 };
     *
     * _.isEqual(object, other);
     * // => true
     *
     * object === other;
     * // => false
     */
    function isEqual(value, other) {
      return baseIsEqual(value, other);
    }

    /**
     * This method is like `_.isEqual` except that it accepts `customizer` which
     * is invoked to compare values. If `customizer` returns `undefined`, comparisons
     * are handled by the method instead. The `customizer` is invoked with up to
     * six arguments: (objValue, othValue [, index|key, object, other, stack]).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to compare.
     * @param {*} other The other value to compare.
     * @param {Function} [customizer] The function to customize comparisons.
     * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
     * @example
     *
     * function isGreeting(value) {
     *   return /^h(?:i|ello)$/.test(value);
     * }
     *
     * function customizer(objValue, othValue) {
     *   if (isGreeting(objValue) && isGreeting(othValue)) {
     *     return true;
     *   }
     * }
     *
     * var array = ['hello', 'goodbye'];
     * var other = ['hi', 'goodbye'];
     *
     * _.isEqualWith(array, other, customizer);
     * // => true
     */
    function isEqualWith(value, other, customizer) {
      customizer = typeof customizer == 'function' ? customizer : undefined;
      var result = customizer ? customizer(value, other) : undefined;
      return result === undefined ? baseIsEqual(value, other, undefined, customizer) : !!result;
    }

    /**
     * Checks if `value` is an `Error`, `EvalError`, `RangeError`, `ReferenceError`,
     * `SyntaxError`, `TypeError`, or `URIError` object.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an error object, else `false`.
     * @example
     *
     * _.isError(new Error);
     * // => true
     *
     * _.isError(Error);
     * // => false
     */
    function isError(value) {
      if (!isObjectLike(value)) {
        return false;
      }
      var tag = baseGetTag(value);
      return tag == errorTag || tag == domExcTag ||
        (typeof value.message == 'string' && typeof value.name == 'string' && !isPlainObject(value));
    }

    /**
     * Checks if `value` is a finite primitive number.
     *
     * **Note:** This method is based on
     * [`Number.isFinite`](https://mdn.io/Number/isFinite).
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a finite number, else `false`.
     * @example
     *
     * _.isFinite(3);
     * // => true
     *
     * _.isFinite(Number.MIN_VALUE);
     * // => true
     *
     * _.isFinite(Infinity);
     * // => false
     *
     * _.isFinite('3');
     * // => false
     */
    function isFinite(value) {
      return typeof value == 'number' && nativeIsFinite(value);
    }

    /**
     * Checks if `value` is classified as a `Function` object.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a function, else `false`.
     * @example
     *
     * _.isFunction(_);
     * // => true
     *
     * _.isFunction(/abc/);
     * // => false
     */
    function isFunction(value) {
      if (!isObject(value)) {
        return false;
      }
      // The use of `Object#toString` avoids issues with the `typeof` operator
      // in Safari 9 which returns 'object' for typed arrays and other constructors.
      var tag = baseGetTag(value);
      return tag == funcTag || tag == genTag || tag == asyncTag || tag == proxyTag;
    }

    /**
     * Checks if `value` is an integer.
     *
     * **Note:** This method is based on
     * [`Number.isInteger`](https://mdn.io/Number/isInteger).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an integer, else `false`.
     * @example
     *
     * _.isInteger(3);
     * // => true
     *
     * _.isInteger(Number.MIN_VALUE);
     * // => false
     *
     * _.isInteger(Infinity);
     * // => false
     *
     * _.isInteger('3');
     * // => false
     */
    function isInteger(value) {
      return typeof value == 'number' && value == toInteger(value);
    }

    /**
     * Checks if `value` is a valid array-like length.
     *
     * **Note:** This method is loosely based on
     * [`ToLength`](http://ecma-international.org/ecma-262/7.0/#sec-tolength).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a valid length, else `false`.
     * @example
     *
     * _.isLength(3);
     * // => true
     *
     * _.isLength(Number.MIN_VALUE);
     * // => false
     *
     * _.isLength(Infinity);
     * // => false
     *
     * _.isLength('3');
     * // => false
     */
    function isLength(value) {
      return typeof value == 'number' &&
        value > -1 && value % 1 == 0 && value <= MAX_SAFE_INTEGER;
    }

    /**
     * Checks if `value` is the
     * [language type](http://www.ecma-international.org/ecma-262/7.0/#sec-ecmascript-language-types)
     * of `Object`. (e.g. arrays, functions, objects, regexes, `new Number(0)`, and `new String('')`)
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an object, else `false`.
     * @example
     *
     * _.isObject({});
     * // => true
     *
     * _.isObject([1, 2, 3]);
     * // => true
     *
     * _.isObject(_.noop);
     * // => true
     *
     * _.isObject(null);
     * // => false
     */
    function isObject(value) {
      var type = typeof value;
      return value != null && (type == 'object' || type == 'function');
    }

    /**
     * Checks if `value` is object-like. A value is object-like if it's not `null`
     * and has a `typeof` result of "object".
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
     * @example
     *
     * _.isObjectLike({});
     * // => true
     *
     * _.isObjectLike([1, 2, 3]);
     * // => true
     *
     * _.isObjectLike(_.noop);
     * // => false
     *
     * _.isObjectLike(null);
     * // => false
     */
    function isObjectLike(value) {
      return value != null && typeof value == 'object';
    }

    /**
     * Checks if `value` is classified as a `Map` object.
     *
     * @static
     * @memberOf _
     * @since 4.3.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a map, else `false`.
     * @example
     *
     * _.isMap(new Map);
     * // => true
     *
     * _.isMap(new WeakMap);
     * // => false
     */
    var isMap = nodeIsMap ? baseUnary(nodeIsMap) : baseIsMap;

    /**
     * Performs a partial deep comparison between `object` and `source` to
     * determine if `object` contains equivalent property values.
     *
     * **Note:** This method is equivalent to `_.matches` when `source` is
     * partially applied.
     *
     * Partial comparisons will match empty array and empty object `source`
     * values against any array or object value, respectively. See `_.isEqual`
     * for a list of supported value comparisons.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Lang
     * @param {Object} object The object to inspect.
     * @param {Object} source The object of property values to match.
     * @returns {boolean} Returns `true` if `object` is a match, else `false`.
     * @example
     *
     * var object = { 'a': 1, 'b': 2 };
     *
     * _.isMatch(object, { 'b': 2 });
     * // => true
     *
     * _.isMatch(object, { 'b': 1 });
     * // => false
     */
    function isMatch(object, source) {
      return object === source || baseIsMatch(object, source, getMatchData(source));
    }

    /**
     * This method is like `_.isMatch` except that it accepts `customizer` which
     * is invoked to compare values. If `customizer` returns `undefined`, comparisons
     * are handled by the method instead. The `customizer` is invoked with five
     * arguments: (objValue, srcValue, index|key, object, source).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {Object} object The object to inspect.
     * @param {Object} source The object of property values to match.
     * @param {Function} [customizer] The function to customize comparisons.
     * @returns {boolean} Returns `true` if `object` is a match, else `false`.
     * @example
     *
     * function isGreeting(value) {
     *   return /^h(?:i|ello)$/.test(value);
     * }
     *
     * function customizer(objValue, srcValue) {
     *   if (isGreeting(objValue) && isGreeting(srcValue)) {
     *     return true;
     *   }
     * }
     *
     * var object = { 'greeting': 'hello' };
     * var source = { 'greeting': 'hi' };
     *
     * _.isMatchWith(object, source, customizer);
     * // => true
     */
    function isMatchWith(object, source, customizer) {
      customizer = typeof customizer == 'function' ? customizer : undefined;
      return baseIsMatch(object, source, getMatchData(source), customizer);
    }

    /**
     * Checks if `value` is `NaN`.
     *
     * **Note:** This method is based on
     * [`Number.isNaN`](https://mdn.io/Number/isNaN) and is not the same as
     * global [`isNaN`](https://mdn.io/isNaN) which returns `true` for
     * `undefined` and other non-number values.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is `NaN`, else `false`.
     * @example
     *
     * _.isNaN(NaN);
     * // => true
     *
     * _.isNaN(new Number(NaN));
     * // => true
     *
     * isNaN(undefined);
     * // => true
     *
     * _.isNaN(undefined);
     * // => false
     */
    function isNaN(value) {
      // An `NaN` primitive is the only value that is not equal to itself.
      // Perform the `toStringTag` check first to avoid errors with some
      // ActiveX objects in IE.
      return isNumber(value) && value != +value;
    }

    /**
     * Checks if `value` is a pristine native function.
     *
     * **Note:** This method can't reliably detect native functions in the presence
     * of the core-js package because core-js circumvents this kind of detection.
     * Despite multiple requests, the core-js maintainer has made it clear: any
     * attempt to fix the detection will be obstructed. As a result, we're left
     * with little choice but to throw an error. Unfortunately, this also affects
     * packages, like [babel-polyfill](https://www.npmjs.com/package/babel-polyfill),
     * which rely on core-js.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a native function,
     *  else `false`.
     * @example
     *
     * _.isNative(Array.prototype.push);
     * // => true
     *
     * _.isNative(_);
     * // => false
     */
    function isNative(value) {
      if (isMaskable(value)) {
        throw new Error(CORE_ERROR_TEXT);
      }
      return baseIsNative(value);
    }

    /**
     * Checks if `value` is `null`.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is `null`, else `false`.
     * @example
     *
     * _.isNull(null);
     * // => true
     *
     * _.isNull(void 0);
     * // => false
     */
    function isNull(value) {
      return value === null;
    }

    /**
     * Checks if `value` is `null` or `undefined`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is nullish, else `false`.
     * @example
     *
     * _.isNil(null);
     * // => true
     *
     * _.isNil(void 0);
     * // => true
     *
     * _.isNil(NaN);
     * // => false
     */
    function isNil(value) {
      return value == null;
    }

    /**
     * Checks if `value` is classified as a `Number` primitive or object.
     *
     * **Note:** To exclude `Infinity`, `-Infinity`, and `NaN`, which are
     * classified as numbers, use the `_.isFinite` method.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a number, else `false`.
     * @example
     *
     * _.isNumber(3);
     * // => true
     *
     * _.isNumber(Number.MIN_VALUE);
     * // => true
     *
     * _.isNumber(Infinity);
     * // => true
     *
     * _.isNumber('3');
     * // => false
     */
    function isNumber(value) {
      return typeof value == 'number' ||
        (isObjectLike(value) && baseGetTag(value) == numberTag);
    }

    /**
     * Checks if `value` is a plain object, that is, an object created by the
     * `Object` constructor or one with a `[[Prototype]]` of `null`.
     *
     * @static
     * @memberOf _
     * @since 0.8.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a plain object, else `false`.
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     * }
     *
     * _.isPlainObject(new Foo);
     * // => false
     *
     * _.isPlainObject([1, 2, 3]);
     * // => false
     *
     * _.isPlainObject({ 'x': 0, 'y': 0 });
     * // => true
     *
     * _.isPlainObject(Object.create(null));
     * // => true
     */
    function isPlainObject(value) {
      if (!isObjectLike(value) || baseGetTag(value) != objectTag) {
        return false;
      }
      var proto = getPrototype(value);
      if (proto === null) {
        return true;
      }
      var Ctor = hasOwnProperty.call(proto, 'constructor') && proto.constructor;
      return typeof Ctor == 'function' && Ctor instanceof Ctor &&
        funcToString.call(Ctor) == objectCtorString;
    }

    /**
     * Checks if `value` is classified as a `RegExp` object.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a regexp, else `false`.
     * @example
     *
     * _.isRegExp(/abc/);
     * // => true
     *
     * _.isRegExp('/abc/');
     * // => false
     */
    var isRegExp = nodeIsRegExp ? baseUnary(nodeIsRegExp) : baseIsRegExp;

    /**
     * Checks if `value` is a safe integer. An integer is safe if it's an IEEE-754
     * double precision number which isn't the result of a rounded unsafe integer.
     *
     * **Note:** This method is based on
     * [`Number.isSafeInteger`](https://mdn.io/Number/isSafeInteger).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a safe integer, else `false`.
     * @example
     *
     * _.isSafeInteger(3);
     * // => true
     *
     * _.isSafeInteger(Number.MIN_VALUE);
     * // => false
     *
     * _.isSafeInteger(Infinity);
     * // => false
     *
     * _.isSafeInteger('3');
     * // => false
     */
    function isSafeInteger(value) {
      return isInteger(value) && value >= -MAX_SAFE_INTEGER && value <= MAX_SAFE_INTEGER;
    }

    /**
     * Checks if `value` is classified as a `Set` object.
     *
     * @static
     * @memberOf _
     * @since 4.3.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a set, else `false`.
     * @example
     *
     * _.isSet(new Set);
     * // => true
     *
     * _.isSet(new WeakSet);
     * // => false
     */
    var isSet = nodeIsSet ? baseUnary(nodeIsSet) : baseIsSet;

    /**
     * Checks if `value` is classified as a `String` primitive or object.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a string, else `false`.
     * @example
     *
     * _.isString('abc');
     * // => true
     *
     * _.isString(1);
     * // => false
     */
    function isString(value) {
      return typeof value == 'string' ||
        (!isArray(value) && isObjectLike(value) && baseGetTag(value) == stringTag);
    }

    /**
     * Checks if `value` is classified as a `Symbol` primitive or object.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a symbol, else `false`.
     * @example
     *
     * _.isSymbol(Symbol.iterator);
     * // => true
     *
     * _.isSymbol('abc');
     * // => false
     */
    function isSymbol(value) {
      return typeof value == 'symbol' ||
        (isObjectLike(value) && baseGetTag(value) == symbolTag);
    }

    /**
     * Checks if `value` is classified as a typed array.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a typed array, else `false`.
     * @example
     *
     * _.isTypedArray(new Uint8Array);
     * // => true
     *
     * _.isTypedArray([]);
     * // => false
     */
    var isTypedArray = nodeIsTypedArray ? baseUnary(nodeIsTypedArray) : baseIsTypedArray;

    /**
     * Checks if `value` is `undefined`.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is `undefined`, else `false`.
     * @example
     *
     * _.isUndefined(void 0);
     * // => true
     *
     * _.isUndefined(null);
     * // => false
     */
    function isUndefined(value) {
      return value === undefined;
    }

    /**
     * Checks if `value` is classified as a `WeakMap` object.
     *
     * @static
     * @memberOf _
     * @since 4.3.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a weak map, else `false`.
     * @example
     *
     * _.isWeakMap(new WeakMap);
     * // => true
     *
     * _.isWeakMap(new Map);
     * // => false
     */
    function isWeakMap(value) {
      return isObjectLike(value) && getTag(value) == weakMapTag;
    }

    /**
     * Checks if `value` is classified as a `WeakSet` object.
     *
     * @static
     * @memberOf _
     * @since 4.3.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a weak set, else `false`.
     * @example
     *
     * _.isWeakSet(new WeakSet);
     * // => true
     *
     * _.isWeakSet(new Set);
     * // => false
     */
    function isWeakSet(value) {
      return isObjectLike(value) && baseGetTag(value) == weakSetTag;
    }

    /**
     * Checks if `value` is less than `other`.
     *
     * @static
     * @memberOf _
     * @since 3.9.0
     * @category Lang
     * @param {*} value The value to compare.
     * @param {*} other The other value to compare.
     * @returns {boolean} Returns `true` if `value` is less than `other`,
     *  else `false`.
     * @see _.gt
     * @example
     *
     * _.lt(1, 3);
     * // => true
     *
     * _.lt(3, 3);
     * // => false
     *
     * _.lt(3, 1);
     * // => false
     */
    var lt = createRelationalOperation(baseLt);

    /**
     * Checks if `value` is less than or equal to `other`.
     *
     * @static
     * @memberOf _
     * @since 3.9.0
     * @category Lang
     * @param {*} value The value to compare.
     * @param {*} other The other value to compare.
     * @returns {boolean} Returns `true` if `value` is less than or equal to
     *  `other`, else `false`.
     * @see _.gte
     * @example
     *
     * _.lte(1, 3);
     * // => true
     *
     * _.lte(3, 3);
     * // => true
     *
     * _.lte(3, 1);
     * // => false
     */
    var lte = createRelationalOperation(function(value, other) {
      return value <= other;
    });

    /**
     * Converts `value` to an array.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Lang
     * @param {*} value The value to convert.
     * @returns {Array} Returns the converted array.
     * @example
     *
     * _.toArray({ 'a': 1, 'b': 2 });
     * // => [1, 2]
     *
     * _.toArray('abc');
     * // => ['a', 'b', 'c']
     *
     * _.toArray(1);
     * // => []
     *
     * _.toArray(null);
     * // => []
     */
    function toArray(value) {
      if (!value) {
        return [];
      }
      if (isArrayLike(value)) {
        return isString(value) ? stringToArray(value) : copyArray(value);
      }
      if (symIterator && value[symIterator]) {
        return iteratorToArray(value[symIterator]());
      }
      var tag = getTag(value),
          func = tag == mapTag ? mapToArray : (tag == setTag ? setToArray : values);

      return func(value);
    }

    /**
     * Converts `value` to a finite number.
     *
     * @static
     * @memberOf _
     * @since 4.12.0
     * @category Lang
     * @param {*} value The value to convert.
     * @returns {number} Returns the converted number.
     * @example
     *
     * _.toFinite(3.2);
     * // => 3.2
     *
     * _.toFinite(Number.MIN_VALUE);
     * // => 5e-324
     *
     * _.toFinite(Infinity);
     * // => 1.7976931348623157e+308
     *
     * _.toFinite('3.2');
     * // => 3.2
     */
    function toFinite(value) {
      if (!value) {
        return value === 0 ? value : 0;
      }
      value = toNumber(value);
      if (value === INFINITY || value === -INFINITY) {
        var sign = (value < 0 ? -1 : 1);
        return sign * MAX_INTEGER;
      }
      return value === value ? value : 0;
    }

    /**
     * Converts `value` to an integer.
     *
     * **Note:** This method is loosely based on
     * [`ToInteger`](http://www.ecma-international.org/ecma-262/7.0/#sec-tointeger).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to convert.
     * @returns {number} Returns the converted integer.
     * @example
     *
     * _.toInteger(3.2);
     * // => 3
     *
     * _.toInteger(Number.MIN_VALUE);
     * // => 0
     *
     * _.toInteger(Infinity);
     * // => 1.7976931348623157e+308
     *
     * _.toInteger('3.2');
     * // => 3
     */
    function toInteger(value) {
      var result = toFinite(value),
          remainder = result % 1;

      return result === result ? (remainder ? result - remainder : result) : 0;
    }

    /**
     * Converts `value` to an integer suitable for use as the length of an
     * array-like object.
     *
     * **Note:** This method is based on
     * [`ToLength`](http://ecma-international.org/ecma-262/7.0/#sec-tolength).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to convert.
     * @returns {number} Returns the converted integer.
     * @example
     *
     * _.toLength(3.2);
     * // => 3
     *
     * _.toLength(Number.MIN_VALUE);
     * // => 0
     *
     * _.toLength(Infinity);
     * // => 4294967295
     *
     * _.toLength('3.2');
     * // => 3
     */
    function toLength(value) {
      return value ? baseClamp(toInteger(value), 0, MAX_ARRAY_LENGTH) : 0;
    }

    /**
     * Converts `value` to a number.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to process.
     * @returns {number} Returns the number.
     * @example
     *
     * _.toNumber(3.2);
     * // => 3.2
     *
     * _.toNumber(Number.MIN_VALUE);
     * // => 5e-324
     *
     * _.toNumber(Infinity);
     * // => Infinity
     *
     * _.toNumber('3.2');
     * // => 3.2
     */
    function toNumber(value) {
      if (typeof value == 'number') {
        return value;
      }
      if (isSymbol(value)) {
        return NAN;
      }
      if (isObject(value)) {
        var other = typeof value.valueOf == 'function' ? value.valueOf() : value;
        value = isObject(other) ? (other + '') : other;
      }
      if (typeof value != 'string') {
        return value === 0 ? value : +value;
      }
      value = baseTrim(value);
      var isBinary = reIsBinary.test(value);
      return (isBinary || reIsOctal.test(value))
        ? freeParseInt(value.slice(2), isBinary ? 2 : 8)
        : (reIsBadHex.test(value) ? NAN : +value);
    }

    /**
     * Converts `value` to a plain object flattening inherited enumerable string
     * keyed properties of `value` to own properties of the plain object.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Lang
     * @param {*} value The value to convert.
     * @returns {Object} Returns the converted plain object.
     * @example
     *
     * function Foo() {
     *   this.b = 2;
     * }
     *
     * Foo.prototype.c = 3;
     *
     * _.assign({ 'a': 1 }, new Foo);
     * // => { 'a': 1, 'b': 2 }
     *
     * _.assign({ 'a': 1 }, _.toPlainObject(new Foo));
     * // => { 'a': 1, 'b': 2, 'c': 3 }
     */
    function toPlainObject(value) {
      return copyObject(value, keysIn(value));
    }

    /**
     * Converts `value` to a safe integer. A safe integer can be compared and
     * represented correctly.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to convert.
     * @returns {number} Returns the converted integer.
     * @example
     *
     * _.toSafeInteger(3.2);
     * // => 3
     *
     * _.toSafeInteger(Number.MIN_VALUE);
     * // => 0
     *
     * _.toSafeInteger(Infinity);
     * // => 9007199254740991
     *
     * _.toSafeInteger('3.2');
     * // => 3
     */
    function toSafeInteger(value) {
      return value
        ? baseClamp(toInteger(value), -MAX_SAFE_INTEGER, MAX_SAFE_INTEGER)
        : (value === 0 ? value : 0);
    }

    /**
     * Converts `value` to a string. An empty string is returned for `null`
     * and `undefined` values. The sign of `-0` is preserved.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to convert.
     * @returns {string} Returns the converted string.
     * @example
     *
     * _.toString(null);
     * // => ''
     *
     * _.toString(-0);
     * // => '-0'
     *
     * _.toString([1, 2, 3]);
     * // => '1,2,3'
     */
    function toString(value) {
      return value == null ? '' : baseToString(value);
    }

    /*------------------------------------------------------------------------*/

    /**
     * Assigns own enumerable string keyed properties of source objects to the
     * destination object. Source objects are applied from left to right.
     * Subsequent sources overwrite property assignments of previous sources.
     *
     * **Note:** This method mutates `object` and is loosely based on
     * [`Object.assign`](https://mdn.io/Object/assign).
     *
     * @static
     * @memberOf _
     * @since 0.10.0
     * @category Object
     * @param {Object} object The destination object.
     * @param {...Object} [sources] The source objects.
     * @returns {Object} Returns `object`.
     * @see _.assignIn
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     * }
     *
     * function Bar() {
     *   this.c = 3;
     * }
     *
     * Foo.prototype.b = 2;
     * Bar.prototype.d = 4;
     *
     * _.assign({ 'a': 0 }, new Foo, new Bar);
     * // => { 'a': 1, 'c': 3 }
     */
    var assign = createAssigner(function(object, source) {
      if (isPrototype(source) || isArrayLike(source)) {
        copyObject(source, keys(source), object);
        return;
      }
      for (var key in source) {
        if (hasOwnProperty.call(source, key)) {
          assignValue(object, key, source[key]);
        }
      }
    });

    /**
     * This method is like `_.assign` except that it iterates over own and
     * inherited source properties.
     *
     * **Note:** This method mutates `object`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @alias extend
     * @category Object
     * @param {Object} object The destination object.
     * @param {...Object} [sources] The source objects.
     * @returns {Object} Returns `object`.
     * @see _.assign
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     * }
     *
     * function Bar() {
     *   this.c = 3;
     * }
     *
     * Foo.prototype.b = 2;
     * Bar.prototype.d = 4;
     *
     * _.assignIn({ 'a': 0 }, new Foo, new Bar);
     * // => { 'a': 1, 'b': 2, 'c': 3, 'd': 4 }
     */
    var assignIn = createAssigner(function(object, source) {
      copyObject(source, keysIn(source), object);
    });

    /**
     * This method is like `_.assignIn` except that it accepts `customizer`
     * which is invoked to produce the assigned values. If `customizer` returns
     * `undefined`, assignment is handled by the method instead. The `customizer`
     * is invoked with five arguments: (objValue, srcValue, key, object, source).
     *
     * **Note:** This method mutates `object`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @alias extendWith
     * @category Object
     * @param {Object} object The destination object.
     * @param {...Object} sources The source objects.
     * @param {Function} [customizer] The function to customize assigned values.
     * @returns {Object} Returns `object`.
     * @see _.assignWith
     * @example
     *
     * function customizer(objValue, srcValue) {
     *   return _.isUndefined(objValue) ? srcValue : objValue;
     * }
     *
     * var defaults = _.partialRight(_.assignInWith, customizer);
     *
     * defaults({ 'a': 1 }, { 'b': 2 }, { 'a': 3 });
     * // => { 'a': 1, 'b': 2 }
     */
    var assignInWith = createAssigner(function(object, source, srcIndex, customizer) {
      copyObject(source, keysIn(source), object, customizer);
    });

    /**
     * This method is like `_.assign` except that it accepts `customizer`
     * which is invoked to produce the assigned values. If `customizer` returns
     * `undefined`, assignment is handled by the method instead. The `customizer`
     * is invoked with five arguments: (objValue, srcValue, key, object, source).
     *
     * **Note:** This method mutates `object`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Object
     * @param {Object} object The destination object.
     * @param {...Object} sources The source objects.
     * @param {Function} [customizer] The function to customize assigned values.
     * @returns {Object} Returns `object`.
     * @see _.assignInWith
     * @example
     *
     * function customizer(objValue, srcValue) {
     *   return _.isUndefined(objValue) ? srcValue : objValue;
     * }
     *
     * var defaults = _.partialRight(_.assignWith, customizer);
     *
     * defaults({ 'a': 1 }, { 'b': 2 }, { 'a': 3 });
     * // => { 'a': 1, 'b': 2 }
     */
    var assignWith = createAssigner(function(object, source, srcIndex, customizer) {
      copyObject(source, keys(source), object, customizer);
    });

    /**
     * Creates an array of values corresponding to `paths` of `object`.
     *
     * @static
     * @memberOf _
     * @since 1.0.0
     * @category Object
     * @param {Object} object The object to iterate over.
     * @param {...(string|string[])} [paths] The property paths to pick.
     * @returns {Array} Returns the picked values.
     * @example
     *
     * var object = { 'a': [{ 'b': { 'c': 3 } }, 4] };
     *
     * _.at(object, ['a[0].b.c', 'a[1]']);
     * // => [3, 4]
     */
    var at = flatRest(baseAt);

    /**
     * Creates an object that inherits from the `prototype` object. If a
     * `properties` object is given, its own enumerable string keyed properties
     * are assigned to the created object.
     *
     * @static
     * @memberOf _
     * @since 2.3.0
     * @category Object
     * @param {Object} prototype The object to inherit from.
     * @param {Object} [properties] The properties to assign to the object.
     * @returns {Object} Returns the new object.
     * @example
     *
     * function Shape() {
     *   this.x = 0;
     *   this.y = 0;
     * }
     *
     * function Circle() {
     *   Shape.call(this);
     * }
     *
     * Circle.prototype = _.create(Shape.prototype, {
     *   'constructor': Circle
     * });
     *
     * var circle = new Circle;
     * circle instanceof Circle;
     * // => true
     *
     * circle instanceof Shape;
     * // => true
     */
    function create(prototype, properties) {
      var result = baseCreate(prototype);
      return properties == null ? result : baseAssign(result, properties);
    }

    /**
     * Assigns own and inherited enumerable string keyed properties of source
     * objects to the destination object for all destination properties that
     * resolve to `undefined`. Source objects are applied from left to right.
     * Once a property is set, additional values of the same property are ignored.
     *
     * **Note:** This method mutates `object`.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Object
     * @param {Object} object The destination object.
     * @param {...Object} [sources] The source objects.
     * @returns {Object} Returns `object`.
     * @see _.defaultsDeep
     * @example
     *
     * _.defaults({ 'a': 1 }, { 'b': 2 }, { 'a': 3 });
     * // => { 'a': 1, 'b': 2 }
     */
    var defaults = baseRest(function(object, sources) {
      object = Object(object);

      var index = -1;
      var length = sources.length;
      var guard = length > 2 ? sources[2] : undefined;

      if (guard && isIterateeCall(sources[0], sources[1], guard)) {
        length = 1;
      }

      while (++index < length) {
        var source = sources[index];
        var props = keysIn(source);
        var propsIndex = -1;
        var propsLength = props.length;

        while (++propsIndex < propsLength) {
          var key = props[propsIndex];
          var value = object[key];

          if (value === undefined ||
              (eq(value, objectProto[key]) && !hasOwnProperty.call(object, key))) {
            object[key] = source[key];
          }
        }
      }

      return object;
    });

    /**
     * This method is like `_.defaults` except that it recursively assigns
     * default properties.
     *
     * **Note:** This method mutates `object`.
     *
     * @static
     * @memberOf _
     * @since 3.10.0
     * @category Object
     * @param {Object} object The destination object.
     * @param {...Object} [sources] The source objects.
     * @returns {Object} Returns `object`.
     * @see _.defaults
     * @example
     *
     * _.defaultsDeep({ 'a': { 'b': 2 } }, { 'a': { 'b': 1, 'c': 3 } });
     * // => { 'a': { 'b': 2, 'c': 3 } }
     */
    var defaultsDeep = baseRest(function(args) {
      args.push(undefined, customDefaultsMerge);
      return apply(mergeWith, undefined, args);
    });

    /**
     * This method is like `_.find` except that it returns the key of the first
     * element `predicate` returns truthy for instead of the element itself.
     *
     * @static
     * @memberOf _
     * @since 1.1.0
     * @category Object
     * @param {Object} object The object to inspect.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @returns {string|undefined} Returns the key of the matched element,
     *  else `undefined`.
     * @example
     *
     * var users = {
     *   'barney':  { 'age': 36, 'active': true },
     *   'fred':    { 'age': 40, 'active': false },
     *   'pebbles': { 'age': 1,  'active': true }
     * };
     *
     * _.findKey(users, function(o) { return o.age < 40; });
     * // => 'barney' (iteration order is not guaranteed)
     *
     * // The `_.matches` iteratee shorthand.
     * _.findKey(users, { 'age': 1, 'active': true });
     * // => 'pebbles'
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.findKey(users, ['active', false]);
     * // => 'fred'
     *
     * // The `_.property` iteratee shorthand.
     * _.findKey(users, 'active');
     * // => 'barney'
     */
    function findKey(object, predicate) {
      return baseFindKey(object, getIteratee(predicate, 3), baseForOwn);
    }

    /**
     * This method is like `_.findKey` except that it iterates over elements of
     * a collection in the opposite order.
     *
     * @static
     * @memberOf _
     * @since 2.0.0
     * @category Object
     * @param {Object} object The object to inspect.
     * @param {Function} [predicate=_.identity] The function invoked per iteration.
     * @returns {string|undefined} Returns the key of the matched element,
     *  else `undefined`.
     * @example
     *
     * var users = {
     *   'barney':  { 'age': 36, 'active': true },
     *   'fred':    { 'age': 40, 'active': false },
     *   'pebbles': { 'age': 1,  'active': true }
     * };
     *
     * _.findLastKey(users, function(o) { return o.age < 40; });
     * // => returns 'pebbles' assuming `_.findKey` returns 'barney'
     *
     * // The `_.matches` iteratee shorthand.
     * _.findLastKey(users, { 'age': 36, 'active': true });
     * // => 'barney'
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.findLastKey(users, ['active', false]);
     * // => 'fred'
     *
     * // The `_.property` iteratee shorthand.
     * _.findLastKey(users, 'active');
     * // => 'pebbles'
     */
    function findLastKey(object, predicate) {
      return baseFindKey(object, getIteratee(predicate, 3), baseForOwnRight);
    }

    /**
     * Iterates over own and inherited enumerable string keyed properties of an
     * object and invokes `iteratee` for each property. The iteratee is invoked
     * with three arguments: (value, key, object). Iteratee functions may exit
     * iteration early by explicitly returning `false`.
     *
     * @static
     * @memberOf _
     * @since 0.3.0
     * @category Object
     * @param {Object} object The object to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @returns {Object} Returns `object`.
     * @see _.forInRight
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     *   this.b = 2;
     * }
     *
     * Foo.prototype.c = 3;
     *
     * _.forIn(new Foo, function(value, key) {
     *   console.log(key);
     * });
     * // => Logs 'a', 'b', then 'c' (iteration order is not guaranteed).
     */
    function forIn(object, iteratee) {
      return object == null
        ? object
        : baseFor(object, getIteratee(iteratee, 3), keysIn);
    }

    /**
     * This method is like `_.forIn` except that it iterates over properties of
     * `object` in the opposite order.
     *
     * @static
     * @memberOf _
     * @since 2.0.0
     * @category Object
     * @param {Object} object The object to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @returns {Object} Returns `object`.
     * @see _.forIn
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     *   this.b = 2;
     * }
     *
     * Foo.prototype.c = 3;
     *
     * _.forInRight(new Foo, function(value, key) {
     *   console.log(key);
     * });
     * // => Logs 'c', 'b', then 'a' assuming `_.forIn` logs 'a', 'b', then 'c'.
     */
    function forInRight(object, iteratee) {
      return object == null
        ? object
        : baseForRight(object, getIteratee(iteratee, 3), keysIn);
    }

    /**
     * Iterates over own enumerable string keyed properties of an object and
     * invokes `iteratee` for each property. The iteratee is invoked with three
     * arguments: (value, key, object). Iteratee functions may exit iteration
     * early by explicitly returning `false`.
     *
     * @static
     * @memberOf _
     * @since 0.3.0
     * @category Object
     * @param {Object} object The object to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @returns {Object} Returns `object`.
     * @see _.forOwnRight
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     *   this.b = 2;
     * }
     *
     * Foo.prototype.c = 3;
     *
     * _.forOwn(new Foo, function(value, key) {
     *   console.log(key);
     * });
     * // => Logs 'a' then 'b' (iteration order is not guaranteed).
     */
    function forOwn(object, iteratee) {
      return object && baseForOwn(object, getIteratee(iteratee, 3));
    }

    /**
     * This method is like `_.forOwn` except that it iterates over properties of
     * `object` in the opposite order.
     *
     * @static
     * @memberOf _
     * @since 2.0.0
     * @category Object
     * @param {Object} object The object to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @returns {Object} Returns `object`.
     * @see _.forOwn
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     *   this.b = 2;
     * }
     *
     * Foo.prototype.c = 3;
     *
     * _.forOwnRight(new Foo, function(value, key) {
     *   console.log(key);
     * });
     * // => Logs 'b' then 'a' assuming `_.forOwn` logs 'a' then 'b'.
     */
    function forOwnRight(object, iteratee) {
      return object && baseForOwnRight(object, getIteratee(iteratee, 3));
    }

    /**
     * Creates an array of function property names from own enumerable properties
     * of `object`.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Object
     * @param {Object} object The object to inspect.
     * @returns {Array} Returns the function names.
     * @see _.functionsIn
     * @example
     *
     * function Foo() {
     *   this.a = _.constant('a');
     *   this.b = _.constant('b');
     * }
     *
     * Foo.prototype.c = _.constant('c');
     *
     * _.functions(new Foo);
     * // => ['a', 'b']
     */
    function functions(object) {
      return object == null ? [] : baseFunctions(object, keys(object));
    }

    /**
     * Creates an array of function property names from own and inherited
     * enumerable properties of `object`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Object
     * @param {Object} object The object to inspect.
     * @returns {Array} Returns the function names.
     * @see _.functions
     * @example
     *
     * function Foo() {
     *   this.a = _.constant('a');
     *   this.b = _.constant('b');
     * }
     *
     * Foo.prototype.c = _.constant('c');
     *
     * _.functionsIn(new Foo);
     * // => ['a', 'b', 'c']
     */
    function functionsIn(object) {
      return object == null ? [] : baseFunctions(object, keysIn(object));
    }

    /**
     * Gets the value at `path` of `object`. If the resolved value is
     * `undefined`, the `defaultValue` is returned in its place.
     *
     * @static
     * @memberOf _
     * @since 3.7.0
     * @category Object
     * @param {Object} object The object to query.
     * @param {Array|string} path The path of the property to get.
     * @param {*} [defaultValue] The value returned for `undefined` resolved values.
     * @returns {*} Returns the resolved value.
     * @example
     *
     * var object = { 'a': [{ 'b': { 'c': 3 } }] };
     *
     * _.get(object, 'a[0].b.c');
     * // => 3
     *
     * _.get(object, ['a', '0', 'b', 'c']);
     * // => 3
     *
     * _.get(object, 'a.b.c', 'default');
     * // => 'default'
     */
    function get(object, path, defaultValue) {
      var result = object == null ? undefined : baseGet(object, path);
      return result === undefined ? defaultValue : result;
    }

    /**
     * Checks if `path` is a direct property of `object`.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Object
     * @param {Object} object The object to query.
     * @param {Array|string} path The path to check.
     * @returns {boolean} Returns `true` if `path` exists, else `false`.
     * @example
     *
     * var object = { 'a': { 'b': 2 } };
     * var other = _.create({ 'a': _.create({ 'b': 2 }) });
     *
     * _.has(object, 'a');
     * // => true
     *
     * _.has(object, 'a.b');
     * // => true
     *
     * _.has(object, ['a', 'b']);
     * // => true
     *
     * _.has(other, 'a');
     * // => false
     */
    function has(object, path) {
      return object != null && hasPath(object, path, baseHas);
    }

    /**
     * Checks if `path` is a direct or inherited property of `object`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Object
     * @param {Object} object The object to query.
     * @param {Array|string} path The path to check.
     * @returns {boolean} Returns `true` if `path` exists, else `false`.
     * @example
     *
     * var object = _.create({ 'a': _.create({ 'b': 2 }) });
     *
     * _.hasIn(object, 'a');
     * // => true
     *
     * _.hasIn(object, 'a.b');
     * // => true
     *
     * _.hasIn(object, ['a', 'b']);
     * // => true
     *
     * _.hasIn(object, 'b');
     * // => false
     */
    function hasIn(object, path) {
      return object != null && hasPath(object, path, baseHasIn);
    }

    /**
     * Creates an object composed of the inverted keys and values of `object`.
     * If `object` contains duplicate values, subsequent values overwrite
     * property assignments of previous values.
     *
     * @static
     * @memberOf _
     * @since 0.7.0
     * @category Object
     * @param {Object} object The object to invert.
     * @returns {Object} Returns the new inverted object.
     * @example
     *
     * var object = { 'a': 1, 'b': 2, 'c': 1 };
     *
     * _.invert(object);
     * // => { '1': 'c', '2': 'b' }
     */
    var invert = createInverter(function(result, value, key) {
      if (value != null &&
          typeof value.toString != 'function') {
        value = nativeObjectToString.call(value);
      }

      result[value] = key;
    }, constant(identity));

    /**
     * This method is like `_.invert` except that the inverted object is generated
     * from the results of running each element of `object` thru `iteratee`. The
     * corresponding inverted value of each inverted key is an array of keys
     * responsible for generating the inverted value. The iteratee is invoked
     * with one argument: (value).
     *
     * @static
     * @memberOf _
     * @since 4.1.0
     * @category Object
     * @param {Object} object The object to invert.
     * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
     * @returns {Object} Returns the new inverted object.
     * @example
     *
     * var object = { 'a': 1, 'b': 2, 'c': 1 };
     *
     * _.invertBy(object);
     * // => { '1': ['a', 'c'], '2': ['b'] }
     *
     * _.invertBy(object, function(value) {
     *   return 'group' + value;
     * });
     * // => { 'group1': ['a', 'c'], 'group2': ['b'] }
     */
    var invertBy = createInverter(function(result, value, key) {
      if (value != null &&
          typeof value.toString != 'function') {
        value = nativeObjectToString.call(value);
      }

      if (hasOwnProperty.call(result, value)) {
        result[value].push(key);
      } else {
        result[value] = [key];
      }
    }, getIteratee);

    /**
     * Invokes the method at `path` of `object`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Object
     * @param {Object} object The object to query.
     * @param {Array|string} path The path of the method to invoke.
     * @param {...*} [args] The arguments to invoke the method with.
     * @returns {*} Returns the result of the invoked method.
     * @example
     *
     * var object = { 'a': [{ 'b': { 'c': [1, 2, 3, 4] } }] };
     *
     * _.invoke(object, 'a[0].b.c.slice', 1, 3);
     * // => [2, 3]
     */
    var invoke = baseRest(baseInvoke);

    /**
     * Creates an array of the own enumerable property names of `object`.
     *
     * **Note:** Non-object values are coerced to objects. See the
     * [ES spec](http://ecma-international.org/ecma-262/7.0/#sec-object.keys)
     * for more details.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Object
     * @param {Object} object The object to query.
     * @returns {Array} Returns the array of property names.
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     *   this.b = 2;
     * }
     *
     * Foo.prototype.c = 3;
     *
     * _.keys(new Foo);
     * // => ['a', 'b'] (iteration order is not guaranteed)
     *
     * _.keys('hi');
     * // => ['0', '1']
     */
    function keys(object) {
      return isArrayLike(object) ? arrayLikeKeys(object) : baseKeys(object);
    }

    /**
     * Creates an array of the own and inherited enumerable property names of `object`.
     *
     * **Note:** Non-object values are coerced to objects.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Object
     * @param {Object} object The object to query.
     * @returns {Array} Returns the array of property names.
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     *   this.b = 2;
     * }
     *
     * Foo.prototype.c = 3;
     *
     * _.keysIn(new Foo);
     * // => ['a', 'b', 'c'] (iteration order is not guaranteed)
     */
    function keysIn(object) {
      return isArrayLike(object) ? arrayLikeKeys(object, true) : baseKeysIn(object);
    }

    /**
     * The opposite of `_.mapValues`; this method creates an object with the
     * same values as `object` and keys generated by running each own enumerable
     * string keyed property of `object` thru `iteratee`. The iteratee is invoked
     * with three arguments: (value, key, object).
     *
     * @static
     * @memberOf _
     * @since 3.8.0
     * @category Object
     * @param {Object} object The object to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @returns {Object} Returns the new mapped object.
     * @see _.mapValues
     * @example
     *
     * _.mapKeys({ 'a': 1, 'b': 2 }, function(value, key) {
     *   return key + value;
     * });
     * // => { 'a1': 1, 'b2': 2 }
     */
    function mapKeys(object, iteratee) {
      var result = {};
      iteratee = getIteratee(iteratee, 3);

      baseForOwn(object, function(value, key, object) {
        baseAssignValue(result, iteratee(value, key, object), value);
      });
      return result;
    }

    /**
     * Creates an object with the same keys as `object` and values generated
     * by running each own enumerable string keyed property of `object` thru
     * `iteratee`. The iteratee is invoked with three arguments:
     * (value, key, object).
     *
     * @static
     * @memberOf _
     * @since 2.4.0
     * @category Object
     * @param {Object} object The object to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @returns {Object} Returns the new mapped object.
     * @see _.mapKeys
     * @example
     *
     * var users = {
     *   'fred':    { 'user': 'fred',    'age': 40 },
     *   'pebbles': { 'user': 'pebbles', 'age': 1 }
     * };
     *
     * _.mapValues(users, function(o) { return o.age; });
     * // => { 'fred': 40, 'pebbles': 1 } (iteration order is not guaranteed)
     *
     * // The `_.property` iteratee shorthand.
     * _.mapValues(users, 'age');
     * // => { 'fred': 40, 'pebbles': 1 } (iteration order is not guaranteed)
     */
    function mapValues(object, iteratee) {
      var result = {};
      iteratee = getIteratee(iteratee, 3);

      baseForOwn(object, function(value, key, object) {
        baseAssignValue(result, key, iteratee(value, key, object));
      });
      return result;
    }

    /**
     * This method is like `_.assign` except that it recursively merges own and
     * inherited enumerable string keyed properties of source objects into the
     * destination object. Source properties that resolve to `undefined` are
     * skipped if a destination value exists. Array and plain object properties
     * are merged recursively. Other objects and value types are overridden by
     * assignment. Source objects are applied from left to right. Subsequent
     * sources overwrite property assignments of previous sources.
     *
     * **Note:** This method mutates `object`.
     *
     * @static
     * @memberOf _
     * @since 0.5.0
     * @category Object
     * @param {Object} object The destination object.
     * @param {...Object} [sources] The source objects.
     * @returns {Object} Returns `object`.
     * @example
     *
     * var object = {
     *   'a': [{ 'b': 2 }, { 'd': 4 }]
     * };
     *
     * var other = {
     *   'a': [{ 'c': 3 }, { 'e': 5 }]
     * };
     *
     * _.merge(object, other);
     * // => { 'a': [{ 'b': 2, 'c': 3 }, { 'd': 4, 'e': 5 }] }
     */
    var merge = createAssigner(function(object, source, srcIndex) {
      baseMerge(object, source, srcIndex);
    });

    /**
     * This method is like `_.merge` except that it accepts `customizer` which
     * is invoked to produce the merged values of the destination and source
     * properties. If `customizer` returns `undefined`, merging is handled by the
     * method instead. The `customizer` is invoked with six arguments:
     * (objValue, srcValue, key, object, source, stack).
     *
     * **Note:** This method mutates `object`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Object
     * @param {Object} object The destination object.
     * @param {...Object} sources The source objects.
     * @param {Function} customizer The function to customize assigned values.
     * @returns {Object} Returns `object`.
     * @example
     *
     * function customizer(objValue, srcValue) {
     *   if (_.isArray(objValue)) {
     *     return objValue.concat(srcValue);
     *   }
     * }
     *
     * var object = { 'a': [1], 'b': [2] };
     * var other = { 'a': [3], 'b': [4] };
     *
     * _.mergeWith(object, other, customizer);
     * // => { 'a': [1, 3], 'b': [2, 4] }
     */
    var mergeWith = createAssigner(function(object, source, srcIndex, customizer) {
      baseMerge(object, source, srcIndex, customizer);
    });

    /**
     * The opposite of `_.pick`; this method creates an object composed of the
     * own and inherited enumerable property paths of `object` that are not omitted.
     *
     * **Note:** This method is considerably slower than `_.pick`.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Object
     * @param {Object} object The source object.
     * @param {...(string|string[])} [paths] The property paths to omit.
     * @returns {Object} Returns the new object.
     * @example
     *
     * var object = { 'a': 1, 'b': '2', 'c': 3 };
     *
     * _.omit(object, ['a', 'c']);
     * // => { 'b': '2' }
     */
    var omit = flatRest(function(object, paths) {
      var result = {};
      if (object == null) {
        return result;
      }
      var isDeep = false;
      paths = arrayMap(paths, function(path) {
        path = castPath(path, object);
        isDeep || (isDeep = path.length > 1);
        return path;
      });
      copyObject(object, getAllKeysIn(object), result);
      if (isDeep) {
        result = baseClone(result, CLONE_DEEP_FLAG | CLONE_FLAT_FLAG | CLONE_SYMBOLS_FLAG, customOmitClone);
      }
      var length = paths.length;
      while (length--) {
        baseUnset(result, paths[length]);
      }
      return result;
    });

    /**
     * The opposite of `_.pickBy`; this method creates an object composed of
     * the own and inherited enumerable string keyed properties of `object` that
     * `predicate` doesn't return truthy for. The predicate is invoked with two
     * arguments: (value, key).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Object
     * @param {Object} object The source object.
     * @param {Function} [predicate=_.identity] The function invoked per property.
     * @returns {Object} Returns the new object.
     * @example
     *
     * var object = { 'a': 1, 'b': '2', 'c': 3 };
     *
     * _.omitBy(object, _.isNumber);
     * // => { 'b': '2' }
     */
    function omitBy(object, predicate) {
      return pickBy(object, negate(getIteratee(predicate)));
    }

    /**
     * Creates an object composed of the picked `object` properties.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Object
     * @param {Object} object The source object.
     * @param {...(string|string[])} [paths] The property paths to pick.
     * @returns {Object} Returns the new object.
     * @example
     *
     * var object = { 'a': 1, 'b': '2', 'c': 3 };
     *
     * _.pick(object, ['a', 'c']);
     * // => { 'a': 1, 'c': 3 }
     */
    var pick = flatRest(function(object, paths) {
      return object == null ? {} : basePick(object, paths);
    });

    /**
     * Creates an object composed of the `object` properties `predicate` returns
     * truthy for. The predicate is invoked with two arguments: (value, key).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Object
     * @param {Object} object The source object.
     * @param {Function} [predicate=_.identity] The function invoked per property.
     * @returns {Object} Returns the new object.
     * @example
     *
     * var object = { 'a': 1, 'b': '2', 'c': 3 };
     *
     * _.pickBy(object, _.isNumber);
     * // => { 'a': 1, 'c': 3 }
     */
    function pickBy(object, predicate) {
      if (object == null) {
        return {};
      }
      var props = arrayMap(getAllKeysIn(object), function(prop) {
        return [prop];
      });
      predicate = getIteratee(predicate);
      return basePickBy(object, props, function(value, path) {
        return predicate(value, path[0]);
      });
    }

    /**
     * This method is like `_.get` except that if the resolved value is a
     * function it's invoked with the `this` binding of its parent object and
     * its result is returned.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Object
     * @param {Object} object The object to query.
     * @param {Array|string} path The path of the property to resolve.
     * @param {*} [defaultValue] The value returned for `undefined` resolved values.
     * @returns {*} Returns the resolved value.
     * @example
     *
     * var object = { 'a': [{ 'b': { 'c1': 3, 'c2': _.constant(4) } }] };
     *
     * _.result(object, 'a[0].b.c1');
     * // => 3
     *
     * _.result(object, 'a[0].b.c2');
     * // => 4
     *
     * _.result(object, 'a[0].b.c3', 'default');
     * // => 'default'
     *
     * _.result(object, 'a[0].b.c3', _.constant('default'));
     * // => 'default'
     */
    function result(object, path, defaultValue) {
      path = castPath(path, object);

      var index = -1,
          length = path.length;

      // Ensure the loop is entered when path is empty.
      if (!length) {
        length = 1;
        object = undefined;
      }
      while (++index < length) {
        var value = object == null ? undefined : object[toKey(path[index])];
        if (value === undefined) {
          index = length;
          value = defaultValue;
        }
        object = isFunction(value) ? value.call(object) : value;
      }
      return object;
    }

    /**
     * Sets the value at `path` of `object`. If a portion of `path` doesn't exist,
     * it's created. Arrays are created for missing index properties while objects
     * are created for all other missing properties. Use `_.setWith` to customize
     * `path` creation.
     *
     * **Note:** This method mutates `object`.
     *
     * @static
     * @memberOf _
     * @since 3.7.0
     * @category Object
     * @param {Object} object The object to modify.
     * @param {Array|string} path The path of the property to set.
     * @param {*} value The value to set.
     * @returns {Object} Returns `object`.
     * @example
     *
     * var object = { 'a': [{ 'b': { 'c': 3 } }] };
     *
     * _.set(object, 'a[0].b.c', 4);
     * console.log(object.a[0].b.c);
     * // => 4
     *
     * _.set(object, ['x', '0', 'y', 'z'], 5);
     * console.log(object.x[0].y.z);
     * // => 5
     */
    function set(object, path, value) {
      return object == null ? object : baseSet(object, path, value);
    }

    /**
     * This method is like `_.set` except that it accepts `customizer` which is
     * invoked to produce the objects of `path`.  If `customizer` returns `undefined`
     * path creation is handled by the method instead. The `customizer` is invoked
     * with three arguments: (nsValue, key, nsObject).
     *
     * **Note:** This method mutates `object`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Object
     * @param {Object} object The object to modify.
     * @param {Array|string} path The path of the property to set.
     * @param {*} value The value to set.
     * @param {Function} [customizer] The function to customize assigned values.
     * @returns {Object} Returns `object`.
     * @example
     *
     * var object = {};
     *
     * _.setWith(object, '[0][1]', 'a', Object);
     * // => { '0': { '1': 'a' } }
     */
    function setWith(object, path, value, customizer) {
      customizer = typeof customizer == 'function' ? customizer : undefined;
      return object == null ? object : baseSet(object, path, value, customizer);
    }

    /**
     * Creates an array of own enumerable string keyed-value pairs for `object`
     * which can be consumed by `_.fromPairs`. If `object` is a map or set, its
     * entries are returned.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @alias entries
     * @category Object
     * @param {Object} object The object to query.
     * @returns {Array} Returns the key-value pairs.
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     *   this.b = 2;
     * }
     *
     * Foo.prototype.c = 3;
     *
     * _.toPairs(new Foo);
     * // => [['a', 1], ['b', 2]] (iteration order is not guaranteed)
     */
    var toPairs = createToPairs(keys);

    /**
     * Creates an array of own and inherited enumerable string keyed-value pairs
     * for `object` which can be consumed by `_.fromPairs`. If `object` is a map
     * or set, its entries are returned.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @alias entriesIn
     * @category Object
     * @param {Object} object The object to query.
     * @returns {Array} Returns the key-value pairs.
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     *   this.b = 2;
     * }
     *
     * Foo.prototype.c = 3;
     *
     * _.toPairsIn(new Foo);
     * // => [['a', 1], ['b', 2], ['c', 3]] (iteration order is not guaranteed)
     */
    var toPairsIn = createToPairs(keysIn);

    /**
     * An alternative to `_.reduce`; this method transforms `object` to a new
     * `accumulator` object which is the result of running each of its own
     * enumerable string keyed properties thru `iteratee`, with each invocation
     * potentially mutating the `accumulator` object. If `accumulator` is not
     * provided, a new object with the same `[[Prototype]]` will be used. The
     * iteratee is invoked with four arguments: (accumulator, value, key, object).
     * Iteratee functions may exit iteration early by explicitly returning `false`.
     *
     * @static
     * @memberOf _
     * @since 1.3.0
     * @category Object
     * @param {Object} object The object to iterate over.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @param {*} [accumulator] The custom accumulator value.
     * @returns {*} Returns the accumulated value.
     * @example
     *
     * _.transform([2, 3, 4], function(result, n) {
     *   result.push(n *= n);
     *   return n % 2 == 0;
     * }, []);
     * // => [4, 9]
     *
     * _.transform({ 'a': 1, 'b': 2, 'c': 1 }, function(result, value, key) {
     *   (result[value] || (result[value] = [])).push(key);
     * }, {});
     * // => { '1': ['a', 'c'], '2': ['b'] }
     */
    function transform(object, iteratee, accumulator) {
      var isArr = isArray(object),
          isArrLike = isArr || isBuffer(object) || isTypedArray(object);

      iteratee = getIteratee(iteratee, 4);
      if (accumulator == null) {
        var Ctor = object && object.constructor;
        if (isArrLike) {
          accumulator = isArr ? new Ctor : [];
        }
        else if (isObject(object)) {
          accumulator = isFunction(Ctor) ? baseCreate(getPrototype(object)) : {};
        }
        else {
          accumulator = {};
        }
      }
      (isArrLike ? arrayEach : baseForOwn)(object, function(value, index, object) {
        return iteratee(accumulator, value, index, object);
      });
      return accumulator;
    }

    /**
     * Removes the property at `path` of `object`.
     *
     * **Note:** This method mutates `object`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Object
     * @param {Object} object The object to modify.
     * @param {Array|string} path The path of the property to unset.
     * @returns {boolean} Returns `true` if the property is deleted, else `false`.
     * @example
     *
     * var object = { 'a': [{ 'b': { 'c': 7 } }] };
     * _.unset(object, 'a[0].b.c');
     * // => true
     *
     * console.log(object);
     * // => { 'a': [{ 'b': {} }] };
     *
     * _.unset(object, ['a', '0', 'b', 'c']);
     * // => true
     *
     * console.log(object);
     * // => { 'a': [{ 'b': {} }] };
     */
    function unset(object, path) {
      return object == null ? true : baseUnset(object, path);
    }

    /**
     * This method is like `_.set` except that accepts `updater` to produce the
     * value to set. Use `_.updateWith` to customize `path` creation. The `updater`
     * is invoked with one argument: (value).
     *
     * **Note:** This method mutates `object`.
     *
     * @static
     * @memberOf _
     * @since 4.6.0
     * @category Object
     * @param {Object} object The object to modify.
     * @param {Array|string} path The path of the property to set.
     * @param {Function} updater The function to produce the updated value.
     * @returns {Object} Returns `object`.
     * @example
     *
     * var object = { 'a': [{ 'b': { 'c': 3 } }] };
     *
     * _.update(object, 'a[0].b.c', function(n) { return n * n; });
     * console.log(object.a[0].b.c);
     * // => 9
     *
     * _.update(object, 'x[0].y.z', function(n) { return n ? n + 1 : 0; });
     * console.log(object.x[0].y.z);
     * // => 0
     */
    function update(object, path, updater) {
      return object == null ? object : baseUpdate(object, path, castFunction(updater));
    }

    /**
     * This method is like `_.update` except that it accepts `customizer` which is
     * invoked to produce the objects of `path`.  If `customizer` returns `undefined`
     * path creation is handled by the method instead. The `customizer` is invoked
     * with three arguments: (nsValue, key, nsObject).
     *
     * **Note:** This method mutates `object`.
     *
     * @static
     * @memberOf _
     * @since 4.6.0
     * @category Object
     * @param {Object} object The object to modify.
     * @param {Array|string} path The path of the property to set.
     * @param {Function} updater The function to produce the updated value.
     * @param {Function} [customizer] The function to customize assigned values.
     * @returns {Object} Returns `object`.
     * @example
     *
     * var object = {};
     *
     * _.updateWith(object, '[0][1]', _.constant('a'), Object);
     * // => { '0': { '1': 'a' } }
     */
    function updateWith(object, path, updater, customizer) {
      customizer = typeof customizer == 'function' ? customizer : undefined;
      return object == null ? object : baseUpdate(object, path, castFunction(updater), customizer);
    }

    /**
     * Creates an array of the own enumerable string keyed property values of `object`.
     *
     * **Note:** Non-object values are coerced to objects.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Object
     * @param {Object} object The object to query.
     * @returns {Array} Returns the array of property values.
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     *   this.b = 2;
     * }
     *
     * Foo.prototype.c = 3;
     *
     * _.values(new Foo);
     * // => [1, 2] (iteration order is not guaranteed)
     *
     * _.values('hi');
     * // => ['h', 'i']
     */
    function values(object) {
      return object == null ? [] : baseValues(object, keys(object));
    }

    /**
     * Creates an array of the own and inherited enumerable string keyed property
     * values of `object`.
     *
     * **Note:** Non-object values are coerced to objects.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Object
     * @param {Object} object The object to query.
     * @returns {Array} Returns the array of property values.
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     *   this.b = 2;
     * }
     *
     * Foo.prototype.c = 3;
     *
     * _.valuesIn(new Foo);
     * // => [1, 2, 3] (iteration order is not guaranteed)
     */
    function valuesIn(object) {
      return object == null ? [] : baseValues(object, keysIn(object));
    }

    /*------------------------------------------------------------------------*/

    /**
     * Clamps `number` within the inclusive `lower` and `upper` bounds.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Number
     * @param {number} number The number to clamp.
     * @param {number} [lower] The lower bound.
     * @param {number} upper The upper bound.
     * @returns {number} Returns the clamped number.
     * @example
     *
     * _.clamp(-10, -5, 5);
     * // => -5
     *
     * _.clamp(10, -5, 5);
     * // => 5
     */
    function clamp(number, lower, upper) {
      if (upper === undefined) {
        upper = lower;
        lower = undefined;
      }
      if (upper !== undefined) {
        upper = toNumber(upper);
        upper = upper === upper ? upper : 0;
      }
      if (lower !== undefined) {
        lower = toNumber(lower);
        lower = lower === lower ? lower : 0;
      }
      return baseClamp(toNumber(number), lower, upper);
    }

    /**
     * Checks if `n` is between `start` and up to, but not including, `end`. If
     * `end` is not specified, it's set to `start` with `start` then set to `0`.
     * If `start` is greater than `end` the params are swapped to support
     * negative ranges.
     *
     * @static
     * @memberOf _
     * @since 3.3.0
     * @category Number
     * @param {number} number The number to check.
     * @param {number} [start=0] The start of the range.
     * @param {number} end The end of the range.
     * @returns {boolean} Returns `true` if `number` is in the range, else `false`.
     * @see _.range, _.rangeRight
     * @example
     *
     * _.inRange(3, 2, 4);
     * // => true
     *
     * _.inRange(4, 8);
     * // => true
     *
     * _.inRange(4, 2);
     * // => false
     *
     * _.inRange(2, 2);
     * // => false
     *
     * _.inRange(1.2, 2);
     * // => true
     *
     * _.inRange(5.2, 4);
     * // => false
     *
     * _.inRange(-3, -2, -6);
     * // => true
     */
    function inRange(number, start, end) {
      start = toFinite(start);
      if (end === undefined) {
        end = start;
        start = 0;
      } else {
        end = toFinite(end);
      }
      number = toNumber(number);
      return baseInRange(number, start, end);
    }

    /**
     * Produces a random number between the inclusive `lower` and `upper` bounds.
     * If only one argument is provided a number between `0` and the given number
     * is returned. If `floating` is `true`, or either `lower` or `upper` are
     * floats, a floating-point number is returned instead of an integer.
     *
     * **Note:** JavaScript follows the IEEE-754 standard for resolving
     * floating-point values which can produce unexpected results.
     *
     * @static
     * @memberOf _
     * @since 0.7.0
     * @category Number
     * @param {number} [lower=0] The lower bound.
     * @param {number} [upper=1] The upper bound.
     * @param {boolean} [floating] Specify returning a floating-point number.
     * @returns {number} Returns the random number.
     * @example
     *
     * _.random(0, 5);
     * // => an integer between 0 and 5
     *
     * _.random(5);
     * // => also an integer between 0 and 5
     *
     * _.random(5, true);
     * // => a floating-point number between 0 and 5
     *
     * _.random(1.2, 5.2);
     * // => a floating-point number between 1.2 and 5.2
     */
    function random(lower, upper, floating) {
      if (floating && typeof floating != 'boolean' && isIterateeCall(lower, upper, floating)) {
        upper = floating = undefined;
      }
      if (floating === undefined) {
        if (typeof upper == 'boolean') {
          floating = upper;
          upper = undefined;
        }
        else if (typeof lower == 'boolean') {
          floating = lower;
          lower = undefined;
        }
      }
      if (lower === undefined && upper === undefined) {
        lower = 0;
        upper = 1;
      }
      else {
        lower = toFinite(lower);
        if (upper === undefined) {
          upper = lower;
          lower = 0;
        } else {
          upper = toFinite(upper);
        }
      }
      if (lower > upper) {
        var temp = lower;
        lower = upper;
        upper = temp;
      }
      if (floating || lower % 1 || upper % 1) {
        var rand = nativeRandom();
        return nativeMin(lower + (rand * (upper - lower + freeParseFloat('1e-' + ((rand + '').length - 1)))), upper);
      }
      return baseRandom(lower, upper);
    }

    /*------------------------------------------------------------------------*/

    /**
     * Converts `string` to [camel case](https://en.wikipedia.org/wiki/CamelCase).
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category String
     * @param {string} [string=''] The string to convert.
     * @returns {string} Returns the camel cased string.
     * @example
     *
     * _.camelCase('Foo Bar');
     * // => 'fooBar'
     *
     * _.camelCase('--foo-bar--');
     * // => 'fooBar'
     *
     * _.camelCase('__FOO_BAR__');
     * // => 'fooBar'
     */
    var camelCase = createCompounder(function(result, word, index) {
      word = word.toLowerCase();
      return result + (index ? capitalize(word) : word);
    });

    /**
     * Converts the first character of `string` to upper case and the remaining
     * to lower case.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category String
     * @param {string} [string=''] The string to capitalize.
     * @returns {string} Returns the capitalized string.
     * @example
     *
     * _.capitalize('FRED');
     * // => 'Fred'
     */
    function capitalize(string) {
      return upperFirst(toString(string).toLowerCase());
    }

    /**
     * Deburrs `string` by converting
     * [Latin-1 Supplement](https://en.wikipedia.org/wiki/Latin-1_Supplement_(Unicode_block)#Character_table)
     * and [Latin Extended-A](https://en.wikipedia.org/wiki/Latin_Extended-A)
     * letters to basic Latin letters and removing
     * [combining diacritical marks](https://en.wikipedia.org/wiki/Combining_Diacritical_Marks).
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category String
     * @param {string} [string=''] The string to deburr.
     * @returns {string} Returns the deburred string.
     * @example
     *
     * _.deburr('déjà vu');
     * // => 'deja vu'
     */
    function deburr(string) {
      string = toString(string);
      return string && string.replace(reLatin, deburrLetter).replace(reComboMark, '');
    }

    /**
     * Checks if `string` ends with the given target string.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category String
     * @param {string} [string=''] The string to inspect.
     * @param {string} [target] The string to search for.
     * @param {number} [position=string.length] The position to search up to.
     * @returns {boolean} Returns `true` if `string` ends with `target`,
     *  else `false`.
     * @example
     *
     * _.endsWith('abc', 'c');
     * // => true
     *
     * _.endsWith('abc', 'b');
     * // => false
     *
     * _.endsWith('abc', 'b', 2);
     * // => true
     */
    function endsWith(string, target, position) {
      string = toString(string);
      target = baseToString(target);

      var length = string.length;
      position = position === undefined
        ? length
        : baseClamp(toInteger(position), 0, length);

      var end = position;
      position -= target.length;
      return position >= 0 && string.slice(position, end) == target;
    }

    /**
     * Converts the characters "&", "<", ">", '"', and "'" in `string` to their
     * corresponding HTML entities.
     *
     * **Note:** No other characters are escaped. To escape additional
     * characters use a third-party library like [_he_](https://mths.be/he).
     *
     * Though the ">" character is escaped for symmetry, characters like
     * ">" and "/" don't need escaping in HTML and have no special meaning
     * unless they're part of a tag or unquoted attribute value. See
     * [Mathias Bynens's article](https://mathiasbynens.be/notes/ambiguous-ampersands)
     * (under "semi-related fun fact") for more details.
     *
     * When working with HTML you should always
     * [quote attribute values](http://wonko.com/post/html-escaping) to reduce
     * XSS vectors.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category String
     * @param {string} [string=''] The string to escape.
     * @returns {string} Returns the escaped string.
     * @example
     *
     * _.escape('fred, barney, & pebbles');
     * // => 'fred, barney, &amp; pebbles'
     */
    function escape(string) {
      string = toString(string);
      return (string && reHasUnescapedHtml.test(string))
        ? string.replace(reUnescapedHtml, escapeHtmlChar)
        : string;
    }

    /**
     * Escapes the `RegExp` special characters "^", "$", "\", ".", "*", "+",
     * "?", "(", ")", "[", "]", "{", "}", and "|" in `string`.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category String
     * @param {string} [string=''] The string to escape.
     * @returns {string} Returns the escaped string.
     * @example
     *
     * _.escapeRegExp('[lodash](https://lodash.com/)');
     * // => '\[lodash\]\(https://lodash\.com/\)'
     */
    function escapeRegExp(string) {
      string = toString(string);
      return (string && reHasRegExpChar.test(string))
        ? string.replace(reRegExpChar, '\\$&')
        : string;
    }

    /**
     * Converts `string` to
     * [kebab case](https://en.wikipedia.org/wiki/Letter_case#Special_case_styles).
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category String
     * @param {string} [string=''] The string to convert.
     * @returns {string} Returns the kebab cased string.
     * @example
     *
     * _.kebabCase('Foo Bar');
     * // => 'foo-bar'
     *
     * _.kebabCase('fooBar');
     * // => 'foo-bar'
     *
     * _.kebabCase('__FOO_BAR__');
     * // => 'foo-bar'
     */
    var kebabCase = createCompounder(function(result, word, index) {
      return result + (index ? '-' : '') + word.toLowerCase();
    });

    /**
     * Converts `string`, as space separated words, to lower case.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category String
     * @param {string} [string=''] The string to convert.
     * @returns {string} Returns the lower cased string.
     * @example
     *
     * _.lowerCase('--Foo-Bar--');
     * // => 'foo bar'
     *
     * _.lowerCase('fooBar');
     * // => 'foo bar'
     *
     * _.lowerCase('__FOO_BAR__');
     * // => 'foo bar'
     */
    var lowerCase = createCompounder(function(result, word, index) {
      return result + (index ? ' ' : '') + word.toLowerCase();
    });

    /**
     * Converts the first character of `string` to lower case.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category String
     * @param {string} [string=''] The string to convert.
     * @returns {string} Returns the converted string.
     * @example
     *
     * _.lowerFirst('Fred');
     * // => 'fred'
     *
     * _.lowerFirst('FRED');
     * // => 'fRED'
     */
    var lowerFirst = createCaseFirst('toLowerCase');

    /**
     * Pads `string` on the left and right sides if it's shorter than `length`.
     * Padding characters are truncated if they can't be evenly divided by `length`.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category String
     * @param {string} [string=''] The string to pad.
     * @param {number} [length=0] The padding length.
     * @param {string} [chars=' '] The string used as padding.
     * @returns {string} Returns the padded string.
     * @example
     *
     * _.pad('abc', 8);
     * // => '  abc   '
     *
     * _.pad('abc', 8, '_-');
     * // => '_-abc_-_'
     *
     * _.pad('abc', 3);
     * // => 'abc'
     */
    function pad(string, length, chars) {
      string = toString(string);
      length = toInteger(length);

      var strLength = length ? stringSize(string) : 0;
      if (!length || strLength >= length) {
        return string;
      }
      var mid = (length - strLength) / 2;
      return (
        createPadding(nativeFloor(mid), chars) +
        string +
        createPadding(nativeCeil(mid), chars)
      );
    }

    /**
     * Pads `string` on the right side if it's shorter than `length`. Padding
     * characters are truncated if they exceed `length`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category String
     * @param {string} [string=''] The string to pad.
     * @param {number} [length=0] The padding length.
     * @param {string} [chars=' '] The string used as padding.
     * @returns {string} Returns the padded string.
     * @example
     *
     * _.padEnd('abc', 6);
     * // => 'abc   '
     *
     * _.padEnd('abc', 6, '_-');
     * // => 'abc_-_'
     *
     * _.padEnd('abc', 3);
     * // => 'abc'
     */
    function padEnd(string, length, chars) {
      string = toString(string);
      length = toInteger(length);

      var strLength = length ? stringSize(string) : 0;
      return (length && strLength < length)
        ? (string + createPadding(length - strLength, chars))
        : string;
    }

    /**
     * Pads `string` on the left side if it's shorter than `length`. Padding
     * characters are truncated if they exceed `length`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category String
     * @param {string} [string=''] The string to pad.
     * @param {number} [length=0] The padding length.
     * @param {string} [chars=' '] The string used as padding.
     * @returns {string} Returns the padded string.
     * @example
     *
     * _.padStart('abc', 6);
     * // => '   abc'
     *
     * _.padStart('abc', 6, '_-');
     * // => '_-_abc'
     *
     * _.padStart('abc', 3);
     * // => 'abc'
     */
    function padStart(string, length, chars) {
      string = toString(string);
      length = toInteger(length);

      var strLength = length ? stringSize(string) : 0;
      return (length && strLength < length)
        ? (createPadding(length - strLength, chars) + string)
        : string;
    }

    /**
     * Converts `string` to an integer of the specified radix. If `radix` is
     * `undefined` or `0`, a `radix` of `10` is used unless `value` is a
     * hexadecimal, in which case a `radix` of `16` is used.
     *
     * **Note:** This method aligns with the
     * [ES5 implementation](https://es5.github.io/#x15.1.2.2) of `parseInt`.
     *
     * @static
     * @memberOf _
     * @since 1.1.0
     * @category String
     * @param {string} string The string to convert.
     * @param {number} [radix=10] The radix to interpret `value` by.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {number} Returns the converted integer.
     * @example
     *
     * _.parseInt('08');
     * // => 8
     *
     * _.map(['6', '08', '10'], _.parseInt);
     * // => [6, 8, 10]
     */
    function parseInt(string, radix, guard) {
      if (guard || radix == null) {
        radix = 0;
      } else if (radix) {
        radix = +radix;
      }
      return nativeParseInt(toString(string).replace(reTrimStart, ''), radix || 0);
    }

    /**
     * Repeats the given string `n` times.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category String
     * @param {string} [string=''] The string to repeat.
     * @param {number} [n=1] The number of times to repeat the string.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {string} Returns the repeated string.
     * @example
     *
     * _.repeat('*', 3);
     * // => '***'
     *
     * _.repeat('abc', 2);
     * // => 'abcabc'
     *
     * _.repeat('abc', 0);
     * // => ''
     */
    function repeat(string, n, guard) {
      if ((guard ? isIterateeCall(string, n, guard) : n === undefined)) {
        n = 1;
      } else {
        n = toInteger(n);
      }
      return baseRepeat(toString(string), n);
    }

    /**
     * Replaces matches for `pattern` in `string` with `replacement`.
     *
     * **Note:** This method is based on
     * [`String#replace`](https://mdn.io/String/replace).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category String
     * @param {string} [string=''] The string to modify.
     * @param {RegExp|string} pattern The pattern to replace.
     * @param {Function|string} replacement The match replacement.
     * @returns {string} Returns the modified string.
     * @example
     *
     * _.replace('Hi Fred', 'Fred', 'Barney');
     * // => 'Hi Barney'
     */
    function replace() {
      var args = arguments,
          string = toString(args[0]);

      return args.length < 3 ? string : string.replace(args[1], args[2]);
    }

    /**
     * Converts `string` to
     * [snake case](https://en.wikipedia.org/wiki/Snake_case).
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category String
     * @param {string} [string=''] The string to convert.
     * @returns {string} Returns the snake cased string.
     * @example
     *
     * _.snakeCase('Foo Bar');
     * // => 'foo_bar'
     *
     * _.snakeCase('fooBar');
     * // => 'foo_bar'
     *
     * _.snakeCase('--FOO-BAR--');
     * // => 'foo_bar'
     */
    var snakeCase = createCompounder(function(result, word, index) {
      return result + (index ? '_' : '') + word.toLowerCase();
    });

    /**
     * Splits `string` by `separator`.
     *
     * **Note:** This method is based on
     * [`String#split`](https://mdn.io/String/split).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category String
     * @param {string} [string=''] The string to split.
     * @param {RegExp|string} separator The separator pattern to split by.
     * @param {number} [limit] The length to truncate results to.
     * @returns {Array} Returns the string segments.
     * @example
     *
     * _.split('a-b-c', '-', 2);
     * // => ['a', 'b']
     */
    function split(string, separator, limit) {
      if (limit && typeof limit != 'number' && isIterateeCall(string, separator, limit)) {
        separator = limit = undefined;
      }
      limit = limit === undefined ? MAX_ARRAY_LENGTH : limit >>> 0;
      if (!limit) {
        return [];
      }
      string = toString(string);
      if (string && (
            typeof separator == 'string' ||
            (separator != null && !isRegExp(separator))
          )) {
        separator = baseToString(separator);
        if (!separator && hasUnicode(string)) {
          return castSlice(stringToArray(string), 0, limit);
        }
      }
      return string.split(separator, limit);
    }

    /**
     * Converts `string` to
     * [start case](https://en.wikipedia.org/wiki/Letter_case#Stylistic_or_specialised_usage).
     *
     * @static
     * @memberOf _
     * @since 3.1.0
     * @category String
     * @param {string} [string=''] The string to convert.
     * @returns {string} Returns the start cased string.
     * @example
     *
     * _.startCase('--foo-bar--');
     * // => 'Foo Bar'
     *
     * _.startCase('fooBar');
     * // => 'Foo Bar'
     *
     * _.startCase('__FOO_BAR__');
     * // => 'FOO BAR'
     */
    var startCase = createCompounder(function(result, word, index) {
      return result + (index ? ' ' : '') + upperFirst(word);
    });

    /**
     * Checks if `string` starts with the given target string.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category String
     * @param {string} [string=''] The string to inspect.
     * @param {string} [target] The string to search for.
     * @param {number} [position=0] The position to search from.
     * @returns {boolean} Returns `true` if `string` starts with `target`,
     *  else `false`.
     * @example
     *
     * _.startsWith('abc', 'a');
     * // => true
     *
     * _.startsWith('abc', 'b');
     * // => false
     *
     * _.startsWith('abc', 'b', 1);
     * // => true
     */
    function startsWith(string, target, position) {
      string = toString(string);
      position = position == null
        ? 0
        : baseClamp(toInteger(position), 0, string.length);

      target = baseToString(target);
      return string.slice(position, position + target.length) == target;
    }

    /**
     * Creates a compiled template function that can interpolate data properties
     * in "interpolate" delimiters, HTML-escape interpolated data properties in
     * "escape" delimiters, and execute JavaScript in "evaluate" delimiters. Data
     * properties may be accessed as free variables in the template. If a setting
     * object is given, it takes precedence over `_.templateSettings` values.
     *
     * **Note:** In the development build `_.template` utilizes
     * [sourceURLs](http://www.html5rocks.com/en/tutorials/developertools/sourcemaps/#toc-sourceurl)
     * for easier debugging.
     *
     * For more information on precompiling templates see
     * [lodash's custom builds documentation](https://lodash.com/custom-builds).
     *
     * For more information on Chrome extension sandboxes see
     * [Chrome's extensions documentation](https://developer.chrome.com/extensions/sandboxingEval).
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category String
     * @param {string} [string=''] The template string.
     * @param {Object} [options={}] The options object.
     * @param {RegExp} [options.escape=_.templateSettings.escape]
     *  The HTML "escape" delimiter.
     * @param {RegExp} [options.evaluate=_.templateSettings.evaluate]
     *  The "evaluate" delimiter.
     * @param {Object} [options.imports=_.templateSettings.imports]
     *  An object to import into the template as free variables.
     * @param {RegExp} [options.interpolate=_.templateSettings.interpolate]
     *  The "interpolate" delimiter.
     * @param {string} [options.sourceURL='lodash.templateSources[n]']
     *  The sourceURL of the compiled template.
     * @param {string} [options.variable='obj']
     *  The data object variable name.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {Function} Returns the compiled template function.
     * @example
     *
     * // Use the "interpolate" delimiter to create a compiled template.
     * var compiled = _.template('hello <%= user %>!');
     * compiled({ 'user': 'fred' });
     * // => 'hello fred!'
     *
     * // Use the HTML "escape" delimiter to escape data property values.
     * var compiled = _.template('<b><%- value %></b>');
     * compiled({ 'value': '<script>' });
     * // => '<b>&lt;script&gt;</b>'
     *
     * // Use the "evaluate" delimiter to execute JavaScript and generate HTML.
     * var compiled = _.template('<% _.forEach(users, function(user) { %><li><%- user %></li><% }); %>');
     * compiled({ 'users': ['fred', 'barney'] });
     * // => '<li>fred</li><li>barney</li>'
     *
     * // Use the internal `print` function in "evaluate" delimiters.
     * var compiled = _.template('<% print("hello " + user); %>!');
     * compiled({ 'user': 'barney' });
     * // => 'hello barney!'
     *
     * // Use the ES template literal delimiter as an "interpolate" delimiter.
     * // Disable support by replacing the "interpolate" delimiter.
     * var compiled = _.template('hello ${ user }!');
     * compiled({ 'user': 'pebbles' });
     * // => 'hello pebbles!'
     *
     * // Use backslashes to treat delimiters as plain text.
     * var compiled = _.template('<%= "\\<%- value %\\>" %>');
     * compiled({ 'value': 'ignored' });
     * // => '<%- value %>'
     *
     * // Use the `imports` option to import `jQuery` as `jq`.
     * var text = '<% jq.each(users, function(user) { %><li><%- user %></li><% }); %>';
     * var compiled = _.template(text, { 'imports': { 'jq': jQuery } });
     * compiled({ 'users': ['fred', 'barney'] });
     * // => '<li>fred</li><li>barney</li>'
     *
     * // Use the `sourceURL` option to specify a custom sourceURL for the template.
     * var compiled = _.template('hello <%= user %>!', { 'sourceURL': '/basic/greeting.jst' });
     * compiled(data);
     * // => Find the source of "greeting.jst" under the Sources tab or Resources panel of the web inspector.
     *
     * // Use the `variable` option to ensure a with-statement isn't used in the compiled template.
     * var compiled = _.template('hi <%= data.user %>!', { 'variable': 'data' });
     * compiled.source;
     * // => function(data) {
     * //   var __t, __p = '';
     * //   __p += 'hi ' + ((__t = ( data.user )) == null ? '' : __t) + '!';
     * //   return __p;
     * // }
     *
     * // Use custom template delimiters.
     * _.templateSettings.interpolate = /{{([\s\S]+?)}}/g;
     * var compiled = _.template('hello {{ user }}!');
     * compiled({ 'user': 'mustache' });
     * // => 'hello mustache!'
     *
     * // Use the `source` property to inline compiled templates for meaningful
     * // line numbers in error messages and stack traces.
     * fs.writeFileSync(path.join(process.cwd(), 'jst.js'), '\
     *   var JST = {\
     *     "main": ' + _.template(mainText).source + '\
     *   };\
     * ');
     */
    function template(string, options, guard) {
      // Based on John Resig's `tmpl` implementation
      // (http://ejohn.org/blog/javascript-micro-templating/)
      // and Laura Doktorova's doT.js (https://github.com/olado/doT).
      var settings = lodash.templateSettings;

      if (guard && isIterateeCall(string, options, guard)) {
        options = undefined;
      }
      string = toString(string);
      options = assignInWith({}, options, settings, customDefaultsAssignIn);

      var imports = assignInWith({}, options.imports, settings.imports, customDefaultsAssignIn),
          importsKeys = keys(imports),
          importsValues = baseValues(imports, importsKeys);

      var isEscaping,
          isEvaluating,
          index = 0,
          interpolate = options.interpolate || reNoMatch,
          source = "__p += '";

      // Compile the regexp to match each delimiter.
      var reDelimiters = RegExp(
        (options.escape || reNoMatch).source + '|' +
        interpolate.source + '|' +
        (interpolate === reInterpolate ? reEsTemplate : reNoMatch).source + '|' +
        (options.evaluate || reNoMatch).source + '|$'
      , 'g');

      // Use a sourceURL for easier debugging.
      // The sourceURL gets injected into the source that's eval-ed, so be careful
      // to normalize all kinds of whitespace, so e.g. newlines (and unicode versions of it) can't sneak in
      // and escape the comment, thus injecting code that gets evaled.
      var sourceURL = '//# sourceURL=' +
        (hasOwnProperty.call(options, 'sourceURL')
          ? (options.sourceURL + '').replace(/\s/g, ' ')
          : ('lodash.templateSources[' + (++templateCounter) + ']')
        ) + '\n';

      string.replace(reDelimiters, function(match, escapeValue, interpolateValue, esTemplateValue, evaluateValue, offset) {
        interpolateValue || (interpolateValue = esTemplateValue);

        // Escape characters that can't be included in string literals.
        source += string.slice(index, offset).replace(reUnescapedString, escapeStringChar);

        // Replace delimiters with snippets.
        if (escapeValue) {
          isEscaping = true;
          source += "' +\n__e(" + escapeValue + ") +\n'";
        }
        if (evaluateValue) {
          isEvaluating = true;
          source += "';\n" + evaluateValue + ";\n__p += '";
        }
        if (interpolateValue) {
          source += "' +\n((__t = (" + interpolateValue + ")) == null ? '' : __t) +\n'";
        }
        index = offset + match.length;

        // The JS engine embedded in Adobe products needs `match` returned in
        // order to produce the correct `offset` value.
        return match;
      });

      source += "';\n";

      // If `variable` is not specified wrap a with-statement around the generated
      // code to add the data object to the top of the scope chain.
      var variable = hasOwnProperty.call(options, 'variable') && options.variable;
      if (!variable) {
        source = 'with (obj) {\n' + source + '\n}\n';
      }
      // Throw an error if a forbidden character was found in `variable`, to prevent
      // potential command injection attacks.
      else if (reForbiddenIdentifierChars.test(variable)) {
        throw new Error(INVALID_TEMPL_VAR_ERROR_TEXT);
      }

      // Cleanup code by stripping empty strings.
      source = (isEvaluating ? source.replace(reEmptyStringLeading, '') : source)
        .replace(reEmptyStringMiddle, '$1')
        .replace(reEmptyStringTrailing, '$1;');

      // Frame code as the function body.
      source = 'function(' + (variable || 'obj') + ') {\n' +
        (variable
          ? ''
          : 'obj || (obj = {});\n'
        ) +
        "var __t, __p = ''" +
        (isEscaping
           ? ', __e = _.escape'
           : ''
        ) +
        (isEvaluating
          ? ', __j = Array.prototype.join;\n' +
            "function print() { __p += __j.call(arguments, '') }\n"
          : ';\n'
        ) +
        source +
        'return __p\n}';

      var result = attempt(function() {
        return Function(importsKeys, sourceURL + 'return ' + source)
          .apply(undefined, importsValues);
      });

      // Provide the compiled function's source by its `toString` method or
      // the `source` property as a convenience for inlining compiled templates.
      result.source = source;
      if (isError(result)) {
        throw result;
      }
      return result;
    }

    /**
     * Converts `string`, as a whole, to lower case just like
     * [String#toLowerCase](https://mdn.io/toLowerCase).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category String
     * @param {string} [string=''] The string to convert.
     * @returns {string} Returns the lower cased string.
     * @example
     *
     * _.toLower('--Foo-Bar--');
     * // => '--foo-bar--'
     *
     * _.toLower('fooBar');
     * // => 'foobar'
     *
     * _.toLower('__FOO_BAR__');
     * // => '__foo_bar__'
     */
    function toLower(value) {
      return toString(value).toLowerCase();
    }

    /**
     * Converts `string`, as a whole, to upper case just like
     * [String#toUpperCase](https://mdn.io/toUpperCase).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category String
     * @param {string} [string=''] The string to convert.
     * @returns {string} Returns the upper cased string.
     * @example
     *
     * _.toUpper('--foo-bar--');
     * // => '--FOO-BAR--'
     *
     * _.toUpper('fooBar');
     * // => 'FOOBAR'
     *
     * _.toUpper('__foo_bar__');
     * // => '__FOO_BAR__'
     */
    function toUpper(value) {
      return toString(value).toUpperCase();
    }

    /**
     * Removes leading and trailing whitespace or specified characters from `string`.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category String
     * @param {string} [string=''] The string to trim.
     * @param {string} [chars=whitespace] The characters to trim.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {string} Returns the trimmed string.
     * @example
     *
     * _.trim('  abc  ');
     * // => 'abc'
     *
     * _.trim('-_-abc-_-', '_-');
     * // => 'abc'
     *
     * _.map(['  foo  ', '  bar  '], _.trim);
     * // => ['foo', 'bar']
     */
    function trim(string, chars, guard) {
      string = toString(string);
      if (string && (guard || chars === undefined)) {
        return baseTrim(string);
      }
      if (!string || !(chars = baseToString(chars))) {
        return string;
      }
      var strSymbols = stringToArray(string),
          chrSymbols = stringToArray(chars),
          start = charsStartIndex(strSymbols, chrSymbols),
          end = charsEndIndex(strSymbols, chrSymbols) + 1;

      return castSlice(strSymbols, start, end).join('');
    }

    /**
     * Removes trailing whitespace or specified characters from `string`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category String
     * @param {string} [string=''] The string to trim.
     * @param {string} [chars=whitespace] The characters to trim.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {string} Returns the trimmed string.
     * @example
     *
     * _.trimEnd('  abc  ');
     * // => '  abc'
     *
     * _.trimEnd('-_-abc-_-', '_-');
     * // => '-_-abc'
     */
    function trimEnd(string, chars, guard) {
      string = toString(string);
      if (string && (guard || chars === undefined)) {
        return string.slice(0, trimmedEndIndex(string) + 1);
      }
      if (!string || !(chars = baseToString(chars))) {
        return string;
      }
      var strSymbols = stringToArray(string),
          end = charsEndIndex(strSymbols, stringToArray(chars)) + 1;

      return castSlice(strSymbols, 0, end).join('');
    }

    /**
     * Removes leading whitespace or specified characters from `string`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category String
     * @param {string} [string=''] The string to trim.
     * @param {string} [chars=whitespace] The characters to trim.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {string} Returns the trimmed string.
     * @example
     *
     * _.trimStart('  abc  ');
     * // => 'abc  '
     *
     * _.trimStart('-_-abc-_-', '_-');
     * // => 'abc-_-'
     */
    function trimStart(string, chars, guard) {
      string = toString(string);
      if (string && (guard || chars === undefined)) {
        return string.replace(reTrimStart, '');
      }
      if (!string || !(chars = baseToString(chars))) {
        return string;
      }
      var strSymbols = stringToArray(string),
          start = charsStartIndex(strSymbols, stringToArray(chars));

      return castSlice(strSymbols, start).join('');
    }

    /**
     * Truncates `string` if it's longer than the given maximum string length.
     * The last characters of the truncated string are replaced with the omission
     * string which defaults to "...".
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category String
     * @param {string} [string=''] The string to truncate.
     * @param {Object} [options={}] The options object.
     * @param {number} [options.length=30] The maximum string length.
     * @param {string} [options.omission='...'] The string to indicate text is omitted.
     * @param {RegExp|string} [options.separator] The separator pattern to truncate to.
     * @returns {string} Returns the truncated string.
     * @example
     *
     * _.truncate('hi-diddly-ho there, neighborino');
     * // => 'hi-diddly-ho there, neighbo...'
     *
     * _.truncate('hi-diddly-ho there, neighborino', {
     *   'length': 24,
     *   'separator': ' '
     * });
     * // => 'hi-diddly-ho there,...'
     *
     * _.truncate('hi-diddly-ho there, neighborino', {
     *   'length': 24,
     *   'separator': /,? +/
     * });
     * // => 'hi-diddly-ho there...'
     *
     * _.truncate('hi-diddly-ho there, neighborino', {
     *   'omission': ' [...]'
     * });
     * // => 'hi-diddly-ho there, neig [...]'
     */
    function truncate(string, options) {
      var length = DEFAULT_TRUNC_LENGTH,
          omission = DEFAULT_TRUNC_OMISSION;

      if (isObject(options)) {
        var separator = 'separator' in options ? options.separator : separator;
        length = 'length' in options ? toInteger(options.length) : length;
        omission = 'omission' in options ? baseToString(options.omission) : omission;
      }
      string = toString(string);

      var strLength = string.length;
      if (hasUnicode(string)) {
        var strSymbols = stringToArray(string);
        strLength = strSymbols.length;
      }
      if (length >= strLength) {
        return string;
      }
      var end = length - stringSize(omission);
      if (end < 1) {
        return omission;
      }
      var result = strSymbols
        ? castSlice(strSymbols, 0, end).join('')
        : string.slice(0, end);

      if (separator === undefined) {
        return result + omission;
      }
      if (strSymbols) {
        end += (result.length - end);
      }
      if (isRegExp(separator)) {
        if (string.slice(end).search(separator)) {
          var match,
              substring = result;

          if (!separator.global) {
            separator = RegExp(separator.source, toString(reFlags.exec(separator)) + 'g');
          }
          separator.lastIndex = 0;
          while ((match = separator.exec(substring))) {
            var newEnd = match.index;
          }
          result = result.slice(0, newEnd === undefined ? end : newEnd);
        }
      } else if (string.indexOf(baseToString(separator), end) != end) {
        var index = result.lastIndexOf(separator);
        if (index > -1) {
          result = result.slice(0, index);
        }
      }
      return result + omission;
    }

    /**
     * The inverse of `_.escape`; this method converts the HTML entities
     * `&amp;`, `&lt;`, `&gt;`, `&quot;`, and `&#39;` in `string` to
     * their corresponding characters.
     *
     * **Note:** No other HTML entities are unescaped. To unescape additional
     * HTML entities use a third-party library like [_he_](https://mths.be/he).
     *
     * @static
     * @memberOf _
     * @since 0.6.0
     * @category String
     * @param {string} [string=''] The string to unescape.
     * @returns {string} Returns the unescaped string.
     * @example
     *
     * _.unescape('fred, barney, &amp; pebbles');
     * // => 'fred, barney, & pebbles'
     */
    function unescape(string) {
      string = toString(string);
      return (string && reHasEscapedHtml.test(string))
        ? string.replace(reEscapedHtml, unescapeHtmlChar)
        : string;
    }

    /**
     * Converts `string`, as space separated words, to upper case.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category String
     * @param {string} [string=''] The string to convert.
     * @returns {string} Returns the upper cased string.
     * @example
     *
     * _.upperCase('--foo-bar');
     * // => 'FOO BAR'
     *
     * _.upperCase('fooBar');
     * // => 'FOO BAR'
     *
     * _.upperCase('__foo_bar__');
     * // => 'FOO BAR'
     */
    var upperCase = createCompounder(function(result, word, index) {
      return result + (index ? ' ' : '') + word.toUpperCase();
    });

    /**
     * Converts the first character of `string` to upper case.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category String
     * @param {string} [string=''] The string to convert.
     * @returns {string} Returns the converted string.
     * @example
     *
     * _.upperFirst('fred');
     * // => 'Fred'
     *
     * _.upperFirst('FRED');
     * // => 'FRED'
     */
    var upperFirst = createCaseFirst('toUpperCase');

    /**
     * Splits `string` into an array of its words.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category String
     * @param {string} [string=''] The string to inspect.
     * @param {RegExp|string} [pattern] The pattern to match words.
     * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
     * @returns {Array} Returns the words of `string`.
     * @example
     *
     * _.words('fred, barney, & pebbles');
     * // => ['fred', 'barney', 'pebbles']
     *
     * _.words('fred, barney, & pebbles', /[^, ]+/g);
     * // => ['fred', 'barney', '&', 'pebbles']
     */
    function words(string, pattern, guard) {
      string = toString(string);
      pattern = guard ? undefined : pattern;

      if (pattern === undefined) {
        return hasUnicodeWord(string) ? unicodeWords(string) : asciiWords(string);
      }
      return string.match(pattern) || [];
    }

    /*------------------------------------------------------------------------*/

    /**
     * Attempts to invoke `func`, returning either the result or the caught error
     * object. Any additional arguments are provided to `func` when it's invoked.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Util
     * @param {Function} func The function to attempt.
     * @param {...*} [args] The arguments to invoke `func` with.
     * @returns {*} Returns the `func` result or error object.
     * @example
     *
     * // Avoid throwing errors for invalid selectors.
     * var elements = _.attempt(function(selector) {
     *   return document.querySelectorAll(selector);
     * }, '>_>');
     *
     * if (_.isError(elements)) {
     *   elements = [];
     * }
     */
    var attempt = baseRest(function(func, args) {
      try {
        return apply(func, undefined, args);
      } catch (e) {
        return isError(e) ? e : new Error(e);
      }
    });

    /**
     * Binds methods of an object to the object itself, overwriting the existing
     * method.
     *
     * **Note:** This method doesn't set the "length" property of bound functions.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Util
     * @param {Object} object The object to bind and assign the bound methods to.
     * @param {...(string|string[])} methodNames The object method names to bind.
     * @returns {Object} Returns `object`.
     * @example
     *
     * var view = {
     *   'label': 'docs',
     *   'click': function() {
     *     console.log('clicked ' + this.label);
     *   }
     * };
     *
     * _.bindAll(view, ['click']);
     * jQuery(element).on('click', view.click);
     * // => Logs 'clicked docs' when clicked.
     */
    var bindAll = flatRest(function(object, methodNames) {
      arrayEach(methodNames, function(key) {
        key = toKey(key);
        baseAssignValue(object, key, bind(object[key], object));
      });
      return object;
    });

    /**
     * Creates a function that iterates over `pairs` and invokes the corresponding
     * function of the first predicate to return truthy. The predicate-function
     * pairs are invoked with the `this` binding and arguments of the created
     * function.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Util
     * @param {Array} pairs The predicate-function pairs.
     * @returns {Function} Returns the new composite function.
     * @example
     *
     * var func = _.cond([
     *   [_.matches({ 'a': 1 }),           _.constant('matches A')],
     *   [_.conforms({ 'b': _.isNumber }), _.constant('matches B')],
     *   [_.stubTrue,                      _.constant('no match')]
     * ]);
     *
     * func({ 'a': 1, 'b': 2 });
     * // => 'matches A'
     *
     * func({ 'a': 0, 'b': 1 });
     * // => 'matches B'
     *
     * func({ 'a': '1', 'b': '2' });
     * // => 'no match'
     */
    function cond(pairs) {
      var length = pairs == null ? 0 : pairs.length,
          toIteratee = getIteratee();

      pairs = !length ? [] : arrayMap(pairs, function(pair) {
        if (typeof pair[1] != 'function') {
          throw new TypeError(FUNC_ERROR_TEXT);
        }
        return [toIteratee(pair[0]), pair[1]];
      });

      return baseRest(function(args) {
        var index = -1;
        while (++index < length) {
          var pair = pairs[index];
          if (apply(pair[0], this, args)) {
            return apply(pair[1], this, args);
          }
        }
      });
    }

    /**
     * Creates a function that invokes the predicate properties of `source` with
     * the corresponding property values of a given object, returning `true` if
     * all predicates return truthy, else `false`.
     *
     * **Note:** The created function is equivalent to `_.conformsTo` with
     * `source` partially applied.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Util
     * @param {Object} source The object of property predicates to conform to.
     * @returns {Function} Returns the new spec function.
     * @example
     *
     * var objects = [
     *   { 'a': 2, 'b': 1 },
     *   { 'a': 1, 'b': 2 }
     * ];
     *
     * _.filter(objects, _.conforms({ 'b': function(n) { return n > 1; } }));
     * // => [{ 'a': 1, 'b': 2 }]
     */
    function conforms(source) {
      return baseConforms(baseClone(source, CLONE_DEEP_FLAG));
    }

    /**
     * Creates a function that returns `value`.
     *
     * @static
     * @memberOf _
     * @since 2.4.0
     * @category Util
     * @param {*} value The value to return from the new function.
     * @returns {Function} Returns the new constant function.
     * @example
     *
     * var objects = _.times(2, _.constant({ 'a': 1 }));
     *
     * console.log(objects);
     * // => [{ 'a': 1 }, { 'a': 1 }]
     *
     * console.log(objects[0] === objects[1]);
     * // => true
     */
    function constant(value) {
      return function() {
        return value;
      };
    }

    /**
     * Checks `value` to determine whether a default value should be returned in
     * its place. The `defaultValue` is returned if `value` is `NaN`, `null`,
     * or `undefined`.
     *
     * @static
     * @memberOf _
     * @since 4.14.0
     * @category Util
     * @param {*} value The value to check.
     * @param {*} defaultValue The default value.
     * @returns {*} Returns the resolved value.
     * @example
     *
     * _.defaultTo(1, 10);
     * // => 1
     *
     * _.defaultTo(undefined, 10);
     * // => 10
     */
    function defaultTo(value, defaultValue) {
      return (value == null || value !== value) ? defaultValue : value;
    }

    /**
     * Creates a function that returns the result of invoking the given functions
     * with the `this` binding of the created function, where each successive
     * invocation is supplied the return value of the previous.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Util
     * @param {...(Function|Function[])} [funcs] The functions to invoke.
     * @returns {Function} Returns the new composite function.
     * @see _.flowRight
     * @example
     *
     * function square(n) {
     *   return n * n;
     * }
     *
     * var addSquare = _.flow([_.add, square]);
     * addSquare(1, 2);
     * // => 9
     */
    var flow = createFlow();

    /**
     * This method is like `_.flow` except that it creates a function that
     * invokes the given functions from right to left.
     *
     * @static
     * @since 3.0.0
     * @memberOf _
     * @category Util
     * @param {...(Function|Function[])} [funcs] The functions to invoke.
     * @returns {Function} Returns the new composite function.
     * @see _.flow
     * @example
     *
     * function square(n) {
     *   return n * n;
     * }
     *
     * var addSquare = _.flowRight([square, _.add]);
     * addSquare(1, 2);
     * // => 9
     */
    var flowRight = createFlow(true);

    /**
     * This method returns the first argument it receives.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Util
     * @param {*} value Any value.
     * @returns {*} Returns `value`.
     * @example
     *
     * var object = { 'a': 1 };
     *
     * console.log(_.identity(object) === object);
     * // => true
     */
    function identity(value) {
      return value;
    }

    /**
     * Creates a function that invokes `func` with the arguments of the created
     * function. If `func` is a property name, the created function returns the
     * property value for a given element. If `func` is an array or object, the
     * created function returns `true` for elements that contain the equivalent
     * source properties, otherwise it returns `false`.
     *
     * @static
     * @since 4.0.0
     * @memberOf _
     * @category Util
     * @param {*} [func=_.identity] The value to convert to a callback.
     * @returns {Function} Returns the callback.
     * @example
     *
     * var users = [
     *   { 'user': 'barney', 'age': 36, 'active': true },
     *   { 'user': 'fred',   'age': 40, 'active': false }
     * ];
     *
     * // The `_.matches` iteratee shorthand.
     * _.filter(users, _.iteratee({ 'user': 'barney', 'active': true }));
     * // => [{ 'user': 'barney', 'age': 36, 'active': true }]
     *
     * // The `_.matchesProperty` iteratee shorthand.
     * _.filter(users, _.iteratee(['user', 'fred']));
     * // => [{ 'user': 'fred', 'age': 40 }]
     *
     * // The `_.property` iteratee shorthand.
     * _.map(users, _.iteratee('user'));
     * // => ['barney', 'fred']
     *
     * // Create custom iteratee shorthands.
     * _.iteratee = _.wrap(_.iteratee, function(iteratee, func) {
     *   return !_.isRegExp(func) ? iteratee(func) : function(string) {
     *     return func.test(string);
     *   };
     * });
     *
     * _.filter(['abc', 'def'], /ef/);
     * // => ['def']
     */
    function iteratee(func) {
      return baseIteratee(typeof func == 'function' ? func : baseClone(func, CLONE_DEEP_FLAG));
    }

    /**
     * Creates a function that performs a partial deep comparison between a given
     * object and `source`, returning `true` if the given object has equivalent
     * property values, else `false`.
     *
     * **Note:** The created function is equivalent to `_.isMatch` with `source`
     * partially applied.
     *
     * Partial comparisons will match empty array and empty object `source`
     * values against any array or object value, respectively. See `_.isEqual`
     * for a list of supported value comparisons.
     *
     * **Note:** Multiple values can be checked by combining several matchers
     * using `_.overSome`
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Util
     * @param {Object} source The object of property values to match.
     * @returns {Function} Returns the new spec function.
     * @example
     *
     * var objects = [
     *   { 'a': 1, 'b': 2, 'c': 3 },
     *   { 'a': 4, 'b': 5, 'c': 6 }
     * ];
     *
     * _.filter(objects, _.matches({ 'a': 4, 'c': 6 }));
     * // => [{ 'a': 4, 'b': 5, 'c': 6 }]
     *
     * // Checking for several possible values
     * _.filter(objects, _.overSome([_.matches({ 'a': 1 }), _.matches({ 'a': 4 })]));
     * // => [{ 'a': 1, 'b': 2, 'c': 3 }, { 'a': 4, 'b': 5, 'c': 6 }]
     */
    function matches(source) {
      return baseMatches(baseClone(source, CLONE_DEEP_FLAG));
    }

    /**
     * Creates a function that performs a partial deep comparison between the
     * value at `path` of a given object to `srcValue`, returning `true` if the
     * object value is equivalent, else `false`.
     *
     * **Note:** Partial comparisons will match empty array and empty object
     * `srcValue` values against any array or object value, respectively. See
     * `_.isEqual` for a list of supported value comparisons.
     *
     * **Note:** Multiple values can be checked by combining several matchers
     * using `_.overSome`
     *
     * @static
     * @memberOf _
     * @since 3.2.0
     * @category Util
     * @param {Array|string} path The path of the property to get.
     * @param {*} srcValue The value to match.
     * @returns {Function} Returns the new spec function.
     * @example
     *
     * var objects = [
     *   { 'a': 1, 'b': 2, 'c': 3 },
     *   { 'a': 4, 'b': 5, 'c': 6 }
     * ];
     *
     * _.find(objects, _.matchesProperty('a', 4));
     * // => { 'a': 4, 'b': 5, 'c': 6 }
     *
     * // Checking for several possible values
     * _.filter(objects, _.overSome([_.matchesProperty('a', 1), _.matchesProperty('a', 4)]));
     * // => [{ 'a': 1, 'b': 2, 'c': 3 }, { 'a': 4, 'b': 5, 'c': 6 }]
     */
    function matchesProperty(path, srcValue) {
      return baseMatchesProperty(path, baseClone(srcValue, CLONE_DEEP_FLAG));
    }

    /**
     * Creates a function that invokes the method at `path` of a given object.
     * Any additional arguments are provided to the invoked method.
     *
     * @static
     * @memberOf _
     * @since 3.7.0
     * @category Util
     * @param {Array|string} path The path of the method to invoke.
     * @param {...*} [args] The arguments to invoke the method with.
     * @returns {Function} Returns the new invoker function.
     * @example
     *
     * var objects = [
     *   { 'a': { 'b': _.constant(2) } },
     *   { 'a': { 'b': _.constant(1) } }
     * ];
     *
     * _.map(objects, _.method('a.b'));
     * // => [2, 1]
     *
     * _.map(objects, _.method(['a', 'b']));
     * // => [2, 1]
     */
    var method = baseRest(function(path, args) {
      return function(object) {
        return baseInvoke(object, path, args);
      };
    });

    /**
     * The opposite of `_.method`; this method creates a function that invokes
     * the method at a given path of `object`. Any additional arguments are
     * provided to the invoked method.
     *
     * @static
     * @memberOf _
     * @since 3.7.0
     * @category Util
     * @param {Object} object The object to query.
     * @param {...*} [args] The arguments to invoke the method with.
     * @returns {Function} Returns the new invoker function.
     * @example
     *
     * var array = _.times(3, _.constant),
     *     object = { 'a': array, 'b': array, 'c': array };
     *
     * _.map(['a[2]', 'c[0]'], _.methodOf(object));
     * // => [2, 0]
     *
     * _.map([['a', '2'], ['c', '0']], _.methodOf(object));
     * // => [2, 0]
     */
    var methodOf = baseRest(function(object, args) {
      return function(path) {
        return baseInvoke(object, path, args);
      };
    });

    /**
     * Adds all own enumerable string keyed function properties of a source
     * object to the destination object. If `object` is a function, then methods
     * are added to its prototype as well.
     *
     * **Note:** Use `_.runInContext` to create a pristine `lodash` function to
     * avoid conflicts caused by modifying the original.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Util
     * @param {Function|Object} [object=lodash] The destination object.
     * @param {Object} source The object of functions to add.
     * @param {Object} [options={}] The options object.
     * @param {boolean} [options.chain=true] Specify whether mixins are chainable.
     * @returns {Function|Object} Returns `object`.
     * @example
     *
     * function vowels(string) {
     *   return _.filter(string, function(v) {
     *     return /[aeiou]/i.test(v);
     *   });
     * }
     *
     * _.mixin({ 'vowels': vowels });
     * _.vowels('fred');
     * // => ['e']
     *
     * _('fred').vowels().value();
     * // => ['e']
     *
     * _.mixin({ 'vowels': vowels }, { 'chain': false });
     * _('fred').vowels();
     * // => ['e']
     */
    function mixin(object, source, options) {
      var props = keys(source),
          methodNames = baseFunctions(source, props);

      if (options == null &&
          !(isObject(source) && (methodNames.length || !props.length))) {
        options = source;
        source = object;
        object = this;
        methodNames = baseFunctions(source, keys(source));
      }
      var chain = !(isObject(options) && 'chain' in options) || !!options.chain,
          isFunc = isFunction(object);

      arrayEach(methodNames, function(methodName) {
        var func = source[methodName];
        object[methodName] = func;
        if (isFunc) {
          object.prototype[methodName] = function() {
            var chainAll = this.__chain__;
            if (chain || chainAll) {
              var result = object(this.__wrapped__),
                  actions = result.__actions__ = copyArray(this.__actions__);

              actions.push({ 'func': func, 'args': arguments, 'thisArg': object });
              result.__chain__ = chainAll;
              return result;
            }
            return func.apply(object, arrayPush([this.value()], arguments));
          };
        }
      });

      return object;
    }

    /**
     * Reverts the `_` variable to its previous value and returns a reference to
     * the `lodash` function.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Util
     * @returns {Function} Returns the `lodash` function.
     * @example
     *
     * var lodash = _.noConflict();
     */
    function noConflict() {
      if (root._ === this) {
        root._ = oldDash;
      }
      return this;
    }

    /**
     * This method returns `undefined`.
     *
     * @static
     * @memberOf _
     * @since 2.3.0
     * @category Util
     * @example
     *
     * _.times(2, _.noop);
     * // => [undefined, undefined]
     */
    function noop() {
      // No operation performed.
    }

    /**
     * Creates a function that gets the argument at index `n`. If `n` is negative,
     * the nth argument from the end is returned.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Util
     * @param {number} [n=0] The index of the argument to return.
     * @returns {Function} Returns the new pass-thru function.
     * @example
     *
     * var func = _.nthArg(1);
     * func('a', 'b', 'c', 'd');
     * // => 'b'
     *
     * var func = _.nthArg(-2);
     * func('a', 'b', 'c', 'd');
     * // => 'c'
     */
    function nthArg(n) {
      n = toInteger(n);
      return baseRest(function(args) {
        return baseNth(args, n);
      });
    }

    /**
     * Creates a function that invokes `iteratees` with the arguments it receives
     * and returns their results.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Util
     * @param {...(Function|Function[])} [iteratees=[_.identity]]
     *  The iteratees to invoke.
     * @returns {Function} Returns the new function.
     * @example
     *
     * var func = _.over([Math.max, Math.min]);
     *
     * func(1, 2, 3, 4);
     * // => [4, 1]
     */
    var over = createOver(arrayMap);

    /**
     * Creates a function that checks if **all** of the `predicates` return
     * truthy when invoked with the arguments it receives.
     *
     * Following shorthands are possible for providing predicates.
     * Pass an `Object` and it will be used as an parameter for `_.matches` to create the predicate.
     * Pass an `Array` of parameters for `_.matchesProperty` and the predicate will be created using them.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Util
     * @param {...(Function|Function[])} [predicates=[_.identity]]
     *  The predicates to check.
     * @returns {Function} Returns the new function.
     * @example
     *
     * var func = _.overEvery([Boolean, isFinite]);
     *
     * func('1');
     * // => true
     *
     * func(null);
     * // => false
     *
     * func(NaN);
     * // => false
     */
    var overEvery = createOver(arrayEvery);

    /**
     * Creates a function that checks if **any** of the `predicates` return
     * truthy when invoked with the arguments it receives.
     *
     * Following shorthands are possible for providing predicates.
     * Pass an `Object` and it will be used as an parameter for `_.matches` to create the predicate.
     * Pass an `Array` of parameters for `_.matchesProperty` and the predicate will be created using them.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Util
     * @param {...(Function|Function[])} [predicates=[_.identity]]
     *  The predicates to check.
     * @returns {Function} Returns the new function.
     * @example
     *
     * var func = _.overSome([Boolean, isFinite]);
     *
     * func('1');
     * // => true
     *
     * func(null);
     * // => true
     *
     * func(NaN);
     * // => false
     *
     * var matchesFunc = _.overSome([{ 'a': 1 }, { 'a': 2 }])
     * var matchesPropertyFunc = _.overSome([['a', 1], ['a', 2]])
     */
    var overSome = createOver(arraySome);

    /**
     * Creates a function that returns the value at `path` of a given object.
     *
     * @static
     * @memberOf _
     * @since 2.4.0
     * @category Util
     * @param {Array|string} path The path of the property to get.
     * @returns {Function} Returns the new accessor function.
     * @example
     *
     * var objects = [
     *   { 'a': { 'b': 2 } },
     *   { 'a': { 'b': 1 } }
     * ];
     *
     * _.map(objects, _.property('a.b'));
     * // => [2, 1]
     *
     * _.map(_.sortBy(objects, _.property(['a', 'b'])), 'a.b');
     * // => [1, 2]
     */
    function property(path) {
      return isKey(path) ? baseProperty(toKey(path)) : basePropertyDeep(path);
    }

    /**
     * The opposite of `_.property`; this method creates a function that returns
     * the value at a given path of `object`.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Util
     * @param {Object} object The object to query.
     * @returns {Function} Returns the new accessor function.
     * @example
     *
     * var array = [0, 1, 2],
     *     object = { 'a': array, 'b': array, 'c': array };
     *
     * _.map(['a[2]', 'c[0]'], _.propertyOf(object));
     * // => [2, 0]
     *
     * _.map([['a', '2'], ['c', '0']], _.propertyOf(object));
     * // => [2, 0]
     */
    function propertyOf(object) {
      return function(path) {
        return object == null ? undefined : baseGet(object, path);
      };
    }

    /**
     * Creates an array of numbers (positive and/or negative) progressing from
     * `start` up to, but not including, `end`. A step of `-1` is used if a negative
     * `start` is specified without an `end` or `step`. If `end` is not specified,
     * it's set to `start` with `start` then set to `0`.
     *
     * **Note:** JavaScript follows the IEEE-754 standard for resolving
     * floating-point values which can produce unexpected results.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Util
     * @param {number} [start=0] The start of the range.
     * @param {number} end The end of the range.
     * @param {number} [step=1] The value to increment or decrement by.
     * @returns {Array} Returns the range of numbers.
     * @see _.inRange, _.rangeRight
     * @example
     *
     * _.range(4);
     * // => [0, 1, 2, 3]
     *
     * _.range(-4);
     * // => [0, -1, -2, -3]
     *
     * _.range(1, 5);
     * // => [1, 2, 3, 4]
     *
     * _.range(0, 20, 5);
     * // => [0, 5, 10, 15]
     *
     * _.range(0, -4, -1);
     * // => [0, -1, -2, -3]
     *
     * _.range(1, 4, 0);
     * // => [1, 1, 1]
     *
     * _.range(0);
     * // => []
     */
    var range = createRange();

    /**
     * This method is like `_.range` except that it populates values in
     * descending order.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Util
     * @param {number} [start=0] The start of the range.
     * @param {number} end The end of the range.
     * @param {number} [step=1] The value to increment or decrement by.
     * @returns {Array} Returns the range of numbers.
     * @see _.inRange, _.range
     * @example
     *
     * _.rangeRight(4);
     * // => [3, 2, 1, 0]
     *
     * _.rangeRight(-4);
     * // => [-3, -2, -1, 0]
     *
     * _.rangeRight(1, 5);
     * // => [4, 3, 2, 1]
     *
     * _.rangeRight(0, 20, 5);
     * // => [15, 10, 5, 0]
     *
     * _.rangeRight(0, -4, -1);
     * // => [-3, -2, -1, 0]
     *
     * _.rangeRight(1, 4, 0);
     * // => [1, 1, 1]
     *
     * _.rangeRight(0);
     * // => []
     */
    var rangeRight = createRange(true);

    /**
     * This method returns a new empty array.
     *
     * @static
     * @memberOf _
     * @since 4.13.0
     * @category Util
     * @returns {Array} Returns the new empty array.
     * @example
     *
     * var arrays = _.times(2, _.stubArray);
     *
     * console.log(arrays);
     * // => [[], []]
     *
     * console.log(arrays[0] === arrays[1]);
     * // => false
     */
    function stubArray() {
      return [];
    }

    /**
     * This method returns `false`.
     *
     * @static
     * @memberOf _
     * @since 4.13.0
     * @category Util
     * @returns {boolean} Returns `false`.
     * @example
     *
     * _.times(2, _.stubFalse);
     * // => [false, false]
     */
    function stubFalse() {
      return false;
    }

    /**
     * This method returns a new empty object.
     *
     * @static
     * @memberOf _
     * @since 4.13.0
     * @category Util
     * @returns {Object} Returns the new empty object.
     * @example
     *
     * var objects = _.times(2, _.stubObject);
     *
     * console.log(objects);
     * // => [{}, {}]
     *
     * console.log(objects[0] === objects[1]);
     * // => false
     */
    function stubObject() {
      return {};
    }

    /**
     * This method returns an empty string.
     *
     * @static
     * @memberOf _
     * @since 4.13.0
     * @category Util
     * @returns {string} Returns the empty string.
     * @example
     *
     * _.times(2, _.stubString);
     * // => ['', '']
     */
    function stubString() {
      return '';
    }

    /**
     * This method returns `true`.
     *
     * @static
     * @memberOf _
     * @since 4.13.0
     * @category Util
     * @returns {boolean} Returns `true`.
     * @example
     *
     * _.times(2, _.stubTrue);
     * // => [true, true]
     */
    function stubTrue() {
      return true;
    }

    /**
     * Invokes the iteratee `n` times, returning an array of the results of
     * each invocation. The iteratee is invoked with one argument; (index).
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Util
     * @param {number} n The number of times to invoke `iteratee`.
     * @param {Function} [iteratee=_.identity] The function invoked per iteration.
     * @returns {Array} Returns the array of results.
     * @example
     *
     * _.times(3, String);
     * // => ['0', '1', '2']
     *
     *  _.times(4, _.constant(0));
     * // => [0, 0, 0, 0]
     */
    function times(n, iteratee) {
      n = toInteger(n);
      if (n < 1 || n > MAX_SAFE_INTEGER) {
        return [];
      }
      var index = MAX_ARRAY_LENGTH,
          length = nativeMin(n, MAX_ARRAY_LENGTH);

      iteratee = getIteratee(iteratee);
      n -= MAX_ARRAY_LENGTH;

      var result = baseTimes(length, iteratee);
      while (++index < n) {
        iteratee(index);
      }
      return result;
    }

    /**
     * Converts `value` to a property path array.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Util
     * @param {*} value The value to convert.
     * @returns {Array} Returns the new property path array.
     * @example
     *
     * _.toPath('a.b.c');
     * // => ['a', 'b', 'c']
     *
     * _.toPath('a[0].b.c');
     * // => ['a', '0', 'b', 'c']
     */
    function toPath(value) {
      if (isArray(value)) {
        return arrayMap(value, toKey);
      }
      return isSymbol(value) ? [value] : copyArray(stringToPath(toString(value)));
    }

    /**
     * Generates a unique ID. If `prefix` is given, the ID is appended to it.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Util
     * @param {string} [prefix=''] The value to prefix the ID with.
     * @returns {string} Returns the unique ID.
     * @example
     *
     * _.uniqueId('contact_');
     * // => 'contact_104'
     *
     * _.uniqueId();
     * // => '105'
     */
    function uniqueId(prefix) {
      var id = ++idCounter;
      return toString(prefix) + id;
    }

    /*------------------------------------------------------------------------*/

    /**
     * Adds two numbers.
     *
     * @static
     * @memberOf _
     * @since 3.4.0
     * @category Math
     * @param {number} augend The first number in an addition.
     * @param {number} addend The second number in an addition.
     * @returns {number} Returns the total.
     * @example
     *
     * _.add(6, 4);
     * // => 10
     */
    var add = createMathOperation(function(augend, addend) {
      return augend + addend;
    }, 0);

    /**
     * Computes `number` rounded up to `precision`.
     *
     * @static
     * @memberOf _
     * @since 3.10.0
     * @category Math
     * @param {number} number The number to round up.
     * @param {number} [precision=0] The precision to round up to.
     * @returns {number} Returns the rounded up number.
     * @example
     *
     * _.ceil(4.006);
     * // => 5
     *
     * _.ceil(6.004, 2);
     * // => 6.01
     *
     * _.ceil(6040, -2);
     * // => 6100
     */
    var ceil = createRound('ceil');

    /**
     * Divide two numbers.
     *
     * @static
     * @memberOf _
     * @since 4.7.0
     * @category Math
     * @param {number} dividend The first number in a division.
     * @param {number} divisor The second number in a division.
     * @returns {number} Returns the quotient.
     * @example
     *
     * _.divide(6, 4);
     * // => 1.5
     */
    var divide = createMathOperation(function(dividend, divisor) {
      return dividend / divisor;
    }, 1);

    /**
     * Computes `number` rounded down to `precision`.
     *
     * @static
     * @memberOf _
     * @since 3.10.0
     * @category Math
     * @param {number} number The number to round down.
     * @param {number} [precision=0] The precision to round down to.
     * @returns {number} Returns the rounded down number.
     * @example
     *
     * _.floor(4.006);
     * // => 4
     *
     * _.floor(0.046, 2);
     * // => 0.04
     *
     * _.floor(4060, -2);
     * // => 4000
     */
    var floor = createRound('floor');

    /**
     * Computes the maximum value of `array`. If `array` is empty or falsey,
     * `undefined` is returned.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Math
     * @param {Array} array The array to iterate over.
     * @returns {*} Returns the maximum value.
     * @example
     *
     * _.max([4, 2, 8, 6]);
     * // => 8
     *
     * _.max([]);
     * // => undefined
     */
    function max(array) {
      return (array && array.length)
        ? baseExtremum(array, identity, baseGt)
        : undefined;
    }

    /**
     * This method is like `_.max` except that it accepts `iteratee` which is
     * invoked for each element in `array` to generate the criterion by which
     * the value is ranked. The iteratee is invoked with one argument: (value).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Math
     * @param {Array} array The array to iterate over.
     * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
     * @returns {*} Returns the maximum value.
     * @example
     *
     * var objects = [{ 'n': 1 }, { 'n': 2 }];
     *
     * _.maxBy(objects, function(o) { return o.n; });
     * // => { 'n': 2 }
     *
     * // The `_.property` iteratee shorthand.
     * _.maxBy(objects, 'n');
     * // => { 'n': 2 }
     */
    function maxBy(array, iteratee) {
      return (array && array.length)
        ? baseExtremum(array, getIteratee(iteratee, 2), baseGt)
        : undefined;
    }

    /**
     * Computes the mean of the values in `array`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Math
     * @param {Array} array The array to iterate over.
     * @returns {number} Returns the mean.
     * @example
     *
     * _.mean([4, 2, 8, 6]);
     * // => 5
     */
    function mean(array) {
      return baseMean(array, identity);
    }

    /**
     * This method is like `_.mean` except that it accepts `iteratee` which is
     * invoked for each element in `array` to generate the value to be averaged.
     * The iteratee is invoked with one argument: (value).
     *
     * @static
     * @memberOf _
     * @since 4.7.0
     * @category Math
     * @param {Array} array The array to iterate over.
     * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
     * @returns {number} Returns the mean.
     * @example
     *
     * var objects = [{ 'n': 4 }, { 'n': 2 }, { 'n': 8 }, { 'n': 6 }];
     *
     * _.meanBy(objects, function(o) { return o.n; });
     * // => 5
     *
     * // The `_.property` iteratee shorthand.
     * _.meanBy(objects, 'n');
     * // => 5
     */
    function meanBy(array, iteratee) {
      return baseMean(array, getIteratee(iteratee, 2));
    }

    /**
     * Computes the minimum value of `array`. If `array` is empty or falsey,
     * `undefined` is returned.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Math
     * @param {Array} array The array to iterate over.
     * @returns {*} Returns the minimum value.
     * @example
     *
     * _.min([4, 2, 8, 6]);
     * // => 2
     *
     * _.min([]);
     * // => undefined
     */
    function min(array) {
      return (array && array.length)
        ? baseExtremum(array, identity, baseLt)
        : undefined;
    }

    /**
     * This method is like `_.min` except that it accepts `iteratee` which is
     * invoked for each element in `array` to generate the criterion by which
     * the value is ranked. The iteratee is invoked with one argument: (value).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Math
     * @param {Array} array The array to iterate over.
     * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
     * @returns {*} Returns the minimum value.
     * @example
     *
     * var objects = [{ 'n': 1 }, { 'n': 2 }];
     *
     * _.minBy(objects, function(o) { return o.n; });
     * // => { 'n': 1 }
     *
     * // The `_.property` iteratee shorthand.
     * _.minBy(objects, 'n');
     * // => { 'n': 1 }
     */
    function minBy(array, iteratee) {
      return (array && array.length)
        ? baseExtremum(array, getIteratee(iteratee, 2), baseLt)
        : undefined;
    }

    /**
     * Multiply two numbers.
     *
     * @static
     * @memberOf _
     * @since 4.7.0
     * @category Math
     * @param {number} multiplier The first number in a multiplication.
     * @param {number} multiplicand The second number in a multiplication.
     * @returns {number} Returns the product.
     * @example
     *
     * _.multiply(6, 4);
     * // => 24
     */
    var multiply = createMathOperation(function(multiplier, multiplicand) {
      return multiplier * multiplicand;
    }, 1);

    /**
     * Computes `number` rounded to `precision`.
     *
     * @static
     * @memberOf _
     * @since 3.10.0
     * @category Math
     * @param {number} number The number to round.
     * @param {number} [precision=0] The precision to round to.
     * @returns {number} Returns the rounded number.
     * @example
     *
     * _.round(4.006);
     * // => 4
     *
     * _.round(4.006, 2);
     * // => 4.01
     *
     * _.round(4060, -2);
     * // => 4100
     */
    var round = createRound('round');

    /**
     * Subtract two numbers.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Math
     * @param {number} minuend The first number in a subtraction.
     * @param {number} subtrahend The second number in a subtraction.
     * @returns {number} Returns the difference.
     * @example
     *
     * _.subtract(6, 4);
     * // => 2
     */
    var subtract = createMathOperation(function(minuend, subtrahend) {
      return minuend - subtrahend;
    }, 0);

    /**
     * Computes the sum of the values in `array`.
     *
     * @static
     * @memberOf _
     * @since 3.4.0
     * @category Math
     * @param {Array} array The array to iterate over.
     * @returns {number} Returns the sum.
     * @example
     *
     * _.sum([4, 2, 8, 6]);
     * // => 20
     */
    function sum(array) {
      return (array && array.length)
        ? baseSum(array, identity)
        : 0;
    }

    /**
     * This method is like `_.sum` except that it accepts `iteratee` which is
     * invoked for each element in `array` to generate the value to be summed.
     * The iteratee is invoked with one argument: (value).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Math
     * @param {Array} array The array to iterate over.
     * @param {Function} [iteratee=_.identity] The iteratee invoked per element.
     * @returns {number} Returns the sum.
     * @example
     *
     * var objects = [{ 'n': 4 }, { 'n': 2 }, { 'n': 8 }, { 'n': 6 }];
     *
     * _.sumBy(objects, function(o) { return o.n; });
     * // => 20
     *
     * // The `_.property` iteratee shorthand.
     * _.sumBy(objects, 'n');
     * // => 20
     */
    function sumBy(array, iteratee) {
      return (array && array.length)
        ? baseSum(array, getIteratee(iteratee, 2))
        : 0;
    }

    /*------------------------------------------------------------------------*/

    // Add methods that return wrapped values in chain sequences.
    lodash.after = after;
    lodash.ary = ary;
    lodash.assign = assign;
    lodash.assignIn = assignIn;
    lodash.assignInWith = assignInWith;
    lodash.assignWith = assignWith;
    lodash.at = at;
    lodash.before = before;
    lodash.bind = bind;
    lodash.bindAll = bindAll;
    lodash.bindKey = bindKey;
    lodash.castArray = castArray;
    lodash.chain = chain;
    lodash.chunk = chunk;
    lodash.compact = compact;
    lodash.concat = concat;
    lodash.cond = cond;
    lodash.conforms = conforms;
    lodash.constant = constant;
    lodash.countBy = countBy;
    lodash.create = create;
    lodash.curry = curry;
    lodash.curryRight = curryRight;
    lodash.debounce = debounce;
    lodash.defaults = defaults;
    lodash.defaultsDeep = defaultsDeep;
    lodash.defer = defer;
    lodash.delay = delay;
    lodash.difference = difference;
    lodash.differenceBy = differenceBy;
    lodash.differenceWith = differenceWith;
    lodash.drop = drop;
    lodash.dropRight = dropRight;
    lodash.dropRightWhile = dropRightWhile;
    lodash.dropWhile = dropWhile;
    lodash.fill = fill;
    lodash.filter = filter;
    lodash.flatMap = flatMap;
    lodash.flatMapDeep = flatMapDeep;
    lodash.flatMapDepth = flatMapDepth;
    lodash.flatten = flatten;
    lodash.flattenDeep = flattenDeep;
    lodash.flattenDepth = flattenDepth;
    lodash.flip = flip;
    lodash.flow = flow;
    lodash.flowRight = flowRight;
    lodash.fromPairs = fromPairs;
    lodash.functions = functions;
    lodash.functionsIn = functionsIn;
    lodash.groupBy = groupBy;
    lodash.initial = initial;
    lodash.intersection = intersection;
    lodash.intersectionBy = intersectionBy;
    lodash.intersectionWith = intersectionWith;
    lodash.invert = invert;
    lodash.invertBy = invertBy;
    lodash.invokeMap = invokeMap;
    lodash.iteratee = iteratee;
    lodash.keyBy = keyBy;
    lodash.keys = keys;
    lodash.keysIn = keysIn;
    lodash.map = map;
    lodash.mapKeys = mapKeys;
    lodash.mapValues = mapValues;
    lodash.matches = matches;
    lodash.matchesProperty = matchesProperty;
    lodash.memoize = memoize;
    lodash.merge = merge;
    lodash.mergeWith = mergeWith;
    lodash.method = method;
    lodash.methodOf = methodOf;
    lodash.mixin = mixin;
    lodash.negate = negate;
    lodash.nthArg = nthArg;
    lodash.omit = omit;
    lodash.omitBy = omitBy;
    lodash.once = once;
    lodash.orderBy = orderBy;
    lodash.over = over;
    lodash.overArgs = overArgs;
    lodash.overEvery = overEvery;
    lodash.overSome = overSome;
    lodash.partial = partial;
    lodash.partialRight = partialRight;
    lodash.partition = partition;
    lodash.pick = pick;
    lodash.pickBy = pickBy;
    lodash.property = property;
    lodash.propertyOf = propertyOf;
    lodash.pull = pull;
    lodash.pullAll = pullAll;
    lodash.pullAllBy = pullAllBy;
    lodash.pullAllWith = pullAllWith;
    lodash.pullAt = pullAt;
    lodash.range = range;
    lodash.rangeRight = rangeRight;
    lodash.rearg = rearg;
    lodash.reject = reject;
    lodash.remove = remove;
    lodash.rest = rest;
    lodash.reverse = reverse;
    lodash.sampleSize = sampleSize;
    lodash.set = set;
    lodash.setWith = setWith;
    lodash.shuffle = shuffle;
    lodash.slice = slice;
    lodash.sortBy = sortBy;
    lodash.sortedUniq = sortedUniq;
    lodash.sortedUniqBy = sortedUniqBy;
    lodash.split = split;
    lodash.spread = spread;
    lodash.tail = tail;
    lodash.take = take;
    lodash.takeRight = takeRight;
    lodash.takeRightWhile = takeRightWhile;
    lodash.takeWhile = takeWhile;
    lodash.tap = tap;
    lodash.throttle = throttle;
    lodash.thru = thru;
    lodash.toArray = toArray;
    lodash.toPairs = toPairs;
    lodash.toPairsIn = toPairsIn;
    lodash.toPath = toPath;
    lodash.toPlainObject = toPlainObject;
    lodash.transform = transform;
    lodash.unary = unary;
    lodash.union = union;
    lodash.unionBy = unionBy;
    lodash.unionWith = unionWith;
    lodash.uniq = uniq;
    lodash.uniqBy = uniqBy;
    lodash.uniqWith = uniqWith;
    lodash.unset = unset;
    lodash.unzip = unzip;
    lodash.unzipWith = unzipWith;
    lodash.update = update;
    lodash.updateWith = updateWith;
    lodash.values = values;
    lodash.valuesIn = valuesIn;
    lodash.without = without;
    lodash.words = words;
    lodash.wrap = wrap;
    lodash.xor = xor;
    lodash.xorBy = xorBy;
    lodash.xorWith = xorWith;
    lodash.zip = zip;
    lodash.zipObject = zipObject;
    lodash.zipObjectDeep = zipObjectDeep;
    lodash.zipWith = zipWith;

    // Add aliases.
    lodash.entries = toPairs;
    lodash.entriesIn = toPairsIn;
    lodash.extend = assignIn;
    lodash.extendWith = assignInWith;

    // Add methods to `lodash.prototype`.
    mixin(lodash, lodash);

    /*------------------------------------------------------------------------*/

    // Add methods that return unwrapped values in chain sequences.
    lodash.add = add;
    lodash.attempt = attempt;
    lodash.camelCase = camelCase;
    lodash.capitalize = capitalize;
    lodash.ceil = ceil;
    lodash.clamp = clamp;
    lodash.clone = clone;
    lodash.cloneDeep = cloneDeep;
    lodash.cloneDeepWith = cloneDeepWith;
    lodash.cloneWith = cloneWith;
    lodash.conformsTo = conformsTo;
    lodash.deburr = deburr;
    lodash.defaultTo = defaultTo;
    lodash.divide = divide;
    lodash.endsWith = endsWith;
    lodash.eq = eq;
    lodash.escape = escape;
    lodash.escapeRegExp = escapeRegExp;
    lodash.every = every;
    lodash.find = find;
    lodash.findIndex = findIndex;
    lodash.findKey = findKey;
    lodash.findLast = findLast;
    lodash.findLastIndex = findLastIndex;
    lodash.findLastKey = findLastKey;
    lodash.floor = floor;
    lodash.forEach = forEach;
    lodash.forEachRight = forEachRight;
    lodash.forIn = forIn;
    lodash.forInRight = forInRight;
    lodash.forOwn = forOwn;
    lodash.forOwnRight = forOwnRight;
    lodash.get = get;
    lodash.gt = gt;
    lodash.gte = gte;
    lodash.has = has;
    lodash.hasIn = hasIn;
    lodash.head = head;
    lodash.identity = identity;
    lodash.includes = includes;
    lodash.indexOf = indexOf;
    lodash.inRange = inRange;
    lodash.invoke = invoke;
    lodash.isArguments = isArguments;
    lodash.isArray = isArray;
    lodash.isArrayBuffer = isArrayBuffer;
    lodash.isArrayLike = isArrayLike;
    lodash.isArrayLikeObject = isArrayLikeObject;
    lodash.isBoolean = isBoolean;
    lodash.isBuffer = isBuffer;
    lodash.isDate = isDate;
    lodash.isElement = isElement;
    lodash.isEmpty = isEmpty;
    lodash.isEqual = isEqual;
    lodash.isEqualWith = isEqualWith;
    lodash.isError = isError;
    lodash.isFinite = isFinite;
    lodash.isFunction = isFunction;
    lodash.isInteger = isInteger;
    lodash.isLength = isLength;
    lodash.isMap = isMap;
    lodash.isMatch = isMatch;
    lodash.isMatchWith = isMatchWith;
    lodash.isNaN = isNaN;
    lodash.isNative = isNative;
    lodash.isNil = isNil;
    lodash.isNull = isNull;
    lodash.isNumber = isNumber;
    lodash.isObject = isObject;
    lodash.isObjectLike = isObjectLike;
    lodash.isPlainObject = isPlainObject;
    lodash.isRegExp = isRegExp;
    lodash.isSafeInteger = isSafeInteger;
    lodash.isSet = isSet;
    lodash.isString = isString;
    lodash.isSymbol = isSymbol;
    lodash.isTypedArray = isTypedArray;
    lodash.isUndefined = isUndefined;
    lodash.isWeakMap = isWeakMap;
    lodash.isWeakSet = isWeakSet;
    lodash.join = join;
    lodash.kebabCase = kebabCase;
    lodash.last = last;
    lodash.lastIndexOf = lastIndexOf;
    lodash.lowerCase = lowerCase;
    lodash.lowerFirst = lowerFirst;
    lodash.lt = lt;
    lodash.lte = lte;
    lodash.max = max;
    lodash.maxBy = maxBy;
    lodash.mean = mean;
    lodash.meanBy = meanBy;
    lodash.min = min;
    lodash.minBy = minBy;
    lodash.stubArray = stubArray;
    lodash.stubFalse = stubFalse;
    lodash.stubObject = stubObject;
    lodash.stubString = stubString;
    lodash.stubTrue = stubTrue;
    lodash.multiply = multiply;
    lodash.nth = nth;
    lodash.noConflict = noConflict;
    lodash.noop = noop;
    lodash.now = now;
    lodash.pad = pad;
    lodash.padEnd = padEnd;
    lodash.padStart = padStart;
    lodash.parseInt = parseInt;
    lodash.random = random;
    lodash.reduce = reduce;
    lodash.reduceRight = reduceRight;
    lodash.repeat = repeat;
    lodash.replace = replace;
    lodash.result = result;
    lodash.round = round;
    lodash.runInContext = runInContext;
    lodash.sample = sample;
    lodash.size = size;
    lodash.snakeCase = snakeCase;
    lodash.some = some;
    lodash.sortedIndex = sortedIndex;
    lodash.sortedIndexBy = sortedIndexBy;
    lodash.sortedIndexOf = sortedIndexOf;
    lodash.sortedLastIndex = sortedLastIndex;
    lodash.sortedLastIndexBy = sortedLastIndexBy;
    lodash.sortedLastIndexOf = sortedLastIndexOf;
    lodash.startCase = startCase;
    lodash.startsWith = startsWith;
    lodash.subtract = subtract;
    lodash.sum = sum;
    lodash.sumBy = sumBy;
    lodash.template = template;
    lodash.times = times;
    lodash.toFinite = toFinite;
    lodash.toInteger = toInteger;
    lodash.toLength = toLength;
    lodash.toLower = toLower;
    lodash.toNumber = toNumber;
    lodash.toSafeInteger = toSafeInteger;
    lodash.toString = toString;
    lodash.toUpper = toUpper;
    lodash.trim = trim;
    lodash.trimEnd = trimEnd;
    lodash.trimStart = trimStart;
    lodash.truncate = truncate;
    lodash.unescape = unescape;
    lodash.uniqueId = uniqueId;
    lodash.upperCase = upperCase;
    lodash.upperFirst = upperFirst;

    // Add aliases.
    lodash.each = forEach;
    lodash.eachRight = forEachRight;
    lodash.first = head;

    mixin(lodash, (function() {
      var source = {};
      baseForOwn(lodash, function(func, methodName) {
        if (!hasOwnProperty.call(lodash.prototype, methodName)) {
          source[methodName] = func;
        }
      });
      return source;
    }()), { 'chain': false });

    /*------------------------------------------------------------------------*/

    /**
     * The semantic version number.
     *
     * @static
     * @memberOf _
     * @type {string}
     */
    lodash.VERSION = VERSION;

    // Assign default placeholders.
    arrayEach(['bind', 'bindKey', 'curry', 'curryRight', 'partial', 'partialRight'], function(methodName) {
      lodash[methodName].placeholder = lodash;
    });

    // Add `LazyWrapper` methods for `_.drop` and `_.take` variants.
    arrayEach(['drop', 'take'], function(methodName, index) {
      LazyWrapper.prototype[methodName] = function(n) {
        n = n === undefined ? 1 : nativeMax(toInteger(n), 0);

        var result = (this.__filtered__ && !index)
          ? new LazyWrapper(this)
          : this.clone();

        if (result.__filtered__) {
          result.__takeCount__ = nativeMin(n, result.__takeCount__);
        } else {
          result.__views__.push({
            'size': nativeMin(n, MAX_ARRAY_LENGTH),
            'type': methodName + (result.__dir__ < 0 ? 'Right' : '')
          });
        }
        return result;
      };

      LazyWrapper.prototype[methodName + 'Right'] = function(n) {
        return this.reverse()[methodName](n).reverse();
      };
    });

    // Add `LazyWrapper` methods that accept an `iteratee` value.
    arrayEach(['filter', 'map', 'takeWhile'], function(methodName, index) {
      var type = index + 1,
          isFilter = type == LAZY_FILTER_FLAG || type == LAZY_WHILE_FLAG;

      LazyWrapper.prototype[methodName] = function(iteratee) {
        var result = this.clone();
        result.__iteratees__.push({
          'iteratee': getIteratee(iteratee, 3),
          'type': type
        });
        result.__filtered__ = result.__filtered__ || isFilter;
        return result;
      };
    });

    // Add `LazyWrapper` methods for `_.head` and `_.last`.
    arrayEach(['head', 'last'], function(methodName, index) {
      var takeName = 'take' + (index ? 'Right' : '');

      LazyWrapper.prototype[methodName] = function() {
        return this[takeName](1).value()[0];
      };
    });

    // Add `LazyWrapper` methods for `_.initial` and `_.tail`.
    arrayEach(['initial', 'tail'], function(methodName, index) {
      var dropName = 'drop' + (index ? '' : 'Right');

      LazyWrapper.prototype[methodName] = function() {
        return this.__filtered__ ? new LazyWrapper(this) : this[dropName](1);
      };
    });

    LazyWrapper.prototype.compact = function() {
      return this.filter(identity);
    };

    LazyWrapper.prototype.find = function(predicate) {
      return this.filter(predicate).head();
    };

    LazyWrapper.prototype.findLast = function(predicate) {
      return this.reverse().find(predicate);
    };

    LazyWrapper.prototype.invokeMap = baseRest(function(path, args) {
      if (typeof path == 'function') {
        return new LazyWrapper(this);
      }
      return this.map(function(value) {
        return baseInvoke(value, path, args);
      });
    });

    LazyWrapper.prototype.reject = function(predicate) {
      return this.filter(negate(getIteratee(predicate)));
    };

    LazyWrapper.prototype.slice = function(start, end) {
      start = toInteger(start);

      var result = this;
      if (result.__filtered__ && (start > 0 || end < 0)) {
        return new LazyWrapper(result);
      }
      if (start < 0) {
        result = result.takeRight(-start);
      } else if (start) {
        result = result.drop(start);
      }
      if (end !== undefined) {
        end = toInteger(end);
        result = end < 0 ? result.dropRight(-end) : result.take(end - start);
      }
      return result;
    };

    LazyWrapper.prototype.takeRightWhile = function(predicate) {
      return this.reverse().takeWhile(predicate).reverse();
    };

    LazyWrapper.prototype.toArray = function() {
      return this.take(MAX_ARRAY_LENGTH);
    };

    // Add `LazyWrapper` methods to `lodash.prototype`.
    baseForOwn(LazyWrapper.prototype, function(func, methodName) {
      var checkIteratee = /^(?:filter|find|map|reject)|While$/.test(methodName),
          isTaker = /^(?:head|last)$/.test(methodName),
          lodashFunc = lodash[isTaker ? ('take' + (methodName == 'last' ? 'Right' : '')) : methodName],
          retUnwrapped = isTaker || /^find/.test(methodName);

      if (!lodashFunc) {
        return;
      }
      lodash.prototype[methodName] = function() {
        var value = this.__wrapped__,
            args = isTaker ? [1] : arguments,
            isLazy = value instanceof LazyWrapper,
            iteratee = args[0],
            useLazy = isLazy || isArray(value);

        var interceptor = function(value) {
          var result = lodashFunc.apply(lodash, arrayPush([value], args));
          return (isTaker && chainAll) ? result[0] : result;
        };

        if (useLazy && checkIteratee && typeof iteratee == 'function' && iteratee.length != 1) {
          // Avoid lazy use if the iteratee has a "length" value other than `1`.
          isLazy = useLazy = false;
        }
        var chainAll = this.__chain__,
            isHybrid = !!this.__actions__.length,
            isUnwrapped = retUnwrapped && !chainAll,
            onlyLazy = isLazy && !isHybrid;

        if (!retUnwrapped && useLazy) {
          value = onlyLazy ? value : new LazyWrapper(this);
          var result = func.apply(value, args);
          result.__actions__.push({ 'func': thru, 'args': [interceptor], 'thisArg': undefined });
          return new LodashWrapper(result, chainAll);
        }
        if (isUnwrapped && onlyLazy) {
          return func.apply(this, args);
        }
        result = this.thru(interceptor);
        return isUnwrapped ? (isTaker ? result.value()[0] : result.value()) : result;
      };
    });

    // Add `Array` methods to `lodash.prototype`.
    arrayEach(['pop', 'push', 'shift', 'sort', 'splice', 'unshift'], function(methodName) {
      var func = arrayProto[methodName],
          chainName = /^(?:push|sort|unshift)$/.test(methodName) ? 'tap' : 'thru',
          retUnwrapped = /^(?:pop|shift)$/.test(methodName);

      lodash.prototype[methodName] = function() {
        var args = arguments;
        if (retUnwrapped && !this.__chain__) {
          var value = this.value();
          return func.apply(isArray(value) ? value : [], args);
        }
        return this[chainName](function(value) {
          return func.apply(isArray(value) ? value : [], args);
        });
      };
    });

    // Map minified method names to their real names.
    baseForOwn(LazyWrapper.prototype, function(func, methodName) {
      var lodashFunc = lodash[methodName];
      if (lodashFunc) {
        var key = lodashFunc.name + '';
        if (!hasOwnProperty.call(realNames, key)) {
          realNames[key] = [];
        }
        realNames[key].push({ 'name': methodName, 'func': lodashFunc });
      }
    });

    realNames[createHybrid(undefined, WRAP_BIND_KEY_FLAG).name] = [{
      'name': 'wrapper',
      'func': undefined
    }];

    // Add methods to `LazyWrapper`.
    LazyWrapper.prototype.clone = lazyClone;
    LazyWrapper.prototype.reverse = lazyReverse;
    LazyWrapper.prototype.value = lazyValue;

    // Add chain sequence methods to the `lodash` wrapper.
    lodash.prototype.at = wrapperAt;
    lodash.prototype.chain = wrapperChain;
    lodash.prototype.commit = wrapperCommit;
    lodash.prototype.next = wrapperNext;
    lodash.prototype.plant = wrapperPlant;
    lodash.prototype.reverse = wrapperReverse;
    lodash.prototype.toJSON = lodash.prototype.valueOf = lodash.prototype.value = wrapperValue;

    // Add lazy aliases.
    lodash.prototype.first = lodash.prototype.head;

    if (symIterator) {
      lodash.prototype[symIterator] = wrapperToIterator;
    }
    return lodash;
  });

  /*--------------------------------------------------------------------------*/

  // Export lodash.
  var _ = runInContext();

  // Some AMD build optimizers, like r.js, check for condition patterns like:
  if (true) {
    // Expose Lodash on the global object to prevent errors when Lodash is
    // loaded by a script tag in the presence of an AMD loader.
    // See http://requirejs.org/docs/errors.html#mismatch for more details.
    // Use `_.noConflict` to remove Lodash from the global object.
    root._ = _;

    // Define as an anonymous module so, through path mapping, it can be
    // referenced as the "underscore" module.
    !(__WEBPACK_AMD_DEFINE_RESULT__ = (function() {
      return _;
    }).call(exports, __webpack_require__, exports, module),
		__WEBPACK_AMD_DEFINE_RESULT__ !== undefined && (module.exports = __WEBPACK_AMD_DEFINE_RESULT__));
  }
  // Check for `exports` after `define` in case a build optimizer adds it.
  else {}
}.call(this));


/***/ }),

/***/ "./node_modules/lodash/lodash.min.js":
/*!*******************************************!*\
  !*** ./node_modules/lodash/lodash.min.js ***!
  \*******************************************/
/***/ (function(module, exports, __webpack_require__) {

/* module decorator */ module = __webpack_require__.nmd(module);
var __WEBPACK_AMD_DEFINE_RESULT__;/**
 * @license
 * Lodash <https://lodash.com/>
 * Copyright OpenJS Foundation and other contributors <https://openjsf.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */
(function(){function n(n,t,r){switch(r.length){case 0:return n.call(t);case 1:return n.call(t,r[0]);case 2:return n.call(t,r[0],r[1]);case 3:return n.call(t,r[0],r[1],r[2])}return n.apply(t,r)}function t(n,t,r,e){for(var u=-1,i=null==n?0:n.length;++u<i;){var o=n[u];t(e,o,r(o),n)}return e}function r(n,t){for(var r=-1,e=null==n?0:n.length;++r<e&&t(n[r],r,n)!==!1;);return n}function e(n,t){for(var r=null==n?0:n.length;r--&&t(n[r],r,n)!==!1;);return n}function u(n,t){for(var r=-1,e=null==n?0:n.length;++r<e;)if(!t(n[r],r,n))return!1;
return!0}function i(n,t){for(var r=-1,e=null==n?0:n.length,u=0,i=[];++r<e;){var o=n[r];t(o,r,n)&&(i[u++]=o)}return i}function o(n,t){return!!(null==n?0:n.length)&&y(n,t,0)>-1}function f(n,t,r){for(var e=-1,u=null==n?0:n.length;++e<u;)if(r(t,n[e]))return!0;return!1}function c(n,t){for(var r=-1,e=null==n?0:n.length,u=Array(e);++r<e;)u[r]=t(n[r],r,n);return u}function a(n,t){for(var r=-1,e=t.length,u=n.length;++r<e;)n[u+r]=t[r];return n}function l(n,t,r,e){var u=-1,i=null==n?0:n.length;for(e&&i&&(r=n[++u]);++u<i;)r=t(r,n[u],u,n);
return r}function s(n,t,r,e){var u=null==n?0:n.length;for(e&&u&&(r=n[--u]);u--;)r=t(r,n[u],u,n);return r}function h(n,t){for(var r=-1,e=null==n?0:n.length;++r<e;)if(t(n[r],r,n))return!0;return!1}function p(n){return n.split("")}function _(n){return n.match($t)||[]}function v(n,t,r){var e;return r(n,function(n,r,u){if(t(n,r,u))return e=r,!1}),e}function g(n,t,r,e){for(var u=n.length,i=r+(e?1:-1);e?i--:++i<u;)if(t(n[i],i,n))return i;return-1}function y(n,t,r){return t===t?Z(n,t,r):g(n,b,r)}function d(n,t,r,e){
for(var u=r-1,i=n.length;++u<i;)if(e(n[u],t))return u;return-1}function b(n){return n!==n}function w(n,t){var r=null==n?0:n.length;return r?k(n,t)/r:Cn}function m(n){return function(t){return null==t?X:t[n]}}function x(n){return function(t){return null==n?X:n[t]}}function j(n,t,r,e,u){return u(n,function(n,u,i){r=e?(e=!1,n):t(r,n,u,i)}),r}function A(n,t){var r=n.length;for(n.sort(t);r--;)n[r]=n[r].value;return n}function k(n,t){for(var r,e=-1,u=n.length;++e<u;){var i=t(n[e]);i!==X&&(r=r===X?i:r+i);
}return r}function O(n,t){for(var r=-1,e=Array(n);++r<n;)e[r]=t(r);return e}function I(n,t){return c(t,function(t){return[t,n[t]]})}function R(n){return n?n.slice(0,H(n)+1).replace(Lt,""):n}function z(n){return function(t){return n(t)}}function E(n,t){return c(t,function(t){return n[t]})}function S(n,t){return n.has(t)}function W(n,t){for(var r=-1,e=n.length;++r<e&&y(t,n[r],0)>-1;);return r}function L(n,t){for(var r=n.length;r--&&y(t,n[r],0)>-1;);return r}function C(n,t){for(var r=n.length,e=0;r--;)n[r]===t&&++e;
return e}function U(n){return"\\"+Yr[n]}function B(n,t){return null==n?X:n[t]}function T(n){return Nr.test(n)}function $(n){return Pr.test(n)}function D(n){for(var t,r=[];!(t=n.next()).done;)r.push(t.value);return r}function M(n){var t=-1,r=Array(n.size);return n.forEach(function(n,e){r[++t]=[e,n]}),r}function F(n,t){return function(r){return n(t(r))}}function N(n,t){for(var r=-1,e=n.length,u=0,i=[];++r<e;){var o=n[r];o!==t&&o!==cn||(n[r]=cn,i[u++]=r)}return i}function P(n){var t=-1,r=Array(n.size);
return n.forEach(function(n){r[++t]=n}),r}function q(n){var t=-1,r=Array(n.size);return n.forEach(function(n){r[++t]=[n,n]}),r}function Z(n,t,r){for(var e=r-1,u=n.length;++e<u;)if(n[e]===t)return e;return-1}function K(n,t,r){for(var e=r+1;e--;)if(n[e]===t)return e;return e}function V(n){return T(n)?J(n):_e(n)}function G(n){return T(n)?Y(n):p(n)}function H(n){for(var t=n.length;t--&&Ct.test(n.charAt(t)););return t}function J(n){for(var t=Mr.lastIndex=0;Mr.test(n);)++t;return t}function Y(n){return n.match(Mr)||[];
}function Q(n){return n.match(Fr)||[]}var X,nn="4.17.21",tn=200,rn="Unsupported core-js use. Try https://npms.io/search?q=ponyfill.",en="Expected a function",un="Invalid `variable` option passed into `_.template`",on="__lodash_hash_undefined__",fn=500,cn="__lodash_placeholder__",an=1,ln=2,sn=4,hn=1,pn=2,_n=1,vn=2,gn=4,yn=8,dn=16,bn=32,wn=64,mn=128,xn=256,jn=512,An=30,kn="...",On=800,In=16,Rn=1,zn=2,En=3,Sn=1/0,Wn=9007199254740991,Ln=1.7976931348623157e308,Cn=NaN,Un=4294967295,Bn=Un-1,Tn=Un>>>1,$n=[["ary",mn],["bind",_n],["bindKey",vn],["curry",yn],["curryRight",dn],["flip",jn],["partial",bn],["partialRight",wn],["rearg",xn]],Dn="[object Arguments]",Mn="[object Array]",Fn="[object AsyncFunction]",Nn="[object Boolean]",Pn="[object Date]",qn="[object DOMException]",Zn="[object Error]",Kn="[object Function]",Vn="[object GeneratorFunction]",Gn="[object Map]",Hn="[object Number]",Jn="[object Null]",Yn="[object Object]",Qn="[object Promise]",Xn="[object Proxy]",nt="[object RegExp]",tt="[object Set]",rt="[object String]",et="[object Symbol]",ut="[object Undefined]",it="[object WeakMap]",ot="[object WeakSet]",ft="[object ArrayBuffer]",ct="[object DataView]",at="[object Float32Array]",lt="[object Float64Array]",st="[object Int8Array]",ht="[object Int16Array]",pt="[object Int32Array]",_t="[object Uint8Array]",vt="[object Uint8ClampedArray]",gt="[object Uint16Array]",yt="[object Uint32Array]",dt=/\b__p \+= '';/g,bt=/\b(__p \+=) '' \+/g,wt=/(__e\(.*?\)|\b__t\)) \+\n'';/g,mt=/&(?:amp|lt|gt|quot|#39);/g,xt=/[&<>"']/g,jt=RegExp(mt.source),At=RegExp(xt.source),kt=/<%-([\s\S]+?)%>/g,Ot=/<%([\s\S]+?)%>/g,It=/<%=([\s\S]+?)%>/g,Rt=/\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,zt=/^\w*$/,Et=/[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g,St=/[\\^$.*+?()[\]{}|]/g,Wt=RegExp(St.source),Lt=/^\s+/,Ct=/\s/,Ut=/\{(?:\n\/\* \[wrapped with .+\] \*\/)?\n?/,Bt=/\{\n\/\* \[wrapped with (.+)\] \*/,Tt=/,? & /,$t=/[^\x00-\x2f\x3a-\x40\x5b-\x60\x7b-\x7f]+/g,Dt=/[()=,{}\[\]\/\s]/,Mt=/\\(\\)?/g,Ft=/\$\{([^\\}]*(?:\\.[^\\}]*)*)\}/g,Nt=/\w*$/,Pt=/^[-+]0x[0-9a-f]+$/i,qt=/^0b[01]+$/i,Zt=/^\[object .+?Constructor\]$/,Kt=/^0o[0-7]+$/i,Vt=/^(?:0|[1-9]\d*)$/,Gt=/[\xc0-\xd6\xd8-\xf6\xf8-\xff\u0100-\u017f]/g,Ht=/($^)/,Jt=/['\n\r\u2028\u2029\\]/g,Yt="\\ud800-\\udfff",Qt="\\u0300-\\u036f",Xt="\\ufe20-\\ufe2f",nr="\\u20d0-\\u20ff",tr=Qt+Xt+nr,rr="\\u2700-\\u27bf",er="a-z\\xdf-\\xf6\\xf8-\\xff",ur="\\xac\\xb1\\xd7\\xf7",ir="\\x00-\\x2f\\x3a-\\x40\\x5b-\\x60\\x7b-\\xbf",or="\\u2000-\\u206f",fr=" \\t\\x0b\\f\\xa0\\ufeff\\n\\r\\u2028\\u2029\\u1680\\u180e\\u2000\\u2001\\u2002\\u2003\\u2004\\u2005\\u2006\\u2007\\u2008\\u2009\\u200a\\u202f\\u205f\\u3000",cr="A-Z\\xc0-\\xd6\\xd8-\\xde",ar="\\ufe0e\\ufe0f",lr=ur+ir+or+fr,sr="['\u2019]",hr="["+Yt+"]",pr="["+lr+"]",_r="["+tr+"]",vr="\\d+",gr="["+rr+"]",yr="["+er+"]",dr="[^"+Yt+lr+vr+rr+er+cr+"]",br="\\ud83c[\\udffb-\\udfff]",wr="(?:"+_r+"|"+br+")",mr="[^"+Yt+"]",xr="(?:\\ud83c[\\udde6-\\uddff]){2}",jr="[\\ud800-\\udbff][\\udc00-\\udfff]",Ar="["+cr+"]",kr="\\u200d",Or="(?:"+yr+"|"+dr+")",Ir="(?:"+Ar+"|"+dr+")",Rr="(?:"+sr+"(?:d|ll|m|re|s|t|ve))?",zr="(?:"+sr+"(?:D|LL|M|RE|S|T|VE))?",Er=wr+"?",Sr="["+ar+"]?",Wr="(?:"+kr+"(?:"+[mr,xr,jr].join("|")+")"+Sr+Er+")*",Lr="\\d*(?:1st|2nd|3rd|(?![123])\\dth)(?=\\b|[A-Z_])",Cr="\\d*(?:1ST|2ND|3RD|(?![123])\\dTH)(?=\\b|[a-z_])",Ur=Sr+Er+Wr,Br="(?:"+[gr,xr,jr].join("|")+")"+Ur,Tr="(?:"+[mr+_r+"?",_r,xr,jr,hr].join("|")+")",$r=RegExp(sr,"g"),Dr=RegExp(_r,"g"),Mr=RegExp(br+"(?="+br+")|"+Tr+Ur,"g"),Fr=RegExp([Ar+"?"+yr+"+"+Rr+"(?="+[pr,Ar,"$"].join("|")+")",Ir+"+"+zr+"(?="+[pr,Ar+Or,"$"].join("|")+")",Ar+"?"+Or+"+"+Rr,Ar+"+"+zr,Cr,Lr,vr,Br].join("|"),"g"),Nr=RegExp("["+kr+Yt+tr+ar+"]"),Pr=/[a-z][A-Z]|[A-Z]{2}[a-z]|[0-9][a-zA-Z]|[a-zA-Z][0-9]|[^a-zA-Z0-9 ]/,qr=["Array","Buffer","DataView","Date","Error","Float32Array","Float64Array","Function","Int8Array","Int16Array","Int32Array","Map","Math","Object","Promise","RegExp","Set","String","Symbol","TypeError","Uint8Array","Uint8ClampedArray","Uint16Array","Uint32Array","WeakMap","_","clearTimeout","isFinite","parseInt","setTimeout"],Zr=-1,Kr={};
Kr[at]=Kr[lt]=Kr[st]=Kr[ht]=Kr[pt]=Kr[_t]=Kr[vt]=Kr[gt]=Kr[yt]=!0,Kr[Dn]=Kr[Mn]=Kr[ft]=Kr[Nn]=Kr[ct]=Kr[Pn]=Kr[Zn]=Kr[Kn]=Kr[Gn]=Kr[Hn]=Kr[Yn]=Kr[nt]=Kr[tt]=Kr[rt]=Kr[it]=!1;var Vr={};Vr[Dn]=Vr[Mn]=Vr[ft]=Vr[ct]=Vr[Nn]=Vr[Pn]=Vr[at]=Vr[lt]=Vr[st]=Vr[ht]=Vr[pt]=Vr[Gn]=Vr[Hn]=Vr[Yn]=Vr[nt]=Vr[tt]=Vr[rt]=Vr[et]=Vr[_t]=Vr[vt]=Vr[gt]=Vr[yt]=!0,Vr[Zn]=Vr[Kn]=Vr[it]=!1;var Gr={"\xc0":"A","\xc1":"A","\xc2":"A","\xc3":"A","\xc4":"A","\xc5":"A","\xe0":"a","\xe1":"a","\xe2":"a","\xe3":"a","\xe4":"a","\xe5":"a",
"\xc7":"C","\xe7":"c","\xd0":"D","\xf0":"d","\xc8":"E","\xc9":"E","\xca":"E","\xcb":"E","\xe8":"e","\xe9":"e","\xea":"e","\xeb":"e","\xcc":"I","\xcd":"I","\xce":"I","\xcf":"I","\xec":"i","\xed":"i","\xee":"i","\xef":"i","\xd1":"N","\xf1":"n","\xd2":"O","\xd3":"O","\xd4":"O","\xd5":"O","\xd6":"O","\xd8":"O","\xf2":"o","\xf3":"o","\xf4":"o","\xf5":"o","\xf6":"o","\xf8":"o","\xd9":"U","\xda":"U","\xdb":"U","\xdc":"U","\xf9":"u","\xfa":"u","\xfb":"u","\xfc":"u","\xdd":"Y","\xfd":"y","\xff":"y","\xc6":"Ae",
"\xe6":"ae","\xde":"Th","\xfe":"th","\xdf":"ss","\u0100":"A","\u0102":"A","\u0104":"A","\u0101":"a","\u0103":"a","\u0105":"a","\u0106":"C","\u0108":"C","\u010a":"C","\u010c":"C","\u0107":"c","\u0109":"c","\u010b":"c","\u010d":"c","\u010e":"D","\u0110":"D","\u010f":"d","\u0111":"d","\u0112":"E","\u0114":"E","\u0116":"E","\u0118":"E","\u011a":"E","\u0113":"e","\u0115":"e","\u0117":"e","\u0119":"e","\u011b":"e","\u011c":"G","\u011e":"G","\u0120":"G","\u0122":"G","\u011d":"g","\u011f":"g","\u0121":"g",
"\u0123":"g","\u0124":"H","\u0126":"H","\u0125":"h","\u0127":"h","\u0128":"I","\u012a":"I","\u012c":"I","\u012e":"I","\u0130":"I","\u0129":"i","\u012b":"i","\u012d":"i","\u012f":"i","\u0131":"i","\u0134":"J","\u0135":"j","\u0136":"K","\u0137":"k","\u0138":"k","\u0139":"L","\u013b":"L","\u013d":"L","\u013f":"L","\u0141":"L","\u013a":"l","\u013c":"l","\u013e":"l","\u0140":"l","\u0142":"l","\u0143":"N","\u0145":"N","\u0147":"N","\u014a":"N","\u0144":"n","\u0146":"n","\u0148":"n","\u014b":"n","\u014c":"O",
"\u014e":"O","\u0150":"O","\u014d":"o","\u014f":"o","\u0151":"o","\u0154":"R","\u0156":"R","\u0158":"R","\u0155":"r","\u0157":"r","\u0159":"r","\u015a":"S","\u015c":"S","\u015e":"S","\u0160":"S","\u015b":"s","\u015d":"s","\u015f":"s","\u0161":"s","\u0162":"T","\u0164":"T","\u0166":"T","\u0163":"t","\u0165":"t","\u0167":"t","\u0168":"U","\u016a":"U","\u016c":"U","\u016e":"U","\u0170":"U","\u0172":"U","\u0169":"u","\u016b":"u","\u016d":"u","\u016f":"u","\u0171":"u","\u0173":"u","\u0174":"W","\u0175":"w",
"\u0176":"Y","\u0177":"y","\u0178":"Y","\u0179":"Z","\u017b":"Z","\u017d":"Z","\u017a":"z","\u017c":"z","\u017e":"z","\u0132":"IJ","\u0133":"ij","\u0152":"Oe","\u0153":"oe","\u0149":"'n","\u017f":"s"},Hr={"&":"&amp;","<":"&lt;",">":"&gt;",'"':"&quot;","'":"&#39;"},Jr={"&amp;":"&","&lt;":"<","&gt;":">","&quot;":'"',"&#39;":"'"},Yr={"\\":"\\","'":"'","\n":"n","\r":"r","\u2028":"u2028","\u2029":"u2029"},Qr=parseFloat,Xr=parseInt,ne="object"==typeof __webpack_require__.g&&__webpack_require__.g&&__webpack_require__.g.Object===Object&&__webpack_require__.g,te="object"==typeof self&&self&&self.Object===Object&&self,re=ne||te||Function("return this")(),ee= true&&exports&&!exports.nodeType&&exports,ue=ee&&"object"=="object"&&module&&!module.nodeType&&module,ie=ue&&ue.exports===ee,oe=ie&&ne.process,fe=function(){
try{var n=ue&&ue.require&&ue.require("util").types;return n?n:oe&&oe.binding&&oe.binding("util")}catch(n){}}(),ce=fe&&fe.isArrayBuffer,ae=fe&&fe.isDate,le=fe&&fe.isMap,se=fe&&fe.isRegExp,he=fe&&fe.isSet,pe=fe&&fe.isTypedArray,_e=m("length"),ve=x(Gr),ge=x(Hr),ye=x(Jr),de=function p(x){function Z(n){if(cc(n)&&!bh(n)&&!(n instanceof Ct)){if(n instanceof Y)return n;if(bl.call(n,"__wrapped__"))return eo(n)}return new Y(n)}function J(){}function Y(n,t){this.__wrapped__=n,this.__actions__=[],this.__chain__=!!t,
this.__index__=0,this.__values__=X}function Ct(n){this.__wrapped__=n,this.__actions__=[],this.__dir__=1,this.__filtered__=!1,this.__iteratees__=[],this.__takeCount__=Un,this.__views__=[]}function $t(){var n=new Ct(this.__wrapped__);return n.__actions__=Tu(this.__actions__),n.__dir__=this.__dir__,n.__filtered__=this.__filtered__,n.__iteratees__=Tu(this.__iteratees__),n.__takeCount__=this.__takeCount__,n.__views__=Tu(this.__views__),n}function Yt(){if(this.__filtered__){var n=new Ct(this);n.__dir__=-1,
n.__filtered__=!0}else n=this.clone(),n.__dir__*=-1;return n}function Qt(){var n=this.__wrapped__.value(),t=this.__dir__,r=bh(n),e=t<0,u=r?n.length:0,i=Oi(0,u,this.__views__),o=i.start,f=i.end,c=f-o,a=e?f:o-1,l=this.__iteratees__,s=l.length,h=0,p=Hl(c,this.__takeCount__);if(!r||!e&&u==c&&p==c)return wu(n,this.__actions__);var _=[];n:for(;c--&&h<p;){a+=t;for(var v=-1,g=n[a];++v<s;){var y=l[v],d=y.iteratee,b=y.type,w=d(g);if(b==zn)g=w;else if(!w){if(b==Rn)continue n;break n}}_[h++]=g}return _}function Xt(n){
var t=-1,r=null==n?0:n.length;for(this.clear();++t<r;){var e=n[t];this.set(e[0],e[1])}}function nr(){this.__data__=is?is(null):{},this.size=0}function tr(n){var t=this.has(n)&&delete this.__data__[n];return this.size-=t?1:0,t}function rr(n){var t=this.__data__;if(is){var r=t[n];return r===on?X:r}return bl.call(t,n)?t[n]:X}function er(n){var t=this.__data__;return is?t[n]!==X:bl.call(t,n)}function ur(n,t){var r=this.__data__;return this.size+=this.has(n)?0:1,r[n]=is&&t===X?on:t,this}function ir(n){
var t=-1,r=null==n?0:n.length;for(this.clear();++t<r;){var e=n[t];this.set(e[0],e[1])}}function or(){this.__data__=[],this.size=0}function fr(n){var t=this.__data__,r=Wr(t,n);return!(r<0)&&(r==t.length-1?t.pop():Ll.call(t,r,1),--this.size,!0)}function cr(n){var t=this.__data__,r=Wr(t,n);return r<0?X:t[r][1]}function ar(n){return Wr(this.__data__,n)>-1}function lr(n,t){var r=this.__data__,e=Wr(r,n);return e<0?(++this.size,r.push([n,t])):r[e][1]=t,this}function sr(n){var t=-1,r=null==n?0:n.length;for(this.clear();++t<r;){
var e=n[t];this.set(e[0],e[1])}}function hr(){this.size=0,this.__data__={hash:new Xt,map:new(ts||ir),string:new Xt}}function pr(n){var t=xi(this,n).delete(n);return this.size-=t?1:0,t}function _r(n){return xi(this,n).get(n)}function vr(n){return xi(this,n).has(n)}function gr(n,t){var r=xi(this,n),e=r.size;return r.set(n,t),this.size+=r.size==e?0:1,this}function yr(n){var t=-1,r=null==n?0:n.length;for(this.__data__=new sr;++t<r;)this.add(n[t])}function dr(n){return this.__data__.set(n,on),this}function br(n){
return this.__data__.has(n)}function wr(n){this.size=(this.__data__=new ir(n)).size}function mr(){this.__data__=new ir,this.size=0}function xr(n){var t=this.__data__,r=t.delete(n);return this.size=t.size,r}function jr(n){return this.__data__.get(n)}function Ar(n){return this.__data__.has(n)}function kr(n,t){var r=this.__data__;if(r instanceof ir){var e=r.__data__;if(!ts||e.length<tn-1)return e.push([n,t]),this.size=++r.size,this;r=this.__data__=new sr(e)}return r.set(n,t),this.size=r.size,this}function Or(n,t){
var r=bh(n),e=!r&&dh(n),u=!r&&!e&&mh(n),i=!r&&!e&&!u&&Oh(n),o=r||e||u||i,f=o?O(n.length,hl):[],c=f.length;for(var a in n)!t&&!bl.call(n,a)||o&&("length"==a||u&&("offset"==a||"parent"==a)||i&&("buffer"==a||"byteLength"==a||"byteOffset"==a)||Ci(a,c))||f.push(a);return f}function Ir(n){var t=n.length;return t?n[tu(0,t-1)]:X}function Rr(n,t){return Xi(Tu(n),Mr(t,0,n.length))}function zr(n){return Xi(Tu(n))}function Er(n,t,r){(r===X||Gf(n[t],r))&&(r!==X||t in n)||Br(n,t,r)}function Sr(n,t,r){var e=n[t];
bl.call(n,t)&&Gf(e,r)&&(r!==X||t in n)||Br(n,t,r)}function Wr(n,t){for(var r=n.length;r--;)if(Gf(n[r][0],t))return r;return-1}function Lr(n,t,r,e){return ys(n,function(n,u,i){t(e,n,r(n),i)}),e}function Cr(n,t){return n&&$u(t,Pc(t),n)}function Ur(n,t){return n&&$u(t,qc(t),n)}function Br(n,t,r){"__proto__"==t&&Tl?Tl(n,t,{configurable:!0,enumerable:!0,value:r,writable:!0}):n[t]=r}function Tr(n,t){for(var r=-1,e=t.length,u=il(e),i=null==n;++r<e;)u[r]=i?X:Mc(n,t[r]);return u}function Mr(n,t,r){return n===n&&(r!==X&&(n=n<=r?n:r),
t!==X&&(n=n>=t?n:t)),n}function Fr(n,t,e,u,i,o){var f,c=t&an,a=t&ln,l=t&sn;if(e&&(f=i?e(n,u,i,o):e(n)),f!==X)return f;if(!fc(n))return n;var s=bh(n);if(s){if(f=zi(n),!c)return Tu(n,f)}else{var h=zs(n),p=h==Kn||h==Vn;if(mh(n))return Iu(n,c);if(h==Yn||h==Dn||p&&!i){if(f=a||p?{}:Ei(n),!c)return a?Mu(n,Ur(f,n)):Du(n,Cr(f,n))}else{if(!Vr[h])return i?n:{};f=Si(n,h,c)}}o||(o=new wr);var _=o.get(n);if(_)return _;o.set(n,f),kh(n)?n.forEach(function(r){f.add(Fr(r,t,e,r,n,o))}):jh(n)&&n.forEach(function(r,u){
f.set(u,Fr(r,t,e,u,n,o))});var v=l?a?di:yi:a?qc:Pc,g=s?X:v(n);return r(g||n,function(r,u){g&&(u=r,r=n[u]),Sr(f,u,Fr(r,t,e,u,n,o))}),f}function Nr(n){var t=Pc(n);return function(r){return Pr(r,n,t)}}function Pr(n,t,r){var e=r.length;if(null==n)return!e;for(n=ll(n);e--;){var u=r[e],i=t[u],o=n[u];if(o===X&&!(u in n)||!i(o))return!1}return!0}function Gr(n,t,r){if("function"!=typeof n)throw new pl(en);return Ws(function(){n.apply(X,r)},t)}function Hr(n,t,r,e){var u=-1,i=o,a=!0,l=n.length,s=[],h=t.length;
if(!l)return s;r&&(t=c(t,z(r))),e?(i=f,a=!1):t.length>=tn&&(i=S,a=!1,t=new yr(t));n:for(;++u<l;){var p=n[u],_=null==r?p:r(p);if(p=e||0!==p?p:0,a&&_===_){for(var v=h;v--;)if(t[v]===_)continue n;s.push(p)}else i(t,_,e)||s.push(p)}return s}function Jr(n,t){var r=!0;return ys(n,function(n,e,u){return r=!!t(n,e,u)}),r}function Yr(n,t,r){for(var e=-1,u=n.length;++e<u;){var i=n[e],o=t(i);if(null!=o&&(f===X?o===o&&!bc(o):r(o,f)))var f=o,c=i}return c}function ne(n,t,r,e){var u=n.length;for(r=kc(r),r<0&&(r=-r>u?0:u+r),
e=e===X||e>u?u:kc(e),e<0&&(e+=u),e=r>e?0:Oc(e);r<e;)n[r++]=t;return n}function te(n,t){var r=[];return ys(n,function(n,e,u){t(n,e,u)&&r.push(n)}),r}function ee(n,t,r,e,u){var i=-1,o=n.length;for(r||(r=Li),u||(u=[]);++i<o;){var f=n[i];t>0&&r(f)?t>1?ee(f,t-1,r,e,u):a(u,f):e||(u[u.length]=f)}return u}function ue(n,t){return n&&bs(n,t,Pc)}function oe(n,t){return n&&ws(n,t,Pc)}function fe(n,t){return i(t,function(t){return uc(n[t])})}function _e(n,t){t=ku(t,n);for(var r=0,e=t.length;null!=n&&r<e;)n=n[no(t[r++])];
return r&&r==e?n:X}function de(n,t,r){var e=t(n);return bh(n)?e:a(e,r(n))}function we(n){return null==n?n===X?ut:Jn:Bl&&Bl in ll(n)?ki(n):Ki(n)}function me(n,t){return n>t}function xe(n,t){return null!=n&&bl.call(n,t)}function je(n,t){return null!=n&&t in ll(n)}function Ae(n,t,r){return n>=Hl(t,r)&&n<Gl(t,r)}function ke(n,t,r){for(var e=r?f:o,u=n[0].length,i=n.length,a=i,l=il(i),s=1/0,h=[];a--;){var p=n[a];a&&t&&(p=c(p,z(t))),s=Hl(p.length,s),l[a]=!r&&(t||u>=120&&p.length>=120)?new yr(a&&p):X}p=n[0];
var _=-1,v=l[0];n:for(;++_<u&&h.length<s;){var g=p[_],y=t?t(g):g;if(g=r||0!==g?g:0,!(v?S(v,y):e(h,y,r))){for(a=i;--a;){var d=l[a];if(!(d?S(d,y):e(n[a],y,r)))continue n}v&&v.push(y),h.push(g)}}return h}function Oe(n,t,r,e){return ue(n,function(n,u,i){t(e,r(n),u,i)}),e}function Ie(t,r,e){r=ku(r,t),t=Gi(t,r);var u=null==t?t:t[no(jo(r))];return null==u?X:n(u,t,e)}function Re(n){return cc(n)&&we(n)==Dn}function ze(n){return cc(n)&&we(n)==ft}function Ee(n){return cc(n)&&we(n)==Pn}function Se(n,t,r,e,u){
return n===t||(null==n||null==t||!cc(n)&&!cc(t)?n!==n&&t!==t:We(n,t,r,e,Se,u))}function We(n,t,r,e,u,i){var o=bh(n),f=bh(t),c=o?Mn:zs(n),a=f?Mn:zs(t);c=c==Dn?Yn:c,a=a==Dn?Yn:a;var l=c==Yn,s=a==Yn,h=c==a;if(h&&mh(n)){if(!mh(t))return!1;o=!0,l=!1}if(h&&!l)return i||(i=new wr),o||Oh(n)?pi(n,t,r,e,u,i):_i(n,t,c,r,e,u,i);if(!(r&hn)){var p=l&&bl.call(n,"__wrapped__"),_=s&&bl.call(t,"__wrapped__");if(p||_){var v=p?n.value():n,g=_?t.value():t;return i||(i=new wr),u(v,g,r,e,i)}}return!!h&&(i||(i=new wr),vi(n,t,r,e,u,i));
}function Le(n){return cc(n)&&zs(n)==Gn}function Ce(n,t,r,e){var u=r.length,i=u,o=!e;if(null==n)return!i;for(n=ll(n);u--;){var f=r[u];if(o&&f[2]?f[1]!==n[f[0]]:!(f[0]in n))return!1}for(;++u<i;){f=r[u];var c=f[0],a=n[c],l=f[1];if(o&&f[2]){if(a===X&&!(c in n))return!1}else{var s=new wr;if(e)var h=e(a,l,c,n,t,s);if(!(h===X?Se(l,a,hn|pn,e,s):h))return!1}}return!0}function Ue(n){return!(!fc(n)||Di(n))&&(uc(n)?kl:Zt).test(to(n))}function Be(n){return cc(n)&&we(n)==nt}function Te(n){return cc(n)&&zs(n)==tt;
}function $e(n){return cc(n)&&oc(n.length)&&!!Kr[we(n)]}function De(n){return"function"==typeof n?n:null==n?La:"object"==typeof n?bh(n)?Ze(n[0],n[1]):qe(n):Fa(n)}function Me(n){if(!Mi(n))return Vl(n);var t=[];for(var r in ll(n))bl.call(n,r)&&"constructor"!=r&&t.push(r);return t}function Fe(n){if(!fc(n))return Zi(n);var t=Mi(n),r=[];for(var e in n)("constructor"!=e||!t&&bl.call(n,e))&&r.push(e);return r}function Ne(n,t){return n<t}function Pe(n,t){var r=-1,e=Hf(n)?il(n.length):[];return ys(n,function(n,u,i){
e[++r]=t(n,u,i)}),e}function qe(n){var t=ji(n);return 1==t.length&&t[0][2]?Ni(t[0][0],t[0][1]):function(r){return r===n||Ce(r,n,t)}}function Ze(n,t){return Bi(n)&&Fi(t)?Ni(no(n),t):function(r){var e=Mc(r,n);return e===X&&e===t?Nc(r,n):Se(t,e,hn|pn)}}function Ke(n,t,r,e,u){n!==t&&bs(t,function(i,o){if(u||(u=new wr),fc(i))Ve(n,t,o,r,Ke,e,u);else{var f=e?e(Ji(n,o),i,o+"",n,t,u):X;f===X&&(f=i),Er(n,o,f)}},qc)}function Ve(n,t,r,e,u,i,o){var f=Ji(n,r),c=Ji(t,r),a=o.get(c);if(a)return Er(n,r,a),X;var l=i?i(f,c,r+"",n,t,o):X,s=l===X;
if(s){var h=bh(c),p=!h&&mh(c),_=!h&&!p&&Oh(c);l=c,h||p||_?bh(f)?l=f:Jf(f)?l=Tu(f):p?(s=!1,l=Iu(c,!0)):_?(s=!1,l=Wu(c,!0)):l=[]:gc(c)||dh(c)?(l=f,dh(f)?l=Rc(f):fc(f)&&!uc(f)||(l=Ei(c))):s=!1}s&&(o.set(c,l),u(l,c,e,i,o),o.delete(c)),Er(n,r,l)}function Ge(n,t){var r=n.length;if(r)return t+=t<0?r:0,Ci(t,r)?n[t]:X}function He(n,t,r){t=t.length?c(t,function(n){return bh(n)?function(t){return _e(t,1===n.length?n[0]:n)}:n}):[La];var e=-1;return t=c(t,z(mi())),A(Pe(n,function(n,r,u){return{criteria:c(t,function(t){
return t(n)}),index:++e,value:n}}),function(n,t){return Cu(n,t,r)})}function Je(n,t){return Ye(n,t,function(t,r){return Nc(n,r)})}function Ye(n,t,r){for(var e=-1,u=t.length,i={};++e<u;){var o=t[e],f=_e(n,o);r(f,o)&&fu(i,ku(o,n),f)}return i}function Qe(n){return function(t){return _e(t,n)}}function Xe(n,t,r,e){var u=e?d:y,i=-1,o=t.length,f=n;for(n===t&&(t=Tu(t)),r&&(f=c(n,z(r)));++i<o;)for(var a=0,l=t[i],s=r?r(l):l;(a=u(f,s,a,e))>-1;)f!==n&&Ll.call(f,a,1),Ll.call(n,a,1);return n}function nu(n,t){for(var r=n?t.length:0,e=r-1;r--;){
var u=t[r];if(r==e||u!==i){var i=u;Ci(u)?Ll.call(n,u,1):yu(n,u)}}return n}function tu(n,t){return n+Nl(Ql()*(t-n+1))}function ru(n,t,r,e){for(var u=-1,i=Gl(Fl((t-n)/(r||1)),0),o=il(i);i--;)o[e?i:++u]=n,n+=r;return o}function eu(n,t){var r="";if(!n||t<1||t>Wn)return r;do t%2&&(r+=n),t=Nl(t/2),t&&(n+=n);while(t);return r}function uu(n,t){return Ls(Vi(n,t,La),n+"")}function iu(n){return Ir(ra(n))}function ou(n,t){var r=ra(n);return Xi(r,Mr(t,0,r.length))}function fu(n,t,r,e){if(!fc(n))return n;t=ku(t,n);
for(var u=-1,i=t.length,o=i-1,f=n;null!=f&&++u<i;){var c=no(t[u]),a=r;if("__proto__"===c||"constructor"===c||"prototype"===c)return n;if(u!=o){var l=f[c];a=e?e(l,c,f):X,a===X&&(a=fc(l)?l:Ci(t[u+1])?[]:{})}Sr(f,c,a),f=f[c]}return n}function cu(n){return Xi(ra(n))}function au(n,t,r){var e=-1,u=n.length;t<0&&(t=-t>u?0:u+t),r=r>u?u:r,r<0&&(r+=u),u=t>r?0:r-t>>>0,t>>>=0;for(var i=il(u);++e<u;)i[e]=n[e+t];return i}function lu(n,t){var r;return ys(n,function(n,e,u){return r=t(n,e,u),!r}),!!r}function su(n,t,r){
var e=0,u=null==n?e:n.length;if("number"==typeof t&&t===t&&u<=Tn){for(;e<u;){var i=e+u>>>1,o=n[i];null!==o&&!bc(o)&&(r?o<=t:o<t)?e=i+1:u=i}return u}return hu(n,t,La,r)}function hu(n,t,r,e){var u=0,i=null==n?0:n.length;if(0===i)return 0;t=r(t);for(var o=t!==t,f=null===t,c=bc(t),a=t===X;u<i;){var l=Nl((u+i)/2),s=r(n[l]),h=s!==X,p=null===s,_=s===s,v=bc(s);if(o)var g=e||_;else g=a?_&&(e||h):f?_&&h&&(e||!p):c?_&&h&&!p&&(e||!v):!p&&!v&&(e?s<=t:s<t);g?u=l+1:i=l}return Hl(i,Bn)}function pu(n,t){for(var r=-1,e=n.length,u=0,i=[];++r<e;){
var o=n[r],f=t?t(o):o;if(!r||!Gf(f,c)){var c=f;i[u++]=0===o?0:o}}return i}function _u(n){return"number"==typeof n?n:bc(n)?Cn:+n}function vu(n){if("string"==typeof n)return n;if(bh(n))return c(n,vu)+"";if(bc(n))return vs?vs.call(n):"";var t=n+"";return"0"==t&&1/n==-Sn?"-0":t}function gu(n,t,r){var e=-1,u=o,i=n.length,c=!0,a=[],l=a;if(r)c=!1,u=f;else if(i>=tn){var s=t?null:ks(n);if(s)return P(s);c=!1,u=S,l=new yr}else l=t?[]:a;n:for(;++e<i;){var h=n[e],p=t?t(h):h;if(h=r||0!==h?h:0,c&&p===p){for(var _=l.length;_--;)if(l[_]===p)continue n;
t&&l.push(p),a.push(h)}else u(l,p,r)||(l!==a&&l.push(p),a.push(h))}return a}function yu(n,t){return t=ku(t,n),n=Gi(n,t),null==n||delete n[no(jo(t))]}function du(n,t,r,e){return fu(n,t,r(_e(n,t)),e)}function bu(n,t,r,e){for(var u=n.length,i=e?u:-1;(e?i--:++i<u)&&t(n[i],i,n););return r?au(n,e?0:i,e?i+1:u):au(n,e?i+1:0,e?u:i)}function wu(n,t){var r=n;return r instanceof Ct&&(r=r.value()),l(t,function(n,t){return t.func.apply(t.thisArg,a([n],t.args))},r)}function mu(n,t,r){var e=n.length;if(e<2)return e?gu(n[0]):[];
for(var u=-1,i=il(e);++u<e;)for(var o=n[u],f=-1;++f<e;)f!=u&&(i[u]=Hr(i[u]||o,n[f],t,r));return gu(ee(i,1),t,r)}function xu(n,t,r){for(var e=-1,u=n.length,i=t.length,o={};++e<u;){r(o,n[e],e<i?t[e]:X)}return o}function ju(n){return Jf(n)?n:[]}function Au(n){return"function"==typeof n?n:La}function ku(n,t){return bh(n)?n:Bi(n,t)?[n]:Cs(Ec(n))}function Ou(n,t,r){var e=n.length;return r=r===X?e:r,!t&&r>=e?n:au(n,t,r)}function Iu(n,t){if(t)return n.slice();var r=n.length,e=zl?zl(r):new n.constructor(r);
return n.copy(e),e}function Ru(n){var t=new n.constructor(n.byteLength);return new Rl(t).set(new Rl(n)),t}function zu(n,t){return new n.constructor(t?Ru(n.buffer):n.buffer,n.byteOffset,n.byteLength)}function Eu(n){var t=new n.constructor(n.source,Nt.exec(n));return t.lastIndex=n.lastIndex,t}function Su(n){return _s?ll(_s.call(n)):{}}function Wu(n,t){return new n.constructor(t?Ru(n.buffer):n.buffer,n.byteOffset,n.length)}function Lu(n,t){if(n!==t){var r=n!==X,e=null===n,u=n===n,i=bc(n),o=t!==X,f=null===t,c=t===t,a=bc(t);
if(!f&&!a&&!i&&n>t||i&&o&&c&&!f&&!a||e&&o&&c||!r&&c||!u)return 1;if(!e&&!i&&!a&&n<t||a&&r&&u&&!e&&!i||f&&r&&u||!o&&u||!c)return-1}return 0}function Cu(n,t,r){for(var e=-1,u=n.criteria,i=t.criteria,o=u.length,f=r.length;++e<o;){var c=Lu(u[e],i[e]);if(c){if(e>=f)return c;return c*("desc"==r[e]?-1:1)}}return n.index-t.index}function Uu(n,t,r,e){for(var u=-1,i=n.length,o=r.length,f=-1,c=t.length,a=Gl(i-o,0),l=il(c+a),s=!e;++f<c;)l[f]=t[f];for(;++u<o;)(s||u<i)&&(l[r[u]]=n[u]);for(;a--;)l[f++]=n[u++];return l;
}function Bu(n,t,r,e){for(var u=-1,i=n.length,o=-1,f=r.length,c=-1,a=t.length,l=Gl(i-f,0),s=il(l+a),h=!e;++u<l;)s[u]=n[u];for(var p=u;++c<a;)s[p+c]=t[c];for(;++o<f;)(h||u<i)&&(s[p+r[o]]=n[u++]);return s}function Tu(n,t){var r=-1,e=n.length;for(t||(t=il(e));++r<e;)t[r]=n[r];return t}function $u(n,t,r,e){var u=!r;r||(r={});for(var i=-1,o=t.length;++i<o;){var f=t[i],c=e?e(r[f],n[f],f,r,n):X;c===X&&(c=n[f]),u?Br(r,f,c):Sr(r,f,c)}return r}function Du(n,t){return $u(n,Is(n),t)}function Mu(n,t){return $u(n,Rs(n),t);
}function Fu(n,r){return function(e,u){var i=bh(e)?t:Lr,o=r?r():{};return i(e,n,mi(u,2),o)}}function Nu(n){return uu(function(t,r){var e=-1,u=r.length,i=u>1?r[u-1]:X,o=u>2?r[2]:X;for(i=n.length>3&&"function"==typeof i?(u--,i):X,o&&Ui(r[0],r[1],o)&&(i=u<3?X:i,u=1),t=ll(t);++e<u;){var f=r[e];f&&n(t,f,e,i)}return t})}function Pu(n,t){return function(r,e){if(null==r)return r;if(!Hf(r))return n(r,e);for(var u=r.length,i=t?u:-1,o=ll(r);(t?i--:++i<u)&&e(o[i],i,o)!==!1;);return r}}function qu(n){return function(t,r,e){
for(var u=-1,i=ll(t),o=e(t),f=o.length;f--;){var c=o[n?f:++u];if(r(i[c],c,i)===!1)break}return t}}function Zu(n,t,r){function e(){return(this&&this!==re&&this instanceof e?i:n).apply(u?r:this,arguments)}var u=t&_n,i=Gu(n);return e}function Ku(n){return function(t){t=Ec(t);var r=T(t)?G(t):X,e=r?r[0]:t.charAt(0),u=r?Ou(r,1).join(""):t.slice(1);return e[n]()+u}}function Vu(n){return function(t){return l(Ra(ca(t).replace($r,"")),n,"")}}function Gu(n){return function(){var t=arguments;switch(t.length){
case 0:return new n;case 1:return new n(t[0]);case 2:return new n(t[0],t[1]);case 3:return new n(t[0],t[1],t[2]);case 4:return new n(t[0],t[1],t[2],t[3]);case 5:return new n(t[0],t[1],t[2],t[3],t[4]);case 6:return new n(t[0],t[1],t[2],t[3],t[4],t[5]);case 7:return new n(t[0],t[1],t[2],t[3],t[4],t[5],t[6])}var r=gs(n.prototype),e=n.apply(r,t);return fc(e)?e:r}}function Hu(t,r,e){function u(){for(var o=arguments.length,f=il(o),c=o,a=wi(u);c--;)f[c]=arguments[c];var l=o<3&&f[0]!==a&&f[o-1]!==a?[]:N(f,a);
return o-=l.length,o<e?oi(t,r,Qu,u.placeholder,X,f,l,X,X,e-o):n(this&&this!==re&&this instanceof u?i:t,this,f)}var i=Gu(t);return u}function Ju(n){return function(t,r,e){var u=ll(t);if(!Hf(t)){var i=mi(r,3);t=Pc(t),r=function(n){return i(u[n],n,u)}}var o=n(t,r,e);return o>-1?u[i?t[o]:o]:X}}function Yu(n){return gi(function(t){var r=t.length,e=r,u=Y.prototype.thru;for(n&&t.reverse();e--;){var i=t[e];if("function"!=typeof i)throw new pl(en);if(u&&!o&&"wrapper"==bi(i))var o=new Y([],!0)}for(e=o?e:r;++e<r;){
i=t[e];var f=bi(i),c="wrapper"==f?Os(i):X;o=c&&$i(c[0])&&c[1]==(mn|yn|bn|xn)&&!c[4].length&&1==c[9]?o[bi(c[0])].apply(o,c[3]):1==i.length&&$i(i)?o[f]():o.thru(i)}return function(){var n=arguments,e=n[0];if(o&&1==n.length&&bh(e))return o.plant(e).value();for(var u=0,i=r?t[u].apply(this,n):e;++u<r;)i=t[u].call(this,i);return i}})}function Qu(n,t,r,e,u,i,o,f,c,a){function l(){for(var y=arguments.length,d=il(y),b=y;b--;)d[b]=arguments[b];if(_)var w=wi(l),m=C(d,w);if(e&&(d=Uu(d,e,u,_)),i&&(d=Bu(d,i,o,_)),
y-=m,_&&y<a){return oi(n,t,Qu,l.placeholder,r,d,N(d,w),f,c,a-y)}var x=h?r:this,j=p?x[n]:n;return y=d.length,f?d=Hi(d,f):v&&y>1&&d.reverse(),s&&c<y&&(d.length=c),this&&this!==re&&this instanceof l&&(j=g||Gu(j)),j.apply(x,d)}var s=t&mn,h=t&_n,p=t&vn,_=t&(yn|dn),v=t&jn,g=p?X:Gu(n);return l}function Xu(n,t){return function(r,e){return Oe(r,n,t(e),{})}}function ni(n,t){return function(r,e){var u;if(r===X&&e===X)return t;if(r!==X&&(u=r),e!==X){if(u===X)return e;"string"==typeof r||"string"==typeof e?(r=vu(r),
e=vu(e)):(r=_u(r),e=_u(e)),u=n(r,e)}return u}}function ti(t){return gi(function(r){return r=c(r,z(mi())),uu(function(e){var u=this;return t(r,function(t){return n(t,u,e)})})})}function ri(n,t){t=t===X?" ":vu(t);var r=t.length;if(r<2)return r?eu(t,n):t;var e=eu(t,Fl(n/V(t)));return T(t)?Ou(G(e),0,n).join(""):e.slice(0,n)}function ei(t,r,e,u){function i(){for(var r=-1,c=arguments.length,a=-1,l=u.length,s=il(l+c),h=this&&this!==re&&this instanceof i?f:t;++a<l;)s[a]=u[a];for(;c--;)s[a++]=arguments[++r];
return n(h,o?e:this,s)}var o=r&_n,f=Gu(t);return i}function ui(n){return function(t,r,e){return e&&"number"!=typeof e&&Ui(t,r,e)&&(r=e=X),t=Ac(t),r===X?(r=t,t=0):r=Ac(r),e=e===X?t<r?1:-1:Ac(e),ru(t,r,e,n)}}function ii(n){return function(t,r){return"string"==typeof t&&"string"==typeof r||(t=Ic(t),r=Ic(r)),n(t,r)}}function oi(n,t,r,e,u,i,o,f,c,a){var l=t&yn,s=l?o:X,h=l?X:o,p=l?i:X,_=l?X:i;t|=l?bn:wn,t&=~(l?wn:bn),t&gn||(t&=~(_n|vn));var v=[n,t,u,p,s,_,h,f,c,a],g=r.apply(X,v);return $i(n)&&Ss(g,v),g.placeholder=e,
Yi(g,n,t)}function fi(n){var t=al[n];return function(n,r){if(n=Ic(n),r=null==r?0:Hl(kc(r),292),r&&Zl(n)){var e=(Ec(n)+"e").split("e");return e=(Ec(t(e[0]+"e"+(+e[1]+r)))+"e").split("e"),+(e[0]+"e"+(+e[1]-r))}return t(n)}}function ci(n){return function(t){var r=zs(t);return r==Gn?M(t):r==tt?q(t):I(t,n(t))}}function ai(n,t,r,e,u,i,o,f){var c=t&vn;if(!c&&"function"!=typeof n)throw new pl(en);var a=e?e.length:0;if(a||(t&=~(bn|wn),e=u=X),o=o===X?o:Gl(kc(o),0),f=f===X?f:kc(f),a-=u?u.length:0,t&wn){var l=e,s=u;
e=u=X}var h=c?X:Os(n),p=[n,t,r,e,u,l,s,i,o,f];if(h&&qi(p,h),n=p[0],t=p[1],r=p[2],e=p[3],u=p[4],f=p[9]=p[9]===X?c?0:n.length:Gl(p[9]-a,0),!f&&t&(yn|dn)&&(t&=~(yn|dn)),t&&t!=_n)_=t==yn||t==dn?Hu(n,t,f):t!=bn&&t!=(_n|bn)||u.length?Qu.apply(X,p):ei(n,t,r,e);else var _=Zu(n,t,r);return Yi((h?ms:Ss)(_,p),n,t)}function li(n,t,r,e){return n===X||Gf(n,gl[r])&&!bl.call(e,r)?t:n}function si(n,t,r,e,u,i){return fc(n)&&fc(t)&&(i.set(t,n),Ke(n,t,X,si,i),i.delete(t)),n}function hi(n){return gc(n)?X:n}function pi(n,t,r,e,u,i){
var o=r&hn,f=n.length,c=t.length;if(f!=c&&!(o&&c>f))return!1;var a=i.get(n),l=i.get(t);if(a&&l)return a==t&&l==n;var s=-1,p=!0,_=r&pn?new yr:X;for(i.set(n,t),i.set(t,n);++s<f;){var v=n[s],g=t[s];if(e)var y=o?e(g,v,s,t,n,i):e(v,g,s,n,t,i);if(y!==X){if(y)continue;p=!1;break}if(_){if(!h(t,function(n,t){if(!S(_,t)&&(v===n||u(v,n,r,e,i)))return _.push(t)})){p=!1;break}}else if(v!==g&&!u(v,g,r,e,i)){p=!1;break}}return i.delete(n),i.delete(t),p}function _i(n,t,r,e,u,i,o){switch(r){case ct:if(n.byteLength!=t.byteLength||n.byteOffset!=t.byteOffset)return!1;
n=n.buffer,t=t.buffer;case ft:return!(n.byteLength!=t.byteLength||!i(new Rl(n),new Rl(t)));case Nn:case Pn:case Hn:return Gf(+n,+t);case Zn:return n.name==t.name&&n.message==t.message;case nt:case rt:return n==t+"";case Gn:var f=M;case tt:var c=e&hn;if(f||(f=P),n.size!=t.size&&!c)return!1;var a=o.get(n);if(a)return a==t;e|=pn,o.set(n,t);var l=pi(f(n),f(t),e,u,i,o);return o.delete(n),l;case et:if(_s)return _s.call(n)==_s.call(t)}return!1}function vi(n,t,r,e,u,i){var o=r&hn,f=yi(n),c=f.length;if(c!=yi(t).length&&!o)return!1;
for(var a=c;a--;){var l=f[a];if(!(o?l in t:bl.call(t,l)))return!1}var s=i.get(n),h=i.get(t);if(s&&h)return s==t&&h==n;var p=!0;i.set(n,t),i.set(t,n);for(var _=o;++a<c;){l=f[a];var v=n[l],g=t[l];if(e)var y=o?e(g,v,l,t,n,i):e(v,g,l,n,t,i);if(!(y===X?v===g||u(v,g,r,e,i):y)){p=!1;break}_||(_="constructor"==l)}if(p&&!_){var d=n.constructor,b=t.constructor;d!=b&&"constructor"in n&&"constructor"in t&&!("function"==typeof d&&d instanceof d&&"function"==typeof b&&b instanceof b)&&(p=!1)}return i.delete(n),
i.delete(t),p}function gi(n){return Ls(Vi(n,X,_o),n+"")}function yi(n){return de(n,Pc,Is)}function di(n){return de(n,qc,Rs)}function bi(n){for(var t=n.name+"",r=fs[t],e=bl.call(fs,t)?r.length:0;e--;){var u=r[e],i=u.func;if(null==i||i==n)return u.name}return t}function wi(n){return(bl.call(Z,"placeholder")?Z:n).placeholder}function mi(){var n=Z.iteratee||Ca;return n=n===Ca?De:n,arguments.length?n(arguments[0],arguments[1]):n}function xi(n,t){var r=n.__data__;return Ti(t)?r["string"==typeof t?"string":"hash"]:r.map;
}function ji(n){for(var t=Pc(n),r=t.length;r--;){var e=t[r],u=n[e];t[r]=[e,u,Fi(u)]}return t}function Ai(n,t){var r=B(n,t);return Ue(r)?r:X}function ki(n){var t=bl.call(n,Bl),r=n[Bl];try{n[Bl]=X;var e=!0}catch(n){}var u=xl.call(n);return e&&(t?n[Bl]=r:delete n[Bl]),u}function Oi(n,t,r){for(var e=-1,u=r.length;++e<u;){var i=r[e],o=i.size;switch(i.type){case"drop":n+=o;break;case"dropRight":t-=o;break;case"take":t=Hl(t,n+o);break;case"takeRight":n=Gl(n,t-o)}}return{start:n,end:t}}function Ii(n){var t=n.match(Bt);
return t?t[1].split(Tt):[]}function Ri(n,t,r){t=ku(t,n);for(var e=-1,u=t.length,i=!1;++e<u;){var o=no(t[e]);if(!(i=null!=n&&r(n,o)))break;n=n[o]}return i||++e!=u?i:(u=null==n?0:n.length,!!u&&oc(u)&&Ci(o,u)&&(bh(n)||dh(n)))}function zi(n){var t=n.length,r=new n.constructor(t);return t&&"string"==typeof n[0]&&bl.call(n,"index")&&(r.index=n.index,r.input=n.input),r}function Ei(n){return"function"!=typeof n.constructor||Mi(n)?{}:gs(El(n))}function Si(n,t,r){var e=n.constructor;switch(t){case ft:return Ru(n);
case Nn:case Pn:return new e(+n);case ct:return zu(n,r);case at:case lt:case st:case ht:case pt:case _t:case vt:case gt:case yt:return Wu(n,r);case Gn:return new e;case Hn:case rt:return new e(n);case nt:return Eu(n);case tt:return new e;case et:return Su(n)}}function Wi(n,t){var r=t.length;if(!r)return n;var e=r-1;return t[e]=(r>1?"& ":"")+t[e],t=t.join(r>2?", ":" "),n.replace(Ut,"{\n/* [wrapped with "+t+"] */\n")}function Li(n){return bh(n)||dh(n)||!!(Cl&&n&&n[Cl])}function Ci(n,t){var r=typeof n;
return t=null==t?Wn:t,!!t&&("number"==r||"symbol"!=r&&Vt.test(n))&&n>-1&&n%1==0&&n<t}function Ui(n,t,r){if(!fc(r))return!1;var e=typeof t;return!!("number"==e?Hf(r)&&Ci(t,r.length):"string"==e&&t in r)&&Gf(r[t],n)}function Bi(n,t){if(bh(n))return!1;var r=typeof n;return!("number"!=r&&"symbol"!=r&&"boolean"!=r&&null!=n&&!bc(n))||(zt.test(n)||!Rt.test(n)||null!=t&&n in ll(t))}function Ti(n){var t=typeof n;return"string"==t||"number"==t||"symbol"==t||"boolean"==t?"__proto__"!==n:null===n}function $i(n){
var t=bi(n),r=Z[t];if("function"!=typeof r||!(t in Ct.prototype))return!1;if(n===r)return!0;var e=Os(r);return!!e&&n===e[0]}function Di(n){return!!ml&&ml in n}function Mi(n){var t=n&&n.constructor;return n===("function"==typeof t&&t.prototype||gl)}function Fi(n){return n===n&&!fc(n)}function Ni(n,t){return function(r){return null!=r&&(r[n]===t&&(t!==X||n in ll(r)))}}function Pi(n){var t=Cf(n,function(n){return r.size===fn&&r.clear(),n}),r=t.cache;return t}function qi(n,t){var r=n[1],e=t[1],u=r|e,i=u<(_n|vn|mn),o=e==mn&&r==yn||e==mn&&r==xn&&n[7].length<=t[8]||e==(mn|xn)&&t[7].length<=t[8]&&r==yn;
if(!i&&!o)return n;e&_n&&(n[2]=t[2],u|=r&_n?0:gn);var f=t[3];if(f){var c=n[3];n[3]=c?Uu(c,f,t[4]):f,n[4]=c?N(n[3],cn):t[4]}return f=t[5],f&&(c=n[5],n[5]=c?Bu(c,f,t[6]):f,n[6]=c?N(n[5],cn):t[6]),f=t[7],f&&(n[7]=f),e&mn&&(n[8]=null==n[8]?t[8]:Hl(n[8],t[8])),null==n[9]&&(n[9]=t[9]),n[0]=t[0],n[1]=u,n}function Zi(n){var t=[];if(null!=n)for(var r in ll(n))t.push(r);return t}function Ki(n){return xl.call(n)}function Vi(t,r,e){return r=Gl(r===X?t.length-1:r,0),function(){for(var u=arguments,i=-1,o=Gl(u.length-r,0),f=il(o);++i<o;)f[i]=u[r+i];
i=-1;for(var c=il(r+1);++i<r;)c[i]=u[i];return c[r]=e(f),n(t,this,c)}}function Gi(n,t){return t.length<2?n:_e(n,au(t,0,-1))}function Hi(n,t){for(var r=n.length,e=Hl(t.length,r),u=Tu(n);e--;){var i=t[e];n[e]=Ci(i,r)?u[i]:X}return n}function Ji(n,t){if(("constructor"!==t||"function"!=typeof n[t])&&"__proto__"!=t)return n[t]}function Yi(n,t,r){var e=t+"";return Ls(n,Wi(e,ro(Ii(e),r)))}function Qi(n){var t=0,r=0;return function(){var e=Jl(),u=In-(e-r);if(r=e,u>0){if(++t>=On)return arguments[0]}else t=0;
return n.apply(X,arguments)}}function Xi(n,t){var r=-1,e=n.length,u=e-1;for(t=t===X?e:t;++r<t;){var i=tu(r,u),o=n[i];n[i]=n[r],n[r]=o}return n.length=t,n}function no(n){if("string"==typeof n||bc(n))return n;var t=n+"";return"0"==t&&1/n==-Sn?"-0":t}function to(n){if(null!=n){try{return dl.call(n)}catch(n){}try{return n+""}catch(n){}}return""}function ro(n,t){return r($n,function(r){var e="_."+r[0];t&r[1]&&!o(n,e)&&n.push(e)}),n.sort()}function eo(n){if(n instanceof Ct)return n.clone();var t=new Y(n.__wrapped__,n.__chain__);
return t.__actions__=Tu(n.__actions__),t.__index__=n.__index__,t.__values__=n.__values__,t}function uo(n,t,r){t=(r?Ui(n,t,r):t===X)?1:Gl(kc(t),0);var e=null==n?0:n.length;if(!e||t<1)return[];for(var u=0,i=0,o=il(Fl(e/t));u<e;)o[i++]=au(n,u,u+=t);return o}function io(n){for(var t=-1,r=null==n?0:n.length,e=0,u=[];++t<r;){var i=n[t];i&&(u[e++]=i)}return u}function oo(){var n=arguments.length;if(!n)return[];for(var t=il(n-1),r=arguments[0],e=n;e--;)t[e-1]=arguments[e];return a(bh(r)?Tu(r):[r],ee(t,1));
}function fo(n,t,r){var e=null==n?0:n.length;return e?(t=r||t===X?1:kc(t),au(n,t<0?0:t,e)):[]}function co(n,t,r){var e=null==n?0:n.length;return e?(t=r||t===X?1:kc(t),t=e-t,au(n,0,t<0?0:t)):[]}function ao(n,t){return n&&n.length?bu(n,mi(t,3),!0,!0):[]}function lo(n,t){return n&&n.length?bu(n,mi(t,3),!0):[]}function so(n,t,r,e){var u=null==n?0:n.length;return u?(r&&"number"!=typeof r&&Ui(n,t,r)&&(r=0,e=u),ne(n,t,r,e)):[]}function ho(n,t,r){var e=null==n?0:n.length;if(!e)return-1;var u=null==r?0:kc(r);
return u<0&&(u=Gl(e+u,0)),g(n,mi(t,3),u)}function po(n,t,r){var e=null==n?0:n.length;if(!e)return-1;var u=e-1;return r!==X&&(u=kc(r),u=r<0?Gl(e+u,0):Hl(u,e-1)),g(n,mi(t,3),u,!0)}function _o(n){return(null==n?0:n.length)?ee(n,1):[]}function vo(n){return(null==n?0:n.length)?ee(n,Sn):[]}function go(n,t){return(null==n?0:n.length)?(t=t===X?1:kc(t),ee(n,t)):[]}function yo(n){for(var t=-1,r=null==n?0:n.length,e={};++t<r;){var u=n[t];e[u[0]]=u[1]}return e}function bo(n){return n&&n.length?n[0]:X}function wo(n,t,r){
var e=null==n?0:n.length;if(!e)return-1;var u=null==r?0:kc(r);return u<0&&(u=Gl(e+u,0)),y(n,t,u)}function mo(n){return(null==n?0:n.length)?au(n,0,-1):[]}function xo(n,t){return null==n?"":Kl.call(n,t)}function jo(n){var t=null==n?0:n.length;return t?n[t-1]:X}function Ao(n,t,r){var e=null==n?0:n.length;if(!e)return-1;var u=e;return r!==X&&(u=kc(r),u=u<0?Gl(e+u,0):Hl(u,e-1)),t===t?K(n,t,u):g(n,b,u,!0)}function ko(n,t){return n&&n.length?Ge(n,kc(t)):X}function Oo(n,t){return n&&n.length&&t&&t.length?Xe(n,t):n;
}function Io(n,t,r){return n&&n.length&&t&&t.length?Xe(n,t,mi(r,2)):n}function Ro(n,t,r){return n&&n.length&&t&&t.length?Xe(n,t,X,r):n}function zo(n,t){var r=[];if(!n||!n.length)return r;var e=-1,u=[],i=n.length;for(t=mi(t,3);++e<i;){var o=n[e];t(o,e,n)&&(r.push(o),u.push(e))}return nu(n,u),r}function Eo(n){return null==n?n:Xl.call(n)}function So(n,t,r){var e=null==n?0:n.length;return e?(r&&"number"!=typeof r&&Ui(n,t,r)?(t=0,r=e):(t=null==t?0:kc(t),r=r===X?e:kc(r)),au(n,t,r)):[]}function Wo(n,t){
return su(n,t)}function Lo(n,t,r){return hu(n,t,mi(r,2))}function Co(n,t){var r=null==n?0:n.length;if(r){var e=su(n,t);if(e<r&&Gf(n[e],t))return e}return-1}function Uo(n,t){return su(n,t,!0)}function Bo(n,t,r){return hu(n,t,mi(r,2),!0)}function To(n,t){if(null==n?0:n.length){var r=su(n,t,!0)-1;if(Gf(n[r],t))return r}return-1}function $o(n){return n&&n.length?pu(n):[]}function Do(n,t){return n&&n.length?pu(n,mi(t,2)):[]}function Mo(n){var t=null==n?0:n.length;return t?au(n,1,t):[]}function Fo(n,t,r){
return n&&n.length?(t=r||t===X?1:kc(t),au(n,0,t<0?0:t)):[]}function No(n,t,r){var e=null==n?0:n.length;return e?(t=r||t===X?1:kc(t),t=e-t,au(n,t<0?0:t,e)):[]}function Po(n,t){return n&&n.length?bu(n,mi(t,3),!1,!0):[]}function qo(n,t){return n&&n.length?bu(n,mi(t,3)):[]}function Zo(n){return n&&n.length?gu(n):[]}function Ko(n,t){return n&&n.length?gu(n,mi(t,2)):[]}function Vo(n,t){return t="function"==typeof t?t:X,n&&n.length?gu(n,X,t):[]}function Go(n){if(!n||!n.length)return[];var t=0;return n=i(n,function(n){
if(Jf(n))return t=Gl(n.length,t),!0}),O(t,function(t){return c(n,m(t))})}function Ho(t,r){if(!t||!t.length)return[];var e=Go(t);return null==r?e:c(e,function(t){return n(r,X,t)})}function Jo(n,t){return xu(n||[],t||[],Sr)}function Yo(n,t){return xu(n||[],t||[],fu)}function Qo(n){var t=Z(n);return t.__chain__=!0,t}function Xo(n,t){return t(n),n}function nf(n,t){return t(n)}function tf(){return Qo(this)}function rf(){return new Y(this.value(),this.__chain__)}function ef(){this.__values__===X&&(this.__values__=jc(this.value()));
var n=this.__index__>=this.__values__.length;return{done:n,value:n?X:this.__values__[this.__index__++]}}function uf(){return this}function of(n){for(var t,r=this;r instanceof J;){var e=eo(r);e.__index__=0,e.__values__=X,t?u.__wrapped__=e:t=e;var u=e;r=r.__wrapped__}return u.__wrapped__=n,t}function ff(){var n=this.__wrapped__;if(n instanceof Ct){var t=n;return this.__actions__.length&&(t=new Ct(this)),t=t.reverse(),t.__actions__.push({func:nf,args:[Eo],thisArg:X}),new Y(t,this.__chain__)}return this.thru(Eo);
}function cf(){return wu(this.__wrapped__,this.__actions__)}function af(n,t,r){var e=bh(n)?u:Jr;return r&&Ui(n,t,r)&&(t=X),e(n,mi(t,3))}function lf(n,t){return(bh(n)?i:te)(n,mi(t,3))}function sf(n,t){return ee(yf(n,t),1)}function hf(n,t){return ee(yf(n,t),Sn)}function pf(n,t,r){return r=r===X?1:kc(r),ee(yf(n,t),r)}function _f(n,t){return(bh(n)?r:ys)(n,mi(t,3))}function vf(n,t){return(bh(n)?e:ds)(n,mi(t,3))}function gf(n,t,r,e){n=Hf(n)?n:ra(n),r=r&&!e?kc(r):0;var u=n.length;return r<0&&(r=Gl(u+r,0)),
dc(n)?r<=u&&n.indexOf(t,r)>-1:!!u&&y(n,t,r)>-1}function yf(n,t){return(bh(n)?c:Pe)(n,mi(t,3))}function df(n,t,r,e){return null==n?[]:(bh(t)||(t=null==t?[]:[t]),r=e?X:r,bh(r)||(r=null==r?[]:[r]),He(n,t,r))}function bf(n,t,r){var e=bh(n)?l:j,u=arguments.length<3;return e(n,mi(t,4),r,u,ys)}function wf(n,t,r){var e=bh(n)?s:j,u=arguments.length<3;return e(n,mi(t,4),r,u,ds)}function mf(n,t){return(bh(n)?i:te)(n,Uf(mi(t,3)))}function xf(n){return(bh(n)?Ir:iu)(n)}function jf(n,t,r){return t=(r?Ui(n,t,r):t===X)?1:kc(t),
(bh(n)?Rr:ou)(n,t)}function Af(n){return(bh(n)?zr:cu)(n)}function kf(n){if(null==n)return 0;if(Hf(n))return dc(n)?V(n):n.length;var t=zs(n);return t==Gn||t==tt?n.size:Me(n).length}function Of(n,t,r){var e=bh(n)?h:lu;return r&&Ui(n,t,r)&&(t=X),e(n,mi(t,3))}function If(n,t){if("function"!=typeof t)throw new pl(en);return n=kc(n),function(){if(--n<1)return t.apply(this,arguments)}}function Rf(n,t,r){return t=r?X:t,t=n&&null==t?n.length:t,ai(n,mn,X,X,X,X,t)}function zf(n,t){var r;if("function"!=typeof t)throw new pl(en);
return n=kc(n),function(){return--n>0&&(r=t.apply(this,arguments)),n<=1&&(t=X),r}}function Ef(n,t,r){t=r?X:t;var e=ai(n,yn,X,X,X,X,X,t);return e.placeholder=Ef.placeholder,e}function Sf(n,t,r){t=r?X:t;var e=ai(n,dn,X,X,X,X,X,t);return e.placeholder=Sf.placeholder,e}function Wf(n,t,r){function e(t){var r=h,e=p;return h=p=X,d=t,v=n.apply(e,r)}function u(n){return d=n,g=Ws(f,t),b?e(n):v}function i(n){var r=n-y,e=n-d,u=t-r;return w?Hl(u,_-e):u}function o(n){var r=n-y,e=n-d;return y===X||r>=t||r<0||w&&e>=_;
}function f(){var n=fh();return o(n)?c(n):(g=Ws(f,i(n)),X)}function c(n){return g=X,m&&h?e(n):(h=p=X,v)}function a(){g!==X&&As(g),d=0,h=y=p=g=X}function l(){return g===X?v:c(fh())}function s(){var n=fh(),r=o(n);if(h=arguments,p=this,y=n,r){if(g===X)return u(y);if(w)return As(g),g=Ws(f,t),e(y)}return g===X&&(g=Ws(f,t)),v}var h,p,_,v,g,y,d=0,b=!1,w=!1,m=!0;if("function"!=typeof n)throw new pl(en);return t=Ic(t)||0,fc(r)&&(b=!!r.leading,w="maxWait"in r,_=w?Gl(Ic(r.maxWait)||0,t):_,m="trailing"in r?!!r.trailing:m),
s.cancel=a,s.flush=l,s}function Lf(n){return ai(n,jn)}function Cf(n,t){if("function"!=typeof n||null!=t&&"function"!=typeof t)throw new pl(en);var r=function(){var e=arguments,u=t?t.apply(this,e):e[0],i=r.cache;if(i.has(u))return i.get(u);var o=n.apply(this,e);return r.cache=i.set(u,o)||i,o};return r.cache=new(Cf.Cache||sr),r}function Uf(n){if("function"!=typeof n)throw new pl(en);return function(){var t=arguments;switch(t.length){case 0:return!n.call(this);case 1:return!n.call(this,t[0]);case 2:
return!n.call(this,t[0],t[1]);case 3:return!n.call(this,t[0],t[1],t[2])}return!n.apply(this,t)}}function Bf(n){return zf(2,n)}function Tf(n,t){if("function"!=typeof n)throw new pl(en);return t=t===X?t:kc(t),uu(n,t)}function $f(t,r){if("function"!=typeof t)throw new pl(en);return r=null==r?0:Gl(kc(r),0),uu(function(e){var u=e[r],i=Ou(e,0,r);return u&&a(i,u),n(t,this,i)})}function Df(n,t,r){var e=!0,u=!0;if("function"!=typeof n)throw new pl(en);return fc(r)&&(e="leading"in r?!!r.leading:e,u="trailing"in r?!!r.trailing:u),
Wf(n,t,{leading:e,maxWait:t,trailing:u})}function Mf(n){return Rf(n,1)}function Ff(n,t){return ph(Au(t),n)}function Nf(){if(!arguments.length)return[];var n=arguments[0];return bh(n)?n:[n]}function Pf(n){return Fr(n,sn)}function qf(n,t){return t="function"==typeof t?t:X,Fr(n,sn,t)}function Zf(n){return Fr(n,an|sn)}function Kf(n,t){return t="function"==typeof t?t:X,Fr(n,an|sn,t)}function Vf(n,t){return null==t||Pr(n,t,Pc(t))}function Gf(n,t){return n===t||n!==n&&t!==t}function Hf(n){return null!=n&&oc(n.length)&&!uc(n);
}function Jf(n){return cc(n)&&Hf(n)}function Yf(n){return n===!0||n===!1||cc(n)&&we(n)==Nn}function Qf(n){return cc(n)&&1===n.nodeType&&!gc(n)}function Xf(n){if(null==n)return!0;if(Hf(n)&&(bh(n)||"string"==typeof n||"function"==typeof n.splice||mh(n)||Oh(n)||dh(n)))return!n.length;var t=zs(n);if(t==Gn||t==tt)return!n.size;if(Mi(n))return!Me(n).length;for(var r in n)if(bl.call(n,r))return!1;return!0}function nc(n,t){return Se(n,t)}function tc(n,t,r){r="function"==typeof r?r:X;var e=r?r(n,t):X;return e===X?Se(n,t,X,r):!!e;
}function rc(n){if(!cc(n))return!1;var t=we(n);return t==Zn||t==qn||"string"==typeof n.message&&"string"==typeof n.name&&!gc(n)}function ec(n){return"number"==typeof n&&Zl(n)}function uc(n){if(!fc(n))return!1;var t=we(n);return t==Kn||t==Vn||t==Fn||t==Xn}function ic(n){return"number"==typeof n&&n==kc(n)}function oc(n){return"number"==typeof n&&n>-1&&n%1==0&&n<=Wn}function fc(n){var t=typeof n;return null!=n&&("object"==t||"function"==t)}function cc(n){return null!=n&&"object"==typeof n}function ac(n,t){
return n===t||Ce(n,t,ji(t))}function lc(n,t,r){return r="function"==typeof r?r:X,Ce(n,t,ji(t),r)}function sc(n){return vc(n)&&n!=+n}function hc(n){if(Es(n))throw new fl(rn);return Ue(n)}function pc(n){return null===n}function _c(n){return null==n}function vc(n){return"number"==typeof n||cc(n)&&we(n)==Hn}function gc(n){if(!cc(n)||we(n)!=Yn)return!1;var t=El(n);if(null===t)return!0;var r=bl.call(t,"constructor")&&t.constructor;return"function"==typeof r&&r instanceof r&&dl.call(r)==jl}function yc(n){
return ic(n)&&n>=-Wn&&n<=Wn}function dc(n){return"string"==typeof n||!bh(n)&&cc(n)&&we(n)==rt}function bc(n){return"symbol"==typeof n||cc(n)&&we(n)==et}function wc(n){return n===X}function mc(n){return cc(n)&&zs(n)==it}function xc(n){return cc(n)&&we(n)==ot}function jc(n){if(!n)return[];if(Hf(n))return dc(n)?G(n):Tu(n);if(Ul&&n[Ul])return D(n[Ul]());var t=zs(n);return(t==Gn?M:t==tt?P:ra)(n)}function Ac(n){if(!n)return 0===n?n:0;if(n=Ic(n),n===Sn||n===-Sn){return(n<0?-1:1)*Ln}return n===n?n:0}function kc(n){
var t=Ac(n),r=t%1;return t===t?r?t-r:t:0}function Oc(n){return n?Mr(kc(n),0,Un):0}function Ic(n){if("number"==typeof n)return n;if(bc(n))return Cn;if(fc(n)){var t="function"==typeof n.valueOf?n.valueOf():n;n=fc(t)?t+"":t}if("string"!=typeof n)return 0===n?n:+n;n=R(n);var r=qt.test(n);return r||Kt.test(n)?Xr(n.slice(2),r?2:8):Pt.test(n)?Cn:+n}function Rc(n){return $u(n,qc(n))}function zc(n){return n?Mr(kc(n),-Wn,Wn):0===n?n:0}function Ec(n){return null==n?"":vu(n)}function Sc(n,t){var r=gs(n);return null==t?r:Cr(r,t);
}function Wc(n,t){return v(n,mi(t,3),ue)}function Lc(n,t){return v(n,mi(t,3),oe)}function Cc(n,t){return null==n?n:bs(n,mi(t,3),qc)}function Uc(n,t){return null==n?n:ws(n,mi(t,3),qc)}function Bc(n,t){return n&&ue(n,mi(t,3))}function Tc(n,t){return n&&oe(n,mi(t,3))}function $c(n){return null==n?[]:fe(n,Pc(n))}function Dc(n){return null==n?[]:fe(n,qc(n))}function Mc(n,t,r){var e=null==n?X:_e(n,t);return e===X?r:e}function Fc(n,t){return null!=n&&Ri(n,t,xe)}function Nc(n,t){return null!=n&&Ri(n,t,je);
}function Pc(n){return Hf(n)?Or(n):Me(n)}function qc(n){return Hf(n)?Or(n,!0):Fe(n)}function Zc(n,t){var r={};return t=mi(t,3),ue(n,function(n,e,u){Br(r,t(n,e,u),n)}),r}function Kc(n,t){var r={};return t=mi(t,3),ue(n,function(n,e,u){Br(r,e,t(n,e,u))}),r}function Vc(n,t){return Gc(n,Uf(mi(t)))}function Gc(n,t){if(null==n)return{};var r=c(di(n),function(n){return[n]});return t=mi(t),Ye(n,r,function(n,r){return t(n,r[0])})}function Hc(n,t,r){t=ku(t,n);var e=-1,u=t.length;for(u||(u=1,n=X);++e<u;){var i=null==n?X:n[no(t[e])];
i===X&&(e=u,i=r),n=uc(i)?i.call(n):i}return n}function Jc(n,t,r){return null==n?n:fu(n,t,r)}function Yc(n,t,r,e){return e="function"==typeof e?e:X,null==n?n:fu(n,t,r,e)}function Qc(n,t,e){var u=bh(n),i=u||mh(n)||Oh(n);if(t=mi(t,4),null==e){var o=n&&n.constructor;e=i?u?new o:[]:fc(n)&&uc(o)?gs(El(n)):{}}return(i?r:ue)(n,function(n,r,u){return t(e,n,r,u)}),e}function Xc(n,t){return null==n||yu(n,t)}function na(n,t,r){return null==n?n:du(n,t,Au(r))}function ta(n,t,r,e){return e="function"==typeof e?e:X,
null==n?n:du(n,t,Au(r),e)}function ra(n){return null==n?[]:E(n,Pc(n))}function ea(n){return null==n?[]:E(n,qc(n))}function ua(n,t,r){return r===X&&(r=t,t=X),r!==X&&(r=Ic(r),r=r===r?r:0),t!==X&&(t=Ic(t),t=t===t?t:0),Mr(Ic(n),t,r)}function ia(n,t,r){return t=Ac(t),r===X?(r=t,t=0):r=Ac(r),n=Ic(n),Ae(n,t,r)}function oa(n,t,r){if(r&&"boolean"!=typeof r&&Ui(n,t,r)&&(t=r=X),r===X&&("boolean"==typeof t?(r=t,t=X):"boolean"==typeof n&&(r=n,n=X)),n===X&&t===X?(n=0,t=1):(n=Ac(n),t===X?(t=n,n=0):t=Ac(t)),n>t){
var e=n;n=t,t=e}if(r||n%1||t%1){var u=Ql();return Hl(n+u*(t-n+Qr("1e-"+((u+"").length-1))),t)}return tu(n,t)}function fa(n){return Qh(Ec(n).toLowerCase())}function ca(n){return n=Ec(n),n&&n.replace(Gt,ve).replace(Dr,"")}function aa(n,t,r){n=Ec(n),t=vu(t);var e=n.length;r=r===X?e:Mr(kc(r),0,e);var u=r;return r-=t.length,r>=0&&n.slice(r,u)==t}function la(n){return n=Ec(n),n&&At.test(n)?n.replace(xt,ge):n}function sa(n){return n=Ec(n),n&&Wt.test(n)?n.replace(St,"\\$&"):n}function ha(n,t,r){n=Ec(n),t=kc(t);
var e=t?V(n):0;if(!t||e>=t)return n;var u=(t-e)/2;return ri(Nl(u),r)+n+ri(Fl(u),r)}function pa(n,t,r){n=Ec(n),t=kc(t);var e=t?V(n):0;return t&&e<t?n+ri(t-e,r):n}function _a(n,t,r){n=Ec(n),t=kc(t);var e=t?V(n):0;return t&&e<t?ri(t-e,r)+n:n}function va(n,t,r){return r||null==t?t=0:t&&(t=+t),Yl(Ec(n).replace(Lt,""),t||0)}function ga(n,t,r){return t=(r?Ui(n,t,r):t===X)?1:kc(t),eu(Ec(n),t)}function ya(){var n=arguments,t=Ec(n[0]);return n.length<3?t:t.replace(n[1],n[2])}function da(n,t,r){return r&&"number"!=typeof r&&Ui(n,t,r)&&(t=r=X),
(r=r===X?Un:r>>>0)?(n=Ec(n),n&&("string"==typeof t||null!=t&&!Ah(t))&&(t=vu(t),!t&&T(n))?Ou(G(n),0,r):n.split(t,r)):[]}function ba(n,t,r){return n=Ec(n),r=null==r?0:Mr(kc(r),0,n.length),t=vu(t),n.slice(r,r+t.length)==t}function wa(n,t,r){var e=Z.templateSettings;r&&Ui(n,t,r)&&(t=X),n=Ec(n),t=Sh({},t,e,li);var u,i,o=Sh({},t.imports,e.imports,li),f=Pc(o),c=E(o,f),a=0,l=t.interpolate||Ht,s="__p += '",h=sl((t.escape||Ht).source+"|"+l.source+"|"+(l===It?Ft:Ht).source+"|"+(t.evaluate||Ht).source+"|$","g"),p="//# sourceURL="+(bl.call(t,"sourceURL")?(t.sourceURL+"").replace(/\s/g," "):"lodash.templateSources["+ ++Zr+"]")+"\n";
n.replace(h,function(t,r,e,o,f,c){return e||(e=o),s+=n.slice(a,c).replace(Jt,U),r&&(u=!0,s+="' +\n__e("+r+") +\n'"),f&&(i=!0,s+="';\n"+f+";\n__p += '"),e&&(s+="' +\n((__t = ("+e+")) == null ? '' : __t) +\n'"),a=c+t.length,t}),s+="';\n";var _=bl.call(t,"variable")&&t.variable;if(_){if(Dt.test(_))throw new fl(un)}else s="with (obj) {\n"+s+"\n}\n";s=(i?s.replace(dt,""):s).replace(bt,"$1").replace(wt,"$1;"),s="function("+(_||"obj")+") {\n"+(_?"":"obj || (obj = {});\n")+"var __t, __p = ''"+(u?", __e = _.escape":"")+(i?", __j = Array.prototype.join;\nfunction print() { __p += __j.call(arguments, '') }\n":";\n")+s+"return __p\n}";
var v=Xh(function(){return cl(f,p+"return "+s).apply(X,c)});if(v.source=s,rc(v))throw v;return v}function ma(n){return Ec(n).toLowerCase()}function xa(n){return Ec(n).toUpperCase()}function ja(n,t,r){if(n=Ec(n),n&&(r||t===X))return R(n);if(!n||!(t=vu(t)))return n;var e=G(n),u=G(t);return Ou(e,W(e,u),L(e,u)+1).join("")}function Aa(n,t,r){if(n=Ec(n),n&&(r||t===X))return n.slice(0,H(n)+1);if(!n||!(t=vu(t)))return n;var e=G(n);return Ou(e,0,L(e,G(t))+1).join("")}function ka(n,t,r){if(n=Ec(n),n&&(r||t===X))return n.replace(Lt,"");
if(!n||!(t=vu(t)))return n;var e=G(n);return Ou(e,W(e,G(t))).join("")}function Oa(n,t){var r=An,e=kn;if(fc(t)){var u="separator"in t?t.separator:u;r="length"in t?kc(t.length):r,e="omission"in t?vu(t.omission):e}n=Ec(n);var i=n.length;if(T(n)){var o=G(n);i=o.length}if(r>=i)return n;var f=r-V(e);if(f<1)return e;var c=o?Ou(o,0,f).join(""):n.slice(0,f);if(u===X)return c+e;if(o&&(f+=c.length-f),Ah(u)){if(n.slice(f).search(u)){var a,l=c;for(u.global||(u=sl(u.source,Ec(Nt.exec(u))+"g")),u.lastIndex=0;a=u.exec(l);)var s=a.index;
c=c.slice(0,s===X?f:s)}}else if(n.indexOf(vu(u),f)!=f){var h=c.lastIndexOf(u);h>-1&&(c=c.slice(0,h))}return c+e}function Ia(n){return n=Ec(n),n&&jt.test(n)?n.replace(mt,ye):n}function Ra(n,t,r){return n=Ec(n),t=r?X:t,t===X?$(n)?Q(n):_(n):n.match(t)||[]}function za(t){var r=null==t?0:t.length,e=mi();return t=r?c(t,function(n){if("function"!=typeof n[1])throw new pl(en);return[e(n[0]),n[1]]}):[],uu(function(e){for(var u=-1;++u<r;){var i=t[u];if(n(i[0],this,e))return n(i[1],this,e)}})}function Ea(n){
return Nr(Fr(n,an))}function Sa(n){return function(){return n}}function Wa(n,t){return null==n||n!==n?t:n}function La(n){return n}function Ca(n){return De("function"==typeof n?n:Fr(n,an))}function Ua(n){return qe(Fr(n,an))}function Ba(n,t){return Ze(n,Fr(t,an))}function Ta(n,t,e){var u=Pc(t),i=fe(t,u);null!=e||fc(t)&&(i.length||!u.length)||(e=t,t=n,n=this,i=fe(t,Pc(t)));var o=!(fc(e)&&"chain"in e&&!e.chain),f=uc(n);return r(i,function(r){var e=t[r];n[r]=e,f&&(n.prototype[r]=function(){var t=this.__chain__;
if(o||t){var r=n(this.__wrapped__);return(r.__actions__=Tu(this.__actions__)).push({func:e,args:arguments,thisArg:n}),r.__chain__=t,r}return e.apply(n,a([this.value()],arguments))})}),n}function $a(){return re._===this&&(re._=Al),this}function Da(){}function Ma(n){return n=kc(n),uu(function(t){return Ge(t,n)})}function Fa(n){return Bi(n)?m(no(n)):Qe(n)}function Na(n){return function(t){return null==n?X:_e(n,t)}}function Pa(){return[]}function qa(){return!1}function Za(){return{}}function Ka(){return"";
}function Va(){return!0}function Ga(n,t){if(n=kc(n),n<1||n>Wn)return[];var r=Un,e=Hl(n,Un);t=mi(t),n-=Un;for(var u=O(e,t);++r<n;)t(r);return u}function Ha(n){return bh(n)?c(n,no):bc(n)?[n]:Tu(Cs(Ec(n)))}function Ja(n){var t=++wl;return Ec(n)+t}function Ya(n){return n&&n.length?Yr(n,La,me):X}function Qa(n,t){return n&&n.length?Yr(n,mi(t,2),me):X}function Xa(n){return w(n,La)}function nl(n,t){return w(n,mi(t,2))}function tl(n){return n&&n.length?Yr(n,La,Ne):X}function rl(n,t){return n&&n.length?Yr(n,mi(t,2),Ne):X;
}function el(n){return n&&n.length?k(n,La):0}function ul(n,t){return n&&n.length?k(n,mi(t,2)):0}x=null==x?re:be.defaults(re.Object(),x,be.pick(re,qr));var il=x.Array,ol=x.Date,fl=x.Error,cl=x.Function,al=x.Math,ll=x.Object,sl=x.RegExp,hl=x.String,pl=x.TypeError,_l=il.prototype,vl=cl.prototype,gl=ll.prototype,yl=x["__core-js_shared__"],dl=vl.toString,bl=gl.hasOwnProperty,wl=0,ml=function(){var n=/[^.]+$/.exec(yl&&yl.keys&&yl.keys.IE_PROTO||"");return n?"Symbol(src)_1."+n:""}(),xl=gl.toString,jl=dl.call(ll),Al=re._,kl=sl("^"+dl.call(bl).replace(St,"\\$&").replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g,"$1.*?")+"$"),Ol=ie?x.Buffer:X,Il=x.Symbol,Rl=x.Uint8Array,zl=Ol?Ol.allocUnsafe:X,El=F(ll.getPrototypeOf,ll),Sl=ll.create,Wl=gl.propertyIsEnumerable,Ll=_l.splice,Cl=Il?Il.isConcatSpreadable:X,Ul=Il?Il.iterator:X,Bl=Il?Il.toStringTag:X,Tl=function(){
try{var n=Ai(ll,"defineProperty");return n({},"",{}),n}catch(n){}}(),$l=x.clearTimeout!==re.clearTimeout&&x.clearTimeout,Dl=ol&&ol.now!==re.Date.now&&ol.now,Ml=x.setTimeout!==re.setTimeout&&x.setTimeout,Fl=al.ceil,Nl=al.floor,Pl=ll.getOwnPropertySymbols,ql=Ol?Ol.isBuffer:X,Zl=x.isFinite,Kl=_l.join,Vl=F(ll.keys,ll),Gl=al.max,Hl=al.min,Jl=ol.now,Yl=x.parseInt,Ql=al.random,Xl=_l.reverse,ns=Ai(x,"DataView"),ts=Ai(x,"Map"),rs=Ai(x,"Promise"),es=Ai(x,"Set"),us=Ai(x,"WeakMap"),is=Ai(ll,"create"),os=us&&new us,fs={},cs=to(ns),as=to(ts),ls=to(rs),ss=to(es),hs=to(us),ps=Il?Il.prototype:X,_s=ps?ps.valueOf:X,vs=ps?ps.toString:X,gs=function(){
function n(){}return function(t){if(!fc(t))return{};if(Sl)return Sl(t);n.prototype=t;var r=new n;return n.prototype=X,r}}();Z.templateSettings={escape:kt,evaluate:Ot,interpolate:It,variable:"",imports:{_:Z}},Z.prototype=J.prototype,Z.prototype.constructor=Z,Y.prototype=gs(J.prototype),Y.prototype.constructor=Y,Ct.prototype=gs(J.prototype),Ct.prototype.constructor=Ct,Xt.prototype.clear=nr,Xt.prototype.delete=tr,Xt.prototype.get=rr,Xt.prototype.has=er,Xt.prototype.set=ur,ir.prototype.clear=or,ir.prototype.delete=fr,
ir.prototype.get=cr,ir.prototype.has=ar,ir.prototype.set=lr,sr.prototype.clear=hr,sr.prototype.delete=pr,sr.prototype.get=_r,sr.prototype.has=vr,sr.prototype.set=gr,yr.prototype.add=yr.prototype.push=dr,yr.prototype.has=br,wr.prototype.clear=mr,wr.prototype.delete=xr,wr.prototype.get=jr,wr.prototype.has=Ar,wr.prototype.set=kr;var ys=Pu(ue),ds=Pu(oe,!0),bs=qu(),ws=qu(!0),ms=os?function(n,t){return os.set(n,t),n}:La,xs=Tl?function(n,t){return Tl(n,"toString",{configurable:!0,enumerable:!1,value:Sa(t),
writable:!0})}:La,js=uu,As=$l||function(n){return re.clearTimeout(n)},ks=es&&1/P(new es([,-0]))[1]==Sn?function(n){return new es(n)}:Da,Os=os?function(n){return os.get(n)}:Da,Is=Pl?function(n){return null==n?[]:(n=ll(n),i(Pl(n),function(t){return Wl.call(n,t)}))}:Pa,Rs=Pl?function(n){for(var t=[];n;)a(t,Is(n)),n=El(n);return t}:Pa,zs=we;(ns&&zs(new ns(new ArrayBuffer(1)))!=ct||ts&&zs(new ts)!=Gn||rs&&zs(rs.resolve())!=Qn||es&&zs(new es)!=tt||us&&zs(new us)!=it)&&(zs=function(n){var t=we(n),r=t==Yn?n.constructor:X,e=r?to(r):"";
if(e)switch(e){case cs:return ct;case as:return Gn;case ls:return Qn;case ss:return tt;case hs:return it}return t});var Es=yl?uc:qa,Ss=Qi(ms),Ws=Ml||function(n,t){return re.setTimeout(n,t)},Ls=Qi(xs),Cs=Pi(function(n){var t=[];return 46===n.charCodeAt(0)&&t.push(""),n.replace(Et,function(n,r,e,u){t.push(e?u.replace(Mt,"$1"):r||n)}),t}),Us=uu(function(n,t){return Jf(n)?Hr(n,ee(t,1,Jf,!0)):[]}),Bs=uu(function(n,t){var r=jo(t);return Jf(r)&&(r=X),Jf(n)?Hr(n,ee(t,1,Jf,!0),mi(r,2)):[]}),Ts=uu(function(n,t){
var r=jo(t);return Jf(r)&&(r=X),Jf(n)?Hr(n,ee(t,1,Jf,!0),X,r):[]}),$s=uu(function(n){var t=c(n,ju);return t.length&&t[0]===n[0]?ke(t):[]}),Ds=uu(function(n){var t=jo(n),r=c(n,ju);return t===jo(r)?t=X:r.pop(),r.length&&r[0]===n[0]?ke(r,mi(t,2)):[]}),Ms=uu(function(n){var t=jo(n),r=c(n,ju);return t="function"==typeof t?t:X,t&&r.pop(),r.length&&r[0]===n[0]?ke(r,X,t):[]}),Fs=uu(Oo),Ns=gi(function(n,t){var r=null==n?0:n.length,e=Tr(n,t);return nu(n,c(t,function(n){return Ci(n,r)?+n:n}).sort(Lu)),e}),Ps=uu(function(n){
return gu(ee(n,1,Jf,!0))}),qs=uu(function(n){var t=jo(n);return Jf(t)&&(t=X),gu(ee(n,1,Jf,!0),mi(t,2))}),Zs=uu(function(n){var t=jo(n);return t="function"==typeof t?t:X,gu(ee(n,1,Jf,!0),X,t)}),Ks=uu(function(n,t){return Jf(n)?Hr(n,t):[]}),Vs=uu(function(n){return mu(i(n,Jf))}),Gs=uu(function(n){var t=jo(n);return Jf(t)&&(t=X),mu(i(n,Jf),mi(t,2))}),Hs=uu(function(n){var t=jo(n);return t="function"==typeof t?t:X,mu(i(n,Jf),X,t)}),Js=uu(Go),Ys=uu(function(n){var t=n.length,r=t>1?n[t-1]:X;return r="function"==typeof r?(n.pop(),
r):X,Ho(n,r)}),Qs=gi(function(n){var t=n.length,r=t?n[0]:0,e=this.__wrapped__,u=function(t){return Tr(t,n)};return!(t>1||this.__actions__.length)&&e instanceof Ct&&Ci(r)?(e=e.slice(r,+r+(t?1:0)),e.__actions__.push({func:nf,args:[u],thisArg:X}),new Y(e,this.__chain__).thru(function(n){return t&&!n.length&&n.push(X),n})):this.thru(u)}),Xs=Fu(function(n,t,r){bl.call(n,r)?++n[r]:Br(n,r,1)}),nh=Ju(ho),th=Ju(po),rh=Fu(function(n,t,r){bl.call(n,r)?n[r].push(t):Br(n,r,[t])}),eh=uu(function(t,r,e){var u=-1,i="function"==typeof r,o=Hf(t)?il(t.length):[];
return ys(t,function(t){o[++u]=i?n(r,t,e):Ie(t,r,e)}),o}),uh=Fu(function(n,t,r){Br(n,r,t)}),ih=Fu(function(n,t,r){n[r?0:1].push(t)},function(){return[[],[]]}),oh=uu(function(n,t){if(null==n)return[];var r=t.length;return r>1&&Ui(n,t[0],t[1])?t=[]:r>2&&Ui(t[0],t[1],t[2])&&(t=[t[0]]),He(n,ee(t,1),[])}),fh=Dl||function(){return re.Date.now()},ch=uu(function(n,t,r){var e=_n;if(r.length){var u=N(r,wi(ch));e|=bn}return ai(n,e,t,r,u)}),ah=uu(function(n,t,r){var e=_n|vn;if(r.length){var u=N(r,wi(ah));e|=bn;
}return ai(t,e,n,r,u)}),lh=uu(function(n,t){return Gr(n,1,t)}),sh=uu(function(n,t,r){return Gr(n,Ic(t)||0,r)});Cf.Cache=sr;var hh=js(function(t,r){r=1==r.length&&bh(r[0])?c(r[0],z(mi())):c(ee(r,1),z(mi()));var e=r.length;return uu(function(u){for(var i=-1,o=Hl(u.length,e);++i<o;)u[i]=r[i].call(this,u[i]);return n(t,this,u)})}),ph=uu(function(n,t){return ai(n,bn,X,t,N(t,wi(ph)))}),_h=uu(function(n,t){return ai(n,wn,X,t,N(t,wi(_h)))}),vh=gi(function(n,t){return ai(n,xn,X,X,X,t)}),gh=ii(me),yh=ii(function(n,t){
return n>=t}),dh=Re(function(){return arguments}())?Re:function(n){return cc(n)&&bl.call(n,"callee")&&!Wl.call(n,"callee")},bh=il.isArray,wh=ce?z(ce):ze,mh=ql||qa,xh=ae?z(ae):Ee,jh=le?z(le):Le,Ah=se?z(se):Be,kh=he?z(he):Te,Oh=pe?z(pe):$e,Ih=ii(Ne),Rh=ii(function(n,t){return n<=t}),zh=Nu(function(n,t){if(Mi(t)||Hf(t))return $u(t,Pc(t),n),X;for(var r in t)bl.call(t,r)&&Sr(n,r,t[r])}),Eh=Nu(function(n,t){$u(t,qc(t),n)}),Sh=Nu(function(n,t,r,e){$u(t,qc(t),n,e)}),Wh=Nu(function(n,t,r,e){$u(t,Pc(t),n,e);
}),Lh=gi(Tr),Ch=uu(function(n,t){n=ll(n);var r=-1,e=t.length,u=e>2?t[2]:X;for(u&&Ui(t[0],t[1],u)&&(e=1);++r<e;)for(var i=t[r],o=qc(i),f=-1,c=o.length;++f<c;){var a=o[f],l=n[a];(l===X||Gf(l,gl[a])&&!bl.call(n,a))&&(n[a]=i[a])}return n}),Uh=uu(function(t){return t.push(X,si),n(Mh,X,t)}),Bh=Xu(function(n,t,r){null!=t&&"function"!=typeof t.toString&&(t=xl.call(t)),n[t]=r},Sa(La)),Th=Xu(function(n,t,r){null!=t&&"function"!=typeof t.toString&&(t=xl.call(t)),bl.call(n,t)?n[t].push(r):n[t]=[r]},mi),$h=uu(Ie),Dh=Nu(function(n,t,r){
Ke(n,t,r)}),Mh=Nu(function(n,t,r,e){Ke(n,t,r,e)}),Fh=gi(function(n,t){var r={};if(null==n)return r;var e=!1;t=c(t,function(t){return t=ku(t,n),e||(e=t.length>1),t}),$u(n,di(n),r),e&&(r=Fr(r,an|ln|sn,hi));for(var u=t.length;u--;)yu(r,t[u]);return r}),Nh=gi(function(n,t){return null==n?{}:Je(n,t)}),Ph=ci(Pc),qh=ci(qc),Zh=Vu(function(n,t,r){return t=t.toLowerCase(),n+(r?fa(t):t)}),Kh=Vu(function(n,t,r){return n+(r?"-":"")+t.toLowerCase()}),Vh=Vu(function(n,t,r){return n+(r?" ":"")+t.toLowerCase()}),Gh=Ku("toLowerCase"),Hh=Vu(function(n,t,r){
return n+(r?"_":"")+t.toLowerCase()}),Jh=Vu(function(n,t,r){return n+(r?" ":"")+Qh(t)}),Yh=Vu(function(n,t,r){return n+(r?" ":"")+t.toUpperCase()}),Qh=Ku("toUpperCase"),Xh=uu(function(t,r){try{return n(t,X,r)}catch(n){return rc(n)?n:new fl(n)}}),np=gi(function(n,t){return r(t,function(t){t=no(t),Br(n,t,ch(n[t],n))}),n}),tp=Yu(),rp=Yu(!0),ep=uu(function(n,t){return function(r){return Ie(r,n,t)}}),up=uu(function(n,t){return function(r){return Ie(n,r,t)}}),ip=ti(c),op=ti(u),fp=ti(h),cp=ui(),ap=ui(!0),lp=ni(function(n,t){
return n+t},0),sp=fi("ceil"),hp=ni(function(n,t){return n/t},1),pp=fi("floor"),_p=ni(function(n,t){return n*t},1),vp=fi("round"),gp=ni(function(n,t){return n-t},0);return Z.after=If,Z.ary=Rf,Z.assign=zh,Z.assignIn=Eh,Z.assignInWith=Sh,Z.assignWith=Wh,Z.at=Lh,Z.before=zf,Z.bind=ch,Z.bindAll=np,Z.bindKey=ah,Z.castArray=Nf,Z.chain=Qo,Z.chunk=uo,Z.compact=io,Z.concat=oo,Z.cond=za,Z.conforms=Ea,Z.constant=Sa,Z.countBy=Xs,Z.create=Sc,Z.curry=Ef,Z.curryRight=Sf,Z.debounce=Wf,Z.defaults=Ch,Z.defaultsDeep=Uh,
Z.defer=lh,Z.delay=sh,Z.difference=Us,Z.differenceBy=Bs,Z.differenceWith=Ts,Z.drop=fo,Z.dropRight=co,Z.dropRightWhile=ao,Z.dropWhile=lo,Z.fill=so,Z.filter=lf,Z.flatMap=sf,Z.flatMapDeep=hf,Z.flatMapDepth=pf,Z.flatten=_o,Z.flattenDeep=vo,Z.flattenDepth=go,Z.flip=Lf,Z.flow=tp,Z.flowRight=rp,Z.fromPairs=yo,Z.functions=$c,Z.functionsIn=Dc,Z.groupBy=rh,Z.initial=mo,Z.intersection=$s,Z.intersectionBy=Ds,Z.intersectionWith=Ms,Z.invert=Bh,Z.invertBy=Th,Z.invokeMap=eh,Z.iteratee=Ca,Z.keyBy=uh,Z.keys=Pc,Z.keysIn=qc,
Z.map=yf,Z.mapKeys=Zc,Z.mapValues=Kc,Z.matches=Ua,Z.matchesProperty=Ba,Z.memoize=Cf,Z.merge=Dh,Z.mergeWith=Mh,Z.method=ep,Z.methodOf=up,Z.mixin=Ta,Z.negate=Uf,Z.nthArg=Ma,Z.omit=Fh,Z.omitBy=Vc,Z.once=Bf,Z.orderBy=df,Z.over=ip,Z.overArgs=hh,Z.overEvery=op,Z.overSome=fp,Z.partial=ph,Z.partialRight=_h,Z.partition=ih,Z.pick=Nh,Z.pickBy=Gc,Z.property=Fa,Z.propertyOf=Na,Z.pull=Fs,Z.pullAll=Oo,Z.pullAllBy=Io,Z.pullAllWith=Ro,Z.pullAt=Ns,Z.range=cp,Z.rangeRight=ap,Z.rearg=vh,Z.reject=mf,Z.remove=zo,Z.rest=Tf,
Z.reverse=Eo,Z.sampleSize=jf,Z.set=Jc,Z.setWith=Yc,Z.shuffle=Af,Z.slice=So,Z.sortBy=oh,Z.sortedUniq=$o,Z.sortedUniqBy=Do,Z.split=da,Z.spread=$f,Z.tail=Mo,Z.take=Fo,Z.takeRight=No,Z.takeRightWhile=Po,Z.takeWhile=qo,Z.tap=Xo,Z.throttle=Df,Z.thru=nf,Z.toArray=jc,Z.toPairs=Ph,Z.toPairsIn=qh,Z.toPath=Ha,Z.toPlainObject=Rc,Z.transform=Qc,Z.unary=Mf,Z.union=Ps,Z.unionBy=qs,Z.unionWith=Zs,Z.uniq=Zo,Z.uniqBy=Ko,Z.uniqWith=Vo,Z.unset=Xc,Z.unzip=Go,Z.unzipWith=Ho,Z.update=na,Z.updateWith=ta,Z.values=ra,Z.valuesIn=ea,
Z.without=Ks,Z.words=Ra,Z.wrap=Ff,Z.xor=Vs,Z.xorBy=Gs,Z.xorWith=Hs,Z.zip=Js,Z.zipObject=Jo,Z.zipObjectDeep=Yo,Z.zipWith=Ys,Z.entries=Ph,Z.entriesIn=qh,Z.extend=Eh,Z.extendWith=Sh,Ta(Z,Z),Z.add=lp,Z.attempt=Xh,Z.camelCase=Zh,Z.capitalize=fa,Z.ceil=sp,Z.clamp=ua,Z.clone=Pf,Z.cloneDeep=Zf,Z.cloneDeepWith=Kf,Z.cloneWith=qf,Z.conformsTo=Vf,Z.deburr=ca,Z.defaultTo=Wa,Z.divide=hp,Z.endsWith=aa,Z.eq=Gf,Z.escape=la,Z.escapeRegExp=sa,Z.every=af,Z.find=nh,Z.findIndex=ho,Z.findKey=Wc,Z.findLast=th,Z.findLastIndex=po,
Z.findLastKey=Lc,Z.floor=pp,Z.forEach=_f,Z.forEachRight=vf,Z.forIn=Cc,Z.forInRight=Uc,Z.forOwn=Bc,Z.forOwnRight=Tc,Z.get=Mc,Z.gt=gh,Z.gte=yh,Z.has=Fc,Z.hasIn=Nc,Z.head=bo,Z.identity=La,Z.includes=gf,Z.indexOf=wo,Z.inRange=ia,Z.invoke=$h,Z.isArguments=dh,Z.isArray=bh,Z.isArrayBuffer=wh,Z.isArrayLike=Hf,Z.isArrayLikeObject=Jf,Z.isBoolean=Yf,Z.isBuffer=mh,Z.isDate=xh,Z.isElement=Qf,Z.isEmpty=Xf,Z.isEqual=nc,Z.isEqualWith=tc,Z.isError=rc,Z.isFinite=ec,Z.isFunction=uc,Z.isInteger=ic,Z.isLength=oc,Z.isMap=jh,
Z.isMatch=ac,Z.isMatchWith=lc,Z.isNaN=sc,Z.isNative=hc,Z.isNil=_c,Z.isNull=pc,Z.isNumber=vc,Z.isObject=fc,Z.isObjectLike=cc,Z.isPlainObject=gc,Z.isRegExp=Ah,Z.isSafeInteger=yc,Z.isSet=kh,Z.isString=dc,Z.isSymbol=bc,Z.isTypedArray=Oh,Z.isUndefined=wc,Z.isWeakMap=mc,Z.isWeakSet=xc,Z.join=xo,Z.kebabCase=Kh,Z.last=jo,Z.lastIndexOf=Ao,Z.lowerCase=Vh,Z.lowerFirst=Gh,Z.lt=Ih,Z.lte=Rh,Z.max=Ya,Z.maxBy=Qa,Z.mean=Xa,Z.meanBy=nl,Z.min=tl,Z.minBy=rl,Z.stubArray=Pa,Z.stubFalse=qa,Z.stubObject=Za,Z.stubString=Ka,
Z.stubTrue=Va,Z.multiply=_p,Z.nth=ko,Z.noConflict=$a,Z.noop=Da,Z.now=fh,Z.pad=ha,Z.padEnd=pa,Z.padStart=_a,Z.parseInt=va,Z.random=oa,Z.reduce=bf,Z.reduceRight=wf,Z.repeat=ga,Z.replace=ya,Z.result=Hc,Z.round=vp,Z.runInContext=p,Z.sample=xf,Z.size=kf,Z.snakeCase=Hh,Z.some=Of,Z.sortedIndex=Wo,Z.sortedIndexBy=Lo,Z.sortedIndexOf=Co,Z.sortedLastIndex=Uo,Z.sortedLastIndexBy=Bo,Z.sortedLastIndexOf=To,Z.startCase=Jh,Z.startsWith=ba,Z.subtract=gp,Z.sum=el,Z.sumBy=ul,Z.template=wa,Z.times=Ga,Z.toFinite=Ac,Z.toInteger=kc,
Z.toLength=Oc,Z.toLower=ma,Z.toNumber=Ic,Z.toSafeInteger=zc,Z.toString=Ec,Z.toUpper=xa,Z.trim=ja,Z.trimEnd=Aa,Z.trimStart=ka,Z.truncate=Oa,Z.unescape=Ia,Z.uniqueId=Ja,Z.upperCase=Yh,Z.upperFirst=Qh,Z.each=_f,Z.eachRight=vf,Z.first=bo,Ta(Z,function(){var n={};return ue(Z,function(t,r){bl.call(Z.prototype,r)||(n[r]=t)}),n}(),{chain:!1}),Z.VERSION=nn,r(["bind","bindKey","curry","curryRight","partial","partialRight"],function(n){Z[n].placeholder=Z}),r(["drop","take"],function(n,t){Ct.prototype[n]=function(r){
r=r===X?1:Gl(kc(r),0);var e=this.__filtered__&&!t?new Ct(this):this.clone();return e.__filtered__?e.__takeCount__=Hl(r,e.__takeCount__):e.__views__.push({size:Hl(r,Un),type:n+(e.__dir__<0?"Right":"")}),e},Ct.prototype[n+"Right"]=function(t){return this.reverse()[n](t).reverse()}}),r(["filter","map","takeWhile"],function(n,t){var r=t+1,e=r==Rn||r==En;Ct.prototype[n]=function(n){var t=this.clone();return t.__iteratees__.push({iteratee:mi(n,3),type:r}),t.__filtered__=t.__filtered__||e,t}}),r(["head","last"],function(n,t){
var r="take"+(t?"Right":"");Ct.prototype[n]=function(){return this[r](1).value()[0]}}),r(["initial","tail"],function(n,t){var r="drop"+(t?"":"Right");Ct.prototype[n]=function(){return this.__filtered__?new Ct(this):this[r](1)}}),Ct.prototype.compact=function(){return this.filter(La)},Ct.prototype.find=function(n){return this.filter(n).head()},Ct.prototype.findLast=function(n){return this.reverse().find(n)},Ct.prototype.invokeMap=uu(function(n,t){return"function"==typeof n?new Ct(this):this.map(function(r){
return Ie(r,n,t)})}),Ct.prototype.reject=function(n){return this.filter(Uf(mi(n)))},Ct.prototype.slice=function(n,t){n=kc(n);var r=this;return r.__filtered__&&(n>0||t<0)?new Ct(r):(n<0?r=r.takeRight(-n):n&&(r=r.drop(n)),t!==X&&(t=kc(t),r=t<0?r.dropRight(-t):r.take(t-n)),r)},Ct.prototype.takeRightWhile=function(n){return this.reverse().takeWhile(n).reverse()},Ct.prototype.toArray=function(){return this.take(Un)},ue(Ct.prototype,function(n,t){var r=/^(?:filter|find|map|reject)|While$/.test(t),e=/^(?:head|last)$/.test(t),u=Z[e?"take"+("last"==t?"Right":""):t],i=e||/^find/.test(t);
u&&(Z.prototype[t]=function(){var t=this.__wrapped__,o=e?[1]:arguments,f=t instanceof Ct,c=o[0],l=f||bh(t),s=function(n){var t=u.apply(Z,a([n],o));return e&&h?t[0]:t};l&&r&&"function"==typeof c&&1!=c.length&&(f=l=!1);var h=this.__chain__,p=!!this.__actions__.length,_=i&&!h,v=f&&!p;if(!i&&l){t=v?t:new Ct(this);var g=n.apply(t,o);return g.__actions__.push({func:nf,args:[s],thisArg:X}),new Y(g,h)}return _&&v?n.apply(this,o):(g=this.thru(s),_?e?g.value()[0]:g.value():g)})}),r(["pop","push","shift","sort","splice","unshift"],function(n){
var t=_l[n],r=/^(?:push|sort|unshift)$/.test(n)?"tap":"thru",e=/^(?:pop|shift)$/.test(n);Z.prototype[n]=function(){var n=arguments;if(e&&!this.__chain__){var u=this.value();return t.apply(bh(u)?u:[],n)}return this[r](function(r){return t.apply(bh(r)?r:[],n)})}}),ue(Ct.prototype,function(n,t){var r=Z[t];if(r){var e=r.name+"";bl.call(fs,e)||(fs[e]=[]),fs[e].push({name:t,func:r})}}),fs[Qu(X,vn).name]=[{name:"wrapper",func:X}],Ct.prototype.clone=$t,Ct.prototype.reverse=Yt,Ct.prototype.value=Qt,Z.prototype.at=Qs,
Z.prototype.chain=tf,Z.prototype.commit=rf,Z.prototype.next=ef,Z.prototype.plant=of,Z.prototype.reverse=ff,Z.prototype.toJSON=Z.prototype.valueOf=Z.prototype.value=cf,Z.prototype.first=Z.prototype.head,Ul&&(Z.prototype[Ul]=uf),Z},be=de(); true?(re._=be,!(__WEBPACK_AMD_DEFINE_RESULT__ = (function(){return be}).call(exports, __webpack_require__, exports, module),
		__WEBPACK_AMD_DEFINE_RESULT__ !== undefined && (module.exports = __WEBPACK_AMD_DEFINE_RESULT__))):0}).call(this);

/***/ }),

/***/ "./node_modules/semver/semver.js":
/*!***************************************!*\
  !*** ./node_modules/semver/semver.js ***!
  \***************************************/
/***/ ((module, exports) => {

exports = module.exports = SemVer

var debug
/* istanbul ignore next */
if (typeof process === 'object' &&
    process.env &&
    process.env.NODE_DEBUG &&
    /\bsemver\b/i.test(process.env.NODE_DEBUG)) {
  debug = function () {
    var args = Array.prototype.slice.call(arguments, 0)
    args.unshift('SEMVER')
    console.log.apply(console, args)
  }
} else {
  debug = function () {}
}

// Note: this is the semver.org version of the spec that it implements
// Not necessarily the package version of this code.
exports.SEMVER_SPEC_VERSION = '2.0.0'

var MAX_LENGTH = 256
var MAX_SAFE_INTEGER = Number.MAX_SAFE_INTEGER ||
  /* istanbul ignore next */ 9007199254740991

// Max safe segment length for coercion.
var MAX_SAFE_COMPONENT_LENGTH = 16

// The actual regexps go on exports.re
var re = exports.re = []
var src = exports.src = []
var R = 0

// The following Regular Expressions can be used for tokenizing,
// validating, and parsing SemVer version strings.

// ## Numeric Identifier
// A single `0`, or a non-zero digit followed by zero or more digits.

var NUMERICIDENTIFIER = R++
src[NUMERICIDENTIFIER] = '0|[1-9]\\d*'
var NUMERICIDENTIFIERLOOSE = R++
src[NUMERICIDENTIFIERLOOSE] = '[0-9]+'

// ## Non-numeric Identifier
// Zero or more digits, followed by a letter or hyphen, and then zero or
// more letters, digits, or hyphens.

var NONNUMERICIDENTIFIER = R++
src[NONNUMERICIDENTIFIER] = '\\d*[a-zA-Z-][a-zA-Z0-9-]*'

// ## Main Version
// Three dot-separated numeric identifiers.

var MAINVERSION = R++
src[MAINVERSION] = '(' + src[NUMERICIDENTIFIER] + ')\\.' +
                   '(' + src[NUMERICIDENTIFIER] + ')\\.' +
                   '(' + src[NUMERICIDENTIFIER] + ')'

var MAINVERSIONLOOSE = R++
src[MAINVERSIONLOOSE] = '(' + src[NUMERICIDENTIFIERLOOSE] + ')\\.' +
                        '(' + src[NUMERICIDENTIFIERLOOSE] + ')\\.' +
                        '(' + src[NUMERICIDENTIFIERLOOSE] + ')'

// ## Pre-release Version Identifier
// A numeric identifier, or a non-numeric identifier.

var PRERELEASEIDENTIFIER = R++
src[PRERELEASEIDENTIFIER] = '(?:' + src[NUMERICIDENTIFIER] +
                            '|' + src[NONNUMERICIDENTIFIER] + ')'

var PRERELEASEIDENTIFIERLOOSE = R++
src[PRERELEASEIDENTIFIERLOOSE] = '(?:' + src[NUMERICIDENTIFIERLOOSE] +
                                 '|' + src[NONNUMERICIDENTIFIER] + ')'

// ## Pre-release Version
// Hyphen, followed by one or more dot-separated pre-release version
// identifiers.

var PRERELEASE = R++
src[PRERELEASE] = '(?:-(' + src[PRERELEASEIDENTIFIER] +
                  '(?:\\.' + src[PRERELEASEIDENTIFIER] + ')*))'

var PRERELEASELOOSE = R++
src[PRERELEASELOOSE] = '(?:-?(' + src[PRERELEASEIDENTIFIERLOOSE] +
                       '(?:\\.' + src[PRERELEASEIDENTIFIERLOOSE] + ')*))'

// ## Build Metadata Identifier
// Any combination of digits, letters, or hyphens.

var BUILDIDENTIFIER = R++
src[BUILDIDENTIFIER] = '[0-9A-Za-z-]+'

// ## Build Metadata
// Plus sign, followed by one or more period-separated build metadata
// identifiers.

var BUILD = R++
src[BUILD] = '(?:\\+(' + src[BUILDIDENTIFIER] +
             '(?:\\.' + src[BUILDIDENTIFIER] + ')*))'

// ## Full Version String
// A main version, followed optionally by a pre-release version and
// build metadata.

// Note that the only major, minor, patch, and pre-release sections of
// the version string are capturing groups.  The build metadata is not a
// capturing group, because it should not ever be used in version
// comparison.

var FULL = R++
var FULLPLAIN = 'v?' + src[MAINVERSION] +
                src[PRERELEASE] + '?' +
                src[BUILD] + '?'

src[FULL] = '^' + FULLPLAIN + '$'

// like full, but allows v1.2.3 and =1.2.3, which people do sometimes.
// also, 1.0.0alpha1 (prerelease without the hyphen) which is pretty
// common in the npm registry.
var LOOSEPLAIN = '[v=\\s]*' + src[MAINVERSIONLOOSE] +
                 src[PRERELEASELOOSE] + '?' +
                 src[BUILD] + '?'

var LOOSE = R++
src[LOOSE] = '^' + LOOSEPLAIN + '$'

var GTLT = R++
src[GTLT] = '((?:<|>)?=?)'

// Something like "2.*" or "1.2.x".
// Note that "x.x" is a valid xRange identifer, meaning "any version"
// Only the first item is strictly required.
var XRANGEIDENTIFIERLOOSE = R++
src[XRANGEIDENTIFIERLOOSE] = src[NUMERICIDENTIFIERLOOSE] + '|x|X|\\*'
var XRANGEIDENTIFIER = R++
src[XRANGEIDENTIFIER] = src[NUMERICIDENTIFIER] + '|x|X|\\*'

var XRANGEPLAIN = R++
src[XRANGEPLAIN] = '[v=\\s]*(' + src[XRANGEIDENTIFIER] + ')' +
                   '(?:\\.(' + src[XRANGEIDENTIFIER] + ')' +
                   '(?:\\.(' + src[XRANGEIDENTIFIER] + ')' +
                   '(?:' + src[PRERELEASE] + ')?' +
                   src[BUILD] + '?' +
                   ')?)?'

var XRANGEPLAINLOOSE = R++
src[XRANGEPLAINLOOSE] = '[v=\\s]*(' + src[XRANGEIDENTIFIERLOOSE] + ')' +
                        '(?:\\.(' + src[XRANGEIDENTIFIERLOOSE] + ')' +
                        '(?:\\.(' + src[XRANGEIDENTIFIERLOOSE] + ')' +
                        '(?:' + src[PRERELEASELOOSE] + ')?' +
                        src[BUILD] + '?' +
                        ')?)?'

var XRANGE = R++
src[XRANGE] = '^' + src[GTLT] + '\\s*' + src[XRANGEPLAIN] + '$'
var XRANGELOOSE = R++
src[XRANGELOOSE] = '^' + src[GTLT] + '\\s*' + src[XRANGEPLAINLOOSE] + '$'

// Coercion.
// Extract anything that could conceivably be a part of a valid semver
var COERCE = R++
src[COERCE] = '(?:^|[^\\d])' +
              '(\\d{1,' + MAX_SAFE_COMPONENT_LENGTH + '})' +
              '(?:\\.(\\d{1,' + MAX_SAFE_COMPONENT_LENGTH + '}))?' +
              '(?:\\.(\\d{1,' + MAX_SAFE_COMPONENT_LENGTH + '}))?' +
              '(?:$|[^\\d])'

// Tilde ranges.
// Meaning is "reasonably at or greater than"
var LONETILDE = R++
src[LONETILDE] = '(?:~>?)'

var TILDETRIM = R++
src[TILDETRIM] = '(\\s*)' + src[LONETILDE] + '\\s+'
re[TILDETRIM] = new RegExp(src[TILDETRIM], 'g')
var tildeTrimReplace = '$1~'

var TILDE = R++
src[TILDE] = '^' + src[LONETILDE] + src[XRANGEPLAIN] + '$'
var TILDELOOSE = R++
src[TILDELOOSE] = '^' + src[LONETILDE] + src[XRANGEPLAINLOOSE] + '$'

// Caret ranges.
// Meaning is "at least and backwards compatible with"
var LONECARET = R++
src[LONECARET] = '(?:\\^)'

var CARETTRIM = R++
src[CARETTRIM] = '(\\s*)' + src[LONECARET] + '\\s+'
re[CARETTRIM] = new RegExp(src[CARETTRIM], 'g')
var caretTrimReplace = '$1^'

var CARET = R++
src[CARET] = '^' + src[LONECARET] + src[XRANGEPLAIN] + '$'
var CARETLOOSE = R++
src[CARETLOOSE] = '^' + src[LONECARET] + src[XRANGEPLAINLOOSE] + '$'

// A simple gt/lt/eq thing, or just "" to indicate "any version"
var COMPARATORLOOSE = R++
src[COMPARATORLOOSE] = '^' + src[GTLT] + '\\s*(' + LOOSEPLAIN + ')$|^$'
var COMPARATOR = R++
src[COMPARATOR] = '^' + src[GTLT] + '\\s*(' + FULLPLAIN + ')$|^$'

// An expression to strip any whitespace between the gtlt and the thing
// it modifies, so that `> 1.2.3` ==> `>1.2.3`
var COMPARATORTRIM = R++
src[COMPARATORTRIM] = '(\\s*)' + src[GTLT] +
                      '\\s*(' + LOOSEPLAIN + '|' + src[XRANGEPLAIN] + ')'

// this one has to use the /g flag
re[COMPARATORTRIM] = new RegExp(src[COMPARATORTRIM], 'g')
var comparatorTrimReplace = '$1$2$3'

// Something like `1.2.3 - 1.2.4`
// Note that these all use the loose form, because they'll be
// checked against either the strict or loose comparator form
// later.
var HYPHENRANGE = R++
src[HYPHENRANGE] = '^\\s*(' + src[XRANGEPLAIN] + ')' +
                   '\\s+-\\s+' +
                   '(' + src[XRANGEPLAIN] + ')' +
                   '\\s*$'

var HYPHENRANGELOOSE = R++
src[HYPHENRANGELOOSE] = '^\\s*(' + src[XRANGEPLAINLOOSE] + ')' +
                        '\\s+-\\s+' +
                        '(' + src[XRANGEPLAINLOOSE] + ')' +
                        '\\s*$'

// Star ranges basically just allow anything at all.
var STAR = R++
src[STAR] = '(<|>)?=?\\s*\\*'

// Compile to actual regexp objects.
// All are flag-free, unless they were created above with a flag.
for (var i = 0; i < R; i++) {
  debug(i, src[i])
  if (!re[i]) {
    re[i] = new RegExp(src[i])
  }
}

exports.parse = parse
function parse (version, options) {
  if (!options || typeof options !== 'object') {
    options = {
      loose: !!options,
      includePrerelease: false
    }
  }

  if (version instanceof SemVer) {
    return version
  }

  if (typeof version !== 'string') {
    return null
  }

  if (version.length > MAX_LENGTH) {
    return null
  }

  var r = options.loose ? re[LOOSE] : re[FULL]
  if (!r.test(version)) {
    return null
  }

  try {
    return new SemVer(version, options)
  } catch (er) {
    return null
  }
}

exports.valid = valid
function valid (version, options) {
  var v = parse(version, options)
  return v ? v.version : null
}

exports.clean = clean
function clean (version, options) {
  var s = parse(version.trim().replace(/^[=v]+/, ''), options)
  return s ? s.version : null
}

exports.SemVer = SemVer

function SemVer (version, options) {
  if (!options || typeof options !== 'object') {
    options = {
      loose: !!options,
      includePrerelease: false
    }
  }
  if (version instanceof SemVer) {
    if (version.loose === options.loose) {
      return version
    } else {
      version = version.version
    }
  } else if (typeof version !== 'string') {
    throw new TypeError('Invalid Version: ' + version)
  }

  if (version.length > MAX_LENGTH) {
    throw new TypeError('version is longer than ' + MAX_LENGTH + ' characters')
  }

  if (!(this instanceof SemVer)) {
    return new SemVer(version, options)
  }

  debug('SemVer', version, options)
  this.options = options
  this.loose = !!options.loose

  var m = version.trim().match(options.loose ? re[LOOSE] : re[FULL])

  if (!m) {
    throw new TypeError('Invalid Version: ' + version)
  }

  this.raw = version

  // these are actually numbers
  this.major = +m[1]
  this.minor = +m[2]
  this.patch = +m[3]

  if (this.major > MAX_SAFE_INTEGER || this.major < 0) {
    throw new TypeError('Invalid major version')
  }

  if (this.minor > MAX_SAFE_INTEGER || this.minor < 0) {
    throw new TypeError('Invalid minor version')
  }

  if (this.patch > MAX_SAFE_INTEGER || this.patch < 0) {
    throw new TypeError('Invalid patch version')
  }

  // numberify any prerelease numeric ids
  if (!m[4]) {
    this.prerelease = []
  } else {
    this.prerelease = m[4].split('.').map(function (id) {
      if (/^[0-9]+$/.test(id)) {
        var num = +id
        if (num >= 0 && num < MAX_SAFE_INTEGER) {
          return num
        }
      }
      return id
    })
  }

  this.build = m[5] ? m[5].split('.') : []
  this.format()
}

SemVer.prototype.format = function () {
  this.version = this.major + '.' + this.minor + '.' + this.patch
  if (this.prerelease.length) {
    this.version += '-' + this.prerelease.join('.')
  }
  return this.version
}

SemVer.prototype.toString = function () {
  return this.version
}

SemVer.prototype.compare = function (other) {
  debug('SemVer.compare', this.version, this.options, other)
  if (!(other instanceof SemVer)) {
    other = new SemVer(other, this.options)
  }

  return this.compareMain(other) || this.comparePre(other)
}

SemVer.prototype.compareMain = function (other) {
  if (!(other instanceof SemVer)) {
    other = new SemVer(other, this.options)
  }

  return compareIdentifiers(this.major, other.major) ||
         compareIdentifiers(this.minor, other.minor) ||
         compareIdentifiers(this.patch, other.patch)
}

SemVer.prototype.comparePre = function (other) {
  if (!(other instanceof SemVer)) {
    other = new SemVer(other, this.options)
  }

  // NOT having a prerelease is > having one
  if (this.prerelease.length && !other.prerelease.length) {
    return -1
  } else if (!this.prerelease.length && other.prerelease.length) {
    return 1
  } else if (!this.prerelease.length && !other.prerelease.length) {
    return 0
  }

  var i = 0
  do {
    var a = this.prerelease[i]
    var b = other.prerelease[i]
    debug('prerelease compare', i, a, b)
    if (a === undefined && b === undefined) {
      return 0
    } else if (b === undefined) {
      return 1
    } else if (a === undefined) {
      return -1
    } else if (a === b) {
      continue
    } else {
      return compareIdentifiers(a, b)
    }
  } while (++i)
}

// preminor will bump the version up to the next minor release, and immediately
// down to pre-release. premajor and prepatch work the same way.
SemVer.prototype.inc = function (release, identifier) {
  switch (release) {
    case 'premajor':
      this.prerelease.length = 0
      this.patch = 0
      this.minor = 0
      this.major++
      this.inc('pre', identifier)
      break
    case 'preminor':
      this.prerelease.length = 0
      this.patch = 0
      this.minor++
      this.inc('pre', identifier)
      break
    case 'prepatch':
      // If this is already a prerelease, it will bump to the next version
      // drop any prereleases that might already exist, since they are not
      // relevant at this point.
      this.prerelease.length = 0
      this.inc('patch', identifier)
      this.inc('pre', identifier)
      break
    // If the input is a non-prerelease version, this acts the same as
    // prepatch.
    case 'prerelease':
      if (this.prerelease.length === 0) {
        this.inc('patch', identifier)
      }
      this.inc('pre', identifier)
      break

    case 'major':
      // If this is a pre-major version, bump up to the same major version.
      // Otherwise increment major.
      // 1.0.0-5 bumps to 1.0.0
      // 1.1.0 bumps to 2.0.0
      if (this.minor !== 0 ||
          this.patch !== 0 ||
          this.prerelease.length === 0) {
        this.major++
      }
      this.minor = 0
      this.patch = 0
      this.prerelease = []
      break
    case 'minor':
      // If this is a pre-minor version, bump up to the same minor version.
      // Otherwise increment minor.
      // 1.2.0-5 bumps to 1.2.0
      // 1.2.1 bumps to 1.3.0
      if (this.patch !== 0 || this.prerelease.length === 0) {
        this.minor++
      }
      this.patch = 0
      this.prerelease = []
      break
    case 'patch':
      // If this is not a pre-release version, it will increment the patch.
      // If it is a pre-release it will bump up to the same patch version.
      // 1.2.0-5 patches to 1.2.0
      // 1.2.0 patches to 1.2.1
      if (this.prerelease.length === 0) {
        this.patch++
      }
      this.prerelease = []
      break
    // This probably shouldn't be used publicly.
    // 1.0.0 "pre" would become 1.0.0-0 which is the wrong direction.
    case 'pre':
      if (this.prerelease.length === 0) {
        this.prerelease = [0]
      } else {
        var i = this.prerelease.length
        while (--i >= 0) {
          if (typeof this.prerelease[i] === 'number') {
            this.prerelease[i]++
            i = -2
          }
        }
        if (i === -1) {
          // didn't increment anything
          this.prerelease.push(0)
        }
      }
      if (identifier) {
        // 1.2.0-beta.1 bumps to 1.2.0-beta.2,
        // 1.2.0-beta.fooblz or 1.2.0-beta bumps to 1.2.0-beta.0
        if (this.prerelease[0] === identifier) {
          if (isNaN(this.prerelease[1])) {
            this.prerelease = [identifier, 0]
          }
        } else {
          this.prerelease = [identifier, 0]
        }
      }
      break

    default:
      throw new Error('invalid increment argument: ' + release)
  }
  this.format()
  this.raw = this.version
  return this
}

exports.inc = inc
function inc (version, release, loose, identifier) {
  if (typeof (loose) === 'string') {
    identifier = loose
    loose = undefined
  }

  try {
    return new SemVer(version, loose).inc(release, identifier).version
  } catch (er) {
    return null
  }
}

exports.diff = diff
function diff (version1, version2) {
  if (eq(version1, version2)) {
    return null
  } else {
    var v1 = parse(version1)
    var v2 = parse(version2)
    var prefix = ''
    if (v1.prerelease.length || v2.prerelease.length) {
      prefix = 'pre'
      var defaultResult = 'prerelease'
    }
    for (var key in v1) {
      if (key === 'major' || key === 'minor' || key === 'patch') {
        if (v1[key] !== v2[key]) {
          return prefix + key
        }
      }
    }
    return defaultResult // may be undefined
  }
}

exports.compareIdentifiers = compareIdentifiers

var numeric = /^[0-9]+$/
function compareIdentifiers (a, b) {
  var anum = numeric.test(a)
  var bnum = numeric.test(b)

  if (anum && bnum) {
    a = +a
    b = +b
  }

  return a === b ? 0
    : (anum && !bnum) ? -1
    : (bnum && !anum) ? 1
    : a < b ? -1
    : 1
}

exports.rcompareIdentifiers = rcompareIdentifiers
function rcompareIdentifiers (a, b) {
  return compareIdentifiers(b, a)
}

exports.major = major
function major (a, loose) {
  return new SemVer(a, loose).major
}

exports.minor = minor
function minor (a, loose) {
  return new SemVer(a, loose).minor
}

exports.patch = patch
function patch (a, loose) {
  return new SemVer(a, loose).patch
}

exports.compare = compare
function compare (a, b, loose) {
  return new SemVer(a, loose).compare(new SemVer(b, loose))
}

exports.compareLoose = compareLoose
function compareLoose (a, b) {
  return compare(a, b, true)
}

exports.rcompare = rcompare
function rcompare (a, b, loose) {
  return compare(b, a, loose)
}

exports.sort = sort
function sort (list, loose) {
  return list.sort(function (a, b) {
    return exports.compare(a, b, loose)
  })
}

exports.rsort = rsort
function rsort (list, loose) {
  return list.sort(function (a, b) {
    return exports.rcompare(a, b, loose)
  })
}

exports.gt = gt
function gt (a, b, loose) {
  return compare(a, b, loose) > 0
}

exports.lt = lt
function lt (a, b, loose) {
  return compare(a, b, loose) < 0
}

exports.eq = eq
function eq (a, b, loose) {
  return compare(a, b, loose) === 0
}

exports.neq = neq
function neq (a, b, loose) {
  return compare(a, b, loose) !== 0
}

exports.gte = gte
function gte (a, b, loose) {
  return compare(a, b, loose) >= 0
}

exports.lte = lte
function lte (a, b, loose) {
  return compare(a, b, loose) <= 0
}

exports.cmp = cmp
function cmp (a, op, b, loose) {
  switch (op) {
    case '===':
      if (typeof a === 'object')
        a = a.version
      if (typeof b === 'object')
        b = b.version
      return a === b

    case '!==':
      if (typeof a === 'object')
        a = a.version
      if (typeof b === 'object')
        b = b.version
      return a !== b

    case '':
    case '=':
    case '==':
      return eq(a, b, loose)

    case '!=':
      return neq(a, b, loose)

    case '>':
      return gt(a, b, loose)

    case '>=':
      return gte(a, b, loose)

    case '<':
      return lt(a, b, loose)

    case '<=':
      return lte(a, b, loose)

    default:
      throw new TypeError('Invalid operator: ' + op)
  }
}

exports.Comparator = Comparator
function Comparator (comp, options) {
  if (!options || typeof options !== 'object') {
    options = {
      loose: !!options,
      includePrerelease: false
    }
  }

  if (comp instanceof Comparator) {
    if (comp.loose === !!options.loose) {
      return comp
    } else {
      comp = comp.value
    }
  }

  if (!(this instanceof Comparator)) {
    return new Comparator(comp, options)
  }

  debug('comparator', comp, options)
  this.options = options
  this.loose = !!options.loose
  this.parse(comp)

  if (this.semver === ANY) {
    this.value = ''
  } else {
    this.value = this.operator + this.semver.version
  }

  debug('comp', this)
}

var ANY = {}
Comparator.prototype.parse = function (comp) {
  var r = this.options.loose ? re[COMPARATORLOOSE] : re[COMPARATOR]
  var m = comp.match(r)

  if (!m) {
    throw new TypeError('Invalid comparator: ' + comp)
  }

  this.operator = m[1]
  if (this.operator === '=') {
    this.operator = ''
  }

  // if it literally is just '>' or '' then allow anything.
  if (!m[2]) {
    this.semver = ANY
  } else {
    this.semver = new SemVer(m[2], this.options.loose)
  }
}

Comparator.prototype.toString = function () {
  return this.value
}

Comparator.prototype.test = function (version) {
  debug('Comparator.test', version, this.options.loose)

  if (this.semver === ANY) {
    return true
  }

  if (typeof version === 'string') {
    version = new SemVer(version, this.options)
  }

  return cmp(version, this.operator, this.semver, this.options)
}

Comparator.prototype.intersects = function (comp, options) {
  if (!(comp instanceof Comparator)) {
    throw new TypeError('a Comparator is required')
  }

  if (!options || typeof options !== 'object') {
    options = {
      loose: !!options,
      includePrerelease: false
    }
  }

  var rangeTmp

  if (this.operator === '') {
    rangeTmp = new Range(comp.value, options)
    return satisfies(this.value, rangeTmp, options)
  } else if (comp.operator === '') {
    rangeTmp = new Range(this.value, options)
    return satisfies(comp.semver, rangeTmp, options)
  }

  var sameDirectionIncreasing =
    (this.operator === '>=' || this.operator === '>') &&
    (comp.operator === '>=' || comp.operator === '>')
  var sameDirectionDecreasing =
    (this.operator === '<=' || this.operator === '<') &&
    (comp.operator === '<=' || comp.operator === '<')
  var sameSemVer = this.semver.version === comp.semver.version
  var differentDirectionsInclusive =
    (this.operator === '>=' || this.operator === '<=') &&
    (comp.operator === '>=' || comp.operator === '<=')
  var oppositeDirectionsLessThan =
    cmp(this.semver, '<', comp.semver, options) &&
    ((this.operator === '>=' || this.operator === '>') &&
    (comp.operator === '<=' || comp.operator === '<'))
  var oppositeDirectionsGreaterThan =
    cmp(this.semver, '>', comp.semver, options) &&
    ((this.operator === '<=' || this.operator === '<') &&
    (comp.operator === '>=' || comp.operator === '>'))

  return sameDirectionIncreasing || sameDirectionDecreasing ||
    (sameSemVer && differentDirectionsInclusive) ||
    oppositeDirectionsLessThan || oppositeDirectionsGreaterThan
}

exports.Range = Range
function Range (range, options) {
  if (!options || typeof options !== 'object') {
    options = {
      loose: !!options,
      includePrerelease: false
    }
  }

  if (range instanceof Range) {
    if (range.loose === !!options.loose &&
        range.includePrerelease === !!options.includePrerelease) {
      return range
    } else {
      return new Range(range.raw, options)
    }
  }

  if (range instanceof Comparator) {
    return new Range(range.value, options)
  }

  if (!(this instanceof Range)) {
    return new Range(range, options)
  }

  this.options = options
  this.loose = !!options.loose
  this.includePrerelease = !!options.includePrerelease

  // First, split based on boolean or ||
  this.raw = range
  this.set = range.split(/\s*\|\|\s*/).map(function (range) {
    return this.parseRange(range.trim())
  }, this).filter(function (c) {
    // throw out any that are not relevant for whatever reason
    return c.length
  })

  if (!this.set.length) {
    throw new TypeError('Invalid SemVer Range: ' + range)
  }

  this.format()
}

Range.prototype.format = function () {
  this.range = this.set.map(function (comps) {
    return comps.join(' ').trim()
  }).join('||').trim()
  return this.range
}

Range.prototype.toString = function () {
  return this.range
}

Range.prototype.parseRange = function (range) {
  var loose = this.options.loose
  range = range.trim()
  // `1.2.3 - 1.2.4` => `>=1.2.3 <=1.2.4`
  var hr = loose ? re[HYPHENRANGELOOSE] : re[HYPHENRANGE]
  range = range.replace(hr, hyphenReplace)
  debug('hyphen replace', range)
  // `> 1.2.3 < 1.2.5` => `>1.2.3 <1.2.5`
  range = range.replace(re[COMPARATORTRIM], comparatorTrimReplace)
  debug('comparator trim', range, re[COMPARATORTRIM])

  // `~ 1.2.3` => `~1.2.3`
  range = range.replace(re[TILDETRIM], tildeTrimReplace)

  // `^ 1.2.3` => `^1.2.3`
  range = range.replace(re[CARETTRIM], caretTrimReplace)

  // normalize spaces
  range = range.split(/\s+/).join(' ')

  // At this point, the range is completely trimmed and
  // ready to be split into comparators.

  var compRe = loose ? re[COMPARATORLOOSE] : re[COMPARATOR]
  var set = range.split(' ').map(function (comp) {
    return parseComparator(comp, this.options)
  }, this).join(' ').split(/\s+/)
  if (this.options.loose) {
    // in loose mode, throw out any that are not valid comparators
    set = set.filter(function (comp) {
      return !!comp.match(compRe)
    })
  }
  set = set.map(function (comp) {
    return new Comparator(comp, this.options)
  }, this)

  return set
}

Range.prototype.intersects = function (range, options) {
  if (!(range instanceof Range)) {
    throw new TypeError('a Range is required')
  }

  return this.set.some(function (thisComparators) {
    return thisComparators.every(function (thisComparator) {
      return range.set.some(function (rangeComparators) {
        return rangeComparators.every(function (rangeComparator) {
          return thisComparator.intersects(rangeComparator, options)
        })
      })
    })
  })
}

// Mostly just for testing and legacy API reasons
exports.toComparators = toComparators
function toComparators (range, options) {
  return new Range(range, options).set.map(function (comp) {
    return comp.map(function (c) {
      return c.value
    }).join(' ').trim().split(' ')
  })
}

// comprised of xranges, tildes, stars, and gtlt's at this point.
// already replaced the hyphen ranges
// turn into a set of JUST comparators.
function parseComparator (comp, options) {
  debug('comp', comp, options)
  comp = replaceCarets(comp, options)
  debug('caret', comp)
  comp = replaceTildes(comp, options)
  debug('tildes', comp)
  comp = replaceXRanges(comp, options)
  debug('xrange', comp)
  comp = replaceStars(comp, options)
  debug('stars', comp)
  return comp
}

function isX (id) {
  return !id || id.toLowerCase() === 'x' || id === '*'
}

// ~, ~> --> * (any, kinda silly)
// ~2, ~2.x, ~2.x.x, ~>2, ~>2.x ~>2.x.x --> >=2.0.0 <3.0.0
// ~2.0, ~2.0.x, ~>2.0, ~>2.0.x --> >=2.0.0 <2.1.0
// ~1.2, ~1.2.x, ~>1.2, ~>1.2.x --> >=1.2.0 <1.3.0
// ~1.2.3, ~>1.2.3 --> >=1.2.3 <1.3.0
// ~1.2.0, ~>1.2.0 --> >=1.2.0 <1.3.0
function replaceTildes (comp, options) {
  return comp.trim().split(/\s+/).map(function (comp) {
    return replaceTilde(comp, options)
  }).join(' ')
}

function replaceTilde (comp, options) {
  var r = options.loose ? re[TILDELOOSE] : re[TILDE]
  return comp.replace(r, function (_, M, m, p, pr) {
    debug('tilde', comp, _, M, m, p, pr)
    var ret

    if (isX(M)) {
      ret = ''
    } else if (isX(m)) {
      ret = '>=' + M + '.0.0 <' + (+M + 1) + '.0.0'
    } else if (isX(p)) {
      // ~1.2 == >=1.2.0 <1.3.0
      ret = '>=' + M + '.' + m + '.0 <' + M + '.' + (+m + 1) + '.0'
    } else if (pr) {
      debug('replaceTilde pr', pr)
      ret = '>=' + M + '.' + m + '.' + p + '-' + pr +
            ' <' + M + '.' + (+m + 1) + '.0'
    } else {
      // ~1.2.3 == >=1.2.3 <1.3.0
      ret = '>=' + M + '.' + m + '.' + p +
            ' <' + M + '.' + (+m + 1) + '.0'
    }

    debug('tilde return', ret)
    return ret
  })
}

// ^ --> * (any, kinda silly)
// ^2, ^2.x, ^2.x.x --> >=2.0.0 <3.0.0
// ^2.0, ^2.0.x --> >=2.0.0 <3.0.0
// ^1.2, ^1.2.x --> >=1.2.0 <2.0.0
// ^1.2.3 --> >=1.2.3 <2.0.0
// ^1.2.0 --> >=1.2.0 <2.0.0
function replaceCarets (comp, options) {
  return comp.trim().split(/\s+/).map(function (comp) {
    return replaceCaret(comp, options)
  }).join(' ')
}

function replaceCaret (comp, options) {
  debug('caret', comp, options)
  var r = options.loose ? re[CARETLOOSE] : re[CARET]
  return comp.replace(r, function (_, M, m, p, pr) {
    debug('caret', comp, _, M, m, p, pr)
    var ret

    if (isX(M)) {
      ret = ''
    } else if (isX(m)) {
      ret = '>=' + M + '.0.0 <' + (+M + 1) + '.0.0'
    } else if (isX(p)) {
      if (M === '0') {
        ret = '>=' + M + '.' + m + '.0 <' + M + '.' + (+m + 1) + '.0'
      } else {
        ret = '>=' + M + '.' + m + '.0 <' + (+M + 1) + '.0.0'
      }
    } else if (pr) {
      debug('replaceCaret pr', pr)
      if (M === '0') {
        if (m === '0') {
          ret = '>=' + M + '.' + m + '.' + p + '-' + pr +
                ' <' + M + '.' + m + '.' + (+p + 1)
        } else {
          ret = '>=' + M + '.' + m + '.' + p + '-' + pr +
                ' <' + M + '.' + (+m + 1) + '.0'
        }
      } else {
        ret = '>=' + M + '.' + m + '.' + p + '-' + pr +
              ' <' + (+M + 1) + '.0.0'
      }
    } else {
      debug('no pr')
      if (M === '0') {
        if (m === '0') {
          ret = '>=' + M + '.' + m + '.' + p +
                ' <' + M + '.' + m + '.' + (+p + 1)
        } else {
          ret = '>=' + M + '.' + m + '.' + p +
                ' <' + M + '.' + (+m + 1) + '.0'
        }
      } else {
        ret = '>=' + M + '.' + m + '.' + p +
              ' <' + (+M + 1) + '.0.0'
      }
    }

    debug('caret return', ret)
    return ret
  })
}

function replaceXRanges (comp, options) {
  debug('replaceXRanges', comp, options)
  return comp.split(/\s+/).map(function (comp) {
    return replaceXRange(comp, options)
  }).join(' ')
}

function replaceXRange (comp, options) {
  comp = comp.trim()
  var r = options.loose ? re[XRANGELOOSE] : re[XRANGE]
  return comp.replace(r, function (ret, gtlt, M, m, p, pr) {
    debug('xRange', comp, ret, gtlt, M, m, p, pr)
    var xM = isX(M)
    var xm = xM || isX(m)
    var xp = xm || isX(p)
    var anyX = xp

    if (gtlt === '=' && anyX) {
      gtlt = ''
    }

    if (xM) {
      if (gtlt === '>' || gtlt === '<') {
        // nothing is allowed
        ret = '<0.0.0'
      } else {
        // nothing is forbidden
        ret = '*'
      }
    } else if (gtlt && anyX) {
      // we know patch is an x, because we have any x at all.
      // replace X with 0
      if (xm) {
        m = 0
      }
      p = 0

      if (gtlt === '>') {
        // >1 => >=2.0.0
        // >1.2 => >=1.3.0
        // >1.2.3 => >= 1.2.4
        gtlt = '>='
        if (xm) {
          M = +M + 1
          m = 0
          p = 0
        } else {
          m = +m + 1
          p = 0
        }
      } else if (gtlt === '<=') {
        // <=0.7.x is actually <0.8.0, since any 0.7.x should
        // pass.  Similarly, <=7.x is actually <8.0.0, etc.
        gtlt = '<'
        if (xm) {
          M = +M + 1
        } else {
          m = +m + 1
        }
      }

      ret = gtlt + M + '.' + m + '.' + p
    } else if (xm) {
      ret = '>=' + M + '.0.0 <' + (+M + 1) + '.0.0'
    } else if (xp) {
      ret = '>=' + M + '.' + m + '.0 <' + M + '.' + (+m + 1) + '.0'
    }

    debug('xRange return', ret)

    return ret
  })
}

// Because * is AND-ed with everything else in the comparator,
// and '' means "any version", just remove the *s entirely.
function replaceStars (comp, options) {
  debug('replaceStars', comp, options)
  // Looseness is ignored here.  star is always as loose as it gets!
  return comp.trim().replace(re[STAR], '')
}

// This function is passed to string.replace(re[HYPHENRANGE])
// M, m, patch, prerelease, build
// 1.2 - 3.4.5 => >=1.2.0 <=3.4.5
// 1.2.3 - 3.4 => >=1.2.0 <3.5.0 Any 3.4.x will do
// 1.2 - 3.4 => >=1.2.0 <3.5.0
function hyphenReplace ($0,
  from, fM, fm, fp, fpr, fb,
  to, tM, tm, tp, tpr, tb) {
  if (isX(fM)) {
    from = ''
  } else if (isX(fm)) {
    from = '>=' + fM + '.0.0'
  } else if (isX(fp)) {
    from = '>=' + fM + '.' + fm + '.0'
  } else {
    from = '>=' + from
  }

  if (isX(tM)) {
    to = ''
  } else if (isX(tm)) {
    to = '<' + (+tM + 1) + '.0.0'
  } else if (isX(tp)) {
    to = '<' + tM + '.' + (+tm + 1) + '.0'
  } else if (tpr) {
    to = '<=' + tM + '.' + tm + '.' + tp + '-' + tpr
  } else {
    to = '<=' + to
  }

  return (from + ' ' + to).trim()
}

// if ANY of the sets match ALL of its comparators, then pass
Range.prototype.test = function (version) {
  if (!version) {
    return false
  }

  if (typeof version === 'string') {
    version = new SemVer(version, this.options)
  }

  for (var i = 0; i < this.set.length; i++) {
    if (testSet(this.set[i], version, this.options)) {
      return true
    }
  }
  return false
}

function testSet (set, version, options) {
  for (var i = 0; i < set.length; i++) {
    if (!set[i].test(version)) {
      return false
    }
  }

  if (version.prerelease.length && !options.includePrerelease) {
    // Find the set of versions that are allowed to have prereleases
    // For example, ^1.2.3-pr.1 desugars to >=1.2.3-pr.1 <2.0.0
    // That should allow `1.2.3-pr.2` to pass.
    // However, `1.2.4-alpha.notready` should NOT be allowed,
    // even though it's within the range set by the comparators.
    for (i = 0; i < set.length; i++) {
      debug(set[i].semver)
      if (set[i].semver === ANY) {
        continue
      }

      if (set[i].semver.prerelease.length > 0) {
        var allowed = set[i].semver
        if (allowed.major === version.major &&
            allowed.minor === version.minor &&
            allowed.patch === version.patch) {
          return true
        }
      }
    }

    // Version has a -pre, but it's not one of the ones we like.
    return false
  }

  return true
}

exports.satisfies = satisfies
function satisfies (version, range, options) {
  try {
    range = new Range(range, options)
  } catch (er) {
    return false
  }
  return range.test(version)
}

exports.maxSatisfying = maxSatisfying
function maxSatisfying (versions, range, options) {
  var max = null
  var maxSV = null
  try {
    var rangeObj = new Range(range, options)
  } catch (er) {
    return null
  }
  versions.forEach(function (v) {
    if (rangeObj.test(v)) {
      // satisfies(v, range, options)
      if (!max || maxSV.compare(v) === -1) {
        // compare(max, v, true)
        max = v
        maxSV = new SemVer(max, options)
      }
    }
  })
  return max
}

exports.minSatisfying = minSatisfying
function minSatisfying (versions, range, options) {
  var min = null
  var minSV = null
  try {
    var rangeObj = new Range(range, options)
  } catch (er) {
    return null
  }
  versions.forEach(function (v) {
    if (rangeObj.test(v)) {
      // satisfies(v, range, options)
      if (!min || minSV.compare(v) === 1) {
        // compare(min, v, true)
        min = v
        minSV = new SemVer(min, options)
      }
    }
  })
  return min
}

exports.minVersion = minVersion
function minVersion (range, loose) {
  range = new Range(range, loose)

  var minver = new SemVer('0.0.0')
  if (range.test(minver)) {
    return minver
  }

  minver = new SemVer('0.0.0-0')
  if (range.test(minver)) {
    return minver
  }

  minver = null
  for (var i = 0; i < range.set.length; ++i) {
    var comparators = range.set[i]

    comparators.forEach(function (comparator) {
      // Clone to avoid manipulating the comparator's semver object.
      var compver = new SemVer(comparator.semver.version)
      switch (comparator.operator) {
        case '>':
          if (compver.prerelease.length === 0) {
            compver.patch++
          } else {
            compver.prerelease.push(0)
          }
          compver.raw = compver.format()
          /* fallthrough */
        case '':
        case '>=':
          if (!minver || gt(minver, compver)) {
            minver = compver
          }
          break
        case '<':
        case '<=':
          /* Ignore maximum versions */
          break
        /* istanbul ignore next */
        default:
          throw new Error('Unexpected operation: ' + comparator.operator)
      }
    })
  }

  if (minver && range.test(minver)) {
    return minver
  }

  return null
}

exports.validRange = validRange
function validRange (range, options) {
  try {
    // Return '*' instead of '' so that truthiness works.
    // This will throw if it's invalid anyway
    return new Range(range, options).range || '*'
  } catch (er) {
    return null
  }
}

// Determine if version is less than all the versions possible in the range
exports.ltr = ltr
function ltr (version, range, options) {
  return outside(version, range, '<', options)
}

// Determine if version is greater than all the versions possible in the range.
exports.gtr = gtr
function gtr (version, range, options) {
  return outside(version, range, '>', options)
}

exports.outside = outside
function outside (version, range, hilo, options) {
  version = new SemVer(version, options)
  range = new Range(range, options)

  var gtfn, ltefn, ltfn, comp, ecomp
  switch (hilo) {
    case '>':
      gtfn = gt
      ltefn = lte
      ltfn = lt
      comp = '>'
      ecomp = '>='
      break
    case '<':
      gtfn = lt
      ltefn = gte
      ltfn = gt
      comp = '<'
      ecomp = '<='
      break
    default:
      throw new TypeError('Must provide a hilo val of "<" or ">"')
  }

  // If it satisifes the range it is not outside
  if (satisfies(version, range, options)) {
    return false
  }

  // From now on, variable terms are as if we're in "gtr" mode.
  // but note that everything is flipped for the "ltr" function.

  for (var i = 0; i < range.set.length; ++i) {
    var comparators = range.set[i]

    var high = null
    var low = null

    comparators.forEach(function (comparator) {
      if (comparator.semver === ANY) {
        comparator = new Comparator('>=0.0.0')
      }
      high = high || comparator
      low = low || comparator
      if (gtfn(comparator.semver, high.semver, options)) {
        high = comparator
      } else if (ltfn(comparator.semver, low.semver, options)) {
        low = comparator
      }
    })

    // If the edge version comparator has a operator then our version
    // isn't outside it
    if (high.operator === comp || high.operator === ecomp) {
      return false
    }

    // If the lowest version comparator has an operator and our version
    // is less than it then it isn't higher than the range
    if ((!low.operator || low.operator === comp) &&
        ltefn(version, low.semver)) {
      return false
    } else if (low.operator === ecomp && ltfn(version, low.semver)) {
      return false
    }
  }
  return true
}

exports.prerelease = prerelease
function prerelease (version, options) {
  var parsed = parse(version, options)
  return (parsed && parsed.prerelease.length) ? parsed.prerelease : null
}

exports.intersects = intersects
function intersects (r1, r2, options) {
  r1 = new Range(r1, options)
  r2 = new Range(r2, options)
  return r1.intersects(r2)
}

exports.coerce = coerce
function coerce (version) {
  if (version instanceof SemVer) {
    return version
  }

  if (typeof version !== 'string') {
    return null
  }

  var match = version.match(re[COERCE])

  if (match == null) {
    return null
  }

  return parse(match[1] +
    '.' + (match[2] || '0') +
    '.' + (match[3] || '0'))
}


/***/ }),

/***/ "./node_modules/uuid-js/lib/uuid.js":
/*!******************************************!*\
  !*** ./node_modules/uuid-js/lib/uuid.js ***!
  \******************************************/
/***/ ((module) => {

/*
 * UUID-js: A js library to generate and parse UUIDs, TimeUUIDs and generate
 * TimeUUID based on dates for range selections.
 * @see http://www.ietf.org/rfc/rfc4122.txt
 **/

function UUIDjs() {
};

UUIDjs.maxFromBits = function(bits) {
  return Math.pow(2, bits);
};

UUIDjs.limitUI04 = UUIDjs.maxFromBits(4);
UUIDjs.limitUI06 = UUIDjs.maxFromBits(6);
UUIDjs.limitUI08 = UUIDjs.maxFromBits(8);
UUIDjs.limitUI12 = UUIDjs.maxFromBits(12);
UUIDjs.limitUI14 = UUIDjs.maxFromBits(14);
UUIDjs.limitUI16 = UUIDjs.maxFromBits(16);
UUIDjs.limitUI32 = UUIDjs.maxFromBits(32);
UUIDjs.limitUI40 = UUIDjs.maxFromBits(40);
UUIDjs.limitUI48 = UUIDjs.maxFromBits(48);

// Returns a random integer between min and max
// Using Math.round() will give you a non-uniform distribution!
// @see https://developer.mozilla.org/en-US/docs/JavaScript/Reference/Global_Objects/Math/random
function getRandomInt(min, max) {
  return Math.floor(Math.random() * (max - min + 1)) + min;
}

UUIDjs.randomUI04 = function() {
  return getRandomInt(0, UUIDjs.limitUI04-1);
};
UUIDjs.randomUI06 = function() {
  return getRandomInt(0, UUIDjs.limitUI06-1);
};
UUIDjs.randomUI08 = function() {
  return getRandomInt(0, UUIDjs.limitUI08-1);
};
UUIDjs.randomUI12 = function() {
  return getRandomInt(0, UUIDjs.limitUI12-1);
};
UUIDjs.randomUI14 = function() {
  return getRandomInt(0, UUIDjs.limitUI14-1);
};
UUIDjs.randomUI16 = function() {
  return getRandomInt(0, UUIDjs.limitUI16-1);
};
UUIDjs.randomUI32 = function() {
  return getRandomInt(0, UUIDjs.limitUI32-1);
};
UUIDjs.randomUI40 = function() {
  return (0 | Math.random() * (1 << 30)) + (0 | Math.random() * (1 << 40 - 30)) * (1 << 30);
};
UUIDjs.randomUI48 = function() {
  return (0 | Math.random() * (1 << 30)) + (0 | Math.random() * (1 << 48 - 30)) * (1 << 30);
};

UUIDjs.paddedString = function(string, length, z) {
  string = String(string);
  z = (!z) ? '0' : z;
  var i = length - string.length;
  for (; i > 0; i >>>= 1, z += z) {
    if (i & 1) {
      string = z + string;
    }
  }
  return string;
};

UUIDjs.prototype.fromParts = function(timeLow, timeMid, timeHiAndVersion, clockSeqHiAndReserved, clockSeqLow, node) {
  this.version = (timeHiAndVersion >> 12) & 0xF;
  this.hex = UUIDjs.paddedString(timeLow.toString(16), 8)
             + '-'
             + UUIDjs.paddedString(timeMid.toString(16), 4)
             + '-'
             + UUIDjs.paddedString(timeHiAndVersion.toString(16), 4)
             + '-'
             + UUIDjs.paddedString(clockSeqHiAndReserved.toString(16), 2)
             + UUIDjs.paddedString(clockSeqLow.toString(16), 2)
             + '-'
             + UUIDjs.paddedString(node.toString(16), 12);
  return this;
};

UUIDjs.prototype.toString = function() {
  return this.hex;
};
UUIDjs.prototype.toURN = function() {
  return 'urn:uuid:' + this.hex;
};

UUIDjs.prototype.toBytes = function() {
  var parts = this.hex.split('-');
  var ints = [];
  var intPos = 0;
  for (var i = 0; i < parts.length; i++) {
    for (var j = 0; j < parts[i].length; j+=2) {
      ints[intPos++] = parseInt(parts[i].substr(j, 2), 16);
    }
  }
  return ints;
};

UUIDjs.prototype.equals = function(uuid) {
  if (!(uuid instanceof UUID)) {
    return false;
  }
  if (this.hex !== uuid.hex) {
    return false;
  }
  return true;
};

UUIDjs.getTimeFieldValues = function(time) {
  var ts = time - Date.UTC(1582, 9, 15);
  var hm = ((ts / 0x100000000) * 10000) & 0xFFFFFFF;
  return { low: ((ts & 0xFFFFFFF) * 10000) % 0x100000000,
            mid: hm & 0xFFFF, hi: hm >>> 16, timestamp: ts };
};

UUIDjs._create4 = function() {
  return new UUIDjs().fromParts(
    UUIDjs.randomUI32(),
    UUIDjs.randomUI16(),
    0x4000 | UUIDjs.randomUI12(),
    0x80   | UUIDjs.randomUI06(),
    UUIDjs.randomUI08(),
    UUIDjs.randomUI48()
  );
};

UUIDjs._create1 = function() {
  var now = new Date().getTime();
  var sequence = UUIDjs.randomUI14();
  var node = (UUIDjs.randomUI08() | 1) * 0x10000000000 + UUIDjs.randomUI40();
  var tick = UUIDjs.randomUI04();
  var timestamp = 0;
  var timestampRatio = 1/4;

  if (now != timestamp) {
    if (now < timestamp) {
      sequence++;
    }
    timestamp = now;
    tick = UUIDjs.randomUI04();
  } else if (Math.random() < timestampRatio && tick < 9984) {
    tick += 1 + UUIDjs.randomUI04();
  } else {
    sequence++;
  }

  var tf = UUIDjs.getTimeFieldValues(timestamp);
  var tl = tf.low + tick;
  var thav = (tf.hi & 0xFFF) | 0x1000;

  sequence &= 0x3FFF;
  var cshar = (sequence >>> 8) | 0x80;
  var csl = sequence & 0xFF;

  return new UUIDjs().fromParts(tl, tf.mid, thav, cshar, csl, node);
};

UUIDjs.create = function(version) {
  version = version || 4;
  return this['_create' + version]();
};

UUIDjs.fromTime = function(time, last) {
  last = (!last) ? false : last;
  var tf = UUIDjs.getTimeFieldValues(time);
  var tl = tf.low;
  var thav = (tf.hi & 0xFFF) | 0x1000;  // set version '0001'
  if (last === false) {
    return new UUIDjs().fromParts(tl, tf.mid, thav, 0, 0, 0);
  } else {
    return new UUIDjs().fromParts(tl, tf.mid, thav, 0x80 | UUIDjs.limitUI06, UUIDjs.limitUI08 - 1, UUIDjs.limitUI48 - 1);
  }
};

UUIDjs.firstFromTime = function(time) {
  return UUIDjs.fromTime(time, false);
};
UUIDjs.lastFromTime = function(time) {
  return UUIDjs.fromTime(time, true);
};

UUIDjs.fromURN = function(strId) {
  var r, p = /^(?:urn:uuid:|\{)?([0-9a-f]{8})-([0-9a-f]{4})-([0-9a-f]{4})-([0-9a-f]{2})([0-9a-f]{2})-([0-9a-f]{12})(?:\})?$/i;
  if ((r = p.exec(strId))) {
    return new UUIDjs().fromParts(parseInt(r[1], 16), parseInt(r[2], 16),
                            parseInt(r[3], 16), parseInt(r[4], 16),
                            parseInt(r[5], 16), parseInt(r[6], 16));
  }
  return null;
};

UUIDjs.fromBytes = function(ints) {
  if (ints.length < 5) {
    return null;
  }
  var str = '';
  var pos = 0;
  var parts = [4, 2, 2, 2, 6];
  for (var i = 0; i < parts.length; i++) {
    for (var j = 0; j < parts[i]; j++) {
      var octet = ints[pos++].toString(16);
      if (octet.length == 1) {
        octet = '0' + octet;
      }
      str += octet;
    }
    if (parts[i] !== 6) {
      str += '-';
    }
  }
  return UUIDjs.fromURN(str);
};

UUIDjs.fromBinary = function(binary) {
  var ints = [];
  for (var i = 0; i < binary.length; i++) {
    ints[i] = binary.charCodeAt(i);
    if (ints[i] > 255 || ints[i] < 0) {
      throw new Error('Unexpected byte in binary data.');
    }
  }
  return UUIDjs.fromBytes(ints);
};

// Aliases to support legacy code. Do not use these when writing new code as
// they may be removed in future versions!
UUIDjs.new = function() {
  return this.create(4);
};
UUIDjs.newTS = function() {
  return this.create(1);
};

module.exports = UUIDjs;


/***/ })

/******/ 	});
/************************************************************************/
/******/ 	// The module cache
/******/ 	var __webpack_module_cache__ = {};
/******/ 	
/******/ 	// The require function
/******/ 	function __webpack_require__(moduleId) {
/******/ 		// Check if module is in cache
/******/ 		var cachedModule = __webpack_module_cache__[moduleId];
/******/ 		if (cachedModule !== undefined) {
/******/ 			return cachedModule.exports;
/******/ 		}
/******/ 		// Create a new module (and put it into the cache)
/******/ 		var module = __webpack_module_cache__[moduleId] = {
/******/ 			id: moduleId,
/******/ 			loaded: false,
/******/ 			exports: {}
/******/ 		};
/******/ 	
/******/ 		// Execute the module function
/******/ 		__webpack_modules__[moduleId].call(module.exports, module, module.exports, __webpack_require__);
/******/ 	
/******/ 		// Flag the module as loaded
/******/ 		module.loaded = true;
/******/ 	
/******/ 		// Return the exports of the module
/******/ 		return module.exports;
/******/ 	}
/******/ 	
/************************************************************************/
/******/ 	/* webpack/runtime/global */
/******/ 	(() => {
/******/ 		__webpack_require__.g = (function() {
/******/ 			if (typeof globalThis === 'object') return globalThis;
/******/ 			try {
/******/ 				return this || new Function('return this')();
/******/ 			} catch (e) {
/******/ 				if (typeof window === 'object') return window;
/******/ 			}
/******/ 		})();
/******/ 	})();
/******/ 	
/******/ 	/* webpack/runtime/node module decorator */
/******/ 	(() => {
/******/ 		__webpack_require__.nmd = (module) => {
/******/ 			module.paths = [];
/******/ 			if (!module.children) module.children = [];
/******/ 			return module;
/******/ 		};
/******/ 	})();
/******/ 	
/************************************************************************/
var __webpack_exports__ = {};
// This entry need to be wrapped in an IIFE because it need to be in strict mode.
(() => {
"use strict";
var exports = __webpack_exports__;
/*!***********************************!*\
  !*** ./js/macos/exporter/main.js ***!
  \***********************************/


Object.defineProperty(exports, "__esModule", ({
  value: true
}));
Object.defineProperty(exports, "Guest", ({
  enumerable: true,
  get: function () {
    return _native.Guest;
  }
}));
exports.startMacOsExporter = startMacOsExporter;

var _exporter = __webpack_require__(/*! macos/exporter/exporter */ "./js/macos/exporter/exporter.js");

var _native = __webpack_require__(/*! core/native */ "./js/core/native.js");

var exporterApp = null;

function startMacOsExporter(options) {
  exporterApp = new _exporter.ExporterAppDesktop(options);
  exporterApp.start();
}
})();

MacOSExporter = __webpack_exports__;
/******/ })()
;
//# sourceMappingURL=exporter-macos.js.map